#! /usr/bin/env python3
# Julian to Greenwich and vice versa
# with my own hvfTime class, works also for BC dates and leap years
# hvf 16.09.2021 26.12.2021 11.11.2022 13.04.2023
# make hvfTime() hashable -- hvf 02.10.2023

from datetime import datetime, timedelta
from math import floor
import sys
import re
import dateutil.parser

VersionJulian = "V1.8 hvf 2023-10-02 HVFTime.py based on Meeus"


# JulianDay from year, month, day -  hvf 16.11.2021
def getJulianDayfromYMD(year, month, day, hour=0, minute=0, second=0, 
                        verbose = False):
    if verbose: 
        print ("getJulianDayfromYMD: ", year, month, day, hour, minute, second)
    day  += hour/24. + \
            minute/1440. + \
            second/86400.
    isGregor = False    # 04.10.1582 is last day before Gregorian calendar
    if year > 1582 or year == 1582 and (month > 10 or month == 10 and day > 4):
        isGregor = True
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    # JD 0 is -4712-01-01.5 i.e. -4712-01-01 12h
    # N.B. JD 1524.5 is -4708-03-05 00h
    if verbose: print ("JD = ", JD)
    return JD


def roundhvfTime(hvfDT = None, roundTo = 3600):
    # Round a hvfTime object to anytime lapse in seconds
    # hvfDT: hvftime object, default now
    # roundTo: closest number of seconds to round to, default 1 hour.
    # Author: Harald von Fellenberg 2021
    # Problem: 2021-11-21 1223 returns 2021-11-21 12:22:60.000
    # hvf 2022-09-09
    verbose=False
    if not isinstance(hvfDT, hvfTime): hvfDT = hvfTime.now()
    h, m, s = hvfDT.getTime()
    seconds = h*3600 + m*60 + s
    fraction = seconds - floor(seconds)     # fractional seconds
    if verbose:
        print ("seconds, fraction, floor, rounding = ", seconds, fraction, floor(seconds))
    seconds = floor(seconds)
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    if verbose: print (rounding)
    return hvfDT + timedelta(0, rounding-seconds, -fraction*1e6)

# returns yyyymmdd hh for a hvfTime object, rounded
def nicehvfTimehh (hvfDT):
    hvfDTrounded = roundhvfTime(hvfDT)
    yy, mm, dd = hvfDTrounded.getDate()
    hh, dummy, dummy = hvfDTrounded.getTime()

    s = ''
    if yy >= 0:
        s = " %04d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    else:
        s = "%05d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    return s

#######################################################################
# replaces astropy.time.Time since that does not support leap days 
# before 100BC
# values are represented in Julian days
# creator parameter can be:
# - a hvfTime object
# - a JD JulianDay object (float)
# - a datetime object
# - date and time as six values yy, mm, dd, [hh, MM, ss]
#   dd and ss may have fractions
# hvf 26.12.2021

class hvfTime:
    def __init__(self, *arg):
        # print ("__init__: arg = ", arg)
        # argument is a hvfTime object
        if isinstance(arg[0], hvfTime):
            JD = arg[0].getJD()
            self.__JD = JD
        # argument is a JD
        elif isinstance(arg[0], float) and len(arg) == 1:
            JD = arg[0]
            self.__JD = JD
        # argument is a datetime object
        elif isinstance(arg[0], datetime):
            dt = arg[0]
            yy = dt.year
            mm = dt.month
            dd = dt.day
            hh = dt.hour
            MM = dt.minute
            ss = dt.second + dt.microsecond*1e-6
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        # argument is yy mm dd [hh MM ss]
        elif (isinstance(arg[0], int) or isinstance(arg[0], float)) and \
            len(arg) >= 3:
            hh = 0; MM = 0; ss = 0
            yy = arg[0]; mm = arg[1]; dd = arg[2]
            if len(arg) > 3: hh = arg[3]
            if len(arg) > 4: MM = arg[4]
            if len(arg) > 5: ss = arg[5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD

        # argument is (yy mm dd [hh MM ss])
        elif isinstance(arg[0], tuple) or isinstance(arg[0], list):
            hh = 0; MM = 0; ss = 0
            yy = arg[0][0]; mm = arg[0][1]; dd = arg[0][2]
            if len(arg[0]) > 3: hh = arg[0][3]
            if len(arg[0]) > 4: MM = arg[0][4]
            if len(arg[0]) > 5: ss = arg[0][5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        else:
            print ("hvfTime instance not supported: ", arg[0])
            sys.exit(1)
            
    def getJD(self):
        return self.__JD

    def getDate(self, verbose = False):  # returns (yy, mm, dd), 
                                        # where dd may have fraction
        # works also for negative JDs! hvf 29.12.2021
        JD = self.__JD
        jd1 = JD + 0.5
        I = floor(jd1)      # integer part, may be negative
        F = jd1 - I         # fractional part, always positive
        if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
            A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
            B = I + A - int(A/4) + 1
        else:   B = I       # Julian Calendar for earlier dates
        C = B + 1524
        D = floor((C-122.1)/365.25)
        E = floor(365.25 * D)
        G = floor((C-E)/30.6001)
        d = C - E + F - floor(30.6001 * G)
        if G < 13.5:    m = G - 1
        else:           m = G - 13
        if m > 2.5:     y = D - 4716
        else:           y = D - 4715

        if verbose:
            print ("getDate: JD, jd1, I, F = ", JD, jd1, I, F)
            print ("getDate: B, C, D, E, G = ", B, C, D, E, G)
            print ("getDate: y, m, d       = ", y, m, d)
        return (y, m, d)

    def getTime(self):      # seconds have 3 digi fractions (millisec)
        dummy, dummy, t = self.getDate()
        t -= floor(t)
        # t is fraction of day - convert to sec and round
        daysec = round(t * 86400., 3)
        hms = daysec / 3600.
        #hms = t*24.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 3)
        #s = int((ms - m) * 60)
        #print ("daysec, hms, h, ms, m, s = ", daysec, hms, h, ms, m, s)
    
        return (h, m, s)

    def __str__(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, hh, mm, ss)

    def getStr(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        daysec = round(ss + mm*60 + hh*3600, 0)
        hms = daysec / 3600.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 0)

        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, h, m, s)

    # Ernesti, Python3, p. 380
    # hvfTime = hvfTime + timedelta
    # hvfTime = hvfTime + day   (float or int)
    def __add__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__add__: wrong type ", type(td))
            sys.exit(1)
        return hvfTime(self.__JD + sec/86400.)

    # hvfTime += timedelta
    def __iadd__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__iadd__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD += sec/86400.
        return self

    # hvfTime = hvfTime - timedelta
    # and   n = hvfTime - hvfTime       n in days
    def __sub__(self, td):
        # subtract a timedelta object 
        if isinstance (td, timedelta):
            sec = td.total_seconds()
            return hvfTime(self.__JD - sec/86400.)
        elif isinstance(td, hvfTime):
            # return JDs in days as difference between two hvfTime objects
            return self.__JD - td.getJD()
        else:
            print ("hvfTime.__sub__: wrong type ", type(td), td)
            sys.exit(1)

    # hvfTime -= timedelta
    def __isub__(self, td):
        # subtract a timedelta object 
        try:    
            sec = td.total_seconds()
        except:
            print ("hvfTime.__isub__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD -= sec/86400.
        return self

    # hvfTime < hvfTime
    def __lt__(self, other):
        return self.__JD < other.getJD()

    # hvfTime <= hvfTime
    def __le__(self, other):
        return self.__JD <= other.getJD()

    # hvfTime == hvfTime
    def __eq__(self, other):
        return self.__JD == other.getJD()

    # hash - makes hvfTime hashable, i.e. as key in dict or set 
    def __hash__(self):
        return hash(self.__JD)

    # hvfTime != hvfTime
    def __ne__(self, other):
        return self.__JD != other.getJD()

    # hvfTime > hvfTime
    def __gt__(self, other):
        return self.__JD > other.getJD()

    # hvfTime >= hvfTime
    def __ge__(self, other):
        return self.__JD >= other.getJD()

    def now():  # returns current time (h, m, s), uses datetime.now()
                # roundedd to integer seconds
        return hvfTime(datetime.now()).getTime()

    def today():    # returns current date (y, m, d), uses datetime.today()
                    # d has fractions
        return hvfTime(datetime.today()).getDate()
    
    def getDatetime(self):
        (yy, mm, dd) = self.getDate()
        dd = floor(dd)
        (HH, MM, SS) = self.getTime()
        SS = int(round(SS, 0))
        return datetime(yy, mm, dd, HH, MM, SS)
#######################################################################

# parse date time
# returns a hvfTime object
# hvf 19.11.2021 25.12.2021
# allowed patterns:
# 2021-11-21
# 2021-11-21 12:23:45.635
# 2021-11-21T12:23:45.635
# 2021-11-21 122345.635
# 2021-11-21 122345
# 2021-11-21 1223
# 2021-11-21 12
# -4712-1-1.5 
# 25.12.2021 12:34:56.789
# 1.1.2022
# 2451545.0 (a Julian Day)
# works for dates between -9999-01-01 and 9999-12-31

def parseDateTime(s, verbose = False):
    HH = MM = SS = 0
    # check if two parts separated by blank or T
    if verbose: print ("parseDateTime input: ", s)
    s = s.strip()
    dpart = s
    tpart = None
    parts = re.split (r"[\sT]", s)
    if len(parts) >= 2:
        dpart = parts[0]    # date part
        tpart = parts[1]    # time part
        if verbose:
            print ("date part = ", dpart)
            print ("time part = ", tpart)

    # parse date  2021-12-24 or  4712-1-1.5
    if re.fullmatch(r'\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        ymd = re.split(r'-', dpart)
        yy = ymd[0]
        mm = ymd[1]
        dd = ymd[2]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date 24.12.2021 or 1.1.2022
    elif re.fullmatch(r'\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        ymd = re.split(r'\.', dpart)
        yy = ymd[2]
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -4712-01-01.5
    elif re.fullmatch(r'-\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        symd = re.split(r'-', dpart)
        yy = -int(symd[1])
        mm = symd[2]
        dd = symd[3]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -24.12.2021 or -1.1.2022
    elif re.fullmatch(r'-\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        symd = re.split(r'-', dpart)
        ymd = re.split(r'\.', symd[1])
        yy = -int(ymd[2])
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    else:
        print ("parseDateTime format unknown: ", s)
    # parse date  2021-12-24 or  4712-1-1.5
    # parse date 24.12.2021 or 1.1.2022
    # parse date -4712-01-01.5
    # parse date -24.12.2021 or -1.1.2022
        print ("Allowed date formats include 2021-12-24, 4712-1-1.5,")
        print ("    24.12.2021, 1.1.2022, -4712-01-01.5, -24.12.2021, -1.1.2022")
        print ("Allowed range of years is from -9999 to 9999")
        sys.exit(1)

    # parse time using dateutil.parser.isoparse - this is easier
    # than checking all the different possible input formats
    if tpart:
        deltaT = dateutil.parser.isoparse("2021-12-24 " + tpart) \
                - datetime(2021, 12, 24)
        Secs = round(deltaT.total_seconds(), 3)    # avoid rounding error
        HH = Secs // 3600
        Secs -= HH * 3600
        MM = Secs // 60
        Secs -= MM * 60
        SS = Secs
        if verbose:
            print ("HH, MM, SS = ", HH, MM, SS)

    #return hvfTime object
    return hvfTime(int(yy), int(mm), float(dd), int(HH), int(MM), float(SS))

#######################################################################
# catculate weekday for any date from -9999-01-01 to 9999-12-31
# Meeus 1991 p. 65

def getDayofWeek(dt):
    name =["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday"]
    hvfdt = hvfTime(dt)
    (yy, mm, dd) = hvfdt.getDate()
    hvfdtMidnight = hvfTime(yy, mm, int(dd), 0, 0, 0.0)
    JD0 = hvfdtMidnight.getJD()
    weekday = int(JD0 + 1.51) % 7
    return name[weekday]

#######################################################################

#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-

# Der folgende Python-Code fuer Sonne und Mond 
# basiert auf der Portierung des java-Scripts von Arnold Barmettler
# http://lexikon.astronomie.info/java/sunmoon/sunmoon.html
#
#     Entfernen Sie folgende Informationen auf keinen Fall: / 
#     Do not remove following text:
#     Source code by Arnold Barmettler, www.astronomie.info / www.CalSky.com
#     and Harald von Fellenberg, www.hvf.ch
#     based on algorithms by Peter Duffett-Smith's great and easy book
#     'Practical Astronomy with your Calculator or Spreadsheet'.
#     and Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Totale Ueberarbeitung durch Harald von Fellenberg, Dezember 2020
#   Portierung auf Python 3
#   Ergaenzung fuer Mitternachtssonne, Polarnacht
#   Ergaenzungen fuer Daemmerung, Hoehe der Sonne beim Transit tags und nachts
#   Ergaenzung mit Zeitgleichung, wahre und mittlere Ortszeit
#   Azimut für Sonnenaufgang, Sonnenuntergang
#   Liste einiger Städte mit ihren Koordinaten
#   Korrekte Berechnung in allen Zeitzonen
#   Behandlung aller Planeten
#
#   Die Behandlung von Sonne und Mond wurde ersetzt durch die Algorithmen in
#   Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Die Daten aller Planeten beruhen auf VSPO87
#   Pierre Bretagnon, Gerard Francou
#   Bureau des Longitudes
#   77, Avenue Denfert-Rochereau
#   F75014, Paris, France
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/
#
#   Die Resultate wurden mit denen auf https://www.timeanddate.com/ verglichen,
#   die Uebereinstimmung ist gut innerhalb von wenigen Minuten.
#   Beim Mond ist stets das Datum zu beachten!
#   Liegt die lokale Zeit zwischen Mitternacht und "zone", so wird für den Mond
#   zum Teil auf den Vortag zurückgegriffen. Dies ist besonders für Orte mit
#   grosser Zeitzonendifferenz wichtig, wie Sydney oder Honolulu.
#   Für Orte in den Polarregionen wird Mitternachtssonne und Polarnacht
#   korrekt ausgegeben, ausser am Tag, wo die Sonne unter- oder aufgeht.
#   Ebenso fuer den Mond.
#
#   Anmerkung: die Webseiten astronomie.info sowie calsky.com existieren
#   seit Oktober 2020 nicht mehr.
#
#   Alle 8 Planeten sind ebenfalls behandelt, gemaess den Algorithmen in
#   Jean Meeus, Astronomical Algorithms.
#   Die Daten zu den 8 Planeten stammen aus den VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Version 1.0
#   Version 2.0
#   Version 3.0
#   Version 3.1
#   Version 3.4
#   Version 4.0 inkl Sun-Events, Moon-Events
#   Version 5.1 
#   Version 5.4 add getTimezone()
Version = "V5.4 hvf 2024-05-08 SunHead.py using VSOP87 ephemerides and HVFTime"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(Version)

import time
from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import argparse
import sys
import dateutil.parser
import os

# alle Konstanten sind auf die Epoche 2010.0, JD=2455196.5 bezogen
# falls nicht anders dokumentiert.

def getEpoch2010():
    return 2455196.5

# https://www.geeksforgeeks.org/how-to-print-to-stderr-and-stdout-in-python/
def print_to_stderr(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stderr) 

def print_to_stdout(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stdout) 

# produces a datetime object output string without microseconds
# seconds get rounded

def nicedatetime(dt):       # yyyy-mm-dd HH:MM:SS
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        string = dt.getStr()
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%Y-%m-%d %H:%M:%S")
    return string

def niceshortdatetime(dt):  # dd.mm.yy HH:MM:SS
    # only meaningful for known century
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        (yy, mm, dd) = dt.getDate()
        yy = int(abs(yy)%100)    # only last 2 ciphers, also for negative yyyy
        (hh, MM, ss) = dt.getTime()
        # round seconds! use %f fmt - hvf 13.04.2023
        #string = "%02d.%02d.%02d %02d:%02d:%02d" % (dd, mm, yy, hh, MM, ss)
        string = "%02d.%02d.%02d %02d:%02d:%02.0f" % (dd, mm, yy, hh, MM, ss)
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%d.%m.%y %H:%M:%S")
    return string

def nicedateshorttime(dt):  # yyyy-mm-dd HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=30)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        (h, m, s) = dt.getTime()
        string = "%4d-%02d-%02d %02d:%02d" % (y, mm, dd, h, m)
    else:
        string = dt.strftime("%Y-%m-%d %H:%M")
    return string

def niceshortdateshorttime(dt):  # dd.mm.yy HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        dtrounded = roundhvfTime(dt, 60)
        #print ("dt, dtrounded = ", dt, dtrounded)
        (y, mm, dd) = dtrounded.getDate()
        yy = abs(y)%100
        dd = floor(dd)
        (h, m, s) = dtrounded.getTime()
        string = "%02d.%02d.%02d %02d:%02d" % (dd, mm, yy, h, m)
    else:
        dt += timedelta(seconds=30)
        string = dt.strftime("%d.%m.%y %H:%M")
    return string

def nicedate(dt):           # yyyy-mm-dd
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=0.5)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        string = "%04d-%02d-%02d" % (y, mm, dd)
    else:
        string = dt.strftime("%Y-%m-%d")
    return string

def nicetime(dt):           # HH:MM:SS
    if isinstance(dt, type(None)): return None
    try:
        if isinstance(dt, hvfTime):
            (h, m, s) = dt.getTime()
            s =round(s, 0)
            string = "%02d:%02d:%02d" % (h, m, s)
        else:
            dt += timedelta(seconds=0.5)
            string = dt.strftime("%H:%M:%S")
    except:
        string = ' '*8
    return string

def nicetimedelta(dt):      # +-HH:MM:SS
    if isinstance(dt, type(None)): return None
    #dt += timedelta(seconds=0.5)
    # treat also -00:00:05 correctly! hvf 16.06.2021
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60

    string = "%1s%2d:%02d:%02d" % (sign, hours, minutes, seconds)
    return string

def nicedeltadayhhmm(dt):   # +-days HH.MM
    if isinstance(dt, type(None)): return None
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    days  = int(totalsecs/86400.); totalsecs -= days*86400.
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60:
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60
    if hours >= 24:
        days += 1; hours -= 24

    string = "%1s%d days %02d:%02d" % (sign, days, hours, minutes)
    return string

def niceminsec(secs):       # +- M:SS  for Time Equation - Zeitgleichung
    if isinstance(secs, type(None)): return None
    sign = ' '
    if secs < 0: sign = '-'; secs = abs(secs)
    minutes = int(secs/60.); secs -= minutes*60.
    seconds = round(secs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60

    string = "%1s%d:%02d" % (sign, minutes, seconds)
    return string

# printZeile(kopf, Liste, format, zeile=79)
# schreibt eine Zeile der max. Laenge zeile mit kopf und den
# Werten aus Liste, formattiert je mit format.
# Padding vor jedem Listenelement
# Die  formattierten Werte muessen gleich lang sein.
# hvf 31.12.2020

def printZeile(kopf, Liste, format, zeile=79):
    rest = zeile
    string = kopf
    rest -= len(string)
    n = len(Liste)      # Anzahl Werte
    test = format % Liste[0]
    testlen = len(test)
    padlen = int((rest - n*testlen)/n)
    padding = ' '*padlen

    for item in Liste:
        string += padding
        string += format % item

    print (string)
    return string

# modulus in Python has always the same sign as b
# Mod(-3, 24) gives 21
def Mod(a, b):
    return (a % b)

# return integer value, closer to 0
def Int(x): return int(x)

# return fractional part, closer to 0 (i.e. may be nagative)
def Frac(x): return (x - int(x))

# Winkelfunktionen fuer Grad Eingaben
def Sin(x):
    return sin(radians(Mod(x, 360.)))

def Cos(x):
    return cos(radians(Mod(x, 360.)))

def Tan(x):
    return tan(radians(Mod(x, 360.)))

def Deg(x):
    return degrees(x)

# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D


#####################################################################
# Quelle: https://de.wikipedia.org/wiki/Julianisches_Datum
# abgerufen am 20.10.2019
# Name geaendert in getJulianDay (siehe Meeus, chapter 7) - hvf 24.03.2023

# Calculate Julian day for all Gregorian and Julian dates
# Greenwich date is supplied as a hvfTime object

def getJulianDay(dtGreenwichDate, isGregor=True ):
    hvfdt = hvfTime(dtGreenwichDate)
    return hvfdt.getJD()

################################################################
# calculate Greenwich date from Julian date
# Ref. Practical Astronomy (see above), p.11
# returns a hvfTime object for midnight 

def getGreenwichDate(JD):

    return hvfTime(float(JD))

#   HMSDH convert hours, minutes, seconds to decimal hours
def HMSDH(hh, mm, ss):  # hh may be negative
    habs = abs(hh)
    habs += mm/60. + ss/3600.
    h = habs
    if hh < 0: h = -habs
    return h

#   DHHMS converts decimal hours into hours, minutes, seconds
#   seconds may have fractions, no rounding
def DHHMS(h):   # h may be negative, then only hours will be negative
    sign = 1
    if h < 0: sign = -1
    habs = abs(h)
    hh = int(habs)
    mm = int(Frac(habs)* 60.)
    ss = Frac(habs*60)*60.
    hh = hh * sign
    return (hh, mm, ss)

#   returns HH:MM and a rounded value of h
#   h must be positive or None or ''
def HHMM(hh):
    if (hh == None or hh == ''):
        return('--')
    m = Frac(hh) * 60.0
    h = Int(hh)
    if (m >= 59.5):
        h += 1
        m -= 60.0
    m = int(round(m))
    string = "%02d:%02d = " % (h, m) + str(round(hh, 4))
    return (string)

#   returns HH:MM:SS and a rounded value of h
#   h must be positive or None or ''
def HHMMSS(hh):
    if (hh == 0):
        return('--')
    m = Frac(hh) * 60
    h = Int(hh)
    s = Frac(m) * 60.0
    m = Int(m)
    if (s >= 59.5):
        m += 1
        s -= 60.0
    if (m >= 60):
        h += 1
        m -= 60
    s = int(round(s))
    string = "%02d:%02d:%02d = " % (h, m, s) + str(round(hh, 5))
    return (string)

def hmsstring(hour):
    # hour may also be negative, e.g. if used to convert declination
    (hh, mm, ss) = DHHMS(hour)
    str = "%d:%02d:%02d" % (hh, mm, ss)
    return str


#   local civic time to Universal Time
#   returns a hvfTime object 

def getUTfromCivicTime(dtLocalCivicTime, zone):    
    # dtLocalCivicTime is a hvfTime object
    # zone is time zone offset in hours, positive going East
    dtUT = dtLocalCivicTime - timedelta(hours=zone)
    return dtUT

def getCivicTimefromUT(dtUT, zone):
    dtLocalCivicTime = dtUT + timedelta(hours=zone)
    return dtLocalCivicTime

#   UTC to mean local time 
#   dtUT: Universal Time hvfTime object
#   longitude: Eastern longitude, degrees
#   returns a datetime object, rounded to seconds

def getMeanLocalTime(dtUT, longitude):
    dtmeanLT = dtUT + timedelta(seconds=round(240.*longitude, 0))
    return dtmeanLT

def getUTfromMeanLocalTime(dtmeanLT, longitude):
    dtUT =  dtmeanLT - timedelta(seconds=round(240.*longitude, 0))
    return dtUT

#   datetime.datetime or datetime.time to decimal hours
#   dttime is a hvfTime object

def getDecimalHour(dttime):
    (hour, min, sec) = hvfTime(dttime).getTime()
    return hour    

# calculate horizontal coordinates (Azimut, Altitude)
# from ecliptic coordinates (Hour angle, Declination)
#   Parameters:
#   Hour    Hour angle counted from meridian to West, degrees
#   Decl    Declination, degrees
#   latitude    degrees
#   Returns:
#   azimut  counted from North, degrees
#   alt     altitude above horizon, degrees
#   x, y, z Cartesian coords in horizontal system, x pointing South, y West
#   https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen

def getAziAltfromHourDecl(Hour, Decl, latitude):
    # Hour angle and Declination into azimut and altitude
    azimut = atan2(Sin(Hour), Cos(Hour)*Sin(latitude)-Tan(Decl)*Cos(latitude))
    alt    = asin(Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour))
    azimut = Mod(degrees(azimut)+180., 360.)
    alt    = degrees(alt)
    # Cartesian coords in horizon system
    x = Cos(alt)*Cos(azimut-180.)
    y = Cos(alt)*Sin(azimut-180.)
    z = Sin(alt)
    # dito calculated from  Hour, Decl, latitude
    xx = -Cos(latitude)*Sin(Decl)+Sin(latitude)*Cos(Decl)*Cos(Hour)
    yy = Cos(Decl)*Sin(Hour)
    zz = Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour)
    xyzDict = {
        "x":    x,
        "y":    y,
        "z":    z, 
        "xx":   xx,
        "yy":   yy,
        "zz":   zz, }
    return (azimut, alt, xyzDict)

# Zodiac signs for Sun and Moon
#   lon ecliptic longitude, degrees
def getSign(lon):
    Signs_de = ("Widder", "Stier", "Zwillinge", "Krebs", "Löwe", "Jungfrau", \
        "Waage", "Skorpion", "Schütze", "Steinbock", "Wassermann", "Fische")
    Signs_en = ("Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo", \
        "Libra", "Scorpio", "Sagittarius", "Capricorn", "Aquarius", "Pisces")
    return (Signs_en[int(floor(lon / 30))])

########################################################################
#
# getNutation(dtCivicTime, zone)
# Nutation in ecliptic longitude and ecliptic latitude
# see Meeus, Astronomical Algorithms, chapter 21

def getNutation(dtCivicTime, zone, verbose=False):
    # Epoch is J2000.0
    # JD 2451545
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    T = (JD - 2451545.0)/36525. # Centuries after the Epoch, 21.1

    # Moon Ascending Node
    Omega = 125.04452 - 1934.136261 * T + 0.0020708 * T*T + T*T*T/450000.
    Omega = Mod(Omega, 360.)
    O = radians(Omega)

    # accuracy of 0.5 arcsec in longitude, 0.1 arcsec in latitude
    # Sun Mean Longitude
    LmeanSun = 280.4665 + 36000.7698 * T
    LmeanSun = Mod(LmeanSun, 360.)
    LS = radians(LmeanSun)

    # Moon Mean Longitude
    LmeanMoon = 218.3165 + 481267.8813 * T
    LmeanMoon = Mod(LmeanMoon, 360.)
    LM = radians(LmeanMoon)

    # Effect on ecliptic longitude and latitude, arcsecs
    deltaELong = -17.20*sin(O) -1.32*sin(2.*LS) -0.23*sin(2*LM) +0.21*sin(2*O)
    deltaELat  = 9.20*cos(O) + 0.57*cos(2.*LS) + 0.10*cos(2*LM) -0.09*cos(2*O)

    # obliquity of the ecliptic, 21.2
    eps0 = 23+26/60.+21.448/3600. - 46.8150/3600.*T - 0.00059/3600.*T*T \
            +0.001813/3600*T*T*T
    eps = eps0 + deltaELat/3600.

    if verbose:
        print ("getNutation")
        print ("JD = ", JD)
        print ("T = ", T)
        print ("LmeanSun = ", LmeanSun)
        print ("LmeanMoon = ", LmeanMoon)
        print ("getNutation deltaELong (arcsec) = %.2f" % deltaELong)
        print ("getNutation deltaELat  (arcsec) = %.2f" % deltaELat )
        print ("eps0 (degrees) = %.6f" % eps0, ", ", HHMMSS(eps0))
        print ("eps  (degrees) = %.6f" % eps, ", ", HHMMSS(eps) )

    # return values in degrees
    return (deltaELong/3600., deltaELat/3600., eps)

#################################################################
#   Funktionen für die Berechnung der Zeitgleichung

def getSunStufffromTT(TT, verbose = False):
    # TT Tage nach Epoch2010 inkl. DeltaT Korrektur
    # returns RA, epsilon in degrees

    def atanLage(x, Lage):    # arctangens mit Lageparameter, alle Werte in Grad
        # siehe https://en.wikipedia.org/wiki/Equation_of_time
        # Resultat in Grad
        xrad = radians(x)
        Lagerad = radians(Lage)
        y = atan(xrad)
        valuerad = y + pi * round((Lagerad-y)/pi, 0)
        value = degrees(valuerad)
        return value
    
    def getE(t):    # numerische Exzentrizität zur Zeit t
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        e = 0.01670491 - 4.208E-5*T - 1.26E-7*T*T
        return e
    
    def getEps(t):  # Schiefe der Ekliptik in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
        return eps
    
    def getOmegabar(t):     # Länge des Perihels in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        omegabar = 283.11243767 + 1.720171*T + 0.000452778*T*T
        return omegabar
    
    def getL(t):    # mittlere Länge der Sonne in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        L = 279.557285 + 36000.769585*T + 0.0003025*T*T
        return L
    
    def getC(e, M):    # Mittelpunktsgleichung equation of the center
        # berechnet wahre minus mittlere Anomalie als Fkt der Exzentrizität
        # und der mittleren Anomalie in dritter Näherung in e
        # e numerische Exzentrizität
        # M mittlere Anomalie der Sonne in Grad
        # Resultat in Grad
        # Danby, Fundamentals of Celestial Mechanics, p.438
        Mrad = radians(M)
        Crad = (2*e-e*e*e/4)*sin(Mrad) + (1.25*e*e -11./24.*e**4)*sin(2*Mrad)\
             + 13./12.*e*e*e*sin(3*Mrad) + 103./96.*e**4*sin(4*Mrad)
        C = degrees(Crad)
        return C

    t = TT
    e = getE(t)                 # Exzentrizität der Erdbahn, Zeit t
    eps = getEps(t)             # Schiefe der Ekliptik in Grad, Zeit t
    epsRad = radians(eps)       # Schiefe der Ekliptik in Radians
    omegabar = getOmegabar(t)   # Länge des Perihels in Grad, Zeit t
    L = getL(t)                 # mittlere Länge der Sonne in Grad, Zeit t
    Lhaupt = Mod(L, 360)        # Hauptwert der Länge der Sonne
    M = Lhaupt - omegabar       # Mittlere Anomalie der Sonne in Grad, Zeit t
    Mhaupt = Mod(M, 360)        # Hauptwert der mittleren Anomalie
    C = getC(e, Mhaupt)         # Mittelpunktsgleichung, wahre minus mittlere A
    Lwahr = L + C               # wahre Länge der Sonne in Grad, Zeit t
    LwahrRad = radians(Lwahr)   # wahre Länge der Sonne in Radians
    Lwahrhaupt = Mod(Lwahr, 360)# Hauptwert der wahren Länge
    LwahrHauptRad = radians(Lwahrhaupt)  # dito, in Radians


    # Rektaszension der Sonne in Funktion der wahren Länge der Sonne und der 
    # Schiefe der Ekliptik
    alpha = atanLage(degrees(tan(LwahrRad)*cos(epsRad)), Lwahr)
    alphahaupt = atanLage(degrees(tan(LwahrHauptRad)*cos(epsRad)), Lwahrhaupt)

    Dict = { "Object"   : "SunStuff",
            "TT"        : t,            # Tage nach Epoch2010 inkl. deltaT
            "MPEq"      : C,            # wahre - mittlere Anomalie
            "omegabar"  : omegabar,     # Länge des perihels in Grad, Zeit t
            "M"         : M,            # mittlere Anomalie, Grad
            "nu"        : C+M,          # wahre Anomalie
            "Long"      : Lwahrhaupt,   # wahre Länge der Sonne, Grad
            "eps"       : eps,          # Schiefe der Ekliptik, Grad
            "exc"       : e,            # Exzentrizität der Erdbahn
            }
    if verbose:
        print ("Numerische Exzentrizitaet e = ", "%.6f" % e)
        print ("Omegabar, Laenge des Perihels = ", "%.3f Grad" % omegabar)
        print ("Lhaupt, Hauptwert Mittlere Laenge = ", "%.3f Grad" % Lhaupt)
        print ("M Mittlere Anomalie = ", "%.3f Grad" % M)
        print ("Mhaupt, Hauptwert Mittlere Anomalie = ", "%.3f Grad" % Mhaupt)
        print ("C Mittelpunktsgleichung = ", "%.3f Grad" % C)
        print ("Alpha Wahre Anomalie = ", "%.3f Grad" % alpha)
    return (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict)

#######################################################################    
def Zeitgleichung(JD, verbose=False):
    # berechnet die Zeitgleichung in ganzen Sekunden
    # JD Julianisches Datum
    # verbose aktiviert zusätzlichen Output
    # Resultat in Sekunden ohne Rundung
    # Epoche konsistent 2010.0, 31.12.2009 0h UT, JD 2455196.5
    dtEpoch2010 = datetime(2009, 12, 31)
    tbase = getJulianDay (dtEpoch2010)
    t = JD - tbase
    TT = JD - tbase + getDeltaT(JD)/86400.

    (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict) = getSunStufffromTT(TT, verbose)

    # Zeitgleichung erste Ursache Exzentrizität der Erdbahn (epsilon = 0)
    ZGstar0 = L - Lwahr         # in Grad
    ZG0 = 240. * ZGstar0        # in Sekunden

    # zweite Ursache Anteil Schiefe der Ekliptik
    ZGstareps = Lwahrhaupt - alphahaupt # in Grad
    ZGeps = 240. * ZGstareps            # in Sekunden

    # Summe der beiden Ursachen
    ZGstar = ZGstar0 + ZGstareps        # in Grad
    ZGsecs = ZG0 + ZGeps                # in Sekunden

    # ZGmin, ZGsec for printing purposes 
    ZGstr = niceminsec(ZGsecs)
    if verbose:
        print ("Epoch2010 = ", dtEpoch2010)
        print ("Epoch Julian Date = ", tbase)
        print ("TT = ", TT)
        print ("Schiefe der Ekliptik eps = ", "%.3f Grad" % eps)
        print ("L, Mittlere Laenge = ", "%.3f Grad" % L)
        print ("Lwahr Wahre Sonnenlaenge = ", "%.3f Grad" % Lwahr)
        print ("Lwahrhaupt Hauptwert Wahre Sonnenlaenge = ", "%.3f Grad" \
                    % Lwahrhaupt)
        print ("Alphahaupt Wahre Anomalie = ", "%.3f Grad" % alphahaupt)
        print ("ZGstar0 = ", "%.3f" % ZGstar0, " Grad")
        print ("ZG0 = ", "%.2f" % ZG0, " Sekunden")
        print ("ZGstareps = ", "%.3f" % ZGstareps, " Grad")
        print ("ZGeps = ", "%.2f" % ZGeps, " Sekunden")
        print ("ZGstar = ", "%.3f" % ZGstar, " Grad")
        print ("ZGsecs = ", "%.2f" % ZGsecs, " Sekunden")
        print ("ZG = %s" % ZGstr)

    return (ZGsecs, ZGstr, Dict)

def getZeitgleichung(JD, verbose=False):
    # returns a timedelta object and two int's for printing
    zgsecs, ZGstr, Dict = Zeitgleichung(JD, verbose)
    return (timedelta(seconds=zgsecs), ZGstr, Dict)

# Greenwich Mean Sidereal Time from Julian Date

def getGMSTfromJD(JD):
    # Meeus chapter 11 p. 83
    T = (JD - 2451545.0) / 36525.   # (11.1)
    theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) \
            + 0.000387933 * T*T - T*T*T/387110000   # (11.4)
    # normalize to 360 degrees
    theta = Mod(theta0, 360.)
    # add to midnight
    JD0 = floor(JD-0.5) + 0.5
    Gdate = hvfTime(JD0 + theta/360.)
    return Gdate

# Local Mean Sideral Time from GMST, longitude
# returns a datetime object

def getLMSTfromGMST(gmst, longitude):
    # gmst datetime object
    # longitude in degrees
    # returns a datetime object
    lmst = gmst + timedelta(hours=longitude/15.)
    return lmst

# Greenwich Mean Sideral Time from LMST, longitude
# returns a datetime object

def getGMSTfromLMST(lmst, longitude):
    # lmst datetime object
    # longitude in degrees
    # returns a datetime object
    #gmst = datetime.time(hour=lmst) - timedelta(hours=longitude/15.)
    gmst = lmst - timedelta(hours=longitude/15.)
    return gmst
    

# deltaT, the difference between Terrestrial Time TT and
# Universal Time 
# see https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
# for Julian Date JD

def getDeltaT(JD): 
    # convert JD into year, month, day
    (year, month, day) = hvfTime(JD).getDate()
    # We define the decimal year "y" as follows:
    y = year + (month - 0.5)/12

    if year < -500:
        u = (year-1820.)/100.
        deltaT = -20 + 32*u*u
    elif year < 500:
        u = y/100.
        deltaT = 10583.6 - 1014.41*u + 33.78311*u*u - 5.952053*u**3 \
                - 0.1798452*u**4 + 0.022174192*u**5 + 0.0090316521*u**6
    elif year < 1600:
        u = (y-1000.)/100.
        deltaT = 1574.2 - 556.01*u + 71.23472*u*u + 0.319781*u**3 \
                - 0.8503463*u**4 - 0.005050998*u**5 + 0.0083572073*u**6
    elif year < 1700:
        t = y - 1600.
        deltaT = 120. -0.9808*t - 0.01532*t*t + t**3/7129
    elif year < 1800:
        t = y - 1700.
        deltaT = 8.83 + 0.1603*t - 0.0059285*t*t + 0.00013336*t**3 \
                - t**4/1174000.
    elif year < 1860:
        t = y - 1800.
        deltaT = 13.72 - 0.332447*t + 0.0068612*t*t + 0.0041116*t**3 \
                - 0.00037436*t**4 + 0.0000121272*t**5 \
                - 0.0000001699*t**6 + 0.000000000875*t**7
    elif year < 1900:
        t = y - 1860.
        deltaT = 7.62 + 0.5737*t - 0.251754*t*t + 0.01680668*t**3 \
                - 0.0004473624*t**4 + t**5/233174.
    elif year < 1920:
        t = y - 1900.
        deltaT = -2.79 + 1.494119*t - 0.0598939*t*t + 0.0061966*t**3 \
                -0.000197*t**4
    elif year < 1941:
        t = y - 1920.
        deltaT = 21.20 + 0.84493*t - 0.076100*t*t + 0.0020936*t**3
    elif year < 1961:
        t = y - 1950.
        deltaT = 29.07 + 0.407*t - t*t/233. + t**3/2547.
    elif year < 1986:
        t = y - 1975.
        deltaT = 45.45 + 1.067*t - t*t/260. - t**3/718.
    elif year < 2005:
        t = y - 2000.
        deltaT = 63.86 + 0.3345*t - 0.060374*t*t + 0.0017275*t**3 \
                + 0.000651814*t**4 + 0.00002373599*t**5
    elif year < 2050:
        t = y - 2000.
        deltaT = 62.92 + 0.32217*t + 0.005589*t*t
    elif year < 2150:
        deltaT = -20. + 32*((y-1820.)/100)**2 - 0.5628*(2150.-y)
    else:   # after 2150
        u = (year-1820.)/100.
        deltaT = -20. + 32.*u*u
        # https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
        # small correction beyond 2150 (only needed for Perihelion calc's)
        c = -0.000012932 * (y - 1955.)**2
        deltaT += c

    return deltaT

def getEps(t):  # Schiefe der Ekliptik in Grad
    # t in Tage nach Epoch2010
    # Resultat in Grad
    T = t / 36525
    eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
    return eps


#######################################################################
# Jean Meeus, Astronomical Algorithms, Willmann-Bell, Richmond VA, 1991
#
# hvf 03.12.2020

# quadratic interpolation
# xi 3 Stuetzstellen, yi 3 Funktionswerte
# Meeus p. 25
def interp(xi, yi, x):
    a10 = (yi[1]-yi[0])/(xi[1]-xi[0])
    a21 = (yi[2]-yi[1])/(xi[2]-xi[1])
    a210 = (a21-a10)/(xi[2]-xi[0])
    res = yi[0] + a10*(x-xi[0]) + a210*(x-xi[0])*(x-xi[1])
    return res

#############################################################################
#! /usr/bin/env python3
# getTimezone(verbose)
# returns (server) timezone using /etc/timezone, timedatectl, or a default
# hvf 07.05.2024

import re
import os
import subprocess

def getTimezone(verbose = False):
    # default timezone is Europe/Zurich
    timezone = "Europe/Zurich"
    # easiest: use timezone file and read its content§
    TimezoneFile = "/etc/timezone"
    try:
        file = open(TimezoneFile, "r") 
    except (Exception, FileNotFoundError, OSError) as  e:
        #print ("Error ", e, "Thrown.")
        #print ("file not found: ", filename, "\n")
        pass

    else:
        for line in file:
            #print ("Line: ", line)
            #length = len(line); print ("Length =", length)
            line = line.strip()
            #print ("Line after stripping: ", line)
            #length = len(line); print ("Length =", length)
            timezone = line
            if verbose:
                print ("Timezone from ", TimezoneFile," is: ", timezone)
            return timezone

    # next method: use timedatectl
    # Define the command you want to run as a list of arguments
    cmd = ["/usr/bin/timedatectl", "show", "-p", "Timezone"]

    # Use subprocess.Popen to start the process
    try:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    except Exception as e:
        #print ("Error: ", e, "\n")
        pass
    else:
        # Communicate with the process and get its output and errors
        output, errors = process.communicate()

        # Print the output and errors (if any)
        #print("Output:", output)
        #print("Errors:", errors)

        # Get the return code of the process (0 usually means success)
        return_code = process.returncode
        #print("Return Code:", return_code)
        lines = output.split("\n")
        for line in lines:
            if (line == ''): break
            #print ("Line = ", line)
            #print ("length = ", len(line))
            line = line.strip()
            #print (line)
            #print ("length = ", len(line))
            # presumably parse line now
            lArr = re.split("=", line)
            #print ("After splitting: ", lArr)
            timezone = lArr[1]
            if verbose:
                print ("Timezone from ", cmd[0]," is: ", timezone)
            return timezone

    # default timezone
    if verbose:
        print ("Default timezone: ", timezone)
    return timezone
'''
tz = getTimezone(True)
print ("Timezone: ", tz)
tz = getTimezone(False)
print ("Timezone: ", tz)
tz = getTimezone(1)
print ("Timezone: ", tz)
tz = getTimezone(0)
print ("Timezone: ", tz)
tz = getTimezone(None)
print ("Timezone: ", tz)
'''

########################################################################
#
#   getMoonData(dtCivicTime, zone, verbose=False):
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       MoonDist    distance Earth - Moon, km
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonHP      horizontal parallax, degrees
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getMoonData.py"
VersionList.append(Version)

def getMoonData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.

    T = TT / 36525.    # eq. 21.1

    # obliquity of the ecliptic, eq. 21.2, degrees
    epsilon = 23.+26./60+21.448/3600. - 46.8150/3600.*T \
        - 0.00059/3600.*T*T + 0.001813/3600.*T*T*T

    # Moon mean longitude (mean equinox of the date, incl, effect of light-time)
    # eq. 45.1
    Lprime  = 218.3164591 + T * (481267.88134236 + T * (-0.0013268 \
            + T * (1./538841. - T / 65194000)))

    # Mean elongation of the Moon
    # eq. 45.2
    D       = 297.8502042 + T * (445267.1115168 + T * (-0.0016300 \
            + T * (1./545868.  -T / 113065000.)))

    # Sun's mean anomaly
    # eq. 45.3
    M       = 357.5291092 + T * (35999.0502909 + T * (-0.0001536 \
            + T / 24490000))

    # Moon's mean anomaly
    # eq 45.4
    MM      = 134.9634114 + T * (477198.8676313 + T * (0.0089970 \
            + T * (1./69699.  - T / 14712000.)))

    # Moon's argument of latitude (mean distance of the Moon
    # from its ascending node)
    # eq 45.5
    F       = 93.2720993 + T * (483202.0175273 + T * (-0.0034029 \
            + T * (-1./3526000. + T / 863310000.)))

    # three further arguments are needed
    A1      = 119.75 + 131.849 * T          # action of Venus
    A2      =  53.09 + 479264.290 * T       # action of Jupiter
    A3      = 313.45 + 481266.484 * T

    # effect of decreasing eccentricity of Earth's orbit
    E       = 1. + T * (-0.002516 + T * -0.0000074)
    EE      = E*E

    # Longitude, distance: now calculate sums sigL, sigR
    # see Table 45.A, p.309
    arg = MM;           sigL = 6288774*Sin(arg);    sigR = -20905355*Cos(arg)
    arg = 2*D - MM;     sigL += 1274027*Sin(arg);   sigR += -3699111*Cos(arg)
    arg = 2*D;          sigL += 658314*Sin(arg);    sigR += -2955968*Cos(arg)
    arg = 2*MM;         sigL += 213618*Sin(arg);    sigR += -569925*Cos(arg)  
    arg = M;            sigL += -185116*Sin(arg)*E; sigR += 48888*Cos(arg)*E
    arg = 2*F;          sigL += -114332*Sin(arg);   sigR += -3149*Cos(arg)
    arg = 2*D - 2*MM;   sigL += 58793*Sin(arg);     sigR += 246158*Cos(arg)
    arg = 2*D - M - MM; sigL += 57066*Sin(arg)*E;   sigR += -152138*Cos(arg)*E
    arg = 2*D + MM;     sigL += 53322*Sin(arg);     sigR += -170733*Cos(arg)
    arg = 2*D - M;      sigL += 45758*Sin(arg)*E;   sigR += -204586*Cos(arg)*E
    arg = M - MM;       sigL += -40923*Sin(arg)*E;  sigR += -129620*Cos(arg)*E
    arg = D;            sigL += -34720*Sin(arg);    sigR += 108743*Cos(arg)
    arg = M + MM;       sigL += -30383*Sin(arg)*E;  sigR +=  104755*Cos(arg)*E
    arg = 2*D - 2*F;    sigL += 15327*Sin(arg);     sigR += 10321*Cos(arg)
    arg = MM + 2*F;     sigL += -12528*Sin(arg)    
    arg = MM - 2*F;     sigL += 10980*Sin(arg);     sigR += 79661*Cos(arg)
    arg = 4*D-MM;       sigL += 10675*Sin(arg);     sigR += -34782*Cos(arg)
    arg = 3*MM;         sigL += 10034*Sin(arg);     sigR += -23210*Cos(arg)
    arg = 4*D-2*MM;     sigL +=  8548*Sin(arg);     sigR += -21636*Cos(arg)
    arg = 2*D + M - MM; sigL += -7888*Sin(arg)*E;   sigR += 24208*Cos(arg)*E
    arg = 2*D + M ;     sigL += -6766*Sin(arg)*E;   sigR += 30824*Cos(arg)*E
    arg = D - MM;       sigL += -5163*Sin(arg);     sigR += -8379*Cos(arg)
    arg = D + M;        sigL += 4987*Sin(arg)*E;    sigR += -16675*Cos(arg)*E
    arg = 2*D - M + MM; sigL += 4036*Sin(arg)*E;    sigR += -12831*Cos(arg)*E
    arg = 2*D + 2*MM;   sigL += 3994*Sin(arg);      sigR += -10445*Cos(arg)
    arg = 4*D;          sigL += 3861*Sin(arg);      sigR += -11650*Cos(arg)
    arg = 2*D - 3*MM;   sigL += 3665*Sin(arg);      sigR += 14403*Cos(arg)
    arg = M - 2*MM;     sigL += -2689*Sin(arg)*E;   sigR += -7003*Cos(arg)*E
    arg = 2*D - MM + 2*F; sigL += -2602*Sin(arg)    
    arg = 2*D - M - 2*MM; sigL += 2390*Sin(arg)*E;  sigR += 10056*Cos(arg)*E
    arg = D + MM;       sigL += -2348*Sin(arg);     sigR += 6322*Cos(arg)
    arg = 2*D - 2*M ;   sigL += 2236*Sin(arg)*EE;   sigR += -9884*Cos(arg)*EE
    arg = M + 2*MM;     sigL += -2120*Sin(arg)*E;   sigR += 5751*Cos(arg)*E
    arg = 2*M ;         sigL += -2069*Sin(arg)*EE
    arg = 2*D - 2*M - MM; sigL += 2048*Sin(arg)*EE; sigR += -4950*Cos(arg)*EE
    arg = 2*D + MM - 2*F;   sigL += -1773*Sin(arg); sigR += 4130*Cos(arg)
    arg = 2*D + 2*F;    sigL += -1595*Sin(arg)
    arg = 4*D - M - MM; sigL += 1215*Sin(arg)*E;    sigR += -3958*Cos(arg)*E
    arg = 2*MM + 2*F;   sigL += -1110*Sin(arg)
    arg = 3*D - MM;     sigL += -892*Sin(arg);      sigR += 3258*Cos(arg)
    arg = 2*D + M + MM; sigL += -810*Sin(arg)*E;    sigR += 2616*Cos(arg)*E
    arg = 4*D - M - 2*MM; sigL += 759*Sin(arg)*E;   sigR += -1897*Cos(arg)*E
    arg = 2*M - MM;     sigL += -713*Sin(arg)*EE;   sigR += -2117*Cos(arg)*EE
    arg = 2*D + 2*M - MM; sigL += -700*Sin(arg)*EE; sigR += 2354*Cos(arg)*EE
    arg = 2*D + M - 2*MM;   sigL += 691*Sin(arg)*E
    arg = 2*D - M - 2*F;    sigL += 596*Sin(arg)*E
    arg = 4*D + MM;     sigL += 549*Sin(arg);       sigR += -1423*Cos(arg)
    arg = 4*MM;         sigL += 537*Sin(arg);       sigR += -1117*Cos(arg)
    arg = 4*D - M;      sigL += 520*Sin(arg)*E;     sigR += -1571*Cos(arg)*E
    arg = D - 2*MM;     sigL += -487*Sin(arg);      sigR += -1739*Cos(arg)
    arg = 2*D + M - 2*F; sigL += -399*Sin(arg)*E
    arg = 2*MM - 2*F;   sigL += -381*Sin(arg);      sigR += -4421*Cos(arg)
    arg = D + M + MM ;  sigL += 351*Sin(arg)*E
    arg = 3*D - 2*MM;   sigL += -340*Sin(arg)
    arg = 4*D - 3*MM;   sigL += 330*Sin(arg)
    arg = 2*D - M + 2*MM ;  sigL += 327*Sin(arg)*E
    arg = 2*M + MM;     sigL += -323*Sin(arg)*EE;   sigR += 1165*Cos(arg)*EE
    arg = D + M - MM ;  sigL += 299*Sin(arg)*E
    arg = 2*D + 3*MM;   sigL += 294*Sin(arg)
    arg = 2*D - MM - 2*F;                           sigR += 8752*Cos(arg)

    # Latitude: calculate sumB, see Table 45.B, p.311
    arg = F;                    sigB = 5128122*Sin(arg)
    arg = MM + F;               sigB += 280602*Sin(arg)
    arg = MM - F;               sigB += 277693*Sin(arg)
    arg = 2*D - F;              sigB += 173237*Sin(arg)
    arg = 2*D - MM + F;         sigB += 55413*Sin(arg)
    arg = 2*D - MM - F;         sigB += 46271*Sin(arg)
    arg = 2*D + F;              sigB += 32573*Sin(arg)
    arg = 2*MM + F;             sigB += 17198*Sin(arg)
    arg = 2*D + MM - F;         sigB += 9266*Sin(arg)
    arg = 2*MM - F;             sigB += 8822*Sin(arg)
    arg = 2*D - M - F;          sigB += 8216*Sin(arg)*E
    arg = 2*D - 2*MM - F;       sigB += 4324*Sin(arg)
    arg = 2*D + MM + F;         sigB += 4200*Sin(arg)
    arg = 2*D + M - F;          sigB += -3359*Sin(arg)*E
    arg = 2*D - M - MM + F;     sigB += 2463*Sin(arg)*E
    arg = 2*D - M + F;          sigB += 2211*Sin(arg)*E
    arg = 2*D - M - MM - F;     sigB += 2065*Sin(arg)*E
    arg = M - MM - F;           sigB += -1870*Sin(arg)*E
    arg = 4*D - MM - F;         sigB += 1828*Sin(arg)
    arg = M + F;                sigB += -1794*Sin(arg)*E
    arg = 3*F;                  sigB += -1749*Sin(arg)
    arg = M - MM + F;           sigB += -1565*Sin(arg)*E
    arg = D + F;                sigB += -1491*Sin(arg)
    arg = M + MM + F;           sigB += -1475*Sin(arg)*E
    arg = M + MM - F;           sigB += -1410*Sin(arg)*E
    arg = M - F;                sigB += -1344*Sin(arg)*E
    arg = D - F;                sigB += -1335*Sin(arg)
    arg = 3*MM + F;             sigB += 1107*Sin(arg)
    arg = 4*D - F;              sigB += 1021*Sin(arg)
    arg = 4*D - M + F;          sigB += 833*Sin(arg)
    arg = MM - 3*F;             sigB += 777*Sin(arg)
    arg = 4*D - 2*MM + F;       sigB += 671*Sin(arg)
    arg = 2*D - 3*F;            sigB += 607*Sin(arg)
    arg = 2*D + 2*MM - F;       sigB += 596*Sin(arg)
    arg = 2*D - M + MM - F;     sigB += 491*Sin(arg)*E
    arg = 2*D - 2*MM + F;       sigB += -451*Sin(arg)
    arg = 3*MM - F;             sigB += 439*Sin(arg)
    arg = 2*D + 2*MM + F;       sigB += 422*Sin(arg)
    arg = 2*D - 3*MM - F;       sigB += 421*Sin(arg)
    arg = 2*D + M - MM + F;     sigB += -366*Sin(arg)*E
    arg = 2*D + M + F;          sigB += -351*Sin(arg)*E
    arg = 4*D + F;              sigB += 331*Sin(arg)
    arg = 2*D - M + MM + F;     sigB += 315*Sin(arg)*E
    arg = 2*D - 2*M - F;        sigB += 302*Sin(arg)*EE
    arg = MM + 3*F;             sigB += -283*Sin(arg)
    arg = 2*D + M + MM - F;     sigB += -229*Sin(arg)*E
    arg = D + M - F;            sigB += 223*Sin(arg)*E
    arg = D + M + F;            sigB += 223*Sin(arg)*E
    arg = M - 2*MM - F;         sigB += -220*Sin(arg)*E
    arg = 2*D + M - MM - F;     sigB += -220*Sin(arg)*E
    arg = D + MM + F;           sigB += -185*Sin(arg)
    arg = 2*D - M - 2*MM - F;   sigB += 181*Sin(arg)*E
    arg = M + 2*MM + F;         sigB += -177*Sin(arg)*E
    arg = 4*D - 2*MM - F;       sigB += 176*Sin(arg)
    arg = 4*D - M - MM - F;     sigB += 166*Sin(arg)*E
    arg = D + MM - F;           sigB += -164*Sin(arg)
    arg = 4*D + MM - F;         sigB += 132*Sin(arg)
    arg = D - MM - F;           sigB += -119*Sin(arg)
    arg = 4*D - M - F;          sigB += 115*Sin(arg)*E
    arg = 2*D - 2*M + F;        sigB += 107*Sin(arg)*EE

    # Additives to sigL (p.112)
    sigL +=   3958*Sin(A1) \
            + 1962*Sin(Lprime-F) \
            + 318*Sin(A2)

    # Additives to sigB
    sigB += -2235*Sin(Lprime) \
            + 382*Sin(A3) \
            + 175*Sin(A1-F) \
            + 175*Sin(A1+F) \
            + 127*Sin(Lprime-MM) \
            - 115*Sin(Lprime+MM)

    MoonLong = Lprime + sigL/1e6                # degrees
    MoonLong = Mod(MoonLong, 360.)              # degrees, normnalized
    MoonLat  = sigB/1e6                         # degrees
    MoonDist = 385000.56 + sigR/1e3             # km
    MoonHP   = degrees(asin(6378.14/MoonDist))  # degrees
    MoonDistAU = MoonDist / 149597870.700       # au
    X = MoonDistAU * Cos(MoonLat) * Cos(MoonLong) # au, geocentric X
    Y = MoonDistAU * Cos(MoonLat) * Sin(MoonLong) # au, geocentric y
    Z = MoonDistAU * Sin(MoonLat)               # au, geocentric Z

    # Lunar Node and lunar Perigee
    # p.313, eq. 45.7

    MoonNode = 125.0445550 + T * (-1934.1361849 + T * (0.0020762 \
        + T * (1./467410. - T / 60616000.)))
    MoonNode = Mod(MoonNode, 360.)
    MoonPerigee = 83.3532430 + T * (4069.0137111 + T * (-0.0103238 \
        + T * (-1./80053 + T / 18999000.)))
    MoonPerigee = Mod(MoonPerigee, 360.)

    MoonRA = atan2(Sin(MoonLong)*Cos(epsilon)-Tan(MoonLat)*Sin(epsilon), \
                Cos(MoonLong))
    MoonDecl = asin(Sin(MoonLat)*Cos(epsilon) \
            + Cos(MoonLat)*Sin(epsilon)*Sin(MoonLong))
    MoonRA = Mod(degrees(MoonRA), 360.); MoonDecl = degrees(MoonDecl)

    # Moon altitude at rise, set (see p.98)
    Alt = 0.7275 * MoonHP - 34./60. # degrees

    if verbose:
        print ("JD  = ", JD)
        print ("T   = ", T)
        print ("Lprime = ", Deg(Lprime))
        print ("D   = ", Deg(D))
        print ("M   = ", Deg(M))
        print ("MM  = ", Deg(MM))
        print ("F   = ", Deg(F))
        print ("A1  = ", Deg(A1))
        print ("A2  = ", Deg(A2))
        print ("A3  = ", Deg(A3))
        print ("E   = ", E)
        print ("sigL= %.1f" % sigL)
        print ("sigB= %.1f" % sigB)
        print ("sigR= %.1f" % sigR)
        print ("MoonLong    = %10.6f degrees" % MoonLong)
        print ("MoonLat     = %10.6f degrees" % MoonLat )
        print ("MoonDist    =   %.1f km" % MoonDist)
        print ("MoonHP      = %10.6f degrees" % MoonHP  )
        print ("MoonNode    = %10.6f degrees" % MoonNode)
        print ("MoonPerigee = %10.6f degrees" % MoonPerigee)
        print ("MoonRA      = %10.6f degrees" % MoonRA     )
        print ("MoonDecl    = %10.6f degrees" % MoonDecl)
        print ("Alt         = %10.6f degrees" % Alt     )
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       RA          Rektaszension
#       Decl        Deklination
#       MoonDist    distance Earth - Moon, km
#       MoonHP      horizontal parallax, degrees
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)
#       suffix      "++" if MoonDist <= 360000, "--" if MoonDist >= 405000
    suffix = '  '
    if MoonDist <= 360000. : suffix = "++" # Super Moon
    if MoonDist >= 405000. : suffix = "--" # Mini Moon

    MoonDict = {
        "Object"    : "Moon",
        "MoonLong"  : MoonLong,
        "MoonLat"   : MoonLat,
        "RA"        : MoonRA,
        "Decl"      : MoonDecl,
        "MoonDist"  : MoonDist,
        "MoonHP"    : MoonHP,
        "MoonDistAU": MoonDistAU,
        "X"         : X,
        "Y"         : Y,
        "Z"         : Z,
        "MoonNode"  : MoonNode,
        "MoonPerigee": MoonPerigee,
        "Alt"       : Alt,
        "suffix"    : suffix}
    

    return (MoonRA, MoonDecl, MoonDict)

################ End of getMoonData() #################################
Version = "V3.1 hvf 2023-11-29 MoonPhasesMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(Version)
# Phases of the Moon, Meeus Chapter 47
# hvf 08.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87
# includes Solar Eclipses - hvf 28.06.2023

# getPhasesJDEbyK (k, verbose = False)
# k counts the number of new moons since 2000-01-06
# Meeus chapter 47
# returns the Julian Ephemeris Days (JDE) time of the mean phase of the Moon
# returns also the elements for solar eclipse (for new moon)

def getPhasesJDEbyK(k, verbose = False):
    # k integer:        New Moon    - Solar eclipse possible
    # k integer + 0.25: First Quarter
    # k integer + 0.5 : Full Moon   - Lunar eclipse possible
    # k integer + 0.75: Last Quarter

    T  = k / 1236.85            # Julian centuries after Epoch2000 (47.3)
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # Sun's Mean anomaly at time JDE (47.4)
    M = 2.5534 + 29.10535669*k - 0.0000218*T*T - 0.00000011*T*T*T
    M = Mod(M, 360.)

    # Moon's Mean anomaly (47.5)
    MM = 201.5643 + 385.81693528*k + 0.0107438*T*T \
            + 0.00001239*T**3 - 0.000000058*T**4
    MM = Mod(MM, 360.)

    # Moon's argument of Latitude (47.6)
    F = 160.7108 + 390.67050274*k - 0.0016341*T*T \
            - 0.00000227*T**3 + 0.000000011*T**4
    F = Mod(F, 360.)

    # Longitude of the Ascending Node of the lunar orbit (47.7)
    Omega = 124.7746 - 1.56375580*k \
            + 0.0020691*T*T + 0.00000215*T**3
    Omega = Mod(Omega, 360.)

    # chapter 52 Solar Eclipses  p.350
    F1 = F - 0.02665 * Sin(Omega)

    # Planetary Arguments

    A1  = 299.77 +  0.107408*k - 0.009173*T*T
    A2  = 251.88 +  0.016321*k
    A3  = 251.83 + 26.651886*k
    A4  = 349.42 + 36.412478*k
    A5  =  84.66 + 18.206239*k
    A6  = 141.74 + 53.303771*k
    A7  = 207.14 +  2.453732*k
    A8  = 154.84 +  7.306860*k
    A9  =  34.52 + 27.261239*k
    A10 = 207.19 +  0.121824*k
    A11 = 291.34 +  1.844379*k
    A12 = 161.72 + 24.198154*k
    A13 = 239.56 + 25.513099*k
    A14 = 331.55 +  3.592518*k

    # Corrections (in days) to JDE (p. 321)
    # for New Moon and Full Moon

    arg = MM;       sNew =  -0.40720*Sin(arg);   sFull =  -0.40614*Sin(arg)
    arg = M;        sNew +=  0.17241*E*Sin(arg); sFull +=  0.17302*E*Sin(arg)
    arg = 2*MM;     sNew +=  0.01608*Sin(arg);   sFull +=  0.01614*Sin(arg)
    arg = 2*F;      sNew +=  0.01039*Sin(arg);   sFull +=  0.01043*Sin(arg)
    arg = MM-M;     sNew +=  0.00739*E*Sin(arg); sFull +=  0.00734*E*Sin(arg)
    arg = MM+M;     sNew += -0.00514*E*Sin(arg); sFull += -0.00515*E*Sin(arg)
    arg = 2*M;      sNew +=  0.00208*EE*Sin(arg); sFull += 0.00209*EE*Sin(arg)
    arg = MM-2*F;   sNew += -0.00111*Sin(arg);   sFull += -0.00111*Sin(arg)
    arg = MM+2*F;   sNew += -0.00057*Sin(arg);   sFull += -0.00057*Sin(arg)
    arg = 2*MM+M;   sNew +=  0.00056*E*Sin(arg); sFull +=  0.00056*E*Sin(arg)
    arg = 3*MM;     sNew += -0.00042*Sin(arg);   sFull += -0.00042*Sin(arg)
    arg = M+2*F;    sNew +=  0.00042*E*Sin(arg); sFull +=  0.00042*E*Sin(arg)
    arg = M-2*F;    sNew +=  0.00038*E*Sin(arg); sFull +=  0.00038*E*Sin(arg)
    arg = 2*MM-M;   sNew += -0.00024*E*Sin(arg); sFull += -0.00024*E*Sin(arg)
    arg = Omega;    sNew += -0.00017*Sin(arg);   sFull += -0.00017*Sin(arg)
    arg = MM+2*M;   sNew += -0.00007*Sin(arg);   sFull += -0.00007*Sin(arg)
    arg = 2*MM-2*F; sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = 3*M;      sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = MM+M-2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = 2*MM+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM+M+2*F; sNew += -0.00003*Sin(arg);   sFull += -0.00003*Sin(arg)
    arg = MM-M+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM-M-2*F; sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 3*MM+M;   sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 4*MM;     sNew +=  0.00002*Sin(arg);   sFull +=  0.00002*Sin(arg)

    # Correction (in days) for First and Last Quarter (p.322)

    arg = MM;       sQ =  -0.62801*Sin(arg); 
    arg = M;        sQ +=  0.17172*E*Sin(arg); 
    arg = MM+M;     sQ += -0.01183*E*Sin(arg); 
    arg = 2*MM;     sQ +=  0.00862*Sin(arg); 
    arg = 2*F;      sQ +=  0.00804*Sin(arg); 
    arg = MM-M;     sQ +=  0.00454*E*Sin(arg); 
    arg = 2*M;      sQ +=  0.00204*EE*Sin(arg); 
    arg = MM-2*F;   sQ += -0.00180*Sin(arg); 
    arg = MM+2*F;   sQ += -0.00070*Sin(arg); 
    arg = 3*MM;     sQ += -0.00040*Sin(arg); 
    arg = 2*MM-M;   sQ += -0.00034*E*Sin(arg); 
    arg = M+2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = M-2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = 2*M+MM;   sQ += -0.00028*EE*Sin(arg); 
    arg = 2*MM+M;   sQ +=  0.00027*E*Sin(arg); 
    arg = Omega;    sQ += -0.00017*Sin(arg); 
    arg = MM-M-2*F; sQ += -0.00005*Sin(arg); 
    arg = 2*MM+2*F; sQ +=  0.00004*Sin(arg); 
    arg = MM+M+2*F; sQ += -0.00004*Sin(arg); 
    arg = MM-2*M;   sQ +=  0.00004*Sin(arg); 
    arg = MM+M-2*F; sQ +=  0.00003*Sin(arg); 
    arg = 3*M;      sQ +=  0.00003*Sin(arg); 
    arg = 2*MM-2*F; sQ +=  0.00002*Sin(arg); 
    arg = MM-M+2*F; sQ +=  0.00002*Sin(arg); 
    arg = 3*MM+M;   sQ += -0.00002*Sin(arg); 

    # additional correction for Quarter phases only

    W = 0.00306 - 0.00038*E*Cos(M) + 0.00026*Cos(MM) \
        - 0.00002*Cos(MM-M) + 0.00002*Cos(MM+M) + 0.00002*Cos(2*F)
    # for first Quarter: +W
    # for last Quarter:  -W

    # additional correction for all phases

    addCorr = 0.000325*Sin(A1)  + 0.000165*Sin(A2)  + 0.000164*Sin(A3) \
            + 0.000126*Sin(A4)  + 0.000110*Sin(A5)  + 0.000062*Sin(A6) \
            + 0.000060*Sin(A7)  + 0.000056*Sin(A8)  + 0.000047*Sin(A9) \
            + 0.000042*Sin(A10) + 0.000040*Sin(A11) + 0.000037*Sin(A12) \
            + 0.000035*Sin(A13) + 0.000023*Sin(A14)

    # times of the mean phases of the Moon, already affected by the Sun's
    # aberration and by the Moon's light-time (eq. 47.1)

    JDE = 2451550.09765 + 29.530588853*k \
            + 0.0001337*T*T \
            - 0.000000150*T*T*T \
            + 0.00000000073*T**4
    JDEsave = JDE

    # Eclipses correction to time of max eclipse (52.1)
    # sN: New Moon; sF: Full Moon
    arg = 2*MM;     sN  = 0.0161*Sin(arg); 
    arg = 2*F1;     sN += -0.0097*Sin(arg)
    arg = MM-M;     sN += 0.0073*E*Sin(arg)
    arg = MM+M;     sN += -0.0050*E*Sin(arg)
    arg = MM-2*F1;  sN += -0.0023*Sin(arg)
    arg = 2*M;      sN += 0.0021*E*Sin(arg)
    arg = MM+2*F1;  sN += 0.0012*Sin(arg)
    arg = 2*MM+M;   sN += 0.0006*E*Sin(arg)
    arg = 3*MM;     sN += -0.0004*Sin(arg)
    arg = M+2*F1;   sN += -0.0003*E*Sin(arg)
    arg = A1;       sN += 0.0003*Sin(arg)
    arg = M-2*F1;   sN += -0.0002*E*Sin(arg)
    arg = 2*MM-M;   sN += -0.0002*E*Sin(arg)
    arg = Omega;    sN += -0.0002*Sin(arg)

    sF = sN

    arg = MM;   sN += -0.4075*Sin(arg);     sF += -0.4065*Sin(arg)
    arg = M;    sN += 0.1721*E*Sin(arg);    sF += 0.1727*E*Sin(arg)

    # P and Q - p. 351
    Q = 5.2207
    arg = M;    P  = 0.2070*E*Sin(arg);     Q += -0.0048*E*Cos(arg)
    arg = 2*M;  P += 0.0024*E*Sin(arg);     Q += 0.0020*E*Cos(arg)
    arg = MM;   P += -0.0392*Sin(arg);      Q += -0.3299*Cos(arg)
    arg = 2*MM; P += 0.0116*Sin(arg)
    arg = MM+M; P += -0.0073*E*Sin(arg);    Q += -0.0060*E*Cos(arg)
    arg = MM-M; P += 0.0067*E*Sin(arg);     Q += 0.0041*E*Cos(arg)
    arg = 2*F1; P += 0.0118*Sin(arg)

    # WW, gamma, u
    # For a Solar eclipse, gamma is the least distance from the axis 
    # of the Moon's shadow to the center of the Earth, 
    # in units of the equatorial radius of the Earth.
    # gamma is positive if the shadow passes North of the center.
    # when gamma is between +0.9972 and -0.9972, the solar eclipse is central:
    # there exists a line of central eclipse on the Earth's surface.
    WW = abs(Cos(F1))
    gamma = (P*Cos(F1) + Q*Sin(F1)) * (1. - 0.0048*WW)

    # u denotes the radius of the Moon's umbral cone in the fundamental plane.
    # The radius of the penumbral cone in the fundamental plane is
    #   u + 0.5461

    u = 0.0059
    u += 0.0046*E*Cos(M)
    u += -0.0182*Cos(MM)
    u += 0.0004*Cos(2*MM)
    u += -0.0005*Cos(M+MM)

    # if abs(gamma) is between 0.9972 and 1.5433+u, the eclipse is not central.
    # If abs(gamma) is between 0.9972 and 1.5433 + u, the eclipse is not
    # central. In most cases, it is then a partial eclipse. However, when
    # abs(gamma) is between 0.9972 and 1.0260, a part of the umbral cone may
    # touch the surface of the Earth (within the polar regions), while the
    # axis of the cone does not touch the Earth. These non-central total
    # or annular eclipses occur when 0.9972 < abs(gamma) < 0.9972 + |u|.


    # CAREFUL: these tests must work for negative years as well!

    def isNewMoon(k):   # k must be integer
        test = (abs(k - int(k)) < 0.001)
        # if verbose: print ("isNewMoon k, test = ", k, test)
        return test     # True or False

    def isFullMoon(k):  # k must be half-integer
        test = (abs(abs(k - int(k)) - 0.5) < 0.001)
        # if verbose: print ("isFullMoon k, test = ", k, test)
        return test     # True or False 

    def isFirstQuarter(k):  # k must be .25
        test = abs(abs(k - floor(k)) - 0.25) < 0.001
        # if verbose: print ("isFirstQuarter k, test = ", k, test)
        return test     # True or False 

    def isLastQuarter(k):  # k must be .75
        test = abs(abs(ceil(k) - k) - 0.25) < 0.001
        # if verbose: print ("isLastQuarter k, test = ", k, test)
        return test     # True or False 

    JDEeclipse = None
    if isNewMoon(k):
        JDE += sNew
        JDEeclipse = hvfTime(JDEsave + sN)
        # if verbose: print ("New Moon")
    elif isFullMoon(k):
        JDE += sFull
        # if verbose: print ("Full Moon")
    elif isFirstQuarter(k):
        JDE += sQ
        JDE += W
        # if verbose: print ("First Quarter")
    elif isLastQuarter(k):
        JDE += sQ
        JDE -= W
        # if verbose: print ("Last Quarter")
    else:
        print ("k not for New or Full Moon: ", k)
    JDE += addCorr
    eclType = None
    omega = None
    SEpossible = False
    if isNewMoon(k) and abs(gamma) <= 1.5433+u: 
        SEpossible = True
        if abs(gamma) < 0.9972:
            if u < 0:
                eclType = ("Total Central Solar Eclipse")
            elif u < 0.0047:
                omega = 0.00464*sqrt(1-gamma**2)
                if u < omega:
                    eclType = ("Annular-total Central Solar Eclipse")
                else:
                    eclType = ("Annular Central Solar Eclipse")
            elif u >= 0.0047:
                eclType = ("Annular Central Solar Eclipse")
        elif abs(gamma) < 0.9972+abs(u):
            eclType = ("Non-central total or annular Solar eclipse")
        elif abs(gamma) < 1.5433+u:
            eclType = ("Non-central Solar Eclipse")
        if verbose:
            print ("getPhasesJDEbyK: k = %.2f" % k)
            print ("JDE = %.6f, %s" % (JDE, nicedatetime(hvfTime(JDE))))
            print ("JDEsave = %.6f, %s" % (JDEsave, nicedatetime(hvfTime(JDEsave))))
            print ("corrections sN, sF = ", sN, sF)
            #if JDEeclipse: print ("JDEeclipse = %.6f, %s" % (JDEeclipse, nicedatetime(hvfTime(JDEeclipse))))
            if JDEeclipse: print ("JDEeclipse = %s" % (nicedatetime(hvfTime(JDEeclipse))))
            print ("E, M(Sun), MM(Moon), F, Omega = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (E, M, MM, F, Omega)) 
            print ("F1, WW, gamma, u = %.4f, %.4f, %.4f, %.4f" % (F1, WW, gamma, u))
            print ("gamma = %.4f, closest distance of shadow to Earth center"%gamma)
            print ("Radius of umbral cone u = %.4f" % u)
            print ()
    
    EclipseDict = { "Object" : "Sun",
            "possible"  : SEpossible,
            "JDE"   : JDE,
            "JDEeclipse"   : JDEeclipse,
            "dt"    : JDEeclipse,
            "u"     : u,
            "gamma" : gamma,
            "omega" : omega,
            "event" : eclType,
            }
    if SEpossible and verbose: 
        print (EclipseDict)
        print ("JDEclipse = ", nicedatetime(hvfTime(JDEeclipse)))
        print ()


    return JDE, EclipseDict

# getMoonPhases(year, useUT)
#   returns:
#   newMoon
#   firstQuarter
#   fullMoon
#   lastQuarter
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
from operator import itemgetter

def getMoonPhases(year, useUT):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.)*12.3685  # relative moon phase at new year
    if (k0 == int(k0)): k = k0
    else: k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85            # Julian centuries after Epoch2000
    y  = year
    EventList = []
    EventDistList = []
    eventType = ["New Moon", "First Quarter", "Full Moon", "Last Quarter"]
    EventTypeList = []
    # starting values
    #print ("getMoonPhases start: k, T, y =", k, T, y)

    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    while y == year:
        #print ("y, k, ", y, k)
        dateList = ['', '', '', ''] # new,first q, full, last q
        distList = ['', '']         # new, full moon
        for i in range(4):
            #verb = k == int(k)      # verbose
            verb = False
            JDE, EclipseDict = getPhasesJDEbyK(k, verb)  # TT
            dtE = getGreenwichDate(JDE)      # TT
            deltaT = timedelta(seconds=getDeltaT(JDE))
            dtUTC = dtE - deltaT
            dtDisplay = dtE
            if useUT: dtDisplay = dtUTC

            # getMoonData expects UTC time
            (MoonRA, MoonDecl, MoonDict) = getMoonData(dtUTC, 0, False)
            MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "  "
            if MoonDist <= 360000. : suffix = "++" # Super Moon
            if MoonDist >= 405000. : suffix = "--" # Mini Moon
            (yyyy, dummy, dummy) = dtUTC.getDate()

            # Event
            Event = {'Object': MoonDict["Object"],  # Moon
                    'dt'    : dtDisplay,
                    'event' : eventType[i],
                    'mode'  : eventType[i],
                    'Dict'  : MoonDict,
                    }
            EventTypeList.append(Event)
            # just pass eclipse dict, adjust dt for UTC or TT
            if EclipseDict["possible"] :
                JDEeclipseDisplay = EclipseDict["JDEeclipse"]
                if useUT:
                    JDEeclipseDisplay = JDEeclipseDisplay - deltaT    
                EclipseDict["dt"] = JDEeclipseDisplay        

                EventTypeList.append(EclipseDict)
            #if (dt.year == year): 
            if (yyyy == year): 
                dateList[i] = niceshortdatetime(dtDisplay)
                if i == 0:
                    distList[0] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
                if i == 2:
                    distList[1] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
            k += 0.25
            T = k / 1236.85
            #print ("k, T, yyyy =", k, T, yyyy)
        EventList.append(dateList)
        if len(distList[0]+distList[1]) > 0: EventDistList.append(distList)
        T = k / 1236.85
        #y = int(T * 100 + 2000)
        y = floor(T * 100 + 2000)
        #print ("k, T, y =", k, T, y)
   # sort events by time dt in EventTypeList (solar eclipses)
    SortedEventTypeList = sorted(EventTypeList, key = itemgetter('dt'))

    return (EventList, EventDistList, SortedEventTypeList)

#########################################################      

# getMoonPhasesxxx(dtCivicTime, zone, verbose=False)
#   
#   returns:
#   newMoon
#   firstQuarter
#   fullMoon
#   lastQuarter
#   for about 30 days before and after dtCivicTime
#
from operator import itemgetter

def getMoonPhasesxxx(dtCivicTime, zone, verbose=False):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 
    #k0 = (yearxxx-2000.)*12.3685  # relative moon phase at given 
    k0 = Ty * 12.3685  # relative moon phase at given (47.2)
    k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85      # Julian centuries after Epoch2000
    kmin = k - 2
    kmax = k + 3
    EventList = []
    EventDistList = []
    eventType = ["New Moon", "First Quarter", "Full Moon", "Last Quarter"]
    EventTypeList = []
    # starting values
    #print ("getMoonPhases start: k, T, y =", k, T, y)
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    k = kmin
    for ki in range(kmin, kmax+1):
    #while y == year:
        #print ("y, k, ", y, k)
        dateList = ['', '', '', ''] # new,first q, full, last q
        distList = ['', '']         # new, full moon
        for i in range(4):
            #verb = k == int(k)      # verbose
            verb = False
            JDE, EclipseDict = getPhasesJDEbyK(k, verb)  # TT
            # only +- 3 weeks
            
            dtE = getGreenwichDate(JDE)      # TT
            deltaT = timedelta(seconds=getDeltaT(JDE))
            dtUTC = dtE - deltaT
            dtDisplay = dtE
            if useUT: dtDisplay = dtUTC

            # getMoonData expects UTC time
            (MoonRA, MoonDecl, MoonDict) = getMoonData(dtUTC, 0, False)
            MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "  "
            if MoonDist <= 360000. : suffix = "++" # Super Moon
            if MoonDist >= 405000. : suffix = "--" # Mini Moon
            (yyyy, dummy, dummy) = dtUTC.getDate()

            dtCT = getCivicTimefromUT(dtUTC, Zone)
            # Event
            #Event = {'Object': MoonDict["Object"],  # Moon
            Event = {'Object': "Moon Phases",
                    'dt'    : dtUTC,
                    'dtCivicTime' : dtCT,
                    'event' : eventType[i],
                    'mode'  : eventType[i],
                    'decl'  : MoonDecl,
                    'dist'  : MoonDist,
                    'suffix': suffix,
   
                    'Dict'  : MoonDict,
                    }
            EventTypeList.append(Event)
            # just pass eclipse dict, adjust dt for UTC or TT
            if EclipseDict["possible"] :
                JDEeclipseDisplay = EclipseDict["JDEeclipse"]
                if useUT:
                    JDEeclipseDisplay = JDEeclipseDisplay - deltaT    
                EclipseDict["dt"] = JDEeclipseDisplay        
                EclipseDict["dtCivicTime"] = \
                    getCivicTimefromUT(JDEeclipseDisplay, Zone)        
                EventTypeList.append(EclipseDict)
            if True:
                dateList[i] = niceshortdatetime(dtDisplay)
                if i == 0:
                    distList[0] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
                if i == 2:
                    distList[1] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
            k += 0.25
            T = k / 1236.85
        EventList.append(dateList)
        if len(distList[0]+distList[1]) > 0: EventDistList.append(distList)
        T = k / 1236.85
        #y = int(T * 100 + 2000)
        y = floor(T * 100 + 2000)
        #print ("k, T, y =", k, T, y)
   # sort events by time dt in EventTypeList (solar eclipses)
    SortedEventTypeList = sorted(EventTypeList, key = itemgetter('dt'))

    return (EventList, EventDistList, SortedEventTypeList)

#########################################################      
import sys

VersionMoon = "V2.1 hvf 2023-06-30 MoonPhases.py using ELP-2000/82 and VSOP87 Ephemerides"
VersionMoonShort = "V2.1 hvf 2023-06-30 using ELP-2000/82 and VSOP87 Ephemerides"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(VersionMoon)
def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    print ("Usage: %s [UTC|TT] fromyear-untilyear ..." % progname)
    sys.exit(0)

#print ("###########################################################")
print ()
print ("%s %s" % (sys.argv[0], VersionMoonShort))
print ("Solar Eclipses included in Event list")
print ()
for v in VersionList:
    print (v)
    
print ()

import re
# year-year or year is possible

# read years
useUT = True
foundOne = False
while len(sys.argv) > 1:
    arg = sys.argv[1]
    if arg.upper() == 'UT' or arg.upper() == 'UTC':
        useUT = True
        del sys.argv[1]
        continue
    elif arg.upper() == 'TT':
        useUT = False
        del sys.argv[1]
        continue
    else:
        yearList = re.split(r'-', arg)
        # we have to consider negative years as well!
        if len(yearList) == 2:  # startYear-lastYear, both positive 
            try:
                if len(yearList[0]) > 0:
                    startYear = int(yearList[0])
                    lastYear  = int(yearList[1])
                else:           # -startyear, negative, e.g. -1234
                    startYear = -int(yearList[1])
                    lastYear  = startYear
            except:
                Usage(sys.argv[0])
        elif len(yearList) == 3:  # -0003-0003 from -3 to +3
            try:
                startYear = -int(yearList[1])
                lastYear  = int(yearList[2])
            except:
                Usage(sys.argv[0])
        elif len(yearList) == 4:  # -0003--0002 from -3 to -2
            try:
                startYear = -int(yearList[1])
                lastYear  = -int(yearList[3])
            except:
                Usage(sys.argv[0])
        else:
            try:
                startYear = int(yearList[0])
                lastYear  = startYear
            except:
                Usage(sys.argv[0])


        if useUT: comment = "Universal Time Coordinated (UTC)"
        else: comment = "Terrestrial Time (TT)"
        for year in range(startYear, lastYear+1):
            print (" Moon Phases for the Year %4d" % year, comment)
            print (" -----------------------------", "-"*len(comment))
            print ()
# four phase events per line
            print (
"  New Moon           First Quarter      Full Moon          Last Quarter")
            (EventList, EventDistList, EventTypeList) = getMoonPhases(year, useUT)
            while len(EventList) > 0:
                printZeile("", EventList[0], "%17s", 79)
                del EventList[0]
            print()
# New moon, full Moon events incl. distance per line
            print (
"  New Moon                              Full Moon")
            print (
"    Date     Time      Distance           Date     Time      Distance")
            while len(EventDistList) > 0:
                print ("  %30s"%EventDistList[0][0], " "*6,"%30s"%EventDistList[0][1])
                del EventDistList[0]
            print()
# Moon and Sun Events
            print ("%-4s Events:" % EventTypeList[0]["Object"])
            for e in EventTypeList:
                print ("%-4s %-36s %s" % (e["Object"], e["event"],
                    nicedatetime(e["dt"])))
            print()
            foundOne = True
    del sys.argv[1]

if not foundOne: Usage(sys.argv[0])
