#! /usr/bin/env python3
# Julian to Greenwich and vice versa
# with my own hvfTime class, works also for BC dates and leap years
# hvf 16.09.2021 26.12.2021 11.11.2022 13.04.2023
# make hvfTime() hashable -- hvf 02.10.2023

from datetime import datetime, timedelta
from math import floor
import sys
import re
import dateutil.parser

VersionJulian = "V1.8 hvf 2023-10-02 HVFTime.py based on Meeus"


# JulianDay from year, month, day -  hvf 16.11.2021
def getJulianDayfromYMD(year, month, day, hour=0, minute=0, second=0, 
                        verbose = False):
    if verbose: 
        print ("getJulianDayfromYMD: ", year, month, day, hour, minute, second)
    day  += hour/24. + \
            minute/1440. + \
            second/86400.
    isGregor = False    # 04.10.1582 is last day before Gregorian calendar
    if year > 1582 or year == 1582 and (month > 10 or month == 10 and day > 4):
        isGregor = True
    if month <= 2:
        month += 12
        year -= 1
    B = 0
    # Gregorianische Korrektur
    if (isGregor): B = 2 - floor(year/100.) + floor(year/400.)
    JD = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) \
         + day + B - 1524.5
    # JD 0 is -4712-01-01.5 i.e. -4712-01-01 12h
    # N.B. JD 1524.5 is -4708-03-05 00h
    if verbose: print ("JD = ", JD)
    return JD


def roundhvfTime(hvfDT = None, roundTo = 3600):
    # Round a hvfTime object to anytime lapse in seconds
    # hvfDT: hvftime object, default now
    # roundTo: closest number of seconds to round to, default 1 hour.
    # Author: Harald von Fellenberg 2021
    # Problem: 2021-11-21 1223 returns 2021-11-21 12:22:60.000
    # hvf 2022-09-09
    verbose=False
    if not isinstance(hvfDT, hvfTime): hvfDT = hvfTime.now()
    h, m, s = hvfDT.getTime()
    seconds = h*3600 + m*60 + s
    fraction = seconds - floor(seconds)     # fractional seconds
    if verbose:
        print ("seconds, fraction, floor, rounding = ", seconds, fraction, floor(seconds))
    seconds = floor(seconds)
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    if verbose: print (rounding)
    return hvfDT + timedelta(0, rounding-seconds, -fraction*1e6)

# returns yyyymmdd hh for a hvfTime object, rounded
def nicehvfTimehh (hvfDT):
    hvfDTrounded = roundhvfTime(hvfDT)
    yy, mm, dd = hvfDTrounded.getDate()
    hh, dummy, dummy = hvfDTrounded.getTime()

    s = ''
    if yy >= 0:
        s = " %04d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    else:
        s = "%05d-%02d-%02d %02dh" % (yy, mm, dd, hh)
    return s

#######################################################################
# replaces astropy.time.Time since that does not support leap days 
# before 100BC
# values are represented in Julian days
# creator parameter can be:
# - a hvfTime object
# - a JD JulianDay object (float)
# - a datetime object
# - date and time as six values yy, mm, dd, [hh, MM, ss]
#   dd and ss may have fractions
# hvf 26.12.2021

class hvfTime:
    def __init__(self, *arg):
        # print ("__init__: arg = ", arg)
        # argument is a hvfTime object
        if isinstance(arg[0], hvfTime):
            JD = arg[0].getJD()
            self.__JD = JD
        # argument is a JD
        elif isinstance(arg[0], float) and len(arg) == 1:
            JD = arg[0]
            self.__JD = JD
        # argument is a datetime object
        elif isinstance(arg[0], datetime):
            dt = arg[0]
            yy = dt.year
            mm = dt.month
            dd = dt.day
            hh = dt.hour
            MM = dt.minute
            ss = dt.second + dt.microsecond*1e-6
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        # argument is yy mm dd [hh MM ss]
        elif (isinstance(arg[0], int) or isinstance(arg[0], float)) and \
            len(arg) >= 3:
            hh = 0; MM = 0; ss = 0
            yy = arg[0]; mm = arg[1]; dd = arg[2]
            if len(arg) > 3: hh = arg[3]
            if len(arg) > 4: MM = arg[4]
            if len(arg) > 5: ss = arg[5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD

        # argument is (yy mm dd [hh MM ss])
        elif isinstance(arg[0], tuple) or isinstance(arg[0], list):
            hh = 0; MM = 0; ss = 0
            yy = arg[0][0]; mm = arg[0][1]; dd = arg[0][2]
            if len(arg[0]) > 3: hh = arg[0][3]
            if len(arg[0]) > 4: MM = arg[0][4]
            if len(arg[0]) > 5: ss = arg[0][5]
            if not hh: hh = 0
            if not MM: MM = 0
            if not ss: ss = 0
            JD = getJulianDayfromYMD(yy, mm, dd, hh, MM, ss)
            self.__JD = JD
        else:
            print ("hvfTime instance not supported: ", arg[0])
            sys.exit(1)
            
    def getJD(self):
        return self.__JD

    def getDate(self, verbose = False):  # returns (yy, mm, dd), 
                                        # where dd may have fraction
        # works also for negative JDs! hvf 29.12.2021
        JD = self.__JD
        jd1 = JD + 0.5
        I = floor(jd1)      # integer part, may be negative
        F = jd1 - I         # fractional part, always positive
        if I > 2299160:     # 4.10.1582, last day before Gregorian calendar
            A = int((I - 1867216.25) / 36524.25)    # 28.2.400 AD 18:00h
            B = I + A - int(A/4) + 1
        else:   B = I       # Julian Calendar for earlier dates
        C = B + 1524
        D = floor((C-122.1)/365.25)
        E = floor(365.25 * D)
        G = floor((C-E)/30.6001)
        d = C - E + F - floor(30.6001 * G)
        if G < 13.5:    m = G - 1
        else:           m = G - 13
        if m > 2.5:     y = D - 4716
        else:           y = D - 4715

        if verbose:
            print ("getDate: JD, jd1, I, F = ", JD, jd1, I, F)
            print ("getDate: B, C, D, E, G = ", B, C, D, E, G)
            print ("getDate: y, m, d       = ", y, m, d)
        return (y, m, d)

    def getTime(self):      # seconds have 3 digi fractions (millisec)
        dummy, dummy, t = self.getDate()
        t -= floor(t)
        # t is fraction of day - convert to sec and round
        daysec = round(t * 86400., 3)
        hms = daysec / 3600.
        #hms = t*24.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 3)
        #s = int((ms - m) * 60)
        #print ("daysec, hms, h, ms, m, s = ", daysec, hms, h, ms, m, s)
    
        return (h, m, s)

    def __str__(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, hh, mm, ss)

    def getStr(self):      # YYYY-MM-DD hh:mm:ss, no fractions
        YY, MM, DD = self.getDate()
        DD = floor(DD)    # no fractions
        hh, mm, ss = self.getTime()
        daysec = round(ss + mm*60 + hh*3600, 0)
        hms = daysec / 3600.
        h = floor(hms)
        ms = (hms - h) * 60. + 1e-6 # avoid cutoff error
        m = floor(ms)
        s = round((ms - m) * 60, 0)

        #ss = round(ss, 0)
        return "%04d-%02d-%02d %02d:%02d:%02d" % (YY, MM, DD, h, m, s)

    # Ernesti, Python3, p. 380
    # hvfTime = hvfTime + timedelta
    # hvfTime = hvfTime + day   (float or int)
    def __add__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__add__: wrong type ", type(td))
            sys.exit(1)
        return hvfTime(self.__JD + sec/86400.)

    # hvfTime += timedelta
    def __iadd__(self, td):
        # add a timedelta object 
        try:    
            if isinstance(td, timedelta):
                sec = td.total_seconds()
            if isinstance(td, float) or isinstance(td, int):
                sec = td*86400.     # td is given in days
        except:
            print ("hvfTime.__iadd__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD += sec/86400.
        return self

    # hvfTime = hvfTime - timedelta
    # and   n = hvfTime - hvfTime       n in days
    def __sub__(self, td):
        # subtract a timedelta object 
        if isinstance (td, timedelta):
            sec = td.total_seconds()
            return hvfTime(self.__JD - sec/86400.)
        elif isinstance(td, hvfTime):
            # return JDs in days as difference between two hvfTime objects
            return self.__JD - td.getJD()
        else:
            print ("hvfTime.__sub__: wrong type ", type(td), td)
            sys.exit(1)

    # hvfTime -= timedelta
    def __isub__(self, td):
        # subtract a timedelta object 
        try:    
            sec = td.total_seconds()
        except:
            print ("hvfTime.__isub__: wrong type ", type(td), td)
            sys.exit(1)
        self.__JD -= sec/86400.
        return self

    # hvfTime < hvfTime
    def __lt__(self, other):
        return self.__JD < other.getJD()

    # hvfTime <= hvfTime
    def __le__(self, other):
        return self.__JD <= other.getJD()

    # hvfTime == hvfTime
    def __eq__(self, other):
        return self.__JD == other.getJD()

    # hash - makes hvfTime hashable, i.e. as key in dict or set 
    def __hash__(self):
        return hash(self.__JD)

    # hvfTime != hvfTime
    def __ne__(self, other):
        return self.__JD != other.getJD()

    # hvfTime > hvfTime
    def __gt__(self, other):
        return self.__JD > other.getJD()

    # hvfTime >= hvfTime
    def __ge__(self, other):
        return self.__JD >= other.getJD()

    def now():  # returns current time (h, m, s), uses datetime.now()
                # roundedd to integer seconds
        return hvfTime(datetime.now()).getTime()

    def today():    # returns current date (y, m, d), uses datetime.today()
                    # d has fractions
        return hvfTime(datetime.today()).getDate()
    
    def getDatetime(self):
        (yy, mm, dd) = self.getDate()
        dd = floor(dd)
        (HH, MM, SS) = self.getTime()
        SS = int(round(SS, 0))
        return datetime(yy, mm, dd, HH, MM, SS)
#######################################################################

# parse date time
# returns a hvfTime object
# hvf 19.11.2021 25.12.2021
# allowed patterns:
# 2021-11-21
# 2021-11-21 12:23:45.635
# 2021-11-21T12:23:45.635
# 2021-11-21 122345.635
# 2021-11-21 122345
# 2021-11-21 1223
# 2021-11-21 12
# -4712-1-1.5 
# 25.12.2021 12:34:56.789
# 1.1.2022
# 2451545.0 (a Julian Day)
# works for dates between -9999-01-01 and 9999-12-31

def parseDateTime(s, verbose = False):
    HH = MM = SS = 0
    # check if two parts separated by blank or T
    if verbose: print ("parseDateTime input: ", s)
    s = s.strip()
    dpart = s
    tpart = None
    parts = re.split (r"[\sT]", s)
    if len(parts) >= 2:
        dpart = parts[0]    # date part
        tpart = parts[1]    # time part
        if verbose:
            print ("date part = ", dpart)
            print ("time part = ", tpart)

    # parse date  2021-12-24 or  4712-1-1.5
    if re.fullmatch(r'\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        ymd = re.split(r'-', dpart)
        yy = ymd[0]
        mm = ymd[1]
        dd = ymd[2]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date 24.12.2021 or 1.1.2022
    elif re.fullmatch(r'\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        ymd = re.split(r'\.', dpart)
        yy = ymd[2]
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -4712-01-01.5
    elif re.fullmatch(r'-\d{4}-\d{1,2}-\d{1,2}(\.\d*)?', dpart):
        symd = re.split(r'-', dpart)
        yy = -int(symd[1])
        mm = symd[2]
        dd = symd[3]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    # parse date -24.12.2021 or -1.1.2022
    elif re.fullmatch(r'-\d{1,2}\.\d{1,2}\.\d{4}', dpart):
        symd = re.split(r'-', dpart)
        ymd = re.split(r'\.', symd[1])
        yy = -int(ymd[2])
        mm = ymd[1]
        dd = ymd[0]
        if verbose:
            print ("yy, mm, dd = ", yy, mm, dd)

    else:
        print ("parseDateTime format unknown: ", s)
    # parse date  2021-12-24 or  4712-1-1.5
    # parse date 24.12.2021 or 1.1.2022
    # parse date -4712-01-01.5
    # parse date -24.12.2021 or -1.1.2022
        print ("Allowed date formats include 2021-12-24, 4712-1-1.5,")
        print ("    24.12.2021, 1.1.2022, -4712-01-01.5, -24.12.2021, -1.1.2022")
        print ("Allowed range of years is from -9999 to 9999")
        sys.exit(1)

    # parse time using dateutil.parser.isoparse - this is easier
    # than checking all the different possible input formats
    if tpart:
        deltaT = dateutil.parser.isoparse("2021-12-24 " + tpart) \
                - datetime(2021, 12, 24)
        Secs = round(deltaT.total_seconds(), 3)    # avoid rounding error
        HH = Secs // 3600
        Secs -= HH * 3600
        MM = Secs // 60
        Secs -= MM * 60
        SS = Secs
        if verbose:
            print ("HH, MM, SS = ", HH, MM, SS)

    #return hvfTime object
    return hvfTime(int(yy), int(mm), float(dd), int(HH), int(MM), float(SS))

#######################################################################
# catculate weekday for any date from -9999-01-01 to 9999-12-31
# Meeus 1991 p. 65

def getDayofWeek(dt):
    name =["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday"]
    hvfdt = hvfTime(dt)
    (yy, mm, dd) = hvfdt.getDate()
    hvfdtMidnight = hvfTime(yy, mm, int(dd), 0, 0, 0.0)
    JD0 = hvfdtMidnight.getJD()
    weekday = int(JD0 + 1.51) % 7
    return name[weekday]

#######################################################################

#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-

# Der folgende Python-Code fuer Sonne und Mond 
# basiert auf der Portierung des java-Scripts von Arnold Barmettler
# http://lexikon.astronomie.info/java/sunmoon/sunmoon.html
#
#     Entfernen Sie folgende Informationen auf keinen Fall: / 
#     Do not remove following text:
#     Source code by Arnold Barmettler, www.astronomie.info / www.CalSky.com
#     and Harald von Fellenberg, www.hvf.ch
#     based on algorithms by Peter Duffett-Smith's great and easy book
#     'Practical Astronomy with your Calculator or Spreadsheet'.
#     and Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Totale Ueberarbeitung durch Harald von Fellenberg, Dezember 2020
#   Portierung auf Python 3
#   Ergaenzung fuer Mitternachtssonne, Polarnacht
#   Ergaenzungen fuer Daemmerung, Hoehe der Sonne beim Transit tags und nachts
#   Ergaenzung mit Zeitgleichung, wahre und mittlere Ortszeit
#   Azimut für Sonnenaufgang, Sonnenuntergang
#   Liste einiger Städte mit ihren Koordinaten
#   Korrekte Berechnung in allen Zeitzonen
#   Behandlung aller Planeten
#
#   Die Behandlung von Sonne und Mond wurde ersetzt durch die Algorithmen in
#   Jean Meeus, Astronomical Algorithms, Willmann-Bell cop., Richmond,
#       Virginia, 1991
#
#   Die Daten aller Planeten beruhen auf VSPO87
#   Pierre Bretagnon, Gerard Francou
#   Bureau des Longitudes
#   77, Avenue Denfert-Rochereau
#   F75014, Paris, France
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/
#
#   Die Resultate wurden mit denen auf https://www.timeanddate.com/ verglichen,
#   die Uebereinstimmung ist gut innerhalb von wenigen Minuten.
#   Beim Mond ist stets das Datum zu beachten!
#   Liegt die lokale Zeit zwischen Mitternacht und "zone", so wird für den Mond
#   zum Teil auf den Vortag zurückgegriffen. Dies ist besonders für Orte mit
#   grosser Zeitzonendifferenz wichtig, wie Sydney oder Honolulu.
#   Für Orte in den Polarregionen wird Mitternachtssonne und Polarnacht
#   korrekt ausgegeben, ausser am Tag, wo die Sonne unter- oder aufgeht.
#   Ebenso fuer den Mond.
#
#   Anmerkung: die Webseiten astronomie.info sowie calsky.com existieren
#   seit Oktober 2020 nicht mehr.
#
#   Alle 8 Planeten sind ebenfalls behandelt, gemaess den Algorithmen in
#   Jean Meeus, Astronomical Algorithms.
#   Die Daten zu den 8 Planeten stammen aus den VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Version 1.0
#   Version 2.0
#   Version 3.0
#   Version 3.1
#   Version 3.4
#   Version 4.0 inkl Sun-Events, Moon-Events
#   Version 5.1 
#   Version 5.4 add getTimezone()
Version = "V5.4 hvf 2024-05-08 SunHead.py using VSOP87 ephemerides and HVFTime"
if 'VersionList' not in globals().keys(): VersionList = []
VersionList.append(Version)

import time
from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import argparse
import sys
import dateutil.parser
import os

# alle Konstanten sind auf die Epoche 2010.0, JD=2455196.5 bezogen
# falls nicht anders dokumentiert.

def getEpoch2010():
    return 2455196.5

# https://www.geeksforgeeks.org/how-to-print-to-stderr-and-stdout-in-python/
def print_to_stderr(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stderr) 

def print_to_stdout(*a): 
    # Here a is the array holding the objects 
    # passed as the arguement of the function 
    print(*a, file = sys.stdout) 

# produces a datetime object output string without microseconds
# seconds get rounded

def nicedatetime(dt):       # yyyy-mm-dd HH:MM:SS
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        string = dt.getStr()
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%Y-%m-%d %H:%M:%S")
    return string

def niceshortdatetime(dt):  # dd.mm.yy HH:MM:SS
    # only meaningful for known century
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        (yy, mm, dd) = dt.getDate()
        yy = int(abs(yy)%100)    # only last 2 ciphers, also for negative yyyy
        (hh, MM, ss) = dt.getTime()
        # round seconds! use %f fmt - hvf 13.04.2023
        #string = "%02d.%02d.%02d %02d:%02d:%02d" % (dd, mm, yy, hh, MM, ss)
        string = "%02d.%02d.%02d %02d:%02d:%02.0f" % (dd, mm, yy, hh, MM, ss)
    else: 
        dt += timedelta(seconds=0.5)
        string = dt.strftime("%d.%m.%y %H:%M:%S")
    return string

def nicedateshorttime(dt):  # yyyy-mm-dd HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=30)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        (h, m, s) = dt.getTime()
        string = "%4d-%02d-%02d %02d:%02d" % (y, mm, dd, h, m)
    else:
        string = dt.strftime("%Y-%m-%d %H:%M")
    return string

def niceshortdateshorttime(dt):  # dd.mm.yy HH:MM (rounded)
    if isinstance(dt, type(None)): return None
    if isinstance(dt, hvfTime):
        dtrounded = roundhvfTime(dt, 60)
        #print ("dt, dtrounded = ", dt, dtrounded)
        (y, mm, dd) = dtrounded.getDate()
        yy = abs(y)%100
        dd = floor(dd)
        (h, m, s) = dtrounded.getTime()
        string = "%02d.%02d.%02d %02d:%02d" % (dd, mm, yy, h, m)
    else:
        dt += timedelta(seconds=30)
        string = dt.strftime("%d.%m.%y %H:%M")
    return string

def nicedate(dt):           # yyyy-mm-dd
    if isinstance(dt, type(None)): return None
    dt += timedelta(seconds=0.5)
    if isinstance(dt, hvfTime):
        (y, mm, dd) = dt.getDate()
        dd = floor(dd)
        string = "%04d-%02d-%02d" % (y, mm, dd)
    else:
        string = dt.strftime("%Y-%m-%d")
    return string

def nicetime(dt):           # HH:MM:SS
    if isinstance(dt, type(None)): return None
    try:
        if isinstance(dt, hvfTime):
            (h, m, s) = dt.getTime()
            s =round(s, 0)
            string = "%02d:%02d:%02d" % (h, m, s)
        else:
            dt += timedelta(seconds=0.5)
            string = dt.strftime("%H:%M:%S")
    except:
        string = ' '*8
    return string

def nicetimedelta(dt):      # +-HH:MM:SS
    if isinstance(dt, type(None)): return None
    #dt += timedelta(seconds=0.5)
    # treat also -00:00:05 correctly! hvf 16.06.2021
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60

    string = "%1s%2d:%02d:%02d" % (sign, hours, minutes, seconds)
    return string

def nicedeltadayhhmm(dt):   # +-days HH.MM
    if isinstance(dt, type(None)): return None
    if isinstance(dt, timedelta):
        totalsecs = dt.total_seconds()
    if isinstance(dt, float):
        totalsecs = dt*86400.
    sign = ' '
    if totalsecs < 0: sign = '-'; totalsecs = abs(totalsecs)
    days  = int(totalsecs/86400.); totalsecs -= days*86400.
    hours = int(totalsecs/3600.); totalsecs -= hours*3600.
    minutes = int(totalsecs/60.); totalsecs -= minutes*60.
    seconds = round(totalsecs, 0)
    if seconds >= 60:
        minutes += 1; seconds -= 60
    if minutes >= 60:
        hours += 1; minutes -= 60
    if hours >= 24:
        days += 1; hours -= 24

    string = "%1s%d days %02d:%02d" % (sign, days, hours, minutes)
    return string

def niceminsec(secs):       # +- M:SS  for Time Equation - Zeitgleichung
    if isinstance(secs, type(None)): return None
    sign = ' '
    if secs < 0: sign = '-'; secs = abs(secs)
    minutes = int(secs/60.); secs -= minutes*60.
    seconds = round(secs, 0)
    if seconds >= 60: 
        minutes += 1; seconds -= 60

    string = "%1s%d:%02d" % (sign, minutes, seconds)
    return string

# printZeile(kopf, Liste, format, zeile=79)
# schreibt eine Zeile der max. Laenge zeile mit kopf und den
# Werten aus Liste, formattiert je mit format.
# Padding vor jedem Listenelement
# Die  formattierten Werte muessen gleich lang sein.
# hvf 31.12.2020

def printZeile(kopf, Liste, format, zeile=79):
    rest = zeile
    string = kopf
    rest -= len(string)
    n = len(Liste)      # Anzahl Werte
    test = format % Liste[0]
    testlen = len(test)
    padlen = int((rest - n*testlen)/n)
    padding = ' '*padlen

    for item in Liste:
        string += padding
        string += format % item

    print (string)
    return string

# modulus in Python has always the same sign as b
# Mod(-3, 24) gives 21
def Mod(a, b):
    return (a % b)

# return integer value, closer to 0
def Int(x): return int(x)

# return fractional part, closer to 0 (i.e. may be nagative)
def Frac(x): return (x - int(x))

# Winkelfunktionen fuer Grad Eingaben
def Sin(x):
    return sin(radians(Mod(x, 360.)))

def Cos(x):
    return cos(radians(Mod(x, 360.)))

def Tan(x):
    return tan(radians(Mod(x, 360.)))

def Deg(x):
    return degrees(x)

# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D


#####################################################################
# Quelle: https://de.wikipedia.org/wiki/Julianisches_Datum
# abgerufen am 20.10.2019
# Name geaendert in getJulianDay (siehe Meeus, chapter 7) - hvf 24.03.2023

# Calculate Julian day for all Gregorian and Julian dates
# Greenwich date is supplied as a hvfTime object

def getJulianDay(dtGreenwichDate, isGregor=True ):
    hvfdt = hvfTime(dtGreenwichDate)
    return hvfdt.getJD()

################################################################
# calculate Greenwich date from Julian date
# Ref. Practical Astronomy (see above), p.11
# returns a hvfTime object for midnight 

def getGreenwichDate(JD):

    return hvfTime(float(JD))

#   HMSDH convert hours, minutes, seconds to decimal hours
def HMSDH(hh, mm, ss):  # hh may be negative
    habs = abs(hh)
    habs += mm/60. + ss/3600.
    h = habs
    if hh < 0: h = -habs
    return h

#   DHHMS converts decimal hours into hours, minutes, seconds
#   seconds may have fractions, no rounding
def DHHMS(h):   # h may be negative, then only hours will be negative
    sign = 1
    if h < 0: sign = -1
    habs = abs(h)
    hh = int(habs)
    mm = int(Frac(habs)* 60.)
    ss = Frac(habs*60)*60.
    hh = hh * sign
    return (hh, mm, ss)

#   returns HH:MM and a rounded value of h
#   h must be positive or None or ''
def HHMM(hh):
    if (hh == None or hh == ''):
        return('--')
    m = Frac(hh) * 60.0
    h = Int(hh)
    if (m >= 59.5):
        h += 1
        m -= 60.0
    m = int(round(m))
    string = "%02d:%02d = " % (h, m) + str(round(hh, 4))
    return (string)

#   returns HH:MM:SS and a rounded value of h
#   h must be positive or None or ''
def HHMMSS(hh):
    if (hh == 0):
        return('--')
    m = Frac(hh) * 60
    h = Int(hh)
    s = Frac(m) * 60.0
    m = Int(m)
    if (s >= 59.5):
        m += 1
        s -= 60.0
    if (m >= 60):
        h += 1
        m -= 60
    s = int(round(s))
    string = "%02d:%02d:%02d = " % (h, m, s) + str(round(hh, 5))
    return (string)

def hmsstring(hour):
    # hour may also be negative, e.g. if used to convert declination
    (hh, mm, ss) = DHHMS(hour)
    str = "%d:%02d:%02d" % (hh, mm, ss)
    return str


#   local civic time to Universal Time
#   returns a hvfTime object 

def getUTfromCivicTime(dtLocalCivicTime, zone):    
    # dtLocalCivicTime is a hvfTime object
    # zone is time zone offset in hours, positive going East
    dtUT = dtLocalCivicTime - timedelta(hours=zone)
    return dtUT

def getCivicTimefromUT(dtUT, zone):
    dtLocalCivicTime = dtUT + timedelta(hours=zone)
    return dtLocalCivicTime

#   UTC to mean local time 
#   dtUT: Universal Time hvfTime object
#   longitude: Eastern longitude, degrees
#   returns a datetime object, rounded to seconds

def getMeanLocalTime(dtUT, longitude):
    dtmeanLT = dtUT + timedelta(seconds=round(240.*longitude, 0))
    return dtmeanLT

def getUTfromMeanLocalTime(dtmeanLT, longitude):
    dtUT =  dtmeanLT - timedelta(seconds=round(240.*longitude, 0))
    return dtUT

#   datetime.datetime or datetime.time to decimal hours
#   dttime is a hvfTime object

def getDecimalHour(dttime):
    (hour, min, sec) = hvfTime(dttime).getTime()
    return hour    

# calculate horizontal coordinates (Azimut, Altitude)
# from ecliptic coordinates (Hour angle, Declination)
#   Parameters:
#   Hour    Hour angle counted from meridian to West, degrees
#   Decl    Declination, degrees
#   latitude    degrees
#   Returns:
#   azimut  counted from North, degrees
#   alt     altitude above horizon, degrees
#   x, y, z Cartesian coords in horizontal system, x pointing South, y West
#   https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen

def getAziAltfromHourDecl(Hour, Decl, latitude):
    # Hour angle and Declination into azimut and altitude
    azimut = atan2(Sin(Hour), Cos(Hour)*Sin(latitude)-Tan(Decl)*Cos(latitude))
    alt    = asin(Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour))
    azimut = Mod(degrees(azimut)+180., 360.)
    alt    = degrees(alt)
    # Cartesian coords in horizon system
    x = Cos(alt)*Cos(azimut-180.)
    y = Cos(alt)*Sin(azimut-180.)
    z = Sin(alt)
    # dito calculated from  Hour, Decl, latitude
    xx = -Cos(latitude)*Sin(Decl)+Sin(latitude)*Cos(Decl)*Cos(Hour)
    yy = Cos(Decl)*Sin(Hour)
    zz = Sin(latitude)*Sin(Decl)+Cos(latitude)*Cos(Decl)*Cos(Hour)
    xyzDict = {
        "x":    x,
        "y":    y,
        "z":    z, 
        "xx":   xx,
        "yy":   yy,
        "zz":   zz, }
    return (azimut, alt, xyzDict)

# Zodiac signs for Sun and Moon
#   lon ecliptic longitude, degrees
def getSign(lon):
    Signs_de = ("Widder", "Stier", "Zwillinge", "Krebs", "Löwe", "Jungfrau", \
        "Waage", "Skorpion", "Schütze", "Steinbock", "Wassermann", "Fische")
    Signs_en = ("Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo", \
        "Libra", "Scorpio", "Sagittarius", "Capricorn", "Aquarius", "Pisces")
    return (Signs_en[int(floor(lon / 30))])

########################################################################
#
# getNutation(dtCivicTime, zone)
# Nutation in ecliptic longitude and ecliptic latitude
# see Meeus, Astronomical Algorithms, chapter 21

def getNutation(dtCivicTime, zone, verbose=False):
    # Epoch is J2000.0
    # JD 2451545
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    T = (JD - 2451545.0)/36525. # Centuries after the Epoch, 21.1

    # Moon Ascending Node
    Omega = 125.04452 - 1934.136261 * T + 0.0020708 * T*T + T*T*T/450000.
    Omega = Mod(Omega, 360.)
    O = radians(Omega)

    # accuracy of 0.5 arcsec in longitude, 0.1 arcsec in latitude
    # Sun Mean Longitude
    LmeanSun = 280.4665 + 36000.7698 * T
    LmeanSun = Mod(LmeanSun, 360.)
    LS = radians(LmeanSun)

    # Moon Mean Longitude
    LmeanMoon = 218.3165 + 481267.8813 * T
    LmeanMoon = Mod(LmeanMoon, 360.)
    LM = radians(LmeanMoon)

    # Effect on ecliptic longitude and latitude, arcsecs
    deltaELong = -17.20*sin(O) -1.32*sin(2.*LS) -0.23*sin(2*LM) +0.21*sin(2*O)
    deltaELat  = 9.20*cos(O) + 0.57*cos(2.*LS) + 0.10*cos(2*LM) -0.09*cos(2*O)

    # obliquity of the ecliptic, 21.2
    eps0 = 23+26/60.+21.448/3600. - 46.8150/3600.*T - 0.00059/3600.*T*T \
            +0.001813/3600*T*T*T
    eps = eps0 + deltaELat/3600.

    if verbose:
        print ("getNutation")
        print ("JD = ", JD)
        print ("T = ", T)
        print ("LmeanSun = ", LmeanSun)
        print ("LmeanMoon = ", LmeanMoon)
        print ("getNutation deltaELong (arcsec) = %.2f" % deltaELong)
        print ("getNutation deltaELat  (arcsec) = %.2f" % deltaELat )
        print ("eps0 (degrees) = %.6f" % eps0, ", ", HHMMSS(eps0))
        print ("eps  (degrees) = %.6f" % eps, ", ", HHMMSS(eps) )

    # return values in degrees
    return (deltaELong/3600., deltaELat/3600., eps)

#################################################################
#   Funktionen für die Berechnung der Zeitgleichung

def getSunStufffromTT(TT, verbose = False):
    # TT Tage nach Epoch2010 inkl. DeltaT Korrektur
    # returns RA, epsilon in degrees

    def atanLage(x, Lage):    # arctangens mit Lageparameter, alle Werte in Grad
        # siehe https://en.wikipedia.org/wiki/Equation_of_time
        # Resultat in Grad
        xrad = radians(x)
        Lagerad = radians(Lage)
        y = atan(xrad)
        valuerad = y + pi * round((Lagerad-y)/pi, 0)
        value = degrees(valuerad)
        return value
    
    def getE(t):    # numerische Exzentrizität zur Zeit t
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        e = 0.01670491 - 4.208E-5*T - 1.26E-7*T*T
        return e
    
    def getEps(t):  # Schiefe der Ekliptik in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
        return eps
    
    def getOmegabar(t):     # Länge des Perihels in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        omegabar = 283.11243767 + 1.720171*T + 0.000452778*T*T
        return omegabar
    
    def getL(t):    # mittlere Länge der Sonne in Grad
        # t in Tage nach Epoch2010
        # Resultat in Grad
        T = t / 36525
        L = 279.557285 + 36000.769585*T + 0.0003025*T*T
        return L
    
    def getC(e, M):    # Mittelpunktsgleichung equation of the center
        # berechnet wahre minus mittlere Anomalie als Fkt der Exzentrizität
        # und der mittleren Anomalie in dritter Näherung in e
        # e numerische Exzentrizität
        # M mittlere Anomalie der Sonne in Grad
        # Resultat in Grad
        # Danby, Fundamentals of Celestial Mechanics, p.438
        Mrad = radians(M)
        Crad = (2*e-e*e*e/4)*sin(Mrad) + (1.25*e*e -11./24.*e**4)*sin(2*Mrad)\
             + 13./12.*e*e*e*sin(3*Mrad) + 103./96.*e**4*sin(4*Mrad)
        C = degrees(Crad)
        return C

    t = TT
    e = getE(t)                 # Exzentrizität der Erdbahn, Zeit t
    eps = getEps(t)             # Schiefe der Ekliptik in Grad, Zeit t
    epsRad = radians(eps)       # Schiefe der Ekliptik in Radians
    omegabar = getOmegabar(t)   # Länge des Perihels in Grad, Zeit t
    L = getL(t)                 # mittlere Länge der Sonne in Grad, Zeit t
    Lhaupt = Mod(L, 360)        # Hauptwert der Länge der Sonne
    M = Lhaupt - omegabar       # Mittlere Anomalie der Sonne in Grad, Zeit t
    Mhaupt = Mod(M, 360)        # Hauptwert der mittleren Anomalie
    C = getC(e, Mhaupt)         # Mittelpunktsgleichung, wahre minus mittlere A
    Lwahr = L + C               # wahre Länge der Sonne in Grad, Zeit t
    LwahrRad = radians(Lwahr)   # wahre Länge der Sonne in Radians
    Lwahrhaupt = Mod(Lwahr, 360)# Hauptwert der wahren Länge
    LwahrHauptRad = radians(Lwahrhaupt)  # dito, in Radians


    # Rektaszension der Sonne in Funktion der wahren Länge der Sonne und der 
    # Schiefe der Ekliptik
    alpha = atanLage(degrees(tan(LwahrRad)*cos(epsRad)), Lwahr)
    alphahaupt = atanLage(degrees(tan(LwahrHauptRad)*cos(epsRad)), Lwahrhaupt)

    Dict = { "Object"   : "SunStuff",
            "TT"        : t,            # Tage nach Epoch2010 inkl. deltaT
            "MPEq"      : C,            # wahre - mittlere Anomalie
            "omegabar"  : omegabar,     # Länge des perihels in Grad, Zeit t
            "M"         : M,            # mittlere Anomalie, Grad
            "nu"        : C+M,          # wahre Anomalie
            "Long"      : Lwahrhaupt,   # wahre Länge der Sonne, Grad
            "eps"       : eps,          # Schiefe der Ekliptik, Grad
            "exc"       : e,            # Exzentrizität der Erdbahn
            }
    if verbose:
        print ("Numerische Exzentrizitaet e = ", "%.6f" % e)
        print ("Omegabar, Laenge des Perihels = ", "%.3f Grad" % omegabar)
        print ("Lhaupt, Hauptwert Mittlere Laenge = ", "%.3f Grad" % Lhaupt)
        print ("M Mittlere Anomalie = ", "%.3f Grad" % M)
        print ("Mhaupt, Hauptwert Mittlere Anomalie = ", "%.3f Grad" % Mhaupt)
        print ("C Mittelpunktsgleichung = ", "%.3f Grad" % C)
        print ("Alpha Wahre Anomalie = ", "%.3f Grad" % alpha)
    return (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict)

#######################################################################    
def Zeitgleichung(JD, verbose=False):
    # berechnet die Zeitgleichung in ganzen Sekunden
    # JD Julianisches Datum
    # verbose aktiviert zusätzlichen Output
    # Resultat in Sekunden ohne Rundung
    # Epoche konsistent 2010.0, 31.12.2009 0h UT, JD 2455196.5
    dtEpoch2010 = datetime(2009, 12, 31)
    tbase = getJulianDay (dtEpoch2010)
    t = JD - tbase
    TT = JD - tbase + getDeltaT(JD)/86400.

    (alphahaupt, eps, L, Lwahr, Lwahrhaupt, Dict) = getSunStufffromTT(TT, verbose)

    # Zeitgleichung erste Ursache Exzentrizität der Erdbahn (epsilon = 0)
    ZGstar0 = L - Lwahr         # in Grad
    ZG0 = 240. * ZGstar0        # in Sekunden

    # zweite Ursache Anteil Schiefe der Ekliptik
    ZGstareps = Lwahrhaupt - alphahaupt # in Grad
    ZGeps = 240. * ZGstareps            # in Sekunden

    # Summe der beiden Ursachen
    ZGstar = ZGstar0 + ZGstareps        # in Grad
    ZGsecs = ZG0 + ZGeps                # in Sekunden

    # ZGmin, ZGsec for printing purposes 
    ZGstr = niceminsec(ZGsecs)
    if verbose:
        print ("Epoch2010 = ", dtEpoch2010)
        print ("Epoch Julian Date = ", tbase)
        print ("TT = ", TT)
        print ("Schiefe der Ekliptik eps = ", "%.3f Grad" % eps)
        print ("L, Mittlere Laenge = ", "%.3f Grad" % L)
        print ("Lwahr Wahre Sonnenlaenge = ", "%.3f Grad" % Lwahr)
        print ("Lwahrhaupt Hauptwert Wahre Sonnenlaenge = ", "%.3f Grad" \
                    % Lwahrhaupt)
        print ("Alphahaupt Wahre Anomalie = ", "%.3f Grad" % alphahaupt)
        print ("ZGstar0 = ", "%.3f" % ZGstar0, " Grad")
        print ("ZG0 = ", "%.2f" % ZG0, " Sekunden")
        print ("ZGstareps = ", "%.3f" % ZGstareps, " Grad")
        print ("ZGeps = ", "%.2f" % ZGeps, " Sekunden")
        print ("ZGstar = ", "%.3f" % ZGstar, " Grad")
        print ("ZGsecs = ", "%.2f" % ZGsecs, " Sekunden")
        print ("ZG = %s" % ZGstr)

    return (ZGsecs, ZGstr, Dict)

def getZeitgleichung(JD, verbose=False):
    # returns a timedelta object and two int's for printing
    zgsecs, ZGstr, Dict = Zeitgleichung(JD, verbose)
    return (timedelta(seconds=zgsecs), ZGstr, Dict)

# Greenwich Mean Sidereal Time from Julian Date

def getGMSTfromJD(JD):
    # Meeus chapter 11 p. 83
    T = (JD - 2451545.0) / 36525.   # (11.1)
    theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) \
            + 0.000387933 * T*T - T*T*T/387110000   # (11.4)
    # normalize to 360 degrees
    theta = Mod(theta0, 360.)
    # add to midnight
    JD0 = floor(JD-0.5) + 0.5
    Gdate = hvfTime(JD0 + theta/360.)
    return Gdate

# Local Mean Sideral Time from GMST, longitude
# returns a datetime object

def getLMSTfromGMST(gmst, longitude):
    # gmst datetime object
    # longitude in degrees
    # returns a datetime object
    lmst = gmst + timedelta(hours=longitude/15.)
    return lmst

# Greenwich Mean Sideral Time from LMST, longitude
# returns a datetime object

def getGMSTfromLMST(lmst, longitude):
    # lmst datetime object
    # longitude in degrees
    # returns a datetime object
    #gmst = datetime.time(hour=lmst) - timedelta(hours=longitude/15.)
    gmst = lmst - timedelta(hours=longitude/15.)
    return gmst
    

# deltaT, the difference between Terrestrial Time TT and
# Universal Time 
# see https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
# for Julian Date JD

def getDeltaT(JD): 
    # convert JD into year, month, day
    (year, month, day) = hvfTime(JD).getDate()
    # We define the decimal year "y" as follows:
    y = year + (month - 0.5)/12

    if year < -500:
        u = (year-1820.)/100.
        deltaT = -20 + 32*u*u
    elif year < 500:
        u = y/100.
        deltaT = 10583.6 - 1014.41*u + 33.78311*u*u - 5.952053*u**3 \
                - 0.1798452*u**4 + 0.022174192*u**5 + 0.0090316521*u**6
    elif year < 1600:
        u = (y-1000.)/100.
        deltaT = 1574.2 - 556.01*u + 71.23472*u*u + 0.319781*u**3 \
                - 0.8503463*u**4 - 0.005050998*u**5 + 0.0083572073*u**6
    elif year < 1700:
        t = y - 1600.
        deltaT = 120. -0.9808*t - 0.01532*t*t + t**3/7129
    elif year < 1800:
        t = y - 1700.
        deltaT = 8.83 + 0.1603*t - 0.0059285*t*t + 0.00013336*t**3 \
                - t**4/1174000.
    elif year < 1860:
        t = y - 1800.
        deltaT = 13.72 - 0.332447*t + 0.0068612*t*t + 0.0041116*t**3 \
                - 0.00037436*t**4 + 0.0000121272*t**5 \
                - 0.0000001699*t**6 + 0.000000000875*t**7
    elif year < 1900:
        t = y - 1860.
        deltaT = 7.62 + 0.5737*t - 0.251754*t*t + 0.01680668*t**3 \
                - 0.0004473624*t**4 + t**5/233174.
    elif year < 1920:
        t = y - 1900.
        deltaT = -2.79 + 1.494119*t - 0.0598939*t*t + 0.0061966*t**3 \
                -0.000197*t**4
    elif year < 1941:
        t = y - 1920.
        deltaT = 21.20 + 0.84493*t - 0.076100*t*t + 0.0020936*t**3
    elif year < 1961:
        t = y - 1950.
        deltaT = 29.07 + 0.407*t - t*t/233. + t**3/2547.
    elif year < 1986:
        t = y - 1975.
        deltaT = 45.45 + 1.067*t - t*t/260. - t**3/718.
    elif year < 2005:
        t = y - 2000.
        deltaT = 63.86 + 0.3345*t - 0.060374*t*t + 0.0017275*t**3 \
                + 0.000651814*t**4 + 0.00002373599*t**5
    elif year < 2050:
        t = y - 2000.
        deltaT = 62.92 + 0.32217*t + 0.005589*t*t
    elif year < 2150:
        deltaT = -20. + 32*((y-1820.)/100)**2 - 0.5628*(2150.-y)
    else:   # after 2150
        u = (year-1820.)/100.
        deltaT = -20. + 32.*u*u
        # https://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
        # small correction beyond 2150 (only needed for Perihelion calc's)
        c = -0.000012932 * (y - 1955.)**2
        deltaT += c

    return deltaT

def getEps(t):  # Schiefe der Ekliptik in Grad
    # t in Tage nach Epoch2010
    # Resultat in Grad
    T = t / 36525
    eps = 23.4379916 - 0.0130042*T - 1.5875E-8*T*T + 5.0278E-7*T*T*T
    return eps


#######################################################################
# Jean Meeus, Astronomical Algorithms, Willmann-Bell, Richmond VA, 1991
#
# hvf 03.12.2020

# quadratic interpolation
# xi 3 Stuetzstellen, yi 3 Funktionswerte
# Meeus p. 25
def interp(xi, yi, x):
    a10 = (yi[1]-yi[0])/(xi[1]-xi[0])
    a21 = (yi[2]-yi[1])/(xi[2]-xi[1])
    a210 = (a21-a10)/(xi[2]-xi[0])
    res = yi[0] + a10*(x-xi[0]) + a210*(x-xi[0])*(x-xi[1])
    return res

#############################################################################
#! /usr/bin/env python3
# getTimezone(verbose)
# returns (server) timezone using /etc/timezone, timedatectl, or a default
# hvf 07.05.2024

import re
import os
import subprocess

def getTimezone(verbose = False):
    # default timezone is Europe/Zurich
    timezone = "Europe/Zurich"
    # easiest: use timezone file and read its content§
    TimezoneFile = "/etc/timezone"
    try:
        file = open(TimezoneFile, "r") 
    except (Exception, FileNotFoundError, OSError) as  e:
        #print ("Error ", e, "Thrown.")
        #print ("file not found: ", filename, "\n")
        pass

    else:
        for line in file:
            #print ("Line: ", line)
            #length = len(line); print ("Length =", length)
            line = line.strip()
            #print ("Line after stripping: ", line)
            #length = len(line); print ("Length =", length)
            timezone = line
            if verbose:
                print ("Timezone from ", TimezoneFile," is: ", timezone)
            return timezone

    # next method: use timedatectl
    # Define the command you want to run as a list of arguments
    cmd = ["/usr/bin/timedatectl", "show", "-p", "Timezone"]

    # Use subprocess.Popen to start the process
    try:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    except Exception as e:
        #print ("Error: ", e, "\n")
        pass
    else:
        # Communicate with the process and get its output and errors
        output, errors = process.communicate()

        # Print the output and errors (if any)
        #print("Output:", output)
        #print("Errors:", errors)

        # Get the return code of the process (0 usually means success)
        return_code = process.returncode
        #print("Return Code:", return_code)
        lines = output.split("\n")
        for line in lines:
            if (line == ''): break
            #print ("Line = ", line)
            #print ("length = ", len(line))
            line = line.strip()
            #print (line)
            #print ("length = ", len(line))
            # presumably parse line now
            lArr = re.split("=", line)
            #print ("After splitting: ", lArr)
            timezone = lArr[1]
            if verbose:
                print ("Timezone from ", cmd[0]," is: ", timezone)
            return timezone

    # default timezone
    if verbose:
        print ("Default timezone: ", timezone)
    return timezone
'''
tz = getTimezone(True)
print ("Timezone: ", tz)
tz = getTimezone(False)
print ("Timezone: ", tz)
tz = getTimezone(1)
print ("Timezone: ", tz)
tz = getTimezone(0)
print ("Timezone: ", tz)
tz = getTimezone(None)
print ("Timezone: ", tz)
'''


from datetime import datetime
global EarthdtUTimes
EarthdtUTimes = {}  # empty dict of times
#!/usr/bin/env python
# coding: utf-8
#   getEarthXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Earth"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getEarthXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getEarthXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################


    #strdtUT = dtUT.isoformat(sep=' ')     # convert to string and use as key to dict
    #strdtUT = str(dtUT)     # usually yyyy-mm-dd HH:MM:SS.sss
    strdtUT = dtUT.getStr()     # usually yyyy-mm-dd HH:MM:SS.sss
    # check if we have been here before
    if strdtUT in EarthdtUTimes:
        Data = EarthdtUTimes[strdtUT]
        # print ("getEarthXYZData using stored data %s" % strdtUT)
        return Data # X, Y, Z, Dict
    # print ("getEarthXYZData calculate fresh: dtUT = %s" % strdtUT)

    # EARTH X Coord
    sumX0  = 0
    sumX0  +=  99986069.925 * cos(1.75347045757  + 6283.31966747490 *T)
    sumX0  +=  2506324.281 * cos(4.93819429098  + 0.24381748350 *T)
    sumX0  +=  835274.807 * cos(1.71033525539  + 12566.39551746630 *T)
    sumX0  +=   10466.796 * cos(1.66721984219  + 18849.47136745770 *T)
    sumX0  +=    3490.608 * cos(4.44373803231  + 6282.83203250789 *T)
    sumX0  +=    3110.838 * cos(0.66875189331  + 83997.09113559539 *T)
    sumX0  +=    2561.408 * cos(0.58588607490  + 529.93478257810 *T)
    sumX0  +=    2142.365 * cos(1.09204474884  + 1577.09972496430 *T)
    sumX0  +=    1709.161 * cos(0.49540863237  + 6279.79654912590 *T)
    sumX0  +=    1707.934 * cos(6.15314019418  + 6286.84278582391 *T)
    sumX0  +=    1442.753 * cos(3.47210398336  + 2353.10997125530 *T)
    sumX0  +=    1113.027 * cos(3.69621650479  + 5223.45010231870 *T)
    sumX0  +=     934.484 * cos(6.07385378286  + 12036.70455237170 *T)
    sumX0  +=     899.854 * cos(3.17607463681  + 10213.52936369450 *T)
    sumX0  +=     566.820 * cos(2.15241946891  + 1059.62574767270 *T)
    sumX0  +=     661.858 * cos(1.31175222119  + 5753.62870238030 *T)
    sumX0  +=     739.644 * cos(4.36662524112  + 398.39282089170 *T)
    sumX0  +=     681.381 * cos(2.21815397940  + 4705.97612502710 *T)
    sumX0  +=     611.293 * cos(5.38470180335  + 6813.01063256950 *T)
    sumX0  +=     451.129 * cos(6.09315891204  + 5885.17066406670 *T)
    sumX0  +=     451.985 * cos(1.27931036318  + 6257.02134767510 *T)
    sumX0  +=     449.743 * cos(5.36941929064  + 6309.61798727471 *T)
    sumX0  +=     406.291 * cos(0.54369369385  + 6681.46867088311 *T)
    sumX0  +=     541.115 * cos(0.78670634299  + 775.76642880750 *T)
    sumX0  +=     546.682 * cos(1.46109463961  + 14143.73905991410 *T)
    sumX0  +=     513.780 * cos(4.43695037790  + 7860.66320992270 *T)
    sumX0  +=     263.916 * cos(5.39551752630  + 5507.79705615090 *T)
    sumX0  +=     217.220 * cos(4.51265697015  + 11507.01358727710 *T)
    sumX0  +=     292.146 * cos(2.51372357301  + 5507.30942118390 *T)
    sumX0  +=     227.915 * cos(1.23940916733  + 7058.84227879890 *T)
    sumX0  +=     226.835 * cos(3.27356345008  + 4693.75913722409 *T)
    sumX0  +=     255.833 * cos(2.26545070197  + 12168.24651405810 *T)
    sumX0  +=     256.157 * cos(1.45478864804  + 710.17686604180 *T)
    sumX0  +=     178.679 * cos(2.96645630109  + 796.54182429990 *T)
    sumX0  +=     178.303 * cos(0.40472561766  + 6283.25235717211 *T)
    sumX0  +=     178.303 * cos(6.24380795499  + 6283.38697777770 *T)
    sumX0  +=     155.471 * cos(1.62414733243  + 25132.54721744910 *T)
    sumX0  +=     155.246 * cos(1.41969109555  + 5487.02166065850 *T)
    sumX0  +=     207.640 * cos(5.85326148962  + 11790.87290614230 *T)
    sumX0  +=     199.988 * cos(4.07216830110  + 17790.08943726851 *T)
    sumX0  +=     128.970 * cos(5.21698329136  + 7079.61767429131 *T)
    sumX0  +=     128.155 * cos(4.80258701251  + 3739.00524759150 *T)
    sumX0  +=     152.241 * cos(0.86901699174  + 213.54291292150 *T)
    sumX0  +=      78.109 * cos(1.84890519772  + 8827.63408735830 *T)
    sumX0  +=      73.503 * cos(2.69517697923  + 1589.31671276730 *T)
    sumX0  +=      93.001 * cos(3.93777887053  + 9437.51911740350 *T)
    sumX0  +=      65.526 * cos(3.65568333160  + 11770.09751064990 *T)
    sumX0  +=      63.240 * cos(2.24124127597  + 6262.54427198250 *T)
    sumX0  +=      62.961 * cos(4.40729385308  + 6304.09506296730 *T)
    sumX0  +=      85.870 * cos(3.02430778985  + 167284.00540514899 *T)
    sumX0  +=      83.233 * cos(6.19780778434  + 3340.85624418330 *T)
    sumX0  +=      62.850 * cos(2.03823135444  + 7.35736448430 *T)
    sumX0  +=      79.971 * cos(2.49335782552  + 6070.02057203690 *T)
    sumX0  +=      60.043 * cos(3.39955163445  + 4137.15425099970 *T)
    sumX0  +=      75.613 * cos(4.15948190002  + 6496.61876291290 *T)
    sumX0  +=      53.731 * cos(1.56268892040  + 1194.69082770810 *T)
    sumX0  +=      57.574 * cos(3.96971590315  + 6284.29998854310 *T)
    sumX0  +=      57.574 * cos(2.67881766950  + 6282.33934640670 *T)
    sumX0  +=      72.385 * cos(4.85114866885  + 10976.83498721549 *T)
    sumX0  +=      51.135 * cos(1.31237333336  + 6290.43321447570 *T)
    sumX0  +=      50.604 * cos(5.34520074613  + 6276.20612047410 *T)
    sumX0  +=      49.613 * cos(2.64370047145  + 6127.89926804070 *T)
    sumX0  +=      48.009 * cos(4.00571885519  + 6438.74006690910 *T)
    sumX0  +=      59.233 * cos(3.78322753249  + 3154.44326741210 *T)
    sumX0  +=      48.264 * cos(4.77126143981  + 801.57711364030 *T)
    sumX0  +=      47.368 * cos(6.10696378485  + 3128.63258257930 *T)
    sumX0  +=      39.225 * cos(3.25281924495  + 8429.48508395010 *T)
    sumX0  +=      51.307 * cos(5.78448747277  + 12353.09642202830 *T)
    sumX0  +=      50.639 * cos(2.05215225623  + 11856.46246890800 *T)
    sumX0  +=      39.555 * cos(4.93022654622  + 5481.49873635110 *T)
    sumX0  +=      42.757 * cos(3.64063452002  + 17260.39847217390 *T)
    sumX0  +=      48.941 * cos(0.45355151773  + 9225.78309076650 *T)
    sumX0  +=      48.302 * cos(0.60366334071  + 2544.07060239989 *T)
    sumX0  +=      37.971 * cos(3.04093006964  + 426.84200835950 *T)
    sumX0  +=      43.643 * cos(3.14688065147  + 5856.23384163189 *T)
    sumX0  +=      31.830 * cos(2.00885127406  + 2145.92159899169 *T)
    sumX0  +=      33.959 * cos(1.72230567102  + 7085.14059859870 *T)
    sumX0  +=      30.184 * cos(3.96882361878  + 10977.32262218251 *T)
    sumX0  +=      32.562 * cos(2.84647033523  + 4164.06817212950 *T)
    sumX0  +=      38.887 * cos(5.36022568355  + 13367.72881362310 *T)
    sumX0  +=      28.515 * cos(6.10933021457  + 12569.91863581531 *T)
    sumX0  +=      28.873 * cos(1.01228967614  + 3930.45351370310 *T)
    sumX0  +=      36.633 * cos(2.90798803424  + 18073.94875613370 *T)
    sumX0  +=      24.984 * cos(2.17919076736  + 26.05450231630 *T)
    sumX0  +=      26.364 * cos(0.54621601823  + 9438.00675237050 *T)
    sumX0  +=      30.326 * cos(1.99372292658  + 4535.30325440790 *T)
    sumX0  +=      28.166 * cos(0.48127325062  + 12562.87239911730 *T)
    sumX0  +=      28.001 * cos(2.82603231075  + 5088.87265725030 *T)
    sumX0  +=      21.157 * cos(5.07256589577  + 11371.94850724170 *T)
    sumX0  +=      23.113 * cos(0.85996930170  + 5223.93773728571 *T)
    sumX0  +=      21.642 * cos(4.63524157818  + 12559.28197046550 *T)
    sumX0  +=      19.448 * cos(4.40136974621  + 12565.90788249930 *T)
    sumX0  +=      26.055 * cos(0.30340363466  + 71430.45180064559 *T)
    sumX0  +=      18.232 * cos(3.41362095130  + 1747.77259558350 *T)
    sumX0  +=      23.248 * cos(6.19009992884  + 18319.78040236310 *T)
    sumX0  +=      22.785 * cos(1.16821848960  + 5753.14106741329 *T)
    sumX0  +=      17.091 * cos(0.58584172932  + 4933.45225781610 *T)
    sumX0  +=      19.399 * cos(4.43188457477  + 10447.14402212089 *T)
    sumX0  +=      21.067 * cos(3.80033599994  + 7477.76667769950 *T)
    sumX0  +=      21.310 * cos(4.65919931667  + 8031.33608054190 *T)
    sumX0  +=      14.189 * cos(1.30988546287  + 2942.70724077510 *T)
    sumX0  +=      13.864 * cos(4.59350621319  + 8636.18582124671 *T)
    sumX0  +=      18.884 * cos(2.42694418444  + 156137.71980228281 *T)
    sumX0  +=      17.875 * cos(0.15660631596  + 1592.83983111630 *T)
    sumX0  +=      18.602 * cos(0.50024326019  + 17297.93850984270 *T)
    sumX0  +=      13.605 * cos(5.19720916665  + 13096.08648256090 *T)
    sumX0  +=      15.064 * cos(3.12148038350  + 16496.60521368590 *T)
    sumX0  +=      12.969 * cos(6.06491705231  + 7633.18707713370 *T)
    sumX0  +=      14.403 * cos(1.43725922780  + 20426.81490990550 *T)
    sumX0  +=      13.311 * cos(3.60732173221  + 12139.79732659030 *T)
    sumX0  +=      13.645 * cos(5.14981984839  + 13368.21644859010 *T)
    sumX0  +=      11.583 * cos(5.46063496654  + 5331.60126122430 *T)
    sumX0  +=      11.405 * cos(4.40016025355  + 529.44714761109 *T)
    sumX0  +=      15.057 * cos(5.21434071429  + 7084.65296363170 *T)
    sumX0  +=      14.330 * cos(6.02775140319  + 7342.70159766410 *T)
    sumX0  +=      12.375 * cos(3.10941922715  + 6279.72923882311 *T)
    sumX0  +=      12.366 * cos(3.53911446993  + 6286.91009612670 *T)
    sumX0  +=      10.016 * cos(2.24542867964  + 397.90518592470 *T)
    sumX0  +=      12.435 * cos(1.06626484546  + 15110.70993734970 *T)
    sumX0  +=       9.513 * cos(1.19056065248  + 7235.03807372550 *T)
    sumX0  +=      10.997 * cos(2.19198404583  + 10989.05197501850 *T)
    sumX0  +=      11.583 * cos(4.05250824543  + 5729.75026463250 *T)
    sumX0  +=       9.034 * cos(5.42875830514  + 9623.93209417470 *T)
    sumX0  +=      11.842 * cos(4.16957446176  + 15721.08260236190 *T)
    sumX0  +=       8.869 * cos(5.04690919891  + 6148.25458743950 *T)
    sumX0  +=       8.614 * cos(1.60161199451  + 6418.38474751031 *T)
    sumX0  +=      10.312 * cos(2.59789642414  + 6836.88907031730 *T)
    sumX0  +=       8.411 * cos(2.44368872285  + 14712.56093394150 *T)
    sumX0  +=       7.911 * cos(3.15428722892  + 2119.00767786191 *T)
    sumX0  +=      10.379 * cos(4.81870974023  + 1349.62359217529 *T)
    sumX0  +=      10.374 * cos(2.40569215505  + 5486.53402569149 *T)
    sumX0  +=      10.480 * cos(2.95288921695  + 5999.46034860970 *T)
    sumX0  +=       7.404 * cos(5.41419049870  + 6040.59106350090 *T)
    sumX0  +=       7.357 * cos(3.83057351958  + 5088.38502228330 *T)
    sumX0  +=       9.820 * cos(3.69556241034  + 6567.17898634010 *T)
    sumX0  +=       7.018 * cos(1.23420241153  + 6526.04827144891 *T)
    sumX0  +=       9.668 * cos(1.92238869394  + 21228.14820606230 *T)
    sumX0  +=       7.545 * cos(1.23604547808  + 12540.09719766650 *T)
    sumX0  +=       7.202 * cos(5.88391537700  + 6245.29199483910 *T)
    sumX0  +=       7.144 * cos(0.76464656573  + 6321.34734011070 *T)
    sumX0  +=       8.883 * cos(3.89630774866  + 5642.93474619389 *T)
    sumX0  +=       6.626 * cos(0.70123567856  + 5327.71992586630 *T)
    sumX0  +=       7.015 * cos(4.13622360700  + 7860.17557495569 *T)
    sumX0  +=       6.159 * cos(0.94592300466  + 12964.54452087450 *T)
    sumX0  +=       7.413 * cos(5.03498725246  + 23543.47432216530 *T)
    sumX0  +=       6.355 * cos(5.03541099132  + 1990.98883452450 *T)
    sumX0  +=       7.286 * cos(3.58219614228  + 537.04832957890 *T)
    sumX0  +=       6.314 * cos(6.22485628824  + 951.47458876710 *T)
    sumX0  +=       6.257 * cos(4.23208117686  + 4690.72365384210 *T)
    sumX0  +=       6.484 * cos(3.20813291319  + 16730.70750707931 *T)
    sumX0  +=       5.803 * cos(6.09476150956  + 955.35592412510 *T)
    sumX0  +=       6.391 * cos(3.97601526750  + 24073.16528725990 *T)
    sumX0  +=       7.530 * cos(0.20298795021  + 10973.79950383350 *T)
    sumX0  +=       6.921 * cos(0.69969701816  + 522.82123557730 *T)
    sumX0  +=       7.626 * cos(4.29336813013  + 22004.15845235330 *T)
    sumX0  +=       7.531 * cos(3.51464444079  + 18422.87317658170 *T)
    sumX0  +=       6.902 * cos(2.75534627331  + 155.17658195069 *T)
    sumX0  +=       5.434 * cos(5.94625981048  + 7238.91940908350 *T)
    sumX0  +=       6.223 * cos(2.24396249102  + 18451.32236404950 *T)
    sumX0  +=       7.296 * cos(5.98982040270  + 16730.21987211229 *T)
    sumX0  +=       5.833 * cos(2.32930728499  + 640.14110379750 *T)
    sumX0  +=       7.171 * cos(6.14759594277  + 3929.96587873609 *T)
    sumX0  +=       6.251 * cos(2.19374266424  + 6277.79674316750 *T)
    sumX0  +=       6.251 * cos(4.45479090841  + 6288.84259178230 *T)
    sumX0  +=       5.683 * cos(3.49552871951  + 1551.28904013150 *T)
    sumX0  +=       6.383 * cos(0.64667497225  + 5216.33655531789 *T)
    sumX0  +=       6.194 * cos(3.56522608256  + 5230.56364931950 *T)
    sumX0  +=       5.036 * cos(3.83694551305  + 14314.41193053330 *T)
    sumX0  +=       6.774 * cos(1.99358309023  + 102.84895673509 *T)
    sumX0  +=       6.224 * cos(0.35231413534  + 11014.86265985130 *T)
    sumX0  +=       4.621 * cos(1.34668136864  + 553.32558535889 *T)
    sumX0  +=       6.293 * cos(2.32035231613  + 5650.53592816170 *T)
    sumX0  +=       6.297 * cos(4.40170593853  + 26088.14695905770 *T)
    sumX0  +=       5.186 * cos(0.71187314766  + 77714.01528560399 *T)
    sumX0  +=       6.196 * cos(2.58705857336  + 84672.23202702119 *T)
    sumX0  +=       6.160 * cos(4.78250009298  + 239424.63407183639 *T)
    sumX0  +=       5.414 * cos(2.99961957880  + 6180.22689325630 *T)
    sumX0  +=       5.413 * cos(3.65415218952  + 6386.41244169350 *T)
    sumX0  +=       4.863 * cos(3.76722359858  + 90280.16698558679 *T)
    sumX0  +=       5.711 * cos(4.32712250291  + 6916.10340678810 *T)
    sumX0  +=       5.150 * cos(4.10583483822  + 1577.58735993130 *T)
    sumX0  +=       4.269 * cos(2.38358764863  + 7875.91568110771 *T)
    sumX0  +=       4.017 * cos(3.78580903100  + 6254.87048000710 *T)
    sumX0  +=       4.069 * cos(2.42056660765  + 3634.37720703489 *T)
    sumX0  +=       3.993 * cos(2.86272128776  + 6311.76885494270 *T)
    sumX0  +=       4.016 * cos(5.24041079205  + 4690.23601887509 *T)
    sumX0  +=       5.290 * cos(3.34482936009  + 25158.35790228190 *T)
    sumX0  +=       4.831 * cos(2.83916520281  + 5760.74224938110 *T)
    sumX0  +=       3.487 * cos(3.51068129706  + 10447.63165708790 *T)
    sumX0  +=       4.120 * cos(3.29656037233  + 6709.91785835091 *T)
    sumX0  +=       4.410 * cos(3.80589939396  + 6805.89708556871 *T)
    sumX0  +=       4.023 * cos(6.22474978871  + 4731.78680985990 *T)
    sumX0  +=       3.980 * cos(3.31931976159  + 5856.72147659890 *T)
    sumX0  +=       3.173 * cos(5.19128151293  + 1066.73929467350 *T)
    sumX0  +=       3.293 * cos(4.01637480858  + 9917.45305702629 *T)
    sumX0  +=       2.935 * cos(3.67590980046  + 12592.69383726610 *T)
    sumX0  +=       2.979 * cos(6.20067573402  + 12566.46282776910 *T)
    sumX0  +=       3.980 * cos(1.19681754169  + 13341.91812879030 *T)
    sumX0  +=       3.892 * cos(5.52571426134  + 75.02541605080 *T)
    sumX0  +=       3.042 * cos(3.19425693799  + 18053.17336064130 *T)
    sumX0  +=       3.608 * cos(5.60952778146  + 6208.53806890760 *T)
    sumX0  +=       3.721 * cos(5.22104546686  + 5966.92779781830 *T)
    sumX0  +=       3.581 * cos(1.03995231220  + 6358.10126604220 *T)
    sumX0  +=       2.880 * cos(1.61225556126  + 10575.65050042530 *T)
    sumX0  +=       2.806 * cos(4.21551614220  + 5863.83502359970 *T)
    sumX0  +=       3.626 * cos(1.42632163253  + 6599.71153713150 *T)
    sumX0  +=       2.794 * cos(5.69269191706  + 8030.84844557490 *T)
    sumX0  +=       2.729 * cos(4.17319557124  + 3.76693583251 *T)
    sumX0  +=       3.625 * cos(3.00192313835  + 6020.23574410210 *T)
    sumX0  +=       3.206 * cos(2.15397091751  + 11506.52595231009 *T)
    sumX0  +=       3.556 * cos(4.28947576556  + 8428.99744898310 *T)
    sumX0  +=       2.856 * cos(3.33625143564  + 12721.81591690050 *T)
    sumX0  +=       2.611 * cos(2.43261867894  + 6702.80431135010 *T)
    sumX0  +=       2.535 * cos(1.58310702384  + 31415.62306744050 *T)
    sumX0  +=       3.520 * cos(5.37986371510  + 250570.91967470257 *T)
    sumX0  +=       3.434 * cos(3.64656923781  + 6546.40359084770 *T)
    sumX0  +=       2.572 * cos(4.79831420582  + 5884.68302909969 *T)
    sumX0  +=       2.647 * cos(3.81972272861  + 2352.62233628830 *T)
    sumX0  +=       3.026 * cos(5.24078074980  + 13916.26292712510 *T)
    sumX0  +=       2.378 * cos(3.62833713669  + 2389.13783793270 *T)
    sumX0  +=       2.979 * cos(0.36159339669  + 12566.32820716351 *T)
    sumX0  +=       3.082 * cos(1.77514222265  + 14945.07235607090 *T)
    sumX0  +=       2.716 * cos(3.06935389211  + 12029.59100537090 *T)
    sumX0  +=       2.215 * cos(5.26643739208  + 13362.69352428270 *T)
    sumX0  +=       3.032 * cos(4.76553829426  + 29088.56759850150 *T)
    sumX0  +=       2.421 * cos(3.15326621321  + 11015.35029481830 *T)
    sumX0  +=       2.275 * cos(5.95465292259  + 6077.13411903770 *T)
    sumX0  +=       2.466 * cos(5.59166943387  + 16200.52890701769 *T)
    sumX0  +=       2.529 * cos(5.94709131586  + 12043.81809937250 *T)
    sumX0  +=       2.015 * cos(2.48432239832  + 6262.96434807611 *T)
    sumX0  +=       2.218 * cos(0.69324011482  + 6489.50521591210 *T)
    sumX0  +=       2.005 * cos(4.16423892775  + 6303.67498687370 *T)
    sumX0  +=       2.010 * cos(4.11254992503  + 4700.87268422489 *T)
    sumX0  +=       2.641 * cos(4.28275048815  + 4590.66636300550 *T)
    sumX0  +=       2.438 * cos(6.12910205560  + 149.80701461810 *T)
    sumX0  +=       1.893 * cos(0.98057464900  + 6279.43833211690 *T)
    sumX0  +=       1.893 * cos(5.66795892365  + 6287.20100283290 *T)
    sumX0  +=       2.156 * cos(2.00902236290  + 18139.53831889940 *T)
    sumX0  +=       2.071 * cos(4.32179144410  + 11926.49823115230 *T)
    sumX0  +=       2.395 * cos(0.13148357341  + 1162.71852189130 *T)
    sumX0  +=       1.954 * cos(0.37275781655  + 13518.11392371690 *T)
    sumX0  +=       2.006 * cos(0.31048315304  + 3154.93090237910 *T)
    sumX0  +=       1.776 * cos(0.14855204887  + 13521.99525907490 *T)
    sumX0  +=       1.964 * cos(0.25103153440  + 4686.64559022330 *T)
    sumX0  +=       2.332 * cos(4.40649103371  + 10022.08109758290 *T)
    sumX0  +=       1.863 * cos(5.26030609599  + 242.97242145751 *T)
    sumX0  +=       1.928 * cos(4.59250450967  + 5746.51515537950 *T)
    sumX0  +=       2.436 * cos(4.40770823480  + 95143.37673846159 *T)
    sumX0  +=       2.163 * cos(1.41797907369  + 6037.48802124550 *T)
    sumX0  +=       2.250 * cos(2.39728922664  + 12669.48829168490 *T)
    sumX0  +=       2.052 * cos(5.23006608728  + 6529.15131370430 *T)
    sumX0  +=       1.886 * cos(3.61213584641  + 7238.43177411650 *T)
    sumX0  +=       2.170 * cos(5.49596575387  + 19651.29229858150 *T)
    sumX0  +=       2.058 * cos(4.69821533065  + 23013.78335707070 *T)
    sumX0  +=       1.774 * cos(2.08482289391  + 6820.12417957030 *T)
    sumX0  +=       1.841 * cos(0.34653293659  + 65147.37595065419 *T)
    sumX0  +=       1.578 * cos(5.64311385009  + 4292.57465043390 *T)
    sumX0  +=       1.577 * cos(0.80779559190  + 7632.69944216670 *T)
    sumX0  +=       1.541 * cos(1.80359816798  + 13119.96492030870 *T)
    sumX0  +=       2.012 * cos(0.36995312887  + 4292.08701546690 *T)
    sumX0  +=       1.543 * cos(0.68755069426  + 6293.95633282471 *T)
    sumX0  +=       1.543 * cos(5.96098287839  + 6272.68300212510 *T)
    sumX0  +=       1.513 * cos(0.65061890512  + 20.53157800890 *T)
    sumX0  +=       1.489 * cos(2.95205535091  + 110.45013870291 *T)
    sumX0  +=       1.639 * cos(4.44076102657  + 17655.02435723310 *T)
    sumX0  +=       1.443 * cos(2.26329309493  + 1052.51220067191 *T)
    sumX0  +=       1.337 * cos(3.06401553951  + 2544.55823736690 *T)
    sumX0  +=       1.797 * cos(6.18654785499  + 33018.77729472110 *T)
    sumX0  +=       1.488 * cos(5.67450527487  + 76.50988875911 *T)
    sumX0  +=       1.453 * cos(5.46078654600  + 633.02755679670 *T)
    sumX0  +=       1.349 * cos(5.75097749599  + 6016.71262575310 *T)
    sumX0  +=       1.477 * cos(4.14004774348  + 12779.69461290430 *T)
    sumX0  +=       1.688 * cos(3.20087734666  + 18875.28205229050 *T)
    sumX0  +=       1.491 * cos(6.22974184712  + 9411.70843257070 *T)
    sumX0  +=       1.471 * cos(5.75779240419  + 18636.17227201970 *T)
    sumX0  +=       1.263 * cos(0.91728971350  + 12573.50906446710 *T)
    sumX0  +=       1.685 * cos(5.70877589839  + 25934.36814857290 *T)
    sumX0  +=       1.375 * cos(1.08927160683  + 2378.92065608810 *T)
    sumX0  +=       1.641 * cos(6.27664360507  + 6.86972951729 *T)
    sumX0  +=       1.200 * cos(1.47556274227  + 11499.90004027630 *T)
    sumX0  +=       1.272 * cos(0.89738913067  + 6549.92670919670 *T)
    sumX0  +=       1.214 * cos(1.22660393299  + 10973.31186886650 *T)
    sumX0  +=       1.406 * cos(5.67596795657  + 11790.38527117530 *T)
    sumX0  +=       1.524 * cos(4.27066787572  + 18209.57408114370 *T)
    sumX0  +=       1.149 * cos(3.01961706159  + 10177.50149701710 *T)
    sumX0  +=       1.547 * cos(4.02915828753  + 11926.01059618529 *T)
    sumX0  +=       1.460 * cos(4.97235137781  + 246.07546371290 *T)
    sumX0  +=       1.323 * cos(1.38571450422  + 8661.99650607950 *T)
    sumX0  +=       1.252 * cos(1.41166880118  + 6993.25271603320 *T)
    sumX0  +=       1.459 * cos(5.29807880334  + 38.37685312130 *T)
    sumX0  +=       1.235 * cos(2.84096519816  + 24357.02460612510 *T)
    sumX0  +=       1.028 * cos(2.51029997292  + 6112.64679685570 *T)
    sumX0  +=       1.330 * cos(1.00065727327  + 2146.40923395870 *T)
    sumX0  +=       1.312 * cos(1.10091584033  + 12491.61391889900 *T)
    sumX0  +=       1.246 * cos(4.66430542078  + 5429.63565075589 *T)
    sumX0  +=       0.992 * cos(4.13817744468  + 6453.99253809410 *T)
    sumX0  +=       0.988 * cos(0.97969587833  + 8274.06468451590 *T)
    sumX0  +=       1.365 * cos(6.11146862666  + 11371.46087227470 *T)
    sumX0  +=       1.020 * cos(2.82319817767  + 4732.27444482690 *T)
    sumX0  +=       1.191 * cos(3.43008039922  + 6290.36590417291 *T)
    sumX0  +=       1.191 * cos(3.21845317343  + 6276.27343077690 *T)
    sumX0  +=       0.978 * cos(4.73852758176  + 6247.29180079750 *T)
    sumX0  +=       0.971 * cos(1.91000724147  + 6319.34753415231 *T)
    sumX0  +=       0.962 * cos(2.63568544853  + 12565.41519639810 *T)
    sumX0  +=       1.025 * cos(2.21768401152  + 12545.62012197390 *T)
    sumX0  +=       0.990 * cos(0.39039931315  + 4694.24677219110 *T)
    sumX0  +=       0.964 * cos(1.79192739930  + 3893.93801205869 *T)
    sumX0  +=       0.944 * cos(5.86379546667  + 6259.44122972711 *T)
    sumX0  +=       0.944 * cos(0.78473810598  + 6307.19810522270 *T)
    sumX0  +=       1.092 * cos(3.92065202481  + 17797.20298426930 *T)
    sumX0  +=       1.132 * cos(1.00628522970  + 17782.97589026770 *T)
    sumX0  +=       0.859 * cos(1.48109589441  + 1692.40948698591 *T)
    sumX0  +=       1.213 * cos(4.04235990086  + 82576.73740351178 *T)
    sumX0  +=       0.942 * cos(5.22790737110  + 6298.57213865991 *T)
    sumX0  +=       0.942 * cos(1.42062620155  + 6268.06719628990 *T)
    sumX0  +=       1.045 * cos(0.46195584294  + 15508.85894075790 *T)
    sumX0  +=       0.851 * cos(5.19678367921  + 6173.11334625550 *T)
    sumX0  +=       0.850 * cos(1.45204280805  + 6393.52598869430 *T)
    sumX0  +=       0.895 * cos(2.40089211752  + 3904.15519390330 *T)
    sumX0  +=       1.020 * cos(0.71539416375  + 220.65645992230 *T)
    sumX0  +=       0.845 * cos(5.75499771232  + 17256.87535382490 *T)
    sumX0  +=       0.863 * cos(4.43158774725  + 949.41942645330 *T)
    sumX0  +=       0.853 * cos(2.75614337272  + 16201.01654198470 *T)
    sumX0  +=       0.811 * cos(6.13459778682  + 4803.96545844350 *T)
    sumX0  +=       0.966 * cos(4.14597367445  + 206.42936592071 *T)
    sumX0  +=       0.917 * cos(2.47007329129  + 149854.64395229140 *T)
    sumX0  +=       1.083 * cos(1.32514794865  + 36948.98699094070 *T)
    sumX0  +=       0.779 * cos(3.58339682996  + 2648.69864295650 *T)
    sumX0  +=       0.964 * cos(0.77431665261  + 796.05418933290 *T)
    sumX0  +=       0.995 * cos(0.25423339373  + 11403.92081305850 *T)
    sumX0  +=       0.962 * cos(3.92658368218  + 12567.37583853451 *T)
    sumX0  +=       0.779 * cos(0.45683397157  + 10213.04172872750 *T)
    sumX0  +=       0.980 * cos(4.62590378445  + 22805.49174851010 *T)
    sumX0  +=       0.930 * cos(2.22316684933  + 2787.28684134090 *T)
    sumX0  +=       0.829 * cos(5.09643797682  + 5120.35732810009 *T)
    sumX0  +=       0.733 * cos(2.60949856930  + 10575.16286545830 *T)
    sumX0  +=       0.743 * cos(3.79905998257  + 7834.36489012290 *T)
    sumX0  +=       0.818 * cos(1.64368328490  + 5572.89898394960 *T)
    sumX0  +=       0.905 * cos(3.81538816339  + 6284.80414018320 *T)
    sumX0  +=       0.905 * cos(2.83314540925  + 6281.83519476660 *T)
    sumX0  +=       0.798 * cos(2.62501051906  + 12410.97511803210 *T)
    sumX0  +=       0.817 * cos(0.52083502825  + 12416.83232033170 *T)
    sumX0  +=       0.884 * cos(0.84319899817  + 22483.60475700909 *T)
    sumX0  +=       0.755 * cos(3.82941056281  + 4060.97539791089 *T)
    sumX0  +=       0.651 * cos(3.93683821152  + 17259.91083720689 *T)
    sumX0  +=       0.616 * cos(5.70474451820  + 1596.43025976811 *T)
    sumX0  +=       0.651 * cos(2.41491885723  + 1748.26023055050 *T)
    sumX0  +=       0.723 * cos(3.06743374106  + 161000.92955515761 *T)
    sumX0  +=       0.784 * cos(0.87758060898  + 4907.05823266209 *T)
    sumX0  +=       0.784 * cos(2.20169255256  + 7234.55043875850 *T)
    sumX0  +=       0.785 * cos(3.43654258027  + 846.32665223470 *T)
    sumX0  +=       0.611 * cos(1.53152362816  + 853.44019923550 *T)
    sumX0  +=       0.576 * cos(1.22798260830  + 12323.66691349230 *T)
    sumX0  +=       0.570 * cos(4.54665961586  + 12587.17091295870 *T)
    sumX0  +=       0.576 * cos(3.06908128494  + 13915.77529215809 *T)
    sumX0  +=       0.770 * cos(5.60311452827  + 6069.53293706989 *T)
    sumX0  +=       0.773 * cos(0.21363273558  + 6133.26883537330 *T)
    sumX0  +=       0.581 * cos(1.34025743471  + 11933.61177815310 *T)
    sumX0  +=       0.719 * cos(0.90840576142  + 15720.59496739490 *T)
    sumX0  +=       0.578 * cos(4.26997200274  + 8662.48414104651 *T)
    sumX0  +=       0.536 * cos(6.06660590951  + 18852.99448580670 *T)
    sumX0  +=       0.633 * cos(1.17258360285  + 52176.05010063190 *T)
    sumX0  +=       0.581 * cos(2.92798889289  + 5334.14405850510 *T)
    sumX0  +=       0.532 * cos(4.58762178742  + 18842.35782045690 *T)
    sumX0  +=       0.657 * cos(3.25871870022  + 5849.12029463110 *T)
    sumX0  +=       0.599 * cos(2.11793230261  + 6151.77770578850 *T)
    sumX0  +=       0.540 * cos(3.24412649701  + 6286.60602489270 *T)
    sumX0  +=       0.540 * cos(3.40440707563  + 6280.03331005710 *T)
    sumX0  +=       0.628 * cos(0.31668062965  + 17298.42614480970 *T)
    sumX0  +=       0.532 * cos(4.41105068397  + 11514.12713427790 *T)
    sumX0  +=       0.513 * cos(3.70291605445  + 12456.18919624690 *T)
    sumX0  +=       0.649 * cos(4.90568628758  + 11764.57458634250 *T)
    sumX0  +=       0.582 * cos(4.53092429894  + 6414.86162916130 *T)
    sumX0  +=       0.517 * cos(4.06133972007  + 3340.36860921629 *T)
    sumX0  +=       0.609 * cos(3.94494306228  + 10420.23010099111 *T)
    sumX0  +=       0.615 * cos(4.71270964329  + 10983.94853421629 *T)
    sumX0  +=       0.622 * cos(5.53858241889  + 5326.54287653730 *T)
    sumX0  +=       0.536 * cos(3.70224967645  + 7232.49527644471 *T)
    sumX0  +=       0.516 * cos(2.92986975188  + 433.95555536030 *T)
    sumX0  +=       0.613 * cos(1.79095811234  + 10969.72144021469 *T)
    sumX0  +=       0.583 * cos(6.15311267171  + 5863.34738863270 *T)
    sumX0  +=       0.636 * cos(6.06042539245  + 26735.70144472970 *T)
    sumX0  +=       0.665 * cos(2.74682435671  + 40879.19668716030 *T)
    sumX0  +=       0.583 * cos(2.82325235087  + 12592.20620229910 *T)
    sumX0  +=       0.479 * cos(1.33548820034  + 5547.44315394310 *T)
    sumX0  +=       0.571 * cos(2.58673571504  + 6062.90702503610 *T)
    sumX0  +=       0.565 * cos(3.92092198777  + 4171.18171913030 *T)
    sumX0  +=       0.462 * cos(1.96630641591  + 3104.68624194030 *T)
    sumX0  +=       0.551 * cos(4.03874430699  + 6503.73230991370 *T)
    sumX0  +=       0.504 * cos(5.14547341413  + 15670.83794192309 *T)
    sumX0  +=       0.535 * cos(6.12277243019  + 173567.08125514039 *T)
    sumX0  +=       0.460 * cos(3.18401450277  + 3495.78900865049 *T)
    sumX0  +=       0.449 * cos(5.09833649130  + 4274.27449334889 *T)
    sumX0  +=       0.453 * cos(3.57566005736  + 9387.76209193169 *T)
    sumX0  +=       0.527 * cos(6.24827137997  + 24602.85625235450 *T)
    sumX0  +=       0.502 * cos(0.97407216668  + 12490.12944619070 *T)
    sumX0  +=       0.603 * cos(0.85487069842  + 322711.54834138998 *T)
    sumX0  +=       0.532 * cos(1.83659178149  + 5120.84496306710 *T)
    sumX0  +=       0.530 * cos(0.43491203890  + 18845.94824910870 *T)
    sumX0  +=       0.431 * cos(5.31145947559  + 7019.19618100671 *T)
    sumX0  +=       0.465 * cos(5.83298981980  + 8827.14645239129 *T)
    sumX0  +=       0.559 * cos(1.71560880010  + 1582.20316576650 *T)
    sumX0  +=       0.485 * cos(6.13610027053  + 29296.85920706210 *T)
    sumX0  +=       0.584 * cos(0.07138791653  + 72850.80553272919 *T)
    sumX0  +=       0.550 * cos(3.95560592819  + 213.05527795450 *T)
    sumX0  +=       0.481 * cos(1.86946119801  + 14.47091148511 *T)
    sumX0  +=       0.569 * cos(2.95240683895  + 97238.87136197100 *T)
    sumX0  +=       0.551 * cos(1.64539130884  + 14313.92429556630 *T)
    sumX0  +=       0.389 * cos(3.43047350256  + 6245.18663183710 *T)
    sumX0  +=       0.389 * cos(3.21806007009  + 6321.45270311270 *T)
    sumX0  +=       0.388 * cos(1.13963224225  + 6297.54676147650 *T)
    sumX0  +=       0.388 * cos(5.50890133040  + 6269.09257347330 *T)
    sumX0  +=       0.515 * cos(1.67075264819  + 12320.56387123691 *T)
    sumX0  +=       0.394 * cos(6.10437801091  + 4156.95462512869 *T)
    sumX0  +=       0.491 * cos(5.96788884074  + 1479.11039154791 *T)
    sumX0  +=       0.440 * cos(0.97316688838  + 5650.04829319470 *T)
    sumX0  +=       0.382 * cos(3.07660544855  + 9917.94069199330 *T)
    sumX0  +=       0.503 * cos(4.65767245093  + 17157.30569795530 *T)
    sumX0  +=       0.439 * cos(4.82562869150  + 233141.55822184496 *T)
    sumX0  +=       0.381 * cos(2.02484381212  + 14143.25142494710 *T)
    sumX0  +=       0.371 * cos(4.59067170554  + 5643.42238116090 *T)
    sumX0  +=       0.516 * cos(5.61424407240  + 135.30889751891 *T)
    sumX0  +=       0.417 * cos(3.74461559548  + 13760.84252769090 *T)
    sumX0  +=       0.483 * cos(4.41750912984  + 9779.35249360890 *T)
    sumX0  +=       0.381 * cos(4.36951119319  + 14919.26167123810 *T)
    sumX0  +=       0.378 * cos(4.51952349691  + 17267.51201917470 *T)
    sumX0  +=       0.393 * cos(1.56678612595  + 7872.39256275871 *T)
    sumX0  +=       0.376 * cos(4.45581102621  + 13517.62628874989 *T)
    sumX0  +=       0.365 * cos(2.11020607854  + 6923.21695378890 *T)
    sumX0  +=       0.371 * cos(5.76233679446  + 13625.77744765550 *T)
    sumX0  +=       0.355 * cos(5.24317955258  + 10874.22984796390 *T)
    sumX0  +=       0.499 * cos(2.72562606986  + 161710.86260371591 *T)
    sumX0  +=       0.379 * cos(0.81621597880  + 3185.43584474911 *T)
    sumX0  +=       0.471 * cos(4.78294711940  + 11712.71150074729 *T)
    sumX0  +=       0.417 * cos(3.33073591690  + 22779.68106367730 *T)
    sumX0  +=       0.485 * cos(1.35444198321  + 12528.26248182851 *T)
    sumX0  +=       0.358 * cos(3.49115357527  + 6295.04902031090 *T)
    sumX0  +=       0.358 * cos(3.15737999738  + 6271.59031463890 *T)
    sumX0  +=       0.365 * cos(3.62050447597  + 6836.40143535029 *T)
    sumX0  +=       0.351 * cos(2.21887831196  + 11617.21990849651 *T)
    sumX0  +=       0.351 * cos(2.96356670458  + 205.94173095370 *T)
    sumX0  +=       0.429 * cos(0.79790217165  + 3894.42564702570 *T)
    sumX0  +=       0.388 * cos(0.49080052992  + 956.53297345411 *T)
    sumX0  +=       0.464 * cos(0.69464877419  + 23581.50199480110 *T)
    sumX0  +=       0.396 * cos(2.35908361907  + 5231.05128428650 *T)
    sumX0  +=       0.421 * cos(4.82739444320  + 7445.79437188270 *T)
    sumX0  +=       0.366 * cos(0.62612486089  + 17253.28492517310 *T)
    sumX0  +=       0.360 * cos(0.81896878015  + 21393.78578734110 *T)
    sumX0  +=       0.336 * cos(1.41330667750  + 6279.38751421180 *T)
    sumX0  +=       0.336 * cos(5.23522689515  + 6287.25182073800 *T)
    sumX0  +=       0.326 * cos(4.93781139708  + 1059.13811270570 *T)
    sumX0  +=       0.395 * cos(0.53943839302  + 5642.44206009270 *T)
    sumX0  +=       0.450 * cos(0.23150267191  + 1385.31775746760 *T)
    sumX0  +=       0.363 * cos(4.29615345291  + 22484.09239197610 *T)
    sumX0  +=       0.399 * cos(0.80548840004  + 16858.72635041670 *T)
    sumX0  +=       0.316 * cos(4.03075171777  + 20995.63678393290 *T)
    sumX0  +=       0.430 * cos(2.39021484637  + 19650.80466361450 *T)
    sumX0  +=       0.363 * cos(4.29069778349  + 7335.58805066330 *T)
    sumX0  +=       0.317 * cos(1.52889301942  + 11769.60987568290 *T)
    sumX0  +=       0.395 * cos(5.39427984473  + 5905.94605955911 *T)
    sumX0  +=       0.324 * cos(3.70774573147  + 37.78385515230 *T)
    sumX0  +=       0.299 * cos(3.86306301151  + 641.12142486571 *T)
    sumX0  +=       0.304 * cos(0.15985948407  + 5750.10558403130 *T)
    sumX0  +=       0.298 * cos(3.82763680613  + 1350.11122714230 *T)
    sumX0  +=       0.414 * cos(4.17890604161  + 44809.40638337990 *T)
    sumX0  +=       0.339 * cos(4.29176984719  + 3.68833577960 *T)
    sumX0  +=       0.351 * cos(5.93395494462  + 12345.98287502750 *T)
    sumX0  +=       0.362 * cos(0.42214982034  + 21953.91379191449 *T)
    sumX0  +=       0.304 * cos(4.93815385347  + 29826.55017215670 *T)
    sumX0  +=       0.395 * cos(0.07721971830  + 4176.28515993250 *T)
    sumX0  +=       0.320 * cos(2.49901226571  + 10818.37910439930 *T)
    sumX0  +=       0.384 * cos(4.01872268681  + 10177.01386205009 *T)
    sumX0  +=       0.279 * cos(0.96409855547  + 10970.20907518170 *T)
    sumX0  +=       0.365 * cos(1.25425361388  + 6660.69327539070 *T)
    sumX0  +=       0.377 * cos(0.82514906124  + 29864.57784479250 *T)
    sumX0  +=       0.275 * cos(5.73507538514  + 20597.48778052470 *T)
    sumX0  +=       0.316 * cos(2.59688143484  + 316.63568714010 *T)
    sumX0  +=       0.339 * cos(6.10829032638  + 6924.19727485710 *T)
    sumX0  +=       0.386 * cos(2.29032835664  + 2636.96929012050 *T)
    sumX0  +=       0.273 * cos(1.09519308783  + 26709.89075989690 *T)
    sumX0  +=       0.314 * cos(1.47223989608  + 14945.55999103790 *T)
    sumX0  +=       0.281 * cos(4.86653427016  + 16858.23871544969 *T)
    sumX0  +=       0.287 * cos(3.30150277561  + 18073.46112116669 *T)
    sumX0  +=       0.265 * cos(5.11010310272  + 19379.16233255230 *T)
    sumX0  +=       0.325 * cos(2.49278036693  + 12360.20996902910 *T)
    sumX0  +=       0.362 * cos(1.18157001338  + 30665.91114094930 *T)
    sumX0  +=       0.269 * cos(0.20537228942  + 6816.53375091851 *T)
    sumX0  +=       0.344 * cos(3.51559724211  + 6147.69434246491 *T)
    sumX0  +=       0.255 * cos(4.94623609008  + 1376.01761732930 *T)
    sumX0  +=       0.334 * cos(3.13303124260  + 6418.94499248490 *T)
    sumX0  +=       0.316 * cos(5.54205330844  + 6055.84353468590 *T)
    sumX0  +=       0.283 * cos(4.23095234926  + 28287.23430234470 *T)
    sumX0  +=       0.305 * cos(4.26252488384  + 16522.41589851870 *T)
    sumX0  +=       0.270 * cos(0.05215867170  + 283.61550138170 *T)
    sumX0  +=       0.243 * cos(2.54680473739  + 6255.91811137810 *T)
    sumX0  +=       0.243 * cos(4.10172883526  + 6310.72122357170 *T)
    sumX0  +=       0.296 * cos(0.07899510796  + 6129.54085699010 *T)
    sumX0  +=       0.301 * cos(1.10648037942  + 6510.79580026390 *T)
    sumX0  +=       0.329 * cos(1.15683223901  + 377.61742539930 *T)
    sumX0  +=       0.267 * cos(3.42011505320  + 24705.94902657310 *T)
    sumX0  +=       0.285 * cos(2.50494570372  + 5469.76938351510 *T)
    sumX0  +=       0.286 * cos(0.28647364691  + 6437.09847795970 *T)
    sumX0  +=       0.311 * cos(3.21063889136  + 11720.31268271510 *T)
    sumX0  +=       0.319 * cos(4.71059907821  + 169379.50002865840 *T)
    sumX0  +=       0.241 * cos(2.99439330770  + 632.53992182970 *T)
    sumX0  +=       0.229 * cos(1.01522459538  + 1265.81129610991 *T)
    sumX0  +=       0.235 * cos(3.36978882786  + 4487.57358878689 *T)
    sumX0  +=       0.253 * cos(0.51356087882  + 4377.36726756750 *T)
    sumX0  +=       0.277 * cos(1.58719268637  + 419.24082639170 *T)
    sumX0  +=       0.234 * cos(5.12574585310  + 11713.19913571430 *T)
    sumX0  +=       0.216 * cos(5.01445941939  + 10454.25756912169 *T)
    sumX0  +=       0.241 * cos(5.44670796756  + 103.33659170210 *T)
    sumX0  +=       0.214 * cos(2.09481715597  + 2222.10045208050 *T)
    sumX0  +=       0.213 * cos(3.79438383373  + 30356.24113725130 *T)
    sumX0  +=       0.240 * cos(2.57405501441  + 6309.13035230770 *T)
    sumX0  +=       0.212 * cos(2.36637350989  + 262.84010588929 *T)
    sumX0  +=       0.259 * cos(3.21725974670  + 6283.56348495841 *T)
    sumX0  +=       0.259 * cos(3.43127382595  + 6283.07584999140 *T)
    sumX0  +=       0.222 * cos(1.42789121927  + 10440.03047512009 *T)
    sumX0  +=       0.201 * cos(5.11101601554  + 5746.02752041250 *T)
    sumX0  +=       0.274 * cos(3.85387261790  + 23581.01435983410 *T)
    sumX0  +=       0.243 * cos(4.14346519403  + 7096.86995143470 *T)
    sumX0  +=       0.234 * cos(5.23697266935  + 5573.38661891660 *T)
    sumX0  +=       0.203 * cos(0.00533428175  + 16460.08971204149 *T)
    sumX0  +=       0.246 * cos(5.83654379480  + 8672.21368792411 *T)
    sumX0  +=       0.216 * cos(4.93054195434  + 5437.23683272371 *T)
    sumX0  +=       0.210 * cos(5.82498971470  + 9381.20349020070 *T)
    sumX0  +=       0.199 * cos(0.93528415346  + 11216.52810780750 *T)
    sumX0  +=       0.270 * cos(0.40019527935  + 284.10313634870 *T)
    sumX0  +=       0.207 * cos(3.06628694399  + 12562.80508881451 *T)
    sumX0  +=       0.212 * cos(1.75471606720  + 7129.40250222610 *T)
    sumX0  +=       0.269 * cos(1.52668960070  + 70755.31090921978 *T)
    sumX0  +=       0.221 * cos(3.40184768346  + 77713.52765063698 *T)
    sumX0  +=       0.225 * cos(4.01628844723  + 5635.82119919310 *T)
    sumX0  +=       0.197 * cos(3.36122016267  + 14712.07329897449 *T)
    sumX0  +=       0.192 * cos(2.04581063866  + 17272.12782500990 *T)
    sumX0  +=       0.241 * cos(5.09422826000  + 15907.00794416610 *T)
    sumX0  +=       0.261 * cos(5.60213035381  + 48739.61607959950 *T)
    sumX0  +=       0.255 * cos(0.53150571182  + 6206.56596123230 *T)
    sumX0  +=       0.256 * cos(2.30148639986  + 224.58861318540 *T)
    sumX0  +=       0.182 * cos(4.41917983264  + 18848.98373249069 *T)
    sumX0  +=       0.250 * cos(4.24183231344  + 5934.39524702691 *T)
    sumX0  +=       0.196 * cos(2.21911266897  + 16460.57734700850 *T)
    sumX0  +=       0.187 * cos(4.75761261391  + 22003.67081738630 *T)
    sumX0  +=       0.229 * cos(0.16732487654  + 2942.21960580809 *T)
    sumX0  +=       0.175 * cos(5.41173900219  + 11614.67711121570 *T)
    sumX0  +=       0.187 * cos(5.47425531847  + 9778.86485864190 *T)
    sumX0  +=       0.175 * cos(4.58223305763  + 3744.58352825430 *T)
    sumX0  +=       0.241 * cos(0.28794939753  + 8390.35417501730 *T)
    sumX0  +=       0.236 * cos(2.02485303450  + 1.72829019180 *T)
    sumX0  +=       0.217 * cos(5.02460023035  + 17996.27498570570 *T)
    sumX0  +=       0.173 * cos(1.08399000143  + 6275.71848550709 *T)
    sumX0  +=       0.191 * cos(6.27756282518  + 394.86970254271 *T)
    sumX0  +=       0.231 * cos(2.60472823534  + 34596.12083716890 *T)
    sumX0  +=       0.169 * cos(5.11378923747  + 6438.25243194210 *T)
    sumX0  +=       0.235 * cos(3.47396101480  + 17256.38771885789 *T)
    sumX0  +=       0.188 * cos(5.59662915440  + 401.91593924071 *T)
    sumX0  +=       0.217 * cos(4.97442863783  + 10984.43616918330 *T)
    sumX0  +=       0.230 * cos(2.40692408305  + 6632.24408792290 *T)
    sumX0  +=       0.208 * cos(0.72059454942  + 11087.52894340190 *T)
    sumX0  +=       0.165 * cos(2.37182548424  + 743.23387801611 *T)
    sumX0  +=       0.177 * cos(4.73067975820  + 4796.85191144269 *T)
    sumX0  +=       0.208 * cos(4.59043597135  + 3097.64000524229 *T)
    sumX0  +=       0.167 * cos(3.97387947664  + 5714.49779344750 *T)
    sumX0  +=       0.198 * cos(3.26732644351  + 5539.84197197530 *T)
    sumX0  +=       0.169 * cos(3.73983013574  + 12132.68377958950 *T)
    sumX0  +=       0.214 * cos(4.34216809137  + 24492.64993113510 *T)
    sumX0  +=       0.172 * cos(1.25880044898  + 6233.56264200310 *T)
    sumX0  +=       0.172 * cos(5.38973312367  + 6333.07669294670 *T)
    sumX0  +=       0.223 * cos(4.99990847060  + 266.85085920531 *T)
    sumX0  +=       0.159 * cos(3.90246086071  + 13199.17925677950 *T)
    sumX0  +=       0.219 * cos(5.40805183208  + 10344.05124790229 *T)
    sumX0  +=       0.207 * cos(3.49598218672  + 12569.98594611810 *T)
    sumX0  +=       0.176 * cos(4.05199209274  + 12012.82611462390 *T)
    sumX0  +=       0.170 * cos(6.01039539440  + 6294.36536773881 *T)
    sumX0  +=       0.170 * cos(0.63813817825  + 6272.27396721100 *T)
    sumX0  +=       0.205 * cos(5.87166506948  + 13119.47728534169 *T)
    sumX0  +=       0.167 * cos(5.54385526640  + 17686.99666304990 *T)
    sumX0  +=       0.210 * cos(0.88175440791  + 13521.50762410789 *T)
    sumX0  +=       0.163 * cos(1.46796231870  + 802.06474860730 *T)
    sumX0  +=       0.159 * cos(4.85018558111  + 5017.26455388150 *T)
    sumX0  +=       0.148 * cos(0.75075299347  + 419.72846135871 *T)
    sumX0  +=       0.165 * cos(0.87966880312  + 20199.33877711650 *T)
    sumX0  +=       0.196 * cos(3.57999903444  + 33326.33491569069 *T)
    sumX0  +=       0.146 * cos(0.39484750798  + 19800.70213874130 *T)
    sumX0  +=       0.148 * cos(2.67465416652  + 6852.14154150230 *T)
    sumX0  +=       0.146 * cos(0.23882077577  + 17370.60479339330 *T)
    sumX0  +=       0.143 * cos(0.85202349514  + 5618.56362234490 *T)
    sumX0  +=       0.142 * cos(5.37141800193  + 17654.53672226610 *T)
    sumX0  +=       0.140 * cos(1.22241558577  + 2008.80135664250 *T)
    sumX0  +=       0.153 * cos(1.79747194391  + 5436.74919775670 *T)
    sumX0  +=       0.162 * cos(3.70215026876  + 775.27879384050 *T)
    sumX0  +=       0.162 * cos(4.78754155685  + 12552.16842346470 *T)
    sumX0  +=       0.143 * cos(2.32218951292  + 5010.15100688070 *T)
    sumX0  +=       0.188 * cos(3.68702080121  + 28.69300495130 *T)
    sumX0  +=       0.155 * cos(6.15650399253  + 11610.10636149570 *T)
    sumX0  +=       0.174 * cos(5.71964679538  + 20452.62559473830 *T)
    sumX0  +=       0.179 * cos(5.30407615411  + 27511.22405605370 *T)
    sumX0  +=       0.135 * cos(0.07815962516  + 12431.33043743090 *T)
    sumX0  +=       0.164 * cos(5.60920106543  + 28767.16824196750 *T)
    sumX0  +=       0.134 * cos(5.44099179137  + 16840.91382829870 *T)
    sumX0  +=       0.174 * cos(4.27755511799  + 19805.07110906630 *T)
    sumX0  +=       0.146 * cos(1.55848605346  + 12701.46059750170 *T)
    sumX0  +=       0.137 * cos(2.60101256262  + 11.28951774740 *T)
    sumX0  +=       0.167 * cos(1.27059509188  + 17473.69756761190 *T)
    sumX0  +=       0.152 * cos(4.19104806587  + 16627.61473286070 *T)
    sumX0  +=       0.124 * cos(5.79651030598  + 6948.07571260490 *T)
    sumX0  +=       0.135 * cos(3.38729000615  + 3531.28443281630 *T)
    sumX0  +=       0.172 * cos(4.94261447200  + 167959.14629657479 *T)
    sumX0  +=       0.123 * cos(5.21069118841  + 23013.29572210369 *T)
    sumX0  +=       0.122 * cos(5.69263982892  + 3583.58484815730 *T)
    sumX0  +=       0.170 * cos(1.45223428169  + 333857.83394425618 *T)
    sumX0  +=       0.155 * cos(4.05745332726  + 6058.48723680599 *T)
    sumX0  +=       0.120 * cos(1.19140907580  + 12809.12412144031 *T)
    sumX0  +=       0.141 * cos(5.52535299558  + 162420.79565227419 *T)
    sumX0  +=       0.120 * cos(5.84076902832  + 12528.36784483050 *T)
    sumX0  +=       0.123 * cos(6.20892018031  + 25933.88051360590 *T)
    sumX0  +=       0.144 * cos(2.13726910032  + 95.73540973430 *T)
    sumX0  +=       0.165 * cos(0.74320909184  + 52669.82577581910 *T)
    sumX0  +=       0.119 * cos(1.19695838858  + 19247.62037086590 *T)
    sumX0  +=       0.117 * cos(0.73795876397  + 11610.79577585770 *T)
    sumX0  +=       0.163 * cos(3.11536370558  + 661.47674426450 *T)
    sumX0  +=       0.163 * cos(4.35849093916  + 9929.67004482930 *T)
    sumX0  +=       0.119 * cos(4.28708216617  + 12250.00364780970 *T)
    sumX0  +=       0.145 * cos(3.90603326867  + 6205.64589704690 *T)
    sumX0  +=       0.145 * cos(2.74250030398  + 6360.99343790291 *T)
    sumX0  +=       0.159 * cos(4.18513637906  + 228278.34846897016 *T)
    sumX0  +=       0.114 * cos(1.84089435963  + 19402.55313533309 *T)
    sumX0  +=       0.147 * cos(4.06094470975  + 38526.33053338850 *T)
    sumX0  +=       0.156 * cos(2.57788751226  + 4307.82712161890 *T)
    sumX0  +=       0.131 * cos(1.70667997590  + 21228.63584102931 *T)
    sumX0  +=       0.118 * cos(5.40927777429  + 6263.64990657511 *T)
    sumX0  +=       0.118 * cos(1.23925579836  + 6302.98942837470 *T)
    sumX0  +=       0.137 * cos(2.54415078163  + 6315.85221826630 *T)
    sumX0  +=       0.137 * cos(4.10438279102  + 6250.78711668350 *T)
    sumX0  +=       0.109 * cos(2.76724250809  + 11925.51791008410 *T)
    sumX0  +=       0.111 * cos(4.37410999057  + 6226.42129253930 *T)
    sumX0  +=       0.111 * cos(2.27442358208  + 6340.21804241050 *T)
    sumX0  +=       0.123 * cos(2.37617280556  + 24734.39821404090 *T)
    sumX0  +=       0.107 * cos(2.16622176546  + 12463.30274324771 *T)
    sumX0  +=       0.133 * cos(3.03066152119  + 18875.76968725751 *T)
    sumX0  +=       0.123 * cos(5.64258939044  + 6260.54446602410 *T)
    sumX0  +=       0.123 * cos(1.00594418221  + 6306.09486892571 *T)
    sumX0  +=       0.113 * cos(1.95402250286  + 2111.89413086110 *T)
    sumX0  +=       0.131 * cos(3.61193663544  + 18415.75962958090 *T)
    sumX0  +=       0.110 * cos(1.76186229885  + 6289.94822637491 *T)
    sumX0  +=       0.110 * cos(4.88667127380  + 6276.69110857489 *T)
    sumX0  +=       0.103 * cos(3.89043805400  + 6241.76887649010 *T)
    sumX0  +=       0.103 * cos(2.75809551865  + 6324.87045845970 *T)
    sumX0  +=       0.113 * cos(2.20489389956  + 3496.27664361750 *T)
    sumX0  +=       0.103 * cos(4.57936207476  + 10344.53888286930 *T)
    sumX0  +=       0.115 * cos(2.66876079900  + 24336.24921063270 *T)
    sumX0  +=       0.107 * cos(1.72027504380  + 83974.07917321339 *T)
    sumX0  +=       0.107 * cos(2.75882626470  + 84020.10309797739 *T)
    sumX0  +=       0.106 * cos(3.27680652019  + 12772.58106590351 *T)
    sumX0  +=       0.143 * cos(1.51224865614  + 2069.25065239010 *T)
    sumX0  +=       0.143 * cos(0.91984229557  + 18773.20529618210 *T)
    sumX0  +=       0.115 * cos(3.78326005147  + 3641.49075403570 *T)
    sumX0  +=       0.112 * cos(5.73842498855  + 11499.41240530929 *T)
    sumX0  +=       0.099 * cos(1.79019915134  + 11190.62171762050 *T)
    sumX0  +=       0.126 * cos(1.19294266238  + 18823.17304765790 *T)
    sumX0  +=       0.115 * cos(4.66544469064  + 12570.32767072940 *T)
    sumX0  +=       0.100 * cos(3.52903073097  + 16062.42834360030 *T)
    sumX0  +=       0.107 * cos(4.02566335272  + 5216.82419028490 *T)
    sumX0  +=       0.120 * cos(4.96595619835  + 9814.36028280769 *T)
    sumX0  +=       0.138 * cos(4.55918305844  + 6210.02254161590 *T)
    sumX0  +=       0.138 * cos(2.08935051421  + 6356.61679333390 *T)
    sumX0  +=       0.105 * cos(1.03351724810  + 12721.32828193349 *T)
    sumX0  +=       0.135 * cos(0.57056498904  + 18699.90817032310 *T)
    sumX0  +=       0.104 * cos(2.15061044327  + 12560.87259315890 *T)
    sumX0  +=       0.098 * cos(5.57514638218  + 5815.35467712050 *T)
    sumX0  +=       0.111 * cos(3.97355954616  + 10239.34004852730 *T)
    sumX0  +=       0.100 * cos(4.55824596261  + 263.32774085630 *T)
    sumX0  +=       0.095 * cos(1.74741902247  + 155.66421691770 *T)
    sumX0  +=       0.130 * cos(2.12273849218  + 27511.71169102070 *T)
    sumX0  +=       0.120 * cos(0.27869701115  + 31441.43375227330 *T)
    sumX0  +=       0.114 * cos(3.75372093481  + 6155.30082413750 *T)
    sumX0  +=       0.114 * cos(2.89481263784  + 6411.33851081230 *T)
    sumX0  +=       0.119 * cos(5.21922596573  + 951.96222373410 *T)
    sumX0  +=       0.117 * cos(1.90078630587  + 28236.98964190589 *T)
    sumX0  +=       0.107 * cos(1.90914871846  + 21.01921297590 *T)
    sumX0  +=       0.095 * cos(5.63307991850  + 11300.82803883990 *T)
    sumX0  +=       0.110 * cos(2.79120257993  + 6312.74917601091 *T)
    sumX0  +=       0.110 * cos(3.85733099272  + 6253.89015893890 *T)
    sumX0  +=       0.097 * cos(4.22068998833  + 78263.95324220609 *T)
    sumX0  +=       0.095 * cos(5.98375623165  + 23938.10020722450 *T)
    sumX0  +=       0.091 * cos(2.99457476326  + 12829.47944083910 *T)
    sumX0  +=       0.112 * cos(4.31165167302  + 16737.82105408010 *T)
    sumX0  +=       0.126 * cos(3.10862438843  + 3.27930086550 *T)
    sumX0  +=       0.100 * cos(3.03286435761  + 6133.75647034030 *T)
    sumX0  +=       0.127 * cos(5.41503454763  + 1293.24040609909 *T)
    sumX0  +=       0.119 * cos(5.99165259734  + 17893.18221148710 *T)
    sumX0  +=       0.089 * cos(1.33458093593  + 23539.95120381630 *T)
    sumX0  +=       0.126 * cos(0.25750692887  + 311565.26273852371 *T)
    sumX0  +=       0.093 * cos(5.67599033319  + 736.12033101530 *T)
    sumX0  +=       0.120 * cos(2.08797711521  + 14158.99153109910 *T)
    sumX0  +=       0.118 * cos(1.43212365415  + 16061.94070863329 *T)
    sumX0  +=       0.099 * cos(3.61958161583  + 6432.88286460950 *T)
    sumX0  +=       0.095 * cos(5.99715593195  + 2699.49100183409 *T)
    sumX0  +=       0.089 * cos(2.30094279014  + 15671.32557689010 *T)
    sumX0  +=       0.124 * cos(0.48008010598  + 178430.29100801519 *T)
    sumX0  +=       0.089 * cos(1.07338684504  + 6751.28465782931 *T)
    sumX0  +=       0.096 * cos(2.64494034094  + 7349.81514466491 *T)
    sumX0  +=       0.096 * cos(0.74221743876  + 24066.05174025910 *T)
    sumX0  +=       0.103 * cos(4.37747381017  + 18202.46053414290 *T)
    sumX0  +=       0.099 * cos(6.06683620849  + 6252.40554183991 *T)
    sumX0  +=       0.099 * cos(0.58169736416  + 6314.23379310990 *T)
    sumX0  +=       0.090 * cos(0.38965345976  + 58864.30010066279 *T)
    sumX0  +=       0.086 * cos(0.59831141560  + 9380.71585523369 *T)
    sumX0  +=       0.086 * cos(0.93909542206  + 10557.35034334029 *T)
    sumX0  +=       0.098 * cos(3.41472072058  + 6439.72038797730 *T)
    sumX0  +=       0.098 * cos(3.23381285207  + 6126.91894697251 *T)
    sumX0  +=       0.086 * cos(1.34148300572  + 23123.98967829010 *T)
    sumX0  +=       0.118 * cos(3.73844359380  + 13951.95709241740 *T)
    sumX0  +=       0.117 * cos(4.10902635534  + 89570.23393702849 *T)
    sumX0  +=       0.083 * cos(3.14133645510  + 8858.07112683710 *T)
    sumX0  +=       0.092 * cos(4.01226998426  + 12342.05072176440 *T)
    sumX0  +=       0.085 * cos(2.29415888429  + 5017.75218884850 *T)
    sumX0  +=       0.101 * cos(0.23020335948  + 18429.98672358250 *T)
    sumX0  +=       0.088 * cos(3.76744436736  + 17054.21292373670 *T)
    sumX0  +=       0.113 * cos(2.05049922518  + 12985.88016134151 *T)
    sumX0  +=       0.083 * cos(0.90577628609  + 20597.00014555770 *T)
    sumX0  +=       0.082 * cos(0.19375496576  + 5483.49854230950 *T)
    sumX0  +=       0.087 * cos(0.10696172464  + 21424.22282681990 *T)
    sumX0  +=       0.107 * cos(2.21757059467  + 522.33360061030 *T)
    sumX0  +=       0.084 * cos(2.41107951633  + 6187.34044025710 *T)
    sumX0  +=       0.084 * cos(4.23745405632  + 6379.29889469270 *T)
    sumX0  +=       0.107 * cos(0.88149755532  + 24382.83529095790 *T)
    sumX0  +=       0.082 * cos(0.94957627406  + 8983.05448679250 *T)
    sumX0  +=       0.105 * cos(3.87374571694  + 6131.42238638970 *T)
    sumX0  +=       0.105 * cos(2.77478785571  + 6435.21694856010 *T)
    sumX0  +=       0.114 * cos(4.07080782247  + 8258.81221333091 *T)
    sumX0  +=       0.082 * cos(3.82951941927  + 3957.88262369230 *T)
    sumX0  +=       0.112 * cos(2.70888594040  + 3738.51761262449 *T)
    sumX0  +=       0.102 * cos(2.89783593458  + 5767.85579638190 *T)
    sumX0  +=       0.102 * cos(3.75069763806  + 6798.78353856790 *T)
    sumX0  +=       0.092 * cos(1.35671084183  + 18216.68762814450 *T)
    sumX0  +=       0.083 * cos(1.35246903804  + 29864.09020982550 *T)
    sumX0  +=       0.112 * cos(5.48905138227  + 12189.02190955050 *T)
    sumX0  +=       0.080 * cos(3.59069348639  + 24080.27883426070 *T)
    sumX0  +=       0.080 * cos(4.80792276932  + 15141.14697682849 *T)
    sumX0  +=       0.088 * cos(3.00407783628  + 1573.57660661529 *T)
    sumX0  +=       0.081 * cos(3.72088724498  + 1550.80140516450 *T)
    sumX0  +=       0.092 * cos(4.36458397294  + 101426.45258845300 *T)
    sumX0  +=       0.110 * cos(0.37007007331  + 78423.94833416230 *T)
    sumX0  +=       0.087 * cos(2.32197271603  + 1580.62284331330 *T)
    sumX0  +=       0.085 * cos(6.14240130116  + 27043.25906569930 *T)
    sumX0  +=       0.078 * cos(1.45118645187  + 6812.52299760250 *T)
    sumX0  +=       0.080 * cos(3.36946826019  + 6081.06627230081 *T)
    sumX0  +=       0.080 * cos(3.27906531246  + 6485.57306264900 *T)
    sumX0  +=       0.079 * cos(4.53816488559  + 36109.62602214810 *T)
    sumX0  +=       0.088 * cos(0.87069012239  + 16944.00660251730 *T)
    sumX0  +=       0.081 * cos(1.41865192078  + 5.27910682390 *T)
    sumX0  +=       0.093 * cos(5.85619230131  + 16737.33341911309 *T)
    sumX0  +=       0.088 * cos(3.24901264480  + 12537.94632999850 *T)
    sumX0  +=       0.096 * cos(5.19394745813  + 20198.85114214949 *T)
    sumX0  +=       0.106 * cos(2.16665694222  + 56600.03547203870 *T)
    sumX0  +=       0.097 * cos(0.14089796919  + 6040.10342853389 *T)
    sumX0  +=       0.081 * cos(3.62140277149  + 3956.26419853590 *T)
    sumX0  +=       0.077 * cos(1.39348893935  + 40796.58274015770 *T)
    sumX0  +=       0.076 * cos(2.20964866296  + 22743.16556203289 *T)
    sumX0  +=       0.098 * cos(5.48885750823  + 42456.54022960810 *T)
    sumX0  +=       0.083 * cos(2.20640709608  + 19801.18977370830 *T)
    sumX0  +=       0.079 * cos(4.50777033878  + 5622.08674069391 *T)
    sumX0  +=       0.096 * cos(1.98497858569  + 5888.69378241570 *T)
    sumX0  +=       0.096 * cos(4.66355498696  + 6677.94555253410 *T)
    sumX0  +=       0.075 * cos(6.24025577336  + 41194.73174356590 *T)
    sumX0  +=       0.074 * cos(0.66729085356  + 6261.98402700790 *T)
    sumX0  +=       0.074 * cos(5.98124271908  + 6304.65530794190 *T)
    sumX0  +=       0.073 * cos(4.02724017625  + 5870.94857060050 *T)
    sumX0  +=       0.073 * cos(2.62129339640  + 6695.69076434930 *T)
    sumX0  +=       0.085 * cos(3.65241534932  + 12850.25483633150 *T)
    sumX0  +=       0.072 * cos(0.04178130909  + 6253.49822932610 *T)
    sumX0  +=       0.072 * cos(0.32356695638  + 6313.14110562371 *T)
    sumX0  +=       0.100 * cos(5.37794526764  + 5316.34879003930 *T)
    sumX0  +=       0.085 * cos(2.90983978106  + 12282.53619860110 *T)
    sumX0  +=       0.075 * cos(1.96590105616  + 24422.61416889080 *T)
    sumX0  +=       0.076 * cos(1.98227687206  + 63659.12156832110 *T)
    sumX0  +=       0.072 * cos(0.25938390676  + 16723.59396007850 *T)
    sumX0  +=       0.097 * cos(4.92418361725  + 17995.78735073869 *T)
    sumX0  +=       0.087 * cos(3.63417247018  + 18106.48130692510 *T)
    sumX0  +=       0.080 * cos(5.17587638062  + 17363.49124639250 *T)
    sumX0  +=       0.070 * cos(4.63132769081  + 6124.37614969171 *T)
    sumX0  +=       0.070 * cos(2.01720588184  + 6442.26318525810 *T)
    sumX0  +=       0.070 * cos(2.85493203288  + 4705.48849006010 *T)
    sumX0  +=       0.078 * cos(4.94849285823  + 23550.58786916610 *T)
    sumX0  +=       0.070 * cos(2.27501820444  + 12036.21691740469 *T)
    sumX0  +=       0.077 * cos(3.43499900992  + 5237.67719632029 *T)
    sumX0  +=       0.069 * cos(6.02127545231  + 16207.64245401849 *T)
    sumX0  +=       0.082 * cos(5.48703292615  + 6774.98295993371 *T)
    sumX0  +=       0.067 * cos(0.17045217443  + 7083.14079264031 *T)
    sumX0  +=       0.067 * cos(1.83540783927  + 6394.50630976251 *T)
    sumX0  +=       0.067 * cos(4.81312573338  + 6172.13302518730 *T)
    sumX0  +=       0.077 * cos(4.99636632687  + 9924.56660402710 *T)
    sumX0  +=       0.080 * cos(5.36758481138  + 22380.99961775750 *T)
    sumX0  +=       0.070 * cos(4.60174692010  + 6390.98319141350 *T)
    sumX0  +=       0.070 * cos(2.04678665255  + 6175.65614353630 *T)
    sumX0  +=       0.074 * cos(2.55353968467  + 16193.41536001689 *T)
    sumX0  +=       0.076 * cos(5.64799287134  + 32217.44399856430 *T)
    sumX0  +=       0.091 * cos(2.47526506655  + 6653.01948341530 *T)
    sumX0  +=       0.091 * cos(4.17326850610  + 5913.61985153450 *T)
    sumX0  +=       0.067 * cos(0.23293815799  + 6265.50714535691 *T)
    sumX0  +=       0.067 * cos(0.13241010748  + 6301.13218959290 *T)
    sumX0  +=       0.067 * cos(1.44106493302  + 5959.32661585050 *T)
    sumX0  +=       0.083 * cos(2.94569161083  + 16723.10632511149 *T)
    sumX0  +=       0.080 * cos(0.77941152413  + 23646.56709638390 *T)
    sumX0  +=       0.068 * cos(3.64312239532  + 4897.42439113870 *T)
    sumX0  +=       0.069 * cos(2.14076318935  + 6944.55259425590 *T)
    sumX0  +=       0.089 * cos(1.99980743200  + 10660.44311755889 *T)
    sumX0  +=       0.079 * cos(1.67507258686  + 35371.64344849290 *T)
    sumX0  +=       0.073 * cos(2.11438932219  + 6370.62787201471 *T)
    sumX0  +=       0.073 * cos(4.53414425046  + 6196.01146293510 *T)
    sumX0  +=       0.065 * cos(3.66592557921  + 22345.01655862470 *T)
    sumX0  +=       0.064 * cos(4.23436375372  + 15265.64270181689 *T)
    sumX0  +=       0.066 * cos(2.69579215047  + 6315.29197329170 *T)
    sumX0  +=       0.066 * cos(3.95274142218  + 6251.34736165811 *T)
    sumX0  +=       0.065 * cos(5.01328627787  + 323.74923414091 *T)
    sumX0  +=       0.076 * cos(6.17321822941  + 10873.74221299689 *T)
    sumX0  +=       0.084 * cos(5.21213167261  + 109.96250373590 *T)
    sumX0  +=       0.063 * cos(4.27995662426  + 11823.40545693370 *T)
    sumX0  +=       0.064 * cos(0.09093186873  + 28774.28178896830 *T)
    sumX0  +=       0.072 * cos(0.54185576116  + 18099.75944096650 *T)
    sumX0  +=       0.066 * cos(2.78134607936  + 17576.79034183050 *T)
    sumX0  +=       0.072 * cos(1.59783352501  + 245707.70992182777 *T)
    sumX0  +=       0.061 * cos(5.48721212671  + 10557.83797830730 *T)
    sumX0  +=       0.064 * cos(0.75473231286  + 71430.93943561260 *T)
    sumX0  +=       0.068 * cos(2.90814349886  + 28760.05469496669 *T)
    sumX0  +=       0.062 * cos(0.67966569221  + 3854.78984947370 *T)
    sumX0  +=       0.073 * cos(6.18332621846  + 23440.38154794670 *T)
    sumX0  +=       0.060 * cos(3.22950754418  + 13088.97293556010 *T)
    sumX0  +=       0.078 * cos(4.56094059497  + 12564.91104475801 *T)
    sumX0  +=       0.060 * cos(4.47214922427  + 7548.88714610130 *T)
    sumX0  +=       0.059 * cos(2.55184049976  + 6286.35515085690 *T)
    sumX0  +=       0.075 * cos(4.42124944201  + 18625.12657175580 *T)
    sumX0  +=       0.058 * cos(0.47683719349  + 35050.24409195890 *T)
    sumX0  +=       0.060 * cos(0.20507190989  + 553.81322032590 *T)
    sumX0  +=       0.069 * cos(5.82409350668  + 167993.93845370729 *T)
    sumX0  +=       0.059 * cos(0.42390666617  + 41991.02975038230 *T)
    sumX0  +=       0.062 * cos(2.87556762555  + 15663.79170522509 *T)
    sumX0  +=       0.081 * cos(1.27058846781  + 7250.29054491051 *T)
    sumX0  +=       0.059 * cos(2.68580345393  + 6277.25375184510 *T)
    sumX0  +=       0.059 * cos(3.96273011871  + 6289.38558310470 *T)
    sumX0  +=       0.063 * cos(2.13380171799  + 18003.38853270651 *T)
    sumX0  +=       0.057 * cos(4.87646676285  + 793.01870595090 *T)
    sumX0  +=       0.057 * cos(0.02714811761  + 10027.65937824569 *T)
    sumX0  +=       0.060 * cos(2.20528431296  + 647.25465079831 *T)
    sumX0  +=       0.072 * cos(4.16486482248  + 4597.77991000629 *T)
    sumX0  +=       0.059 * cos(3.82312915284  + 6279.30891415889 *T)
    sumX0  +=       0.068 * cos(2.71579349818  + 3166.66025521511 *T)
    sumX0  +=       0.056 * cos(4.67855976016  + 6226.51640530510 *T)
    sumX0  +=       0.056 * cos(1.96997381249  + 6340.12292964471 *T)
    sumX0  +=       0.056 * cos(2.52728721608  + 12303.31159409350 *T)
    sumX0  +=       0.057 * cos(3.09052199581  + 5864.39526857430 *T)
    sumX0  +=       0.057 * cos(3.55801157684  + 6702.24406637550 *T)
    sumX0  +=       0.069 * cos(1.90302432953  + 23536.36077516450 *T)
    sumX0  +=       0.059 * cos(0.79584524101  + 5760.25461441409 *T)
    sumX0  +=       0.065 * cos(0.39208551086  + 12139.30969162329 *T)
    sumX0  +=       0.059 * cos(3.38503180730  + 67589.33126454070 *T)
    sumX0  +=       0.073 * cos(0.51774809611  + 11079.92776143409 *T)
    sumX0  +=       0.076 * cos(2.34882980338  + 23227.08245250870 *T)
    sumX0  +=       0.067 * cos(2.11713130342  + 4804.45309341050 *T)
    sumX0  +=       0.073 * cos(6.00949211634  + 30349.12759025050 *T)
    sumX0  +=       0.074 * cos(4.74855693255  + 92747.93298430610 *T)
    sumX0  +=       0.070 * cos(1.56579326622  + 12299.78847574450 *T)
    sumX0  +=       0.060 * cos(0.37222072501  + 12171.76963240710 *T)
    sumX0  +=       0.059 * cos(2.60537764043  + 15.49628866851 *T)
    sumX0  +=       0.074 * cos(5.91207332800  + 3684.13423250670 *T)
    sumX0  +=       0.058 * cos(3.15618940951  + 6717.03140535170 *T)
    sumX0  +=       0.053 * cos(2.11594966811  + 26087.65932409069 *T)
    sumX0  +=       0.059 * cos(1.05069506955  + 12164.72339570910 *T)
    sumX0  +=       0.063 * cos(3.66132366701  + 5219.51794905560 *T)
    sumX0  +=       0.056 * cos(4.02457482788  + 3178.38960805111 *T)
    sumX0  +=       0.063 * cos(0.59069175384  + 5227.38225558179 *T)
    sumX0  +=       0.070 * cos(3.26289105386  + 19004.40413192490 *T)
    sumX0  +=       0.072 * cos(1.05516042130  + 4583.55281600469 *T)
    sumX0  +=       0.052 * cos(5.67763480740  + 3627.26366003410 *T)
    sumX0  +=       0.052 * cos(4.87358219235  + 6411.27120050950 *T)
    sumX0  +=       0.052 * cos(1.77495138030  + 6155.36813444030 *T)
    sumX0  +=       0.054 * cos(3.61747287023  + 5849.60792959810 *T)
    sumX0  +=       0.066 * cos(4.36702417592  + 5791.16874004909 *T)
    sumX0  +=       0.066 * cos(1.22543152233  + 5791.65637501610 *T)
    sumX0  +=       0.051 * cos(0.84172627776  + 5113.24378109930 *T)
    sumX0  +=       0.058 * cos(4.29307138428  + 30775.72578112650 *T)
    sumX0  +=       0.061 * cos(5.44071514919  + 14169.54974474690 *T)
    sumX0  +=       0.054 * cos(4.73181482869  + 10454.74520408871 *T)
    sumX0  +=       0.056 * cos(2.77114790322  + 23141.80220040810 *T)
    sumX0  +=       0.067 * cos(2.31997560907  + 28313.04498717750 *T)
    sumX0  +=       0.056 * cos(1.94933767144  + 5244.29305668450 *T)
    sumX0  +=       0.062 * cos(2.18007483018  + 5657.64947516250 *T)
    sumX0  +=       0.062 * cos(4.46845874247  + 6908.98985978730 *T)
    sumX0  +=       0.050 * cos(1.53914825533  + 2574.99527684569 *T)
    sumX0  +=       0.050 * cos(1.32272515661  + 536.56069461190 *T)
    sumX0  +=       0.050 * cos(1.95397557344  + 6684.99178923210 *T)
    sumX0  +=       0.050 * cos(4.69455799921  + 5881.64754571770 *T)
    sumX0  +=       0.049 * cos(5.07674655291  + 16310.73522823709 *T)
    sumX0  +=       0.049 * cos(5.07674655291  + 16311.22286320410 *T)
    sumX0  +=       0.065 * cos(0.63077087865  + 46386.74992582770 *T)
    sumX0  +=       0.069 * cos(3.59016732594  + 60530.24516825830 *T)
    sumX0  +=       0.068 * cos(6.19031950148  + 96678.14268052568 *T)
    sumX0  +=       0.049 * cos(2.44170258748  + 12323.17927852529 *T)
    sumX0  +=       0.050 * cos(3.95175048363  + 21954.40142688150 *T)
    sumX0  +=       0.049 * cos(0.26987394090  + 4164.55580709650 *T)
    sumX0  +=       0.059 * cos(2.11910349639  + 33794.78754101210 *T)
    sumX0  +=       0.050 * cos(3.21674377716  + 43739.04616344930 *T)
    sumX0  +=       0.059 * cos(4.28455887932  + 3981.73385156551 *T)
    sumX0  +=       0.058 * cos(1.59500717257  + 27707.29867681129 *T)
    sumX0  +=       0.051 * cos(3.00541030027  + 7669.21494381111 *T)
    sumX0  +=       0.048 * cos(4.80655958998  + 3646.59419483791 *T)
    sumX0  +=       0.065 * cos(3.20870046710  + 82534.64743809510 *T)
    sumX0  +=       0.056 * cos(4.19440679704  + 800.06494264891 *T)
    sumX0  +=       0.057 * cos(0.99121344005  + 11609.61872652869 *T)
    sumX0  +=       0.048 * cos(5.23363777755  + 26.54213728330 *T)
    sumX0  +=       0.047 * cos(1.10988199958  + 30640.10045611650 *T)
    sumX0  +=       0.060 * cos(1.97427905695  + 8982.56685182550 *T)
    sumX0  +=       0.049 * cos(3.72444308522  + 96563.24283557820 *T)
    sumX0  +=       0.050 * cos(5.17708029533  + 17583.90388883130 *T)
    sumX0  +=       0.046 * cos(2.75949485809  + 23539.46356884929 *T)
    sumX0  +=       0.048 * cos(2.83397317642  + 170.42905313570 *T)
    sumX0  +=       0.055 * cos(4.77764015710  + 735.63269604830 *T)
    sumX0  +=       0.054 * cos(4.33153690679  + 6680.98103591609 *T)
    sumX0  +=       0.060 * cos(1.72047044215  + 44033.88377205590 *T)
    sumX0  +=       0.064 * cos(4.40652103235  + 80181.29364935629 *T)
    sumX0  +=       0.059 * cos(5.18038765510  + 24279.35083569710 *T)
    sumX0  +=       0.049 * cos(5.16566622453  + 5490.54477900750 *T)
    sumX0  +=       0.058 * cos(1.24075349030  + 6490.02040477150 *T)
    sumX0  +=       0.058 * cos(5.40778008235  + 6076.61893017830 *T)
    sumX0  +=       0.046 * cos(6.06699951300  + 18208.10612510850 *T)
    sumX0  +=       0.045 * cos(1.62616739636  + 4480.46004178609 *T)
    sumX0  +=       0.055 * cos(3.56883579861  + 16097.92376776610 *T)
    sumX0  +=       0.055 * cos(0.42724314502  + 16097.43613279909 *T)
    sumX0  +=       0.061 * cos(5.10427594640  + 491.90710994230 *T)
    sumX0  +=       0.054 * cos(2.79730041476  + 6006.57389561050 *T)
    sumX0  +=       0.054 * cos(3.85123315789  + 6560.06543933930 *T)
    sumX0  +=       0.044 * cos(3.20288072834  + 5333.65642353809 *T)
    sumX0  +=       0.052 * cos(5.42298629738  + 244287.84382471119 *T)
    sumX0  +=       0.044 * cos(1.12093494405  + 2301.34199842589 *T)
    sumX0  +=       0.045 * cos(2.78871592108  + 33794.29990604510 *T)
    sumX0  +=       0.046 * cos(2.27794751316  + 4384.48081456829 *T)
    sumX0  +=       0.048 * cos(6.20709507263  + 35579.93505705350 *T)
    sumX0  +=       0.051 * cos(0.20647678216  + 18326.89394936390 *T)
    sumX0  +=       0.045 * cos(4.21002666289  + 533.45790092711 *T)
    sumX0  +=       0.046 * cos(4.72419290173  + 16703.30595098250 *T)
    sumX0  +=       0.047 * cos(0.82323755553  + 5209.22300831710 *T)
    sumX0  +=       0.051 * cos(2.88192111957  + 38650.41732368250 *T)
    sumX0  +=       0.046 * cos(4.04752097938  + 12555.34981720240 *T)
    sumX0  +=       0.052 * cos(6.21526623133  + 5992.34680160890 *T)
    sumX0  +=       0.052 * cos(0.43326734132  + 6574.29253334090 *T)
    sumX0  +=       0.044 * cos(5.37917268925  + 69942.19741831250 *T)
    sumX0  +=       0.049 * cos(5.19212576524  + 12146.91087359110 *T)
    sumX0  +=       0.053 * cos(3.10372910020  + 39301.85314471250 *T)
    sumX0  +=       0.045 * cos(6.22187415000  + 46360.93924099490 *T)
    sumX0  +=       0.043 * cos(3.11607958976  + 8402.08352785330 *T)
    sumX0  +=       0.043 * cos(1.76871622593  + 44137.19516685750 *T)
    sumX0  +=       0.049 * cos(1.97645700808  + 22029.96913718610 *T)
    sumX0  +=       0.044 * cos(3.04557845219  + 31441.92138724031 *T)
    sumX0  +=       0.042 * cos(1.28858480303  + 9225.29545579949 *T)
    sumX0  +=       0.045 * cos(4.79112566977  + 71519.54096076029 *T)
    sumX0  +=       0.051 * cos(4.52703512787  + 7762.18624153930 *T)
    sumX0  +=       0.041 * cos(0.43156198612  + 3735.48212924251 *T)
    sumX0  +=       0.057 * cos(2.81178684757  + 40398.43373674950 *T)
    sumX0  +=       0.047 * cos(4.46091937774  + 4707.98623122570 *T)
    sumX0  +=       0.056 * cos(0.65497165024  + 28664.07546774890 *T)
    sumX0  +=       0.050 * cos(4.15809563419  + 11919.38468415150 *T)
    sumX0  +=       0.047 * cos(0.89798744348  + 316428.47249139857 *T)
    sumX0  +=       0.041 * cos(4.27548568888  + 27278.71263392430 *T)
    sumX0  +=       0.056 * cos(2.85120735939  + 83659.52068988250 *T)
    sumX0  +=       0.048 * cos(2.50261741205  + 12662.37474468410 *T)
    sumX0  +=       0.054 * cos(0.83831240155  + 16627.12709789369 *T)
    sumX0  +=       0.052 * cos(4.41165168727  + 12571.91844177370 *T)
    sumX0  +=       0.043 * cos(4.69793524808  + 7322.34627826531 *T)
    sumX0  +=       0.041 * cos(3.56154893542  + 664.99986261351 *T)
    sumX0  +=       0.042 * cos(1.54738888492  + 5641.95442512570 *T)
    sumX0  +=       0.044 * cos(1.44765859641  + 6425.49829451110 *T)
    sumX0  +=       0.044 * cos(5.20087497624  + 6141.14104043870 *T)
    sumX0  +=       0.039 * cos(0.38205897719  + 22594.29871319550 *T)
    sumX0  +=       0.040 * cos(5.88641926833  + 20894.53186172529 *T)
    sumX0  +=       0.044 * cos(5.28307931868  + 5540.32960694230 *T)
    sumX0  +=       0.050 * cos(5.69294803866  + 17782.48825530069 *T)
    sumX0  +=       0.054 * cos(4.69164296803  + 6006.28467373350 *T)
    sumX0  +=       0.054 * cos(1.95689060462  + 6560.35466121630 *T)
    sumX0  +=       0.045 * cos(5.54499872381  + 142861.14741877469 *T)
    sumX0  +=       0.047 * cos(1.12158335918  + 24499.76347813590 *T)
    sumX0  +=       0.043 * cos(1.36615498859  + 7026.30972800750 *T)
    sumX0  +=       0.047 * cos(1.95434456607  + 955.84355909210 *T)
    sumX0  +=       0.047 * cos(4.50829221159  + 24485.53638413430 *T)
    sumX0  +=       0.040 * cos(2.60727518415  + 9388.24972689870 *T)
    sumX0  +=       0.043 * cos(5.81439105801  + 6359.58573875050 *T)
    sumX0  +=       0.042 * cos(3.96208965107  + 8635.69818627969 *T)
    sumX0  +=       0.051 * cos(2.46325624829  + 27831.79440179969 *T)
    sumX0  +=       0.042 * cos(1.94770311943  + 6334.60000533731 *T)
    sumX0  +=       0.042 * cos(4.70083045322  + 6232.03932961249 *T)
    sumX0  +=       0.052 * cos(3.53311914292  + 11905.16259068530 *T)
    sumX0  +=       0.040 * cos(1.50948151820  + 7076.09455594230 *T)
    sumX0  +=       0.040 * cos(0.29378456988  + 5444.35037972450 *T)
    sumX0  +=       0.040 * cos(0.07156369559  + 7122.28895522530 *T)
    sumX0  +=       0.038 * cos(2.24525725009  + 44933.49317367390 *T)
    sumX0  +=       0.037 * cos(0.27418012894  + 5715.60102974450 *T)
    sumX0  +=       0.037 * cos(0.09116813653  + 6851.03830520530 *T)
    sumX0  +=       0.042 * cos(0.79913949254  + 36147.65369478390 *T)
    sumX0  +=       0.042 * cos(2.85822305358  + 6238.17844783830 *T)
    sumX0  +=       0.042 * cos(3.79031051907  + 6328.46088711150 *T)
    sumX0  +=       0.040 * cos(3.64347872660  + 3214.89256315970 *T)
    sumX0  +=       0.043 * cos(0.65521409425  + 23141.31456544110 *T)
    sumX0  +=       0.041 * cos(3.95515629208  + 19804.58347409930 *T)
    sumX0  +=       0.038 * cos(2.90982798521  + 20.11150191529 *T)
    sumX0  +=       0.050 * cos(1.62788860301  + 84334.66158130829 *T)
    sumX0  +=       0.050 * cos(2.84763289309  + 60284.16619777939 *T)

    # EARTH X Coord
    sumX1  = 0
    sumX1  +=  154550.744 * cos(0.64605836878  + 0.24381748350 *T)
    sumX1  +=   51503.383 * cos(6.00263199393  + 12566.39551746630 *T)
    sumX1  +=    1290.763 * cos(5.95941652859  + 18849.47136745770 *T)
    sumX1  +=     702.576 * cos(1.75347943445  + 6283.31966747490 *T)
    sumX1  +=     430.422 * cos(2.45216492471  + 6282.83203250789 *T)
    sumX1  +=     212.689 * cos(1.73380190045  + 6279.79654912590 *T)
    sumX1  +=     212.524 * cos(4.91484799365  + 6286.84278582391 *T)
    sumX1  +=      62.308 * cos(0.36193297710  + 4705.97612502710 *T)
    sumX1  +=      59.808 * cos(3.81187956678  + 6257.02134767510 *T)
    sumX1  +=      59.474 * cos(2.83659652893  + 6309.61798727471 *T)
    sumX1  +=      48.914 * cos(5.21332670904  + 775.76642880750 *T)
    sumX1  +=      42.814 * cos(0.43864033644  + 1059.62574767270 *T)
    sumX1  +=      46.457 * cos(0.01089905463  + 7860.66320992270 *T)
    sumX1  +=      36.653 * cos(2.19635713914  + 5753.62870238030 *T)
    sumX1  +=      35.649 * cos(1.44752011203  + 5885.17066406670 *T)
    sumX1  +=      35.362 * cos(4.47215447256  + 6813.01063256950 *T)
    sumX1  +=      32.151 * cos(5.19677261546  + 6681.46867088311 *T)
    sumX1  +=      28.763 * cos(5.91618988907  + 25132.54721744910 *T)
    sumX1  +=      28.447 * cos(1.14976229491  + 6127.89926804070 *T)
    sumX1  +=      27.537 * cos(5.49920443604  + 6438.74006690910 *T)
    sumX1  +=      24.819 * cos(2.92230738957  + 5487.02166065850 *T)
    sumX1  +=      20.615 * cos(3.71821460067  + 7079.61767429131 *T)
    sumX1  +=      19.621 * cos(2.88650825392  + 5507.79705615090 *T)
    sumX1  +=      18.378 * cos(1.46521342874  + 11790.87290614230 *T)
    sumX1  +=      16.494 * cos(6.21543454587  + 11507.01358727710 *T)
    sumX1  +=      16.756 * cos(3.81938434057  + 7058.84227879890 *T)
    sumX1  +=      14.558 * cos(5.97530104893  + 6290.43321447570 *T)
    sumX1  +=      14.404 * cos(0.68122522191  + 6276.20612047410 *T)
    sumX1  +=      14.052 * cos(1.45747121073  + 796.54182429990 *T)
    sumX1  +=      12.261 * cos(4.15789187748  + 4693.75913722409 *T)
    sumX1  +=      12.758 * cos(0.34831751321  + 7.35736448430 *T)
    sumX1  +=      10.086 * cos(3.29242550455  + 3739.00524759150 *T)
    sumX1  +=       9.469 * cos(4.63920457850  + 6070.02057203690 *T)
    sumX1  +=      10.425 * cos(2.38242547117  + 6284.29998854310 *T)
    sumX1  +=      10.425 * cos(4.26610810148  + 6282.33934640670 *T)
    sumX1  +=       9.570 * cos(1.89427100067  + 4137.15425099970 *T)
    sumX1  +=       9.044 * cos(2.00972967096  + 6496.61876291290 *T)
    sumX1  +=       8.544 * cos(0.05731253621  + 1194.69082770810 *T)
    sumX1  +=       8.214 * cos(1.25401494002  + 1589.31671276730 *T)
    sumX1  +=       6.157 * cos(3.35676469316  + 8827.63408735830 *T)
    sumX1  +=       6.271 * cos(4.75475820102  + 8429.48508395010 *T)
    sumX1  +=       5.524 * cos(5.36941014190  + 4933.45225781610 *T)
    sumX1  +=       7.314 * cos(0.49102526454  + 4535.30325440790 *T)
    sumX1  +=       5.157 * cos(5.14965163127  + 11770.09751064990 *T)
    sumX1  +=       6.749 * cos(4.32513020462  + 5088.87265725030 *T)
    sumX1  +=       4.853 * cos(0.62462823219  + 6040.59106350090 *T)
    sumX1  +=       5.304 * cos(5.64244320178  + 3154.44326741210 *T)
    sumX1  +=       4.985 * cos(4.62777616911  + 12569.91863581531 *T)
    sumX1  +=       4.709 * cos(3.96386608147  + 5331.60126122430 *T)
    sumX1  +=       4.601 * cos(6.02376124126  + 6526.04827144891 *T)
    sumX1  +=       4.201 * cos(1.28125193997  + 7633.18707713370 *T)
    sumX1  +=       5.607 * cos(2.55800509408  + 5729.75026463250 *T)
    sumX1  +=       4.257 * cos(4.12450123624  + 3930.45351370310 *T)
    sumX1  +=       4.394 * cos(0.04709285339  + 12559.28197046550 *T)
    sumX1  +=       3.870 * cos(2.68681123019  + 7235.03807372550 *T)
    sumX1  +=       5.154 * cos(6.15930214427  + 8031.33608054190 *T)
    sumX1  +=       5.036 * cos(4.09009014859  + 6836.88907031730 *T)
    sumX1  +=       5.076 * cos(2.30606090748  + 7477.76667769950 *T)
    sumX1  +=       3.593 * cos(2.40570544354  + 12565.90788249930 *T)
    sumX1  +=       3.384 * cos(5.40354177904  + 10977.32262218251 *T)
    sumX1  +=       3.371 * cos(0.28961103245  + 11371.94850724170 *T)
    sumX1  +=       3.536 * cos(3.73353119602  + 4164.06817212950 *T)
    sumX1  +=       4.292 * cos(4.93786086114  + 1592.83983111630 *T)
    sumX1  +=       3.407 * cos(2.40599273330  + 3128.63258257930 *T)
    sumX1  +=       3.152 * cos(1.73356525688  + 5223.93773728571 *T)
    sumX1  +=       2.883 * cos(4.92077247219  + 1747.77259558350 *T)
    sumX1  +=       2.751 * cos(5.24449444697  + 7342.70159766410 *T)
    sumX1  +=       3.323 * cos(1.91880015934  + 801.57711364030 *T)
    sumX1  +=       2.999 * cos(2.25355241223  + 8636.18582124671 *T)
    sumX1  +=       2.496 * cos(3.52078273708  + 2145.92159899169 *T)
    sumX1  +=       3.358 * cos(4.25271343539  + 155.17658195069 *T)
    sumX1  +=       2.297 * cos(4.53559364224  + 17260.39847217390 *T)
    sumX1  +=       2.047 * cos(3.53660263792  + 1990.98883452450 *T)
    sumX1  +=       2.778 * cos(1.46152428305  + 5481.49873635110 *T)
    sumX1  +=       2.025 * cos(1.44003078455  + 951.47458876710 *T)
    sumX1  +=       2.663 * cos(5.71398175957  + 26.05450231630 *T)
    sumX1  +=       2.009 * cos(5.73390835065  + 4690.72365384210 *T)
    sumX1  +=       2.148 * cos(1.89688144273  + 537.04832957890 *T)
    sumX1  +=       1.853 * cos(2.84762029989  + 553.32558535889 *T)
    sumX1  +=       2.478 * cos(0.03933370984  + 1349.62359217529 *T)
    sumX1  +=       2.023 * cos(2.39095052485  + 522.82123557730 *T)
    sumX1  +=       1.735 * cos(2.91944410412  + 529.44714761109 *T)
    sumX1  +=       2.417 * cos(5.16772880951  + 7085.14059859870 *T)
    sumX1  +=       1.809 * cos(0.78457177103  + 397.90518592470 *T)
    sumX1  +=       1.893 * cos(4.22746619412  + 9438.00675237050 *T)
    sumX1  +=       1.708 * cos(0.29122424525  + 10989.05197501850 *T)
    sumX1  +=       1.898 * cos(2.34007370780  + 5216.33655531789 *T)
    sumX1  +=       1.820 * cos(1.87728974480  + 5230.56364931950 *T)
    sumX1  +=       1.541 * cos(1.10077450497  + 426.84200835950 *T)
    sumX1  +=       1.555 * cos(3.67004127739  + 13096.08648256090 *T)
    sumX1  +=       1.779 * cos(1.80817581905  + 12562.87239911730 *T)
    sumX1  +=       1.448 * cos(0.18764842398  + 5753.14106741329 *T)
    sumX1  +=       1.450 * cos(0.99962470222  + 6262.96434807611 *T)
    sumX1  +=       1.450 * cos(5.64890887043  + 6303.67498687370 *T)
    sumX1  +=       1.822 * cos(1.70526332053  + 10973.79950383350 *T)
    sumX1  +=       1.372 * cos(0.89281758077  + 7875.91568110771 *T)
    sumX1  +=       1.375 * cos(0.76961577554  + 7084.65296363170 *T)
    sumX1  +=       1.400 * cos(4.85224030747  + 12721.81591690050 *T)
    sumX1  +=       1.247 * cos(1.86305212941  + 2119.00767786191 *T)
    sumX1  +=       1.231 * cos(4.10244270612  + 18319.78040236310 *T)
    sumX1  +=       1.141 * cos(3.49241515175  + 1066.73929467350 *T)
    sumX1  +=       1.138 * cos(2.80935582144  + 2942.70724077510 *T)
    sumX1  +=       1.047 * cos(6.02902034733  + 5642.93474619389 *T)
    sumX1  +=       1.058 * cos(3.76081315695  + 242.97242145751 *T)
    sumX1  +=       1.025 * cos(5.32138364615  + 10447.14402212089 *T)
    sumX1  +=       0.973 * cos(0.24159402695  + 640.14110379750 *T)
    sumX1  +=       1.114 * cos(5.94057908115  + 15721.08260236190 *T)
    sumX1  +=       0.963 * cos(2.13180404708  + 2389.13783793270 *T)
    sumX1  +=       0.941 * cos(5.78706319005  + 20426.81490990550 *T)
    sumX1  +=       1.068 * cos(0.37393702270  + 529.93478257810 *T)
    sumX1  +=       0.929 * cos(3.11135132378  + 10575.65050042530 *T)
    sumX1  +=       0.899 * cos(1.07875107899  + 16496.60521368590 *T)
    sumX1  +=       0.889 * cos(0.29953586851  + 6277.79674316750 *T)
    sumX1  +=       0.889 * cos(0.06581239696  + 6288.84259178230 *T)
    sumX1  +=       1.018 * cos(4.23379398746  + 12540.09719766650 *T)
    sumX1  +=       1.060 * cos(1.05626464973  + 5760.74224938110 *T)
    sumX1  +=       1.017 * cos(5.59172649862  + 6805.89708556871 *T)
    sumX1  +=       0.797 * cos(5.36416276600  + 14314.41193053330 *T)
    sumX1  +=       0.795 * cos(1.67821896162  + 6286.60602489270 *T)
    sumX1  +=       0.795 * cos(4.97031461103  + 6280.03331005710 *T)
    sumX1  +=       0.800 * cos(4.75445763497  + 12029.59100537090 *T)
    sumX1  +=       0.719 * cos(3.86746073982  + 9623.93209417470 *T)
    sumX1  +=       0.698 * cos(0.23124353834  + 6148.25458743950 *T)
    sumX1  +=       0.686 * cos(5.36448356218  + 5856.72147659890 *T)
    sumX1  +=       0.673 * cos(5.56184807840  + 12964.54452087450 *T)
    sumX1  +=       0.902 * cos(2.26778049127  + 13368.21644859010 *T)
    sumX1  +=       0.678 * cos(0.13408295783  + 6418.38474751031 *T)
    sumX1  +=       0.671 * cos(1.28940504615  + 6709.91785835091 *T)
    sumX1  +=       0.744 * cos(4.25820705300  + 12043.81809937250 *T)
    sumX1  +=       0.706 * cos(4.09030374624  + 16730.70750707931 *T)
    sumX1  +=       0.661 * cos(3.99296008106  + 14712.56093394150 *T)
    sumX1  +=       0.633 * cos(0.86914955896  + 4292.57465043390 *T)
    sumX1  +=       0.621 * cos(3.31200729066  + 13119.96492030870 *T)
    sumX1  +=       0.638 * cos(0.47109166303  + 4690.23601887509 *T)
    sumX1  +=       0.625 * cos(1.88693051693  + 13518.11392371690 *T)
    sumX1  +=       0.637 * cos(6.16733887432  + 5746.51515537950 *T)
    sumX1  +=       0.628 * cos(1.85369027207  + 4686.64559022330 *T)
    sumX1  +=       0.653 * cos(4.28657823528  + 3929.96587873609 *T)
    sumX1  +=       0.578 * cos(5.33616352365  + 5088.38502228330 *T)
    sumX1  +=       0.563 * cos(4.82044552939  + 10447.63165708790 *T)
    sumX1  +=       0.610 * cos(0.48716946879  + 6820.12417957030 *T)
    sumX1  +=       0.666 * cos(3.30009380787  + 3634.37720703489 *T)
    sumX1  +=       0.720 * cos(0.47572131456  + 13916.26292712510 *T)
    sumX1  +=       0.620 * cos(5.87390167695  + 31415.62306744050 *T)
    sumX1  +=       0.553 * cos(4.28779208332  + 6259.44122972711 *T)
    sumX1  +=       0.553 * cos(2.36074148933  + 6307.19810522270 *T)
    sumX1  +=       0.523 * cos(0.75768878632  + 14143.73905991410 *T)
    sumX1  +=       0.524 * cos(5.54806074141  + 12139.79732659030 *T)
    sumX1  +=       0.569 * cos(0.12658241682  + 12036.70455237170 *T)
    sumX1  +=       0.514 * cos(2.30528734098  + 4700.87268422489 *T)
    sumX1  +=       0.563 * cos(2.21509710432  + 11014.86265985130 *T)
    sumX1  +=       0.495 * cos(3.58067034949  + 13362.69352428270 *T)
    sumX1  +=       0.481 * cos(2.50655660751  + 6279.43833211690 *T)
    sumX1  +=       0.481 * cos(4.14197696514  + 6287.20100283290 *T)
    sumX1  +=       0.466 * cos(4.51385713822  + 10177.50149701710 *T)
    sumX1  +=       0.440 * cos(0.45716811540  + 1577.58735993130 *T)
    sumX1  +=       0.437 * cos(3.18132586326  + 11499.90004027630 *T)
    sumX1  +=       0.452 * cos(5.55409870988  + 12573.50906446710 *T)
    sumX1  +=       0.499 * cos(1.08746970886  + 12410.97511803210 *T)
    sumX1  +=       0.440 * cos(2.15051971988  + 24073.16528725990 *T)
    sumX1  +=       0.587 * cos(4.71917094180  + 6.86972951729 *T)
    sumX1  +=       0.388 * cos(2.04118602600  + 7860.17557495569 *T)
    sumX1  +=       0.398 * cos(5.77147994683  + 8274.06468451590 *T)
    sumX1  +=       0.423 * cos(1.16599451585  + 4694.24677219110 *T)
    sumX1  +=       0.530 * cos(3.00672799320  + 12592.69383726610 *T)
    sumX1  +=       0.379 * cos(2.02388854283  + 6180.22689325630 *T)
    sumX1  +=       0.379 * cos(4.62464502982  + 6386.41244169350 *T)
    sumX1  +=       0.377 * cos(5.90352173005  + 7872.39256275871 *T)
    sumX1  +=       0.419 * cos(5.62219874401  + 5327.71992586630 *T)
    sumX1  +=       0.383 * cos(6.24432820927  + 6247.29180079750 *T)
    sumX1  +=       0.383 * cos(0.40420536338  + 6319.34753415231 *T)
    sumX1  +=       0.375 * cos(5.62315844522  + 2352.62233628830 *T)
    sumX1  +=       0.457 * cos(4.19200440747  + 6077.13411903770 *T)
    sumX1  +=       0.457 * cos(2.45652916518  + 6489.50521591210 *T)
    sumX1  +=       0.481 * cos(1.89483272077  + 4292.08701546690 *T)
    sumX1  +=       0.344 * cos(0.28668049892  + 18451.32236404950 *T)
    sumX1  +=       0.453 * cos(0.72577236980  + 2787.28684134090 *T)
    sumX1  +=       0.364 * cos(4.39782645564  + 6245.29199483910 *T)
    sumX1  +=       0.364 * cos(2.25070711701  + 6321.34734011070 *T)
    sumX1  +=       0.327 * cos(2.71208539186  + 12323.66691349230 *T)
    sumX1  +=       0.358 * cos(4.87210297876  + 9917.45305702629 *T)
    sumX1  +=       0.342 * cos(0.54833716779  + 6262.54427198250 *T)
    sumX1  +=       0.342 * cos(6.10019640486  + 6304.09506296730 *T)
    sumX1  +=       0.318 * cos(0.13364568371  + 11926.49823115230 *T)
    sumX1  +=       0.320 * cos(2.70295484544  + 77714.01528560399 *T)
    sumX1  +=       0.354 * cos(1.02467818407  + 7238.91940908350 *T)
    sumX1  +=       0.318 * cos(5.32503015719  + 6254.87048000710 *T)
    sumX1  +=       0.318 * cos(1.32350341546  + 6311.76885494270 *T)
    sumX1  +=       0.304 * cos(2.00970548291  + 12779.69461290430 *T)
    sumX1  +=       0.343 * cos(3.71302827314  + 6298.57213865991 *T)
    sumX1  +=       0.343 * cos(2.93550529951  + 6268.06719628990 *T)
    sumX1  +=       0.359 * cos(4.09955266779  + 1052.51220067191 *T)
    sumX1  +=       0.398 * cos(0.12298117861  + 1551.28904013150 *T)
    sumX1  +=       0.306 * cos(2.58873546625  + 5863.83502359970 *T)
    sumX1  +=       0.300 * cos(1.77608786075  + 90280.16698558679 *T)
    sumX1  +=       0.307 * cos(3.29263568042  + 3893.93801205869 *T)
    sumX1  +=       0.297 * cos(0.51221148619  + 5429.63565075589 *T)
    sumX1  +=       0.365 * cos(2.74542548477  + 10022.08109758290 *T)
    sumX1  +=       0.336 * cos(2.72364909606  + 17782.97589026770 *T)
    sumX1  +=       0.295 * cos(4.05850212848  + 6702.80431135010 *T)
    sumX1  +=       0.288 * cos(2.18795828306  + 18073.94875613370 *T)
    sumX1  +=       0.287 * cos(3.33776996109  + 1577.09972496430 *T)
    sumX1  +=       0.378 * cos(0.39946174567  + 2353.10997125530 *T)
    sumX1  +=       0.289 * cos(3.04361347297  + 213.54291292150 *T)
    sumX1  +=       0.347 * cos(0.72838501040  + 5223.45010231870 *T)
    sumX1  +=       0.321 * cos(2.23787969105  + 17797.20298426930 *T)
    sumX1  +=       0.289 * cos(5.30650717041  + 220.65645992230 *T)
    sumX1  +=       0.260 * cos(1.04284839158  + 955.35592412510 *T)
    sumX1  +=       0.268 * cos(3.62847841862  + 14945.07235607090 *T)
    sumX1  +=       0.242 * cos(4.34386674771  + 2544.55823736690 *T)
    sumX1  +=       0.245 * cos(2.31719119888  + 7632.69944216670 *T)
    sumX1  +=       0.234 * cos(4.07393437960  + 1596.43025976811 *T)
    sumX1  +=       0.267 * cos(5.83055956746  + 206.42936592071 *T)
    sumX1  +=       0.282 * cos(4.50260656021  + 13341.91812879030 *T)
    sumX1  +=       0.279 * cos(3.13773429847  + 4731.78680985990 *T)
    sumX1  +=       0.290 * cos(2.03216763842  + 5642.44206009270 *T)
    sumX1  +=       0.298 * cos(0.93822937797  + 17790.08943726851 *T)
    sumX1  +=       0.270 * cos(2.16529088691  + 12168.24651405810 *T)
    sumX1  +=       0.294 * cos(2.52895912407  + 2146.40923395870 *T)
    sumX1  +=       0.214 * cos(0.93543623291  + 8030.84844557490 *T)
    sumX1  +=       0.207 * cos(1.60966002008  + 213.05527795450 *T)
    sumX1  +=       0.216 * cos(5.60316710178  + 3185.43584474911 *T)
    sumX1  +=       0.195 * cos(2.87143133187  + 5884.68302909969 *T)
    sumX1  +=       0.199 * cos(3.93236191602  + 1748.26023055050 *T)
    sumX1  +=       0.250 * cos(4.61596023287  + 6924.19727485710 *T)
    sumX1  +=       0.194 * cos(2.36749003244  + 641.12142486571 *T)
    sumX1  +=       0.203 * cos(2.25356528795  + 6503.73230991370 *T)
    sumX1  +=       0.203 * cos(4.39496828470  + 6062.90702503610 *T)
    sumX1  +=       0.258 * cos(5.63025007570  + 796.05418933290 *T)
    sumX1  +=       0.189 * cos(1.63377883836  + 11506.52595231009 *T)
    sumX1  +=       0.186 * cos(0.12949416246  + 18209.57408114370 *T)
    sumX1  +=       0.184 * cos(4.20955149479  + 12566.46282776910 *T)
    sumX1  +=       0.180 * cos(5.76808682495  + 853.44019923550 *T)
    sumX1  +=       0.185 * cos(4.68506649147  + 3495.78900865049 *T)
    sumX1  +=       0.205 * cos(4.93663483706  + 5849.12029463110 *T)
    sumX1  +=       0.187 * cos(5.91135912176  + 10213.52936369450 *T)
    sumX1  +=       0.187 * cos(2.05685668523  + 6290.36590417291 *T)
    sumX1  +=       0.187 * cos(4.59167688742  + 6276.27343077690 *T)
    sumX1  +=       0.235 * cos(5.91611965384  + 9779.35249360890 *T)
    sumX1  +=       0.173 * cos(1.38702654435  + 19651.29229858150 *T)
    sumX1  +=       0.184 * cos(4.65365446464  + 12566.32820716351 *T)
    sumX1  +=       0.230 * cos(2.23774195109  + 12567.37583853451 *T)
    sumX1  +=       0.166 * cos(1.32597581910  + 110.45013870291 *T)
    sumX1  +=       0.189 * cos(2.84387211432  + 3.76693583251 *T)
    sumX1  +=       0.166 * cos(1.19482070134  + 433.95555536030 *T)
    sumX1  +=       0.186 * cos(4.46969004938  + 5863.34738863270 *T)
    sumX1  +=       0.184 * cos(3.01631602481  + 10983.94853421629 *T)
    sumX1  +=       0.184 * cos(3.38049875173  + 6037.48802124550 *T)
    sumX1  +=       0.156 * cos(2.34433370292  + 2648.69864295650 *T)
    sumX1  +=       0.182 * cos(3.47271515850  + 10969.72144021469 *T)
    sumX1  +=       0.175 * cos(3.26771548230  + 6529.15131370430 *T)
    sumX1  +=       0.154 * cos(5.67725689556  + 6453.99253809410 *T)
    sumX1  +=       0.154 * cos(0.97127667708  + 6112.64679685570 *T)
    sumX1  +=       0.164 * cos(4.52273465985  + 12353.09642202830 *T)
    sumX1  +=       0.146 * cos(4.05729660897  + 149.80701461810 *T)
    sumX1  +=       0.184 * cos(0.20513580951  + 11015.35029481830 *T)
    sumX1  +=       0.191 * cos(2.29402324776  + 3894.42564702570 *T)
    sumX1  +=       0.144 * cos(3.76178854682  + 18636.17227201970 *T)
    sumX1  +=       0.146 * cos(1.99164382254  + 13760.84252769090 *T)
    sumX1  +=       0.142 * cos(1.34565609994  + 4156.95462512869 *T)
    sumX1  +=       0.185 * cos(3.72101619927  + 7234.55043875850 *T)
    sumX1  +=       0.133 * cos(0.01824676794  + 18139.53831889940 *T)
    sumX1  +=       0.131 * cos(4.75893185547  + 5507.30942118390 *T)
    sumX1  +=       0.146 * cos(5.03857686201  + 18875.28205229050 *T)
    sumX1  +=       0.128 * cos(3.41348197806  + 6069.53293706989 *T)
    sumX1  +=       0.128 * cos(0.18086618365  + 16200.52890701769 *T)
    sumX1  +=       0.123 * cos(5.45134858107  + 20.53157800890 *T)
    sumX1  +=       0.121 * cos(0.98195264587  + 17256.87535382490 *T)
    sumX1  +=       0.123 * cos(0.10275700246  + 6393.52598869430 *T)
    sumX1  +=       0.123 * cos(0.26259126301  + 6173.11334625550 *T)
    sumX1  +=       0.117 * cos(4.47711713050  + 18852.99448580670 *T)
    sumX1  +=       0.120 * cos(1.03913122140  + 9381.20349020070 *T)
    sumX1  +=       0.113 * cos(2.33706395068  + 65147.37595065419 *T)
    sumX1  +=       0.137 * cos(3.68199834086  + 16201.01654198470 *T)
    sumX1  +=       0.113 * cos(5.07439785395  + 6836.40143535029 *T)
    sumX1  +=       0.123 * cos(4.42839393502  + 12565.41519639810 *T)
    sumX1  +=       0.132 * cos(6.08414353296  + 17655.02435723310 *T)
    sumX1  +=       0.113 * cos(4.12294240934  + 10575.16286545830 *T)
    sumX1  +=       0.113 * cos(2.09696400140  + 10213.04172872750 *T)
    sumX1  +=       0.116 * cos(2.23828708218  + 17253.28492517310 *T)
    sumX1  +=       0.120 * cos(2.28034205827  + 10420.23010099111 *T)
    sumX1  +=       0.109 * cos(0.50597705280  + 633.02755679670 *T)
    sumX1  +=       0.103 * cos(0.11832394316  + 6295.04902031090 *T)
    sumX1  +=       0.103 * cos(0.24702432231  + 6271.59031463890 *T)
    sumX1  +=       0.104 * cos(5.58102765204  + 23013.78335707070 *T)
    sumX1  +=       0.105 * cos(3.75605193168  + 11790.38527117530 *T)
    sumX1  +=       0.109 * cos(2.65465737803  + 3340.36860921629 *T)
    sumX1  +=       0.102 * cos(2.59681923716  + 10970.20907518170 *T)
    sumX1  +=       0.101 * cos(4.42190708814  + 4803.96545844350 *T)
    sumX1  +=       0.114 * cos(2.21750546661  + 4171.18171913030 *T)
    sumX1  +=       0.111 * cos(3.66232090402  + 1582.20316576650 *T)
    sumX1  +=       0.107 * cos(2.47220154064  + 23543.47432216530 *T)
    sumX1  +=       0.099 * cos(2.67705199399  + 10973.31186886650 *T)
    sumX1  +=       0.102 * cos(5.67042027327  + 6549.92670919670 *T)
    sumX1  +=       0.102 * cos(0.97811329938  + 6016.71262575310 *T)
    sumX1  +=       0.120 * cos(2.49695982999  + 11514.12713427790 *T)
    sumX1  +=       0.095 * cos(2.69903868611  + 17267.51201917470 *T)
    sumX1  +=       0.121 * cos(0.05336928445  + 13625.77744765550 *T)
    sumX1  +=       0.100 * cos(2.04355401267  + 10976.83498721549 *T)
    sumX1  +=       0.090 * cos(3.68840859712  + 76.50988875911 *T)
    sumX1  +=       0.100 * cos(4.37665365923  + 8661.99650607950 *T)
    sumX1  +=       0.089 * cos(5.33255005639  + 1350.11122714230 *T)
    sumX1  +=       0.087 * cos(4.33390332130  + 9917.94069199330 *T)
    sumX1  +=       0.086 * cos(5.93949751968  + 12809.12412144031 *T)
    sumX1  +=       0.095 * cos(1.34358559692  + 12345.98287502750 *T)
    sumX1  +=       0.084 * cos(1.54530867800  + 775.27879384050 *T)
    sumX1  +=       0.106 * cos(5.14280637669  + 11216.52810780750 *T)
    sumX1  +=       0.095 * cos(2.50046187937  + 12012.82611462390 *T)
    sumX1  +=       0.082 * cos(0.01600918310  + 5643.42238116090 *T)
    sumX1  +=       0.081 * cos(3.84772162011  + 11614.67711121570 *T)
    sumX1  +=       0.084 * cos(1.14755262027  + 6255.91811137810 *T)
    sumX1  +=       0.084 * cos(5.50098095238  + 6310.72122357170 *T)
    sumX1  +=       0.080 * cos(0.34711727404  + 6923.21695378890 *T)
    sumX1  +=       0.085 * cos(5.79487336888  + 5635.82119919310 *T)
    sumX1  +=       0.084 * cos(3.04423309467  + 10440.03047512009 *T)
    sumX1  +=       0.080 * cos(4.19702378753  + 3583.58484815730 *T)
    sumX1  +=       0.092 * cos(0.87932993874  + 10818.37910439930 *T)
    sumX1  +=       0.077 * cos(5.70308186355  + 6993.25271603320 *T)
    sumX1  +=       0.080 * cos(1.36958890049  + 13521.99525907490 *T)
    sumX1  +=       0.104 * cos(5.42045444776  + 5650.04829319470 *T)
    sumX1  +=       0.093 * cos(0.29330855376  + 4732.27444482690 *T)
    sumX1  +=       0.085 * cos(0.17489631763  + 22805.49174851010 *T)
    sumX1  +=       0.086 * cos(0.81273058913  + 12360.20996902910 *T)
    sumX1  +=       0.075 * cos(4.00156639703  + 5120.35732810009 *T)
    sumX1  +=       0.090 * cos(0.62558959530  + 6370.62787201471 *T)
    sumX1  +=       0.090 * cos(6.02294397735  + 6196.01146293510 *T)
    sumX1  +=       0.082 * cos(0.19767948100  + 18842.35782045690 *T)
    sumX1  +=       0.101 * cos(6.08995730057  + 3097.64000524229 *T)
    sumX1  +=       0.094 * cos(5.45896547971  + 10177.01386205009 *T)
    sumX1  +=       0.083 * cos(5.31992956371  + 18415.75962958090 *T)
    sumX1  +=       0.073 * cos(6.09826272325  + 949.41942645330 *T)
    sumX1  +=       0.101 * cos(2.86248284119  + 3104.68624194030 *T)
    sumX1  +=       0.074 * cos(5.96635290969  + 13517.62628874989 *T)
    sumX1  +=       0.092 * cos(2.35039346805  + 16858.72635041670 *T)
    sumX1  +=       0.075 * cos(0.00198703030  + 1059.13811270570 *T)
    sumX1  +=       0.085 * cos(6.09242867296  + 398.39282089170 *T)
    sumX1  +=       0.069 * cos(0.89145782253  + 11713.19913571430 *T)
    sumX1  +=       0.084 * cos(4.35477901935  + 2378.92065608810 *T)
    sumX1  +=       0.069 * cos(0.61262618132  + 11712.71150074729 *T)
    sumX1  +=       0.075 * cos(1.62832949134  + 30356.24113725130 *T)
    sumX1  +=       0.067 * cos(4.61214364579  + 3154.93090237910 *T)
    sumX1  +=       0.076 * cos(4.83480996889  + 18429.98672358250 *T)
    sumX1  +=       0.068 * cos(4.26265140929  + 11925.51791008410 *T)
    sumX1  +=       0.067 * cos(3.24609638890  + 10454.25756912169 *T)
    sumX1  +=       0.073 * cos(6.07047724510  + 15670.83794192309 *T)
    sumX1  +=       0.065 * cos(0.20786131115  + 6438.25243194210 *T)
    sumX1  +=       0.069 * cos(3.29499829591  + 17298.42614480970 *T)
    sumX1  +=       0.087 * cos(5.09754892709  + 3904.15519390330 *T)
    sumX1  +=       0.076 * cos(0.73384165663  + 5231.05128428650 *T)
    sumX1  +=       0.064 * cos(3.69394501139  + 3496.27664361750 *T)
    sumX1  +=       0.078 * cos(4.31639568529  + 8672.21368792411 *T)
    sumX1  +=       0.066 * cos(3.41816290372  + 14143.25142494710 *T)
    sumX1  +=       0.063 * cos(0.99537559299  + 24357.02460612510 *T)
    sumX1  +=       0.063 * cos(5.29204726636  + 15720.59496739490 *T)
    sumX1  +=       0.061 * cos(3.72699038600  + 16460.57734700850 *T)
    sumX1  +=       0.073 * cos(4.50408758477  + 9387.76209193169 *T)
    sumX1  +=       0.057 * cos(2.80099560279  + 17259.91083720689 *T)
    sumX1  +=       0.059 * cos(0.69432195316  + 9778.86485864190 *T)
    sumX1  +=       0.058 * cos(3.24953926506  + 155.66421691770 *T)
    sumX1  +=       0.060 * cos(3.51548937348  + 34570.31015233610 *T)
    sumX1  +=       0.057 * cos(4.46074610322  + 149854.64395229140 *T)
    sumX1  +=       0.056 * cos(5.33393809517  + 12456.18919624690 *T)
    sumX1  +=       0.056 * cos(0.91240244730  + 17996.27498570570 *T)
    sumX1  +=       0.056 * cos(2.17966368743  + 11764.57458634250 *T)
    sumX1  +=       0.054 * cos(2.96145051332  + 4705.48849006010 *T)
    sumX1  +=       0.055 * cos(4.53949491324  + 13915.77529215809 *T)
    sumX1  +=       0.061 * cos(0.15191781763  + 28237.47727687290 *T)
    sumX1  +=       0.065 * cos(5.87045161229  + 7335.58805066330 *T)
    sumX1  +=       0.072 * cos(0.77556931591  + 6055.84353468590 *T)
    sumX1  +=       0.072 * cos(5.87296425674  + 6510.79580026390 *T)
    sumX1  +=       0.071 * cos(0.23083621209  + 12545.62012197390 *T)
    sumX1  +=       0.061 * cos(1.27148175728  + 6312.74917601091 *T)
    sumX1  +=       0.061 * cos(5.37705181537  + 6253.89015893890 *T)
    sumX1  +=       0.052 * cos(0.26640147565  + 5326.54287653730 *T)
    sumX1  +=       0.054 * cos(1.21229891621  + 2699.49100183409 *T)
    sumX1  +=       0.052 * cos(2.44818566509  + 8983.05448679250 *T)
    sumX1  +=       0.066 * cos(2.35337977576  + 14919.26167123810 *T)
    sumX1  +=       0.050 * cos(5.93502263343  + 5572.89898394960 *T)
    sumX1  +=       0.051 * cos(0.76420455409  + 16522.41589851870 *T)
    sumX1  +=       0.049 * cos(4.18738074622  + 35579.93505705350 *T)
    sumX1  +=       0.053 * cos(1.36303997543  + 6208.53806890760 *T)
    sumX1  +=       0.053 * cos(5.28549359722  + 6358.10126604220 *T)
    sumX1  +=       0.049 * cos(4.92672897893  + 21393.78578734110 *T)
    sumX1  +=       0.049 * cos(6.04327871377  + 5547.44315394310 *T)
    sumX1  +=       0.049 * cos(0.60525485888  + 7019.19618100671 *T)
    sumX1  +=       0.049 * cos(2.59123701567  + 12416.83232033170 *T)
    sumX1  +=       0.050 * cos(5.65157500725  + 12592.20620229910 *T)
    sumX1  +=       0.050 * cos(0.07520350695  + 26084.26562369970 *T)
    sumX1  +=       0.051 * cos(3.40680645218  + 23006.66981006990 *T)
    sumX1  +=       0.058 * cos(2.46543782438  + 15141.14697682849 *T)
    sumX1  +=       0.047 * cos(0.97665096734  + 24279.35083569710 *T)
    sumX1  +=       0.049 * cos(4.98006572526  + 6816.53375091851 *T)
    sumX1  +=       0.049 * cos(1.66846784739  + 5750.10558403130 *T)
    sumX1  +=       0.046 * cos(3.45134537804  + 19379.16233255230 *T)
    sumX1  +=       0.052 * cos(4.17729143118  + 16737.33341911309 *T)
    sumX1  +=       0.047 * cos(2.09298679166  + 24066.05174025910 *T)
    sumX1  +=       0.046 * cos(5.73796654718  + 104351.85638378028 *T)
    sumX1  +=       0.063 * cos(1.54245276158  + 8662.48414104651 *T)
    sumX1  +=       0.048 * cos(1.71424177408  + 26735.70144472970 *T)
    sumX1  +=       0.046 * cos(5.48854983581  + 12132.68377958950 *T)
    sumX1  +=       0.050 * cos(4.49443808542  + 24602.85625235450 *T)
    sumX1  +=       0.054 * cos(0.03975141461  + 7834.36489012290 *T)
    sumX1  +=       0.045 * cos(5.05887207146  + 161000.92955515761 *T)
    sumX1  +=       0.049 * cos(4.51484833854  + 29303.97275406290 *T)
    sumX1  +=       0.049 * cos(3.00236481482  + 10984.43616918330 *T)
    sumX1  +=       0.045 * cos(5.42122497288  + 45584.92899470390 *T)
    sumX1  +=       0.046 * cos(3.44074843825  + 5244.29305668450 *T)
    sumX1  +=       0.048 * cos(1.96431249949  + 23581.01435983410 *T)
    sumX1  +=       0.049 * cos(0.60097746518  + 19650.80466361450 *T)
    sumX1  +=       0.049 * cos(4.98977220505  + 29289.74566006130 *T)
    sumX1  +=       0.044 * cos(3.01088620106  + 6390.98319141350 *T)
    sumX1  +=       0.044 * cos(3.63764737159  + 6175.65614353630 *T)
    sumX1  +=       0.051 * cos(1.08500510566  + 17789.60180230149 *T)
    sumX1  +=       0.042 * cos(1.47304377234  + 24705.94902657310 *T)
    sumX1  +=       0.042 * cos(4.33726197214  + 15663.79170522509 *T)
    sumX1  +=       0.043 * cos(1.51170891404  + 16460.08971204149 *T)
    sumX1  +=       0.043 * cos(1.85670214007  + 553.81322032590 *T)
    sumX1  +=       0.041 * cos(4.78695133916  + 17370.60479339330 *T)
    sumX1  +=       0.053 * cos(1.02071442021  + 13119.47728534169 *T)
    sumX1  +=       0.042 * cos(0.40756986853  + 4164.55580709650 *T)
    sumX1  +=       0.041 * cos(3.62600931231  + 23020.89690407150 *T)
    sumX1  +=       0.045 * cos(2.13148338898  + 54247.16931826690 *T)
    sumX1  +=       0.040 * cos(3.00393862246  + 32243.25468339710 *T)
    sumX1  +=       0.040 * cos(0.45945430396  + 24336.24921063270 *T)
    sumX1  +=       0.040 * cos(0.14772670564  + 11769.60987568290 *T)
    sumX1  +=       0.045 * cos(4.68905632727  + 16723.10632511149 *T)
    sumX1  +=       0.045 * cos(1.54746367368  + 16723.59396007850 *T)
    sumX1  +=       0.042 * cos(4.32247442790  + 8827.14645239129 *T)
    sumX1  +=       0.040 * cos(0.00254937606  + 8402.08352785330 *T)
    sumX1  +=       0.039 * cos(5.06793905702  + 16062.42834360030 *T)
    sumX1  +=       0.044 * cos(2.87431244066  + 5856.23384163189 *T)
    sumX1  +=       0.038 * cos(2.71291798518  + 6297.54676147650 *T)
    sumX1  +=       0.038 * cos(3.93561558747  + 6269.09257347330 *T)
    sumX1  +=       0.041 * cos(4.95313128130  + 15508.85894075790 *T)
    sumX1  +=       0.038 * cos(0.25558032671  + 6394.50630976251 *T)
    sumX1  +=       0.038 * cos(0.10976793876  + 6172.13302518730 *T)
    sumX1  +=       0.039 * cos(3.31895424299  + 29826.55017215670 *T)
    sumX1  +=       0.046 * cos(5.02561530595  + 17297.93850984270 *T)
    sumX1  +=       0.041 * cos(2.47236149910  + 6309.13035230770 *T)
    sumX1  +=       0.038 * cos(1.11170997081  + 29089.05523346850 *T)
    sumX1  +=       0.038 * cos(6.16488744565  + 11933.61177815310 *T)
    sumX1  +=       0.037 * cos(4.91170884792  + 26482.41462710790 *T)
    sumX1  +=       0.037 * cos(2.07642652738  + 20452.62559473830 *T)
    sumX1  +=       0.037 * cos(5.02138933475  + 18073.46112116669 *T)
    sumX1  +=       0.036 * cos(1.41330554556  + 11499.41240530929 *T)
    sumX1  +=       0.045 * cos(0.04410300421  + 3340.85624418330 *T)
    sumX1  +=       0.049 * cos(5.78167784671  + 18216.68762814450 *T)
    sumX1  +=       0.038 * cos(6.16266694488  + 18202.46053414290 *T)
    sumX1  +=       0.041 * cos(5.48085477947  + 5216.82419028490 *T)
    sumX1  +=       0.036 * cos(4.64059514280  + 4060.97539791089 *T)
    sumX1  +=       0.048 * cos(0.43106257187  + 951.96222373410 *T)
    sumX1  +=       0.041 * cos(2.33852383906  + 18848.98373249069 *T)
    sumX1  +=       0.038 * cos(6.17960213131  + 26709.89075989690 *T)
    sumX1  +=       0.036 * cos(4.69601378640  + 28230.43104017490 *T)
    sumX1  +=       0.036 * cos(0.39760453863  + 24491.18197509989 *T)
    sumX1  +=       0.036 * cos(6.28050310342  + 9924.56660402710 *T)
    sumX1  +=       0.048 * cos(0.32621587842  + 8982.56685182550 *T)
    sumX1  +=       0.037 * cos(5.17748996647  + 22484.09239197610 *T)
    sumX1  +=       0.047 * cos(1.64189885412  + 6040.10342853389 *T)
    sumX1  +=       0.034 * cos(5.26052334707  + 6418.94499248490 *T)
    sumX1  +=       0.034 * cos(1.38801022558  + 6147.69434246491 *T)
    sumX1  +=       0.034 * cos(2.09927479760  + 9380.71585523369 *T)
    sumX1  +=       0.036 * cos(5.01642644412  + 7238.43177411650 *T)
    sumX1  +=       0.034 * cos(1.97967235505  + 17272.12782500990 *T)
    sumX1  +=       0.037 * cos(2.80234472663  + 58953.38926077750 *T)
    sumX1  +=       0.034 * cos(1.90506635542  + 29026.72904699130 *T)
    sumX1  +=       0.033 * cos(2.30422309015  + 12721.32828193349 *T)
    sumX1  +=       0.035 * cos(3.20661909223  + 7322.34627826531 *T)
    sumX1  +=       0.044 * cos(2.79016934175  + 3981.73385156551 *T)
    sumX1  +=       0.033 * cos(2.02741588975  + 19804.58347409930 *T)
    sumX1  +=       0.033 * cos(4.13127522861  + 173567.08125514039 *T)
    sumX1  +=       0.033 * cos(4.59578048276  + 205.94173095370 *T)
    sumX1  +=       0.033 * cos(3.87810188817  + 23013.29572210369 *T)
    sumX1  +=       0.034 * cos(0.56182222293  + 49515.13869092350 *T)
    sumX1  +=       0.034 * cos(4.95130100456  + 38500.51984855570 *T)
    sumX1  +=       0.032 * cos(4.38300866562  + 17686.99666304990 *T)
    sumX1  +=       0.036 * cos(0.10343586811  + 5641.95442512570 *T)
    sumX1  +=       0.032 * cos(4.48170445266  + 21228.63584102931 *T)
    sumX1  +=       0.033 * cos(1.70919210465  + 34520.55312686430 *T)
    sumX1  +=       0.033 * cos(4.08300924175  + 8635.69818627969 *T)
    sumX1  +=       0.031 * cos(0.30962555206  + 5746.02752041250 *T)
    sumX1  +=       0.041 * cos(4.09101330171  + 21202.33752122950 *T)
    sumX1  +=       0.037 * cos(1.21597687953  + 7349.81514466491 *T)
    sumX1  +=       0.034 * cos(0.61019544663  + 16496.11757871889 *T)
    sumX1  +=       0.032 * cos(2.47223060088  + 33019.26492968810 *T)
    sumX1  +=       0.032 * cos(6.25412877228  + 21150.56954840009 *T)
    sumX1  +=       0.037 * cos(3.41910013684  + 29296.85920706210 *T)
    sumX1  +=       0.036 * cos(3.45170661535  + 27511.22405605370 *T)
    sumX1  +=       0.035 * cos(2.20855027459  + 11933.12414318609 *T)
    sumX1  +=       0.035 * cos(2.69261850768  + 11918.89704918449 *T)
    sumX1  +=       0.035 * cos(0.83074174405  + 29062.75691366870 *T)
    sumX1  +=       0.030 * cos(2.96012497927  + 12490.12944619070 *T)
    sumX1  +=       0.032 * cos(0.38688555099  + 6334.60000533731 *T)
    sumX1  +=       0.032 * cos(6.26164802166  + 6232.03932961249 *T)
    sumX1  +=       0.029 * cos(2.48914084100  + 419.24082639170 *T)
    sumX1  +=       0.029 * cos(2.48914084100  + 419.72846135871 *T)
    sumX1  +=       0.029 * cos(0.09098498823  + 16858.23871544969 *T)
    sumX1  +=       0.030 * cos(4.30269873097  + 16208.13008898551 *T)
    sumX1  +=       0.029 * cos(2.26254947152  + 27278.71263392430 *T)
    sumX1  +=       0.032 * cos(3.16446260317  + 5017.26455388150 *T)
    sumX1  +=       0.028 * cos(1.87847577782  + 24080.27883426070 *T)
    sumX1  +=       0.030 * cos(2.57345584103  + 23539.95120381630 *T)
    sumX1  +=       0.029 * cos(6.17121903514  + 18606.25512851669 *T)
    sumX1  +=       0.030 * cos(3.16071544619  + 30665.91114094930 *T)
    sumX1  +=       0.034 * cos(5.86279015300  + 6660.69327539070 *T)
    sumX1  +=       0.034 * cos(0.78574341965  + 5905.94605955911 *T)
    sumX1  +=       0.038 * cos(5.12222686816  + 31571.04346687470 *T)
    sumX1  +=       0.038 * cos(2.95935499043  + 16061.94070863329 *T)
    sumX1  +=       0.029 * cos(0.07805875622  + 23539.46356884929 *T)
    sumX1  +=       0.033 * cos(1.31621330654  + 3738.51761262449 *T)
    sumX1  +=       0.028 * cos(4.45185289388  + 2942.21960580809 *T)
    sumX1  +=       0.032 * cos(4.32330584281  + 15110.70993734970 *T)
    sumX1  +=       0.030 * cos(1.39195112722  + 55023.17956455790 *T)
    sumX1  +=       0.027 * cos(0.53286078729  + 233141.55822184496 *T)
    sumX1  +=       0.030 * cos(2.32405719366  + 12587.17091295870 *T)
    sumX1  +=       0.028 * cos(5.54771281521  + 34911.65589357450 *T)
    sumX1  +=       0.027 * cos(6.17188911074  + 1692.40948698591 *T)
    sumX1  +=       0.027 * cos(4.56663555607  + 2544.07060239989 *T)
    sumX1  +=       0.027 * cos(3.42500443489  + 24382.83529095790 *T)
    sumX1  +=       0.027 * cos(5.99215939135  + 6205.64589704690 *T)
    sumX1  +=       0.027 * cos(0.65637418130  + 6360.99343790291 *T)
    sumX1  +=       0.027 * cos(0.24480889857  + 31172.89446346650 *T)
    sumX1  +=       0.026 * cos(3.76734683925  + 36949.47462590770 *T)
    sumX1  +=       0.034 * cos(6.02674277322  + 18053.17336064130 *T)
    sumX1  +=       0.027 * cos(2.32047706036  + 32367.34147369110 *T)
    sumX1  +=       0.027 * cos(1.77181133426  + 4487.57358878689 *T)
    sumX1  +=       0.028 * cos(3.80331914659  + 10239.34004852730 *T)
    sumX1  +=       0.026 * cos(6.01470923824  + 4796.85191144269 *T)
    sumX1  +=       0.026 * cos(6.27941236443  + 6226.42129253930 *T)
    sumX1  +=       0.026 * cos(0.36912120822  + 6340.21804241050 *T)
    sumX1  +=       0.029 * cos(0.18788001431  + 5974.04134481910 *T)
    sumX1  +=       0.029 * cos(0.17746825116  + 6592.59799013070 *T)
    sumX1  +=       0.026 * cos(3.39870851542  + 19402.55313533309 *T)
    sumX1  +=       0.028 * cos(2.42845493949  + 46386.74992582770 *T)
    sumX1  +=       0.032 * cos(0.33223863826  + 9225.78309076650 *T)
    sumX1  +=       0.026 * cos(2.54665382210  + 9910.33951002549 *T)
    sumX1  +=       0.025 * cos(5.74078884001  + 15265.64270181689 *T)
    sumX1  +=       0.029 * cos(5.90495655038  + 52176.05010063190 *T)
    sumX1  +=       0.027 * cos(4.22338082584  + 62883.59895699710 *T)
    sumX1  +=       0.024 * cos(0.55027097677  + 11617.21990849651 *T)
    sumX1  +=       0.024 * cos(2.62016281496  + 5120.84496306710 *T)
    sumX1  +=       0.028 * cos(3.25064295131  + 6680.98103591609 *T)
    sumX1  +=       0.024 * cos(4.40689701323  + 522.33360061030 *T)
    sumX1  +=       0.025 * cos(3.86362291910  + 6263.64990657511 *T)
    sumX1  +=       0.025 * cos(2.78491065355  + 6302.98942837470 *T)
    sumX1  +=       0.029 * cos(2.44405132476  + 23581.50199480110 *T)
    sumX1  +=       0.024 * cos(0.50709291410  + 29424.87805039950 *T)
    sumX1  +=       0.028 * cos(2.66240986547  + 16737.82105408010 *T)
    sumX1  +=       0.024 * cos(1.98880441412  + 22743.16556203289 *T)
    sumX1  +=       0.024 * cos(5.13039706771  + 22743.65319699990 *T)
    sumX1  +=       0.024 * cos(5.85302121512  + 25158.84553724890 *T)
    sumX1  +=       0.025 * cos(3.95115098429  + 21424.22282681990 *T)
    sumX1  +=       0.025 * cos(0.80955833070  + 21424.71046178690 *T)
    sumX1  +=       0.024 * cos(1.92863541994  + 14712.07329897449 *T)
    sumX1  +=       0.028 * cos(4.50601984314  + 5760.25461441409 *T)
    sumX1  +=       0.024 * cos(3.73455640738  + 31969.19247028290 *T)
    sumX1  +=       0.023 * cos(0.62345777089  + 17893.18221148710 *T)
    sumX1  +=       0.030 * cos(3.85563309519  + 8584.90548338430 *T)
    sumX1  +=       0.024 * cos(3.21746306158  + 23550.58786916610 *T)
    sumX1  +=       0.023 * cos(1.36063534536  + 6717.03140535170 *T)
    sumX1  +=       0.022 * cos(3.99741791962  + 7445.79437188270 *T)
    sumX1  +=       0.023 * cos(5.13650786055  + 5849.60792959810 *T)
    sumX1  +=       0.025 * cos(0.81555419576  + 14155.46841275010 *T)
    sumX1  +=       0.023 * cos(3.58569412080  + 6294.36536773881 *T)
    sumX1  +=       0.023 * cos(3.06283945185  + 6272.27396721100 *T)
    sumX1  +=       0.024 * cos(4.24597953395  + 18208.10612510850 *T)
    sumX1  +=       0.024 * cos(1.10438688036  + 18208.59376007550 *T)
    sumX1  +=       0.021 * cos(5.78251945962  + 22779.68106367730 *T)
    sumX1  +=       0.029 * cos(3.31448732534  + 28628.58004358310 *T)
    sumX1  +=       0.021 * cos(5.13427576878  + 40879.68432212730 *T)
    sumX1  +=       0.021 * cos(4.08575884387  + 30221.17605721590 *T)
    sumX1  +=       0.022 * cos(1.38262024424  + 4379.88285497370 *T)
    sumX1  +=       0.022 * cos(5.26591980546  + 8186.75647997610 *T)
    sumX1  +=       0.022 * cos(4.52225402167  + 28760.05469496669 *T)
    sumX1  +=       0.026 * cos(3.41053115091  + 18422.87317658170 *T)
    sumX1  +=       0.020 * cos(0.80720267360  + 12036.21691740469 *T)
    sumX1  +=       0.020 * cos(4.78724181293  + 28313.04498717750 *T)
    sumX1  +=       0.027 * cos(0.67000770864  + 34513.50689016630 *T)
    sumX1  +=       0.020 * cos(2.07630932980  + 34115.35788675810 *T)
    sumX1  +=       0.020 * cos(1.72148703277  + 21548.71855180830 *T)
    sumX1  +=       0.020 * cos(3.68576430300  + 17654.53672226610 *T)
    sumX1  +=       0.025 * cos(3.53209868087  + 55798.70217588190 *T)
    sumX1  +=       0.020 * cos(0.81961551511  + 20597.48778052470 *T)
    sumX1  +=       0.020 * cos(2.60684350676  + 18875.76968725751 *T)
    sumX1  +=       0.022 * cos(0.44774356245  + 1588.82907780029 *T)
    sumX1  +=       0.024 * cos(1.20852424503  + 9411.70843257070 *T)
    sumX1  +=       0.019 * cos(5.85945140949  + 536.56069461190 *T)
    sumX1  +=       0.020 * cos(4.99560905506  + 6187.34044025710 *T)
    sumX1  +=       0.020 * cos(1.65292451759  + 6379.29889469270 *T)
    sumX1  +=       0.022 * cos(4.44471803531  + 18326.89394936390 *T)
    sumX1  +=       0.022 * cos(2.18348213630  + 6233.56264200310 *T)
    sumX1  +=       0.022 * cos(4.46505143635  + 6333.07669294670 *T)
    sumX1  +=       0.018 * cos(2.03036294822  + 5972.47886860650 *T)
    sumX1  +=       0.018 * cos(4.61817062443  + 6594.16046634330 *T)
    sumX1  +=       0.019 * cos(4.28397604583  + 6948.07571260490 *T)
    sumX1  +=       0.019 * cos(2.36455752682  + 5618.56362234490 *T)
    sumX1  +=       0.019 * cos(0.18461366541  + 83974.07917321339 *T)
    sumX1  +=       0.019 * cos(4.29574443999  + 84020.10309797739 *T)
    sumX1  +=       0.020 * cos(3.84957027230  + 50316.95962204730 *T)
    sumX1  +=       0.020 * cos(2.85391796921  + 9070.36269133230 *T)

    # EARTH X Coord
    sumX2  = 0
    sumX2  +=   52911.498 * cos(3.32403354915  + 6283.31966747490 *T)
    sumX2  +=    6074.441 * cos(2.05485843872  + 0.24381748350 *T)
    sumX2  +=    2406.871 * cos(4.23086027149  + 12566.39551746630 *T)
    sumX2  +=      96.033 * cos(4.14485769516  + 18849.47136745770 *T)
    sumX2  +=      29.888 * cos(0.73436192593  + 6282.83203250789 *T)
    sumX2  +=      14.021 * cos(3.21121684298  + 6279.79654912590 *T)
    sumX2  +=      13.375 * cos(3.31400005464  + 6286.84278582391 *T)
    sumX2  +=       8.148 * cos(2.19949162987  + 6309.61798727471 *T)
    sumX2  +=       8.195 * cos(4.38954108022  + 6257.02134767510 *T)
    sumX2  +=       8.001 * cos(5.91956929382  + 6127.89926804070 *T)
    sumX2  +=       7.754 * cos(0.72231700953  + 6438.74006690910 *T)
    sumX2  +=       7.223 * cos(5.38114271434  + 83997.09113559539 *T)
    sumX2  +=       4.616 * cos(6.16703227370  + 710.17686604180 *T)
    sumX2  +=       3.060 * cos(4.83964573312  + 4705.97612502710 *T)
    sumX2  +=       3.717 * cos(2.45040909403  + 5507.79705615090 *T)
    sumX2  +=       2.991 * cos(4.05993241608  + 25132.54721744910 *T)
    sumX2  +=       3.235 * cos(4.10584554352  + 7058.84227879890 *T)
    sumX2  +=       2.393 * cos(3.40776365917  + 775.76642880750 *T)
    sumX2  +=       2.034 * cos(4.34141229883  + 6290.43321447570 *T)
    sumX2  +=       2.028 * cos(2.28023581144  + 6276.20612047410 *T)
    sumX2  +=       2.064 * cos(1.58527926677  + 7860.66320992270 *T)
    sumX2  +=       1.914 * cos(4.42314112548  + 5487.02166065850 *T)
    sumX2  +=       1.770 * cos(5.05862783743  + 1059.62574767270 *T)
    sumX2  +=       1.761 * cos(3.37902852940  + 5753.62870238030 *T)
    sumX2  +=       1.558 * cos(2.13660228192  + 7079.61767429131 *T)
    sumX2  +=       1.551 * cos(2.12932375831  + 6040.59106350090 *T)
    sumX2  +=       1.436 * cos(2.85243683562  + 5885.17066406670 *T)
    sumX2  +=       1.469 * cos(4.51404554645  + 6526.04827144891 *T)
    sumX2  +=       1.319 * cos(3.46234944603  + 6681.46867088311 *T)
    sumX2  +=       1.180 * cos(2.99170595830  + 6813.01063256950 *T)
    sumX2  +=       1.337 * cos(1.03982518814  + 5729.75026463250 *T)
    sumX2  +=       1.039 * cos(5.82323058792  + 6282.33934640670 *T)
    sumX2  +=       1.313 * cos(2.16977183919  + 529.93478257810 *T)
    sumX2  +=       1.041 * cos(0.76562479642  + 6284.29998854310 *T)
    sumX2  +=       1.202 * cos(5.59909179750  + 6836.88907031730 *T)
    sumX2  +=       0.904 * cos(2.42848578586  + 5331.60126122430 *T)
    sumX2  +=       0.870 * cos(3.83740101245  + 4933.45225781610 *T)
    sumX2  +=       1.149 * cos(4.59530313333  + 167284.00540514899 *T)
    sumX2  +=       1.069 * cos(5.80442130983  + 1577.09972496430 *T)
    sumX2  +=       0.772 * cos(4.23369730112  + 7235.03807372550 *T)
    sumX2  +=       0.846 * cos(2.93878609085  + 11790.87290614230 *T)
    sumX2  +=       0.742 * cos(0.30768536261  + 4137.15425099970 *T)
    sumX2  +=       0.966 * cos(3.62307276164  + 11856.46246890800 *T)
    sumX2  +=       0.664 * cos(2.78855764568  + 7633.18707713370 *T)
    sumX2  +=       0.662 * cos(4.76048207091  + 1194.69082770810 *T)
    sumX2  +=       0.659 * cos(1.23774551244  + 11507.01358727710 *T)
    sumX2  +=       0.853 * cos(5.22418483353  + 4535.30325440790 *T)
    sumX2  +=       0.823 * cos(5.77365627974  + 155.17658195069 *T)
    sumX2  +=       0.783 * cos(5.87375907691  + 5088.87265725030 *T)
    sumX2  +=       0.785 * cos(5.03738396313  + 2353.10997125530 *T)
    sumX2  +=       0.694 * cos(1.93706172082  + 801.57711364030 *T)
    sumX2  +=       0.566 * cos(6.00002295544  + 796.54182429990 *T)
    sumX2  +=       0.684 * cos(2.68347359202  + 3128.63258257930 *T)
    sumX2  +=       0.499 * cos(2.99674498554  + 12569.91863581531 *T)
    sumX2  +=       0.596 * cos(1.38539217884  + 8031.33608054190 *T)
    sumX2  +=       0.486 * cos(6.11316978391  + 1589.31671276730 *T)
    sumX2  +=       0.483 * cos(6.25283705744  + 8429.48508395010 *T)
    sumX2  +=       0.441 * cos(0.27873568624  + 6496.61876291290 *T)
    sumX2  +=       0.412 * cos(5.58840963851  + 4693.75913722409 *T)
    sumX2  +=       0.561 * cos(0.75091675093  + 7477.76667769950 *T)
    sumX2  +=       0.390 * cos(6.20968064366  + 6070.02057203690 *T)
    sumX2  +=       0.405 * cos(1.55299195368  + 3739.00524759150 *T)
    sumX2  +=       0.501 * cos(3.39229516665  + 1592.83983111630 *T)
    sumX2  +=       0.392 * cos(5.39455294636  + 26.05450231630 *T)
    sumX2  +=       0.369 * cos(4.36331143713  + 553.32558535889 *T)
    sumX2  +=       0.483 * cos(1.36538995021  + 12036.70455237170 *T)
    sumX2  +=       0.489 * cos(2.13588971106  + 5223.45010231870 *T)
    sumX2  +=       0.481 * cos(4.74275484789  + 10213.52936369450 *T)
    sumX2  +=       0.448 * cos(0.85608336591  + 156137.71980228281 *T)
    sumX2  +=       0.321 * cos(2.96911879180  + 951.47458876710 *T)
    sumX2  +=       0.320 * cos(2.00218171333  + 1990.98883452450 *T)
    sumX2  +=       0.361 * cos(0.59519501420  + 12565.90788249930 *T)
    sumX2  +=       0.361 * cos(0.08584353459  + 12721.81591690050 *T)
    sumX2  +=       0.412 * cos(5.91338851626  + 398.39282089170 *T)
    sumX2  +=       0.304 * cos(0.96839213314  + 4690.72365384210 *T)
    sumX2  +=       0.356 * cos(1.28855678128  + 5481.49873635110 *T)
    sumX2  +=       0.283 * cos(2.23992178340  + 242.97242145751 *T)
    sumX2  +=       0.362 * cos(3.88633843420  + 9438.00675237050 *T)
    sumX2  +=       0.248 * cos(4.75322854174  + 8827.63408735830 *T)
    sumX2  +=       0.252 * cos(1.16164987849  + 3154.44326741210 *T)
    sumX2  +=       0.257 * cos(1.79452152688  + 11371.94850724170 *T)
    sumX2  +=       0.327 * cos(3.02081233401  + 14143.73905991410 *T)
    sumX2  +=       0.230 * cos(0.23367695814  + 1747.77259558350 *T)
    sumX2  +=       0.289 * cos(5.24697142787  + 7085.14059859870 *T)
    sumX2  +=       0.297 * cos(1.59965866765  + 1349.62359217529 *T)
    sumX2  +=       0.209 * cos(0.27341674720  + 11770.09751064990 *T)
    sumX2  +=       0.207 * cos(4.92862210539  + 4164.06817212950 *T)
    sumX2  +=       0.188 * cos(5.64673454004  + 7875.91568110771 *T)
    sumX2  +=       0.189 * cos(0.60545229996  + 2389.13783793270 *T)
    sumX2  +=       0.194 * cos(0.41046959310  + 10977.32262218251 *T)
    sumX2  +=       0.176 * cos(4.08570919572  + 7342.70159766410 *T)
    sumX2  +=       0.164 * cos(2.79150408226  + 5223.93773728571 *T)
    sumX2  +=       0.205 * cos(3.21370431881  + 10973.79950383350 *T)
    sumX2  +=       0.205 * cos(2.25101785528  + 13368.21644859010 *T)
    sumX2  +=       0.139 * cos(4.61869477858  + 10575.65050042530 *T)
    sumX2  +=       0.155 * cos(5.81412821499  + 12410.97511803210 *T)
    sumX2  +=       0.152 * cos(0.10500467147  + 3930.45351370310 *T)
    sumX2  +=       0.129 * cos(5.52193535090  + 397.90518592470 *T)
    sumX2  +=       0.127 * cos(0.96654930095  + 12592.69383726610 *T)
    sumX2  +=       0.132 * cos(0.16052766322  + 8636.18582124671 *T)
    sumX2  +=       0.122 * cos(4.78358483655  + 13119.96492030870 *T)
    sumX2  +=       0.163 * cos(0.93280092706  + 5507.30942118390 *T)
    sumX2  +=       0.114 * cos(5.70431693679  + 17260.39847217390 *T)
    sumX2  +=       0.118 * cos(2.37854817453  + 4292.57465043390 *T)
    sumX2  +=       0.148 * cos(5.63695738547  + 17790.08943726851 *T)
    sumX2  +=       0.123 * cos(3.75785910050  + 12562.87239911730 *T)
    sumX2  +=       0.103 * cos(3.36994741884  + 13518.11392371690 *T)
    sumX2  +=       0.138 * cos(3.83865460083  + 12168.24651405810 *T)
    sumX2  +=       0.107 * cos(4.33977826153  + 10989.05197501850 *T)
    sumX2  +=       0.104 * cos(5.27997888785  + 2145.92159899169 *T)
    sumX2  +=       0.096 * cos(1.97792917854  + 6283.25235717211 *T)
    sumX2  +=       0.099 * cos(2.27831564394  + 13096.08648256090 *T)
    sumX2  +=       0.096 * cos(1.53382620869  + 6283.38697777770 *T)
    sumX2  +=       0.084 * cos(5.20596620689  + 16730.70750707931 *T)
    sumX2  +=       0.111 * cos(3.96322491245  + 12540.09719766650 *T)
    sumX2  +=       0.087 * cos(6.02515097465  + 10177.50149701710 *T)
    sumX2  +=       0.078 * cos(4.18422621711  + 12323.66691349230 *T)
    sumX2  +=       0.076 * cos(5.70862044549  + 6709.91785835091 *T)
    sumX2  +=       0.105 * cos(3.53845405180  + 5642.44206009270 *T)
    sumX2  +=       0.073 * cos(0.89693480363  + 5856.72147659890 *T)
    sumX2  +=       0.102 * cos(0.66758153957  + 250570.91967470257 *T)
    sumX2  +=       0.070 * cos(5.22505818809  + 5753.14106741329 *T)
    sumX2  +=       0.071 * cos(0.56166089035  + 14314.41193053330 *T)
    sumX2  +=       0.092 * cos(1.96485842296  + 13916.26292712510 *T)
    sumX2  +=       0.070 * cos(2.53450052921  + 7084.65296363170 *T)
    sumX2  +=       0.091 * cos(3.10960565577  + 6924.19727485710 *T)
    sumX2  +=       0.083 * cos(3.98989749432  + 31415.62306744050 *T)
    sumX2  +=       0.088 * cos(5.01539409795  + 71430.45180064559 *T)
    sumX2  +=       0.074 * cos(1.96538390591  + 529.44714761109 *T)
    sumX2  +=       0.085 * cos(5.97852929798  + 95143.37673846159 *T)
    sumX2  +=       0.061 * cos(4.23114381660  + 8274.06468451590 *T)
    sumX2  +=       0.060 * cos(5.52155170015  + 16496.60521368590 *T)
    sumX2  +=       0.058 * cos(1.47891111226  + 5642.93474619389 *T)
    sumX2  +=       0.078 * cos(2.41141378465  + 213.54291292150 *T)
    sumX2  +=       0.059 * cos(4.67526538731  + 640.14110379750 *T)
    sumX2  +=       0.057 * cos(2.06129454632  + 4690.23601887509 *T)
    sumX2  +=       0.056 * cos(3.80170948025  + 13341.91812879030 *T)
    sumX2  +=       0.065 * cos(3.38872777921  + 4731.78680985990 *T)
    sumX2  +=       0.051 * cos(4.81168622601  + 3893.93801205869 *T)
    sumX2  +=       0.050 * cos(4.17369676080  + 12573.50906446710 *T)
    sumX2  +=       0.050 * cos(4.41669813786  + 3634.37720703489 *T)
    sumX2  +=       0.049 * cos(2.27712508812  + 18319.78040236310 *T)
    sumX2  +=       0.066 * cos(5.49585584500  + 2787.28684134090 *T)
    sumX2  +=       0.049 * cos(1.99905022377  + 13362.69352428270 *T)
    sumX2  +=       0.048 * cos(3.76233131596  + 12964.54452087450 *T)
    sumX2  +=       0.053 * cos(4.12374270394  + 14919.26167123810 *T)
    sumX2  +=       0.058 * cos(1.23083144160  + 1577.58735993130 *T)
    sumX2  +=       0.052 * cos(1.27828874996  + 15721.08260236190 *T)
    sumX2  +=       0.061 * cos(3.42876383277  + 4292.08701546690 *T)
    sumX2  +=       0.044 * cos(0.31046230867  + 24073.16528725990 *T)
    sumX2  +=       0.044 * cos(0.38954461131  + 10447.14402212089 *T)
    sumX2  +=       0.042 * cos(4.09986255075  + 20426.81490990550 *T)
    sumX2  +=       0.059 * cos(1.01599200374  + 84672.23202702119 *T)
    sumX2  +=       0.051 * cos(5.38279738858  + 6370.62787201471 *T)
    sumX2  +=       0.051 * cos(1.26573618407  + 6196.01146293510 *T)
    sumX2  +=       0.046 * cos(0.52171400874  + 2119.00767786191 *T)
    sumX2  +=       0.041 * cos(4.09327874029  + 3185.43584474911 *T)
    sumX2  +=       0.056 * cos(3.27189314076  + 10976.83498721549 *T)
    sumX2  +=       0.051 * cos(2.37524851068  + 9437.51911740350 *T)
    sumX2  +=       0.049 * cos(3.21159567950  + 239424.63407183639 *T)
    sumX2  +=       0.036 * cos(6.20393463674  + 3495.78900865049 *T)
    sumX2  +=       0.044 * cos(3.69396549327  + 18139.53831889940 *T)
    sumX2  +=       0.035 * cos(3.61386877088  + 3929.96587873609 *T)
    sumX2  +=       0.044 * cos(1.66014811397  + 8661.99650607950 *T)
    sumX2  +=       0.046 * cos(3.78615165617  + 3894.42564702570 *T)
    sumX2  +=       0.034 * cos(3.81444510796  + 6262.54427198250 *T)
    sumX2  +=       0.034 * cos(5.98049768477  + 6304.09506296730 *T)
    sumX2  +=       0.035 * cos(3.58408872678  + 18875.28205229050 *T)
    sumX2  +=       0.035 * cos(6.09075949155  + 10420.23010099111 *T)
    sumX2  +=       0.032 * cos(0.62535692558  + 23013.78335707070 *T)
    sumX2  +=       0.034 * cos(3.61147349841  + 7.35736448430 *T)
    sumX2  +=       0.045 * cos(1.48783646485  + 3340.85624418330 *T)
    sumX2  +=       0.035 * cos(0.38590076970  + 10213.04172872750 *T)
    sumX2  +=       0.031 * cos(5.38472872675  + 14712.56093394150 *T)
    sumX2  +=       0.030 * cos(4.39715700541  + 12809.12412144031 *T)
    sumX2  +=       0.039 * cos(1.14782571257  + 9779.35249360890 *T)
    sumX2  +=       0.041 * cos(5.61324415849  + 82576.73740351178 *T)
    sumX2  +=       0.031 * cos(0.76672410297  + 22779.68106367730 *T)
    sumX2  +=       0.028 * cos(4.07975589699  + 2942.70724077510 *T)
    sumX2  +=       0.028 * cos(2.36060849192  + 5429.63565075589 *T)
    sumX2  +=       0.027 * cos(4.04275590321  + 11014.86265985130 *T)
    sumX2  +=       0.027 * cos(2.11023646997  + 9623.93209417470 *T)
    sumX2  +=       0.025 * cos(1.72579553961  + 18209.57408114370 *T)
    sumX2  +=       0.026 * cos(2.55064187749  + 9381.20349020070 *T)
    sumX2  +=       0.025 * cos(2.68812943588  + 3583.58484815730 *T)
    sumX2  +=       0.030 * cos(5.98905764002  + 10447.63165708790 *T)
    sumX2  +=       0.032 * cos(4.08880306095  + 796.05418933290 *T)
    sumX2  +=       0.023 * cos(6.09814489767  + 9917.45305702629 *T)
    sumX2  +=       0.026 * cos(0.90493478768  + 12012.82611462390 *T)
    sumX2  +=       0.024 * cos(1.88327707702  + 14143.25142494710 *T)
    sumX2  +=       0.023 * cos(2.26727226510  + 26709.89075989690 *T)
    sumX2  +=       0.021 * cos(3.86895197800  + 7632.69944216670 *T)
    sumX2  +=       0.021 * cos(2.37138453193  + 17256.87535382490 *T)
    sumX2  +=       0.028 * cos(1.07450535792  + 12353.09642202830 *T)
    sumX2  +=       0.022 * cos(4.83463880427  + 77714.01528560399 *T)
    sumX2  +=       0.021 * cos(5.98744320172  + 6993.25271603320 *T)
    sumX2  +=       0.020 * cos(0.31265663265  + 6836.40143535029 *T)
    sumX2  +=       0.020 * cos(5.44332565028  + 1748.26023055050 *T)
    sumX2  +=       0.026 * cos(2.02675372626  + 9225.78309076650 *T)
    sumX2  +=       0.023 * cos(0.80832702606  + 149854.64395229140 *T)
    sumX2  +=       0.026 * cos(5.31358634432  + 2544.07060239989 *T)
    sumX2  +=       0.018 * cos(2.28255799994  + 11614.67711121570 *T)
    sumX2  +=       0.020 * cos(4.61413285151  + 426.84200835950 *T)
    sumX2  +=       0.019 * cos(3.66100662326  + 30640.10045611650 *T)
    sumX2  +=       0.018 * cos(4.77794649120  + 155.66421691770 *T)
    sumX2  +=       0.018 * cos(1.89582517195  + 11926.49823115230 *T)
    sumX2  +=       0.022 * cos(3.74943338657  + 17298.42614480970 *T)
    sumX2  +=       0.019 * cos(3.25511318252  + 18073.46112116669 *T)
    sumX2  +=       0.020 * cos(2.60030261119  + 19651.29229858150 *T)
    sumX2  +=       0.024 * cos(1.57388276311  + 5856.23384163189 *T)
    sumX2  +=       0.023 * cos(4.78368329384  + 72850.80553272919 *T)
    sumX2  +=       0.017 * cos(3.95703430839  + 8983.05448679250 *T)
    sumX2  +=       0.016 * cos(5.17069195690  + 16460.57734700850 *T)
    sumX2  +=       0.016 * cos(1.20099644560  + 17259.91083720689 *T)
    sumX2  +=       0.021 * cos(3.77366558760  + 16858.72635041670 *T)
    sumX2  +=       0.020 * cos(0.39427566601  + 23543.47432216530 *T)
    sumX2  +=       0.021 * cos(3.79445696120  + 13367.72881362310 *T)
    sumX2  +=       0.020 * cos(4.04662298246  + 2146.40923395870 *T)
    sumX2  +=       0.014 * cos(0.14269366851  + 22484.09239197610 *T)
    sumX2  +=       0.014 * cos(4.66600220934  + 16200.52890701769 *T)
    sumX2  +=       0.015 * cos(4.21847977142  + 20199.33877711650 *T)
    sumX2  +=       0.019 * cos(2.81420484440  + 8672.21368792411 *T)
    sumX2  +=       0.019 * cos(4.46915157584  + 18073.94875613370 *T)
    sumX2  +=       0.013 * cos(4.71429939378  + 16201.01654198470 *T)
    sumX2  +=       0.017 * cos(1.38925168648  + 7860.17557495569 *T)
    sumX2  +=       0.013 * cos(4.69427401677  + 18451.32236404950 *T)
    sumX2  +=       0.016 * cos(3.33196861521  + 5572.89898394960 *T)
    sumX2  +=       0.012 * cos(2.70989366066  + 17996.27498570570 *T)
    sumX2  +=       0.012 * cos(0.72221452149  + 19403.04077030010 *T)
    sumX2  +=       0.012 * cos(5.53894834507  + 14945.07235607090 *T)
    sumX2  +=       0.015 * cos(2.04879024946  + 12559.28197046550 *T)
    sumX2  +=       0.012 * cos(4.37879205675  + 10818.37910439930 *T)
    sumX2  +=       0.015 * cos(0.57625721169  + 12567.37583853451 *T)
    sumX2  +=       0.011 * cos(1.78471307843  + 28767.16824196750 *T)
    sumX2  +=       0.013 * cos(1.09547571668  + 90280.16698558679 *T)
    sumX2  +=       0.011 * cos(5.02230424673  + 22805.49174851010 *T)
    sumX2  +=       0.011 * cos(2.24605985689  + 35372.13108345990 *T)
    sumX2  +=       0.010 * cos(6.26009139838  + 11506.52595231009 *T)
    sumX2  +=       0.014 * cos(1.93802583909  + 21228.63584102931 *T)
    sumX2  +=       0.010 * cos(0.13064402023  + 12779.69461290430 *T)
    sumX2  +=       0.010 * cos(6.05309072952  + 23141.80220040810 *T)
    sumX2  +=       0.010 * cos(1.18002886182  + 1350.11122714230 *T)
    sumX2  +=       0.010 * cos(2.50762435685  + 10575.16286545830 *T)
    sumX2  +=       0.010 * cos(2.55643759758  + 22345.50419359170 *T)
    sumX2  +=       0.010 * cos(5.33273079625  + 9778.86485864190 *T)
    sumX2  +=       0.009 * cos(1.49151911226  + 29826.55017215670 *T)
    sumX2  +=       0.010 * cos(4.24000912367  + 23581.50199480110 *T)
    sumX2  +=       0.011 * cos(0.39610887139  + 25158.84553724890 *T)

    # EARTH X Coord
    sumX3  = 0
    sumX3  +=     232.790 * cos(3.40634928966  + 0.24381748350 *T)
    sumX3  +=      76.843 * cos(2.52439403387  + 12566.39551746630 *T)
    sumX3  +=      35.331 * cos(3.34616699853  + 6283.31966747490 *T)
    sumX3  +=       5.282 * cos(2.40489818589  + 18849.47136745770 *T)
    sumX3  +=       1.631 * cos(5.43859760597  + 6282.83203250789 *T)
    sumX3  +=       1.483 * cos(2.22678849266  + 6438.74006690910 *T)
    sumX3  +=       1.479 * cos(4.40151760084  + 6127.89926804070 *T)
    sumX3  +=       0.652 * cos(4.70876083993  + 6279.79654912590 *T)
    sumX3  +=       0.656 * cos(1.58695055624  + 6286.84278582391 *T)
    sumX3  +=       0.317 * cos(3.00237328519  + 6526.04827144891 *T)
    sumX3  +=       0.318 * cos(3.63024148359  + 6040.59106350090 *T)
    sumX3  +=       0.227 * cos(2.27961613476  + 25132.54721744910 *T)
    sumX3  +=       0.201 * cos(0.81376672334  + 6836.88907031730 *T)
    sumX3  +=       0.201 * cos(5.80642900510  + 5729.75026463250 *T)
    sumX3  +=       0.036 * cos(1.33765000112  + 12569.91863581531 *T)
    sumX3  +=       0.034 * cos(1.93992455025  + 4705.97612502710 *T)
    sumX3  +=       0.033 * cos(4.22745351005  + 12410.97511803210 *T)
    sumX3  +=       0.032 * cos(5.38348719867  + 6257.02134767510 *T)
    sumX3  +=       0.032 * cos(4.41212819179  + 6309.61798727471 *T)
    sumX3  +=       0.026 * cos(0.50334838608  + 775.76642880750 *T)
    sumX3  +=       0.023 * cos(2.01776542272  + 1059.62574767270 *T)
    sumX3  +=       0.025 * cos(1.58405990040  + 7860.66320992270 *T)
    sumX3  +=       0.027 * cos(5.17969059866  + 12565.90788249930 *T)
    sumX3  +=       0.020 * cos(3.76292921295  + 5753.62870238030 *T)
    sumX3  +=       0.019 * cos(3.01515389134  + 5885.17066406670 *T)
    sumX3  +=       0.019 * cos(6.05056350877  + 6813.01063256950 *T)
    sumX3  +=       0.018 * cos(2.29290644779  + 12721.81591690050 *T)
    sumX3  +=       0.017 * cos(0.49235818139  + 6681.46867088311 *T)
    sumX3  +=       0.013 * cos(4.49557806784  + 5487.02166065850 *T)
    sumX3  +=       0.011 * cos(5.29141894629  + 7079.61767429131 *T)
    sumX3  +=       0.010 * cos(4.46246654504  + 5507.79705615090 *T)
    sumX3  +=       0.010 * cos(3.03842434664  + 11790.87290614230 *T)
    sumX3  +=       0.009 * cos(1.50545167256  + 11507.01358727710 *T)
    sumX3  +=       0.011 * cos(6.25880735981  + 12592.69383726610 *T)
    sumX3  +=       0.009 * cos(5.39258795934  + 7058.84227879890 *T)
    sumX3  +=       0.008 * cos(1.26531983919  + 6290.43321447570 *T)
    sumX3  +=       0.008 * cos(2.25393064493  + 6276.20612047410 *T)
    sumX3  +=       0.008 * cos(3.03070423945  + 796.54182429990 *T)
    sumX3  +=       0.007 * cos(2.58386169059  + 4693.75913722409 *T)
    sumX3  +=       0.007 * cos(1.92157668691  + 7.35736448430 *T)
    sumX3  +=       0.005 * cos(4.86563371701  + 3739.00524759150 *T)
    sumX3  +=       0.005 * cos(6.21240311845  + 6070.02057203690 *T)
    sumX3  +=       0.006 * cos(3.95562930301  + 6284.29998854310 *T)
    sumX3  +=       0.006 * cos(5.83931193332  + 6282.33934640670 *T)
    sumX3  +=       0.005 * cos(3.46747665337  + 4137.15425099970 *T)
    sumX3  +=       0.005 * cos(3.58293336989  + 6496.61876291290 *T)

    # EARTH X Coord
    sumX4  = 0
    sumX4  +=     114.918 * cos(0.06053023506  + 6283.31966747490 *T)
    sumX4  +=       6.817 * cos(4.47624663983  + 0.24381748350 *T)
    sumX4  +=       3.158 * cos(0.47910545815  + 12566.39551746630 *T)
    sumX4  +=       0.211 * cos(3.72351646714  + 6438.74006690910 *T)
    sumX4  +=       0.210 * cos(2.88489921647  + 6127.89926804070 *T)
    sumX4  +=       0.253 * cos(0.64665875683  + 18849.47136745770 *T)
    sumX4  +=       0.073 * cos(3.92376796583  + 6282.83203250789 *T)
    sumX4  +=       0.042 * cos(4.60322795936  + 710.17686604180 *T)
    sumX4  +=       0.023 * cos(6.04513529243  + 6279.79654912590 *T)
    sumX4  +=       0.022 * cos(6.04140763882  + 6286.84278582391 *T)
    sumX4  +=       0.012 * cos(0.51640078868  + 25132.54721744910 *T)
    sumX4  +=       0.009 * cos(4.21008055560  + 83997.09113559539 *T)
    sumX4  +=       0.009 * cos(5.19392488383  + 11856.46246890800 *T)
    sumX4  +=       0.008 * cos(6.16587249238  + 167284.00540514899 *T)
    sumX4  +=       0.005 * cos(5.97913996910  + 6257.02134767510 *T)
    sumX4  +=       0.005 * cos(3.78878463461  + 6309.61798727471 *T)
    sumX4  +=       0.005 * cos(2.68915538708  + 12410.97511803210 *T)
    sumX4  +=       0.005 * cos(5.56850457487  + 156137.71980228281 *T)
    sumX4  +=       0.004 * cos(3.40118346072  + 12565.90788249930 *T)
    sumX4  +=       0.003 * cos(5.17221494066  + 529.93478257810 *T)

    # EARTH X Coord
    sumX5  = 0
    sumX5  +=       0.877 * cos(0.16136296068  + 6283.31966747490 *T)
    sumX5  +=       0.305 * cos(5.62973274927  + 0.24381748350 *T)
    sumX5  +=       0.101 * cos(4.88076699149  + 12566.39551746630 *T)
    sumX5  +=       0.025 * cos(5.30593958386  + 6438.74006690910 *T)
    sumX5  +=       0.025 * cos(1.27879113841  + 6127.89926804070 *T)
    sumX5  +=       0.010 * cos(5.11399286577  + 18849.47136745770 *T)
    sumX5  +=       0.003 * cos(2.56205735968  + 6282.83203250789 *T)

    # EARTH Y Coord
    sumY0  = 0
    sumY0  +=  99986069.925 * cos(0.18267413078  + 6283.31966747490 *T)
    sumY0  +=  2506324.281 * cos(3.36739796418  + 0.24381748350 *T)
    sumY0  +=  835274.807 * cos(0.13953892859  + 12566.39551746630 *T)
    sumY0  +=   10466.796 * cos(0.09642351540  + 18849.47136745770 *T)
    sumY0  +=    3490.608 * cos(6.01453435910  + 6282.83203250789 *T)
    sumY0  +=    3110.838 * cos(5.38114087369  + 83997.09113559539 *T)
    sumY0  +=    2561.408 * cos(5.29827505528  + 529.93478257810 *T)
    sumY0  +=    2142.365 * cos(2.66284107563  + 1577.09972496430 *T)
    sumY0  +=    1709.161 * cos(5.20779761275  + 6279.79654912590 *T)
    sumY0  +=    1707.934 * cos(4.58234386738  + 6286.84278582391 *T)
    sumY0  +=    1442.753 * cos(1.90130765657  + 2353.10997125530 *T)
    sumY0  +=    1113.027 * cos(5.26701283158  + 5223.45010231870 *T)
    sumY0  +=     934.484 * cos(4.50305745607  + 12036.70455237170 *T)
    sumY0  +=     899.854 * cos(1.60527831001  + 10213.52936369450 *T)
    sumY0  +=     566.820 * cos(0.58162314212  + 1059.62574767270 *T)
    sumY0  +=     661.858 * cos(6.02414120157  + 5753.62870238030 *T)
    sumY0  +=     739.644 * cos(2.79582891432  + 398.39282089170 *T)
    sumY0  +=     681.381 * cos(0.64735765260  + 4705.97612502710 *T)
    sumY0  +=     611.293 * cos(3.81390547656  + 6813.01063256950 *T)
    sumY0  +=     451.129 * cos(4.52236258525  + 5885.17066406670 *T)
    sumY0  +=     451.985 * cos(5.99169934357  + 6257.02134767510 *T)
    sumY0  +=     449.743 * cos(3.79862296384  + 6309.61798727471 *T)
    sumY0  +=     406.291 * cos(5.25608267423  + 6681.46867088311 *T)
    sumY0  +=     541.115 * cos(5.49909532338  + 775.76642880750 *T)
    sumY0  +=     546.682 * cos(6.17348361999  + 14143.73905991410 *T)
    sumY0  +=     513.780 * cos(2.86615405111  + 7860.66320992270 *T)
    sumY0  +=     263.916 * cos(3.82472119950  + 5507.79705615090 *T)
    sumY0  +=     217.220 * cos(2.94186064336  + 11507.01358727710 *T)
    sumY0  +=     292.146 * cos(4.08451989980  + 5507.30942118390 *T)
    sumY0  +=     227.915 * cos(5.95179814771  + 7058.84227879890 *T)
    sumY0  +=     226.835 * cos(4.84435977688  + 4693.75913722409 *T)
    sumY0  +=     255.833 * cos(0.69465437518  + 12168.24651405810 *T)
    sumY0  +=     256.157 * cos(6.16717762843  + 710.17686604180 *T)
    sumY0  +=     178.679 * cos(1.39565997429  + 796.54182429990 *T)
    sumY0  +=     178.303 * cos(5.11711459804  + 6283.25235717211 *T)
    sumY0  +=     178.303 * cos(4.67301162820  + 6283.38697777770 *T)
    sumY0  +=     155.471 * cos(0.05335100564  + 25132.54721744910 *T)
    sumY0  +=     155.246 * cos(6.13208007594  + 5487.02166065850 *T)
    sumY0  +=     207.640 * cos(4.28246516283  + 11790.87290614230 *T)
    sumY0  +=     199.988 * cos(2.50137197430  + 17790.08943726851 *T)
    sumY0  +=     128.970 * cos(3.64618696457  + 7079.61767429131 *T)
    sumY0  +=     128.155 * cos(3.23179068571  + 3739.00524759150 *T)
    sumY0  +=     152.241 * cos(5.58140597212  + 213.54291292150 *T)
    sumY0  +=      78.109 * cos(0.27810887092  + 8827.63408735830 *T)
    sumY0  +=      73.503 * cos(1.12438065243  + 1589.31671276730 *T)
    sumY0  +=      93.001 * cos(5.50857519732  + 9437.51911740350 *T)
    sumY0  +=      65.526 * cos(2.08488700481  + 11770.09751064990 *T)
    sumY0  +=      63.240 * cos(0.67044494917  + 6262.54427198250 *T)
    sumY0  +=      62.961 * cos(2.83649752628  + 6304.09506296730 *T)
    sumY0  +=      85.870 * cos(1.45351146305  + 167284.00540514899 *T)
    sumY0  +=      83.233 * cos(4.62701145755  + 3340.85624418330 *T)
    sumY0  +=      62.850 * cos(0.46743502764  + 7.35736448430 *T)
    sumY0  +=      79.971 * cos(0.92256149872  + 6070.02057203690 *T)
    sumY0  +=      60.043 * cos(1.82875530765  + 4137.15425099970 *T)
    sumY0  +=      75.613 * cos(2.58868557323  + 6496.61876291290 *T)
    sumY0  +=      53.731 * cos(6.27507790078  + 1194.69082770810 *T)
    sumY0  +=      57.574 * cos(2.39891957636  + 6284.29998854310 *T)
    sumY0  +=      57.574 * cos(1.10802134270  + 6282.33934640670 *T)
    sumY0  +=      72.385 * cos(0.13875968847  + 10976.83498721549 *T)
    sumY0  +=      51.135 * cos(6.02476231374  + 6290.43321447570 *T)
    sumY0  +=      50.604 * cos(3.77440441934  + 6276.20612047410 *T)
    sumY0  +=      49.613 * cos(1.07290414466  + 6127.89926804070 *T)
    sumY0  +=      48.009 * cos(2.43492252840  + 6438.74006690910 *T)
    sumY0  +=      59.233 * cos(5.35402385929  + 3154.44326741210 *T)
    sumY0  +=      48.264 * cos(0.05887245943  + 801.57711364030 *T)
    sumY0  +=      47.368 * cos(4.53616745806  + 3128.63258257930 *T)
    sumY0  +=      39.225 * cos(1.68202291816  + 8429.48508395010 *T)
    sumY0  +=      51.307 * cos(4.21369114598  + 12353.09642202830 *T)
    sumY0  +=      50.639 * cos(0.48135592944  + 11856.46246890800 *T)
    sumY0  +=      39.555 * cos(3.35943021943  + 5481.49873635110 *T)
    sumY0  +=      42.757 * cos(2.06983819322  + 17260.39847217390 *T)
    sumY0  +=      48.941 * cos(5.16594049812  + 9225.78309076650 *T)
    sumY0  +=      48.302 * cos(2.17445966751  + 2544.07060239989 *T)
    sumY0  +=      37.971 * cos(1.47013374285  + 426.84200835950 *T)
    sumY0  +=      43.643 * cos(4.71767697827  + 5856.23384163189 *T)
    sumY0  +=      31.830 * cos(3.57964760085  + 2145.92159899169 *T)
    sumY0  +=      33.959 * cos(0.15150934422  + 7085.14059859870 *T)
    sumY0  +=      30.184 * cos(2.39802729198  + 10977.32262218251 *T)
    sumY0  +=      32.562 * cos(4.41726666203  + 4164.06817212950 *T)
    sumY0  +=      38.887 * cos(0.64783670316  + 13367.72881362310 *T)
    sumY0  +=      28.515 * cos(4.53853388778  + 12569.91863581531 *T)
    sumY0  +=      28.873 * cos(5.72467865653  + 3930.45351370310 *T)
    sumY0  +=      36.633 * cos(1.33719170745  + 18073.94875613370 *T)
    sumY0  +=      24.984 * cos(3.74998709416  + 26.05450231630 *T)
    sumY0  +=      26.364 * cos(5.25860499861  + 9438.00675237050 *T)
    sumY0  +=      30.326 * cos(0.42292659979  + 4535.30325440790 *T)
    sumY0  +=      28.166 * cos(5.19366223100  + 12562.87239911730 *T)
    sumY0  +=      28.001 * cos(1.25523598395  + 5088.87265725030 *T)
    sumY0  +=      21.157 * cos(3.50176956898  + 11371.94850724170 *T)
    sumY0  +=      23.113 * cos(5.57235828208  + 5223.93773728571 *T)
    sumY0  +=      21.642 * cos(3.06444525139  + 12559.28197046550 *T)
    sumY0  +=      19.448 * cos(5.97216607301  + 12565.90788249930 *T)
    sumY0  +=      26.055 * cos(1.87419996146  + 71430.45180064559 *T)
    sumY0  +=      18.232 * cos(4.98441727809  + 1747.77259558350 *T)
    sumY0  +=      23.248 * cos(4.61930360205  + 18319.78040236310 *T)
    sumY0  +=      22.785 * cos(2.73901481639  + 5753.14106741329 *T)
    sumY0  +=      17.091 * cos(5.29823070971  + 4933.45225781610 *T)
    sumY0  +=      19.399 * cos(6.00268090156  + 10447.14402212089 *T)
    sumY0  +=      21.067 * cos(2.22953967314  + 7477.76667769950 *T)
    sumY0  +=      21.310 * cos(3.08840298987  + 8031.33608054190 *T)
    sumY0  +=      14.189 * cos(6.02227444325  + 2942.70724077510 *T)
    sumY0  +=      13.864 * cos(3.02270988640  + 8636.18582124671 *T)
    sumY0  +=      18.884 * cos(0.85614785764  + 156137.71980228281 *T)
    sumY0  +=      17.875 * cos(4.86899529635  + 1592.83983111630 *T)
    sumY0  +=      18.602 * cos(2.07103958699  + 17297.93850984270 *T)
    sumY0  +=      13.605 * cos(3.62641283985  + 13096.08648256090 *T)
    sumY0  +=      15.064 * cos(1.55068405670  + 16496.60521368590 *T)
    sumY0  +=      12.969 * cos(4.49412072551  + 7633.18707713370 *T)
    sumY0  +=      14.403 * cos(6.14964820819  + 20426.81490990550 *T)
    sumY0  +=      13.311 * cos(2.03652540541  + 12139.79732659030 *T)
    sumY0  +=      13.645 * cos(3.57902352159  + 13368.21644859010 *T)
    sumY0  +=      11.583 * cos(3.88983863974  + 5331.60126122430 *T)
    sumY0  +=      11.405 * cos(5.97095658034  + 529.44714761109 *T)
    sumY0  +=      15.057 * cos(0.50195173390  + 7084.65296363170 *T)
    sumY0  +=      14.330 * cos(4.45695507640  + 7342.70159766410 *T)
    sumY0  +=      12.375 * cos(1.53862290035  + 6279.72923882311 *T)
    sumY0  +=      12.366 * cos(1.96831814313  + 6286.91009612670 *T)
    sumY0  +=      10.016 * cos(3.81622500644  + 397.90518592470 *T)
    sumY0  +=      12.435 * cos(5.77865382585  + 15110.70993734970 *T)
    sumY0  +=       9.513 * cos(5.90294963286  + 7235.03807372550 *T)
    sumY0  +=      10.997 * cos(0.62118771904  + 10989.05197501850 *T)
    sumY0  +=      11.583 * cos(2.48171191864  + 5729.75026463250 *T)
    sumY0  +=       9.034 * cos(3.85796197835  + 9623.93209417470 *T)
    sumY0  +=      11.842 * cos(2.59877813496  + 15721.08260236190 *T)
    sumY0  +=       8.869 * cos(3.47611287212  + 6148.25458743950 *T)
    sumY0  +=       8.614 * cos(0.03081566771  + 6418.38474751031 *T)
    sumY0  +=      10.312 * cos(1.02710009735  + 6836.88907031730 *T)
    sumY0  +=       8.411 * cos(0.87289239606  + 14712.56093394150 *T)
    sumY0  +=       7.911 * cos(1.58349090213  + 2119.00767786191 *T)
    sumY0  +=      10.379 * cos(0.10632075985  + 1349.62359217529 *T)
    sumY0  +=      10.374 * cos(3.97648848184  + 5486.53402569149 *T)
    sumY0  +=      10.480 * cos(1.38209289015  + 5999.46034860970 *T)
    sumY0  +=       7.404 * cos(3.84339417190  + 6040.59106350090 *T)
    sumY0  +=       7.357 * cos(5.40136984638  + 5088.38502228330 *T)
    sumY0  +=       9.820 * cos(2.12476608355  + 6567.17898634010 *T)
    sumY0  +=       7.018 * cos(5.94659139192  + 6526.04827144891 *T)
    sumY0  +=       9.668 * cos(3.49318502073  + 21228.14820606230 *T)
    sumY0  +=       7.545 * cos(5.94843445846  + 12540.09719766650 *T)
    sumY0  +=       7.202 * cos(4.31311905021  + 6245.29199483910 *T)
    sumY0  +=       7.144 * cos(5.47703554612  + 6321.34734011070 *T)
    sumY0  +=       8.883 * cos(5.46710407546  + 5642.93474619389 *T)
    sumY0  +=       6.626 * cos(5.41362465894  + 5327.71992586630 *T)
    sumY0  +=       7.015 * cos(5.70701993379  + 7860.17557495569 *T)
    sumY0  +=       6.159 * cos(5.65831198505  + 12964.54452087450 *T)
    sumY0  +=       7.413 * cos(3.46419092567  + 23543.47432216530 *T)
    sumY0  +=       6.355 * cos(3.46461466453  + 1990.98883452450 *T)
    sumY0  +=       7.286 * cos(2.01139981548  + 537.04832957890 *T)
    sumY0  +=       6.314 * cos(1.51246730786  + 951.47458876710 *T)
    sumY0  +=       6.257 * cos(2.66128485006  + 4690.72365384210 *T)
    sumY0  +=       6.484 * cos(1.63733658639  + 16730.70750707931 *T)
    sumY0  +=       5.803 * cos(1.38237252918  + 955.35592412510 *T)
    sumY0  +=       6.391 * cos(2.40521894071  + 24073.16528725990 *T)
    sumY0  +=       7.530 * cos(4.91537693059  + 10973.79950383350 *T)
    sumY0  +=       6.921 * cos(5.41208599854  + 522.82123557730 *T)
    sumY0  +=       7.626 * cos(2.72257180334  + 22004.15845235330 *T)
    sumY0  +=       7.531 * cos(1.94384811400  + 18422.87317658170 *T)
    sumY0  +=       6.902 * cos(4.32614260011  + 155.17658195069 *T)
    sumY0  +=       5.434 * cos(4.37546348368  + 7238.91940908350 *T)
    sumY0  +=       6.223 * cos(0.67316616422  + 18451.32236404950 *T)
    sumY0  +=       7.296 * cos(1.27743142232  + 16730.21987211229 *T)
    sumY0  +=       5.833 * cos(0.75851095819  + 640.14110379750 *T)
    sumY0  +=       7.171 * cos(1.43520696239  + 3929.96587873609 *T)
    sumY0  +=       6.251 * cos(0.62294633745  + 6277.79674316750 *T)
    sumY0  +=       6.251 * cos(2.88399458161  + 6288.84259178230 *T)
    sumY0  +=       5.683 * cos(1.92473239272  + 1551.28904013150 *T)
    sumY0  +=       6.383 * cos(2.21747129905  + 5216.33655531789 *T)
    sumY0  +=       6.194 * cos(5.13602240936  + 5230.56364931950 *T)
    sumY0  +=       5.036 * cos(2.26614918626  + 14314.41193053330 *T)
    sumY0  +=       6.774 * cos(3.56437941703  + 102.84895673509 *T)
    sumY0  +=       6.224 * cos(1.92311046214  + 11014.86265985130 *T)
    sumY0  +=       4.621 * cos(2.91747769543  + 553.32558535889 *T)
    sumY0  +=       6.293 * cos(0.74955598933  + 5650.53592816170 *T)
    sumY0  +=       6.297 * cos(2.83090961174  + 26088.14695905770 *T)
    sumY0  +=       5.186 * cos(5.42426212805  + 77714.01528560399 *T)
    sumY0  +=       6.196 * cos(4.15785490016  + 84672.23202702119 *T)
    sumY0  +=       6.160 * cos(3.21170376619  + 239424.63407183639 *T)
    sumY0  +=       5.414 * cos(1.42882325200  + 6180.22689325630 *T)
    sumY0  +=       5.413 * cos(2.08335586272  + 6386.41244169350 *T)
    sumY0  +=       4.863 * cos(2.19642727179  + 90280.16698558679 *T)
    sumY0  +=       5.711 * cos(2.75632617612  + 6916.10340678810 *T)
    sumY0  +=       5.150 * cos(2.53503851142  + 1577.58735993130 *T)
    sumY0  +=       4.269 * cos(0.81279132183  + 7875.91568110771 *T)
    sumY0  +=       4.017 * cos(2.21501270421  + 6254.87048000710 *T)
    sumY0  +=       4.069 * cos(3.99136293444  + 3634.37720703489 *T)
    sumY0  +=       3.993 * cos(1.29192496097  + 6311.76885494270 *T)
    sumY0  +=       4.016 * cos(0.52802181166  + 4690.23601887509 *T)
    sumY0  +=       5.290 * cos(4.91562568689  + 25158.35790228190 *T)
    sumY0  +=       4.831 * cos(1.26836887601  + 5760.74224938110 *T)
    sumY0  +=       3.487 * cos(1.93988497026  + 10447.63165708790 *T)
    sumY0  +=       4.120 * cos(1.72576404553  + 6709.91785835091 *T)
    sumY0  +=       4.410 * cos(2.23510306717  + 6805.89708556871 *T)
    sumY0  +=       4.023 * cos(1.51236080833  + 4731.78680985990 *T)
    sumY0  +=       3.980 * cos(1.74852343480  + 5856.72147659890 *T)
    sumY0  +=       3.173 * cos(3.62048518613  + 1066.73929467350 *T)
    sumY0  +=       3.293 * cos(5.58717113537  + 9917.45305702629 *T)
    sumY0  +=       2.935 * cos(2.10511347366  + 12592.69383726610 *T)
    sumY0  +=       2.979 * cos(4.62987940723  + 12566.46282776910 *T)
    sumY0  +=       3.980 * cos(5.90920652207  + 13341.91812879030 *T)
    sumY0  +=       3.892 * cos(3.95491793455  + 75.02541605080 *T)
    sumY0  +=       3.042 * cos(1.62346061120  + 18053.17336064130 *T)
    sumY0  +=       3.608 * cos(4.03873145467  + 6208.53806890760 *T)
    sumY0  +=       3.721 * cos(3.65024914006  + 5966.92779781830 *T)
    sumY0  +=       3.581 * cos(5.75234129258  + 6358.10126604220 *T)
    sumY0  +=       2.880 * cos(0.04145923446  + 10575.65050042530 *T)
    sumY0  +=       2.806 * cos(2.64471981541  + 5863.83502359970 *T)
    sumY0  +=       3.626 * cos(6.13871061291  + 6599.71153713150 *T)
    sumY0  +=       2.794 * cos(0.98030293667  + 8030.84844557490 *T)
    sumY0  +=       2.729 * cos(2.60239924444  + 3.76693583251 *T)
    sumY0  +=       3.625 * cos(1.43112681156  + 6020.23574410210 *T)
    sumY0  +=       3.206 * cos(3.72476724430  + 11506.52595231009 *T)
    sumY0  +=       3.556 * cos(5.86027209236  + 8428.99744898310 *T)
    sumY0  +=       2.856 * cos(1.76545510884  + 12721.81591690050 *T)
    sumY0  +=       2.611 * cos(0.86182235214  + 6702.80431135010 *T)
    sumY0  +=       2.535 * cos(0.01231069705  + 31415.62306744050 *T)
    sumY0  +=       3.520 * cos(3.80906738831  + 250570.91967470257 *T)
    sumY0  +=       3.434 * cos(2.07577291101  + 6546.40359084770 *T)
    sumY0  +=       2.572 * cos(0.08592522543  + 5884.68302909969 *T)
    sumY0  +=       2.647 * cos(5.39051905541  + 2352.62233628830 *T)
    sumY0  +=       3.026 * cos(3.66998442300  + 13916.26292712510 *T)
    sumY0  +=       2.378 * cos(2.05754080989  + 2389.13783793270 *T)
    sumY0  +=       2.979 * cos(5.07398237707  + 12566.32820716351 *T)
    sumY0  +=       3.082 * cos(3.34593854945  + 14945.07235607090 *T)
    sumY0  +=       2.716 * cos(1.49855756532  + 12029.59100537090 *T)
    sumY0  +=       2.215 * cos(3.69564106529  + 13362.69352428270 *T)
    sumY0  +=       3.032 * cos(0.05314931388  + 29088.56759850150 *T)
    sumY0  +=       2.421 * cos(1.58246988642  + 11015.35029481830 *T)
    sumY0  +=       2.275 * cos(4.38385659579  + 6077.13411903770 *T)
    sumY0  +=       2.466 * cos(0.87928045349  + 16200.52890701769 *T)
    sumY0  +=       2.529 * cos(4.37629498907  + 12043.81809937250 *T)
    sumY0  +=       2.015 * cos(0.91352607152  + 6262.96434807611 *T)
    sumY0  +=       2.218 * cos(5.40562909521  + 6489.50521591210 *T)
    sumY0  +=       2.005 * cos(2.59344260095  + 6303.67498687370 *T)
    sumY0  +=       2.010 * cos(5.68334625182  + 4700.87268422489 *T)
    sumY0  +=       2.641 * cos(5.85354681494  + 4590.66636300550 *T)
    sumY0  +=       2.438 * cos(4.55830572881  + 149.80701461810 *T)
    sumY0  +=       1.893 * cos(5.69296362938  + 6279.43833211690 *T)
    sumY0  +=       1.893 * cos(4.09716259686  + 6287.20100283290 *T)
    sumY0  +=       2.156 * cos(0.43822603611  + 18139.53831889940 *T)
    sumY0  +=       2.071 * cos(2.75099511730  + 11926.49823115230 *T)
    sumY0  +=       2.395 * cos(4.84387255380  + 1162.71852189130 *T)
    sumY0  +=       1.954 * cos(5.08514679694  + 13518.11392371690 *T)
    sumY0  +=       2.006 * cos(5.02287213343  + 3154.93090237910 *T)
    sumY0  +=       1.776 * cos(4.86094102925  + 13521.99525907490 *T)
    sumY0  +=       1.964 * cos(1.82182786120  + 4686.64559022330 *T)
    sumY0  +=       2.332 * cos(2.83569470692  + 10022.08109758290 *T)
    sumY0  +=       1.863 * cos(3.68950976920  + 242.97242145751 *T)
    sumY0  +=       1.928 * cos(3.02170818288  + 5746.51515537950 *T)
    sumY0  +=       2.436 * cos(2.83691190801  + 95143.37673846159 *T)
    sumY0  +=       2.163 * cos(6.13036805408  + 6037.48802124550 *T)
    sumY0  +=       2.250 * cos(0.82649289984  + 12669.48829168490 *T)
    sumY0  +=       2.052 * cos(3.65926976048  + 6529.15131370430 *T)
    sumY0  +=       1.886 * cos(5.18293217320  + 7238.43177411650 *T)
    sumY0  +=       2.170 * cos(3.92516942708  + 19651.29229858150 *T)
    sumY0  +=       2.058 * cos(3.12741900385  + 23013.78335707070 *T)
    sumY0  +=       1.774 * cos(0.51402656711  + 6820.12417957030 *T)
    sumY0  +=       1.841 * cos(1.91732926338  + 65147.37595065419 *T)
    sumY0  +=       1.578 * cos(4.07231752329  + 4292.57465043390 *T)
    sumY0  +=       1.577 * cos(2.37859191870  + 7632.69944216670 *T)
    sumY0  +=       1.541 * cos(0.23280184119  + 13119.96492030870 *T)
    sumY0  +=       2.012 * cos(1.94074945567  + 4292.08701546690 *T)
    sumY0  +=       1.543 * cos(5.39993967465  + 6293.95633282471 *T)
    sumY0  +=       1.543 * cos(4.39018655159  + 6272.68300212510 *T)
    sumY0  +=       1.513 * cos(2.22141523192  + 20.53157800890 *T)
    sumY0  +=       1.489 * cos(1.38125902411  + 110.45013870291 *T)
    sumY0  +=       1.639 * cos(2.86996469977  + 17655.02435723310 *T)
    sumY0  +=       1.443 * cos(0.69249676813  + 1052.51220067191 *T)
    sumY0  +=       1.337 * cos(1.49321921272  + 2544.55823736690 *T)
    sumY0  +=       1.797 * cos(1.47415887460  + 33018.77729472110 *T)
    sumY0  +=       1.488 * cos(4.10370894807  + 76.50988875911 *T)
    sumY0  +=       1.453 * cos(3.88999021921  + 633.02755679670 *T)
    sumY0  +=       1.349 * cos(4.18018116919  + 6016.71262575310 *T)
    sumY0  +=       1.477 * cos(2.56925141669  + 12779.69461290430 *T)
    sumY0  +=       1.688 * cos(4.77167367345  + 18875.28205229050 *T)
    sumY0  +=       1.491 * cos(4.65894552032  + 9411.70843257070 *T)
    sumY0  +=       1.471 * cos(4.18699607739  + 18636.17227201970 *T)
    sumY0  +=       1.263 * cos(5.62967869388  + 12573.50906446710 *T)
    sumY0  +=       1.685 * cos(4.13797957160  + 25934.36814857290 *T)
    sumY0  +=       1.375 * cos(2.66006793362  + 2378.92065608810 *T)
    sumY0  +=       1.641 * cos(1.56425462469  + 6.86972951729 *T)
    sumY0  +=       1.200 * cos(6.18795172265  + 11499.90004027630 *T)
    sumY0  +=       1.272 * cos(5.60977811105  + 6549.92670919670 *T)
    sumY0  +=       1.214 * cos(2.79740025979  + 10973.31186886650 *T)
    sumY0  +=       1.406 * cos(0.96357897619  + 11790.38527117530 *T)
    sumY0  +=       1.524 * cos(2.69987154893  + 18209.57408114370 *T)
    sumY0  +=       1.149 * cos(1.44882073480  + 10177.50149701710 *T)
    sumY0  +=       1.547 * cos(5.59995461433  + 11926.01059618529 *T)
    sumY0  +=       1.460 * cos(3.40155505101  + 246.07546371290 *T)
    sumY0  +=       1.323 * cos(2.95651083102  + 8661.99650607950 *T)
    sumY0  +=       1.252 * cos(6.12405778157  + 6993.25271603320 *T)
    sumY0  +=       1.459 * cos(3.72728247655  + 38.37685312130 *T)
    sumY0  +=       1.235 * cos(1.27016887136  + 24357.02460612510 *T)
    sumY0  +=       1.028 * cos(0.93950364612  + 6112.64679685570 *T)
    sumY0  +=       1.330 * cos(5.71304625366  + 2146.40923395870 *T)
    sumY0  +=       1.312 * cos(5.81330482072  + 12491.61391889900 *T)
    sumY0  +=       1.246 * cos(6.23510174757  + 5429.63565075589 *T)
    sumY0  +=       0.992 * cos(2.56738111789  + 6453.99253809410 *T)
    sumY0  +=       0.988 * cos(5.69208485871  + 8274.06468451590 *T)
    sumY0  +=       1.365 * cos(1.39907964627  + 11371.46087227470 *T)
    sumY0  +=       1.020 * cos(1.25240185088  + 4732.27444482690 *T)
    sumY0  +=       1.191 * cos(1.85928407242  + 6290.36590417291 *T)
    sumY0  +=       1.191 * cos(1.64765684663  + 6276.27343077690 *T)
    sumY0  +=       0.978 * cos(3.16773125497  + 6247.29180079750 *T)
    sumY0  +=       0.971 * cos(0.33921091467  + 6319.34753415231 *T)
    sumY0  +=       0.962 * cos(1.06488912173  + 12565.41519639810 *T)
    sumY0  +=       1.025 * cos(0.64688768472  + 12545.62012197390 *T)
    sumY0  +=       0.990 * cos(5.10278829353  + 4694.24677219110 *T)
    sumY0  +=       0.964 * cos(3.36272372610  + 3893.93801205869 *T)
    sumY0  +=       0.944 * cos(4.29299913988  + 6259.44122972711 *T)
    sumY0  +=       0.944 * cos(5.49712708636  + 6307.19810522270 *T)
    sumY0  +=       1.092 * cos(2.34985569802  + 17797.20298426930 *T)
    sumY0  +=       1.132 * cos(5.71867421008  + 17782.97589026770 *T)
    sumY0  +=       0.859 * cos(6.19348487479  + 1692.40948698591 *T)
    sumY0  +=       1.213 * cos(5.61315622766  + 82576.73740351178 *T)
    sumY0  +=       0.942 * cos(3.65711104430  + 6298.57213865991 *T)
    sumY0  +=       0.942 * cos(6.13301518193  + 6268.06719628990 *T)
    sumY0  +=       1.045 * cos(5.17434482333  + 15508.85894075790 *T)
    sumY0  +=       0.851 * cos(3.62598735241  + 6173.11334625550 *T)
    sumY0  +=       0.850 * cos(6.16443178844  + 6393.52598869430 *T)
    sumY0  +=       0.895 * cos(0.83009579073  + 3904.15519390330 *T)
    sumY0  +=       1.020 * cos(5.42778314413  + 220.65645992230 *T)
    sumY0  +=       0.845 * cos(4.18420138552  + 17256.87535382490 *T)
    sumY0  +=       0.863 * cos(2.86079142045  + 949.41942645330 *T)
    sumY0  +=       0.853 * cos(1.18534704592  + 16201.01654198470 *T)
    sumY0  +=       0.811 * cos(1.42220880644  + 4803.96545844350 *T)
    sumY0  +=       0.966 * cos(2.57517734766  + 206.42936592071 *T)
    sumY0  +=       0.917 * cos(0.89927696450  + 149854.64395229140 *T)
    sumY0  +=       1.083 * cos(2.89594427545  + 36948.98699094070 *T)
    sumY0  +=       0.779 * cos(2.01260050316  + 2648.69864295650 *T)
    sumY0  +=       0.964 * cos(2.34511297941  + 796.05418933290 *T)
    sumY0  +=       0.995 * cos(4.96662237412  + 11403.92081305850 *T)
    sumY0  +=       0.962 * cos(2.35578735539  + 12567.37583853451 *T)
    sumY0  +=       0.779 * cos(2.02763029837  + 10213.04172872750 *T)
    sumY0  +=       0.980 * cos(6.19670011124  + 22805.49174851010 *T)
    sumY0  +=       0.930 * cos(0.65237052253  + 2787.28684134090 *T)
    sumY0  +=       0.829 * cos(0.38404899643  + 5120.35732810009 *T)
    sumY0  +=       0.733 * cos(4.18029489609  + 10575.16286545830 *T)
    sumY0  +=       0.743 * cos(2.22826365578  + 7834.36489012290 *T)
    sumY0  +=       0.818 * cos(3.21447961169  + 5572.89898394960 *T)
    sumY0  +=       0.905 * cos(2.24459183660  + 6284.80414018320 *T)
    sumY0  +=       0.905 * cos(1.26234908246  + 6281.83519476660 *T)
    sumY0  +=       0.798 * cos(1.05421419226  + 12410.97511803210 *T)
    sumY0  +=       0.817 * cos(5.23322400864  + 12416.83232033170 *T)
    sumY0  +=       0.884 * cos(2.41399532496  + 22483.60475700909 *T)
    sumY0  +=       0.755 * cos(5.40020688961  + 4060.97539791089 *T)
    sumY0  +=       0.651 * cos(5.50763453832  + 17259.91083720689 *T)
    sumY0  +=       0.616 * cos(4.13394819140  + 1596.43025976811 *T)
    sumY0  +=       0.651 * cos(0.84412253043  + 1748.26023055050 *T)
    sumY0  +=       0.723 * cos(1.49663741427  + 161000.92955515761 *T)
    sumY0  +=       0.784 * cos(2.44837693578  + 4907.05823266209 *T)
    sumY0  +=       0.784 * cos(3.77248887935  + 7234.55043875850 *T)
    sumY0  +=       0.785 * cos(1.86574625348  + 846.32665223470 *T)
    sumY0  +=       0.611 * cos(6.24391260854  + 853.44019923550 *T)
    sumY0  +=       0.576 * cos(5.94037158869  + 12323.66691349230 *T)
    sumY0  +=       0.570 * cos(2.97586328906  + 12587.17091295870 *T)
    sumY0  +=       0.576 * cos(4.63987761174  + 13915.77529215809 *T)
    sumY0  +=       0.770 * cos(0.89072554788  + 6069.53293706989 *T)
    sumY0  +=       0.773 * cos(1.78442906238  + 6133.26883537330 *T)
    sumY0  +=       0.581 * cos(6.05264641509  + 11933.61177815310 *T)
    sumY0  +=       0.719 * cos(2.47920208822  + 15720.59496739490 *T)
    sumY0  +=       0.578 * cos(2.69917567595  + 8662.48414104651 *T)
    sumY0  +=       0.536 * cos(4.49580958272  + 18852.99448580670 *T)
    sumY0  +=       0.633 * cos(5.88497258324  + 52176.05010063190 *T)
    sumY0  +=       0.581 * cos(1.35719256610  + 5334.14405850510 *T)
    sumY0  +=       0.532 * cos(3.01682546062  + 18842.35782045690 *T)
    sumY0  +=       0.657 * cos(4.82951502701  + 5849.12029463110 *T)
    sumY0  +=       0.599 * cos(0.54713597581  + 6151.77770578850 *T)
    sumY0  +=       0.540 * cos(1.67333017022  + 6286.60602489270 *T)
    sumY0  +=       0.540 * cos(1.83361074884  + 6280.03331005710 *T)
    sumY0  +=       0.628 * cos(5.02906961003  + 17298.42614480970 *T)
    sumY0  +=       0.532 * cos(2.84025435717  + 11514.12713427790 *T)
    sumY0  +=       0.513 * cos(2.13211972766  + 12456.18919624690 *T)
    sumY0  +=       0.649 * cos(3.33488996078  + 11764.57458634250 *T)
    sumY0  +=       0.582 * cos(2.96012797214  + 6414.86162916130 *T)
    sumY0  +=       0.517 * cos(5.63213604686  + 3340.36860921629 *T)
    sumY0  +=       0.609 * cos(2.37414673548  + 10420.23010099111 *T)
    sumY0  +=       0.615 * cos(0.00032066291  + 10983.94853421629 *T)
    sumY0  +=       0.622 * cos(0.82619343851  + 5326.54287653730 *T)
    sumY0  +=       0.536 * cos(2.13145334966  + 7232.49527644471 *T)
    sumY0  +=       0.516 * cos(1.35907342509  + 433.95555536030 *T)
    sumY0  +=       0.613 * cos(3.36175443914  + 10969.72144021469 *T)
    sumY0  +=       0.583 * cos(1.44072369133  + 5863.34738863270 *T)
    sumY0  +=       0.636 * cos(1.34803641207  + 26735.70144472970 *T)
    sumY0  +=       0.665 * cos(4.31762068350  + 40879.19668716030 *T)
    sumY0  +=       0.583 * cos(4.39404867767  + 12592.20620229910 *T)
    sumY0  +=       0.479 * cos(6.04787718072  + 5547.44315394310 *T)
    sumY0  +=       0.571 * cos(1.01593938825  + 6062.90702503610 *T)
    sumY0  +=       0.565 * cos(5.49171831457  + 4171.18171913030 *T)
    sumY0  +=       0.462 * cos(3.53710274271  + 3104.68624194030 *T)
    sumY0  +=       0.551 * cos(2.46794798020  + 6503.73230991370 *T)
    sumY0  +=       0.504 * cos(0.43308443375  + 15670.83794192309 *T)
    sumY0  +=       0.535 * cos(4.55197610339  + 173567.08125514039 *T)
    sumY0  +=       0.460 * cos(4.75481082956  + 3495.78900865049 *T)
    sumY0  +=       0.449 * cos(0.38594751092  + 4274.27449334889 *T)
    sumY0  +=       0.453 * cos(5.14645638415  + 9387.76209193169 *T)
    sumY0  +=       0.527 * cos(4.67747505318  + 24602.85625235450 *T)
    sumY0  +=       0.502 * cos(5.68646114706  + 12490.12944619070 *T)
    sumY0  +=       0.603 * cos(5.56725967880  + 322711.54834138998 *T)
    sumY0  +=       0.532 * cos(0.26579545469  + 5120.84496306710 *T)
    sumY0  +=       0.530 * cos(5.14730101928  + 18845.94824910870 *T)
    sumY0  +=       0.431 * cos(3.74066314880  + 7019.19618100671 *T)
    sumY0  +=       0.465 * cos(1.12060083942  + 8827.14645239129 *T)
    sumY0  +=       0.559 * cos(0.14481247330  + 1582.20316576650 *T)
    sumY0  +=       0.485 * cos(4.56530394374  + 29296.85920706210 *T)
    sumY0  +=       0.584 * cos(4.78377689692  + 72850.80553272919 *T)
    sumY0  +=       0.550 * cos(5.52640225498  + 213.05527795450 *T)
    sumY0  +=       0.481 * cos(0.29866487122  + 14.47091148511 *T)
    sumY0  +=       0.569 * cos(1.38161051216  + 97238.87136197100 *T)
    sumY0  +=       0.551 * cos(3.21618763564  + 14313.92429556630 *T)
    sumY0  +=       0.389 * cos(1.85967717576  + 6245.18663183710 *T)
    sumY0  +=       0.389 * cos(1.64726374330  + 6321.45270311270 *T)
    sumY0  +=       0.388 * cos(5.85202122263  + 6297.54676147650 *T)
    sumY0  +=       0.388 * cos(3.93810500361  + 6269.09257347330 *T)
    sumY0  +=       0.515 * cos(0.09995632139  + 12320.56387123691 *T)
    sumY0  +=       0.394 * cos(1.39198903052  + 4156.95462512869 *T)
    sumY0  +=       0.491 * cos(4.39709251394  + 1479.11039154791 *T)
    sumY0  +=       0.440 * cos(2.54396321518  + 5650.04829319470 *T)
    sumY0  +=       0.382 * cos(1.50580912176  + 9917.94069199330 *T)
    sumY0  +=       0.503 * cos(3.08687612414  + 17157.30569795530 *T)
    sumY0  +=       0.439 * cos(3.25483236471  + 233141.55822184496 *T)
    sumY0  +=       0.381 * cos(3.59564013891  + 14143.25142494710 *T)
    sumY0  +=       0.371 * cos(3.01987537874  + 5643.42238116090 *T)
    sumY0  +=       0.516 * cos(4.04344774561  + 135.30889751891 *T)
    sumY0  +=       0.417 * cos(2.17381926869  + 13760.84252769090 *T)
    sumY0  +=       0.483 * cos(2.84671280305  + 9779.35249360890 *T)
    sumY0  +=       0.381 * cos(2.79871486639  + 14919.26167123810 *T)
    sumY0  +=       0.378 * cos(2.94872717012  + 17267.51201917470 *T)
    sumY0  +=       0.393 * cos(6.27917510634  + 7872.39256275871 *T)
    sumY0  +=       0.376 * cos(6.02660735301  + 13517.62628874989 *T)
    sumY0  +=       0.365 * cos(0.53940975175  + 6923.21695378890 *T)
    sumY0  +=       0.371 * cos(4.19154046767  + 13625.77744765550 *T)
    sumY0  +=       0.355 * cos(3.67238322579  + 10874.22984796390 *T)
    sumY0  +=       0.499 * cos(1.15482974306  + 161710.86260371591 *T)
    sumY0  +=       0.379 * cos(5.52860495919  + 3185.43584474911 *T)
    sumY0  +=       0.471 * cos(0.07055813902  + 11712.71150074729 *T)
    sumY0  +=       0.417 * cos(1.75993959010  + 22779.68106367730 *T)
    sumY0  +=       0.485 * cos(6.06683096360  + 12528.26248182851 *T)
    sumY0  +=       0.358 * cos(1.92035724848  + 6295.04902031090 *T)
    sumY0  +=       0.358 * cos(1.58658367058  + 6271.59031463890 *T)
    sumY0  +=       0.365 * cos(5.19130080276  + 6836.40143535029 *T)
    sumY0  +=       0.351 * cos(0.64808198517  + 11617.21990849651 *T)
    sumY0  +=       0.351 * cos(4.53436303138  + 205.94173095370 *T)
    sumY0  +=       0.429 * cos(5.51029115203  + 3894.42564702570 *T)
    sumY0  +=       0.388 * cos(5.20318951030  + 956.53297345411 *T)
    sumY0  +=       0.464 * cos(5.40703775457  + 23581.50199480110 *T)
    sumY0  +=       0.396 * cos(0.78828729228  + 5231.05128428650 *T)
    sumY0  +=       0.421 * cos(3.25659811640  + 7445.79437188270 *T)
    sumY0  +=       0.366 * cos(5.33851384127  + 17253.28492517310 *T)
    sumY0  +=       0.360 * cos(5.53135776054  + 21393.78578734110 *T)
    sumY0  +=       0.336 * cos(6.12569565789  + 6279.38751421180 *T)
    sumY0  +=       0.336 * cos(3.66443056835  + 6287.25182073800 *T)
    sumY0  +=       0.326 * cos(0.22542241669  + 1059.13811270570 *T)
    sumY0  +=       0.395 * cos(5.25182737341  + 5642.44206009270 *T)
    sumY0  +=       0.450 * cos(1.80229899871  + 1385.31775746760 *T)
    sumY0  +=       0.363 * cos(2.72535712612  + 22484.09239197610 *T)
    sumY0  +=       0.399 * cos(5.51787738042  + 16858.72635041670 *T)
    sumY0  +=       0.316 * cos(2.45995539098  + 20995.63678393290 *T)
    sumY0  +=       0.430 * cos(3.96101117316  + 19650.80466361450 *T)
    sumY0  +=       0.363 * cos(2.71990145670  + 7335.58805066330 *T)
    sumY0  +=       0.317 * cos(3.09968934621  + 11769.60987568290 *T)
    sumY0  +=       0.395 * cos(3.82348351794  + 5905.94605955911 *T)
    sumY0  +=       0.324 * cos(5.27854205827  + 37.78385515230 *T)
    sumY0  +=       0.299 * cos(2.29226668471  + 641.12142486571 *T)
    sumY0  +=       0.304 * cos(4.87224846445  + 5750.10558403130 *T)
    sumY0  +=       0.298 * cos(2.25684047934  + 1350.11122714230 *T)
    sumY0  +=       0.414 * cos(5.74970236841  + 44809.40638337990 *T)
    sumY0  +=       0.339 * cos(5.86256617398  + 3.68833577960 *T)
    sumY0  +=       0.351 * cos(4.36315861783  + 12345.98287502750 *T)
    sumY0  +=       0.362 * cos(1.99294614713  + 21953.91379191449 *T)
    sumY0  +=       0.304 * cos(3.36735752668  + 29826.55017215670 *T)
    sumY0  +=       0.395 * cos(4.78960869869  + 4176.28515993250 *T)
    sumY0  +=       0.320 * cos(0.92821593891  + 10818.37910439930 *T)
    sumY0  +=       0.384 * cos(5.58951901360  + 10177.01386205009 *T)
    sumY0  +=       0.279 * cos(5.67648753585  + 10970.20907518170 *T)
    sumY0  +=       0.365 * cos(5.96664259426  + 6660.69327539070 *T)
    sumY0  +=       0.377 * cos(5.53753804163  + 29864.57784479250 *T)
    sumY0  +=       0.275 * cos(4.16427905835  + 20597.48778052470 *T)
    sumY0  +=       0.316 * cos(1.02608510804  + 316.63568714010 *T)
    sumY0  +=       0.339 * cos(4.53749399958  + 6924.19727485710 *T)
    sumY0  +=       0.386 * cos(0.71953202985  + 2636.96929012050 *T)
    sumY0  +=       0.273 * cos(5.80758206821  + 26709.89075989690 *T)
    sumY0  +=       0.314 * cos(6.18462887647  + 14945.55999103790 *T)
    sumY0  +=       0.281 * cos(0.15414528977  + 16858.23871544969 *T)
    sumY0  +=       0.287 * cos(4.87229910240  + 18073.46112116669 *T)
    sumY0  +=       0.265 * cos(3.53930677592  + 19379.16233255230 *T)
    sumY0  +=       0.325 * cos(0.92198404013  + 12360.20996902910 *T)
    sumY0  +=       0.362 * cos(2.75236634018  + 30665.91114094930 *T)
    sumY0  +=       0.269 * cos(4.91776126981  + 6816.53375091851 *T)
    sumY0  +=       0.344 * cos(1.94480091531  + 6147.69434246491 *T)
    sumY0  +=       0.255 * cos(3.37543976329  + 1376.01761732930 *T)
    sumY0  +=       0.334 * cos(1.56223491581  + 6418.94499248490 *T)
    sumY0  +=       0.316 * cos(3.97125698165  + 6055.84353468590 *T)
    sumY0  +=       0.283 * cos(2.66015602247  + 28287.23430234470 *T)
    sumY0  +=       0.305 * cos(5.83332121064  + 16522.41589851870 *T)
    sumY0  +=       0.270 * cos(1.62295499850  + 283.61550138170 *T)
    sumY0  +=       0.243 * cos(0.97600841060  + 6255.91811137810 *T)
    sumY0  +=       0.243 * cos(2.53093250846  + 6310.72122357170 *T)
    sumY0  +=       0.296 * cos(4.79138408835  + 6129.54085699010 *T)
    sumY0  +=       0.301 * cos(5.81886935980  + 6510.79580026390 *T)
    sumY0  +=       0.329 * cos(5.86922121940  + 377.61742539930 *T)
    sumY0  +=       0.267 * cos(1.84931872641  + 24705.94902657310 *T)
    sumY0  +=       0.285 * cos(0.93414937692  + 5469.76938351510 *T)
    sumY0  +=       0.286 * cos(4.99886262730  + 6437.09847795970 *T)
    sumY0  +=       0.311 * cos(1.63984256457  + 11720.31268271510 *T)
    sumY0  +=       0.319 * cos(3.13980275142  + 169379.50002865840 *T)
    sumY0  +=       0.241 * cos(4.56518963450  + 632.53992182970 *T)
    sumY0  +=       0.229 * cos(5.72761357577  + 1265.81129610991 *T)
    sumY0  +=       0.235 * cos(4.94058515466  + 4487.57358878689 *T)
    sumY0  +=       0.253 * cos(2.08435720561  + 4377.36726756750 *T)
    sumY0  +=       0.277 * cos(3.15798901317  + 419.24082639170 *T)
    sumY0  +=       0.234 * cos(3.55494952630  + 11713.19913571430 *T)
    sumY0  +=       0.216 * cos(0.30207043900  + 10454.25756912169 *T)
    sumY0  +=       0.241 * cos(3.87591164077  + 103.33659170210 *T)
    sumY0  +=       0.214 * cos(0.52402082918  + 2222.10045208050 *T)
    sumY0  +=       0.213 * cos(2.22358750694  + 30356.24113725130 *T)
    sumY0  +=       0.240 * cos(4.14485134121  + 6309.13035230770 *T)
    sumY0  +=       0.212 * cos(3.93716983669  + 262.84010588929 *T)
    sumY0  +=       0.259 * cos(1.64646341991  + 6283.56348495841 *T)
    sumY0  +=       0.259 * cos(1.86047749915  + 6283.07584999140 *T)
    sumY0  +=       0.222 * cos(2.99868754607  + 10440.03047512009 *T)
    sumY0  +=       0.201 * cos(0.39862703515  + 5746.02752041250 *T)
    sumY0  +=       0.274 * cos(5.42466894469  + 23581.01435983410 *T)
    sumY0  +=       0.243 * cos(2.57266886723  + 7096.86995143470 *T)
    sumY0  +=       0.234 * cos(3.66617634256  + 5573.38661891660 *T)
    sumY0  +=       0.203 * cos(1.57613060854  + 16460.08971204149 *T)
    sumY0  +=       0.246 * cos(4.26574746801  + 8672.21368792411 *T)
    sumY0  +=       0.216 * cos(3.35974562754  + 5437.23683272371 *T)
    sumY0  +=       0.210 * cos(4.25419338790  + 9381.20349020070 *T)
    sumY0  +=       0.199 * cos(5.64767313384  + 11216.52810780750 *T)
    sumY0  +=       0.270 * cos(5.11258425974  + 284.10313634870 *T)
    sumY0  +=       0.207 * cos(1.49549061720  + 12562.80508881451 *T)
    sumY0  +=       0.212 * cos(0.18391974040  + 7129.40250222610 *T)
    sumY0  +=       0.269 * cos(6.23907858108  + 70755.31090921978 *T)
    sumY0  +=       0.221 * cos(4.97264401025  + 77713.52765063698 *T)
    sumY0  +=       0.225 * cos(5.58708477402  + 5635.82119919310 *T)
    sumY0  +=       0.197 * cos(4.93201648946  + 14712.07329897449 *T)
    sumY0  +=       0.192 * cos(0.47501431186  + 17272.12782500990 *T)
    sumY0  +=       0.241 * cos(3.52343193321  + 15907.00794416610 *T)
    sumY0  +=       0.261 * cos(0.88974137342  + 48739.61607959950 *T)
    sumY0  +=       0.255 * cos(2.10230203862  + 6206.56596123230 *T)
    sumY0  +=       0.256 * cos(0.73069007306  + 224.58861318540 *T)
    sumY0  +=       0.182 * cos(5.98997615944  + 18848.98373249069 *T)
    sumY0  +=       0.250 * cos(2.67103598665  + 5934.39524702691 *T)
    sumY0  +=       0.196 * cos(0.64831634218  + 16460.57734700850 *T)
    sumY0  +=       0.187 * cos(0.04522363352  + 22003.67081738630 *T)
    sumY0  +=       0.229 * cos(1.73812120333  + 2942.21960580809 *T)
    sumY0  +=       0.175 * cos(3.84094267540  + 11614.67711121570 *T)
    sumY0  +=       0.187 * cos(0.76186633809  + 9778.86485864190 *T)
    sumY0  +=       0.175 * cos(6.15302938442  + 3744.58352825430 *T)
    sumY0  +=       0.241 * cos(5.00033837792  + 8390.35417501730 *T)
    sumY0  +=       0.236 * cos(0.45405670770  + 1.72829019180 *T)
    sumY0  +=       0.217 * cos(3.45380390356  + 17996.27498570570 *T)
    sumY0  +=       0.173 * cos(2.65478632822  + 6275.71848550709 *T)
    sumY0  +=       0.191 * cos(4.70676649838  + 394.86970254271 *T)
    sumY0  +=       0.231 * cos(4.17552456214  + 34596.12083716890 *T)
    sumY0  +=       0.169 * cos(0.40140025709  + 6438.25243194210 *T)
    sumY0  +=       0.235 * cos(5.04475734159  + 17256.38771885789 *T)
    sumY0  +=       0.188 * cos(4.02583282760  + 401.91593924071 *T)
    sumY0  +=       0.217 * cos(3.40363231104  + 10984.43616918330 *T)
    sumY0  +=       0.230 * cos(0.83612775625  + 6632.24408792290 *T)
    sumY0  +=       0.208 * cos(5.43298352980  + 11087.52894340190 *T)
    sumY0  +=       0.165 * cos(0.80102915744  + 743.23387801611 *T)
    sumY0  +=       0.177 * cos(0.01829077781  + 4796.85191144269 *T)
    sumY0  +=       0.208 * cos(6.16123229814  + 3097.64000524229 *T)
    sumY0  +=       0.167 * cos(2.40308314985  + 5714.49779344750 *T)
    sumY0  +=       0.198 * cos(4.83812277030  + 5539.84197197530 *T)
    sumY0  +=       0.169 * cos(2.16903380895  + 12132.68377958950 *T)
    sumY0  +=       0.214 * cos(2.77137176457  + 24492.64993113510 *T)
    sumY0  +=       0.172 * cos(5.97118942936  + 6233.56264200310 *T)
    sumY0  +=       0.172 * cos(3.81893679688  + 6333.07669294670 *T)
    sumY0  +=       0.223 * cos(3.42911214381  + 266.85085920531 *T)
    sumY0  +=       0.159 * cos(2.33166453392  + 13199.17925677950 *T)
    sumY0  +=       0.219 * cos(0.69566285170  + 10344.05124790229 *T)
    sumY0  +=       0.207 * cos(1.92518585993  + 12569.98594611810 *T)
    sumY0  +=       0.176 * cos(2.48119576594  + 12012.82611462390 *T)
    sumY0  +=       0.170 * cos(4.43959906760  + 6294.36536773881 *T)
    sumY0  +=       0.170 * cos(5.35052715864  + 6272.27396721100 *T)
    sumY0  +=       0.205 * cos(1.15927608909  + 13119.47728534169 *T)
    sumY0  +=       0.167 * cos(3.97305893961  + 17686.99666304990 *T)
    sumY0  +=       0.210 * cos(2.45255073470  + 13521.50762410789 *T)
    sumY0  +=       0.163 * cos(6.18035129909  + 802.06474860730 *T)
    sumY0  +=       0.159 * cos(0.13779660073  + 5017.26455388150 *T)
    sumY0  +=       0.148 * cos(5.46314197385  + 419.72846135871 *T)
    sumY0  +=       0.165 * cos(5.59205778351  + 20199.33877711650 *T)
    sumY0  +=       0.196 * cos(5.15079536124  + 33326.33491569069 *T)
    sumY0  +=       0.146 * cos(1.96564383477  + 19800.70213874130 *T)
    sumY0  +=       0.148 * cos(1.10385783973  + 6852.14154150230 *T)
    sumY0  +=       0.146 * cos(4.95120975616  + 17370.60479339330 *T)
    sumY0  +=       0.143 * cos(5.56441247552  + 5618.56362234490 *T)
    sumY0  +=       0.142 * cos(0.65902902155  + 17654.53672226610 *T)
    sumY0  +=       0.140 * cos(5.93480456615  + 2008.80135664250 *T)
    sumY0  +=       0.153 * cos(3.36826827070  + 5436.74919775670 *T)
    sumY0  +=       0.162 * cos(5.27294659555  + 775.27879384050 *T)
    sumY0  +=       0.162 * cos(3.21674523006  + 12552.16842346470 *T)
    sumY0  +=       0.143 * cos(3.89298583971  + 5010.15100688070 *T)
    sumY0  +=       0.188 * cos(2.11622447442  + 28.69300495130 *T)
    sumY0  +=       0.155 * cos(4.58570766573  + 11610.10636149570 *T)
    sumY0  +=       0.174 * cos(1.00725781500  + 20452.62559473830 *T)
    sumY0  +=       0.179 * cos(0.59168717373  + 27511.22405605370 *T)
    sumY0  +=       0.135 * cos(4.79054860554  + 12431.33043743090 *T)
    sumY0  +=       0.164 * cos(4.03840473864  + 28767.16824196750 *T)
    sumY0  +=       0.134 * cos(3.87019546457  + 16840.91382829870 *T)
    sumY0  +=       0.174 * cos(2.70675879120  + 19805.07110906630 *T)
    sumY0  +=       0.146 * cos(6.27087503385  + 12701.46059750170 *T)
    sumY0  +=       0.137 * cos(1.03021623583  + 11.28951774740 *T)
    sumY0  +=       0.167 * cos(5.98298407227  + 17473.69756761190 *T)
    sumY0  +=       0.152 * cos(2.62025173907  + 16627.61473286070 *T)
    sumY0  +=       0.124 * cos(4.22571397919  + 6948.07571260490 *T)
    sumY0  +=       0.135 * cos(4.95808633295  + 3531.28443281630 *T)
    sumY0  +=       0.172 * cos(0.23022549161  + 167959.14629657479 *T)
    sumY0  +=       0.123 * cos(0.49830220802  + 23013.29572210369 *T)
    sumY0  +=       0.122 * cos(4.12184350213  + 3583.58484815730 *T)
    sumY0  +=       0.170 * cos(6.16462326207  + 333857.83394425618 *T)
    sumY0  +=       0.155 * cos(5.62824965405  + 6058.48723680599 *T)
    sumY0  +=       0.120 * cos(5.90379805618  + 12809.12412144031 *T)
    sumY0  +=       0.141 * cos(3.95455666878  + 162420.79565227419 *T)
    sumY0  +=       0.120 * cos(4.26997270152  + 12528.36784483050 *T)
    sumY0  +=       0.123 * cos(1.49653119992  + 25933.88051360590 *T)
    sumY0  +=       0.144 * cos(3.70806542711  + 95.73540973430 *T)
    sumY0  +=       0.165 * cos(2.31400541864  + 52669.82577581910 *T)
    sumY0  +=       0.119 * cos(5.90934736896  + 19247.62037086590 *T)
    sumY0  +=       0.117 * cos(5.45034774436  + 11610.79577585770 *T)
    sumY0  +=       0.163 * cos(1.54456737879  + 661.47674426450 *T)
    sumY0  +=       0.163 * cos(2.78769461236  + 9929.67004482930 *T)
    sumY0  +=       0.119 * cos(2.71628583938  + 12250.00364780970 *T)
    sumY0  +=       0.145 * cos(2.33523694188  + 6205.64589704690 *T)
    sumY0  +=       0.145 * cos(1.17170397718  + 6360.99343790291 *T)
    sumY0  +=       0.159 * cos(2.61434005227  + 228278.34846897016 *T)
    sumY0  +=       0.114 * cos(3.41169068642  + 19402.55313533309 *T)
    sumY0  +=       0.147 * cos(5.63174103655  + 38526.33053338850 *T)
    sumY0  +=       0.156 * cos(1.00709118546  + 4307.82712161890 *T)
    sumY0  +=       0.131 * cos(0.13588364911  + 21228.63584102931 *T)
    sumY0  +=       0.118 * cos(3.83848144749  + 6263.64990657511 *T)
    sumY0  +=       0.118 * cos(5.95164477875  + 6302.98942837470 *T)
    sumY0  +=       0.137 * cos(0.97335445483  + 6315.85221826630 *T)
    sumY0  +=       0.137 * cos(2.53358646423  + 6250.78711668350 *T)
    sumY0  +=       0.109 * cos(1.19644618130  + 11925.51791008410 *T)
    sumY0  +=       0.111 * cos(2.80331366377  + 6226.42129253930 *T)
    sumY0  +=       0.111 * cos(0.70362725529  + 6340.21804241050 *T)
    sumY0  +=       0.123 * cos(0.80537647876  + 24734.39821404090 *T)
    sumY0  +=       0.107 * cos(0.59542543867  + 12463.30274324771 *T)
    sumY0  +=       0.133 * cos(1.45986519440  + 18875.76968725751 *T)
    sumY0  +=       0.123 * cos(4.07179306364  + 6260.54446602410 *T)
    sumY0  +=       0.123 * cos(5.71833316259  + 6306.09486892571 *T)
    sumY0  +=       0.113 * cos(0.38322617607  + 2111.89413086110 *T)
    sumY0  +=       0.131 * cos(2.04114030865  + 18415.75962958090 *T)
    sumY0  +=       0.110 * cos(0.19106597205  + 6289.94822637491 *T)
    sumY0  +=       0.110 * cos(3.31587494701  + 6276.69110857489 *T)
    sumY0  +=       0.103 * cos(2.31964172720  + 6241.76887649010 *T)
    sumY0  +=       0.103 * cos(1.18729919186  + 6324.87045845970 *T)
    sumY0  +=       0.113 * cos(0.63409757277  + 3496.27664361750 *T)
    sumY0  +=       0.103 * cos(3.00856574797  + 10344.53888286930 *T)
    sumY0  +=       0.115 * cos(1.09796447220  + 24336.24921063270 *T)
    sumY0  +=       0.107 * cos(0.14947871700  + 83974.07917321339 *T)
    sumY0  +=       0.107 * cos(1.18802993790  + 84020.10309797739 *T)
    sumY0  +=       0.106 * cos(1.70601019340  + 12772.58106590351 *T)
    sumY0  +=       0.143 * cos(6.22463763653  + 2069.25065239010 *T)
    sumY0  +=       0.143 * cos(5.63223127595  + 18773.20529618210 *T)
    sumY0  +=       0.115 * cos(5.35405637827  + 3641.49075403570 *T)
    sumY0  +=       0.112 * cos(1.02603600816  + 11499.41240530929 *T)
    sumY0  +=       0.099 * cos(0.21940282455  + 11190.62171762050 *T)
    sumY0  +=       0.126 * cos(5.90533164276  + 18823.17304765790 *T)
    sumY0  +=       0.115 * cos(3.09464836385  + 12570.32767072940 *T)
    sumY0  +=       0.100 * cos(1.95823440417  + 16062.42834360030 *T)
    sumY0  +=       0.107 * cos(2.45486702592  + 5216.82419028490 *T)
    sumY0  +=       0.120 * cos(0.25356721797  + 9814.36028280769 *T)
    sumY0  +=       0.138 * cos(2.98838673165  + 6210.02254161590 *T)
    sumY0  +=       0.138 * cos(0.51855418741  + 6356.61679333390 *T)
    sumY0  +=       0.105 * cos(2.60431357489  + 12721.32828193349 *T)
    sumY0  +=       0.135 * cos(5.28295396943  + 18699.90817032310 *T)
    sumY0  +=       0.104 * cos(0.57981411648  + 12560.87259315890 *T)
    sumY0  +=       0.098 * cos(4.00435005538  + 5815.35467712050 *T)
    sumY0  +=       0.111 * cos(5.54435587296  + 10239.34004852730 *T)
    sumY0  +=       0.100 * cos(2.98744963581  + 263.32774085630 *T)
    sumY0  +=       0.095 * cos(0.17662269567  + 155.66421691770 *T)
    sumY0  +=       0.130 * cos(0.55194216539  + 27511.71169102070 *T)
    sumY0  +=       0.120 * cos(1.84949333794  + 31441.43375227330 *T)
    sumY0  +=       0.114 * cos(2.18292460802  + 6155.30082413750 *T)
    sumY0  +=       0.114 * cos(1.32401631104  + 6411.33851081230 *T)
    sumY0  +=       0.119 * cos(3.64842963894  + 951.96222373410 *T)
    sumY0  +=       0.117 * cos(3.47158263266  + 28236.98964190589 *T)
    sumY0  +=       0.107 * cos(0.33835239167  + 21.01921297590 *T)
    sumY0  +=       0.095 * cos(4.06228359170  + 11300.82803883990 *T)
    sumY0  +=       0.110 * cos(1.22040625314  + 6312.74917601091 *T)
    sumY0  +=       0.110 * cos(2.28653466592  + 6253.89015893890 *T)
    sumY0  +=       0.097 * cos(2.64989366154  + 78263.95324220609 *T)
    sumY0  +=       0.095 * cos(4.41295990485  + 23938.10020722450 *T)
    sumY0  +=       0.091 * cos(1.42377843647  + 12829.47944083910 *T)
    sumY0  +=       0.112 * cos(2.74085534623  + 16737.82105408010 *T)
    sumY0  +=       0.126 * cos(4.67942071523  + 3.27930086550 *T)
    sumY0  +=       0.100 * cos(1.46206803081  + 6133.75647034030 *T)
    sumY0  +=       0.127 * cos(0.70264556724  + 1293.24040609909 *T)
    sumY0  +=       0.119 * cos(4.42085627054  + 17893.18221148710 *T)
    sumY0  +=       0.089 * cos(6.04696991631  + 23539.95120381630 *T)
    sumY0  +=       0.126 * cos(4.96989590925  + 311565.26273852371 *T)
    sumY0  +=       0.093 * cos(4.10519400640  + 736.12033101530 *T)
    sumY0  +=       0.120 * cos(0.51718078841  + 14158.99153109910 *T)
    sumY0  +=       0.118 * cos(3.00291998094  + 16061.94070863329 *T)
    sumY0  +=       0.099 * cos(2.04878528903  + 6432.88286460950 *T)
    sumY0  +=       0.095 * cos(1.28476695157  + 2699.49100183409 *T)
    sumY0  +=       0.089 * cos(0.73014646335  + 15671.32557689010 *T)
    sumY0  +=       0.124 * cos(5.19246908636  + 178430.29100801519 *T)
    sumY0  +=       0.089 * cos(5.78577582542  + 6751.28465782931 *T)
    sumY0  +=       0.096 * cos(1.07414401415  + 7349.81514466491 *T)
    sumY0  +=       0.096 * cos(5.45460641914  + 24066.05174025910 *T)
    sumY0  +=       0.103 * cos(2.80667748337  + 18202.46053414290 *T)
    sumY0  +=       0.099 * cos(4.49603988169  + 6252.40554183991 *T)
    sumY0  +=       0.099 * cos(5.29408634455  + 6314.23379310990 *T)
    sumY0  +=       0.090 * cos(1.96044978655  + 58864.30010066279 *T)
    sumY0  +=       0.086 * cos(2.16910774240  + 9380.71585523369 *T)
    sumY0  +=       0.086 * cos(2.50989174885  + 10557.35034334029 *T)
    sumY0  +=       0.098 * cos(1.84392439378  + 6439.72038797730 *T)
    sumY0  +=       0.098 * cos(1.66301652528  + 6126.91894697251 *T)
    sumY0  +=       0.086 * cos(6.05387198611  + 23123.98967829010 *T)
    sumY0  +=       0.118 * cos(2.16764726700  + 13951.95709241740 *T)
    sumY0  +=       0.117 * cos(2.53823002855  + 89570.23393702849 *T)
    sumY0  +=       0.083 * cos(4.71213278190  + 8858.07112683710 *T)
    sumY0  +=       0.092 * cos(2.44147365747  + 12342.05072176440 *T)
    sumY0  +=       0.085 * cos(0.72336255749  + 5017.75218884850 *T)
    sumY0  +=       0.101 * cos(4.94259233986  + 18429.98672358250 *T)
    sumY0  +=       0.088 * cos(2.19664804056  + 17054.21292373670 *T)
    sumY0  +=       0.113 * cos(0.47970289838  + 12985.88016134151 *T)
    sumY0  +=       0.083 * cos(2.47657261289  + 20597.00014555770 *T)
    sumY0  +=       0.082 * cos(4.90614394615  + 5483.49854230950 *T)
    sumY0  +=       0.087 * cos(1.67775805144  + 21424.22282681990 *T)
    sumY0  +=       0.107 * cos(3.78836692146  + 522.33360061030 *T)
    sumY0  +=       0.084 * cos(0.84028318953  + 6187.34044025710 *T)
    sumY0  +=       0.084 * cos(2.66665772953  + 6379.29889469270 *T)
    sumY0  +=       0.107 * cos(2.45229388212  + 24382.83529095790 *T)
    sumY0  +=       0.082 * cos(5.66196525445  + 8983.05448679250 *T)
    sumY0  +=       0.105 * cos(2.30294939014  + 6131.42238638970 *T)
    sumY0  +=       0.105 * cos(1.20399152892  + 6435.21694856010 *T)
    sumY0  +=       0.114 * cos(2.50001149568  + 8258.81221333091 *T)
    sumY0  +=       0.082 * cos(5.40031574607  + 3957.88262369230 *T)
    sumY0  +=       0.112 * cos(4.27968226719  + 3738.51761262449 *T)
    sumY0  +=       0.102 * cos(1.32703960779  + 5767.85579638190 *T)
    sumY0  +=       0.102 * cos(2.17990131127  + 6798.78353856790 *T)
    sumY0  +=       0.092 * cos(6.06909982221  + 18216.68762814450 *T)
    sumY0  +=       0.083 * cos(2.92326536483  + 29864.09020982550 *T)
    sumY0  +=       0.112 * cos(3.91825505548  + 12189.02190955050 *T)
    sumY0  +=       0.080 * cos(2.01989715959  + 24080.27883426070 *T)
    sumY0  +=       0.080 * cos(0.09553378894  + 15141.14697682849 *T)
    sumY0  +=       0.088 * cos(4.57487416308  + 1573.57660661529 *T)
    sumY0  +=       0.081 * cos(5.29168357177  + 1550.80140516450 *T)
    sumY0  +=       0.092 * cos(2.79378764615  + 101426.45258845300 *T)
    sumY0  +=       0.110 * cos(5.08245905369  + 78423.94833416230 *T)
    sumY0  +=       0.087 * cos(3.89276904283  + 1580.62284331330 *T)
    sumY0  +=       0.085 * cos(1.43001232077  + 27043.25906569930 *T)
    sumY0  +=       0.078 * cos(3.02198277867  + 6812.52299760250 *T)
    sumY0  +=       0.080 * cos(1.79867193339  + 6081.06627230081 *T)
    sumY0  +=       0.080 * cos(1.70826898567  + 6485.57306264900 *T)
    sumY0  +=       0.079 * cos(2.96736855880  + 36109.62602214810 *T)
    sumY0  +=       0.088 * cos(5.58307910278  + 16944.00660251730 *T)
    sumY0  +=       0.081 * cos(2.98944824758  + 5.27910682390 *T)
    sumY0  +=       0.093 * cos(1.14380332092  + 16737.33341911309 *T)
    sumY0  +=       0.088 * cos(1.67821631800  + 12537.94632999850 *T)
    sumY0  +=       0.096 * cos(0.48155847774  + 20198.85114214949 *T)
    sumY0  +=       0.106 * cos(3.73745326901  + 56600.03547203870 *T)
    sumY0  +=       0.097 * cos(1.71169429599  + 6040.10342853389 *T)
    sumY0  +=       0.081 * cos(5.19219909829  + 3956.26419853590 *T)
    sumY0  +=       0.077 * cos(6.10587791973  + 40796.58274015770 *T)
    sumY0  +=       0.076 * cos(3.78044498975  + 22743.16556203289 *T)
    sumY0  +=       0.098 * cos(0.77646852784  + 42456.54022960810 *T)
    sumY0  +=       0.083 * cos(0.63561076928  + 19801.18977370830 *T)
    sumY0  +=       0.079 * cos(2.93697401198  + 5622.08674069391 *T)
    sumY0  +=       0.096 * cos(0.41418225890  + 5888.69378241570 *T)
    sumY0  +=       0.096 * cos(3.09275866016  + 6677.94555253410 *T)
    sumY0  +=       0.075 * cos(4.66945944657  + 41194.73174356590 *T)
    sumY0  +=       0.074 * cos(5.37967983395  + 6261.98402700790 *T)
    sumY0  +=       0.074 * cos(4.41044639229  + 6304.65530794190 *T)
    sumY0  +=       0.073 * cos(2.45644384945  + 5870.94857060050 *T)
    sumY0  +=       0.073 * cos(1.05049706961  + 6695.69076434930 *T)
    sumY0  +=       0.085 * cos(2.08161902252  + 12850.25483633150 *T)
    sumY0  +=       0.072 * cos(4.75417028947  + 6253.49822932610 *T)
    sumY0  +=       0.072 * cos(5.03595593677  + 6313.14110562371 *T)
    sumY0  +=       0.100 * cos(3.80714894085  + 5316.34879003930 *T)
    sumY0  +=       0.085 * cos(1.33904345426  + 12282.53619860110 *T)
    sumY0  +=       0.075 * cos(0.39510472937  + 24422.61416889080 *T)
    sumY0  +=       0.076 * cos(0.41148054527  + 63659.12156832110 *T)
    sumY0  +=       0.072 * cos(4.97177288715  + 16723.59396007850 *T)
    sumY0  +=       0.097 * cos(0.21179463686  + 17995.78735073869 *T)
    sumY0  +=       0.087 * cos(2.06337614338  + 18106.48130692510 *T)
    sumY0  +=       0.080 * cos(3.60508005382  + 17363.49124639250 *T)
    sumY0  +=       0.070 * cos(3.06053136401  + 6124.37614969171 *T)
    sumY0  +=       0.070 * cos(0.44640955505  + 6442.26318525810 *T)
    sumY0  +=       0.070 * cos(4.42572835968  + 4705.48849006010 *T)
    sumY0  +=       0.078 * cos(3.37769653143  + 23550.58786916610 *T)
    sumY0  +=       0.070 * cos(3.84581453123  + 12036.21691740469 *T)
    sumY0  +=       0.077 * cos(5.00579533671  + 5237.67719632029 *T)
    sumY0  +=       0.069 * cos(1.30888647192  + 16207.64245401849 *T)
    sumY0  +=       0.082 * cos(3.91623659935  + 6774.98295993371 *T)
    sumY0  +=       0.067 * cos(4.88284115481  + 7083.14079264031 *T)
    sumY0  +=       0.067 * cos(0.26461151247  + 6394.50630976251 *T)
    sumY0  +=       0.067 * cos(3.24232940659  + 6172.13302518730 *T)
    sumY0  +=       0.077 * cos(0.28397734649  + 9924.56660402710 *T)
    sumY0  +=       0.080 * cos(3.79678848458  + 22380.99961775750 *T)
    sumY0  +=       0.070 * cos(3.03095059330  + 6390.98319141350 *T)
    sumY0  +=       0.070 * cos(0.47599032576  + 6175.65614353630 *T)
    sumY0  +=       0.074 * cos(4.12433601146  + 16193.41536001689 *T)
    sumY0  +=       0.076 * cos(4.07719654455  + 32217.44399856430 *T)
    sumY0  +=       0.091 * cos(0.90446873976  + 6653.01948341530 *T)
    sumY0  +=       0.091 * cos(2.60247217930  + 5913.61985153450 *T)
    sumY0  +=       0.067 * cos(4.94532713838  + 6265.50714535691 *T)
    sumY0  +=       0.067 * cos(4.84479908786  + 6301.13218959290 *T)
    sumY0  +=       0.067 * cos(3.01186125982  + 5959.32661585050 *T)
    sumY0  +=       0.083 * cos(4.51648793763  + 16723.10632511149 *T)
    sumY0  +=       0.080 * cos(5.49180050451  + 23646.56709638390 *T)
    sumY0  +=       0.068 * cos(2.07232606852  + 4897.42439113870 *T)
    sumY0  +=       0.069 * cos(0.56996686256  + 6944.55259425590 *T)
    sumY0  +=       0.089 * cos(3.57060375879  + 10660.44311755889 *T)
    sumY0  +=       0.079 * cos(3.24586891365  + 35371.64344849290 *T)
    sumY0  +=       0.073 * cos(0.54359299539  + 6370.62787201471 *T)
    sumY0  +=       0.073 * cos(2.96334792367  + 6196.01146293510 *T)
    sumY0  +=       0.065 * cos(5.23672190600  + 22345.01655862470 *T)
    sumY0  +=       0.064 * cos(5.80516008052  + 15265.64270181689 *T)
    sumY0  +=       0.066 * cos(1.12499582368  + 6315.29197329170 *T)
    sumY0  +=       0.066 * cos(2.38194509538  + 6251.34736165811 *T)
    sumY0  +=       0.065 * cos(3.44248995108  + 323.74923414091 *T)
    sumY0  +=       0.076 * cos(1.46082924902  + 10873.74221299689 *T)
    sumY0  +=       0.084 * cos(0.49974269222  + 109.96250373590 *T)
    sumY0  +=       0.063 * cos(2.70916029747  + 11823.40545693370 *T)
    sumY0  +=       0.064 * cos(4.80332084911  + 28774.28178896830 *T)
    sumY0  +=       0.072 * cos(2.11265208796  + 18099.75944096650 *T)
    sumY0  +=       0.066 * cos(1.21054975257  + 17576.79034183050 *T)
    sumY0  +=       0.072 * cos(0.02703719822  + 245707.70992182777 *T)
    sumY0  +=       0.061 * cos(3.91641579991  + 10557.83797830730 *T)
    sumY0  +=       0.064 * cos(5.46712129325  + 71430.93943561260 *T)
    sumY0  +=       0.068 * cos(1.33734717206  + 28760.05469496669 *T)
    sumY0  +=       0.062 * cos(2.25046201901  + 3854.78984947370 *T)
    sumY0  +=       0.073 * cos(4.61252989167  + 23440.38154794670 *T)
    sumY0  +=       0.060 * cos(1.65871121739  + 13088.97293556010 *T)
    sumY0  +=       0.078 * cos(2.99014426818  + 12564.91104475801 *T)
    sumY0  +=       0.060 * cos(2.90135289748  + 7548.88714610130 *T)
    sumY0  +=       0.059 * cos(4.12263682656  + 6286.35515085690 *T)
    sumY0  +=       0.075 * cos(2.85045311521  + 18625.12657175580 *T)
    sumY0  +=       0.058 * cos(5.18922617388  + 35050.24409195890 *T)
    sumY0  +=       0.060 * cos(4.91746089028  + 553.81322032590 *T)
    sumY0  +=       0.069 * cos(4.25329717989  + 167993.93845370729 *T)
    sumY0  +=       0.059 * cos(5.13629564655  + 41991.02975038230 *T)
    sumY0  +=       0.062 * cos(4.44636395235  + 15663.79170522509 *T)
    sumY0  +=       0.081 * cos(5.98297744820  + 7250.29054491051 *T)
    sumY0  +=       0.059 * cos(1.11500712714  + 6277.25375184510 *T)
    sumY0  +=       0.059 * cos(2.39193379192  + 6289.38558310470 *T)
    sumY0  +=       0.063 * cos(0.56300539120  + 18003.38853270651 *T)
    sumY0  +=       0.057 * cos(3.30567043605  + 793.01870595090 *T)
    sumY0  +=       0.057 * cos(1.59794444440  + 10027.65937824569 *T)
    sumY0  +=       0.060 * cos(0.63448798616  + 647.25465079831 *T)
    sumY0  +=       0.072 * cos(5.73566114928  + 4597.77991000629 *T)
    sumY0  +=       0.059 * cos(5.39392547964  + 6279.30891415889 *T)
    sumY0  +=       0.068 * cos(1.14499717138  + 3166.66025521511 *T)
    sumY0  +=       0.056 * cos(3.10776343336  + 6226.51640530510 *T)
    sumY0  +=       0.056 * cos(0.39917748570  + 6340.12292964471 *T)
    sumY0  +=       0.056 * cos(0.95649088929  + 12303.31159409350 *T)
    sumY0  +=       0.057 * cos(1.51972566902  + 5864.39526857430 *T)
    sumY0  +=       0.057 * cos(1.98721525004  + 6702.24406637550 *T)
    sumY0  +=       0.069 * cos(0.33222800273  + 23536.36077516450 *T)
    sumY0  +=       0.059 * cos(2.36664156781  + 5760.25461441409 *T)
    sumY0  +=       0.065 * cos(1.96288183765  + 12139.30969162329 *T)
    sumY0  +=       0.059 * cos(1.81423548051  + 67589.33126454070 *T)
    sumY0  +=       0.073 * cos(2.08854442290  + 11079.92776143409 *T)
    sumY0  +=       0.076 * cos(0.77803347659  + 23227.08245250870 *T)
    sumY0  +=       0.067 * cos(0.54633497662  + 4804.45309341050 *T)
    sumY0  +=       0.073 * cos(4.43869578954  + 30349.12759025050 *T)
    sumY0  +=       0.074 * cos(3.17776060575  + 92747.93298430610 *T)
    sumY0  +=       0.070 * cos(6.27818224660  + 12299.78847574450 *T)
    sumY0  +=       0.060 * cos(5.08460970539  + 12171.76963240710 *T)
    sumY0  +=       0.059 * cos(1.03458131364  + 15.49628866851 *T)
    sumY0  +=       0.074 * cos(1.19968434762  + 3684.13423250670 *T)
    sumY0  +=       0.058 * cos(1.58539308271  + 6717.03140535170 *T)
    sumY0  +=       0.053 * cos(3.68674599490  + 26087.65932409069 *T)
    sumY0  +=       0.059 * cos(5.76308404993  + 12164.72339570910 *T)
    sumY0  +=       0.063 * cos(5.23211999381  + 5219.51794905560 *T)
    sumY0  +=       0.056 * cos(2.45377850108  + 3178.38960805111 *T)
    sumY0  +=       0.063 * cos(2.16148808063  + 5227.38225558179 *T)
    sumY0  +=       0.070 * cos(4.83368738065  + 19004.40413192490 *T)
    sumY0  +=       0.072 * cos(2.62595674809  + 4583.55281600469 *T)
    sumY0  +=       0.052 * cos(0.96524582701  + 3627.26366003410 *T)
    sumY0  +=       0.052 * cos(3.30278586555  + 6411.27120050950 *T)
    sumY0  +=       0.052 * cos(0.20415505351  + 6155.36813444030 *T)
    sumY0  +=       0.054 * cos(2.04667654344  + 5849.60792959810 *T)
    sumY0  +=       0.066 * cos(5.93782050272  + 5791.16874004909 *T)
    sumY0  +=       0.066 * cos(5.93782050272  + 5791.65637501610 *T)
    sumY0  +=       0.051 * cos(2.41252260456  + 5113.24378109930 *T)
    sumY0  +=       0.058 * cos(2.72227505748  + 30775.72578112650 *T)
    sumY0  +=       0.061 * cos(0.72832616881  + 14169.54974474690 *T)
    sumY0  +=       0.054 * cos(3.16101850189  + 10454.74520408871 *T)
    sumY0  +=       0.056 * cos(1.20035157643  + 23141.80220040810 *T)
    sumY0  +=       0.067 * cos(3.89077193587  + 28313.04498717750 *T)
    sumY0  +=       0.056 * cos(0.37854134464  + 5244.29305668450 *T)
    sumY0  +=       0.062 * cos(0.60927850338  + 5657.64947516250 *T)
    sumY0  +=       0.062 * cos(2.89766241568  + 6908.98985978730 *T)
    sumY0  +=       0.050 * cos(3.10994458213  + 2574.99527684569 *T)
    sumY0  +=       0.050 * cos(2.89352148340  + 536.56069461190 *T)
    sumY0  +=       0.050 * cos(0.38317924664  + 6684.99178923210 *T)
    sumY0  +=       0.050 * cos(3.12376167242  + 5881.64754571770 *T)
    sumY0  +=       0.049 * cos(0.36435757253  + 16310.73522823709 *T)
    sumY0  +=       0.049 * cos(3.50595022612  + 16311.22286320410 *T)
    sumY0  +=       0.065 * cos(2.20156720545  + 46386.74992582770 *T)
    sumY0  +=       0.069 * cos(5.16096365273  + 60530.24516825830 *T)
    sumY0  +=       0.068 * cos(4.61952317468  + 96678.14268052568 *T)
    sumY0  +=       0.049 * cos(4.01249891427  + 12323.17927852529 *T)
    sumY0  +=       0.050 * cos(2.38095415684  + 21954.40142688150 *T)
    sumY0  +=       0.049 * cos(4.98226292129  + 4164.55580709650 *T)
    sumY0  +=       0.059 * cos(0.54830716960  + 33794.78754101210 *T)
    sumY0  +=       0.050 * cos(1.64594745036  + 43739.04616344930 *T)
    sumY0  +=       0.059 * cos(2.71376255252  + 3981.73385156551 *T)
    sumY0  +=       0.058 * cos(3.16580349936  + 27707.29867681129 *T)
    sumY0  +=       0.051 * cos(1.43461397348  + 7669.21494381111 *T)
    sumY0  +=       0.048 * cos(3.23576326319  + 3646.59419483791 *T)
    sumY0  +=       0.065 * cos(1.63790414031  + 82534.64743809510 *T)
    sumY0  +=       0.056 * cos(2.62361047024  + 800.06494264891 *T)
    sumY0  +=       0.057 * cos(2.56200976684  + 11609.61872652869 *T)
    sumY0  +=       0.048 * cos(3.66284145075  + 26.54213728330 *T)
    sumY0  +=       0.047 * cos(5.82227097996  + 30640.10045611650 *T)
    sumY0  +=       0.060 * cos(3.54507538375  + 8982.56685182550 *T)
    sumY0  +=       0.049 * cos(2.15364675842  + 96563.24283557820 *T)
    sumY0  +=       0.050 * cos(3.60628396854  + 17583.90388883130 *T)
    sumY0  +=       0.046 * cos(4.33029118489  + 23539.46356884929 *T)
    sumY0  +=       0.048 * cos(4.40476950322  + 170.42905313570 *T)
    sumY0  +=       0.055 * cos(0.06525117672  + 735.63269604830 *T)
    sumY0  +=       0.054 * cos(5.90233323358  + 6680.98103591609 *T)
    sumY0  +=       0.060 * cos(3.29126676894  + 44033.88377205590 *T)
    sumY0  +=       0.064 * cos(5.97731735914  + 80181.29364935629 *T)
    sumY0  +=       0.059 * cos(3.60959132830  + 24279.35083569710 *T)
    sumY0  +=       0.049 * cos(3.59486989773  + 5490.54477900750 *T)
    sumY0  +=       0.058 * cos(5.95314247069  + 6490.02040477150 *T)
    sumY0  +=       0.058 * cos(3.83698375555  + 6076.61893017830 *T)
    sumY0  +=       0.046 * cos(1.35461053261  + 18208.10612510850 *T)
    sumY0  +=       0.045 * cos(3.19696372316  + 4480.46004178609 *T)
    sumY0  +=       0.055 * cos(1.99803947181  + 16097.92376776610 *T)
    sumY0  +=       0.055 * cos(1.99803947181  + 16097.43613279909 *T)
    sumY0  +=       0.061 * cos(3.53347961961  + 491.90710994230 *T)
    sumY0  +=       0.054 * cos(1.22650408797  + 6006.57389561050 *T)
    sumY0  +=       0.054 * cos(2.28043683109  + 6560.06543933930 *T)
    sumY0  +=       0.044 * cos(4.77367705513  + 5333.65642353809 *T)
    sumY0  +=       0.052 * cos(3.85218997059  + 244287.84382471119 *T)
    sumY0  +=       0.044 * cos(2.69173127084  + 2301.34199842589 *T)
    sumY0  +=       0.045 * cos(4.35951224787  + 33794.29990604510 *T)
    sumY0  +=       0.046 * cos(3.84874383995  + 4384.48081456829 *T)
    sumY0  +=       0.048 * cos(4.63629874583  + 35579.93505705350 *T)
    sumY0  +=       0.051 * cos(4.91886576255  + 18326.89394936390 *T)
    sumY0  +=       0.045 * cos(2.63923033609  + 533.45790092711 *T)
    sumY0  +=       0.046 * cos(3.15339657493  + 16703.30595098250 *T)
    sumY0  +=       0.047 * cos(2.39403388232  + 5209.22300831710 *T)
    sumY0  +=       0.051 * cos(1.31112479278  + 38650.41732368250 *T)
    sumY0  +=       0.046 * cos(2.47672465258  + 12555.34981720240 *T)
    sumY0  +=       0.052 * cos(4.64446990453  + 5992.34680160890 *T)
    sumY0  +=       0.052 * cos(5.14565632171  + 6574.29253334090 *T)
    sumY0  +=       0.044 * cos(3.80837636245  + 69942.19741831250 *T)
    sumY0  +=       0.049 * cos(3.62132943845  + 12146.91087359110 *T)
    sumY0  +=       0.053 * cos(4.67452542700  + 39301.85314471250 *T)
    sumY0  +=       0.045 * cos(4.65107782321  + 46360.93924099490 *T)
    sumY0  +=       0.043 * cos(1.54528326296  + 8402.08352785330 *T)
    sumY0  +=       0.043 * cos(0.19791989913  + 44137.19516685750 *T)
    sumY0  +=       0.049 * cos(3.54725333487  + 22029.96913718610 *T)
    sumY0  +=       0.044 * cos(1.47478212539  + 31441.92138724031 *T)
    sumY0  +=       0.042 * cos(2.85938112982  + 9225.29545579949 *T)
    sumY0  +=       0.045 * cos(3.22032934298  + 71519.54096076029 *T)
    sumY0  +=       0.051 * cos(2.95623880108  + 7762.18624153930 *T)
    sumY0  +=       0.041 * cos(5.14395096650  + 3735.48212924251 *T)
    sumY0  +=       0.057 * cos(1.24099052077  + 40398.43373674950 *T)
    sumY0  +=       0.047 * cos(6.03171570454  + 4707.98623122570 *T)
    sumY0  +=       0.056 * cos(5.36736063063  + 28664.07546774890 *T)
    sumY0  +=       0.050 * cos(2.58729930740  + 11919.38468415150 *T)
    sumY0  +=       0.047 * cos(5.61037642387  + 316428.47249139857 *T)
    sumY0  +=       0.041 * cos(2.70468936208  + 27278.71263392430 *T)
    sumY0  +=       0.056 * cos(1.28041103259  + 83659.52068988250 *T)
    sumY0  +=       0.048 * cos(0.93182108526  + 12662.37474468410 *T)
    sumY0  +=       0.054 * cos(2.40910872834  + 16627.12709789369 *T)
    sumY0  +=       0.052 * cos(2.84085536047  + 12571.91844177370 *T)
    sumY0  +=       0.043 * cos(3.12713892129  + 7322.34627826531 *T)
    sumY0  +=       0.041 * cos(1.99075260862  + 664.99986261351 *T)
    sumY0  +=       0.042 * cos(3.11818521171  + 5641.95442512570 *T)
    sumY0  +=       0.044 * cos(6.16004757680  + 6425.49829451110 *T)
    sumY0  +=       0.044 * cos(3.63007864944  + 6141.14104043870 *T)
    sumY0  +=       0.039 * cos(5.09444795757  + 22594.29871319550 *T)
    sumY0  +=       0.040 * cos(1.17403028794  + 20894.53186172529 *T)
    sumY0  +=       0.044 * cos(3.71228299188  + 5540.32960694230 *T)
    sumY0  +=       0.050 * cos(0.98055905828  + 17782.48825530069 *T)
    sumY0  +=       0.054 * cos(3.12084664124  + 6006.28467373350 *T)
    sumY0  +=       0.054 * cos(0.38609427782  + 6560.35466121630 *T)
    sumY0  +=       0.045 * cos(0.83260974343  + 142861.14741877469 *T)
    sumY0  +=       0.047 * cos(5.83397233956  + 24499.76347813590 *T)
    sumY0  +=       0.043 * cos(6.07854396898  + 7026.30972800750 *T)
    sumY0  +=       0.047 * cos(0.38354823927  + 955.84355909210 *T)
    sumY0  +=       0.047 * cos(2.93749588479  + 24485.53638413430 *T)
    sumY0  +=       0.040 * cos(1.03647885736  + 9388.24972689870 *T)
    sumY0  +=       0.043 * cos(4.24359473122  + 6359.58573875050 *T)
    sumY0  +=       0.042 * cos(5.53288597787  + 8635.69818627969 *T)
    sumY0  +=       0.051 * cos(4.03405257508  + 27831.79440179969 *T)
    sumY0  +=       0.042 * cos(0.37690679264  + 6334.60000533731 *T)
    sumY0  +=       0.042 * cos(3.13003412642  + 6232.03932961249 *T)
    sumY0  +=       0.052 * cos(1.96232281613  + 11905.16259068530 *T)
    sumY0  +=       0.040 * cos(6.22187049858  + 7076.09455594230 *T)
    sumY0  +=       0.040 * cos(5.00617355027  + 5444.35037972450 *T)
    sumY0  +=       0.040 * cos(4.78395267597  + 7122.28895522530 *T)
    sumY0  +=       0.038 * cos(0.67446092329  + 44933.49317367390 *T)
    sumY0  +=       0.037 * cos(4.98656910933  + 5715.60102974450 *T)
    sumY0  +=       0.037 * cos(4.80355711691  + 6851.03830520530 *T)
    sumY0  +=       0.042 * cos(5.51152847292  + 36147.65369478390 *T)
    sumY0  +=       0.042 * cos(1.28742672679  + 6238.17844783830 *T)
    sumY0  +=       0.042 * cos(2.21951419227  + 6328.46088711150 *T)
    sumY0  +=       0.040 * cos(5.21427505339  + 3214.89256315970 *T)
    sumY0  +=       0.043 * cos(2.22601042104  + 23141.31456544110 *T)
    sumY0  +=       0.041 * cos(5.52595261887  + 19804.58347409930 *T)
    sumY0  +=       0.038 * cos(4.48062431201  + 20.11150191529 *T)
    sumY0  +=       0.050 * cos(0.05709227621  + 84334.66158130829 *T)
    sumY0  +=       0.050 * cos(4.41842921988  + 60284.16619777939 *T)

    # EARTH Y Coord
    sumY1  = 0
    sumY1  +=  154550.744 * cos(5.35844734917  + 0.24381748350 *T)
    sumY1  +=   51503.383 * cos(4.43183566713  + 12566.39551746630 *T)
    sumY1  +=    1290.763 * cos(4.38862020180  + 18849.47136745770 *T)
    sumY1  +=     702.576 * cos(0.18268310766  + 6283.31966747490 *T)
    sumY1  +=     430.422 * cos(4.02296125150  + 6282.83203250789 *T)
    sumY1  +=     212.689 * cos(0.16300557365  + 6279.79654912590 *T)
    sumY1  +=     212.524 * cos(3.34405166686  + 6286.84278582391 *T)
    sumY1  +=      62.308 * cos(5.07432195748  + 4705.97612502710 *T)
    sumY1  +=      59.808 * cos(2.24108323998  + 6257.02134767510 *T)
    sumY1  +=      59.474 * cos(1.26580020213  + 6309.61798727471 *T)
    sumY1  +=      48.914 * cos(3.64253038225  + 775.76642880750 *T)
    sumY1  +=      42.814 * cos(5.15102931682  + 1059.62574767270 *T)
    sumY1  +=      46.457 * cos(4.72328803502  + 7860.66320992270 *T)
    sumY1  +=      36.653 * cos(0.62556081234  + 5753.62870238030 *T)
    sumY1  +=      35.649 * cos(6.15990909241  + 5885.17066406670 *T)
    sumY1  +=      35.362 * cos(2.90135814576  + 6813.01063256950 *T)
    sumY1  +=      32.151 * cos(3.62597628866  + 6681.46867088311 *T)
    sumY1  +=      28.763 * cos(4.34539356227  + 25132.54721744910 *T)
    sumY1  +=      28.447 * cos(5.86215127530  + 6127.89926804070 *T)
    sumY1  +=      27.537 * cos(3.92840810925  + 6438.74006690910 *T)
    sumY1  +=      24.819 * cos(1.35151106278  + 5487.02166065850 *T)
    sumY1  +=      20.615 * cos(2.14741827387  + 7079.61767429131 *T)
    sumY1  +=      19.621 * cos(1.31571192713  + 5507.79705615090 *T)
    sumY1  +=      18.378 * cos(6.17760240913  + 11790.87290614230 *T)
    sumY1  +=      16.494 * cos(4.64463821908  + 11507.01358727710 *T)
    sumY1  +=      16.756 * cos(2.24858801378  + 7058.84227879890 *T)
    sumY1  +=      14.558 * cos(4.40450472214  + 6290.43321447570 *T)
    sumY1  +=      14.404 * cos(5.39361420229  + 6276.20612047410 *T)
    sumY1  +=      14.052 * cos(6.16986019111  + 796.54182429990 *T)
    sumY1  +=      12.261 * cos(5.72868820428  + 4693.75913722409 *T)
    sumY1  +=      12.758 * cos(5.06070649359  + 7.35736448430 *T)
    sumY1  +=      10.086 * cos(1.72162917776  + 3739.00524759150 *T)
    sumY1  +=       9.469 * cos(3.06840825171  + 6070.02057203690 *T)
    sumY1  +=      10.425 * cos(0.81162914438  + 6284.29998854310 *T)
    sumY1  +=      10.425 * cos(2.69531177468  + 6282.33934640670 *T)
    sumY1  +=       9.570 * cos(0.32347467387  + 4137.15425099970 *T)
    sumY1  +=       9.044 * cos(0.43893334417  + 6496.61876291290 *T)
    sumY1  +=       8.544 * cos(4.76970151659  + 1194.69082770810 *T)
    sumY1  +=       8.214 * cos(5.96640392040  + 1589.31671276730 *T)
    sumY1  +=       6.157 * cos(1.78596836636  + 8827.63408735830 *T)
    sumY1  +=       6.271 * cos(3.18396187423  + 8429.48508395010 *T)
    sumY1  +=       5.524 * cos(3.79861381511  + 4933.45225781610 *T)
    sumY1  +=       7.314 * cos(5.20341424492  + 4535.30325440790 *T)
    sumY1  +=       5.157 * cos(3.57885530448  + 11770.09751064990 *T)
    sumY1  +=       6.749 * cos(2.75433387783  + 5088.87265725030 *T)
    sumY1  +=       4.853 * cos(5.33701721257  + 6040.59106350090 *T)
    sumY1  +=       5.304 * cos(0.93005422139  + 3154.44326741210 *T)
    sumY1  +=       4.985 * cos(3.05697984231  + 12569.91863581531 *T)
    sumY1  +=       4.709 * cos(2.39306975468  + 5331.60126122430 *T)
    sumY1  +=       4.601 * cos(4.45296491447  + 6526.04827144891 *T)
    sumY1  +=       4.201 * cos(5.99364092036  + 7633.18707713370 *T)
    sumY1  +=       5.607 * cos(0.98720876729  + 5729.75026463250 *T)
    sumY1  +=       4.257 * cos(2.55370490945  + 3930.45351370310 *T)
    sumY1  +=       4.394 * cos(4.75948183378  + 12559.28197046550 *T)
    sumY1  +=       3.870 * cos(1.11601490340  + 7235.03807372550 *T)
    sumY1  +=       5.154 * cos(4.58850581748  + 8031.33608054190 *T)
    sumY1  +=       5.036 * cos(2.51929382179  + 6836.88907031730 *T)
    sumY1  +=       5.076 * cos(0.73526458069  + 7477.76667769950 *T)
    sumY1  +=       3.593 * cos(3.97650177033  + 12565.90788249930 *T)
    sumY1  +=       3.384 * cos(3.83274545225  + 10977.32262218251 *T)
    sumY1  +=       3.371 * cos(5.00200001284  + 11371.94850724170 *T)
    sumY1  +=       3.536 * cos(5.30432752281  + 4164.06817212950 *T)
    sumY1  +=       4.292 * cos(3.36706453434  + 1592.83983111630 *T)
    sumY1  +=       3.407 * cos(0.83519640650  + 3128.63258257930 *T)
    sumY1  +=       3.152 * cos(0.16276893009  + 5223.93773728571 *T)
    sumY1  +=       2.883 * cos(0.20838349180  + 1747.77259558350 *T)
    sumY1  +=       2.751 * cos(3.67369812017  + 7342.70159766410 *T)
    sumY1  +=       3.323 * cos(3.48959648614  + 801.57711364030 *T)
    sumY1  +=       2.999 * cos(0.68275608544  + 8636.18582124671 *T)
    sumY1  +=       2.496 * cos(5.09157906387  + 2145.92159899169 *T)
    sumY1  +=       3.358 * cos(5.82350976219  + 155.17658195069 *T)
    sumY1  +=       2.297 * cos(2.96479731544  + 17260.39847217390 *T)
    sumY1  +=       2.047 * cos(1.96580631113  + 1990.98883452450 *T)
    sumY1  +=       2.778 * cos(6.17391326343  + 5481.49873635110 *T)
    sumY1  +=       2.025 * cos(3.01082711135  + 951.47458876710 *T)
    sumY1  +=       2.663 * cos(1.00159277919  + 26.05450231630 *T)
    sumY1  +=       2.009 * cos(4.16311202386  + 4690.72365384210 *T)
    sumY1  +=       2.148 * cos(0.32608511593  + 537.04832957890 *T)
    sumY1  +=       1.853 * cos(4.41841662668  + 553.32558535889 *T)
    sumY1  +=       2.478 * cos(1.61013003664  + 1349.62359217529 *T)
    sumY1  +=       2.023 * cos(0.82015419806  + 522.82123557730 *T)
    sumY1  +=       1.735 * cos(4.49024043092  + 529.44714761109 *T)
    sumY1  +=       2.417 * cos(3.59693248271  + 7085.14059859870 *T)
    sumY1  +=       1.809 * cos(2.35536809783  + 397.90518592470 *T)
    sumY1  +=       1.893 * cos(2.65666986732  + 9438.00675237050 *T)
    sumY1  +=       1.708 * cos(5.00361322564  + 10989.05197501850 *T)
    sumY1  +=       1.898 * cos(3.91087003459  + 5216.33655531789 *T)
    sumY1  +=       1.820 * cos(3.44808607160  + 5230.56364931950 *T)
    sumY1  +=       1.541 * cos(5.81316348536  + 426.84200835950 *T)
    sumY1  +=       1.555 * cos(2.09924495059  + 13096.08648256090 *T)
    sumY1  +=       1.779 * cos(0.23737949225  + 12562.87239911730 *T)
    sumY1  +=       1.448 * cos(1.75844475078  + 5753.14106741329 *T)
    sumY1  +=       1.450 * cos(5.71201368260  + 6262.96434807611 *T)
    sumY1  +=       1.450 * cos(4.07811254364  + 6303.67498687370 *T)
    sumY1  +=       1.822 * cos(0.13446699374  + 10973.79950383350 *T)
    sumY1  +=       1.372 * cos(5.60520656116  + 7875.91568110771 *T)
    sumY1  +=       1.375 * cos(2.34041210234  + 7084.65296363170 *T)
    sumY1  +=       1.400 * cos(3.28144398068  + 12721.81591690050 *T)
    sumY1  +=       1.247 * cos(0.29225580262  + 2119.00767786191 *T)
    sumY1  +=       1.231 * cos(2.53164637932  + 18319.78040236310 *T)
    sumY1  +=       1.141 * cos(1.92161882496  + 1066.73929467350 *T)
    sumY1  +=       1.138 * cos(1.23855949465  + 2942.70724077510 *T)
    sumY1  +=       1.047 * cos(1.31663136695  + 5642.93474619389 *T)
    sumY1  +=       1.058 * cos(2.19001683016  + 242.97242145751 *T)
    sumY1  +=       1.025 * cos(0.60899466577  + 10447.14402212089 *T)
    sumY1  +=       0.973 * cos(4.95398300734  + 640.14110379750 *T)
    sumY1  +=       1.114 * cos(4.36978275435  + 15721.08260236190 *T)
    sumY1  +=       0.963 * cos(0.56100772029  + 2389.13783793270 *T)
    sumY1  +=       0.941 * cos(4.21626686326  + 20426.81490990550 *T)
    sumY1  +=       1.068 * cos(5.08632600308  + 529.93478257810 *T)
    sumY1  +=       0.929 * cos(1.54055499699  + 10575.65050042530 *T)
    sumY1  +=       0.899 * cos(5.79114005938  + 16496.60521368590 *T)
    sumY1  +=       0.889 * cos(5.01192484889  + 6277.79674316750 *T)
    sumY1  +=       0.889 * cos(4.77820137734  + 6288.84259178230 *T)
    sumY1  +=       1.018 * cos(2.66299766067  + 12540.09719766650 *T)
    sumY1  +=       1.060 * cos(5.76865363011  + 5760.74224938110 *T)
    sumY1  +=       1.017 * cos(4.02093017183  + 6805.89708556871 *T)
    sumY1  +=       0.797 * cos(3.79336643920  + 14314.41193053330 *T)
    sumY1  +=       0.795 * cos(0.10742263482  + 6286.60602489270 *T)
    sumY1  +=       0.795 * cos(3.39951828424  + 6280.03331005710 *T)
    sumY1  +=       0.800 * cos(3.18366130818  + 12029.59100537090 *T)
    sumY1  +=       0.719 * cos(2.29666441302  + 9623.93209417470 *T)
    sumY1  +=       0.698 * cos(4.94363251872  + 6148.25458743950 *T)
    sumY1  +=       0.686 * cos(3.79368723538  + 5856.72147659890 *T)
    sumY1  +=       0.673 * cos(3.99105175160  + 12964.54452087450 *T)
    sumY1  +=       0.902 * cos(0.69698416448  + 13368.21644859010 *T)
    sumY1  +=       0.678 * cos(4.84647193821  + 6418.38474751031 *T)
    sumY1  +=       0.671 * cos(6.00179402654  + 6709.91785835091 *T)
    sumY1  +=       0.744 * cos(2.68741072621  + 12043.81809937250 *T)
    sumY1  +=       0.706 * cos(2.51950741945  + 16730.70750707931 *T)
    sumY1  +=       0.661 * cos(2.42216375426  + 14712.56093394150 *T)
    sumY1  +=       0.633 * cos(5.58153853935  + 4292.57465043390 *T)
    sumY1  +=       0.621 * cos(1.74121096386  + 13119.96492030870 *T)
    sumY1  +=       0.638 * cos(2.04188798982  + 4690.23601887509 *T)
    sumY1  +=       0.625 * cos(0.31613419014  + 13518.11392371690 *T)
    sumY1  +=       0.637 * cos(4.59654254753  + 5746.51515537950 *T)
    sumY1  +=       0.628 * cos(3.42448659886  + 4686.64559022330 *T)
    sumY1  +=       0.653 * cos(5.85737456207  + 3929.96587873609 *T)
    sumY1  +=       0.578 * cos(0.62377454326  + 5088.38502228330 *T)
    sumY1  +=       0.563 * cos(3.24964920260  + 10447.63165708790 *T)
    sumY1  +=       0.610 * cos(5.19955844918  + 6820.12417957030 *T)
    sumY1  +=       0.666 * cos(4.87089013467  + 3634.37720703489 *T)
    sumY1  +=       0.720 * cos(5.18811029494  + 13916.26292712510 *T)
    sumY1  +=       0.620 * cos(4.30310535016  + 31415.62306744050 *T)
    sumY1  +=       0.553 * cos(2.71699575652  + 6259.44122972711 *T)
    sumY1  +=       0.553 * cos(0.78994516254  + 6307.19810522270 *T)
    sumY1  +=       0.523 * cos(5.47007776670  + 14143.73905991410 *T)
    sumY1  +=       0.524 * cos(3.97726441461  + 12139.79732659030 *T)
    sumY1  +=       0.569 * cos(4.83897139720  + 12036.70455237170 *T)
    sumY1  +=       0.514 * cos(3.87608366777  + 4700.87268422489 *T)
    sumY1  +=       0.563 * cos(3.78589343112  + 11014.86265985130 *T)
    sumY1  +=       0.495 * cos(2.00987402269  + 13362.69352428270 *T)
    sumY1  +=       0.481 * cos(0.93576028071  + 6279.43833211690 *T)
    sumY1  +=       0.481 * cos(2.57118063835  + 6287.20100283290 *T)
    sumY1  +=       0.466 * cos(2.94306081142  + 10177.50149701710 *T)
    sumY1  +=       0.440 * cos(5.16955709579  + 1577.58735993130 *T)
    sumY1  +=       0.437 * cos(1.61052953646  + 11499.90004027630 *T)
    sumY1  +=       0.452 * cos(3.98330238308  + 12573.50906446710 *T)
    sumY1  +=       0.499 * cos(5.79985868924  + 12410.97511803210 *T)
    sumY1  +=       0.440 * cos(0.57972339309  + 24073.16528725990 *T)
    sumY1  +=       0.587 * cos(0.00678196141  + 6.86972951729 *T)
    sumY1  +=       0.388 * cos(3.61198235279  + 7860.17557495569 *T)
    sumY1  +=       0.398 * cos(4.20068362004  + 8274.06468451590 *T)
    sumY1  +=       0.423 * cos(5.87838349624  + 4694.24677219110 *T)
    sumY1  +=       0.530 * cos(1.43593166640  + 12592.69383726610 *T)
    sumY1  +=       0.379 * cos(0.45309221604  + 6180.22689325630 *T)
    sumY1  +=       0.379 * cos(3.05384870302  + 6386.41244169350 *T)
    sumY1  +=       0.377 * cos(4.33272540326  + 7872.39256275871 *T)
    sumY1  +=       0.419 * cos(4.05140241721  + 5327.71992586630 *T)
    sumY1  +=       0.383 * cos(4.67353188247  + 6247.29180079750 *T)
    sumY1  +=       0.383 * cos(5.11659434377  + 6319.34753415231 *T)
    sumY1  +=       0.375 * cos(0.91076946484  + 2352.62233628830 *T)
    sumY1  +=       0.457 * cos(2.62120808067  + 6077.13411903770 *T)
    sumY1  +=       0.457 * cos(0.88573283839  + 6489.50521591210 *T)
    sumY1  +=       0.481 * cos(3.46562904756  + 4292.08701546690 *T)
    sumY1  +=       0.344 * cos(4.99906947930  + 18451.32236404950 *T)
    sumY1  +=       0.453 * cos(5.43816135018  + 2787.28684134090 *T)
    sumY1  +=       0.364 * cos(2.82703012884  + 6245.29199483910 *T)
    sumY1  +=       0.364 * cos(0.67991079022  + 6321.34734011070 *T)
    sumY1  +=       0.327 * cos(1.14128906507  + 12323.66691349230 *T)
    sumY1  +=       0.358 * cos(0.15971399837  + 9917.45305702629 *T)
    sumY1  +=       0.342 * cos(5.26072614817  + 6262.54427198250 *T)
    sumY1  +=       0.342 * cos(4.52940007806  + 6304.09506296730 *T)
    sumY1  +=       0.318 * cos(4.84603466410  + 11926.49823115230 *T)
    sumY1  +=       0.320 * cos(1.13215851865  + 77714.01528560399 *T)
    sumY1  +=       0.354 * cos(5.73706716445  + 7238.91940908350 *T)
    sumY1  +=       0.318 * cos(3.75423383039  + 6254.87048000710 *T)
    sumY1  +=       0.318 * cos(6.03589239585  + 6311.76885494270 *T)
    sumY1  +=       0.304 * cos(0.43890915611  + 12779.69461290430 *T)
    sumY1  +=       0.343 * cos(2.14223194635  + 6298.57213865991 *T)
    sumY1  +=       0.343 * cos(1.36470897271  + 6268.06719628990 *T)
    sumY1  +=       0.359 * cos(2.52875634099  + 1052.51220067191 *T)
    sumY1  +=       0.398 * cos(4.83537015899  + 1551.28904013150 *T)
    sumY1  +=       0.306 * cos(1.01793913946  + 5863.83502359970 *T)
    sumY1  +=       0.300 * cos(0.20529153395  + 90280.16698558679 *T)
    sumY1  +=       0.307 * cos(4.86343200722  + 3893.93801205869 *T)
    sumY1  +=       0.297 * cos(2.08300781298  + 5429.63565075589 *T)
    sumY1  +=       0.365 * cos(1.17462915798  + 10022.08109758290 *T)
    sumY1  +=       0.336 * cos(1.15285276926  + 17782.97589026770 *T)
    sumY1  +=       0.295 * cos(2.48770580169  + 6702.80431135010 *T)
    sumY1  +=       0.288 * cos(0.61716195626  + 18073.94875613370 *T)
    sumY1  +=       0.287 * cos(4.90856628789  + 1577.09972496430 *T)
    sumY1  +=       0.378 * cos(5.11185072605  + 2353.10997125530 *T)
    sumY1  +=       0.289 * cos(1.47281714617  + 213.54291292150 *T)
    sumY1  +=       0.347 * cos(2.29918133719  + 5223.45010231870 *T)
    sumY1  +=       0.321 * cos(0.66708336426  + 17797.20298426930 *T)
    sumY1  +=       0.289 * cos(3.73571084362  + 220.65645992230 *T)
    sumY1  +=       0.260 * cos(2.61364471837  + 955.35592412510 *T)
    sumY1  +=       0.268 * cos(5.19927474541  + 14945.07235607090 *T)
    sumY1  +=       0.242 * cos(2.77307042091  + 2544.55823736690 *T)
    sumY1  +=       0.245 * cos(3.88798752568  + 7632.69944216670 *T)
    sumY1  +=       0.234 * cos(2.50313805280  + 1596.43025976811 *T)
    sumY1  +=       0.267 * cos(4.25976324066  + 206.42936592071 *T)
    sumY1  +=       0.282 * cos(2.93181023342  + 13341.91812879030 *T)
    sumY1  +=       0.279 * cos(4.70853062526  + 4731.78680985990 *T)
    sumY1  +=       0.290 * cos(0.46137131163  + 5642.44206009270 *T)
    sumY1  +=       0.298 * cos(5.65061835836  + 17790.08943726851 *T)
    sumY1  +=       0.270 * cos(0.59449456012  + 12168.24651405810 *T)
    sumY1  +=       0.294 * cos(0.95816279728  + 2146.40923395870 *T)
    sumY1  +=       0.214 * cos(2.50623255971  + 8030.84844557490 *T)
    sumY1  +=       0.207 * cos(3.18045634687  + 213.05527795450 *T)
    sumY1  +=       0.216 * cos(4.03237077499  + 3185.43584474911 *T)
    sumY1  +=       0.195 * cos(4.44222765867  + 5884.68302909969 *T)
    sumY1  +=       0.199 * cos(2.36156558922  + 1748.26023055050 *T)
    sumY1  +=       0.250 * cos(3.04516390608  + 6924.19727485710 *T)
    sumY1  +=       0.194 * cos(0.79669370564  + 641.12142486571 *T)
    sumY1  +=       0.203 * cos(0.68276896116  + 6503.73230991370 *T)
    sumY1  +=       0.203 * cos(2.82417195790  + 6062.90702503610 *T)
    sumY1  +=       0.258 * cos(0.91786109532  + 796.05418933290 *T)
    sumY1  +=       0.189 * cos(3.20457516516  + 11506.52595231009 *T)
    sumY1  +=       0.186 * cos(4.84188314285  + 18209.57408114370 *T)
    sumY1  +=       0.184 * cos(2.63875516800  + 12566.46282776910 *T)
    sumY1  +=       0.180 * cos(4.19729049815  + 853.44019923550 *T)
    sumY1  +=       0.185 * cos(6.25586281827  + 3495.78900865049 *T)
    sumY1  +=       0.205 * cos(0.22424585668  + 5849.12029463110 *T)
    sumY1  +=       0.187 * cos(4.34056279496  + 10213.52936369450 *T)
    sumY1  +=       0.187 * cos(0.48606035843  + 6290.36590417291 *T)
    sumY1  +=       0.187 * cos(3.02088056063  + 6276.27343077690 *T)
    sumY1  +=       0.235 * cos(4.34532332705  + 9779.35249360890 *T)
    sumY1  +=       0.173 * cos(6.09941552473  + 19651.29229858150 *T)
    sumY1  +=       0.184 * cos(3.08285813784  + 12566.32820716351 *T)
    sumY1  +=       0.230 * cos(0.66694562429  + 12567.37583853451 *T)
    sumY1  +=       0.166 * cos(6.03836479949  + 110.45013870291 *T)
    sumY1  +=       0.189 * cos(1.27307578752  + 3.76693583251 *T)
    sumY1  +=       0.166 * cos(5.90720968173  + 433.95555536030 *T)
    sumY1  +=       0.186 * cos(6.04048637618  + 5863.34738863270 *T)
    sumY1  +=       0.184 * cos(4.58711235161  + 10983.94853421629 *T)
    sumY1  +=       0.184 * cos(1.80970242494  + 6037.48802124550 *T)
    sumY1  +=       0.156 * cos(0.77353737613  + 2648.69864295650 *T)
    sumY1  +=       0.182 * cos(5.04351148530  + 10969.72144021469 *T)
    sumY1  +=       0.175 * cos(1.69691915551  + 6529.15131370430 *T)
    sumY1  +=       0.154 * cos(4.10646056877  + 6453.99253809410 *T)
    sumY1  +=       0.154 * cos(5.68366565747  + 6112.64679685570 *T)
    sumY1  +=       0.164 * cos(2.95193833305  + 12353.09642202830 *T)
    sumY1  +=       0.146 * cos(2.48650028218  + 149.80701461810 *T)
    sumY1  +=       0.184 * cos(4.91752478989  + 11015.35029481830 *T)
    sumY1  +=       0.191 * cos(0.72322692096  + 3894.42564702570 *T)
    sumY1  +=       0.144 * cos(2.19099222002  + 18636.17227201970 *T)
    sumY1  +=       0.146 * cos(0.42084749575  + 13760.84252769090 *T)
    sumY1  +=       0.142 * cos(2.91645242674  + 4156.95462512869 *T)
    sumY1  +=       0.185 * cos(5.29181252606  + 7234.55043875850 *T)
    sumY1  +=       0.133 * cos(4.73063574833  + 18139.53831889940 *T)
    sumY1  +=       0.131 * cos(0.04654287508  + 5507.30942118390 *T)
    sumY1  +=       0.146 * cos(0.32618788162  + 18875.28205229050 *T)
    sumY1  +=       0.128 * cos(4.98427830485  + 6069.53293706989 *T)
    sumY1  +=       0.128 * cos(1.75166251045  + 16200.52890701769 *T)
    sumY1  +=       0.123 * cos(0.73895960068  + 20.53157800890 *T)
    sumY1  +=       0.121 * cos(5.69434162626  + 17256.87535382490 *T)
    sumY1  +=       0.123 * cos(4.81514598285  + 6393.52598869430 *T)
    sumY1  +=       0.123 * cos(4.97498024339  + 6173.11334625550 *T)
    sumY1  +=       0.117 * cos(2.90632080371  + 18852.99448580670 *T)
    sumY1  +=       0.120 * cos(5.75152020178  + 9381.20349020070 *T)
    sumY1  +=       0.113 * cos(3.90786027747  + 65147.37595065419 *T)
    sumY1  +=       0.137 * cos(2.11120201406  + 16201.01654198470 *T)
    sumY1  +=       0.113 * cos(0.36200887356  + 6836.40143535029 *T)
    sumY1  +=       0.123 * cos(2.85759760823  + 12565.41519639810 *T)
    sumY1  +=       0.132 * cos(4.51334720616  + 17655.02435723310 *T)
    sumY1  +=       0.113 * cos(5.69373873614  + 10575.16286545830 *T)
    sumY1  +=       0.113 * cos(3.66776032820  + 10213.04172872750 *T)
    sumY1  +=       0.116 * cos(0.66749075538  + 17253.28492517310 *T)
    sumY1  +=       0.120 * cos(0.70954573147  + 10420.23010099111 *T)
    sumY1  +=       0.109 * cos(5.21836603318  + 633.02755679670 *T)
    sumY1  +=       0.103 * cos(4.83071292354  + 6295.04902031090 *T)
    sumY1  +=       0.103 * cos(4.95941330270  + 6271.59031463890 *T)
    sumY1  +=       0.104 * cos(4.01023132524  + 23013.78335707070 *T)
    sumY1  +=       0.105 * cos(5.32684825848  + 11790.38527117530 *T)
    sumY1  +=       0.109 * cos(4.22545370482  + 3340.36860921629 *T)
    sumY1  +=       0.102 * cos(1.02602291037  + 10970.20907518170 *T)
    sumY1  +=       0.101 * cos(5.99270341494  + 4803.96545844350 *T)
    sumY1  +=       0.114 * cos(3.78830179341  + 4171.18171913030 *T)
    sumY1  +=       0.111 * cos(2.09152457722  + 1582.20316576650 *T)
    sumY1  +=       0.107 * cos(0.90140521385  + 23543.47432216530 *T)
    sumY1  +=       0.099 * cos(4.24784832079  + 10973.31186886650 *T)
    sumY1  +=       0.102 * cos(4.09962394648  + 6549.92670919670 *T)
    sumY1  +=       0.102 * cos(5.69050227976  + 6016.71262575310 *T)
    sumY1  +=       0.120 * cos(0.92616350319  + 11514.12713427790 *T)
    sumY1  +=       0.095 * cos(1.12824235931  + 17267.51201917470 *T)
    sumY1  +=       0.121 * cos(4.76575826483  + 13625.77744765550 *T)
    sumY1  +=       0.100 * cos(3.61435033946  + 10976.83498721549 *T)
    sumY1  +=       0.090 * cos(2.11761227033  + 76.50988875911 *T)
    sumY1  +=       0.100 * cos(5.94744998603  + 8661.99650607950 *T)
    sumY1  +=       0.089 * cos(3.76175372959  + 1350.11122714230 *T)
    sumY1  +=       0.087 * cos(2.76310699450  + 9917.94069199330 *T)
    sumY1  +=       0.086 * cos(4.36870119289  + 12809.12412144031 *T)
    sumY1  +=       0.095 * cos(6.05597457730  + 12345.98287502750 *T)
    sumY1  +=       0.084 * cos(3.11610500480  + 775.27879384050 *T)
    sumY1  +=       0.106 * cos(3.57201004990  + 11216.52810780750 *T)
    sumY1  +=       0.095 * cos(0.92966555257  + 12012.82611462390 *T)
    sumY1  +=       0.082 * cos(4.72839816349  + 5643.42238116090 *T)
    sumY1  +=       0.081 * cos(2.27692529332  + 11614.67711121570 *T)
    sumY1  +=       0.084 * cos(5.85994160066  + 6255.91811137810 *T)
    sumY1  +=       0.084 * cos(3.93018462558  + 6310.72122357170 *T)
    sumY1  +=       0.080 * cos(5.05950625443  + 6923.21695378890 *T)
    sumY1  +=       0.085 * cos(1.08248438849  + 5635.82119919310 *T)
    sumY1  +=       0.084 * cos(4.61502942146  + 10440.03047512009 *T)
    sumY1  +=       0.080 * cos(2.62622746074  + 3583.58484815730 *T)
    sumY1  +=       0.092 * cos(5.59171891913  + 10818.37910439930 *T)
    sumY1  +=       0.077 * cos(4.13228553675  + 6993.25271603320 *T)
    sumY1  +=       0.080 * cos(6.08197788088  + 13521.99525907490 *T)
    sumY1  +=       0.104 * cos(0.70806546737  + 5650.04829319470 *T)
    sumY1  +=       0.093 * cos(5.00569753414  + 4732.27444482690 *T)
    sumY1  +=       0.085 * cos(1.74569264442  + 22805.49174851010 *T)
    sumY1  +=       0.086 * cos(5.52511956951  + 12360.20996902910 *T)
    sumY1  +=       0.075 * cos(5.57236272383  + 5120.35732810009 *T)
    sumY1  +=       0.090 * cos(5.33797857569  + 6370.62787201471 *T)
    sumY1  +=       0.090 * cos(4.45214765055  + 6196.01146293510 *T)
    sumY1  +=       0.082 * cos(4.91006846138  + 18842.35782045690 *T)
    sumY1  +=       0.101 * cos(1.37756832018  + 3097.64000524229 *T)
    sumY1  +=       0.094 * cos(0.74657649932  + 10177.01386205009 *T)
    sumY1  +=       0.083 * cos(3.74913323692  + 18415.75962958090 *T)
    sumY1  +=       0.073 * cos(4.52746639645  + 949.41942645330 *T)
    sumY1  +=       0.101 * cos(4.43327916798  + 3104.68624194030 *T)
    sumY1  +=       0.074 * cos(1.25396392931  + 13517.62628874989 *T)
    sumY1  +=       0.092 * cos(0.77959714125  + 16858.72635041670 *T)
    sumY1  +=       0.075 * cos(1.57278335709  + 1059.13811270570 *T)
    sumY1  +=       0.085 * cos(4.52163234616  + 398.39282089170 *T)
    sumY1  +=       0.069 * cos(5.60384680292  + 11713.19913571430 *T)
    sumY1  +=       0.084 * cos(5.92557534615  + 2378.92065608810 *T)
    sumY1  +=       0.069 * cos(2.18342250812  + 11712.71150074729 *T)
    sumY1  +=       0.075 * cos(0.05753316454  + 30356.24113725130 *T)
    sumY1  +=       0.067 * cos(3.04134731899  + 3154.93090237910 *T)
    sumY1  +=       0.076 * cos(3.26401364210  + 18429.98672358250 *T)
    sumY1  +=       0.068 * cos(2.69185508250  + 11925.51791008410 *T)
    sumY1  +=       0.067 * cos(4.81689271569  + 10454.25756912169 *T)
    sumY1  +=       0.073 * cos(1.35808826471  + 15670.83794192309 *T)
    sumY1  +=       0.065 * cos(1.77865763794  + 6438.25243194210 *T)
    sumY1  +=       0.069 * cos(1.72420196911  + 17298.42614480970 *T)
    sumY1  +=       0.087 * cos(3.52675260029  + 3904.15519390330 *T)
    sumY1  +=       0.076 * cos(5.44623063701  + 5231.05128428650 *T)
    sumY1  +=       0.064 * cos(2.12314868459  + 3496.27664361750 *T)
    sumY1  +=       0.078 * cos(2.74559935849  + 8672.21368792411 *T)
    sumY1  +=       0.066 * cos(4.98895923052  + 14143.25142494710 *T)
    sumY1  +=       0.063 * cos(5.70776457338  + 24357.02460612510 *T)
    sumY1  +=       0.063 * cos(0.57965828598  + 15720.59496739490 *T)
    sumY1  +=       0.061 * cos(2.15619405920  + 16460.57734700850 *T)
    sumY1  +=       0.073 * cos(6.07488391156  + 9387.76209193169 *T)
    sumY1  +=       0.057 * cos(4.37179192959  + 17259.91083720689 *T)
    sumY1  +=       0.059 * cos(2.26511827996  + 9778.86485864190 *T)
    sumY1  +=       0.058 * cos(1.67874293827  + 155.66421691770 *T)
    sumY1  +=       0.060 * cos(1.94469304669  + 34570.31015233610 *T)
    sumY1  +=       0.057 * cos(2.88994977642  + 149854.64395229140 *T)
    sumY1  +=       0.056 * cos(3.76314176838  + 12456.18919624690 *T)
    sumY1  +=       0.056 * cos(5.62479142768  + 17996.27498570570 *T)
    sumY1  +=       0.056 * cos(0.60886736064  + 11764.57458634250 *T)
    sumY1  +=       0.054 * cos(4.53224684011  + 4705.48849006010 *T)
    sumY1  +=       0.055 * cos(6.11029124003  + 13915.77529215809 *T)
    sumY1  +=       0.061 * cos(4.86430679802  + 28237.47727687290 *T)
    sumY1  +=       0.065 * cos(4.29965528550  + 7335.58805066330 *T)
    sumY1  +=       0.072 * cos(5.48795829629  + 6055.84353468590 *T)
    sumY1  +=       0.072 * cos(4.30216792995  + 6510.79580026390 *T)
    sumY1  +=       0.071 * cos(4.94322519248  + 12545.62012197390 *T)
    sumY1  +=       0.061 * cos(5.98387073766  + 6312.74917601091 *T)
    sumY1  +=       0.061 * cos(3.80625548858  + 6253.89015893890 *T)
    sumY1  +=       0.052 * cos(1.83719780245  + 5326.54287653730 *T)
    sumY1  +=       0.054 * cos(2.78309524301  + 2699.49100183409 *T)
    sumY1  +=       0.052 * cos(0.87738933830  + 8983.05448679250 *T)
    sumY1  +=       0.066 * cos(0.78258344897  + 14919.26167123810 *T)
    sumY1  +=       0.050 * cos(1.22263365304  + 5572.89898394960 *T)
    sumY1  +=       0.051 * cos(2.33500088089  + 16522.41589851870 *T)
    sumY1  +=       0.049 * cos(2.61658441943  + 35579.93505705350 *T)
    sumY1  +=       0.053 * cos(6.07542895582  + 6208.53806890760 *T)
    sumY1  +=       0.053 * cos(3.71469727042  + 6358.10126604220 *T)
    sumY1  +=       0.049 * cos(3.35593265214  + 21393.78578734110 *T)
    sumY1  +=       0.049 * cos(4.47248238697  + 5547.44315394310 *T)
    sumY1  +=       0.049 * cos(5.31764383927  + 7019.19618100671 *T)
    sumY1  +=       0.049 * cos(1.02044068888  + 12416.83232033170 *T)
    sumY1  +=       0.050 * cos(0.93918602686  + 12592.20620229910 *T)
    sumY1  +=       0.050 * cos(4.78759248733  + 26084.26562369970 *T)
    sumY1  +=       0.051 * cos(1.83601012538  + 23006.66981006990 *T)
    sumY1  +=       0.058 * cos(4.03623415117  + 15141.14697682849 *T)
    sumY1  +=       0.047 * cos(5.68903994773  + 24279.35083569710 *T)
    sumY1  +=       0.049 * cos(3.40926939847  + 6816.53375091851 *T)
    sumY1  +=       0.049 * cos(0.09767152059  + 5750.10558403130 *T)
    sumY1  +=       0.046 * cos(1.88054905125  + 19379.16233255230 *T)
    sumY1  +=       0.052 * cos(5.74808775798  + 16737.33341911309 *T)
    sumY1  +=       0.047 * cos(0.52219046486  + 24066.05174025910 *T)
    sumY1  +=       0.046 * cos(4.16717022039  + 104351.85638378028 *T)
    sumY1  +=       0.063 * cos(6.25484174196  + 8662.48414104651 *T)
    sumY1  +=       0.048 * cos(3.28503810088  + 26735.70144472970 *T)
    sumY1  +=       0.046 * cos(3.91775350902  + 12132.68377958950 *T)
    sumY1  +=       0.050 * cos(2.92364175862  + 24602.85625235450 *T)
    sumY1  +=       0.054 * cos(4.75214039499  + 7834.36489012290 *T)
    sumY1  +=       0.045 * cos(3.48807574467  + 161000.92955515761 *T)
    sumY1  +=       0.049 * cos(2.94405201174  + 29303.97275406290 *T)
    sumY1  +=       0.049 * cos(1.43156848803  + 10984.43616918330 *T)
    sumY1  +=       0.045 * cos(0.70883599249  + 45584.92899470390 *T)
    sumY1  +=       0.046 * cos(1.86995211145  + 5244.29305668450 *T)
    sumY1  +=       0.048 * cos(3.53510882628  + 23581.01435983410 *T)
    sumY1  +=       0.049 * cos(2.17177379198  + 19650.80466361450 *T)
    sumY1  +=       0.049 * cos(3.41897587826  + 29289.74566006130 *T)
    sumY1  +=       0.044 * cos(1.44008987426  + 6390.98319141350 *T)
    sumY1  +=       0.044 * cos(2.06685104480  + 6175.65614353630 *T)
    sumY1  +=       0.051 * cos(2.65580143245  + 17789.60180230149 *T)
    sumY1  +=       0.042 * cos(6.18543275273  + 24705.94902657310 *T)
    sumY1  +=       0.042 * cos(5.90805829894  + 15663.79170522509 *T)
    sumY1  +=       0.043 * cos(3.08250524084  + 16460.08971204149 *T)
    sumY1  +=       0.043 * cos(0.28590581328  + 553.81322032590 *T)
    sumY1  +=       0.041 * cos(3.21615501236  + 17370.60479339330 *T)
    sumY1  +=       0.053 * cos(2.59151074700  + 13119.47728534169 *T)
    sumY1  +=       0.042 * cos(5.11995884892  + 4164.55580709650 *T)
    sumY1  +=       0.041 * cos(2.05521298551  + 23020.89690407150 *T)
    sumY1  +=       0.045 * cos(3.70227971577  + 54247.16931826690 *T)
    sumY1  +=       0.040 * cos(4.57473494926  + 32243.25468339710 *T)
    sumY1  +=       0.040 * cos(5.17184328434  + 24336.24921063270 *T)
    sumY1  +=       0.040 * cos(1.71852303243  + 11769.60987568290 *T)
    sumY1  +=       0.045 * cos(6.25985265406  + 16723.10632511149 *T)
    sumY1  +=       0.045 * cos(6.25985265406  + 16723.59396007850 *T)
    sumY1  +=       0.042 * cos(5.89327075469  + 8827.14645239129 *T)
    sumY1  +=       0.040 * cos(4.71493835645  + 8402.08352785330 *T)
    sumY1  +=       0.039 * cos(3.49714273023  + 16062.42834360030 *T)
    sumY1  +=       0.044 * cos(4.44510876745  + 5856.23384163189 *T)
    sumY1  +=       0.038 * cos(1.14212165839  + 6297.54676147650 *T)
    sumY1  +=       0.038 * cos(2.36481926067  + 6269.09257347330 *T)
    sumY1  +=       0.041 * cos(3.38233495451  + 15508.85894075790 *T)
    sumY1  +=       0.038 * cos(4.96796930709  + 6394.50630976251 *T)
    sumY1  +=       0.038 * cos(4.82215691915  + 6172.13302518730 *T)
    sumY1  +=       0.039 * cos(1.74815791619  + 29826.55017215670 *T)
    sumY1  +=       0.046 * cos(0.31322632556  + 17297.93850984270 *T)
    sumY1  +=       0.041 * cos(4.04315782590  + 6309.13035230770 *T)
    sumY1  +=       0.038 * cos(5.82409895120  + 29089.05523346850 *T)
    sumY1  +=       0.038 * cos(4.59409111885  + 11933.61177815310 *T)
    sumY1  +=       0.037 * cos(3.34091252113  + 26482.41462710790 *T)
    sumY1  +=       0.037 * cos(3.64722285417  + 20452.62559473830 *T)
    sumY1  +=       0.037 * cos(0.30900035436  + 18073.46112116669 *T)
    sumY1  +=       0.036 * cos(2.98410187235  + 11499.41240530929 *T)
    sumY1  +=       0.045 * cos(4.75649198460  + 3340.85624418330 *T)
    sumY1  +=       0.049 * cos(4.21088151991  + 18216.68762814450 *T)
    sumY1  +=       0.038 * cos(4.59187061808  + 18202.46053414290 *T)
    sumY1  +=       0.041 * cos(3.91005845267  + 5216.82419028490 *T)
    sumY1  +=       0.036 * cos(6.21139146959  + 4060.97539791089 *T)
    sumY1  +=       0.048 * cos(5.14345155225  + 951.96222373410 *T)
    sumY1  +=       0.041 * cos(3.90932016585  + 18848.98373249069 *T)
    sumY1  +=       0.038 * cos(4.60880580452  + 26709.89075989690 *T)
    sumY1  +=       0.036 * cos(3.12521745961  + 28230.43104017490 *T)
    sumY1  +=       0.036 * cos(1.96840086543  + 24491.18197509989 *T)
    sumY1  +=       0.036 * cos(1.56811412304  + 9924.56660402710 *T)
    sumY1  +=       0.048 * cos(1.89701220521  + 8982.56685182550 *T)
    sumY1  +=       0.037 * cos(3.60669363968  + 22484.09239197610 *T)
    sumY1  +=       0.047 * cos(3.21269518092  + 6040.10342853389 *T)
    sumY1  +=       0.034 * cos(3.68972702028  + 6418.94499248490 *T)
    sumY1  +=       0.034 * cos(6.10039920596  + 6147.69434246491 *T)
    sumY1  +=       0.034 * cos(3.67007112439  + 9380.71585523369 *T)
    sumY1  +=       0.036 * cos(0.30403746373  + 7238.43177411650 *T)
    sumY1  +=       0.034 * cos(0.40887602826  + 17272.12782500990 *T)
    sumY1  +=       0.037 * cos(1.23154839984  + 58953.38926077750 *T)
    sumY1  +=       0.034 * cos(0.33427002863  + 29026.72904699130 *T)
    sumY1  +=       0.033 * cos(3.87501941695  + 12721.32828193349 *T)
    sumY1  +=       0.035 * cos(1.63582276543  + 7322.34627826531 *T)
    sumY1  +=       0.044 * cos(1.21937301495  + 3981.73385156551 *T)
    sumY1  +=       0.033 * cos(3.59821221655  + 19804.58347409930 *T)
    sumY1  +=       0.033 * cos(2.56047890182  + 173567.08125514039 *T)
    sumY1  +=       0.033 * cos(6.16657680956  + 205.94173095370 *T)
    sumY1  +=       0.033 * cos(5.44889821497  + 23013.29572210369 *T)
    sumY1  +=       0.034 * cos(2.13261854972  + 49515.13869092350 *T)
    sumY1  +=       0.034 * cos(3.38050467777  + 38500.51984855570 *T)
    sumY1  +=       0.032 * cos(2.81221233882  + 17686.99666304990 *T)
    sumY1  +=       0.036 * cos(1.67423219491  + 5641.95442512570 *T)
    sumY1  +=       0.032 * cos(2.91090812586  + 21228.63584102931 *T)
    sumY1  +=       0.033 * cos(0.13839577785  + 34520.55312686430 *T)
    sumY1  +=       0.033 * cos(5.65380556855  + 8635.69818627969 *T)
    sumY1  +=       0.031 * cos(1.88042187886  + 5746.02752041250 *T)
    sumY1  +=       0.041 * cos(2.52021697492  + 21202.33752122950 *T)
    sumY1  +=       0.037 * cos(5.92836585991  + 7349.81514466491 *T)
    sumY1  +=       0.034 * cos(2.18099177342  + 16496.11757871889 *T)
    sumY1  +=       0.032 * cos(0.90143427409  + 33019.26492968810 *T)
    sumY1  +=       0.032 * cos(1.54173979189  + 21150.56954840009 *T)
    sumY1  +=       0.037 * cos(1.84830381005  + 29296.85920706210 *T)
    sumY1  +=       0.036 * cos(5.02250294215  + 27511.22405605370 *T)
    sumY1  +=       0.035 * cos(3.77934660139  + 11933.12414318609 *T)
    sumY1  +=       0.035 * cos(4.26341483447  + 11918.89704918449 *T)
    sumY1  +=       0.035 * cos(5.54313072444  + 29062.75691366870 *T)
    sumY1  +=       0.030 * cos(1.38932865248  + 12490.12944619070 *T)
    sumY1  +=       0.032 * cos(5.09927453137  + 6334.60000533731 *T)
    sumY1  +=       0.032 * cos(4.69085169487  + 6232.03932961249 *T)
    sumY1  +=       0.029 * cos(4.05993716780  + 419.24082639170 *T)
    sumY1  +=       0.029 * cos(0.91834451421  + 419.72846135871 *T)
    sumY1  +=       0.029 * cos(1.66178131503  + 16858.23871544969 *T)
    sumY1  +=       0.030 * cos(2.73190240418  + 16208.13008898551 *T)
    sumY1  +=       0.029 * cos(0.69175314473  + 27278.71263392430 *T)
    sumY1  +=       0.032 * cos(4.73525892997  + 5017.26455388150 *T)
    sumY1  +=       0.028 * cos(0.30767945102  + 24080.27883426070 *T)
    sumY1  +=       0.030 * cos(1.00265951424  + 23539.95120381630 *T)
    sumY1  +=       0.029 * cos(1.45883005475  + 18606.25512851669 *T)
    sumY1  +=       0.030 * cos(4.73151177298  + 30665.91114094930 *T)
    sumY1  +=       0.034 * cos(4.29199382621  + 6660.69327539070 *T)
    sumY1  +=       0.034 * cos(5.49813240003  + 5905.94605955911 *T)
    sumY1  +=       0.038 * cos(3.55143054136  + 31571.04346687470 *T)
    sumY1  +=       0.038 * cos(4.53015131722  + 16061.94070863329 *T)
    sumY1  +=       0.029 * cos(1.64885508302  + 23539.46356884929 *T)
    sumY1  +=       0.033 * cos(2.88700963334  + 3738.51761262449 *T)
    sumY1  +=       0.028 * cos(6.02264922068  + 2942.21960580809 *T)
    sumY1  +=       0.032 * cos(2.75250951602  + 15110.70993734970 *T)
    sumY1  +=       0.030 * cos(6.10434010761  + 55023.17956455790 *T)
    sumY1  +=       0.027 * cos(5.24524976767  + 233141.55822184496 *T)
    sumY1  +=       0.030 * cos(0.75326086687  + 12587.17091295870 *T)
    sumY1  +=       0.028 * cos(3.97691648841  + 34911.65589357450 *T)
    sumY1  +=       0.027 * cos(4.60109278395  + 1692.40948698591 *T)
    sumY1  +=       0.027 * cos(6.13743188286  + 2544.07060239989 *T)
    sumY1  +=       0.027 * cos(4.99580076168  + 24382.83529095790 *T)
    sumY1  +=       0.027 * cos(4.42136306455  + 6205.64589704690 *T)
    sumY1  +=       0.027 * cos(5.36876316169  + 6360.99343790291 *T)
    sumY1  +=       0.027 * cos(4.95719787895  + 31172.89446346650 *T)
    sumY1  +=       0.026 * cos(2.19655051246  + 36949.47462590770 *T)
    sumY1  +=       0.034 * cos(4.45594644643  + 18053.17336064130 *T)
    sumY1  +=       0.027 * cos(0.74968073356  + 32367.34147369110 *T)
    sumY1  +=       0.027 * cos(3.34260766105  + 4487.57358878689 *T)
    sumY1  +=       0.028 * cos(5.37411547339  + 10239.34004852730 *T)
    sumY1  +=       0.026 * cos(1.30232025786  + 4796.85191144269 *T)
    sumY1  +=       0.026 * cos(4.70861603764  + 6226.42129253930 *T)
    sumY1  +=       0.026 * cos(5.08151018860  + 6340.21804241050 *T)
    sumY1  +=       0.029 * cos(4.90026899469  + 5974.04134481910 *T)
    sumY1  +=       0.029 * cos(4.88985723155  + 6592.59799013070 *T)
    sumY1  +=       0.026 * cos(4.96950484222  + 19402.55313533309 *T)
    sumY1  +=       0.028 * cos(3.99925126629  + 46386.74992582770 *T)
    sumY1  +=       0.032 * cos(5.04462761864  + 9225.78309076650 *T)
    sumY1  +=       0.026 * cos(4.11745014890  + 9910.33951002549 *T)
    sumY1  +=       0.025 * cos(1.02839985963  + 15265.64270181689 *T)
    sumY1  +=       0.029 * cos(4.33416022359  + 52176.05010063190 *T)
    sumY1  +=       0.027 * cos(2.65258449905  + 62883.59895699710 *T)
    sumY1  +=       0.024 * cos(5.26265995716  + 11617.21990849651 *T)
    sumY1  +=       0.024 * cos(1.04936648816  + 5120.84496306710 *T)
    sumY1  +=       0.028 * cos(4.82143927810  + 6680.98103591609 *T)
    sumY1  +=       0.024 * cos(5.97769334002  + 522.33360061030 *T)
    sumY1  +=       0.025 * cos(2.29282659231  + 6263.64990657511 *T)
    sumY1  +=       0.025 * cos(1.21411432675  + 6302.98942837470 *T)
    sumY1  +=       0.029 * cos(0.87325499796  + 23581.50199480110 *T)
    sumY1  +=       0.024 * cos(5.21948189448  + 29424.87805039950 *T)
    sumY1  +=       0.028 * cos(1.09161353868  + 16737.82105408010 *T)
    sumY1  +=       0.024 * cos(3.55960074091  + 22743.16556203289 *T)
    sumY1  +=       0.024 * cos(3.55960074091  + 22743.65319699990 *T)
    sumY1  +=       0.024 * cos(4.28222488832  + 25158.84553724890 *T)
    sumY1  +=       0.025 * cos(5.52194731109  + 21424.22282681990 *T)
    sumY1  +=       0.025 * cos(5.52194731109  + 21424.71046178690 *T)
    sumY1  +=       0.024 * cos(3.49943174674  + 14712.07329897449 *T)
    sumY1  +=       0.028 * cos(6.07681616993  + 5760.25461441409 *T)
    sumY1  +=       0.024 * cos(2.16376008059  + 31969.19247028290 *T)
    sumY1  +=       0.023 * cos(5.33584675127  + 17893.18221148710 *T)
    sumY1  +=       0.030 * cos(2.28483676839  + 8584.90548338430 *T)
    sumY1  +=       0.024 * cos(1.64666673478  + 23550.58786916610 *T)
    sumY1  +=       0.023 * cos(6.07302432574  + 6717.03140535170 *T)
    sumY1  +=       0.022 * cos(2.42662159283  + 7445.79437188270 *T)
    sumY1  +=       0.023 * cos(3.56571153376  + 5849.60792959810 *T)
    sumY1  +=       0.025 * cos(5.52794317615  + 14155.46841275010 *T)
    sumY1  +=       0.023 * cos(2.01489779401  + 6294.36536773881 *T)
    sumY1  +=       0.023 * cos(1.49204312505  + 6272.27396721100 *T)
    sumY1  +=       0.024 * cos(5.81677586075  + 18208.10612510850 *T)
    sumY1  +=       0.024 * cos(5.81677586075  + 18208.59376007550 *T)
    sumY1  +=       0.021 * cos(4.21172313283  + 22779.68106367730 *T)
    sumY1  +=       0.029 * cos(1.74369099855  + 28628.58004358310 *T)
    sumY1  +=       0.021 * cos(3.56347944199  + 40879.68432212730 *T)
    sumY1  +=       0.021 * cos(2.51496251708  + 30221.17605721590 *T)
    sumY1  +=       0.022 * cos(6.09500922463  + 4379.88285497370 *T)
    sumY1  +=       0.022 * cos(3.69512347867  + 8186.75647997610 *T)
    sumY1  +=       0.022 * cos(2.95145769488  + 28760.05469496669 *T)
    sumY1  +=       0.026 * cos(1.83973482412  + 18422.87317658170 *T)
    sumY1  +=       0.020 * cos(2.37799900039  + 12036.21691740469 *T)
    sumY1  +=       0.020 * cos(0.07485283254  + 28313.04498717750 *T)
    sumY1  +=       0.027 * cos(5.38239668902  + 34513.50689016630 *T)
    sumY1  +=       0.020 * cos(0.50551300300  + 34115.35788675810 *T)
    sumY1  +=       0.020 * cos(3.29228335956  + 21548.71855180830 *T)
    sumY1  +=       0.020 * cos(5.25656062979  + 17654.53672226610 *T)
    sumY1  +=       0.025 * cos(1.96130235408  + 55798.70217588190 *T)
    sumY1  +=       0.020 * cos(5.53200449549  + 20597.48778052470 *T)
    sumY1  +=       0.020 * cos(1.03604717996  + 18875.76968725751 *T)
    sumY1  +=       0.022 * cos(2.01853988925  + 1588.82907780029 *T)
    sumY1  +=       0.024 * cos(5.92091322542  + 9411.70843257070 *T)
    sumY1  +=       0.019 * cos(1.14706242910  + 536.56069461190 *T)
    sumY1  +=       0.020 * cos(3.42481272827  + 6187.34044025710 *T)
    sumY1  +=       0.020 * cos(0.08212819079  + 6379.29889469270 *T)
    sumY1  +=       0.022 * cos(2.87392170851  + 18326.89394936390 *T)
    sumY1  +=       0.022 * cos(0.61268580951  + 6233.56264200310 *T)
    sumY1  +=       0.022 * cos(2.89425510955  + 6333.07669294670 *T)
    sumY1  +=       0.018 * cos(0.45956662142  + 5972.47886860650 *T)
    sumY1  +=       0.018 * cos(3.04737429764  + 6594.16046634330 *T)
    sumY1  +=       0.019 * cos(2.71317971903  + 6948.07571260490 *T)
    sumY1  +=       0.019 * cos(0.79376120003  + 5618.56362234490 *T)
    sumY1  +=       0.019 * cos(4.89700264579  + 83974.07917321339 *T)
    sumY1  +=       0.019 * cos(2.72494811319  + 84020.10309797739 *T)
    sumY1  +=       0.020 * cos(5.42036659910  + 50316.95962204730 *T)
    sumY1  +=       0.020 * cos(1.28312164242  + 9070.36269133230 *T)

    # EARTH Y Coord
    sumY2  = 0
    sumY2  +=   52911.498 * cos(1.75323722235  + 6283.31966747490 *T)
    sumY2  +=    6074.441 * cos(0.48406211192  + 0.24381748350 *T)
    sumY2  +=    2406.871 * cos(2.66006394470  + 12566.39551746630 *T)
    sumY2  +=      96.033 * cos(2.57406136836  + 18849.47136745770 *T)
    sumY2  +=      29.888 * cos(2.30515825272  + 6282.83203250789 *T)
    sumY2  +=      14.021 * cos(1.64042051619  + 6279.79654912590 *T)
    sumY2  +=      13.375 * cos(1.74320372785  + 6286.84278582391 *T)
    sumY2  +=       8.148 * cos(0.62869530307  + 6309.61798727471 *T)
    sumY2  +=       8.195 * cos(2.81874475342  + 6257.02134767510 *T)
    sumY2  +=       8.001 * cos(4.34877296703  + 6127.89926804070 *T)
    sumY2  +=       7.754 * cos(5.43470598991  + 6438.74006690910 *T)
    sumY2  +=       7.223 * cos(3.81034638754  + 83997.09113559539 *T)
    sumY2  +=       4.616 * cos(4.59623594691  + 710.17686604180 *T)
    sumY2  +=       3.060 * cos(3.26884940633  + 4705.97612502710 *T)
    sumY2  +=       3.717 * cos(0.87961276723  + 5507.79705615090 *T)
    sumY2  +=       2.991 * cos(2.48913608928  + 25132.54721744910 *T)
    sumY2  +=       3.235 * cos(2.53504921673  + 7058.84227879890 *T)
    sumY2  +=       2.393 * cos(1.83696733237  + 775.76642880750 *T)
    sumY2  +=       2.034 * cos(2.77061597204  + 6290.43321447570 *T)
    sumY2  +=       2.028 * cos(0.70943948464  + 6276.20612047410 *T)
    sumY2  +=       2.064 * cos(0.01448293997  + 7860.66320992270 *T)
    sumY2  +=       1.914 * cos(2.85234479868  + 5487.02166065850 *T)
    sumY2  +=       1.770 * cos(3.48783151064  + 1059.62574767270 *T)
    sumY2  +=       1.761 * cos(1.80823220260  + 5753.62870238030 *T)
    sumY2  +=       1.558 * cos(0.56580595512  + 7079.61767429131 *T)
    sumY2  +=       1.551 * cos(0.55852743151  + 6040.59106350090 *T)
    sumY2  +=       1.436 * cos(1.28164050883  + 5885.17066406670 *T)
    sumY2  +=       1.469 * cos(2.94324921966  + 6526.04827144891 *T)
    sumY2  +=       1.319 * cos(1.89155311923  + 6681.46867088311 *T)
    sumY2  +=       1.180 * cos(1.42090963151  + 6813.01063256950 *T)
    sumY2  +=       1.337 * cos(5.75221416852  + 5729.75026463250 *T)
    sumY2  +=       1.039 * cos(4.25243426113  + 6282.33934640670 *T)
    sumY2  +=       1.313 * cos(0.59897551240  + 529.93478257810 *T)
    sumY2  +=       1.041 * cos(5.47801377680  + 6284.29998854310 *T)
    sumY2  +=       1.202 * cos(4.02829547071  + 6836.88907031730 *T)
    sumY2  +=       0.904 * cos(0.85768945907  + 5331.60126122430 *T)
    sumY2  +=       0.870 * cos(2.26660468565  + 4933.45225781610 *T)
    sumY2  +=       1.149 * cos(3.02450680654  + 167284.00540514899 *T)
    sumY2  +=       1.069 * cos(1.09203232944  + 1577.09972496430 *T)
    sumY2  +=       0.772 * cos(2.66290097432  + 7235.03807372550 *T)
    sumY2  +=       0.846 * cos(1.36798976406  + 11790.87290614230 *T)
    sumY2  +=       0.742 * cos(5.02007434299  + 4137.15425099970 *T)
    sumY2  +=       0.966 * cos(2.05227643484  + 11856.46246890800 *T)
    sumY2  +=       0.664 * cos(1.21776131888  + 7633.18707713370 *T)
    sumY2  +=       0.662 * cos(3.18968574411  + 1194.69082770810 *T)
    sumY2  +=       0.659 * cos(5.95013449282  + 11507.01358727710 *T)
    sumY2  +=       0.853 * cos(3.65338850674  + 4535.30325440790 *T)
    sumY2  +=       0.823 * cos(1.06126729936  + 155.17658195069 *T)
    sumY2  +=       0.783 * cos(4.30296275012  + 5088.87265725030 *T)
    sumY2  +=       0.785 * cos(3.46658763633  + 2353.10997125530 *T)
    sumY2  +=       0.694 * cos(3.50785804761  + 801.57711364030 *T)
    sumY2  +=       0.566 * cos(4.42922662864  + 796.54182429990 *T)
    sumY2  +=       0.684 * cos(1.11267726523  + 3128.63258257930 *T)
    sumY2  +=       0.499 * cos(1.42594865875  + 12569.91863581531 *T)
    sumY2  +=       0.596 * cos(6.09778115923  + 8031.33608054190 *T)
    sumY2  +=       0.486 * cos(4.54237345712  + 1589.31671276730 *T)
    sumY2  +=       0.483 * cos(4.68204073064  + 8429.48508395010 *T)
    sumY2  +=       0.441 * cos(4.99112466662  + 6496.61876291290 *T)
    sumY2  +=       0.412 * cos(0.87602065813  + 4693.75913722409 *T)
    sumY2  +=       0.561 * cos(5.46330573132  + 7477.76667769950 *T)
    sumY2  +=       0.390 * cos(4.63888431687  + 6070.02057203690 *T)
    sumY2  +=       0.405 * cos(6.26538093406  + 3739.00524759150 *T)
    sumY2  +=       0.501 * cos(1.82149883986  + 1592.83983111630 *T)
    sumY2  +=       0.392 * cos(0.68216396598  + 26.05450231630 *T)
    sumY2  +=       0.369 * cos(5.93410776393  + 553.32558535889 *T)
    sumY2  +=       0.483 * cos(6.07777893059  + 12036.70455237170 *T)
    sumY2  +=       0.489 * cos(3.70668603785  + 5223.45010231870 *T)
    sumY2  +=       0.481 * cos(3.17195852109  + 10213.52936369450 *T)
    sumY2  +=       0.448 * cos(5.56847234629  + 156137.71980228281 *T)
    sumY2  +=       0.321 * cos(4.53991511859  + 951.47458876710 *T)
    sumY2  +=       0.320 * cos(0.43138538654  + 1990.98883452450 *T)
    sumY2  +=       0.361 * cos(2.16599134100  + 12565.90788249930 *T)
    sumY2  +=       0.361 * cos(4.79823251497  + 12721.81591690050 *T)
    sumY2  +=       0.412 * cos(4.34259218946  + 398.39282089170 *T)
    sumY2  +=       0.304 * cos(5.68078111352  + 4690.72365384210 *T)
    sumY2  +=       0.356 * cos(6.00094576166  + 5481.49873635110 *T)
    sumY2  +=       0.283 * cos(0.66912545661  + 242.97242145751 *T)
    sumY2  +=       0.362 * cos(2.31554210740  + 9438.00675237050 *T)
    sumY2  +=       0.248 * cos(3.18243221495  + 8827.63408735830 *T)
    sumY2  +=       0.252 * cos(2.73244620528  + 3154.44326741210 *T)
    sumY2  +=       0.257 * cos(0.22372520009  + 11371.94850724170 *T)
    sumY2  +=       0.327 * cos(1.45001600721  + 14143.73905991410 *T)
    sumY2  +=       0.230 * cos(1.80447328493  + 1747.77259558350 *T)
    sumY2  +=       0.289 * cos(3.67617510107  + 7085.14059859870 *T)
    sumY2  +=       0.297 * cos(3.17045499445  + 1349.62359217529 *T)
    sumY2  +=       0.209 * cos(4.98580572759  + 11770.09751064990 *T)
    sumY2  +=       0.207 * cos(0.21623312501  + 4164.06817212950 *T)
    sumY2  +=       0.188 * cos(4.07593821324  + 7875.91568110771 *T)
    sumY2  +=       0.189 * cos(5.31784128034  + 2389.13783793270 *T)
    sumY2  +=       0.194 * cos(5.12285857348  + 10977.32262218251 *T)
    sumY2  +=       0.176 * cos(2.51491286893  + 7342.70159766410 *T)
    sumY2  +=       0.164 * cos(1.22070775546  + 5223.93773728571 *T)
    sumY2  +=       0.205 * cos(1.64290799201  + 10973.79950383350 *T)
    sumY2  +=       0.205 * cos(0.68022152849  + 13368.21644859010 *T)
    sumY2  +=       0.139 * cos(3.04789845179  + 10575.65050042530 *T)
    sumY2  +=       0.155 * cos(4.24333188820  + 12410.97511803210 *T)
    sumY2  +=       0.152 * cos(4.81739365185  + 3930.45351370310 *T)
    sumY2  +=       0.129 * cos(0.80954637051  + 397.90518592470 *T)
    sumY2  +=       0.127 * cos(5.67893828134  + 12592.69383726610 *T)
    sumY2  +=       0.132 * cos(4.87291664360  + 8636.18582124671 *T)
    sumY2  +=       0.122 * cos(3.21278850975  + 13119.96492030870 *T)
    sumY2  +=       0.163 * cos(2.50359725386  + 5507.30942118390 *T)
    sumY2  +=       0.114 * cos(4.13352060999  + 17260.39847217390 *T)
    sumY2  +=       0.118 * cos(0.80775184774  + 4292.57465043390 *T)
    sumY2  +=       0.148 * cos(4.06616105867  + 17790.08943726851 *T)
    sumY2  +=       0.123 * cos(2.18706277371  + 12562.87239911730 *T)
    sumY2  +=       0.103 * cos(1.79915109205  + 13518.11392371690 *T)
    sumY2  +=       0.138 * cos(2.26785827403  + 12168.24651405810 *T)
    sumY2  +=       0.107 * cos(2.76898193474  + 10989.05197501850 *T)
    sumY2  +=       0.104 * cos(0.56758990747  + 2145.92159899169 *T)
    sumY2  +=       0.096 * cos(0.40713285174  + 6283.25235717211 *T)
    sumY2  +=       0.099 * cos(0.70751931715  + 13096.08648256090 *T)
    sumY2  +=       0.096 * cos(6.24621518908  + 6283.38697777770 *T)
    sumY2  +=       0.084 * cos(3.63516988010  + 16730.70750707931 *T)
    sumY2  +=       0.111 * cos(2.39242858566  + 12540.09719766650 *T)
    sumY2  +=       0.087 * cos(4.45435464785  + 10177.50149701710 *T)
    sumY2  +=       0.078 * cos(2.61342989032  + 12323.66691349230 *T)
    sumY2  +=       0.076 * cos(4.13782411870  + 6709.91785835091 *T)
    sumY2  +=       0.105 * cos(1.96765772501  + 5642.44206009270 *T)
    sumY2  +=       0.073 * cos(5.60932378402  + 5856.72147659890 *T)
    sumY2  +=       0.102 * cos(5.37997051996  + 250570.91967470257 *T)
    sumY2  +=       0.070 * cos(0.51266920771  + 5753.14106741329 *T)
    sumY2  +=       0.071 * cos(5.27404987073  + 14314.41193053330 *T)
    sumY2  +=       0.092 * cos(0.39406209616  + 13916.26292712510 *T)
    sumY2  +=       0.070 * cos(4.10529685600  + 7084.65296363170 *T)
    sumY2  +=       0.091 * cos(1.53880932897  + 6924.19727485710 *T)
    sumY2  +=       0.083 * cos(2.41910116753  + 31415.62306744050 *T)
    sumY2  +=       0.088 * cos(0.30300511757  + 71430.45180064559 *T)
    sumY2  +=       0.074 * cos(3.53618023271  + 529.44714761109 *T)
    sumY2  +=       0.085 * cos(4.40773297118  + 95143.37673846159 *T)
    sumY2  +=       0.061 * cos(2.66034748981  + 8274.06468451590 *T)
    sumY2  +=       0.060 * cos(3.95075537336  + 16496.60521368590 *T)
    sumY2  +=       0.058 * cos(3.04970743906  + 5642.93474619389 *T)
    sumY2  +=       0.078 * cos(0.84061745785  + 213.54291292150 *T)
    sumY2  +=       0.059 * cos(3.10446906052  + 640.14110379750 *T)
    sumY2  +=       0.057 * cos(3.63209087311  + 4690.23601887509 *T)
    sumY2  +=       0.056 * cos(2.23091315346  + 13341.91812879030 *T)
    sumY2  +=       0.065 * cos(4.95952410600  + 4731.78680985990 *T)
    sumY2  +=       0.051 * cos(0.09929724563  + 3893.93801205869 *T)
    sumY2  +=       0.050 * cos(2.60290043400  + 12573.50906446710 *T)
    sumY2  +=       0.050 * cos(5.98749446465  + 3634.37720703489 *T)
    sumY2  +=       0.049 * cos(0.70632876133  + 18319.78040236310 *T)
    sumY2  +=       0.066 * cos(3.92505951821  + 2787.28684134090 *T)
    sumY2  +=       0.049 * cos(0.42825389697  + 13362.69352428270 *T)
    sumY2  +=       0.048 * cos(2.19153498916  + 12964.54452087450 *T)
    sumY2  +=       0.053 * cos(2.55294637714  + 14919.26167123810 *T)
    sumY2  +=       0.058 * cos(5.94322042199  + 1577.58735993130 *T)
    sumY2  +=       0.052 * cos(5.99067773035  + 15721.08260236190 *T)
    sumY2  +=       0.061 * cos(4.99956015956  + 4292.08701546690 *T)
    sumY2  +=       0.044 * cos(5.02285128905  + 24073.16528725990 *T)
    sumY2  +=       0.044 * cos(1.96034093811  + 10447.14402212089 *T)
    sumY2  +=       0.042 * cos(2.52906622396  + 20426.81490990550 *T)
    sumY2  +=       0.059 * cos(2.58678833053  + 84672.23202702119 *T)
    sumY2  +=       0.051 * cos(3.81200106179  + 6370.62787201471 *T)
    sumY2  +=       0.051 * cos(5.97812516445  + 6196.01146293510 *T)
    sumY2  +=       0.046 * cos(5.23410298912  + 2119.00767786191 *T)
    sumY2  +=       0.041 * cos(2.52248241350  + 3185.43584474911 *T)
    sumY2  +=       0.056 * cos(4.84268946755  + 10976.83498721549 *T)
    sumY2  +=       0.051 * cos(3.94604483748  + 9437.51911740350 *T)
    sumY2  +=       0.049 * cos(1.64079935270  + 239424.63407183639 *T)
    sumY2  +=       0.036 * cos(1.49154565635  + 3495.78900865049 *T)
    sumY2  +=       0.044 * cos(2.12316916648  + 18139.53831889940 *T)
    sumY2  +=       0.035 * cos(5.18466509768  + 3929.96587873609 *T)
    sumY2  +=       0.044 * cos(3.23094444076  + 8661.99650607950 *T)
    sumY2  +=       0.046 * cos(2.21535532938  + 3894.42564702570 *T)
    sumY2  +=       0.034 * cos(2.24364878116  + 6262.54427198250 *T)
    sumY2  +=       0.034 * cos(4.40970135798  + 6304.09506296730 *T)
    sumY2  +=       0.035 * cos(5.15488505358  + 18875.28205229050 *T)
    sumY2  +=       0.035 * cos(4.51996316476  + 10420.23010099111 *T)
    sumY2  +=       0.032 * cos(5.33774590597  + 23013.78335707070 *T)
    sumY2  +=       0.034 * cos(2.04067717162  + 7.35736448430 *T)
    sumY2  +=       0.045 * cos(6.20022544523  + 3340.85624418330 *T)
    sumY2  +=       0.035 * cos(1.95669709650  + 10213.04172872750 *T)
    sumY2  +=       0.031 * cos(3.81393239996  + 14712.56093394150 *T)
    sumY2  +=       0.030 * cos(2.82636067861  + 12809.12412144031 *T)
    sumY2  +=       0.039 * cos(5.86021469295  + 9779.35249360890 *T)
    sumY2  +=       0.041 * cos(0.90085517811  + 82576.73740351178 *T)
    sumY2  +=       0.031 * cos(5.47911308336  + 22779.68106367730 *T)
    sumY2  +=       0.028 * cos(2.50895957020  + 2942.70724077510 *T)
    sumY2  +=       0.028 * cos(3.93140481871  + 5429.63565075589 *T)
    sumY2  +=       0.027 * cos(5.61355223000  + 11014.86265985130 *T)
    sumY2  +=       0.027 * cos(0.53944014317  + 9623.93209417470 *T)
    sumY2  +=       0.025 * cos(0.15499921281  + 18209.57408114370 *T)
    sumY2  +=       0.026 * cos(0.97984555070  + 9381.20349020070 *T)
    sumY2  +=       0.025 * cos(1.11733310909  + 3583.58484815730 *T)
    sumY2  +=       0.030 * cos(4.41826131323  + 10447.63165708790 *T)
    sumY2  +=       0.032 * cos(5.65959938774  + 796.05418933290 *T)
    sumY2  +=       0.023 * cos(1.38575591729  + 9917.45305702629 *T)
    sumY2  +=       0.026 * cos(5.61732376806  + 12012.82611462390 *T)
    sumY2  +=       0.024 * cos(3.45407340381  + 14143.25142494710 *T)
    sumY2  +=       0.023 * cos(0.69647593831  + 26709.89075989690 *T)
    sumY2  +=       0.021 * cos(5.43974830479  + 7632.69944216670 *T)
    sumY2  +=       0.021 * cos(0.80058820514  + 17256.87535382490 *T)
    sumY2  +=       0.028 * cos(5.78689433830  + 12353.09642202830 *T)
    sumY2  +=       0.022 * cos(3.26384247748  + 77714.01528560399 *T)
    sumY2  +=       0.021 * cos(4.41664687493  + 6993.25271603320 *T)
    sumY2  +=       0.020 * cos(1.88345295944  + 6836.40143535029 *T)
    sumY2  +=       0.020 * cos(3.87252932348  + 1748.26023055050 *T)
    sumY2  +=       0.026 * cos(0.45595739946  + 9225.78309076650 *T)
    sumY2  +=       0.023 * cos(5.52071600644  + 149854.64395229140 *T)
    sumY2  +=       0.026 * cos(0.60119736394  + 2544.07060239989 *T)
    sumY2  +=       0.018 * cos(0.71176167314  + 11614.67711121570 *T)
    sumY2  +=       0.020 * cos(3.04333652471  + 426.84200835950 *T)
    sumY2  +=       0.019 * cos(2.09021029647  + 30640.10045611650 *T)
    sumY2  +=       0.018 * cos(3.20715016441  + 155.66421691770 *T)
    sumY2  +=       0.018 * cos(0.32502884515  + 11926.49823115230 *T)
    sumY2  +=       0.022 * cos(2.17863705978  + 17298.42614480970 *T)
    sumY2  +=       0.019 * cos(4.82590950931  + 18073.46112116669 *T)
    sumY2  +=       0.020 * cos(1.02950628439  + 19651.29229858150 *T)
    sumY2  +=       0.024 * cos(3.14467908991  + 5856.23384163189 *T)
    sumY2  +=       0.023 * cos(3.21288696705  + 72850.80553272919 *T)
    sumY2  +=       0.017 * cos(2.38623798159  + 8983.05448679250 *T)
    sumY2  +=       0.016 * cos(3.59989563011  + 16460.57734700850 *T)
    sumY2  +=       0.016 * cos(2.77179277240  + 17259.91083720689 *T)
    sumY2  +=       0.021 * cos(2.20286926080  + 16858.72635041670 *T)
    sumY2  +=       0.020 * cos(5.10666464640  + 23543.47432216530 *T)
    sumY2  +=       0.021 * cos(5.36525328799  + 13367.72881362310 *T)
    sumY2  +=       0.020 * cos(2.47582665567  + 2146.40923395870 *T)
    sumY2  +=       0.014 * cos(4.85508264889  + 22484.09239197610 *T)
    sumY2  +=       0.014 * cos(6.23679853613  + 16200.52890701769 *T)
    sumY2  +=       0.015 * cos(2.64768344463  + 20199.33877711650 *T)
    sumY2  +=       0.019 * cos(1.24340851761  + 8672.21368792411 *T)
    sumY2  +=       0.019 * cos(2.89835524904  + 18073.94875613370 *T)
    sumY2  +=       0.013 * cos(3.14350306699  + 16201.01654198470 *T)
    sumY2  +=       0.017 * cos(2.96004801328  + 7860.17557495569 *T)
    sumY2  +=       0.013 * cos(3.12347768998  + 18451.32236404950 *T)
    sumY2  +=       0.016 * cos(4.90276494200  + 5572.89898394960 *T)
    sumY2  +=       0.012 * cos(1.13909733386  + 17996.27498570570 *T)
    sumY2  +=       0.012 * cos(5.43460350187  + 19403.04077030010 *T)
    sumY2  +=       0.012 * cos(0.82655936469  + 14945.07235607090 *T)
    sumY2  +=       0.015 * cos(0.47799392266  + 12559.28197046550 *T)
    sumY2  +=       0.012 * cos(2.80799572996  + 10818.37910439930 *T)
    sumY2  +=       0.015 * cos(5.28864619208  + 12567.37583853451 *T)
    sumY2  +=       0.011 * cos(0.21391675163  + 28767.16824196750 *T)
    sumY2  +=       0.013 * cos(5.80786469706  + 90280.16698558679 *T)
    sumY2  +=       0.011 * cos(0.30991526634  + 22805.49174851010 *T)
    sumY2  +=       0.011 * cos(0.67526353010  + 35372.13108345990 *T)
    sumY2  +=       0.010 * cos(1.54770241799  + 11506.52595231009 *T)
    sumY2  +=       0.014 * cos(0.36722951229  + 21228.63584102931 *T)
    sumY2  +=       0.010 * cos(4.84303300061  + 12779.69461290430 *T)
    sumY2  +=       0.010 * cos(4.48229440272  + 23141.80220040810 *T)
    sumY2  +=       0.010 * cos(5.89241784221  + 1350.11122714230 *T)
    sumY2  +=       0.010 * cos(4.07842068364  + 10575.16286545830 *T)
    sumY2  +=       0.010 * cos(0.98564127078  + 22345.50419359170 *T)
    sumY2  +=       0.010 * cos(0.62034181586  + 9778.86485864190 *T)
    sumY2  +=       0.009 * cos(6.20390809264  + 29826.55017215670 *T)
    sumY2  +=       0.010 * cos(2.66921279687  + 23581.50199480110 *T)
    sumY2  +=       0.011 * cos(5.10849785178  + 25158.84553724890 *T)

    # EARTH Y Coord
    sumY3  = 0
    sumY3  +=     232.790 * cos(1.83555296287  + 0.24381748350 *T)
    sumY3  +=      76.843 * cos(0.95359770708  + 12566.39551746630 *T)
    sumY3  +=      35.331 * cos(1.77537067174  + 6283.31966747490 *T)
    sumY3  +=       5.282 * cos(0.83410185910  + 18849.47136745770 *T)
    sumY3  +=       1.631 * cos(0.72620862558  + 6282.83203250789 *T)
    sumY3  +=       1.483 * cos(0.65599216587  + 6438.74006690910 *T)
    sumY3  +=       1.479 * cos(2.83072127404  + 6127.89926804070 *T)
    sumY3  +=       0.652 * cos(3.13796451313  + 6279.79654912590 *T)
    sumY3  +=       0.656 * cos(0.01615422945  + 6286.84278582391 *T)
    sumY3  +=       0.317 * cos(1.43157695840  + 6526.04827144891 *T)
    sumY3  +=       0.318 * cos(2.05944515679  + 6040.59106350090 *T)
    sumY3  +=       0.227 * cos(0.70881980797  + 25132.54721744910 *T)
    sumY3  +=       0.201 * cos(5.52615570373  + 6836.88907031730 *T)
    sumY3  +=       0.201 * cos(4.23563267830  + 5729.75026463250 *T)
    sumY3  +=       0.036 * cos(6.05003898151  + 12569.91863581531 *T)
    sumY3  +=       0.034 * cos(0.36912822345  + 4705.97612502710 *T)
    sumY3  +=       0.033 * cos(2.65665718326  + 12410.97511803210 *T)
    sumY3  +=       0.032 * cos(3.81269087187  + 6257.02134767510 *T)
    sumY3  +=       0.032 * cos(2.84133186500  + 6309.61798727471 *T)
    sumY3  +=       0.026 * cos(5.21573736647  + 775.76642880750 *T)
    sumY3  +=       0.023 * cos(0.44696909593  + 1059.62574767270 *T)
    sumY3  +=       0.025 * cos(0.01326357360  + 7860.66320992270 *T)
    sumY3  +=       0.027 * cos(0.46730161827  + 12565.90788249930 *T)
    sumY3  +=       0.020 * cos(2.19213288615  + 5753.62870238030 *T)
    sumY3  +=       0.019 * cos(1.44435756454  + 5885.17066406670 *T)
    sumY3  +=       0.019 * cos(4.47976718197  + 6813.01063256950 *T)
    sumY3  +=       0.018 * cos(0.72211012099  + 12721.81591690050 *T)
    sumY3  +=       0.017 * cos(5.20474716177  + 6681.46867088311 *T)
    sumY3  +=       0.013 * cos(2.92478174105  + 5487.02166065850 *T)
    sumY3  +=       0.011 * cos(3.72062261949  + 7079.61767429131 *T)
    sumY3  +=       0.010 * cos(2.89167021825  + 5507.79705615090 *T)
    sumY3  +=       0.010 * cos(1.46762801985  + 11790.87290614230 *T)
    sumY3  +=       0.009 * cos(6.21784065295  + 11507.01358727710 *T)
    sumY3  +=       0.011 * cos(4.68801103302  + 12592.69383726610 *T)
    sumY3  +=       0.009 * cos(3.82179163254  + 7058.84227879890 *T)
    sumY3  +=       0.008 * cos(5.97770881958  + 6290.43321447570 *T)
    sumY3  +=       0.008 * cos(0.68313431813  + 6276.20612047410 *T)
    sumY3  +=       0.008 * cos(1.45990791265  + 796.54182429990 *T)
    sumY3  +=       0.007 * cos(4.15465801739  + 4693.75913722409 *T)
    sumY3  +=       0.007 * cos(0.35078036011  + 7.35736448430 *T)
    sumY3  +=       0.005 * cos(3.29483739022  + 3739.00524759150 *T)
    sumY3  +=       0.005 * cos(4.64160679165  + 6070.02057203690 *T)
    sumY3  +=       0.006 * cos(2.38483297622  + 6284.29998854310 *T)
    sumY3  +=       0.006 * cos(4.26851560652  + 6282.33934640670 *T)
    sumY3  +=       0.005 * cos(1.89668032657  + 4137.15425099970 *T)
    sumY3  +=       0.005 * cos(2.01213704309  + 6496.61876291290 *T)

    # EARTH Y Coord
    sumY4  = 0
    sumY4  +=     114.918 * cos(4.77291921544  + 6283.31966747490 *T)
    sumY4  +=       6.817 * cos(2.90545031303  + 0.24381748350 *T)
    sumY4  +=       3.158 * cos(5.19149443854  + 12566.39551746630 *T)
    sumY4  +=       0.211 * cos(2.15272014034  + 6438.74006690910 *T)
    sumY4  +=       0.210 * cos(1.31410288968  + 6127.89926804070 *T)
    sumY4  +=       0.253 * cos(5.35904773722  + 18849.47136745770 *T)
    sumY4  +=       0.073 * cos(5.49456429263  + 6282.83203250789 *T)
    sumY4  +=       0.042 * cos(3.03243163257  + 710.17686604180 *T)
    sumY4  +=       0.023 * cos(4.47433896564  + 6279.79654912590 *T)
    sumY4  +=       0.022 * cos(4.47061131203  + 6286.84278582391 *T)
    sumY4  +=       0.012 * cos(5.22878976906  + 25132.54721744910 *T)
    sumY4  +=       0.009 * cos(2.63928422881  + 83997.09113559539 *T)
    sumY4  +=       0.009 * cos(3.62312855703  + 11856.46246890800 *T)
    sumY4  +=       0.008 * cos(4.59507616558  + 167284.00540514899 *T)
    sumY4  +=       0.005 * cos(4.40834364231  + 6257.02134767510 *T)
    sumY4  +=       0.005 * cos(2.21798830782  + 6309.61798727471 *T)
    sumY4  +=       0.005 * cos(1.11835906029  + 12410.97511803210 *T)
    sumY4  +=       0.005 * cos(3.99770824807  + 156137.71980228281 *T)
    sumY4  +=       0.004 * cos(4.97197978751  + 12565.90788249930 *T)
    sumY4  +=       0.003 * cos(3.60141861387  + 529.93478257810 *T)

    # EARTH Y Coord
    sumY5  = 0
    sumY5  +=       0.877 * cos(4.87375194107  + 6283.31966747490 *T)
    sumY5  +=       0.305 * cos(4.05893642248  + 0.24381748350 *T)
    sumY5  +=       0.101 * cos(3.30997066470  + 12566.39551746630 *T)
    sumY5  +=       0.025 * cos(3.73514325706  + 6438.74006690910 *T)
    sumY5  +=       0.025 * cos(5.99118011879  + 6127.89926804070 *T)
    sumY5  +=       0.010 * cos(3.54319653898  + 18849.47136745770 *T)
    sumY5  +=       0.003 * cos(4.13285368647  + 6282.83203250789 *T)

    # EARTH Z Coord
    sumZ0  = 0
    sumZ0  +=     279.620 * cos(3.19870156017  + 84334.66158130829 *T)
    sumZ0  +=     101.625 * cos(5.42248110597  + 5507.55323866740 *T)
    sumZ0  +=      80.461 * cos(3.88027157914  + 5223.69391980220 *T)
    sumZ0  +=      43.855 * cos(3.70430347723  + 2352.86615377180 *T)
    sumZ0  +=      31.853 * cos(3.99997840986  + 1577.34354244780 *T)
    sumZ0  +=      22.724 * cos(3.98473831560  + 1047.74731175470 *T)
    sumZ0  +=      16.395 * cos(3.56456162523  + 5856.47765911540 *T)
    sumZ0  +=      18.150 * cos(4.98479613321  + 6283.07584999140 *T)
    sumZ0  +=      14.477 * cos(3.70258423465  + 9437.76293488700 *T)
    sumZ0  +=      14.325 * cos(3.41020246931  + 10213.28554621100 *T)
    sumZ0  +=      11.204 * cos(4.83021499098  + 14143.49524243060 *T)
    sumZ0  +=      10.896 * cos(2.08023708321  + 6812.76681508600 *T)
    sumZ0  +=       9.716 * cos(3.47560205064  + 4694.00295470760 *T)
    sumZ0  +=      10.363 * cos(4.05663928093  + 71092.88135493269 *T)
    sumZ0  +=       8.792 * cos(4.44890729280  + 5753.38488489680 *T)
    sumZ0  +=       8.337 * cos(4.99167706048  + 7084.89678111520 *T)
    sumZ0  +=       6.921 * cos(4.32559054073  + 6275.96230299060 *T)
    sumZ0  +=       9.145 * cos(1.14182646613  + 6620.89011318780 *T)
    sumZ0  +=       7.205 * cos(3.62441599378  + 529.69096509460 *T)
    sumZ0  +=       7.698 * cos(5.55425745881  + 167621.57585086187 *T)
    sumZ0  +=       5.186 * cos(6.25384533126  + 18073.70493865020 *T)
    sumZ0  +=       5.027 * cos(2.49727910749  + 4705.73230754360 *T)
    sumZ0  +=       4.529 * cos(2.33827747356  + 6309.37416979120 *T)
    sumZ0  +=       4.754 * cos(0.71100829534  + 5884.92684658320 *T)
    sumZ0  +=       4.292 * cos(1.10034019036  + 6681.22485339960 *T)
    sumZ0  +=       3.855 * cos(1.82331477760  + 5486.77784317500 *T)
    sumZ0  +=       5.256 * cos(4.42445744523  + 7860.41939243920 *T)
    sumZ0  +=       4.025 * cos(5.11990285618  + 13367.97263110660 *T)
    sumZ0  +=       4.061 * cos(6.02923989849  + 3930.20969621960 *T)
    sumZ0  +=       3.798 * cos(0.44370219832  + 3154.68708489560 *T)
    sumZ0  +=       2.933 * cos(5.12428135363  + 1059.38193018920 *T)
    sumZ0  +=       3.424 * cos(5.42065191700  + 6069.77675455340 *T)
    sumZ0  +=       3.670 * cos(4.58210192227  + 12194.03291462090 *T)
    sumZ0  +=       2.912 * cos(1.92688087702  + 10977.07880469900 *T)
    sumZ0  +=       2.506 * cos(0.60484952637  + 6496.37494542940 *T)
    sumZ0  +=       2.680 * cos(1.39497359287  + 22003.91463486980 *T)
    sumZ0  +=       2.120 * cos(4.30691000285  + 5643.17856367740 *T)
    sumZ0  +=       2.235 * cos(0.81363184041  + 8635.94200376320 *T)
    sumZ0  +=       1.819 * cos(3.75748003107  + 3340.61242669980 *T)
    sumZ0  +=       2.230 * cos(2.76407822446  + 12036.46073488820 *T)
    sumZ0  +=       1.945 * cos(5.69949789177  + 11790.62908865880 *T)
    sumZ0  +=       1.529 * cos(1.98639348122  + 398.14900340820 *T)
    sumZ0  +=       1.581 * cos(3.19976230948  + 5088.62883976680 *T)
    sumZ0  +=       1.424 * cos(6.26116472313  + 2544.31441988340 *T)
    sumZ0  +=       1.405 * cos(4.69394873481  + 7058.59846131540 *T)
    sumZ0  +=       1.588 * cos(0.25704624784  + 17298.18232732620 *T)
    sumZ0  +=       1.478 * cos(2.81808207569  + 25934.12433108940 *T)
    sumZ0  +=       1.693 * cos(4.95689385293  + 156475.29024799570 *T)
    sumZ0  +=       1.123 * cos(2.38605285936  + 3738.76143010800 *T)
    sumZ0  +=       0.990 * cos(4.29600699156  + 9225.53927328300 *T)
    sumZ0  +=       0.923 * cos(3.07195736431  + 4164.31198961300 *T)
    sumZ0  +=       0.863 * cos(0.55119493097  + 8429.24126646660 *T)
    sumZ0  +=       0.984 * cos(5.96784225392  + 7079.37385680780 *T)
    sumZ0  +=       1.086 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ0  +=       0.824 * cos(1.50050055147  + 10447.38783960440 *T)
    sumZ0  +=       0.829 * cos(3.51336978167  + 11506.76976979360 *T)
    sumZ0  +=       0.908 * cos(0.12102885474  + 11015.10647733480 *T)
    sumZ0  +=       0.742 * cos(1.99159139281  + 26087.90314157420 *T)
    sumZ0  +=       0.850 * cos(4.24120016095  + 29864.33402730900 *T)
    sumZ0  +=       0.762 * cos(2.90028034892  + 4732.03062734340 *T)
    sumZ0  +=       0.711 * cos(1.38596151724  + 2146.16541647520 *T)
    sumZ0  +=       0.766 * cos(0.61397570576  + 796.29800681640 *T)
    sumZ0  +=       0.708 * cos(1.91406542362  + 8031.09226305840 *T)
    sumZ0  +=       0.810 * cos(5.13961498599  + 2942.46342329160 *T)
    sumZ0  +=       0.751 * cos(1.67479850166  + 21228.39202354580 *T)
    sumZ0  +=       0.589 * cos(2.02193316047  + 775.52261132400 *T)
    sumZ0  +=       0.513 * cos(2.60734651128  + 12566.15169998280 *T)
    sumZ0  +=       0.550 * cos(1.61119298963  + 801.82093112380 *T)
    sumZ0  +=       0.592 * cos(4.59481504319  + 4690.47983635860 *T)
    sumZ0  +=       0.519 * cos(5.72925039114  + 8827.39026987480 *T)
    sumZ0  +=       0.515 * cos(4.09976487552  + 64809.80550494129 *T)
    sumZ0  +=       0.503 * cos(5.66433137112  + 33794.54372352860 *T)
    sumZ0  +=       0.476 * cos(3.69640472448  + 213.29909543800 *T)
    sumZ0  +=       0.463 * cos(1.24987240483  + 15720.83878487840 *T)
    sumZ0  +=       0.411 * cos(0.10769444525  + 3128.38876509580 *T)
    sumZ0  +=       0.504 * cos(3.26207669160  + 7632.94325965020 *T)
    sumZ0  +=       0.552 * cos(1.02926440457  + 239762.20451754928 *T)
    sumZ0  +=       0.394 * cos(5.81496021156  + 426.59819087600 *T)
    sumZ0  +=       0.476 * cos(3.52810472421  + 16496.36139620240 *T)
    sumZ0  +=       0.503 * cos(4.85802444134  + 6290.18939699220 *T)
    sumZ0  +=       0.417 * cos(0.81920713533  + 5216.58037280140 *T)
    sumZ0  +=       0.363 * cos(5.70015720452  + 1589.07289528380 *T)
    sumZ0  +=       0.363 * cos(1.28376436579  + 6206.80977871580 *T)
    sumZ0  +=       0.353 * cos(4.70059133110  + 7234.79425624200 *T)
    sumZ0  +=       0.381 * cos(2.61890749829  + 7342.45778018060 *T)
    sumZ0  +=       0.420 * cos(0.97821135132  + 4136.91043351620 *T)
    sumZ0  +=       0.335 * cos(6.22057967320  + 12168.00269657460 *T)
    sumZ0  +=       0.387 * cos(3.09145061418  + 25158.60171976540 *T)
    sumZ0  +=       0.349 * cos(2.94840720073  + 9623.68827669120 *T)
    sumZ0  +=       0.418 * cos(3.75759994446  + 5230.80746680300 *T)
    sumZ0  +=       0.396 * cos(1.22507712354  + 6438.49624942560 *T)
    sumZ0  +=       0.322 * cos(1.21162178805  + 8662.24032356300 *T)
    sumZ0  +=       0.323 * cos(5.50254808471  + 1194.44701022460 *T)
    sumZ0  +=       0.372 * cos(1.72256107938  + 14945.31617355440 *T)
    sumZ0  +=       0.305 * cos(0.80429352049  + 37724.75341974820 *T)
    sumZ0  +=       0.265 * cos(6.10358507671  + 6836.64525283380 *T)
    sumZ0  +=       0.250 * cos(4.56452895547  + 7477.52286021600 *T)
    sumZ0  +=       0.266 * cos(2.62926282354  + 7238.67559160000 *T)
    sumZ0  +=       0.262 * cos(1.34297269105  + 11769.85369316640 *T)
    sumZ0  +=       0.263 * cos(6.22089501237  + 6133.51265285680 *T)
    sumZ0  +=       0.302 * cos(2.81139045072  + 1748.01641306700 *T)
    sumZ0  +=       0.316 * cos(1.62662805006  + 250908.49012041549 *T)
    sumZ0  +=       0.226 * cos(2.42346415873  + 11371.70468975820 *T)
    sumZ0  +=       0.216 * cos(3.68721275185  + 5849.36411211460 *T)
    sumZ0  +=       0.276 * cos(2.33399767923  + 19651.04848109800 *T)
    sumZ0  +=       0.230 * cos(0.36165162947  + 5863.59120611620 *T)
    sumZ0  +=       0.200 * cos(5.86073159059  + 4535.05943692440 *T)
    sumZ0  +=       0.277 * cos(4.65400292395  + 82239.16695779889 *T)
    sumZ0  +=       0.199 * cos(5.05186622555  + 5429.87946823940 *T)
    sumZ0  +=       0.206 * cos(3.70556027982  + 10973.55568635000 *T)
    sumZ0  +=       0.210 * cos(4.50691909144  + 29088.81141598500 *T)
    sumZ0  +=       0.181 * cos(6.00294783127  + 4292.33083295040 *T)
    sumZ0  +=       0.249 * cos(0.12900984422  + 154379.79562448629 *T)
    sumZ0  +=       0.167 * cos(6.27348370813  + 10988.80815753500 *T)
    sumZ0  +=       0.183 * cos(3.19836584401  + 16730.46368959580 *T)
    sumZ0  +=       0.225 * cos(3.18339652605  + 18875.52586977400 *T)
    sumZ0  +=       0.192 * cos(0.87192032514  + 77375.95720492408 *T)
    sumZ0  +=       0.188 * cos(2.22746128596  + 41654.96311596780 *T)
    sumZ0  +=       0.160 * cos(5.20207795189  + 5481.25491886760 *T)
    sumZ0  +=       0.192 * cos(3.89678943865  + 17789.84561978500 *T)
    sumZ0  +=       0.159 * cos(5.32590846475  + 7.11354700080 *T)
    sumZ0  +=       0.214 * cos(4.08171403539  + 337.81426319640 *T)
    sumZ0  +=       0.184 * cos(3.75465072369  + 23581.25817731760 *T)
    sumZ0  +=       0.160 * cos(4.53897529970  + 18477.10876461230 *T)
    sumZ0  +=       0.144 * cos(5.19486374726  + 20426.57109242200 *T)
    sumZ0  +=       0.126 * cos(1.69446958184  + 15110.46611986620 *T)
    sumZ0  +=       0.149 * cos(1.55306832217  + 10021.83728009940 *T)
    sumZ0  +=       0.127 * cos(5.31068233515  + 639.89728631400 *T)
    sumZ0  +=       0.118 * cos(2.73268114062  + 18849.22754997420 *T)
    sumZ0  +=       0.123 * cos(2.55361087146  + 12592.45001978260 *T)
    sumZ0  +=       0.113 * cos(6.07178776726  + 6709.67404086740 *T)
    sumZ0  +=       0.130 * cos(3.94161517411  + 7330.82316174610 *T)
    sumZ0  +=       0.110 * cos(3.51203379263  + 18052.92954315780 *T)
    sumZ0  +=       0.147 * cos(4.63371971408  + 22805.73556599360 *T)
    sumZ0  +=       0.113 * cos(1.20711498905  + 14919.01785375460 *T)
    sumZ0  +=       0.108 * cos(3.76359880753  + 14314.16811304980 *T)
    sumZ0  +=       0.148 * cos(0.65447253687  + 95480.94718417450 *T)
    sumZ0  +=       0.121 * cos(3.13748782783  + 14712.31711645800 *T)
    sumZ0  +=       0.119 * cos(5.92110458985  + 33019.02111220460 *T)
    sumZ0  +=       0.124 * cos(2.93638421840  + 32217.20018108080 *T)
    sumZ0  +=       0.117 * cos(3.65065271640  + 45585.17281218740 *T)
    sumZ0  +=       0.127 * cos(4.74596574209  + 49515.38250840700 *T)
    sumZ0  +=       0.120 * cos(1.04211499785  + 6915.85958930460 *T)
    sumZ0  +=       0.120 * cos(5.60638811846  + 5650.29211067820 *T)
    sumZ0  +=       0.127 * cos(2.70766997307  + 12352.85260454480 *T)
    sumZ0  +=       0.113 * cos(5.39691616452  + 5235.32853823670 *T)
    sumZ0  +=       0.097 * cos(1.07908724794  + 9917.69687450980 *T)
    sumZ0  +=       0.123 * cos(5.15992792934  + 27511.46787353720 *T)
    sumZ0  +=       0.099 * cos(4.45774681732  + 4933.20844033260 *T)
    sumZ0  +=       0.087 * cos(3.93637812950  + 17654.78053974960 *T)
    sumZ0  +=       0.122 * cos(2.23956068680  + 83997.09113559539 *T)
    sumZ0  +=       0.102 * cos(3.97386522171  + 10818.13528691580 *T)
    sumZ0  +=       0.084 * cos(4.21241958885  + 22779.43724619380 *T)
    sumZ0  +=       0.103 * cos(4.63519841733  + 1349.86740965880 *T)
    sumZ0  +=       0.081 * cos(4.01655240430  + 1592.59601363280 *T)
    sumZ0  +=       0.092 * cos(3.24182657569  + 78051.58573131690 *T)
    sumZ0  +=       0.094 * cos(4.33873972800  + 36147.40987730040 *T)
    sumZ0  +=       0.076 * cos(2.83897554420  + 17260.15465469040 *T)
    sumZ0  +=       0.100 * cos(6.07733097102  + 26735.94526221320 *T)
    sumZ0  +=       0.078 * cos(0.72462427283  + 12779.45079542080 *T)
    sumZ0  +=       0.086 * cos(5.26029638250  + 28313.28880466100 *T)
    sumZ0  +=       0.093 * cos(4.31900620254  + 44809.65020086340 *T)
    sumZ0  +=       0.076 * cos(6.22743405935  + 13521.75144159140 *T)
    sumZ0  +=       0.081 * cos(2.16420552322  + 13095.84266507740 *T)
    sumZ0  +=       0.100 * cos(1.38002787119  + 28286.99048486120 *T)
    sumZ0  +=       0.072 * cos(1.55820597747  + 6256.77753019160 *T)
    sumZ0  +=       0.082 * cos(4.95202664555  + 10575.40668294180 *T)
    sumZ0  +=       0.082 * cos(1.69647647075  + 1990.74501704100 *T)
    sumZ0  +=       0.096 * cos(6.16038106485  + 24356.78078864160 *T)
    sumZ0  +=       0.075 * cos(2.29836095644  + 3634.62102451840 *T)
    sumZ0  +=       0.075 * cos(2.66367876557  + 16200.77272450120 *T)
    sumZ0  +=       0.087 * cos(0.26630214764  + 31441.67756975680 *T)
    sumZ0  +=       0.080 * cos(5.00001604436  + 150192.21439800429 *T)
    sumZ0  +=       0.079 * cos(0.01398391548  + 90617.73743129970 *T)
    sumZ0  +=       0.054 * cos(5.59738773448  + 161338.50000087050 *T)
    sumZ0  +=       0.056 * cos(2.60133794851  + 73188.37597844210 *T)
    sumZ0  +=       0.055 * cos(5.81483150022  + 143233.51002162008 *T)
    sumZ0  +=       0.054 * cos(3.38482031504  + 323049.11878710287 *T)

    # EARTH Z Coord
    sumZ1  = 0
    sumZ1  +=       9.031 * cos(3.89751156799  + 5507.55323866740 *T)
    sumZ1  +=       6.179 * cos(1.73051337995  + 5223.69391980220 *T)
    sumZ1  +=       3.793 * cos(5.24575814515  + 2352.86615377180 *T)
    sumZ1  +=       2.840 * cos(2.47694599818  + 1577.34354244780 *T)
    sumZ1  +=       1.817 * cos(0.41874743765  + 6283.07584999140 *T)
    sumZ1  +=       1.499 * cos(1.83320979291  + 5856.47765911540 *T)
    sumZ1  +=       1.463 * cos(5.68891324948  + 5753.38488489680 *T)
    sumZ1  +=       1.302 * cos(2.18800611215  + 9437.76293488700 *T)
    sumZ1  +=       1.239 * cos(4.95327097854  + 10213.28554621100 *T)
    sumZ1  +=       1.029 * cos(0.09367831611  + 7860.41939243920 *T)
    sumZ1  +=       0.975 * cos(0.08833256880  + 14143.49524243060 *T)
    sumZ1  +=       0.851 * cos(1.79547916132  + 3930.20969621960 *T)
    sumZ1  +=       0.581 * cos(2.26949174067  + 5884.92684658320 *T)
    sumZ1  +=       0.515 * cos(5.64196681593  + 529.69096509460 *T)
    sumZ1  +=       0.473 * cos(6.22750969242  + 6309.37416979120 *T)
    sumZ1  +=       0.449 * cos(1.52767374606  + 18073.70493865020 *T)
    sumZ1  +=       0.360 * cos(3.62344325932  + 13367.97263110660 *T)
    sumZ1  +=       0.372 * cos(3.22470721320  + 6275.96230299060 *T)
    sumZ1  +=       0.260 * cos(1.87645933242  + 11790.62908865880 *T)
    sumZ1  +=       0.322 * cos(0.94084045832  + 6069.77675455340 *T)
    sumZ1  +=       0.232 * cos(0.27531852596  + 7058.59846131540 *T)
    sumZ1  +=       0.217 * cos(6.03652873142  + 10977.07880469900 *T)
    sumZ1  +=       0.232 * cos(2.93325646109  + 22003.91463486980 *T)
    sumZ1  +=       0.204 * cos(3.86264841382  + 6496.37494542940 *T)
    sumZ1  +=       0.189 * cos(2.88937704419  + 15720.83878487840 *T)
    sumZ1  +=       0.179 * cos(4.90537280911  + 12036.46073488820 *T)
    sumZ1  +=       0.222 * cos(3.98495366315  + 6812.76681508600 *T)
    sumZ1  +=       0.213 * cos(1.57516933652  + 4694.00295470760 *T)
    sumZ1  +=       0.157 * cos(1.08259734788  + 5643.17856367740 *T)
    sumZ1  +=       0.154 * cos(5.99434678412  + 5486.77784317500 *T)
    sumZ1  +=       0.137 * cos(2.67760436027  + 6290.18939699220 *T)
    sumZ1  +=       0.179 * cos(2.05905949693  + 7084.89678111520 *T)
    sumZ1  +=       0.121 * cos(5.90212574947  + 9225.53927328300 *T)
    sumZ1  +=       0.125 * cos(2.24111392416  + 1059.38193018920 *T)
    sumZ1  +=       0.150 * cos(2.00175038718  + 5230.80746680300 *T)
    sumZ1  +=       0.149 * cos(5.06157254516  + 17298.18232732620 *T)
    sumZ1  +=       0.118 * cos(5.39979058038  + 3340.61242669980 *T)
    sumZ1  +=       0.161 * cos(3.32421999691  + 6283.31966747490 *T)
    sumZ1  +=       0.115 * cos(5.92406672373  + 4705.73230754360 *T)
    sumZ1  +=       0.118 * cos(4.40207874911  + 19651.04848109800 *T)
    sumZ1  +=       0.128 * cos(4.35489873365  + 25934.12433108940 *T)
    sumZ1  +=       0.109 * cos(2.52157834166  + 6438.49624942560 *T)
    sumZ1  +=       0.099 * cos(2.70727488041  + 5216.58037280140 *T)
    sumZ1  +=       0.122 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ1  +=       0.103 * cos(0.93782340879  + 8827.39026987480 *T)
    sumZ1  +=       0.083 * cos(4.12473540351  + 8635.94200376320 *T)
    sumZ1  +=       0.097 * cos(5.50959692365  + 29864.33402730900 *T)
    sumZ1  +=       0.073 * cos(1.73905345744  + 11506.76976979360 *T)
    sumZ1  +=       0.083 * cos(5.69169692653  + 775.52261132400 *T)
    sumZ1  +=       0.072 * cos(0.21891639822  + 21228.39202354580 *T)
    sumZ1  +=       0.071 * cos(2.86755026812  + 6681.22485339960 *T)
    sumZ1  +=       0.074 * cos(2.20184828895  + 37724.75341974820 *T)
    sumZ1  +=       0.063 * cos(4.45586625948  + 7079.37385680780 *T)
    sumZ1  +=       0.047 * cos(2.04946724464  + 3128.38876509580 *T)
    sumZ1  +=       0.061 * cos(0.63918772258  + 33794.54372352860 *T)
    sumZ1  +=       0.047 * cos(3.32543843300  + 26087.90314157420 *T)
    sumZ1  +=       0.049 * cos(1.60680905005  + 6702.56049386660 *T)
    sumZ1  +=       0.057 * cos(0.11215813438  + 29088.81141598500 *T)
    sumZ1  +=       0.049 * cos(3.02832204050  + 20426.57109242200 *T)
    sumZ1  +=       0.041 * cos(5.55329394890  + 11015.10647733480 *T)
    sumZ1  +=       0.041 * cos(5.91861144924  + 23581.25817731760 *T)
    sumZ1  +=       0.045 * cos(2.00068583743  + 426.59819087600 *T)
    sumZ1  +=       0.045 * cos(4.95273290181  + 5863.59120611620 *T)
    sumZ1  +=       0.050 * cos(3.62740835096  + 41654.96311596780 *T)
    sumZ1  +=       0.046 * cos(1.65798680284  + 25158.60171976540 *T)
    sumZ1  +=       0.036 * cos(5.61836577943  + 12566.15169998280 *T)
    sumZ1  +=       0.036 * cos(6.24373396652  + 6283.14316029419 *T)
    sumZ1  +=       0.036 * cos(0.40465162918  + 6283.00853968860 *T)
    sumZ1  +=       0.032 * cos(6.09025731476  + 64809.80550494129 *T)
    sumZ1  +=       0.032 * cos(6.03707103538  + 2942.46342329160 *T)
    sumZ1  +=       0.041 * cos(4.86809570283  + 1592.59601363280 *T)
    sumZ1  +=       0.037 * cos(1.04055368426  + 213.29909543800 *T)
    sumZ1  +=       0.031 * cos(3.62641145030  + 13095.84266507740 *T)
    sumZ1  +=       0.028 * cos(4.38359423735  + 7632.94325965020 *T)
    sumZ1  +=       0.030 * cos(2.03616887071  + 12139.55350910680 *T)
    sumZ1  +=       0.028 * cos(6.03334294232  + 17789.84561978500 *T)
    sumZ1  +=       0.026 * cos(3.88971333608  + 5331.35744374080 *T)
    sumZ1  +=       0.031 * cos(1.44666331503  + 16730.46368959580 *T)
    sumZ1  +=       0.026 * cos(6.26376705837  + 23543.23050468179 *T)
    sumZ1  +=       0.023 * cos(4.44388985550  + 18849.22754997420 *T)
    sumZ1  +=       0.025 * cos(4.13395006026  + 3154.68708489560 *T)
    sumZ1  +=       0.028 * cos(1.53862289477  + 6279.48542133960 *T)
    sumZ1  +=       0.028 * cos(1.96831814872  + 6286.66627864320 *T)
    sumZ1  +=       0.028 * cos(5.78094918529  + 15110.46611986620 *T)
    sumZ1  +=       0.025 * cos(0.62040343662  + 10988.80815753500 *T)
    sumZ1  +=       0.022 * cos(6.02390113954  + 16496.36139620240 *T)
    sumZ1  +=       0.026 * cos(2.48165809843  + 5729.50644714900 *T)
    sumZ1  +=       0.020 * cos(3.85655029499  + 9623.68827669120 *T)
    sumZ1  +=       0.021 * cos(5.83006047147  + 7234.79425624200 *T)
    sumZ1  +=       0.021 * cos(0.69628570421  + 398.14900340820 *T)
    sumZ1  +=       0.022 * cos(5.02222806555  + 6127.65545055720 *T)
    sumZ1  +=       0.020 * cos(3.47611265290  + 6148.01076995600 *T)
    sumZ1  +=       0.020 * cos(0.90769829044  + 5481.25491886760 *T)
    sumZ1  +=       0.020 * cos(0.03081589303  + 6418.14093002680 *T)
    sumZ1  +=       0.020 * cos(3.74220084927  + 1589.07289528380 *T)
    sumZ1  +=       0.018 * cos(1.58348238359  + 2118.76386037840 *T)
    sumZ1  +=       0.019 * cos(0.85407021371  + 14712.31711645800 *T)

    # EARTH Z Coord
    sumZ2  = 0
    sumZ2  +=       1.662 * cos(1.62703209173  + 84334.66158130829 *T)
    sumZ2  +=       0.492 * cos(2.41382223971  + 1047.74731175470 *T)
    sumZ2  +=       0.344 * cos(2.24353004539  + 5507.55323866740 *T)
    sumZ2  +=       0.258 * cos(6.00906896311  + 5223.69391980220 *T)
    sumZ2  +=       0.131 * cos(0.95447345240  + 6283.07584999140 *T)
    sumZ2  +=       0.086 * cos(1.69139956945  + 7860.41939243920 *T)
    sumZ2  +=       0.090 * cos(0.97606804452  + 1577.34354244780 *T)
    sumZ2  +=       0.090 * cos(0.37899871725  + 2352.86615377180 *T)
    sumZ2  +=       0.089 * cos(6.25807507963  + 10213.28554621100 *T)
    sumZ2  +=       0.075 * cos(0.84213523741  + 167621.57585086187 *T)
    sumZ2  +=       0.052 * cos(1.70501566089  + 14143.49524243060 *T)
    sumZ2  +=       0.057 * cos(6.15295833679  + 12194.03291462090 *T)
    sumZ2  +=       0.051 * cos(1.27616016740  + 5753.38488489680 *T)
    sumZ2  +=       0.051 * cos(5.37229738682  + 6812.76681508600 *T)
    sumZ2  +=       0.034 * cos(1.73672994279  + 7058.59846131540 *T)
    sumZ2  +=       0.038 * cos(2.77761031485  + 10988.80815753500 *T)
    sumZ2  +=       0.046 * cos(3.38617099014  + 156475.29024799570 *T)
    sumZ2  +=       0.021 * cos(1.95248349228  + 8827.39026987480 *T)
    sumZ2  +=       0.018 * cos(3.33419222028  + 8429.24126646660 *T)
    sumZ2  +=       0.019 * cos(4.32945160287  + 17789.84561978500 *T)
    sumZ2  +=       0.017 * cos(0.66191210656  + 6283.00853968860 *T)
    sumZ2  +=       0.018 * cos(3.74885333072  + 11769.85369316640 *T)
    sumZ2  +=       0.017 * cos(4.23058370776  + 10977.07880469900 *T)
    sumZ2  +=       0.017 * cos(1.78116162721  + 5486.77784317500 *T)
    sumZ2  +=       0.018 * cos(6.14348240716  + 11790.62908865880 *T)
    sumZ2  +=       0.022 * cos(1.34376831204  + 12036.46073488820 *T)
    sumZ2  +=       0.017 * cos(2.79601092529  + 796.29800681640 *T)
    sumZ2  +=       0.015 * cos(1.17032155085  + 213.29909543800 *T)
    sumZ2  +=       0.018 * cos(2.85221514199  + 5088.62883976680 *T)
    sumZ2  +=       0.017 * cos(0.21780913672  + 6283.14316029419 *T)
    sumZ2  +=       0.013 * cos(1.21201504386  + 25132.30339996560 *T)
    sumZ2  +=       0.012 * cos(5.13714452592  + 7079.37385680780 *T)
    sumZ2  +=       0.013 * cos(3.79842135217  + 4933.20844033260 *T)
    sumZ2  +=       0.012 * cos(4.89407978213  + 3738.76143010800 *T)
    sumZ2  +=       0.015 * cos(6.05682328852  + 398.14900340820 *T)
    sumZ2  +=       0.014 * cos(4.81029291856  + 4694.00295470760 *T)
    sumZ2  +=       0.011 * cos(0.61684523405  + 3128.38876509580 *T)
    sumZ2  +=       0.011 * cos(5.32876538500  + 6040.34724601740 *T)
    sumZ2  +=       0.014 * cos(5.27227350286  + 4535.05943692440 *T)
    sumZ2  +=       0.011 * cos(2.39292099451  + 5331.35744374080 *T)
    sumZ2  +=       0.010 * cos(4.45296532710  + 6525.80445396540 *T)
    sumZ2  +=       0.014 * cos(4.66400985037  + 8031.09226305840 *T)
    sumZ2  +=       0.010 * cos(3.22472385926  + 9437.76293488700 *T)
    sumZ2  +=       0.011 * cos(3.80913404437  + 801.82093112380 *T)
    sumZ2  +=       0.010 * cos(5.15032130575  + 11371.70468975820 *T)
    sumZ2  +=       0.013 * cos(0.98720797401  + 5729.50644714900 *T)
    sumZ2  +=       0.009 * cos(5.94191743597  + 7632.94325965020 *T)

    # EARTH Z Coord
    sumZ3  = 0
    sumZ3  +=       0.011 * cos(0.23877262399  + 7860.41939243920 *T)
    sumZ3  +=       0.009 * cos(1.16069982609  + 5507.55323866740 *T)
    sumZ3  +=       0.008 * cos(1.65357552925  + 5884.92684658320 *T)
    sumZ3  +=       0.008 * cos(2.86720038197  + 7058.59846131540 *T)
    sumZ3  +=       0.007 * cos(3.04818741666  + 5486.77784317500 *T)
    sumZ3  +=       0.007 * cos(2.59437103785  + 529.69096509460 *T)
    sumZ3  +=       0.008 * cos(4.02863090524  + 6256.77753019160 *T)
    sumZ3  +=       0.008 * cos(2.42003508927  + 5753.38488489680 *T)
    sumZ3  +=       0.006 * cos(0.84181087594  + 6275.96230299060 *T)
    sumZ3  +=       0.006 * cos(5.40160929468  + 1577.34354244780 *T)
    sumZ3  +=       0.007 * cos(2.73399865247  + 6309.37416979120 *T)

    # EARTH Z Coord
    sumZ4  = 0
    sumZ4  +=       0.004 * cos(0.79662198849  + 6438.49624942560 *T)
    sumZ4  +=       0.005 * cos(0.84308705203  + 1047.74731175470 *T)
    sumZ4  +=       0.005 * cos(0.05711572303  + 84334.66158130829 *T)
    sumZ4  +=       0.003 * cos(3.46779895686  + 6279.55273164240 *T)
    sumZ4  +=       0.003 * cos(2.89822201212  + 6127.65545055720 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Earth",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    # save parameters and values
    EarthdtUTimes[strdtUT] = Data


    return Data
#######################################################################



def getEarthData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getEarthXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getJupiterXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Jupiter"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getJupiterXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getJupiterXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # JUPITER X Coord
    sumX0  = 0
    sumX0  +=  519591755.961 * cos(0.59954672200  + 529.93478257810 *T)
    sumX0  +=  37839498.798 * cos(3.39164799011  + 0.24381748350 *T)
    sumX0  +=  12593400.247 * cos(0.94916456487  + 1059.62574767270 *T)
    sumX0  +=  1504469.362 * cos(0.72934997067  + 522.82123557730 *T)
    sumX0  +=  1476016.965 * cos(3.61748058581  + 537.04832957890 *T)
    sumX0  +=  457751.771 * cos(1.29886948102  + 1589.31671276730 *T)
    sumX0  +=  301186.623 * cos(5.17699571796  + 7.35736448430 *T)
    sumX0  +=  381996.183 * cos(1.98605213956  + 102.84895673509 *T)
    sumX0  +=  193179.179 * cos(5.02674864982  + 426.84200835950 *T)
    sumX0  +=  148291.722 * cos(6.15618324076  + 110.45013870291 *T)
    sumX0  +=  142171.400 * cos(5.56556633177  + 633.02755679670 *T)
    sumX0  +=  133443.745 * cos(0.89287673636  + 213.54291292150 *T)
    sumX0  +=  102719.450 * cos(6.17016418155  + 1052.51220067191 *T)
    sumX0  +=   87327.523 * cos(6.27541436099  + 529.44714761109 *T)
    sumX0  +=  114122.292 * cos(0.01655163411  + 1162.71852189130 *T)
    sumX0  +=   72093.376 * cos(3.96122861070  + 1066.73929467350 *T)
    sumX0  +=   59268.404 * cos(4.46187501371  + 949.41942645330 *T)
    sumX0  +=   68329.094 * cos(3.43833771050  + 846.32665223470 *T)
    sumX0  +=   30209.026 * cos(2.97106959244  + 205.94173095370 *T)
    sumX0  +=   34007.392 * cos(1.74694593876  + 419.24082639170 *T)
    sumX0  +=   26616.582 * cos(4.85226363695  + 323.74923414091 *T)
    sumX0  +=   19727.427 * cos(1.64894896707  + 2119.00767786191 *T)
    sumX0  +=   17749.540 * cos(0.84241151614  + 419.72846135871 *T)
    sumX0  +=   17560.618 * cos(1.95511771720  + 316.63568714010 *T)
    sumX0  +=   16308.761 * cos(5.73600933766  + 736.12033101530 *T)
    sumX0  +=   16227.856 * cos(0.86857414313  + 515.70768857651 *T)
    sumX0  +=   14406.626 * cos(2.06284119096  + 743.23387801611 *T)
    sumX0  +=   11421.526 * cos(3.48157539864  + 544.16187657970 *T)
    sumX0  +=   10835.166 * cos(3.19395745172  + 103.33659170210 *T)
    sumX0  +=   12154.356 * cos(3.75230765024  + 526.00262931501 *T)
    sumX0  +=   11992.571 * cos(0.58508657370  + 533.86693584120 *T)
    sumX0  +=    8455.662 * cos(3.47881139969  + 640.14110379750 *T)
    sumX0  +=    8193.182 * cos(5.57030621208  + 1479.11039154791 *T)
    sumX0  +=    8713.455 * cos(0.38315206501  + 1692.40948698591 *T)
    sumX0  +=    7420.814 * cos(5.98301259011  + 956.53297345411 *T)
    sumX0  +=    7565.993 * cos(0.90590258528  + 1265.81129610991 *T)
    sumX0  +=    8202.922 * cos(1.41656499571  + 735.63269604830 *T)
    sumX0  +=    7717.111 * cos(0.13134820979  + 1582.20316576650 *T)
    sumX0  +=    5320.837 * cos(1.10497796931  + 526.75338884040 *T)
    sumX0  +=    5217.759 * cos(3.23246701096  + 533.11617631580 *T)
    sumX0  +=    5238.212 * cos(3.11227009303  + 522.33360061030 *T)
    sumX0  +=    5774.716 * cos(5.03860697265  + 14.47091148511 *T)
    sumX0  +=    5907.614 * cos(4.98319674710  + 109.96250373590 *T)
    sumX0  +=    4575.512 * cos(3.75638993330  + 1376.01761732930 *T)
    sumX0  +=    3939.870 * cos(4.30901425634  + 1596.43025976811 *T)
    sumX0  +=    4568.269 * cos(2.14654320251  + 95.73540973430 *T)
    sumX0  +=    3036.375 * cos(3.94492954790  + 309.03450517230 *T)
    sumX0  +=    2856.672 * cos(6.01338099597  + 117.56368570370 *T)
    sumX0  +=    2820.138 * cos(5.26515556940  + 632.53992182970 *T)
    sumX0  +=    2442.859 * cos(4.24056243704  + 433.95555536030 *T)
    sumX0  +=    2534.517 * cos(6.18089912855  + 625.91400979590 *T)
    sumX0  +=    2329.307 * cos(5.93999996459  + 213.05527795450 *T)
    sumX0  +=    2430.903 * cos(3.89413393897  + 220.65645992230 *T)
    sumX0  +=    2673.686 * cos(3.18830432379  + 1169.83206889211 *T)
    sumX0  +=    2396.896 * cos(5.93300784335  + 1045.39865367110 *T)
    sumX0  +=    2503.065 * cos(1.15446452871  + 1052.02456570489 *T)
    sumX0  +=    1817.118 * cos(2.38828009078  + 838.72547026689 *T)
    sumX0  +=    2014.206 * cos(0.16819149452  + 1155.60497489050 *T)
    sumX0  +=    1501.452 * cos(4.60588864812  + 942.30587945250 *T)
    sumX0  +=    1454.916 * cos(5.75976807023  + 853.44019923550 *T)
    sumX0  +=    1628.897 * cos(5.49183726268  + 75.02541605080 *T)
    sumX0  +=    1106.760 * cos(4.07652960629  + 6.86972951729 *T)
    sumX0  +=    1080.777 * cos(5.15189894960  + 316.14805217309 *T)
    sumX0  +=    1041.755 * cos(0.10701428561  + 1795.50226120450 *T)
    sumX0  +=     996.914 * cos(3.08308908686  + 1272.92484311070 *T)
    sumX0  +=     928.877 * cos(3.11398170655  + 198.82818395289 *T)
    sumX0  +=     934.490 * cos(1.99952244525  + 2648.69864295650 *T)
    sumX0  +=     858.575 * cos(3.71317000991  + 529.88659846831 *T)
    sumX0  +=     858.481 * cos(0.62779584572  + 529.98296668790 *T)
    sumX0  +=    1088.006 * cos(1.13420142953  + 527.48710202330 *T)
    sumX0  +=    1080.077 * cos(3.20528251357  + 532.38246313290 *T)
    sumX0  +=     958.229 * cos(1.34776425757  + 149.80701461810 *T)
    sumX0  +=     830.762 * cos(3.68869745476  + 625.42637482889 *T)
    sumX0  +=     815.889 * cos(0.09783778677  + 1059.13811270570 *T)
    sumX0  +=     838.684 * cos(4.25222669612  + 3.68833577960 *T)
    sumX0  +=     956.674 * cos(5.34413307275  + 379.88395047650 *T)
    sumX0  +=     694.759 * cos(2.12065531862  + 1155.11733992350 *T)
    sumX0  +=     673.874 * cos(6.09243781947  + 2008.80135664250 *T)
    sumX0  +=     865.777 * cos(4.81335091362  + 426.35437339250 *T)
    sumX0  +=     897.730 * cos(0.91007512660  + 1368.41643536150 *T)
    sumX0  +=     667.422 * cos(1.95480301512  + 1368.90407032850 *T)
    sumX0  +=     747.747 * cos(4.71007343847  + 330.86278114170 *T)
    sumX0  +=     693.620 * cos(1.33864708601  + 1063.55790093581 *T)
    sumX0  +=     787.813 * cos(0.40190120466  + 839.21310523390 *T)
    sumX0  +=     614.978 * cos(0.12549858875  + 2111.89413086110 *T)
    sumX0  +=     680.352 * cos(0.29385269041  + 528.97157473160 *T)
    sumX0  +=     678.592 * cos(4.04690202000  + 530.89799042461 *T)
    sumX0  +=     574.130 * cos(0.74562235425  + 2222.10045208050 *T)
    sumX0  +=     644.748 * cos(1.92762797797  + 412.12727939089 *T)
    sumX0  +=     742.423 * cos(3.95986148740  + 206.42936592071 *T)
    sumX0  +=     563.602 * cos(4.06358227588  + 1055.69359440960 *T)
    sumX0  +=     559.331 * cos(5.87731196423  + 729.00678401450 *T)
    sumX0  +=     647.539 * cos(5.12542038621  + 984.84414910540 *T)
    sumX0  +=     460.071 * cos(3.82654707910  + 1073.85284167431 *T)
    sumX0  +=     403.270 * cos(0.99319532178  + 412.61491435790 *T)
    sumX0  +=     539.827 * cos(5.31630620093  + 38.37685312130 *T)
    sumX0  +=     407.824 * cos(2.98828635038  + 910.06255053810 *T)
    sumX0  +=     319.575 * cos(3.35886072316  + 941.81824448550 *T)
    sumX0  +=     301.187 * cos(4.06443054226  + 1905.70858242391 *T)
    sumX0  +=     330.186 * cos(3.60882237102  + 948.93179148630 *T)
    sumX0  +=     334.161 * cos(0.95340181567  + 76.50988875911 *T)
    sumX0  +=     292.241 * cos(1.85350584256  + 1471.50920958009 *T)
    sumX0  +=     388.960 * cos(6.06054382912  + 604.71638114540 *T)
    sumX0  +=     387.023 * cos(4.52724734369  + 455.15318401080 *T)
    sumX0  +=     303.530 * cos(5.06433323115  + 529.41351771630 *T)
    sumX0  +=     303.139 * cos(5.55956712580  + 530.45604743991 *T)
    sumX0  +=     276.704 * cos(0.65474572491  + 1486.22393854870 *T)
    sumX0  +=     383.222 * cos(4.89517559942  + 3.42521122121 *T)
    sumX0  +=     267.964 * cos(2.01438259050  + 309.52214013930 *T)
    sumX0  +=     270.827 * cos(1.04116519472  + 1258.69774910910 *T)
    sumX0  +=     264.824 * cos(2.94766261780  + 536.56069461190 *T)
    sumX0  +=     342.262 * cos(0.65840844301  + 1684.80830501809 *T)
    sumX0  +=     254.392 * cos(3.58469491330  + 1062.80714141041 *T)
    sumX0  +=     289.891 * cos(4.38886336303  + 742.74624304909 *T)
    sumX0  +=     227.013 * cos(4.65796948353  + 2126.12122486270 *T)
    sumX0  +=     219.942 * cos(3.05429299635  + 515.22005360950 *T)
    sumX0  +=     304.124 * cos(1.75362149529  + 6283.31966747490 *T)
    sumX0  +=     289.590 * cos(2.97007145679  + 305.10235190919 *T)
    sumX0  +=     225.217 * cos(3.50147007971  + 1699.52303398670 *T)
    sumX0  +=     192.734 * cos(1.02002841562  + 508.59414157570 *T)
    sumX0  +=     181.955 * cos(0.77882678658  + 1898.59503542310 *T)
    sumX0  +=     180.824 * cos(2.76562890852  + 1265.32366114289 *T)
    sumX0  +=     234.431 * cos(0.92363421460  + 1056.44435393500 *T)
    sumX0  +=     210.412 * cos(4.02038896426  + 490.57790666290 *T)
    sumX0  +=     178.960 * cos(6.10154360210  + 527.01402127131 *T)
    sumX0  +=     184.077 * cos(5.59641213885  + 1471.99684454710 *T)
    sumX0  +=     177.005 * cos(4.52323003424  + 532.85554388490 *T)
    sumX0  +=     197.754 * cos(6.22787965841  + 1575.08961876571 *T)
    sumX0  +=     201.548 * cos(1.60035448064  + 728.51914904750 *T)
    sumX0  +=     164.871 * cos(3.47686660224  + 529.18402305270 *T)
    sumX0  +=     164.585 * cos(0.86380543039  + 530.68554210351 *T)
    sumX0  +=     191.757 * cos(0.30748865043  + 569.29165849331 *T)
    sumX0  +=     155.012 * cos(1.25790672042  + 519.63984183961 *T)
    sumX0  +=     216.423 * cos(5.68405931071  + 453.18107633550 *T)
    sumX0  +=     208.270 * cos(3.39324145545  + 1439.75351563271 *T)
    sumX0  +=     195.461 * cos(0.69780042310  + 1685.29593998510 *T)
    sumX0  +=     158.021 * cos(4.32825610392  + 525.74199688410 *T)
    sumX0  +=     155.478 * cos(0.01448171646  + 534.12756827210 *T)
    sumX0  +=     142.916 * cos(2.35979415786  + 11.28951774740 *T)
    sumX0  +=     166.137 * cos(3.66026536347  + 224.58861318540 *T)
    sumX0  +=     139.568 * cos(3.35575676150  + 647.25465079831 *T)
    sumX0  +=     194.718 * cos(5.30500289646  + 1021.49271203491 *T)
    sumX0  +=     142.687 * cos(3.06398958840  + 1258.21011414209 *T)
    sumX0  +=     136.501 * cos(6.00364083075  + 99.40443843901 *T)
    sumX0  +=     155.156 * cos(5.30496926203  + 845.83901726770 *T)
    sumX0  +=     148.749 * cos(4.08138644957  + 301.92095817150 *T)
    sumX0  +=     142.483 * cos(3.75685487130  + 227.77000692311 *T)
    sumX0  +=     128.042 * cos(5.39609470815  + 1162.23088692429 *T)
    sumX0  +=     129.196 * cos(1.58923571595  + 1787.90107923670 *T)
    sumX0  +=     176.884 * cos(3.17597373171  + 10213.52936369450 *T)
    sumX0  +=     120.579 * cos(3.08299978310  + 540.22972331660 *T)
    sumX0  +=     163.945 * cos(2.01248197600  + 1.72829019180 *T)
    sumX0  +=     135.003 * cos(3.79885672041  + 540.98048284201 *T)
    sumX0  +=     121.846 * cos(3.78904805348  + 524.30570828560 *T)
    sumX0  +=     120.032 * cos(0.55319497448  + 535.56385687061 *T)
    sumX0  +=     124.238 * cos(3.38846678226  + 983.35967639710 *T)
    sumX0  +=     123.682 * cos(3.34428700111  + 525.26891613211 *T)
    sumX0  +=     104.430 * cos(0.57516865787  + 530.09484127250 *T)
    sumX0  +=     104.430 * cos(3.76551699050  + 529.77472388370 *T)
    sumX0  +=     136.458 * cos(0.41278945595  + 2001.20017467470 *T)
    sumX0  +=     115.988 * cos(3.35547392183  + 750.34742501690 *T)
    sumX0  +=     104.731 * cos(3.32722729895  + 551.27542358051 *T)
    sumX0  +=     100.945 * cos(2.35927152342  + 1581.71553079950 *T)
    sumX0  +=     116.158 * cos(0.99764725383  + 534.60064902410 *T)
    sumX0  +=     122.805 * cos(4.90630450089  + 524.51815660670 *T)
    sumX0  +=     103.104 * cos(4.41419723891  + 963.64652045490 *T)
    sumX0  +=     115.142 * cos(0.31814375241  + 2325.19322629910 *T)
    sumX0  +=     120.928 * cos(5.71584928874  + 535.35140854951 *T)
    sumX0  +=      94.293 * cos(3.24393572880  + 1802.61580820530 *T)
    sumX0  +=      92.521 * cos(0.02282195426  + 1383.13116433011 *T)
    sumX0  +=      85.964 * cos(2.53905253590  + 831.61192326610 *T)
    sumX0  +=      91.882 * cos(3.03470202176  + 441.06910236111 *T)
    sumX0  +=      90.884 * cos(2.64222941727  + 4.17597074660 *T)
    sumX0  +=      96.153 * cos(4.88678063240  + 21.58445848590 *T)
    sumX0  +=      73.111 * cos(5.81311465775  + 415.79630809560 *T)
    sumX0  +=      72.096 * cos(0.31807246212  + 529.72233425701 *T)
    sumX0  +=      72.096 * cos(4.02261318625  + 530.14723089920 *T)
    sumX0  +=      95.408 * cos(5.93750222727  + 35.66854013560 *T)
    sumX0  +=      79.513 * cos(0.68230348252  + 835.28095197081 *T)
    sumX0  +=      68.517 * cos(2.77354883412  + 1574.60198379870 *T)
    sumX0  +=      69.848 * cos(0.27519157939  + 518.88908231420 *T)
    sumX0  +=      73.288 * cos(0.55107431240  + 832.09955823310 *T)
    sumX0  +=      73.790 * cos(1.37824862036  + 1044.91101870410 *T)
    sumX0  +=      62.635 * cos(0.34734323123  + 2214.98690507971 *T)
    sumX0  +=      72.372 * cos(5.64987811590  + 511.77553531340 *T)
    sumX0  +=      79.774 * cos(1.04707348735  + 618.80046279510 *T)
    sumX0  +=      84.465 * cos(0.44987072707  + 834.79331700380 *T)
    sumX0  +=      82.357 * cos(2.58250672354  + 531.41925528640 *T)
    sumX0  +=      68.826 * cos(4.98304145215  + 548.09402984280 *T)
    sumX0  +=      58.994 * cos(1.32874569406  + 2104.29294889330 *T)
    sumX0  +=      69.760 * cos(5.12306055524  + 98.91680347199 *T)
    sumX0  +=      74.013 * cos(4.41911169510  + 174.18603700630 *T)
    sumX0  +=      55.680 * cos(4.79831502417  + 644.07325706060 *T)
    sumX0  +=      77.878 * cos(1.73526543836  + 528.45030986980 *T)
    sumX0  +=      54.479 * cos(2.02247283767  + 1898.10740045609 *T)
    sumX0  +=      55.045 * cos(2.14129882164  + 639.65346883050 *T)
    sumX0  +=      53.073 * cos(1.97157293394  + 1048.58004740881 *T)
    sumX0  +=      52.282 * cos(4.62008175590  + 960.46512671720 *T)
    sumX0  +=      51.560 * cos(1.64669753295  + 514.95692905111 *T)
    sumX0  +=      52.703 * cos(0.30289030550  + 2641.58509595570 *T)
    sumX0  +=      66.188 * cos(0.59773291524  + 230.32075334190 *T)
    sumX0  +=      49.711 * cos(2.69410394468  + 544.91263610510 *T)
    sumX0  +=      49.029 * cos(3.17600907963  + 387.48513244430 *T)
    sumX0  +=      65.960 * cos(1.03002405667  + 1364.97191706540 *T)
    sumX0  +=      52.715 * cos(1.48379190348  + 1057.17806711790 *T)
    sumX0  +=      58.796 * cos(4.35893030098  + 39.11305843170 *T)
    sumX0  +=      59.809 * cos(0.66614084819  + 91.80325647120 *T)
    sumX0  +=      46.231 * cos(0.25459657690  + 2538.49232173710 *T)
    sumX0  +=      48.382 * cos(5.62148696439  + 2428.28600051770 *T)
    sumX0  +=      60.070 * cos(3.77993719329  + 1062.07342822750 *T)
    sumX0  +=      56.428 * cos(0.17441147343  + 2317.59204433130 *T)
    sumX0  +=      59.945 * cos(0.89318521903  + 494.51005992600 *T)
    sumX0  +=      43.719 * cos(0.36566607858  + 2.20386307129 *T)
    sumX0  +=      48.181 * cos(3.78650322165  + 618.31282782810 *T)
    sumX0  +=      49.751 * cos(1.88142062033  + 679.49797971270 *T)
    sumX0  +=      58.551 * cos(2.31298978175  + 1058.14127496441 *T)
    sumX0  +=      47.067 * cos(2.35309949230  + 3178.38960805111 *T)
    sumX0  +=      52.897 * cos(1.94858672370  + 106.78110999820 *T)
    sumX0  +=      41.687 * cos(0.97719980018  + 1059.67393178251 *T)
    sumX0  +=      41.782 * cos(3.44686195540  + 71.09326278771 *T)
    sumX0  +=      43.674 * cos(1.21900999269  + 430.77416162260 *T)
    sumX0  +=      44.776 * cos(4.88736591079  + 312.70353387701 *T)
    sumX0  +=      48.969 * cos(5.94887655147  + 1038.28510667030 *T)
    sumX0  +=      38.530 * cos(2.27809788058  + 1148.00379292269 *T)
    sumX0  +=      52.006 * cos(3.46735017105  + 565.35950523021 *T)
    sumX0  +=      47.878 * cos(3.92919910061  + 1375.52998236229 *T)
    sumX0  +=      38.108 * cos(0.29006450181  + 1788.38871420371 *T)
    sumX0  +=      49.073 * cos(4.15252949988  + 523.78444342380 *T)
    sumX0  +=      38.297 * cos(5.94013989158  + 106.51798543980 *T)
    sumX0  +=      41.687 * cos(4.06257397973  + 1059.57756356290 *T)
    sumX0  +=      37.366 * cos(1.12685707466  + 2751.79141717511 *T)
    sumX0  +=      41.634 * cos(4.80654439187  + 1478.62275658089 *T)
    sumX0  +=      41.532 * cos(5.44681505603  + 1514.53511420000 *T)
    sumX0  +=      41.308 * cos(4.63852007603  + 10.53875822200 *T)
    sumX0  +=      40.633 * cos(0.18959097751  + 18.40306474821 *T)
    sumX0  +=      42.800 * cos(0.11312892780  + 113.89465699899 *T)
    sumX0  +=      48.985 * cos(6.20232924224  + 3340.85624418330 *T)
    sumX0  +=      44.449 * cos(0.18999902170  + 536.08512173241 *T)
    sumX0  +=      34.928 * cos(1.15350567398  + 672.38443271190 *T)
    sumX0  +=      36.386 * cos(2.02390109067  + 1361.79052332771 *T)
    sumX0  +=      36.844 * cos(4.88632014089  + 530.19541500900 *T)
    sumX0  +=      36.844 * cos(5.73755081465  + 529.67415014720 *T)
    sumX0  +=      34.444 * cos(2.48976796512  + 1890.99385345530 *T)
    sumX0  +=      41.479 * cos(1.78003006526  + 1593.24886603040 *T)
    sumX0  +=      46.613 * cos(3.74069024826  + 1513.05064149171 *T)
    sumX0  +=      32.971 * cos(3.84810585282  + 522.77305146750 *T)
    sumX0  +=      35.418 * cos(0.77769804196  + 522.86941968710 *T)
    sumX0  +=      32.742 * cos(5.83069185951  + 124.67723270450 *T)
    sumX0  +=      34.850 * cos(4.49564263533  + 1060.58895551920 *T)
    sumX0  +=      30.476 * cos(4.38612921635  + 1585.38455950421 *T)
    sumX0  +=      29.011 * cos(1.17948661211  + 121.49583896680 *T)
    sumX0  +=      33.477 * cos(2.97189030450  + 629.09540353360 *T)
    sumX0  +=      28.909 * cos(0.86697673022  + 2001.68780964171 *T)
    sumX0  +=      29.909 * cos(3.55918762072  + 537.00014546910 *T)
    sumX0  +=      27.497 * cos(0.49110595355  + 537.09651368871 *T)
    sumX0  +=      28.353 * cos(1.72487234310  + 2214.49927011270 *T)
    sumX0  +=      27.748 * cos(1.07211390256  + 2420.68481854989 *T)
    sumX0  +=      33.057 * cos(6.08453905425  + 568.06781821590 *T)
    sumX0  +=      27.675 * cos(5.33832022976  + 528.66275819090 *T)
    sumX0  +=      29.521 * cos(1.16643867883  + 1361.30288836070 *T)
    sumX0  +=      27.586 * cos(5.28550988684  + 531.20680696530 *T)
    sumX0  +=      33.930 * cos(6.20456328242  + 885.68352814990 *T)
    sumX0  +=      27.514 * cos(3.02979777846  + 114.38229196601 *T)
    sumX0  +=      24.495 * cos(2.50371137844  + 99.66756299740 *T)
    sumX0  +=      24.344 * cos(4.77377573492  + 935.19233245171 *T)
    sumX0  +=      26.492 * cos(4.72075048576  + 422.90985509640 *T)
    sumX0  +=      30.581 * cos(4.75568574212  + 142.20583265029 *T)
    sumX0  +=      32.993 * cos(0.64338422669  + 1058.66253982621 *T)
    sumX0  +=      26.693 * cos(5.72614338710  + 747.16603127920 *T)
    sumX0  +=      23.218 * cos(3.10531480612  + 216.98743121760 *T)
    sumX0  +=      31.446 * cos(0.08123535736  + 437.88770862340 *T)
    sumX0  +=      23.990 * cos(5.67950841232  + 523.34250043910 *T)
    sumX0  +=      25.905 * cos(1.13267038207  + 10.80188278040 *T)
    sumX0  +=      30.584 * cos(4.71116923713  + 1024.20102502061 *T)
    sumX0  +=      22.355 * cos(4.58469057077  + 106.03035047280 *T)
    sumX0  +=      27.098 * cos(0.93569436946  + 497.69145366371 *T)
    sumX0  +=      29.688 * cos(4.36436889437  + 760.01171843649 *T)
    sumX0  +=      26.696 * cos(3.19374047498  + 1158.78636862821 *T)
    sumX0  +=      25.786 * cos(5.76825883047  + 636.95971005981 *T)
    sumX0  +=      24.825 * cos(0.15973617739  + 1134.40734624000 *T)
    sumX0  +=      22.840 * cos(2.40514320530  + 459.08533727390 *T)
    sumX0  +=      21.870 * cos(4.94490093150  + 536.52706471711 *T)
    sumX0  +=      22.257 * cos(0.47018335123  + 521.85802773080 *T)
    sumX0  +=      26.904 * cos(6.12428331703  + 299.37021175271 *T)
    sumX0  +=      20.168 * cos(6.20868850257  + 2104.78058386030 *T)
    sumX0  +=      19.591 * cos(4.17431782455  + 1603.54380676890 *T)
    sumX0  +=      23.983 * cos(6.22729049577  + 2633.98391398789 *T)
    sumX0  +=      23.778 * cos(3.38484692408  + 562.17811149250 *T)
    sumX0  +=      19.633 * cos(4.12778045809  + 491.80174694030 *T)
    sumX0  +=      26.288 * cos(4.52163096435  + 454.66554904380 *T)
    sumX0  +=      21.128 * cos(3.50042367011  + 934.70469748469 *T)
    sumX0  +=      19.074 * cos(2.11128326565  + 327.68138740401 *T)
    sumX0  +=      18.465 * cos(2.01761953412  + 1464.39566257930 *T)
    sumX0  +=      21.220 * cos(0.04126742412  + 495.99453263430 *T)
    sumX0  +=      24.986 * cos(5.13828768722  + 860.55374623631 *T)
    sumX0  +=      18.596 * cos(2.14973780639  + 5.87289177600 *T)
    sumX0  +=      24.136 * cos(2.13158664025  + 380.37158544351 *T)
    sumX0  +=      18.303 * cos(4.62800343377  + 953.35157971640 *T)
    sumX0  +=      24.099 * cos(4.23660015333  + 621.98185653281 *T)
    sumX0  +=      19.533 * cos(1.90765352683  + 1098.98262358790 *T)
    sumX0  +=      17.623 * cos(2.20903684953  + 2207.38572311189 *T)
    sumX0  +=      18.076 * cos(1.99926940915  + 600.78422788230 *T)
    sumX0  +=      17.281 * cos(3.28385778370  + 191.71463695210 *T)
    sumX0  +=      17.873 * cos(3.86678308802  + 538.01153742540 *T)
    sumX0  +=      17.226 * cos(4.25912073762  + 2435.39954751850 *T)
    sumX0  +=      21.244 * cos(0.00678624260  + 1166.65067515440 *T)
    sumX0  +=      20.337 * cos(4.39625051286  + 1795.01462623750 *T)
    sumX0  +=      16.743 * cos(0.09936960208  + 2531.37877473630 *T)
    sumX0  +=      19.199 * cos(4.29791815751  + 563.87503252191 *T)
    sumX0  +=      16.833 * cos(4.00808166209  + 24.62283987170 *T)
    sumX0  +=      21.175 * cos(4.99199643071  + 88.62186273350 *T)
    sumX0  +=      16.556 * cos(3.85103221392  + 2229.21399908130 *T)
    sumX0  +=      16.153 * cos(0.18779034779  + 1049.33080693420 *T)
    sumX0  +=      19.785 * cos(0.93062834115  + 505.55576018990 *T)
    sumX0  +=      18.422 * cos(3.75327275320  + 1969.44448072730 *T)
    sumX0  +=      19.816 * cos(3.66020647553  + 96.22304470130 *T)
    sumX0  +=      16.272 * cos(3.90786214802  + 527.22646959241 *T)
    sumX0  +=      16.170 * cos(0.43269809157  + 532.64309556381 *T)
    sumX0  +=      15.510 * cos(0.95686644280  + 988.77630236850 *T)
    sumX0  +=      15.594 * cos(4.85697058261  + 39.60069339870 *T)
    sumX0  +=      15.853 * cos(3.02589971757  + 501.62360692681 *T)
    sumX0  +=      16.153 * cos(0.61159863731  + 31.99951143090 *T)
    sumX0  +=      18.509 * cos(3.41681808770  + 554.31380496631 *T)
    sumX0  +=      15.127 * cos(3.55269023427  + 460.78225830330 *T)
    sumX0  +=      16.461 * cos(2.78340595427  + 220.16882495529 *T)
    sumX0  +=      19.878 * cos(3.75956771122  + 1691.92185201890 *T)
    sumX0  +=      14.712 * cos(5.41386260565  + 1059.10448281091 *T)
    sumX0  +=      17.901 * cos(3.55352098378  + 107.26874496520 *T)
    sumX0  +=      14.484 * cos(4.85838274505  + 1070.67144793660 *T)
    sumX0  +=      15.909 * cos(2.53709217872  + 4.90968392950 *T)
    sumX0  +=      14.674 * cos(1.30453945233  + 558.24595822940 *T)
    sumX0  +=      14.766 * cos(5.45303717130  + 2744.67787017430 *T)
    sumX0  +=      18.715 * cos(0.73543610732  + 523.71500745461 *T)
    sumX0  +=      14.896 * cos(6.07608976526  + 1588.82907780029 *T)
    sumX0  +=      17.816 * cos(2.22215138151  + 74.53778108379 *T)
    sumX0  +=      14.616 * cos(1.43483029983  + 2530.89113976930 *T)
    sumX0  +=      18.292 * cos(4.97649486970  + 1151.67282162740 *T)
    sumX0  +=      18.426 * cos(3.60525731370  + 536.15455770160 *T)
    sumX0  +=      13.673 * cos(2.25783413626  + 732.18817775220 *T)
    sumX0  +=      13.765 * cos(4.47849985236  + 67.42423408300 *T)
    sumX0  +=      14.986 * cos(1.65285454596  + 2015.91490364330 *T)
    sumX0  +=      13.750 * cos(2.09723381259  + 451.22103074770 *T)
    sumX0  +=      13.701 * cos(1.07017817726  + 1586.13531902961 *T)
    sumX0  +=      13.415 * cos(0.81149327947  + 2737.07668820650 *T)
    sumX0  +=      13.707 * cos(3.60868974559  + 498.91529394111 *T)
    sumX0  +=      13.957 * cos(3.93496730766  + 1592.49810650500 *T)
    sumX0  +=      12.833 * cos(2.22104975837  + 423.66061462181 *T)
    sumX0  +=      12.935 * cos(0.78943775071  + 599.08730685290 *T)
    sumX0  +=      13.524 * cos(5.01956369633  + 2655.81218995730 *T)
    sumX0  +=      15.347 * cos(3.07625567879  + 323.26159917390 *T)
    sumX0  +=      14.250 * cos(3.29327530307  + 462.26673101160 *T)
    sumX0  +=      14.314 * cos(5.64098428550  + 1551.18367712950 *T)
    sumX0  +=      16.350 * cos(4.95142719083  + 1290.19031849811 *T)
    sumX0  +=      12.174 * cos(0.91067652575  + 2318.07967929830 *T)
    sumX0  +=      14.758 * cos(1.09361060094  + 945.48727319021 *T)
    sumX0  +=      13.727 * cos(5.64051351782  + 113.63153244060 *T)
    sumX0  +=      12.014 * cos(3.16779032587  + 938.37372618940 *T)
    sumX0  +=      12.669 * cos(0.88657082206  + 597.60283414461 *T)
    sumX0  +=      12.684 * cos(0.73280624012  + 560.95427121510 *T)
    sumX0  +=      12.686 * cos(2.81489159967  + 528.23786154870 *T)
    sumX0  +=      12.686 * cos(1.52579404870  + 531.63170360751 *T)
    sumX0  +=      11.902 * cos(1.87001916874  + 636.20895053440 *T)
    sumX0  +=      15.029 * cos(4.51212876715  + 155.53915477459 *T)
    sumX0  +=      14.712 * cos(5.90909648144  + 1060.14701253450 *T)
    sumX0  +=      13.153 * cos(5.81211132947  + 34.18406742730 *T)
    sumX0  +=      11.492 * cos(1.17570579774  + 0.71939036300 *T)
    sumX0  +=      11.728 * cos(2.23882691795  + 608.64853440851 *T)
    sumX0  +=      14.802 * cos(4.49062834163  + 760.49935340350 *T)
    sumX0  +=      14.208 * cos(0.26437682229  + 151.29148732640 *T)
    sumX0  +=      13.247 * cos(3.17700507895  + 5.17280848790 *T)
    sumX0  +=      13.608 * cos(4.18261824771  + 28.06735816780 *T)
    sumX0  +=      11.186 * cos(2.71818106129  + 531.37107117660 *T)
    sumX0  +=      11.186 * cos(1.62250458708  + 528.49849397960 *T)
    sumX0  +=      14.801 * cos(0.14206552505  + 378.39947776819 *T)
    sumX0  +=      11.873 * cos(0.97351968238  + 1677.69475801730 *T)
    sumX0  +=      10.898 * cos(6.16454751785  + 629.84616305900 *T)
    sumX0  +=      14.422 * cos(1.11940784009  + 381.36842318480 *T)
    sumX0  +=      10.799 * cos(0.67317304481  + 1148.49142788971 *T)
    sumX0  +=      10.862 * cos(4.63986010898  + 248.96763557360 *T)
    sumX0  +=      10.584 * cos(4.42476612376  + 430.02340209721 *T)
    sumX0  +=      14.240 * cos(4.58422194293  + 1173.76422215520 *T)
    sumX0  +=      10.278 * cos(3.13159900994  + 209.87388421679 *T)
    sumX0  +=      11.280 * cos(5.48473876275  + 46.71424039951 *T)
    sumX0  +=      10.327 * cos(1.24113581631  + 606.20085385370 *T)
    sumX0  +=      13.270 * cos(4.37719157411  + 1041.46650040800 *T)
    sumX0  +=      13.786 * cos(4.55090300929  + 337.97632814250 *T)
    sumX0  +=      10.252 * cos(0.92652948440  + 1891.48148842230 *T)
    sumX0  +=      13.304 * cos(5.56269483757  + 63.49208081989 *T)
    sumX0  +=       9.721 * cos(0.53326540754  + 1159.53712815361 *T)
    sumX0  +=       9.974 * cos(4.64682563136  + 9.80504503910 *T)
    sumX0  +=      10.564 * cos(3.20765136591  + 1251.09656714130 *T)
    sumX0  +=      12.937 * cos(0.98121958090  + 944.73900579230 *T)
    sumX0  +=      12.666 * cos(6.06707633877  + 721.89323701370 *T)
    sumX0  +=       9.141 * cos(1.74908467583  + 1780.78753223589 *T)
    sumX0  +=       9.377 * cos(1.95219143915  + 2523.77759276850 *T)
    sumX0  +=       8.992 * cos(2.62152527313  + 1165.89991562901 *T)
    sumX0  +=       9.074 * cos(3.84732904999  + 754.27957828001 *T)
    sumX0  +=       8.886 * cos(2.45050457262  + 528.71094230071 *T)
    sumX0  +=       8.886 * cos(1.89018107575  + 531.15862285550 *T)
    sumX0  +=       8.815 * cos(2.08073903214  + 302.40859313851 *T)
    sumX0  +=      12.114 * cos(4.85216986598  + 1020.26887175750 *T)
    sumX0  +=       8.872 * cos(1.62855962032  + 194.89603068979 *T)
    sumX0  +=      10.983 * cos(5.00727208394  + 1.24065522479 *T)
    sumX0  +=      12.190 * cos(6.14672657104  + 72.31710306511 *T)
    sumX0  +=       8.635 * cos(0.16436387127  + 1056.70498636590 *T)
    sumX0  +=      10.308 * cos(6.00917440923  + 2950.37578364450 *T)
    sumX0  +=      11.595 * cos(1.86026762055  + 1289.70268353110 *T)
    sumX0  +=       8.931 * cos(0.42213138869  + 956.04533848709 *T)
    sumX0  +=      10.790 * cos(3.70406181805  + 529.04101070080 *T)
    sumX0  +=      10.790 * cos(0.63662383032  + 530.82855445540 *T)
    sumX0  +=       8.192 * cos(3.78940462638  + 13.98327651810 *T)
    sumX0  +=       7.989 * cos(3.82641746989  + 1058.87498814731 *T)
    sumX0  +=      10.336 * cos(3.40289007586  + 73.54094334250 *T)
    sumX0  +=       8.394 * cos(2.58792157805  + 525.79018099390 *T)
    sumX0  +=      10.766 * cos(3.96400751980  + 2111.40649589409 *T)
    sumX0  +=       7.749 * cos(1.75600486693  + 522.07047605190 *T)
    sumX0  +=       8.309 * cos(1.75332287431  + 534.07938416230 *T)
    sumX0  +=      10.415 * cos(4.58881028655  + 319.81708087780 *T)
    sumX0  +=       9.365 * cos(4.84977681203  + 415.30867312860 *T)
    sumX0  +=       7.601 * cos(2.58490575206  + 537.79908910430 *T)
    sumX0  +=       8.635 * cos(4.87197749696  + 1062.54650897950 *T)
    sumX0  +=       7.859 * cos(0.38734357220  + 1065.04237364410 *T)
    sumX0  +=       7.602 * cos(0.36299136019  + 1063.81853336670 *T)
    sumX0  +=      10.315 * cos(3.38211998665  + 456.63765671910 *T)
    sumX0  +=       7.989 * cos(1.21335631002  + 1060.37650719810 *T)
    sumX0  +=       8.083 * cos(1.68073025462  + 533.37930087419 *T)
    sumX0  +=       7.529 * cos(1.15606058501  + 2847.28300942589 *T)
    sumX0  +=       8.490 * cos(5.93185778563  + 202.00957769060 *T)
    sumX0  +=       7.329 * cos(5.32490948993  + 148.32254190981 *T)
    sumX0  +=       7.187 * cos(4.26147498381  + 30.77567115350 *T)
    sumX0  +=       9.756 * cos(3.52626213949  + 2008.31372167550 *T)
    sumX0  +=       8.727 * cos(2.59290505793  + 1202.07539780650 *T)
    sumX0  +=       7.941 * cos(3.50505477575  + 2332.30677329990 *T)
    sumX0  +=       6.801 * cos(2.80122649992  + 1054.95988122670 *T)
    sumX0  +=       8.387 * cos(5.98467407510  + 2957.97696561230 *T)
    sumX0  +=       7.512 * cos(0.64381751060  + 952.60082019101 *T)
    sumX0  +=       7.552 * cos(0.20225772775  + 842.39449897160 *T)
    sumX0  +=       7.680 * cos(6.15861344699  + 334.79493440480 *T)
    sumX0  +=       9.231 * cos(0.70359932145  + 2854.88419139370 *T)
    sumX0  +=       8.791 * cos(1.10109597117  + 483.46435966210 *T)
    sumX0  +=       7.389 * cos(2.29761310414  + 202.49721265760 *T)
    sumX0  +=       6.847 * cos(2.25408688945  + 320.56784040320 *T)
    sumX0  +=       8.929 * cos(0.96089681844  + 603.23190843711 *T)
    sumX0  +=       6.520 * cos(0.54950232146  + 3053.46855786309 *T)
    sumX0  +=       6.721 * cos(0.46202136795  + 1912.82212942470 *T)
    sumX0  +=       8.859 * cos(3.82736358661  + 217.47506618460 *T)
    sumX0  +=       8.710 * cos(3.26439990653  + 576.40520549410 *T)
    sumX0  +=       7.535 * cos(3.00289317558  + 1061.11022038100 *T)
    sumX0  +=       6.826 * cos(2.18157664615  + 284.39235822571 *T)
    sumX0  +=       6.401 * cos(4.64009991006  + 117.07605073670 *T)
    sumX0  +=       6.884 * cos(5.25234540574  + 277.27881122490 *T)
    sumX0  +=       6.632 * cos(6.12206657880  + 2.93757625420 *T)
    sumX0  +=       7.529 * cos(1.69514413761  + 423.17297965479 *T)
    sumX0  +=       6.885 * cos(4.98621366157  + 946.23803271561 *T)
    sumX0  +=       6.200 * cos(5.72324814885  + 1272.43720814370 *T)
    sumX0  +=       6.730 * cos(2.20794164333  + 531.57931398080 *T)
    sumX0  +=       6.730 * cos(2.13274400504  + 528.29025117541 *T)
    sumX0  +=       6.259 * cos(5.74959705113  + 142.69346761730 *T)
    sumX0  +=       7.602 * cos(4.67235401511  + 1055.43296197871 *T)
    sumX0  +=       7.329 * cos(6.05862254648  + 5.97595764000 *T)
    sumX0  +=       5.901 * cos(1.20913339550  + 1251.58420210830 *T)
    sumX0  +=       6.013 * cos(0.39504567840  + 1035.24672528451 *T)
    sumX0  +=       5.953 * cos(4.50565796160  + 498.44221318910 *T)
    sumX0  +=       7.833 * cos(4.37166198972  + 68.90870679129 *T)
    sumX0  +=       6.343 * cos(2.74168510429  + 199.31581891990 *T)
    sumX0  +=       5.843 * cos(0.90437055622  + 1276.85699637381 *T)
    sumX0  +=       7.666 * cos(4.79082431872  + 247.48316286530 *T)
    sumX0  +=       5.663 * cos(4.73243855855  + 1091.86907658711 *T)
    sumX0  +=       5.605 * cos(2.83447163903  + 513.47245634280 *T)
    sumX0  +=       5.524 * cos(6.17355324978  + 2847.77064439291 *T)
    sumX0  +=       7.612 * cos(5.67631106398  + 352.06040979221 *T)
    sumX0  +=       5.914 * cos(5.25663173098  + 1054.20912170131 *T)
    sumX0  +=       7.627 * cos(1.64993486675  + 405.50136735710 *T)
    sumX0  +=       5.507 * cos(6.11802918315  + 561.42735196710 *T)
    sumX0  +=       5.497 * cos(5.97722134510  + 810.90192958261 *T)
    sumX0  +=       5.666 * cos(1.34586791176  + 1064.29161411871 *T)
    sumX0  +=       5.378 * cos(1.50621399976  + 546.39710881341 *T)
    sumX0  +=       6.319 * cos(4.76637123715  + 1493.33748554950 *T)
    sumX0  +=       6.876 * cos(0.37866209699  + 408.19512612779 *T)
    sumX0  +=       5.745 * cos(5.91871265537  + 430.28652665560 *T)
    sumX0  +=       6.791 * cos(4.05903259226  + 303.61787920090 *T)
    sumX0  +=       7.093 * cos(5.97150143560  + 740.05248427840 *T)
    sumX0  +=       6.815 * cos(3.11788100877  + 1053.99667338020 *T)
    sumX0  +=       5.111 * cos(4.19087253985  + 1176.94561589290 *T)
    sumX0  +=       5.065 * cos(4.11506105627  + 1059.46568897830 *T)
    sumX0  +=       6.697 * cos(2.00713771131  + 685.23011986920 *T)
    sumX0  +=       5.674 * cos(3.00549520381  + 1567.48843679789 *T)
    sumX0  +=       5.759 * cos(2.66366184710  + 166.58485503849 *T)
    sumX0  +=       4.933 * cos(5.71470503937  + 245.29860686890 *T)
    sumX0  +=       5.446 * cos(3.42551680794  + 757.46097201770 *T)
    sumX0  +=       5.053 * cos(1.65431850894  + 1678.18239298430 *T)
    sumX0  +=       5.999 * cos(4.09275084512  + 908.57807782981 *T)
    sumX0  +=       5.382 * cos(3.62538539878  + 512.66930719071 *T)
    sumX0  +=       4.943 * cos(1.68108509143  + 2840.16946242510 *T)
    sumX0  +=       5.065 * cos(0.92471272365  + 1059.78580636710 *T)
    sumX0  +=       5.075 * cos(3.24596774005  + 508.10650660869 *T)
    sumX0  +=       4.955 * cos(3.41906236842  + 1069.92068841121 *T)
    sumX0  +=       5.153 * cos(0.71530024343  + 547.20025796550 *T)
    sumX0  +=       4.607 * cos(0.55203312702  + 2634.47154895491 *T)
    sumX0  +=       4.755 * cos(2.46962174787  + 415.04804069769 *T)
    sumX0  +=       4.775 * cos(0.82589283410  + 1994.08662767389 *T)
    sumX0  +=       4.590 * cos(1.48644692945  + 2097.17940189250 *T)
    sumX0  +=       4.911 * cos(1.51731908389  + 512.52629483880 *T)
    sumX0  +=       4.553 * cos(3.05075479447  + 509.48791345300 *T)
    sumX0  +=       4.672 * cos(3.60789441008  + 234.88355392390 *T)
    sumX0  +=       4.578 * cos(0.86543502074  + 305.58998687620 *T)
    sumX0  +=       5.193 * cos(4.02461886486  + 320.08020543620 *T)
    sumX0  +=       6.202 * cos(4.38664626654  + 26088.14695905770 *T)
    sumX0  +=       5.852 * cos(3.60701404156  + 2427.79836555070 *T)
    sumX0  +=       4.324 * cos(1.28993084923  + 550.38165170320 *T)
    sumX0  +=       4.410 * cos(2.46921347430  + 1127.29379923920 *T)
    sumX0  +=       4.238 * cos(2.94796150753  + 92.29089143820 *T)
    sumX0  +=       4.881 * cos(4.47400985269  + 725.07463075140 *T)
    sumX0  +=       4.688 * cos(0.82987676346  + 504.80500066450 *T)
    sumX0  +=       4.432 * cos(4.78229538893  + 1490.15609181181 *T)
    sumX0  +=       4.144 * cos(3.28540589450  + 970.76006745570 *T)
    sumX0  +=       5.737 * cos(4.94804445005  + 707.80915536400 *T)
    sumX0  +=       4.256 * cos(4.52622797251  + 512.40867475980 *T)
    sumX0  +=       4.139 * cos(0.44147895230  + 470.13103753780 *T)
    sumX0  +=       4.547 * cos(2.09665176175  + 857.37235249860 *T)
    sumX0  +=       4.431 * cos(6.20712541641  + 519.90047427050 *T)
    sumX0  +=       5.723 * cos(0.80521226018  + 1065.25482196520 *T)
    sumX0  +=       4.278 * cos(2.08187064363  + 739.30172475301 *T)
    sumX0  +=       5.621 * cos(3.19131378401  + 280.72332952100 *T)
    sumX0  +=       5.007 * cos(6.09093261803  + 492.06237937121 *T)
    sumX0  +=       3.987 * cos(2.63450729840  + 448.18264936190 *T)
    sumX0  +=       4.559 * cos(5.79888016601  + 3266.76765330109 *T)
    sumX0  +=       4.072 * cos(5.22974977136  + 3061.06973983091 *T)
    sumX0  +=       4.072 * cos(6.09764298210  + 547.46089039640 *T)
    sumX0  +=       5.087 * cos(4.84521270183  + 114.64292439690 *T)
    sumX0  +=       4.392 * cos(3.51029744842  + 555.06456449170 *T)
    sumX0  +=       4.733 * cos(2.24217612600  + 1087.93692332400 *T)
    sumX0  +=       4.590 * cos(5.53010211558  + 209.12312469139 *T)
    sumX0  +=       4.061 * cos(2.41966232758  + 1578.27101250340 *T)
    sumX0  +=       4.043 * cos(4.88146457506  + 917.17609753890 *T)
    sumX0  +=       4.781 * cos(2.52650341453  + 355.99256305531 *T)
    sumX0  +=       3.840 * cos(1.91188325761  + 479.53220639900 *T)
    sumX0  +=       4.229 * cos(3.44950557661  + 202.76033721600 *T)
    sumX0  +=       5.079 * cos(2.53028620261  + 100.40127618030 *T)
    sumX0  +=       3.809 * cos(0.33163870572  + 63.97971578690 *T)
    sumX0  +=       3.869 * cos(0.90238558045  + 3163.67487908250 *T)
    sumX0  +=       4.615 * cos(5.43148824800  + 8.32057233081 *T)
    sumX0  +=       4.615 * cos(5.43148824800  + 7.83293736379 *T)
    sumX0  +=       4.628 * cos(4.53025832657  + 567.80718578500 *T)
    sumX0  +=       4.429 * cos(4.90564275884  + 1025.68549772890 *T)
    sumX0  +=       4.166 * cos(2.66734467932  + 1587.83224005900 *T)
    sumX0  +=       4.232 * cos(4.10293029058  + 2042.74160658630 *T)
    sumX0  +=       3.934 * cos(5.18155455773  + 1894.66288216001 *T)
    sumX0  +=       4.119 * cos(5.42073132205  + 782.59075393131 *T)
    sumX0  +=       4.305 * cos(4.04299412868  + 237.92193530970 *T)
    sumX0  +=       4.647 * cos(4.25285731978  + 294.80741117069 *T)
    sumX0  +=       3.571 * cos(3.65947756157  + 1080.96638867510 *T)
    sumX0  +=       3.908 * cos(2.16984161430  + 775.47720693050 *T)
    sumX0  +=       4.183 * cos(0.65593677894  + 3171.27606105030 *T)
    sumX0  +=       3.544 * cos(5.95475674739  + 849.50804597240 *T)
    sumX0  +=       4.668 * cos(2.13834164389  + 391.41728570740 *T)
    sumX0  +=       3.825 * cos(5.15730861384  + 1013.15532475670 *T)
    sumX0  +=       4.490 * cos(2.90215227898  + 1066.25165970650 *T)
    sumX0  +=       3.497 * cos(4.37215725203  + 1059.83819599381 *T)
    sumX0  +=       4.592 * cos(4.61227234774  + 105.29663728990 *T)
    sumX0  +=       3.507 * cos(3.90043589953  + 589.73852761840 *T)
    sumX0  +=       3.460 * cos(4.57095415442  + 210.36151918381 *T)
    sumX0  +=       3.472 * cos(2.42806893940  + 580.33735875721 *T)
    sumX0  +=       4.742 * cos(5.78282912740  + 1214.92108496379 *T)
    sumX0  +=       4.236 * cos(4.52262475548  + 458.59770230690 *T)
    sumX0  +=       4.571 * cos(3.28632828787  + 2324.70559133209 *T)
    sumX0  +=       3.680 * cos(2.78358155559  + 547.34327031741 *T)
    sumX0  +=       3.342 * cos(1.72110734407  + 611.68691579430 *T)
    sumX0  +=       3.327 * cos(2.23284112556  + 416.05943265400 *T)
    sumX0  +=       3.357 * cos(2.00040904789  + 102.89714084490 *T)
    sumX0  +=       3.597 * cos(5.41599821164  + 81.50831573269 *T)
    sumX0  +=       3.697 * cos(4.35495083583  + 326.93062787860 *T)
    sumX0  +=       3.305 * cos(0.91377513656  + 3068.18328683170 *T)
    sumX0  +=       3.332 * cos(5.10225181759  + 102.80077262530 *T)
    sumX0  +=       3.850 * cos(4.79682693839  + 223.10414047710 *T)
    sumX0  +=       3.389 * cos(5.97871270107  + 1050.06452011710 *T)
    sumX0  +=       3.693 * cos(3.87928741090  + 100.88891114730 *T)
    sumX0  +=       3.520 * cos(4.33292989715  + 355.50492808829 *T)
    sumX0  +=       3.747 * cos(1.69335109674  + 101.88574888859 *T)
    sumX0  +=       3.497 * cos(0.66761652789  + 1059.41329935160 *T)
    sumX0  +=       4.041 * cos(2.14362833723  + 80.75755620730 *T)
    sumX0  +=       3.215 * cos(0.53534058688  + 2207.87335807890 *T)
    sumX0  +=       3.350 * cos(5.33479503934  + 1077.78499493740 *T)
    sumX0  +=       3.628 * cos(6.13897925009  + 1567.97607176490 *T)
    sumX0  +=       4.303 * cos(2.23046220012  + 374.15181032000 *T)
    sumX0  +=       3.455 * cos(4.99170868278  + 489.09343395460 *T)
    sumX0  +=       4.402 * cos(4.47863845219  + 987.55246209111 *T)
    sumX0  +=       3.638 * cos(5.42646055999  + 103.81216458159 *T)
    sumX0  +=       3.112 * cos(4.34851347009  + 422.42222012939 *T)
    sumX0  +=       3.124 * cos(0.32512970919  + 3369.86042751970 *T)
    sumX0  +=       3.303 * cos(4.42204851380  + 539.96909088571 *T)
    sumX0  +=       3.803 * cos(2.07273212148  + 852.95256426850 *T)
    sumX0  +=       3.982 * cos(0.94890253623  + 986.32862181370 *T)
    sumX0  +=       3.666 * cos(3.00278414954  + 77.99436146740 *T)
    sumX0  +=       4.012 * cos(2.19807079078  + 668.45227944880 *T)
    sumX0  +=       3.922 * cos(2.56685919071  + 1215.40871993080 *T)
    sumX0  +=       3.967 * cos(2.90360763886  + 486.64575339981 *T)
    sumX0  +=       3.197 * cos(2.07944029870  + 1169.34443392509 *T)
    sumX0  +=       3.122 * cos(5.39978748620  + 69.39634175830 *T)
    sumX0  +=       3.954 * cos(3.61115285346  + 1123.36164597610 *T)
    sumX0  +=       3.201 * cos(3.66652395167  + 145.87486135500 *T)
    sumX0  +=       3.935 * cos(1.80977709754  + 306.58682461749 *T)
    sumX0  +=       2.958 * cos(4.65595664577  + 531.94052014820 *T)
    sumX0  +=       2.958 * cos(5.96791430978  + 527.92904500800 *T)
    sumX0  +=       3.054 * cos(2.21604720422  + 654.36819779910 *T)
    sumX0  +=       2.909 * cos(5.97782239901  + 536.93645499430 *T)
    sumX0  +=       2.909 * cos(4.64604855654  + 522.93311016190 *T)
    sumX0  +=       3.738 * cos(5.20971126454  + 977.73060210460 *T)
    sumX0  +=       3.893 * cos(5.83257009844  + 812.38640229090 *T)
    sumX0  +=       3.604 * cos(5.17763880188  + 78.46993434690 *T)
    sumX0  +=       2.867 * cos(1.07832422526  + 1269.74344937300 *T)
    sumX0  +=       3.118 * cos(5.63216227640  + 570.77613120160 *T)
    sumX0  +=       3.676 * cos(2.63995604671  + 226.07308589371 *T)
    sumX0  +=       3.463 * cos(4.17288178571  + 1591.76439332211 *T)
    sumX0  +=       2.870 * cos(1.83388013858  + 1586.86903221250 *T)
    sumX0  +=       2.889 * cos(5.88273410848  + 504.66198831261 *T)
    sumX0  +=       3.107 * cos(4.97048599434  + 1905.22094745690 *T)
    sumX0  +=       2.833 * cos(5.34654613316  + 911.54702324640 *T)
    sumX0  +=       3.315 * cos(4.08289970624  + 1065.84552279621 *T)
    sumX0  +=       3.556 * cos(1.43723870007  + 573.22381175640 *T)
    sumX0  +=       2.918 * cos(1.02389443723  + 967.57867371801 *T)
    sumX0  +=       2.676 * cos(1.60354882374  + 3281.48238226970 *T)
    sumX0  +=       3.689 * cos(3.27567307401  + 850.25880549781 *T)
    sumX0  +=       3.016 * cos(5.84238370205  + 3274.36883526891 *T)
    sumX0  +=       3.204 * cos(3.20846751019  + 2744.19023520730 *T)
    sumX0  +=       2.667 * cos(0.77811362765  + 97.91996573070 *T)
    sumX0  +=       2.647 * cos(2.45257224390  + 134.82916109110 *T)
    sumX0  +=       2.715 * cos(4.74043154929  + 555.20757684360 *T)
    sumX0  +=       3.135 * cos(1.58908870559  + 433.46792039329 *T)
    sumX0  +=       2.639 * cos(3.83183781056  + 144.39038864671 *T)
    sumX0  +=       2.912 * cos(2.70342002444  + 1883.88030645450 *T)
    sumX0  +=       3.601 * cos(2.04225476495  + 991.95769610620 *T)
    sumX0  +=       2.745 * cos(5.11304244810  + 33.69643246030 *T)
    sumX0  +=       2.589 * cos(5.15515901130  + 93.28772917949 *T)
    sumX0  +=       2.531 * cos(3.52886647530  + 108.00245814810 *T)
    sumX0  +=       2.821 * cos(2.63736215369  + 503.10807963510 *T)
    sumX0  +=       2.585 * cos(1.42843919503  + 3156.56133208170 *T)
    sumX0  +=       3.142 * cos(1.28656971724  + 520.37355502251 *T)
    sumX0  +=       3.142 * cos(6.19570189655  + 539.49601013370 *T)
    sumX0  +=       2.731 * cos(4.37970098385  + 537.16020416351 *T)
    sumX0  +=       2.731 * cos(6.24416997170  + 522.70936099270 *T)
    sumX0  +=       3.288 * cos(2.11811777567  + 685.71775483620 *T)
    sumX0  +=       2.456 * cos(3.04358904051  + 1074.60360119970 *T)
    sumX0  +=       2.382 * cos(5.50759187807  + 112.89781925770 *T)
    sumX0  +=       2.385 * cos(2.46238935625  + 1475.17823828480 *T)
    sumX0  +=       2.456 * cos(1.99618473941  + 1044.64789414570 *T)
    sumX0  +=       2.435 * cos(0.75544262993  + 824.98601123230 *T)
    sumX0  +=       2.637 * cos(1.70331908888  + 556.76148552110 *T)
    sumX0  +=       3.031 * cos(5.67687846191  + 538.53280228720 *T)
    sumX0  +=       3.031 * cos(4.94699249364  + 521.33676286901 *T)
    sumX0  +=       2.342 * cos(0.06228259686  + 518.40144734719 *T)
    sumX0  +=       2.443 * cos(1.03845059435  + 749.85979004990 *T)
    sumX0  +=       3.226 * cos(4.44994968410  + 536.83588125780 *T)
    sumX0  +=       3.226 * cos(6.17392127144  + 523.03368389840 *T)
    sumX0  +=       2.356 * cos(4.85533213884  + 285.61619850311 *T)
    sumX0  +=       2.452 * cos(2.19107576775  + 1090.64523630970 *T)
    sumX0  +=       2.986 * cos(4.62414537014  + 491.31411197330 *T)
    sumX0  +=       2.323 * cos(0.16164499756  + 530.56792202451 *T)
    sumX0  +=       2.323 * cos(4.17904065081  + 529.30164313170 *T)
    sumX0  +=       3.003 * cos(3.08627097096  + 108.96566599461 *T)
    sumX0  +=       3.049 * cos(4.48513403718  + 1438.26904292440 *T)
    sumX0  +=       3.159 * cos(0.76092054251  + 457.86149699650 *T)
    sumX0  +=       3.023 * cos(1.39945420180  + 532.90372799470 *T)
    sumX0  +=       3.023 * cos(2.94123144657  + 526.96583716151 *T)
    sumX0  +=       2.581 * cos(5.61581142863  + 1483.04254481100 *T)
    sumX0  +=       2.851 * cos(4.62796468589  + 731.70054278519 *T)
    sumX0  +=       2.627 * cos(3.80975501995  + 1095.05047032480 *T)
    sumX0  +=       2.198 * cos(1.26120223291  + 2413.57127154910 *T)
    sumX0  +=       2.533 * cos(1.67483276763  + 12566.39551746630 *T)
    sumX0  +=       2.183 * cos(6.04766236935  + 2421.17245351690 *T)
    sumX0  +=       2.273 * cos(1.32684908668  + 1589.36489687710 *T)
    sumX0  +=       2.201 * cos(5.72550531705  + 2044.22607929460 *T)
    sumX0  +=       2.304 * cos(4.41348832810  + 1053.47540851841 *T)
    sumX0  +=       2.878 * cos(2.85936483769  + 28.55499313480 *T)
    sumX0  +=       2.523 * cos(6.18014679261  + 909.57491557110 *T)
    sumX0  +=       2.847 * cos(5.46944309481  + 1441.23798834100 *T)
    sumX0  +=       2.211 * cos(0.29447557006  + 376.43943218040 *T)
    sumX0  +=       2.767 * cos(1.11249816364  + 39.86132582961 *T)
    sumX0  +=       2.242 * cos(0.27111162703  + 6681.46867088311 *T)
    sumX0  +=       2.284 * cos(3.14190849220  + 902.94900353731 *T)
    sumX0  +=       2.122 * cos(2.81881363812  + 543.67424161270 *T)
    sumX0  +=       2.251 * cos(3.73285596763  + 1067.70250252000 *T)
    sumX0  +=       2.782 * cos(3.62990010057  + 451.69660362719 *T)
    sumX0  +=       2.107 * cos(2.78853830817  + 824.49837626530 *T)
    sumX0  +=       2.308 * cos(6.27924595004  + 732.93893727761 *T)
    sumX0  +=       2.055 * cos(4.34456901846  + 526.26575387340 *T)
    sumX0  +=       2.624 * cos(3.68204703869  + 1261.87914284680 *T)
    sumX0  +=       2.059 * cos(5.93579199305  + 1254.76559584600 *T)
    sumX0  +=       2.523 * cos(5.46151498365  + 372.77040347569 *T)
    sumX0  +=       2.035 * cos(5.20593524984  + 535.77630519170 *T)
    sumX0  +=       2.035 * cos(5.41793570571  + 524.09325996451 *T)
    sumX0  +=       2.005 * cos(3.97142686370  + 843.14525849700 *T)
    sumX0  +=       2.366 * cos(2.81479040414  + 3708.08057314570 *T)
    sumX0  +=       1.959 * cos(2.33278855728  + 163.14033674240 *T)
    sumX0  +=       2.451 * cos(1.41485691262  + 739.56484931140 *T)
    sumX0  +=       2.474 * cos(2.34380492610  + 2122.93983112501 *T)
    sumX0  +=       1.935 * cos(5.95122170645  + 3164.16251404950 *T)
    sumX0  +=       2.548 * cos(2.30031435926  + 1128.77827194750 *T)
    sumX0  +=       2.140 * cos(5.32310438264  + 408.68276109481 *T)
    sumX0  +=       2.089 * cos(0.70684572014  + 60.04756252381 *T)
    sumX0  +=       2.555 * cos(5.75500802854  + 231.80522605020 *T)
    sumX0  +=       2.289 * cos(4.10142481033  + 429.28968891431 *T)
    sumX0  +=       2.652 * cos(3.24034861472  + 405.01373239010 *T)
    sumX0  +=       2.044 * cos(3.27128458744  + 241.36645360580 *T)
    sumX0  +=       1.959 * cos(0.64242562063  + 3480.06674873909 *T)
    sumX0  +=       1.890 * cos(1.30374554288  + 501.48059457490 *T)
    sumX0  +=       2.361 * cos(2.60496924654  + 1819.88128359270 *T)
    sumX0  +=       2.290 * cos(2.07662910000  + 386.99749747730 *T)
    sumX0  +=       1.839 * cos(1.47204895914  + 511.28790034640 *T)
    sumX0  +=       1.873 * cos(2.16131201791  + 1209.18894480731 *T)
    sumX0  +=       1.837 * cos(0.13057602825  + 111.41334654940 *T)
    sumX0  +=       2.510 * cos(3.56540915410  + 602.00806815971 *T)
    sumX0  +=       2.186 * cos(5.05661256477  + 1553.89199011520 *T)
    sumX0  +=       1.875 * cos(4.96409845324  + 1590.27992061380 *T)
    sumX0  +=       2.325 * cos(2.99238523654  + 488.88098563350 *T)
    sumX0  +=       1.960 * cos(2.67757828738  + 109.48693085641 *T)
    sumX0  +=       2.511 * cos(1.33963728695  + 167.07249000551 *T)
    sumX0  +=       2.012 * cos(5.57171500563  + 3583.15952295770 *T)
    sumX0  +=       2.198 * cos(2.81677718687  + 297.98880490840 *T)
    sumX0  +=       2.129 * cos(0.30008608118  + 821.94762984651 *T)
    sumX0  +=       2.219 * cos(3.10449171167  + 2641.09746098870 *T)
    sumX0  +=       1.954 * cos(0.78506313670  + 1415.37449324450 *T)
    sumX0  +=       1.787 * cos(5.23586420667  + 1059.88638010361 *T)
    sumX0  +=       1.933 * cos(3.60934798531  + 1688.47733372280 *T)
    sumX0  +=       2.173 * cos(0.53973434031  + 1065.77608682700 *T)
    sumX0  +=       2.273 * cos(4.41222326623  + 1589.26852865751 *T)
    sumX0  +=       2.311 * cos(0.11223855924  + 36.89238041300 *T)
    sumX0  +=       2.005 * cos(5.77162207371  + 746.67839631219 *T)
    sumX0  +=       1.716 * cos(4.22634329515  + 504.07128748160 *T)
    sumX0  +=       1.970 * cos(4.25083421565  + 363.10611005610 *T)
    sumX0  +=       1.879 * cos(1.53326034532  + 1994.57426264090 *T)
    sumX0  +=       1.715 * cos(0.40254693953  + 2950.86341861150 *T)
    sumX0  +=       1.859 * cos(4.76472866777  + 2221.61281711350 *T)
    sumX0  +=       1.760 * cos(5.81144237350  + 1051.54899282540 *T)
    sumX0  +=       1.679 * cos(0.73859095289  + 2310.47849733050 *T)
    sumX0  +=       2.127 * cos(5.79555857294  + 945.22664075931 *T)
    sumX0  +=       1.827 * cos(2.98105578454  + 110.49832281271 *T)
    sumX0  +=       1.954 * cos(0.06688544879  + 1446.86706263350 *T)
    sumX0  +=       2.077 * cos(0.12761917926  + 724.58699578440 *T)
    sumX0  +=       2.096 * cos(1.34829485372  + 570.98857952271 *T)
    sumX0  +=       1.615 * cos(0.93368590926  + 1052.56038478170 *T)
    sumX0  +=       1.608 * cos(0.11434235358  + 555.79827767460 *T)
    sumX0  +=       1.681 * cos(5.56771625121  + 516.44140175940 *T)
    sumX0  +=       1.681 * cos(5.05615470433  + 543.42816339680 *T)
    sumX0  +=       1.618 * cos(2.03312852847  + 112.41018429070 *T)
    sumX0  +=       1.962 * cos(5.75393639029  + 836.76542467910 *T)
    sumX0  +=       1.962 * cos(2.61234373671  + 836.27778971210 *T)
    sumX0  +=       1.620 * cos(4.04601847199  + 683.43013297580 *T)
    sumX0  +=       1.551 * cos(2.34280930104  + 593.67068088151 *T)
    sumX0  +=       1.819 * cos(2.88219564243  + 3060.58210486390 *T)
    sumX0  +=       1.634 * cos(4.46994256209  + 56.37853381910 *T)
    sumX0  +=       1.791 * cos(0.10409892110  + 502.84744720420 *T)
    sumX0  +=       2.008 * cos(2.52248957084  + 487.39651292520 *T)
    sumX0  +=       1.551 * cos(6.21389515122  + 3686.25229717630 *T)
    sumX0  +=       1.511 * cos(1.75866420206  + 1518.46726746311 *T)
    sumX0  +=       1.910 * cos(1.68897371752  + 228.83628063359 *T)
    sumX0  +=       1.738 * cos(0.70952679097  + 1226.45442019471 *T)
    sumX0  +=       1.471 * cos(6.05957994367  + 350.57593708390 *T)
    sumX0  +=       1.466 * cos(4.96867723168  + 426.79382424970 *T)
    sumX0  +=       1.636 * cos(2.43292318068  + 2200.27217611110 *T)
    sumX0  +=       1.940 * cos(3.42777184231  + 1432.63996863190 *T)
    sumX0  +=       1.484 * cos(3.98627195351  + 995.88984936931 *T)
    sumX0  +=       1.673 * cos(4.23658671292  + 557.02211795201 *T)
    sumX0  +=       2.009 * cos(3.88072147559  + 417.28078080390 *T)
    sumX0  +=       1.573 * cos(1.54985228466  + 425.87880051300 *T)
    sumX0  +=       1.562 * cos(1.17133703517  + 416.54706762100 *T)
    sumX0  +=       1.419 * cos(0.22434479279  + 216.72430665921 *T)
    sumX0  +=       1.787 * cos(6.08709488043  + 1059.36511524180 *T)
    sumX0  +=       1.600 * cos(3.53298161884  + 1268.99268984760 *T)
    sumX0  +=       1.459 * cos(1.86682228155  + 426.89019246930 *T)
    sumX0  +=       1.860 * cos(0.59126369009  + 1160.27084133651 *T)
    sumX0  +=       1.409 * cos(4.21180105945  + 383.81610373959 *T)
    sumX0  +=       1.599 * cos(0.99304321709  + 1588.35350492080 *T)
    sumX0  +=       1.385 * cos(1.96310882827  + 466.19888427470 *T)
    sumX0  +=       1.645 * cos(0.25601088543  + 630.57987624190 *T)
    sumX0  +=       1.867 * cos(3.30529498913  + 1744.61205005840 *T)
    sumX0  +=       1.796 * cos(2.27749593221  + 424.39432780471 *T)
    sumX0  +=       1.802 * cos(1.81980262725  + 572.47305223100 *T)
    sumX0  +=       1.563 * cos(5.91937041266  + 312.21589891000 *T)
    sumX0  +=       1.811 * cos(1.18645718578  + 1340.59289467720 *T)
    sumX0  +=       1.342 * cos(2.52735020510  + 849.77117053080 *T)
    sumX0  +=       1.409 * cos(1.19028163106  + 3472.95320173830 *T)
    sumX0  +=       1.340 * cos(5.68788468184  + 1058.35372328550 *T)
    sumX0  +=       1.342 * cos(5.03842872086  + 3377.46160948750 *T)
    sumX0  +=       1.764 * cos(4.18037485648  + 312.96665843540 *T)
    sumX0  +=       1.801 * cos(1.64093908538  + 833.79647926250 *T)
    sumX0  +=       1.801 * cos(1.64093908538  + 833.30884429549 *T)
    sumX0  +=       1.504 * cos(3.98849552736  + 1052.46401656211 *T)
    sumX0  +=       1.346 * cos(0.84106067159  + 1066.78747878330 *T)
    sumX0  +=       1.711 * cos(2.11130834468  + 1363.48744435711 *T)
    sumX0  +=       1.382 * cos(3.43383168337  + 24.13520490470 *T)
    sumX0  +=       1.636 * cos(0.25591141604  + 422.64922266551 *T)
    sumX0  +=       1.714 * cos(2.66553451397  + 1165.16620244611 *T)
    sumX0  +=       1.547 * cos(0.43253240963  + 1097.75878331050 *T)
    sumX0  +=       1.706 * cos(0.40870999807  + 1696.34164024900 *T)
    sumX0  +=       1.293 * cos(2.98652787499  + 394.59867944510 *T)
    sumX0  +=       1.361 * cos(0.44504530774  + 1579.02177202880 *T)
    sumX0  +=       1.329 * cos(2.37930563293  + 1628.67358868250 *T)
    sumX0  +=       1.657 * cos(2.95697828072  + 473.31243127551 *T)
    sumX0  +=       1.342 * cos(0.89798927248  + 827.67977000299 *T)
    sumX0  +=       1.353 * cos(4.16385577520  + 2499.13544582190 *T)
    sumX0  +=       1.322 * cos(4.93972524369  + 1464.88329754630 *T)
    sumX0  +=       1.528 * cos(0.29983648075  + 138.76131435421 *T)
    sumX0  +=       1.507 * cos(4.31558459369  + 774.98957196349 *T)
    sumX0  +=       1.375 * cos(3.98046179765  + 1093.56599761650 *T)
    sumX0  +=       1.393 * cos(0.72013842017  + 913.99470380120 *T)
    sumX0  +=       1.302 * cos(1.30247879244  + 494.72250824710 *T)
    sumX0  +=       1.266 * cos(0.05484214471  + 1162.76670600110 *T)
    sumX0  +=       1.602 * cos(3.97091760080  + 611.19928082730 *T)
    sumX0  +=       1.243 * cos(5.45126497249  + 561.69047652550 *T)
    sumX0  +=       1.234 * cos(1.89694179274  + 436.40323591510 *T)
    sumX0  +=       1.258 * cos(3.15497421074  + 1162.67033778151 *T)
    sumX0  +=       1.490 * cos(4.44362414485  + 1468.06469128401 *T)
    sumX0  +=       1.419 * cos(3.11151373878  + 1138.33949950310 *T)
    sumX0  +=       1.394 * cos(0.44413366244  + 958.98065400890 *T)
    sumX0  +=       1.290 * cos(5.15722188790  + 990.47322339791 *T)
    sumX0  +=       1.199 * cos(0.62664940478  + 414.31183538731 *T)
    sumX0  +=       1.364 * cos(3.81523518368  + 252.89978883670 *T)
    sumX0  +=       1.494 * cos(1.51661927481  + 1485.73630358170 *T)
    sumX0  +=       1.367 * cos(1.47066835723  + 527.96267490280 *T)
    sumX0  +=       1.464 * cos(3.90911748944  + 1066.69111056371 *T)
    sumX0  +=       1.248 * cos(3.01032171032  + 9.31741007210 *T)
    sumX0  +=       1.194 * cos(5.78008609111  + 774.25336665311 *T)
    sumX0  +=       1.328 * cos(2.33133276999  + 482.97672469510 *T)
    sumX0  +=       1.176 * cos(5.65248871466  + 633.07574090651 *T)
    sumX0  +=       1.352 * cos(0.11794612755  + 696.76345510010 *T)
    sumX0  +=       1.235 * cos(1.29434837440  + 1027.38241875830 *T)
    sumX0  +=       1.192 * cos(3.03820680249  + 565.14705690910 *T)
    sumX0  +=       1.337 * cos(5.18331371440  + 427.80521620600 *T)
    sumX0  +=       1.581 * cos(0.42492742452  + 642.58878435231 *T)
    sumX0  +=       1.186 * cos(4.67061724657  + 2115.07552459880 *T)
    sumX0  +=       1.339 * cos(6.03072294407  + 1161.75531404480 *T)
    sumX0  +=       1.260 * cos(2.78492021778  + 633.99076464320 *T)
    sumX0  +=       1.309 * cos(3.48082829130  + 1163.68172973780 *T)
    sumX0  +=       1.221 * cos(4.50279420601  + 2538.00468677009 *T)
    sumX0  +=       1.133 * cos(4.12042287083  + 738.81408978600 *T)
    sumX0  +=       1.172 * cos(2.47236884204  + 632.97937268690 *T)
    sumX0  +=       1.120 * cos(4.59706955368  + 709.29362807231 *T)
    sumX0  +=       1.340 * cos(5.63507440526  + 1060.89777205990 *T)
    sumX0  +=       1.126 * cos(1.12456010627  + 2729.96314120569 *T)
    sumX0  +=       1.417 * cos(0.94598908604  + 1364.48428209840 *T)
    sumX0  +=       1.112 * cos(5.29416320994  + 1066.21802981170 *T)
    sumX0  +=       1.312 * cos(1.35334544403  + 981.87520368880 *T)
    sumX0  +=       1.496 * cos(5.90360357966  + 388.70897272171 *T)
    sumX0  +=       1.423 * cos(1.37785950879  + 586.55713388070 *T)
    sumX0  +=       1.221 * cos(3.63981220961  + 269.67762925710 *T)
    sumX0  +=       1.088 * cos(1.29669383617  + 1280.03839011150 *T)
    sumX0  +=       1.089 * cos(5.17675407458  + 946.97174589851 *T)
    sumX0  +=       1.132 * cos(0.51083517505  + 1664.09831133461 *T)
    sumX0  +=       1.433 * cos(2.05202099214  + 635.47523735150 *T)
    sumX0  +=       1.427 * cos(2.97093867648  + 597.11519917759 *T)
    sumX0  +=       1.389 * cos(4.08780249338  + 636.47207509279 *T)
    sumX0  +=       1.191 * cos(5.09661092725  + 1887.54933515920 *T)
    sumX0  +=       1.420 * cos(3.40607940834  + 1140.13948639650 *T)
    sumX0  +=       1.420 * cos(0.26448675475  + 1140.62712136350 *T)
    sumX0  +=       1.082 * cos(0.52075379881  + 915.47917650950 *T)
    sumX0  +=       1.134 * cos(0.14574453315  + 102.32769187330 *T)
    sumX0  +=       1.061 * cos(1.90789304117  + 925.04040406510 *T)
    sumX0  +=       1.302 * cos(4.07370610626  + 637.22034249070 *T)
    sumX0  +=       1.036 * cos(2.10557063879  + 1798.68365494221 *T)
    sumX0  +=       1.210 * cos(1.65592259758  + 31.73638687251 *T)
    sumX0  +=       1.210 * cos(1.65592259758  + 31.24875190549 *T)
    sumX0  +=       1.439 * cos(3.00324185430  + 244.07476659150 *T)
    sumX0  +=       1.020 * cos(2.05257171773  + 732.45130231059 *T)
    sumX0  +=       1.025 * cos(0.81896733353  + 951.86710700811 *T)
    sumX0  +=       1.018 * cos(2.44176419905  + 623.46632924110 *T)
    sumX0  +=       1.027 * cos(0.61184452819  + 1474.91760585390 *T)
    sumX0  +=       1.114 * cos(0.64613208093  + 103.37022159689 *T)
    sumX0  +=       1.040 * cos(2.40824650249  + 314.18800658530 *T)
    sumX0  +=       1.371 * cos(1.13698163960  + 4590.66636300550 *T)
    sumX0  +=       1.256 * cos(3.34855371696  + 73.05330837550 *T)
    sumX0  +=       1.059 * cos(5.69757951835  + 3590.76070492550 *T)
    sumX0  +=       1.008 * cos(2.53682498993  + 1140.89024592190 *T)
    sumX0  +=       1.288 * cos(5.39489850289  + 2004.60857094851 *T)
    sumX0  +=       1.033 * cos(2.65742933066  + 1028.60625903570 *T)
    sumX0  +=       1.085 * cos(3.78966042020  + 395.34943897051 *T)
    sumX0  +=       0.971 * cos(1.34946807177  + 665.27088571111 *T)
    sumX0  +=       1.358 * cos(4.62268264011  + 448.93340888730 *T)
    sumX0  +=       0.980 * cos(3.55994269076  + 961.94959942551 *T)
    sumX0  +=       0.994 * cos(0.40039858601  + 3796.45861839570 *T)
    sumX0  +=       1.202 * cos(5.54033969791  + 453.66871130251 *T)
    sumX0  +=       1.028 * cos(0.48891270332  + 2524.26522773550 *T)
    sumX0  +=       1.040 * cos(5.47890663354  + 170.51700830159 *T)
    sumX0  +=       0.955 * cos(3.71242781522  + 645.55772976890 *T)
    sumX0  +=       1.241 * cos(4.39110413012  + 1048.09241244180 *T)
    sumX0  +=       1.076 * cos(2.97284479790  + 2957.48933064529 *T)
    sumX0  +=       1.230 * cos(2.33341924042  + 1354.67697632690 *T)
    sumX0  +=       1.240 * cos(5.94758800259  + 610.93615626891 *T)
    sumX0  +=       1.240 * cos(5.94758800259  + 610.44852130189 *T)
    sumX0  +=       1.022 * cos(3.69588867080  + 276.79117625789 *T)
    sumX0  +=       0.924 * cos(2.41983440650  + 842.65762352999 *T)
    sumX0  +=       1.083 * cos(1.92054379034  + 1901.77642916080 *T)
    sumX0  +=       1.243 * cos(4.71980053276  + 671.16059243450 *T)
    sumX0  +=       1.013 * cos(2.54761722905  + 3376.97397452050 *T)
    sumX0  +=       1.100 * cos(5.92802097748  + 1053.03346553370 *T)
    sumX0  +=       1.099 * cos(2.80854803975  + 611.82992814620 *T)
    sumX0  +=       1.099 * cos(4.67370721199  + 448.03963701001 *T)
    sumX0  +=       1.033 * cos(5.44055724404  + 632.06434895021 *T)
    sumX0  +=       0.957 * cos(3.89598622213  + 2965.09051261311 *T)
    sumX0  +=       1.030 * cos(3.34763414892  + 1141.37788088890 *T)
    sumX0  +=       1.196 * cos(2.56247891942  + 8520.87605922509 *T)
    sumX0  +=       0.967 * cos(0.19420587029  + 829.06117684730 *T)
    sumX0  +=       1.166 * cos(1.68067516200  + 1511.56616878340 *T)
    sumX0  +=       1.147 * cos(5.62238808631  + 1390.24471133090 *T)
    sumX0  +=       0.873 * cos(1.78825991988  + 322.26476143261 *T)
    sumX0  +=       1.137 * cos(1.25214902883  + 950.90389916161 *T)
    sumX0  +=       1.201 * cos(2.52928106613  + 1436.29693524910 *T)
    sumX0  +=       0.903 * cos(0.04697695828  + 1475.92899781020 *T)
    sumX0  +=       0.858 * cos(5.65232336931  + 3480.55438370610 *T)
    sumX0  +=       1.061 * cos(5.43420240433  + 1014.37916503410 *T)
    sumX0  +=       0.852 * cos(3.85820066843  + 1305.16817202511 *T)
    sumX0  +=       1.121 * cos(3.79729993404  + 2861.99773839450 *T)
    sumX0  +=       0.978 * cos(1.58128047477  + 1262.62990237220 *T)
    sumX0  +=       0.918 * cos(3.86741534242  + 341.01470952831 *T)
    sumX0  +=       0.886 * cos(2.01115688146  + 1482.29178528560 *T)
    sumX0  +=       1.153 * cos(1.03307352866  + 1055.95671896799 *T)
    sumX0  +=       1.011 * cos(2.13326219711  + 1379.94977059240 *T)
    sumX0  +=       0.844 * cos(4.80951949791  + 231.70216018621 *T)
    sumX0  +=       0.939 * cos(4.90771963793  + 1703.45518724980 *T)
    sumX0  +=       1.029 * cos(2.73261618718  + 2545.60586873790 *T)
    sumX0  +=       1.042 * cos(0.93684376629  + 297.88573904440 *T)
    sumX0  +=       1.170 * cos(1.11117592168  + 46.22660543249 *T)
    sumX0  +=       0.962 * cos(3.48479854663  + 37.88921815429 *T)
    sumX0  +=       0.877 * cos(2.57722816875  + 357.68948408471 *T)
    sumX0  +=       0.988 * cos(3.77368109282  + 7211.15963597770 *T)
    sumX0  +=       0.892 * cos(5.41512877404  + 1571.15746550260 *T)
    sumX0  +=       1.109 * cos(1.10719434979  + 348.12825652910 *T)
    sumX0  +=       0.827 * cos(1.26246251342  + 20956.50587500010 *T)
    sumX0  +=       0.934 * cos(3.64280443414  + 525.51499434800 *T)
    sumX0  +=       0.914 * cos(1.85188487603  + 1031.31457202140 *T)
    sumX0  +=       0.998 * cos(3.87130911943  + 1505.93709449090 *T)
    sumX0  +=       1.003 * cos(5.35180604374  + 1681.36378672201 *T)
    sumX0  +=       1.043 * cos(1.20439768274  + 3384.57515648830 *T)
    sumX0  +=       1.133 * cos(1.03655442301  + 1357.85837006460 *T)
    sumX0  +=       1.118 * cos(3.02473429596  + 892.79707515070 *T)
    sumX0  +=       0.818 * cos(3.19817529862  + 1153.15729433570 *T)
    sumX0  +=       0.883 * cos(4.94847300730  + 533.45790092711 *T)
    sumX0  +=       0.883 * cos(5.67539794824  + 526.41166422910 *T)
    sumX0  +=       0.966 * cos(0.88110178776  + 381.85605815180 *T)
    sumX0  +=       1.014 * cos(1.05859637798  + 50.64639366260 *T)
    sumX0  +=       1.014 * cos(1.05859637798  + 50.15875869559 *T)
    sumX0  +=       0.977 * cos(3.07712828393  + 25035.39094090229 *T)
    sumX0  +=       0.933 * cos(3.89970622027  + 671.89679774490 *T)
    sumX0  +=       0.898 * cos(2.17511781419  + 2516.66404576769 *T)
    sumX0  +=       0.787 * cos(4.25746871915  + 1056.91743468700 *T)
    sumX0  +=       0.788 * cos(6.10879216964  + 4002.64416683289 *T)
    sumX0  +=       0.950 * cos(2.54359450579  + 532.62854134880 *T)
    sumX0  +=       0.839 * cos(3.46904179365  + 4686.64559022330 *T)
    sumX0  +=       0.965 * cos(0.47908461253  + 1520.16418849250 *T)
    sumX0  +=       1.017 * cos(3.23008830093  + 1020.00823932660 *T)
    sumX0  +=       0.820 * cos(0.33194380569  + 348.39138108750 *T)
    sumX0  +=       0.786 * cos(3.36629143378  + 178.11819026941 *T)
    sumX0  +=       0.929 * cos(2.33109721672  + 1130.47519297691 *T)
    sumX0  +=       0.779 * cos(6.05972875769  + 224.10097821840 *T)
    sumX0  +=       0.951 * cos(0.17977055776  + 25021.16384690069 *T)
    sumX0  +=       0.917 * cos(5.17636386869  + 3899.55139261430 *T)
    sumX0  +=       0.763 * cos(6.06912796746  + 432.25863433090 *T)
    sumX0  +=       0.841 * cos(3.14888189898  + 643.32249753521 *T)
    sumX0  +=       0.782 * cos(2.23847303429  + 1731.76636290111 *T)
    sumX0  +=       0.945 * cos(5.41706658717  + 758.52724572820 *T)
    sumX0  +=       1.032 * cos(4.77033383542  + 3075.29683383250 *T)
    sumX0  +=       0.899 * cos(1.08525568253  + 1053.40597254920 *T)
    sumX0  +=       0.822 * cos(3.63842218035  + 1354.18934135989 *T)
    sumX0  +=       0.742 * cos(2.00349570279  + 896.72922841380 *T)
    sumX0  +=       0.989 * cos(0.49078513563  + 493.28621964860 *T)
    sumX0  +=       0.989 * cos(3.84990051274  + 566.58334550761 *T)
    sumX0  +=       0.742 * cos(4.21700173393  + 948.45621860680 *T)
    sumX0  +=       0.743 * cos(0.15399551563  + 3267.25528826810 *T)
    sumX0  +=       0.785 * cos(4.75769615946  + 5091.90814063229 *T)
    sumX0  +=       0.735 * cos(1.17205944527  + 1689.22809324820 *T)
    sumX0  +=       0.902 * cos(0.31191653892  + 3487.66793070691 *T)
    sumX0  +=       0.766 * cos(4.59479371633  + 402.46298597130 *T)
    sumX0  +=       0.870 * cos(5.19110249005  + 522.29997071550 *T)
    sumX0  +=       0.870 * cos(2.29117581191  + 537.56959444070 *T)
    sumX0  +=       0.789 * cos(1.67089119717  + 950.38263429980 *T)
    sumX0  +=       0.729 * cos(0.18108618274  + 2226.03260534360 *T)
    sumX0  +=       0.967 * cos(1.47776388922  + 149.31937965110 *T)
    sumX0  +=       0.740 * cos(5.30434390465  + 172.00148100990 *T)
    sumX0  +=       0.732 * cos(5.42947892699  + 1600.36241303120 *T)
    sumX0  +=       0.743 * cos(3.66751928973  + 6011.68934095890 *T)
    sumX0  +=       0.773 * cos(4.89422681258  + 8616.85528644289 *T)
    sumX0  +=       0.718 * cos(3.85060206529  + 1055.20595944260 *T)
    sumX0  +=       0.814 * cos(0.89760005882  + 5.38525680900 *T)
    sumX0  +=       0.780 * cos(1.36192178535  + 949.37124234350 *T)
    sumX0  +=       0.736 * cos(2.50427265642  + 23968.89546371229 *T)
    sumX0  +=       0.703 * cos(6.05341470629  + 2108.71273712340 *T)
    sumX0  +=       0.705 * cos(4.27109184095  + 260.01333583751 *T)
    sumX0  +=       0.837 * cos(6.23244350461  + 97.43233076370 *T)
    sumX0  +=       0.708 * cos(1.83471250896  + 628.60776856660 *T)
    sumX0  +=       0.700 * cos(3.30963068355  + 78.20680978850 *T)
    sumX0  +=       0.704 * cos(3.63963986047  + 621.49422156580 *T)
    sumX0  +=       0.972 * cos(0.79938922309  + 806.96977631951 *T)
    sumX0  +=       0.902 * cos(2.81734866519  + 3259.39098174191 *T)
    sumX0  +=       0.806 * cos(2.20722357743  + 703.87700210090 *T)
    sumX0  +=       0.861 * cos(0.12846059397  + 215.99059347631 *T)
    sumX0  +=       0.761 * cos(1.26132938995  + 2310.96613229750 *T)
    sumX0  +=       0.705 * cos(6.12372523380  + 657.40657918490 *T)
    sumX0  +=       0.758 * cos(0.59861605652  + 1382.64352936309 *T)
    sumX0  +=       0.727 * cos(4.99142796126  + 64.71592109730 *T)
    sumX0  +=       0.687 * cos(2.94155817065  + 952.86394474940 *T)
    sumX0  +=       0.692 * cos(2.45791725703  + 10.05112325500 *T)
    sumX0  +=       0.963 * cos(4.33870402832  + 6916.10340678810 *T)
    sumX0  +=       0.906 * cos(6.05623889142  + 110.40195459310 *T)
    sumX0  +=       0.729 * cos(0.96200157729  + 3789.34507139489 *T)
    sumX0  +=       0.678 * cos(0.26412480699  + 2218.91905834281 *T)
    sumX0  +=       0.810 * cos(3.48317459389  + 678.01350700440 *T)
    sumX0  +=       0.761 * cos(0.98944003909  + 1542.84628985130 *T)
    sumX0  +=       0.734 * cos(1.45019492995  + 4863.44173055270 *T)
    sumX0  +=       0.793 * cos(2.38971782488  + 209.61075965840 *T)
    sumX0  +=       0.666 * cos(3.51087567158  + 4127.56521702090 *T)
    sumX0  +=       0.849 * cos(0.87603037716  + 368.83825021259 *T)
    sumX0  +=       0.815 * cos(2.42173445767  + 359.17395679300 *T)
    sumX0  +=       0.865 * cos(3.38745286825  + 157.02362748290 *T)
    sumX0  +=       0.799 * cos(3.33480033880  + 519.15220687259 *T)
    sumX0  +=       0.730 * cos(5.68999754324  + 98.65617104110 *T)
    sumX0  +=       0.923 * cos(0.04640766407  + 179.60266297770 *T)
    sumX0  +=       0.862 * cos(0.85096575069  + 1669.83045149109 *T)
    sumX0  +=       0.903 * cos(0.10254863796  + 1745.09968502540 *T)
    sumX0  +=       0.704 * cos(0.48323040130  + 718.85485562790 *T)
    sumX0  +=       0.765 * cos(2.39228400458  + 1834.85913711970 *T)
    sumX0  +=       0.651 * cos(0.55942905431  + 153.25153291419 *T)
    sumX0  +=       0.670 * cos(5.53772543546  + 5702.41101830310 *T)
    sumX0  +=       0.689 * cos(4.05205162416  + 540.49284787499 *T)
    sumX0  +=       0.677 * cos(1.95007474747  + 745.68155857090 *T)
    sumX0  +=       0.646 * cos(0.12458575010  + 2118.52004289490 *T)
    sumX0  +=       0.744 * cos(4.47249588475  + 305.85311143459 *T)
    sumX0  +=       0.837 * cos(3.51443539039  + 1818.39681088440 *T)
    sumX0  +=       0.795 * cos(1.49835916545  + 376.91500505990 *T)
    sumX0  +=       0.639 * cos(0.44760408452  + 5282.92637442791 *T)
    sumX0  +=       0.639 * cos(0.08134772596  + 651.18680406140 *T)
    sumX0  +=       0.791 * cos(5.22053399158  + 2432.21815378081 *T)
    sumX0  +=       0.666 * cos(3.04828183923  + 2439.33170078160 *T)
    sumX0  +=       0.699 * cos(3.61168485832  + 999.82200263240 *T)
    sumX0  +=       0.848 * cos(4.72505501580  + 445.00125562421 *T)
    sumX0  +=       0.784 * cos(0.87407256063  + 1372.34858862460 *T)
    sumX0  +=       0.784 * cos(3.91355130181  + 494.02242495900 *T)
    sumX0  +=       0.826 * cos(5.66627016660  + 1063.07026596880 *T)
    sumX0  +=       0.773 * cos(1.76079397166  + 3334.17258030920 *T)
    sumX0  +=       0.766 * cos(3.62656366077  + 927.59115048390 *T)
    sumX0  +=       0.629 * cos(5.55045637886  + 444.10748374690 *T)
    sumX0  +=       0.680 * cos(4.21509885032  + 2854.39655642670 *T)
    sumX0  +=       0.652 * cos(1.27664994367  + 99.92819542829 *T)
    sumX0  +=       0.628 * cos(4.74686224811  + 3693.85347914410 *T)
    sumX0  +=       0.810 * cos(3.03838651852  + 721.40560204669 *T)
    sumX0  +=       0.854 * cos(6.18289625105  + 1366.45638977370 *T)
    sumX0  +=       0.867 * cos(5.76374497089  + 10846.31310300770 *T)
    sumX0  +=       0.788 * cos(6.12952239651  + 1040.97886544099 *T)
    sumX0  +=       0.626 * cos(4.22299641050  + 841.90686400460 *T)
    sumX0  +=       0.630 * cos(5.17292925619  + 291.76902978489 *T)
    sumX0  +=       0.650 * cos(2.73748814643  + 558.38897058130 *T)
    sumX0  +=       0.611 * cos(3.22751051338  + 4443.95708667750 *T)
    sumX0  +=       0.726 * cos(2.09447836425  + 2151.25100677631 *T)
    sumX0  +=       0.751 * cos(5.81420327516  + 446.06752933469 *T)
    sumX0  +=       0.615 * cos(5.59117038277  + 208.38941150850 *T)
    sumX0  +=       0.745 * cos(2.60964521069  + 451.97179027311 *T)
    sumX0  +=       0.614 * cos(3.69094819583  + 6121.89566217830 *T)
    sumX0  +=       0.603 * cos(0.30175769208  + 181.56270856549 *T)
    sumX0  +=       0.616 * cos(0.71796395195  + 1144.07163965960 *T)
    sumX0  +=       0.636 * cos(0.81670492532  + 1751.29571393250 *T)
    sumX0  +=       0.596 * cos(2.57571835843  + 737.60480372361 *T)
    sumX0  +=       0.619 * cos(4.37153268900  + 326.19691469570 *T)
    sumX0  +=       0.787 * cos(1.03322103547  + 1802.12817323829 *T)
    sumX0  +=       0.781 * cos(3.79546900139  + 88.35873817511 *T)
    sumX0  +=       0.617 * cos(3.74482275875  + 921.10825080200 *T)
    sumX0  +=       0.591 * cos(2.16133870119  + 1166.16304018740 *T)
    sumX0  +=       0.590 * cos(5.76599972625  + 3634.37720703489 *T)
    sumX0  +=       0.722 * cos(2.97658953647  + 1791.57010794140 *T)
    sumX0  +=       0.820 * cos(4.24711097101  + 1372.08546406621 *T)
    sumX0  +=       0.590 * cos(4.55770998857  + 627.61093082531 *T)
    sumX0  +=       0.611 * cos(2.00459977476  + 392.90175841570 *T)
    sumX0  +=       0.611 * cos(2.33608587361  + 666.96780674051 *T)
    sumX0  +=       0.584 * cos(5.71119009710  + 2737.56432317350 *T)
    sumX0  +=       0.624 * cos(1.77518234702  + 702.18008107150 *T)
    sumX0  +=       0.600 * cos(3.53895345579  + 203.49405039889 *T)
    sumX0  +=       0.720 * cos(3.17911251266  + 3333.74269718250 *T)
    sumX0  +=       0.715 * cos(6.07736623315  + 3347.96979118411 *T)
    sumX0  +=       0.604 * cos(0.54072229600  + 427.36327322130 *T)
    sumX0  +=       0.578 * cos(0.68384333266  + 881.75137488680 *T)
    sumX0  +=       0.752 * cos(2.27484845625  + 799.36859435169 *T)
    sumX0  +=       0.581 * cos(5.43194940731  + 2954.04481234920 *T)
    sumX0  +=       0.754 * cos(4.23362271932  + 1022.97718474320 *T)
    sumX0  +=       0.772 * cos(3.22944539114  + 711.74130862710 *T)
    sumX0  +=       0.577 * cos(3.90731032425  + 700.20797339620 *T)
    sumX0  +=       0.736 * cos(2.17294584865  + 458.12212942741 *T)
    sumX0  +=       0.736 * cos(2.16773979972  + 601.74743572880 *T)
    sumX0  +=       0.693 * cos(4.97718562978  + 1967.96000801901 *T)
    sumX0  +=       0.740 * cos(2.23775921738  + 416.79314583689 *T)
    sumX0  +=       0.704 * cos(2.75287525372  + 425.35753565121 *T)
    sumX0  +=       0.708 * cos(4.59755586960  + 3391.68870348911 *T)
    sumX0  +=       0.607 * cos(5.77643174191  + 105.76971804189 *T)
    sumX0  +=       0.559 * cos(4.05584510248  + 20426.81490990550 *T)
    sumX0  +=       0.628 * cos(6.14517735977  + 1809.72935520611 *T)
    sumX0  +=       0.696 * cos(3.48103715785  + 378.88711273520 *T)
    sumX0  +=       0.554 * cos(4.24043679096  + 386.00065973601 *T)
    sumX0  +=       0.589 * cos(2.84804112106  + 3273.88120030190 *T)
    sumX0  +=       0.633 * cos(5.05901975525  + 131.16013238639 *T)
    sumX0  +=       0.562 * cos(2.97186811483  + 4803.96545844350 *T)
    sumX0  +=       0.560 * cos(4.71048501869  + 421.21293406700 *T)
    sumX0  +=       0.580 * cos(1.24952540618  + 5179.83360020930 *T)
    sumX0  +=       0.740 * cos(2.67735727689  + 913.50706883420 *T)
    sumX0  +=       0.699 * cos(1.23224313252  + 2627.35800195410 *T)
    sumX0  +=       0.699 * cos(1.23224313252  + 2626.87036698710 *T)
    sumX0  +=       0.747 * cos(1.73127494354  + 90.31878376290 *T)
    sumX0  +=       0.575 * cos(1.40772978331  + 1656.98476433380 *T)
    sumX0  +=       0.555 * cos(2.46441478919  + 6702.80431135010 *T)
    sumX0  +=       0.559 * cos(5.27994346504  + 388.96960515260 *T)
    sumX0  +=       0.672 * cos(3.95084883531  + 2417.24030025381 *T)
    sumX0  +=       0.606 * cos(2.25540936996  + 3693.36584417709 *T)
    sumX0  +=       0.538 * cos(0.92849193558  + 7564.58690325449 *T)
    sumX0  +=       0.539 * cos(5.86855643077  + 828.16740497000 *T)
    sumX0  +=       0.548 * cos(0.57206390796  + 4700.87268422489 *T)
    sumX0  +=       0.599 * cos(1.91496502312  + 108.47803102759 *T)
    sumX0  +=       0.550 * cos(0.94455632585  + 3046.35501086230 *T)
    sumX0  +=       0.652 * cos(3.44170706330  + 282.20780222929 *T)
    sumX0  +=       0.586 * cos(4.87892797654  + 878.08234618209 *T)
    sumX0  +=       0.622 * cos(1.44606819015  + 568.80402352629 *T)
    sumX0  +=       0.728 * cos(4.06001131597  + 344.45922782439 *T)
    sumX0  +=       0.544 * cos(3.88849597870  + 10633.01400756970 *T)
    sumX0  +=       0.657 * cos(2.19781111367  + 464.97504399731 *T)
    sumX0  +=       0.546 * cos(0.09522235899  + 4112.85048805230 *T)
    sumX0  +=       0.541 * cos(2.90330545324  + 121.00820399980 *T)
    sumX0  +=       0.695 * cos(4.33330180196  + 421.68850694650 *T)
    sumX0  +=       0.601 * cos(0.02013991441  + 426.32074349770 *T)
    sumX0  +=       0.525 * cos(2.67444074803  + 76674.39271195528 *T)
    sumX0  +=       0.532 * cos(4.39556016683  + 8734.17515466309 *T)
    sumX0  +=       0.622 * cos(4.02967281763  + 1091.38144162010 *T)
    sumX0  +=       0.553 * cos(5.66376860324  + 604.22874617840 *T)
    sumX0  +=       0.633 * cos(3.70162370380  + 2274.30301515300 *T)
    sumX0  +=       0.564 * cos(1.73577608543  + 241.85408857280 *T)
    sumX0  +=       0.516 * cos(0.26364140815  + 5599.31824408450 *T)
    sumX0  +=       0.645 * cos(1.30657929230  + 10110.43658947590 *T)
    sumX0  +=       0.509 * cos(2.11728726417  + 1159.04949318660 *T)
    sumX0  +=       0.507 * cos(5.07344432036  + 615.76208140930 *T)
    sumX0  +=       0.599 * cos(4.14881340222  + 185.23173727020 *T)
    sumX0  +=       0.522 * cos(0.92209254895  + 1884.36794142151 *T)
    sumX0  +=       0.520 * cos(2.30383555351  + 1457.28211557850 *T)
    sumX0  +=       0.527 * cos(0.12531819468  + 1512.56300652469 *T)
    sumX0  +=       0.503 * cos(3.39203314539  + 1037.79747170330 *T)
    sumX0  +=       0.517 * cos(5.27606872473  + 6018.80288795970 *T)
    sumX0  +=       0.613 * cos(1.73070833822  + 607.89777488310 *T)
    sumX0  +=       0.601 * cos(1.95549963379  + 700.69560836320 *T)
    sumX0  +=       0.505 * cos(3.18507574383  + 5908.59656674030 *T)
    sumX0  +=       0.679 * cos(5.71225833484  + 1357.37073509760 *T)
    sumX0  +=       0.496 * cos(1.76244328747  + 1048.84317196720 *T)
    sumX0  +=       0.602 * cos(0.81687029998  + 680.98245242101 *T)
    sumX0  +=       0.621 * cos(3.11600964819  + 1560.37488979710 *T)
    sumX0  +=       0.623 * cos(2.89379873996  + 120.01136625851 *T)
    sumX0  +=       0.494 * cos(1.02217524155  + 523.57199510270 *T)
    sumX0  +=       0.494 * cos(0.17691775323  + 536.29757005351 *T)
    sumX0  +=       0.500 * cos(0.07757434568  + 2840.65709739210 *T)
    sumX0  +=       0.657 * cos(1.97087292015  + 2351.05672139560 *T)
    sumX0  +=       0.587 * cos(5.17430255862  + 215.50295850929 *T)
    sumX0  +=       0.609 * cos(5.71512479774  + 614.86830953200 *T)
    sumX0  +=       0.609 * cos(2.57353214415  + 614.38067456499 *T)
    sumX0  +=       0.502 * cos(4.96490318108  + 1379.46213562540 *T)
    sumX0  +=       0.492 * cos(1.74083007620  + 419.28901050149 *T)
    sumX0  +=       0.485 * cos(3.23499089371  + 1937.95191133830 *T)
    sumX0  +=       0.484 * cos(4.83554210114  + 102.09819720970 *T)
    sumX0  +=       0.491 * cos(4.82626130082  + 419.19264228190 *T)
    sumX0  +=       0.636 * cos(0.80534214264  + 664.52012618570 *T)
    sumX0  +=       0.478 * cos(2.68779652903  + 204.97852310719 *T)
    sumX0  +=       0.494 * cos(4.49966350328  + 2758.90496417590 *T)
    sumX0  +=       0.495 * cos(3.15153574636  + 398.53083270820 *T)
    sumX0  +=       0.483 * cos(2.53161064938  + 89.10949770051 *T)
    sumX0  +=       0.589 * cos(1.83933111516  + 2740.74571691120 *T)
    sumX0  +=       0.600 * cos(1.57257401507  + 634.51202950500 *T)
    sumX0  +=       0.486 * cos(5.04467319390  + 2203.94120481581 *T)
    sumX0  +=       0.515 * cos(6.13663617559  + 1091.11831706170 *T)
    sumX0  +=       0.615 * cos(1.11374614520  + 3700.96702614491 *T)
    sumX0  +=       0.490 * cos(1.91408549288  + 8631.08238044450 *T)
    sumX0  +=       0.494 * cos(4.66411522593  + 984.35651413839 *T)
    sumX0  +=       0.475 * cos(2.27777755361  + 103.59971626049 *T)
    sumX0  +=       0.475 * cos(2.59023496451  + 835.54407652920 *T)
    sumX0  +=       0.512 * cos(5.15116159496  + 397.90018538930 *T)
    sumX0  +=       0.655 * cos(1.80371823253  + 127.22797912329 *T)
    sumX0  +=       0.468 * cos(0.94842717325  + 582.62498061761 *T)
    sumX0  +=       0.468 * cos(3.39225847512  + 477.24458453860 *T)
    sumX0  +=       0.543 * cos(2.77376969444  + 481.17673780170 *T)
    sumX0  +=       0.543 * cos(1.56691595393  + 578.69282735451 *T)
    sumX0  +=       0.517 * cos(0.86310995182  + 76.02225379210 *T)
    sumX0  +=       0.573 * cos(1.37924154661  + 1132.92287353170 *T)
    sumX0  +=       0.649 * cos(0.37556266722  + 3184.60938317460 *T)
    sumX0  +=       0.480 * cos(4.52427818977  + 2133.23477186350 *T)
    sumX0  +=       0.468 * cos(2.06153810948  + 2048.15823255770 *T)
    sumX0  +=       0.594 * cos(0.39773738437  + 390.92965074040 *T)
    sumX0  +=       0.459 * cos(1.62165138981  + 937.88609122240 *T)
    sumX0  +=       0.481 * cos(2.13958103959  + 1443.68566889580 *T)
    sumX0  +=       0.463 * cos(5.78478013513  + 429.76276966630 *T)
    sumX0  +=       0.545 * cos(0.16089867082  + 3157.04896704871 *T)
    sumX0  +=       0.475 * cos(5.50914223201  + 3907.15257458210 *T)
    sumX0  +=       0.609 * cos(4.81927839884  + 1595.53648789080 *T)
    sumX0  +=       0.473 * cos(6.26679260619  + 733.67265046051 *T)
    sumX0  +=       0.541 * cos(1.14751616188  + 52176.05010063190 *T)
    sumX0  +=       0.533 * cos(5.98390379378  + 382.85289589309 *T)
    sumX0  +=       0.534 * cos(4.28038540288  + 843.87897167990 *T)
    sumX0  +=       0.451 * cos(2.42625593348  + 5371.30441967789 *T)
    sumX0  +=       0.471 * cos(4.45737409056  + 2572.43257168090 *T)
    sumX0  +=       0.558 * cos(2.14287458249  + 594.89452115890 *T)
    sumX0  +=       0.588 * cos(1.67758273586  + 458.33457774851 *T)
    sumX0  +=       0.482 * cos(4.23247478761  + 6114.78211517750 *T)
    sumX0  +=       0.449 * cos(4.70348907548  + 1221.60474883790 *T)
    sumX0  +=       0.447 * cos(5.21836347199  + 584.26951202031 *T)
    sumX0  +=       0.447 * cos(5.40550748355  + 475.60005313590 *T)
    sumX0  +=       0.466 * cos(3.79996346648  + 632.50629193491 *T)
    sumX0  +=       0.624 * cos(2.89552279007  + 10529.92123335110 *T)
    sumX0  +=       0.605 * cos(5.04020796752  + 300.85468446100 *T)
    sumX0  +=       0.487 * cos(2.27435976567  + 321.30155358610 *T)
    sumX0  +=       0.487 * cos(0.14720235737  + 206.90493880020 *T)
    sumX0  +=       0.526 * cos(1.60016547325  + 1336.66074141410 *T)
    sumX0  +=       0.450 * cos(4.45402230413  + 518.62844988331 *T)
    sumX0  +=       0.450 * cos(3.02825599783  + 541.24111527290 *T)
    sumX0  +=       0.452 * cos(2.45557763807  + 252.41215386970 *T)
    sumX0  +=       0.569 * cos(6.06714280091  + 207.42620366200 *T)
    sumX0  +=       0.442 * cos(0.71239558825  + 10976.83498721549 *T)
    sumX0  +=       0.435 * cos(2.04457315580  + 218.47190392590 *T)
    sumX0  +=       0.536 * cos(1.16069752348  + 1781.27516720290 *T)
    sumX0  +=       0.606 * cos(1.47437041783  + 6599.71153713150 *T)
    sumX0  +=       0.430 * cos(6.27151324176  + 2535.31092799940 *T)
    sumX0  +=       0.465 * cos(3.76774297614  + 1084.00477006090 *T)
    sumX0  +=       0.528 * cos(3.23700299396  + 557.75832326239 *T)
    sumX0  +=       0.602 * cos(1.89096655273  + 354.76872277791 *T)
    sumX0  +=       0.602 * cos(2.44971909564  + 705.10084237830 *T)
    sumX0  +=       0.579 * cos(2.05383962420  + 462.05428269051 *T)
    sumX0  +=       0.579 * cos(2.28684602417  + 597.81528246570 *T)
    sumX0  +=       0.480 * cos(5.15414760143  + 1106.09617058870 *T)
    sumX0  +=       0.561 * cos(0.48075957659  + 1859.23815950790 *T)
    sumX0  +=       0.565 * cos(3.20740832233  + 531.89482816590 *T)
    sumX0  +=       0.593 * cos(3.87608972833  + 1291.67479120640 *T)
    sumX0  +=       0.593 * cos(3.87608972833  + 1291.18715623940 *T)
    sumX0  +=       0.467 * cos(1.83754367407  + 2833.05591542430 *T)
    sumX0  +=       0.423 * cos(0.03642510895  + 799.85622931870 *T)
    sumX0  +=       0.591 * cos(2.46012843727  + 1819.39364862570 *T)
    sumX0  +=       0.442 * cos(1.08908006468  + 1069.18697522831 *T)
    sumX0  +=       0.538 * cos(1.02927939001  + 1065.35788782919 *T)
    sumX0  +=       0.472 * cos(2.28157145696  + 460.29462333630 *T)
    sumX0  +=       0.432 * cos(6.00978313661  + 2080.87464222410 *T)
    sumX0  +=       0.415 * cos(2.96241779304  + 4760.34895633411 *T)
    sumX0  +=       0.455 * cos(3.03366464226  + 2527.44662147320 *T)
    sumX0  +=       0.506 * cos(0.36247155090  + 535.66692273460 *T)
    sumX0  +=       0.408 * cos(3.08015718534  + 237.43430034270 *T)
    sumX0  +=       0.558 * cos(1.03001492173  + 2348.08777597900 *T)
    sumX0  +=       0.506 * cos(5.25352948963  + 2424.35384725460 *T)
    sumX0  +=       0.496 * cos(6.06044529767  + 4120.45167002010 *T)
    sumX0  +=       0.496 * cos(2.91885264408  + 4119.96403505309 *T)
    sumX0  +=       0.432 * cos(1.29582024184  + 1301.23601876200 *T)
    sumX0  +=       0.550 * cos(5.43475108780  + 8837.26792888169 *T)
    sumX0  +=       0.546 * cos(3.31132091883  + 761.98382611181 *T)
    sumX0  +=       0.546 * cos(0.16972826524  + 761.49619114479 *T)
    sumX0  +=       0.435 * cos(6.11673595133  + 205.89354684390 *T)
    sumX0  +=       0.463 * cos(4.32054580531  + 633.54882165850 *T)
    sumX0  +=       0.408 * cos(1.79247263237  + 1482.55490984399 *T)
    sumX0  +=       0.483 * cos(0.34268842153  + 2879.26321378190 *T)
    sumX0  +=       0.511 * cos(3.60313597493  + 1670.58121101650 *T)
    sumX0  +=       0.421 * cos(5.74575757785  + 638.65663108921 *T)
    sumX0  +=       0.512 * cos(4.63462743085  + 1474.42997088690 *T)
    sumX0  +=       0.388 * cos(5.80743867551  + 2425.10460678000 *T)
    sumX0  +=       0.483 * cos(5.61088897560  + 154.05468206630 *T)
    sumX0  +=       0.412 * cos(3.95817641351  + 83.62343575990 *T)
    sumX0  +=       0.384 * cos(0.10024887062  + 673.86890542020 *T)
    sumX0  +=       0.400 * cos(2.77474684816  + 2755.72357043820 *T)
    sumX0  +=       0.500 * cos(0.77216947100  + 1973.37663399041 *T)
    sumX0  +=       0.380 * cos(6.00308091678  + 3583.64715792471 *T)
    sumX0  +=       0.509 * cos(6.23294075938  + 5856.23384163189 *T)
    sumX0  +=       0.392 * cos(3.41709820264  + 6328.08121061550 *T)
    sumX0  +=       0.391 * cos(5.34392743186  + 670.89996000361 *T)
    sumX0  +=       0.397 * cos(5.90731747027  + 50586.48957038109 *T)
    sumX0  +=       0.468 * cos(5.85291318311  + 1970.92895343560 *T)
    sumX0  +=       0.491 * cos(6.06026074383  + 1288.70584578980 *T)
    sumX0  +=       0.491 * cos(2.91866809024  + 1288.21821082279 *T)
    sumX0  +=       0.472 * cos(5.64551789445  + 416.53251340600 *T)
    sumX0  +=       0.491 * cos(2.66310290421  + 601.53498740770 *T)
    sumX0  +=       0.476 * cos(0.29096139335  + 2015.42726867629 *T)

    # JUPITER X Coord
    sumX1  = 0
    sumX1  +=  1912556.490 * cos(4.23275123829  + 0.24381748350 *T)
    sumX1  +=  634902.259 * cos(0.10706507632  + 1059.62574767270 *T)
    sumX1  +=  600483.021 * cos(2.42939944495  + 522.82123557730 *T)
    sumX1  +=  589005.176 * cos(1.91564604125  + 537.04832957890 *T)
    sumX1  +=   81432.741 * cos(3.47331266459  + 7.35736448430 *T)
    sumX1  +=   46191.617 * cos(0.45737291603  + 1589.31671276730 *T)
    sumX1  +=   37242.394 * cos(3.38536248935  + 529.93478257810 *T)
    sumX1  +=   32508.552 * cos(1.74680960116  + 1052.51220067191 *T)
    sumX1  +=   31231.184 * cos(2.34701987017  + 1066.73929467350 *T)
    sumX1  +=   25950.740 * cos(5.52780779320  + 529.44714761109 *T)
    sumX1  +=   20232.242 * cos(4.46840569934  + 110.45013870291 *T)
    sumX1  +=   17941.640 * cos(4.02099530935  + 426.84200835950 *T)
    sumX1  +=   13643.757 * cos(0.30427911223  + 633.02755679670 *T)
    sumX1  +=   12962.263 * cos(2.56767956763  + 515.70768857651 *T)
    sumX1  +=    9123.979 * cos(1.78092724073  + 544.16187657970 *T)
    sumX1  +=    8098.978 * cos(6.15662077289  + 949.41942645330 *T)
    sumX1  +=    7143.102 * cos(3.17367507069  + 323.74923414091 *T)
    sumX1  +=    4115.125 * cos(4.68496126162  + 205.94173095370 *T)
    sumX1  +=    4370.464 * cos(1.14540545130  + 736.12033101530 *T)
    sumX1  +=    3400.571 * cos(2.90100216943  + 526.75338884040 *T)
    sumX1  +=    3332.481 * cos(1.43702485996  + 533.11617631580 *T)
    sumX1  +=    3869.655 * cos(3.33803842759  + 14.47091148511 *T)
    sumX1  +=    3043.986 * cos(1.65416267419  + 526.00262931501 *T)
    sumX1  +=    3013.156 * cos(2.59693904643  + 419.72846135871 *T)
    sumX1  +=    3001.886 * cos(2.68361926800  + 533.86693584120 *T)
    sumX1  +=    2892.493 * cos(4.90922246003  + 103.33659170210 *T)
    sumX1  +=    2393.125 * cos(3.62223412280  + 316.63568714010 *T)
    sumX1  +=    2989.944 * cos(0.80713415295  + 2119.00767786191 *T)
    sumX1  +=    2354.966 * cos(3.28995924205  + 109.96250373590 *T)
    sumX1  +=    1853.775 * cos(2.76966317285  + 1596.43025976811 *T)
    sumX1  +=    1972.856 * cos(5.08955117905  + 956.53297345411 *T)
    sumX1  +=    1820.762 * cos(0.66887392674  + 743.23387801611 *T)
    sumX1  +=    1810.752 * cos(3.83905864684  + 95.73540973430 *T)
    sumX1  +=    1531.701 * cos(4.31865285485  + 117.56368570370 *T)
    sumX1  +=    1894.874 * cos(1.86455327598  + 1582.20316576650 *T)
    sumX1  +=    1527.752 * cos(1.49162793541  + 640.14110379750 *T)
    sumX1  +=    1634.391 * cos(1.39765618229  + 1045.39865367110 *T)
    sumX1  +=    1088.512 * cos(1.51681538001  + 625.91400979590 *T)
    sumX1  +=    1025.988 * cos(2.57282118549  + 433.95555536030 *T)
    sumX1  +=    1045.564 * cos(2.53075471830  + 1265.81129610991 *T)
    sumX1  +=     977.087 * cos(2.19851142066  + 220.65645992230 *T)
    sumX1  +=     984.636 * cos(1.42900900580  + 1169.83206889211 *T)
    sumX1  +=     805.130 * cos(0.01887809686  + 942.30587945250 *T)
    sumX1  +=     821.829 * cos(5.66460920839  + 309.03450517230 *T)
    sumX1  +=     798.891 * cos(1.85669438695  + 1155.60497489050 *T)
    sumX1  +=     834.956 * cos(5.16168526363  + 522.33360061030 *T)
    sumX1  +=     600.935 * cos(2.35527790752  + 6.86972951729 *T)
    sumX1  +=     752.130 * cos(0.12867542563  + 1162.71852189130 *T)
    sumX1  +=     539.923 * cos(4.24283648453  + 853.44019923550 *T)
    sumX1  +=     599.196 * cos(4.21898955063  + 213.05527795450 *T)
    sumX1  +=     497.018 * cos(4.81516413781  + 198.82818395289 *T)
    sumX1  +=     476.930 * cos(5.90061449066  + 1692.40948698591 *T)
    sumX1  +=     499.703 * cos(3.02089381667  + 330.86278114170 *T)
    sumX1  +=     383.795 * cos(2.17189096891  + 1073.85284167431 *T)
    sumX1  +=     504.945 * cos(5.88136662932  + 1059.13811270570 *T)
    sumX1  +=     402.231 * cos(6.21084762941  + 1479.11039154791 *T)
    sumX1  +=     373.538 * cos(1.28962855954  + 729.00678401450 *T)
    sumX1  +=     349.467 * cos(3.10812701391  + 426.35437339250 *T)
    sumX1  +=     316.847 * cos(2.11759539145  + 839.21310523390 *T)
    sumX1  +=     324.562 * cos(5.71130479446  + 206.42936592071 *T)
    sumX1  +=     246.849 * cos(4.10602399001  + 838.72547026689 *T)
    sumX1  +=     255.191 * cos(3.62607142490  + 412.12727939089 *T)
    sumX1  +=     230.121 * cos(2.72254614569  + 508.59414157570 *T)
    sumX1  +=     246.851 * cos(1.75011280000  + 1272.92484311070 *T)
    sumX1  +=     222.270 * cos(2.70702942772  + 412.61491435790 *T)
    sumX1  +=     213.383 * cos(2.90120840681  + 1376.01761732930 *T)
    sumX1  +=     225.567 * cos(5.42992965137  + 625.42637482889 *T)
    sumX1  +=     212.049 * cos(3.04058050625  + 3.42521122121 *T)
    sumX1  +=     181.234 * cos(3.64994746675  + 1368.90407032850 *T)
    sumX1  +=     174.181 * cos(1.85281706799  + 1062.80714141041 *T)
    sumX1  +=     189.391 * cos(1.15749117460  + 2648.69864295650 *T)
    sumX1  +=     161.119 * cos(3.02399692553  + 519.63984183961 *T)
    sumX1  +=     157.874 * cos(1.64594194339  + 536.56069461190 *T)
    sumX1  +=     145.420 * cos(2.71846376946  + 1258.69774910910 *T)
    sumX1  +=     144.205 * cos(2.08426886196  + 1055.69359440960 *T)
    sumX1  +=     143.507 * cos(3.70576716798  + 309.52214013930 *T)
    sumX1  +=     126.188 * cos(3.31694506826  + 1063.55790093581 *T)
    sumX1  +=     124.788 * cos(1.62052640087  + 551.27542358051 *T)
    sumX1  +=     125.329 * cos(1.31439998076  + 540.22972331660 *T)
    sumX1  +=     146.669 * cos(2.72335559452  + 1056.44435393500 *T)
    sumX1  +=     115.893 * cos(3.18162677570  + 2126.12122486270 *T)
    sumX1  +=     124.485 * cos(0.17472854156  + 3.68833577960 *T)
    sumX1  +=     118.218 * cos(4.80555513709  + 515.22005360950 *T)
    sumX1  +=     114.208 * cos(2.05820677440  + 227.77000692311 *T)
    sumX1  +=     113.852 * cos(2.70245764618  + 742.74624304909 *T)
    sumX1  +=     144.447 * cos(6.08317056171  + 1486.22393854870 *T)
    sumX1  +=     124.849 * cos(1.63225929060  + 1575.08961876571 *T)
    sumX1  +=      93.951 * cos(3.84985265371  + 1155.11733992350 *T)
    sumX1  +=     100.599 * cos(2.02088520418  + 2111.89413086110 *T)
    sumX1  +=     103.583 * cos(5.06428969516  + 419.24082639170 *T)
    sumX1  +=      92.556 * cos(1.85907721166  + 1699.52303398670 *T)
    sumX1  +=      99.479 * cos(5.78783063560  + 301.92095817150 *T)
    sumX1  +=      82.832 * cos(1.28886276595  + 441.06910236111 *T)
    sumX1  +=      86.688 * cos(5.09274860535  + 941.81824448550 *T)
    sumX1  +=      95.969 * cos(2.12504520091  + 102.84895673509 *T)
    sumX1  +=      80.324 * cos(0.88712083078  + 1471.99684454710 *T)
    sumX1  +=     102.333 * cos(3.18857661435  + 21.58445848590 *T)
    sumX1  +=      78.195 * cos(5.09322017801  + 525.26891613211 *T)
    sumX1  +=      74.123 * cos(1.57689883069  + 647.25465079831 *T)
    sumX1  +=      79.270 * cos(3.27341766644  + 728.51914904750 *T)
    sumX1  +=      74.807 * cos(5.52867705621  + 534.60064902410 *T)
    sumX1  +=      87.759 * cos(4.16413990980  + 632.53992182970 *T)
    sumX1  +=      69.800 * cos(2.80403510206  + 618.80046279510 *T)
    sumX1  +=      60.886 * cos(4.07058850374  + 846.32665223470 *T)
    sumX1  +=      58.777 * cos(2.25774461225  + 832.09955823310 *T)
    sumX1  +=      61.003 * cos(6.24178613664  + 2222.10045208050 *T)
    sumX1  +=      66.693 * cos(2.28288216120  + 1685.29593998510 *T)
    sumX1  +=      50.999 * cos(5.78085041063  + 2008.80135664250 *T)
    sumX1  +=      50.585 * cos(4.75410728625  + 213.54291292150 *T)
    sumX1  +=      45.850 * cos(4.24082041831  + 831.61192326610 *T)
    sumX1  +=      52.339 * cos(1.38343334733  + 1038.28510667030 *T)
    sumX1  +=      46.409 * cos(2.49832780835  + 524.51815660670 *T)
    sumX1  +=      49.446 * cos(2.75500829134  + 963.64652045490 *T)
    sumX1  +=      45.782 * cos(1.84235784093  + 535.35140854951 *T)
    sumX1  +=      61.385 * cos(2.63233647245  + 1795.50226120450 *T)
    sumX1  +=      58.958 * cos(2.91371154675  + 316.14805217309 *T)
    sumX1  +=      59.833 * cos(4.54238626333  + 4.17597074660 *T)
    sumX1  +=      43.525 * cos(0.40754709239  + 149.80701461810 *T)
    sumX1  +=      39.973 * cos(4.82513939066  + 1383.13116433011 *T)
    sumX1  +=      53.170 * cos(1.55759345804  + 750.34742501690 *T)
    sumX1  +=      39.316 * cos(3.58850609580  + 1471.50920958009 *T)
    sumX1  +=      40.288 * cos(1.77691319997  + 948.93179148630 *T)
    sumX1  +=      38.303 * cos(1.25612538854  + 511.77553531340 *T)
    sumX1  +=      36.523 * cos(3.08500012455  + 548.09402984280 *T)
    sumX1  +=      38.571 * cos(4.79060991224  + 1258.21011414209 *T)
    sumX1  +=      35.601 * cos(2.81741907830  + 10.53875822200 *T)
    sumX1  +=      36.868 * cos(1.53995872607  + 527.01402127131 *T)
    sumX1  +=      36.603 * cos(2.80095469601  + 532.85554388490 *T)
    sumX1  +=      41.197 * cos(3.47741533158  + 845.83901726770 *T)
    sumX1  +=      30.575 * cos(4.13554799825  + 124.67723270450 *T)
    sumX1  +=      27.619 * cos(1.69648523137  + 1802.61580820530 *T)
    sumX1  +=      32.289 * cos(5.50255743191  + 618.31282782810 *T)
    sumX1  +=      28.470 * cos(3.04847955006  + 1044.91101870410 *T)
    sumX1  +=      25.452 * cos(1.34642958378  + 1265.32366114289 *T)
    sumX1  +=      27.183 * cos(2.69831398387  + 1898.59503542310 *T)
    sumX1  +=      22.512 * cos(0.44856506675  + 639.65346883050 *T)
    sumX1  +=      22.622 * cos(0.18357944836  + 935.19233245171 *T)
    sumX1  +=      24.508 * cos(3.71033407609  + 1361.79052332771 *T)
    sumX1  +=      25.260 * cos(1.81233714136  + 540.98048284201 *T)
    sumX1  +=      20.677 * cos(3.98002843517  + 1148.00379292269 *T)
    sumX1  +=      28.190 * cos(3.07761522702  + 1905.70858242391 *T)
    sumX1  +=      20.297 * cos(3.81657536112  + 1162.23088692429 *T)
    sumX1  +=      25.685 * cos(1.55199744523  + 735.63269604830 *T)
    sumX1  +=      19.769 * cos(5.53593575435  + 11.28951774740 *T)
    sumX1  +=      17.928 * cos(3.04845555075  + 98.91680347199 *T)
    sumX1  +=      17.469 * cos(3.32708371179  + 1787.90107923670 *T)
    sumX1  +=      18.703 * cos(4.50425954841  + 1574.60198379870 *T)
    sumX1  +=      18.135 * cos(3.93430947434  + 910.06255053810 *T)
    sumX1  +=      17.089 * cos(2.56125921461  + 1603.54380676890 *T)
    sumX1  +=      17.538 * cos(6.04152227093  + 525.74199688410 *T)
    sumX1  +=      17.368 * cos(4.58219600456  + 534.12756827210 *T)
    sumX1  +=      19.204 * cos(2.23161769695  + 1375.52998236229 *T)
    sumX1  +=      17.455 * cos(0.10703246236  + 76.50988875911 *T)
    sumX1  +=      16.153 * cos(4.30082408762  + 99.66756299740 *T)
    sumX1  +=      16.460 * cos(2.09636027394  + 1788.38871420371 *T)
    sumX1  +=      15.683 * cos(3.99052583747  + 2.93757625420 *T)
    sumX1  +=      16.034 * cos(4.98317709627  + 191.71463695210 *T)
    sumX1  +=      15.051 * cos(2.78980787117  + 106.03035047280 *T)
    sumX1  +=      14.892 * cos(1.77707927575  + 518.88908231420 *T)
    sumX1  +=      15.187 * cos(6.07616641818  + 1588.82907780029 *T)
    sumX1  +=      17.388 * cos(4.43775098797  + 18.40306474821 *T)
    sumX1  +=      14.796 * cos(2.06469993479  + 1049.33080693420 *T)
    sumX1  +=      15.408 * cos(4.06617777635  + 106.51798543980 *T)
    sumX1  +=      17.378 * cos(0.47153358689  + 88.62186273350 *T)
    sumX1  +=      13.474 * cos(1.04099539964  + 1581.71553079950 *T)
    sumX1  +=      13.574 * cos(4.03698274106  + 106.78110999820 *T)
    sumX1  +=      17.733 * cos(3.53345150510  + 860.55374623631 *T)
    sumX1  +=      14.099 * cos(5.20790939624  + 934.70469748469 *T)
    sumX1  +=      14.690 * cos(2.86077681205  + 337.97632814250 *T)
    sumX1  +=      11.717 * cos(1.50309851310  + 3178.38960805111 *T)
    sumX1  +=      10.988 * cos(2.13993350041  + 2214.98690507971 *T)
    sumX1  +=      10.909 * cos(0.16362139771  + 1052.02456570489 *T)
    sumX1  +=      11.465 * cos(2.83736836823  + 1361.30288836070 *T)
    sumX1  +=      10.506 * cos(1.65935656006  + 2104.78058386030 *T)
    sumX1  +=      12.808 * cos(5.23966663914  + 96.22304470130 *T)
    sumX1  +=      13.402 * cos(1.46045916222  + 721.89323701370 *T)
    sumX1  +=      10.003 * cos(3.71949533579  + 1464.39566257930 *T)
    sumX1  +=      11.029 * cos(4.75763401948  + 604.71638114540 *T)
    sumX1  +=       9.807 * cos(5.38496623744  + 107.26874496520 *T)
    sumX1  +=      11.312 * cos(1.04980298350  + 220.16882495529 *T)
    sumX1  +=      10.028 * cos(5.51845919925  + 497.69145366371 *T)
    sumX1  +=       9.515 * cos(4.20231641978  + 1890.99385345530 *T)
    sumX1  +=      10.988 * cos(0.55033431492  + 2015.91490364330 *T)
    sumX1  +=      10.134 * cos(2.26013428987  + 1592.49810650500 *T)
    sumX1  +=       8.954 * cos(5.58285753900  + 10.80188278040 *T)
    sumX1  +=       8.739 * cos(3.10143619037  + 91.80325647120 *T)
    sumX1  +=       8.399 * cos(2.39323702495  + 1148.49142788971 *T)
    sumX1  +=       8.408 * cos(2.53671229460  + 1585.38455950421 *T)
    sumX1  +=       9.230 * cos(5.10523054826  + 562.17811149250 *T)
    sumX1  +=       8.186 * cos(3.07470498245  + 2104.29294889330 *T)
    sumX1  +=       8.259 * cos(3.79122069331  + 302.40859313851 *T)
    sumX1  +=       9.900 * cos(3.08797905190  + 1478.62275658089 *T)
    sumX1  +=      10.878 * cos(3.84174776466  + 113.63153244060 *T)
    sumX1  +=       9.582 * cos(5.95011959797  + 455.15318401080 *T)
    sumX1  +=       7.712 * cos(2.87491912382  + 422.90985509640 *T)
    sumX1  +=       7.595 * cos(1.59710858352  + 629.84616305900 *T)
    sumX1  +=       8.496 * cos(2.61146913566  + 224.58861318540 *T)
    sumX1  +=       7.414 * cos(2.03258534769  + 39.11305843170 *T)
    sumX1  +=       7.292 * cos(2.69782256595  + 430.02340209721 *T)
    sumX1  +=       7.357 * cos(3.64822171992  + 2655.81218995730 *T)
    sumX1  +=       7.633 * cos(1.32269581525  + 636.95971005981 *T)
    sumX1  +=       7.369 * cos(2.28675718790  + 2229.21399908130 *T)
    sumX1  +=       7.258 * cos(4.08894525395  + 423.66061462181 *T)
    sumX1  +=       7.648 * cos(2.32798852949  + 2001.68780964171 *T)
    sumX1  +=       8.081 * cos(2.88299363492  + 1586.13531902961 *T)
    sumX1  +=       7.334 * cos(5.14442280849  + 529.18402305270 *T)
    sumX1  +=       7.334 * cos(5.47944814706  + 530.68554210351 *T)
    sumX1  +=       7.088 * cos(2.33493448028  + 13.98327651810 *T)
    sumX1  +=       6.944 * cos(2.85039309523  + 99.40443843901 *T)
    sumX1  +=       7.575 * cos(5.37079456066  + 121.49583896680 *T)
    sumX1  +=       6.848 * cos(0.78796625546  + 1898.10740045609 *T)
    sumX1  +=       6.870 * cos(1.15013907149  + 1158.78636862821 *T)
    sumX1  +=       6.772 * cos(5.07892665697  + 1048.58004740881 *T)
    sumX1  +=       7.746 * cos(1.97831751281  + 1691.92185201890 *T)
    sumX1  +=       6.468 * cos(0.83654880503  + 1165.89991562901 *T)
    sumX1  +=       6.936 * cos(0.21365116245  + 953.35157971640 *T)
    sumX1  +=       6.382 * cos(2.33142676077  + 1159.53712815361 *T)
    sumX1  +=       6.464 * cos(3.96902820569  + 113.89465699899 *T)
    sumX1  +=       6.270 * cos(1.41636529387  + 387.48513244430 *T)
    sumX1  +=       6.557 * cos(0.09772332548  + 636.20895053440 *T)
    sumX1  +=       7.405 * cos(3.41801223400  + 405.50136735710 *T)
    sumX1  +=       6.424 * cos(2.56898493091  + 312.70353387701 *T)
    sumX1  +=       6.311 * cos(4.22916699759  + 983.35967639710 *T)
    sumX1  +=       7.257 * cos(4.97217454857  + 1251.09656714130 *T)
    sumX1  +=       6.999 * cos(1.52695047635  + 323.26159917390 *T)
    sumX1  +=       5.842 * cos(2.07617926504  + 1166.65067515440 *T)
    sumX1  +=       6.084 * cos(6.08154567006  + 2538.49232173710 *T)
    sumX1  +=       5.618 * cos(1.65396088949  + 71.09326278771 *T)
    sumX1  +=       5.590 * cos(1.91532661942  + 234.88355392390 *T)
    sumX1  +=       5.829 * cos(0.16935049171  + 1041.46650040800 *T)
    sumX1  +=       6.815 * cos(4.44264173400  + 305.10235190919 *T)
    sumX1  +=       5.423 * cos(5.67735456927  + 114.38229196601 *T)
    sumX1  +=       5.434 * cos(2.90645289424  + 1251.58420210830 *T)
    sumX1  +=       5.984 * cos(2.75504343237  + 319.81708087780 *T)
    sumX1  +=       5.352 * cos(2.47648610539  + 415.79630809560 *T)
    sumX1  +=       5.826 * cos(2.69993180753  + 1891.48148842230 *T)
    sumX1  +=       5.519 * cos(3.64099912123  + 1593.24886603040 *T)
    sumX1  +=       5.189 * cos(0.91842109203  + 448.18264936190 *T)
    sumX1  +=       5.286 * cos(4.46266267934  + 505.55576018990 *T)
    sumX1  +=       5.876 * cos(4.11769538608  + 459.08533727390 *T)
    sumX1  +=       5.578 * cos(4.41596318124  + 437.88770862340 *T)
    sumX1  +=       4.953 * cos(3.44649774547  + 1780.78753223589 *T)
    sumX1  +=       4.928 * cos(4.68268114080  + 199.31581891990 *T)
    sumX1  +=       4.972 * cos(6.16120827874  + 554.31380496631 *T)
    sumX1  +=       4.889 * cos(3.00346445874  + 117.07605073670 *T)
    sumX1  +=       5.942 * cos(3.32585444966  + 512.52629483880 *T)
    sumX1  +=       4.807 * cos(1.83942967124  + 494.51005992600 *T)
    sumX1  +=       5.528 * cos(3.37702275374  + 430.77416162260 *T)
    sumX1  +=       4.806 * cos(4.93123850862  + 956.04533848709 *T)
    sumX1  +=       6.190 * cos(0.33303059089  + 2751.79141717511 *T)
    sumX1  +=       5.306 * cos(1.69049301545  + 1069.92068841121 *T)
    sumX1  +=       4.939 * cos(3.92798617287  + 2207.38572311189 *T)
    sumX1  +=       5.203 * cos(1.87963213473  + 498.91529394111 *T)
    sumX1  +=       4.397 * cos(2.50128659500  + 565.35950523021 *T)
    sumX1  +=       4.342 * cos(2.91932588132  + 672.38443271190 *T)
    sumX1  +=       4.528 * cos(2.64641812201  + 1677.69475801730 *T)
    sumX1  +=       4.334 * cos(3.43386241697  + 611.68691579430 *T)
    sumX1  +=       4.608 * cos(0.94871692824  + 679.49797971270 *T)
    sumX1  +=       4.812 * cos(2.46105351857  + 560.95427121510 *T)
    sumX1  +=       4.305 * cos(3.17292686307  + 1070.67144793660 *T)
    sumX1  +=       4.662 * cos(6.19278511564  + 621.98185653281 *T)
    sumX1  +=       4.362 * cos(0.81856641043  + 81.50831573269 *T)
    sumX1  +=       4.781 * cos(0.23034918069  + 600.78422788230 *T)
    sumX1  +=       4.148 * cos(1.91465810869  + 1080.96638867510 *T)
    sumX1  +=       4.478 * cos(0.83734666663  + 629.09540353360 *T)
    sumX1  +=       4.108 * cos(1.85129748483  + 644.07325706060 *T)
    sumX1  +=       4.654 * cos(4.97986041266  + 508.10650660869 *T)
    sumX1  +=       4.260 * cos(4.99902513065  + 209.87388421679 *T)
    sumX1  +=       5.131 * cos(0.48588112186  + 946.23803271561 *T)
    sumX1  +=       3.845 * cos(1.59767295215  + 970.76006745570 *T)
    sumX1  +=       3.839 * cos(2.82406421175  + 2420.68481854989 *T)
    sumX1  +=       3.898 * cos(5.93876139914  + 1064.29161411871 *T)
    sumX1  +=       4.274 * cos(2.31117388618  + 31.99951143090 *T)
    sumX1  +=       3.986 * cos(5.18457061537  + 952.60082019101 *T)
    sumX1  +=       3.880 * cos(5.29550041805  + 504.80500066450 *T)
    sumX1  +=       5.205 * cos(2.65346011281  + 1795.01462623750 *T)
    sumX1  +=       3.879 * cos(1.78725414335  + 747.16603127920 *T)
    sumX1  +=       3.965 * cos(1.33979952763  + 501.62360692681 *T)
    sumX1  +=       4.541 * cos(1.67944661862  + 757.46097201770 *T)
    sumX1  +=       3.663 * cos(2.53934739854  + 2318.07967929830 *T)
    sumX1  +=       4.938 * cos(5.95351997396  + 294.80741117069 *T)
    sumX1  +=       3.643 * cos(5.32837052634  + 555.06456449170 *T)
    sumX1  +=       3.850 * cos(1.68007336193  + 835.28095197081 *T)
    sumX1  +=       3.693 * cos(3.00088611994  + 558.24595822940 *T)
    sumX1  +=       3.818 * cos(5.13364214511  + 2325.19322629910 *T)
    sumX1  +=       3.422 * cos(0.51968162936  + 2214.49927011270 *T)
    sumX1  +=       3.726 * cos(1.58879941368  + 740.05248427840 *T)
    sumX1  +=       3.306 * cos(2.13594948327  + 2641.58509595570 *T)
    sumX1  +=       3.331 * cos(2.97883947073  + 501.48059457490 *T)
    sumX1  +=       3.289 * cos(2.06996431796  + 1066.25165970650 *T)
    sumX1  +=       4.224 * cos(0.94453108652  + 547.34327031741 *T)
    sumX1  +=       3.214 * cos(6.21073356050  + 498.44221318910 *T)
    sumX1  +=       3.771 * cos(3.19812168530  + 1493.33748554950 *T)
    sumX1  +=       3.357 * cos(4.06089125495  + 1272.43720814370 *T)
    sumX1  +=       3.469 * cos(3.39231437882  + 1678.18239298430 *T)
    sumX1  +=       3.147 * cos(0.44528787143  + 654.36819779910 *T)
    sumX1  +=       3.122 * cos(5.26948049817  + 1912.82212942470 *T)
    sumX1  +=       3.766 * cos(1.55302482141  + 1567.97607176490 *T)
    sumX1  +=       3.766 * cos(4.69461747500  + 1567.48843679789 *T)
    sumX1  +=       3.651 * cos(1.76784115018  + 2008.31372167550 *T)
    sumX1  +=       3.350 * cos(3.60014817718  + 6276.20612047410 *T)
    sumX1  +=       2.969 * cos(4.41313741154  + 561.42735196710 *T)
    sumX1  +=       2.926 * cos(2.44841800418  + 824.98601123230 *T)
    sumX1  +=       3.251 * cos(3.05741893468  + 6290.43321447570 *T)
    sumX1  +=       3.134 * cos(5.02262053935  + 10206.41581669370 *T)
    sumX1  +=       3.154 * cos(2.88790610293  + 483.46435966210 *T)
    sumX1  +=       2.968 * cos(2.52814022000  + 597.60283414461 *T)
    sumX1  +=       3.023 * cos(5.26034603701  + 938.37372618940 *T)
    sumX1  +=       3.496 * cos(5.37662649242  + 1054.95988122670 *T)
    sumX1  +=       2.718 * cos(1.12612274568  + 1151.67282162740 *T)
    sumX1  +=       3.041 * cos(4.48045335531  + 10220.64291069530 *T)
    sumX1  +=       2.652 * cos(1.50411451235  + 960.46512671720 *T)
    sumX1  +=       2.720 * cos(2.58012487497  + 326.93062787860 *T)
    sumX1  +=       2.672 * cos(2.37743868437  + 537.79908910430 *T)
    sumX1  +=       2.672 * cos(1.96324696400  + 522.07047605190 *T)
    sumX1  +=       2.641 * cos(1.93870259756  + 2531.37877473630 *T)
    sumX1  +=       2.658 * cos(2.71532856876  + 1062.07342822750 *T)
    sumX1  +=       2.567 * cos(2.71690582841  + 415.30867312860 *T)
    sumX1  +=       2.847 * cos(2.92204886379  + 1176.94561589290 *T)
    sumX1  +=       2.545 * cos(0.06405909754  + 1098.98262358790 *T)
    sumX1  +=       2.658 * cos(0.64454112605  + 1057.17806711790 *T)
    sumX1  +=       3.084 * cos(4.14043630979  + 320.56784040320 *T)
    sumX1  +=       3.075 * cos(1.56268674394  + 462.26673101160 *T)
    sumX1  +=       2.804 * cos(1.45328370683  + 576.40520549410 *T)
    sumX1  +=       2.606 * cos(3.64826504072  + 2523.77759276850 *T)
    sumX1  +=       3.000 * cos(2.08528609490  + 230.32075334190 *T)
    sumX1  +=       2.425 * cos(2.52434943707  + 422.42222012939 *T)
    sumX1  +=       2.483 * cos(3.69036238721  + 209.12312469139 *T)
    sumX1  +=       2.390 * cos(0.07414982415  + 210.36151918381 *T)
    sumX1  +=       2.327 * cos(3.99650789580  + 416.05943265400 *T)
    sumX1  +=       2.797 * cos(1.87832948272  + 2332.30677329990 *T)
    sumX1  +=       2.273 * cos(4.89984409290  + 945.48727319021 *T)
    sumX1  +=       2.980 * cos(5.28454893190  + 202.76033721600 *T)
    sumX1  +=       2.267 * cos(3.20625354263  + 2097.17940189250 *T)
    sumX1  +=       2.330 * cos(4.18274089040  + 849.50804597240 *T)
    sumX1  +=       2.940 * cos(2.32020734015  + 2111.40649589409 *T)
    sumX1  +=       2.403 * cos(6.15190475337  + 433.46792039329 *T)
    sumX1  +=       2.256 * cos(2.98060305636  + 1054.20912170131 *T)
    sumX1  +=       2.203 * cos(5.95488550794  + 217.47506618460 *T)
    sumX1  +=       2.189 * cos(4.68499930418  + 1514.53511420000 *T)
    sumX1  +=       2.120 * cos(3.75884388832  + 423.17297965479 *T)
    sumX1  +=       2.093 * cos(1.34498726275  + 543.67424161270 *T)
    sumX1  +=       2.373 * cos(4.97528466796  + 536.15455770160 *T)
    sumX1  +=       2.373 * cos(5.64858628759  + 523.71500745461 *T)
    sumX1  +=       2.040 * cos(2.72132861555  + 988.77630236850 *T)
    sumX1  +=       2.105 * cos(3.22262200918  + 1059.57756356290 *T)
    sumX1  +=       2.005 * cos(1.80102545599  + 460.78225830330 *T)
    sumX1  +=       1.997 * cos(1.58652758554  + 504.66198831261 *T)
    sumX1  +=       2.033 * cos(2.03694665564  + 1173.76422215520 *T)
    sumX1  +=       1.976 * cos(3.51326639371  + 46.71424039951 *T)
    sumX1  +=       2.284 * cos(1.65248273100  + 531.57931398080 *T)
    sumX1  +=       2.284 * cos(2.68820291737  + 528.29025117541 *T)
    sumX1  +=       2.219 * cos(5.00103752698  + 5230.56364931950 *T)
    sumX1  +=       2.084 * cos(5.36575264228  + 490.57790666290 *T)
    sumX1  +=       2.020 * cos(4.84607190143  + 568.06781821590 *T)
    sumX1  +=       1.939 * cos(2.20836538491  + 1065.04237364410 *T)
    sumX1  +=       1.972 * cos(2.67204098380  + 2428.28600051770 *T)
    sumX1  +=       2.080 * cos(3.74170619496  + 194.89603068979 *T)
    sumX1  +=       1.994 * cos(5.72172751294  + 749.85979004990 *T)
    sumX1  +=       1.905 * cos(3.51155751559  + 1077.78499493740 *T)
    sumX1  +=       1.909 * cos(2.72297523968  + 2737.07668820650 *T)
    sumX1  +=       1.874 * cos(2.75415806536  + 555.20757684360 *T)
    sumX1  +=       2.388 * cos(4.22827702093  + 334.79493440480 *T)
    sumX1  +=       2.111 * cos(5.51827565803  + 5216.33655531789 *T)
    sumX1  +=       2.086 * cos(3.31238187572  + 1062.54650897950 *T)
    sumX1  +=       2.063 * cos(0.14160166361  + 9160.77334553909 *T)
    sumX1  +=       1.930 * cos(2.31808019838  + 2207.87335807890 *T)
    sumX1  +=       1.877 * cos(1.77046886437  + 732.93893727761 *T)
    sumX1  +=       1.836 * cos(2.50051244655  + 1994.08662767389 *T)
    sumX1  +=       1.907 * cos(4.49445370726  + 824.49837626530 *T)
    sumX1  +=       1.819 * cos(5.68655037935  + 491.80174694030 *T)
    sumX1  +=       1.892 * cos(5.26005735036  + 569.29165849331 *T)
    sumX1  +=       1.770 * cos(4.43317433717  + 842.39449897160 *T)
    sumX1  +=       1.786 * cos(5.25911906099  + 327.68138740401 *T)
    sumX1  +=       1.761 * cos(6.10459116553  + 67.42423408300 *T)
    sumX1  +=       1.872 * cos(0.48305166447  + 1169.34443392509 *T)
    sumX1  +=       1.951 * cos(1.20156127405  + 142.69346761730 *T)
    sumX1  +=       1.737 * cos(3.47351315761  + 202.00957769060 *T)
    sumX1  +=       1.965 * cos(0.65764447635  + 9146.54625153749 *T)
    sumX1  +=       1.703 * cos(0.22482006123  + 2530.89113976930 *T)
    sumX1  +=       1.717 * cos(3.13603496824  + 380.37158544351 *T)
    sumX1  +=       1.684 * cos(2.53972202455  + 599.08730685290 *T)
    sumX1  +=       1.669 * cos(2.13331218955  + 2634.47154895491 *T)
    sumX1  +=       1.707 * cos(4.70440451897  + 92.29089143820 *T)
    sumX1  +=       1.666 * cos(6.08661752005  + 1058.66253982621 *T)
    sumX1  +=       2.249 * cos(0.32860344346  + 852.95256426850 *T)
    sumX1  +=       2.105 * cos(0.13724782964  + 1059.67393178251 *T)
    sumX1  +=       1.971 * cos(4.40376628788  + 1883.88030645450 *T)
    sumX1  +=       1.682 * cos(3.48278576722  + 1905.22094745690 *T)
    sumX1  +=       1.591 * cos(3.28337636744  + 2435.39954751850 *T)
    sumX1  +=       1.870 * cos(0.19806856983  + 739.30172475301 *T)
    sumX1  +=       1.831 * cos(4.94275349020  + 405.01373239010 *T)
    sumX1  +=       1.741 * cos(5.44783098353  + 1280.03839011150 *T)
    sumX1  +=       1.437 * cos(1.72233118576  + 1056.70498636590 *T)
    sumX1  +=       1.666 * cos(3.55643756155  + 1060.58895551920 *T)
    sumX1  +=       1.646 * cos(4.89045222906  + 299.37021175271 *T)
    sumX1  +=       1.423 * cos(0.35443031090  + 1464.88329754630 *T)
    sumX1  +=       1.879 * cos(1.06964842472  + 531.41925528640 *T)
    sumX1  +=       1.879 * cos(3.27103722365  + 528.45030986980 *T)
    sumX1  +=       1.460 * cos(2.99740366174  + 1269.74344937300 *T)
    sumX1  +=       1.684 * cos(2.47723234250  + 1364.97191706540 *T)
    sumX1  +=       1.453 * cos(6.26510844895  + 526.26575387340 *T)
    sumX1  +=       1.700 * cos(0.15073432597  + 725.07463075140 *T)
    sumX1  +=       1.360 * cos(3.51016366749  + 917.17609753890 *T)
    sumX1  +=       1.426 * cos(3.38257862218  + 2840.16946242510 *T)
    sumX1  +=       1.498 * cos(5.40291606338  + 1134.40734624000 *T)
    sumX1  +=       1.410 * cos(0.25814998758  + 1091.86907658711 *T)
    sumX1  +=       1.520 * cos(0.66510646757  + 977.73060210460 *T)
    sumX1  +=       1.747 * cos(1.60904440146  + 2324.70559133209 *T)
    sumX1  +=       1.288 * cos(5.52938584132  + 732.18817775220 *T)
    sumX1  +=       1.713 * cos(2.03572918983  + 2427.79836555070 *T)
    sumX1  +=       1.282 * cos(0.82379362487  + 1475.17823828480 *T)
    sumX1  +=       1.517 * cos(3.20491072001  + 1439.75351563271 *T)
    sumX1  +=       1.271 * cos(5.76578238196  + 843.14525849700 *T)
    sumX1  +=       1.308 * cos(1.16466426738  + 558.38897058130 *T)
    sumX1  +=       1.677 * cos(5.68726278883  + 611.19928082730 *T)
    sumX1  +=       1.256 * cos(5.98835555735  + 2854.88419139370 *T)
    sumX1  +=       1.366 * cos(4.72040446216  + 1446.86706263350 *T)
    sumX1  +=       1.435 * cos(0.26721377820  + 991.95769610620 *T)
    sumX1  +=       1.144 * cos(0.57883702331  + 1482.29178528560 *T)
    sumX1  +=       1.142 * cos(2.82387136290  + 408.19512612779 *T)
    sumX1  +=       1.147 * cos(0.57710279821  + 3068.18328683170 *T)
    sumX1  +=       1.501 * cos(0.82356141490  + 75.02541605080 *T)
    sumX1  +=       1.226 * cos(5.02777555864  + 1141.37788088890 *T)
    sumX1  +=       1.323 * cos(0.45303236447  + 284.39235822571 *T)
    sumX1  +=       1.150 * cos(3.64731637589  + 430.28652665560 *T)
    sumX1  +=       1.292 * cos(3.22691222108  + 1994.57426264090 *T)
    sumX1  +=       1.226 * cos(0.90830574042  + 372.77040347569 *T)
    sumX1  +=       1.132 * cos(5.19504872814  + 1063.81853336670 *T)
    sumX1  +=       1.079 * cos(1.83169919397  + 1483.04254481100 *T)
    sumX1  +=       1.111 * cos(3.09681405003  + 2413.57127154910 *T)
    sumX1  +=       1.252 * cos(1.06008408031  + 473.31243127551 *T)
    sumX1  +=       1.242 * cos(4.51992844109  + 539.49601013370 *T)
    sumX1  +=       1.242 * cos(2.96234986087  + 520.37355502251 *T)
    sumX1  +=       1.169 * cos(0.37626190654  + 386.99749747730 *T)
    sumX1  +=       1.003 * cos(6.17382979199  + 216.98743121760 *T)
    sumX1  +=       1.030 * cos(5.35594718648  + 850.25880549781 *T)
    sumX1  +=       1.165 * cos(3.23424220409  + 2758.90496417590 *T)
    sumX1  +=       0.984 * cos(5.54057958497  + 522.77305146750 *T)
    sumX1  +=       0.984 * cos(1.94169871699  + 537.09651368871 *T)
    sumX1  +=       1.015 * cos(0.51819766710  + 451.22103074770 *T)
    sumX1  +=       1.310 * cos(4.02666314741  + 1354.67697632690 *T)
    sumX1  +=       1.121 * cos(0.67317844435  + 4164.06817212950 *T)
    sumX1  +=       0.956 * cos(2.02124975965  + 2950.86341861150 *T)
    sumX1  +=       0.940 * cos(0.33627498720  + 39.60069339870 *T)
    sumX1  +=       0.949 * cos(2.09328292248  + 1475.92899781020 *T)
    sumX1  +=       0.992 * cos(2.88019144631  + 1969.44448072730 *T)
    sumX1  +=       0.935 * cos(2.32097978196  + 3053.46855786309 *T)
    sumX1  +=       1.085 * cos(3.28059335022  + 586.55713388070 *T)
    sumX1  +=       0.901 * cos(2.95406440747  + 248.96763557360 *T)
    sumX1  +=       0.894 * cos(6.19201666773  + 2847.28300942589 *T)
    sumX1  +=       0.948 * cos(3.23576370557  + 2221.61281711350 *T)
    sumX1  +=       0.923 * cos(1.57948501053  + 2421.17245351690 *T)
    sumX1  +=       1.070 * cos(4.12769395162  + 2200.27217611110 *T)
    sumX1  +=       1.169 * cos(0.96073817242  + 100.40127618030 *T)
    sumX1  +=       1.005 * cos(5.20924948749  + 1432.63996863190 *T)
    sumX1  +=       1.017 * cos(2.10326291360  + 8094.27786834910 *T)
    sumX1  +=       0.984 * cos(5.02707289654  + 537.00014546910 *T)
    sumX1  +=       0.984 * cos(2.45520540542  + 522.86941968710 *T)
    sumX1  +=       1.118 * cos(2.99407073283  + 416.54706762100 *T)
    sumX1  +=       0.914 * cos(4.24495681812  + 1140.89024592190 *T)
    sumX1  +=       0.977 * cos(6.06165201031  + 312.96665843540 *T)
    sumX1  +=       0.840 * cos(2.42530575361  + 1579.02177202880 *T)
    sumX1  +=       0.891 * cos(1.62042692734  + 486.64575339981 *T)
    sumX1  +=       0.849 * cos(5.66050445306  + 834.79331700380 *T)
    sumX1  +=       0.926 * cos(1.06469662901  + 532.62854134880 *T)
    sumX1  +=       0.832 * cos(3.82248800441  + 608.64853440851 *T)
    sumX1  +=       0.975 * cos(1.83569716432  + 21.09682351890 *T)
    sumX1  +=       0.794 * cos(0.55277759585  + 1791.57010794140 *T)
    sumX1  +=       1.006 * cos(3.29673383179  + 2310.96613229750 *T)
    sumX1  +=       0.825 * cos(0.24960267153  + 8616.85528644289 *T)
    sumX1  +=       0.827 * cos(2.11307142672  + 2545.60586873790 *T)
    sumX1  +=       0.955 * cos(5.80717048495  + 760.01171843649 *T)
    sumX1  +=       0.955 * cos(5.80717048495  + 760.49935340350 *T)
    sumX1  +=       0.936 * cos(5.99339737103  + 155.53915477459 *T)
    sumX1  +=       0.779 * cos(1.83263011252  + 1792.32086746680 *T)
    sumX1  +=       0.766 * cos(2.28443434916  + 3811.17334736430 *T)
    sumX1  +=       0.773 * cos(2.27805515523  + 738.81408978600 *T)
    sumX1  +=       0.788 * cos(3.85951320467  + 732.45130231059 *T)
    sumX1  +=       0.801 * cos(1.48502354739  + 77204.08367704989 *T)
    sumX1  +=       0.937 * cos(4.69566509817  + 297.98880490840 *T)
    sumX1  +=       0.779 * cos(5.36095004640  + 538.01153742540 *T)
    sumX1  +=       0.779 * cos(2.12132825556  + 521.85802773080 *T)
    sumX1  +=       0.755 * cos(0.53346289480  + 8631.08238044450 *T)
    sumX1  +=       0.743 * cos(5.06914451090  + 1060.14701253450 *T)
    sumX1  +=       0.790 * cos(4.76601235989  + 902.94900353731 *T)
    sumX1  +=       0.897 * cos(5.90672024499  + 1485.73630358170 *T)
    sumX1  +=       0.991 * cos(1.76419853445  + 2744.19023520730 *T)
    sumX1  +=       0.740 * cos(0.93130456921  + 277.27881122490 *T)
    sumX1  +=       0.780 * cos(0.67747505651  + 606.20085385370 *T)
    sumX1  +=       0.738 * cos(3.53234955097  + 4856.32818355191 *T)
    sumX1  +=       0.799 * cos(2.72025873636  + 573.22381175640 *T)
    sumX1  +=       0.802 * cos(3.92777211068  + 984.84414910540 *T)
    sumX1  +=       0.761 * cos(2.31403909443  + 2744.67787017430 *T)
    sumX1  +=       0.735 * cos(5.48605031473  + 4650.14263511470 *T)
    sumX1  +=       0.721 * cos(2.25748057238  + 731.70054278519 *T)
    sumX1  +=       0.850 * cos(3.90105539794  + 775.47720693050 *T)
    sumX1  +=       0.717 * cos(3.64971882351  + 5746.02752041250 *T)
    sumX1  +=       0.754 * cos(4.58241793030  + 202.49721265760 *T)
    sumX1  +=       0.787 * cos(6.16479913324  + 320.08020543620 *T)
    sumX1  +=       0.739 * cos(5.82129294779  + 967.57867371801 *T)
    sumX1  +=       0.744 * cos(1.81557790720  + 2847.77064439291 *T)
    sumX1  +=       0.684 * cos(0.30776904368  + 1798.68365494221 *T)
    sumX1  +=       0.700 * cos(6.09650846364  + 572.47305223100 *T)
    sumX1  +=       0.700 * cos(4.52736249190  + 487.39651292520 *T)
    sumX1  +=       0.690 * cos(6.09169936894  + 2118.52004289490 *T)
    sumX1  +=       0.670 * cos(1.19443093892  + 245.29860686890 *T)
    sumX1  +=       0.668 * cos(3.63109061354  + 739.56484931140 *T)
    sumX1  +=       0.667 * cos(0.87583989763  + 1013.15532475670 *T)
    sumX1  +=       0.762 * cos(4.85397542040  + 9153.65979853830 *T)
    sumX1  +=       0.711 * cos(3.11169852570  + 3156.56133208170 *T)
    sumX1  +=       0.679 * cos(3.67936511602  + 24498.58642880689 *T)
    sumX1  +=       0.729 * cos(0.09495555814  + 4547.04986089611 *T)
    sumX1  +=       0.692 * cos(6.19117624873  + 5062.51373198910 *T)
    sumX1  +=       0.649 * cos(1.56313876075  + 1799.43441446760 *T)
    sumX1  +=       0.677 * cos(0.68187945383  + 10.05112325500 *T)
    sumX1  +=       0.784 * cos(0.71180416773  + 3281.48238226970 *T)
    sumX1  +=       0.733 * cos(3.85073191118  + 453.66871130251 *T)
    sumX1  +=       0.643 * cos(2.57202346935  + 2310.47849733050 *T)
    sumX1  +=       0.658 * cos(0.00955649379  + 305.85311143459 *T)
    sumX1  +=       0.784 * cos(0.92101367041  + 74.53778108379 *T)
    sumX1  +=       0.704 * cos(0.42049120814  + 5165.60650620771 *T)
    sumX1  +=       0.648 * cos(5.17262659684  + 1382.64352936309 *T)
    sumX1  +=       0.786 * cos(0.17119550214  + 1290.19031849811 *T)
    sumX1  +=       0.844 * cos(4.57333845552  + 312.21589891000 *T)
    sumX1  +=       0.703 * cos(3.18188887722  + 2288.10022602790 *T)
    sumX1  +=       0.703 * cos(3.68329176097  + 2273.87313202630 *T)
    sumX1  +=       0.797 * cos(1.30518182972  + 2641.09746098870 *T)
    sumX1  +=       0.635 * cos(0.57438614150  + 397.90018538930 *T)
    sumX1  +=       0.607 * cos(2.87691632991  + 1689.22809324820 *T)
    sumX1  +=       0.626 * cos(1.96304211270  + 2524.26522773550 *T)
    sumX1  +=       0.636 * cos(3.42015645415  + 511.28790034640 *T)
    sumX1  +=       0.613 * cos(5.27453086724  + 1037.79747170330 *T)
    sumX1  +=       0.614 * cos(4.75169588095  + 2957.97696561230 *T)
    sumX1  +=       0.747 * cos(3.57894990009  + 5223.45010231870 *T)
    sumX1  +=       0.606 * cos(5.48577834442  + 1751.29571393250 *T)
    sumX1  +=       0.779 * cos(5.87445673484  + 523.78444342380 *T)
    sumX1  +=       0.779 * cos(1.60782156712  + 536.08512173241 *T)
    sumX1  +=       0.607 * cos(2.92319803825  + 2133.23477186350 *T)
    sumX1  +=       0.744 * cos(1.33557042847  + 643.32249753521 *T)
    sumX1  +=       0.697 * cos(5.37775664790  + 1354.18934135989 *T)
    sumX1  +=       0.669 * cos(4.82929349311  + 5069.62727898990 *T)
    sumX1  +=       0.762 * cos(3.36471924953  + 1262.62990237220 *T)
    sumX1  +=       0.765 * cos(5.32617475783  + 927.59115048390 *T)
    sumX1  +=       0.720 * cos(3.06031913027  + 3075.29683383250 *T)
    sumX1  +=       0.573 * cos(1.26296826321  + 394.59867944510 *T)
    sumX1  +=       0.634 * cos(1.38494179693  + 470.13103753780 *T)
    sumX1  +=       0.559 * cos(2.49825062445  + 445.00125562421 *T)
    sumX1  +=       0.671 * cos(3.15851735580  + 454.66554904380 *T)
    sumX1  +=       0.755 * cos(6.14426019579  + 105.29663728990 *T)
    sumX1  +=       0.567 * cos(0.62012587567  + 1887.54933515920 *T)
    sumX1  +=       0.593 * cos(4.42709128942  + 842.65762352999 *T)
    sumX1  +=       0.564 * cos(4.09053029799  + 1127.29379923920 *T)
    sumX1  +=       0.547 * cos(2.00134392136  + 4127.56521702090 *T)
    sumX1  +=       0.624 * cos(1.38182148571  + 25035.39094090229 *T)
    sumX1  +=       0.545 * cos(1.38452164358  + 810.90192958261 *T)
    sumX1  +=       0.641 * cos(2.77533084404  + 305.58998687620 *T)
    sumX1  +=       0.593 * cos(6.12084605791  + 1063.07026596880 *T)
    sumX1  +=       0.574 * cos(4.93770315169  + 1699.03539901970 *T)
    sumX1  +=       0.616 * cos(1.88320700397  + 25021.16384690069 *T)
    sumX1  +=       0.721 * cos(3.98272042008  + 1819.88128359270 *T)
    sumX1  +=       0.537 * cos(3.42267089674  + 782.59075393131 *T)
    sumX1  +=       0.597 * cos(2.50100642147  + 379.88395047650 *T)
    sumX1  +=       0.543 * cos(4.69778069172  + 216.72430665921 *T)
    sumX1  +=       0.552 * cos(1.82100424033  + 1268.99268984760 *T)
    sumX1  +=       0.534 * cos(1.45757104363  + 437.13694909800 *T)
    sumX1  +=       0.588 * cos(3.86118609001  + 460.29462333630 *T)
    sumX1  +=       0.524 * cos(2.90061539997  + 518.40144734719 *T)
    sumX1  +=       0.701 * cos(5.29663931018  + 1670.58121101650 *T)
    sumX1  +=       0.632 * cos(1.12269993255  + 4525.70921989370 *T)
    sumX1  +=       0.546 * cos(4.73323913866  + 89.10949770051 *T)
    sumX1  +=       0.496 * cos(6.08219208052  + 3171.27606105030 *T)
    sumX1  +=       0.501 * cos(0.04295581721  + 32.48714639791 *T)
    sumX1  +=       0.508 * cos(3.08019966842  + 827.67977000299 *T)
    sumX1  +=       0.493 * cos(5.43229903686  + 10022.08109758290 *T)
    sumX1  +=       0.550 * cos(3.74627720943  + 2516.66404576769 *T)
    sumX1  +=       0.546 * cos(2.95574391589  + 589.73852761840 *T)
    sumX1  +=       0.509 * cos(5.07088390420  + 4686.64559022330 *T)
    sumX1  +=       0.505 * cos(1.78087149512  + 754.27957828001 *T)
    sumX1  +=       0.479 * cos(4.72565705175  + 5223.93773728571 *T)
    sumX1  +=       0.461 * cos(0.99543408630  + 1595.94262480110 *T)
    sumX1  +=       0.472 * cos(5.67013290312  + 3163.67487908250 *T)
    sumX1  +=       0.472 * cos(2.52854024953  + 3164.16251404950 *T)
    sumX1  +=       0.509 * cos(0.84552728990  + 69.39634175830 *T)
    sumX1  +=       0.546 * cos(2.01514901803  + 2840.65709739210 *T)
    sumX1  +=       0.615 * cos(5.03969710796  + 519.15220687259 *T)
    sumX1  +=       0.579 * cos(2.45280217192  + 525.51499434800 *T)
    sumX1  +=       0.455 * cos(0.84441579380  + 1468.06469128401 *T)
    sumX1  +=       0.451 * cos(1.84241635349  + 614.86830953200 *T)
    sumX1  +=       0.451 * cos(0.27109462263  + 2107.96197759801 *T)
    sumX1  +=       0.450 * cos(5.54300486570  + 3480.06674873909 *T)
    sumX1  +=       0.450 * cos(2.40141221211  + 3480.55438370610 *T)
    sumX1  +=       0.447 * cos(1.49071427438  + 3267.25528826810 *T)
    sumX1  +=       0.467 * cos(1.66996550533  + 1490.15609181181 *T)
    sumX1  +=       0.472 * cos(3.97777002593  + 1457.28211557850 *T)
    sumX1  +=       0.462 * cos(0.76906085237  + 26617.83792415230 *T)
    sumX1  +=       0.440 * cos(3.74033577848  + 3384.08752152129 *T)
    sumX1  +=       0.440 * cos(0.59874312489  + 3384.57515648830 *T)
    sumX1  +=       0.453 * cos(6.16538796545  + 9154.14743350530 *T)
    sumX1  +=       0.440 * cos(3.20211909619  + 2538.00468677009 *T)
    sumX1  +=       0.616 * cos(4.82336088817  + 1560.37488979710 *T)
    sumX1  +=       0.437 * cos(5.35668949324  + 4966.53450477130 *T)
    sumX1  +=       0.433 * cos(5.44090883319  + 4700.87268422489 *T)
    sumX1  +=       0.439 * cos(1.57953949229  + 2108.71273712340 *T)
    sumX1  +=       0.452 * cos(4.81735477031  + 1551.18367712950 *T)
    sumX1  +=       0.483 * cos(5.25822765428  + 629.35852809199 *T)
    sumX1  +=       0.455 * cos(0.52481636476  + 1368.41643536150 *T)
    sumX1  +=       0.454 * cos(2.63963867488  + 1884.36794142151 *T)
    sumX1  +=       0.476 * cos(1.27217476850  + 3060.58210486390 *T)
    sumX1  +=       0.414 * cos(4.36171182245  + 533.37930087419 *T)
    sumX1  +=       0.434 * cos(2.83878055329  + 622.24498109120 *T)
    sumX1  +=       0.434 * cos(2.83878055329  + 622.73261605820 *T)
    sumX1  +=       0.423 * cos(4.52676867100  + 415.04804069769 *T)
    sumX1  +=       0.411 * cos(0.64284710617  + 4539.44867892829 *T)
    sumX1  +=       0.411 * cos(3.78443975976  + 4539.93631389530 *T)
    sumX1  +=       0.507 * cos(0.89077366361  + 3252.54055929950 *T)
    sumX1  +=       0.401 * cos(5.31716419569  + 1040.97886544099 *T)
    sumX1  +=       0.424 * cos(1.74548317826  + 1254.76559584600 *T)
    sumX1  +=       0.407 * cos(3.07774750544  + 665.27088571111 *T)
    sumX1  +=       0.428 * cos(2.52282816585  + 12036.21691740469 *T)
    sumX1  +=       0.487 * cos(0.27735959651  + 1901.77642916080 *T)
    sumX1  +=       0.484 * cos(2.76465853394  + 1357.85837006460 *T)
    sumX1  +=       0.435 * cos(2.99417344584  + 2729.96314120569 *T)
    sumX1  +=       0.435 * cos(2.99417344584  + 2730.45077617271 *T)
    sumX1  +=       0.393 * cos(2.13115023199  + 3803.57216539650 *T)
    sumX1  +=       0.393 * cos(5.27274288558  + 3804.05980036350 *T)
    sumX1  +=       0.396 * cos(3.86752181242  + 1020.26887175750 *T)
    sumX1  +=       0.413 * cos(2.47801246921  + 1902.52718868620 *T)
    sumX1  +=       0.438 * cos(1.11008336467  + 611.82992814620 *T)
    sumX1  +=       0.404 * cos(3.24724828605  + 2042.74160658630 *T)
    sumX1  +=       0.412 * cos(6.18108169678  + 4863.44173055270 *T)
    sumX1  +=       0.379 * cos(1.72808594571  + 4443.95708667750 *T)
    sumX1  +=       0.528 * cos(4.53641477848  + 721.40560204669 *T)
    sumX1  +=       0.392 * cos(3.97639191399  + 2001.20017467470 *T)
    sumX1  +=       0.379 * cos(5.06316213675  + 9676.23721663209 *T)
    sumX1  +=       0.379 * cos(5.06316213675  + 9676.72485159910 *T)
    sumX1  +=       0.372 * cos(1.50891676381  + 3273.88120030190 *T)
    sumX1  +=       0.372 * cos(4.65050941740  + 3274.36883526891 *T)

    # JUPITER X Coord
    sumX2  = 0
    sumX2  +=  202462.055 * cos(2.16872090770  + 529.93478257810 *T)
    sumX2  +=  132313.738 * cos(5.46375601476  + 0.24381748350 *T)
    sumX2  +=  123752.958 * cos(4.12933545743  + 522.82123557730 *T)
    sumX2  +=  121682.952 * cos(0.20529040863  + 537.04832957890 *T)
    sumX2  +=   35719.700 * cos(5.02591678600  + 1059.62574767270 *T)
    sumX2  +=   11394.908 * cos(1.72910824402  + 7.35736448430 *T)
    sumX2  +=    7053.660 * cos(0.73618974864  + 1066.73929467350 *T)
    sumX2  +=    5684.733 * cos(3.62301903765  + 1052.51220067191 *T)
    sumX2  +=    5262.466 * cos(4.26860629303  + 515.70768857651 *T)
    sumX2  +=    3815.469 * cos(5.55769158033  + 1589.31671276730 *T)
    sumX2  +=    3706.857 * cos(0.07620773039  + 544.16187657970 *T)
    sumX2  +=    1449.544 * cos(2.64864661582  + 110.45013870291 *T)
    sumX2  +=    1046.768 * cos(2.32475473247  + 426.84200835950 *T)
    sumX2  +=    1322.480 * cos(1.62777310244  + 14.47091148511 *T)
    sumX2  +=     858.857 * cos(1.87732946593  + 633.02755679670 *T)
    sumX2  +=     887.013 * cos(4.66395505443  + 526.75338884040 *T)
    sumX2  +=     866.197 * cos(5.95310584104  + 533.11617631580 *T)
    sumX2  +=     997.140 * cos(1.45788897945  + 323.74923414091 *T)
    sumX2  +=     563.455 * cos(1.62040359556  + 949.41942645330 *T)
    sumX2  +=     571.107 * cos(3.15712293933  + 1045.39865367110 *T)
    sumX2  +=     519.207 * cos(5.96329235901  + 526.00262931501 *T)
    sumX2  +=     607.942 * cos(2.84828776126  + 736.12033101530 *T)
    sumX2  +=     506.922 * cos(4.64788690486  + 533.86693584120 *T)
    sumX2  +=     485.331 * cos(1.60461640700  + 109.96250373590 *T)
    sumX2  +=     475.072 * cos(4.21127530175  + 529.44714761109 *T)
    sumX2  +=     458.270 * cos(1.23999732854  + 1596.43025976811 *T)
    sumX2  +=     418.781 * cos(2.61159590787  + 117.56368570370 *T)
    sumX2  +=     525.152 * cos(0.41184029160  + 102.84895673509 *T)
    sumX2  +=     374.437 * cos(5.55383016603  + 95.73540973430 *T)
    sumX2  +=     412.815 * cos(0.32973940832  + 103.33659170210 *T)
    sumX2  +=     320.928 * cos(4.24368441414  + 419.72846135871 *T)
    sumX2  +=     322.793 * cos(5.98934797904  + 2119.00767786191 *T)
    sumX2  +=     302.747 * cos(0.27601683091  + 205.94173095370 *T)
    sumX2  +=     280.926 * cos(3.69548040149  + 1582.20316576650 *T)
    sumX2  +=     243.003 * cos(3.14355186105  + 625.91400979590 *T)
    sumX2  +=     260.213 * cos(3.67206938453  + 956.53297345411 *T)
    sumX2  +=     223.206 * cos(0.88087474540  + 433.95555536030 *T)
    sumX2  +=     219.874 * cos(1.72167133402  + 942.30587945250 *T)
    sumX2  +=     202.904 * cos(0.48756623733  + 220.65645992230 *T)
    sumX2  +=     196.744 * cos(5.99963553149  + 1169.83206889211 *T)
    sumX2  +=     171.475 * cos(0.62368794765  + 6.86972951729 *T)
    sumX2  +=     195.569 * cos(5.98737256847  + 640.14110379750 *T)
    sumX2  +=     165.172 * cos(5.28462686074  + 316.63568714010 *T)
    sumX2  +=     163.913 * cos(0.52317209564  + 1073.85284167431 *T)
    sumX2  +=     164.904 * cos(3.55715812496  + 1155.60497489050 *T)
    sumX2  +=     138.497 * cos(4.42900020166  + 508.59414157570 *T)
    sumX2  +=     135.640 * cos(0.24898267917  + 198.82818395289 *T)
    sumX2  +=     170.382 * cos(1.32273885498  + 330.86278114170 *T)
    sumX2  +=     123.801 * cos(5.24272600896  + 743.23387801611 *T)
    sumX2  +=     105.922 * cos(2.67322273860  + 853.44019923550 *T)
    sumX2  +=     104.749 * cos(1.04943852587  + 522.33360061030 *T)
    sumX2  +=     115.508 * cos(1.14912313221  + 309.03450517230 *T)
    sumX2  +=     127.421 * cos(2.98933095130  + 729.00678401450 *T)
    sumX2  +=     104.374 * cos(2.46273763192  + 213.54291292150 *T)
    sumX2  +=      74.731 * cos(6.19108417338  + 551.27542358051 *T)
    sumX2  +=      76.056 * cos(4.16147111300  + 1265.81129610991 *T)
    sumX2  +=      84.864 * cos(4.83454215641  + 519.63984183961 *T)
    sumX2  +=      85.433 * cos(2.50456948380  + 213.05527795450 *T)
    sumX2  +=      70.250 * cos(1.40859137196  + 426.35437339250 *T)
    sumX2  +=      70.226 * cos(1.17306752912  + 206.42936592071 *T)
    sumX2  +=      65.869 * cos(3.83439660231  + 839.21310523390 *T)
    sumX2  +=      62.918 * cos(4.41103639601  + 412.61491435790 *T)
    sumX2  +=      67.605 * cos(5.76604168472  + 540.22972331660 *T)
    sumX2  +=      71.970 * cos(0.19069755890  + 419.24082639170 *T)
    sumX2  +=      53.452 * cos(1.25726970711  + 3.42521122121 *T)
    sumX2  +=      52.805 * cos(5.35851382984  + 412.12727939089 *T)
    sumX2  +=      55.599 * cos(1.47903917952  + 21.58445848590 *T)
    sumX2  +=      46.667 * cos(0.35527018271  + 227.77000692311 *T)
    sumX2  +=      48.911 * cos(0.17388705856  + 1062.80714141041 *T)
    sumX2  +=      46.046 * cos(0.12865245318  + 536.56069461190 *T)
    sumX2  +=      49.381 * cos(4.53307122517  + 1162.71852189130 *T)
    sumX2  +=      39.853 * cos(4.40185904785  + 1258.69774910910 *T)
    sumX2  +=      38.964 * cos(5.39489558888  + 309.52214013930 *T)
    sumX2  +=      38.082 * cos(5.82106990307  + 441.06910236111 *T)
    sumX2  +=      32.665 * cos(4.64318627449  + 1692.40948698591 *T)
    sumX2  +=      31.801 * cos(1.71799573531  + 2126.12122486270 *T)
    sumX2  +=      36.510 * cos(4.40226967417  + 1056.44435393500 *T)
    sumX2  +=      40.831 * cos(3.33818544372  + 1575.08961876571 *T)
    sumX2  +=      31.242 * cos(4.56981926575  + 618.80046279510 *T)
    sumX2  +=      32.827 * cos(0.30468062826  + 515.22005360950 *T)
    sumX2  +=      28.849 * cos(0.32258423359  + 1055.69359440960 *T)
    sumX2  +=      32.224 * cos(0.24894992339  + 1272.92484311070 *T)
    sumX2  +=      28.682 * cos(3.10923644274  + 1038.28510667030 *T)
    sumX2  +=      29.430 * cos(4.79051983333  + 1486.22393854870 *T)
    sumX2  +=      34.150 * cos(1.22671657337  + 301.92095817150 *T)
    sumX2  +=      32.088 * cos(0.95269083826  + 625.42637482889 *T)
    sumX2  +=      24.170 * cos(3.96546667147  + 832.09955823310 *T)
    sumX2  +=      23.462 * cos(5.23275815839  + 1059.13811270570 *T)
    sumX2  +=      23.622 * cos(1.03079849725  + 742.74624304909 *T)
    sumX2  +=      25.413 * cos(5.36457820071  + 1368.90407032850 *T)
    sumX2  +=      25.008 * cos(0.10163753957  + 2648.69864295650 *T)
    sumX2  +=      19.641 * cos(4.86937041561  + 1063.55790093581 *T)
    sumX2  +=      20.824 * cos(0.24917335293  + 1699.52303398670 *T)
    sumX2  +=      18.958 * cos(6.04301166316  + 838.72547026689 *T)
    sumX2  +=      19.085 * cos(1.84501493463  + 1479.11039154791 *T)
    sumX2  +=      20.467 * cos(6.09881928033  + 647.25465079831 *T)
    sumX2  +=      18.387 * cos(2.47199072977  + 1471.99684454710 *T)
    sumX2  +=      16.332 * cos(4.98457997872  + 728.51914904750 *T)
    sumX2  +=      18.784 * cos(6.11069390571  + 735.63269604830 *T)
    sumX2  +=      13.140 * cos(1.62064062804  + 1376.01761732930 *T)
    sumX2  +=      14.219 * cos(2.42409257216  + 124.67723270450 *T)
    sumX2  +=      12.676 * cos(5.96326773737  + 831.61192326610 *T)
    sumX2  +=      12.289 * cos(1.04197442517  + 10.53875822200 *T)
    sumX2  +=      11.203 * cos(3.16486860174  + 511.77553531340 *T)
    sumX2  +=      12.667 * cos(1.06449203791  + 963.64652045490 *T)
    sumX2  +=      13.805 * cos(3.83367170604  + 2111.89413086110 *T)
    sumX2  +=      10.749 * cos(1.17035514019  + 548.09402984280 *T)
    sumX2  +=      12.863 * cos(6.04094024122  + 750.34742501690 *T)
    sumX2  +=      12.393 * cos(3.91459301015  + 1685.29593998510 *T)
    sumX2  +=      12.549 * cos(0.60189314422  + 941.81824448550 *T)
    sumX2  +=      10.468 * cos(1.87537642305  + 935.19233245171 *T)
    sumX2  +=       8.995 * cos(3.32040576908  + 1383.13116433011 *T)
    sumX2  +=      11.159 * cos(0.93936211173  + 618.31282782810 *T)
    sumX2  +=       7.631 * cos(3.48899971728  + 632.53992182970 *T)
    sumX2  +=       7.503 * cos(0.95949200774  + 1603.54380676890 *T)
    sumX2  +=       7.708 * cos(0.38373971235  + 191.71463695210 *T)
    sumX2  +=       7.383 * cos(5.84748639962  + 1155.11733992350 *T)
    sumX2  +=       8.531 * cos(5.41891495754  + 1361.79052332771 *T)
    sumX2  +=       8.011 * cos(1.16879967744  + 337.97632814250 *T)
    sumX2  +=       7.223 * cos(2.27526633054  + 88.62186273350 *T)
    sumX2  +=       7.778 * cos(6.08089864211  + 1052.02456570489 *T)
    sumX2  +=       7.476 * cos(3.16644140740  + 721.89323701370 *T)
    sumX2  +=       5.663 * cos(5.09194448812  + 2222.10045208050 *T)
    sumX2  +=       5.661 * cos(4.71699843448  + 1044.91101870410 *T)
    sumX2  +=       5.318 * cos(5.67177401754  + 1148.00379292269 *T)
    sumX2  +=       4.988 * cos(5.11228308670  + 639.65346883050 *T)
    sumX2  +=       5.093 * cos(1.01868695076  + 106.03035047280 *T)
    sumX2  +=       6.533 * cos(1.90999732711  + 845.83901726770 *T)
    sumX2  +=       6.385 * cos(1.92471190846  + 860.55374623631 *T)
    sumX2  +=       4.988 * cos(6.12592952219  + 99.66756299740 *T)
    sumX2  +=       5.571 * cos(4.38524037140  + 1795.50226120450 *T)
    sumX2  +=       5.541 * cos(3.99278179683  + 1049.33080693420 *T)
    sumX2  +=       5.247 * cos(0.31216331680  + 1258.21011414209 *T)
    sumX2  +=       4.445 * cos(5.36838660295  + 846.32665223470 *T)
    sumX2  +=       3.990 * cos(0.03159217662  + 1802.61580820530 *T)
    sumX2  +=       3.926 * cos(3.73297740864  + 497.69145366371 *T)
    sumX2  +=       3.892 * cos(5.41749662246  + 302.40859313851 *T)
    sumX2  +=       3.809 * cos(2.16830295705  + 2.93757625420 *T)
    sumX2  +=       3.620 * cos(0.60770824733  + 562.17811149250 *T)
    sumX2  +=       4.760 * cos(0.65404095225  + 934.70469748469 *T)
    sumX2  +=       4.154 * cos(0.58444106217  + 96.22304470130 *T)
    sumX2  +=       3.877 * cos(0.63005503524  + 1375.52998236229 *T)
    sumX2  +=       3.522 * cos(4.66663934261  + 501.48059457490 *T)
    sumX2  +=       3.445 * cos(4.10178765579  + 1148.49142788971 *T)
    sumX2  +=       3.382 * cos(0.17821445242  + 234.88355392390 *T)
    sumX2  +=       3.592 * cos(5.23722943024  + 405.50136735710 *T)
    sumX2  +=       3.808 * cos(3.92687517258  + 1788.38871420371 *T)
    sumX2  +=       3.336 * cos(5.59164932879  + 1471.50920958009 *T)
    sumX2  +=       3.758 * cos(2.02589350773  + 113.63153244060 *T)
    sumX2  +=       3.174 * cos(5.47701611962  + 448.18264936190 *T)
    sumX2  +=       3.698 * cos(5.72579798285  + 220.16882495529 *T)
    sumX2  +=       3.101 * cos(1.61252122505  + 316.14805217309 *T)
    sumX2  +=       2.746 * cos(2.60898959388  + 81.50831573269 *T)
    sumX2  +=       2.816 * cos(5.64817967676  + 2015.91490364330 *T)
    sumX2  +=       2.754 * cos(0.98878483675  + 430.02340209721 *T)
    sumX2  +=       2.783 * cos(3.38652113978  + 2104.78058386030 *T)
    sumX2  +=       3.501 * cos(6.18907958990  + 1069.92068841121 *T)
    sumX2  +=       2.580 * cos(1.13656236739  + 98.91680347199 *T)
    sumX2  +=       2.605 * cos(5.15330137752  + 611.68691579430 *T)
    sumX2  +=       2.955 * cos(0.62556987115  + 1592.49810650500 *T)
    sumX2  +=       3.418 * cos(5.62641717211  + 1368.41643536150 *T)
    sumX2  +=       2.482 * cos(4.62391263198  + 1361.30288836070 *T)
    sumX2  +=       2.409 * cos(0.93859403652  + 107.26874496520 *T)
    sumX2  +=       2.450 * cos(0.25718396069  + 948.93179148630 *T)
    sumX2  +=       2.454 * cos(4.64671390500  + 1251.58420210830 *T)
    sumX2  +=       2.443 * cos(5.51785304335  + 1464.39566257930 *T)
    sumX2  +=       2.848 * cos(4.80044770776  + 1898.59503542310 *T)
    sumX2  +=       2.434 * cos(0.16399375449  + 1080.96638867510 *T)
    sumX2  +=       2.240 * cos(3.34539121485  + 629.84616305900 *T)
    sumX2  +=       2.676 * cos(1.36442566154  + 294.80741117069 *T)
    sumX2  +=       2.252 * cos(5.77491969067  + 423.66061462181 *T)
    sumX2  +=       2.207 * cos(5.98168647453  + 106.78110999820 *T)
    sumX2  +=       2.966 * cos(2.18772217641  + 106.51798543980 *T)
    sumX2  +=       2.114 * cos(1.40054172319  + 117.07605073670 *T)
    sumX2  +=       2.234 * cos(5.28783438446  + 6276.20612047410 *T)
    sumX2  +=       2.232 * cos(1.36347902456  + 6290.43321447570 *T)
    sumX2  +=       2.060 * cos(2.18476343972  + 2655.81218995730 *T)
    sumX2  +=       2.492 * cos(6.26929425827  + 1574.60198379870 *T)
    sumX2  +=       2.405 * cos(1.83175482621  + 1162.23088692429 *T)
    sumX2  +=       2.094 * cos(2.69213826264  + 13.98327651810 *T)
    sumX2  +=       1.997 * cos(4.16493009666  + 2008.80135664250 *T)
    sumX2  +=       1.998 * cos(5.23976804747  + 149.80701461810 *T)
    sumX2  +=       2.075 * cos(0.42755107557  + 10206.41581669370 *T)
    sumX2  +=       2.073 * cos(2.78638288310  + 10220.64291069530 *T)
    sumX2  +=       1.833 * cos(4.18008301702  + 824.98601123230 *T)
    sumX2  +=       1.940 * cos(0.37528581380  + 199.31581891990 *T)
    sumX2  +=       1.989 * cos(0.05550838377  + 323.26159917390 *T)
    sumX2  +=       1.764 * cos(6.26232235027  + 970.76006745570 *T)
    sumX2  +=       2.079 * cos(1.73475090893  + 1905.70858242391 *T)
    sumX2  +=       1.815 * cos(0.03040681609  + 1265.32366114289 *T)
    sumX2  +=       1.815 * cos(5.85864386233  + 558.38897058130 *T)
    sumX2  +=       1.819 * cos(5.36203973035  + 1165.89991562901 *T)
    sumX2  +=       2.303 * cos(0.38034815736  + 1251.09656714130 *T)
    sumX2  +=       1.797 * cos(0.77607919875  + 1585.38455950421 *T)
    sumX2  +=       1.785 * cos(4.17120693216  + 1159.53712815361 *T)
    sumX2  +=       1.719 * cos(4.75204120917  + 654.36819779910 *T)
    sumX2  +=       2.106 * cos(0.43427607956  + 508.10650660869 *T)
    sumX2  +=       1.746 * cos(4.42490277475  + 1586.13531902961 *T)
    sumX2  +=       1.693 * cos(0.29049773833  + 1691.92185201890 *T)
    sumX2  +=       1.560 * cos(4.06032111240  + 2001.68780964171 *T)
    sumX2  +=       1.635 * cos(5.33907949351  + 1787.90107923670 *T)
    sumX2  +=       1.492 * cos(1.04533558456  + 422.90985509640 *T)
    sumX2  +=       1.558 * cos(4.80176978113  + 636.20895053440 *T)
    sumX2  +=       1.753 * cos(0.13061364698  + 757.46097201770 *T)
    sumX2  +=       1.757 * cos(0.60186557399  + 3178.38960805111 *T)
    sumX2  +=       1.411 * cos(3.25105098307  + 956.04533848709 *T)
    sumX2  +=       1.925 * cos(3.26959207194  + 1567.97607176490 *T)
    sumX2  +=       1.563 * cos(4.39444066103  + 1891.48148842230 *T)
    sumX2  +=       1.383 * cos(3.30206953671  + 5230.56364931950 *T)
    sumX2  +=       1.393 * cos(4.07489654140  + 1280.03839011150 *T)
    sumX2  +=       1.381 * cos(0.94379151453  + 5216.33655531789 *T)
    sumX2  +=       1.548 * cos(2.31717712129  + 946.23803271561 *T)
    sumX2  +=       1.239 * cos(6.05197725659  + 543.67424161270 *T)
    sumX2  +=       1.373 * cos(5.88274816049  + 1890.99385345530 *T)
    sumX2  +=       1.198 * cos(0.79241931733  + 326.93062787860 *T)
    sumX2  +=       1.220 * cos(0.92943402597  + 319.81708087780 *T)
    sumX2  +=       1.357 * cos(0.99707936433  + 2229.21399908130 *T)
    sumX2  +=       1.271 * cos(4.72505663718  + 9160.77334553909 *T)
    sumX2  +=       1.268 * cos(2.36679203236  + 9146.54625153749 *T)
    sumX2  +=       1.183 * cos(3.25890608260  + 636.95971005981 *T)
    sumX2  +=       1.254 * cos(4.50024168543  + 433.46792039329 *T)
    sumX2  +=       1.180 * cos(5.26395330152  + 1780.78753223589 *T)
    sumX2  +=       1.421 * cos(5.18461492344  + 1678.18239298430 *T)
    sumX2  +=       1.111 * cos(5.11243140232  + 910.06255053810 *T)
    sumX2  +=       1.135 * cos(5.56380149178  + 540.98048284201 *T)
    sumX2  +=       1.425 * cos(5.31039737830  + 1684.80830501809 *T)
    sumX2  +=       1.101 * cos(1.90347864340  + 518.88908231420 *T)
    sumX2  +=       1.052 * cos(2.44195881689  + 1493.33748554950 *T)
    sumX2  +=       0.954 * cos(0.56086592982  + 1066.25165970650 *T)
    sumX2  +=       1.305 * cos(2.11338311992  + 5223.45010231870 *T)
    sumX2  +=       0.932 * cos(3.38781033249  + 952.60082019101 *T)
    sumX2  +=       1.152 * cos(5.72701789718  + 3281.48238226970 *T)
    sumX2  +=       1.271 * cos(1.34644154763  + 1478.62275658089 *T)
    sumX2  +=       0.993 * cos(0.85140525988  + 2427.79836555070 *T)
    sumX2  +=       1.202 * cos(0.09319048042  + 1567.48843679789 *T)
    sumX2  +=       0.952 * cos(4.40542256077  + 2325.19322629910 *T)
    sumX2  +=       0.904 * cos(5.45851579572  + 1158.78636862821 *T)
    sumX2  +=       1.185 * cos(4.39736411513  + 2214.98690507971 *T)
    sumX2  +=       0.965 * cos(3.98841084090  + 2428.28600051770 *T)
    sumX2  +=       0.862 * cos(1.88793880359  + 5753.14106741329 *T)
    sumX2  +=       1.183 * cos(2.13750411181  + 953.35157971640 *T)
    sumX2  +=       0.858 * cos(4.03268202733  + 749.85979004990 *T)
    sumX2  +=       0.857 * cos(4.35000353836  + 1677.69475801730 *T)
    sumX2  +=       0.896 * cos(2.48010732688  + 1272.43720814370 *T)
    sumX2  +=       0.838 * cos(3.42424958269  + 740.05248427840 *T)
    sumX2  +=       0.810 * cos(3.56166190184  + 732.93893727761 *T)
    sumX2  +=       0.899 * cos(6.16076732455  + 1581.71553079950 *T)
    sumX2  +=       1.107 * cos(3.10071050352  + 1912.82212942470 *T)
    sumX2  +=       1.099 * cos(0.06283564514  + 21.09682351890 *T)
    sumX2  +=       0.856 * cos(1.83949886373  + 210.36151918381 *T)
    sumX2  +=       0.792 * cos(3.31088624099  + 9683.35076363289 *T)
    sumX2  +=       0.803 * cos(4.00091268006  + 2641.58509595570 *T)
    sumX2  +=       1.013 * cos(4.97621633767  + 852.95256426850 *T)
    sumX2  +=       0.790 * cos(4.03009368633  + 1166.65067515440 *T)
    sumX2  +=       0.927 * cos(0.83468452012  + 202.76033721600 *T)
    sumX2  +=       0.955 * cos(1.05973862416  + 611.19928082730 *T)
    sumX2  +=       0.848 * cos(6.24143056133  + 824.49837626530 *T)
    sumX2  +=       0.721 * cos(2.30689352602  + 4693.75913722409 *T)
    sumX2  +=       0.732 * cos(0.54378112448  + 1141.37788088890 *T)
    sumX2  +=       0.717 * cos(5.04553206489  + 2538.49232173710 *T)
    sumX2  +=       0.764 * cos(2.59169068637  + 3789.83270636190 *T)
    sumX2  +=       0.757 * cos(0.72064468326  + 422.42222012939 *T)
    sumX2  +=       0.870 * cos(3.27554624699  + 3900.03902758130 *T)
    sumX2  +=       0.745 * cos(5.82790167193  + 416.05943265400 *T)
    sumX2  +=       0.786 * cos(3.44863092226  + 3590.76070492550 *T)
    sumX2  +=       0.722 * cos(2.26242549861  + 1809.72935520611 *T)
    sumX2  +=       0.731 * cos(6.27879280918  + 2008.31372167550 *T)
    sumX2  +=       0.784 * cos(0.83516247967  + 1795.01462623750 *T)
    sumX2  +=       0.655 * cos(3.72870036603  + 8623.96883344369 *T)
    sumX2  +=       0.859 * cos(6.02318696083  + 459.08533727390 *T)
    sumX2  +=       0.787 * cos(5.46578436956  + 1671.06884598351 *T)
    sumX2  +=       0.668 * cos(2.03555764252  + 2545.60586873790 *T)
    sumX2  +=       0.707 * cos(5.16062924903  + 1169.34443392509 *T)
    sumX2  +=       0.679 * cos(5.40421465627  + 1588.82907780029 *T)
    sumX2  +=       0.721 * cos(5.17345152360  + 2104.29294889330 *T)
    sumX2  +=       0.869 * cos(3.62787610341  + 3576.53361092390 *T)
    sumX2  +=       0.878 * cos(0.78188086789  + 75.02541605080 *T)
    sumX2  +=       0.883 * cos(4.44629342777  + 3384.57515648830 *T)
    sumX2  +=       0.627 * cos(1.03465619593  + 405.01373239010 *T)
    sumX2  +=       0.776 * cos(0.79670919548  + 927.59115048390 *T)
    sumX2  +=       0.751 * cos(5.49611274777  + 3597.87425192630 *T)
    sumX2  +=       0.821 * cos(4.90808802861  + 3487.66793070691 *T)
    sumX2  +=       0.603 * cos(3.81342348025  + 2318.07967929830 *T)
    sumX2  +=       0.766 * cos(5.46484704372  + 1354.67697632690 *T)
    sumX2  +=       0.679 * cos(5.65072392656  + 2207.38572311189 *T)
    sumX2  +=       0.820 * cos(0.17433301545  + 6283.31966747490 *T)
    sumX2  +=       0.638 * cos(2.47793884910  + 3370.34806248671 *T)
    sumX2  +=       0.711 * cos(6.00519368625  + 1883.88030645450 *T)
    sumX2  +=       0.666 * cos(2.80033054652  + 3693.85347914410 *T)
    sumX2  +=       0.785 * cos(1.92447383793  + 3149.44778508089 *T)
    sumX2  +=       0.566 * cos(1.70107169172  + 2861.99773839450 *T)
    sumX2  +=       0.781 * cos(2.98565268549  + 3796.94625336271 *T)
    sumX2  +=       0.719 * cos(4.60068400707  + 600.78422788230 *T)
    sumX2  +=       0.586 * cos(2.40576117729  + 732.45130231059 *T)
    sumX2  +=       0.546 * cos(2.44839181979  + 437.88770862340 *T)
    sumX2  +=       0.539 * cos(1.82254051811  + 209.12312469139 *T)
    sumX2  +=       0.659 * cos(4.16439637438  + 2207.87335807890 *T)
    sumX2  +=       0.617 * cos(3.66057473715  + 2744.67787017430 *T)
    sumX2  +=       0.617 * cos(0.51898208356  + 2744.19023520730 *T)
    sumX2  +=       0.687 * cos(0.58854485387  + 209.87388421679 *T)
    sumX2  +=       0.566 * cos(3.13556615267  + 526.26575387340 *T)
    sumX2  +=       0.640 * cos(5.17909893148  + 1987.46071564010 *T)
    sumX2  +=       0.623 * cos(3.78657888107  + 3274.36883526891 *T)
    sumX2  +=       0.623 * cos(0.64498622748  + 3273.88120030190 *T)
    sumX2  +=       0.682 * cos(0.06918162441  + 2332.30677329990 *T)
    sumX2  +=       0.719 * cos(3.54267214135  + 9153.65979853830 *T)
    sumX2  +=       0.547 * cos(5.05475414650  + 2097.17940189250 *T)
    sumX2  +=       0.705 * cos(4.16571456887  + 3700.96702614491 *T)
    sumX2  +=       0.574 * cos(4.99043344927  + 430.77416162260 *T)
    sumX2  +=       0.543 * cos(5.59320658894  + 1457.28211557850 *T)
    sumX2  +=       0.693 * cos(5.04264798056  + 2001.20017467470 *T)
    sumX2  +=       0.659 * cos(4.53919606995  + 3804.05980036350 *T)
    sumX2  +=       0.581 * cos(0.58799150667  + 2111.40649589409 *T)
    sumX2  +=       0.540 * cos(4.56823438295  + 2729.96314120569 *T)
    sumX2  +=       0.653 * cos(4.52246789790  + 2626.87036698710 *T)
    sumX2  +=       0.673 * cos(4.65921419822  + 416.54706762100 *T)
    sumX2  +=       0.541 * cos(5.68847817907  + 2214.49927011270 *T)
    sumX2  +=       0.463 * cos(5.28637314313  + 529.88659846831 *T)
    sumX2  +=       0.463 * cos(2.20099897893  + 529.98296668790 *T)
    sumX2  +=       0.453 * cos(0.47524333960  + 5753.62870238030 *T)
    sumX2  +=       0.499 * cos(3.31827894302  + 2421.17245351690 *T)
    sumX2  +=       0.456 * cos(4.85909697014  + 76.50988875911 *T)
    sumX2  +=       0.587 * cos(2.70740471670  + 527.48710202330 *T)
    sumX2  +=       0.505 * cos(5.77419802762  + 532.62854134880 *T)
    sumX2  +=       0.582 * cos(4.77848510541  + 532.38246313290 *T)
    sumX2  +=       0.517 * cos(5.40871139161  + 2523.77759276850 *T)
    sumX2  +=       0.475 * cos(4.45896757052  + 2531.37877473630 *T)
    sumX2  +=       0.433 * cos(1.89229373784  + 621.98185653281 *T)
    sumX2  +=       0.530 * cos(4.89239591125  + 512.52629483880 *T)
    sumX2  +=       0.530 * cos(2.58988239071  + 547.34327031741 *T)
    sumX2  +=       0.478 * cos(5.07132599131  + 1593.24886603040 *T)
    sumX2  +=       0.423 * cos(1.89822370785  + 9683.83839859990 *T)
    sumX2  +=       0.592 * cos(0.09761552238  + 2324.70559133209 *T)
    sumX2  +=       0.410 * cos(2.56036863491  + 2751.30378220809 *T)
    sumX2  +=       0.507 * cos(0.52941934961  + 1243.98302014049 *T)
    sumX2  +=       0.521 * cos(3.03081020519  + 4216.43089723790 *T)
    sumX2  +=       0.458 * cos(5.60102642086  + 629.09540353360 *T)
    sumX2  +=       0.508 * cos(4.89635261381  + 1994.57426264090 *T)
    sumX2  +=       0.508 * cos(4.89635261381  + 1994.08662767389 *T)
    sumX2  +=       0.452 * cos(2.67905981945  + 3.68833577960 *T)
    sumX2  +=       0.527 * cos(4.41239342340  + 597.60283414461 *T)
    sumX2  +=       0.527 * cos(6.21147753215  + 462.26673101160 *T)
    sumX2  +=       0.555 * cos(3.37057288042  + 3892.92548058051 *T)
    sumX2  +=       0.464 * cos(1.74325066579  + 1905.22094745690 *T)
    sumX2  +=       0.535 * cos(4.17138356606  + 2943.74987161071 *T)
    sumX2  +=       0.535 * cos(4.17138356606  + 2943.26223664369 *T)
    sumX2  +=       0.450 * cos(5.30706754834  + 3914.26612158291 *T)
    sumX2  +=       0.390 * cos(5.69507695721  + 2751.79141717511 *T)
    sumX2  +=       0.511 * cos(6.03743992199  + 643.32249753521 *T)
    sumX2  +=       0.431 * cos(2.42207911772  + 849.50804597240 *T)
    sumX2  +=       0.464 * cos(4.69682366854  + 2633.98391398789 *T)
    sumX2  +=       0.486 * cos(4.84373317918  + 2303.85258529670 *T)
    sumX2  +=       0.412 * cos(1.48548182287  + 2758.90496417590 *T)
    sumX2  +=       0.423 * cos(5.88986413302  + 1898.10740045609 *T)
    sumX2  +=       0.457 * cos(3.69523724581  + 519.15220687259 *T)
    sumX2  +=       0.530 * cos(1.25410178254  + 2221.61281711350 *T)
    sumX2  +=       0.443 * cos(0.57624758775  + 89.10949770051 *T)
    sumX2  +=       0.478 * cos(5.10332943513  + 2420.68481854989 *T)

    # JUPITER X Coord
    sumX3  = 0
    sumX3  +=   16990.902 * cos(5.84326964724  + 522.82123557730 *T)
    sumX3  +=   16795.793 * cos(4.75643108252  + 537.04832957890 *T)
    sumX3  +=    5535.565 * cos(0.69030286987  + 0.24381748350 *T)
    sumX3  +=    1711.756 * cos(3.37150503700  + 1059.62574767270 *T)
    sumX3  +=    1437.893 * cos(5.97752044582  + 515.70768857651 *T)
    sumX3  +=    1080.832 * cos(5.40982686420  + 1066.73929467350 *T)
    sumX3  +=    1014.927 * cos(4.64326947068  + 544.16187657970 *T)
    sumX3  +=    1080.320 * cos(6.20988832893  + 7.35736448430 *T)
    sumX3  +=     700.816 * cos(5.48708920252  + 1052.51220067191 *T)
    sumX3  +=     245.290 * cos(4.20564828134  + 1589.31671276730 *T)
    sumX3  +=     301.422 * cos(6.18874877941  + 14.47091148511 *T)
    sumX3  +=     233.420 * cos(1.79483227465  + 529.93478257810 *T)
    sumX3  +=     141.715 * cos(3.21962130849  + 529.44714761109 *T)
    sumX3  +=     134.439 * cos(4.93447341944  + 1045.39865367110 *T)
    sumX3  +=      78.118 * cos(6.00083895472  + 1596.43025976811 *T)
    sumX3  +=      76.833 * cos(0.88537843817  + 117.56368570370 *T)
    sumX3  +=      92.926 * cos(5.97429663193  + 323.74923414091 *T)
    sumX3  +=      78.223 * cos(0.68857971258  + 110.45013870291 *T)
    sumX3  +=      65.879 * cos(6.20010983926  + 109.96250373590 *T)
    sumX3  +=      58.147 * cos(6.14329253529  + 508.59414157570 *T)
    sumX3  +=      51.718 * cos(1.01900575462  + 95.73540973430 *T)
    sumX3  +=      58.000 * cos(0.45369239592  + 426.84200835950 *T)
    sumX3  +=      47.055 * cos(5.14517645295  + 1073.85284167431 *T)
    sumX3  +=      41.297 * cos(3.61714630670  + 633.02755679670 *T)
    sumX3  +=      56.002 * cos(4.58170066805  + 736.12033101530 *T)
    sumX3  +=      40.590 * cos(3.44189359091  + 942.30587945250 *T)
    sumX3  +=      36.838 * cos(4.76515425506  + 625.91400979590 *T)
    sumX3  +=      32.307 * cos(4.46032442444  + 551.27542358051 *T)
    sumX3  +=      38.649 * cos(5.90144372395  + 330.86278114170 *T)
    sumX3  +=      31.637 * cos(5.57280111823  + 1582.20316576650 *T)
    sumX3  +=      32.315 * cos(5.45702982437  + 433.95555536030 *T)
    sumX3  +=      37.135 * cos(2.05207397741  + 103.33659170210 *T)
    sumX3  +=      27.869 * cos(5.02736594295  + 220.65645992230 *T)
    sumX3  +=      27.130 * cos(4.72533930097  + 2119.00767786191 *T)
    sumX3  +=      27.831 * cos(3.47658936637  + 949.41942645330 *T)
    sumX3  +=      26.529 * cos(4.28803355000  + 1169.83206889211 *T)
    sumX3  +=      26.303 * cos(5.95970150804  + 419.72846135871 *T)
    sumX3  +=      27.724 * cos(5.01596447062  + 6.86972951729 *T)
    sumX3  +=      28.942 * cos(4.70013339156  + 729.00678401450 *T)
    sumX3  +=      24.801 * cos(1.98471609425  + 198.82818395289 *T)
    sumX3  +=      22.666 * cos(5.28389674207  + 1155.60497489050 *T)
    sumX3  +=      22.463 * cos(2.05657515244  + 956.53297345411 *T)
    sumX3  +=      24.855 * cos(0.38561886393  + 519.63984183961 *T)
    sumX3  +=      17.587 * cos(6.09056550431  + 21.58445848590 *T)
    sumX3  +=      20.511 * cos(3.90902530966  + 540.22972331660 *T)
    sumX3  +=      17.573 * cos(2.28775742180  + 205.94173095370 *T)
    sumX3  +=      19.314 * cos(4.20495563145  + 640.14110379750 *T)
    sumX3  +=      13.850 * cos(1.05364244327  + 853.44019923550 *T)
    sumX3  +=      13.374 * cos(4.92762205841  + 227.77000692311 *T)
    sumX3  +=      12.226 * cos(3.01621301164  + 522.33360061030 *T)
    sumX3  +=      11.511 * cos(4.04684166302  + 441.06910236111 *T)
    sumX3  +=      11.785 * cos(6.11408212844  + 412.61491435790 *T)
    sumX3  +=       9.912 * cos(2.91727446112  + 206.42936592071 *T)
    sumX3  +=       9.439 * cos(0.08289221852  + 618.80046279510 *T)
    sumX3  +=       9.398 * cos(4.84246640550  + 1038.28510667030 *T)
    sumX3  +=       9.377 * cos(5.60289942767  + 839.21310523390 *T)
    sumX3  +=       9.606 * cos(6.01169153580  + 426.35437339250 *T)
    sumX3  +=      11.272 * cos(2.94227399823  + 309.03450517230 *T)
    sumX3  +=       8.495 * cos(4.87340017026  + 536.56069461190 *T)
    sumX3  +=       7.775 * cos(0.77828391265  + 316.63568714010 *T)
    sumX3  +=       7.402 * cos(0.87870227189  + 412.12727939089 *T)
    sumX3  +=       9.022 * cos(5.05061886203  + 1575.08961876571 *T)
    sumX3  +=       7.220 * cos(6.11239376290  + 1258.69774910910 *T)
    sumX3  +=       6.726 * cos(5.70739192447  + 832.09955823310 *T)
    sumX3  +=       6.944 * cos(0.79854239865  + 309.52214013930 *T)
    sumX3  +=       7.778 * cos(2.92871658834  + 301.92095817150 *T)
    sumX3  +=       7.986 * cos(0.83091189689  + 213.05527795450 *T)
    sumX3  +=       5.933 * cos(0.24709144312  + 2126.12122486270 *T)
    sumX3  +=       5.844 * cos(3.31006297989  + 743.23387801611 *T)
    sumX3  +=       5.965 * cos(2.10263322732  + 515.22005360950 *T)
    sumX3  +=       4.458 * cos(3.94363208534  + 1059.13811270570 *T)
    sumX3  +=       3.905 * cos(5.84264336718  + 1265.81129610991 *T)
    sumX3  +=       4.118 * cos(0.75516336086  + 124.67723270450 *T)
    sumX3  +=       3.669 * cos(3.41470956201  + 1486.22393854870 *T)
    sumX3  +=       3.534 * cos(2.81641405410  + 625.42637482889 *T)
    sumX3  +=       3.134 * cos(4.93499952524  + 1699.52303398670 *T)
    sumX3  +=       2.973 * cos(5.75709909562  + 742.74624304909 *T)
    sumX3  +=       2.686 * cos(4.21604360019  + 88.62186273350 *T)
    sumX3  +=       3.079 * cos(3.56661827942  + 935.19233245171 *T)
    sumX3  +=       2.579 * cos(5.52509554050  + 1062.80714141041 *T)
    sumX3  +=       2.378 * cos(5.22398482496  + 2648.69864295650 *T)
    sumX3  +=       2.346 * cos(0.57872206639  + 728.51914904750 *T)
    sumX3  +=       3.011 * cos(4.13745406105  + 647.25465079831 *T)
    sumX3  +=       2.322 * cos(5.62530333683  + 1603.54380676890 *T)
    sumX3  +=       2.809 * cos(5.09173475362  + 1272.92484311070 *T)
    sumX3  +=       2.701 * cos(4.29236849013  + 1471.99684454710 *T)
    sumX3  +=       2.369 * cos(4.87717253226  + 721.89323701370 *T)
    sumX3  +=       2.359 * cos(5.74427423810  + 337.97632814250 *T)
    sumX3  +=       2.423 * cos(2.71447765750  + 618.31282782810 *T)
    sumX3  +=       2.070 * cos(1.48052642228  + 831.61192326610 *T)
    sumX3  +=       1.958 * cos(5.45174663546  + 526.75338884040 *T)
    sumX3  +=       1.935 * cos(2.01598572418  + 533.11617631580 *T)
    sumX3  +=       2.297 * cos(0.82936512542  + 1368.90407032850 *T)
    sumX3  +=       2.015 * cos(2.11557967911  + 191.71463695210 *T)
    sumX3  +=       2.184 * cos(5.63673637257  + 963.64652045490 *T)
    sumX3  +=       1.685 * cos(0.02501611955  + 1685.29593998510 *T)
    sumX3  +=       1.816 * cos(3.53877940742  + 1162.71852189130 *T)
    sumX3  +=       1.733 * cos(4.27157077977  + 750.34742501690 *T)
    sumX3  +=       1.699 * cos(5.67639711035  + 1055.69359440960 *T)
    sumX3  +=       1.717 * cos(3.76114243385  + 448.18264936190 *T)
    sumX3  +=       1.811 * cos(0.22568163048  + 860.55374623631 *T)
    sumX3  +=       1.432 * cos(4.43210549604  + 81.50831573269 *T)
    sumX3  +=       1.779 * cos(4.17528036440  + 1056.44435393500 *T)
    sumX3  +=       1.397 * cos(0.57951822575  + 611.68691579430 *T)
    sumX3  +=       1.563 * cos(1.99224070438  + 838.72547026689 *T)
    sumX3  +=       1.726 * cos(0.87041746914  + 1361.79052332771 *T)
    sumX3  +=       1.152 * cos(1.70030030447  + 1383.13116433011 *T)
    sumX3  +=       1.200 * cos(2.12159906502  + 632.53992182970 *T)
    sumX3  +=       1.145 * cos(0.71912374834  + 405.50136735710 *T)
    sumX3  +=       1.078 * cos(3.60318956091  + 1479.11039154791 *T)
    sumX3  +=       1.223 * cos(4.59470445130  + 21.09682351890 *T)
    sumX3  +=       1.051 * cos(2.88376336111  + 526.00262931501 *T)
    sumX3  +=       1.414 * cos(2.28596268994  + 941.81824448550 *T)
    sumX3  +=       1.030 * cos(4.60899071471  + 533.86693584120 *T)
    sumX3  +=       1.211 * cos(4.41389386454  + 1069.92068841121 *T)
    sumX3  +=       1.240 * cos(2.35782844542  + 934.70469748469 *T)
    sumX3  +=       0.944 * cos(0.29242241358  + 1044.91101870410 *T)
    sumX3  +=       1.091 * cos(5.67396621358  + 2111.89413086110 *T)
    sumX3  +=       0.986 * cos(3.57242992336  + 2097.17940189250 *T)
    sumX3  +=       0.967 * cos(3.71520398332  + 1588.82907780029 *T)
    sumX3  +=       0.939 * cos(1.24033341734  + 13.98327651810 *T)
    sumX3  +=       0.870 * cos(3.50165707483  + 639.65346883050 *T)
    sumX3  +=       1.013 * cos(0.73509414390  + 1049.33080693420 *T)
    sumX3  +=       0.989 * cos(1.29758643303  + 1148.00379292269 *T)
    sumX3  +=       1.043 * cos(4.55359910492  + 1080.96638867510 *T)
    sumX3  +=       0.824 * cos(2.60527227567  + 433.46792039329 *T)
    sumX3  +=       1.002 * cos(5.58426670661  + 1788.38871420371 *T)
    sumX3  +=       0.823 * cos(3.61010169574  + 654.36819779910 *T)
    sumX3  +=       0.801 * cos(4.98281196861  + 1375.52998236229 *T)
    sumX3  +=       0.890 * cos(3.08710678609  + 294.80741117069 *T)
    sumX3  +=       0.763 * cos(4.57978148957  + 2421.17245351690 *T)
    sumX3  +=       0.793 * cos(4.53708667061  + 234.88355392390 *T)
    sumX3  +=       0.752 * cos(6.13454078435  + 1251.58420210830 *T)
    sumX3  +=       0.788 * cos(0.83160211061  + 302.40859313851 *T)
    sumX3  +=       0.719 * cos(1.71413188590  + 1787.90107923670 *T)
    sumX3  +=       0.760 * cos(5.06618363392  + 2104.78058386030 *T)
    sumX3  +=       0.839 * cos(0.13142328969  + 1898.59503542310 *T)
    sumX3  +=       0.962 * cos(0.09321612606  + 2318.07967929830 *T)
    sumX3  +=       0.832 * cos(0.14656820498  + 2001.68780964171 *T)
    sumX3  +=       0.930 * cos(2.01393320120  + 96.22304470130 *T)
    sumX3  +=       0.879 * cos(0.21241160083  + 845.83901726770 *T)
    sumX3  +=       0.780 * cos(5.93658042409  + 611.19928082730 *T)
    sumX3  +=       0.900 * cos(3.12757954089  + 1162.23088692429 *T)
    sumX3  +=       0.759 * cos(1.64043504780  + 2008.31372167550 *T)
    sumX3  +=       0.715 * cos(0.86869868609  + 1464.39566257930 *T)
    sumX3  +=       0.780 * cos(6.10062998423  + 2214.98690507971 *T)
    sumX3  +=       0.679 * cos(3.26847622660  + 2435.39954751850 *T)
    sumX3  +=       0.809 * cos(0.41722409643  + 1994.57426264090 *T)
    sumX3  +=       0.772 * cos(3.89077003744  + 220.16882495529 *T)
    sumX3  +=       0.622 * cos(0.43134297347  + 316.14805217309 *T)
    sumX3  +=       0.612 * cos(5.54465900470  + 749.85979004990 *T)
    sumX3  +=       0.612 * cos(5.84476543520  + 117.07605073670 *T)
    sumX3  +=       0.713 * cos(4.45141904690  + 1140.89024592190 *T)
    sumX3  +=       0.626 * cos(2.69322845144  + 1692.40948698591 *T)
    sumX3  +=       0.676 * cos(5.06415053228  + 2840.65709739210 *T)
    sumX3  +=       0.594 * cos(6.08554261343  + 102.84895673509 *T)
    sumX3  +=       0.790 * cos(1.96038787639  + 1251.09656714130 *T)
    sumX3  +=       0.756 * cos(6.21273062765  + 1148.49142788971 *T)
    sumX3  +=       0.568 * cos(2.41561339810  + 1141.37788088890 *T)
    sumX3  +=       0.616 * cos(5.32006450469  + 2524.26522773550 *T)
    sumX3  +=       0.560 * cos(1.32052284871  + 2104.29294889330 *T)
    sumX3  +=       0.686 * cos(6.16886011032  + 2627.35800195410 *T)
    sumX3  +=       0.720 * cos(2.07248159341  + 1258.21011414209 *T)
    sumX3  +=       0.733 * cos(1.40692108080  + 1884.36794142151 *T)
    sumX3  +=       0.583 * cos(3.07032575150  + 2751.79141717511 *T)
    sumX3  +=       0.546 * cos(4.30523957810  + 543.67424161270 *T)
    sumX3  +=       0.530 * cos(5.36356787938  + 1066.25165970650 *T)
    sumX3  +=       0.581 * cos(0.74040675650  + 1780.78753223589 *T)
    sumX3  +=       0.523 * cos(2.35523231154  + 1471.50920958009 *T)
    sumX3  +=       0.508 * cos(1.53637927432  + 508.10650660869 *T)
    sumX3  +=       0.645 * cos(5.04939320209  + 2015.91490364330 *T)
    sumX3  +=       0.553 * cos(3.77158096643  + 852.95256426850 *T)
    sumX3  +=       0.480 * cos(1.48751125961  + 1677.69475801730 *T)
    sumX3  +=       0.480 * cos(1.48751125961  + 1678.18239298430 *T)
    sumX3  +=       0.618 * cos(4.58568946876  + 1905.70858242391 *T)
    sumX3  +=       0.463 * cos(2.69346041416  + 89.10949770051 *T)
    sumX3  +=       0.512 * cos(1.92658763463  + 1155.11733992350 *T)
    sumX3  +=       0.497 * cos(5.24109179268  + 1376.01761732930 *T)
    sumX3  +=       0.529 * cos(3.98646203660  + 1802.61580820530 *T)
    sumX3  +=       0.617 * cos(6.12621305930  + 2634.47154895491 *T)
    sumX3  +=       0.459 * cos(4.38908201999  + 1691.92185201890 *T)
    sumX3  +=       0.503 * cos(5.14894098049  + 948.93179148630 *T)
    sumX3  +=       0.547 * cos(0.16414935856  + 2310.96613229750 *T)
    sumX3  +=       0.432 * cos(2.24745181738  + 1567.48843679789 *T)
    sumX3  +=       0.432 * cos(5.38904447097  + 1567.97607176490 *T)
    sumX3  +=       0.419 * cos(0.18825223729  + 1361.30288836070 *T)
    sumX3  +=       0.415 * cos(1.22669638769  + 558.38897058130 *T)
    sumX3  +=       0.415 * cos(6.25558191427  + 501.48059457490 *T)
    sumX3  +=       0.413 * cos(5.02515661889  + 1883.88030645450 *T)
    sumX3  +=       0.436 * cos(4.68517073519  + 1464.88329754630 *T)

    # JUPITER X Coord
    sumX4  = 0
    sumX4  +=    1739.145 * cos(2.99660018877  + 537.04832957890 *T)
    sumX4  +=    1740.807 * cos(1.29193923641  + 522.82123557730 *T)
    sumX4  +=     593.812 * cos(5.24154438099  + 529.93478257810 *T)
    sumX4  +=     303.714 * cos(1.42080900852  + 515.70768857651 *T)
    sumX4  +=     216.971 * cos(2.90416346781  + 544.16187657970 *T)
    sumX4  +=     201.301 * cos(2.23147367590  + 0.24381748350 *T)
    sumX4  +=     128.314 * cos(3.79998656469  + 1066.73929467350 *T)
    sumX4  +=      84.519 * cos(4.38013747176  + 7.35736448430 *T)
    sumX4  +=      66.713 * cos(1.02569869600  + 1052.51220067191 *T)
    sumX4  +=      50.914 * cos(4.46050047710  + 14.47091148511 *T)
    sumX4  +=      56.189 * cos(1.24300511536  + 1059.62574767270 *T)
    sumX4  +=      23.551 * cos(0.46257856947  + 1045.39865367110 *T)
    sumX4  +=      14.975 * cos(1.64416256689  + 508.59414157570 *T)
    sumX4  +=      12.320 * cos(2.75903076030  + 1589.31671276730 *T)
    sumX4  +=      10.412 * cos(3.45236443524  + 1073.85284167431 *T)
    sumX4  +=      10.436 * cos(4.47365597327  + 1596.43025976811 *T)
    sumX4  +=      10.171 * cos(5.43199507160  + 117.56368570370 *T)
    sumX4  +=       7.028 * cos(4.53417325848  + 109.96250373590 *T)
    sumX4  +=       7.310 * cos(2.57204763203  + 551.27542358051 *T)
    sumX4  +=       5.867 * cos(2.76327576305  + 95.73540973430 *T)
    sumX4  +=       5.482 * cos(4.10723362761  + 323.74923414091 *T)
    sumX4  +=       5.308 * cos(4.17493851179  + 330.86278114170 *T)
    sumX4  +=       5.155 * cos(5.15140142409  + 942.30587945250 *T)
    sumX4  +=       4.404 * cos(0.12688032419  + 729.00678401450 *T)
    sumX4  +=       3.429 * cos(0.28270437324  + 625.91400979590 *T)
    sumX4  +=       3.596 * cos(4.90660665484  + 110.45013870291 *T)
    sumX4  +=       3.379 * cos(3.68348076405  + 198.82818395289 *T)
    sumX4  +=       2.870 * cos(1.39677778907  + 529.44714761109 *T)
    sumX4  +=       3.465 * cos(0.05238096193  + 736.12033101530 *T)
    sumX4  +=       2.781 * cos(1.08815967748  + 1582.20316576650 *T)
    sumX4  +=       2.710 * cos(2.38998980124  + 441.06910236111 *T)
    sumX4  +=       2.621 * cos(2.53533210188  + 1169.83206889211 *T)
    sumX4  +=       2.851 * cos(3.77906778388  + 103.33659170210 *T)
    sumX4  +=       3.222 * cos(3.34043716023  + 433.95555536030 *T)
    sumX4  +=       2.252 * cos(0.76740108556  + 1155.60497489050 *T)
    sumX4  +=       2.643 * cos(4.66187220612  + 426.84200835950 *T)
    sumX4  +=       2.565 * cos(6.07676238286  + 6.86972951729 *T)
    sumX4  +=       2.154 * cos(1.93925275699  + 618.80046279510 *T)
    sumX4  +=       2.044 * cos(1.48052835765  + 419.72846135871 *T)
    sumX4  +=       1.513 * cos(3.17740670423  + 220.65645992230 *T)
    sumX4  +=       1.619 * cos(5.33278235904  + 633.02755679670 *T)
    sumX4  +=       1.503 * cos(3.23684234354  + 2119.00767786191 *T)
    sumX4  +=       1.422 * cos(3.21419588854  + 227.77000692311 *T)
    sumX4  +=       1.448 * cos(0.63548633830  + 1575.08961876571 *T)
    sumX4  +=       1.354 * cos(0.37531194856  + 956.53297345411 *T)
    sumX4  +=       1.398 * cos(4.66226122647  + 301.92095817150 *T)
    sumX4  +=       1.551 * cos(2.51719031322  + 640.14110379750 *T)
    sumX4  +=       1.193 * cos(5.69090306000  + 853.44019923550 *T)
    sumX4  +=       1.228 * cos(3.32529407238  + 536.56069461190 *T)
    sumX4  +=       1.153 * cos(4.24056902908  + 426.35437339250 *T)
    sumX4  +=       1.111 * cos(1.82993073798  + 309.52214013930 *T)
    sumX4  +=       1.322 * cos(1.52760953063  + 412.61491435790 *T)
    sumX4  +=       1.137 * cos(1.63346191003  + 1258.69774910910 *T)
    sumX4  +=       0.987 * cos(5.01376257502  + 522.33360061030 *T)
    sumX4  +=       0.937 * cos(1.26232145651  + 618.31282782810 *T)
    sumX4  +=       0.830 * cos(2.55897013457  + 412.12727939089 *T)
    sumX4  +=       0.868 * cos(0.67939405545  + 1038.28510667030 *T)
    sumX4  +=       0.819 * cos(0.85929624974  + 742.74624304909 *T)
    sumX4  +=       0.952 * cos(0.08637090339  + 831.61192326610 *T)
    sumX4  +=       0.903 * cos(5.28542287743  + 949.41942645330 *T)
    sumX4  +=       0.740 * cos(1.11419070098  + 832.09955823310 *T)
    sumX4  +=       0.726 * cos(1.10542550627  + 839.21310523390 *T)
    sumX4  +=       0.890 * cos(2.70182272881  + 647.25465079831 *T)
    sumX4  +=       0.672 * cos(3.82581500735  + 102.84895673509 *T)
    sumX4  +=       0.598 * cos(4.03229295405  + 337.97632814250 *T)
    sumX4  +=       0.711 * cos(3.79919685217  + 515.22005360950 *T)
    sumX4  +=       0.557 * cos(2.29520680686  + 1486.22393854870 *T)
    sumX4  +=       0.506 * cos(2.29510505753  + 1059.13811270570 *T)
    sumX4  +=       0.464 * cos(1.33006467962  + 117.07605073670 *T)
    sumX4  +=       0.479 * cos(6.24011460202  + 526.75338884040 *T)
    sumX4  +=       0.451 * cos(0.61581620782  + 1471.99684454710 *T)
    sumX4  +=       0.466 * cos(1.24608768657  + 533.11617631580 *T)
    sumX4  +=       0.396 * cos(3.90384143043  + 1603.54380676890 *T)

    # JUPITER X Coord
    sumX5  = 0
    sumX5  +=     131.509 * cos(1.09720765135  + 537.04832957890 *T)
    sumX5  +=     126.839 * cos(3.14566028359  + 522.82123557730 *T)
    sumX5  +=      48.033 * cos(3.18247744923  + 515.70768857651 *T)
    sumX5  +=      34.241 * cos(1.11319864554  + 544.16187657970 *T)
    sumX5  +=      11.284 * cos(2.13882736419  + 1066.73929467350 *T)
    sumX5  +=       5.229 * cos(2.44608299466  + 0.24381748350 *T)
    sumX5  +=       3.712 * cos(3.28596952033  + 1052.51220067191 *T)
    sumX5  +=       4.457 * cos(5.28755623884  + 529.93478257810 *T)
    sumX5  +=       1.949 * cos(1.73980382124  + 1073.85284167431 *T)
    sumX5  +=       1.738 * cos(4.99242464275  + 1059.62574767270 *T)
    sumX5  +=       1.716 * cos(3.05846955779  + 1045.39865367110 *T)
    sumX5  +=       0.879 * cos(3.22039268284  + 508.59414157570 *T)
    sumX5  +=       0.886 * cos(4.22728564482  + 551.27542358051 *T)
    sumX5  +=       0.665 * cos(1.57102457622  + 7.35736448430 *T)
    sumX5  +=       0.466 * cos(2.85877844081  + 1596.43025976811 *T)

    # JUPITER Y Coord
    sumY0  = 0
    sumY0  +=  519591755.961 * cos(5.31193570238  + 529.93478257810 *T)
    sumY0  +=  37839498.798 * cos(1.82085166331  + 0.24381748350 *T)
    sumY0  +=  12593400.247 * cos(5.66155354525  + 1059.62574767270 *T)
    sumY0  +=  1504469.362 * cos(5.44173895105  + 522.82123557730 *T)
    sumY0  +=  1476016.965 * cos(2.04668425902  + 537.04832957890 *T)
    sumY0  +=  457751.771 * cos(6.01125846140  + 1589.31671276730 *T)
    sumY0  +=  301186.623 * cos(3.60619939116  + 7.35736448430 *T)
    sumY0  +=  381996.183 * cos(3.55684846636  + 102.84895673509 *T)
    sumY0  +=  193179.179 * cos(3.45595232302  + 426.84200835950 *T)
    sumY0  +=  148291.722 * cos(4.58538691397  + 110.45013870291 *T)
    sumY0  +=  142171.400 * cos(3.99477000498  + 633.02755679670 *T)
    sumY0  +=  133443.745 * cos(5.60526571675  + 213.54291292150 *T)
    sumY0  +=  102719.450 * cos(4.59936785476  + 1052.51220067191 *T)
    sumY0  +=   87327.523 * cos(1.56302538060  + 529.44714761109 *T)
    sumY0  +=  114122.292 * cos(4.72894061449  + 1162.71852189130 *T)
    sumY0  +=   72093.376 * cos(2.39043228390  + 1066.73929467350 *T)
    sumY0  +=   59268.404 * cos(2.89107868692  + 949.41942645330 *T)
    sumY0  +=   68329.094 * cos(1.86754138371  + 846.32665223470 *T)
    sumY0  +=   30209.026 * cos(4.54186591923  + 205.94173095370 *T)
    sumY0  +=   34007.392 * cos(3.31774226555  + 419.24082639170 *T)
    sumY0  +=   26616.582 * cos(3.28146731015  + 323.74923414091 *T)
    sumY0  +=   19727.427 * cos(0.07815264028  + 2119.00767786191 *T)
    sumY0  +=   17749.540 * cos(5.55480049652  + 419.72846135871 *T)
    sumY0  +=   17560.618 * cos(0.38432139040  + 316.63568714010 *T)
    sumY0  +=   16308.761 * cos(4.16521301086  + 736.12033101530 *T)
    sumY0  +=   16227.856 * cos(5.58096312351  + 515.70768857651 *T)
    sumY0  +=   14406.626 * cos(0.49204486417  + 743.23387801611 *T)
    sumY0  +=   11421.526 * cos(1.91077907185  + 544.16187657970 *T)
    sumY0  +=   10835.166 * cos(1.62316112493  + 103.33659170210 *T)
    sumY0  +=   12154.356 * cos(2.18151132345  + 526.00262931501 *T)
    sumY0  +=   11992.571 * cos(5.29747555409  + 533.86693584120 *T)
    sumY0  +=    8455.662 * cos(1.90801507290  + 640.14110379750 *T)
    sumY0  +=    8193.182 * cos(3.99950988528  + 1479.11039154791 *T)
    sumY0  +=    8713.455 * cos(5.09554104540  + 1692.40948698591 *T)
    sumY0  +=    7420.814 * cos(4.41221626331  + 956.53297345411 *T)
    sumY0  +=    7565.993 * cos(5.61829156567  + 1265.81129610991 *T)
    sumY0  +=    8202.922 * cos(2.98736132250  + 735.63269604830 *T)
    sumY0  +=    7717.111 * cos(4.84373719017  + 1582.20316576650 *T)
    sumY0  +=    5320.837 * cos(5.81736694969  + 526.75338884040 *T)
    sumY0  +=    5217.759 * cos(1.66167068417  + 533.11617631580 *T)
    sumY0  +=    5238.212 * cos(4.68306641983  + 522.33360061030 *T)
    sumY0  +=    5774.716 * cos(3.46781064586  + 14.47091148511 *T)
    sumY0  +=    5907.614 * cos(0.27080776672  + 109.96250373590 *T)
    sumY0  +=    4575.512 * cos(2.18559360651  + 1376.01761732930 *T)
    sumY0  +=    3939.870 * cos(2.73821792955  + 1596.43025976811 *T)
    sumY0  +=    4568.269 * cos(3.71733952931  + 95.73540973430 *T)
    sumY0  +=    3036.375 * cos(5.51572587470  + 309.03450517230 *T)
    sumY0  +=    2856.672 * cos(4.44258466917  + 117.56368570370 *T)
    sumY0  +=    2820.138 * cos(0.55276658901  + 632.53992182970 *T)
    sumY0  +=    2442.859 * cos(2.66976611024  + 433.95555536030 *T)
    sumY0  +=    2534.517 * cos(4.61010280175  + 625.91400979590 *T)
    sumY0  +=    2329.307 * cos(1.22761098421  + 213.05527795450 *T)
    sumY0  +=    2430.903 * cos(2.32333761217  + 220.65645992230 *T)
    sumY0  +=    2673.686 * cos(1.61750799699  + 1169.83206889211 *T)
    sumY0  +=    2396.896 * cos(4.36221151655  + 1045.39865367110 *T)
    sumY0  +=    2503.065 * cos(2.72526085550  + 1052.02456570489 *T)
    sumY0  +=    1817.118 * cos(3.95907641758  + 838.72547026689 *T)
    sumY0  +=    2014.206 * cos(4.88058047490  + 1155.60497489050 *T)
    sumY0  +=    1501.452 * cos(3.03509232132  + 942.30587945250 *T)
    sumY0  +=    1454.916 * cos(4.18897174343  + 853.44019923550 *T)
    sumY0  +=    1628.897 * cos(3.92104093589  + 75.02541605080 *T)
    sumY0  +=    1106.760 * cos(5.64732593309  + 6.86972951729 *T)
    sumY0  +=    1080.777 * cos(0.43950996921  + 316.14805217309 *T)
    sumY0  +=    1041.755 * cos(4.81940326599  + 1795.50226120450 *T)
    sumY0  +=     996.914 * cos(1.51229276007  + 1272.92484311070 *T)
    sumY0  +=     928.877 * cos(4.68477803335  + 198.82818395289 *T)
    sumY0  +=     934.490 * cos(0.42872611846  + 2648.69864295650 *T)
    sumY0  +=     858.575 * cos(2.14237368311  + 529.88659846831 *T)
    sumY0  +=     858.481 * cos(5.34018482611  + 529.98296668790 *T)
    sumY0  +=    1088.006 * cos(5.84659040992  + 527.48710202330 *T)
    sumY0  +=    1080.077 * cos(1.63448618678  + 532.38246313290 *T)
    sumY0  +=     958.229 * cos(6.06015323795  + 149.80701461810 *T)
    sumY0  +=     830.762 * cos(5.25949378156  + 625.42637482889 *T)
    sumY0  +=     815.889 * cos(1.66863411356  + 1059.13811270570 *T)
    sumY0  +=     838.684 * cos(5.82302302292  + 3.68833577960 *T)
    sumY0  +=     956.674 * cos(0.63174409236  + 379.88395047650 *T)
    sumY0  +=     694.759 * cos(3.69145164541  + 1155.11733992350 *T)
    sumY0  +=     673.874 * cos(4.52164149268  + 2008.80135664250 *T)
    sumY0  +=     865.777 * cos(0.10096193323  + 426.35437339250 *T)
    sumY0  +=     897.730 * cos(2.48087145339  + 1368.41643536150 *T)
    sumY0  +=     667.422 * cos(0.38400668832  + 1368.90407032850 *T)
    sumY0  +=     747.747 * cos(3.13927711168  + 330.86278114170 *T)
    sumY0  +=     693.620 * cos(6.05103606639  + 1063.55790093581 *T)
    sumY0  +=     787.813 * cos(5.11429018504  + 839.21310523390 *T)
    sumY0  +=     614.978 * cos(4.83788756913  + 2111.89413086110 *T)
    sumY0  +=     680.352 * cos(5.00624167079  + 528.97157473160 *T)
    sumY0  +=     678.592 * cos(2.47610569321  + 530.89799042461 *T)
    sumY0  +=     574.130 * cos(5.45801133464  + 2222.10045208050 *T)
    sumY0  +=     644.748 * cos(3.49842430477  + 412.12727939089 *T)
    sumY0  +=     742.423 * cos(2.38906516060  + 206.42936592071 *T)
    sumY0  +=     563.602 * cos(2.49278594908  + 1055.69359440960 *T)
    sumY0  +=     559.331 * cos(4.30651563744  + 729.00678401450 *T)
    sumY0  +=     647.539 * cos(3.55462405942  + 984.84414910540 *T)
    sumY0  +=     460.071 * cos(2.25575075230  + 1073.85284167431 *T)
    sumY0  +=     403.270 * cos(5.70558430216  + 412.61491435790 *T)
    sumY0  +=     539.827 * cos(3.74550987414  + 38.37685312130 *T)
    sumY0  +=     407.824 * cos(1.41749002358  + 910.06255053810 *T)
    sumY0  +=     319.575 * cos(4.92965704995  + 941.81824448550 *T)
    sumY0  +=     301.187 * cos(2.49363421546  + 1905.70858242391 *T)
    sumY0  +=     330.186 * cos(5.17961869781  + 948.93179148630 *T)
    sumY0  +=     334.161 * cos(5.66579079606  + 76.50988875911 *T)
    sumY0  +=     292.241 * cos(3.42430216936  + 1471.50920958009 *T)
    sumY0  +=     388.960 * cos(4.48974750232  + 604.71638114540 *T)
    sumY0  +=     387.023 * cos(2.95645101689  + 455.15318401080 *T)
    sumY0  +=     303.530 * cos(3.49353690436  + 529.41351771630 *T)
    sumY0  +=     303.139 * cos(3.98877079900  + 530.45604743991 *T)
    sumY0  +=     276.704 * cos(5.36713470530  + 1486.22393854870 *T)
    sumY0  +=     383.222 * cos(3.32437927262  + 3.42521122121 *T)
    sumY0  +=     267.964 * cos(0.44358626370  + 309.52214013930 *T)
    sumY0  +=     270.827 * cos(5.75355417511  + 1258.69774910910 *T)
    sumY0  +=     264.824 * cos(4.51845894459  + 536.56069461190 *T)
    sumY0  +=     342.262 * cos(2.22920476980  + 1684.80830501809 *T)
    sumY0  +=     254.392 * cos(2.01389858651  + 1062.80714141041 *T)
    sumY0  +=     289.891 * cos(5.95965968983  + 742.74624304909 *T)
    sumY0  +=     227.013 * cos(3.08717315673  + 2126.12122486270 *T)
    sumY0  +=     219.942 * cos(4.62508932315  + 515.22005360950 *T)
    sumY0  +=     304.124 * cos(0.18282516849  + 6283.31966747490 *T)
    sumY0  +=     289.590 * cos(4.54086778359  + 305.10235190919 *T)
    sumY0  +=     225.217 * cos(1.93067375292  + 1699.52303398670 *T)
    sumY0  +=     192.734 * cos(5.73241739600  + 508.59414157570 *T)
    sumY0  +=     181.955 * cos(5.49121576696  + 1898.59503542310 *T)
    sumY0  +=     180.824 * cos(4.33642523532  + 1265.32366114289 *T)
    sumY0  +=     234.431 * cos(5.63602319498  + 1056.44435393500 *T)
    sumY0  +=     210.412 * cos(2.44959263747  + 490.57790666290 *T)
    sumY0  +=     178.960 * cos(4.53074727531  + 527.01402127131 *T)
    sumY0  +=     184.077 * cos(4.02561581206  + 1471.99684454710 *T)
    sumY0  +=     177.005 * cos(2.95243370745  + 532.85554388490 *T)
    sumY0  +=     197.754 * cos(4.65708333161  + 1575.08961876571 *T)
    sumY0  +=     201.548 * cos(3.17115080743  + 728.51914904750 *T)
    sumY0  +=     164.871 * cos(1.90607027544  + 529.18402305270 *T)
    sumY0  +=     164.585 * cos(5.57619441078  + 530.68554210351 *T)
    sumY0  +=     191.757 * cos(5.01987763082  + 569.29165849331 *T)
    sumY0  +=     155.012 * cos(5.97029570081  + 519.63984183961 *T)
    sumY0  +=     216.423 * cos(0.97167033032  + 453.18107633550 *T)
    sumY0  +=     208.270 * cos(1.82244512865  + 1439.75351563271 *T)
    sumY0  +=     195.461 * cos(5.41018940348  + 1685.29593998510 *T)
    sumY0  +=     158.021 * cos(2.75745977713  + 525.74199688410 *T)
    sumY0  +=     155.478 * cos(4.72687069685  + 534.12756827210 *T)
    sumY0  +=     142.916 * cos(0.78899783107  + 11.28951774740 *T)
    sumY0  +=     166.137 * cos(2.08946903667  + 224.58861318540 *T)
    sumY0  +=     139.568 * cos(1.78496043471  + 647.25465079831 *T)
    sumY0  +=     194.718 * cos(3.73420656967  + 1021.49271203491 *T)
    sumY0  +=     142.687 * cos(4.63478591519  + 1258.21011414209 *T)
    sumY0  +=     136.501 * cos(4.43284450396  + 99.40443843901 *T)
    sumY0  +=     155.156 * cos(0.59258028164  + 845.83901726770 *T)
    sumY0  +=     148.749 * cos(5.65218277636  + 301.92095817150 *T)
    sumY0  +=     142.483 * cos(2.18605854450  + 227.77000692311 *T)
    sumY0  +=     128.042 * cos(0.68370572776  + 1162.23088692429 *T)
    sumY0  +=     129.196 * cos(3.16003204274  + 1787.90107923670 *T)
    sumY0  +=     176.884 * cos(1.60517740492  + 10213.52936369450 *T)
    sumY0  +=     120.579 * cos(1.51220345630  + 540.22972331660 *T)
    sumY0  +=     163.945 * cos(0.44168564921  + 1.72829019180 *T)
    sumY0  +=     135.003 * cos(2.22806039362  + 540.98048284201 *T)
    sumY0  +=     121.846 * cos(2.21825172668  + 524.30570828560 *T)
    sumY0  +=     120.032 * cos(5.26558395487  + 535.56385687061 *T)
    sumY0  +=     124.238 * cos(1.81767045547  + 983.35967639710 *T)
    sumY0  +=     123.682 * cos(1.77349067432  + 525.26891613211 *T)
    sumY0  +=     104.430 * cos(5.28755763826  + 530.09484127250 *T)
    sumY0  +=     104.430 * cos(2.19472066370  + 529.77472388370 *T)
    sumY0  +=     136.458 * cos(1.98358578274  + 2001.20017467470 *T)
    sumY0  +=     115.988 * cos(1.78467759503  + 750.34742501690 *T)
    sumY0  +=     104.731 * cos(1.75643097216  + 551.27542358051 *T)
    sumY0  +=     100.945 * cos(3.93006785022  + 1581.71553079950 *T)
    sumY0  +=     116.158 * cos(5.71003623422  + 534.60064902410 *T)
    sumY0  +=     122.805 * cos(3.33550817409  + 524.51815660670 *T)
    sumY0  +=     103.104 * cos(2.84340091211  + 963.64652045490 *T)
    sumY0  +=     115.142 * cos(5.03053273280  + 2325.19322629910 *T)
    sumY0  +=     120.928 * cos(4.14505296195  + 535.35140854951 *T)
    sumY0  +=      94.293 * cos(1.67313940200  + 1802.61580820530 *T)
    sumY0  +=      92.521 * cos(4.73521093464  + 1383.13116433011 *T)
    sumY0  +=      85.964 * cos(4.10984886270  + 831.61192326610 *T)
    sumY0  +=      91.882 * cos(1.46390569496  + 441.06910236111 *T)
    sumY0  +=      90.884 * cos(1.07143309047  + 4.17597074660 *T)
    sumY0  +=      96.153 * cos(3.31598430560  + 21.58445848590 *T)
    sumY0  +=      73.111 * cos(4.24231833095  + 415.79630809560 *T)
    sumY0  +=      72.096 * cos(5.03046144250  + 529.72233425701 *T)
    sumY0  +=      72.096 * cos(2.45181685946  + 530.14723089920 *T)
    sumY0  +=      95.408 * cos(4.36670590047  + 35.66854013560 *T)
    sumY0  +=      79.513 * cos(5.39469246290  + 835.28095197081 *T)
    sumY0  +=      68.517 * cos(4.34434516091  + 1574.60198379870 *T)
    sumY0  +=      69.848 * cos(4.98758055977  + 518.88908231420 *T)
    sumY0  +=      73.288 * cos(5.26346329278  + 832.09955823310 *T)
    sumY0  +=      73.790 * cos(2.94904494716  + 1044.91101870410 *T)
    sumY0  +=      62.635 * cos(5.05973221161  + 2214.98690507971 *T)
    sumY0  +=      72.372 * cos(4.07908178910  + 511.77553531340 *T)
    sumY0  +=      79.774 * cos(5.75946246773  + 618.80046279510 *T)
    sumY0  +=      84.465 * cos(2.02066705387  + 834.79331700380 *T)
    sumY0  +=      82.357 * cos(1.01171039674  + 531.41925528640 *T)
    sumY0  +=      68.826 * cos(3.41224512535  + 548.09402984280 *T)
    sumY0  +=      58.994 * cos(2.89954202085  + 2104.29294889330 *T)
    sumY0  +=      69.760 * cos(0.41067157485  + 98.91680347199 *T)
    sumY0  +=      74.013 * cos(2.84831536831  + 174.18603700630 *T)
    sumY0  +=      55.680 * cos(3.22751869738  + 644.07325706060 *T)
    sumY0  +=      77.878 * cos(0.16446911156  + 528.45030986980 *T)
    sumY0  +=      54.479 * cos(3.59326916447  + 1898.10740045609 *T)
    sumY0  +=      55.045 * cos(3.71209514844  + 639.65346883050 *T)
    sumY0  +=      53.073 * cos(0.40077660714  + 1048.58004740881 *T)
    sumY0  +=      52.282 * cos(3.04928542911  + 960.46512671720 *T)
    sumY0  +=      51.560 * cos(0.07590120615  + 514.95692905111 *T)
    sumY0  +=      52.703 * cos(5.01527928588  + 2641.58509595570 *T)
    sumY0  +=      66.188 * cos(2.16852924204  + 230.32075334190 *T)
    sumY0  +=      49.711 * cos(1.12330761789  + 544.91263610510 *T)
    sumY0  +=      49.029 * cos(1.60521275283  + 387.48513244430 *T)
    sumY0  +=      65.960 * cos(5.74241303706  + 1364.97191706540 *T)
    sumY0  +=      52.715 * cos(6.19618088387  + 1057.17806711790 *T)
    sumY0  +=      58.796 * cos(5.92972662777  + 39.11305843170 *T)
    sumY0  +=      59.809 * cos(2.23693717498  + 91.80325647120 *T)
    sumY0  +=      46.231 * cos(4.96698555728  + 2538.49232173710 *T)
    sumY0  +=      48.382 * cos(4.05069063760  + 2428.28600051770 *T)
    sumY0  +=      60.070 * cos(2.20914086650  + 1062.07342822750 *T)
    sumY0  +=      56.428 * cos(1.74520780023  + 2317.59204433130 *T)
    sumY0  +=      59.945 * cos(5.60557419941  + 494.51005992600 *T)
    sumY0  +=      43.719 * cos(1.93646240538  + 2.20386307129 *T)
    sumY0  +=      48.181 * cos(5.35729954844  + 618.31282782810 *T)
    sumY0  +=      49.751 * cos(0.31062429353  + 679.49797971270 *T)
    sumY0  +=      58.551 * cos(0.74219345496  + 1058.14127496441 *T)
    sumY0  +=      47.067 * cos(0.78230316551  + 3178.38960805111 *T)
    sumY0  +=      52.897 * cos(3.51938305050  + 106.78110999820 *T)
    sumY0  +=      41.687 * cos(5.68958878057  + 1059.67393178251 *T)
    sumY0  +=      41.782 * cos(1.87606562860  + 71.09326278771 *T)
    sumY0  +=      43.674 * cos(5.93139897307  + 430.77416162260 *T)
    sumY0  +=      44.776 * cos(3.31656958399  + 312.70353387701 *T)
    sumY0  +=      48.969 * cos(4.37808022467  + 1038.28510667030 *T)
    sumY0  +=      38.530 * cos(3.84889420737  + 1148.00379292269 *T)
    sumY0  +=      52.006 * cos(1.89655384425  + 565.35950523021 *T)
    sumY0  +=      47.878 * cos(5.49999542741  + 1375.52998236229 *T)
    sumY0  +=      38.108 * cos(5.00245348220  + 1788.38871420371 *T)
    sumY0  +=      49.073 * cos(2.58173317309  + 523.78444342380 *T)
    sumY0  +=      38.297 * cos(4.36934356478  + 106.51798543980 *T)
    sumY0  +=      41.687 * cos(2.49177765294  + 1059.57756356290 *T)
    sumY0  +=      37.366 * cos(5.83924605505  + 2751.79141717511 *T)
    sumY0  +=      41.634 * cos(0.09415541149  + 1478.62275658089 *T)
    sumY0  +=      41.532 * cos(3.87601872923  + 1514.53511420000 *T)
    sumY0  +=      41.308 * cos(3.06772374924  + 10.53875822200 *T)
    sumY0  +=      40.633 * cos(4.90197995789  + 18.40306474821 *T)
    sumY0  +=      42.800 * cos(1.68392525459  + 113.89465699899 *T)
    sumY0  +=      48.985 * cos(4.63153291545  + 3340.85624418330 *T)
    sumY0  +=      44.449 * cos(4.90238800208  + 536.08512173241 *T)
    sumY0  +=      34.928 * cos(5.86589465436  + 672.38443271190 *T)
    sumY0  +=      36.386 * cos(0.45310476388  + 1361.79052332771 *T)
    sumY0  +=      36.844 * cos(3.31552381410  + 530.19541500900 *T)
    sumY0  +=      36.844 * cos(4.16675448786  + 529.67415014720 *T)
    sumY0  +=      34.444 * cos(4.06056429191  + 1890.99385345530 *T)
    sumY0  +=      41.479 * cos(0.20923373847  + 1593.24886603040 *T)
    sumY0  +=      46.613 * cos(2.16989392147  + 1513.05064149171 *T)
    sumY0  +=      32.971 * cos(2.27730952602  + 522.77305146750 *T)
    sumY0  +=      35.418 * cos(5.49008702234  + 522.86941968710 *T)
    sumY0  +=      32.742 * cos(4.25989553271  + 124.67723270450 *T)
    sumY0  +=      34.850 * cos(2.92484630853  + 1060.58895551920 *T)
    sumY0  +=      30.476 * cos(2.81533288955  + 1585.38455950421 *T)
    sumY0  +=      29.011 * cos(5.89187559250  + 121.49583896680 *T)
    sumY0  +=      33.477 * cos(1.40109397770  + 629.09540353360 *T)
    sumY0  +=      28.909 * cos(5.57936571060  + 2001.68780964171 *T)
    sumY0  +=      29.909 * cos(1.98839129392  + 537.00014546910 *T)
    sumY0  +=      27.497 * cos(5.20349493393  + 537.09651368871 *T)
    sumY0  +=      28.353 * cos(3.29566866990  + 2214.49927011270 *T)
    sumY0  +=      27.748 * cos(2.64291022936  + 2420.68481854989 *T)
    sumY0  +=      33.057 * cos(4.51374272746  + 568.06781821590 *T)
    sumY0  +=      27.675 * cos(3.76752390296  + 528.66275819090 *T)
    sumY0  +=      29.521 * cos(2.73723500563  + 1361.30288836070 *T)
    sumY0  +=      27.586 * cos(3.71471356005  + 531.20680696530 *T)
    sumY0  +=      33.930 * cos(4.63376695563  + 885.68352814990 *T)
    sumY0  +=      27.514 * cos(1.45900145166  + 114.38229196601 *T)
    sumY0  +=      24.495 * cos(4.07450770524  + 99.66756299740 *T)
    sumY0  +=      24.344 * cos(3.20297940813  + 935.19233245171 *T)
    sumY0  +=      26.492 * cos(3.14995415897  + 422.90985509640 *T)
    sumY0  +=      30.581 * cos(0.04329676173  + 142.20583265029 *T)
    sumY0  +=      32.993 * cos(5.35577320707  + 1058.66253982621 *T)
    sumY0  +=      26.693 * cos(4.15534706030  + 747.16603127920 *T)
    sumY0  +=      23.218 * cos(4.67611113291  + 216.98743121760 *T)
    sumY0  +=      31.446 * cos(4.79362433774  + 437.88770862340 *T)
    sumY0  +=      23.990 * cos(4.10871208553  + 523.34250043910 *T)
    sumY0  +=      25.905 * cos(2.70346670886  + 10.80188278040 *T)
    sumY0  +=      30.584 * cos(3.14037291034  + 1024.20102502061 *T)
    sumY0  +=      22.355 * cos(6.15548689757  + 106.03035047280 *T)
    sumY0  +=      27.098 * cos(5.64808334984  + 497.69145366371 *T)
    sumY0  +=      29.688 * cos(5.93516522116  + 760.01171843649 *T)
    sumY0  +=      26.696 * cos(1.62294414819  + 1158.78636862821 *T)
    sumY0  +=      25.786 * cos(4.19746250367  + 636.95971005981 *T)
    sumY0  +=      24.825 * cos(4.87212515778  + 1134.40734624000 *T)
    sumY0  +=      22.840 * cos(0.83434687851  + 459.08533727390 *T)
    sumY0  +=      21.870 * cos(3.37410460471  + 536.52706471711 *T)
    sumY0  +=      22.257 * cos(5.18257233161  + 521.85802773080 *T)
    sumY0  +=      26.904 * cos(4.55348699023  + 299.37021175271 *T)
    sumY0  +=      20.168 * cos(4.63789217577  + 2104.78058386030 *T)
    sumY0  +=      19.591 * cos(2.60352149776  + 1603.54380676890 *T)
    sumY0  +=      23.983 * cos(1.51490151538  + 2633.98391398789 *T)
    sumY0  +=      23.778 * cos(1.81405059728  + 562.17811149250 *T)
    sumY0  +=      19.633 * cos(2.55698413130  + 491.80174694030 *T)
    sumY0  +=      26.288 * cos(6.09242729115  + 454.66554904380 *T)
    sumY0  +=      21.128 * cos(5.07121999690  + 934.70469748469 *T)
    sumY0  +=      19.074 * cos(0.54048693886  + 327.68138740401 *T)
    sumY0  +=      18.465 * cos(3.58841586091  + 1464.39566257930 *T)
    sumY0  +=      21.220 * cos(4.75365640451  + 495.99453263430 *T)
    sumY0  +=      24.986 * cos(3.56749136042  + 860.55374623631 *T)
    sumY0  +=      18.596 * cos(0.57894147959  + 5.87289177600 *T)
    sumY0  +=      24.136 * cos(0.56079031346  + 380.37158544351 *T)
    sumY0  +=      18.303 * cos(3.05720710697  + 953.35157971640 *T)
    sumY0  +=      24.099 * cos(2.66580382654  + 621.98185653281 *T)
    sumY0  +=      19.533 * cos(0.33685720004  + 1098.98262358790 *T)
    sumY0  +=      17.623 * cos(3.77983317632  + 2207.38572311189 *T)
    sumY0  +=      18.076 * cos(0.42847308235  + 600.78422788230 *T)
    sumY0  +=      17.281 * cos(4.85465411050  + 191.71463695210 *T)
    sumY0  +=      17.873 * cos(2.29598676123  + 538.01153742540 *T)
    sumY0  +=      17.226 * cos(2.68832441082  + 2435.39954751850 *T)
    sumY0  +=      21.244 * cos(4.71917522298  + 1166.65067515440 *T)
    sumY0  +=      20.337 * cos(5.96704683966  + 1795.01462623750 *T)
    sumY0  +=      16.743 * cos(4.81175858246  + 2531.37877473630 *T)
    sumY0  +=      19.199 * cos(2.72712183072  + 563.87503252191 *T)
    sumY0  +=      16.833 * cos(2.43728533530  + 24.62283987170 *T)
    sumY0  +=      21.175 * cos(0.27960745033  + 88.62186273350 *T)
    sumY0  +=      16.556 * cos(2.28023588713  + 2229.21399908130 *T)
    sumY0  +=      16.153 * cos(4.90017932817  + 1049.33080693420 *T)
    sumY0  +=      19.785 * cos(5.64301732154  + 505.55576018990 *T)
    sumY0  +=      18.422 * cos(2.18247642641  + 1969.44448072730 *T)
    sumY0  +=      19.816 * cos(2.08941014873  + 96.22304470130 *T)
    sumY0  +=      16.272 * cos(2.33706582122  + 527.22646959241 *T)
    sumY0  +=      16.170 * cos(5.14508707196  + 532.64309556381 *T)
    sumY0  +=      15.510 * cos(5.66925542318  + 988.77630236850 *T)
    sumY0  +=      15.594 * cos(3.28617425582  + 39.60069339870 *T)
    sumY0  +=      15.853 * cos(1.45510339077  + 501.62360692681 *T)
    sumY0  +=      16.153 * cos(2.18239496410  + 31.99951143090 *T)
    sumY0  +=      18.509 * cos(1.84602176090  + 554.31380496631 *T)
    sumY0  +=      15.127 * cos(1.98189390748  + 460.78225830330 *T)
    sumY0  +=      16.461 * cos(4.35420228107  + 220.16882495529 *T)
    sumY0  +=      19.878 * cos(5.33036403802  + 1691.92185201890 *T)
    sumY0  +=      14.712 * cos(3.84306627886  + 1059.10448281091 *T)
    sumY0  +=      17.901 * cos(1.98272465699  + 107.26874496520 *T)
    sumY0  +=      14.484 * cos(3.28758641825  + 1070.67144793660 *T)
    sumY0  +=      15.909 * cos(0.96629585192  + 4.90968392950 *T)
    sumY0  +=      14.674 * cos(6.01692843272  + 558.24595822940 *T)
    sumY0  +=      14.766 * cos(3.88224084450  + 2744.67787017430 *T)
    sumY0  +=      18.715 * cos(5.44782508771  + 523.71500745461 *T)
    sumY0  +=      14.896 * cos(1.36370078487  + 1588.82907780029 *T)
    sumY0  +=      17.816 * cos(3.79294770830  + 74.53778108379 *T)
    sumY0  +=      14.616 * cos(3.00562662663  + 2530.89113976930 *T)
    sumY0  +=      18.292 * cos(3.40569854290  + 1151.67282162740 *T)
    sumY0  +=      18.426 * cos(2.03446098690  + 536.15455770160 *T)
    sumY0  +=      13.673 * cos(0.68703780947  + 732.18817775220 *T)
    sumY0  +=      13.765 * cos(6.04929617915  + 67.42423408300 *T)
    sumY0  +=      14.986 * cos(0.08205821916  + 2015.91490364330 *T)
    sumY0  +=      13.750 * cos(0.52643748579  + 451.22103074770 *T)
    sumY0  +=      13.701 * cos(5.78256715764  + 1586.13531902961 *T)
    sumY0  +=      13.415 * cos(2.38228960626  + 2737.07668820650 *T)
    sumY0  +=      13.707 * cos(2.03789341879  + 498.91529394111 *T)
    sumY0  +=      13.957 * cos(2.36417098087  + 1592.49810650500 *T)
    sumY0  +=      12.833 * cos(0.65025343158  + 423.66061462181 *T)
    sumY0  +=      12.935 * cos(5.50182673109  + 599.08730685290 *T)
    sumY0  +=      13.524 * cos(3.44876736954  + 2655.81218995730 *T)
    sumY0  +=      15.347 * cos(4.64705200559  + 323.26159917390 *T)
    sumY0  +=      14.250 * cos(1.72247897628  + 462.26673101160 *T)
    sumY0  +=      14.314 * cos(4.07018795871  + 1551.18367712950 *T)
    sumY0  +=      16.350 * cos(3.38063086403  + 1290.19031849811 *T)
    sumY0  +=      12.174 * cos(5.62306550613  + 2318.07967929830 *T)
    sumY0  +=      14.758 * cos(5.80599958132  + 945.48727319021 *T)
    sumY0  +=      13.727 * cos(4.06971719103  + 113.63153244060 *T)
    sumY0  +=      12.014 * cos(1.59699399907  + 938.37372618940 *T)
    sumY0  +=      12.669 * cos(5.59895980244  + 597.60283414461 *T)
    sumY0  +=      12.684 * cos(5.44519522050  + 560.95427121510 *T)
    sumY0  +=      12.686 * cos(1.24409527288  + 528.23786154870 *T)
    sumY0  +=      12.686 * cos(6.23818302908  + 531.63170360751 *T)
    sumY0  +=      11.902 * cos(0.29922284195  + 636.20895053440 *T)
    sumY0  +=      15.029 * cos(6.08292509395  + 155.53915477459 *T)
    sumY0  +=      14.712 * cos(4.33830015465  + 1060.14701253450 *T)
    sumY0  +=      13.153 * cos(4.24131500267  + 34.18406742730 *T)
    sumY0  +=      11.492 * cos(2.74650212454  + 0.71939036300 *T)
    sumY0  +=      11.728 * cos(0.66803059115  + 608.64853440851 *T)
    sumY0  +=      14.802 * cos(2.91983201484  + 760.49935340350 *T)
    sumY0  +=      14.208 * cos(4.97676580268  + 151.29148732640 *T)
    sumY0  +=      13.247 * cos(4.74780140574  + 5.17280848790 *T)
    sumY0  +=      13.608 * cos(5.75341457451  + 28.06735816780 *T)
    sumY0  +=      11.186 * cos(1.14738473449  + 531.37107117660 *T)
    sumY0  +=      11.186 * cos(0.05170826029  + 528.49849397960 *T)
    sumY0  +=      14.801 * cos(1.71286185185  + 378.39947776819 *T)
    sumY0  +=      11.873 * cos(2.54431600917  + 1677.69475801730 *T)
    sumY0  +=      10.898 * cos(4.59375119105  + 629.84616305900 *T)
    sumY0  +=      14.422 * cos(2.69020416688  + 381.36842318480 *T)
    sumY0  +=      10.799 * cos(5.38556202520  + 1148.49142788971 *T)
    sumY0  +=      10.862 * cos(3.06906378219  + 248.96763557360 *T)
    sumY0  +=      10.584 * cos(2.85396979697  + 430.02340209721 *T)
    sumY0  +=      14.240 * cos(3.01342561614  + 1173.76422215520 *T)
    sumY0  +=      10.278 * cos(4.70239533674  + 209.87388421679 *T)
    sumY0  +=      11.280 * cos(3.91394243595  + 46.71424039951 *T)
    sumY0  +=      10.327 * cos(5.95352479669  + 606.20085385370 *T)
    sumY0  +=      13.270 * cos(2.80639524732  + 1041.46650040800 *T)
    sumY0  +=      13.786 * cos(2.98010668249  + 337.97632814250 *T)
    sumY0  +=      10.252 * cos(5.63891846478  + 1891.48148842230 *T)
    sumY0  +=      13.304 * cos(0.85030585719  + 63.49208081989 *T)
    sumY0  +=       9.721 * cos(5.24565438793  + 1159.53712815361 *T)
    sumY0  +=       9.974 * cos(3.07602930457  + 9.80504503910 *T)
    sumY0  +=      10.564 * cos(4.77844769270  + 1251.09656714130 *T)
    sumY0  +=      12.937 * cos(2.55201590769  + 944.73900579230 *T)
    sumY0  +=      12.666 * cos(4.49628001198  + 721.89323701370 *T)
    sumY0  +=       9.141 * cos(3.31988100263  + 1780.78753223589 *T)
    sumY0  +=       9.377 * cos(3.52298776594  + 2523.77759276850 *T)
    sumY0  +=       8.992 * cos(1.05072894634  + 1165.89991562901 *T)
    sumY0  +=       9.074 * cos(2.27653272320  + 754.27957828001 *T)
    sumY0  +=       8.886 * cos(0.87970824582  + 528.71094230071 *T)
    sumY0  +=       8.886 * cos(0.31938474896  + 531.15862285550 *T)
    sumY0  +=       8.815 * cos(0.50994270535  + 302.40859313851 *T)
    sumY0  +=      12.114 * cos(3.28137353918  + 1020.26887175750 *T)
    sumY0  +=       8.872 * cos(3.19935594712  + 194.89603068979 *T)
    sumY0  +=      10.983 * cos(0.29488310355  + 1.24065522479 *T)
    sumY0  +=      12.190 * cos(4.57593024424  + 72.31710306511 *T)
    sumY0  +=       8.635 * cos(4.87675285165  + 1056.70498636590 *T)
    sumY0  +=      10.308 * cos(1.29678542884  + 2950.37578364450 *T)
    sumY0  +=      11.595 * cos(3.43106394734  + 1289.70268353110 *T)
    sumY0  +=       8.931 * cos(1.99292771548  + 956.04533848709 *T)
    sumY0  +=      10.790 * cos(2.13326549126  + 529.04101070080 *T)
    sumY0  +=      10.790 * cos(5.34901281070  + 530.82855445540 *T)
    sumY0  +=       8.192 * cos(5.36020095318  + 13.98327651810 *T)
    sumY0  +=       7.989 * cos(2.25562114310  + 1058.87498814731 *T)
    sumY0  +=      10.336 * cos(1.83209374906  + 73.54094334250 *T)
    sumY0  +=       8.394 * cos(1.01712525126  + 525.79018099390 *T)
    sumY0  +=      10.766 * cos(5.53480384660  + 2111.40649589409 *T)
    sumY0  +=       7.749 * cos(0.18520854013  + 522.07047605190 *T)
    sumY0  +=       8.309 * cos(0.18252654751  + 534.07938416230 *T)
    sumY0  +=      10.415 * cos(3.01801395976  + 319.81708087780 *T)
    sumY0  +=       9.365 * cos(0.13738783165  + 415.30867312860 *T)
    sumY0  +=       7.601 * cos(1.01410942526  + 537.79908910430 *T)
    sumY0  +=       8.635 * cos(3.30118117017  + 1062.54650897950 *T)
    sumY0  +=       7.859 * cos(5.09973255259  + 1065.04237364410 *T)
    sumY0  +=       7.602 * cos(5.07538034057  + 1063.81853336670 *T)
    sumY0  +=      10.315 * cos(1.81132365985  + 456.63765671910 *T)
    sumY0  +=       7.989 * cos(5.92574529041  + 1060.37650719810 *T)
    sumY0  +=       8.083 * cos(3.25152658141  + 533.37930087419 *T)
    sumY0  +=       7.529 * cos(2.72685691181  + 2847.28300942589 *T)
    sumY0  +=       8.490 * cos(1.21946880524  + 202.00957769060 *T)
    sumY0  +=       7.329 * cos(3.75411316314  + 148.32254190981 *T)
    sumY0  +=       7.187 * cos(5.83227131060  + 30.77567115350 *T)
    sumY0  +=       9.756 * cos(5.09705846628  + 2008.31372167550 *T)
    sumY0  +=       8.727 * cos(1.02210873113  + 1202.07539780650 *T)
    sumY0  +=       7.941 * cos(1.93425844896  + 2332.30677329990 *T)
    sumY0  +=       6.801 * cos(1.23043017312  + 1054.95988122670 *T)
    sumY0  +=       8.387 * cos(4.41387774831  + 2957.97696561230 *T)
    sumY0  +=       7.512 * cos(5.35620649098  + 952.60082019101 *T)
    sumY0  +=       7.552 * cos(4.91464670813  + 842.39449897160 *T)
    sumY0  +=       7.680 * cos(4.58781712019  + 334.79493440480 *T)
    sumY0  +=       9.231 * cos(5.41598830183  + 2854.88419139370 *T)
    sumY0  +=       8.791 * cos(5.81348495155  + 483.46435966210 *T)
    sumY0  +=       7.389 * cos(0.72681677735  + 202.49721265760 *T)
    sumY0  +=       6.847 * cos(0.68329056266  + 320.56784040320 *T)
    sumY0  +=       8.929 * cos(5.67328579883  + 603.23190843711 *T)
    sumY0  +=       6.520 * cos(2.12029864826  + 3053.46855786309 *T)
    sumY0  +=       6.721 * cos(5.17441034834  + 1912.82212942470 *T)
    sumY0  +=       8.859 * cos(2.25656725982  + 217.47506618460 *T)
    sumY0  +=       8.710 * cos(1.69360357974  + 576.40520549410 *T)
    sumY0  +=       7.535 * cos(1.43209684879  + 1061.11022038100 *T)
    sumY0  +=       6.826 * cos(0.61078031935  + 284.39235822571 *T)
    sumY0  +=       6.401 * cos(6.21089623686  + 117.07605073670 *T)
    sumY0  +=       6.884 * cos(3.68154907894  + 277.27881122490 *T)
    sumY0  +=       6.632 * cos(1.40967759841  + 2.93757625420 *T)
    sumY0  +=       7.529 * cos(3.26594046441  + 423.17297965479 *T)
    sumY0  +=       6.885 * cos(3.41541733477  + 946.23803271561 *T)
    sumY0  +=       6.200 * cos(1.01085916847  + 1272.43720814370 *T)
    sumY0  +=       6.730 * cos(0.63714531653  + 531.57931398080 *T)
    sumY0  +=       6.730 * cos(0.56194767825  + 528.29025117541 *T)
    sumY0  +=       6.259 * cos(4.17880072433  + 142.69346761730 *T)
    sumY0  +=       7.602 * cos(3.10155768831  + 1055.43296197871 *T)
    sumY0  +=       7.329 * cos(1.34623356610  + 5.97595764000 *T)
    sumY0  +=       5.901 * cos(5.92152237589  + 1251.58420210830 *T)
    sumY0  +=       6.013 * cos(5.10743465878  + 1035.24672528451 *T)
    sumY0  +=       5.953 * cos(2.93486163481  + 498.44221318910 *T)
    sumY0  +=       7.833 * cos(5.94245831651  + 68.90870679129 *T)
    sumY0  +=       6.343 * cos(1.17088877750  + 199.31581891990 *T)
    sumY0  +=       5.843 * cos(5.61675953661  + 1276.85699637381 *T)
    sumY0  +=       7.666 * cos(3.22002799193  + 247.48316286530 *T)
    sumY0  +=       5.663 * cos(3.16164223175  + 1091.86907658711 *T)
    sumY0  +=       5.605 * cos(1.26367531224  + 513.47245634280 *T)
    sumY0  +=       5.524 * cos(4.60275692299  + 2847.77064439291 *T)
    sumY0  +=       7.612 * cos(4.10551473719  + 352.06040979221 *T)
    sumY0  +=       5.914 * cos(3.68583540418  + 1054.20912170131 *T)
    sumY0  +=       7.627 * cos(0.07913853996  + 405.50136735710 *T)
    sumY0  +=       5.507 * cos(4.54723285635  + 561.42735196710 *T)
    sumY0  +=       5.497 * cos(4.40642501831  + 810.90192958261 *T)
    sumY0  +=       5.666 * cos(6.05825689215  + 1064.29161411871 *T)
    sumY0  +=       5.378 * cos(6.21860298014  + 546.39710881341 *T)
    sumY0  +=       6.319 * cos(3.19557491036  + 1493.33748554950 *T)
    sumY0  +=       6.876 * cos(1.94945842378  + 408.19512612779 *T)
    sumY0  +=       5.745 * cos(1.20632367499  + 430.28652665560 *T)
    sumY0  +=       6.791 * cos(5.62982891906  + 303.61787920090 *T)
    sumY0  +=       7.093 * cos(4.40070510881  + 740.05248427840 *T)
    sumY0  +=       6.815 * cos(1.54708468197  + 1053.99667338020 *T)
    sumY0  +=       5.111 * cos(2.62007621305  + 1176.94561589290 *T)
    sumY0  +=       5.065 * cos(2.54426472948  + 1059.46568897830 *T)
    sumY0  +=       6.697 * cos(3.57793403810  + 685.23011986920 *T)
    sumY0  +=       5.674 * cos(4.57629153060  + 1567.48843679789 *T)
    sumY0  +=       5.759 * cos(4.23445817390  + 166.58485503849 *T)
    sumY0  +=       4.933 * cos(1.00231605898  + 245.29860686890 *T)
    sumY0  +=       5.446 * cos(1.85472048115  + 757.46097201770 *T)
    sumY0  +=       5.053 * cos(0.08352218214  + 1678.18239298430 *T)
    sumY0  +=       5.999 * cos(2.52195451833  + 908.57807782981 *T)
    sumY0  +=       5.382 * cos(2.05458907199  + 512.66930719071 *T)
    sumY0  +=       4.943 * cos(3.25188141823  + 2840.16946242510 *T)
    sumY0  +=       5.065 * cos(5.63710170403  + 1059.78580636710 *T)
    sumY0  +=       5.075 * cos(4.81676406685  + 508.10650660869 *T)
    sumY0  +=       4.955 * cos(1.84826604162  + 1069.92068841121 *T)
    sumY0  +=       5.153 * cos(5.42768922381  + 547.20025796550 *T)
    sumY0  +=       4.607 * cos(5.26442210740  + 2634.47154895491 *T)
    sumY0  +=       4.755 * cos(4.04041807467  + 415.04804069769 *T)
    sumY0  +=       4.775 * cos(2.39668916090  + 1994.08662767389 *T)
    sumY0  +=       4.590 * cos(3.05724325625  + 2097.17940189250 *T)
    sumY0  +=       4.911 * cos(6.22970806427  + 512.52629483880 *T)
    sumY0  +=       4.553 * cos(1.47995846767  + 509.48791345300 *T)
    sumY0  +=       4.672 * cos(2.03709808328  + 234.88355392390 *T)
    sumY0  +=       4.578 * cos(5.57782400113  + 305.58998687620 *T)
    sumY0  +=       5.193 * cos(5.59541519165  + 320.08020543620 *T)
    sumY0  +=       6.202 * cos(2.81584993974  + 26088.14695905770 *T)
    sumY0  +=       5.852 * cos(5.17781036835  + 2427.79836555070 *T)
    sumY0  +=       4.324 * cos(6.00231982961  + 550.38165170320 *T)
    sumY0  +=       4.410 * cos(0.89841714751  + 1127.29379923920 *T)
    sumY0  +=       4.238 * cos(1.37716518074  + 92.29089143820 *T)
    sumY0  +=       4.881 * cos(2.90321352590  + 725.07463075140 *T)
    sumY0  +=       4.688 * cos(5.54226574385  + 504.80500066450 *T)
    sumY0  +=       4.432 * cos(3.21149906214  + 1490.15609181181 *T)
    sumY0  +=       4.144 * cos(1.71460956771  + 970.76006745570 *T)
    sumY0  +=       5.737 * cos(3.37724812325  + 707.80915536400 *T)
    sumY0  +=       4.256 * cos(2.95543164571  + 512.40867475980 *T)
    sumY0  +=       4.139 * cos(5.15386793268  + 470.13103753780 *T)
    sumY0  +=       4.547 * cos(0.52585543495  + 857.37235249860 *T)
    sumY0  +=       4.431 * cos(4.63632908961  + 519.90047427050 *T)
    sumY0  +=       5.723 * cos(5.51760124056  + 1065.25482196520 *T)
    sumY0  +=       4.278 * cos(0.51107431683  + 739.30172475301 *T)
    sumY0  +=       5.621 * cos(4.76211011081  + 280.72332952100 *T)
    sumY0  +=       5.007 * cos(4.52013629124  + 492.06237937121 *T)
    sumY0  +=       3.987 * cos(1.06371097161  + 448.18264936190 *T)
    sumY0  +=       4.559 * cos(1.08649118563  + 3266.76765330109 *T)
    sumY0  +=       4.072 * cos(3.65895344457  + 3061.06973983091 *T)
    sumY0  +=       4.072 * cos(4.52684665530  + 547.46089039640 *T)
    sumY0  +=       5.087 * cos(3.27441637503  + 114.64292439690 *T)
    sumY0  +=       4.392 * cos(1.93950112163  + 555.06456449170 *T)
    sumY0  +=       4.733 * cos(0.67137979921  + 1087.93692332400 *T)
    sumY0  +=       4.590 * cos(0.81771313519  + 209.12312469139 *T)
    sumY0  +=       4.061 * cos(0.84886600079  + 1578.27101250340 *T)
    sumY0  +=       4.043 * cos(3.31066824827  + 917.17609753890 *T)
    sumY0  +=       4.781 * cos(0.95570708773  + 355.99256305531 *T)
    sumY0  +=       3.840 * cos(0.34108693081  + 479.53220639900 *T)
    sumY0  +=       4.229 * cos(5.02030190341  + 202.76033721600 *T)
    sumY0  +=       5.079 * cos(4.10108252940  + 100.40127618030 *T)
    sumY0  +=       3.809 * cos(5.04402768610  + 63.97971578690 *T)
    sumY0  +=       3.869 * cos(2.47318190725  + 3163.67487908250 *T)
    sumY0  +=       4.615 * cos(3.86069192120  + 8.32057233081 *T)
    sumY0  +=       4.615 * cos(0.71909926761  + 7.83293736379 *T)
    sumY0  +=       4.628 * cos(2.95946199977  + 567.80718578500 *T)
    sumY0  +=       4.429 * cos(3.33484643205  + 1025.68549772890 *T)
    sumY0  +=       4.166 * cos(1.09654835252  + 1587.83224005900 *T)
    sumY0  +=       4.232 * cos(2.53213396378  + 2042.74160658630 *T)
    sumY0  +=       3.934 * cos(3.61075823093  + 1894.66288216001 *T)
    sumY0  +=       4.119 * cos(3.84993499525  + 782.59075393131 *T)
    sumY0  +=       4.305 * cos(2.47219780188  + 237.92193530970 *T)
    sumY0  +=       4.647 * cos(5.82365364657  + 294.80741117069 *T)
    sumY0  +=       3.571 * cos(2.08868123477  + 1080.96638867510 *T)
    sumY0  +=       3.908 * cos(0.59904528751  + 775.47720693050 *T)
    sumY0  +=       4.183 * cos(5.36832575933  + 3171.27606105030 *T)
    sumY0  +=       3.544 * cos(4.38396042060  + 849.50804597240 *T)
    sumY0  +=       4.668 * cos(0.56754531710  + 391.41728570740 *T)
    sumY0  +=       3.825 * cos(3.58651228704  + 1013.15532475670 *T)
    sumY0  +=       4.490 * cos(4.47294860578  + 1066.25165970650 *T)
    sumY0  +=       3.497 * cos(2.80136092523  + 1059.83819599381 *T)
    sumY0  +=       4.592 * cos(6.18306867454  + 105.29663728990 *T)
    sumY0  +=       3.507 * cos(2.32963957274  + 589.73852761840 *T)
    sumY0  +=       3.460 * cos(3.00015782763  + 210.36151918381 *T)
    sumY0  +=       3.472 * cos(0.85727261261  + 580.33735875721 *T)
    sumY0  +=       4.742 * cos(1.07044014702  + 1214.92108496379 *T)
    sumY0  +=       4.236 * cos(6.09342108227  + 458.59770230690 *T)
    sumY0  +=       4.571 * cos(4.85712461466  + 2324.70559133209 *T)
    sumY0  +=       3.680 * cos(1.21278522879  + 547.34327031741 *T)
    sumY0  +=       3.342 * cos(0.15031101728  + 611.68691579430 *T)
    sumY0  +=       3.327 * cos(3.80363745236  + 416.05943265400 *T)
    sumY0  +=       3.357 * cos(3.57120537469  + 102.89714084490 *T)
    sumY0  +=       3.597 * cos(0.70360923125  + 81.50831573269 *T)
    sumY0  +=       3.697 * cos(2.78415450903  + 326.93062787860 *T)
    sumY0  +=       3.305 * cos(5.62616411694  + 3068.18328683170 *T)
    sumY0  +=       3.332 * cos(0.38986283721  + 102.80077262530 *T)
    sumY0  +=       3.850 * cos(3.22603061160  + 223.10414047710 *T)
    sumY0  +=       3.389 * cos(4.40791637427  + 1050.06452011710 *T)
    sumY0  +=       3.693 * cos(2.30849108411  + 100.88891114730 *T)
    sumY0  +=       3.520 * cos(5.90372622394  + 355.50492808829 *T)
    sumY0  +=       3.747 * cos(3.26414742353  + 101.88574888859 *T)
    sumY0  +=       3.497 * cos(5.38000550828  + 1059.41329935160 *T)
    sumY0  +=       4.041 * cos(3.71442466403  + 80.75755620730 *T)
    sumY0  +=       3.215 * cos(5.24772956727  + 2207.87335807890 *T)
    sumY0  +=       3.350 * cos(3.76399871254  + 1077.78499493740 *T)
    sumY0  +=       3.628 * cos(4.56818292330  + 1567.97607176490 *T)
    sumY0  +=       4.303 * cos(0.65966587333  + 374.15181032000 *T)
    sumY0  +=       3.455 * cos(3.42091235598  + 489.09343395460 *T)
    sumY0  +=       4.402 * cos(2.90784212539  + 987.55246209111 *T)
    sumY0  +=       3.638 * cos(0.71407157961  + 103.81216458159 *T)
    sumY0  +=       3.112 * cos(5.91930979689  + 422.42222012939 *T)
    sumY0  +=       3.124 * cos(1.89592603599  + 3369.86042751970 *T)
    sumY0  +=       3.303 * cos(2.85125218700  + 539.96909088571 *T)
    sumY0  +=       3.803 * cos(3.64352844828  + 852.95256426850 *T)
    sumY0  +=       3.982 * cos(5.66129151662  + 986.32862181370 *T)
    sumY0  +=       3.666 * cos(1.43198782274  + 77.99436146740 *T)
    sumY0  +=       4.012 * cos(0.62727446399  + 668.45227944880 *T)
    sumY0  +=       3.922 * cos(0.99606286392  + 1215.40871993080 *T)
    sumY0  +=       3.967 * cos(1.33281131206  + 486.64575339981 *T)
    sumY0  +=       3.197 * cos(3.65023662550  + 1169.34443392509 *T)
    sumY0  +=       3.122 * cos(3.82899115941  + 69.39634175830 *T)
    sumY0  +=       3.954 * cos(2.04035652666  + 1123.36164597610 *T)
    sumY0  +=       3.201 * cos(2.09572762487  + 145.87486135500 *T)
    sumY0  +=       3.935 * cos(3.38057342433  + 306.58682461749 *T)
    sumY0  +=       2.958 * cos(3.08516031897  + 531.94052014820 *T)
    sumY0  +=       2.958 * cos(4.39711798299  + 527.92904500800 *T)
    sumY0  +=       3.054 * cos(0.64525087742  + 654.36819779910 *T)
    sumY0  +=       2.909 * cos(4.40702607221  + 536.93645499430 *T)
    sumY0  +=       2.909 * cos(3.07525222975  + 522.93311016190 *T)
    sumY0  +=       3.738 * cos(3.63891493774  + 977.73060210460 *T)
    sumY0  +=       3.893 * cos(4.26177377165  + 812.38640229090 *T)
    sumY0  +=       3.604 * cos(0.46524982149  + 78.46993434690 *T)
    sumY0  +=       2.867 * cos(5.79071320564  + 1269.74344937300 *T)
    sumY0  +=       3.118 * cos(4.06136594960  + 570.77613120160 *T)
    sumY0  +=       3.676 * cos(1.06915971991  + 226.07308589371 *T)
    sumY0  +=       3.463 * cos(2.60208545891  + 1591.76439332211 *T)
    sumY0  +=       2.870 * cos(0.26308381178  + 1586.86903221250 *T)
    sumY0  +=       2.889 * cos(4.31193778169  + 504.66198831261 *T)
    sumY0  +=       3.107 * cos(0.25809701396  + 1905.22094745690 *T)
    sumY0  +=       2.833 * cos(3.77574980636  + 911.54702324640 *T)
    sumY0  +=       3.315 * cos(2.51210337945  + 1065.84552279621 *T)
    sumY0  +=       3.556 * cos(6.14962768045  + 573.22381175640 *T)
    sumY0  +=       2.918 * cos(5.73628341762  + 967.57867371801 *T)
    sumY0  +=       2.676 * cos(0.03275249694  + 3281.48238226970 *T)
    sumY0  +=       3.689 * cos(1.70487674721  + 850.25880549781 *T)
    sumY0  +=       3.016 * cos(4.27158737525  + 3274.36883526891 *T)
    sumY0  +=       3.204 * cos(4.77926383698  + 2744.19023520730 *T)
    sumY0  +=       2.667 * cos(5.49050260804  + 97.91996573070 *T)
    sumY0  +=       2.647 * cos(0.88177591711  + 134.82916109110 *T)
    sumY0  +=       2.715 * cos(3.16963522250  + 555.20757684360 *T)
    sumY0  +=       3.135 * cos(3.15988503239  + 433.46792039329 *T)
    sumY0  +=       2.639 * cos(2.26104148377  + 144.39038864671 *T)
    sumY0  +=       2.912 * cos(4.27421635124  + 1883.88030645450 *T)
    sumY0  +=       3.601 * cos(0.47145843816  + 991.95769610620 *T)
    sumY0  +=       2.745 * cos(0.40065346771  + 33.69643246030 *T)
    sumY0  +=       2.589 * cos(0.44277003092  + 93.28772917949 *T)
    sumY0  +=       2.531 * cos(1.95807014850  + 108.00245814810 *T)
    sumY0  +=       2.821 * cos(1.06656582689  + 503.10807963510 *T)
    sumY0  +=       2.585 * cos(2.99923552183  + 3156.56133208170 *T)
    sumY0  +=       3.142 * cos(5.99895869762  + 520.37355502251 *T)
    sumY0  +=       3.142 * cos(4.62490556976  + 539.49601013370 *T)
    sumY0  +=       2.731 * cos(2.80890465706  + 537.16020416351 *T)
    sumY0  +=       2.731 * cos(4.67337364490  + 522.70936099270 *T)
    sumY0  +=       3.288 * cos(0.54732144888  + 685.71775483620 *T)
    sumY0  +=       2.456 * cos(1.47279271371  + 1074.60360119970 *T)
    sumY0  +=       2.382 * cos(3.93679555128  + 112.89781925770 *T)
    sumY0  +=       2.385 * cos(0.89159302945  + 1475.17823828480 *T)
    sumY0  +=       2.456 * cos(0.42538841262  + 1044.64789414570 *T)
    sumY0  +=       2.435 * cos(5.46783161031  + 824.98601123230 *T)
    sumY0  +=       2.637 * cos(0.13252276208  + 556.76148552110 *T)
    sumY0  +=       3.031 * cos(4.10608213511  + 538.53280228720 *T)
    sumY0  +=       3.031 * cos(3.37619616684  + 521.33676286901 *T)
    sumY0  +=       2.342 * cos(1.63307892366  + 518.40144734719 *T)
    sumY0  +=       2.443 * cos(2.60924692115  + 749.85979004990 *T)
    sumY0  +=       3.226 * cos(2.87915335731  + 536.83588125780 *T)
    sumY0  +=       3.226 * cos(4.60312494465  + 523.03368389840 *T)
    sumY0  +=       2.356 * cos(3.28453581204  + 285.61619850311 *T)
    sumY0  +=       2.452 * cos(0.62027944096  + 1090.64523630970 *T)
    sumY0  +=       2.986 * cos(6.19494169694  + 491.31411197330 *T)
    sumY0  +=       2.323 * cos(4.87403397794  + 530.56792202451 *T)
    sumY0  +=       2.323 * cos(2.60824432402  + 529.30164313170 *T)
    sumY0  +=       3.003 * cos(1.51547464417  + 108.96566599461 *T)
    sumY0  +=       3.049 * cos(2.91433771038  + 1438.26904292440 *T)
    sumY0  +=       3.159 * cos(5.47330952290  + 457.86149699650 *T)
    sumY0  +=       3.023 * cos(6.11184318218  + 532.90372799470 *T)
    sumY0  +=       3.023 * cos(1.37043511978  + 526.96583716151 *T)
    sumY0  +=       2.581 * cos(4.04501510183  + 1483.04254481100 *T)
    sumY0  +=       2.851 * cos(6.19876101268  + 731.70054278519 *T)
    sumY0  +=       2.627 * cos(2.23895869315  + 1095.05047032480 *T)
    sumY0  +=       2.198 * cos(2.83199855971  + 2413.57127154910 *T)
    sumY0  +=       2.533 * cos(0.10403644084  + 12566.39551746630 *T)
    sumY0  +=       2.183 * cos(4.47686604255  + 2421.17245351690 *T)
    sumY0  +=       2.273 * cos(6.03923806707  + 1589.36489687710 *T)
    sumY0  +=       2.201 * cos(4.15470899026  + 2044.22607929460 *T)
    sumY0  +=       2.304 * cos(2.84269200130  + 1053.47540851841 *T)
    sumY0  +=       2.878 * cos(1.28856851089  + 28.55499313480 *T)
    sumY0  +=       2.523 * cos(1.46775781223  + 909.57491557110 *T)
    sumY0  +=       2.847 * cos(3.89864676802  + 1441.23798834100 *T)
    sumY0  +=       2.211 * cos(5.00686455045  + 376.43943218040 *T)
    sumY0  +=       2.767 * cos(5.82488714403  + 39.86132582961 *T)
    sumY0  +=       2.242 * cos(4.98350060741  + 6681.46867088311 *T)
    sumY0  +=       2.284 * cos(1.57111216540  + 902.94900353731 *T)
    sumY0  +=       2.122 * cos(4.38960996491  + 543.67424161270 *T)
    sumY0  +=       2.251 * cos(2.16205964084  + 1067.70250252000 *T)
    sumY0  +=       2.782 * cos(5.20069642737  + 451.69660362719 *T)
    sumY0  +=       2.107 * cos(4.35933463496  + 824.49837626530 *T)
    sumY0  +=       2.308 * cos(4.70844962325  + 732.93893727761 *T)
    sumY0  +=       2.055 * cos(5.91536534525  + 526.26575387340 *T)
    sumY0  +=       2.624 * cos(2.11125071189  + 1261.87914284680 *T)
    sumY0  +=       2.059 * cos(4.36499566626  + 1254.76559584600 *T)
    sumY0  +=       2.523 * cos(0.74912600326  + 372.77040347569 *T)
    sumY0  +=       2.035 * cos(3.63513892304  + 535.77630519170 *T)
    sumY0  +=       2.035 * cos(3.84713937892  + 524.09325996451 *T)
    sumY0  +=       2.005 * cos(2.40063053691  + 843.14525849700 *T)
    sumY0  +=       2.366 * cos(1.24399407734  + 3708.08057314570 *T)
    sumY0  +=       1.959 * cos(0.76199223049  + 163.14033674240 *T)
    sumY0  +=       2.451 * cos(2.98565323941  + 739.56484931140 *T)
    sumY0  +=       2.474 * cos(0.77300859930  + 2122.93983112501 *T)
    sumY0  +=       1.935 * cos(4.38042537965  + 3164.16251404950 *T)
    sumY0  +=       2.548 * cos(0.72951803247  + 1128.77827194750 *T)
    sumY0  +=       2.140 * cos(3.75230805584  + 408.68276109481 *T)
    sumY0  +=       2.089 * cos(5.41923470052  + 60.04756252381 *T)
    sumY0  +=       2.555 * cos(1.04261904815  + 231.80522605020 *T)
    sumY0  +=       2.289 * cos(2.53062848354  + 429.28968891431 *T)
    sumY0  +=       2.652 * cos(4.81114494152  + 405.01373239010 *T)
    sumY0  +=       2.044 * cos(4.84208091423  + 241.36645360580 *T)
    sumY0  +=       1.959 * cos(2.21322194743  + 3480.06674873909 *T)
    sumY0  +=       1.890 * cos(6.01613452327  + 501.48059457490 *T)
    sumY0  +=       2.361 * cos(1.03417291975  + 1819.88128359270 *T)
    sumY0  +=       2.290 * cos(3.64742542679  + 386.99749747730 *T)
    sumY0  +=       1.839 * cos(3.04284528593  + 511.28790034640 *T)
    sumY0  +=       1.873 * cos(0.59051569111  + 1209.18894480731 *T)
    sumY0  +=       1.837 * cos(4.84296500864  + 111.41334654940 *T)
    sumY0  +=       2.510 * cos(1.99461282730  + 602.00806815971 *T)
    sumY0  +=       2.186 * cos(3.48581623798  + 1553.89199011520 *T)
    sumY0  +=       1.875 * cos(3.39330212644  + 1590.27992061380 *T)
    sumY0  +=       2.325 * cos(1.42158890974  + 488.88098563350 *T)
    sumY0  +=       1.960 * cos(1.10678196058  + 109.48693085641 *T)
    sumY0  +=       2.511 * cos(6.05202626734  + 167.07249000551 *T)
    sumY0  +=       2.012 * cos(0.85932602524  + 3583.15952295770 *T)
    sumY0  +=       2.198 * cos(4.38757351366  + 297.98880490840 *T)
    sumY0  +=       2.129 * cos(5.01247506157  + 821.94762984651 *T)
    sumY0  +=       2.219 * cos(4.67528803846  + 2641.09746098870 *T)
    sumY0  +=       1.954 * cos(5.49745211708  + 1415.37449324450 *T)
    sumY0  +=       1.787 * cos(3.66506787987  + 1059.88638010361 *T)
    sumY0  +=       1.933 * cos(2.03855165852  + 1688.47733372280 *T)
    sumY0  +=       2.173 * cos(5.25212332069  + 1065.77608682700 *T)
    sumY0  +=       2.273 * cos(2.84142693944  + 1589.26852865751 *T)
    sumY0  +=       2.311 * cos(4.82462753962  + 36.89238041300 *T)
    sumY0  +=       2.005 * cos(1.05923309333  + 746.67839631219 *T)
    sumY0  +=       1.716 * cos(2.65554696835  + 504.07128748160 *T)
    sumY0  +=       1.970 * cos(2.68003788885  + 363.10611005610 *T)
    sumY0  +=       1.879 * cos(6.24564932571  + 1994.57426264090 *T)
    sumY0  +=       1.715 * cos(5.11493591992  + 2950.86341861150 *T)
    sumY0  +=       1.859 * cos(0.05233968738  + 2221.61281711350 *T)
    sumY0  +=       1.760 * cos(4.24064604671  + 1051.54899282540 *T)
    sumY0  +=       1.679 * cos(2.30938727968  + 2310.47849733050 *T)
    sumY0  +=       2.127 * cos(4.22476224615  + 945.22664075931 *T)
    sumY0  +=       1.827 * cos(1.41025945774  + 110.49832281271 *T)
    sumY0  +=       1.954 * cos(4.77927442918  + 1446.86706263350 *T)
    sumY0  +=       2.077 * cos(1.69841550605  + 724.58699578440 *T)
    sumY0  +=       2.096 * cos(6.06068383411  + 570.98857952271 *T)
    sumY0  +=       1.615 * cos(5.64607488965  + 1052.56038478170 *T)
    sumY0  +=       1.608 * cos(4.82673133397  + 555.79827767460 *T)
    sumY0  +=       1.681 * cos(3.99691992442  + 516.44140175940 *T)
    sumY0  +=       1.681 * cos(3.48535837754  + 543.42816339680 *T)
    sumY0  +=       1.618 * cos(3.60392485527  + 112.41018429070 *T)
    sumY0  +=       1.962 * cos(4.18314006350  + 836.76542467910 *T)
    sumY0  +=       1.962 * cos(4.18314006350  + 836.27778971210 *T)
    sumY0  +=       1.620 * cos(2.47522214520  + 683.43013297580 *T)
    sumY0  +=       1.551 * cos(0.77201297425  + 593.67068088151 *T)
    sumY0  +=       1.819 * cos(4.45299196922  + 3060.58210486390 *T)
    sumY0  +=       1.634 * cos(6.04073888889  + 56.37853381910 *T)
    sumY0  +=       1.791 * cos(4.81648790148  + 502.84744720420 *T)
    sumY0  +=       2.008 * cos(0.95169324404  + 487.39651292520 *T)
    sumY0  +=       1.551 * cos(1.50150617084  + 3686.25229717630 *T)
    sumY0  +=       1.511 * cos(0.18786787527  + 1518.46726746311 *T)
    sumY0  +=       1.910 * cos(3.25977004432  + 228.83628063359 *T)
    sumY0  +=       1.738 * cos(5.42191577135  + 1226.45442019471 *T)
    sumY0  +=       1.471 * cos(4.48878361688  + 350.57593708390 *T)
    sumY0  +=       1.466 * cos(3.39788090489  + 426.79382424970 *T)
    sumY0  +=       1.636 * cos(4.00371950748  + 2200.27217611110 *T)
    sumY0  +=       1.940 * cos(1.85697551551  + 1432.63996863190 *T)
    sumY0  +=       1.484 * cos(2.41547562672  + 995.88984936931 *T)
    sumY0  +=       1.673 * cos(2.66579038613  + 557.02211795201 *T)
    sumY0  +=       2.009 * cos(2.30992514879  + 417.28078080390 *T)
    sumY0  +=       1.573 * cos(6.26224126504  + 425.87880051300 *T)
    sumY0  +=       1.562 * cos(5.88372601556  + 416.54706762100 *T)
    sumY0  +=       1.419 * cos(4.93673377317  + 216.72430665921 *T)
    sumY0  +=       1.787 * cos(4.51629855363  + 1059.36511524180 *T)
    sumY0  +=       1.600 * cos(1.96218529204  + 1268.99268984760 *T)
    sumY0  +=       1.459 * cos(0.29602595476  + 426.89019246930 *T)
    sumY0  +=       1.860 * cos(5.30365267047  + 1160.27084133651 *T)
    sumY0  +=       1.409 * cos(5.78259738624  + 383.81610373959 *T)
    sumY0  +=       1.599 * cos(5.70543219748  + 1588.35350492080 *T)
    sumY0  +=       1.385 * cos(0.39231250148  + 466.19888427470 *T)
    sumY0  +=       1.645 * cos(4.96839986582  + 630.57987624190 *T)
    sumY0  +=       1.867 * cos(4.87609131592  + 1744.61205005840 *T)
    sumY0  +=       1.796 * cos(0.70669960542  + 424.39432780471 *T)
    sumY0  +=       1.802 * cos(0.24900630046  + 572.47305223100 *T)
    sumY0  +=       1.563 * cos(1.20698143227  + 312.21589891000 *T)
    sumY0  +=       1.811 * cos(5.89884616616  + 1340.59289467720 *T)
    sumY0  +=       1.342 * cos(4.09814653190  + 849.77117053080 *T)
    sumY0  +=       1.409 * cos(2.76107795785  + 3472.95320173830 *T)
    sumY0  +=       1.340 * cos(4.11708835504  + 1058.35372328550 *T)
    sumY0  +=       1.342 * cos(3.46763239406  + 3377.46160948750 *T)
    sumY0  +=       1.764 * cos(5.75117118327  + 312.96665843540 *T)
    sumY0  +=       1.801 * cos(0.07014275858  + 833.79647926250 *T)
    sumY0  +=       1.801 * cos(3.21173541218  + 833.30884429549 *T)
    sumY0  +=       1.504 * cos(2.41769920056  + 1052.46401656211 *T)
    sumY0  +=       1.346 * cos(5.55344965198  + 1066.78747878330 *T)
    sumY0  +=       1.711 * cos(0.54051201789  + 1363.48744435711 *T)
    sumY0  +=       1.382 * cos(5.00462801017  + 24.13520490470 *T)
    sumY0  +=       1.636 * cos(4.96830039642  + 422.64922266551 *T)
    sumY0  +=       1.714 * cos(1.09473818717  + 1165.16620244611 *T)
    sumY0  +=       1.547 * cos(5.14492139001  + 1097.75878331050 *T)
    sumY0  +=       1.706 * cos(5.12109897845  + 1696.34164024900 *T)
    sumY0  +=       1.293 * cos(1.41573154819  + 394.59867944510 *T)
    sumY0  +=       1.361 * cos(5.15743428812  + 1579.02177202880 *T)
    sumY0  +=       1.329 * cos(0.80850930613  + 1628.67358868250 *T)
    sumY0  +=       1.657 * cos(1.38618195393  + 473.31243127551 *T)
    sumY0  +=       1.342 * cos(2.46878559927  + 827.67977000299 *T)
    sumY0  +=       1.353 * cos(2.59305944840  + 2499.13544582190 *T)
    sumY0  +=       1.322 * cos(3.36892891689  + 1464.88329754630 *T)
    sumY0  +=       1.528 * cos(5.01222546113  + 138.76131435421 *T)
    sumY0  +=       1.507 * cos(5.88638092049  + 774.98957196349 *T)
    sumY0  +=       1.375 * cos(2.40966547086  + 1093.56599761650 *T)
    sumY0  +=       1.393 * cos(5.43252740056  + 913.99470380120 *T)
    sumY0  +=       1.302 * cos(6.01486777282  + 494.72250824710 *T)
    sumY0  +=       1.266 * cos(4.76723112510  + 1162.76670600110 *T)
    sumY0  +=       1.602 * cos(5.54171392759  + 611.19928082730 *T)
    sumY0  +=       1.243 * cos(0.73887599211  + 561.69047652550 *T)
    sumY0  +=       1.234 * cos(0.32614546594  + 436.40323591510 *T)
    sumY0  +=       1.258 * cos(1.58417788394  + 1162.67033778151 *T)
    sumY0  +=       1.490 * cos(2.87282781805  + 1468.06469128401 *T)
    sumY0  +=       1.419 * cos(1.54071741199  + 1138.33949950310 *T)
    sumY0  +=       1.394 * cos(5.15652264283  + 958.98065400890 *T)
    sumY0  +=       1.290 * cos(3.58642556110  + 990.47322339791 *T)
    sumY0  +=       1.199 * cos(5.33903838516  + 414.31183538731 *T)
    sumY0  +=       1.364 * cos(2.24443885688  + 252.89978883670 *T)
    sumY0  +=       1.494 * cos(3.08741560160  + 1485.73630358170 *T)
    sumY0  +=       1.367 * cos(3.04146468402  + 527.96267490280 *T)
    sumY0  +=       1.464 * cos(2.33832116265  + 1066.69111056371 *T)
    sumY0  +=       1.248 * cos(4.58111803711  + 9.31741007210 *T)
    sumY0  +=       1.194 * cos(4.20928976432  + 774.25336665311 *T)
    sumY0  +=       1.328 * cos(3.90212909678  + 482.97672469510 *T)
    sumY0  +=       1.176 * cos(4.08169238786  + 633.07574090651 *T)
    sumY0  +=       1.352 * cos(4.83033510793  + 696.76345510010 *T)
    sumY0  +=       1.235 * cos(6.00673735479  + 1027.38241875830 *T)
    sumY0  +=       1.192 * cos(1.46741047570  + 565.14705690910 *T)
    sumY0  +=       1.337 * cos(3.61251738760  + 427.80521620600 *T)
    sumY0  +=       1.581 * cos(5.13731640491  + 642.58878435231 *T)
    sumY0  +=       1.186 * cos(3.09982091978  + 2115.07552459880 *T)
    sumY0  +=       1.339 * cos(4.45992661727  + 1161.75531404480 *T)
    sumY0  +=       1.260 * cos(1.21412389098  + 633.99076464320 *T)
    sumY0  +=       1.309 * cos(1.91003196451  + 1163.68172973780 *T)
    sumY0  +=       1.221 * cos(6.07359053280  + 2538.00468677009 *T)
    sumY0  +=       1.133 * cos(5.69121919762  + 738.81408978600 *T)
    sumY0  +=       1.172 * cos(0.90157251525  + 632.97937268690 *T)
    sumY0  +=       1.120 * cos(3.02627322689  + 709.29362807231 *T)
    sumY0  +=       1.340 * cos(4.06427807846  + 1060.89777205990 *T)
    sumY0  +=       1.126 * cos(2.69535643306  + 2729.96314120569 *T)
    sumY0  +=       1.417 * cos(2.51678541283  + 1364.48428209840 *T)
    sumY0  +=       1.112 * cos(3.72336688315  + 1066.21802981170 *T)
    sumY0  +=       1.312 * cos(6.06573442442  + 981.87520368880 *T)
    sumY0  +=       1.496 * cos(4.33280725286  + 388.70897272171 *T)
    sumY0  +=       1.423 * cos(6.09024848918  + 586.55713388070 *T)
    sumY0  +=       1.221 * cos(5.21060853641  + 269.67762925710 *T)
    sumY0  +=       1.088 * cos(6.00908281655  + 1280.03839011150 *T)
    sumY0  +=       1.089 * cos(3.60595774779  + 946.97174589851 *T)
    sumY0  +=       1.132 * cos(5.22322415543  + 1664.09831133461 *T)
    sumY0  +=       1.433 * cos(0.48122466534  + 635.47523735150 *T)
    sumY0  +=       1.427 * cos(4.54173500327  + 597.11519917759 *T)
    sumY0  +=       1.389 * cos(5.65859882018  + 636.47207509279 *T)
    sumY0  +=       1.191 * cos(3.52581460046  + 1887.54933515920 *T)
    sumY0  +=       1.420 * cos(4.97687573513  + 1140.13948639650 *T)
    sumY0  +=       1.420 * cos(4.97687573513  + 1140.62712136350 *T)
    sumY0  +=       1.082 * cos(5.23314277920  + 915.47917650950 *T)
    sumY0  +=       1.134 * cos(1.71654085995  + 102.32769187330 *T)
    sumY0  +=       1.061 * cos(0.33709671438  + 925.04040406510 *T)
    sumY0  +=       1.302 * cos(2.50290977947  + 637.22034249070 *T)
    sumY0  +=       1.036 * cos(0.53477431200  + 1798.68365494221 *T)
    sumY0  +=       1.210 * cos(0.08512627078  + 31.73638687251 *T)
    sumY0  +=       1.210 * cos(3.22671892437  + 31.24875190549 *T)
    sumY0  +=       1.439 * cos(4.57403818109  + 244.07476659150 *T)
    sumY0  +=       1.020 * cos(3.62336804453  + 732.45130231059 *T)
    sumY0  +=       1.025 * cos(5.53135631392  + 951.86710700811 *T)
    sumY0  +=       1.018 * cos(0.87096787225  + 623.46632924110 *T)
    sumY0  +=       1.027 * cos(5.32423350857  + 1474.91760585390 *T)
    sumY0  +=       1.114 * cos(2.21692840772  + 103.37022159689 *T)
    sumY0  +=       1.040 * cos(0.83745017570  + 314.18800658530 *T)
    sumY0  +=       1.371 * cos(2.70777796640  + 4590.66636300550 *T)
    sumY0  +=       1.256 * cos(4.91935004376  + 73.05330837550 *T)
    sumY0  +=       1.059 * cos(4.12678319156  + 3590.76070492550 *T)
    sumY0  +=       1.008 * cos(4.10762131673  + 1140.89024592190 *T)
    sumY0  +=       1.288 * cos(3.82410217609  + 2004.60857094851 *T)
    sumY0  +=       1.033 * cos(1.08663300387  + 1028.60625903570 *T)
    sumY0  +=       1.085 * cos(2.21886409341  + 395.34943897051 *T)
    sumY0  +=       0.971 * cos(6.06185705215  + 665.27088571111 *T)
    sumY0  +=       1.358 * cos(3.05188631331  + 448.93340888730 *T)
    sumY0  +=       0.980 * cos(1.98914636396  + 961.94959942551 *T)
    sumY0  +=       0.994 * cos(1.97119491280  + 3796.45861839570 *T)
    sumY0  +=       1.202 * cos(3.96954337111  + 453.66871130251 *T)
    sumY0  +=       1.028 * cos(5.20130168371  + 2524.26522773550 *T)
    sumY0  +=       1.040 * cos(0.76651765316  + 170.51700830159 *T)
    sumY0  +=       0.955 * cos(2.14163148842  + 645.55772976890 *T)
    sumY0  +=       1.241 * cos(5.96190045692  + 1048.09241244180 *T)
    sumY0  +=       1.076 * cos(4.54364112469  + 2957.48933064529 *T)
    sumY0  +=       1.230 * cos(0.76262291362  + 1354.67697632690 *T)
    sumY0  +=       1.240 * cos(4.37679167579  + 610.93615626891 *T)
    sumY0  +=       1.240 * cos(1.23519902220  + 610.44852130189 *T)
    sumY0  +=       1.022 * cos(5.26668499759  + 276.79117625789 *T)
    sumY0  +=       0.924 * cos(3.99063073330  + 842.65762352999 *T)
    sumY0  +=       1.083 * cos(0.34974746355  + 1901.77642916080 *T)
    sumY0  +=       1.243 * cos(3.14900420596  + 671.16059243450 *T)
    sumY0  +=       1.013 * cos(4.11841355585  + 3376.97397452050 *T)
    sumY0  +=       1.100 * cos(4.35722465068  + 1053.03346553370 *T)
    sumY0  +=       1.099 * cos(1.23775171296  + 611.82992814620 *T)
    sumY0  +=       1.099 * cos(3.10291088519  + 448.03963701001 *T)
    sumY0  +=       1.033 * cos(3.86976091725  + 632.06434895021 *T)
    sumY0  +=       0.957 * cos(2.32518989533  + 2965.09051261311 *T)
    sumY0  +=       1.030 * cos(1.77683782212  + 1141.37788088890 *T)
    sumY0  +=       1.196 * cos(4.13327524621  + 8520.87605922509 *T)
    sumY0  +=       0.967 * cos(4.90659485068  + 829.06117684730 *T)
    sumY0  +=       1.166 * cos(0.10987883520  + 1511.56616878340 *T)
    sumY0  +=       1.147 * cos(4.05159175951  + 1390.24471133090 *T)
    sumY0  +=       0.873 * cos(0.21746359309  + 322.26476143261 *T)
    sumY0  +=       1.137 * cos(5.96453800922  + 950.90389916161 *T)
    sumY0  +=       1.201 * cos(4.10007739292  + 1436.29693524910 *T)
    sumY0  +=       0.903 * cos(4.75936593867  + 1475.92899781020 *T)
    sumY0  +=       0.858 * cos(4.08152704252  + 3480.55438370610 *T)
    sumY0  +=       1.061 * cos(3.86340607753  + 1014.37916503410 *T)
    sumY0  +=       0.852 * cos(2.28740434163  + 1305.16817202511 *T)
    sumY0  +=       1.121 * cos(2.22650360725  + 2861.99773839450 *T)
    sumY0  +=       0.978 * cos(0.01048414798  + 1262.62990237220 *T)
    sumY0  +=       0.918 * cos(2.29661901563  + 341.01470952831 *T)
    sumY0  +=       0.886 * cos(0.44036055466  + 1482.29178528560 *T)
    sumY0  +=       1.153 * cos(2.60386985545  + 1055.95671896799 *T)
    sumY0  +=       1.011 * cos(0.56246587032  + 1379.94977059240 *T)
    sumY0  +=       0.844 * cos(3.23872317111  + 231.70216018621 *T)
    sumY0  +=       0.939 * cos(3.33692331113  + 1703.45518724980 *T)
    sumY0  +=       1.029 * cos(1.16181986038  + 2545.60586873790 *T)
    sumY0  +=       1.042 * cos(5.64923274668  + 297.88573904440 *T)
    sumY0  +=       1.170 * cos(2.68197224847  + 46.22660543249 *T)
    sumY0  +=       0.962 * cos(5.05559487343  + 37.88921815429 *T)
    sumY0  +=       0.877 * cos(1.00643184196  + 357.68948408471 *T)
    sumY0  +=       0.988 * cos(2.20288476603  + 7211.15963597770 *T)
    sumY0  +=       0.892 * cos(3.84433244725  + 1571.15746550260 *T)
    sumY0  +=       1.109 * cos(5.81958333017  + 348.12825652910 *T)
    sumY0  +=       0.827 * cos(5.97485149381  + 20956.50587500010 *T)
    sumY0  +=       0.934 * cos(5.21360076093  + 525.51499434800 *T)
    sumY0  +=       0.914 * cos(0.28108854924  + 1031.31457202140 *T)
    sumY0  +=       0.998 * cos(2.30051279264  + 1505.93709449090 *T)
    sumY0  +=       1.003 * cos(3.78100971695  + 1681.36378672201 *T)
    sumY0  +=       1.043 * cos(5.91678666313  + 3384.57515648830 *T)
    sumY0  +=       1.133 * cos(5.74894340340  + 1357.85837006460 *T)
    sumY0  +=       1.118 * cos(1.45393796917  + 892.79707515070 *T)
    sumY0  +=       0.818 * cos(1.62737897182  + 1153.15729433570 *T)
    sumY0  +=       0.883 * cos(3.37767668051  + 533.45790092711 *T)
    sumY0  +=       0.883 * cos(4.10460162145  + 526.41166422910 *T)
    sumY0  +=       0.966 * cos(5.59349076815  + 381.85605815180 *T)
    sumY0  +=       1.014 * cos(5.77098535837  + 50.64639366260 *T)
    sumY0  +=       1.014 * cos(2.62939270478  + 50.15875869559 *T)
    sumY0  +=       0.977 * cos(4.64792461073  + 25035.39094090229 *T)
    sumY0  +=       0.933 * cos(5.47050254707  + 671.89679774490 *T)
    sumY0  +=       0.898 * cos(3.74591414099  + 2516.66404576769 *T)
    sumY0  +=       0.787 * cos(2.68667239235  + 1056.91743468700 *T)
    sumY0  +=       0.788 * cos(1.39640318925  + 4002.64416683289 *T)
    sumY0  +=       0.950 * cos(4.11439083259  + 532.62854134880 *T)
    sumY0  +=       0.839 * cos(5.03983812044  + 4686.64559022330 *T)
    sumY0  +=       0.965 * cos(5.19147359292  + 1520.16418849250 *T)
    sumY0  +=       1.017 * cos(1.65929197414  + 1020.00823932660 *T)
    sumY0  +=       0.820 * cos(1.90274013249  + 348.39138108750 *T)
    sumY0  +=       0.786 * cos(1.79549510698  + 178.11819026941 *T)
    sumY0  +=       0.929 * cos(0.76030088993  + 1130.47519297691 *T)
    sumY0  +=       0.779 * cos(1.34733977730  + 224.10097821840 *T)
    sumY0  +=       0.951 * cos(1.75056688455  + 25021.16384690069 *T)
    sumY0  +=       0.917 * cos(0.46397488831  + 3899.55139261430 *T)
    sumY0  +=       0.763 * cos(4.49833164066  + 432.25863433090 *T)
    sumY0  +=       0.841 * cos(1.57808557219  + 643.32249753521 *T)
    sumY0  +=       0.782 * cos(0.66767670749  + 1731.76636290111 *T)
    sumY0  +=       0.945 * cos(0.70467760679  + 758.52724572820 *T)
    sumY0  +=       1.032 * cos(3.19953750862  + 3075.29683383250 *T)
    sumY0  +=       0.899 * cos(5.79764466292  + 1053.40597254920 *T)
    sumY0  +=       0.822 * cos(5.20921850714  + 1354.18934135989 *T)
    sumY0  +=       0.742 * cos(0.43269937600  + 896.72922841380 *T)
    sumY0  +=       0.989 * cos(5.20317411602  + 493.28621964860 *T)
    sumY0  +=       0.989 * cos(2.27910418594  + 566.58334550761 *T)
    sumY0  +=       0.742 * cos(2.64620540713  + 948.45621860680 *T)
    sumY0  +=       0.743 * cos(4.86638449602  + 3267.25528826810 *T)
    sumY0  +=       0.785 * cos(0.04530717907  + 5091.90814063229 *T)
    sumY0  +=       0.735 * cos(5.88444842565  + 1689.22809324820 *T)
    sumY0  +=       0.902 * cos(5.02430551931  + 3487.66793070691 *T)
    sumY0  +=       0.766 * cos(3.02399738953  + 402.46298597130 *T)
    sumY0  +=       0.870 * cos(3.62030616326  + 522.29997071550 *T)
    sumY0  +=       0.870 * cos(0.72037948511  + 537.56959444070 *T)
    sumY0  +=       0.789 * cos(0.10009487038  + 950.38263429980 *T)
    sumY0  +=       0.729 * cos(4.89347516312  + 2226.03260534360 *T)
    sumY0  +=       0.967 * cos(3.04856021602  + 149.31937965110 *T)
    sumY0  +=       0.740 * cos(0.59195492427  + 172.00148100990 *T)
    sumY0  +=       0.732 * cos(3.85868260020  + 1600.36241303120 *T)
    sumY0  +=       0.743 * cos(2.09672296294  + 6011.68934095890 *T)
    sumY0  +=       0.773 * cos(0.18183783219  + 8616.85528644289 *T)
    sumY0  +=       0.718 * cos(5.42139839209  + 1055.20595944260 *T)
    sumY0  +=       0.814 * cos(2.46839638562  + 5.38525680900 *T)
    sumY0  +=       0.780 * cos(6.07431076573  + 949.37124234350 *T)
    sumY0  +=       0.736 * cos(4.07506898321  + 23968.89546371229 *T)
    sumY0  +=       0.703 * cos(4.48261837950  + 2108.71273712340 *T)
    sumY0  +=       0.705 * cos(2.70029551415  + 260.01333583751 *T)
    sumY0  +=       0.837 * cos(1.52005452422  + 97.43233076370 *T)
    sumY0  +=       0.708 * cos(3.40550883575  + 628.60776856660 *T)
    sumY0  +=       0.700 * cos(1.73883435675  + 78.20680978850 *T)
    sumY0  +=       0.704 * cos(5.21043618727  + 621.49422156580 *T)
    sumY0  +=       0.972 * cos(5.51177820347  + 806.96977631951 *T)
    sumY0  +=       0.902 * cos(1.24655233839  + 3259.39098174191 *T)
    sumY0  +=       0.806 * cos(0.63642725064  + 703.87700210090 *T)
    sumY0  +=       0.861 * cos(4.84084957436  + 215.99059347631 *T)
    sumY0  +=       0.761 * cos(5.97371837033  + 2310.96613229750 *T)
    sumY0  +=       0.705 * cos(4.55292890701  + 657.40657918490 *T)
    sumY0  +=       0.758 * cos(2.16941238331  + 1382.64352936309 *T)
    sumY0  +=       0.727 * cos(0.27903898087  + 64.71592109730 *T)
    sumY0  +=       0.687 * cos(4.51235449745  + 952.86394474940 *T)
    sumY0  +=       0.692 * cos(4.02871358383  + 10.05112325500 *T)
    sumY0  +=       0.963 * cos(2.76790770152  + 6916.10340678810 *T)
    sumY0  +=       0.906 * cos(4.48544256463  + 110.40195459310 *T)
    sumY0  +=       0.729 * cos(2.53279790409  + 3789.34507139489 *T)
    sumY0  +=       0.678 * cos(4.97651378738  + 2218.91905834281 *T)
    sumY0  +=       0.810 * cos(1.91237826709  + 678.01350700440 *T)
    sumY0  +=       0.761 * cos(5.70182901948  + 1542.84628985130 *T)
    sumY0  +=       0.734 * cos(6.16258391033  + 4863.44173055270 *T)
    sumY0  +=       0.793 * cos(0.81892149809  + 209.61075965840 *T)
    sumY0  +=       0.666 * cos(1.94007934478  + 4127.56521702090 *T)
    sumY0  +=       0.849 * cos(2.44682670396  + 368.83825021259 *T)
    sumY0  +=       0.815 * cos(0.85093813087  + 359.17395679300 *T)
    sumY0  +=       0.865 * cos(4.95824919504  + 157.02362748290 *T)
    sumY0  +=       0.799 * cos(4.90559666559  + 519.15220687259 *T)
    sumY0  +=       0.730 * cos(0.97760856285  + 98.65617104110 *T)
    sumY0  +=       0.923 * cos(4.75879664446  + 179.60266297770 *T)
    sumY0  +=       0.862 * cos(2.42176207749  + 1669.83045149109 *T)
    sumY0  +=       0.903 * cos(4.81493761835  + 1745.09968502540 *T)
    sumY0  +=       0.704 * cos(5.19561938169  + 718.85485562790 *T)
    sumY0  +=       0.765 * cos(0.82148767779  + 1834.85913711970 *T)
    sumY0  +=       0.651 * cos(2.13022538110  + 153.25153291419 *T)
    sumY0  +=       0.670 * cos(3.96692910866  + 5702.41101830310 *T)
    sumY0  +=       0.689 * cos(5.62284795095  + 540.49284787499 *T)
    sumY0  +=       0.677 * cos(0.37927842067  + 745.68155857090 *T)
    sumY0  +=       0.646 * cos(1.69538207689  + 2118.52004289490 *T)
    sumY0  +=       0.744 * cos(6.04329221154  + 305.85311143459 *T)
    sumY0  +=       0.837 * cos(1.94363906360  + 1818.39681088440 *T)
    sumY0  +=       0.795 * cos(3.06915549224  + 376.91500505990 *T)
    sumY0  +=       0.639 * cos(5.15999306490  + 5282.92637442791 *T)
    sumY0  +=       0.639 * cos(4.79373670634  + 651.18680406140 *T)
    sumY0  +=       0.791 * cos(3.64973766479  + 2432.21815378081 *T)
    sumY0  +=       0.666 * cos(1.47748551243  + 2439.33170078160 *T)
    sumY0  +=       0.699 * cos(2.04088853152  + 999.82200263240 *T)
    sumY0  +=       0.848 * cos(3.15425868901  + 445.00125562421 *T)
    sumY0  +=       0.784 * cos(2.44486888742  + 1372.34858862460 *T)
    sumY0  +=       0.784 * cos(5.48434762861  + 494.02242495900 *T)
    sumY0  +=       0.826 * cos(0.95388118622  + 1063.07026596880 *T)
    sumY0  +=       0.773 * cos(0.18999764486  + 3334.17258030920 *T)
    sumY0  +=       0.766 * cos(5.19735998756  + 927.59115048390 *T)
    sumY0  +=       0.629 * cos(3.97966005206  + 444.10748374690 *T)
    sumY0  +=       0.680 * cos(5.78589517711  + 2854.39655642670 *T)
    sumY0  +=       0.652 * cos(2.84744627046  + 99.92819542829 *T)
    sumY0  +=       0.628 * cos(3.17606592131  + 3693.85347914410 *T)
    sumY0  +=       0.810 * cos(4.60918284532  + 721.40560204669 *T)
    sumY0  +=       0.854 * cos(4.61209992426  + 1366.45638977370 *T)
    sumY0  +=       0.867 * cos(4.19294864410  + 10846.31310300770 *T)
    sumY0  +=       0.788 * cos(1.41713341612  + 1040.97886544099 *T)
    sumY0  +=       0.626 * cos(5.79379273730  + 841.90686400460 *T)
    sumY0  +=       0.630 * cos(0.46054027581  + 291.76902978489 *T)
    sumY0  +=       0.650 * cos(1.16669181964  + 558.38897058130 *T)
    sumY0  +=       0.611 * cos(1.65671418659  + 4443.95708667750 *T)
    sumY0  +=       0.726 * cos(0.52368203745  + 2151.25100677631 *T)
    sumY0  +=       0.751 * cos(1.10181429478  + 446.06752933469 *T)
    sumY0  +=       0.615 * cos(0.87878140238  + 208.38941150850 *T)
    sumY0  +=       0.745 * cos(1.03884888390  + 451.97179027311 *T)
    sumY0  +=       0.614 * cos(2.12015186904  + 6121.89566217830 *T)
    sumY0  +=       0.603 * cos(1.87255401888  + 181.56270856549 *T)
    sumY0  +=       0.616 * cos(2.28876027875  + 1144.07163965960 *T)
    sumY0  +=       0.636 * cos(2.38750125211  + 1751.29571393250 *T)
    sumY0  +=       0.596 * cos(1.00492203164  + 737.60480372361 *T)
    sumY0  +=       0.619 * cos(2.80073636220  + 326.19691469570 *T)
    sumY0  +=       0.787 * cos(2.60401736226  + 1802.12817323829 *T)
    sumY0  +=       0.781 * cos(2.22467267459  + 88.35873817511 *T)
    sumY0  +=       0.617 * cos(2.17402643196  + 921.10825080200 *T)
    sumY0  +=       0.591 * cos(3.73213502799  + 1166.16304018740 *T)
    sumY0  +=       0.590 * cos(1.05361074587  + 3634.37720703489 *T)
    sumY0  +=       0.722 * cos(1.40579320968  + 1791.57010794140 *T)
    sumY0  +=       0.820 * cos(2.67631464422  + 1372.08546406621 *T)
    sumY0  +=       0.590 * cos(2.98691366177  + 627.61093082531 *T)
    sumY0  +=       0.611 * cos(0.43380344797  + 392.90175841570 *T)
    sumY0  +=       0.611 * cos(0.76528954681  + 666.96780674051 *T)
    sumY0  +=       0.584 * cos(4.14039377030  + 2737.56432317350 *T)
    sumY0  +=       0.624 * cos(0.20438602023  + 702.18008107150 *T)
    sumY0  +=       0.600 * cos(5.10974978259  + 203.49405039889 *T)
    sumY0  +=       0.720 * cos(1.60831618587  + 3333.74269718250 *T)
    sumY0  +=       0.715 * cos(4.50656990635  + 3347.96979118411 *T)
    sumY0  +=       0.604 * cos(5.25311127639  + 427.36327322130 *T)
    sumY0  +=       0.578 * cos(5.39623231304  + 881.75137488680 *T)
    sumY0  +=       0.752 * cos(3.84564478305  + 799.36859435169 *T)
    sumY0  +=       0.581 * cos(3.86115308052  + 2954.04481234920 *T)
    sumY0  +=       0.754 * cos(2.66282639252  + 1022.97718474320 *T)
    sumY0  +=       0.772 * cos(1.65864906434  + 711.74130862710 *T)
    sumY0  +=       0.577 * cos(5.47810665105  + 700.20797339620 *T)
    sumY0  +=       0.736 * cos(0.60214952186  + 458.12212942741 *T)
    sumY0  +=       0.736 * cos(0.59694347292  + 601.74743572880 *T)
    sumY0  +=       0.693 * cos(3.40638930298  + 1967.96000801901 *T)
    sumY0  +=       0.740 * cos(3.80855554417  + 416.79314583689 *T)
    sumY0  +=       0.704 * cos(1.18207892692  + 425.35753565121 *T)
    sumY0  +=       0.708 * cos(3.02675954280  + 3391.68870348911 *T)
    sumY0  +=       0.607 * cos(1.06404276153  + 105.76971804189 *T)
    sumY0  +=       0.559 * cos(2.48504877569  + 20426.81490990550 *T)
    sumY0  +=       0.628 * cos(4.57438103298  + 1809.72935520611 *T)
    sumY0  +=       0.696 * cos(1.91024083106  + 378.88711273520 *T)
    sumY0  +=       0.554 * cos(2.66964046417  + 386.00065973601 *T)
    sumY0  +=       0.589 * cos(4.41883744785  + 3273.88120030190 *T)
    sumY0  +=       0.633 * cos(0.34663077486  + 131.16013238639 *T)
    sumY0  +=       0.562 * cos(4.54266444163  + 4803.96545844350 *T)
    sumY0  +=       0.560 * cos(3.13968869190  + 421.21293406700 *T)
    sumY0  +=       0.580 * cos(5.96191438656  + 5179.83360020930 *T)
    sumY0  +=       0.740 * cos(4.24815360369  + 913.50706883420 *T)
    sumY0  +=       0.699 * cos(5.94463211291  + 2627.35800195410 *T)
    sumY0  +=       0.699 * cos(2.80303945932  + 2626.87036698710 *T)
    sumY0  +=       0.747 * cos(3.30207127034  + 90.31878376290 *T)
    sumY0  +=       0.575 * cos(6.12011876370  + 1656.98476433380 *T)
    sumY0  +=       0.555 * cos(0.89361846240  + 6702.80431135010 *T)
    sumY0  +=       0.559 * cos(3.70914713824  + 388.96960515260 *T)
    sumY0  +=       0.672 * cos(2.38005250851  + 2417.24030025381 *T)
    sumY0  +=       0.606 * cos(3.82620569675  + 3693.36584417709 *T)
    sumY0  +=       0.538 * cos(2.49928826237  + 7564.58690325449 *T)
    sumY0  +=       0.539 * cos(4.29776010398  + 828.16740497000 *T)
    sumY0  +=       0.548 * cos(2.14286023475  + 4700.87268422489 *T)
    sumY0  +=       0.599 * cos(3.48576134992  + 108.47803102759 *T)
    sumY0  +=       0.550 * cos(2.51535265264  + 3046.35501086230 *T)
    sumY0  +=       0.652 * cos(5.01250339010  + 282.20780222929 *T)
    sumY0  +=       0.586 * cos(0.16653899616  + 878.08234618209 *T)
    sumY0  +=       0.622 * cos(3.01686451694  + 568.80402352629 *T)
    sumY0  +=       0.728 * cos(5.63080764277  + 344.45922782439 *T)
    sumY0  +=       0.544 * cos(2.31769965190  + 10633.01400756970 *T)
    sumY0  +=       0.657 * cos(0.62701478688  + 464.97504399731 *T)
    sumY0  +=       0.546 * cos(1.66601868578  + 4112.85048805230 *T)
    sumY0  +=       0.541 * cos(4.47410178003  + 121.00820399980 *T)
    sumY0  +=       0.695 * cos(5.90409812875  + 421.68850694650 *T)
    sumY0  +=       0.601 * cos(4.73252889479  + 426.32074349770 *T)
    sumY0  +=       0.525 * cos(4.24523707483  + 76674.39271195528 *T)
    sumY0  +=       0.532 * cos(5.96635649363  + 8734.17515466309 *T)
    sumY0  +=       0.622 * cos(5.60046914442  + 1091.38144162010 *T)
    sumY0  +=       0.553 * cos(0.95137962286  + 604.22874617840 *T)
    sumY0  +=       0.633 * cos(5.27242003059  + 2274.30301515300 *T)
    sumY0  +=       0.564 * cos(0.16497975863  + 241.85408857280 *T)
    sumY0  +=       0.516 * cos(4.97603038853  + 5599.31824408450 *T)
    sumY0  +=       0.645 * cos(6.01896827268  + 10110.43658947590 *T)
    sumY0  +=       0.509 * cos(3.68808359097  + 1159.04949318660 *T)
    sumY0  +=       0.507 * cos(3.50264799357  + 615.76208140930 *T)
    sumY0  +=       0.599 * cos(2.57801707543  + 185.23173727020 *T)
    sumY0  +=       0.522 * cos(5.63448152934  + 1884.36794142151 *T)
    sumY0  +=       0.520 * cos(3.87463188031  + 1457.28211557850 *T)
    sumY0  +=       0.527 * cos(1.69611452147  + 1512.56300652469 *T)
    sumY0  +=       0.503 * cos(4.96282947219  + 1037.79747170330 *T)
    sumY0  +=       0.517 * cos(3.70527239794  + 6018.80288795970 *T)
    sumY0  +=       0.613 * cos(0.15991201142  + 607.89777488310 *T)
    sumY0  +=       0.601 * cos(0.38470330699  + 700.69560836320 *T)
    sumY0  +=       0.505 * cos(1.61427941704  + 5908.59656674030 *T)
    sumY0  +=       0.679 * cos(0.99986935446  + 1357.37073509760 *T)
    sumY0  +=       0.496 * cos(3.33323961426  + 1048.84317196720 *T)
    sumY0  +=       0.602 * cos(5.52925928036  + 680.98245242101 *T)
    sumY0  +=       0.621 * cos(4.68680597499  + 1560.37488979710 *T)
    sumY0  +=       0.623 * cos(1.32300241317  + 120.01136625851 *T)
    sumY0  +=       0.494 * cos(5.73456422193  + 523.57199510270 *T)
    sumY0  +=       0.494 * cos(4.88930673362  + 536.29757005351 *T)
    sumY0  +=       0.500 * cos(4.78996332607  + 2840.65709739210 *T)
    sumY0  +=       0.657 * cos(0.40007659336  + 2351.05672139560 *T)
    sumY0  +=       0.587 * cos(0.46191357824  + 215.50295850929 *T)
    sumY0  +=       0.609 * cos(4.14432847094  + 614.86830953200 *T)
    sumY0  +=       0.609 * cos(4.14432847094  + 614.38067456499 *T)
    sumY0  +=       0.502 * cos(0.25251420069  + 1379.46213562540 *T)
    sumY0  +=       0.492 * cos(3.31162640299  + 419.28901050149 *T)
    sumY0  +=       0.485 * cos(1.66419456691  + 1937.95191133830 *T)
    sumY0  +=       0.484 * cos(0.12315312076  + 102.09819720970 *T)
    sumY0  +=       0.491 * cos(0.11387232043  + 419.19264228190 *T)
    sumY0  +=       0.636 * cos(5.51773112302  + 664.52012618570 *T)
    sumY0  +=       0.478 * cos(4.25859285582  + 204.97852310719 *T)
    sumY0  +=       0.494 * cos(2.92886717648  + 2758.90496417590 *T)
    sumY0  +=       0.495 * cos(1.58073941957  + 398.53083270820 *T)
    sumY0  +=       0.483 * cos(0.96081432259  + 89.10949770051 *T)
    sumY0  +=       0.589 * cos(0.26853478837  + 2740.74571691120 *T)
    sumY0  +=       0.600 * cos(0.00177768827  + 634.51202950500 *T)
    sumY0  +=       0.486 * cos(3.47387686711  + 2203.94120481581 *T)
    sumY0  +=       0.515 * cos(4.56583984880  + 1091.11831706170 *T)
    sumY0  +=       0.615 * cos(5.82613512558  + 3700.96702614491 *T)
    sumY0  +=       0.490 * cos(3.48488181967  + 8631.08238044450 *T)
    sumY0  +=       0.494 * cos(6.23491155272  + 984.35651413839 *T)
    sumY0  +=       0.475 * cos(3.84857388040  + 103.59971626049 *T)
    sumY0  +=       0.475 * cos(4.16103129130  + 835.54407652920 *T)
    sumY0  +=       0.512 * cos(0.43877261458  + 397.90018538930 *T)
    sumY0  +=       0.655 * cos(3.37451455932  + 127.22797912329 *T)
    sumY0  +=       0.468 * cos(5.66081615363  + 582.62498061761 *T)
    sumY0  +=       0.468 * cos(1.82146214833  + 477.24458453860 *T)
    sumY0  +=       0.543 * cos(1.20297336764  + 481.17673780170 *T)
    sumY0  +=       0.543 * cos(6.27930493431  + 578.69282735451 *T)
    sumY0  +=       0.517 * cos(2.43390627861  + 76.02225379210 *T)
    sumY0  +=       0.573 * cos(6.09163052699  + 1132.92287353170 *T)
    sumY0  +=       0.649 * cos(5.08795164761  + 3184.60938317460 *T)
    sumY0  +=       0.480 * cos(2.95348186297  + 2133.23477186350 *T)
    sumY0  +=       0.468 * cos(0.49074178268  + 2048.15823255770 *T)
    sumY0  +=       0.594 * cos(1.96853371116  + 390.92965074040 *T)
    sumY0  +=       0.459 * cos(3.19244771660  + 937.88609122240 *T)
    sumY0  +=       0.481 * cos(0.56878471280  + 1443.68566889580 *T)
    sumY0  +=       0.463 * cos(4.21398380834  + 429.76276966630 *T)
    sumY0  +=       0.545 * cos(4.87328765120  + 3157.04896704871 *T)
    sumY0  +=       0.475 * cos(3.93834590521  + 3907.15257458210 *T)
    sumY0  +=       0.609 * cos(3.24848207205  + 1595.53648789080 *T)
    sumY0  +=       0.473 * cos(4.69599627940  + 733.67265046051 *T)
    sumY0  +=       0.541 * cos(5.85990514227  + 52176.05010063190 *T)
    sumY0  +=       0.533 * cos(1.27151481340  + 382.85289589309 *T)
    sumY0  +=       0.534 * cos(2.70958907609  + 843.87897167990 *T)
    sumY0  +=       0.451 * cos(3.99705226028  + 5371.30441967789 *T)
    sumY0  +=       0.471 * cos(2.88657776376  + 2572.43257168090 *T)
    sumY0  +=       0.558 * cos(0.57207825570  + 594.89452115890 *T)
    sumY0  +=       0.588 * cos(0.10678640907  + 458.33457774851 *T)
    sumY0  +=       0.482 * cos(2.66167846082  + 6114.78211517750 *T)
    sumY0  +=       0.449 * cos(6.27428540227  + 1221.60474883790 *T)
    sumY0  +=       0.447 * cos(3.64756714520  + 584.26951202031 *T)
    sumY0  +=       0.447 * cos(3.83471115676  + 475.60005313590 *T)
    sumY0  +=       0.466 * cos(2.22916713968  + 632.50629193491 *T)
    sumY0  +=       0.624 * cos(1.32472646328  + 10529.92123335110 *T)
    sumY0  +=       0.605 * cos(3.46941164072  + 300.85468446100 *T)
    sumY0  +=       0.487 * cos(0.70356343888  + 321.30155358610 *T)
    sumY0  +=       0.487 * cos(1.71799868416  + 206.90493880020 *T)
    sumY0  +=       0.526 * cos(0.02936914645  + 1336.66074141410 *T)
    sumY0  +=       0.450 * cos(2.88322597733  + 518.62844988331 *T)
    sumY0  +=       0.450 * cos(1.45745967104  + 541.24111527290 *T)
    sumY0  +=       0.452 * cos(4.02637396486  + 252.41215386970 *T)
    sumY0  +=       0.569 * cos(1.35475382052  + 207.42620366200 *T)
    sumY0  +=       0.442 * cos(2.28319191504  + 10976.83498721549 *T)
    sumY0  +=       0.435 * cos(3.61536948260  + 218.47190392590 *T)
    sumY0  +=       0.536 * cos(5.87308650386  + 1781.27516720290 *T)
    sumY0  +=       0.606 * cos(6.18675939821  + 6599.71153713150 *T)
    sumY0  +=       0.430 * cos(4.70071691496  + 2535.31092799940 *T)
    sumY0  +=       0.465 * cos(2.19694664935  + 1084.00477006090 *T)
    sumY0  +=       0.528 * cos(4.80779932075  + 557.75832326239 *T)
    sumY0  +=       0.602 * cos(0.32017022594  + 354.76872277791 *T)
    sumY0  +=       0.602 * cos(0.87892276884  + 705.10084237830 *T)
    sumY0  +=       0.579 * cos(0.48304329741  + 462.05428269051 *T)
    sumY0  +=       0.579 * cos(0.71604969737  + 597.81528246570 *T)
    sumY0  +=       0.480 * cos(3.58335127463  + 1106.09617058870 *T)
    sumY0  +=       0.561 * cos(5.19314855697  + 1859.23815950790 *T)
    sumY0  +=       0.565 * cos(4.77820464913  + 531.89482816590 *T)
    sumY0  +=       0.593 * cos(2.30529340153  + 1291.67479120640 *T)
    sumY0  +=       0.593 * cos(5.44688605512  + 1291.18715623940 *T)
    sumY0  +=       0.467 * cos(3.40834000086  + 2833.05591542430 *T)
    sumY0  +=       0.423 * cos(4.74881408933  + 799.85622931870 *T)
    sumY0  +=       0.591 * cos(4.03092476406  + 1819.39364862570 *T)
    sumY0  +=       0.442 * cos(5.80146904507  + 1069.18697522831 *T)
    sumY0  +=       0.538 * cos(2.60007571681  + 1065.35788782919 *T)
    sumY0  +=       0.472 * cos(3.85236778376  + 460.29462333630 *T)
    sumY0  +=       0.432 * cos(4.43898680981  + 2080.87464222410 *T)
    sumY0  +=       0.415 * cos(1.39162146625  + 4760.34895633411 *T)
    sumY0  +=       0.455 * cos(1.46286831547  + 2527.44662147320 *T)
    sumY0  +=       0.506 * cos(1.93326787770  + 535.66692273460 *T)
    sumY0  +=       0.408 * cos(4.65095351213  + 237.43430034270 *T)
    sumY0  +=       0.558 * cos(5.74240390212  + 2348.08777597900 *T)
    sumY0  +=       0.506 * cos(3.68273316284  + 2424.35384725460 *T)
    sumY0  +=       0.496 * cos(4.48964897087  + 4120.45167002010 *T)
    sumY0  +=       0.496 * cos(4.48964897088  + 4119.96403505309 *T)
    sumY0  +=       0.432 * cos(6.00820922222  + 1301.23601876200 *T)
    sumY0  +=       0.550 * cos(0.72236210742  + 8837.26792888169 *T)
    sumY0  +=       0.546 * cos(1.74052459203  + 761.98382611181 *T)
    sumY0  +=       0.546 * cos(1.74052459203  + 761.49619114479 *T)
    sumY0  +=       0.435 * cos(1.40434697094  + 205.89354684390 *T)
    sumY0  +=       0.463 * cos(2.74974947852  + 633.54882165850 *T)
    sumY0  +=       0.408 * cos(3.36326895916  + 1482.55490984399 *T)
    sumY0  +=       0.483 * cos(5.05507740192  + 2879.26321378190 *T)
    sumY0  +=       0.511 * cos(5.17393230173  + 1670.58121101650 *T)
    sumY0  +=       0.421 * cos(4.17496125106  + 638.65663108921 *T)
    sumY0  +=       0.512 * cos(6.20542375764  + 1474.42997088690 *T)
    sumY0  +=       0.388 * cos(4.23664234872  + 2425.10460678000 *T)
    sumY0  +=       0.483 * cos(0.89849999521  + 154.05468206630 *T)
    sumY0  +=       0.412 * cos(2.38738008672  + 83.62343575990 *T)
    sumY0  +=       0.384 * cos(4.81263785100  + 673.86890542020 *T)
    sumY0  +=       0.400 * cos(1.20395052136  + 2755.72357043820 *T)
    sumY0  +=       0.500 * cos(5.48455845139  + 1973.37663399041 *T)
    sumY0  +=       0.380 * cos(4.43228458999  + 3583.64715792471 *T)
    sumY0  +=       0.509 * cos(1.52055177900  + 5856.23384163189 *T)
    sumY0  +=       0.392 * cos(1.84630187584  + 6328.08121061550 *T)
    sumY0  +=       0.391 * cos(3.77313110506  + 670.89996000361 *T)
    sumY0  +=       0.397 * cos(1.19492848989  + 50586.48957038109 *T)
    sumY0  +=       0.468 * cos(4.28211685631  + 1970.92895343560 *T)
    sumY0  +=       0.491 * cos(4.48946441703  + 1288.70584578980 *T)
    sumY0  +=       0.491 * cos(4.48946441703  + 1288.21821082279 *T)
    sumY0  +=       0.472 * cos(0.93312891407  + 416.53251340600 *T)
    sumY0  +=       0.491 * cos(1.09230657741  + 601.53498740770 *T)
    sumY0  +=       0.476 * cos(1.86175772015  + 2015.42726867629 *T)

    # JUPITER Y Coord
    sumY1  = 0
    sumY1  +=  1912556.490 * cos(2.66195491149  + 0.24381748350 *T)
    sumY1  +=  634902.259 * cos(4.81945405671  + 1059.62574767270 *T)
    sumY1  +=  600483.021 * cos(0.85860311815  + 522.82123557730 *T)
    sumY1  +=  589005.176 * cos(0.34484971445  + 537.04832957890 *T)
    sumY1  +=   81432.741 * cos(1.90251633780  + 7.35736448430 *T)
    sumY1  +=   46191.617 * cos(5.16976189642  + 1589.31671276730 *T)
    sumY1  +=   37242.394 * cos(1.81456616256  + 529.93478257810 *T)
    sumY1  +=   32508.552 * cos(0.17601327437  + 1052.51220067191 *T)
    sumY1  +=   31231.184 * cos(0.77622354337  + 1066.73929467350 *T)
    sumY1  +=   25950.740 * cos(0.81541881282  + 529.44714761109 *T)
    sumY1  +=   20232.242 * cos(2.89760937255  + 110.45013870291 *T)
    sumY1  +=   17941.640 * cos(2.45019898255  + 426.84200835950 *T)
    sumY1  +=   13643.757 * cos(5.01666809262  + 633.02755679670 *T)
    sumY1  +=   12962.263 * cos(0.99688324083  + 515.70768857651 *T)
    sumY1  +=    9123.979 * cos(0.21013091394  + 544.16187657970 *T)
    sumY1  +=    8098.978 * cos(4.58582444610  + 949.41942645330 *T)
    sumY1  +=    7143.102 * cos(1.60287874390  + 323.74923414091 *T)
    sumY1  +=    4115.125 * cos(6.25575758842  + 205.94173095370 *T)
    sumY1  +=    4370.464 * cos(5.85779443169  + 736.12033101530 *T)
    sumY1  +=    3400.571 * cos(1.33020584264  + 526.75338884040 *T)
    sumY1  +=    3332.481 * cos(6.14941384034  + 533.11617631580 *T)
    sumY1  +=    3869.655 * cos(1.76724210080  + 14.47091148511 *T)
    sumY1  +=    3043.986 * cos(0.08336634740  + 526.00262931501 *T)
    sumY1  +=    3013.156 * cos(1.02614271964  + 419.72846135871 *T)
    sumY1  +=    3001.886 * cos(1.11282294120  + 533.86693584120 *T)
    sumY1  +=    2892.493 * cos(3.33842613324  + 103.33659170210 *T)
    sumY1  +=    2393.125 * cos(2.05143779600  + 316.63568714010 *T)
    sumY1  +=    2989.944 * cos(5.51952313334  + 2119.00767786191 *T)
    sumY1  +=    2354.966 * cos(4.86075556884  + 109.96250373590 *T)
    sumY1  +=    1853.775 * cos(1.19886684605  + 1596.43025976811 *T)
    sumY1  +=    1972.856 * cos(3.51875485226  + 956.53297345411 *T)
    sumY1  +=    1820.762 * cos(5.38126290712  + 743.23387801611 *T)
    sumY1  +=    1810.752 * cos(5.40985497363  + 95.73540973430 *T)
    sumY1  +=    1531.701 * cos(2.74785652806  + 117.56368570370 *T)
    sumY1  +=    1894.874 * cos(0.29375694918  + 1582.20316576650 *T)
    sumY1  +=    1527.752 * cos(6.20401691579  + 640.14110379750 *T)
    sumY1  +=    1634.391 * cos(6.11004516268  + 1045.39865367110 *T)
    sumY1  +=    1088.512 * cos(6.22920436039  + 625.91400979590 *T)
    sumY1  +=    1025.988 * cos(1.00202485870  + 433.95555536030 *T)
    sumY1  +=    1045.564 * cos(0.95995839150  + 1265.81129610991 *T)
    sumY1  +=     977.087 * cos(0.62771509387  + 220.65645992230 *T)
    sumY1  +=     984.636 * cos(6.14139798618  + 1169.83206889211 *T)
    sumY1  +=     805.130 * cos(4.73126707725  + 942.30587945250 *T)
    sumY1  +=     821.829 * cos(0.95222022801  + 309.03450517230 *T)
    sumY1  +=     798.891 * cos(0.28589806015  + 1155.60497489050 *T)
    sumY1  +=     834.956 * cos(0.44929628324  + 522.33360061030 *T)
    sumY1  +=     600.935 * cos(3.92607423432  + 6.86972951729 *T)
    sumY1  +=     752.130 * cos(4.84106440602  + 1162.71852189130 *T)
    sumY1  +=     539.923 * cos(2.67204015773  + 853.44019923550 *T)
    sumY1  +=     599.196 * cos(5.78978587742  + 213.05527795450 *T)
    sumY1  +=     497.018 * cos(0.10277515742  + 198.82818395289 *T)
    sumY1  +=     476.930 * cos(4.32981816387  + 1692.40948698591 *T)
    sumY1  +=     499.703 * cos(1.45009748987  + 330.86278114170 *T)
    sumY1  +=     383.795 * cos(0.60109464211  + 1073.85284167431 *T)
    sumY1  +=     504.945 * cos(1.16897764893  + 1059.13811270570 *T)
    sumY1  +=     402.231 * cos(4.64005130261  + 1479.11039154791 *T)
    sumY1  +=     373.538 * cos(6.00201753992  + 729.00678401450 *T)
    sumY1  +=     349.467 * cos(4.67892334070  + 426.35437339250 *T)
    sumY1  +=     316.847 * cos(0.54679906466  + 839.21310523390 *T)
    sumY1  +=     324.562 * cos(4.14050846767  + 206.42936592071 *T)
    sumY1  +=     246.849 * cos(5.67682031680  + 838.72547026689 *T)
    sumY1  +=     255.191 * cos(5.19686775170  + 412.12727939089 *T)
    sumY1  +=     230.121 * cos(1.15174981889  + 508.59414157570 *T)
    sumY1  +=     246.851 * cos(0.17931647320  + 1272.92484311070 *T)
    sumY1  +=     222.270 * cos(1.13623310093  + 412.61491435790 *T)
    sumY1  +=     213.383 * cos(1.33041208002  + 1376.01761732930 *T)
    sumY1  +=     225.567 * cos(0.71754067098  + 625.42637482889 *T)
    sumY1  +=     212.049 * cos(1.46978417946  + 3.42521122121 *T)
    sumY1  +=     181.234 * cos(2.07915113995  + 1368.90407032850 *T)
    sumY1  +=     174.181 * cos(0.28202074120  + 1062.80714141041 *T)
    sumY1  +=     189.391 * cos(5.86988015498  + 2648.69864295650 *T)
    sumY1  +=     161.119 * cos(1.45320059873  + 519.63984183961 *T)
    sumY1  +=     157.874 * cos(3.21673827019  + 536.56069461190 *T)
    sumY1  +=     145.420 * cos(1.14766744266  + 1258.69774910910 *T)
    sumY1  +=     144.205 * cos(0.51347253517  + 1055.69359440960 *T)
    sumY1  +=     143.507 * cos(2.13497084119  + 309.52214013930 *T)
    sumY1  +=     126.188 * cos(1.74614874146  + 1063.55790093581 *T)
    sumY1  +=     124.788 * cos(0.04973007408  + 551.27542358051 *T)
    sumY1  +=     125.329 * cos(6.02678896114  + 540.22972331660 *T)
    sumY1  +=     146.669 * cos(1.15255926772  + 1056.44435393500 *T)
    sumY1  +=     115.893 * cos(1.61083044890  + 2126.12122486270 *T)
    sumY1  +=     124.485 * cos(1.74552486835  + 3.68833577960 *T)
    sumY1  +=     118.218 * cos(0.09316615671  + 515.22005360950 *T)
    sumY1  +=     114.208 * cos(0.48741044761  + 227.77000692311 *T)
    sumY1  +=     113.852 * cos(4.27325397298  + 742.74624304909 *T)
    sumY1  +=     144.447 * cos(4.51237423492  + 1486.22393854870 *T)
    sumY1  +=     124.849 * cos(0.06146296380  + 1575.08961876571 *T)
    sumY1  +=      93.951 * cos(5.42064898050  + 1155.11733992350 *T)
    sumY1  +=     100.599 * cos(0.45008887739  + 2111.89413086110 *T)
    sumY1  +=     103.583 * cos(0.35190071477  + 419.24082639170 *T)
    sumY1  +=      92.556 * cos(0.28828088487  + 1699.52303398670 *T)
    sumY1  +=      99.479 * cos(1.07544165521  + 301.92095817150 *T)
    sumY1  +=      82.832 * cos(6.00125174633  + 441.06910236111 *T)
    sumY1  +=      86.688 * cos(0.38035962497  + 941.81824448550 *T)
    sumY1  +=      95.969 * cos(3.69584152771  + 102.84895673509 *T)
    sumY1  +=      80.324 * cos(5.59950981116  + 1471.99684454710 *T)
    sumY1  +=     102.333 * cos(1.61778028756  + 21.58445848590 *T)
    sumY1  +=      78.195 * cos(3.52242385121  + 525.26891613211 *T)
    sumY1  +=      74.123 * cos(0.00610250389  + 647.25465079831 *T)
    sumY1  +=      79.270 * cos(4.84421399324  + 728.51914904750 *T)
    sumY1  +=      74.807 * cos(3.95788072941  + 534.60064902410 *T)
    sumY1  +=      87.759 * cos(5.73493623659  + 632.53992182970 *T)
    sumY1  +=      69.800 * cos(1.23323877527  + 618.80046279510 *T)
    sumY1  +=      60.886 * cos(2.49979217695  + 846.32665223470 *T)
    sumY1  +=      58.777 * cos(0.68694828546  + 832.09955823310 *T)
    sumY1  +=      61.003 * cos(4.67098980984  + 2222.10045208050 *T)
    sumY1  +=      66.693 * cos(0.71208583440  + 1685.29593998510 *T)
    sumY1  +=      50.999 * cos(4.21005408384  + 2008.80135664250 *T)
    sumY1  +=      50.585 * cos(3.18331095946  + 213.54291292150 *T)
    sumY1  +=      45.850 * cos(5.81161674510  + 831.61192326610 *T)
    sumY1  +=      52.339 * cos(6.09582232771  + 1038.28510667030 *T)
    sumY1  +=      46.409 * cos(0.92753148155  + 524.51815660670 *T)
    sumY1  +=      49.446 * cos(1.18421196454  + 963.64652045490 *T)
    sumY1  +=      45.782 * cos(0.27156151413  + 535.35140854951 *T)
    sumY1  +=      61.385 * cos(1.06154014565  + 1795.50226120450 *T)
    sumY1  +=      58.958 * cos(4.48450787354  + 316.14805217309 *T)
    sumY1  +=      59.833 * cos(2.97158993654  + 4.17597074660 *T)
    sumY1  +=      43.525 * cos(5.11993607277  + 149.80701461810 *T)
    sumY1  +=      39.973 * cos(3.25434306386  + 1383.13116433011 *T)
    sumY1  +=      53.170 * cos(6.26998243842  + 750.34742501690 *T)
    sumY1  +=      39.316 * cos(5.15930242259  + 1471.50920958009 *T)
    sumY1  +=      40.288 * cos(3.34770952676  + 948.93179148630 *T)
    sumY1  +=      38.303 * cos(5.96851436892  + 511.77553531340 *T)
    sumY1  +=      36.523 * cos(1.51420379775  + 548.09402984280 *T)
    sumY1  +=      38.571 * cos(0.07822093186  + 1258.21011414209 *T)
    sumY1  +=      35.601 * cos(1.24662275151  + 10.53875822200 *T)
    sumY1  +=      36.868 * cos(6.25234770646  + 527.01402127131 *T)
    sumY1  +=      36.603 * cos(1.23015836922  + 532.85554388490 *T)
    sumY1  +=      41.197 * cos(5.04821165837  + 845.83901726770 *T)
    sumY1  +=      30.575 * cos(2.56475167146  + 124.67723270450 *T)
    sumY1  +=      27.619 * cos(0.12568890457  + 1802.61580820530 *T)
    sumY1  +=      32.289 * cos(0.79016845152  + 618.31282782810 *T)
    sumY1  +=      28.470 * cos(4.61927587685  + 1044.91101870410 *T)
    sumY1  +=      25.452 * cos(2.91722591057  + 1265.32366114289 *T)
    sumY1  +=      27.183 * cos(1.12751765708  + 1898.59503542310 *T)
    sumY1  +=      22.512 * cos(2.01936139354  + 639.65346883050 *T)
    sumY1  +=      22.622 * cos(4.89596842874  + 935.19233245171 *T)
    sumY1  +=      24.508 * cos(2.13953774929  + 1361.79052332771 *T)
    sumY1  +=      25.260 * cos(0.24154081456  + 540.98048284201 *T)
    sumY1  +=      20.677 * cos(5.55082476196  + 1148.00379292269 *T)
    sumY1  +=      28.190 * cos(1.50681890023  + 1905.70858242391 *T)
    sumY1  +=      20.297 * cos(5.38737168791  + 1162.23088692429 *T)
    sumY1  +=      25.685 * cos(3.12279377202  + 735.63269604830 *T)
    sumY1  +=      19.769 * cos(3.96513942756  + 11.28951774740 *T)
    sumY1  +=      17.928 * cos(4.61925187754  + 98.91680347199 *T)
    sumY1  +=      17.469 * cos(4.89788003858  + 1787.90107923670 *T)
    sumY1  +=      18.703 * cos(6.07505587521  + 1574.60198379870 *T)
    sumY1  +=      18.135 * cos(2.36351314755  + 910.06255053810 *T)
    sumY1  +=      17.089 * cos(0.99046288782  + 1603.54380676890 *T)
    sumY1  +=      17.538 * cos(4.47072594413  + 525.74199688410 *T)
    sumY1  +=      17.368 * cos(3.01139967776  + 534.12756827210 *T)
    sumY1  +=      19.204 * cos(3.80241402375  + 1375.52998236229 *T)
    sumY1  +=      17.455 * cos(4.81942144274  + 76.50988875911 *T)
    sumY1  +=      16.153 * cos(5.87162041441  + 99.66756299740 *T)
    sumY1  +=      16.460 * cos(0.52556394714  + 1788.38871420371 *T)
    sumY1  +=      15.683 * cos(5.56132216427  + 2.93757625420 *T)
    sumY1  +=      16.034 * cos(0.27078811588  + 191.71463695210 *T)
    sumY1  +=      15.051 * cos(4.36060419796  + 106.03035047280 *T)
    sumY1  +=      14.892 * cos(0.20628294896  + 518.88908231420 *T)
    sumY1  +=      15.187 * cos(1.36377743780  + 1588.82907780029 *T)
    sumY1  +=      17.388 * cos(2.86695466117  + 18.40306474821 *T)
    sumY1  +=      14.796 * cos(0.49390360799  + 1049.33080693420 *T)
    sumY1  +=      15.408 * cos(2.49538144956  + 106.51798543980 *T)
    sumY1  +=      17.378 * cos(2.04232991369  + 88.62186273350 *T)
    sumY1  +=      13.474 * cos(2.61179172644  + 1581.71553079950 *T)
    sumY1  +=      13.574 * cos(5.60777906786  + 106.78110999820 *T)
    sumY1  +=      17.733 * cos(1.96265517830  + 860.55374623631 *T)
    sumY1  +=      14.099 * cos(0.49552041586  + 934.70469748469 *T)
    sumY1  +=      14.690 * cos(1.28998048526  + 337.97632814250 *T)
    sumY1  +=      11.717 * cos(6.21548749349  + 3178.38960805111 *T)
    sumY1  +=      10.988 * cos(0.56913717361  + 2214.98690507971 *T)
    sumY1  +=      10.909 * cos(1.73441772451  + 1052.02456570489 *T)
    sumY1  +=      11.465 * cos(4.40816469502  + 1361.30288836070 *T)
    sumY1  +=      10.506 * cos(0.08856023326  + 2104.78058386030 *T)
    sumY1  +=      12.808 * cos(3.66887031235  + 96.22304470130 *T)
    sumY1  +=      13.402 * cos(6.17284814260  + 721.89323701370 *T)
    sumY1  +=      10.003 * cos(5.29029166259  + 1464.39566257930 *T)
    sumY1  +=      11.029 * cos(3.18683769269  + 604.71638114540 *T)
    sumY1  +=       9.807 * cos(3.81416991065  + 107.26874496520 *T)
    sumY1  +=      11.312 * cos(2.62059931029  + 220.16882495529 *T)
    sumY1  +=      10.028 * cos(3.94766287245  + 497.69145366371 *T)
    sumY1  +=       9.515 * cos(5.77311274657  + 1890.99385345530 *T)
    sumY1  +=      10.988 * cos(5.26272329530  + 2015.91490364330 *T)
    sumY1  +=      10.134 * cos(0.68933796308  + 1592.49810650500 *T)
    sumY1  +=       8.954 * cos(0.87046855862  + 10.80188278040 *T)
    sumY1  +=       8.739 * cos(4.67223251717  + 91.80325647120 *T)
    sumY1  +=       8.399 * cos(0.82244069816  + 1148.49142788971 *T)
    sumY1  +=       8.408 * cos(0.96591596781  + 1585.38455950421 *T)
    sumY1  +=       9.230 * cos(3.53443422146  + 562.17811149250 *T)
    sumY1  +=       8.186 * cos(4.64550130925  + 2104.29294889330 *T)
    sumY1  +=       8.259 * cos(2.22042436652  + 302.40859313851 *T)
    sumY1  +=       9.900 * cos(4.65877537869  + 1478.62275658089 *T)
    sumY1  +=      10.878 * cos(2.27095143786  + 113.63153244060 *T)
    sumY1  +=       9.582 * cos(4.37932327117  + 455.15318401080 *T)
    sumY1  +=       7.712 * cos(1.30412279702  + 422.90985509640 *T)
    sumY1  +=       7.595 * cos(0.02631225673  + 629.84616305900 *T)
    sumY1  +=       8.496 * cos(1.04067280887  + 224.58861318540 *T)
    sumY1  +=       7.414 * cos(3.60338167449  + 39.11305843170 *T)
    sumY1  +=       7.292 * cos(1.12702623915  + 430.02340209721 *T)
    sumY1  +=       7.357 * cos(2.07742539313  + 2655.81218995730 *T)
    sumY1  +=       7.633 * cos(6.03508479564  + 636.95971005981 *T)
    sumY1  +=       7.369 * cos(0.71596086110  + 2229.21399908130 *T)
    sumY1  +=       7.258 * cos(2.51814892716  + 423.66061462181 *T)
    sumY1  +=       7.648 * cos(0.75719220270  + 2001.68780964171 *T)
    sumY1  +=       8.081 * cos(1.31219730813  + 1586.13531902961 *T)
    sumY1  +=       7.334 * cos(3.57362648170  + 529.18402305270 *T)
    sumY1  +=       7.334 * cos(3.90865182026  + 530.68554210351 *T)
    sumY1  +=       7.088 * cos(3.90573080708  + 13.98327651810 *T)
    sumY1  +=       6.944 * cos(1.27959676844  + 99.40443843901 *T)
    sumY1  +=       7.575 * cos(3.79999823386  + 121.49583896680 *T)
    sumY1  +=       6.848 * cos(2.35876258225  + 1898.10740045609 *T)
    sumY1  +=       6.870 * cos(5.86252805188  + 1158.78636862821 *T)
    sumY1  +=       6.772 * cos(3.50813033017  + 1048.58004740881 *T)
    sumY1  +=       7.746 * cos(3.54911383960  + 1691.92185201890 *T)
    sumY1  +=       6.468 * cos(5.54893778542  + 1165.89991562901 *T)
    sumY1  +=       6.936 * cos(4.92604014284  + 953.35157971640 *T)
    sumY1  +=       6.382 * cos(0.76063043397  + 1159.53712815361 *T)
    sumY1  +=       6.464 * cos(5.53982453249  + 113.89465699899 *T)
    sumY1  +=       6.270 * cos(6.12875427425  + 387.48513244430 *T)
    sumY1  +=       6.557 * cos(4.81011230587  + 636.20895053440 *T)
    sumY1  +=       7.405 * cos(1.84721590721  + 405.50136735710 *T)
    sumY1  +=       6.424 * cos(0.99818860411  + 312.70353387701 *T)
    sumY1  +=       6.311 * cos(2.65837067079  + 983.35967639710 *T)
    sumY1  +=       7.257 * cos(0.25978556818  + 1251.09656714130 *T)
    sumY1  +=       6.999 * cos(3.09774680314  + 323.26159917390 *T)
    sumY1  +=       5.842 * cos(0.50538293824  + 1166.65067515440 *T)
    sumY1  +=       6.084 * cos(4.51074934327  + 2538.49232173710 *T)
    sumY1  +=       5.618 * cos(0.08316456270  + 71.09326278771 *T)
    sumY1  +=       5.590 * cos(0.34453029262  + 234.88355392390 *T)
    sumY1  +=       5.829 * cos(4.88173947209  + 1041.46650040800 *T)
    sumY1  +=       6.815 * cos(6.01343806079  + 305.10235190919 *T)
    sumY1  +=       5.423 * cos(4.10655824247  + 114.38229196601 *T)
    sumY1  +=       5.434 * cos(1.33565656744  + 1251.58420210830 *T)
    sumY1  +=       5.984 * cos(1.18424710557  + 319.81708087780 *T)
    sumY1  +=       5.352 * cos(0.90568977859  + 415.79630809560 *T)
    sumY1  +=       5.826 * cos(1.12913548074  + 1891.48148842230 *T)
    sumY1  +=       5.519 * cos(2.07020279444  + 1593.24886603040 *T)
    sumY1  +=       5.189 * cos(5.63081007242  + 448.18264936190 *T)
    sumY1  +=       5.286 * cos(2.89186635254  + 505.55576018990 *T)
    sumY1  +=       5.876 * cos(2.54689905929  + 459.08533727390 *T)
    sumY1  +=       5.578 * cos(2.84516685445  + 437.88770862340 *T)
    sumY1  +=       4.953 * cos(5.01729407226  + 1780.78753223589 *T)
    sumY1  +=       4.928 * cos(3.11188481400  + 199.31581891990 *T)
    sumY1  +=       4.972 * cos(4.59041195194  + 554.31380496631 *T)
    sumY1  +=       4.889 * cos(4.57426078554  + 117.07605073670 *T)
    sumY1  +=       5.942 * cos(1.75505812286  + 512.52629483880 *T)
    sumY1  +=       4.807 * cos(0.26863334445  + 494.51005992600 *T)
    sumY1  +=       5.528 * cos(1.80622642695  + 430.77416162260 *T)
    sumY1  +=       4.806 * cos(0.21884952823  + 956.04533848709 *T)
    sumY1  +=       6.190 * cos(5.04541957128  + 2751.79141717511 *T)
    sumY1  +=       5.306 * cos(0.11969668866  + 1069.92068841121 *T)
    sumY1  +=       4.939 * cos(5.49878249966  + 2207.38572311189 *T)
    sumY1  +=       5.203 * cos(0.30883580794  + 498.91529394111 *T)
    sumY1  +=       4.397 * cos(0.93049026820  + 565.35950523021 *T)
    sumY1  +=       4.342 * cos(1.34852955452  + 672.38443271190 *T)
    sumY1  +=       4.528 * cos(4.21721444880  + 1677.69475801730 *T)
    sumY1  +=       4.334 * cos(1.86306609018  + 611.68691579430 *T)
    sumY1  +=       4.608 * cos(5.66110590863  + 679.49797971270 *T)
    sumY1  +=       4.812 * cos(0.89025719177  + 560.95427121510 *T)
    sumY1  +=       4.305 * cos(1.60213053627  + 1070.67144793660 *T)
    sumY1  +=       4.662 * cos(4.62198878885  + 621.98185653281 *T)
    sumY1  +=       4.362 * cos(2.38936273723  + 81.50831573269 *T)
    sumY1  +=       4.781 * cos(4.94273816108  + 600.78422788230 *T)
    sumY1  +=       4.148 * cos(0.34386178190  + 1080.96638867510 *T)
    sumY1  +=       4.478 * cos(5.54973564701  + 629.09540353360 *T)
    sumY1  +=       4.108 * cos(0.28050115804  + 644.07325706060 *T)
    sumY1  +=       4.654 * cos(0.26747143227  + 508.10650660869 *T)
    sumY1  +=       4.260 * cos(0.28663615026  + 209.87388421679 *T)
    sumY1  +=       5.131 * cos(5.19827010225  + 946.23803271561 *T)
    sumY1  +=       3.845 * cos(0.02687662536  + 970.76006745570 *T)
    sumY1  +=       3.839 * cos(4.39486053854  + 2420.68481854989 *T)
    sumY1  +=       3.898 * cos(4.36796507234  + 1064.29161411871 *T)
    sumY1  +=       4.274 * cos(3.88197021297  + 31.99951143090 *T)
    sumY1  +=       3.986 * cos(3.61377428858  + 952.60082019101 *T)
    sumY1  +=       3.880 * cos(3.72470409125  + 504.80500066450 *T)
    sumY1  +=       5.205 * cos(4.22425643960  + 1795.01462623750 *T)
    sumY1  +=       3.879 * cos(0.21645781655  + 747.16603127920 *T)
    sumY1  +=       3.965 * cos(6.05218850802  + 501.62360692681 *T)
    sumY1  +=       4.541 * cos(0.10865029183  + 757.46097201770 *T)
    sumY1  +=       3.663 * cos(0.96855107175  + 2318.07967929830 *T)
    sumY1  +=       4.938 * cos(1.24113099358  + 294.80741117069 *T)
    sumY1  +=       3.643 * cos(3.75757419955  + 555.06456449170 *T)
    sumY1  +=       3.850 * cos(0.10927703513  + 835.28095197081 *T)
    sumY1  +=       3.693 * cos(1.43008979314  + 558.24595822940 *T)
    sumY1  +=       3.818 * cos(3.56284581831  + 2325.19322629910 *T)
    sumY1  +=       3.422 * cos(2.09047795615  + 2214.49927011270 *T)
    sumY1  +=       3.726 * cos(0.01800308688  + 740.05248427840 *T)
    sumY1  +=       3.306 * cos(0.56515315647  + 2641.58509595570 *T)
    sumY1  +=       3.331 * cos(1.40804314394  + 501.48059457490 *T)
    sumY1  +=       3.289 * cos(3.64076064476  + 1066.25165970650 *T)
    sumY1  +=       4.224 * cos(5.65692006690  + 547.34327031741 *T)
    sumY1  +=       3.214 * cos(4.63993723371  + 498.44221318910 *T)
    sumY1  +=       3.771 * cos(1.62732535851  + 1493.33748554950 *T)
    sumY1  +=       3.357 * cos(5.63168758175  + 1272.43720814370 *T)
    sumY1  +=       3.469 * cos(1.82151805203  + 1678.18239298430 *T)
    sumY1  +=       3.147 * cos(5.15767685182  + 654.36819779910 *T)
    sumY1  +=       3.122 * cos(3.69868417138  + 1912.82212942470 *T)
    sumY1  +=       3.766 * cos(6.26541380179  + 1567.97607176490 *T)
    sumY1  +=       3.766 * cos(6.26541380179  + 1567.48843679789 *T)
    sumY1  +=       3.651 * cos(3.33863747697  + 2008.31372167550 *T)
    sumY1  +=       3.350 * cos(2.02935185038  + 6276.20612047410 *T)
    sumY1  +=       2.969 * cos(2.84234108474  + 561.42735196710 *T)
    sumY1  +=       2.926 * cos(0.87762167738  + 824.98601123230 *T)
    sumY1  +=       3.251 * cos(1.48662260788  + 6290.43321447570 *T)
    sumY1  +=       3.134 * cos(3.45182421255  + 10206.41581669370 *T)
    sumY1  +=       3.154 * cos(1.31710977614  + 483.46435966210 *T)
    sumY1  +=       2.968 * cos(0.95734389320  + 597.60283414461 *T)
    sumY1  +=       3.023 * cos(3.68954971021  + 938.37372618940 *T)
    sumY1  +=       3.496 * cos(3.80583016563  + 1054.95988122670 *T)
    sumY1  +=       2.718 * cos(5.83851172606  + 1151.67282162740 *T)
    sumY1  +=       3.041 * cos(2.90965702851  + 10220.64291069530 *T)
    sumY1  +=       2.652 * cos(6.21650349274  + 960.46512671720 *T)
    sumY1  +=       2.720 * cos(1.00932854817  + 326.93062787860 *T)
    sumY1  +=       2.672 * cos(0.80664235757  + 537.79908910430 *T)
    sumY1  +=       2.672 * cos(0.39245063721  + 522.07047605190 *T)
    sumY1  +=       2.641 * cos(0.36790627076  + 2531.37877473630 *T)
    sumY1  +=       2.658 * cos(1.14453224196  + 1062.07342822750 *T)
    sumY1  +=       2.567 * cos(4.28770215521  + 415.30867312860 *T)
    sumY1  +=       2.847 * cos(1.35125253700  + 1176.94561589290 *T)
    sumY1  +=       2.545 * cos(4.77644807792  + 1098.98262358790 *T)
    sumY1  +=       2.658 * cos(5.35693010643  + 1057.17806711790 *T)
    sumY1  +=       3.084 * cos(2.56963998299  + 320.56784040320 *T)
    sumY1  +=       3.075 * cos(6.27507572432  + 462.26673101160 *T)
    sumY1  +=       2.804 * cos(6.16567268722  + 576.40520549410 *T)
    sumY1  +=       2.606 * cos(5.21906136752  + 2523.77759276850 *T)
    sumY1  +=       3.000 * cos(3.65608242169  + 230.32075334190 *T)
    sumY1  +=       2.425 * cos(4.09514576387  + 422.42222012939 *T)
    sumY1  +=       2.483 * cos(5.26115871401  + 209.12312469139 *T)
    sumY1  +=       2.390 * cos(4.78653880453  + 210.36151918381 *T)
    sumY1  +=       2.327 * cos(5.56730422259  + 416.05943265400 *T)
    sumY1  +=       2.797 * cos(0.30753315592  + 2332.30677329990 *T)
    sumY1  +=       2.273 * cos(3.32904776610  + 945.48727319021 *T)
    sumY1  +=       2.980 * cos(0.57215995151  + 202.76033721600 *T)
    sumY1  +=       2.267 * cos(4.77704986942  + 2097.17940189250 *T)
    sumY1  +=       2.330 * cos(2.61194456361  + 849.50804597240 *T)
    sumY1  +=       2.940 * cos(3.89100366694  + 2111.40649589409 *T)
    sumY1  +=       2.403 * cos(1.43951577299  + 433.46792039329 *T)
    sumY1  +=       2.256 * cos(1.40980672956  + 1054.20912170131 *T)
    sumY1  +=       2.203 * cos(4.38408918114  + 217.47506618460 *T)
    sumY1  +=       2.189 * cos(3.11420297739  + 1514.53511420000 *T)
    sumY1  +=       2.120 * cos(5.32964021512  + 423.17297965479 *T)
    sumY1  +=       2.093 * cos(2.91578358954  + 543.67424161270 *T)
    sumY1  +=       2.373 * cos(3.40448834117  + 536.15455770160 *T)
    sumY1  +=       2.373 * cos(4.07778996079  + 523.71500745461 *T)
    sumY1  +=       2.040 * cos(1.15053228876  + 988.77630236850 *T)
    sumY1  +=       2.105 * cos(1.65182568239  + 1059.57756356290 *T)
    sumY1  +=       2.005 * cos(0.23022912919  + 460.78225830330 *T)
    sumY1  +=       1.997 * cos(0.01573125874  + 504.66198831261 *T)
    sumY1  +=       2.033 * cos(0.46615032885  + 1173.76422215520 *T)
    sumY1  +=       1.976 * cos(1.94247006692  + 46.71424039951 *T)
    sumY1  +=       2.284 * cos(0.08168640420  + 531.57931398080 *T)
    sumY1  +=       2.284 * cos(1.11740659058  + 528.29025117541 *T)
    sumY1  +=       2.219 * cos(0.28864854659  + 5230.56364931950 *T)
    sumY1  +=       2.084 * cos(3.79495631549  + 490.57790666290 *T)
    sumY1  +=       2.020 * cos(3.27527557464  + 568.06781821590 *T)
    sumY1  +=       1.939 * cos(0.63756905811  + 1065.04237364410 *T)
    sumY1  +=       1.972 * cos(1.10124465701  + 2428.28600051770 *T)
    sumY1  +=       2.080 * cos(5.31250252176  + 194.89603068979 *T)
    sumY1  +=       1.994 * cos(1.00933853256  + 749.85979004990 *T)
    sumY1  +=       1.905 * cos(1.94076118880  + 1077.78499493740 *T)
    sumY1  +=       1.909 * cos(4.29377156648  + 2737.07668820650 *T)
    sumY1  +=       1.874 * cos(1.18336173856  + 555.20757684360 *T)
    sumY1  +=       2.388 * cos(2.65748069414  + 334.79493440480 *T)
    sumY1  +=       2.111 * cos(0.80588667765  + 5216.33655531789 *T)
    sumY1  +=       2.086 * cos(1.74158554893  + 1062.54650897950 *T)
    sumY1  +=       2.063 * cos(1.71239799040  + 9160.77334553909 *T)
    sumY1  +=       1.930 * cos(0.74728387158  + 2207.87335807890 *T)
    sumY1  +=       1.877 * cos(0.19967253757  + 732.93893727761 *T)
    sumY1  +=       1.836 * cos(4.07130877335  + 1994.08662767389 *T)
    sumY1  +=       1.907 * cos(6.06525003405  + 824.49837626530 *T)
    sumY1  +=       1.819 * cos(4.11575405255  + 491.80174694030 *T)
    sumY1  +=       1.892 * cos(3.68926102357  + 569.29165849331 *T)
    sumY1  +=       1.770 * cos(2.86237801037  + 842.39449897160 *T)
    sumY1  +=       1.786 * cos(3.68832273420  + 327.68138740401 *T)
    sumY1  +=       1.761 * cos(1.39220218514  + 67.42423408300 *T)
    sumY1  +=       1.872 * cos(2.05384799126  + 1169.34443392509 *T)
    sumY1  +=       1.951 * cos(5.91395025443  + 142.69346761730 *T)
    sumY1  +=       1.737 * cos(5.04430948440  + 202.00957769060 *T)
    sumY1  +=       1.965 * cos(2.22844080314  + 9146.54625153749 *T)
    sumY1  +=       1.703 * cos(1.79561638802  + 2530.89113976930 *T)
    sumY1  +=       1.717 * cos(1.56523864144  + 380.37158544351 *T)
    sumY1  +=       1.684 * cos(0.96892569776  + 599.08730685290 *T)
    sumY1  +=       1.669 * cos(0.56251586275  + 2634.47154895491 *T)
    sumY1  +=       1.707 * cos(3.13360819218  + 92.29089143820 *T)
    sumY1  +=       1.666 * cos(4.51582119325  + 1058.66253982621 *T)
    sumY1  +=       2.249 * cos(1.89939977026  + 852.95256426850 *T)
    sumY1  +=       2.105 * cos(4.84963681002  + 1059.67393178251 *T)
    sumY1  +=       1.971 * cos(5.97456261468  + 1883.88030645450 *T)
    sumY1  +=       1.682 * cos(5.05358209402  + 1905.22094745690 *T)
    sumY1  +=       1.591 * cos(1.71258004064  + 2435.39954751850 *T)
    sumY1  +=       1.870 * cos(4.91045755022  + 739.30172475301 *T)
    sumY1  +=       1.831 * cos(0.23036450981  + 405.01373239010 *T)
    sumY1  +=       1.741 * cos(3.87703465673  + 1280.03839011150 *T)
    sumY1  +=       1.437 * cos(0.15153485897  + 1056.70498636590 *T)
    sumY1  +=       1.666 * cos(1.98564123475  + 1060.58895551920 *T)
    sumY1  +=       1.646 * cos(3.31965590226  + 299.37021175271 *T)
    sumY1  +=       1.423 * cos(5.06681929129  + 1464.88329754630 *T)
    sumY1  +=       1.879 * cos(5.78203740511  + 531.41925528640 *T)
    sumY1  +=       1.879 * cos(1.70024089685  + 528.45030986980 *T)
    sumY1  +=       1.460 * cos(1.42660733495  + 1269.74344937300 *T)
    sumY1  +=       1.684 * cos(0.90643601570  + 1364.97191706540 *T)
    sumY1  +=       1.453 * cos(1.55271946857  + 526.26575387340 *T)
    sumY1  +=       1.700 * cos(4.86312330636  + 725.07463075140 *T)
    sumY1  +=       1.360 * cos(1.93936734070  + 917.17609753890 *T)
    sumY1  +=       1.426 * cos(4.95337494898  + 2840.16946242510 *T)
    sumY1  +=       1.498 * cos(3.83211973659  + 1134.40734624000 *T)
    sumY1  +=       1.410 * cos(4.97053896796  + 1091.86907658711 *T)
    sumY1  +=       1.520 * cos(5.37749544795  + 977.73060210460 *T)
    sumY1  +=       1.747 * cos(3.17984072825  + 2324.70559133209 *T)
    sumY1  +=       1.288 * cos(3.95858951452  + 732.18817775220 *T)
    sumY1  +=       1.713 * cos(3.60652551663  + 2427.79836555070 *T)
    sumY1  +=       1.282 * cos(5.53618260525  + 1475.17823828480 *T)
    sumY1  +=       1.517 * cos(1.63411439321  + 1439.75351563271 *T)
    sumY1  +=       1.271 * cos(4.19498605517  + 843.14525849700 *T)
    sumY1  +=       1.308 * cos(5.87705324776  + 558.38897058130 *T)
    sumY1  +=       1.677 * cos(0.97487380844  + 611.19928082730 *T)
    sumY1  +=       1.256 * cos(4.41755923055  + 2854.88419139370 *T)
    sumY1  +=       1.366 * cos(3.14960813536  + 1446.86706263350 *T)
    sumY1  +=       1.435 * cos(4.97960275859  + 991.95769610620 *T)
    sumY1  +=       1.144 * cos(5.29122600369  + 1482.29178528560 *T)
    sumY1  +=       1.142 * cos(4.39466768970  + 408.19512612779 *T)
    sumY1  +=       1.147 * cos(5.28949177860  + 3068.18328683170 *T)
    sumY1  +=       1.501 * cos(5.53595039529  + 75.02541605080 *T)
    sumY1  +=       1.226 * cos(3.45697923185  + 1141.37788088890 *T)
    sumY1  +=       1.323 * cos(5.16542134486  + 284.39235822571 *T)
    sumY1  +=       1.150 * cos(5.21811270268  + 430.28652665560 *T)
    sumY1  +=       1.292 * cos(1.65611589428  + 1994.57426264090 *T)
    sumY1  +=       1.226 * cos(2.47910206722  + 372.77040347569 *T)
    sumY1  +=       1.132 * cos(3.62425240134  + 1063.81853336670 *T)
    sumY1  +=       1.079 * cos(0.26090286718  + 1483.04254481100 *T)
    sumY1  +=       1.111 * cos(4.66761037682  + 2413.57127154910 *T)
    sumY1  +=       1.252 * cos(5.77247306070  + 473.31243127551 *T)
    sumY1  +=       1.242 * cos(2.94913211429  + 539.49601013370 *T)
    sumY1  +=       1.242 * cos(1.39155353408  + 520.37355502251 *T)
    sumY1  +=       1.169 * cos(1.94705823334  + 386.99749747730 *T)
    sumY1  +=       1.003 * cos(1.46144081160  + 216.98743121760 *T)
    sumY1  +=       1.030 * cos(3.78515085968  + 850.25880549781 *T)
    sumY1  +=       1.165 * cos(1.66344587729  + 2758.90496417590 *T)
    sumY1  +=       0.984 * cos(3.96978325817  + 522.77305146750 *T)
    sumY1  +=       0.984 * cos(0.37090239019  + 537.09651368871 *T)
    sumY1  +=       1.015 * cos(5.23058664749  + 451.22103074770 *T)
    sumY1  +=       1.310 * cos(2.45586682061  + 1354.67697632690 *T)
    sumY1  +=       1.121 * cos(2.24397477114  + 4164.06817212950 *T)
    sumY1  +=       0.956 * cos(0.45045343285  + 2950.86341861150 *T)
    sumY1  +=       0.940 * cos(5.04866396759  + 39.60069339870 *T)
    sumY1  +=       0.949 * cos(0.52248659569  + 1475.92899781020 *T)
    sumY1  +=       0.992 * cos(1.30939511951  + 1969.44448072730 *T)
    sumY1  +=       0.935 * cos(3.89177610875  + 3053.46855786309 *T)
    sumY1  +=       1.085 * cos(1.70979702342  + 586.55713388070 *T)
    sumY1  +=       0.901 * cos(1.38326808068  + 248.96763557360 *T)
    sumY1  +=       0.894 * cos(1.47962768735  + 2847.28300942589 *T)
    sumY1  +=       0.948 * cos(4.80656003236  + 2221.61281711350 *T)
    sumY1  +=       0.923 * cos(0.00868868373  + 2421.17245351690 *T)
    sumY1  +=       1.070 * cos(5.69849027842  + 2200.27217611110 *T)
    sumY1  +=       1.169 * cos(2.53153449921  + 100.40127618030 *T)
    sumY1  +=       1.005 * cos(3.63845316069  + 1432.63996863190 *T)
    sumY1  +=       1.017 * cos(3.67405924039  + 8094.27786834910 *T)
    sumY1  +=       0.984 * cos(3.45627656974  + 537.00014546910 *T)
    sumY1  +=       0.984 * cos(0.88440907863  + 522.86941968710 *T)
    sumY1  +=       1.118 * cos(1.42327440603  + 416.54706762100 *T)
    sumY1  +=       0.914 * cos(5.81575314491  + 1140.89024592190 *T)
    sumY1  +=       0.977 * cos(1.34926302993  + 312.96665843540 *T)
    sumY1  +=       0.840 * cos(0.85450942681  + 1579.02177202880 *T)
    sumY1  +=       0.891 * cos(0.04963060055  + 486.64575339981 *T)
    sumY1  +=       0.849 * cos(0.94811547267  + 834.79331700380 *T)
    sumY1  +=       0.926 * cos(2.63549295580  + 532.62854134880 *T)
    sumY1  +=       0.832 * cos(2.25169167762  + 608.64853440851 *T)
    sumY1  +=       0.975 * cos(3.40649349112  + 21.09682351890 *T)
    sumY1  +=       0.794 * cos(5.26516657623  + 1791.57010794140 *T)
    sumY1  +=       1.006 * cos(1.72593750499  + 2310.96613229750 *T)
    sumY1  +=       0.825 * cos(1.82039899833  + 8616.85528644289 *T)
    sumY1  +=       0.827 * cos(0.54227509993  + 2545.60586873790 *T)
    sumY1  +=       0.955 * cos(1.09478150456  + 760.01171843649 *T)
    sumY1  +=       0.955 * cos(4.23637415815  + 760.49935340350 *T)
    sumY1  +=       0.936 * cos(1.28100839064  + 155.53915477459 *T)
    sumY1  +=       0.779 * cos(0.26183378572  + 1792.32086746680 *T)
    sumY1  +=       0.766 * cos(0.71363802237  + 3811.17334736430 *T)
    sumY1  +=       0.773 * cos(3.84885148203  + 738.81408978600 *T)
    sumY1  +=       0.788 * cos(5.43030953147  + 732.45130231059 *T)
    sumY1  +=       0.801 * cos(3.05581987419  + 77204.08367704989 *T)
    sumY1  +=       0.937 * cos(6.26646142496  + 297.98880490840 *T)
    sumY1  +=       0.779 * cos(3.79015371961  + 538.01153742540 *T)
    sumY1  +=       0.779 * cos(0.55053192876  + 521.85802773080 *T)
    sumY1  +=       0.755 * cos(2.10425922159  + 8631.08238044450 *T)
    sumY1  +=       0.743 * cos(3.49834818410  + 1060.14701253450 *T)
    sumY1  +=       0.790 * cos(3.19521603309  + 902.94900353731 *T)
    sumY1  +=       0.897 * cos(1.19433126461  + 1485.73630358170 *T)
    sumY1  +=       0.991 * cos(3.33499486125  + 2744.19023520730 *T)
    sumY1  +=       0.740 * cos(5.64369354959  + 277.27881122490 *T)
    sumY1  +=       0.780 * cos(5.38986403689  + 606.20085385370 *T)
    sumY1  +=       0.738 * cos(1.96155322417  + 4856.32818355191 *T)
    sumY1  +=       0.799 * cos(1.14946240957  + 573.22381175640 *T)
    sumY1  +=       0.802 * cos(2.35697578389  + 984.84414910540 *T)
    sumY1  +=       0.761 * cos(0.74324276764  + 2744.67787017430 *T)
    sumY1  +=       0.735 * cos(3.91525398793  + 4650.14263511470 *T)
    sumY1  +=       0.721 * cos(3.82827689917  + 731.70054278519 *T)
    sumY1  +=       0.850 * cos(2.33025907115  + 775.47720693050 *T)
    sumY1  +=       0.717 * cos(5.22051515031  + 5746.02752041250 *T)
    sumY1  +=       0.754 * cos(3.01162160350  + 202.49721265760 *T)
    sumY1  +=       0.787 * cos(1.45241015286  + 320.08020543620 *T)
    sumY1  +=       0.739 * cos(4.25049662100  + 967.57867371801 *T)
    sumY1  +=       0.744 * cos(0.24478158041  + 2847.77064439291 *T)
    sumY1  +=       0.684 * cos(5.02015802407  + 1798.68365494221 *T)
    sumY1  +=       0.700 * cos(4.52571213685  + 572.47305223100 *T)
    sumY1  +=       0.700 * cos(2.95656616511  + 487.39651292520 *T)
    sumY1  +=       0.690 * cos(1.37931038856  + 2118.52004289490 *T)
    sumY1  +=       0.670 * cos(2.76522726571  + 245.29860686890 *T)
    sumY1  +=       0.668 * cos(5.20188694033  + 739.56484931140 *T)
    sumY1  +=       0.667 * cos(5.58822887801  + 1013.15532475670 *T)
    sumY1  +=       0.762 * cos(0.14158644002  + 9153.65979853830 *T)
    sumY1  +=       0.711 * cos(4.68249485250  + 3156.56133208170 *T)
    sumY1  +=       0.679 * cos(5.25016144281  + 24498.58642880689 *T)
    sumY1  +=       0.729 * cos(4.80734453852  + 4547.04986089611 *T)
    sumY1  +=       0.692 * cos(4.62037992194  + 5062.51373198910 *T)
    sumY1  +=       0.649 * cos(6.27552774113  + 1799.43441446760 *T)
    sumY1  +=       0.677 * cos(2.25267578062  + 10.05112325500 *T)
    sumY1  +=       0.784 * cos(5.42419314811  + 3281.48238226970 *T)
    sumY1  +=       0.733 * cos(2.27993558439  + 453.66871130251 *T)
    sumY1  +=       0.643 * cos(4.14281979615  + 2310.47849733050 *T)
    sumY1  +=       0.658 * cos(1.58035282058  + 305.85311143459 *T)
    sumY1  +=       0.784 * cos(2.49180999721  + 74.53778108379 *T)
    sumY1  +=       0.704 * cos(5.13288018853  + 5165.60650620771 *T)
    sumY1  +=       0.648 * cos(0.46023761646  + 1382.64352936309 *T)
    sumY1  +=       0.786 * cos(4.88358448252  + 1290.19031849811 *T)
    sumY1  +=       0.844 * cos(6.14413478231  + 312.21589891000 *T)
    sumY1  +=       0.703 * cos(4.75268520401  + 2288.10022602790 *T)
    sumY1  +=       0.703 * cos(5.25408808776  + 2273.87313202630 *T)
    sumY1  +=       0.797 * cos(2.87597815652  + 2641.09746098870 *T)
    sumY1  +=       0.635 * cos(2.14518246830  + 397.90018538930 *T)
    sumY1  +=       0.607 * cos(1.30612000311  + 1689.22809324820 *T)
    sumY1  +=       0.626 * cos(0.39224578590  + 2524.26522773550 *T)
    sumY1  +=       0.636 * cos(4.99095278095  + 511.28790034640 *T)
    sumY1  +=       0.613 * cos(0.56214188686  + 1037.79747170330 *T)
    sumY1  +=       0.614 * cos(3.18089955415  + 2957.97696561230 *T)
    sumY1  +=       0.747 * cos(5.14974622689  + 5223.45010231870 *T)
    sumY1  +=       0.606 * cos(0.77338936404  + 1751.29571393250 *T)
    sumY1  +=       0.779 * cos(4.30366040804  + 523.78444342380 *T)
    sumY1  +=       0.779 * cos(0.03702524033  + 536.08512173241 *T)
    sumY1  +=       0.607 * cos(1.35240171145  + 2133.23477186350 *T)
    sumY1  +=       0.744 * cos(6.04795940886  + 643.32249753521 *T)
    sumY1  +=       0.697 * cos(0.66536766751  + 1354.18934135989 *T)
    sumY1  +=       0.669 * cos(3.25849716631  + 5069.62727898990 *T)
    sumY1  +=       0.762 * cos(1.79392292274  + 1262.62990237220 *T)
    sumY1  +=       0.765 * cos(0.61378577745  + 927.59115048390 *T)
    sumY1  +=       0.720 * cos(1.48952280347  + 3075.29683383250 *T)
    sumY1  +=       0.573 * cos(5.97535724360  + 394.59867944510 *T)
    sumY1  +=       0.634 * cos(6.09733077732  + 470.13103753780 *T)
    sumY1  +=       0.559 * cos(0.92745429765  + 445.00125562421 *T)
    sumY1  +=       0.671 * cos(4.72931368260  + 454.66554904380 *T)
    sumY1  +=       0.755 * cos(1.43187121540  + 105.29663728990 *T)
    sumY1  +=       0.567 * cos(5.33251485606  + 1887.54933515920 *T)
    sumY1  +=       0.593 * cos(5.99788761622  + 842.65762352999 *T)
    sumY1  +=       0.564 * cos(2.51973397119  + 1127.29379923920 *T)
    sumY1  +=       0.547 * cos(0.43054759456  + 4127.56521702090 *T)
    sumY1  +=       0.624 * cos(2.95261781250  + 25035.39094090229 *T)
    sumY1  +=       0.545 * cos(6.09691062396  + 810.90192958261 *T)
    sumY1  +=       0.641 * cos(1.20453451725  + 305.58998687620 *T)
    sumY1  +=       0.593 * cos(1.40845707752  + 1063.07026596880 *T)
    sumY1  +=       0.574 * cos(0.22531417131  + 1699.03539901970 *T)
    sumY1  +=       0.616 * cos(3.45400333076  + 25021.16384690069 *T)
    sumY1  +=       0.721 * cos(2.41192409328  + 1819.88128359270 *T)
    sumY1  +=       0.537 * cos(1.85187456995  + 782.59075393131 *T)
    sumY1  +=       0.597 * cos(4.07180274826  + 379.88395047650 *T)
    sumY1  +=       0.543 * cos(3.12698436492  + 216.72430665921 *T)
    sumY1  +=       0.552 * cos(0.25020791353  + 1268.99268984760 *T)
    sumY1  +=       0.534 * cos(6.16996002402  + 437.13694909800 *T)
    sumY1  +=       0.588 * cos(5.43198241681  + 460.29462333630 *T)
    sumY1  +=       0.524 * cos(4.47141172676  + 518.40144734719 *T)
    sumY1  +=       0.701 * cos(0.58425032979  + 1670.58121101650 *T)
    sumY1  +=       0.632 * cos(5.83508891293  + 4525.70921989370 *T)
    sumY1  +=       0.546 * cos(3.16244281187  + 89.10949770051 *T)
    sumY1  +=       0.496 * cos(4.51139575372  + 3171.27606105030 *T)
    sumY1  +=       0.501 * cos(4.75534479759  + 32.48714639791 *T)
    sumY1  +=       0.508 * cos(4.65099599522  + 827.67977000299 *T)
    sumY1  +=       0.493 * cos(3.86150271006  + 10022.08109758290 *T)
    sumY1  +=       0.550 * cos(5.31707353623  + 2516.66404576769 *T)
    sumY1  +=       0.546 * cos(1.38494758910  + 589.73852761840 *T)
    sumY1  +=       0.509 * cos(0.35849492381  + 4686.64559022330 *T)
    sumY1  +=       0.505 * cos(0.21007516833  + 754.27957828001 *T)
    sumY1  +=       0.479 * cos(3.15486072496  + 5223.93773728571 *T)
    sumY1  +=       0.461 * cos(2.56623041309  + 1595.94262480110 *T)
    sumY1  +=       0.472 * cos(0.95774392273  + 3163.67487908250 *T)
    sumY1  +=       0.472 * cos(0.95774392273  + 3164.16251404950 *T)
    sumY1  +=       0.509 * cos(5.55791627029  + 69.39634175830 *T)
    sumY1  +=       0.546 * cos(0.44435269123  + 2840.65709739210 *T)
    sumY1  +=       0.615 * cos(0.32730812757  + 519.15220687259 *T)
    sumY1  +=       0.579 * cos(4.02359849872  + 525.51499434800 *T)
    sumY1  +=       0.455 * cos(5.55680477419  + 1468.06469128401 *T)
    sumY1  +=       0.451 * cos(0.27162002670  + 614.86830953200 *T)
    sumY1  +=       0.451 * cos(4.98348360302  + 2107.96197759801 *T)
    sumY1  +=       0.450 * cos(0.83061588531  + 3480.06674873909 *T)
    sumY1  +=       0.450 * cos(0.83061588531  + 3480.55438370610 *T)
    sumY1  +=       0.447 * cos(6.20310325477  + 3267.25528826810 *T)
    sumY1  +=       0.467 * cos(0.09916917853  + 1490.15609181181 *T)
    sumY1  +=       0.472 * cos(5.54856635272  + 1457.28211557850 *T)
    sumY1  +=       0.462 * cos(5.48144983275  + 26617.83792415230 *T)
    sumY1  +=       0.440 * cos(5.31113210528  + 3384.08752152129 *T)
    sumY1  +=       0.440 * cos(5.31113210528  + 3384.57515648830 *T)
    sumY1  +=       0.453 * cos(4.59459163866  + 9154.14743350530 *T)
    sumY1  +=       0.440 * cos(4.77291542298  + 2538.00468677009 *T)
    sumY1  +=       0.616 * cos(0.11097190779  + 1560.37488979710 *T)
    sumY1  +=       0.437 * cos(3.78589316644  + 4966.53450477130 *T)
    sumY1  +=       0.433 * cos(0.72851985281  + 4700.87268422489 *T)
    sumY1  +=       0.439 * cos(0.00874316549  + 2108.71273712340 *T)
    sumY1  +=       0.452 * cos(3.24655844352  + 1551.18367712950 *T)
    sumY1  +=       0.483 * cos(0.54583867389  + 629.35852809199 *T)
    sumY1  +=       0.455 * cos(2.09561269155  + 1368.41643536150 *T)
    sumY1  +=       0.454 * cos(1.06884234808  + 1884.36794142151 *T)
    sumY1  +=       0.476 * cos(2.84297109529  + 3060.58210486390 *T)
    sumY1  +=       0.414 * cos(5.93250814925  + 533.37930087419 *T)
    sumY1  +=       0.434 * cos(4.40957688009  + 622.24498109120 *T)
    sumY1  +=       0.434 * cos(1.26798422650  + 622.73261605820 *T)
    sumY1  +=       0.423 * cos(6.09756499780  + 415.04804069769 *T)
    sumY1  +=       0.411 * cos(2.21364343296  + 4539.44867892829 *T)
    sumY1  +=       0.411 * cos(2.21364343296  + 4539.93631389530 *T)
    sumY1  +=       0.507 * cos(2.46156999040  + 3252.54055929950 *T)
    sumY1  +=       0.401 * cos(0.60477521531  + 1040.97886544099 *T)
    sumY1  +=       0.424 * cos(0.17468685146  + 1254.76559584600 *T)
    sumY1  +=       0.407 * cos(1.50695117865  + 665.27088571111 *T)
    sumY1  +=       0.428 * cos(4.09362449264  + 12036.21691740469 *T)
    sumY1  +=       0.487 * cos(4.98974857689  + 1901.77642916080 *T)
    sumY1  +=       0.484 * cos(1.19386220714  + 1357.85837006460 *T)
    sumY1  +=       0.435 * cos(4.56496977263  + 2729.96314120569 *T)
    sumY1  +=       0.435 * cos(1.42337711904  + 2730.45077617271 *T)
    sumY1  +=       0.393 * cos(3.70194655878  + 3803.57216539650 *T)
    sumY1  +=       0.393 * cos(3.70194655878  + 3804.05980036350 *T)
    sumY1  +=       0.396 * cos(2.29672548563  + 1020.26887175750 *T)
    sumY1  +=       0.413 * cos(0.90721614242  + 1902.52718868620 *T)
    sumY1  +=       0.438 * cos(5.82247234506  + 611.82992814620 *T)
    sumY1  +=       0.404 * cos(1.67645195925  + 2042.74160658630 *T)
    sumY1  +=       0.412 * cos(4.61028536998  + 4863.44173055270 *T)
    sumY1  +=       0.379 * cos(0.15728961892  + 4443.95708667750 *T)
    sumY1  +=       0.528 * cos(6.10721110527  + 721.40560204669 *T)
    sumY1  +=       0.392 * cos(5.54718824078  + 2001.20017467470 *T)
    sumY1  +=       0.379 * cos(0.35077315636  + 9676.23721663209 *T)
    sumY1  +=       0.379 * cos(3.49236580995  + 9676.72485159910 *T)
    sumY1  +=       0.372 * cos(3.07971309061  + 3273.88120030190 *T)
    sumY1  +=       0.372 * cos(3.07971309060  + 3274.36883526891 *T)

    # JUPITER Y Coord
    sumY2  = 0
    sumY2  +=  202462.055 * cos(0.59792458091  + 529.93478257810 *T)
    sumY2  +=  132313.738 * cos(3.89295968796  + 0.24381748350 *T)
    sumY2  +=  123752.958 * cos(2.55853913064  + 522.82123557730 *T)
    sumY2  +=  121682.952 * cos(4.91767938901  + 537.04832957890 *T)
    sumY2  +=   35719.700 * cos(3.45512045921  + 1059.62574767270 *T)
    sumY2  +=   11394.908 * cos(0.15831191722  + 7.35736448430 *T)
    sumY2  +=    7053.660 * cos(5.44857872902  + 1066.73929467350 *T)
    sumY2  +=    5684.733 * cos(2.05222271085  + 1052.51220067191 *T)
    sumY2  +=    5262.466 * cos(2.69780996624  + 515.70768857651 *T)
    sumY2  +=    3815.469 * cos(3.98689525353  + 1589.31671276730 *T)
    sumY2  +=    3706.857 * cos(4.78859671078  + 544.16187657970 *T)
    sumY2  +=    1449.544 * cos(1.07785028902  + 110.45013870291 *T)
    sumY2  +=    1046.768 * cos(0.75395840567  + 426.84200835950 *T)
    sumY2  +=    1322.480 * cos(0.05697677565  + 14.47091148511 *T)
    sumY2  +=     858.857 * cos(0.30653313914  + 633.02755679670 *T)
    sumY2  +=     887.013 * cos(3.09315872764  + 526.75338884040 *T)
    sumY2  +=     866.197 * cos(4.38230951425  + 533.11617631580 *T)
    sumY2  +=     997.140 * cos(6.17027795984  + 323.74923414091 *T)
    sumY2  +=     563.455 * cos(0.04960726876  + 949.41942645330 *T)
    sumY2  +=     571.107 * cos(1.58632661254  + 1045.39865367110 *T)
    sumY2  +=     519.207 * cos(4.39249603222  + 526.00262931501 *T)
    sumY2  +=     607.942 * cos(1.27749143447  + 736.12033101530 *T)
    sumY2  +=     506.922 * cos(3.07709057807  + 533.86693584120 *T)
    sumY2  +=     485.331 * cos(3.17541273379  + 109.96250373590 *T)
    sumY2  +=     475.072 * cos(5.78207162854  + 529.44714761109 *T)
    sumY2  +=     458.270 * cos(5.95238630892  + 1596.43025976811 *T)
    sumY2  +=     418.781 * cos(1.04079958107  + 117.56368570370 *T)
    sumY2  +=     525.152 * cos(1.98263661840  + 102.84895673509 *T)
    sumY2  +=     374.437 * cos(0.84144118564  + 95.73540973430 *T)
    sumY2  +=     412.815 * cos(5.04212838870  + 103.33659170210 *T)
    sumY2  +=     320.928 * cos(2.67288808735  + 419.72846135871 *T)
    sumY2  +=     322.793 * cos(4.41855165225  + 2119.00767786191 *T)
    sumY2  +=     302.747 * cos(1.84681315770  + 205.94173095370 *T)
    sumY2  +=     280.926 * cos(2.12468407469  + 1582.20316576650 *T)
    sumY2  +=     243.003 * cos(1.57275553426  + 625.91400979590 *T)
    sumY2  +=     260.213 * cos(2.10127305773  + 956.53297345411 *T)
    sumY2  +=     223.206 * cos(5.59326372578  + 433.95555536030 *T)
    sumY2  +=     219.874 * cos(0.15087500723  + 942.30587945250 *T)
    sumY2  +=     202.904 * cos(5.19995521771  + 220.65645992230 *T)
    sumY2  +=     196.744 * cos(4.42883920470  + 1169.83206889211 *T)
    sumY2  +=     171.475 * cos(2.19448427444  + 6.86972951729 *T)
    sumY2  +=     195.569 * cos(4.41657624168  + 640.14110379750 *T)
    sumY2  +=     165.172 * cos(3.71383053394  + 316.63568714010 *T)
    sumY2  +=     163.913 * cos(5.23556107603  + 1073.85284167431 *T)
    sumY2  +=     164.904 * cos(1.98636179817  + 1155.60497489050 *T)
    sumY2  +=     138.497 * cos(2.85820387486  + 508.59414157570 *T)
    sumY2  +=     135.640 * cos(1.81977900596  + 198.82818395289 *T)
    sumY2  +=     170.382 * cos(6.03512783536  + 330.86278114170 *T)
    sumY2  +=     123.801 * cos(3.67192968217  + 743.23387801611 *T)
    sumY2  +=     105.922 * cos(1.10242641180  + 853.44019923550 *T)
    sumY2  +=     104.749 * cos(2.62023485267  + 522.33360061030 *T)
    sumY2  +=     115.508 * cos(2.71991945901  + 309.03450517230 *T)
    sumY2  +=     127.421 * cos(1.41853462450  + 729.00678401450 *T)
    sumY2  +=     104.374 * cos(0.89194130513  + 213.54291292150 *T)
    sumY2  +=      74.731 * cos(4.62028784658  + 551.27542358051 *T)
    sumY2  +=      76.056 * cos(2.59067478621  + 1265.81129610991 *T)
    sumY2  +=      84.864 * cos(3.26374582961  + 519.63984183961 *T)
    sumY2  +=      85.433 * cos(4.07536581059  + 213.05527795450 *T)
    sumY2  +=      70.250 * cos(2.97938769876  + 426.35437339250 *T)
    sumY2  +=      70.226 * cos(5.88545650951  + 206.42936592071 *T)
    sumY2  +=      65.869 * cos(2.26360027552  + 839.21310523390 *T)
    sumY2  +=      62.918 * cos(2.84024006922  + 412.61491435790 *T)
    sumY2  +=      67.605 * cos(4.19524535793  + 540.22972331660 *T)
    sumY2  +=      71.970 * cos(1.76149388570  + 419.24082639170 *T)
    sumY2  +=      53.452 * cos(5.96965868749  + 3.42521122121 *T)
    sumY2  +=      52.805 * cos(0.64612484946  + 412.12727939089 *T)
    sumY2  +=      55.599 * cos(6.19142815990  + 21.58445848590 *T)
    sumY2  +=      46.667 * cos(5.06765916310  + 227.77000692311 *T)
    sumY2  +=      48.911 * cos(4.88627603894  + 1062.80714141041 *T)
    sumY2  +=      46.046 * cos(1.69944877998  + 536.56069461190 *T)
    sumY2  +=      49.381 * cos(2.96227489837  + 1162.71852189130 *T)
    sumY2  +=      39.853 * cos(2.83106272106  + 1258.69774910910 *T)
    sumY2  +=      38.964 * cos(3.82409926208  + 309.52214013930 *T)
    sumY2  +=      38.082 * cos(4.25027357628  + 441.06910236111 *T)
    sumY2  +=      32.665 * cos(3.07238994770  + 1692.40948698591 *T)
    sumY2  +=      31.801 * cos(0.14719940852  + 2126.12122486270 *T)
    sumY2  +=      36.510 * cos(2.83147334737  + 1056.44435393500 *T)
    sumY2  +=      40.831 * cos(1.76738911692  + 1575.08961876571 *T)
    sumY2  +=      31.242 * cos(2.99902293895  + 618.80046279510 *T)
    sumY2  +=      32.827 * cos(1.87547695506  + 515.22005360950 *T)
    sumY2  +=      28.849 * cos(5.03497321398  + 1055.69359440960 *T)
    sumY2  +=      32.224 * cos(4.96133890377  + 1272.92484311070 *T)
    sumY2  +=      28.682 * cos(1.53844011595  + 1038.28510667030 *T)
    sumY2  +=      29.430 * cos(3.21972350654  + 1486.22393854870 *T)
    sumY2  +=      34.150 * cos(2.79751290016  + 301.92095817150 *T)
    sumY2  +=      32.088 * cos(2.52348716505  + 625.42637482889 *T)
    sumY2  +=      24.170 * cos(2.39467034468  + 832.09955823310 *T)
    sumY2  +=      23.462 * cos(0.52036917801  + 1059.13811270570 *T)
    sumY2  +=      23.622 * cos(2.60159482405  + 742.74624304909 *T)
    sumY2  +=      25.413 * cos(3.79378187392  + 1368.90407032850 *T)
    sumY2  +=      25.008 * cos(4.81402651995  + 2648.69864295650 *T)
    sumY2  +=      19.641 * cos(3.29857408882  + 1063.55790093581 *T)
    sumY2  +=      20.824 * cos(4.96156233332  + 1699.52303398670 *T)
    sumY2  +=      18.958 * cos(1.33062268278  + 838.72547026689 *T)
    sumY2  +=      19.085 * cos(0.27421860784  + 1479.11039154791 *T)
    sumY2  +=      20.467 * cos(4.52802295353  + 647.25465079831 *T)
    sumY2  +=      18.387 * cos(0.90119440297  + 1471.99684454710 *T)
    sumY2  +=      16.332 * cos(0.27219099833  + 728.51914904750 *T)
    sumY2  +=      18.784 * cos(1.39830492533  + 735.63269604830 *T)
    sumY2  +=      13.140 * cos(0.04984430125  + 1376.01761732930 *T)
    sumY2  +=      14.219 * cos(0.85329624536  + 124.67723270450 *T)
    sumY2  +=      12.676 * cos(1.25087875699  + 831.61192326610 *T)
    sumY2  +=      12.289 * cos(5.75436340555  + 10.53875822200 *T)
    sumY2  +=      11.203 * cos(1.59407227495  + 511.77553531340 *T)
    sumY2  +=      12.667 * cos(5.77688101830  + 963.64652045490 *T)
    sumY2  +=      13.805 * cos(2.26287537924  + 2111.89413086110 *T)
    sumY2  +=      10.749 * cos(5.88274412057  + 548.09402984280 *T)
    sumY2  +=      12.863 * cos(4.47014391443  + 750.34742501690 *T)
    sumY2  +=      12.393 * cos(2.34379668336  + 1685.29593998510 *T)
    sumY2  +=      12.549 * cos(2.17268947101  + 941.81824448550 *T)
    sumY2  +=      10.468 * cos(0.30458009626  + 935.19233245171 *T)
    sumY2  +=       8.995 * cos(1.74960944229  + 1383.13116433011 *T)
    sumY2  +=      11.159 * cos(2.51015843852  + 618.31282782810 *T)
    sumY2  +=       7.631 * cos(5.05979604407  + 632.53992182970 *T)
    sumY2  +=       7.503 * cos(5.67188098812  + 1603.54380676890 *T)
    sumY2  +=       7.708 * cos(1.95453603914  + 191.71463695210 *T)
    sumY2  +=       7.383 * cos(1.13509741924  + 1155.11733992350 *T)
    sumY2  +=       8.531 * cos(3.84811863075  + 1361.79052332771 *T)
    sumY2  +=       8.011 * cos(5.88118865782  + 337.97632814250 *T)
    sumY2  +=       7.223 * cos(3.84606265734  + 88.62186273350 *T)
    sumY2  +=       7.778 * cos(1.36850966173  + 1052.02456570489 *T)
    sumY2  +=       7.476 * cos(1.59564508061  + 721.89323701370 *T)
    sumY2  +=       5.663 * cos(3.52114816133  + 2222.10045208050 *T)
    sumY2  +=       5.661 * cos(0.00460945410  + 1044.91101870410 *T)
    sumY2  +=       5.318 * cos(0.95938503716  + 1148.00379292269 *T)
    sumY2  +=       4.988 * cos(0.39989410631  + 639.65346883050 *T)
    sumY2  +=       5.093 * cos(2.58948327756  + 106.03035047280 *T)
    sumY2  +=       6.533 * cos(3.48079365390  + 845.83901726770 *T)
    sumY2  +=       6.385 * cos(0.35391558166  + 860.55374623631 *T)
    sumY2  +=       4.988 * cos(1.41354054180  + 99.66756299740 *T)
    sumY2  +=       5.571 * cos(2.81444404460  + 1795.50226120450 *T)
    sumY2  +=       5.541 * cos(2.42198547004  + 1049.33080693420 *T)
    sumY2  +=       5.247 * cos(1.88295964360  + 1258.21011414209 *T)
    sumY2  +=       4.445 * cos(3.79759027615  + 846.32665223470 *T)
    sumY2  +=       3.990 * cos(4.74398115700  + 1802.61580820530 *T)
    sumY2  +=       3.926 * cos(2.16218108184  + 497.69145366371 *T)
    sumY2  +=       3.892 * cos(3.84670029567  + 302.40859313851 *T)
    sumY2  +=       3.809 * cos(3.73909928384  + 2.93757625420 *T)
    sumY2  +=       3.620 * cos(5.32009722771  + 562.17811149250 *T)
    sumY2  +=       4.760 * cos(2.22483727905  + 934.70469748469 *T)
    sumY2  +=       4.154 * cos(5.29683004255  + 96.22304470130 *T)
    sumY2  +=       3.877 * cos(2.20085136203  + 1375.52998236229 *T)
    sumY2  +=       3.522 * cos(3.09584301581  + 501.48059457490 *T)
    sumY2  +=       3.445 * cos(2.53099132900  + 1148.49142788971 *T)
    sumY2  +=       3.382 * cos(4.89060343280  + 234.88355392390 *T)
    sumY2  +=       3.592 * cos(3.66643310344  + 405.50136735710 *T)
    sumY2  +=       3.808 * cos(2.35607884579  + 1788.38871420371 *T)
    sumY2  +=       3.336 * cos(0.87926034840  + 1471.50920958009 *T)
    sumY2  +=       3.758 * cos(0.45509718094  + 113.63153244060 *T)
    sumY2  +=       3.174 * cos(3.90621979282  + 448.18264936190 *T)
    sumY2  +=       3.698 * cos(1.01340900246  + 220.16882495529 *T)
    sumY2  +=       3.101 * cos(3.18331755184  + 316.14805217309 *T)
    sumY2  +=       2.746 * cos(4.17978592068  + 81.50831573269 *T)
    sumY2  +=       2.816 * cos(4.07738334997  + 2015.91490364330 *T)
    sumY2  +=       2.754 * cos(5.70117381714  + 430.02340209721 *T)
    sumY2  +=       2.783 * cos(1.81572481299  + 2104.78058386030 *T)
    sumY2  +=       3.501 * cos(4.61828326310  + 1069.92068841121 *T)
    sumY2  +=       2.580 * cos(2.70735869418  + 98.91680347199 *T)
    sumY2  +=       2.605 * cos(3.58250505073  + 611.68691579430 *T)
    sumY2  +=       2.955 * cos(5.33795885153  + 1592.49810650500 *T)
    sumY2  +=       3.418 * cos(0.91402819173  + 1368.41643536150 *T)
    sumY2  +=       2.482 * cos(6.19470895878  + 1361.30288836070 *T)
    sumY2  +=       2.409 * cos(5.65098301690  + 107.26874496520 *T)
    sumY2  +=       2.450 * cos(1.82798028749  + 948.93179148630 *T)
    sumY2  +=       2.454 * cos(3.07591757821  + 1251.58420210830 *T)
    sumY2  +=       2.443 * cos(0.80546406296  + 1464.39566257930 *T)
    sumY2  +=       2.848 * cos(3.22965138097  + 1898.59503542310 *T)
    sumY2  +=       2.434 * cos(4.87638273487  + 1080.96638867510 *T)
    sumY2  +=       2.240 * cos(1.77459488806  + 629.84616305900 *T)
    sumY2  +=       2.676 * cos(2.93522198834  + 294.80741117069 *T)
    sumY2  +=       2.252 * cos(4.20412336387  + 423.66061462181 *T)
    sumY2  +=       2.207 * cos(1.26929749414  + 106.78110999820 *T)
    sumY2  +=       2.966 * cos(0.61692584962  + 106.51798543980 *T)
    sumY2  +=       2.114 * cos(2.97133804999  + 117.07605073670 *T)
    sumY2  +=       2.234 * cos(3.71703805766  + 6276.20612047410 *T)
    sumY2  +=       2.232 * cos(6.07586800494  + 6290.43321447570 *T)
    sumY2  +=       2.060 * cos(0.61396711293  + 2655.81218995730 *T)
    sumY2  +=       2.492 * cos(1.55690527788  + 1574.60198379870 *T)
    sumY2  +=       2.405 * cos(3.40255115300  + 1162.23088692429 *T)
    sumY2  +=       2.094 * cos(4.26293458943  + 13.98327651810 *T)
    sumY2  +=       1.997 * cos(2.59413376987  + 2008.80135664250 *T)
    sumY2  +=       1.998 * cos(3.66897172067  + 149.80701461810 *T)
    sumY2  +=       2.075 * cos(5.13994005595  + 10206.41581669370 *T)
    sumY2  +=       2.073 * cos(1.21558655630  + 10220.64291069530 *T)
    sumY2  +=       1.833 * cos(2.60928669023  + 824.98601123230 *T)
    sumY2  +=       1.940 * cos(5.08767479418  + 199.31581891990 *T)
    sumY2  +=       1.989 * cos(1.62630471057  + 323.26159917390 *T)
    sumY2  +=       1.764 * cos(4.69152602347  + 970.76006745570 *T)
    sumY2  +=       2.079 * cos(0.16395458213  + 1905.70858242391 *T)
    sumY2  +=       1.815 * cos(1.60120314289  + 1265.32366114289 *T)
    sumY2  +=       1.815 * cos(4.28784753553  + 558.38897058130 *T)
    sumY2  +=       1.819 * cos(3.79124340355  + 1165.89991562901 *T)
    sumY2  +=       2.303 * cos(1.95114448415  + 1251.09656714130 *T)
    sumY2  +=       1.797 * cos(5.48846817914  + 1585.38455950421 *T)
    sumY2  +=       1.785 * cos(2.60041060537  + 1159.53712815361 *T)
    sumY2  +=       1.719 * cos(3.18124488238  + 654.36819779910 *T)
    sumY2  +=       2.106 * cos(2.00507240636  + 508.10650660869 *T)
    sumY2  +=       1.746 * cos(2.85410644796  + 1586.13531902961 *T)
    sumY2  +=       1.693 * cos(1.86129406512  + 1691.92185201890 *T)
    sumY2  +=       1.560 * cos(2.48952478561  + 2001.68780964171 *T)
    sumY2  +=       1.635 * cos(0.62669051313  + 1787.90107923670 *T)
    sumY2  +=       1.492 * cos(5.75772456494  + 422.90985509640 *T)
    sumY2  +=       1.558 * cos(3.23097345434  + 636.20895053440 *T)
    sumY2  +=       1.753 * cos(4.84300262737  + 757.46097201770 *T)
    sumY2  +=       1.757 * cos(5.31425455438  + 3178.38960805111 *T)
    sumY2  +=       1.411 * cos(4.82184730987  + 956.04533848709 *T)
    sumY2  +=       1.925 * cos(1.69879574514  + 1567.97607176490 *T)
    sumY2  +=       1.563 * cos(2.82364433424  + 1891.48148842230 *T)
    sumY2  +=       1.383 * cos(4.87286586350  + 5230.56364931950 *T)
    sumY2  +=       1.393 * cos(2.50410021460  + 1280.03839011150 *T)
    sumY2  +=       1.381 * cos(2.51458784132  + 5216.33655531789 *T)
    sumY2  +=       1.548 * cos(0.74638079450  + 946.23803271561 *T)
    sumY2  +=       1.239 * cos(1.33958827620  + 543.67424161270 *T)
    sumY2  +=       1.373 * cos(1.17035918011  + 1890.99385345530 *T)
    sumY2  +=       1.198 * cos(5.50480829771  + 326.93062787860 *T)
    sumY2  +=       1.220 * cos(5.64182300635  + 319.81708087780 *T)
    sumY2  +=       1.357 * cos(5.70946834472  + 2229.21399908130 *T)
    sumY2  +=       1.271 * cos(0.01266765679  + 9160.77334553909 *T)
    sumY2  +=       1.268 * cos(3.93758835915  + 9146.54625153749 *T)
    sumY2  +=       1.183 * cos(1.68810975581  + 636.95971005981 *T)
    sumY2  +=       1.254 * cos(6.07103801223  + 433.46792039329 *T)
    sumY2  +=       1.180 * cos(0.55156432114  + 1780.78753223589 *T)
    sumY2  +=       1.421 * cos(3.61381859664  + 1678.18239298430 *T)
    sumY2  +=       1.111 * cos(3.54163507553  + 910.06255053810 *T)
    sumY2  +=       1.135 * cos(3.99300516499  + 540.98048284201 *T)
    sumY2  +=       1.425 * cos(0.59800839792  + 1684.80830501809 *T)
    sumY2  +=       1.101 * cos(0.33268231661  + 518.88908231420 *T)
    sumY2  +=       1.052 * cos(0.87116249009  + 1493.33748554950 *T)
    sumY2  +=       0.954 * cos(2.13166225661  + 1066.25165970650 *T)
    sumY2  +=       1.305 * cos(3.68417944671  + 5223.45010231870 *T)
    sumY2  +=       0.932 * cos(1.81701400569  + 952.60082019101 *T)
    sumY2  +=       1.152 * cos(4.15622157039  + 3281.48238226970 *T)
    sumY2  +=       1.271 * cos(2.91723787442  + 1478.62275658089 *T)
    sumY2  +=       0.993 * cos(2.42220158667  + 2427.79836555070 *T)
    sumY2  +=       1.202 * cos(1.66398680721  + 1567.48843679789 *T)
    sumY2  +=       0.952 * cos(2.83462623397  + 2325.19322629910 *T)
    sumY2  +=       0.904 * cos(3.88771946892  + 1158.78636862821 *T)
    sumY2  +=       1.185 * cos(2.82656778833  + 2214.98690507971 *T)
    sumY2  +=       0.965 * cos(2.41761451411  + 2428.28600051770 *T)
    sumY2  +=       0.862 * cos(3.45873513038  + 5753.14106741329 *T)
    sumY2  +=       1.183 * cos(0.56670778502  + 953.35157971640 *T)
    sumY2  +=       0.858 * cos(5.60347835412  + 749.85979004990 *T)
    sumY2  +=       0.857 * cos(5.92079986515  + 1677.69475801730 *T)
    sumY2  +=       0.896 * cos(4.05090365368  + 1272.43720814370 *T)
    sumY2  +=       0.838 * cos(1.85345325590  + 740.05248427840 *T)
    sumY2  +=       0.810 * cos(1.99086557504  + 732.93893727761 *T)
    sumY2  +=       0.899 * cos(1.44837834416  + 1581.71553079950 *T)
    sumY2  +=       1.107 * cos(1.52991417673  + 1912.82212942470 *T)
    sumY2  +=       1.099 * cos(1.63363197193  + 21.09682351890 *T)
    sumY2  +=       0.856 * cos(0.26870253693  + 210.36151918381 *T)
    sumY2  +=       0.792 * cos(4.88168256779  + 9683.35076363289 *T)
    sumY2  +=       0.803 * cos(2.43011635326  + 2641.58509595570 *T)
    sumY2  +=       1.013 * cos(0.26382735728  + 852.95256426850 *T)
    sumY2  +=       0.790 * cos(2.45929735954  + 1166.65067515440 *T)
    sumY2  +=       0.927 * cos(2.40548084691  + 202.76033721600 *T)
    sumY2  +=       0.955 * cos(2.63053495096  + 611.19928082730 *T)
    sumY2  +=       0.848 * cos(1.52904158095  + 824.49837626530 *T)
    sumY2  +=       0.721 * cos(3.87768985281  + 4693.75913722409 *T)
    sumY2  +=       0.732 * cos(5.25617010486  + 1141.37788088890 *T)
    sumY2  +=       0.717 * cos(3.47473573810  + 2538.49232173710 *T)
    sumY2  +=       0.764 * cos(1.02089435958  + 3789.83270636190 *T)
    sumY2  +=       0.757 * cos(2.29144101005  + 422.42222012939 *T)
    sumY2  +=       0.870 * cos(1.70474992020  + 3900.03902758130 *T)
    sumY2  +=       0.745 * cos(1.11551269154  + 416.05943265400 *T)
    sumY2  +=       0.786 * cos(1.87783459546  + 3590.76070492550 *T)
    sumY2  +=       0.722 * cos(0.69162917181  + 1809.72935520611 *T)
    sumY2  +=       0.731 * cos(1.56640382879  + 2008.31372167550 *T)
    sumY2  +=       0.784 * cos(2.40595880647  + 1795.01462623750 *T)
    sumY2  +=       0.655 * cos(5.29949669283  + 8623.96883344369 *T)
    sumY2  +=       0.859 * cos(4.45239063404  + 459.08533727390 *T)
    sumY2  +=       0.787 * cos(3.89498804276  + 1671.06884598351 *T)
    sumY2  +=       0.668 * cos(0.46476131572  + 2545.60586873790 *T)
    sumY2  +=       0.707 * cos(0.44824026865  + 1169.34443392509 *T)
    sumY2  +=       0.679 * cos(0.69182567589  + 1588.82907780029 *T)
    sumY2  +=       0.721 * cos(0.46106254321  + 2104.29294889330 *T)
    sumY2  +=       0.869 * cos(2.05707977661  + 3576.53361092390 *T)
    sumY2  +=       0.878 * cos(5.49426984828  + 75.02541605080 *T)
    sumY2  +=       0.883 * cos(2.87549710098  + 3384.57515648830 *T)
    sumY2  +=       0.627 * cos(2.60545252272  + 405.01373239010 *T)
    sumY2  +=       0.776 * cos(2.36750552228  + 927.59115048390 *T)
    sumY2  +=       0.751 * cos(3.92531642097  + 3597.87425192630 *T)
    sumY2  +=       0.821 * cos(3.33729170182  + 3487.66793070691 *T)
    sumY2  +=       0.603 * cos(2.24262715346  + 2318.07967929830 *T)
    sumY2  +=       0.766 * cos(3.89405071692  + 1354.67697632690 *T)
    sumY2  +=       0.679 * cos(0.93833494617  + 2207.38572311189 *T)
    sumY2  +=       0.820 * cos(4.88672199584  + 6283.31966747490 *T)
    sumY2  +=       0.638 * cos(0.90714252231  + 3370.34806248671 *T)
    sumY2  +=       0.711 * cos(1.29280470587  + 1883.88030645450 *T)
    sumY2  +=       0.666 * cos(1.22953421972  + 3693.85347914410 *T)
    sumY2  +=       0.785 * cos(3.49527016473  + 3149.44778508089 *T)
    sumY2  +=       0.566 * cos(0.13027536492  + 2861.99773839450 *T)
    sumY2  +=       0.781 * cos(1.41485635870  + 3796.94625336271 *T)
    sumY2  +=       0.719 * cos(3.02988768028  + 600.78422788230 *T)
    sumY2  +=       0.586 * cos(3.97655750409  + 732.45130231059 *T)
    sumY2  +=       0.546 * cos(0.87759549299  + 437.88770862340 *T)
    sumY2  +=       0.539 * cos(3.39333684490  + 209.12312469139 *T)
    sumY2  +=       0.659 * cos(2.59360004759  + 2207.87335807890 *T)
    sumY2  +=       0.617 * cos(2.08977841036  + 2744.67787017430 *T)
    sumY2  +=       0.617 * cos(2.08977841036  + 2744.19023520730 *T)
    sumY2  +=       0.687 * cos(2.15934118066  + 209.87388421679 *T)
    sumY2  +=       0.566 * cos(4.70636247947  + 526.26575387340 *T)
    sumY2  +=       0.640 * cos(3.60830260468  + 1987.46071564010 *T)
    sumY2  +=       0.623 * cos(2.21578255427  + 3274.36883526891 *T)
    sumY2  +=       0.623 * cos(2.21578255427  + 3273.88120030190 *T)
    sumY2  +=       0.682 * cos(4.78157060479  + 2332.30677329990 *T)
    sumY2  +=       0.719 * cos(5.11346846815  + 9153.65979853830 *T)
    sumY2  +=       0.547 * cos(0.34236516611  + 2097.17940189250 *T)
    sumY2  +=       0.705 * cos(2.59491824208  + 3700.96702614491 *T)
    sumY2  +=       0.574 * cos(3.41963712247  + 430.77416162260 *T)
    sumY2  +=       0.543 * cos(0.88081760855  + 1457.28211557850 *T)
    sumY2  +=       0.693 * cos(0.33025900018  + 2001.20017467470 *T)
    sumY2  +=       0.659 * cos(2.96839974316  + 3804.05980036350 *T)
    sumY2  +=       0.581 * cos(2.15878783346  + 2111.40649589409 *T)
    sumY2  +=       0.540 * cos(6.13903070975  + 2729.96314120569 *T)
    sumY2  +=       0.653 * cos(6.09326422469  + 2626.87036698710 *T)
    sumY2  +=       0.673 * cos(3.08841787143  + 416.54706762100 *T)
    sumY2  +=       0.541 * cos(0.97608919869  + 2214.49927011270 *T)
    sumY2  +=       0.463 * cos(3.71557681634  + 529.88659846831 *T)
    sumY2  +=       0.463 * cos(0.63020265214  + 529.98296668790 *T)
    sumY2  +=       0.453 * cos(5.18763231998  + 5753.62870238030 *T)
    sumY2  +=       0.499 * cos(1.74748261623  + 2421.17245351690 *T)
    sumY2  +=       0.456 * cos(3.28830064335  + 76.50988875911 *T)
    sumY2  +=       0.587 * cos(1.13660838990  + 527.48710202330 *T)
    sumY2  +=       0.505 * cos(1.06180904724  + 532.62854134880 *T)
    sumY2  +=       0.582 * cos(3.20768877862  + 532.38246313290 *T)
    sumY2  +=       0.517 * cos(0.69632241122  + 2523.77759276850 *T)
    sumY2  +=       0.475 * cos(2.88817124373  + 2531.37877473630 *T)
    sumY2  +=       0.433 * cos(0.32149741104  + 621.98185653281 *T)
    sumY2  +=       0.530 * cos(3.32159958445  + 512.52629483880 *T)
    sumY2  +=       0.530 * cos(1.01908606392  + 547.34327031741 *T)
    sumY2  +=       0.478 * cos(3.50052966452  + 1593.24886603040 *T)
    sumY2  +=       0.423 * cos(0.32742738106  + 9683.83839859990 *T)
    sumY2  +=       0.592 * cos(1.66841184917  + 2324.70559133209 *T)
    sumY2  +=       0.410 * cos(4.13116496171  + 2751.30378220809 *T)
    sumY2  +=       0.507 * cos(2.10021567641  + 1243.98302014049 *T)
    sumY2  +=       0.521 * cos(1.46001387840  + 4216.43089723790 *T)
    sumY2  +=       0.458 * cos(4.03023009406  + 629.09540353360 *T)
    sumY2  +=       0.508 * cos(3.32555628701  + 1994.57426264090 *T)
    sumY2  +=       0.508 * cos(0.18396363342  + 1994.08662767389 *T)
    sumY2  +=       0.452 * cos(4.24985614624  + 3.68833577960 *T)
    sumY2  +=       0.527 * cos(2.84159709661  + 597.60283414461 *T)
    sumY2  +=       0.527 * cos(4.64068120535  + 462.26673101160 *T)
    sumY2  +=       0.555 * cos(1.79977655362  + 3892.92548058051 *T)
    sumY2  +=       0.464 * cos(3.31404699259  + 1905.22094745690 *T)
    sumY2  +=       0.535 * cos(2.60058723927  + 2943.74987161071 *T)
    sumY2  +=       0.535 * cos(5.74217989286  + 2943.26223664369 *T)
    sumY2  +=       0.450 * cos(3.73627122155  + 3914.26612158291 *T)
    sumY2  +=       0.390 * cos(4.12428063042  + 2751.79141717511 *T)
    sumY2  +=       0.511 * cos(4.46664359519  + 643.32249753521 *T)
    sumY2  +=       0.431 * cos(0.85128279093  + 849.50804597240 *T)
    sumY2  +=       0.464 * cos(6.26761999534  + 2633.98391398789 *T)
    sumY2  +=       0.486 * cos(3.27293685238  + 2303.85258529670 *T)
    sumY2  +=       0.412 * cos(6.19787080325  + 2758.90496417590 *T)
    sumY2  +=       0.423 * cos(1.17747515264  + 1898.10740045609 *T)
    sumY2  +=       0.457 * cos(5.26603357261  + 519.15220687259 *T)
    sumY2  +=       0.530 * cos(2.82489810933  + 2221.61281711350 *T)
    sumY2  +=       0.443 * cos(5.28863656813  + 89.10949770051 *T)
    sumY2  +=       0.478 * cos(0.39094045474  + 2420.68481854989 *T)

    # JUPITER Y Coord
    sumY3  = 0
    sumY3  +=   16990.902 * cos(4.27247332045  + 522.82123557730 *T)
    sumY3  +=   16795.793 * cos(3.18563475573  + 537.04832957890 *T)
    sumY3  +=    5535.565 * cos(5.40269185025  + 0.24381748350 *T)
    sumY3  +=    1711.756 * cos(1.80070871020  + 1059.62574767270 *T)
    sumY3  +=    1437.893 * cos(4.40672411903  + 515.70768857651 *T)
    sumY3  +=    1080.832 * cos(3.83903053740  + 1066.73929467350 *T)
    sumY3  +=    1014.927 * cos(3.07247314388  + 544.16187657970 *T)
    sumY3  +=    1080.320 * cos(4.63909200214  + 7.35736448430 *T)
    sumY3  +=     700.816 * cos(3.91629287572  + 1052.51220067191 *T)
    sumY3  +=     245.290 * cos(2.63485195455  + 1589.31671276730 *T)
    sumY3  +=     301.422 * cos(4.61795245261  + 14.47091148511 *T)
    sumY3  +=     233.420 * cos(0.22403594785  + 529.93478257810 *T)
    sumY3  +=     141.715 * cos(4.79041763529  + 529.44714761109 *T)
    sumY3  +=     134.439 * cos(3.36367709264  + 1045.39865367110 *T)
    sumY3  +=      78.118 * cos(4.43004262792  + 1596.43025976811 *T)
    sumY3  +=      76.833 * cos(5.59776741855  + 117.56368570370 *T)
    sumY3  +=      92.926 * cos(4.40350030513  + 323.74923414091 *T)
    sumY3  +=      78.223 * cos(5.40096869297  + 110.45013870291 *T)
    sumY3  +=      65.879 * cos(1.48772085888  + 109.96250373590 *T)
    sumY3  +=      58.147 * cos(4.57249620850  + 508.59414157570 *T)
    sumY3  +=      51.718 * cos(2.58980208142  + 95.73540973430 *T)
    sumY3  +=      58.000 * cos(5.16608137630  + 426.84200835950 *T)
    sumY3  +=      47.055 * cos(3.57438012615  + 1073.85284167431 *T)
    sumY3  +=      41.297 * cos(2.04634997990  + 633.02755679670 *T)
    sumY3  +=      56.002 * cos(3.01090434125  + 736.12033101530 *T)
    sumY3  +=      40.590 * cos(1.87109726412  + 942.30587945250 *T)
    sumY3  +=      36.838 * cos(3.19435792826  + 625.91400979590 *T)
    sumY3  +=      32.307 * cos(2.88952809764  + 551.27542358051 *T)
    sumY3  +=      38.649 * cos(4.33064739716  + 330.86278114170 *T)
    sumY3  +=      31.637 * cos(4.00200479143  + 1582.20316576650 *T)
    sumY3  +=      32.315 * cos(3.88623349757  + 433.95555536030 *T)
    sumY3  +=      37.135 * cos(0.48127765061  + 103.33659170210 *T)
    sumY3  +=      27.869 * cos(3.45656961616  + 220.65645992230 *T)
    sumY3  +=      27.130 * cos(3.15454297418  + 2119.00767786191 *T)
    sumY3  +=      27.831 * cos(1.90579303958  + 949.41942645330 *T)
    sumY3  +=      26.529 * cos(2.71723722320  + 1169.83206889211 *T)
    sumY3  +=      26.303 * cos(4.38890518124  + 419.72846135871 *T)
    sumY3  +=      27.724 * cos(0.30357549024  + 6.86972951729 *T)
    sumY3  +=      28.942 * cos(3.12933706476  + 729.00678401450 *T)
    sumY3  +=      24.801 * cos(3.55551242105  + 198.82818395289 *T)
    sumY3  +=      22.666 * cos(3.71310041528  + 1155.60497489050 *T)
    sumY3  +=      22.463 * cos(0.48577882565  + 956.53297345411 *T)
    sumY3  +=      24.855 * cos(5.09800784432  + 519.63984183961 *T)
    sumY3  +=      17.587 * cos(4.51976917752  + 21.58445848590 *T)
    sumY3  +=      20.511 * cos(2.33822898286  + 540.22972331660 *T)
    sumY3  +=      17.573 * cos(3.85855374859  + 205.94173095370 *T)
    sumY3  +=      19.314 * cos(2.63415930465  + 640.14110379750 *T)
    sumY3  +=      13.850 * cos(5.76603142366  + 853.44019923550 *T)
    sumY3  +=      13.374 * cos(3.35682573162  + 227.77000692311 *T)
    sumY3  +=      12.226 * cos(4.58700933844  + 522.33360061030 *T)
    sumY3  +=      11.511 * cos(2.47604533623  + 441.06910236111 *T)
    sumY3  +=      11.785 * cos(4.54328580165  + 412.61491435790 *T)
    sumY3  +=       9.912 * cos(1.34647813433  + 206.42936592071 *T)
    sumY3  +=       9.439 * cos(4.79528119891  + 618.80046279510 *T)
    sumY3  +=       9.398 * cos(3.27167007870  + 1038.28510667030 *T)
    sumY3  +=       9.377 * cos(4.03210310088  + 839.21310523390 *T)
    sumY3  +=       9.606 * cos(1.29930255542  + 426.35437339250 *T)
    sumY3  +=      11.272 * cos(4.51307032502  + 309.03450517230 *T)
    sumY3  +=       8.495 * cos(0.16101118987  + 536.56069461190 *T)
    sumY3  +=       7.775 * cos(5.49067289304  + 316.63568714010 *T)
    sumY3  +=       7.402 * cos(2.44949859869  + 412.12727939089 *T)
    sumY3  +=       9.022 * cos(3.47982253523  + 1575.08961876571 *T)
    sumY3  +=       7.220 * cos(4.54159743611  + 1258.69774910910 *T)
    sumY3  +=       6.726 * cos(4.13659559768  + 832.09955823310 *T)
    sumY3  +=       6.944 * cos(5.51093137903  + 309.52214013930 *T)
    sumY3  +=       7.778 * cos(4.49951291514  + 301.92095817150 *T)
    sumY3  +=       7.986 * cos(2.40170822369  + 213.05527795450 *T)
    sumY3  +=       5.933 * cos(4.95948042350  + 2126.12122486270 *T)
    sumY3  +=       5.844 * cos(1.73926665310  + 743.23387801611 *T)
    sumY3  +=       5.965 * cos(3.67342955411  + 515.22005360950 *T)
    sumY3  +=       4.458 * cos(5.51442841213  + 1059.13811270570 *T)
    sumY3  +=       3.905 * cos(4.27184704039  + 1265.81129610991 *T)
    sumY3  +=       4.118 * cos(5.46755234125  + 124.67723270450 *T)
    sumY3  +=       3.669 * cos(1.84391323522  + 1486.22393854870 *T)
    sumY3  +=       3.534 * cos(4.38721038089  + 625.42637482889 *T)
    sumY3  +=       3.134 * cos(3.36420319845  + 1699.52303398670 *T)
    sumY3  +=       2.973 * cos(1.04471011524  + 742.74624304909 *T)
    sumY3  +=       2.686 * cos(5.78683992698  + 88.62186273350 *T)
    sumY3  +=       3.079 * cos(1.99582195262  + 935.19233245171 *T)
    sumY3  +=       2.579 * cos(3.95429921370  + 1062.80714141041 *T)
    sumY3  +=       2.378 * cos(3.65318849817  + 2648.69864295650 *T)
    sumY3  +=       2.346 * cos(2.14951839318  + 728.51914904750 *T)
    sumY3  +=       3.011 * cos(2.56665773426  + 647.25465079831 *T)
    sumY3  +=       2.322 * cos(4.05450701004  + 1603.54380676890 *T)
    sumY3  +=       2.809 * cos(3.52093842683  + 1272.92484311070 *T)
    sumY3  +=       2.701 * cos(2.72157216334  + 1471.99684454710 *T)
    sumY3  +=       2.369 * cos(3.30637620547  + 721.89323701370 *T)
    sumY3  +=       2.359 * cos(4.17347791131  + 337.97632814250 *T)
    sumY3  +=       2.423 * cos(4.28527398429  + 618.31282782810 *T)
    sumY3  +=       2.070 * cos(3.05132274908  + 831.61192326610 *T)
    sumY3  +=       1.958 * cos(3.88095030867  + 526.75338884040 *T)
    sumY3  +=       1.935 * cos(0.44518939738  + 533.11617631580 *T)
    sumY3  +=       2.297 * cos(5.54175410580  + 1368.90407032850 *T)
    sumY3  +=       2.015 * cos(3.68637600590  + 191.71463695210 *T)
    sumY3  +=       2.184 * cos(4.06594004577  + 963.64652045490 *T)
    sumY3  +=       1.685 * cos(4.73740509993  + 1685.29593998510 *T)
    sumY3  +=       1.816 * cos(1.96798308062  + 1162.71852189130 *T)
    sumY3  +=       1.733 * cos(2.70077445297  + 750.34742501690 *T)
    sumY3  +=       1.699 * cos(4.10560078356  + 1055.69359440960 *T)
    sumY3  +=       1.717 * cos(2.19034610706  + 448.18264936190 *T)
    sumY3  +=       1.811 * cos(4.93807061086  + 860.55374623631 *T)
    sumY3  +=       1.432 * cos(6.00290182284  + 81.50831573269 *T)
    sumY3  +=       1.779 * cos(2.60448403761  + 1056.44435393500 *T)
    sumY3  +=       1.397 * cos(5.29190720613  + 611.68691579430 *T)
    sumY3  +=       1.563 * cos(3.56303703117  + 838.72547026689 *T)
    sumY3  +=       1.726 * cos(5.58280644952  + 1361.79052332771 *T)
    sumY3  +=       1.152 * cos(0.12950397768  + 1383.13116433011 *T)
    sumY3  +=       1.200 * cos(3.69239539182  + 632.53992182970 *T)
    sumY3  +=       1.145 * cos(5.43151272872  + 405.50136735710 *T)
    sumY3  +=       1.078 * cos(2.03239323412  + 1479.11039154791 *T)
    sumY3  +=       1.223 * cos(6.16550077809  + 21.09682351890 *T)
    sumY3  +=       1.051 * cos(1.31296703432  + 526.00262931501 *T)
    sumY3  +=       1.414 * cos(3.85675901673  + 941.81824448550 *T)
    sumY3  +=       1.030 * cos(3.03819438792  + 533.86693584120 *T)
    sumY3  +=       1.211 * cos(2.84309753774  + 1069.92068841121 *T)
    sumY3  +=       1.240 * cos(3.92862477221  + 934.70469748469 *T)
    sumY3  +=       0.944 * cos(1.86321874038  + 1044.91101870410 *T)
    sumY3  +=       1.091 * cos(4.10316988679  + 2111.89413086110 *T)
    sumY3  +=       0.986 * cos(5.14322625016  + 2097.17940189250 *T)
    sumY3  +=       0.967 * cos(5.28600031012  + 1588.82907780029 *T)
    sumY3  +=       0.939 * cos(2.81112974414  + 13.98327651810 *T)
    sumY3  +=       0.870 * cos(5.07245340162  + 639.65346883050 *T)
    sumY3  +=       1.013 * cos(5.44748312429  + 1049.33080693420 *T)
    sumY3  +=       0.989 * cos(2.86838275982  + 1148.00379292269 *T)
    sumY3  +=       1.043 * cos(2.98280277812  + 1080.96638867510 *T)
    sumY3  +=       0.824 * cos(4.17606860247  + 433.46792039329 *T)
    sumY3  +=       1.002 * cos(4.01347037981  + 1788.38871420371 *T)
    sumY3  +=       0.823 * cos(2.03930536894  + 654.36819779910 *T)
    sumY3  +=       0.801 * cos(0.27042298823  + 1375.52998236229 *T)
    sumY3  +=       0.890 * cos(4.65790311289  + 294.80741117069 *T)
    sumY3  +=       0.763 * cos(3.00898516277  + 2421.17245351690 *T)
    sumY3  +=       0.793 * cos(2.96629034382  + 234.88355392390 *T)
    sumY3  +=       0.752 * cos(4.56374445756  + 1251.58420210830 *T)
    sumY3  +=       0.788 * cos(5.54399109100  + 302.40859313851 *T)
    sumY3  +=       0.719 * cos(3.28492821269  + 1787.90107923670 *T)
    sumY3  +=       0.760 * cos(3.49538730713  + 2104.78058386030 *T)
    sumY3  +=       0.839 * cos(4.84381227007  + 1898.59503542310 *T)
    sumY3  +=       0.962 * cos(4.80560510645  + 2318.07967929830 *T)
    sumY3  +=       0.832 * cos(4.85895718536  + 2001.68780964171 *T)
    sumY3  +=       0.930 * cos(0.44313687441  + 96.22304470130 *T)
    sumY3  +=       0.879 * cos(1.78320792762  + 845.83901726770 *T)
    sumY3  +=       0.780 * cos(1.22419144371  + 611.19928082730 *T)
    sumY3  +=       0.900 * cos(4.69837586769  + 1162.23088692429 *T)
    sumY3  +=       0.759 * cos(3.21123137459  + 2008.31372167550 *T)
    sumY3  +=       0.715 * cos(2.43949501289  + 1464.39566257930 *T)
    sumY3  +=       0.780 * cos(4.52983365743  + 2214.98690507971 *T)
    sumY3  +=       0.679 * cos(1.69767989980  + 2435.39954751850 *T)
    sumY3  +=       0.809 * cos(5.12961307681  + 1994.57426264090 *T)
    sumY3  +=       0.772 * cos(5.46156636423  + 220.16882495529 *T)
    sumY3  +=       0.622 * cos(2.00213930026  + 316.14805217309 *T)
    sumY3  +=       0.612 * cos(0.83227002431  + 749.85979004990 *T)
    sumY3  +=       0.612 * cos(1.13237645481  + 117.07605073670 *T)
    sumY3  +=       0.713 * cos(6.02221537369  + 1140.89024592190 *T)
    sumY3  +=       0.626 * cos(1.12243212464  + 1692.40948698591 *T)
    sumY3  +=       0.676 * cos(3.49335420549  + 2840.65709739210 *T)
    sumY3  +=       0.594 * cos(1.37315363305  + 102.84895673509 *T)
    sumY3  +=       0.790 * cos(3.53118420318  + 1251.09656714130 *T)
    sumY3  +=       0.756 * cos(4.64193430086  + 1148.49142788971 *T)
    sumY3  +=       0.568 * cos(0.84481707130  + 1141.37788088890 *T)
    sumY3  +=       0.616 * cos(3.74926817789  + 2524.26522773550 *T)
    sumY3  +=       0.560 * cos(2.89131917550  + 2104.29294889330 *T)
    sumY3  +=       0.686 * cos(4.59806378353  + 2627.35800195410 *T)
    sumY3  +=       0.720 * cos(3.64327792020  + 1258.21011414209 *T)
    sumY3  +=       0.733 * cos(6.11931006119  + 1884.36794142151 *T)
    sumY3  +=       0.583 * cos(1.49952942471  + 2751.79141717511 *T)
    sumY3  +=       0.546 * cos(5.87603590490  + 543.67424161270 *T)
    sumY3  +=       0.530 * cos(0.65117889899  + 1066.25165970650 *T)
    sumY3  +=       0.581 * cos(2.31120308329  + 1780.78753223589 *T)
    sumY3  +=       0.523 * cos(3.92602863833  + 1471.50920958009 *T)
    sumY3  +=       0.508 * cos(3.10717560112  + 508.10650660869 *T)
    sumY3  +=       0.645 * cos(3.47859687529  + 2015.91490364330 *T)
    sumY3  +=       0.553 * cos(5.34237729322  + 852.95256426850 *T)
    sumY3  +=       0.480 * cos(3.05830758640  + 1677.69475801730 *T)
    sumY3  +=       0.480 * cos(6.19990023999  + 1678.18239298430 *T)
    sumY3  +=       0.618 * cos(3.01489314196  + 1905.70858242391 *T)
    sumY3  +=       0.463 * cos(1.12266408736  + 89.10949770051 *T)
    sumY3  +=       0.512 * cos(3.49738396142  + 1155.11733992350 *T)
    sumY3  +=       0.497 * cos(3.67029546589  + 1376.01761732930 *T)
    sumY3  +=       0.529 * cos(2.41566570980  + 1802.61580820530 *T)
    sumY3  +=       0.617 * cos(4.55541673250  + 2634.47154895491 *T)
    sumY3  +=       0.459 * cos(5.95987834679  + 1691.92185201890 *T)
    sumY3  +=       0.503 * cos(0.43655200011  + 948.93179148630 *T)
    sumY3  +=       0.547 * cos(4.87653833894  + 2310.96613229750 *T)
    sumY3  +=       0.432 * cos(3.81824814417  + 1567.48843679789 *T)
    sumY3  +=       0.432 * cos(3.81824814417  + 1567.97607176490 *T)
    sumY3  +=       0.419 * cos(1.75904856409  + 1361.30288836070 *T)
    sumY3  +=       0.415 * cos(5.93908536808  + 558.38897058130 *T)
    sumY3  +=       0.415 * cos(4.68478558747  + 501.48059457490 *T)
    sumY3  +=       0.413 * cos(0.31276763851  + 1883.88030645450 *T)
    sumY3  +=       0.436 * cos(3.11437440839  + 1464.88329754630 *T)

    # JUPITER Y Coord
    sumY4  = 0
    sumY4  +=    1739.145 * cos(1.42580386197  + 537.04832957890 *T)
    sumY4  +=    1740.807 * cos(6.00432821679  + 522.82123557730 *T)
    sumY4  +=     593.812 * cos(3.67074805419  + 529.93478257810 *T)
    sumY4  +=     303.714 * cos(6.13319798890  + 515.70768857651 *T)
    sumY4  +=     216.971 * cos(1.33336714102  + 544.16187657970 *T)
    sumY4  +=     201.301 * cos(0.66067734911  + 0.24381748350 *T)
    sumY4  +=     128.314 * cos(2.22919023790  + 1066.73929467350 *T)
    sumY4  +=      84.519 * cos(2.80934114497  + 7.35736448430 *T)
    sumY4  +=      66.713 * cos(5.73808767639  + 1052.51220067191 *T)
    sumY4  +=      50.914 * cos(2.88970415030  + 14.47091148511 *T)
    sumY4  +=      56.189 * cos(5.95539409574  + 1059.62574767270 *T)
    sumY4  +=      23.551 * cos(5.17496754986  + 1045.39865367110 *T)
    sumY4  +=      14.975 * cos(0.07336624009  + 508.59414157570 *T)
    sumY4  +=      12.320 * cos(1.18823443350  + 1589.31671276730 *T)
    sumY4  +=      10.412 * cos(1.88156810845  + 1073.85284167431 *T)
    sumY4  +=      10.436 * cos(2.90285964648  + 1596.43025976811 *T)
    sumY4  +=      10.171 * cos(3.86119874481  + 117.56368570370 *T)
    sumY4  +=       7.028 * cos(6.10496958528  + 109.96250373590 *T)
    sumY4  +=       7.310 * cos(1.00125130523  + 551.27542358051 *T)
    sumY4  +=       5.867 * cos(4.33407208985  + 95.73540973430 *T)
    sumY4  +=       5.482 * cos(2.53643730082  + 323.74923414091 *T)
    sumY4  +=       5.308 * cos(2.60414218500  + 330.86278114170 *T)
    sumY4  +=       5.155 * cos(3.58060509730  + 942.30587945250 *T)
    sumY4  +=       4.404 * cos(4.83926930457  + 729.00678401450 *T)
    sumY4  +=       3.429 * cos(4.99509335363  + 625.91400979590 *T)
    sumY4  +=       3.596 * cos(3.33581032804  + 110.45013870291 *T)
    sumY4  +=       3.379 * cos(5.25427709084  + 198.82818395289 *T)
    sumY4  +=       2.870 * cos(2.96757411587  + 529.44714761109 *T)
    sumY4  +=       3.465 * cos(4.76476994232  + 736.12033101530 *T)
    sumY4  +=       2.781 * cos(5.80054865786  + 1582.20316576650 *T)
    sumY4  +=       2.710 * cos(0.81919347445  + 441.06910236111 *T)
    sumY4  +=       2.621 * cos(0.96453577508  + 1169.83206889211 *T)
    sumY4  +=       2.851 * cos(2.20827145709  + 103.33659170210 *T)
    sumY4  +=       3.222 * cos(1.76964083344  + 433.95555536030 *T)
    sumY4  +=       2.252 * cos(5.47979006595  + 1155.60497489050 *T)
    sumY4  +=       2.643 * cos(3.09107587932  + 426.84200835950 *T)
    sumY4  +=       2.565 * cos(1.36437340248  + 6.86972951729 *T)
    sumY4  +=       2.154 * cos(0.36845643020  + 618.80046279510 *T)
    sumY4  +=       2.044 * cos(6.19291733804  + 419.72846135871 *T)
    sumY4  +=       1.513 * cos(1.60661037744  + 220.65645992230 *T)
    sumY4  +=       1.619 * cos(3.76198603224  + 633.02755679670 *T)
    sumY4  +=       1.503 * cos(1.66604601675  + 2119.00767786191 *T)
    sumY4  +=       1.422 * cos(1.64339956175  + 227.77000692311 *T)
    sumY4  +=       1.448 * cos(5.34787531868  + 1575.08961876571 *T)
    sumY4  +=       1.354 * cos(5.08770092894  + 956.53297345411 *T)
    sumY4  +=       1.398 * cos(6.23305755326  + 301.92095817150 *T)
    sumY4  +=       1.551 * cos(0.94639398642  + 640.14110379750 *T)
    sumY4  +=       1.193 * cos(4.12010673320  + 853.44019923550 *T)
    sumY4  +=       1.228 * cos(4.89609039918  + 536.56069461190 *T)
    sumY4  +=       1.153 * cos(5.81136535588  + 426.35437339250 *T)
    sumY4  +=       1.111 * cos(0.25913441118  + 309.52214013930 *T)
    sumY4  +=       1.322 * cos(6.23999851102  + 412.61491435790 *T)
    sumY4  +=       1.137 * cos(0.06266558324  + 1258.69774910910 *T)
    sumY4  +=       0.987 * cos(0.30137359463  + 522.33360061030 *T)
    sumY4  +=       0.937 * cos(2.83311778330  + 618.31282782810 *T)
    sumY4  +=       0.830 * cos(4.12976646136  + 412.12727939089 *T)
    sumY4  +=       0.868 * cos(5.39178303584  + 1038.28510667030 *T)
    sumY4  +=       0.819 * cos(2.43009257653  + 742.74624304909 *T)
    sumY4  +=       0.952 * cos(1.65716723019  + 831.61192326610 *T)
    sumY4  +=       0.903 * cos(3.71462655064  + 949.41942645330 *T)
    sumY4  +=       0.740 * cos(5.82657968137  + 832.09955823310 *T)
    sumY4  +=       0.726 * cos(5.81781448666  + 839.21310523390 *T)
    sumY4  +=       0.890 * cos(1.13102640201  + 647.25465079831 *T)
    sumY4  +=       0.672 * cos(5.39661133415  + 102.84895673509 *T)
    sumY4  +=       0.598 * cos(2.46149662725  + 337.97632814250 *T)
    sumY4  +=       0.711 * cos(5.36999317897  + 515.22005360950 *T)
    sumY4  +=       0.557 * cos(0.72441048007  + 1486.22393854870 *T)
    sumY4  +=       0.506 * cos(3.86590138433  + 1059.13811270570 *T)
    sumY4  +=       0.464 * cos(2.90086100642  + 117.07605073670 *T)
    sumY4  +=       0.479 * cos(4.66931827522  + 526.75338884040 *T)
    sumY4  +=       0.451 * cos(5.32820518820  + 1471.99684454710 *T)
    sumY4  +=       0.466 * cos(5.95847666696  + 533.11617631580 *T)
    sumY4  +=       0.396 * cos(2.33304510363  + 1603.54380676890 *T)

    # JUPITER Y Coord
    sumY5  = 0
    sumY5  +=     131.509 * cos(5.80959663173  + 537.04832957890 *T)
    sumY5  +=     126.839 * cos(1.57486395680  + 522.82123557730 *T)
    sumY5  +=      48.033 * cos(1.61168112243  + 515.70768857651 *T)
    sumY5  +=      34.241 * cos(5.82558762592  + 544.16187657970 *T)
    sumY5  +=      11.284 * cos(0.56803103740  + 1066.73929467350 *T)
    sumY5  +=       5.229 * cos(0.87528666787  + 0.24381748350 *T)
    sumY5  +=       3.712 * cos(1.71517319353  + 1052.51220067191 *T)
    sumY5  +=       4.457 * cos(3.71675991204  + 529.93478257810 *T)
    sumY5  +=       1.949 * cos(0.16900749445  + 1073.85284167431 *T)
    sumY5  +=       1.738 * cos(3.42162831595  + 1059.62574767270 *T)
    sumY5  +=       1.716 * cos(1.48767323100  + 1045.39865367110 *T)
    sumY5  +=       0.879 * cos(1.64959635605  + 508.59414157570 *T)
    sumY5  +=       0.886 * cos(2.65648931803  + 551.27542358051 *T)
    sumY5  +=       0.665 * cos(0.00022824942  + 7.35736448430 *T)
    sumY5  +=       0.466 * cos(1.28798211401  + 1596.43025976811 *T)

    # JUPITER Z Coord
    sumZ0  = 0
    sumZ0  +=  11823100.489 * cos(3.55844646343  + 529.69096509460 *T)
    sumZ0  +=  859031.952 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ0  +=  286562.094 * cos(3.90812238338  + 1059.38193018920 *T)
    sumZ0  +=   42388.592 * cos(3.60144191032  + 522.57741809380 *T)
    sumZ0  +=   33295.491 * cos(0.30297050585  + 536.80451209540 *T)
    sumZ0  +=   10416.160 * cos(4.25764593061  + 1589.07289528380 *T)
    sumZ0  +=    7449.294 * cos(5.24213104150  + 103.09277421860 *T)
    sumZ0  +=    6910.102 * cos(1.75032945752  + 7.11354700080 *T)
    sumZ0  +=    5292.012 * cos(1.68231447192  + 426.59819087600 *T)
    sumZ0  +=    4313.598 * cos(3.70673689841  + 213.29909543800 *T)
    sumZ0  +=    3784.265 * cos(2.71522544491  + 110.20632121940 *T)
    sumZ0  +=    3798.016 * cos(2.16715743175  + 632.78373931320 *T)
    sumZ0  +=    2455.385 * cos(2.96904135659  + 1052.26838318840 *T)
    sumZ0  +=    2461.547 * cos(2.99889460411  + 1162.47470440780 *T)
    sumZ0  +=    2001.451 * cos(2.68535838309  + 419.48464387520 *T)
    sumZ0  +=    2163.471 * cos(6.26718259854  + 846.08283475120 *T)
    sumZ0  +=    1633.653 * cos(0.64194743493  + 1066.49547719000 *T)
    sumZ0  +=    1450.672 * cos(1.17108416193  + 949.17560896980 *T)
    sumZ0  +=     693.095 * cos(5.14278041161  + 316.39186965660 *T)
    sumZ0  +=     715.042 * cos(1.41211197820  + 323.50541665740 *T)
    sumZ0  +=     549.720 * cos(4.84164274378  + 742.99006053260 *T)
    sumZ0  +=     447.831 * cos(4.60746588621  + 2118.76386037840 *T)
    sumZ0  +=     543.619 * cos(3.69636561822  + 515.46387109300 *T)
    sumZ0  +=     483.852 * cos(4.71796110160  + 735.87651353180 *T)
    sumZ0  +=     365.100 * cos(6.06065925437  + 206.18554843720 *T)
    sumZ0  +=     288.980 * cos(0.40266293090  + 525.75881183150 *T)
    sumZ0  +=     254.938 * cos(0.17978560944  + 543.91805909620 *T)
    sumZ0  +=     195.137 * cos(3.35374802237  + 1692.16566950240 *T)
    sumZ0  +=     253.254 * cos(3.57485315487  + 533.62311835770 *T)
    sumZ0  +=     196.691 * cos(2.65526923569  + 956.28915597060 *T)
    sumZ0  +=     154.379 * cos(1.47050669470  + 625.67019231240 *T)
    sumZ0  +=     145.972 * cos(5.96112903396  + 838.96928775040 *T)
    sumZ0  +=     162.371 * cos(3.12160596807  + 1581.95934828300 *T)
    sumZ0  +=     135.954 * cos(2.23721600610  + 1478.86657406440 *T)
    sumZ0  +=     135.376 * cos(4.03180075278  + 1265.56747862640 *T)
    sumZ0  +=     118.734 * cos(6.19132554294  + 532.87235883230 *T)
    sumZ0  +=     121.147 * cos(4.06252466827  + 526.50957135690 *T)
    sumZ0  +=     113.963 * cos(0.34443034869  + 1375.77379984580 *T)
    sumZ0  +=     133.095 * cos(1.61283648080  + 14.22709400160 *T)
    sumZ0  +=      89.416 * cos(0.98707919437  + 1596.18644228460 *T)
    sumZ0  +=      91.948 * cos(5.29521569721  + 95.97922721780 *T)
    sumZ0  +=      71.808 * cos(2.60370257221  + 117.31986822020 *T)
    sumZ0  +=      63.527 * cos(0.97603653343  + 433.71173787680 *T)
    sumZ0  +=      64.121 * cos(4.63524391670  + 1368.66025284500 *T)
    sumZ0  +=      59.116 * cos(0.42038566170  + 220.41264243880 *T)
    sumZ0  +=      60.054 * cos(1.05875226211  + 942.06206196900 *T)
    sumZ0  +=      54.868 * cos(6.14868344705  + 1169.58825140860 *T)
    sumZ0  +=      47.058 * cos(2.39069042869  + 853.19638175200 *T)
    sumZ0  +=      56.617 * cos(5.78644786940  + 639.89728631400 *T)
    sumZ0  +=      41.028 * cos(5.07649815801  + 309.27832265580 *T)
    sumZ0  +=      54.980 * cos(2.72414041857  + 1045.15483618760 *T)
    sumZ0  +=      39.937 * cos(2.63014962920  + 380.12776796000 *T)
    sumZ0  +=      34.887 * cos(3.06644858707  + 412.37109687440 *T)
    sumZ0  +=      33.112 * cos(2.21084011524  + 74.78159856730 *T)
    sumZ0  +=      24.753 * cos(4.09311535367  + 527.24328453980 *T)
    sumZ0  +=      21.473 * cos(5.68324848411  + 1272.68102562720 *T)
    sumZ0  +=      21.259 * cos(4.15648418623  + 1685.05212250160 *T)
    sumZ0  +=      21.214 * cos(4.95749714252  + 2648.45482547300 *T)
    sumZ0  +=      19.531 * cos(3.58685387193  + 529.73914920440 *T)
    sumZ0  +=      24.578 * cos(6.16442414965  + 532.13864564940 *T)
    sumZ0  +=      21.217 * cos(4.46071227915  + 1155.36115740700 *T)
    sumZ0  +=      21.279 * cos(3.98105217713  + 1795.25844372100 *T)
    sumZ0  +=      19.533 * cos(0.38904271302  + 529.64278098480 *T)
    sumZ0  +=      19.570 * cos(4.37171598624  + 149.56319713460 *T)
    sumZ0  +=      19.447 * cos(1.27663598466  + 330.61896365820 *T)
    sumZ0  +=      15.440 * cos(0.72260027140  + 530.65417294110 *T)
    sumZ0  +=      14.562 * cos(5.50135998822  + 1471.75302706360 *T)
    sumZ0  +=      17.624 * cos(1.31667827871  + 3.93215326310 *T)
    sumZ0  +=      13.309 * cos(2.77897533854  + 2008.55753915900 *T)
    sumZ0  +=      13.206 * cos(3.64976502017  + 2221.85663459700 *T)
    sumZ0  +=      13.142 * cos(0.72687086889  + 1055.44977692610 *T)
    sumZ0  +=      15.413 * cos(4.31958059396  + 1063.31408345230 *T)
    sumZ0  +=      14.193 * cos(3.67387943991  + 2111.65031337760 *T)
    sumZ0  +=      12.996 * cos(1.31000927869  + 453.42489381900 *T)
    sumZ0  +=      11.738 * cos(6.14331082365  + 199.07200143640 *T)
    sumZ0  +=      15.482 * cos(3.25291782438  + 528.72775724810 *T)
    sumZ0  +=      10.389 * cos(0.51378228980  + 1073.60902419080 *T)
    sumZ0  +=      12.990 * cos(1.87234255615  + 984.60033162190 *T)
    sumZ0  +=      12.280 * cos(1.99030236725  + 38.13303563780 *T)
    sumZ0  +=      10.473 * cos(0.26689008608  + 10213.28554621100 *T)
    sumZ0  +=       8.875 * cos(0.13793367408  + 305.34616939270 *T)
    sumZ0  +=       7.692 * cos(5.08407965373  + 1788.14489672020 *T)
    sumZ0  +=       7.861 * cos(3.79832245754  + 76.26607127560 *T)
    sumZ0  +=       8.714 * cos(1.58825029218  + 3.18139373770 *T)
    sumZ0  +=       7.143 * cos(4.01912438773  + 2001.44399215820 *T)
    sumZ0  +=       6.752 * cos(5.50960892986  + 1898.35121793960 *T)
    sumZ0  +=       7.441 * cos(1.30439352743  + 618.55664531160 *T)
    sumZ0  +=       6.896 * cos(1.74025919863  + 529.16970023280 *T)
    sumZ0  +=       7.365 * cos(3.26347830221  + 1485.98012106520 *T)
    sumZ0  +=       6.518 * cos(0.90320354569  + 1905.46476494040 *T)
    sumZ0  +=       8.347 * cos(5.96531853825  + 909.81873305460 *T)
    sumZ0  +=       7.728 * cos(4.61604781975  + 728.76296653100 *T)
    sumZ0  +=       6.887 * cos(2.23549334986  + 530.21222995640 *T)
    sumZ0  +=       5.783 * cos(6.02338915633  + 831.85574074960 *T)
    sumZ0  +=       5.788 * cos(0.26042181508  + 1062.56332392690 *T)
    sumZ0  +=       5.581 * cos(6.17489228666  + 1258.45393162560 *T)
    sumZ0  +=       5.154 * cos(1.33592252764  + 2125.87740737920 *T)
    sumZ0  +=       4.916 * cos(0.18728472832  + 1699.27921650320 *T)
    sumZ0  +=       5.480 * cos(3.12818768952  + 604.47256366190 *T)
    sumZ0  +=       4.967 * cos(3.92882210242  + 508.35032409220 *T)
    sumZ0  +=       4.994 * cos(0.03902921434  + 1439.50969814920 *T)
    sumZ0  +=       5.269 * cos(4.06576160798  + 1056.20053645150 *T)
    sumZ0  +=       4.711 * cos(0.87975326503  + 490.33408917940 *T)
    sumZ0  +=       4.221 * cos(2.57304081485  + 99.16062095550 *T)
    sumZ0  +=       4.005 * cos(2.96062737592  + 526.77020378780 *T)
    sumZ0  +=       3.959 * cos(1.38163787157  + 532.61172640140 *T)
    sumZ0  +=       4.844 * cos(1.47818569722  + 454.90936652730 *T)
    sumZ0  +=       3.683 * cos(0.33555577415  + 528.94020556920 *T)
    sumZ0  +=       3.681 * cos(4.00539820601  + 530.44172462000 *T)
    sumZ0  +=       3.623 * cos(0.28674357330  + 227.52618943960 *T)
    sumZ0  +=       4.295 * cos(3.45202038167  + 569.04784100980 *T)
    sumZ0  +=       3.458 * cos(4.40401766789  + 519.39602435610 *T)
    sumZ0  +=       3.435 * cos(5.34191505443  + 11.04570026390 *T)
    sumZ0  +=       3.355 * cos(4.80063248866  + 2104.53676637680 *T)
    sumZ0  +=       3.535 * cos(1.18645749024  + 525.49817940060 *T)
    sumZ0  +=       3.477 * cos(3.15607509055  + 533.88375078860 *T)
    sumZ0  +=       3.219 * cos(5.37130793285  + 2214.74308759620 *T)
    sumZ0  +=       4.358 * cos(2.16300555603  + 1021.24889455140 *T)
    sumZ0  +=       3.334 * cos(0.38640772450  + 224.34479570190 *T)
    sumZ0  +=       2.697 * cos(6.22459411775  + 539.98590583310 *T)
    sumZ0  +=       3.143 * cos(0.15836271478  + 750.10360753340 *T)
    sumZ0  +=       2.649 * cos(3.80434469443  + 2317.83586181480 *T)
    sumZ0  +=       2.856 * cos(0.29274516197  + 983.11585891360 *T)
    sumZ0  +=       3.424 * cos(5.16456871383  + 1.48447270830 *T)
    sumZ0  +=       2.725 * cos(0.64710955846  + 524.06189080210 *T)
    sumZ0  +=       2.623 * cos(0.25280896703  + 551.03160609700 *T)
    sumZ0  +=       2.685 * cos(3.69478866795  + 535.32003938710 *T)
    sumZ0  +=       2.939 * cos(0.64629595693  + 540.73666535850 *T)
    sumZ0  +=       2.398 * cos(4.36615854926  + 2428.04218303420 *T)
    sumZ0  +=       2.597 * cos(1.15824869762  + 963.40270297140 *T)
    sumZ0  +=       2.769 * cos(0.20040217534  + 525.02509864860 *T)
    sumZ0  +=       2.336 * cos(3.71676131146  + 529.85102378900 *T)
    sumZ0  +=       2.336 * cos(0.62392433691  + 529.53090640020 *T)
    sumZ0  +=       2.491 * cos(2.84297287947  + 1382.88734684660 *T)
    sumZ0  +=       2.598 * cos(4.13924061941  + 534.35683154060 *T)
    sumZ0  +=       2.747 * cos(1.76471074936  + 524.27433912320 *T)
    sumZ0  +=       2.475 * cos(0.99313692673  + 934.94851496820 *T)
    sumZ0  +=       2.704 * cos(2.57406093768  + 535.10759106600 *T)
    sumZ0  +=       2.302 * cos(5.84295144929  + 1148.24761040620 *T)
    sumZ0  +=       2.068 * cos(5.96196623799  + 440.82528487760 *T)
    sumZ0  +=       2.012 * cos(4.95946680265  + 1361.54670584420 *T)
    sumZ0  +=       1.987 * cos(0.18375525793  + 1802.37199072180 *T)
    sumZ0  +=       1.934 * cos(2.82066156035  + 835.03713448730 *T)
    sumZ0  +=       2.147 * cos(5.96724060411  + 1891.23767093880 *T)
    sumZ0  +=       1.940 * cos(3.63687343812  + 2324.94940881560 *T)
    sumZ0  +=       1.763 * cos(2.57382733259  + 415.55249061210 *T)
    sumZ0  +=       1.860 * cos(5.24230405559  + 2531.13495725280 *T)
    sumZ0  +=       1.707 * cos(5.38883058569  + 302.16477565500 *T)
    sumZ0  +=       2.152 * cos(1.74557444220  + 21.34064100240 *T)
    sumZ0  +=       1.613 * cos(3.45966511571  + 529.47851677350 *T)
    sumZ0  +=       1.613 * cos(0.88102053266  + 529.90341341570 *T)
    sumZ0  +=       2.134 * cos(2.79657990895  + 35.42472265210 *T)
    sumZ0  +=       1.593 * cos(1.37312370449  + 1574.84580128220 *T)
    sumZ0  +=       1.510 * cos(3.41611736454  + 518.64526483070 *T)
    sumZ0  +=       1.579 * cos(3.96836848379  + 2420.92863603340 *T)
    sumZ0  +=       1.630 * cos(1.31973482281  + 39.35687591520 *T)
    sumZ0  +=       1.598 * cos(2.53043088824  + 511.53171782990 *T)
    sumZ0  +=       1.841 * cos(5.72475961436  + 531.17543780290 *T)
    sumZ0  +=       1.537 * cos(1.84394223501  + 547.85021235930 *T)
    sumZ0  +=       1.771 * cos(4.87234280655  + 528.20649238630 *T)
    sumZ0  +=       1.656 * cos(1.27834742762  + 173.94221952280 *T)
    sumZ0  +=       1.242 * cos(1.65064670901  + 643.82943957710 *T)
    sumZ0  +=       1.302 * cos(3.54050271463  + 3340.61242669980 *T)
    sumZ0  +=       1.236 * cos(5.41985106155  + 107.02492748170 *T)
    sumZ0  +=       1.165 * cos(1.48429677606  + 960.22130923370 *T)
    sumZ0  +=       1.166 * cos(5.12888892088  + 1048.33622992530 *T)
    sumZ0  +=       1.153 * cos(4.78828764413  + 514.71311156760 *T)
    sumZ0  +=       1.153 * cos(4.22743390789  + 2744.43405269080 *T)
    sumZ0  +=       1.478 * cos(3.73912254261  + 230.56457082540 *T)
    sumZ0  +=       1.121 * cos(6.27688038656  + 387.24131496080 *T)
    sumZ0  +=       1.112 * cos(5.83569926578  + 544.66881862160 *T)
    sumZ0  +=       1.170 * cos(1.05508911315  + 1512.80682400820 *T)
    sumZ0  +=       1.113 * cos(5.74217831509  + 3480.31056622260 *T)
    sumZ0  +=       1.507 * cos(4.16985097397  + 1364.72809958190 *T)
    sumZ0  +=       1.120 * cos(5.02238271412  + 1464.63948006280 *T)
    sumZ0  +=       1.232 * cos(2.83088412944  + 2538.24850425360 *T)
    sumZ0  +=       1.179 * cos(4.62538441124  + 1056.93424963440 *T)
    sumZ0  +=       1.278 * cos(3.86396842379  + 92.04707395470 *T)
    sumZ0  +=       1.094 * cos(2.73378417498  + 299.12639426920 *T)
    sumZ0  +=       1.345 * cos(0.63863230611  + 1061.82961074400 *T)
    sumZ0  +=       0.977 * cos(3.50831165685  + 2.44768055480 *T)
    sumZ0  +=       1.323 * cos(4.03638019689  + 494.26624244250 *T)
    sumZ0  +=       1.310 * cos(5.45458264580  + 1057.89745748090 *T)
    sumZ0  +=       0.966 * cos(3.51363341772  + 2634.22773147140 *T)
    sumZ0  +=       1.053 * cos(5.49394122920  + 3178.14579056760 *T)
    sumZ0  +=       0.984 * cos(4.22960042766  + 430.53034413910 *T)
    sumZ0  +=       0.937 * cos(3.77590974564  + 2641.34127847220 *T)
    sumZ0  +=       0.932 * cos(4.11879245377  + 1059.43011429900 *T)
    sumZ0  +=       0.941 * cos(0.30660455816  + 70.84944530420 *T)
    sumZ0  +=       0.925 * cos(1.29091785470  + 2435.15573003500 *T)
    sumZ0  +=       0.950 * cos(3.88567620136  + 3377.21779200400 *T)
    sumZ0  +=       1.051 * cos(5.62777473038  + 2207.62954059540 *T)
    sumZ0  +=       1.020 * cos(1.77520625500  + 312.45971639350 *T)
    sumZ0  +=       0.946 * cos(4.13718116942  + 1677.93857550080 *T)
    sumZ0  +=       1.080 * cos(2.32932073464  + 106.27416795630 *T)
    sumZ0  +=       0.890 * cos(0.03582895580  + 647.01083331480 *T)
    sumZ0  +=       1.086 * cos(2.80157071056  + 1038.04128918680 *T)
    sumZ0  +=       1.158 * cos(0.32499630540  + 565.11568774670 *T)
    sumZ0  +=       1.098 * cos(1.00937080256  + 523.54062594030 *T)
    sumZ0  +=       0.932 * cos(0.92098132614  + 1059.33374607940 *T)
    sumZ0  +=       0.821 * cos(3.41471589329  + 2950.61960112800 *T)
    sumZ0  +=       0.929 * cos(2.30521635468  + 1514.29129671650 *T)
    sumZ0  +=       0.915 * cos(1.48306891526  + 10.29494073850 *T)
    sumZ0  +=       0.911 * cos(3.32547809355  + 18.15924726470 *T)
    sumZ0  +=       0.786 * cos(2.82286526802  + 1134.16352875650 *T)
    sumZ0  +=       0.954 * cos(0.13401891198  + 3906.90875709860 *T)
    sumZ0  +=       0.994 * cos(3.33159283750  + 535.84130424890 *T)
    sumZ0  +=       0.792 * cos(5.10203109941  + 2524.02141025200 *T)
    sumZ0  +=       0.824 * cos(1.74472748730  + 529.95159752550 *T)
    sumZ0  +=       0.824 * cos(2.59595816107  + 529.43033266370 *T)
    sumZ0  +=       0.762 * cos(4.30564108086  + 672.14061522840 *T)
    sumZ0  +=       0.872 * cos(5.03498959444  + 679.25416222920 *T)
    sumZ0  +=       0.839 * cos(3.34536065236  + 124.43341522100 *T)
    sumZ0  +=       0.927 * cos(4.91765336511  + 1593.00504854690 *T)
    sumZ0  +=       0.826 * cos(5.05738343335  + 2847.52682690940 *T)
    sumZ0  +=       0.738 * cos(0.70435921398  + 522.52923398400 *T)

    # JUPITER Z Coord
    sumZ1  = 0
    sumZ1  +=  922338.114 * cos(5.70129376981  + 529.69096509460 *T)
    sumZ1  +=   16158.151 * cos(5.47088355549  + 522.57741809380 *T)
    sumZ1  +=   11443.757 * cos(4.73284385916  + 536.80451209540 *T)
    sumZ1  +=   13239.917 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ1  +=    8389.696 * cos(5.77702424773  + 1059.38193018920 *T)
    sumZ1  +=    1518.380 * cos(6.17585226995  + 7.11354700080 *T)
    sumZ1  +=     909.709 * cos(4.72812859336  + 1052.26838318840 *T)
    sumZ1  +=     620.246 * cos(2.72253251427  + 103.09277421860 *T)
    sumZ1  +=     608.445 * cos(5.18778213162  + 1066.49547719000 *T)
    sumZ1  +=     532.577 * cos(3.42162653671  + 632.78373931320 *T)
    sumZ1  +=     380.990 * cos(0.66271432082  + 419.48464387520 *T)
    sumZ1  +=     388.254 * cos(5.49417723877  + 515.46387109300 *T)
    sumZ1  +=     298.589 * cos(2.90253153613  + 949.17560896980 *T)
    sumZ1  +=     349.922 * cos(0.54665309808  + 110.20632121940 *T)
    sumZ1  +=     279.305 * cos(3.88961739871  + 1589.07289528380 *T)
    sumZ1  +=     361.932 * cos(5.53643296927  + 213.29909543800 *T)
    sumZ1  +=     225.182 * cos(3.89570879874  + 735.87651353180 *T)
    sumZ1  +=     187.607 * cos(4.69299591719  + 543.91805909620 *T)
    sumZ1  +=     179.235 * cos(5.12740491355  + 1162.47470440780 *T)
    sumZ1  +=     200.039 * cos(6.12017192088  + 316.39186965660 *T)
    sumZ1  +=     182.523 * cos(1.80179763504  + 846.08283475120 *T)
    sumZ1  +=     144.548 * cos(5.86543306509  + 323.50541665740 *T)
    sumZ1  +=      86.282 * cos(5.89687025782  + 526.50957135690 *T)
    sumZ1  +=      89.546 * cos(5.65296765520  + 533.62311835770 *T)
    sumZ1  +=      81.378 * cos(6.14717900879  + 14.22709400160 *T)
    sumZ1  +=      74.353 * cos(2.14164008505  + 206.18554843720 *T)
    sumZ1  +=      69.673 * cos(4.30083628252  + 532.87235883230 *T)
    sumZ1  +=      62.471 * cos(4.30070044341  + 525.75881183150 *T)
    sumZ1  +=      57.223 * cos(1.37578902299  + 426.59819087600 *T)
    sumZ1  +=      53.295 * cos(4.89194899977  + 1581.95934828300 *T)
    sumZ1  +=      40.084 * cos(3.76141485240  + 625.67019231240 *T)
    sumZ1  +=      36.247 * cos(5.63125611734  + 1596.18644228460 *T)
    sumZ1  +=      35.548 * cos(1.83858046513  + 956.28915597060 *T)
    sumZ1  +=      41.340 * cos(4.43813420705  + 1045.15483618760 *T)
    sumZ1  +=      31.563 * cos(0.93258298806  + 95.97922721780 *T)
    sumZ1  +=      33.208 * cos(0.84342977892  + 117.31986822020 *T)
    sumZ1  +=      30.354 * cos(3.48783919024  + 639.89728631400 *T)
    sumZ1  +=      30.115 * cos(5.67382431748  + 1265.56747862640 *T)
    sumZ1  +=      25.977 * cos(2.95191149293  + 942.06206196900 *T)
    sumZ1  +=      33.745 * cos(3.93467189724  + 2118.76386037840 *T)
    sumZ1  +=      22.626 * cos(5.48825527865  + 433.71173787680 *T)
    sumZ1  +=      19.878 * cos(4.91628562475  + 220.41264243880 *T)
    sumZ1  +=      17.659 * cos(4.23943297681  + 1169.58825140860 *T)
    sumZ1  +=      18.816 * cos(4.88638541834  + 1155.36115740700 *T)
    sumZ1  +=      13.710 * cos(0.75869975183  + 853.19638175200 *T)
    sumZ1  +=      11.629 * cos(5.82902179758  + 330.61896365820 *T)
    sumZ1  +=       9.953 * cos(3.77150415749  + 1478.86657406440 *T)
    sumZ1  +=       9.435 * cos(5.19056014088  + 412.37109687440 *T)
    sumZ1  +=       7.949 * cos(5.09235140984  + 1073.60902419080 *T)
    sumZ1  +=       7.394 * cos(3.77882877747  + 742.99006053260 *T)
    sumZ1  +=       6.394 * cos(1.73926334857  + 199.07200143640 *T)
    sumZ1  +=       7.268 * cos(1.08212079652  + 309.27832265580 *T)
    sumZ1  +=       6.126 * cos(5.65302722173  + 508.35032409220 *T)
    sumZ1  +=       5.709 * cos(5.24271332802  + 1692.16566950240 *T)
    sumZ1  +=       4.446 * cos(0.13248846755  + 1063.31408345230 *T)
    sumZ1  +=       4.664 * cos(6.01570799148  + 1795.25844372100 *T)
    sumZ1  +=       4.631 * cos(2.23071927945  + 1368.66025284500 *T)
    sumZ1  +=       5.404 * cos(4.55754877469  + 728.76296653100 *T)
    sumZ1  +=       3.944 * cos(6.15098833747  + 519.39602435610 *T)
    sumZ1  +=       4.113 * cos(6.19598758673  + 3.18139373770 *T)
    sumZ1  +=       3.543 * cos(3.09506092197  + 618.55664531160 *T)
    sumZ1  +=       3.392 * cos(1.77987055903  + 1375.77379984580 *T)
    sumZ1  +=       3.322 * cos(4.96703910975  + 1062.56332392690 *T)
    sumZ1  +=       3.774 * cos(5.36462989641  + 2111.65031337760 *T)
    sumZ1  +=       3.799 * cos(5.83406327157  + 1056.20053645150 *T)
    sumZ1  +=       2.813 * cos(4.52082617477  + 1272.68102562720 *T)
    sumZ1  +=       2.782 * cos(4.79754662680  + 551.03160609700 *T)
    sumZ1  +=       2.602 * cos(4.23853778386  + 1574.84580128220 *T)
    sumZ1  +=       2.563 * cos(4.89379767599  + 227.52618943960 *T)
    sumZ1  +=       2.684 * cos(4.21379358590  + 2648.45482547300 *T)
    sumZ1  +=       2.532 * cos(4.43605994513  + 539.98590583310 *T)
    sumZ1  +=       2.233 * cos(3.23226667541  + 1685.05212250160 *T)
    sumZ1  +=       2.724 * cos(4.04040213922  + 74.78159856730 *T)
    sumZ1  +=       2.366 * cos(4.85623283641  + 1055.44977692610 *T)
    sumZ1  +=       2.278 * cos(2.57618076227  + 1485.98012106520 *T)
    sumZ1  +=       2.080 * cos(0.06312365899  + 2125.87740737920 *T)
    sumZ1  +=       2.460 * cos(1.72405831407  + 532.13864564940 *T)
    sumZ1  +=       1.956 * cos(2.23190744786  + 529.64278098480 *T)
    sumZ1  +=       1.955 * cos(5.42971857629  + 529.73914920440 *T)
    sumZ1  +=       2.026 * cos(1.92698506631  + 525.02509864860 *T)
    sumZ1  +=       1.820 * cos(6.00083773693  + 1258.45393162560 *T)
    sumZ1  +=       2.478 * cos(5.93625415892  + 527.24328453980 *T)
    sumZ1  +=       1.744 * cos(4.12331874845  + 440.82528487760 *T)
    sumZ1  +=       2.234 * cos(0.54491713117  + 380.12776796000 *T)
    sumZ1  +=       2.181 * cos(6.14968267326  + 149.56319713460 *T)
    sumZ1  +=       1.629 * cos(4.73547025109  + 1699.27921650320 *T)
    sumZ1  +=       1.576 * cos(5.79923230523  + 838.96928775040 *T)
    sumZ1  +=       1.718 * cos(1.47385390454  + 831.85574074960 *T)
    sumZ1  +=       2.078 * cos(0.02343346752  + 21.34064100240 *T)
    sumZ1  +=       1.415 * cos(6.10139082852  + 2428.04218303420 *T)
    sumZ1  +=       1.549 * cos(5.09589374634  + 528.72775724810 *T)
    sumZ1  +=       1.415 * cos(2.35346960340  + 534.35683154060 *T)
    sumZ1  +=       1.305 * cos(2.13001780938  + 302.16477565500 *T)
    sumZ1  +=       1.546 * cos(2.56584512211  + 530.65417294110 *T)
    sumZ1  +=       1.225 * cos(4.81644489422  + 535.10759106600 *T)
    sumZ1  +=       1.227 * cos(4.63162956792  + 526.77020378780 *T)
    sumZ1  +=       1.475 * cos(3.64456015927  + 984.60033162190 *T)
    sumZ1  +=       1.121 * cos(3.02317463559  + 2001.44399215820 *T)
    sumZ1  +=       1.035 * cos(3.65466665401  + 1471.75302706360 *T)
    sumZ1  +=       1.161 * cos(4.51697446536  + 1038.04128918680 *T)
    sumZ1  +=       1.060 * cos(3.43347467903  + 3.93215326310 *T)
    sumZ1  +=       1.007 * cos(4.35426216497  + 511.53171782990 *T)
    sumZ1  +=       1.025 * cos(2.87748450885  + 2317.83586181480 *T)
    sumZ1  +=       1.112 * cos(4.47223399610  + 750.10360753340 *T)
    sumZ1  +=       1.226 * cos(3.74723115522  + 38.13303563780 *T)
    sumZ1  +=       0.876 * cos(5.40141749419  + 524.27433912320 *T)
    sumZ1  +=       0.809 * cos(1.79288461783  + 2214.74308759620 *T)
    sumZ1  +=       0.788 * cos(5.94278119191  + 10.29494073850 *T)
    sumZ1  +=       0.931 * cos(1.41779899532  + 909.81873305460 *T)
    sumZ1  +=       0.741 * cos(1.37687026697  + 1382.88734684660 *T)
    sumZ1  +=       0.670 * cos(6.15206333598  + 547.85021235930 *T)
    sumZ1  +=       0.688 * cos(1.04145468936  + 1905.46476494040 *T)
    sumZ1  +=       0.757 * cos(5.66593038440  + 76.26607127560 *T)
    sumZ1  +=       0.882 * cos(4.49029761988  + 604.47256366190 *T)
    sumZ1  +=       0.690 * cos(3.49297492409  + 529.16970023280 *T)
    sumZ1  +=       0.689 * cos(3.98820790955  + 530.21222995640 *T)
    sumZ1  +=       0.819 * cos(2.95686029502  + 454.90936652730 *T)
    sumZ1  +=       0.570 * cos(1.96990183237  + 1898.35121793960 *T)
    sumZ1  +=       0.690 * cos(0.18267926862  + 6283.07584999140 *T)
    sumZ1  +=       0.663 * cos(4.55436664519  + 305.34616939270 *T)
    sumZ1  +=       0.477 * cos(2.44863388631  + 490.33408917940 *T)
    sumZ1  +=       0.398 * cos(4.50217443276  + 2008.55753915900 *T)
    sumZ1  +=       0.375 * cos(1.90578907085  + 528.94020556920 *T)
    sumZ1  +=       0.374 * cos(5.57619428876  + 530.44172462000 *T)

    # JUPITER Z Coord
    sumZ2  = 0
    sumZ2  +=   42203.037 * cos(1.45965899387  + 529.69096509460 *T)
    sumZ2  +=    6345.355 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ2  +=    3895.469 * cos(0.95407663893  + 522.57741809380 *T)
    sumZ2  +=    2065.004 * cos(2.89817590660  + 536.80451209540 *T)
    sumZ2  +=     893.912 * cos(1.44537531474  + 1059.38193018920 *T)
    sumZ2  +=     193.882 * cos(0.40346918252  + 1052.26838318840 *T)
    sumZ2  +=     177.093 * cos(4.24934061763  + 7.11354700080 *T)
    sumZ2  +=     154.339 * cos(0.98627236886  + 515.46387109300 *T)
    sumZ2  +=     119.867 * cos(3.48023127478  + 1066.49547719000 *T)
    sumZ2  +=      71.483 * cos(2.91929643864  + 543.91805909620 *T)
    sumZ2  +=      58.309 * cos(5.21153342767  + 632.78373931320 *T)
    sumZ2  +=      49.489 * cos(1.90363556247  + 1589.07289528380 *T)
    sumZ2  +=      33.191 * cos(4.84947257364  + 949.17560896980 *T)
    sumZ2  +=      28.865 * cos(4.59812610654  + 110.20632121940 *T)
    sumZ2  +=      26.279 * cos(1.43562959449  + 526.50957135690 *T)
    sumZ2  +=      25.332 * cos(4.38538322886  + 14.22709400160 *T)
    sumZ2  +=      17.862 * cos(1.38653335267  + 533.62311835770 *T)
    sumZ2  +=      16.652 * cos(2.14142688267  + 316.39186965660 *T)
    sumZ2  +=      18.475 * cos(0.44768511310  + 419.48464387520 *T)
    sumZ2  +=      15.818 * cos(1.42318045904  + 103.09277421860 *T)
    sumZ2  +=      20.087 * cos(6.06937699639  + 735.87651353180 *T)
    sumZ2  +=      15.944 * cos(3.98909127597  + 323.50541665740 *T)
    sumZ2  +=      16.057 * cos(6.20772505653  + 1045.15483618760 *T)
    sumZ2  +=      12.863 * cos(4.00884384040  + 426.59819087600 *T)
    sumZ2  +=      15.906 * cos(2.39783235711  + 532.87235883230 *T)
    sumZ2  +=      13.892 * cos(0.94813323601  + 213.29909543800 *T)
    sumZ2  +=      10.214 * cos(0.99254309899  + 1162.47470440780 *T)
    sumZ2  +=      10.216 * cos(0.51784128572  + 1581.95934828300 *T)
    sumZ2  +=       7.926 * cos(4.00901193840  + 1596.18644228460 *T)
    sumZ2  +=       7.954 * cos(3.77963260387  + 846.08283475120 *T)
    sumZ2  +=       9.013 * cos(2.24077766707  + 525.75881183150 *T)
    sumZ2  +=       9.087 * cos(5.73971133695  + 625.67019231240 *T)
    sumZ2  +=       6.668 * cos(4.76731608799  + 942.06206196900 *T)
    sumZ2  +=       6.568 * cos(2.86210728523  + 95.97922721780 *T)
    sumZ2  +=       7.860 * cos(5.36569633982  + 117.31986822020 *T)
    sumZ2  +=       6.892 * cos(4.23401086250  + 206.18554843720 *T)
    sumZ2  +=       5.090 * cos(0.36658695963  + 1155.36115740700 *T)
    sumZ2  +=       4.274 * cos(3.72888580140  + 433.71173787680 *T)
    sumZ2  +=       3.753 * cos(3.16074830315  + 220.41264243880 *T)
    sumZ2  +=       3.669 * cos(0.07721721390  + 956.28915597060 *T)
    sumZ2  +=       3.591 * cos(1.27574344579  + 508.35032409220 *T)
    sumZ2  +=       3.432 * cos(5.12494175250  + 742.99006053260 *T)
    sumZ2  +=       3.262 * cos(1.85564958119  + 639.89728631400 *T)
    sumZ2  +=       3.246 * cos(2.42226150944  + 1169.58825140860 *T)
    sumZ2  +=       3.879 * cos(2.47671375339  + 2118.76386037840 *T)
    sumZ2  +=       2.719 * cos(3.57668078721  + 1073.60902419080 *T)
    sumZ2  +=       3.564 * cos(6.14378298704  + 728.76296653100 *T)
    sumZ2  +=       2.630 * cos(1.72012208968  + 1265.56747862640 *T)
    sumZ2  +=       2.800 * cos(4.30633268838  + 330.61896365820 *T)
    sumZ2  +=       2.361 * cos(5.36600615712  + 853.19638175200 *T)
    sumZ2  +=       2.261 * cos(1.65056479007  + 519.39602435610 *T)
    sumZ2  +=       2.143 * cos(3.59903064599  + 199.07200143640 *T)
    sumZ2  +=       2.064 * cos(0.72183122308  + 838.96928775040 *T)
    sumZ2  +=       2.327 * cos(1.57523321082  + 412.37109687440 *T)
    sumZ2  +=       1.483 * cos(5.22462726605  + 1478.86657406440 *T)
    sumZ2  +=       1.392 * cos(3.02125335643  + 551.03160609700 *T)
    sumZ2  +=       1.187 * cos(1.24411330169  + 1258.45393162560 *T)
    sumZ2  +=       1.238 * cos(2.56643737684  + 539.98590583310 *T)
    sumZ2  +=       1.019 * cos(4.56433477067  + 21.34064100240 *T)
    sumZ2  +=       0.787 * cos(3.45332815915  + 227.52618943960 *T)
    sumZ2  +=       0.789 * cos(4.40210167179  + 3.18139373770 *T)
    sumZ2  +=       0.933 * cos(2.99359495481  + 309.27832265580 *T)
    sumZ2  +=       0.668 * cos(2.62794838154  + 440.82528487760 *T)
    sumZ2  +=       0.701 * cos(3.25895071673  + 1062.56332392690 *T)
    sumZ2  +=       0.655 * cos(2.81138161833  + 1485.98012106520 *T)
    sumZ2  +=       0.530 * cos(1.07172768238  + 1692.16566950240 *T)
    sumZ2  +=       0.491 * cos(1.99562035757  + 1368.66025284500 *T)
    sumZ2  +=       0.470 * cos(0.26830712431  + 1272.68102562720 *T)

    # JUPITER Z Coord
    sumZ3  = 0
    sumZ3  +=    1302.949 * cos(3.38307505290  + 529.69096509460 *T)
    sumZ3  +=     638.849 * cos(2.73076320051  + 522.57741809380 *T)
    sumZ3  +=     252.188 * cos(1.03735897029  + 536.80451209540 *T)
    sumZ3  +=      42.424 * cos(2.76335793506  + 515.46387109300 *T)
    sumZ3  +=      28.787 * cos(2.30679311994  + 1052.26838318840 *T)
    sumZ3  +=      18.636 * cos(1.12686228293  + 543.91805909620 *T)
    sumZ3  +=      16.037 * cos(1.78194169646  + 1066.49547719000 *T)
    sumZ3  +=      21.893 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ3  +=      15.010 * cos(2.24882166949  + 7.11354700080 *T)
    sumZ3  +=      18.770 * cos(4.26444493597  + 1059.38193018920 *T)
    sumZ3  +=       5.506 * cos(2.63836917983  + 14.22709400160 *T)
    sumZ3  +=       4.515 * cos(1.81963459999  + 1045.15483618760 *T)
    sumZ3  +=       4.499 * cos(0.43487771424  + 632.78373931320 *T)
    sumZ3  +=       3.150 * cos(2.90272527852  + 110.20632121940 *T)
    sumZ3  +=       2.649 * cos(3.67990759642  + 419.48464387520 *T)
    sumZ3  +=       2.448 * cos(5.72936768843  + 103.09277421860 *T)
    sumZ3  +=       3.025 * cos(1.52033973511  + 735.87651353180 *T)
    sumZ3  +=       2.257 * cos(0.17284429054  + 949.17560896980 *T)
    sumZ3  +=       2.354 * cos(3.18232334886  + 526.50957135690 *T)
    sumZ3  +=       1.614 * cos(2.98017326384  + 508.35032409220 *T)
    sumZ3  +=       1.671 * cos(4.40642025933  + 532.87235883230 *T)
    sumZ3  +=       1.483 * cos(1.64073529408  + 625.67019231240 *T)
    sumZ3  +=       1.665 * cos(0.44117653607  + 1589.07289528380 *T)
    sumZ3  +=       1.469 * cos(3.11087801399  + 533.62311835770 *T)
    sumZ3  +=       1.412 * cos(0.26761069900  + 942.06206196900 *T)
    sumZ3  +=       1.270 * cos(0.70083660931  + 426.59819087600 *T)
    sumZ3  +=       0.899 * cos(2.62337071525  + 1596.18644228460 *T)
    sumZ3  +=       0.951 * cos(4.34670868038  + 525.75881183150 *T)
    sumZ3  +=       0.973 * cos(6.07278030998  + 639.89728631400 *T)
    sumZ3  +=       0.829 * cos(2.20068062832  + 1581.95934828300 *T)
    sumZ3  +=       1.015 * cos(3.67348757646  + 117.31986822020 *T)
    sumZ3  +=       0.849 * cos(1.53481384034  + 323.50541665740 *T)
    sumZ3  +=       0.681 * cos(1.96048278513  + 1073.60902419080 *T)
    sumZ3  +=       0.665 * cos(3.82697658502  + 316.39186965660 *T)
    sumZ3  +=       0.609 * cos(4.70141431381  + 95.97922721780 *T)
    sumZ3  +=       0.518 * cos(4.61207781194  + 2118.76386037840 *T)
    sumZ3  +=       0.512 * cos(5.55216613299  + 433.71173787680 *T)
    sumZ3  +=       0.459 * cos(3.26874502411  + 213.29909543800 *T)
    sumZ3  +=       0.401 * cos(2.27671168160  + 1162.47470440780 *T)

    # JUPITER Z Coord
    sumZ4  = 0
    sumZ4  +=      78.889 * cos(4.52764322985  + 522.57741809380 *T)
    sumZ4  +=      28.644 * cos(4.45318845059  + 529.69096509460 *T)
    sumZ4  +=      26.536 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ4  +=      23.071 * cos(5.44321158713  + 536.80451209540 *T)
    sumZ4  +=       9.686 * cos(4.52165553359  + 515.46387109300 *T)
    sumZ4  +=       3.848 * cos(5.59513666444  + 543.91805909620 *T)
    sumZ4  +=       3.620 * cos(4.19174561337  + 1052.26838318840 *T)
    sumZ4  +=       0.945 * cos(5.96517767658  + 7.11354700080 *T)
    sumZ4  +=       0.957 * cos(6.25072054635  + 1066.49547719000 *T)
    sumZ4  +=       0.722 * cos(3.64496080495  + 1059.38193018920 *T)
    sumZ4  +=       0.582 * cos(1.16718383135  + 14.22709400160 *T)

    # JUPITER Z Coord
    sumZ5  = 0
    sumZ5  +=       7.519 * cos(0.09198554072  + 522.57741809380 *T)
    sumZ5  +=       1.913 * cos(0.00874408003  + 515.46387109300 *T)
    sumZ5  +=       1.582 * cos(3.27902945138  + 536.80451209540 *T)
    sumZ5  +=       0.704 * cos(0.27563461391  + 529.69096509460 *T)
    sumZ5  +=       0.495 * cos(1.29305954542  + 543.91805909620 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Jupiter",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getJupiterData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getJupiterXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getMarsXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Mars"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getMarsXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getMarsXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # MARS X Coord
    sumX0  = 0
    sumX0  +=  151664432.758 * cos(6.20347631684  + 3340.85624418330 *T)
    sumX0  +=  21337343.470 * cos(2.72390342700  + 0.24381748350 *T)
    sumX0  +=  7067734.657 * cos(0.25841679630  + 6681.46867088311 *T)
    sumX0  +=  494034.875 * cos(0.59654023167  + 10022.08109758290 *T)
    sumX0  +=  135189.385 * cos(0.59603904502  + 3340.36860921629 *T)
    sumX0  +=   40927.830 * cos(0.93466338314  + 13362.69352428270 *T)
    sumX0  +=   21051.988 * cos(1.80377821704  + 3337.33312583431 *T)
    sumX0  +=   21027.270 * cos(1.17839785680  + 3344.37936253230 *T)
    sumX0  +=   11370.204 * cos(4.83238126700  + 1059.62574767270 *T)
    sumX0  +=   13426.076 * cos(0.63015591517  + 529.93478257810 *T)
    sumX0  +=    6769.403 * cos(3.61729412618  + 3340.83899053111 *T)
    sumX0  +=    6769.355 * cos(5.64806575768  + 3340.87349783550 *T)
    sumX0  +=    8285.925 * cos(1.86109008374  + 2280.98667902709 *T)
    sumX0  +=    5433.910 * cos(1.11806362109  + 2942.70724077510 *T)
    sumX0  +=    4813.330 * cos(1.85023048147  + 3739.00524759150 *T)
    sumX0  +=    4893.493 * cos(4.42760397790  + 5622.08674069391 *T)
    sumX0  +=    5269.759 * cos(2.33019561512  + 6151.77770578850 *T)
    sumX0  +=    3774.767 * cos(0.82510516838  + 6680.98103591609 *T)
    sumX0  +=    3623.045 * cos(6.10862866873  + 796.54182429990 *T)
    sumX0  +=    3725.058 * cos(1.27278649428  + 16703.30595098250 *T)
    sumX0  +=    3767.289 * cos(1.22204844262  + 398.39282089170 *T)
    sumX0  +=    2396.177 * cos(0.63496836160  + 3149.40797807170 *T)
    sumX0  +=    2262.756 * cos(2.34867489134  + 3532.30451029490 *T)
    sumX0  +=    1976.674 * cos(2.14058651048  + 6677.94555253410 *T)
    sumX0  +=    2227.632 * cos(1.69533548136  + 3340.78893388051 *T)
    sumX0  +=    2227.573 * cos(1.28683909451  + 3340.92355448610 *T)
    sumX0  +=    1950.645 * cos(2.72809117200  + 2544.55823736690 *T)
    sumX0  +=    2179.897 * cos(1.69534718656  + 6283.31966747490 *T)
    sumX0  +=    1684.046 * cos(3.14928536452  + 5885.17066406670 *T)
    sumX0  +=    2237.564 * cos(4.82103877835  + 8962.69916739370 *T)
    sumX0  +=    1629.438 * cos(0.24065025766  + 4137.15425099970 *T)
    sumX0  +=    1957.286 * cos(1.51885757913  + 6684.99178923210 *T)
    sumX0  +=    1382.161 * cos(2.17182492289  + 1751.29571393250 *T)
    sumX0  +=    1364.523 * cos(0.46059571034  + 2811.16527908870 *T)
    sumX0  +=    1285.835 * cos(4.70544160337  + 1194.69082770810 *T)
    sumX0  +=    1468.951 * cos(1.87682106061  + 3870.54720927791 *T)
    sumX0  +=    1572.210 * cos(4.84784531890  + 1589.31671276730 *T)
    sumX0  +=     955.879 * cos(5.37058897391  + 426.84200835950 *T)
    sumX0  +=     819.304 * cos(5.15875122140  + 4400.23817437250 *T)
    sumX0  +=     714.429 * cos(2.92303461187  + 191.20444862810 *T)
    sumX0  +=     719.663 * cos(2.65486751128  + 9492.39013248830 *T)
    sumX0  +=     643.260 * cos(4.55789424001  + 5487.02166065850 *T)
    sumX0  +=     795.172 * cos(4.39660517598  + 5092.39577559930 *T)
    sumX0  +=     630.906 * cos(3.95541777233  + 6681.45141723091 *T)
    sumX0  +=     630.906 * cos(5.98618940383  + 6681.48592453530 *T)
    sumX0  +=     782.173 * cos(1.16801410137  + 5091.90814063229 *T)
    sumX0  +=     703.756 * cos(0.96657545612  + 213.54291292150 *T)
    sumX0  +=     581.847 * cos(3.95231836766  + 3185.43584474911 *T)
    sumX0  +=     553.454 * cos(5.31409832853  + 3496.27664361750 *T)
    sumX0  +=     663.460 * cos(3.74354794256  + 2544.07060239989 *T)
    sumX0  +=     491.963 * cos(2.17574433513  + 7079.61767429131 *T)
    sumX0  +=     426.327 * cos(5.14926594152  + 2145.92159899169 *T)
    sumX0  +=     533.910 * cos(3.30010366565  + 1592.83983111630 *T)
    sumX0  +=     386.953 * cos(5.07337650750  + 8433.00820229911 *T)
    sumX0  +=     417.478 * cos(3.96425682190  + 3339.87592311510 *T)
    sumX0  +=     417.367 * cos(5.30109299779  + 3341.83656525151 *T)
    sumX0  +=     359.961 * cos(1.61089747813  + 20043.91837768230 *T)
    sumX0  +=     352.641 * cos(3.88893523415  + 6254.87048000710 *T)
    sumX0  +=     351.819 * cos(5.16992059902  + 12303.31159409350 *T)
    sumX0  +=     321.320 * cos(1.94435133813  + 3554.15533962130 *T)
    sumX0  +=     377.309 * cos(4.13476329642  + 2146.40923395870 *T)
    sumX0  +=     307.274 * cos(0.94421931529  + 3127.55714874530 *T)
    sumX0  +=     284.287 * cos(2.03419702203  + 7.35736448430 *T)
    sumX0  +=     259.684 * cos(1.89385284662  + 1990.98883452450 *T)
    sumX0  +=     246.886 * cos(0.27164096909  + 1747.77259558350 *T)
    sumX0  +=     313.688 * cos(5.10599974151  + 4535.30325440790 *T)
    sumX0  +=     271.483 * cos(1.94778284200  + 6468.16957544510 *T)
    sumX0  +=     298.122 * cos(5.96616996596  + 5088.87265725030 *T)
    sumX0  +=     223.194 * cos(3.55284314213  + 3320.08084869090 *T)
    sumX0  +=     221.461 * cos(5.71228247661  + 3361.63163967570 *T)
    sumX0  +=     233.899 * cos(2.03689241948  + 9623.93209417470 *T)
    sumX0  +=     207.617 * cos(2.03345912239  + 6681.40136058031 *T)
    sumX0  +=     205.148 * cos(1.85775607260  + 10025.60421593190 *T)
    sumX0  +=     207.784 * cos(2.47827795124  + 10018.55797923390 *T)
    sumX0  +=     193.198 * cos(4.93193682847  + 2119.00767786191 *T)
    sumX0  +=     218.630 * cos(0.97526221043  + 6490.02040477150 *T)
    sumX0  +=     247.990 * cos(1.31792500524  + 2913.77041834030 *T)
    sumX0  +=     207.617 * cos(1.62496273555  + 6681.53598118590 *T)
    sumX0  +=     178.587 * cos(2.54120991611  + 1221.60474883790 *T)
    sumX0  +=     178.400 * cos(1.02824622233  + 10021.59346261590 *T)
    sumX0  +=     164.293 * cos(1.09050742498  + 4690.72365384210 *T)
    sumX0  +=     157.091 * cos(0.48700512124  + 2389.13783793270 *T)
    sumX0  +=     151.308 * cos(1.43281018748  + 4562.21717553770 *T)
    sumX0  +=     177.027 * cos(4.27007713386  + 2957.95971196010 *T)
    sumX0  +=     147.685 * cos(2.47506585562  + 7211.15963597770 *T)
    sumX0  +=     164.646 * cos(0.57567045734  + 7477.76667769950 *T)
    sumX0  +=     153.884 * cos(4.44526715317  + 640.14110379750 *T)
    sumX0  +=     157.989 * cos(4.99618947427  + 3723.75277640650 *T)
    sumX0  +=     146.757 * cos(4.41545990951  + 3337.26581553151 *T)
    sumX0  +=     166.434 * cos(3.16925511951  + 10213.52936369450 *T)
    sumX0  +=     146.132 * cos(4.84659253705  + 3344.44667283510 *T)
    sumX0  +=     118.082 * cos(4.29228989356  + 4562.70481050471 *T)
    sumX0  +=     119.770 * cos(4.65884659981  + 11773.62062899890 *T)
    sumX0  +=     113.695 * cos(2.49810201078  + 4292.57465043390 *T)
    sumX0  +=     143.954 * cos(5.36191839973  + 2787.28684134090 *T)
    sumX0  +=     113.828 * cos(3.49219829771  + 3333.74269718250 *T)
    sumX0  +=     112.131 * cos(0.07204419182  + 3205.79116414790 *T)
    sumX0  +=     144.684 * cos(1.67820772946  + 1349.62359217529 *T)
    sumX0  +=     118.108 * cos(5.18561000649  + 4929.92913946711 *T)
    sumX0  +=     106.690 * cos(2.90949586529  + 3475.92132421871 *T)
    sumX0  +=     120.237 * cos(3.40929466701  + 9225.78309076650 *T)
    sumX0  +=     119.599 * cos(3.90588399352  + 3894.42564702570 *T)
    sumX0  +=     104.910 * cos(5.51977893032  + 382.65271473970 *T)
    sumX0  +=      95.920 * cos(5.81540533813  + 3347.96979118411 *T)
    sumX0  +=      93.778 * cos(5.54256544882  + 1748.26023055050 *T)
    sumX0  +=      88.754 * cos(2.98682152461  + 12833.00255918811 *T)
    sumX0  +=     112.794 * cos(5.89797835701  + 155.17658195069 *T)
    sumX0  +=      91.596 * cos(3.08454019471  + 951.47458876710 *T)
    sumX0  +=      86.667 * cos(0.43803962703  + 3098.12764020930 *T)
    sumX0  +=     118.500 * cos(0.52380467505  + 7902.82960223750 *T)
    sumX0  +=      89.906 * cos(2.29494981588  + 2942.21960580809 *T)
    sumX0  +=      85.205 * cos(4.21443205483  + 2281.47431399410 *T)
    sumX0  +=     113.660 * cos(5.58557229219  + 5486.53402569149 *T)
    sumX0  +=      80.640 * cos(2.54197906469  + 3583.58484815730 *T)
    sumX0  +=      85.344 * cos(3.68498025243  + 4933.45225781610 *T)
    sumX0  +=      84.447 * cos(0.97974209820  + 6674.35512388230 *T)
    sumX0  +=      79.513 * cos(0.70956398510  + 5088.38502228330 *T)
    sumX0  +=     101.901 * cos(0.62374863015  + 529.44714761109 *T)
    sumX0  +=      73.054 * cos(5.03109665468  + 1052.51220067191 *T)
    sumX0  +=      73.997 * cos(5.88971113539  + 3767.45443505930 *T)
    sumX0  +=      70.372 * cos(4.49111053246  + 553.32558535889 *T)
    sumX0  +=      91.916 * cos(4.26029980883  + 9595.48290670690 *T)
    sumX0  +=      73.607 * cos(5.48588623719  + 7740.85060107230 *T)
    sumX0  +=      66.068 * cos(1.60352414710  + 1066.73929467350 *T)
    sumX0  +=      81.219 * cos(4.00417692850  + 2810.67764412170 *T)
    sumX0  +=      66.430 * cos(2.30115866241  + 2700.47132290230 *T)
    sumX0  +=      54.145 * cos(4.29245819665  + 6526.04827144891 *T)
    sumX0  +=      59.768 * cos(4.81363180968  + 6041.57138456910 *T)
    sumX0  +=      66.173 * cos(0.04112658538  + 10022.09835123510 *T)
    sumX0  +=      49.883 * cos(5.09321137459  + 3312.40705671550 *T)
    sumX0  +=      49.178 * cos(4.17234313530  + 3369.30543165110 *T)
    sumX0  +=      48.202 * cos(5.03755934935  + 397.90518592470 *T)
    sumX0  +=      47.310 * cos(5.51367266321  + 15643.92402079330 *T)
    sumX0  +=      48.401 * cos(1.00254354626  + 3.76693583251 *T)
    sumX0  +=      66.173 * cos(4.29354026107  + 10022.06384393071 *T)
    sumX0  +=      43.167 * cos(2.11330967332  + 4690.23601887509 *T)
    sumX0  +=      52.329 * cos(5.35612779574  + 7903.31723720450 *T)
    sumX0  +=      43.258 * cos(1.94634380811  + 149.80701461810 *T)
    sumX0  +=      53.318 * cos(2.50854201221  + 10420.23010099111 *T)
    sumX0  +=      38.925 * cos(5.63931982315  + 6682.44899195131 *T)
    sumX0  +=      43.058 * cos(2.71172733627  + 3325.60377299830 *T)
    sumX0  +=      42.835 * cos(0.27036084223  + 3356.10871536831 *T)
    sumX0  +=      49.042 * cos(5.09197877772  + 2273.87313202630 *T)
    sumX0  +=      45.897 * cos(1.73176272729  + 2288.10022602790 *T)
    sumX0  +=      36.267 * cos(1.94902636545  + 23384.53080438210 *T)
    sumX0  +=      50.027 * cos(4.92232111087  + 5621.59910572690 *T)
    sumX0  +=      41.152 * cos(3.60849241104  + 537.04832957890 *T)
    sumX0  +=      38.722 * cos(0.77097024087  + 522.82123557730 *T)
    sumX0  +=      44.751 * cos(4.31225718691  + 3077.77232081051 *T)
    sumX0  +=      31.695 * cos(4.22324584554  + 5629.20028769470 *T)
    sumX0  +=      31.478 * cos(2.26370333180  + 5331.60126122430 *T)
    sumX0  +=      33.349 * cos(2.12149635572  + 242.97242145751 *T)
    sumX0  +=      36.130 * cos(2.27448966151  + 9808.78200214491 *T)
    sumX0  +=      31.660 * cos(5.11359333219  + 2648.69864295650 *T)
    sumX0  +=      39.058 * cos(4.30336342995  + 6680.48834981490 *T)
    sumX0  +=      32.374 * cos(5.44019764272  + 7875.91568110771 *T)
    sumX0  +=      38.493 * cos(2.35050388066  + 6894.76776632111 *T)
    sumX0  +=      40.044 * cos(4.95433661511  + 3603.94016755610 *T)
    sumX0  +=      27.893 * cos(0.75588690511  + 7373.13863714289 *T)
    sumX0  +=      27.665 * cos(2.37934374634  + 12964.54452087450 *T)
    sumX0  +=      27.878 * cos(0.66739091469  + 1350.11122714230 *T)
    sumX0  +=      28.322 * cos(2.53961838260  + 8030.84844557490 *T)
    sumX0  +=      28.486 * cos(1.37297559963  + 5614.97319369310 *T)
    sumX0  +=      29.761 * cos(5.41923539807  + 3178.38960805111 *T)
    sumX0  +=      35.657 * cos(1.13413038335  + 8428.99744898310 *T)
    sumX0  +=      28.159 * cos(3.83299371524  + 3503.32288031550 *T)
    sumX0  +=      27.295 * cos(5.85818083120  + 213.05527795450 *T)
    sumX0  +=      24.083 * cos(4.10582287293  + 4033.01384541010 *T)
    sumX0  +=      23.726 * cos(3.79203465807  + 3320.50092478451 *T)
    sumX0  +=      23.543 * cos(5.47205540117  + 3361.21156358210 *T)
    sumX0  +=      24.407 * cos(1.62000405809  + 76.50988875911 *T)
    sumX0  +=      22.360 * cos(4.86162540597  + 15114.23305569870 *T)
    sumX0  +=      22.681 * cos(2.19545260317  + 13366.21664263171 *T)
    sumX0  +=      30.767 * cos(1.97333726282  + 102.84895673509 *T)
    sumX0  +=      23.116 * cos(4.24443125854  + 1692.40948698591 *T)
    sumX0  +=      24.795 * cos(2.94450891135  + 691.91378374330 *T)
    sumX0  +=      21.819 * cos(1.96227087233  + 10022.14840788570 *T)
    sumX0  +=      21.420 * cos(3.90434583630  + 6660.69327539070 *T)
    sumX0  +=      20.686 * cos(4.09172464650  + 1582.20316576650 *T)
    sumX0  +=      25.165 * cos(0.35004555829  + 3335.33331987590 *T)
    sumX0  +=      21.819 * cos(2.37076725917  + 10022.01378728011 *T)
    sumX0  +=      25.110 * cos(2.63051920012  + 3346.37916849070 *T)
    sumX0  +=      22.778 * cos(1.31432064900  + 9830.63283147130 *T)
    sumX0  +=      21.875 * cos(4.80561875471  + 11243.92966390430 *T)
    sumX0  +=      19.487 * cos(2.47923244785  + 3336.84549086730 *T)
    sumX0  +=      25.048 * cos(2.47380350738  + 1648.20293971390 *T)
    sumX0  +=      19.285 * cos(0.96932439937  + 3330.21957883350 *T)
    sumX0  +=      19.206 * cos(2.01113330365  + 3351.49290953311 *T)
    sumX0  +=      19.577 * cos(0.82781613472  + 6438.74006690910 *T)
    sumX0  +=      23.713 * cos(0.78696317220  + 796.05418933290 *T)
    sumX0  +=      18.024 * cos(0.80164082178  + 3.27930086550 *T)
    sumX0  +=      17.457 * cos(0.90670155624  + 6872.91693699471 *T)
    sumX0  +=      20.758 * cos(2.68199279076  + 2914.25805330730 *T)
    sumX0  +=      23.325 * cos(2.81224294003  + 1758.40926093329 *T)
    sumX0  +=      19.762 * cos(1.19465593585  + 2818.27882608950 *T)
    sumX0  +=      18.903 * cos(1.70198303484  + 3863.43366227710 *T)
    sumX0  +=      18.548 * cos(1.86435352340  + 3343.89172756530 *T)
    sumX0  +=      20.024 * cos(5.23513746258  + 9382.18381126891 *T)
    sumX0  +=      23.004 * cos(2.81536425224  + 13359.17040593370 *T)
    sumX0  +=      20.686 * cos(6.04843065741  + 6702.24406637550 *T)
    sumX0  +=      17.546 * cos(0.76946873055  + 3074.24920246151 *T)
    sumX0  +=      21.953 * cos(6.16726175116  + 10713.75106384269 *T)
    sumX0  +=      22.478 * cos(3.36629130965  + 3531.81687532789 *T)
    sumX0  +=      17.443 * cos(6.23925536469  + 7063.87756813929 *T)
    sumX0  +=      16.475 * cos(1.75316661727  + 8432.52056733209 *T)
    sumX0  +=      16.284 * cos(4.60826689329  + 6298.57213865991 *T)
    sumX0  +=      16.296 * cos(2.90679614817  + 10551.77206267750 *T)
    sumX0  +=      21.367 * cos(3.52004981807  + 4292.08701546690 *T)
    sumX0  +=      15.862 * cos(5.88391024825  + 3723.26514143949 *T)
    sumX0  +=      16.182 * cos(3.94577566715  + 7632.69944216670 *T)
    sumX0  +=      18.728 * cos(1.73152268876  + 4032.52621044309 *T)
    sumX0  +=      17.711 * cos(0.91113854617  + 10818.37910439930 *T)
    sumX0  +=      16.214 * cos(2.21183941484  + 3607.46328590511 *T)
    sumX0  +=      20.207 * cos(0.50430731047  + 2708.07250487011 *T)
    sumX0  +=      17.851 * cos(5.47859973724  + 6836.88907031730 *T)
    sumX0  +=      14.367 * cos(3.47695884338  + 8429.48508395010 *T)
    sumX0  +=      16.523 * cos(2.14637482599  + 6158.89125278930 *T)
    sumX0  +=      19.608 * cos(3.51602051779  + 853.44019923550 *T)
    sumX0  +=      14.968 * cos(1.03757514888  + 6531.90547374850 *T)
    sumX0  +=      13.797 * cos(2.36391887749  + 419.72846135871 *T)
    sumX0  +=      13.414 * cos(2.94536295838  + 3127.06951377830 *T)
    sumX0  +=      15.245 * cos(2.24543615829  + 7064.36520310630 *T)
    sumX0  +=      15.674 * cos(4.46460349265  + 3333.81000748530 *T)
    sumX0  +=      13.513 * cos(5.18432808372  + 6685.05909953490 *T)
    sumX0  +=      16.177 * cos(0.84378599845  + 5729.75026463250 *T)
    sumX0  +=      16.167 * cos(5.63583810469  + 6144.66415878770 *T)
    sumX0  +=      13.924 * cos(4.60961152326  + 12936.09533340670 *T)
    sumX0  +=      12.758 * cos(2.28807897894  + 3336.97490882530 *T)
    sumX0  +=      16.755 * cos(2.44526644281  + 3973.63998349650 *T)
    sumX0  +=      12.740 * cos(0.69410204881  + 3344.73757954131 *T)
    sumX0  +=      15.216 * cos(5.27437629135  + 5459.62010456170 *T)
    sumX0  +=      13.755 * cos(5.25003721422  + 433.95555536030 *T)
    sumX0  +=      13.727 * cos(5.70869872406  + 6127.89926804070 *T)
    sumX0  +=      16.418 * cos(5.42287303075  + 75.02541605080 *T)
    sumX0  +=      12.379 * cos(3.30493783022  + 2487.17222746430 *T)
    sumX0  +=      12.121 * cos(4.87336947084  + 3980.75353049730 *T)
    sumX0  +=      12.145 * cos(4.64108642144  + 3281.48238226970 *T)
    sumX0  +=      16.915 * cos(0.30896398254  + 10404.48999483909 *T)
    sumX0  +=      13.732 * cos(4.74929587262  + 6677.87824223130 *T)
    sumX0  +=      12.163 * cos(4.37039838160  + 10973.31186886650 *T)
    sumX0  +=      11.839 * cos(4.62276702188  + 3400.23010609691 *T)
    sumX0  +=      13.460 * cos(1.58419502448  + 1596.43025976811 *T)
    sumX0  +=      11.950 * cos(5.43572222924  + 1744.18216693170 *T)
    sumX0  +=      14.031 * cos(4.77979856866  + 8031.33608054190 *T)
    sumX0  +=      13.621 * cos(5.59064545004  + 3415.63784275061 *T)
    sumX0  +=      11.890 * cos(5.01320389678  + 4989.30300138070 *T)
    sumX0  +=      11.133 * cos(5.52124766174  + 8270.54156616690 *T)
    sumX0  +=      11.115 * cos(3.25426586017  + 6816.53375091851 *T)
    sumX0  +=      10.961 * cos(2.86101048703  + 110.45013870291 *T)
    sumX0  +=      12.029 * cos(4.05861575489  + 1751.78334889950 *T)
    sumX0  +=      12.954 * cos(4.66570452942  + 8969.81271439450 *T)
    sumX0  +=      12.975 * cos(3.69069473524  + 3266.07464561600 *T)
    sumX0  +=      13.898 * cos(0.43379973859  + 6546.40359084770 *T)
    sumX0  +=      10.787 * cos(3.68885960945  + 11770.09751064990 *T)
    sumX0  +=      13.676 * cos(4.64950162211  + 3347.90248088131 *T)
    sumX0  +=      14.429 * cos(0.35374653823  + 6041.08374960209 *T)
    sumX0  +=      10.265 * cos(5.15008405880  + 5099.50932260011 *T)
    sumX0  +=      10.855 * cos(0.89301290681  + 3316.97780643551 *T)
    sumX0  +=      13.203 * cos(1.76870797269  + 8955.58562039291 *T)
    sumX0  +=      10.750 * cos(2.08914960461  + 3364.73468193110 *T)
    sumX0  +=      12.106 * cos(3.38174169479  + 6.86972951729 *T)
    sumX0  +=      10.745 * cos(3.32222106136  + 16173.61498588790 *T)
    sumX0  +=      10.862 * cos(6.16671167895  + 7633.18707713370 *T)
    sumX0  +=      13.130 * cos(6.21487916243  + 6688.58221788390 *T)
    sumX0  +=      12.618 * cos(2.02052426773  + 3443.94901840191 *T)
    sumX0  +=      10.348 * cos(0.91493796382  + 5884.68302909969 *T)
    sumX0  +=      11.408 * cos(0.87656413019  + 3237.76346996470 *T)
    sumX0  +=      10.014 * cos(4.92767694376  + 3893.93801205869 *T)
    sumX0  +=      13.330 * cos(2.96193815385  + 11371.46087227470 *T)
    sumX0  +=       9.390 * cos(3.13708066819  + 6151.29007082150 *T)
    sumX0  +=       9.317 * cos(6.04347781776  + 3304.82837750590 *T)
    sumX0  +=       9.588 * cos(3.93497646953  + 14584.54209060410 *T)
    sumX0  +=       9.187 * cos(3.22190702341  + 3376.88411086071 *T)
    sumX0  +=       8.922 * cos(4.01975646099  + 8274.06468451590 *T)
    sumX0  +=      12.186 * cos(1.30797225041  + 10014.96755058210 *T)
    sumX0  +=       8.852 * cos(2.35096921245  + 2921.37160030811 *T)
    sumX0  +=      11.498 * cos(3.39107154427  + 3024.46437452670 *T)
    sumX0  +=       8.932 * cos(2.04695690729  + 3340.43616808970 *T)
    sumX0  +=       8.932 * cos(0.93521735776  + 3341.27632027691 *T)
    sumX0  +=      10.791 * cos(1.20222738595  + 13362.20588931570 *T)
    sumX0  +=      10.440 * cos(5.86847040226  + 3657.24811383991 *T)
    sumX0  +=      10.488 * cos(2.33198909218  + 12168.24651405810 *T)
    sumX0  +=       7.791 * cos(3.37156314108  + 162.22281864870 *T)
    sumX0  +=       8.278 * cos(1.36101802767  + 6605.20259960750 *T)
    sumX0  +=       7.730 * cos(1.32895910157  + 7235.03807372550 *T)
    sumX0  +=      10.793 * cos(2.47932423892  + 135.30889751891 *T)
    sumX0  +=       9.855 * cos(2.07588323732  + 951.96222373410 *T)
    sumX0  +=       7.570 * cos(4.28016049737  + 1861.50203515190 *T)
    sumX0  +=       7.347 * cos(0.73950074439  + 3760.34088805850 *T)
    sumX0  +=       7.458 * cos(5.07772886658  + 11371.94850724170 *T)
    sumX0  +=       7.415 * cos(5.77777265612  + 10575.16286545830 *T)
    sumX0  +=       7.387 * cos(5.61828044516  + 7373.62627210991 *T)
    sumX0  +=       7.330 * cos(0.37925022199  + 13362.71077793490 *T)
    sumX0  +=       7.563 * cos(4.15753542979  + 3134.67069574611 *T)
    sumX0  +=       8.273 * cos(0.81705527879  + 633.02755679670 *T)
    sumX0  +=       9.017 * cos(4.88682687946  + 6784.56144510170 *T)
    sumX0  +=       7.330 * cos(4.63166389767  + 13362.67627063050 *T)
    sumX0  +=       6.892 * cos(3.80196693431  + 3170.18337356410 *T)
    sumX0  +=       9.499 * cos(1.85422005926  + 191.69208359510 *T)
    sumX0  +=       8.390 * cos(3.35818184456  + 12566.39551746630 *T)
    sumX0  +=       7.258 * cos(5.07622910043  + 3547.04179262050 *T)
    sumX0  +=       7.563 * cos(3.43861002197  + 3209.31428249690 *T)
    sumX0  +=       7.609 * cos(5.08370549642  + 3304.25087887911 *T)
    sumX0  +=       7.490 * cos(2.87715807563  + 6924.19727485710 *T)
    sumX0  +=       6.532 * cos(2.77753405872  + 1118.51197461930 *T)
    sumX0  +=       6.446 * cos(3.48151772817  + 3340.42636105660 *T)
    sumX0  +=       6.446 * cos(5.78384184406  + 3341.28612731000 *T)
    sumX0  +=       8.809 * cos(3.95809596395  + 8827.63408735830 *T)
    sumX0  +=       6.453 * cos(5.46311522509  + 3511.52911480250 *T)
    sumX0  +=       7.497 * cos(4.18166418924  + 3377.46160948750 *T)
    sumX0  +=       6.647 * cos(1.41691169841  + 1376.01761732930 *T)
    sumX0  +=       6.444 * cos(2.33688278488  + 1479.11039154791 *T)
    sumX0  +=       6.777 * cos(5.65281173718  + 1194.20319274110 *T)
    sumX0  +=       6.272 * cos(1.03860361018  + 11243.44202893729 *T)
    sumX0  +=       7.647 * cos(5.81483718115  + 11081.46302777211 *T)
    sumX0  +=       6.372 * cos(2.85534407747  + 8827.14645239129 *T)
    sumX0  +=       8.206 * cos(5.35126987297  + 7234.55043875850 *T)
    sumX0  +=       6.106 * cos(0.08823631309  + 10184.06009874809 *T)
    sumX0  +=       6.255 * cos(2.84195845072  + 956.53297345411 *T)
    sumX0  +=       6.895 * cos(5.82543451964  + 3472.39820586971 *T)
    sumX0  +=       5.956 * cos(0.71294472658  + 641.12142486571 *T)
    sumX0  +=       6.342 * cos(5.95803230213  + 1162.71852189130 *T)
    sumX0  +=       8.015 * cos(5.74289052633  + 5828.27228913110 *T)
    sumX0  +=       6.858 * cos(1.37458045220  + 5085.28222859850 *T)
    sumX0  +=       5.564 * cos(4.52500489740  + 949.41942645330 *T)
    sumX0  +=       5.685 * cos(6.19978628309  + 13915.77529215809 *T)
    sumX0  +=       5.989 * cos(2.84289974537  + 13760.84252769090 *T)
    sumX0  +=       6.614 * cos(1.02737793204  + 5099.02168763310 *T)
    sumX0  +=       6.046 * cos(5.85461434376  + 18984.53644749310 *T)
    sumX0  +=       5.388 * cos(3.18378253913  + 1228.71829583870 *T)
    sumX0  +=       6.477 * cos(3.19289948037  + 2177.89390480850 *T)
    sumX0  +=       6.716 * cos(4.42516229875  + 5084.79459363150 *T)
    sumX0  +=       6.834 * cos(3.87407074810  + 6606.68707231580 *T)
    sumX0  +=       6.185 * cos(0.69582667501  + 220.65645992230 *T)
    sumX0  +=       5.372 * cos(1.74198025456  + 2766.51144584850 *T)
    sumX0  +=       6.939 * cos(5.44031621555  + 8329.91542808050 *T)
    sumX0  +=       5.250 * cos(4.43412282387  + 224.58861318540 *T)
    sumX0  +=       5.178 * cos(3.38283339119  + 3274.36883526891 *T)
    sumX0  +=       5.636 * cos(4.62896008864  + 9866.66069814871 *T)
    sumX0  +=       5.018 * cos(5.87780776958  + 3407.34365309770 *T)
    sumX0  +=       6.287 * cos(4.27847352716  + 2222.10045208050 *T)
    sumX0  +=       6.027 * cos(4.29283875964  + 3340.35135556410 *T)
    sumX0  +=       6.581 * cos(5.36035626468  + 38.37685312130 *T)
    sumX0  +=       6.293 * cos(5.74754269848  + 8440.12174929990 *T)
    sumX0  +=       4.882 * cos(2.57565728394  + 10404.97762980610 *T)
    sumX0  +=       4.795 * cos(1.93507580186  + 4407.35172137330 *T)
    sumX0  +=       6.476 * cos(4.86709384177  + 3297.13733187830 *T)
    sumX0  +=       6.362 * cos(4.39851396252  + 3384.57515648830 *T)
    sumX0  +=       5.306 * cos(4.27892955220  + 206.42936592071 *T)
    sumX0  +=       6.484 * cos(0.01882718386  + 7108.06686175911 *T)
    sumX0  +=       4.752 * cos(0.76109516197  + 6261.98402700790 *T)
    sumX0  +=       6.261 * cos(5.30955220110  + 1964.59480937050 *T)
    sumX0  +=       5.039 * cos(4.00989442055  + 6247.75693300630 *T)
    sumX0  +=       6.027 * cos(0.04042508285  + 3340.38586286849 *T)
    sumX0  +=       4.586 * cos(4.50717493600  + 6709.91785835091 *T)
    sumX0  +=       4.635 * cos(1.37708243435  + 263.32774085630 *T)
    sumX0  +=       4.870 * cos(5.44486723246  + 4393.12462737170 *T)
    sumX0  +=       6.261 * cos(4.40252342346  + 26088.14695905770 *T)
    sumX0  +=       4.434 * cos(4.89543652156  + 155.66421691770 *T)
    sumX0  +=       4.383 * cos(3.06060463606  + 6666.21619969810 *T)
    sumX0  +=       4.389 * cos(2.60327063667  + 13149.39442884470 *T)
    sumX0  +=       4.433 * cos(1.50884053543  + 2807.64216073970 *T)
    sumX0  +=       4.980 * cos(1.32305016589  + 5827.78465416409 *T)
    sumX0  +=       4.443 * cos(1.24237319799  + 3915.20104251810 *T)
    sumX0  +=       4.299 * cos(5.46246418866  + 8962.21153242670 *T)
    sumX0  +=       4.356 * cos(2.71657046567  + 10235.38019302090 *T)
    sumX0  +=       4.498 * cos(3.48518081494  + 553.81322032590 *T)
    sumX0  +=       5.019 * cos(2.56129005058  + 3738.51761262449 *T)
    sumX0  +=       4.315 * cos(4.84313966445  + 3877.66075627870 *T)
    sumX0  +=       4.072 * cos(1.76821812939  + 574.10098085129 *T)
    sumX0  +=       4.107 * cos(4.64137843892  + 10021.10077651470 *T)
    sumX0  +=       3.964 * cos(0.04229553627  + 3495.78900865049 *T)
    sumX0  +=       4.785 * cos(1.84722253955  + 2699.97863680110 *T)
    sumX0  +=       5.396 * cos(4.78692961565  + 14313.92429556630 *T)
    sumX0  +=       3.822 * cos(2.21613892883  + 3490.41944131790 *T)
    sumX0  +=       4.095 * cos(5.97738996422  + 10023.06141865111 *T)
    sumX0  +=       3.697 * cos(0.08404470764  + 6571.26234966370 *T)
    sumX0  +=       3.779 * cos(1.32451866449  + 13517.62628874989 *T)
    sumX0  +=       3.860 * cos(0.47636916025  + 6836.40143535029 *T)
    sumX0  +=       4.702 * cos(4.02327796277  + 2111.89413086110 *T)
    sumX0  +=       3.630 * cos(4.72154111022  + 11769.60987568290 *T)
    sumX0  +=       4.499 * cos(5.53792455232  + 13524.67252544789 *T)
    sumX0  +=       3.497 * cos(2.54545971953  + 6282.83203250789 *T)
    sumX0  +=       3.489 * cos(5.77181356746  + 11216.52810780750 *T)
    sumX0  +=       3.591 * cos(1.55498611814  + 3874.07032762691 *T)
    sumX0  +=       3.976 * cos(3.74269735212  + 2804.05173208791 *T)
    sumX0  +=       4.680 * cos(0.19530618850  + 10973.79950383350 *T)
    sumX0  +=       3.961 * cos(1.75399025490  + 6048.68493156990 *T)
    sumX0  +=       3.954 * cos(5.40430940067  + 6653.01948341530 *T)
    sumX0  +=       3.352 * cos(2.05742522098  + 8425.89465529830 *T)
    sumX0  +=       4.016 * cos(1.12722498592  + 3981.73385156551 *T)
    sumX0  +=       3.356 * cos(6.22809815165  + 9168.88471583091 *T)
    sumX0  +=       4.260 * cos(1.53810484916  + 3333.25506221550 *T)
    sumX0  +=       3.936 * cos(1.03505935194  + 6843.44767204830 *T)
    sumX0  +=       3.767 * cos(2.28715050291  + 26725.14323108190 *T)
    sumX0  +=       3.388 * cos(5.15160395842  + 18454.84548239850 *T)
    sumX0  +=       3.337 * cos(3.21108403976  + 13936.55068765050 *T)
    sumX0  +=       3.230 * cos(5.56219394317  + 262.84010588929 *T)
    sumX0  +=       4.019 * cos(0.60853180976  + 6696.72114206811 *T)
    sumX0  +=       3.967 * cos(1.75879114367  + 266.85085920531 *T)
    sumX0  +=       3.932 * cos(3.75548243540  + 2384.07945324570 *T)
    sumX0  +=       3.087 * cos(0.64142856871  + 5202.60209681870 *T)
    sumX0  +=       3.922 * cos(4.95967156091  + 4459.61203628611 *T)
    sumX0  +=       3.717 * cos(1.43162034575  + 2906.65687133950 *T)
    sumX0  +=       3.266 * cos(2.71766260163  + 16305.15694757430 *T)
    sumX0  +=       3.104 * cos(1.55447444218  + 5305.69487103730 *T)
    sumX0  +=       3.000 * cos(5.12142568188  + 10714.23869880971 *T)
    sumX0  +=       3.921 * cos(0.90004199829  + 10177.01386205009 *T)
    sumX0  +=       3.084 * cos(0.06473956323  + 1059.13811270570 *T)
    sumX0  +=       3.253 * cos(3.22894876740  + 5518.99396647530 *T)
    sumX0  +=       2.861 * cos(3.34583941548  + 1331.81107005730 *T)
    sumX0  +=       3.304 * cos(2.59879254413  + 8672.21368792411 *T)
    sumX0  +=       2.975 * cos(1.60725252181  + 10575.65050042530 *T)
    sumX0  +=       2.756 * cos(2.25672622330  + 6040.59106350090 *T)
    sumX0  +=       2.950 * cos(5.25312754456  + 6944.55259425590 *T)
    sumX0  +=       3.005 * cos(5.51425578501  + 2707.58486990309 *T)
    sumX0  +=       2.717 * cos(5.76321921177  + 6519.00203475090 *T)
    sumX0  +=       2.801 * cos(1.74557749980  + 16858.23871544969 *T)
    sumX0  +=       2.897 * cos(3.96845114135  + 2700.95895786930 *T)
    sumX0  +=       3.274 * cos(4.32367822866  + 2920.88396534110 *T)
    sumX0  +=       2.645 * cos(5.46774026264  + 5989.31106965630 *T)
    sumX0  +=       3.228 * cos(4.33146477085  + 647.25465079831 *T)
    sumX0  +=       2.528 * cos(1.65349433627  + 13171.24525817110 *T)
    sumX0  +=       3.531 * cos(4.66472793851  + 3190.80541208169 *T)
    sumX0  +=       2.590 * cos(0.13630377168  + 5725.17951491250 *T)
    sumX0  +=       2.501 * cos(4.63633723841  + 383.14034970670 *T)
    sumX0  +=       2.483 * cos(3.87259614726  + 3313.45468808651 *T)
    sumX0  +=       2.456 * cos(5.39283937660  + 3368.25780028011 *T)
    sumX0  +=       2.450 * cos(4.75373878367  + 5732.29306191330 *T)
    sumX0  +=       2.623 * cos(3.15434704817  + 16699.78283263350 *T)
    sumX0  +=       3.010 * cos(3.37698665041  + 2008.80135664250 *T)
    sumX0  +=       3.148 * cos(1.43176306010  + 3097.64000524229 *T)
    sumX0  +=       2.412 * cos(1.77643225234  + 3561.26888662211 *T)
    sumX0  +=       2.837 * cos(1.97746221043  + 4569.33072253850 *T)
    sumX0  +=       2.416 * cos(2.30120840231  + 13362.76083458550 *T)
    sumX0  +=       2.532 * cos(4.86337013936  + 6475.28312244591 *T)
    sumX0  +=       2.407 * cos(0.34730498643  + 134.82126255190 *T)
    sumX0  +=       2.351 * cos(4.11732696406  + 17925.15451730390 *T)
    sumX0  +=       3.311 * cos(3.56285236974  + 17277.16311435029 *T)
    sumX0  +=       2.384 * cos(5.99225497880  + 9485.27658548750 *T)
    sumX0  +=       2.320 * cos(5.45576909738  + 14712.56093394150 *T)
    sumX0  +=       2.350 * cos(1.08442308251  + 533.45790092711 *T)
    sumX0  +=       2.433 * cos(3.97615373043  + 3171.27606105030 *T)
    sumX0  +=       2.860 * cos(5.20645325605  + 4569.81835750550 *T)
    sumX0  +=       2.382 * cos(2.62777493812  + 15508.85894075790 *T)
    sumX0  +=       2.274 * cos(2.96629550939  + 6686.99159519050 *T)
    sumX0  +=       2.448 * cos(4.59094306357  + 6418.38474751031 *T)
    sumX0  +=       2.223 * cos(2.85609320263  + 10595.93826095070 *T)
    sumX0  +=       2.362 * cos(5.17828434191  + 6843.93530701530 *T)
    sumX0  +=       2.247 * cos(0.76007819691  + 3338.10473357230 *T)
    sumX0  +=       2.247 * cos(2.22209606814  + 3343.60775479431 *T)
    sumX0  +=       2.877 * cos(3.74013026534  + 3870.05957431090 *T)
    sumX0  +=       2.167 * cos(1.68859563420  + 2126.12122486270 *T)
    sumX0  +=       2.861 * cos(1.98905363521  + 3342.34071689161 *T)
    sumX0  +=       2.948 * cos(1.15364316687  + 6414.86162916130 *T)
    sumX0  +=       2.832 * cos(0.98955987530  + 3339.37177147500 *T)
    sumX0  +=       2.257 * cos(3.99764421756  + 14054.85112550950 *T)
    sumX0  +=       2.582 * cos(2.53488965696  + 16706.82906933150 *T)
    sumX0  +=       2.284 * cos(5.29122374255  + 3510.43642731630 *T)
    sumX0  +=       2.213 * cos(1.07850009278  + 3120.44360174450 *T)
    sumX0  +=       2.518 * cos(5.12492489789  + 12722.79623796870 *T)
    sumX0  +=       2.156 * cos(5.25903110334  + 1795.50226120450 *T)
    sumX0  +=       2.120 * cos(0.29007019173  + 14712.07329897449 *T)
    sumX0  +=       2.894 * cos(6.25806113022  + 661.47674426450 *T)
    sumX0  +=       2.197 * cos(2.83017943957  + 7314.25241019630 *T)
    sumX0  +=       2.352 * cos(0.68899979183  + 6675.94574657570 *T)
    sumX0  +=       2.592 * cos(3.56656255149  + 846.32665223470 *T)
    sumX0  +=       2.011 * cos(5.67994909502  + 7800.22446298591 *T)
    sumX0  +=       2.009 * cos(4.22466408986  + 1592.35219614930 *T)
    sumX0  +=       1.985 * cos(1.24704608967  + 14158.99153109910 *T)
    sumX0  +=       2.025 * cos(3.15545503806  + 16460.08971204149 *T)
    sumX0  +=       2.756 * cos(1.37300158683  + 9872.51790044830 *T)
    sumX0  +=       2.586 * cos(4.03791457497  + 15110.70993734970 *T)
    sumX0  +=       1.977 * cos(2.37034987225  + 3340.30129891350 *T)
    sumX0  +=       2.046 * cos(5.79910745940  + 1214.49120183710 *T)
    sumX0  +=       2.049 * cos(1.54711450837  + 1545.11016549530 *T)
    sumX0  +=       2.016 * cos(5.80333241967  + 6701.82399028190 *T)
    sumX0  +=       2.591 * cos(1.78624782563  + 4459.12440131909 *T)
    sumX0  +=       1.944 * cos(1.56628544269  + 793.01870595090 *T)
    sumX0  +=       1.969 * cos(0.98471628518  + 8543.21452351850 *T)
    sumX0  +=       2.221 * cos(4.13093986960  + 6661.11335148431 *T)
    sumX0  +=       2.185 * cos(0.10437662983  + 10042.85649307530 *T)
    sumX0  +=       1.909 * cos(0.37281275566  + 14054.36349054249 *T)
    sumX0  +=       2.355 * cos(2.21625400824  + 6461.05602844430 *T)
    sumX0  +=       1.945 * cos(2.09907033054  + 12296.19804709270 *T)
    sumX0  +=       1.856 * cos(1.97074291339  + 3417.12231545890 *T)
    sumX0  +=       1.964 * cos(3.73559626157  + 3230.64992296391 *T)
    sumX0  +=       1.841 * cos(4.95467577248  + 16276.70776010650 *T)
    sumX0  +=       1.867 * cos(1.87467602822  + 6438.25243194210 *T)
    sumX0  +=       2.124 * cos(1.88480889248  + 14.47091148511 *T)
    sumX0  +=       1.889 * cos(2.84240400113  + 2946.23035912410 *T)
    sumX0  +=       2.273 * cos(4.24795887217  + 10001.30570209050 *T)
    sumX0  +=       1.977 * cos(1.96185346705  + 3340.43591951909 *T)
    sumX0  +=       1.852 * cos(4.40722802585  + 2751.79141717511 *T)
    sumX0  +=       1.888 * cos(2.02744109255  + 3502.83524534850 *T)
    sumX0  +=       1.924 * cos(2.83351941687  + 205.94173095370 *T)
    sumX0  +=       1.900 * cos(2.30804703412  + 9778.86485864190 *T)
    sumX0  +=       1.840 * cos(4.83205314537  + 6457.12387518120 *T)
    sumX0  +=       2.215 * cos(2.48697776679  + 9499.50367948910 *T)
    sumX0  +=       1.806 * cos(2.18544314025  + 2676.10019905331 *T)
    sumX0  +=       1.805 * cos(2.35009306355  + 6692.10533623291 *T)
    sumX0  +=       1.869 * cos(0.41860742717  + 13554.14179039430 *T)
    sumX0  +=       2.416 * cos(2.70970478916  + 13362.62621397991 *T)
    sumX0  +=       2.281 * cos(3.85791196552  + 6643.33563524530 *T)
    sumX0  +=       2.183 * cos(4.94105874519  + 21.01921297590 *T)
    sumX0  +=       1.774 * cos(0.68167607326  + 323.74923414091 *T)
    sumX0  +=       1.748 * cos(4.57092956799  + 5033.02191368570 *T)
    sumX0  +=       1.767 * cos(4.66859088892  + 4555.10362853689 *T)
    sumX0  +=       1.973 * cos(5.03143174563  + 12310.42514109431 *T)
    sumX0  +=       2.109 * cos(2.11500033422  + 7204.04608897691 *T)
    sumX0  +=       2.169 * cos(1.31433847015  + 3148.92034310470 *T)
    sumX0  +=       1.672 * cos(2.55080387851  + 6948.07571260490 *T)
    sumX0  +=       1.685 * cos(5.64872462506  + 1434.90384427590 *T)
    sumX0  +=       2.274 * cos(0.32514053197  + 17256.38771885789 *T)
    sumX0  +=       1.724 * cos(0.89161382376  + 103.33659170210 *T)
    sumX0  +=       1.921 * cos(1.09598172014  + 800.06494264891 *T)
    sumX0  +=       1.651 * cos(5.91048794937  + 6756.25026945040 *T)
    sumX0  +=       1.660 * cos(4.94562345671  + 9639.18456535971 *T)
    sumX0  +=       1.575 * cos(6.09206295741  + 7910.43078420530 *T)
    sumX0  +=       1.556 * cos(5.31271132117  + 743.23387801611 *T)
    sumX0  +=       1.578 * cos(2.33595912935  + 5614.48555872609 *T)
    sumX0  +=       2.080 * cos(2.73283134198  + 13119.47728534169 *T)
    sumX0  +=       2.163 * cos(0.52943676546  + 28.69300495130 *T)
    sumX0  +=       1.532 * cos(0.78840953974  + 4005.61228931331 *T)
    sumX0  +=       1.548 * cos(1.53438288417  + 6578.37589666450 *T)
    sumX0  +=       2.036 * cos(1.14812219907  + 9779.35249360890 *T)
    sumX0  +=       1.534 * cos(5.67314504699  + 2679.62331740231 *T)
    sumX0  +=       1.855 * cos(5.59012601568  + 1039.27042827391 *T)
    sumX0  +=       1.554 * cos(0.58249651167  + 14314.41193053330 *T)
    sumX0  +=       1.555 * cos(0.27389127072  + 10029.19464458370 *T)
    sumX0  +=       1.553 * cos(3.89948887717  + 4194.05262593530 *T)
    sumX0  +=       1.869 * cos(0.58467331713  + 3113.38011139430 *T)
    sumX0  +=       1.547 * cos(1.64100920813  + 13355.57997728191 *T)
    sumX0  +=       1.463 * cos(2.71025544519  + 3362.70707350971 *T)
    sumX0  +=       1.463 * cos(0.27191881986  + 3319.00541485690 *T)
    sumX0  +=       1.468 * cos(0.04316622889  + 632.53992182970 *T)
    sumX0  +=       1.867 * cos(3.32731443087  + 13892.38448937730 *T)
    sumX0  +=       1.483 * cos(5.79145984345  + 2604.97973065150 *T)
    sumX0  +=       1.445 * cos(5.72146800920  + 12994.98156035330 *T)
    sumX0  +=       1.444 * cos(5.09058625219  + 10018.49066893110 *T)
    sumX0  +=       1.862 * cos(1.31686189323  + 6670.83200553330 *T)
    sumX0  +=       1.428 * cos(6.05106658591  + 9468.51169474051 *T)
    sumX0  +=       1.440 * cos(3.90166704313  + 2641.58509595570 *T)
    sumX0  +=       1.456 * cos(4.84813606347  + 3929.92107119150 *T)
    sumX0  +=       1.515 * cos(4.40854499352  + 4922.81559246630 *T)
    sumX0  +=       1.642 * cos(1.18330968138  + 9070.36269133230 *T)
    sumX0  +=       1.501 * cos(0.37274885135  + 664.99986261351 *T)
    sumX0  +=       1.434 * cos(5.73698094815  + 3283.95786924770 *T)
    sumX0  +=       1.756 * cos(5.88702162431  + 4672.91113172411 *T)
    sumX0  +=       1.552 * cos(0.75524999537  + 9887.01601754750 *T)
    sumX0  +=       1.430 * cos(5.52560820233  + 10025.67152623470 *T)
    sumX0  +=       1.459 * cos(2.43643676312  + 2693.35777590150 *T)
    sumX0  +=       1.423 * cos(3.57254753277  + 19800.70213874130 *T)
    sumX0  +=       1.403 * cos(3.52793861890  + 3397.75461911890 *T)
    sumX0  +=       1.451 * cos(2.92581468083  + 13745.59005650590 *T)
    sumX0  +=       1.366 * cos(1.27672056082  + 4555.59126350391 *T)
    sumX0  +=       1.580 * cos(3.47199309807  + 699.02733074409 *T)
    sumX0  +=       1.700 * cos(2.39703980797  + 3568.33237697230 *T)
    sumX0  +=       1.599 * cos(5.02449812818  + 3212.83740084591 *T)
    sumX0  +=       1.492 * cos(3.59408030079  + 6155.30082413750 *T)
    sumX0  +=       1.463 * cos(2.61399249960  + 1015.41920040070 *T)
    sumX0  +=       1.302 * cos(5.67390860842  + 5625.60985904290 *T)
    sumX0  +=       1.329 * cos(6.06855511718  + 3334.17258030920 *T)
    sumX0  +=       1.329 * cos(3.19680445505  + 3347.53990805741 *T)
    sumX0  +=       1.503 * cos(0.26668514170  + 3378.98927982111 *T)
    sumX0  +=       1.397 * cos(3.16602524069  + 588.82100952470 *T)
    sumX0  +=       1.500 * cos(4.30672667001  + 3.68833577960 *T)
    sumX0  +=       1.522 * cos(4.24249365559  + 3468.87508752070 *T)
    sumX0  +=       1.286 * cos(1.87461279962  + 3152.93109642071 *T)
    sumX0  +=       1.600 * cos(1.40701843693  + 1905.70858242391 *T)
    sumX0  +=       1.339 * cos(2.85153623164  + 2699.49100183409 *T)
    sumX0  +=       1.563 * cos(4.74033234784  + 6674.42243418510 *T)
    sumX0  +=       1.455 * cos(5.62085725030  + 8800.23253126150 *T)
    sumX0  +=       1.337 * cos(3.86347754064  + 2277.46356067810 *T)
    sumX0  +=       1.259 * cos(3.26565882972  + 3742.52836594050 *T)
    sumX0  +=       1.315 * cos(0.15620262737  + 1191.16770935910 *T)
    sumX0  +=       1.298 * cos(3.23907368211  + 17395.46355220930 *T)
    sumX0  +=       1.692 * cos(1.98709811396  + 8646.30729773711 *T)
    sumX0  +=       1.571 * cos(5.17275397472  + 3451.06256540270 *T)
    sumX0  +=       1.377 * cos(2.49780736724  + 7255.32583425090 *T)
    sumX0  +=       1.223 * cos(2.13508431892  + 17654.53672226610 *T)
    sumX0  +=       1.272 * cos(6.25716699988  + 6997.86054053970 *T)
    sumX0  +=       1.293 * cos(5.24149508087  + 10125.17387180150 *T)
    sumX0  +=       1.205 * cos(2.87986284577  + 74.53778108379 *T)
    sumX0  +=       1.193 * cos(1.03222295834  + 6685.35000624110 *T)
    sumX0  +=       1.184 * cos(0.91434121954  + 2178.38153977550 *T)
    sumX0  +=       1.175 * cos(3.46872073630  + 4076.73275771510 *T)
    sumX0  +=       1.203 * cos(5.75196130140  + 11670.52785478030 *T)
    sumX0  +=       1.334 * cos(3.53686819132  + 59.13004443009 *T)
    sumX0  +=       1.144 * cos(0.93257993677  + 3191.29304704871 *T)
    sumX0  +=       1.205 * cos(4.89787234183  + 6688.51490758111 *T)
    sumX0  +=       1.228 * cos(1.09370444041  + 2391.68063521350 *T)
    sumX0  +=       1.148 * cos(4.01619565855  + 4886.21022716211 *T)
    sumX0  +=       1.116 * cos(6.08825399135  + 15.49628866851 *T)
    sumX0  +=       1.253 * cos(3.48476988361  + 66.24359143090 *T)
    sumX0  +=       1.122 * cos(5.35041736715  + 4040.12739241091 *T)
    sumX0  +=       1.358 * cos(3.77878000346  + 7895.71605523670 *T)
    sumX0  +=       1.102 * cos(3.98446017631  + 6365.07680122650 *T)
    sumX0  +=       1.406 * cos(4.37889002094  + 9588.36935970610 *T)
    sumX0  +=       1.171 * cos(5.20706500181  + 2772.03437015590 *T)
    sumX0  +=       1.123 * cos(5.00444898311  + 19402.55313533309 *T)
    sumX0  +=       1.128 * cos(4.98002753288  + 6622.09480896950 *T)
    sumX0  +=       1.312 * cos(5.17135511580  + 3311.42673564730 *T)
    sumX0  +=       1.312 * cos(4.09400445643  + 3370.28575271930 *T)
    sumX0  +=       1.101 * cos(5.78974461016  + 3774.56798206011 *T)
    sumX0  +=       1.088 * cos(3.28977902817  + 6148.25458743950 *T)
    sumX0  +=       1.229 * cos(2.53762292335  + 3145.88485972270 *T)
    sumX0  +=       1.354 * cos(0.93200823976  + 4136.66661603269 *T)
    sumX0  +=       1.266 * cos(5.61557722066  + 6887.65421932030 *T)
    sumX0  +=       1.192 * cos(3.10049513646  + 2284.50979737610 *T)
    sumX0  +=       1.059 * cos(5.93800224665  + 1198.21394605711 *T)
    sumX0  +=       1.222 * cos(4.54200893982  + 3184.45552368090 *T)
    sumX0  +=       1.094 * cos(3.07328576578  + 2074.80113058990 *T)
    sumX0  +=       1.086 * cos(4.32081022849  + 2325.19322629910 *T)
    sumX0  +=       1.116 * cos(5.86191156738  + 11611.15399286670 *T)
    sumX0  +=       1.041 * cos(2.09105531099  + 6684.50415426510 *T)
    sumX0  +=       1.105 * cos(6.08678077494  + 5618.56362234490 *T)
    sumX0  +=       1.218 * cos(3.59654741580  + 10157.14617761831 *T)
    sumX0  +=       1.292 * cos(0.40837478006  + 7909.94314923829 *T)
    sumX0  +=       1.193 * cos(2.62619989298  + 6677.58733552510 *T)
    sumX0  +=       1.011 * cos(0.31388757278  + 2494.28577446509 *T)
    sumX0  +=       1.102 * cos(5.26415591186  + 7321.36595719711 *T)
    sumX0  +=       1.001 * cos(1.63432488835  + 11766.50708199810 *T)
    sumX0  +=       1.048 * cos(4.13993269564  + 12721.32828193349 *T)
    sumX0  +=       0.990 * cos(5.95408825469  + 3181.91272640011 *T)
    sumX0  +=       1.146 * cos(4.72386914625  + 3497.25696468570 *T)
    sumX0  +=       0.991 * cos(1.49586957068  + 4140.67736934870 *T)
    sumX0  +=       1.269 * cos(5.19929518288  + 3188.95896309811 *T)
    sumX0  +=       1.254 * cos(4.92591337114  + 6034.45783756830 *T)
    sumX0  +=       1.293 * cos(1.92197905113  + 1.72829019180 *T)
    sumX0  +=       1.206 * cos(5.33211209828  + 6740.84253279671 *T)
    sumX0  +=       1.073 * cos(2.22955826533  + 9389.29735826970 *T)
    sumX0  +=       1.361 * cos(1.69731268075  + 9945.81502630730 *T)
    sumX0  +=       0.977 * cos(1.92768164917  + 3319.52060371631 *T)
    sumX0  +=       0.977 * cos(1.05449261588  + 3362.19188465030 *T)
    sumX0  +=       0.969 * cos(6.07412163761  + 1272.92484311070 *T)
    sumX0  +=       1.176 * cos(1.01032312893  + 9602.59645370770 *T)
    sumX0  +=       1.347 * cos(3.63306292592  + 19514.22741258770 *T)
    sumX0  +=       1.073 * cos(2.70812081471  + 6677.45791756710 *T)
    sumX0  +=       0.968 * cos(5.81625170021  + 522.33360061030 *T)
    sumX0  +=       1.192 * cos(3.92438816098  + 1222.09238380491 *T)
    sumX0  +=       1.237 * cos(5.72667761991  + 5835.38583613191 *T)
    sumX0  +=       1.057 * cos(1.29430538890  + 2597.37854868370 *T)
    sumX0  +=       0.957 * cos(0.10887247031  + 6645.44080420570 *T)
    sumX0  +=       0.971 * cos(0.15243087645  + 20809.22380716170 *T)
    sumX0  +=       0.991 * cos(2.10746769634  + 131.78577916990 *T)
    sumX0  +=       1.159 * cos(0.42658815264  + 9065.79194161230 *T)
    sumX0  +=       0.943 * cos(5.50078031935  + 3337.31587218211 *T)
    sumX0  +=       0.943 * cos(0.62298659929  + 3344.39661618450 *T)
    sumX0  +=       0.929 * cos(3.64745111813  + 4002.08917096430 *T)
    sumX0  +=       0.977 * cos(3.69400431600  + 2538.49232173710 *T)
    sumX0  +=       0.938 * cos(5.95942050303  + 931.60690433531 *T)
    sumX0  +=       1.272 * cos(2.46943992451  + 3302.72320854550 *T)
    sumX0  +=       0.960 * cos(4.05574189493  + 3909.67811821070 *T)
    sumX0  +=       0.909 * cos(3.63898193941  + 3326.62915018171 *T)
    sumX0  +=       0.943 * cos(4.87540027498  + 3344.36210888011 *T)
    sumX0  +=       0.943 * cos(1.24836664366  + 3337.35037948650 *T)
    sumX0  +=       0.886 * cos(0.45694838479  + 36.84918278771 *T)
    sumX0  +=       0.890 * cos(3.60250232314  + 169.33636564950 *T)
    sumX0  +=       0.925 * cos(3.39789693407  + 1486.22393854870 *T)
    sumX0  +=       1.156 * cos(4.57673762061  + 16061.94070863329 *T)
    sumX0  +=       0.879 * cos(5.66441531374  + 3355.08333818490 *T)
    sumX0  +=       0.965 * cos(4.35386546492  + 11614.67711121570 *T)
    sumX0  +=       0.874 * cos(3.48389519074  + 316.14805217309 *T)
    sumX0  +=       0.896 * cos(0.49130824985  + 2067.68758358909 *T)
    sumX0  +=       0.856 * cos(2.38006031197  + 6705.34710863090 *T)
    sumX0  +=       1.200 * cos(5.49932526392  + 2971.15642824290 *T)
    sumX0  +=       1.091 * cos(5.80152513875  + 9168.39708086389 *T)
    sumX0  +=       0.871 * cos(3.68107437791  + 9380.71585523369 *T)
    sumX0  +=       1.070 * cos(4.75389970378  + 426.35437339250 *T)
    sumX0  +=       1.039 * cos(1.92437940169  + 4937.04268646790 *T)
    sumX0  +=       0.968 * cos(0.44516536655  + 8959.17604904470 *T)
    sumX0  +=       0.852 * cos(0.05056797182  + 8013.52355842391 *T)
    sumX0  +=       0.868 * cos(4.22726507605  + 9947.29949901560 *T)
    sumX0  +=       0.890 * cos(0.04474613839  + 6747.95607979750 *T)
    sumX0  +=       0.834 * cos(2.38508120043  + 6681.04859478950 *T)
    sumX0  +=       1.025 * cos(2.15735900007  + 3914.71340755109 *T)
    sumX0  +=       0.838 * cos(3.57907497775  + 4672.42349675709 *T)
    sumX0  +=       0.872 * cos(5.33485611894  + 11780.73417599970 *T)
    sumX0  +=       0.998 * cos(4.91720456967  + 16335.59398705309 *T)
    sumX0  +=       0.814 * cos(4.96978756082  + 16703.28869733031 *T)
    sumX0  +=       1.043 * cos(4.06758688367  + 3492.75352526850 *T)
    sumX0  +=       0.878 * cos(3.45694168055  + 3233.19272024471 *T)
    sumX0  +=       0.814 * cos(0.71737388513  + 16703.32320463470 *T)
    sumX0  +=       0.821 * cos(1.52684790788  + 3323.04372206531 *T)
    sumX0  +=       0.821 * cos(1.45532635717  + 3358.66876630130 *T)
    sumX0  +=       0.861 * cos(6.07747716764  + 170.42905313570 *T)
    sumX0  +=       0.895 * cos(6.08078315422  + 8966.22228574270 *T)
    sumX0  +=       0.809 * cos(5.26204419900  + 1045.39865367110 *T)
    sumX0  +=       0.986 * cos(0.34218165194  + 9654.36913365350 *T)
    sumX0  +=       0.841 * cos(3.77708723958  + 13745.10242153889 *T)
    sumX0  +=       0.792 * cos(0.96214307294  + 7100.95331475830 *T)
    sumX0  +=       0.779 * cos(3.28806023341  + 3499.79976196650 *T)
    sumX0  +=       0.841 * cos(5.80837213979  + 3448.51976812190 *T)
    sumX0  +=       0.788 * cos(6.15063940263  + 14422.07545447190 *T)
    sumX0  +=       1.013 * cos(5.94832458905  + 8859.60639317511 *T)
    sumX0  +=       1.009 * cos(1.23105531513  + 6657.59023313531 *T)
    sumX0  +=       1.030 * cos(3.76569779004  + 3710.55606012370 *T)
    sumX0  +=       0.787 * cos(5.30046515807  + 3308.88393836651 *T)
    sumX0  +=       0.787 * cos(3.96489441416  + 3372.82855000010 *T)
    sumX0  +=       0.731 * cos(3.22922044128  + 2803.56409712089 *T)
    sumX0  +=       0.725 * cos(6.20613636370  + 3229.66960189570 *T)
    sumX0  +=       0.749 * cos(4.97775385354  + 4356.51926206750 *T)
    sumX0  +=       0.720 * cos(3.15038755289  + 3553.66770465429 *T)
    sumX0  +=       0.818 * cos(3.37077170078  + 802.12010496270 *T)
    sumX0  +=       0.825 * cos(5.84107737328  + 3253.54803964350 *T)
    sumX0  +=       0.793 * cos(2.67246545703  + 11.28951774740 *T)
    sumX0  +=       0.864 * cos(3.56004399666  + 6717.49653756051 *T)
    sumX0  +=       0.969 * cos(2.14091772308  + 20198.85114214949 *T)
    sumX0  +=       0.917 * cos(1.55560807260  + 4776.00390594270 *T)
    sumX0  +=       0.729 * cos(5.39749363536  + 22743.16556203289 *T)
    sumX0  +=       0.730 * cos(5.37089940702  + 1063.14886602171 *T)
    sumX0  +=       0.860 * cos(3.11478671194  + 242.48478649050 *T)
    sumX0  +=       0.713 * cos(3.97545007472  + 20597.00014555770 *T)
    sumX0  +=       0.694 * cos(3.05919399377  + 3452.04288647091 *T)
    sumX0  +=       0.694 * cos(3.98809162040  + 20426.81490990550 *T)
    sumX0  +=       0.972 * cos(1.98698742024  + 13916.26292712510 *T)
    sumX0  +=       0.692 * cos(3.93965986059  + 3068.18328683170 *T)
    sumX0  +=       0.730 * cos(2.57797232879  + 6016.71262575310 *T)
    sumX0  +=       0.795 * cos(3.42432574767  + 3428.16444872311 *T)
    sumX0  +=       0.686 * cos(4.76971489412  + 7255.81346921790 *T)
    sumX0  +=       0.712 * cos(5.60162243479  + 4143.22016662950 *T)
    sumX0  +=       0.727 * cos(3.20806928644  + 2301.82963339290 *T)
    sumX0  +=       0.836 * cos(3.67296892402  + 5642.44206009270 *T)
    sumX0  +=       0.672 * cos(1.47909475288  + 2541.03511901791 *T)
    sumX0  +=       0.725 * cos(3.81718268491  + 15907.00794416610 *T)
    sumX0  +=       0.723 * cos(5.49638926477  + 3735.48212924251 *T)
    sumX0  +=       0.659 * cos(1.96302087770  + 3929.43343622450 *T)
    sumX0  +=       0.752 * cos(2.52557851793  + 401.91593924071 *T)
    sumX0  +=       0.790 * cos(3.21400960006  + 10264.80970155691 *T)
    sumX0  +=       0.833 * cos(3.27818134172  + 6040.10342853389 *T)
    sumX0  +=       0.652 * cos(3.12933985544  + 3212.90471114871 *T)
    sumX0  +=       0.634 * cos(3.17304979978  + 17101.45495439070 *T)
    sumX0  +=       0.636 * cos(6.11222406149  + 17468.61138046190 *T)
    sumX0  +=       0.831 * cos(2.90780121075  + 2409.49315733150 *T)
    sumX0  +=       0.670 * cos(5.98800797760  + 15663.79170522509 *T)
    sumX0  +=       0.636 * cos(3.99964515410  + 5938.47861035050 *T)
    sumX0  +=       0.767 * cos(1.34735128779  + 16702.81831601549 *T)
    sumX0  +=       0.649 * cos(2.88860634271  + 485.72823530609 *T)
    sumX0  +=       0.720 * cos(4.17180002134  + 316.63568714010 *T)
    sumX0  +=       0.685 * cos(4.17683462000  + 1437.41943168211 *T)
    sumX0  +=       0.874 * cos(5.00761171285  + 3264.10253794069 *T)
    sumX0  +=       0.631 * cos(3.92104776965  + 3351.90194444720 *T)
    sumX0  +=       0.631 * cos(5.34431180258  + 3329.81054391940 *T)
    sumX0  +=       0.617 * cos(5.30944605959  + 3613.52920153491 *T)
    sumX0  +=       0.631 * cos(0.15035258857  + 3343.39904146410 *T)
    sumX0  +=       0.834 * cos(1.27334165089  + 6681.88874697671 *T)
    sumX0  +=       0.618 * cos(2.29471318966  + 6357.96325422570 *T)
    sumX0  +=       0.620 * cos(4.22618374757  + 2301.34199842589 *T)
    sumX0  +=       0.619 * cos(6.13399739970  + 3468.80777721790 *T)
    sumX0  +=       0.856 * cos(2.79699777939  + 6254.38284504009 *T)
    sumX0  +=       0.779 * cos(6.20441370524  + 22325.14887419290 *T)
    sumX0  +=       0.621 * cos(0.55328275382  + 22345.01655862470 *T)
    sumX0  +=       0.683 * cos(3.20202486335  + 3344.78839744640 *T)
    sumX0  +=       0.683 * cos(6.06333470888  + 3336.92409092021 *T)
    sumX0  +=       0.637 * cos(0.86301009432  + 151.29148732640 *T)
    sumX0  +=       0.600 * cos(2.64922528930  + 3338.31344690251 *T)
    sumX0  +=       0.737 * cos(2.16246318145  + 3116.02381351440 *T)
    sumX0  +=       0.843 * cos(0.53396130094  + 24149.83623386150 *T)
    sumX0  +=       0.595 * cos(2.29953648123  + 4039.63975744390 *T)
    sumX0  +=       0.621 * cos(0.40498928750  + 10448.67928845890 *T)
    sumX0  +=       0.649 * cos(6.17082186672  + 148.32254190981 *T)
    sumX0  +=       0.706 * cos(4.51954003602  + 6718.07403618730 *T)
    sumX0  +=       0.615 * cos(2.10277179076  + 1169.83206889211 *T)
    sumX0  +=       0.597 * cos(3.77194671714  + 2494.77340943211 *T)
    sumX0  +=       0.580 * cos(1.12862006570  + 3528.78139194590 *T)
    sumX0  +=       0.680 * cos(2.77754353712  + 1990.50119955750 *T)
    sumX0  +=       0.593 * cos(3.34298351276  + 9491.90249752129 *T)
    sumX0  +=       0.704 * cos(2.61371769364  + 3547.55698147990 *T)
    sumX0  +=       0.602 * cos(6.12196613719  + 6681.89855400980 *T)
    sumX0  +=       0.750 * cos(2.93237946903  + 6020.23574410210 *T)
    sumX0  +=       0.672 * cos(1.12165707971  + 6667.24157688150 *T)
    sumX0  +=       0.691 * cos(1.16386423121  + 6679.98419817480 *T)
    sumX0  +=       0.644 * cos(1.94503669392  + 4290.03185315310 *T)
    sumX0  +=       0.778 * cos(2.85190829858  + 394.86970254271 *T)
    sumX0  +=       0.606 * cos(5.75275909849  + 736.12033101530 *T)
    sumX0  +=       0.620 * cos(2.39266198193  + 1265.81129610991 *T)
    sumX0  +=       0.610 * cos(4.41144858406  + 18451.32236404950 *T)
    sumX0  +=       0.602 * cos(6.04397580328  + 20.11150191529 *T)
    sumX0  +=       0.579 * cos(1.80661717855  + 2391.19300024650 *T)
    sumX0  +=       0.574 * cos(0.81989247879  + 9381.69617630189 *T)
    sumX0  +=       0.585 * cos(5.02389462382  + 2170.78035780770 *T)
    sumX0  +=       0.610 * cos(0.51796263381  + 9911.87477636350 *T)
    sumX0  +=       0.648 * cos(2.64891747863  + 4133.63113265070 *T)
    sumX0  +=       0.583 * cos(4.50653081612  + 8226.82265386190 *T)
    sumX0  +=       0.602 * cos(4.20891405026  + 2854.88419139370 *T)
    sumX0  +=       0.555 * cos(5.99624223583  + 16865.28495214769 *T)
    sumX0  +=       0.600 * cos(4.73583940116  + 6725.18758318811 *T)
    sumX0  +=       0.766 * cos(4.61041261070  + 4775.51627097570 *T)
    sumX0  +=       0.577 * cos(4.65659204889  + 3184.94820978210 *T)
    sumX0  +=       0.587 * cos(5.57127850279  + 3194.04455765970 *T)
    sumX0  +=       0.613 * cos(3.89681403357  + 6549.92670919670 *T)
    sumX0  +=       0.696 * cos(0.15859294835  + 19004.40413192490 *T)
    sumX0  +=       0.634 * cos(4.14004370715  + 6510.79580026390 *T)
    sumX0  +=       0.634 * cos(1.16000431262  + 52176.05010063190 *T)
    sumX0  +=       0.603 * cos(2.33995115129  + 7896.20369020371 *T)
    sumX0  +=       0.658 * cos(2.46906007046  + 4988.81536641370 *T)
    sumX0  +=       0.602 * cos(3.81964202131  + 6681.03878775640 *T)
    sumX0  +=       0.520 * cos(2.82724916780  + 6467.68194047809 *T)
    sumX0  +=       0.522 * cos(5.42478409574  + 3339.67560153121 *T)
    sumX0  +=       0.522 * cos(3.84057547649  + 3342.03688683541 *T)
    sumX0  +=       0.528 * cos(4.56727960946  + 109.96250373590 *T)
    sumX0  +=       0.518 * cos(5.72791791977  + 1588.82907780029 *T)
    sumX0  +=       0.553 * cos(3.69268120443  + 3487.66793070691 *T)
    sumX0  +=       0.513 * cos(0.64641381344  + 299.37021175271 *T)
    sumX0  +=       0.715 * cos(1.53227325055  + 4503.33094859110 *T)
    sumX0  +=       0.511 * cos(2.84483152183  + 16490.00685554450 *T)
    sumX0  +=       0.505 * cos(4.83997458246  + 1063.55790093581 *T)
    sumX0  +=       0.698 * cos(6.07857734312  + 43.96272978850 *T)
    sumX0  +=       0.609 * cos(0.36018986098  + 3134.15550688670 *T)
    sumX0  +=       0.645 * cos(5.62278484839  + 1641.08939271309 *T)
    sumX0  +=       0.542 * cos(4.52460942145  + 1596.36294946531 *T)
    sumX0  +=       0.502 * cos(5.53009144929  + 12323.17927852529 *T)
    sumX0  +=       0.569 * cos(1.24397627234  + 7380.25218414370 *T)
    sumX0  +=       0.530 * cos(1.06670380363  + 1699.52303398670 *T)
    sumX0  +=       0.491 * cos(5.38693036025  + 11140.83688968570 *T)
    sumX0  +=       0.518 * cos(5.49646744261  + 3443.46138343489 *T)
    sumX0  +=       0.545 * cos(2.20958810438  + 2288.58786099491 *T)
    sumX0  +=       0.509 * cos(5.97803958339  + 6813.01063256950 *T)
    sumX0  +=       0.489 * cos(3.06702523150  + 13575.99261972070 *T)
    sumX0  +=       0.486 * cos(1.17003274180  + 2274.36076699330 *T)
    sumX0  +=       0.579 * cos(4.18326711934  + 10177.50149701710 *T)
    sumX0  +=       0.535 * cos(4.96826186440  + 9072.90548861311 *T)
    sumX0  +=       0.536 * cos(0.22940923175  + 3198.67761714710 *T)
    sumX0  +=       0.478 * cos(5.77347222021  + 114.64292439690 *T)
    sumX0  +=       0.619 * cos(4.96804844467  + 13207.27312484851 *T)
    sumX0  +=       0.527 * cos(3.25574629584  + 3392.13658204571 *T)
    sumX0  +=       0.527 * cos(6.00961327639  + 3289.57590632090 *T)
    sumX0  +=       0.491 * cos(1.56769780695  + 3393.11655909610 *T)
    sumX0  +=       0.491 * cos(1.41447645810  + 3288.59592927050 *T)
    sumX0  +=       0.518 * cos(5.06326236759  + 3826.82829697291 *T)
    sumX0  +=       0.532 * cos(1.66587585857  + 1055.69359440960 *T)
    sumX0  +=       0.475 * cos(2.35694443121  + 8116.61633264250 *T)
    sumX0  +=       0.473 * cos(3.22587789900  + 4297.14540015391 *T)
    sumX0  +=       0.483 * cos(0.44610561215  + 131.29814420289 *T)
    sumX0  +=       0.483 * cos(5.34340560368  + 3346.92215981310 *T)
    sumX0  +=       0.483 * cos(3.92195396855  + 3334.79032855351 *T)
    sumX0  +=       0.460 * cos(1.30188709712  + 2281.00393267929 *T)
    sumX0  +=       0.460 * cos(5.55430077281  + 2280.96942537490 *T)
    sumX0  +=       0.488 * cos(5.88179921344  + 5195.48854981791 *T)
    sumX0  +=       0.604 * cos(0.33182831620  + 2373.88536674770 *T)
    sumX0  +=       0.491 * cos(5.58315217747  + 11251.04321090511 *T)
    sumX0  +=       0.482 * cos(2.83595657742  + 23.63462026429 *T)
    sumX0  +=       0.455 * cos(4.28453161275  + 8962.71642104590 *T)
    sumX0  +=       0.455 * cos(2.25375998126  + 8962.68191374151 *T)
    sumX0  +=       0.455 * cos(3.21511534225  + 5408.78764525591 *T)
    sumX0  +=       0.458 * cos(1.20850177387  + 6313.75670695370 *T)
    sumX0  +=       0.525 * cos(0.42290029660  + 3535.82762864390 *T)
    sumX0  +=       0.506 * cos(2.75255831190  + 3483.03487121951 *T)
    sumX0  +=       0.453 * cos(3.59219997124  + 2957.97696561230 *T)
    sumX0  +=       0.633 * cos(0.33856779888  + 2963.48263626750 *T)
    sumX0  +=       0.532 * cos(2.38112136358  + 11773.13299403189 *T)
    sumX0  +=       0.502 * cos(4.62390876188  + 5562.71287878030 *T)
    sumX0  +=       0.487 * cos(5.66137463397  + 2548.08135571590 *T)
    sumX0  +=       0.604 * cos(4.92504643981  + 3311.36442587990 *T)
    sumX0  +=       0.604 * cos(4.34031313242  + 3370.34806248671 *T)
    sumX0  +=       0.448 * cos(4.97947122194  + 13361.71320321450 *T)
    sumX0  +=       0.455 * cos(2.05317333710  + 227.71995027250 *T)
    sumX0  +=       0.500 * cos(2.53819791260  + 4819.72281824770 *T)
    sumX0  +=       0.452 * cos(6.19160666400  + 6851.04885401611 *T)
    sumX0  +=       0.503 * cos(3.23595850746  + 382.63546108750 *T)
    sumX0  +=       0.440 * cos(4.69308097742  + 3565.20103988520 *T)
    sumX0  +=       0.607 * cos(3.16411651832  + 5195.00091485090 *T)
    sumX0  +=       0.477 * cos(5.99667940023  + 4379.88285497370 *T)
    sumX0  +=       0.468 * cos(5.32421647470  + 4186.93907893450 *T)
    sumX0  +=       0.474 * cos(1.58560587121  + 2295.21377302870 *T)
    sumX0  +=       0.430 * cos(1.74064675227  + 4716.63004402911 *T)
    sumX0  +=       0.536 * cos(3.39391143643  + 860.55374623631 *T)
    sumX0  +=       0.561 * cos(0.07745940312  + 4272.21933103511 *T)
    sumX0  +=       0.427 * cos(3.93324943357  + 1575.08961876571 *T)
    sumX0  +=       0.456 * cos(3.39763687697  + 10006.82862639790 *T)
    sumX0  +=       0.455 * cos(3.65437854757  + 1854.38848815109 *T)
    sumX0  +=       0.421 * cos(3.47192756045  + 5753.62870238030 *T)
    sumX0  +=       0.539 * cos(2.16591609141  + 3185.50315505190 *T)
    sumX0  +=       0.440 * cos(2.84607699756  + 5490.54477900750 *T)
    sumX0  +=       0.411 * cos(3.44829890606  + 2480.05868046350 *T)
    sumX0  +=       0.433 * cos(4.06813058765  + 3376.41897865191 *T)
    sumX0  +=       0.433 * cos(5.19722898458  + 3305.29350971470 *T)
    sumX0  +=       0.448 * cos(0.47019823361  + 8955.09798542589 *T)
    sumX0  +=       0.416 * cos(5.80949609188  + 23539.46356884929 *T)
    sumX0  +=       0.423 * cos(0.94668022805  + 10037.33356876791 *T)
    sumX0  +=       0.481 * cos(1.03096497224  + 6685.40082414620 *T)
    sumX0  +=       0.538 * cos(4.80873309420  + 6637.74975857810 *T)
    sumX0  +=       0.403 * cos(0.45532779122  + 394.38206757569 *T)
    sumX0  +=       0.414 * cos(5.88932386622  + 3016.86319255889 *T)
    sumX0  +=       0.402 * cos(1.79274198343  + 6151.79495944070 *T)
    sumX0  +=       0.403 * cos(2.79426969488  + 18849.47136745770 *T)
    sumX0  +=       0.402 * cos(6.04515565912  + 6151.76045213631 *T)
    sumX0  +=       0.430 * cos(1.85443040441  + 2655.81218995730 *T)
    sumX0  +=       0.435 * cos(5.32997749844  + 9375.07026426810 *T)
    sumX0  +=       0.509 * cos(0.81630529644  + 3496.20933331471 *T)
    sumX0  +=       0.548 * cos(5.48508184425  + 10184.54773371510 *T)
    sumX0  +=       0.397 * cos(5.67378728474  + 3723.73552275430 *T)
    sumX0  +=       0.406 * cos(3.29797001828  + 5750.10558403130 *T)
    sumX0  +=       0.477 * cos(3.72133708668  + 6614.98126196870 *T)
    sumX0  +=       0.555 * cos(2.64209868471  + 3718.22985209910 *T)
    sumX0  +=       0.425 * cos(5.57143459419  + 7218.27318297850 *T)
    sumX0  +=       0.408 * cos(2.99345508442  + 2149.44471734070 *T)
    sumX0  +=       0.402 * cos(4.31245835258  + 1685.29593998510 *T)
    sumX0  +=       0.411 * cos(4.42878307560  + 21265.76694400370 *T)
    sumX0  +=       0.486 * cos(4.84700116898  + 10050.53028505070 *T)
    sumX0  +=       0.394 * cos(1.43738424753  + 5888.69378241570 *T)
    sumX0  +=       0.487 * cos(2.30357691290  + 1655.31648671470 *T)
    sumX0  +=       0.421 * cos(3.53490544380  + 13518.11392371690 *T)
    sumX0  +=       0.395 * cos(4.02660480719  + 7366.02509014209 *T)
    sumX0  +=       0.392 * cos(0.97128057325  + 17655.02435723310 *T)
    sumX0  +=       0.443 * cos(0.74922233388  + 2075.28876555691 *T)
    sumX0  +=       0.404 * cos(5.22642617943  + 6644.86330557891 *T)
    sumX0  +=       0.381 * cos(4.92982666503  + 15805.90302195849 *T)
    sumX0  +=       0.414 * cos(2.26879885935  + 7747.96414807311 *T)
    sumX0  +=       0.395 * cos(1.49962602130  + 18606.25512851669 *T)
    sumX0  +=       0.377 * cos(6.04852731229  + 8564.55016398550 *T)
    sumX0  +=       0.475 * cos(0.05735729625  + 7380.73981911070 *T)
    sumX0  +=       0.378 * cos(2.64754791943  + 9801.66845514410 *T)
    sumX0  +=       0.430 * cos(5.91707320028  + 905.21287918130 *T)
    sumX0  +=       0.384 * cos(1.10487298901  + 3205.30352918089 *T)
    sumX0  +=       0.444 * cos(1.82982329075  + 2277.05452576400 *T)
    sumX0  +=       0.397 * cos(0.95323066030  + 515.70768857651 *T)
    sumX0  +=       0.471 * cos(5.87580736949  + 7270.53349789130 *T)
    sumX0  +=       0.367 * cos(1.76058008964  + 912.32642618209 *T)
    sumX0  +=       0.430 * cos(1.89342601713  + 5881.64754571770 *T)
    sumX0  +=       0.469 * cos(1.15366899488  + 7270.04586292429 *T)
    sumX0  +=       0.433 * cos(5.04314920496  + 2284.91883229020 *T)
    sumX0  +=       0.438 * cos(5.12080144325  + 9797.73630188101 *T)
    sumX0  +=       0.506 * cos(2.90113109352  + 1765.52280793410 *T)
    sumX0  +=       0.360 * cos(3.50809061650  + 20736.07597890910 *T)
    sumX0  +=       0.395 * cos(5.60289307177  + 7697.13168876730 *T)
    sumX0  +=       0.364 * cos(5.97461535482  + 3341.34893028450 *T)
    sumX0  +=       0.379 * cos(4.68258479838  + 5641.95442512570 *T)
    sumX0  +=       0.476 * cos(4.55315477415  + 1329.75590774351 *T)
    sumX0  +=       0.473 * cos(5.45656145574  + 21795.45790909830 *T)
    sumX0  +=       0.351 * cos(5.70650606381  + 6852.14154150230 *T)
    sumX0  +=       0.351 * cos(1.54032212267  + 2547.59372074890 *T)
    sumX0  +=       0.452 * cos(2.07881654365  + 6107.12387254830 *T)
    sumX0  +=       0.344 * cos(0.65486948739  + 16894.75421709410 *T)
    sumX0  +=       0.364 * cos(2.28651674734  + 6682.95314359140 *T)
    sumX0  +=       0.344 * cos(3.35027689955  + 5483.49854230950 *T)
    sumX0  +=       0.372 * cos(4.65630251193  + 1.22413855171 *T)
    sumX0  +=       0.452 * cos(2.64378641805  + 4307.82712161890 *T)
    sumX0  +=       0.349 * cos(3.97626348636  + 5636.31383469550 *T)
    sumX0  +=       0.362 * cos(3.43448634413  + 27681.89692667289 *T)
    sumX0  +=       0.376 * cos(4.65142839394  + 3339.38828814809 *T)
    sumX0  +=       0.347 * cos(0.90845197085  + 25685.62898532449 *T)
    sumX0  +=       0.351 * cos(4.37546900283  + 5452.50655756091 *T)
    sumX0  +=       0.348 * cos(3.23789957314  + 16865.77258711470 *T)
    sumX0  +=       0.345 * cos(2.38118381027  + 25287.47998191629 *T)
    sumX0  +=       0.358 * cos(1.48353909010  + 14584.05445563709 *T)
    sumX0  +=       0.336 * cos(2.98820379900  + 1882.83767561889 *T)
    sumX0  +=       0.436 * cos(1.98136341941  + 21946.86755521649 *T)
    sumX0  +=       0.434 * cos(1.89642061669  + 7214.68275432671 *T)
    sumX0  +=       0.395 * cos(5.73856202838  + 9993.63191011510 *T)
    sumX0  +=       0.351 * cos(1.69713753658  + 13213.13032714810 *T)
    sumX0  +=       0.327 * cos(5.61960069444  + 3354.09801658130 *T)
    sumX0  +=       0.327 * cos(3.64575887779  + 3327.61447178531 *T)
    sumX0  +=       0.361 * cos(5.11177050140  + 3.34661116829 *T)
    sumX0  +=       0.328 * cos(1.43829359971  + 2435.39954751850 *T)
    sumX0  +=       0.447 * cos(0.03231232062  + 13363.67384535091 *T)
    sumX0  +=       0.389 * cos(2.62527245368  + 30065.75565778170 *T)
    sumX0  +=       0.326 * cos(0.96031549475  + 3264.59017290771 *T)
    sumX0  +=       0.395 * cos(3.32438313829  + 3337.40043613710 *T)
    sumX0  +=       0.339 * cos(6.08213730855  + 14557.14053450730 *T)
    sumX0  +=       0.334 * cos(6.05623235898  + 147.05550400710 *T)
    sumX0  +=       0.335 * cos(0.79225648740  + 20617.77554105009 *T)
    sumX0  +=       0.347 * cos(5.62869239068  + 5223.93773728571 *T)
    sumX0  +=       0.344 * cos(1.14994835781  + 4025.90029840930 *T)
    sumX0  +=       0.314 * cos(2.86750999648  + 4142.73253166250 *T)
    sumX0  +=       0.347 * cos(4.19150197941  + 3265.58701064900 *T)
    sumX0  +=       0.408 * cos(1.39255335689  + 11883.82695021830 *T)
    sumX0  +=       0.317 * cos(0.53690715547  + 3347.48215621710 *T)
    sumX0  +=       0.335 * cos(3.39833290956  + 544.16187657970 *T)
    sumX0  +=       0.345 * cos(5.10704504215  + 5244.29305668450 *T)
    sumX0  +=       0.318 * cos(5.77952261520  + 12302.82395912650 *T)
    sumX0  +=       0.384 * cos(5.87247896307  + 7733.73705407150 *T)
    sumX0  +=       0.402 * cos(5.10624732337  + 8982.56685182550 *T)
    sumX0  +=       0.409 * cos(3.94804561206  + 23141.31456544110 *T)
    sumX0  +=       0.335 * cos(4.32453175361  + 3061.06973983091 *T)
    sumX0  +=       0.415 * cos(5.56961683825  + 12509.49714253070 *T)
    sumX0  +=       0.355 * cos(1.45991925898  + 7322.34627826531 *T)
    sumX0  +=       0.316 * cos(0.87910526351  + 3440.42590005290 *T)
    sumX0  +=       0.316 * cos(2.10306900154  + 3241.28658831371 *T)
    sumX0  +=       0.354 * cos(2.21314787069  + 4606.42372280970 *T)
    sumX0  +=       0.347 * cos(5.76966973208  + 14127.99895376209 *T)
    sumX0  +=       0.310 * cos(0.93053726029  + 6453.99253809410 *T)
    sumX0  +=       0.300 * cos(3.78819115153  + 1987.46571617550 *T)
    sumX0  +=       0.304 * cos(1.07677018782  + 15265.64270181689 *T)
    sumX0  +=       0.299 * cos(5.63009118135  + 10285.16502095570 *T)
    sumX0  +=       0.379 * cos(3.85337142681  + 1461.29786942991 *T)
    sumX0  +=       0.294 * cos(3.49253072198  + 20040.39525933330 *T)
    sumX0  +=       0.295 * cos(5.12372688658  + 2487.65986243130 *T)
    sumX0  +=       0.313 * cos(2.87465266242  + 6286.84278582391 *T)
    sumX0  +=       0.307 * cos(1.51056904066  + 9755.47405586110 *T)
    sumX0  +=       0.306 * cos(6.04685205074  + 8756.51361895650 *T)
    sumX0  +=       0.313 * cos(1.89942417078  + 4982.18945437991 *T)
    sumX0  +=       0.392 * cos(3.05316745117  + 19645.76937427410 *T)
    sumX0  +=       0.361 * cos(6.17342387386  + 684.80023674250 *T)
    sumX0  +=       0.328 * cos(5.65545524347  + 3284.05298201350 *T)
    sumX0  +=       0.328 * cos(3.60990432876  + 3397.65950635310 *T)
    sumX0  +=       0.291 * cos(2.47748718892  + 11986.91972443691 *T)
    sumX0  +=       0.341 * cos(5.74548283796  + 18053.17336064130 *T)
    sumX0  +=       0.345 * cos(5.01202054200  + 16063.40866466850 *T)
    sumX0  +=       0.314 * cos(4.21854332735  + 9983.94806194510 *T)
    sumX0  +=       0.331 * cos(0.26449397604  + 6680.99828956829 *T)
    sumX0  +=       0.329 * cos(3.71339298820  + 2277.95119564511 *T)
    sumX0  +=       0.301 * cos(4.17619017646  + 1251.07401378430 *T)
    sumX0  +=       0.339 * cos(1.15206040807  + 2825.39237309031 *T)
    sumX0  +=       0.339 * cos(1.83011385698  + 3856.32011527630 *T)
    sumX0  +=       0.383 * cos(4.85726262562  + 2921.93184528270 *T)
    sumX0  +=       0.302 * cos(0.20789359091  + 12825.88901218730 *T)
    sumX0  +=       0.310 * cos(5.32681384370  + 2060.57403658830 *T)
    sumX0  +=       0.310 * cos(3.86907443553  + 5622.10399434610 *T)
    sumX0  +=       0.292 * cos(2.39430170560  + 17256.87535382490 *T)
    sumX0  +=       0.286 * cos(2.53378899320  + 279.54268686890 *T)
    sumX0  +=       0.330 * cos(1.47284455054  + 1073.85284167431 *T)
    sumX0  +=       0.382 * cos(2.75070755631  + 3267.55911832430 *T)
    sumX0  +=       0.382 * cos(0.23146670874  + 3414.15337004230 *T)
    sumX0  +=       0.304 * cos(4.94294697674  + 3620.64274853570 *T)
    sumX0  +=       0.326 * cos(0.65821317943  + 533.86693584120 *T)
    sumX0  +=       0.322 * cos(1.32651672788  + 2388.65020296570 *T)
    sumX0  +=       0.289 * cos(2.79104178187  + 36.27168416091 *T)
    sumX0  +=       0.287 * cos(0.85216793603  + 5511.88041947450 *T)
    sumX0  +=       0.283 * cos(2.78331249725  + 12840.11610618890 *T)
    sumX0  +=       0.358 * cos(4.64860381703  + 568.57805654390 *T)
    sumX0  +=       0.282 * cos(2.30704296928  + 595.93455652550 *T)
    sumX0  +=       0.380 * cos(3.49134141596  + 2973.63191522091 *T)
    sumX0  +=       0.283 * cos(6.10276906215  + 9859.61446145070 *T)
    sumX0  +=       0.269 * cos(2.39532753439  + 15636.81047379250 *T)
    sumX0  +=       0.280 * cos(1.99086077526  + 16511.85768487090 *T)
    sumX0  +=       0.296 * cos(5.47787166920  + 1.24065522479 *T)
    sumX0  +=       0.300 * cos(4.16267400886  + 13525.16016041490 *T)
    sumX0  +=       0.331 * cos(4.51690765173  + 6680.96378226390 *T)
    sumX0  +=       0.310 * cos(1.83830280403  + 5622.06948704171 *T)
    sumX0  +=       0.323 * cos(5.09944111205  + 2266.75958502550 *T)
    sumX0  +=       0.284 * cos(4.75295349811  + 5085.34953890130 *T)
    sumX0  +=       0.320 * cos(5.98266609774  + 10720.86461084350 *T)
    sumX0  +=       0.317 * cos(3.36262525597  + 6048.19729660290 *T)
    sumX0  +=       0.269 * cos(4.46031273955  + 2744.67787017430 *T)
    sumX0  +=       0.318 * cos(3.75471974270  + 526.00262931501 *T)
    sumX0  +=       0.257 * cos(3.06293713227  + 24341.28449997309 *T)
    sumX0  +=       0.266 * cos(2.03059365371  + 206.94455478010 *T)
    sumX0  +=       0.258 * cos(3.36613687261  + 8749.40007195570 *T)
    sumX0  +=       0.271 * cos(3.84748092274  + 3345.42699390331 *T)
    sumX0  +=       0.271 * cos(5.41787864949  + 3336.28549446330 *T)
    sumX0  +=       0.301 * cos(3.14400326371  + 10706.63751684189 *T)
    sumX0  +=       0.251 * cos(5.53940659235  + 3343.95928643870 *T)
    sumX0  +=       0.251 * cos(3.72595297988  + 3337.75320192791 *T)
    sumX0  +=       0.251 * cos(4.94239589656  + 9758.99717421010 *T)
    sumX0  +=       0.267 * cos(2.53115640127  + 6831.03186801770 *T)
    sumX0  +=       0.306 * cos(2.93208929772  + 12012.82611462390 *T)
    sumX0  +=       0.249 * cos(1.56864578441  + 4246.31294084810 *T)
    sumX0  +=       0.286 * cos(0.19225481574  + 1898.59503542310 *T)
    sumX0  +=       0.320 * cos(3.66652720680  + 536.56069461190 *T)
    sumX0  +=       0.290 * cos(2.87280676776  + 20047.44149603130 *T)
    sumX0  +=       0.268 * cos(3.04782864984  + 16703.23864067971 *T)
    sumX0  +=       0.250 * cos(2.52435118750  + 9698.57568092550 *T)
    sumX0  +=       0.248 * cos(5.82252924753  + 4534.81561944089 *T)
    sumX0  +=       0.246 * cos(5.46974928195  + 15651.03756779411 *T)
    sumX0  +=       0.309 * cos(5.65012767043  + 1903.19299501769 *T)
    sumX0  +=       0.247 * cos(1.02711909982  + 10016.55817327550 *T)
    sumX0  +=       0.262 * cos(5.09398079638  + 4025.41266344230 *T)
    sumX0  +=       0.312 * cos(0.44934890599  + 6033.97020260130 *T)
    sumX0  +=       0.245 * cos(4.18381763277  + 2939.18412242610 *T)
    sumX0  +=       0.317 * cos(2.05559964887  + 95.73540973430 *T)
    sumX0  +=       0.256 * cos(3.14277515921  + 1727.90491115171 *T)
    sumX0  +=       0.330 * cos(1.40678844479  + 735.63269604830 *T)
    sumX0  +=       0.276 * cos(0.25330294458  + 6670.42297061920 *T)
    sumX0  +=       0.321 * cos(0.33298486449  + 11905.16259068530 *T)
    sumX0  +=       0.252 * cos(5.12405452534  + 5430.63847458231 *T)
    sumX0  +=       0.252 * cos(3.32857386946  + 21548.71855180830 *T)
    sumX0  +=       0.237 * cos(1.01475797851  + 194.72756697710 *T)
    sumX0  +=       0.258 * cos(4.25971949042  + 9488.86701413930 *T)
    sumX0  +=       0.269 * cos(3.40948701309  + 3867.02409092890 *T)
    sumX0  +=       0.265 * cos(4.47702156214  + 3053.89388306271 *T)
    sumX0  +=       0.240 * cos(3.30477326315  + 10027.60402189030 *T)
    sumX0  +=       0.321 * cos(4.39956903033  + 3759.78064308390 *T)
    sumX0  +=       0.319 * cos(5.16195488471  + 9815.89554914570 *T)
    sumX0  +=       0.251 * cos(1.69379264716  + 1280.03839011150 *T)
    sumX0  +=       0.239 * cos(1.35799257039  + 26481.92699214089 *T)
    sumX0  +=       0.251 * cos(0.30789864433  + 2790.80995968991 *T)
    sumX0  +=       0.326 * cos(5.77590839437  + 3708.08057314570 *T)
    sumX0  +=       0.235 * cos(5.32419238086  + 5106.62286960090 *T)
    sumX0  +=       0.293 * cos(4.82940686086  + 5628.71265272770 *T)
    sumX0  +=       0.323 * cos(0.15779112892  + 17086.20248320570 *T)
    sumX0  +=       0.311 * cos(1.69858307705  + 3290.02378487750 *T)
    sumX0  +=       0.311 * cos(1.28359118800  + 3391.68870348911 *T)
    sumX0  +=       0.242 * cos(5.03369279862  + 4996.41654838150 *T)
    sumX0  +=       0.233 * cos(3.70327934902  + 2679.13568243529 *T)
    sumX0  +=       0.230 * cos(3.19672105861  + 10654.86483689610 *T)
    sumX0  +=       0.292 * cos(0.94155136266  + 3050.37076471370 *T)
    sumX0  +=       0.230 * cos(5.31033620112  + 1744.24947723449 *T)
    sumX0  +=       0.235 * cos(3.67403387076  + 5945.59215735130 *T)
    sumX0  +=       0.301 * cos(3.92360235288  + 9495.91325083730 *T)
    sumX0  +=       0.318 * cos(2.84216979877  + 2907.14450630650 *T)
    sumX0  +=       0.229 * cos(3.22262757156  + 5.27910682390 *T)
    sumX0  +=       0.248 * cos(0.37197955444  + 1442.01739127670 *T)
    sumX0  +=       0.226 * cos(4.14053859051  + 1868.61558215270 *T)
    sumX0  +=       0.234 * cos(3.57298597343  + 2817.79119112250 *T)
    sumX0  +=       0.230 * cos(5.02142639697  + 4245.82530588109 *T)
    sumX0  +=       0.312 * cos(5.79393471603  + 4403.76129272151 *T)
    sumX0  +=       0.222 * cos(2.77556068119  + 1835.56843509031 *T)
    sumX0  +=       0.277 * cos(2.78377312092  + 7079.13003932430 *T)
    sumX0  +=       0.289 * cos(2.22892326730  + 6901.88131332190 *T)
    sumX0  +=       0.244 * cos(3.78614066645  + 5349.41378334230 *T)
    sumX0  +=       0.268 * cos(2.63933226300  + 16703.37326128530 *T)
    sumX0  +=       0.250 * cos(1.57631353673  + 187.68133027909 *T)
    sumX0  +=       0.220 * cos(4.31591174490  + 6511.88848775011 *T)
    sumX0  +=       0.224 * cos(1.47201980663  + 3138.60284900920 *T)
    sumX0  +=       0.224 * cos(1.51015445842  + 3543.10963935740 *T)
    sumX0  +=       0.234 * cos(1.63153005672  + 17499.60395779890 *T)
    sumX0  +=       0.256 * cos(5.60816219579  + 367.46814644590 *T)
    sumX0  +=       0.272 * cos(5.08797661427  + 2428.28600051770 *T)
    sumX0  +=       0.218 * cos(6.20908726956  + 2786.79920637390 *T)
    sumX0  +=       0.218 * cos(1.17556461134  + 6283.33692112710 *T)
    sumX0  +=       0.270 * cos(1.81634201038  + 11236.81611690350 *T)
    sumX0  +=       0.218 * cos(5.42797828703  + 6283.30241382270 *T)
    sumX0  +=       0.220 * cos(4.76793118335  + 3987.86707749811 *T)
    sumX0  +=       0.238 * cos(1.21800672651  + 2814.68839743771 *T)
    sumX0  +=       0.222 * cos(1.93943525542  + 1368.90407032850 *T)
    sumX0  +=       0.215 * cos(0.44293995073  + 13383.46891977510 *T)
    sumX0  +=       0.224 * cos(3.68637424288  + 77.99436146740 *T)
    sumX0  +=       0.221 * cos(4.81153466860  + 3937.03461819230 *T)
    sumX0  +=       0.226 * cos(4.13866805425  + 9279.09103705030 *T)
    sumX0  +=       0.256 * cos(5.10926173053  + 10596.42589591770 *T)
    sumX0  +=       0.211 * cos(0.32856158622  + 12200.21881987490 *T)
    sumX0  +=       0.212 * cos(1.71628960778  + 2392.66095628171 *T)
    sumX0  +=       0.216 * cos(4.73005287854  + 12928.98178640591 *T)
    sumX0  +=       0.237 * cos(4.78833782578  + 3627.81860530390 *T)
    sumX0  +=       0.220 * cos(2.11904228547  + 6530.42100104020 *T)
    sumX0  +=       0.213 * cos(4.70667890897  + 1484.68867221069 *T)
    sumX0  +=       0.262 * cos(0.21389223894  + 20.53157800890 *T)
    sumX0  +=       0.254 * cos(4.24958766273  + 19146.51544865829 *T)
    sumX0  +=       0.222 * cos(1.24391244474  + 8166.40116057730 *T)
    sumX0  +=       0.263 * cos(0.61066006216  + 4531.78013605890 *T)
    sumX0  +=       0.249 * cos(4.15411886603  + 3084.88586781131 *T)
    sumX0  +=       0.239 * cos(5.60863730462  + 5095.91889394830 *T)
    sumX0  +=       0.239 * cos(2.46704465103  + 5095.43125898130 *T)
    sumX0  +=       0.270 * cos(3.81676548381  + 24889.33097850809 *T)
    sumX0  +=       0.252 * cos(4.65459224520  + 3299.30545319850 *T)
    sumX0  +=       0.252 * cos(4.61076732703  + 3382.40703516810 *T)
    sumX0  +=       0.207 * cos(4.40367735236  + 5618.15458743080 *T)
    sumX0  +=       0.239 * cos(3.20969165945  + 3344.31205222951 *T)
    sumX0  +=       0.262 * cos(2.04023447861  + 3631.34172365290 *T)
    sumX0  +=       0.206 * cos(3.37185879421  + 7083.14079264031 *T)
    sumX0  +=       0.249 * cos(2.30381341333  + 2149.93235230770 *T)
    sumX0  +=       0.210 * cos(1.09820249005  + 6678.71716027210 *T)
    sumX0  +=       0.255 * cos(3.12446085063  + 1994.51195287351 *T)
    sumX0  +=       0.204 * cos(5.48441633924  + 4694.24677219110 *T)
    sumX0  +=       0.274 * cos(0.08158298999  + 12306.83471244250 *T)
    sumX0  +=       0.262 * cos(5.18218468663  + 532.97026596010 *T)
    sumX0  +=       0.214 * cos(2.76700465239  + 9623.44445920770 *T)
    sumX0  +=       0.246 * cos(4.44352898187  + 151.65346360169 *T)
    sumX0  +=       0.223 * cos(4.55939375869  + 3345.21828057310 *T)
    sumX0  +=       0.223 * cos(4.70596581354  + 3336.49420779350 *T)
    sumX0  +=       0.255 * cos(0.57678604074  + 2942.72449442730 *T)
    sumX0  +=       0.224 * cos(3.09503100625  + 6533.38994645680 *T)
    sumX0  +=       0.245 * cos(3.93191759940  + 2142.39848064269 *T)
    sumX0  +=       0.201 * cos(0.64350792344  + 9360.84817080190 *T)
    sumX0  +=       0.282 * cos(5.79937846262  + 12465.29059525869 *T)
    sumX0  +=       0.199 * cos(4.83221191762  + 3030.01544531491 *T)
    sumX0  +=       0.243 * cos(5.80725236762  + 9329.92349635610 *T)
    sumX0  +=       0.233 * cos(4.46399451272  + 5834.89820116490 *T)
    sumX0  +=       0.205 * cos(5.87364808478  + 7167.44072367270 *T)
    sumX0  +=       0.199 * cos(4.98231898069  + 2957.47207699310 *T)
    sumX0  +=       0.220 * cos(1.28938253421  + 5626.01889395700 *T)
    sumX0  +=       0.197 * cos(0.25728561615  + 3244.87701696550 *T)
    sumX0  +=       0.197 * cos(2.72488864890  + 3436.83547140110 *T)
    sumX0  +=       0.200 * cos(3.33617729049  + 6567.06956396970 *T)
    sumX0  +=       0.231 * cos(4.21063730317  + 6654.06711478630 *T)
    sumX0  +=       0.228 * cos(4.31246262185  + 3319.59321372389 *T)
    sumX0  +=       0.206 * cos(3.05101379573  + 7483.34495836230 *T)
    sumX0  +=       0.206 * cos(6.19260644932  + 7483.83259332930 *T)
    sumX0  +=       0.198 * cos(5.49214722476  + 3094.60452186030 *T)
    sumX0  +=       0.215 * cos(2.54030230126  + 2540.54748405090 *T)
    sumX0  +=       0.223 * cos(4.58369899094  + 13341.91812879030 *T)
    sumX0  +=       0.199 * cos(3.86871213743  + 2409.00552236449 *T)
    sumX0  +=       0.192 * cos(2.31728673105  + 6382.34227661390 *T)
    sumX0  +=       0.205 * cos(2.67383805570  + 3890.90252867670 *T)
    sumX0  +=       0.191 * cos(2.18933385410  + 2928.48514730890 *T)
    sumX0  +=       0.191 * cos(0.79284041095  + 3753.22734105771 *T)
    sumX0  +=       0.193 * cos(2.64989993830  + 117.56368570370 *T)
    sumX0  +=       0.194 * cos(2.90629271112  + 18208.10612510850 *T)
    sumX0  +=       0.197 * cos(2.64056587295  + 20206.38501381450 *T)
    sumX0  +=       0.194 * cos(2.75330979101  + 28628.09240861609 *T)
    sumX0  +=       0.212 * cos(0.12612105707  + 4538.82637275690 *T)
    sumX0  +=       0.192 * cos(5.68657153593  + 3365.71500299931 *T)
    sumX0  +=       0.192 * cos(3.57878803630  + 3315.99748536730 *T)
    sumX0  +=       0.218 * cos(2.79753939558  + 2142.88611560970 *T)
    sumX0  +=       0.226 * cos(5.11126309189  + 3596.82662055530 *T)
    sumX0  +=       0.253 * cos(0.88845534596  + 5415.90119225670 *T)
    sumX0  +=       0.187 * cos(1.76835954015  + 15107.11950869790 *T)
    sumX0  +=       0.207 * cos(2.95174513706  + 419.24082639170 *T)
    sumX0  +=       0.255 * cos(4.82919971643  + 2942.68998712290 *T)
    sumX0  +=       0.191 * cos(2.45354674704  + 10544.65851567670 *T)
    sumX0  +=       0.190 * cos(4.20925049248  + 28229.94340520789 *T)
    sumX0  +=       0.205 * cos(2.24855213910  + 1059.60849402051 *T)
    sumX0  +=       0.233 * cos(5.87086760494  + 5821.15874213030 *T)
    sumX0  +=       0.187 * cos(3.39707563848  + 5401.18646328809 *T)
    sumX0  +=       0.188 * cos(1.22198431905  + 3318.08774757391 *T)
    sumX0  +=       0.188 * cos(1.76018994600  + 3363.62474079271 *T)
    sumX0  +=       0.231 * cos(4.09170497266  + 9094.24112908010 *T)
    sumX0  +=       0.194 * cos(2.01920817535  + 2736.00650347850 *T)
    sumX0  +=       0.182 * cos(2.90203674966  + 5960.32943967690 *T)
    sumX0  +=       0.193 * cos(4.00292264985  + 7417.34518441490 *T)
    sumX0  +=       0.186 * cos(0.33834024196  + 11354.13598512370 *T)
    sumX0  +=       0.181 * cos(3.45042796588  + 1754.81883228150 *T)
    sumX0  +=       0.222 * cos(1.29486520765  + 3070.65877380970 *T)
    sumX0  +=       0.181 * cos(0.71220143855  + 4356.03162710049 *T)
    sumX0  +=       0.177 * cos(4.43301273664  + 3651.69704305170 *T)
    sumX0  +=       0.189 * cos(1.80837972860  + 6673.86748891530 *T)
    sumX0  +=       0.179 * cos(5.05077866044  + 15121.34660269950 *T)
    sumX0  +=       0.232 * cos(4.23011116665  + 3517.54997431711 *T)
    sumX0  +=       0.232 * cos(5.03524840558  + 3164.16251404950 *T)
    sumX0  +=       0.217 * cos(5.77707162486  + 220.16882495529 *T)
    sumX0  +=       0.176 * cos(4.70413738332  + 1903.68062998470 *T)
    sumX0  +=       0.180 * cos(2.39448209299  + 3348.95011225231 *T)
    sumX0  +=       0.180 * cos(0.58769217206  + 3332.76237611430 *T)
    sumX0  +=       0.179 * cos(3.77310628062  + 3587.10796650631 *T)
    sumX0  +=       0.205 * cos(4.27932377060  + 1059.64300132490 *T)
    sumX0  +=       0.187 * cos(2.61826471479  + 2619.71701297711 *T)
    sumX0  +=       0.176 * cos(5.47433094670  + 13465.78629850130 *T)
    sumX0  +=       0.242 * cos(0.93268242658  + 5408.30001028890 *T)
    sumX0  +=       0.178 * cos(2.51501672448  + 2973.14428025389 *T)
    sumX0  +=       0.226 * cos(0.65351089705  + 11925.03027511709 *T)
    sumX0  +=       0.174 * cos(6.12894045783  + 10096.86269615020 *T)
    sumX0  +=       0.188 * cos(5.50048096325  + 2575.06317973690 *T)
    sumX0  +=       0.175 * cos(0.98347992481  + 12299.78847574450 *T)
    sumX0  +=       0.232 * cos(4.71901702635  + 5351.95658062310 *T)
    sumX0  +=       0.174 * cos(6.06374124229  + 5243.80542171749 *T)
    sumX0  +=       0.177 * cos(4.28915238170  + 6872.42930202769 *T)
    sumX0  +=       0.177 * cos(0.60438030609  + 2573.00742483470 *T)
    sumX0  +=       0.210 * cos(2.56022036127  + 6684.22018149410 *T)
    sumX0  +=       0.175 * cos(5.70361271383  + 6708.87022697990 *T)
    sumX0  +=       0.174 * cos(0.12430770186  + 3156.00633621310 *T)
    sumX0  +=       0.182 * cos(4.32115413536  + 2011.34415392330 *T)
    sumX0  +=       0.169 * cos(5.11813247600  + 6684.97453557991 *T)
    sumX0  +=       0.173 * cos(2.47259223958  + 4885.72259219509 *T)
    sumX0  +=       0.239 * cos(3.79200199968  + 31022.50935337269 *T)
    sumX0  +=       0.169 * cos(0.86571880031  + 6685.00904288430 *T)
    sumX0  +=       0.191 * cos(5.18490906634  + 11567.43508056170 *T)
    sumX0  +=       0.207 * cos(4.46904668742  + 10001.72577818411 *T)
    sumX0  +=       0.181 * cos(0.23355817402  + 8584.41784841729 *T)
    sumX0  +=       0.189 * cos(1.06059790633  + 2783.76372299191 *T)
    sumX0  +=       0.233 * cos(3.39745394763  + 23958.87560271690 *T)
    sumX0  +=       0.219 * cos(5.87651375075  + 3070.72608411250 *T)
    sumX0  +=       0.177 * cos(4.57327453091  + 2978.73510745250 *T)
    sumX0  +=       0.179 * cos(3.71447292171  + 17232.99691607710 *T)
    sumX0  +=       0.168 * cos(5.95159678891  + 1111.39842761850 *T)
    sumX0  +=       0.172 * cos(0.33588249395  + 4061.99547538951 *T)
    sumX0  +=       0.184 * cos(1.40136911195  + 12943.20888040750 *T)
    sumX0  +=       0.179 * cos(1.59854865789  + 8852.49284617430 *T)
    sumX0  +=       0.162 * cos(1.97243550807  + 16696.19240398170 *T)
    sumX0  +=       0.163 * cos(5.34659176540  + 19676.20641375289 *T)
    sumX0  +=       0.204 * cos(4.15805864546  + 4253.42648784890 *T)
    sumX0  +=       0.200 * cos(1.68724699588  + 3611.05371455690 *T)
    sumX0  +=       0.188 * cos(2.44881959486  + 2385.61471958371 *T)
    sumX0  +=       0.166 * cos(5.16059611000  + 24491.18197509989 *T)
    sumX0  +=       0.168 * cos(2.00217090683  + 6166.00479979010 *T)
    sumX0  +=       0.201 * cos(5.56233470044  + 3738.98799393931 *T)
    sumX0  +=       0.184 * cos(6.06162121787  + 1155.60497489050 *T)
    sumX0  +=       0.163 * cos(1.26202958512  + 3335.65705301750 *T)
    sumX0  +=       0.163 * cos(1.72014467993  + 3346.05543534910 *T)
    sumX0  +=       0.158 * cos(3.39261805220  + 5490.05714404049 *T)
    sumX0  +=       0.162 * cos(3.55591572493  + 272.42913986809 *T)
    sumX0  +=       0.173 * cos(3.27771655682  + 4.17597074660 *T)
    sumX0  +=       0.166 * cos(6.26161653226  + 21393.78578734110 *T)
    sumX0  +=       0.175 * cos(1.20396202464  + 2757.85733280490 *T)
    sumX0  +=       0.154 * cos(1.26391171796  + 4289.05153208490 *T)
    sumX0  +=       0.214 * cos(3.94790035362  + 1386.13909381971 *T)
    sumX0  +=       0.171 * cos(1.64736764034  + 10011.44443223310 *T)
    sumX0  +=       0.156 * cos(5.65947538133  + 10081.45495949650 *T)
    sumX0  +=       0.163 * cos(2.04192539510  + 13286.42745300710 *T)
    sumX0  +=       0.155 * cos(5.55556139897  + 8771.25090128210 *T)
    sumX0  +=       0.151 * cos(6.02869301104  + 15011.14028148010 *T)
    sumX0  +=       0.155 * cos(5.93813726450  + 1879.31455726989 *T)
    sumX0  +=       0.155 * cos(2.60087355671  + 2089.53841291549 *T)
    sumX0  +=       0.194 * cos(0.09021705575  + 3361.14400470870 *T)
    sumX0  +=       0.169 * cos(6.19326528274  + 7064.34794945411 *T)
    sumX0  +=       0.166 * cos(0.77365213441  + 6719.60170652090 *T)
    sumX0  +=       0.158 * cos(1.56915594044  + 8439.63411433290 *T)
    sumX0  +=       0.151 * cos(5.90185701300  + 10228.26664602010 *T)
    sumX0  +=       0.151 * cos(2.82948943010  + 3525.70615215350 *T)
    sumX0  +=       0.149 * cos(1.89551632354  + 3664.36166084070 *T)
    sumX0  +=       0.196 * cos(3.39238995578  + 3610.98640425411 *T)
    sumX0  +=       0.166 * cos(2.03190415107  + 5466.73365156250 *T)
    sumX0  +=       0.167 * cos(2.89181496866  + 10288.68813930470 *T)
    sumX0  +=       0.156 * cos(0.79100595475  + 17394.97591724229 *T)
    sumX0  +=       0.147 * cos(1.14044441078  + 14577.42854360330 *T)
    sumX0  +=       0.147 * cos(2.18386267585  + 1898.57778177091 *T)
    sumX0  +=       0.154 * cos(1.77068494981  + 11457.22875934230 *T)
    sumX0  +=       0.172 * cos(5.39541277180  + 19617.32018680630 *T)
    sumX0  +=       0.166 * cos(2.73543927114  + 6908.94480367210 *T)
    sumX0  +=       0.160 * cos(0.44981752738  + 1346.10047382629 *T)
    sumX0  +=       0.180 * cos(5.45674212617  + 5220.41461893670 *T)
    sumX0  +=       0.159 * cos(0.77814105937  + 1505.04399160949 *T)
    sumX0  +=       0.171 * cos(5.60623731826  + 4399.75053940550 *T)
    sumX0  +=       0.178 * cos(2.31756312240  + 2807.15452577270 *T)
    sumX0  +=       0.201 * cos(1.30992102475  + 3739.02250124370 *T)
    sumX0  +=       0.148 * cos(0.98641859240  + 9154.14743350530 *T)
    sumX0  +=       0.146 * cos(0.62125880196  + 13369.80707128350 *T)
    sumX0  +=       0.170 * cos(5.74209007670  + 6677.92829888190 *T)
    sumX0  +=       0.147 * cos(1.87608878363  + 9918.98832336430 *T)
    sumX0  +=       0.147 * cos(5.20291449518  + 14481.44931638550 *T)
    sumX0  +=       0.147 * cos(1.47480421685  + 6850.56121904909 *T)
    sumX0  +=       0.159 * cos(0.03020279035  + 3334.99904188371 *T)
    sumX0  +=       0.159 * cos(2.95197147470  + 3346.71344648291 *T)
    sumX0  +=       0.150 * cos(2.06659306486  + 15110.22230238269 *T)
    sumX0  +=       0.143 * cos(1.10166083747  + 13227.62844424730 *T)
    sumX0  +=       0.189 * cos(6.14146200362  + 10042.43641698170 *T)
    sumX0  +=       0.153 * cos(4.69206111649  + 3702.97738091410 *T)
    sumX0  +=       0.168 * cos(2.68821315216  + 10032.71776293271 *T)
    sumX0  +=       0.155 * cos(3.47829704410  + 369.94363342390 *T)
    sumX0  +=       0.151 * cos(2.48862413425  + 1859.44687283811 *T)
    sumX0  +=       0.163 * cos(1.42057059488  + 2965.09051261311 *T)
    sumX0  +=       0.179 * cos(2.60241794828  + 3966.52643649571 *T)
    sumX0  +=       0.179 * cos(0.37975631677  + 2715.18605187090 *T)
    sumX0  +=       0.195 * cos(3.84830983656  + 16858.72635041670 *T)
    sumX0  +=       0.140 * cos(4.86382612147  + 18461.95902939930 *T)
    sumX0  +=       0.183 * cos(0.37676347949  + 6791.67499210250 *T)
    sumX0  +=       0.150 * cos(4.97392660722  + 1879.80219223690 *T)
    sumX0  +=       0.150 * cos(5.20281046756  + 2922.35192137631 *T)
    sumX0  +=       0.182 * cos(2.41452857008  + 8005.92237645609 *T)
    sumX0  +=       0.146 * cos(1.91269041932  + 2715.23110798610 *T)
    sumX0  +=       0.165 * cos(1.13937556996  + 7346.22471601311 *T)
    sumX0  +=       0.141 * cos(0.78224513843  + 3017.35082752591 *T)
    sumX0  +=       0.191 * cos(2.19520817628  + 3202.26804579890 *T)
    sumX0  +=       0.191 * cos(0.78696608877  + 3479.44444256770 *T)
    sumX0  +=       0.139 * cos(0.20210254749  + 4846.14405327630 *T)
    sumX0  +=       0.138 * cos(5.80974078574  + 1537.99661849450 *T)
    sumX0  +=       0.193 * cos(2.76342547478  + 3650.13456683910 *T)
    sumX0  +=       0.160 * cos(3.79608374479  + 3394.16419046711 *T)
    sumX0  +=       0.160 * cos(5.46927582744  + 3287.54829789951 *T)
    sumX0  +=       0.193 * cos(0.22143365956  + 3031.57792152750 *T)
    sumX0  +=       0.169 * cos(6.13731522421  + 4687.20053549310 *T)
    sumX0  +=       0.155 * cos(1.91928194110  + 3897.94876537470 *T)
    sumX0  +=       0.162 * cos(1.51359934194  + 5607.85964669230 *T)
    sumX0  +=       0.136 * cos(5.24657523026  + 3336.77818056450 *T)
    sumX0  +=       0.144 * cos(0.96151831208  + 3945.70598488811 *T)
    sumX0  +=       0.155 * cos(1.92058338472  + 7799.73682801889 *T)
    sumX0  +=       0.170 * cos(1.48967640102  + 6677.96280618630 *T)
    sumX0  +=       0.155 * cos(5.68296928057  + 25136.07033579810 *T)
    sumX0  +=       0.141 * cos(3.83758141126  + 6279.79654912590 *T)
    sumX0  +=       0.174 * cos(4.01650767670  + 7586.43773258089 *T)
    sumX0  +=       0.138 * cos(4.77290767341  + 6343.22597190010 *T)
    sumX0  +=       0.176 * cos(0.53515924546  + 605.09355818830 *T)
    sumX0  +=       0.167 * cos(5.28260459939  + 8337.02897508130 *T)
    sumX0  +=       0.185 * cos(5.57942546824  + 27831.79440179969 *T)
    sumX0  +=       0.151 * cos(3.22384573444  + 2281.05398932989 *T)
    sumX0  +=       0.151 * cos(6.20601999951  + 8962.76647769650 *T)
    sumX0  +=       0.136 * cos(2.37641738096  + 12729.90978496950 *T)
    sumX0  +=       0.159 * cos(4.80632202174  + 8958.76701413060 *T)
    sumX0  +=       0.151 * cos(0.33133107918  + 8962.63185709091 *T)
    sumX0  +=       0.151 * cos(3.63234212128  + 2280.91936872430 *T)
    sumX0  +=       0.137 * cos(3.04837973833  + 6703.31950020950 *T)
    sumX0  +=       0.159 * cos(1.73569021890  + 8966.63132065680 *T)
    sumX0  +=       0.132 * cos(6.26868643109  + 12171.76963240710 *T)
    sumX0  +=       0.136 * cos(5.59329741401  + 8696.09212567190 *T)
    sumX0  +=       0.143 * cos(3.76386314585  + 4106.64930862970 *T)
    sumX0  +=       0.140 * cos(5.97498967370  + 27.15773861329 *T)
    sumX0  +=       0.179 * cos(3.80633362749  + 5724.69187994550 *T)
    sumX0  +=       0.133 * cos(4.74386275251  + 21150.56954840009 *T)
    sumX0  +=       0.131 * cos(6.16697098787  + 948.93179148630 *T)
    sumX0  +=       0.132 * cos(0.76534708942  + 1861.98967011891 *T)
    sumX0  +=       0.133 * cos(4.29217117055  + 2472.43494513869 *T)
    sumX0  +=       0.138 * cos(0.40513336194  + 34554.57004618409 *T)
    sumX0  +=       0.139 * cos(5.19191345372  + 3.83424613530 *T)
    sumX0  +=       0.130 * cos(4.35517837249  + 529.91752892591 *T)
    sumX0  +=       0.157 * cos(5.37876152794  + 10787.38652706230 *T)
    sumX0  +=       0.133 * cos(6.07532283158  + 6624.57029594750 *T)
    sumX0  +=       0.148 * cos(3.97691997672  + 7051.16848682350 *T)
    sumX0  +=       0.130 * cos(4.65904545064  + 7594.03891454870 *T)
    sumX0  +=       0.132 * cos(4.88502683468  + 9594.99527173989 *T)
    sumX0  +=       0.138 * cos(0.90971001825  + 12406.40436831210 *T)
    sumX0  +=       0.130 * cos(0.10276469681  + 529.95203623031 *T)
    sumX0  +=       0.143 * cos(4.86125646505  + 10015.03486088490 *T)
    sumX0  +=       0.149 * cos(1.84050714895  + 10.88048283331 *T)
    sumX0  +=       0.178 * cos(3.08917546737  + 2675.61256408629 *T)
    sumX0  +=       0.133 * cos(3.17506743739  + 29424.39041543249 *T)
    sumX0  +=       0.158 * cos(4.74264934391  + 6092.40384387490 *T)
    sumX0  +=       0.131 * cos(3.61329007760  + 158.69970029970 *T)
    sumX0  +=       0.134 * cos(6.18217952972  + 2954.43659361110 *T)
    sumX0  +=       0.142 * cos(5.93252825590  + 3041.24221494710 *T)
    sumX0  +=       0.164 * cos(5.73696427086  + 26083.77798873270 *T)
    sumX0  +=       0.145 * cos(4.90613886816  + 8373.63434038550 *T)
    sumX0  +=       0.172 * cos(6.07544204816  + 1086.53966880249 *T)
    sumX0  +=       0.128 * cos(0.30967307338  + 155.10927164789 *T)
    sumX0  +=       0.137 * cos(5.86605729001  + 6137.55061178690 *T)
    sumX0  +=       0.167 * cos(5.43341203989  + 107.90734142210 *T)
    sumX0  +=       0.133 * cos(0.23815011517  + 12167.75887909109 *T)
    sumX0  +=       0.137 * cos(5.46181201519  + 14581.01897225510 *T)
    sumX0  +=       0.137 * cos(5.81885310871  + 20995.14914896590 *T)
    sumX0  +=       0.126 * cos(4.56442495497  + 37.88921815429 *T)
    sumX0  +=       0.130 * cos(2.37779375339  + 4108.70506353190 *T)
    sumX0  +=       0.137 * cos(4.00489189426  + 18052.68572567429 *T)
    sumX0  +=       0.131 * cos(3.67362007014  + 692.40141871030 *T)
    sumX0  +=       0.132 * cos(4.04092682792  + 27490.44866056129 *T)
    sumX0  +=       0.144 * cos(0.85129982457  + 1802.61580820530 *T)
    sumX0  +=       0.146 * cos(3.74761052478  + 3475.43368925169 *T)
    sumX0  +=       0.126 * cos(1.85382678462  + 18447.73193539770 *T)
    sumX0  +=       0.141 * cos(1.56404343366  + 3716.62197575350 *T)
    sumX0  +=       0.124 * cos(6.09263518701  + 31172.40682849949 *T)
    sumX0  +=       0.124 * cos(3.53492874819  + 6688.15233475721 *T)
    sumX0  +=       0.121 * cos(2.97155384439  + 3590.76070492550 *T)
    sumX0  +=       0.121 * cos(0.01062042066  + 3090.95178344111 *T)
    sumX0  +=       0.122 * cos(1.08207213909  + 2257.59587624631 *T)
    sumX0  +=       0.123 * cos(4.43212874956  + 6710.89817941910 *T)
    sumX0  +=       0.129 * cos(1.62910369718  + 3073.76156749449 *T)
    sumX0  +=       0.131 * cos(1.25301701217  + 5401.67409825510 *T)
    sumX0  +=       0.166 * cos(5.13636005426  + 6553.44982754571 *T)
    sumX0  +=       0.134 * cos(2.74325583907  + 4783.11745294350 *T)
    sumX0  +=       0.122 * cos(3.27465422739  + 1235.83184283950 *T)
    sumX0  +=       0.128 * cos(2.31642376876  + 3081.29543915950 *T)
    sumX0  +=       0.128 * cos(0.66575049629  + 3600.41704920711 *T)
    sumX0  +=       0.121 * cos(1.73585336392  + 2126.05391455991 *T)
    sumX0  +=       0.124 * cos(1.52639839263  + 5820.67110716330 *T)
    sumX0  +=       0.122 * cos(5.72445961080  + 3343.95903786810 *T)
    sumX0  +=       0.140 * cos(0.61176367150  + 10191.17364574889 *T)
    sumX0  +=       0.127 * cos(4.06254894391  + 3759.36056699030 *T)
    sumX0  +=       0.143 * cos(1.09962062110  + 2229.21399908130 *T)
    sumX0  +=       0.120 * cos(4.94879981221  + 1648.69057468091 *T)
    sumX0  +=       0.152 * cos(1.03111182783  + 3340.82173687890 *T)
    sumX0  +=       0.152 * cos(5.09265509081  + 3340.89075148770 *T)
    sumX0  +=       0.137 * cos(0.53719849293  + 4296.09776878290 *T)
    sumX0  +=       0.140 * cos(1.36695700894  + 2233.71743849851 *T)
    sumX0  +=       0.118 * cos(4.24520508177  + 7534.66505263510 *T)
    sumX0  +=       0.137 * cos(4.79183279956  + 14588.06520895310 *T)
    sumX0  +=       0.123 * cos(2.62418379139  + 9381.20349020070 *T)
    sumX0  +=       0.116 * cos(0.31023473755  + 10338.47296723950 *T)
    sumX0  +=       0.137 * cos(2.44663137240  + 8322.80188107970 *T)
    sumX0  +=       0.126 * cos(5.72232356435  + 4083.84630471591 *T)
    sumX0  +=       0.133 * cos(1.77129764655  + 3923.85515556170 *T)
    sumX0  +=       0.124 * cos(0.75314756792  + 9124.67816855889 *T)
    sumX0  +=       0.115 * cos(1.14717177200  + 6740.35489782970 *T)
    sumX0  +=       0.126 * cos(2.90864771221  + 3509.94879234930 *T)
    sumX0  +=       0.121 * cos(3.21008073536  + 3862.94602731009 *T)
    sumX0  +=       0.118 * cos(5.38340024977  + 1052.02456570489 *T)
    sumX0  +=       0.134 * cos(0.59423378961  + 6659.61784155670 *T)
    sumX0  +=       0.133 * cos(5.00091518953  + 28.20536998430 *T)
    sumX0  +=       0.113 * cos(0.40887729438  + 8436.53132064810 *T)
    sumX0  +=       0.133 * cos(2.32355321105  + 6147.84555252540 *T)
    sumX0  +=       0.113 * cos(4.81201285270  + 3116.51144848140 *T)
    sumX0  +=       0.154 * cos(5.94823211725  + 401.42830427370 *T)
    sumX0  +=       0.119 * cos(2.45406070822  + 1471.99684454710 *T)
    sumX0  +=       0.116 * cos(3.08299355103  + 3727.27589475550 *T)
    sumX0  +=       0.120 * cos(1.90192822830  + 3826.34066200590 *T)
    sumX0  +=       0.111 * cos(5.06172588434  + 6837.86939138550 *T)
    sumX0  +=       0.147 * cos(5.92093564346  + 8447.23529630070 *T)
    sumX0  +=       0.120 * cos(2.69810029066  + 2185.00745180929 *T)
    sumX0  +=       0.114 * cos(1.06948367588  + 3966.48138038050 *T)
    sumX0  +=       0.126 * cos(2.58719720597  + 7586.92536754791 *T)
    sumX0  +=       0.129 * cos(4.59879750309  + 3580.06172980831 *T)
    sumX0  +=       0.129 * cos(4.66656206913  + 3101.65075855830 *T)
    sumX0  +=       0.123 * cos(1.60493496065  + 5511.39278450750 *T)
    sumX0  +=       0.133 * cos(3.86628532692  + 6738.36704581871 *T)
    sumX0  +=       0.124 * cos(3.83529166494  + 7210.67200101070 *T)
    sumX0  +=       0.107 * cos(2.18213862825  + 6681.04834621889 *T)
    sumX0  +=       0.139 * cos(2.48335135146  + 14867.49369840869 *T)
    sumX0  +=       0.112 * cos(0.70994163466  + 27433.64539839149 *T)
    sumX0  +=       0.112 * cos(3.44331255043  + 10176.94655174729 *T)
    sumX0  +=       0.126 * cos(3.62916769632  + 6314.24434192070 *T)
    sumX0  +=       0.146 * cos(3.29717245103  + 3282.97754817950 *T)
    sumX0  +=       0.146 * cos(5.96818712120  + 3398.73494018711 *T)
    sumX0  +=       0.112 * cos(6.11830849841  + 4953.80757721490 *T)
    sumX0  +=       0.115 * cos(4.55565863446  + 4179.82553193370 *T)
    sumX0  +=       0.115 * cos(4.70981861448  + 2501.88695643291 *T)
    sumX0  +=       0.130 * cos(5.51502060547  + 6155.70985905160 *T)
    sumX0  +=       0.134 * cos(1.01177681780  + 4825.78873387751 *T)
    sumX0  +=       0.138 * cos(0.56483252235  + 5813.53500680550 *T)
    sumX0  +=       0.115 * cos(4.55081877486  + 8976.92626139531 *T)
    sumX0  +=       0.147 * cos(3.69174223041  + 5223.45010231870 *T)
    sumX0  +=       0.105 * cos(5.41291051440  + 17822.06174308530 *T)
    sumX0  +=       0.118 * cos(4.32124194486  + 5483.01090734249 *T)
    sumX0  +=       0.103 * cos(0.32432296727  + 1059.55843736991 *T)
    sumX0  +=       0.104 * cos(3.26220203929  + 549.80246700989 *T)
    sumX0  +=       0.107 * cos(3.74838397930  + 7076.09455594230 *T)
    sumX0  +=       0.105 * cos(5.49367983392  + 6961.25517523550 *T)
    sumX0  +=       0.129 * cos(5.50018660820  + 2961.48283030911 *T)
    sumX0  +=       0.129 * cos(3.76517296403  + 3720.22965805750 *T)
    sumX0  +=       0.102 * cos(4.98025675709  + 3230.16228799689 *T)
    sumX0  +=       0.122 * cos(2.69290805626  + 27676.86163733250 *T)
    sumX0  +=       0.136 * cos(4.29617974542  + 765.54924696289 *T)
    sumX0  +=       0.116 * cos(0.32033806866  + 10213.04172872750 *T)
    sumX0  +=       0.115 * cos(6.24129475028  + 2631.86644190671 *T)
    sumX0  +=       0.102 * cos(3.50373962589  + 639.65346883050 *T)
    sumX0  +=       0.104 * cos(6.25966851912  + 31450.07086023530 *T)
    sumX0  +=       0.141 * cos(5.37004252010  + 6456.63624021419 *T)
    sumX0  +=       0.138 * cos(4.70210522548  + 12295.71041212569 *T)
    sumX0  +=       0.127 * cos(5.95836944391  + 2655.74487965451 *T)
    sumX0  +=       0.101 * cos(4.76282120148  + 21791.93479074930 *T)
    sumX0  +=       0.103 * cos(5.89952296685  + 3134.18306077910 *T)
    sumX0  +=       0.100 * cos(0.05360387756  + 31213.95761948429 *T)
    sumX0  +=       0.115 * cos(2.72333533541  + 21000.67207327329 *T)
    sumX0  +=       0.138 * cos(4.45448589312  + 3201.15810466050 *T)
    sumX0  +=       0.138 * cos(4.81087367911  + 3480.55438370610 *T)
    sumX0  +=       0.099 * cos(2.15773607989  + 6953.65399326770 *T)
    sumX0  +=       0.129 * cos(0.22745708157  + 4878.60904519429 *T)
    sumX0  +=       0.106 * cos(2.17957220226  + 3023.97673955970 *T)
    sumX0  +=       0.100 * cos(1.77158830078  + 8741.79888998789 *T)
    sumX0  +=       0.100 * cos(1.77158830078  + 8742.28652495491 *T)
    sumX0  +=       0.101 * cos(3.30035420675  + 6534.65698435950 *T)
    sumX0  +=       0.100 * cos(2.36711969354  + 2652.22176130550 *T)
    sumX0  +=       0.108 * cos(5.70132621502  + 250.14827822570 *T)
    sumX0  +=       0.113 * cos(5.46854342462  + 13937.03832261750 *T)
    sumX0  +=       0.100 * cos(5.39233849058  + 3340.77168022831 *T)
    sumX0  +=       0.100 * cos(0.73142842806  + 3340.94080813830 *T)
    sumX0  +=       0.098 * cos(6.11395858344  + 6144.17652382070 *T)
    sumX0  +=       0.104 * cos(2.06636969565  + 11526.88127170889 *T)
    sumX0  +=       0.099 * cos(1.07049660080  + 5312.80841803811 *T)
    sumX0  +=       0.113 * cos(0.61738035609  + 10080.96732452950 *T)
    sumX0  +=       0.107 * cos(2.59063501509  + 6680.91372561330 *T)
    sumX0  +=       0.095 * cos(4.93046121541  + 8425.40702033129 *T)
    sumX0  +=       0.095 * cos(4.56982042451  + 3310.97420057310 *T)
    sumX0  +=       0.095 * cos(4.69553914772  + 3370.73828779350 *T)
    sumX0  +=       0.098 * cos(4.50493848722  + 18616.82448356369 *T)
    sumX0  +=       0.095 * cos(0.79771279792  + 4783.13470659570 *T)
    sumX0  +=       0.096 * cos(5.71058433425  + 1353.14671052430 *T)
    sumX0  +=       0.100 * cos(4.98384210374  + 3340.90630083390 *T)
    sumX0  +=       0.100 * cos(1.13992481490  + 3340.80618753271 *T)
    sumX0  +=       0.112 * cos(5.73113213169  + 5084.86190393429 *T)
    sumX0  +=       0.098 * cos(5.27414574207  + 7277.64704489211 *T)
    sumX0  +=       0.097 * cos(6.16604306709  + 12509.00950756369 *T)
    sumX0  +=       0.100 * cos(5.37057272396  + 1755.22786719560 *T)
    sumX0  +=       0.098 * cos(2.55063075141  + 1784.28809722790 *T)
    sumX0  +=       0.129 * cos(4.50770814605  + 4591.68644048410 *T)
    sumX0  +=       0.098 * cos(4.36330521693  + 4981.70181941289 *T)
    sumX0  +=       0.100 * cos(0.88244732253  + 6474.76793358650 *T)
    sumX0  +=       0.125 * cos(3.41428539212  + 30228.22229391390 *T)
    sumX0  +=       0.102 * cos(4.70479717845  + 2214.98690507971 *T)
    sumX0  +=       0.092 * cos(3.28323563949  + 1383.13116433011 *T)
    sumX0  +=       0.101 * cos(4.40159342227  + 3302.23557357850 *T)
    sumX0  +=       0.100 * cos(2.64848792786  + 13517.55897844709 *T)
    sumX0  +=       0.096 * cos(5.28550942561  + 16975.97890833410 *T)
    sumX0  +=       0.096 * cos(1.50961429884  + 34114.87025179110 *T)
    sumX0  +=       0.118 * cos(0.67935665521  + 3343.82518959990 *T)
    sumX0  +=       0.118 * cos(2.30281760984  + 3337.88729876671 *T)
    sumX0  +=       0.105 * cos(5.50998714425  + 42389.17875379050 *T)
    sumX0  +=       0.118 * cos(0.33297168104  + 2524.20291796811 *T)
    sumX0  +=       0.114 * cos(1.85606000482  + 947.95147041809 *T)
    sumX0  +=       0.102 * cos(5.06357443797  + 27946.99179740330 *T)
    sumX0  +=       0.103 * cos(4.21339332479  + 20235.36664379391 *T)
    sumX0  +=       0.094 * cos(1.77869076759  + 11507.01358727710 *T)
    sumX0  +=       0.118 * cos(3.61771545773  + 3284.47305810711 *T)
    sumX0  +=       0.118 * cos(5.64764411450  + 3397.23943025950 *T)
    sumX0  +=       0.115 * cos(1.48299242949  + 640.63378989870 *T)
    sumX0  +=       0.104 * cos(0.07874415195  + 3149.42523172390 *T)
    sumX0  +=       0.104 * cos(6.04497292118  + 3532.28725664270 *T)
    sumX0  +=       0.091 * cos(0.29342992971  + 24093.03297169169 *T)
    sumX0  +=       0.093 * cos(2.25704229812  + 987.99009041150 *T)
    sumX0  +=       0.093 * cos(3.19860248214  + 1957.48126236970 *T)
    sumX0  +=       0.116 * cos(5.06230565784  + 49225.82400662430 *T)
    sumX0  +=       0.099 * cos(1.45477069665  + 6489.53276980449 *T)
    sumX0  +=       0.113 * cos(3.26669234129  + 2597.86618365071 *T)
    sumX0  +=       0.107 * cos(1.04892077554  + 1744.66980189871 *T)
    sumX0  +=       0.094 * cos(5.35074037672  + 3334.72301825071 *T)
    sumX0  +=       0.094 * cos(3.91461919551  + 3346.98947011590 *T)
    sumX0  +=       0.090 * cos(0.96613791810  + 15747.01679501190 *T)
    sumX0  +=       0.091 * cos(2.71602822239  + 50818.42002025710 *T)
    sumX0  +=       0.108 * cos(2.01528048558  + 1971.70835637129 *T)
    sumX0  +=       0.111 * cos(4.02376942461  + 27417.30083230870 *T)
    sumX0  +=       0.104 * cos(5.13876361969  + 441.06910236111 *T)
    sumX0  +=       0.089 * cos(0.76619322850  + 3313.38737778371 *T)
    sumX0  +=       0.089 * cos(2.21598103655  + 3368.32511058290 *T)
    sumX0  +=       0.103 * cos(6.19901188761  + 1059.69305797550 *T)
    sumX0  +=       0.112 * cos(4.36311428624  + 1744.73711220151 *T)
    sumX0  +=       0.092 * cos(1.76217757007  + 32509.45279042450 *T)
    sumX0  +=       0.100 * cos(5.42847993770  + 10611.14592459111 *T)
    sumX0  +=       0.112 * cos(6.05049937833  + 10721.35224581050 *T)
    sumX0  +=       0.100 * cos(4.15719372404  + 309.03450517230 *T)
    sumX0  +=       0.105 * cos(4.65669488804  + 24606.37937070350 *T)
    sumX0  +=       0.095 * cos(2.41537501026  + 529.86747227531 *T)
    sumX0  +=       0.120 * cos(1.39317335094  + 3202.33874731261 *T)
    sumX0  +=       0.120 * cos(1.58900091411  + 3479.37374105400 *T)
    sumX0  +=       0.086 * cos(2.34324816413  + 2918.82880302731 *T)
    sumX0  +=       0.086 * cos(0.63892610092  + 3762.88368533930 *T)
    sumX0  +=       0.099 * cos(3.62602219574  + 3458.66904707530 *T)
    sumX0  +=       0.099 * cos(5.63933737649  + 3223.04344129130 *T)
    sumX0  +=       0.089 * cos(3.31586686845  + 22531.33442263010 *T)
    sumX0  +=       0.107 * cos(5.26232652340  + 3311.87461420391 *T)
    sumX0  +=       0.107 * cos(4.00303304883  + 3369.83787416271 *T)
    sumX0  +=       0.090 * cos(4.95245301445  + 4670.36833444331 *T)
    sumX0  +=       0.094 * cos(3.52526410634  + 796.52457064770 *T)
    sumX0  +=       0.085 * cos(4.38880338786  + 809.23365196350 *T)
    sumX0  +=       0.084 * cos(4.57429862732  + 4778.03185838191 *T)
    sumX0  +=       0.118 * cos(0.22160767097  + 8219.22147189409 *T)
    sumX0  +=       0.104 * cos(4.33115782764  + 3149.39072441950 *T)
    sumX0  +=       0.104 * cos(1.79255924549  + 3532.32176394710 *T)
    sumX0  +=       0.084 * cos(2.65547973490  + 1125.62552162009 *T)
    sumX0  +=       0.085 * cos(1.24355302775  + 4555.65857380670 *T)
    sumX0  +=       0.116 * cos(0.77903465821  + 14694.74841182350 *T)
    sumX0  +=       0.086 * cos(1.37625658794  + 12410.97511803210 *T)
    sumX0  +=       0.084 * cos(2.20272544801  + 1106.89498820129 *T)
    sumX0  +=       0.117 * cos(1.12167690201  + 30774.25782509130 *T)
    sumX0  +=       0.090 * cos(4.84535420677  + 32366.85383872410 *T)
    sumX0  +=       0.091 * cos(0.58267443363  + 9161.77116883010 *T)
    sumX0  +=       0.089 * cos(1.52114958439  + 3348.53003615871 *T)
    sumX0  +=       0.089 * cos(1.46102468066  + 3333.18245220790 *T)
    sumX0  +=       0.093 * cos(3.90251764282  + 9225.29545579949 *T)
    sumX0  +=       0.102 * cos(4.73756255850  + 4466.72558328691 *T)
    sumX0  +=       0.102 * cos(1.59596990491  + 4466.23794831990 *T)
    sumX0  +=       0.099 * cos(2.08209672992  + 4879.09668016130 *T)
    sumX0  +=       0.086 * cos(2.48786705786  + 3613.04156656790 *T)
    sumX0  +=       0.083 * cos(2.77680182693  + 7899.30648388849 *T)
    sumX0  +=       0.082 * cos(6.02206548524  + 1265.32366114289 *T)
    sumX0  +=       0.084 * cos(2.49874468506  + 2942.77455107790 *T)
    sumX0  +=       0.083 * cos(2.13736601614  + 3077.28468584349 *T)
    sumX0  +=       0.082 * cos(2.21746232443  + 3351.42559923031 *T)
    sumX0  +=       0.082 * cos(0.76471194062  + 3330.28688913630 *T)
    sumX0  +=       0.112 * cos(3.24551167664  + 14900.93396026070 *T)
    sumX0  +=       0.082 * cos(2.15484008375  + 22318.03532719209 *T)
    sumX0  +=       0.084 * cos(4.77917502205  + 3597.87425192630 *T)
    sumX0  +=       0.084 * cos(4.48618455018  + 3083.83823644030 *T)
    sumX0  +=       0.082 * cos(1.35034038247  + 1552.22371249610 *T)
    sumX0  +=       0.095 * cos(2.00687862342  + 530.00209288090 *T)
    sumX0  +=       0.081 * cos(2.73020648895  + 2980.25782725470 *T)
    sumX0  +=       0.085 * cos(0.15610612016  + 1056.10262932370 *T)
    sumX0  +=       0.097 * cos(1.61255624836  + 4447.99504986810 *T)
    sumX0  +=       0.084 * cos(2.96706110524  + 845.83901726770 *T)
    sumX0  +=       0.114 * cos(2.85738890734  + 3063.82125044190 *T)
    sumX0  +=       0.114 * cos(0.12478535771  + 3617.89123792471 *T)
    sumX0  +=       0.085 * cos(1.29452650597  + 23937.61257225750 *T)
    sumX0  +=       0.085 * cos(4.25238937662  + 54790.18050746150 *T)
    sumX0  +=       0.086 * cos(4.98647530125  + 23010.19292841890 *T)
    sumX0  +=       0.085 * cos(1.10962056312  + 8322.31424611270 *T)
    sumX0  +=       0.083 * cos(3.80381782412  + 20103.29223959590 *T)
    sumX0  +=       0.112 * cos(5.31910031613  + 5295.57339454690 *T)
    sumX0  +=       0.081 * cos(2.24487635939  + 3490.75371931010 *T)
    sumX0  +=       0.081 * cos(0.73729790566  + 3190.95876905650 *T)
    sumX0  +=       0.099 * cos(1.21508005987  + 4845.65641830930 *T)
    sumX0  +=       0.094 * cos(5.55603573783  + 796.55907795210 *T)
    sumX0  +=       0.092 * cos(1.95151196671  + 4452.49848928530 *T)
    sumX0  +=       0.092 * cos(5.09310462030  + 4452.01085431830 *T)
    sumX0  +=       0.091 * cos(0.14656244570  + 2544.54098371471 *T)
    sumX0  +=       0.081 * cos(0.07757124603  + 7166.95308870570 *T)
    sumX0  +=       0.089 * cos(5.72126604208  + 6311.76885494270 *T)
    sumX0  +=       0.084 * cos(0.49185882297  + 4822.26561552850 *T)
    sumX0  +=       0.089 * cos(3.02340140636  + 4049.84604645990 *T)
    sumX0  +=       0.085 * cos(0.12419247098  + 14716.08405229050 *T)
    sumX0  +=       0.081 * cos(1.13581368137  + 34363.12178007249 *T)
    sumX0  +=       0.094 * cos(5.59053061063  + 24532.73276608469 *T)
    sumX0  +=       0.102 * cos(3.18943570469  + 3524.72583108530 *T)
    sumX0  +=       0.102 * cos(6.07592386754  + 3156.98665728130 *T)
    sumX0  +=       0.084 * cos(2.90724107191  + 2942.63993047231 *T)
    sumX0  +=       0.084 * cos(4.30324979650  + 4801.91029612970 *T)
    sumX0  +=       0.084 * cos(4.20311871432  + 78263.95324220609 *T)
    sumX0  +=       0.079 * cos(1.90010229816  + 4424.11661212030 *T)
    sumX0  +=       0.081 * cos(3.35310168176  + 3333.72544353031 *T)
    sumX0  +=       0.081 * cos(5.91225789047  + 3347.98704483631 *T)
    sumX0  +=       0.079 * cos(0.61325160212  + 4029.49072706111 *T)
    sumX0  +=       0.083 * cos(1.86215319352  + 24336.24921063270 *T)
    sumX0  +=       0.102 * cos(0.51044670726  + 3119.95596677750 *T)
    sumX0  +=       0.096 * cos(1.45358314071  + 3324.02404313351 *T)
    sumX0  +=       0.096 * cos(1.52859112434  + 3357.68844523310 *T)
    sumX0  +=       0.080 * cos(2.69227009829  + 12987.86801335249 *T)
    sumX0  +=       0.099 * cos(5.59593346094  + 3376.46403476710 *T)
    sumX0  +=       0.099 * cos(3.66942611129  + 3305.24845359951 *T)
    sumX0  +=       0.090 * cos(5.64056301828  + 18028.24729152250 *T)
    sumX0  +=       0.086 * cos(4.08226718914  + 3296.20241094311 *T)
    sumX0  +=       0.086 * cos(5.18309238309  + 3385.51007742350 *T)
    sumX0  +=       0.085 * cos(0.33690474046  + 3487.65067705471 *T)
    sumX0  +=       0.085 * cos(2.64526952459  + 3194.06181131190 *T)
    sumX0  +=       0.085 * cos(5.64237858602  + 1758.89689590030 *T)
    sumX0  +=       0.088 * cos(1.84603493900  + 2751.25897466350 *T)
    sumX0  +=       0.082 * cos(5.31753219651  + 16460.57734700850 *T)
    sumX0  +=       0.101 * cos(0.12565896078  + 1162.23088692429 *T)
    sumX0  +=       0.079 * cos(1.54738520834  + 8186.26884500910 *T)
    sumX0  +=       0.097 * cos(3.30698966692  + 4025.96760871210 *T)
    sumX0  +=       0.077 * cos(4.93338786819  + 7446.77410036250 *T)
    sumX0  +=       0.077 * cos(1.79179521460  + 7447.26173532950 *T)
    sumX0  +=       0.077 * cos(2.27672892520  + 430.77416162260 *T)
    sumX0  +=       0.102 * cos(0.88442764654  + 574.58861581830 *T)
    sumX0  +=       0.079 * cos(5.30896262392  + 1593.24886603040 *T)
    sumX0  +=       0.090 * cos(6.20338338719  + 12891.88878613469 *T)
    sumX0  +=       0.079 * cos(4.62780146592  + 101010.75632211348 *T)
    sumX0  +=       0.100 * cos(1.08134927193  + 1481.16555386170 *T)
    sumX0  +=       0.106 * cos(3.21478248142  + 13.98327651810 *T)
    sumX0  +=       0.076 * cos(0.77814474339  + 31979.76182532990 *T)
    sumX0  +=       0.102 * cos(0.64416843915  + 52168.28742991590 *T)
    sumX0  +=       0.103 * cos(4.98287803658  + 6777.44789810091 *T)
    sumX0  +=       0.080 * cos(4.95264640866  + 1954.96096784710 *T)
    sumX0  +=       0.075 * cos(2.02296081747  + 14797.84118604210 *T)
    sumX0  +=       0.091 * cos(4.16695567867  + 6836.33412504750 *T)
    sumX0  +=       0.077 * cos(2.05190836413  + 51770.13842650770 *T)
    sumX0  +=       0.078 * cos(5.96632605031  + 8831.15720570730 *T)
    sumX0  +=       0.085 * cos(1.00204631140  + 369.45599845690 *T)
    sumX0  +=       0.091 * cos(2.17733407719  + 2544.57549101910 *T)
    sumX0  +=       0.081 * cos(2.52401673663  + 3277.12034587990 *T)
    sumX0  +=       0.081 * cos(0.45815752842  + 3404.59214248671 *T)
    sumX0  +=       0.076 * cos(2.28366173471  + 10025.11658096489 *T)
    sumX0  +=       0.084 * cos(1.03922100960  + 45331.64217708210 *T)
    sumX0  +=       0.077 * cos(5.30919338072  + 1324.69752305650 *T)
    sumX0  +=       0.091 * cos(4.40755313323  + 45494.33761226529 *T)
    sumX0  +=       0.081 * cos(3.60186938439  + 3325.11613803129 *T)
    sumX0  +=       0.080 * cos(3.36745124689  + 569.06569151090 *T)
    sumX0  +=       0.075 * cos(0.53983177923  + 9090.71801073110 *T)
    sumX0  +=       0.090 * cos(1.60142738772  + 1747.36356066940 *T)
    sumX0  +=       0.073 * cos(2.65616186351  + 7476.23141136150 *T)
    sumX0  +=       0.091 * cos(0.10088212205  + 8536.10097651771 *T)
    sumX0  +=       0.091 * cos(3.24247477564  + 8535.61334155070 *T)
    sumX0  +=       0.102 * cos(3.02727638000  + 8692.08137235589 *T)
    sumX0  +=       0.073 * cos(0.49752926048  + 16062.42834360030 *T)
    sumX0  +=       0.087 * cos(5.83216237330  + 15117.75617404771 *T)
    sumX0  +=       0.077 * cos(3.04573883254  + 20618.26317601710 *T)
    sumX0  +=       0.086 * cos(4.85081525532  + 5355.47969897210 *T)
    sumX0  +=       0.081 * cos(5.67577359419  + 2917.29353668930 *T)
    sumX0  +=       0.077 * cos(5.62600904801  + 1008.30565339989 *T)
    sumX0  +=       0.081 * cos(0.08809861865  + 2910.24729999130 *T)
    sumX0  +=       0.072 * cos(5.81652451955  + 928.08378598631 *T)
    sumX0  +=       0.073 * cos(4.57978609101  + 6809.48751422050 *T)
    sumX0  +=       0.075 * cos(4.66223023270  + 4106.16167366270 *T)
    sumX0  +=       0.074 * cos(2.51037789906  + 30376.10882168309 *T)
    sumX0  +=       0.094 * cos(4.31600433309  + 17809.95712170029 *T)
    sumX0  +=       0.080 * cos(3.70927773647  + 6151.84501609130 *T)
    sumX0  +=       0.075 * cos(1.66172010069  + 81604.56566890588 *T)
    sumX0  +=       0.076 * cos(4.72368572998  + 44534.85653529869 *T)
    sumX0  +=       0.097 * cos(4.28214074273  + 5749.61794906429 *T)
    sumX0  +=       0.071 * cos(2.42412250787  + 9384.59719059169 *T)
    sumX0  +=       0.074 * cos(5.57898561678  + 6603.71812689920 *T)
    sumX0  +=       0.071 * cos(5.34550834972  + 14061.96467251030 *T)
    sumX0  +=       0.073 * cos(5.34904276015  + 13635.36648163430 *T)
    sumX0  +=       0.071 * cos(5.78968211370  + 5622.15405099670 *T)
    sumX0  +=       0.071 * cos(5.87925588598  + 23539.88389351350 *T)
    sumX0  +=       0.081 * cos(2.95315282923  + 19573.60127450130 *T)
    sumX0  +=       0.100 * cos(2.10255812973  + 3399.74247112990 *T)
    sumX0  +=       0.075 * cos(1.83278893940  + 1214.55851213989 *T)
    sumX0  +=       0.094 * cos(1.75965102296  + 10816.84383806129 *T)
    sumX0  +=       0.071 * cos(5.95705764996  + 12140.84495796130 *T)
    sumX0  +=       0.072 * cos(2.94505111531  + 26887.60986721410 *T)
    sumX0  +=       0.072 * cos(5.27569870166  + 2010.85651895630 *T)
    sumX0  +=       0.072 * cos(2.76090980030  + 11560.32153356091 *T)
    sumX0  +=       0.072 * cos(3.91771440218  + 21957.43691026350 *T)
    sumX0  +=       0.071 * cos(2.84795147281  + 1619.75375224610 *T)
    sumX0  +=       0.074 * cos(0.11394719873  + 20205.89737884749 *T)
    sumX0  +=       0.075 * cos(1.13246079000  + 7477.27904273249 *T)
    sumX0  +=       0.073 * cos(2.21781912263  + 6493.54352312051 *T)
    sumX0  +=       0.069 * cos(1.97183985363  + 5784.06574185909 *T)
    sumX0  +=       0.084 * cos(5.72884329247  + 2118.52004289490 *T)
    sumX0  +=       0.086 * cos(4.60823055211  + 19404.02109136830 *T)
    sumX0  +=       0.095 * cos(3.88399746483  + 2104.78058386030 *T)
    sumX0  +=       0.075 * cos(1.20420559875  + 8748.91243698870 *T)
    sumX0  +=       0.076 * cos(5.90435441090  + 3268.80051082850 *T)
    sumX0  +=       0.076 * cos(3.36100516133  + 3412.91197753810 *T)
    sumX0  +=       0.089 * cos(0.07046438358  + 15643.43638582629 *T)
    sumX0  +=       0.075 * cos(2.83027542239  + 36106.10290379910 *T)
    sumX0  +=       0.069 * cos(4.22526525232  + 21272.88049100450 *T)
    sumX0  +=       0.071 * cos(1.81785521705  + 58857.27495403150 *T)
    sumX0  +=       0.071 * cos(3.09739469496  + 5298.09368906950 *T)
    sumX0  +=       0.071 * cos(6.23898734855  + 5298.58132403650 *T)
    sumX0  +=       0.080 * cos(4.11777412331  + 6151.71039548571 *T)
    sumX0  +=       0.068 * cos(4.63938903003  + 942.30587945250 *T)
    sumX0  +=       0.089 * cos(4.89960612432  + 30640.10045611650 *T)
    sumX0  +=       0.068 * cos(5.87799590803  + 11247.45278225330 *T)
    sumX0  +=       0.068 * cos(0.82684530792  + 8550.32807051931 *T)
    sumX0  +=       0.078 * cos(5.32553737695  + 422.90985509640 *T)
    sumX0  +=       0.068 * cos(3.37824489645  + 4665.30994975630 *T)
    sumX0  +=       0.068 * cos(0.23665224286  + 4665.79758472330 *T)
    sumX0  +=       0.070 * cos(1.18575062405  + 2332.30677329990 *T)
    sumX0  +=       0.070 * cos(4.74024249826  + 13105.18788157269 *T)
    sumX0  +=       0.071 * cos(0.83016269932  + 3309.86425943471 *T)
    sumX0  +=       0.071 * cos(2.15201156573  + 3371.84822893190 *T)
    sumX0  +=       0.080 * cos(4.96945014581  + 9705.68922792630 *T)
    sumX0  +=       0.079 * cos(2.82952117654  + 16065.82204399129 *T)
    sumX0  +=       0.070 * cos(2.49836750115  + 7843.94337529091 *T)
    sumX0  +=       0.068 * cos(5.11241672674  + 9175.99826283170 *T)
    sumX0  +=       0.081 * cos(0.77055887901  + 94329.53146871387 *T)
    sumX0  +=       0.067 * cos(5.57581450600  + 6112.64679685570 *T)
    sumX0  +=       0.075 * cos(1.66052000409  + 1585.38455950421 *T)
    sumX0  +=       0.091 * cos(2.77249023737  + 7768.25215716910 *T)
    sumX0  +=       0.068 * cos(1.17229514872  + 11663.41430777950 *T)
    sumX0  +=       0.069 * cos(2.60276940326  + 23554.11098751510 *T)
    sumX0  +=       0.072 * cos(0.87830980344  + 3267.22078096370 *T)
    sumX0  +=       0.072 * cos(2.10386446161  + 3414.49170740290 *T)
    sumX0  +=       0.073 * cos(3.91879555351  + 2693.84541086851 *T)
    sumX0  +=       0.069 * cos(1.24929360061  + 3494.78147770770 *T)
    sumX0  +=       0.069 * cos(1.73288066444  + 3186.93101065891 *T)
    sumX0  +=       0.068 * cos(0.65498506864  + 5202.11446185170 *T)
    sumX0  +=       0.085 * cos(2.64920255148  + 4157.50957039850 *T)
    sumX0  +=       0.066 * cos(5.42206372629  + 15636.32283882549 *T)
    sumX0  +=       0.093 * cos(3.60627494992  + 1826.07731249979 *T)
    sumX0  +=       0.075 * cos(5.40680087602  + 15276.21205686389 *T)
    sumX0  +=       0.067 * cos(2.59734357148  + 290.24166198609 *T)
    sumX0  +=       0.073 * cos(2.30862354300  + 3237.27583499770 *T)
    sumX0  +=       0.066 * cos(2.80946310059  + 10823.95738506209 *T)
    sumX0  +=       0.066 * cos(5.95105575418  + 10824.44502002911 *T)
    sumX0  +=       0.066 * cos(3.87880650690  + 16056.29511766771 *T)
    sumX0  +=       0.066 * cos(3.36268639559  + 226.07308589371 *T)
    sumX0  +=       0.068 * cos(0.42764873920  + 4897.42439113870 *T)
    sumX0  +=       0.065 * cos(3.82963818162  + 3324.09135343631 *T)
    sumX0  +=       0.065 * cos(5.43572139061  + 3357.62113493030 *T)
    sumX0  +=       0.066 * cos(2.06780245805  + 176.44991265029 *T)
    sumX0  +=       0.069 * cos(1.31713075410  + 1442.03464492889 *T)
    sumX0  +=       0.065 * cos(2.23384330119  + 12725.20961729149 *T)
    sumX0  +=       0.071 * cos(0.81327740269  + 11250.55557593809 *T)
    sumX0  +=       0.066 * cos(0.93383696418  + 415.79630809560 *T)
    sumX0  +=       0.066 * cos(3.64037605585  + 3738.93793728871 *T)
    sumX0  +=       0.090 * cos(1.66801934841  + 2885.94687765600 *T)
    sumX0  +=       0.090 * cos(1.31415491664  + 3795.76561071060 *T)
    sumX0  +=       0.067 * cos(5.42902273412  + 13359.10309563091 *T)
    sumX0  +=       0.069 * cos(4.43980027024  + 4932.96462284910 *T)
    sumX0  +=       0.067 * cos(1.57727720508  + 20043.43074271529 *T)
    sumX0  +=       0.067 * cos(3.84155718574  + 19247.62037086590 *T)
    sumX0  +=       0.064 * cos(1.06581237341  + 8983.05448679250 *T)
    sumX0  +=       0.081 * cos(5.58958887514  + 5136.11468790430 *T)
    sumX0  +=       0.076 * cos(6.10249673727  + 6677.53651762000 *T)
    sumX0  +=       0.064 * cos(0.33226555661  + 6564.14880266290 *T)
    sumX0  +=       0.063 * cos(2.10390607796  + 10294.26641996750 *T)
    sumX0  +=       0.063 * cos(5.24549873155  + 10294.75405493450 *T)
    sumX0  +=       0.065 * cos(5.56809088940  + 0.73650358470 *T)
    sumX0  +=       0.067 * cos(0.73015466933  + 4396.71505602351 *T)
    sumX0  +=       0.070 * cos(4.18413149495  + 5088.46362233621 *T)
    sumX0  +=       0.070 * cos(1.04253884136  + 5087.97598736919 *T)
    sumX0  +=       0.066 * cos(3.51965735519  + 3332.68810760091 *T)
    sumX0  +=       0.066 * cos(5.74570221704  + 3349.02438076571 *T)
    sumX0  +=       0.069 * cos(3.86382217258  + 24076.68840560890 *T)
    sumX0  +=       0.063 * cos(4.22297067160  + 15004.02673447930 *T)
    sumX0  +=       0.066 * cos(4.86919893241  + 3303.86065357230 *T)
    sumX0  +=       0.066 * cos(4.39616063982  + 3377.85183479431 *T)
    sumX0  +=       0.062 * cos(0.97042406466  + 423.31889001050 *T)
    sumX0  +=       0.073 * cos(1.49145392699  + 7250.29054491051 *T)
    sumX0  +=       0.066 * cos(2.48697755670  + 1083.01655045349 *T)
    sumX0  +=       0.066 * cos(3.23187966901  + 3739.07255789430 *T)
    sumX0  +=       0.065 * cos(3.95948324523  + 3454.58568375171 *T)
    sumX0  +=       0.065 * cos(5.30587632700  + 3227.12680461490 *T)
    sumX0  +=       0.069 * cos(2.59648260174  + 5885.18791771890 *T)
    sumX0  +=       0.069 * cos(5.75230198725  + 2384.56708821271 *T)
    sumX0  +=       0.082 * cos(3.81994976939  + 6573.80514694450 *T)
    sumX0  +=       0.078 * cos(5.28528678757  + 12979.79699205951 *T)
    sumX0  +=       0.063 * cos(5.02148754857  + 6298.08450369290 *T)
    sumX0  +=       0.063 * cos(4.29059352218  + 379.12959639069 *T)
    sumX0  +=       0.078 * cos(0.06919505661  + 17079.07168255270 *T)
    sumX0  +=       0.063 * cos(5.55828474147  + 4295.61013381590 *T)
    sumX0  +=       0.072 * cos(2.77553583313  + 7366.51272510910 *T)
    sumX0  +=       0.064 * cos(1.22906259266  + 21258.65339700290 *T)
    sumX0  +=       0.061 * cos(5.82279175439  + 3870.52995562571 *T)
    sumX0  +=       0.069 * cos(2.31182487954  + 6757.73474215870 *T)
    sumX0  +=       0.061 * cos(1.57037807871  + 3870.56446293010 *T)
    sumX0  +=       0.061 * cos(2.32660945835  + 8528.49979454989 *T)
    sumX0  +=       0.072 * cos(1.46495213208  + 3337.95522989351 *T)
    sumX0  +=       0.072 * cos(1.51722213297  + 3343.75725847310 *T)
    sumX0  +=       0.081 * cos(4.85945415811  + 17498.55632642790 *T)
    sumX0  +=       0.080 * cos(4.09102247218  + 68241.62832713968 *T)
    sumX0  +=       0.070 * cos(1.26769104700  + 5096.32792886240 *T)
    sumX0  +=       0.070 * cos(4.40928370059  + 5095.84029389540 *T)
    sumX0  +=       0.062 * cos(3.77365809945  + 629.09540353360 *T)
    sumX0  +=       0.061 * cos(4.23156701226  + 7056.74676748629 *T)
    sumX0  +=       0.061 * cos(2.12062311073  + 27035.49639498329 *T)
    sumX0  +=       0.075 * cos(5.58983904787  + 13531.78607244869 *T)
    sumX0  +=       0.062 * cos(1.40905160179  + 14591.16800263789 *T)
    sumX0  +=       0.062 * cos(4.55064425538  + 14591.65563760490 *T)
    sumX0  +=       0.071 * cos(6.19817850055  + 5622.01943039111 *T)
    sumX0  +=       0.066 * cos(0.70285666709  + 954.99770711610 *T)
    sumX0  +=       0.068 * cos(4.29528632626  + 3930.45351370310 *T)
    sumX0  +=       0.066 * cos(1.57540103682  + 2899.54332433869 *T)
    sumX0  +=       0.061 * cos(2.92470112475  + 2711.51702316619 *T)
    sumX0  +=       0.063 * cos(2.75001640915  + 2125.56627959290 *T)
    sumX0  +=       0.073 * cos(6.02852675275  + 6695.69576488470 *T)
    sumX0  +=       0.069 * cos(0.56571097024  + 5885.15341041451 *T)
    sumX0  +=       0.060 * cos(3.57090765756  + 41427.24316569529 *T)
    sumX0  +=       0.080 * cos(2.18652340630  + 4576.44426953929 *T)
    sumX0  +=       0.080 * cos(5.32811605989  + 4576.93190450630 *T)
    sumX0  +=       0.059 * cos(0.92319299629  + 9374.58262930109 *T)
    sumX0  +=       0.074 * cos(3.32088958533  + 3583.09721319030 *T)
    sumX0  +=       0.082 * cos(4.24713604667  + 2636.73542459890 *T)
    sumX0  +=       0.058 * cos(3.21277912522  + 720.89541372269 *T)
    sumX0  +=       0.058 * cos(3.21277912522  + 721.38304868970 *T)
    sumX0  +=       0.059 * cos(2.00330526657  + 14319.44721987369 *T)
    sumX0  +=       0.065 * cos(2.85461191452  + 48274.10560037370 *T)
    sumX0  +=       0.058 * cos(2.94023903950  + 110.94282480410 *T)
    sumX0  +=       0.072 * cos(3.25665693297  + 195.21520194411 *T)
    sumX0  +=       0.059 * cos(1.03271362744  + 13738.45925585290 *T)
    sumX0  +=       0.063 * cos(2.63685401817  + 9722.95470331370 *T)
    sumX0  +=       0.080 * cos(6.03050339434  + 3304.20768125380 *T)
    sumX0  +=       0.080 * cos(3.23485617788  + 3377.50480711280 *T)
    sumX0  +=       0.058 * cos(0.73211712408  + 16070.52221166930 *T)
    sumX0  +=       0.070 * cos(4.06333381597  + 11787.84772300050 *T)
    sumX0  +=       0.062 * cos(3.37913079081  + 4558.69405718870 *T)
    sumX0  +=       0.076 * cos(0.42812166341  + 1612.70751554809 *T)
    sumX0  +=       0.071 * cos(2.87138479505  + 6486.49728642250 *T)
    sumX0  +=       0.062 * cos(0.38524528784  + 437.88770862340 *T)
    sumX0  +=       0.059 * cos(3.33501995812  + 37057.33367508269 *T)
    sumX0  +=       0.070 * cos(1.22310296909  + 8116.12869767550 *T)
    sumX0  +=       0.058 * cos(1.38401229102  + 1969.44448072730 *T)
    sumX0  +=       0.058 * cos(1.59816197403  + 4712.26800763931 *T)
    sumX0  +=       0.057 * cos(2.80514625199  + 4565.74029388670 *T)
    sumX0  +=       0.065 * cos(3.23325458609  + 5526.10751347610 *T)
    sumX0  +=       0.072 * cos(1.75331960583  + 8948.47207339210 *T)
    sumX0  +=       0.059 * cos(4.01427864314  + 22854.83983928750 *T)
    sumX0  +=       0.074 * cos(0.45536940410  + 3192.28988479000 *T)
    sumX0  +=       0.074 * cos(5.73581209791  + 3189.32093937340 *T)
    sumX0  +=       0.074 * cos(1.18271789974  + 5106.13523463390 *T)
    sumX0  +=       0.060 * cos(3.98760857931  + 3333.27756497370 *T)
    sumX0  +=       0.060 * cos(5.27775099292  + 3348.43492339290 *T)
    sumX0  +=       0.057 * cos(2.42715474002  + 1022.53274740150 *T)
    sumX0  +=       0.062 * cos(5.31656329809  + 792.53107098389 *T)
    sumX0  +=       0.055 * cos(5.92010875776  + 3355.59352650890 *T)
    sumX0  +=       0.055 * cos(3.34525081447  + 3326.11896185771 *T)
    sumX0  +=       0.057 * cos(3.55845650996  + 3973.15234852950 *T)
    sumX0  +=       0.057 * cos(2.89164664491  + 15842.99602222970 *T)
    sumX0  +=       0.071 * cos(4.64519741711  + 418.26050532349 *T)
    sumX0  +=       0.061 * cos(5.33074649278  + 223.10414047710 *T)
    sumX0  +=       0.063 * cos(5.97450695245  + 4137.17150465190 *T)
    sumX0  +=       0.067 * cos(0.28907282931  + 11240.40654555530 *T)
    sumX0  +=       0.055 * cos(0.37389864441  + 25665.76130089270 *T)
    sumX0  +=       0.076 * cos(2.93332951447  + 33716.72124838289 *T)
    sumX0  +=       0.061 * cos(2.45900432418  + 7630.64427985291 *T)
    sumX0  +=       0.065 * cos(3.96741643938  + 14267.66258598049 *T)
    sumX0  +=       0.065 * cos(0.82582378579  + 14268.15022094750 *T)
    sumX0  +=       0.060 * cos(4.38893033458  + 1368.41643536150 *T)
    sumX0  +=       0.061 * cos(4.63971906183  + 1118.99960958630 *T)
    sumX0  +=       0.067 * cos(5.86396730577  + 13366.28395293450 *T)

    # MARS X Coord
    sumX1  = 0
    sumX1  +=  1668487.239 * cos(4.16976892466  + 0.24381748350 *T)
    sumX1  +=  551520.815 * cos(5.09364818449  + 6681.46867088311 *T)
    sumX1  +=   77122.404 * cos(5.43227875215  + 10022.08109758290 *T)
    sumX1  +=   33905.789 * cos(5.71116244693  + 3340.36860921629 *T)
    sumX1  +=   13417.687 * cos(2.71465693005  + 3340.85624418330 *T)
    sumX1  +=    9584.524 * cos(5.77057197269  + 13362.69352428270 *T)
    sumX1  +=    2619.567 * cos(3.04292130471  + 3337.33312583431 *T)
    sumX1  +=    2615.345 * cos(6.22267850729  + 3344.37936253230 *T)
    sumX1  +=    1454.057 * cos(6.00537388308  + 6680.98103591609 *T)
    sumX1  +=    1163.163 * cos(6.10878130495  + 16703.30595098250 *T)
    sumX1  +=     900.588 * cos(3.31444263639  + 1059.62574767270 *T)
    sumX1  +=     429.589 * cos(2.75628228081  + 2942.70724077510 *T)
    sumX1  +=     388.556 * cos(5.95081376697  + 5622.08674069391 *T)
    sumX1  +=     380.679 * cos(0.22066211311  + 3739.00524759150 *T)
    sumX1  +=     394.611 * cos(0.19955281144  + 6684.99178923210 *T)
    sumX1  +=     336.947 * cos(5.21225980605  + 3149.40797807170 *T)
    sumX1  +=     333.888 * cos(2.45815226028  + 3185.43584474911 *T)
    sumX1  +=     318.443 * cos(4.05173927787  + 3532.30451029490 *T)
    sumX1  +=     317.603 * cos(0.52499208882  + 3496.27664361750 *T)
    sumX1  +=     312.062 * cos(4.23079808787  + 2544.55823736690 *T)
    sumX1  +=     285.965 * cos(4.60046760581  + 796.54182429990 *T)
    sumX1  +=     259.967 * cos(5.02487716401  + 4137.15425099970 *T)
    sumX1  +=     205.452 * cos(3.20172034630  + 1194.69082770810 *T)
    sumX1  +=     140.526 * cos(0.16384177624  + 20043.91837768230 *T)
    sumX1  +=     159.180 * cos(3.57652346172  + 1589.31671276730 *T)
    sumX1  +=     133.487 * cos(4.63998807852  + 5885.17066406670 *T)
    sumX1  +=     132.246 * cos(3.70031751325  + 4400.23817437250 *T)
    sumX1  +=     126.940 * cos(2.82015450407  + 6677.94555253410 *T)
    sumX1  +=     102.941 * cos(6.05722597437  + 5487.02166065850 *T)
    sumX1  +=     128.977 * cos(1.79935813905  + 1592.83983111630 *T)
    sumX1  +=     101.403 * cos(5.54137030567  + 3339.87592311510 *T)
    sumX1  +=     101.343 * cos(3.72398711211  + 3341.83656525151 *T)
    sumX1  +=      89.600 * cos(5.87965689031  + 3870.54720927791 *T)
    sumX1  +=      84.159 * cos(0.39524478086  + 1990.98883452450 *T)
    sumX1  +=      93.032 * cos(0.00825302856  + 10021.59346261590 *T)
    sumX1  +=      79.379 * cos(3.91303424877  + 426.84200835950 *T)
    sumX1  +=      77.187 * cos(0.64135737142  + 7079.61767429131 *T)
    sumX1  +=      81.451 * cos(5.66551932020  + 5092.39577559930 *T)
    sumX1  +=      91.218 * cos(5.63424494104  + 2146.40923395870 *T)
    sumX1  +=      63.951 * cos(5.27338440626  + 2389.13783793270 *T)
    sumX1  +=      57.311 * cos(0.50275055126  + 10025.60421593190 *T)
    sumX1  +=      56.935 * cos(1.93125711419  + 3098.12764020930 *T)
    sumX1  +=      75.514 * cos(3.61046388189  + 4535.30325440790 *T)
    sumX1  +=      55.567 * cos(1.21839855581  + 9492.39013248830 *T)
    sumX1  +=      72.076 * cos(1.18180609896  + 5088.87265725030 *T)
    sumX1  +=      52.967 * cos(1.04882176228  + 3583.58484815730 *T)
    sumX1  +=      53.267 * cos(2.58802816719  + 4690.72365384210 *T)
    sumX1  +=      70.583 * cos(3.86667384573  + 2787.28684134090 *T)
    sumX1  +=      57.299 * cos(0.33406941177  + 7.35736448430 *T)
    sumX1  +=      54.075 * cos(2.75074198097  + 2957.95971196010 *T)
    sumX1  +=      50.146 * cos(4.61860820816  + 191.20444862810 *T)
    sumX1  +=      46.293 * cos(3.99427986110  + 4292.57465043390 *T)
    sumX1  +=      49.234 * cos(2.50739423398  + 6681.45141723091 *T)
    sumX1  +=      58.634 * cos(5.40086631171  + 3894.42564702570 *T)
    sumX1  +=      48.264 * cos(0.23146545625  + 3723.75277640650 *T)
    sumX1  +=      49.234 * cos(4.53816586556  + 6681.48592453530 *T)
    sumX1  +=      55.021 * cos(1.11115546675  + 155.17658195069 *T)
    sumX1  +=      47.402 * cos(5.64225040832  + 6490.02040477150 *T)
    sumX1  +=      39.275 * cos(1.77443036655  + 1747.77259558350 *T)
    sumX1  +=      37.297 * cos(0.06860934281  + 3554.15533962130 *T)
    sumX1  +=      36.066 * cos(2.88742336978  + 3127.55714874530 *T)
    sumX1  +=      35.298 * cos(2.80382718345  + 6526.04827144891 *T)
    sumX1  +=      33.577 * cos(0.36673291911  + 2145.92159899169 *T)
    sumX1  +=      32.981 * cos(5.07340957623  + 3333.74269718250 *T)
    sumX1  +=      30.372 * cos(0.75673237899  + 1748.26023055050 *T)
    sumX1  +=      39.114 * cos(5.37767054154  + 7477.76667769950 *T)
    sumX1  +=      29.430 * cos(5.34497951848  + 6254.87048000710 *T)
    sumX1  +=      29.559 * cos(4.58393732854  + 951.47458876710 *T)
    sumX1  +=      28.511 * cos(5.98947307886  + 553.32558535889 *T)
    sumX1  +=      27.567 * cos(3.74859884789  + 12303.31159409350 *T)
    sumX1  +=      27.467 * cos(2.19208047399  + 4933.45225781610 *T)
    sumX1  +=      28.130 * cos(4.24458535882  + 3347.96979118411 *T)
    sumX1  +=      34.806 * cos(3.17911079826  + 1349.62359217529 *T)
    sumX1  +=      24.557 * cos(3.16432663502  + 2811.16527908870 *T)
    sumX1  +=      24.445 * cos(6.22731125031  + 1066.73929467350 *T)
    sumX1  +=      29.230 * cos(2.65841667918  + 640.14110379750 *T)
    sumX1  +=      21.272 * cos(3.84506725342  + 4929.92913946711 *T)
    sumX1  +=      24.027 * cos(0.74730655765  + 382.65271473970 *T)
    sumX1  +=      24.805 * cos(3.86501438648  + 2119.00767786191 *T)
    sumX1  +=      19.016 * cos(0.62486368468  + 242.97242145751 *T)
    sumX1  +=      18.321 * cos(0.60915509257  + 9623.93209417470 *T)
    sumX1  +=      17.069 * cos(2.30803476254  + 3320.50092478451 *T)
    sumX1  +=      16.952 * cos(0.67381702286  + 3361.21156358210 *T)
    sumX1  +=      16.989 * cos(0.50197441357  + 23384.53080438210 *T)
    sumX1  +=      16.214 * cos(0.58543208283  + 6681.40136058031 *T)
    sumX1  +=      16.696 * cos(5.51495886637  + 2281.47431399410 *T)
    sumX1  +=      16.004 * cos(0.45066540723  + 1052.51220067191 *T)
    sumX1  +=      17.047 * cos(2.70239305232  + 6674.35512388230 *T)
    sumX1  +=      14.590 * cos(1.92122996858  + 10018.55797923390 *T)
    sumX1  +=      16.713 * cos(0.66892539570  + 7211.15963597770 *T)
    sumX1  +=      17.707 * cos(4.03769647872  + 7740.85060107230 *T)
    sumX1  +=      16.047 * cos(4.24883931225  + 3325.60377299830 *T)
    sumX1  +=      15.961 * cos(5.01645836303  + 3356.10871536831 *T)
    sumX1  +=      16.214 * cos(0.17693569958  + 6681.53598118590 *T)
    sumX1  +=      13.780 * cos(1.54940676591  + 12833.00255918811 *T)
    sumX1  +=      14.630 * cos(0.50614270793  + 2273.87313202630 *T)
    sumX1  +=      12.478 * cos(4.09396331041  + 6682.44899195131 *T)
    sumX1  +=      12.677 * cos(0.77220148195  + 5331.60126122430 *T)
    sumX1  +=      13.489 * cos(0.04091965178  + 2288.10022602790 *T)
    sumX1  +=      15.563 * cos(5.36370101191  + 4562.70481050471 *T)
    sumX1  +=      13.052 * cos(4.49166387640  + 3767.45443505930 *T)
    sumX1  +=      11.294 * cos(2.16352014717  + 1350.11122714230 *T)
    sumX1  +=      11.234 * cos(2.31197110602  + 6438.74006690910 *T)
    sumX1  +=      12.654 * cos(5.55709903692  + 529.44714761109 *T)
    sumX1  +=      12.118 * cos(1.92353601510  + 537.04832957890 *T)
    sumX1  +=      11.323 * cos(2.45817739410  + 522.82123557730 *T)
    sumX1  +=      10.300 * cos(3.95714627958  + 7875.91568110771 *T)
    sumX1  +=      10.510 * cos(3.03392108066  + 5614.97319369310 *T)
    sumX1  +=      12.502 * cos(1.00420796908  + 10420.23010099111 *T)
    sumX1  +=      10.338 * cos(4.87546518042  + 10022.09835123510 *T)
    sumX1  +=       8.834 * cos(1.54079392488  + 3205.79116414790 *T)
    sumX1  +=       9.393 * cos(0.60719486474  + 6836.88907031730 *T)
    sumX1  +=      11.331 * cos(0.31834272198  + 6041.57138456910 *T)
    sumX1  +=       8.381 * cos(1.44095675543  + 3475.92132421871 *T)
    sumX1  +=       9.243 * cos(0.29804306221  + 529.93478257810 *T)
    sumX1  +=      10.338 * cos(2.84469357959  + 10022.06384393071 *T)
    sumX1  +=       7.850 * cos(4.26599008644  + 2700.47132290230 *T)
    sumX1  +=       8.102 * cos(0.82050642338  + 13366.21664263171 *T)
    sumX1  +=       7.470 * cos(3.66222275723  + 397.90518592470 *T)
    sumX1  +=       7.391 * cos(4.07636898152  + 15643.92402079330 *T)
    sumX1  +=       7.798 * cos(4.21954637832  + 6127.89926804070 *T)
    sumX1  +=       7.318 * cos(3.76030403702  + 1751.29571393250 *T)
    sumX1  +=       7.398 * cos(0.83000682555  + 2942.21960580809 *T)
    sumX1  +=       7.000 * cos(0.67061782231  + 3343.89172756530 *T)
    sumX1  +=       6.961 * cos(2.51913263895  + 5629.20028769470 *T)
    sumX1  +=       7.023 * cos(0.28436722670  + 13362.20588931570 *T)
    sumX1  +=       6.810 * cos(3.62419194021  + 4690.23601887509 *T)
    sumX1  +=       6.681 * cos(4.61242157950  + 3178.38960805111 *T)
    sumX1  +=       6.243 * cos(3.10320630386  + 6298.57213865991 *T)
    sumX1  +=       6.294 * cos(4.65274248196  + 3503.32288031550 *T)
    sumX1  +=       6.358 * cos(5.59134148149  + 3316.97780643551 *T)
    sumX1  +=       7.854 * cos(5.63695074863  + 5729.75026463250 *T)
    sumX1  +=       6.174 * cos(2.23158424330  + 5088.38502228330 *T)
    sumX1  +=       6.298 * cos(3.67401727587  + 3364.73468193110 *T)
    sumX1  +=       6.734 * cos(6.02579303320  + 9830.63283147130 *T)
    sumX1  +=       7.590 * cos(4.92016746886  + 8962.69916739370 *T)
    sumX1  +=       6.074 * cos(5.14201333067  + 6151.77770578850 *T)
    sumX1  +=       7.279 * cos(0.63824331472  + 6894.76776632111 *T)
    sumX1  +=       6.015 * cos(1.64381432255  + 853.44019923550 *T)
    sumX1  +=       5.586 * cos(5.72111629827  + 10818.37910439930 *T)
    sumX1  +=       5.539 * cos(0.97704679440  + 1758.40926093329 *T)
    sumX1  +=       5.309 * cos(4.22207781323  + 2648.69864295650 *T)
    sumX1  +=       6.531 * cos(5.82376804487  + 6680.48834981490 *T)
    sumX1  +=       5.498 * cos(1.38930771941  + 6924.19727485710 *T)
    sumX1  +=       5.070 * cos(6.24979955975  + 1596.43025976811 *T)
    sumX1  +=       4.841 * cos(0.45977971146  + 6158.89125278930 *T)
    sumX1  +=       5.011 * cos(3.60259248893  + 433.95555536030 *T)
    sumX1  +=       5.999 * cos(5.01730803001  + 2280.98667902709 *T)
    sumX1  +=       4.334 * cos(0.94400971144  + 12964.54452087450 *T)
    sumX1  +=       5.929 * cos(5.59565118419  + 796.05418933290 *T)
    sumX1  +=       4.781 * cos(1.04203007490  + 6144.66415878770 *T)
    sumX1  +=       4.633 * cos(4.21476805887  + 2914.25805330730 *T)
    sumX1  +=       4.948 * cos(4.14593775556  + 213.05527795450 *T)
    sumX1  +=       3.970 * cos(0.34201182432  + 3312.40705671550 *T)
    sumX1  +=       4.094 * cos(1.37911547191  + 2810.67764412170 *T)
    sumX1  +=       3.913 * cos(2.63739420106  + 3369.30543165110 *T)
    sumX1  +=       3.870 * cos(5.50156644179  + 641.12142486571 *T)
    sumX1  +=       4.135 * cos(5.02138732910  + 4033.01384541010 *T)
    sumX1  +=       5.101 * cos(5.02688997011  + 4292.08701546690 *T)
    sumX1  +=       3.721 * cos(4.59533823527  + 3346.37916849070 *T)
    sumX1  +=       3.721 * cos(4.67002133696  + 3335.33331987590 *T)
    sumX1  +=       4.814 * cos(4.64078611480  + 6688.58221788390 *T)
    sumX1  +=       3.684 * cos(3.92161821873  + 5621.59910572690 *T)
    sumX1  +=       3.561 * cos(2.53539651638  + 8274.06468451590 *T)
    sumX1  +=       4.122 * cos(3.14545069009  + 9866.66069814871 *T)
    sumX1  +=       3.629 * cos(1.25761657663  + 3304.82837750590 *T)
    sumX1  +=       4.333 * cos(1.75353700500  + 6.86972951729 *T)
    sumX1  +=       4.795 * cos(3.57148917867  + 951.96222373410 *T)
    sumX1  +=       3.622 * cos(5.99777439310  + 1582.20316576650 *T)
    sumX1  +=       3.577 * cos(1.72457610482  + 3376.88411086071 *T)
    sumX1  +=       4.037 * cos(5.59805067484  + 8433.00820229911 *T)
    sumX1  +=       3.824 * cos(2.99247114805  + 8969.81271439450 *T)
    sumX1  +=       3.414 * cos(0.51623493472  + 10022.14840788570 *T)
    sumX1  +=       4.384 * cos(4.08397917912  + 11773.62062899890 *T)
    sumX1  +=       4.071 * cos(3.45831553278  + 3863.43366227710 *T)
    sumX1  +=       3.905 * cos(3.45281463337  + 8955.58562039291 *T)
    sumX1  +=       4.291 * cos(5.71320129335  + 2818.27882608950 *T)
    sumX1  +=       3.570 * cos(0.88507363603  + 1744.18216693170 *T)
    sumX1  +=       3.461 * cos(1.44846796079  + 7633.18707713370 *T)
    sumX1  +=       3.246 * cos(0.50620427647  + 149.80701461810 *T)
    sumX1  +=       3.160 * cos(3.18977931848  + 3980.75353049730 *T)
    sumX1  +=       4.113 * cos(5.50708372849  + 213.54291292150 *T)
    sumX1  +=       3.117 * cos(3.81812196204  + 3336.97490882530 *T)
    sumX1  +=       3.117 * cos(5.44723761019  + 3344.73757954131 *T)
    sumX1  +=       3.201 * cos(0.14972615341  + 3893.93801205869 *T)
    sumX1  +=       3.517 * cos(3.81060565872  + 7064.36520310630 *T)
    sumX1  +=       3.078 * cos(2.90200490041  + 7235.03807372550 *T)
    sumX1  +=       3.414 * cos(0.92482016698  + 10022.01378728011 *T)
    sumX1  +=       2.996 * cos(4.05759423372  + 419.72846135871 *T)
    sumX1  +=       3.014 * cos(1.27034835453  + 10551.77206267750 *T)
    sumX1  +=       3.089 * cos(1.78845183673  + 13359.17040593370 *T)
    sumX1  +=       3.393 * cos(1.79839879835  + 3336.84549086730 *T)
    sumX1  +=       2.930 * cos(5.25870617721  + 2487.17222746430 *T)
    sumX1  +=       2.899 * cos(0.10521459121  + 155.66421691770 *T)
    sumX1  +=       2.868 * cos(4.15178337256  + 8270.54156616690 *T)
    sumX1  +=       3.341 * cos(0.09875387834  + 8031.33608054190 *T)
    sumX1  +=       2.771 * cos(5.49404613753  + 3.27930086550 *T)
    sumX1  +=       3.548 * cos(3.33993922239  + 2699.97863680110 *T)
    sumX1  +=       3.139 * cos(4.06806080343  + 5459.62010456170 *T)
    sumX1  +=       2.993 * cos(6.05593455661  + 3.76693583251 *T)
    sumX1  +=       2.558 * cos(5.45850946255  + 7632.69944216670 *T)
    sumX1  +=       2.973 * cos(1.40668693913  + 398.39282089170 *T)
    sumX1  +=       2.555 * cos(4.97943235904  + 553.81322032590 *T)
    sumX1  +=       2.436 * cos(1.11031101787  + 3723.26514143949 *T)
    sumX1  +=       3.398 * cos(3.58689364383  + 9225.78309076650 *T)
    sumX1  +=       2.528 * cos(5.78755929713  + 3333.81000748530 *T)
    sumX1  +=       2.381 * cos(4.54326140537  + 1221.60474883790 *T)
    sumX1  +=       2.367 * cos(5.24552499745  + 8429.48508395010 *T)
    sumX1  +=       2.452 * cos(0.21632651454  + 2766.51144584850 *T)
    sumX1  +=       3.001 * cos(5.91830671191  + 3981.73385156551 *T)
    sumX1  +=       2.325 * cos(0.86902446422  + 9382.18381126891 *T)
    sumX1  +=       3.095 * cos(1.72392494967  + 6283.31966747490 *T)
    sumX1  +=       2.582 * cos(2.98943082072  + 5085.28222859850 *T)
    sumX1  +=       2.407 * cos(0.81452071345  + 9808.78200214491 *T)
    sumX1  +=       2.189 * cos(0.31848733903  + 4407.35172137330 *T)
    sumX1  +=       2.276 * cos(4.18243740519  + 1194.20319274110 *T)
    sumX1  +=       2.512 * cos(1.88451079218  + 16173.61498588790 *T)
    sumX1  +=       2.182 * cos(5.33696085883  + 1751.78334889950 *T)
    sumX1  +=       2.219 * cos(4.06490625389  + 8030.84844557490 *T)
    sumX1  +=       2.433 * cos(4.37190902188  + 11081.46302777211 *T)
    sumX1  +=       2.433 * cos(1.32659636118  + 5828.27228913110 *T)
    sumX1  +=       2.109 * cos(3.63361421871  + 15114.23305569870 *T)
    sumX1  +=       2.719 * cos(3.70485969332  + 6468.16957544510 *T)
    sumX1  +=       2.025 * cos(2.77266249790  + 3915.20104251810 *T)
    sumX1  +=       1.887 * cos(0.17014232268  + 76.50988875911 *T)
    sumX1  +=       1.853 * cos(1.81086841675  + 6151.29007082150 *T)
    sumX1  +=       1.844 * cos(2.78954932950  + 1692.40948698591 *T)
    sumX1  +=       2.127 * cos(3.21512647219  + 3347.90248088131 *T)
    sumX1  +=       1.808 * cos(3.74769873780  + 6040.59106350090 *T)
    sumX1  +=       1.865 * cos(1.35270983937  + 13760.84252769090 *T)
    sumX1  +=       1.998 * cos(6.12517040311  + 5084.79459363150 *T)
    sumX1  +=       1.821 * cos(3.24689091699  + 5099.50932260011 *T)
    sumX1  +=       1.970 * cos(5.21835467754  + 8827.63408735830 *T)
    sumX1  +=       1.948 * cos(5.63032630522  + 5099.02168763310 *T)
    sumX1  +=       1.717 * cos(5.21576113138  + 13362.71077793490 *T)
    sumX1  +=       1.785 * cos(2.44739830297  + 6660.69327539070 *T)
    sumX1  +=       1.744 * cos(1.79434895453  + 6816.53375091851 *T)
    sumX1  +=       1.717 * cos(3.18498950602  + 13362.67627063050 *T)
    sumX1  +=       1.955 * cos(3.63998937844  + 191.69208359510 *T)
    sumX1  +=       1.839 * cos(5.31098518915  + 220.65645992230 *T)
    sumX1  +=       1.739 * cos(3.64901391277  + 3281.48238226970 *T)
    sumX1  +=       2.059 * cos(0.84035144431  + 26725.14323108190 *T)
    sumX1  +=       1.828 * cos(5.65900245619  + 6247.75693300630 *T)
    sumX1  +=       2.136 * cos(1.71869001756  + 6872.91693699471 *T)
    sumX1  +=       1.698 * cos(5.61638598924  + 3400.23010609691 *T)
    sumX1  +=       1.534 * cos(3.28685934658  + 574.10098085129 *T)
    sumX1  +=       1.533 * cos(3.03625020180  + 3877.66075627870 *T)
    sumX1  +=       1.541 * cos(1.53159919865  + 3495.78900865049 *T)
    sumX1  +=       1.628 * cos(4.45153470681  + 10023.06141865111 *T)
    sumX1  +=       1.807 * cos(5.37059175260  + 6696.72114206811 *T)
    sumX1  +=       1.767 * cos(2.64991180130  + 6661.11335148431 *T)
    sumX1  +=       1.525 * cos(5.97153443990  + 206.42936592071 *T)
    sumX1  +=       1.957 * cos(0.57667231930  + 7234.55043875850 *T)
    sumX1  +=       1.381 * cos(4.29497602044  + 11216.52810780750 *T)
    sumX1  +=       1.688 * cos(4.89795336991  + 7108.06686175911 *T)
    sumX1  +=       1.517 * cos(5.15570920014  + 3340.38586286849 *T)
    sumX1  +=       1.506 * cos(3.91589541430  + 8440.12174929990 *T)
    sumX1  +=       1.379 * cos(1.26307005719  + 3127.06951377830 *T)
    sumX1  +=       1.360 * cos(2.27637695603  + 3074.24920246151 *T)
    sumX1  +=       1.516 * cos(3.12494539721  + 3340.35135556410 *T)
    sumX1  +=       1.302 * cos(4.55773751868  + 6666.21619969810 *T)
    sumX1  +=       1.588 * cos(1.11386226420  + 8672.21368792411 *T)
    sumX1  +=       1.305 * cos(0.97554659415  + 6701.82399028190 *T)
    sumX1  +=       1.529 * cos(0.56980933613  + 3547.04179262050 *T)
    sumX1  +=       1.317 * cos(1.45792516402  + 7063.87756813929 *T)
    sumX1  +=       1.316 * cos(6.26946114201  + 8432.52056733209 *T)
    sumX1  +=       1.510 * cos(2.38931811093  + 3134.67069574611 *T)
    sumX1  +=       1.232 * cos(1.35650102913  + 1228.71829583870 *T)
    sumX1  +=       1.498 * cos(3.04227517263  + 10014.96755058210 *T)
    sumX1  +=       1.240 * cos(1.97737579625  + 6836.40143535029 *T)
    sumX1  +=       1.407 * cos(4.40841782051  + 18984.53644749310 *T)
    sumX1  +=       1.219 * cos(0.70593882337  + 3607.46328590511 *T)
    sumX1  +=       1.443 * cos(4.48857942604  + 1059.13811270570 *T)
    sumX1  +=       1.157 * cos(1.26136880250  + 110.45013870291 *T)
    sumX1  +=       1.195 * cos(0.29423080541  + 11371.94850724170 *T)
    sumX1  +=       1.173 * cos(1.00908500027  + 10575.16286545830 *T)
    sumX1  +=       1.175 * cos(1.90125230900  + 3320.08084869090 *T)
    sumX1  +=       1.171 * cos(1.07324612452  + 3361.63163967570 *T)
    sumX1  +=       1.509 * cos(4.60862746462  + 6702.24406637550 *T)
    sumX1  +=       1.157 * cos(1.07614409578  + 10235.38019302090 *T)
    sumX1  +=       1.097 * cos(3.27118145757  + 12936.09533340670 *T)
    sumX1  +=       1.517 * cos(2.92274898055  + 3097.64000524229 *T)
    sumX1  +=       1.089 * cos(2.28516107614  + 3170.18337356410 *T)
    sumX1  +=       1.068 * cos(3.73801386165  + 6685.05909953490 *T)
    sumX1  +=       1.082 * cos(2.47496216267  + 6531.90547374850 *T)
    sumX1  +=       1.365 * cos(4.09642340658  + 1039.27042827391 *T)
    sumX1  +=       1.415 * cos(2.62020577266  + 647.25465079831 *T)
    sumX1  +=       1.028 * cos(5.31972930210  + 6261.98402700790 *T)
    sumX1  +=       1.020 * cos(0.69687736418  + 3511.52911480250 *T)
    sumX1  +=       1.072 * cos(3.31077938125  + 6677.87824223130 *T)
    sumX1  +=       0.994 * cos(0.71631497535  + 2921.37160030811 *T)
    sumX1  +=       0.960 * cos(6.15441067021  + 383.14034970670 *T)
    sumX1  +=       1.295 * cos(1.19205099554  + 3738.51761262449 *T)
    sumX1  +=       1.081 * cos(3.12534442949  + 2906.65687133950 *T)
    sumX1  +=       0.958 * cos(0.18373503494  + 4989.30300138070 *T)
    sumX1  +=       0.985 * cos(3.76783660715  + 8425.89465529830 *T)
    sumX1  +=       0.957 * cos(3.10714680748  + 10575.65050042530 *T)
    sumX1  +=       0.922 * cos(4.56395067955  + 9468.51169474051 *T)
    sumX1  +=       0.997 * cos(2.53736303913  + 4562.21717553770 *T)
    sumX1  +=       0.918 * cos(0.10711401132  + 13171.24525817110 *T)
    sumX1  +=       0.929 * cos(5.90027659664  + 10973.31186886650 *T)
    sumX1  +=       0.902 * cos(0.05631596164  + 3561.26888662211 *T)
    sumX1  +=       1.212 * cos(5.52073488118  + 2804.05173208791 *T)
    sumX1  +=       0.887 * cos(2.35910260079  + 3760.34088805850 *T)
    sumX1  +=       0.876 * cos(1.69293209630  + 956.53297345411 *T)
    sumX1  +=       1.037 * cos(2.59899224317  + 9779.35249360890 *T)
    sumX1  +=       0.929 * cos(5.98408419364  + 9070.36269133230 *T)
    sumX1  +=       0.857 * cos(2.76482019861  + 1592.35219614930 *T)
    sumX1  +=       0.840 * cos(0.12599209397  + 2126.12122486270 *T)
    sumX1  +=       0.845 * cos(2.81776300450  + 3120.44360174450 *T)
    sumX1  +=       0.954 * cos(2.62936052400  + 2920.88396534110 *T)
    sumX1  +=       0.880 * cos(2.47516322984  + 3313.45468808651 *T)
    sumX1  +=       1.010 * cos(1.04761636483  + 3253.54803964350 *T)
    sumX1  +=       1.124 * cos(1.69207540981  + 10973.79950383350 *T)
    sumX1  +=       0.870 * cos(0.50701374292  + 3368.25780028011 *T)
    sumX1  +=       0.983 * cos(4.50403465372  + 9595.48290670690 *T)
    sumX1  +=       0.852 * cos(5.23864143891  + 11770.09751064990 *T)
    sumX1  +=       0.833 * cos(4.23151062947  + 8962.21153242670 *T)
    sumX1  +=       0.824 * cos(2.58971694742  + 1861.50203515190 *T)
    sumX1  +=       0.974 * cos(1.93463517204  + 3428.16444872311 *T)
    sumX1  +=       1.121 * cos(1.14643572651  + 16706.82906933150 *T)
    sumX1  +=       0.794 * cos(1.89288433548  + 9168.88471583091 *T)
    sumX1  +=       1.033 * cos(4.24581346686  + 5091.90814063229 *T)
    sumX1  +=       0.766 * cos(2.74496789725  + 3443.94901840191 *T)
    sumX1  +=       0.797 * cos(1.27369538416  + 6282.83203250789 *T)
    sumX1  +=       0.758 * cos(6.06572380841  + 14158.99153109910 *T)
    sumX1  +=       0.758 * cos(4.05984927529  + 10404.97762980610 *T)
    sumX1  +=       0.769 * cos(3.45170944257  + 9639.18456535971 *T)
    sumX1  +=       0.762 * cos(4.34975730450  + 2699.49100183409 *T)
    sumX1  +=       0.740 * cos(3.37186639218  + 6438.25243194210 *T)
    sumX1  +=       0.761 * cos(3.02473744131  + 633.02755679670 *T)
    sumX1  +=       0.800 * cos(4.79153045650  + 6519.00203475090 *T)
    sumX1  +=       0.725 * cos(3.17644612903  + 6475.28312244591 *T)
    sumX1  +=       0.730 * cos(3.02102816704  + 6709.91785835091 *T)
    sumX1  +=       0.766 * cos(6.23897845769  + 3148.92034310470 *T)
    sumX1  +=       0.795 * cos(1.28741844431  + 16305.15694757430 *T)
    sumX1  +=       0.701 * cos(0.13673007891  + 4569.33072253850 *T)
    sumX1  +=       0.750 * cos(6.08088023817  + 3297.13733187830 *T)
    sumX1  +=       0.762 * cos(3.46520920498  + 2544.07060239989 *T)
    sumX1  +=       0.695 * cos(5.64254870753  + 2700.95895786930 *T)
    sumX1  +=       0.688 * cos(5.00398855432  + 10029.19464458370 *T)
    sumX1  +=       0.923 * cos(2.40915411891  + 10177.01386205009 *T)
    sumX1  +=       0.737 * cos(3.18449445655  + 3384.57515648830 *T)
    sumX1  +=       0.674 * cos(2.97028549448  + 2807.64216073970 *T)
    sumX1  +=       0.845 * cos(0.83945644325  + 9499.50367948910 *T)
    sumX1  +=       0.645 * cos(1.31214998001  + 1214.49120183710 *T)
    sumX1  +=       0.629 * cos(0.28992570288  + 3237.76346996470 *T)
    sumX1  +=       0.627 * cos(1.11050059243  + 13149.39442884470 *T)
    sumX1  +=       0.632 * cos(2.81036789867  + 6605.20259960750 *T)
    sumX1  +=       0.710 * cos(3.89496660928  + 6461.05602844430 *T)
    sumX1  +=       0.701 * cos(2.59503044929  + 3370.28575271930 *T)
    sumX1  +=       0.701 * cos(0.38714381576  + 3311.42673564730 *T)
    sumX1  +=       0.607 * cos(6.17244952085  + 949.41942645330 *T)
    sumX1  +=       0.603 * cos(2.72434120082  + 12566.39551746630 *T)
    sumX1  +=       0.605 * cos(4.55268583903  + 11243.92966390430 *T)
    sumX1  +=       0.719 * cos(3.39790369460  + 12310.42514109431 *T)
    sumX1  +=       0.585 * cos(3.28767074464  + 5827.78465416409 *T)
    sumX1  +=       0.622 * cos(5.63029890672  + 7903.31723720450 *T)
    sumX1  +=       0.633 * cos(5.14454720349  + 691.91378374330 *T)
    sumX1  +=       0.601 * cos(3.81540155391  + 9778.86485864190 *T)
    sumX1  +=       0.580 * cos(0.54616275524  + 16702.81831601549 *T)
    sumX1  +=       0.598 * cos(2.15895090295  + 3340.87349783550 *T)
    sumX1  +=       0.587 * cos(2.84018169062  + 13517.62628874989 *T)
    sumX1  +=       0.646 * cos(5.77534161373  + 2111.89413086110 *T)
    sumX1  +=       0.592 * cos(4.73533629827  + 2301.82963339290 *T)
    sumX1  +=       0.598 * cos(0.12817703817  + 3340.83899053111 *T)
    sumX1  +=       0.551 * cos(4.21218171420  + 2693.35777590150 *T)
    sumX1  +=       0.544 * cos(1.97116952941  + 16699.78283263350 *T)
    sumX1  +=       0.623 * cos(1.72895571381  + 10264.80970155691 *T)
    sumX1  +=       0.539 * cos(0.82048737136  + 3490.41944131790 *T)
    sumX1  +=       0.568 * cos(2.68750434726  + 1437.41943168211 *T)
    sumX1  +=       0.557 * cos(0.85462599459  + 13362.76083458550 *T)
    sumX1  +=       0.530 * cos(2.83578333137  + 4555.59126350391 *T)
    sumX1  +=       0.647 * cos(0.93860283834  + 4393.12462737170 *T)
    sumX1  +=       0.575 * cos(5.57774190632  + 5884.68302909969 *T)
    sumX1  +=       0.529 * cos(6.20686668225  + 1376.01761732930 *T)
    sumX1  +=       0.537 * cos(0.00980371467  + 3874.07032762691 *T)
    sumX1  +=       0.584 * cos(2.12240201386  + 4194.05262593530 *T)
    sumX1  +=       0.507 * cos(2.19313676678  + 2177.89390480850 *T)
    sumX1  +=       0.505 * cos(1.37897438890  + 9485.27658548750 *T)
    sumX1  +=       0.510 * cos(5.74338619548  + 11243.44202893729 *T)
    sumX1  +=       0.528 * cos(0.09425635417  + 4555.10362853689 *T)
    sumX1  +=       0.499 * cos(0.79138578393  + 3340.43591951909 *T)
    sumX1  +=       0.501 * cos(2.93965768481  + 224.58861318540 *T)
    sumX1  +=       0.525 * cos(0.94887121150  + 6684.50415426510 *T)
    sumX1  +=       0.670 * cos(5.94569787504  + 6657.59023313531 *T)
    sumX1  +=       0.634 * cos(1.62056513361  + 242.48478649050 *T)
    sumX1  +=       0.513 * cos(4.94013592097  + 3233.19272024471 *T)
    sumX1  +=       0.490 * cos(4.17557345892  + 3774.56798206011 *T)
    sumX1  +=       0.634 * cos(3.69885606720  + 2707.58486990309 *T)
    sumX1  +=       0.613 * cos(4.40032363311  + 5989.31106965630 *T)
    sumX1  +=       0.492 * cos(4.32515466341  + 3448.51976812190 *T)
    sumX1  +=       0.616 * cos(5.16594414687  + 5642.44206009270 *T)
    sumX1  +=       0.555 * cos(3.80576965537  + 18454.84548239850 *T)
    sumX1  +=       0.463 * cos(1.38847352998  + 3229.66960189570 *T)
    sumX1  +=       0.507 * cos(5.87248459784  + 4136.66661603269 *T)
    sumX1  +=       0.477 * cos(3.66215350045  + 12296.19804709270 *T)
    sumX1  +=       0.462 * cos(1.34104474279  + 8827.14645239129 *T)
    sumX1  +=       0.444 * cos(3.95877289512  + 6705.34710863090 *T)
    sumX1  +=       0.549 * cos(3.18688547095  + 3333.25506221550 *T)
    sumX1  +=       0.443 * cos(1.59367641062  + 3452.04288647091 *T)
    sumX1  +=       0.557 * cos(1.26314936469  + 13362.62621397991 *T)
    sumX1  +=       0.437 * cos(1.45419351715  + 13915.77529215809 *T)
    sumX1  +=       0.498 * cos(1.19917091183  + 3340.30129891350 *T)
    sumX1  +=       0.424 * cos(5.72891764326  + 2301.34199842589 *T)
    sumX1  +=       0.433 * cos(4.48268877222  + 6843.93530701530 *T)
    sumX1  +=       0.494 * cos(3.48540091775  + 13207.27312484851 *T)
    sumX1  +=       0.402 * cos(1.71994710298  + 6571.26234966370 *T)
    sumX1  +=       0.453 * cos(3.24616258868  + 3184.94820978210 *T)
    sumX1  +=       0.413 * cos(4.34476522794  + 8800.23253126150 *T)
    sumX1  +=       0.407 * cos(6.08186909440  + 6048.68493156990 *T)
    sumX1  +=       0.488 * cos(2.98211910254  + 2222.10045208050 *T)
    sumX1  +=       0.556 * cos(0.32985654758  + 6034.45783756830 *T)
    sumX1  +=       0.421 * cos(4.49869717068  + 4379.88285497370 *T)
    sumX1  +=       0.472 * cos(0.33931417528  + 4937.04268646790 *T)
    sumX1  +=       0.392 * cos(5.13497859630  + 134.82126255190 *T)
    sumX1  +=       0.515 * cos(5.18562869638  + 6675.94574657570 *T)
    sumX1  +=       0.407 * cos(1.72332746589  + 3392.13658204571 *T)
    sumX1  +=       0.407 * cos(1.25884679916  + 3289.57590632090 *T)
    sumX1  +=       0.474 * cos(1.74795831788  + 13892.38448937730 *T)
    sumX1  +=       0.426 * cos(2.86953666046  + 11614.67711121570 *T)
    sumX1  +=       0.384 * cos(5.80577939753  + 6685.35000624110 *T)
    sumX1  +=       0.407 * cos(3.46091021814  + 4569.81835750550 *T)
    sumX1  +=       0.495 * cos(4.23130410948  + 13119.47728534169 *T)
    sumX1  +=       0.366 * cos(2.89115269594  + 263.32774085630 *T)
    sumX1  +=       0.370 * cos(1.36093537897  + 3274.36883526891 *T)
    sumX1  +=       0.359 * cos(4.24198911604  + 5614.48555872609 *T)
    sumX1  +=       0.358 * cos(4.30213026610  + 7910.43078420530 *T)
    sumX1  +=       0.360 * cos(1.62131716357  + 3407.34365309770 *T)
    sumX1  +=       0.402 * cos(5.43912274152  + 7895.71605523670 *T)
    sumX1  +=       0.453 * cos(2.09910428652  + 3113.38011139430 *T)
    sumX1  +=       0.394 * cos(6.09769421438  + 9588.36935970610 *T)
    sumX1  +=       0.341 * cos(4.58419197314  + 15.49628866851 *T)
    sumX1  +=       0.341 * cos(4.67783034753  + 2384.07945324570 *T)
    sumX1  +=       0.358 * cos(1.27303671253  + 20.11150191529 *T)
    sumX1  +=       0.352 * cos(5.65764885671  + 12721.32828193349 *T)
    sumX1  +=       0.422 * cos(2.19333082001  + 19514.22741258770 *T)
    sumX1  +=       0.463 * cos(2.10905707905  + 3870.05957431090 *T)
    sumX1  +=       0.334 * cos(4.35518721630  + 10595.93826095070 *T)
    sumX1  +=       0.333 * cos(1.29837025236  + 5725.17951491250 *T)
    sumX1  +=       0.342 * cos(5.18889491099  + 9380.71585523369 *T)
    sumX1  +=       0.373 * cos(6.01149621113  + 10021.10077651470 *T)
    sumX1  +=       0.380 * cos(5.03329108256  + 7909.94314923829 *T)
    sumX1  +=       0.375 * cos(4.47589680364  + 11611.15399286670 *T)
    sumX1  +=       0.335 * cos(4.55687373480  + 2939.18412242610 *T)
    sumX1  +=       0.340 * cos(3.64506451118  + 7321.36595719711 *T)
    sumX1  +=       0.374 * cos(4.03375047167  + 3415.63784275061 *T)
    sumX1  +=       0.330 * cos(1.88368883608  + 6677.45791756710 *T)
    sumX1  +=       0.412 * cos(0.88251814469  + 3568.33237697230 *T)
    sumX1  +=       0.310 * cos(4.70847967924  + 14422.07545447190 *T)
    sumX1  +=       0.318 * cos(4.69941081301  + 16460.08971204149 *T)
    sumX1  +=       0.338 * cos(5.91131346510  + 10177.50149701710 *T)
    sumX1  +=       0.330 * cos(0.91597037671  + 3362.70707350971 *T)
    sumX1  +=       0.330 * cos(2.06620388834  + 3319.00541485690 *T)
    sumX1  +=       0.348 * cos(5.61730962300  + 9602.59645370770 *T)
    sumX1  +=       0.329 * cos(4.94303630669  + 10042.85649307530 *T)
    sumX1  +=       0.312 * cos(5.55211289042  + 323.74923414091 *T)
    sumX1  +=       0.340 * cos(4.78508277647  + 22325.14887419290 *T)
    sumX1  +=       0.403 * cos(4.77937379248  + 6040.10342853389 *T)
    sumX1  +=       0.390 * cos(2.10755320496  + 6717.49653756051 *T)
    sumX1  +=       0.324 * cos(3.81410761918  + 14584.54209060410 *T)
    sumX1  +=       0.298 * cos(0.56258586387  + 3417.12231545890 *T)
    sumX1  +=       0.291 * cos(1.54713222563  + 6645.44080420570 *T)
    sumX1  +=       0.293 * cos(3.68939166735  + 2676.10019905331 *T)
    sumX1  +=       0.332 * cos(6.23779631324  + 4459.61203628611 *T)
    sumX1  +=       0.279 * cos(3.08511995000  + 5732.29306191330 *T)
    sumX1  +=       0.304 * cos(3.91894030488  + 7204.04608897691 *T)
    sumX1  +=       0.280 * cos(1.35602318924  + 3283.95786924770 *T)
    sumX1  +=       0.280 * cos(1.62615107581  + 3397.75461911890 *T)
    sumX1  +=       0.283 * cos(4.01285717255  + 262.84010588929 *T)
    sumX1  +=       0.325 * cos(5.23850564888  + 3266.07464561600 *T)
    sumX1  +=       0.280 * cos(3.40896567674  + 6688.51490758111 *T)
    sumX1  +=       0.284 * cos(0.30082021041  + 5244.29305668450 *T)
    sumX1  +=       0.363 * cos(2.79281358783  + 10001.30570209050 *T)
    sumX1  +=       0.267 * cos(0.02468832444  + 7255.81346921790 *T)
    sumX1  +=       0.348 * cos(1.35684937218  + 1990.50119955750 *T)
    sumX1  +=       0.329 * cos(5.60955627373  + 162.22281864870 *T)
    sumX1  +=       0.356 * cos(3.23261573293  + 1162.71852189130 *T)
    sumX1  +=       0.345 * cos(3.39135824465  + 426.35437339250 *T)
    sumX1  +=       0.312 * cos(4.00206322359  + 7255.32583425090 *T)
    sumX1  +=       0.260 * cos(3.24406180206  + 11766.50708199810 *T)
    sumX1  +=       0.312 * cos(3.70486580950  + 3914.71340755109 *T)
    sumX1  +=       0.261 * cos(4.36606169264  + 1588.82907780029 *T)
    sumX1  +=       0.260 * cos(3.19061283477  + 11769.60987568290 *T)
    sumX1  +=       0.259 * cos(3.13096549005  + 5033.02191368570 *T)
    sumX1  +=       0.257 * cos(4.58563664538  + 205.94173095370 *T)
    sumX1  +=       0.253 * cos(1.68968336145  + 17101.45495439070 *T)
    sumX1  +=       0.253 * cos(3.52319644222  + 16703.28869733031 *T)
    sumX1  +=       0.253 * cos(1.08133024526  + 6948.07571260490 *T)
    sumX1  +=       0.278 * cos(6.24540446055  + 7322.34627826531 *T)
    sumX1  +=       0.255 * cos(3.63050380615  + 11780.73417599970 *T)
    sumX1  +=       0.282 * cos(1.66208146566  + 860.55374623631 *T)
    sumX1  +=       0.253 * cos(5.55396848472  + 16703.32320463470 *T)
    sumX1  +=       0.256 * cos(4.75382802148  + 13936.55068765050 *T)
    sumX1  +=       0.240 * cos(2.67113783155  + 2274.36076699330 *T)
    sumX1  +=       0.246 * cos(0.25187790496  + 3377.46160948750 *T)
    sumX1  +=       0.246 * cos(2.73029636009  + 3304.25087887911 *T)
    sumX1  +=       0.270 * cos(5.07892919701  + 1222.09238380491 *T)
    sumX1  +=       0.243 * cos(3.58179875490  + 16276.70776010650 *T)
    sumX1  +=       0.253 * cos(2.03967667378  + 14314.41193053330 *T)
    sumX1  +=       0.225 * cos(3.64551843619  + 10018.49066893110 *T)
    sumX1  +=       0.227 * cos(3.05024629694  + 5305.69487103730 *T)
    sumX1  +=       0.284 * cos(2.13762517288  + 10157.14617761831 *T)
    sumX1  +=       0.244 * cos(1.34723380846  + 2074.80113058990 *T)
    sumX1  +=       0.243 * cos(1.17853314043  + 30065.75565778170 *T)
    sumX1  +=       0.218 * cos(3.34943504189  + 3030.01544531491 *T)
    sumX1  +=       0.220 * cos(3.22940204842  + 16858.23871544969 *T)
    sumX1  +=       0.221 * cos(5.57676911186  + 4005.61228931331 *T)
    sumX1  +=       0.213 * cos(6.12180241864  + 533.45790092711 *T)
    sumX1  +=       0.224 * cos(4.07746567192  + 10025.67152623470 *T)
    sumX1  +=       0.216 * cos(5.71759573965  + 10037.33356876791 *T)
    sumX1  +=       0.213 * cos(5.27835525808  + 10448.67928845890 *T)
    sumX1  +=       0.219 * cos(4.37173752188  + 23.63462026429 *T)
    sumX1  +=       0.218 * cos(6.18590261959  + 5641.95442512570 *T)
    sumX1  +=       0.211 * cos(3.90230684295  + 6944.55259425590 *T)
    sumX1  +=       0.215 * cos(4.18957088171  + 3212.83740084591 *T)
    sumX1  +=       0.234 * cos(0.76565690804  + 3185.50315505190 *T)
    sumX1  +=       0.212 * cos(1.32558805223  + 3553.66770465429 *T)
    sumX1  +=       0.270 * cos(6.14786843780  + 16061.94070863329 *T)
    sumX1  +=       0.214 * cos(6.25312379495  + 1545.11016549530 *T)
    sumX1  +=       0.213 * cos(2.95953405061  + 17925.15451730390 *T)
    sumX1  +=       0.206 * cos(1.17842868532  + 15663.79170522509 *T)
    sumX1  +=       0.201 * cos(1.28322953170  + 1835.56843509031 *T)
    sumX1  +=       0.211 * cos(5.31180316254  + 3230.64992296391 *T)
    sumX1  +=       0.198 * cos(0.22514658211  + 6578.37589666450 *T)
    sumX1  +=       0.204 * cos(0.75052474665  + 12323.17927852529 *T)
    sumX1  +=       0.204 * cos(5.07567496442  + 3468.87508752070 *T)
    sumX1  +=       0.213 * cos(0.67647172011  + 7747.96414807311 *T)
    sumX1  +=       0.197 * cos(4.08128556688  + 3340.92355448610 *T)
    sumX1  +=       0.221 * cos(2.21670811630  + 3496.20933331471 *T)
    sumX1  +=       0.193 * cos(5.91467798973  + 3651.69704305170 *T)
    sumX1  +=       0.191 * cos(0.58881365615  + 14712.56093394150 *T)
    sumX1  +=       0.242 * cos(3.82869102683  + 6622.09480896950 *T)
    sumX1  +=       0.191 * cos(5.23241625121  + 2480.05868046350 *T)
    sumX1  +=       0.188 * cos(2.57440320373  + 103.33659170210 *T)
    sumX1  +=       0.190 * cos(1.14046863706  + 15508.85894075790 *T)
    sumX1  +=       0.182 * cos(5.56209368259  + 8543.21452351850 *T)
    sumX1  +=       0.233 * cos(3.46250833088  + 13916.26292712510 *T)
    sumX1  +=       0.200 * cos(4.07950076032  + 6677.58733552510 *T)
    sumX1  +=       0.249 * cos(1.64032039460  + 699.02733074409 *T)
    sumX1  +=       0.237 * cos(1.54894672444  + 7373.13863714289 *T)
    sumX1  +=       0.202 * cos(1.48932512281  + 12012.82611462390 *T)
    sumX1  +=       0.205 * cos(1.56629832109  + 2946.23035912410 *T)
    sumX1  +=       0.230 * cos(0.87749290251  + 1903.19299501769 *T)
    sumX1  +=       0.239 * cos(0.56550721465  + 6107.12387254830 *T)
    sumX1  +=       0.200 * cos(2.99049384035  + 10001.72577818411 *T)
    sumX1  +=       0.174 * cos(0.03064105150  + 6457.12387518120 *T)
    sumX1  +=       0.188 * cos(6.17205698509  + 2388.65020296570 *T)
    sumX1  +=       0.171 * cos(4.43743888209  + 6756.25026945040 *T)
    sumX1  +=       0.174 * cos(3.24256094891  + 6418.38474751031 *T)
    sumX1  +=       0.182 * cos(1.09751490284  + 6887.65421932030 *T)
    sumX1  +=       0.172 * cos(0.31527807188  + 19402.55313533309 *T)
    sumX1  +=       0.169 * cos(1.34967221969  + 522.33360061030 *T)
    sumX1  +=       0.176 * cos(4.11592405892  + 7896.20369020371 *T)
    sumX1  +=       0.169 * cos(1.19861060296  + 5618.56362234490 *T)
    sumX1  +=       0.168 * cos(1.47006471713  + 13575.99261972070 *T)
    sumX1  +=       0.184 * cos(3.93805925267  + 7218.27318297850 *T)
    sumX1  +=       0.166 * cos(3.82808880048  + 15651.03756779411 *T)
    sumX1  +=       0.212 * cos(4.80268839794  + 13363.67384535091 *T)
    sumX1  +=       0.169 * cos(0.37705302403  + 2655.81218995730 *T)
    sumX1  +=       0.194 * cos(2.09860374727  + 9491.90249752129 *T)
    sumX1  +=       0.168 * cos(1.37306425270  + 7314.25241019630 *T)
    sumX1  +=       0.177 * cos(4.60890307799  + 3341.34893028450 *T)
    sumX1  +=       0.160 * cos(6.16917404327  + 1903.68062998470 *T)
    sumX1  +=       0.159 * cos(2.63543785839  + 2008.80135664250 *T)
    sumX1  +=       0.177 * cos(5.21096301825  + 6686.99159519050 *T)
    sumX1  +=       0.210 * cos(3.08946859688  + 4032.52621044309 *T)
    sumX1  +=       0.192 * cos(3.27966303360  + 2751.79141717511 *T)
    sumX1  +=       0.153 * cos(4.89714620461  + 2803.56409712089 *T)
    sumX1  +=       0.153 * cos(4.40731948478  + 3334.17258030920 *T)
    sumX1  +=       0.153 * cos(4.85804008745  + 3347.53990805741 *T)
    sumX1  +=       0.152 * cos(3.30181617069  + 5486.53402569149 *T)
    sumX1  +=       0.154 * cos(1.98358917407  + 3499.79976196650 *T)
    sumX1  +=       0.151 * cos(5.12277493842  + 14054.36349054249 *T)
    sumX1  +=       0.169 * cos(1.72774686313  + 19645.76937427410 *T)
    sumX1  +=       0.197 * cos(4.49011102737  + 3340.78893388051 *T)
    sumX1  +=       0.195 * cos(0.42823378367  + 6546.40359084770 *T)
    sumX1  +=       0.185 * cos(2.81424337458  + 1331.81107005730 *T)
    sumX1  +=       0.148 * cos(5.02521612753  + 14712.07329897449 *T)
    sumX1  +=       0.196 * cos(0.33560277948  + 8982.56685182550 *T)
    sumX1  +=       0.158 * cos(4.60992813529  + 14557.14053450730 *T)
    sumX1  +=       0.171 * cos(4.49539596939  + 1063.14886602171 *T)
    sumX1  +=       0.140 * cos(4.74314220170  + 2786.79920637390 *T)
    sumX1  +=       0.149 * cos(5.05608716763  + 13518.11392371690 *T)
    sumX1  +=       0.150 * cos(1.03888799126  + 736.12033101530 *T)
    sumX1  +=       0.140 * cos(0.63946494957  + 4039.63975744390 *T)
    sumX1  +=       0.148 * cos(3.00259617784  + 3053.89388306271 *T)
    sumX1  +=       0.141 * cos(5.70073619213  + 7380.25218414370 *T)
    sumX1  +=       0.187 * cos(4.37697824709  + 2409.49315733150 *T)
    sumX1  +=       0.140 * cos(0.90450305556  + 6692.10533623291 *T)
    sumX1  +=       0.171 * cos(3.32683055485  + 14.47091148511 *T)
    sumX1  +=       0.135 * cos(5.16323033758  + 4076.73275771510 *T)
    sumX1  +=       0.135 * cos(4.10212923465  + 2604.97973065150 *T)
    sumX1  +=       0.134 * cos(5.28028126267  + 13554.14179039430 *T)
    sumX1  +=       0.133 * cos(2.22525164413  + 3191.29304704871 *T)
    sumX1  +=       0.177 * cos(4.97584254328  + 2277.46356067810 *T)
    sumX1  +=       0.177 * cos(1.67473064239  + 8959.17604904470 *T)
    sumX1  +=       0.135 * cos(5.19738274919  + 3347.48215621710 *T)
    sumX1  +=       0.178 * cos(1.87756312184  + 2284.50979737610 *T)
    sumX1  +=       0.175 * cos(4.86177792832  + 8966.22228574270 *T)
    sumX1  +=       0.131 * cos(0.83104395212  + 9375.07026426810 *T)
    sumX1  +=       0.160 * cos(4.45288273211  + 2494.28577446509 *T)
    sumX1  +=       0.129 * cos(5.46603984965  + 6680.99828956829 *T)
    sumX1  +=       0.152 * cos(1.54903747444  + 12722.79623796870 *T)
    sumX1  +=       0.138 * cos(1.47418494330  + 20047.44149603130 *T)
    sumX1  +=       0.171 * cos(0.83416282874  + 75.02541605080 *T)
    sumX1  +=       0.129 * cos(1.71104572673  + 4846.14405327630 *T)
    sumX1  +=       0.134 * cos(2.62552732791  + 4025.90029840930 *T)
    sumX1  +=       0.129 * cos(4.40753732369  + 11670.52785478030 *T)
    sumX1  +=       0.129 * cos(3.43526215532  + 6680.96378226390 *T)
    sumX1  +=       0.157 * cos(0.45500626871  + 9389.29735826970 *T)
    sumX1  +=       0.126 * cos(4.78407774469  + 6148.25458743950 *T)
    sumX1  +=       0.176 * cos(0.06847011749  + 21.01921297590 *T)
    sumX1  +=       0.169 * cos(1.68451776572  + 19004.40413192490 *T)
    sumX1  +=       0.155 * cos(5.99614758178  + 3929.92107119150 *T)
    sumX1  +=       0.150 * cos(4.51700078093  + 3077.77232081051 *T)
    sumX1  +=       0.133 * cos(1.12151471008  + 12840.11610618890 *T)
    sumX1  +=       0.137 * cos(5.81509659853  + 6674.42243418510 *T)
    sumX1  +=       0.129 * cos(3.00992741103  + 18606.25512851669 *T)
    sumX1  +=       0.146 * cos(0.88857962035  + 3284.05298201350 *T)
    sumX1  +=       0.146 * cos(2.09359464470  + 3397.65950635310 *T)
    sumX1  +=       0.130 * cos(2.27123580336  + 1505.04399160949 *T)
    sumX1  +=       0.133 * cos(6.26353175980  + 3627.81860530390 *T)
    sumX1  +=       0.125 * cos(2.00888455156  + 4672.42349675709 *T)
    sumX1  +=       0.124 * cos(1.40431448992  + 6594.16046634330 *T)
    sumX1  +=       0.135 * cos(0.17263957625  + 7902.82960223750 *T)
    sumX1  +=       0.147 * cos(2.02223322465  + 536.56069461190 *T)
    sumX1  +=       0.123 * cos(1.97449162762  + 1486.22393854870 *T)
    sumX1  +=       0.125 * cos(2.58633115739  + 15265.64270181689 *T)
    sumX1  +=       0.121 * cos(4.43373574929  + 4534.81561944089 *T)
    sumX1  +=       0.134 * cos(0.01049298570  + 2288.58786099491 *T)
    sumX1  +=       0.119 * cos(1.42038025370  + 74.53778108379 *T)
    sumX1  +=       0.146 * cos(4.33987734620  + 1479.11039154791 *T)
    sumX1  +=       0.117 * cos(3.63700868579  + 3344.36210888011 *T)
    sumX1  +=       0.117 * cos(2.48675823285  + 3337.35037948650 *T)
    sumX1  +=       0.147 * cos(2.64686332853  + 6510.79580026390 *T)
    sumX1  +=       0.118 * cos(2.66713380790  + 3531.81687532789 *T)
    sumX1  +=       0.160 * cos(3.48000571787  + 3451.06256540270 *T)
    sumX1  +=       0.117 * cos(0.45598660135  + 3337.31587218211 *T)
    sumX1  +=       0.117 * cos(5.66778031729  + 3344.39661618450 *T)
    sumX1  +=       0.118 * cos(5.77425794908  + 9887.01601754750 *T)
    sumX1  +=       0.119 * cos(2.86055893849  + 15907.00794416610 *T)
    sumX1  +=       0.154 * cos(2.03788737462  + 3134.15550688670 *T)
    sumX1  +=       0.118 * cos(2.90358676347  + 3363.62474079271 *T)
    sumX1  +=       0.118 * cos(0.07858750158  + 3318.08774757391 *T)
    sumX1  +=       0.117 * cos(2.63259162370  + 3184.45552368090 *T)
    sumX1  +=       0.117 * cos(0.34958264135  + 3497.25696468570 *T)
    sumX1  +=       0.115 * cos(3.42928363203  + 6725.18758318811 *T)
    sumX1  +=       0.116 * cos(1.31026929252  + 5243.80542171749 *T)
    sumX1  +=       0.124 * cos(4.75071984619  + 2631.86644190671 *T)
    sumX1  +=       0.135 * cos(4.74864603951  + 3603.94016755610 *T)
    sumX1  +=       0.112 * cos(1.39378918277  + 9065.79194161230 *T)
    sumX1  +=       0.152 * cos(3.41529704725  + 1.22413855171 *T)
    sumX1  +=       0.127 * cos(5.86796114412  + 7366.02509014209 *T)
    sumX1  +=       0.119 * cos(0.46273681956  + 16511.85768487090 *T)
    sumX1  +=       0.113 * cos(4.31237181889  + 13745.59005650590 *T)
    sumX1  +=       0.155 * cos(2.36789782817  + 6155.30082413750 *T)
    sumX1  +=       0.115 * cos(4.80039490722  + 6997.86054053970 *T)
    sumX1  +=       0.114 * cos(4.71044015292  + 632.53992182970 *T)
    sumX1  +=       0.111 * cos(1.50115320372  + 87.55202202331 *T)
    sumX1  +=       0.114 * cos(3.09765449003  + 3987.86707749811 *T)
    sumX1  +=       0.125 * cos(1.31378413061  + 5821.15874213030 *T)
    sumX1  +=       0.138 * cos(6.15830900546  + 2233.71743849851 *T)
    sumX1  +=       0.149 * cos(4.65381217458  + 7373.62627210991 *T)
    sumX1  +=       0.131 * cos(0.53786654009  + 6901.88131332190 *T)
    sumX1  +=       0.107 * cos(4.29058441705  + 2973.63191522091 *T)
    sumX1  +=       0.107 * cos(0.28658465594  + 4672.91113172411 *T)
    sumX1  +=       0.145 * cos(1.00064082557  + 2060.57403658830 *T)
    sumX1  +=       0.111 * cos(3.72631581546  + 6357.96325422570 *T)
    sumX1  +=       0.112 * cos(3.97073393136  + 2575.06317973690 *T)
    sumX1  +=       0.139 * cos(0.94733057212  + 3547.55698147990 *T)
    sumX1  +=       0.122 * cos(4.41872868475  + 3346.71344648291 *T)
    sumX1  +=       0.122 * cos(4.84663088748  + 3334.99904188371 *T)
    sumX1  +=       0.114 * cos(4.59065890561  + 931.60690433531 *T)
    sumX1  +=       0.112 * cos(1.44715487405  + 7079.13003932430 *T)
    sumX1  +=       0.108 * cos(1.73727679166  + 8584.41784841729 *T)
    sumX1  +=       0.141 * cos(4.45396803446  + 2907.14450630650 *T)
    sumX1  +=       0.139 * cos(6.03004639149  + 5518.99396647530 *T)
    sumX1  +=       0.115 * cos(4.56654646438  + 4399.75053940550 *T)
    sumX1  +=       0.103 * cos(0.56675408892  + 2487.65986243130 *T)
    sumX1  +=       0.140 * cos(6.09857806608  + 6670.83200553330 *T)
    sumX1  +=       0.105 * cos(3.60113930342  + 2957.47207699310 *T)
    sumX1  +=       0.102 * cos(1.89951054497  + 316.14805217309 *T)
    sumX1  +=       0.100 * cos(2.58249576503  + 17918.04097030310 *T)
    sumX1  +=       0.120 * cos(0.00199458528  + 3311.36442587990 *T)
    sumX1  +=       0.120 * cos(2.98017967977  + 3370.34806248671 *T)
    sumX1  +=       0.118 * cos(2.40172384456  + 6606.68707231580 *T)
    sumX1  +=       0.098 * cos(1.31051304917  + 17468.61138046190 *T)
    sumX1  +=       0.097 * cos(5.18663624065  + 8859.60639317511 *T)
    sumX1  +=       0.107 * cos(0.39903321622  + 2391.19300024650 *T)
    sumX1  +=       0.132 * cos(5.74613926898  + 3050.37076471370 *T)
    sumX1  +=       0.106 * cos(5.77391871927  + 2641.58509595570 *T)
    sumX1  +=       0.097 * cos(2.15228847060  + 22345.01655862470 *T)
    sumX1  +=       0.101 * cos(3.76810604314  + 10125.17387180150 *T)
    sumX1  +=       0.104 * cos(1.92314989559  + 4297.14540015391 *T)
    sumX1  +=       0.096 * cos(0.92848572409  + 2679.62331740231 *T)
    sumX1  +=       0.111 * cos(1.28913302911  + 10042.43641698170 *T)
    sumX1  +=       0.110 * cos(4.68759919971  + 2913.77041834030 *T)
    sumX1  +=       0.096 * cos(4.84024975949  + 10006.82862639790 *T)
    sumX1  +=       0.112 * cos(3.79531321228  + 11251.04321090511 *T)
    sumX1  +=       0.096 * cos(0.08944555474  + 59.13004443009 *T)
    sumX1  +=       0.108 * cos(1.41251308034  + 3735.48212924251 *T)
    sumX1  +=       0.092 * cos(4.97016796965  + 3708.08057314570 *T)
    sumX1  +=       0.097 * cos(3.70144815543  + 2922.35192137631 *T)
    sumX1  +=       0.127 * cos(4.88627440223  + 4272.21933103511 *T)
    sumX1  +=       0.117 * cos(5.31775060658  + 12168.24651405810 *T)
    sumX1  +=       0.116 * cos(1.49729294895  + 10213.52936369450 *T)
    sumX1  +=       0.115 * cos(3.52057619150  + 9815.89554914570 *T)
    sumX1  +=       0.093 * cos(3.71897405268  + 2772.03437015590 *T)
    sumX1  +=       0.104 * cos(2.14819026672  + 6033.97020260130 *T)
    sumX1  +=       0.091 * cos(2.04346250814  + 169.33636564950 *T)
    sumX1  +=       0.090 * cos(6.01161563012  + 2541.03511901791 *T)
    sumX1  +=       0.100 * cos(1.66100640202  + 26887.60986721410 *T)
    sumX1  +=       0.096 * cos(0.19390309418  + 640.63378989870 *T)
    sumX1  +=       0.090 * cos(2.85101513299  + 846.32665223470 *T)
    sumX1  +=       0.089 * cos(1.82000345485  + 2817.79119112250 *T)
    sumX1  +=       0.115 * cos(2.32563511814  + 1648.20293971390 *T)
    sumX1  +=       0.096 * cos(4.51369552987  + 4049.84604645990 *T)
    sumX1  +=       0.113 * cos(0.21883056920  + 4922.81559246630 *T)
    sumX1  +=       0.087 * cos(5.73909456316  + 1118.51197461930 *T)
    sumX1  +=       0.094 * cos(4.82163973372  + 1272.92484311070 *T)
    sumX1  +=       0.110 * cos(4.05120684954  + 21795.45790909830 *T)
    sumX1  +=       0.103 * cos(4.54724340391  + 12302.82395912650 *T)
    sumX1  +=       0.088 * cos(0.58289504579  + 17654.53672226610 *T)
    sumX1  +=       0.119 * cos(3.51883656664  + 3631.34172365290 *T)
    sumX1  +=       0.094 * cos(4.87154048069  + 10706.63751684189 *T)
    sumX1  +=       0.090 * cos(5.20227178872  + 19800.70213874130 *T)
    sumX1  +=       0.117 * cos(6.22319339964  + 1905.70858242391 *T)
    sumX1  +=       0.098 * cos(2.27543509085  + 6768.77687542291 *T)
    sumX1  +=       0.084 * cos(0.40663774606  + 3152.93109642071 *T)
    sumX1  +=       0.084 * cos(5.71712518213  + 3528.78139194590 *T)
    sumX1  +=       0.088 * cos(2.48359504580  + 36.84918278771 *T)
    sumX1  +=       0.112 * cos(2.17506174010  + 11925.03027511709 *T)
    sumX1  +=       0.093 * cos(2.83649458406  + 2538.49232173710 *T)
    sumX1  +=       0.091 * cos(4.87521452246  + 10714.23869880971 *T)
    sumX1  +=       0.094 * cos(1.69069857507  + 6048.19729660290 *T)
    sumX1  +=       0.110 * cos(3.42458146112  + 21946.86755521649 *T)
    sumX1  +=       0.085 * cos(1.40842506800  + 16490.00685554450 *T)
    sumX1  +=       0.097 * cos(0.03832228360  + 2170.78035780770 *T)
    sumX1  +=       0.094 * cos(6.04928950566  + 12943.20888040750 *T)
    sumX1  +=       0.088 * cos(3.52169323662  + 4040.12739241091 *T)
    sumX1  +=       0.107 * cos(1.79263478591  + 6286.84278582391 *T)
    sumX1  +=       0.082 * cos(4.84136227056  + 21548.71855180830 *T)
    sumX1  +=       0.081 * cos(0.90487003246  + 3355.08333818490 *T)
    sumX1  +=       0.081 * cos(2.07730423259  + 3326.62915018171 *T)
    sumX1  +=       0.085 * cos(5.29390218197  + 4106.64930862970 *T)
    sumX1  +=       0.082 * cos(4.45108278321  + 18208.10612510850 *T)
    sumX1  +=       0.101 * cos(3.27831724076  + 5628.71265272770 *T)
    sumX1  +=       0.079 * cos(3.08021794234  + 4778.03185838191 *T)
    sumX1  +=       0.083 * cos(5.56389512532  + 3759.36056699030 *T)
    sumX1  +=       0.081 * cos(5.24671474039  + 1442.01739127670 *T)
    sumX1  +=       0.079 * cos(4.99629515155  + 3302.72320854550 *T)
    sumX1  +=       0.079 * cos(4.26906442068  + 3378.98927982111 *T)
    sumX1  +=       0.109 * cos(5.82044508542  + 765.54924696289 *T)
    sumX1  +=       0.079 * cos(2.75167670672  + 9947.29949901560 *T)
    sumX1  +=       0.087 * cos(0.74054571271  + 3181.91272640011 *T)
    sumX1  +=       0.092 * cos(1.81679365007  + 684.80023674250 *T)
    sumX1  +=       0.078 * cos(0.54443036545  + 4025.41266344230 *T)
    sumX1  +=       0.106 * cos(0.39437492385  + 7214.68275432671 *T)
    sumX1  +=       0.079 * cos(3.24883118288  + 5562.71287878030 *T)
    sumX1  +=       0.094 * cos(3.86479398938  + 1655.31648671470 *T)
    sumX1  +=       0.077 * cos(4.37397781320  + 42942.74815663290 *T)
    sumX1  +=       0.096 * cos(3.10433237844  + 4447.99504986810 *T)
    sumX1  +=       0.079 * cos(0.57783819707  + 1169.83206889211 *T)
    sumX1  +=       0.077 * cos(2.05082432743  + 4002.08917096430 *T)
    sumX1  +=       0.075 * cos(1.45657411241  + 4988.81536641370 *T)
    sumX1  +=       0.103 * cos(4.88591124475  + 6279.79654912590 *T)
    sumX1  +=       0.075 * cos(2.93994155162  + 6710.89817941910 *T)
    sumX1  +=       0.100 * cos(2.80459580595  + 6654.06711478630 *T)
    sumX1  +=       0.075 * cos(3.69701859291  + 1106.89498820129 *T)
    sumX1  +=       0.085 * cos(4.64963704633  + 3031.57792152750 *T)
    sumX1  +=       0.085 * cos(4.61572252590  + 3650.13456683910 *T)
    sumX1  +=       0.102 * cos(4.01749947544  + 220.16882495529 *T)
    sumX1  +=       0.077 * cos(5.54308532188  + 3909.67811821070 *T)
    sumX1  +=       0.080 * cos(2.35937385984  + 42153.72518556549 *T)
    sumX1  +=       0.090 * cos(5.20545196866  + 2277.95119564511 *T)
    sumX1  +=       0.074 * cos(4.85392128576  + 10184.54773371510 *T)
    sumX1  +=       0.102 * cos(4.85789296676  + 3282.97754817950 *T)
    sumX1  +=       0.102 * cos(4.40746660547  + 3398.73494018711 *T)
    sumX1  +=       0.077 * cos(0.97150754740  + 20199.27146681370 *T)
    sumX1  +=       0.073 * cos(5.35302732038  + 13369.80707128350 *T)
    sumX1  +=       0.097 * cos(1.78236800999  + 6784.56144510170 *T)
    sumX1  +=       0.092 * cos(2.49268078499  + 3492.75352526850 *T)
    sumX1  +=       0.076 * cos(2.72604775599  + 19404.02109136830 *T)
    sumX1  +=       0.073 * cos(2.28914210106  + 6365.07680122650 *T)
    sumX1  +=       0.080 * cos(4.31946976428  + 8756.51361895650 *T)
    sumX1  +=       0.077 * cos(2.43358799510  + 6549.92670919670 *T)
    sumX1  +=       0.072 * cos(1.60251649866  + 2067.68758358909 *T)
    sumX1  +=       0.073 * cos(1.66952560844  + 20809.22380716170 *T)
    sumX1  +=       0.073 * cos(2.18603588252  + 8955.09798542589 *T)
    sumX1  +=       0.088 * cos(6.05370075400  + 11883.82695021830 *T)
    sumX1  +=       0.070 * cos(0.54574968829  + 6254.38284504009 *T)
    sumX1  +=       0.093 * cos(3.37029033662  + 10050.53028505070 *T)
    sumX1  +=       0.074 * cos(5.10982595496  + 12725.20961729149 *T)
    sumX1  +=       0.069 * cos(6.03979719075  + 1280.03839011150 *T)
    sumX1  +=       0.090 * cos(6.03814884327  + 10713.75106384269 *T)
    sumX1  +=       0.082 * cos(5.40449104981  + 5401.18646328809 *T)
    sumX1  +=       0.072 * cos(2.91500867928  + 18451.32236404950 *T)
    sumX1  +=       0.077 * cos(0.47267389836  + 5466.73365156250 *T)
    sumX1  +=       0.079 * cos(1.60573790161  + 3351.90194444720 *T)
    sumX1  +=       0.079 * cos(1.37643636344  + 3329.81054391940 *T)
    sumX1  +=       0.077 * cos(4.93901006797  + 15636.81047379250 *T)
    sumX1  +=       0.070 * cos(5.75932631153  + 2011.34415392330 *T)
    sumX1  +=       0.068 * cos(3.77776938096  + 9381.20349020070 *T)
    sumX1  +=       0.074 * cos(4.22015140549  + 25136.07033579810 *T)
    sumX1  +=       0.068 * cos(0.23736554031  + 6489.53276980449 *T)
    sumX1  +=       0.078 * cos(1.87967755594  + 3583.09721319030 *T)
    sumX1  +=       0.069 * cos(5.04882196655  + 11354.13598512370 *T)
    sumX1  +=       0.070 * cos(3.84126763579  + 25287.47998191629 *T)
    sumX1  +=       0.070 * cos(3.95696412828  + 35707.46626542390 *T)
    sumX1  +=       0.088 * cos(3.09948179111  + 3068.18328683170 *T)
    sumX1  +=       0.088 * cos(6.16582202980  + 3613.52920153491 *T)
    sumX1  +=       0.065 * cos(3.39284738601  + 15121.34660269950 *T)
    sumX1  +=       0.077 * cos(0.99849955209  + 14127.99895376209 *T)
    sumX1  +=       0.069 * cos(1.59437276472  + 43739.04616344930 *T)
    sumX1  +=       0.066 * cos(1.03311700868  + 5195.48854981791 *T)
    sumX1  +=       0.066 * cos(4.17470966227  + 5195.00091485090 *T)
    sumX1  +=       0.065 * cos(0.76344895489  + 6708.87022697990 *T)
    sumX1  +=       0.071 * cos(5.13055527150  + 10176.94655174729 *T)
    sumX1  +=       0.090 * cos(5.15547486786  + 6740.84253279671 *T)
    sumX1  +=       0.065 * cos(0.10668764820  + 4140.67736934870 *T)
    sumX1  +=       0.064 * cos(4.35311548094  + 9801.66845514410 *T)
    sumX1  +=       0.085 * cos(2.97975451165  + 43340.89716004110 *T)
    sumX1  +=       0.087 * cos(5.05289595532  + 9859.61446145070 *T)
    sumX1  +=       0.074 * cos(0.32205798976  + 10596.42589591770 *T)
    sumX1  +=       0.065 * cos(0.93949119244  + 6681.04859478950 *T)
    sumX1  +=       0.071 * cos(0.13869046417  + 4143.22016662950 *T)
    sumX1  +=       0.066 * cos(1.89967622669  + 3267.25528826810 *T)
    sumX1  +=       0.066 * cos(1.08249803836  + 3414.45720009850 *T)
    sumX1  +=       0.063 * cos(1.22371219128  + 6467.68194047809 *T)
    sumX1  +=       0.064 * cos(2.27838099501  + 3487.66793070691 *T)
    sumX1  +=       0.064 * cos(0.70379327004  + 3194.04455765970 *T)
    sumX1  +=       0.085 * cos(4.60745592882  + 10720.86461084350 *T)
    sumX1  +=       0.062 * cos(1.44691455296  + 35.78404919390 *T)
    sumX1  +=       0.062 * cos(1.44691455296  + 36.27168416091 *T)
    sumX1  +=       0.064 * cos(0.50126921709  + 11773.13299403189 *T)
    sumX1  +=       0.064 * cos(3.39664176030  + 41747.81351144129 *T)
    sumX1  +=       0.068 * cos(4.85567854407  + 12809.12412144031 *T)
    sumX1  +=       0.062 * cos(3.24713366848  + 18461.95902939930 *T)
    sumX1  +=       0.068 * cos(1.99986373490  + 6614.98126196870 *T)
    sumX1  +=       0.079 * cos(3.63912822114  + 11236.81611690350 *T)
    sumX1  +=       0.066 * cos(0.00739681430  + 9998.20265983511 *T)
    sumX1  +=       0.064 * cos(3.57531615713  + 11526.88127170889 *T)
    sumX1  +=       0.070 * cos(1.91919982448  + 16858.65904011390 *T)
    sumX1  +=       0.065 * cos(2.25710269131  + 20597.00014555770 *T)
    sumX1  +=       0.064 * cos(6.00619877627  + 40796.58274015770 *T)
    sumX1  +=       0.079 * cos(3.24064854543  + 10730.58326489249 *T)
    sumX1  +=       0.062 * cos(2.19007422012  + 20040.39525933330 *T)
    sumX1  +=       0.066 * cos(4.61504680225  + 8428.99744898310 *T)
    sumX1  +=       0.079 * cos(1.14777889484  + 40398.43373674950 *T)
    sumX1  +=       0.059 * cos(4.95064610711  + 44959.30385850670 *T)
    sumX1  +=       0.063 * cos(1.57149098969  + 20743.18952590990 *T)
    sumX1  +=       0.060 * cos(5.59502027006  + 10016.55817327550 *T)
    sumX1  +=       0.059 * cos(2.37148961772  + 13745.10242153889 *T)
    sumX1  +=       0.067 * cos(4.66236805003  + 9329.92349635610 *T)
    sumX1  +=       0.078 * cos(3.20257989132  + 24606.37937070350 *T)
    sumX1  +=       0.080 * cos(4.78522853089  + 5202.60209681870 *T)
    sumX1  +=       0.057 * cos(0.14787405732  + 12928.98178640591 *T)
    sumX1  +=       0.080 * cos(2.57142552579  + 27417.30083230870 *T)
    sumX1  +=       0.077 * cos(2.68569752236  + 4845.65641830930 *T)
    sumX1  +=       0.062 * cos(1.28402834481  + 12987.86801335249 *T)
    sumX1  +=       0.057 * cos(3.31195041686  + 5820.67110716330 *T)
    sumX1  +=       0.057 * cos(4.71242438007  + 25841.04938475869 *T)
    sumX1  +=       0.059 * cos(1.99234809567  + 5945.59215735130 *T)
    sumX1  +=       0.061 * cos(2.21763600824  + 3264.59017290771 *T)
    sumX1  +=       0.058 * cos(0.15438737921  + 15423.51137835450 *T)
    sumX1  +=       0.074 * cos(1.62116168437  + 9623.44445920770 *T)
    sumX1  +=       0.062 * cos(0.01729652290  + 3324.02404313351 *T)
    sumX1  +=       0.062 * cos(2.96487774215  + 3357.68844523310 *T)
    sumX1  +=       0.060 * cos(3.21935705290  + 9225.29545579949 *T)
    sumX1  +=       0.061 * cos(0.87383375009  + 5223.93773728571 *T)
    sumX1  +=       0.056 * cos(6.08238895883  + 3702.97738091410 *T)
    sumX1  +=       0.056 * cos(3.18297061340  + 2978.73510745250 *T)
    sumX1  +=       0.055 * cos(5.55769169501  + 12715.68269096791 *T)
    sumX1  +=       0.058 * cos(6.20637133104  + 14584.05445563709 *T)
    sumX1  +=       0.075 * cos(3.03936450956  + 4776.00390594270 *T)
    sumX1  +=       0.054 * cos(6.20587218477  + 45885.21157992450 *T)
    sumX1  +=       0.055 * cos(0.20267751971  + 2126.05391455991 *T)
    sumX1  +=       0.054 * cos(5.63419925281  + 9381.69617630189 *T)
    sumX1  +=       0.054 * cos(1.14797661971  + 5607.37201172530 *T)
    sumX1  +=       0.053 * cos(4.66773371614  + 6789.13219482170 *T)
    sumX1  +=       0.058 * cos(1.18695310382  + 3268.80051082850 *T)
    sumX1  +=       0.058 * cos(1.79522116123  + 3412.91197753810 *T)
    sumX1  +=       0.053 * cos(4.87665472917  + 12509.00950756369 *T)
    sumX1  +=       0.053 * cos(1.73506207558  + 12509.49714253070 *T)
    sumX1  +=       0.073 * cos(3.99821780645  + 14867.49369840869 *T)
    sumX1  +=       0.053 * cos(3.76863341135  + 14481.44931638550 *T)
    sumX1  +=       0.054 * cos(0.37546635249  + 24491.18197509989 *T)
    sumX1  +=       0.052 * cos(4.72089853524  + 6813.01063256950 *T)
    sumX1  +=       0.054 * cos(3.09595317364  + 8186.26884500910 *T)
    sumX1  +=       0.054 * cos(0.00461175410  + 21150.56954840009 *T)
    sumX1  +=       0.054 * cos(2.87852685560  + 8226.82265386190 *T)
    sumX1  +=       0.054 * cos(1.09489146543  + 18849.47136745770 *T)
    sumX1  +=       0.068 * cos(0.66923543045  + 20213.49856081530 *T)
    sumX1  +=       0.062 * cos(1.20435016011  + 9175.99826283170 *T)
    sumX1  +=       0.068 * cos(0.82610831691  + 2391.68063521350 *T)
    sumX1  +=       0.055 * cos(4.11195400528  + 23539.46356884929 *T)
    sumX1  +=       0.072 * cos(0.36267682866  + 3342.34071689161 *T)
    sumX1  +=       0.072 * cos(2.61949743639  + 3339.37177147500 *T)
    sumX1  +=       0.055 * cos(2.02910200345  + 6843.44767204830 *T)
    sumX1  +=       0.060 * cos(0.59388124144  + 4885.72259219509 *T)
    sumX1  +=       0.052 * cos(0.73783930820  + 7799.73682801889 *T)
    sumX1  +=       0.063 * cos(3.16724562752  + 21265.76694400370 *T)
    sumX1  +=       0.069 * cos(5.26085614502  + 24889.33097850809 *T)
    sumX1  +=       0.051 * cos(4.53705971735  + 37895.67010785090 *T)
    sumX1  +=       0.052 * cos(3.81855873195  + 45088.42593814109 *T)
    sumX1  +=       0.051 * cos(6.16736633171  + 12410.97511803210 *T)
    sumX1  +=       0.068 * cos(4.79486677262  + 1854.38848815109 *T)
    sumX1  +=       0.054 * cos(3.63347214209  + 2736.00650347850 *T)
    sumX1  +=       0.052 * cos(0.45870559028  + 16328.48044005229 *T)
    sumX1  +=       0.051 * cos(2.71896338300  + 16063.40866466850 *T)
    sumX1  +=       0.061 * cos(4.37753041368  + 2655.74487965451 *T)
    sumX1  +=       0.056 * cos(6.12270264253  + 8439.63411433290 *T)
    sumX1  +=       0.050 * cos(1.66638675730  + 1727.90491115171 *T)
    sumX1  +=       0.054 * cos(4.41236851357  + 5408.78764525591 *T)
    sumX1  +=       0.059 * cos(5.82807691891  + 12516.61068953150 *T)
    sumX1  +=       0.052 * cos(2.44824121469  + 4356.03162710049 *T)
    sumX1  +=       0.050 * cos(5.98909587083  + 3116.51144848140 *T)
    sumX1  +=       0.050 * cos(3.27629040044  + 3565.20103988520 *T)
    sumX1  +=       0.049 * cos(3.04482195507  + 55516.66252733170 *T)
    sumX1  +=       0.050 * cos(3.76474251754  + 3742.52836594050 *T)
    sumX1  +=       0.048 * cos(1.55364396872  + 1882.83767561889 *T)
    sumX1  +=       0.054 * cos(1.33668194066  + 20426.81490990550 *T)
    sumX1  +=       0.048 * cos(4.53136736783  + 16865.28495214769 *T)
    sumX1  +=       0.061 * cos(3.94417006930  + 19617.32018680630 *T)
    sumX1  +=       0.048 * cos(1.52363711730  + 19190.72199593030 *T)
    sumX1  +=       0.048 * cos(4.57834518343  + 24341.28449997309 *T)
    sumX1  +=       0.056 * cos(0.62265341478  + 10787.38652706230 *T)
    sumX1  +=       0.048 * cos(1.47112939196  + 44768.34322736210 *T)
    sumX1  +=       0.047 * cos(4.52109505540  + 13517.55897844709 *T)
    sumX1  +=       0.051 * cos(4.70693407374  + 5750.10558403130 *T)
    sumX1  +=       0.053 * cos(0.31337304133  + 1971.70835637129 *T)
    sumX1  +=       0.047 * cos(0.76973666278  + 11876.71340321750 *T)
    sumX1  +=       0.047 * cos(5.80540140583  + 2257.59587624631 *T)
    sumX1  +=       0.053 * cos(0.80336106618  + 20043.43074271529 *T)
    sumX1  +=       0.050 * cos(2.12975781859  + 32765.00284213229 *T)
    sumX1  +=       0.047 * cos(0.13685654516  + 39048.56632709070 *T)
    sumX1  +=       0.048 * cos(3.00714578495  + 135.30889751891 *T)
    sumX1  +=       0.064 * cos(3.04700054519  + 1329.75590774351 *T)
    sumX1  +=       0.051 * cos(5.44594778277  + 2693.84541086851 *T)
    sumX1  +=       0.055 * cos(2.85027352182  + 5834.89820116490 *T)
    sumX1  +=       0.046 * cos(0.27718268152  + 1795.50226120450 *T)
    sumX1  +=       0.055 * cos(4.04664328548  + 4403.76129272151 *T)
    sumX1  +=       0.048 * cos(0.33882252916  + 14955.28953791550 *T)
    sumX1  +=       0.059 * cos(2.13527332496  + 4290.03185315310 *T)
    sumX1  +=       0.047 * cos(1.62007289183  + 35464.73766144989 *T)
    sumX1  +=       0.047 * cos(4.76166554542  + 35465.22529641690 *T)
    sumX1  +=       0.045 * cos(0.23464093612  + 5835.38583613191 *T)
    sumX1  +=       0.052 * cos(3.84343144307  + 17660.05964657349 *T)
    sumX1  +=       0.048 * cos(0.93046633940  + 1859.44687283811 *T)
    sumX1  +=       0.056 * cos(3.86355359519  + 532.97026596010 *T)
    sumX1  +=       0.052 * cos(0.80583034945  + 3718.22985209910 *T)
    sumX1  +=       0.052 * cos(2.17634391560  + 2963.48263626750 *T)
    sumX1  +=       0.045 * cos(0.08670687387  + 22487.12787535809 *T)
    sumX1  +=       0.061 * cos(4.60323457975  + 7380.73981911070 *T)
    sumX1  +=       0.045 * cos(2.80587535059  + 38252.26832027430 *T)
    sumX1  +=       0.045 * cos(3.33165976697  + 8741.79888998789 *T)
    sumX1  +=       0.054 * cos(1.04355837542  + 1056.10262932370 *T)
    sumX1  +=       0.045 * cos(2.35300742152  + 5511.88041947450 *T)
    sumX1  +=       0.045 * cos(5.22801084585  + 664.99986261351 *T)
    sumX1  +=       0.045 * cos(2.08641819226  + 664.51222764649 *T)
    sumX1  +=       0.045 * cos(3.15300518185  + 3973.63998349650 *T)
    sumX1  +=       0.044 * cos(3.22051341282  + 5401.67409825510 *T)
    sumX1  +=       0.054 * cos(1.65601444750  + 931.11926936830 *T)
    sumX1  +=       0.046 * cos(0.11012295064  + 5881.64754571770 *T)
    sumX1  +=       0.050 * cos(2.52491431413  + 1744.66980189871 *T)
    sumX1  +=       0.044 * cos(3.45607612983  + 3244.87701696550 *T)
    sumX1  +=       0.044 * cos(5.80928344240  + 3436.83547140110 *T)
    sumX1  +=       0.045 * cos(4.66994870660  + 1265.81129610991 *T)
    sumX1  +=       0.044 * cos(6.06839188413  + 17499.60395779890 *T)
    sumX1  +=       0.043 * cos(5.89648781204  + 97670.14389541368 *T)
    sumX1  +=       0.057 * cos(4.80979061249  + 46283.36058333270 *T)
    sumX1  +=       0.044 * cos(4.99066032403  + 10191.17364574889 *T)
    sumX1  +=       0.043 * cos(2.87708345328  + 33163.63948050750 *T)
    sumX1  +=       0.047 * cos(3.84961762116  + 17256.87535382490 *T)
    sumX1  +=       0.044 * cos(6.17241939552  + 2708.07250487011 *T)
    sumX1  +=       0.059 * cos(5.12086252013  + 2524.20291796811 *T)
    sumX1  +=       0.053 * cos(0.56166352340  + 8337.02897508130 *T)
    sumX1  +=       0.053 * cos(4.32226690599  + 14054.85112550950 *T)
    sumX1  +=       0.043 * cos(5.31336835297  + 12200.21881987490 *T)
    sumX1  +=       0.046 * cos(5.92396722498  + 1648.69057468091 *T)
    sumX1  +=       0.042 * cos(2.74816043776  + 845.83901726770 *T)
    sumX1  +=       0.043 * cos(4.63799596013  + 41194.73174356590 *T)
    sumX1  +=       0.047 * cos(4.18521043124  + 13531.78607244869 *T)
    sumX1  +=       0.047 * cos(1.11517601592  + 15850.10956923050 *T)
    sumX1  +=       0.042 * cos(5.91221095313  + 12729.90978496950 *T)
    sumX1  +=       0.043 * cos(3.52376715745  + 15107.11950869790 *T)
    sumX1  +=       0.043 * cos(1.90839497963  + 8123.72987964330 *T)
    sumX1  +=       0.046 * cos(1.99766003910  + 2325.19322629910 *T)
    sumX1  +=       0.043 * cos(5.33982186623  + 14469.34469500049 *T)
    sumX1  +=       0.043 * cos(2.75590007489  + 2854.88419139370 *T)
    sumX1  +=       0.043 * cos(0.22627419016  + 3826.82829697291 *T)
    sumX1  +=       0.041 * cos(4.48535365931  + 42544.59915322470 *T)
    sumX1  +=       0.042 * cos(0.81046200996  + 2652.22176130550 *T)
    sumX1  +=       0.042 * cos(4.75852639657  + 38086.63073899550 *T)
    sumX1  +=       0.045 * cos(5.07021316431  + 3361.14400470870 *T)
    sumX1  +=       0.043 * cos(5.68935522674  + 28229.94340520789 *T)
    sumX1  +=       0.048 * cos(1.95422763977  + 12825.88901218730 *T)
    sumX1  +=       0.052 * cos(5.96852999978  + 6667.24157688150 *T)
    sumX1  +=       0.041 * cos(2.26231059465  + 3296.20241094311 *T)
    sumX1  +=       0.041 * cos(0.71986367040  + 3385.51007742350 *T)
    sumX1  +=       0.043 * cos(3.41833244496  + 46681.50958674090 *T)
    sumX1  +=       0.040 * cos(4.95644488553  + 6851.04885401611 *T)
    sumX1  +=       0.040 * cos(2.54157100544  + 8983.05448679250 *T)
    sumX1  +=       0.043 * cos(5.68028575617  + 20205.89737884749 *T)
    sumX1  +=       0.040 * cos(1.93822990752  + 20728.96243190830 *T)
    sumX1  +=       0.044 * cos(0.93531414815  + 29698.53132881930 *T)
    sumX1  +=       0.041 * cos(3.51275275749  + 4670.36833444331 *T)
    sumX1  +=       0.052 * cos(4.40335159538  + 3299.30545319850 *T)
    sumX1  +=       0.052 * cos(4.86200797685  + 3382.40703516810 *T)
    sumX1  +=       0.041 * cos(1.79203239583  + 48827.67500321610 *T)
    sumX1  +=       0.046 * cos(3.58085963587  + 19247.62037086590 *T)
    sumX1  +=       0.054 * cos(2.95533474059  + 3319.59321372389 *T)
    sumX1  +=       0.039 * cos(3.66982465499  + 743.23387801611 *T)
    sumX1  +=       0.039 * cos(3.73873153389  + 26357.43126715249 *T)
    sumX1  +=       0.039 * cos(0.59713888030  + 26357.91890211950 *T)
    sumX1  +=       0.049 * cos(4.88931538366  + 4025.96760871210 *T)
    sumX1  +=       0.054 * cos(2.40625176831  + 574.58861581830 *T)
    sumX1  +=       0.052 * cos(2.35515997528  + 1461.29786942991 *T)
    sumX1  +=       0.041 * cos(1.31959971528  + 45487.06257651630 *T)
    sumX1  +=       0.051 * cos(5.06339472166  + 102.84895673509 *T)
    sumX1  +=       0.045 * cos(1.82993488286  + 2715.18605187090 *T)
    sumX1  +=       0.045 * cos(1.15223938219  + 3966.52643649571 *T)

    # MARS X Coord
    sumX2  = 0
    sumX2  +=   68999.338 * cos(5.55726456394  + 0.24381748350 *T)
    sumX2  +=   82190.105 * cos(1.49108837587  + 3340.85624418330 *T)
    sumX2  +=   21826.465 * cos(3.36498183687  + 6681.46867088311 *T)
    sumX2  +=    6036.287 * cos(3.88765559226  + 10022.08109758290 *T)
    sumX2  +=    1291.999 * cos(3.87907870157  + 3340.36860921629 *T)
    sumX2  +=    1125.170 * cos(4.26813980903  + 13362.69352428270 *T)
    sumX2  +=     182.064 * cos(4.62383908892  + 16703.30595098250 *T)
    sumX2  +=     172.815 * cos(4.51953182882  + 3337.33312583431 *T)
    sumX2  +=     164.573 * cos(4.61946633733  + 3344.37936253230 *T)
    sumX2  +=     128.077 * cos(4.44271682598  + 6680.98103591609 *T)
    sumX2  +=      93.944 * cos(0.94466158704  + 3185.43584474911 *T)
    sumX2  +=      89.463 * cos(2.03179886201  + 3496.27664361750 *T)
    sumX2  +=      40.244 * cos(5.00717764881  + 6684.99178923210 *T)
    sumX2  +=      36.661 * cos(1.60353028933  + 1059.62574767270 *T)
    sumX2  +=      27.490 * cos(4.97123271543  + 20043.91837768230 *T)
    sumX2  +=      24.847 * cos(3.58016850622  + 3149.40797807170 *T)
    sumX2  +=      23.186 * cos(5.57924030683  + 3532.30451029490 *T)
    sumX2  +=      24.495 * cos(5.73028409392  + 2544.55823736690 *T)
    sumX2  +=      20.269 * cos(3.44189573522  + 4137.15425099970 *T)
    sumX2  +=      18.207 * cos(3.43690977415  + 3098.12764020930 *T)
    sumX2  +=      17.365 * cos(4.15861973569  + 2942.70724077510 *T)
    sumX2  +=      16.922 * cos(5.82133860801  + 3583.58484815730 *T)
    sumX2  +=      16.265 * cos(1.04984131412  + 5622.08674069391 *T)
    sumX2  +=      15.988 * cos(1.61453370816  + 1194.69082770810 *T)
    sumX2  +=      15.705 * cos(4.76611870892  + 3739.00524759150 *T)
    sumX2  +=      13.211 * cos(5.14469570759  + 1990.98883452450 *T)
    sumX2  +=      16.835 * cos(2.34831657807  + 2787.28684134090 *T)
    sumX2  +=      12.623 * cos(3.74877336094  + 2389.13783793270 *T)
    sumX2  +=      13.191 * cos(4.82160322627  + 10021.59346261590 *T)
    sumX2  +=      12.309 * cos(0.81667763054  + 3339.87592311510 *T)
    sumX2  +=      15.092 * cos(0.24928051607  + 1592.83983111630 *T)
    sumX2  +=      11.292 * cos(1.29869510013  + 6526.04827144891 *T)
    sumX2  +=      12.308 * cos(2.12893393689  + 3341.83656525151 *T)
    sumX2  +=      11.493 * cos(2.85941655933  + 796.54182429990 *T)
    sumX2  +=      14.005 * cos(0.62664625135  + 3894.42564702570 *T)
    sumX2  +=      10.521 * cos(2.12777102096  + 4400.23817437250 *T)
    sumX2  +=      13.157 * cos(2.63040425407  + 155.17658195069 *T)
    sumX2  +=       9.155 * cos(5.50505997499  + 4292.57465043390 *T)
    sumX2  +=       8.383 * cos(4.10001775855  + 4690.72365384210 *T)
    sumX2  +=      10.715 * cos(0.86217417712  + 2146.40923395870 *T)
    sumX2  +=       8.087 * cos(5.34109961713  + 10025.60421593190 *T)
    sumX2  +=       8.070 * cos(1.27034526598  + 5487.02166065850 *T)
    sumX2  +=       8.313 * cos(1.22840731664  + 2957.95971196010 *T)
    sumX2  +=       7.119 * cos(2.20406602795  + 1589.31671276730 *T)
    sumX2  +=       7.434 * cos(1.73090220698  + 3723.75277640650 *T)
    sumX2  +=       8.463 * cos(2.69211325650  + 5088.87265725030 *T)
    sumX2  +=       8.851 * cos(2.06610908753  + 4535.30325440790 *T)
    sumX2  +=       6.118 * cos(5.33366394762  + 7079.61767429131 *T)
    sumX2  +=       5.446 * cos(6.02298801760  + 5885.17066406670 *T)
    sumX2  +=       5.632 * cos(1.23083025270  + 553.32558535889 *T)
    sumX2  +=       5.307 * cos(5.39355782839  + 242.97242145751 *T)
    sumX2  +=       4.779 * cos(2.26833646163  + 1748.26023055050 *T)
    sumX2  +=       5.798 * cos(4.89156562350  + 7.35736448430 *T)
    sumX2  +=       5.029 * cos(2.19814519647  + 2273.87313202630 *T)
    sumX2  +=       4.779 * cos(0.35723903723  + 3333.74269718250 *T)
    sumX2  +=       4.645 * cos(6.11788961406  + 951.47458876710 *T)
    sumX2  +=       5.308 * cos(4.04427431331  + 6490.02040477150 *T)
    sumX2  +=       5.680 * cos(2.22833819331  + 529.93478257810 *T)
    sumX2  +=       4.316 * cos(0.66431258202  + 4933.45225781610 *T)
    sumX2  +=       4.455 * cos(4.63572690053  + 2288.10022602790 *T)
    sumX2  +=       4.106 * cos(0.58655299185  + 5092.39577559930 *T)
    sumX2  +=       4.484 * cos(3.97431621601  + 3870.54720927791 *T)
    sumX2  +=       4.129 * cos(2.65817400046  + 3347.96979118411 *T)
    sumX2  +=       3.982 * cos(5.31410167478  + 23384.53080438210 *T)
    sumX2  +=       3.668 * cos(5.18866421914  + 3340.83899053111 *T)
    sumX2  +=       3.668 * cos(0.93622666271  + 3340.87349783550 *T)
    sumX2  +=       4.603 * cos(3.84811665743  + 7477.76667769950 *T)
    sumX2  +=       3.137 * cos(3.80753274993  + 6438.74006690910 *T)
    sumX2  +=       3.223 * cos(2.20207720882  + 426.84200835950 *T)
    sumX2  +=       4.083 * cos(4.72879097145  + 1349.62359217529 *T)
    sumX2  +=       3.073 * cos(3.35893621939  + 1747.77259558350 *T)
    sumX2  +=       3.724 * cos(3.92616253855  + 6151.77770578850 *T)
    sumX2  +=       3.580 * cos(0.29090747369  + 2280.98667902709 *T)
    sumX2  +=       2.506 * cos(5.53763489777  + 5331.60126122430 *T)
    sumX2  +=       2.389 * cos(2.03493789554  + 6836.88907031730 *T)
    sumX2  +=       2.777 * cos(2.28004564558  + 382.65271473970 *T)
    sumX2  +=       2.243 * cos(3.67333516968  + 1350.11122714230 *T)
    sumX2  +=       2.179 * cos(2.71210505704  + 6127.89926804070 *T)
    sumX2  +=       2.794 * cos(4.55180867021  + 6677.94555253410 *T)
    sumX2  +=       2.007 * cos(2.53260318164  + 6682.44899195131 *T)
    sumX2  +=       2.204 * cos(5.71912902267  + 9492.39013248830 *T)
    sumX2  +=       1.869 * cos(2.42716744125  + 4929.92913946711 *T)
    sumX2  +=       1.955 * cos(0.78128655783  + 6681.45141723091 *T)
    sumX2  +=       2.432 * cos(0.97455241566  + 640.14110379750 *T)
    sumX2  +=       1.785 * cos(4.64994614803  + 3554.15533962130 *T)
    sumX2  +=       1.956 * cos(2.81122973926  + 6681.48592453530 *T)
    sumX2  +=       2.128 * cos(2.52171277812  + 7740.85060107230 *T)
    sumX2  +=       2.132 * cos(2.76821785309  + 398.39282089170 *T)
    sumX2  +=       1.915 * cos(0.11819972923  + 191.20444862810 *T)
    sumX2  +=       1.620 * cos(2.44109389132  + 7875.91568110771 *T)
    sumX2  +=       1.540 * cos(4.39464833734  + 3127.55714874530 *T)
    sumX2  +=       1.960 * cos(6.16925557034  + 6924.19727485710 *T)
    sumX2  +=       1.559 * cos(0.77292154394  + 2281.47431399410 *T)
    sumX2  +=       1.569 * cos(1.33136574759  + 8969.81271439450 *T)
    sumX2  +=       1.559 * cos(5.17485623460  + 8955.58562039291 *T)
    sumX2  +=       1.390 * cos(5.15686093227  + 13362.20588931570 *T)
    sumX2  +=       1.868 * cos(4.12485939066  + 5729.75026463250 *T)
    sumX2  +=       1.714 * cos(4.40106234765  + 6674.35512388230 *T)
    sumX2  +=       1.444 * cos(2.78777227082  + 2119.00767786191 *T)
    sumX2  +=       1.461 * cos(5.66929508293  + 13366.21664263171 *T)
    sumX2  +=       1.339 * cos(2.11058695972  + 2145.92159899169 *T)
    sumX2  +=       1.222 * cos(0.48334105773  + 6254.87048000710 *T)
    sumX2  +=       1.257 * cos(3.98999018614  + 641.12142486571 *T)
    sumX2  +=       1.485 * cos(1.64745398687  + 9866.66069814871 *T)
    sumX2  +=       1.566 * cos(0.18233365390  + 8962.69916739370 *T)
    sumX2  +=       1.160 * cos(1.59307514109  + 6298.57213865991 *T)
    sumX2  +=       1.124 * cos(6.27801587166  + 12833.00255918811 *T)
    sumX2  +=       1.479 * cos(5.74615387809  + 10420.23010099111 *T)
    sumX2  +=       1.201 * cos(2.89742393965  + 3767.45443505930 *T)
    sumX2  +=       1.207 * cos(3.26671107441  + 3340.78893388051 *T)
    sumX2  +=       1.207 * cos(2.85824918128  + 3340.92355448610 *T)
    sumX2  +=       1.403 * cos(3.27423655733  + 6283.31966747490 *T)
    sumX2  +=       1.082 * cos(2.00077916742  + 12303.31159409350 *T)
    sumX2  +=       1.021 * cos(0.11004697104  + 4562.70481050471 *T)
    sumX2  +=       1.086 * cos(1.57123280785  + 1751.29571393250 *T)
    sumX2  +=       1.153 * cos(4.95873813423  + 7211.15963597770 *T)
    sumX2  +=       1.277 * cos(4.84268967955  + 2699.97863680110 *T)
    sumX2  +=       0.930 * cos(6.12678273780  + 2811.16527908870 *T)
    sumX2  +=       0.918 * cos(1.60703520232  + 155.66421691770 *T)
    sumX2  +=       0.890 * cos(5.56079732420  + 3343.89172756530 *T)
    sumX2  +=       0.991 * cos(4.45489640942  + 9830.63283147130 *T)
    sumX2  +=       1.152 * cos(5.07946933275  + 951.96222373410 *T)
    sumX2  +=       0.877 * cos(4.21408679267  + 10818.37910439930 *T)
    sumX2  +=       1.028 * cos(4.41166071466  + 3981.73385156551 *T)
    sumX2  +=       0.724 * cos(0.20350508903  + 553.81322032590 *T)
    sumX2  +=       0.813 * cos(3.33403943461  + 10022.09835123510 *T)
    sumX2  +=       0.728 * cos(6.18625603354  + 853.44019923550 *T)
    sumX2  +=       0.746 * cos(5.16496010034  + 9623.93209417470 *T)
    sumX2  +=       0.684 * cos(1.02857269442  + 8274.06468451590 *T)
    sumX2  +=       0.813 * cos(1.30336771571  + 10022.06384393071 *T)
    sumX2  +=       0.641 * cos(5.14516311024  + 6681.40136058031 *T)
    sumX2  +=       0.632 * cos(0.73190374320  + 10018.55797923390 *T)
    sumX2  +=       0.616 * cos(4.48165745116  + 7235.03807372550 *T)
    sumX2  +=       0.846 * cos(3.03682654998  + 6688.58221788390 *T)
    sumX2  +=       0.857 * cos(1.92920837540  + 6041.57138456910 *T)
    sumX2  +=       0.572 * cos(4.49573046647  + 529.44714761109 *T)
    sumX2  +=       0.578 * cos(5.24981714161  + 6040.59106350090 *T)
    sumX2  +=       0.568 * cos(2.55555530886  + 15643.92402079330 *T)
    sumX2  +=       0.560 * cos(2.22514515488  + 397.90518592470 *T)
    sumX2  +=       0.661 * cos(5.25544946005  + 6894.76776632111 *T)
    sumX2  +=       0.538 * cos(5.21568797326  + 4690.23601887509 *T)
    sumX2  +=       0.701 * cos(3.64136458654  + 3178.38960805111 *T)
    sumX2  +=       0.689 * cos(4.08112540175  + 796.05418933290 *T)
    sumX2  +=       0.688 * cos(5.61433085796  + 3503.32288031550 *T)
    sumX2  +=       0.545 * cos(4.98081015653  + 2766.51144584850 *T)
    sumX2  +=       0.550 * cos(3.02101538887  + 7633.18707713370 *T)
    sumX2  +=       0.640 * cos(4.73664982480  + 6681.53598118590 *T)
    sumX2  +=       0.510 * cos(1.69224093861  + 3893.93801205869 *T)
    sumX2  +=       0.617 * cos(2.56761131336  + 3253.54803964350 *T)
    sumX2  +=       0.596 * cos(0.41459493687  + 3428.16444872311 *T)
    sumX2  +=       0.629 * cos(0.07153257669  + 6.86972951729 *T)
    sumX2  +=       0.603 * cos(0.30001085052  + 4292.08701546690 *T)
    sumX2  +=       0.548 * cos(1.00098902826  + 6680.48834981490 *T)
    sumX2  +=       0.444 * cos(4.27401799503  + 3915.20104251810 *T)
    sumX2  +=       0.411 * cos(1.56657247102  + 3980.75353049730 *T)
    sumX2  +=       0.560 * cos(5.65797499774  + 26725.14323108190 *T)
    sumX2  +=       0.399 * cos(5.70475834066  + 12964.54452087450 *T)
    sumX2  +=       0.383 * cos(5.96972895101  + 2700.47132290230 *T)
    sumX2  +=       0.409 * cos(5.38057050876  + 7064.36520310630 *T)
    sumX2  +=       0.499 * cos(6.03750484858  + 2914.25805330730 *T)
    sumX2  +=       0.408 * cos(3.26187625259  + 2648.69864295650 *T)
    sumX2  +=       0.494 * cos(2.58791492755  + 1039.27042827391 *T)
    sumX2  +=       0.367 * cos(2.73062985578  + 8270.54156616690 *T)
    sumX2  +=       0.379 * cos(2.88219149441  + 11081.46302777211 *T)
    sumX2  +=       0.354 * cos(2.66145180389  + 1194.20319274110 *T)
    sumX2  +=       0.403 * cos(1.70308177622  + 8031.33608054190 *T)
    sumX2  +=       0.449 * cos(2.42384589322  + 213.05527795450 *T)
    sumX2  +=       0.397 * cos(2.26969467217  + 2544.07060239989 *T)
    sumX2  +=       0.344 * cos(5.97650048429  + 4033.01384541010 *T)
    sumX2  +=       0.314 * cos(4.31468219581  + 6872.91693699471 *T)
    sumX2  +=       0.410 * cos(5.85701038391  + 8433.00820229911 *T)
    sumX2  +=       0.297 * cos(4.80643849748  + 574.10098085129 *T)
    sumX2  +=       0.326 * cos(2.91113862688  + 10023.06141865111 *T)
    sumX2  +=       0.294 * cos(2.83275017779  + 5459.62010456170 *T)
    sumX2  +=       0.278 * cos(2.81997653776  + 11216.52810780750 *T)
    sumX2  +=       0.274 * cos(5.82142747819  + 10551.77206267750 *T)
    sumX2  +=       0.355 * cos(5.89906958059  + 8672.21368792411 *T)
    sumX2  +=       0.287 * cos(3.03599407466  + 3495.78900865049 *T)
    sumX2  +=       0.282 * cos(6.12544103738  + 13760.84252769090 *T)
    sumX2  +=       0.359 * cos(4.43440472664  + 3097.64000524229 *T)
    sumX2  +=       0.266 * cos(5.25677383400  + 10022.14840788570 *T)
    sumX2  +=       0.284 * cos(3.03494190652  + 5828.27228913110 *T)
    sumX2  +=       0.314 * cos(0.37126567078  + 16173.61498588790 *T)
    sumX2  +=       0.263 * cos(3.04592616684  + 9468.51169474051 *T)
    sumX2  +=       0.277 * cos(2.72843223796  + 3336.84549086730 *T)
    sumX2  +=       0.251 * cos(5.79261076520  + 2942.21960580809 *T)
    sumX2  +=       0.269 * cos(0.77221793249  + 2487.17222746430 *T)
    sumX2  +=       0.256 * cos(3.95807126956  + 5088.38502228330 *T)
    sumX2  +=       0.265 * cos(5.66503537702  + 10022.01378728011 *T)
    sumX2  +=       0.241 * cos(6.22048805000  + 2301.82963339290 *T)
    sumX2  +=       0.232 * cos(0.63600637107  + 13359.17040593370 *T)
    sumX2  +=       0.261 * cos(1.96413122632  + 1221.60474883790 *T)
    sumX2  +=       0.225 * cos(0.50307942308  + 1751.78334889950 *T)
    sumX2  +=       0.252 * cos(4.68430989389  + 1066.73929467350 *T)
    sumX2  +=       0.246 * cos(0.23565072433  + 10264.80970155691 *T)
    sumX2  +=       0.285 * cos(1.72684624328  + 12310.42514109431 *T)
    sumX2  +=       0.243 * cos(4.04327156952  + 9779.35249360890 *T)
    sumX2  +=       0.220 * cos(0.63056559109  + 8429.48508395010 *T)
    sumX2  +=       0.280 * cos(5.94773306444  + 5091.90814063229 *T)
    sumX2  +=       0.214 * cos(5.86156666088  + 2699.49100183409 *T)
    sumX2  +=       0.234 * cos(3.39608420004  + 6468.16957544510 *T)
    sumX2  +=       0.247 * cos(2.62514291276  + 213.54291292150 *T)
    sumX2  +=       0.265 * cos(2.95115218410  + 18984.53644749310 *T)
    sumX2  +=       0.206 * cos(3.53273929640  + 6836.40143535029 *T)
    sumX2  +=       0.198 * cos(3.71412641694  + 13362.71077793490 *T)
    sumX2  +=       0.199 * cos(1.68344049522  + 13362.67627063050 *T)
    sumX2  +=       0.267 * cos(1.77892178245  + 5614.97319369310 *T)
    sumX2  +=       0.200 * cos(0.77310064303  + 7632.69944216670 *T)
    sumX2  +=       0.238 * cos(5.99888163760  + 16706.82906933150 *T)
    sumX2  +=       0.189 * cos(2.67826488440  + 3723.26514143949 *T)
    sumX2  +=       0.187 * cos(1.24354577352  + 1592.35219614930 *T)
    sumX2  +=       0.211 * cos(3.39936925167  + 7108.06686175911 *T)
    sumX2  +=       0.244 * cos(0.11291910384  + 242.48478649050 *T)
    sumX2  +=       0.193 * cos(4.43444237494  + 9070.36269133230 *T)
    sumX2  +=       0.198 * cos(5.26919917088  + 191.69208359510 *T)
    sumX2  +=       0.172 * cos(4.83979658671  + 13171.24525817110 *T)
    sumX2  +=       0.196 * cos(1.99723494678  + 13207.27312484851 *T)
    sumX2  +=       0.234 * cos(2.13538732396  + 7234.55043875850 *T)
    sumX2  +=       0.177 * cos(1.94425535548  + 9639.18456535971 *T)
    sumX2  +=       0.168 * cos(1.75411639556  + 1052.51220067191 *T)
    sumX2  +=       0.170 * cos(2.99535940889  + 4379.88285497370 *T)
    sumX2  +=       0.222 * cos(0.39232779431  + 5642.44206009270 *T)
    sumX2  +=       0.157 * cos(1.36555252862  + 383.14034970670 *T)
    sumX2  +=       0.155 * cos(5.48011073140  + 12296.19804709270 *T)
    sumX2  +=       0.154 * cos(4.61716677676  + 10575.65050042530 *T)
    sumX2  +=       0.155 * cos(5.46273803177  + 16702.81831601549 *T)
    sumX2  +=       0.171 * cos(1.78790613674  + 3184.94820978210 *T)
    sumX2  +=       0.148 * cos(5.73788219136  + 10235.38019302090 *T)
    sumX2  +=       0.140 * cos(4.92817953315  + 6438.25243194210 *T)
    sumX2  +=       0.141 * cos(3.41744392254  + 10029.19464458370 *T)
    sumX2  +=       0.154 * cos(5.87338987371  + 2810.67764412170 *T)
    sumX2  +=       0.137 * cos(1.19383404130  + 1437.41943168211 *T)
    sumX2  +=       0.133 * cos(0.19722245975  + 8827.63408735830 *T)
    sumX2  +=       0.131 * cos(1.82245337339  + 3030.01544531491 *T)
    sumX2  +=       0.134 * cos(0.93479121392  + 2301.34199842589 *T)
    sumX2  +=       0.128 * cos(2.39973337332  + 9382.18381126891 *T)
    sumX2  +=       0.119 * cos(5.77157495793  + 5614.48555872609 *T)
    sumX2  +=       0.120 * cos(5.12604491583  + 3320.08084869090 *T)
    sumX2  +=       0.119 * cos(1.00229912761  + 3361.63163967570 *T)
    sumX2  +=       0.116 * cos(1.15960987836  + 3651.69704305170 *T)
    sumX2  +=       0.117 * cos(5.85304070266  + 6701.82399028190 *T)
    sumX2  +=       0.146 * cos(5.93167541862  + 3738.51761262449 *T)
    sumX2  +=       0.129 * cos(3.89120051809  + 6696.72114206811 *T)
    sumX2  +=       0.118 * cos(5.04634450495  + 149.80701461810 *T)
    sumX2  +=       0.107 * cos(5.12041826560  + 4407.35172137330 *T)
    sumX2  +=       0.114 * cos(4.75751094726  + 7322.34627826531 *T)
    sumX2  +=       0.105 * cos(3.15741664634  + 6666.21619969810 *T)
    sumX2  +=       0.108 * cos(2.60224731647  + 10575.16286545830 *T)
    sumX2  +=       0.115 * cos(3.79062631878  + 9168.88471583091 *T)
    sumX2  +=       0.130 * cos(3.18226883156  + 10973.79950383350 *T)
    sumX2  +=       0.101 * cos(4.45451620205  + 14158.99153109910 *T)
    sumX2  +=       0.102 * cos(1.14962329478  + 4562.21717553770 *T)
    sumX2  +=       0.099 * cos(2.51353107397  + 5621.59910572690 *T)
    sumX2  +=       0.101 * cos(5.32874387689  + 9778.86485864190 *T)
    sumX2  +=       0.115 * cos(4.59270876787  + 3148.92034310470 *T)
    sumX2  +=       0.131 * cos(1.18803355863  + 6661.11335148431 *T)
    sumX2  +=       0.134 * cos(6.02271435541  + 2913.77041834030 *T)
    sumX2  +=       0.113 * cos(3.71424984396  + 6519.00203475090 *T)
    sumX2  +=       0.096 * cos(0.98095706382  + 2700.95895786930 *T)
    sumX2  +=       0.094 * cos(0.24327692845  + 6816.53375091851 *T)
    sumX2  +=       0.097 * cos(1.78626013160  + 11371.94850724170 *T)
    sumX2  +=       0.118 * cos(4.70652198213  + 10213.52936369450 *T)
    sumX2  +=       0.088 * cos(6.06276290521  + 1835.56843509031 *T)
    sumX2  +=       0.096 * cos(2.19847165068  + 15114.23305569870 *T)
    sumX2  +=       0.119 * cos(3.97087141578  + 10177.01386205009 *T)
    sumX2  +=       0.096 * cos(5.83220204707  + 6684.50415426510 *T)
    sumX2  +=       0.117 * cos(5.16366852876  + 9225.78309076650 *T)
    sumX2  +=       0.095 * cos(4.29171920346  + 5486.53402569149 *T)
    sumX2  +=       0.088 * cos(5.74190057756  + 8030.84844557490 *T)
    sumX2  +=       0.085 * cos(0.27003870811  + 6151.29007082150 *T)
    sumX2  +=       0.089 * cos(1.17370697759  + 10177.50149701710 *T)
    sumX2  +=       0.081 * cos(1.80921862202  + 5244.29305668450 *T)
    sumX2  +=       0.081 * cos(3.79172529620  + 1505.04399160949 *T)
    sumX2  +=       0.079 * cos(5.56747633751  + 10404.97762980610 *T)
    sumX2  +=       0.095 * cos(6.07783916606  + 16305.15694757430 *T)
    sumX2  +=       0.090 * cos(4.66517176290  + 10014.96755058210 *T)
    sumX2  +=       0.098 * cos(5.10446957982  + 9808.78200214491 *T)
    sumX2  +=       0.096 * cos(2.86554772752  + 1059.13811270570 *T)
    sumX2  +=       0.073 * cos(0.72497936154  + 6660.69327539070 *T)
    sumX2  +=       0.077 * cos(4.35359928029  + 4136.66661603269 *T)
    sumX2  +=       0.071 * cos(1.03916311153  + 2274.36076699330 *T)
    sumX2  +=       0.075 * cos(5.63557700884  + 7903.31723720450 *T)
    sumX2  +=       0.098 * cos(0.02076697156  + 6040.10342853389 *T)
    sumX2  +=       0.085 * cos(0.44510087290  + 4194.05262593530 *T)
    sumX2  +=       0.073 * cos(0.44168846203  + 9380.71585523369 *T)
    sumX2  +=       0.093 * cos(6.09456085621  + 1990.50119955750 *T)
    sumX2  +=       0.086 * cos(6.18283823038  + 20752.42054499190 *T)
    sumX2  +=       0.069 * cos(1.38553111275  + 1903.68062998470 *T)
    sumX2  +=       0.070 * cos(4.70293199621  + 2388.65020296570 *T)
    sumX2  +=       0.069 * cos(2.17037139628  + 8962.21153242670 *T)
    sumX2  +=       0.066 * cos(4.49413894792  + 13517.62628874989 *T)
    sumX2  +=       0.069 * cos(6.00365908535  + 30065.75565778170 *T)
    sumX2  +=       0.079 * cos(5.98866196477  + 3337.26581553151 *T)
    sumX2  +=       0.079 * cos(0.13660951723  + 3344.44667283510 *T)
    sumX2  +=       0.064 * cos(2.30763395826  + 25287.47998191629 *T)
    sumX2  +=       0.065 * cos(3.24445550921  + 2631.86644190671 *T)
    sumX2  +=       0.065 * cos(1.03862495257  + 20199.33877711650 *T)
    sumX2  +=       0.061 * cos(5.30466523023  + 4032.52621044309 *T)
    sumX2  +=       0.062 * cos(2.05245090503  + 27035.49639498329 *T)
    sumX2  +=       0.062 * cos(1.42815172261  + 5641.95442512570 *T)
    sumX2  +=       0.063 * cos(1.53293676835  + 12936.09533340670 *T)
    sumX2  +=       0.084 * cos(2.38574993037  + 1903.19299501769 *T)
    sumX2  +=       0.071 * cos(5.95376472970  + 11770.09751064990 *T)
    sumX2  +=       0.072 * cos(5.64693968962  + 13119.47728534169 *T)
    sumX2  +=       0.063 * cos(5.63682315335  + 13362.76083458550 *T)
    sumX2  +=       0.061 * cos(5.64246036249  + 17411.80811829210 *T)
    sumX2  +=       0.060 * cos(1.64524258736  + 3205.79116414790 *T)
    sumX2  +=       0.058 * cos(1.51670643508  + 7255.81346921790 *T)
    sumX2  +=       0.080 * cos(2.35306753256  + 21795.45790909830 *T)
    sumX2  +=       0.064 * cos(0.76110443806  + 6768.77687542291 *T)
    sumX2  +=       0.063 * cos(3.56114214131  + 3340.38586286849 *T)
    sumX2  +=       0.064 * cos(5.16926753358  + 3914.71340755109 *T)
    sumX2  +=       0.056 * cos(3.47086999510  + 27434.13303335850 *T)
    sumX2  +=       0.057 * cos(0.87381633431  + 12721.32828193349 *T)
    sumX2  +=       0.058 * cos(4.48269709673  + 3475.92132421871 *T)
    sumX2  +=       0.066 * cos(5.64545659289  + 14127.99895376209 *T)
    sumX2  +=       0.055 * cos(3.21479909652  + 4846.14405327630 *T)
    sumX2  +=       0.061 * cos(3.21825869339  + 5989.31106965630 *T)
    sumX2  +=       0.056 * cos(0.73623323904  + 16699.78283263350 *T)
    sumX2  +=       0.063 * cos(1.53292797875  + 3340.35135556410 *T)
    sumX2  +=       0.054 * cos(3.27785291788  + 2786.79920637390 *T)
    sumX2  +=       0.059 * cos(3.22300251676  + 7063.87756813929 *T)
    sumX2  +=       0.052 * cos(3.36345782490  + 11128.73226830069 *T)
    sumX2  +=       0.067 * cos(4.65085357599  + 2233.71743849851 *T)
    sumX2  +=       0.051 * cos(2.44574956886  + 30376.10882168309 *T)
    sumX2  +=       0.050 * cos(4.18360328454  + 17924.66688233689 *T)
    sumX2  +=       0.050 * cos(1.01622762006  + 17925.15451730390 *T)
    sumX2  +=       0.049 * cos(3.23997362723  + 14422.07545447190 *T)
    sumX2  +=       0.050 * cos(2.88028372970  + 6594.16046634330 *T)
    sumX2  +=       0.065 * cos(5.34147580845  + 6107.12387254830 *T)
    sumX2  +=       0.050 * cos(6.02043294912  + 4049.84604645990 *T)
    sumX2  +=       0.049 * cos(1.41928479603  + 14584.05445563709 *T)
    sumX2  +=       0.056 * cos(2.84568635986  + 6702.24406637550 *T)
    sumX2  +=       0.051 * cos(4.61875474436  + 8432.52056733209 *T)
    sumX2  +=       0.060 * cos(4.23711909745  + 22743.65319699990 *T)
    sumX2  +=       0.050 * cos(1.38036390033  + 426.35437339250 *T)
    sumX2  +=       0.056 * cos(5.63954798239  + 9499.50367948910 *T)
    sumX2  +=       0.048 * cos(1.32224646794  + 21265.76694400370 *T)
    sumX2  +=       0.045 * cos(5.84170955710  + 13149.39442884470 *T)
    sumX2  +=       0.055 * cos(6.04214625243  + 13362.62621397991 *T)
    sumX2  +=       0.047 * cos(2.92094264898  + 9381.69617630189 *T)
    sumX2  +=       0.051 * cos(0.76284557905  + 19514.22741258770 *T)
    sumX2  +=       0.047 * cos(2.03226039424  + 7321.36595719711 *T)
    sumX2  +=       0.048 * cos(2.85045948416  + 23141.80220040810 *T)
    sumX2  +=       0.062 * cos(5.07812840159  + 6041.08374960209 *T)
    sumX2  +=       0.056 * cos(5.76598331597  + 9595.48290670690 *T)
    sumX2  +=       0.045 * cos(5.50975189318  + 17256.38771885789 *T)
    sumX2  +=       0.043 * cos(6.27646091354  + 9485.27658548750 *T)
    sumX2  +=       0.055 * cos(1.28106527129  + 16061.94070863329 *T)
    sumX2  +=       0.042 * cos(3.07366795744  + 8800.23253126150 *T)
    sumX2  +=       0.044 * cos(5.51875626180  + 17256.87535382490 *T)
    sumX2  +=       0.055 * cos(1.15180858550  + 8425.89465529830 *T)
    sumX2  +=       0.043 * cos(2.56622033374  + 6705.34710863090 *T)
    sumX2  +=       0.042 * cos(1.19860454413  + 25287.96761688330 *T)
    sumX2  +=       0.050 * cos(2.38959203838  + 11773.62062899890 *T)
    sumX2  +=       0.046 * cos(1.39809566223  + 10973.31186886650 *T)
    sumX2  +=       0.042 * cos(2.01717790098  + 6685.05909953490 *T)
    sumX2  +=       0.043 * cos(6.18001810512  + 16460.08971204149 *T)
    sumX2  +=       0.042 * cos(4.51566406515  + 6158.89125278930 *T)
    sumX2  +=       0.043 * cos(1.46895638191  + 6709.91785835091 *T)
    sumX2  +=       0.040 * cos(3.81761467851  + 6531.90547374850 *T)
    sumX2  +=       0.050 * cos(0.14621919981  + 13892.38448937730 *T)
    sumX2  +=       0.039 * cos(2.35834767375  + 691.91378374330 *T)
    sumX2  +=       0.053 * cos(6.10905519875  + 3347.90248088131 *T)
    sumX2  +=       0.041 * cos(2.27487072554  + 12323.17927852529 *T)
    sumX2  +=       0.050 * cos(2.32783343015  + 18454.84548239850 *T)
    sumX2  +=       0.040 * cos(3.67259125733  + 13554.14179039430 *T)
    sumX2  +=       0.054 * cos(6.27305419537  + 3333.81000748530 *T)
    sumX2  +=       0.049 * cos(4.48594234475  + 6657.59023313531 *T)
    sumX2  +=       0.049 * cos(3.05979125217  + 11611.15399286670 *T)
    sumX2  +=       0.039 * cos(2.75953361592  + 5243.80542171749 *T)
    sumX2  +=       0.036 * cos(1.79555223900  + 19801.18977370830 *T)
    sumX2  +=       0.038 * cos(0.48175253779  + 8186.75647997610 *T)
    sumX2  +=       0.045 * cos(1.66666762132  + 5628.71265272770 *T)
    sumX2  +=       0.038 * cos(1.09319366612  + 11614.67711121570 *T)
    sumX2  +=       0.042 * cos(1.58323618910  + 6677.87824223130 *T)
    sumX2  +=       0.037 * cos(0.09890574681  + 8428.99744898310 *T)
    sumX2  +=       0.036 * cos(4.11399535323  + 17655.02435723310 *T)
    sumX2  +=       0.036 * cos(4.93312159758  + 19800.70213874130 *T)
    sumX2  +=       0.038 * cos(3.33015953455  + 22325.14887419290 *T)
    sumX2  +=       0.039 * cos(0.38152703835  + 3583.09721319030 *T)
    sumX2  +=       0.044 * cos(6.07368194572  + 25686.11662029150 *T)
    sumX2  +=       0.046 * cos(4.61289468261  + 4447.99504986810 *T)
    sumX2  +=       0.047 * cos(1.84502012984  + 8982.56685182550 *T)
    sumX2  +=       0.042 * cos(2.98971092818  + 6254.38284504009 *T)
    sumX2  +=       0.035 * cos(4.50369247266  + 18606.25512851669 *T)
    sumX2  +=       0.035 * cos(1.76189804108  + 5099.50932260011 *T)
    sumX2  +=       0.034 * cos(4.98411585506  + 5827.78465416409 *T)
    sumX2  +=       0.044 * cos(5.32026095241  + 7902.82960223750 *T)
    sumX2  +=       0.034 * cos(2.70658486935  + 15663.79170522509 *T)
    sumX2  +=       0.034 * cos(3.06682390120  + 28230.43104017490 *T)
    sumX2  +=       0.042 * cos(3.27968491376  + 13363.67384535091 *T)
    sumX2  +=       0.034 * cos(5.24068810869  + 3.27930086550 *T)
    sumX2  +=       0.038 * cos(5.33550411635  + 6843.93530701530 *T)
    sumX2  +=       0.033 * cos(4.24587603414  + 11243.44202893729 *T)
    sumX2  +=       0.033 * cos(1.58483385818  + 4778.03185838191 *T)
    sumX2  +=       0.035 * cos(4.68394755805  + 26084.26562369970 *T)
    sumX2  +=       0.042 * cos(3.47886429128  + 31571.04346687470 *T)
    sumX2  +=       0.034 * cos(2.69688154052  + 692.40141871030 *T)
    sumX2  +=       0.034 * cos(3.61044061282  + 3347.48215621710 *T)
    sumX2  +=       0.037 * cos(3.09345116660  + 3341.34893028450 *T)
    sumX2  +=       0.037 * cos(0.85261167200  + 3127.06951377830 *T)
    sumX2  +=       0.036 * cos(1.40223718038  + 24606.37937070350 *T)
    sumX2  +=       0.032 * cos(3.21575612349  + 8584.41784841729 *T)
    sumX2  +=       0.039 * cos(5.52528074071  + 7255.32583425090 *T)
    sumX2  +=       0.035 * cos(4.88108821778  + 3333.25506221550 *T)
    sumX2  +=       0.031 * cos(5.90785295838  + 10595.93826095070 *T)
    sumX2  +=       0.030 * cos(0.37640913202  + 20736.07597890910 *T)
    sumX2  +=       0.030 * cos(5.34266413467  + 10713.75106384269 *T)
    sumX2  +=       0.032 * cos(3.96275214977  + 1744.18216693170 *T)
    sumX2  +=       0.030 * cos(2.12718216703  + 3.76693583251 *T)
    sumX2  +=       0.036 * cos(5.85230454873  + 15113.74542073169 *T)
    sumX2  +=       0.030 * cos(0.02303147918  + 1222.09238380491 *T)
    sumX2  +=       0.031 * cos(2.44338692629  + 2575.06317973690 *T)
    sumX2  +=       0.031 * cos(0.53878733876  + 4106.64930862970 *T)
    sumX2  +=       0.029 * cos(6.26168191376  + 8827.14645239129 *T)
    sumX2  +=       0.030 * cos(4.97271935437  + 640.63378989870 *T)
    sumX2  +=       0.037 * cos(1.22493098074  + 14712.56093394150 *T)
    sumX2  +=       0.028 * cos(1.67581087356  + 12509.49714253070 *T)
    sumX2  +=       0.030 * cos(2.88869644863  + 6282.83203250789 *T)
    sumX2  +=       0.029 * cos(2.04175207609  + 16276.70776010650 *T)
    sumX2  +=       0.028 * cos(0.37303060773  + 8440.12174929990 *T)
    sumX2  +=       0.032 * cos(6.06171055220  + 11769.60987568290 *T)
    sumX2  +=       0.030 * cos(0.66860843732  + 24076.68840560890 *T)
    sumX2  +=       0.028 * cos(2.16704382280  + 15651.03756779411 *T)
    sumX2  +=       0.035 * cos(3.10903764323  + 19004.40413192490 *T)
    sumX2  +=       0.035 * cos(1.22211763949  + 14712.07329897449 *T)
    sumX2  +=       0.028 * cos(5.99561319996  + 3553.66770465429 *T)
    sumX2  +=       0.027 * cos(0.10827338046  + 17101.45495439070 *T)
    sumX2  +=       0.029 * cos(5.93234517199  + 27831.79440179969 *T)
    sumX2  +=       0.029 * cos(2.79075251840  + 27832.28203676670 *T)
    sumX2  +=       0.027 * cos(4.82458640917  + 31172.89446346650 *T)
    sumX2  +=       0.027 * cos(0.38322826891  + 3312.40705671550 *T)
    sumX2  +=       0.026 * cos(5.35970768977  + 7373.62627210991 *T)
    sumX2  +=       0.033 * cos(4.81267899723  + 5629.20028769470 *T)
    sumX2  +=       0.027 * cos(5.74554498508  + 3369.30543165110 *T)
    sumX2  +=       0.027 * cos(4.16679929541  + 15265.64270181689 *T)
    sumX2  +=       0.034 * cos(1.66093650866  + 28628.58004358310 *T)
    sumX2  +=       0.027 * cos(1.74573120715  + 19402.55313533309 *T)
    sumX2  +=       0.027 * cos(0.24189728224  + 29026.72904699130 *T)
    sumX2  +=       0.026 * cos(0.95064703918  + 16065.82204399129 *T)
    sumX2  +=       0.026 * cos(0.09426007080  + 21548.71855180830 *T)
    sumX2  +=       0.028 * cos(3.61885820364  + 3531.81687532789 *T)
    sumX2  +=       0.033 * cos(0.50594568985  + 3870.05957431090 *T)
    sumX2  +=       0.025 * cos(1.32229101061  + 12722.79623796870 *T)
    sumX2  +=       0.024 * cos(5.04445688037  + 22747.04689739089 *T)
    sumX2  +=       0.025 * cos(2.16587276266  + 2957.47207699310 *T)
    sumX2  +=       0.032 * cos(0.05221443708  + 12302.82395912650 *T)
    sumX2  +=       0.023 * cos(6.01606332444  + 20597.48778052470 *T)
    sumX2  +=       0.026 * cos(0.77786131261  + 3877.66075627870 *T)
    sumX2  +=       0.025 * cos(0.88197322398  + 27433.64539839149 *T)
    sumX2  +=       0.022 * cos(2.17540514796  + 17654.53672226610 *T)
    sumX2  +=       0.021 * cos(1.97613972525  + 5085.28222859850 *T)
    sumX2  +=       0.022 * cos(4.45819756524  + 13745.10242153889 *T)
    sumX2  +=       0.023 * cos(4.28492772247  + 3325.60377299830 *T)
    sumX2  +=       0.022 * cos(2.49953342580  + 10714.23869880971 *T)
    sumX2  +=       0.023 * cos(1.84356256708  + 3356.10871536831 *T)
    sumX2  +=       0.022 * cos(4.79335421406  + 13916.26292712510 *T)
    sumX2  +=       0.020 * cos(0.07454477807  + 13518.11392371690 *T)
    sumX2  +=       0.020 * cos(0.61605943900  + 14054.36349054249 *T)
    sumX2  +=       0.020 * cos(0.61605943900  + 14054.85112550950 *T)
    sumX2  +=       0.020 * cos(5.18934224543  + 14071.19569159229 *T)
    sumX2  +=       0.021 * cos(1.21172782919  + 20354.27154158369 *T)
    sumX2  +=       0.027 * cos(1.65038286221  + 23694.88396828350 *T)
    sumX2  +=       0.027 * cos(3.71231869252  + 11925.03027511709 *T)
    sumX2  +=       0.023 * cos(3.11201039786  + 1588.82907780029 *T)
    sumX2  +=       0.020 * cos(6.21683766974  + 2817.79119112250 *T)
    sumX2  +=       0.020 * cos(3.92465652923  + 29977.95981827489 *T)
    sumX2  +=       0.019 * cos(0.91394718054  + 27035.98402995030 *T)
    sumX2  +=       0.020 * cos(5.18320463527  + 9381.20349020070 *T)
    sumX2  +=       0.020 * cos(3.28496788336  + 1063.14886602171 *T)
    sumX2  +=       0.020 * cos(3.64150715316  + 36659.67230664150 *T)
    sumX2  +=       0.021 * cos(1.85047642439  + 10596.42589591770 *T)
    sumX2  +=       0.019 * cos(5.49838475903  + 3340.43591951909 *T)
    sumX2  +=       0.020 * cos(3.94260953954  + 34911.65589357450 *T)
    sumX2  +=       0.019 * cos(1.47273454627  + 23539.95120381630 *T)
    sumX2  +=       0.019 * cos(4.12041236337  + 40000.28473334130 *T)
    sumX2  +=       0.020 * cos(0.43561923981  + 2803.56409712089 *T)
    sumX2  +=       0.024 * cos(2.07709746581  + 9225.29545579949 *T)
    sumX2  +=       0.024 * cos(1.38288674852  + 30376.59645665010 *T)
    sumX2  +=       0.022 * cos(5.18167767699  + 537.04832957890 *T)
    sumX2  +=       0.019 * cos(6.22455531250  + 30774.74546005830 *T)
    sumX2  +=       0.024 * cos(4.96348676377  + 21946.86755521649 *T)
    sumX2  +=       0.020 * cos(5.27217035189  + 10787.38652706230 *T)
    sumX2  +=       0.018 * cos(2.50890520490  + 23938.10020722450 *T)
    sumX2  +=       0.018 * cos(2.22783415294  + 2487.65986243130 *T)
    sumX2  +=       0.018 * cos(1.88162325221  + 24491.18197509989 *T)
    sumX2  +=       0.021 * cos(2.34416498409  + 522.82123557730 *T)
    sumX2  +=       0.023 * cos(0.88923789924  + 37455.97031345790 *T)
    sumX2  +=       0.024 * cos(5.88546419101  + 3077.77232081051 *T)
    sumX2  +=       0.018 * cos(5.95267160454  + 18208.10612510850 *T)
    sumX2  +=       0.024 * cos(5.81264582562  + 4393.12462737170 *T)
    sumX2  +=       0.022 * cos(3.63700445171  + 13915.77529215809 *T)
    sumX2  +=       0.020 * cos(5.90812888983  + 3340.30129891350 *T)
    sumX2  +=       0.017 * cos(3.57017871636  + 22345.01655862470 *T)
    sumX2  +=       0.018 * cos(3.83484681339  + 3492.75352526850 *T)
    sumX2  +=       0.018 * cos(2.28892010525  + 3188.95896309811 *T)
    sumX2  +=       0.017 * cos(0.73986431664  + 17013.65911488389 *T)
    sumX2  +=       0.017 * cos(4.16857207200  + 28229.94340520789 *T)
    sumX2  +=       0.017 * cos(4.56595596791  + 8186.26884500910 *T)
    sumX2  +=       0.017 * cos(2.84716533138  + 5618.56362234490 *T)
    sumX2  +=       0.016 * cos(2.75060309814  + 29978.44745324190 *T)
    sumX2  +=       0.016 * cos(3.77900213334  + 10448.67928845890 *T)
    sumX2  +=       0.022 * cos(0.24435502443  + 3603.94016755610 *T)
    sumX2  +=       0.021 * cos(3.48477603406  + 26637.34739157509 *T)
    sumX2  +=       0.015 * cos(2.15480342032  + 11371.46087227470 *T)
    sumX2  +=       0.018 * cos(2.23032843213  + 18451.32236404950 *T)
    sumX2  +=       0.015 * cos(5.97010854409  + 23546.99744051430 *T)
    sumX2  +=       0.020 * cos(3.19548468220  + 33319.05987994170 *T)
    sumX2  +=       0.015 * cos(2.57021608840  + 17277.65074931730 *T)
    sumX2  +=       0.020 * cos(5.73090127884  + 20043.43074271529 *T)

    # MARS X Coord
    sumX3  = 0
    sumX3  +=    1905.181 * cos(0.55093544444  + 0.24381748350 *T)
    sumX3  +=     638.154 * cos(1.43815993842  + 6681.46867088311 *T)
    sumX3  +=     315.844 * cos(2.25892403607  + 10022.08109758290 *T)
    sumX3  +=     167.935 * cos(2.76973197145  + 3340.36860921629 *T)
    sumX3  +=      88.045 * cos(2.71691445968  + 13362.69352428270 *T)
    sumX3  +=      34.296 * cos(1.41715335205  + 3340.85624418330 *T)
    sumX3  +=      19.004 * cos(3.10474615459  + 16703.30595098250 *T)
    sumX3  +=      17.309 * cos(3.53485555448  + 3496.27664361750 *T)
    sumX3  +=      17.052 * cos(5.71088724992  + 3185.43584474911 *T)
    sumX3  +=      12.081 * cos(2.98816202142  + 6680.98103591609 *T)
    sumX3  +=       8.036 * cos(6.01800755163  + 3337.33312583431 *T)
    sumX3  +=       8.075 * cos(2.89215175615  + 3344.37936253230 *T)
    sumX3  +=       3.586 * cos(3.46976109002  + 20043.91837768230 *T)
    sumX3  +=       3.728 * cos(4.30810867123  + 3583.58484815730 *T)
    sumX3  +=       3.708 * cos(4.94088205435  + 3098.12764020930 *T)
    sumX3  +=       2.838 * cos(3.43415575124  + 6684.99178923210 *T)
    sumX3  +=       2.391 * cos(6.07237527440  + 6526.04827144891 *T)
    sumX3  +=       2.389 * cos(2.12960839543  + 3894.42564702570 *T)
    sumX3  +=       2.358 * cos(0.82395288049  + 2787.28684134090 *T)
    sumX3  +=       2.166 * cos(4.15770883257  + 155.17658195069 *T)
    sumX3  +=       1.411 * cos(3.32468261407  + 10021.59346261590 *T)
    sumX3  +=       1.300 * cos(0.72724281150  + 4292.57465043390 *T)
    sumX3  +=       1.268 * cos(2.20906408368  + 2389.13783793270 *T)
    sumX3  +=       1.168 * cos(1.89867328281  + 3149.40797807170 *T)
    sumX3  +=       1.170 * cos(0.77806649581  + 3532.30451029490 *T)
    sumX3  +=       1.149 * cos(1.82712058880  + 4137.15425099970 *T)
    sumX3  +=       1.000 * cos(2.35347706930  + 3339.87592311510 *T)
    sumX3  +=       1.053 * cos(0.86855974610  + 2544.55823736690 *T)
    sumX3  +=       0.909 * cos(5.60595748036  + 4690.72365384210 *T)
    sumX3  +=       0.997 * cos(0.51493534476  + 3341.83656525151 *T)
    sumX3  +=       0.874 * cos(3.57638185545  + 1990.98883452450 *T)
    sumX3  +=       0.773 * cos(3.83045600398  + 10025.60421593190 *T)
    sumX3  +=       0.800 * cos(3.22880764813  + 3723.75277640650 *T)
    sumX3  +=       0.785 * cos(5.96799151563  + 2957.95971196010 *T)
    sumX3  +=       0.624 * cos(3.82443513781  + 23384.53080438210 *T)
    sumX3  +=       0.586 * cos(5.53512020922  + 1059.62574767270 *T)
    sumX3  +=       0.721 * cos(0.50440664244  + 4535.30325440790 *T)
    sumX3  +=       0.636 * cos(4.17907998332  + 5088.87265725030 *T)
    sumX3  +=       0.615 * cos(2.33680720167  + 2146.40923395870 *T)
    sumX3  +=       0.474 * cos(5.37074646349  + 2942.70724077510 *T)
    sumX3  +=       0.604 * cos(4.91081510372  + 1592.83983111630 *T)
    sumX3  +=       0.450 * cos(1.89317206758  + 3333.74269718250 *T)
    sumX3  +=       0.486 * cos(2.92245527853  + 3739.00524759150 *T)
    sumX3  +=       0.417 * cos(1.04177275312  + 3347.96979118411 *T)
    sumX3  +=       0.400 * cos(0.45853342120  + 4400.23817437250 *T)
    sumX3  +=       0.374 * cos(3.38310907177  + 6836.88907031730 *T)
    sumX3  +=       0.488 * cos(4.66492474940  + 6924.19727485710 *T)
    sumX3  +=       0.399 * cos(2.43777775214  + 6490.02040477150 *T)
    sumX3  +=       0.356 * cos(1.18950177450  + 6127.89926804070 *T)
    sumX3  +=       0.362 * cos(2.14655068421  + 5622.08674069391 *T)
    sumX3  +=       0.295 * cos(3.68111290408  + 7079.61767429131 *T)
    sumX3  +=       0.356 * cos(0.14502929692  + 9866.66069814871 *T)
    sumX3  +=       0.340 * cos(2.28749442266  + 7477.76667769950 *T)
    sumX3  +=       0.239 * cos(5.17898361533  + 6438.74006690910 *T)
    sumX3  +=       0.214 * cos(0.95650348151  + 6682.44899195131 *T)
    sumX3  +=       0.184 * cos(3.67133608119  + 13362.20588931570 *T)
    sumX3  +=       0.176 * cos(3.96314611309  + 5331.60126122430 *T)
    sumX3  +=       0.215 * cos(2.58410594073  + 5729.75026463250 *T)
    sumX3  +=       0.159 * cos(5.31902457668  + 4933.45225781610 *T)
    sumX3  +=       0.148 * cos(0.07133203651  + 6298.57213865991 *T)
    sumX3  +=       0.154 * cos(5.64200281325  + 6677.94555253410 *T)
    sumX3  +=       0.153 * cos(0.89551511976  + 7875.91568110771 *T)
    sumX3  +=       0.157 * cos(6.03012929865  + 796.54182429990 *T)
    sumX3  +=       0.172 * cos(4.18899228750  + 13366.21664263171 *T)
    sumX3  +=       0.143 * cos(2.39578247400  + 7235.03807372550 *T)
    sumX3  +=       0.155 * cos(0.96938893909  + 7740.85060107230 *T)
    sumX3  +=       0.116 * cos(4.64918761388  + 1194.69082770810 *T)
    sumX3  +=       0.106 * cos(3.00887462201  + 155.66421691770 *T)
    sumX3  +=       0.089 * cos(4.02630363689  + 3343.89172756530 *T)
    sumX3  +=       0.100 * cos(2.87689937674  + 9830.63283147130 *T)
    sumX3  +=       0.108 * cos(4.17718423869  + 10420.23010099111 *T)
    sumX3  +=       0.084 * cos(2.68446580877  + 10818.37910439930 *T)
    sumX3  +=       0.098 * cos(4.17525996193  + 26725.14323108190 *T)
    sumX3  +=       0.076 * cos(0.87920756584  + 4929.92913946711 *T)
    sumX3  +=       0.074 * cos(1.00299622836  + 7633.18707713370 *T)
    sumX3  +=       0.086 * cos(5.15622842132  + 1589.31671276730 *T)
    sumX3  +=       0.072 * cos(6.20405951094  + 5885.17066406670 *T)
    sumX3  +=       0.062 * cos(1.68960801507  + 5487.02166065850 *T)
    sumX3  +=       0.065 * cos(5.02060335506  + 10264.80970155691 *T)
    sumX3  +=       0.062 * cos(3.09389208181  + 7211.15963597770 *T)
    sumX3  +=       0.082 * cos(1.44324153466  + 6688.58221788390 *T)
    sumX3  +=       0.057 * cos(4.66555911574  + 12833.00255918811 *T)
    sumX3  +=       0.057 * cos(1.54372665386  + 9468.51169474051 *T)
    sumX3  +=       0.056 * cos(5.13139205932  + 6681.45141723091 *T)
    sumX3  +=       0.056 * cos(0.87897817015  + 6681.48592453530 *T)
    sumX3  +=       0.069 * cos(2.58172462044  + 796.05418933290 *T)
    sumX3  +=       0.065 * cos(1.91690600798  + 3870.54720927791 *T)
    sumX3  +=       0.067 * cos(3.74829679592  + 9492.39013248830 *T)
    sumX3  +=       0.051 * cos(5.79032862137  + 8274.06468451590 *T)
    sumX3  +=       0.048 * cos(3.88765170229  + 397.90518592470 *T)
    sumX3  +=       0.050 * cos(0.50123821286  + 13207.27312484851 *T)
    sumX3  +=       0.043 * cos(5.49155036342  + 426.84200835950 *T)
    sumX3  +=       0.044 * cos(5.87908345436  + 8031.33608054190 *T)
    sumX3  +=       0.039 * cos(1.20945071946  + 1194.20319274110 *T)
    sumX3  +=       0.043 * cos(1.35709156628  + 10023.06141865111 *T)
    sumX3  +=       0.044 * cos(0.94709794694  + 5092.39577559930 *T)
    sumX3  +=       0.037 * cos(0.18533157058  + 8827.63408735830 *T)
    sumX3  +=       0.042 * cos(1.70772874502  + 10022.09835123510 *T)
    sumX3  +=       0.043 * cos(0.30947981672  + 3184.94820978210 *T)
    sumX3  +=       0.033 * cos(2.40776286352  + 1349.62359217529 *T)
    sumX3  +=       0.035 * cos(2.95444208499  + 9070.36269133230 *T)
    sumX3  +=       0.036 * cos(1.36035547703  + 11081.46302777211 *T)
    sumX3  +=       0.044 * cos(0.12174745615  + 12303.31159409350 *T)
    sumX3  +=       0.033 * cos(0.88839870426  + 15643.92402079330 *T)
    sumX3  +=       0.042 * cos(5.96028045375  + 10022.06384393071 *T)
    sumX3  +=       0.032 * cos(5.66439939781  + 10018.55797923390 *T)
    sumX3  +=       0.029 * cos(1.54951760073  + 3495.78900865049 *T)
    sumX3  +=       0.032 * cos(2.40303298785  + 6680.48834981490 *T)
    sumX3  +=       0.031 * cos(1.90739176085  + 7.35736448430 *T)
    sumX3  +=       0.036 * cos(3.46372584197  + 7064.36520310630 *T)
    sumX3  +=       0.026 * cos(1.27046456409  + 11216.52810780750 *T)
    sumX3  +=       0.035 * cos(4.70034992379  + 3336.84549086730 *T)
    sumX3  +=       0.025 * cos(3.03604811102  + 10575.65050042530 *T)
    sumX3  +=       0.033 * cos(4.36445511764  + 8672.21368792411 *T)
    sumX3  +=       0.026 * cos(5.99843166178  + 1592.35219614930 *T)
    sumX3  +=       0.027 * cos(3.98200757587  + 16702.81831601549 *T)
    sumX3  +=       0.025 * cos(4.57948088729  + 13760.84252769090 *T)
    sumX3  +=       0.027 * cos(3.09776670837  + 191.20444862810 *T)
    sumX3  +=       0.026 * cos(3.54518463791  + 6894.76776632111 *T)
    sumX3  +=       0.023 * cos(5.85147228213  + 2281.47431399410 *T)
    sumX3  +=       0.027 * cos(4.51774923051  + 16706.82906933150 *T)
    sumX3  +=       0.026 * cos(5.53508596353  + 8433.00820229911 *T)
    sumX3  +=       0.028 * cos(3.36273080390  + 9623.93209417470 *T)
    sumX3  +=       0.027 * cos(5.85014917702  + 2145.92159899169 *T)
    sumX3  +=       0.024 * cos(1.25782822684  + 8270.54156616690 *T)
    sumX3  +=       0.026 * cos(1.08150011094  + 1747.77259558350 *T)
    sumX3  +=       0.022 * cos(0.90453576854  + 3767.45443505930 *T)
    sumX3  +=       0.025 * cos(3.82100205156  + 2811.16527908870 *T)
    sumX3  +=       0.020 * cos(1.64178140869  + 3554.15533962130 *T)
    sumX3  +=       0.019 * cos(4.46076005938  + 3127.55714874530 *T)
    sumX3  +=       0.022 * cos(5.17308651919  + 553.32558535889 *T)
    sumX3  +=       0.023 * cos(3.76515587292  + 951.47458876710 *T)
    sumX3  +=       0.019 * cos(1.67275856465  + 553.81322032590 *T)
    sumX3  +=       0.026 * cos(0.16277946818  + 8962.69916739370 *T)
    sumX3  +=       0.020 * cos(0.43011448080  + 9639.18456535971 *T)
    sumX3  +=       0.019 * cos(4.43079189104  + 2810.67764412170 *T)
    sumX3  +=       0.021 * cos(0.21950833051  + 951.96222373410 *T)
    sumX3  +=       0.017 * cos(3.25327411593  + 13171.24525817110 *T)
    sumX3  +=       0.022 * cos(4.38492748210  + 3738.51761262449 *T)
    sumX3  +=       0.019 * cos(3.21726717973  + 6681.40136058031 *T)
    sumX3  +=       0.022 * cos(1.65280044894  + 10973.79950383350 *T)
    sumX3  +=       0.016 * cos(2.94665873224  + 4136.66661603269 *T)
    sumX3  +=       0.019 * cos(2.80877068966  + 6681.53598118590 *T)
    sumX3  +=       0.021 * cos(2.92855757827  + 1748.26023055050 *T)
    sumX3  +=       0.017 * cos(5.71391599259  + 382.65271473970 *T)
    sumX3  +=       0.020 * cos(0.18316423734  + 2280.98667902709 *T)
    sumX3  +=       0.022 * cos(4.74779376737  + 1350.11122714230 *T)
    sumX3  +=       0.016 * cos(0.63462673047  + 6254.87048000710 *T)
    sumX3  +=       0.016 * cos(3.93662181429  + 10551.77206267750 *T)
    sumX3  +=       0.015 * cos(3.10357108475  + 10177.50149701710 *T)
    sumX3  +=       0.014 * cos(5.09381800673  + 2914.25805330730 *T)
    sumX3  +=       0.014 * cos(1.36653791314  + 5459.62010456170 *T)
    sumX3  +=       0.019 * cos(4.60528090517  + 1990.50119955750 *T)
    sumX3  +=       0.014 * cos(1.82944391532  + 10029.19464458370 *T)
    sumX3  +=       0.013 * cos(4.04119501897  + 5614.97319369310 *T)
    sumX3  +=       0.015 * cos(2.66542384043  + 6674.35512388230 *T)
    sumX3  +=       0.013 * cos(1.51732945573  + 1066.73929467350 *T)
    sumX3  +=       0.016 * cos(4.23161759812  + 640.14110379750 *T)
    sumX3  +=       0.013 * cos(3.65983424195  + 5621.59910572690 *T)
    sumX3  +=       0.013 * cos(1.47843905604  + 242.97242145751 *T)
    sumX3  +=       0.014 * cos(0.24731425362  + 12310.42514109431 *T)
    sumX3  +=       0.013 * cos(0.42091404951  + 191.69208359510 *T)
    sumX3  +=       0.011 * cos(3.19331361004  + 2388.65020296570 *T)
    sumX3  +=       0.012 * cos(3.12268751453  + 3148.92034310470 *T)
    sumX3  +=       0.013 * cos(5.43821523436  + 2119.00767786191 *T)
    sumX3  +=       0.011 * cos(5.19736362540  + 3583.09721319030 *T)
    sumX3  +=       0.011 * cos(1.99832694470  + 3340.38586286849 *T)
    sumX3  +=       0.010 * cos(1.49714805225  + 4534.81561944089 *T)
    sumX3  +=       0.011 * cos(6.25129481290  + 3340.35135556410 *T)
    sumX3  +=       0.010 * cos(4.88853298993  + 8429.48508395010 *T)
    sumX3  +=       0.010 * cos(3.31545628452  + 10014.96755058210 *T)
    sumX3  +=       0.009 * cos(3.88123559083  + 3320.50092478451 *T)
    sumX3  +=       0.011 * cos(4.66037741803  + 8962.21153242670 *T)
    sumX3  +=       0.009 * cos(2.24701790497  + 3361.21156358210 *T)
    sumX3  +=       0.009 * cos(4.10537369930  + 12296.19804709270 *T)
    sumX3  +=       0.010 * cos(5.06205358587  + 16173.61498588790 *T)
    sumX3  +=       0.009 * cos(2.02386810043  + 1052.51220067191 *T)
    sumX3  +=       0.009 * cos(3.68063634198  + 5614.48555872609 *T)
    sumX3  +=       0.010 * cos(1.99400297925  + 529.93478257810 *T)
    sumX3  +=       0.008 * cos(6.05827415687  + 3893.93801205869 *T)
    sumX3  +=       0.009 * cos(5.82204012446  + 3325.60377299830 *T)
    sumX3  +=       0.008 * cos(1.82638763774  + 2786.79920637390 *T)
    sumX3  +=       0.009 * cos(0.30647315712  + 3356.10871536831 *T)
    sumX3  +=       0.011 * cos(4.85651046171  + 242.48478649050 *T)
    sumX3  +=       0.009 * cos(1.42556082626  + 1751.78334889950 *T)
    sumX3  +=       0.007 * cos(0.31910199684  + 2957.47207699310 *T)
    sumX3  +=       0.008 * cos(4.58643573119  + 4292.08701546690 *T)
    sumX3  +=       0.007 * cos(3.02875762795  + 14158.99153109910 *T)
    sumX3  +=       0.010 * cos(3.75880088864  + 529.44714761109 *T)
    sumX3  +=       0.010 * cos(4.75162534448  + 12302.82395912650 *T)
    sumX3  +=       0.009 * cos(4.40822524545  + 6684.50415426510 *T)
    sumX3  +=       0.008 * cos(5.21609653228  + 2273.87313202630 *T)

    # MARS X Coord
    sumX4  = 0
    sumX4  +=     174.339 * cos(4.50461485471  + 3340.85624418330 *T)
    sumX4  +=      65.332 * cos(1.45277537677  + 0.24381748350 *T)
    sumX4  +=      21.612 * cos(5.47577416210  + 6681.46867088311 *T)
    sumX4  +=      12.491 * cos(0.51687441739  + 10022.08109758290 *T)
    sumX4  +=       5.152 * cos(1.11384612445  + 13362.69352428270 *T)
    sumX4  +=       2.432 * cos(5.03376242803  + 3496.27664361750 *T)
    sumX4  +=       2.407 * cos(4.19068604389  + 3185.43584474911 *T)
    sumX4  +=       1.485 * cos(1.55324524407  + 16703.30595098250 *T)
    sumX4  +=       1.453 * cos(1.17593053226  + 3340.36860921629 *T)
    sumX4  +=       0.592 * cos(2.79217502706  + 3583.58484815730 *T)
    sumX4  +=       0.596 * cos(0.15627931299  + 3098.12764020930 *T)
    sumX4  +=       0.649 * cos(1.53914017168  + 6680.98103591609 *T)
    sumX4  +=       0.380 * cos(4.55600229997  + 6526.04827144891 *T)
    sumX4  +=       0.351 * cos(1.94518064214  + 20043.91837768230 *T)
    sumX4  +=       0.281 * cos(1.02784973125  + 3344.37936253230 *T)
    sumX4  +=       0.275 * cos(1.11119927155  + 3337.33312583431 *T)
    sumX4  +=       0.287 * cos(3.61921656726  + 3894.42564702570 *T)
    sumX4  +=       0.285 * cos(5.58839678189  + 2787.28684134090 *T)
    sumX4  +=       0.151 * cos(1.78910865682  + 6684.99178923210 *T)
    sumX4  +=       0.128 * cos(5.24790443912  + 6836.88907031730 *T)
    sumX4  +=       0.111 * cos(1.85011350887  + 10021.59346261590 *T)
    sumX4  +=       0.088 * cos(3.15545951202  + 6924.19727485710 *T)
    sumX4  +=       0.071 * cos(2.31380588062  + 23384.53080438210 *T)
    sumX4  +=       0.055 * cos(2.26852590135  + 10025.60421593190 *T)
    sumX4  +=       0.048 * cos(5.95592456117  + 6127.89926804070 *T)
    sumX4  +=       0.063 * cos(4.91767924460  + 9866.66069814871 *T)
    sumX4  +=       0.034 * cos(0.35023274977  + 6438.74006690910 *T)
    sumX4  +=       0.033 * cos(3.27139275821  + 1059.62574767270 *T)
    sumX4  +=       0.019 * cos(2.14803379863  + 13362.20588931570 *T)
    sumX4  +=       0.018 * cos(0.78782665763  + 6490.02040477150 *T)
    sumX4  +=       0.016 * cos(5.21278468290  + 3149.40797807170 *T)
    sumX4  +=       0.016 * cos(5.81489292910  + 2942.70724077510 *T)
    sumX4  +=       0.015 * cos(0.86980577909  + 3532.30451029490 *T)
    sumX4  +=       0.015 * cos(1.05784197172  + 2544.55823736690 *T)
    sumX4  +=       0.014 * cos(2.75650621307  + 5622.08674069391 *T)
    sumX4  +=       0.014 * cos(0.20093505699  + 3739.00524759150 *T)
    sumX4  +=       0.013 * cos(4.83578063472  + 4400.23817437250 *T)
    sumX4  +=       0.013 * cos(5.02472221311  + 4137.15425099970 *T)
    sumX4  +=       0.012 * cos(3.15784569291  + 1194.69082770810 *T)
    sumX4  +=       0.012 * cos(5.66215318851  + 6682.44899195131 *T)
    sumX4  +=       0.015 * cos(0.65316905429  + 7477.76667769950 *T)
    sumX4  +=       0.011 * cos(1.92062639148  + 7079.61767429131 *T)
    sumX4  +=       0.015 * cos(5.27033110100  + 529.93478257810 *T)
    sumX4  +=       0.013 * cos(5.72126392176  + 10177.50149701710 *T)
    sumX4  +=       0.014 * cos(3.84659935898  + 7235.03807372550 *T)
    sumX4  +=       0.011 * cos(3.51933198406  + 10264.80970155691 *T)
    sumX4  +=       0.013 * cos(2.66285978116  + 13366.21664263171 *T)
    sumX4  +=       0.010 * cos(4.52594962497  + 796.54182429990 *T)
    sumX4  +=       0.010 * cos(2.37075551462  + 5331.60126122430 *T)
    sumX4  +=       0.012 * cos(1.03827349416  + 5729.75026463250 *T)
    sumX4  +=       0.011 * cos(2.64809487107  + 26725.14323108190 *T)
    sumX4  +=       0.008 * cos(4.81245299131  + 6298.57213865991 *T)
    sumX4  +=       0.008 * cos(3.95118300322  + 3340.87349783550 *T)
    sumX4  +=       0.008 * cos(1.92041137172  + 3340.83899053111 *T)
    sumX4  +=       0.007 * cos(0.43533123258  + 1990.98883452450 *T)
    sumX4  +=       0.009 * cos(3.50915514440  + 2280.98667902709 *T)
    sumX4  +=       0.007 * cos(5.60821148288  + 7875.91568110771 *T)
    sumX4  +=       0.008 * cos(3.59280368451  + 4933.45225781610 *T)
    sumX4  +=       0.007 * cos(5.32208951844  + 2389.13783793270 *T)
    sumX4  +=       0.009 * cos(1.82497486389  + 1592.83983111630 *T)
    sumX4  +=       0.007 * cos(2.39846706486  + 3339.87592311510 *T)
    sumX4  +=       0.007 * cos(3.71269284859  + 3341.83656525151 *T)
    sumX4  +=       0.007 * cos(2.58444553246  + 3343.89172756530 *T)
    sumX4  +=       0.006 * cos(0.55736215127  + 6151.77770578850 *T)
    sumX4  +=       0.007 * cos(5.10772331644  + 3184.94820978210 *T)
    sumX4  +=       0.007 * cos(1.05272897981  + 155.17658195069 *T)
    sumX4  +=       0.006 * cos(3.65850030907  + 1589.31671276730 *T)
    sumX4  +=       0.005 * cos(2.88251144144  + 5487.02166065850 *T)
    sumX4  +=       0.005 * cos(1.46784159994  + 5885.17066406670 *T)
    sumX4  +=       0.005 * cos(0.80161683332  + 4292.57465043390 *T)
    sumX4  +=       0.006 * cos(2.45234336003  + 2146.40923395870 *T)
    sumX4  +=       0.005 * cos(5.68329332382  + 4690.72365384210 *T)
    sumX4  +=       0.005 * cos(2.80269016526  + 2957.95971196010 *T)
    sumX4  +=       0.004 * cos(5.28325791601  + 13207.27312484851 *T)
    sumX4  +=       0.005 * cos(3.64094054061  + 4535.30325440790 *T)
    sumX4  +=       0.004 * cos(3.31472278070  + 3723.75277640650 *T)
    sumX4  +=       0.005 * cos(4.28302336857  + 5088.87265725030 *T)
    sumX4  +=       0.004 * cos(0.05439896723  + 9468.51169474051 *T)
    sumX4  +=       0.004 * cos(5.79414027536  + 398.39282089170 *T)
    sumX4  +=       0.003 * cos(6.01492767118  + 3870.54720927791 *T)
    sumX4  +=       0.003 * cos(5.81270604686  + 7633.18707713370 *T)

    # MARS X Coord
    sumX5  = 0
    sumX5  +=       2.822 * cos(2.53736821761  + 0.24381748350 *T)
    sumX5  +=       1.310 * cos(4.62457430290  + 3340.85624418330 *T)
    sumX5  +=       0.954 * cos(3.55909372543  + 6681.46867088311 *T)
    sumX5  +=       0.429 * cos(4.88697356497  + 10022.08109758290 *T)
    sumX5  +=       0.275 * cos(0.24510755091  + 3496.27664361750 *T)
    sumX5  +=       0.272 * cos(2.66437498180  + 3185.43584474911 *T)
    sumX5  +=       0.241 * cos(5.73070473130  + 13362.69352428270 *T)
    sumX5  +=       0.262 * cos(6.24547382421  + 3340.36860921629 *T)
    sumX5  +=       0.092 * cos(6.24977577619  + 16703.30595098250 *T)
    sumX5  +=       0.076 * cos(1.27309430645  + 3583.58484815730 *T)
    sumX5  +=       0.077 * cos(1.65718815142  + 3098.12764020930 *T)
    sumX5  +=       0.049 * cos(3.03838424904  + 6526.04827144891 *T)
    sumX5  +=       0.032 * cos(0.11875540710  + 6680.98103591609 *T)
    sumX5  +=       0.026 * cos(0.39487275836  + 20043.91837768230 *T)
    sumX5  +=       0.027 * cos(5.12576113855  + 3894.42564702570 *T)
    sumX5  +=       0.026 * cos(4.04351110154  + 2787.28684134090 *T)
    sumX5  +=       0.013 * cos(0.43187059389  + 6836.88907031730 *T)
    sumX5  +=       0.013 * cos(1.64004823919  + 6924.19727485710 *T)
    sumX5  +=       0.006 * cos(0.36382978226  + 10021.59346261590 *T)
    sumX5  +=       0.007 * cos(1.40323245136  + 3337.33312583431 *T)
    sumX5  +=       0.007 * cos(4.61746707728  + 3344.37936253230 *T)
    sumX5  +=       0.005 * cos(4.43055095713  + 6127.89926804070 *T)
    sumX5  +=       0.007 * cos(3.39474539876  + 9866.66069814871 *T)
    sumX5  +=       0.006 * cos(0.04415481578  + 6684.99178923210 *T)
    sumX5  +=       0.005 * cos(0.78203649698  + 23384.53080438210 *T)
    sumX5  +=       0.004 * cos(1.85425114024  + 6438.74006690910 *T)

    # MARS Y Coord
    sumY0  = 0
    sumY0  +=  151664432.758 * cos(4.63267999004  + 3340.85624418330 *T)
    sumY0  +=  21337343.470 * cos(1.15310710021  + 0.24381748350 *T)
    sumY0  +=  7067734.657 * cos(4.97080577669  + 6681.46867088311 *T)
    sumY0  +=  494034.875 * cos(5.30892921206  + 10022.08109758290 *T)
    sumY0  +=  135189.385 * cos(2.16683537182  + 3340.36860921629 *T)
    sumY0  +=   40927.830 * cos(5.64705236353  + 13362.69352428270 *T)
    sumY0  +=   21051.988 * cos(0.23298189024  + 3337.33312583431 *T)
    sumY0  +=   21027.270 * cos(5.89078683719  + 3344.37936253230 *T)
    sumY0  +=   11370.204 * cos(3.26158494021  + 1059.62574767270 *T)
    sumY0  +=   13426.076 * cos(5.34254489555  + 529.93478257810 *T)
    sumY0  +=    6769.403 * cos(2.04649779939  + 3340.83899053111 *T)
    sumY0  +=    6769.355 * cos(4.07726943088  + 3340.87349783550 *T)
    sumY0  +=    8285.925 * cos(3.43188641054  + 2280.98667902709 *T)
    sumY0  +=    5433.910 * cos(5.83045260147  + 2942.70724077510 *T)
    sumY0  +=    4813.330 * cos(0.27943415467  + 3739.00524759150 *T)
    sumY0  +=    4893.493 * cos(2.85680765111  + 5622.08674069391 *T)
    sumY0  +=    5269.759 * cos(0.75939928833  + 6151.77770578850 *T)
    sumY0  +=    3774.767 * cos(2.39590149517  + 6680.98103591609 *T)
    sumY0  +=    3623.045 * cos(4.53783234193  + 796.54182429990 *T)
    sumY0  +=    3725.058 * cos(5.98517547467  + 16703.30595098250 *T)
    sumY0  +=    3767.289 * cos(5.93443742300  + 398.39282089170 *T)
    sumY0  +=    2396.177 * cos(5.34735734199  + 3149.40797807170 *T)
    sumY0  +=    2262.756 * cos(0.77787856455  + 3532.30451029490 *T)
    sumY0  +=    1976.674 * cos(0.56979018369  + 6677.94555253410 *T)
    sumY0  +=    2227.632 * cos(0.12453915456  + 3340.78893388051 *T)
    sumY0  +=    2227.573 * cos(5.99922807490  + 3340.92355448610 *T)
    sumY0  +=    1950.645 * cos(1.15729484520  + 2544.55823736690 *T)
    sumY0  +=    2179.897 * cos(0.12455085977  + 6283.31966747490 *T)
    sumY0  +=    1684.046 * cos(1.57848903772  + 5885.17066406670 *T)
    sumY0  +=    2237.564 * cos(3.25024245156  + 8962.69916739370 *T)
    sumY0  +=    1629.438 * cos(4.95303923805  + 4137.15425099970 *T)
    sumY0  +=    1957.286 * cos(6.23124655952  + 6684.99178923210 *T)
    sumY0  +=    1382.161 * cos(3.74262124968  + 1751.29571393250 *T)
    sumY0  +=    1364.523 * cos(5.17298469073  + 2811.16527908870 *T)
    sumY0  +=    1285.835 * cos(3.13464527657  + 1194.69082770810 *T)
    sumY0  +=    1468.951 * cos(0.30602473382  + 3870.54720927791 *T)
    sumY0  +=    1572.210 * cos(3.27704899210  + 1589.31671276730 *T)
    sumY0  +=     955.879 * cos(3.79979264711  + 426.84200835950 *T)
    sumY0  +=     819.304 * cos(3.58795489460  + 4400.23817437250 *T)
    sumY0  +=     714.429 * cos(4.49383093867  + 191.20444862810 *T)
    sumY0  +=     719.663 * cos(1.08407118449  + 9492.39013248830 *T)
    sumY0  +=     643.260 * cos(2.98709791321  + 5487.02166065850 *T)
    sumY0  +=     795.172 * cos(2.82580884918  + 5092.39577559930 *T)
    sumY0  +=     630.906 * cos(2.38462144554  + 6681.45141723091 *T)
    sumY0  +=     630.906 * cos(4.41539307703  + 6681.48592453530 *T)
    sumY0  +=     782.173 * cos(2.73881042817  + 5091.90814063229 *T)
    sumY0  +=     703.756 * cos(5.67896443651  + 213.54291292150 *T)
    sumY0  +=     581.847 * cos(2.38152204086  + 3185.43584474911 *T)
    sumY0  +=     553.454 * cos(3.74330200173  + 3496.27664361750 *T)
    sumY0  +=     663.460 * cos(5.31434426935  + 2544.07060239989 *T)
    sumY0  +=     491.963 * cos(0.60494800834  + 7079.61767429131 *T)
    sumY0  +=     426.327 * cos(0.43687696114  + 2145.92159899169 *T)
    sumY0  +=     533.910 * cos(1.72930733886  + 1592.83983111630 *T)
    sumY0  +=     386.953 * cos(3.50258018070  + 8433.00820229911 *T)
    sumY0  +=     417.478 * cos(2.39346049510  + 3339.87592311510 *T)
    sumY0  +=     417.367 * cos(3.73029667099  + 3341.83656525151 *T)
    sumY0  +=     359.961 * cos(0.04010115133  + 20043.91837768230 *T)
    sumY0  +=     352.641 * cos(2.31813890735  + 6254.87048000710 *T)
    sumY0  +=     351.819 * cos(3.59912427222  + 12303.31159409350 *T)
    sumY0  +=     321.320 * cos(0.37355501133  + 3554.15533962130 *T)
    sumY0  +=     377.309 * cos(2.56396696963  + 2146.40923395870 *T)
    sumY0  +=     307.274 * cos(5.65660829568  + 3127.55714874530 *T)
    sumY0  +=     284.287 * cos(0.46340069524  + 7.35736448430 *T)
    sumY0  +=     259.684 * cos(0.32305651983  + 1990.98883452450 *T)
    sumY0  +=     246.886 * cos(1.84243729588  + 1747.77259558350 *T)
    sumY0  +=     313.688 * cos(3.53520341472  + 4535.30325440790 *T)
    sumY0  +=     271.483 * cos(0.37698651521  + 6468.16957544510 *T)
    sumY0  +=     298.122 * cos(4.39537363917  + 5088.87265725030 *T)
    sumY0  +=     223.194 * cos(1.98204681533  + 3320.08084869090 *T)
    sumY0  +=     221.461 * cos(4.14148614982  + 3361.63163967570 *T)
    sumY0  +=     233.899 * cos(0.46609609268  + 9623.93209417470 *T)
    sumY0  +=     207.617 * cos(0.46266279560  + 6681.40136058031 *T)
    sumY0  +=     205.148 * cos(0.28695974581  + 10025.60421593190 *T)
    sumY0  +=     207.784 * cos(0.90748162445  + 10018.55797923390 *T)
    sumY0  +=     193.198 * cos(3.36114050168  + 2119.00767786191 *T)
    sumY0  +=     218.630 * cos(5.68765119081  + 6490.02040477150 *T)
    sumY0  +=     247.990 * cos(2.88872133203  + 2913.77041834030 *T)
    sumY0  +=     207.617 * cos(0.05416640875  + 6681.53598118590 *T)
    sumY0  +=     178.587 * cos(4.11200624291  + 1221.60474883790 *T)
    sumY0  +=     178.400 * cos(2.59904254913  + 10021.59346261590 *T)
    sumY0  +=     164.293 * cos(5.80289640537  + 4690.72365384210 *T)
    sumY0  +=     157.091 * cos(5.19939410162  + 2389.13783793270 *T)
    sumY0  +=     151.308 * cos(3.00360651428  + 4562.21717553770 *T)
    sumY0  +=     177.027 * cos(2.69928080706  + 2957.95971196010 *T)
    sumY0  +=     147.685 * cos(0.90426952883  + 7211.15963597770 *T)
    sumY0  +=     164.646 * cos(5.28805943773  + 7477.76667769950 *T)
    sumY0  +=     153.884 * cos(2.87447082637  + 640.14110379750 *T)
    sumY0  +=     157.989 * cos(3.42539314748  + 3723.75277640650 *T)
    sumY0  +=     146.757 * cos(2.84466358271  + 3337.26581553151 *T)
    sumY0  +=     166.434 * cos(1.59845879272  + 10213.52936369450 *T)
    sumY0  +=     146.132 * cos(3.27579621025  + 3344.44667283510 *T)
    sumY0  +=     118.082 * cos(2.72149356676  + 4562.70481050471 *T)
    sumY0  +=     119.770 * cos(3.08805027302  + 11773.62062899890 *T)
    sumY0  +=     113.695 * cos(0.92730568398  + 4292.57465043390 *T)
    sumY0  +=     143.954 * cos(3.79112207293  + 2787.28684134090 *T)
    sumY0  +=     113.828 * cos(1.92140197091  + 3333.74269718250 *T)
    sumY0  +=     112.131 * cos(4.78443317220  + 3205.79116414790 *T)
    sumY0  +=     144.684 * cos(3.24900405626  + 1349.62359217529 *T)
    sumY0  +=     118.108 * cos(3.61481367969  + 4929.92913946711 *T)
    sumY0  +=     106.690 * cos(1.33869953850  + 3475.92132421871 *T)
    sumY0  +=     120.237 * cos(1.83849834021  + 9225.78309076650 *T)
    sumY0  +=     119.599 * cos(2.33508766672  + 3894.42564702570 *T)
    sumY0  +=     104.910 * cos(0.80738994994  + 382.65271473970 *T)
    sumY0  +=      95.920 * cos(4.24460901134  + 3347.96979118411 *T)
    sumY0  +=      93.778 * cos(3.97176912203  + 1748.26023055050 *T)
    sumY0  +=      88.754 * cos(1.41602519781  + 12833.00255918811 *T)
    sumY0  +=     112.794 * cos(1.18558937662  + 155.17658195069 *T)
    sumY0  +=      91.596 * cos(4.65533652150  + 951.47458876710 *T)
    sumY0  +=      86.667 * cos(5.15042860741  + 3098.12764020930 *T)
    sumY0  +=     118.500 * cos(2.09460100184  + 7902.82960223750 *T)
    sumY0  +=      89.906 * cos(3.86574614268  + 2942.21960580809 *T)
    sumY0  +=      85.205 * cos(2.64363572803  + 2281.47431399410 *T)
    sumY0  +=     113.660 * cos(0.87318331181  + 5486.53402569149 *T)
    sumY0  +=      80.640 * cos(0.97118273790  + 3583.58484815730 *T)
    sumY0  +=      85.344 * cos(2.11418392564  + 4933.45225781610 *T)
    sumY0  +=      84.447 * cos(5.69213107859  + 6674.35512388230 *T)
    sumY0  +=      79.513 * cos(2.28036031190  + 5088.38502228330 *T)
    sumY0  +=     101.901 * cos(2.19454495695  + 529.44714761109 *T)
    sumY0  +=      73.054 * cos(3.46030032788  + 1052.51220067191 *T)
    sumY0  +=      73.997 * cos(4.31891480859  + 3767.45443505930 *T)
    sumY0  +=      70.372 * cos(6.06190685925  + 553.32558535889 *T)
    sumY0  +=      91.916 * cos(2.68950348204  + 9595.48290670690 *T)
    sumY0  +=      73.607 * cos(3.91508991039  + 7740.85060107230 *T)
    sumY0  +=      66.068 * cos(0.03272782030  + 1066.73929467350 *T)
    sumY0  +=      81.219 * cos(5.57497325530  + 2810.67764412170 *T)
    sumY0  +=      66.430 * cos(3.87195498920  + 2700.47132290230 *T)
    sumY0  +=      54.145 * cos(2.72166186985  + 6526.04827144891 *T)
    sumY0  +=      59.768 * cos(3.24283548288  + 6041.57138456910 *T)
    sumY0  +=      66.173 * cos(4.75351556577  + 10022.09835123510 *T)
    sumY0  +=      49.883 * cos(3.52241504779  + 3312.40705671550 *T)
    sumY0  +=      49.178 * cos(2.60154680851  + 3369.30543165110 *T)
    sumY0  +=      48.202 * cos(0.32517036897  + 397.90518592470 *T)
    sumY0  +=      47.310 * cos(3.94287633641  + 15643.92402079330 *T)
    sumY0  +=      48.401 * cos(5.71493252665  + 3.76693583251 *T)
    sumY0  +=      66.173 * cos(2.72274393427  + 10022.06384393071 *T)
    sumY0  +=      43.167 * cos(3.68410600012  + 4690.23601887509 *T)
    sumY0  +=      52.329 * cos(3.78533146894  + 7903.31723720450 *T)
    sumY0  +=      43.258 * cos(0.37554748132  + 149.80701461810 *T)
    sumY0  +=      53.318 * cos(0.93774568542  + 10420.23010099111 *T)
    sumY0  +=      38.925 * cos(4.06852349635  + 6682.44899195131 *T)
    sumY0  +=      43.058 * cos(1.14093100948  + 3325.60377299830 *T)
    sumY0  +=      42.835 * cos(4.98274982261  + 3356.10871536831 *T)
    sumY0  +=      49.042 * cos(0.37958979733  + 2273.87313202630 *T)
    sumY0  +=      45.897 * cos(3.30255905408  + 2288.10022602790 *T)
    sumY0  +=      36.267 * cos(0.37823003866  + 23384.53080438210 *T)
    sumY0  +=      50.027 * cos(0.20993213048  + 5621.59910572690 *T)
    sumY0  +=      41.152 * cos(2.03769608425  + 537.04832957890 *T)
    sumY0  +=      38.722 * cos(5.48335922125  + 522.82123557730 *T)
    sumY0  +=      44.751 * cos(2.74146086011  + 3077.77232081051 *T)
    sumY0  +=      31.695 * cos(2.65244951875  + 5629.20028769470 *T)
    sumY0  +=      31.478 * cos(0.69290700501  + 5331.60126122430 *T)
    sumY0  +=      33.349 * cos(0.55070002892  + 242.97242145751 *T)
    sumY0  +=      36.130 * cos(0.70369333471  + 9808.78200214491 *T)
    sumY0  +=      31.660 * cos(3.54279700539  + 2648.69864295650 *T)
    sumY0  +=      39.058 * cos(2.73256710316  + 6680.48834981490 *T)
    sumY0  +=      32.374 * cos(3.86940131593  + 7875.91568110771 *T)
    sumY0  +=      38.493 * cos(0.77970755386  + 6894.76776632111 *T)
    sumY0  +=      40.044 * cos(3.38354028832  + 3603.94016755610 *T)
    sumY0  +=      27.893 * cos(2.32668323191  + 7373.13863714289 *T)
    sumY0  +=      27.665 * cos(0.80854741955  + 12964.54452087450 *T)
    sumY0  +=      27.878 * cos(5.37977989507  + 1350.11122714230 *T)
    sumY0  +=      28.322 * cos(4.11041470939  + 8030.84844557490 *T)
    sumY0  +=      28.486 * cos(6.08536458002  + 5614.97319369310 *T)
    sumY0  +=      29.761 * cos(3.84843907128  + 3178.38960805111 *T)
    sumY0  +=      35.657 * cos(2.70492671014  + 8428.99744898310 *T)
    sumY0  +=      28.159 * cos(2.26219738844  + 3503.32288031550 *T)
    sumY0  +=      27.295 * cos(1.14579185082  + 213.05527795450 *T)
    sumY0  +=      24.083 * cos(2.53502654613  + 4033.01384541010 *T)
    sumY0  +=      23.726 * cos(2.22123833128  + 3320.50092478451 *T)
    sumY0  +=      23.543 * cos(3.90125907437  + 3361.21156358210 *T)
    sumY0  +=      24.407 * cos(0.04920773130  + 76.50988875911 *T)
    sumY0  +=      22.360 * cos(3.29082907917  + 15114.23305569870 *T)
    sumY0  +=      22.681 * cos(0.62465627638  + 13366.21664263171 *T)
    sumY0  +=      30.767 * cos(3.54413358961  + 102.84895673509 *T)
    sumY0  +=      23.116 * cos(2.67363493175  + 1692.40948698591 *T)
    sumY0  +=      24.795 * cos(4.51530523814  + 691.91378374330 *T)
    sumY0  +=      21.819 * cos(0.39147454554  + 10022.14840788570 *T)
    sumY0  +=      21.420 * cos(2.33354950951  + 6660.69327539070 *T)
    sumY0  +=      20.686 * cos(2.52092831971  + 1582.20316576650 *T)
    sumY0  +=      25.165 * cos(5.06243453868  + 3335.33331987590 *T)
    sumY0  +=      21.819 * cos(0.79997093238  + 10022.01378728011 *T)
    sumY0  +=      25.110 * cos(1.05972287332  + 3346.37916849070 *T)
    sumY0  +=      22.778 * cos(6.02670962938  + 9830.63283147130 *T)
    sumY0  +=      21.875 * cos(3.23482242791  + 11243.92966390430 *T)
    sumY0  +=      19.487 * cos(4.05002877465  + 3336.84549086730 *T)
    sumY0  +=      25.048 * cos(4.04459983417  + 1648.20293971390 *T)
    sumY0  +=      19.285 * cos(5.68171337975  + 3330.21957883350 *T)
    sumY0  +=      19.206 * cos(0.44033697686  + 3351.49290953311 *T)
    sumY0  +=      19.577 * cos(5.54020511510  + 6438.74006690910 *T)
    sumY0  +=      23.713 * cos(2.35775949900  + 796.05418933290 *T)
    sumY0  +=      18.024 * cos(2.37243714858  + 3.27930086550 *T)
    sumY0  +=      17.457 * cos(5.61909053663  + 6872.91693699471 *T)
    sumY0  +=      20.758 * cos(1.11119646396  + 2914.25805330730 *T)
    sumY0  +=      23.325 * cos(4.38303926683  + 1758.40926093329 *T)
    sumY0  +=      19.762 * cos(5.90704491623  + 2818.27882608950 *T)
    sumY0  +=      18.903 * cos(0.13118670804  + 3863.43366227710 *T)
    sumY0  +=      18.548 * cos(3.43514985020  + 3343.89172756530 *T)
    sumY0  +=      20.024 * cos(3.66434113579  + 9382.18381126891 *T)
    sumY0  +=      23.004 * cos(1.24456792545  + 13359.17040593370 *T)
    sumY0  +=      20.686 * cos(4.47763433062  + 6702.24406637550 *T)
    sumY0  +=      17.546 * cos(5.48185771093  + 3074.24920246151 *T)
    sumY0  +=      21.953 * cos(1.45487277078  + 10713.75106384269 *T)
    sumY0  +=      22.478 * cos(4.93708763645  + 3531.81687532789 *T)
    sumY0  +=      17.443 * cos(1.52686638431  + 7063.87756813929 *T)
    sumY0  +=      16.475 * cos(3.32396294407  + 8432.52056733209 *T)
    sumY0  +=      16.284 * cos(3.03747056649  + 6298.57213865991 *T)
    sumY0  +=      16.296 * cos(1.33599982138  + 10551.77206267750 *T)
    sumY0  +=      21.367 * cos(5.09084614486  + 4292.08701546690 *T)
    sumY0  +=      15.862 * cos(1.17152126786  + 3723.26514143949 *T)
    sumY0  +=      16.182 * cos(5.51657199394  + 7632.69944216670 *T)
    sumY0  +=      18.728 * cos(3.30231901555  + 4032.52621044309 *T)
    sumY0  +=      17.711 * cos(5.62352752655  + 10818.37910439930 *T)
    sumY0  +=      16.214 * cos(0.64104308804  + 3607.46328590511 *T)
    sumY0  +=      20.207 * cos(5.21669629085  + 2708.07250487011 *T)
    sumY0  +=      17.851 * cos(3.90780341044  + 6836.88907031730 *T)
    sumY0  +=      14.367 * cos(1.90616251659  + 8429.48508395010 *T)
    sumY0  +=      16.523 * cos(0.57557849919  + 6158.89125278930 *T)
    sumY0  +=      19.608 * cos(1.94522419099  + 853.44019923550 *T)
    sumY0  +=      14.968 * cos(5.74996412926  + 6531.90547374850 *T)
    sumY0  +=      13.797 * cos(0.79312255069  + 419.72846135871 *T)
    sumY0  +=      13.414 * cos(4.51615928517  + 3127.06951377830 *T)
    sumY0  +=      15.245 * cos(0.67463983149  + 7064.36520310630 *T)
    sumY0  +=      15.674 * cos(2.89380716586  + 3333.81000748530 *T)
    sumY0  +=      13.513 * cos(3.61353175693  + 6685.05909953490 *T)
    sumY0  +=      16.177 * cos(5.55617497884  + 5729.75026463250 *T)
    sumY0  +=      16.167 * cos(4.06504177790  + 6144.66415878770 *T)
    sumY0  +=      13.924 * cos(3.03881519646  + 12936.09533340670 *T)
    sumY0  +=      12.758 * cos(0.71728265214  + 3336.97490882530 *T)
    sumY0  +=      16.755 * cos(0.87447011602  + 3973.63998349650 *T)
    sumY0  +=      12.740 * cos(5.40649102919  + 3344.73757954131 *T)
    sumY0  +=      15.216 * cos(3.70357996455  + 5459.62010456170 *T)
    sumY0  +=      13.755 * cos(3.67924088742  + 433.95555536030 *T)
    sumY0  +=      13.727 * cos(4.13790239727  + 6127.89926804070 *T)
    sumY0  +=      16.418 * cos(3.85207670396  + 75.02541605080 *T)
    sumY0  +=      12.379 * cos(4.87573415701  + 2487.17222746430 *T)
    sumY0  +=      12.121 * cos(3.30257314405  + 3980.75353049730 *T)
    sumY0  +=      12.145 * cos(3.07029009464  + 3281.48238226970 *T)
    sumY0  +=      16.915 * cos(1.87976030933  + 10404.48999483909 *T)
    sumY0  +=      13.732 * cos(3.17849954583  + 6677.87824223130 *T)
    sumY0  +=      12.163 * cos(5.94119470839  + 10973.31186886650 *T)
    sumY0  +=      11.839 * cos(3.05197069508  + 3400.23010609691 *T)
    sumY0  +=      13.460 * cos(0.01339869768  + 1596.43025976811 *T)
    sumY0  +=      11.950 * cos(0.72333324885  + 1744.18216693170 *T)
    sumY0  +=      14.031 * cos(3.20900224186  + 8031.33608054190 *T)
    sumY0  +=      13.621 * cos(4.01984912325  + 3415.63784275061 *T)
    sumY0  +=      11.890 * cos(3.44240756999  + 4989.30300138070 *T)
    sumY0  +=      11.133 * cos(3.95045133494  + 8270.54156616690 *T)
    sumY0  +=      11.115 * cos(1.68346953337  + 6816.53375091851 *T)
    sumY0  +=      10.961 * cos(1.29021416023  + 110.45013870291 *T)
    sumY0  +=      12.029 * cos(2.48781942810  + 1751.78334889950 *T)
    sumY0  +=      12.954 * cos(3.09490820262  + 8969.81271439450 *T)
    sumY0  +=      12.975 * cos(2.11989840845  + 3266.07464561600 *T)
    sumY0  +=      13.898 * cos(5.14618871897  + 6546.40359084770 *T)
    sumY0  +=      10.787 * cos(2.11806328265  + 11770.09751064990 *T)
    sumY0  +=      13.676 * cos(3.07870529532  + 3347.90248088131 *T)
    sumY0  +=      14.429 * cos(1.92454286503  + 6041.08374960209 *T)
    sumY0  +=      10.265 * cos(3.57928773201  + 5099.50932260011 *T)
    sumY0  +=      10.855 * cos(5.60540188720  + 3316.97780643551 *T)
    sumY0  +=      13.203 * cos(0.19791164589  + 8955.58562039291 *T)
    sumY0  +=      10.750 * cos(0.51835327782  + 3364.73468193110 *T)
    sumY0  +=      12.106 * cos(4.95253802158  + 6.86972951729 *T)
    sumY0  +=      10.745 * cos(1.75142473456  + 16173.61498588790 *T)
    sumY0  +=      10.862 * cos(4.59591535215  + 7633.18707713370 *T)
    sumY0  +=      13.130 * cos(4.64408283563  + 6688.58221788390 *T)
    sumY0  +=      12.618 * cos(0.44972794093  + 3443.94901840191 *T)
    sumY0  +=      10.348 * cos(2.48573429062  + 5884.68302909969 *T)
    sumY0  +=      11.408 * cos(5.58895311057  + 3237.76346996470 *T)
    sumY0  +=      10.014 * cos(0.21528796338  + 3893.93801205869 *T)
    sumY0  +=      13.330 * cos(4.53273448064  + 11371.46087227470 *T)
    sumY0  +=       9.390 * cos(4.70787699498  + 6151.29007082150 *T)
    sumY0  +=       9.317 * cos(4.47268149096  + 3304.82837750590 *T)
    sumY0  +=       9.588 * cos(2.36418014273  + 14584.54209060410 *T)
    sumY0  +=       9.187 * cos(1.65111069662  + 3376.88411086071 *T)
    sumY0  +=       8.922 * cos(2.44896013420  + 8274.06468451590 *T)
    sumY0  +=      12.186 * cos(6.02036123080  + 10014.96755058210 *T)
    sumY0  +=       8.852 * cos(0.78017288566  + 2921.37160030811 *T)
    sumY0  +=      11.498 * cos(1.82027521748  + 3024.46437452670 *T)
    sumY0  +=       8.932 * cos(0.47616058050  + 3340.43616808970 *T)
    sumY0  +=       8.932 * cos(5.64760633814  + 3341.27632027691 *T)
    sumY0  +=      10.791 * cos(2.77302371275  + 13362.20588931570 *T)
    sumY0  +=      10.440 * cos(4.29767407547  + 3657.24811383991 *T)
    sumY0  +=      10.488 * cos(0.76119276539  + 12168.24651405810 *T)
    sumY0  +=       7.791 * cos(4.94235946787  + 162.22281864870 *T)
    sumY0  +=       8.278 * cos(6.07340700806  + 6605.20259960750 *T)
    sumY0  +=       7.730 * cos(6.04134808195  + 7235.03807372550 *T)
    sumY0  +=      10.793 * cos(0.90852791213  + 135.30889751891 *T)
    sumY0  +=       9.855 * cos(0.50508691053  + 951.96222373410 *T)
    sumY0  +=       7.570 * cos(5.85095682416  + 1861.50203515190 *T)
    sumY0  +=       7.347 * cos(5.45188972477  + 3760.34088805850 *T)
    sumY0  +=       7.458 * cos(3.50693253979  + 11371.94850724170 *T)
    sumY0  +=       7.415 * cos(1.06538367574  + 10575.16286545830 *T)
    sumY0  +=       7.387 * cos(4.04748411837  + 7373.62627210991 *T)
    sumY0  +=       7.330 * cos(5.09163920237  + 13362.71077793490 *T)
    sumY0  +=       7.563 * cos(2.58673910300  + 3134.67069574611 *T)
    sumY0  +=       8.273 * cos(5.52944425917  + 633.02755679670 *T)
    sumY0  +=       9.017 * cos(3.31603055267  + 6784.56144510170 *T)
    sumY0  +=       7.330 * cos(3.06086757088  + 13362.67627063050 *T)
    sumY0  +=       6.892 * cos(2.23117060751  + 3170.18337356410 *T)
    sumY0  +=       9.499 * cos(0.28342373247  + 191.69208359510 *T)
    sumY0  +=       8.390 * cos(1.78738551777  + 12566.39551746630 *T)
    sumY0  +=       7.258 * cos(3.50543277363  + 3547.04179262050 *T)
    sumY0  +=       7.563 * cos(1.86781369518  + 3209.31428249690 *T)
    sumY0  +=       7.609 * cos(3.51290916963  + 3304.25087887911 *T)
    sumY0  +=       7.490 * cos(1.30636174884  + 6924.19727485710 *T)
    sumY0  +=       6.532 * cos(4.34833038552  + 1118.51197461930 *T)
    sumY0  +=       6.446 * cos(1.91072140138  + 3340.42636105660 *T)
    sumY0  +=       6.446 * cos(4.21304551726  + 3341.28612731000 *T)
    sumY0  +=       8.809 * cos(2.38729963716  + 8827.63408735830 *T)
    sumY0  +=       6.453 * cos(3.89231889829  + 3511.52911480250 *T)
    sumY0  +=       7.497 * cos(2.61086786245  + 3377.46160948750 *T)
    sumY0  +=       6.647 * cos(6.12930067880  + 1376.01761732930 *T)
    sumY0  +=       6.444 * cos(0.76608645809  + 1479.11039154791 *T)
    sumY0  +=       6.777 * cos(0.94042275680  + 1194.20319274110 *T)
    sumY0  +=       6.272 * cos(2.60939993697  + 11243.44202893729 *T)
    sumY0  +=       7.647 * cos(4.24404085436  + 11081.46302777211 *T)
    sumY0  +=       6.372 * cos(4.42614040426  + 8827.14645239129 *T)
    sumY0  +=       8.206 * cos(0.63888089259  + 7234.55043875850 *T)
    sumY0  +=       6.106 * cos(1.65903263989  + 10184.06009874809 *T)
    sumY0  +=       6.255 * cos(1.27116212393  + 956.53297345411 *T)
    sumY0  +=       6.895 * cos(4.25463819285  + 3472.39820586971 *T)
    sumY0  +=       5.956 * cos(5.42533370697  + 641.12142486571 *T)
    sumY0  +=       6.342 * cos(4.38723597534  + 1162.71852189130 *T)
    sumY0  +=       8.015 * cos(4.17209419953  + 5828.27228913110 *T)
    sumY0  +=       6.858 * cos(6.08696943259  + 5085.28222859850 *T)
    sumY0  +=       5.564 * cos(2.95420857060  + 949.41942645330 *T)
    sumY0  +=       5.685 * cos(1.48739730270  + 13915.77529215809 *T)
    sumY0  +=       5.989 * cos(1.27210341857  + 13760.84252769090 *T)
    sumY0  +=       6.614 * cos(2.59817425884  + 5099.02168763310 *T)
    sumY0  +=       6.046 * cos(4.28381801697  + 18984.53644749310 *T)
    sumY0  +=       5.388 * cos(4.75457886592  + 1228.71829583870 *T)
    sumY0  +=       6.477 * cos(4.76369580716  + 2177.89390480850 *T)
    sumY0  +=       6.716 * cos(5.99595862554  + 5084.79459363150 *T)
    sumY0  +=       6.834 * cos(2.30327442130  + 6606.68707231580 *T)
    sumY0  +=       6.185 * cos(5.40821565539  + 220.65645992230 *T)
    sumY0  +=       5.372 * cos(0.17118392777  + 2766.51144584850 *T)
    sumY0  +=       6.939 * cos(3.86951988876  + 8329.91542808050 *T)
    sumY0  +=       5.250 * cos(2.86332649708  + 224.58861318540 *T)
    sumY0  +=       5.178 * cos(1.81203706439  + 3274.36883526891 *T)
    sumY0  +=       5.636 * cos(3.05816376185  + 9866.66069814871 *T)
    sumY0  +=       5.018 * cos(4.30701144279  + 3407.34365309770 *T)
    sumY0  +=       6.287 * cos(2.70767720037  + 2222.10045208050 *T)
    sumY0  +=       6.027 * cos(5.86363508643  + 3340.35135556410 *T)
    sumY0  +=       6.581 * cos(3.78955993789  + 38.37685312130 *T)
    sumY0  +=       6.293 * cos(4.17674637169  + 8440.12174929990 *T)
    sumY0  +=       4.882 * cos(1.00486095714  + 10404.97762980610 *T)
    sumY0  +=       4.795 * cos(0.36427947507  + 4407.35172137330 *T)
    sumY0  +=       6.476 * cos(3.29629751497  + 3297.13733187830 *T)
    sumY0  +=       6.362 * cos(2.82771763572  + 3384.57515648830 *T)
    sumY0  +=       5.306 * cos(2.70813322540  + 206.42936592071 *T)
    sumY0  +=       6.484 * cos(4.73121616424  + 7108.06686175911 *T)
    sumY0  +=       4.752 * cos(5.47348414235  + 6261.98402700790 *T)
    sumY0  +=       6.261 * cos(0.59716322071  + 1964.59480937050 *T)
    sumY0  +=       5.039 * cos(2.43909809375  + 6247.75693300630 *T)
    sumY0  +=       6.027 * cos(1.61122140965  + 3340.38586286849 *T)
    sumY0  +=       4.586 * cos(2.93637860920  + 6709.91785835091 *T)
    sumY0  +=       4.635 * cos(6.08947141473  + 263.32774085630 *T)
    sumY0  +=       4.870 * cos(3.87407090566  + 4393.12462737170 *T)
    sumY0  +=       6.261 * cos(2.83172709667  + 26088.14695905770 *T)
    sumY0  +=       4.434 * cos(3.32464019476  + 155.66421691770 *T)
    sumY0  +=       4.383 * cos(1.48980830926  + 6666.21619969810 *T)
    sumY0  +=       4.389 * cos(1.03247430987  + 13149.39442884470 *T)
    sumY0  +=       4.433 * cos(6.22122951581  + 2807.64216073970 *T)
    sumY0  +=       4.980 * cos(2.89384649269  + 5827.78465416409 *T)
    sumY0  +=       4.443 * cos(5.95476217838  + 3915.20104251810 *T)
    sumY0  +=       4.299 * cos(0.75007520827  + 8962.21153242670 *T)
    sumY0  +=       4.356 * cos(1.14577413888  + 10235.38019302090 *T)
    sumY0  +=       4.498 * cos(1.91438448815  + 553.81322032590 *T)
    sumY0  +=       5.019 * cos(4.13208637738  + 3738.51761262449 *T)
    sumY0  +=       4.315 * cos(3.27234333766  + 3877.66075627870 *T)
    sumY0  +=       4.072 * cos(3.33901445618  + 574.10098085129 *T)
    sumY0  +=       4.107 * cos(3.07058211212  + 10021.10077651470 *T)
    sumY0  +=       3.964 * cos(1.61309186306  + 3495.78900865049 *T)
    sumY0  +=       4.785 * cos(0.27642621276  + 2699.97863680110 *T)
    sumY0  +=       5.396 * cos(0.07454063526  + 14313.92429556630 *T)
    sumY0  +=       3.822 * cos(0.64534260204  + 3490.41944131790 *T)
    sumY0  +=       4.095 * cos(4.40659363743  + 10023.06141865111 *T)
    sumY0  +=       3.697 * cos(4.79643368803  + 6571.26234966370 *T)
    sumY0  +=       3.779 * cos(2.89531499128  + 13517.62628874989 *T)
    sumY0  +=       3.860 * cos(2.04716548704  + 6836.40143535029 *T)
    sumY0  +=       4.702 * cos(2.45248163598  + 2111.89413086110 *T)
    sumY0  +=       3.630 * cos(0.00915212983  + 11769.60987568290 *T)
    sumY0  +=       4.499 * cos(0.82553557194  + 13524.67252544789 *T)
    sumY0  +=       3.497 * cos(4.11625604633  + 6282.83203250789 *T)
    sumY0  +=       3.489 * cos(4.20101724067  + 11216.52810780750 *T)
    sumY0  +=       3.591 * cos(6.26737509852  + 3874.07032762691 *T)
    sumY0  +=       3.976 * cos(2.17190102533  + 2804.05173208791 *T)
    sumY0  +=       4.680 * cos(4.90769516889  + 10973.79950383350 *T)
    sumY0  +=       3.961 * cos(0.18319392811  + 6048.68493156990 *T)
    sumY0  +=       3.954 * cos(3.83351307387  + 6653.01948341530 *T)
    sumY0  +=       3.352 * cos(0.48662889418  + 8425.89465529830 *T)
    sumY0  +=       4.016 * cos(5.83961396630  + 3981.73385156551 *T)
    sumY0  +=       3.356 * cos(4.65730182486  + 9168.88471583091 *T)
    sumY0  +=       4.260 * cos(3.10890117595  + 3333.25506221550 *T)
    sumY0  +=       3.936 * cos(2.60585567874  + 6843.44767204830 *T)
    sumY0  +=       3.767 * cos(0.71635417612  + 26725.14323108190 *T)
    sumY0  +=       3.388 * cos(3.58080763162  + 18454.84548239850 *T)
    sumY0  +=       3.337 * cos(4.78188036655  + 13936.55068765050 *T)
    sumY0  +=       3.230 * cos(0.84980496278  + 262.84010588929 *T)
    sumY0  +=       4.019 * cos(5.32092079015  + 6696.72114206811 *T)
    sumY0  +=       3.967 * cos(0.18799481688  + 266.85085920531 *T)
    sumY0  +=       3.932 * cos(5.32627876219  + 2384.07945324570 *T)
    sumY0  +=       3.087 * cos(5.35381754909  + 5202.60209681870 *T)
    sumY0  +=       3.922 * cos(3.38887523411  + 4459.61203628611 *T)
    sumY0  +=       3.717 * cos(3.00241667255  + 2906.65687133950 *T)
    sumY0  +=       3.266 * cos(1.14686627484  + 16305.15694757430 *T)
    sumY0  +=       3.104 * cos(6.26686342256  + 5305.69487103730 *T)
    sumY0  +=       3.000 * cos(3.55062935508  + 10714.23869880971 *T)
    sumY0  +=       3.921 * cos(2.47083832508  + 10177.01386205009 *T)
    sumY0  +=       3.084 * cos(1.63553589002  + 1059.13811270570 *T)
    sumY0  +=       3.253 * cos(1.65815244061  + 5518.99396647530 *T)
    sumY0  +=       2.861 * cos(4.91663574228  + 1331.81107005730 *T)
    sumY0  +=       3.304 * cos(1.02799621733  + 8672.21368792411 *T)
    sumY0  +=       2.975 * cos(0.03645619502  + 10575.65050042530 *T)
    sumY0  +=       2.756 * cos(0.68592989650  + 6040.59106350090 *T)
    sumY0  +=       2.950 * cos(3.68233121776  + 6944.55259425590 *T)
    sumY0  +=       3.005 * cos(0.80186680463  + 2707.58486990309 *T)
    sumY0  +=       2.717 * cos(4.19242288497  + 6519.00203475090 *T)
    sumY0  +=       2.801 * cos(3.31637382659  + 16858.23871544969 *T)
    sumY0  +=       2.897 * cos(2.39765481455  + 2700.95895786930 *T)
    sumY0  +=       3.274 * cos(5.89447455545  + 2920.88396534110 *T)
    sumY0  +=       2.645 * cos(3.89694393584  + 5989.31106965630 *T)
    sumY0  +=       3.228 * cos(2.76066844405  + 647.25465079831 *T)
    sumY0  +=       2.528 * cos(0.08269800947  + 13171.24525817110 *T)
    sumY0  +=       3.531 * cos(6.23552426531  + 3190.80541208169 *T)
    sumY0  +=       2.590 * cos(4.84869275207  + 5725.17951491250 *T)
    sumY0  +=       2.501 * cos(3.06554091162  + 383.14034970670 *T)
    sumY0  +=       2.483 * cos(2.30179982047  + 3313.45468808651 *T)
    sumY0  +=       2.456 * cos(3.82204304981  + 3368.25780028011 *T)
    sumY0  +=       2.450 * cos(3.18294245687  + 5732.29306191330 *T)
    sumY0  +=       2.623 * cos(1.58355072138  + 16699.78283263350 *T)
    sumY0  +=       3.010 * cos(1.80619032361  + 2008.80135664250 *T)
    sumY0  +=       3.148 * cos(3.00255938689  + 3097.64000524229 *T)
    sumY0  +=       2.412 * cos(0.20563592554  + 3561.26888662211 *T)
    sumY0  +=       2.837 * cos(3.54825853722  + 4569.33072253850 *T)
    sumY0  +=       2.416 * cos(0.73041207552  + 13362.76083458550 *T)
    sumY0  +=       2.532 * cos(3.29257381257  + 6475.28312244591 *T)
    sumY0  +=       2.407 * cos(1.91810131323  + 134.82126255190 *T)
    sumY0  +=       2.351 * cos(2.54653063726  + 17925.15451730390 *T)
    sumY0  +=       3.311 * cos(5.13364869654  + 17277.16311435029 *T)
    sumY0  +=       2.384 * cos(4.42145865201  + 9485.27658548750 *T)
    sumY0  +=       2.320 * cos(3.88497277059  + 14712.56093394150 *T)
    sumY0  +=       2.350 * cos(5.79681206289  + 533.45790092711 *T)
    sumY0  +=       2.433 * cos(2.40535740363  + 3171.27606105030 *T)
    sumY0  +=       2.860 * cos(3.63565692925  + 4569.81835750550 *T)
    sumY0  +=       2.382 * cos(1.05697861132  + 15508.85894075790 *T)
    sumY0  +=       2.274 * cos(1.39549918259  + 6686.99159519050 *T)
    sumY0  +=       2.448 * cos(3.02014673677  + 6418.38474751031 *T)
    sumY0  +=       2.223 * cos(4.42688952943  + 10595.93826095070 *T)
    sumY0  +=       2.362 * cos(3.60748801511  + 6843.93530701530 *T)
    sumY0  +=       2.247 * cos(5.47246717730  + 3338.10473357230 *T)
    sumY0  +=       2.247 * cos(0.65129974134  + 3343.60775479431 *T)
    sumY0  +=       2.877 * cos(5.31092659214  + 3870.05957431090 *T)
    sumY0  +=       2.167 * cos(0.11779930741  + 2126.12122486270 *T)
    sumY0  +=       2.861 * cos(0.41825730841  + 3342.34071689161 *T)
    sumY0  +=       2.948 * cos(5.86603214726  + 6414.86162916130 *T)
    sumY0  +=       2.832 * cos(5.70194885568  + 3339.37177147500 *T)
    sumY0  +=       2.257 * cos(2.42684789077  + 14054.85112550950 *T)
    sumY0  +=       2.582 * cos(0.96409333017  + 16706.82906933150 *T)
    sumY0  +=       2.284 * cos(3.72042741576  + 3510.43642731630 *T)
    sumY0  +=       2.213 * cos(5.79088907317  + 3120.44360174450 *T)
    sumY0  +=       2.518 * cos(3.55412857109  + 12722.79623796870 *T)
    sumY0  +=       2.156 * cos(3.68823477654  + 1795.50226120450 *T)
    sumY0  +=       2.120 * cos(1.86086651852  + 14712.07329897449 *T)
    sumY0  +=       2.894 * cos(4.68726480342  + 661.47674426450 *T)
    sumY0  +=       2.197 * cos(1.25938311277  + 7314.25241019630 *T)
    sumY0  +=       2.352 * cos(5.40138877221  + 6675.94574657570 *T)
    sumY0  +=       2.592 * cos(1.99576622470  + 846.32665223470 *T)
    sumY0  +=       2.011 * cos(4.10915276822  + 7800.22446298591 *T)
    sumY0  +=       2.009 * cos(5.79546041665  + 1592.35219614930 *T)
    sumY0  +=       1.985 * cos(5.95943507005  + 14158.99153109910 *T)
    sumY0  +=       2.025 * cos(4.72625136485  + 16460.08971204149 *T)
    sumY0  +=       2.756 * cos(6.08539056721  + 9872.51790044830 *T)
    sumY0  +=       2.586 * cos(2.46711824817  + 15110.70993734970 *T)
    sumY0  +=       1.977 * cos(3.94114619904  + 3340.30129891350 *T)
    sumY0  +=       2.046 * cos(1.08671847901  + 1214.49120183710 *T)
    sumY0  +=       2.049 * cos(3.11791083516  + 1545.11016549530 *T)
    sumY0  +=       2.016 * cos(4.23253609288  + 6701.82399028190 *T)
    sumY0  +=       2.591 * cos(3.35704415242  + 4459.12440131909 *T)
    sumY0  +=       1.944 * cos(6.27867442308  + 793.01870595090 *T)
    sumY0  +=       1.969 * cos(5.69710526556  + 8543.21452351850 *T)
    sumY0  +=       2.221 * cos(2.56014354281  + 6661.11335148431 *T)
    sumY0  +=       2.185 * cos(4.81676561022  + 10042.85649307530 *T)
    sumY0  +=       1.909 * cos(1.94360908246  + 14054.36349054249 *T)
    sumY0  +=       2.355 * cos(0.64545768145  + 6461.05602844430 *T)
    sumY0  +=       1.945 * cos(0.52827400374  + 12296.19804709270 *T)
    sumY0  +=       1.856 * cos(0.39994658659  + 3417.12231545890 *T)
    sumY0  +=       1.964 * cos(2.16479993477  + 3230.64992296391 *T)
    sumY0  +=       1.841 * cos(3.38387944568  + 16276.70776010650 *T)
    sumY0  +=       1.867 * cos(3.44547235502  + 6438.25243194210 *T)
    sumY0  +=       2.124 * cos(0.31401256569  + 14.47091148511 *T)
    sumY0  +=       1.889 * cos(1.27160767434  + 2946.23035912410 *T)
    sumY0  +=       2.273 * cos(2.67716254538  + 10001.30570209050 *T)
    sumY0  +=       1.977 * cos(3.53264979384  + 3340.43591951909 *T)
    sumY0  +=       1.852 * cos(2.83643169906  + 2751.79141717511 *T)
    sumY0  +=       1.888 * cos(3.59823741934  + 3502.83524534850 *T)
    sumY0  +=       1.924 * cos(4.40431574367  + 205.94173095370 *T)
    sumY0  +=       1.900 * cos(3.87884336092  + 9778.86485864190 *T)
    sumY0  +=       1.840 * cos(3.26125681857  + 6457.12387518120 *T)
    sumY0  +=       2.215 * cos(0.91618144000  + 9499.50367948910 *T)
    sumY0  +=       1.806 * cos(0.61464681346  + 2676.10019905331 *T)
    sumY0  +=       1.805 * cos(0.77929673675  + 6692.10533623291 *T)
    sumY0  +=       1.869 * cos(5.13099640755  + 13554.14179039430 *T)
    sumY0  +=       2.416 * cos(1.13890846236  + 13362.62621397991 *T)
    sumY0  +=       2.281 * cos(2.28711563872  + 6643.33563524530 *T)
    sumY0  +=       2.183 * cos(3.37026241840  + 21.01921297590 *T)
    sumY0  +=       1.774 * cos(5.39406505365  + 323.74923414091 *T)
    sumY0  +=       1.748 * cos(3.00013324119  + 5033.02191368570 *T)
    sumY0  +=       1.767 * cos(6.23938721571  + 4555.10362853689 *T)
    sumY0  +=       1.973 * cos(3.46063541883  + 12310.42514109431 *T)
    sumY0  +=       2.109 * cos(0.54420400743  + 7204.04608897691 *T)
    sumY0  +=       2.169 * cos(2.88513479695  + 3148.92034310470 *T)
    sumY0  +=       1.672 * cos(0.98000755171  + 6948.07571260490 *T)
    sumY0  +=       1.685 * cos(0.93633564467  + 1434.90384427590 *T)
    sumY0  +=       2.274 * cos(1.89593685876  + 17256.38771885789 *T)
    sumY0  +=       1.724 * cos(5.60400280415  + 103.33659170210 *T)
    sumY0  +=       1.921 * cos(5.80837070052  + 800.06494264891 *T)
    sumY0  +=       1.651 * cos(4.33969162257  + 6756.25026945040 *T)
    sumY0  +=       1.660 * cos(3.37482712991  + 9639.18456535971 *T)
    sumY0  +=       1.575 * cos(4.52126663062  + 7910.43078420530 *T)
    sumY0  +=       1.556 * cos(3.74191499438  + 743.23387801611 *T)
    sumY0  +=       1.578 * cos(3.90675545614  + 5614.48555872609 *T)
    sumY0  +=       2.080 * cos(4.30362766877  + 13119.47728534169 *T)
    sumY0  +=       2.163 * cos(5.24182574584  + 28.69300495130 *T)
    sumY0  +=       1.532 * cos(5.50079852012  + 4005.61228931331 *T)
    sumY0  +=       1.548 * cos(6.24677186455  + 6578.37589666450 *T)
    sumY0  +=       2.036 * cos(5.86051117946  + 9779.35249360890 *T)
    sumY0  +=       1.534 * cos(4.10234872019  + 2679.62331740231 *T)
    sumY0  +=       1.855 * cos(4.01932968889  + 1039.27042827391 *T)
    sumY0  +=       1.554 * cos(5.29488549205  + 14314.41193053330 *T)
    sumY0  +=       1.555 * cos(4.98628025111  + 10029.19464458370 *T)
    sumY0  +=       1.553 * cos(2.32869255037  + 4194.05262593530 *T)
    sumY0  +=       1.869 * cos(5.29706229752  + 3113.38011139430 *T)
    sumY0  +=       1.547 * cos(0.07021288134  + 13355.57997728191 *T)
    sumY0  +=       1.463 * cos(1.13945911840  + 3362.70707350971 *T)
    sumY0  +=       1.463 * cos(4.98430780024  + 3319.00541485690 *T)
    sumY0  +=       1.468 * cos(1.61396255568  + 632.53992182970 *T)
    sumY0  +=       1.867 * cos(1.75651810408  + 13892.38448937730 *T)
    sumY0  +=       1.483 * cos(4.22066351666  + 2604.97973065150 *T)
    sumY0  +=       1.445 * cos(1.00907902882  + 12994.98156035330 *T)
    sumY0  +=       1.444 * cos(3.51978992540  + 10018.49066893110 *T)
    sumY0  +=       1.862 * cos(6.02925087361  + 6670.83200553330 *T)
    sumY0  +=       1.428 * cos(4.48027025912  + 9468.51169474051 *T)
    sumY0  +=       1.440 * cos(2.33087071634  + 2641.58509595570 *T)
    sumY0  +=       1.456 * cos(3.27733973667  + 3929.92107119150 *T)
    sumY0  +=       1.515 * cos(2.83774866672  + 4922.81559246630 *T)
    sumY0  +=       1.642 * cos(5.89569866176  + 9070.36269133230 *T)
    sumY0  +=       1.501 * cos(5.08513783174  + 664.99986261351 *T)
    sumY0  +=       1.434 * cos(4.16618462135  + 3283.95786924770 *T)
    sumY0  +=       1.756 * cos(4.31622529752  + 4672.91113172411 *T)
    sumY0  +=       1.552 * cos(5.46763897576  + 9887.01601754750 *T)
    sumY0  +=       1.430 * cos(3.95481187554  + 10025.67152623470 *T)
    sumY0  +=       1.459 * cos(4.00723308991  + 2693.35777590150 *T)
    sumY0  +=       1.423 * cos(5.14334385957  + 19800.70213874130 *T)
    sumY0  +=       1.403 * cos(1.95714229211  + 3397.75461911890 *T)
    sumY0  +=       1.451 * cos(1.35501835403  + 13745.59005650590 *T)
    sumY0  +=       1.366 * cos(5.98910954120  + 4555.59126350391 *T)
    sumY0  +=       1.580 * cos(5.04278942486  + 699.02733074409 *T)
    sumY0  +=       1.700 * cos(0.82624348118  + 3568.33237697230 *T)
    sumY0  +=       1.599 * cos(3.45370180139  + 3212.83740084591 *T)
    sumY0  +=       1.492 * cos(2.02328397399  + 6155.30082413750 *T)
    sumY0  +=       1.463 * cos(4.18478882640  + 1015.41920040070 *T)
    sumY0  +=       1.302 * cos(4.10311228163  + 5625.60985904290 *T)
    sumY0  +=       1.329 * cos(4.49775879038  + 3334.17258030920 *T)
    sumY0  +=       1.329 * cos(1.62600812826  + 3347.53990805741 *T)
    sumY0  +=       1.503 * cos(4.97907412209  + 3378.98927982111 *T)
    sumY0  +=       1.397 * cos(4.73682156749  + 588.82100952470 *T)
    sumY0  +=       1.500 * cos(5.87752299680  + 3.68833577960 *T)
    sumY0  +=       1.522 * cos(2.67169732879  + 3468.87508752070 *T)
    sumY0  +=       1.286 * cos(0.30381647283  + 3152.93109642071 *T)
    sumY0  +=       1.600 * cos(6.11940741732  + 1905.70858242391 *T)
    sumY0  +=       1.339 * cos(4.42233255843  + 2699.49100183409 *T)
    sumY0  +=       1.563 * cos(3.16953602104  + 6674.42243418510 *T)
    sumY0  +=       1.455 * cos(4.05006092350  + 8800.23253126150 *T)
    sumY0  +=       1.337 * cos(5.43427386744  + 2277.46356067810 *T)
    sumY0  +=       1.259 * cos(1.69486250292  + 3742.52836594050 *T)
    sumY0  +=       1.315 * cos(4.86859160776  + 1191.16770935910 *T)
    sumY0  +=       1.298 * cos(1.66827735532  + 17395.46355220930 *T)
    sumY0  +=       1.692 * cos(0.41630178717  + 8646.30729773711 *T)
    sumY0  +=       1.571 * cos(3.60195764792  + 3451.06256540270 *T)
    sumY0  +=       1.377 * cos(4.06860369404  + 7255.32583425090 *T)
    sumY0  +=       1.223 * cos(3.70588064572  + 17654.53672226610 *T)
    sumY0  +=       1.272 * cos(4.68637067309  + 6997.86054053970 *T)
    sumY0  +=       1.293 * cos(3.67069875408  + 10125.17387180150 *T)
    sumY0  +=       1.205 * cos(4.45065917256  + 74.53778108379 *T)
    sumY0  +=       1.193 * cos(5.74461193872  + 6685.35000624110 *T)
    sumY0  +=       1.184 * cos(5.62673019992  + 2178.38153977550 *T)
    sumY0  +=       1.175 * cos(1.89792440951  + 4076.73275771510 *T)
    sumY0  +=       1.203 * cos(4.18116497461  + 11670.52785478030 *T)
    sumY0  +=       1.334 * cos(5.10766451811  + 59.13004443009 *T)
    sumY0  +=       1.144 * cos(5.64496891715  + 3191.29304704871 *T)
    sumY0  +=       1.205 * cos(3.32707601504  + 6688.51490758111 *T)
    sumY0  +=       1.228 * cos(5.80609342080  + 2391.68063521350 *T)
    sumY0  +=       1.148 * cos(2.44539933176  + 4886.21022716211 *T)
    sumY0  +=       1.116 * cos(4.51745766455  + 15.49628866851 *T)
    sumY0  +=       1.253 * cos(5.05556621040  + 66.24359143090 *T)
    sumY0  +=       1.122 * cos(3.77962104036  + 4040.12739241091 *T)
    sumY0  +=       1.358 * cos(5.34957633026  + 7895.71605523670 *T)
    sumY0  +=       1.102 * cos(2.41366384952  + 6365.07680122650 *T)
    sumY0  +=       1.406 * cos(2.80809369415  + 9588.36935970610 *T)
    sumY0  +=       1.171 * cos(3.63626867502  + 2772.03437015590 *T)
    sumY0  +=       1.123 * cos(0.29206000273  + 19402.55313533309 *T)
    sumY0  +=       1.128 * cos(3.40923120609  + 6622.09480896950 *T)
    sumY0  +=       1.312 * cos(3.60055878901  + 3311.42673564730 *T)
    sumY0  +=       1.312 * cos(2.52320812963  + 3370.28575271930 *T)
    sumY0  +=       1.101 * cos(4.21894828336  + 3774.56798206011 *T)
    sumY0  +=       1.088 * cos(1.71898270137  + 6148.25458743950 *T)
    sumY0  +=       1.229 * cos(0.96682659655  + 3145.88485972270 *T)
    sumY0  +=       1.354 * cos(2.50280456655  + 4136.66661603269 *T)
    sumY0  +=       1.266 * cos(4.04478089387  + 6887.65421932030 *T)
    sumY0  +=       1.192 * cos(4.67129146326  + 2284.50979737610 *T)
    sumY0  +=       1.059 * cos(4.36720591986  + 1198.21394605711 *T)
    sumY0  +=       1.222 * cos(2.97121261302  + 3184.45552368090 *T)
    sumY0  +=       1.094 * cos(4.64408209257  + 2074.80113058990 *T)
    sumY0  +=       1.086 * cos(2.75001390170  + 2325.19322629910 *T)
    sumY0  +=       1.116 * cos(4.29111524058  + 11611.15399286670 *T)
    sumY0  +=       1.041 * cos(3.66185163778  + 6684.50415426510 *T)
    sumY0  +=       1.105 * cos(4.51598444814  + 5618.56362234490 *T)
    sumY0  +=       1.218 * cos(2.02575108900  + 10157.14617761831 *T)
    sumY0  +=       1.292 * cos(1.97917110685  + 7909.94314923829 *T)
    sumY0  +=       1.193 * cos(1.05540356619  + 6677.58733552510 *T)
    sumY0  +=       1.011 * cos(1.88468389957  + 2494.28577446509 *T)
    sumY0  +=       1.102 * cos(3.69335958507  + 7321.36595719711 *T)
    sumY0  +=       1.001 * cos(0.06352856156  + 11766.50708199810 *T)
    sumY0  +=       1.048 * cos(5.71072902244  + 12721.32828193349 *T)
    sumY0  +=       0.990 * cos(4.38329192789  + 3181.91272640011 *T)
    sumY0  +=       1.146 * cos(3.15307281945  + 3497.25696468570 *T)
    sumY0  +=       0.991 * cos(6.20825855106  + 4140.67736934870 *T)
    sumY0  +=       1.269 * cos(3.62849885608  + 3188.95896309811 *T)
    sumY0  +=       1.254 * cos(3.35511704434  + 6034.45783756830 *T)
    sumY0  +=       1.293 * cos(0.35118272433  + 1.72829019180 *T)
    sumY0  +=       1.206 * cos(3.76131577148  + 6740.84253279671 *T)
    sumY0  +=       1.073 * cos(0.65876193854  + 9389.29735826970 *T)
    sumY0  +=       1.361 * cos(0.12651635396  + 9945.81502630730 *T)
    sumY0  +=       0.977 * cos(0.35688532237  + 3319.52060371631 *T)
    sumY0  +=       0.977 * cos(5.76688159627  + 3362.19188465030 *T)
    sumY0  +=       0.969 * cos(4.50332531081  + 1272.92484311070 *T)
    sumY0  +=       1.176 * cos(5.72271210931  + 9602.59645370770 *T)
    sumY0  +=       1.347 * cos(2.06226659912  + 19514.22741258770 *T)
    sumY0  +=       1.073 * cos(4.27891714150  + 6677.45791756710 *T)
    sumY0  +=       0.968 * cos(1.10386271982  + 522.33360061030 *T)
    sumY0  +=       1.192 * cos(2.35359183418  + 1222.09238380491 *T)
    sumY0  +=       1.237 * cos(4.15588129311  + 5835.38583613191 *T)
    sumY0  +=       1.057 * cos(2.86510171570  + 2597.37854868370 *T)
    sumY0  +=       0.957 * cos(4.82126145069  + 6645.44080420570 *T)
    sumY0  +=       0.971 * cos(1.72322720325  + 20809.22380716170 *T)
    sumY0  +=       0.991 * cos(0.53667136955  + 131.78577916990 *T)
    sumY0  +=       1.159 * cos(5.13897713303  + 9065.79194161230 *T)
    sumY0  +=       0.943 * cos(3.92998399256  + 3337.31587218211 *T)
    sumY0  +=       0.943 * cos(5.33537557967  + 3344.39661618450 *T)
    sumY0  +=       0.929 * cos(2.07665479133  + 4002.08917096430 *T)
    sumY0  +=       0.977 * cos(2.12320798920  + 2538.49232173710 *T)
    sumY0  +=       0.938 * cos(4.38862417624  + 931.60690433531 *T)
    sumY0  +=       1.272 * cos(0.89864359772  + 3302.72320854550 *T)
    sumY0  +=       0.960 * cos(2.48494556814  + 3909.67811821070 *T)
    sumY0  +=       0.909 * cos(2.06818561261  + 3326.62915018171 *T)
    sumY0  +=       0.943 * cos(3.30460394818  + 3344.36210888011 *T)
    sumY0  +=       0.943 * cos(5.96075562405  + 3337.35037948650 *T)
    sumY0  +=       0.886 * cos(5.16933736517  + 36.84918278771 *T)
    sumY0  +=       0.890 * cos(5.17329864993  + 169.33636564950 *T)
    sumY0  +=       0.925 * cos(1.82710060728  + 1486.22393854870 *T)
    sumY0  +=       1.156 * cos(6.14753394741  + 16061.94070863329 *T)
    sumY0  +=       0.879 * cos(4.09361898695  + 3355.08333818490 *T)
    sumY0  +=       0.965 * cos(2.78306913812  + 11614.67711121570 *T)
    sumY0  +=       0.874 * cos(5.05469151754  + 316.14805217309 *T)
    sumY0  +=       0.896 * cos(2.06210457665  + 2067.68758358909 *T)
    sumY0  +=       0.856 * cos(0.80926398517  + 6705.34710863090 *T)
    sumY0  +=       1.200 * cos(3.92852893712  + 2971.15642824290 *T)
    sumY0  +=       1.091 * cos(1.08913615837  + 9168.39708086389 *T)
    sumY0  +=       0.871 * cos(5.25187070471  + 9380.71585523369 *T)
    sumY0  +=       1.070 * cos(0.04151072339  + 426.35437339250 *T)
    sumY0  +=       1.039 * cos(0.35358307489  + 4937.04268646790 *T)
    sumY0  +=       0.968 * cos(5.15755434693  + 8959.17604904470 *T)
    sumY0  +=       0.852 * cos(4.76295695220  + 8013.52355842391 *T)
    sumY0  +=       0.868 * cos(2.65646874925  + 9947.29949901560 *T)
    sumY0  +=       0.890 * cos(4.75713511878  + 6747.95607979750 *T)
    sumY0  +=       0.834 * cos(0.81428487363  + 6681.04859478950 *T)
    sumY0  +=       1.025 * cos(3.72815532687  + 3914.71340755109 *T)
    sumY0  +=       0.838 * cos(5.14987130454  + 4672.42349675709 *T)
    sumY0  +=       0.872 * cos(3.76405979215  + 11780.73417599970 *T)
    sumY0  +=       0.998 * cos(0.20481558928  + 16335.59398705309 *T)
    sumY0  +=       0.814 * cos(3.39899123402  + 16703.28869733031 *T)
    sumY0  +=       1.043 * cos(2.49679055688  + 3492.75352526850 *T)
    sumY0  +=       0.878 * cos(1.88614535376  + 3233.19272024471 *T)
    sumY0  +=       0.814 * cos(5.42976286552  + 16703.32320463470 *T)
    sumY0  +=       0.821 * cos(6.23923688826  + 3323.04372206531 *T)
    sumY0  +=       0.821 * cos(6.16771533756  + 3358.66876630130 *T)
    sumY0  +=       0.861 * cos(1.36508818725  + 170.42905313570 *T)
    sumY0  +=       0.895 * cos(4.50998682743  + 8966.22228574270 *T)
    sumY0  +=       0.809 * cos(3.69124787220  + 1045.39865367110 *T)
    sumY0  +=       0.986 * cos(1.91297797873  + 9654.36913365350 *T)
    sumY0  +=       0.841 * cos(5.34788356637  + 13745.10242153889 *T)
    sumY0  +=       0.792 * cos(5.67453205333  + 7100.95331475830 *T)
    sumY0  +=       0.779 * cos(1.71726390662  + 3499.79976196650 *T)
    sumY0  +=       0.841 * cos(4.23757581300  + 3448.51976812190 *T)
    sumY0  +=       0.788 * cos(4.57984307584  + 14422.07545447190 *T)
    sumY0  +=       1.013 * cos(4.37752826226  + 8859.60639317511 *T)
    sumY0  +=       1.009 * cos(5.94344429551  + 6657.59023313531 *T)
    sumY0  +=       1.030 * cos(2.19490146324  + 3710.55606012370 *T)
    sumY0  +=       0.787 * cos(3.72966883127  + 3308.88393836651 *T)
    sumY0  +=       0.787 * cos(2.39409808737  + 3372.82855000010 *T)
    sumY0  +=       0.731 * cos(4.80001676808  + 2803.56409712089 *T)
    sumY0  +=       0.725 * cos(4.63534003691  + 3229.66960189570 *T)
    sumY0  +=       0.749 * cos(3.40695752675  + 4356.51926206750 *T)
    sumY0  +=       0.720 * cos(4.72118387969  + 3553.66770465429 *T)
    sumY0  +=       0.818 * cos(4.94156802757  + 802.12010496270 *T)
    sumY0  +=       0.825 * cos(4.27028104649  + 3253.54803964350 *T)
    sumY0  +=       0.793 * cos(1.10166913023  + 11.28951774740 *T)
    sumY0  +=       0.864 * cos(1.98924766986  + 6717.49653756051 *T)
    sumY0  +=       0.969 * cos(3.71171404987  + 20198.85114214949 *T)
    sumY0  +=       0.917 * cos(6.26799705299  + 4776.00390594270 *T)
    sumY0  +=       0.729 * cos(0.68510465497  + 22743.16556203289 *T)
    sumY0  +=       0.730 * cos(3.80010308022  + 1063.14886602171 *T)
    sumY0  +=       0.860 * cos(4.68558303874  + 242.48478649050 *T)
    sumY0  +=       0.713 * cos(5.54624640151  + 20597.00014555770 *T)
    sumY0  +=       0.694 * cos(1.48839766697  + 3452.04288647091 *T)
    sumY0  +=       0.694 * cos(2.41729529361  + 20426.81490990550 *T)
    sumY0  +=       0.972 * cos(0.41619109344  + 13916.26292712510 *T)
    sumY0  +=       0.692 * cos(2.36886353379  + 3068.18328683170 *T)
    sumY0  +=       0.730 * cos(1.00717600200  + 6016.71262575310 *T)
    sumY0  +=       0.795 * cos(1.85352942088  + 3428.16444872311 *T)
    sumY0  +=       0.686 * cos(3.19891856733  + 7255.81346921790 *T)
    sumY0  +=       0.712 * cos(4.03082610800  + 4143.22016662950 *T)
    sumY0  +=       0.727 * cos(1.63727295965  + 2301.82963339290 *T)
    sumY0  +=       0.836 * cos(2.10217259723  + 5642.44206009270 *T)
    sumY0  +=       0.672 * cos(6.19148373326  + 2541.03511901791 *T)
    sumY0  +=       0.725 * cos(2.24638635812  + 15907.00794416610 *T)
    sumY0  +=       0.723 * cos(3.92559293798  + 3735.48212924251 *T)
    sumY0  +=       0.659 * cos(3.53381720449  + 3929.43343622450 *T)
    sumY0  +=       0.752 * cos(0.95478219114  + 401.91593924071 *T)
    sumY0  +=       0.790 * cos(1.64321327327  + 10264.80970155691 *T)
    sumY0  +=       0.833 * cos(4.84897766851  + 6040.10342853389 *T)
    sumY0  +=       0.652 * cos(1.55854352865  + 3212.90471114871 *T)
    sumY0  +=       0.634 * cos(1.60225347298  + 17101.45495439070 *T)
    sumY0  +=       0.636 * cos(1.39983508111  + 17468.61138046190 *T)
    sumY0  +=       0.831 * cos(1.33700488395  + 2409.49315733150 *T)
    sumY0  +=       0.670 * cos(1.27561899722  + 15663.79170522509 *T)
    sumY0  +=       0.636 * cos(2.42884882731  + 5938.47861035050 *T)
    sumY0  +=       0.767 * cos(2.91814761459  + 16702.81831601549 *T)
    sumY0  +=       0.649 * cos(4.45940266951  + 485.72823530609 *T)
    sumY0  +=       0.720 * cos(2.60100369454  + 316.63568714010 *T)
    sumY0  +=       0.685 * cos(2.60603829320  + 1437.41943168211 *T)
    sumY0  +=       0.874 * cos(0.29522273247  + 3264.10253794069 *T)
    sumY0  +=       0.631 * cos(2.35025144285  + 3351.90194444720 *T)
    sumY0  +=       0.631 * cos(3.77351547579  + 3329.81054391940 *T)
    sumY0  +=       0.617 * cos(3.73864973279  + 3613.52920153491 *T)
    sumY0  +=       0.631 * cos(4.86274156896  + 3343.39904146410 *T)
    sumY0  +=       0.834 * cos(5.98573063128  + 6681.88874697671 *T)
    sumY0  +=       0.618 * cos(0.72391686286  + 6357.96325422570 *T)
    sumY0  +=       0.620 * cos(5.79698007436  + 2301.34199842589 *T)
    sumY0  +=       0.619 * cos(4.56320107291  + 3468.80777721790 *T)
    sumY0  +=       0.856 * cos(4.36779410619  + 6254.38284504009 *T)
    sumY0  +=       0.779 * cos(4.63361737844  + 22325.14887419290 *T)
    sumY0  +=       0.621 * cos(2.12407908062  + 22345.01655862470 *T)
    sumY0  +=       0.683 * cos(1.63122853655  + 3344.78839744640 *T)
    sumY0  +=       0.683 * cos(4.49253838209  + 3336.92409092021 *T)
    sumY0  +=       0.637 * cos(5.57539907470  + 151.29148732640 *T)
    sumY0  +=       0.600 * cos(1.07842896250  + 3338.31344690251 *T)
    sumY0  +=       0.737 * cos(3.73325950825  + 3116.02381351440 *T)
    sumY0  +=       0.843 * cos(2.10475762774  + 24149.83623386150 *T)
    sumY0  +=       0.595 * cos(3.87033280802  + 4039.63975744390 *T)
    sumY0  +=       0.621 * cos(5.11737826788  + 10448.67928845890 *T)
    sumY0  +=       0.649 * cos(4.60002553992  + 148.32254190981 *T)
    sumY0  +=       0.706 * cos(2.94874370922  + 6718.07403618730 *T)
    sumY0  +=       0.615 * cos(0.53197546397  + 1169.83206889211 *T)
    sumY0  +=       0.597 * cos(2.20115039035  + 2494.77340943211 *T)
    sumY0  +=       0.580 * cos(5.84100904608  + 3528.78139194590 *T)
    sumY0  +=       0.680 * cos(4.34833986391  + 1990.50119955750 *T)
    sumY0  +=       0.593 * cos(4.91377983956  + 9491.90249752129 *T)
    sumY0  +=       0.704 * cos(1.04292136685  + 3547.55698147990 *T)
    sumY0  +=       0.602 * cos(4.55116981040  + 6681.89855400980 *T)
    sumY0  +=       0.750 * cos(1.36158314223  + 6020.23574410210 *T)
    sumY0  +=       0.672 * cos(5.83404606009  + 6667.24157688150 *T)
    sumY0  +=       0.691 * cos(5.87625321159  + 6679.98419817480 *T)
    sumY0  +=       0.644 * cos(0.37424036712  + 4290.03185315310 *T)
    sumY0  +=       0.778 * cos(1.28111197179  + 394.86970254271 *T)
    sumY0  +=       0.606 * cos(4.18196277170  + 736.12033101530 *T)
    sumY0  +=       0.620 * cos(0.82186565514  + 1265.81129610991 *T)
    sumY0  +=       0.610 * cos(2.84065225726  + 18451.32236404950 *T)
    sumY0  +=       0.602 * cos(1.33158682290  + 20.11150191529 *T)
    sumY0  +=       0.579 * cos(3.37741350534  + 2391.19300024650 *T)
    sumY0  +=       0.574 * cos(2.39068880559  + 9381.69617630189 *T)
    sumY0  +=       0.585 * cos(0.31150564344  + 2170.78035780770 *T)
    sumY0  +=       0.610 * cos(5.23035161420  + 9911.87477636350 *T)
    sumY0  +=       0.648 * cos(1.07812115183  + 4133.63113265070 *T)
    sumY0  +=       0.583 * cos(2.93573448932  + 8226.82265386190 *T)
    sumY0  +=       0.602 * cos(2.63811772346  + 2854.88419139370 *T)
    sumY0  +=       0.555 * cos(1.28385325545  + 16865.28495214769 *T)
    sumY0  +=       0.600 * cos(3.16504307436  + 6725.18758318811 *T)
    sumY0  +=       0.766 * cos(6.18120893749  + 4775.51627097570 *T)
    sumY0  +=       0.577 * cos(6.22738837569  + 3184.94820978210 *T)
    sumY0  +=       0.587 * cos(4.00048217600  + 3194.04455765970 *T)
    sumY0  +=       0.613 * cos(2.32601770677  + 6549.92670919670 *T)
    sumY0  +=       0.696 * cos(1.72938927515  + 19004.40413192490 *T)
    sumY0  +=       0.634 * cos(2.56924738035  + 6510.79580026390 *T)
    sumY0  +=       0.634 * cos(5.87239329301  + 52176.05010063190 *T)
    sumY0  +=       0.603 * cos(0.76915482449  + 7896.20369020371 *T)
    sumY0  +=       0.658 * cos(4.03985639725  + 4988.81536641370 *T)
    sumY0  +=       0.602 * cos(2.24884569451  + 6681.03878775640 *T)
    sumY0  +=       0.520 * cos(4.39804549459  + 6467.68194047809 *T)
    sumY0  +=       0.522 * cos(3.85398776894  + 3339.67560153121 *T)
    sumY0  +=       0.522 * cos(2.26977914970  + 3342.03688683541 *T)
    sumY0  +=       0.528 * cos(6.13807593625  + 109.96250373590 *T)
    sumY0  +=       0.518 * cos(1.01552893939  + 1588.82907780029 *T)
    sumY0  +=       0.553 * cos(2.12188487764  + 3487.66793070691 *T)
    sumY0  +=       0.513 * cos(5.35880279382  + 299.37021175271 *T)
    sumY0  +=       0.715 * cos(6.24466223093  + 4503.33094859110 *T)
    sumY0  +=       0.511 * cos(1.27403519504  + 16490.00685554450 *T)
    sumY0  +=       0.505 * cos(3.26917825567  + 1063.55790093581 *T)
    sumY0  +=       0.698 * cos(4.50778101632  + 43.96272978850 *T)
    sumY0  +=       0.609 * cos(5.07257884137  + 3134.15550688670 *T)
    sumY0  +=       0.645 * cos(0.91039586800  + 1641.08939271309 *T)
    sumY0  +=       0.542 * cos(2.95381309465  + 1596.36294946531 *T)
    sumY0  +=       0.502 * cos(0.81770246890  + 12323.17927852529 *T)
    sumY0  +=       0.569 * cos(2.81477259914  + 7380.25218414370 *T)
    sumY0  +=       0.530 * cos(5.77909278402  + 1699.52303398670 *T)
    sumY0  +=       0.491 * cos(3.81613403345  + 11140.83688968570 *T)
    sumY0  +=       0.518 * cos(0.78407846223  + 3443.46138343489 *T)
    sumY0  +=       0.545 * cos(0.63879177759  + 2288.58786099491 *T)
    sumY0  +=       0.509 * cos(4.40724325660  + 6813.01063256950 *T)
    sumY0  +=       0.489 * cos(1.49622890471  + 13575.99261972070 *T)
    sumY0  +=       0.486 * cos(5.88242172218  + 2274.36076699330 *T)
    sumY0  +=       0.579 * cos(2.61247079255  + 10177.50149701710 *T)
    sumY0  +=       0.535 * cos(3.39746553760  + 9072.90548861311 *T)
    sumY0  +=       0.536 * cos(4.94179821214  + 3198.67761714710 *T)
    sumY0  +=       0.478 * cos(4.20267589342  + 114.64292439690 *T)
    sumY0  +=       0.619 * cos(3.39725211788  + 13207.27312484851 *T)
    sumY0  +=       0.527 * cos(1.68494996905  + 3392.13658204571 *T)
    sumY0  +=       0.527 * cos(4.43881694959  + 3289.57590632090 *T)
    sumY0  +=       0.491 * cos(6.28008678733  + 3393.11655909610 *T)
    sumY0  +=       0.491 * cos(6.12686543849  + 3288.59592927050 *T)
    sumY0  +=       0.518 * cos(3.49246604080  + 3826.82829697291 *T)
    sumY0  +=       0.532 * cos(0.09507953177  + 1055.69359440960 *T)
    sumY0  +=       0.475 * cos(0.78614810441  + 8116.61633264250 *T)
    sumY0  +=       0.473 * cos(1.65508157221  + 4297.14540015391 *T)
    sumY0  +=       0.483 * cos(2.01690193894  + 131.29814420289 *T)
    sumY0  +=       0.483 * cos(3.77260927689  + 3346.92215981310 *T)
    sumY0  +=       0.483 * cos(2.35115764175  + 3334.79032855351 *T)
    sumY0  +=       0.460 * cos(2.87268342392  + 2281.00393267929 *T)
    sumY0  +=       0.460 * cos(0.84191179243  + 2280.96942537490 *T)
    sumY0  +=       0.488 * cos(4.31100288665  + 5195.48854981791 *T)
    sumY0  +=       0.604 * cos(5.04421729659  + 2373.88536674770 *T)
    sumY0  +=       0.491 * cos(4.01235585068  + 11251.04321090511 *T)
    sumY0  +=       0.482 * cos(4.40675290422  + 23.63462026429 *T)
    sumY0  +=       0.455 * cos(2.71373528596  + 8962.71642104590 *T)
    sumY0  +=       0.455 * cos(0.68296365446  + 8962.68191374151 *T)
    sumY0  +=       0.455 * cos(1.64431901546  + 5408.78764525591 *T)
    sumY0  +=       0.458 * cos(2.77929810066  + 6313.75670695370 *T)
    sumY0  +=       0.525 * cos(5.13528927698  + 3535.82762864390 *T)
    sumY0  +=       0.506 * cos(1.18176198510  + 3483.03487121951 *T)
    sumY0  +=       0.453 * cos(2.02140364445  + 2957.97696561230 *T)
    sumY0  +=       0.633 * cos(5.05095677926  + 2963.48263626750 *T)
    sumY0  +=       0.532 * cos(3.95191769037  + 11773.13299403189 *T)
    sumY0  +=       0.502 * cos(3.05311243509  + 5562.71287878030 *T)
    sumY0  +=       0.487 * cos(4.09057830717  + 2548.08135571590 *T)
    sumY0  +=       0.604 * cos(3.35425011302  + 3311.36442587990 *T)
    sumY0  +=       0.604 * cos(2.76951680562  + 3370.34806248671 *T)
    sumY0  +=       0.448 * cos(3.40867489514  + 13361.71320321450 *T)
    sumY0  +=       0.455 * cos(0.48237701031  + 227.71995027250 *T)
    sumY0  +=       0.500 * cos(0.96740158580  + 4819.72281824770 *T)
    sumY0  +=       0.452 * cos(4.62081033720  + 6851.04885401611 *T)
    sumY0  +=       0.503 * cos(4.80675483426  + 382.63546108750 *T)
    sumY0  +=       0.440 * cos(3.12228465063  + 3565.20103988520 *T)
    sumY0  +=       0.607 * cos(4.73491284511  + 5195.00091485090 *T)
    sumY0  +=       0.477 * cos(4.42588307343  + 4379.88285497370 *T)
    sumY0  +=       0.468 * cos(3.75342014790  + 4186.93907893450 *T)
    sumY0  +=       0.474 * cos(3.15640219800  + 2295.21377302870 *T)
    sumY0  +=       0.430 * cos(0.16985042547  + 4716.63004402911 *T)
    sumY0  +=       0.536 * cos(1.82311510964  + 860.55374623631 *T)
    sumY0  +=       0.561 * cos(4.78984838350  + 4272.21933103511 *T)
    sumY0  +=       0.427 * cos(2.36245310678  + 1575.08961876571 *T)
    sumY0  +=       0.456 * cos(1.82684055018  + 10006.82862639790 *T)
    sumY0  +=       0.455 * cos(5.22517487437  + 1854.38848815109 *T)
    sumY0  +=       0.421 * cos(1.90113123366  + 5753.62870238030 *T)
    sumY0  +=       0.539 * cos(0.59511976462  + 3185.50315505190 *T)
    sumY0  +=       0.440 * cos(1.27528067077  + 5490.54477900750 *T)
    sumY0  +=       0.411 * cos(5.01909523285  + 2480.05868046350 *T)
    sumY0  +=       0.433 * cos(2.49733426085  + 3376.41897865191 *T)
    sumY0  +=       0.433 * cos(3.62643265779  + 3305.29350971470 *T)
    sumY0  +=       0.448 * cos(2.04099456040  + 8955.09798542589 *T)
    sumY0  +=       0.416 * cos(1.09710711150  + 23539.46356884929 *T)
    sumY0  +=       0.423 * cos(5.65906920843  + 10037.33356876791 *T)
    sumY0  +=       0.481 * cos(5.74335395262  + 6685.40082414620 *T)
    sumY0  +=       0.538 * cos(3.23793676740  + 6637.74975857810 *T)
    sumY0  +=       0.403 * cos(2.02612411802  + 394.38206757569 *T)
    sumY0  +=       0.414 * cos(1.17693488584  + 3016.86319255889 *T)
    sumY0  +=       0.402 * cos(0.22194565664  + 6151.79495944070 *T)
    sumY0  +=       0.403 * cos(1.22347336809  + 18849.47136745770 *T)
    sumY0  +=       0.402 * cos(4.47435933232  + 6151.76045213631 *T)
    sumY0  +=       0.430 * cos(0.28363407761  + 2655.81218995730 *T)
    sumY0  +=       0.435 * cos(3.75918117164  + 9375.07026426810 *T)
    sumY0  +=       0.509 * cos(5.52869427682  + 3496.20933331471 *T)
    sumY0  +=       0.548 * cos(3.91428551746  + 10184.54773371510 *T)
    sumY0  +=       0.397 * cos(4.10299095794  + 3723.73552275430 *T)
    sumY0  +=       0.406 * cos(1.72717369149  + 5750.10558403130 *T)
    sumY0  +=       0.477 * cos(2.15054075988  + 6614.98126196870 *T)
    sumY0  +=       0.555 * cos(1.07130235791  + 3718.22985209910 *T)
    sumY0  +=       0.425 * cos(4.00063826739  + 7218.27318297850 *T)
    sumY0  +=       0.408 * cos(4.56425141121  + 2149.44471734070 *T)
    sumY0  +=       0.402 * cos(2.74166202579  + 1685.29593998510 *T)
    sumY0  +=       0.411 * cos(2.85798674880  + 21265.76694400370 *T)
    sumY0  +=       0.486 * cos(3.27620484219  + 10050.53028505070 *T)
    sumY0  +=       0.394 * cos(6.14977322791  + 5888.69378241570 *T)
    sumY0  +=       0.487 * cos(3.87437323970  + 1655.31648671470 *T)
    sumY0  +=       0.421 * cos(1.96410911700  + 13518.11392371690 *T)
    sumY0  +=       0.395 * cos(5.59740113398  + 7366.02509014209 *T)
    sumY0  +=       0.392 * cos(5.68366955363  + 17655.02435723310 *T)
    sumY0  +=       0.443 * cos(5.46161131426  + 2075.28876555691 *T)
    sumY0  +=       0.404 * cos(3.65562985264  + 6644.86330557891 *T)
    sumY0  +=       0.381 * cos(0.21743768465  + 15805.90302195849 *T)
    sumY0  +=       0.414 * cos(0.69800253256  + 7747.96414807311 *T)
    sumY0  +=       0.395 * cos(3.07042234810  + 18606.25512851669 *T)
    sumY0  +=       0.377 * cos(4.47773098550  + 8564.55016398550 *T)
    sumY0  +=       0.475 * cos(4.76974627664  + 7380.73981911070 *T)
    sumY0  +=       0.378 * cos(1.07675159263  + 9801.66845514410 *T)
    sumY0  +=       0.430 * cos(1.20468421989  + 905.21287918130 *T)
    sumY0  +=       0.384 * cos(2.67566931580  + 3205.30352918089 *T)
    sumY0  +=       0.444 * cos(3.40061961754  + 2277.05452576400 *T)
    sumY0  +=       0.397 * cos(5.66561964068  + 515.70768857651 *T)
    sumY0  +=       0.471 * cos(4.30501104270  + 7270.53349789130 *T)
    sumY0  +=       0.367 * cos(3.33137641644  + 912.32642618209 *T)
    sumY0  +=       0.430 * cos(0.32262969033  + 5881.64754571770 *T)
    sumY0  +=       0.469 * cos(2.72446532167  + 7270.04586292429 *T)
    sumY0  +=       0.433 * cos(0.33076022458  + 2284.91883229020 *T)
    sumY0  +=       0.438 * cos(3.55000511645  + 9797.73630188101 *T)
    sumY0  +=       0.506 * cos(4.47192742031  + 1765.52280793410 *T)
    sumY0  +=       0.360 * cos(1.93729428971  + 20736.07597890910 *T)
    sumY0  +=       0.395 * cos(4.03209674497  + 7697.13168876730 *T)
    sumY0  +=       0.364 * cos(1.26222637443  + 3341.34893028450 *T)
    sumY0  +=       0.379 * cos(6.25338112517  + 5641.95442512570 *T)
    sumY0  +=       0.476 * cos(2.98235844735  + 1329.75590774351 *T)
    sumY0  +=       0.473 * cos(3.88576512895  + 21795.45790909830 *T)
    sumY0  +=       0.351 * cos(4.13570973701  + 6852.14154150230 *T)
    sumY0  +=       0.351 * cos(3.11111844946  + 2547.59372074890 *T)
    sumY0  +=       0.452 * cos(0.50802021686  + 6107.12387254830 *T)
    sumY0  +=       0.344 * cos(5.36725846777  + 16894.75421709410 *T)
    sumY0  +=       0.364 * cos(0.71572042055  + 6682.95314359140 *T)
    sumY0  +=       0.344 * cos(1.77948057276  + 5483.49854230950 *T)
    sumY0  +=       0.372 * cos(3.08550618514  + 1.22413855171 *T)
    sumY0  +=       0.452 * cos(1.07299009125  + 4307.82712161890 *T)
    sumY0  +=       0.349 * cos(2.40546715957  + 5636.31383469550 *T)
    sumY0  +=       0.362 * cos(5.00528267092  + 27681.89692667289 *T)
    sumY0  +=       0.376 * cos(6.22222472073  + 3339.38828814809 *T)
    sumY0  +=       0.347 * cos(2.47924829765  + 25685.62898532449 *T)
    sumY0  +=       0.351 * cos(2.80467267603  + 5452.50655756091 *T)
    sumY0  +=       0.348 * cos(1.66710324635  + 16865.77258711470 *T)
    sumY0  +=       0.345 * cos(3.95198013706  + 25287.47998191629 *T)
    sumY0  +=       0.358 * cos(3.05433541689  + 14584.05445563709 *T)
    sumY0  +=       0.336 * cos(4.55900012579  + 1882.83767561889 *T)
    sumY0  +=       0.436 * cos(3.55215974621  + 21946.86755521649 *T)
    sumY0  +=       0.434 * cos(0.32562428989  + 7214.68275432671 *T)
    sumY0  +=       0.395 * cos(4.16776570158  + 9993.63191011510 *T)
    sumY0  +=       0.351 * cos(0.12634120978  + 13213.13032714810 *T)
    sumY0  +=       0.327 * cos(4.04880436765  + 3354.09801658130 *T)
    sumY0  +=       0.327 * cos(2.07496255099  + 3327.61447178531 *T)
    sumY0  +=       0.361 * cos(0.39938152102  + 3.34661116829 *T)
    sumY0  +=       0.328 * cos(6.15068258009  + 2435.39954751850 *T)
    sumY0  +=       0.447 * cos(4.74470130101  + 13363.67384535091 *T)
    sumY0  +=       0.389 * cos(1.05447612688  + 30065.75565778170 *T)
    sumY0  +=       0.326 * cos(5.67270447514  + 3264.59017290771 *T)
    sumY0  +=       0.395 * cos(1.75358681150  + 3337.40043613710 *T)
    sumY0  +=       0.339 * cos(4.51134098176  + 14557.14053450730 *T)
    sumY0  +=       0.334 * cos(4.48543603219  + 147.05550400710 *T)
    sumY0  +=       0.335 * cos(2.36305281420  + 20617.77554105009 *T)
    sumY0  +=       0.347 * cos(4.05789606388  + 5223.93773728571 *T)
    sumY0  +=       0.344 * cos(5.86233733820  + 4025.90029840930 *T)
    sumY0  +=       0.314 * cos(4.43830632328  + 4142.73253166250 *T)
    sumY0  +=       0.347 * cos(5.76229830621  + 3265.58701064900 *T)
    sumY0  +=       0.408 * cos(6.10494233727  + 11883.82695021830 *T)
    sumY0  +=       0.317 * cos(2.10770348226  + 3347.48215621710 *T)
    sumY0  +=       0.335 * cos(1.82753658277  + 544.16187657970 *T)
    sumY0  +=       0.345 * cos(3.53624871536  + 5244.29305668450 *T)
    sumY0  +=       0.318 * cos(1.06713363481  + 12302.82395912650 *T)
    sumY0  +=       0.384 * cos(4.30168263627  + 7733.73705407150 *T)
    sumY0  +=       0.402 * cos(0.39385834298  + 8982.56685182550 *T)
    sumY0  +=       0.409 * cos(5.51884193886  + 23141.31456544110 *T)
    sumY0  +=       0.335 * cos(2.75373542682  + 3061.06973983091 *T)
    sumY0  +=       0.415 * cos(3.99882051146  + 12509.49714253070 *T)
    sumY0  +=       0.355 * cos(6.17230823937  + 7322.34627826531 *T)
    sumY0  +=       0.316 * cos(5.59149424389  + 3440.42590005290 *T)
    sumY0  +=       0.316 * cos(0.53227267475  + 3241.28658831371 *T)
    sumY0  +=       0.354 * cos(0.64235154390  + 4606.42372280970 *T)
    sumY0  +=       0.347 * cos(1.05728075170  + 14127.99895376209 *T)
    sumY0  +=       0.310 * cos(5.64292624067  + 6453.99253809410 *T)
    sumY0  +=       0.300 * cos(2.21739482474  + 1987.46571617550 *T)
    sumY0  +=       0.304 * cos(2.64756651462  + 15265.64270181689 *T)
    sumY0  +=       0.299 * cos(4.05929485456  + 10285.16502095570 *T)
    sumY0  +=       0.379 * cos(2.28257510001  + 1461.29786942991 *T)
    sumY0  +=       0.294 * cos(1.92173439518  + 20040.39525933330 *T)
    sumY0  +=       0.295 * cos(3.55293055978  + 2487.65986243130 *T)
    sumY0  +=       0.313 * cos(1.30385633562  + 6286.84278582391 *T)
    sumY0  +=       0.307 * cos(6.22295802105  + 9755.47405586110 *T)
    sumY0  +=       0.306 * cos(4.47605572395  + 8756.51361895650 *T)
    sumY0  +=       0.313 * cos(0.32862784398  + 4982.18945437991 *T)
    sumY0  +=       0.392 * cos(1.48237112438  + 19645.76937427410 *T)
    sumY0  +=       0.361 * cos(1.46103489347  + 684.80023674250 *T)
    sumY0  +=       0.328 * cos(4.08465891668  + 3284.05298201350 *T)
    sumY0  +=       0.328 * cos(2.03910800196  + 3397.65950635310 *T)
    sumY0  +=       0.291 * cos(0.90669086212  + 11986.91972443691 *T)
    sumY0  +=       0.341 * cos(4.17468651117  + 18053.17336064130 *T)
    sumY0  +=       0.345 * cos(3.44122421521  + 16063.40866466850 *T)
    sumY0  +=       0.314 * cos(2.64774700056  + 9983.94806194510 *T)
    sumY0  +=       0.331 * cos(1.83529030284  + 6680.99828956829 *T)
    sumY0  +=       0.329 * cos(2.14259666141  + 2277.95119564511 *T)
    sumY0  +=       0.301 * cos(2.60539384966  + 1251.07401378430 *T)
    sumY0  +=       0.339 * cos(5.86444938845  + 2825.39237309031 *T)
    sumY0  +=       0.339 * cos(0.25931753019  + 3856.32011527630 *T)
    sumY0  +=       0.383 * cos(3.28646629882  + 2921.93184528270 *T)
    sumY0  +=       0.302 * cos(4.92028257129  + 12825.88901218730 *T)
    sumY0  +=       0.310 * cos(0.61442486331  + 2060.57403658830 *T)
    sumY0  +=       0.310 * cos(2.29827810873  + 5622.10399434610 *T)
    sumY0  +=       0.292 * cos(0.82350537881  + 17256.87535382490 *T)
    sumY0  +=       0.286 * cos(4.10458531999  + 279.54268686890 *T)
    sumY0  +=       0.330 * cos(6.18523353093  + 1073.85284167431 *T)
    sumY0  +=       0.382 * cos(1.17991122952  + 3267.55911832430 *T)
    sumY0  +=       0.382 * cos(4.94385568912  + 3414.15337004230 *T)
    sumY0  +=       0.304 * cos(3.37215064995  + 3620.64274853570 *T)
    sumY0  +=       0.326 * cos(5.37060215981  + 533.86693584120 *T)
    sumY0  +=       0.322 * cos(2.89731305468  + 2388.65020296570 *T)
    sumY0  +=       0.289 * cos(1.22024545508  + 36.27168416091 *T)
    sumY0  +=       0.287 * cos(5.56455691642  + 5511.88041947450 *T)
    sumY0  +=       0.283 * cos(1.21251617045  + 12840.11610618890 *T)
    sumY0  +=       0.358 * cos(6.21940014383  + 568.57805654390 *T)
    sumY0  +=       0.282 * cos(3.87783929607  + 595.93455652550 *T)
    sumY0  +=       0.380 * cos(1.92054508917  + 2973.63191522091 *T)
    sumY0  +=       0.283 * cos(4.53197273536  + 9859.61446145070 *T)
    sumY0  +=       0.269 * cos(0.82453120759  + 15636.81047379250 *T)
    sumY0  +=       0.280 * cos(0.42006444847  + 16511.85768487090 *T)
    sumY0  +=       0.296 * cos(0.76548268882  + 1.24065522479 *T)
    sumY0  +=       0.300 * cos(2.59187768206  + 13525.16016041490 *T)
    sumY0  +=       0.331 * cos(6.08770397853  + 6680.96378226390 *T)
    sumY0  +=       0.310 * cos(0.26750647724  + 5622.06948704171 *T)
    sumY0  +=       0.323 * cos(0.38705213166  + 2266.75958502550 *T)
    sumY0  +=       0.284 * cos(3.18215717131  + 5085.34953890130 *T)
    sumY0  +=       0.320 * cos(1.27027711736  + 10720.86461084350 *T)
    sumY0  +=       0.317 * cos(4.93342158277  + 6048.19729660290 *T)
    sumY0  +=       0.269 * cos(2.88951641275  + 2744.67787017430 *T)
    sumY0  +=       0.318 * cos(2.18392341590  + 526.00262931501 *T)
    sumY0  +=       0.257 * cos(4.63373345906  + 24341.28449997309 *T)
    sumY0  +=       0.266 * cos(0.45979732691  + 206.94455478010 *T)
    sumY0  +=       0.258 * cos(1.79534054581  + 8749.40007195570 *T)
    sumY0  +=       0.271 * cos(2.27668459594  + 3345.42699390331 *T)
    sumY0  +=       0.271 * cos(3.84708232270  + 3336.28549446330 *T)
    sumY0  +=       0.301 * cos(4.71479959051  + 10706.63751684189 *T)
    sumY0  +=       0.251 * cos(3.96861026556  + 3343.95928643870 *T)
    sumY0  +=       0.251 * cos(2.15515665308  + 3337.75320192791 *T)
    sumY0  +=       0.251 * cos(3.37159956977  + 9758.99717421010 *T)
    sumY0  +=       0.267 * cos(0.96036007447  + 6831.03186801770 *T)
    sumY0  +=       0.306 * cos(1.36129297092  + 12012.82611462390 *T)
    sumY0  +=       0.249 * cos(6.28103476480  + 4246.31294084810 *T)
    sumY0  +=       0.286 * cos(4.90464379613  + 1898.59503542310 *T)
    sumY0  +=       0.320 * cos(5.23732353360  + 536.56069461190 *T)
    sumY0  +=       0.290 * cos(1.30201044096  + 20047.44149603130 *T)
    sumY0  +=       0.268 * cos(1.47703232305  + 16703.23864067971 *T)
    sumY0  +=       0.250 * cos(0.95355486070  + 9698.57568092550 *T)
    sumY0  +=       0.248 * cos(1.11014026714  + 4534.81561944089 *T)
    sumY0  +=       0.246 * cos(3.89895295515  + 15651.03756779411 *T)
    sumY0  +=       0.309 * cos(0.93773869004  + 1903.19299501769 *T)
    sumY0  +=       0.247 * cos(5.73950808020  + 10016.55817327550 *T)
    sumY0  +=       0.262 * cos(0.38159181599  + 4025.41266344230 *T)
    sumY0  +=       0.312 * cos(2.02014523279  + 6033.97020260130 *T)
    sumY0  +=       0.245 * cos(2.61302130597  + 2939.18412242610 *T)
    sumY0  +=       0.317 * cos(3.62639597567  + 95.73540973430 *T)
    sumY0  +=       0.256 * cos(1.57197883241  + 1727.90491115171 *T)
    sumY0  +=       0.330 * cos(2.97758477159  + 735.63269604830 *T)
    sumY0  +=       0.276 * cos(4.96569192497  + 6670.42297061920 *T)
    sumY0  +=       0.321 * cos(5.04537384488  + 11905.16259068530 *T)
    sumY0  +=       0.252 * cos(3.55325819855  + 5430.63847458231 *T)
    sumY0  +=       0.252 * cos(4.89937019626  + 21548.71855180830 *T)
    sumY0  +=       0.237 * cos(2.58555430531  + 194.72756697710 *T)
    sumY0  +=       0.258 * cos(2.68892316362  + 9488.86701413930 *T)
    sumY0  +=       0.269 * cos(1.83869068629  + 3867.02409092890 *T)
    sumY0  +=       0.265 * cos(2.90622523535  + 3053.89388306271 *T)
    sumY0  +=       0.240 * cos(1.73397693635  + 10027.60402189030 *T)
    sumY0  +=       0.321 * cos(2.82877270353  + 3759.78064308390 *T)
    sumY0  +=       0.319 * cos(3.59115855791  + 9815.89554914570 *T)
    sumY0  +=       0.251 * cos(0.12299632037  + 1280.03839011150 *T)
    sumY0  +=       0.239 * cos(2.92878889718  + 26481.92699214089 *T)
    sumY0  +=       0.251 * cos(5.02028762471  + 2790.80995968991 *T)
    sumY0  +=       0.326 * cos(4.20511206757  + 3708.08057314570 *T)
    sumY0  +=       0.235 * cos(3.75339605407  + 5106.62286960090 *T)
    sumY0  +=       0.293 * cos(0.11701788048  + 5628.71265272770 *T)
    sumY0  +=       0.323 * cos(4.87018010930  + 17086.20248320570 *T)
    sumY0  +=       0.311 * cos(0.12778675025  + 3290.02378487750 *T)
    sumY0  +=       0.311 * cos(5.99598016839  + 3391.68870348911 *T)
    sumY0  +=       0.242 * cos(3.46289647182  + 4996.41654838150 *T)
    sumY0  +=       0.233 * cos(5.27407567581  + 2679.13568243529 *T)
    sumY0  +=       0.230 * cos(1.62592473181  + 10654.86483689610 *T)
    sumY0  +=       0.292 * cos(5.65394034305  + 3050.37076471370 *T)
    sumY0  +=       0.230 * cos(0.59794722074  + 1744.24947723449 *T)
    sumY0  +=       0.235 * cos(2.10323754397  + 5945.59215735130 *T)
    sumY0  +=       0.301 * cos(2.35280602608  + 9495.91325083730 *T)
    sumY0  +=       0.318 * cos(1.27137347197  + 2907.14450630650 *T)
    sumY0  +=       0.229 * cos(4.79342389836  + 5.27910682390 *T)
    sumY0  +=       0.248 * cos(1.94277588124  + 1442.01739127670 *T)
    sumY0  +=       0.226 * cos(5.71133491731  + 1868.61558215270 *T)
    sumY0  +=       0.234 * cos(5.14378230022  + 2817.79119112250 *T)
    sumY0  +=       0.230 * cos(0.30903741658  + 4245.82530588109 *T)
    sumY0  +=       0.312 * cos(4.22313838923  + 4403.76129272151 *T)
    sumY0  +=       0.222 * cos(1.20476435440  + 1835.56843509031 *T)
    sumY0  +=       0.277 * cos(4.35456944772  + 7079.13003932430 *T)
    sumY0  +=       0.289 * cos(0.65812694051  + 6901.88131332190 *T)
    sumY0  +=       0.244 * cos(2.21534433966  + 5349.41378334230 *T)
    sumY0  +=       0.268 * cos(1.06853593620  + 16703.37326128530 *T)
    sumY0  +=       0.250 * cos(3.14710986353  + 187.68133027909 *T)
    sumY0  +=       0.220 * cos(2.74511541811  + 6511.88848775011 *T)
    sumY0  +=       0.224 * cos(6.18440878701  + 3138.60284900920 *T)
    sumY0  +=       0.224 * cos(6.22254343881  + 3543.10963935740 *T)
    sumY0  +=       0.234 * cos(0.06073372993  + 17499.60395779890 *T)
    sumY0  +=       0.256 * cos(4.03736586899  + 367.46814644590 *T)
    sumY0  +=       0.272 * cos(3.51718028747  + 2428.28600051770 *T)
    sumY0  +=       0.218 * cos(1.49669828918  + 2786.79920637390 *T)
    sumY0  +=       0.218 * cos(5.88795359173  + 6283.33692112710 *T)
    sumY0  +=       0.270 * cos(0.24554568359  + 11236.81611690350 *T)
    sumY0  +=       0.218 * cos(3.85718196023  + 6283.30241382270 *T)
    sumY0  +=       0.220 * cos(3.19713485655  + 3987.86707749811 *T)
    sumY0  +=       0.238 * cos(5.93039570689  + 2814.68839743771 *T)
    sumY0  +=       0.222 * cos(0.36863892862  + 1368.90407032850 *T)
    sumY0  +=       0.215 * cos(5.15532893112  + 13383.46891977510 *T)
    sumY0  +=       0.224 * cos(2.11557791608  + 77.99436146740 *T)
    sumY0  +=       0.221 * cos(3.24073834180  + 3937.03461819230 *T)
    sumY0  +=       0.226 * cos(2.56787172746  + 9279.09103705030 *T)
    sumY0  +=       0.256 * cos(3.53846540374  + 10596.42589591770 *T)
    sumY0  +=       0.211 * cos(5.04095056660  + 12200.21881987490 *T)
    sumY0  +=       0.212 * cos(0.14549328098  + 2392.66095628171 *T)
    sumY0  +=       0.216 * cos(3.15925655175  + 12928.98178640591 *T)
    sumY0  +=       0.237 * cos(3.21754149899  + 3627.81860530390 *T)
    sumY0  +=       0.220 * cos(0.54824595868  + 6530.42100104020 *T)
    sumY0  +=       0.213 * cos(6.27747523577  + 1484.68867221069 *T)
    sumY0  +=       0.262 * cos(1.78468856574  + 20.53157800890 *T)
    sumY0  +=       0.254 * cos(5.82038398953  + 19146.51544865829 *T)
    sumY0  +=       0.222 * cos(5.95630142513  + 8166.40116057730 *T)
    sumY0  +=       0.263 * cos(5.32304904255  + 4531.78013605890 *T)
    sumY0  +=       0.249 * cos(2.58332253923  + 3084.88586781131 *T)
    sumY0  +=       0.239 * cos(4.03784097782  + 5095.91889394830 *T)
    sumY0  +=       0.239 * cos(4.03784097782  + 5095.43125898130 *T)
    sumY0  +=       0.270 * cos(5.38756181061  + 24889.33097850809 *T)
    sumY0  +=       0.252 * cos(3.08379591840  + 3299.30545319850 *T)
    sumY0  +=       0.252 * cos(3.03997100024  + 3382.40703516810 *T)
    sumY0  +=       0.207 * cos(2.83288102556  + 5618.15458743080 *T)
    sumY0  +=       0.239 * cos(1.63889533266  + 3344.31205222951 *T)
    sumY0  +=       0.262 * cos(0.46943815181  + 3631.34172365290 *T)
    sumY0  +=       0.206 * cos(1.80106246742  + 7083.14079264031 *T)
    sumY0  +=       0.249 * cos(0.73301708654  + 2149.93235230770 *T)
    sumY0  +=       0.210 * cos(5.81059147043  + 6678.71716027210 *T)
    sumY0  +=       0.255 * cos(1.55366452384  + 1994.51195287351 *T)
    sumY0  +=       0.204 * cos(3.91362001245  + 4694.24677219110 *T)
    sumY0  +=       0.274 * cos(4.79397197037  + 12306.83471244250 *T)
    sumY0  +=       0.262 * cos(0.46979570624  + 532.97026596010 *T)
    sumY0  +=       0.214 * cos(4.33780097919  + 9623.44445920770 *T)
    sumY0  +=       0.246 * cos(6.01432530866  + 151.65346360169 *T)
    sumY0  +=       0.223 * cos(2.98859743190  + 3345.21828057310 *T)
    sumY0  +=       0.223 * cos(3.13516948674  + 3336.49420779350 *T)
    sumY0  +=       0.255 * cos(5.28917502113  + 2942.72449442730 *T)
    sumY0  +=       0.224 * cos(1.52423467945  + 6533.38994645680 *T)
    sumY0  +=       0.245 * cos(5.50271392619  + 2142.39848064269 *T)
    sumY0  +=       0.201 * cos(5.35589690383  + 9360.84817080190 *T)
    sumY0  +=       0.282 * cos(1.08698948224  + 12465.29059525869 *T)
    sumY0  +=       0.199 * cos(3.26141559083  + 3030.01544531491 *T)
    sumY0  +=       0.243 * cos(4.23645604083  + 9329.92349635610 *T)
    sumY0  +=       0.233 * cos(6.03479083951  + 5834.89820116490 *T)
    sumY0  +=       0.205 * cos(4.30285175798  + 7167.44072367270 *T)
    sumY0  +=       0.199 * cos(0.26993000031  + 2957.47207699310 *T)
    sumY0  +=       0.220 * cos(6.00177151459  + 5626.01889395700 *T)
    sumY0  +=       0.197 * cos(4.96967459654  + 3244.87701696550 *T)
    sumY0  +=       0.197 * cos(1.15409232210  + 3436.83547140110 *T)
    sumY0  +=       0.200 * cos(1.76538096369  + 6567.06956396970 *T)
    sumY0  +=       0.231 * cos(2.63984097637  + 6654.06711478630 *T)
    sumY0  +=       0.228 * cos(5.88325894864  + 3319.59321372389 *T)
    sumY0  +=       0.206 * cos(4.62181012252  + 7483.34495836230 *T)
    sumY0  +=       0.206 * cos(4.62181012252  + 7483.83259332930 *T)
    sumY0  +=       0.198 * cos(3.92135089796  + 3094.60452186030 *T)
    sumY0  +=       0.215 * cos(4.11109862806  + 2540.54748405090 *T)
    sumY0  +=       0.223 * cos(3.01290266414  + 13341.91812879030 *T)
    sumY0  +=       0.199 * cos(5.43950846422  + 2409.00552236449 *T)
    sumY0  +=       0.192 * cos(0.74649040425  + 6382.34227661390 *T)
    sumY0  +=       0.205 * cos(1.10304172891  + 3890.90252867670 *T)
    sumY0  +=       0.191 * cos(0.61853752731  + 2928.48514730890 *T)
    sumY0  +=       0.191 * cos(5.50522939133  + 3753.22734105771 *T)
    sumY0  +=       0.193 * cos(1.07910361151  + 117.56368570370 *T)
    sumY0  +=       0.194 * cos(4.47708903791  + 18208.10612510850 *T)
    sumY0  +=       0.197 * cos(1.06976954616  + 20206.38501381450 *T)
    sumY0  +=       0.194 * cos(4.32410611780  + 28628.09240861609 *T)
    sumY0  +=       0.212 * cos(4.83851003745  + 4538.82637275690 *T)
    sumY0  +=       0.192 * cos(4.11577520914  + 3365.71500299931 *T)
    sumY0  +=       0.192 * cos(2.00799170950  + 3315.99748536730 *T)
    sumY0  +=       0.218 * cos(1.22674306878  + 2142.88611560970 *T)
    sumY0  +=       0.226 * cos(3.54046676510  + 3596.82662055530 *T)
    sumY0  +=       0.253 * cos(5.60084432635  + 5415.90119225670 *T)
    sumY0  +=       0.187 * cos(0.19756321336  + 15107.11950869790 *T)
    sumY0  +=       0.207 * cos(4.52254146385  + 419.24082639170 *T)
    sumY0  +=       0.255 * cos(3.25840338963  + 2942.68998712290 *T)
    sumY0  +=       0.191 * cos(0.88275042024  + 10544.65851567670 *T)
    sumY0  +=       0.190 * cos(5.78004681928  + 28229.94340520789 *T)
    sumY0  +=       0.205 * cos(0.67775581231  + 1059.60849402051 *T)
    sumY0  +=       0.233 * cos(4.30007127815  + 5821.15874213030 *T)
    sumY0  +=       0.187 * cos(4.96787196527  + 5401.18646328809 *T)
    sumY0  +=       0.188 * cos(5.93437329943  + 3318.08774757391 *T)
    sumY0  +=       0.188 * cos(0.18939361921  + 3363.62474079271 *T)
    sumY0  +=       0.231 * cos(2.52090864586  + 9094.24112908010 *T)
    sumY0  +=       0.194 * cos(0.44841184856  + 2736.00650347850 *T)
    sumY0  +=       0.182 * cos(1.33124042287  + 5960.32943967690 *T)
    sumY0  +=       0.193 * cos(2.43212632306  + 7417.34518441490 *T)
    sumY0  +=       0.186 * cos(5.05072922235  + 11354.13598512370 *T)
    sumY0  +=       0.181 * cos(5.02122429267  + 1754.81883228150 *T)
    sumY0  +=       0.222 * cos(6.00725418804  + 3070.65877380970 *T)
    sumY0  +=       0.181 * cos(2.28299776534  + 4356.03162710049 *T)
    sumY0  +=       0.177 * cos(2.86221640984  + 3651.69704305170 *T)
    sumY0  +=       0.189 * cos(3.37917605540  + 6673.86748891530 *T)
    sumY0  +=       0.179 * cos(3.47998233365  + 15121.34660269950 *T)
    sumY0  +=       0.232 * cos(2.65931483985  + 3517.54997431711 *T)
    sumY0  +=       0.232 * cos(3.46445207879  + 3164.16251404950 *T)
    sumY0  +=       0.217 * cos(1.06468264448  + 220.16882495529 *T)
    sumY0  +=       0.176 * cos(3.13334105652  + 1903.68062998470 *T)
    sumY0  +=       0.180 * cos(0.82368576619  + 3348.95011225231 *T)
    sumY0  +=       0.180 * cos(5.30008115245  + 3332.76237611430 *T)
    sumY0  +=       0.179 * cos(2.20230995382  + 3587.10796650631 *T)
    sumY0  +=       0.205 * cos(2.70852744380  + 1059.64300132490 *T)
    sumY0  +=       0.187 * cos(1.04746838799  + 2619.71701297711 *T)
    sumY0  +=       0.176 * cos(3.90353461991  + 13465.78629850130 *T)
    sumY0  +=       0.242 * cos(2.50347875338  + 5408.30001028890 *T)
    sumY0  +=       0.178 * cos(4.08581305127  + 2973.14428025389 *T)
    sumY0  +=       0.226 * cos(2.22430722385  + 11925.03027511709 *T)
    sumY0  +=       0.174 * cos(4.55814413103  + 10096.86269615020 *T)
    sumY0  +=       0.188 * cos(3.92968463646  + 2575.06317973690 *T)
    sumY0  +=       0.175 * cos(5.69586890519  + 12299.78847574450 *T)
    sumY0  +=       0.232 * cos(3.14822069955  + 5351.95658062310 *T)
    sumY0  +=       0.174 * cos(1.35135226191  + 5243.80542171749 *T)
    sumY0  +=       0.177 * cos(5.85994870849  + 6872.42930202769 *T)
    sumY0  +=       0.177 * cos(5.31676928648  + 2573.00742483470 *T)
    sumY0  +=       0.210 * cos(0.98942403448  + 6684.22018149410 *T)
    sumY0  +=       0.175 * cos(4.13281638704  + 6708.87022697990 *T)
    sumY0  +=       0.174 * cos(4.83669668224  + 3156.00633621310 *T)
    sumY0  +=       0.182 * cos(2.75035780857  + 2011.34415392330 *T)
    sumY0  +=       0.169 * cos(3.54733614920  + 6684.97453557991 *T)
    sumY0  +=       0.173 * cos(4.04338856638  + 4885.72259219509 *T)
    sumY0  +=       0.239 * cos(5.36279832648  + 31022.50935337269 *T)
    sumY0  +=       0.169 * cos(5.57810778070  + 6685.00904288430 *T)
    sumY0  +=       0.191 * cos(3.61411273955  + 11567.43508056170 *T)
    sumY0  +=       0.207 * cos(2.89825036062  + 10001.72577818411 *T)
    sumY0  +=       0.181 * cos(1.80435450081  + 8584.41784841729 *T)
    sumY0  +=       0.189 * cos(5.77298688672  + 2783.76372299191 *T)
    sumY0  +=       0.233 * cos(1.82665762083  + 23958.87560271690 *T)
    sumY0  +=       0.219 * cos(4.30571742396  + 3070.72608411250 *T)
    sumY0  +=       0.177 * cos(3.00247820411  + 2978.73510745250 *T)
    sumY0  +=       0.179 * cos(2.14367659492  + 17232.99691607710 *T)
    sumY0  +=       0.168 * cos(1.23920780852  + 1111.39842761850 *T)
    sumY0  +=       0.172 * cos(5.04827147433  + 4061.99547538951 *T)
    sumY0  +=       0.184 * cos(6.11375809233  + 12943.20888040750 *T)
    sumY0  +=       0.179 * cos(0.02775233109  + 8852.49284617430 *T)
    sumY0  +=       0.162 * cos(0.40163918127  + 16696.19240398170 *T)
    sumY0  +=       0.163 * cos(0.63420278502  + 19676.20641375289 *T)
    sumY0  +=       0.204 * cos(2.58726231867  + 4253.42648784890 *T)
    sumY0  +=       0.200 * cos(0.11645066909  + 3611.05371455690 *T)
    sumY0  +=       0.188 * cos(0.87802326806  + 2385.61471958371 *T)
    sumY0  +=       0.166 * cos(0.44820712961  + 24491.18197509989 *T)
    sumY0  +=       0.168 * cos(0.43137458003  + 6166.00479979010 *T)
    sumY0  +=       0.201 * cos(3.99153837364  + 3738.98799393931 *T)
    sumY0  +=       0.184 * cos(4.49082489107  + 1155.60497489050 *T)
    sumY0  +=       0.163 * cos(5.97441856550  + 3335.65705301750 *T)
    sumY0  +=       0.163 * cos(0.14934835314  + 3346.05543534910 *T)
    sumY0  +=       0.158 * cos(4.96341437900  + 5490.05714404049 *T)
    sumY0  +=       0.162 * cos(5.12671205173  + 272.42913986809 *T)
    sumY0  +=       0.173 * cos(1.70692023003  + 4.17597074660 *T)
    sumY0  +=       0.166 * cos(4.69082020547  + 21393.78578734110 *T)
    sumY0  +=       0.175 * cos(5.91635100503  + 2757.85733280490 *T)
    sumY0  +=       0.154 * cos(5.97630069834  + 4289.05153208490 *T)
    sumY0  +=       0.214 * cos(2.37710402682  + 1386.13909381971 *T)
    sumY0  +=       0.171 * cos(0.07657131354  + 10011.44443223310 *T)
    sumY0  +=       0.156 * cos(4.08867905453  + 10081.45495949650 *T)
    sumY0  +=       0.163 * cos(0.47112906830  + 13286.42745300710 *T)
    sumY0  +=       0.155 * cos(3.98476507217  + 8771.25090128210 *T)
    sumY0  +=       0.151 * cos(4.45789668424  + 15011.14028148010 *T)
    sumY0  +=       0.155 * cos(1.22574828411  + 1879.31455726989 *T)
    sumY0  +=       0.155 * cos(4.17166988351  + 2089.53841291549 *T)
    sumY0  +=       0.194 * cos(1.66101338255  + 3361.14400470870 *T)
    sumY0  +=       0.169 * cos(4.62246895595  + 7064.34794945411 *T)
    sumY0  +=       0.166 * cos(5.48604111479  + 6719.60170652090 *T)
    sumY0  +=       0.158 * cos(3.13995226724  + 8439.63411433290 *T)
    sumY0  +=       0.151 * cos(4.33106068620  + 10228.26664602010 *T)
    sumY0  +=       0.151 * cos(1.25869310331  + 3525.70615215350 *T)
    sumY0  +=       0.149 * cos(0.32471999674  + 3664.36166084070 *T)
    sumY0  +=       0.196 * cos(1.82159362898  + 3610.98640425411 *T)
    sumY0  +=       0.166 * cos(0.46110782427  + 5466.73365156250 *T)
    sumY0  +=       0.167 * cos(1.32101864186  + 10288.68813930470 *T)
    sumY0  +=       0.156 * cos(2.36180228155  + 17394.97591724229 *T)
    sumY0  +=       0.147 * cos(5.85283339117  + 14577.42854360330 *T)
    sumY0  +=       0.147 * cos(0.61306634906  + 1898.57778177091 *T)
    sumY0  +=       0.154 * cos(0.19988862301  + 11457.22875934230 *T)
    sumY0  +=       0.172 * cos(3.82461644501  + 19617.32018680630 *T)
    sumY0  +=       0.166 * cos(1.16464294435  + 6908.94480367210 *T)
    sumY0  +=       0.160 * cos(2.02061385417  + 1346.10047382629 *T)
    sumY0  +=       0.180 * cos(3.88594579937  + 5220.41461893670 *T)
    sumY0  +=       0.159 * cos(2.34893738617  + 1505.04399160949 *T)
    sumY0  +=       0.171 * cos(0.89384833788  + 4399.75053940550 *T)
    sumY0  +=       0.178 * cos(3.88835944920  + 2807.15452577270 *T)
    sumY0  +=       0.201 * cos(6.02231000514  + 3739.02250124370 *T)
    sumY0  +=       0.148 * cos(5.69880757278  + 9154.14743350530 *T)
    sumY0  +=       0.146 * cos(5.33364778234  + 13369.80707128350 *T)
    sumY0  +=       0.170 * cos(4.17129374991  + 6677.92829888190 *T)
    sumY0  +=       0.147 * cos(0.30529245684  + 9918.98832336430 *T)
    sumY0  +=       0.147 * cos(3.63211816839  + 14481.44931638550 *T)
    sumY0  +=       0.147 * cos(3.04560054365  + 6850.56121904909 *T)
    sumY0  +=       0.159 * cos(4.74259177074  + 3334.99904188371 *T)
    sumY0  +=       0.159 * cos(1.38117514790  + 3346.71344648291 *T)
    sumY0  +=       0.150 * cos(3.63738939165  + 15110.22230238269 *T)
    sumY0  +=       0.143 * cos(5.81404981785  + 13227.62844424730 *T)
    sumY0  +=       0.189 * cos(4.57066567683  + 10042.43641698170 *T)
    sumY0  +=       0.153 * cos(3.12126478969  + 3702.97738091410 *T)
    sumY0  +=       0.168 * cos(1.11741682536  + 10032.71776293271 *T)
    sumY0  +=       0.155 * cos(1.90750071730  + 369.94363342390 *T)
    sumY0  +=       0.151 * cos(0.91782780746  + 1859.44687283811 *T)
    sumY0  +=       0.163 * cos(6.13295957527  + 2965.09051261311 *T)
    sumY0  +=       0.179 * cos(1.03162162149  + 3966.52643649571 *T)
    sumY0  +=       0.179 * cos(5.09214529715  + 2715.18605187090 *T)
    sumY0  +=       0.195 * cos(2.27751350976  + 16858.72635041670 *T)
    sumY0  +=       0.140 * cos(3.29302979467  + 18461.95902939930 *T)
    sumY0  +=       0.183 * cos(5.08915245987  + 6791.67499210250 *T)
    sumY0  +=       0.150 * cos(3.40313028042  + 1879.80219223690 *T)
    sumY0  +=       0.150 * cos(3.63201414077  + 2922.35192137631 *T)
    sumY0  +=       0.182 * cos(3.98532489687  + 8005.92237645609 *T)
    sumY0  +=       0.146 * cos(0.34189409252  + 2715.23110798610 *T)
    sumY0  +=       0.165 * cos(5.85176455035  + 7346.22471601311 *T)
    sumY0  +=       0.141 * cos(5.49463411882  + 3017.35082752591 *T)
    sumY0  +=       0.191 * cos(0.62441184949  + 3202.26804579890 *T)
    sumY0  +=       0.191 * cos(5.49935506915  + 3479.44444256770 *T)
    sumY0  +=       0.139 * cos(4.91449152787  + 4846.14405327630 *T)
    sumY0  +=       0.138 * cos(1.09735180536  + 1537.99661849450 *T)
    sumY0  +=       0.193 * cos(1.19262914799  + 3650.13456683910 *T)
    sumY0  +=       0.160 * cos(2.22528741800  + 3394.16419046711 *T)
    sumY0  +=       0.160 * cos(3.89847950064  + 3287.54829789951 *T)
    sumY0  +=       0.193 * cos(4.93382263995  + 3031.57792152750 *T)
    sumY0  +=       0.169 * cos(4.56651889742  + 4687.20053549310 *T)
    sumY0  +=       0.155 * cos(0.34848561430  + 3897.94876537470 *T)
    sumY0  +=       0.162 * cos(6.22598832232  + 5607.85964669230 *T)
    sumY0  +=       0.136 * cos(0.53418624988  + 3336.77818056450 *T)
    sumY0  +=       0.144 * cos(5.67390729246  + 3945.70598488811 *T)
    sumY0  +=       0.155 * cos(3.49137971151  + 7799.73682801889 *T)
    sumY0  +=       0.170 * cos(6.20206538140  + 6677.96280618630 *T)
    sumY0  +=       0.155 * cos(4.11217295378  + 25136.07033579810 *T)
    sumY0  +=       0.141 * cos(2.26678508447  + 6279.79654912590 *T)
    sumY0  +=       0.174 * cos(5.58730400350  + 7586.43773258089 *T)
    sumY0  +=       0.138 * cos(3.20211134662  + 6343.22597190010 *T)
    sumY0  +=       0.176 * cos(5.24754822584  + 605.09355818830 *T)
    sumY0  +=       0.167 * cos(3.71180827260  + 8337.02897508130 *T)
    sumY0  +=       0.185 * cos(0.86703648786  + 27831.79440179969 *T)
    sumY0  +=       0.151 * cos(4.79464206123  + 2281.05398932989 *T)
    sumY0  +=       0.151 * cos(4.63522367272  + 8962.76647769650 *T)
    sumY0  +=       0.136 * cos(0.80562105416  + 12729.90978496950 *T)
    sumY0  +=       0.159 * cos(3.23552569495  + 8958.76701413060 *T)
    sumY0  +=       0.151 * cos(5.04372005956  + 8962.63185709091 *T)
    sumY0  +=       0.151 * cos(5.20313844808  + 2280.91936872430 *T)
    sumY0  +=       0.137 * cos(1.47758341153  + 6703.31950020950 *T)
    sumY0  +=       0.159 * cos(0.16489389211  + 8966.63132065680 *T)
    sumY0  +=       0.132 * cos(4.69789010430  + 12171.76963240710 *T)
    sumY0  +=       0.136 * cos(4.02250108722  + 8696.09212567190 *T)
    sumY0  +=       0.143 * cos(2.19306681905  + 4106.64930862970 *T)
    sumY0  +=       0.140 * cos(1.26260069331  + 27.15773861329 *T)
    sumY0  +=       0.179 * cos(5.37712995429  + 5724.69187994550 *T)
    sumY0  +=       0.133 * cos(0.03147377213  + 21150.56954840009 *T)
    sumY0  +=       0.131 * cos(1.45458200748  + 948.93179148630 *T)
    sumY0  +=       0.132 * cos(5.47773606981  + 1861.98967011891 *T)
    sumY0  +=       0.133 * cos(5.86296749734  + 2472.43494513869 *T)
    sumY0  +=       0.138 * cos(1.97592968874  + 34554.57004618409 *T)
    sumY0  +=       0.139 * cos(3.62111712693  + 3.83424613530 *T)
    sumY0  +=       0.130 * cos(2.78438204570  + 529.91752892591 *T)
    sumY0  +=       0.157 * cos(0.66637254756  + 10787.38652706230 *T)
    sumY0  +=       0.133 * cos(4.50452650479  + 6624.57029594750 *T)
    sumY0  +=       0.148 * cos(2.40612364993  + 7051.16848682350 *T)
    sumY0  +=       0.130 * cos(3.08824912384  + 7594.03891454870 *T)
    sumY0  +=       0.132 * cos(0.17263785430  + 9594.99527173989 *T)
    sumY0  +=       0.138 * cos(5.62209899863  + 12406.40436831210 *T)
    sumY0  +=       0.130 * cos(4.81515367719  + 529.95203623031 *T)
    sumY0  +=       0.143 * cos(3.29046013826  + 10015.03486088490 *T)
    sumY0  +=       0.149 * cos(0.26971082215  + 10.88048283331 *T)
    sumY0  +=       0.178 * cos(4.65997179417  + 2675.61256408629 *T)
    sumY0  +=       0.133 * cos(4.74586376419  + 29424.39041543249 *T)
    sumY0  +=       0.158 * cos(3.17185301711  + 6092.40384387490 *T)
    sumY0  +=       0.131 * cos(5.18408640439  + 158.69970029970 *T)
    sumY0  +=       0.134 * cos(4.61138320293  + 2954.43659361110 *T)
    sumY0  +=       0.142 * cos(1.22013927552  + 3041.24221494710 *T)
    sumY0  +=       0.164 * cos(1.02457529047  + 26083.77798873270 *T)
    sumY0  +=       0.145 * cos(3.33534254136  + 8373.63434038550 *T)
    sumY0  +=       0.172 * cos(1.36305306778  + 1086.53966880249 *T)
    sumY0  +=       0.128 * cos(1.88046940017  + 155.10927164789 *T)
    sumY0  +=       0.137 * cos(4.29526096322  + 6137.55061178690 *T)
    sumY0  +=       0.167 * cos(3.86261571309  + 107.90734142210 *T)
    sumY0  +=       0.133 * cos(1.80894644197  + 12167.75887909109 *T)
    sumY0  +=       0.137 * cos(3.89101568839  + 14581.01897225510 *T)
    sumY0  +=       0.137 * cos(1.10646412832  + 20995.14914896590 *T)
    sumY0  +=       0.126 * cos(6.13522128177  + 37.88921815429 *T)
    sumY0  +=       0.130 * cos(0.80699742659  + 4108.70506353190 *T)
    sumY0  +=       0.137 * cos(5.57568822106  + 18052.68572567429 *T)
    sumY0  +=       0.131 * cos(2.10282374334  + 692.40141871030 *T)
    sumY0  +=       0.132 * cos(5.61172315472  + 27490.44866056129 *T)
    sumY0  +=       0.144 * cos(5.56368880495  + 1802.61580820530 *T)
    sumY0  +=       0.146 * cos(5.31840685158  + 3475.43368925169 *T)
    sumY0  +=       0.126 * cos(0.28303045782  + 18447.73193539770 *T)
    sumY0  +=       0.141 * cos(6.27643241405  + 3716.62197575350 *T)
    sumY0  +=       0.124 * cos(1.38024620663  + 31172.40682849949 *T)
    sumY0  +=       0.124 * cos(1.96413242139  + 6688.15233475721 *T)
    sumY0  +=       0.121 * cos(1.40075751759  + 3590.76070492550 *T)
    sumY0  +=       0.121 * cos(4.72300940105  + 3090.95178344111 *T)
    sumY0  +=       0.122 * cos(5.79446111948  + 2257.59587624631 *T)
    sumY0  +=       0.123 * cos(2.86133242277  + 6710.89817941910 *T)
    sumY0  +=       0.129 * cos(3.19990002398  + 3073.76156749449 *T)
    sumY0  +=       0.131 * cos(5.96540599255  + 5401.67409825510 *T)
    sumY0  +=       0.166 * cos(3.56556372747  + 6553.44982754571 *T)
    sumY0  +=       0.134 * cos(1.17245951228  + 4783.11745294350 *T)
    sumY0  +=       0.122 * cos(4.84545055419  + 1235.83184283950 *T)
    sumY0  +=       0.128 * cos(0.74562744196  + 3081.29543915950 *T)
    sumY0  +=       0.128 * cos(5.37813947668  + 3600.41704920711 *T)
    sumY0  +=       0.121 * cos(0.16505703713  + 2126.05391455991 *T)
    sumY0  +=       0.124 * cos(3.09719471943  + 5820.67110716330 *T)
    sumY0  +=       0.122 * cos(1.01207063042  + 3343.95903786810 *T)
    sumY0  +=       0.140 * cos(2.18255999829  + 10191.17364574889 *T)
    sumY0  +=       0.127 * cos(2.49175261711  + 3759.36056699030 *T)
    sumY0  +=       0.143 * cos(5.81200960148  + 2229.21399908130 *T)
    sumY0  +=       0.120 * cos(3.37800348541  + 1648.69057468091 *T)
    sumY0  +=       0.152 * cos(5.74350080821  + 3340.82173687890 *T)
    sumY0  +=       0.152 * cos(3.52185876402  + 3340.89075148770 *T)
    sumY0  +=       0.137 * cos(5.24958747332  + 4296.09776878290 *T)
    sumY0  +=       0.140 * cos(6.07934598932  + 2233.71743849851 *T)
    sumY0  +=       0.118 * cos(2.67440875498  + 7534.66505263510 *T)
    sumY0  +=       0.137 * cos(3.22103647277  + 14588.06520895310 *T)
    sumY0  +=       0.123 * cos(1.05338746460  + 9381.20349020070 *T)
    sumY0  +=       0.116 * cos(5.02262371793  + 10338.47296723950 *T)
    sumY0  +=       0.137 * cos(0.87583504561  + 8322.80188107970 *T)
    sumY0  +=       0.126 * cos(4.15152723756  + 4083.84630471591 *T)
    sumY0  +=       0.133 * cos(0.20050131976  + 3923.85515556170 *T)
    sumY0  +=       0.124 * cos(2.32394389472  + 9124.67816855889 *T)
    sumY0  +=       0.115 * cos(2.71796809879  + 6740.35489782970 *T)
    sumY0  +=       0.126 * cos(4.47944403901  + 3509.94879234930 *T)
    sumY0  +=       0.121 * cos(4.78087706215  + 3862.94602731009 *T)
    sumY0  +=       0.118 * cos(0.67101126938  + 1052.02456570489 *T)
    sumY0  +=       0.134 * cos(5.30662276999  + 6659.61784155670 *T)
    sumY0  +=       0.133 * cos(0.28852620914  + 28.20536998430 *T)
    sumY0  +=       0.113 * cos(5.12126627476  + 8436.53132064810 *T)
    sumY0  +=       0.133 * cos(0.75275688425  + 6147.84555252540 *T)
    sumY0  +=       0.113 * cos(3.24121652591  + 3116.51144848140 *T)
    sumY0  +=       0.154 * cos(1.23584313686  + 401.42830427370 *T)
    sumY0  +=       0.119 * cos(0.88326438143  + 1471.99684454710 *T)
    sumY0  +=       0.116 * cos(1.51219722423  + 3727.27589475550 *T)
    sumY0  +=       0.120 * cos(3.47272455509  + 3826.34066200590 *T)
    sumY0  +=       0.111 * cos(3.49092955754  + 6837.86939138550 *T)
    sumY0  +=       0.147 * cos(4.35013931667  + 8447.23529630070 *T)
    sumY0  +=       0.120 * cos(4.26889661746  + 2185.00745180929 *T)
    sumY0  +=       0.114 * cos(5.78187265627  + 3966.48138038050 *T)
    sumY0  +=       0.126 * cos(1.01640087918  + 7586.92536754791 *T)
    sumY0  +=       0.129 * cos(3.02800117630  + 3580.06172980831 *T)
    sumY0  +=       0.129 * cos(3.09576574234  + 3101.65075855830 *T)
    sumY0  +=       0.123 * cos(3.17573128745  + 5511.39278450750 *T)
    sumY0  +=       0.133 * cos(2.29548900012  + 6738.36704581871 *T)
    sumY0  +=       0.124 * cos(5.40608799173  + 7210.67200101070 *T)
    sumY0  +=       0.107 * cos(3.75293495504  + 6681.04834621889 *T)
    sumY0  +=       0.139 * cos(4.05414767826  + 14867.49369840869 *T)
    sumY0  +=       0.112 * cos(2.28073796145  + 27433.64539839149 *T)
    sumY0  +=       0.112 * cos(5.01410887723  + 10176.94655174729 *T)
    sumY0  +=       0.126 * cos(2.05837136952  + 6314.24434192070 *T)
    sumY0  +=       0.146 * cos(1.72637612424  + 3282.97754817950 *T)
    sumY0  +=       0.146 * cos(4.39739079440  + 3398.73494018711 *T)
    sumY0  +=       0.112 * cos(4.54751217161  + 4953.80757721490 *T)
    sumY0  +=       0.115 * cos(2.98486230767  + 4179.82553193370 *T)
    sumY0  +=       0.115 * cos(3.13902228769  + 2501.88695643291 *T)
    sumY0  +=       0.130 * cos(3.94422427867  + 6155.70985905160 *T)
    sumY0  +=       0.134 * cos(5.72416579818  + 4825.78873387751 *T)
    sumY0  +=       0.138 * cos(5.27722150274  + 5813.53500680550 *T)
    sumY0  +=       0.115 * cos(2.98002244807  + 8976.92626139531 *T)
    sumY0  +=       0.147 * cos(5.26253855720  + 5223.45010231870 *T)
    sumY0  +=       0.105 * cos(3.84211418761  + 17822.06174308530 *T)
    sumY0  +=       0.118 * cos(5.89203827165  + 5483.01090734249 *T)
    sumY0  +=       0.103 * cos(5.03671194765  + 1059.55843736991 *T)
    sumY0  +=       0.104 * cos(4.83299836609  + 549.80246700989 *T)
    sumY0  +=       0.107 * cos(2.17758765251  + 7076.09455594230 *T)
    sumY0  +=       0.105 * cos(3.92288350713  + 6961.25517523550 *T)
    sumY0  +=       0.129 * cos(3.92939028141  + 2961.48283030911 *T)
    sumY0  +=       0.129 * cos(2.19437663723  + 3720.22965805750 *T)
    sumY0  +=       0.102 * cos(0.26786777670  + 3230.16228799689 *T)
    sumY0  +=       0.122 * cos(1.12211172947  + 27676.86163733250 *T)
    sumY0  +=       0.136 * cos(5.86697607222  + 765.54924696289 *T)
    sumY0  +=       0.116 * cos(1.89113439545  + 10213.04172872750 *T)
    sumY0  +=       0.115 * cos(4.67049842349  + 2631.86644190671 *T)
    sumY0  +=       0.102 * cos(5.07453595268  + 639.65346883050 *T)
    sumY0  +=       0.104 * cos(4.68887219232  + 31450.07086023530 *T)
    sumY0  +=       0.141 * cos(0.65765353971  + 6456.63624021419 *T)
    sumY0  +=       0.138 * cos(6.27290155228  + 12295.71041212569 *T)
    sumY0  +=       0.127 * cos(4.38757311711  + 2655.74487965451 *T)
    sumY0  +=       0.101 * cos(3.19202487469  + 21791.93479074930 *T)
    sumY0  +=       0.103 * cos(1.18713398647  + 3134.18306077910 *T)
    sumY0  +=       0.100 * cos(1.62440020436  + 31213.95761948429 *T)
    sumY0  +=       0.115 * cos(4.29413166220  + 21000.67207327329 *T)
    sumY0  +=       0.138 * cos(2.88368956632  + 3201.15810466050 *T)
    sumY0  +=       0.138 * cos(3.24007735232  + 3480.55438370610 *T)
    sumY0  +=       0.099 * cos(3.72853240669  + 6953.65399326770 *T)
    sumY0  +=       0.129 * cos(1.79825340837  + 4878.60904519429 *T)
    sumY0  +=       0.106 * cos(3.75036852906  + 3023.97673955970 *T)
    sumY0  +=       0.100 * cos(3.34238462758  + 8741.79888998789 *T)
    sumY0  +=       0.100 * cos(0.20079197399  + 8742.28652495491 *T)
    sumY0  +=       0.101 * cos(1.72955787995  + 6534.65698435950 *T)
    sumY0  +=       0.100 * cos(0.79632336674  + 2652.22176130550 *T)
    sumY0  +=       0.108 * cos(4.13052988823  + 250.14827822570 *T)
    sumY0  +=       0.113 * cos(3.89774709782  + 13937.03832261750 *T)
    sumY0  +=       0.100 * cos(3.82154216379  + 3340.77168022831 *T)
    sumY0  +=       0.100 * cos(5.44381740844  + 3340.94080813830 *T)
    sumY0  +=       0.098 * cos(1.40156960306  + 6144.17652382070 *T)
    sumY0  +=       0.104 * cos(3.63716602244  + 11526.88127170889 *T)
    sumY0  +=       0.099 * cos(5.78288558119  + 5312.80841803811 *T)
    sumY0  +=       0.113 * cos(2.18817668288  + 10080.96732452950 *T)
    sumY0  +=       0.107 * cos(4.16143134189  + 6680.91372561330 *T)
    sumY0  +=       0.095 * cos(0.21807223502  + 8425.40702033129 *T)
    sumY0  +=       0.095 * cos(2.99902409772  + 3310.97420057310 *T)
    sumY0  +=       0.095 * cos(3.12474282092  + 3370.73828779350 *T)
    sumY0  +=       0.098 * cos(6.07573481401  + 18616.82448356369 *T)
    sumY0  +=       0.095 * cos(5.51010177831  + 4783.13470659570 *T)
    sumY0  +=       0.096 * cos(0.99819535387  + 1353.14671052430 *T)
    sumY0  +=       0.100 * cos(3.41304577695  + 3340.90630083390 *T)
    sumY0  +=       0.100 * cos(5.85231379528  + 3340.80618753271 *T)
    sumY0  +=       0.112 * cos(1.01874315131  + 5084.86190393429 *T)
    sumY0  +=       0.098 * cos(3.70334941527  + 7277.64704489211 *T)
    sumY0  +=       0.097 * cos(1.45365408670  + 12509.00950756369 *T)
    sumY0  +=       0.100 * cos(0.65818374358  + 1755.22786719560 *T)
    sumY0  +=       0.098 * cos(0.97983442462  + 1784.28809722790 *T)
    sumY0  +=       0.129 * cos(2.93691181926  + 4591.68644048410 *T)
    sumY0  +=       0.098 * cos(5.93410154372  + 4981.70181941289 *T)
    sumY0  +=       0.100 * cos(5.59483630292  + 6474.76793358650 *T)
    sumY0  +=       0.125 * cos(1.84348906532  + 30228.22229391390 *T)
    sumY0  +=       0.102 * cos(3.13400085165  + 2214.98690507971 *T)
    sumY0  +=       0.092 * cos(1.71243931270  + 1383.13116433011 *T)
    sumY0  +=       0.101 * cos(5.97238974906  + 3302.23557357850 *T)
    sumY0  +=       0.100 * cos(4.21928425466  + 13517.55897844709 *T)
    sumY0  +=       0.096 * cos(3.71471309882  + 16975.97890833410 *T)
    sumY0  +=       0.096 * cos(3.08041062563  + 34114.87025179110 *T)
    sumY0  +=       0.118 * cos(5.39174563560  + 3343.82518959990 *T)
    sumY0  +=       0.118 * cos(0.73202128304  + 3337.88729876671 *T)
    sumY0  +=       0.105 * cos(3.93919081745  + 42389.17875379050 *T)
    sumY0  +=       0.118 * cos(5.04536066143  + 2524.20291796811 *T)
    sumY0  +=       0.114 * cos(3.42685633162  + 947.95147041809 *T)
    sumY0  +=       0.102 * cos(3.49277811118  + 27946.99179740330 *T)
    sumY0  +=       0.103 * cos(2.64259699800  + 20235.36664379391 *T)
    sumY0  +=       0.094 * cos(0.20789444079  + 11507.01358727710 *T)
    sumY0  +=       0.118 * cos(2.04691913094  + 3284.47305810711 *T)
    sumY0  +=       0.118 * cos(4.07684778770  + 3397.23943025950 *T)
    sumY0  +=       0.115 * cos(3.05378875629  + 640.63378989870 *T)
    sumY0  +=       0.104 * cos(4.79113313233  + 3149.42523172390 *T)
    sumY0  +=       0.104 * cos(4.47417659438  + 3532.28725664270 *T)
    sumY0  +=       0.091 * cos(1.86422625651  + 24093.03297169169 *T)
    sumY0  +=       0.093 * cos(0.68624597132  + 987.99009041150 *T)
    sumY0  +=       0.093 * cos(4.76939880894  + 1957.48126236970 *T)
    sumY0  +=       0.116 * cos(3.49150933105  + 49225.82400662430 *T)
    sumY0  +=       0.099 * cos(3.02556702345  + 6489.53276980449 *T)
    sumY0  +=       0.113 * cos(1.69589601449  + 2597.86618365071 *T)
    sumY0  +=       0.107 * cos(5.76130975592  + 1744.66980189871 *T)
    sumY0  +=       0.094 * cos(3.77994404993  + 3334.72301825071 *T)
    sumY0  +=       0.094 * cos(2.34382286871  + 3346.98947011590 *T)
    sumY0  +=       0.090 * cos(5.67852689848  + 15747.01679501190 *T)
    sumY0  +=       0.091 * cos(1.14523189560  + 50818.42002025710 *T)
    sumY0  +=       0.108 * cos(3.58607681238  + 1971.70835637129 *T)
    sumY0  +=       0.111 * cos(2.45297309782  + 27417.30083230870 *T)
    sumY0  +=       0.104 * cos(3.56796729290  + 441.06910236111 *T)
    sumY0  +=       0.089 * cos(5.47858220889  + 3313.38737778371 *T)
    sumY0  +=       0.089 * cos(0.64518470975  + 3368.32511058290 *T)
    sumY0  +=       0.103 * cos(4.62821556081  + 1059.69305797550 *T)
    sumY0  +=       0.112 * cos(2.79231795944  + 1744.73711220151 *T)
    sumY0  +=       0.092 * cos(0.19138124327  + 32509.45279042450 *T)
    sumY0  +=       0.100 * cos(3.85768361090  + 10611.14592459111 *T)
    sumY0  +=       0.112 * cos(4.47970305153  + 10721.35224581050 *T)
    sumY0  +=       0.100 * cos(5.72799005083  + 309.03450517230 *T)
    sumY0  +=       0.105 * cos(3.08589856125  + 24606.37937070350 *T)
    sumY0  +=       0.095 * cos(0.84457868347  + 529.86747227531 *T)
    sumY0  +=       0.120 * cos(6.10556233132  + 3202.33874731261 *T)
    sumY0  +=       0.120 * cos(0.01820458732  + 3479.37374105400 *T)
    sumY0  +=       0.086 * cos(0.77245183733  + 2918.82880302731 *T)
    sumY0  +=       0.086 * cos(5.35131508131  + 3762.88368533930 *T)
    sumY0  +=       0.099 * cos(2.05522586895  + 3458.66904707530 *T)
    sumY0  +=       0.099 * cos(4.06854104969  + 3223.04344129130 *T)
    sumY0  +=       0.089 * cos(1.74507054165  + 22531.33442263010 *T)
    sumY0  +=       0.107 * cos(3.69153019661  + 3311.87461420391 *T)
    sumY0  +=       0.107 * cos(2.43223672203  + 3369.83787416271 *T)
    sumY0  +=       0.090 * cos(3.38165668766  + 4670.36833444331 *T)
    sumY0  +=       0.094 * cos(1.95446777954  + 796.52457064770 *T)
    sumY0  +=       0.085 * cos(5.95959971466  + 809.23365196350 *T)
    sumY0  +=       0.084 * cos(3.00350230053  + 4778.03185838191 *T)
    sumY0  +=       0.118 * cos(1.79240399777  + 8219.22147189409 *T)
    sumY0  +=       0.104 * cos(2.76036150084  + 3149.39072441950 *T)
    sumY0  +=       0.104 * cos(0.22176291869  + 3532.32176394710 *T)
    sumY0  +=       0.084 * cos(4.22627606170  + 1125.62552162009 *T)
    sumY0  +=       0.085 * cos(5.95594200813  + 4555.65857380670 *T)
    sumY0  +=       0.116 * cos(5.49142363859  + 14694.74841182350 *T)
    sumY0  +=       0.086 * cos(6.08864556833  + 12410.97511803210 *T)
    sumY0  +=       0.084 * cos(3.77352177481  + 1106.89498820129 *T)
    sumY0  +=       0.117 * cos(2.69247322881  + 30774.25782509130 *T)
    sumY0  +=       0.090 * cos(0.13296522638  + 32366.85383872410 *T)
    sumY0  +=       0.091 * cos(5.29506341401  + 9161.77116883010 *T)
    sumY0  +=       0.089 * cos(6.23353856478  + 3348.53003615871 *T)
    sumY0  +=       0.089 * cos(6.17341366104  + 3333.18245220790 *T)
    sumY0  +=       0.093 * cos(5.47331396961  + 9225.29545579949 *T)
    sumY0  +=       0.102 * cos(3.16676623170  + 4466.72558328691 *T)
    sumY0  +=       0.102 * cos(3.16676623170  + 4466.23794831990 *T)
    sumY0  +=       0.099 * cos(0.51130040313  + 4879.09668016130 *T)
    sumY0  +=       0.086 * cos(4.05866338465  + 3613.04156656790 *T)
    sumY0  +=       0.083 * cos(4.34759815373  + 7899.30648388849 *T)
    sumY0  +=       0.082 * cos(1.30967650485  + 1265.32366114289 *T)
    sumY0  +=       0.084 * cos(0.92794835827  + 2942.77455107790 *T)
    sumY0  +=       0.083 * cos(3.70816234293  + 3077.28468584349 *T)
    sumY0  +=       0.082 * cos(0.64666599764  + 3351.42559923031 *T)
    sumY0  +=       0.082 * cos(5.47710092100  + 3330.28688913630 *T)
    sumY0  +=       0.112 * cos(1.67471534984  + 14900.93396026070 *T)
    sumY0  +=       0.082 * cos(0.58404375695  + 22318.03532719209 *T)
    sumY0  +=       0.084 * cos(3.20837869526  + 3597.87425192630 *T)
    sumY0  +=       0.084 * cos(2.91538822338  + 3083.83823644030 *T)
    sumY0  +=       0.082 * cos(2.92113670927  + 1552.22371249610 *T)
    sumY0  +=       0.095 * cos(0.43608229663  + 530.00209288090 *T)
    sumY0  +=       0.081 * cos(4.30100281575  + 2980.25782725470 *T)
    sumY0  +=       0.085 * cos(4.86849510055  + 1056.10262932370 *T)
    sumY0  +=       0.097 * cos(0.04175992157  + 4447.99504986810 *T)
    sumY0  +=       0.084 * cos(4.53785743204  + 845.83901726770 *T)
    sumY0  +=       0.114 * cos(1.28659258054  + 3063.82125044190 *T)
    sumY0  +=       0.114 * cos(4.83717433810  + 3617.89123792471 *T)
    sumY0  +=       0.085 * cos(2.86532283276  + 23937.61257225750 *T)
    sumY0  +=       0.085 * cos(2.68159304983  + 54790.18050746150 *T)
    sumY0  +=       0.086 * cos(3.41567897446  + 23010.19292841890 *T)
    sumY0  +=       0.085 * cos(2.68041688991  + 8322.31424611270 *T)
    sumY0  +=       0.083 * cos(2.23302149732  + 20103.29223959590 *T)
    sumY0  +=       0.112 * cos(3.74830398933  + 5295.57339454690 *T)
    sumY0  +=       0.081 * cos(0.67408003259  + 3490.75371931010 *T)
    sumY0  +=       0.081 * cos(5.44968688605  + 3190.95876905650 *T)
    sumY0  +=       0.099 * cos(2.78587638666  + 4845.65641830930 *T)
    sumY0  +=       0.094 * cos(3.98523941104  + 796.55907795210 *T)
    sumY0  +=       0.092 * cos(0.38071563992  + 4452.49848928530 *T)
    sumY0  +=       0.092 * cos(0.38071563992  + 4452.01085431830 *T)
    sumY0  +=       0.091 * cos(4.85895142608  + 2544.54098371471 *T)
    sumY0  +=       0.081 * cos(1.64836757283  + 7166.95308870570 *T)
    sumY0  +=       0.089 * cos(4.15046971529  + 6311.76885494270 *T)
    sumY0  +=       0.084 * cos(5.20424780335  + 4822.26561552850 *T)
    sumY0  +=       0.089 * cos(1.45260507956  + 4049.84604645990 *T)
    sumY0  +=       0.085 * cos(4.83658145136  + 14716.08405229050 *T)
    sumY0  +=       0.081 * cos(2.70661000816  + 34363.12178007249 *T)
    sumY0  +=       0.094 * cos(0.87814163025  + 24532.73276608469 *T)
    sumY0  +=       0.102 * cos(1.61863937790  + 3524.72583108530 *T)
    sumY0  +=       0.102 * cos(4.50512754074  + 3156.98665728130 *T)
    sumY0  +=       0.084 * cos(1.33644474511  + 2942.63993047231 *T)
    sumY0  +=       0.084 * cos(2.73245346971  + 4801.91029612970 *T)
    sumY0  +=       0.084 * cos(2.63232238753  + 78263.95324220609 *T)
    sumY0  +=       0.079 * cos(0.32930597136  + 4424.11661212030 *T)
    sumY0  +=       0.081 * cos(1.78230535496  + 3333.72544353031 *T)
    sumY0  +=       0.081 * cos(4.34146156368  + 3347.98704483631 *T)
    sumY0  +=       0.079 * cos(5.32564058250  + 4029.49072706111 *T)
    sumY0  +=       0.083 * cos(0.29135686673  + 24336.24921063270 *T)
    sumY0  +=       0.102 * cos(2.08124303406  + 3119.95596677750 *T)
    sumY0  +=       0.096 * cos(6.16597212110  + 3324.02404313351 *T)
    sumY0  +=       0.096 * cos(6.24098010472  + 3357.68844523310 *T)
    sumY0  +=       0.080 * cos(4.26306642509  + 12987.86801335249 *T)
    sumY0  +=       0.099 * cos(4.02513713414  + 3376.46403476710 *T)
    sumY0  +=       0.099 * cos(2.09862978450  + 3305.24845359951 *T)
    sumY0  +=       0.090 * cos(4.06976669148  + 18028.24729152250 *T)
    sumY0  +=       0.086 * cos(2.51147086235  + 3296.20241094311 *T)
    sumY0  +=       0.086 * cos(3.61229605629  + 3385.51007742350 *T)
    sumY0  +=       0.085 * cos(5.04929372084  + 3487.65067705471 *T)
    sumY0  +=       0.085 * cos(1.07447319780  + 3194.06181131190 *T)
    sumY0  +=       0.085 * cos(4.07158225922  + 1758.89689590030 *T)
    sumY0  +=       0.088 * cos(0.27523861221  + 2751.25897466350 *T)
    sumY0  +=       0.082 * cos(3.74673586971  + 16460.57734700850 *T)
    sumY0  +=       0.101 * cos(1.69645528757  + 1162.23088692429 *T)
    sumY0  +=       0.079 * cos(3.11818153513  + 8186.26884500910 *T)
    sumY0  +=       0.097 * cos(1.73619334013  + 4025.96760871210 *T)
    sumY0  +=       0.077 * cos(0.22099888781  + 7446.77410036250 *T)
    sumY0  +=       0.077 * cos(0.22099888781  + 7447.26173532950 *T)
    sumY0  +=       0.077 * cos(0.70593259841  + 430.77416162260 *T)
    sumY0  +=       0.102 * cos(5.59681662693  + 574.58861581830 *T)
    sumY0  +=       0.079 * cos(3.73816629713  + 1593.24886603040 *T)
    sumY0  +=       0.090 * cos(1.49099440681  + 12891.88878613469 *T)
    sumY0  +=       0.079 * cos(6.19859779271  + 101010.75632211348 *T)
    sumY0  +=       0.100 * cos(2.65214559872  + 1481.16555386170 *T)
    sumY0  +=       0.106 * cos(4.78557880821  + 13.98327651810 *T)
    sumY0  +=       0.076 * cos(5.49053372377  + 31979.76182532990 *T)
    sumY0  +=       0.102 * cos(5.35655741954  + 52168.28742991590 *T)
    sumY0  +=       0.103 * cos(3.41208170979  + 6777.44789810091 *T)
    sumY0  +=       0.080 * cos(3.38185008186  + 1954.96096784710 *T)
    sumY0  +=       0.075 * cos(0.45216449067  + 14797.84118604210 *T)
    sumY0  +=       0.091 * cos(5.73775200546  + 6836.33412504750 *T)
    sumY0  +=       0.077 * cos(0.48111203734  + 51770.13842650770 *T)
    sumY0  +=       0.078 * cos(4.39552972351  + 8831.15720570730 *T)
    sumY0  +=       0.085 * cos(2.57284263819  + 369.45599845690 *T)
    sumY0  +=       0.091 * cos(0.60653775040  + 2544.57549101910 *T)
    sumY0  +=       0.081 * cos(0.95322040984  + 3277.12034587990 *T)
    sumY0  +=       0.081 * cos(5.17054650880  + 3404.59214248671 *T)
    sumY0  +=       0.076 * cos(3.85445806150  + 10025.11658096489 *T)
    sumY0  +=       0.084 * cos(5.75160998999  + 45331.64217708210 *T)
    sumY0  +=       0.077 * cos(0.59680440034  + 1324.69752305650 *T)
    sumY0  +=       0.091 * cos(5.97834946002  + 45494.33761226529 *T)
    sumY0  +=       0.081 * cos(5.17266571119  + 3325.11613803129 *T)
    sumY0  +=       0.080 * cos(1.79665492010  + 569.06569151090 *T)
    sumY0  +=       0.075 * cos(5.25222075961  + 9090.71801073110 *T)
    sumY0  +=       0.090 * cos(3.17222371452  + 1747.36356066940 *T)
    sumY0  +=       0.073 * cos(4.22695819030  + 7476.23141136150 *T)
    sumY0  +=       0.091 * cos(4.81327110244  + 8536.10097651771 *T)
    sumY0  +=       0.091 * cos(4.81327110244  + 8535.61334155070 *T)
    sumY0  +=       0.102 * cos(4.59807270679  + 8692.08137235589 *T)
    sumY0  +=       0.073 * cos(5.20991824086  + 16062.42834360030 *T)
    sumY0  +=       0.087 * cos(4.26136604650  + 15117.75617404771 *T)
    sumY0  +=       0.077 * cos(1.47494250575  + 20618.26317601710 *T)
    sumY0  +=       0.086 * cos(3.28001892853  + 5355.47969897210 *T)
    sumY0  +=       0.081 * cos(0.96338461381  + 2917.29353668930 *T)
    sumY0  +=       0.077 * cos(0.91362006762  + 1008.30565339989 *T)
    sumY0  +=       0.081 * cos(1.65889494544  + 2910.24729999130 *T)
    sumY0  +=       0.072 * cos(4.24572819275  + 928.08378598631 *T)
    sumY0  +=       0.073 * cos(3.00898976422  + 6809.48751422050 *T)
    sumY0  +=       0.075 * cos(6.23302655949  + 4106.16167366270 *T)
    sumY0  +=       0.074 * cos(4.08117422586  + 30376.10882168309 *T)
    sumY0  +=       0.094 * cos(5.88680065989  + 17809.95712170029 *T)
    sumY0  +=       0.080 * cos(2.13848140967  + 6151.84501609130 *T)
    sumY0  +=       0.075 * cos(0.09092377390  + 81604.56566890588 *T)
    sumY0  +=       0.076 * cos(0.01129674960  + 44534.85653529869 *T)
    sumY0  +=       0.097 * cos(5.85293706952  + 5749.61794906429 *T)
    sumY0  +=       0.071 * cos(3.99491883467  + 9384.59719059169 *T)
    sumY0  +=       0.074 * cos(4.00818928999  + 6603.71812689920 *T)
    sumY0  +=       0.071 * cos(3.77471202293  + 14061.96467251030 *T)
    sumY0  +=       0.073 * cos(3.77824643335  + 13635.36648163430 *T)
    sumY0  +=       0.071 * cos(4.21888578691  + 5622.15405099670 *T)
    sumY0  +=       0.071 * cos(4.30845955918  + 23539.88389351350 *T)
    sumY0  +=       0.081 * cos(1.38235650244  + 19573.60127450130 *T)
    sumY0  +=       0.100 * cos(3.67335445653  + 3399.74247112990 *T)
    sumY0  +=       0.075 * cos(3.40358526619  + 1214.55851213989 *T)
    sumY0  +=       0.094 * cos(3.33044734976  + 10816.84383806129 *T)
    sumY0  +=       0.071 * cos(4.38626132317  + 12140.84495796130 *T)
    sumY0  +=       0.072 * cos(1.37425478851  + 26887.60986721410 *T)
    sumY0  +=       0.072 * cos(0.56330972127  + 2010.85651895630 *T)
    sumY0  +=       0.072 * cos(1.19011347351  + 11560.32153356091 *T)
    sumY0  +=       0.072 * cos(5.48851072898  + 21957.43691026350 *T)
    sumY0  +=       0.071 * cos(4.41874779960  + 1619.75375224610 *T)
    sumY0  +=       0.074 * cos(1.68474352553  + 20205.89737884749 *T)
    sumY0  +=       0.075 * cos(2.70325711680  + 7477.27904273249 *T)
    sumY0  +=       0.073 * cos(0.64702279583  + 6493.54352312051 *T)
    sumY0  +=       0.069 * cos(3.54263618042  + 5784.06574185909 *T)
    sumY0  +=       0.084 * cos(1.01645431208  + 2118.52004289490 *T)
    sumY0  +=       0.086 * cos(3.03743422532  + 19404.02109136830 *T)
    sumY0  +=       0.095 * cos(2.31320113804  + 2104.78058386030 *T)
    sumY0  +=       0.075 * cos(2.77500192554  + 8748.91243698870 *T)
    sumY0  +=       0.076 * cos(4.33355808410  + 3268.80051082850 *T)
    sumY0  +=       0.076 * cos(1.79020883454  + 3412.91197753810 *T)
    sumY0  +=       0.089 * cos(1.64126071038  + 15643.43638582629 *T)
    sumY0  +=       0.075 * cos(1.25947909560  + 36106.10290379910 *T)
    sumY0  +=       0.069 * cos(2.65446892553  + 21272.88049100450 *T)
    sumY0  +=       0.071 * cos(0.24705889025  + 58857.27495403150 *T)
    sumY0  +=       0.071 * cos(4.66819102175  + 5298.09368906950 *T)
    sumY0  +=       0.071 * cos(4.66819102175  + 5298.58132403650 *T)
    sumY0  +=       0.080 * cos(2.54697779651  + 6151.71039548571 *T)
    sumY0  +=       0.068 * cos(3.06859270324  + 942.30587945250 *T)
    sumY0  +=       0.089 * cos(3.32880979752  + 30640.10045611650 *T)
    sumY0  +=       0.068 * cos(4.30719958123  + 11247.45278225330 *T)
    sumY0  +=       0.068 * cos(5.53923428830  + 8550.32807051931 *T)
    sumY0  +=       0.078 * cos(3.75474105015  + 422.90985509640 *T)
    sumY0  +=       0.068 * cos(4.94904122324  + 4665.30994975630 *T)
    sumY0  +=       0.068 * cos(4.94904122324  + 4665.79758472330 *T)
    sumY0  +=       0.070 * cos(5.89813960443  + 2332.30677329990 *T)
    sumY0  +=       0.070 * cos(0.02785351787  + 13105.18788157269 *T)
    sumY0  +=       0.071 * cos(5.54255167971  + 3309.86425943471 *T)
    sumY0  +=       0.071 * cos(0.58121523893  + 3371.84822893190 *T)
    sumY0  +=       0.080 * cos(3.39865381902  + 9705.68922792630 *T)
    sumY0  +=       0.079 * cos(4.40031750333  + 16065.82204399129 *T)
    sumY0  +=       0.070 * cos(0.92757117435  + 7843.94337529091 *T)
    sumY0  +=       0.068 * cos(3.54162039995  + 9175.99826283170 *T)
    sumY0  +=       0.081 * cos(2.34135520581  + 94329.53146871387 *T)
    sumY0  +=       0.067 * cos(4.00501817920  + 6112.64679685570 *T)
    sumY0  +=       0.075 * cos(0.08972367730  + 1585.38455950421 *T)
    sumY0  +=       0.091 * cos(1.20169391057  + 7768.25215716910 *T)
    sumY0  +=       0.068 * cos(5.88468412911  + 11663.41430777950 *T)
    sumY0  +=       0.069 * cos(1.03197307646  + 23554.11098751510 *T)
    sumY0  +=       0.072 * cos(5.59069878383  + 3267.22078096370 *T)
    sumY0  +=       0.072 * cos(0.53306813481  + 3414.49170740290 *T)
    sumY0  +=       0.073 * cos(2.34799922671  + 2693.84541086851 *T)
    sumY0  +=       0.069 * cos(5.96168258099  + 3494.78147770770 *T)
    sumY0  +=       0.069 * cos(0.16208433765  + 3186.93101065891 *T)
    sumY0  +=       0.068 * cos(2.22578139544  + 5202.11446185170 *T)
    sumY0  +=       0.085 * cos(1.07840622469  + 4157.50957039850 *T)
    sumY0  +=       0.066 * cos(0.70967474590  + 15636.32283882549 *T)
    sumY0  +=       0.093 * cos(5.17707127671  + 1826.07731249979 *T)
    sumY0  +=       0.075 * cos(0.69441189563  + 15276.21205686389 *T)
    sumY0  +=       0.067 * cos(4.16813989827  + 290.24166198609 *T)
    sumY0  +=       0.073 * cos(3.87941986980  + 3237.27583499770 *T)
    sumY0  +=       0.066 * cos(4.38025942738  + 10823.95738506209 *T)
    sumY0  +=       0.066 * cos(4.38025942738  + 10824.44502002911 *T)
    sumY0  +=       0.066 * cos(2.30801018010  + 16056.29511766771 *T)
    sumY0  +=       0.066 * cos(1.79189006879  + 226.07308589371 *T)
    sumY0  +=       0.068 * cos(5.14003771959  + 4897.42439113870 *T)
    sumY0  +=       0.065 * cos(2.25884185483  + 3324.09135343631 *T)
    sumY0  +=       0.065 * cos(3.86492506381  + 3357.62113493030 *T)
    sumY0  +=       0.066 * cos(3.63859878484  + 176.44991265029 *T)
    sumY0  +=       0.069 * cos(2.88792708089  + 1442.03464492889 *T)
    sumY0  +=       0.065 * cos(3.80463962799  + 12725.20961729149 *T)
    sumY0  +=       0.071 * cos(2.38407372949  + 11250.55557593809 *T)
    sumY0  +=       0.066 * cos(5.64622594457  + 415.79630809560 *T)
    sumY0  +=       0.066 * cos(2.06957972906  + 3738.93793728871 *T)
    sumY0  +=       0.090 * cos(0.09722302162  + 2885.94687765600 *T)
    sumY0  +=       0.090 * cos(6.02654389702  + 3795.76561071060 *T)
    sumY0  +=       0.067 * cos(3.85822640732  + 13359.10309563091 *T)
    sumY0  +=       0.069 * cos(6.01059659703  + 4932.96462284910 *T)
    sumY0  +=       0.067 * cos(3.14807353187  + 20043.43074271529 *T)
    sumY0  +=       0.067 * cos(2.27076085895  + 19247.62037086590 *T)
    sumY0  +=       0.064 * cos(5.77820135380  + 8983.05448679250 *T)
    sumY0  +=       0.081 * cos(4.01879254835  + 5136.11468790430 *T)
    sumY0  +=       0.076 * cos(4.53170041047  + 6677.53651762000 *T)
    sumY0  +=       0.064 * cos(5.04465453700  + 6564.14880266290 *T)
    sumY0  +=       0.063 * cos(3.67470240475  + 10294.26641996750 *T)
    sumY0  +=       0.063 * cos(3.67470240475  + 10294.75405493450 *T)
    sumY0  +=       0.065 * cos(0.85570190902  + 0.73650358470 *T)
    sumY0  +=       0.067 * cos(5.44254364972  + 4396.71505602351 *T)
    sumY0  +=       0.070 * cos(2.61333516815  + 5088.46362233621 *T)
    sumY0  +=       0.070 * cos(2.61333516815  + 5087.97598736919 *T)
    sumY0  +=       0.066 * cos(1.94886102839  + 3332.68810760091 *T)
    sumY0  +=       0.066 * cos(4.17490589025  + 3349.02438076571 *T)
    sumY0  +=       0.069 * cos(2.29302584578  + 24076.68840560890 *T)
    sumY0  +=       0.063 * cos(2.65217434480  + 15004.02673447930 *T)
    sumY0  +=       0.066 * cos(3.29840260562  + 3303.86065357230 *T)
    sumY0  +=       0.066 * cos(2.82536431302  + 3377.85183479431 *T)
    sumY0  +=       0.062 * cos(5.68281304505  + 423.31889001050 *T)
    sumY0  +=       0.073 * cos(6.20384290738  + 7250.29054491051 *T)
    sumY0  +=       0.066 * cos(4.05777388350  + 1083.01655045349 *T)
    sumY0  +=       0.066 * cos(1.66108334221  + 3739.07255789430 *T)
    sumY0  +=       0.065 * cos(2.38868691844  + 3454.58568375171 *T)
    sumY0  +=       0.065 * cos(3.73508000020  + 3227.12680461490 *T)
    sumY0  +=       0.069 * cos(1.02568627494  + 5885.18791771890 *T)
    sumY0  +=       0.069 * cos(4.18150566046  + 2384.56708821271 *T)
    sumY0  +=       0.082 * cos(2.24915344259  + 6573.80514694450 *T)
    sumY0  +=       0.078 * cos(3.71449046077  + 12979.79699205951 *T)
    sumY0  +=       0.063 * cos(0.30909856819  + 6298.08450369290 *T)
    sumY0  +=       0.063 * cos(5.86138984897  + 379.12959639069 *T)
    sumY0  +=       0.078 * cos(4.78158403700  + 17079.07168255270 *T)
    sumY0  +=       0.063 * cos(0.84589576108  + 4295.61013381590 *T)
    sumY0  +=       0.072 * cos(1.20473950633  + 7366.51272510910 *T)
    sumY0  +=       0.064 * cos(5.94145157304  + 21258.65339700290 *T)
    sumY0  +=       0.061 * cos(4.25199542760  + 3870.52995562571 *T)
    sumY0  +=       0.069 * cos(0.74102855275  + 6757.73474215870 *T)
    sumY0  +=       0.061 * cos(6.28276705909  + 3870.56446293010 *T)
    sumY0  +=       0.061 * cos(3.89740578515  + 8528.49979454989 *T)
    sumY0  +=       0.072 * cos(6.17734111246  + 3337.95522989351 *T)
    sumY0  +=       0.072 * cos(6.22961111336  + 3343.75725847310 *T)
    sumY0  +=       0.081 * cos(3.28865783131  + 17498.55632642790 *T)
    sumY0  +=       0.080 * cos(5.66181879898  + 68241.62832713968 *T)
    sumY0  +=       0.070 * cos(5.98008002738  + 5096.32792886240 *T)
    sumY0  +=       0.070 * cos(5.98008002738  + 5095.84029389540 *T)
    sumY0  +=       0.062 * cos(2.20286177266  + 629.09540353360 *T)
    sumY0  +=       0.061 * cos(5.80236333905  + 7056.74676748629 *T)
    sumY0  +=       0.061 * cos(3.69141943753  + 27035.49639498329 *T)
    sumY0  +=       0.075 * cos(0.87745006748  + 13531.78607244869 *T)
    sumY0  +=       0.062 * cos(2.97984792859  + 14591.16800263789 *T)
    sumY0  +=       0.062 * cos(2.97984792859  + 14591.65563760490 *T)
    sumY0  +=       0.071 * cos(4.62738217375  + 5622.01943039111 *T)
    sumY0  +=       0.066 * cos(2.27365299388  + 954.99770711610 *T)
    sumY0  +=       0.068 * cos(2.72448999946  + 3930.45351370310 *T)
    sumY0  +=       0.066 * cos(3.14619736362  + 2899.54332433869 *T)
    sumY0  +=       0.061 * cos(4.49549745154  + 2711.51702316619 *T)
    sumY0  +=       0.063 * cos(4.32081273594  + 2125.56627959290 *T)
    sumY0  +=       0.073 * cos(4.45773042596  + 6695.69576488470 *T)
    sumY0  +=       0.069 * cos(5.27809995063  + 5885.15341041451 *T)
    sumY0  +=       0.060 * cos(5.14170398436  + 41427.24316569529 *T)
    sumY0  +=       0.080 * cos(3.75731973309  + 4576.44426953929 *T)
    sumY0  +=       0.080 * cos(3.75731973309  + 4576.93190450630 *T)
    sumY0  +=       0.059 * cos(2.49398932309  + 9374.58262930109 *T)
    sumY0  +=       0.074 * cos(4.89168591212  + 3583.09721319030 *T)
    sumY0  +=       0.082 * cos(5.81793237346  + 2636.73542459890 *T)
    sumY0  +=       0.058 * cos(4.78357545202  + 720.89541372269 *T)
    sumY0  +=       0.058 * cos(1.64198279843  + 721.38304868970 *T)
    sumY0  +=       0.059 * cos(3.57410159337  + 14319.44721987369 *T)
    sumY0  +=       0.065 * cos(1.28381558772  + 48274.10560037370 *T)
    sumY0  +=       0.058 * cos(4.51103536629  + 110.94282480410 *T)
    sumY0  +=       0.072 * cos(1.68586060618  + 195.21520194411 *T)
    sumY0  +=       0.059 * cos(5.74510260782  + 13738.45925585290 *T)
    sumY0  +=       0.063 * cos(1.06605769138  + 9722.95470331370 *T)
    sumY0  +=       0.080 * cos(4.45970706755  + 3304.20768125380 *T)
    sumY0  +=       0.080 * cos(1.66405985109  + 3377.50480711280 *T)
    sumY0  +=       0.058 * cos(5.44450610447  + 16070.52221166930 *T)
    sumY0  +=       0.070 * cos(2.49253748917  + 11787.84772300050 *T)
    sumY0  +=       0.062 * cos(4.94992711761  + 4558.69405718870 *T)
    sumY0  +=       0.076 * cos(1.99891799021  + 1612.70751554809 *T)
    sumY0  +=       0.071 * cos(1.30058846825  + 6486.49728642250 *T)
    sumY0  +=       0.062 * cos(5.09763426823  + 437.88770862340 *T)
    sumY0  +=       0.059 * cos(4.90581628492  + 37057.33367508269 *T)
    sumY0  +=       0.070 * cos(2.79389929588  + 8116.12869767550 *T)
    sumY0  +=       0.058 * cos(6.09640127140  + 1969.44448072730 *T)
    sumY0  +=       0.058 * cos(0.02736564724  + 4712.26800763931 *T)
    sumY0  +=       0.057 * cos(4.37594257879  + 4565.74029388670 *T)
    sumY0  +=       0.065 * cos(1.66245825930  + 5526.10751347610 *T)
    sumY0  +=       0.072 * cos(0.18252327904  + 8948.47207339210 *T)
    sumY0  +=       0.059 * cos(2.44348231635  + 22854.83983928750 *T)
    sumY0  +=       0.074 * cos(2.02616573090  + 3192.28988479000 *T)
    sumY0  +=       0.074 * cos(1.02342311752  + 3189.32093937340 *T)
    sumY0  +=       0.074 * cos(2.75351422654  + 5106.13523463390 *T)
    sumY0  +=       0.060 * cos(2.41681225251  + 3333.27756497370 *T)
    sumY0  +=       0.060 * cos(3.70695466613  + 3348.43492339290 *T)
    sumY0  +=       0.057 * cos(3.99795106682  + 1022.53274740150 *T)
    sumY0  +=       0.062 * cos(0.60417431770  + 792.53107098389 *T)
    sumY0  +=       0.055 * cos(4.34931243096  + 3355.59352650890 *T)
    sumY0  +=       0.055 * cos(1.77445448768  + 3326.11896185771 *T)
    sumY0  +=       0.057 * cos(5.12925283675  + 3973.15234852950 *T)
    sumY0  +=       0.057 * cos(1.32085031811  + 15842.99602222970 *T)
    sumY0  +=       0.071 * cos(6.21599374391  + 418.26050532349 *T)
    sumY0  +=       0.061 * cos(3.75995016598  + 223.10414047710 *T)
    sumY0  +=       0.063 * cos(4.40371062565  + 4137.17150465190 *T)
    sumY0  +=       0.067 * cos(5.00146180970  + 11240.40654555530 *T)
    sumY0  +=       0.055 * cos(5.08628762479  + 25665.76130089270 *T)
    sumY0  +=       0.076 * cos(4.50412584127  + 33716.72124838289 *T)
    sumY0  +=       0.061 * cos(0.88820799738  + 7630.64427985291 *T)
    sumY0  +=       0.065 * cos(5.53821276618  + 14267.66258598049 *T)
    sumY0  +=       0.065 * cos(5.53821276618  + 14268.15022094750 *T)
    sumY0  +=       0.060 * cos(5.95972666137  + 1368.41643536150 *T)
    sumY0  +=       0.061 * cos(3.06892273504  + 1118.99960958630 *T)
    sumY0  +=       0.067 * cos(4.29317097897  + 13366.28395293450 *T)

    # MARS Y Coord
    sumY1  = 0
    sumY1  +=  1668487.239 * cos(2.59897259786  + 0.24381748350 *T)
    sumY1  +=  551520.815 * cos(3.52285185770  + 6681.46867088311 *T)
    sumY1  +=   77122.404 * cos(3.86148242536  + 10022.08109758290 *T)
    sumY1  +=   33905.789 * cos(0.99877346654  + 3340.36860921629 *T)
    sumY1  +=   13417.687 * cos(1.14386060325  + 3340.85624418330 *T)
    sumY1  +=    9584.524 * cos(4.19977564590  + 13362.69352428270 *T)
    sumY1  +=    2619.567 * cos(1.47212497791  + 3337.33312583431 *T)
    sumY1  +=    2615.345 * cos(4.65188218049  + 3344.37936253230 *T)
    sumY1  +=    1454.057 * cos(1.29298490269  + 6680.98103591609 *T)
    sumY1  +=    1163.163 * cos(4.53798497815  + 16703.30595098250 *T)
    sumY1  +=     900.588 * cos(1.74364630959  + 1059.62574767270 *T)
    sumY1  +=     429.589 * cos(1.18548595402  + 2942.70724077510 *T)
    sumY1  +=     388.556 * cos(4.38001744017  + 5622.08674069391 *T)
    sumY1  +=     380.679 * cos(4.93305109350  + 3739.00524759150 *T)
    sumY1  +=     394.611 * cos(4.91194179182  + 6684.99178923210 *T)
    sumY1  +=     336.947 * cos(3.64146347925  + 3149.40797807170 *T)
    sumY1  +=     333.888 * cos(0.88735593348  + 3185.43584474911 *T)
    sumY1  +=     318.443 * cos(2.48094295107  + 3532.30451029490 *T)
    sumY1  +=     317.603 * cos(5.23738106920  + 3496.27664361750 *T)
    sumY1  +=     312.062 * cos(2.66000176108  + 2544.55823736690 *T)
    sumY1  +=     285.965 * cos(3.02967127901  + 796.54182429990 *T)
    sumY1  +=     259.967 * cos(3.45408083721  + 4137.15425099970 *T)
    sumY1  +=     205.452 * cos(1.63092401950  + 1194.69082770810 *T)
    sumY1  +=     140.526 * cos(4.87623075662  + 20043.91837768230 *T)
    sumY1  +=     159.180 * cos(2.00572713492  + 1589.31671276730 *T)
    sumY1  +=     133.487 * cos(3.06919175173  + 5885.17066406670 *T)
    sumY1  +=     132.246 * cos(2.12952118645  + 4400.23817437250 *T)
    sumY1  +=     126.940 * cos(1.24935817727  + 6677.94555253410 *T)
    sumY1  +=     102.941 * cos(4.48642964758  + 5487.02166065850 *T)
    sumY1  +=     128.977 * cos(0.22856181225  + 1592.83983111630 *T)
    sumY1  +=     101.403 * cos(3.97057397888  + 3339.87592311510 *T)
    sumY1  +=     101.343 * cos(2.15319078531  + 3341.83656525151 *T)
    sumY1  +=      89.600 * cos(4.30886056351  + 3870.54720927791 *T)
    sumY1  +=      84.159 * cos(5.10763376124  + 1990.98883452450 *T)
    sumY1  +=      93.032 * cos(1.57904935535  + 10021.59346261590 *T)
    sumY1  +=      79.379 * cos(2.34223792198  + 426.84200835950 *T)
    sumY1  +=      77.187 * cos(5.35374635180  + 7079.61767429131 *T)
    sumY1  +=      81.451 * cos(4.09472299340  + 5092.39577559930 *T)
    sumY1  +=      91.218 * cos(4.06344861425  + 2146.40923395870 *T)
    sumY1  +=      63.951 * cos(3.70258807947  + 2389.13783793270 *T)
    sumY1  +=      57.311 * cos(5.21513953164  + 10025.60421593190 *T)
    sumY1  +=      56.935 * cos(0.36046078739  + 3098.12764020930 *T)
    sumY1  +=      75.514 * cos(2.03966755509  + 4535.30325440790 *T)
    sumY1  +=      55.567 * cos(5.93078753620  + 9492.39013248830 *T)
    sumY1  +=      72.076 * cos(5.89419507935  + 5088.87265725030 *T)
    sumY1  +=      52.967 * cos(5.76121074267  + 3583.58484815730 *T)
    sumY1  +=      53.267 * cos(1.01723184039  + 4690.72365384210 *T)
    sumY1  +=      70.583 * cos(2.29587751894  + 2787.28684134090 *T)
    sumY1  +=      57.299 * cos(5.04645839215  + 7.35736448430 *T)
    sumY1  +=      54.075 * cos(1.17994565418  + 2957.95971196010 *T)
    sumY1  +=      50.146 * cos(6.18940453496  + 191.20444862810 *T)
    sumY1  +=      46.293 * cos(2.42348353431  + 4292.57465043390 *T)
    sumY1  +=      49.234 * cos(0.93659790719  + 6681.45141723091 *T)
    sumY1  +=      58.634 * cos(3.83006998491  + 3894.42564702570 *T)
    sumY1  +=      48.264 * cos(4.94385443663  + 3723.75277640650 *T)
    sumY1  +=      49.234 * cos(2.96736953876  + 6681.48592453530 *T)
    sumY1  +=      55.021 * cos(2.68195179354  + 155.17658195069 *T)
    sumY1  +=      47.402 * cos(4.07145408153  + 6490.02040477150 *T)
    sumY1  +=      39.275 * cos(3.34522669334  + 1747.77259558350 *T)
    sumY1  +=      37.297 * cos(4.78099832320  + 3554.15533962130 *T)
    sumY1  +=      36.066 * cos(1.31662704298  + 3127.55714874530 *T)
    sumY1  +=      35.298 * cos(1.23303085666  + 6526.04827144891 *T)
    sumY1  +=      33.577 * cos(1.93752924591  + 2145.92159899169 *T)
    sumY1  +=      32.981 * cos(3.50261324943  + 3333.74269718250 *T)
    sumY1  +=      30.372 * cos(5.46912135938  + 1748.26023055050 *T)
    sumY1  +=      39.114 * cos(3.80687421474  + 7477.76667769950 *T)
    sumY1  +=      29.430 * cos(3.77418319168  + 6254.87048000710 *T)
    sumY1  +=      29.559 * cos(6.15473365534  + 951.47458876710 *T)
    sumY1  +=      28.511 * cos(1.27708409847  + 553.32558535889 *T)
    sumY1  +=      27.567 * cos(2.17780252110  + 12303.31159409350 *T)
    sumY1  +=      27.467 * cos(0.62128414719  + 4933.45225781610 *T)
    sumY1  +=      28.130 * cos(2.67378903203  + 3347.96979118411 *T)
    sumY1  +=      34.806 * cos(4.74990712506  + 1349.62359217529 *T)
    sumY1  +=      24.557 * cos(1.59353030822  + 2811.16527908870 *T)
    sumY1  +=      24.445 * cos(4.65651492351  + 1066.73929467350 *T)
    sumY1  +=      29.230 * cos(1.08762035239  + 640.14110379750 *T)
    sumY1  +=      21.272 * cos(2.27427092663  + 4929.92913946711 *T)
    sumY1  +=      24.027 * cos(2.31810288444  + 382.65271473970 *T)
    sumY1  +=      24.805 * cos(2.29421805969  + 2119.00767786191 *T)
    sumY1  +=      19.016 * cos(5.33725266506  + 242.97242145751 *T)
    sumY1  +=      18.321 * cos(5.32154407296  + 9623.93209417470 *T)
    sumY1  +=      17.069 * cos(0.73723843575  + 3320.50092478451 *T)
    sumY1  +=      16.952 * cos(5.38620600325  + 3361.21156358210 *T)
    sumY1  +=      16.989 * cos(5.21436339395  + 23384.53080438210 *T)
    sumY1  +=      16.214 * cos(5.29782106321  + 6681.40136058031 *T)
    sumY1  +=      16.696 * cos(3.94416253958  + 2281.47431399410 *T)
    sumY1  +=      16.004 * cos(5.16305438762  + 1052.51220067191 *T)
    sumY1  +=      17.047 * cos(1.13159672552  + 6674.35512388230 *T)
    sumY1  +=      14.590 * cos(0.35043364178  + 10018.55797923390 *T)
    sumY1  +=      16.713 * cos(5.38131437608  + 7211.15963597770 *T)
    sumY1  +=      17.707 * cos(2.46690015193  + 7740.85060107230 *T)
    sumY1  +=      16.047 * cos(2.67804298545  + 3325.60377299830 *T)
    sumY1  +=      15.961 * cos(3.44566203623  + 3356.10871536831 *T)
    sumY1  +=      16.214 * cos(4.88932467996  + 6681.53598118590 *T)
    sumY1  +=      13.780 * cos(6.26179574629  + 12833.00255918811 *T)
    sumY1  +=      14.630 * cos(2.07693903473  + 2273.87313202630 *T)
    sumY1  +=      12.478 * cos(2.52316698362  + 6682.44899195131 *T)
    sumY1  +=      12.677 * cos(5.48459046233  + 5331.60126122430 *T)
    sumY1  +=      13.489 * cos(1.61171597857  + 2288.10022602790 *T)
    sumY1  +=      15.563 * cos(3.79290468512  + 4562.70481050471 *T)
    sumY1  +=      13.052 * cos(2.92086754961  + 3767.45443505930 *T)
    sumY1  +=      11.294 * cos(0.59272382037  + 1350.11122714230 *T)
    sumY1  +=      11.234 * cos(0.74117477922  + 6438.74006690910 *T)
    sumY1  +=      12.654 * cos(0.84471005653  + 529.44714761109 *T)
    sumY1  +=      12.118 * cos(0.35273968830  + 537.04832957890 *T)
    sumY1  +=      11.323 * cos(0.88738106730  + 522.82123557730 *T)
    sumY1  +=      10.300 * cos(2.38634995278  + 7875.91568110771 *T)
    sumY1  +=      10.510 * cos(1.46312475386  + 5614.97319369310 *T)
    sumY1  +=      12.502 * cos(5.71659694947  + 10420.23010099111 *T)
    sumY1  +=      10.338 * cos(3.30466885363  + 10022.09835123510 *T)
    sumY1  +=       8.834 * cos(6.25318290526  + 3205.79116414790 *T)
    sumY1  +=       9.393 * cos(5.31958384512  + 6836.88907031730 *T)
    sumY1  +=      11.331 * cos(5.03073170237  + 6041.57138456910 *T)
    sumY1  +=       8.381 * cos(6.15334573581  + 3475.92132421871 *T)
    sumY1  +=       9.243 * cos(5.01043204259  + 529.93478257810 *T)
    sumY1  +=      10.338 * cos(1.27389725279  + 10022.06384393071 *T)
    sumY1  +=       7.850 * cos(5.83678641323  + 2700.47132290230 *T)
    sumY1  +=       8.102 * cos(5.53289540376  + 13366.21664263171 *T)
    sumY1  +=       7.470 * cos(5.23301908403  + 397.90518592470 *T)
    sumY1  +=       7.391 * cos(2.50557265473  + 15643.92402079330 *T)
    sumY1  +=       7.798 * cos(2.64875005153  + 6127.89926804070 *T)
    sumY1  +=       7.318 * cos(5.33110036381  + 1751.29571393250 *T)
    sumY1  +=       7.398 * cos(2.40080315235  + 2942.21960580809 *T)
    sumY1  +=       7.000 * cos(2.24141414910  + 3343.89172756530 *T)
    sumY1  +=       6.961 * cos(0.94833631216  + 5629.20028769470 *T)
    sumY1  +=       7.023 * cos(1.85516355349  + 13362.20588931570 *T)
    sumY1  +=       6.810 * cos(5.19498826700  + 4690.23601887509 *T)
    sumY1  +=       6.681 * cos(3.04162525271  + 3178.38960805111 *T)
    sumY1  +=       6.243 * cos(1.53240997706  + 6298.57213865991 *T)
    sumY1  +=       6.294 * cos(3.08194615517  + 3503.32288031550 *T)
    sumY1  +=       6.358 * cos(4.02054515470  + 3316.97780643551 *T)
    sumY1  +=       7.854 * cos(4.06615442184  + 5729.75026463250 *T)
    sumY1  +=       6.174 * cos(3.80238057009  + 5088.38502228330 *T)
    sumY1  +=       6.298 * cos(2.10322094908  + 3364.73468193110 *T)
    sumY1  +=       6.734 * cos(4.45499670640  + 9830.63283147130 *T)
    sumY1  +=       7.590 * cos(3.34937114207  + 8962.69916739370 *T)
    sumY1  +=       6.074 * cos(3.57121700387  + 6151.77770578850 *T)
    sumY1  +=       7.279 * cos(5.35063229510  + 6894.76776632111 *T)
    sumY1  +=       6.015 * cos(0.07301799575  + 853.44019923550 *T)
    sumY1  +=       5.586 * cos(4.15031997147  + 10818.37910439930 *T)
    sumY1  +=       5.539 * cos(2.54784312119  + 1758.40926093329 *T)
    sumY1  +=       5.309 * cos(2.65128148644  + 2648.69864295650 *T)
    sumY1  +=       6.531 * cos(4.25297171807  + 6680.48834981490 *T)
    sumY1  +=       5.498 * cos(6.10169669980  + 6924.19727485710 *T)
    sumY1  +=       5.070 * cos(4.67900323296  + 1596.43025976811 *T)
    sumY1  +=       4.841 * cos(5.17216869184  + 6158.89125278930 *T)
    sumY1  +=       5.011 * cos(2.03179616213  + 433.95555536030 *T)
    sumY1  +=       5.999 * cos(0.30491904963  + 2280.98667902709 *T)
    sumY1  +=       4.334 * cos(5.65639869182  + 12964.54452087450 *T)
    sumY1  +=       5.929 * cos(0.88326220381  + 796.05418933290 *T)
    sumY1  +=       4.781 * cos(5.75441905529  + 6144.66415878770 *T)
    sumY1  +=       4.633 * cos(2.64397173207  + 2914.25805330730 *T)
    sumY1  +=       4.948 * cos(5.71673408235  + 213.05527795450 *T)
    sumY1  +=       3.970 * cos(5.05440080471  + 3312.40705671550 *T)
    sumY1  +=       4.094 * cos(2.94991179870  + 2810.67764412170 *T)
    sumY1  +=       3.913 * cos(1.06659787427  + 3369.30543165110 *T)
    sumY1  +=       3.870 * cos(3.93077011499  + 641.12142486571 *T)
    sumY1  +=       4.135 * cos(3.45059100231  + 4033.01384541010 *T)
    sumY1  +=       5.101 * cos(0.31450098972  + 4292.08701546690 *T)
    sumY1  +=       3.721 * cos(3.02454190848  + 3346.37916849070 *T)
    sumY1  +=       3.721 * cos(3.09922501016  + 3335.33331987590 *T)
    sumY1  +=       4.814 * cos(3.06998978800  + 6688.58221788390 *T)
    sumY1  +=       3.684 * cos(5.49241454553  + 5621.59910572690 *T)
    sumY1  +=       3.561 * cos(0.96460018959  + 8274.06468451590 *T)
    sumY1  +=       4.122 * cos(1.57465436330  + 9866.66069814871 *T)
    sumY1  +=       3.629 * cos(5.97000555701  + 3304.82837750590 *T)
    sumY1  +=       4.333 * cos(3.32433333179  + 6.86972951729 *T)
    sumY1  +=       4.795 * cos(2.00069285188  + 951.96222373410 *T)
    sumY1  +=       3.622 * cos(4.42697806631  + 1582.20316576650 *T)
    sumY1  +=       3.577 * cos(0.15377977803  + 3376.88411086071 *T)
    sumY1  +=       4.037 * cos(4.02725434804  + 8433.00820229911 *T)
    sumY1  +=       3.824 * cos(1.42167482126  + 8969.81271439450 *T)
    sumY1  +=       3.414 * cos(5.22862391511  + 10022.14840788570 *T)
    sumY1  +=       4.384 * cos(2.51318285232  + 11773.62062899890 *T)
    sumY1  +=       4.071 * cos(1.88751920598  + 3863.43366227710 *T)
    sumY1  +=       3.905 * cos(1.88201830658  + 8955.58562039291 *T)
    sumY1  +=       4.291 * cos(4.14240496655  + 2818.27882608950 *T)
    sumY1  +=       3.570 * cos(2.45586996283  + 1744.18216693170 *T)
    sumY1  +=       3.461 * cos(6.16085694117  + 7633.18707713370 *T)
    sumY1  +=       3.246 * cos(5.21859325685  + 149.80701461810 *T)
    sumY1  +=       3.160 * cos(1.61898299169  + 3980.75353049730 *T)
    sumY1  +=       4.113 * cos(3.93628740170  + 213.54291292150 *T)
    sumY1  +=       3.117 * cos(2.24732563525  + 3336.97490882530 *T)
    sumY1  +=       3.117 * cos(3.87644128339  + 3344.73757954131 *T)
    sumY1  +=       3.201 * cos(1.72052248020  + 3893.93801205869 *T)
    sumY1  +=       3.517 * cos(2.23980933192  + 7064.36520310630 *T)
    sumY1  +=       3.078 * cos(1.33120857361  + 7235.03807372550 *T)
    sumY1  +=       3.414 * cos(5.63720914737  + 10022.01378728011 *T)
    sumY1  +=       2.996 * cos(2.48679790692  + 419.72846135871 *T)
    sumY1  +=       3.014 * cos(5.98273733492  + 10551.77206267750 *T)
    sumY1  +=       3.089 * cos(0.21765550994  + 13359.17040593370 *T)
    sumY1  +=       3.393 * cos(3.36919512514  + 3336.84549086730 *T)
    sumY1  +=       2.930 * cos(0.54631719683  + 2487.17222746430 *T)
    sumY1  +=       2.899 * cos(4.81760357159  + 155.66421691770 *T)
    sumY1  +=       2.868 * cos(2.58098704577  + 8270.54156616690 *T)
    sumY1  +=       3.341 * cos(4.81114285872  + 8031.33608054190 *T)
    sumY1  +=       2.771 * cos(0.78165715714  + 3.27930086550 *T)
    sumY1  +=       3.548 * cos(1.76914289559  + 2699.97863680110 *T)
    sumY1  +=       3.139 * cos(2.49726447663  + 5459.62010456170 *T)
    sumY1  +=       2.993 * cos(4.48513822981  + 3.76693583251 *T)
    sumY1  +=       2.558 * cos(0.74612048217  + 7632.69944216670 *T)
    sumY1  +=       2.973 * cos(6.11907591951  + 398.39282089170 *T)
    sumY1  +=       2.555 * cos(3.40863603224  + 553.81322032590 *T)
    sumY1  +=       2.436 * cos(2.68110734466  + 3723.26514143949 *T)
    sumY1  +=       3.398 * cos(2.01609731704  + 9225.78309076650 *T)
    sumY1  +=       2.528 * cos(4.21676297033  + 3333.81000748530 *T)
    sumY1  +=       2.381 * cos(6.11405773217  + 1221.60474883790 *T)
    sumY1  +=       2.367 * cos(3.67472867066  + 8429.48508395010 *T)
    sumY1  +=       2.452 * cos(4.92871549492  + 2766.51144584850 *T)
    sumY1  +=       3.001 * cos(4.34751038511  + 3981.73385156551 *T)
    sumY1  +=       2.325 * cos(5.58141344460  + 9382.18381126891 *T)
    sumY1  +=       3.095 * cos(0.15312862288  + 6283.31966747490 *T)
    sumY1  +=       2.582 * cos(1.41863449393  + 5085.28222859850 *T)
    sumY1  +=       2.407 * cos(5.52690969383  + 9808.78200214491 *T)
    sumY1  +=       2.189 * cos(5.03087631941  + 4407.35172137330 *T)
    sumY1  +=       2.276 * cos(5.75323373198  + 1194.20319274110 *T)
    sumY1  +=       2.512 * cos(0.31371446538  + 16173.61498588790 *T)
    sumY1  +=       2.182 * cos(3.76616453203  + 1751.78334889950 *T)
    sumY1  +=       2.219 * cos(5.63570258068  + 8030.84844557490 *T)
    sumY1  +=       2.433 * cos(2.80111269508  + 11081.46302777211 *T)
    sumY1  +=       2.433 * cos(6.03898534156  + 5828.27228913110 *T)
    sumY1  +=       2.109 * cos(2.06281789191  + 15114.23305569870 *T)
    sumY1  +=       2.719 * cos(2.13406336652  + 6468.16957544510 *T)
    sumY1  +=       2.025 * cos(1.20186617110  + 3915.20104251810 *T)
    sumY1  +=       1.887 * cos(4.88253130307  + 76.50988875911 *T)
    sumY1  +=       1.853 * cos(3.38166474355  + 6151.29007082150 *T)
    sumY1  +=       1.844 * cos(1.21875300270  + 1692.40948698591 *T)
    sumY1  +=       2.127 * cos(1.64433014539  + 3347.90248088131 *T)
    sumY1  +=       1.808 * cos(2.17690241100  + 6040.59106350090 *T)
    sumY1  +=       1.865 * cos(6.06509881975  + 13760.84252769090 *T)
    sumY1  +=       1.998 * cos(1.41278142272  + 5084.79459363150 *T)
    sumY1  +=       1.821 * cos(1.67609459019  + 5099.50932260011 *T)
    sumY1  +=       1.970 * cos(3.64755835075  + 8827.63408735830 *T)
    sumY1  +=       1.948 * cos(0.91793732483  + 5099.02168763310 *T)
    sumY1  +=       1.717 * cos(3.64496480459  + 13362.71077793490 *T)
    sumY1  +=       1.785 * cos(0.87660197618  + 6660.69327539070 *T)
    sumY1  +=       1.744 * cos(0.22355262774  + 6816.53375091851 *T)
    sumY1  +=       1.717 * cos(1.61419317923  + 13362.67627063050 *T)
    sumY1  +=       1.955 * cos(2.06919305164  + 191.69208359510 *T)
    sumY1  +=       1.839 * cos(3.74018886236  + 220.65645992230 *T)
    sumY1  +=       1.739 * cos(2.07821758598  + 3281.48238226970 *T)
    sumY1  +=       2.059 * cos(5.55274042469  + 26725.14323108190 *T)
    sumY1  +=       1.828 * cos(4.08820612940  + 6247.75693300630 *T)
    sumY1  +=       2.136 * cos(0.14789369077  + 6872.91693699471 *T)
    sumY1  +=       1.698 * cos(4.04558966245  + 3400.23010609691 *T)
    sumY1  +=       1.534 * cos(4.85765567337  + 574.10098085129 *T)
    sumY1  +=       1.533 * cos(1.46545387500  + 3877.66075627870 *T)
    sumY1  +=       1.541 * cos(3.10239552544  + 3495.78900865049 *T)
    sumY1  +=       1.628 * cos(2.88073838002  + 10023.06141865111 *T)
    sumY1  +=       1.807 * cos(3.79979542581  + 6696.72114206811 *T)
    sumY1  +=       1.767 * cos(1.07911547450  + 6661.11335148431 *T)
    sumY1  +=       1.525 * cos(4.40073811310  + 206.42936592071 *T)
    sumY1  +=       1.957 * cos(2.14746864610  + 7234.55043875850 *T)
    sumY1  +=       1.381 * cos(2.72417969364  + 11216.52810780750 *T)
    sumY1  +=       1.688 * cos(3.32715704311  + 7108.06686175911 *T)
    sumY1  +=       1.517 * cos(0.44332021975  + 3340.38586286849 *T)
    sumY1  +=       1.506 * cos(2.34509908751  + 8440.12174929990 *T)
    sumY1  +=       1.379 * cos(2.83386638398  + 3127.06951377830 *T)
    sumY1  +=       1.360 * cos(0.70558062923  + 3074.24920246151 *T)
    sumY1  +=       1.516 * cos(4.69574172401  + 3340.35135556410 *T)
    sumY1  +=       1.302 * cos(2.98694119189  + 6666.21619969810 *T)
    sumY1  +=       1.588 * cos(5.82625124459  + 8672.21368792411 *T)
    sumY1  +=       1.305 * cos(5.68793557454  + 6701.82399028190 *T)
    sumY1  +=       1.529 * cos(5.28219831652  + 3547.04179262050 *T)
    sumY1  +=       1.317 * cos(3.02872149081  + 7063.87756813929 *T)
    sumY1  +=       1.316 * cos(1.55707216163  + 8432.52056733209 *T)
    sumY1  +=       1.510 * cos(0.81852178414  + 3134.67069574611 *T)
    sumY1  +=       1.232 * cos(2.92729735592  + 1228.71829583870 *T)
    sumY1  +=       1.498 * cos(1.47147884583  + 10014.96755058210 *T)
    sumY1  +=       1.240 * cos(3.54817212305  + 6836.40143535029 *T)
    sumY1  +=       1.407 * cos(2.83762149372  + 18984.53644749310 *T)
    sumY1  +=       1.219 * cos(5.41832780376  + 3607.46328590511 *T)
    sumY1  +=       1.443 * cos(6.05937575283  + 1059.13811270570 *T)
    sumY1  +=       1.157 * cos(5.97375778288  + 110.45013870291 *T)
    sumY1  +=       1.195 * cos(5.00661978579  + 11371.94850724170 *T)
    sumY1  +=       1.173 * cos(2.57988132706  + 10575.16286545830 *T)
    sumY1  +=       1.175 * cos(0.33045598221  + 3320.08084869090 *T)
    sumY1  +=       1.171 * cos(5.78563510490  + 3361.63163967570 *T)
    sumY1  +=       1.509 * cos(3.03783113783  + 6702.24406637550 *T)
    sumY1  +=       1.157 * cos(5.78853307616  + 10235.38019302090 *T)
    sumY1  +=       1.097 * cos(1.70038513078  + 12936.09533340670 *T)
    sumY1  +=       1.517 * cos(4.49354530734  + 3097.64000524229 *T)
    sumY1  +=       1.089 * cos(0.71436474935  + 3170.18337356410 *T)
    sumY1  +=       1.068 * cos(2.16721753485  + 6685.05909953490 *T)
    sumY1  +=       1.082 * cos(0.90416583587  + 6531.90547374850 *T)
    sumY1  +=       1.365 * cos(2.52562707978  + 1039.27042827391 *T)
    sumY1  +=       1.415 * cos(1.04940944586  + 647.25465079831 *T)
    sumY1  +=       1.028 * cos(3.74893297531  + 6261.98402700790 *T)
    sumY1  +=       1.020 * cos(5.40926634456  + 3511.52911480250 *T)
    sumY1  +=       1.072 * cos(1.73998305445  + 6677.87824223130 *T)
    sumY1  +=       0.994 * cos(5.42870395574  + 2921.37160030811 *T)
    sumY1  +=       0.960 * cos(4.58361434342  + 383.14034970670 *T)
    sumY1  +=       1.295 * cos(2.76284732234  + 3738.51761262449 *T)
    sumY1  +=       1.081 * cos(4.69614075629  + 2906.65687133950 *T)
    sumY1  +=       0.958 * cos(4.89612401533  + 4989.30300138070 *T)
    sumY1  +=       0.985 * cos(2.19704028035  + 8425.89465529830 *T)
    sumY1  +=       0.957 * cos(1.53635048068  + 10575.65050042530 *T)
    sumY1  +=       0.922 * cos(2.99315435276  + 9468.51169474051 *T)
    sumY1  +=       0.997 * cos(4.10815936593  + 4562.21717553770 *T)
    sumY1  +=       0.918 * cos(4.81950299170  + 13171.24525817110 *T)
    sumY1  +=       0.929 * cos(1.18788761625  + 10973.31186886650 *T)
    sumY1  +=       0.902 * cos(4.76870494203  + 3561.26888662211 *T)
    sumY1  +=       1.212 * cos(3.94993855439  + 2804.05173208791 *T)
    sumY1  +=       0.887 * cos(0.78830627399  + 3760.34088805850 *T)
    sumY1  +=       0.876 * cos(0.12213576951  + 956.53297345411 *T)
    sumY1  +=       1.037 * cos(1.02819591637  + 9779.35249360890 *T)
    sumY1  +=       0.929 * cos(4.41328786684  + 9070.36269133230 *T)
    sumY1  +=       0.857 * cos(4.33561652540  + 1592.35219614930 *T)
    sumY1  +=       0.840 * cos(4.83838107436  + 2126.12122486270 *T)
    sumY1  +=       0.845 * cos(1.24696667770  + 3120.44360174450 *T)
    sumY1  +=       0.954 * cos(4.20015685079  + 2920.88396534110 *T)
    sumY1  +=       0.880 * cos(0.90436690304  + 3313.45468808651 *T)
    sumY1  +=       1.010 * cos(5.76000534522  + 3253.54803964350 *T)
    sumY1  +=       1.124 * cos(0.12127908301  + 10973.79950383350 *T)
    sumY1  +=       0.870 * cos(5.21940272331  + 3368.25780028011 *T)
    sumY1  +=       0.983 * cos(2.93323832693  + 9595.48290670690 *T)
    sumY1  +=       0.852 * cos(3.66784511212  + 11770.09751064990 *T)
    sumY1  +=       0.833 * cos(5.80230695626  + 8962.21153242670 *T)
    sumY1  +=       0.824 * cos(4.16051327422  + 1861.50203515190 *T)
    sumY1  +=       0.974 * cos(0.36383884525  + 3428.16444872311 *T)
    sumY1  +=       1.121 * cos(5.85882470690  + 16706.82906933150 *T)
    sumY1  +=       0.794 * cos(0.32208800868  + 9168.88471583091 *T)
    sumY1  +=       1.033 * cos(5.81660979366  + 5091.90814063229 *T)
    sumY1  +=       0.766 * cos(1.17417157046  + 3443.94901840191 *T)
    sumY1  +=       0.797 * cos(2.84449171095  + 6282.83203250789 *T)
    sumY1  +=       0.758 * cos(4.49492748161  + 14158.99153109910 *T)
    sumY1  +=       0.758 * cos(2.48905294849  + 10404.97762980610 *T)
    sumY1  +=       0.769 * cos(1.88091311578  + 9639.18456535971 *T)
    sumY1  +=       0.762 * cos(5.92055363129  + 2699.49100183409 *T)
    sumY1  +=       0.740 * cos(4.94266271898  + 6438.25243194210 *T)
    sumY1  +=       0.761 * cos(1.45394111452  + 633.02755679670 *T)
    sumY1  +=       0.800 * cos(3.22073412970  + 6519.00203475090 *T)
    sumY1  +=       0.725 * cos(1.60564980223  + 6475.28312244591 *T)
    sumY1  +=       0.730 * cos(1.45023184025  + 6709.91785835091 *T)
    sumY1  +=       0.766 * cos(1.52658947730  + 3148.92034310470 *T)
    sumY1  +=       0.795 * cos(5.99980742469  + 16305.15694757430 *T)
    sumY1  +=       0.701 * cos(1.70752640570  + 4569.33072253850 *T)
    sumY1  +=       0.750 * cos(4.51008391137  + 3297.13733187830 *T)
    sumY1  +=       0.762 * cos(5.03600553177  + 2544.07060239989 *T)
    sumY1  +=       0.695 * cos(4.07175238073  + 2700.95895786930 *T)
    sumY1  +=       0.688 * cos(3.43319222752  + 10029.19464458370 *T)
    sumY1  +=       0.923 * cos(3.97995044570  + 10177.01386205009 *T)
    sumY1  +=       0.737 * cos(1.61369812975  + 3384.57515648830 *T)
    sumY1  +=       0.674 * cos(1.39948916768  + 2807.64216073970 *T)
    sumY1  +=       0.845 * cos(5.55184542363  + 9499.50367948910 *T)
    sumY1  +=       0.645 * cos(2.88294630681  + 1214.49120183710 *T)
    sumY1  +=       0.629 * cos(5.00231468326  + 3237.76346996470 *T)
    sumY1  +=       0.627 * cos(5.82288957282  + 13149.39442884470 *T)
    sumY1  +=       0.632 * cos(1.23957157188  + 6605.20259960750 *T)
    sumY1  +=       0.710 * cos(2.32417028248  + 6461.05602844430 *T)
    sumY1  +=       0.701 * cos(1.02423412249  + 3370.28575271930 *T)
    sumY1  +=       0.701 * cos(5.09953279615  + 3311.42673564730 *T)
    sumY1  +=       0.607 * cos(4.60165319406  + 949.41942645330 *T)
    sumY1  +=       0.603 * cos(1.15354487403  + 12566.39551746630 *T)
    sumY1  +=       0.605 * cos(2.98188951224  + 11243.92966390430 *T)
    sumY1  +=       0.719 * cos(1.82710736781  + 12310.42514109431 *T)
    sumY1  +=       0.585 * cos(4.85846707144  + 5827.78465416409 *T)
    sumY1  +=       0.622 * cos(4.05950257993  + 7903.31723720450 *T)
    sumY1  +=       0.633 * cos(0.43215822310  + 691.91378374330 *T)
    sumY1  +=       0.601 * cos(5.38619788070  + 9778.86485864190 *T)
    sumY1  +=       0.580 * cos(2.11695908203  + 16702.81831601549 *T)
    sumY1  +=       0.598 * cos(0.58815457615  + 3340.87349783550 *T)
    sumY1  +=       0.587 * cos(4.41097801741  + 13517.62628874989 *T)
    sumY1  +=       0.646 * cos(4.20454528693  + 2111.89413086110 *T)
    sumY1  +=       0.592 * cos(3.16453997147  + 2301.82963339290 *T)
    sumY1  +=       0.598 * cos(4.84056601855  + 3340.83899053111 *T)
    sumY1  +=       0.551 * cos(5.78297804099  + 2693.35777590150 *T)
    sumY1  +=       0.544 * cos(0.40037320261  + 16699.78283263350 *T)
    sumY1  +=       0.623 * cos(0.15815938701  + 10264.80970155691 *T)
    sumY1  +=       0.539 * cos(5.53287635175  + 3490.41944131790 *T)
    sumY1  +=       0.568 * cos(1.11670802046  + 1437.41943168211 *T)
    sumY1  +=       0.557 * cos(5.56701497497  + 13362.76083458550 *T)
    sumY1  +=       0.530 * cos(1.26498700457  + 4555.59126350391 *T)
    sumY1  +=       0.647 * cos(5.65099181873  + 4393.12462737170 *T)
    sumY1  +=       0.575 * cos(0.86535292594  + 5884.68302909969 *T)
    sumY1  +=       0.529 * cos(4.63607035545  + 1376.01761732930 *T)
    sumY1  +=       0.537 * cos(4.72219269506  + 3874.07032762691 *T)
    sumY1  +=       0.584 * cos(0.55160568707  + 4194.05262593530 *T)
    sumY1  +=       0.507 * cos(3.76393309357  + 2177.89390480850 *T)
    sumY1  +=       0.505 * cos(6.09136336929  + 9485.27658548750 *T)
    sumY1  +=       0.510 * cos(1.03099721510  + 11243.44202893729 *T)
    sumY1  +=       0.528 * cos(1.66505268096  + 4555.10362853689 *T)
    sumY1  +=       0.499 * cos(2.36218211072  + 3340.43591951909 *T)
    sumY1  +=       0.501 * cos(1.36886135802  + 224.58861318540 *T)
    sumY1  +=       0.525 * cos(2.51966753829  + 6684.50415426510 *T)
    sumY1  +=       0.670 * cos(4.37490154824  + 6657.59023313531 *T)
    sumY1  +=       0.634 * cos(3.19136146041  + 242.48478649050 *T)
    sumY1  +=       0.513 * cos(3.36933959417  + 3233.19272024471 *T)
    sumY1  +=       0.490 * cos(2.60477713212  + 3774.56798206011 *T)
    sumY1  +=       0.634 * cos(5.26965239400  + 2707.58486990309 *T)
    sumY1  +=       0.613 * cos(2.82952730632  + 5989.31106965630 *T)
    sumY1  +=       0.492 * cos(2.75435833662  + 3448.51976812190 *T)
    sumY1  +=       0.616 * cos(3.59514782008  + 5642.44206009270 *T)
    sumY1  +=       0.555 * cos(2.23497332858  + 18454.84548239850 *T)
    sumY1  +=       0.463 * cos(6.10086251037  + 3229.66960189570 *T)
    sumY1  +=       0.507 * cos(1.16009561745  + 4136.66661603269 *T)
    sumY1  +=       0.477 * cos(2.09135717365  + 12296.19804709270 *T)
    sumY1  +=       0.462 * cos(2.91184106959  + 8827.14645239129 *T)
    sumY1  +=       0.444 * cos(2.38797656832  + 6705.34710863090 *T)
    sumY1  +=       0.549 * cos(4.75768179775  + 3333.25506221550 *T)
    sumY1  +=       0.443 * cos(0.02288008382  + 3452.04288647091 *T)
    sumY1  +=       0.557 * cos(5.97553834508  + 13362.62621397991 *T)
    sumY1  +=       0.437 * cos(3.02498984395  + 13915.77529215809 *T)
    sumY1  +=       0.498 * cos(2.76996723863  + 3340.30129891350 *T)
    sumY1  +=       0.424 * cos(1.01652866287  + 2301.34199842589 *T)
    sumY1  +=       0.433 * cos(2.91189244542  + 6843.93530701530 *T)
    sumY1  +=       0.494 * cos(1.91460459095  + 13207.27312484851 *T)
    sumY1  +=       0.402 * cos(0.14915077619  + 6571.26234966370 *T)
    sumY1  +=       0.453 * cos(4.81695891547  + 3184.94820978210 *T)
    sumY1  +=       0.413 * cos(2.77396890115  + 8800.23253126150 *T)
    sumY1  +=       0.407 * cos(4.51107276760  + 6048.68493156990 *T)
    sumY1  +=       0.488 * cos(1.41132277574  + 2222.10045208050 *T)
    sumY1  +=       0.556 * cos(5.04224552796  + 6034.45783756830 *T)
    sumY1  +=       0.421 * cos(2.92790084389  + 4379.88285497370 *T)
    sumY1  +=       0.472 * cos(5.05170315567  + 4937.04268646790 *T)
    sumY1  +=       0.392 * cos(0.42258961591  + 134.82126255190 *T)
    sumY1  +=       0.515 * cos(3.61483236959  + 6675.94574657570 *T)
    sumY1  +=       0.407 * cos(0.15253113909  + 3392.13658204571 *T)
    sumY1  +=       0.407 * cos(5.97123577955  + 3289.57590632090 *T)
    sumY1  +=       0.474 * cos(0.17716199109  + 13892.38448937730 *T)
    sumY1  +=       0.426 * cos(1.29874033367  + 11614.67711121570 *T)
    sumY1  +=       0.384 * cos(4.23498307074  + 6685.35000624110 *T)
    sumY1  +=       0.407 * cos(1.89011389134  + 4569.81835750550 *T)
    sumY1  +=       0.495 * cos(5.80210043627  + 13119.47728534169 *T)
    sumY1  +=       0.366 * cos(1.32035636914  + 263.32774085630 *T)
    sumY1  +=       0.370 * cos(6.07332435935  + 3274.36883526891 *T)
    sumY1  +=       0.359 * cos(5.81278544284  + 5614.48555872609 *T)
    sumY1  +=       0.358 * cos(2.73133393931  + 7910.43078420530 *T)
    sumY1  +=       0.360 * cos(0.05052083677  + 3407.34365309770 *T)
    sumY1  +=       0.402 * cos(0.72673376114  + 7895.71605523670 *T)
    sumY1  +=       0.453 * cos(0.52830795973  + 3113.38011139430 *T)
    sumY1  +=       0.394 * cos(4.52689788759  + 9588.36935970610 *T)
    sumY1  +=       0.341 * cos(3.01339564635  + 15.49628866851 *T)
    sumY1  +=       0.341 * cos(6.24862667433  + 2384.07945324570 *T)
    sumY1  +=       0.358 * cos(2.84383303933  + 20.11150191529 *T)
    sumY1  +=       0.352 * cos(0.94525987632  + 12721.32828193349 *T)
    sumY1  +=       0.422 * cos(0.62253449322  + 19514.22741258770 *T)
    sumY1  +=       0.463 * cos(3.67985340584  + 3870.05957431090 *T)
    sumY1  +=       0.334 * cos(5.92598354309  + 10595.93826095070 *T)
    sumY1  +=       0.333 * cos(6.01075923274  + 5725.17951491250 *T)
    sumY1  +=       0.342 * cos(0.47650593061  + 9380.71585523369 *T)
    sumY1  +=       0.373 * cos(4.44069988433  + 10021.10077651470 *T)
    sumY1  +=       0.380 * cos(0.32090210218  + 7909.94314923829 *T)
    sumY1  +=       0.375 * cos(2.90510047685  + 11611.15399286670 *T)
    sumY1  +=       0.335 * cos(2.98607740801  + 2939.18412242610 *T)
    sumY1  +=       0.340 * cos(2.07426818439  + 7321.36595719711 *T)
    sumY1  +=       0.374 * cos(2.46295414488  + 3415.63784275061 *T)
    sumY1  +=       0.330 * cos(3.45448516287  + 6677.45791756710 *T)
    sumY1  +=       0.412 * cos(5.59490712507  + 3568.33237697230 *T)
    sumY1  +=       0.310 * cos(3.13768335245  + 14422.07545447190 *T)
    sumY1  +=       0.318 * cos(6.27020713981  + 16460.08971204149 *T)
    sumY1  +=       0.338 * cos(4.34051713830  + 10177.50149701710 *T)
    sumY1  +=       0.330 * cos(5.62835935709  + 3362.70707350971 *T)
    sumY1  +=       0.330 * cos(0.49540756155  + 3319.00541485690 *T)
    sumY1  +=       0.348 * cos(4.04651329620  + 9602.59645370770 *T)
    sumY1  +=       0.329 * cos(3.37223997989  + 10042.85649307530 *T)
    sumY1  +=       0.312 * cos(3.98131656363  + 323.74923414091 *T)
    sumY1  +=       0.340 * cos(3.21428644968  + 22325.14887419290 *T)
    sumY1  +=       0.403 * cos(0.06698481210  + 6040.10342853389 *T)
    sumY1  +=       0.390 * cos(0.53675687816  + 6717.49653756051 *T)
    sumY1  +=       0.324 * cos(2.24331129239  + 14584.54209060410 *T)
    sumY1  +=       0.298 * cos(5.27497484425  + 3417.12231545890 *T)
    sumY1  +=       0.291 * cos(6.25952120602  + 6645.44080420570 *T)
    sumY1  +=       0.293 * cos(2.11859534056  + 2676.10019905331 *T)
    sumY1  +=       0.332 * cos(4.66699998645  + 4459.61203628611 *T)
    sumY1  +=       0.279 * cos(1.51432362320  + 5732.29306191330 *T)
    sumY1  +=       0.304 * cos(2.34814397808  + 7204.04608897691 *T)
    sumY1  +=       0.280 * cos(6.06841216963  + 3283.95786924770 *T)
    sumY1  +=       0.280 * cos(0.05535474901  + 3397.75461911890 *T)
    sumY1  +=       0.283 * cos(5.58365349934  + 262.84010588929 *T)
    sumY1  +=       0.325 * cos(3.66770932208  + 3266.07464561600 *T)
    sumY1  +=       0.280 * cos(1.83816934995  + 6688.51490758111 *T)
    sumY1  +=       0.284 * cos(5.01320919080  + 5244.29305668450 *T)
    sumY1  +=       0.363 * cos(1.22201726104  + 10001.30570209050 *T)
    sumY1  +=       0.267 * cos(4.73707730483  + 7255.81346921790 *T)
    sumY1  +=       0.348 * cos(2.92764569897  + 1990.50119955750 *T)
    sumY1  +=       0.329 * cos(0.89716729335  + 162.22281864870 *T)
    sumY1  +=       0.356 * cos(1.66181940614  + 1162.71852189130 *T)
    sumY1  +=       0.345 * cos(4.96215457144  + 426.35437339250 *T)
    sumY1  +=       0.312 * cos(5.57285955038  + 7255.32583425090 *T)
    sumY1  +=       0.260 * cos(1.67326547526  + 11766.50708199810 *T)
    sumY1  +=       0.312 * cos(5.27566213630  + 3914.71340755109 *T)
    sumY1  +=       0.261 * cos(5.93685801944  + 1588.82907780029 *T)
    sumY1  +=       0.260 * cos(4.76140916156  + 11769.60987568290 *T)
    sumY1  +=       0.259 * cos(1.56016916326  + 5033.02191368570 *T)
    sumY1  +=       0.257 * cos(6.15643297218  + 205.94173095370 *T)
    sumY1  +=       0.253 * cos(0.11888703466  + 17101.45495439070 *T)
    sumY1  +=       0.253 * cos(1.95240011543  + 16703.28869733031 *T)
    sumY1  +=       0.253 * cos(5.79371922564  + 6948.07571260490 *T)
    sumY1  +=       0.278 * cos(4.67460813376  + 7322.34627826531 *T)
    sumY1  +=       0.255 * cos(2.05970747935  + 11780.73417599970 *T)
    sumY1  +=       0.282 * cos(0.09128513887  + 860.55374623631 *T)
    sumY1  +=       0.253 * cos(3.98317215792  + 16703.32320463470 *T)
    sumY1  +=       0.256 * cos(0.04143904110  + 13936.55068765050 *T)
    sumY1  +=       0.240 * cos(1.10034150476  + 2274.36076699330 *T)
    sumY1  +=       0.246 * cos(4.96426688535  + 3377.46160948750 *T)
    sumY1  +=       0.246 * cos(1.15950003329  + 3304.25087887911 *T)
    sumY1  +=       0.270 * cos(3.50813287021  + 1222.09238380491 *T)
    sumY1  +=       0.243 * cos(2.01100242811  + 16276.70776010650 *T)
    sumY1  +=       0.253 * cos(0.46888034698  + 14314.41193053330 *T)
    sumY1  +=       0.225 * cos(2.07472210939  + 10018.49066893110 *T)
    sumY1  +=       0.227 * cos(1.47944997014  + 5305.69487103730 *T)
    sumY1  +=       0.284 * cos(0.56682884608  + 10157.14617761831 *T)
    sumY1  +=       0.244 * cos(2.91803013526  + 2074.80113058990 *T)
    sumY1  +=       0.243 * cos(5.89092212082  + 30065.75565778170 *T)
    sumY1  +=       0.218 * cos(1.77863871509  + 3030.01544531491 *T)
    sumY1  +=       0.220 * cos(4.80019837521  + 16858.23871544969 *T)
    sumY1  +=       0.221 * cos(4.00597278506  + 4005.61228931331 *T)
    sumY1  +=       0.213 * cos(4.55100609184  + 533.45790092711 *T)
    sumY1  +=       0.224 * cos(2.50666934513  + 10025.67152623470 *T)
    sumY1  +=       0.216 * cos(4.14679941286  + 10037.33356876791 *T)
    sumY1  +=       0.213 * cos(3.70755893129  + 10448.67928845890 *T)
    sumY1  +=       0.219 * cos(5.94253384867  + 23.63462026429 *T)
    sumY1  +=       0.218 * cos(1.47351363921  + 5641.95442512570 *T)
    sumY1  +=       0.211 * cos(2.33151051616  + 6944.55259425590 *T)
    sumY1  +=       0.215 * cos(2.61877455491  + 3212.83740084591 *T)
    sumY1  +=       0.234 * cos(5.47804588842  + 3185.50315505190 *T)
    sumY1  +=       0.212 * cos(2.89638437903  + 3553.66770465429 *T)
    sumY1  +=       0.270 * cos(1.43547945741  + 16061.94070863329 *T)
    sumY1  +=       0.214 * cos(1.54073481456  + 1545.11016549530 *T)
    sumY1  +=       0.213 * cos(1.38873772382  + 17925.15451730390 *T)
    sumY1  +=       0.206 * cos(2.74922501212  + 15663.79170522509 *T)
    sumY1  +=       0.201 * cos(5.99561851209  + 1835.56843509031 *T)
    sumY1  +=       0.211 * cos(3.74100683574  + 3230.64992296391 *T)
    sumY1  +=       0.198 * cos(4.93753556250  + 6578.37589666450 *T)
    sumY1  +=       0.204 * cos(2.32132107344  + 12323.17927852529 *T)
    sumY1  +=       0.204 * cos(3.50487863763  + 3468.87508752070 *T)
    sumY1  +=       0.213 * cos(5.38886070050  + 7747.96414807311 *T)
    sumY1  +=       0.197 * cos(2.51048924008  + 3340.92355448610 *T)
    sumY1  +=       0.221 * cos(0.64591178951  + 3496.20933331471 *T)
    sumY1  +=       0.193 * cos(4.34388166294  + 3651.69704305170 *T)
    sumY1  +=       0.191 * cos(5.30120263654  + 14712.56093394150 *T)
    sumY1  +=       0.242 * cos(2.25789470004  + 6622.09480896950 *T)
    sumY1  +=       0.191 * cos(0.52002727082  + 2480.05868046350 *T)
    sumY1  +=       0.188 * cos(1.00360687694  + 103.33659170210 *T)
    sumY1  +=       0.190 * cos(5.85285761744  + 15508.85894075790 *T)
    sumY1  +=       0.182 * cos(3.99129735579  + 8543.21452351850 *T)
    sumY1  +=       0.233 * cos(1.89171200408  + 13916.26292712510 *T)
    sumY1  +=       0.200 * cos(2.50870443352  + 6677.58733552510 *T)
    sumY1  +=       0.249 * cos(3.21111672139  + 699.02733074409 *T)
    sumY1  +=       0.237 * cos(3.11974305123  + 7373.13863714289 *T)
    sumY1  +=       0.202 * cos(6.20171410320  + 12012.82611462390 *T)
    sumY1  +=       0.205 * cos(6.27868730147  + 2946.23035912410 *T)
    sumY1  +=       0.230 * cos(2.44828922930  + 1903.19299501769 *T)
    sumY1  +=       0.239 * cos(5.27789619503  + 6107.12387254830 *T)
    sumY1  +=       0.200 * cos(1.41969751355  + 10001.72577818411 *T)
    sumY1  +=       0.174 * cos(4.74303003189  + 6457.12387518120 *T)
    sumY1  +=       0.188 * cos(1.45966800471  + 2388.65020296570 *T)
    sumY1  +=       0.171 * cos(2.86664255530  + 6756.25026945040 *T)
    sumY1  +=       0.174 * cos(1.67176462211  + 6418.38474751031 *T)
    sumY1  +=       0.182 * cos(5.80990388322  + 6887.65421932030 *T)
    sumY1  +=       0.172 * cos(1.88607439867  + 19402.55313533309 *T)
    sumY1  +=       0.169 * cos(2.92046854649  + 522.33360061030 *T)
    sumY1  +=       0.176 * cos(2.54512773213  + 7896.20369020371 *T)
    sumY1  +=       0.169 * cos(5.91099958335  + 5618.56362234490 *T)
    sumY1  +=       0.168 * cos(6.18245369751  + 13575.99261972070 *T)
    sumY1  +=       0.184 * cos(2.36726292588  + 7218.27318297850 *T)
    sumY1  +=       0.166 * cos(2.25729247368  + 15651.03756779411 *T)
    sumY1  +=       0.212 * cos(3.23189207115  + 13363.67384535091 *T)
    sumY1  +=       0.169 * cos(5.08944200441  + 2655.81218995730 *T)
    sumY1  +=       0.194 * cos(3.66940007407  + 9491.90249752129 *T)
    sumY1  +=       0.168 * cos(6.08545323308  + 7314.25241019630 *T)
    sumY1  +=       0.177 * cos(6.17969940479  + 3341.34893028450 *T)
    sumY1  +=       0.160 * cos(4.59837771648  + 1903.68062998470 *T)
    sumY1  +=       0.159 * cos(1.06464153160  + 2008.80135664250 *T)
    sumY1  +=       0.177 * cos(3.64016669145  + 6686.99159519050 *T)
    sumY1  +=       0.210 * cos(4.66026492368  + 4032.52621044309 *T)
    sumY1  +=       0.192 * cos(1.70886670681  + 2751.79141717511 *T)
    sumY1  +=       0.153 * cos(0.18475722423  + 2803.56409712089 *T)
    sumY1  +=       0.153 * cos(2.83652315799  + 3334.17258030920 *T)
    sumY1  +=       0.153 * cos(3.28724376065  + 3347.53990805741 *T)
    sumY1  +=       0.152 * cos(4.87261249749  + 5486.53402569149 *T)
    sumY1  +=       0.154 * cos(0.41279284728  + 3499.79976196650 *T)
    sumY1  +=       0.151 * cos(0.41038595803  + 14054.36349054249 *T)
    sumY1  +=       0.169 * cos(0.15695053633  + 19645.76937427410 *T)
    sumY1  +=       0.197 * cos(2.91931470058  + 3340.78893388051 *T)
    sumY1  +=       0.195 * cos(5.14062276406  + 6546.40359084770 *T)
    sumY1  +=       0.185 * cos(4.38503970138  + 1331.81107005730 *T)
    sumY1  +=       0.148 * cos(0.31282714714  + 14712.07329897449 *T)
    sumY1  +=       0.196 * cos(1.90639910627  + 8982.56685182550 *T)
    sumY1  +=       0.158 * cos(3.03913180850  + 14557.14053450730 *T)
    sumY1  +=       0.171 * cos(2.92459964259  + 1063.14886602171 *T)
    sumY1  +=       0.140 * cos(0.03075322131  + 2786.79920637390 *T)
    sumY1  +=       0.149 * cos(3.48529084084  + 13518.11392371690 *T)
    sumY1  +=       0.150 * cos(5.75127697164  + 736.12033101530 *T)
    sumY1  +=       0.140 * cos(2.21026127637  + 4039.63975744390 *T)
    sumY1  +=       0.148 * cos(1.43179985105  + 3053.89388306271 *T)
    sumY1  +=       0.141 * cos(0.98834721175  + 7380.25218414370 *T)
    sumY1  +=       0.187 * cos(2.80618192029  + 2409.49315733150 *T)
    sumY1  +=       0.140 * cos(5.61689203595  + 6692.10533623291 *T)
    sumY1  +=       0.171 * cos(1.75603422806  + 14.47091148511 *T)
    sumY1  +=       0.135 * cos(3.59243401079  + 4076.73275771510 *T)
    sumY1  +=       0.135 * cos(2.53133290785  + 2604.97973065150 *T)
    sumY1  +=       0.134 * cos(3.70948493588  + 13554.14179039430 *T)
    sumY1  +=       0.133 * cos(0.65445531734  + 3191.29304704871 *T)
    sumY1  +=       0.177 * cos(0.26345356290  + 2277.46356067810 *T)
    sumY1  +=       0.177 * cos(0.10393431560  + 8959.17604904470 *T)
    sumY1  +=       0.135 * cos(0.48499376881  + 3347.48215621710 *T)
    sumY1  +=       0.178 * cos(3.44835944864  + 2284.50979737610 *T)
    sumY1  +=       0.175 * cos(3.29098160153  + 8966.22228574270 *T)
    sumY1  +=       0.131 * cos(5.54343293251  + 9375.07026426810 *T)
    sumY1  +=       0.160 * cos(6.02367905890  + 2494.28577446509 *T)
    sumY1  +=       0.129 * cos(0.75365086927  + 6680.99828956829 *T)
    sumY1  +=       0.152 * cos(6.26142645482  + 12722.79623796870 *T)
    sumY1  +=       0.138 * cos(6.18657392368  + 20047.44149603130 *T)
    sumY1  +=       0.171 * cos(5.54655180913  + 75.02541605080 *T)
    sumY1  +=       0.129 * cos(0.14024939993  + 4846.14405327630 *T)
    sumY1  +=       0.134 * cos(1.05473100111  + 4025.90029840930 *T)
    sumY1  +=       0.129 * cos(2.83674099689  + 11670.52785478030 *T)
    sumY1  +=       0.129 * cos(5.00605848212  + 6680.96378226390 *T)
    sumY1  +=       0.157 * cos(5.16739524910  + 9389.29735826970 *T)
    sumY1  +=       0.126 * cos(3.21328141790  + 6148.25458743950 *T)
    sumY1  +=       0.176 * cos(4.78085909788  + 21.01921297590 *T)
    sumY1  +=       0.169 * cos(3.25531409251  + 19004.40413192490 *T)
    sumY1  +=       0.155 * cos(4.42535125499  + 3929.92107119150 *T)
    sumY1  +=       0.150 * cos(2.94620445414  + 3077.77232081051 *T)
    sumY1  +=       0.133 * cos(5.83390369046  + 12840.11610618890 *T)
    sumY1  +=       0.137 * cos(4.24430027174  + 6674.42243418510 *T)
    sumY1  +=       0.129 * cos(4.58072373782  + 18606.25512851669 *T)
    sumY1  +=       0.146 * cos(5.60096860073  + 3284.05298201350 *T)
    sumY1  +=       0.146 * cos(0.52279831791  + 3397.65950635310 *T)
    sumY1  +=       0.130 * cos(3.84203213016  + 1505.04399160949 *T)
    sumY1  +=       0.133 * cos(4.69273543300  + 3627.81860530390 *T)
    sumY1  +=       0.125 * cos(3.57968087835  + 4672.42349675709 *T)
    sumY1  +=       0.124 * cos(6.11670347030  + 6594.16046634330 *T)
    sumY1  +=       0.135 * cos(1.74343590305  + 7902.82960223750 *T)
    sumY1  +=       0.147 * cos(3.59302955145  + 536.56069461190 *T)
    sumY1  +=       0.123 * cos(0.40369530083  + 1486.22393854870 *T)
    sumY1  +=       0.125 * cos(4.15712748419  + 15265.64270181689 *T)
    sumY1  +=       0.121 * cos(6.00453207609  + 4534.81561944089 *T)
    sumY1  +=       0.134 * cos(4.72288196608  + 2288.58786099491 *T)
    sumY1  +=       0.119 * cos(2.99117658049  + 74.53778108379 *T)
    sumY1  +=       0.146 * cos(2.76908101941  + 1479.11039154791 *T)
    sumY1  +=       0.117 * cos(2.06621235900  + 3344.36210888011 *T)
    sumY1  +=       0.117 * cos(0.91596190605  + 3337.35037948650 *T)
    sumY1  +=       0.147 * cos(1.07606700174  + 6510.79580026390 *T)
    sumY1  +=       0.118 * cos(4.23793013469  + 3531.81687532789 *T)
    sumY1  +=       0.160 * cos(1.90920939107  + 3451.06256540270 *T)
    sumY1  +=       0.117 * cos(5.16837558174  + 3337.31587218211 *T)
    sumY1  +=       0.117 * cos(4.09698399049  + 3344.39661618450 *T)
    sumY1  +=       0.118 * cos(4.20346162229  + 9887.01601754750 *T)
    sumY1  +=       0.119 * cos(1.28976261169  + 15907.00794416610 *T)
    sumY1  +=       0.154 * cos(0.46709104782  + 3134.15550688670 *T)
    sumY1  +=       0.118 * cos(1.33279043667  + 3363.62474079271 *T)
    sumY1  +=       0.118 * cos(4.79097648197  + 3318.08774757391 *T)
    sumY1  +=       0.117 * cos(1.06179529690  + 3184.45552368090 *T)
    sumY1  +=       0.117 * cos(5.06197162174  + 3497.25696468570 *T)
    sumY1  +=       0.115 * cos(1.85848730524  + 6725.18758318811 *T)
    sumY1  +=       0.116 * cos(2.88106561931  + 5243.80542171749 *T)
    sumY1  +=       0.124 * cos(3.17992351939  + 2631.86644190671 *T)
    sumY1  +=       0.135 * cos(3.17784971271  + 3603.94016755610 *T)
    sumY1  +=       0.112 * cos(6.10617816315  + 9065.79194161230 *T)
    sumY1  +=       0.152 * cos(1.84450072046  + 1.22413855171 *T)
    sumY1  +=       0.127 * cos(1.15557216374  + 7366.02509014209 *T)
    sumY1  +=       0.119 * cos(5.17512579994  + 16511.85768487090 *T)
    sumY1  +=       0.113 * cos(2.74157549210  + 13745.59005650590 *T)
    sumY1  +=       0.155 * cos(0.79710150138  + 6155.30082413750 *T)
    sumY1  +=       0.115 * cos(3.22959858043  + 6997.86054053970 *T)
    sumY1  +=       0.114 * cos(6.28123647971  + 632.53992182970 *T)
    sumY1  +=       0.111 * cos(6.21354218410  + 87.55202202331 *T)
    sumY1  +=       0.114 * cos(1.52685816324  + 3987.86707749811 *T)
    sumY1  +=       0.125 * cos(6.02617311099  + 5821.15874213030 *T)
    sumY1  +=       0.138 * cos(4.58751267867  + 2233.71743849851 *T)
    sumY1  +=       0.149 * cos(3.08301584779  + 7373.62627210991 *T)
    sumY1  +=       0.131 * cos(5.25025552048  + 6901.88131332190 *T)
    sumY1  +=       0.107 * cos(2.71978809026  + 2973.63191522091 *T)
    sumY1  +=       0.107 * cos(4.99897363633  + 4672.91113172411 *T)
    sumY1  +=       0.145 * cos(2.57143715237  + 2060.57403658830 *T)
    sumY1  +=       0.111 * cos(2.15551948866  + 6357.96325422570 *T)
    sumY1  +=       0.112 * cos(2.39993760456  + 2575.06317973690 *T)
    sumY1  +=       0.139 * cos(5.65971955250  + 3547.55698147990 *T)
    sumY1  +=       0.122 * cos(2.84793235795  + 3346.71344648291 *T)
    sumY1  +=       0.122 * cos(3.27583456069  + 3334.99904188371 *T)
    sumY1  +=       0.114 * cos(3.01986257881  + 931.60690433531 *T)
    sumY1  +=       0.112 * cos(3.01795120085  + 7079.13003932430 *T)
    sumY1  +=       0.108 * cos(3.30807311845  + 8584.41784841729 *T)
    sumY1  +=       0.141 * cos(2.88317170767  + 2907.14450630650 *T)
    sumY1  +=       0.139 * cos(4.45925006470  + 5518.99396647530 *T)
    sumY1  +=       0.115 * cos(6.13734279118  + 4399.75053940550 *T)
    sumY1  +=       0.103 * cos(5.27914306931  + 2487.65986243130 *T)
    sumY1  +=       0.140 * cos(4.52778173929  + 6670.83200553330 *T)
    sumY1  +=       0.105 * cos(5.17193563022  + 2957.47207699310 *T)
    sumY1  +=       0.102 * cos(3.47030687176  + 316.14805217309 *T)
    sumY1  +=       0.100 * cos(1.01169943823  + 17918.04097030310 *T)
    sumY1  +=       0.120 * cos(4.71438356567  + 3311.36442587990 *T)
    sumY1  +=       0.120 * cos(1.40938335297  + 3370.34806248671 *T)
    sumY1  +=       0.118 * cos(0.83092751777  + 6606.68707231580 *T)
    sumY1  +=       0.098 * cos(2.88130937597  + 17468.61138046190 *T)
    sumY1  +=       0.097 * cos(3.61583991386  + 8859.60639317511 *T)
    sumY1  +=       0.107 * cos(1.96982954301  + 2391.19300024650 *T)
    sumY1  +=       0.132 * cos(4.17534294218  + 3050.37076471370 *T)
    sumY1  +=       0.106 * cos(4.20312239248  + 2641.58509595570 *T)
    sumY1  +=       0.097 * cos(3.72308479740  + 22345.01655862470 *T)
    sumY1  +=       0.101 * cos(2.19730971635  + 10125.17387180150 *T)
    sumY1  +=       0.104 * cos(0.35235356880  + 4297.14540015391 *T)
    sumY1  +=       0.096 * cos(5.64087470447  + 2679.62331740231 *T)
    sumY1  +=       0.111 * cos(6.00152200950  + 10042.43641698170 *T)
    sumY1  +=       0.110 * cos(6.25839552650  + 2913.77041834030 *T)
    sumY1  +=       0.096 * cos(3.26945343269  + 10006.82862639790 *T)
    sumY1  +=       0.112 * cos(2.22451688549  + 11251.04321090511 *T)
    sumY1  +=       0.096 * cos(1.66024188153  + 59.13004443009 *T)
    sumY1  +=       0.108 * cos(6.12490206072  + 3735.48212924251 *T)
    sumY1  +=       0.092 * cos(3.39937164286  + 3708.08057314570 *T)
    sumY1  +=       0.097 * cos(2.13065182863  + 2922.35192137631 *T)
    sumY1  +=       0.127 * cos(3.31547807544  + 4272.21933103511 *T)
    sumY1  +=       0.117 * cos(3.74695427979  + 12168.24651405810 *T)
    sumY1  +=       0.116 * cos(6.20968192934  + 10213.52936369450 *T)
    sumY1  +=       0.115 * cos(1.94977986471  + 9815.89554914570 *T)
    sumY1  +=       0.093 * cos(2.14817772588  + 2772.03437015590 *T)
    sumY1  +=       0.104 * cos(3.71898659352  + 6033.97020260130 *T)
    sumY1  +=       0.091 * cos(3.61425883493  + 169.33636564950 *T)
    sumY1  +=       0.090 * cos(4.44081930332  + 2541.03511901791 *T)
    sumY1  +=       0.100 * cos(0.09021007523  + 26887.60986721410 *T)
    sumY1  +=       0.096 * cos(1.76469942098  + 640.63378989870 *T)
    sumY1  +=       0.090 * cos(1.28021880619  + 846.32665223470 *T)
    sumY1  +=       0.089 * cos(3.39079978165  + 2817.79119112250 *T)
    sumY1  +=       0.115 * cos(3.89643144493  + 1648.20293971390 *T)
    sumY1  +=       0.096 * cos(2.94289920307  + 4049.84604645990 *T)
    sumY1  +=       0.113 * cos(4.93121954959  + 4922.81559246630 *T)
    sumY1  +=       0.087 * cos(1.02670558277  + 1118.51197461930 *T)
    sumY1  +=       0.094 * cos(3.25084340693  + 1272.92484311070 *T)
    sumY1  +=       0.110 * cos(2.48041052275  + 21795.45790909830 *T)
    sumY1  +=       0.103 * cos(6.11803973071  + 12302.82395912650 *T)
    sumY1  +=       0.088 * cos(2.15369137259  + 17654.53672226610 *T)
    sumY1  +=       0.119 * cos(1.94804023984  + 3631.34172365290 *T)
    sumY1  +=       0.094 * cos(0.15915150031  + 10706.63751684189 *T)
    sumY1  +=       0.090 * cos(0.48988280834  + 19800.70213874130 *T)
    sumY1  +=       0.117 * cos(4.65239707285  + 1905.70858242391 *T)
    sumY1  +=       0.098 * cos(0.70463876405  + 6768.77687542291 *T)
    sumY1  +=       0.084 * cos(5.11902672644  + 3152.93109642071 *T)
    sumY1  +=       0.084 * cos(4.14632885534  + 3528.78139194590 *T)
    sumY1  +=       0.088 * cos(0.91279871901  + 36.84918278771 *T)
    sumY1  +=       0.112 * cos(3.74585806690  + 11925.03027511709 *T)
    sumY1  +=       0.093 * cos(1.26569825727  + 2538.49232173710 *T)
    sumY1  +=       0.091 * cos(3.30441819566  + 10714.23869880971 *T)
    sumY1  +=       0.094 * cos(3.26149490187  + 6048.19729660290 *T)
    sumY1  +=       0.110 * cos(4.99537778792  + 21946.86755521649 *T)
    sumY1  +=       0.085 * cos(6.12081404839  + 16490.00685554450 *T)
    sumY1  +=       0.097 * cos(1.60911861040  + 2170.78035780770 *T)
    sumY1  +=       0.094 * cos(4.47849317886  + 12943.20888040750 *T)
    sumY1  +=       0.088 * cos(1.95089690983  + 4040.12739241091 *T)
    sumY1  +=       0.107 * cos(0.22183845911  + 6286.84278582391 *T)
    sumY1  +=       0.082 * cos(0.12897329018  + 21548.71855180830 *T)
    sumY1  +=       0.081 * cos(5.61725901284  + 3355.08333818490 *T)
    sumY1  +=       0.081 * cos(0.50650790580  + 3326.62915018171 *T)
    sumY1  +=       0.085 * cos(3.72310585517  + 4106.64930862970 *T)
    sumY1  +=       0.082 * cos(6.02187911000  + 18208.10612510850 *T)
    sumY1  +=       0.101 * cos(4.84911356755  + 5628.71265272770 *T)
    sumY1  +=       0.079 * cos(1.50942161554  + 4778.03185838191 *T)
    sumY1  +=       0.083 * cos(3.99309879852  + 3759.36056699030 *T)
    sumY1  +=       0.081 * cos(0.53432576001  + 1442.01739127670 *T)
    sumY1  +=       0.079 * cos(3.42549882475  + 3302.72320854550 *T)
    sumY1  +=       0.079 * cos(2.69826809389  + 3378.98927982111 *T)
    sumY1  +=       0.109 * cos(1.10805610503  + 765.54924696289 *T)
    sumY1  +=       0.079 * cos(1.18088037992  + 9947.29949901560 *T)
    sumY1  +=       0.087 * cos(5.45293469310  + 3181.91272640011 *T)
    sumY1  +=       0.092 * cos(3.38758997686  + 684.80023674250 *T)
    sumY1  +=       0.078 * cos(2.11522669224  + 4025.41266344230 *T)
    sumY1  +=       0.106 * cos(5.10676390423  + 7214.68275432671 *T)
    sumY1  +=       0.079 * cos(1.67803485609  + 5562.71287878030 *T)
    sumY1  +=       0.094 * cos(5.43559031617  + 1655.31648671470 *T)
    sumY1  +=       0.077 * cos(2.80318148641  + 42942.74815663290 *T)
    sumY1  +=       0.096 * cos(1.53353605164  + 4447.99504986810 *T)
    sumY1  +=       0.079 * cos(5.29022717745  + 1169.83206889211 *T)
    sumY1  +=       0.077 * cos(0.48002800064  + 4002.08917096430 *T)
    sumY1  +=       0.075 * cos(3.02737043920  + 4988.81536641370 *T)
    sumY1  +=       0.103 * cos(3.31511491795  + 6279.79654912590 *T)
    sumY1  +=       0.075 * cos(1.36914522483  + 6710.89817941910 *T)
    sumY1  +=       0.100 * cos(1.23379947915  + 6654.06711478630 *T)
    sumY1  +=       0.075 * cos(5.26781491970  + 1106.89498820129 *T)
    sumY1  +=       0.085 * cos(3.07884071954  + 3031.57792152750 *T)
    sumY1  +=       0.085 * cos(3.04492619910  + 3650.13456683910 *T)
    sumY1  +=       0.102 * cos(5.58829580223  + 220.16882495529 *T)
    sumY1  +=       0.077 * cos(3.97228899508  + 3909.67811821070 *T)
    sumY1  +=       0.080 * cos(3.93017018663  + 42153.72518556549 *T)
    sumY1  +=       0.090 * cos(3.63465564186  + 2277.95119564511 *T)
    sumY1  +=       0.074 * cos(3.28312495896  + 10184.54773371510 *T)
    sumY1  +=       0.102 * cos(3.28709663996  + 3282.97754817950 *T)
    sumY1  +=       0.102 * cos(2.83667027868  + 3398.73494018711 *T)
    sumY1  +=       0.077 * cos(5.68389652778  + 20199.27146681370 *T)
    sumY1  +=       0.073 * cos(3.78223099359  + 13369.80707128350 *T)
    sumY1  +=       0.097 * cos(0.21157168319  + 6784.56144510170 *T)
    sumY1  +=       0.092 * cos(0.92188445820  + 3492.75352526850 *T)
    sumY1  +=       0.076 * cos(1.15525142920  + 19404.02109136830 *T)
    sumY1  +=       0.073 * cos(0.71834577426  + 6365.07680122650 *T)
    sumY1  +=       0.080 * cos(2.74867343749  + 8756.51361895650 *T)
    sumY1  +=       0.077 * cos(0.86279166830  + 6549.92670919670 *T)
    sumY1  +=       0.072 * cos(3.17331282546  + 2067.68758358909 *T)
    sumY1  +=       0.073 * cos(3.24032193523  + 20809.22380716170 *T)
    sumY1  +=       0.073 * cos(3.75683220932  + 8955.09798542589 *T)
    sumY1  +=       0.088 * cos(4.48290442721  + 11883.82695021830 *T)
    sumY1  +=       0.070 * cos(2.11654601508  + 6254.38284504009 *T)
    sumY1  +=       0.093 * cos(1.79949400982  + 10050.53028505070 *T)
    sumY1  +=       0.074 * cos(0.39743697458  + 12725.20961729149 *T)
    sumY1  +=       0.069 * cos(4.46900086396  + 1280.03839011150 *T)
    sumY1  +=       0.090 * cos(1.32575986288  + 10713.75106384269 *T)
    sumY1  +=       0.082 * cos(0.69210206943  + 5401.18646328809 *T)
    sumY1  +=       0.072 * cos(1.34421235248  + 18451.32236404950 *T)
    sumY1  +=       0.077 * cos(5.18506287875  + 5466.73365156250 *T)
    sumY1  +=       0.079 * cos(0.03494157481  + 3351.90194444720 *T)
    sumY1  +=       0.079 * cos(6.08882534383  + 3329.81054391940 *T)
    sumY1  +=       0.077 * cos(3.36821374117  + 15636.81047379250 *T)
    sumY1  +=       0.070 * cos(4.18852998473  + 2011.34415392330 *T)
    sumY1  +=       0.068 * cos(2.20697305416  + 9381.20349020070 *T)
    sumY1  +=       0.074 * cos(2.64935507870  + 25136.07033579810 *T)
    sumY1  +=       0.068 * cos(1.80816186710  + 6489.53276980449 *T)
    sumY1  +=       0.078 * cos(3.45047388274  + 3583.09721319030 *T)
    sumY1  +=       0.069 * cos(3.47802563975  + 11354.13598512370 *T)
    sumY1  +=       0.070 * cos(5.41206396259  + 25287.47998191629 *T)
    sumY1  +=       0.070 * cos(5.52776045508  + 35707.46626542390 *T)
    sumY1  +=       0.088 * cos(1.52868546432  + 3068.18328683170 *T)
    sumY1  +=       0.088 * cos(4.59502570300  + 3613.52920153491 *T)
    sumY1  +=       0.065 * cos(1.82205105922  + 15121.34660269950 *T)
    sumY1  +=       0.077 * cos(2.56929587888  + 14127.99895376209 *T)
    sumY1  +=       0.069 * cos(0.02357643792  + 43739.04616344930 *T)
    sumY1  +=       0.066 * cos(5.74550598907  + 5195.48854981791 *T)
    sumY1  +=       0.066 * cos(5.74550598907  + 5195.00091485090 *T)
    sumY1  +=       0.065 * cos(5.47583793528  + 6708.87022697990 *T)
    sumY1  +=       0.071 * cos(0.41816629111  + 10176.94655174729 *T)
    sumY1  +=       0.090 * cos(3.58467854106  + 6740.84253279671 *T)
    sumY1  +=       0.065 * cos(4.81907662858  + 4140.67736934870 *T)
    sumY1  +=       0.064 * cos(2.78231915414  + 9801.66845514410 *T)
    sumY1  +=       0.085 * cos(1.40895818486  + 43340.89716004110 *T)
    sumY1  +=       0.087 * cos(3.48209962852  + 9859.61446145070 *T)
    sumY1  +=       0.074 * cos(5.03444697014  + 10596.42589591770 *T)
    sumY1  +=       0.065 * cos(5.65188017283  + 6681.04859478950 *T)
    sumY1  +=       0.071 * cos(4.85107944455  + 4143.22016662950 *T)
    sumY1  +=       0.066 * cos(0.32887989990  + 3267.25528826810 *T)
    sumY1  +=       0.066 * cos(5.79488701874  + 3414.45720009850 *T)
    sumY1  +=       0.063 * cos(2.79450851807  + 6467.68194047809 *T)
    sumY1  +=       0.064 * cos(0.70758466822  + 3487.66793070691 *T)
    sumY1  +=       0.064 * cos(5.41618225042  + 3194.04455765970 *T)
    sumY1  +=       0.085 * cos(6.17825225561  + 10720.86461084350 *T)
    sumY1  +=       0.062 * cos(3.01771087976  + 35.78404919390 *T)
    sumY1  +=       0.062 * cos(6.15930353335  + 36.27168416091 *T)
    sumY1  +=       0.064 * cos(2.07206554388  + 11773.13299403189 *T)
    sumY1  +=       0.064 * cos(4.96743808710  + 41747.81351144129 *T)
    sumY1  +=       0.068 * cos(3.28488221728  + 12809.12412144031 *T)
    sumY1  +=       0.062 * cos(1.67633734169  + 18461.95902939930 *T)
    sumY1  +=       0.068 * cos(0.42906740811  + 6614.98126196870 *T)
    sumY1  +=       0.079 * cos(2.06833189434  + 11236.81611690350 *T)
    sumY1  +=       0.066 * cos(4.71978579468  + 9998.20265983511 *T)
    sumY1  +=       0.064 * cos(5.14611248392  + 11526.88127170889 *T)
    sumY1  +=       0.070 * cos(0.34840349768  + 16858.65904011390 *T)
    sumY1  +=       0.065 * cos(3.82789901811  + 20597.00014555770 *T)
    sumY1  +=       0.064 * cos(4.43540244948  + 40796.58274015770 *T)
    sumY1  +=       0.079 * cos(4.81144487223  + 10730.58326489249 *T)
    sumY1  +=       0.062 * cos(0.61927789332  + 20040.39525933330 *T)
    sumY1  +=       0.066 * cos(6.18584312904  + 8428.99744898310 *T)
    sumY1  +=       0.079 * cos(5.86016787523  + 40398.43373674950 *T)
    sumY1  +=       0.059 * cos(0.23825712673  + 44959.30385850670 *T)
    sumY1  +=       0.063 * cos(0.00069466290  + 20743.18952590990 *T)
    sumY1  +=       0.060 * cos(4.02422394326  + 10016.55817327550 *T)
    sumY1  +=       0.059 * cos(3.94228594452  + 13745.10242153889 *T)
    sumY1  +=       0.067 * cos(3.09157172324  + 9329.92349635610 *T)
    sumY1  +=       0.078 * cos(1.63178356452  + 24606.37937070350 *T)
    sumY1  +=       0.080 * cos(3.21443220409  + 5202.60209681870 *T)
    sumY1  +=       0.057 * cos(4.86026303770  + 12928.98178640591 *T)
    sumY1  +=       0.080 * cos(1.00062919899  + 27417.30083230870 *T)
    sumY1  +=       0.077 * cos(4.25649384915  + 4845.65641830930 *T)
    sumY1  +=       0.062 * cos(2.85482467160  + 12987.86801335249 *T)
    sumY1  +=       0.057 * cos(4.88274674366  + 5820.67110716330 *T)
    sumY1  +=       0.057 * cos(0.00003539968  + 25841.04938475869 *T)
    sumY1  +=       0.059 * cos(0.42155176888  + 5945.59215735130 *T)
    sumY1  +=       0.061 * cos(0.64683968145  + 3264.59017290771 *T)
    sumY1  +=       0.058 * cos(4.86677635959  + 15423.51137835450 *T)
    sumY1  +=       0.074 * cos(3.19195801116  + 9623.44445920770 *T)
    sumY1  +=       0.062 * cos(4.72968550328  + 3324.02404313351 *T)
    sumY1  +=       0.062 * cos(1.39408141536  + 3357.68844523310 *T)
    sumY1  +=       0.060 * cos(4.79015337969  + 9225.29545579949 *T)
    sumY1  +=       0.061 * cos(5.58622273047  + 5223.93773728571 *T)
    sumY1  +=       0.056 * cos(4.51159263204  + 3702.97738091410 *T)
    sumY1  +=       0.056 * cos(1.61217428660  + 2978.73510745250 *T)
    sumY1  +=       0.055 * cos(3.98689536822  + 12715.68269096791 *T)
    sumY1  +=       0.058 * cos(1.49398235066  + 14584.05445563709 *T)
    sumY1  +=       0.075 * cos(1.46856818276  + 4776.00390594270 *T)
    sumY1  +=       0.054 * cos(4.63507585798  + 45885.21157992450 *T)
    sumY1  +=       0.055 * cos(4.91506650009  + 2126.05391455991 *T)
    sumY1  +=       0.054 * cos(0.92181027242  + 9381.69617630189 *T)
    sumY1  +=       0.054 * cos(2.71877294650  + 5607.37201172530 *T)
    sumY1  +=       0.053 * cos(3.09693738935  + 6789.13219482170 *T)
    sumY1  +=       0.058 * cos(5.89934208420  + 3268.80051082850 *T)
    sumY1  +=       0.058 * cos(0.22442483444  + 3412.91197753810 *T)
    sumY1  +=       0.053 * cos(0.16426574878  + 12509.00950756369 *T)
    sumY1  +=       0.053 * cos(0.16426574878  + 12509.49714253070 *T)
    sumY1  +=       0.073 * cos(5.56901413325  + 14867.49369840869 *T)
    sumY1  +=       0.053 * cos(2.19783708455  + 14481.44931638550 *T)
    sumY1  +=       0.054 * cos(1.94626267929  + 24491.18197509989 *T)
    sumY1  +=       0.052 * cos(3.15010220845  + 6813.01063256950 *T)
    sumY1  +=       0.054 * cos(4.66674950044  + 8186.26884500910 *T)
    sumY1  +=       0.054 * cos(1.57540808090  + 21150.56954840009 *T)
    sumY1  +=       0.054 * cos(1.30773052880  + 8226.82265386190 *T)
    sumY1  +=       0.054 * cos(5.80728044581  + 18849.47136745770 *T)
    sumY1  +=       0.068 * cos(5.38162441083  + 20213.49856081530 *T)
    sumY1  +=       0.062 * cos(5.91673914049  + 9175.99826283170 *T)
    sumY1  +=       0.068 * cos(5.53849729729  + 2391.68063521350 *T)
    sumY1  +=       0.055 * cos(5.68275033208  + 23539.46356884929 *T)
    sumY1  +=       0.072 * cos(5.07506580904  + 3342.34071689161 *T)
    sumY1  +=       0.072 * cos(1.04870110960  + 3339.37177147500 *T)
    sumY1  +=       0.055 * cos(3.59989833025  + 6843.44767204830 *T)
    sumY1  +=       0.060 * cos(2.16467756824  + 4885.72259219509 *T)
    sumY1  +=       0.052 * cos(2.30863563499  + 7799.73682801889 *T)
    sumY1  +=       0.063 * cos(1.59644930072  + 21265.76694400370 *T)
    sumY1  +=       0.069 * cos(0.54846716464  + 24889.33097850809 *T)
    sumY1  +=       0.051 * cos(2.96626339056  + 37895.67010785090 *T)
    sumY1  +=       0.052 * cos(5.38935505875  + 45088.42593814109 *T)
    sumY1  +=       0.051 * cos(4.59657000491  + 12410.97511803210 *T)
    sumY1  +=       0.068 * cos(0.08247779224  + 1854.38848815109 *T)
    sumY1  +=       0.054 * cos(2.06267581529  + 2736.00650347850 *T)
    sumY1  +=       0.052 * cos(2.02950191708  + 16328.48044005229 *T)
    sumY1  +=       0.051 * cos(1.14816705620  + 16063.40866466850 *T)
    sumY1  +=       0.061 * cos(2.80673408688  + 2655.74487965451 *T)
    sumY1  +=       0.056 * cos(1.41031366215  + 8439.63411433290 *T)
    sumY1  +=       0.050 * cos(0.09559043051  + 1727.90491115171 *T)
    sumY1  +=       0.054 * cos(2.84157218677  + 5408.78764525591 *T)
    sumY1  +=       0.059 * cos(4.25728059212  + 12516.61068953150 *T)
    sumY1  +=       0.052 * cos(4.01903754148  + 4356.03162710049 *T)
    sumY1  +=       0.050 * cos(4.41829954403  + 3116.51144848140 *T)
    sumY1  +=       0.050 * cos(1.70549407364  + 3565.20103988520 *T)
    sumY1  +=       0.049 * cos(1.47402562828  + 55516.66252733170 *T)
    sumY1  +=       0.050 * cos(2.19394619075  + 3742.52836594050 *T)
    sumY1  +=       0.048 * cos(3.12444029552  + 1882.83767561889 *T)
    sumY1  +=       0.054 * cos(6.04907092105  + 20426.81490990550 *T)
    sumY1  +=       0.048 * cos(6.10216369462  + 16865.28495214769 *T)
    sumY1  +=       0.061 * cos(2.37337374250  + 19617.32018680630 *T)
    sumY1  +=       0.048 * cos(6.23602609768  + 19190.72199593030 *T)
    sumY1  +=       0.048 * cos(6.14914151023  + 24341.28449997309 *T)
    sumY1  +=       0.056 * cos(2.19344974157  + 10787.38652706230 *T)
    sumY1  +=       0.048 * cos(6.18351837234  + 44768.34322736210 *T)
    sumY1  +=       0.047 * cos(6.09189138220  + 13517.55897844709 *T)
    sumY1  +=       0.051 * cos(3.13613774695  + 5750.10558403130 *T)
    sumY1  +=       0.053 * cos(1.88416936813  + 1971.70835637129 *T)
    sumY1  +=       0.047 * cos(5.48212564317  + 11876.71340321750 *T)
    sumY1  +=       0.047 * cos(4.23460507904  + 2257.59587624631 *T)
    sumY1  +=       0.053 * cos(2.37415739297  + 20043.43074271529 *T)
    sumY1  +=       0.050 * cos(3.70055414539  + 32765.00284213229 *T)
    sumY1  +=       0.047 * cos(4.84924552555  + 39048.56632709070 *T)
    sumY1  +=       0.048 * cos(1.43634945815  + 135.30889751891 *T)
    sumY1  +=       0.064 * cos(1.47620421839  + 1329.75590774351 *T)
    sumY1  +=       0.051 * cos(3.87515145598  + 2693.84541086851 *T)
    sumY1  +=       0.055 * cos(4.42106984861  + 5834.89820116490 *T)
    sumY1  +=       0.046 * cos(4.98957166191  + 1795.50226120450 *T)
    sumY1  +=       0.055 * cos(2.47584695868  + 4403.76129272151 *T)
    sumY1  +=       0.048 * cos(5.05121150955  + 14955.28953791550 *T)
    sumY1  +=       0.059 * cos(0.56447699816  + 4290.03185315310 *T)
    sumY1  +=       0.047 * cos(3.19086921862  + 35464.73766144989 *T)
    sumY1  +=       0.047 * cos(3.19086921862  + 35465.22529641690 *T)
    sumY1  +=       0.045 * cos(4.94702991651  + 5835.38583613191 *T)
    sumY1  +=       0.052 * cos(5.41422776986  + 17660.05964657349 *T)
    sumY1  +=       0.048 * cos(5.64285531978  + 1859.44687283811 *T)
    sumY1  +=       0.056 * cos(5.43434992198  + 532.97026596010 *T)
    sumY1  +=       0.052 * cos(5.51821932984  + 3718.22985209910 *T)
    sumY1  +=       0.052 * cos(0.60554758880  + 2963.48263626750 *T)
    sumY1  +=       0.045 * cos(1.65750320067  + 22487.12787535809 *T)
    sumY1  +=       0.061 * cos(3.03243825296  + 7380.73981911070 *T)
    sumY1  +=       0.045 * cos(1.23507902380  + 38252.26832027430 *T)
    sumY1  +=       0.045 * cos(4.90245609377  + 8741.79888998789 *T)
    sumY1  +=       0.054 * cos(5.75594735581  + 1056.10262932370 *T)
    sumY1  +=       0.045 * cos(0.78221109473  + 5511.88041947450 *T)
    sumY1  +=       0.045 * cos(3.65721451906  + 664.99986261351 *T)
    sumY1  +=       0.045 * cos(3.65721451906  + 664.51222764649 *T)
    sumY1  +=       0.045 * cos(1.58220885506  + 3973.63998349650 *T)
    sumY1  +=       0.044 * cos(1.64971708603  + 5401.67409825510 *T)
    sumY1  +=       0.054 * cos(3.22681077430  + 931.11926936830 *T)
    sumY1  +=       0.046 * cos(4.82251193103  + 5881.64754571770 *T)
    sumY1  +=       0.050 * cos(0.95411798734  + 1744.66980189871 *T)
    sumY1  +=       0.044 * cos(1.88527980304  + 3244.87701696550 *T)
    sumY1  +=       0.044 * cos(4.23848711560  + 3436.83547140110 *T)
    sumY1  +=       0.045 * cos(3.09915237980  + 1265.81129610991 *T)
    sumY1  +=       0.044 * cos(4.49759555733  + 17499.60395779890 *T)
    sumY1  +=       0.043 * cos(1.18409883166  + 97670.14389541368 *T)
    sumY1  +=       0.057 * cos(3.23899428569  + 46283.36058333270 *T)
    sumY1  +=       0.044 * cos(0.27827134364  + 10191.17364574889 *T)
    sumY1  +=       0.043 * cos(1.30628712649  + 33163.63948050750 *T)
    sumY1  +=       0.047 * cos(2.27882129437  + 17256.87535382490 *T)
    sumY1  +=       0.044 * cos(4.60162306872  + 2708.07250487011 *T)
    sumY1  +=       0.059 * cos(3.55006619333  + 2524.20291796811 *T)
    sumY1  +=       0.053 * cos(5.27405250378  + 8337.02897508130 *T)
    sumY1  +=       0.053 * cos(2.75147057919  + 14054.85112550950 *T)
    sumY1  +=       0.043 * cos(3.74257202617  + 12200.21881987490 *T)
    sumY1  +=       0.046 * cos(4.35317089818  + 1648.69057468091 *T)
    sumY1  +=       0.042 * cos(4.31895676455  + 845.83901726770 *T)
    sumY1  +=       0.043 * cos(3.06719963334  + 41194.73174356590 *T)
    sumY1  +=       0.047 * cos(5.75600675804  + 13531.78607244869 *T)
    sumY1  +=       0.047 * cos(5.82756499631  + 15850.10956923050 *T)
    sumY1  +=       0.042 * cos(4.34141462634  + 12729.90978496950 *T)
    sumY1  +=       0.043 * cos(1.95297083066  + 15107.11950869790 *T)
    sumY1  +=       0.043 * cos(0.33759865283  + 8123.72987964330 *T)
    sumY1  +=       0.046 * cos(0.42686371231  + 2325.19322629910 *T)
    sumY1  +=       0.043 * cos(0.62743288585  + 14469.34469500049 *T)
    sumY1  +=       0.043 * cos(1.18510374809  + 2854.88419139370 *T)
    sumY1  +=       0.043 * cos(4.93866317055  + 3826.82829697291 *T)
    sumY1  +=       0.041 * cos(2.91455733252  + 42544.59915322470 *T)
    sumY1  +=       0.042 * cos(5.52285099034  + 2652.22176130550 *T)
    sumY1  +=       0.042 * cos(0.04613741618  + 38086.63073899550 *T)
    sumY1  +=       0.045 * cos(0.35782418393  + 3361.14400470870 *T)
    sumY1  +=       0.043 * cos(0.97696624636  + 28229.94340520789 *T)
    sumY1  +=       0.048 * cos(0.38343131297  + 12825.88901218730 *T)
    sumY1  +=       0.052 * cos(4.39773367298  + 6667.24157688150 *T)
    sumY1  +=       0.041 * cos(0.69151426786  + 3296.20241094311 *T)
    sumY1  +=       0.041 * cos(5.43225265078  + 3385.51007742350 *T)
    sumY1  +=       0.043 * cos(1.84753611816  + 46681.50958674090 *T)
    sumY1  +=       0.040 * cos(3.38564855874  + 6851.04885401611 *T)
    sumY1  +=       0.040 * cos(0.97077467865  + 8983.05448679250 *T)
    sumY1  +=       0.043 * cos(0.96789677579  + 20205.89737884749 *T)
    sumY1  +=       0.040 * cos(0.36743358073  + 20728.96243190830 *T)
    sumY1  +=       0.044 * cos(5.64770312853  + 29698.53132881930 *T)
    sumY1  +=       0.041 * cos(1.94195643070  + 4670.36833444331 *T)
    sumY1  +=       0.052 * cos(2.83255526859  + 3299.30545319850 *T)
    sumY1  +=       0.052 * cos(3.29121165005  + 3382.40703516810 *T)
    sumY1  +=       0.041 * cos(0.22123606904  + 48827.67500321610 *T)
    sumY1  +=       0.046 * cos(2.01006330908  + 19247.62037086590 *T)
    sumY1  +=       0.054 * cos(4.52613106738  + 3319.59321372389 *T)
    sumY1  +=       0.039 * cos(2.09902832820  + 743.23387801611 *T)
    sumY1  +=       0.039 * cos(5.30952786068  + 26357.43126715249 *T)
    sumY1  +=       0.039 * cos(5.30952786068  + 26357.91890211950 *T)
    sumY1  +=       0.049 * cos(3.31851905687  + 4025.96760871210 *T)
    sumY1  +=       0.054 * cos(0.83545544152  + 574.58861581830 *T)
    sumY1  +=       0.052 * cos(0.78436364849  + 1461.29786942991 *T)
    sumY1  +=       0.041 * cos(6.03198869566  + 45487.06257651630 *T)
    sumY1  +=       0.051 * cos(0.35100574128  + 102.84895673509 *T)
    sumY1  +=       0.045 * cos(0.25913855607  + 2715.18605187090 *T)
    sumY1  +=       0.045 * cos(5.86462836257  + 3966.52643649571 *T)

    # MARS Y Coord
    sumY2  = 0
    sumY2  +=   68999.338 * cos(3.98646823714  + 0.24381748350 *T)
    sumY2  +=   82190.105 * cos(6.20347735626  + 3340.85624418330 *T)
    sumY2  +=   21826.465 * cos(1.79418551007  + 6681.46867088311 *T)
    sumY2  +=    6036.287 * cos(2.31685926546  + 10022.08109758290 *T)
    sumY2  +=    1291.999 * cos(5.44987502837  + 3340.36860921629 *T)
    sumY2  +=    1125.170 * cos(2.69734348224  + 13362.69352428270 *T)
    sumY2  +=     182.064 * cos(3.05304276213  + 16703.30595098250 *T)
    sumY2  +=     172.815 * cos(2.94873550203  + 3337.33312583431 *T)
    sumY2  +=     164.573 * cos(3.04867001053  + 3344.37936253230 *T)
    sumY2  +=     128.077 * cos(6.01351315277  + 6680.98103591609 *T)
    sumY2  +=      93.944 * cos(5.65705056742  + 3185.43584474911 *T)
    sumY2  +=      89.463 * cos(0.46100253521  + 3496.27664361750 *T)
    sumY2  +=      40.244 * cos(3.43638132202  + 6684.99178923210 *T)
    sumY2  +=      36.661 * cos(0.03273396254  + 1059.62574767270 *T)
    sumY2  +=      27.490 * cos(3.40043638863  + 20043.91837768230 *T)
    sumY2  +=      24.847 * cos(2.00937217942  + 3149.40797807170 *T)
    sumY2  +=      23.186 * cos(4.00844398003  + 3532.30451029490 *T)
    sumY2  +=      24.495 * cos(4.15948776712  + 2544.55823736690 *T)
    sumY2  +=      20.269 * cos(1.87109940843  + 4137.15425099970 *T)
    sumY2  +=      18.207 * cos(1.86611344735  + 3098.12764020930 *T)
    sumY2  +=      17.365 * cos(2.58782340890  + 2942.70724077510 *T)
    sumY2  +=      16.922 * cos(4.25054228122  + 3583.58484815730 *T)
    sumY2  +=      16.265 * cos(5.76223029450  + 5622.08674069391 *T)
    sumY2  +=      15.988 * cos(0.04373738136  + 1194.69082770810 *T)
    sumY2  +=      15.705 * cos(3.19532238212  + 3739.00524759150 *T)
    sumY2  +=      13.211 * cos(3.57389938080  + 1990.98883452450 *T)
    sumY2  +=      16.835 * cos(0.77752025128  + 2787.28684134090 *T)
    sumY2  +=      12.623 * cos(2.17797703414  + 2389.13783793270 *T)
    sumY2  +=      13.191 * cos(0.10921424588  + 10021.59346261590 *T)
    sumY2  +=      12.309 * cos(5.52906661092  + 3339.87592311510 *T)
    sumY2  +=      15.092 * cos(4.96166949645  + 1592.83983111630 *T)
    sumY2  +=      11.292 * cos(6.01108408051  + 6526.04827144891 *T)
    sumY2  +=      12.308 * cos(0.55813761009  + 3341.83656525151 *T)
    sumY2  +=      11.493 * cos(1.28862023253  + 796.54182429990 *T)
    sumY2  +=      14.005 * cos(5.33903523173  + 3894.42564702570 *T)
    sumY2  +=      10.521 * cos(0.55697469416  + 4400.23817437250 *T)
    sumY2  +=      13.157 * cos(4.20120058086  + 155.17658195069 *T)
    sumY2  +=       9.155 * cos(3.93426364819  + 4292.57465043390 *T)
    sumY2  +=       8.383 * cos(2.52922143175  + 4690.72365384210 *T)
    sumY2  +=      10.715 * cos(5.57456315750  + 2146.40923395870 *T)
    sumY2  +=       8.087 * cos(3.77030329034  + 10025.60421593190 *T)
    sumY2  +=       8.070 * cos(5.98273424636  + 5487.02166065850 *T)
    sumY2  +=       8.313 * cos(5.94079629702  + 2957.95971196010 *T)
    sumY2  +=       7.119 * cos(0.63326970116  + 1589.31671276730 *T)
    sumY2  +=       7.434 * cos(0.16010588018  + 3723.75277640650 *T)
    sumY2  +=       8.463 * cos(1.12131692970  + 5088.87265725030 *T)
    sumY2  +=       8.851 * cos(0.49531276074  + 4535.30325440790 *T)
    sumY2  +=       6.118 * cos(3.76286762082  + 7079.61767429131 *T)
    sumY2  +=       5.446 * cos(4.45219169081  + 5885.17066406670 *T)
    sumY2  +=       5.632 * cos(2.80162657949  + 553.32558535889 *T)
    sumY2  +=       5.307 * cos(3.82276150159  + 242.97242145751 *T)
    sumY2  +=       4.779 * cos(0.69754013484  + 1748.26023055050 *T)
    sumY2  +=       5.798 * cos(3.32076929671  + 7.35736448430 *T)
    sumY2  +=       5.029 * cos(3.76894152326  + 2273.87313202630 *T)
    sumY2  +=       4.779 * cos(5.06962801761  + 3333.74269718250 *T)
    sumY2  +=       4.645 * cos(1.40550063367  + 951.47458876710 *T)
    sumY2  +=       5.308 * cos(2.47347798651  + 6490.02040477150 *T)
    sumY2  +=       5.680 * cos(0.65754186652  + 529.93478257810 *T)
    sumY2  +=       4.316 * cos(5.37670156241  + 4933.45225781610 *T)
    sumY2  +=       4.455 * cos(6.20652322732  + 2288.10022602790 *T)
    sumY2  +=       4.106 * cos(5.29894197224  + 5092.39577559930 *T)
    sumY2  +=       4.484 * cos(2.40351988921  + 3870.54720927791 *T)
    sumY2  +=       4.129 * cos(1.08737767367  + 3347.96979118411 *T)
    sumY2  +=       3.982 * cos(3.74330534798  + 23384.53080438210 *T)
    sumY2  +=       3.668 * cos(3.61786789234  + 3340.83899053111 *T)
    sumY2  +=       3.668 * cos(5.64861564310  + 3340.87349783550 *T)
    sumY2  +=       4.603 * cos(2.27732033063  + 7477.76667769950 *T)
    sumY2  +=       3.137 * cos(2.23673642314  + 6438.74006690910 *T)
    sumY2  +=       3.223 * cos(0.63128088203  + 426.84200835950 *T)
    sumY2  +=       4.083 * cos(0.01640199107  + 1349.62359217529 *T)
    sumY2  +=       3.073 * cos(4.92973254619  + 1747.77259558350 *T)
    sumY2  +=       3.724 * cos(2.35536621176  + 6151.77770578850 *T)
    sumY2  +=       3.580 * cos(1.86170380048  + 2280.98667902709 *T)
    sumY2  +=       2.506 * cos(3.96683857097  + 5331.60126122430 *T)
    sumY2  +=       2.389 * cos(0.46414156874  + 6836.88907031730 *T)
    sumY2  +=       2.777 * cos(3.85084197237  + 382.65271473970 *T)
    sumY2  +=       2.243 * cos(2.10253884289  + 1350.11122714230 *T)
    sumY2  +=       2.179 * cos(1.14130873025  + 6127.89926804070 *T)
    sumY2  +=       2.794 * cos(2.98101234342  + 6677.94555253410 *T)
    sumY2  +=       2.007 * cos(0.96180685485  + 6682.44899195131 *T)
    sumY2  +=       2.204 * cos(4.14833269588  + 9492.39013248830 *T)
    sumY2  +=       1.869 * cos(0.85637111445  + 4929.92913946711 *T)
    sumY2  +=       1.955 * cos(5.49367553822  + 6681.45141723091 *T)
    sumY2  +=       2.432 * cos(5.68694139605  + 640.14110379750 *T)
    sumY2  +=       1.785 * cos(3.07914982123  + 3554.15533962130 *T)
    sumY2  +=       1.956 * cos(1.24043341247  + 6681.48592453530 *T)
    sumY2  +=       2.128 * cos(0.95091645133  + 7740.85060107230 *T)
    sumY2  +=       2.132 * cos(1.19742152629  + 398.39282089170 *T)
    sumY2  +=       1.915 * cos(1.68899605602  + 191.20444862810 *T)
    sumY2  +=       1.620 * cos(0.87029756453  + 7875.91568110771 *T)
    sumY2  +=       1.540 * cos(2.82385201055  + 3127.55714874530 *T)
    sumY2  +=       1.960 * cos(4.59845924354  + 6924.19727485710 *T)
    sumY2  +=       1.559 * cos(5.48531052433  + 2281.47431399410 *T)
    sumY2  +=       1.569 * cos(6.04375472798  + 8969.81271439450 *T)
    sumY2  +=       1.559 * cos(3.60405990781  + 8955.58562039291 *T)
    sumY2  +=       1.390 * cos(0.44447195188  + 13362.20588931570 *T)
    sumY2  +=       1.868 * cos(2.55406306387  + 5729.75026463250 *T)
    sumY2  +=       1.714 * cos(2.83026602085  + 6674.35512388230 *T)
    sumY2  +=       1.444 * cos(1.21697594403  + 2119.00767786191 *T)
    sumY2  +=       1.461 * cos(4.09849875613  + 13366.21664263171 *T)
    sumY2  +=       1.339 * cos(3.68138328651  + 2145.92159899169 *T)
    sumY2  +=       1.222 * cos(5.19573003811  + 6254.87048000710 *T)
    sumY2  +=       1.257 * cos(2.41919385934  + 641.12142486571 *T)
    sumY2  +=       1.485 * cos(0.07665766008  + 9866.66069814871 *T)
    sumY2  +=       1.566 * cos(4.89472263428  + 8962.69916739370 *T)
    sumY2  +=       1.160 * cos(0.02227881429  + 6298.57213865991 *T)
    sumY2  +=       1.124 * cos(4.70721954486  + 12833.00255918811 *T)
    sumY2  +=       1.479 * cos(4.17535755129  + 10420.23010099111 *T)
    sumY2  +=       1.201 * cos(1.32662761285  + 3767.45443505930 *T)
    sumY2  +=       1.207 * cos(1.69591474761  + 3340.78893388051 *T)
    sumY2  +=       1.207 * cos(1.28745285448  + 3340.92355448610 *T)
    sumY2  +=       1.403 * cos(1.70344023054  + 6283.31966747490 *T)
    sumY2  +=       1.082 * cos(0.42998284063  + 12303.31159409350 *T)
    sumY2  +=       1.021 * cos(4.82243595142  + 4562.70481050471 *T)
    sumY2  +=       1.086 * cos(3.14202913465  + 1751.29571393250 *T)
    sumY2  +=       1.153 * cos(3.38794180744  + 7211.15963597770 *T)
    sumY2  +=       1.277 * cos(3.27189335276  + 2699.97863680110 *T)
    sumY2  +=       0.930 * cos(4.55598641101  + 2811.16527908870 *T)
    sumY2  +=       0.918 * cos(0.03623887552  + 155.66421691770 *T)
    sumY2  +=       0.890 * cos(0.84840834382  + 3343.89172756530 *T)
    sumY2  +=       0.991 * cos(2.88410008262  + 9830.63283147130 *T)
    sumY2  +=       1.152 * cos(3.50867300596  + 951.96222373410 *T)
    sumY2  +=       0.877 * cos(2.64329046587  + 10818.37910439930 *T)
    sumY2  +=       1.028 * cos(2.84086438787  + 3981.73385156551 *T)
    sumY2  +=       0.724 * cos(4.91589406941  + 553.81322032590 *T)
    sumY2  +=       0.813 * cos(1.76324310782  + 10022.09835123510 *T)
    sumY2  +=       0.728 * cos(4.61545970674  + 853.44019923550 *T)
    sumY2  +=       0.746 * cos(3.59416377355  + 9623.93209417470 *T)
    sumY2  +=       0.684 * cos(5.74096167481  + 8274.06468451590 *T)
    sumY2  +=       0.813 * cos(6.01575669610  + 10022.06384393071 *T)
    sumY2  +=       0.641 * cos(3.57436678344  + 6681.40136058031 *T)
    sumY2  +=       0.632 * cos(5.44429272359  + 10018.55797923390 *T)
    sumY2  +=       0.616 * cos(2.91086112437  + 7235.03807372550 *T)
    sumY2  +=       0.846 * cos(1.46603022319  + 6688.58221788390 *T)
    sumY2  +=       0.857 * cos(0.35841204861  + 6041.57138456910 *T)
    sumY2  +=       0.572 * cos(6.06652679327  + 529.44714761109 *T)
    sumY2  +=       0.578 * cos(3.67902081482  + 6040.59106350090 *T)
    sumY2  +=       0.568 * cos(0.98475898207  + 15643.92402079330 *T)
    sumY2  +=       0.560 * cos(3.79594148168  + 397.90518592470 *T)
    sumY2  +=       0.661 * cos(3.68465313325  + 6894.76776632111 *T)
    sumY2  +=       0.538 * cos(0.50329899287  + 4690.23601887509 *T)
    sumY2  +=       0.701 * cos(2.07056825974  + 3178.38960805111 *T)
    sumY2  +=       0.689 * cos(5.65192172854  + 796.05418933290 *T)
    sumY2  +=       0.688 * cos(4.04353453116  + 3503.32288031550 *T)
    sumY2  +=       0.545 * cos(3.41001382974  + 2766.51144584850 *T)
    sumY2  +=       0.550 * cos(1.45021906208  + 7633.18707713370 *T)
    sumY2  +=       0.640 * cos(3.16585349800  + 6681.53598118590 *T)
    sumY2  +=       0.510 * cos(3.26303726541  + 3893.93801205869 *T)
    sumY2  +=       0.617 * cos(0.99681498656  + 3253.54803964350 *T)
    sumY2  +=       0.596 * cos(5.12698391726  + 3428.16444872311 *T)
    sumY2  +=       0.629 * cos(1.64232890349  + 6.86972951729 *T)
    sumY2  +=       0.603 * cos(1.87080717732  + 4292.08701546690 *T)
    sumY2  +=       0.548 * cos(5.71337800864  + 6680.48834981490 *T)
    sumY2  +=       0.444 * cos(2.70322166823  + 3915.20104251810 *T)
    sumY2  +=       0.411 * cos(6.27896145140  + 3980.75353049730 *T)
    sumY2  +=       0.560 * cos(4.08717867095  + 26725.14323108190 *T)
    sumY2  +=       0.399 * cos(4.13396201386  + 12964.54452087450 *T)
    sumY2  +=       0.383 * cos(1.25733997062  + 2700.47132290230 *T)
    sumY2  +=       0.409 * cos(3.80977418197  + 7064.36520310630 *T)
    sumY2  +=       0.499 * cos(4.46670852178  + 2914.25805330730 *T)
    sumY2  +=       0.408 * cos(1.69107992579  + 2648.69864295650 *T)
    sumY2  +=       0.494 * cos(1.01711860076  + 1039.27042827391 *T)
    sumY2  +=       0.367 * cos(1.15983352898  + 8270.54156616690 *T)
    sumY2  +=       0.379 * cos(1.31139516762  + 11081.46302777211 *T)
    sumY2  +=       0.354 * cos(4.23224813069  + 1194.20319274110 *T)
    sumY2  +=       0.403 * cos(0.13228544943  + 8031.33608054190 *T)
    sumY2  +=       0.449 * cos(3.99464222002  + 213.05527795450 *T)
    sumY2  +=       0.397 * cos(3.84049099896  + 2544.07060239989 *T)
    sumY2  +=       0.344 * cos(4.40570415749  + 4033.01384541010 *T)
    sumY2  +=       0.314 * cos(2.74388586902  + 6872.91693699471 *T)
    sumY2  +=       0.410 * cos(4.28621405712  + 8433.00820229911 *T)
    sumY2  +=       0.297 * cos(0.09404951710  + 574.10098085129 *T)
    sumY2  +=       0.326 * cos(1.34034230009  + 10023.06141865111 *T)
    sumY2  +=       0.294 * cos(1.26195385099  + 5459.62010456170 *T)
    sumY2  +=       0.278 * cos(1.24918021097  + 11216.52810780750 *T)
    sumY2  +=       0.274 * cos(4.25063115139  + 10551.77206267750 *T)
    sumY2  +=       0.355 * cos(4.32827325380  + 8672.21368792411 *T)
    sumY2  +=       0.287 * cos(4.60679040146  + 3495.78900865049 *T)
    sumY2  +=       0.282 * cos(4.55464471058  + 13760.84252769090 *T)
    sumY2  +=       0.359 * cos(6.00520105343  + 3097.64000524229 *T)
    sumY2  +=       0.266 * cos(3.68597750720  + 10022.14840788570 *T)
    sumY2  +=       0.284 * cos(1.46414557972  + 5828.27228913110 *T)
    sumY2  +=       0.314 * cos(5.08365465116  + 16173.61498588790 *T)
    sumY2  +=       0.263 * cos(1.47512984004  + 9468.51169474051 *T)
    sumY2  +=       0.277 * cos(4.29922856475  + 3336.84549086730 *T)
    sumY2  +=       0.251 * cos(1.08022178481  + 2942.21960580809 *T)
    sumY2  +=       0.269 * cos(2.34301425928  + 2487.17222746430 *T)
    sumY2  +=       0.256 * cos(5.52886759635  + 5088.38502228330 *T)
    sumY2  +=       0.265 * cos(4.09423905022  + 10022.01378728011 *T)
    sumY2  +=       0.241 * cos(4.64969172321  + 2301.82963339290 *T)
    sumY2  +=       0.232 * cos(5.34839535146  + 13359.17040593370 *T)
    sumY2  +=       0.261 * cos(3.53492755311  + 1221.60474883790 *T)
    sumY2  +=       0.225 * cos(5.21546840346  + 1751.78334889950 *T)
    sumY2  +=       0.252 * cos(3.11351356709  + 1066.73929467350 *T)
    sumY2  +=       0.246 * cos(4.94803970472  + 10264.80970155691 *T)
    sumY2  +=       0.285 * cos(0.15604991648  + 12310.42514109431 *T)
    sumY2  +=       0.243 * cos(2.47247524273  + 9779.35249360890 *T)
    sumY2  +=       0.220 * cos(5.34295457148  + 8429.48508395010 *T)
    sumY2  +=       0.280 * cos(1.23534408406  + 5091.90814063229 *T)
    sumY2  +=       0.214 * cos(1.14917768049  + 2699.49100183409 *T)
    sumY2  +=       0.234 * cos(1.82528787325  + 6468.16957544510 *T)
    sumY2  +=       0.247 * cos(1.05434658596  + 213.54291292150 *T)
    sumY2  +=       0.265 * cos(1.38035585731  + 18984.53644749310 *T)
    sumY2  +=       0.206 * cos(5.10353562320  + 6836.40143535029 *T)
    sumY2  +=       0.198 * cos(2.14333009014  + 13362.71077793490 *T)
    sumY2  +=       0.199 * cos(0.11264416842  + 13362.67627063050 *T)
    sumY2  +=       0.267 * cos(0.20812545566  + 5614.97319369310 *T)
    sumY2  +=       0.200 * cos(2.34389696982  + 7632.69944216670 *T)
    sumY2  +=       0.238 * cos(4.42808531081  + 16706.82906933150 *T)
    sumY2  +=       0.189 * cos(4.24906121119  + 3723.26514143949 *T)
    sumY2  +=       0.187 * cos(2.81434210031  + 1592.35219614930 *T)
    sumY2  +=       0.211 * cos(1.82857292487  + 7108.06686175911 *T)
    sumY2  +=       0.244 * cos(1.68371543063  + 242.48478649050 *T)
    sumY2  +=       0.193 * cos(2.86364604815  + 9070.36269133230 *T)
    sumY2  +=       0.198 * cos(3.69840284408  + 191.69208359510 *T)
    sumY2  +=       0.172 * cos(3.26900025992  + 13171.24525817110 *T)
    sumY2  +=       0.196 * cos(0.42643861998  + 13207.27312484851 *T)
    sumY2  +=       0.234 * cos(3.70618365075  + 7234.55043875850 *T)
    sumY2  +=       0.177 * cos(0.37345902868  + 9639.18456535971 *T)
    sumY2  +=       0.168 * cos(0.18332006876  + 1052.51220067191 *T)
    sumY2  +=       0.170 * cos(1.42456308210  + 4379.88285497370 *T)
    sumY2  +=       0.222 * cos(5.10471677470  + 5642.44206009270 *T)
    sumY2  +=       0.157 * cos(6.07794150900  + 383.14034970670 *T)
    sumY2  +=       0.155 * cos(3.90931440460  + 12296.19804709270 *T)
    sumY2  +=       0.154 * cos(3.04637044996  + 10575.65050042530 *T)
    sumY2  +=       0.155 * cos(0.75034905139  + 16702.81831601549 *T)
    sumY2  +=       0.171 * cos(3.35870246354  + 3184.94820978210 *T)
    sumY2  +=       0.148 * cos(4.16708586457  + 10235.38019302090 *T)
    sumY2  +=       0.140 * cos(0.21579055276  + 6438.25243194210 *T)
    sumY2  +=       0.141 * cos(1.84664759575  + 10029.19464458370 *T)
    sumY2  +=       0.154 * cos(1.16100089333  + 2810.67764412170 *T)
    sumY2  +=       0.137 * cos(5.90622302168  + 1437.41943168211 *T)
    sumY2  +=       0.133 * cos(4.90961144014  + 8827.63408735830 *T)
    sumY2  +=       0.131 * cos(0.25165704659  + 3030.01544531491 *T)
    sumY2  +=       0.134 * cos(2.50558754072  + 2301.34199842589 *T)
    sumY2  +=       0.128 * cos(0.82893704653  + 9382.18381126891 *T)
    sumY2  +=       0.119 * cos(1.05918597755  + 5614.48555872609 *T)
    sumY2  +=       0.120 * cos(3.55524858904  + 3320.08084869090 *T)
    sumY2  +=       0.119 * cos(5.71468810799  + 3361.63163967570 *T)
    sumY2  +=       0.116 * cos(5.87199885875  + 3651.69704305170 *T)
    sumY2  +=       0.117 * cos(4.28224437586  + 6701.82399028190 *T)
    sumY2  +=       0.146 * cos(1.21928643824  + 3738.51761262449 *T)
    sumY2  +=       0.129 * cos(2.32040419130  + 6696.72114206811 *T)
    sumY2  +=       0.118 * cos(3.47554817815  + 149.80701461810 *T)
    sumY2  +=       0.107 * cos(3.54962193881  + 4407.35172137330 *T)
    sumY2  +=       0.114 * cos(3.18671462047  + 7322.34627826531 *T)
    sumY2  +=       0.105 * cos(1.58662031954  + 6666.21619969810 *T)
    sumY2  +=       0.108 * cos(4.17304364327  + 10575.16286545830 *T)
    sumY2  +=       0.115 * cos(2.21982999199  + 9168.88471583091 *T)
    sumY2  +=       0.130 * cos(1.61147250477  + 10973.79950383350 *T)
    sumY2  +=       0.101 * cos(2.88371987525  + 14158.99153109910 *T)
    sumY2  +=       0.102 * cos(2.72041962157  + 4562.21717553770 *T)
    sumY2  +=       0.099 * cos(4.08432740077  + 5621.59910572690 *T)
    sumY2  +=       0.101 * cos(0.61635489651  + 9778.86485864190 *T)
    sumY2  +=       0.115 * cos(6.16350509467  + 3148.92034310470 *T)
    sumY2  +=       0.131 * cos(5.90042253901  + 6661.11335148431 *T)
    sumY2  +=       0.134 * cos(1.31032537502  + 2913.77041834030 *T)
    sumY2  +=       0.113 * cos(2.14345351717  + 6519.00203475090 *T)
    sumY2  +=       0.096 * cos(5.69334604420  + 2700.95895786930 *T)
    sumY2  +=       0.094 * cos(4.95566590883  + 6816.53375091851 *T)
    sumY2  +=       0.097 * cos(0.21546380481  + 11371.94850724170 *T)
    sumY2  +=       0.118 * cos(3.13572565534  + 10213.52936369450 *T)
    sumY2  +=       0.088 * cos(4.49196657842  + 1835.56843509031 *T)
    sumY2  +=       0.096 * cos(0.62767532388  + 15114.23305569870 *T)
    sumY2  +=       0.119 * cos(5.54166774257  + 10177.01386205009 *T)
    sumY2  +=       0.096 * cos(1.11981306669  + 6684.50415426510 *T)
    sumY2  +=       0.117 * cos(3.59287220197  + 9225.78309076650 *T)
    sumY2  +=       0.095 * cos(5.86251553025  + 5486.53402569149 *T)
    sumY2  +=       0.088 * cos(1.02951159717  + 8030.84844557490 *T)
    sumY2  +=       0.085 * cos(1.84083503490  + 6151.29007082150 *T)
    sumY2  +=       0.089 * cos(5.88609595798  + 10177.50149701710 *T)
    sumY2  +=       0.081 * cos(0.23842229522  + 5244.29305668450 *T)
    sumY2  +=       0.081 * cos(5.36252162300  + 1505.04399160949 *T)
    sumY2  +=       0.079 * cos(3.99668001072  + 10404.97762980610 *T)
    sumY2  +=       0.095 * cos(4.50704283927  + 16305.15694757430 *T)
    sumY2  +=       0.090 * cos(3.09437543611  + 10014.96755058210 *T)
    sumY2  +=       0.098 * cos(3.53367325303  + 9808.78200214491 *T)
    sumY2  +=       0.096 * cos(4.43634405431  + 1059.13811270570 *T)
    sumY2  +=       0.073 * cos(5.43736834192  + 6660.69327539070 *T)
    sumY2  +=       0.077 * cos(5.92439560709  + 4136.66661603269 *T)
    sumY2  +=       0.071 * cos(5.75155209191  + 2274.36076699330 *T)
    sumY2  +=       0.075 * cos(4.06478068204  + 7903.31723720450 *T)
    sumY2  +=       0.098 * cos(1.59156329835  + 6040.10342853389 *T)
    sumY2  +=       0.085 * cos(5.15748985329  + 4194.05262593530 *T)
    sumY2  +=       0.073 * cos(2.01248478882  + 9380.71585523369 *T)
    sumY2  +=       0.093 * cos(1.38217187583  + 1990.50119955750 *T)
    sumY2  +=       0.086 * cos(1.47044925000  + 20752.42054499190 *T)
    sumY2  +=       0.069 * cos(6.09792009314  + 1903.68062998470 *T)
    sumY2  +=       0.070 * cos(6.27372832300  + 2388.65020296570 *T)
    sumY2  +=       0.069 * cos(3.74116772307  + 8962.21153242670 *T)
    sumY2  +=       0.066 * cos(6.06493527472  + 13517.62628874989 *T)
    sumY2  +=       0.069 * cos(4.43286275856  + 30065.75565778170 *T)
    sumY2  +=       0.079 * cos(4.41786563798  + 3337.26581553151 *T)
    sumY2  +=       0.079 * cos(4.84899849761  + 3344.44667283510 *T)
    sumY2  +=       0.064 * cos(3.87843028506  + 25287.47998191629 *T)
    sumY2  +=       0.065 * cos(1.67365918241  + 2631.86644190671 *T)
    sumY2  +=       0.065 * cos(5.75101393296  + 20199.33877711650 *T)
    sumY2  +=       0.061 * cos(0.59227624984  + 4032.52621044309 *T)
    sumY2  +=       0.062 * cos(3.62324723182  + 27035.49639498329 *T)
    sumY2  +=       0.062 * cos(2.99894804940  + 5641.95442512570 *T)
    sumY2  +=       0.063 * cos(6.24532574874  + 12936.09533340670 *T)
    sumY2  +=       0.084 * cos(3.95654625717  + 1903.19299501769 *T)
    sumY2  +=       0.071 * cos(4.38296840291  + 11770.09751064990 *T)
    sumY2  +=       0.072 * cos(0.93455070924  + 13119.47728534169 *T)
    sumY2  +=       0.063 * cos(4.06602682655  + 13362.76083458550 *T)
    sumY2  +=       0.061 * cos(0.93007138211  + 17411.80811829210 *T)
    sumY2  +=       0.060 * cos(0.07444626057  + 3205.79116414790 *T)
    sumY2  +=       0.058 * cos(6.22909541546  + 7255.81346921790 *T)
    sumY2  +=       0.080 * cos(0.78227120577  + 21795.45790909830 *T)
    sumY2  +=       0.064 * cos(5.47349341844  + 6768.77687542291 *T)
    sumY2  +=       0.063 * cos(5.13193846811  + 3340.38586286849 *T)
    sumY2  +=       0.064 * cos(0.45687855320  + 3914.71340755109 *T)
    sumY2  +=       0.056 * cos(1.90007366831  + 27434.13303335850 *T)
    sumY2  +=       0.057 * cos(2.44461266111  + 12721.32828193349 *T)
    sumY2  +=       0.058 * cos(2.91190076994  + 3475.92132421871 *T)
    sumY2  +=       0.066 * cos(0.93306761250  + 14127.99895376209 *T)
    sumY2  +=       0.055 * cos(1.64400276973  + 4846.14405327630 *T)
    sumY2  +=       0.061 * cos(1.64746236660  + 5989.31106965630 *T)
    sumY2  +=       0.056 * cos(5.44862221942  + 16699.78283263350 *T)
    sumY2  +=       0.063 * cos(3.10372430555  + 3340.35135556410 *T)
    sumY2  +=       0.054 * cos(4.84864924468  + 2786.79920637390 *T)
    sumY2  +=       0.059 * cos(4.79379884356  + 7063.87756813929 *T)
    sumY2  +=       0.052 * cos(4.93425415169  + 11128.73226830069 *T)
    sumY2  +=       0.067 * cos(3.08005724919  + 2233.71743849851 *T)
    sumY2  +=       0.051 * cos(4.01654589565  + 30376.10882168309 *T)
    sumY2  +=       0.050 * cos(5.75439961134  + 17924.66688233689 *T)
    sumY2  +=       0.050 * cos(5.72861660044  + 17925.15451730390 *T)
    sumY2  +=       0.049 * cos(1.66917730044  + 14422.07545447190 *T)
    sumY2  +=       0.050 * cos(1.30948740291  + 6594.16046634330 *T)
    sumY2  +=       0.065 * cos(3.77067948165  + 6107.12387254830 *T)
    sumY2  +=       0.050 * cos(4.44963662233  + 4049.84604645990 *T)
    sumY2  +=       0.049 * cos(2.99008112282  + 14584.05445563709 *T)
    sumY2  +=       0.056 * cos(1.27489003307  + 6702.24406637550 *T)
    sumY2  +=       0.051 * cos(6.18955107115  + 8432.52056733209 *T)
    sumY2  +=       0.060 * cos(2.66632277065  + 22743.65319699990 *T)
    sumY2  +=       0.050 * cos(2.95116022713  + 426.35437339250 *T)
    sumY2  +=       0.056 * cos(4.06875165560  + 9499.50367948910 *T)
    sumY2  +=       0.048 * cos(6.03463544833  + 21265.76694400370 *T)
    sumY2  +=       0.045 * cos(4.27091323030  + 13149.39442884470 *T)
    sumY2  +=       0.055 * cos(4.47134992564  + 13362.62621397991 *T)
    sumY2  +=       0.047 * cos(4.49173897577  + 9381.69617630189 *T)
    sumY2  +=       0.051 * cos(5.47523455944  + 19514.22741258770 *T)
    sumY2  +=       0.047 * cos(0.46146406745  + 7321.36595719711 *T)
    sumY2  +=       0.048 * cos(1.27966315736  + 23141.80220040810 *T)
    sumY2  +=       0.062 * cos(0.36573942120  + 6041.08374960209 *T)
    sumY2  +=       0.056 * cos(4.19518698917  + 9595.48290670690 *T)
    sumY2  +=       0.045 * cos(0.79736291280  + 17256.38771885789 *T)
    sumY2  +=       0.043 * cos(4.70566458674  + 9485.27658548750 *T)
    sumY2  +=       0.055 * cos(2.85186159808  + 16061.94070863329 *T)
    sumY2  +=       0.042 * cos(1.50287163065  + 8800.23253126150 *T)
    sumY2  +=       0.044 * cos(3.94795993501  + 17256.87535382490 *T)
    sumY2  +=       0.055 * cos(5.86419756588  + 8425.89465529830 *T)
    sumY2  +=       0.043 * cos(0.99542400694  + 6705.34710863090 *T)
    sumY2  +=       0.042 * cos(5.91099352451  + 25287.96761688330 *T)
    sumY2  +=       0.050 * cos(0.81879571159  + 11773.62062899890 *T)
    sumY2  +=       0.046 * cos(2.96889198903  + 10973.31186886650 *T)
    sumY2  +=       0.042 * cos(0.44638157419  + 6685.05909953490 *T)
    sumY2  +=       0.043 * cos(1.46762912473  + 16460.08971204149 *T)
    sumY2  +=       0.042 * cos(2.94486773835  + 6158.89125278930 *T)
    sumY2  +=       0.043 * cos(6.18134536230  + 6709.91785835091 *T)
    sumY2  +=       0.040 * cos(2.24681835171  + 6531.90547374850 *T)
    sumY2  +=       0.050 * cos(4.85860818020  + 13892.38448937730 *T)
    sumY2  +=       0.039 * cos(3.92914400055  + 691.91378374330 *T)
    sumY2  +=       0.053 * cos(4.53825887195  + 3347.90248088131 *T)
    sumY2  +=       0.041 * cos(3.84566705233  + 12323.17927852529 *T)
    sumY2  +=       0.050 * cos(0.75703710335  + 18454.84548239850 *T)
    sumY2  +=       0.040 * cos(2.10179493054  + 13554.14179039430 *T)
    sumY2  +=       0.054 * cos(4.70225786858  + 3333.81000748530 *T)
    sumY2  +=       0.049 * cos(2.91514601795  + 6657.59023313531 *T)
    sumY2  +=       0.049 * cos(1.48899492537  + 11611.15399286670 *T)
    sumY2  +=       0.039 * cos(4.33032994271  + 5243.80542171749 *T)
    sumY2  +=       0.036 * cos(0.22475591221  + 19801.18977370830 *T)
    sumY2  +=       0.038 * cos(5.19414151817  + 8186.75647997610 *T)
    sumY2  +=       0.045 * cos(3.23746394812  + 5628.71265272770 *T)
    sumY2  +=       0.038 * cos(5.80558264651  + 11614.67711121570 *T)
    sumY2  +=       0.042 * cos(0.01243986230  + 6677.87824223130 *T)
    sumY2  +=       0.037 * cos(1.66970207361  + 8428.99744898310 *T)
    sumY2  +=       0.036 * cos(2.54319902644  + 17655.02435723310 *T)
    sumY2  +=       0.036 * cos(0.22073261720  + 19800.70213874130 *T)
    sumY2  +=       0.038 * cos(1.75936320775  + 22325.14887419290 *T)
    sumY2  +=       0.039 * cos(1.95232336514  + 3583.09721319030 *T)
    sumY2  +=       0.044 * cos(4.50288561893  + 25686.11662029150 *T)
    sumY2  +=       0.046 * cos(3.04209835581  + 4447.99504986810 *T)
    sumY2  +=       0.047 * cos(3.41581645664  + 8982.56685182550 *T)
    sumY2  +=       0.042 * cos(4.56050725497  + 6254.38284504009 *T)
    sumY2  +=       0.035 * cos(6.07448879945  + 18606.25512851669 *T)
    sumY2  +=       0.035 * cos(0.19110171429  + 5099.50932260011 *T)
    sumY2  +=       0.034 * cos(0.27172687467  + 5827.78465416409 *T)
    sumY2  +=       0.044 * cos(0.60787197203  + 7902.82960223750 *T)
    sumY2  +=       0.034 * cos(4.27738119614  + 15663.79170522509 *T)
    sumY2  +=       0.034 * cos(1.49602757441  + 28230.43104017490 *T)
    sumY2  +=       0.042 * cos(1.70888858696  + 13363.67384535091 *T)
    sumY2  +=       0.034 * cos(0.52829912831  + 3.27930086550 *T)
    sumY2  +=       0.038 * cos(3.76470778956  + 6843.93530701530 *T)
    sumY2  +=       0.033 * cos(5.81667236094  + 11243.44202893729 *T)
    sumY2  +=       0.033 * cos(0.01403753138  + 4778.03185838191 *T)
    sumY2  +=       0.035 * cos(3.11315123125  + 26084.26562369970 *T)
    sumY2  +=       0.042 * cos(1.90806796448  + 31571.04346687470 *T)
    sumY2  +=       0.034 * cos(1.12608521373  + 692.40141871030 *T)
    sumY2  +=       0.034 * cos(5.18123693961  + 3347.48215621710 *T)
    sumY2  +=       0.037 * cos(4.66424749339  + 3341.34893028450 *T)
    sumY2  +=       0.037 * cos(2.42340799880  + 3127.06951377830 *T)
    sumY2  +=       0.036 * cos(6.11462616077  + 24606.37937070350 *T)
    sumY2  +=       0.032 * cos(4.78655245028  + 8584.41784841729 *T)
    sumY2  +=       0.039 * cos(0.81289176033  + 7255.32583425090 *T)
    sumY2  +=       0.035 * cos(0.16869923739  + 3333.25506221550 *T)
    sumY2  +=       0.031 * cos(1.19546397800  + 10595.93826095070 *T)
    sumY2  +=       0.030 * cos(5.08879811241  + 20736.07597890910 *T)
    sumY2  +=       0.030 * cos(0.63027515429  + 10713.75106384269 *T)
    sumY2  +=       0.032 * cos(5.53354847657  + 1744.18216693170 *T)
    sumY2  +=       0.030 * cos(0.55638584024  + 3.76693583251 *T)
    sumY2  +=       0.036 * cos(1.13991556835  + 15113.74542073169 *T)
    sumY2  +=       0.030 * cos(4.73542045957  + 1222.09238380491 *T)
    sumY2  +=       0.031 * cos(0.87259059949  + 2575.06317973690 *T)
    sumY2  +=       0.031 * cos(5.25117631915  + 4106.64930862970 *T)
    sumY2  +=       0.029 * cos(1.54929293338  + 8827.14645239129 *T)
    sumY2  +=       0.030 * cos(0.26033037399  + 640.63378989870 *T)
    sumY2  +=       0.037 * cos(5.93731996113  + 14712.56093394150 *T)
    sumY2  +=       0.028 * cos(0.10501454676  + 12509.49714253070 *T)
    sumY2  +=       0.030 * cos(4.45949277543  + 6282.83203250789 *T)
    sumY2  +=       0.029 * cos(0.47095574930  + 16276.70776010650 *T)
    sumY2  +=       0.028 * cos(5.08541958812  + 8440.12174929990 *T)
    sumY2  +=       0.032 * cos(1.34932157181  + 11769.60987568290 *T)
    sumY2  +=       0.030 * cos(5.38099741770  + 24076.68840560890 *T)
    sumY2  +=       0.028 * cos(0.59624749600  + 15651.03756779411 *T)
    sumY2  +=       0.035 * cos(4.67983397003  + 19004.40413192490 *T)
    sumY2  +=       0.035 * cos(2.79291396629  + 14712.07329897449 *T)
    sumY2  +=       0.028 * cos(1.28322421957  + 3553.66770465429 *T)
    sumY2  +=       0.027 * cos(4.82066236084  + 17101.45495439070 *T)
    sumY2  +=       0.029 * cos(1.21995619161  + 27831.79440179969 *T)
    sumY2  +=       0.029 * cos(1.21995619161  + 27832.28203676670 *T)
    sumY2  +=       0.027 * cos(3.25379008238  + 31172.89446346650 *T)
    sumY2  +=       0.027 * cos(5.09561724929  + 3312.40705671550 *T)
    sumY2  +=       0.026 * cos(3.78891136297  + 7373.62627210991 *T)
    sumY2  +=       0.033 * cos(3.24188267043  + 5629.20028769470 *T)
    sumY2  +=       0.027 * cos(4.17474865828  + 3369.30543165110 *T)
    sumY2  +=       0.027 * cos(5.73759562221  + 15265.64270181689 *T)
    sumY2  +=       0.034 * cos(0.09014018186  + 28628.58004358310 *T)
    sumY2  +=       0.027 * cos(3.31652753395  + 19402.55313533309 *T)
    sumY2  +=       0.027 * cos(4.95428626263  + 29026.72904699130 *T)
    sumY2  +=       0.026 * cos(2.52144336597  + 16065.82204399129 *T)
    sumY2  +=       0.026 * cos(1.66505639760  + 21548.71855180830 *T)
    sumY2  +=       0.028 * cos(5.18965453043  + 3531.81687532789 *T)
    sumY2  +=       0.033 * cos(2.07674201664  + 3870.05957431090 *T)
    sumY2  +=       0.025 * cos(6.03467999100  + 12722.79623796870 *T)
    sumY2  +=       0.024 * cos(0.33206789998  + 22747.04689739089 *T)
    sumY2  +=       0.025 * cos(3.73666908945  + 2957.47207699310 *T)
    sumY2  +=       0.032 * cos(1.62301076387  + 12302.82395912650 *T)
    sumY2  +=       0.023 * cos(4.44526699765  + 20597.48778052470 *T)
    sumY2  +=       0.026 * cos(5.49025029300  + 3877.66075627870 *T)
    sumY2  +=       0.025 * cos(2.45276955078  + 27433.64539839149 *T)
    sumY2  +=       0.022 * cos(3.74620147475  + 17654.53672226610 *T)
    sumY2  +=       0.021 * cos(0.40534339845  + 5085.28222859850 *T)
    sumY2  +=       0.022 * cos(6.02899389204  + 13745.10242153889 *T)
    sumY2  +=       0.023 * cos(2.71413139567  + 3325.60377299830 *T)
    sumY2  +=       0.022 * cos(0.92873709901  + 10714.23869880971 *T)
    sumY2  +=       0.023 * cos(0.27276624029  + 3356.10871536831 *T)
    sumY2  +=       0.022 * cos(3.22255788727  + 13916.26292712510 *T)
    sumY2  +=       0.020 * cos(4.78693375846  + 13518.11392371690 *T)
    sumY2  +=       0.020 * cos(2.18685576579  + 14054.36349054249 *T)
    sumY2  +=       0.020 * cos(5.32844841938  + 14054.85112550950 *T)
    sumY2  +=       0.020 * cos(0.47695326504  + 14071.19569159229 *T)
    sumY2  +=       0.021 * cos(2.78252415599  + 20354.27154158369 *T)
    sumY2  +=       0.027 * cos(3.22117918900  + 23694.88396828350 *T)
    sumY2  +=       0.027 * cos(5.28311501931  + 11925.03027511709 *T)
    sumY2  +=       0.023 * cos(4.68280672465  + 1588.82907780029 *T)
    sumY2  +=       0.020 * cos(1.50444868935  + 2817.79119112250 *T)
    sumY2  +=       0.020 * cos(5.49545285602  + 29977.95981827489 *T)
    sumY2  +=       0.019 * cos(5.62633616092  + 27035.98402995030 *T)
    sumY2  +=       0.020 * cos(3.61240830848  + 9381.20349020070 *T)
    sumY2  +=       0.020 * cos(1.71417155656  + 1063.14886602171 *T)
    sumY2  +=       0.020 * cos(2.07071082637  + 36659.67230664150 *T)
    sumY2  +=       0.021 * cos(0.27968009759  + 10596.42589591770 *T)
    sumY2  +=       0.019 * cos(0.78599577864  + 3340.43591951909 *T)
    sumY2  +=       0.020 * cos(2.37181321274  + 34911.65589357450 *T)
    sumY2  +=       0.019 * cos(6.18512352666  + 23539.95120381630 *T)
    sumY2  +=       0.019 * cos(2.54961603658  + 40000.28473334130 *T)
    sumY2  +=       0.020 * cos(2.00641556660  + 2803.56409712089 *T)
    sumY2  +=       0.024 * cos(3.64789379261  + 9225.29545579949 *T)
    sumY2  +=       0.024 * cos(6.09527572891  + 30376.59645665010 *T)
    sumY2  +=       0.022 * cos(3.61088135020  + 537.04832957890 *T)
    sumY2  +=       0.019 * cos(4.65375898570  + 30774.74546005830 *T)
    sumY2  +=       0.024 * cos(0.25109778339  + 21946.86755521649 *T)
    sumY2  +=       0.020 * cos(0.55978137151  + 10787.38652706230 *T)
    sumY2  +=       0.018 * cos(0.93810887810  + 23938.10020722450 *T)
    sumY2  +=       0.018 * cos(0.65703782614  + 2487.65986243130 *T)
    sumY2  +=       0.018 * cos(3.45241957900  + 24491.18197509989 *T)
    sumY2  +=       0.021 * cos(0.77336865730  + 522.82123557730 *T)
    sumY2  +=       0.023 * cos(5.60162687963  + 37455.97031345790 *T)
    sumY2  +=       0.024 * cos(4.31466786422  + 3077.77232081051 *T)
    sumY2  +=       0.018 * cos(1.24028262416  + 18208.10612510850 *T)
    sumY2  +=       0.024 * cos(4.24184949882  + 4393.12462737170 *T)
    sumY2  +=       0.022 * cos(5.20780077851  + 13915.77529215809 *T)
    sumY2  +=       0.020 * cos(1.19573990945  + 3340.30129891350 *T)
    sumY2  +=       0.017 * cos(5.14097504316  + 22345.01655862470 *T)
    sumY2  +=       0.018 * cos(2.26405048660  + 3492.75352526850 *T)
    sumY2  +=       0.018 * cos(0.71812377845  + 3188.95896309811 *T)
    sumY2  +=       0.017 * cos(2.31066064344  + 17013.65911488389 *T)
    sumY2  +=       0.017 * cos(5.73936839879  + 28229.94340520789 *T)
    sumY2  +=       0.017 * cos(6.13675229471  + 8186.26884500910 *T)
    sumY2  +=       0.017 * cos(1.27636900458  + 5618.56362234490 *T)
    sumY2  +=       0.016 * cos(1.17980677135  + 29978.44745324190 *T)
    sumY2  +=       0.016 * cos(2.20820580655  + 10448.67928845890 *T)
    sumY2  +=       0.022 * cos(4.95674400481  + 3603.94016755610 *T)
    sumY2  +=       0.021 * cos(5.05557236085  + 26637.34739157509 *T)
    sumY2  +=       0.015 * cos(3.72559974712  + 11371.46087227470 *T)
    sumY2  +=       0.018 * cos(0.65953210534  + 18451.32236404950 *T)
    sumY2  +=       0.015 * cos(4.39931221730  + 23546.99744051430 *T)
    sumY2  +=       0.020 * cos(1.62468835541  + 33319.05987994170 *T)
    sumY2  +=       0.015 * cos(0.99941976160  + 17277.65074931730 *T)
    sumY2  +=       0.020 * cos(1.01851229845  + 20043.43074271529 *T)

    # MARS Y Coord
    sumY3  = 0
    sumY3  +=    1905.181 * cos(5.26332442482  + 0.24381748350 *T)
    sumY3  +=     638.154 * cos(6.15054891881  + 6681.46867088311 *T)
    sumY3  +=     315.844 * cos(0.68812770928  + 10022.08109758290 *T)
    sumY3  +=     167.935 * cos(4.34052829824  + 3340.36860921629 *T)
    sumY3  +=      88.045 * cos(1.14611813288  + 13362.69352428270 *T)
    sumY3  +=      34.296 * cos(6.12954233244  + 3340.85624418330 *T)
    sumY3  +=      19.004 * cos(1.53394982779  + 16703.30595098250 *T)
    sumY3  +=      17.309 * cos(1.96405922769  + 3496.27664361750 *T)
    sumY3  +=      17.052 * cos(4.14009092313  + 3185.43584474911 *T)
    sumY3  +=      12.081 * cos(4.55895834821  + 6680.98103591609 *T)
    sumY3  +=       8.036 * cos(4.44721122484  + 3337.33312583431 *T)
    sumY3  +=       8.075 * cos(1.32135542935  + 3344.37936253230 *T)
    sumY3  +=       3.586 * cos(1.89896476323  + 20043.91837768230 *T)
    sumY3  +=       3.728 * cos(2.73731234443  + 3583.58484815730 *T)
    sumY3  +=       3.708 * cos(3.37008572755  + 3098.12764020930 *T)
    sumY3  +=       2.838 * cos(1.86335942445  + 6684.99178923210 *T)
    sumY3  +=       2.391 * cos(4.50157894761  + 6526.04827144891 *T)
    sumY3  +=       2.389 * cos(0.55881206863  + 3894.42564702570 *T)
    sumY3  +=       2.358 * cos(5.53634186088  + 2787.28684134090 *T)
    sumY3  +=       2.166 * cos(5.72850515937  + 155.17658195069 *T)
    sumY3  +=       1.411 * cos(4.89547894086  + 10021.59346261590 *T)
    sumY3  +=       1.300 * cos(5.43963179189  + 4292.57465043390 *T)
    sumY3  +=       1.268 * cos(0.63826775689  + 2389.13783793270 *T)
    sumY3  +=       1.168 * cos(0.32787695602  + 3149.40797807170 *T)
    sumY3  +=       1.170 * cos(5.49045547620  + 3532.30451029490 *T)
    sumY3  +=       1.149 * cos(0.25632426200  + 4137.15425099970 *T)
    sumY3  +=       1.000 * cos(0.78268074251  + 3339.87592311510 *T)
    sumY3  +=       1.053 * cos(5.58094872648  + 2544.55823736690 *T)
    sumY3  +=       0.909 * cos(4.03516115356  + 4690.72365384210 *T)
    sumY3  +=       0.997 * cos(5.22732432514  + 3341.83656525151 *T)
    sumY3  +=       0.874 * cos(2.00558552866  + 1990.98883452450 *T)
    sumY3  +=       0.773 * cos(2.25965967719  + 10025.60421593190 *T)
    sumY3  +=       0.800 * cos(1.65801132134  + 3723.75277640650 *T)
    sumY3  +=       0.785 * cos(4.39719518884  + 2957.95971196010 *T)
    sumY3  +=       0.624 * cos(2.25363881101  + 23384.53080438210 *T)
    sumY3  +=       0.586 * cos(3.96432388242  + 1059.62574767270 *T)
    sumY3  +=       0.721 * cos(5.21679562282  + 4535.30325440790 *T)
    sumY3  +=       0.636 * cos(2.60828365653  + 5088.87265725030 *T)
    sumY3  +=       0.615 * cos(0.76601087488  + 2146.40923395870 *T)
    sumY3  +=       0.474 * cos(3.79995013670  + 2942.70724077510 *T)
    sumY3  +=       0.604 * cos(3.34001877693  + 1592.83983111630 *T)
    sumY3  +=       0.450 * cos(0.32237574079  + 3333.74269718250 *T)
    sumY3  +=       0.486 * cos(1.35165895173  + 3739.00524759150 *T)
    sumY3  +=       0.417 * cos(5.75416173350  + 3347.96979118411 *T)
    sumY3  +=       0.400 * cos(5.17092240158  + 4400.23817437250 *T)
    sumY3  +=       0.374 * cos(1.81231274497  + 6836.88907031730 *T)
    sumY3  +=       0.488 * cos(3.09412842261  + 6924.19727485710 *T)
    sumY3  +=       0.399 * cos(0.86698142534  + 6490.02040477150 *T)
    sumY3  +=       0.356 * cos(5.90189075488  + 6127.89926804070 *T)
    sumY3  +=       0.362 * cos(0.57575435741  + 5622.08674069391 *T)
    sumY3  +=       0.295 * cos(2.11031657729  + 7079.61767429131 *T)
    sumY3  +=       0.356 * cos(4.85741827730  + 9866.66069814871 *T)
    sumY3  +=       0.340 * cos(0.71669809587  + 7477.76667769950 *T)
    sumY3  +=       0.239 * cos(3.60818728854  + 6438.74006690910 *T)
    sumY3  +=       0.214 * cos(5.66889246190  + 6682.44899195131 *T)
    sumY3  +=       0.184 * cos(5.24213240799  + 13362.20588931570 *T)
    sumY3  +=       0.176 * cos(2.39234978630  + 5331.60126122430 *T)
    sumY3  +=       0.215 * cos(1.01330961394  + 5729.75026463250 *T)
    sumY3  +=       0.159 * cos(3.74822824989  + 4933.45225781610 *T)
    sumY3  +=       0.148 * cos(4.78372101689  + 6298.57213865991 *T)
    sumY3  +=       0.154 * cos(4.07120648646  + 6677.94555253410 *T)
    sumY3  +=       0.153 * cos(5.60790410015  + 7875.91568110771 *T)
    sumY3  +=       0.157 * cos(4.45933297186  + 796.54182429990 *T)
    sumY3  +=       0.172 * cos(2.61819596070  + 13366.21664263171 *T)
    sumY3  +=       0.143 * cos(0.82498614720  + 7235.03807372550 *T)
    sumY3  +=       0.155 * cos(5.68177791947  + 7740.85060107230 *T)
    sumY3  +=       0.116 * cos(3.07839128709  + 1194.69082770810 *T)
    sumY3  +=       0.106 * cos(1.43807829522  + 155.66421691770 *T)
    sumY3  +=       0.089 * cos(5.59709996368  + 3343.89172756530 *T)
    sumY3  +=       0.100 * cos(1.30610304994  + 9830.63283147130 *T)
    sumY3  +=       0.108 * cos(2.60638791189  + 10420.23010099111 *T)
    sumY3  +=       0.084 * cos(1.11366948198  + 10818.37910439930 *T)
    sumY3  +=       0.098 * cos(2.60446363513  + 26725.14323108190 *T)
    sumY3  +=       0.076 * cos(5.59159654622  + 4929.92913946711 *T)
    sumY3  +=       0.074 * cos(5.71538520874  + 7633.18707713370 *T)
    sumY3  +=       0.086 * cos(3.58543209453  + 1589.31671276730 *T)
    sumY3  +=       0.072 * cos(4.63326318414  + 5885.17066406670 *T)
    sumY3  +=       0.062 * cos(0.11881168827  + 5487.02166065850 *T)
    sumY3  +=       0.065 * cos(3.44980702826  + 10264.80970155691 *T)
    sumY3  +=       0.062 * cos(1.52309575501  + 7211.15963597770 *T)
    sumY3  +=       0.082 * cos(6.15563051504  + 6688.58221788390 *T)
    sumY3  +=       0.057 * cos(3.09476278894  + 12833.00255918811 *T)
    sumY3  +=       0.057 * cos(6.25611563425  + 9468.51169474051 *T)
    sumY3  +=       0.056 * cos(3.56059573252  + 6681.45141723091 *T)
    sumY3  +=       0.056 * cos(5.59136715053  + 6681.48592453530 *T)
    sumY3  +=       0.069 * cos(4.15252094724  + 796.05418933290 *T)
    sumY3  +=       0.065 * cos(0.34610968118  + 3870.54720927791 *T)
    sumY3  +=       0.067 * cos(2.17750046913  + 9492.39013248830 *T)
    sumY3  +=       0.051 * cos(4.21953229457  + 8274.06468451590 *T)
    sumY3  +=       0.048 * cos(5.45844802909  + 397.90518592470 *T)
    sumY3  +=       0.050 * cos(5.21362719324  + 13207.27312484851 *T)
    sumY3  +=       0.043 * cos(3.92075403663  + 426.84200835950 *T)
    sumY3  +=       0.044 * cos(4.30828712757  + 8031.33608054190 *T)
    sumY3  +=       0.039 * cos(2.78024704625  + 1194.20319274110 *T)
    sumY3  +=       0.043 * cos(6.06948054666  + 10023.06141865111 *T)
    sumY3  +=       0.044 * cos(5.65948692732  + 5092.39577559930 *T)
    sumY3  +=       0.037 * cos(4.89772055096  + 8827.63408735830 *T)
    sumY3  +=       0.042 * cos(0.13693241822  + 10022.09835123510 *T)
    sumY3  +=       0.043 * cos(1.88027614352  + 3184.94820978210 *T)
    sumY3  +=       0.033 * cos(3.97855919031  + 1349.62359217529 *T)
    sumY3  +=       0.035 * cos(1.38364575820  + 9070.36269133230 *T)
    sumY3  +=       0.036 * cos(6.07274445742  + 11081.46302777211 *T)
    sumY3  +=       0.044 * cos(4.83413643654  + 12303.31159409350 *T)
    sumY3  +=       0.033 * cos(5.60078768464  + 15643.92402079330 *T)
    sumY3  +=       0.042 * cos(4.38948412695  + 10022.06384393071 *T)
    sumY3  +=       0.032 * cos(4.09360307102  + 10018.55797923390 *T)
    sumY3  +=       0.029 * cos(3.12031392752  + 3495.78900865049 *T)
    sumY3  +=       0.032 * cos(0.83223666105  + 6680.48834981490 *T)
    sumY3  +=       0.031 * cos(0.33659543405  + 7.35736448430 *T)
    sumY3  +=       0.036 * cos(1.89292951517  + 7064.36520310630 *T)
    sumY3  +=       0.026 * cos(5.98285354447  + 11216.52810780750 *T)
    sumY3  +=       0.035 * cos(6.27114625059  + 3336.84549086730 *T)
    sumY3  +=       0.025 * cos(1.46525178422  + 10575.65050042530 *T)
    sumY3  +=       0.033 * cos(2.79365879085  + 8672.21368792411 *T)
    sumY3  +=       0.026 * cos(1.28604268140  + 1592.35219614930 *T)
    sumY3  +=       0.027 * cos(5.55280390266  + 16702.81831601549 *T)
    sumY3  +=       0.025 * cos(3.00868456049  + 13760.84252769090 *T)
    sumY3  +=       0.027 * cos(4.66856303516  + 191.20444862810 *T)
    sumY3  +=       0.026 * cos(1.97438831111  + 6894.76776632111 *T)
    sumY3  +=       0.023 * cos(4.28067595534  + 2281.47431399410 *T)
    sumY3  +=       0.027 * cos(2.94695290371  + 16706.82906933150 *T)
    sumY3  +=       0.026 * cos(3.96428963673  + 8433.00820229911 *T)
    sumY3  +=       0.028 * cos(1.79193447711  + 9623.93209417470 *T)
    sumY3  +=       0.027 * cos(1.13776019664  + 2145.92159899169 *T)
    sumY3  +=       0.024 * cos(5.97021720723  + 8270.54156616690 *T)
    sumY3  +=       0.026 * cos(2.65229643774  + 1747.77259558350 *T)
    sumY3  +=       0.022 * cos(5.61692474893  + 3767.45443505930 *T)
    sumY3  +=       0.025 * cos(2.25020572476  + 2811.16527908870 *T)
    sumY3  +=       0.020 * cos(0.07098508189  + 3554.15533962130 *T)
    sumY3  +=       0.019 * cos(2.88996373259  + 3127.55714874530 *T)
    sumY3  +=       0.022 * cos(0.46069753881  + 553.32558535889 *T)
    sumY3  +=       0.023 * cos(5.33595219972  + 951.47458876710 *T)
    sumY3  +=       0.019 * cos(0.10196223786  + 553.81322032590 *T)
    sumY3  +=       0.026 * cos(4.87516844857  + 8962.69916739370 *T)
    sumY3  +=       0.020 * cos(5.14250346119  + 9639.18456535971 *T)
    sumY3  +=       0.019 * cos(6.00158821783  + 2810.67764412170 *T)
    sumY3  +=       0.021 * cos(4.93189731089  + 951.96222373410 *T)
    sumY3  +=       0.017 * cos(1.68247778914  + 13171.24525817110 *T)
    sumY3  +=       0.022 * cos(5.95572380889  + 3738.51761262449 *T)
    sumY3  +=       0.019 * cos(1.64647085293  + 6681.40136058031 *T)
    sumY3  +=       0.022 * cos(0.08200412214  + 10973.79950383350 *T)
    sumY3  +=       0.016 * cos(4.51745505904  + 4136.66661603269 *T)
    sumY3  +=       0.019 * cos(1.23797436286  + 6681.53598118590 *T)
    sumY3  +=       0.021 * cos(1.35776125147  + 1748.26023055050 *T)
    sumY3  +=       0.017 * cos(1.00152701220  + 382.65271473970 *T)
    sumY3  +=       0.020 * cos(1.75396056414  + 2280.98667902709 *T)
    sumY3  +=       0.022 * cos(3.17699744058  + 1350.11122714230 *T)
    sumY3  +=       0.016 * cos(5.34701571085  + 6254.87048000710 *T)
    sumY3  +=       0.016 * cos(2.36582548750  + 10551.77206267750 *T)
    sumY3  +=       0.015 * cos(1.53277475796  + 10177.50149701710 *T)
    sumY3  +=       0.014 * cos(3.52302167993  + 2914.25805330730 *T)
    sumY3  +=       0.014 * cos(6.07892689352  + 5459.62010456170 *T)
    sumY3  +=       0.019 * cos(6.17607723197  + 1990.50119955750 *T)
    sumY3  +=       0.014 * cos(0.25864758853  + 10029.19464458370 *T)
    sumY3  +=       0.013 * cos(2.47039869218  + 5614.97319369310 *T)
    sumY3  +=       0.015 * cos(1.09462751364  + 6674.35512388230 *T)
    sumY3  +=       0.013 * cos(6.22971843612  + 1066.73929467350 *T)
    sumY3  +=       0.016 * cos(2.66082127133  + 640.14110379750 *T)
    sumY3  +=       0.013 * cos(5.23063056874  + 5621.59910572690 *T)
    sumY3  +=       0.013 * cos(6.19082803643  + 242.97242145751 *T)
    sumY3  +=       0.014 * cos(4.95970323400  + 12310.42514109431 *T)
    sumY3  +=       0.013 * cos(5.13330302990  + 191.69208359510 *T)
    sumY3  +=       0.011 * cos(4.76410993683  + 2388.65020296570 *T)
    sumY3  +=       0.012 * cos(4.69348384132  + 3148.92034310470 *T)
    sumY3  +=       0.013 * cos(3.86741890756  + 2119.00767786191 *T)
    sumY3  +=       0.011 * cos(0.48497464502  + 3583.09721319030 *T)
    sumY3  +=       0.011 * cos(3.56912327150  + 3340.38586286849 *T)
    sumY3  +=       0.010 * cos(3.06794437904  + 4534.81561944089 *T)
    sumY3  +=       0.011 * cos(1.53890583251  + 3340.35135556410 *T)
    sumY3  +=       0.010 * cos(3.31773666314  + 8429.48508395010 *T)
    sumY3  +=       0.010 * cos(1.74465995773  + 10014.96755058210 *T)
    sumY3  +=       0.009 * cos(2.31043926403  + 3320.50092478451 *T)
    sumY3  +=       0.011 * cos(6.23117374482  + 8962.21153242670 *T)
    sumY3  +=       0.009 * cos(0.67622157817  + 3361.21156358210 *T)
    sumY3  +=       0.009 * cos(2.53457737250  + 12296.19804709270 *T)
    sumY3  +=       0.010 * cos(3.49125725908  + 16173.61498588790 *T)
    sumY3  +=       0.009 * cos(0.45307177363  + 1052.51220067191 *T)
    sumY3  +=       0.009 * cos(5.25143266877  + 5614.48555872609 *T)
    sumY3  +=       0.010 * cos(0.42320665245  + 529.93478257810 *T)
    sumY3  +=       0.008 * cos(1.34588517648  + 3893.93801205869 *T)
    sumY3  +=       0.009 * cos(4.25124379767  + 3325.60377299830 *T)
    sumY3  +=       0.008 * cos(3.39718396453  + 2786.79920637390 *T)
    sumY3  +=       0.009 * cos(5.01886213751  + 3356.10871536831 *T)
    sumY3  +=       0.011 * cos(0.14412148132  + 242.48478649050 *T)
    sumY3  +=       0.009 * cos(6.13794980664  + 1751.78334889950 *T)
    sumY3  +=       0.007 * cos(1.88989832363  + 2957.47207699310 *T)
    sumY3  +=       0.008 * cos(6.15723205798  + 4292.08701546690 *T)
    sumY3  +=       0.007 * cos(1.45796130115  + 14158.99153109910 *T)
    sumY3  +=       0.010 * cos(5.32959721544  + 529.44714761109 *T)
    sumY3  +=       0.010 * cos(0.03923636410  + 12302.82395912650 *T)
    sumY3  +=       0.009 * cos(5.97902157224  + 6684.50415426510 *T)
    sumY3  +=       0.008 * cos(0.50370755190  + 2273.87313202630 *T)

    # MARS Y Coord
    sumY4  = 0
    sumY4  +=     174.339 * cos(2.93381852792  + 3340.85624418330 *T)
    sumY4  +=      65.332 * cos(6.16516435716  + 0.24381748350 *T)
    sumY4  +=      21.612 * cos(3.90497783530  + 6681.46867088311 *T)
    sumY4  +=      12.491 * cos(5.22926339778  + 10022.08109758290 *T)
    sumY4  +=       5.152 * cos(5.82623510484  + 13362.69352428270 *T)
    sumY4  +=       2.432 * cos(3.46296610123  + 3496.27664361750 *T)
    sumY4  +=       2.407 * cos(2.61988971710  + 3185.43584474911 *T)
    sumY4  +=       1.485 * cos(6.26563422445  + 16703.30595098250 *T)
    sumY4  +=       1.453 * cos(2.74672685906  + 3340.36860921629 *T)
    sumY4  +=       0.592 * cos(1.22137870027  + 3583.58484815730 *T)
    sumY4  +=       0.596 * cos(4.86866829338  + 3098.12764020930 *T)
    sumY4  +=       0.649 * cos(3.10993649847  + 6680.98103591609 *T)
    sumY4  +=       0.380 * cos(2.98520597318  + 6526.04827144891 *T)
    sumY4  +=       0.351 * cos(0.37438431534  + 20043.91837768230 *T)
    sumY4  +=       0.281 * cos(5.74023871163  + 3344.37936253230 *T)
    sumY4  +=       0.275 * cos(5.82358825194  + 3337.33312583431 *T)
    sumY4  +=       0.287 * cos(2.04842024047  + 3894.42564702570 *T)
    sumY4  +=       0.285 * cos(4.01760045509  + 2787.28684134090 *T)
    sumY4  +=       0.151 * cos(0.21831233003  + 6684.99178923210 *T)
    sumY4  +=       0.128 * cos(3.67710811233  + 6836.88907031730 *T)
    sumY4  +=       0.111 * cos(3.42090983567  + 10021.59346261590 *T)
    sumY4  +=       0.088 * cos(1.58466318522  + 6924.19727485710 *T)
    sumY4  +=       0.071 * cos(0.74300955382  + 23384.53080438210 *T)
    sumY4  +=       0.055 * cos(0.69772957456  + 10025.60421593190 *T)
    sumY4  +=       0.048 * cos(4.38512823437  + 6127.89926804070 *T)
    sumY4  +=       0.063 * cos(3.34688291780  + 9866.66069814871 *T)
    sumY4  +=       0.034 * cos(5.06262173016  + 6438.74006690910 *T)
    sumY4  +=       0.033 * cos(1.70059643142  + 1059.62574767270 *T)
    sumY4  +=       0.019 * cos(3.71883012543  + 13362.20588931570 *T)
    sumY4  +=       0.018 * cos(5.50021563801  + 6490.02040477150 *T)
    sumY4  +=       0.016 * cos(3.64198835610  + 3149.40797807170 *T)
    sumY4  +=       0.016 * cos(4.24409660230  + 2942.70724077510 *T)
    sumY4  +=       0.015 * cos(5.58219475947  + 3532.30451029490 *T)
    sumY4  +=       0.015 * cos(5.77023095211  + 2544.55823736690 *T)
    sumY4  +=       0.014 * cos(1.18570988628  + 5622.08674069391 *T)
    sumY4  +=       0.014 * cos(4.91332403738  + 3739.00524759150 *T)
    sumY4  +=       0.013 * cos(3.26498430793  + 4400.23817437250 *T)
    sumY4  +=       0.013 * cos(3.45392588632  + 4137.15425099970 *T)
    sumY4  +=       0.012 * cos(1.58704936612  + 1194.69082770810 *T)
    sumY4  +=       0.012 * cos(4.09135686171  + 6682.44899195131 *T)
    sumY4  +=       0.015 * cos(5.36555803467  + 7477.76667769950 *T)
    sumY4  +=       0.011 * cos(0.34983006468  + 7079.61767429131 *T)
    sumY4  +=       0.015 * cos(3.69953477421  + 529.93478257810 *T)
    sumY4  +=       0.013 * cos(4.15046759496  + 10177.50149701710 *T)
    sumY4  +=       0.014 * cos(2.27580303219  + 7235.03807372550 *T)
    sumY4  +=       0.011 * cos(1.94853565727  + 10264.80970155691 *T)
    sumY4  +=       0.013 * cos(1.09206345437  + 13366.21664263171 *T)
    sumY4  +=       0.010 * cos(2.95515329817  + 796.54182429990 *T)
    sumY4  +=       0.010 * cos(0.79995918783  + 5331.60126122430 *T)
    sumY4  +=       0.012 * cos(5.75066247454  + 5729.75026463250 *T)
    sumY4  +=       0.011 * cos(1.07729854427  + 26725.14323108190 *T)
    sumY4  +=       0.008 * cos(3.24165666451  + 6298.57213865991 *T)
    sumY4  +=       0.008 * cos(2.38038667642  + 3340.87349783550 *T)
    sumY4  +=       0.008 * cos(0.34961504493  + 3340.83899053111 *T)
    sumY4  +=       0.007 * cos(5.14772021296  + 1990.98883452450 *T)
    sumY4  +=       0.009 * cos(5.07995147120  + 2280.98667902709 *T)
    sumY4  +=       0.007 * cos(4.03741515609  + 7875.91568110771 *T)
    sumY4  +=       0.008 * cos(2.02200735771  + 4933.45225781610 *T)
    sumY4  +=       0.007 * cos(3.75129319164  + 2389.13783793270 *T)
    sumY4  +=       0.009 * cos(0.25417853710  + 1592.83983111630 *T)
    sumY4  +=       0.007 * cos(0.82767073807  + 3339.87592311510 *T)
    sumY4  +=       0.007 * cos(2.14189652179  + 3341.83656525151 *T)
    sumY4  +=       0.007 * cos(4.15524185925  + 3343.89172756530 *T)
    sumY4  +=       0.006 * cos(5.26975113166  + 6151.77770578850 *T)
    sumY4  +=       0.007 * cos(0.39533433606  + 3184.94820978210 *T)
    sumY4  +=       0.007 * cos(2.62352530661  + 155.17658195069 *T)
    sumY4  +=       0.006 * cos(2.08770398228  + 1589.31671276730 *T)
    sumY4  +=       0.005 * cos(1.31171511465  + 5487.02166065850 *T)
    sumY4  +=       0.005 * cos(6.18023058032  + 5885.17066406670 *T)
    sumY4  +=       0.005 * cos(5.51400581371  + 4292.57465043390 *T)
    sumY4  +=       0.006 * cos(0.88154703323  + 2146.40923395870 *T)
    sumY4  +=       0.005 * cos(4.11249699702  + 4690.72365384210 *T)
    sumY4  +=       0.005 * cos(1.23189383847  + 2957.95971196010 *T)
    sumY4  +=       0.004 * cos(3.71246158922  + 13207.27312484851 *T)
    sumY4  +=       0.005 * cos(2.07014421381  + 4535.30325440790 *T)
    sumY4  +=       0.004 * cos(1.74392645390  + 3723.75277640650 *T)
    sumY4  +=       0.005 * cos(2.71222704178  + 5088.87265725030 *T)
    sumY4  +=       0.004 * cos(4.76678794761  + 9468.51169474051 *T)
    sumY4  +=       0.004 * cos(4.22334394857  + 398.39282089170 *T)
    sumY4  +=       0.003 * cos(4.44413134438  + 3870.54720927791 *T)
    sumY4  +=       0.003 * cos(4.24190972006  + 7633.18707713370 *T)

    # MARS Y Coord
    sumY5  = 0
    sumY5  +=       2.822 * cos(0.96657189081  + 0.24381748350 *T)
    sumY5  +=       1.310 * cos(3.05377797611  + 3340.85624418330 *T)
    sumY5  +=       0.954 * cos(1.98829739864  + 6681.46867088311 *T)
    sumY5  +=       0.429 * cos(3.31617723817  + 10022.08109758290 *T)
    sumY5  +=       0.275 * cos(4.95749653129  + 3496.27664361750 *T)
    sumY5  +=       0.272 * cos(1.09357865500  + 3185.43584474911 *T)
    sumY5  +=       0.241 * cos(4.15990840451  + 13362.69352428270 *T)
    sumY5  +=       0.262 * cos(1.53308484382  + 3340.36860921629 *T)
    sumY5  +=       0.092 * cos(4.67897944940  + 16703.30595098250 *T)
    sumY5  +=       0.076 * cos(5.98548328683  + 3583.58484815730 *T)
    sumY5  +=       0.077 * cos(0.08639182463  + 3098.12764020930 *T)
    sumY5  +=       0.049 * cos(1.46758792224  + 6526.04827144891 *T)
    sumY5  +=       0.032 * cos(1.68955173389  + 6680.98103591609 *T)
    sumY5  +=       0.026 * cos(5.10726173875  + 20043.91837768230 *T)
    sumY5  +=       0.027 * cos(3.55496481176  + 3894.42564702570 *T)
    sumY5  +=       0.026 * cos(2.47271477475  + 2787.28684134090 *T)
    sumY5  +=       0.013 * cos(5.14425957427  + 6836.88907031730 *T)
    sumY5  +=       0.013 * cos(0.06925191240  + 6924.19727485710 *T)
    sumY5  +=       0.006 * cos(1.93462610905  + 10021.59346261590 *T)
    sumY5  +=       0.007 * cos(6.11562143175  + 3337.33312583431 *T)
    sumY5  +=       0.007 * cos(3.04667075048  + 3344.37936253230 *T)
    sumY5  +=       0.005 * cos(2.85975463034  + 6127.89926804070 *T)
    sumY5  +=       0.007 * cos(1.82394907196  + 9866.66069814871 *T)
    sumY5  +=       0.006 * cos(4.75654379617  + 6684.99178923210 *T)
    sumY5  +=       0.005 * cos(5.49442547736  + 23384.53080438210 *T)
    sumY5  +=       0.004 * cos(0.28345481344  + 6438.74006690910 *T)

    # MARS Z Coord
    sumZ0  = 0
    sumZ0  +=  4901207.220 * cos(3.76712324286  + 3340.61242669980 *T)
    sumZ0  +=  660669.541 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ0  +=  228333.904 * cos(4.10544022266  + 6681.22485339960 *T)
    sumZ0  +=   15958.402 * cos(4.44367058261  + 10021.83728009940 *T)
    sumZ0  +=    1321.976 * cos(4.78186604114  + 13362.44970679920 *T)
    sumZ0  +=     679.660 * cos(5.65109977813  + 3337.08930835080 *T)
    sumZ0  +=     679.219 * cos(5.02527030899  + 3344.13554504880 *T)
    sumZ0  +=     531.140 * cos(3.86748390045  + 2281.23049651060 *T)
    sumZ0  +=     374.993 * cos(4.33338216773  + 529.69096509460 *T)
    sumZ0  +=     325.315 * cos(2.24562508217  + 1059.38193018920 *T)
    sumZ0  +=     218.762 * cos(1.18094849702  + 3340.59517304760 *T)
    sumZ0  +=     218.761 * cos(3.21172012852  + 3340.62968035200 *T)
    sumZ0  +=     152.333 * cos(6.08109566130  + 6151.53388830500 *T)
    sumZ0  +=     146.631 * cos(5.09927022855  + 398.14900340820 *T)
    sumZ0  +=     152.443 * cos(2.13254535850  + 5621.84292321040 *T)
    sumZ0  +=     120.328 * cos(5.12010663983  + 16703.06213349900 *T)
    sumZ0  +=     126.016 * cos(3.67803475473  + 796.29800681640 *T)
    sumZ0  +=     116.766 * cos(4.86913583010  + 2942.46342329160 *T)
    sumZ0  +=     112.925 * cos(5.76548397681  + 3738.76143010800 *T)
    sumZ0  +=      74.290 * cos(2.43508234938  + 8962.45534991020 *T)
    sumZ0  +=      71.989 * cos(5.54217679275  + 3340.54511639700 *T)
    sumZ0  +=      68.644 * cos(5.54532471548  + 6283.07584999140 *T)
    sumZ0  +=      71.987 * cos(5.13368040597  + 3340.67973700260 *T)
    sumZ0  +=      63.219 * cos(5.36590967098  + 6684.74797174860 *T)
    sumZ0  +=      75.858 * cos(4.51645136094  + 3149.16416058820 *T)
    sumZ0  +=      63.502 * cos(4.16748385740  + 1751.53953141600 *T)
    sumZ0  +=      63.828 * cos(5.98786823499  + 6677.70173505060 *T)
    sumZ0  +=      59.355 * cos(0.72879549961  + 5884.92684658320 *T)
    sumZ0  +=      50.332 * cos(1.96033823263  + 2810.92146160520 *T)
    sumZ0  +=      51.292 * cos(2.64769056433  + 2914.01423582380 *T)
    sumZ0  +=      44.615 * cos(5.88956380171  + 3532.06069281140 *T)
    sumZ0  +=      45.001 * cos(2.29942389620  + 1589.07289528380 *T)
    sumZ0  +=      45.526 * cos(2.27415776311  + 1194.44701022460 *T)
    sumZ0  +=      45.098 * cos(4.08200115074  + 4136.91043351620 *T)
    sumZ0  +=      33.148 * cos(2.01609063016  + 5486.77784317500 *T)
    sumZ0  +=      29.003 * cos(2.64937555759  + 426.59819087600 *T)
    sumZ0  +=      33.968 * cos(0.46851841759  + 2544.31441988340 *T)
    sumZ0  +=      27.697 * cos(5.54319865094  + 191.44826611160 *T)
    sumZ0  +=      30.473 * cos(4.98782051819  + 3870.30339179440 *T)
    sumZ0  +=      22.262 * cos(0.19101027284  + 9492.14631500480 *T)
    sumZ0  +=      21.566 * cos(2.45059201892  + 3333.49887969900 *T)
    sumZ0  +=      19.900 * cos(4.78493597626  + 213.29909543800 *T)
    sumZ0  +=      19.445 * cos(3.36733718353  + 5088.62883976680 *T)
    sumZ0  +=      22.415 * cos(2.80792353370  + 3496.03282613400 *T)
    sumZ0  +=      19.109 * cos(0.86825937304  + 1592.59601363280 *T)
    sumZ0  +=      24.101 * cos(2.57371991774  + 4399.99435688900 *T)
    sumZ0  +=      20.382 * cos(3.55004736979  + 6681.24210705180 *T)
    sumZ0  +=      19.188 * cos(1.51640815942  + 3185.19202726560 *T)
    sumZ0  +=      20.382 * cos(1.51927573830  + 6681.20759974740 *T)
    sumZ0  +=      13.488 * cos(2.86475891587  + 3341.59274776800 *T)
    sumZ0  +=      13.492 * cos(1.52791184830  + 3339.63210563160 *T)
    sumZ0  +=      11.777 * cos(4.79772279974  + 3553.91152213780 *T)
    sumZ0  +=      12.741 * cos(6.05559229749  + 7079.37385680780 *T)
    sumZ0  +=      11.907 * cos(1.65033513222  + 6254.62666252360 *T)
    sumZ0  +=      11.511 * cos(2.75835305381  + 12303.06777661000 *T)
    sumZ0  +=      11.629 * cos(5.45832237690  + 20043.67456019880 *T)
    sumZ0  +=      11.780 * cos(4.75307421274  + 4690.47983635860 *T)
    sumZ0  +=      10.587 * cos(6.14125670809  + 7.11354700080 *T)
    sumZ0  +=       8.746 * cos(1.07533372055  + 9225.53927328300 *T)
    sumZ0  +=       9.300 * cos(5.74426412364  + 1990.74501704100 *T)
    sumZ0  +=       9.010 * cos(5.91031650213  + 9623.68827669120 *T)
    sumZ0  +=       9.099 * cos(4.14734627288  + 3127.31333126180 *T)
    sumZ0  +=       8.399 * cos(2.65942772730  + 4535.05943692440 *T)
    sumZ0  +=       9.575 * cos(0.26669816734  + 10213.28554621100 *T)
    sumZ0  +=      11.208 * cos(2.64763421097  + 8432.76438481560 *T)
    sumZ0  +=       8.425 * cos(5.67806346712  + 6467.92575796160 *T)
    sumZ0  +=       7.879 * cos(6.16301318418  + 4292.33083295040 *T)
    sumZ0  +=       6.986 * cos(1.31599425886  + 3894.18182954220 *T)
    sumZ0  +=       7.013 * cos(4.83160093113  + 6489.77658728800 *T)
    sumZ0  +=       6.707 * cos(5.47201048198  + 6681.29216370240 *T)
    sumZ0  +=       8.732 * cos(5.94456391651  + 5092.15195811580 *T)
    sumZ0  +=       6.427 * cos(6.10941025846  + 4562.46099302120 *T)
    sumZ0  +=       8.234 * cos(3.68630710748  + 2700.71514038580 *T)
    sumZ0  +=       7.214 * cos(1.11647593098  + 3319.83703120740 *T)
    sumZ0  +=       6.128 * cos(4.51741845562  + 1221.84856632140 *T)
    sumZ0  +=       6.627 * cos(5.70488581385  + 10025.36039844840 *T)
    sumZ0  +=       6.712 * cos(0.04223488375  + 10018.31416175040 *T)
    sumZ0  +=       7.157 * cos(3.27593626835  + 3361.38782219220 *T)
    sumZ0  +=       5.667 * cos(1.84572324920  + 2957.71589447660 *T)
    sumZ0  +=       6.707 * cos(5.88050686882  + 6681.15754309680 *T)
    sumZ0  +=       4.916 * cos(2.93046960145  + 2787.04302385740 *T)
    sumZ0  +=       4.722 * cos(2.41041908083  + 3344.20285535160 *T)
    sumZ0  +=       4.650 * cos(1.71276888306  + 639.89728631400 *T)
    sumZ0  +=       4.742 * cos(1.97920373919  + 3337.02199804800 *T)
    sumZ0  +=       4.597 * cos(4.41002877977  + 7477.52286021600 *T)
    sumZ0  +=       5.433 * cos(4.33760555117  + 2388.89402044920 *T)
    sumZ0  +=       5.641 * cos(2.43527331240  + 2118.76386037840 *T)
    sumZ0  +=       4.363 * cos(2.24320799714  + 11773.37681151540 *T)
    sumZ0  +=       4.127 * cos(2.04411881734  + 2146.16541647520 *T)
    sumZ0  +=       3.613 * cos(3.61081984797  + 8429.24126646660 *T)
    sumZ0  +=       3.478 * cos(2.64660968975  + 4929.68532198360 *T)
    sumZ0  +=       3.714 * cos(1.76418965622  + 382.89653222320 *T)
    sumZ0  +=       3.331 * cos(6.17753431621  + 7210.91581849420 *T)
    sumZ0  +=       3.580 * cos(3.90470994536  + 7903.07341972100 *T)
    sumZ0  +=       3.111 * cos(2.03406809335  + 155.42039943420 *T)
    sumZ0  +=       3.351 * cos(0.47359753009  + 3475.67750673520 *T)
    sumZ0  +=       3.025 * cos(1.91083270977  + 9595.23908922340 *T)
    sumZ0  +=       3.493 * cos(2.25666784516  + 3723.50895892300 *T)
    sumZ0  +=       3.509 * cos(3.91758478184  + 3205.54734666440 *T)
    sumZ0  +=       3.007 * cos(0.81247341820  + 2274.11694950980 *T)
    sumZ0  +=       2.982 * cos(3.68257004005  + 2288.34404351140 *T)
    sumZ0  +=       2.567 * cos(4.96092533051  + 8031.09226305840 *T)
    sumZ0  +=       2.649 * cos(2.98742483865  + 3767.21061757580 *T)
    sumZ0  +=       2.322 * cos(1.23249592342  + 4933.20844033260 *T)
    sumZ0  +=       2.733 * cos(0.51395284395  + 12832.75874170460 *T)
    sumZ0  +=       2.136 * cos(3.88842223350  + 10021.85453375160 *T)
    sumZ0  +=       2.136 * cos(1.85765060200  + 10021.82002644720 *T)
    sumZ0  +=       2.772 * cos(2.43968968772  + 8827.39026987480 *T)
    sumZ0  +=       2.162 * cos(3.01828376513  + 6041.32756708560 *T)
    sumZ0  +=       2.091 * cos(3.80489059877  + 6872.67311951120 *T)
    sumZ0  +=       2.233 * cos(2.94758000961  + 7740.60678358880 *T)
    sumZ0  +=       2.614 * cos(4.45694073527  + 6674.11130639880 *T)
    sumZ0  +=       1.930 * cos(5.30759100687  + 1066.49547719000 *T)
    sumZ0  +=       2.513 * cos(0.10458597053  + 3583.34103067380 *T)
    sumZ0  +=       2.104 * cos(2.45809927011  + 1052.26838318840 *T)
    sumZ0  +=       1.861 * cos(4.13522134185  + 1349.86740965880 *T)
    sumZ0  +=       1.728 * cos(2.95584201533  + 12168.00269657460 *T)
    sumZ0  +=       1.840 * cos(0.20978853293  + 4032.77002792660 *T)
    sumZ0  +=       1.669 * cos(5.50497155791  + 951.71840625060 *T)
    sumZ0  +=       1.911 * cos(2.94107630035  + 1748.01641306700 *T)
    sumZ0  +=       1.442 * cos(1.87731237490  + 3077.52850332700 *T)
    sumZ0  +=       1.589 * cos(1.73601297200  + 3369.06161416760 *T)
    sumZ0  +=       1.379 * cos(0.09013023527  + 10419.98628350760 *T)
    sumZ0  +=       1.392 * cos(0.27504136647  + 3325.35995551480 *T)
    sumZ0  +=       1.580 * cos(4.48665176300  + 3.52311834900 *T)
    sumZ0  +=       1.766 * cos(1.85661467458  + 6525.80445396540 *T)
    sumZ0  +=       1.292 * cos(2.51829136056  + 3603.69635007260 *T)
    sumZ0  +=       1.297 * cos(5.48481699939  + 11371.70468975820 *T)
    sumZ0  +=       1.613 * cos(0.62341029242  + 553.56940284240 *T)
    sumZ0  +=       1.384 * cos(4.11719633248  + 3355.86489788480 *T)
    sumZ0  +=       1.261 * cos(1.86730474733  + 6680.24453233140 *T)
    sumZ0  +=       1.610 * cos(2.65671079949  + 3312.16323923200 *T)
    sumZ0  +=       1.274 * cos(0.55350948232  + 3191.04922956520 *T)
    sumZ0  +=       1.704 * cos(4.50757815133  + 3097.88382272579 *T)
    sumZ0  +=       1.460 * cos(0.04101454115  + 7632.94325965020 *T)
    sumZ0  +=       1.242 * cos(5.75816506092  + 149.56319713460 *T)
    sumZ0  +=       1.230 * cos(1.10948154340  + 536.80451209540 *T)
    sumZ0  +=       1.216 * cos(2.98306649362  + 6836.64525283380 *T)
    sumZ0  +=       1.161 * cos(4.60505013406  + 1648.44675719740 *T)
    sumZ0  +=       1.579 * cos(3.09053337023  + 15643.68020330980 *T)
    sumZ0  +=       1.118 * cos(6.05429252683  + 9808.53818466140 *T)
    sumZ0  +=       1.376 * cos(0.81116680514  + 3503.07906283200 *T)
    sumZ0  +=       1.101 * cos(4.62263227757  + 522.57741809380 *T)
    sumZ0  +=       1.320 * cos(3.23812050004  + 3347.72597370060 *T)
    sumZ0  +=       1.419 * cos(0.32932000507  + 3264.34635542420 *T)
    sumZ0  +=       1.103 * cos(5.79565205719  + 6894.52394883760 *T)
    sumZ0  +=       1.257 * cos(3.20326473314  + 6682.20517446780 *T)
    sumZ0  +=       0.968 * cos(4.43650051215  + 103.09277421860 *T)
    sumZ0  +=       1.004 * cos(4.71380449727  + 2487.41604494780 *T)
    sumZ0  +=       1.014 * cos(0.55883970461  + 10973.55568635000 *T)
    sumZ0  +=       0.950 * cos(2.98201757904  + 3178.14579056760 *T)
    sumZ0  +=       0.958 * cos(1.45300862079  + 12566.15169998280 *T)
    sumZ0  +=       0.964 * cos(5.98684645223  + 242.72860397400 *T)
    sumZ0  +=       0.925 * cos(5.46131905325  + 5614.72937620960 *T)
    sumZ0  +=       0.935 * cos(4.94933845312  + 1758.65307841680 *T)
    sumZ0  +=       1.171 * cos(5.79646129252  + 23384.28698689860 *T)
    sumZ0  +=       0.969 * cos(2.65689194156  + 2648.45482547300 *T)
    sumZ0  +=       0.853 * cos(4.81086337808  + 15110.46611986620 *T)
    sumZ0  +=       0.901 * cos(5.67928823390  + 2921.12778282460 *T)
    sumZ0  +=       1.125 * cos(6.27341969245  + 12964.30070339100 *T)
    sumZ0  +=       0.844 * cos(2.97906384874  + 7875.67186362420 *T)
    sumZ0  +=       0.811 * cos(0.19379046862  + 3346.13535100720 *T)
    sumZ0  +=       0.812 * cos(4.19639785587  + 3335.08950239240 *T)
    sumZ0  +=       0.774 * cos(1.38696439205  + 7234.79425624200 *T)
    sumZ0  +=       0.761 * cos(1.99221081181  + 26087.90314157420 *T)
    sumZ0  +=       0.742 * cos(0.37908087243  + 13358.92658845020 *T)
    sumZ0  +=       0.732 * cos(5.16784948008  + 9830.38901398780 *T)
    sumZ0  +=       0.946 * cos(1.88936722079  + 5628.95647021120 *T)
    sumZ0  +=       0.820 * cos(4.37824637524  + 7373.38245462640 *T)
    sumZ0  +=       0.671 * cos(5.25224406605  + 76.26607127560 *T)
    sumZ0  +=       0.785 * cos(4.59990880607  + 6438.49624942560 *T)
    sumZ0  +=       0.668 * cos(3.61130377946  + 6702.00024889200 *T)
    sumZ0  +=       0.655 * cos(2.60002275981  + 11243.68584642080 *T)
    sumZ0  +=       0.760 * cos(3.03501058836  + 3360.96774609859 *T)
    sumZ0  +=       0.764 * cos(4.37581912554  + 11769.85369316640 *T)
    sumZ0  +=       0.704 * cos(5.80908299008  + 10021.90459040220 *T)
    sumZ0  +=       0.840 * cos(6.11799044851  + 5331.35744374080 *T)
    sumZ0  +=       0.641 * cos(1.73269627789  + 1692.16566950240 *T)
    sumZ0  +=       0.629 * cos(1.09880202633  + 14314.16811304980 *T)
    sumZ0  +=       0.704 * cos(6.21757937692  + 10021.76996979660 *T)
    sumZ0  +=       0.646 * cos(3.01919436853  + 9381.93999378540 *T)
    sumZ0  +=       0.612 * cos(3.40750819704  + 10713.99488132620 *T)
    sumZ0  +=       0.732 * cos(6.04237219046  + 13365.97282514820 *T)
    sumZ0  +=       0.771 * cos(2.47252701375  + 15113.98923821520 *T)
    sumZ0  +=       0.621 * cos(5.85690842904  + 3351.24909204960 *T)
    sumZ0  +=       0.767 * cos(1.35441792357  + 3320.25710730100 *T)
    sumZ0  +=       0.586 * cos(0.79051565116  + 853.19638175200 *T)
    sumZ0  +=       0.692 * cos(1.46718257115  + 6660.44945790720 *T)
    sumZ0  +=       0.669 * cos(0.51352618116  + 7064.12138562280 *T)
    sumZ0  +=       0.528 * cos(3.12611933096  + 74.78159856730 *T)
    sumZ0  +=       0.600 * cos(5.04547651146  + 692.15760122680 *T)
    sumZ0  +=       0.630 * cos(1.58386830258  + 1581.95934828300 *T)
    sumZ0  +=       0.623 * cos(4.81491846507  + 3329.97576135000 *T)
    sumZ0  +=       0.661 * cos(2.79286199869  + 2906.90068882300 *T)
    sumZ0  +=       0.570 * cos(4.61307898923  + 3074.00538497800 *T)
    sumZ0  +=       0.524 * cos(6.05770135572  + 3607.21946842160 *T)
    sumZ0  +=       0.516 * cos(2.43629649577  + 13916.01910964160 *T)
    sumZ0  +=       0.507 * cos(2.02939502894  + 3333.56619000180 *T)
    sumZ0  +=       0.605 * cos(1.93804011864  + 10575.40668294180 *T)
    sumZ0  +=       0.488 * cos(4.67812564849  + 5729.50644714900 *T)
    sumZ0  +=       0.465 * cos(5.56224766580  + 8955.34180290940 *T)
    sumZ0  +=       0.467 * cos(3.02812406788  + 6144.42034130420 *T)
    sumZ0  +=       0.524 * cos(2.17529898104  + 6298.32832117640 *T)
    sumZ0  +=       0.449 * cos(4.27905913954  + 6546.15977336420 *T)
    sumZ0  +=       0.442 * cos(2.21279250958  + 3347.65866339780 *T)
    sumZ0  +=       0.442 * cos(0.35835774849  + 18052.92954315780 *T)
    sumZ0  +=       0.452 * cos(4.74868579573  + 5828.02847164760 *T)
    sumZ0  +=       0.444 * cos(2.31126753824  + 6677.63442474780 *T)
    sumZ0  +=       0.454 * cos(5.96502230469  + 419.48464387520 *T)
    sumZ0  +=       0.457 * cos(2.76482172070  + 5459.37628707820 *T)
    sumZ0  +=       0.452 * cos(2.18667537557  + 12935.85151592320 *T)
    sumZ0  +=       0.440 * cos(2.28024553720  + 8969.56889691100 *T)
    sumZ0  +=       0.439 * cos(3.26823913184  + 6127.65545055720 *T)
    sumZ0  +=       0.417 * cos(2.62418009712  + 433.71173787680 *T)
    sumZ0  +=       0.450 * cos(4.97549718782  + 6531.66165626500 *T)
    sumZ0  +=       0.467 * cos(5.81074791371  + 6158.64743530580 *T)
    sumZ0  +=       0.430 * cos(4.78793443738  + 3863.18984479360 *T)
    sumZ0  +=       0.409 * cos(1.99836071636  + 3980.50971301380 *T)
    sumZ0  +=       0.399 * cos(4.74255258433  + 10818.13528691580 *T)
    sumZ0  +=       0.453 * cos(6.26634249241  + 14712.31711645800 *T)
    sumZ0  +=       0.437 * cos(2.74634913436  + 6684.81528205140 *T)
    sumZ0  +=       0.400 * cos(2.77865299135  + 4989.05918389720 *T)
    sumZ0  +=       0.506 * cos(1.18461823734  + 1744.42598441520 *T)
    sumZ0  +=       0.403 * cos(5.36937826882  + 1596.18644228460 *T)
    sumZ0  +=       0.380 * cos(4.97818849120  + 10014.72373309860 *T)
    sumZ0  +=       0.371 * cos(3.33405966002  + 10177.25767953360 *T)
    sumZ0  +=       0.361 * cos(5.24088703687  + 2178.13772229200 *T)
    sumZ0  +=       0.400 * cos(1.62063267593  + 14584.29827312060 *T)
    sumZ0  +=       0.403 * cos(1.82423390219  + 3399.98628861340 *T)
    sumZ0  +=       0.410 * cos(4.54392931383  + 3344.49376205780 *T)
    sumZ0  +=       0.418 * cos(4.88395861492  + 3443.70520091840 *T)
    sumZ0  +=       0.341 * cos(3.76526341028  + 6688.33840040040 *T)
    sumZ0  +=       0.387 * cos(2.20624213901  + 3281.23856478620 *T)
    sumZ0  +=       0.328 * cos(1.60706573380  + 3546.79797513700 *T)
    sumZ0  +=       0.346 * cos(5.93876607715  + 3364.49086444760 *T)
    sumZ0  +=       0.427 * cos(0.42766619939  + 10551.52824519400 *T)
    sumZ0  +=       0.333 * cos(2.97391198780  + 17256.63153634140 *T)
    sumZ0  +=       0.351 * cos(0.89374033083  + 16173.37116840440 *T)
    sumZ0  +=       0.387 * cos(5.50555914948  + 3973.39616601300 *T)
    sumZ0  +=       0.410 * cos(6.13790219128  + 3336.73109134180 *T)
    sumZ0  +=       0.349 * cos(4.74282066857  + 3316.73398895200 *T)
    sumZ0  +=       0.366 * cos(0.54436256149  + 110.20632121940 *T)
    sumZ0  +=       0.284 * cos(2.49529211057  + 6784.31762761820 *T)
    sumZ0  +=       0.338 * cos(3.01411006339  + 8270.29774868340 *T)
    sumZ0  +=       0.279 * cos(0.15295518242  + 135.06508003540 *T)
    sumZ0  +=       0.290 * cos(4.30595018325  + 16858.48253293320 *T)
    sumZ0  +=       0.277 * cos(3.21266616465  + 8329.67161059700 *T)
    sumZ0  +=       0.303 * cos(4.39325865345  + 2818.03500860600 *T)
    sumZ0  +=       0.302 * cos(5.76020823903  + 2384.32327072920 *T)
    sumZ0  +=       0.360 * cos(0.81576341528  + 6816.28993343500 *T)
    sumZ0  +=       0.287 * cos(5.89677690833  + 3340.19235060619 *T)
    sumZ0  +=       0.300 * cos(3.61030466166  + 3304.58456002240 *T)
    sumZ0  +=       0.264 * cos(5.45859266580  + 6604.95878212400 *T)
    sumZ0  +=       0.330 * cos(3.81421832191  + 13517.87010623340 *T)
    sumZ0  +=       0.287 * cos(4.78503735880  + 3341.03250279340 *T)
    sumZ0  +=       0.343 * cos(1.02525479822  + 3230.40610548040 *T)
    sumZ0  +=       0.246 * cos(2.64795288862  + 3304.00706139560 *T)
    sumZ0  +=       0.310 * cos(1.76905174573  + 10404.73381232260 *T)
    sumZ0  +=       0.241 * cos(3.33235757149  + 11081.21921028860 *T)
    sumZ0  +=       0.245 * cos(1.00254307216  + 3209.07046501340 *T)
    sumZ0  +=       0.295 * cos(0.78856158130  + 3376.64029337720 *T)
    sumZ0  +=       0.247 * cos(4.74687572675  + 9779.10867612540 *T)
    sumZ0  +=       0.241 * cos(1.74832761256  + 3377.21779200400 *T)
    sumZ0  +=       0.264 * cos(1.82146900201  + 17654.78053974960 *T)
    sumZ0  +=       0.274 * cos(2.60650323538  + 3657.00429635640 *T)
    sumZ0  +=       0.238 * cos(0.43806562773  + 6923.95345737360 *T)
    sumZ0  +=       0.227 * cos(2.88048108848  + 3415.39402526710 *T)
    sumZ0  +=       0.300 * cos(0.70241789966  + 3134.42687826260 *T)
    sumZ0  +=       0.237 * cos(2.19737291235  + 13362.43245314700 *T)
    sumZ0  +=       0.220 * cos(1.43836209680  + 6606.44325483230 *T)
    sumZ0  +=       0.222 * cos(3.39627559384  + 3472.15438838620 *T)
    sumZ0  +=       0.217 * cos(2.18530503085  + 20995.39296644940 *T)
    sumZ0  +=       0.254 * cos(2.93819031378  + 5085.03841111500 *T)
    sumZ0  +=       0.236 * cos(4.23042638023  + 13362.46696045140 *T)
    sumZ0  +=       0.242 * cos(0.41755696147  + 1861.74585263540 *T)
    sumZ0  +=       0.285 * cos(0.10171887281  + 17085.95866572220 *T)
    sumZ0  +=       0.218 * cos(0.33151885058  + 16304.91313009080 *T)
    sumZ0  +=       0.215 * cos(5.26457047836  + 1375.77379984580 *T)
    sumZ0  +=       0.207 * cos(3.60591178333  + 7107.82304427560 *T)
    sumZ0  +=       0.196 * cos(3.34505476799  + 8439.87793181640 *T)
    sumZ0  +=       0.192 * cos(5.08475415071  + 13119.72110282519 *T)
    sumZ0  +=       0.269 * cos(0.19698472851  + 3024.22055704320 *T)
    sumZ0  +=       0.194 * cos(4.52507414451  + 220.41264243880 *T)
    sumZ0  +=       0.186 * cos(1.75940958561  + 2221.85663459700 *T)
    sumZ0  +=       0.194 * cos(3.41534671398  + 18984.29263000960 *T)
    sumZ0  +=       0.208 * cos(3.02738720944  + 3511.28529731900 *T)
    sumZ0  +=       0.204 * cos(6.00463791835  + 162.46663613220 *T)
    sumZ0  +=       0.183 * cos(3.62962279261  + 9168.64089834740 *T)
    sumZ0  +=       0.175 * cos(4.04165128738  + 3760.09707057500 *T)
    sumZ0  +=       0.174 * cos(4.51117927233  + 23539.70738633280 *T)
    sumZ0  +=       0.193 * cos(5.88968824988  + 2803.80791460440 *T)
    sumZ0  +=       0.190 * cos(6.17637484900  + 13745.34623902240 *T)
    sumZ0  +=       0.208 * cos(3.34809361979  + 3341.04230982650 *T)
    sumZ0  +=       0.198 * cos(0.97311347960  + 3265.83082813250 *T)
    sumZ0  +=       0.205 * cos(4.20157952693  + 6843.69148953180 *T)
    sumZ0  +=       0.222 * cos(1.37049171707  + 3169.93955608060 *T)
    sumZ0  +=       0.183 * cos(2.19041678580  + 9866.41688066520 *T)
    sumZ0  +=       0.225 * cos(0.32450219510  + 13149.15061136120 *T)
    sumZ0  +=       0.211 * cos(4.38136759956  + 640.87760738220 *T)
    sumZ0  +=       0.200 * cos(3.00112781370  + 1162.47470440780 *T)
    sumZ0  +=       0.208 * cos(6.18496348629  + 1478.86657406440 *T)
    sumZ0  +=       0.173 * cos(3.90520749444  + 5099.26550511660 *T)
    sumZ0  +=       0.175 * cos(0.38064797785  + 956.28915597060 *T)
    sumZ0  +=       0.158 * cos(2.98574812713  + 13524.91634293140 *T)
    sumZ0  +=       0.170 * cos(3.37406232659  + 15508.61512327440 *T)
    sumZ0  +=       0.184 * cos(2.09449845330  + 949.17560896980 *T)
    sumZ0  +=       0.171 * cos(4.75570011379  + 6261.74020952440 *T)
    sumZ0  +=       0.188 * cos(5.56241899175  + 16460.33352952499 *T)
    sumZ0  +=       0.205 * cos(5.01940138208  + 1118.75579210280 *T)
    sumZ0  +=       0.156 * cos(3.68058177082  + 13553.89797291080 *T)
    sumZ0  +=       0.170 * cos(2.00050957424  + 224.34479570190 *T)
    sumZ0  +=       0.208 * cos(1.04576950390  + 3340.18254357310 *T)
    sumZ0  +=       0.156 * cos(4.90620162208  + 20199.09495963300 *T)
    sumZ0  +=       0.174 * cos(5.58778318130  + 2766.26762836500 *T)
    sumZ0  +=       0.169 * cos(3.71127165203  + 10184.30391623160 *T)
    sumZ0  +=       0.171 * cos(3.60600294119  + 20597.24396304120 *T)
    sumZ0  +=       0.140 * cos(1.54110979029  + 2111.65031337760 *T)
    sumZ0  +=       0.145 * cos(6.17161140540  + 19800.94595622480 *T)
    sumZ0  +=       0.156 * cos(0.43599559375  + 4459.36821880260 *T)
    sumZ0  +=       0.143 * cos(0.19892133285  + 10235.13637553740 *T)
    sumZ0  +=       0.184 * cos(1.58017928752  + 8273.82086703240 *T)
    sumZ0  +=       0.133 * cos(2.86759235449  + 20618.01935853360 *T)
    sumZ0  +=       0.151 * cos(1.59809012076  + 13892.14067189380 *T)
    sumZ0  +=       0.133 * cos(6.14918807002  + 17277.40693183380 *T)
    sumZ0  +=       0.167 * cos(0.94707767481  + 3274.12501778540 *T)
    sumZ0  +=       0.143 * cos(4.48621138334  + 632.78373931320 *T)
    sumZ0  +=       0.126 * cos(4.54286582844  + 3914.95722503460 *T)
    sumZ0  +=       0.145 * cos(2.88581037055  + 4392.88080988820 *T)
    sumZ0  +=       0.165 * cos(5.44939669734  + 1228.96211332220 *T)
    sumZ0  +=       0.124 * cos(2.73197400851  + 3407.09983561420 *T)
    sumZ0  +=       0.145 * cos(5.65866985683  + 4407.10790388980 *T)
    sumZ0  +=       0.164 * cos(2.20529053904  + 38.13303563780 *T)
    sumZ0  +=       0.158 * cos(1.72206104768  + 3296.89351439480 *T)
    sumZ0  +=       0.116 * cos(0.02904609570  + 26482.17080962440 *T)
    sumZ0  +=       0.156 * cos(1.25691413032  + 3384.33133900480 *T)
    sumZ0  +=       0.117 * cos(4.96153079924  + 3981.49003408200 *T)
    sumZ0  +=       0.154 * cos(2.17255031800  + 1964.83862685400 *T)
    sumZ0  +=       0.113 * cos(3.06854044437  + 206.18554843720 *T)
    sumZ0  +=       0.117 * cos(2.13763640035  + 29424.63423291600 *T)
    sumZ0  +=       0.129 * cos(2.20198223551  + 3450.81874791920 *T)
    sumZ0  +=       0.115 * cos(5.24850953150  + 3237.51965248120 *T)
    sumZ0  +=       0.112 * cos(6.26358102825  + 2125.87740737920 *T)
    sumZ0  +=       0.112 * cos(4.67861613977  + 2807.39834325620 *T)
    sumZ0  +=       0.115 * cos(0.38603739544  + 12721.57209941700 *T)
    sumZ0  +=       0.113 * cos(1.36558076768  + 6709.67404086740 *T)
    sumZ0  +=       0.144 * cos(0.22567806033  + 263.08392337280 *T)
    sumZ0  +=       0.109 * cos(5.27187307952  + 5518.75014899180 *T)
    sumZ0  +=       0.125 * cos(2.20397088123  + 30220.93223973240 *T)
    sumZ0  +=       0.108 * cos(6.20219802547  + 6665.97238221460 *T)
    sumZ0  +=       0.127 * cos(2.68652319514  + 7321.12213971360 *T)
    sumZ0  +=       0.110 * cos(0.18586807658  + 29822.78323632420 *T)
    sumZ0  +=       0.107 * cos(0.50226646577  + 22743.40937951640 *T)
    sumZ0  +=       0.107 * cos(3.59856046864  + 5989.06725217280 *T)
    sumZ0  +=       0.106 * cos(5.69966166034  + 33561.54466643220 *T)
    sumZ0  +=       0.105 * cos(3.97375976080  + 9380.95967271720 *T)

    # MARS Z Coord
    sumZ1  = 0
    sumZ1  +=  533220.761 * cos(5.37045188878  + 3340.61242669980 *T)
    sumZ1  +=   32258.788 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ1  +=    7302.119 * cos(5.48197049223  + 6681.22485339960 *T)
    sumZ1  +=     777.092 * cos(3.20452658315  + 10021.83728009940 *T)
    sumZ1  +=     166.695 * cos(3.41487685068  + 13362.44970679920 *T)
    sumZ1  +=     156.044 * cos(0.77594576545  + 3337.08930835080 *T)
    sumZ1  +=      52.883 * cos(2.55566894166  + 2281.23049651060 *T)
    sumZ1  +=      42.031 * cos(6.05247110697  + 529.69096509460 *T)
    sumZ1  +=      31.177 * cos(3.47957118843  + 5621.84292321040 *T)
    sumZ1  +=      24.541 * cos(3.72250523350  + 16703.06213349900 *T)
    sumZ1  +=      25.907 * cos(4.79004064527  + 3344.13554504880 *T)
    sumZ1  +=      23.800 * cos(2.78437789651  + 3340.59517304760 *T)
    sumZ1  +=      17.724 * cos(0.33832536017  + 398.14900340820 *T)
    sumZ1  +=      24.098 * cos(0.26741008281  + 2942.46342329160 *T)
    sumZ1  +=      23.800 * cos(4.81514959722  + 3340.62968035200 *T)
    sumZ1  +=      17.059 * cos(1.42825204765  + 6151.53388830500 *T)
    sumZ1  +=      19.956 * cos(1.69346504279  + 3532.06069281140 *T)
    sumZ1  +=      15.117 * cos(1.79886272527  + 2544.31441988340 *T)
    sumZ1  +=      13.619 * cos(6.03064463759  + 2810.92146160520 *T)
    sumZ1  +=      13.681 * cos(4.34937268643  + 3496.03282613400 *T)
    sumZ1  +=       9.980 * cos(0.97276584150  + 6677.70173505060 *T)
    sumZ1  +=      11.657 * cos(2.25529456107  + 5884.92684658320 *T)
    sumZ1  +=       9.888 * cos(4.14137402777  + 1059.38193018920 *T)
    sumZ1  +=       8.253 * cos(3.93247597797  + 8962.45534991020 *T)
    sumZ1  +=       8.936 * cos(0.03236493895  + 3185.19202726560 *T)
    sumZ1  +=       7.843 * cos(0.86090694397  + 3340.54511639700 *T)
    sumZ1  +=       7.463 * cos(0.85796039982  + 6283.07584999140 *T)
    sumZ1  +=       7.838 * cos(0.45374819664  + 3340.67973700260 *T)
    sumZ1  +=       6.532 * cos(3.46604878226  + 5092.15195811580 *T)
    sumZ1  +=       7.000 * cos(2.91161281175  + 1751.53953141600 *T)
    sumZ1  +=       6.215 * cos(4.29108610819  + 6684.74797174860 *T)
    sumZ1  +=       5.486 * cos(3.73127537680  + 5486.77784317500 *T)
    sumZ1  +=       4.744 * cos(3.11292686055  + 3339.63210563160 *T)
    sumZ1  +=       4.768 * cos(1.11702153656  + 2914.01423582380 *T)
    sumZ1  +=       3.697 * cos(5.07727152864  + 5088.62883976680 *T)
    sumZ1  +=       3.696 * cos(5.17619762565  + 796.29800681640 *T)
    sumZ1  +=       4.194 * cos(1.30289365634  + 3738.76143010800 *T)
    sumZ1  +=       3.104 * cos(2.90264842875  + 3894.18182954220 *T)
    sumZ1  +=       3.078 * cos(3.23232337445  + 2146.16541647520 *T)
    sumZ1  +=       3.060 * cos(1.47684271283  + 3870.30339179440 *T)
    sumZ1  +=       3.279 * cos(4.04708484838  + 20043.67456019880 *T)
    sumZ1  +=       2.645 * cos(6.20328748216  + 213.29909543800 *T)
    sumZ1  +=       2.591 * cos(5.67618677721  + 1592.59601363280 *T)
    sumZ1  +=       2.686 * cos(2.27518970275  + 3149.16416058820 *T)
    sumZ1  +=       3.065 * cos(5.50490756275  + 3127.31333126180 *T)
    sumZ1  +=       2.618 * cos(1.49434953131  + 4292.33083295040 *T)
    sumZ1  +=       2.929 * cos(0.13035493149  + 4690.47983635860 *T)
    sumZ1  +=       2.433 * cos(0.83360306332  + 1194.44701022460 *T)
    sumZ1  +=       1.893 * cos(1.44037290348  + 2787.04302385740 *T)
    sumZ1  +=       2.056 * cos(2.69646867769  + 4136.91043351620 *T)
    sumZ1  +=       1.896 * cos(4.12121426717  + 3723.50895892300 *T)
    sumZ1  +=       1.810 * cos(1.25670128975  + 3341.59274776800 *T)
    sumZ1  +=       1.948 * cos(4.26663375622  + 1990.74501704100 *T)
    sumZ1  +=       2.146 * cos(2.79934106364  + 6254.62666252360 *T)
    sumZ1  +=       1.699 * cos(5.88106551599  + 3097.88382272579 *T)
    sumZ1  +=       1.644 * cos(2.85985547231  + 2388.89402044920 *T)
    sumZ1  +=       1.210 * cos(3.52826008702  + 155.42039943420 *T)
    sumZ1  +=       1.638 * cos(4.23138744183  + 8432.76438481560 *T)
    sumZ1  +=       1.151 * cos(4.07168420281  + 7.11354700080 *T)
    sumZ1  +=       1.111 * cos(0.33771745106  + 2957.71589447660 *T)
    sumZ1  +=       1.375 * cos(4.90390057340  + 3583.34103067380 *T)
    sumZ1  +=       1.075 * cos(2.89446881499  + 1589.07289528380 *T)
    sumZ1  +=       0.994 * cos(1.22724079631  + 4535.05943692440 *T)
    sumZ1  +=       0.973 * cos(2.54415383282  + 9225.53927328300 *T)
    sumZ1  +=       1.053 * cos(2.17507915001  + 2288.34404351140 *T)
    sumZ1  +=       1.234 * cos(1.07118451631  + 4399.99435688900 *T)
    sumZ1  +=       0.937 * cos(0.33440700581  + 6674.11130639880 *T)
    sumZ1  +=       0.964 * cos(0.92522802179  + 6467.92575796160 *T)
    sumZ1  +=       1.180 * cos(1.46657018316  + 3333.49887969900 *T)
    sumZ1  +=       1.149 * cos(4.46586649175  + 10025.36039844840 *T)
    sumZ1  +=       0.787 * cos(4.21744543042  + 426.59819087600 *T)
    sumZ1  +=       0.734 * cos(1.59169132873  + 9492.14631500480 *T)
    sumZ1  +=       0.744 * cos(3.21417897895  + 1221.84856632140 *T)
    sumZ1  +=       0.850 * cos(1.84579754127  + 10213.28554621100 *T)
    sumZ1  +=       0.960 * cos(0.37818281067  + 6525.80445396540 *T)
    sumZ1  +=       0.748 * cos(3.14653223844  + 6489.77658728800 *T)
    sumZ1  +=       0.939 * cos(3.74411147052  + 4562.46099302120 *T)
    sumZ1  +=       0.816 * cos(4.88051263913  + 3361.38782219220 *T)
    sumZ1  +=       0.670 * cos(1.82692939869  + 3325.35995551480 *T)
    sumZ1  +=       0.787 * cos(4.62073861040  + 1748.01641306700 *T)
    sumZ1  +=       0.746 * cos(2.72473273276  + 3319.83703120740 *T)
    sumZ1  +=       0.615 * cos(3.00346220351  + 7477.52286021600 *T)
    sumZ1  +=       0.612 * cos(0.94873589103  + 10018.31416175040 *T)
    sumZ1  +=       0.726 * cos(4.21670769259  + 1052.26838318840 *T)
    sumZ1  +=       0.562 * cos(2.51075037144  + 2274.11694950980 *T)
    sumZ1  +=       0.676 * cos(5.46592420299  + 3205.54734666440 *T)
    sumZ1  +=       0.652 * cos(4.92680960161  + 6681.24210705180 *T)
    sumZ1  +=       0.580 * cos(1.95948510974  + 3553.91152213780 *T)
    sumZ1  +=       0.620 * cos(4.53706892447  + 3360.96774609859 *T)
    sumZ1  +=       0.568 * cos(4.48134936442  + 6836.64525283380 *T)
    sumZ1  +=       0.516 * cos(3.66568073238  + 1066.49547719000 *T)
    sumZ1  +=       0.652 * cos(2.89603807936  + 6681.20759974740 *T)
    sumZ1  +=       0.528 * cos(3.91265871028  + 6041.32756708560 *T)
    sumZ1  +=       0.590 * cos(2.14286730936  + 553.56940284240 *T)
    sumZ1  +=       0.481 * cos(3.65133581406  + 3337.02199804800 *T)
    sumZ1  +=       0.486 * cos(0.03539800900  + 522.57741809380 *T)
    sumZ1  +=       0.479 * cos(4.08237627559  + 3344.20285535160 *T)
    sumZ1  +=       0.491 * cos(3.26656040639  + 382.89653222320 *T)
    sumZ1  +=       0.599 * cos(3.48114565787  + 191.44826611160 *T)
    sumZ1  +=       0.460 * cos(3.60938831179  + 11773.37681151540 *T)
    sumZ1  +=       0.484 * cos(4.53788344821  + 7079.37385680780 *T)
    sumZ1  +=       0.457 * cos(4.49932663564  + 242.72860397400 *T)
    sumZ1  +=       0.410 * cos(3.93129679071  + 12303.06777661000 *T)
    sumZ1  +=       0.432 * cos(0.62719640964  + 5614.72937620960 *T)
    sumZ1  +=       0.415 * cos(5.78715397470  + 1349.86740965880 *T)
    sumZ1  +=       0.449 * cos(6.07546222965  + 4933.20844033260 *T)
    sumZ1  +=       0.414 * cos(0.76551632852  + 951.71840625060 *T)
    sumZ1  +=       0.477 * cos(6.15909631316  + 3320.25710730100 *T)
    sumZ1  +=       0.421 * cos(4.37788398088  + 23384.28698689860 *T)
    sumZ1  +=       0.418 * cos(6.18476783317  + 6438.49624942560 *T)
    sumZ1  +=       0.482 * cos(3.91728938067  + 8827.39026987480 *T)
    sumZ1  +=       0.365 * cos(2.58064909826  + 3355.86489788480 *T)
    sumZ1  +=       0.333 * cos(3.42873039925  + 6680.24453233140 *T)
    sumZ1  +=       0.330 * cos(3.24081124683  + 9595.23908922340 *T)
    sumZ1  +=       0.317 * cos(1.00610206365  + 9623.68827669120 *T)
    sumZ1  +=       0.365 * cos(5.85846191272  + 639.89728631400 *T)
    sumZ1  +=       0.293 * cos(1.52957859377  + 7740.60678358880 *T)
    sumZ1  +=       0.331 * cos(0.22404636497  + 2700.71514038580 *T)
    sumZ1  +=       0.255 * cos(3.99432797748  + 2921.12778282460 *T)
    sumZ1  +=       0.304 * cos(4.22948972595  + 3312.16323923200 *T)
    sumZ1  +=       0.328 * cos(2.40258646743  + 2118.76386037840 *T)
    sumZ1  +=       0.236 * cos(3.40020727050  + 1758.65307841680 *T)
    sumZ1  +=       0.251 * cos(5.66519309695  + 536.80451209540 *T)
    sumZ1  +=       0.237 * cos(1.24529852403  + 3364.49086444760 *T)
    sumZ1  +=       0.214 * cos(0.56442646442  + 6681.29216370240 *T)
    sumZ1  +=       0.245 * cos(1.46405275777  + 4929.68532198360 *T)
    sumZ1  +=       0.261 * cos(1.37575395801  + 3347.72597370060 *T)
    sumZ1  +=       0.277 * cos(2.45714538470  + 3503.07906283200 *T)
    sumZ1  +=       0.257 * cos(4.63989625540  + 5331.35744374080 *T)
    sumZ1  +=       0.216 * cos(1.78813616163  + 6127.65545055720 *T)
    sumZ1  +=       0.266 * cos(1.65519778241  + 6682.20517446780 *T)
    sumZ1  +=       0.204 * cos(4.87400049907  + 6144.42034130420 *T)
    sumZ1  +=       0.214 * cos(0.97289996058  + 6681.15754309680 *T)
    sumZ1  +=       0.193 * cos(2.86664857796  + 7210.91581849420 *T)
    sumZ1  +=       0.189 * cos(3.19369904946  + 5729.50644714900 *T)
    sumZ1  +=       0.185 * cos(4.60022172333  + 12168.00269657460 *T)
    sumZ1  +=       0.184 * cos(4.90073216006  + 10419.98628350760 *T)
    sumZ1  +=       0.196 * cos(2.02607662846  + 3346.13535100720 *T)
    sumZ1  +=       0.191 * cos(4.71901518542  + 13365.97282514820 *T)
    sumZ1  +=       0.174 * cos(1.31405252638  + 7064.12138562280 *T)
    sumZ1  +=       0.167 * cos(0.99843143165  + 8955.34180290940 *T)
    sumZ1  +=       0.170 * cos(1.48043717586  + 7875.67186362420 *T)
    sumZ1  +=       0.197 * cos(3.29154622691  + 17654.78053974960 *T)
    sumZ1  +=       0.170 * cos(2.88198960502  + 6872.67311951120 *T)
    sumZ1  +=       0.157 * cos(2.56733171598  + 3178.14579056760 *T)
    sumZ1  +=       0.148 * cos(0.86328445077  + 2699.73481931760 *T)
    sumZ1  +=       0.152 * cos(5.23944955455  + 6923.95345737360 *T)
    sumZ1  +=       0.171 * cos(3.14240135508  + 3316.73398895200 *T)
    sumZ1  +=       0.174 * cos(1.48894564475  + 3767.21061757580 *T)
    sumZ1  +=       0.144 * cos(3.57581507564  + 9830.38901398780 *T)
    sumZ1  +=       0.140 * cos(3.50121767027  + 3077.52850332700 *T)
    sumZ1  +=       0.147 * cos(0.72898424375  + 2487.41604494780 *T)
    sumZ1  +=       0.165 * cos(1.35817246585  + 3.52311834900 *T)
    sumZ1  +=       0.171 * cos(5.87779648432  + 11769.85369316640 *T)
    sumZ1  +=       0.146 * cos(5.13010621713  + 3304.58456002240 *T)
    sumZ1  +=       0.126 * cos(4.14449585210  + 3603.69635007260 *T)
    sumZ1  +=       0.150 * cos(0.67031927264  + 6298.32832117640 *T)
    sumZ1  +=       0.169 * cos(3.37958950467  + 1581.95934828300 *T)
    sumZ1  +=       0.132 * cos(1.36026397982  + 419.48464387520 *T)
    sumZ1  +=       0.137 * cos(3.91255974972  + 7903.07341972100 *T)
    sumZ1  +=       0.119 * cos(1.11588866360  + 433.71173787680 *T)
    sumZ1  +=       0.116 * cos(4.20840451833  + 9381.93999378540 *T)
    sumZ1  +=       0.132 * cos(5.61925791338  + 12832.75874170460 *T)
    sumZ1  +=       0.119 * cos(3.29620748546  + 10818.13528691580 *T)
    sumZ1  +=       0.139 * cos(1.48720938048  + 14712.31711645800 *T)
    sumZ1  +=       0.121 * cos(3.78071153759  + 1596.18644228460 *T)
    sumZ1  +=       0.116 * cos(0.71600721713  + 9866.41688066520 *T)
    sumZ1  +=       0.107 * cos(2.41569477061  + 2648.45482547300 *T)
    sumZ1  +=       0.141 * cos(1.40883592065  + 3336.73109134180 *T)
    sumZ1  +=       0.124 * cos(5.27111992916  + 853.19638175200 *T)
    sumZ1  +=       0.120 * cos(0.24314469015  + 3863.18984479360 *T)
    sumZ1  +=       0.106 * cos(1.89681319705  + 3475.67750673520 *T)
    sumZ1  +=       0.104 * cos(0.61683878316  + 10021.82002644720 *T)
    sumZ1  +=       0.099 * cos(2.78700157494  + 10575.40668294180 *T)
    sumZ1  +=       0.100 * cos(1.75288902946  + 8429.24126646660 *T)
    sumZ1  +=       0.105 * cos(3.95535664942  + 11243.68584642080 *T)
    sumZ1  +=       0.104 * cos(2.64763687393  + 10021.85453375160 *T)
    sumZ1  +=       0.097 * cos(3.70870547299  + 4032.77002792660 *T)
    sumZ1  +=       0.099 * cos(0.46508871986  + 149.56319713460 *T)
    sumZ1  +=       0.094 * cos(0.92726193663  + 7632.94325965020 *T)
    sumZ1  +=       0.098 * cos(1.72351303556  + 18052.92954315780 *T)
    sumZ1  +=       0.108 * cos(2.41308938131  + 7234.79425624200 *T)
    sumZ1  +=       0.093 * cos(2.12448381740  + 6688.33840040040 *T)
    sumZ1  +=       0.106 * cos(4.22500281625  + 6158.64743530580 *T)
    sumZ1  +=       0.109 * cos(3.05520401747  + 640.87760738220 *T)
    sumZ1  +=       0.110 * cos(0.14136188594  + 5628.95647021120 *T)
    sumZ1  +=       0.112 * cos(3.90540165638  + 6894.52394883760 *T)
    sumZ1  +=       0.077 * cos(3.42693027158  + 1744.42598441520 *T)
    sumZ1  +=       0.083 * cos(0.48188810071  + 8969.56889691100 *T)
    sumZ1  +=       0.078 * cos(3.48895806928  + 7373.38245462640 *T)
    sumZ1  +=       0.098 * cos(1.09842815785  + 5085.03841111500 *T)
    sumZ1  +=       0.087 * cos(5.57913927474  + 3376.64029337720 *T)
    sumZ1  +=       0.071 * cos(0.19809257252  + 3980.50971301380 *T)
    sumZ1  +=       0.072 * cos(4.45423986754  + 17256.63153634140 *T)
    sumZ1  +=       0.082 * cos(0.70225541331  + 9808.53818466140 *T)
    sumZ1  +=       0.075 * cos(1.88840044542  + 15643.68020330980 *T)
    sumZ1  +=       0.071 * cos(4.47046372301  + 10177.25767953360 *T)
    sumZ1  +=       0.070 * cos(3.37789486799  + 13517.87010623340 *T)
    sumZ1  +=       0.075 * cos(2.27088357802  + 12566.15169998280 *T)
    sumZ1  +=       0.066 * cos(0.13337760475  + 15110.46611986620 *T)
    sumZ1  +=       0.080 * cos(1.32412177054  + 21393.54196985760 *T)
    sumZ1  +=       0.076 * cos(3.32363330267  + 13916.01910964160 *T)
    sumZ1  +=       0.067 * cos(4.66693287754  + 5828.02847164760 *T)
    sumZ1  +=       0.063 * cos(6.16425576512  + 3914.95722503460 *T)
    sumZ1  +=       0.073 * cos(3.44104044998  + 692.15760122680 *T)
    sumZ1  +=       0.064 * cos(5.11183842129  + 3191.04922956520 *T)
    sumZ1  +=       0.076 * cos(4.86164511219  + 20597.24396304120 *T)
    sumZ1  +=       0.061 * cos(1.44371703655  + 11081.21921028860 *T)
    sumZ1  +=       0.060 * cos(1.62689158397  + 3335.08950239240 *T)
    sumZ1  +=       0.075 * cos(4.78215371632  + 12964.30070339100 *T)
    sumZ1  +=       0.061 * cos(3.35711733735  + 20995.39296644940 *T)
    sumZ1  +=       0.053 * cos(6.24589101777  + 20199.09495963300 *T)
    sumZ1  +=       0.055 * cos(0.28321587193  + 5459.37628707820 *T)
    sumZ1  +=       0.052 * cos(4.94957858677  + 24734.15439655740 *T)
    sumZ1  +=       0.051 * cos(3.29132317156  + 15113.98923821520 *T)
    sumZ1  +=       0.068 * cos(3.59488057957  + 103.09277421860 *T)
    sumZ1  +=       0.057 * cos(2.70607036374  + 10713.99488132620 *T)
    sumZ1  +=       0.060 * cos(3.00951950244  + 3344.49376205780 *T)
    sumZ1  +=       0.065 * cos(2.78876221132  + 3981.49003408200 *T)
    sumZ1  +=       0.049 * cos(3.13153581502  + 2766.26762836500 *T)
    sumZ1  +=       0.049 * cos(2.33304379441  + 6660.44945790720 *T)
    sumZ1  +=       0.060 * cos(4.17222973491  + 4407.10790388980 *T)
    sumZ1  +=       0.049 * cos(0.10301634940  + 13358.92658845020 *T)
    sumZ1  +=       0.046 * cos(0.97092241478  + 8270.29774868340 *T)
    sumZ1  +=       0.061 * cos(1.61644555817  + 14314.16811304980 *T)
    sumZ1  +=       0.057 * cos(4.04165606796  + 1648.44675719740 *T)
    sumZ1  +=       0.060 * cos(0.05239748726  + 8273.82086703240 *T)
    sumZ1  +=       0.044 * cos(5.68073325802  + 3329.97576135000 *T)
    sumZ1  +=       0.044 * cos(0.43949269085  + 3351.24909204960 *T)
    sumZ1  +=       0.043 * cos(4.87531812567  + 10021.76996979660 *T)
    sumZ1  +=       0.048 * cos(1.42334455474  + 76.26607127560 *T)
    sumZ1  +=       0.045 * cos(4.07917825234  + 1692.16566950240 *T)
    sumZ1  +=       0.046 * cos(3.33348460172  + 3369.06161416760 *T)
    sumZ1  +=       0.045 * cos(5.89893308094  + 2818.03500860600 *T)
    sumZ1  +=       0.046 * cos(1.69497220340  + 10404.73381232260 *T)
    sumZ1  +=       0.041 * cos(5.80219476378  + 6702.00024889200 *T)
    sumZ1  +=       0.040 * cos(5.47626814223  + 3074.00538497800 *T)
    sumZ1  +=       0.044 * cos(5.36428737888  + 2707.82868738660 *T)
    sumZ1  +=       0.037 * cos(0.64022063667  + 3607.21946842160 *T)
    sumZ1  +=       0.043 * cos(4.47402411722  + 10021.90459040220 *T)
    sumZ1  +=       0.034 * cos(3.20375305228  + 5099.26550511660 *T)
    sumZ1  +=       0.034 * cos(5.74767895336  + 6531.66165626500 *T)
    sumZ1  +=       0.032 * cos(4.71476653314  + 16173.37116840440 *T)
    sumZ1  +=       0.035 * cos(4.78248662801  + 10551.52824519400 *T)
    sumZ1  +=       0.031 * cos(2.45682708815  + 10973.55568635000 *T)
    sumZ1  +=       0.036 * cos(2.89209321078  + 3333.56619000180 *T)
    sumZ1  +=       0.031 * cos(3.61192484167  + 6684.81528205140 *T)
    sumZ1  +=       0.041 * cos(4.63075392785  + 8031.09226305840 *T)
    sumZ1  +=       0.032 * cos(3.03761491100  + 12935.85151592320 *T)
    sumZ1  +=       0.038 * cos(0.87141658782  + 3973.39616601300 *T)
    sumZ1  +=       0.040 * cos(3.89104366040  + 74.78159856730 *T)
    sumZ1  +=       0.028 * cos(3.07347183013  + 3281.23856478620 *T)

    # MARS Z Coord
    sumZ2  = 0
    sumZ2  +=   25937.973 * cos(0.59444055009  + 3340.61242669980 *T)
    sumZ2  +=   11396.105 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ2  +=     237.211 * cos(5.64509583638  + 6681.22485339960 *T)
    sumZ2  +=      18.120 * cos(2.24234887624  + 3337.08930835080 *T)
    sumZ2  +=      11.005 * cos(0.65490267432  + 10021.83728009940 *T)
    sumZ2  +=       9.574 * cos(1.87721957302  + 13362.44970679920 *T)
    sumZ2  +=       4.373 * cos(5.89380048765  + 3496.03282613400 *T)
    sumZ2  +=       2.387 * cos(1.18064647446  + 2281.23049651060 *T)
    sumZ2  +=       2.425 * cos(2.24521403500  + 16703.06213349900 *T)
    sumZ2  +=       2.528 * cos(5.16130960942  + 5621.84292321040 *T)
    sumZ2  +=       1.966 * cos(1.32977724844  + 529.69096509460 *T)
    sumZ2  +=       2.006 * cos(4.80335600212  + 3185.19202726560 *T)
    sumZ2  +=       2.202 * cos(3.18250102468  + 3532.06069281140 *T)
    sumZ2  +=       1.723 * cos(3.39265057715  + 2544.31441988340 *T)
    sumZ2  +=       2.050 * cos(1.83403249506  + 2942.46342329160 *T)
    sumZ2  +=       1.453 * cos(0.37005682859  + 3344.13554504880 *T)
    sumZ2  +=       1.154 * cos(4.29461877948  + 3340.59517304760 *T)
    sumZ2  +=       0.942 * cos(5.17754009029  + 5486.77784317500 *T)
    sumZ2  +=       1.154 * cos(0.04220516563  + 3340.62968035200 *T)
    sumZ2  +=       1.059 * cos(3.80803028238  + 5884.92684658320 *T)
    sumZ2  +=       0.802 * cos(4.47943934542  + 3894.18182954220 *T)
    sumZ2  +=       0.790 * cos(2.35155821746  + 6677.70173505060 *T)
    sumZ2  +=       0.825 * cos(4.68895461946  + 3339.63210563160 *T)
    sumZ2  +=       0.712 * cos(0.32050354263  + 5088.62883976680 *T)
    sumZ2  +=       0.708 * cos(3.10465264690  + 6151.53388830500 *T)
    sumZ2  +=       0.753 * cos(1.05772254525  + 3097.88382272579 *T)
    sumZ2  +=       0.681 * cos(5.28693721347  + 2810.92146160520 *T)
    sumZ2  +=       0.576 * cos(2.00619713133  + 398.14900340820 *T)
    sumZ2  +=       0.534 * cos(4.77461983159  + 2146.16541647520 *T)
    sumZ2  +=       0.558 * cos(4.73319922901  + 5092.15195811580 *T)
    sumZ2  +=       0.532 * cos(3.07796101643  + 4292.33083295040 *T)
    sumZ2  +=       0.431 * cos(5.63787706168  + 3723.50895892300 *T)
    sumZ2  +=       0.486 * cos(1.71152918305  + 4690.47983635860 *T)
    sumZ2  +=       0.395 * cos(5.46836844513  + 8962.45534991020 *T)
    sumZ2  +=       0.380 * cos(2.37265090321  + 3340.54511639700 *T)
    sumZ2  +=       0.379 * cos(2.21710594361  + 6283.07584999140 *T)
    sumZ2  +=       0.380 * cos(1.96415613563  + 3340.67973700260 *T)
    sumZ2  +=       0.455 * cos(2.58588447058  + 20043.67456019880 *T)
    sumZ2  +=       0.340 * cos(6.21625640724  + 2787.04302385740 *T)
    sumZ2  +=       0.347 * cos(3.54462815350  + 3583.34103067380 *T)
    sumZ2  +=       0.287 * cos(4.23196046901  + 3333.49887969900 *T)
    sumZ2  +=       0.233 * cos(5.05398236760  + 155.42039943420 *T)
    sumZ2  +=       0.230 * cos(1.34020611446  + 3127.31333126180 *T)
    sumZ2  +=       0.236 * cos(2.56255875970  + 6684.74797174860 *T)
    sumZ2  +=       0.260 * cos(6.15858109011  + 1748.01641306700 *T)
    sumZ2  +=       0.206 * cos(3.72535272141  + 1059.38193018920 *T)
    sumZ2  +=       0.248 * cos(5.15522357344  + 6525.80445396540 *T)
    sumZ2  +=       0.192 * cos(0.80270977901  + 2288.34404351140 *T)
    sumZ2  +=       0.204 * cos(4.67148878659  + 6254.62666252360 *T)
    sumZ2  +=       0.240 * cos(1.10355321364  + 2388.89402044920 *T)
    sumZ2  +=       0.178 * cos(2.70046492113  + 1990.74501704100 *T)
    sumZ2  +=       0.151 * cos(4.03002963282  + 1592.59601363280 *T)
    sumZ2  +=       0.171 * cos(1.51083272272  + 1751.53953141600 *T)
    sumZ2  +=       0.178 * cos(5.84283595443  + 2914.01423582380 *T)
    sumZ2  +=       0.136 * cos(5.31523052413  + 3738.76143010800 *T)
    sumZ2  +=       0.136 * cos(1.17176449808  + 1349.86740965880 *T)
    sumZ2  +=       0.118 * cos(1.93283002344  + 6674.11130639880 *T)
    sumZ2  +=       0.114 * cos(5.86339475096  + 3341.59274776800 *T)
    sumZ2  +=       0.136 * cos(1.41161175400  + 6438.49624942560 *T)
    sumZ2  +=       0.109 * cos(5.09540016416  + 2957.71589447660 *T)
    sumZ2  +=       0.114 * cos(2.87540718990  + 242.72860397400 *T)
    sumZ2  +=       0.122 * cos(4.72332419375  + 2274.11694950980 *T)
    sumZ2  +=       0.116 * cos(5.94036127941  + 6836.64525283380 *T)
    sumZ2  +=       0.101 * cos(2.47526676901  + 951.71840625060 *T)
    sumZ2  +=       0.091 * cos(2.24799296656  + 7.11354700080 *T)
    sumZ2  +=       0.101 * cos(3.67031931828  + 553.56940284240 *T)
    sumZ2  +=       0.085 * cos(4.86490039033  + 796.29800681640 *T)
    sumZ2  +=       0.083 * cos(4.64956502978  + 4562.46099302120 *T)
    sumZ2  +=       0.095 * cos(3.01019726572  + 10025.36039844840 *T)
    sumZ2  +=       0.082 * cos(2.24784970600  + 213.29909543800 *T)
    sumZ2  +=       0.072 * cos(2.91435793605  + 23384.28698689860 *T)
    sumZ2  +=       0.070 * cos(5.55642055050  + 1066.49547719000 *T)
    sumZ2  +=       0.063 * cos(2.72606022638  + 8955.34180290940 *T)
    sumZ2  +=       0.058 * cos(4.22927578363  + 7210.91581849420 *T)
    sumZ2  +=       0.060 * cos(5.55658292278  + 6041.32756708560 *T)
    sumZ2  +=       0.071 * cos(3.27534690079  + 1589.07289528380 *T)
    sumZ2  +=       0.056 * cos(3.44816851943  + 3325.35995551480 *T)
    sumZ2  +=       0.063 * cos(3.79468594256  + 3870.30339179440 *T)
    sumZ2  +=       0.057 * cos(4.55390406368  + 4933.20844033260 *T)
    sumZ2  +=       0.049 * cos(4.29985057106  + 3355.86489788480 *T)
    sumZ2  +=       0.048 * cos(1.97226084372  + 10018.31416175040 *T)
    sumZ2  +=       0.047 * cos(5.89481794819  + 4535.05943692440 *T)
    sumZ2  +=       0.053 * cos(6.14813219827  + 3360.96774609859 *T)
    sumZ2  +=       0.045 * cos(3.79109648033  + 6923.95345737360 *T)
    sumZ2  +=       0.053 * cos(1.49899662719  + 3320.25710730100 *T)
    sumZ2  +=       0.047 * cos(0.28071441637  + 6127.65545055720 *T)
    sumZ2  +=       0.049 * cos(0.19353472631  + 3347.72597370060 *T)
    sumZ2  +=       0.058 * cos(5.89204904964  + 1052.26838318840 *T)
    sumZ2  +=       0.045 * cos(4.60130468759  + 4399.99435688900 *T)
    sumZ2  +=       0.038 * cos(3.96782009416  + 426.59819087600 *T)
    sumZ2  +=       0.040 * cos(1.57658451682  + 6489.77658728800 *T)
    sumZ2  +=       0.037 * cos(1.62198165732  + 5729.50644714900 *T)
    sumZ2  +=       0.041 * cos(1.99669605118  + 2699.73481931760 *T)
    sumZ2  +=       0.048 * cos(2.84634781523  + 3149.16416058820 *T)
    sumZ2  +=       0.036 * cos(5.09960798057  + 8432.76438481560 *T)
    sumZ2  +=       0.037 * cos(0.36652988837  + 6681.29216370240 *T)
    sumZ2  +=       0.039 * cos(5.13641267319  + 8969.56889691100 *T)
    sumZ2  +=       0.034 * cos(4.67496434823  + 1194.44701022460 *T)
    sumZ2  +=       0.043 * cos(3.02127238214  + 5331.35744374080 *T)
    sumZ2  +=       0.037 * cos(1.53879532057  + 191.44826611160 *T)
    sumZ2  +=       0.039 * cos(2.54138775081  + 2118.76386037840 *T)
    sumZ2  +=       0.032 * cos(1.49561091356  + 10419.98628350760 *T)
    sumZ2  +=       0.029 * cos(1.59835843850  + 7477.52286021600 *T)
    sumZ2  +=       0.031 * cos(0.15682384975  + 3205.54734666440 *T)
    sumZ2  +=       0.029 * cos(0.99282533197  + 522.57741809380 *T)
    sumZ2  +=       0.028 * cos(4.76871968197  + 382.89653222320 *T)
    sumZ2  +=       0.028 * cos(4.49319887552  + 9866.41688066520 *T)
    sumZ2  +=       0.029 * cos(5.20891382825  + 4136.91043351620 *T)
    sumZ2  +=       0.033 * cos(2.19423390657  + 6467.92575796160 *T)
    sumZ2  +=       0.028 * cos(6.20715072868  + 6682.20517446780 *T)
    sumZ2  +=       0.023 * cos(0.54135050421  + 640.87760738220 *T)
    sumZ2  +=       0.026 * cos(2.01536526590  + 9492.14631500480 *T)
    sumZ2  +=       0.024 * cos(5.95184917338  + 3361.38782219220 *T)
    sumZ2  +=       0.022 * cos(3.34551257447  + 13365.97282514820 *T)
    sumZ2  +=       0.024 * cos(1.43082762360  + 5614.72937620960 *T)
    sumZ2  +=       0.022 * cos(5.01088515186  + 10021.85453375160 *T)
    sumZ2  +=       0.023 * cos(0.39604919460  + 536.80451209540 *T)
    sumZ2  +=       0.020 * cos(4.75287585505  + 3553.91152213780 *T)
    sumZ2  +=       0.021 * cos(6.15131706326  + 7875.67186362420 *T)
    sumZ2  +=       0.019 * cos(3.85960262286  + 3319.83703120740 *T)
    sumZ2  +=       0.022 * cos(2.98009613873  + 10021.82002644720 *T)
    sumZ2  +=       0.021 * cos(5.09018656324  + 6681.24210705180 *T)
    sumZ2  +=       0.019 * cos(5.37557953087  + 9623.68827669120 *T)
    sumZ2  +=       0.021 * cos(3.05785674731  + 6681.20759974740 *T)
    sumZ2  +=       0.017 * cos(5.33111011611  + 6681.15754309680 *T)
    sumZ2  +=       0.016 * cos(0.88926952692  + 5628.95647021120 *T)
    sumZ2  +=       0.019 * cos(4.38572360867  + 6680.24453233140 *T)
    sumZ2  +=       0.015 * cos(3.04079800434  + 3178.14579056760 *T)
    sumZ2  +=       0.018 * cos(5.87932589084  + 7740.60678358880 *T)
    sumZ2  +=       0.014 * cos(4.19183102447  + 12303.06777661000 *T)
    sumZ2  +=       0.014 * cos(3.08467357092  + 3503.07906283200 *T)
    sumZ2  +=       0.014 * cos(4.02054439730  + 3316.73398895200 *T)
    sumZ2  +=       0.014 * cos(2.10530552076  + 3364.49086444760 *T)

    # MARS Z Coord
    sumZ3  = 0
    sumZ3  +=     897.035 * cos(1.98437168368  + 3340.61242669980 *T)
    sumZ3  +=      97.387 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ3  +=       8.489 * cos(1.73419562297  + 6681.22485339960 *T)
    sumZ3  +=       1.419 * cos(3.74979881372  + 3337.08930835080 *T)
    sumZ3  +=       0.928 * cos(3.63625528948  + 10021.83728009940 *T)
    sumZ3  +=       0.926 * cos(1.14567082339  + 3496.03282613400 *T)
    sumZ3  +=       0.286 * cos(0.11838177607  + 13362.44970679920 *T)
    sumZ3  +=       0.276 * cos(3.29100886638  + 3185.19202726560 *T)
    sumZ3  +=       0.190 * cos(2.55743259713  + 3097.88382272579 *T)
    sumZ3  +=       0.179 * cos(4.90901018351  + 2544.31441988340 *T)
    sumZ3  +=       0.186 * cos(4.74419057060  + 3532.06069281140 *T)
    sumZ3  +=       0.149 * cos(0.69534389042  + 16703.06213349900 *T)
    sumZ3  +=       0.133 * cos(6.04294120509  + 3894.18182954220 *T)
    sumZ3  +=       0.165 * cos(0.31646002710  + 5621.84292321040 *T)
    sumZ3  +=       0.176 * cos(3.40942833158  + 2942.46342329160 *T)
    sumZ3  +=       0.094 * cos(6.25738949900  + 3339.63210563160 *T)
    sumZ3  +=       0.083 * cos(2.80242797318  + 529.69096509460 *T)
    sumZ3  +=       0.079 * cos(4.65398630855  + 4292.33083295040 *T)
    sumZ3  +=       0.073 * cos(0.02372395336  + 2146.16541647520 *T)
    sumZ3  +=       0.069 * cos(1.90616132633  + 3583.34103067380 *T)
    sumZ3  +=       0.059 * cos(1.90076914463  + 5088.62883976680 *T)
    sumZ3  +=       0.057 * cos(0.91841960965  + 3723.50895892300 *T)
    sumZ3  +=       0.066 * cos(3.26000189494  + 4690.47983635860 *T)
    sumZ3  +=       0.049 * cos(6.24342258336  + 2281.23049651060 *T)
    sumZ3  +=       0.040 * cos(0.33925158823  + 5486.77784317500 *T)
    sumZ3  +=       0.038 * cos(0.33382735668  + 155.42039943420 *T)
    sumZ3  +=       0.037 * cos(5.38905349684  + 3340.59517304760 *T)
    sumZ3  +=       0.043 * cos(3.67789144879  + 6525.80445396540 *T)
    sumZ3  +=       0.037 * cos(1.13663982116  + 3340.62968035200 *T)
    sumZ3  +=       0.034 * cos(0.54674814571  + 1059.38193018920 *T)
    sumZ3  +=       0.033 * cos(3.70799920482  + 6677.70173505060 *T)
    sumZ3  +=       0.042 * cos(1.06901082266  + 20043.67456019880 *T)
    sumZ3  +=       0.040 * cos(5.18845744550  + 5884.92684658320 *T)
    sumZ3  +=       0.028 * cos(5.53170427847  + 3333.49887969900 *T)
    sumZ3  +=       0.027 * cos(4.74965318878  + 2787.04302385740 *T)
    sumZ3  +=       0.032 * cos(2.94765501590  + 398.14900340820 *T)
    sumZ3  +=       0.026 * cos(0.80891817293  + 1194.44701022460 *T)
    sumZ3  +=       0.020 * cos(5.71161307356  + 5092.15195811580 *T)
    sumZ3  +=       0.014 * cos(0.59214435004  + 4399.99435688900 *T)
    sumZ3  +=       0.012 * cos(0.57596701083  + 1748.01641306700 *T)
    sumZ3  +=       0.015 * cos(5.61526640204  + 7079.37385680780 *T)
    sumZ3  +=       0.014 * cos(4.18010272073  + 6151.53388830500 *T)
    sumZ3  +=       0.012 * cos(5.88319051616  + 1592.59601363280 *T)
    sumZ3  +=       0.012 * cos(3.74719789330  + 242.72860397400 *T)
    sumZ3  +=       0.011 * cos(3.77687231272  + 2274.11694950980 *T)
    sumZ3  +=       0.011 * cos(3.28827114872  + 3738.76143010800 *T)
    sumZ3  +=       0.011 * cos(6.10972583214  + 23384.28698689860 *T)
    sumZ3  +=       0.010 * cos(6.20023805343  + 2288.34404351140 *T)
    sumZ3  +=       0.011 * cos(1.61096472294  + 3344.13554504880 *T)
    sumZ3  +=       0.009 * cos(2.89522242682  + 553.56940284240 *T)
    sumZ3  +=       0.011 * cos(4.21212605505  + 6489.77658728800 *T)
    sumZ3  +=       0.008 * cos(2.95652295638  + 3127.31333126180 *T)
    sumZ3  +=       0.007 * cos(0.49773686048  + 6684.74797174860 *T)
    sumZ3  +=       0.007 * cos(5.43009944036  + 4933.20844033260 *T)
    sumZ3  +=       0.007 * cos(2.19921948725  + 6438.49624942560 *T)
    sumZ3  +=       0.009 * cos(3.04008808528  + 7.11354700080 *T)

    # MARS Z Coord
    sumZ4  = 0
    sumZ4  +=      30.554 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ4  +=      19.442 * cos(3.44959674542  + 3340.61242669980 *T)
    sumZ4  +=       0.371 * cos(0.36613555151  + 6681.22485339960 *T)
    sumZ4  +=       0.155 * cos(2.66185835593  + 3496.03282613400 *T)
    sumZ4  +=       0.081 * cos(5.27888218929  + 3337.08930835080 *T)
    sumZ4  +=       0.073 * cos(1.31158212762  + 10021.83728009940 *T)
    sumZ4  +=       0.034 * cos(4.09971603267  + 3097.88382272579 *T)
    sumZ4  +=       0.033 * cos(1.76655964001  + 3185.19202726560 *T)
    sumZ4  +=       0.015 * cos(0.95989595600  + 3894.18182954220 *T)
    sumZ4  +=       0.009 * cos(0.31833513882  + 3583.34103067380 *T)
    sumZ4  +=       0.006 * cos(3.76221637426  + 6684.74797174860 *T)
    sumZ4  +=       0.006 * cos(1.44439054717  + 2787.04302385740 *T)
    sumZ4  +=       0.008 * cos(1.39569739639  + 3344.13554504880 *T)
    sumZ4  +=       0.006 * cos(5.83516395431  + 3532.06069281140 *T)
    sumZ4  +=       0.006 * cos(5.97538779583  + 2544.31441988340 *T)
    sumZ4  +=       0.005 * cos(5.41788972302  + 16703.06213349900 *T)
    sumZ4  +=       0.006 * cos(1.96708787649  + 6525.80445396540 *T)
    sumZ4  +=       0.004 * cos(5.61196522691  + 4292.33083295040 *T)
    sumZ4  +=       0.004 * cos(4.54689889636  + 2942.46342329160 *T)
    sumZ4  +=       0.005 * cos(2.79859082923  + 13362.44970679920 *T)

    # MARS Z Coord
    sumZ5  = 0
    sumZ5  +=       0.597 * cos(4.83347154755  + 3340.61242669980 *T)
    sumZ5  +=       0.028 * cos(5.25340658633  + 6681.22485339960 *T)
    sumZ5  +=       0.021 * cos(4.17736925293  + 3496.03282613400 *T)
    sumZ5  +=       0.011 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ5  +=       0.006 * cos(5.72895319056  + 10021.83728009940 *T)
    sumZ5  +=       0.002 * cos(0.08188351604  + 16703.06213349900 *T)
    sumZ5  +=       0.003 * cos(0.40135650339  + 3185.19202726560 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Mars",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getMarsData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getMarsXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getMercuryXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Mercury"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getMercuryXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getMercuryXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # MERCURY X Coord
    sumX0  = 0
    sumX0  +=  37749277.893 * cos(4.40259139579  + 26088.14695905770 *T)
    sumX0  +=  11918926.148 * cos(4.49027758439  + 0.24381748350 *T)
    sumX0  +=  3840153.904 * cos(1.17015646101  + 52176.05010063190 *T)
    sumX0  +=  585979.278 * cos(4.22090402969  + 78263.95324220609 *T)
    sumX0  +=  305833.424 * cos(2.10298673336  + 26087.65932409069 *T)
    sumX0  +=  105974.941 * cos(0.98846517420  + 104351.85638378028 *T)
    sumX0  +=   24906.132 * cos(5.26305668971  + 52175.56246566489 *T)
    sumX0  +=   21056.065 * cos(4.03921104994  + 130439.75952535449 *T)
    sumX0  +=    4441.671 * cos(0.80677167456  + 156527.66266692869 *T)
    sumX0  +=    3347.442 * cos(2.10873209401  + 78263.46560723908 *T)
    sumX0  +=     976.617 * cos(3.85751708019  + 182615.56580850287 *T)
    sumX0  +=     703.903 * cos(4.46040321768  + 24978.76840696430 *T)
    sumX0  +=     640.528 * cos(1.20142454428  + 27197.52551115110 *T)
    sumX0  +=     558.149 * cos(5.21749759764  + 104351.36874881329 *T)
    sumX0  +=     447.775 * cos(2.97591075236  + 1059.62574767270 *T)
    sumX0  +=     400.281 * cos(1.86944903196  + 20426.81490990550 *T)
    sumX0  +=     283.882 * cos(3.79350724209  + 31749.47900820990 *T)
    sumX0  +=     191.007 * cos(4.27537754512  + 53285.42865272530 *T)
    sumX0  +=     181.454 * cos(4.95018365158  + 1109.13473460990 *T)
    sumX0  +=     193.895 * cos(0.67633415224  + 4552.19731454230 *T)
    sumX0  +=     221.371 * cos(0.62507677073  + 208703.46895007708 *T)
    sumX0  +=     139.001 * cos(1.22954955274  + 51066.67154853850 *T)
    sumX0  +=     183.933 * cos(1.19500306041  + 5661.08823166870 *T)
    sumX0  +=     153.303 * cos(2.68451868547  + 51116.66817044270 *T)
    sumX0  +=     104.855 * cos(2.02959651352  + 130439.27189038748 *T)
    sumX0  +=     115.202 * cos(0.56691147306  + 57837.38214978410 *T)
    sumX0  +=      90.628 * cos(0.25792152823  + 529.93478257810 *T)
    sumX0  +=      75.318 * cos(6.07215712399  + 27147.52888924690 *T)
    sumX0  +=      72.957 * cos(5.00412566043  + 21536.19346199890 *T)
    sumX0  +=      74.803 * cos(4.92914440147  + 46514.71805147970 *T)
    sumX0  +=      68.710 * cos(0.20379815554  + 25132.54721744910 *T)
    sumX0  +=      83.504 * cos(3.17939445535  + 10213.52936369450 *T)
    sumX0  +=      60.985 * cos(5.45128332631  + 27043.74670066630 *T)
    sumX0  +=      57.987 * cos(5.70479654236  + 14765.48286075331 *T)
    sumX0  +=      60.953 * cos(5.00244327341  + 47624.09660357310 *T)
    sumX0  +=      53.356 * cos(2.29756051926  + 26617.83792415230 *T)
    sumX0  +=      54.964 * cos(5.31144278536  + 12566.39551746630 *T)
    sumX0  +=      47.881 * cos(1.04626983477  + 79373.33179429949 *T)
    sumX0  +=      51.263 * cos(3.53744966658  + 426.84200835950 *T)
    sumX0  +=      51.361 * cos(3.67582188055  + 234791.37209165128 *T)
    sumX0  +=      50.374 * cos(3.60286953784  + 25558.45599396310 *T)
    sumX0  +=      48.449 * cos(3.51993896091  + 1589.31671276730 *T)
    sumX0  +=      48.867 * cos(5.73548681205  + 77204.57131201689 *T)
    sumX0  +=      34.739 * cos(5.68510489030  + 51646.35913553730 *T)
    sumX0  +=      32.851 * cos(2.99883640297  + 955.35592412510 *T)
    sumX0  +=      44.998 * cos(2.48247314947  + 41962.76455442090 *T)
    sumX0  +=      31.618 * cos(6.23993881626  + 37410.81105736210 *T)
    sumX0  +=      28.634 * cos(0.71509820566  + 30640.10045611650 *T)
    sumX0  +=      31.480 * cos(4.28075443320  + 77154.57469011268 *T)
    sumX0  +=      32.102 * cos(3.61877734434  + 83925.28529135829 *T)
    sumX0  +=      27.670 * cos(0.34889366405  + 39609.89840064910 *T)
    sumX0  +=      26.872 * cos(6.26000952694  + 36301.43250526870 *T)
    sumX0  +=      23.054 * cos(2.29052909388  + 53131.64984224050 *T)
    sumX0  +=      26.193 * cos(3.69823179514  + 11322.42028082089 *T)
    sumX0  +=      21.235 * cos(5.11548231564  + 156527.17503196167 *T)
    sumX0  +=      26.946 * cos(1.76252593609  + 6283.31966747490 *T)
    sumX0  +=      20.875 * cos(5.66887296875  + 15874.86141284670 *T)
    sumX0  +=      16.454 * cos(1.69895735267  + 72602.62119305388 *T)
    sumX0  +=      16.101 * cos(3.22620073160  + 9104.15081160110 *T)
    sumX0  +=      22.241 * cos(4.03643224642  + 13521.50762410789 *T)
    sumX0  +=      21.188 * cos(1.77076461428  + 73711.99974514729 *T)
    sumX0  +=      15.123 * cos(2.81544648697  + 18849.47136745770 *T)
    sumX0  +=      17.255 * cos(4.95346284309  + 27197.03787618409 *T)
    sumX0  +=      17.287 * cos(2.12451701243  + 51749.45190975589 *T)
    sumX0  +=      13.862 * cos(2.86403540216  + 53235.43203082109 *T)
    sumX0  +=      15.239 * cos(0.39591934360  + 7238.43177411650 *T)
    sumX0  +=      16.863 * cos(2.14228359969  + 50586.97720534809 *T)
    sumX0  +=      13.546 * cos(5.21986559185  + 52705.74106572650 *T)
    sumX0  +=      13.134 * cos(5.78471170881  + 25028.76502886850 *T)
    sumX0  +=      14.145 * cos(3.03635370573  + 63498.71419893629 *T)
    sumX0  +=      13.081 * cos(1.77889761864  + 26068.47719815791 *T)
    sumX0  +=      13.058 * cos(3.88464356096  + 26107.81671995749 *T)
    sumX0  +=      13.609 * cos(3.25855054515  + 51220.45035902330 *T)
    sumX0  +=      13.539 * cos(2.50293964820  + 103292.47445359109 *T)
    sumX0  +=      12.138 * cos(0.44338161953  + 260879.27523322543 *T)
    sumX0  +=      11.433 * cos(1.98105303514  + 19317.43635781211 *T)
    sumX0  +=      10.174 * cos(2.39402085953  + 77734.26227711148 *T)
    sumX0  +=      12.016 * cos(5.53519559373  + 68050.66769599509 *T)
    sumX0  +=      12.094 * cos(4.09914561799  + 105461.23493587368 *T)
    sumX0  +=       9.215 * cos(2.48225447254  + 40853.38600232750 *T)
    sumX0  +=       9.157 * cos(2.82670347755  + 33326.82255065770 *T)
    sumX0  +=       9.552 * cos(2.11034031277  + 38654.29865904050 *T)
    sumX0  +=      12.500 * cos(3.89244437690  + 45892.97425064050 *T)
    sumX0  +=       8.978 * cos(0.47617343198  + 26084.26562369970 *T)
    sumX0  +=       8.975 * cos(5.18743195952  + 26092.02829441570 *T)
    sumX0  +=      12.160 * cos(3.40664771625  + 65697.80154222329 *T)
    sumX0  +=       8.880 * cos(0.33806954264  + 26514.74514993370 *T)
    sumX0  +=       8.892 * cos(0.74770854519  + 3442.81876244891 *T)
    sumX0  +=       8.503 * cos(0.38676770328  + 110013.18843293248 *T)
    sumX0  +=      10.244 * cos(2.50326457964  + 21535.70582703189 *T)
    sumX0  +=       8.816 * cos(4.86930066154  + 32371.22280904910 *T)
    sumX0  +=       9.705 * cos(5.44815918458  + 529.44714761109 *T)
    sumX0  +=       7.663 * cos(3.73484444482  + 56728.00359769070 *T)
    sumX0  +=       7.852 * cos(0.65554683317  + 5661.57586663570 *T)
    sumX0  +=       8.166 * cos(0.32786623322  + 27677.21985434150 *T)
    sumX0  +=       7.299 * cos(1.14994821071  + 6770.46678376209 *T)
    sumX0  +=       7.437 * cos(3.67894554888  + 32858.85756030330 *T)
    sumX0  +=       7.796 * cos(1.18993742629  + 31748.99137324289 *T)
    sumX0  +=       6.786 * cos(2.43978460253  + 43072.14310651430 *T)
    sumX0  +=       7.408 * cos(4.54789639858  + 23869.38985487090 *T)
    sumX0  +=       5.938 * cos(5.34593352715  + 79219.55298381469 *T)
    sumX0  +=       7.553 * cos(1.04863475066  + 103242.47783168688 *T)
    sumX0  +=       5.988 * cos(3.04417882527  + 62389.33564684289 *T)
    sumX0  +=       6.154 * cos(5.99522472277  + 59414.72569223190 *T)
    sumX0  +=       6.213 * cos(0.75145729813  + 19805.07110906630 *T)
    sumX0  +=       6.507 * cos(1.06344526474  + 213.54291292150 *T)
    sumX0  +=       6.284 * cos(4.91987886888  + 22645.57201409229 *T)
    sumX0  +=       6.068 * cos(4.82277476434  + 99799.90288672148 *T)
    sumX0  +=       6.276 * cos(2.74748364541  + 640.14110379750 *T)
    sumX0  +=       6.192 * cos(1.11265105508  + 28306.90406324449 *T)
    sumX0  +=       6.081 * cos(5.23278439947  + 25874.84786361970 *T)
    sumX0  +=       4.693 * cos(6.19313162709  + 16983.75232997309 *T)
    sumX0  +=       4.527 * cos(1.91140424398  + 182615.07817353585 *T)
    sumX0  +=       5.085 * cos(0.33812112463  + 26301.44605449570 *T)
    sumX0  +=       4.658 * cos(4.11167846402  + 26095.26050605850 *T)
    sumX0  +=       5.323 * cos(2.15023457135  + 24978.28077199729 *T)
    sumX0  +=       4.497 * cos(1.59603827995  + 26081.03341205689 *T)
    sumX0  +=       5.479 * cos(5.17786692376  + 76674.88034692229 *T)
    sumX0  +=       3.898 * cos(4.75161525594  + 98690.52433462808 *T)
    sumX0  +=       5.352 * cos(5.17556723587  + 77837.35505133009 *T)
    sumX0  +=       3.694 * cos(0.48936321493  + 58946.76070187749 *T)
    sumX0  +=       3.631 * cos(5.55357343180  + 129380.37759516528 *T)
    sumX0  +=       4.821 * cos(3.64928863412  + 37410.32342239509 *T)
    sumX0  +=       3.323 * cos(1.95435021777  + 78793.64420730069 *T)
    sumX0  +=       3.932 * cos(0.76247078505  + 29530.72190402310 *T)
    sumX0  +=       4.022 * cos(4.06373546409  + 2119.00767786191 *T)
    sumX0  +=       3.208 * cos(5.89844713322  + 25557.96835899609 *T)
    sumX0  +=       3.611 * cos(0.17577280462  + 91785.70468379748 *T)
    sumX0  +=       3.062 * cos(5.92483558903  + 79323.33517239528 *T)
    sumX0  +=       3.480 * cos(0.66265476262  + 71980.87739221469 *T)
    sumX0  +=       3.093 * cos(2.30327965785  + 94138.57083756928 *T)
    sumX0  +=       3.172 * cos(3.00172025362  + 27043.25906569930 *T)
    sumX0  +=       2.912 * cos(3.49412631679  + 286967.17837479961 *T)
    sumX0  +=       3.746 * cos(6.08415745758  + 89586.61734051049 *T)
    sumX0  +=       2.870 * cos(5.51397391523  + 69160.04624808849 *T)
    sumX0  +=       2.730 * cos(5.88663339742  + 44937.37450903190 *T)
    sumX0  +=       2.725 * cos(5.42457603553  + 103822.16541868568 *T)
    sumX0  +=       3.214 * cos(4.87298940652  + 2218.51328670329 *T)
    sumX0  +=       2.660 * cos(4.82966654366  + 52156.38033973211 *T)
    sumX0  +=       3.257 * cos(2.71278189877  + 25934.36814857290 *T)
    sumX0  +=       2.620 * cos(5.77949301476  + 20426.32727493849 *T)
    sumX0  +=       3.227 * cos(2.95128548648  + 26241.92576954250 *T)
    sumX0  +=       3.086 * cos(0.02730531846  + 77308.35350059748 *T)
    sumX0  +=       3.099 * cos(0.86870364149  + 131549.13807744789 *T)
    sumX0  +=       2.731 * cos(0.19114675485  + 149.80701461810 *T)
    sumX0  +=       2.486 * cos(3.12136440583  + 1052.51220067191 *T)
    sumX0  +=       3.167 * cos(0.38944502659  + 33326.33491569069 *T)
    sumX0  +=       2.680 * cos(0.65295917827  + 52195.71986153169 *T)
    sumX0  +=       2.231 * cos(3.43792316659  + 136101.09157450669 *T)
    sumX0  +=       2.653 * cos(1.68316051523  + 7.35736448430 *T)
    sumX0  +=       2.074 * cos(4.96316067320  + 48733.47515566649 *T)
    sumX0  +=       1.974 * cos(5.16962188890  + 64742.20180061470 *T)
    sumX0  +=       2.064 * cos(0.21494181057  + 51116.18053547569 *T)
    sumX0  +=       2.533 * cos(6.02031593619  + 1066.73929467350 *T)
    sumX0  +=       1.949 * cos(1.68714181050  + 58459.12595062330 *T)
    sumX0  +=       2.010 * cos(2.44539246304  + 16984.23996494010 *T)
    sumX0  +=       1.876 * cos(0.49249036640  + 82815.90673926489 *T)
    sumX0  +=       1.832 * cos(1.95461512625  + 52179.93143598990 *T)
    sumX0  +=       2.306 * cos(4.20103483106  + 54394.80720481869 *T)
    sumX0  +=       1.837 * cos(5.53329054515  + 66941.28914390170 *T)
    sumX0  +=       1.783 * cos(1.94386918249  + 25620.18196870330 *T)
    sumX0  +=       1.826 * cos(3.52694014247  + 52172.16876527390 *T)
    sumX0  +=       1.791 * cos(5.32229718021  + 25661.54876818170 *T)
    sumX0  +=       2.126 * cos(2.91445652095  + 51536.15281431789 *T)
    sumX0  +=       1.714 * cos(3.72047839107  + 26556.11194941211 *T)
    sumX0  +=       1.972 * cos(5.03869626311  + 45405.33949938630 *T)
    sumX0  +=       1.711 * cos(1.11320004200  + 32858.36992533629 *T)
    sumX0  +=       1.643 * cos(3.40283361772  + 52602.64829150790 *T)
    sumX0  +=       1.665 * cos(3.83085129266  + 5327.71992586630 *T)
    sumX0  +=       1.930 * cos(5.78785787875  + 13656.10430865991 *T)
    sumX0  +=       1.655 * cos(1.59098456165  + 125887.80602829569 *T)
    sumX0  +=       1.554 * cos(3.40019334049  + 13521.99525907490 *T)
    sumX0  +=       1.517 * cos(3.39842653971  + 53765.12299591570 *T)
    sumX0  +=       1.870 * cos(4.09951350071  + 129330.38097326108 *T)
    sumX0  +=       1.705 * cos(6.14540598785  + 43071.65547154729 *T)
    sumX0  +=       1.683 * cos(4.32584442162  + 52168.93655363109 *T)
    sumX0  +=       1.449 * cos(6.10448204410  + 88477.23878841709 *T)
    sumX0  +=       1.466 * cos(5.27211735766  + 24499.07406377390 *T)
    sumX0  +=       1.655 * cos(6.12815645892  + 76.50988875911 *T)
    sumX0  +=       1.539 * cos(2.12176953420  + 105307.45612538888 *T)
    sumX0  +=       1.542 * cos(1.96595530367  + 102762.78348849648 *T)
    sumX0  +=       1.400 * cos(1.26185265808  + 26081.51840015769 *T)
    sumX0  +=       1.400 * cos(4.40173881619  + 26094.77551795771 *T)
    sumX0  +=       1.415 * cos(1.31785110641  + 49957.29299644510 *T)
    sumX0  +=       1.545 * cos(2.74460517130  + 85502.62883380610 *T)
    sumX0  +=       1.259 * cos(4.06949140437  + 51962.75100519390 *T)
    sumX0  +=       1.464 * cos(1.94328769190  + 103925.25819290428 *T)
    sumX0  +=       1.265 * cos(6.23687279994  + 35192.05395317530 *T)
    sumX0  +=       1.563 * cos(1.53248979312  + 6681.46867088311 *T)
    sumX0  +=       1.517 * cos(4.14158249910  + 213.05527795450 *T)
    sumX0  +=       1.171 * cos(4.44560816323  + 57836.89451481709 *T)
    sumX0  +=       1.547 * cos(3.43699337530  + 52389.34919606990 *T)
    sumX0  +=       1.423 * cos(1.60092130514  + 50057.28624025350 *T)
    sumX0  +=       1.238 * cos(0.84389836177  + 52183.16364763270 *T)
    sumX0  +=       1.070 * cos(5.82812348558  + 26728.04424537170 *T)
    sumX0  +=       1.020 * cos(3.63373915432  + 12431.79883291429 *T)
    sumX0  +=       0.965 * cos(2.32109036696  + 155468.28073673949 *T)
    sumX0  +=       1.018 * cos(5.37034692069  + 24498.58642880689 *T)
    sumX0  +=       0.994 * cos(3.22732624356  + 117873.60782537168 *T)
    sumX0  +=       1.002 * cos(4.98556762896  + 208702.98131511008 *T)
    sumX0  +=       1.214 * cos(2.58462840069  + 25661.06113321469 *T)
    sumX0  +=       0.958 * cos(1.52001998159  + 124778.42747620228 *T)
    sumX0  +=       0.893 * cos(3.52720868596  + 85034.66384345168 *T)
    sumX0  +=       0.875 * cos(1.72393184919  + 74821.37829724069 *T)
    sumX0  +=       0.871 * cos(0.27938200831  + 1059.13811270570 *T)
    sumX0  +=       0.910 * cos(3.71329662418  + 98068.78053378889 *T)
    sumX0  +=       0.827 * cos(2.44624289116  + 22645.08437912529 *T)
    sumX0  +=       0.829 * cos(4.98909621401  + 104881.54734887488 *T)
    sumX0  +=       0.870 * cos(5.31437509457  + 51066.18391357149 *T)
    sumX0  +=       0.978 * cos(6.15736380216  + 38520.18960945550 *T)
    sumX0  +=       0.915 * cos(1.34262350391  + 6.86972951729 *T)
    sumX0  +=       0.855 * cos(2.53346806421  + 51123.78171744350 *T)
    sumX0  +=       0.795 * cos(5.35307581508  + 120226.47397914348 *T)
    sumX0  +=       0.778 * cos(6.26214759185  + 11322.90791578790 *T)
    sumX0  +=       0.913 * cos(5.46819895663  + 52026.48690349730 *T)
    sumX0  +=       0.930 * cos(2.84523614939  + 115674.52048208468 *T)
    sumX0  +=       0.932 * cos(3.80512706428  + 55618.62504559730 *T)
    sumX0  +=       0.707 * cos(0.26168602915  + 313055.08151637384 *T)
    sumX0  +=       0.803 * cos(3.92107317901  + 157637.04121902207 *T)
    sumX0  +=       0.730 * cos(2.20347820113  + 53284.94101775829 *T)
    sumX0  +=       0.725 * cos(2.69780707537  + 105411.23831396949 *T)
    sumX0  +=       0.788 * cos(5.97840774266  + 3340.85624418330 *T)
    sumX0  +=       0.867 * cos(5.92376156240  + 51109.55462344189 *T)
    sumX0  +=       0.718 * cos(2.18164724460  + 129910.06856025988 *T)
    sumX0  +=       0.688 * cos(0.61823279820  + 419.72846135871 *T)
    sumX0  +=       0.710 * cos(2.26735831674  + 95247.94938966268 *T)
    sumX0  +=       0.694 * cos(5.40455859489  + 26617.35028918529 *T)
    sumX0  +=       0.621 * cos(1.24687546366  + 4551.70967957530 *T)
    sumX0  +=       0.661 * cos(4.35118955440  + 10213.04172872750 *T)
    sumX0  +=       0.674 * cos(0.86923628110  + 28206.91081943610 *T)
    sumX0  +=       0.609 * cos(1.59724799045  + 78244.28348130629 *T)
    sumX0  +=       0.661 * cos(5.76395663117  + 52022.27129014710 *T)
    sumX0  +=       0.741 * cos(3.07875547199  + 103396.25664217168 *T)
    sumX0  +=       0.585 * cos(0.20603891239  + 162188.99471608087 *T)
    sumX0  +=       0.599 * cos(1.83731791923  + 46848.57399224910 *T)
    sumX0  +=       0.634 * cos(4.05934263309  + 39609.41076568209 *T)
    sumX0  +=       0.731 * cos(3.42015997539  + 433.95555536030 *T)
    sumX0  +=       0.591 * cos(2.65948363134  + 71025.27765060609 *T)
    sumX0  +=       0.576 * cos(2.53160242412  + 47623.60896860609 *T)
    sumX0  +=       0.545 * cos(2.34985134231  + 48732.98752069949 *T)
    sumX0  +=       0.528 * cos(4.82834838426  + 23754.95056618569 *T)
    sumX0  +=       0.531 * cos(6.23166694924  + 41962.27691945389 *T)
    sumX0  +=       0.525 * cos(2.22737583220  + 10022.08109758290 *T)
    sumX0  +=       0.578 * cos(0.96705890639  + 80482.71034639288 *T)
    sumX0  +=       0.675 * cos(6.00003515273  + 52329.82891111670 *T)
    sumX0  +=       0.664 * cos(5.96665839431  + 77624.05595589208 *T)
    sumX0  +=       0.541 * cos(2.93745553127  + 6282.83203250789 *T)
    sumX0  +=       0.639 * cos(4.14642362107  + 45494.82524723229 *T)
    sumX0  +=       0.594 * cos(3.31128236497  + 7994.77225950771 *T)
    sumX0  +=       0.467 * cos(4.75719880131  + 84547.02909219750 *T)
    sumX0  +=       0.552 * cos(5.81449540251  + 52099.78402935630 *T)
    sumX0  +=       0.490 * cos(2.96490072888  + 64608.09275102969 *T)
    sumX0  +=       0.553 * cos(2.39742743080  + 1692.40948698591 *T)
    sumX0  +=       0.456 * cos(1.94416761157  + 90830.10494218887 *T)
    sumX0  +=       0.470 * cos(3.29329750332  + 26237.71015619230 *T)
    sumX0  +=       0.501 * cos(4.18799339117  + 25132.05958248210 *T)
    sumX0  +=       0.615 * cos(3.70395803529  + 78283.62300310588 *T)
    sumX0  +=       0.441 * cos(6.26869790938  + 27140.41534224609 *T)
    sumX0  +=       0.445 * cos(4.64236190444  + 151975.70916986989 *T)
    sumX0  +=       0.437 * cos(5.66855783361  + 20760.18321570790 *T)
    sumX0  +=       0.472 * cos(0.86711075193  + 155418.28411483529 *T)
    sumX0  +=       0.428 * cos(3.39576463667  + 77204.08367704989 *T)
    sumX0  +=       0.469 * cos(3.54106330325  + 108903.80988083908 *T)
    sumX0  +=       0.478 * cos(0.48558080454  + 63498.22656396929 *T)
    sumX0  +=       0.448 * cos(1.04430189969  + 78256.83969520529 *T)
    sumX0  +=       0.591 * cos(1.93028393360  + 853.44019923550 *T)
    sumX0  +=       0.438 * cos(3.00642339238  + 51645.87150057029 *T)
    sumX0  +=       0.440 * cos(4.84626648414  + 28306.41642827749 *T)
    sumX0  +=       0.413 * cos(3.60129277287  + 38519.70197448849 *T)
    sumX0  +=       0.456 * cos(2.01311166346  + 26038.15033715350 *T)
    sumX0  +=       0.456 * cos(3.65054925615  + 26138.14358096190 *T)
    sumX0  +=       0.470 * cos(1.08701476541  + 58946.27306691049 *T)
    sumX0  +=       0.434 * cos(2.31499427060  + 29428.75938575750 *T)
    sumX0  +=       0.439 * cos(0.83378397651  + 2333.44021035551 *T)
    sumX0  +=       0.406 * cos(2.57203150650  + 1582.20316576650 *T)
    sumX0  +=       0.419 * cos(5.03633652112  + 128850.68663007068 *T)
    sumX0  +=       0.403 * cos(0.84418373392  + 28421.34335192970 *T)
    sumX0  +=       0.423 * cos(4.33257670397  + 50586.48957038109 *T)
    sumX0  +=       0.420 * cos(5.00524088112  + 78267.83457756409 *T)
    sumX0  +=       0.508 * cos(0.82135338502  + 26720.93069837089 *T)
    sumX0  +=       0.445 * cos(4.23188832059  + 1109.62236957690 *T)
    sumX0  +=       0.452 * cos(5.82726780405  + 25985.05418483910 *T)
    sumX0  +=       0.440 * cos(1.02799745863  + 78050.65414676809 *T)
    sumX0  +=       0.427 * cos(2.83364183360  + 27154.64243624770 *T)
    sumX0  +=       0.422 * cos(2.30681425357  + 93029.19228547587 *T)
    sumX0  +=       0.387 * cos(3.49964872741  + 22747.53453235790 *T)
    sumX0  +=       0.451 * cos(6.18526475915  + 26191.23973327629 *T)
    sumX0  +=       0.525 * cos(4.94137283297  + 25455.36321974450 *T)
    sumX0  +=       0.439 * cos(2.68207148004  + 46514.23041651269 *T)
    sumX0  +=       0.390 * cos(4.99420452598  + 130013.16133447849 *T)
    sumX0  +=       0.364 * cos(0.17463222983  + 78690.55143308209 *T)
    sumX0  +=       0.468 * cos(4.61226407005  + 76145.18938182769 *T)
    sumX0  +=       0.361 * cos(2.87164735708  + 114565.14192999128 *T)
    sumX0  +=       0.354 * cos(0.48572193966  + 52644.01509098630 *T)
    sumX0  +=       0.418 * cos(0.29451759125  + 78260.07190684808 *T)
    sumX0  +=       0.366 * cos(0.57462664693  + 25028.27739390149 *T)
    sumX0  +=       0.394 * cos(0.21360682748  + 78477.25233764409 *T)
    sumX0  +=       0.337 * cos(0.17366623994  + 79853.02613748988 *T)
    sumX0  +=       0.356 * cos(5.33654783666  + 633.02755679670 *T)
    sumX0  +=       0.427 * cos(4.30697428113  + 19804.58347409930 *T)
    sumX0  +=       0.360 * cos(6.08031778998  + 48835.43767393209 *T)
    sumX0  +=       0.431 * cos(4.76277333184  + 30639.61282114949 *T)
    sumX0  +=       0.336 * cos(2.07023419375  + 18208.05780571871 *T)
    sumX0  +=       0.317 * cos(4.36862743851  + 76045.19613801930 *T)
    sumX0  +=       0.403 * cos(5.18083081721  + 131395.35926696309 *T)
    sumX0  +=       0.300 * cos(4.63629291392  + 25973.70767037247 *T)
    sumX0  +=       0.365 * cos(5.56868437730  + 25565.56954096390 *T)
    sumX0  +=       0.417 * cos(1.80624647547  + 71493.24264096050 *T)
    sumX0  +=       0.298 * cos(1.02394718540  + 26202.58624774291 *T)
    sumX0  +=       0.411 * cos(0.25672822810  + 1596.43025976811 *T)
    sumX0  +=       0.286 * cos(1.16931638049  + 52182.67865953191 *T)
    sumX0  +=       0.304 * cos(4.98265595281  + 51708.08511027749 *T)
    sumX0  +=       0.288 * cos(2.94996952897  + 26164.41303033330 *T)
    sumX0  +=       0.326 * cos(0.06254421557  + 26610.72437715149 *T)
    sumX0  +=       0.375 * cos(3.57829900047  + 26162.92855762500 *T)
    sumX0  +=       0.268 * cos(6.27766671307  + 143961.51096694588 *T)
    sumX0  +=       0.367 * cos(5.77470435266  + 111590.53197538028 *T)
    sumX0  +=       0.264 * cos(4.93317721296  + 72936.47713382330 *T)
    sumX0  +=       0.301 * cos(2.58808348930  + 224.58861318540 *T)
    sumX0  +=       0.286 * cos(5.68119587232  + 46848.08635728210 *T)
    sumX0  +=       0.257 * cos(5.37176115578  + 181556.18387831369 *T)
    sumX0  +=       0.352 * cos(2.05335376553  + 26013.36536049040 *T)
    sumX0  +=       0.274 * cos(1.10717834982  + 956.53297345411 *T)
    sumX0  +=       0.253 * cos(0.21818001465  + 75.02541605080 *T)
    sumX0  +=       0.256 * cos(2.67592097640  + 537.04832957890 *T)
    sumX0  +=       0.245 * cos(4.84007767717  + 51756.56545675670 *T)
    sumX0  +=       0.286 * cos(4.31029470055  + 52169.42154173189 *T)
    sumX0  +=       0.326 * cos(1.57460855395  + 25771.75508940110 *T)
    sumX0  +=       0.252 * cos(0.65486372524  + 25668.66231518250 *T)
    sumX0  +=       0.239 * cos(3.45236126203  + 42154.21282053250 *T)
    sumX0  +=       0.324 * cos(4.09374409332  + 26404.53882871430 *T)
    sumX0  +=       0.285 * cos(1.43554886267  + 1162.71852189130 *T)
    sumX0  +=       0.237 * cos(0.47964640927  + 124156.68367536308 *T)
    sumX0  +=       0.271 * cos(1.47052171454  + 15874.37377787969 *T)
    sumX0  +=       0.307 * cos(3.88495459836  + 78271.06678920689 *T)
    sumX0  +=       0.323 * cos(1.43180313540  + 19406.43447069109 *T)
    sumX0  +=       0.245 * cos(0.30807019834  + 111122.56698502589 *T)
    sumX0  +=       0.229 * cos(3.58086611266  + 36300.94487030169 *T)
    sumX0  +=       0.243 * cos(5.83259460773  + 7238.91940908350 *T)
    sumX0  +=       0.242 * cos(5.89612108076  + 51748.96427478889 *T)
    sumX0  +=       0.275 * cos(5.58726658873  + 77211.68485901768 *T)
    sumX0  +=       0.234 * cos(1.35631291904  + 83924.79765639128 *T)
    sumX0  +=       0.279 * cos(2.69169755313  + 77197.45776501608 *T)
    sumX0  +=       0.260 * cos(4.48257071630  + 32769.37181245730 *T)
    sumX0  +=       0.222 * cos(4.75592425527  + 100909.28143881488 *T)
    sumX0  +=       0.279 * cos(4.66876901800  + 26507.63160293289 *T)
    sumX0  +=       0.299 * cos(4.60166208316  + 2648.69864295650 *T)
    sumX0  +=       0.213 * cos(0.53490943339  + 1479.11039154791 *T)
    sumX0  +=       0.221 * cos(2.46213816436  + 102.84895673509 *T)
    sumX0  +=       0.231 * cos(5.89019952812  + 141762.42362365889 *T)
    sumX0  +=       0.208 * cos(2.11068814642  + 146314.37712071769 *T)
    sumX0  +=       0.236 * cos(3.31514829524  + 14764.99522578630 *T)
    sumX0  +=       0.219 * cos(0.89003648017  + 110.45013870291 *T)
    sumX0  +=       0.241 * cos(4.57135620023  + 150866.33061777649 *T)
    sumX0  +=       0.210 * cos(0.69084085552  + 183724.94436059627 *T)
    sumX0  +=       0.280 * cos(2.23623280041  + 78114.39004507149 *T)
    sumX0  +=       0.211 * cos(1.74754356172  + 130969.45049044908 *T)
    sumX0  +=       0.244 * cos(1.23689090868  + 11610.79577585770 *T)
    sumX0  +=       0.265 * cos(2.55973299169  + 39744.00745023410 *T)
    sumX0  +=       0.200 * cos(0.58765256464  + 16065.82204399129 *T)
    sumX0  +=       0.222 * cos(0.56541187975  + 81706.52818717150 *T)
    sumX0  +=       0.197 * cos(2.60824419320  + 52815.94738694590 *T)
    sumX0  +=       0.197 * cos(2.88503643876  + 12565.90788249930 *T)
    sumX0  +=       0.241 * cos(3.88588376012  + 27147.04125427989 *T)
    sumX0  +=       0.205 * cos(3.58973790844  + 33968.23611239669 *T)
    sumX0  +=       0.248 * cos(0.88668879236  + 71582.72838880648 *T)
    sumX0  +=       0.246 * cos(2.24144190197  + 51742.33836275509 *T)
    sumX0  +=       0.203 * cos(2.34530658749  + 25881.96141062050 *T)
    sumX0  +=       0.202 * cos(6.11911124355  + 25448.24967274370 *T)
    sumX0  +=       0.228 * cos(1.77307991536  + 234790.88445668429 *T)
    sumX0  +=       0.198 * cos(3.28340120527  + 26294.33250749489 *T)
    sumX0  +=       0.231 * cos(2.35481338711  + 44181.52165860769 *T)
    sumX0  +=       0.185 * cos(3.58471190592  + 64607.60511606269 *T)
    sumX0  +=       0.226 * cos(1.07085750404  + 7879.84533585549 *T)
    sumX0  +=       0.176 * cos(2.16160994172  + 77154.08705514569 *T)
    sumX0  +=       0.179 * cos(1.96834324181  + 14477.10736571650 *T)
    sumX0  +=       0.195 * cos(3.51218600925  + 59414.23805726489 *T)
    sumX0  +=       0.207 * cos(5.81274289850  + 1376.01761732930 *T)
    sumX0  +=       0.174 * cos(3.31014261106  + 339142.98465794802 *T)
    sumX0  +=       0.174 * cos(0.97426408403  + 51543.26636131870 *T)
    sumX0  +=       0.184 * cos(6.12768134790  + 129484.15978374588 *T)
    sumX0  +=       0.173 * cos(4.63181171916  + 23968.89546371229 *T)
    sumX0  +=       0.207 * cos(1.74596827081  + 20760.67085067490 *T)
    sumX0  +=       0.166 * cos(5.44721121225  + 79372.84415933248 *T)
    sumX0  +=       0.185 * cos(2.73373869094  + 103711.95909746629 *T)
    sumX0  +=       0.161 * cos(5.70596905124  + 73711.51211018028 *T)
    sumX0  +=       0.194 * cos(3.26236269810  + 50483.88443112950 *T)
    sumX0  +=       0.220 * cos(6.11849377619  + 69159.55861312148 *T)
    sumX0  +=       0.191 * cos(5.22716664164  + 155997.97170183409 *T)
    sumX0  +=       0.216 * cos(3.03668046944  + 53131.16220727349 *T)
    sumX0  +=       0.156 * cos(4.03054314136  + 106570.61348796709 *T)
    sumX0  +=       0.181 * cos(6.12275012681  + 18093.13088206649 *T)
    sumX0  +=       0.152 * cos(2.52925607729  + 78110.17443172129 *T)
    sumX0  +=       0.149 * cos(1.42595419867  + 45892.48661567349 *T)
    sumX0  +=       0.182 * cos(5.31752226443  + 121335.85253123687 *T)
    sumX0  +=       0.155 * cos(3.25690697601  + 188276.89785765507 *T)
    sumX0  +=       0.178 * cos(5.74884910934  + 131499.14145554369 *T)
    sumX0  +=       0.166 * cos(3.75288461237  + 468.20880783791 *T)
    sumX0  +=       0.157 * cos(2.94486244891  + 61279.95709474950 *T)
    sumX0  +=       0.203 * cos(3.72989631697  + 51322.85371887989 *T)
    sumX0  +=       0.190 * cos(5.02157979409  + 25874.36022865269 *T)
    sumX0  +=       0.142 * cos(3.08736192154  + 50594.09075234889 *T)
    sumX0  +=       0.175 * cos(5.19081352879  + 26624.95147115310 *T)
    sumX0  +=       0.148 * cos(4.64578202487  + 104332.18662288050 *T)
    sumX0  +=       0.153 * cos(5.28337526633  + 52101.26850206459 *T)
    sumX0  +=       0.164 * cos(2.33110807572  + 74820.89066227368 *T)
    sumX0  +=       0.158 * cos(2.76429934042  + 78417.73205269089 *T)
    sumX0  +=       0.140 * cos(2.46992319261  + 52279.14287485049 *T)
    sumX0  +=       0.163 * cos(0.92047983596  + 25780.58933808811 *T)
    sumX0  +=       0.160 * cos(4.73851791777  + 26395.70458002730 *T)
    sumX0  +=       0.132 * cos(4.64443067312  + 22760.01130277751 *T)
    sumX0  +=       0.129 * cos(4.81820902062  + 54394.31956985169 *T)
    sumX0  +=       0.150 * cos(0.46939872247  + 104371.52614468009 *T)
    sumX0  +=       0.166 * cos(0.65776448592  + 522.82123557730 *T)
    sumX0  +=       0.149 * cos(4.09099834194  + 23889.05961577070 *T)
    sumX0  +=       0.125 * cos(0.99047205683  + 65697.31390725628 *T)
    sumX0  +=       0.127 * cos(3.92999950538  + 54294.81396101029 *T)
    sumX0  +=       0.136 * cos(1.42756921141  + 102233.09252340188 *T)
    sumX0  +=       0.140 * cos(3.82491395518  + 52808.83383994509 *T)
    sumX0  +=       0.171 * cos(2.58259586325  + 78187.68717093048 *T)
    sumX0  +=       0.119 * cos(1.41260374923  + 178063.61231144407 *T)
    sumX0  +=       0.121 * cos(3.91577422121  + 181506.18725640947 *T)
    sumX0  +=       0.150 * cos(2.94179740209  + 13362.69352428270 *T)
    sumX0  +=       0.160 * cos(2.07478360203  + 955.84355909210 *T)
    sumX0  +=       0.122 * cos(2.54246856782  + 52072.95732641330 *T)
    sumX0  +=       0.139 * cos(5.71618549148  + 97113.18079218028 *T)
    sumX0  +=       0.116 * cos(4.06977577347  + 104344.74283677948 *T)
    sumX0  +=       0.127 * cos(0.21852764054  + 26521.85869693450 *T)
    sumX0  +=       0.113 * cos(1.81808535780  + 154938.58977164488 *T)
    sumX0  +=       0.116 * cos(1.53733085408  + 110634.93223377167 *T)
    sumX0  +=       0.120 * cos(4.73916175135  + 23969.38309867930 *T)
    sumX0  +=       0.117 * cos(2.35333449828  + 18093.61851703350 *T)
    sumX0  +=       0.114 * cos(0.81259306552  + 25754.29101828830 *T)
    sumX0  +=       0.150 * cos(0.60833287071  + 25551.34244696229 *T)
    sumX0  +=       0.109 * cos(2.00086908622  + 40565.01050729069 *T)
    sumX0  +=       0.110 * cos(2.36264207110  + 24176.94747584050 *T)
    sumX0  +=       0.107 * cos(0.12185511336  + 68242.11596210669 *T)
    sumX0  +=       0.112 * cos(4.84462127363  + 26422.00289982710 *T)
    sumX0  +=       0.107 * cos(5.97982710676  + 90695.99589260388 *T)
    sumX0  +=       0.104 * cos(1.75659774658  + 156101.06447605268 *T)
    sumX0  +=       0.111 * cos(4.99889923860  + 116918.00808376308 *T)
    sumX0  +=       0.121 * cos(1.56849356350  + 28287.23430234470 *T)
    sumX0  +=       0.108 * cos(6.08227647204  + 44181.03402364069 *T)
    sumX0  +=       0.102 * cos(4.15125783923  + 26198.35328027710 *T)
    sumX0  +=       0.143 * cos(5.40448013475  + 50579.86365834729 *T)
    sumX0  +=       0.108 * cos(1.57765174204  + 26107.32908499049 *T)
    sumX0  +=       0.103 * cos(1.01910656324  + 29416.28261533789 *T)
    sumX0  +=       0.102 * cos(2.83999953225  + 74923.34081550628 *T)
    sumX0  +=       0.111 * cos(3.14675866322  + 51639.24558853649 *T)
    sumX0  +=       0.100 * cos(5.42391958785  + 25035.87857586930 *T)
    sumX0  +=       0.097 * cos(5.13819824333  + 323.74923414091 *T)
    sumX0  +=       0.100 * cos(3.26784777828  + 104565.15547921829 *T)
    sumX0  +=       0.099 * cos(0.40287203553  + 31415.62306744050 *T)
    sumX0  +=       0.120 * cos(0.30671438664  + 134991.71302241329 *T)
    sumX0  +=       0.131 * cos(2.63461906713  + 647.25465079831 *T)
    sumX0  +=       0.128 * cos(4.13013360989  + 104138.55728834229 *T)
    sumX0  +=       0.110 * cos(0.16249290358  + 24357.02460612510 *T)
    sumX0  +=       0.105 * cos(5.76370877998  + 26067.98956319090 *T)
    sumX0  +=       0.095 * cos(2.53553314645  + 52705.25343075949 *T)
    sumX0  +=       0.105 * cos(5.29014619681  + 55516.66252733170 *T)
    sumX0  +=       0.093 * cos(5.06220815245  + 52126.05347872769 *T)
    sumX0  +=       0.092 * cos(3.89022682970  + 54509.24649350390 *T)
    sumX0  +=       0.096 * cos(3.48344687819  + 743.23387801611 *T)
    sumX0  +=       0.109 * cos(4.33764824375  + 632.53992182970 *T)
    sumX0  +=       0.090 * cos(3.28421553573  + 27999.34644227490 *T)
    sumX0  +=       0.094 * cos(0.26566762837  + 103291.98681862408 *T)
    sumX0  +=       0.111 * cos(4.40219993403  + 40565.49814225770 *T)
    sumX0  +=       0.101 * cos(3.06854770182  + 51951.70530492999 *T)
    sumX0  +=       0.102 * cos(1.77060203080  + 104355.73771913828 *T)
    sumX0  +=       0.088 * cos(0.98000029962  + 1588.82907780029 *T)
    sumX0  +=       0.102 * cos(5.00530419072  + 26941.34334080970 *T)
    sumX0  +=       0.109 * cos(0.41184272678  + 60056.13925397089 *T)
    sumX0  +=       0.093 * cos(5.48203494654  + 27780.31262856009 *T)
    sumX0  +=       0.089 * cos(3.15455699832  + 68050.18006102808 *T)
    sumX0  +=       0.106 * cos(3.72898790593  + 6770.95441872910 *T)
    sumX0  +=       0.101 * cos(0.37716551293  + 52250.83169919920 *T)
    sumX0  +=       0.086 * cos(0.08251374650  + 52325.61329776649 *T)
    sumX0  +=       0.092 * cos(5.92652248722  + 140653.04507156549 *T)
    sumX0  +=       0.107 * cos(1.95514516893  + 157483.26240853727 *T)
    sumX0  +=       0.087 * cos(2.25938740860  + 25938.58376192310 *T)
    sumX0  +=       0.094 * cos(4.55110857685  + 51219.76094466129 *T)
    sumX0  +=       0.089 * cos(5.24168644118  + 36109.98423915710 *T)
    sumX0  +=       0.093 * cos(3.23654834978  + 103.33659170210 *T)
    sumX0  +=       0.098 * cos(5.50021939322  + 27819.26931199030 *T)
    sumX0  +=       0.084 * cos(3.07323794777  + 53228.31848382029 *T)
    sumX0  +=       0.087 * cos(4.71717400578  + 52065.84377941249 *T)
    sumX0  +=       0.099 * cos(4.86243388147  + 97581.14578253467 *T)
    sumX0  +=       0.105 * cos(5.43405944979  + 70269.42480018188 *T)
    sumX0  +=       0.083 * cos(1.07166783581  + 51219.96272405629 *T)
    sumX0  +=       0.101 * cos(0.37289734688  + 52226.04672253610 *T)
    sumX0  +=       0.087 * cos(3.22552375788  + 104778.45457465628 *T)
    sumX0  +=       0.102 * cos(6.17051419926  + 12432.28646788130 *T)
    sumX0  +=       0.082 * cos(3.53594042739  + 78731.91823256049 *T)
    sumX0  +=       0.101 * cos(5.35965052966  + 119117.09542705010 *T)
    sumX0  +=       0.108 * cos(1.06265370367  + 49527.59527515889 *T)
    sumX0  +=       0.092 * cos(4.58719441891  + 51859.65823097529 *T)
    sumX0  +=       0.080 * cos(5.90301150782  + 53242.54557782190 *T)
    sumX0  +=       0.107 * cos(3.33033819724  + 25447.76203777669 *T)
    sumX0  +=       0.088 * cos(6.01318518947  + 12725.20961729149 *T)
    sumX0  +=       0.091 * cos(3.91995839732  + 26300.95841952869 *T)
    sumX0  +=       0.080 * cos(3.22630056109  + 105940.92927906408 *T)
    sumX0  +=       0.076 * cos(1.13669532403  + 102133.09927959349 *T)
    sumX0  +=       0.079 * cos(4.83610406929  + 80482.22271142589 *T)
    sumX0  +=       0.091 * cos(4.23329610745  + 51013.57539622409 *T)
    sumX0  +=       0.094 * cos(2.96601881549  + 2222.10045208050 *T)
    sumX0  +=       0.089 * cos(2.52058786962  + 137678.43511695447 *T)
    sumX0  +=       0.075 * cos(4.78568609813  + 3327.89183879669 *T)
    sumX0  +=       0.089 * cos(0.85977762746  + 1910.95566573370 *T)
    sumX0  +=       0.098 * cos(3.32594040773  + 50056.79860528649 *T)
    sumX0  +=       0.074 * cos(1.41768963459  + 25977.94063783829 *T)
    sumX0  +=       0.089 * cos(0.85266701038  + 52492.44197028850 *T)
    sumX0  +=       0.073 * cos(5.13639330582  + 50697.18352656750 *T)
    sumX0  +=       0.102 * cos(3.34303114673  + 104347.97504842227 *T)
    sumX0  +=       0.072 * cos(3.04850836782  + 170049.41410852008 *T)
    sumX0  +=       0.078 * cos(5.73381249578  + 103285.36090659029 *T)
    sumX0  +=       0.075 * cos(1.55658438412  + 52595.53474450709 *T)
    sumX0  +=       0.074 * cos(2.73392405787  + 25021.65148186770 *T)
    sumX0  +=       0.075 * cos(1.65865555855  + 77844.46859833089 *T)
    sumX0  +=       0.078 * cos(1.07533201236  + 74.53778108379 *T)
    sumX0  +=       0.094 * cos(3.12911978302  + 52698.62751872569 *T)
    sumX0  +=       0.095 * cos(5.99575918602  + 25938.09612695609 *T)
    sumX0  +=       0.087 * cos(5.82357433280  + 72602.13355808688 *T)
    sumX0  +=       0.070 * cos(5.67887850233  + 27670.10630734069 *T)
    sumX0  +=       0.076 * cos(0.64559937719  + 104358.96993078108 *T)
    sumX0  +=       0.074 * cos(5.28706150347  + 25954.03790947270 *T)
    sumX0  +=       0.070 * cos(1.27402715914  + 66653.40128383189 *T)
    sumX0  +=       0.071 * cos(3.97675324486  + 24998.43816786410 *T)
    sumX0  +=       0.073 * cos(0.37426023469  + 26222.25600864270 *T)
    sumX0  +=       0.081 * cos(4.55346032047  + 8194.03151572510 *T)
    sumX0  +=       0.066 * cos(4.21780979859  + 78270.58180110609 *T)
    sumX0  +=       0.072 * cos(1.38206164812  + 426.35437339250 *T)
    sumX0  +=       0.078 * cos(5.28440082300  + 104202.29318664568 *T)
    sumX0  +=       0.066 * cos(3.65294907987  + 89586.12970554348 *T)
    sumX0  +=       0.081 * cos(4.51389667692  + 26080.54577708989 *T)
    sumX0  +=       0.068 * cos(2.13854484878  + 207644.08701988787 *T)
    sumX0  +=       0.065 * cos(5.87225172961  + 12546.72575656651 *T)
    sumX0  +=       0.074 * cos(3.91991873665  + 97670.63153038068 *T)
    sumX0  +=       0.067 * cos(2.82169598550  + 2111.89413086110 *T)
    sumX0  +=       0.079 * cos(5.28726952189  + 77830.24150432929 *T)
    sumX0  +=       0.083 * cos(3.78654945074  + 522.33360061030 *T)
    sumX0  +=       0.079 * cos(3.79951632035  + 25446.73339731871 *T)
    sumX0  +=       0.063 * cos(6.08370597700  + 70268.93716521488 *T)
    sumX0  +=       0.074 * cos(2.86346034860  + 95247.46175469569 *T)
    sumX0  +=       0.085 * cos(5.74305012713  + 72935.98949885629 *T)
    sumX0  +=       0.062 * cos(3.50617407914  + 150244.58681693728 *T)
    sumX0  +=       0.081 * cos(6.24341823549  + 77733.77464214448 *T)
    sumX0  +=       0.077 * cos(1.16703890175  + 316.63568714010 *T)
    sumX0  +=       0.075 * cos(1.04471353660  + 33967.74847742969 *T)
    sumX0  +=       0.071 * cos(3.33594132090  + 27684.33340134230 *T)
    sumX0  +=       0.061 * cos(4.37869496731  + 846.32665223470 *T)
    sumX0  +=       0.064 * cos(1.68211518159  + 27177.85575025130 *T)
    sumX0  +=       0.073 * cos(4.46095079744  + 26083.77798873270 *T)
    sumX0  +=       0.075 * cos(1.85897859977  + 26729.56052079669 *T)
    sumX0  +=       0.071 * cos(4.93586530137  + 17893.87162584910 *T)
    sumX0  +=       0.063 * cos(4.07467864823  + 52290.48938931711 *T)
    sumX0  +=       0.072 * cos(6.13433603432  + 50800.27630078609 *T)
    sumX0  +=       0.059 * cos(1.40204104749  + 52061.61081194667 *T)
    sumX0  +=       0.077 * cos(2.75393031467  + 38813.60039383270 *T)
    sumX0  +=       0.080 * cos(0.16011950416  + 26089.63143176599 *T)
    sumX0  +=       0.069 * cos(1.59730247279  + 56727.51596272369 *T)
    sumX0  +=       0.080 * cos(5.51239927463  + 26086.66248634940 *T)
    sumX0  +=       0.058 * cos(2.41520369138  + 419.24082639170 *T)
    sumX0  +=       0.066 * cos(1.67365233236  + 2008.80135664250 *T)
    sumX0  +=       0.065 * cos(2.06432984048  + 66652.91364886488 *T)
    sumX0  +=       0.058 * cos(1.62097861042  + 99024.38027539747 *T)
    sumX0  +=       0.077 * cos(2.34966674240  + 103299.58800059187 *T)
    sumX0  +=       0.069 * cos(1.46687065035  + 316.14805217309 *T)
    sumX0  +=       0.061 * cos(1.53694228847  + 126997.18458038908 *T)
    sumX0  +=       0.068 * cos(3.36042091101  + 137210.47012660009 *T)
    sumX0  +=       0.058 * cos(2.64478381351  + 167850.32676523307 *T)
    sumX0  +=       0.057 * cos(2.63258373208  + 26011.88088778210 *T)
    sumX0  +=       0.062 * cos(5.91411269940  + 19316.94872284510 *T)
    sumX0  +=       0.073 * cos(2.88502409591  + 26091.54065944870 *T)
    sumX0  +=       0.055 * cos(3.74371053813  + 209812.84750217048 *T)
    sumX0  +=       0.055 * cos(5.12214562228  + 44295.96094729290 *T)
    sumX0  +=       0.058 * cos(1.54167828285  + 25508.45937205890 *T)
    sumX0  +=       0.073 * cos(0.48335623767  + 18848.98373249069 *T)
    sumX0  +=       0.056 * cos(5.41229627249  + 22625.90225319250 *T)
    sumX0  +=       0.064 * cos(0.50377008046  + 77410.75686045409 *T)
    sumX0  +=       0.054 * cos(6.01642065343  + 52252.31617190749 *T)
    sumX0  +=       0.066 * cos(1.74674543912  + 77795.98825185168 *T)
    sumX0  +=       0.055 * cos(3.61457759506  + 107794.43132874568 *T)
    sumX0  +=       0.057 * cos(5.00011319395  + 24449.07744186970 *T)
    sumX0  +=       0.055 * cos(5.16272010710  + 172402.28026229187 *T)
    sumX0  +=       0.067 * cos(1.36451686419  + 19.42594341629 *T)
    sumX0  +=       0.053 * cos(2.13528819255  + 32370.73517408209 *T)
    sumX0  +=       0.052 * cos(5.50451319252  + 3340.36860921629 *T)
    sumX0  +=       0.055 * cos(5.18023594496  + 220.65645992230 *T)
    sumX0  +=       0.072 * cos(4.73998437139  + 24864.32911827909 *T)
    sumX0  +=       0.052 * cos(2.23631389572  + 23868.90221990390 *T)
    sumX0  +=       0.060 * cos(3.57826041100  + 90695.50825763687 *T)
    sumX0  +=       0.052 * cos(5.67658506171  + 26312.49175475960 *T)
    sumX0  +=       0.053 * cos(1.86041754394  + 19406.92210565810 *T)
    sumX0  +=       0.053 * cos(4.13410058005  + 26667.83454605650 *T)
    sumX0  +=       0.070 * cos(1.65225887205  + 2199.33116077051 *T)
    sumX0  +=       0.059 * cos(5.11557005640  + 38.37685312130 *T)
    sumX0  +=       0.051 * cos(1.23099026352  + 58857.27495403150 *T)
    sumX0  +=       0.050 * cos(5.77314673421  + 129799.86223904048 *T)
    sumX0  +=       0.067 * cos(1.08017341186  + 78257.32468330608 *T)
    sumX0  +=       0.050 * cos(0.63851425030  + 27727.21647624570 *T)
    sumX0  +=       0.048 * cos(4.17567122141  + 27044.43611502830 *T)
    sumX0  +=       0.047 * cos(5.67782985051  + 51969.86455219470 *T)
    sumX0  +=       0.055 * cos(4.78351928938  + 157057.35363202327 *T)
    sumX0  +=       0.060 * cos(1.14600976483  + 85034.17620848469 *T)
    sumX0  +=       0.049 * cos(3.86127420297  + 77631.16950289288 *T)
    sumX0  +=       0.064 * cos(0.02229686602  + 76571.78757270369 *T)
    sumX0  +=       0.046 * cos(2.81829667996  + 77836.86741636308 *T)
    sumX0  +=       0.054 * cos(5.78884650295  + 23919.38647677510 *T)
    sumX0  +=       0.047 * cos(2.89825161341  + 155572.06292532009 *T)
    sumX0  +=       0.064 * cos(0.92372709577  + 27311.96479983631 *T)
    sumX0  +=       0.046 * cos(6.08737593166  + 76681.99389392308 *T)
    sumX0  +=       0.046 * cos(4.76636361010  + 29530.23426905610 *T)
    sumX0  +=       0.053 * cos(5.00175567579  + 49842.85370775989 *T)
    sumX0  +=       0.055 * cos(0.24135443971  + 27037.32256802749 *T)
    sumX0  +=       0.054 * cos(3.68631996655  + 949.41942645330 *T)
    sumX0  +=       0.047 * cos(0.24536622662  + 29550.39166492290 *T)
    sumX0  +=       0.056 * cos(0.74089796514  + 53234.94439585409 *T)
    sumX0  +=       0.062 * cos(0.02840313751  + 26011.39325281509 *T)
    sumX0  +=       0.051 * cos(1.39538802358  + 28736.60178453070 *T)
    sumX0  +=       0.045 * cos(4.11327342358  + 55504.18575691209 *T)
    sumX0  +=       0.062 * cos(1.33928421875  + 176954.23375935067 *T)
    sumX0  +=       0.043 * cos(5.66151866873  + 78903.85052852010 *T)
    sumX0  +=       0.044 * cos(3.67807658881  + 16703.30595098250 *T)
    sumX0  +=       0.052 * cos(4.49985681008  + 110012.70079796547 *T)
    sumX0  +=       0.050 * cos(5.46692304514  + 25138.97135008790 *T)
    sumX0  +=       0.043 * cos(4.35523617525  + 1272.92484311070 *T)
    sumX0  +=       0.044 * cos(2.19769750104  + 105460.74730090667 *T)
    sumX0  +=       0.058 * cos(1.25267627496  + 26094.77287109149 *T)
    sumX0  +=       0.042 * cos(2.03273940863  + 1486.22393854870 *T)
    sumX0  +=       0.042 * cos(2.08186025120  + 78189.17164363878 *T)
    sumX0  +=       0.048 * cos(5.60414814999  + 65831.91059180829 *T)
    sumX0  +=       0.054 * cos(0.13039414614  + 52382.23564906909 *T)
    sumX0  +=       0.041 * cos(3.34731027231  + 1795.50226120450 *T)
    sumX0  +=       0.043 * cos(1.03324982448  + 60055.65161900389 *T)
    sumX0  +=       0.046 * cos(0.15692625477  + 54087.24958384910 *T)
    sumX0  +=       0.053 * cos(4.84593569859  + 260878.78759825832 *T)
    sumX0  +=       0.048 * cos(0.50977666152  + 25466.40315821850 *T)
    sumX0  +=       0.048 * cos(5.15384836969  + 26709.89075989690 *T)
    sumX0  +=       0.048 * cos(2.14565331737  + 76667.76679992149 *T)
    sumX0  +=       0.047 * cos(5.63034544251  + 104275.59031250468 *T)
    sumX0  +=       0.039 * cos(0.78144245702  + 132658.51662954129 *T)
    sumX0  +=       0.040 * cos(1.64169511886  + 26824.02347258949 *T)
    sumX0  +=       0.040 * cos(4.01282714257  + 25352.27044552590 *T)
    sumX0  +=       0.043 * cos(0.07999630612  + 365230.88779952226 *T)
    sumX0  +=       0.044 * cos(1.85242708019  + 52712.85461272729 *T)
    sumX0  +=       0.045 * cos(6.15705935255  + 28256.90744134030 *T)
    sumX0  +=       0.039 * cos(5.27554854731  + 103241.99019671988 *T)
    sumX0  +=       0.051 * cos(1.99101186951  + 182085.87484340827 *T)
    sumX0  +=       0.050 * cos(2.09130353046  + 147423.75567281109 *T)
    sumX0  +=       0.043 * cos(0.02949186925  + 214364.80099922928 *T)
    sumX0  +=       0.039 * cos(2.59117531794  + 99799.41525175449 *T)
    sumX0  +=       0.044 * cos(5.29675829214  + 25867.73431661889 *T)
    sumX0  +=       0.043 * cos(1.77519064144  + 1265.81129610991 *T)
    sumX0  +=       0.038 * cos(5.81703538519  + 104505.63519426508 *T)
    sumX0  +=       0.037 * cos(5.58241031795  + 104198.07757329548 *T)
    sumX0  +=       0.037 * cos(4.53399815239  + 128320.99566497609 *T)
    sumX0  +=       0.042 * cos(5.19727486409  + 25170.21667307590 *T)
    sumX0  +=       0.040 * cos(5.22993102913  + 206.42936592071 *T)
    sumX0  +=       0.038 * cos(3.52230881007  + 130459.42928625426 *T)
    sumX0  +=       0.037 * cos(3.93629350281  + 26102.37405305930 *T)
    sumX0  +=       0.050 * cos(0.24736089771  + 153.53499300130 *T)
    sumX0  +=       0.037 * cos(1.67080581846  + 24203.24579564031 *T)
    sumX0  +=       0.041 * cos(0.69210311059  + 34282.42229226630 *T)
    sumX0  +=       0.036 * cos(3.01321329685  + 26118.47382006210 *T)
    sumX0  +=       0.036 * cos(2.65041173436  + 26057.82009805329 *T)
    sumX0  +=       0.036 * cos(4.58551065069  + 34281.93465729930 *T)
    sumX0  +=       0.042 * cos(6.16382020852  + 38653.81102407349 *T)
    sumX0  +=       0.035 * cos(1.73963103644  + 26073.91986505609 *T)
    sumX0  +=       0.039 * cos(0.45985553144  + 27006.07724503950 *T)
    sumX0  +=       0.039 * cos(5.40193552573  + 151.29148732640 *T)
    sumX0  +=       0.037 * cos(5.43157910397  + 78367.04601642469 *T)
    sumX0  +=       0.034 * cos(3.56564955468  + 27567.01353312209 *T)
    sumX0  +=       0.035 * cos(0.56899615907  + 78896.73698151928 *T)
    sumX0  +=       0.037 * cos(4.00811594346  + 25773.96077918809 *T)
    sumX0  +=       0.037 * cos(1.65550908775  + 26402.33313892731 *T)
    sumX0  +=       0.045 * cos(0.79814686706  + 2126.12122486270 *T)
    sumX0  +=       0.037 * cos(5.12681406758  + 9384.59719059169 *T)
    sumX0  +=       0.033 * cos(2.80453550266  + 116783.89903417807 *T)
    sumX0  +=       0.035 * cos(2.62159531521  + 27463.92075890350 *T)
    sumX0  +=       0.033 * cos(0.53015735881  + 51852.54468397449 *T)
    sumX0  +=       0.037 * cos(1.41559511533  + 130420.08976445468 *T)
    sumX0  +=       0.035 * cos(0.57763840050  + 62388.84801187589 *T)
    sumX0  +=       0.033 * cos(1.99538084924  + 26190.10947732330 *T)
    sumX0  +=       0.033 * cos(3.66824418197  + 25986.18444079209 *T)
    sumX0  +=       0.035 * cos(3.17259770035  + 51653.47268253809 *T)
    sumX0  +=       0.032 * cos(6.04952225858  + 39629.56816154889 *T)
    sumX0  +=       0.040 * cos(4.40766389323  + 148.32254190981 *T)
    sumX0  +=       0.043 * cos(2.52314764326  + 157587.04459711787 *T)
    sumX0  +=       0.034 * cos(3.99284017002  + 27973.04812247509 *T)
    sumX0  +=       0.043 * cos(0.54047908222  + 26308.55960149650 *T)
    sumX0  +=       0.040 * cos(0.99428555932  + 205.94173095370 *T)
    sumX0  +=       0.034 * cos(3.54482897451  + 13541.17738500769 *T)
    sumX0  +=       0.031 * cos(0.68583013925  + 207594.09039798367 *T)
    sumX0  +=       0.032 * cos(4.46011355827  + 204151.51545301828 *T)
    sumX0  +=       0.034 * cos(3.97349842615  + 51868.49247966230 *T)
    sumX0  +=       0.043 * cos(3.01518313757  + 51529.03926731709 *T)
    sumX0  +=       0.031 * cos(5.58184574820  + 78160.86046798748 *T)
    sumX0  +=       0.036 * cos(0.93475827193  + 130226.46042991648 *T)
    sumX0  +=       0.030 * cos(0.82681675219  + 130432.64597835368 *T)
    sumX0  +=       0.039 * cos(5.59499527090  + 64901.50353540690 *T)
    sumX0  +=       0.031 * cos(2.16793872992  + 51433.06004009930 *T)
    sumX0  +=       0.041 * cos(0.15772285796  + 40852.89836736050 *T)
    sumX0  +=       0.039 * cos(1.91317014888  + 76674.39271195528 *T)
    sumX0  +=       0.031 * cos(3.23141999011  + 94330.01910368088 *T)
    sumX0  +=       0.030 * cos(4.87996329072  + 181026.49291321909 *T)
    sumX0  +=       0.030 * cos(6.07201296464  + 96356.84030678908 *T)
    sumX0  +=       0.029 * cos(4.81126203138  + 182188.96761762688 *T)
    sumX0  +=       0.034 * cos(4.11459561196  + 111122.07935005888 *T)
    sumX0  +=       0.031 * cos(4.20399330248  + 26514.25751496669 *T)
    sumX0  +=       0.036 * cos(4.04094788403  + 75615.49841673308 *T)
    sumX0  +=       0.028 * cos(1.03310392380  + 86143.55476057807 *T)
    sumX0  +=       0.028 * cos(0.70757951529  + 80382.71710258449 *T)
    sumX0  +=       0.037 * cos(4.30143869314  + 37698.69891743190 *T)
    sumX0  +=       0.030 * cos(0.97985183975  + 77101.47853779829 *T)
    sumX0  +=       0.029 * cos(4.41729316787  + 114.64292439690 *T)
    sumX0  +=       0.030 * cos(4.17789908809  + 26084.21480579459 *T)
    sumX0  +=       0.027 * cos(0.61831878834  + 26241.43813457550 *T)
    sumX0  +=       0.027 * cos(4.67046992186  + 60369.83779887349 *T)
    sumX0  +=       0.027 * cos(5.92007475132  + 101011.24395708049 *T)
    sumX0  +=       0.027 * cos(1.62628759375  + 52509.90604140130 *T)
    sumX0  +=       0.029 * cos(1.49521060767  + 26092.07911232079 *T)
    sumX0  +=       0.033 * cos(6.04618436459  + 42153.72518556549 *T)
    sumX0  +=       0.033 * cos(2.68337298749  + 49954.19346603489 *T)
    sumX0  +=       0.027 * cos(0.40084478333  + 25933.88051360590 *T)
    sumX0  +=       0.027 * cos(4.17930801308  + 91785.21704883048 *T)
    sumX0  +=       0.026 * cos(2.02354598068  + 35472.98796713290 *T)
    sumX0  +=       0.034 * cos(1.25359434355  + 22004.15845235330 *T)
    sumX0  +=       0.026 * cos(4.60481238290  + 71980.38975724769 *T)
    sumX0  +=       0.026 * cos(0.04548898361  + 130653.05862079248 *T)
    sumX0  +=       0.029 * cos(4.59660921219  + 136722.83537534589 *T)
    sumX0  +=       0.035 * cos(1.50172306531  + 52483.60772160150 *T)
    sumX0  +=       0.027 * cos(4.62222063183  + 54375.13744391890 *T)
    sumX0  +=       0.026 * cos(5.92161181229  + 26050.01392341990 *T)
    sumX0  +=       0.033 * cos(6.20772055762  + 77727.14873011068 *T)
    sumX0  +=       0.031 * cos(6.01164257984  + 87367.86023632369 *T)
    sumX0  +=       0.028 * cos(5.00871771816  + 183571.16555011147 *T)
    sumX0  +=       0.034 * cos(0.00412760548  + 1905.70858242391 *T)
    sumX0  +=       0.028 * cos(2.05766266145  + 51226.87449166209 *T)
    sumX0  +=       0.032 * cos(1.32626829577  + 77307.66408623548 *T)
    sumX0  +=       0.033 * cos(6.12204372905  + 78039.60844650418 *T)
    sumX0  +=       0.031 * cos(3.88641044819  + 60370.32543384050 *T)
    sumX0  +=       0.033 * cos(2.48500437212  + 123201.08393375447 *T)
    sumX0  +=       0.027 * cos(0.91898929060  + 1224.06165826211 *T)
    sumX0  +=       0.025 * cos(3.43998049238  + 78338.73484077338 *T)
    sumX0  +=       0.029 * cos(2.28677197427  + 49842.36607279289 *T)
    sumX0  +=       0.025 * cos(1.89094361261  + 76785.08666814168 *T)
    sumX0  +=       0.024 * cos(2.66952173476  + 24824.98959647950 *T)
    sumX0  +=       0.031 * cos(3.35325440781  + 161079.61616398749 *T)
    sumX0  +=       0.031 * cos(1.70259252563  + 75930.75684933408 *T)
    sumX0  +=       0.024 * cos(3.24402392538  + 1045.39865367110 *T)
    sumX0  +=       0.029 * cos(5.58510707424  + 25600.51220780350 *T)
    sumX0  +=       0.029 * cos(0.07851795697  + 26575.78171031190 *T)
    sumX0  +=       0.026 * cos(5.97420147293  + 39450.59666585690 *T)
    sumX0  +=       0.027 * cos(0.73776364996  + 27171.22719135130 *T)
    sumX0  +=       0.027 * cos(4.92586138125  + 25005.06672676410 *T)
    sumX0  +=       0.032 * cos(0.79345370764  + 9103.66317663410 *T)
    sumX0  +=       0.024 * cos(2.49242891465  + 129373.26404816448 *T)
    sumX0  +=       0.023 * cos(0.40602877121  + 85502.14119883909 *T)
    sumX0  +=       0.024 * cos(3.27434836834  + 52609.76183850870 *T)
    sumX0  +=       0.025 * cos(4.99949553531  + 299.37021175271 *T)
    sumX0  +=       0.027 * cos(1.76554597629  + 143005.91122533727 *T)
    sumX0  +=       0.027 * cos(1.29828079538  + 24073.16528725990 *T)
    sumX0  +=       0.025 * cos(5.44789800662  + 25654.43522118089 *T)
    sumX0  +=       0.024 * cos(0.97156468916  + 52286.25642185129 *T)
    sumX0  +=       0.023 * cos(5.20850734570  + 38813.11275886569 *T)
    sumX0  +=       0.026 * cos(4.38403477094  + 27676.73221937449 *T)
    sumX0  +=       0.023 * cos(2.99428107393  + 27351.30432163590 *T)
    sumX0  +=       0.025 * cos(1.41127895754  + 48847.91444435170 *T)
    sumX0  +=       0.028 * cos(4.21226965873  + 24925.67225464990 *T)
    sumX0  +=       0.029 * cos(4.41166263353  + 30172.13546576210 *T)
    sumX0  +=       0.026 * cos(1.50614892755  + 78153.74692098668 *T)
    sumX0  +=       0.023 * cos(2.83196594205  + 50064.39978725430 *T)
    sumX0  +=       0.024 * cos(1.97684419653  + 51962.26337022689 *T)
    sumX0  +=       0.023 * cos(1.32747738833  + 77947.56137254949 *T)
    sumX0  +=       0.023 * cos(1.63082397395  + 123669.04892410888 *T)
    sumX0  +=       0.022 * cos(1.83245731848  + 78213.95662030189 *T)
    sumX0  +=       0.025 * cos(5.93871501925  + 26725.14323108190 *T)
    sumX0  +=       0.025 * cos(6.00809531914  + 25451.15068703350 *T)
    sumX0  +=       0.022 * cos(5.41900898206  + 50264.85061741470 *T)
    sumX0  +=       0.030 * cos(2.27535725604  + 53907.17245356450 *T)
    sumX0  +=       0.030 * cos(2.41447471924  + 23754.46293121869 *T)
    sumX0  +=       0.025 * cos(3.33106077360  + 1.72829019180 *T)
    sumX0  +=       0.025 * cos(3.33106077360  + 1.24065522479 *T)
    sumX0  +=       0.024 * cos(6.19145581405  + 78786.53066029989 *T)
    sumX0  +=       0.022 * cos(3.40212459403  + 6885.39370741431 *T)
    sumX0  +=       0.025 * cos(0.11282707964  + 130435.87818999648 *T)
    sumX0  +=       0.022 * cos(3.21338836945  + 50444.92774769930 *T)
    sumX0  +=       0.028 * cos(6.25386153205  + 6129.54085699010 *T)
    sumX0  +=       0.025 * cos(4.36482847694  + 28103.12863085550 *T)
    sumX0  +=       0.023 * cos(2.05346110066  + 103918.14464590348 *T)
    sumX0  +=       0.025 * cos(2.29706808754  + 52174.56562792359 *T)
    sumX0  +=       0.022 * cos(1.18783541749  + 2219.00092167030 *T)
    sumX0  +=       0.022 * cos(3.88895575058  + 78580.34511186268 *T)
    sumX0  +=       0.029 * cos(2.91258588959  + 1730.87853544909 *T)
    sumX0  +=       0.022 * cos(0.65836013072  + 80597.14963507809 *T)
    sumX0  +=       0.024 * cos(3.86546209537  + 51842.19415986250 *T)
    sumX0  +=       0.020 * cos(0.93295366770  + 51955.63745819309 *T)
    sumX0  +=       0.028 * cos(2.18732774825  + 536.56069461190 *T)
    sumX0  +=       0.020 * cos(3.94160971202  + 23439.20449861769 *T)
    sumX0  +=       0.023 * cos(5.40527491438  + 129387.49114216608 *T)
    sumX0  +=       0.021 * cos(5.54505025605  + 163766.33825852867 *T)
    sumX0  +=       0.025 * cos(2.01245845269  + 81604.56566890588 *T)
    sumX0  +=       0.021 * cos(3.01232876030  + 1063.55790093581 *T)
    sumX0  +=       0.023 * cos(4.96870238168  + 16342.82640320110 *T)
    sumX0  +=       0.022 * cos(2.15856505754  + 96357.32794175608 *T)
    sumX0  +=       0.023 * cos(1.14266322173  + 76144.70174686068 *T)
    sumX0  +=       0.020 * cos(1.29887924947  + 51329.96726588070 *T)
    sumX0  +=       0.020 * cos(1.71833485025  + 1089.95260867711 *T)
    sumX0  +=       0.020 * cos(0.66958086764  + 123758.53467195488 *T)
    sumX0  +=       0.019 * cos(0.72592280877  + 51535.66517935089 *T)
    sumX0  +=       0.022 * cos(3.99083726125  + 3462.00088838170 *T)
    sumX0  +=       0.021 * cos(2.05460557007  + 130290.19632821988 *T)
    sumX0  +=       0.023 * cos(3.93119144363  + 50167.49256147289 *T)
    sumX0  +=       0.020 * cos(6.09924257691  + 196137.31725009429 *T)
    sumX0  +=       0.023 * cos(2.69769554120  + 166740.94821313969 *T)
    sumX0  +=       0.026 * cos(6.04075681178  + 22747.04689739089 *T)
    sumX0  +=       0.022 * cos(0.05797120861  + 26126.27999469550 *T)
    sumX0  +=       0.022 * cos(5.71570675867  + 26735.15779237250 *T)
    sumX0  +=       0.019 * cos(4.66884897067  + 78683.43788608129 *T)
    sumX0  +=       0.020 * cos(2.25947325898  + 75930.26921436709 *T)
    sumX0  +=       0.024 * cos(2.12831639297  + 145204.99856862429 *T)
    sumX0  +=       0.019 * cos(3.69323819425  + 130446.87307235529 *T)
    sumX0  +=       0.024 * cos(3.04675754545  + 52177.53457334019 *T)
    sumX0  +=       0.025 * cos(4.82366657336  + 130443.64086071248 *T)
    sumX0  +=       0.020 * cos(2.64704504078  + 98690.03669966108 *T)
    sumX0  +=       0.019 * cos(3.52306398899  + 103498.66000202828 *T)
    sumX0  +=       0.025 * cos(2.89149847485  + 76888.17944236028 *T)
    sumX0  +=       0.022 * cos(0.83649888960  + 24952.47008716450 *T)
    sumX0  +=       0.022 * cos(4.82712614161  + 27223.82383095090 *T)
    sumX0  +=       0.018 * cos(3.15171375774  + 78413.51643934069 *T)
    sumX0  +=       0.020 * cos(0.30552664552  + 104819.82137413470 *T)
    sumX0  +=       0.022 * cos(3.38793394354  + 129379.88996019828 *T)
    sumX0  +=       0.019 * cos(6.23315092064  + 26248.55432844250 *T)
    sumX0  +=       0.019 * cos(5.71365941775  + 25927.73958967290 *T)
    sumX0  +=       0.019 * cos(1.78991551728  + 53029.24648238390 *T)
    sumX0  +=       0.018 * cos(4.18932552893  + 128221.00242116768 *T)
    sumX0  +=       0.020 * cos(5.93600662958  + 25908.06982877310 *T)
    sumX0  +=       0.020 * cos(6.01080370881  + 26268.22408934230 *T)
    sumX0  +=       0.019 * cos(4.93432108959  + 52817.46366237089 *T)
    sumX0  +=       0.018 * cos(1.87899794614  + 26411.65237571510 *T)
    sumX0  +=       0.022 * cos(4.75027274504  + 103932.37173990508 *T)
    sumX0  +=       0.018 * cos(4.74269631350  + 52195.23222656469 *T)
    sumX0  +=       0.024 * cos(3.42780024885  + 86144.04239554508 *T)
    sumX0  +=       0.020 * cos(0.38316404093  + 12412.61670698150 *T)
    sumX0  +=       0.025 * cos(3.36113877175  + 78313.94986411028 *T)
    sumX0  +=       0.020 * cos(5.18325339676  + 233731.99016146208 *T)
    sumX0  +=       0.020 * cos(0.60562813850  + 25234.95057730570 *T)
    sumX0  +=       0.017 * cos(2.29216844390  + 53868.21577013429 *T)
    sumX0  +=       0.017 * cos(4.19571093375  + 77307.86586563049 *T)
    sumX0  +=       0.017 * cos(0.77039059907  + 24602.85625235450 *T)
    sumX0  +=       0.017 * cos(4.89323443214  + 27573.43766576090 *T)
    sumX0  +=       0.023 * cos(1.65926065951  + 14.47091148511 *T)
    sumX0  +=       0.018 * cos(0.80516343452  + 52137.91706499410 *T)
    sumX0  +=       0.020 * cos(6.28264844209  + 130866.35771623049 *T)
    sumX0  +=       0.016 * cos(2.69293004860  + 79330.44871939608 *T)
    sumX0  +=       0.018 * cos(6.15817325038  + 79316.22162539448 *T)
    sumX0  +=       0.016 * cos(2.44285814230  + 3.63751787450 *T)
    sumX0  +=       0.017 * cos(5.95313986190  + 121335.36489626988 *T)
    sumX0  +=       0.019 * cos(0.00239064860  + 132028.83242063827 *T)
    sumX0  +=       0.018 * cos(3.07016978555  + 102659.69071427788 *T)
    sumX0  +=       0.018 * cos(2.48538976177  + 99023.89264043048 *T)
    sumX0  +=       0.016 * cos(1.30832465047  + 50903.36907500470 *T)
    sumX0  +=       0.019 * cos(0.09235363676  + 94138.08320260228 *T)
    sumX0  +=       0.016 * cos(6.10168410633  + 28310.00359365469 *T)
    sumX0  +=       0.016 * cos(5.36118674347  + 25668.17468021549 *T)
    sumX0  +=       0.016 * cos(0.06029661634  + 24395.98128955530 *T)
    sumX0  +=       0.019 * cos(5.79930577989  + 78793.15657233368 *T)
    sumX0  +=       0.015 * cos(2.51997196564  + 155887.76538061467 *T)
    sumX0  +=       0.022 * cos(6.08296146382  + 1055.69359440960 *T)
    sumX0  +=       0.016 * cos(2.88270090422  + 65717.47130312308 *T)
    sumX0  +=       0.021 * cos(5.12943933290  + 3178.38960805111 *T)
    sumX0  +=       0.019 * cos(4.86287967507  + 106570.12585300008 *T)
    sumX0  +=       0.015 * cos(3.64115942517  + 50690.06997956669 *T)
    sumX0  +=       0.017 * cos(3.32848323709  + 55638.29480649710 *T)
    sumX0  +=       0.016 * cos(0.98987078340  + 104358.48494268030 *T)
    sumX0  +=       0.015 * cos(4.27283735340  + 6043.98476389190 *T)
    sumX0  +=       0.016 * cos(0.28894936273  + 176332.48995851149 *T)
    sumX0  +=       0.017 * cos(2.00324514114  + 62197.88738073129 *T)
    sumX0  +=       0.017 * cos(4.71989777168  + 86457.74094044768 *T)
    sumX0  +=       0.015 * cos(3.42730813171  + 52310.15915021690 *T)
    sumX0  +=       0.015 * cos(0.50958115781  + 235900.75064374463 *T)
    sumX0  +=       0.015 * cos(4.72190436199  + 61560.89110870710 *T)
    sumX0  +=       0.017 * cos(3.12766472647  + 103821.67778371868 *T)
    sumX0  +=       0.017 * cos(4.57987044915  + 153085.08772196327 *T)
    sumX0  +=       0.020 * cos(4.57785189845  + 3.68833577960 *T)
    sumX0  +=       0.014 * cos(5.68581925640  + 193938.22990680728 *T)
    sumX0  +=       0.015 * cos(2.61658035144  + 50007.28961834930 *T)
    sumX0  +=       0.015 * cos(0.06488742456  + 26831.13701959030 *T)
    sumX0  +=       0.014 * cos(0.38204630080  + 133882.33447031991 *T)
    sumX0  +=       0.017 * cos(3.50167044497  + 39629.08052658189 *T)
    sumX0  +=       0.016 * cos(1.91399253711  + 77.99436146740 *T)
    sumX0  +=       0.015 * cos(4.72920790161  + 125112.28341697170 *T)
    sumX0  +=       0.014 * cos(2.29306969321  + 50380.79165691089 *T)
    sumX0  +=       0.016 * cos(4.13091799370  + 104345.22782488028 *T)
    sumX0  +=       0.014 * cos(1.95123243708  + 26247.44869384990 *T)
    sumX0  +=       0.014 * cos(3.71239259413  + 25928.84522426550 *T)
    sumX0  +=       0.014 * cos(0.78821341455  + 16066.30967895830 *T)
    sumX0  +=       0.016 * cos(3.86619726000  + 50910.48262200550 *T)
    sumX0  +=       0.018 * cos(2.12623271525  + 92740.81679043909 *T)
    sumX0  +=       0.014 * cos(5.78782811767  + 1169.83206889211 *T)
    sumX0  +=       0.016 * cos(3.71525448527  + 53815.11961781990 *T)
    sumX0  +=       0.014 * cos(4.45337551699  + 78149.51395352086 *T)
    sumX0  +=       0.016 * cos(4.39061350798  + 203042.13690092487 *T)
    sumX0  +=       0.013 * cos(0.37425161178  + 639.65346883050 *T)
    sumX0  +=       0.015 * cos(6.07909581296  + 77616.94240889128 *T)
    sumX0  +=       0.014 * cos(4.68939663529  + 82815.41910429788 *T)
    sumX0  +=       0.014 * cos(4.74773700930  + 29415.79498037089 *T)
    sumX0  +=       0.014 * cos(5.20331959316  + 102755.66994149568 *T)
    sumX0  +=       0.014 * cos(1.72247029033  + 109.96250373590 *T)
    sumX0  +=       0.017 * cos(0.16324703337  + 163298.37326817427 *T)
    sumX0  +=       0.015 * cos(1.68035003786  + 11.28951774740 *T)
    sumX0  +=       0.014 * cos(5.16225967454  + 173511.65881438527 *T)
    sumX0  +=       0.014 * cos(4.39150540314  + 52161.82300663029 *T)
    sumX0  +=       0.012 * cos(2.27730262051  + 102018.17235594128 *T)
    sumX0  +=       0.017 * cos(2.64158525193  + 52155.89270476510 *T)
    sumX0  +=       0.013 * cos(2.88005600662  + 102769.89703549728 *T)
    sumX0  +=       0.014 * cos(0.84211332161  + 78378.39253089130 *T)
    sumX0  +=       0.014 * cos(3.21771971652  + 26077.10125879380 *T)
    sumX0  +=       0.014 * cos(2.44590531469  + 26099.19265932160 *T)
    sumX0  +=       0.013 * cos(3.84393935941  + 79322.84753742829 *T)
    sumX0  +=       0.016 * cos(0.56654525193  + 51534.63653889291 *T)
    sumX0  +=       0.014 * cos(4.45047217505  + 20043.91837768230 *T)
    sumX0  +=       0.016 * cos(4.85429966913  + 50050.17269325269 *T)
    sumX0  +=       0.013 * cos(3.70788935281  + 44936.88687406489 *T)
    sumX0  +=       0.014 * cos(6.05588092072  + 24506.18761077469 *T)
    sumX0  +=       0.014 * cos(0.68454442549  + 35833.46751491430 *T)
    sumX0  +=       0.012 * cos(0.96763097373  + 52388.86156110289 *T)
    sumX0  +=       0.014 * cos(1.93515089311  + 198490.18340386607 *T)
    sumX0  +=       0.013 * cos(0.54258695046  + 103719.07264446709 *T)
    sumX0  +=       0.013 * cos(2.24721025792  + 24491.96051677309 *T)
    sumX0  +=       0.013 * cos(2.99089102997  + 52025.99926853029 *T)
    sumX0  +=       0.012 * cos(1.26244334627  + 25984.36477047710 *T)
    sumX0  +=       0.012 * cos(4.40118168494  + 26191.92914763830 *T)
    sumX0  +=       0.014 * cos(4.74013545104  + 53265.75889182550 *T)
    sumX0  +=       0.016 * cos(2.69508248660  + 25653.94758621389 *T)
    sumX0  +=       0.012 * cos(5.94982180077  + 181659.96606689427 *T)
    sumX0  +=       0.014 * cos(3.98127038415  + 53399.86794141051 *T)
    sumX0  +=       0.012 * cos(0.13992189685  + 53772.23654291649 *T)
    sumX0  +=       0.012 * cos(1.72523320675  + 43981.77476742330 *T)
    sumX0  +=       0.012 * cos(5.92436505942  + 28199.79727243529 *T)
    sumX0  +=       0.015 * cos(4.44881823724  + 1638.82569970449 *T)
    sumX0  +=       0.016 * cos(6.01974460542  + 34082.67540108190 *T)
    sumX0  +=       0.012 * cos(4.59110069577  + 51596.36251363310 *T)
    sumX0  +=       0.013 * cos(0.25880243151  + 52025.00243078899 *T)
    sumX0  +=       0.011 * cos(4.64769916430  + 52911.92661416369 *T)
    sumX0  +=       0.011 * cos(1.89874077048  + 70383.86408886709 *T)
    sumX0  +=       0.012 * cos(1.02764449556  + 112231.45790215228 *T)
    sumX0  +=       0.015 * cos(4.79472245666  + 103883.89139342587 *T)
    sumX0  +=       0.012 * cos(1.34514817706  + 52171.68113030689 *T)
    sumX0  +=       0.014 * cos(1.43559106475  + 26555.62431444510 *T)
    sumX0  +=       0.013 * cos(1.61506412907  + 286966.69073983253 *T)
    sumX0  +=       0.013 * cos(5.01690772702  + 208173.77798498247 *T)
    sumX0  +=       0.013 * cos(5.62315849993  + 79219.06534884768 *T)
    sumX0  +=       0.011 * cos(1.70747594337  + 52168.44891866409 *T)
    sumX0  +=       0.011 * cos(5.83137810614  + 26012.87772552339 *T)
    sumX0  +=       0.011 * cos(2.17980505276  + 48713.80539476670 *T)
    sumX0  +=       0.014 * cos(0.87464188435  + 27998.85880730789 *T)
    sumX0  +=       0.014 * cos(1.00856446032  + 52755.73768763070 *T)
    sumX0  +=       0.014 * cos(2.05272684647  + 52041.94105104690 *T)
    sumX0  +=       0.012 * cos(5.20664534748  + 131548.65044248087 *T)
    sumX0  +=       0.012 * cos(2.40230884933  + 130363.49345407888 *T)
    sumX0  +=       0.011 * cos(5.14348161985  + 104277.07478521300 *T)
    sumX0  +=       0.014 * cos(4.30522472177  + 27250.62166346549 *T)
    sumX0  +=       0.011 * cos(3.57984794643  + 51109.06698847489 *T)
    sumX0  +=       0.011 * cos(0.50291413498  + 115674.03284711768 *T)
    sumX0  +=       0.014 * cos(2.57900377541  + 49527.10764019189 *T)
    sumX0  +=       0.013 * cos(1.25530162700  + 52027.97137620559 *T)
    sumX0  +=       0.011 * cos(3.18085491379  + 57503.52620901470 *T)
    sumX0  +=       0.013 * cos(3.20863649406  + 78470.13879064328 *T)
    sumX0  +=       0.014 * cos(1.85519169440  + 860.55374623631 *T)
    sumX0  +=       0.011 * cos(4.15960190462  + 92741.30442540610 *T)
    sumX0  +=       0.013 * cos(5.32460666801  + 126996.69694542208 *T)
    sumX0  +=       0.011 * cos(1.25079104877  + 20894.77990025990 *T)
    sumX0  +=       0.012 * cos(0.71489552500  + 51440.17358710009 *T)
    sumX0  +=       0.015 * cos(0.61708921425  + 24395.49365458829 *T)
    sumX0  +=       0.012 * cos(4.22401388053  + 52179.98225389499 *T)
    sumX0  +=       0.013 * cos(3.48954455220  + 65716.98366815608 *T)
    sumX0  +=       0.014 * cos(4.39584924296  + 3442.33112748190 *T)
    sumX0  +=       0.011 * cos(5.05191127458  + 736.12033101530 *T)
    sumX0  +=       0.010 * cos(2.32271342326  + 333.61212328590 *T)
    sumX0  +=       0.011 * cos(1.70456192528  + 50536.98058344389 *T)
    sumX0  +=       0.010 * cos(2.59099487043  + 26190.75209830929 *T)
    sumX0  +=       0.012 * cos(2.51295739282  + 53758.00944891489 *T)
    sumX0  +=       0.011 * cos(5.25664726449  + 26081.03076519069 *T)
    sumX0  +=       0.010 * cos(2.79700804589  + 78340.21931348169 *T)
    sumX0  +=       0.011 * cos(4.88405989087  + 78800.75775430149 *T)
    sumX0  +=       0.012 * cos(6.04572661714  + 52179.44380102290 *T)
    sumX0  +=       0.010 * cos(4.25864792126  + 84945.17809560569 *T)
    sumX0  +=       0.011 * cos(0.86477927749  + 81592.08889848628 *T)
    sumX0  +=       0.011 * cos(5.65005720167  + 641.16974425549 *T)
    sumX0  +=       0.014 * cos(1.53374794610  + 183145.25677359747 *T)
    sumX0  +=       0.010 * cos(6.17476996345  + 25863.80216335580 *T)
    sumX0  +=       0.010 * cos(3.55930973128  + 53093.98038661369 *T)
    sumX0  +=       0.011 * cos(0.09944117691  + 51123.29408247649 *T)
    sumX0  +=       0.014 * cos(0.87408194721  + 25242.06412430650 *T)
    sumX0  +=       0.010 * cos(1.40003580429  + 154408.89880655028 *T)
    sumX0  +=       0.014 * cos(2.93953436186  + 54344.81058291449 *T)
    sumX0  +=       0.011 * cos(2.53121281826  + 78057.76769376888 *T)
    sumX0  +=       0.012 * cos(1.32462443773  + 26149.87293379789 *T)
    sumX0  +=       0.012 * cos(4.33900059348  + 26026.42098431750 *T)
    sumX0  +=       0.010 * cos(0.70338242506  + 637.24008950771 *T)
    sumX0  +=       0.010 * cos(2.36288136655  + 91919.81373338249 *T)
    sumX0  +=       0.014 * cos(3.26945543294  + 53125.22570960169 *T)
    sumX0  +=       0.011 * cos(2.11182562121  + 26094.28788299070 *T)
    sumX0  +=       0.012 * cos(1.35040938945  + 136100.60393953967 *T)
    sumX0  +=       0.010 * cos(4.17716450281  + 6751.28465782931 *T)
    sumX0  +=       0.010 * cos(0.88468292764  + 137209.98249163307 *T)
    sumX0  +=       0.010 * cos(3.56286217757  + 77940.44782554869 *T)
    sumX0  +=       0.010 * cos(2.79243370962  + 45404.85186441930 *T)
    sumX0  +=       0.013 * cos(5.65469308747  + 50270.58533569150 *T)
    sumX0  +=       0.011 * cos(5.25078615822  + 77520.96318167349 *T)
    sumX0  +=       0.011 * cos(5.80424854653  + 19958.84991955110 *T)
    sumX0  +=       0.010 * cos(6.13036479176  + 25619.69433373629 *T)
    sumX0  +=       0.011 * cos(5.27912616615  + 39763.67721113390 *T)
    sumX0  +=       0.010 * cos(1.03100691224  + 54086.76194888209 *T)
    sumX0  +=       0.009 * cos(2.58296270636  + 130593.53833583929 *T)
    sumX0  +=       0.009 * cos(1.93570258198  + 26514.05573557170 *T)
    sumX0  +=       0.009 * cos(3.72792244923  + 25662.23818254370 *T)
    sumX0  +=       0.010 * cos(0.87681259763  + 101703.40155830728 *T)
    sumX0  +=       0.013 * cos(4.75978960525  + 26161.44408491669 *T)
    sumX0  +=       0.013 * cos(0.90383542596  + 26014.84983319870 *T)
    sumX0  +=       0.010 * cos(2.43967897596  + 104991.75367009427 *T)
    sumX0  +=       0.010 * cos(1.24342292300  + 52061.65099371849 *T)
    sumX0  +=       0.009 * cos(1.43295817370  + 25131.85780308709 *T)
    sumX0  +=       0.011 * cos(5.57723740117  + 183674.94773869208 *T)
    sumX0  +=       0.010 * cos(0.49259119482  + 28791.76311373330 *T)
    sumX0  +=       0.010 * cos(5.98998894020  + 4372.12018425770 *T)
    sumX0  +=       0.010 * cos(3.84228475284  + 158746.41977111547 *T)
    sumX0  +=       0.009 * cos(2.25433851888  + 45290.90021070109 *T)
    sumX0  +=       0.010 * cos(5.16716938460  + 23384.53080438210 *T)
    sumX0  +=       0.012 * cos(5.74392096253  + 76042.09660760908 *T)
    sumX0  +=       0.012 * cos(3.47156983874  + 52396.46274307070 *T)
    sumX0  +=       0.013 * cos(5.95557437514  + 1073.85284167431 *T)
    sumX0  +=       0.010 * cos(6.00837282954  + 103924.77055793729 *T)
    sumX0  +=       0.010 * cos(3.46837882181  + 91804.88680973028 *T)
    sumX0  +=       0.009 * cos(3.96963705288  + 103189.38167937248 *T)
    sumX0  +=       0.009 * cos(0.21796861386  + 156547.33242782846 *T)
    sumX0  +=       0.009 * cos(5.23543221261  + 735.63269604830 *T)
    sumX0  +=       0.009 * cos(1.95816420727  + 45494.33761226529 *T)
    sumX0  +=       0.009 * cos(1.00178607588  + 114.68310616872 *T)
    sumX0  +=       0.009 * cos(2.34791148855  + 130285.98071486969 *T)
    sumX0  +=       0.013 * cos(3.54674823405  + 2168.51666479910 *T)
    sumX0  +=       0.010 * cos(1.19765211170  + 32132.37554043310 *T)
    sumX0  +=       0.009 * cos(3.61118357787  + 104984.64012309349 *T)
    sumX0  +=       0.011 * cos(3.38074365009  + 19367.43297971630 *T)
    sumX0  +=       0.010 * cos(2.44976316483  + 90989.40667698107 *T)
    sumX0  +=       0.009 * cos(3.11869996984  + 25550.85481199529 *T)
    sumX0  +=       0.009 * cos(0.74322450432  + 76255.39570304709 *T)
    sumX0  +=       0.009 * cos(4.41624596522  + 31281.51401785550 *T)
    sumX0  +=       0.012 * cos(5.37072201396  + 1699.52303398670 *T)
    sumX0  +=       0.011 * cos(3.93547727008  + 220.16882495529 *T)
    sumX0  +=       0.009 * cos(2.41712554824  + 51130.89526444429 *T)
    sumX0  +=       0.011 * cos(1.25172069529  + 4022.50634944771 *T)
    sumX0  +=       0.009 * cos(1.96969875107  + 52797.79390147110 *T)
    sumX0  +=       0.009 * cos(5.38585404056  + 24844.65935737930 *T)
    sumX0  +=       0.009 * cos(0.27777099065  + 27331.63456073610 *T)
    sumX0  +=       0.010 * cos(1.25844371970  + 230239.41859459243 *T)
    sumX0  +=       0.009 * cos(4.03973091884  + 156314.36357149068 *T)
    sumX0  +=       0.011 * cos(3.17025202968  + 80175.15272542329 *T)
    sumX0  +=       0.009 * cos(2.05859040472  + 129329.89333829407 *T)
    sumX0  +=       0.008 * cos(2.65734511175  + 38634.62889814070 *T)
    sumX0  +=       0.011 * cos(5.69260296788  + 46046.75306112530 *T)
    sumX0  +=       0.009 * cos(0.74288994112  + 52190.27719463350 *T)
    sumX0  +=       0.009 * cos(1.39536424747  + 25454.87558477749 *T)
    sumX0  +=       0.009 * cos(4.15458502385  + 23439.69213358470 *T)
    sumX0  +=       0.008 * cos(5.65796860157  + 102762.29585352947 *T)
    sumX0  +=       0.011 * cos(2.58616852636  + 2538.49232173710 *T)
    sumX0  +=       0.010 * cos(3.10787144924  + 391318.79094109638 *T)
    sumX0  +=       0.008 * cos(1.00986052839  + 80174.66509045628 *T)
    sumX0  +=       0.008 * cos(5.80399155144  + 142871.80217575227 *T)
    sumX0  +=       0.009 * cos(5.83837700511  + 76152.30292882849 *T)
    sumX0  +=       0.010 * cos(0.76586105008  + 54060.95126404929 *T)
    sumX0  +=       0.009 * cos(4.80200590567  + 4.17597074660 *T)
    sumX0  +=       0.010 * cos(4.30161237035  + 35472.50033216589 *T)
    sumX0  +=       0.010 * cos(5.71617817105  + 125887.31839332868 *T)
    sumX0  +=       0.008 * cos(1.10337825429  + 27360.82798468490 *T)
    sumX0  +=       0.009 * cos(2.54194822126  + 1685.29593998510 *T)
    sumX0  +=       0.009 * cos(5.51043601697  + 58458.63831565629 *T)
    sumX0  +=       0.009 * cos(5.26195704910  + 25977.45300287129 *T)
    sumX0  +=       0.009 * cos(2.40795090884  + 1485.04688921970 *T)
    sumX0  +=       0.008 * cos(2.74586842231  + 2325.19322629910 *T)
    sumX0  +=       0.009 * cos(3.83628763541  + 22595.57539218810 *T)
    sumX0  +=       0.009 * cos(0.89908538700  + 52172.11794736879 *T)
    sumX0  +=       0.008 * cos(5.40598309652  + 26934.22979380890 *T)
    sumX0  +=       0.008 * cos(2.16624725476  + 1575.08961876571 *T)
    sumX0  +=       0.008 * cos(0.66150337713  + 51876.92370636270 *T)
    sumX0  +=       0.010 * cos(3.52742187765  + 2751.79141717511 *T)
    sumX0  +=       0.008 * cos(2.46382310966  + 10681.49435404891 *T)
    sumX0  +=       0.011 * cos(4.78783764732  + 28096.70449821670 *T)
    sumX0  +=       0.008 * cos(0.74099606624  + 77956.39562123650 *T)
    sumX0  +=       0.008 * cos(2.46215429910  + 52400.39489633380 *T)
    sumX0  +=       0.008 * cos(4.46259452267  + 54824.50492610490 *T)
    sumX0  +=       0.010 * cos(3.07618265192  + 240452.70414080343 *T)
    sumX0  +=       0.008 * cos(0.55741619385  + 77741.37582411229 *T)
    sumX0  +=       0.009 * cos(2.33738686148  + 24491.47288180609 *T)
    sumX0  +=       0.008 * cos(0.42516315180  + 42791.20909255670 *T)
    sumX0  +=       0.008 * cos(4.70625369395  + 52490.23628050151 *T)
    sumX0  +=       0.008 * cos(3.86744165099  + 156520.54911992789 *T)
    sumX0  +=       0.007 * cos(6.06395790305  + 52206.37696163630 *T)
    sumX0  +=       0.008 * cos(2.15830939123  + 17098.67925362531 *T)
    sumX0  +=       0.007 * cos(4.72162069749  + 50291.14893721450 *T)
    sumX0  +=       0.010 * cos(4.49723907752  + 22065.88442709350 *T)
    sumX0  +=       0.008 * cos(2.65028306976  + 51112.73601717959 *T)
    sumX0  +=       0.009 * cos(2.14235988572  + 104454.94915799888 *T)
    sumX0  +=       0.010 * cos(3.27688619254  + 441.06910236111 *T)
    sumX0  +=       0.010 * cos(1.07651496998  + 49.75280442070 *T)
    sumX0  +=       0.007 * cos(3.67002518147  + 7880.33297082250 *T)
    sumX0  +=       0.007 * cos(4.65878049481  + 52182.67601266569 *T)
    sumX0  +=       0.008 * cos(3.31050910615  + 52099.29639438929 *T)
    sumX0  +=       0.007 * cos(5.29571821879  + 122445.23108333028 *T)
    sumX0  +=       0.008 * cos(3.30784894855  + 66940.80150893469 *T)
    sumX0  +=       0.008 * cos(0.11825864367  + 187167.51930556167 *T)
    sumX0  +=       0.009 * cos(4.39445007595  + 156507.99290602890 *T)
    sumX0  +=       0.007 * cos(0.74222476670  + 51086.34130943830 *T)
    sumX0  +=       0.008 * cos(0.44253932824  + 430.77416162260 *T)
    sumX0  +=       0.008 * cos(5.53617232430  + 149288.98707532868 *T)
    sumX0  +=       0.008 * cos(5.41205762148  + 49956.80536147810 *T)
    sumX0  +=       0.007 * cos(0.95302460841  + 53132.33925660249 *T)
    sumX0  +=       0.008 * cos(2.62377772169  + 100908.79380384788 *T)
    sumX0  +=       0.007 * cos(3.70441477670  + 233681.99353955788 *T)
    sumX0  +=       0.007 * cos(2.34002720858  + 104248.76360956168 *T)
    sumX0  +=       0.008 * cos(5.03600399329  + 77314.77763323628 *T)
    sumX0  +=       0.008 * cos(0.25407509038  + 112231.94553711927 *T)
    sumX0  +=       0.007 * cos(1.59729628741  + 29396.61285443810 *T)
    sumX0  +=       0.007 * cos(4.06632874380  + 22779.68106367730 *T)
    sumX0  +=       0.009 * cos(3.00366222190  + 64741.71416564769 *T)
    sumX0  +=       0.007 * cos(5.95614295602  + 26290.40035423180 *T)
    sumX0  +=       0.007 * cos(5.99066738237  + 25885.89356388359 *T)
    sumX0  +=       0.009 * cos(3.58307674710  + 25863.31452838879 *T)
    sumX0  +=       0.007 * cos(3.41090615216  + 13655.61667369290 *T)
    sumX0  +=       0.007 * cos(3.03792135842  + 156740.96176236667 *T)
    sumX0  +=       0.008 * cos(4.55619991306  + 78571.51086317569 *T)
    sumX0  +=       0.007 * cos(3.14378763001  + 52663.68485188610 *T)
    sumX0  +=       0.007 * cos(1.56777604890  + 2118.52004289490 *T)
    sumX0  +=       0.008 * cos(4.91321996165  + 102872.98980971587 *T)
    sumX0  +=       0.010 * cos(3.40349349020  + 2703.37233719210 *T)
    sumX0  +=       0.007 * cos(0.43580348099  + 52074.08758236629 *T)
    sumX0  +=       0.007 * cos(2.04943751114  + 209659.06869168568 *T)
    sumX0  +=       0.007 * cos(3.67671851452  + 149846.43781352908 *T)
    sumX0  +=       0.007 * cos(4.30875978873  + 77417.87040745489 *T)
    sumX0  +=       0.009 * cos(1.15423490298  + 56260.03860733629 *T)
    sumX0  +=       0.007 * cos(3.50070425163  + 117892.78995130444 *T)
    sumX0  +=       0.009 * cos(2.88385440366  + 18073.94875613370 *T)
    sumX0  +=       0.007 * cos(4.35051921098  + 917.68646849830 *T)
    sumX0  +=       0.008 * cos(2.28326188901  + 32808.86093839910 *T)
    sumX0  +=       0.008 * cos(1.76307802021  + 207114.39605479327 *T)
    sumX0  +=       0.008 * cos(5.32581127623  + 3308.22207789689 *T)
    sumX0  +=       0.007 * cos(2.91658615206  + 23267.31581493150 *T)
    sumX0  +=       0.007 * cos(2.74703887915  + 28908.97810318390 *T)
    sumX0  +=       0.007 * cos(3.11128439659  + 37.88921815429 *T)
    sumX0  +=       0.007 * cos(6.13048599151  + 19202.50943415989 *T)
    sumX0  +=       0.007 * cos(5.70115634056  + 52145.72323962749 *T)
    sumX0  +=       0.008 * cos(4.37424371597  + 103395.56722780968 *T)
    sumX0  +=       0.007 * cos(1.57841023127  + 208276.87075920109 *T)
    sumX0  +=       0.009 * cos(1.81069664930  + 2014.73785431429 *T)
    sumX0  +=       0.007 * cos(1.08271855123  + 63786.60205900610 *T)
    sumX0  +=       0.007 * cos(2.83741634124  + 24712.37315921190 *T)
    sumX0  +=       0.007 * cos(2.64521183634  + 127099.14709865468 *T)
    sumX0  +=       0.008 * cos(0.10378895366  + 120417.92224525507 *T)
    sumX0  +=       0.008 * cos(2.96158619363  + 103815.05187168487 *T)
    sumX0  +=       0.007 * cos(0.99819510291  + 8989.22388794889 *T)
    sumX0  +=       0.008 * cos(1.62417504708  + 579.44376951530 *T)
    sumX0  +=       0.007 * cos(2.35955703224  + 1135.43305440970 *T)
    sumX0  +=       0.006 * cos(5.07797134858  + 27574.12708012290 *T)
    sumX0  +=       0.007 * cos(5.91237630109  + 102976.08258393448 *T)
    sumX0  +=       0.006 * cos(3.83307059240  + 28774.86905359890 *T)
    sumX0  +=       0.006 * cos(1.83055443881  + 23401.42486451650 *T)
    sumX0  +=       0.006 * cos(1.38927552842  + 80463.04058549309 *T)
    sumX0  +=       0.008 * cos(1.96535036231  + 51258.11981465010 *T)
    sumX0  +=       0.006 * cos(3.61870647781  + 23704.95394428149 *T)
    sumX0  +=       0.006 * cos(2.04491855340  + 28471.33997383390 *T)
    sumX0  +=       0.006 * cos(1.68248146387  + 55618.13741063030 *T)
    sumX0  +=       0.007 * cos(1.61837202878  + 53311.72697252510 *T)
    sumX0  +=       0.007 * cos(5.53141035478  + 1802.61580820530 *T)
    sumX0  +=       0.008 * cos(4.68334848992  + 102018.65999090829 *T)
    sumX0  +=       0.007 * cos(5.32851714437  + 79995.07559513870 *T)
    sumX0  +=       0.007 * cos(5.04612545544  + 52278.01261889750 *T)
    sumX0  +=       0.008 * cos(4.88698976103  + 1162.23088692429 *T)
    sumX0  +=       0.006 * cos(6.06855779639  + 53439.20746321010 *T)
    sumX0  +=       0.007 * cos(1.80548735062  + 29580.71852592730 *T)
    sumX0  +=       0.007 * cos(3.67914527744  + 27388.97377726269 *T)
    sumX0  +=       0.007 * cos(1.98447975377  + 24787.32014085270 *T)
    sumX0  +=       0.006 * cos(3.77150246945  + 106470.62024415868 *T)
    sumX0  +=       0.007 * cos(2.88869753965  + 104127.51158807840 *T)
    sumX0  +=       0.007 * cos(0.34091991598  + 27624.12370202710 *T)
    sumX0  +=       0.007 * cos(5.32270511523  + 24552.17021608830 *T)
    sumX0  +=       0.008 * cos(3.83070609952  + 948.93179148630 *T)
    sumX0  +=       0.007 * cos(5.94865065189  + 6680.98103591609 *T)
    sumX0  +=       0.008 * cos(0.77567524248  + 51861.86392076229 *T)
    sumX0  +=       0.008 * cos(3.58552572007  + 51554.30629979270 *T)
    sumX0  +=       0.007 * cos(4.23028089495  + 26720.44306340389 *T)
    sumX0  +=       0.007 * cos(2.84651949085  + 52214.18313626970 *T)
    sumX0  +=       0.007 * cos(5.58797516267  + 25572.68308796469 *T)
    sumX0  +=       0.007 * cos(0.07564986854  + 26603.61083015070 *T)
    sumX0  +=       0.007 * cos(5.38283352699  + 415.79630809560 *T)
    sumX0  +=       0.008 * cos(1.12257158980  + 54191.03177242970 *T)
    sumX0  +=       0.006 * cos(0.30285518092  + 24505.69997580769 *T)
    sumX0  +=       0.007 * cos(4.29992052709  + 2974.85377209451 *T)
    sumX0  +=       0.007 * cos(4.30362880096  + 48092.06159392750 *T)
    sumX0  +=       0.006 * cos(3.84286985256  + 78043.54059976728 *T)
    sumX0  +=       0.006 * cos(2.33701938638  + 19202.99706912689 *T)
    sumX0  +=       0.006 * cos(1.70009289223  + 53661.34080733510 *T)
    sumX0  +=       0.006 * cos(0.21405293391  + 104426.63798234760 *T)
    sumX0  +=       0.006 * cos(0.26081525757  + 27883.40540277870 *T)
    sumX0  +=       0.006 * cos(3.09307209285  + 156523.78133157067 *T)
    sumX0  +=       0.008 * cos(5.86283230643  + 51120.60032370579 *T)
    sumX0  +=       0.006 * cos(1.52305102135  + 226.07308589371 *T)
    sumX0  +=       0.008 * cos(0.51879015762  + 48997.90431006430 *T)
    sumX0  +=       0.006 * cos(4.68111037881  + 149756.95206568309 *T)
    sumX0  +=       0.007 * cos(1.82650871102  + 4083.74468922089 *T)
    sumX0  +=       0.008 * cos(3.80478888311  + 53259.13033292550 *T)
    sumX0  +=       0.006 * cos(5.07036834716  + 130006.04778747768 *T)
    sumX0  +=       0.007 * cos(1.36856874001  + 162810.73851692007 *T)
    sumX0  +=       0.006 * cos(2.21234472657  + 27441.89570407450 *T)
    sumX0  +=       0.006 * cos(3.45128030464  + 24734.39821404090 *T)
    sumX0  +=       0.007 * cos(4.80045013467  + 55503.69812194508 *T)
    sumX0  +=       0.008 * cos(3.49135266112  + 422.90985509640 *T)
    sumX0  +=       0.006 * cos(0.40283917391  + 76777.97312114088 *T)
    sumX0  +=       0.006 * cos(5.52356103113  + 155461.16718973868 *T)
    sumX0  +=       0.006 * cos(4.38638116269  + 50161.06842883410 *T)
    sumX0  +=       0.006 * cos(5.35041562795  + 78262.46876949779 *T)
    sumX0  +=       0.006 * cos(3.67530335628  + 223.10414047710 *T)
    sumX0  +=       0.006 * cos(3.76321444127  + 78225.82020656829 *T)
    sumX0  +=       0.006 * cos(5.36628440371  + 76468.69479848508 *T)
    sumX0  +=       0.006 * cos(4.67707783228  + 78597.80918297550 *T)
    sumX0  +=       0.007 * cos(1.16259543363  + 30110.40949102190 *T)
    sumX0  +=       0.006 * cos(3.49467754329  + 35077.61466449009 *T)
    sumX0  +=       0.007 * cos(1.41920132868  + 1884.65734593389 *T)
    sumX0  +=       0.007 * cos(4.54853659672  + 1082.83641481010 *T)
    sumX0  +=       0.006 * cos(1.96577053321  + 5193.12324131429 *T)
    sumX0  +=       0.006 * cos(5.17864942773  + 171292.90171019849 *T)
    sumX0  +=       0.006 * cos(2.86805602516  + 25984.56654987209 *T)
    sumX0  +=       0.006 * cos(4.20102757189  + 29864.57784479250 *T)
    sumX0  +=       0.006 * cos(1.46259745932  + 22311.71607332290 *T)
    sumX0  +=       0.006 * cos(4.38736983699  + 76991.27221657889 *T)
    sumX0  +=       0.005 * cos(0.68949263605  + 27410.82460658910 *T)
    sumX0  +=       0.005 * cos(4.97413239516  + 24765.46931152630 *T)
    sumX0  +=       0.006 * cos(1.69366740663  + 51092.96986833830 *T)
    sumX0  +=       0.007 * cos(3.84496680361  + 88476.75115345008 *T)
    sumX0  +=       0.007 * cos(2.78025711521  + 34102.34516198170 *T)
    sumX0  +=       0.005 * cos(1.26084797472  + 2686.47827705770 *T)
    sumX0  +=       0.006 * cos(5.55737889155  + 31903.25781869470 *T)
    sumX0  +=       0.006 * cos(0.10624613966  + 20273.03609942070 *T)
    sumX0  +=       0.006 * cos(2.86083846875  + 104874.43380187408 *T)
    sumX0  +=       0.005 * cos(2.16809698297  + 155475.39428374029 *T)
    sumX0  +=       0.005 * cos(6.14186631974  + 128747.59385585209 *T)
    sumX0  +=       0.005 * cos(5.05100044910  + 26500.51805593209 *T)
    sumX0  +=       0.005 * cos(0.61262458211  + 25675.77586218330 *T)
    sumX0  +=       0.007 * cos(4.13768447366  + 25234.46294233869 *T)
    sumX0  +=       0.005 * cos(5.69197138572  + 53551.82390047770 *T)
    sumX0  +=       0.005 * cos(0.34616736613  + 60170.57854265610 *T)
    sumX0  +=       0.005 * cos(6.05655614068  + 27353.71443768409 *T)
    sumX0  +=       0.007 * cos(1.18001608430  + 52602.16065654089 *T)
    sumX0  +=       0.005 * cos(2.35031471718  + 128106.07549751546 *T)
    sumX0  +=       0.005 * cos(2.02403333135  + 629.09540353360 *T)
    sumX0  +=       0.005 * cos(1.31146516422  + 35191.56631820829 *T)
    sumX0  +=       0.006 * cos(2.18238318065  + 3492.81538435310 *T)
    sumX0  +=       0.007 * cos(4.81830760175  + 169093.81436691148 *T)
    sumX0  +=       0.006 * cos(4.71794828628  + 81591.60126351927 *T)
    sumX0  +=       0.006 * cos(4.84978213740  + 25462.47676674530 *T)
    sumX0  +=       0.006 * cos(0.81384289381  + 26713.81715137009 *T)
    sumX0  +=       0.007 * cos(2.97261820220  + 2641.58509595570 *T)
    sumX0  +=       0.005 * cos(3.79651203289  + 58220.27868200730 *T)
    sumX0  +=       0.005 * cos(2.70097411082  + 52813.04637265610 *T)
    sumX0  +=       0.006 * cos(1.52077024889  + 156531.54400228668 *T)
    sumX0  +=       0.005 * cos(1.19912051466  + 26236.22568348400 *T)
    sumX0  +=       0.007 * cos(4.44424756295  + 9745.07673837309 *T)
    sumX0  +=       0.005 * cos(4.73755202618  + 26035.05080674330 *T)
    sumX0  +=       0.005 * cos(0.92607300503  + 26141.24311137209 *T)
    sumX0  +=       0.006 * cos(1.08638725845  + 117873.12019040468 *T)
    sumX0  +=       0.006 * cos(3.87149168929  + 28214.02436643689 *T)
    sumX0  +=       0.005 * cos(3.94775717381  + 487.39093377069 *T)
    sumX0  +=       0.006 * cos(1.27113821429  + 53764.63536094869 *T)
    sumX0  +=       0.005 * cos(4.36934946235  + 104035.46451412368 *T)
    sumX0  +=       0.006 * cos(4.83390346226  + 437.88770862340 *T)
    sumX0  +=       0.005 * cos(3.74810256898  + 52098.29955664799 *T)
    sumX0  +=       0.006 * cos(1.53446440230  + 98068.29289882188 *T)
    sumX0  +=       0.005 * cos(2.74138703040  + 65538.49980743109 *T)
    sumX0  +=       0.005 * cos(2.17866438309  + 189854.24140010288 *T)
    sumX0  +=       0.005 * cos(1.65056263900  + 76138.07583482689 *T)
    sumX0  +=       0.005 * cos(4.88320109198  + 104301.85976187608 *T)
    sumX0  +=       0.005 * cos(0.46719543123  + 1243.73141916191 *T)
    sumX0  +=       0.005 * cos(1.20837858717  + 118829.20756698029 *T)
    sumX0  +=       0.006 * cos(2.74486813288  + 13675.28643459270 *T)
    sumX0  +=       0.005 * cos(2.13216570647  + 153084.60008699627 *T)
    sumX0  +=       0.006 * cos(1.50606009977  + 13.98327651810 *T)
    sumX0  +=       0.005 * cos(0.04883294046  + 78697.66498008289 *T)
    sumX0  +=       0.005 * cos(3.63277479562  + 26610.23674218449 *T)
    sumX0  +=       0.005 * cos(5.00629124334  + 1052.02456570489 *T)
    sumX0  +=       0.005 * cos(3.70422350101  + 149.31937965110 *T)
    sumX0  +=       0.006 * cos(6.13258483711  + 122444.74344836328 *T)
    sumX0  +=       0.006 * cos(5.32441467055  + 24822.57948043130 *T)
    sumX0  +=       0.005 * cos(0.80608286136  + 26162.44092265799 *T)
    sumX0  +=       0.005 * cos(6.05508383397  + 154.02262796830 *T)
    sumX0  +=       0.005 * cos(0.11277607968  + 104401.85300568448 *T)
    sumX0  +=       0.005 * cos(0.32519606443  + 53654.91667469629 *T)
    sumX0  +=       0.005 * cos(5.10038810847  + 118828.71993201328 *T)
    sumX0  +=       0.005 * cos(6.11943807574  + 191.20444862810 *T)
    sumX0  +=       0.005 * cos(5.77124771600  + 124777.93984123527 *T)
    sumX0  +=       0.005 * cos(6.14333001423  + 51006.46184922329 *T)
    sumX0  +=       0.006 * cos(4.56143654817  + 104241.65006256088 *T)
    sumX0  +=       0.005 * cos(0.65495415751  + 104668.24825343688 *T)
    sumX0  +=       0.004 * cos(6.26437463075  + 76532.83088927349 *T)
    sumX0  +=       0.006 * cos(0.69389985894  + 75615.01078176608 *T)
    sumX0  +=       0.004 * cos(2.34685033934  + 24608.79275002629 *T)
    sumX0  +=       0.004 * cos(2.34685033934  + 24609.28038499330 *T)
    sumX0  +=       0.005 * cos(5.16993727796  + 23389.69551168050 *T)
    sumX0  +=       0.005 * cos(0.49368775325  + 28786.59840643490 *T)
    sumX0  +=       0.004 * cos(1.13517998278  + 31775.77732800970 *T)
    sumX0  +=       0.004 * cos(4.52844504843  + 20400.51659010570 *T)
    sumX0  +=       0.005 * cos(4.70021525766  + 107679.50440509344 *T)
    sumX0  +=       0.004 * cos(6.12765105459  + 13541.66501997470 *T)
    sumX0  +=       0.005 * cos(5.57977129989  + 29428.27175079049 *T)
    sumX0  +=       0.005 * cos(5.30763173110  + 26.05450231630 *T)
    sumX0  +=       0.004 * cos(3.86073160191  + 25764.64154240029 *T)
    sumX0  +=       0.005 * cos(6.09620933760  + 78265.43771491438 *T)
    sumX0  +=       0.006 * cos(2.35286352659  + 51688.41534937770 *T)
    sumX0  +=       0.006 * cos(6.24688491106  + 23865.80268949369 *T)
    sumX0  +=       0.004 * cos(2.73813240429  + 35211.72371407509 *T)
    sumX0  +=       0.004 * cos(2.92549262692  + 16964.57020404031 *T)
    sumX0  +=       0.004 * cos(5.93215370895  + 47804.17373385770 *T)
    sumX0  +=       0.004 * cos(3.83250941463  + 77623.56832092509 *T)
    sumX0  +=       0.005 * cos(0.50234970795  + 86458.22857541469 *T)
    sumX0  +=       0.004 * cos(1.08485860111  + 27040.22358231730 *T)
    sumX0  +=       0.004 * cos(4.57876643010  + 25136.07033579810 *T)
    sumX0  +=       0.005 * cos(4.59959229753  + 25778.86863640189 *T)
    sumX0  +=       0.005 * cos(1.06403273368  + 26397.42528171349 *T)
    sumX0  +=       0.005 * cos(4.02283478855  + 78374.15956342549 *T)
    sumX0  +=       0.004 * cos(3.47793310153  + 48834.95003896509 *T)
    sumX0  +=       0.005 * cos(5.01835542656  + 25771.26745443409 *T)
    sumX0  +=       0.004 * cos(1.67917095725  + 33480.60136114250 *T)
    sumX0  +=       0.004 * cos(3.98445407396  + 18695.69255697290 *T)
    sumX0  +=       0.005 * cos(0.59989929417  + 703.87700210090 *T)
    sumX0  +=       0.004 * cos(5.10698405361  + 156378.09946979408 *T)
    sumX0  +=       0.005 * cos(0.28534518128  + 155467.79310177249 *T)
    sumX0  +=       0.005 * cos(1.29884621572  + 26404.05119374729 *T)
    sumX0  +=       0.004 * cos(0.18983170026  + 9123.33293753389 *T)
    sumX0  +=       0.005 * cos(2.32620215234  + 2015.91490364330 *T)
    sumX0  +=       0.004 * cos(0.61980953649  + 76998.38576357969 *T)
    sumX0  +=       0.004 * cos(0.13065109470  + 129586.56314360249 *T)
    sumX0  +=       0.004 * cos(2.99427045342  + 1368.90407032850 *T)
    sumX0  +=       0.004 * cos(1.49149882620  + 45425.00926028610 *T)
    sumX0  +=       0.004 * cos(5.39873919994  + 28928.64786408370 *T)
    sumX0  +=       0.004 * cos(0.26488583127  + 23247.64605403170 *T)
    sumX0  +=       0.004 * cos(5.36778217078  + 23866.29032446070 *T)
    sumX0  +=       0.004 * cos(2.35188051852  + 20453.11322970530 *T)
    sumX0  +=       0.004 * cos(3.31174451269  + 31723.18068841010 *T)
    sumX0  +=       0.004 * cos(2.30331784549  + 26239.19462890059 *T)
    sumX0  +=       0.004 * cos(1.97810159548  + 199599.56195595945 *T)
    sumX0  +=       0.004 * cos(2.77836300783  + 52356.12723091650 *T)
    sumX0  +=       0.004 * cos(1.24931397565  + 26137.65594599489 *T)
    sumX0  +=       0.004 * cos(3.46342382133  + 51539.05382860770 *T)
    sumX0  +=       0.004 * cos(3.69944590805  + 32279.16997330450 *T)
    sumX0  +=       0.004 * cos(1.96417912316  + 19897.12394481090 *T)
    sumX0  +=       0.004 * cos(1.70188026990  + 78905.36680394509 *T)
    sumX0  +=       0.004 * cos(3.00071021966  + 52336.45747001670 *T)
    sumX0  +=       0.004 * cos(5.68676014339  + 50912.89273805370 *T)
    sumX0  +=       0.005 * cos(0.53917560687  + 26085.17801364109 *T)
    sumX0  +=       0.005 * cos(5.12444942434  + 26091.11590447430 *T)
    sumX0  +=       0.004 * cos(6.12827810146  + 25441.13612574289 *T)
    sumX0  +=       0.004 * cos(3.34711834324  + 25923.71104846829 *T)
    sumX0  +=       0.004 * cos(2.31650668797  + 26252.58286964711 *T)
    sumX0  +=       0.004 * cos(0.93197083282  + 24925.18461968289 *T)
    sumX0  +=       0.005 * cos(0.84223070790  + 13706.10093056410 *T)
    sumX0  +=       0.005 * cos(4.82139432331  + 38470.19298755129 *T)
    sumX0  +=       0.005 * cos(4.69896171289  + 18227.72756661850 *T)
    sumX0  +=       0.005 * cos(0.96466331832  + 33948.56635149690 *T)
    sumX0  +=       0.005 * cos(6.00043896187  + 10.80188278040 *T)
    sumX0  +=       0.004 * cos(2.80981383368  + 147423.26803784407 *T)
    sumX0  +=       0.004 * cos(3.37539797378  + 120225.98634417648 *T)
    sumX0  +=       0.004 * cos(2.65298230712  + 95.73540973430 *T)
    sumX0  +=       0.004 * cos(4.35824496747  + 74935.81758592589 *T)
    sumX0  +=       0.004 * cos(0.27953957902  + 142871.31454078527 *T)
    sumX0  +=       0.004 * cos(0.37589110708  + 26131.03003396110 *T)
    sumX0  +=       0.004 * cos(5.28773392413  + 26045.26388415430 *T)
    sumX0  +=       0.004 * cos(3.24717162624  + 1066.25165970650 *T)
    sumX0  +=       0.004 * cos(0.73794406026  + 25035.39094090229 *T)
    sumX0  +=       0.005 * cos(4.22425932829  + 1577.09972496430 *T)
    sumX0  +=       0.004 * cos(0.56307156026  + 1262.91354509469 *T)
    sumX0  +=       0.005 * cos(6.00635216618  + 51102.44107644109 *T)
    sumX0  +=       0.004 * cos(1.12630290328  + 106262.56823203048 *T)
    sumX0  +=       0.004 * cos(1.48479477953  + 104771.34102765548 *T)
    sumX0  +=       0.005 * cos(1.39791854337  + 179172.99086353747 *T)
    sumX0  +=       0.004 * cos(3.88144405304  + 52329.34127614969 *T)
    sumX0  +=       0.005 * cos(1.62731568621  + 151200.18655854589 *T)
    sumX0  +=       0.004 * cos(3.34025458568  + 18319.78040236310 *T)
    sumX0  +=       0.004 * cos(2.32337044553  + 33856.51351575230 *T)
    sumX0  +=       0.005 * cos(3.80744702100  + 51315.74017187909 *T)
    sumX0  +=       0.004 * cos(3.26438690958  + 159970.23761189409 *T)
    sumX0  +=       0.005 * cos(5.02554220823  + 107692.46881048007 *T)
    sumX0  +=       0.004 * cos(3.12725019808  + 24182.19455915029 *T)
    sumX0  +=       0.004 * cos(3.79384310544  + 25021.16384690069 *T)
    sumX0  +=       0.005 * cos(3.05069290377  + 49637.80159637830 *T)
    sumX0  +=       0.004 * cos(3.62578024170  + 64901.01590043989 *T)
    sumX0  +=       0.005 * cos(0.63333030208  + 77930.09730143669 *T)
    sumX0  +=       0.004 * cos(3.70617022011  + 106685.05277665229 *T)
    sumX0  +=       0.004 * cos(2.87321007380  + 25249.17767130730 *T)
    sumX0  +=       0.004 * cos(2.79041495741  + 26927.11624680809 *T)
    sumX0  +=       0.005 * cos(1.54666149626  + 130020.27488147929 *T)
    sumX0  +=       0.004 * cos(1.73976987963  + 42430.72954477530 *T)
    sumX0  +=       0.004 * cos(3.38176805408  + 189386.27640974848 *T)
    sumX0  +=       0.004 * cos(3.25018015096  + 130907.72451570888 *T)
    sumX0  +=       0.004 * cos(2.06353106693  + 76352.75375898890 *T)
    sumX0  +=       0.004 * cos(5.31959837801  + 78050.16651180109 *T)
    sumX0  +=       0.004 * cos(2.24224128580  + 26507.14396796589 *T)
    sumX0  +=       0.004 * cos(3.53906955839  + 21715.78295731649 *T)
    sumX0  +=       0.004 * cos(3.82039111865  + 50690.75939392870 *T)
    sumX0  +=       0.003 * cos(5.68926517353  + 76095.19275992349 *T)
    sumX0  +=       0.004 * cos(0.40278258619  + 27047.26981901530 *T)
    sumX0  +=       0.004 * cos(5.26084244502  + 25129.02409910010 *T)
    sumX0  +=       0.003 * cos(4.44502926835  + 25980.84165212810 *T)
    sumX0  +=       0.003 * cos(1.21859576286  + 26195.45226598730 *T)
    sumX0  +=       0.004 * cos(4.63561794007  + 52165.00440036799 *T)
    sumX0  +=       0.004 * cos(2.69531678715  + 104881.05971390787 *T)
    sumX0  +=       0.004 * cos(4.60182936648  + 8044.76888141190 *T)
    sumX0  +=       0.005 * cos(0.15112505254  + 45455.33612129050 *T)
    sumX0  +=       0.004 * cos(5.01042912978  + 25565.08190599689 *T)
    sumX0  +=       0.004 * cos(2.55785749371  + 215942.14454167709 *T)
    sumX0  +=       0.004 * cos(5.41986884632  + 25992.16773183990 *T)
    sumX0  +=       0.004 * cos(0.24375618489  + 26184.12618627550 *T)
    sumX0  +=       0.003 * cos(0.06938261642  + 129909.58092529287 *T)
    sumX0  +=       0.004 * cos(4.35391878884  + 48153.78756866770 *T)
    sumX0  +=       0.005 * cos(1.35637771566  + 10059.75055320970 *T)
    sumX0  +=       0.004 * cos(0.44819649652  + 16496.60521368590 *T)
    sumX0  +=       0.004 * cos(5.21542853469  + 35679.68870442950 *T)
    sumX0  +=       0.004 * cos(1.03867130112  + 103395.76900720468 *T)
    sumX0  +=       0.004 * cos(3.88762025023  + 51040.37322873870 *T)
    sumX0  +=       0.003 * cos(5.63440461335  + 125111.79578200469 *T)
    sumX0  +=       0.004 * cos(2.46419290723  + 128843.57308306987 *T)
    sumX0  +=       0.004 * cos(2.48121871677  + 52015.64273124709 *T)
    sumX0  +=       0.004 * cos(5.17960970329  + 621.98761832270 *T)
    sumX0  +=       0.003 * cos(1.35017943249  + 23113.53700444670 *T)
    sumX0  +=       0.003 * cos(4.31344559872  + 29062.75691366870 *T)
    sumX0  +=       0.003 * cos(0.13992300676  + 25794.73617468050 *T)
    sumX0  +=       0.003 * cos(5.52370202445  + 26381.55774343490 *T)
    sumX0  +=       0.003 * cos(1.65047629311  + 78283.13536813887 *T)
    sumX0  +=       0.004 * cos(6.01757036855  + 26037.66270218649 *T)
    sumX0  +=       0.004 * cos(0.78738841944  + 23543.47432216530 *T)
    sumX0  +=       0.004 * cos(4.87623661177  + 28632.81959595010 *T)
    sumX0  +=       0.003 * cos(2.34677335126  + 26279.59522516930 *T)
    sumX0  +=       0.003 * cos(3.31685167995  + 25896.69869294610 *T)
    sumX0  +=       0.004 * cos(4.17686572959  + 25998.66121121170 *T)
    sumX0  +=       0.004 * cos(1.48675930162  + 26177.63270690369 *T)
    sumX0  +=       0.003 * cos(1.45345885107  + 24079.10178493169 *T)
    sumX0  +=       0.003 * cos(2.83722079462  + 49851.10069181629 *T)
    sumX0  +=       0.004 * cos(0.04334468913  + 14235.79189565871 *T)
    sumX0  +=       0.004 * cos(5.62028034208  + 37940.50202245669 *T)
    sumX0  +=       0.003 * cos(3.19680435173  + 41494.79956406649 *T)
    sumX0  +=       0.004 * cos(3.72413363777  + 261988.65378531881 *T)
    sumX0  +=       0.003 * cos(3.23280579449  + 61560.40347374009 *T)
    sumX0  +=       0.004 * cos(6.14807840422  + 32217.44399856430 *T)
    sumX0  +=       0.003 * cos(5.81361645381  + 78243.79584633929 *T)
    sumX0  +=       0.004 * cos(2.70356592860  + 51995.97297034730 *T)
    sumX0  +=       0.003 * cos(0.89067609209  + 154308.90556274189 *T)
    sumX0  +=       0.003 * cos(0.19486754591  + 78398.06229179110 *T)
    sumX0  +=       0.003 * cos(3.22218438408  + 202420.39310008567 *T)
    sumX0  +=       0.003 * cos(4.85336780642  + 79117.14962395809 *T)
    sumX0  +=       0.004 * cos(4.81266956480  + 28420.85571696270 *T)
    sumX0  +=       0.003 * cos(4.64236443125  + 313054.59388140670 *T)
    sumX0  +=       0.003 * cos(1.82348628585  + 26114.44527885750 *T)
    sumX0  +=       0.003 * cos(3.84013874536  + 26061.84863925790 *T)
    sumX0  +=       0.003 * cos(6.16611553760  + 116783.41139921108 *T)
    sumX0  +=       0.003 * cos(6.22418249204  + 2169.00429976610 *T)
    sumX0  +=       0.003 * cos(3.34714723948  + 32131.88790546609 *T)
    sumX0  +=       0.003 * cos(2.95057028322  + 68241.62832713968 *T)
    sumX0  +=       0.004 * cos(2.12743436214  + 49424.50250094030 *T)
    sumX0  +=       0.003 * cos(3.11077916313  + 27993.61172399810 *T)
    sumX0  +=       0.004 * cos(3.05439250241  + 156954.26085780468 *T)
    sumX0  +=       0.004 * cos(3.67142845852  + 52021.78365518009 *T)
    sumX0  +=       0.004 * cos(5.73212182199  + 192828.85135471387 *T)
    sumX0  +=       0.003 * cos(0.27361838208  + 25345.15689852510 *T)
    sumX0  +=       0.003 * cos(1.81581655351  + 87648.79425028129 *T)
    sumX0  +=       0.003 * cos(0.68500664823  + 105410.75067900248 *T)
    sumX0  +=       0.003 * cos(3.93077791188  + 130446.38808425447 *T)
    sumX0  +=       0.003 * cos(6.05546250987  + 45290.41257573408 *T)
    sumX0  +=       0.003 * cos(1.95812396188  + 251775.36823910783 *T)
    sumX0  +=       0.004 * cos(1.27498750708  + 50952.23225985329 *T)
    sumX0  +=       0.003 * cos(1.43670342990  + 23421.09462541629 *T)
    sumX0  +=       0.003 * cos(4.22692160131  + 28755.19929269911 *T)
    sumX0  +=       0.003 * cos(3.98270762104  + 163297.88563320728 *T)
    sumX0  +=       0.003 * cos(3.47547087612  + 26727.55661040469 *T)
    sumX0  +=       0.004 * cos(3.43694245255  + 24711.88552424489 *T)
    sumX0  +=       0.003 * cos(3.68091678909  + 111590.04434041328 *T)
    sumX0  +=       0.004 * cos(0.29131403040  + 50476.77088412869 *T)
    sumX0  +=       0.003 * cos(4.21457884017  + 52072.26791205130 *T)
    sumX0  +=       0.003 * cos(3.61792324623  + 77622.53968046710 *T)
    sumX0  +=       0.003 * cos(0.47995189315  + 52016.74836583970 *T)
    sumX0  +=       0.003 * cos(1.68763185836  + 25970.82709083750 *T)
    sumX0  +=       0.003 * cos(3.97599317285  + 26205.46682727790 *T)
    sumX0  +=       0.004 * cos(5.27937253398  + 27511.71169102070 *T)
    sumX0  +=       0.004 * cos(0.38425249723  + 24664.58222709470 *T)
    sumX0  +=       0.003 * cos(2.30032754148  + 33918.23949049249 *T)
    sumX0  +=       0.003 * cos(3.36329748973  + 18258.05442762290 *T)
    sumX0  +=       0.004 * cos(4.83285053411  + 112545.64408202188 *T)
    sumX0  +=       0.004 * cos(3.76448219227  + 61921.37065648850 *T)
    sumX0  +=       0.003 * cos(6.06735340898  + 51741.85072778809 *T)
    sumX0  +=       0.003 * cos(2.88367503068  + 222225.22039166847 *T)
    sumX0  +=       0.003 * cos(1.15891911019  + 78249.72614820449 *T)
    sumX0  +=       0.004 * cos(3.05981714265  + 158116.73556221247 *T)
    sumX0  +=       0.003 * cos(3.68443191743  + 73891.58924046489 *T)
    sumX0  +=       0.003 * cos(1.88338582885  + 26387.27335332690 *T)
    sumX0  +=       0.003 * cos(6.09657221455  + 12725.69725225850 *T)
    sumX0  +=       0.003 * cos(5.04796024742  + 26984.22641571310 *T)
    sumX0  +=       0.003 * cos(0.61566478379  + 25192.06750240230 *T)
    sumX0  +=       0.003 * cos(3.48855867987  + 103498.17236706128 *T)
    sumX0  +=       0.003 * cos(0.74835060715  + 79487.77108298469 *T)
    sumX0  +=       0.003 * cos(1.44150895108  + 50579.37602338029 *T)
    sumX0  +=       0.003 * cos(3.44346861237  + 143980.69309287865 *T)
    sumX0  +=       0.003 * cos(5.63232630472  + 91805.37444469729 *T)
    sumX0  +=       0.003 * cos(3.11959949863  + 50490.99797813029 *T)
    sumX0  +=       0.003 * cos(2.99543245253  + 54879.66625530750 *T)
    sumX0  +=       0.003 * cos(4.69667538764  + 55484.51599601230 *T)
    sumX0  +=       0.003 * cos(4.64572031653  + 2332.95257538850 *T)
    sumX0  +=       0.003 * cos(1.39528143088  + 1383.13116433011 *T)
    sumX0  +=       0.003 * cos(1.78559566583  + 128106.56313248246 *T)
    sumX0  +=       0.003 * cos(0.70542302818  + 1471.99684454710 *T)
    sumX0  +=       0.003 * cos(4.72601280222  + 146505.82538682927 *T)
    sumX0  +=       0.004 * cos(4.23386749461  + 102232.60488843489 *T)
    sumX0  +=       0.003 * cos(5.04493378788  + 24510.80341660990 *T)
    sumX0  +=       0.003 * cos(0.61869124333  + 27665.49050150550 *T)
    sumX0  +=       0.004 * cos(4.60583023308  + 26365.18195279910 *T)
    sumX0  +=       0.004 * cos(1.05779479813  + 25811.11196531629 *T)
    sumX0  +=       0.003 * cos(0.02744721053  + 106263.05586699749 *T)
    sumX0  +=       0.003 * cos(3.49736618564  + 50601.20429934970 *T)
    sumX0  +=       0.003 * cos(0.30011567778  + 179967.11098302988 *T)
    sumX0  +=       0.003 * cos(1.53690934138  + 210921.73841929680 *T)
    sumX0  +=       0.003 * cos(1.31624723631  + 26202.54606597110 *T)
    sumX0  +=       0.003 * cos(2.75823464337  + 2821.07496160971 *T)
    sumX0  +=       0.003 * cos(6.26421306637  + 5020.16230489671 *T)
    sumX0  +=       0.003 * cos(4.80001740407  + 229463.89598326848 *T)
    sumX0  +=       0.003 * cos(6.02378503125  + 71378.31571730827 *T)
    sumX0  +=       0.003 * cos(1.42673173544  + 29001.03093892850 *T)
    sumX0  +=       0.003 * cos(4.23689329577  + 23175.26297918689 *T)
    sumX0  +=       0.003 * cos(0.68304283752  + 15406.40878752529 *T)
    sumX0  +=       0.003 * cos(4.32693015716  + 25940.06823463139 *T)
    sumX0  +=       0.003 * cos(2.38990502393  + 220026.13304838145 *T)
    sumX0  +=       0.003 * cos(0.59649942463  + 963.64652045490 *T)
    sumX0  +=       0.003 * cos(2.87116210123  + 113455.76337789788 *T)
    sumX0  +=       0.003 * cos(1.35908128792  + 49201.44014602090 *T)
    sumX0  +=       0.003 * cos(5.82324458614  + 143165.21296012949 *T)
    sumX0  +=       0.003 * cos(2.38673232157  + 51735.22481575429 *T)
    sumX0  +=       0.003 * cos(2.84532132702  + 25914.69838767311 *T)
    sumX0  +=       0.003 * cos(2.81830370419  + 26261.59553044229 *T)
    sumX0  +=       0.003 * cos(5.22041687007  + 51745.51975649279 *T)
    sumX0  +=       0.002 * cos(3.03566163825  + 25554.52384069999 *T)
    sumX0  +=       0.002 * cos(2.62796339296  + 26621.77007741539 *T)
    sumX0  +=       0.003 * cos(5.20753478644  + 25530.69622085730 *T)
    sumX0  +=       0.003 * cos(0.45609024477  + 26645.59769725810 *T)
    sumX0  +=       0.002 * cos(3.73081171564  + 99979.49238203908 *T)
    sumX0  +=       0.003 * cos(5.03565808597  + 25626.81052760331 *T)
    sumX0  +=       0.003 * cos(0.62796694524  + 26549.48339051209 *T)
    sumX0  +=       0.003 * cos(5.59916996136  + 23976.49664568010 *T)
    sumX0  +=       0.003 * cos(5.34361577728  + 20956.50587500010 *T)
    sumX0  +=       0.003 * cos(0.32000925393  + 31219.78804311530 *T)
    sumX0  +=       0.002 * cos(0.74211300250  + 242985.64742485987 *T)
    sumX0  +=       0.002 * cos(4.68727528203  + 133767.40754666767 *T)
    sumX0  +=       0.003 * cos(4.44342016483  + 283524.60342983418 *T)
    sumX0  +=       0.002 * cos(1.15603737137  + 127790.81706491447 *T)
    sumX0  +=       0.002 * cos(4.29763002496  + 127791.30469988148 *T)
    sumX0  +=       0.002 * cos(5.43119958754  + 71378.80335227528 *T)
    sumX0  +=       0.002 * cos(4.58931827360  + 21202.33752122950 *T)
    sumX0  +=       0.002 * cos(1.07430675761  + 30973.95639688590 *T)
    sumX0  +=       0.003 * cos(3.47073853524  + 2353.10997125530 *T)
    sumX0  +=       0.003 * cos(2.48443590939  + 76358.48847726568 *T)
    sumX0  +=       0.003 * cos(0.03897228819  + 24292.40088036969 *T)
    sumX0  +=       0.003 * cos(2.04739291424  + 157636.55358405507 *T)
    sumX0  +=       0.003 * cos(0.83289676584  + 25361.66493918750 *T)
    sumX0  +=       0.003 * cos(4.83072826537  + 26814.62897892790 *T)
    sumX0  +=       0.002 * cos(1.24497717708  + 57369.41715942969 *T)
    sumX0  +=       0.003 * cos(5.41525530380  + 52187.09580089580 *T)
    sumX0  +=       0.003 * cos(4.02922617142  + 49953.70583106789 *T)
    sumX0  +=       0.002 * cos(1.46800292953  + 110396.08496515568 *T)
    sumX0  +=       0.002 * cos(2.56755430966  + 273311.31788362318 *T)
    sumX0  +=       0.002 * cos(0.95339846585  + 7994.28462454070 *T)
    sumX0  +=       0.002 * cos(2.63620968801  + 31127.73520737070 *T)
    sumX0  +=       0.002 * cos(3.02741534320  + 21048.55871074470 *T)
    sumX0  +=       0.002 * cos(0.16837305460  + 27133.30179524529 *T)
    sumX0  +=       0.002 * cos(3.81308180912  + 289185.93547898642 *T)
    sumX0  +=       0.002 * cos(0.71138534841  + 5634.78991186890 *T)
    sumX0  +=       0.002 * cos(1.76112647723  + 108903.32224587208 *T)
    sumX0  +=       0.003 * cos(6.06940283285  + 128857.80017707148 *T)
    sumX0  +=       0.002 * cos(4.81801153794  + 5687.38655146850 *T)
    sumX0  +=       0.003 * cos(4.73128149211  + 21650.63275068411 *T)
    sumX0  +=       0.003 * cos(0.93234353910  + 30525.66116743129 *T)
    sumX0  +=       0.003 * cos(0.49698769500  + 117.56368570370 *T)
    sumX0  +=       0.002 * cos(2.18169836096  + 29266.29274962530 *T)
    sumX0  +=       0.002 * cos(5.80090003502  + 24176.45984087349 *T)
    sumX0  +=       0.002 * cos(3.25477992512  + 102129.99974918328 *T)
    sumX0  +=       0.003 * cos(4.34402660372  + 53338.52480503969 *T)
    sumX0  +=       0.003 * cos(6.27379601640  + 67608.51339047348 *T)
    sumX0  +=       0.002 * cos(3.21341452307  + 267649.98583447101 *T)
    sumX0  +=       0.002 * cos(1.92153809084  + 544.16187657970 *T)
    sumX0  +=       0.002 * cos(0.63318713248  + 54862.77219517310 *T)
    sumX0  +=       0.002 * cos(4.71897270316  + 162188.50708111387 *T)
    sumX0  +=       0.002 * cos(5.95465982913  + 123554.12200045666 *T)
    sumX0  +=       0.003 * cos(3.65120791864  + 47803.68609889069 *T)
    sumX0  +=       0.002 * cos(3.82189241243  + 112546.13171698888 *T)
    sumX0  +=       0.002 * cos(0.77739174044  + 71024.79001563908 *T)
    sumX0  +=       0.003 * cos(3.39676430082  + 24971.65485996349 *T)
    sumX0  +=       0.003 * cos(2.26686073039  + 27204.63905815190 *T)
    sumX0  +=       0.002 * cos(5.01817813007  + 26165.89750304159 *T)
    sumX0  +=       0.003 * cos(4.38762823834  + 14477.59500068350 *T)
    sumX0  +=       0.002 * cos(6.05328448229  + 97466.21885888244 *T)
    sumX0  +=       0.002 * cos(5.24033215662  + 8482.40701076191 *T)
    sumX0  +=       0.003 * cos(3.13012289596  + 206584.70508969869 *T)
    sumX0  +=       0.003 * cos(0.28818731494  + 25881.47377565349 *T)
    sumX0  +=       0.002 * cos(0.63582632095  + 105306.96849042187 *T)
    sumX0  +=       0.003 * cos(6.00821622727  + 7768.12273921109 *T)
    sumX0  +=       0.002 * cos(0.75081825828  + 263097.54470244516 *T)
    sumX0  +=       0.002 * cos(0.75081825828  + 263098.03233741224 *T)
    sumX0  +=       0.002 * cos(0.39253998165  + 90988.91904201408 *T)
    sumX0  +=       0.003 * cos(3.01604861450  + 5814.86704215349 *T)
    sumX0  +=       0.003 * cos(0.86557994215  + 49976.96275734490 *T)
    sumX0  +=       0.003 * cos(2.68824003168  + 27161.75598324850 *T)
    sumX0  +=       0.002 * cos(6.12457483767  + 27151.46104250999 *T)
    sumX0  +=       0.003 * cos(1.61229350276  + 29051.02756083270 *T)
    sumX0  +=       0.003 * cos(4.05133152845  + 23125.26635728270 *T)
    sumX0  +=       0.002 * cos(3.41958084674  + 32973.29684898851 *T)
    sumX0  +=       0.003 * cos(0.44711181004  + 134.35286706851 *T)
    sumX0  +=       0.002 * cos(0.64120353129  + 78484.36588464488 *T)
    sumX0  +=       0.002 * cos(0.13220595262  + 50373.67810991009 *T)
    sumX0  +=       0.003 * cos(2.58526525309  + 246114.03618995569 *T)
    sumX0  +=       0.002 * cos(5.24756751342  + 19336.61848374489 *T)
    sumX0  +=       0.002 * cos(2.97251385048  + 25014.53793486689 *T)
    sumX0  +=       0.002 * cos(0.46042924320  + 156534.77621392949 *T)
    sumX0  +=       0.003 * cos(4.30654322236  + 78115.87451777978 *T)
    sumX0  +=       0.002 * cos(0.28012264229  + 51760.49761001979 *T)
    sumX0  +=       0.002 * cos(1.24434330997  + 25867.24668165189 *T)
    sumX0  +=       0.002 * cos(3.40478353479  + 26089.14379679899 *T)
    sumX0  +=       0.002 * cos(5.12603604161  + 30486.32164563170 *T)
    sumX0  +=       0.002 * cos(0.53758898960  + 21689.97227248370 *T)
    sumX0  +=       0.003 * cos(3.31169456914  + 78112.90557236319 *T)
    sumX0  +=       0.002 * cos(4.36305168844  + 1898.59503542310 *T)
    sumX0  +=       0.002 * cos(5.21464317983  + 235746.97183325983 *T)
    sumX0  +=       0.002 * cos(4.71956373810  + 26151.88285736110 *T)
    sumX0  +=       0.002 * cos(0.94406129310  + 26024.41106075430 *T)
    sumX0  +=       0.002 * cos(5.77039379556  + 50049.68505828569 *T)
    sumX0  +=       0.003 * cos(4.30724193035  + 42116.54336490570 *T)
    sumX0  +=       0.002 * cos(0.39973619044  + 61165.51780606429 *T)
    sumX0  +=       0.002 * cos(1.17288571933  + 249268.72327485128 *T)
    sumX0  +=       0.002 * cos(4.26445898022  + 184526.76529172008 *T)
    sumX0  +=       0.003 * cos(6.06473923651  + 26226.66445592840 *T)
    sumX0  +=       0.003 * cos(5.88207110188  + 25949.62946218700 *T)
    sumX0  +=       0.002 * cos(4.41949123763  + 223180.33249831005 *T)
    sumX0  +=       0.002 * cos(4.41949123763  + 223180.82013327704 *T)
    sumX0  +=       0.003 * cos(2.07289744094  + 65851.58035270809 *T)
    sumX0  +=       0.002 * cos(3.33696976419  + 4525.89899474250 *T)
    sumX0  +=       0.003 * cos(1.90326122005  + 26376.03481912750 *T)
    sumX0  +=       0.003 * cos(3.76036381116  + 25800.25909898790 *T)
    sumX0  +=       0.002 * cos(1.19438265065  + 26049.52628845289 *T)
    sumX0  +=       0.002 * cos(5.14117998003  + 4398.41850405751 *T)
    sumX0  +=       0.003 * cos(3.79796015832  + 23581.50199480110 *T)
    sumX0  +=       0.003 * cos(1.86566487289  + 28594.79192331430 *T)
    sumX0  +=       0.002 * cos(3.55520237633  + 1593.24886603040 *T)
    sumX0  +=       0.002 * cos(3.13962245506  + 78267.34694259708 *T)
    sumX0  +=       0.002 * cos(0.92505130149  + 54508.75885853690 *T)
    sumX0  +=       0.002 * cos(4.31177771871  + 78476.76470267709 *T)
    sumX0  +=       0.002 * cos(0.41611414366  + 1280.03839011150 *T)
    sumX0  +=       0.002 * cos(1.10174173283  + 35077.12702952309 *T)
    sumX0  +=       0.003 * cos(1.43403842688  + 26124.79552198719 *T)
    sumX0  +=       0.003 * cos(4.22958660433  + 26051.49839612820 *T)
    sumX0  +=       0.002 * cos(1.72941826067  + 158745.93213614845 *T)
    sumX0  +=       0.002 * cos(5.48576882408  + 52367.49836674349 *T)
    sumX0  +=       0.002 * cos(6.20502032580  + 41520.61024889929 *T)
    sumX0  +=       0.002 * cos(4.14056597313  + 51950.22083222170 *T)
    sumX0  +=       0.002 * cos(2.55940503302  + 28133.45549185990 *T)
    sumX0  +=       0.002 * cos(3.10421999819  + 24042.83842625550 *T)
    sumX0  +=       0.003 * cos(2.17160326158  + 51753.38406301899 *T)
    sumX0  +=       0.003 * cos(6.15199422925  + 25633.23759253039 *T)
    sumX0  +=       0.003 * cos(5.79481610914  + 26543.05632558499 *T)
    sumX0  +=       0.002 * cos(3.97512757833  + 24999.54380245671 *T)
    sumX0  +=       0.002 * cos(1.68849745288  + 27176.75011565869 *T)
    sumX0  +=       0.002 * cos(0.30844272162  + 309.03450517230 *T)
    sumX0  +=       0.002 * cos(2.40438367720  + 1155.60497489050 *T)
    sumX0  +=       0.002 * cos(0.34200603730  + 236702.57157486849 *T)
    sumX0  +=       0.002 * cos(3.00390396144  + 76887.69180739328 *T)
    sumX0  +=       0.002 * cos(0.98682277986  + 4084.23232418790 *T)
    sumX0  +=       0.002 * cos(1.59074157843  + 24601.67920302549 *T)
    sumX0  +=       0.002 * cos(2.80906034932  + 27818.78167702329 *T)
    sumX0  +=       0.002 * cos(2.26687572256  + 2198.84352580350 *T)
    sumX0  +=       0.002 * cos(4.53134788435  + 24292.88851533670 *T)
    sumX0  +=       0.002 * cos(1.58580178730  + 22759.52366781050 *T)
    sumX0  +=       0.002 * cos(2.91202347140  + 27143.59673598380 *T)
    sumX0  +=       0.002 * cos(1.70541081745  + 25973.22003540546 *T)
    sumX0  +=       0.002 * cos(0.14177660173  + 10021.59346261590 *T)
    sumX0  +=       0.002 * cos(1.51917223670  + 20618.26317601710 *T)
    sumX0  +=       0.002 * cos(4.14445279451  + 31558.03074209830 *T)
    sumX0  +=       0.002 * cos(2.33949764052  + 2854.88419139370 *T)
    sumX0  +=       0.002 * cos(1.98831285254  + 51953.18977763829 *T)
    sumX0  +=       0.002 * cos(0.41604857841  + 51951.21766996299 *T)
    sumX0  +=       0.002 * cos(1.57183044674  + 52249.34722649089 *T)
    sumX0  +=       0.002 * cos(1.70818461045  + 138633.54722359608 *T)
    sumX0  +=       0.002 * cos(5.13575806013  + 225687.46509753368 *T)
    sumX0  +=       0.002 * cos(1.56258965853  + 138319.36104372647 *T)
    sumX0  +=       0.002 * cos(2.18966503711  + 32768.88417749029 *T)
    sumX0  +=       0.002 * cos(0.33098015159  + 24079.58941989870 *T)
    sumX0  +=       0.002 * cos(4.71436622522  + 52202.34842043170 *T)
    sumX0  +=       0.002 * cos(4.96992436419  + 28206.42318446909 *T)
    sumX0  +=       0.002 * cos(4.96216423435  + 180496.80194812448 *T)
    sumX0  +=       0.002 * cos(4.73660629226  + 78259.58427188109 *T)
    sumX0  +=       0.002 * cos(2.28401532959  + 26624.46383618609 *T)
    sumX0  +=       0.002 * cos(0.07960474111  + 23549.89845480410 *T)
    sumX0  +=       0.002 * cos(0.37327493336  + 26293.84487252789 *T)
    sumX0  +=       0.002 * cos(5.83913193896  + 4601.70630147949 *T)
    sumX0  +=       0.002 * cos(6.23015270329  + 21716.27059228350 *T)
    sumX0  +=       0.002 * cos(3.66936461651  + 330.86278114170 *T)
    sumX0  +=       0.002 * cos(1.51579465920  + 51219.27330969429 *T)
    sumX0  +=       0.002 * cos(5.05016618926  + 78256.35206023828 *T)
    sumX0  +=       0.002 * cos(3.58281342347  + 50063.91215228729 *T)
    sumX0  +=       0.002 * cos(1.42199745004  + 132658.02899457427 *T)
    sumX0  +=       0.002 * cos(3.28152370063  + 96.22304470130 *T)
    sumX0  +=       0.002 * cos(4.38858567715  + 1265.32366114289 *T)
    sumX0  +=       0.002 * cos(2.49109566786  + 52823.06093394670 *T)
    sumX0  +=       0.002 * cos(1.26800894276  + 78270.57915423988 *T)
    sumX0  +=       0.002 * cos(5.41071683840  + 7392.21058460129 *T)
    sumX0  +=       0.002 * cos(1.10270458425  + 61165.03017109729 *T)
    sumX0  +=       0.002 * cos(2.14950038663  + 2544.42881940890 *T)
    sumX0  +=       0.002 * cos(1.97162097641  + 26421.51526486010 *T)
    sumX0  +=       0.002 * cos(1.33257889355  + 2655.81218995730 *T)
    sumX0  +=       0.002 * cos(0.79025165001  + 24705.94902657310 *T)
    sumX0  +=       0.002 * cos(4.87337338120  + 27470.34489154230 *T)
    sumX0  +=       0.002 * cos(6.14740336072  + 1353.50492753330 *T)
    sumX0  +=       0.002 * cos(0.11906484226  + 1603.54380676890 *T)
    sumX0  +=       0.002 * cos(3.46701006002  + 40738.45907867528 *T)
    sumX0  +=       0.002 * cos(5.71504799272  + 30460.02332583190 *T)
    sumX0  +=       0.002 * cos(1.16042497962  + 4578.49563434210 *T)
    sumX0  +=       0.002 * cos(0.14527391572  + 78113.90241010448 *T)
    sumX0  +=       0.002 * cos(4.19897532608  + 32839.18779940350 *T)
    sumX0  +=       0.002 * cos(0.13374340654  + 14297.51787039890 *T)
    sumX0  +=       0.002 * cos(5.52988162467  + 37878.77604771650 *T)
    sumX0  +=       0.002 * cos(6.20542886765  + 104501.41958091488 *T)
    sumX0  +=       0.002 * cos(2.61097814223  + 25138.48371512089 *T)
    sumX0  +=       0.002 * cos(1.06395351693  + 44131.52503670349 *T)
    sumX0  +=       0.002 * cos(4.10426443523  + 104358.48229581409 *T)
    sumX0  +=       0.002 * cos(5.39518200455  + 23336.59935936610 *T)
    sumX0  +=       0.002 * cos(5.39518200455  + 23336.11172439909 *T)
    sumX0  +=       0.002 * cos(6.04499536111  + 224.10097821840 *T)
    sumX0  +=       0.002 * cos(3.63958668308  + 51547.19851458179 *T)
    sumX0  +=       0.002 * cos(3.16690823789  + 202.00957769060 *T)
    sumX0  +=       0.002 * cos(3.16690823789  + 202.49721265760 *T)
    sumX0  +=       0.002 * cos(6.25775046727  + 31108.06544647091 *T)
    sumX0  +=       0.002 * cos(5.68905987112  + 21068.22847164450 *T)
    sumX0  +=       0.002 * cos(2.96587029379  + 117076.82218358829 *T)
    sumX0  +=       0.002 * cos(6.10746294738  + 117077.30981855528 *T)
    sumX0  +=       0.002 * cos(5.38291113550  + 24814.97829846349 *T)
    sumX0  +=       0.002 * cos(1.98570847468  + 27299.48802941670 *T)
    sumX0  +=       0.002 * cos(3.67791655653  + 24876.80588869870 *T)
    sumX0  +=       0.002 * cos(2.83769263825  + 26086.17485138239 *T)
    sumX0  +=       0.002 * cos(5.74298703182  + 105418.35186097027 *T)
    sumX0  +=       0.002 * cos(2.69099617398  + 49534.70882215970 *T)
    sumX0  +=       0.002 * cos(4.25460776712  + 107679.99204006046 *T)
    sumX0  +=       0.002 * cos(1.43038533012  + 19337.10611871190 *T)
    sumX0  +=       0.002 * cos(1.80708255079  + 24957.99301147190 *T)
    sumX0  +=       0.002 * cos(3.85654248042  + 27218.30090664350 *T)
    sumX0  +=       0.002 * cos(3.46898572008  + 66826.36222024946 *T)
    sumX0  +=       0.002 * cos(3.11677200697  + 27257.73521046630 *T)
    sumX0  +=       0.002 * cos(1.54813681705  + 26197.86564531009 *T)
    sumX0  +=       0.002 * cos(2.26084282341  + 151199.69892357889 *T)
    sumX0  +=       0.002 * cos(2.41505887981  + 426.15259399750 *T)
    sumX0  +=       0.002 * cos(2.05501019547  + 76784.59903317469 *T)
    sumX0  +=       0.002 * cos(5.17891248404  + 179172.50322857045 *T)
    sumX0  +=       0.002 * cos(3.03839569379  + 168959.21768235945 *T)
    sumX0  +=       0.002 * cos(3.03839569378  + 168959.70531732647 *T)
    sumX0  +=       0.002 * cos(6.05144227865  + 7829.84871395129 *T)
    sumX0  +=       0.002 * cos(2.93845777192  + 151975.22153490287 *T)
    sumX0  +=       0.002 * cos(5.56610117919  + 20235.36664379391 *T)
    sumX0  +=       0.002 * cos(0.09752385202  + 31940.92727432150 *T)
    sumX0  +=       0.002 * cos(0.34264055281  + 1585.38455950421 *T)
    sumX0  +=       0.002 * cos(3.25664376155  + 76.02225379210 *T)
    sumX0  +=       0.002 * cos(4.50673995632  + 1581.71553079950 *T)
    sumX0  +=       0.002 * cos(4.26102899883  + 52643.52745601930 *T)
    sumX0  +=       0.002 * cos(0.89847733804  + 130433.13096645444 *T)
    sumX0  +=       0.002 * cos(5.35295047368  + 79956.11891170849 *T)
    sumX0  +=       0.002 * cos(1.38123247936  + 190809.35350674443 *T)
    sumX0  +=       0.002 * cos(5.22748899717  + 155417.79647986827 *T)
    sumX0  +=       0.002 * cos(2.25472465461  + 58219.79104704029 *T)
    sumX0  +=       0.002 * cos(5.67352931908  + 299399.22102519736 *T)
    sumX0  +=       0.002 * cos(0.82971507937  + 51738.40620949199 *T)
    sumX0  +=       0.002 * cos(2.97072309926  + 2045.55235028570 *T)
    sumX0  +=       0.002 * cos(4.40886857181  + 27953.37836157530 *T)
    sumX0  +=       0.002 * cos(1.25475645940  + 24222.91555654010 *T)
    sumX0  +=       0.002 * cos(4.61765528518  + 164721.45036517028 *T)
    sumX0  +=       0.002 * cos(0.02963994293  + 28441.01311282950 *T)
    sumX0  +=       0.002 * cos(5.63398508828  + 23735.28080528590 *T)
    sumX0  +=       0.002 * cos(4.94052786966  + 26948.45688781050 *T)
    sumX0  +=       0.002 * cos(0.72309716155  + 25227.83703030489 *T)
    sumX0  +=       0.002 * cos(2.26852923431  + 148532.64658993747 *T)
    sumX0  +=       0.002 * cos(2.26852923431  + 148533.13422490444 *T)
    sumX0  +=       0.002 * cos(1.51852948979  + 77307.17645126847 *T)
    sumX0  +=       0.002 * cos(6.10588157940  + 191.69208359510 *T)
    sumX0  +=       0.002 * cos(0.34186830904  + 23976.00901071309 *T)
    sumX0  +=       0.002 * cos(4.08150026692  + 206370.27255720508 *T)
    sumX0  +=       0.002 * cos(5.05627960945  + 4268.33799567710 *T)
    sumX0  +=       0.002 * cos(1.84151604889  + 23962.26955167850 *T)
    sumX0  +=       0.002 * cos(1.84151604889  + 23961.78191671149 *T)
    sumX0  +=       0.002 * cos(4.35044814692  + 3776.67470321830 *T)
    sumX0  +=       0.002 * cos(2.20524904392  + 24918.55870764909 *T)
    sumX0  +=       0.002 * cos(3.21408526505  + 102769.40940053028 *T)
    sumX0  +=       0.002 * cos(2.12858110666  + 25889.07495762130 *T)
    sumX0  +=       0.002 * cos(3.53504392455  + 26287.21896049409 *T)
    sumX0  +=       0.002 * cos(1.66367572246  + 2648.21100798950 *T)
    sumX0  +=       0.002 * cos(5.71672028658  + 80432.71372448868 *T)
    sumX0  +=       0.002 * cos(2.91629438354  + 13989.96024942930 *T)
    sumX0  +=       0.002 * cos(2.74733064767  + 38186.33366868610 *T)
    sumX0  +=       0.002 * cos(3.61656667123  + 8194.51915069210 *T)
    sumX0  +=       0.002 * cos(4.13822438475  + 237009.64156087104 *T)
    sumX0  +=       0.002 * cos(2.25489644524  + 24863.84148331209 *T)
    sumX0  +=       0.002 * cos(0.79435355035  + 76064.86589891909 *T)
    sumX0  +=       0.002 * cos(6.11153635623  + 71492.75500599349 *T)
    sumX0  +=       0.002 * cos(1.47457311729  + 30831.54872222810 *T)
    sumX0  +=       0.002 * cos(4.18905191391  + 21344.74519588730 *T)
    sumX0  +=       0.002 * cos(2.64103330733  + 124663.98818751708 *T)
    sumX0  +=       0.001 * cos(4.41626377495  + 101702.91392334028 *T)
    sumX0  +=       0.002 * cos(2.92553614920  + 105404.12476696869 *T)
    sumX0  +=       0.002 * cos(1.90962776233  + 307.31380348609 *T)
    sumX0  +=       0.002 * cos(1.20163121188  + 184834.32291268968 *T)
    sumX0  +=       0.002 * cos(4.34322386547  + 184833.83527772260 *T)
    sumX0  +=       0.001 * cos(5.20673509792  + 124156.19604039608 *T)
    sumX0  +=       0.001 * cos(3.29097711242  + 22909.51353352309 *T)
    sumX0  +=       0.001 * cos(3.29097711242  + 22910.00116849010 *T)
    sumX0  +=       0.001 * cos(4.22627101320  + 958.87904247409 *T)
    sumX0  +=       0.002 * cos(0.74624553136  + 77196.97013004909 *T)
    sumX0  +=       0.002 * cos(2.61684173604  + 9591.29792788829 *T)
    sumX0  +=       0.002 * cos(1.03105335724  + 189385.78877478145 *T)
    sumX0  +=       0.001 * cos(0.57809610749  + 24919.24812201110 *T)
    sumX0  +=       0.001 * cos(5.08552892372  + 27257.04579610430 *T)
    sumX0  +=       0.002 * cos(3.46440981619  + 25440.64849077589 *T)
    sumX0  +=       0.001 * cos(0.45556161499  + 26010.39641507380 *T)
    sumX0  +=       0.002 * cos(0.42518108376  + 43693.88690735350 *T)
    sumX0  +=       0.002 * cos(3.47315264300  + 25014.05029989989 *T)
    sumX0  +=       0.001 * cos(5.34981963981  + 9385.08482555870 *T)
    sumX0  +=       0.002 * cos(4.60292662632  + 26202.09861277591 *T)
    sumX0  +=       0.001 * cos(4.92214692877  + 52499.55551728929 *T)
    sumX0  +=       0.002 * cos(3.46155131960  + 14650.55593710108 *T)
    sumX0  +=       0.002 * cos(5.08759156576  + 78690.06379811508 *T)
    sumX0  +=       0.001 * cos(0.08764164508  + 5039.34443082949 *T)
    sumX0  +=       0.002 * cos(4.10611692906  + 2214.98690507971 *T)
    sumX0  +=       0.001 * cos(3.50305053765  + 77211.19722405069 *T)
    sumX0  +=       0.001 * cos(5.68985540916  + 233202.29919636747 *T)
    sumX0  +=       0.002 * cos(5.49174220323  + 21516.52370109911 *T)
    sumX0  +=       0.002 * cos(0.17188282798  + 30659.77021701629 *T)
    sumX0  +=       0.002 * cos(5.30439852095  + 621.49998335569 *T)
    sumX0  +=       0.002 * cos(1.13720629039  + 25764.15390743329 *T)
    sumX0  +=       0.001 * cos(5.32131294248  + 26.54213728330 *T)
    sumX0  +=       0.001 * cos(2.30545913825  + 226796.35601466001 *T)
    sumX0  +=       0.001 * cos(5.44705179184  + 226796.84364962703 *T)
    sumX0  +=       0.002 * cos(0.33446329498  + 28839.69455874929 *T)
    sumX0  +=       0.002 * cos(5.50347633955  + 102225.97897640108 *T)
    sumX0  +=       0.002 * cos(1.56059892812  + 129971.79453500008 *T)
    sumX0  +=       0.002 * cos(4.06465171282  + 27139.92770727909 *T)
    sumX0  +=       0.001 * cos(2.62468643294  + 50807.38984778689 *T)
    sumX0  +=       0.001 * cos(6.16306543539  + 78187.19953596349 *T)
    sumX0  +=       0.002 * cos(5.06371993416  + 51472.41691601449 *T)

    # MERCURY X Coord
    sumX1  = 0
    sumX1  +=  328639.517 * cos(6.04028758995  + 0.24381748350 *T)
    sumX1  +=  106107.047 * cos(5.91538469937  + 52176.05010063190 *T)
    sumX1  +=   32448.440 * cos(2.68404164136  + 78263.95324220609 *T)
    sumX1  +=    9699.418 * cos(5.42935843059  + 26087.65932409069 *T)
    sumX1  +=    8808.490 * cos(5.73515025839  + 104351.85638378028 *T)
    sumX1  +=    4833.058 * cos(0.21935428474  + 26088.14695905770 *T)
    sumX1  +=    2334.329 * cos(2.50289115433  + 130439.75952535449 *T)
    sumX1  +=     890.250 * cos(2.61758665086  + 52175.56246566489 *T)
    sumX1  +=     615.643 * cos(5.55374456387  + 156527.66266692869 *T)
    sumX1  +=     147.225 * cos(6.09730240676  + 78263.46560723908 *T)
    sumX1  +=     162.460 * cos(2.32137690587  + 182615.56580850287 *T)
    sumX1  +=      69.502 * cos(2.91939587909  + 24978.76840696430 *T)
    sumX1  +=      63.529 * cos(2.74219954754  + 27197.52551115110 *T)
    sumX1  +=      42.967 * cos(5.37217235288  + 208703.46895007708 *T)
    sumX1  +=      32.514 * cos(3.18436046111  + 104351.36874881329 *T)
    sumX1  +=      17.597 * cos(5.97213886087  + 51066.67154853850 *T)
    sumX1  +=      13.287 * cos(5.79146748296  + 53285.42865272530 *T)
    sumX1  +=      11.758 * cos(0.18693736842  + 1109.13473460990 *T)
    sumX1  +=      12.502 * cos(1.40596439354  + 1059.62574767270 *T)
    sumX1  +=      11.468 * cos(0.35008608561  + 20426.81490990550 *T)
    sumX1  +=      11.394 * cos(2.13977164486  + 234791.37209165128 *T)
    sumX1  +=       8.221 * cos(5.31229385932  + 31749.47900820990 *T)
    sumX1  +=       8.067 * cos(0.15951495713  + 130439.27189038748 *T)
    sumX1  +=       6.176 * cos(5.46098436501  + 4552.19731454230 *T)
    sumX1  +=       4.784 * cos(0.26415468148  + 21536.19346199890 *T)
    sumX1  +=       4.858 * cos(2.74077664990  + 77154.57469011268 *T)
    sumX1  +=       4.269 * cos(3.39637722107  + 46514.71805147970 *T)
    sumX1  +=       4.412 * cos(5.10134190424  + 25132.54721744910 *T)
    sumX1  +=       3.922 * cos(2.37978972370  + 1589.31671276730 *T)
    sumX1  +=       4.017 * cos(4.41672907758  + 27147.52888924690 *T)
    sumX1  +=       3.763 * cos(4.17909693196  + 14765.48286075331 *T)
    sumX1  +=       4.285 * cos(4.27106878562  + 51116.66817044270 *T)
    sumX1  +=       3.901 * cos(0.54893072973  + 27043.74670066630 *T)
    sumX1  +=       3.221 * cos(1.09802577229  + 26617.83792415230 *T)
    sumX1  +=       3.080 * cos(4.68247133566  + 25558.45599396310 *T)
    sumX1  +=       3.030 * cos(5.19056049817  + 260879.27523322543 *T)
    sumX1  +=       2.907 * cos(0.23044150312  + 26068.47719815791 *T)
    sumX1  +=       2.906 * cos(5.43307929185  + 26107.81671995749 *T)
    sumX1  +=       2.383 * cos(2.07830748691  + 26084.26562369970 *T)
    sumX1  +=       2.381 * cos(3.58525530417  + 26092.02829441570 *T)
    sumX1  +=       2.069 * cos(1.48127323869  + 37410.81105736210 *T)
    sumX1  +=       2.028 * cos(5.44501246598  + 30640.10045611650 *T)
    sumX1  +=       2.088 * cos(3.34536117513  + 156527.17503196167 *T)
    sumX1  +=       1.998 * cos(2.50305248201  + 79373.33179429949 *T)
    sumX1  +=       1.880 * cos(0.22580120983  + 47624.09660357310 *T)
    sumX1  +=       1.580 * cos(1.69158612681  + 9104.15081160110 *T)
    sumX1  +=       1.575 * cos(4.31668154619  + 955.35592412510 *T)
    sumX1  +=       1.508 * cos(0.43559873501  + 19317.43635781211 *T)
    sumX1  +=       1.397 * cos(0.16380111329  + 72602.62119305388 *T)
    sumX1  +=       1.363 * cos(3.28693470602  + 50586.97720534809 *T)
    sumX1  +=       1.495 * cos(3.77822350663  + 12566.39551746630 *T)
    sumX1  +=       1.401 * cos(2.04862276452  + 426.84200835950 *T)
    sumX1  +=       1.215 * cos(3.93785769383  + 52705.74106572650 *T)
    sumX1  +=       1.310 * cos(2.48605199154  + 26095.26050605850 *T)
    sumX1  +=       1.272 * cos(3.22244702438  + 26081.03341205689 *T)
    sumX1  +=       1.476 * cos(2.99369021823  + 23869.38985487090 *T)
    sumX1  +=       1.169 * cos(5.49111664316  + 3442.81876244891 *T)
    sumX1  +=       1.103 * cos(1.21903918539  + 53235.43203082109 *T)
    sumX1  +=       1.375 * cos(5.79233621902  + 103242.47783168688 *T)
    sumX1  +=       1.244 * cos(1.82501834169  + 51220.45035902330 *T)
    sumX1  +=       1.238 * cos(2.66650381891  + 28306.90406324449 *T)
    sumX1  +=       0.967 * cos(3.57175258487  + 53131.64984224050 *T)
    sumX1  +=       0.986 * cos(5.22314307595  + 32858.85756030330 *T)
    sumX1  +=       0.911 * cos(4.29268130103  + 529.44714761109 *T)
    sumX1  +=       0.859 * cos(0.94524893916  + 40853.38600232750 *T)
    sumX1  +=       0.855 * cos(5.33250267421  + 27677.21985434150 *T)
    sumX1  +=       0.844 * cos(1.03648635199  + 25874.84786361970 *T)
    sumX1  +=       1.042 * cos(0.19040742290  + 22645.57201409229 *T)
    sumX1  +=       0.924 * cos(0.74224500219  + 640.14110379750 *T)
    sumX1  +=       0.810 * cos(4.64191507416  + 26301.44605449570 *T)
    sumX1  +=       0.807 * cos(1.95815243056  + 286967.17837479961 *T)
    sumX1  +=       0.875 * cos(2.13143145763  + 83925.28529135829 *T)
    sumX1  +=       0.874 * cos(5.55044306137  + 529.93478257810 *T)
    sumX1  +=       0.821 * cos(4.77032702914  + 36301.43250526870 *T)
    sumX1  +=       0.823 * cos(5.19902587368  + 11322.42028082089 *T)
    sumX1  +=       0.738 * cos(2.19981448222  + 56728.00359769070 *T)
    sumX1  +=       0.724 * cos(1.50055745214  + 18849.47136745770 *T)
    sumX1  +=       0.714 * cos(2.67898291505  + 6770.46678376209 *T)
    sumX1  +=       0.667 * cos(3.97361924647  + 43072.14310651430 *T)
    sumX1  +=       0.665 * cos(3.28272656551  + 52156.38033973211 *T)
    sumX1  +=       0.893 * cos(0.42706797774  + 27197.03787618409 *T)
    sumX1  +=       0.752 * cos(1.88765944457  + 39609.89840064910 *T)
    sumX1  +=       0.673 * cos(4.81316066279  + 1052.51220067191 *T)
    sumX1  +=       0.688 * cos(0.86035048942  + 15874.86141284670 *T)
    sumX1  +=       0.618 * cos(2.12715472204  + 5661.57586663570 *T)
    sumX1  +=       0.808 * cos(4.33645956081  + 1066.73929467350 *T)
    sumX1  +=       0.616 * cos(0.70456047849  + 51646.35913553730 *T)
    sumX1  +=       0.650 * cos(1.39099602371  + 25028.76502886850 *T)
    sumX1  +=       0.550 * cos(0.20488240066  + 182615.07817353585 *T)
    sumX1  +=       0.554 * cos(3.01484203719  + 2119.00767786191 *T)
    sumX1  +=       0.535 * cos(0.35919946298  + 52179.93143598990 *T)
    sumX1  +=       0.657 * cos(5.49249173952  + 29530.72190402310 *T)
    sumX1  +=       0.527 * cos(4.93321108764  + 26514.74514993370 *T)
    sumX1  +=       0.502 * cos(0.52469020589  + 38654.29865904050 *T)
    sumX1  +=       0.568 * cos(0.32194599748  + 24978.28077199729 *T)
    sumX1  +=       0.498 * cos(4.51537827269  + 63498.71419893629 *T)
    sumX1  +=       0.469 * cos(5.15675930919  + 110013.18843293248 *T)
    sumX1  +=       0.521 * cos(2.19480576214  + 52195.71986153169 *T)
    sumX1  +=       0.440 * cos(3.21479581797  + 98690.52433462808 *T)
    sumX1  +=       0.580 * cos(6.26369012947  + 7.35736448430 *T)
    sumX1  +=       0.438 * cos(4.12030207268  + 33326.82255065770 *T)
    sumX1  +=       0.530 * cos(0.13282096675  + 2218.51328670329 *T)
    sumX1  +=       0.483 * cos(3.64743298930  + 51749.45190975589 *T)
    sumX1  +=       0.434 * cos(5.12850749811  + 52172.16876527390 *T)
    sumX1  +=       0.408 * cos(1.65432828623  + 7238.43177411650 *T)
    sumX1  +=       0.390 * cos(0.61831253528  + 78793.64420730069 *T)
    sumX1  +=       0.374 * cos(2.01855068333  + 58946.76070187749 *T)
    sumX1  +=       0.404 * cos(6.01175659056  + 52168.93655363109 *T)
    sumX1  +=       0.356 * cos(1.51807956509  + 62389.33564684289 *T)
    sumX1  +=       0.373 * cos(0.98588361546  + 103292.47445359109 *T)
    sumX1  +=       0.326 * cos(4.29903268783  + 79323.33517239528 *T)
    sumX1  +=       0.393 * cos(2.56051977629  + 129330.38097326108 *T)
    sumX1  +=       0.385 * cos(5.51554890357  + 52183.16364763270 *T)
    sumX1  +=       0.316 * cos(6.11147866851  + 76674.88034692229 *T)
    sumX1  +=       0.390 * cos(5.74481892138  + 54394.80720481869 *T)
    sumX1  +=       0.323 * cos(6.04898382500  + 49957.29299644510 *T)
    sumX1  +=       0.299 * cos(1.43100773503  + 16983.75232997309 *T)
    sumX1  +=       0.328 * cos(4.04864136289  + 68050.66769599509 *T)
    sumX1  +=       0.314 * cos(4.92007418019  + 51536.15281431789 *T)
    sumX1  +=       0.367 * cos(4.85256595117  + 77308.35350059748 *T)
    sumX1  +=       0.316 * cos(3.49203616890  + 45405.33949938630 *T)
    sumX1  +=       0.319 * cos(4.23988546277  + 13656.10430865991 *T)
    sumX1  +=       0.266 * cos(0.22278100987  + 48733.47515566649 *T)
    sumX1  +=       0.289 * cos(6.05400760825  + 6.86972951729 *T)
    sumX1  +=       0.261 * cos(4.00184287590  + 16984.23996494010 *T)
    sumX1  +=       0.261 * cos(1.49214547675  + 52389.34919606990 *T)
    sumX1  +=       0.233 * cos(0.84231286084  + 51123.78171744350 *T)
    sumX1  +=       0.231 * cos(5.24319932827  + 82815.90673926489 *T)
    sumX1  +=       0.239 * cos(2.80376783399  + 213.54291292150 *T)
    sumX1  +=       0.222 * cos(3.99420764167  + 66941.28914390170 *T)
    sumX1  +=       0.216 * cos(5.00937403559  + 313055.08151637384 *T)
    sumX1  +=       0.276 * cos(1.32498030782  + 51109.55462344189 *T)
    sumX1  +=       0.222 * cos(3.36848698260  + 32371.22280904910 *T)
    sumX1  +=       0.198 * cos(2.05297081511  + 53765.12299591570 *T)
    sumX1  +=       0.197 * cos(0.74093273246  + 69160.04624808849 *T)
    sumX1  +=       0.196 * cos(2.65111539002  + 50057.28624025350 *T)
    sumX1  +=       0.202 * cos(4.02603981938  + 129380.37759516528 *T)
    sumX1  +=       0.201 * cos(4.46760354980  + 44937.37450903190 *T)
    sumX1  +=       0.216 * cos(2.15128003627  + 213.05527795450 *T)
    sumX1  +=       0.185 * cos(1.91579542249  + 136101.09157450669 *T)
    sumX1  +=       0.183 * cos(2.26311522221  + 419.72846135871 *T)
    sumX1  +=       0.186 * cos(1.00434945598  + 57837.38214978410 *T)
    sumX1  +=       0.178 * cos(5.27169711768  + 105461.23493587368 *T)
    sumX1  +=       0.177 * cos(3.75548067168  + 51962.75100519390 *T)
    sumX1  +=       0.170 * cos(0.05181048718  + 78244.28348130629 *T)
    sumX1  +=       0.229 * cos(1.74351509136  + 433.95555536030 *T)
    sumX1  +=       0.173 * cos(0.79032586326  + 94138.57083756928 *T)
    sumX1  +=       0.157 * cos(3.59089906997  + 64742.20180061470 *T)
    sumX1  +=       0.153 * cos(0.87597586259  + 59414.72569223190 *T)
    sumX1  +=       0.159 * cos(6.16899315901  + 853.44019923550 *T)
    sumX1  +=       0.155 * cos(4.68975805923  + 35192.05395317530 *T)
    sumX1  +=       0.149 * cos(0.33674359016  + 24499.07406377390 *T)
    sumX1  +=       0.152 * cos(6.14951984841  + 4551.70967957530 *T)
    sumX1  +=       0.147 * cos(1.70740828389  + 52602.64829150790 *T)
    sumX1  +=       0.147 * cos(2.15567882351  + 19805.07110906630 *T)
    sumX1  +=       0.182 * cos(2.25669630329  + 55618.62504559730 *T)
    sumX1  +=       0.140 * cos(0.10429379711  + 23754.95056618569 *T)
    sumX1  +=       0.146 * cos(3.32146505695  + 208702.98131511008 *T)
    sumX1  +=       0.154 * cos(3.39064691795  + 99799.90288672148 *T)
    sumX1  +=       0.139 * cos(6.20098571989  + 25973.70767037247 *T)
    sumX1  +=       0.138 * cos(5.74451336242  + 26202.58624774291 *T)
    sumX1  +=       0.191 * cos(3.87484611458  + 26728.04424537170 *T)
    sumX1  +=       0.152 * cos(5.74228778845  + 1109.62236957690 *T)
    sumX1  +=       0.162 * cos(1.42094114156  + 38520.18960945550 *T)
    sumX1  +=       0.141 * cos(1.24257990116  + 77734.26227711148 *T)
    sumX1  +=       0.135 * cos(3.41542832234  + 78267.83457756409 *T)
    sumX1  +=       0.128 * cos(4.56986923193  + 88477.23878841709 *T)
    sumX1  +=       0.147 * cos(5.33327922722  + 31748.99137324289 *T)
    sumX1  +=       0.135 * cos(6.26697445915  + 124778.42747620228 *T)
    sumX1  +=       0.148 * cos(1.14618449835  + 27154.64243624770 *T)
    sumX1  +=       0.134 * cos(5.16840701859  + 12431.79883291429 *T)
    sumX1  +=       0.121 * cos(3.61695346173  + 104881.54734887488 *T)
    sumX1  +=       0.116 * cos(4.76247090725  + 25934.36814857290 *T)
    sumX1  +=       0.116 * cos(2.64947988255  + 32858.36992533629 *T)
    sumX1  +=       0.115 * cos(0.89972324521  + 26241.92576954250 *T)
    sumX1  +=       0.148 * cos(1.02099168260  + 25661.54876818170 *T)
    sumX1  +=       0.110 * cos(5.99966246087  + 28206.91081943610 *T)
    sumX1  +=       0.108 * cos(1.68778367228  + 27140.41534224609 *T)
    sumX1  +=       0.115 * cos(4.01860742052  + 103822.16541868568 *T)
    sumX1  +=       0.109 * cos(3.92589684326  + 53284.94101775829 *T)
    sumX1  +=       0.107 * cos(5.56945579461  + 28421.34335192970 *T)
    sumX1  +=       0.143 * cos(4.93475595866  + 1596.43025976811 *T)
    sumX1  +=       0.112 * cos(5.60947833730  + 155418.28411483529 *T)
    sumX1  +=       0.113 * cos(2.46236993488  + 5327.71992586630 *T)
    sumX1  +=       0.101 * cos(3.51555637450  + 25620.18196870330 *T)
    sumX1  +=       0.122 * cos(3.03742531641  + 10213.52936369450 *T)
    sumX1  +=       0.102 * cos(5.56482383222  + 2333.44021035551 *T)
    sumX1  +=       0.097 * cos(2.14556254447  + 26556.11194941211 *T)
    sumX1  +=       0.118 * cos(1.76061169391  + 7994.77225950771 *T)
    sumX1  +=       0.109 * cos(1.60651799859  + 103396.25664217168 *T)
    sumX1  +=       0.109 * cos(6.11285140847  + 79219.55298381469 *T)
    sumX1  +=       0.095 * cos(0.10552520847  + 58459.12595062330 *T)
    sumX1  +=       0.095 * cos(2.74369425701  + 78256.83969520529 *T)
    sumX1  +=       0.097 * cos(1.08624310756  + 105411.23831396949 *T)
    sumX1  +=       0.092 * cos(2.26275916512  + 10213.04172872750 *T)
    sumX1  +=       0.094 * cos(4.41890699093  + 1582.20316576650 *T)
    sumX1  +=       0.103 * cos(4.94789948755  + 91785.70468379748 *T)
    sumX1  +=       0.101 * cos(3.48509343222  + 20426.32727493849 *T)
    sumX1  +=       0.089 * cos(3.24400432046  + 74821.37829724069 *T)
    sumX1  +=       0.089 * cos(0.10163739667  + 125887.80602829569 *T)
    sumX1  +=       0.103 * cos(3.50686067995  + 51066.18391357149 *T)
    sumX1  +=       0.087 * cos(4.36360162781  + 27043.25906569930 *T)
    sumX1  +=       0.083 * cos(2.62141457558  + 73711.99974514729 *T)
    sumX1  +=       0.083 * cos(1.77797295395  + 77624.05595589208 *T)
    sumX1  +=       0.097 * cos(5.36704254243  + 71980.87739221469 *T)
    sumX1  +=       0.081 * cos(0.79181721859  + 155468.28073673949 *T)
    sumX1  +=       0.080 * cos(4.00356912047  + 22645.08437912529 *T)
    sumX1  +=       0.081 * cos(2.81631624727  + 76045.19613801930 *T)
    sumX1  +=       0.081 * cos(5.04291780878  + 26138.14358096190 *T)
    sumX1  +=       0.080 * cos(0.61687993534  + 26038.15033715350 *T)
    sumX1  +=       0.104 * cos(2.28163680118  + 78271.06678920689 *T)
    sumX1  +=       0.103 * cos(5.23374561959  + 78283.62300310588 *T)
    sumX1  +=       0.083 * cos(2.50636376987  + 80482.71034639288 *T)
    sumX1  +=       0.089 * cos(1.89337139570  + 78260.07190684808 *T)
    sumX1  +=       0.073 * cos(4.21987024947  + 26617.35028918529 *T)
    sumX1  +=       0.077 * cos(0.51899049759  + 18208.05780571871 *T)
    sumX1  +=       0.082 * cos(3.91217027287  + 77211.68485901768 *T)
    sumX1  +=       0.077 * cos(2.88917318730  + 41962.76455442090 *T)
    sumX1  +=       0.082 * cos(4.38267763386  + 77197.45776501608 *T)
    sumX1  +=       0.067 * cos(5.02424385988  + 85034.66384345168 *T)
    sumX1  +=       0.070 * cos(1.09572627877  + 537.04832957890 *T)
    sumX1  +=       0.067 * cos(0.33161014347  + 5661.08823166870 *T)
    sumX1  +=       0.067 * cos(3.82822220045  + 120226.47397914348 *T)
    sumX1  +=       0.065 * cos(4.95852563263  + 162188.99471608087 *T)
    sumX1  +=       0.075 * cos(4.93326519789  + 149.80701461810 *T)
    sumX1  +=       0.071 * cos(2.00854094820  + 108903.80988083908 *T)
    sumX1  +=       0.063 * cos(3.29036116867  + 51756.56545675670 *T)
    sumX1  +=       0.062 * cos(2.43463735809  + 102762.78348849648 *T)
    sumX1  +=       0.067 * cos(4.50074668009  + 64608.09275102969 *T)
    sumX1  +=       0.064 * cos(5.16392570404  + 13521.99525907490 *T)
    sumX1  +=       0.062 * cos(0.12968651638  + 28306.41642827749 *T)
    sumX1  +=       0.058 * cos(1.62746286468  + 11322.90791578790 *T)
    sumX1  +=       0.058 * cos(0.27966370230  + 24498.58642880689 *T)
    sumX1  +=       0.077 * cos(3.79522760165  + 25565.56954096390 *T)
    sumX1  +=       0.076 * cos(4.60149373597  + 78477.25233764409 *T)
    sumX1  +=       0.063 * cos(0.77089662682  + 93029.19228547587 *T)
    sumX1  +=       0.078 * cos(0.26117589680  + 71493.24264096050 *T)
    sumX1  +=       0.077 * cos(3.91821281814  + 51742.33836275509 *T)
    sumX1  +=       0.057 * cos(3.59228773257  + 2648.69864295650 *T)
    sumX1  +=       0.058 * cos(1.77431779111  + 339142.98465794802 *T)
    sumX1  +=       0.056 * cos(5.53301773648  + 76145.18938182769 *T)
    sumX1  +=       0.059 * cos(5.94806683967  + 21535.70582703189 *T)
    sumX1  +=       0.056 * cos(1.70277932228  + 117873.60782537168 *T)
    sumX1  +=       0.053 * cos(4.64718015139  + 25557.96835899609 *T)
    sumX1  +=       0.060 * cos(1.21018656116  + 71025.27765060609 *T)
    sumX1  +=       0.052 * cos(0.56762366419  + 10022.08109758290 *T)
    sumX1  +=       0.055 * cos(5.77897482747  + 51322.85371887989 *T)
    sumX1  +=       0.058 * cos(1.40115688303  + 43071.65547154729 *T)
    sumX1  +=       0.053 * cos(5.07195556849  + 79853.02613748988 *T)
    sumX1  +=       0.053 * cos(0.21642667454  + 78050.65414676809 *T)
    sumX1  +=       0.053 * cos(2.16306876234  + 98068.78053378889 *T)
    sumX1  +=       0.057 * cos(1.24363255483  + 633.02755679670 *T)
    sumX1  +=       0.049 * cos(0.37407469111  + 90830.10494218887 *T)
    sumX1  +=       0.067 * cos(1.84374645383  + 26610.72437715149 *T)
    sumX1  +=       0.050 * cos(6.00585950729  + 131549.13807744789 *T)
    sumX1  +=       0.049 * cos(5.14290001765  + 33968.23611239669 *T)
    sumX1  +=       0.050 * cos(5.30537322308  + 81706.52818717150 *T)
    sumX1  +=       0.051 * cos(0.73226920314  + 129910.06856025988 *T)
    sumX1  +=       0.057 * cos(3.56358445708  + 26624.95147115310 *T)
    sumX1  +=       0.045 * cos(3.36854845214  + 51116.18053547569 *T)
    sumX1  +=       0.045 * cos(3.10088248442  + 104332.18662288050 *T)
    sumX1  +=       0.051 * cos(0.85422591600  + 647.25465079831 *T)
    sumX1  +=       0.043 * cos(4.78281150158  + 78690.55143308209 *T)
    sumX1  +=       0.054 * cos(2.25324121444  + 522.82123557730 *T)
    sumX1  +=       0.042 * cos(1.33828179439  + 114565.14192999128 *T)
    sumX1  +=       0.041 * cos(5.13406465270  + 38519.70197448849 *T)
    sumX1  +=       0.041 * cos(3.95146540461  + 25028.27739390149 *T)
    sumX1  +=       0.052 * cos(1.00660903298  + 39744.00745023410 *T)
    sumX1  +=       0.043 * cos(0.49413326170  + 103925.25819290428 *T)
    sumX1  +=       0.045 * cos(4.58276908488  + 76.50988875911 *T)
    sumX1  +=       0.042 * cos(3.21789877684  + 46848.57399224910 *T)
    sumX1  +=       0.047 * cos(3.90827664098  + 44181.52165860769 *T)
    sumX1  +=       0.038 * cos(3.08592038431  + 22760.01130277751 *T)
    sumX1  +=       0.037 * cos(3.12927309026  + 151975.70916986989 *T)
    sumX1  +=       0.050 * cos(2.23521826581  + 25551.34244696229 *T)
    sumX1  +=       0.050 * cos(0.72721768819  + 50579.86365834729 *T)
    sumX1  +=       0.045 * cos(2.61026174635  + 7879.84533585549 *T)
    sumX1  +=       0.043 * cos(4.81034248928  + 26521.85869693450 *T)
    sumX1  +=       0.037 * cos(2.72331919328  + 30639.61282114949 *T)
    sumX1  +=       0.037 * cos(3.88891756517  + 48732.98752069949 *T)
    sumX1  +=       0.035 * cos(3.18195761886  + 84547.02909219750 *T)
    sumX1  +=       0.040 * cos(0.70421187083  + 52815.94738694590 *T)
    sumX1  +=       0.041 * cos(4.14956775757  + 1059.13811270570 *T)
    sumX1  +=       0.041 * cos(3.03500918971  + 150866.33061777649 *T)
    sumX1  +=       0.040 * cos(5.01484347149  + 26294.33250749489 *T)
    sumX1  +=       0.041 * cos(0.58688251264  + 25881.96141062050 *T)
    sumX1  +=       0.034 * cos(0.93821904321  + 26191.23973327629 *T)
    sumX1  +=       0.034 * cos(4.73103289822  + 25985.05418483910 *T)
    sumX1  +=       0.034 * cos(0.83186142058  + 89586.61734051049 *T)
    sumX1  +=       0.041 * cos(4.17323142245  + 52329.82891111670 *T)
    sumX1  +=       0.035 * cos(2.52595229567  + 157637.04121902207 *T)
    sumX1  +=       0.042 * cos(6.02261134030  + 26507.63160293289 *T)
    sumX1  +=       0.039 * cos(0.14372316696  + 234790.88445668429 *T)
    sumX1  +=       0.037 * cos(1.88707685753  + 57836.89451481709 *T)
    sumX1  +=       0.036 * cos(0.34370159925  + 130969.45049044908 *T)
    sumX1  +=       0.032 * cos(2.37775558754  + 181506.18725640947 *T)
    sumX1  +=       0.033 * cos(1.24575523402  + 50594.09075234889 *T)
    sumX1  +=       0.044 * cos(6.11332634879  + 6681.46867088311 *T)
    sumX1  +=       0.036 * cos(0.18721130545  + 104355.73771913828 *T)
    sumX1  +=       0.032 * cos(4.64549491762  + 129484.15978374588 *T)
    sumX1  +=       0.030 * cos(5.20724500900  + 52644.01509098630 *T)
    sumX1  +=       0.031 * cos(2.56009147472  + 29416.28261533789 *T)
    sumX1  +=       0.031 * cos(1.30643526573  + 105307.45612538888 *T)
    sumX1  +=       0.030 * cos(3.71640911346  + 95247.94938966268 *T)
    sumX1  +=       0.031 * cos(2.51344352354  + 52290.48938931711 *T)
    sumX1  +=       0.029 * cos(3.83470620242  + 181556.18387831369 *T)
    sumX1  +=       0.039 * cos(1.79502663396  + 25448.24967274370 *T)
    sumX1  +=       0.032 * cos(4.74231453232  + 51639.24558853649 *T)
    sumX1  +=       0.029 * cos(4.21913805837  + 53242.54557782190 *T)
    sumX1  +=       0.029 * cos(5.39418594332  + 25668.66231518250 *T)
    sumX1  +=       0.028 * cos(6.18231601923  + 77204.57131201689 *T)
    sumX1  +=       0.036 * cos(2.35417973488  + 25132.05958248210 *T)
    sumX1  +=       0.027 * cos(2.33120714494  + 54509.24649350390 *T)
    sumX1  +=       0.026 * cos(2.96005545403  + 52061.61081194667 *T)
    sumX1  +=       0.025 * cos(0.44641100649  + 29428.75938575750 *T)
    sumX1  +=       0.031 * cos(1.36909644374  + 18093.13088206649 *T)
    sumX1  +=       0.030 * cos(2.99318128420  + 26941.34334080970 *T)
    sumX1  +=       0.025 * cos(0.01970755306  + 956.53297345411 *T)
    sumX1  +=       0.027 * cos(3.92481990123  + 18093.61851703350 *T)
    sumX1  +=       0.028 * cos(5.33056586298  + 104358.96993078108 *T)
    sumX1  +=       0.026 * cos(1.68523834841  + 26237.71015619230 *T)
    sumX1  +=       0.024 * cos(5.34360042933  + 22747.53453235790 *T)
    sumX1  +=       0.024 * cos(0.60759776379  + 46514.23041651269 *T)
    sumX1  +=       0.029 * cos(4.14852842199  + 131499.14145554369 *T)
    sumX1  +=       0.024 * cos(5.11295227085  + 42154.21282053250 *T)
    sumX1  +=       0.024 * cos(2.72205818316  + 54294.81396101029 *T)
    sumX1  +=       0.023 * cos(0.59370019139  + 146314.37712071769 *T)
    sumX1  +=       0.024 * cos(3.26900809848  + 26107.32908499049 *T)
    sumX1  +=       0.026 * cos(4.45438431206  + 25021.65148186770 *T)
    sumX1  +=       0.023 * cos(0.68577711601  + 20760.18321570790 *T)
    sumX1  +=       0.023 * cos(0.40104725527  + 77154.08705514569 *T)
    sumX1  +=       0.024 * cos(4.07999331304  + 26067.98956319090 *T)
    sumX1  +=       0.023 * cos(3.75077334681  + 25035.87857586930 *T)
    sumX1  +=       0.023 * cos(4.73851847693  + 143961.51096694588 *T)
    sumX1  +=       0.026 * cos(1.72805748634  + 27684.33340134230 *T)
    sumX1  +=       0.022 * cos(3.46952873618  + 19804.58347409930 *T)
    sumX1  +=       0.025 * cos(6.21468652611  + 26080.54577708989 *T)
    sumX1  +=       0.023 * cos(3.49763456236  + 130013.16133447849 *T)
    sumX1  +=       0.022 * cos(0.73153517786  + 79372.84415933248 *T)
    sumX1  +=       0.025 * cos(0.72123980539  + 52026.48690349730 *T)
    sumX1  +=       0.021 * cos(5.79201875326  + 104344.74283677948 *T)
    sumX1  +=       0.021 * cos(5.87156389227  + 102133.09927959349 *T)
    sumX1  +=       0.023 * cos(1.39284020284  + 61279.95709474950 *T)
    sumX1  +=       0.021 * cos(2.36405581509  + 85502.62883380610 *T)
    sumX1  +=       0.023 * cos(0.10713155060  + 77844.46859833089 *T)
    sumX1  +=       0.022 * cos(1.72894324326  + 188276.89785765507 *T)
    sumX1  +=       0.021 * cos(5.21984877761  + 124156.68367536308 *T)
    sumX1  +=       0.027 * cos(0.68325003204  + 103299.58800059187 *T)
    sumX1  +=       0.020 * cos(5.77389955341  + 23968.89546371229 *T)
    sumX1  +=       0.023 * cos(0.69010991788  + 77830.24150432929 *T)
    sumX1  +=       0.021 * cos(1.14745744394  + 103285.36090659029 *T)
    sumX1  +=       0.020 * cos(2.07158691051  + 49527.59527515889 *T)
    sumX1  +=       0.020 * cos(2.29184774399  + 1479.11039154791 *T)
    sumX1  +=       0.020 * cos(5.92355125832  + 23969.38309867930 *T)
    sumX1  +=       0.020 * cos(4.95913484029  + 103711.95909746629 *T)
    sumX1  +=       0.019 * cos(3.65190813330  + 52126.05347872769 *T)
    sumX1  +=       0.019 * cos(0.04924351217  + 3327.89183879669 *T)
    sumX1  +=       0.022 * cos(1.42243873408  + 104565.15547921829 *T)
    sumX1  +=       0.026 * cos(1.25974103229  + 13362.69352428270 *T)
    sumX1  +=       0.025 * cos(0.01235262119  + 24864.32911827909 *T)
    sumX1  +=       0.023 * cos(4.22016173229  + 15874.37377787969 *T)
    sumX1  +=       0.022 * cos(4.72905553277  + 6283.31966747490 *T)
    sumX1  +=       0.025 * cos(5.64927135524  + 27311.96479983631 *T)
    sumX1  +=       0.019 * cos(4.78388065302  + 53228.31848382029 *T)
    sumX1  +=       0.021 * cos(3.31277231518  + 97581.14578253467 *T)
    sumX1  +=       0.018 * cos(5.56704384876  + 106570.61348796709 *T)
    sumX1  +=       0.021 * cos(1.92247469083  + 60056.13925397089 *T)
    sumX1  +=       0.019 * cos(6.17553462363  + 26083.77798873270 *T)
    sumX1  +=       0.021 * cos(1.98793084523  + 104371.52614468009 *T)
    sumX1  +=       0.017 * cos(2.47269230880  + 1162.71852189130 *T)
    sumX1  +=       0.020 * cos(4.30765634281  + 41962.27691945389 *T)
    sumX1  +=       0.020 * cos(1.17324939187  + 26091.54065944870 *T)
    sumX1  +=       0.017 * cos(2.58918808487  + 77410.75686045409 *T)
    sumX1  +=       0.018 * cos(1.56564486854  + 115674.52048208468 *T)
    sumX1  +=       0.021 * cos(5.05205604193  + 134991.71302241329 *T)
    sumX1  +=       0.017 * cos(3.10172524257  + 104138.55728834229 *T)
    sumX1  +=       0.017 * cos(5.66948357862  + 37410.32342239509 *T)
    sumX1  +=       0.019 * cos(5.53464997520  + 522.33360061030 *T)
    sumX1  +=       0.019 * cos(4.91833775610  + 52698.62751872569 *T)
    sumX1  +=       0.019 * cos(3.92340038849  + 131395.35926696309 *T)
    sumX1  +=       0.016 * cos(4.33041124879  + 12546.72575656651 *T)
    sumX1  +=       0.016 * cos(6.22611768107  + 100909.28143881488 *T)
    sumX1  +=       0.016 * cos(0.82606858430  + 25867.73431661889 *T)
    sumX1  +=       0.020 * cos(5.89802475411  + 26094.77287109149 *T)
    sumX1  +=       0.018 * cos(5.08722928383  + 47623.60896860609 *T)
    sumX1  +=       0.018 * cos(1.45687313804  + 12432.28646788130 *T)
    sumX1  +=       0.020 * cos(6.13261748755  + 11610.79577585770 *T)
    sumX1  +=       0.015 * cos(5.51634191387  + 183724.94436059627 *T)
    sumX1  +=       0.017 * cos(4.79626245128  + 51529.03926731709 *T)
    sumX1  +=       0.020 * cos(1.77003688612  + 426.35437339250 *T)
    sumX1  +=       0.016 * cos(4.90993271420  + 26308.55960149650 *T)
    sumX1  +=       0.016 * cos(4.79764512586  + 128850.68663007068 *T)
    sumX1  +=       0.019 * cos(4.56495311793  + 51543.26636131870 *T)
    sumX1  +=       0.019 * cos(3.74798808655  + 155997.97170183409 *T)
    sumX1  +=       0.015 * cos(1.34224435786  + 26164.41303033330 *T)
    sumX1  +=       0.018 * cos(3.66440627266  + 25874.36022865269 *T)
    sumX1  +=       0.014 * cos(2.04202638738  + 36300.94487030169 *T)
    sumX1  +=       0.019 * cos(5.88442577075  + 45494.82524723229 *T)
    sumX1  +=       0.014 * cos(3.57916105055  + 44295.96094729290 *T)
    sumX1  +=       0.018 * cos(3.82094554949  + 119117.09542705010 *T)
    sumX1  +=       0.016 * cos(3.76272474790  + 76667.76679992149 *T)
    sumX1  +=       0.014 * cos(2.07119234799  + 107794.43132874568 *T)
    sumX1  +=       0.014 * cos(2.36814985653  + 16065.82204399129 *T)
    sumX1  +=       0.014 * cos(1.33222225896  + 44181.03402364069 *T)
    sumX1  +=       0.014 * cos(4.76386004796  + 2111.89413086110 *T)
    sumX1  +=       0.015 * cos(1.74258697688  + 52226.04672253610 *T)
    sumX1  +=       0.014 * cos(4.76443919805  + 27999.34644227490 *T)
    sumX1  +=       0.019 * cos(4.94519146654  + 104347.97504842227 *T)
    sumX1  +=       0.014 * cos(3.43033111839  + 116918.00808376308 *T)
    sumX1  +=       0.015 * cos(4.82510142643  + 365230.88779952226 *T)
    sumX1  +=       0.016 * cos(0.25720185897  + 52712.85461272729 *T)
    sumX1  +=       0.017 * cos(4.24342377575  + 97113.18079218028 *T)
    sumX1  +=       0.013 * cos(0.92750711335  + 24176.94747584050 *T)
    sumX1  +=       0.017 * cos(3.10108098380  + 20760.67085067490 *T)
    sumX1  +=       0.014 * cos(0.49576253379  + 25954.03790947270 *T)
    sumX1  +=       0.014 * cos(5.16786249742  + 26222.25600864270 *T)
    sumX1  +=       0.017 * cos(0.70506585057  + 70269.42480018188 *T)
    sumX1  +=       0.013 * cos(6.17002721316  + 178063.61231144407 *T)
    sumX1  +=       0.015 * cos(1.07987898155  + 52099.78402935630 *T)
    sumX1  +=       0.013 * cos(3.68965474753  + 323.74923414091 *T)
    sumX1  +=       0.013 * cos(1.62156829158  + 39609.41076568209 *T)
    sumX1  +=       0.013 * cos(0.38134163255  + 206.42936592071 *T)
    sumX1  +=       0.014 * cos(1.82134412060  + 105940.92927906408 *T)
    sumX1  +=       0.013 * cos(2.20656460343  + 102233.09252340188 *T)
    sumX1  +=       0.014 * cos(0.61259865461  + 77204.08367704989 *T)
    sumX1  +=       0.014 * cos(0.58031216793  + 13521.50762410789 *T)
    sumX1  +=       0.013 * cos(4.69319019775  + 111590.53197538028 *T)
    sumX1  +=       0.012 * cos(4.27702379029  + 76681.99389392308 *T)
    sumX1  +=       0.015 * cos(1.28652822525  + 27670.10630734069 *T)
    sumX1  +=       0.013 * cos(2.50362342037  + 58946.27306691049 *T)
    sumX1  +=       0.015 * cos(2.00793708614  + 26300.95841952869 *T)
    sumX1  +=       0.013 * cos(3.31492062191  + 14477.10736571650 *T)
    sumX1  +=       0.013 * cos(1.07027376830  + 1692.40948698591 *T)
    sumX1  +=       0.013 * cos(4.39268185141  + 140653.04507156549 *T)
    sumX1  +=       0.012 * cos(1.34267704514  + 52072.95732641330 *T)
    sumX1  +=       0.012 * cos(6.26475550018  + 54394.31956985169 *T)
    sumX1  +=       0.012 * cos(1.17105305254  + 7238.91940908350 *T)
    sumX1  +=       0.012 * cos(5.40922931059  + 25447.76203777669 *T)
    sumX1  +=       0.012 * cos(6.20873329955  + 72936.47713382330 *T)
    sumX1  +=       0.015 * cos(5.49315509340  + 2126.12122486270 *T)
    sumX1  +=       0.012 * cos(5.66744055627  + 55504.18575691209 *T)
    sumX1  +=       0.012 * cos(3.61008361501  + 220.65645992230 *T)
    sumX1  +=       0.012 * cos(1.76003505598  + 111122.56698502589 *T)
    sumX1  +=       0.012 * cos(1.22583703139  + 90695.99589260388 *T)
    sumX1  +=       0.012 * cos(1.56652249391  + 104778.45457465628 *T)
    sumX1  +=       0.012 * cos(5.12697804116  + 64607.60511606269 *T)
    sumX1  +=       0.012 * cos(6.24535629479  + 110634.93223377167 *T)
    sumX1  +=       0.011 * cos(3.57080417047  + 36109.98423915710 *T)
    sumX1  +=       0.013 * cos(0.24854762034  + 49842.85370775989 *T)
    sumX1  +=       0.014 * cos(1.03533656892  + 78417.73205269089 *T)
    sumX1  +=       0.011 * cos(2.67619239702  + 26198.35328027710 *T)
    sumX1  +=       0.011 * cos(2.98743263419  + 25977.94063783829 *T)
    sumX1  +=       0.012 * cos(5.20745322961  + 31415.62306744050 *T)
    sumX1  +=       0.012 * cos(5.36057333309  + 6770.95441872910 *T)
    sumX1  +=       0.011 * cos(4.12038634021  + 51969.86455219470 *T)
    sumX1  +=       0.010 * cos(0.53057090724  + 23868.90221990390 *T)
    sumX1  +=       0.012 * cos(6.15491126193  + 130420.08976445468 *T)
    sumX1  +=       0.014 * cos(2.66131661996  + 32769.37181245730 *T)
    sumX1  +=       0.013 * cos(2.57684550665  + 33967.74847742969 *T)
    sumX1  +=       0.011 * cos(0.63840837632  + 51645.87150057029 *T)
    sumX1  +=       0.010 * cos(0.30214091711  + 28736.60178453070 *T)
    sumX1  +=       0.010 * cos(1.03579590518  + 52705.25343075949 *T)
    sumX1  +=       0.014 * cos(5.20517620734  + 77837.35505133009 *T)
    sumX1  +=       0.010 * cos(4.48203927159  + 141762.42362365889 *T)
    sumX1  +=       0.012 * cos(0.89324430712  + 224.58861318540 *T)
    sumX1  +=       0.010 * cos(1.51553715870  + 170049.41410852008 *T)
    sumX1  +=       0.013 * cos(4.39783434076  + 38813.60039383270 *T)
    sumX1  +=       0.010 * cos(1.45494488062  + 27147.04125427989 *T)
    sumX1  +=       0.010 * cos(3.80634126779  + 78903.85052852010 *T)
    sumX1  +=       0.013 * cos(4.94493446189  + 6282.83203250789 *T)
    sumX1  +=       0.010 * cos(0.60091290149  + 207644.08701988787 *T)
    sumX1  +=       0.012 * cos(0.27544235334  + 26395.70458002730 *T)
    sumX1  +=       0.012 * cos(5.38818267787  + 25780.58933808811 *T)
    sumX1  +=       0.010 * cos(0.84101636752  + 12565.90788249930 *T)
    sumX1  +=       0.012 * cos(4.18148179610  + 26725.14323108190 *T)
    sumX1  +=       0.012 * cos(1.48214323511  + 25451.15068703350 *T)
    sumX1  +=       0.009 * cos(5.62764518642  + 51219.76094466129 *T)
    sumX1  +=       0.009 * cos(1.40970085491  + 155572.06292532009 *T)
    sumX1  +=       0.009 * cos(1.98340877635  + 78731.91823256049 *T)
    sumX1  +=       0.009 * cos(1.03302022359  + 14764.99522578630 *T)
    sumX1  +=       0.011 * cos(3.25069263252  + 65697.80154222329 *T)
    sumX1  +=       0.012 * cos(2.28213947586  + 25446.73339731871 *T)
    sumX1  +=       0.012 * cos(3.38148555535  + 26729.56052079669 *T)
    sumX1  +=       0.012 * cos(1.48681981161  + 12725.20961729149 *T)
    sumX1  +=       0.010 * cos(4.04518614257  + 65831.91059180829 *T)
    sumX1  +=       0.009 * cos(1.98008802000  + 16703.30595098250 *T)
    sumX1  +=       0.009 * cos(0.19460006823  + 51708.08511027749 *T)
    sumX1  +=       0.009 * cos(5.42812340959  + 207594.09039798367 *T)
    sumX1  +=       0.009 * cos(0.25321773369  + 156101.06447605268 *T)
    sumX1  +=       0.012 * cos(6.08676982525  + 176954.23375935067 *T)
    sumX1  +=       0.009 * cos(1.85021451264  + 68242.11596210669 *T)
    sumX1  +=       0.009 * cos(2.19168596252  + 1910.95566573370 *T)
    sumX1  +=       0.009 * cos(1.01460352352  + 50586.48957038109 *T)
    sumX1  +=       0.009 * cos(1.60908201613  + 52609.76183850870 *T)
    sumX1  +=       0.008 * cos(3.02342551822  + 29530.23426905610 *T)
    sumX1  +=       0.009 * cos(2.83792438251  + 205.94173095370 *T)
    sumX1  +=       0.009 * cos(3.48790336223  + 55516.66252733170 *T)
    sumX1  +=       0.011 * cos(2.35570208491  + 52022.27129014710 *T)
    sumX1  +=       0.011 * cos(3.36653560664  + 157057.35363202327 *T)
    sumX1  +=       0.008 * cos(1.31427690069  + 39629.56816154889 *T)
    sumX1  +=       0.010 * cos(1.95365576783  + 52382.23564906909 *T)
    sumX1  +=       0.008 * cos(5.91790300447  + 52182.67865953191 *T)
    sumX1  +=       0.010 * cos(5.89081572614  + 40565.49814225770 *T)
    sumX1  +=       0.008 * cos(4.44673072353  + 103.33659170210 *T)
    sumX1  +=       0.008 * cos(0.60254149340  + 157483.26240853727 *T)
    sumX1  +=       0.008 * cos(0.07994652584  + 27177.85575025130 *T)
    sumX1  +=       0.008 * cos(5.58367850537  + 24998.43816786410 *T)
    sumX1  +=       0.009 * cos(3.23707766893  + 19.42594341629 *T)
    sumX1  +=       0.010 * cos(0.69943263197  + 536.56069461190 *T)
    sumX1  +=       0.010 * cos(3.22927666969  + 260878.78759825832 *T)
    sumX1  +=       0.008 * cos(6.13315569962  + 48847.91444435170 *T)
    sumX1  +=       0.008 * cos(6.23628865773  + 860.55374623631 *T)
    sumX1  +=       0.008 * cos(4.33455466105  + 26720.93069837089 *T)
    sumX1  +=       0.009 * cos(0.87327236491  + 25654.43522118089 *T)
    sumX1  +=       0.009 * cos(4.71644893845  + 63498.22656396929 *T)
    sumX1  +=       0.008 * cos(2.95337805438  + 52595.53474450709 *T)
    sumX1  +=       0.008 * cos(2.09085109942  + 130446.87307235529 *T)
    sumX1  +=       0.007 * cos(3.35858694831  + 50697.18352656750 *T)
    sumX1  +=       0.008 * cos(5.66351050239  + 1224.06165826211 *T)
    sumX1  +=       0.009 * cos(4.15418404000  + 19316.94872284510 *T)
    sumX1  +=       0.007 * cos(5.10075848971  + 13541.17738500769 *T)
    sumX1  +=       0.009 * cos(3.23441439127  + 130443.64086071248 *T)
    sumX1  +=       0.007 * cos(2.30584818184  + 209812.84750217048 *T)
    sumX1  +=       0.007 * cos(0.64823715471  + 9384.59719059169 *T)
    sumX1  +=       0.007 * cos(0.61661379288  + 46848.08635728210 *T)
    sumX1  +=       0.007 * cos(3.64199179530  + 172402.28026229187 *T)
    sumX1  +=       0.007 * cos(0.75404668288  + 846.32665223470 *T)
    sumX1  +=       0.007 * cos(3.05134673809  + 27044.43611502830 *T)
    sumX1  +=       0.007 * cos(1.99335698169  + 150244.58681693728 *T)
    sumX1  +=       0.009 * cos(3.93094161276  + 26735.15779237250 *T)
    sumX1  +=       0.007 * cos(4.75721708287  + 52325.61329776649 *T)
    sumX1  +=       0.007 * cos(1.46600264602  + 45892.97425064050 *T)
    sumX1  +=       0.008 * cos(2.77799241232  + 52169.42154173189 *T)
    sumX1  +=       0.008 * cos(5.13696206564  + 83924.79765639128 *T)
    sumX1  +=       0.007 * cos(5.38422375135  + 80597.14963507809 *T)
    sumX1  +=       0.006 * cos(3.78360025001  + 52279.14287485049 *T)
    sumX1  +=       0.006 * cos(3.14185350605  + 51013.57539622409 *T)
    sumX1  +=       0.008 * cos(3.74001326185  + 129387.49114216608 *T)
    sumX1  +=       0.009 * cos(3.71121213994  + 955.84355909210 *T)
    sumX1  +=       0.008 * cos(1.50347216209  + 77727.14873011068 *T)
    sumX1  +=       0.006 * cos(1.00543740922  + 79330.44871939608 *T)
    sumX1  +=       0.007 * cos(5.56521824483  + 78378.39253089130 *T)
    sumX1  +=       0.007 * cos(5.54385794189  + 51219.96272405629 *T)
    sumX1  +=       0.006 * cos(4.20431417253  + 129373.26404816448 *T)
    sumX1  +=       0.007 * cos(1.63093386109  + 26162.92855762500 *T)
    sumX1  +=       0.006 * cos(3.53855282447  + 103241.99019671988 *T)
    sumX1  +=       0.006 * cos(3.70872609374  + 3340.36860921629 *T)
    sumX1  +=       0.006 * cos(3.67291216777  + 35472.98796713290 *T)
    sumX1  +=       0.006 * cos(4.50914445356  + 130653.05862079248 *T)
    sumX1  +=       0.006 * cos(6.20953170755  + 110.45013870291 *T)
    sumX1  +=       0.007 * cos(1.43696950392  + 77631.16950289288 *T)
    sumX1  +=       0.006 * cos(1.84528711156  + 53131.16220727349 *T)
    sumX1  +=       0.006 * cos(4.02329719726  + 26013.36536049040 *T)
    sumX1  +=       0.006 * cos(0.25065144064  + 25508.45937205890 *T)
    sumX1  +=       0.006 * cos(5.41297359057  + 26667.83454605650 *T)
    sumX1  +=       0.006 * cos(1.27873892607  + 70268.93716521488 *T)
    sumX1  +=       0.006 * cos(3.75153393501  + 103918.14464590348 *T)
    sumX1  +=       0.008 * cos(0.93245740025  + 157587.04459711787 *T)
    sumX1  +=       0.006 * cos(1.85606762274  + 6885.39370741431 *T)
    sumX1  +=       0.007 * cos(4.77971152102  + 214364.80099922928 *T)
    sumX1  +=       0.007 * cos(4.36588080270  + 25754.29101828830 *T)
    sumX1  +=       0.007 * cos(1.29774422851  + 26422.00289982710 *T)
    sumX1  +=       0.006 * cos(2.52395990409  + 60055.65161900389 *T)
    sumX1  +=       0.006 * cos(6.00893693971  + 78149.51395352086 *T)
    sumX1  +=       0.005 * cos(0.89617169225  + 50064.39978725430 *T)
    sumX1  +=       0.006 * cos(2.63212801251  + 128221.00242116768 *T)
    sumX1  +=       0.006 * cos(4.95615912967  + 75615.49841673308 *T)
    sumX1  +=       0.005 * cos(1.27204675441  + 137678.43511695447 *T)
    sumX1  +=       0.007 * cos(5.96534970508  + 56727.51596272369 *T)
    sumX1  +=       0.007 * cos(5.86123151663  + 8194.03151572510 *T)
    sumX1  +=       0.005 * cos(6.05965648339  + 130226.46042991648 *T)
    sumX1  +=       0.006 * cos(5.74195245077  + 80382.71710258449 *T)
    sumX1  +=       0.006 * cos(2.26934702805  + 51653.47268253809 *T)
    sumX1  +=       0.006 * cos(0.06833462464  + 123669.04892410888 *T)
    sumX1  +=       0.006 * cos(3.85232597863  + 72602.13355808688 *T)
    sumX1  +=       0.006 * cos(4.43482215622  + 50056.79860528649 *T)
    sumX1  +=       0.005 * cos(1.17789065193  + 102769.89703549728 *T)
    sumX1  +=       0.006 * cos(3.77265893918  + 74820.89066227368 *T)
    sumX1  +=       0.006 * cos(1.81837030655  + 161079.61616398749 *T)
    sumX1  +=       0.006 * cos(3.17899498144  + 103932.37173990508 *T)
    sumX1  +=       0.006 * cos(2.61088164463  + 25234.95057730570 *T)
    sumX1  +=       0.007 * cos(1.11412129297  + 154938.58977164488 *T)
    sumX1  +=       0.005 * cos(3.94989870298  + 419.24082639170 *T)
    sumX1  +=       0.005 * cos(3.88779112513  + 19406.92210565810 *T)
    sumX1  +=       0.006 * cos(3.53467301913  + 1265.81129610991 *T)
    sumX1  +=       0.007 * cos(0.99187081299  + 64901.50353540690 *T)
    sumX1  +=       0.005 * cos(2.45437001625  + 73711.51211018028 *T)
    sumX1  +=       0.005 * cos(0.40751082829  + 78213.95662030189 *T)
    sumX1  +=       0.006 * cos(0.47609620784  + 182085.87484340827 *T)
    sumX1  +=       0.005 * cos(0.72963478402  + 102755.66994149568 *T)
    sumX1  +=       0.005 * cos(3.23310371062  + 40565.01050729069 *T)
    sumX1  +=       0.005 * cos(4.56090628241  + 102.84895673509 *T)
    sumX1  +=       0.005 * cos(2.73032624940  + 51748.96427478889 *T)
    sumX1  +=       0.005 * cos(2.42117857215  + 53399.86794141051 *T)
    sumX1  +=       0.005 * cos(3.99699244254  + 27780.31262856009 *T)
    sumX1  +=       0.005 * cos(6.27879128389  + 80482.22271142589 *T)
    sumX1  +=       0.005 * cos(4.04842386586  + 25938.58376192310 *T)
    sumX1  +=       0.005 * cos(0.51237606895  + 26402.33313892731 *T)
    sumX1  +=       0.005 * cos(5.15124896226  + 25773.96077918809 *T)
    sumX1  +=       0.005 * cos(1.58549672991  + 77616.94240889128 *T)
    sumX1  +=       0.005 * cos(2.07804016284  + 34282.42229226630 *T)
    sumX1  +=       0.006 * cos(0.18266035721  + 50050.17269325269 *T)
    sumX1  +=       0.005 * cos(3.56668263605  + 17893.87162584910 *T)
    sumX1  +=       0.004 * cos(2.54682197432  + 130432.64597835368 *T)
    sumX1  +=       0.004 * cos(1.97821363637  + 51852.54468397449 *T)
    sumX1  +=       0.004 * cos(0.66387666740  + 32370.73517408209 *T)
    sumX1  +=       0.005 * cos(2.92531330514  + 204151.51545301828 *T)
    sumX1  +=       0.006 * cos(3.31682565623  + 75930.75684933408 *T)
    sumX1  +=       0.004 * cos(3.69733037417  + 25668.17468021549 *T)
    sumX1  +=       0.005 * cos(6.08000873981  + 6043.98476389190 *T)
    sumX1  +=       0.005 * cos(4.31437334805  + 39450.59666585690 *T)
    sumX1  +=       0.005 * cos(3.85449175904  + 24491.96051677309 *T)
    sumX1  +=       0.004 * cos(2.80245159203  + 51955.63745819309 *T)
    sumX1  +=       0.004 * cos(5.66172545911  + 103498.66000202828 *T)
    sumX1  +=       0.005 * cos(1.63974900109  + 52396.46274307070 *T)
    sumX1  +=       0.004 * cos(4.13434256469  + 104505.63519426508 *T)
    sumX1  +=       0.005 * cos(4.01690246624  + 23754.46293121869 *T)
    sumX1  +=       0.005 * cos(0.99791835393  + 123201.08393375447 *T)
    sumX1  +=       0.004 * cos(1.49755025165  + 54087.24958384910 *T)
    sumX1  +=       0.004 * cos(5.20879448063  + 637.24008950771 *T)
    sumX1  +=       0.004 * cos(2.16539647390  + 28791.76311373330 *T)
    sumX1  +=       0.004 * cos(3.49822855731  + 23384.53080438210 *T)
    sumX1  +=       0.005 * cos(6.09748705519  + 53029.24648238390 *T)
    sumX1  +=       0.005 * cos(4.96740570225  + 18848.98373249069 *T)
    sumX1  +=       0.004 * cos(3.38348270644  + 26824.02347258949 *T)
    sumX1  +=       0.004 * cos(2.28014232477  + 25352.27044552590 *T)
    sumX1  +=       0.004 * cos(5.11054976948  + 133882.33447031991 *T)
    sumX1  +=       0.004 * cos(2.44432694792  + 132658.51662954129 *T)
    sumX1  +=       0.004 * cos(4.79464066723  + 53772.23654291649 *T)
    sumX1  +=       0.004 * cos(3.58245667544  + 183571.16555011147 *T)
    sumX1  +=       0.004 * cos(1.18282836563  + 167850.32676523307 *T)
    sumX1  +=       0.005 * cos(0.57469466163  + 145204.99856862429 *T)
    sumX1  +=       0.004 * cos(2.91942328210  + 27171.22719135130 *T)
    sumX1  +=       0.004 * cos(2.74420174911  + 25005.06672676410 *T)
    sumX1  +=       0.004 * cos(4.57300561972  + 50483.88443112950 *T)
    sumX1  +=       0.004 * cos(1.71297602638  + 78786.53066029989 *T)
    sumX1  +=       0.004 * cos(4.98597716653  + 949.41942645330 *T)
    sumX1  +=       0.005 * cos(5.48294820542  + 45892.48661567349 *T)
    sumX1  +=       0.004 * cos(5.09042463265  + 1795.50226120450 *T)
    sumX1  +=       0.005 * cos(3.41506026848  + 26709.89075989690 *T)
    sumX1  +=       0.005 * cos(2.24856476273  + 25466.40315821850 *T)
    sumX1  +=       0.004 * cos(5.84172747790  + 52065.84377941249 *T)
    sumX1  +=       0.004 * cos(1.67956658507  + 2222.10045208050 *T)
    sumX1  +=       0.004 * cos(2.68517686010  + 78270.58180110609 *T)
    sumX1  +=       0.004 * cos(4.37036821346  + 25653.94758621389 *T)
    sumX1  +=       0.004 * cos(3.79249696812  + 49842.36607279289 *T)
    sumX1  +=       0.004 * cos(0.19068653380  + 143005.91122533727 *T)
    sumX1  +=       0.004 * cos(3.27752582001  + 78800.75775430149 *T)
    sumX1  +=       0.004 * cos(4.85511136724  + 65697.31390725628 *T)
    sumX1  +=       0.003 * cos(0.19189003895  + 52195.23222656469 *T)
    sumX1  +=       0.004 * cos(5.02476065705  + 130459.42928625426 *T)
    sumX1  +=       0.004 * cos(3.97588615914  + 24491.47288180609 *T)
    sumX1  +=       0.004 * cos(4.14710532879  + 3178.38960805111 *T)
    sumX1  +=       0.004 * cos(2.12747516147  + 220.16882495529 *T)
    sumX1  +=       0.004 * cos(4.96500381777  + 52250.83169919920 *T)
    sumX1  +=       0.003 * cos(4.30211830006  + 78160.86046798748 *T)
    sumX1  +=       0.004 * cos(4.46270461199  + 87367.86023632369 *T)
    sumX1  +=       0.004 * cos(1.72859821695  + 130435.87818999648 *T)
    sumX1  +=       0.004 * cos(4.85546199905  + 53234.94439585409 *T)
    sumX1  +=       0.004 * cos(1.63790933088  + 26575.78171031190 *T)
    sumX1  +=       0.004 * cos(4.02571570033  + 25600.51220780350 *T)
    sumX1  +=       0.003 * cos(0.66645810174  + 1486.22393854870 *T)
    sumX1  +=       0.005 * cos(4.92448477691  + 86144.04239554508 *T)
    sumX1  +=       0.004 * cos(2.91181818574  + 32132.37554043310 *T)
    sumX1  +=       0.004 * cos(2.61773549272  + 66653.40128383189 *T)
    sumX1  +=       0.003 * cos(1.93014248251  + 52310.15915021690 *T)
    sumX1  +=       0.003 * cos(3.81387230068  + 45290.90021070109 *T)
    sumX1  +=       0.003 * cos(0.34402480925  + 70383.86408886709 *T)
    sumX1  +=       0.003 * cos(4.41765442359  + 52252.31617190749 *T)
    sumX1  +=       0.005 * cos(1.98227688514  + 52808.83383994509 *T)
    sumX1  +=       0.003 * cos(5.60337796614  + 1588.82907780029 *T)
    sumX1  +=       0.004 * cos(5.75007070234  + 58857.27495403150 *T)
    sumX1  +=       0.004 * cos(4.77171683373  + 51951.70530492999 *T)
    sumX1  +=       0.003 * cos(3.97223691848  + 50264.85061741470 *T)
    sumX1  +=       0.004 * cos(2.74364830901  + 20043.91837768230 *T)
    sumX1  +=       0.004 * cos(5.35376721002  + 128320.99566497609 *T)
    sumX1  +=       0.003 * cos(5.53627007155  + 25986.18444079209 *T)
    sumX1  +=       0.003 * cos(0.12735495966  + 26190.10947732330 *T)
    sumX1  +=       0.004 * cos(3.03031240281  + 136722.83537534589 *T)
    sumX1  +=       0.003 * cos(4.41334655990  + 65717.47130312308 *T)
    sumX1  +=       0.003 * cos(4.19581525920  + 181026.49291321909 *T)
    sumX1  +=       0.003 * cos(2.72755670550  + 51535.66517935089 *T)
    sumX1  +=       0.004 * cos(1.89805960768  + 129799.86223904048 *T)
    sumX1  +=       0.003 * cos(0.20996976206  + 26073.91986505609 *T)
    sumX1  +=       0.003 * cos(5.45365526915  + 26102.37405305930 *T)
    sumX1  +=       0.003 * cos(3.40818167058  + 52168.44891866409 *T)
    sumX1  +=       0.004 * cos(0.98011046066  + 52155.89270476510 *T)
    sumX1  +=       0.004 * cos(2.86756899369  + 37698.69891743190 *T)
    sumX1  +=       0.003 * cos(5.28107423901  + 51109.06698847489 *T)
    sumX1  +=       0.003 * cos(3.44421079458  + 26247.44869384990 *T)
    sumX1  +=       0.003 * cos(2.21941423663  + 25928.84522426550 *T)
    sumX1  +=       0.004 * cos(1.88058917739  + 468.20880783791 *T)
    sumX1  +=       0.004 * cos(4.84952063164  + 132028.83242063827 *T)
    sumX1  +=       0.004 * cos(4.04256757305  + 25661.06113321469 *T)
    sumX1  +=       0.003 * cos(5.18635157447  + 2703.37233719210 *T)
    sumX1  +=       0.004 * cos(1.14854014554  + 166740.94821313969 *T)
    sumX1  +=       0.004 * cos(1.66492426174  + 59414.23805726489 *T)
    sumX1  +=       0.003 * cos(0.11263836231  + 61560.89110870710 *T)
    sumX1  +=       0.003 * cos(1.57772720091  + 79316.22162539448 *T)
    sumX1  +=       0.003 * cos(4.62863855126  + 130866.35771623049 *T)
    sumX1  +=       0.003 * cos(4.83516317774  + 79219.06534884768 *T)
    sumX1  +=       0.003 * cos(4.03532450201  + 24506.18761077469 *T)
    sumX1  +=       0.003 * cos(2.23216290209  + 51868.49247966230 *T)
    sumX1  +=       0.003 * cos(1.28804069436  + 19202.50943415989 *T)
    sumX1  +=       0.003 * cos(4.56968797112  + 40852.89836736050 *T)
    sumX1  +=       0.003 * cos(4.38565150092  + 116783.89903417807 *T)
    sumX1  +=       0.003 * cos(4.07787007514  + 105460.74730090667 *T)
    sumX1  +=       0.003 * cos(1.85194703747  + 48835.43767393209 *T)
    sumX1  +=       0.003 * cos(5.69146662946  + 114.68310616872 *T)
    sumX1  +=       0.003 * cos(0.19144773585  + 76785.08666814168 *T)
    sumX1  +=       0.003 * cos(2.42893576806  + 81592.08889848628 *T)
    sumX1  +=       0.003 * cos(1.24692506910  + 95247.46175469569 *T)
    sumX1  +=       0.003 * cos(3.12006782424  + 52171.68113030689 *T)
    sumX1  +=       0.004 * cos(0.40583501169  + 78110.17443172129 *T)
    sumX1  +=       0.003 * cos(0.88564181183  + 68050.18006102808 *T)
    sumX1  +=       0.003 * cos(4.02600027527  + 38653.81102407349 *T)
    sumX1  +=       0.003 * cos(2.89741869348  + 2219.00092167030 *T)
    sumX1  +=       0.003 * cos(4.35261728870  + 52179.44380102290 *T)
    sumX1  +=       0.003 * cos(5.72167155686  + 52286.25642185129 *T)
    sumX1  +=       0.003 * cos(4.88587761079  + 76674.39271195528 *T)
    sumX1  +=       0.003 * cos(2.50139495886  + 126997.18458038908 *T)
    sumX1  +=       0.003 * cos(6.06197330359  + 77101.47853779829 *T)
    sumX1  +=       0.003 * cos(5.97751979418  + 77741.37582411229 *T)
    sumX1  +=       0.003 * cos(2.52437891388  + 8989.22388794889 *T)
    sumX1  +=       0.003 * cos(3.70897028008  + 77733.77464214448 *T)
    sumX1  +=       0.003 * cos(6.24188448720  + 286966.69073983253 *T)
    sumX1  +=       0.003 * cos(5.68828427997  + 74.53778108379 *T)
    sumX1  +=       0.003 * cos(2.64767376731  + 632.53992182970 *T)
    sumX1  +=       0.004 * cos(6.05038593627  + 28287.23430234470 *T)
    sumX1  +=       0.004 * cos(5.89642440212  + 23889.05961577070 *T)
    sumX1  +=       0.003 * cos(3.64902489843  + 75930.26921436709 *T)
    sumX1  +=       0.002 * cos(3.26392643567  + 182188.96761762688 *T)
    sumX1  +=       0.003 * cos(6.24797687447  + 27250.62166346549 *T)
    sumX1  +=       0.003 * cos(5.71051457818  + 51315.74017187909 *T)
    sumX1  +=       0.003 * cos(2.95050506838  + 52182.67601266569 *T)
    sumX1  +=       0.002 * cos(0.14560269876  + 121335.85253123687 *T)
    sumX1  +=       0.002 * cos(2.31563547899  + 110012.70079796547 *T)
    sumX1  +=       0.002 * cos(2.79146825409  + 2008.80135664250 *T)
    sumX1  +=       0.002 * cos(5.33939376534  + 24925.67225464990 *T)
    sumX1  +=       0.003 * cos(5.02309471101  + 39629.08052658189 *T)
    sumX1  +=       0.003 * cos(4.81867653228  + 3340.85624418330 *T)
    sumX1  +=       0.002 * cos(5.06629871936  + 23439.20449861769 *T)
    sumX1  +=       0.002 * cos(6.21328369214  + 29415.79498037089 *T)
    sumX1  +=       0.003 * cos(2.40182139304  + 25131.85780308709 *T)
    sumX1  +=       0.003 * cos(4.68865672548  + 51123.29408247649 *T)
    sumX1  +=       0.003 * cos(5.00160560591  + 90695.50825763687 *T)
    sumX1  +=       0.002 * cos(4.84919130926  + 24505.69997580769 *T)
    sumX1  +=       0.003 * cos(4.38371673981  + 26011.88088778210 *T)
    sumX1  +=       0.002 * cos(4.91686748502  + 51329.96726588070 *T)
    sumX1  +=       0.003 * cos(0.23735872795  + 81604.56566890588 *T)
    sumX1  +=       0.003 * cos(0.95743283729  + 52813.04637265610 *T)
    sumX1  +=       0.003 * cos(1.59253054642  + 391318.79094109638 *T)
    sumX1  +=       0.002 * cos(4.54043739808  + 196137.31725009429 *T)
    sumX1  +=       0.003 * cos(2.85252650926  + 156507.99290602890 *T)
    sumX1  +=       0.003 * cos(2.50847406676  + 27727.21647624570 *T)
    sumX1  +=       0.003 * cos(3.15515096445  + 24449.07744186970 *T)
    sumX1  +=       0.002 * cos(1.06183386008  + 78057.76769376888 *T)
    sumX1  +=       0.003 * cos(4.75030126782  + 78313.94986411028 *T)
    sumX1  +=       0.003 * cos(2.83857270203  + 203042.13690092487 *T)
    sumX1  +=       0.003 * cos(4.41522279829  + 34082.67540108190 *T)
    sumX1  +=       0.002 * cos(1.11682523208  + 38634.62889814070 *T)
    sumX1  +=       0.003 * cos(5.82930728472  + 78257.32468330608 *T)
    sumX1  +=       0.002 * cos(3.32397187418  + 49527.10764019189 *T)
    sumX1  +=       0.003 * cos(3.52707759552  + 52041.94105104690 *T)
    sumX1  +=       0.002 * cos(5.03728139459  + 104819.82137413470 *T)
    sumX1  +=       0.002 * cos(0.34249656430  + 62197.88738073129 *T)
    sumX1  +=       0.003 * cos(3.81629136441  + 103291.98681862408 *T)
    sumX1  +=       0.002 * cos(0.38902235709  + 51962.26337022689 *T)
    sumX1  +=       0.002 * cos(5.06187057536  + 35077.61466449009 *T)
    sumX1  +=       0.003 * cos(3.50433403816  + 96357.32794175608 *T)
    sumX1  +=       0.002 * cos(5.63169833454  + 25021.16384690069 *T)
    sumX1  +=       0.002 * cos(3.17912670229  + 66652.91364886488 *T)
    sumX1  +=       0.002 * cos(0.74572876767  + 199599.07432099240 *T)
    sumX1  +=       0.002 * cos(2.95582870605  + 9103.66317663410 *T)
    sumX1  +=       0.002 * cos(1.09917701972  + 74923.34081550628 *T)
    sumX1  +=       0.002 * cos(0.60560213674  + 17098.67925362531 *T)
    sumX1  +=       0.002 * cos(5.11946015221  + 78470.13879064328 *T)
    sumX1  +=       0.002 * cos(2.53118536507  + 101703.40155830728 *T)
    sumX1  +=       0.002 * cos(2.20613703147  + 77307.66408623548 *T)
    sumX1  +=       0.002 * cos(0.59446302071  + 104991.75367009427 *T)
    sumX1  +=       0.002 * cos(2.42950804559  + 86143.55476057807 *T)
    sumX1  +=       0.002 * cos(4.03918468742  + 26312.49175475960 *T)
    sumX1  +=       0.003 * cos(4.60683796484  + 69159.55861312148 *T)
    sumX1  +=       0.002 * cos(2.79539090828  + 27676.73221937449 *T)
    sumX1  +=       0.002 * cos(0.80954348332  + 91919.81373338249 *T)
    sumX1  +=       0.002 * cos(3.64345901161  + 225687.46509753368 *T)
    sumX1  +=       0.002 * cos(6.09474483307  + 60369.83779887349 *T)
    sumX1  +=       0.002 * cos(4.02958262076  + 26184.12618627550 *T)
    sumX1  +=       0.002 * cos(1.63404241045  + 25992.16773183990 *T)
    sumX1  +=       0.002 * cos(4.78056230516  + 62388.84801187589 *T)
    sumX1  +=       0.002 * cos(5.29281787520  + 7880.33297082250 *T)
    sumX1  +=       0.003 * cos(5.23217095654  + 51534.63653889291 *T)
    sumX1  +=       0.002 * cos(1.02304147154  + 24608.79275002629 *T)
    sumX1  +=       0.002 * cos(0.08655833344  + 183145.25677359747 *T)
    sumX1  +=       0.002 * cos(3.38094681034  + 177288.08970012009 *T)
    sumX1  +=       0.002 * cos(2.73964259090  + 26397.42528171349 *T)
    sumX1  +=       0.002 * cos(2.92398244031  + 25778.86863640189 *T)
    sumX1  +=       0.002 * cos(5.05032388283  + 42791.20909255670 *T)
    sumX1  +=       0.002 * cos(0.32496221869  + 38813.11275886569 *T)
    sumX1  +=       0.002 * cos(5.11101145665  + 51539.05382860770 *T)
    sumX1  +=       0.002 * cos(4.02707586748  + 24952.47008716450 *T)
    sumX1  +=       0.002 * cos(1.63654916373  + 27223.82383095090 *T)
    sumX1  +=       0.002 * cos(4.56989494681  + 103719.07264446709 *T)
    sumX1  +=       0.002 * cos(3.21686304521  + 25984.36477047710 *T)
    sumX1  +=       0.002 * cos(2.44676198600  + 26191.92914763830 *T)
    sumX1  +=       0.002 * cos(0.73451437899  + 25927.73958967290 *T)
    sumX1  +=       0.002 * cos(4.92911065222  + 26248.55432844250 *T)
    sumX1  +=       0.002 * cos(5.25833412712  + 113341.32408921266 *T)
    sumX1  +=       0.002 * cos(3.14486706475  + 53265.75889182550 *T)
    sumX1  +=       0.002 * cos(1.46628877500  + 51842.19415986250 *T)
    sumX1  +=       0.002 * cos(2.47984490464  + 16066.30967895830 *T)
    sumX1  +=       0.002 * cos(2.09863898672  + 29550.39166492290 *T)
    sumX1  +=       0.002 * cos(3.56498604449  + 22625.90225319250 *T)
    sumX1  +=       0.002 * cos(1.44463204026  + 27037.32256802749 *T)
    sumX1  +=       0.002 * cos(4.21899299095  + 25138.97135008790 *T)
    sumX1  +=       0.002 * cos(0.93669101848  + 157482.77477357027 *T)
    sumX1  +=       0.002 * cos(0.12919974221  + 52817.46366237089 *T)
    sumX1  +=       0.002 * cos(3.79805429136  + 104202.29318664568 *T)
    sumX1  +=       0.002 * cos(4.69592084098  + 25550.85481199529 *T)
    sumX1  +=       0.002 * cos(5.21767038269  + 52388.86156110289 *T)
    sumX1  +=       0.002 * cos(3.65031025117  + 233731.99016146208 *T)
    sumX1  +=       0.002 * cos(3.29338443989  + 54824.50492610490 *T)
    sumX1  +=       0.002 * cos(1.73836504660  + 13655.61667369290 *T)
    sumX1  +=       0.002 * cos(4.20276068024  + 78793.15657233368 *T)
    sumX1  +=       0.002 * cos(4.96143878613  + 168959.21768235945 *T)
    sumX1  +=       0.002 * cos(6.25349918418  + 205260.89400511168 *T)
    sumX1  +=       0.002 * cos(3.21379506167  + 128106.07549751546 *T)
    sumX1  +=       0.002 * cos(2.12894187827  + 50910.48262200550 *T)
    sumX1  +=       0.002 * cos(3.08974476939  + 51859.65823097529 *T)
    sumX1  +=       0.002 * cos(1.51881438178  + 25863.80216335580 *T)
    sumX1  +=       0.002 * cos(4.40672336509  + 53758.00944891489 *T)
    sumX1  +=       0.002 * cos(5.60768237774  + 52492.44197028850 *T)
    sumX1  +=       0.002 * cos(0.61004720971  + 735.63269604830 *T)
    sumX1  +=       0.002 * cos(0.61004720971  + 736.12033101530 *T)
    sumX1  +=       0.002 * cos(1.54306118494  + 1066.25165970650 *T)
    sumX1  +=       0.002 * cos(1.65860835610  + 215474.17955132265 *T)
    sumX1  +=       0.002 * cos(5.50882009325  + 26077.10125879380 *T)
    sumX1  +=       0.002 * cos(0.15480493796  + 26099.19265932160 *T)
    sumX1  +=       0.002 * cos(1.15953086967  + 96356.84030678908 *T)
    sumX1  +=       0.002 * cos(2.82477485660  + 50579.37602338029 *T)
    sumX1  +=       0.002 * cos(4.95075276458  + 176332.48995851149 *T)
    sumX1  +=       0.002 * cos(5.17774922517  + 78896.73698151928 *T)
    sumX1  +=       0.002 * cos(0.26673303771  + 26411.65237571510 *T)
    sumX1  +=       0.002 * cos(5.49340805923  + 87253.42094763846 *T)
    sumX1  +=       0.002 * cos(4.42579128459  + 181659.96606689427 *T)
    sumX1  +=       0.002 * cos(4.56681824452  + 23976.00901071309 *T)
    sumX1  +=       0.002 * cos(2.69925276343  + 102232.60488843489 *T)
    sumX1  +=       0.002 * cos(3.44678527119  + 52483.60772160150 *T)
    sumX1  +=       0.002 * cos(0.27890618815  + 199599.56195595945 *T)
    sumX1  +=       0.002 * cos(3.49834714879  + 25933.88051360590 *T)
    sumX1  +=       0.002 * cos(5.12143437332  + 23439.69213358470 *T)
    sumX1  +=       0.001 * cos(4.63971494884  + 27351.30432163590 *T)
    sumX1  +=       0.001 * cos(1.02391008237  + 24824.98959647950 *T)
    sumX1  +=       0.001 * cos(4.66880968358  + 78697.66498008289 *T)
    sumX1  +=       0.002 * cos(5.51544937611  + 60370.32543384050 *T)
    sumX1  +=       0.001 * cos(3.24905067191  + 51596.36251363310 *T)
    sumX1  +=       0.002 * cos(4.53665275519  + 103815.05187168487 *T)
    sumX1  +=       0.002 * cos(1.48707290600  + 33326.33491569069 *T)
    sumX1  +=       0.002 * cos(0.70194343478  + 51226.87449166209 *T)
    sumX1  +=       0.001 * cos(2.48229161835  + 77307.86586563049 *T)
    sumX1  +=       0.002 * cos(2.89900338624  + 50952.23225985329 *T)
    sumX1  +=       0.002 * cos(5.06785546369  + 65716.98366815608 *T)
    sumX1  +=       0.001 * cos(3.03622093333  + 231348.79714668583 *T)
    sumX1  +=       0.001 * cos(1.22871462641  + 163765.85062356168 *T)
    sumX1  +=       0.002 * cos(5.81358049715  + 34281.93465729930 *T)
    sumX1  +=       0.001 * cos(0.01895053315  + 25234.46294233869 *T)
    sumX1  +=       0.001 * cos(5.99824107896  + 147423.75567281109 *T)
    sumX1  +=       0.002 * cos(1.73430601364  + 26514.25751496669 *T)
    sumX1  +=       0.001 * cos(0.46522663694  + 3442.33112748190 *T)
    sumX1  +=       0.001 * cos(2.66766894498  + 102762.29585352947 *T)
    sumX1  +=       0.001 * cos(4.19740391458  + 189854.24140010288 *T)
    sumX1  +=       0.001 * cos(4.71704307559  + 54879.66625530750 *T)
    sumX1  +=       0.002 * cos(4.94622107659  + 94330.01910368088 *T)
    sumX1  +=       0.002 * cos(3.22708232430  + 22747.04689739089 *T)
    sumX1  +=       0.001 * cos(5.03032609104  + 138319.84867869344 *T)
    sumX1  +=       0.001 * cos(4.79191528728  + 52509.90604140130 *T)
    sumX1  +=       0.001 * cos(5.71338507031  + 35191.56631820829 *T)
    sumX1  +=       0.002 * cos(5.64922837609  + 26713.81715137009 *T)
    sumX1  +=       0.002 * cos(0.01439665512  + 25462.47676674530 *T)
    sumX1  +=       0.001 * cos(6.11047772238  + 76667.27916495448 *T)
    sumX1  +=       0.001 * cos(1.01940165259  + 116783.41139921108 *T)
    sumX1  +=       0.002 * cos(4.11674730434  + 23919.38647677510 *T)
    sumX1  +=       0.002 * cos(1.54687772687  + 28256.90744134030 *T)
    sumX1  +=       0.002 * cos(2.55506634692  + 99024.38027539747 *T)
    sumX1  +=       0.001 * cos(5.53588784644  + 112546.13171698888 *T)
    sumX1  +=       0.002 * cos(4.36940598830  + 26241.43813457550 *T)
    sumX1  +=       0.001 * cos(0.47478664591  + 129329.89333829407 *T)
    sumX1  +=       0.002 * cos(3.32047464917  + 118828.71993201328 *T)
    sumX1  +=       0.001 * cos(3.81811027049  + 49956.80536147810 *T)
    sumX1  +=       0.001 * cos(1.58300591770  + 28199.79727243529 *T)
    sumX1  +=       0.001 * cos(1.36639947851  + 68241.62832713968 *T)
    sumX1  +=       0.001 * cos(5.16896085902  + 103821.67778371868 *T)
    sumX1  +=       0.002 * cos(1.67830961245  + 25441.13612574289 *T)
    sumX1  +=       0.001 * cos(4.61230342742  + 80175.15272542329 *T)
    sumX1  +=       0.002 * cos(0.48315631972  + 23962.26955167850 *T)
    sumX1  +=       0.001 * cos(3.50023943727  + 26934.22979380890 *T)
    sumX1  +=       0.001 * cos(5.56740631565  + 42153.72518556549 *T)
    sumX1  +=       0.001 * cos(0.57318968993  + 50380.79165691089 *T)
    sumX1  +=       0.001 * cos(4.87428065409  + 241562.08269289689 *T)
    sumX1  +=       0.001 * cos(6.24255788428  + 55618.13741063030 *T)
    sumX1  +=       0.001 * cos(4.91389387543  + 92740.81679043909 *T)
    sumX1  +=       0.002 * cos(5.62002771737  + 27567.01353312209 *T)
    sumX1  +=       0.002 * cos(0.86522458907  + 98690.03669966108 *T)
    sumX1  +=       0.001 * cos(3.83786759867  + 173511.17117941825 *T)
    sumX1  +=       0.001 * cos(3.83786759867  + 173511.65881438527 *T)
    sumX1  +=       0.001 * cos(3.73710353066  + 102018.17235594128 *T)
    sumX1  +=       0.001 * cos(3.98058009404  + 49954.19346603489 *T)
    sumX1  +=       0.001 * cos(0.70035139387  + 155461.16718973868 *T)
    sumX1  +=       0.001 * cos(1.23780419541  + 193937.74227184025 *T)
    sumX1  +=       0.001 * cos(4.37321488294  + 26927.11624680809 *T)
    sumX1  +=       0.001 * cos(1.29041014827  + 25249.17767130730 *T)
    sumX1  +=       0.001 * cos(0.70197977264  + 149642.51277699787 *T)
    sumX1  +=       0.001 * cos(2.49619638334  + 112231.45790215228 *T)
    sumX1  +=       0.001 * cos(3.31600584150  + 639.65346883050 *T)
    sumX1  +=       0.001 * cos(2.81858522024  + 28420.85571696270 *T)
    sumX1  +=       0.002 * cos(0.06574072735  + 55503.69812194508 *T)
    sumX1  +=       0.001 * cos(5.90451405265  + 35472.50033216589 *T)
    sumX1  +=       0.001 * cos(1.51650944248  + 48997.90431006430 *T)
    sumX1  +=       0.001 * cos(5.22070652557  + 27006.07724503950 *T)
    sumX1  +=       0.001 * cos(0.44291850564  + 25170.21667307590 *T)
    sumX1  +=       0.001 * cos(1.78735146281  + 106263.05586699749 *T)
    sumX1  +=       0.001 * cos(4.72084316953  + 90989.40667698107 *T)
    sumX1  +=       0.001 * cos(4.99715761034  + 50690.06997956669 *T)
    sumX1  +=       0.001 * cos(0.33557225431  + 77836.86741636308 *T)
    sumX1  +=       0.001 * cos(3.31386459990  + 125112.28341697170 *T)
    sumX1  +=       0.001 * cos(5.55100305938  + 26610.23674218449 *T)
    sumX1  +=       0.001 * cos(4.62894445056  + 143980.69309287865 *T)
    sumX1  +=       0.001 * cos(0.73494247156  + 153084.60008699627 *T)
    sumX1  +=       0.001 * cos(1.17231831682  + 45404.85186441930 *T)
    sumX1  +=       0.001 * cos(2.54701077766  + 2332.95257538850 *T)
    sumX1  +=       0.001 * cos(4.07473618347  + 76152.30292882849 *T)
    sumX1  +=       0.001 * cos(2.13363904714  + 45494.33761226529 *T)
    sumX1  +=       0.001 * cos(0.33894588198  + 24925.18461968289 *T)
    sumX1  +=       0.001 * cos(2.22391897375  + 50593.60311738189 *T)
    sumX1  +=       0.001 * cos(5.06402826535  + 115674.03284711768 *T)
    sumX1  +=       0.001 * cos(6.21839548284  + 151200.18655854589 *T)
    sumX1  +=       0.001 * cos(1.41453859000  + 84308.18182358149 *T)
    sumX1  +=       0.001 * cos(0.92714278131  + 129379.88996019828 *T)
    sumX1  +=       0.001 * cos(5.82040259792  + 24073.16528725990 *T)
    sumX1  +=       0.001 * cos(6.12640774047  + 28103.12863085550 *T)
    sumX1  +=       0.001 * cos(2.07741344436  + 19406.43447069109 *T)
    sumX1  +=       0.001 * cos(0.51650041444  + 24844.65935737930 *T)
    sumX1  +=       0.001 * cos(5.14712461677  + 27331.63456073610 *T)
    sumX1  +=       0.001 * cos(1.95998075094  + 154408.89880655028 *T)
    sumX1  +=       0.001 * cos(6.15204641708  + 163298.37326817427 *T)
    sumX1  +=       0.001 * cos(5.62631785594  + 58220.27868200730 *T)
    sumX1  +=       0.001 * cos(2.38135876415  + 97670.63153038068 *T)
    sumX1  +=       0.001 * cos(0.13326662566  + 22759.52366781050 *T)
    sumX1  +=       0.001 * cos(1.74239852457  + 24395.98128955530 *T)
    sumX1  +=       0.001 * cos(5.76996425564  + 92741.30442540610 *T)
    sumX1  +=       0.001 * cos(5.18001187713  + 128850.19899510368 *T)
    sumX1  +=       0.001 * cos(1.70882749408  + 25984.56654987209 *T)
    sumX1  +=       0.001 * cos(4.63169895688  + 24357.02460612510 *T)
    sumX1  +=       0.001 * cos(1.03192607433  + 27819.26931199030 *T)
    sumX1  +=       0.001 * cos(6.18803487660  + 183724.45672562928 *T)
    sumX1  +=       0.001 * cos(0.75326620837  + 26624.46383618609 *T)
    sumX1  +=       0.001 * cos(1.73910591832  + 25455.36321974450 *T)
    sumX1  +=       0.001 * cos(1.73910591832  + 25454.87558477749 *T)
    sumX1  +=       0.001 * cos(4.91320505829  + 189385.78877478145 *T)
    sumX1  +=       0.001 * cos(4.91320505829  + 189386.27640974848 *T)
    sumX1  +=       0.001 * cos(0.33962058759  + 209659.06869168568 *T)
    sumX1  +=       0.001 * cos(0.57831111391  + 109.96250373590 *T)
    sumX1  +=       0.001 * cos(3.01192529140  + 82815.41910429788 *T)
    sumX1  +=       0.001 * cos(5.28068518753  + 147423.26803784407 *T)
    sumX1  +=       0.001 * cos(5.41436422726  + 23704.95394428149 *T)
    sumX1  +=       0.001 * cos(0.24926080395  + 28471.33997383390 *T)
    sumX1  +=       0.001 * cos(5.21985223980  + 51440.17358710009 *T)
    sumX1  +=       0.001 * cos(5.81513934032  + 25035.39094090229 *T)
    sumX1  +=       0.001 * cos(5.63545263237  + 26555.62431444510 *T)
    sumX1  +=       0.001 * cos(4.02649464248  + 26177.63270690369 *T)
    sumX1  +=       0.001 * cos(1.63713038873  + 25998.66121121170 *T)
    sumX1  +=       0.001 * cos(1.98713499450  + 233681.99353955788 *T)
    sumX1  +=       0.001 * cos(3.17318181579  + 21650.63275068411 *T)
    sumX1  +=       0.001 * cos(2.49044321542  + 30525.66116743129 *T)
    sumX1  +=       0.001 * cos(1.93284114667  + 111590.04434041328 *T)
    sumX1  +=       0.001 * cos(4.15055550488  + 24203.24579564031 *T)
    sumX1  +=       0.001 * cos(1.51306952632  + 27973.04812247509 *T)
    sumX1  +=       0.001 * cos(4.04308594118  + 99799.41525175449 *T)
    sumX1  +=       0.001 * cos(2.47194170898  + 128106.56313248246 *T)
    sumX1  +=       0.001 * cos(3.82198016519  + 104198.07757329548 *T)
    sumX1  +=       0.001 * cos(1.80561843054  + 132350.95900857169 *T)
    sumX1  +=       0.001 * cos(3.89409837101  + 6680.98103591609 *T)
    sumX1  +=       0.001 * cos(3.02028039042  + 26202.09861277591 *T)
    sumX1  +=       0.001 * cos(2.86028052056  + 50167.49256147289 *T)
    sumX1  +=       0.001 * cos(3.13915436914  + 313054.59388140670 *T)
    sumX1  +=       0.001 * cos(0.04271369503  + 131548.65044248087 *T)
    sumX1  +=       0.001 * cos(4.92373111036  + 85502.14119883909 *T)
    sumX1  +=       0.001 * cos(2.02772617744  + 51542.77872635169 *T)
    sumX1  +=       0.001 * cos(5.92130408913  + 102018.65999090829 *T)
    sumX1  +=       0.001 * cos(3.43111189593  + 51322.36608391289 *T)
    sumX1  +=       0.001 * cos(1.54854247637  + 78413.51643934069 *T)
    sumX1  +=       0.001 * cos(6.28162689960  + 156531.54400228668 *T)
    sumX1  +=       0.001 * cos(1.20531548605  + 105306.96849042187 *T)
    sumX1  +=       0.001 * cos(2.58426030730  + 35077.12702952309 *T)
    sumX1  +=       0.001 * cos(1.09812311120  + 155467.79310177249 *T)
    sumX1  +=       0.001 * cos(1.33162407420  + 45290.41257573408 *T)
    sumX1  +=       0.001 * cos(3.52374631003  + 1162.23088692429 *T)
    sumX1  +=       0.001 * cos(4.28752586027  + 77211.19722405069 *T)
    sumX1  +=       0.001 * cos(4.51338458149  + 25152.21697834890 *T)
    sumX1  +=       0.001 * cos(1.15024044972  + 27024.07693976651 *T)
    sumX1  +=       0.001 * cos(4.26343278626  + 247223.41474204903 *T)
    sumX1  +=       0.001 * cos(3.96461816746  + 183674.94773869208 *T)
    sumX1  +=       0.001 * cos(3.14345130679  + 25131.37016812009 *T)
    sumX1  +=       0.001 * cos(6.06920679029  + 26126.27999469550 *T)
    sumX1  +=       0.001 * cos(5.87760354810  + 26050.01392341990 *T)
    sumX1  +=       0.001 * cos(0.97484682364  + 71980.38975724769 *T)
    sumX1  +=       0.001 * cos(5.34892363416  + 27388.97377726269 *T)
    sumX1  +=       0.001 * cos(0.31470139705  + 24787.32014085270 *T)
    sumX1  +=       0.001 * cos(0.51689218530  + 155475.39428374029 *T)
    sumX1  +=       0.001 * cos(0.66134795423  + 51756.07782178969 *T)
    sumX1  +=       0.001 * cos(2.18850969113  + 54086.76194888209 *T)
    sumX1  +=       0.001 * cos(4.14165011369  + 94138.08320260228 *T)
    sumX1  +=       0.001 * cos(5.17339556279  + 24602.85625235450 *T)
    sumX1  +=       0.001 * cos(0.49022946842  + 27573.43766576090 *T)
    sumX1  +=       0.001 * cos(4.40553516948  + 121335.36489626988 *T)
    sumX1  +=       0.001 * cos(0.40614537067  + 198490.18340386607 *T)
    sumX1  +=       0.001 * cos(0.84588819348  + 181555.69624334667 *T)
    sumX1  +=       0.001 * cos(5.13509102157  + 156534.77621392949 *T)
    sumX1  +=       0.001 * cos(1.73614423183  + 12725.69725225850 *T)
    sumX1  +=       0.001 * cos(0.52115552294  + 130006.04778747768 *T)
    sumX1  +=       0.001 * cos(1.59903474540  + 89586.12970554348 *T)
    sumX1  +=       0.001 * cos(3.74413194011  + 131394.87163199607 *T)
    sumX1  +=       0.001 * cos(0.61669479944  + 77039.64776646046 *T)

    # MERCURY X Coord
    sumX2  = 0
    sumX2  +=   20000.263 * cos(5.96893489541  + 26088.14695905770 *T)
    sumX2  +=    8268.782 * cos(0.41593027178  + 0.24381748350 *T)
    sumX2  +=    2498.175 * cos(3.38864668295  + 52176.05010063190 *T)
    sumX2  +=     949.347 * cos(0.79533142414  + 78263.95324220609 *T)
    sumX2  +=     776.108 * cos(5.52747825416  + 26087.65932409069 *T)
    sumX2  +=     371.109 * cos(4.03419264198  + 104351.85638378028 *T)
    sumX2  +=     130.259 * cos(0.87132938107  + 130439.75952535449 *T)
    sumX2  +=      69.902 * cos(2.02066051600  + 52175.56246566489 *T)
    sumX2  +=      42.865 * cos(3.95497707255  + 156527.66266692869 *T)
    sumX2  +=      13.564 * cos(0.74070016642  + 182615.56580850287 *T)
    sumX2  +=      11.344 * cos(4.89785767709  + 78263.46560723908 *T)
    sumX2  +=       4.184 * cos(3.80256859065  + 208703.46895007708 *T)
    sumX2  +=       2.952 * cos(1.21119140823  + 24978.76840696430 *T)
    sumX2  +=       2.715 * cos(4.19935661957  + 27197.52551115110 *T)
    sumX2  +=       2.412 * cos(1.63037150116  + 104351.36874881329 *T)
    sumX2  +=       1.268 * cos(0.57748866225  + 234791.37209165128 *T)
    sumX2  +=       1.017 * cos(4.33432055001  + 51066.67154853850 *T)
    sumX2  +=       0.591 * cos(4.73978384246  + 130439.27189038748 *T)
    sumX2  +=       0.372 * cos(0.82364022383  + 53285.42865272530 *T)
    sumX2  +=       0.379 * cos(3.63308186625  + 260879.27523322543 *T)
    sumX2  +=       0.353 * cos(1.13633330939  + 77154.57469011268 *T)
    sumX2  +=       0.353 * cos(2.07182576415  + 1109.13473460990 *T)
    sumX2  +=       0.259 * cos(4.03007203572  + 20426.81490990550 *T)
    sumX2  +=       0.259 * cos(5.19867290786  + 1059.62574767270 *T)
    sumX2  +=       0.198 * cos(1.36980571393  + 27043.74670066630 *T)
    sumX2  +=       0.178 * cos(5.96725527175  + 31749.47900820990 *T)
    sumX2  +=       0.161 * cos(1.00641443984  + 1589.31671276730 *T)
    sumX2  +=       0.148 * cos(2.97453779319  + 4552.19731454230 *T)
    sumX2  +=       0.145 * cos(1.59448029290  + 21536.19346199890 *T)
    sumX2  +=       0.158 * cos(1.60582346371  + 156527.17503196167 *T)
    sumX2  +=       0.169 * cos(4.00782299083  + 25132.54721744910 *T)
    sumX2  +=       0.134 * cos(5.93157203334  + 25558.45599396310 *T)
    sumX2  +=       0.112 * cos(0.40462785686  + 286967.17837479961 *T)
    sumX2  +=       0.111 * cos(5.65704047760  + 26617.83792415230 *T)
    sumX2  +=       0.122 * cos(1.49577379421  + 46514.71805147970 *T)
    sumX2  +=       0.125 * cos(1.39636100360  + 23869.38985487090 *T)
    sumX2  +=       0.111 * cos(2.28972892610  + 14765.48286075331 *T)
    sumX2  +=       0.106 * cos(4.77037689879  + 51116.66817044270 *T)
    sumX2  +=       0.120 * cos(4.20635563605  + 103242.47783168688 *T)
    sumX2  +=       0.113 * cos(2.57148364836  + 27147.52888924690 *T)
    sumX2  +=       0.105 * cos(4.19895688280  + 28306.90406324449 *T)
    sumX2  +=       0.086 * cos(5.07571410391  + 19317.43635781211 *T)
    sumX2  +=       0.103 * cos(5.90310312996  + 5661.08823166870 *T)
    sumX2  +=       0.067 * cos(6.26109568890  + 9104.15081160110 *T)
    sumX2  +=       0.076 * cos(1.89488008069  + 529.93478257810 *T)
    sumX2  +=       0.066 * cos(4.44479405797  + 50586.97720534809 *T)
    sumX2  +=       0.066 * cos(3.84543572024  + 3442.81876244891 *T)
    sumX2  +=       0.061 * cos(2.83683010519  + 37410.81105736210 *T)
    sumX2  +=       0.059 * cos(4.74002403670  + 72602.62119305388 *T)
    sumX2  +=       0.074 * cos(1.71392170541  + 22645.57201409229 *T)
    sumX2  +=       0.056 * cos(2.44778877932  + 52705.74106572650 *T)
    sumX2  +=       0.061 * cos(3.55810846013  + 30640.10045611650 *T)
    sumX2  +=       0.053 * cos(5.29509726913  + 640.14110379750 *T)
    sumX2  +=       0.056 * cos(0.44039665586  + 32858.85756030330 *T)
    sumX2  +=       0.062 * cos(0.55305609443  + 51220.45035902330 *T)
    sumX2  +=       0.050 * cos(4.18169929034  + 53131.64984224050 *T)
    sumX2  +=       0.046 * cos(3.94824823145  + 27677.21985434150 *T)
    sumX2  +=       0.045 * cos(1.59652029747  + 25973.70767037247 *T)
    sumX2  +=       0.045 * cos(4.05512833234  + 26202.58624774291 *T)
    sumX2  +=       0.044 * cos(4.75905220926  + 182615.07817353585 *T)
    sumX2  +=       0.046 * cos(2.53699777188  + 27197.03787618409 *T)
    sumX2  +=       0.046 * cos(5.78816990516  + 53235.43203082109 *T)
    sumX2  +=       0.058 * cos(2.08536631307  + 57837.38214978410 *T)
    sumX2  +=       0.043 * cos(3.15018018958  + 529.44714761109 *T)
    sumX2  +=       0.042 * cos(1.10095861133  + 47624.09660357310 *T)
    sumX2  +=       0.041 * cos(2.92440512287  + 26301.44605449570 *T)
    sumX2  +=       0.039 * cos(5.52297144186  + 40853.38600232750 *T)
    sumX2  +=       0.038 * cos(2.59002993610  + 25874.84786361970 *T)
    sumX2  +=       0.045 * cos(3.87761344478  + 29530.72190402310 *T)
    sumX2  +=       0.045 * cos(4.83071642182  + 10213.52936369450 *T)
    sumX2  +=       0.035 * cos(1.77114967266  + 2119.00767786191 *T)
    sumX2  +=       0.034 * cos(0.50537233993  + 56728.00359769070 *T)
    sumX2  +=       0.040 * cos(4.99855315823  + 24978.28077199729 *T)
    sumX2  +=       0.040 * cos(0.98737142456  + 129330.38097326108 *T)
    sumX2  +=       0.033 * cos(3.45515093268  + 313055.08151637384 *T)
    sumX2  +=       0.039 * cos(5.42516034342  + 955.35592412510 *T)
    sumX2  +=       0.031 * cos(4.38256920252  + 6770.46678376209 *T)
    sumX2  +=       0.038 * cos(1.73621782583  + 2218.51328670329 *T)
    sumX2  +=       0.031 * cos(1.18390533268  + 12566.39551746630 *T)
    sumX2  +=       0.032 * cos(4.46582914668  + 49957.29299644510 *T)
    sumX2  +=       0.029 * cos(5.65827133315  + 426.84200835950 *T)
    sumX2  +=       0.028 * cos(5.46107883558  + 43072.14310651430 *T)
    sumX2  +=       0.025 * cos(1.58123453631  + 98690.52433462808 *T)
    sumX2  +=       0.030 * cos(3.15402941421  + 79373.33179429949 *T)
    sumX2  +=       0.023 * cos(5.44646488235  + 78793.64420730069 *T)
    sumX2  +=       0.030 * cos(0.97936081436  + 77204.57131201689 *T)
    sumX2  +=       0.020 * cos(1.34579383399  + 11322.42028082089 *T)
    sumX2  +=       0.027 * cos(0.95498900102  + 54394.80720481869 *T)
    sumX2  +=       0.023 * cos(3.49479332748  + 77308.35350059748 *T)
    sumX2  +=       0.019 * cos(5.90875785254  + 83925.28529135829 *T)
    sumX2  +=       0.022 * cos(1.88106368245  + 45405.33949938630 *T)
    sumX2  +=       0.020 * cos(2.33312866023  + 36301.43250526870 *T)
    sumX2  +=       0.018 * cos(2.63764225914  + 79323.33517239528 *T)
    sumX2  +=       0.021 * cos(2.63713620085  + 13656.10430865991 *T)
    sumX2  +=       0.020 * cos(4.83742385337  + 33326.82255065770 *T)
    sumX2  +=       0.018 * cos(1.65998530501  + 52156.38033973211 *T)
    sumX2  +=       0.023 * cos(4.13597342584  + 41962.76455442090 *T)
    sumX2  +=       0.018 * cos(6.07003180743  + 52389.34919606990 *T)
    sumX2  +=       0.017 * cos(3.47243675916  + 58946.76070187749 *T)
    sumX2  +=       0.016 * cos(1.64800690653  + 23754.95056618569 *T)
    sumX2  +=       0.019 * cos(2.36847359761  + 39609.89840064910 *T)
    sumX2  +=       0.017 * cos(0.25513432917  + 51536.15281431789 *T)
    sumX2  +=       0.018 * cos(0.34702974356  + 18849.47136745770 *T)
    sumX2  +=       0.015 * cos(0.95866304490  + 76674.88034692229 *T)
    sumX2  +=       0.015 * cos(1.75514605179  + 48733.47515566649 *T)
    sumX2  +=       0.015 * cos(4.34995902728  + 853.44019923550 *T)
    sumX2  +=       0.014 * cos(3.61473044000  + 82815.90673926489 *T)
    sumX2  +=       0.014 * cos(5.02500731242  + 52179.93143598990 *T)
    sumX2  +=       0.014 * cos(1.69891784853  + 15874.86141284670 *T)
    sumX2  +=       0.014 * cos(2.36729572184  + 66941.28914390170 *T)
    sumX2  +=       0.015 * cos(5.49473150242  + 16984.23996494010 *T)
    sumX2  +=       0.015 * cos(5.05704278229  + 38654.29865904050 *T)
    sumX2  +=       0.013 * cos(0.64687702919  + 53765.12299591570 *T)
    sumX2  +=       0.013 * cos(1.62167132693  + 208702.98131511008 *T)
    sumX2  +=       0.013 * cos(3.82178316362  + 50057.28624025350 *T)
    sumX2  +=       0.016 * cos(2.05228361219  + 51646.35913553730 *T)
    sumX2  +=       0.012 * cos(3.97912977913  + 28421.34335192970 *T)
    sumX2  +=       0.013 * cos(4.04493324307  + 155418.28411483529 *T)
    sumX2  +=       0.016 * cos(0.65996554595  + 55618.62504559730 *T)
    sumX2  +=       0.012 * cos(3.62569930460  + 5661.57586663570 *T)
    sumX2  +=       0.015 * cos(0.75515577986  + 52195.71986153169 *T)
    sumX2  +=       0.014 * cos(2.81046707561  + 26514.74514993370 *T)
    sumX2  +=       0.014 * cos(3.30095837282  + 110013.18843293248 *T)
    sumX2  +=       0.013 * cos(0.56442607871  + 213.05527795450 *T)
    sumX2  +=       0.014 * cos(2.10561918260  + 26728.04424537170 *T)
    sumX2  +=       0.013 * cos(3.66951257967  + 52172.16876527390 *T)
    sumX2  +=       0.011 * cos(5.46295418561  + 63498.71419893629 *T)
    sumX2  +=       0.011 * cos(0.61230507574  + 79219.55298381469 *T)
    sumX2  +=       0.010 * cos(4.66712610746  + 103292.47445359109 *T)
    sumX2  +=       0.011 * cos(5.20556952105  + 31748.99137324289 *T)
    sumX2  +=       0.014 * cos(3.33453776277  + 6283.31966747490 *T)
    sumX2  +=       0.010 * cos(0.83705988710  + 52290.48938931711 *T)
    sumX2  +=       0.012 * cos(2.60449391926  + 13521.50762410789 *T)
    sumX2  +=       0.009 * cos(3.03188275218  + 35192.05395317530 *T)
    sumX2  +=       0.009 * cos(1.24305000805  + 76045.19613801930 *T)
    sumX2  +=       0.011 * cos(5.89042281055  + 62389.33564684289 *T)
    sumX2  +=       0.010 * cos(4.16418363493  + 51749.45190975589 *T)
    sumX2  +=       0.009 * cos(4.66321261687  + 124778.42747620228 *T)
    sumX2  +=       0.009 * cos(0.22808750229  + 339142.98465794802 *T)
    sumX2  +=       0.009 * cos(4.75513146544  + 78244.28348130629 *T)
    sumX2  +=       0.009 * cos(2.64262688750  + 20426.32727493849 *T)
    sumX2  +=       0.011 * cos(2.92721365751  + 38520.18960945550 *T)
    sumX2  +=       0.010 * cos(3.30244477369  + 16983.75232997309 *T)
    sumX2  +=       0.009 * cos(1.16009282528  + 1109.62236957690 *T)
    sumX2  +=       0.009 * cos(4.65609022876  + 52061.61081194667 *T)
    sumX2  +=       0.009 * cos(2.15470374859  + 104881.54734887488 *T)
    sumX2  +=       0.008 * cos(1.58129311338  + 68050.66769599509 *T)
    sumX2  +=       0.008 * cos(4.01456932062  + 2333.44021035551 *T)
    sumX2  +=       0.008 * cos(0.50473636519  + 12431.79883291429 *T)
    sumX2  +=       0.008 * cos(4.71148146957  + 28206.91081943610 *T)
    sumX2  +=       0.008 * cos(0.22428318331  + 136101.09157450669 *T)
    sumX2  +=       0.008 * cos(5.13855904986  + 213.54291292150 *T)
    sumX2  +=       0.008 * cos(1.53446312549  + 24499.07406377390 *T)
    sumX2  +=       0.008 * cos(0.19632874417  + 103396.25664217168 *T)
    sumX2  +=       0.007 * cos(4.83084149979  + 52168.93655363109 *T)
    sumX2  +=       0.009 * cos(0.68229010552  + 21535.70582703189 *T)
    sumX2  +=       0.007 * cos(5.67305147777  + 53284.94101775829 *T)
    sumX2  +=       0.007 * cos(4.91493686906  + 105461.23493587368 *T)
    sumX2  +=       0.007 * cos(2.13257749573  + 69160.04624808849 *T)
    sumX2  +=       0.009 * cos(0.17767669866  + 7994.77225950771 *T)
    sumX2  +=       0.009 * cos(3.44284310428  + 73711.99974514729 *T)
    sumX2  +=       0.008 * cos(1.40334354171  + 59414.72569223190 *T)
    sumX2  +=       0.008 * cos(3.92015227538  + 52183.16364763270 *T)
    sumX2  +=       0.007 * cos(1.95913527567  + 64742.20180061470 *T)
    sumX2  +=       0.007 * cos(1.83770637593  + 78267.83457756409 *T)
    sumX2  +=       0.007 * cos(5.50637575808  + 26107.81671995749 *T)
    sumX2  +=       0.007 * cos(3.29290677599  + 26068.47719815791 *T)
    sumX2  +=       0.006 * cos(2.48089512320  + 25028.76502886850 *T)
    sumX2  +=       0.007 * cos(6.10010912620  + 51962.75100519390 *T)
    sumX2  +=       0.007 * cos(5.20852269117  + 18208.05780571871 *T)
    sumX2  +=       0.008 * cos(3.20036201627  + 44937.37450903190 *T)
    sumX2  +=       0.008 * cos(1.81720177562  + 51066.18391357149 *T)
    sumX2  +=       0.006 * cos(5.72654755255  + 105411.23831396949 *T)
    sumX2  +=       0.007 * cos(4.86085983191  + 65697.80154222329 *T)
    sumX2  +=       0.006 * cos(2.84309765149  + 88477.23878841709 *T)
    sumX2  +=       0.006 * cos(2.90025697766  + 78477.25233764409 *T)
    sumX2  +=       0.006 * cos(6.10721529063  + 52602.64829150790 *T)
    sumX2  +=       0.007 * cos(3.76564310534  + 78283.62300310588 *T)
    sumX2  +=       0.006 * cos(2.54939515992  + 19805.07110906630 *T)
    sumX2  +=       0.005 * cos(0.40790961084  + 108903.80988083908 *T)
    sumX2  +=       0.006 * cos(2.08187740275  + 129380.37759516528 *T)
    sumX2  +=       0.005 * cos(4.74517813371  + 74821.37829724069 *T)
    sumX2  +=       0.007 * cos(5.47163902652  + 45892.97425064050 *T)
    sumX2  +=       0.005 * cos(3.96385649366  + 1059.13811270570 *T)
    sumX2  +=       0.005 * cos(2.11024903764  + 26084.26562369970 *T)
    sumX2  +=       0.005 * cos(0.40769424239  + 26092.02829441570 *T)
    sumX2  +=       0.007 * cos(4.95181465375  + 71493.24264096050 *T)
    sumX2  +=       0.005 * cos(3.72498845601  + 81706.52818717150 *T)
    sumX2  +=       0.005 * cos(1.28192740378  + 51322.85371887989 *T)
    sumX2  +=       0.005 * cos(0.38902168011  + 33968.23611239669 *T)
    sumX2  +=       0.005 * cos(1.36259031496  + 78256.83969520529 *T)
    sumX2  +=       0.005 * cos(2.94315074391  + 26617.35028918529 *T)
    sumX2  +=       0.006 * cos(0.40666986327  + 78260.07190684808 *T)
    sumX2  +=       0.004 * cos(3.59234736829  + 25661.54876818170 *T)
    sumX2  +=       0.005 * cos(5.20687247513  + 94138.57083756928 *T)
    sumX2  +=       0.004 * cos(5.73436372945  + 22645.08437912529 *T)
    sumX2  +=       0.005 * cos(5.43911125012  + 93029.19228547587 *T)
    sumX2  +=       0.004 * cos(1.51966869369  + 22760.01130277751 *T)
    sumX2  +=       0.004 * cos(4.11367481672  + 29416.28261533789 *T)
    sumX2  +=       0.004 * cos(0.33419603419  + 76145.18938182769 *T)
    sumX2  +=       0.004 * cos(0.81789657863  + 181506.18725640947 *T)
    sumX2  +=       0.004 * cos(5.87548494754  + 13521.99525907490 *T)
    sumX2  +=       0.004 * cos(0.98530676575  + 99799.90288672148 *T)
    sumX2  +=       0.004 * cos(3.64991710996  + 79853.02613748988 *T)
    sumX2  +=       0.005 * cos(5.51168134551  + 44181.52165860769 *T)
    sumX2  +=       0.004 * cos(4.17969593700  + 4551.70967957530 *T)
    sumX2  +=       0.004 * cos(1.07013306418  + 32371.22280904910 *T)
    sumX2  +=       0.004 * cos(4.56317466224  + 32858.36992533629 *T)
    sumX2  +=       0.004 * cos(0.76064444657  + 54509.24649350390 *T)
    sumX2  +=       0.004 * cos(5.34376558738  + 155468.28073673949 *T)
    sumX2  +=       0.003 * cos(3.33102397097  + 162188.99471608087 *T)
    sumX2  +=       0.003 * cos(1.90784099869  + 24498.58642880689 *T)
    sumX2  +=       0.005 * cos(0.70628118340  + 78271.06678920689 *T)
    sumX2  +=       0.004 * cos(4.75154050928  + 234790.88445668429 *T)
    sumX2  +=       0.003 * cos(5.20128667447  + 77734.26227711148 *T)
    sumX2  +=       0.004 * cos(4.16379409199  + 7879.84533585549 *T)
    sumX2  +=       0.004 * cos(5.89383677298  + 64608.09275102969 *T)
    sumX2  +=       0.004 * cos(4.51512332677  + 51116.18053547569 *T)
    sumX2  +=       0.003 * cos(1.82486201671  + 28306.41642827749 *T)
    sumX2  +=       0.004 * cos(5.71282679191  + 39744.00745023410 *T)
    sumX2  +=       0.003 * cos(1.53805667122  + 104332.18662288050 *T)
    sumX2  +=       0.004 * cos(3.89140851351  + 80482.71034639288 *T)
    sumX2  +=       0.003 * cos(1.85783050998  + 37410.32342239509 *T)
    sumX2  +=       0.004 * cos(1.55416627058  + 24864.32911827909 *T)
    sumX2  +=       0.003 * cos(2.39011669525  + 103822.16541868568 *T)
    sumX2  +=       0.004 * cos(4.09022608915  + 27311.96479983631 *T)
    sumX2  +=       0.003 * cos(2.12318372488  + 120226.47397914348 *T)
    sumX2  +=       0.003 * cos(4.30318589016  + 102133.09927959349 *T)
    sumX2  +=       0.003 * cos(1.44971105205  + 150866.33061777649 *T)
    sumX2  +=       0.003 * cos(5.04167948013  + 90830.10494218887 *T)
    sumX2  +=       0.003 * cos(3.27662875658  + 77624.05595589208 *T)
    sumX2  +=       0.003 * cos(3.21324303917  + 129484.15978374588 *T)
    sumX2  +=       0.003 * cos(4.58517957507  + 125887.80602829569 *T)
    sumX2  +=       0.003 * cos(5.16830743449  + 130969.45049044908 *T)
    sumX2  +=       0.003 * cos(4.95343110449  + 58459.12595062330 *T)
    sumX2  +=       0.002 * cos(0.76903109719  + 6282.83203250789 *T)
    sumX2  +=       0.003 * cos(1.18091352454  + 26941.34334080970 *T)
    sumX2  +=       0.002 * cos(1.64968376160  + 3327.89183879669 *T)
    sumX2  +=       0.003 * cos(2.92429834126  + 18093.13088206649 *T)
    sumX2  +=       0.002 * cos(1.24139720028  + 25132.05958248210 *T)
    sumX2  +=       0.003 * cos(0.05117365042  + 26095.26050605850 *T)
    sumX2  +=       0.003 * cos(6.12547181968  + 71025.27765060609 *T)
    sumX2  +=       0.003 * cos(3.14802602217  + 43071.65547154729 *T)
    sumX2  +=       0.002 * cos(5.96655122637  + 114565.14192999128 *T)
    sumX2  +=       0.002 * cos(2.80186013990  + 71980.87739221469 *T)
    sumX2  +=       0.002 * cos(2.32214550421  + 91785.70468379748 *T)
    sumX2  +=       0.002 * cos(2.48797331613  + 2648.69864295650 *T)
    sumX2  +=       0.002 * cos(3.89856574340  + 78378.39253089130 *T)
    sumX2  +=       0.003 * cos(4.90084849261  + 104355.73771913828 *T)
    sumX2  +=       0.002 * cos(2.45748414015  + 26081.03341205689 *T)
    sumX2  +=       0.002 * cos(2.90684371055  + 77197.45776501608 *T)
    sumX2  +=       0.002 * cos(3.27787980454  + 365230.88779952226 *T)
    sumX2  +=       0.002 * cos(2.96609297657  + 78690.55143308209 *T)
    sumX2  +=       0.003 * cos(5.20911364004  + 52815.94738694590 *T)
    sumX2  +=       0.002 * cos(3.65607552967  + 25448.24967274370 *T)
    sumX2  +=       0.003 * cos(0.46560289429  + 77837.35505133009 *T)
    sumX2  +=       0.002 * cos(0.51377942249  + 104371.52614468009 *T)
    sumX2  +=       0.002 * cos(1.04438989611  + 25028.27739390149 *T)
    sumX2  +=       0.002 * cos(2.95995089299  + 11322.90791578790 *T)
    sumX2  +=       0.002 * cos(2.29166292494  + 77211.68485901768 *T)
    sumX2  +=       0.002 * cos(3.65747965646  + 105307.45612538888 *T)
    sumX2  +=       0.002 * cos(3.59068071614  + 78050.65414676809 *T)
    sumX2  +=       0.002 * cos(0.09837196170  + 27043.25906569930 *T)
    sumX2  +=       0.002 * cos(4.33096627724  + 104344.74283677948 *T)
    sumX2  +=       0.002 * cos(5.54465700187  + 46514.23041651269 *T)
    sumX2  +=       0.002 * cos(4.94013298829  + 77154.08705514569 *T)
    sumX2  +=       0.002 * cos(3.18110797568  + 25451.15068703350 *T)
    sumX2  +=       0.002 * cos(2.48251705553  + 26725.14323108190 *T)
    sumX2  +=       0.002 * cos(2.52179042190  + 131499.14145554369 *T)
    sumX2  +=       0.002 * cos(6.07056602539  + 61279.95709474950 *T)
    sumX2  +=       0.002 * cos(1.27106707156  + 89586.61734051049 *T)
    sumX2  +=       0.002 * cos(1.28866377354  + 30639.61282114949 *T)
    sumX2  +=       0.002 * cos(1.72917997527  + 97581.14578253467 *T)
    sumX2  +=       0.002 * cos(1.43340926720  + 78149.51395352086 *T)
    sumX2  +=       0.002 * cos(4.01850243495  + 7238.43177411650 *T)
    sumX2  +=       0.002 * cos(5.54249111609  + 426.35437339250 *T)
    sumX2  +=       0.002 * cos(0.49218153901  + 107794.43132874568 *T)
    sumX2  +=       0.002 * cos(0.34191520237  + 47623.60896860609 *T)
    sumX2  +=       0.002 * cos(5.99315014021  + 104565.15547921829 *T)
    sumX2  +=       0.002 * cos(1.99101623607  + 44295.96094729290 *T)
    sumX2  +=       0.002 * cos(5.47567807736  + 18093.61851703350 *T)
    sumX2  +=       0.002 * cos(5.22941762292  + 10213.04172872750 *T)
    sumX2  +=       0.002 * cos(3.76589759239  + 104358.96993078108 *T)
    sumX2  +=       0.002 * cos(3.44277374163  + 104347.97504842227 *T)
    sumX2  +=       0.002 * cos(3.46735728702  + 134991.71302241329 *T)
    sumX2  +=       0.002 * cos(5.38147823295  + 129910.06856025988 *T)
    sumX2  +=       0.002 * cos(1.91253822385  + 57836.89451481709 *T)
    sumX2  +=       0.002 * cos(3.81440405932  + 25557.96835899609 *T)
    sumX2  +=       0.002 * cos(1.41016954122  + 54294.81396101029 *T)
    sumX2  +=       0.002 * cos(3.48149032535  + 102762.78348849648 *T)
    sumX2  +=       0.001 * cos(1.23200675306  + 7238.91940908350 *T)
    sumX2  +=       0.002 * cos(4.28600144274  + 25934.36814857290 *T)
    sumX2  +=       0.001 * cos(1.40803647408  + 151975.70916986989 *T)
    sumX2  +=       0.002 * cos(4.52450590998  + 26241.92576954250 *T)
    sumX2  +=       0.001 * cos(1.50678411527  + 131549.13807744789 *T)
    sumX2  +=       0.001 * cos(1.06085484841  + 23968.89546371229 *T)
    sumX2  +=       0.002 * cos(2.95379110730  + 25021.65148186770 *T)
    sumX2  +=       0.002 * cos(5.72041106584  + 27154.64243624770 *T)
    sumX2  +=       0.001 * cos(2.16358753423  + 181556.18387831369 *T)
    sumX2  +=       0.001 * cos(0.24053830744  + 27140.41534224609 *T)
    sumX2  +=       0.001 * cos(1.76435464195  + 149.80701461810 *T)
    sumX2  +=       0.001 * cos(3.46637889336  + 50586.48957038109 *T)
    sumX2  +=       0.001 * cos(4.03245836538  + 85502.62883380610 *T)
    sumX2  +=       0.001 * cos(4.69456883146  + 1052.51220067191 *T)
    sumX2  +=       0.001 * cos(1.69300525500  + 84547.02909219750 *T)
    sumX2  +=       0.001 * cos(3.42418885871  + 41962.27691945389 *T)
    sumX2  +=       0.002 * cos(1.10596292069  + 1066.73929467350 *T)
    sumX2  +=       0.001 * cos(4.90928501927  + 33326.33491569069 *T)
    sumX2  +=       0.001 * cos(5.93281924072  + 103285.36090659029 *T)
    sumX2  +=       0.001 * cos(0.91877297183  + 55504.18575691209 *T)
    sumX2  +=       0.001 * cos(6.04736307912  + 117873.60782537168 *T)
    sumX2  +=       0.001 * cos(2.53958781081  + 25035.87857586930 *T)
    sumX2  +=       0.001 * cos(4.09517185988  + 58946.27306691049 *T)
    sumX2  +=       0.001 * cos(0.11578152669  + 188276.89785765507 *T)
    sumX2  +=       0.001 * cos(1.73884599121  + 49842.85370775989 *T)
    sumX2  +=       0.001 * cos(4.77533799776  + 29428.75938575750 *T)
    sumX2  +=       0.001 * cos(5.22199381468  + 146314.37712071769 *T)
    sumX2  +=       0.001 * cos(2.12230022464  + 119117.09542705010 *T)
    sumX2  +=       0.001 * cos(2.86720953628  + 39629.56816154889 *T)
    sumX2  +=       0.001 * cos(2.74763511093  + 12546.72575656651 *T)
    sumX2  +=       0.001 * cos(0.25887140846  + 64607.60511606269 *T)
    sumX2  +=       0.001 * cos(5.22258225356  + 23868.90221990390 *T)
    sumX2  +=       0.001 * cos(4.58859350492  + 48847.91444435170 *T)
    sumX2  +=       0.001 * cos(1.51099098110  + 260878.78759825832 *T)
    sumX2  +=       0.001 * cos(3.51708181436  + 25620.18196870330 *T)
    sumX2  +=       0.001 * cos(0.18832551491  + 98068.78053378889 *T)
    sumX2  +=       0.001 * cos(0.48744281934  + 22747.53453235790 *T)
    sumX2  +=       0.001 * cos(5.29369767440  + 26556.11194941211 *T)
    sumX2  +=       0.001 * cos(5.89974555271  + 85034.66384345168 *T)
    sumX2  +=       0.001 * cos(5.40405552306  + 5327.71992586630 *T)
    sumX2  +=       0.001 * cos(3.99805833912  + 63498.22656396929 *T)
    sumX2  +=       0.001 * cos(0.48047428009  + 26300.95841952869 *T)
    sumX2  +=       0.001 * cos(0.36034529603  + 105940.92927906408 *T)
    sumX2  +=       0.001 * cos(2.48936428540  + 65831.91059180829 *T)
    sumX2  +=       0.001 * cos(2.46490999297  + 79372.84415933248 *T)
    sumX2  +=       0.001 * cos(3.19576682217  + 7.35736448430 *T)
    sumX2  +=       0.001 * cos(0.86957384492  + 53399.86794141051 *T)
    sumX2  +=       0.001 * cos(3.82636106300  + 80597.14963507809 *T)
    sumX2  +=       0.001 * cos(3.10426657434  + 49527.59527515889 *T)
    sumX2  +=       0.001 * cos(0.58848037656  + 36300.94487030169 *T)
    sumX2  +=       0.001 * cos(1.70956045888  + 116918.00808376308 *T)
    sumX2  +=       0.001 * cos(4.85073507517  + 26080.54577708989 *T)
    sumX2  +=       0.001 * cos(3.44422946655  + 60056.13925397089 *T)
    sumX2  +=       0.001 * cos(4.81916140005  + 10022.08109758290 *T)
    sumX2  +=       0.001 * cos(5.11488921531  + 51123.78171744350 *T)
    sumX2  +=       0.001 * cos(0.47588667079  + 51645.87150057029 *T)
    sumX2  +=       0.001 * cos(1.41817736905  + 76.50988875911 *T)
    sumX2  +=       0.001 * cos(0.99594891808  + 106570.61348796709 *T)
    sumX2  +=       0.001 * cos(0.28396033479  + 25021.16384690069 *T)
    sumX2  +=       0.001 * cos(1.51089793343  + 15874.37377787969 *T)
    sumX2  +=       0.001 * cos(0.30377336401  + 86144.04239554508 *T)
    sumX2  +=       0.001 * cos(3.39681294015  + 52644.01509098630 *T)
    sumX2  +=       0.001 * cos(5.97495759877  + 26094.77551795771 *T)
    sumX2  +=       0.001 * cos(2.83507136702  + 26081.51840015769 *T)
    sumX2  +=       0.001 * cos(5.89040270497  + 6.86972951729 *T)
    sumX2  +=       0.001 * cos(0.29019267678  + 16703.30595098250 *T)
    sumX2  +=       0.001 * cos(5.32811862019  + 35472.98796713290 *T)
    sumX2  +=       0.001 * cos(0.32321972304  + 42154.21282053250 *T)
    sumX2  +=       0.001 * cos(4.39149562944  + 77410.75686045409 *T)
    sumX2  +=       0.001 * cos(3.51650331906  + 103925.25819290428 *T)
    sumX2  +=       0.001 * cos(2.00607418956  + 52329.82891111670 *T)
    sumX2  +=       0.001 * cos(3.10569882865  + 6681.46867088311 *T)
    sumX2  +=       0.001 * cos(0.56719469804  + 6770.95441872910 *T)
    sumX2  +=       0.001 * cos(4.66201946613  + 955.84355909210 *T)
    sumX2  +=       0.001 * cos(0.75784126065  + 51109.55462344189 *T)
    sumX2  +=       0.001 * cos(1.25573296054  + 29530.23426905610 *T)
    sumX2  +=       0.001 * cos(1.77919762464  + 26067.98956319090 *T)
    sumX2  +=       0.001 * cos(4.03762558727  + 25447.76203777669 *T)
    sumX2  +=       0.001 * cos(5.95598082912  + 52705.25343075949 *T)
    sumX2  +=       0.001 * cos(1.01747250713  + 33967.74847742969 *T)
    sumX2  +=       0.001 * cos(0.01664299877  + 25035.39094090229 *T)
    sumX2  +=       0.001 * cos(6.20560095721  + 26107.32908499049 *T)
    sumX2  +=       0.001 * cos(0.93578815320  + 23969.38309867930 *T)
    sumX2  +=       0.001 * cos(2.52616943216  + 19316.94872284510 *T)
    sumX2  +=       0.001 * cos(5.09896744500  + 26091.54065944870 *T)
    sumX2  +=       0.001 * cos(2.39439324238  + 53131.16220727349 *T)
    sumX2  +=       0.001 * cos(2.74115190658  + 95247.94938966268 *T)
    sumX2  +=       0.001 * cos(5.36396008776  + 103299.58800059187 *T)
    sumX2  +=       0.001 * cos(4.59387367112  + 26094.77287109149 *T)
    sumX2  +=       0.001 * cos(3.12740178962  + 102233.09252340188 *T)
    sumX2  +=       0.001 * cos(2.15814286033  + 25668.17468021549 *T)
    sumX2  +=       0.001 * cos(3.29290475293  + 12432.28646788130 *T)
    sumX2  +=       0.001 * cos(4.11678654325  + 115674.52048208468 *T)
    sumX2  +=       0.001 * cos(0.81921278001  + 25661.06113321469 *T)
    sumX2  +=       0.001 * cos(1.94160256445  + 39609.41076568209 *T)
    sumX2  +=       0.001 * cos(2.85651249975  + 26083.77798873270 *T)

    # MERCURY X Coord
    sumX3  = 0
    sumX3  +=     180.660 * cos(1.53524873089  + 0.24381748350 *T)
    sumX3  +=      56.109 * cos(1.50368825372  + 52176.05010063190 *T)
    sumX3  +=      23.909 * cos(5.09285389011  + 78263.95324220609 *T)
    sumX3  +=      11.317 * cos(2.22514154728  + 104351.85638378028 *T)
    sumX3  +=      11.202 * cos(6.20552374893  + 26088.14695905770 *T)
    sumX3  +=       7.321 * cos(1.49034347852  + 26087.65932409069 *T)
    sumX3  +=       5.011 * cos(5.45015928644  + 130439.75952535449 *T)
    sumX3  +=       2.025 * cos(2.30601224164  + 156527.66266692869 *T)
    sumX3  +=       1.152 * cos(0.12743555397  + 52175.56246566489 *T)
    sumX3  +=       0.763 * cos(5.40668920792  + 182615.56580850287 *T)
    sumX3  +=       0.387 * cos(3.30772457389  + 78263.46560723908 *T)
    sumX3  +=       0.274 * cos(2.20509049018  + 208703.46895007708 *T)
    sumX3  +=       0.112 * cos(0.07801111545  + 104351.36874881329 *T)
    sumX3  +=       0.095 * cos(5.27592058522  + 234791.37209165128 *T)
    sumX3  +=       0.037 * cos(4.50900105290  + 24978.76840696430 *T)
    sumX3  +=       0.034 * cos(4.28000177876  + 27197.52551115110 *T)
    sumX3  +=       0.032 * cos(2.05675254224  + 260879.27523322543 *T)
    sumX3  +=       0.024 * cos(2.45956768858  + 51066.67154853850 *T)
    sumX3  +=       0.032 * cos(3.14716159333  + 130439.27189038748 *T)
    sumX3  +=       0.013 * cos(5.72061024737  + 77154.57469011268 *T)
    sumX3  +=       0.010 * cos(5.11454206540  + 286967.17837479961 *T)
    sumX3  +=       0.013 * cos(0.11908599274  + 53285.42865272530 *T)
    sumX3  +=       0.010 * cos(6.24570451816  + 156527.17503196167 *T)
    sumX3  +=       0.006 * cos(4.82972887929  + 1109.13473460990 *T)
    sumX3  +=       0.007 * cos(2.99175820133  + 1059.62574767270 *T)
    sumX3  +=       0.006 * cos(1.93767779760  + 20426.81490990550 *T)
    sumX3  +=       0.006 * cos(2.56259446748  + 103242.47783168688 *T)
    sumX3  +=       0.004 * cos(0.57540517960  + 31749.47900820990 *T)
    sumX3  +=       0.004 * cos(2.89306042580  + 79373.33179429949 *T)
    sumX3  +=       0.003 * cos(1.89223297598  + 313055.08151637384 *T)
    sumX3  +=       0.003 * cos(0.79688696210  + 4552.19731454230 *T)
    sumX3  +=       0.003 * cos(3.08934162665  + 182615.07817353585 *T)
    sumX3  +=       0.003 * cos(5.67790423166  + 46514.71805147970 *T)
    sumX3  +=       0.003 * cos(1.82353994272  + 21536.19346199890 *T)
    sumX3  +=       0.002 * cos(0.08756125776  + 27147.52888924690 *T)
    sumX3  +=       0.002 * cos(3.96079675679  + 1589.31671276730 *T)
    sumX3  +=       0.002 * cos(0.44816937719  + 25132.54721744910 *T)
    sumX3  +=       0.002 * cos(5.80684121758  + 14765.48286075331 *T)
    sumX3  +=       0.002 * cos(5.81670079373  + 51116.66817044270 *T)
    sumX3  +=       0.002 * cos(5.64942685187  + 129330.38097326108 *T)
    sumX3  +=       0.002 * cos(4.47727329739  + 27197.03787618409 *T)
    sumX3  +=       0.002 * cos(2.76458797335  + 26617.83792415230 *T)
    sumX3  +=       0.002 * cos(2.01320187052  + 27043.74670066630 *T)
    sumX3  +=       0.002 * cos(6.24606226113  + 25558.45599396310 *T)
    sumX3  +=       0.002 * cos(2.88488990350  + 72602.62119305388 *T)
    sumX3  +=       0.002 * cos(1.80402690775  + 26068.47719815791 *T)
    sumX3  +=       0.002 * cos(0.72272570426  + 26107.81671995749 *T)
    sumX3  +=       0.001 * cos(5.15880419190  + 26092.02829441570 *T)
    sumX3  +=       0.001 * cos(3.65122349702  + 26084.26562369970 *T)
    sumX3  +=       0.001 * cos(3.47204379646  + 24978.28077199729 *T)
    sumX3  +=       0.001 * cos(3.04599592472  + 37410.81105736210 *T)
    sumX3  +=       0.001 * cos(0.31369621903  + 30640.10045611650 *T)
    sumX3  +=       0.001 * cos(0.38979376906  + 52705.74106572650 *T)
    sumX3  +=       0.001 * cos(5.01792294793  + 51220.45035902330 *T)
    sumX3  +=       0.001 * cos(1.77400727778  + 47624.09660357310 *T)
    sumX3  +=       0.001 * cos(6.21364236354  + 208702.98131511008 *T)
    sumX3  +=       0.001 * cos(3.26982842147  + 9104.15081160110 *T)
    sumX3  +=       0.001 * cos(6.10085930771  + 98690.52433462808 *T)
    sumX3  +=       0.001 * cos(2.72881790021  + 955.35592412510 *T)
    sumX3  +=       0.001 * cos(2.74013903592  + 1109.62236957690 *T)
    sumX3  +=       0.001 * cos(2.00927169251  + 19317.43635781211 *T)
    sumX3  +=       0.001 * cos(4.85871407247  + 50586.97720534809 *T)
    sumX3  +=       0.001 * cos(5.37897907969  + 12566.39551746630 *T)
    sumX3  +=       0.001 * cos(3.51249866668  + 40853.38600232750 *T)
    sumX3  +=       0.001 * cos(3.74067924304  + 53235.43203082109 *T)
    sumX3  +=       0.001 * cos(4.93930525226  + 339142.98465794802 *T)
    sumX3  +=       0.001 * cos(3.64879489376  + 426.84200835950 *T)
    sumX3  +=       0.001 * cos(2.70663061614  + 49957.29299644510 *T)
    sumX3  +=       0.001 * cos(4.05940415323  + 26095.26050605850 *T)
    sumX3  +=       0.001 * cos(4.79523771458  + 26081.03341205689 *T)
    sumX3  +=       0.001 * cos(4.56727257443  + 23869.38985487090 *T)
    sumX3  +=       0.001 * cos(1.91316161387  + 77308.35350059748 *T)
    sumX3  +=       0.001 * cos(5.81582715174  + 105461.23493587368 *T)
    sumX3  +=       0.001 * cos(1.67826436610  + 51646.35913553730 *T)
    sumX3  +=       0.001 * cos(2.42681597375  + 155418.28411483529 *T)
    sumX3  +=       0.001 * cos(4.76066610745  + 56728.00359769070 *T)
    sumX3  +=       0.001 * cos(3.51433188972  + 53131.64984224050 *T)
    sumX3  +=       0.001 * cos(0.78156168692  + 3442.81876244891 *T)
    sumX3  +=       0.001 * cos(3.65095652022  + 78793.64420730069 *T)
    sumX3  +=       0.001 * cos(6.11514198482  + 54394.80720481869 *T)
    sumX3  +=       0.001 * cos(0.70996446104  + 79323.33517239528 *T)
    sumX3  +=       0.001 * cos(4.23955129941  + 28306.90406324449 *T)
    sumX3  +=       0.001 * cos(0.04814737416  + 45405.33949938630 *T)
    sumX3  +=       0.001 * cos(0.51322200896  + 32858.85756030330 *T)
    sumX3  +=       0.001 * cos(1.65918694650  + 25028.76502886850 *T)
    sumX3  +=       0.001 * cos(1.76282246473  + 22645.57201409229 *T)
    sumX3  +=       0.001 * cos(1.59769152646  + 27677.21985434150 *T)

    # MERCURY X Coord
    sumX4  = 0
    sumX4  +=      43.303 * cos(2.70854317703  + 26088.14695905770 *T)
    sumX4  +=      16.746 * cos(2.85109602051  + 0.24381748350 *T)
    sumX4  +=       5.097 * cos(5.82035608585  + 52176.05010063190 *T)
    sumX4  +=       1.110 * cos(2.85416500039  + 78263.95324220609 *T)
    sumX4  +=       0.968 * cos(3.18837996543  + 26087.65932409069 *T)
    sumX4  +=       0.362 * cos(0.10660634045  + 104351.85638378028 *T)
    sumX4  +=       0.161 * cos(3.56341896847  + 130439.75952535449 *T)
    sumX4  +=       0.075 * cos(0.55603774289  + 156527.66266692869 *T)
    sumX4  +=       0.076 * cos(6.07252064875  + 52175.56246566489 *T)
    sumX4  +=       0.033 * cos(3.72826875982  + 182615.56580850287 *T)
    sumX4  +=       0.014 * cos(2.40716156772  + 78263.46560723908 *T)
    sumX4  +=       0.014 * cos(0.56399976721  + 208703.46895007708 *T)
    sumX4  +=       0.005 * cos(3.66056233199  + 234791.37209165128 *T)
    sumX4  +=       0.004 * cos(5.13174489459  + 104351.36874881329 *T)
    sumX4  +=       0.002 * cos(2.87490875788  + 24978.76840696430 *T)
    sumX4  +=       0.002 * cos(5.84989349950  + 27197.52551115110 *T)
    sumX4  +=       0.002 * cos(0.45146054812  + 260879.27523322543 *T)
    sumX4  +=       0.001 * cos(1.74376603369  + 130439.27189038748 *T)
    sumX4  +=       0.001 * cos(3.51635178690  + 286967.17837479961 *T)
    sumX4  +=       0.001 * cos(0.06797049294  + 51066.67154853850 *T)
    sumX4  +=       0.001 * cos(1.33488273309  + 1059.62574767270 *T)
    sumX4  +=       0.001 * cos(0.19743580633  + 20426.81490990550 *T)

    # MERCURY X Coord
    sumX5  = 0
    sumX5  +=       0.414 * cos(4.09017660105  + 0.24381748350 *T)
    sumX5  +=       0.327 * cos(2.83894329980  + 26088.14695905770 *T)
    sumX5  +=       0.134 * cos(4.51536199764  + 52176.05010063190 *T)
    sumX5  +=       0.046 * cos(1.23267980717  + 78263.95324220609 *T)
    sumX5  +=       0.016 * cos(4.45794773259  + 104351.85638378028 *T)
    sumX5  +=       0.007 * cos(5.57199695746  + 26087.65932409069 *T)
    sumX5  +=       0.006 * cos(1.55726840151  + 130439.75952535449 *T)
    sumX5  +=       0.003 * cos(4.94192470876  + 156527.66266692869 *T)
    sumX5  +=       0.001 * cos(1.93929522999  + 182615.56580850287 *T)
    sumX5  +=       0.001 * cos(4.16271885947  + 52175.56246566489 *T)
    sumX5  +=       0.001 * cos(5.10948519211  + 208703.46895007708 *T)

    # MERCURY Y Coord
    sumY0  = 0
    sumY0  +=  37749277.893 * cos(2.83179506899  + 26088.14695905770 *T)
    sumY0  +=  11918926.148 * cos(2.91948125760  + 0.24381748350 *T)
    sumY0  +=  3840153.904 * cos(5.88254544140  + 52176.05010063190 *T)
    sumY0  +=  585979.278 * cos(2.65010770289  + 78263.95324220609 *T)
    sumY0  +=  305833.424 * cos(3.67378306016  + 26087.65932409069 *T)
    sumY0  +=  105974.941 * cos(5.70085415459  + 104351.85638378028 *T)
    sumY0  +=   24906.132 * cos(0.55066770933  + 52175.56246566489 *T)
    sumY0  +=   21056.065 * cos(2.46841472315  + 130439.75952535449 *T)
    sumY0  +=    4441.671 * cos(5.51916065495  + 156527.66266692869 *T)
    sumY0  +=    3347.442 * cos(3.67952842081  + 78263.46560723908 *T)
    sumY0  +=     976.617 * cos(2.28672075339  + 182615.56580850287 *T)
    sumY0  +=     703.903 * cos(2.88960689088  + 24978.76840696430 *T)
    sumY0  +=     640.528 * cos(5.91381352466  + 27197.52551115110 *T)
    sumY0  +=     558.149 * cos(0.50510861725  + 104351.36874881329 *T)
    sumY0  +=     447.775 * cos(1.40511442556  + 1059.62574767270 *T)
    sumY0  +=     400.281 * cos(0.29865270517  + 20426.81490990550 *T)
    sumY0  +=     283.882 * cos(2.22271091529  + 31749.47900820990 *T)
    sumY0  +=     191.007 * cos(2.70458121832  + 53285.42865272530 *T)
    sumY0  +=     181.454 * cos(0.23779467119  + 1109.13473460990 *T)
    sumY0  +=     193.895 * cos(5.38872313263  + 4552.19731454230 *T)
    sumY0  +=     221.371 * cos(5.33746575111  + 208703.46895007708 *T)
    sumY0  +=     139.001 * cos(5.94193853313  + 51066.67154853850 *T)
    sumY0  +=     183.933 * cos(2.76579938720  + 5661.08823166870 *T)
    sumY0  +=     153.303 * cos(1.11372235868  + 51116.66817044270 *T)
    sumY0  +=     104.855 * cos(3.60039284031  + 130439.27189038748 *T)
    sumY0  +=     115.202 * cos(5.27930045344  + 57837.38214978410 *T)
    sumY0  +=      90.628 * cos(4.97031050861  + 529.93478257810 *T)
    sumY0  +=      75.318 * cos(4.50136079719  + 27147.52888924690 *T)
    sumY0  +=      72.957 * cos(3.43332933363  + 21536.19346199890 *T)
    sumY0  +=      74.803 * cos(3.35834807468  + 46514.71805147970 *T)
    sumY0  +=      68.710 * cos(4.91618713593  + 25132.54721744910 *T)
    sumY0  +=      83.504 * cos(1.60859812856  + 10213.52936369450 *T)
    sumY0  +=      60.985 * cos(3.88048699951  + 27043.74670066630 *T)
    sumY0  +=      57.987 * cos(4.13400021557  + 14765.48286075331 *T)
    sumY0  +=      60.953 * cos(3.43164694662  + 47624.09660357310 *T)
    sumY0  +=      53.356 * cos(0.72676419246  + 26617.83792415230 *T)
    sumY0  +=      54.964 * cos(3.74064645856  + 12566.39551746630 *T)
    sumY0  +=      47.881 * cos(5.75865881516  + 79373.33179429949 *T)
    sumY0  +=      51.263 * cos(1.96665333979  + 426.84200835950 *T)
    sumY0  +=      51.361 * cos(2.10502555376  + 234791.37209165128 *T)
    sumY0  +=      50.374 * cos(2.03207321104  + 25558.45599396310 *T)
    sumY0  +=      48.449 * cos(1.94914263411  + 1589.31671276730 *T)
    sumY0  +=      48.867 * cos(4.16469048525  + 77204.57131201689 *T)
    sumY0  +=      34.739 * cos(4.11430856351  + 51646.35913553730 *T)
    sumY0  +=      32.851 * cos(4.56963272977  + 955.35592412510 *T)
    sumY0  +=      44.998 * cos(0.91167682267  + 41962.76455442090 *T)
    sumY0  +=      31.618 * cos(4.66914248946  + 37410.81105736210 *T)
    sumY0  +=      28.634 * cos(5.42748718604  + 30640.10045611650 *T)
    sumY0  +=      31.480 * cos(2.70995810641  + 77154.57469011268 *T)
    sumY0  +=      32.102 * cos(2.04798101755  + 83925.28529135829 *T)
    sumY0  +=      27.670 * cos(5.06128264443  + 39609.89840064910 *T)
    sumY0  +=      26.872 * cos(4.68921320015  + 36301.43250526870 *T)
    sumY0  +=      23.054 * cos(0.71973276708  + 53131.64984224050 *T)
    sumY0  +=      26.193 * cos(5.26902812193  + 11322.42028082089 *T)
    sumY0  +=      21.235 * cos(0.40309333526  + 156527.17503196167 *T)
    sumY0  +=      26.946 * cos(0.19172960930  + 6283.31966747490 *T)
    sumY0  +=      20.875 * cos(4.09807664196  + 15874.86141284670 *T)
    sumY0  +=      16.454 * cos(0.12816102587  + 72602.62119305388 *T)
    sumY0  +=      16.101 * cos(1.65540440480  + 9104.15081160110 *T)
    sumY0  +=      22.241 * cos(5.60722857321  + 13521.50762410789 *T)
    sumY0  +=      21.188 * cos(0.19996828748  + 73711.99974514729 *T)
    sumY0  +=      15.123 * cos(1.24465016017  + 18849.47136745770 *T)
    sumY0  +=      17.255 * cos(0.24107386270  + 27197.03787618409 *T)
    sumY0  +=      17.287 * cos(0.55372068563  + 51749.45190975589 *T)
    sumY0  +=      13.862 * cos(1.29323907537  + 53235.43203082109 *T)
    sumY0  +=      15.239 * cos(1.96671567039  + 7238.43177411650 *T)
    sumY0  +=      16.863 * cos(0.57148727289  + 50586.97720534809 *T)
    sumY0  +=      13.546 * cos(3.64906926506  + 52705.74106572650 *T)
    sumY0  +=      13.134 * cos(4.21391538202  + 25028.76502886850 *T)
    sumY0  +=      14.145 * cos(1.46555737893  + 63498.71419893629 *T)
    sumY0  +=      13.081 * cos(0.20810129185  + 26068.47719815791 *T)
    sumY0  +=      13.058 * cos(2.31384723416  + 26107.81671995749 *T)
    sumY0  +=      13.609 * cos(1.68775421835  + 51220.45035902330 *T)
    sumY0  +=      13.539 * cos(0.93214332140  + 103292.47445359109 *T)
    sumY0  +=      12.138 * cos(5.15577059992  + 260879.27523322543 *T)
    sumY0  +=      11.433 * cos(0.41025670835  + 19317.43635781211 *T)
    sumY0  +=      10.174 * cos(0.82322453273  + 77734.26227711148 *T)
    sumY0  +=      12.016 * cos(3.96439926693  + 68050.66769599509 *T)
    sumY0  +=      12.094 * cos(2.52834929119  + 105461.23493587368 *T)
    sumY0  +=       9.215 * cos(0.91145814574  + 40853.38600232750 *T)
    sumY0  +=       9.157 * cos(1.25590715075  + 33326.82255065770 *T)
    sumY0  +=       9.552 * cos(0.53954398597  + 38654.29865904050 *T)
    sumY0  +=      12.500 * cos(2.32164805011  + 45892.97425064050 *T)
    sumY0  +=       8.978 * cos(5.18856241237  + 26084.26562369970 *T)
    sumY0  +=       8.975 * cos(3.61663563272  + 26092.02829441570 *T)
    sumY0  +=      12.160 * cos(1.83585138946  + 65697.80154222329 *T)
    sumY0  +=       8.880 * cos(5.05045852302  + 26514.74514993370 *T)
    sumY0  +=       8.892 * cos(5.46009752557  + 3442.81876244891 *T)
    sumY0  +=       8.503 * cos(5.09915668367  + 110013.18843293248 *T)
    sumY0  +=      10.244 * cos(4.07406090643  + 21535.70582703189 *T)
    sumY0  +=       8.816 * cos(3.29850433474  + 32371.22280904910 *T)
    sumY0  +=       9.705 * cos(0.73577020420  + 529.44714761109 *T)
    sumY0  +=       7.663 * cos(2.16404811803  + 56728.00359769070 *T)
    sumY0  +=       7.852 * cos(5.36793581356  + 5661.57586663570 *T)
    sumY0  +=       8.166 * cos(5.04025521360  + 27677.21985434150 *T)
    sumY0  +=       7.299 * cos(2.72074453751  + 6770.46678376209 *T)
    sumY0  +=       7.437 * cos(2.10814922209  + 32858.85756030330 *T)
    sumY0  +=       7.796 * cos(2.76073375309  + 31748.99137324289 *T)
    sumY0  +=       6.786 * cos(0.86898827574  + 43072.14310651430 *T)
    sumY0  +=       7.408 * cos(2.97710007179  + 23869.38985487090 *T)
    sumY0  +=       5.938 * cos(3.77513720036  + 79219.55298381469 *T)
    sumY0  +=       7.553 * cos(5.76102373104  + 103242.47783168688 *T)
    sumY0  +=       5.988 * cos(1.47338249847  + 62389.33564684289 *T)
    sumY0  +=       6.154 * cos(4.42442839598  + 59414.72569223190 *T)
    sumY0  +=       6.213 * cos(5.46384627852  + 19805.07110906630 *T)
    sumY0  +=       6.507 * cos(5.77583424512  + 213.54291292150 *T)
    sumY0  +=       6.284 * cos(3.34908254208  + 22645.57201409229 *T)
    sumY0  +=       6.068 * cos(3.25197843754  + 99799.90288672148 *T)
    sumY0  +=       6.276 * cos(1.17668731862  + 640.14110379750 *T)
    sumY0  +=       6.192 * cos(5.82504003546  + 28306.90406324449 *T)
    sumY0  +=       6.081 * cos(3.66198807267  + 25874.84786361970 *T)
    sumY0  +=       4.693 * cos(1.48074264670  + 16983.75232997309 *T)
    sumY0  +=       4.527 * cos(3.48220057078  + 182615.07817353585 *T)
    sumY0  +=       5.085 * cos(5.05051010501  + 26301.44605449570 *T)
    sumY0  +=       4.658 * cos(2.54088213723  + 26095.26050605850 *T)
    sumY0  +=       5.323 * cos(3.72103089814  + 24978.28077199729 *T)
    sumY0  +=       4.497 * cos(0.02524195316  + 26081.03341205689 *T)
    sumY0  +=       5.479 * cos(3.60707059696  + 76674.88034692229 *T)
    sumY0  +=       3.898 * cos(3.18081892914  + 98690.52433462808 *T)
    sumY0  +=       5.352 * cos(3.60477090908  + 77837.35505133009 *T)
    sumY0  +=       3.694 * cos(5.20175219531  + 58946.76070187749 *T)
    sumY0  +=       3.631 * cos(3.98277710501  + 129380.37759516528 *T)
    sumY0  +=       4.821 * cos(5.22008496092  + 37410.32342239509 *T)
    sumY0  +=       3.323 * cos(0.38355389097  + 78793.64420730069 *T)
    sumY0  +=       3.932 * cos(5.47485976543  + 29530.72190402310 *T)
    sumY0  +=       4.022 * cos(2.49293913730  + 2119.00767786191 *T)
    sumY0  +=       3.208 * cos(1.18605815284  + 25557.96835899609 *T)
    sumY0  +=       3.611 * cos(4.88816178501  + 91785.70468379748 *T)
    sumY0  +=       3.062 * cos(4.35403926224  + 79323.33517239528 *T)
    sumY0  +=       3.480 * cos(5.37504374301  + 71980.87739221469 *T)
    sumY0  +=       3.093 * cos(0.73248333106  + 94138.57083756928 *T)
    sumY0  +=       3.172 * cos(4.57251658041  + 27043.25906569930 *T)
    sumY0  +=       2.912 * cos(1.92332998999  + 286967.17837479961 *T)
    sumY0  +=       3.746 * cos(4.51336113078  + 89586.61734051049 *T)
    sumY0  +=       2.870 * cos(3.94317758843  + 69160.04624808849 *T)
    sumY0  +=       2.730 * cos(4.31583707063  + 44937.37450903190 *T)
    sumY0  +=       2.725 * cos(3.85377970873  + 103822.16541868568 *T)
    sumY0  +=       3.214 * cos(0.16060042614  + 2218.51328670329 *T)
    sumY0  +=       2.660 * cos(3.25887021687  + 52156.38033973211 *T)
    sumY0  +=       3.257 * cos(1.14198557197  + 25934.36814857290 *T)
    sumY0  +=       2.620 * cos(1.06710403438  + 20426.32727493849 *T)
    sumY0  +=       3.227 * cos(1.38048915969  + 26241.92576954250 *T)
    sumY0  +=       3.086 * cos(4.73969429884  + 77308.35350059748 *T)
    sumY0  +=       3.099 * cos(5.58109262188  + 131549.13807744789 *T)
    sumY0  +=       2.731 * cos(4.90353573524  + 149.80701461810 *T)
    sumY0  +=       2.486 * cos(1.55056807904  + 1052.51220067191 *T)
    sumY0  +=       3.167 * cos(1.96024135338  + 33326.33491569069 *T)
    sumY0  +=       2.680 * cos(5.36534815866  + 52195.71986153169 *T)
    sumY0  +=       2.231 * cos(1.86712683980  + 136101.09157450669 *T)
    sumY0  +=       2.653 * cos(0.11236418844  + 7.35736448430 *T)
    sumY0  +=       2.074 * cos(3.39236434641  + 48733.47515566649 *T)
    sumY0  +=       1.974 * cos(3.59882556210  + 64742.20180061470 *T)
    sumY0  +=       2.064 * cos(1.78573813736  + 51116.18053547569 *T)
    sumY0  +=       2.533 * cos(4.44951960940  + 1066.73929467350 *T)
    sumY0  +=       1.949 * cos(0.11634548370  + 58459.12595062330 *T)
    sumY0  +=       2.010 * cos(0.87459613624  + 16984.23996494010 *T)
    sumY0  +=       1.876 * cos(5.20487934678  + 82815.90673926489 *T)
    sumY0  +=       1.832 * cos(0.38381879945  + 52179.93143598990 *T)
    sumY0  +=       2.306 * cos(2.63023850427  + 54394.80720481869 *T)
    sumY0  +=       1.837 * cos(3.96249421836  + 66941.28914390170 *T)
    sumY0  +=       1.783 * cos(0.37307285570  + 25620.18196870330 *T)
    sumY0  +=       1.826 * cos(1.95614381567  + 52172.16876527390 *T)
    sumY0  +=       1.791 * cos(3.75150085341  + 25661.54876818170 *T)
    sumY0  +=       2.126 * cos(1.34366019415  + 51536.15281431789 *T)
    sumY0  +=       1.714 * cos(2.14968206427  + 26556.11194941211 *T)
    sumY0  +=       1.972 * cos(3.46789993631  + 45405.33949938630 *T)
    sumY0  +=       1.711 * cos(2.68399636879  + 32858.36992533629 *T)
    sumY0  +=       1.643 * cos(1.83203729093  + 52602.64829150790 *T)
    sumY0  +=       1.665 * cos(2.26005496586  + 5327.71992586630 *T)
    sumY0  +=       1.930 * cos(4.21706155195  + 13656.10430865991 *T)
    sumY0  +=       1.655 * cos(0.02018823486  + 125887.80602829569 *T)
    sumY0  +=       1.554 * cos(1.82939701370  + 13521.99525907490 *T)
    sumY0  +=       1.517 * cos(1.82763021291  + 53765.12299591570 *T)
    sumY0  +=       1.870 * cos(2.52871717392  + 129330.38097326108 *T)
    sumY0  +=       1.705 * cos(1.43301700747  + 43071.65547154729 *T)
    sumY0  +=       1.683 * cos(2.75504809483  + 52168.93655363109 *T)
    sumY0  +=       1.449 * cos(4.53368571731  + 88477.23878841709 *T)
    sumY0  +=       1.466 * cos(3.70132103086  + 24499.07406377390 *T)
    sumY0  +=       1.655 * cos(4.55736013213  + 76.50988875911 *T)
    sumY0  +=       1.539 * cos(0.55097320740  + 105307.45612538888 *T)
    sumY0  +=       1.542 * cos(0.39515897687  + 102762.78348849648 *T)
    sumY0  +=       1.400 * cos(5.97424163846  + 26081.51840015769 *T)
    sumY0  +=       1.400 * cos(2.83094248939  + 26094.77551795771 *T)
    sumY0  +=       1.415 * cos(6.03024008680  + 49957.29299644510 *T)
    sumY0  +=       1.545 * cos(1.17380884450  + 85502.62883380610 *T)
    sumY0  +=       1.259 * cos(2.49869507757  + 51962.75100519390 *T)
    sumY0  +=       1.464 * cos(0.37249136511  + 103925.25819290428 *T)
    sumY0  +=       1.265 * cos(4.66607647315  + 35192.05395317530 *T)
    sumY0  +=       1.563 * cos(6.24487877351  + 6681.46867088311 *T)
    sumY0  +=       1.517 * cos(5.71237882590  + 213.05527795450 *T)
    sumY0  +=       1.171 * cos(6.01640449002  + 57836.89451481709 *T)
    sumY0  +=       1.547 * cos(1.86619704850  + 52389.34919606990 *T)
    sumY0  +=       1.423 * cos(0.03012497835  + 50057.28624025350 *T)
    sumY0  +=       1.238 * cos(5.55628734215  + 52183.16364763270 *T)
    sumY0  +=       1.070 * cos(4.25732715878  + 26728.04424537170 *T)
    sumY0  +=       1.020 * cos(5.20453548111  + 12431.79883291429 *T)
    sumY0  +=       0.965 * cos(0.75029404016  + 155468.28073673949 *T)
    sumY0  +=       1.018 * cos(0.65795794031  + 24498.58642880689 *T)
    sumY0  +=       0.994 * cos(1.65652991676  + 117873.60782537168 *T)
    sumY0  +=       1.002 * cos(0.27317864857  + 208702.98131511008 *T)
    sumY0  +=       1.214 * cos(4.15542472749  + 25661.06113321469 *T)
    sumY0  +=       0.958 * cos(6.23240896197  + 124778.42747620228 *T)
    sumY0  +=       0.893 * cos(1.95641235916  + 85034.66384345168 *T)
    sumY0  +=       0.875 * cos(0.15313552239  + 74821.37829724069 *T)
    sumY0  +=       0.871 * cos(1.85017833510  + 1059.13811270570 *T)
    sumY0  +=       0.910 * cos(2.14250029739  + 98068.78053378889 *T)
    sumY0  +=       0.827 * cos(4.01703921795  + 22645.08437912529 *T)
    sumY0  +=       0.829 * cos(3.41829988722  + 104881.54734887488 *T)
    sumY0  +=       0.870 * cos(0.60198611419  + 51066.18391357149 *T)
    sumY0  +=       0.978 * cos(4.58656747537  + 38520.18960945550 *T)
    sumY0  +=       0.915 * cos(2.91341983071  + 6.86972951729 *T)
    sumY0  +=       0.855 * cos(0.96267173742  + 51123.78171744350 *T)
    sumY0  +=       0.795 * cos(3.78227948828  + 120226.47397914348 *T)
    sumY0  +=       0.778 * cos(4.69135126506  + 11322.90791578790 *T)
    sumY0  +=       0.913 * cos(3.89740262984  + 52026.48690349730 *T)
    sumY0  +=       0.930 * cos(1.27443982259  + 115674.52048208468 *T)
    sumY0  +=       0.932 * cos(2.23433073748  + 55618.62504559730 *T)
    sumY0  +=       0.707 * cos(4.97407500954  + 313055.08151637384 *T)
    sumY0  +=       0.803 * cos(2.35027685221  + 157637.04121902207 *T)
    sumY0  +=       0.730 * cos(3.77427452793  + 53284.94101775829 *T)
    sumY0  +=       0.725 * cos(1.12701074857  + 105411.23831396949 *T)
    sumY0  +=       0.788 * cos(4.40761141586  + 3340.85624418330 *T)
    sumY0  +=       0.867 * cos(4.35296523561  + 51109.55462344189 *T)
    sumY0  +=       0.718 * cos(0.61085091780  + 129910.06856025988 *T)
    sumY0  +=       0.688 * cos(5.33062177858  + 419.72846135871 *T)
    sumY0  +=       0.710 * cos(0.69656198994  + 95247.94938966268 *T)
    sumY0  +=       0.694 * cos(0.69216961450  + 26617.35028918529 *T)
    sumY0  +=       0.621 * cos(2.81767179045  + 4551.70967957530 *T)
    sumY0  +=       0.661 * cos(5.92198588119  + 10213.04172872750 *T)
    sumY0  +=       0.674 * cos(5.58162526148  + 28206.91081943610 *T)
    sumY0  +=       0.609 * cos(0.02645166365  + 78244.28348130629 *T)
    sumY0  +=       0.661 * cos(4.19316030438  + 52022.27129014710 *T)
    sumY0  +=       0.741 * cos(1.50795914520  + 103396.25664217168 *T)
    sumY0  +=       0.585 * cos(4.91842789278  + 162188.99471608087 *T)
    sumY0  +=       0.599 * cos(0.26652159243  + 46848.57399224910 *T)
    sumY0  +=       0.634 * cos(5.63013895989  + 39609.41076568209 *T)
    sumY0  +=       0.731 * cos(1.84936364859  + 433.95555536030 *T)
    sumY0  +=       0.591 * cos(1.08868730454  + 71025.27765060609 *T)
    sumY0  +=       0.576 * cos(4.10239875091  + 47623.60896860609 *T)
    sumY0  +=       0.545 * cos(3.92064766910  + 48732.98752069949 *T)
    sumY0  +=       0.528 * cos(3.25755205747  + 23754.95056618569 *T)
    sumY0  +=       0.531 * cos(1.51927796885  + 41962.27691945389 *T)
    sumY0  +=       0.525 * cos(0.65657950540  + 10022.08109758290 *T)
    sumY0  +=       0.578 * cos(5.67944788678  + 80482.71034639288 *T)
    sumY0  +=       0.675 * cos(4.42923882594  + 52329.82891111670 *T)
    sumY0  +=       0.664 * cos(4.39586206751  + 77624.05595589208 *T)
    sumY0  +=       0.541 * cos(4.50825185806  + 6282.83203250789 *T)
    sumY0  +=       0.639 * cos(2.57562729428  + 45494.82524723229 *T)
    sumY0  +=       0.594 * cos(1.74048603817  + 7994.77225950771 *T)
    sumY0  +=       0.467 * cos(3.18640247451  + 84547.02909219750 *T)
    sumY0  +=       0.552 * cos(4.24369907572  + 52099.78402935630 *T)
    sumY0  +=       0.490 * cos(1.39410440209  + 64608.09275102969 *T)
    sumY0  +=       0.553 * cos(0.82663110400  + 1692.40948698591 *T)
    sumY0  +=       0.456 * cos(0.37337128477  + 90830.10494218887 *T)
    sumY0  +=       0.470 * cos(1.72250117653  + 26237.71015619230 *T)
    sumY0  +=       0.501 * cos(5.75878971797  + 25132.05958248210 *T)
    sumY0  +=       0.615 * cos(2.13316170849  + 78283.62300310588 *T)
    sumY0  +=       0.441 * cos(4.69790158258  + 27140.41534224609 *T)
    sumY0  +=       0.445 * cos(3.07156557764  + 151975.70916986989 *T)
    sumY0  +=       0.437 * cos(0.95616885323  + 20760.18321570790 *T)
    sumY0  +=       0.472 * cos(5.57949973231  + 155418.28411483529 *T)
    sumY0  +=       0.428 * cos(4.96656096346  + 77204.08367704989 *T)
    sumY0  +=       0.469 * cos(1.97026697646  + 108903.80988083908 *T)
    sumY0  +=       0.478 * cos(2.05637713133  + 63498.22656396929 *T)
    sumY0  +=       0.448 * cos(5.75669088007  + 78256.83969520529 *T)
    sumY0  +=       0.591 * cos(0.35948760681  + 853.44019923550 *T)
    sumY0  +=       0.438 * cos(4.57721971918  + 51645.87150057029 *T)
    sumY0  +=       0.440 * cos(0.13387750376  + 28306.41642827749 *T)
    sumY0  +=       0.413 * cos(5.17208909967  + 38519.70197448849 *T)
    sumY0  +=       0.456 * cos(0.44231533666  + 26038.15033715350 *T)
    sumY0  +=       0.456 * cos(2.07975292935  + 26138.14358096190 *T)
    sumY0  +=       0.470 * cos(2.65781109220  + 58946.27306691049 *T)
    sumY0  +=       0.434 * cos(0.74419794380  + 29428.75938575750 *T)
    sumY0  +=       0.439 * cos(5.54617295689  + 2333.44021035551 *T)
    sumY0  +=       0.406 * cos(1.00123517971  + 1582.20316576650 *T)
    sumY0  +=       0.419 * cos(3.46554019433  + 128850.68663007068 *T)
    sumY0  +=       0.403 * cos(5.55657271431  + 28421.34335192970 *T)
    sumY0  +=       0.423 * cos(5.90337303076  + 50586.48957038109 *T)
    sumY0  +=       0.420 * cos(3.43444455433  + 78267.83457756409 *T)
    sumY0  +=       0.508 * cos(5.53374236541  + 26720.93069837089 *T)
    sumY0  +=       0.445 * cos(2.66109199380  + 1109.62236957690 *T)
    sumY0  +=       0.452 * cos(4.25647147725  + 25985.05418483910 *T)
    sumY0  +=       0.440 * cos(5.74038643901  + 78050.65414676809 *T)
    sumY0  +=       0.427 * cos(1.26284550681  + 27154.64243624770 *T)
    sumY0  +=       0.422 * cos(0.73601792678  + 93029.19228547587 *T)
    sumY0  +=       0.387 * cos(1.92885240062  + 22747.53453235790 *T)
    sumY0  +=       0.451 * cos(4.61446843236  + 26191.23973327629 *T)
    sumY0  +=       0.525 * cos(3.37057650618  + 25455.36321974450 *T)
    sumY0  +=       0.439 * cos(4.25286780683  + 46514.23041651269 *T)
    sumY0  +=       0.390 * cos(3.42340819919  + 130013.16133447849 *T)
    sumY0  +=       0.364 * cos(4.88702121021  + 78690.55143308209 *T)
    sumY0  +=       0.468 * cos(3.04146774326  + 76145.18938182769 *T)
    sumY0  +=       0.361 * cos(1.30085103028  + 114565.14192999128 *T)
    sumY0  +=       0.354 * cos(5.19811092004  + 52644.01509098630 *T)
    sumY0  +=       0.418 * cos(5.00690657163  + 78260.07190684808 *T)
    sumY0  +=       0.366 * cos(2.14542297373  + 25028.27739390149 *T)
    sumY0  +=       0.394 * cos(4.92599580786  + 78477.25233764409 *T)
    sumY0  +=       0.337 * cos(4.88605522032  + 79853.02613748988 *T)
    sumY0  +=       0.356 * cos(3.76575150987  + 633.02755679670 *T)
    sumY0  +=       0.427 * cos(5.87777060793  + 19804.58347409930 *T)
    sumY0  +=       0.360 * cos(4.50952146319  + 48835.43767393209 *T)
    sumY0  +=       0.431 * cos(0.05038435146  + 30639.61282114949 *T)
    sumY0  +=       0.336 * cos(0.49943786696  + 18208.05780571871 *T)
    sumY0  +=       0.317 * cos(2.79783111172  + 76045.19613801930 *T)
    sumY0  +=       0.403 * cos(3.61003449042  + 131395.35926696309 *T)
    sumY0  +=       0.300 * cos(3.06549658713  + 25973.70767037247 *T)
    sumY0  +=       0.365 * cos(3.99788805051  + 25565.56954096390 *T)
    sumY0  +=       0.417 * cos(0.23545014867  + 71493.24264096050 *T)
    sumY0  +=       0.298 * cos(5.73633616578  + 26202.58624774291 *T)
    sumY0  +=       0.411 * cos(4.96911720848  + 1596.43025976811 *T)
    sumY0  +=       0.286 * cos(5.88170536088  + 52182.67865953191 *T)
    sumY0  +=       0.304 * cos(3.41185962602  + 51708.08511027749 *T)
    sumY0  +=       0.288 * cos(1.37917320218  + 26164.41303033330 *T)
    sumY0  +=       0.326 * cos(4.77493319595  + 26610.72437715149 *T)
    sumY0  +=       0.375 * cos(2.00750267367  + 26162.92855762500 *T)
    sumY0  +=       0.268 * cos(4.70687038628  + 143961.51096694588 *T)
    sumY0  +=       0.367 * cos(4.20390802587  + 111590.53197538028 *T)
    sumY0  +=       0.264 * cos(3.36238088617  + 72936.47713382330 *T)
    sumY0  +=       0.301 * cos(1.01728716251  + 224.58861318540 *T)
    sumY0  +=       0.286 * cos(0.96880689193  + 46848.08635728210 *T)
    sumY0  +=       0.257 * cos(3.80096482898  + 181556.18387831369 *T)
    sumY0  +=       0.352 * cos(0.48255743874  + 26013.36536049040 *T)
    sumY0  +=       0.274 * cos(5.81956733021  + 956.53297345411 *T)
    sumY0  +=       0.253 * cos(4.93056899504  + 75.02541605080 *T)
    sumY0  +=       0.256 * cos(1.10512464961  + 537.04832957890 *T)
    sumY0  +=       0.245 * cos(3.26928135038  + 51756.56545675670 *T)
    sumY0  +=       0.286 * cos(2.73949837375  + 52169.42154173189 *T)
    sumY0  +=       0.326 * cos(0.00381222715  + 25771.75508940110 *T)
    sumY0  +=       0.252 * cos(5.36725270563  + 25668.66231518250 *T)
    sumY0  +=       0.239 * cos(1.88156493524  + 42154.21282053250 *T)
    sumY0  +=       0.324 * cos(2.52294776653  + 26404.53882871430 *T)
    sumY0  +=       0.285 * cos(6.14793784305  + 1162.71852189130 *T)
    sumY0  +=       0.237 * cos(5.19203538966  + 124156.68367536308 *T)
    sumY0  +=       0.271 * cos(3.04131804133  + 15874.37377787969 *T)
    sumY0  +=       0.307 * cos(2.31415827156  + 78271.06678920689 *T)
    sumY0  +=       0.323 * cos(3.00259946219  + 19406.43447069109 *T)
    sumY0  +=       0.245 * cos(5.02045917872  + 111122.56698502589 *T)
    sumY0  +=       0.229 * cos(5.15166243945  + 36300.94487030169 *T)
    sumY0  +=       0.243 * cos(4.26179828093  + 7238.91940908350 *T)
    sumY0  +=       0.242 * cos(1.18373210037  + 51748.96427478889 *T)
    sumY0  +=       0.275 * cos(4.01647026193  + 77211.68485901768 *T)
    sumY0  +=       0.234 * cos(2.92710924584  + 83924.79765639128 *T)
    sumY0  +=       0.279 * cos(1.12090122634  + 77197.45776501608 *T)
    sumY0  +=       0.260 * cos(2.91177438950  + 32769.37181245730 *T)
    sumY0  +=       0.222 * cos(3.18512792848  + 100909.28143881488 *T)
    sumY0  +=       0.279 * cos(3.09797269121  + 26507.63160293289 *T)
    sumY0  +=       0.299 * cos(3.03086575636  + 2648.69864295650 *T)
    sumY0  +=       0.213 * cos(5.24729841378  + 1479.11039154791 *T)
    sumY0  +=       0.221 * cos(4.03293449116  + 102.84895673509 *T)
    sumY0  +=       0.231 * cos(4.31940320133  + 141762.42362365889 *T)
    sumY0  +=       0.208 * cos(0.53989181963  + 146314.37712071769 *T)
    sumY0  +=       0.236 * cos(4.88594462204  + 14764.99522578630 *T)
    sumY0  +=       0.219 * cos(5.60242546056  + 110.45013870291 *T)
    sumY0  +=       0.241 * cos(3.00055987343  + 150866.33061777649 *T)
    sumY0  +=       0.210 * cos(5.40322983591  + 183724.94436059627 *T)
    sumY0  +=       0.280 * cos(0.66543647361  + 78114.39004507149 *T)
    sumY0  +=       0.211 * cos(0.17674723492  + 130969.45049044908 *T)
    sumY0  +=       0.244 * cos(5.94927988906  + 11610.79577585770 *T)
    sumY0  +=       0.265 * cos(0.98893666490  + 39744.00745023410 *T)
    sumY0  +=       0.200 * cos(2.15844889143  + 16065.82204399129 *T)
    sumY0  +=       0.222 * cos(5.27780086013  + 81706.52818717150 *T)
    sumY0  +=       0.197 * cos(1.03744786641  + 52815.94738694590 *T)
    sumY0  +=       0.197 * cos(4.45583276555  + 12565.90788249930 *T)
    sumY0  +=       0.241 * cos(5.45668008692  + 27147.04125427989 *T)
    sumY0  +=       0.205 * cos(2.01894158165  + 33968.23611239669 *T)
    sumY0  +=       0.248 * cos(5.59907777274  + 71582.72838880648 *T)
    sumY0  +=       0.246 * cos(0.67064557518  + 51742.33836275509 *T)
    sumY0  +=       0.203 * cos(0.77451026070  + 25881.96141062050 *T)
    sumY0  +=       0.202 * cos(4.54831491675  + 25448.24967274370 *T)
    sumY0  +=       0.228 * cos(3.34387624216  + 234790.88445668429 *T)
    sumY0  +=       0.198 * cos(1.71260487848  + 26294.33250749489 *T)
    sumY0  +=       0.231 * cos(0.78401706031  + 44181.52165860769 *T)
    sumY0  +=       0.185 * cos(5.15550823271  + 64607.60511606269 *T)
    sumY0  +=       0.226 * cos(2.64165383084  + 7879.84533585549 *T)
    sumY0  +=       0.176 * cos(3.73240626851  + 77154.08705514569 *T)
    sumY0  +=       0.179 * cos(3.53913956860  + 14477.10736571650 *T)
    sumY0  +=       0.195 * cos(5.08298233604  + 59414.23805726489 *T)
    sumY0  +=       0.207 * cos(4.24194657171  + 1376.01761732930 *T)
    sumY0  +=       0.174 * cos(1.73934628427  + 339142.98465794802 *T)
    sumY0  +=       0.174 * cos(5.68665306442  + 51543.26636131870 *T)
    sumY0  +=       0.184 * cos(4.55688502110  + 129484.15978374588 *T)
    sumY0  +=       0.173 * cos(6.20260804596  + 23968.89546371229 *T)
    sumY0  +=       0.207 * cos(0.17517194401  + 20760.67085067490 *T)
    sumY0  +=       0.166 * cos(0.73482223187  + 79372.84415933248 *T)
    sumY0  +=       0.185 * cos(1.16294236415  + 103711.95909746629 *T)
    sumY0  +=       0.161 * cos(0.99358007086  + 73711.51211018028 *T)
    sumY0  +=       0.194 * cos(1.69156637131  + 50483.88443112950 *T)
    sumY0  +=       0.220 * cos(1.40610479581  + 69159.55861312148 *T)
    sumY0  +=       0.191 * cos(3.65637031484  + 155997.97170183409 *T)
    sumY0  +=       0.216 * cos(4.60747679624  + 53131.16220727349 *T)
    sumY0  +=       0.156 * cos(2.45974681457  + 106570.61348796709 *T)
    sumY0  +=       0.181 * cos(1.41036114643  + 18093.13088206649 *T)
    sumY0  +=       0.152 * cos(0.95845975049  + 78110.17443172129 *T)
    sumY0  +=       0.149 * cos(2.99675052546  + 45892.48661567349 *T)
    sumY0  +=       0.182 * cos(3.74672593763  + 121335.85253123687 *T)
    sumY0  +=       0.155 * cos(1.68611064922  + 188276.89785765507 *T)
    sumY0  +=       0.178 * cos(4.17805278254  + 131499.14145554369 *T)
    sumY0  +=       0.166 * cos(2.18208828557  + 468.20880783791 *T)
    sumY0  +=       0.157 * cos(1.37406612212  + 61279.95709474950 *T)
    sumY0  +=       0.203 * cos(2.15909999017  + 51322.85371887989 *T)
    sumY0  +=       0.190 * cos(0.30919081370  + 25874.36022865269 *T)
    sumY0  +=       0.142 * cos(1.51656559474  + 50594.09075234889 *T)
    sumY0  +=       0.175 * cos(3.62001720200  + 26624.95147115310 *T)
    sumY0  +=       0.148 * cos(3.07498569808  + 104332.18662288050 *T)
    sumY0  +=       0.153 * cos(3.71257893953  + 52101.26850206459 *T)
    sumY0  +=       0.164 * cos(3.90190440252  + 74820.89066227368 *T)
    sumY0  +=       0.158 * cos(1.19350301362  + 78417.73205269089 *T)
    sumY0  +=       0.140 * cos(0.89912686582  + 52279.14287485049 *T)
    sumY0  +=       0.163 * cos(5.63286881634  + 25780.58933808811 *T)
    sumY0  +=       0.160 * cos(3.16772159097  + 26395.70458002730 *T)
    sumY0  +=       0.132 * cos(3.07363434632  + 22760.01130277751 *T)
    sumY0  +=       0.129 * cos(0.10582004023  + 54394.31956985169 *T)
    sumY0  +=       0.150 * cos(5.18178770285  + 104371.52614468009 *T)
    sumY0  +=       0.166 * cos(5.37015346630  + 522.82123557730 *T)
    sumY0  +=       0.149 * cos(2.52020201514  + 23889.05961577070 *T)
    sumY0  +=       0.125 * cos(2.56126838362  + 65697.31390725628 *T)
    sumY0  +=       0.127 * cos(2.35920317858  + 54294.81396101029 *T)
    sumY0  +=       0.136 * cos(6.13995819179  + 102233.09252340188 *T)
    sumY0  +=       0.140 * cos(2.25411762839  + 52808.83383994509 *T)
    sumY0  +=       0.171 * cos(1.01179953646  + 78187.68717093048 *T)
    sumY0  +=       0.119 * cos(6.12499272961  + 178063.61231144407 *T)
    sumY0  +=       0.121 * cos(2.34497789442  + 181506.18725640947 *T)
    sumY0  +=       0.150 * cos(1.37100107530  + 13362.69352428270 *T)
    sumY0  +=       0.160 * cos(0.50398727524  + 955.84355909210 *T)
    sumY0  +=       0.122 * cos(0.97167224102  + 52072.95732641330 *T)
    sumY0  +=       0.139 * cos(4.14538916469  + 97113.18079218028 *T)
    sumY0  +=       0.116 * cos(2.49897944668  + 104344.74283677948 *T)
    sumY0  +=       0.127 * cos(4.93091662093  + 26521.85869693450 *T)
    sumY0  +=       0.113 * cos(0.24728903100  + 154938.58977164488 *T)
    sumY0  +=       0.116 * cos(6.24971983447  + 110634.93223377167 *T)
    sumY0  +=       0.120 * cos(3.16836542455  + 23969.38309867930 *T)
    sumY0  +=       0.117 * cos(0.78253817148  + 18093.61851703350 *T)
    sumY0  +=       0.114 * cos(5.52498204591  + 25754.29101828830 *T)
    sumY0  +=       0.150 * cos(5.32072185109  + 25551.34244696229 *T)
    sumY0  +=       0.109 * cos(3.57166541301  + 40565.01050729069 *T)
    sumY0  +=       0.110 * cos(0.79184574430  + 24176.94747584050 *T)
    sumY0  +=       0.107 * cos(4.83424409375  + 68242.11596210669 *T)
    sumY0  +=       0.112 * cos(3.27382494684  + 26422.00289982710 *T)
    sumY0  +=       0.107 * cos(4.40903077996  + 90695.99589260388 *T)
    sumY0  +=       0.104 * cos(0.18580141979  + 156101.06447605268 *T)
    sumY0  +=       0.111 * cos(3.42810291181  + 116918.00808376308 *T)
    sumY0  +=       0.121 * cos(6.28088254389  + 28287.23430234470 *T)
    sumY0  +=       0.108 * cos(1.36988749165  + 44181.03402364069 *T)
    sumY0  +=       0.102 * cos(2.58046151243  + 26198.35328027710 *T)
    sumY0  +=       0.143 * cos(3.83368380796  + 50579.86365834729 *T)
    sumY0  +=       0.108 * cos(3.14844806883  + 26107.32908499049 *T)
    sumY0  +=       0.103 * cos(5.73149554363  + 29416.28261533789 *T)
    sumY0  +=       0.102 * cos(1.26920320545  + 74923.34081550628 *T)
    sumY0  +=       0.111 * cos(1.57596233643  + 51639.24558853649 *T)
    sumY0  +=       0.100 * cos(3.85312326105  + 25035.87857586930 *T)
    sumY0  +=       0.097 * cos(3.56740191654  + 323.74923414091 *T)
    sumY0  +=       0.100 * cos(1.69705145148  + 104565.15547921829 *T)
    sumY0  +=       0.099 * cos(5.11526101592  + 31415.62306744050 *T)
    sumY0  +=       0.120 * cos(5.01910336702  + 134991.71302241329 *T)
    sumY0  +=       0.131 * cos(1.06382274034  + 647.25465079831 *T)
    sumY0  +=       0.128 * cos(2.55933728310  + 104138.55728834229 *T)
    sumY0  +=       0.110 * cos(4.87488188396  + 24357.02460612510 *T)
    sumY0  +=       0.105 * cos(1.05131979960  + 26067.98956319090 *T)
    sumY0  +=       0.095 * cos(4.10632947324  + 52705.25343075949 *T)
    sumY0  +=       0.105 * cos(3.71934987002  + 55516.66252733170 *T)
    sumY0  +=       0.093 * cos(3.49141182565  + 52126.05347872769 *T)
    sumY0  +=       0.092 * cos(2.31943050291  + 54509.24649350390 *T)
    sumY0  +=       0.096 * cos(1.91265055140  + 743.23387801611 *T)
    sumY0  +=       0.109 * cos(5.90844457054  + 632.53992182970 *T)
    sumY0  +=       0.090 * cos(1.71341920893  + 27999.34644227490 *T)
    sumY0  +=       0.094 * cos(1.83646395516  + 103291.98681862408 *T)
    sumY0  +=       0.111 * cos(2.83140360723  + 40565.49814225770 *T)
    sumY0  +=       0.101 * cos(1.49775137503  + 51951.70530492999 *T)
    sumY0  +=       0.102 * cos(0.19980570400  + 104355.73771913828 *T)
    sumY0  +=       0.088 * cos(2.55079662642  + 1588.82907780029 *T)
    sumY0  +=       0.102 * cos(3.43450786393  + 26941.34334080970 *T)
    sumY0  +=       0.109 * cos(5.12423170717  + 60056.13925397089 *T)
    sumY0  +=       0.093 * cos(3.91123861975  + 27780.31262856009 *T)
    sumY0  +=       0.089 * cos(4.72535332511  + 68050.18006102808 *T)
    sumY0  +=       0.106 * cos(2.15819157914  + 6770.95441872910 *T)
    sumY0  +=       0.101 * cos(5.08955449331  + 52250.83169919920 *T)
    sumY0  +=       0.086 * cos(4.79490272689  + 52325.61329776649 *T)
    sumY0  +=       0.092 * cos(4.35572616043  + 140653.04507156549 *T)
    sumY0  +=       0.107 * cos(0.38434884214  + 157483.26240853727 *T)
    sumY0  +=       0.087 * cos(0.68859108180  + 25938.58376192310 *T)
    sumY0  +=       0.094 * cos(2.98031225005  + 51219.76094466129 *T)
    sumY0  +=       0.089 * cos(3.67089011438  + 36109.98423915710 *T)
    sumY0  +=       0.093 * cos(1.66575202299  + 103.33659170210 *T)
    sumY0  +=       0.098 * cos(3.92942306643  + 27819.26931199030 *T)
    sumY0  +=       0.084 * cos(1.50244162097  + 53228.31848382029 *T)
    sumY0  +=       0.087 * cos(3.14637767899  + 52065.84377941249 *T)
    sumY0  +=       0.099 * cos(3.29163755467  + 97581.14578253467 *T)
    sumY0  +=       0.105 * cos(3.86326312300  + 70269.42480018188 *T)
    sumY0  +=       0.083 * cos(2.64246416261  + 51219.96272405629 *T)
    sumY0  +=       0.101 * cos(5.08528632726  + 52226.04672253610 *T)
    sumY0  +=       0.087 * cos(1.65472743108  + 104778.45457465628 *T)
    sumY0  +=       0.102 * cos(4.59971787247  + 12432.28646788130 *T)
    sumY0  +=       0.082 * cos(1.96514410059  + 78731.91823256049 *T)
    sumY0  +=       0.101 * cos(3.78885420287  + 119117.09542705010 *T)
    sumY0  +=       0.108 * cos(5.77504268406  + 49527.59527515889 *T)
    sumY0  +=       0.092 * cos(3.01639809211  + 51859.65823097529 *T)
    sumY0  +=       0.080 * cos(4.33221518102  + 53242.54557782190 *T)
    sumY0  +=       0.107 * cos(4.90113452404  + 25447.76203777669 *T)
    sumY0  +=       0.088 * cos(1.30079620909  + 12725.20961729149 *T)
    sumY0  +=       0.091 * cos(5.49075472411  + 26300.95841952869 *T)
    sumY0  +=       0.080 * cos(1.65550423430  + 105940.92927906408 *T)
    sumY0  +=       0.076 * cos(5.84908430441  + 102133.09927959349 *T)
    sumY0  +=       0.079 * cos(0.12371508890  + 80482.22271142589 *T)
    sumY0  +=       0.091 * cos(2.66249978065  + 51013.57539622409 *T)
    sumY0  +=       0.094 * cos(1.39522248869  + 2222.10045208050 *T)
    sumY0  +=       0.089 * cos(0.94979154283  + 137678.43511695447 *T)
    sumY0  +=       0.075 * cos(0.07329711774  + 3327.89183879669 *T)
    sumY0  +=       0.089 * cos(2.43057395425  + 1910.95566573370 *T)
    sumY0  +=       0.098 * cos(4.89673673453  + 50056.79860528649 *T)
    sumY0  +=       0.074 * cos(6.13007861498  + 25977.94063783829 *T)
    sumY0  +=       0.089 * cos(5.56505599077  + 52492.44197028850 *T)
    sumY0  +=       0.073 * cos(3.56559697903  + 50697.18352656750 *T)
    sumY0  +=       0.102 * cos(1.77223481993  + 104347.97504842227 *T)
    sumY0  +=       0.072 * cos(1.47771204102  + 170049.41410852008 *T)
    sumY0  +=       0.078 * cos(4.16301616898  + 103285.36090659029 *T)
    sumY0  +=       0.075 * cos(6.26897336450  + 52595.53474450709 *T)
    sumY0  +=       0.074 * cos(1.16312773108  + 25021.65148186770 *T)
    sumY0  +=       0.075 * cos(0.08785923175  + 77844.46859833089 *T)
    sumY0  +=       0.078 * cos(2.64612833915  + 74.53778108379 *T)
    sumY0  +=       0.094 * cos(1.55832345623  + 52698.62751872569 *T)
    sumY0  +=       0.095 * cos(1.28337020564  + 25938.09612695609 *T)
    sumY0  +=       0.087 * cos(1.11118535241  + 72602.13355808688 *T)
    sumY0  +=       0.070 * cos(4.10808217554  + 27670.10630734069 *T)
    sumY0  +=       0.076 * cos(5.35798835757  + 104358.96993078108 *T)
    sumY0  +=       0.074 * cos(3.71626517668  + 25954.03790947270 *T)
    sumY0  +=       0.070 * cos(5.98641613952  + 66653.40128383189 *T)
    sumY0  +=       0.071 * cos(2.40595691806  + 24998.43816786410 *T)
    sumY0  +=       0.073 * cos(5.08664921507  + 26222.25600864270 *T)
    sumY0  +=       0.081 * cos(6.12425664727  + 8194.03151572510 *T)
    sumY0  +=       0.066 * cos(2.64701347180  + 78270.58180110609 *T)
    sumY0  +=       0.072 * cos(2.95285797491  + 426.35437339250 *T)
    sumY0  +=       0.078 * cos(3.71360449621  + 104202.29318664568 *T)
    sumY0  +=       0.066 * cos(5.22374540666  + 89586.12970554348 *T)
    sumY0  +=       0.081 * cos(6.08469300372  + 26080.54577708989 *T)
    sumY0  +=       0.068 * cos(0.56774852198  + 207644.08701988787 *T)
    sumY0  +=       0.065 * cos(4.30145540281  + 12546.72575656651 *T)
    sumY0  +=       0.074 * cos(2.34912240985  + 97670.63153038068 *T)
    sumY0  +=       0.067 * cos(1.25089965871  + 2111.89413086110 *T)
    sumY0  +=       0.079 * cos(3.71647319510  + 77830.24150432929 *T)
    sumY0  +=       0.083 * cos(5.35734577753  + 522.33360061030 *T)
    sumY0  +=       0.079 * cos(2.22871999356  + 25446.73339731871 *T)
    sumY0  +=       0.063 * cos(1.37131699662  + 70268.93716521488 *T)
    sumY0  +=       0.074 * cos(4.43425667540  + 95247.46175469569 *T)
    sumY0  +=       0.085 * cos(1.03066114675  + 72935.98949885629 *T)
    sumY0  +=       0.062 * cos(1.93537775234  + 150244.58681693728 *T)
    sumY0  +=       0.081 * cos(1.53102925511  + 77733.77464214448 *T)
    sumY0  +=       0.077 * cos(5.87942788213  + 316.63568714010 *T)
    sumY0  +=       0.075 * cos(2.61550986340  + 33967.74847742969 *T)
    sumY0  +=       0.071 * cos(1.76514499411  + 27684.33340134230 *T)
    sumY0  +=       0.061 * cos(2.80789864051  + 846.32665223470 *T)
    sumY0  +=       0.064 * cos(0.11131885479  + 27177.85575025130 *T)
    sumY0  +=       0.073 * cos(6.03174712424  + 26083.77798873270 *T)
    sumY0  +=       0.075 * cos(0.28818227298  + 26729.56052079669 *T)
    sumY0  +=       0.071 * cos(3.36506897458  + 17893.87162584910 *T)
    sumY0  +=       0.063 * cos(2.50388232144  + 52290.48938931711 *T)
    sumY0  +=       0.072 * cos(4.56353970752  + 50800.27630078609 *T)
    sumY0  +=       0.059 * cos(6.11443002788  + 52061.61081194667 *T)
    sumY0  +=       0.077 * cos(1.18313398787  + 38813.60039383270 *T)
    sumY0  +=       0.080 * cos(4.87250848454  + 26089.63143176599 *T)
    sumY0  +=       0.069 * cos(3.16809879958  + 56727.51596272369 *T)
    sumY0  +=       0.080 * cos(3.94160294784  + 26086.66248634940 *T)
    sumY0  +=       0.058 * cos(3.98600001818  + 419.24082639170 *T)
    sumY0  +=       0.066 * cos(0.10285600557  + 2008.80135664250 *T)
    sumY0  +=       0.065 * cos(3.63512616728  + 66652.91364886488 *T)
    sumY0  +=       0.058 * cos(0.05018228362  + 99024.38027539747 *T)
    sumY0  +=       0.077 * cos(0.77887041561  + 103299.58800059187 *T)
    sumY0  +=       0.069 * cos(3.03766697714  + 316.14805217309 *T)
    sumY0  +=       0.061 * cos(6.24933126885  + 126997.18458038908 *T)
    sumY0  +=       0.068 * cos(1.78962458422  + 137210.47012660009 *T)
    sumY0  +=       0.058 * cos(1.07398748672  + 167850.32676523307 *T)
    sumY0  +=       0.057 * cos(1.06178740528  + 26011.88088778210 *T)
    sumY0  +=       0.062 * cos(1.20172371901  + 19316.94872284510 *T)
    sumY0  +=       0.073 * cos(4.45582042270  + 26091.54065944870 *T)
    sumY0  +=       0.055 * cos(2.17291421134  + 209812.84750217048 *T)
    sumY0  +=       0.055 * cos(3.55134929549  + 44295.96094729290 *T)
    sumY0  +=       0.058 * cos(6.25406726323  + 25508.45937205890 *T)
    sumY0  +=       0.073 * cos(2.05415256446  + 18848.98373249069 *T)
    sumY0  +=       0.056 * cos(3.84149994570  + 22625.90225319250 *T)
    sumY0  +=       0.064 * cos(5.21615906084  + 77410.75686045409 *T)
    sumY0  +=       0.054 * cos(4.44562432663  + 52252.31617190749 *T)
    sumY0  +=       0.066 * cos(0.17594911232  + 77795.98825185168 *T)
    sumY0  +=       0.055 * cos(2.04378126827  + 107794.43132874568 *T)
    sumY0  +=       0.057 * cos(3.42931686715  + 24449.07744186970 *T)
    sumY0  +=       0.055 * cos(3.59192378031  + 172402.28026229187 *T)
    sumY0  +=       0.067 * cos(2.93531319099  + 19.42594341629 *T)
    sumY0  +=       0.053 * cos(3.70608451935  + 32370.73517408209 *T)
    sumY0  +=       0.052 * cos(0.79212421214  + 3340.36860921629 *T)
    sumY0  +=       0.055 * cos(3.60943961817  + 220.65645992230 *T)
    sumY0  +=       0.072 * cos(3.16918804459  + 24864.32911827909 *T)
    sumY0  +=       0.052 * cos(3.80711022251  + 23868.90221990390 *T)
    sumY0  +=       0.060 * cos(5.14905673779  + 90695.50825763687 *T)
    sumY0  +=       0.052 * cos(4.10578873491  + 26312.49175475960 *T)
    sumY0  +=       0.053 * cos(0.28962121715  + 19406.92210565810 *T)
    sumY0  +=       0.053 * cos(2.56330425325  + 26667.83454605650 *T)
    sumY0  +=       0.070 * cos(0.08146254525  + 2199.33116077051 *T)
    sumY0  +=       0.059 * cos(3.54477372961  + 38.37685312130 *T)
    sumY0  +=       0.051 * cos(5.94337924391  + 58857.27495403150 *T)
    sumY0  +=       0.050 * cos(4.20235040741  + 129799.86223904048 *T)
    sumY0  +=       0.067 * cos(5.79256239224  + 78257.32468330608 *T)
    sumY0  +=       0.050 * cos(5.35090323068  + 27727.21647624570 *T)
    sumY0  +=       0.048 * cos(2.60487489461  + 27044.43611502830 *T)
    sumY0  +=       0.047 * cos(4.10703352371  + 51969.86455219470 *T)
    sumY0  +=       0.055 * cos(3.21272296258  + 157057.35363202327 *T)
    sumY0  +=       0.060 * cos(2.71680609162  + 85034.17620848469 *T)
    sumY0  +=       0.049 * cos(2.29047787617  + 77631.16950289288 *T)
    sumY0  +=       0.064 * cos(4.73468584640  + 76571.78757270369 *T)
    sumY0  +=       0.046 * cos(4.38909300675  + 77836.86741636308 *T)
    sumY0  +=       0.054 * cos(4.21805017615  + 23919.38647677510 *T)
    sumY0  +=       0.047 * cos(1.32745528662  + 155572.06292532009 *T)
    sumY0  +=       0.064 * cos(5.63611607616  + 27311.96479983631 *T)
    sumY0  +=       0.046 * cos(4.51657960486  + 76681.99389392308 *T)
    sumY0  +=       0.046 * cos(0.05397462972  + 29530.23426905610 *T)
    sumY0  +=       0.053 * cos(3.43095934900  + 49842.85370775989 *T)
    sumY0  +=       0.055 * cos(4.95374342010  + 27037.32256802749 *T)
    sumY0  +=       0.054 * cos(2.11552363976  + 949.41942645330 *T)
    sumY0  +=       0.047 * cos(4.95775520700  + 29550.39166492290 *T)
    sumY0  +=       0.056 * cos(2.31169429193  + 53234.94439585409 *T)
    sumY0  +=       0.062 * cos(1.59919946431  + 26011.39325281509 *T)
    sumY0  +=       0.051 * cos(6.10777700397  + 28736.60178453070 *T)
    sumY0  +=       0.045 * cos(2.54247709679  + 55504.18575691209 *T)
    sumY0  +=       0.062 * cos(6.05167319913  + 176954.23375935067 *T)
    sumY0  +=       0.043 * cos(4.09072234194  + 78903.85052852010 *T)
    sumY0  +=       0.044 * cos(2.10728026202  + 16703.30595098250 *T)
    sumY0  +=       0.052 * cos(6.07065313687  + 110012.70079796547 *T)
    sumY0  +=       0.050 * cos(3.89612671834  + 25138.97135008790 *T)
    sumY0  +=       0.043 * cos(2.78443984846  + 1272.92484311070 *T)
    sumY0  +=       0.044 * cos(3.76849382783  + 105460.74730090667 *T)
    sumY0  +=       0.058 * cos(2.82347260176  + 26094.77287109149 *T)
    sumY0  +=       0.042 * cos(0.46194308183  + 1486.22393854870 *T)
    sumY0  +=       0.042 * cos(0.51106392440  + 78189.17164363878 *T)
    sumY0  +=       0.048 * cos(4.03335182320  + 65831.91059180829 *T)
    sumY0  +=       0.054 * cos(4.84278312652  + 52382.23564906909 *T)
    sumY0  +=       0.041 * cos(1.77651394552  + 1795.50226120450 *T)
    sumY0  +=       0.043 * cos(2.60404615127  + 60055.65161900389 *T)
    sumY0  +=       0.046 * cos(4.86931523516  + 54087.24958384910 *T)
    sumY0  +=       0.053 * cos(0.13354671820  + 260878.78759825832 *T)
    sumY0  +=       0.048 * cos(5.22216564191  + 25466.40315821850 *T)
    sumY0  +=       0.048 * cos(3.58305204289  + 26709.89075989690 *T)
    sumY0  +=       0.048 * cos(0.57485699057  + 76667.76679992149 *T)
    sumY0  +=       0.047 * cos(4.05954911572  + 104275.59031250468 *T)
    sumY0  +=       0.039 * cos(5.49383143740  + 132658.51662954129 *T)
    sumY0  +=       0.040 * cos(0.07089879206  + 26824.02347258949 *T)
    sumY0  +=       0.040 * cos(2.44203081578  + 25352.27044552590 *T)
    sumY0  +=       0.043 * cos(4.79238528651  + 365230.88779952226 *T)
    sumY0  +=       0.044 * cos(0.28163075339  + 52712.85461272729 *T)
    sumY0  +=       0.045 * cos(4.58626302576  + 28256.90744134030 *T)
    sumY0  +=       0.039 * cos(0.56315956692  + 103241.99019671988 *T)
    sumY0  +=       0.051 * cos(0.42021554272  + 182085.87484340827 *T)
    sumY0  +=       0.050 * cos(0.52050720367  + 147423.75567281109 *T)
    sumY0  +=       0.043 * cos(4.74188084964  + 214364.80099922928 *T)
    sumY0  +=       0.039 * cos(4.16197164473  + 99799.41525175449 *T)
    sumY0  +=       0.044 * cos(3.72596196534  + 25867.73431661889 *T)
    sumY0  +=       0.043 * cos(0.20439431464  + 1265.81129610991 *T)
    sumY0  +=       0.038 * cos(4.24623905840  + 104505.63519426508 *T)
    sumY0  +=       0.037 * cos(4.01161399115  + 104198.07757329548 *T)
    sumY0  +=       0.037 * cos(2.96320182560  + 128320.99566497609 *T)
    sumY0  +=       0.042 * cos(3.62647853730  + 25170.21667307590 *T)
    sumY0  +=       0.040 * cos(3.65913470233  + 206.42936592071 *T)
    sumY0  +=       0.038 * cos(1.95151248327  + 130459.42928625426 *T)
    sumY0  +=       0.037 * cos(2.36549717602  + 26102.37405305930 *T)
    sumY0  +=       0.050 * cos(1.81815722451  + 153.53499300130 *T)
    sumY0  +=       0.037 * cos(0.10000949167  + 24203.24579564031 *T)
    sumY0  +=       0.041 * cos(5.40449209098  + 34282.42229226630 *T)
    sumY0  +=       0.036 * cos(1.44241697006  + 26118.47382006210 *T)
    sumY0  +=       0.036 * cos(1.07961540756  + 26057.82009805329 *T)
    sumY0  +=       0.036 * cos(6.15630697749  + 34281.93465729930 *T)
    sumY0  +=       0.042 * cos(1.45143122814  + 38653.81102407349 *T)
    sumY0  +=       0.035 * cos(0.16883470965  + 26073.91986505609 *T)
    sumY0  +=       0.039 * cos(5.17224451182  + 27006.07724503950 *T)
    sumY0  +=       0.039 * cos(3.83113919893  + 151.29148732640 *T)
    sumY0  +=       0.037 * cos(3.86078277718  + 78367.04601642469 *T)
    sumY0  +=       0.034 * cos(1.99485322788  + 27567.01353312209 *T)
    sumY0  +=       0.035 * cos(5.28138513945  + 78896.73698151928 *T)
    sumY0  +=       0.037 * cos(2.43731961667  + 25773.96077918809 *T)
    sumY0  +=       0.037 * cos(0.08471276095  + 26402.33313892731 *T)
    sumY0  +=       0.045 * cos(5.51053584744  + 2126.12122486270 *T)
    sumY0  +=       0.037 * cos(0.41442508719  + 9384.59719059169 *T)
    sumY0  +=       0.033 * cos(1.23373917586  + 116783.89903417807 *T)
    sumY0  +=       0.035 * cos(1.05079898841  + 27463.92075890350 *T)
    sumY0  +=       0.033 * cos(5.24254633919  + 51852.54468397449 *T)
    sumY0  +=       0.037 * cos(6.12798409572  + 130420.08976445468 *T)
    sumY0  +=       0.035 * cos(2.14843472729  + 62388.84801187589 *T)
    sumY0  +=       0.033 * cos(0.42458452244  + 26190.10947732330 *T)
    sumY0  +=       0.033 * cos(2.09744785518  + 25986.18444079209 *T)
    sumY0  +=       0.035 * cos(1.60180137355  + 51653.47268253809 *T)
    sumY0  +=       0.032 * cos(4.47872593179  + 39629.56816154889 *T)
    sumY0  +=       0.040 * cos(2.83686756644  + 148.32254190981 *T)
    sumY0  +=       0.043 * cos(0.95235131647  + 157587.04459711787 *T)
    sumY0  +=       0.034 * cos(2.42204384323  + 27973.04812247509 *T)
    sumY0  +=       0.043 * cos(5.25286806261  + 26308.55960149650 *T)
    sumY0  +=       0.040 * cos(2.56508188611  + 205.94173095370 *T)
    sumY0  +=       0.034 * cos(5.11562530130  + 13541.17738500769 *T)
    sumY0  +=       0.031 * cos(5.39821911963  + 207594.09039798367 *T)
    sumY0  +=       0.032 * cos(2.88931723147  + 204151.51545301828 *T)
    sumY0  +=       0.034 * cos(2.40270209936  + 51868.49247966230 *T)
    sumY0  +=       0.043 * cos(1.44438681078  + 51529.03926731709 *T)
    sumY0  +=       0.031 * cos(4.01104942140  + 78160.86046798748 *T)
    sumY0  +=       0.036 * cos(5.64714725231  + 130226.46042991648 *T)
    sumY0  +=       0.030 * cos(5.53920573257  + 130432.64597835368 *T)
    sumY0  +=       0.039 * cos(4.02419894411  + 64901.50353540690 *T)
    sumY0  +=       0.031 * cos(0.59714240312  + 51433.06004009930 *T)
    sumY0  +=       0.041 * cos(1.72851918476  + 40852.89836736050 *T)
    sumY0  +=       0.039 * cos(3.48396647568  + 76674.39271195528 *T)
    sumY0  +=       0.031 * cos(1.66062366332  + 94330.01910368088 *T)
    sumY0  +=       0.030 * cos(3.30916696393  + 181026.49291321909 *T)
    sumY0  +=       0.030 * cos(1.35962398425  + 96356.84030678908 *T)
    sumY0  +=       0.029 * cos(3.24046570458  + 182188.96761762688 *T)
    sumY0  +=       0.034 * cos(5.68539193875  + 111122.07935005888 *T)
    sumY0  +=       0.031 * cos(5.77478962928  + 26514.25751496669 *T)
    sumY0  +=       0.036 * cos(2.47015155723  + 75615.49841673308 *T)
    sumY0  +=       0.028 * cos(2.60390025059  + 86143.55476057807 *T)
    sumY0  +=       0.028 * cos(5.41996849568  + 80382.71710258449 *T)
    sumY0  +=       0.037 * cos(2.73064236635  + 37698.69891743190 *T)
    sumY0  +=       0.030 * cos(5.69224082014  + 77101.47853779829 *T)
    sumY0  +=       0.029 * cos(2.84649684108  + 114.64292439690 *T)
    sumY0  +=       0.030 * cos(2.60710276130  + 26084.21480579459 *T)
    sumY0  +=       0.027 * cos(2.18911511513  + 26241.43813457550 *T)
    sumY0  +=       0.027 * cos(6.24126624865  + 60369.83779887349 *T)
    sumY0  +=       0.027 * cos(4.34927842452  + 101011.24395708049 *T)
    sumY0  +=       0.027 * cos(0.05549126696  + 52509.90604140130 *T)
    sumY0  +=       0.029 * cos(6.20759958805  + 26092.07911232079 *T)
    sumY0  +=       0.033 * cos(1.33379538421  + 42153.72518556549 *T)
    sumY0  +=       0.033 * cos(1.11257666070  + 49954.19346603489 *T)
    sumY0  +=       0.027 * cos(1.97164111012  + 25933.88051360590 *T)
    sumY0  +=       0.027 * cos(5.75010433988  + 91785.21704883048 *T)
    sumY0  +=       0.026 * cos(0.45274965388  + 35472.98796713290 *T)
    sumY0  +=       0.034 * cos(5.96598332393  + 22004.15845235330 *T)
    sumY0  +=       0.026 * cos(6.17560870969  + 71980.38975724769 *T)
    sumY0  +=       0.026 * cos(4.75787796399  + 130653.05862079248 *T)
    sumY0  +=       0.029 * cos(3.02581288540  + 136722.83537534589 *T)
    sumY0  +=       0.035 * cos(6.21411204570  + 52483.60772160150 *T)
    sumY0  +=       0.027 * cos(3.05142430504  + 54375.13744391890 *T)
    sumY0  +=       0.026 * cos(4.35081548550  + 26050.01392341990 *T)
    sumY0  +=       0.033 * cos(4.63692423082  + 77727.14873011068 *T)
    sumY0  +=       0.031 * cos(4.44084625304  + 87367.86023632369 *T)
    sumY0  +=       0.028 * cos(3.43792139137  + 183571.16555011147 *T)
    sumY0  +=       0.034 * cos(4.71651658587  + 1905.70858242391 *T)
    sumY0  +=       0.028 * cos(0.48686633466  + 51226.87449166209 *T)
    sumY0  +=       0.032 * cos(6.03865727615  + 77307.66408623548 *T)
    sumY0  +=       0.033 * cos(4.55124740226  + 78039.60844650418 *T)
    sumY0  +=       0.031 * cos(2.31561412140  + 60370.32543384050 *T)
    sumY0  +=       0.033 * cos(0.91420804533  + 123201.08393375447 *T)
    sumY0  +=       0.027 * cos(5.63137827099  + 1224.06165826211 *T)
    sumY0  +=       0.025 * cos(1.86918416559  + 78338.73484077338 *T)
    sumY0  +=       0.029 * cos(3.85756830106  + 49842.36607279289 *T)
    sumY0  +=       0.025 * cos(0.32014728581  + 76785.08666814168 *T)
    sumY0  +=       0.024 * cos(1.09872540797  + 24824.98959647950 *T)
    sumY0  +=       0.031 * cos(1.78245808102  + 161079.61616398749 *T)
    sumY0  +=       0.031 * cos(0.13179619883  + 75930.75684933408 *T)
    sumY0  +=       0.024 * cos(1.67322759859  + 1045.39865367110 *T)
    sumY0  +=       0.029 * cos(4.01431074745  + 25600.51220780350 *T)
    sumY0  +=       0.029 * cos(4.79090693735  + 26575.78171031190 *T)
    sumY0  +=       0.026 * cos(4.40340514614  + 39450.59666585690 *T)
    sumY0  +=       0.027 * cos(5.45015263034  + 27171.22719135130 *T)
    sumY0  +=       0.027 * cos(3.35506505446  + 25005.06672676410 *T)
    sumY0  +=       0.032 * cos(2.36425003443  + 9103.66317663410 *T)
    sumY0  +=       0.024 * cos(0.92163258785  + 129373.26404816448 *T)
    sumY0  +=       0.023 * cos(1.97682509801  + 85502.14119883909 *T)
    sumY0  +=       0.024 * cos(1.70355204155  + 52609.76183850870 *T)
    sumY0  +=       0.025 * cos(3.42869920852  + 299.37021175271 *T)
    sumY0  +=       0.027 * cos(0.19474964949  + 143005.91122533727 *T)
    sumY0  +=       0.027 * cos(6.01066977576  + 24073.16528725990 *T)
    sumY0  +=       0.025 * cos(3.87710167982  + 25654.43522118089 *T)
    sumY0  +=       0.024 * cos(5.68395366955  + 52286.25642185129 *T)
    sumY0  +=       0.023 * cos(0.49611836531  + 38813.11275886569 *T)
    sumY0  +=       0.026 * cos(5.95483109773  + 27676.73221937449 *T)
    sumY0  +=       0.023 * cos(1.42348474713  + 27351.30432163590 *T)
    sumY0  +=       0.025 * cos(6.12366793793  + 48847.91444435170 *T)
    sumY0  +=       0.028 * cos(2.64147333193  + 24925.67225464990 *T)
    sumY0  +=       0.029 * cos(2.84086630674  + 30172.13546576210 *T)
    sumY0  +=       0.026 * cos(6.21853790794  + 78153.74692098668 *T)
    sumY0  +=       0.023 * cos(1.26116961525  + 50064.39978725430 *T)
    sumY0  +=       0.024 * cos(3.54764052332  + 51962.26337022689 *T)
    sumY0  +=       0.023 * cos(6.03986636871  + 77947.56137254949 *T)
    sumY0  +=       0.023 * cos(0.06002764715  + 123669.04892410888 *T)
    sumY0  +=       0.022 * cos(0.26166099168  + 78213.95662030189 *T)
    sumY0  +=       0.025 * cos(4.36791869246  + 26725.14323108190 *T)
    sumY0  +=       0.025 * cos(4.43729899234  + 25451.15068703350 *T)
    sumY0  +=       0.022 * cos(3.84821265527  + 50264.85061741470 *T)
    sumY0  +=       0.030 * cos(0.70456092924  + 53907.17245356450 *T)
    sumY0  +=       0.030 * cos(3.98527104604  + 23754.46293121869 *T)
    sumY0  +=       0.025 * cos(1.76026444680  + 1.72829019180 *T)
    sumY0  +=       0.025 * cos(4.90185710039  + 1.24065522479 *T)
    sumY0  +=       0.024 * cos(4.62065948726  + 78786.53066029989 *T)
    sumY0  +=       0.022 * cos(1.83132826724  + 6885.39370741431 *T)
    sumY0  +=       0.025 * cos(4.82521606002  + 130435.87818999648 *T)
    sumY0  +=       0.022 * cos(1.64259204266  + 50444.92774769930 *T)
    sumY0  +=       0.028 * cos(4.68306520525  + 6129.54085699010 *T)
    sumY0  +=       0.025 * cos(2.79403215015  + 28103.12863085550 *T)
    sumY0  +=       0.023 * cos(0.48266477387  + 103918.14464590348 *T)
    sumY0  +=       0.025 * cos(0.72627176075  + 52174.56562792359 *T)
    sumY0  +=       0.022 * cos(5.90022439787  + 2219.00092167030 *T)
    sumY0  +=       0.022 * cos(2.31815942379  + 78580.34511186268 *T)
    sumY0  +=       0.029 * cos(4.48338221638  + 1730.87853544909 *T)
    sumY0  +=       0.022 * cos(5.37074911110  + 80597.14963507809 *T)
    sumY0  +=       0.024 * cos(2.29466576857  + 51842.19415986250 *T)
    sumY0  +=       0.020 * cos(5.64534264808  + 51955.63745819309 *T)
    sumY0  +=       0.028 * cos(3.75812407504  + 536.56069461190 *T)
    sumY0  +=       0.020 * cos(5.51240603881  + 23439.20449861769 *T)
    sumY0  +=       0.023 * cos(3.83447858759  + 129387.49114216608 *T)
    sumY0  +=       0.021 * cos(3.97425392925  + 163766.33825852867 *T)
    sumY0  +=       0.025 * cos(0.44166212589  + 81604.56566890588 *T)
    sumY0  +=       0.021 * cos(1.44153243351  + 1063.55790093581 *T)
    sumY0  +=       0.023 * cos(3.39790605489  + 16342.82640320110 *T)
    sumY0  +=       0.022 * cos(0.58776873075  + 96357.32794175608 *T)
    sumY0  +=       0.023 * cos(2.71345954853  + 76144.70174686068 *T)
    sumY0  +=       0.020 * cos(6.01126822986  + 51329.96726588070 *T)
    sumY0  +=       0.020 * cos(0.14753852345  + 1089.95260867711 *T)
    sumY0  +=       0.020 * cos(5.38196984803  + 123758.53467195488 *T)
    sumY0  +=       0.019 * cos(2.29671913556  + 51535.66517935089 *T)
    sumY0  +=       0.022 * cos(5.56163358805  + 3462.00088838170 *T)
    sumY0  +=       0.021 * cos(0.48380924328  + 130290.19632821988 *T)
    sumY0  +=       0.023 * cos(2.36039511683  + 50167.49256147289 *T)
    sumY0  +=       0.020 * cos(4.52844625012  + 196137.31725009429 *T)
    sumY0  +=       0.023 * cos(1.12689921440  + 166740.94821313969 *T)
    sumY0  +=       0.026 * cos(1.32836783140  + 22747.04689739089 *T)
    sumY0  +=       0.022 * cos(4.77036018900  + 26126.27999469550 *T)
    sumY0  +=       0.022 * cos(4.14491043187  + 26735.15779237250 *T)
    sumY0  +=       0.019 * cos(3.09805264388  + 78683.43788608129 *T)
    sumY0  +=       0.020 * cos(3.83026958578  + 75930.26921436709 *T)
    sumY0  +=       0.024 * cos(0.55752006618  + 145204.99856862429 *T)
    sumY0  +=       0.019 * cos(2.12244186746  + 130446.87307235529 *T)
    sumY0  +=       0.024 * cos(1.47596121866  + 52177.53457334019 *T)
    sumY0  +=       0.025 * cos(3.25287024656  + 130443.64086071248 *T)
    sumY0  +=       0.020 * cos(4.21784136757  + 98690.03669966108 *T)
    sumY0  +=       0.019 * cos(1.95226766220  + 103498.66000202828 *T)
    sumY0  +=       0.025 * cos(1.32070214805  + 76888.17944236028 *T)
    sumY0  +=       0.022 * cos(5.54888786998  + 24952.47008716450 *T)
    sumY0  +=       0.022 * cos(3.25632981482  + 27223.82383095090 *T)
    sumY0  +=       0.018 * cos(1.58091743094  + 78413.51643934069 *T)
    sumY0  +=       0.020 * cos(5.01791562590  + 104819.82137413470 *T)
    sumY0  +=       0.022 * cos(4.95873027034  + 129379.88996019828 *T)
    sumY0  +=       0.019 * cos(4.66235459384  + 26248.55432844250 *T)
    sumY0  +=       0.019 * cos(4.14286309096  + 25927.73958967290 *T)
    sumY0  +=       0.019 * cos(0.21911919049  + 53029.24648238390 *T)
    sumY0  +=       0.018 * cos(2.61852920213  + 128221.00242116768 *T)
    sumY0  +=       0.020 * cos(4.36521030279  + 25908.06982877310 *T)
    sumY0  +=       0.020 * cos(4.44000738201  + 26268.22408934230 *T)
    sumY0  +=       0.019 * cos(3.36352476280  + 52817.46366237089 *T)
    sumY0  +=       0.018 * cos(0.30820161935  + 26411.65237571510 *T)
    sumY0  +=       0.022 * cos(3.17947641824  + 103932.37173990508 *T)
    sumY0  +=       0.018 * cos(0.03030733311  + 52195.23222656469 *T)
    sumY0  +=       0.024 * cos(1.85700392205  + 86144.04239554508 *T)
    sumY0  +=       0.020 * cos(5.09555302131  + 12412.61670698150 *T)
    sumY0  +=       0.025 * cos(1.79034244495  + 78313.94986411028 *T)
    sumY0  +=       0.020 * cos(3.61245706996  + 233731.99016146208 *T)
    sumY0  +=       0.020 * cos(5.31801711889  + 25234.95057730570 *T)
    sumY0  +=       0.017 * cos(0.72137211711  + 53868.21577013429 *T)
    sumY0  +=       0.017 * cos(5.76650726054  + 77307.86586563049 *T)
    sumY0  +=       0.017 * cos(5.48277957945  + 24602.85625235450 *T)
    sumY0  +=       0.017 * cos(3.32243810535  + 27573.43766576090 *T)
    sumY0  +=       0.023 * cos(0.08846433271  + 14.47091148511 *T)
    sumY0  +=       0.018 * cos(5.51755241491  + 52137.91706499410 *T)
    sumY0  +=       0.020 * cos(4.71185211530  + 130866.35771623049 *T)
    sumY0  +=       0.016 * cos(1.12213372181  + 79330.44871939608 *T)
    sumY0  +=       0.018 * cos(4.58737692359  + 79316.22162539448 *T)
    sumY0  +=       0.016 * cos(4.01365446910  + 3.63751787450 *T)
    sumY0  +=       0.017 * cos(1.24075088152  + 121335.36489626988 *T)
    sumY0  +=       0.019 * cos(4.71477962898  + 132028.83242063827 *T)
    sumY0  +=       0.018 * cos(1.49937345876  + 102659.69071427788 *T)
    sumY0  +=       0.018 * cos(4.05618608857  + 99023.89264043048 *T)
    sumY0  +=       0.016 * cos(6.02071363086  + 50903.36907500470 *T)
    sumY0  +=       0.019 * cos(1.66314996355  + 94138.08320260228 *T)
    sumY0  +=       0.016 * cos(4.53088777953  + 28310.00359365469 *T)
    sumY0  +=       0.016 * cos(0.64879776309  + 25668.17468021549 *T)
    sumY0  +=       0.016 * cos(4.77268559673  + 24395.98128955530 *T)
    sumY0  +=       0.019 * cos(1.08691679951  + 78793.15657233368 *T)
    sumY0  +=       0.015 * cos(0.94917563884  + 155887.76538061467 *T)
    sumY0  +=       0.022 * cos(4.51216513703  + 1055.69359440960 *T)
    sumY0  +=       0.016 * cos(1.31190457742  + 65717.47130312308 *T)
    sumY0  +=       0.021 * cos(3.55864300611  + 3178.38960805111 *T)
    sumY0  +=       0.019 * cos(0.15049069469  + 106570.12585300008 *T)
    sumY0  +=       0.015 * cos(2.07036309837  + 50690.06997956669 *T)
    sumY0  +=       0.017 * cos(1.75768691030  + 55638.29480649710 *T)
    sumY0  +=       0.016 * cos(5.70225976378  + 104358.48494268030 *T)
    sumY0  +=       0.015 * cos(5.84363368019  + 6043.98476389190 *T)
    sumY0  +=       0.016 * cos(5.00133834312  + 176332.48995851149 *T)
    sumY0  +=       0.017 * cos(0.43244881435  + 62197.88738073129 *T)
    sumY0  +=       0.017 * cos(0.00750879129  + 86457.74094044768 *T)
    sumY0  +=       0.015 * cos(1.85651180492  + 52310.15915021690 *T)
    sumY0  +=       0.015 * cos(5.22197013820  + 235900.75064374463 *T)
    sumY0  +=       0.015 * cos(3.15110803520  + 61560.89110870710 *T)
    sumY0  +=       0.017 * cos(4.69846105326  + 103821.67778371868 *T)
    sumY0  +=       0.017 * cos(3.00907412235  + 153085.08772196327 *T)
    sumY0  +=       0.020 * cos(6.14864822525  + 3.68833577960 *T)
    sumY0  +=       0.014 * cos(4.11502292960  + 193938.22990680728 *T)
    sumY0  +=       0.015 * cos(1.04578402465  + 50007.28961834930 *T)
    sumY0  +=       0.015 * cos(4.77727640495  + 26831.13701959030 *T)
    sumY0  +=       0.014 * cos(5.09443528119  + 133882.33447031991 *T)
    sumY0  +=       0.017 * cos(5.07246677176  + 39629.08052658189 *T)
    sumY0  +=       0.016 * cos(0.34319621032  + 77.99436146740 *T)
    sumY0  +=       0.015 * cos(3.15841157482  + 125112.28341697170 *T)
    sumY0  +=       0.014 * cos(0.72227336641  + 50380.79165691089 *T)
    sumY0  +=       0.016 * cos(2.56012166691  + 104345.22782488028 *T)
    sumY0  +=       0.014 * cos(0.38043611029  + 26247.44869384990 *T)
    sumY0  +=       0.014 * cos(2.14159626733  + 25928.84522426550 *T)
    sumY0  +=       0.014 * cos(5.50060239494  + 16066.30967895830 *T)
    sumY0  +=       0.016 * cos(2.29540093320  + 50910.48262200550 *T)
    sumY0  +=       0.018 * cos(3.69702904204  + 92740.81679043909 *T)
    sumY0  +=       0.014 * cos(4.21703179087  + 1169.83206889211 *T)
    sumY0  +=       0.016 * cos(2.14445815847  + 53815.11961781990 *T)
    sumY0  +=       0.014 * cos(2.88257919019  + 78149.51395352086 *T)
    sumY0  +=       0.016 * cos(2.81981718118  + 203042.13690092487 *T)
    sumY0  +=       0.013 * cos(1.94504793857  + 639.65346883050 *T)
    sumY0  +=       0.015 * cos(4.50829948617  + 77616.94240889128 *T)
    sumY0  +=       0.014 * cos(6.26019296208  + 82815.41910429788 *T)
    sumY0  +=       0.014 * cos(0.03534802891  + 29415.79498037089 *T)
    sumY0  +=       0.014 * cos(3.63252326636  + 102755.66994149568 *T)
    sumY0  +=       0.014 * cos(3.29326661713  + 109.96250373590 *T)
    sumY0  +=       0.017 * cos(4.87563601376  + 163298.37326817427 *T)
    sumY0  +=       0.015 * cos(0.10955371106  + 11.28951774740 *T)
    sumY0  +=       0.014 * cos(3.59146334775  + 173511.65881438527 *T)
    sumY0  +=       0.014 * cos(2.82070907634  + 52161.82300663029 *T)
    sumY0  +=       0.012 * cos(3.84809894730  + 102018.17235594128 *T)
    sumY0  +=       0.017 * cos(4.21238157873  + 52155.89270476510 *T)
    sumY0  +=       0.013 * cos(1.30925967982  + 102769.89703549728 *T)
    sumY0  +=       0.014 * cos(5.55450230200  + 78378.39253089130 *T)
    sumY0  +=       0.014 * cos(1.64692338972  + 26077.10125879380 *T)
    sumY0  +=       0.014 * cos(0.87510898790  + 26099.19265932160 *T)
    sumY0  +=       0.013 * cos(5.41473568620  + 79322.84753742829 *T)
    sumY0  +=       0.016 * cos(5.27893423231  + 51534.63653889291 *T)
    sumY0  +=       0.014 * cos(2.87967584826  + 20043.91837768230 *T)
    sumY0  +=       0.016 * cos(3.28350334233  + 50050.17269325269 *T)
    sumY0  +=       0.013 * cos(5.27868567961  + 44936.88687406489 *T)
    sumY0  +=       0.014 * cos(4.48508459393  + 24506.18761077469 *T)
    sumY0  +=       0.014 * cos(5.39693340588  + 35833.46751491430 *T)
    sumY0  +=       0.012 * cos(2.53842730053  + 52388.86156110289 *T)
    sumY0  +=       0.014 * cos(0.36435456631  + 198490.18340386607 *T)
    sumY0  +=       0.013 * cos(5.25497593084  + 103719.07264446709 *T)
    sumY0  +=       0.013 * cos(0.67641393113  + 24491.96051677309 *T)
    sumY0  +=       0.013 * cos(4.56168735677  + 52025.99926853029 *T)
    sumY0  +=       0.012 * cos(5.97483232665  + 25984.36477047710 *T)
    sumY0  +=       0.012 * cos(2.83038535815  + 26191.92914763830 *T)
    sumY0  +=       0.014 * cos(3.16933912424  + 53265.75889182550 *T)
    sumY0  +=       0.016 * cos(4.26587881339  + 25653.94758621389 *T)
    sumY0  +=       0.012 * cos(4.37902547397  + 181659.96606689427 *T)
    sumY0  +=       0.014 * cos(2.41047405735  + 53399.86794141051 *T)
    sumY0  +=       0.012 * cos(4.85231087723  + 53772.23654291649 *T)
    sumY0  +=       0.012 * cos(0.15443687995  + 43981.77476742330 *T)
    sumY0  +=       0.012 * cos(4.35356873262  + 28199.79727243529 *T)
    sumY0  +=       0.015 * cos(6.01961456404  + 1638.82569970449 *T)
    sumY0  +=       0.016 * cos(4.44894827862  + 34082.67540108190 *T)
    sumY0  +=       0.012 * cos(3.02030436898  + 51596.36251363310 *T)
    sumY0  +=       0.013 * cos(4.97119141190  + 52025.00243078899 *T)
    sumY0  +=       0.011 * cos(3.07690283750  + 52911.92661416369 *T)
    sumY0  +=       0.011 * cos(0.32794444369  + 70383.86408886709 *T)
    sumY0  +=       0.012 * cos(2.59844082236  + 112231.45790215228 *T)
    sumY0  +=       0.015 * cos(3.22392612987  + 103883.89139342587 *T)
    sumY0  +=       0.012 * cos(2.91594450386  + 52171.68113030689 *T)
    sumY0  +=       0.014 * cos(3.00638739155  + 26555.62431444510 *T)
    sumY0  +=       0.013 * cos(3.18586045586  + 286966.69073983253 *T)
    sumY0  +=       0.013 * cos(3.44611140023  + 208173.77798498247 *T)
    sumY0  +=       0.013 * cos(0.91076951955  + 79219.06534884768 *T)
    sumY0  +=       0.011 * cos(3.27827227016  + 52168.44891866409 *T)
    sumY0  +=       0.011 * cos(1.11898912575  + 26012.87772552339 *T)
    sumY0  +=       0.011 * cos(0.60900872596  + 48713.80539476670 *T)
    sumY0  +=       0.014 * cos(2.44543821114  + 27998.85880730789 *T)
    sumY0  +=       0.014 * cos(5.72095344070  + 52755.73768763070 *T)
    sumY0  +=       0.014 * cos(0.48193051968  + 52041.94105104690 *T)
    sumY0  +=       0.012 * cos(0.49425636709  + 131548.65044248087 *T)
    sumY0  +=       0.012 * cos(0.83151252253  + 130363.49345407888 *T)
    sumY0  +=       0.011 * cos(3.57268529306  + 104277.07478521300 *T)
    sumY0  +=       0.014 * cos(2.73442839497  + 27250.62166346549 *T)
    sumY0  +=       0.011 * cos(5.15064427323  + 51109.06698847489 *T)
    sumY0  +=       0.011 * cos(2.07371046178  + 115674.03284711768 *T)
    sumY0  +=       0.014 * cos(4.14980010220  + 49527.10764019189 *T)
    sumY0  +=       0.013 * cos(5.96769060739  + 52027.97137620559 *T)
    sumY0  +=       0.011 * cos(1.61005858699  + 57503.52620901470 *T)
    sumY0  +=       0.013 * cos(1.63784016727  + 78470.13879064328 *T)
    sumY0  +=       0.014 * cos(0.28439536761  + 860.55374623631 *T)
    sumY0  +=       0.011 * cos(2.58880557782  + 92741.30442540610 *T)
    sumY0  +=       0.013 * cos(0.61221768763  + 126996.69694542208 *T)
    sumY0  +=       0.011 * cos(5.96318002915  + 20894.77990025990 *T)
    sumY0  +=       0.012 * cos(5.42728450538  + 51440.17358710009 *T)
    sumY0  +=       0.015 * cos(2.18788554104  + 24395.49365458829 *T)
    sumY0  +=       0.012 * cos(2.65321755374  + 52179.98225389499 *T)
    sumY0  +=       0.013 * cos(5.06034087900  + 65716.98366815608 *T)
    sumY0  +=       0.014 * cos(5.96664556976  + 3442.33112748190 *T)
    sumY0  +=       0.011 * cos(3.48111494779  + 736.12033101530 *T)
    sumY0  +=       0.010 * cos(3.89350975005  + 333.61212328590 *T)
    sumY0  +=       0.011 * cos(0.13376559849  + 50536.98058344389 *T)
    sumY0  +=       0.010 * cos(4.16179119723  + 26190.75209830929 *T)
    sumY0  +=       0.012 * cos(0.94216106603  + 53758.00944891489 *T)
    sumY0  +=       0.011 * cos(0.54425828410  + 26081.03076519069 *T)
    sumY0  +=       0.010 * cos(1.22621171910  + 78340.21931348169 *T)
    sumY0  +=       0.011 * cos(3.31326356407  + 78800.75775430149 *T)
    sumY0  +=       0.012 * cos(1.33333763675  + 52179.44380102290 *T)
    sumY0  +=       0.010 * cos(2.68785159447  + 84945.17809560569 *T)
    sumY0  +=       0.011 * cos(5.57716825787  + 81592.08889848628 *T)
    sumY0  +=       0.011 * cos(0.93766822129  + 641.16974425549 *T)
    sumY0  +=       0.014 * cos(6.24613692649  + 183145.25677359747 *T)
    sumY0  +=       0.010 * cos(4.60397363665  + 25863.80216335580 *T)
    sumY0  +=       0.010 * cos(1.98851340448  + 53093.98038661369 *T)
    sumY0  +=       0.011 * cos(1.67023750371  + 51123.29408247649 *T)
    sumY0  +=       0.014 * cos(5.58647092759  + 25242.06412430650 *T)
    sumY0  +=       0.010 * cos(6.11242478468  + 154408.89880655028 *T)
    sumY0  +=       0.014 * cos(1.36873803506  + 54344.81058291449 *T)
    sumY0  +=       0.011 * cos(0.96041649146  + 78057.76769376888 *T)
    sumY0  +=       0.012 * cos(6.03701341811  + 26149.87293379789 *T)
    sumY0  +=       0.012 * cos(2.76820426669  + 26026.42098431750 *T)
    sumY0  +=       0.010 * cos(5.41577140545  + 637.24008950771 *T)
    sumY0  +=       0.010 * cos(0.79208503976  + 91919.81373338249 *T)
    sumY0  +=       0.014 * cos(1.69865910615  + 53125.22570960169 *T)
    sumY0  +=       0.011 * cos(3.68262194800  + 26094.28788299070 *T)
    sumY0  +=       0.012 * cos(2.92120571625  + 136100.60393953967 *T)
    sumY0  +=       0.010 * cos(2.60636817601  + 6751.28465782931 *T)
    sumY0  +=       0.010 * cos(2.45547925443  + 137209.98249163307 *T)
    sumY0  +=       0.010 * cos(1.99206585077  + 77940.44782554869 *T)
    sumY0  +=       0.010 * cos(4.36323003642  + 45404.85186441930 *T)
    sumY0  +=       0.013 * cos(4.08389676067  + 50270.58533569150 *T)
    sumY0  +=       0.011 * cos(3.67998983142  + 77520.96318167349 *T)
    sumY0  +=       0.011 * cos(4.23345221973  + 19958.84991955110 *T)
    sumY0  +=       0.010 * cos(1.41797581138  + 25619.69433373629 *T)
    sumY0  +=       0.011 * cos(3.70832983935  + 39763.67721113390 *T)
    sumY0  +=       0.010 * cos(2.60180323904  + 54086.76194888209 *T)
    sumY0  +=       0.009 * cos(1.01216637956  + 130593.53833583929 *T)
    sumY0  +=       0.009 * cos(0.36490625518  + 26514.05573557170 *T)
    sumY0  +=       0.009 * cos(2.15712612244  + 25662.23818254370 *T)
    sumY0  +=       0.010 * cos(5.58920157802  + 101703.40155830728 *T)
    sumY0  +=       0.013 * cos(3.18899327846  + 26161.44408491669 *T)
    sumY0  +=       0.013 * cos(5.61622440634  + 26014.84983319870 *T)
    sumY0  +=       0.010 * cos(0.86888264917  + 104991.75367009427 *T)
    sumY0  +=       0.010 * cos(5.95581190338  + 52061.65099371849 *T)
    sumY0  +=       0.009 * cos(6.14534715408  + 25131.85780308709 *T)
    sumY0  +=       0.011 * cos(4.00644107438  + 183674.94773869208 *T)
    sumY0  +=       0.010 * cos(5.20498017520  + 28791.76311373330 *T)
    sumY0  +=       0.010 * cos(4.41919261340  + 4372.12018425770 *T)
    sumY0  +=       0.010 * cos(2.27148842605  + 158746.41977111547 *T)
    sumY0  +=       0.009 * cos(0.68354219208  + 45290.90021070109 *T)
    sumY0  +=       0.010 * cos(3.59637305781  + 23384.53080438210 *T)
    sumY0  +=       0.012 * cos(4.17312463574  + 76042.09660760908 *T)
    sumY0  +=       0.012 * cos(1.90077351195  + 52396.46274307070 *T)
    sumY0  +=       0.013 * cos(4.38477804835  + 1073.85284167431 *T)
    sumY0  +=       0.010 * cos(1.29598384915  + 103924.77055793729 *T)
    sumY0  +=       0.010 * cos(5.03917514860  + 91804.88680973028 *T)
    sumY0  +=       0.009 * cos(2.39884072609  + 103189.38167937248 *T)
    sumY0  +=       0.009 * cos(4.93035759424  + 156547.33242782846 *T)
    sumY0  +=       0.009 * cos(0.52304323222  + 735.63269604830 *T)
    sumY0  +=       0.009 * cos(3.52896053406  + 45494.33761226529 *T)
    sumY0  +=       0.009 * cos(5.71417505627  + 114.68310616872 *T)
    sumY0  +=       0.009 * cos(0.77711516175  + 130285.98071486969 *T)
    sumY0  +=       0.013 * cos(5.11754456085  + 2168.51666479910 *T)
    sumY0  +=       0.010 * cos(5.91004109209  + 32132.37554043310 *T)
    sumY0  +=       0.009 * cos(2.04038725108  + 104984.64012309349 *T)
    sumY0  +=       0.011 * cos(1.80994732330  + 19367.43297971630 *T)
    sumY0  +=       0.010 * cos(0.87896683804  + 90989.40667698107 *T)
    sumY0  +=       0.009 * cos(4.68949629663  + 25550.85481199529 *T)
    sumY0  +=       0.009 * cos(5.45561348471  + 76255.39570304709 *T)
    sumY0  +=       0.009 * cos(2.84544963842  + 31281.51401785550 *T)
    sumY0  +=       0.012 * cos(3.79992568717  + 1699.52303398670 *T)
    sumY0  +=       0.011 * cos(5.50627359687  + 220.16882495529 *T)
    sumY0  +=       0.009 * cos(0.84632922145  + 51130.89526444429 *T)
    sumY0  +=       0.011 * cos(5.96410967567  + 4022.50634944771 *T)
    sumY0  +=       0.009 * cos(0.39890242427  + 52797.79390147110 *T)
    sumY0  +=       0.009 * cos(3.81505771377  + 24844.65935737930 *T)
    sumY0  +=       0.009 * cos(4.99015997103  + 27331.63456073610 *T)
    sumY0  +=       0.010 * cos(5.97083270008  + 230239.41859459243 *T)
    sumY0  +=       0.009 * cos(2.46893459204  + 156314.36357149068 *T)
    sumY0  +=       0.011 * cos(1.59945570289  + 80175.15272542329 *T)
    sumY0  +=       0.009 * cos(3.62938673151  + 129329.89333829407 *T)
    sumY0  +=       0.008 * cos(1.08654878495  + 38634.62889814070 *T)
    sumY0  +=       0.011 * cos(4.12180664109  + 46046.75306112530 *T)
    sumY0  +=       0.009 * cos(5.45527892151  + 52190.27719463350 *T)
    sumY0  +=       0.009 * cos(2.96616057426  + 25454.87558477749 *T)
    sumY0  +=       0.009 * cos(2.58378869705  + 23439.69213358470 *T)
    sumY0  +=       0.008 * cos(0.94557962119  + 102762.29585352947 *T)
    sumY0  +=       0.011 * cos(1.01537219957  + 2538.49232173710 *T)
    sumY0  +=       0.010 * cos(1.53707512245  + 391318.79094109638 *T)
    sumY0  +=       0.008 * cos(2.58065685518  + 80174.66509045628 *T)
    sumY0  +=       0.008 * cos(4.23319522465  + 142871.80217575227 *T)
    sumY0  +=       0.009 * cos(4.26758067831  + 76152.30292882849 *T)
    sumY0  +=       0.010 * cos(5.47825003046  + 54060.95126404929 *T)
    sumY0  +=       0.009 * cos(3.23120957887  + 4.17597074660 *T)
    sumY0  +=       0.010 * cos(5.87240869715  + 35472.50033216589 *T)
    sumY0  +=       0.010 * cos(1.00378919066  + 125887.31839332868 *T)
    sumY0  +=       0.008 * cos(5.81576723468  + 27360.82798468490 *T)
    sumY0  +=       0.009 * cos(0.97115189447  + 1685.29593998510 *T)
    sumY0  +=       0.009 * cos(0.79804703659  + 58458.63831565629 *T)
    sumY0  +=       0.009 * cos(0.54956806871  + 25977.45300287129 *T)
    sumY0  +=       0.009 * cos(3.97874723563  + 1485.04688921970 *T)
    sumY0  +=       0.008 * cos(1.17507209551  + 2325.19322629910 *T)
    sumY0  +=       0.009 * cos(2.26549130862  + 22595.57539218810 *T)
    sumY0  +=       0.009 * cos(5.61147436738  + 52172.11794736879 *T)
    sumY0  +=       0.008 * cos(3.83518676972  + 26934.22979380890 *T)
    sumY0  +=       0.008 * cos(0.59545092796  + 1575.08961876571 *T)
    sumY0  +=       0.008 * cos(5.37389235751  + 51876.92370636270 *T)
    sumY0  +=       0.010 * cos(1.95662555086  + 2751.79141717511 *T)
    sumY0  +=       0.008 * cos(0.89302678287  + 10681.49435404891 *T)
    sumY0  +=       0.011 * cos(3.21704132053  + 28096.70449821670 *T)
    sumY0  +=       0.008 * cos(5.45338504662  + 77956.39562123650 *T)
    sumY0  +=       0.008 * cos(0.89135797231  + 52400.39489633380 *T)
    sumY0  +=       0.008 * cos(2.89179819588  + 54824.50492610490 *T)
    sumY0  +=       0.010 * cos(1.50538632513  + 240452.70414080343 *T)
    sumY0  +=       0.008 * cos(5.26980517423  + 77741.37582411229 *T)
    sumY0  +=       0.009 * cos(3.90818318828  + 24491.47288180609 *T)
    sumY0  +=       0.008 * cos(5.13755213219  + 42791.20909255670 *T)
    sumY0  +=       0.008 * cos(3.13545736715  + 52490.23628050151 *T)
    sumY0  +=       0.008 * cos(2.29664532419  + 156520.54911992789 *T)
    sumY0  +=       0.007 * cos(4.49316157626  + 52206.37696163630 *T)
    sumY0  +=       0.008 * cos(0.58751306444  + 17098.67925362531 *T)
    sumY0  +=       0.007 * cos(3.15082437070  + 50291.14893721450 *T)
    sumY0  +=       0.010 * cos(2.92644275072  + 22065.88442709350 *T)
    sumY0  +=       0.008 * cos(1.07948674297  + 51112.73601717959 *T)
    sumY0  +=       0.009 * cos(0.57156355893  + 104454.94915799888 *T)
    sumY0  +=       0.010 * cos(1.70608986575  + 441.06910236111 *T)
    sumY0  +=       0.010 * cos(2.64731129677  + 49.75280442070 *T)
    sumY0  +=       0.007 * cos(2.09922885467  + 7880.33297082250 *T)
    sumY0  +=       0.007 * cos(6.22957682161  + 52182.67601266569 *T)
    sumY0  +=       0.008 * cos(4.88130543295  + 52099.29639438929 *T)
    sumY0  +=       0.007 * cos(3.72492189200  + 122445.23108333028 *T)
    sumY0  +=       0.008 * cos(4.87864527535  + 66940.80150893469 *T)
    sumY0  +=       0.008 * cos(4.83064762406  + 187167.51930556167 *T)
    sumY0  +=       0.009 * cos(2.82365374916  + 156507.99290602890 *T)
    sumY0  +=       0.007 * cos(5.45461374708  + 51086.34130943830 *T)
    sumY0  +=       0.008 * cos(5.15492830863  + 430.77416162260 *T)
    sumY0  +=       0.008 * cos(3.96537599750  + 149288.98707532868 *T)
    sumY0  +=       0.008 * cos(0.69966864109  + 49956.80536147810 *T)
    sumY0  +=       0.007 * cos(5.66541358879  + 53132.33925660249 *T)
    sumY0  +=       0.008 * cos(4.19457404849  + 100908.79380384788 *T)
    sumY0  +=       0.007 * cos(2.13361844991  + 233681.99353955788 *T)
    sumY0  +=       0.007 * cos(0.76923088178  + 104248.76360956168 *T)
    sumY0  +=       0.008 * cos(3.46520766649  + 77314.77763323628 *T)
    sumY0  +=       0.008 * cos(4.96646407077  + 112231.94553711927 *T)
    sumY0  +=       0.007 * cos(0.02649996061  + 29396.61285443810 *T)
    sumY0  +=       0.007 * cos(2.49553241701  + 22779.68106367730 *T)
    sumY0  +=       0.009 * cos(4.57445854869  + 64741.71416564769 *T)
    sumY0  +=       0.007 * cos(4.38534662923  + 26290.40035423180 *T)
    sumY0  +=       0.007 * cos(4.41987105557  + 25885.89356388359 *T)
    sumY0  +=       0.009 * cos(5.15387307390  + 25863.31452838879 *T)
    sumY0  +=       0.007 * cos(4.98170247896  + 13655.61667369290 *T)
    sumY0  +=       0.007 * cos(1.46712503162  + 156740.96176236667 *T)
    sumY0  +=       0.008 * cos(2.98540358627  + 78571.51086317569 *T)
    sumY0  +=       0.007 * cos(1.57299130322  + 52663.68485188610 *T)
    sumY0  +=       0.007 * cos(3.13857237569  + 2118.52004289490 *T)
    sumY0  +=       0.008 * cos(3.34242363485  + 102872.98980971587 *T)
    sumY0  +=       0.010 * cos(4.97428981699  + 2703.37233719210 *T)
    sumY0  +=       0.007 * cos(5.14819246138  + 52074.08758236629 *T)
    sumY0  +=       0.007 * cos(0.47864118434  + 209659.06869168568 *T)
    sumY0  +=       0.007 * cos(2.10592218772  + 149846.43781352908 *T)
    sumY0  +=       0.007 * cos(2.73796346194  + 77417.87040745489 *T)
    sumY0  +=       0.009 * cos(5.86662388336  + 56260.03860733629 *T)
    sumY0  +=       0.007 * cos(5.07150057843  + 117892.78995130444 *T)
    sumY0  +=       0.009 * cos(1.31305807687  + 18073.94875613370 *T)
    sumY0  +=       0.007 * cos(5.92131553777  + 917.68646849830 *T)
    sumY0  +=       0.008 * cos(0.71246556222  + 32808.86093839910 *T)
    sumY0  +=       0.008 * cos(0.19228169341  + 207114.39605479327 *T)
    sumY0  +=       0.008 * cos(0.61342229584  + 3308.22207789689 *T)
    sumY0  +=       0.007 * cos(1.34578982527  + 23267.31581493150 *T)
    sumY0  +=       0.007 * cos(1.17624255235  + 28908.97810318390 *T)
    sumY0  +=       0.007 * cos(4.68208072339  + 37.88921815429 *T)
    sumY0  +=       0.007 * cos(1.41809701113  + 19202.50943415989 *T)
    sumY0  +=       0.007 * cos(4.13036001377  + 52145.72323962749 *T)
    sumY0  +=       0.008 * cos(2.80344738918  + 103395.56722780968 *T)
    sumY0  +=       0.007 * cos(0.00761390447  + 208276.87075920109 *T)
    sumY0  +=       0.009 * cos(3.38149297609  + 2014.73785431429 *T)
    sumY0  +=       0.007 * cos(5.79510753162  + 63786.60205900610 *T)
    sumY0  +=       0.007 * cos(1.26662001445  + 24712.37315921190 *T)
    sumY0  +=       0.007 * cos(1.07441550954  + 127099.14709865468 *T)
    sumY0  +=       0.008 * cos(4.81617793405  + 120417.92224525507 *T)
    sumY0  +=       0.008 * cos(1.39078986683  + 103815.05187168487 *T)
    sumY0  +=       0.007 * cos(2.56899142970  + 8989.22388794889 *T)
    sumY0  +=       0.008 * cos(3.19497137388  + 579.44376951530 *T)
    sumY0  +=       0.007 * cos(3.93035335903  + 1135.43305440970 *T)
    sumY0  +=       0.006 * cos(3.50717502178  + 27574.12708012290 *T)
    sumY0  +=       0.007 * cos(4.34157997430  + 102976.08258393448 *T)
    sumY0  +=       0.006 * cos(2.26227426560  + 28774.86905359890 *T)
    sumY0  +=       0.006 * cos(0.25975811202  + 23401.42486451650 *T)
    sumY0  +=       0.006 * cos(6.10166450881  + 80463.04058549309 *T)
    sumY0  +=       0.008 * cos(0.39455403552  + 51258.11981465010 *T)
    sumY0  +=       0.006 * cos(2.04791015101  + 23704.95394428149 *T)
    sumY0  +=       0.006 * cos(0.47412222661  + 28471.33997383390 *T)
    sumY0  +=       0.006 * cos(3.25327779066  + 55618.13741063030 *T)
    sumY0  +=       0.007 * cos(0.04757570199  + 53311.72697252510 *T)
    sumY0  +=       0.007 * cos(3.96061402799  + 1802.61580820530 *T)
    sumY0  +=       0.008 * cos(3.11255216313  + 102018.65999090829 *T)
    sumY0  +=       0.007 * cos(3.75772081758  + 79995.07559513870 *T)
    sumY0  +=       0.007 * cos(3.47532912865  + 52278.01261889750 *T)
    sumY0  +=       0.008 * cos(0.17460078065  + 1162.23088692429 *T)
    sumY0  +=       0.006 * cos(4.49776146960  + 53439.20746321010 *T)
    sumY0  +=       0.007 * cos(0.23469102383  + 29580.71852592730 *T)
    sumY0  +=       0.007 * cos(2.10834895065  + 27388.97377726269 *T)
    sumY0  +=       0.007 * cos(0.41368342697  + 24787.32014085270 *T)
    sumY0  +=       0.006 * cos(2.20070614265  + 106470.62024415868 *T)
    sumY0  +=       0.007 * cos(1.31790121285  + 104127.51158807840 *T)
    sumY0  +=       0.007 * cos(5.05330889636  + 27624.12370202710 *T)
    sumY0  +=       0.007 * cos(3.75190878844  + 24552.17021608830 *T)
    sumY0  +=       0.008 * cos(5.40150242632  + 948.93179148630 *T)
    sumY0  +=       0.007 * cos(1.23626167150  + 6680.98103591609 *T)
    sumY0  +=       0.008 * cos(5.48806422287  + 51861.86392076229 *T)
    sumY0  +=       0.008 * cos(2.01472939327  + 51554.30629979270 *T)
    sumY0  +=       0.007 * cos(5.80107722174  + 26720.44306340389 *T)
    sumY0  +=       0.007 * cos(1.27572316405  + 52214.18313626970 *T)
    sumY0  +=       0.007 * cos(4.01717883587  + 25572.68308796469 *T)
    sumY0  +=       0.007 * cos(4.78803884893  + 26603.61083015070 *T)
    sumY0  +=       0.007 * cos(3.81203720019  + 415.79630809560 *T)
    sumY0  +=       0.008 * cos(5.83496057018  + 54191.03177242970 *T)
    sumY0  +=       0.006 * cos(1.87365150772  + 24505.69997580769 *T)
    sumY0  +=       0.007 * cos(2.72912420029  + 2974.85377209451 *T)
    sumY0  +=       0.007 * cos(2.73283247417  + 48092.06159392750 *T)
    sumY0  +=       0.006 * cos(2.27207352577  + 78043.54059976728 *T)
    sumY0  +=       0.006 * cos(0.76622305958  + 19202.99706912689 *T)
    sumY0  +=       0.006 * cos(0.12929656543  + 53661.34080733510 *T)
    sumY0  +=       0.006 * cos(4.92644191429  + 104426.63798234760 *T)
    sumY0  +=       0.006 * cos(4.97320423795  + 27883.40540277870 *T)
    sumY0  +=       0.006 * cos(1.52227576605  + 156523.78133157067 *T)
    sumY0  +=       0.008 * cos(4.29203597963  + 51120.60032370579 *T)
    sumY0  +=       0.006 * cos(6.23544000173  + 226.07308589371 *T)
    sumY0  +=       0.008 * cos(5.23117913800  + 48997.90431006430 *T)
    sumY0  +=       0.006 * cos(3.11031405202  + 149756.95206568309 *T)
    sumY0  +=       0.007 * cos(3.39730503781  + 4083.74468922089 *T)
    sumY0  +=       0.008 * cos(2.23399255632  + 53259.13033292550 *T)
    sumY0  +=       0.006 * cos(3.49957202036  + 130006.04778747768 *T)
    sumY0  +=       0.007 * cos(6.08095772039  + 162810.73851692007 *T)
    sumY0  +=       0.006 * cos(0.64154839978  + 27441.89570407450 *T)
    sumY0  +=       0.006 * cos(1.88048397784  + 24734.39821404090 *T)
    sumY0  +=       0.007 * cos(0.08806115428  + 55503.69812194508 *T)
    sumY0  +=       0.008 * cos(1.92055633433  + 422.90985509640 *T)
    sumY0  +=       0.006 * cos(5.11522815430  + 76777.97312114088 *T)
    sumY0  +=       0.006 * cos(3.95276470433  + 155461.16718973868 *T)
    sumY0  +=       0.006 * cos(2.81558483589  + 50161.06842883410 *T)
    sumY0  +=       0.006 * cos(3.77961930116  + 78262.46876949779 *T)
    sumY0  +=       0.006 * cos(2.10450702949  + 223.10414047710 *T)
    sumY0  +=       0.006 * cos(2.19241811447  + 78225.82020656829 *T)
    sumY0  +=       0.006 * cos(3.79548807692  + 76468.69479848508 *T)
    sumY0  +=       0.006 * cos(3.10628150548  + 78597.80918297550 *T)
    sumY0  +=       0.007 * cos(5.87498441401  + 30110.40949102190 *T)
    sumY0  +=       0.006 * cos(1.92388121650  + 35077.61466449009 *T)
    sumY0  +=       0.007 * cos(2.98999765548  + 1884.65734593389 *T)
    sumY0  +=       0.007 * cos(6.11933292351  + 1082.83641481010 *T)
    sumY0  +=       0.006 * cos(3.53656686000  + 5193.12324131429 *T)
    sumY0  +=       0.006 * cos(3.60785310094  + 171292.90171019849 *T)
    sumY0  +=       0.006 * cos(4.43885235196  + 25984.56654987209 *T)
    sumY0  +=       0.006 * cos(2.63023124509  + 29864.57784479250 *T)
    sumY0  +=       0.006 * cos(6.17498643970  + 22311.71607332290 *T)
    sumY0  +=       0.006 * cos(2.81657351020  + 76991.27221657889 *T)
    sumY0  +=       0.005 * cos(5.40188161644  + 27410.82460658910 *T)
    sumY0  +=       0.005 * cos(3.40333606836  + 24765.46931152630 *T)
    sumY0  +=       0.006 * cos(0.12287107983  + 51092.96986833830 *T)
    sumY0  +=       0.007 * cos(5.41576313041  + 88476.75115345008 *T)
    sumY0  +=       0.007 * cos(1.20946078842  + 34102.34516198170 *T)
    sumY0  +=       0.005 * cos(2.83164430151  + 2686.47827705770 *T)
    sumY0  +=       0.006 * cos(3.98658256476  + 31903.25781869470 *T)
    sumY0  +=       0.006 * cos(4.81863512004  + 20273.03609942070 *T)
    sumY0  +=       0.006 * cos(1.29004214195  + 104874.43380187408 *T)
    sumY0  +=       0.005 * cos(0.59730065618  + 155475.39428374029 *T)
    sumY0  +=       0.005 * cos(4.57106999295  + 128747.59385585209 *T)
    sumY0  +=       0.005 * cos(3.48020412230  + 26500.51805593209 *T)
    sumY0  +=       0.005 * cos(5.32501356250  + 25675.77586218330 *T)
    sumY0  +=       0.007 * cos(5.70848080046  + 25234.46294233869 *T)
    sumY0  +=       0.005 * cos(4.12117505892  + 53551.82390047770 *T)
    sumY0  +=       0.005 * cos(5.05855634652  + 60170.57854265610 *T)
    sumY0  +=       0.005 * cos(4.48575981388  + 27353.71443768409 *T)
    sumY0  +=       0.007 * cos(2.75081241109  + 52602.16065654089 *T)
    sumY0  +=       0.005 * cos(3.92111104398  + 128106.07549751546 *T)
    sumY0  +=       0.005 * cos(0.45323700456  + 629.09540353360 *T)
    sumY0  +=       0.005 * cos(2.88226149102  + 35191.56631820829 *T)
    sumY0  +=       0.006 * cos(0.61158685386  + 3492.81538435310 *T)
    sumY0  +=       0.007 * cos(3.24751127495  + 169093.81436691148 *T)
    sumY0  +=       0.006 * cos(0.00555930589  + 81591.60126351927 *T)
    sumY0  +=       0.006 * cos(3.27898581061  + 25462.47676674530 *T)
    sumY0  +=       0.006 * cos(5.52623187419  + 26713.81715137009 *T)
    sumY0  +=       0.007 * cos(1.40182187541  + 2641.58509595570 *T)
    sumY0  +=       0.005 * cos(2.22571570610  + 58220.27868200730 *T)
    sumY0  +=       0.005 * cos(1.13017778403  + 52813.04637265610 *T)
    sumY0  +=       0.006 * cos(6.23315922928  + 156531.54400228668 *T)
    sumY0  +=       0.005 * cos(5.91150949505  + 26236.22568348400 *T)
    sumY0  +=       0.007 * cos(6.01504388975  + 9745.07673837309 *T)
    sumY0  +=       0.005 * cos(3.16675569939  + 26035.05080674330 *T)
    sumY0  +=       0.005 * cos(5.63846198541  + 26141.24311137209 *T)
    sumY0  +=       0.006 * cos(2.65718358525  + 117873.12019040468 *T)
    sumY0  +=       0.006 * cos(2.30069536250  + 28214.02436643689 *T)
    sumY0  +=       0.005 * cos(5.51855350061  + 487.39093377069 *T)
    sumY0  +=       0.006 * cos(2.84193454109  + 53764.63536094869 *T)
    sumY0  +=       0.005 * cos(2.79855313555  + 104035.46451412368 *T)
    sumY0  +=       0.006 * cos(3.26310713546  + 437.88770862340 *T)
    sumY0  +=       0.005 * cos(2.17730624219  + 52098.29955664799 *T)
    sumY0  +=       0.006 * cos(3.10526072910  + 98068.29289882188 *T)
    sumY0  +=       0.005 * cos(1.17059070361  + 65538.49980743109 *T)
    sumY0  +=       0.005 * cos(0.60786805630  + 189854.24140010288 *T)
    sumY0  +=       0.005 * cos(0.07976631221  + 76138.07583482689 *T)
    sumY0  +=       0.005 * cos(3.31240476519  + 104301.85976187608 *T)
    sumY0  +=       0.005 * cos(5.17958441161  + 1243.73141916191 *T)
    sumY0  +=       0.005 * cos(5.92076756755  + 118829.20756698029 *T)
    sumY0  +=       0.006 * cos(4.31566445968  + 13675.28643459270 *T)
    sumY0  +=       0.005 * cos(3.70296203327  + 153084.60008699627 *T)
    sumY0  +=       0.006 * cos(3.07685642656  + 13.98327651810 *T)
    sumY0  +=       0.005 * cos(4.76122192085  + 78697.66498008289 *T)
    sumY0  +=       0.005 * cos(5.20357112241  + 26610.23674218449 *T)
    sumY0  +=       0.005 * cos(0.29390226295  + 1052.02456570489 *T)
    sumY0  +=       0.005 * cos(5.27501982781  + 149.31937965110 *T)
    sumY0  +=       0.006 * cos(1.42019585673  + 122444.74344836328 *T)
    sumY0  +=       0.006 * cos(3.75361834376  + 24822.57948043130 *T)
    sumY0  +=       0.005 * cos(2.37687918815  + 26162.44092265799 *T)
    sumY0  +=       0.005 * cos(4.48428750718  + 154.02262796830 *T)
    sumY0  +=       0.005 * cos(4.82516506006  + 104401.85300568448 *T)
    sumY0  +=       0.005 * cos(5.03758504482  + 53654.91667469629 *T)
    sumY0  +=       0.005 * cos(0.38799912808  + 118828.71993201328 *T)
    sumY0  +=       0.005 * cos(1.40704909535  + 191.20444862810 *T)
    sumY0  +=       0.005 * cos(1.05885873562  + 124777.93984123527 *T)
    sumY0  +=       0.005 * cos(4.57253368744  + 51006.46184922329 *T)
    sumY0  +=       0.006 * cos(2.99064022137  + 104241.65006256088 *T)
    sumY0  +=       0.005 * cos(5.36734313790  + 104668.24825343688 *T)
    sumY0  +=       0.004 * cos(4.69357830395  + 76532.83088927349 *T)
    sumY0  +=       0.006 * cos(2.26469618573  + 75615.01078176608 *T)
    sumY0  +=       0.004 * cos(3.91764666613  + 24608.79275002629 *T)
    sumY0  +=       0.004 * cos(0.77605401254  + 24609.28038499330 *T)
    sumY0  +=       0.005 * cos(3.59914095117  + 23389.69551168050 *T)
    sumY0  +=       0.005 * cos(5.20607673363  + 28786.59840643490 *T)
    sumY0  +=       0.004 * cos(5.84756896317  + 31775.77732800970 *T)
    sumY0  +=       0.004 * cos(2.95764872163  + 20400.51659010570 *T)
    sumY0  +=       0.005 * cos(6.27101158445  + 107679.50440509344 *T)
    sumY0  +=       0.004 * cos(4.55685472779  + 13541.66501997470 *T)
    sumY0  +=       0.005 * cos(0.86738231950  + 29428.27175079049 *T)
    sumY0  +=       0.005 * cos(0.59524275071  + 26.05450231630 *T)
    sumY0  +=       0.004 * cos(2.28993527511  + 25764.64154240029 *T)
    sumY0  +=       0.005 * cos(4.52541301080  + 78265.43771491438 *T)
    sumY0  +=       0.006 * cos(0.78206719980  + 51688.41534937770 *T)
    sumY0  +=       0.006 * cos(1.53449593067  + 23865.80268949369 *T)
    sumY0  +=       0.004 * cos(1.16733607750  + 35211.72371407509 *T)
    sumY0  +=       0.004 * cos(1.35469630012  + 16964.57020404031 *T)
    sumY0  +=       0.004 * cos(4.36135738215  + 47804.17373385770 *T)
    sumY0  +=       0.004 * cos(5.40330574142  + 77623.56832092509 *T)
    sumY0  +=       0.005 * cos(5.21473868834  + 86458.22857541469 *T)
    sumY0  +=       0.004 * cos(5.79724758149  + 27040.22358231730 *T)
    sumY0  +=       0.004 * cos(3.00797010331  + 25136.07033579810 *T)
    sumY0  +=       0.005 * cos(3.02879597074  + 25778.86863640189 *T)
    sumY0  +=       0.005 * cos(5.77642171406  + 26397.42528171349 *T)
    sumY0  +=       0.005 * cos(2.45203846176  + 78374.15956342549 *T)
    sumY0  +=       0.004 * cos(5.04872942833  + 48834.95003896509 *T)
    sumY0  +=       0.005 * cos(0.30596644618  + 25771.26745443409 *T)
    sumY0  +=       0.004 * cos(0.10837463046  + 33480.60136114250 *T)
    sumY0  +=       0.004 * cos(2.41365774716  + 18695.69255697290 *T)
    sumY0  +=       0.005 * cos(5.31228827456  + 703.87700210090 *T)
    sumY0  +=       0.004 * cos(3.53618772681  + 156378.09946979408 *T)
    sumY0  +=       0.005 * cos(1.85614150808  + 155467.79310177249 *T)
    sumY0  +=       0.005 * cos(2.86964254252  + 26404.05119374729 *T)
    sumY0  +=       0.004 * cos(1.76062802705  + 9123.33293753389 *T)
    sumY0  +=       0.005 * cos(0.75540582554  + 2015.91490364330 *T)
    sumY0  +=       0.004 * cos(5.33219851688  + 76998.38576357969 *T)
    sumY0  +=       0.004 * cos(4.84304007509  + 129586.56314360249 *T)
    sumY0  +=       0.004 * cos(1.42347412662  + 1368.90407032850 *T)
    sumY0  +=       0.004 * cos(6.20388780658  + 45425.00926028610 *T)
    sumY0  +=       0.004 * cos(3.82794287315  + 28928.64786408370 *T)
    sumY0  +=       0.004 * cos(4.97727481165  + 23247.64605403170 *T)
    sumY0  +=       0.004 * cos(3.79698584398  + 23866.29032446070 *T)
    sumY0  +=       0.004 * cos(0.78108419173  + 20453.11322970530 *T)
    sumY0  +=       0.004 * cos(1.74094818589  + 31723.18068841010 *T)
    sumY0  +=       0.004 * cos(0.73252151870  + 26239.19462890059 *T)
    sumY0  +=       0.004 * cos(0.40730526869  + 199599.56195595945 *T)
    sumY0  +=       0.004 * cos(1.20756668103  + 52356.12723091650 *T)
    sumY0  +=       0.004 * cos(2.82011030245  + 26137.65594599489 *T)
    sumY0  +=       0.004 * cos(1.89262749454  + 51539.05382860770 *T)
    sumY0  +=       0.004 * cos(2.12864958126  + 32279.16997330450 *T)
    sumY0  +=       0.004 * cos(0.39338279636  + 19897.12394481090 *T)
    sumY0  +=       0.004 * cos(0.13108394311  + 78905.36680394509 *T)
    sumY0  +=       0.004 * cos(1.42991389286  + 52336.45747001670 *T)
    sumY0  +=       0.004 * cos(4.11596381659  + 50912.89273805370 *T)
    sumY0  +=       0.005 * cos(5.25156458726  + 26085.17801364109 *T)
    sumY0  +=       0.005 * cos(3.55365309754  + 26091.11590447430 *T)
    sumY0  +=       0.004 * cos(4.55748177466  + 25441.13612574289 *T)
    sumY0  +=       0.004 * cos(1.77632201645  + 25923.71104846829 *T)
    sumY0  +=       0.004 * cos(0.74571036117  + 26252.58286964711 *T)
    sumY0  +=       0.004 * cos(2.50276715961  + 24925.18461968289 *T)
    sumY0  +=       0.005 * cos(5.55461968829  + 13706.10093056410 *T)
    sumY0  +=       0.005 * cos(3.25059799651  + 38470.19298755129 *T)
    sumY0  +=       0.005 * cos(3.12816538610  + 18227.72756661850 *T)
    sumY0  +=       0.005 * cos(5.67705229870  + 33948.56635149690 *T)
    sumY0  +=       0.005 * cos(1.28804998148  + 10.80188278040 *T)
    sumY0  +=       0.004 * cos(4.38061016047  + 147423.26803784407 *T)
    sumY0  +=       0.004 * cos(4.94619430058  + 120225.98634417648 *T)
    sumY0  +=       0.004 * cos(4.22377863391  + 95.73540973430 *T)
    sumY0  +=       0.004 * cos(2.78744864067  + 74935.81758592589 *T)
    sumY0  +=       0.004 * cos(1.85033590581  + 142871.31454078527 *T)
    sumY0  +=       0.004 * cos(5.08828008747  + 26131.03003396110 *T)
    sumY0  +=       0.004 * cos(3.71693759733  + 26045.26388415430 *T)
    sumY0  +=       0.004 * cos(4.81796795304  + 1066.25165970650 *T)
    sumY0  +=       0.004 * cos(2.30874038706  + 25035.39094090229 *T)
    sumY0  +=       0.005 * cos(5.79505565508  + 1577.09972496430 *T)
    sumY0  +=       0.004 * cos(2.13386788705  + 1262.91354509469 *T)
    sumY0  +=       0.005 * cos(4.43555583939  + 51102.44107644109 *T)
    sumY0  +=       0.004 * cos(2.69709923008  + 106262.56823203048 *T)
    sumY0  +=       0.004 * cos(6.19718375992  + 104771.34102765548 *T)
    sumY0  +=       0.005 * cos(6.11030752375  + 179172.99086353747 *T)
    sumY0  +=       0.004 * cos(5.45224037983  + 52329.34127614969 *T)
    sumY0  +=       0.005 * cos(0.05651935942  + 151200.18655854589 *T)
    sumY0  +=       0.004 * cos(1.76945825888  + 18319.78040236310 *T)
    sumY0  +=       0.004 * cos(0.75257411874  + 33856.51351575230 *T)
    sumY0  +=       0.005 * cos(2.23665069420  + 51315.74017187909 *T)
    sumY0  +=       0.004 * cos(1.69359058279  + 159970.23761189409 *T)
    sumY0  +=       0.005 * cos(3.45474588143  + 107692.46881048007 *T)
    sumY0  +=       0.004 * cos(4.69804652488  + 24182.19455915029 *T)
    sumY0  +=       0.004 * cos(5.36463943224  + 25021.16384690069 *T)
    sumY0  +=       0.005 * cos(1.47989657697  + 49637.80159637830 *T)
    sumY0  +=       0.004 * cos(5.19657656850  + 64901.01590043989 *T)
    sumY0  +=       0.005 * cos(5.34571928246  + 77930.09730143669 *T)
    sumY0  +=       0.004 * cos(2.13537389331  + 106685.05277665229 *T)
    sumY0  +=       0.004 * cos(1.30241374700  + 25249.17767130730 *T)
    sumY0  +=       0.004 * cos(1.21961863062  + 26927.11624680809 *T)
    sumY0  +=       0.005 * cos(6.25905047664  + 130020.27488147929 *T)
    sumY0  +=       0.004 * cos(0.16897355283  + 42430.72954477530 *T)
    sumY0  +=       0.004 * cos(1.81097172729  + 189386.27640974848 *T)
    sumY0  +=       0.004 * cos(1.67938382417  + 130907.72451570888 *T)
    sumY0  +=       0.004 * cos(0.49273474013  + 76352.75375898890 *T)
    sumY0  +=       0.004 * cos(0.60720939762  + 78050.16651180109 *T)
    sumY0  +=       0.004 * cos(3.81303761260  + 26507.14396796589 *T)
    sumY0  +=       0.004 * cos(5.10986588518  + 21715.78295731649 *T)
    sumY0  +=       0.004 * cos(2.24959479186  + 50690.75939392870 *T)
    sumY0  +=       0.003 * cos(4.11846884673  + 76095.19275992349 *T)
    sumY0  +=       0.004 * cos(5.11517156658  + 27047.26981901530 *T)
    sumY0  +=       0.004 * cos(3.69004611822  + 25129.02409910010 *T)
    sumY0  +=       0.003 * cos(2.87423294156  + 25980.84165212810 *T)
    sumY0  +=       0.003 * cos(5.93098474324  + 26195.45226598730 *T)
    sumY0  +=       0.004 * cos(3.06482161328  + 52165.00440036799 *T)
    sumY0  +=       0.004 * cos(4.26611311394  + 104881.05971390787 *T)
    sumY0  +=       0.004 * cos(3.03103303969  + 8044.76888141190 *T)
    sumY0  +=       0.005 * cos(4.86351403293  + 45455.33612129050 *T)
    sumY0  +=       0.004 * cos(0.29804014939  + 25565.08190599689 *T)
    sumY0  +=       0.004 * cos(0.98706116692  + 215942.14454167709 *T)
    sumY0  +=       0.004 * cos(3.84907251953  + 25992.16773183990 *T)
    sumY0  +=       0.004 * cos(4.95614516527  + 26184.12618627550 *T)
    sumY0  +=       0.003 * cos(1.64017894322  + 129909.58092529287 *T)
    sumY0  +=       0.004 * cos(2.78312246204  + 48153.78756866770 *T)
    sumY0  +=       0.005 * cos(6.06876669605  + 10059.75055320970 *T)
    sumY0  +=       0.004 * cos(5.16058547691  + 16496.60521368590 *T)
    sumY0  +=       0.004 * cos(3.64463220789  + 35679.68870442950 *T)
    sumY0  +=       0.004 * cos(2.60946762791  + 103395.76900720468 *T)
    sumY0  +=       0.004 * cos(2.31682392343  + 51040.37322873870 *T)
    sumY0  +=       0.003 * cos(0.92201563297  + 125111.79578200469 *T)
    sumY0  +=       0.004 * cos(0.89339658044  + 128843.57308306987 *T)
    sumY0  +=       0.004 * cos(0.91042238998  + 52015.64273124709 *T)
    sumY0  +=       0.004 * cos(3.60881337650  + 621.98761832270 *T)
    sumY0  +=       0.003 * cos(6.06256841288  + 23113.53700444670 *T)
    sumY0  +=       0.003 * cos(2.74264927192  + 29062.75691366870 *T)
    sumY0  +=       0.003 * cos(4.85231198714  + 25794.73617468050 *T)
    sumY0  +=       0.003 * cos(3.95290569766  + 26381.55774343490 *T)
    sumY0  +=       0.003 * cos(3.22127261990  + 78283.13536813887 *T)
    sumY0  +=       0.004 * cos(1.30518138816  + 26037.66270218649 *T)
    sumY0  +=       0.004 * cos(5.49977739982  + 23543.47432216530 *T)
    sumY0  +=       0.004 * cos(3.30544028498  + 28632.81959595010 *T)
    sumY0  +=       0.003 * cos(0.77597702447  + 26279.59522516930 *T)
    sumY0  +=       0.003 * cos(1.74605535315  + 25896.69869294610 *T)
    sumY0  +=       0.004 * cos(2.60606940279  + 25998.66121121170 *T)
    sumY0  +=       0.004 * cos(6.19914828201  + 26177.63270690369 *T)
    sumY0  +=       0.003 * cos(3.02425517786  + 24079.10178493169 *T)
    sumY0  +=       0.003 * cos(1.26642446783  + 49851.10069181629 *T)
    sumY0  +=       0.004 * cos(4.75573366952  + 14235.79189565871 *T)
    sumY0  +=       0.004 * cos(4.04948401528  + 37940.50202245669 *T)
    sumY0  +=       0.003 * cos(1.62600802494  + 41494.79956406649 *T)
    sumY0  +=       0.004 * cos(2.15333731098  + 261988.65378531881 *T)
    sumY0  +=       0.003 * cos(4.80360212129  + 61560.40347374009 *T)
    sumY0  +=       0.004 * cos(4.57728207743  + 32217.44399856430 *T)
    sumY0  +=       0.003 * cos(1.10122747342  + 78243.79584633929 *T)
    sumY0  +=       0.004 * cos(1.13276960181  + 51995.97297034730 *T)
    sumY0  +=       0.003 * cos(5.60306507247  + 154308.90556274189 *T)
    sumY0  +=       0.003 * cos(4.90725652630  + 78398.06229179110 *T)
    sumY0  +=       0.003 * cos(1.65138805728  + 202420.39310008567 *T)
    sumY0  +=       0.003 * cos(3.28257147962  + 79117.14962395809 *T)
    sumY0  +=       0.004 * cos(0.10028058442  + 28420.85571696270 *T)
    sumY0  +=       0.003 * cos(6.21316075804  + 313054.59388140670 *T)
    sumY0  +=       0.003 * cos(0.25268995906  + 26114.44527885750 *T)
    sumY0  +=       0.003 * cos(2.26934241856  + 26061.84863925790 *T)
    sumY0  +=       0.003 * cos(1.45372655722  + 116783.41139921108 *T)
    sumY0  +=       0.003 * cos(4.65338616524  + 2169.00429976610 *T)
    sumY0  +=       0.003 * cos(4.91794356627  + 32131.88790546609 *T)
    sumY0  +=       0.003 * cos(4.52136661001  + 68241.62832713968 *T)
    sumY0  +=       0.004 * cos(0.55663803534  + 49424.50250094030 *T)
    sumY0  +=       0.003 * cos(1.53998283633  + 27993.61172399810 *T)
    sumY0  +=       0.004 * cos(1.48359617562  + 156954.26085780468 *T)
    sumY0  +=       0.004 * cos(5.24222478531  + 52021.78365518009 *T)
    sumY0  +=       0.004 * cos(4.16132549519  + 192828.85135471387 *T)
    sumY0  +=       0.003 * cos(4.98600736246  + 25345.15689852510 *T)
    sumY0  +=       0.003 * cos(0.24502022671  + 87648.79425028129 *T)
    sumY0  +=       0.003 * cos(2.25580297503  + 105410.75067900248 *T)
    sumY0  +=       0.003 * cos(2.35998158508  + 130446.38808425447 *T)
    sumY0  +=       0.003 * cos(1.34307352948  + 45290.41257573408 *T)
    sumY0  +=       0.003 * cos(0.38732763509  + 251775.36823910783 *T)
    sumY0  +=       0.004 * cos(5.98737648746  + 50952.23225985329 *T)
    sumY0  +=       0.003 * cos(6.14909241029  + 23421.09462541629 *T)
    sumY0  +=       0.003 * cos(2.65612527451  + 28755.19929269911 *T)
    sumY0  +=       0.003 * cos(5.55350394783  + 163297.88563320728 *T)
    sumY0  +=       0.003 * cos(5.04626720292  + 26727.55661040469 *T)
    sumY0  +=       0.004 * cos(5.00773877934  + 24711.88552424489 *T)
    sumY0  +=       0.003 * cos(5.25171311588  + 111590.04434041328 *T)
    sumY0  +=       0.004 * cos(5.00370301079  + 50476.77088412869 *T)
    sumY0  +=       0.003 * cos(2.64378251337  + 52072.26791205130 *T)
    sumY0  +=       0.003 * cos(2.04712691944  + 77622.53968046710 *T)
    sumY0  +=       0.003 * cos(5.19234087353  + 52016.74836583970 *T)
    sumY0  +=       0.003 * cos(0.11683553157  + 25970.82709083750 *T)
    sumY0  +=       0.003 * cos(2.40519684605  + 26205.46682727790 *T)
    sumY0  +=       0.004 * cos(3.70857620718  + 27511.71169102070 *T)
    sumY0  +=       0.004 * cos(5.09664147762  + 24664.58222709470 *T)
    sumY0  +=       0.003 * cos(0.72953121469  + 33918.23949049249 *T)
    sumY0  +=       0.003 * cos(1.79250116293  + 18258.05442762290 *T)
    sumY0  +=       0.004 * cos(0.12046155373  + 112545.64408202188 *T)
    sumY0  +=       0.004 * cos(2.19368586548  + 61921.37065648850 *T)
    sumY0  +=       0.003 * cos(1.35496442859  + 51741.85072778809 *T)
    sumY0  +=       0.003 * cos(1.31287870388  + 222225.22039166847 *T)
    sumY0  +=       0.003 * cos(5.87130809057  + 78249.72614820449 *T)
    sumY0  +=       0.004 * cos(1.48902081585  + 158116.73556221247 *T)
    sumY0  +=       0.003 * cos(5.25522824422  + 73891.58924046489 *T)
    sumY0  +=       0.003 * cos(0.31258950206  + 26387.27335332690 *T)
    sumY0  +=       0.003 * cos(4.52577588776  + 12725.69725225850 *T)
    sumY0  +=       0.003 * cos(3.47716392063  + 26984.22641571310 *T)
    sumY0  +=       0.003 * cos(5.32805376417  + 25192.06750240230 *T)
    sumY0  +=       0.003 * cos(5.05935500666  + 103498.17236706128 *T)
    sumY0  +=       0.003 * cos(5.46073958753  + 79487.77108298469 *T)
    sumY0  +=       0.003 * cos(3.01230527787  + 50579.37602338029 *T)
    sumY0  +=       0.003 * cos(5.01426493917  + 143980.69309287865 *T)
    sumY0  +=       0.003 * cos(4.06152997793  + 91805.37444469729 *T)
    sumY0  +=       0.003 * cos(1.54880317183  + 50490.99797813029 *T)
    sumY0  +=       0.003 * cos(1.42463612574  + 54879.66625530750 *T)
    sumY0  +=       0.003 * cos(3.12587906085  + 55484.51599601230 *T)
    sumY0  +=       0.003 * cos(6.21651664332  + 2332.95257538850 *T)
    sumY0  +=       0.003 * cos(6.10767041127  + 1383.13116433011 *T)
    sumY0  +=       0.003 * cos(0.21479933904  + 128106.56313248246 *T)
    sumY0  +=       0.003 * cos(5.41781200857  + 1471.99684454710 *T)
    sumY0  +=       0.003 * cos(3.15521647543  + 146505.82538682927 *T)
    sumY0  +=       0.004 * cos(5.80466382140  + 102232.60488843489 *T)
    sumY0  +=       0.003 * cos(3.47413746108  + 24510.80341660990 *T)
    sumY0  +=       0.003 * cos(5.33108022372  + 27665.49050150550 *T)
    sumY0  +=       0.004 * cos(3.03503390629  + 26365.18195279910 *T)
    sumY0  +=       0.004 * cos(5.77018377851  + 25811.11196531629 *T)
    sumY0  +=       0.003 * cos(4.73983619092  + 106263.05586699749 *T)
    sumY0  +=       0.003 * cos(1.92656985885  + 50601.20429934970 *T)
    sumY0  +=       0.003 * cos(5.01250465817  + 179967.11098302988 *T)
    sumY0  +=       0.003 * cos(3.10770566817  + 210921.73841929680 *T)
    sumY0  +=       0.003 * cos(6.02863621669  + 26202.54606597110 *T)
    sumY0  +=       0.003 * cos(1.18743831658  + 2821.07496160971 *T)
    sumY0  +=       0.003 * cos(4.69341673957  + 5020.16230489671 *T)
    sumY0  +=       0.003 * cos(3.22922107727  + 229463.89598326848 *T)
    sumY0  +=       0.003 * cos(1.31139605087  + 71378.31571730827 *T)
    sumY0  +=       0.003 * cos(6.13912071583  + 29001.03093892850 *T)
    sumY0  +=       0.003 * cos(2.66609696897  + 23175.26297918689 *T)
    sumY0  +=       0.003 * cos(2.25383916432  + 15406.40878752529 *T)
    sumY0  +=       0.003 * cos(2.75613383037  + 25940.06823463139 *T)
    sumY0  +=       0.003 * cos(0.81910869714  + 220026.13304838145 *T)
    sumY0  +=       0.003 * cos(5.30888840501  + 963.64652045490 *T)
    sumY0  +=       0.003 * cos(1.30036577443  + 113455.76337789788 *T)
    sumY0  +=       0.003 * cos(6.07147026831  + 49201.44014602090 *T)
    sumY0  +=       0.003 * cos(4.25244825934  + 143165.21296012949 *T)
    sumY0  +=       0.003 * cos(0.81593599477  + 51735.22481575429 *T)
    sumY0  +=       0.003 * cos(1.27452500023  + 25914.69838767311 *T)
    sumY0  +=       0.003 * cos(1.24750737739  + 26261.59553044229 *T)
    sumY0  +=       0.003 * cos(3.64962054328  + 51745.51975649279 *T)
    sumY0  +=       0.002 * cos(1.46486531145  + 25554.52384069999 *T)
    sumY0  +=       0.002 * cos(1.05716706617  + 26621.77007741539 *T)
    sumY0  +=       0.003 * cos(3.63673845964  + 25530.69622085730 *T)
    sumY0  +=       0.003 * cos(5.16847922516  + 26645.59769725810 *T)
    sumY0  +=       0.002 * cos(5.30160804243  + 99979.49238203908 *T)
    sumY0  +=       0.003 * cos(3.46486175917  + 25626.81052760331 *T)
    sumY0  +=       0.003 * cos(5.34035592563  + 26549.48339051209 *T)
    sumY0  +=       0.003 * cos(4.02837363457  + 23976.49664568010 *T)
    sumY0  +=       0.003 * cos(3.77281945049  + 20956.50587500010 *T)
    sumY0  +=       0.003 * cos(5.03239823431  + 31219.78804311530 *T)
    sumY0  +=       0.002 * cos(5.45450198288  + 242985.64742485987 *T)
    sumY0  +=       0.002 * cos(6.25807160883  + 133767.40754666767 *T)
    sumY0  +=       0.003 * cos(2.87262383804  + 283524.60342983418 *T)
    sumY0  +=       0.002 * cos(2.72683369817  + 127790.81706491447 *T)
    sumY0  +=       0.002 * cos(2.72683369816  + 127791.30469988148 *T)
    sumY0  +=       0.002 * cos(3.86040326075  + 71378.80335227528 *T)
    sumY0  +=       0.002 * cos(3.01852194681  + 21202.33752122950 *T)
    sumY0  +=       0.002 * cos(5.78669573799  + 30973.95639688590 *T)
    sumY0  +=       0.003 * cos(1.89994220844  + 2353.10997125530 *T)
    sumY0  +=       0.003 * cos(0.91363958260  + 76358.48847726568 *T)
    sumY0  +=       0.003 * cos(1.60976861498  + 24292.40088036969 *T)
    sumY0  +=       0.003 * cos(3.61818924103  + 157636.55358405507 *T)
    sumY0  +=       0.003 * cos(5.54528574622  + 25361.66493918750 *T)
    sumY0  +=       0.003 * cos(3.25993193858  + 26814.62897892790 *T)
    sumY0  +=       0.002 * cos(5.95736615747  + 57369.41715942969 *T)
    sumY0  +=       0.003 * cos(3.84445897700  + 52187.09580089580 *T)
    sumY0  +=       0.003 * cos(5.60002249822  + 49953.70583106789 *T)
    sumY0  +=       0.002 * cos(6.18039190992  + 110396.08496515568 *T)
    sumY0  +=       0.002 * cos(0.99675798287  + 273311.31788362318 *T)
    sumY0  +=       0.002 * cos(2.52419479264  + 7994.28462454070 *T)
    sumY0  +=       0.002 * cos(1.06541336121  + 31127.73520737070 *T)
    sumY0  +=       0.002 * cos(1.45661901641  + 21048.55871074470 *T)
    sumY0  +=       0.002 * cos(4.88076203498  + 27133.30179524529 *T)
    sumY0  +=       0.002 * cos(2.24228548233  + 289185.93547898642 *T)
    sumY0  +=       0.002 * cos(2.28218167520  + 5634.78991186890 *T)
    sumY0  +=       0.002 * cos(3.33192280402  + 108903.32224587208 *T)
    sumY0  +=       0.003 * cos(4.49860650606  + 128857.80017707148 *T)
    sumY0  +=       0.002 * cos(0.10562255755  + 5687.38655146850 *T)
    sumY0  +=       0.003 * cos(3.16048516532  + 21650.63275068411 *T)
    sumY0  +=       0.003 * cos(5.64473251948  + 30525.66116743129 *T)
    sumY0  +=       0.003 * cos(5.20937667539  + 117.56368570370 *T)
    sumY0  +=       0.002 * cos(0.61090203417  + 29266.29274962530 *T)
    sumY0  +=       0.002 * cos(1.08851105464  + 24176.45984087349 *T)
    sumY0  +=       0.002 * cos(1.68398359832  + 102129.99974918328 *T)
    sumY0  +=       0.003 * cos(2.77323027693  + 53338.52480503969 *T)
    sumY0  +=       0.003 * cos(1.56140703602  + 67608.51339047348 *T)
    sumY0  +=       0.002 * cos(1.64261819628  + 267649.98583447101 *T)
    sumY0  +=       0.002 * cos(0.35074176404  + 544.16187657970 *T)
    sumY0  +=       0.002 * cos(5.34557611286  + 54862.77219517310 *T)
    sumY0  +=       0.002 * cos(0.00658372278  + 162188.50708111387 *T)
    sumY0  +=       0.002 * cos(1.24227084874  + 123554.12200045666 *T)
    sumY0  +=       0.003 * cos(5.22200424544  + 47803.68609889069 *T)
    sumY0  +=       0.002 * cos(2.25109608564  + 112546.13171698888 *T)
    sumY0  +=       0.002 * cos(2.34818806724  + 71024.79001563908 *T)
    sumY0  +=       0.003 * cos(1.82596797402  + 24971.65485996349 *T)
    sumY0  +=       0.003 * cos(0.69606440360  + 27204.63905815190 *T)
    sumY0  +=       0.002 * cos(3.44738180328  + 26165.89750304159 *T)
    sumY0  +=       0.003 * cos(2.81683191154  + 14477.59500068350 *T)
    sumY0  +=       0.002 * cos(1.34089550191  + 97466.21885888244 *T)
    sumY0  +=       0.002 * cos(3.66953582983  + 8482.40701076191 *T)
    sumY0  +=       0.003 * cos(1.55932656916  + 206584.70508969869 *T)
    sumY0  +=       0.003 * cos(1.85898364174  + 25881.47377565349 *T)
    sumY0  +=       0.002 * cos(2.20662264774  + 105306.96849042187 *T)
    sumY0  +=       0.003 * cos(1.29582724688  + 7768.12273921109 *T)
    sumY0  +=       0.002 * cos(2.32161458508  + 263097.54470244516 *T)
    sumY0  +=       0.002 * cos(5.46320723867  + 263098.03233741224 *T)
    sumY0  +=       0.002 * cos(1.96333630845  + 90988.91904201408 *T)
    sumY0  +=       0.003 * cos(4.58684494130  + 5814.86704215349 *T)
    sumY0  +=       0.003 * cos(5.57796892253  + 49976.96275734490 *T)
    sumY0  +=       0.003 * cos(1.11744370489  + 27161.75598324850 *T)
    sumY0  +=       0.002 * cos(4.55377851088  + 27151.46104250999 *T)
    sumY0  +=       0.003 * cos(0.04149717596  + 29051.02756083270 *T)
    sumY0  +=       0.003 * cos(2.48053520166  + 23125.26635728270 *T)
    sumY0  +=       0.002 * cos(1.84878451995  + 32973.29684898851 *T)
    sumY0  +=       0.003 * cos(5.15950079042  + 134.35286706851 *T)
    sumY0  +=       0.002 * cos(5.35359251167  + 78484.36588464488 *T)
    sumY0  +=       0.002 * cos(4.84459493300  + 50373.67810991009 *T)
    sumY0  +=       0.003 * cos(1.01446892629  + 246114.03618995569 *T)
    sumY0  +=       0.002 * cos(0.53517853304  + 19336.61848374489 *T)
    sumY0  +=       0.002 * cos(1.40171752369  + 25014.53793486689 *T)
    sumY0  +=       0.002 * cos(5.17281822358  + 156534.77621392949 *T)
    sumY0  +=       0.003 * cos(2.73574689557  + 78115.87451777978 *T)
    sumY0  +=       0.002 * cos(4.99251162267  + 51760.49761001979 *T)
    sumY0  +=       0.002 * cos(2.81513963677  + 25867.24668165189 *T)
    sumY0  +=       0.002 * cos(4.97557986159  + 26089.14379679899 *T)
    sumY0  +=       0.002 * cos(3.55523971481  + 30486.32164563170 *T)
    sumY0  +=       0.002 * cos(5.24997796999  + 21689.97227248370 *T)
    sumY0  +=       0.003 * cos(1.74089824235  + 78112.90557236319 *T)
    sumY0  +=       0.002 * cos(2.79225536164  + 1898.59503542310 *T)
    sumY0  +=       0.002 * cos(3.64384685304  + 235746.97183325983 *T)
    sumY0  +=       0.002 * cos(3.14876741131  + 26151.88285736110 *T)
    sumY0  +=       0.002 * cos(5.65645027349  + 26024.41106075430 *T)
    sumY0  +=       0.002 * cos(1.05800481517  + 50049.68505828569 *T)
    sumY0  +=       0.003 * cos(2.73644560356  + 42116.54336490570 *T)
    sumY0  +=       0.002 * cos(5.11212517082  + 61165.51780606429 *T)
    sumY0  +=       0.002 * cos(5.88527469971  + 249268.72327485128 *T)
    sumY0  +=       0.002 * cos(2.69366265342  + 184526.76529172008 *T)
    sumY0  +=       0.003 * cos(4.49394290972  + 26226.66445592840 *T)
    sumY0  +=       0.003 * cos(4.31127477508  + 25949.62946218700 *T)
    sumY0  +=       0.002 * cos(5.99028756442  + 223180.33249831005 *T)
    sumY0  +=       0.002 * cos(2.84869491083  + 223180.82013327704 *T)
    sumY0  +=       0.003 * cos(0.50210111414  + 65851.58035270809 *T)
    sumY0  +=       0.002 * cos(1.76617343740  + 4525.89899474250 *T)
    sumY0  +=       0.003 * cos(0.33246489326  + 26376.03481912750 *T)
    sumY0  +=       0.003 * cos(2.18956748436  + 25800.25909898790 *T)
    sumY0  +=       0.002 * cos(2.76517897745  + 26049.52628845289 *T)
    sumY0  +=       0.002 * cos(3.57038365323  + 4398.41850405751 *T)
    sumY0  +=       0.003 * cos(2.22716383153  + 23581.50199480110 *T)
    sumY0  +=       0.003 * cos(0.29486854609  + 28594.79192331430 *T)
    sumY0  +=       0.002 * cos(1.98440604953  + 1593.24886603040 *T)
    sumY0  +=       0.002 * cos(4.71041878186  + 78267.34694259708 *T)
    sumY0  +=       0.002 * cos(2.49584762828  + 54508.75885853690 *T)
    sumY0  +=       0.002 * cos(5.88257404550  + 78476.76470267709 *T)
    sumY0  +=       0.002 * cos(5.12850312405  + 1280.03839011150 *T)
    sumY0  +=       0.002 * cos(2.67253805962  + 35077.12702952309 *T)
    sumY0  +=       0.003 * cos(6.14642740727  + 26124.79552198719 *T)
    sumY0  +=       0.003 * cos(2.65879027753  + 26051.49839612820 *T)
    sumY0  +=       0.002 * cos(3.30021458746  + 158745.93213614845 *T)
    sumY0  +=       0.002 * cos(3.91497249728  + 52367.49836674349 *T)
    sumY0  +=       0.002 * cos(1.49263134542  + 41520.61024889929 *T)
    sumY0  +=       0.002 * cos(2.56976964633  + 51950.22083222170 *T)
    sumY0  +=       0.002 * cos(0.98860870622  + 28133.45549185990 *T)
    sumY0  +=       0.002 * cos(1.53342367140  + 24042.83842625550 *T)
    sumY0  +=       0.003 * cos(0.60080693478  + 51753.38406301899 *T)
    sumY0  +=       0.003 * cos(4.58119790245  + 25633.23759253039 *T)
    sumY0  +=       0.003 * cos(4.22401978235  + 26543.05632558499 *T)
    sumY0  +=       0.002 * cos(2.40433125154  + 24999.54380245671 *T)
    sumY0  +=       0.002 * cos(0.11770112608  + 27176.75011565869 *T)
    sumY0  +=       0.002 * cos(1.87923904841  + 309.03450517230 *T)
    sumY0  +=       0.002 * cos(0.83358735040  + 1155.60497489050 *T)
    sumY0  +=       0.002 * cos(5.05439501768  + 236702.57157486849 *T)
    sumY0  +=       0.002 * cos(4.57470028824  + 76887.69180739328 *T)
    sumY0  +=       0.002 * cos(5.69921176024  + 4084.23232418790 *T)
    sumY0  +=       0.002 * cos(3.16153790522  + 24601.67920302549 *T)
    sumY0  +=       0.002 * cos(4.37985667611  + 27818.78167702329 *T)
    sumY0  +=       0.002 * cos(3.83767204936  + 2198.84352580350 *T)
    sumY0  +=       0.002 * cos(2.96055155756  + 24292.88851533670 *T)
    sumY0  +=       0.002 * cos(3.15659811410  + 22759.52366781050 *T)
    sumY0  +=       0.002 * cos(1.34122714460  + 27143.59673598380 *T)
    sumY0  +=       0.002 * cos(3.27620714425  + 25973.22003540546 *T)
    sumY0  +=       0.002 * cos(1.71257292852  + 10021.59346261590 *T)
    sumY0  +=       0.002 * cos(6.23156121709  + 20618.26317601710 *T)
    sumY0  +=       0.002 * cos(2.57365646771  + 31558.03074209830 *T)
    sumY0  +=       0.002 * cos(0.76870131372  + 2854.88419139370 *T)
    sumY0  +=       0.002 * cos(0.41751652575  + 51953.18977763829 *T)
    sumY0  +=       0.002 * cos(1.98684490521  + 51951.21766996299 *T)
    sumY0  +=       0.002 * cos(0.00103411994  + 52249.34722649089 *T)
    sumY0  +=       0.002 * cos(3.27898093725  + 138633.54722359608 *T)
    sumY0  +=       0.002 * cos(3.56496173333  + 225687.46509753368 *T)
    sumY0  +=       0.002 * cos(3.13338598533  + 138319.36104372647 *T)
    sumY0  +=       0.002 * cos(3.76046136390  + 32768.88417749029 *T)
    sumY0  +=       0.002 * cos(5.04336913197  + 24079.58941989870 *T)
    sumY0  +=       0.002 * cos(3.14356989843  + 52202.34842043170 *T)
    sumY0  +=       0.002 * cos(0.25753538380  + 28206.42318446909 *T)
    sumY0  +=       0.002 * cos(3.39136790755  + 180496.80194812448 *T)
    sumY0  +=       0.002 * cos(0.02421731187  + 78259.58427188109 *T)
    sumY0  +=       0.002 * cos(3.85481165639  + 26624.46383618609 *T)
    sumY0  +=       0.002 * cos(4.79199372149  + 23549.89845480410 *T)
    sumY0  +=       0.002 * cos(1.94407126016  + 26293.84487252789 *T)
    sumY0  +=       0.002 * cos(1.12674295858  + 4601.70630147949 *T)
    sumY0  +=       0.002 * cos(4.65935637649  + 21716.27059228350 *T)
    sumY0  +=       0.002 * cos(2.09856828971  + 330.86278114170 *T)
    sumY0  +=       0.002 * cos(3.08659098600  + 51219.27330969429 *T)
    sumY0  +=       0.002 * cos(0.33777720887  + 78256.35206023828 *T)
    sumY0  +=       0.002 * cos(5.15360975027  + 50063.91215228729 *T)
    sumY0  +=       0.002 * cos(2.99279377683  + 132658.02899457427 *T)
    sumY0  +=       0.002 * cos(1.71072737383  + 96.22304470130 *T)
    sumY0  +=       0.002 * cos(5.95938200395  + 1265.32366114289 *T)
    sumY0  +=       0.002 * cos(0.92029934107  + 52823.06093394670 *T)
    sumY0  +=       0.002 * cos(2.83880526955  + 78270.57915423988 *T)
    sumY0  +=       0.002 * cos(0.69832785802  + 7392.21058460129 *T)
    sumY0  +=       0.002 * cos(2.67350091104  + 61165.03017109729 *T)
    sumY0  +=       0.002 * cos(3.72029671343  + 2544.42881940890 *T)
    sumY0  +=       0.002 * cos(3.54241730321  + 26421.51526486010 *T)
    sumY0  +=       0.002 * cos(6.04496787393  + 2655.81218995730 *T)
    sumY0  +=       0.002 * cos(5.50264063039  + 24705.94902657310 *T)
    sumY0  +=       0.002 * cos(3.30257705441  + 27470.34489154230 *T)
    sumY0  +=       0.002 * cos(1.43501438033  + 1353.50492753330 *T)
    sumY0  +=       0.002 * cos(4.83145382265  + 1603.54380676890 *T)
    sumY0  +=       0.002 * cos(5.03780638681  + 40738.45907867528 *T)
    sumY0  +=       0.002 * cos(4.14425166593  + 30460.02332583190 *T)
    sumY0  +=       0.002 * cos(5.87281396000  + 4578.49563434210 *T)
    sumY0  +=       0.002 * cos(1.71607024251  + 78113.90241010448 *T)
    sumY0  +=       0.002 * cos(2.62817899928  + 32839.18779940350 *T)
    sumY0  +=       0.002 * cos(4.84613238692  + 14297.51787039890 *T)
    sumY0  +=       0.002 * cos(3.95908529788  + 37878.77604771650 *T)
    sumY0  +=       0.002 * cos(4.63463254085  + 104501.41958091488 *T)
    sumY0  +=       0.002 * cos(4.18177446903  + 25138.48371512089 *T)
    sumY0  +=       0.002 * cos(5.77634249731  + 44131.52503670349 *T)
    sumY0  +=       0.002 * cos(5.67506076203  + 104358.48229581409 *T)
    sumY0  +=       0.002 * cos(3.82438567775  + 23336.59935936610 *T)
    sumY0  +=       0.002 * cos(0.68279302416  + 23336.11172439909 *T)
    sumY0  +=       0.002 * cos(1.33260638073  + 224.10097821840 *T)
    sumY0  +=       0.002 * cos(2.06879035628  + 51547.19851458179 *T)
    sumY0  +=       0.002 * cos(4.73770456468  + 202.00957769060 *T)
    sumY0  +=       0.002 * cos(1.59611191109  + 202.49721265760 *T)
    sumY0  +=       0.002 * cos(4.68695414047  + 31108.06544647091 *T)
    sumY0  +=       0.002 * cos(4.11826354432  + 21068.22847164450 *T)
    sumY0  +=       0.002 * cos(4.53666662059  + 117076.82218358829 *T)
    sumY0  +=       0.002 * cos(4.53666662059  + 117077.30981855528 *T)
    sumY0  +=       0.002 * cos(0.67052215512  + 24814.97829846349 *T)
    sumY0  +=       0.002 * cos(0.41491214788  + 27299.48802941670 *T)
    sumY0  +=       0.002 * cos(2.10712022974  + 24876.80588869870 *T)
    sumY0  +=       0.002 * cos(4.40848896505  + 26086.17485138239 *T)
    sumY0  +=       0.002 * cos(4.17219070503  + 105418.35186097027 *T)
    sumY0  +=       0.002 * cos(1.12019984719  + 49534.70882215970 *T)
    sumY0  +=       0.002 * cos(2.68381144033  + 107679.99204006046 *T)
    sumY0  +=       0.002 * cos(6.14277431051  + 19337.10611871190 *T)
    sumY0  +=       0.002 * cos(0.23628622399  + 24957.99301147190 *T)
    sumY0  +=       0.002 * cos(2.28574615363  + 27218.30090664350 *T)
    sumY0  +=       0.002 * cos(5.03978204688  + 66826.36222024946 *T)
    sumY0  +=       0.002 * cos(1.54597568017  + 27257.73521046630 *T)
    sumY0  +=       0.002 * cos(3.11893314385  + 26197.86564531009 *T)
    sumY0  +=       0.002 * cos(3.83163915021  + 151199.69892357889 *T)
    sumY0  +=       0.002 * cos(0.84426255302  + 426.15259399750 *T)
    sumY0  +=       0.002 * cos(3.62580652226  + 76784.59903317469 *T)
    sumY0  +=       0.002 * cos(0.46652350366  + 179172.50322857045 *T)
    sumY0  +=       0.002 * cos(4.60919202058  + 168959.21768235945 *T)
    sumY0  +=       0.002 * cos(1.46759936699  + 168959.70531732647 *T)
    sumY0  +=       0.002 * cos(1.33905329827  + 7829.84871395129 *T)
    sumY0  +=       0.002 * cos(4.50925409872  + 151975.22153490287 *T)
    sumY0  +=       0.002 * cos(3.99530485240  + 20235.36664379391 *T)
    sumY0  +=       0.002 * cos(4.80991283240  + 31940.92727432150 *T)
    sumY0  +=       0.002 * cos(5.05502953319  + 1585.38455950421 *T)
    sumY0  +=       0.002 * cos(4.82744008834  + 76.02225379210 *T)
    sumY0  +=       0.002 * cos(6.07753628311  + 1581.71553079950 *T)
    sumY0  +=       0.002 * cos(5.83182532562  + 52643.52745601930 *T)
    sumY0  +=       0.002 * cos(5.61086631842  + 130433.13096645444 *T)
    sumY0  +=       0.002 * cos(3.78215414688  + 79956.11891170849 *T)
    sumY0  +=       0.002 * cos(2.95202880616  + 190809.35350674443 *T)
    sumY0  +=       0.002 * cos(0.51510001678  + 155417.79647986827 *T)
    sumY0  +=       0.002 * cos(3.82552098141  + 58219.79104704029 *T)
    sumY0  +=       0.002 * cos(4.10273299229  + 299399.22102519736 *T)
    sumY0  +=       0.002 * cos(5.54210405976  + 51738.40620949199 *T)
    sumY0  +=       0.002 * cos(1.39992677247  + 2045.55235028570 *T)
    sumY0  +=       0.002 * cos(2.83807224501  + 27953.37836157530 *T)
    sumY0  +=       0.002 * cos(5.96714543979  + 24222.91555654010 *T)
    sumY0  +=       0.002 * cos(6.18845161198  + 164721.45036517028 *T)
    sumY0  +=       0.002 * cos(4.74202892331  + 28441.01311282950 *T)
    sumY0  +=       0.002 * cos(4.06318876149  + 23735.28080528590 *T)
    sumY0  +=       0.002 * cos(3.36973154287  + 26948.45688781050 *T)
    sumY0  +=       0.002 * cos(5.43548614193  + 25227.83703030489 *T)
    sumY0  +=       0.002 * cos(3.83932556111  + 148532.64658993747 *T)
    sumY0  +=       0.002 * cos(0.69773290752  + 148533.13422490444 *T)
    sumY0  +=       0.002 * cos(3.08932581659  + 77307.17645126847 *T)
    sumY0  +=       0.002 * cos(4.53508525261  + 191.69208359510 *T)
    sumY0  +=       0.002 * cos(1.91266463584  + 23976.00901071309 *T)
    sumY0  +=       0.002 * cos(2.51070394013  + 206370.27255720508 *T)
    sumY0  +=       0.002 * cos(3.48548328265  + 4268.33799567710 *T)
    sumY0  +=       0.002 * cos(0.27071972210  + 23962.26955167850 *T)
    sumY0  +=       0.002 * cos(3.41231237569  + 23961.78191671149 *T)
    sumY0  +=       0.002 * cos(2.77965182012  + 3776.67470321830 *T)
    sumY0  +=       0.002 * cos(0.63445271713  + 24918.55870764909 *T)
    sumY0  +=       0.002 * cos(4.78488159185  + 102769.40940053028 *T)
    sumY0  +=       0.002 * cos(0.55778477986  + 25889.07495762130 *T)
    sumY0  +=       0.002 * cos(1.96424759776  + 26287.21896049409 *T)
    sumY0  +=       0.002 * cos(3.23447204926  + 2648.21100798950 *T)
    sumY0  +=       0.002 * cos(4.14592395979  + 80432.71372448868 *T)
    sumY0  +=       0.002 * cos(1.34549805675  + 13989.96024942930 *T)
    sumY0  +=       0.002 * cos(1.17653432087  + 38186.33366868610 *T)
    sumY0  +=       0.002 * cos(2.04577034444  + 8194.51915069210 *T)
    sumY0  +=       0.002 * cos(5.70902071155  + 237009.64156087104 *T)
    sumY0  +=       0.002 * cos(3.82569277203  + 24863.84148331209 *T)
    sumY0  +=       0.002 * cos(5.50674253073  + 76064.86589891909 *T)
    sumY0  +=       0.002 * cos(1.39914737585  + 71492.75500599349 *T)
    sumY0  +=       0.002 * cos(6.18696209768  + 30831.54872222810 *T)
    sumY0  +=       0.002 * cos(2.61825558712  + 21344.74519588730 *T)
    sumY0  +=       0.002 * cos(1.07023698053  + 124663.98818751708 *T)
    sumY0  +=       0.001 * cos(5.98706010174  + 101702.91392334028 *T)
    sumY0  +=       0.002 * cos(1.35473982240  + 105404.12476696869 *T)
    sumY0  +=       0.002 * cos(3.48042408913  + 307.31380348609 *T)
    sumY0  +=       0.002 * cos(5.91402019226  + 184834.32291268968 *T)
    sumY0  +=       0.002 * cos(5.91402019226  + 184833.83527772260 *T)
    sumY0  +=       0.001 * cos(0.49434611754  + 124156.19604039608 *T)
    sumY0  +=       0.001 * cos(4.86177343922  + 22909.51353352309 *T)
    sumY0  +=       0.001 * cos(1.72018078563  + 22910.00116849010 *T)
    sumY0  +=       0.001 * cos(5.79706734000  + 958.87904247409 *T)
    sumY0  +=       0.002 * cos(2.31704185815  + 77196.97013004909 *T)
    sumY0  +=       0.002 * cos(4.18763806284  + 9591.29792788829 *T)
    sumY0  +=       0.002 * cos(2.60184968404  + 189385.78877478145 *T)
    sumY0  +=       0.001 * cos(5.29048508788  + 24919.24812201110 *T)
    sumY0  +=       0.001 * cos(3.51473259692  + 27257.04579610430 *T)
    sumY0  +=       0.002 * cos(5.03520614299  + 25440.64849077589 *T)
    sumY0  +=       0.001 * cos(5.16795059538  + 26010.39641507380 *T)
    sumY0  +=       0.002 * cos(5.13757006415  + 43693.88690735350 *T)
    sumY0  +=       0.002 * cos(5.04394896979  + 25014.05029989989 *T)
    sumY0  +=       0.001 * cos(3.77902331302  + 9385.08482555870 *T)
    sumY0  +=       0.002 * cos(6.17372295311  + 26202.09861277591 *T)
    sumY0  +=       0.001 * cos(3.35135060198  + 52499.55551728929 *T)
    sumY0  +=       0.002 * cos(5.03234764639  + 14650.55593710108 *T)
    sumY0  +=       0.002 * cos(0.37520258538  + 78690.06379811508 *T)
    sumY0  +=       0.001 * cos(1.65843797188  + 5039.34443082949 *T)
    sumY0  +=       0.002 * cos(2.53532060227  + 2214.98690507971 *T)
    sumY0  +=       0.001 * cos(5.07384686444  + 77211.19722405069 *T)
    sumY0  +=       0.001 * cos(4.11905908237  + 233202.29919636747 *T)
    sumY0  +=       0.002 * cos(3.92094587643  + 21516.52370109911 *T)
    sumY0  +=       0.002 * cos(4.88427180837  + 30659.77021701629 *T)
    sumY0  +=       0.002 * cos(0.59200954056  + 621.49998335569 *T)
    sumY0  +=       0.002 * cos(2.70800261719  + 25764.15390743329 *T)
    sumY0  +=       0.001 * cos(3.75051661568  + 26.54213728330 *T)
    sumY0  +=       0.001 * cos(3.87625546504  + 226796.35601466001 *T)
    sumY0  +=       0.001 * cos(3.87625546504  + 226796.84364962703 *T)
    sumY0  +=       0.002 * cos(5.04685227536  + 28839.69455874929 *T)
    sumY0  +=       0.002 * cos(3.93268001276  + 102225.97897640108 *T)
    sumY0  +=       0.002 * cos(6.27298790851  + 129971.79453500008 *T)
    sumY0  +=       0.002 * cos(5.63544803962  + 27139.92770727909 *T)
    sumY0  +=       0.001 * cos(1.05389010615  + 50807.38984778689 *T)
    sumY0  +=       0.001 * cos(1.45067645500  + 78187.19953596349 *T)
    sumY0  +=       0.002 * cos(3.49292360736  + 51472.41691601449 *T)

    # MERCURY Y Coord
    sumY1  = 0
    sumY1  +=  328639.517 * cos(4.46949126315  + 0.24381748350 *T)
    sumY1  +=  106107.047 * cos(4.34458837257  + 52176.05010063190 *T)
    sumY1  +=   32448.440 * cos(1.11324531456  + 78263.95324220609 *T)
    sumY1  +=    9699.418 * cos(0.71696945020  + 26087.65932409069 *T)
    sumY1  +=    8808.490 * cos(4.16435393159  + 104351.85638378028 *T)
    sumY1  +=    4833.058 * cos(4.93174326512  + 26088.14695905770 *T)
    sumY1  +=    2334.329 * cos(0.93209482754  + 130439.75952535449 *T)
    sumY1  +=     890.250 * cos(4.18838297765  + 52175.56246566489 *T)
    sumY1  +=     615.643 * cos(3.98294823708  + 156527.66266692869 *T)
    sumY1  +=     147.225 * cos(1.38491342638  + 78263.46560723908 *T)
    sumY1  +=     162.460 * cos(0.75058057907  + 182615.56580850287 *T)
    sumY1  +=      69.502 * cos(1.34859955230  + 24978.76840696430 *T)
    sumY1  +=      63.529 * cos(1.17140322075  + 27197.52551115110 *T)
    sumY1  +=      42.967 * cos(3.80137602608  + 208703.46895007708 *T)
    sumY1  +=      32.514 * cos(4.75515678790  + 104351.36874881329 *T)
    sumY1  +=      17.597 * cos(4.40134253408  + 51066.67154853850 *T)
    sumY1  +=      13.287 * cos(4.22067115617  + 53285.42865272530 *T)
    sumY1  +=      11.758 * cos(1.75773369521  + 1109.13473460990 *T)
    sumY1  +=      12.502 * cos(6.11835337393  + 1059.62574767270 *T)
    sumY1  +=      11.468 * cos(5.06247506599  + 20426.81490990550 *T)
    sumY1  +=      11.394 * cos(0.56897531807  + 234791.37209165128 *T)
    sumY1  +=       8.221 * cos(3.74149753253  + 31749.47900820990 *T)
    sumY1  +=       8.067 * cos(1.73031128392  + 130439.27189038748 *T)
    sumY1  +=       6.176 * cos(3.89018803821  + 4552.19731454230 *T)
    sumY1  +=       4.784 * cos(4.97654366187  + 21536.19346199890 *T)
    sumY1  +=       4.858 * cos(1.16998032311  + 77154.57469011268 *T)
    sumY1  +=       4.269 * cos(1.82558089427  + 46514.71805147970 *T)
    sumY1  +=       4.412 * cos(3.53054557745  + 25132.54721744910 *T)
    sumY1  +=       3.922 * cos(0.80899339690  + 1589.31671276730 *T)
    sumY1  +=       4.017 * cos(2.84593275079  + 27147.52888924690 *T)
    sumY1  +=       3.763 * cos(2.60830060517  + 14765.48286075331 *T)
    sumY1  +=       4.285 * cos(2.70027245882  + 51116.66817044270 *T)
    sumY1  +=       3.901 * cos(5.26131971011  + 27043.74670066630 *T)
    sumY1  +=       3.221 * cos(5.81041475267  + 26617.83792415230 *T)
    sumY1  +=       3.080 * cos(3.11167500886  + 25558.45599396310 *T)
    sumY1  +=       3.030 * cos(3.61976417137  + 260879.27523322543 *T)
    sumY1  +=       2.907 * cos(4.94283048350  + 26068.47719815791 *T)
    sumY1  +=       2.906 * cos(3.86228296506  + 26107.81671995749 *T)
    sumY1  +=       2.383 * cos(0.50751116011  + 26084.26562369970 *T)
    sumY1  +=       2.381 * cos(2.01445897738  + 26092.02829441570 *T)
    sumY1  +=       2.069 * cos(6.19366221907  + 37410.81105736210 *T)
    sumY1  +=       2.028 * cos(3.87421613919  + 30640.10045611650 *T)
    sumY1  +=       2.088 * cos(4.91615750193  + 156527.17503196167 *T)
    sumY1  +=       1.998 * cos(0.93225615522  + 79373.33179429949 *T)
    sumY1  +=       1.880 * cos(4.93819019021  + 47624.09660357310 *T)
    sumY1  +=       1.580 * cos(0.12078980001  + 9104.15081160110 *T)
    sumY1  +=       1.575 * cos(5.88747787298  + 955.35592412510 *T)
    sumY1  +=       1.508 * cos(5.14798771540  + 19317.43635781211 *T)
    sumY1  +=       1.397 * cos(4.87619009368  + 72602.62119305388 *T)
    sumY1  +=       1.363 * cos(1.71613837922  + 50586.97720534809 *T)
    sumY1  +=       1.495 * cos(2.20742717983  + 12566.39551746630 *T)
    sumY1  +=       1.401 * cos(0.47782643773  + 426.84200835950 *T)
    sumY1  +=       1.215 * cos(2.36706136703  + 52705.74106572650 *T)
    sumY1  +=       1.310 * cos(0.91525566474  + 26095.26050605850 *T)
    sumY1  +=       1.272 * cos(1.65165069758  + 26081.03341205689 *T)
    sumY1  +=       1.476 * cos(1.42289389144  + 23869.38985487090 *T)
    sumY1  +=       1.169 * cos(3.92032031637  + 3442.81876244891 *T)
    sumY1  +=       1.103 * cos(5.93142816578  + 53235.43203082109 *T)
    sumY1  +=       1.375 * cos(4.22153989222  + 103242.47783168688 *T)
    sumY1  +=       1.244 * cos(0.25422201490  + 51220.45035902330 *T)
    sumY1  +=       1.238 * cos(1.09570749211  + 28306.90406324449 *T)
    sumY1  +=       0.967 * cos(2.00095625807  + 53131.64984224050 *T)
    sumY1  +=       0.986 * cos(3.65234674915  + 32858.85756030330 *T)
    sumY1  +=       0.911 * cos(5.86347762782  + 529.44714761109 *T)
    sumY1  +=       0.859 * cos(5.65763791955  + 40853.38600232750 *T)
    sumY1  +=       0.855 * cos(3.76170634741  + 27677.21985434150 *T)
    sumY1  +=       0.844 * cos(5.74887533238  + 25874.84786361970 *T)
    sumY1  +=       1.042 * cos(4.90279640329  + 22645.57201409229 *T)
    sumY1  +=       0.924 * cos(5.45463398257  + 640.14110379750 *T)
    sumY1  +=       0.810 * cos(3.07111874736  + 26301.44605449570 *T)
    sumY1  +=       0.807 * cos(0.38735610376  + 286967.17837479961 *T)
    sumY1  +=       0.875 * cos(0.56063513084  + 83925.28529135829 *T)
    sumY1  +=       0.874 * cos(3.97964673458  + 529.93478257810 *T)
    sumY1  +=       0.821 * cos(3.19953070235  + 36301.43250526870 *T)
    sumY1  +=       0.823 * cos(0.48663689329  + 11322.42028082089 *T)
    sumY1  +=       0.738 * cos(0.62901815542  + 56728.00359769070 *T)
    sumY1  +=       0.724 * cos(6.21294643252  + 18849.47136745770 *T)
    sumY1  +=       0.714 * cos(4.24977924185  + 6770.46678376209 *T)
    sumY1  +=       0.667 * cos(2.40282291967  + 43072.14310651430 *T)
    sumY1  +=       0.665 * cos(1.71193023872  + 52156.38033973211 *T)
    sumY1  +=       0.893 * cos(1.99786430454  + 27197.03787618409 *T)
    sumY1  +=       0.752 * cos(0.31686311777  + 39609.89840064910 *T)
    sumY1  +=       0.673 * cos(3.24236433600  + 1052.51220067191 *T)
    sumY1  +=       0.688 * cos(5.57273946980  + 15874.86141284670 *T)
    sumY1  +=       0.618 * cos(0.55635839524  + 5661.57586663570 *T)
    sumY1  +=       0.808 * cos(2.76566323402  + 1066.73929467350 *T)
    sumY1  +=       0.616 * cos(5.41694945887  + 51646.35913553730 *T)
    sumY1  +=       0.650 * cos(6.10338500409  + 25028.76502886850 *T)
    sumY1  +=       0.550 * cos(1.77567872746  + 182615.07817353585 *T)
    sumY1  +=       0.554 * cos(1.44404571039  + 2119.00767786191 *T)
    sumY1  +=       0.535 * cos(5.07158844336  + 52179.93143598990 *T)
    sumY1  +=       0.657 * cos(3.92169541273  + 29530.72190402310 *T)
    sumY1  +=       0.527 * cos(3.36241476084  + 26514.74514993370 *T)
    sumY1  +=       0.502 * cos(5.23707918628  + 38654.29865904050 *T)
    sumY1  +=       0.568 * cos(1.89274232427  + 24978.28077199729 *T)
    sumY1  +=       0.498 * cos(2.94458194590  + 63498.71419893629 *T)
    sumY1  +=       0.469 * cos(3.58596298240  + 110013.18843293248 *T)
    sumY1  +=       0.521 * cos(0.62400943535  + 52195.71986153169 *T)
    sumY1  +=       0.440 * cos(1.64399949118  + 98690.52433462808 *T)
    sumY1  +=       0.580 * cos(4.69289380268  + 7.35736448430 *T)
    sumY1  +=       0.438 * cos(2.54950574588  + 33326.82255065770 *T)
    sumY1  +=       0.530 * cos(1.70361729355  + 2218.51328670329 *T)
    sumY1  +=       0.483 * cos(2.07663666251  + 51749.45190975589 *T)
    sumY1  +=       0.434 * cos(3.55771117131  + 52172.16876527390 *T)
    sumY1  +=       0.408 * cos(3.22512461303  + 7238.43177411650 *T)
    sumY1  +=       0.390 * cos(5.33070151567  + 78793.64420730069 *T)
    sumY1  +=       0.374 * cos(0.44775435653  + 58946.76070187749 *T)
    sumY1  +=       0.404 * cos(4.44096026376  + 52168.93655363109 *T)
    sumY1  +=       0.356 * cos(6.23046854548  + 62389.33564684289 *T)
    sumY1  +=       0.373 * cos(5.69827259585  + 103292.47445359109 *T)
    sumY1  +=       0.326 * cos(2.72823636104  + 79323.33517239528 *T)
    sumY1  +=       0.393 * cos(0.98972344950  + 129330.38097326108 *T)
    sumY1  +=       0.385 * cos(3.94475257678  + 52183.16364763270 *T)
    sumY1  +=       0.316 * cos(4.54068234172  + 76674.88034692229 *T)
    sumY1  +=       0.390 * cos(4.17402259459  + 54394.80720481869 *T)
    sumY1  +=       0.323 * cos(4.47818749821  + 49957.29299644510 *T)
    sumY1  +=       0.299 * cos(3.00180406182  + 16983.75232997309 *T)
    sumY1  +=       0.328 * cos(2.47784503610  + 68050.66769599509 *T)
    sumY1  +=       0.314 * cos(3.34927785340  + 51536.15281431789 *T)
    sumY1  +=       0.367 * cos(3.28176962437  + 77308.35350059748 *T)
    sumY1  +=       0.316 * cos(1.92123984210  + 45405.33949938630 *T)
    sumY1  +=       0.319 * cos(2.66908913597  + 13656.10430865991 *T)
    sumY1  +=       0.266 * cos(4.93516999025  + 48733.47515566649 *T)
    sumY1  +=       0.289 * cos(1.34161862787  + 6.86972951729 *T)
    sumY1  +=       0.261 * cos(2.43104654910  + 16984.23996494010 *T)
    sumY1  +=       0.261 * cos(6.20453445714  + 52389.34919606990 *T)
    sumY1  +=       0.233 * cos(5.55470184122  + 51123.78171744350 *T)
    sumY1  +=       0.231 * cos(3.67240300147  + 82815.90673926489 *T)
    sumY1  +=       0.239 * cos(1.23297150719  + 213.54291292150 *T)
    sumY1  +=       0.222 * cos(2.42341131487  + 66941.28914390170 *T)
    sumY1  +=       0.216 * cos(3.43857770880  + 313055.08151637384 *T)
    sumY1  +=       0.276 * cos(6.03736928820  + 51109.55462344189 *T)
    sumY1  +=       0.222 * cos(1.79769065581  + 32371.22280904910 *T)
    sumY1  +=       0.198 * cos(0.48217448832  + 53765.12299591570 *T)
    sumY1  +=       0.197 * cos(5.45332171285  + 69160.04624808849 *T)
    sumY1  +=       0.196 * cos(1.08031906323  + 50057.28624025350 *T)
    sumY1  +=       0.202 * cos(2.45524349258  + 129380.37759516528 *T)
    sumY1  +=       0.201 * cos(2.89680722301  + 44937.37450903190 *T)
    sumY1  +=       0.216 * cos(3.72207636306  + 213.05527795450 *T)
    sumY1  +=       0.185 * cos(0.34499909570  + 136101.09157450669 *T)
    sumY1  +=       0.183 * cos(0.69231889541  + 419.72846135871 *T)
    sumY1  +=       0.186 * cos(5.71673843636  + 57837.38214978410 *T)
    sumY1  +=       0.178 * cos(3.70090079088  + 105461.23493587368 *T)
    sumY1  +=       0.177 * cos(2.18468434488  + 51962.75100519390 *T)
    sumY1  +=       0.170 * cos(4.76419946756  + 78244.28348130629 *T)
    sumY1  +=       0.229 * cos(0.17271876456  + 433.95555536030 *T)
    sumY1  +=       0.173 * cos(5.50271484365  + 94138.57083756928 *T)
    sumY1  +=       0.157 * cos(2.02010274317  + 64742.20180061470 *T)
    sumY1  +=       0.153 * cos(5.58836484297  + 59414.72569223190 *T)
    sumY1  +=       0.159 * cos(4.59819683221  + 853.44019923550 *T)
    sumY1  +=       0.155 * cos(3.11896173244  + 35192.05395317530 *T)
    sumY1  +=       0.149 * cos(5.04913257055  + 24499.07406377390 *T)
    sumY1  +=       0.152 * cos(1.43713086803  + 4551.70967957530 *T)
    sumY1  +=       0.147 * cos(0.13661195710  + 52602.64829150790 *T)
    sumY1  +=       0.147 * cos(0.58488249671  + 19805.07110906630 *T)
    sumY1  +=       0.182 * cos(0.68589997649  + 55618.62504559730 *T)
    sumY1  +=       0.140 * cos(4.81668277749  + 23754.95056618569 *T)
    sumY1  +=       0.146 * cos(4.89226138375  + 208702.98131511008 *T)
    sumY1  +=       0.154 * cos(1.81985059116  + 99799.90288672148 *T)
    sumY1  +=       0.139 * cos(4.63018939309  + 25973.70767037247 *T)
    sumY1  +=       0.138 * cos(4.17371703563  + 26202.58624774291 *T)
    sumY1  +=       0.191 * cos(2.30404978778  + 26728.04424537170 *T)
    sumY1  +=       0.152 * cos(4.17149146165  + 1109.62236957690 *T)
    sumY1  +=       0.162 * cos(6.13333012195  + 38520.18960945550 *T)
    sumY1  +=       0.141 * cos(5.95496888155  + 77734.26227711148 *T)
    sumY1  +=       0.135 * cos(1.84463199555  + 78267.83457756409 *T)
    sumY1  +=       0.128 * cos(2.99907290513  + 88477.23878841709 *T)
    sumY1  +=       0.147 * cos(0.62089024684  + 31748.99137324289 *T)
    sumY1  +=       0.135 * cos(4.69617813235  + 124778.42747620228 *T)
    sumY1  +=       0.148 * cos(5.85857347873  + 27154.64243624770 *T)
    sumY1  +=       0.134 * cos(0.45601803820  + 12431.79883291429 *T)
    sumY1  +=       0.121 * cos(2.04615713494  + 104881.54734887488 *T)
    sumY1  +=       0.116 * cos(3.19167458045  + 25934.36814857290 *T)
    sumY1  +=       0.116 * cos(4.22027620935  + 32858.36992533629 *T)
    sumY1  +=       0.115 * cos(5.61211222559  + 26241.92576954250 *T)
    sumY1  +=       0.148 * cos(5.73338066299  + 25661.54876818170 *T)
    sumY1  +=       0.110 * cos(4.42886613408  + 28206.91081943610 *T)
    sumY1  +=       0.108 * cos(0.11698734549  + 27140.41534224609 *T)
    sumY1  +=       0.115 * cos(2.44781109373  + 103822.16541868568 *T)
    sumY1  +=       0.109 * cos(5.49669317005  + 53284.94101775829 *T)
    sumY1  +=       0.107 * cos(3.99865946781  + 28421.34335192970 *T)
    sumY1  +=       0.143 * cos(3.36395963187  + 1596.43025976811 *T)
    sumY1  +=       0.112 * cos(4.03868201051  + 155418.28411483529 *T)
    sumY1  +=       0.113 * cos(0.89157360809  + 5327.71992586630 *T)
    sumY1  +=       0.101 * cos(1.94476004771  + 25620.18196870330 *T)
    sumY1  +=       0.122 * cos(1.46662898961  + 10213.52936369450 *T)
    sumY1  +=       0.102 * cos(3.99402750542  + 2333.44021035551 *T)
    sumY1  +=       0.097 * cos(0.57476621767  + 26556.11194941211 *T)
    sumY1  +=       0.118 * cos(0.18981536711  + 7994.77225950771 *T)
    sumY1  +=       0.109 * cos(0.03572167179  + 103396.25664217168 *T)
    sumY1  +=       0.109 * cos(4.54205508167  + 79219.55298381469 *T)
    sumY1  +=       0.095 * cos(4.81791418885  + 58459.12595062330 *T)
    sumY1  +=       0.095 * cos(1.17289793021  + 78256.83969520529 *T)
    sumY1  +=       0.097 * cos(5.79863208794  + 105411.23831396949 *T)
    sumY1  +=       0.092 * cos(3.83355549192  + 10213.04172872750 *T)
    sumY1  +=       0.094 * cos(2.84811066413  + 1582.20316576650 *T)
    sumY1  +=       0.103 * cos(3.37710316075  + 91785.70468379748 *T)
    sumY1  +=       0.101 * cos(5.05588975902  + 20426.32727493849 *T)
    sumY1  +=       0.089 * cos(1.67320799366  + 74821.37829724069 *T)
    sumY1  +=       0.089 * cos(4.81402637705  + 125887.80602829569 *T)
    sumY1  +=       0.103 * cos(5.07765700675  + 51066.18391357149 *T)
    sumY1  +=       0.087 * cos(5.93439795460  + 27043.25906569930 *T)
    sumY1  +=       0.083 * cos(1.05061824879  + 73711.99974514729 *T)
    sumY1  +=       0.083 * cos(0.20717662715  + 77624.05595589208 *T)
    sumY1  +=       0.097 * cos(3.79624621564  + 71980.87739221469 *T)
    sumY1  +=       0.081 * cos(5.50420619898  + 155468.28073673949 *T)
    sumY1  +=       0.080 * cos(5.57436544726  + 22645.08437912529 *T)
    sumY1  +=       0.081 * cos(1.24551992048  + 76045.19613801930 *T)
    sumY1  +=       0.081 * cos(3.47212148199  + 26138.14358096190 *T)
    sumY1  +=       0.080 * cos(5.32926891572  + 26038.15033715350 *T)
    sumY1  +=       0.104 * cos(0.71084047438  + 78271.06678920689 *T)
    sumY1  +=       0.103 * cos(3.66294929279  + 78283.62300310588 *T)
    sumY1  +=       0.083 * cos(0.93556744308  + 80482.71034639288 *T)
    sumY1  +=       0.089 * cos(0.32257506891  + 78260.07190684808 *T)
    sumY1  +=       0.073 * cos(5.79066657626  + 26617.35028918529 *T)
    sumY1  +=       0.077 * cos(5.23137947798  + 18208.05780571871 *T)
    sumY1  +=       0.082 * cos(2.34137394607  + 77211.68485901768 *T)
    sumY1  +=       0.077 * cos(1.31837686051  + 41962.76455442090 *T)
    sumY1  +=       0.082 * cos(2.81188130706  + 77197.45776501608 *T)
    sumY1  +=       0.067 * cos(3.45344753309  + 85034.66384345168 *T)
    sumY1  +=       0.070 * cos(5.80811525915  + 537.04832957890 *T)
    sumY1  +=       0.067 * cos(1.90240647027  + 5661.08823166870 *T)
    sumY1  +=       0.067 * cos(2.25742587365  + 120226.47397914348 *T)
    sumY1  +=       0.065 * cos(3.38772930583  + 162188.99471608087 *T)
    sumY1  +=       0.075 * cos(3.36246887110  + 149.80701461810 *T)
    sumY1  +=       0.071 * cos(0.43774462141  + 108903.80988083908 *T)
    sumY1  +=       0.063 * cos(1.71956484187  + 51756.56545675670 *T)
    sumY1  +=       0.062 * cos(0.86384103129  + 102762.78348849648 *T)
    sumY1  +=       0.067 * cos(2.92995035330  + 64608.09275102969 *T)
    sumY1  +=       0.064 * cos(3.59312937724  + 13521.99525907490 *T)
    sumY1  +=       0.062 * cos(1.70048284317  + 28306.41642827749 *T)
    sumY1  +=       0.058 * cos(0.05666653789  + 11322.90791578790 *T)
    sumY1  +=       0.058 * cos(1.85046002910  + 24498.58642880689 *T)
    sumY1  +=       0.077 * cos(2.22443127486  + 25565.56954096390 *T)
    sumY1  +=       0.076 * cos(3.03069740917  + 78477.25233764409 *T)
    sumY1  +=       0.063 * cos(5.48328560721  + 93029.19228547587 *T)
    sumY1  +=       0.078 * cos(4.97356487718  + 71493.24264096050 *T)
    sumY1  +=       0.077 * cos(2.34741649134  + 51742.33836275509 *T)
    sumY1  +=       0.057 * cos(2.02149140578  + 2648.69864295650 *T)
    sumY1  +=       0.058 * cos(0.20352146431  + 339142.98465794802 *T)
    sumY1  +=       0.056 * cos(3.96222140968  + 76145.18938182769 *T)
    sumY1  +=       0.059 * cos(1.23567785929  + 21535.70582703189 *T)
    sumY1  +=       0.056 * cos(0.13198299548  + 117873.60782537168 *T)
    sumY1  +=       0.053 * cos(6.21797647818  + 25557.96835899609 *T)
    sumY1  +=       0.060 * cos(5.92257554155  + 71025.27765060609 *T)
    sumY1  +=       0.052 * cos(5.28001264458  + 10022.08109758290 *T)
    sumY1  +=       0.055 * cos(4.20817850068  + 51322.85371887989 *T)
    sumY1  +=       0.058 * cos(2.97195320983  + 43071.65547154729 *T)
    sumY1  +=       0.053 * cos(3.50115924170  + 79853.02613748988 *T)
    sumY1  +=       0.053 * cos(4.92881565492  + 78050.65414676809 *T)
    sumY1  +=       0.053 * cos(0.59227243554  + 98068.78053378889 *T)
    sumY1  +=       0.057 * cos(5.95602153522  + 633.02755679670 *T)
    sumY1  +=       0.049 * cos(5.08646367150  + 90830.10494218887 *T)
    sumY1  +=       0.067 * cos(0.27295012704  + 26610.72437715149 *T)
    sumY1  +=       0.050 * cos(4.43506318049  + 131549.13807744789 *T)
    sumY1  +=       0.049 * cos(3.57210369085  + 33968.23611239669 *T)
    sumY1  +=       0.050 * cos(3.73457689628  + 81706.52818717150 *T)
    sumY1  +=       0.051 * cos(5.44465818352  + 129910.06856025988 *T)
    sumY1  +=       0.057 * cos(1.99278813028  + 26624.95147115310 *T)
    sumY1  +=       0.045 * cos(4.93934477894  + 51116.18053547569 *T)
    sumY1  +=       0.045 * cos(1.53008615762  + 104332.18662288050 *T)
    sumY1  +=       0.051 * cos(5.56661489639  + 647.25465079831 *T)
    sumY1  +=       0.043 * cos(3.21201517479  + 78690.55143308209 *T)
    sumY1  +=       0.054 * cos(0.68244488765  + 522.82123557730 *T)
    sumY1  +=       0.042 * cos(6.05067077478  + 114565.14192999128 *T)
    sumY1  +=       0.041 * cos(0.42167567232  + 38519.70197448849 *T)
    sumY1  +=       0.041 * cos(5.52226173141  + 25028.27739390149 *T)
    sumY1  +=       0.052 * cos(5.71899801337  + 39744.00745023410 *T)
    sumY1  +=       0.043 * cos(5.20652224208  + 103925.25819290428 *T)
    sumY1  +=       0.045 * cos(3.01197275808  + 76.50988875911 *T)
    sumY1  +=       0.042 * cos(1.64710245004  + 46848.57399224910 *T)
    sumY1  +=       0.047 * cos(2.33748031419  + 44181.52165860769 *T)
    sumY1  +=       0.038 * cos(1.51512405751  + 22760.01130277751 *T)
    sumY1  +=       0.037 * cos(1.55847676346  + 151975.70916986989 *T)
    sumY1  +=       0.050 * cos(0.66442193902  + 25551.34244696229 *T)
    sumY1  +=       0.050 * cos(5.43960666857  + 50579.86365834729 *T)
    sumY1  +=       0.045 * cos(4.18105807315  + 7879.84533585549 *T)
    sumY1  +=       0.043 * cos(3.23954616248  + 26521.85869693450 *T)
    sumY1  +=       0.037 * cos(4.29411552008  + 30639.61282114949 *T)
    sumY1  +=       0.037 * cos(5.45971389197  + 48732.98752069949 *T)
    sumY1  +=       0.035 * cos(1.61116129207  + 84547.02909219750 *T)
    sumY1  +=       0.040 * cos(5.41660085121  + 52815.94738694590 *T)
    sumY1  +=       0.041 * cos(5.72036408436  + 1059.13811270570 *T)
    sumY1  +=       0.041 * cos(1.46421286291  + 150866.33061777649 *T)
    sumY1  +=       0.040 * cos(3.44404714469  + 26294.33250749489 *T)
    sumY1  +=       0.041 * cos(5.29927149302  + 25881.96141062050 *T)
    sumY1  +=       0.034 * cos(5.65060802359  + 26191.23973327629 *T)
    sumY1  +=       0.034 * cos(3.16023657143  + 25985.05418483910 *T)
    sumY1  +=       0.034 * cos(5.54425040097  + 89586.61734051049 *T)
    sumY1  +=       0.041 * cos(2.60243509566  + 52329.82891111670 *T)
    sumY1  +=       0.035 * cos(0.95515596888  + 157637.04121902207 *T)
    sumY1  +=       0.042 * cos(4.45181501350  + 26507.63160293289 *T)
    sumY1  +=       0.039 * cos(1.71451949376  + 234790.88445668429 *T)
    sumY1  +=       0.037 * cos(3.45787318433  + 57836.89451481709 *T)
    sumY1  +=       0.036 * cos(5.05609057964  + 130969.45049044908 *T)
    sumY1  +=       0.032 * cos(0.80695926075  + 181506.18725640947 *T)
    sumY1  +=       0.033 * cos(5.95814421441  + 50594.09075234889 *T)
    sumY1  +=       0.044 * cos(4.54253002200  + 6681.46867088311 *T)
    sumY1  +=       0.036 * cos(4.89960028584  + 104355.73771913828 *T)
    sumY1  +=       0.032 * cos(3.07469859082  + 129484.15978374588 *T)
    sumY1  +=       0.030 * cos(3.63644868220  + 52644.01509098630 *T)
    sumY1  +=       0.031 * cos(0.98929514792  + 29416.28261533789 *T)
    sumY1  +=       0.031 * cos(6.01882424612  + 105307.45612538888 *T)
    sumY1  +=       0.030 * cos(2.14561278666  + 95247.94938966268 *T)
    sumY1  +=       0.031 * cos(0.94264719675  + 52290.48938931711 *T)
    sumY1  +=       0.029 * cos(2.26390987562  + 181556.18387831369 *T)
    sumY1  +=       0.039 * cos(0.22423030716  + 25448.24967274370 *T)
    sumY1  +=       0.032 * cos(3.17151820552  + 51639.24558853649 *T)
    sumY1  +=       0.029 * cos(2.64834173157  + 53242.54557782190 *T)
    sumY1  +=       0.029 * cos(3.82338961652  + 25668.66231518250 *T)
    sumY1  +=       0.028 * cos(4.61151969244  + 77204.57131201689 *T)
    sumY1  +=       0.036 * cos(3.92497606168  + 25132.05958248210 *T)
    sumY1  +=       0.027 * cos(0.76041081814  + 54509.24649350390 *T)
    sumY1  +=       0.026 * cos(1.38925912724  + 52061.61081194667 *T)
    sumY1  +=       0.025 * cos(5.15879998687  + 29428.75938575750 *T)
    sumY1  +=       0.031 * cos(2.93989277053  + 18093.13088206649 *T)
    sumY1  +=       0.030 * cos(1.42238495741  + 26941.34334080970 *T)
    sumY1  +=       0.025 * cos(4.73209653345  + 956.53297345411 *T)
    sumY1  +=       0.027 * cos(2.35402357443  + 18093.61851703350 *T)
    sumY1  +=       0.028 * cos(3.75976953619  + 104358.96993078108 *T)
    sumY1  +=       0.026 * cos(0.11444202162  + 26237.71015619230 *T)
    sumY1  +=       0.024 * cos(3.77280410253  + 22747.53453235790 *T)
    sumY1  +=       0.024 * cos(2.17839409058  + 46514.23041651269 *T)
    sumY1  +=       0.029 * cos(2.57773209519  + 131499.14145554369 *T)
    sumY1  +=       0.024 * cos(3.54215594406  + 42154.21282053250 *T)
    sumY1  +=       0.024 * cos(1.15126185636  + 54294.81396101029 *T)
    sumY1  +=       0.023 * cos(5.30608917177  + 146314.37712071769 *T)
    sumY1  +=       0.024 * cos(4.83980442527  + 26107.32908499049 *T)
    sumY1  +=       0.026 * cos(2.88358798527  + 25021.65148186770 *T)
    sumY1  +=       0.023 * cos(2.25657344280  + 20760.18321570790 *T)
    sumY1  +=       0.023 * cos(1.97184358207  + 77154.08705514569 *T)
    sumY1  +=       0.024 * cos(5.65078963984  + 26067.98956319090 *T)
    sumY1  +=       0.023 * cos(2.17997702001  + 25035.87857586930 *T)
    sumY1  +=       0.023 * cos(3.16772215014  + 143961.51096694588 *T)
    sumY1  +=       0.026 * cos(0.15726115955  + 27684.33340134230 *T)
    sumY1  +=       0.022 * cos(5.04032506297  + 19804.58347409930 *T)
    sumY1  +=       0.025 * cos(1.50229754572  + 26080.54577708989 *T)
    sumY1  +=       0.023 * cos(1.92683823556  + 130013.16133447849 *T)
    sumY1  +=       0.022 * cos(2.30233150466  + 79372.84415933248 *T)
    sumY1  +=       0.025 * cos(5.43362878578  + 52026.48690349730 *T)
    sumY1  +=       0.021 * cos(4.22122242647  + 104344.74283677948 *T)
    sumY1  +=       0.021 * cos(4.30076756548  + 102133.09927959349 *T)
    sumY1  +=       0.023 * cos(6.10522918323  + 61279.95709474950 *T)
    sumY1  +=       0.021 * cos(0.79325948829  + 85502.62883380610 *T)
    sumY1  +=       0.023 * cos(4.81952053099  + 77844.46859833089 *T)
    sumY1  +=       0.022 * cos(0.15814691647  + 188276.89785765507 *T)
    sumY1  +=       0.021 * cos(3.64905245082  + 124156.68367536308 *T)
    sumY1  +=       0.027 * cos(5.39563901242  + 103299.58800059187 *T)
    sumY1  +=       0.020 * cos(1.06151057303  + 23968.89546371229 *T)
    sumY1  +=       0.023 * cos(5.40249889826  + 77830.24150432929 *T)
    sumY1  +=       0.021 * cos(5.85984642433  + 103285.36090659029 *T)
    sumY1  +=       0.020 * cos(0.50079058372  + 49527.59527515889 *T)
    sumY1  +=       0.020 * cos(0.72105141719  + 1479.11039154791 *T)
    sumY1  +=       0.020 * cos(4.35275493153  + 23969.38309867930 *T)
    sumY1  +=       0.020 * cos(3.38833851349  + 103711.95909746629 *T)
    sumY1  +=       0.019 * cos(2.08111180650  + 52126.05347872769 *T)
    sumY1  +=       0.019 * cos(1.62003983896  + 3327.89183879669 *T)
    sumY1  +=       0.022 * cos(6.13482771447  + 104565.15547921829 *T)
    sumY1  +=       0.026 * cos(5.97213001267  + 13362.69352428270 *T)
    sumY1  +=       0.025 * cos(4.72474160158  + 24864.32911827909 *T)
    sumY1  +=       0.023 * cos(5.79095805909  + 15874.37377787969 *T)
    sumY1  +=       0.022 * cos(3.15825920598  + 6283.31966747490 *T)
    sumY1  +=       0.025 * cos(4.07847502844  + 27311.96479983631 *T)
    sumY1  +=       0.019 * cos(3.21308432623  + 53228.31848382029 *T)
    sumY1  +=       0.021 * cos(1.74197598839  + 97581.14578253467 *T)
    sumY1  +=       0.018 * cos(3.99624752196  + 106570.61348796709 *T)
    sumY1  +=       0.021 * cos(0.35167836404  + 60056.13925397089 *T)
    sumY1  +=       0.019 * cos(1.46314564325  + 26083.77798873270 *T)
    sumY1  +=       0.021 * cos(0.41713451843  + 104371.52614468009 *T)
    sumY1  +=       0.017 * cos(0.90189598200  + 1162.71852189130 *T)
    sumY1  +=       0.020 * cos(5.87845266961  + 41962.27691945389 *T)
    sumY1  +=       0.020 * cos(2.74404571866  + 26091.54065944870 *T)
    sumY1  +=       0.017 * cos(1.01839175808  + 77410.75686045409 *T)
    sumY1  +=       0.018 * cos(6.27803384892  + 115674.52048208468 *T)
    sumY1  +=       0.021 * cos(3.48125971513  + 134991.71302241329 *T)
    sumY1  +=       0.017 * cos(1.53092891578  + 104138.55728834229 *T)
    sumY1  +=       0.017 * cos(0.95709459823  + 37410.32342239509 *T)
    sumY1  +=       0.019 * cos(0.82226099482  + 522.33360061030 *T)
    sumY1  +=       0.019 * cos(3.34754142930  + 52698.62751872569 *T)
    sumY1  +=       0.019 * cos(2.35260406170  + 131395.35926696309 *T)
    sumY1  +=       0.016 * cos(2.75961492200  + 12546.72575656651 *T)
    sumY1  +=       0.016 * cos(4.65532135428  + 100909.28143881488 *T)
    sumY1  +=       0.016 * cos(5.53845756469  + 25867.73431661889 *T)
    sumY1  +=       0.020 * cos(1.18563577372  + 26094.77287109149 *T)
    sumY1  +=       0.018 * cos(0.37484030344  + 47623.60896860609 *T)
    sumY1  +=       0.018 * cos(6.16926211843  + 12432.28646788130 *T)
    sumY1  +=       0.020 * cos(4.56182116076  + 11610.79577585770 *T)
    sumY1  +=       0.015 * cos(3.94554558708  + 183724.94436059627 *T)
    sumY1  +=       0.017 * cos(3.22546612448  + 51529.03926731709 *T)
    sumY1  +=       0.020 * cos(3.34083321292  + 426.35437339250 *T)
    sumY1  +=       0.016 * cos(3.33913638740  + 26308.55960149650 *T)
    sumY1  +=       0.016 * cos(3.22684879906  + 128850.68663007068 *T)
    sumY1  +=       0.019 * cos(2.99415679113  + 51543.26636131870 *T)
    sumY1  +=       0.019 * cos(2.17719175975  + 155997.97170183409 *T)
    sumY1  +=       0.015 * cos(6.05463333824  + 26164.41303033330 *T)
    sumY1  +=       0.018 * cos(5.23520259945  + 25874.36022865269 *T)
    sumY1  +=       0.014 * cos(3.61282271417  + 36300.94487030169 *T)
    sumY1  +=       0.019 * cos(4.31362944396  + 45494.82524723229 *T)
    sumY1  +=       0.014 * cos(2.00836472375  + 44295.96094729290 *T)
    sumY1  +=       0.018 * cos(2.25014922269  + 119117.09542705010 *T)
    sumY1  +=       0.016 * cos(2.19192842111  + 76667.76679992149 *T)
    sumY1  +=       0.014 * cos(0.50039602119  + 107794.43132874568 *T)
    sumY1  +=       0.014 * cos(3.93894618332  + 16065.82204399129 *T)
    sumY1  +=       0.014 * cos(2.90301858576  + 44181.03402364069 *T)
    sumY1  +=       0.014 * cos(3.19306372116  + 2111.89413086110 *T)
    sumY1  +=       0.015 * cos(0.17179065008  + 52226.04672253610 *T)
    sumY1  +=       0.014 * cos(3.19364287125  + 27999.34644227490 *T)
    sumY1  +=       0.019 * cos(3.37439513975  + 104347.97504842227 *T)
    sumY1  +=       0.014 * cos(1.85953479160  + 116918.00808376308 *T)
    sumY1  +=       0.015 * cos(3.25430509963  + 365230.88779952226 *T)
    sumY1  +=       0.016 * cos(4.96959083935  + 52712.85461272729 *T)
    sumY1  +=       0.017 * cos(2.67262744895  + 97113.18079218028 *T)
    sumY1  +=       0.013 * cos(5.63989609373  + 24176.94747584050 *T)
    sumY1  +=       0.017 * cos(1.53028465700  + 20760.67085067490 *T)
    sumY1  +=       0.014 * cos(5.20815151417  + 25954.03790947270 *T)
    sumY1  +=       0.014 * cos(3.59706617063  + 26222.25600864270 *T)
    sumY1  +=       0.017 * cos(5.41745483095  + 70269.42480018188 *T)
    sumY1  +=       0.013 * cos(4.59923088636  + 178063.61231144407 *T)
    sumY1  +=       0.015 * cos(5.79226796194  + 52099.78402935630 *T)
    sumY1  +=       0.013 * cos(2.11885842073  + 323.74923414091 *T)
    sumY1  +=       0.013 * cos(3.19236461837  + 39609.41076568209 *T)
    sumY1  +=       0.013 * cos(5.09373061293  + 206.42936592071 *T)
    sumY1  +=       0.014 * cos(0.25054779380  + 105940.92927906408 *T)
    sumY1  +=       0.013 * cos(0.63576827663  + 102233.09252340188 *T)
    sumY1  +=       0.014 * cos(2.18339498140  + 77204.08367704989 *T)
    sumY1  +=       0.014 * cos(2.15110849472  + 13521.50762410789 *T)
    sumY1  +=       0.013 * cos(3.12239387096  + 111590.53197538028 *T)
    sumY1  +=       0.012 * cos(2.70622746349  + 76681.99389392308 *T)
    sumY1  +=       0.015 * cos(5.99891720563  + 27670.10630734069 *T)
    sumY1  +=       0.013 * cos(4.07441974716  + 58946.27306691049 *T)
    sumY1  +=       0.015 * cos(3.57873341293  + 26300.95841952869 *T)
    sumY1  +=       0.013 * cos(4.88571694871  + 14477.10736571650 *T)
    sumY1  +=       0.013 * cos(5.78266274868  + 1692.40948698591 *T)
    sumY1  +=       0.013 * cos(2.82188552462  + 140653.04507156549 *T)
    sumY1  +=       0.012 * cos(6.05506602553  + 52072.95732641330 *T)
    sumY1  +=       0.012 * cos(1.55236651979  + 54394.31956985169 *T)
    sumY1  +=       0.012 * cos(5.88344203293  + 7238.91940908350 *T)
    sumY1  +=       0.012 * cos(0.69684033021  + 25447.76203777669 *T)
    sumY1  +=       0.012 * cos(4.63793697276  + 72936.47713382330 *T)
    sumY1  +=       0.015 * cos(3.92235876661  + 2126.12122486270 *T)
    sumY1  +=       0.012 * cos(4.09664422947  + 55504.18575691209 *T)
    sumY1  +=       0.012 * cos(2.03928728821  + 220.65645992230 *T)
    sumY1  +=       0.012 * cos(0.18923872919  + 111122.56698502589 *T)
    sumY1  +=       0.012 * cos(5.93822601178  + 90695.99589260388 *T)
    sumY1  +=       0.012 * cos(6.27891147429  + 104778.45457465628 *T)
    sumY1  +=       0.012 * cos(0.41458906078  + 64607.60511606269 *T)
    sumY1  +=       0.012 * cos(4.67455996800  + 110634.93223377167 *T)
    sumY1  +=       0.011 * cos(2.00000784368  + 36109.98423915710 *T)
    sumY1  +=       0.013 * cos(4.96093660073  + 49842.85370775989 *T)
    sumY1  +=       0.014 * cos(5.74772554930  + 78417.73205269089 *T)
    sumY1  +=       0.011 * cos(1.10539607022  + 26198.35328027710 *T)
    sumY1  +=       0.011 * cos(1.41663630740  + 25977.94063783829 *T)
    sumY1  +=       0.012 * cos(3.63665690282  + 31415.62306744050 *T)
    sumY1  +=       0.012 * cos(3.78977700630  + 6770.95441872910 *T)
    sumY1  +=       0.011 * cos(2.54959001341  + 51969.86455219470 *T)
    sumY1  +=       0.010 * cos(2.10136723403  + 23868.90221990390 *T)
    sumY1  +=       0.012 * cos(4.58411493514  + 130420.08976445468 *T)
    sumY1  +=       0.014 * cos(1.09052029316  + 32769.37181245730 *T)
    sumY1  +=       0.013 * cos(4.14764183344  + 33967.74847742969 *T)
    sumY1  +=       0.011 * cos(2.20920470312  + 51645.87150057029 *T)
    sumY1  +=       0.010 * cos(5.01452989749  + 28736.60178453070 *T)
    sumY1  +=       0.010 * cos(2.60659223197  + 52705.25343075949 *T)
    sumY1  +=       0.014 * cos(3.63437988054  + 77837.35505133009 *T)
    sumY1  +=       0.010 * cos(2.91124294480  + 141762.42362365889 *T)
    sumY1  +=       0.012 * cos(5.60563328751  + 224.58861318540 *T)
    sumY1  +=       0.010 * cos(6.22792613909  + 170049.41410852008 *T)
    sumY1  +=       0.013 * cos(2.82703801397  + 38813.60039383270 *T)
    sumY1  +=       0.010 * cos(3.02574120742  + 27147.04125427989 *T)
    sumY1  +=       0.010 * cos(2.23554494099  + 78903.85052852010 *T)
    sumY1  +=       0.013 * cos(0.23254548151  + 6282.83203250789 *T)
    sumY1  +=       0.010 * cos(5.31330188188  + 207644.08701988787 *T)
    sumY1  +=       0.012 * cos(4.98783133373  + 26395.70458002730 *T)
    sumY1  +=       0.012 * cos(3.81738635107  + 25780.58933808811 *T)
    sumY1  +=       0.010 * cos(2.41181269432  + 12565.90788249930 *T)
    sumY1  +=       0.012 * cos(2.61068546931  + 26725.14323108190 *T)
    sumY1  +=       0.012 * cos(6.19453221549  + 25451.15068703350 *T)
    sumY1  +=       0.009 * cos(4.05684885963  + 51219.76094466129 *T)
    sumY1  +=       0.009 * cos(6.12208983529  + 155572.06292532009 *T)
    sumY1  +=       0.009 * cos(0.41261244956  + 78731.91823256049 *T)
    sumY1  +=       0.009 * cos(2.60381655038  + 14764.99522578630 *T)
    sumY1  +=       0.011 * cos(1.67989630572  + 65697.80154222329 *T)
    sumY1  +=       0.012 * cos(0.71134314907  + 25446.73339731871 *T)
    sumY1  +=       0.012 * cos(1.81068922855  + 26729.56052079669 *T)
    sumY1  +=       0.012 * cos(3.05761613841  + 12725.20961729149 *T)
    sumY1  +=       0.010 * cos(2.47438981577  + 65831.91059180829 *T)
    sumY1  +=       0.009 * cos(0.40929169320  + 16703.30595098250 *T)
    sumY1  +=       0.009 * cos(4.90698904861  + 51708.08511027749 *T)
    sumY1  +=       0.009 * cos(3.85732708279  + 207594.09039798367 *T)
    sumY1  +=       0.009 * cos(4.96560671408  + 156101.06447605268 *T)
    sumY1  +=       0.012 * cos(4.51597349846  + 176954.23375935067 *T)
    sumY1  +=       0.009 * cos(0.27941818584  + 68242.11596210669 *T)
    sumY1  +=       0.009 * cos(3.76248228932  + 1910.95566573370 *T)
    sumY1  +=       0.009 * cos(2.58539985032  + 50586.48957038109 *T)
    sumY1  +=       0.009 * cos(0.03828568933  + 52609.76183850870 *T)
    sumY1  +=       0.008 * cos(4.59422184502  + 29530.23426905610 *T)
    sumY1  +=       0.009 * cos(4.40872070931  + 205.94173095370 *T)
    sumY1  +=       0.009 * cos(1.91710703544  + 55516.66252733170 *T)
    sumY1  +=       0.011 * cos(0.78490575811  + 52022.27129014710 *T)
    sumY1  +=       0.011 * cos(1.79573927984  + 157057.35363202327 *T)
    sumY1  +=       0.008 * cos(6.02666588108  + 39629.56816154889 *T)
    sumY1  +=       0.010 * cos(0.38285944104  + 52382.23564906909 *T)
    sumY1  +=       0.008 * cos(4.34710667767  + 52182.67865953191 *T)
    sumY1  +=       0.010 * cos(4.32001939934  + 40565.49814225770 *T)
    sumY1  +=       0.008 * cos(2.87593439674  + 103.33659170210 *T)
    sumY1  +=       0.008 * cos(5.31493047379  + 157483.26240853727 *T)
    sumY1  +=       0.008 * cos(4.79233550623  + 27177.85575025130 *T)
    sumY1  +=       0.008 * cos(4.01288217857  + 24998.43816786410 *T)
    sumY1  +=       0.009 * cos(4.80787399572  + 19.42594341629 *T)
    sumY1  +=       0.010 * cos(2.27022895876  + 536.56069461190 *T)
    sumY1  +=       0.010 * cos(4.80007299648  + 260878.78759825832 *T)
    sumY1  +=       0.008 * cos(4.56235937282  + 48847.91444435170 *T)
    sumY1  +=       0.008 * cos(4.66549233094  + 860.55374623631 *T)
    sumY1  +=       0.008 * cos(2.76375833425  + 26720.93069837089 *T)
    sumY1  +=       0.009 * cos(5.58566134530  + 25654.43522118089 *T)
    sumY1  +=       0.009 * cos(0.00405995806  + 63498.22656396929 *T)
    sumY1  +=       0.008 * cos(1.38258172759  + 52595.53474450709 *T)
    sumY1  +=       0.008 * cos(0.52005477263  + 130446.87307235529 *T)
    sumY1  +=       0.007 * cos(1.78779062151  + 50697.18352656750 *T)
    sumY1  +=       0.008 * cos(4.09271417559  + 1224.06165826211 *T)
    sumY1  +=       0.009 * cos(5.72498036679  + 19316.94872284510 *T)
    sumY1  +=       0.007 * cos(0.38836950932  + 13541.17738500769 *T)
    sumY1  +=       0.009 * cos(1.66361806448  + 130443.64086071248 *T)
    sumY1  +=       0.007 * cos(0.73505185505  + 209812.84750217048 *T)
    sumY1  +=       0.007 * cos(2.21903348151  + 9384.59719059169 *T)
    sumY1  +=       0.007 * cos(2.18741011968  + 46848.08635728210 *T)
    sumY1  +=       0.007 * cos(2.07119546850  + 172402.28026229187 *T)
    sumY1  +=       0.007 * cos(5.46643566327  + 846.32665223470 *T)
    sumY1  +=       0.007 * cos(1.48055041129  + 27044.43611502830 *T)
    sumY1  +=       0.007 * cos(0.42256065489  + 150244.58681693728 *T)
    sumY1  +=       0.009 * cos(2.36014528596  + 26735.15779237250 *T)
    sumY1  +=       0.007 * cos(3.18642075607  + 52325.61329776649 *T)
    sumY1  +=       0.007 * cos(6.17839162640  + 45892.97425064050 *T)
    sumY1  +=       0.008 * cos(1.20719608552  + 52169.42154173189 *T)
    sumY1  +=       0.008 * cos(0.42457308525  + 83924.79765639128 *T)
    sumY1  +=       0.007 * cos(3.81342742455  + 80597.14963507809 *T)
    sumY1  +=       0.006 * cos(2.21280392322  + 52279.14287485049 *T)
    sumY1  +=       0.006 * cos(1.57105717926  + 51013.57539622409 *T)
    sumY1  +=       0.008 * cos(2.16921693506  + 129387.49114216608 *T)
    sumY1  +=       0.009 * cos(2.14041581315  + 955.84355909210 *T)
    sumY1  +=       0.008 * cos(6.21586114247  + 77727.14873011068 *T)
    sumY1  +=       0.006 * cos(5.71782638960  + 79330.44871939608 *T)
    sumY1  +=       0.007 * cos(3.99442191803  + 78378.39253089130 *T)
    sumY1  +=       0.007 * cos(0.83146896151  + 51219.96272405629 *T)
    sumY1  +=       0.006 * cos(2.63351784574  + 129373.26404816448 *T)
    sumY1  +=       0.007 * cos(0.06013753430  + 26162.92855762500 *T)
    sumY1  +=       0.006 * cos(5.10934915127  + 103241.99019671988 *T)
    sumY1  +=       0.006 * cos(5.27952242053  + 3340.36860921629 *T)
    sumY1  +=       0.006 * cos(2.10211584098  + 35472.98796713290 *T)
    sumY1  +=       0.006 * cos(2.93834812676  + 130653.05862079248 *T)
    sumY1  +=       0.006 * cos(4.63873538075  + 110.45013870291 *T)
    sumY1  +=       0.007 * cos(6.14935848430  + 77631.16950289288 *T)
    sumY1  +=       0.006 * cos(3.41608343835  + 53131.16220727349 *T)
    sumY1  +=       0.006 * cos(2.45250087046  + 26013.36536049040 *T)
    sumY1  +=       0.006 * cos(4.96304042103  + 25508.45937205890 *T)
    sumY1  +=       0.006 * cos(3.84217726377  + 26667.83454605650 *T)
    sumY1  +=       0.006 * cos(2.84953525287  + 70268.93716521488 *T)
    sumY1  +=       0.006 * cos(2.18073760821  + 103918.14464590348 *T)
    sumY1  +=       0.008 * cos(5.64484638063  + 157587.04459711787 *T)
    sumY1  +=       0.006 * cos(0.28527129594  + 6885.39370741431 *T)
    sumY1  +=       0.007 * cos(3.20891519423  + 214364.80099922928 *T)
    sumY1  +=       0.007 * cos(2.79508447590  + 25754.29101828830 *T)
    sumY1  +=       0.007 * cos(6.01013320890  + 26422.00289982710 *T)
    sumY1  +=       0.006 * cos(4.09475623089  + 60055.65161900389 *T)
    sumY1  +=       0.006 * cos(4.43814061291  + 78149.51395352086 *T)
    sumY1  +=       0.005 * cos(5.60856067263  + 50064.39978725430 *T)
    sumY1  +=       0.006 * cos(1.06133168572  + 128221.00242116768 *T)
    sumY1  +=       0.006 * cos(3.38536280287  + 75615.49841673308 *T)
    sumY1  +=       0.005 * cos(5.98443573479  + 137678.43511695447 *T)
    sumY1  +=       0.007 * cos(1.25296072469  + 56727.51596272369 *T)
    sumY1  +=       0.007 * cos(1.14884253624  + 8194.03151572510 *T)
    sumY1  +=       0.005 * cos(4.48886015660  + 130226.46042991648 *T)
    sumY1  +=       0.006 * cos(4.17115612398  + 80382.71710258449 *T)
    sumY1  +=       0.006 * cos(0.69855070126  + 51653.47268253809 *T)
    sumY1  +=       0.006 * cos(4.78072360502  + 123669.04892410888 *T)
    sumY1  +=       0.006 * cos(5.42312230542  + 72602.13355808688 *T)
    sumY1  +=       0.006 * cos(6.00561848302  + 50056.79860528649 *T)
    sumY1  +=       0.005 * cos(5.89027963231  + 102769.89703549728 *T)
    sumY1  +=       0.006 * cos(5.34345526598  + 74820.89066227368 *T)
    sumY1  +=       0.006 * cos(0.24757397975  + 161079.61616398749 *T)
    sumY1  +=       0.006 * cos(1.60819865465  + 103932.37173990508 *T)
    sumY1  +=       0.006 * cos(1.04008531783  + 25234.95057730570 *T)
    sumY1  +=       0.007 * cos(5.82651027335  + 154938.58977164488 *T)
    sumY1  +=       0.005 * cos(5.52069502978  + 419.24082639170 *T)
    sumY1  +=       0.005 * cos(2.31699479833  + 19406.92210565810 *T)
    sumY1  +=       0.006 * cos(1.96387669233  + 1265.81129610991 *T)
    sumY1  +=       0.007 * cos(5.70425979338  + 64901.50353540690 *T)
    sumY1  +=       0.005 * cos(4.02516634304  + 73711.51211018028 *T)
    sumY1  +=       0.005 * cos(5.11989980868  + 78213.95662030189 *T)
    sumY1  +=       0.006 * cos(5.18848518823  + 182085.87484340827 *T)
    sumY1  +=       0.005 * cos(5.44202376441  + 102755.66994149568 *T)
    sumY1  +=       0.005 * cos(4.80390003741  + 40565.01050729069 *T)
    sumY1  +=       0.005 * cos(6.13170260920  + 102.84895673509 *T)
    sumY1  +=       0.005 * cos(4.30112257620  + 51748.96427478889 *T)
    sumY1  +=       0.005 * cos(0.85038224536  + 53399.86794141051 *T)
    sumY1  +=       0.005 * cos(2.42619611574  + 27780.31262856009 *T)
    sumY1  +=       0.005 * cos(1.56640230350  + 80482.22271142589 *T)
    sumY1  +=       0.005 * cos(2.47762753907  + 25938.58376192310 *T)
    sumY1  +=       0.005 * cos(5.22476504933  + 26402.33313892731 *T)
    sumY1  +=       0.005 * cos(3.58045263546  + 25773.96077918809 *T)
    sumY1  +=       0.005 * cos(0.01470040312  + 77616.94240889128 *T)
    sumY1  +=       0.005 * cos(0.50724383605  + 34282.42229226630 *T)
    sumY1  +=       0.006 * cos(4.89504933760  + 50050.17269325269 *T)
    sumY1  +=       0.005 * cos(1.99588630926  + 17893.87162584910 *T)
    sumY1  +=       0.004 * cos(0.97602564752  + 130432.64597835368 *T)
    sumY1  +=       0.004 * cos(0.40741730958  + 51852.54468397449 *T)
    sumY1  +=       0.004 * cos(2.23467299420  + 32370.73517408209 *T)
    sumY1  +=       0.005 * cos(1.35451697835  + 204151.51545301828 *T)
    sumY1  +=       0.006 * cos(1.74602932944  + 75930.75684933408 *T)
    sumY1  +=       0.004 * cos(5.26812670096  + 25668.17468021549 *T)
    sumY1  +=       0.005 * cos(1.36761975943  + 6043.98476389190 *T)
    sumY1  +=       0.005 * cos(2.74357702126  + 39450.59666585690 *T)
    sumY1  +=       0.005 * cos(2.28369543224  + 24491.96051677309 *T)
    sumY1  +=       0.004 * cos(1.23165526523  + 51955.63745819309 *T)
    sumY1  +=       0.004 * cos(4.09092913231  + 103498.66000202828 *T)
    sumY1  +=       0.005 * cos(0.06895267430  + 52396.46274307070 *T)
    sumY1  +=       0.004 * cos(2.56354623790  + 104505.63519426508 *T)
    sumY1  +=       0.005 * cos(5.58769879303  + 23754.46293121869 *T)
    sumY1  +=       0.005 * cos(5.71030733431  + 123201.08393375447 *T)
    sumY1  +=       0.004 * cos(6.20993923204  + 54087.24958384910 *T)
    sumY1  +=       0.004 * cos(3.63799815383  + 637.24008950771 *T)
    sumY1  +=       0.004 * cos(0.59460014711  + 28791.76311373330 *T)
    sumY1  +=       0.004 * cos(1.92743223051  + 23384.53080438210 *T)
    sumY1  +=       0.005 * cos(4.52669072840  + 53029.24648238390 *T)
    sumY1  +=       0.005 * cos(0.25501672186  + 18848.98373249069 *T)
    sumY1  +=       0.004 * cos(1.81268637965  + 26824.02347258949 *T)
    sumY1  +=       0.004 * cos(0.70934599797  + 25352.27044552590 *T)
    sumY1  +=       0.004 * cos(3.53975344268  + 133882.33447031991 *T)
    sumY1  +=       0.004 * cos(0.87353062113  + 132658.51662954129 *T)
    sumY1  +=       0.004 * cos(3.22384434044  + 53772.23654291649 *T)
    sumY1  +=       0.004 * cos(2.01166034864  + 183571.16555011147 *T)
    sumY1  +=       0.004 * cos(5.89521734601  + 167850.32676523307 *T)
    sumY1  +=       0.005 * cos(5.28708364201  + 145204.99856862429 *T)
    sumY1  +=       0.004 * cos(1.34862695530  + 27171.22719135130 *T)
    sumY1  +=       0.004 * cos(1.17340542232  + 25005.06672676410 *T)
    sumY1  +=       0.004 * cos(3.00220929293  + 50483.88443112950 *T)
    sumY1  +=       0.004 * cos(0.14217969958  + 78786.53066029989 *T)
    sumY1  +=       0.004 * cos(3.41518083973  + 949.41942645330 *T)
    sumY1  +=       0.005 * cos(0.77055922503  + 45892.48661567349 *T)
    sumY1  +=       0.004 * cos(3.51962830585  + 1795.50226120450 *T)
    sumY1  +=       0.005 * cos(1.84426394169  + 26709.89075989690 *T)
    sumY1  +=       0.005 * cos(0.67776843593  + 25466.40315821850 *T)
    sumY1  +=       0.004 * cos(4.27093115110  + 52065.84377941249 *T)
    sumY1  +=       0.004 * cos(0.10877025827  + 2222.10045208050 *T)
    sumY1  +=       0.004 * cos(1.11438053331  + 78270.58180110609 *T)
    sumY1  +=       0.004 * cos(5.94116454026  + 25653.94758621389 *T)
    sumY1  +=       0.004 * cos(5.36329329492  + 49842.36607279289 *T)
    sumY1  +=       0.004 * cos(4.90307551419  + 143005.91122533727 *T)
    sumY1  +=       0.004 * cos(1.70672949321  + 78800.75775430149 *T)
    sumY1  +=       0.004 * cos(0.14272238686  + 65697.31390725628 *T)
    sumY1  +=       0.003 * cos(1.76268636575  + 52195.23222656469 *T)
    sumY1  +=       0.004 * cos(3.45396433025  + 130459.42928625426 *T)
    sumY1  +=       0.004 * cos(5.54668248593  + 24491.47288180609 *T)
    sumY1  +=       0.004 * cos(2.57630900199  + 3178.38960805111 *T)
    sumY1  +=       0.004 * cos(3.69827148826  + 220.16882495529 *T)
    sumY1  +=       0.004 * cos(3.39420749098  + 52250.83169919920 *T)
    sumY1  +=       0.003 * cos(2.73132197327  + 78160.86046798748 *T)
    sumY1  +=       0.004 * cos(2.89190828519  + 87367.86023632369 *T)
    sumY1  +=       0.004 * cos(0.15780189015  + 130435.87818999648 *T)
    sumY1  +=       0.004 * cos(0.14307301866  + 53234.94439585409 *T)
    sumY1  +=       0.004 * cos(0.06711300408  + 26575.78171031190 *T)
    sumY1  +=       0.004 * cos(2.45491937354  + 25600.51220780350 *T)
    sumY1  +=       0.003 * cos(5.37884708213  + 1486.22393854870 *T)
    sumY1  +=       0.005 * cos(3.35368845011  + 86144.04239554508 *T)
    sumY1  +=       0.004 * cos(1.34102185895  + 32132.37554043310 *T)
    sumY1  +=       0.004 * cos(1.04693916592  + 66653.40128383189 *T)
    sumY1  +=       0.003 * cos(0.35934615572  + 52310.15915021690 *T)
    sumY1  +=       0.003 * cos(2.24307597389  + 45290.90021070109 *T)
    sumY1  +=       0.003 * cos(5.05641378963  + 70383.86408886709 *T)
    sumY1  +=       0.003 * cos(2.84685809679  + 52252.31617190749 *T)
    sumY1  +=       0.005 * cos(0.41148055834  + 52808.83383994509 *T)
    sumY1  +=       0.003 * cos(0.89098898576  + 1588.82907780029 *T)
    sumY1  +=       0.004 * cos(4.17927437554  + 58857.27495403150 *T)
    sumY1  +=       0.004 * cos(3.20092050693  + 51951.70530492999 *T)
    sumY1  +=       0.003 * cos(2.40144059169  + 50264.85061741470 *T)
    sumY1  +=       0.004 * cos(1.17285198222  + 20043.91837768230 *T)
    sumY1  +=       0.004 * cos(3.78297088323  + 128320.99566497609 *T)
    sumY1  +=       0.003 * cos(3.96547374476  + 25986.18444079209 *T)
    sumY1  +=       0.003 * cos(4.83974394004  + 26190.10947732330 *T)
    sumY1  +=       0.004 * cos(1.45951607601  + 136722.83537534589 *T)
    sumY1  +=       0.003 * cos(2.84255023310  + 65717.47130312308 *T)
    sumY1  +=       0.003 * cos(2.62501893240  + 181026.49291321909 *T)
    sumY1  +=       0.003 * cos(4.29835303229  + 51535.66517935089 *T)
    sumY1  +=       0.004 * cos(0.32726328089  + 129799.86223904048 *T)
    sumY1  +=       0.003 * cos(4.92235874244  + 26073.91986505609 *T)
    sumY1  +=       0.003 * cos(3.88285894236  + 26102.37405305930 *T)
    sumY1  +=       0.003 * cos(4.97897799737  + 52168.44891866409 *T)
    sumY1  +=       0.004 * cos(2.55090678746  + 52155.89270476510 *T)
    sumY1  +=       0.004 * cos(1.29677266690  + 37698.69891743190 *T)
    sumY1  +=       0.003 * cos(0.56868525863  + 51109.06698847489 *T)
    sumY1  +=       0.003 * cos(1.87341446779  + 26247.44869384990 *T)
    sumY1  +=       0.003 * cos(0.64861790983  + 25928.84522426550 *T)
    sumY1  +=       0.004 * cos(0.30979285060  + 468.20880783791 *T)
    sumY1  +=       0.004 * cos(3.27872430484  + 132028.83242063827 *T)
    sumY1  +=       0.004 * cos(5.61336389985  + 25661.06113321469 *T)
    sumY1  +=       0.003 * cos(0.47396259409  + 2703.37233719210 *T)
    sumY1  +=       0.004 * cos(5.86092912592  + 166740.94821313969 *T)
    sumY1  +=       0.004 * cos(3.23572058854  + 59414.23805726489 *T)
    sumY1  +=       0.003 * cos(4.82502734269  + 61560.89110870710 *T)
    sumY1  +=       0.003 * cos(0.00693087411  + 79316.22162539448 *T)
    sumY1  +=       0.003 * cos(3.05784222447  + 130866.35771623049 *T)
    sumY1  +=       0.003 * cos(0.12277419735  + 79219.06534884768 *T)
    sumY1  +=       0.003 * cos(2.46452817521  + 24506.18761077469 *T)
    sumY1  +=       0.003 * cos(0.66136657530  + 51868.49247966230 *T)
    sumY1  +=       0.003 * cos(2.85883702115  + 19202.50943415989 *T)
    sumY1  +=       0.003 * cos(6.14048429792  + 40852.89836736050 *T)
    sumY1  +=       0.003 * cos(2.81485517413  + 116783.89903417807 *T)
    sumY1  +=       0.003 * cos(5.64866640193  + 105460.74730090667 *T)
    sumY1  +=       0.003 * cos(0.28115071068  + 48835.43767393209 *T)
    sumY1  +=       0.003 * cos(4.12067030266  + 114.68310616872 *T)
    sumY1  +=       0.003 * cos(4.90383671624  + 76785.08666814168 *T)
    sumY1  +=       0.003 * cos(0.85813944126  + 81592.08889848628 *T)
    sumY1  +=       0.003 * cos(2.81772139589  + 95247.46175469569 *T)
    sumY1  +=       0.003 * cos(4.69086415103  + 52171.68113030689 *T)
    sumY1  +=       0.004 * cos(5.11822399207  + 78110.17443172129 *T)
    sumY1  +=       0.003 * cos(2.45643813863  + 68050.18006102808 *T)
    sumY1  +=       0.003 * cos(5.59679660207  + 38653.81102407349 *T)
    sumY1  +=       0.003 * cos(1.32662236669  + 2219.00092167030 *T)
    sumY1  +=       0.003 * cos(5.92341361550  + 52179.44380102290 *T)
    sumY1  +=       0.003 * cos(4.15087523006  + 52286.25642185129 *T)
    sumY1  +=       0.003 * cos(0.17348863040  + 76674.39271195528 *T)
    sumY1  +=       0.003 * cos(0.93059863207  + 126997.18458038908 *T)
    sumY1  +=       0.003 * cos(4.49117697679  + 77101.47853779829 *T)
    sumY1  +=       0.003 * cos(4.40672346738  + 77741.37582411229 *T)
    sumY1  +=       0.003 * cos(4.09517524067  + 8989.22388794889 *T)
    sumY1  +=       0.003 * cos(5.27976660688  + 77733.77464214448 *T)
    sumY1  +=       0.003 * cos(1.52949550682  + 286966.69073983253 *T)
    sumY1  +=       0.003 * cos(0.97589529959  + 74.53778108379 *T)
    sumY1  +=       0.003 * cos(4.21847009410  + 632.53992182970 *T)
    sumY1  +=       0.004 * cos(4.47958960947  + 28287.23430234470 *T)
    sumY1  +=       0.004 * cos(4.32562807533  + 23889.05961577070 *T)
    sumY1  +=       0.003 * cos(5.21982122522  + 75930.26921436709 *T)
    sumY1  +=       0.002 * cos(1.69313010887  + 182188.96761762688 *T)
    sumY1  +=       0.003 * cos(4.67718054767  + 27250.62166346549 *T)
    sumY1  +=       0.003 * cos(4.13971825138  + 51315.74017187909 *T)
    sumY1  +=       0.003 * cos(4.52130139518  + 52182.67601266569 *T)
    sumY1  +=       0.002 * cos(4.85799167915  + 121335.85253123687 *T)
    sumY1  +=       0.002 * cos(3.88643180579  + 110012.70079796547 *T)
    sumY1  +=       0.002 * cos(1.22067192730  + 2008.80135664250 *T)
    sumY1  +=       0.002 * cos(3.76859743855  + 24925.67225464990 *T)
    sumY1  +=       0.003 * cos(0.31070573063  + 39629.08052658189 *T)
    sumY1  +=       0.003 * cos(3.24788020549  + 3340.85624418330 *T)
    sumY1  +=       0.002 * cos(0.35390973897  + 23439.20449861769 *T)
    sumY1  +=       0.002 * cos(1.50089471176  + 29415.79498037089 *T)
    sumY1  +=       0.003 * cos(0.83102506625  + 25131.85780308709 *T)
    sumY1  +=       0.003 * cos(6.25945305227  + 51123.29408247649 *T)
    sumY1  +=       0.003 * cos(0.28921662553  + 90695.50825763687 *T)
    sumY1  +=       0.002 * cos(0.13680232887  + 24505.69997580769 *T)
    sumY1  +=       0.003 * cos(2.81292041302  + 26011.88088778210 *T)
    sumY1  +=       0.002 * cos(3.34607115822  + 51329.96726588070 *T)
    sumY1  +=       0.003 * cos(4.94974770833  + 81604.56566890588 *T)
    sumY1  +=       0.003 * cos(5.66982181767  + 52813.04637265610 *T)
    sumY1  +=       0.003 * cos(0.02173421963  + 391318.79094109638 *T)
    sumY1  +=       0.002 * cos(2.96964107128  + 196137.31725009429 *T)
    sumY1  +=       0.003 * cos(1.28173018247  + 156507.99290602890 *T)
    sumY1  +=       0.003 * cos(0.93767773997  + 27727.21647624570 *T)
    sumY1  +=       0.003 * cos(1.58435463765  + 24449.07744186970 *T)
    sumY1  +=       0.002 * cos(5.77422284047  + 78057.76769376888 *T)
    sumY1  +=       0.003 * cos(3.17950494103  + 78313.94986411028 *T)
    sumY1  +=       0.003 * cos(1.26777637523  + 203042.13690092487 *T)
    sumY1  +=       0.003 * cos(2.84442647149  + 34082.67540108190 *T)
    sumY1  +=       0.002 * cos(5.82921421247  + 38634.62889814070 *T)
    sumY1  +=       0.003 * cos(4.25851095793  + 78257.32468330608 *T)
    sumY1  +=       0.002 * cos(4.89476820098  + 49527.10764019189 *T)
    sumY1  +=       0.003 * cos(1.95628126873  + 52041.94105104690 *T)
    sumY1  +=       0.002 * cos(3.46648506780  + 104819.82137413470 *T)
    sumY1  +=       0.002 * cos(5.05488554468  + 62197.88738073129 *T)
    sumY1  +=       0.003 * cos(5.38708769121  + 103291.98681862408 *T)
    sumY1  +=       0.002 * cos(1.95981868389  + 51962.26337022689 *T)
    sumY1  +=       0.002 * cos(3.49107424857  + 35077.61466449009 *T)
    sumY1  +=       0.003 * cos(1.93353771137  + 96357.32794175608 *T)
    sumY1  +=       0.002 * cos(0.91930935416  + 25021.16384690069 *T)
    sumY1  +=       0.002 * cos(4.74992302909  + 66652.91364886488 *T)
    sumY1  +=       0.002 * cos(2.31652509447  + 199599.07432099240 *T)
    sumY1  +=       0.002 * cos(4.52662503285  + 9103.66317663410 *T)
    sumY1  +=       0.002 * cos(5.81156600010  + 74923.34081550628 *T)
    sumY1  +=       0.002 * cos(5.31799111712  + 17098.67925362531 *T)
    sumY1  +=       0.002 * cos(3.54866382542  + 78470.13879064328 *T)
    sumY1  +=       0.002 * cos(0.96038903828  + 101703.40155830728 *T)
    sumY1  +=       0.002 * cos(0.63534070468  + 77307.66408623548 *T)
    sumY1  +=       0.002 * cos(5.30685200110  + 104991.75367009427 *T)
    sumY1  +=       0.002 * cos(4.00030437238  + 86143.55476057807 *T)
    sumY1  +=       0.002 * cos(2.46838836063  + 26312.49175475960 *T)
    sumY1  +=       0.003 * cos(6.17763429163  + 69159.55861312148 *T)
    sumY1  +=       0.002 * cos(4.36618723507  + 27676.73221937449 *T)
    sumY1  +=       0.002 * cos(5.52193246370  + 91919.81373338249 *T)
    sumY1  +=       0.002 * cos(2.07266268481  + 225687.46509753368 *T)
    sumY1  +=       0.002 * cos(1.38235585269  + 60369.83779887349 *T)
    sumY1  +=       0.002 * cos(2.45878629397  + 26184.12618627550 *T)
    sumY1  +=       0.002 * cos(0.06324608365  + 25992.16773183990 *T)
    sumY1  +=       0.002 * cos(0.06817332477  + 62388.84801187589 *T)
    sumY1  +=       0.002 * cos(3.72202154841  + 7880.33297082250 *T)
    sumY1  +=       0.003 * cos(3.66137462974  + 51534.63653889291 *T)
    sumY1  +=       0.002 * cos(2.59383779834  + 24608.79275002629 *T)
    sumY1  +=       0.002 * cos(4.79894731382  + 183145.25677359747 *T)
    sumY1  +=       0.002 * cos(1.81015048354  + 177288.08970012009 *T)
    sumY1  +=       0.002 * cos(1.16884626410  + 26397.42528171349 *T)
    sumY1  +=       0.002 * cos(1.35318611352  + 25778.86863640189 *T)
    sumY1  +=       0.002 * cos(3.47952755604  + 42791.20909255670 *T)
    sumY1  +=       0.002 * cos(1.89575854549  + 38813.11275886569 *T)
    sumY1  +=       0.002 * cos(3.54021512986  + 51539.05382860770 *T)
    sumY1  +=       0.002 * cos(2.45627954068  + 24952.47008716450 *T)
    sumY1  +=       0.002 * cos(0.06575283694  + 27223.82383095090 *T)
    sumY1  +=       0.002 * cos(2.99909862001  + 103719.07264446709 *T)
    sumY1  +=       0.002 * cos(1.64606671842  + 25984.36477047710 *T)
    sumY1  +=       0.002 * cos(0.87596565920  + 26191.92914763830 *T)
    sumY1  +=       0.002 * cos(5.44690335938  + 25927.73958967290 *T)
    sumY1  +=       0.002 * cos(3.35831432542  + 26248.55432844250 *T)
    sumY1  +=       0.002 * cos(3.68753780033  + 113341.32408921266 *T)
    sumY1  +=       0.002 * cos(1.57407073796  + 53265.75889182550 *T)
    sumY1  +=       0.002 * cos(6.17867775538  + 51842.19415986250 *T)
    sumY1  +=       0.002 * cos(0.90904857785  + 16066.30967895830 *T)
    sumY1  +=       0.002 * cos(0.52784265993  + 29550.39166492290 *T)
    sumY1  +=       0.002 * cos(1.99418971769  + 22625.90225319250 *T)
    sumY1  +=       0.002 * cos(6.15702102064  + 27037.32256802749 *T)
    sumY1  +=       0.002 * cos(2.64819666416  + 25138.97135008790 *T)
    sumY1  +=       0.002 * cos(2.50748734527  + 157482.77477357027 *T)
    sumY1  +=       0.002 * cos(4.84158872259  + 52817.46366237089 *T)
    sumY1  +=       0.002 * cos(2.22725796457  + 104202.29318664568 *T)
    sumY1  +=       0.002 * cos(6.26671716778  + 25550.85481199529 *T)
    sumY1  +=       0.002 * cos(0.50528140230  + 52388.86156110289 *T)
    sumY1  +=       0.002 * cos(2.07951392438  + 233731.99016146208 *T)
    sumY1  +=       0.002 * cos(1.72258811309  + 54824.50492610490 *T)
    sumY1  +=       0.002 * cos(3.30916137339  + 13655.61667369290 *T)
    sumY1  +=       0.002 * cos(5.77355700704  + 78793.15657233368 *T)
    sumY1  +=       0.002 * cos(0.24904980575  + 168959.21768235945 *T)
    sumY1  +=       0.002 * cos(4.68270285738  + 205260.89400511168 *T)
    sumY1  +=       0.002 * cos(4.78459138847  + 128106.07549751546 *T)
    sumY1  +=       0.002 * cos(0.55814555148  + 50910.48262200550 *T)
    sumY1  +=       0.002 * cos(1.51894844260  + 51859.65823097529 *T)
    sumY1  +=       0.002 * cos(6.23120336216  + 25863.80216335580 *T)
    sumY1  +=       0.002 * cos(2.83592703830  + 53758.00944891489 *T)
    sumY1  +=       0.002 * cos(4.03688605095  + 52492.44197028850 *T)
    sumY1  +=       0.002 * cos(2.18084353651  + 735.63269604830 *T)
    sumY1  +=       0.002 * cos(5.32243619010  + 736.12033101530 *T)
    sumY1  +=       0.002 * cos(3.11385751174  + 1066.25165970650 *T)
    sumY1  +=       0.002 * cos(0.08781202930  + 215474.17955132265 *T)
    sumY1  +=       0.002 * cos(3.93802376645  + 26077.10125879380 *T)
    sumY1  +=       0.002 * cos(4.86719391834  + 26099.19265932160 *T)
    sumY1  +=       0.002 * cos(2.73032719646  + 96356.84030678908 *T)
    sumY1  +=       0.002 * cos(4.39557118339  + 50579.37602338029 *T)
    sumY1  +=       0.002 * cos(3.37995643779  + 176332.48995851149 *T)
    sumY1  +=       0.002 * cos(3.60695289838  + 78896.73698151928 *T)
    sumY1  +=       0.002 * cos(4.97912201809  + 26411.65237571510 *T)
    sumY1  +=       0.002 * cos(3.92261173243  + 87253.42094763846 *T)
    sumY1  +=       0.002 * cos(2.85499495780  + 181659.96606689427 *T)
    sumY1  +=       0.002 * cos(6.13761457132  + 23976.00901071309 *T)
    sumY1  +=       0.002 * cos(4.27004909023  + 102232.60488843489 *T)
    sumY1  +=       0.002 * cos(1.87598894439  + 52483.60772160150 *T)
    sumY1  +=       0.002 * cos(4.99129516853  + 199599.56195595945 *T)
    sumY1  +=       0.002 * cos(5.06914347559  + 25933.88051360590 *T)
    sumY1  +=       0.002 * cos(3.55063804653  + 23439.69213358470 *T)
    sumY1  +=       0.001 * cos(3.06891862205  + 27351.30432163590 *T)
    sumY1  +=       0.001 * cos(5.73629906275  + 24824.98959647950 *T)
    sumY1  +=       0.001 * cos(3.09801335678  + 78697.66498008289 *T)
    sumY1  +=       0.002 * cos(3.94465304931  + 60370.32543384050 *T)
    sumY1  +=       0.001 * cos(1.67825434512  + 51596.36251363310 *T)
    sumY1  +=       0.002 * cos(2.96585642839  + 103815.05187168487 *T)
    sumY1  +=       0.002 * cos(3.05786923280  + 33326.33491569069 *T)
    sumY1  +=       0.002 * cos(5.41433241516  + 51226.87449166209 *T)
    sumY1  +=       0.001 * cos(4.05308794514  + 77307.86586563049 *T)
    sumY1  +=       0.002 * cos(1.32820705945  + 50952.23225985329 *T)
    sumY1  +=       0.002 * cos(0.35546648330  + 65716.98366815608 *T)
    sumY1  +=       0.001 * cos(1.46542460653  + 231348.79714668583 *T)
    sumY1  +=       0.001 * cos(2.79951095321  + 163765.85062356168 *T)
    sumY1  +=       0.002 * cos(1.10119151677  + 34281.93465729930 *T)
    sumY1  +=       0.001 * cos(1.58974685995  + 25234.46294233869 *T)
    sumY1  +=       0.001 * cos(4.42744475216  + 147423.75567281109 *T)
    sumY1  +=       0.002 * cos(3.30510234043  + 26514.25751496669 *T)
    sumY1  +=       0.001 * cos(2.03602296374  + 3442.33112748190 *T)
    sumY1  +=       0.001 * cos(4.23846527178  + 102762.29585352947 *T)
    sumY1  +=       0.001 * cos(2.62660758779  + 189854.24140010288 *T)
    sumY1  +=       0.001 * cos(3.14624674879  + 54879.66625530750 *T)
    sumY1  +=       0.002 * cos(3.37542474980  + 94330.01910368088 *T)
    sumY1  +=       0.002 * cos(4.79787865109  + 22747.04689739089 *T)
    sumY1  +=       0.001 * cos(3.45952976424  + 138319.84867869344 *T)
    sumY1  +=       0.001 * cos(3.22111896048  + 52509.90604140130 *T)
    sumY1  +=       0.001 * cos(1.00099608993  + 35191.56631820829 *T)
    sumY1  +=       0.002 * cos(4.07843204929  + 26713.81715137009 *T)
    sumY1  +=       0.002 * cos(4.72678563551  + 25462.47676674530 *T)
    sumY1  +=       0.001 * cos(1.39808874199  + 76667.27916495448 *T)
    sumY1  +=       0.001 * cos(2.59019797939  + 116783.41139921108 *T)
    sumY1  +=       0.002 * cos(2.54595097754  + 23919.38647677510 *T)
    sumY1  +=       0.002 * cos(6.25926670726  + 28256.90744134030 *T)
    sumY1  +=       0.002 * cos(0.98427002013  + 99024.38027539747 *T)
    sumY1  +=       0.001 * cos(3.96509151964  + 112546.13171698888 *T)
    sumY1  +=       0.002 * cos(5.94020231509  + 26241.43813457550 *T)
    sumY1  +=       0.001 * cos(2.04558297271  + 129329.89333829407 *T)
    sumY1  +=       0.002 * cos(4.89127097596  + 118828.71993201328 *T)
    sumY1  +=       0.001 * cos(5.38890659728  + 49956.80536147810 *T)
    sumY1  +=       0.001 * cos(0.01220959090  + 28199.79727243529 *T)
    sumY1  +=       0.001 * cos(2.93719580530  + 68241.62832713968 *T)
    sumY1  +=       0.001 * cos(0.45657187864  + 103821.67778371868 *T)
    sumY1  +=       0.002 * cos(0.10751328565  + 25441.13612574289 *T)
    sumY1  +=       0.001 * cos(3.04150710062  + 80175.15272542329 *T)
    sumY1  +=       0.002 * cos(5.19554530010  + 23962.26955167850 *T)
    sumY1  +=       0.001 * cos(1.92944311048  + 26934.22979380890 *T)
    sumY1  +=       0.001 * cos(0.85501733527  + 42153.72518556549 *T)
    sumY1  +=       0.001 * cos(5.28557867032  + 50380.79165691089 *T)
    sumY1  +=       0.001 * cos(3.30348432730  + 241562.08269289689 *T)
    sumY1  +=       0.001 * cos(1.53016890390  + 55618.13741063030 *T)
    sumY1  +=       0.001 * cos(0.20150489504  + 92740.81679043909 *T)
    sumY1  +=       0.002 * cos(4.04923139057  + 27567.01353312209 *T)
    sumY1  +=       0.002 * cos(2.43602091587  + 98690.03669966108 *T)
    sumY1  +=       0.001 * cos(5.40866392547  + 173511.17117941825 *T)
    sumY1  +=       0.001 * cos(2.26707127188  + 173511.65881438527 *T)
    sumY1  +=       0.001 * cos(5.30789985746  + 102018.17235594128 *T)
    sumY1  +=       0.001 * cos(2.40978376724  + 49954.19346603489 *T)
    sumY1  +=       0.001 * cos(5.41274037425  + 155461.16718973868 *T)
    sumY1  +=       0.001 * cos(2.80860052221  + 193937.74227184025 *T)
    sumY1  +=       0.001 * cos(2.80241855615  + 26927.11624680809 *T)
    sumY1  +=       0.001 * cos(6.00279912865  + 25249.17767130730 *T)
    sumY1  +=       0.001 * cos(5.41436875302  + 149642.51277699787 *T)
    sumY1  +=       0.001 * cos(4.06699271013  + 112231.45790215228 *T)
    sumY1  +=       0.001 * cos(4.88680216829  + 639.65346883050 *T)
    sumY1  +=       0.001 * cos(4.38938154703  + 28420.85571696270 *T)
    sumY1  +=       0.002 * cos(1.63653705415  + 55503.69812194508 *T)
    sumY1  +=       0.001 * cos(1.19212507226  + 35472.50033216589 *T)
    sumY1  +=       0.001 * cos(6.22889842286  + 48997.90431006430 *T)
    sumY1  +=       0.001 * cos(3.64991019877  + 27006.07724503950 *T)
    sumY1  +=       0.001 * cos(5.15530748603  + 25170.21667307590 *T)
    sumY1  +=       0.001 * cos(0.21655513602  + 106263.05586699749 *T)
    sumY1  +=       0.001 * cos(3.15004684273  + 90989.40667698107 *T)
    sumY1  +=       0.001 * cos(3.42636128355  + 50690.06997956669 *T)
    sumY1  +=       0.001 * cos(1.90636858110  + 77836.86741636308 *T)
    sumY1  +=       0.001 * cos(1.74306827310  + 125112.28341697170 *T)
    sumY1  +=       0.001 * cos(0.83861407900  + 26610.23674218449 *T)
    sumY1  +=       0.001 * cos(6.19974077735  + 143980.69309287865 *T)
    sumY1  +=       0.001 * cos(2.30573879835  + 153084.60008699627 *T)
    sumY1  +=       0.001 * cos(2.74311464361  + 45404.85186441930 *T)
    sumY1  +=       0.001 * cos(4.11780710445  + 2332.95257538850 *T)
    sumY1  +=       0.001 * cos(2.50393985667  + 76152.30292882849 *T)
    sumY1  +=       0.001 * cos(3.70443537393  + 45494.33761226529 *T)
    sumY1  +=       0.001 * cos(1.90974220877  + 24925.18461968289 *T)
    sumY1  +=       0.001 * cos(3.79471530054  + 50593.60311738189 *T)
    sumY1  +=       0.001 * cos(0.35163928497  + 115674.03284711768 *T)
    sumY1  +=       0.001 * cos(4.64759915605  + 151200.18655854589 *T)
    sumY1  +=       0.001 * cos(6.12692757039  + 84308.18182358149 *T)
    sumY1  +=       0.001 * cos(2.49793910811  + 129379.88996019828 *T)
    sumY1  +=       0.001 * cos(4.24960627112  + 24073.16528725990 *T)
    sumY1  +=       0.001 * cos(4.55561141368  + 28103.12863085550 *T)
    sumY1  +=       0.001 * cos(3.64820977115  + 19406.43447069109 *T)
    sumY1  +=       0.001 * cos(5.22888939483  + 24844.65935737930 *T)
    sumY1  +=       0.001 * cos(3.57632828997  + 27331.63456073610 *T)
    sumY1  +=       0.001 * cos(0.38918442414  + 154408.89880655028 *T)
    sumY1  +=       0.001 * cos(4.58125009028  + 163298.37326817427 *T)
    sumY1  +=       0.001 * cos(4.05552152915  + 58220.27868200730 *T)
    sumY1  +=       0.001 * cos(0.81056243735  + 97670.63153038068 *T)
    sumY1  +=       0.001 * cos(1.70406295246  + 22759.52366781050 *T)
    sumY1  +=       0.001 * cos(0.17160219777  + 24395.98128955530 *T)
    sumY1  +=       0.001 * cos(4.19916792884  + 92741.30442540610 *T)
    sumY1  +=       0.001 * cos(0.46762289675  + 128850.19899510368 *T)
    sumY1  +=       0.001 * cos(3.27962382087  + 25984.56654987209 *T)
    sumY1  +=       0.001 * cos(3.06090263008  + 24357.02460612510 *T)
    sumY1  +=       0.001 * cos(5.74431505472  + 27819.26931199030 *T)
    sumY1  +=       0.001 * cos(1.47564589622  + 183724.45672562928 *T)
    sumY1  +=       0.001 * cos(2.32406253516  + 26624.46383618609 *T)
    sumY1  +=       0.001 * cos(0.16830959152  + 25455.36321974450 *T)
    sumY1  +=       0.001 * cos(3.30990224511  + 25454.87558477749 *T)
    sumY1  +=       0.001 * cos(0.20081607790  + 189385.78877478145 *T)
    sumY1  +=       0.001 * cos(3.34240873149  + 189386.27640974848 *T)
    sumY1  +=       0.001 * cos(5.05200956798  + 209659.06869168568 *T)
    sumY1  +=       0.001 * cos(2.14910744070  + 109.96250373590 *T)
    sumY1  +=       0.001 * cos(4.58272161819  + 82815.41910429788 *T)
    sumY1  +=       0.001 * cos(0.56829620714  + 147423.26803784407 *T)
    sumY1  +=       0.001 * cos(3.84356790046  + 23704.95394428149 *T)
    sumY1  +=       0.001 * cos(4.96164978434  + 28471.33997383390 *T)
    sumY1  +=       0.001 * cos(3.64905591301  + 51440.17358710009 *T)
    sumY1  +=       0.001 * cos(1.10275035993  + 25035.39094090229 *T)
    sumY1  +=       0.001 * cos(0.92306365199  + 26555.62431444510 *T)
    sumY1  +=       0.001 * cos(2.45569831568  + 26177.63270690369 *T)
    sumY1  +=       0.001 * cos(0.06633406194  + 25998.66121121170 *T)
    sumY1  +=       0.001 * cos(0.41633866770  + 233681.99353955788 *T)
    sumY1  +=       0.001 * cos(1.60238548900  + 21650.63275068411 *T)
    sumY1  +=       0.001 * cos(0.91964688862  + 30525.66116743129 *T)
    sumY1  +=       0.001 * cos(3.50363747347  + 111590.04434041328 *T)
    sumY1  +=       0.001 * cos(2.57975917809  + 24203.24579564031 *T)
    sumY1  +=       0.001 * cos(6.22545850671  + 27973.04812247509 *T)
    sumY1  +=       0.001 * cos(5.61388226798  + 99799.41525175449 *T)
    sumY1  +=       0.001 * cos(0.90114538219  + 128106.56313248246 *T)
    sumY1  +=       0.001 * cos(2.25118383839  + 104198.07757329548 *T)
    sumY1  +=       0.001 * cos(0.23482210375  + 132350.95900857169 *T)
    sumY1  +=       0.001 * cos(5.46489469781  + 6680.98103591609 *T)
    sumY1  +=       0.001 * cos(4.59107671722  + 26202.09861277591 *T)
    sumY1  +=       0.001 * cos(1.28948419377  + 50167.49256147289 *T)
    sumY1  +=       0.001 * cos(4.70995069593  + 313054.59388140670 *T)
    sumY1  +=       0.001 * cos(1.61351002182  + 131548.65044248087 *T)
    sumY1  +=       0.001 * cos(0.21134212997  + 85502.14119883909 *T)
    sumY1  +=       0.001 * cos(3.59852250423  + 51542.77872635169 *T)
    sumY1  +=       0.001 * cos(4.35050776234  + 102018.65999090829 *T)
    sumY1  +=       0.001 * cos(5.00190822272  + 51322.36608391289 *T)
    sumY1  +=       0.001 * cos(6.26093145675  + 78413.51643934069 *T)
    sumY1  +=       0.001 * cos(4.71083057280  + 156531.54400228668 *T)
    sumY1  +=       0.001 * cos(2.77611181285  + 105306.96849042187 *T)
    sumY1  +=       0.001 * cos(4.15505663410  + 35077.12702952309 *T)
    sumY1  +=       0.001 * cos(2.66891943800  + 155467.79310177249 *T)
    sumY1  +=       0.001 * cos(2.90242040099  + 45290.41257573408 *T)
    sumY1  +=       0.001 * cos(5.09454263682  + 1162.23088692429 *T)
    sumY1  +=       0.001 * cos(5.85832218706  + 77211.19722405069 *T)
    sumY1  +=       0.001 * cos(2.94258825469  + 25152.21697834890 *T)
    sumY1  +=       0.001 * cos(5.86262943011  + 27024.07693976651 *T)
    sumY1  +=       0.001 * cos(2.69263645947  + 247223.41474204903 *T)
    sumY1  +=       0.001 * cos(2.39382184066  + 183674.94773869208 *T)
    sumY1  +=       0.001 * cos(4.71424763358  + 25131.37016812009 *T)
    sumY1  +=       0.001 * cos(4.49841046349  + 26126.27999469550 *T)
    sumY1  +=       0.001 * cos(4.30680722131  + 26050.01392341990 *T)
    sumY1  +=       0.001 * cos(2.54564315044  + 71980.38975724769 *T)
    sumY1  +=       0.001 * cos(3.77812730736  + 27388.97377726269 *T)
    sumY1  +=       0.001 * cos(5.02709037744  + 24787.32014085270 *T)
    sumY1  +=       0.001 * cos(5.22928116569  + 155475.39428374029 *T)
    sumY1  +=       0.001 * cos(2.23214428102  + 51756.07782178969 *T)
    sumY1  +=       0.001 * cos(3.75930601793  + 54086.76194888209 *T)
    sumY1  +=       0.001 * cos(5.71244644048  + 94138.08320260228 *T)
    sumY1  +=       0.001 * cos(3.60259923599  + 24602.85625235450 *T)
    sumY1  +=       0.001 * cos(5.20261844881  + 27573.43766576090 *T)
    sumY1  +=       0.001 * cos(5.97633149627  + 121335.36489626988 *T)
    sumY1  +=       0.001 * cos(5.11853435105  + 198490.18340386607 *T)
    sumY1  +=       0.001 * cos(2.41668452027  + 181555.69624334667 *T)
    sumY1  +=       0.001 * cos(3.56429469477  + 156534.77621392949 *T)
    sumY1  +=       0.001 * cos(0.16534790503  + 12725.69725225850 *T)
    sumY1  +=       0.001 * cos(5.23354450332  + 130006.04778747768 *T)
    sumY1  +=       0.001 * cos(3.16983107219  + 89586.12970554348 *T)
    sumY1  +=       0.001 * cos(5.31492826690  + 131394.87163199607 *T)
    sumY1  +=       0.001 * cos(2.18749112623  + 77039.64776646046 *T)

    # MERCURY Y Coord
    sumY2  = 0
    sumY2  +=   20000.263 * cos(4.39813856862  + 26088.14695905770 *T)
    sumY2  +=    8268.782 * cos(5.12831925217  + 0.24381748350 *T)
    sumY2  +=    2498.175 * cos(1.81785035615  + 52176.05010063190 *T)
    sumY2  +=     949.347 * cos(5.50772040452  + 78263.95324220609 *T)
    sumY2  +=     776.108 * cos(0.81508927377  + 26087.65932409069 *T)
    sumY2  +=     371.109 * cos(2.46339631518  + 104351.85638378028 *T)
    sumY2  +=     130.259 * cos(5.58371836146  + 130439.75952535449 *T)
    sumY2  +=      69.902 * cos(3.59145684280  + 52175.56246566489 *T)
    sumY2  +=      42.865 * cos(2.38418074575  + 156527.66266692869 *T)
    sumY2  +=      13.564 * cos(5.45308914681  + 182615.56580850287 *T)
    sumY2  +=      11.344 * cos(0.18546869670  + 78263.46560723908 *T)
    sumY2  +=       4.184 * cos(2.23177226385  + 208703.46895007708 *T)
    sumY2  +=       2.952 * cos(5.92358038862  + 24978.76840696430 *T)
    sumY2  +=       2.715 * cos(2.62856029278  + 27197.52551115110 *T)
    sumY2  +=       2.412 * cos(3.20116782795  + 104351.36874881329 *T)
    sumY2  +=       1.268 * cos(5.28987764263  + 234791.37209165128 *T)
    sumY2  +=       1.017 * cos(2.76352422322  + 51066.67154853850 *T)
    sumY2  +=       0.591 * cos(0.02739486208  + 130439.27189038748 *T)
    sumY2  +=       0.372 * cos(5.53602920422  + 53285.42865272530 *T)
    sumY2  +=       0.379 * cos(2.06228553945  + 260879.27523322543 *T)
    sumY2  +=       0.353 * cos(5.84872228977  + 77154.57469011268 *T)
    sumY2  +=       0.353 * cos(3.64262209095  + 1109.13473460990 *T)
    sumY2  +=       0.259 * cos(2.45927570893  + 20426.81490990550 *T)
    sumY2  +=       0.259 * cos(3.62787658106  + 1059.62574767270 *T)
    sumY2  +=       0.198 * cos(6.08219469431  + 27043.74670066630 *T)
    sumY2  +=       0.178 * cos(4.39645894496  + 31749.47900820990 *T)
    sumY2  +=       0.161 * cos(5.71880342023  + 1589.31671276730 *T)
    sumY2  +=       0.148 * cos(1.40374146640  + 4552.19731454230 *T)
    sumY2  +=       0.145 * cos(0.02368396611  + 21536.19346199890 *T)
    sumY2  +=       0.158 * cos(3.17661979051  + 156527.17503196167 *T)
    sumY2  +=       0.169 * cos(2.43702666403  + 25132.54721744910 *T)
    sumY2  +=       0.134 * cos(4.36077570655  + 25558.45599396310 *T)
    sumY2  +=       0.112 * cos(5.11701683724  + 286967.17837479961 *T)
    sumY2  +=       0.111 * cos(4.08624415081  + 26617.83792415230 *T)
    sumY2  +=       0.122 * cos(6.20816277459  + 46514.71805147970 *T)
    sumY2  +=       0.125 * cos(6.10874998399  + 23869.38985487090 *T)
    sumY2  +=       0.111 * cos(0.71893259930  + 14765.48286075331 *T)
    sumY2  +=       0.106 * cos(3.19958057200  + 51116.66817044270 *T)
    sumY2  +=       0.120 * cos(2.63555930925  + 103242.47783168688 *T)
    sumY2  +=       0.113 * cos(1.00068732157  + 27147.52888924690 *T)
    sumY2  +=       0.105 * cos(2.62816055600  + 28306.90406324449 *T)
    sumY2  +=       0.086 * cos(3.50491777711  + 19317.43635781211 *T)
    sumY2  +=       0.103 * cos(1.19071414957  + 5661.08823166870 *T)
    sumY2  +=       0.067 * cos(4.69029936211  + 9104.15081160110 *T)
    sumY2  +=       0.076 * cos(0.32408375390  + 529.93478257810 *T)
    sumY2  +=       0.066 * cos(2.87399773118  + 50586.97720534809 *T)
    sumY2  +=       0.066 * cos(2.27463939345  + 3442.81876244891 *T)
    sumY2  +=       0.061 * cos(1.26603377840  + 37410.81105736210 *T)
    sumY2  +=       0.059 * cos(3.16922770991  + 72602.62119305388 *T)
    sumY2  +=       0.074 * cos(0.14312537861  + 22645.57201409229 *T)
    sumY2  +=       0.056 * cos(0.87699245253  + 52705.74106572650 *T)
    sumY2  +=       0.061 * cos(1.98731213334  + 30640.10045611650 *T)
    sumY2  +=       0.053 * cos(3.72430094233  + 640.14110379750 *T)
    sumY2  +=       0.056 * cos(5.15278563625  + 32858.85756030330 *T)
    sumY2  +=       0.062 * cos(5.26544507482  + 51220.45035902330 *T)
    sumY2  +=       0.050 * cos(2.61090296354  + 53131.64984224050 *T)
    sumY2  +=       0.046 * cos(2.37745190465  + 27677.21985434150 *T)
    sumY2  +=       0.045 * cos(0.02572397068  + 25973.70767037247 *T)
    sumY2  +=       0.045 * cos(2.48433200554  + 26202.58624774291 *T)
    sumY2  +=       0.044 * cos(0.04666322888  + 182615.07817353585 *T)
    sumY2  +=       0.046 * cos(4.10779409868  + 27197.03787618409 *T)
    sumY2  +=       0.046 * cos(4.21737357836  + 53235.43203082109 *T)
    sumY2  +=       0.058 * cos(0.51456998627  + 57837.38214978410 *T)
    sumY2  +=       0.043 * cos(4.72097651637  + 529.44714761109 *T)
    sumY2  +=       0.042 * cos(5.81334759172  + 47624.09660357310 *T)
    sumY2  +=       0.041 * cos(1.35360879608  + 26301.44605449570 *T)
    sumY2  +=       0.039 * cos(3.95217511506  + 40853.38600232750 *T)
    sumY2  +=       0.038 * cos(1.01923360930  + 25874.84786361970 *T)
    sumY2  +=       0.045 * cos(2.30681711799  + 29530.72190402310 *T)
    sumY2  +=       0.045 * cos(3.25992009502  + 10213.52936369450 *T)
    sumY2  +=       0.035 * cos(0.20035334587  + 2119.00767786191 *T)
    sumY2  +=       0.034 * cos(5.21776132031  + 56728.00359769070 *T)
    sumY2  +=       0.040 * cos(0.28616417785  + 24978.28077199729 *T)
    sumY2  +=       0.040 * cos(5.69976040495  + 129330.38097326108 *T)
    sumY2  +=       0.033 * cos(1.88435460588  + 313055.08151637384 *T)
    sumY2  +=       0.039 * cos(0.71277136304  + 955.35592412510 *T)
    sumY2  +=       0.031 * cos(5.95336552931  + 6770.46678376209 *T)
    sumY2  +=       0.038 * cos(3.30701415262  + 2218.51328670329 *T)
    sumY2  +=       0.031 * cos(5.89629431306  + 12566.39551746630 *T)
    sumY2  +=       0.032 * cos(2.89503281988  + 49957.29299644510 *T)
    sumY2  +=       0.029 * cos(4.08747500636  + 426.84200835950 *T)
    sumY2  +=       0.028 * cos(3.89028250879  + 43072.14310651430 *T)
    sumY2  +=       0.025 * cos(0.01043820952  + 98690.52433462808 *T)
    sumY2  +=       0.030 * cos(1.58323308741  + 79373.33179429949 *T)
    sumY2  +=       0.023 * cos(3.87566855556  + 78793.64420730069 *T)
    sumY2  +=       0.030 * cos(5.69174979475  + 77204.57131201689 *T)
    sumY2  +=       0.020 * cos(2.91659016078  + 11322.42028082089 *T)
    sumY2  +=       0.027 * cos(5.66737798141  + 54394.80720481869 *T)
    sumY2  +=       0.023 * cos(1.92399700068  + 77308.35350059748 *T)
    sumY2  +=       0.019 * cos(4.33796152575  + 83925.28529135829 *T)
    sumY2  +=       0.022 * cos(0.31026735565  + 45405.33949938630 *T)
    sumY2  +=       0.020 * cos(0.76233233344  + 36301.43250526870 *T)
    sumY2  +=       0.018 * cos(1.06684593234  + 79323.33517239528 *T)
    sumY2  +=       0.021 * cos(1.06633987406  + 13656.10430865991 *T)
    sumY2  +=       0.020 * cos(3.26662752657  + 33326.82255065770 *T)
    sumY2  +=       0.018 * cos(0.08918897822  + 52156.38033973211 *T)
    sumY2  +=       0.023 * cos(2.56517709904  + 41962.76455442090 *T)
    sumY2  +=       0.018 * cos(4.49923548064  + 52389.34919606990 *T)
    sumY2  +=       0.017 * cos(1.90164043236  + 58946.76070187749 *T)
    sumY2  +=       0.016 * cos(0.07721057973  + 23754.95056618569 *T)
    sumY2  +=       0.019 * cos(0.79767727082  + 39609.89840064910 *T)
    sumY2  +=       0.017 * cos(4.96752330955  + 51536.15281431789 *T)
    sumY2  +=       0.018 * cos(5.05941872394  + 18849.47136745770 *T)
    sumY2  +=       0.015 * cos(5.67105202528  + 76674.88034692229 *T)
    sumY2  +=       0.015 * cos(0.18434972499  + 48733.47515566649 *T)
    sumY2  +=       0.015 * cos(2.77916270049  + 853.44019923550 *T)
    sumY2  +=       0.014 * cos(2.04393411320  + 82815.90673926489 *T)
    sumY2  +=       0.014 * cos(3.45421098562  + 52179.93143598990 *T)
    sumY2  +=       0.014 * cos(0.12812152173  + 15874.86141284670 *T)
    sumY2  +=       0.014 * cos(0.79649939505  + 66941.28914390170 *T)
    sumY2  +=       0.015 * cos(3.92393517562  + 16984.23996494010 *T)
    sumY2  +=       0.015 * cos(3.48624645549  + 38654.29865904050 *T)
    sumY2  +=       0.013 * cos(5.35926600957  + 53765.12299591570 *T)
    sumY2  +=       0.013 * cos(3.19246765373  + 208702.98131511008 *T)
    sumY2  +=       0.013 * cos(2.25098683682  + 50057.28624025350 *T)
    sumY2  +=       0.016 * cos(0.48148728540  + 51646.35913553730 *T)
    sumY2  +=       0.012 * cos(2.40833345233  + 28421.34335192970 *T)
    sumY2  +=       0.013 * cos(2.47413691628  + 155418.28411483529 *T)
    sumY2  +=       0.016 * cos(5.37235452634  + 55618.62504559730 *T)
    sumY2  +=       0.012 * cos(2.05490297780  + 5661.57586663570 *T)
    sumY2  +=       0.015 * cos(5.46754476025  + 52195.71986153169 *T)
    sumY2  +=       0.014 * cos(1.23967074882  + 26514.74514993370 *T)
    sumY2  +=       0.014 * cos(1.73016204603  + 110013.18843293248 *T)
    sumY2  +=       0.013 * cos(2.13522240551  + 213.05527795450 *T)
    sumY2  +=       0.014 * cos(0.53482285581  + 26728.04424537170 *T)
    sumY2  +=       0.013 * cos(2.09871625287  + 52172.16876527390 *T)
    sumY2  +=       0.011 * cos(3.89215785881  + 63498.71419893629 *T)
    sumY2  +=       0.011 * cos(5.32469405612  + 79219.55298381469 *T)
    sumY2  +=       0.010 * cos(3.09632978067  + 103292.47445359109 *T)
    sumY2  +=       0.011 * cos(0.49318054067  + 31748.99137324289 *T)
    sumY2  +=       0.014 * cos(1.76374143598  + 6283.31966747490 *T)
    sumY2  +=       0.010 * cos(5.54944886749  + 52290.48938931711 *T)
    sumY2  +=       0.012 * cos(4.17529024605  + 13521.50762410789 *T)
    sumY2  +=       0.009 * cos(1.46108642538  + 35192.05395317530 *T)
    sumY2  +=       0.009 * cos(5.95543898844  + 76045.19613801930 *T)
    sumY2  +=       0.011 * cos(4.31962648376  + 62389.33564684289 *T)
    sumY2  +=       0.010 * cos(2.59338730813  + 51749.45190975589 *T)
    sumY2  +=       0.009 * cos(3.09241629007  + 124778.42747620228 *T)
    sumY2  +=       0.009 * cos(4.94047648268  + 339142.98465794802 *T)
    sumY2  +=       0.009 * cos(3.18433513864  + 78244.28348130629 *T)
    sumY2  +=       0.009 * cos(4.21342321430  + 20426.32727493849 *T)
    sumY2  +=       0.011 * cos(1.35641733071  + 38520.18960945550 *T)
    sumY2  +=       0.010 * cos(4.87324110048  + 16983.75232997309 *T)
    sumY2  +=       0.009 * cos(5.87248180566  + 1109.62236957690 *T)
    sumY2  +=       0.009 * cos(3.08529390196  + 52061.61081194667 *T)
    sumY2  +=       0.009 * cos(0.58390742179  + 104881.54734887488 *T)
    sumY2  +=       0.008 * cos(0.01049678658  + 68050.66769599509 *T)
    sumY2  +=       0.008 * cos(2.44377299382  + 2333.44021035551 *T)
    sumY2  +=       0.008 * cos(2.07553269199  + 12431.79883291429 *T)
    sumY2  +=       0.008 * cos(3.14068514277  + 28206.91081943610 *T)
    sumY2  +=       0.008 * cos(4.93667216370  + 136101.09157450669 *T)
    sumY2  +=       0.008 * cos(3.56776272307  + 213.54291292150 *T)
    sumY2  +=       0.008 * cos(6.24685210587  + 24499.07406377390 *T)
    sumY2  +=       0.008 * cos(4.90871772456  + 103396.25664217168 *T)
    sumY2  +=       0.007 * cos(3.26004517300  + 52168.93655363109 *T)
    sumY2  +=       0.009 * cos(2.25308643231  + 21535.70582703189 *T)
    sumY2  +=       0.007 * cos(0.96066249738  + 53284.94101775829 *T)
    sumY2  +=       0.007 * cos(3.34414054226  + 105461.23493587368 *T)
    sumY2  +=       0.007 * cos(0.56178116893  + 69160.04624808849 *T)
    sumY2  +=       0.009 * cos(4.89006567905  + 7994.77225950771 *T)
    sumY2  +=       0.009 * cos(1.87204677749  + 73711.99974514729 *T)
    sumY2  +=       0.008 * cos(6.11573252209  + 59414.72569223190 *T)
    sumY2  +=       0.008 * cos(2.34935594859  + 52183.16364763270 *T)
    sumY2  +=       0.007 * cos(0.38833894888  + 64742.20180061470 *T)
    sumY2  +=       0.007 * cos(0.26691004914  + 78267.83457756409 *T)
    sumY2  +=       0.007 * cos(3.93557943129  + 26107.81671995749 *T)
    sumY2  +=       0.007 * cos(1.72211044919  + 26068.47719815791 *T)
    sumY2  +=       0.006 * cos(0.91009879641  + 25028.76502886850 *T)
    sumY2  +=       0.007 * cos(4.52931279940  + 51962.75100519390 *T)
    sumY2  +=       0.007 * cos(3.63772636438  + 18208.05780571871 *T)
    sumY2  +=       0.008 * cos(1.62956568948  + 44937.37450903190 *T)
    sumY2  +=       0.008 * cos(3.38799810242  + 51066.18391357149 *T)
    sumY2  +=       0.006 * cos(4.15575122576  + 105411.23831396949 *T)
    sumY2  +=       0.007 * cos(3.29006350511  + 65697.80154222329 *T)
    sumY2  +=       0.006 * cos(1.27230132469  + 88477.23878841709 *T)
    sumY2  +=       0.006 * cos(1.32946065086  + 78477.25233764409 *T)
    sumY2  +=       0.006 * cos(4.53641896383  + 52602.64829150790 *T)
    sumY2  +=       0.007 * cos(2.19484677855  + 78283.62300310588 *T)
    sumY2  +=       0.006 * cos(0.97859883313  + 19805.07110906630 *T)
    sumY2  +=       0.005 * cos(5.12029859122  + 108903.80988083908 *T)
    sumY2  +=       0.006 * cos(0.51108107596  + 129380.37759516528 *T)
    sumY2  +=       0.005 * cos(3.17438180692  + 74821.37829724069 *T)
    sumY2  +=       0.007 * cos(3.90084269972  + 45892.97425064050 *T)
    sumY2  +=       0.005 * cos(5.53465282046  + 1059.13811270570 *T)
    sumY2  +=       0.005 * cos(0.53945271085  + 26084.26562369970 *T)
    sumY2  +=       0.005 * cos(5.12008322277  + 26092.02829441570 *T)
    sumY2  +=       0.007 * cos(3.38101832696  + 71493.24264096050 *T)
    sumY2  +=       0.005 * cos(2.15419212921  + 81706.52818717150 *T)
    sumY2  +=       0.005 * cos(5.99431638416  + 51322.85371887989 *T)
    sumY2  +=       0.005 * cos(5.10141066049  + 33968.23611239669 *T)
    sumY2  +=       0.005 * cos(6.07497929534  + 78256.83969520529 *T)
    sumY2  +=       0.005 * cos(4.51394707071  + 26617.35028918529 *T)
    sumY2  +=       0.006 * cos(5.11905884366  + 78260.07190684808 *T)
    sumY2  +=       0.004 * cos(2.02155104150  + 25661.54876818170 *T)
    sumY2  +=       0.005 * cos(3.63607614834  + 94138.57083756928 *T)
    sumY2  +=       0.004 * cos(1.02197474906  + 22645.08437912529 *T)
    sumY2  +=       0.005 * cos(3.86831492333  + 93029.19228547587 *T)
    sumY2  +=       0.004 * cos(6.23205767407  + 22760.01130277751 *T)
    sumY2  +=       0.004 * cos(2.54287848992  + 29416.28261533789 *T)
    sumY2  +=       0.004 * cos(5.04658501458  + 76145.18938182769 *T)
    sumY2  +=       0.004 * cos(5.53028555901  + 181506.18725640947 *T)
    sumY2  +=       0.004 * cos(4.30468862075  + 13521.99525907490 *T)
    sumY2  +=       0.004 * cos(5.69769574613  + 99799.90288672148 *T)
    sumY2  +=       0.004 * cos(2.07912078316  + 79853.02613748988 *T)
    sumY2  +=       0.005 * cos(3.94088501872  + 44181.52165860769 *T)
    sumY2  +=       0.004 * cos(5.75049226379  + 4551.70967957530 *T)
    sumY2  +=       0.004 * cos(5.78252204456  + 32371.22280904910 *T)
    sumY2  +=       0.004 * cos(6.13397098904  + 32858.36992533629 *T)
    sumY2  +=       0.004 * cos(5.47303342696  + 54509.24649350390 *T)
    sumY2  +=       0.004 * cos(3.77296926059  + 155468.28073673949 *T)
    sumY2  +=       0.003 * cos(1.76022764417  + 162188.99471608087 *T)
    sumY2  +=       0.003 * cos(3.47863732548  + 24498.58642880689 *T)
    sumY2  +=       0.005 * cos(5.41867016378  + 78271.06678920689 *T)
    sumY2  +=       0.004 * cos(0.03915152889  + 234790.88445668429 *T)
    sumY2  +=       0.003 * cos(3.63049034768  + 77734.26227711148 *T)
    sumY2  +=       0.004 * cos(5.73459041878  + 7879.84533585549 *T)
    sumY2  +=       0.004 * cos(4.32304044619  + 64608.09275102969 *T)
    sumY2  +=       0.004 * cos(6.08591965357  + 51116.18053547569 *T)
    sumY2  +=       0.003 * cos(3.39565834350  + 28306.41642827749 *T)
    sumY2  +=       0.004 * cos(4.14203046511  + 39744.00745023410 *T)
    sumY2  +=       0.003 * cos(6.25044565160  + 104332.18662288050 *T)
    sumY2  +=       0.004 * cos(2.32061218672  + 80482.71034639288 *T)
    sumY2  +=       0.003 * cos(3.42862683678  + 37410.32342239509 *T)
    sumY2  +=       0.004 * cos(6.26655525096  + 24864.32911827909 *T)
    sumY2  +=       0.003 * cos(0.81932036845  + 103822.16541868568 *T)
    sumY2  +=       0.004 * cos(2.51942976235  + 27311.96479983631 *T)
    sumY2  +=       0.003 * cos(0.55238739809  + 120226.47397914348 *T)
    sumY2  +=       0.003 * cos(2.73238956336  + 102133.09927959349 *T)
    sumY2  +=       0.003 * cos(6.16210003244  + 150866.33061777649 *T)
    sumY2  +=       0.003 * cos(3.47088315334  + 90830.10494218887 *T)
    sumY2  +=       0.003 * cos(1.70583242978  + 77624.05595589208 *T)
    sumY2  +=       0.003 * cos(1.64244671238  + 129484.15978374588 *T)
    sumY2  +=       0.003 * cos(3.01438324827  + 125887.80602829569 *T)
    sumY2  +=       0.003 * cos(3.59751110770  + 130969.45049044908 *T)
    sumY2  +=       0.003 * cos(3.38263477770  + 58459.12595062330 *T)
    sumY2  +=       0.002 * cos(2.33982742399  + 6282.83203250789 *T)
    sumY2  +=       0.003 * cos(5.89330250493  + 26941.34334080970 *T)
    sumY2  +=       0.002 * cos(3.22048008840  + 3327.89183879669 *T)
    sumY2  +=       0.003 * cos(4.49509466805  + 18093.13088206649 *T)
    sumY2  +=       0.002 * cos(2.81219352708  + 25132.05958248210 *T)
    sumY2  +=       0.003 * cos(4.76356263080  + 26095.26050605850 *T)
    sumY2  +=       0.003 * cos(4.55467549289  + 71025.27765060609 *T)
    sumY2  +=       0.003 * cos(4.71882234896  + 43071.65547154729 *T)
    sumY2  +=       0.002 * cos(4.39575489958  + 114565.14192999128 *T)
    sumY2  +=       0.002 * cos(1.23106381310  + 71980.87739221469 *T)
    sumY2  +=       0.002 * cos(0.75134917742  + 91785.70468379748 *T)
    sumY2  +=       0.002 * cos(0.91717698933  + 2648.69864295650 *T)
    sumY2  +=       0.002 * cos(2.32776941660  + 78378.39253089130 *T)
    sumY2  +=       0.003 * cos(3.33005216581  + 104355.73771913828 *T)
    sumY2  +=       0.002 * cos(0.88668781336  + 26081.03341205689 *T)
    sumY2  +=       0.002 * cos(1.33604738376  + 77197.45776501608 *T)
    sumY2  +=       0.002 * cos(1.70708347774  + 365230.88779952226 *T)
    sumY2  +=       0.002 * cos(1.39529664977  + 78690.55143308209 *T)
    sumY2  +=       0.003 * cos(3.63831731324  + 52815.94738694590 *T)
    sumY2  +=       0.002 * cos(2.08527920288  + 25448.24967274370 *T)
    sumY2  +=       0.003 * cos(5.17799187468  + 77837.35505133009 *T)
    sumY2  +=       0.002 * cos(5.22616840288  + 104371.52614468009 *T)
    sumY2  +=       0.002 * cos(2.61518622291  + 25028.27739390149 *T)
    sumY2  +=       0.002 * cos(1.38915456619  + 11322.90791578790 *T)
    sumY2  +=       0.002 * cos(0.72086659814  + 77211.68485901768 *T)
    sumY2  +=       0.002 * cos(2.08668332966  + 105307.45612538888 *T)
    sumY2  +=       0.002 * cos(2.01988438935  + 78050.65414676809 *T)
    sumY2  +=       0.002 * cos(1.66916828850  + 27043.25906569930 *T)
    sumY2  +=       0.002 * cos(2.76016995044  + 104344.74283677948 *T)
    sumY2  +=       0.002 * cos(0.83226802149  + 46514.23041651269 *T)
    sumY2  +=       0.002 * cos(0.22774400791  + 77154.08705514569 *T)
    sumY2  +=       0.002 * cos(0.91172072873  + 26725.14323108190 *T)
    sumY2  +=       0.002 * cos(1.61031164889  + 25451.15068703350 *T)
    sumY2  +=       0.002 * cos(0.95099409511  + 131499.14145554369 *T)
    sumY2  +=       0.002 * cos(4.49976969859  + 61279.95709474950 *T)
    sumY2  +=       0.002 * cos(5.98345605195  + 89586.61734051049 *T)
    sumY2  +=       0.002 * cos(2.85946010033  + 30639.61282114949 *T)
    sumY2  +=       0.002 * cos(0.15838364848  + 97581.14578253467 *T)
    sumY2  +=       0.002 * cos(6.14579824758  + 78149.51395352086 *T)
    sumY2  +=       0.002 * cos(5.58929876175  + 7238.43177411650 *T)
    sumY2  +=       0.002 * cos(0.83010213570  + 426.35437339250 *T)
    sumY2  +=       0.002 * cos(5.20457051939  + 107794.43132874568 *T)
    sumY2  +=       0.002 * cos(1.91271152917  + 47623.60896860609 *T)
    sumY2  +=       0.002 * cos(4.42235381341  + 104565.15547921829 *T)
    sumY2  +=       0.002 * cos(0.42021990927  + 44295.96094729290 *T)
    sumY2  +=       0.002 * cos(3.90488175057  + 18093.61851703350 *T)
    sumY2  +=       0.002 * cos(0.51702864254  + 10213.04172872750 *T)
    sumY2  +=       0.002 * cos(2.19510126559  + 104358.96993078108 *T)
    sumY2  +=       0.002 * cos(1.87197741484  + 104347.97504842227 *T)
    sumY2  +=       0.002 * cos(1.89656096023  + 134991.71302241329 *T)
    sumY2  +=       0.002 * cos(3.81068190615  + 129910.06856025988 *T)
    sumY2  +=       0.002 * cos(3.48333455064  + 57836.89451481709 *T)
    sumY2  +=       0.002 * cos(5.38520038611  + 25557.96835899609 *T)
    sumY2  +=       0.002 * cos(6.12255852160  + 54294.81396101029 *T)
    sumY2  +=       0.002 * cos(1.91069399856  + 102762.78348849648 *T)
    sumY2  +=       0.001 * cos(5.94439573345  + 7238.91940908350 *T)
    sumY2  +=       0.002 * cos(2.71520511595  + 25934.36814857290 *T)
    sumY2  +=       0.001 * cos(6.12042545447  + 151975.70916986989 *T)
    sumY2  +=       0.002 * cos(2.95370958319  + 26241.92576954250 *T)
    sumY2  +=       0.001 * cos(6.21917309565  + 131549.13807744789 *T)
    sumY2  +=       0.001 * cos(2.63165117521  + 23968.89546371229 *T)
    sumY2  +=       0.002 * cos(1.38299478051  + 25021.65148186770 *T)
    sumY2  +=       0.002 * cos(4.14961473905  + 27154.64243624770 *T)
    sumY2  +=       0.001 * cos(0.59279120743  + 181556.18387831369 *T)
    sumY2  +=       0.001 * cos(4.95292728782  + 27140.41534224609 *T)
    sumY2  +=       0.001 * cos(0.19355831516  + 149.80701461810 *T)
    sumY2  +=       0.001 * cos(5.03717522016  + 50586.48957038109 *T)
    sumY2  +=       0.001 * cos(2.46166203858  + 85502.62883380610 *T)
    sumY2  +=       0.001 * cos(3.12377250466  + 1052.51220067191 *T)
    sumY2  +=       0.001 * cos(0.12220892820  + 84547.02909219750 *T)
    sumY2  +=       0.001 * cos(4.99498518550  + 41962.27691945389 *T)
    sumY2  +=       0.002 * cos(5.81835190108  + 1066.73929467350 *T)
    sumY2  +=       0.001 * cos(0.19689603888  + 33326.33491569069 *T)
    sumY2  +=       0.001 * cos(4.36202291393  + 103285.36090659029 *T)
    sumY2  +=       0.001 * cos(5.63116195221  + 55504.18575691209 *T)
    sumY2  +=       0.001 * cos(4.47656675233  + 117873.60782537168 *T)
    sumY2  +=       0.001 * cos(0.96879148402  + 25035.87857586930 *T)
    sumY2  +=       0.001 * cos(5.66596818667  + 58946.27306691049 *T)
    sumY2  +=       0.001 * cos(4.82817050708  + 188276.89785765507 *T)
    sumY2  +=       0.001 * cos(0.16804966441  + 49842.85370775989 *T)
    sumY2  +=       0.001 * cos(3.20454167097  + 29428.75938575750 *T)
    sumY2  +=       0.001 * cos(3.65119748789  + 146314.37712071769 *T)
    sumY2  +=       0.001 * cos(0.55150389784  + 119117.09542705010 *T)
    sumY2  +=       0.001 * cos(1.29641320948  + 39629.56816154889 *T)
    sumY2  +=       0.001 * cos(1.17683878413  + 12546.72575656651 *T)
    sumY2  +=       0.001 * cos(1.82966773526  + 64607.60511606269 *T)
    sumY2  +=       0.001 * cos(0.51019327318  + 23868.90221990390 *T)
    sumY2  +=       0.001 * cos(3.01779717813  + 48847.91444435170 *T)
    sumY2  +=       0.001 * cos(3.08178730790  + 260878.78759825832 *T)
    sumY2  +=       0.001 * cos(1.94628548757  + 25620.18196870330 *T)
    sumY2  +=       0.001 * cos(4.90071449530  + 98068.78053378889 *T)
    sumY2  +=       0.001 * cos(5.19983179973  + 22747.53453235790 *T)
    sumY2  +=       0.001 * cos(3.72290134761  + 26556.11194941211 *T)
    sumY2  +=       0.001 * cos(4.32894922591  + 85034.66384345168 *T)
    sumY2  +=       0.001 * cos(3.83325919627  + 5327.71992586630 *T)
    sumY2  +=       0.001 * cos(5.56885466592  + 63498.22656396929 *T)
    sumY2  +=       0.001 * cos(2.05127060689  + 26300.95841952869 *T)
    sumY2  +=       0.001 * cos(5.07273427642  + 105940.92927906408 *T)
    sumY2  +=       0.001 * cos(0.91856795860  + 65831.91059180829 *T)
    sumY2  +=       0.001 * cos(4.03570631976  + 79372.84415933248 *T)
    sumY2  +=       0.001 * cos(1.62497049537  + 7.35736448430 *T)
    sumY2  +=       0.001 * cos(5.58196282531  + 53399.86794141051 *T)
    sumY2  +=       0.001 * cos(2.25556473621  + 80597.14963507809 *T)
    sumY2  +=       0.001 * cos(1.53347024755  + 49527.59527515889 *T)
    sumY2  +=       0.001 * cos(2.15927670335  + 36300.94487030169 *T)
    sumY2  +=       0.001 * cos(0.13876413209  + 116918.00808376308 *T)
    sumY2  +=       0.001 * cos(0.13834609478  + 26080.54577708989 *T)
    sumY2  +=       0.001 * cos(1.87343313975  + 60056.13925397089 *T)
    sumY2  +=       0.001 * cos(3.24836507326  + 10022.08109758290 *T)
    sumY2  +=       0.001 * cos(3.54409288852  + 51123.78171744350 *T)
    sumY2  +=       0.001 * cos(2.04668299759  + 51645.87150057029 *T)
    sumY2  +=       0.001 * cos(6.13056634943  + 76.50988875911 *T)
    sumY2  +=       0.001 * cos(5.70833789847  + 106570.61348796709 *T)
    sumY2  +=       0.001 * cos(1.85475666159  + 25021.16384690069 *T)
    sumY2  +=       0.001 * cos(3.08169426022  + 15874.37377787969 *T)
    sumY2  +=       0.001 * cos(5.01616234440  + 86144.04239554508 *T)
    sumY2  +=       0.001 * cos(1.82601661336  + 52644.01509098630 *T)
    sumY2  +=       0.001 * cos(4.40416127197  + 26094.77551795771 *T)
    sumY2  +=       0.001 * cos(1.26427504022  + 26081.51840015769 *T)
    sumY2  +=       0.001 * cos(1.17801372459  + 6.86972951729 *T)
    sumY2  +=       0.001 * cos(5.00258165716  + 16703.30595098250 *T)
    sumY2  +=       0.001 * cos(3.75732229340  + 35472.98796713290 *T)
    sumY2  +=       0.001 * cos(5.03560870343  + 42154.21282053250 *T)
    sumY2  +=       0.001 * cos(2.82069930264  + 77410.75686045409 *T)
    sumY2  +=       0.001 * cos(1.94570699226  + 103925.25819290428 *T)
    sumY2  +=       0.001 * cos(0.43527786277  + 52329.82891111670 *T)
    sumY2  +=       0.001 * cos(1.53490250185  + 6681.46867088311 *T)
    sumY2  +=       0.001 * cos(5.27958367842  + 6770.95441872910 *T)
    sumY2  +=       0.001 * cos(3.09122313934  + 955.84355909210 *T)
    sumY2  +=       0.001 * cos(5.47023024103  + 51109.55462344189 *T)
    sumY2  +=       0.001 * cos(2.82652928734  + 29530.23426905610 *T)
    sumY2  +=       0.001 * cos(3.34999395144  + 26067.98956319090 *T)
    sumY2  +=       0.001 * cos(5.60842191407  + 25447.76203777669 *T)
    sumY2  +=       0.001 * cos(1.24359184874  + 52705.25343075949 *T)
    sumY2  +=       0.001 * cos(2.58826883392  + 33967.74847742969 *T)
    sumY2  +=       0.001 * cos(1.58743932556  + 25035.39094090229 *T)
    sumY2  +=       0.001 * cos(1.49321197683  + 26107.32908499049 *T)
    sumY2  +=       0.001 * cos(5.64817713358  + 23969.38309867930 *T)
    sumY2  +=       0.001 * cos(4.09696575895  + 19316.94872284510 *T)
    sumY2  +=       0.001 * cos(0.38657846461  + 26091.54065944870 *T)
    sumY2  +=       0.001 * cos(3.96518956917  + 53131.16220727349 *T)
    sumY2  +=       0.001 * cos(1.17035557979  + 95247.94938966268 *T)
    sumY2  +=       0.001 * cos(3.79316376096  + 103299.58800059187 *T)
    sumY2  +=       0.001 * cos(6.16466999792  + 26094.77287109149 *T)
    sumY2  +=       0.001 * cos(1.55660546282  + 102233.09252340188 *T)
    sumY2  +=       0.001 * cos(3.72893918712  + 25668.17468021549 *T)
    sumY2  +=       0.001 * cos(1.72210842613  + 12432.28646788130 *T)
    sumY2  +=       0.001 * cos(2.54599021645  + 115674.52048208468 *T)
    sumY2  +=       0.001 * cos(2.39000910681  + 25661.06113321469 *T)
    sumY2  +=       0.001 * cos(3.51239889125  + 39609.41076568209 *T)
    sumY2  +=       0.001 * cos(4.42730882654  + 26083.77798873270 *T)

    # MERCURY Y Coord
    sumY3  = 0
    sumY3  +=     180.660 * cos(6.24763771127  + 0.24381748350 *T)
    sumY3  +=      56.109 * cos(6.21607723411  + 52176.05010063190 *T)
    sumY3  +=      23.909 * cos(3.52205756332  + 78263.95324220609 *T)
    sumY3  +=      11.317 * cos(0.65434522049  + 104351.85638378028 *T)
    sumY3  +=      11.202 * cos(4.63472742214  + 26088.14695905770 *T)
    sumY3  +=       7.321 * cos(3.06113980531  + 26087.65932409069 *T)
    sumY3  +=       5.011 * cos(3.87936295964  + 130439.75952535449 *T)
    sumY3  +=       2.025 * cos(0.73521591485  + 156527.66266692869 *T)
    sumY3  +=       1.152 * cos(1.69823188077  + 52175.56246566489 *T)
    sumY3  +=       0.763 * cos(3.83589288112  + 182615.56580850287 *T)
    sumY3  +=       0.387 * cos(4.87852090068  + 78263.46560723908 *T)
    sumY3  +=       0.274 * cos(0.63429416338  + 208703.46895007708 *T)
    sumY3  +=       0.112 * cos(1.64880744224  + 104351.36874881329 *T)
    sumY3  +=       0.095 * cos(3.70512425843  + 234791.37209165128 *T)
    sumY3  +=       0.037 * cos(2.93820472611  + 24978.76840696430 *T)
    sumY3  +=       0.034 * cos(2.70920545196  + 27197.52551115110 *T)
    sumY3  +=       0.032 * cos(0.48595621544  + 260879.27523322543 *T)
    sumY3  +=       0.024 * cos(0.88877136179  + 51066.67154853850 *T)
    sumY3  +=       0.032 * cos(4.71795792013  + 130439.27189038748 *T)
    sumY3  +=       0.013 * cos(4.14981392057  + 77154.57469011268 *T)
    sumY3  +=       0.010 * cos(3.54374573861  + 286967.17837479961 *T)
    sumY3  +=       0.013 * cos(4.83147497313  + 53285.42865272530 *T)
    sumY3  +=       0.010 * cos(1.53331553777  + 156527.17503196167 *T)
    sumY3  +=       0.006 * cos(0.11733989890  + 1109.13473460990 *T)
    sumY3  +=       0.007 * cos(1.42096187453  + 1059.62574767270 *T)
    sumY3  +=       0.006 * cos(0.36688147081  + 20426.81490990550 *T)
    sumY3  +=       0.006 * cos(0.99179814069  + 103242.47783168688 *T)
    sumY3  +=       0.004 * cos(5.28779415998  + 31749.47900820990 *T)
    sumY3  +=       0.004 * cos(1.32226409901  + 79373.33179429949 *T)
    sumY3  +=       0.003 * cos(0.32143664919  + 313055.08151637384 *T)
    sumY3  +=       0.003 * cos(5.50927594249  + 4552.19731454230 *T)
    sumY3  +=       0.003 * cos(4.66013795345  + 182615.07817353585 *T)
    sumY3  +=       0.003 * cos(4.10710790486  + 46514.71805147970 *T)
    sumY3  +=       0.003 * cos(0.25274361593  + 21536.19346199890 *T)
    sumY3  +=       0.002 * cos(4.79995023814  + 27147.52888924690 *T)
    sumY3  +=       0.002 * cos(2.39000042999  + 1589.31671276730 *T)
    sumY3  +=       0.002 * cos(5.16055835758  + 25132.54721744910 *T)
    sumY3  +=       0.002 * cos(4.23604489079  + 14765.48286075331 *T)
    sumY3  +=       0.002 * cos(4.24590446693  + 51116.66817044270 *T)
    sumY3  +=       0.002 * cos(4.07863052508  + 129330.38097326108 *T)
    sumY3  +=       0.002 * cos(6.04806962418  + 27197.03787618409 *T)
    sumY3  +=       0.002 * cos(1.19379164655  + 26617.83792415230 *T)
    sumY3  +=       0.002 * cos(0.44240554373  + 27043.74670066630 *T)
    sumY3  +=       0.002 * cos(4.67526593433  + 25558.45599396310 *T)
    sumY3  +=       0.002 * cos(1.31409357670  + 72602.62119305388 *T)
    sumY3  +=       0.002 * cos(0.23323058095  + 26068.47719815791 *T)
    sumY3  +=       0.002 * cos(5.43511468464  + 26107.81671995749 *T)
    sumY3  +=       0.001 * cos(3.58800786511  + 26092.02829441570 *T)
    sumY3  +=       0.001 * cos(2.08042717023  + 26084.26562369970 *T)
    sumY3  +=       0.001 * cos(5.04284012325  + 24978.28077199729 *T)
    sumY3  +=       0.001 * cos(1.47519959792  + 37410.81105736210 *T)
    sumY3  +=       0.001 * cos(5.02608519942  + 30640.10045611650 *T)
    sumY3  +=       0.001 * cos(5.10218274945  + 52705.74106572650 *T)
    sumY3  +=       0.001 * cos(3.44712662114  + 51220.45035902330 *T)
    sumY3  +=       0.001 * cos(0.20321095099  + 47624.09660357310 *T)
    sumY3  +=       0.001 * cos(1.50125338315  + 208702.98131511008 *T)
    sumY3  +=       0.001 * cos(1.69903209468  + 9104.15081160110 *T)
    sumY3  +=       0.001 * cos(4.53006298092  + 98690.52433462808 *T)
    sumY3  +=       0.001 * cos(4.29961422700  + 955.35592412510 *T)
    sumY3  +=       0.001 * cos(1.16934270913  + 1109.62236957690 *T)
    sumY3  +=       0.001 * cos(0.43847536572  + 19317.43635781211 *T)
    sumY3  +=       0.001 * cos(3.28791774568  + 50586.97720534809 *T)
    sumY3  +=       0.001 * cos(3.80818275289  + 12566.39551746630 *T)
    sumY3  +=       0.001 * cos(1.94170233989  + 40853.38600232750 *T)
    sumY3  +=       0.001 * cos(2.16988291625  + 53235.43203082109 *T)
    sumY3  +=       0.001 * cos(3.36850892547  + 339142.98465794802 *T)
    sumY3  +=       0.001 * cos(2.07799856696  + 426.84200835950 *T)
    sumY3  +=       0.001 * cos(1.13583428934  + 49957.29299644510 *T)
    sumY3  +=       0.001 * cos(2.48860782643  + 26095.26050605850 *T)
    sumY3  +=       0.001 * cos(3.22444138778  + 26081.03341205689 *T)
    sumY3  +=       0.001 * cos(2.99647624764  + 23869.38985487090 *T)
    sumY3  +=       0.001 * cos(0.34236528707  + 77308.35350059748 *T)
    sumY3  +=       0.001 * cos(4.24503082495  + 105461.23493587368 *T)
    sumY3  +=       0.001 * cos(0.10746803930  + 51646.35913553730 *T)
    sumY3  +=       0.001 * cos(0.85601964696  + 155418.28411483529 *T)
    sumY3  +=       0.001 * cos(3.18986978065  + 56728.00359769070 *T)
    sumY3  +=       0.001 * cos(1.94353556293  + 53131.64984224050 *T)
    sumY3  +=       0.001 * cos(5.49395066730  + 3442.81876244891 *T)
    sumY3  +=       0.001 * cos(2.08016019343  + 78793.64420730069 *T)
    sumY3  +=       0.001 * cos(4.54434565803  + 54394.80720481869 *T)
    sumY3  +=       0.001 * cos(5.42235344143  + 79323.33517239528 *T)
    sumY3  +=       0.001 * cos(2.66875497262  + 28306.90406324449 *T)
    sumY3  +=       0.001 * cos(4.76053635454  + 45405.33949938630 *T)
    sumY3  +=       0.001 * cos(5.22561098934  + 32858.85756030330 *T)
    sumY3  +=       0.001 * cos(0.08839061971  + 25028.76502886850 *T)
    sumY3  +=       0.001 * cos(0.19202613793  + 22645.57201409229 *T)
    sumY3  +=       0.001 * cos(0.02689519967  + 27677.21985434150 *T)

    # MERCURY Y Coord
    sumY4  = 0
    sumY4  +=      43.303 * cos(1.13774685024  + 26088.14695905770 *T)
    sumY4  +=      16.746 * cos(1.28029969371  + 0.24381748350 *T)
    sumY4  +=       5.097 * cos(4.24955975906  + 52176.05010063190 *T)
    sumY4  +=       1.110 * cos(1.28336867360  + 78263.95324220609 *T)
    sumY4  +=       0.968 * cos(4.75917629222  + 26087.65932409069 *T)
    sumY4  +=       0.362 * cos(4.81899532084  + 104351.85638378028 *T)
    sumY4  +=       0.161 * cos(1.99262264168  + 130439.75952535449 *T)
    sumY4  +=       0.075 * cos(5.26842672327  + 156527.66266692869 *T)
    sumY4  +=       0.076 * cos(1.36013166836  + 52175.56246566489 *T)
    sumY4  +=       0.033 * cos(2.15747243303  + 182615.56580850287 *T)
    sumY4  +=       0.014 * cos(3.97795789451  + 78263.46560723908 *T)
    sumY4  +=       0.014 * cos(5.27638874760  + 208703.46895007708 *T)
    sumY4  +=       0.005 * cos(2.08976600519  + 234791.37209165128 *T)
    sumY4  +=       0.004 * cos(0.41935591421  + 104351.36874881329 *T)
    sumY4  +=       0.002 * cos(1.30411243109  + 24978.76840696430 *T)
    sumY4  +=       0.002 * cos(4.27909717270  + 27197.52551115110 *T)
    sumY4  +=       0.002 * cos(5.16384952851  + 260879.27523322543 *T)
    sumY4  +=       0.001 * cos(3.31456236049  + 130439.27189038748 *T)
    sumY4  +=       0.001 * cos(1.94555546011  + 286967.17837479961 *T)
    sumY4  +=       0.001 * cos(4.78035947332  + 51066.67154853850 *T)
    sumY4  +=       0.001 * cos(6.04727171347  + 1059.62574767270 *T)
    sumY4  +=       0.001 * cos(4.90982478671  + 20426.81490990550 *T)

    # MERCURY Y Coord
    sumY5  = 0
    sumY5  +=       0.414 * cos(2.51938027426  + 0.24381748350 *T)
    sumY5  +=       0.327 * cos(1.26814697300  + 26088.14695905770 *T)
    sumY5  +=       0.134 * cos(2.94456567084  + 52176.05010063190 *T)
    sumY5  +=       0.046 * cos(5.94506878756  + 78263.95324220609 *T)
    sumY5  +=       0.016 * cos(2.88715140580  + 104351.85638378028 *T)
    sumY5  +=       0.007 * cos(0.85960797708  + 26087.65932409069 *T)
    sumY5  +=       0.006 * cos(6.26965738190  + 130439.75952535449 *T)
    sumY5  +=       0.003 * cos(3.37112838196  + 156527.66266692869 *T)
    sumY5  +=       0.001 * cos(0.36849890320  + 182615.56580850287 *T)
    sumY5  +=       0.001 * cos(5.73351518627  + 52175.56246566489 *T)
    sumY5  +=       0.001 * cos(3.53868886531  + 208703.46895007708 *T)

    # MERCURY Z Coord
    sumZ0  = 0
    sumZ0  +=  4607665.326 * cos(1.99295081967  + 26087.90314157420 *T)
    sumZ0  +=  708734.365 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ0  +=  469171.617 * cos(5.04215742764  + 52175.80628314840 *T)
    sumZ0  +=   71626.395 * cos(1.80894256071  + 78263.70942472259 *T)
    sumZ0  +=   12957.446 * cos(4.85922032010  + 104351.61256629678 *T)
    sumZ0  +=    2575.002 * cos(1.62646731545  + 130439.51570787100 *T)
    sumZ0  +=     543.259 * cos(4.67698860167  + 156527.41884944518 *T)
    sumZ0  +=     119.462 * cos(1.44437994097  + 182615.32199101936 *T)
    sumZ0  +=      79.477 * cos(4.94442849343  + 27197.28169366760 *T)
    sumZ0  +=      82.635 * cos(2.03127961938  + 24978.52458948080 *T)
    sumZ0  +=      44.712 * cos(0.52659463739  + 1059.38193018920 *T)
    sumZ0  +=      43.131 * cos(5.66922412963  + 20426.57109242200 *T)
    sumZ0  +=      34.677 * cos(2.36736170874  + 25028.52121138500 *T)
    sumZ0  +=      32.337 * cos(1.29099631099  + 31749.23519072640 *T)
    sumZ0  +=      27.465 * cos(3.57680072404  + 5661.33204915220 *T)
    sumZ0  +=      27.081 * cos(4.49499356963  + 208703.22513259359 *T)
    sumZ0  +=      20.978 * cos(4.46131555233  + 4551.95349705880 *T)
    sumZ0  +=      20.602 * cos(1.12768557199  + 1109.37855209340 *T)
    sumZ0  +=      20.252 * cos(0.11722157720  + 51116.42435295920 *T)
    sumZ0  +=      23.561 * cos(1.82361021215  + 53285.18483524180 *T)
    sumZ0  +=      14.150 * cos(4.42485698284  + 57837.13833230060 *T)
    sumZ0  +=      16.632 * cos(5.09118864825  + 51066.42773105500 *T)
    sumZ0  +=      11.609 * cos(3.48812792468  + 21535.94964451540 *T)
    sumZ0  +=       9.897 * cos(2.80940292918  + 27043.50288318280 *T)
    sumZ0  +=       8.380 * cos(2.46820093103  + 46514.47423399620 *T)
    sumZ0  +=       7.950 * cos(0.44767912341  + 10213.28554621100 *T)
    sumZ0  +=       9.234 * cos(3.80445817621  + 529.69096509460 *T)
    sumZ0  +=       7.834 * cos(4.08006285427  + 25132.30339996560 *T)
    sumZ0  +=       7.169 * cos(2.43861625280  + 47623.85278608960 *T)
    sumZ0  +=       6.283 * cos(1.26244592145  + 234791.12827416777 *T)
    sumZ0  +=       6.189 * cos(3.24598311063  + 77204.32749453338 *T)
    sumZ0  +=       6.312 * cos(3.60156222051  + 27147.28507176339 *T)
    sumZ0  +=       5.874 * cos(3.18517530495  + 14765.23904326980 *T)
    sumZ0  +=       5.464 * cos(0.18067146403  + 13521.75144159140 *T)
    sumZ0  +=       5.844 * cos(4.88684915394  + 79373.08797681599 *T)
    sumZ0  +=       5.207 * cos(1.04360609206  + 426.59819087600 *T)
    sumZ0  +=       5.084 * cos(2.88826076001  + 12566.15169998280 *T)
    sumZ0  +=       4.862 * cos(1.05211425705  + 1589.07289528380 *T)
    sumZ0  +=       5.778 * cos(4.74518603718  + 30639.85663863300 *T)
    sumZ0  +=       4.470 * cos(5.94243776120  + 11322.66409830440 *T)
    sumZ0  +=       4.084 * cos(6.10730721947  + 41962.52073693740 *T)
    sumZ0  +=       3.921 * cos(1.19592374632  + 83925.04147387479 *T)
    sumZ0  +=       3.933 * cos(1.60795544806  + 25661.30495069820 *T)
    sumZ0  +=       4.046 * cos(3.08469176646  + 51646.11531805379 *T)
    sumZ0  +=       3.739 * cos(2.04428825509  + 15874.61759536320 *T)
    sumZ0  +=       3.728 * cos(1.74332154184  + 24498.83024629040 *T)
    sumZ0  +=       3.797 * cos(1.86233694539  + 77154.33087262919 *T)
    sumZ0  +=       3.417 * cos(4.10830949198  + 39609.65458316560 *T)
    sumZ0  +=       3.691 * cos(3.50893180252  + 37410.56723987860 *T)
    sumZ0  +=       2.644 * cos(5.60089521868  + 73711.75592766379 *T)
    sumZ0  +=       3.230 * cos(5.39015797984  + 955.59974160860 *T)
    sumZ0  +=       2.653 * cos(5.52156950287  + 26617.59410666880 *T)
    sumZ0  +=       2.526 * cos(5.49112478707  + 6283.07584999140 *T)
    sumZ0  +=       3.017 * cos(6.08863660917  + 53131.40602475700 *T)
    sumZ0  +=       3.020 * cos(5.46768108443  + 25558.21217647960 *T)
    sumZ0  +=       2.169 * cos(5.85843688724  + 50586.73338786459 *T)
    sumZ0  +=       2.217 * cos(5.81930894103  + 51749.20809227239 *T)
    sumZ0  +=       2.036 * cos(2.79411907992  + 7238.67559160000 *T)
    sumZ0  +=       1.904 * cos(0.02429449494  + 33326.57873317420 *T)
    sumZ0  +=       1.841 * cos(3.31987542265  + 36301.18868778519 *T)
    sumZ0  +=       1.693 * cos(0.04022611885  + 103292.23063610760 *T)
    sumZ0  +=       1.814 * cos(0.58414302882  + 63498.47038145279 *T)
    sumZ0  +=       1.910 * cos(5.54302892735  + 72602.37737557039 *T)
    sumZ0  +=       1.560 * cos(0.96215455240  + 65697.55772473979 *T)
    sumZ0  +=       1.605 * cos(0.84854708960  + 51220.20654153979 *T)
    sumZ0  +=       1.595 * cos(5.65216619091  + 26068.23338067440 *T)
    sumZ0  +=       1.487 * cos(1.66950376778  + 105460.99111839020 *T)
    sumZ0  +=       1.595 * cos(1.47238921630  + 26107.57290247399 *T)
    sumZ0  +=       1.485 * cos(4.31307667056  + 260879.03141574195 *T)
    sumZ0  +=       1.344 * cos(3.05188579262  + 68050.42387851159 *T)
    sumZ0  +=       1.601 * cos(0.69333064931  + 9103.90699411760 *T)
    sumZ0  +=       1.224 * cos(3.56543079781  + 19804.82729158280 *T)
    sumZ0  +=       1.258 * cos(2.62817126488  + 52705.49724824299 *T)
    sumZ0  +=       1.386 * cos(0.42909271065  + 53235.18821333759 *T)
    sumZ0  +=       1.392 * cos(0.41896396083  + 18849.22754997420 *T)
    sumZ0  +=       1.104 * cos(1.37365365216  + 45892.73043315699 *T)
    sumZ0  +=       1.191 * cos(5.76715631957  + 19317.19254032860 *T)
    sumZ0  +=       1.218 * cos(6.16642870453  + 77734.01845962799 *T)
    sumZ0  +=       0.991 * cos(0.77508695125  + 32858.61374281979 *T)
    sumZ0  +=       1.039 * cos(4.25003847002  + 110012.94461544900 *T)
    sumZ0  +=       1.155 * cos(1.83291524170  + 16983.99614745660 *T)
    sumZ0  +=       0.913 * cos(4.91386835375  + 103242.23401420339 *T)
    sumZ0  +=       1.096 * cos(2.77782845562  + 26091.78447693220 *T)
    sumZ0  +=       0.843 * cos(2.10283893013  + 23869.14603738740 *T)
    sumZ0  +=       1.147 * cos(1.40069226758  + 56727.75978020720 *T)
    sumZ0  +=       1.096 * cos(4.34969785012  + 26084.02180621620 *T)
    sumZ0  +=       0.888 * cos(5.96405172933  + 38654.05484155699 *T)
    sumZ0  +=       0.798 * cos(5.90979724117  + 43071.89928903080 *T)
    sumZ0  +=       0.894 * cos(6.20891949330  + 40853.14218484400 *T)
    sumZ0  +=       0.751 * cos(2.38280961577  + 99799.65906923798 *T)
    sumZ0  +=       0.858 * cos(3.48194203031  + 59414.48187474840 *T)
    sumZ0  +=       0.903 * cos(4.51539427543  + 3442.57494496540 *T)
    sumZ0  +=       0.816 * cos(4.68549846425  + 28306.66024576099 *T)
    sumZ0  +=       0.739 * cos(4.03898100455  + 26514.50133245020 *T)
    sumZ0  +=       0.874 * cos(3.33462533468  + 22645.32819660879 *T)
    sumZ0  +=       0.685 * cos(2.68994541603  + 76674.63652943878 *T)
    sumZ0  +=       0.933 * cos(3.52008984237  + 6770.71060124560 *T)
    sumZ0  +=       0.668 * cos(2.67447570603  + 77837.11123384659 *T)
    sumZ0  +=       0.688 * cos(4.13735845510  + 27676.97603685800 *T)
    sumZ0  +=       0.643 * cos(0.26212576857  + 639.89728631400 *T)
    sumZ0  +=       0.747 * cos(2.88413728742  + 79219.30916633119 *T)
    sumZ0  +=       0.594 * cos(4.72204801083  + 29530.47808653960 *T)
    sumZ0  +=       0.580 * cos(0.49131159361  + 62389.09182935939 *T)
    sumZ0  +=       0.560 * cos(0.58980434738  + 26080.78959457339 *T)
    sumZ0  +=       0.636 * cos(5.01277329265  + 7.11354700080 *T)
    sumZ0  +=       0.664 * cos(5.40615713600  + 213.29909543800 *T)
    sumZ0  +=       0.460 * cos(2.41008353609  + 25448.00585526019 *T)
    sumZ0  +=       0.453 * cos(4.02646064195  + 91785.46086631398 *T)
    sumZ0  +=       0.441 * cos(3.60120922804  + 89586.37352302698 *T)
    sumZ0  +=       0.473 * cos(3.19147656788  + 25874.60404613620 *T)
    sumZ0  +=       0.452 * cos(3.10205575226  + 129380.13377768178 *T)
    sumZ0  +=       0.406 * cos(1.57776301723  + 2118.76386037840 *T)
    sumZ0  +=       0.495 * cos(4.35094686160  + 58946.51688439399 *T)
    sumZ0  +=       0.395 * cos(0.29737033271  + 25934.12433108940 *T)
    sumZ0  +=       0.392 * cos(0.53546585935  + 26241.68195205900 *T)
    sumZ0  +=       0.382 * cos(4.72731471129  + 131548.89425996438 *T)
    sumZ0  +=       0.382 * cos(4.49198737842  + 71980.63357473118 *T)
    sumZ0  +=       0.368 * cos(3.89892172772  + 77308.10968311399 *T)
    sumZ0  +=       0.351 * cos(1.04745682138  + 2218.75710418680 *T)
    sumZ0  +=       0.356 * cos(6.12309999532  + 94138.32702008578 *T)
    sumZ0  +=       0.458 * cos(2.31746139528  + 98690.28051714458 *T)
    sumZ0  +=       0.327 * cos(4.52406643775  + 52195.47604404819 *T)
    sumZ0  +=       0.393 * cos(3.09725258397  + 69159.80243060499 *T)
    sumZ0  +=       0.310 * cos(1.10653094623  + 23969.13928119580 *T)
    sumZ0  +=       0.356 * cos(1.08062840940  + 286966.93455731618 *T)
    sumZ0  +=       0.345 * cos(5.71957868514  + 78793.40038981718 *T)
    sumZ0  +=       0.293 * cos(1.74897333943  + 54394.56338733519 *T)
    sumZ0  +=       0.275 * cos(4.05372155800  + 149.56319713460 *T)
    sumZ0  +=       0.325 * cos(2.41844865673  + 52156.13652224860 *T)
    sumZ0  +=       0.273 * cos(0.34374545554  + 51535.90899683439 *T)
    sumZ0  +=       0.256 * cos(3.57544759221  + 1066.49547719000 *T)
    sumZ0  +=       0.252 * cos(2.07789295059  + 48733.23133818299 *T)
    sumZ0  +=       0.326 * cos(3.49941746899  + 79323.09135491178 *T)
    sumZ0  +=       0.327 * cos(2.94651885500  + 103821.92160120218 *T)
    sumZ0  +=       0.300 * cos(3.53612598449  + 26301.20223701220 *T)
    sumZ0  +=       0.253 * cos(0.67254115542  + 1052.26838318840 *T)
    sumZ0  +=       0.274 * cos(1.02128544399  + 136100.84775702321 *T)
    sumZ0  +=       0.227 * cos(1.68291413296  + 129330.13715577760 *T)
    sumZ0  +=       0.264 * cos(4.42527687575  + 82815.66292178139 *T)
    sumZ0  +=       0.280 * cos(3.47711681992  + 44937.13069154840 *T)
    sumZ0  +=       0.207 * cos(5.26698515692  + 25938.33994443960 *T)
    sumZ0  +=       0.233 * cos(1.58956295605  + 26095.01668857500 *T)
    sumZ0  +=       0.211 * cos(2.55179422569  + 45405.09568190280 *T)
    sumZ0  +=       0.223 * cos(1.11566460119  + 52171.92494779040 *T)
    sumZ0  +=       0.195 * cos(5.61180266071  + 25021.40766438420 *T)
    sumZ0  +=       0.204 * cos(5.44212122344  + 125887.56221081220 *T)
    sumZ0  +=       0.193 * cos(2.20507758963  + 25035.63475838580 *T)
    sumZ0  +=       0.189 * cos(5.78823648500  + 102762.53967101300 *T)
    sumZ0  +=       0.213 * cos(1.31985764486  + 26555.86813192860 *T)
    sumZ0  +=       0.191 * cos(0.24359410467  + 85502.38501632259 *T)
    sumZ0  +=       0.195 * cos(5.97596513107  + 105307.21230790540 *T)
    sumZ0  +=       0.218 * cos(5.85009667782  + 25619.93815121980 *T)
    sumZ0  +=       0.181 * cos(5.75215140812  + 103925.01437542078 *T)
    sumZ0  +=       0.190 * cos(3.25444605074  + 13655.86049117640 *T)
    sumZ0  +=       0.177 * cos(5.31586012519  + 50057.04242277000 *T)
    sumZ0  +=       0.207 * cos(2.75433176036  + 64741.95798313119 *T)
    sumZ0  +=       0.224 * cos(5.82663546732  + 52179.68761850640 *T)
    sumZ0  +=       0.164 * cos(3.65486891804  + 76.26607127560 *T)
    sumZ0  +=       0.166 * cos(5.16708694657  + 49957.04917896160 *T)
    sumZ0  +=       0.171 * cos(1.99193197138  + 26094.53170047421 *T)
    sumZ0  +=       0.159 * cos(2.10978774324  + 32370.97899156560 *T)
    sumZ0  +=       0.164 * cos(5.82981713266  + 12432.04265039780 *T)
    sumZ0  +=       0.155 * cos(5.39879388941  + 6681.22485339960 *T)
    sumZ0  +=       0.199 * cos(3.05926705479  + 66941.04532641819 *T)
    sumZ0  +=       0.171 * cos(5.13520423660  + 26081.27458267419 *T)
    sumZ0  +=       0.150 * cos(0.90885651206  + 52389.10537858640 *T)
    sumZ0  +=       0.184 * cos(1.74487163243  + 51962.50718771040 *T)
    sumZ0  +=       0.164 * cos(0.91431121775  + 52602.40447402440 *T)
    sumZ0  +=       0.138 * cos(5.40226519416  + 26011.63707029860 *T)
    sumZ0  +=       0.149 * cos(2.17444239275  + 52168.69273614759 *T)
    sumZ0  +=       0.184 * cos(1.61828580097  + 20760.42703319140 *T)
    sumZ0  +=       0.152 * cos(0.96120452018  + 53764.87917843220 *T)
    sumZ0  +=       0.146 * cos(5.57207251195  + 58458.88213313979 *T)
    sumZ0  +=       0.150 * cos(3.58807734183  + 88476.99497093359 *T)
    sumZ0  +=       0.124 * cos(1.42633058836  + 55618.38122811380 *T)
    sumZ0  +=       0.141 * cos(1.44033578815  + 5327.47610838280 *T)
    sumZ0  +=       0.126 * cos(0.80278847447  + 117873.36400788820 *T)
    sumZ0  +=       0.118 * cos(3.91167705851  + 19406.67828817460 *T)
    sumZ0  +=       0.117 * cos(4.68118162480  + 52182.91983014920 *T)
    sumZ0  +=       0.130 * cos(3.12963980623  + 38519.94579197200 *T)
    sumZ0  +=       0.123 * cos(2.90959272560  + 52026.24308601380 *T)
    sumZ0  +=       0.114 * cos(3.34729061863  + 51109.31080595839 *T)
    sumZ0  +=       0.109 * cos(0.37062017000  + 115674.27666460120 *T)
    sumZ0  +=       0.108 * cos(5.35523619654  + 46848.33017476560 *T)
    sumZ0  +=       0.120 * cos(6.15890498480  + 155468.03691925600 *T)
    sumZ0  +=       0.105 * cos(1.26949313866  + 98068.53671630540 *T)
    sumZ0  +=       0.124 * cos(5.59073645014  + 74821.13447975718 *T)
    sumZ0  +=       0.113 * cos(5.37283602123  + 124778.18365871880 *T)
    sumZ0  +=       0.101 * cos(1.50859657746  + 157636.79740153858 *T)
    sumZ0  +=       0.112 * cos(6.24044765899  + 51123.53789995999 *T)
    sumZ0  +=       0.089 * cos(0.66689996363  + 103396.01282468818 *T)
    sumZ0  +=       0.088 * cos(3.24234443321  + 26727.80042788820 *T)
    sumZ0  +=       0.083 * cos(3.46887356727  + 77623.81213840858 *T)
    sumZ0  +=       0.082 * cos(3.58489505049  + 52329.58509363319 *T)
    sumZ0  +=       0.082 * cos(2.76598729250  + 35191.81013569180 *T)
    sumZ0  +=       0.092 * cos(2.89919284588  + 120226.23016166000 *T)
    sumZ0  +=       0.081 * cos(3.34888362248  + 52022.02747266360 *T)
    sumZ0  +=       0.080 * cos(1.61800826930  + 45494.58142974879 *T)
    sumZ0  +=       0.087 * cos(5.99930110618  + 129909.82474277640 *T)
    sumZ0  +=       0.075 * cos(1.29165560989  + 78283.37918562238 *T)
    sumZ0  +=       0.097 * cos(0.94285631414  + 85034.42002596818 *T)
    sumZ0  +=       0.075 * cos(0.93132779667  + 433.71173787680 *T)
    sumZ0  +=       0.087 * cos(3.29295494560  + 3340.61242669980 *T)
    sumZ0  +=       0.083 * cos(4.56076128644  + 419.48464387520 *T)
    sumZ0  +=       0.080 * cos(0.27627260080  + 105410.99449648600 *T)
    sumZ0  +=       0.079 * cos(5.99645052852  + 95247.70557217918 *T)
    sumZ0  +=       0.074 * cos(0.63932580906  + 64607.84893354619 *T)
    sumZ0  +=       0.066 * cos(4.71152255697  + 80482.46652890938 *T)
    sumZ0  +=       0.073 * cos(3.21815072307  + 52099.54021187280 *T)
    sumZ0  +=       0.066 * cos(2.98683203711  + 23754.70674870219 *T)
    sumZ0  +=       0.091 * cos(2.50052149852  + 104881.30353139140 *T)
    sumZ0  +=       0.062 * cos(5.73156315784  + 853.19638175200 *T)
    sumZ0  +=       0.086 * cos(4.13151425024  + 313054.83769889042 *T)
    sumZ0  +=       0.064 * cos(1.17129038283  + 108903.56606335558 *T)
    sumZ0  +=       0.059 * cos(0.79111914848  + 7994.52844202420 *T)
    sumZ0  +=       0.065 * cos(0.25068571900  + 71025.03383312259 *T)
    sumZ0  +=       0.057 * cos(6.23642054033  + 1692.16566950240 *T)
    sumZ0  +=       0.057 * cos(4.67121384993  + 28206.66700195260 *T)
    sumZ0  +=       0.059 * cos(4.92822158961  + 78050.41032928458 *T)
    sumZ0  +=       0.056 * cos(2.09215892689  + 76144.94556434419 *T)
    sumZ0  +=       0.056 * cos(1.73304008242  + 25654.19140369739 *T)
    sumZ0  +=       0.057 * cos(4.73385475413  + 155418.04029735178 *T)
    sumZ0  +=       0.056 * cos(5.88720785632  + 26037.90651966999 *T)
    sumZ0  +=       0.074 * cos(5.46823409002  + 78244.03966382280 *T)
    sumZ0  +=       0.061 * cos(4.50066150046  + 25668.41849769900 *T)
    sumZ0  +=       0.071 * cos(4.06855895670  + 162188.75089859738 *T)
    sumZ0  +=       0.052 * cos(2.74630211595  + 131395.11544947961 *T)
    sumZ0  +=       0.055 * cos(1.23376269855  + 26137.89976347840 *T)
    sumZ0  +=       0.051 * cos(4.16406216247  + 78259.82808936459 *T)
    sumZ0  +=       0.048 * cos(5.64584502444  + 71492.99882347698 *T)
    sumZ0  +=       0.056 * cos(2.21556357721  + 151975.46535238638 *T)
    sumZ0  +=       0.052 * cos(2.58403290426  + 128850.44281258718 *T)
    sumZ0  +=       0.048 * cos(2.53427528158  + 130012.91751699500 *T)
    sumZ0  +=       0.045 * cos(3.30208209778  + 111590.28815789679 *T)
    sumZ0  +=       0.051 * cos(2.59172476683  + 78267.59076008058 *T)
    sumZ0  +=       0.047 * cos(1.65481269433  + 18093.37469954999 *T)
    sumZ0  +=       0.045 * cos(4.61275436981  + 2333.19639287200 *T)
    sumZ0  +=       0.043 * cos(4.08074925073  + 1596.18644228460 *T)
    sumZ0  +=       0.042 * cos(2.94276331302  + 24395.73747207180 *T)
    sumZ0  +=       0.042 * cos(4.02117925826  + 78477.00852016058 *T)
    sumZ0  +=       0.056 * cos(4.77237421864  + 28421.09953444620 *T)
    sumZ0  +=       0.047 * cos(6.11377086686  + 93028.94846799241 *T)
    sumZ0  +=       0.040 * cos(3.23768263734  + 25234.70675982219 *T)
    sumZ0  +=       0.045 * cos(5.03235979949  + 78256.59587772179 *T)
    sumZ0  +=       0.053 * cos(6.07709485433  + 10021.83728009940 *T)
    sumZ0  +=       0.050 * cos(5.81013134251  + 90829.86112470538 *T)
    sumZ0  +=       0.038 * cos(1.96485604602  + 632.78373931320 *T)
    sumZ0  +=       0.051 * cos(3.90465378478  + 74.78159856730 *T)
    sumZ0  +=       0.036 * cos(6.02764042024  + 78114.14622758799 *T)
    sumZ0  +=       0.035 * cos(0.19483010119  + 77197.21394753258 *T)
    sumZ0  +=       0.038 * cos(0.85372328004  + 26237.46633870879 *T)
    sumZ0  +=       0.036 * cos(3.61160802929  + 48835.19385644859 *T)
    sumZ0  +=       0.036 * cos(0.36428632576  + 27154.39861876420 *T)
    sumZ0  +=       0.037 * cos(2.58618864774  + 51707.84129279399 *T)
    sumZ0  +=       0.047 * cos(3.11077813997  + 16066.06586147480 *T)
    sumZ0  +=       0.034 * cos(5.90024410478  + 103.09277421860 *T)
    sumZ0  +=       0.035 * cos(1.89948168639  + 52169.17772424839 *T)
    sumZ0  +=       0.035 * cos(3.09037355598  + 77211.44104153418 *T)
    sumZ0  +=       0.035 * cos(5.86095139046  + 18207.81398823521 *T)
    sumZ0  +=       0.038 * cos(1.94104879109  + 76044.95232053580 *T)
    sumZ0  +=       0.032 * cos(1.45277855585  + 78270.82297172339 *T)
    sumZ0  +=       0.044 * cos(4.36195482299  + 52643.77127350280 *T)
    sumZ0  +=       0.043 * cos(0.13191769323  + 1581.95934828300 *T)
    sumZ0  +=       0.032 * cos(4.70833495807  + 71582.48457132299 *T)
    sumZ0  +=       0.037 * cos(3.75248447021  + 27140.17152476259 *T)
    sumZ0  +=       0.042 * cos(2.34341217670  + 84546.78527471398 *T)
    sumZ0  +=       0.031 * cos(2.09939558148  + 2648.45482547300 *T)
    sumZ0  +=       0.033 * cos(4.22087782315  + 111122.32316754240 *T)
    sumZ0  +=       0.032 * cos(5.93849066304  + 51742.09454527159 *T)
    sumZ0  +=       0.037 * cos(2.22778724531  + 25973.46385288896 *T)
    sumZ0  +=       0.031 * cos(0.06282672228  + 536.80451209540 *T)
    sumZ0  +=       0.030 * cos(0.17325551658  + 224.34479570190 *T)
    sumZ0  +=       0.030 * cos(4.46570213013  + 81706.28436968799 *T)
    sumZ0  +=       0.029 * cos(4.99680928394  + 24491.71669928959 *T)
    sumZ0  +=       0.035 * cos(5.04036468235  + 52182.43484204840 *T)
    sumZ0  +=       0.033 * cos(4.30956969733  + 14477.35118320000 *T)
    sumZ0  +=       0.039 * cos(3.99296635763  + 78690.30761559859 *T)
    sumZ0  +=       0.028 * cos(3.41723065188  + 7880.08915333899 *T)
    sumZ0  +=       0.039 * cos(0.38970280197  + 114564.89811250778 *T)
    sumZ0  +=       0.033 * cos(3.85407359005  + 143961.26714946239 *T)
    sumZ0  +=       0.029 * cos(2.14304937049  + 150866.08680029298 *T)
    sumZ0  +=       0.028 * cos(4.95187621332  + 956.28915597060 *T)
    sumZ0  +=       0.036 * cos(4.89757043872  + 26202.34243025941 *T)
    sumZ0  +=       0.030 * cos(0.55025093927  + 33967.99229491319 *T)
    sumZ0  +=       0.028 * cos(1.52043005918  + 40565.25432477420 *T)
    sumZ0  +=       0.027 * cos(5.29884427476  + 1162.47470440780 *T)
    sumZ0  +=       0.027 * cos(3.44440418351  + 141762.17980617538 *T)
    sumZ0  +=       0.035 * cos(2.50821332798  + 72936.23331633979 *T)
    sumZ0  +=       0.028 * cos(0.83354394743  + 42153.96900304900 *T)
    sumZ0  +=       0.036 * cos(4.02142487381  + 79852.78232000639 *T)
    sumZ0  +=       0.026 * cos(1.15352991801  + 51322.60990139639 *T)
    sumZ0  +=       0.026 * cos(0.13577439886  + 22747.29071487440 *T)
    sumZ0  +=       0.027 * cos(4.82379477303  + 110.20632121940 *T)
    sumZ0  +=       0.025 * cos(0.71097564554  + 50483.64061364600 *T)
    sumZ0  +=       0.024 * cos(2.92248010137  + 121335.60871375341 *T)
    sumZ0  +=       0.026 * cos(4.56151347419  + 183724.70054311279 *T)
    sumZ0  +=       0.024 * cos(2.16545180970  + 32769.12799497380 *T)
    sumZ0  +=       0.031 * cos(2.33336639968  + 51756.32163927320 *T)
    sumZ0  +=       0.032 * cos(2.93016714121  + 181555.94006083015 *T)
    sumZ0  +=       0.027 * cos(4.32935860229  + 124156.43985787958 *T)
    sumZ0  +=       0.023 * cos(6.16949223424  + 39743.76363275060 *T)
    sumZ0  +=       0.023 * cos(2.77333133719  + 155997.72788435058 *T)
    sumZ0  +=       0.022 * cos(2.90047805068  + 25863.55834587229 *T)
    sumZ0  +=       0.023 * cos(0.26296960044  + 103711.71527998280 *T)
    sumZ0  +=       0.022 * cos(0.49237419293  + 23439.44831610119 *T)
    sumZ0  +=       0.027 * cos(2.77269830479  + 24505.94379329119 *T)
    sumZ0  +=       0.022 * cos(0.39938514656  + 26164.16921284980 *T)
    sumZ0  +=       0.022 * cos(0.07820939757  + 78187.44335344699 *T)
    sumZ0  +=       0.022 * cos(3.71330627735  + 129483.91596626240 *T)
    sumZ0  +=       0.023 * cos(5.63887355161  + 522.57741809380 *T)
    sumZ0  +=       0.021 * cos(3.36821177399  + 1375.77379984580 *T)
    sumZ0  +=       0.029 * cos(5.62929042847  + 44181.27784112419 *T)
    sumZ0  +=       0.022 * cos(4.35422573947  + 26720.68688088739 *T)
    sumZ0  +=       0.021 * cos(5.12497732226  + 11610.55195837420 *T)
    sumZ0  +=       0.025 * cos(5.36835518856  + 26013.12154300690 *T)
    sumZ0  +=       0.024 * cos(5.55813303813  + 130969.20667296558 *T)
    sumZ0  +=       0.020 * cos(4.79424075731  + 25780.34552060460 *T)
    sumZ0  +=       0.020 * cos(3.82740705898  + 24925.42843716640 *T)
    sumZ0  +=       0.020 * cos(4.26878389727  + 25131.61398560359 *T)
    sumZ0  +=       0.021 * cos(5.10505150441  + 25977.69682035479 *T)
    sumZ0  +=       0.020 * cos(3.32855955978  + 131498.89763806018 *T)
    sumZ0  +=       0.020 * cos(1.29805086237  + 467.96499035440 *T)
    sumZ0  +=       0.023 * cos(2.06377601353  + 100909.03762133140 *T)
    sumZ0  +=       0.027 * cos(4.46165349883  + 316.39186965660 *T)
    sumZ0  +=       0.019 * cos(2.33132135274  + 26395.46076254379 *T)
    sumZ0  +=       0.019 * cos(0.35229627877  + 78417.48823520739 *T)
    sumZ0  +=       0.019 * cos(2.86176933014  + 50579.61984086379 *T)
    sumZ0  +=       0.018 * cos(4.33826105585  + 104371.28232719658 *T)
    sumZ0  +=       0.022 * cos(1.69381062404  + 106570.36967048360 *T)
    sumZ0  +=       0.021 * cos(0.89581658307  + 339142.74084046454 *T)
    sumZ0  +=       0.025 * cos(5.95532312645  + 146314.13330323418 *T)
    sumZ0  +=       0.020 * cos(0.11725114676  + 52815.70356946240 *T)
    sumZ0  +=       0.018 * cos(1.66756826028  + 23888.81579828719 *T)
    sumZ0  +=       0.022 * cos(4.37150668135  + 1478.86657406440 *T)
    sumZ0  +=       0.017 * cos(4.39063105551  + 60055.89543648739 *T)
    sumZ0  +=       0.017 * cos(3.17668084118  + 70269.18098269838 *T)
    sumZ0  +=       0.017 * cos(1.74029229152  + 104138.31347085880 *T)
    sumZ0  +=       0.017 * cos(5.24411288462  + 102232.84870591838 *T)
    sumZ0  +=       0.019 * cos(0.11691865573  + 78109.93061423779 *T)
    sumZ0  +=       0.017 * cos(3.28697052824  + 25565.32572348040 *T)
    sumZ0  +=       0.019 * cos(0.84289230962  + 188276.65404017159 *T)
    sumZ0  +=       0.018 * cos(0.47252163890  + 27999.10262479140 *T)
    sumZ0  +=       0.016 * cos(2.81320924670  + 52101.02468458109 *T)
    sumZ0  +=       0.016 * cos(4.20710393311  + 134991.46920492980 *T)
    sumZ0  +=       0.019 * cos(0.54267379272  + 50593.84693486539 *T)
    sumZ0  +=       0.015 * cos(5.78613478653  + 24712.12934172840 *T)
    sumZ0  +=       0.018 * cos(2.23136091809  + 104331.94280539700 *T)
    sumZ0  +=       0.016 * cos(3.30957717580  + 97112.93697469680 *T)
    sumZ0  +=       0.016 * cos(1.02655918243  + 26162.68474014150 *T)
    sumZ0  +=       0.015 * cos(0.51501957594  + 13362.44970679920 *T)
    sumZ0  +=       0.015 * cos(5.44457668379  + 28286.99048486120 *T)
    sumZ0  +=       0.015 * cos(3.33697839260  + 206.18554843720 *T)
    sumZ0  +=       0.015 * cos(3.09319808895  + 25984.81036735560 *T)
    sumZ0  +=       0.019 * cos(5.95214817336  + 29428.51556827400 *T)
    sumZ0  +=       0.017 * cos(1.60411390680  + 26507.38778544939 *T)
    sumZ0  +=       0.016 * cos(3.62113927445  + 26610.48055966799 *T)
    sumZ0  +=       0.018 * cos(4.55509949626  + 51543.02254383520 *T)
    sumZ0  +=       0.016 * cos(4.95762784074  + 25771.51127191760 *T)
    sumZ0  +=       0.014 * cos(0.15853597470  + 647.01083331480 *T)
    sumZ0  +=       0.016 * cos(3.53658275362  + 25551.09862947879 *T)
    sumZ0  +=       0.013 * cos(2.28786159405  + 12725.45343477500 *T)
    sumZ0  +=       0.016 * cos(3.05451392033  + 26190.99591579279 *T)
    sumZ0  +=       0.013 * cos(5.81922389967  + 157483.01859105378 *T)
    sumZ0  +=       0.013 * cos(4.03160498458  + 24356.78078864160 *T)
    sumZ0  +=       0.013 * cos(4.49878848868  + 29416.03879785439 *T)
    sumZ0  +=       0.016 * cos(1.14380482430  + 26404.29501123080 *T)
    sumZ0  +=       0.016 * cos(4.77704247519  + 24609.03656750980 *T)
    sumZ0  +=       0.013 * cos(1.33944288754  + 52808.59002246159 *T)
    sumZ0  +=       0.015 * cos(1.50513017421  + 181505.94343892601 *T)
    sumZ0  +=       0.012 * cos(0.92918175452  + 104347.73123093881 *T)
    sumZ0  +=       0.014 * cos(2.44451494167  + 26421.75908234360 *T)
    sumZ0  +=       0.012 * cos(1.66798607577  + 51013.33157874059 *T)
    sumZ0  +=       0.012 * cos(4.24892890325  + 52225.80290505260 *T)
    sumZ0  +=       0.014 * cos(4.68967753702  + 25754.04720080480 *T)
    sumZ0  +=       0.013 * cos(0.63063491269  + 51639.00177105299 *T)
    sumZ0  +=       0.013 * cos(0.51980975221  + 51951.46148744649 *T)
    sumZ0  +=       0.013 * cos(0.11999669275  + 26294.08869001139 *T)
    sumZ0  +=       0.012 * cos(5.63990253843  + 104355.49390165480 *T)
    sumZ0  +=       0.013 * cos(0.37271479520  + 74923.09699802278 *T)
    sumZ0  +=       0.015 * cos(2.19958896111  + 22759.76748529401 *T)
    sumZ0  +=       0.013 * cos(1.47842980598  + 54294.57014352679 *T)
    sumZ0  +=       0.012 * cos(3.10508217703  + 27819.02549450680 *T)
    sumZ0  +=       0.011 * cos(2.42209220629  + 97580.90196505121 *T)
    sumZ0  +=       0.013 * cos(2.58513127775  + 116917.76426627958 *T)
    sumZ0  +=       0.011 * cos(4.76885794954  + 49527.35145767539 *T)
    sumZ0  +=       0.011 * cos(3.92593978045  + 26521.61487945100 *T)
    sumZ0  +=       0.011 * cos(2.89444277066  + 119116.85160956658 *T)
    sumZ0  +=       0.012 * cos(5.58737020094  + 25881.71759313700 *T)
    sumZ0  +=       0.013 * cos(5.59537220560  + 156100.82065856917 *T)
    sumZ0  +=       0.012 * cos(2.01299993075  + 51219.51712717779 *T)
    sumZ0  +=       0.011 * cos(0.99253734697  + 25455.11940226100 *T)
    sumZ0  +=       0.014 * cos(5.68122574937  + 154938.34595416137 *T)
    sumZ0  +=       0.015 * cos(5.27886514517  + 178063.36849396059 *T)
    sumZ0  +=       0.014 * cos(6.12654802167  + 52278.89905736699 *T)
    sumZ0  +=       0.011 * cos(0.06908636848  + 52072.71350892979 *T)
    sumZ0  +=       0.011 * cos(0.80982228740  + 104564.91166173480 *T)
    sumZ0  +=       0.013 * cos(3.96505351243  + 68241.87214462319 *T)
    sumZ0  +=       0.011 * cos(2.65559265085  + 52125.80966124419 *T)
    sumZ0  +=       0.012 * cos(6.26057108211  + 24176.70365835700 *T)
    sumZ0  +=       0.012 * cos(1.73021765373  + 104344.49901929598 *T)
    sumZ0  +=       0.010 * cos(2.77364193598  + 77829.99768684579 *T)
    sumZ0  +=       0.010 * cos(2.83314383566  + 104202.04936916218 *T)
    sumZ0  +=       0.010 * cos(4.05963464737  + 26089.38761428249 *T)
    sumZ0  +=       0.012 * cos(5.40316858194  + 110634.68841628821 *T)
    sumZ0  +=       0.010 * cos(6.17247219966  + 103299.34418310841 *T)
    sumZ0  +=       0.010 * cos(0.02153234343  + 137678.19129947099 *T)
    sumZ0  +=       0.012 * cos(2.25188175428  + 52065.59996192899 *T)
    sumZ0  +=       0.010 * cos(3.11650831039  + 26086.41866886590 *T)
    sumZ0  +=       0.010 * cos(3.27287049435  + 103285.11708910680 *T)
    sumZ0  +=       0.010 * cos(1.38532261128  + 25446.48957983520 *T)
    sumZ0  +=       0.012 * cos(1.52019007531  + 54509.00267602040 *T)
    sumZ0  +=       0.011 * cos(6.21574870463  + 61279.71327726600 *T)
    sumZ0  +=       0.009 * cos(0.49710408415  + 2221.85663459700 *T)
    sumZ0  +=       0.009 * cos(2.88821691711  + 90695.75207512038 *T)
    sumZ0  +=       0.011 * cos(2.16522629249  + 26624.70765366959 *T)
    sumZ0  +=       0.009 * cos(4.80391342687  + 27311.72098235281 *T)
    sumZ0  +=       0.009 * cos(4.23373474411  + 52250.58788171570 *T)
    sumZ0  +=       0.009 * cos(5.74437259170  + 26729.31670331319 *T)
    sumZ0  +=       0.010 * cos(3.47517485960  + 140652.80125408198 *T)
    sumZ0  +=       0.010 * cos(2.80299508279  + 55516.41870984820 *T)
    sumZ0  +=       0.013 * cos(4.07250716348  + 34282.17847478280 *T)
    sumZ0  +=       0.009 * cos(0.66918270135  + 52698.38370124219 *T)
    sumZ0  +=       0.010 * cos(5.43273059228  + 126996.94076290558 *T)
    sumZ0  +=       0.009 * cos(1.51413989531  + 97670.38771289718 *T)
    sumZ0  +=       0.010 * cos(1.07944281160  + 742.99006053260 *T)
    sumZ0  +=       0.009 * cos(3.33909678408  + 1911.19948321720 *T)
    sumZ0  +=       0.009 * cos(6.28167845492  + 147423.51185532758 *T)
    sumZ0  +=       0.010 * cos(0.71259250113  + 137210.22630911658 *T)
    sumZ0  +=       0.009 * cos(3.57005441415  + 50800.03248330259 *T)
    sumZ0  +=       0.009 * cos(5.50160718751  + 2199.08734328700 *T)
    sumZ0  +=       0.009 * cos(2.08845571909  + 51859.41441349179 *T)
    sumZ0  +=       0.009 * cos(4.25308230419  + 26222.01219115920 *T)
    sumZ0  +=       0.009 * cos(2.50157520492  + 24864.08530079559 *T)
    sumZ0  +=       0.008 * cos(2.47794653703  + 25440.89230825939 *T)
    sumZ0  +=       0.009 * cos(2.85482376043  + 25953.79409198919 *T)
    sumZ0  +=       0.008 * cos(2.29776576843  + 23866.04650697719 *T)
    sumZ0  +=       0.009 * cos(0.78685007135  + 104778.21075717278 *T)
    sumZ0  +=       0.008 * cos(4.62944551319  + 52492.19815280499 *T)
    sumZ0  +=       0.009 * cos(5.48506768547  + 77844.22478084739 *T)
    sumZ0  +=       0.010 * cos(1.12934836943  + 78731.67441507700 *T)
    sumZ0  +=       0.008 * cos(2.59549267900  + 26941.09952332620 *T)
    sumZ0  +=       0.008 * cos(4.95780242181  + 78257.08086582259 *T)
    sumZ0  +=       0.008 * cos(2.55696609752  + 17893.62780836560 *T)
    sumZ0  +=       0.008 * cos(4.52002369897  + 104358.72611329758 *T)
    sumZ0  +=       0.008 * cos(3.82529998469  + 76571.54375522019 *T)
    sumZ0  +=       0.010 * cos(2.73498400513  + 323.50541665740 *T)
    sumZ0  +=       0.009 * cos(1.57363188943  + 24998.19435038059 *T)
    sumZ0  +=       0.008 * cos(0.65301230309  + 8194.27533320860 *T)
    sumZ0  +=       0.008 * cos(4.30691637723  + 77410.51304297059 *T)
    sumZ0  +=       0.008 * cos(5.61648727808  + 77795.74443436819 *T)
    sumZ0  +=       0.009 * cos(2.56370604921  + 50696.93970908399 *T)
    sumZ0  +=       0.008 * cos(5.53494196837  + 2008.55753915900 *T)
    sumZ0  +=       0.009 * cos(0.81310827208  + 105940.68546158058 *T)
    sumZ0  +=       0.008 * cos(3.03053180726  + 27780.06881107659 *T)
    sumZ0  +=       0.008 * cos(1.09072579794  + 949.17560896980 *T)
    sumZ0  +=       0.008 * cos(5.21537078773  + 176953.98994186718 *T)
    sumZ0  +=       0.008 * cos(2.88040588282  + 36109.74042167360 *T)
    sumZ0  +=       0.008 * cos(3.90879186991  + 54087.00576636560 *T)
    sumZ0  +=       0.007 * cos(2.09304055777  + 25344.91308104160 *T)
    sumZ0  +=       0.007 * cos(3.77982679537  + 19.66976089979 *T)
    sumZ0  +=       0.008 * cos(2.01169539344  + 31415.37924995700 *T)
    sumZ0  +=       0.008 * cos(6.01136820214  + 207643.84320240439 *T)
    sumZ0  +=       0.008 * cos(6.24227744862  + 38813.35657634920 *T)
    sumZ0  +=       0.008 * cos(1.33108336808  + 55503.94193942859 *T)
    sumZ0  +=       0.007 * cos(3.30804741677  + 23919.14265929160 *T)
    sumZ0  +=       0.008 * cos(3.48746520229  + 53242.30176033840 *T)
    sumZ0  +=       0.008 * cos(5.56477337187  + 27177.61193276780 *T)
    sumZ0  +=       0.009 * cos(5.00034866965  + 102132.85546211000 *T)
    sumZ0  +=       0.007 * cos(0.40751314845  + 2111.65031337760 *T)
    sumZ0  +=       0.008 * cos(1.66471235682  + 52290.24557183361 *T)
    sumZ0  +=       0.007 * cos(0.77215567463  + 26073.67604757259 *T)
    sumZ0  +=       0.007 * cos(5.32306426879  + 52595.29092702359 *T)
    sumZ0  +=       0.007 * cos(2.57023734016  + 24448.83362438620 *T)
    sumZ0  +=       0.007 * cos(1.40330361432  + 150244.34299945380 *T)
    sumZ0  +=       0.008 * cos(3.96541390761  + 52325.36948028299 *T)
    sumZ0  +=       0.008 * cos(0.64215744063  + 53228.07466633679 *T)
    sumZ0  +=       0.007 * cos(5.40994893955  + 25508.21555457540 *T)
    sumZ0  +=       0.009 * cos(0.63097343517  + 170049.17029103660 *T)
    sumZ0  +=       0.006 * cos(2.73735624509  + 153.77881048480 *T)
    sumZ0  +=       0.006 * cos(1.57808561874  + 49842.60989027639 *T)
    sumZ0  +=       0.006 * cos(5.85639973088  + 182085.63102592481 *T)
    sumZ0  +=       0.007 * cos(1.22602076365  + 107794.18751126219 *T)
    sumZ0  +=       0.007 * cos(2.99828544115  + 22625.65843570900 *T)
    sumZ0  +=       0.007 * cos(3.38113527243  + 12546.48193908300 *T)
    sumZ0  +=       0.006 * cos(5.98824736570  + 76667.52298243798 *T)
    sumZ0  +=       0.007 * cos(2.75252330601  + 172402.03644480839 *T)
    sumZ0  +=       0.007 * cos(0.93184314111  + 3328.13565628019 *T)
    sumZ0  +=       0.006 * cos(2.72361970528  + 26709.64694241340 *T)
    sumZ0  +=       0.008 * cos(1.81192469081  + 78270.33798362259 *T)
    sumZ0  +=       0.006 * cos(2.31595886989  + 157057.10981453978 *T)
    sumZ0  +=       0.006 * cos(1.73557063122  + 26667.59072857300 *T)
    sumZ0  +=       0.006 * cos(4.37863242786  + 25466.15934073500 *T)
    sumZ0  +=       0.006 * cos(3.19403160700  + 104275.34649502118 *T)
    sumZ0  +=       0.006 * cos(1.09823646265  + 25867.49049913539 *T)
    sumZ0  +=       0.007 * cos(0.11236053840  + 25764.39772491679 *T)
    sumZ0  +=       0.007 * cos(5.69537766284  + 66653.15746634839 *T)
    sumZ0  +=       0.007 * cos(1.98238737869  + 846.08283475120 *T)
    sumZ0  +=       0.006 * cos(0.65894058731  + 116783.65521669458 *T)
    sumZ0  +=       0.005 * cos(4.66761334131  + 2125.87740737920 *T)
    sumZ0  +=       0.006 * cos(4.52221761723  + 27726.97265876220 *T)
    sumZ0  +=       0.006 * cos(0.09840227420  + 167850.08294774959 *T)
    sumZ0  +=       0.006 * cos(3.36608415071  + 129799.61842155698 *T)
    sumZ0  +=       0.007 * cos(5.27549434329  + 52061.36699446317 *T)
    sumZ0  +=       0.007 * cos(1.32305784951  + 209812.60368468691 *T)
    sumZ0  +=       0.006 * cos(3.60884349448  + 78903.60671103658 *T)
    sumZ0  +=       0.005 * cos(3.76183094594  + 28256.66362385679 *T)
    sumZ0  +=       0.006 * cos(5.26397966812  + 75930.51303185058 *T)
    sumZ0  +=       0.006 * cos(3.23097479261  + 27669.86248985719 *T)
    sumZ0  +=       0.005 * cos(5.68060753615  + 1265.56747862640 *T)
    sumZ0  +=       0.005 * cos(3.87273283224  + 214364.55718174577 *T)
    sumZ0  +=       0.005 * cos(3.93047654027  + 365230.64398203877 *T)
    sumZ0  +=       0.006 * cos(2.60562458198  + 44295.71712980940 *T)
    sumZ0  +=       0.005 * cos(1.56438384826  + 60370.08161635699 *T)
    sumZ0  +=       0.005 * cos(0.91437456221  + 27684.08958385880 *T)
    sumZ0  +=       0.005 * cos(2.78304413268  + 25169.97285559240 *T)
    sumZ0  +=       0.006 * cos(5.76131236705  + 58857.03113654799 *T)
    sumZ0  +=       0.005 * cos(2.67085954875  + 38.13303563780 *T)
    sumZ0  +=       0.006 * cos(0.50453577398  + 155571.81910783658 *T)
    sumZ0  +=       0.005 * cos(3.86250449828  + 52381.99183158559 *T)
    sumZ0  +=       0.005 * cos(4.32361771821  + 27005.83342755599 *T)
    sumZ0  +=       0.005 * cos(1.63783306460  + 34082.43158359840 *T)
    sumZ0  +=       0.006 * cos(4.13235037436  + 29550.14784743939 *T)
    sumZ0  +=       0.005 * cos(0.09069856194  + 13541.42120249119 *T)
    sumZ0  +=       0.005 * cos(5.50663993071  + 26402.08932144380 *T)
    sumZ0  +=       0.004 * cos(5.22554265930  + 28736.35796704720 *T)
    sumZ0  +=       0.005 * cos(5.88997457155  + 78188.92782615528 *T)
    sumZ0  +=       0.005 * cos(1.29427545799  + 77630.92568540938 *T)
    sumZ0  +=       0.006 * cos(3.65818107099  + 76681.75007643958 *T)
    sumZ0  +=       0.005 * cos(1.59378640261  + 25773.71696170459 *T)
    sumZ0  +=       0.005 * cos(5.53985466815  + 24203.00197815680 *T)
    sumZ0  +=       0.005 * cos(1.10121885185  + 130459.18546877075 *T)
    sumZ0  +=       0.005 * cos(3.55866759420  + 52252.07235442399 *T)
    sumZ0  +=       0.004 * cos(5.49843651098  + 52712.61079524379 *T)
    sumZ0  +=       0.005 * cos(3.39959787962  + 104505.39137678158 *T)
    sumZ0  +=       0.005 * cos(1.94175393423  + 1272.68102562720 *T)
    sumZ0  +=       0.004 * cos(0.58115883263  + 26118.23000257860 *T)
    sumZ0  +=       0.004 * cos(0.54682947961  + 157586.80077963436 *T)
    sumZ0  +=       0.004 * cos(2.04212740892  + 148.07872442630 *T)
    sumZ0  +=       0.005 * cos(1.64853656178  + 26198.10946279360 *T)
    sumZ0  +=       0.005 * cos(0.18946812001  + 1.48447270830 *T)
    sumZ0  +=       0.004 * cos(5.04774492766  + 112231.70171963578 *T)
    sumZ0  +=       0.004 * cos(3.24889530044  + 26312.24793727610 *T)
    sumZ0  +=       0.005 * cos(2.14793916248  + 128320.75184749260 *T)
    sumZ0  +=       0.004 * cos(3.03681019540  + 151.04766984290 *T)
    sumZ0  +=       0.004 * cos(0.23608219351  + 26057.57628056979 *T)
    sumZ0  +=       0.004 * cos(2.78449581166  + 25138.72753260440 *T)
    sumZ0  +=       0.004 * cos(1.74750321635  + 27044.19229754480 *T)
    sumZ0  +=       0.004 * cos(5.26585590029  + 138319.60486120995 *T)
    sumZ0  +=       0.004 * cos(2.16361705057  + 78366.80219894118 *T)
    sumZ0  +=       0.005 * cos(1.86438708559  + 51653.22886505459 *T)
    sumZ0  +=       0.004 * cos(3.67745684549  + 148532.89040742096 *T)
    sumZ0  +=       0.005 * cos(3.36418574107  + 86457.98475793119 *T)
    sumZ0  +=       0.004 * cos(4.39817949768  + 78896.49316403578 *T)
    sumZ0  +=       0.004 * cos(3.16495363173  + 104197.83375581198 *T)
    sumZ0  +=       0.004 * cos(5.96720583626  + 1485.98012106520 *T)
    sumZ0  +=       0.003 * cos(3.62369510287  + 51852.30086649099 *T)
    sumZ0  +=       0.004 * cos(4.95850373202  + 80382.47328510099 *T)
    sumZ0  +=       0.003 * cos(0.24273955075  + 128106.31931499895 *T)
    sumZ0  +=       0.004 * cos(6.09423095974  + 51528.79544983359 *T)
    sumZ0  +=       0.004 * cos(2.26832726016  + 64901.25971792339 *T)
    sumZ0  +=       0.004 * cos(5.37231956323  + 99024.13645791401 *T)
    sumZ0  +=       0.004 * cos(5.83099981006  + 26411.40855823160 *T)
    sumZ0  +=       0.003 * cos(5.27877050710  + 14.22709400160 *T)
    sumZ0  +=       0.004 * cos(1.01079656557  + 16703.06213349900 *T)
    sumZ0  +=       0.003 * cos(5.29624442233  + 51432.81622261579 *T)
    sumZ0  +=       0.003 * cos(1.99372517333  + 9384.84100807520 *T)
    sumZ0  +=       0.003 * cos(1.08549694329  + 92741.06060792258 *T)
    sumZ0  +=       0.004 * cos(2.61532511231  + 27250.37784598199 *T)
    sumZ0  +=       0.004 * cos(3.84384246230  + 65831.66677432480 *T)
    sumZ0  +=       0.004 * cos(5.07647835306  + 118828.96374949680 *T)
    sumZ0  +=       0.004 * cos(0.95691561960  + 1795.25844372100 *T)
    sumZ0  +=       0.003 * cos(4.56071055194  + 130419.84594697120 *T)
    sumZ0  +=       0.004 * cos(3.21002699852  + 51969.62073471119 *T)
    sumZ0  +=       0.003 * cos(0.53542089836  + 25985.94062330859 *T)
    sumZ0  +=       0.003 * cos(0.86346795613  + 27972.80430499159 *T)
    sumZ0  +=       0.004 * cos(1.11850373850  + 52609.51802102519 *T)
    sumZ0  +=       0.003 * cos(5.13697350283  + 26189.86565983980 *T)
    sumZ0  +=       0.003 * cos(3.83162907268  + 207593.84658050016 *T)
    sumZ0  +=       0.003 * cos(3.70405700948  + 193937.98608932376 *T)
    sumZ0  +=       0.003 * cos(1.82907554275  + 153084.84390447981 *T)
    sumZ0  +=       0.003 * cos(1.75677414102  + 158746.17595363196 *T)
    sumZ0  +=       0.003 * cos(0.34579536632  + 161079.37234650398 *T)
    sumZ0  +=       0.003 * cos(3.85483830173  + 39629.32434406539 *T)
    sumZ0  +=       0.003 * cos(0.82694378727  + 51868.24866217880 *T)
    sumZ0  +=       0.003 * cos(0.16328069810  + 27463.67694142000 *T)
    sumZ0  +=       0.003 * cos(4.08971647963  + 130226.21661243298 *T)
    sumZ0  +=       0.004 * cos(3.70176043302  + 163766.09444104519 *T)
    sumZ0  +=       0.003 * cos(1.31853873267  + 204151.27163553479 *T)
    sumZ0  +=       0.004 * cos(2.50849600287  + 220.41264243880 *T)
    sumZ0  +=       0.003 * cos(3.29076294593  + 125112.03959948818 *T)
    sumZ0  +=       0.003 * cos(5.82527117029  + 49953.94964855139 *T)
    sumZ0  +=       0.003 * cos(1.83258674312  + 37698.45509994840 *T)
    sumZ0  +=       0.003 * cos(1.84707936363  + 26076.85744131030 *T)
    sumZ0  +=       0.003 * cos(3.96840940578  + 130432.40216087016 *T)
    sumZ0  +=       0.003 * cos(0.01192870778  + 26098.94884183810 *T)
    sumZ0  +=       0.003 * cos(2.24128507678  + 78160.61665050399 *T)
    sumZ0  +=       0.003 * cos(1.67457691866  + 182188.72380014337 *T)
    sumZ0  +=       0.003 * cos(1.36021422148  + 136722.59155786238 *T)

    # MERCURY Z Coord
    sumZ1  = 0
    sumZ1  +=  172388.569 * cos(3.47961470688  + 26087.90314157420 *T)
    sumZ1  +=   84745.328 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ1  +=    4790.619 * cos(6.21921644635  + 52175.80628314840 *T)
    sumZ1  +=    1359.100 * cos(0.49329143004  + 78263.70942472259 *T)
    sumZ1  +=     602.217 * cos(3.41138035633  + 104351.61256629678 *T)
    sumZ1  +=     191.001 * cos(0.14519065991  + 130439.51570787100 *T)
    sumZ1  +=      55.361 * cos(3.18058121998  + 156527.41884944518 *T)
    sumZ1  +=      15.488 * cos(6.22255945229  + 182615.32199101936 *T)
    sumZ1  +=      10.854 * cos(0.29329316365  + 27197.28169366760 *T)
    sumZ1  +=       4.262 * cos(2.98444809115  + 208703.22513259359 *T)
    sumZ1  +=       4.999 * cos(0.56827577867  + 24978.52458948080 *T)
    sumZ1  +=       2.282 * cos(2.87281226699  + 31749.23519072640 *T)
    sumZ1  +=       2.500 * cos(3.37464102020  + 53285.18483524180 *T)
    sumZ1  +=       1.480 * cos(3.59930190722  + 51066.42773105500 *T)
    sumZ1  +=       1.163 * cos(6.03119799366  + 234791.12827416777 *T)
    sumZ1  +=       1.151 * cos(1.94233400975  + 51116.42435295920 *T)
    sumZ1  +=       0.808 * cos(0.63296975442  + 25028.52121138500 *T)
    sumZ1  +=       0.781 * cos(1.83801930165  + 5661.33204915220 *T)
    sumZ1  +=       0.676 * cos(4.44275341808  + 27043.50288318280 *T)
    sumZ1  +=       0.873 * cos(3.99274661843  + 21535.94964451540 *T)
    sumZ1  +=       0.636 * cos(0.60897167697  + 20426.57109242200 *T)
    sumZ1  +=       0.538 * cos(5.92697698725  + 57837.13833230060 *T)
    sumZ1  +=       0.528 * cos(1.89223267777  + 1059.38193018920 *T)
    sumZ1  +=       0.503 * cos(4.12049535056  + 47623.85278608960 *T)
    sumZ1  +=       0.450 * cos(6.08529735298  + 529.69096509460 *T)
    sumZ1  +=       0.553 * cos(2.50090977683  + 1109.37855209340 *T)
    sumZ1  +=       0.445 * cos(0.35693140869  + 77154.33087262919 *T)
    sumZ1  +=       0.459 * cos(0.10915184787  + 79373.08797681599 *T)
    sumZ1  +=       0.370 * cos(2.24953657104  + 10213.28554621100 *T)
    sumZ1  +=       0.414 * cos(3.01450201380  + 26107.57290247399 *T)
    sumZ1  +=       0.417 * cos(5.31402407370  + 37410.56723987860 *T)
    sumZ1  +=       0.306 * cos(3.10769212775  + 25558.21217647960 *T)
    sumZ1  +=       0.316 * cos(2.79570706285  + 260879.03141574195 *T)
    sumZ1  +=       0.296 * cos(4.12556468959  + 26068.23338067440 *T)
    sumZ1  +=       0.332 * cos(5.93758113684  + 26084.02180621620 *T)
    sumZ1  +=       0.234 * cos(3.03370038245  + 25132.30339996560 *T)
    sumZ1  +=       0.250 * cos(1.18440188253  + 26091.78447693220 *T)
    sumZ1  +=       0.215 * cos(1.41611963529  + 41962.52073693740 *T)
    sumZ1  +=       0.206 * cos(4.97267994525  + 77204.32749453338 *T)
    sumZ1  +=       0.208 * cos(5.60196522278  + 4551.95349705880 *T)
    sumZ1  +=       0.251 * cos(3.12319370712  + 30639.85663863300 *T)
    sumZ1  +=       0.230 * cos(0.95862112790  + 50586.73338786459 *T)
    sumZ1  +=       0.193 * cos(4.90260211702  + 13521.75144159140 *T)
    sumZ1  +=       0.243 * cos(0.30112320330  + 1589.07289528380 *T)
    sumZ1  +=       0.186 * cos(5.87650538684  + 39609.65458316560 *T)
    sumZ1  +=       0.211 * cos(1.27593674125  + 53131.40602475700 *T)
    sumZ1  +=       0.188 * cos(0.09339135236  + 28306.66024576099 *T)
    sumZ1  +=       0.191 * cos(4.77084065304  + 51646.11531805379 *T)
    sumZ1  +=       0.167 * cos(2.62739427985  + 32858.61374281979 *T)
    sumZ1  +=       0.162 * cos(5.02172677195  + 15874.61759536320 *T)
    sumZ1  +=       0.140 * cos(1.37984211372  + 51749.20809227239 *T)
    sumZ1  +=       0.134 * cos(0.57882342916  + 23869.14603738740 *T)
    sumZ1  +=       0.132 * cos(3.40338891449  + 103242.23401420339 *T)
    sumZ1  +=       0.144 * cos(4.41836879515  + 22645.32819660879 *T)
    sumZ1  +=       0.143 * cos(5.55896069473  + 25661.30495069820 *T)
    sumZ1  +=       0.131 * cos(2.17510821075  + 63498.47038145279 *T)
    sumZ1  +=       0.143 * cos(1.22805602725  + 46514.47423399620 *T)
    sumZ1  +=       0.136 * cos(2.45569156259  + 24498.83024629040 *T)
    sumZ1  +=       0.120 * cos(2.89969435814  + 7.11354700080 *T)
    sumZ1  +=       0.104 * cos(0.89744950821  + 73711.75592766379 *T)
    sumZ1  +=       0.119 * cos(1.44014554390  + 43071.89928903080 *T)
    sumZ1  +=       0.126 * cos(1.93369301602  + 14765.23904326980 *T)
    sumZ1  +=       0.107 * cos(4.29215813932  + 19317.19254032860 *T)
    sumZ1  +=       0.089 * cos(5.84175706917  + 51220.20654153979 *T)
    sumZ1  +=       0.093 * cos(2.27928877976  + 19804.82729158280 *T)
    sumZ1  +=       0.083 * cos(3.17188415740  + 29530.47808653960 *T)
    sumZ1  +=       0.076 * cos(5.10839374532  + 36301.18868778519 *T)
    sumZ1  +=       0.079 * cos(2.34370769368  + 1052.26838318840 *T)
    sumZ1  +=       0.080 * cos(2.11190196063  + 426.59819087600 *T)
    sumZ1  +=       0.076 * cos(6.05836593763  + 52195.47604404819 *T)
    sumZ1  +=       0.072 * cos(1.89073508118  + 1066.49547719000 *T)
    sumZ1  +=       0.082 * cos(3.03552784633  + 3442.57494496540 *T)
    sumZ1  +=       0.075 * cos(3.11962204135  + 105460.99111839020 *T)
    sumZ1  +=       0.086 * cos(5.84461147075  + 286966.93455731618 *T)
    sumZ1  +=       0.069 * cos(1.89458844998  + 27147.28507176339 *T)
    sumZ1  +=       0.089 * cos(4.10904716360  + 72602.37737557039 *T)
    sumZ1  +=       0.077 * cos(5.34990441360  + 213.29909543800 *T)
    sumZ1  +=       0.087 * cos(5.53929952701  + 9103.90699411760 *T)
    sumZ1  +=       0.070 * cos(4.43072071341  + 639.89728631400 *T)
    sumZ1  +=       0.059 * cos(3.32089922507  + 54394.56338733519 *T)
    sumZ1  +=       0.061 * cos(0.53721787795  + 25035.63475838580 *T)
    sumZ1  +=       0.062 * cos(1.36724381328  + 6283.07584999140 *T)
    sumZ1  +=       0.073 * cos(6.16650137658  + 56727.75978020720 *T)
    sumZ1  +=       0.069 * cos(0.88932506586  + 52156.13652224860 *T)
    sumZ1  +=       0.055 * cos(2.54693938612  + 65697.55772473979 *T)
    sumZ1  +=       0.061 * cos(2.70208782780  + 52171.92494779040 *T)
    sumZ1  +=       0.058 * cos(3.90075823666  + 76674.63652943878 *T)
    sumZ1  +=       0.064 * cos(6.09232375776  + 26080.78959457339 *T)
    sumZ1  +=       0.052 * cos(2.85118534529  + 16983.99614745660 *T)
    sumZ1  +=       0.063 * cos(5.90351345253  + 58946.51688439399 *T)
    sumZ1  +=       0.052 * cos(1.08065292487  + 25021.40766438420 *T)
    sumZ1  +=       0.050 * cos(5.98403755810  + 955.59974160860 *T)
    sumZ1  +=       0.050 * cos(2.34891444356  + 51535.90899683439 *T)
    sumZ1  +=       0.048 * cos(4.15049070651  + 12566.15169998280 *T)
    sumZ1  +=       0.047 * cos(3.35524401909  + 52168.69273614759 *T)
    sumZ1  +=       0.060 * cos(3.40462418234  + 25874.60404613620 *T)
    sumZ1  +=       0.059 * cos(4.98660394607  + 6770.71060124560 *T)
    sumZ1  +=       0.065 * cos(1.57121191195  + 52705.49724824299 *T)
    sumZ1  +=       0.055 * cos(5.05311644127  + 53235.18821333759 *T)
    sumZ1  +=       0.056 * cos(4.79703863542  + 25448.00585526019 *T)
    sumZ1  +=       0.045 * cos(3.87928453463  + 48733.23133818299 *T)
    sumZ1  +=       0.054 * cos(2.11712920969  + 33326.57873317420 *T)
    sumZ1  +=       0.044 * cos(2.58011573987  + 2218.75710418680 *T)
    sumZ1  +=       0.057 * cos(4.23895384399  + 52179.68761850640 *T)
    sumZ1  +=       0.042 * cos(2.38859796348  + 83925.04147387479 *T)
    sumZ1  +=       0.040 * cos(5.07476924931  + 51109.31080595839 *T)
    sumZ1  +=       0.039 * cos(0.16837719613  + 129330.13715577760 *T)
    sumZ1  +=       0.041 * cos(5.06879737373  + 59414.48187474840 *T)
    sumZ1  +=       0.045 * cos(0.74507723970  + 2118.76386037840 *T)
    sumZ1  +=       0.045 * cos(3.04665350718  + 27676.97603685800 *T)
    sumZ1  +=       0.039 * cos(4.70158354540  + 69159.80243060499 *T)
    sumZ1  +=       0.041 * cos(5.93310326966  + 26095.01668857500 *T)
    sumZ1  +=       0.031 * cos(3.33432475040  + 45892.73043315699 *T)
    sumZ1  +=       0.030 * cos(2.52574953749  + 77308.10968311399 *T)
    sumZ1  +=       0.032 * cos(3.64034199518  + 49957.04917896160 *T)
    sumZ1  +=       0.030 * cos(1.16170502578  + 26301.20223701220 *T)
    sumZ1  +=       0.039 * cos(4.76885068068  + 40853.14218484400 *T)
    sumZ1  +=       0.030 * cos(3.03342131333  + 52182.91983014920 *T)
    sumZ1  +=       0.033 * cos(2.00778363573  + 7238.67559160000 *T)
    sumZ1  +=       0.034 * cos(0.83933724452  + 98690.28051714458 *T)
    sumZ1  +=       0.028 * cos(2.06072938372  + 25934.12433108940 *T)
    sumZ1  +=       0.033 * cos(4.21710497504  + 79219.30916633119 *T)
    sumZ1  +=       0.026 * cos(4.99085554385  + 38519.94579197200 *T)
    sumZ1  +=       0.026 * cos(6.08273991212  + 18849.22754997420 *T)
    sumZ1  +=       0.025 * cos(4.07659089438  + 26617.59410666880 *T)
    sumZ1  +=       0.024 * cos(4.39675973371  + 77837.11123384659 *T)
    sumZ1  +=       0.030 * cos(0.25587368521  + 50057.04242277000 *T)
    sumZ1  +=       0.024 * cos(1.03737813626  + 45405.09568190280 *T)
    sumZ1  +=       0.024 * cos(1.77374954914  + 13655.86049117640 *T)
    sumZ1  +=       0.026 * cos(1.35857651349  + 77734.01845962799 *T)
    sumZ1  +=       0.028 * cos(4.51227694460  + 78793.40038981718 *T)
    sumZ1  +=       0.024 * cos(2.91891144270  + 82815.66292178139 *T)
    sumZ1  +=       0.021 * cos(5.13585062761  + 89586.37352302698 *T)
    sumZ1  +=       0.024 * cos(0.58794385685  + 11322.66409830440 *T)
    sumZ1  +=       0.028 * cos(1.99718125621  + 23969.13928119580 *T)
    sumZ1  +=       0.022 * cos(6.06219672387  + 419.48464387520 *T)
    sumZ1  +=       0.020 * cos(5.52785966186  + 433.71173787680 *T)
    sumZ1  +=       0.028 * cos(4.50140690236  + 51123.53789995999 *T)
    sumZ1  +=       0.020 * cos(4.37342000676  + 68050.42387851159 *T)
    sumZ1  +=       0.019 * cos(6.20105117352  + 55618.38122811380 *T)
    sumZ1  +=       0.023 * cos(2.61451042759  + 313054.83769889042 *T)
    sumZ1  +=       0.018 * cos(4.29485153949  + 23754.70674870219 *T)
    sumZ1  +=       0.022 * cos(1.87766446801  + 79323.09135491178 *T)
    sumZ1  +=       0.020 * cos(1.07656223897  + 25619.93815121980 *T)
    sumZ1  +=       0.016 * cos(1.03604660588  + 12432.04265039780 *T)
    sumZ1  +=       0.016 * cos(3.43674109707  + 25654.19140369739 *T)
    sumZ1  +=       0.020 * cos(2.97083013896  + 110012.94461544900 *T)
    sumZ1  +=       0.017 * cos(4.43931851455  + 38654.05484155699 *T)
    sumZ1  +=       0.015 * cos(2.81279895693  + 78283.37918562238 *T)
    sumZ1  +=       0.018 * cos(3.78270508656  + 25973.46385288896 *T)
    sumZ1  +=       0.018 * cos(5.12059646549  + 52389.10537858640 *T)
    sumZ1  +=       0.017 * cos(2.85670047821  + 25668.41849769900 *T)
    sumZ1  +=       0.018 * cos(3.94184263610  + 78244.03966382280 *T)
    sumZ1  +=       0.016 * cos(1.58918523327  + 66941.04532641819 *T)
    sumZ1  +=       0.013 * cos(2.50556550656  + 1596.18644228460 *T)
    sumZ1  +=       0.013 * cos(6.23648492936  + 80482.46652890938 *T)
    sumZ1  +=       0.015 * cos(1.01343393682  + 78267.59076008058 *T)
    sumZ1  +=       0.013 * cos(5.74551369637  + 78259.82808936459 *T)
    sumZ1  +=       0.013 * cos(5.49092142595  + 77623.81213840858 *T)
    sumZ1  +=       0.015 * cos(3.65531240129  + 853.19638175200 *T)
    sumZ1  +=       0.016 * cos(3.34679351675  + 26202.34243025941 *T)
    sumZ1  +=       0.012 * cos(4.99331123174  + 27154.39861876420 *T)
    sumZ1  +=       0.012 * cos(0.15067899984  + 78256.59587772179 *T)
    sumZ1  +=       0.012 * cos(1.23744980330  + 26727.80042788820 *T)
    sumZ1  +=       0.012 * cos(1.88171610012  + 77197.21394753258 *T)
    sumZ1  +=       0.016 * cos(1.02871649406  + 74821.13447975718 *T)
    sumZ1  +=       0.014 * cos(6.00648288256  + 53764.87917843220 *T)
    sumZ1  +=       0.011 * cos(1.03664866140  + 51962.50718771040 *T)
    sumZ1  +=       0.012 * cos(0.48930963657  + 102762.53967101300 *T)
    sumZ1  +=       0.012 * cos(3.21811087935  + 155418.04029735178 *T)
    sumZ1  +=       0.012 * cos(1.85815756639  + 103292.23063610760 *T)
    sumZ1  +=       0.012 * cos(3.87370634811  + 124778.18365871880 *T)
    sumZ1  +=       0.011 * cos(1.35784311134  + 51742.09454527159 *T)
    sumZ1  +=       0.011 * cos(2.25817150238  + 44937.13069154840 *T)
    sumZ1  +=       0.012 * cos(2.64551987584  + 26137.89976347840 *T)
    sumZ1  +=       0.011 * cos(2.75252916894  + 26514.50133245020 *T)
    sumZ1  +=       0.010 * cos(5.54274166657  + 103396.01282468818 *T)
    sumZ1  +=       0.009 * cos(6.12478323746  + 78270.82297172339 *T)
    sumZ1  +=       0.013 * cos(5.85664456058  + 136100.84775702321 *T)
    sumZ1  +=       0.011 * cos(2.43318609972  + 85034.42002596818 *T)
    sumZ1  +=       0.011 * cos(2.32000962427  + 64607.84893354619 *T)
    sumZ1  +=       0.013 * cos(3.22527913647  + 28421.09953444620 *T)
    sumZ1  +=       0.009 * cos(0.40604041041  + 24491.71669928959 *T)
    sumZ1  +=       0.011 * cos(1.88427152362  + 1581.95934828300 *T)
    sumZ1  +=       0.009 * cos(1.56488206261  + 129380.13377768178 *T)
    sumZ1  +=       0.009 * cos(3.73371803074  + 99799.65906923798 *T)
    sumZ1  +=       0.009 * cos(3.14695429609  + 2333.19639287200 *T)
    sumZ1  +=       0.010 * cos(1.21339698089  + 104881.30353139140 *T)
    sumZ1  +=       0.009 * cos(1.39539227982  + 77211.44104153418 *T)
    sumZ1  +=       0.008 * cos(4.87898553246  + 632.78373931320 *T)
    sumZ1  +=       0.009 * cos(5.97666796617  + 131548.89425996438 *T)
    sumZ1  +=       0.008 * cos(5.63267469085  + 7994.52844202420 *T)
    sumZ1  +=       0.008 * cos(0.82184344341  + 46848.33017476560 *T)
    sumZ1  +=       0.008 * cos(3.94765597932  + 25455.11940226100 *T)
    sumZ1  +=       0.010 * cos(5.48154668085  + 27140.17152476259 *T)
    sumZ1  +=       0.009 * cos(3.19254939333  + 76144.94556434419 *T)
    sumZ1  +=       0.007 * cos(4.71909828780  + 536.80451209540 *T)
    sumZ1  +=       0.008 * cos(5.52314941870  + 52602.40447402440 *T)
    sumZ1  +=       0.008 * cos(3.24434973969  + 51322.60990139639 *T)
    sumZ1  +=       0.010 * cos(6.10170513742  + 32370.97899156560 *T)
    sumZ1  +=       0.007 * cos(5.95089721308  + 108903.56606335558 *T)
    sumZ1  +=       0.008 * cos(0.41129689957  + 76044.95232053580 *T)
    sumZ1  +=       0.009 * cos(1.20083825078  + 64741.95798313119 *T)
    sumZ1  +=       0.007 * cos(4.12190282008  + 71492.99882347698 *T)
    sumZ1  +=       0.008 * cos(4.96879374914  + 105410.99449648600 *T)
    sumZ1  +=       0.008 * cos(4.56629366973  + 26037.90651966999 *T)
    sumZ1  +=       0.007 * cos(0.70887703358  + 24505.94379329119 *T)
    sumZ1  +=       0.007 * cos(4.46371843319  + 50579.61984086379 *T)
    sumZ1  +=       0.008 * cos(5.56132218876  + 62389.09182935939 *T)
    sumZ1  +=       0.007 * cos(3.12746642701  + 18093.37469954999 *T)
    sumZ1  +=       0.007 * cos(4.68816053110  + 52026.24308601380 *T)
    sumZ1  +=       0.007 * cos(0.17502761083  + 35191.81013569180 *T)
    sumZ1  +=       0.008 * cos(5.48340707498  + 25234.70675982219 *T)
    sumZ1  +=       0.007 * cos(4.34617041823  + 18207.81398823521 *T)
    sumZ1  +=       0.006 * cos(2.06324851601  + 78477.00852016058 *T)
    sumZ1  +=       0.006 * cos(3.48502574572  + 26094.53170047421 *T)
    sumZ1  +=       0.008 * cos(2.34450938553  + 33967.99229491319 *T)
    sumZ1  +=       0.006 * cos(0.49326410196  + 5327.47610838280 *T)
    sumZ1  +=       0.007 * cos(1.20744112899  + 44181.27784112419 *T)
    sumZ1  +=       0.006 * cos(4.73779414587  + 647.01083331480 *T)
    sumZ1  +=       0.007 * cos(0.69979076833  + 51756.32163927320 *T)
    sumZ1  +=       0.006 * cos(5.63847965133  + 339142.74084046454 *T)
    sumZ1  +=       0.008 * cos(3.68367545759  + 28206.66700195260 *T)
    sumZ1  +=       0.006 * cos(0.34516328945  + 26081.27458267419 *T)
    sumZ1  +=       0.006 * cos(2.90716455101  + 81706.28436968799 *T)
    sumZ1  +=       0.007 * cos(2.12491220048  + 88476.99497093359 *T)
    sumZ1  +=       0.005 * cos(3.34028647657  + 45494.58142974879 *T)
    sumZ1  +=       0.005 * cos(4.65551599269  + 93028.94846799241 *T)
    sumZ1  +=       0.005 * cos(5.37501413358  + 26610.48055966799 *T)
    sumZ1  +=       0.006 * cos(0.81299872270  + 95247.70557217918 *T)
    sumZ1  +=       0.005 * cos(3.47000694038  + 26507.38778544939 *T)
    sumZ1  +=       0.006 * cos(3.28193656342  + 26241.68195205900 *T)
    sumZ1  +=       0.006 * cos(4.64447087045  + 155468.03691925600 *T)
    sumZ1  +=       0.006 * cos(4.82010175617  + 94138.32702008578 *T)
    sumZ1  +=       0.006 * cos(1.32053124690  + 2648.45482547300 *T)
    sumZ1  +=       0.005 * cos(3.22819989009  + 25938.33994443960 *T)
    sumZ1  +=       0.005 * cos(2.61718935389  + 162188.75089859738 *T)
    sumZ1  +=       0.005 * cos(0.67832900959  + 104331.94280539700 *T)
    sumZ1  +=       0.004 * cos(5.21344084187  + 71025.03383312259 *T)
    sumZ1  +=       0.004 * cos(4.01714451988  + 78050.41032928458 *T)
    sumZ1  +=       0.004 * cos(5.08291206251  + 52099.54021187280 *T)
    sumZ1  +=       0.004 * cos(4.09094572411  + 104355.49390165480 *T)
    sumZ1  +=       0.005 * cos(4.64872616599  + 26190.99591579279 *T)
    sumZ1  +=       0.005 * cos(5.32437142333  + 25551.09862947879 *T)
    sumZ1  +=       0.004 * cos(2.72976343615  + 79852.78232000639 *T)
    sumZ1  +=       0.004 * cos(4.86893081526  + 50593.84693486539 *T)
    sumZ1  +=       0.004 * cos(0.65479452219  + 150866.08680029298 *T)
    sumZ1  +=       0.004 * cos(5.05003950789  + 7880.08915333899 *T)
    sumZ1  +=       0.005 * cos(0.56998430384  + 522.57741809380 *T)
    sumZ1  +=       0.004 * cos(2.18434126488  + 51639.00177105299 *T)
    sumZ1  +=       0.004 * cos(5.16982498302  + 149.56319713460 *T)
    sumZ1  +=       0.004 * cos(4.45668113275  + 52815.70356946240 *T)
    sumZ1  +=       0.004 * cos(0.69135494074  + 22759.76748529401 *T)
    sumZ1  +=       0.004 * cos(1.45087522595  + 20760.42703319140 *T)
    sumZ1  +=       0.003 * cos(2.24345480860  + 26521.61487945100 *T)
    sumZ1  +=       0.004 * cos(1.44955412166  + 120226.23016166000 *T)
    sumZ1  +=       0.004 * cos(0.16879575063  + 29416.03879785439 *T)
    sumZ1  +=       0.003 * cos(4.57090150345  + 103299.34418310841 *T)
    sumZ1  +=       0.003 * cos(4.59598906499  + 39743.76363275060 *T)
    sumZ1  +=       0.004 * cos(2.57869053238  + 42153.96900304900 *T)
    sumZ1  +=       0.004 * cos(0.10237777214  + 52290.24557183361 *T)
    sumZ1  +=       0.003 * cos(5.55031090013  + 3340.61242669980 *T)
    sumZ1  +=       0.004 * cos(6.17863510125  + 26555.86813192860 *T)
    sumZ1  +=       0.003 * cos(4.17269750233  + 130969.20667296558 *T)
    sumZ1  +=       0.003 * cos(4.68682814951  + 129909.82474277640 *T)
    sumZ1  +=       0.004 * cos(2.02800775775  + 53242.30176033840 *T)
    sumZ1  +=       0.003 * cos(5.35296159846  + 91785.46086631398 *T)
    sumZ1  +=       0.003 * cos(2.90685522022  + 104358.72611329758 *T)
    sumZ1  +=       0.003 * cos(3.88120838753  + 24864.08530079559 *T)
    sumZ1  +=       0.003 * cos(3.64751061524  + 26011.63707029860 *T)
    sumZ1  +=       0.003 * cos(3.23195009997  + 27311.72098235281 *T)
    sumZ1  +=       0.003 * cos(5.88557504342  + 104371.28232719658 *T)
    sumZ1  +=       0.004 * cos(3.12210700112  + 22747.29071487440 *T)
    sumZ1  +=       0.004 * cos(0.64204557090  + 105307.21230790540 *T)
    sumZ1  +=       0.003 * cos(6.24152399429  + 181505.94343892601 *T)
    sumZ1  +=       0.004 * cos(4.13065969197  + 125887.56221081220 *T)
    sumZ1  +=       0.004 * cos(4.45779901612  + 10021.83728009940 *T)
    sumZ1  +=       0.004 * cos(4.30265545797  + 90829.86112470538 *T)
    sumZ1  +=       0.003 * cos(2.25871507982  + 129483.91596626240 *T)
    sumZ1  +=       0.003 * cos(4.48812754789  + 77829.99768684579 *T)
    sumZ1  +=       0.003 * cos(1.95423245378  + 103821.92160120218 *T)
    sumZ1  +=       0.003 * cos(2.31574317697  + 53228.07466633679 *T)
    sumZ1  +=       0.003 * cos(5.16799762022  + 52022.02747266360 *T)
    sumZ1  +=       0.003 * cos(0.00175159073  + 54509.00267602040 *T)
    sumZ1  +=       0.004 * cos(1.27199373061  + 85502.38501632259 *T)
    sumZ1  +=       0.003 * cos(2.62911266804  + 100909.03762133140 *T)
    sumZ1  +=       0.003 * cos(5.08550718888  + 13362.44970679920 *T)
    sumZ1  +=       0.004 * cos(0.35429190140  + 26624.70765366959 *T)
    sumZ1  +=       0.003 * cos(5.11624187785  + 206.18554843720 *T)
    sumZ1  +=       0.003 * cos(1.69621650635  + 131498.89763806018 *T)
    sumZ1  +=       0.003 * cos(3.11541432650  + 106570.36967048360 *T)
    sumZ1  +=       0.003 * cos(3.95256477489  + 77844.22478084739 *T)
    sumZ1  +=       0.002 * cos(0.69658812162  + 6681.22485339960 *T)
    sumZ1  +=       0.003 * cos(3.36003785702  + 128850.44281258718 *T)
    sumZ1  +=       0.003 * cos(0.51920338716  + 26941.09952332620 *T)
    sumZ1  +=       0.003 * cos(6.07817701344  + 1478.86657406440 *T)
    sumZ1  +=       0.002 * cos(5.59272300692  + 27684.08958385880 *T)
    sumZ1  +=       0.003 * cos(0.55085259267  + 52061.36699446317 *T)
    sumZ1  +=       0.002 * cos(5.99908279167  + 98068.53671630540 *T)
    sumZ1  +=       0.002 * cos(4.99656134147  + 103285.11708910680 *T)
    sumZ1  +=       0.003 * cos(3.71061275134  + 90695.75207512038 *T)
    sumZ1  +=       0.002 * cos(5.86031214770  + 111122.32316754240 *T)
    sumZ1  +=       0.003 * cos(5.17758096862  + 114564.89811250778 *T)
    sumZ1  +=       0.003 * cos(2.34733115087  + 78690.30761559859 *T)
    sumZ1  +=       0.002 * cos(5.09808138852  + 76.26607127560 *T)
    sumZ1  +=       0.002 * cos(5.31294016066  + 104564.91166173480 *T)
    sumZ1  +=       0.002 * cos(0.98459946560  + 97580.90196505121 *T)
    sumZ1  +=       0.003 * cos(5.89794789328  + 49527.35145767539 *T)
    sumZ1  +=       0.003 * cos(3.28062974222  + 104344.49901929598 *T)
    sumZ1  +=       0.003 * cos(2.42068250003  + 103711.71527998280 *T)
    sumZ1  +=       0.002 * cos(1.05901421415  + 51543.02254383520 *T)
    sumZ1  +=       0.002 * cos(5.70093401914  + 117873.36400788820 *T)
    sumZ1  +=       0.002 * cos(3.31505268739  + 58458.88213313979 *T)
    sumZ1  +=       0.003 * cos(0.82349684851  + 151975.46535238638 *T)
    sumZ1  +=       0.002 * cos(1.52195733318  + 181555.94006083015 *T)
    sumZ1  +=       0.003 * cos(2.24270782677  + 25565.32572348040 *T)
    sumZ1  +=       0.003 * cos(5.23472870141  + 60055.89543648739 *T)
    sumZ1  +=       0.002 * cos(4.19499586662  + 956.28915597060 *T)
    sumZ1  +=       0.002 * cos(0.43182759927  + 1162.47470440780 *T)
    sumZ1  +=       0.002 * cos(1.55169232531  + 52329.58509363319 *T)
    sumZ1  +=       0.002 * cos(5.54133200346  + 188276.65404017159 *T)
    sumZ1  +=       0.002 * cos(1.63007261166  + 26294.08869001139 *T)
    sumZ1  +=       0.002 * cos(4.84459398809  + 70269.18098269838 *T)
    sumZ1  +=       0.002 * cos(2.45157941038  + 52698.38370124219 *T)
    sumZ1  +=       0.002 * cos(3.50351080444  + 102132.85546211000 *T)
    sumZ1  +=       0.002 * cos(1.28285332683  + 115674.27666460120 *T)
    sumZ1  +=       0.002 * cos(1.03461044928  + 31415.37924995700 *T)
    sumZ1  +=       0.002 * cos(1.31658394171  + 52125.80966124419 *T)
    sumZ1  +=       0.002 * cos(2.43230222606  + 103.09277421860 *T)
    sumZ1  +=       0.002 * cos(2.41277818747  + 3328.13565628019 *T)
    sumZ1  +=       0.002 * cos(3.38189136419  + 25881.71759313700 *T)
    sumZ1  +=       0.002 * cos(0.33148559099  + 54294.57014352679 *T)
    sumZ1  +=       0.002 * cos(0.15981568000  + 77410.51304297059 *T)
    sumZ1  +=       0.002 * cos(2.90582069750  + 52643.77127350280 *T)
    sumZ1  +=       0.002 * cos(0.45204563942  + 71980.63357473118 *T)
    sumZ1  +=       0.001 * cos(4.12045734648  + 25440.89230825939 *T)
    sumZ1  +=       0.001 * cos(1.18539094152  + 12546.48193908300 *T)
    sumZ1  +=       0.002 * cos(4.49448672471  + 146314.13330323418 *T)
    sumZ1  +=       0.001 * cos(3.99060758894  + 25867.49049913539 *T)
    sumZ1  +=       0.001 * cos(0.11291308321  + 105940.68546158058 *T)
    sumZ1  +=       0.001 * cos(6.21239484932  + 104138.31347085880 *T)
    sumZ1  +=       0.002 * cos(1.60021595888  + 51528.79544983359 *T)
    sumZ1  +=       0.002 * cos(2.03829619352  + 134991.46920492980 *T)
    sumZ1  +=       0.002 * cos(0.76189495741  + 84546.78527471398 *T)
    sumZ1  +=       0.001 * cos(1.19290119810  + 103925.01437542078 *T)
    sumZ1  +=       0.001 * cos(0.26092608641  + 155997.72788435058 *T)
    sumZ1  +=       0.001 * cos(0.91369257474  + 157636.79740153858 *T)
    sumZ1  +=       0.002 * cos(2.43141471440  + 143961.26714946239 *T)
    sumZ1  +=       0.001 * cos(2.58462364639  + 11610.55195837420 *T)
    sumZ1  +=       0.001 * cos(5.08144757931  + 32769.12799497380 *T)
    sumZ1  +=       0.001 * cos(2.90527234589  + 124156.43985787958 *T)
    sumZ1  +=       0.001 * cos(6.04124419781  + 16066.06586147480 *T)
    sumZ1  +=       0.001 * cos(0.82693925421  + 1692.16566950240 *T)
    sumZ1  +=       0.001 * cos(1.72510786190  + 26237.46633870879 *T)
    sumZ1  +=       0.001 * cos(0.74557141998  + 29428.51556827400 *T)
    sumZ1  +=       0.001 * cos(1.18363932898  + 130012.91751699500 *T)
    sumZ1  +=       0.001 * cos(5.53709511500  + 26720.68688088739 *T)
    sumZ1  +=       0.001 * cos(2.22421236230  + 25984.81036735560 *T)
    sumZ1  +=       0.001 * cos(4.51546243970  + 48835.19385644859 *T)
    sumZ1  +=       0.001 * cos(3.61219119527  + 131395.11544947961 *T)

    # MERCURY Z Coord
    sumZ2  = 0
    sumZ2  +=    4475.377 * cos(4.81703460547  + 26087.90314157420 *T)
    sumZ2  +=    1109.532 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ2  +=     180.327 * cos(1.19420644188  + 52175.80628314840 *T)
    sumZ2  +=      32.581 * cos(4.62632031743  + 78263.70942472259 *T)
    sumZ2  +=      17.342 * cos(1.72679813654  + 104351.61256629678 *T)
    sumZ2  +=       7.781 * cos(4.84181443647  + 130439.51570787100 *T)
    sumZ2  +=       2.976 * cos(1.62432394721  + 156527.41884944518 *T)
    sumZ2  +=       1.040 * cos(4.67846690150  + 182615.32199101936 *T)
    sumZ2  +=       0.680 * cos(1.83005840786  + 27197.28169366760 *T)
    sumZ2  +=       0.344 * cos(1.44603022303  + 208703.22513259359 *T)
    sumZ2  +=       0.110 * cos(4.49542230374  + 234791.12827416777 *T)
    sumZ2  +=       0.124 * cos(4.86657119280  + 53285.18483524180 *T)
    sumZ2  +=       0.112 * cos(5.14312161072  + 24978.52458948080 *T)
    sumZ2  +=       0.079 * cos(4.25092448424  + 31749.23519072640 *T)
    sumZ2  +=       0.058 * cos(2.00193276015  + 51066.42773105500 *T)
    sumZ2  +=       0.046 * cos(5.63734699624  + 27043.50288318280 *T)
    sumZ2  +=       0.054 * cos(5.70534232394  + 21535.94964451540 *T)
    sumZ2  +=       0.048 * cos(3.15710696395  + 51116.42435295920 *T)
    sumZ2  +=       0.034 * cos(1.26182479499  + 260879.03141574195 *T)
    sumZ2  +=       0.027 * cos(5.76991833962  + 529.69096509460 *T)
    sumZ2  +=       0.024 * cos(5.07488580193  + 77154.33087262919 *T)
    sumZ2  +=       0.024 * cos(0.95399123610  + 5661.33204915220 *T)
    sumZ2  +=       0.019 * cos(1.78858873244  + 28306.66024576099 *T)
    sumZ2  +=       0.021 * cos(3.57893007020  + 25558.21217647960 *T)
    sumZ2  +=       0.022 * cos(0.13783545204  + 25028.52121138500 *T)
    sumZ2  +=       0.018 * cos(5.52049109830  + 47623.85278608960 *T)
    sumZ2  +=       0.022 * cos(1.64692693544  + 20426.57109242200 *T)
    sumZ2  +=       0.014 * cos(5.59601085083  + 22645.32819660879 *T)
    sumZ2  +=       0.014 * cos(3.31820162449  + 1059.38193018920 *T)
    sumZ2  +=       0.018 * cos(0.54978030810  + 37410.56723987860 *T)
    sumZ2  +=       0.016 * cos(1.54285068472  + 79373.08797681599 *T)
    sumZ2  +=       0.015 * cos(4.49153081628  + 26107.57290247399 *T)
    sumZ2  +=       0.014 * cos(0.87791325175  + 57837.13833230060 *T)
    sumZ2  +=       0.013 * cos(5.18106920630  + 1109.37855209340 *T)
    sumZ2  +=       0.011 * cos(2.36351366554  + 53131.40602475700 *T)
    sumZ2  +=       0.012 * cos(4.34261596283  + 32858.61374281979 *T)
    sumZ2  +=       0.015 * cos(2.28255722706  + 50586.73338786459 *T)
    sumZ2  +=       0.012 * cos(5.63353087898  + 26068.23338067440 *T)
    sumZ2  +=       0.013 * cos(2.08539778974  + 25132.30339996560 *T)
    sumZ2  +=       0.013 * cos(3.31771880069  + 10213.28554621100 *T)
    sumZ2  +=       0.013 * cos(0.37928151736  + 4551.95349705880 *T)
    sumZ2  +=       0.009 * cos(5.32896907669  + 23869.14603738740 *T)
    sumZ2  +=       0.009 * cos(1.85195473467  + 103242.23401420339 *T)
    sumZ2  +=       0.010 * cos(2.69290690744  + 26091.78447693220 *T)
    sumZ2  +=       0.012 * cos(1.13570211007  + 26084.02180621620 *T)
    sumZ2  +=       0.010 * cos(4.31085447586  + 286966.93455731618 *T)
    sumZ2  +=       0.008 * cos(5.82534551009  + 25874.60404613620 *T)
    sumZ2  +=       0.008 * cos(6.17561594460  + 51646.11531805379 *T)
    sumZ2  +=       0.008 * cos(2.73528822352  + 41962.52073693740 *T)
    sumZ2  +=       0.008 * cos(0.86132391139  + 39609.65458316560 *T)
    sumZ2  +=       0.007 * cos(5.90586936446  + 77204.32749453338 *T)
    sumZ2  +=       0.006 * cos(2.23725426587  + 26080.78959457339 *T)
    sumZ2  +=       0.007 * cos(5.14284397401  + 1589.07289528380 *T)
    sumZ2  +=       0.007 * cos(3.13934336281  + 43071.89928903080 *T)
    sumZ2  +=       0.005 * cos(4.86318139233  + 54394.56338733519 *T)
    sumZ2  +=       0.007 * cos(2.45439783289  + 213.29909543800 *T)
    sumZ2  +=       0.006 * cos(1.58442144371  + 26095.01668857500 *T)
    sumZ2  +=       0.006 * cos(5.45060244038  + 25973.46385288896 *T)
    sumZ2  +=       0.005 * cos(4.79894457233  + 46514.47423399620 *T)
    sumZ2  +=       0.005 * cos(2.20020439243  + 6283.07584999140 *T)
    sumZ2  +=       0.005 * cos(5.38203064600  + 15874.61759536320 *T)
    sumZ2  +=       0.005 * cos(2.75432060530  + 19317.19254032860 *T)
    sumZ2  +=       0.004 * cos(2.96762251508  + 33326.57873317420 *T)
    sumZ2  +=       0.005 * cos(3.58879974911  + 63498.47038145279 *T)
    sumZ2  +=       0.004 * cos(6.00635657777  + 7.11354700080 *T)
    sumZ2  +=       0.004 * cos(3.43811673157  + 1066.49547719000 *T)
    sumZ2  +=       0.004 * cos(3.46360896326  + 25661.30495069820 *T)
    sumZ2  +=       0.004 * cos(4.83154673482  + 51220.20654153979 *T)
    sumZ2  +=       0.005 * cos(1.65155678587  + 26202.34243025941 *T)
    sumZ2  +=       0.004 * cos(4.05043524436  + 51535.90899683439 *T)
    sumZ2  +=       0.004 * cos(3.87269622429  + 1052.26838318840 *T)
    sumZ2  +=       0.004 * cos(1.10462215070  + 58946.51688439399 *T)
    sumZ2  +=       0.004 * cos(5.41502198784  + 14765.23904326980 *T)
    sumZ2  +=       0.004 * cos(2.03873056084  + 26617.59410666880 *T)
    sumZ2  +=       0.003 * cos(2.19389680754  + 30639.85663863300 *T)
    sumZ2  +=       0.003 * cos(4.90884264121  + 129330.13715577760 *T)
    sumZ2  +=       0.004 * cos(6.22828785556  + 36301.18868778519 *T)
    sumZ2  +=       0.004 * cos(3.68258570333  + 13521.75144159140 *T)
    sumZ2  +=       0.003 * cos(5.62688231105  + 48733.23133818299 *T)
    sumZ2  +=       0.004 * cos(2.45849991842  + 51749.20809227239 *T)
    sumZ2  +=       0.003 * cos(3.06502787272  + 16983.99614745660 *T)
    sumZ2  +=       0.003 * cos(4.57742754340  + 24498.83024629040 *T)
    sumZ2  +=       0.003 * cos(1.46286237530  + 3442.57494496540 *T)
    sumZ2  +=       0.003 * cos(1.55999070785  + 29530.47808653960 *T)
    sumZ2  +=       0.003 * cos(5.14417203650  + 76674.63652943878 *T)
    sumZ2  +=       0.003 * cos(1.09829861469  + 313054.83769889042 *T)
    sumZ2  +=       0.003 * cos(2.10570480254  + 49957.04917896160 *T)
    sumZ2  +=       0.002 * cos(0.49635100411  + 38519.94579197200 *T)
    sumZ2  +=       0.003 * cos(3.19744325207  + 426.59819087600 *T)
    sumZ2  +=       0.003 * cos(2.27361066514  + 72602.37737557039 *T)
    sumZ2  +=       0.002 * cos(5.42166315666  + 12566.15169998280 *T)
    sumZ2  +=       0.002 * cos(4.11021645418  + 2218.75710418680 *T)
    sumZ2  +=       0.002 * cos(6.01977464255  + 2118.76386037840 *T)
    sumZ2  +=       0.003 * cos(3.82490235305  + 9103.90699411760 *T)
    sumZ2  +=       0.002 * cos(1.55999258470  + 50057.04242277000 *T)
    sumZ2  +=       0.002 * cos(2.83741388675  + 639.89728631400 *T)
    sumZ2  +=       0.003 * cos(5.50522133834  + 23754.70674870219 *T)
    sumZ2  +=       0.002 * cos(4.64349469451  + 56727.75978020720 *T)
    sumZ2  +=       0.002 * cos(1.95986167578  + 73711.75592766379 *T)
    sumZ2  +=       0.002 * cos(0.55672894560  + 955.59974160860 *T)
    sumZ2  +=       0.002 * cos(1.93244538678  + 103292.23063610760 *T)
    sumZ2  +=       0.002 * cos(6.05929958767  + 25448.00585526019 *T)
    sumZ2  +=       0.002 * cos(3.03995453086  + 40853.14218484400 *T)
    sumZ2  +=       0.001 * cos(2.47665559131  + 77734.01845962799 *T)
    sumZ2  +=       0.002 * cos(6.25822253418  + 52705.49724824299 *T)
    sumZ2  +=       0.002 * cos(1.38627663309  + 77308.10968311399 *T)
    sumZ2  +=       0.001 * cos(1.59643416291  + 27147.28507176339 *T)
    sumZ2  +=       0.002 * cos(3.54992576302  + 83925.04147387479 *T)
    sumZ2  +=       0.001 * cos(5.82473062390  + 45405.09568190280 *T)
    sumZ2  +=       0.001 * cos(0.32361625099  + 13655.86049117640 *T)
    sumZ2  +=       0.001 * cos(4.59226455833  + 55618.38122811380 *T)
    sumZ2  +=       0.001 * cos(1.77801233677  + 52156.13652224860 *T)
    sumZ2  +=       0.001 * cos(3.59899238770  + 53235.18821333759 *T)
    sumZ2  +=       0.001 * cos(0.75186170269  + 52195.47604404819 *T)
    sumZ2  +=       0.002 * cos(4.21419422689  + 105460.99111839020 *T)
    sumZ2  +=       0.001 * cos(3.02724177299  + 51962.50718771040 *T)
    sumZ2  +=       0.001 * cos(5.54016425013  + 98690.28051714458 *T)
    sumZ2  +=       0.001 * cos(4.96163083687  + 18849.22754997420 *T)
    sumZ2  +=       0.001 * cos(5.13453198510  + 52179.68761850640 *T)
    sumZ2  +=       0.001 * cos(4.54050452164  + 52168.69273614759 *T)
    sumZ2  +=       0.001 * cos(3.67122666149  + 52171.92494779040 *T)
    sumZ2  +=       0.001 * cos(4.16002499033  + 45892.73043315699 *T)
    sumZ2  +=       0.001 * cos(3.11767834677  + 78793.40038981718 *T)
    sumZ2  +=       0.001 * cos(4.70607503583  + 52290.24557183361 *T)
    sumZ2  +=       0.001 * cos(1.72196789636  + 155418.04029735178 *T)
    sumZ2  +=       0.001 * cos(1.43678225289  + 82815.66292178139 *T)
    sumZ2  +=       0.001 * cos(3.64195400420  + 65697.55772473979 *T)
    sumZ2  +=       0.001 * cos(2.34232573897  + 27676.97603685800 *T)
    sumZ2  +=       0.001 * cos(2.22519315150  + 52061.36699446317 *T)
    sumZ2  +=       0.001 * cos(4.06167314974  + 339142.74084046454 *T)
    sumZ2  +=       0.001 * cos(1.84893906586  + 853.19638175200 *T)
    sumZ2  +=       0.001 * cos(5.42775732068  + 26514.50133245020 *T)
    sumZ2  +=       0.001 * cos(6.03028910593  + 59414.48187474840 *T)
    sumZ2  +=       0.001 * cos(5.09856883029  + 76044.95232053580 *T)
    sumZ2  +=       0.001 * cos(1.68523311659  + 28421.09953444620 *T)
    sumZ2  +=       0.001 * cos(5.32883383904  + 79219.30916633119 *T)
    sumZ2  +=       0.001 * cos(0.84570925458  + 19804.82729158280 *T)
    sumZ2  +=       0.001 * cos(1.06906655760  + 124778.18365871880 *T)
    sumZ2  +=       0.001 * cos(5.63868454693  + 69159.80243060499 *T)
    sumZ2  +=       0.001 * cos(5.38153004139  + 26301.20223701220 *T)
    sumZ2  +=       0.001 * cos(4.77073547131  + 104881.30353139140 *T)
    sumZ2  +=       0.001 * cos(4.74588142049  + 129380.13377768178 *T)
    sumZ2  +=       0.001 * cos(2.25545108286  + 52602.40447402440 *T)
    sumZ2  +=       0.001 * cos(3.98601687793  + 11322.66409830440 *T)
    sumZ2  +=       0.001 * cos(0.07845866028  + 66941.04532641819 *T)
    sumZ2  +=       0.001 * cos(2.88984683172  + 38654.05484155699 *T)
    sumZ2  +=       0.001 * cos(4.00503893845  + 52182.91983014920 *T)
    sumZ2  +=       0.001 * cos(3.49574656538  + 52389.10537858640 *T)
    sumZ2  +=       0.001 * cos(0.33429045009  + 7238.67559160000 *T)
    sumZ2  +=       0.001 * cos(0.31985247795  + 79323.09135491178 *T)
    sumZ2  +=       0.001 * cos(6.10860920195  + 51109.31080595839 *T)
    sumZ2  +=       0.001 * cos(2.84676327671  + 7994.52844202420 *T)
    sumZ2  +=       0.001 * cos(3.11496551907  + 25934.12433108940 *T)
    sumZ2  +=       0.001 * cos(5.40172380413  + 78477.00852016058 *T)
    sumZ2  +=       0.001 * cos(0.41949488018  + 6770.71060124560 *T)
    sumZ2  +=       0.001 * cos(4.62233743377  + 64741.95798313119 *T)
    sumZ2  +=       0.001 * cos(4.27662369360  + 12432.04265039780 *T)
    sumZ2  +=       0.001 * cos(0.17150693291  + 433.71173787680 *T)
    sumZ2  +=       0.001 * cos(5.16942669114  + 77837.11123384659 *T)

    # MERCURY Z Coord
    sumZ3  = 0
    sumZ3  +=      99.250 * cos(0.29443229590  + 26087.90314157420 *T)
    sumZ3  +=      92.976 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ3  +=       3.230 * cos(4.02309993038  + 52175.80628314840 *T)
    sumZ3  +=       0.640 * cos(2.16717843388  + 78263.70942472259 *T)
    sumZ3  +=       0.369 * cos(6.09477192866  + 104351.61256629678 *T)
    sumZ3  +=       0.224 * cos(3.14630379889  + 130439.51570787100 *T)
    sumZ3  +=       0.111 * cos(0.00722359195  + 156527.41884944518 *T)
    sumZ3  +=       0.048 * cos(3.09594836549  + 182615.32199101936 *T)
    sumZ3  +=       0.019 * cos(6.16655603798  + 208703.22513259359 *T)
    sumZ3  +=       0.021 * cos(3.31046636942  + 27197.28169366760 *T)
    sumZ3  +=       0.007 * cos(2.94356012175  + 234791.12827416777 *T)
    sumZ3  +=       0.007 * cos(0.65628834083  + 24978.52458948080 *T)
    sumZ3  +=       0.002 * cos(5.99334261589  + 260879.03141574195 *T)
    sumZ3  +=       0.002 * cos(6.22873921629  + 53285.18483524180 *T)
    sumZ3  +=       0.002 * cos(5.78810325925  + 31749.23519072640 *T)
    sumZ3  +=       0.001 * cos(0.84479650691  + 27043.50288318280 *T)
    sumZ3  +=       0.001 * cos(1.53587211633  + 1109.37855209340 *T)
    sumZ3  +=       0.001 * cos(0.78761433232  + 21535.94964451540 *T)
    sumZ3  +=       0.001 * cos(4.53925772521  + 51116.42435295920 *T)
    sumZ3  +=       0.001 * cos(2.76184993272  + 286966.93455731618 *T)
    sumZ3  +=       0.001 * cos(4.91700706367  + 25558.21217647960 *T)
    sumZ3  +=       0.001 * cos(5.91899219772  + 1059.38193018920 *T)
    sumZ3  +=       0.001 * cos(3.91598529129  + 25132.30339996560 *T)
    sumZ3  +=       0.001 * cos(4.83279660629  + 20426.57109242200 *T)

    # MERCURY Z Coord
    sumZ4  = 0
    sumZ4  +=       1.500 * cos(1.82376153122  + 26087.90314157420 *T)
    sumZ4  +=       0.592 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ4  +=       0.056 * cos(3.48599666355  + 52175.80628314840 *T)
    sumZ4  +=       0.022 * cos(0.17608711147  + 78263.70942472259 *T)
    sumZ4  +=       0.009 * cos(3.89825290813  + 104351.61256629678 *T)
    sumZ4  +=       0.005 * cos(1.27729295657  + 130439.51570787100 *T)
    sumZ4  +=       0.003 * cos(4.59766720323  + 156527.41884944518 *T)
    sumZ4  +=       0.002 * cos(1.46273885253  + 182615.32199101936 *T)
    sumZ4  +=       0.001 * cos(4.58832110028  + 208703.22513259359 *T)

    # MERCURY Z Coord
    sumZ5  = 0
    sumZ5  +=       0.045 * cos(3.82268203713  + 26087.90314157420 *T)
    sumZ5  +=       0.043 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ5  +=       0.003 * cos(1.05444326177  + 52175.80628314840 *T)
    sumZ5  +=       0.001 * cos(4.50060145511  + 78263.70942472259 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Mercury",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getMercuryData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getMercuryXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getNeptuneXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Neptune"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getNeptuneXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getNeptuneXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # NEPTUNE X Coord
    sumX0  = 0
    sumX0  +=  3005973100.580 * cos(5.31188633083  + 38.37685312130 *T)
    sumX0  +=  40567587.218 * cos(3.98149970131  + 0.24381748350 *T)
    sumX0  +=  13506026.414 * cos(3.50055820972  + 76.50988875911 *T)
    sumX0  +=  15716341.901 * cos(0.11310077968  + 36.89238041300 *T)
    sumX0  +=  14935642.614 * cos(1.08477702063  + 39.86132582961 *T)
    sumX0  +=  2590782.232 * cos(1.99609768221  + 1.72829019180 *T)
    sumX0  +=  1073890.204 * cos(5.38477153556  + 75.02541605080 *T)
    sumX0  +=  816388.197 * cos(0.78185518038  + 3.21276290011 *T)
    sumX0  +=  702768.075 * cos(1.45363642119  + 35.40790770471 *T)
    sumX0  +=  687594.822 * cos(0.72075739344  + 37.88921815429 *T)
    sumX0  +=  565555.652 * cos(5.98943773879  + 41.34579853790 *T)
    sumX0  +=  495650.075 * cos(0.59957534348  + 529.93478257810 *T)
    sumX0  +=  306025.380 * cos(0.39916788140  + 73.54094334250 *T)
    sumX0  +=  272446.904 * cos(0.87404115637  + 213.54291292150 *T)
    sumX0  +=  135892.298 * cos(5.54654979922  + 77.99436146740 *T)
    sumX0  +=  122117.697 * cos(1.30863876781  + 34.92027273770 *T)
    sumX0  +=   90968.285 * cos(1.68886748674  + 114.64292439690 *T)
    sumX0  +=   68915.400 * cos(5.83470374400  + 4.69723560840 *T)
    sumX0  +=   40370.680 * cos(2.66129691063  + 33.92343499640 *T)
    sumX0  +=   28891.307 * cos(4.78947715515  + 42.83027124621 *T)
    sumX0  +=   29247.752 * cos(1.62319522731  + 72.05647063421 *T)
    sumX0  +=   25576.289 * cos(1.48342967006  + 71.56883566720 *T)
    sumX0  +=   20517.968 * cos(2.55621077117  + 33.43580002939 *T)
    sumX0  +=   12614.154 * cos(3.56929744338  + 113.15845168861 *T)
    sumX0  +=   12788.929 * cos(2.73769634046  + 111.67397898031 *T)
    sumX0  +=   12013.477 * cos(0.94915799508  + 1059.62574767270 *T)
    sumX0  +=    9854.638 * cos(0.25713641240  + 36.40474544600 *T)
    sumX0  +=    8385.825 * cos(1.65242210861  + 108.21739859670 *T)
    sumX0  +=    7577.585 * cos(0.09970777629  + 426.84200835950 *T)
    sumX0  +=    6452.053 * cos(4.62556526073  + 6.18170831670 *T)
    sumX0  +=    6551.074 * cos(1.91884050790  + 1.24065522479 *T)
    sumX0  +=    4652.534 * cos(0.10344003066  + 37.85558825950 *T)
    sumX0  +=    4732.958 * cos(4.09711900918  + 79.47883417571 *T)
    sumX0  +=    4557.247 * cos(1.09712661798  + 38.89811798311 *T)
    sumX0  +=    4322.550 * cos(2.37744779374  + 38.32866901151 *T)
    sumX0  +=    4315.539 * cos(5.10473140788  + 38.42503723110 *T)
    sumX0  +=    4089.036 * cos(1.99429063701  + 37.41364527480 *T)
    sumX0  +=    4248.658 * cos(5.63379709294  + 28.81562556571 *T)
    sumX0  +=    4622.142 * cos(2.73995451568  + 70.08436295889 *T)
    sumX0  +=    3926.447 * cos(5.48975060892  + 39.34006096781 *T)
    sumX0  +=    3148.422 * cos(5.18755364576  + 76.02225379210 *T)
    sumX0  +=    3940.981 * cos(2.29766376691  + 98.65617104110 *T)
    sumX0  +=    3323.363 * cos(4.68776245279  + 4.43660317750 *T)
    sumX0  +=    3282.964 * cos(2.81551282614  + 39.37369086260 *T)
    sumX0  +=    3110.464 * cos(1.84416897204  + 47.93808067690 *T)
    sumX0  +=    2927.062 * cos(2.83767313961  + 70.57199792590 *T)
    sumX0  +=    2748.919 * cos(3.86990252936  + 32.43896228810 *T)
    sumX0  +=    3316.668 * cos(1.82194084200  + 144.86596152620 *T)
    sumX0  +=    2822.405 * cos(3.78131048254  + 31.95132732110 *T)
    sumX0  +=    2695.972 * cos(3.85276301548  + 110.18950627200 *T)
    sumX0  +=    2522.990 * cos(4.66308619966  + 311.95526647910 *T)
    sumX0  +=    1888.129 * cos(3.20464683230  + 35.92917256650 *T)
    sumX0  +=    1648.229 * cos(4.07040254381  + 30.30009827400 *T)
    sumX0  +=    1826.545 * cos(3.58021128918  + 44.31474395451 *T)
    sumX0  +=    1956.241 * cos(4.14516146871  + 206.42936592071 *T)
    sumX0  +=    1681.257 * cos(4.27560127770  + 40.82453367610 *T)
    sumX0  +=    1533.383 * cos(1.17732213608  + 38.26497853671 *T)
    sumX0  +=    1893.076 * cos(0.75017402977  + 220.65645992230 *T)
    sumX0  +=    1527.526 * cos(0.02173638301  + 38.48872770590 *T)
    sumX0  +=    2085.691 * cos(1.56948272604  + 149.80701461810 *T)
    sumX0  +=    2070.612 * cos(2.82581806721  + 136.78920667889 *T)
    sumX0  +=    1535.699 * cos(0.61413315675  + 73.05330837550 *T)
    sumX0  +=    1667.976 * cos(2.91712458990  + 106.73292588839 *T)
    sumX0  +=    1289.620 * cos(3.39708861100  + 46.45360796860 *T)
    sumX0  +=    1559.811 * cos(0.55870841967  + 38.11622069041 *T)
    sumX0  +=    1545.705 * cos(0.64028776037  + 38.63748555220 *T)
    sumX0  +=    1435.033 * cos(0.72855949679  + 522.82123557730 *T)
    sumX0  +=    1406.206 * cos(3.61717027558  + 537.04832957890 *T)
    sumX0  +=    1256.446 * cos(2.70907758736  + 34.18406742730 *T)
    sumX0  +=    1387.973 * cos(3.71843398082  + 116.12739710521 *T)
    sumX0  +=    1457.739 * cos(1.98981635014  + 181.51452445570 *T)
    sumX0  +=    1228.429 * cos(2.78646343835  + 72.31710306511 *T)
    sumX0  +=    1140.665 * cos(3.96643713353  + 7.83293736379 *T)
    sumX0  +=    1080.801 * cos(4.75483465055  + 42.56963881530 *T)
    sumX0  +=    1201.409 * cos(0.74547986507  + 2.72512793310 *T)
    sumX0  +=    1228.671 * cos(2.65249731727  + 148.32254190981 *T)
    sumX0  +=     722.014 * cos(6.16806714444  + 152.77596003471 *T)
    sumX0  +=     608.545 * cos(4.49536985567  + 35.45609181450 *T)
    sumX0  +=     722.865 * cos(3.09340262825  + 143.38148881789 *T)
    sumX0  +=     632.820 * cos(3.41702130042  + 7.66618102501 *T)
    sumX0  +=     642.369 * cos(3.97490787694  + 68.59989025060 *T)
    sumX0  +=     553.789 * cos(2.98606728111  + 41.29761442810 *T)
    sumX0  +=     682.276 * cos(2.15806346682  + 218.16308738520 *T)
    sumX0  +=     463.186 * cos(2.74420554348  + 31.78457098230 *T)
    sumX0  +=     521.560 * cos(0.34813640632  + 0.71939036300 *T)
    sumX0  +=     437.892 * cos(1.29807722623  + 1589.31671276730 *T)
    sumX0  +=     398.091 * cos(5.50783691510  + 6.34846465550 *T)
    sumX0  +=     384.065 * cos(4.72632236146  + 44.96913526031 *T)
    sumX0  +=     395.583 * cos(5.05527677390  + 108.70503356371 *T)
    sumX0  +=     327.446 * cos(2.69199709491  + 60.52313540329 *T)
    sumX0  +=     358.824 * cos(4.99912098256  + 30.46685461280 *T)
    sumX0  +=     315.179 * cos(0.17468500209  + 74.53778108379 *T)
    sumX0  +=     343.384 * cos(1.74645896957  + 0.76508234530 *T)
    sumX0  +=     399.611 * cos(5.33540800911  + 31.26330612050 *T)
    sumX0  +=     314.611 * cos(2.98803024638  + 419.72846135871 *T)
    sumX0  +=     347.596 * cos(3.26643963659  + 180.03005174739 *T)
    sumX0  +=     382.279 * cos(0.21764578681  + 487.12132627930 *T)
    sumX0  +=     300.918 * cos(4.04922612099  + 69.08752521760 *T)
    sumX0  +=     340.448 * cos(3.90546849629  + 146.83806920150 *T)
    sumX0  +=     298.710 * cos(5.18013539651  + 84.58664360640 *T)
    sumX0  +=     290.629 * cos(1.74873675275  + 110.45013870291 *T)
    sumX0  +=     336.211 * cos(2.14815098729  + 45.49040012211 *T)
    sumX0  +=     305.606 * cos(5.63265481978  + 640.14110379750 *T)
    sumX0  +=     333.702 * cos(2.32938316969  + 254.81165031470 *T)
    sumX0  +=     268.060 * cos(3.30852201658  + 37.00425499760 *T)
    sumX0  +=     264.760 * cos(4.12724058864  + 39.74945124500 *T)
    sumX0  +=     315.240 * cos(2.72241788492  + 388.70897272171 *T)
    sumX0  +=     227.098 * cos(4.59157281152  + 273.82223084130 *T)
    sumX0  +=     306.112 * cos(1.75345186469  + 6283.31966747490 *T)
    sumX0  +=     284.373 * cos(3.36139825385  + 12.77399045571 *T)
    sumX0  +=     221.105 * cos(3.50940363876  + 213.05527795450 *T)
    sumX0  +=     242.568 * cos(2.06437650010  + 14.25846316400 *T)
    sumX0  +=     241.087 * cos(4.16115355874  + 105.24845318010 *T)
    sumX0  +=     226.136 * cos(2.83815938853  + 80.96330688400 *T)
    sumX0  +=     245.904 * cos(0.54462524204  + 27.33115285740 *T)
    sumX0  +=     265.825 * cos(4.10952660358  + 944.73900579230 *T)
    sumX0  +=     207.893 * cos(5.07812851336  + 30.95448957981 *T)
    sumX0  +=     214.661 * cos(2.65402494691  + 316.63568714010 *T)
    sumX0  +=     190.638 * cos(2.32667703756  + 69.39634175830 *T)
    sumX0  +=     246.295 * cos(1.98638488517  + 102.84895673509 *T)
    sumX0  +=     202.915 * cos(0.60029260077  + 415.04804069769 *T)
    sumX0  +=     176.465 * cos(0.14731292877  + 36.78050582840 *T)
    sumX0  +=     193.886 * cos(3.35476299352  + 174.92224231670 *T)
    sumX0  +=     175.209 * cos(1.12575693515  + 39.97320041421 *T)
    sumX0  +=     177.868 * cos(3.43923391414  + 216.67861467689 *T)
    sumX0  +=     138.494 * cos(5.45265920432  + 75.98862389731 *T)
    sumX0  +=     152.234 * cos(4.81662104772  + 11.28951774740 *T)
    sumX0  +=     147.648 * cos(1.68543706672  + 151.29148732640 *T)
    sumX0  +=     156.202 * cos(3.65252575052  + 146.35043423450 *T)
    sumX0  +=     152.289 * cos(0.07345728764  + 23.87457247379 *T)
    sumX0  +=     177.911 * cos(3.17643554721  + 10213.52936369450 *T)
    sumX0  +=     162.474 * cos(4.13351391379  + 63.97971578690 *T)
    sumX0  +=     121.226 * cos(5.10584286197  + 38.16440480021 *T)
    sumX0  +=     129.049 * cos(3.80684906955  + 37.10482873410 *T)
    sumX0  +=     120.334 * cos(2.37637214462  + 38.58930144240 *T)
    sumX0  +=     168.977 * cos(2.49551838497  + 291.46021324420 *T)
    sumX0  +=     121.138 * cos(1.49657109299  + 33.26904369061 *T)
    sumX0  +=     129.366 * cos(2.36903010922  + 45.79921666280 *T)
    sumX0  +=     144.682 * cos(0.63023431786  + 49.42255338521 *T)
    sumX0  +=     122.915 * cos(3.67433526761  + 39.64887750850 *T)
    sumX0  +=     113.400 * cos(0.42160185021  + 83.10217089810 *T)
    sumX0  +=     154.892 * cos(1.74989069653  + 77.47309660560 *T)
    sumX0  +=     106.737 * cos(0.57437068095  + 4.86399194720 *T)
    sumX0  +=     104.756 * cos(5.96272070512  + 43.48466255200 *T)
    sumX0  +=     125.142 * cos(5.82780261996  + 4.20960064140 *T)
    sumX0  +=     103.541 * cos(5.25634741505  + 41.08516610701 *T)
    sumX0  +=     133.573 * cos(3.92147215781  + 182.99899716400 *T)
    sumX0  +=     103.627 * cos(2.29256111601  + 35.66854013560 *T)
    sumX0  +=     116.874 * cos(5.41378396536  + 62.49524307860 *T)
    sumX0  +=      98.063 * cos(3.25654027665  + 9.80504503910 *T)
    sumX0  +=     111.411 * cos(4.34345309647  + 141.89701610960 *T)
    sumX0  +=     114.294 * cos(5.56228935636  + 633.02755679670 *T)
    sumX0  +=     104.705 * cos(6.26072139356  + 433.95555536030 *T)
    sumX0  +=     121.306 * cos(1.44892345337  + 40.85816357090 *T)
    sumX0  +=      96.954 * cos(6.17373469303  + 1052.51220067191 *T)
    sumX0  +=      85.104 * cos(4.79018222360  + 36.67993209190 *T)
    sumX0  +=      85.209 * cos(5.94497188324  + 105.76971804189 *T)
    sumX0  +=      85.291 * cos(2.59495207397  + 109.70187130500 *T)
    sumX0  +=      83.260 * cos(0.00625676877  + 529.44714761109 *T)
    sumX0  +=      80.200 * cos(2.69199769694  + 40.07377415071 *T)
    sumX0  +=     107.927 * cos(0.01570870082  + 1162.71852189130 *T)
    sumX0  +=      95.241 * cos(3.61102256601  + 253.32717760639 *T)
    sumX0  +=      89.535 * cos(3.25178384851  + 32.96022714990 *T)
    sumX0  +=      89.793 * cos(2.76430560225  + 65.46418849521 *T)
    sumX0  +=      72.027 * cos(0.11366576076  + 36.94056452280 *T)
    sumX0  +=      80.381 * cos(5.21057317852  + 67.11541754230 *T)
    sumX0  +=      99.502 * cos(2.53010647936  + 453.18107633550 *T)
    sumX0  +=      88.685 * cos(1.33848394125  + 251.67594855930 *T)
    sumX0  +=      94.971 * cos(4.11602347578  + 219.64756009350 *T)
    sumX0  +=      77.015 * cos(5.30660266172  + 5.66044345490 *T)
    sumX0  +=      69.098 * cos(1.84984192453  + 22.39009976550 *T)
    sumX0  +=      79.079 * cos(4.12824954018  + 44.48150029329 *T)
    sumX0  +=      69.159 * cos(3.95901333551  + 1066.73929467350 *T)
    sumX0  +=      64.446 * cos(4.03076164648  + 66.94866120350 *T)
    sumX0  +=      88.518 * cos(2.66179796694  + 328.10877617370 *T)
    sumX0  +=      65.817 * cos(1.42821476263  + 36.37111555120 *T)
    sumX0  +=      71.422 * cos(4.23104971231  + 43.79347909271 *T)
    sumX0  +=      63.298 * cos(2.21146718451  + 9.15065373330 *T)
    sumX0  +=      77.320 * cos(0.26842720811  + 97.17169833279 *T)
    sumX0  +=      73.912 * cos(1.72397638430  + 2.67694382330 *T)
    sumX0  +=      73.965 * cos(5.55809543248  + 2.95213046920 *T)
    sumX0  +=      56.194 * cos(4.45857439361  + 949.41942645330 *T)
    sumX0  +=      59.173 * cos(1.41372632607  + 100.14064374939 *T)
    sumX0  +=      67.507 * cos(3.94700376513  + 7.14491616321 *T)
    sumX0  +=      71.718 * cos(0.93392607299  + 2.20386307129 *T)
    sumX0  +=      63.606 * cos(5.17175542607  + 25.84668014910 *T)
    sumX0  +=      71.523 * cos(2.05830478088  + 662.28738607949 *T)
    sumX0  +=      57.219 * cos(0.88485013288  + 15.74293587230 *T)
    sumX0  +=      50.322 * cos(1.08310288762  + 37.15301284391 *T)
    sumX0  +=      66.615 * cos(3.42462264660  + 846.32665223470 *T)
    sumX0  +=      56.220 * cos(4.52386924168  + 178.54557903910 *T)
    sumX0  +=      67.883 * cos(3.88546727303  + 224.58861318540 *T)
    sumX0  +=      57.761 * cos(5.16493680948  + 145.35359649321 *T)
    sumX0  +=      53.973 * cos(6.25404762289  + 107.22056085540 *T)
    sumX0  +=      57.588 * cos(4.84839311245  + 25.35904518210 *T)
    sumX0  +=      49.026 * cos(1.27836371915  + 19.25439801010 *T)
    sumX0  +=      63.036 * cos(4.29760573349  + 256.29612302300 *T)
    sumX0  +=      45.304 * cos(0.86492921312  + 4.17597074660 *T)
    sumX0  +=      45.669 * cos(2.17547535945  + 117.61186981350 *T)
    sumX0  +=      52.821 * cos(3.77933473571  + 289.97574053589 *T)
    sumX0  +=      44.016 * cos(2.25498623278  + 32.74777882880 *T)
    sumX0  +=      42.933 * cos(6.21504221321  + 28.98238190449 *T)
    sumX0  +=      38.369 * cos(0.36602717013  + 39.60069339870 *T)
    sumX0  +=      38.805 * cos(4.12403932769  + 103.33659170210 *T)
    sumX0  +=      37.679 * cos(3.40097359574  + 9.31741007210 *T)
    sumX0  +=      40.292 * cos(6.03933270535  + 111.18634401329 *T)
    sumX0  +=      50.011 * cos(6.19966711969  + 221.61966776881 *T)
    sumX0  +=      37.056 * cos(4.63008749202  + 8.32057233081 *T)
    sumX0  +=      36.562 * cos(0.18548635975  + 448.98829064149 *T)
    sumX0  +=      44.628 * cos(3.82762130859  + 525.25436191710 *T)
    sumX0  +=      38.213 * cos(0.28030378709  + 75.54668091261 *T)
    sumX0  +=      45.963 * cos(4.06403723861  + 183.48663213100 *T)
    sumX0  +=      48.222 * cos(2.81328685847  + 364.75733910320 *T)
    sumX0  +=      38.164 * cos(5.23367149002  + 44.00592741381 *T)
    sumX0  +=      47.779 * cos(6.19272750750  + 3340.85624418330 *T)
    sumX0  +=      42.228 * cos(5.64690940917  + 77.03115362090 *T)
    sumX0  +=      35.247 * cos(0.20766845689  + 34.75351639890 *T)
    sumX0  +=      46.804 * cos(3.96902162832  + 33.69643246030 *T)
    sumX0  +=      34.352 * cos(1.08289070011  + 33.71098667531 *T)
    sumX0  +=      34.949 * cos(2.01384094499  + 3.37951923889 *T)
    sumX0  +=      36.030 * cos(2.17275904548  + 71.09326278771 *T)
    sumX0  +=      38.112 * cos(5.65470955047  + 45.96597300160 *T)
    sumX0  +=      33.119 * cos(5.27794057043  + 7.35736448430 *T)
    sumX0  +=      32.049 * cos(4.61840704188  + 34.44469985821 *T)
    sumX0  +=      31.910 * cos(1.77890975693  + 81.61769818981 *T)
    sumX0  +=      38.697 * cos(2.66910057126  + 184.97110483931 *T)
    sumX0  +=      41.486 * cos(2.58550378076  + 310.47079377080 *T)
    sumX0  +=      38.631 * cos(2.31715796823  + 50.90702609350 *T)
    sumX0  +=      42.711 * cos(2.19232104972  + 1021.49271203491 *T)
    sumX0  +=      32.006 * cos(0.97590559431  + 42.00018984371 *T)
    sumX0  +=      38.436 * cos(0.31352578874  + 5.92107588581 *T)
    sumX0  +=      38.880 * cos(3.29381198979  + 76.55807286891 *T)
    sumX0  +=      41.190 * cos(4.58002024645  + 563.87503252191 *T)
    sumX0  +=      29.786 * cos(1.00565266044  + 77.50672650040 *T)
    sumX0  +=      40.604 * cos(4.47511985144  + 292.94468595250 *T)
    sumX0  +=      35.275 * cos(1.67517293934  + 304.84171947829 *T)
    sumX0  +=      38.242 * cos(2.80091349300  + 17.76992530181 *T)
    sumX0  +=      34.445 * cos(4.48124108827  + 319.06881347989 *T)
    sumX0  +=      28.725 * cos(5.51593817617  + 67.63668240410 *T)
    sumX0  +=      32.809 * cos(5.57900930431  + 91.54262404029 *T)
    sumX0  +=      38.880 * cos(0.56654650956  + 76.46170464930 *T)
    sumX0  +=      30.731 * cos(5.22467991145  + 67.60305250931 *T)
    sumX0  +=      28.459 * cos(0.11298908847  + 43.04271956730 *T)
    sumX0  +=      35.368 * cos(3.56936550095  + 313.43973918739 *T)
    sumX0  +=      35.703 * cos(0.06787236157  + 258.26823069831 *T)
    sumX0  +=      32.317 * cos(2.30071476395  + 78.95756931390 *T)
    sumX0  +=      29.243 * cos(0.30724049567  + 61.01077037031 *T)
    sumX0  +=      26.235 * cos(3.88058959053  + 137.27684164590 *T)
    sumX0  +=      26.519 * cos(6.20266742881  + 57.49930823250 *T)
    sumX0  +=      24.931 * cos(5.73688334159  + 42.99702758500 *T)
    sumX0  +=      27.608 * cos(5.39681935370  + 103.76398047180 *T)
    sumX0  +=      28.680 * cos(4.65490114562  + 215.19414196860 *T)
    sumX0  +=      25.052 * cos(5.70195779765  + 350.08830211689 *T)
    sumX0  +=      31.386 * cos(4.10756442698  + 22.22334342671 *T)
    sumX0  +=      27.545 * cos(1.30787829275  + 100.62827871640 *T)
    sumX0  +=      22.617 * cos(3.46251776435  + 36.84419630320 *T)
    sumX0  +=      24.909 * cos(0.20851017271  + 24.36220744081 *T)
    sumX0  +=      26.216 * cos(4.94808817995  + 491.80174694030 *T)
    sumX0  +=      28.040 * cos(2.83295165264  + 11.55015017831 *T)
    sumX0  +=      23.047 * cos(4.24570423583  + 35.51978228931 *T)
    sumX0  +=      27.067 * cos(3.95547247738  + 326.62430346539 *T)
    sumX0  +=      26.192 * cos(2.35959813381  + 20.73887071840 *T)
    sumX0  +=      23.134 * cos(2.59485537406  + 68.43313391180 *T)
    sumX0  +=      21.423 * cos(0.87822750255  + 39.90950993941 *T)
    sumX0  +=      25.696 * cos(0.32414101638  + 186.45557754760 *T)
    sumX0  +=      26.985 * cos(3.53264939991  + 69.60879007940 *T)
    sumX0  +=      23.284 * cos(2.71588030137  + 79.43065006591 *T)
    sumX0  +=      22.894 * cos(0.61847067768  + 227.77000692311 *T)
    sumX0  +=      22.482 * cos(0.72349596890  + 39.81314171980 *T)
    sumX0  +=      23.480 * cos(4.39643703557  + 30.98811947460 *T)
    sumX0  +=      20.858 * cos(3.23577429095  + 41.23392395330 *T)
    sumX0  +=      20.327 * cos(1.15567976096  + 39.03124442710 *T)
    sumX0  +=      20.327 * cos(0.04331485179  + 37.72246181551 *T)
    sumX0  +=      22.639 * cos(0.21515321589  + 0.98002279390 *T)
    sumX0  +=      22.639 * cos(0.21515321589  + 1.46765776091 *T)
    sumX0  +=      19.139 * cos(0.03506366059  + 205.94173095370 *T)
    sumX0  +=      19.118 * cos(1.62564867989  + 2119.00767786191 *T)
    sumX0  +=      25.698 * cos(2.97643019475  + 401.40590203270 *T)
    sumX0  +=      21.582 * cos(4.29532713983  + 81.13006322279 *T)
    sumX0  +=      25.509 * cos(4.64829559110  + 329.59324888200 *T)
    sumX0  +=      24.296 * cos(2.11682013072  + 62.00760811160 *T)
    sumX0  +=      23.969 * cos(0.88887585882  + 135.30473397060 *T)
    sumX0  +=      20.599 * cos(4.51946091131  + 491.31411197330 *T)
    sumX0  +=      16.829 * cos(5.63589438225  + 3.16457879030 *T)
    sumX0  +=      20.030 * cos(4.02146628228  + 217.47506618460 *T)
    sumX0  +=      20.377 * cos(0.89378346451  + 209.61075965840 *T)
    sumX0  +=      17.251 * cos(2.57319624936  + 350.57593708390 *T)
    sumX0  +=      19.625 * cos(6.12382765898  + 129.67565967810 *T)
    sumX0  +=      22.707 * cos(5.69106089810  + 1436.29693524910 *T)
    sumX0  +=      17.142 * cos(0.00501932570  + 29.47001687150 *T)
    sumX0  +=      16.188 * cos(4.90861200887  + 39.00999256771 *T)
    sumX0  +=      16.188 * cos(2.57356791106  + 37.74371367490 *T)
    sumX0  +=      20.858 * cos(4.67505024087  + 58.98378094080 *T)
    sumX0  +=      15.747 * cos(1.88900821622  + 154.26043274300 *T)
    sumX0  +=      19.714 * cos(0.33238117487  + 294.91679362781 *T)
    sumX0  +=      19.078 * cos(2.73754913300  + 202.49721265760 *T)
    sumX0  +=      21.530 * cos(3.37996249680  + 114.15528942990 *T)
    sumX0  +=      19.068 * cos(1.82733694293  + 138.27367938720 *T)
    sumX0  +=      18.723 * cos(6.21404671018  + 323.74923414091 *T)
    sumX0  +=      18.916 * cos(5.47002080885  + 40.38259069140 *T)
    sumX0  +=      15.843 * cos(0.27660393480  + 72.57773549600 *T)
    sumX0  +=      20.695 * cos(5.32080415125  + 86.07111631471 *T)
    sumX0  +=      15.895 * cos(5.73200518668  + 736.12033101530 *T)
    sumX0  +=      14.983 * cos(2.13549071268  + 743.23387801611 *T)
    sumX0  +=      14.928 * cos(0.78464963633  + 34.23225153711 *T)
    sumX0  +=      15.461 * cos(6.04598420333  + 20.85074530300 *T)
    sumX0  +=      16.206 * cos(6.05974800797  + 138.76131435421 *T)
    sumX0  +=      15.978 * cos(0.85734083354  + 515.70768857651 *T)
    sumX0  +=      14.173 * cos(2.99587831656  + 99.14380600810 *T)
    sumX0  +=      18.749 * cos(3.37545937432  + 54.53036281590 *T)
    sumX0  +=      13.971 * cos(5.11256155147  + 76.77052119001 *T)
    sumX0  +=      13.971 * cos(5.03098185419  + 76.24925632820 *T)
    sumX0  +=      14.035 * cos(4.45768361334  + 235.68919520349 *T)
    sumX0  +=      18.894 * cos(4.59865824553  + 31.47575444160 *T)
    sumX0  +=      14.967 * cos(0.97104009185  + 52.39149880180 *T)
    sumX0  +=      17.392 * cos(1.69348450373  + 74.06220820430 *T)
    sumX0  +=      14.788 * cos(5.00944229014  + 56.01483552421 *T)
    sumX0  +=      15.758 * cos(5.97423795440  + 208.86249226050 *T)
    sumX0  +=      12.911 * cos(0.41434497695  + 42.52145470550 *T)
    sumX0  +=      14.356 * cos(4.89778066710  + 251.84270489810 *T)
    sumX0  +=      16.266 * cos(4.96350311575  + 853.44019923550 *T)
    sumX0  +=      15.513 * cos(1.02523907534  + 59.03866269500 *T)
    sumX0  +=      12.783 * cos(2.34267333656  + 107.52937739611 *T)
    sumX0  +=      16.075 * cos(4.73335524561  + 366.24181181149 *T)
    sumX0  +=      14.277 * cos(4.88488299527  + 19.36627259471 *T)
    sumX0  +=      14.742 * cos(1.55115458505  + 82.44777959230 *T)
    sumX0  +=      15.111 * cos(4.13629021798  + 363.27286639489 *T)
    sumX0  +=      14.981 * cos(5.88358063018  + 82.61453593110 *T)
    sumX0  +=      14.840 * cos(0.62836299110  + 44.05411152360 *T)
    sumX0  +=      15.592 * cos(1.03195525294  + 8.62938887150 *T)
    sumX0  +=      14.568 * cos(2.02105422692  + 73.80157577341 *T)
    sumX0  +=      12.251 * cos(1.18824225128  + 47.28368937111 *T)
    sumX0  +=      11.447 * cos(0.91374266731  + 175.40987728371 *T)
    sumX0  +=      13.900 * cos(5.64591952885  + 700.42042171730 *T)
    sumX0  +=      15.583 * cos(3.88966860773  + 837.45344587970 *T)
    sumX0  +=      12.109 * cos(2.10142517621  + 33.00841125970 *T)
    sumX0  +=      12.379 * cos(5.59016916358  + 140.41254340130 *T)
    sumX0  +=      11.481 * cos(5.22670638349  + 39.20693452380 *T)
    sumX0  +=      11.481 * cos(2.25547353643  + 37.54677171881 *T)
    sumX0  +=      11.452 * cos(1.21111994028  + 529.41351771630 *T)
    sumX0  +=      10.981 * cos(0.01852111423  + 63.49208081989 *T)
    sumX0  +=      12.137 * cos(2.33017731448  + 42.30900638440 *T)
    sumX0  +=      13.771 * cos(4.49397894473  + 76.62176334371 *T)
    sumX0  +=      11.036 * cos(3.16457889057  + 530.45604743991 *T)
    sumX0  +=      11.537 * cos(4.29449656032  + 199.31581891990 *T)
    sumX0  +=      11.189 * cos(3.24467764115  + 80.13322548150 *T)
    sumX0  +=      12.835 * cos(1.26831311464  + 38.85242600079 *T)
    sumX0  +=      12.879 * cos(4.74400685998  + 5.69407334969 *T)
    sumX0  +=      13.663 * cos(3.12818073078  + 438.05446496220 *T)
    sumX0  +=      10.132 * cos(4.37559264666  + 187.94005025590 *T)
    sumX0  +=      12.619 * cos(4.66177013386  + 65.20355606430 *T)
    sumX0  +=      10.088 * cos(6.12382762451  + 26.58288545949 *T)
    sumX0  +=      11.959 * cos(5.90953104234  + 64.71592109730 *T)
    sumX0  +=      11.578 * cos(4.24710384177  + 275.30670354960 *T)
    sumX0  +=      12.795 * cos(3.23836197733  + 17.88179988640 *T)
    sumX0  +=      13.771 * cos(5.64956481971  + 76.39801417450 *T)
    sumX0  +=      10.044 * cos(0.10145082472  + 147.83490694279 *T)
    sumX0  +=      13.632 * cos(2.86683446064  + 45.27795180100 *T)
    sumX0  +=      11.660 * cos(2.65801239040  + 143.90275367970 *T)
    sumX0  +=       9.938 * cos(4.21970476320  + 6.86972951729 *T)
    sumX0  +=       9.719 * cos(6.05786462616  + 956.53297345411 *T)
    sumX0  +=      11.441 * cos(0.61314587598  + 533.86693584120 *T)
    sumX0  +=      10.240 * cos(2.91846731922  + 80.70267445310 *T)
    sumX0  +=      10.031 * cos(5.38075474506  + 43.74529498291 *T)
    sumX0  +=      10.063 * cos(5.77064020369  + 0.27744737829 *T)
    sumX0  +=      11.428 * cos(3.77013145660  + 526.00262931501 *T)
    sumX0  +=       9.279 * cos(6.16721103485  + 79.64559051450 *T)
    sumX0  +=      10.172 * cos(2.46540726742  + 568.06781821590 *T)
    sumX0  +=       9.198 * cos(5.07759437389  + 112.67081672160 *T)
    sumX0  +=       9.831 * cos(2.49002547943  + 20.90562705720 *T)
    sumX0  +=       9.830 * cos(3.51040521049  + 544.16187657970 *T)
    sumX0  +=       8.646 * cos(4.49185896918  + 30.77567115350 *T)
    sumX0  +=       9.315 * cos(0.15689765715  + 65.63094483399 *T)
    sumX0  +=       9.201 * cos(0.09219461091  + 184.48346987229 *T)
    sumX0  +=       8.674 * cos(2.01170720350  + 624.15435044170 *T)
    sumX0  +=      10.739 * cos(0.49719235939  + 331.56535655731 *T)
    sumX0  +=       9.612 * cos(5.38629260665  + 182.00215942271 *T)
    sumX0  +=       8.664 * cos(5.62437013922  + 1479.11039154791 *T)
    sumX0  +=       8.092 * cos(5.65922856578  + 6.83609962250 *T)
    sumX0  +=      10.092 * cos(4.71596617075  + 419.24082639170 *T)
    sumX0  +=      10.233 * cos(4.88231209018  + 402.89037474099 *T)
    sumX0  +=       8.502 * cos(2.03567120581  + 17.39416491939 *T)
    sumX0  +=      10.189 * cos(2.58985636739  + 21.70207856490 *T)
    sumX0  +=       9.829 * cos(5.23644081358  + 121.23520653590 *T)
    sumX0  +=       8.406 * cos(2.47191018350  + 376.91500505990 *T)
    sumX0  +=       8.060 * cos(5.62304271115  + 415.79630809560 *T)
    sumX0  +=       9.455 * cos(0.06796991442  + 167.80869531589 *T)
    sumX0  +=       7.941 * cos(1.43287391293  + 526.75338884040 *T)
    sumX0  +=       7.870 * cos(2.90339733997  + 533.11617631580 *T)
    sumX0  +=       7.695 * cos(0.92731028198  + 906.60597015449 *T)
    sumX0  +=       7.862 * cos(0.91484097138  + 1265.81129610991 *T)
    sumX0  +=       8.062 * cos(1.12885573257  + 105.73608814710 *T)
    sumX0  +=       8.904 * cos(4.30824949636  + 399.92142932440 *T)
    sumX0  +=       8.050 * cos(0.14722556593  + 143.86912378490 *T)
    sumX0  +=       9.102 * cos(4.77518241515  + 348.17644063891 *T)
    sumX0  +=       7.137 * cos(1.26110622464  + 117.56368570370 *T)
    sumX0  +=       7.076 * cos(3.19957487812  + 26.84351789039 *T)
    sumX0  +=       8.418 * cos(1.48515415206  + 77.73372903651 *T)
    sumX0  +=       8.257 * cos(4.44435970504  + 117.77862615229 *T)
    sumX0  +=       7.868 * cos(5.07706724776  + 288.49126782760 *T)
    sumX0  +=       8.093 * cos(0.41458983168  + 1692.40948698591 *T)
    sumX0  +=       6.910 * cos(0.44789832682  + 216.72430665921 *T)
    sumX0  +=       7.092 * cos(0.01337002281  + 452.65981147369 *T)
    sumX0  +=       7.060 * cos(1.93108090539  + 453.70234119730 *T)
    sumX0  +=       8.233 * cos(3.50880140177  + 480.00777927849 *T)
    sumX0  +=       6.772 * cos(4.46250089888  + 210.36151918381 *T)
    sumX0  +=       7.025 * cos(1.42668370417  + 55.90296093960 *T)
    sumX0  +=       8.356 * cos(2.10000097648  + 95.73540973430 *T)
    sumX0  +=       7.404 * cos(1.00293545057  + 75.28604848170 *T)
    sumX0  +=       6.839 * cos(0.99943444853  + 41.51255487670 *T)
    sumX0  +=       7.909 * cos(1.64368221183  + 36.63174798211 *T)
    sumX0  +=       7.909 * cos(2.69690505451  + 40.12195826051 *T)
    sumX0  +=       6.362 * cos(0.26347531595  + 29.99128173331 *T)
    sumX0  +=       6.712 * cos(0.84138813413  + 133.82026126229 *T)
    sumX0  +=       7.571 * cos(2.81738238064  + 23.70781613500 *T)
    sumX0  +=       6.677 * cos(0.10164158344  + 1.20702533000 *T)
    sumX0  +=       7.600 * cos(0.07294781428  + 494.23487328010 *T)
    sumX0  +=       8.009 * cos(0.39086308190  + 170.72945662269 *T)
    sumX0  +=       7.584 * cos(6.04989436828  + 119.26309886060 *T)
    sumX0  +=       6.599 * cos(2.25520576507  + 32.22651396700 *T)
    sumX0  +=       6.085 * cos(4.97064703625  + 322.00412900171 *T)
    sumX0  +=       5.953 * cos(2.49854544351  + 52214.18313626970 *T)
    sumX0  +=       7.827 * cos(3.28593277837  + 474.70302789170 *T)
    sumX0  +=       7.907 * cos(4.46293464979  + 485.63685357099 *T)
    sumX0  +=       7.372 * cos(4.88712847504  + 55.05162767771 *T)
    sumX0  +=       6.966 * cos(5.60552242454  + 647.25465079831 *T)
    sumX0  +=       6.266 * cos(5.78133779594  + 177.06110633080 *T)
    sumX0  +=       5.900 * cos(4.92602771915  + 52061.16335875149 *T)
    sumX0  +=       6.221 * cos(2.35523958706  + 602.00806815971 *T)
    sumX0  +=       5.552 * cos(5.87735995607  + 223.10414047710 *T)
    sumX0  +=       5.976 * cos(1.83099110545  + 10.80188278040 *T)
    sumX0  +=       7.600 * cos(5.33804556108  + 488.60579898760 *T)
    sumX0  +=       6.831 * cos(0.04615498459  + 1582.20316576650 *T)
    sumX0  +=       5.654 * cos(3.04032114806  + 12604.52855310410 *T)
    sumX0  +=       5.798 * cos(1.13675043219  + 27.49790919620 *T)
    sumX0  +=       7.216 * cos(0.18192294134  + 739.04109232210 *T)
    sumX0  +=       6.579 * cos(3.94809746775  + 2.69149803831 *T)
    sumX0  +=       5.758 * cos(2.82344188087  + 30.03946584310 *T)
    sumX0  +=       5.270 * cos(3.46743079634  + 6166.94845288619 *T)
    sumX0  +=       7.398 * cos(0.58333334375  + 709.72101684200 *T)
    sumX0  +=       5.679 * cos(5.91776083103  + 17.22740858061 *T)
    sumX0  +=       5.205 * cos(2.61017638124  + 426.35437339250 *T)
    sumX0  +=       5.146 * cos(0.81172664742  + 46.76242450930 *T)
    sumX0  +=       5.694 * cos(2.94913098744  + 168.98435148349 *T)
    sumX0  +=       6.627 * cos(6.07668723879  + 221.13203280179 *T)
    sumX0  +=       5.443 * cos(4.34867602386  + 525.74199688410 *T)
    sumX0  +=       6.475 * cos(2.52364293984  + 591.07424248041 *T)
    sumX0  +=       4.984 * cos(4.89088409053  + 10097.15814910579 *T)
    sumX0  +=       5.318 * cos(5.22697316848  + 44.52719227561 *T)
    sumX0  +=       6.699 * cos(2.95047965393  + 2157.14071349970 *T)
    sumX0  +=       6.443 * cos(5.65068156930  + 675.04456158780 *T)
    sumX0  +=       5.078 * cos(0.96513123174  + 101.62511645769 *T)
    sumX0  +=       5.394 * cos(0.88948762211  + 368.21391948681 *T)
    sumX0  +=       5.072 * cos(2.52597530610  + 272.33775813299 *T)
    sumX0  +=       5.208 * cos(4.53150187093  + 277.27881122490 *T)
    sumX0  +=       5.332 * cos(1.28621962216  + 280.93577784210 *T)
    sumX0  +=       5.989 * cos(5.89271411050  + 93.02709674860 *T)
    sumX0  +=       6.329 * cos(0.49570607842  + 18.87863762769 *T)
    sumX0  +=       5.551 * cos(2.57045763275  + 57.38743364790 *T)
    sumX0  +=       6.471 * cos(0.04463535540  + 68.12431737110 *T)
    sumX0  +=       4.708 * cos(2.23921095477  + 95.68722562449 *T)
    sumX0  +=       5.891 * cos(5.96441381591  + 381.59542572090 *T)
    sumX0  +=       4.717 * cos(4.31682479516  + 104.28524533360 *T)
    sumX0  +=       5.675 * cos(1.71229301179  + 1165.63928319810 *T)
    sumX0  +=       5.888 * cos(0.43219504278  + 42.34263627919 *T)
    sumX0  +=       5.587 * cos(4.09170092519  + 459.60660213570 *T)
    sumX0  +=       5.456 * cos(1.50864831442  + 75.50098893029 *T)
    sumX0  +=       5.940 * cos(6.28075673596  + 6318.48375769610 *T)
    sumX0  +=       5.207 * cos(4.55134069280  + 436.56999225390 *T)
    sumX0  +=       6.160 * cos(4.76046448210  + 749.82616015511 *T)
    sumX0  +=       6.137 * cos(4.59348226478  + 713.17759722561 *T)
    sumX0  +=       4.547 * cos(2.39218547281  + 32.47259218289 *T)
    sumX0  +=       5.246 * cos(4.97888240032  + 109.96250373590 *T)
    sumX0  +=       5.244 * cos(2.33674770879  + 73.58912745230 *T)
    sumX0  +=       5.572 * cos(6.12038028190  + 102.11275142471 *T)
    sumX0  +=       5.638 * cos(1.42053892188  + 10248.69345391570 *T)
    sumX0  +=       4.513 * cos(1.62848698862  + 1272.92484311070 *T)
    sumX0  +=       4.340 * cos(2.36449866810  + 384.02855206069 *T)
    sumX0  +=       4.263 * cos(4.24631269159  + 1577.52274510549 *T)
    sumX0  +=       5.964 * cos(4.92643136579  + 786.47472308461 *T)
    sumX0  +=       4.962 * cos(6.09839378254  + 257.78059573129 *T)
    sumX0  +=       5.327 * cos(5.70215230442  + 107.74182571721 *T)
    sumX0  +=       5.572 * cos(0.87438107795  + 291.29345690540 *T)
    sumX0  +=       4.336 * cos(5.80113193852  + 53.40958840249 *T)
    sumX0  +=       4.427 * cos(3.00157250839  + 189.42452296421 *T)
    sumX0  +=       4.157 * cos(3.46647899628  + 29.50364676630 *T)
    sumX0  +=       4.646 * cos(2.87774169214  + 13285.93981804009 *T)
    sumX0  +=       5.507 * cos(4.27464738844  + 178.11819026941 *T)
    sumX0  +=       5.348 * cos(1.42468292991  + 24.88347230261 *T)
    sumX0  +=       5.339 * cos(3.91840662285  + 314.66357946480 *T)
    sumX0  +=       4.678 * cos(4.43608792406  + 1474.42997088690 *T)
    sumX0  +=       4.090 * cos(3.35633664186  + 765.38016029810 *T)
    sumX0  +=       5.008 * cos(5.85701520659  + 352.06040979221 *T)
    sumX0  +=       5.562 * cos(0.40887335705  + 6248.15557725370 *T)
    sumX0  +=       4.983 * cos(3.16236150253  + 1055.43296197871 *T)
    sumX0  +=       4.566 * cos(5.25700629292  + 325.13983075710 *T)
    sumX0  +=       5.327 * cos(5.25347269162  + 439.53893767049 *T)
    sumX0  +=       5.121 * cos(5.84825704577  + 711.69312451730 *T)
    sumX0  +=       4.181 * cos(1.11749590962  + 6606.19943734880 *T)
    sumX0  +=       4.293 * cos(4.65873798886  + 46.71424039951 *T)
    sumX0  +=       5.532 * cos(0.53479774781  + 320.03202132639 *T)
    sumX0  +=       4.492 * cos(0.09912827297  + 52177.53457334019 *T)
    sumX0  +=       4.312 * cos(1.38883413817  + 22.87773473250 *T)
    sumX0  +=       5.332 * cos(1.83070192574  + 10178.36527347330 *T)
    sumX0  +=       4.593 * cos(0.14820750962  + 1025.68549772890 *T)
    sumX0  +=       5.439 * cos(5.09447580219  + 823.12328601411 *T)
    sumX0  +=       3.870 * cos(4.27995377915  + 1596.43025976811 *T)
    sumX0  +=       3.892 * cos(2.11564791977  + 226.07308589371 *T)
    sumX0  +=       4.891 * cos(2.80814026706  + 8.14175390450 *T)
    sumX0  +=       4.689 * cos(3.52062924653  + 276.79117625789 *T)
    sumX0  +=       4.268 * cos(2.59269427473  + 374.15181032000 *T)
    sumX0  +=       3.828 * cos(2.28076604659  + 2138.23319883710 *T)
    sumX0  +=       4.592 * cos(3.87527577295  + 1376.01761732930 *T)
    sumX0  +=       4.629 * cos(0.97709160917  + 122.71967924421 *T)
    sumX0  +=       3.871 * cos(3.17548325596  + 531.41925528640 *T)
    sumX0  +=       4.995 * cos(0.32063762943  + 32.69959471901 *T)
    sumX0  +=       4.711 * cos(0.43748317622  + 52252.31617190749 *T)
    sumX0  +=       3.893 * cos(0.12475334110  + 116.29415344400 *T)
    sumX0  +=       4.481 * cos(4.66479841820  + 53.04589010760 *T)
    sumX0  +=       4.136 * cos(2.59386926777  + 503.10807963510 *T)
    sumX0  +=       4.508 * cos(4.38574998818  + 562.12992738271 *T)
    sumX0  +=       5.025 * cos(0.39865233659  + 283.38345839689 *T)
    sumX0  +=       4.789 * cos(2.68692249791  + 627.72280540990 *T)
    sumX0  +=       4.021 * cos(0.14454426922  + 6603.23049193219 *T)
    sumX0  +=       5.163 * cos(4.77460676620  + 25519.83532335829 *T)
    sumX0  +=       4.150 * cos(3.86319541901  + 27.44302744200 *T)
    sumX0  +=       3.623 * cos(2.29457319711  + 1665.58278404290 *T)
    sumX0  +=       4.634 * cos(1.79141170909  + 3227.45397501119 *T)
    sumX0  +=       4.060 * cos(6.21658618282  + 304.47802118340 *T)
    sumX0  +=       3.862 * cos(0.50812728673  + 74.50415118900 *T)
    sumX0  +=       3.561 * cos(4.92971224760  + 358.65269193120 *T)
    sumX0  +=       4.557 * cos(6.27521064672  + 25974.74468988559 *T)
    sumX0  +=       4.264 * cos(1.56884112199  + 634.93941827469 *T)
    sumX0  +=       4.482 * cos(1.70550805319  + 342.61105682121 *T)
    sumX0  +=       3.539 * cos(0.56907944763  + 119.75073382760 *T)
    sumX0  +=       4.304 * cos(0.63784646457  + 12567.87999017460 *T)
    sumX0  +=       4.138 * cos(4.03567139847  + 107.25419075020 *T)
    sumX0  +=       4.284 * cos(0.05420881503  + 294.42915866079 *T)
    sumX0  +=       3.723 * cos(5.58644401851  + 987.32545955500 *T)
    sumX0  +=       3.723 * cos(5.58644401851  + 987.81309452200 *T)
    sumX0  +=       4.606 * cos(5.49553530451  + 14.42521950279 *T)
    sumX0  +=       4.236 * cos(6.22240593144  + 155.91166179010 *T)
    sumX0  +=       4.458 * cos(2.64590572483  + 395.82251972250 *T)
    sumX0  +=       4.798 * cos(5.23929868658  + 530.19541500900 *T)
    sumX0  +=       3.640 * cos(2.22734915897  + 2564.83138971310 *T)
    sumX0  +=       3.563 * cos(5.37459598926  + 12451.50877558589 *T)
    sumX0  +=       3.443 * cos(2.13809774331  + 245.25042275910 *T)
    sumX0  +=       3.429 * cos(4.73423412994  + 530.04665716270 *T)
    sumX0  +=       3.872 * cos(4.09217464449  + 308.98632106249 *T)
    sumX0  +=       3.406 * cos(5.88979864779  + 529.82290799351 *T)
    sumX0  +=       4.348 * cos(1.52419659995  + 20311.92816802509 *T)
    sumX0  +=       4.589 * cos(5.24153025487  + 181.08713568601 *T)
    sumX0  +=       3.854 * cos(5.92510183178  + 12564.91104475801 *T)
    sumX0  +=       3.789 * cos(4.29351893525  + 3101.63590180850 *T)
    sumX0  +=       3.783 * cos(0.26936683978  + 1614.17130803499 *T)
    sumX0  +=       3.904 * cos(0.00421090422  + 369.80145805910 *T)
    sumX0  +=       3.765 * cos(4.70889835066  + 1025.94613015981 *T)
    sumX0  +=       4.231 * cos(5.35914297519  + 31.52393855141 *T)
    sumX0  +=       4.303 * cos(4.97345150272  + 396.78572756900 *T)
    sumX0  +=       4.085 * cos(1.80921070558  + 14.47091148511 *T)
    sumX0  +=       4.085 * cos(1.80921070558  + 13.98327651810 *T)
    sumX0  +=       3.346 * cos(4.91522066963  + 20351.54567637119 *T)
    sumX0  +=       4.021 * cos(6.08537487228  + 748.34168744680 *T)
    sumX0  +=       3.753 * cos(1.17204243376  + 524.99372948619 *T)
    sumX0  +=       3.935 * cos(1.24122122244  + 1617.14025345159 *T)
    sumX0  +=       4.432 * cos(3.45778366813  + 511.35159082120 *T)
    sumX0  +=       4.170 * cos(4.42864444413  + 274.87931477991 *T)
    sumX0  +=       3.317 * cos(1.79347554880  + 266.70868384049 *T)
    sumX0  +=       4.545 * cos(4.56531161641  + 244.56240155850 *T)
    sumX0  +=       3.589 * cos(1.55384880430  + 59.52629766200 *T)
    sumX0  +=       3.464 * cos(0.37736158688  + 102.27950776349 *T)
    sumX0  +=       4.526 * cos(4.55402483522  + 525.79018099390 *T)
    sumX0  +=       4.603 * cos(4.40260765490  + 26088.14695905770 *T)
    sumX0  +=       4.021 * cos(5.38581853850  + 52174.56562792359 *T)
    sumX0  +=       3.276 * cos(1.95663025139  + 1306.37745808750 *T)
    sumX0  +=       3.214 * cos(3.94235488355  + 20348.57673095459 *T)
    sumX0  +=       3.706 * cos(5.25360971143  + 27.07052042651 *T)
    sumX0  +=       3.759 * cos(4.32245166720  + 164.83974989929 *T)
    sumX0  +=       3.184 * cos(2.01654309849  + 538.01153742540 *T)
    sumX0  +=       4.430 * cos(5.37917502758  + 529.67415014720 *T)
    sumX0  +=       4.064 * cos(1.03322736236  + 6130.29988995670 *T)
    sumX0  +=       3.918 * cos(4.20575585414  + 375.43053235159 *T)
    sumX0  +=       4.058 * cos(5.13313296042  + 433.43429049850 *T)
    sumX0  +=       3.919 * cos(0.36694469487  + 1092.81773021860 *T)
    sumX0  +=       3.919 * cos(0.36694469487  + 1093.30536518560 *T)
    sumX0  +=       3.175 * cos(1.14568678321  + 241.36645360580 *T)
    sumX0  +=       3.135 * cos(5.81037649777  + 127.22797912329 *T)
    sumX0  +=       3.834 * cos(1.84941829775  + 14.31334491820 *T)
    sumX0  +=       4.022 * cos(1.72079825603  + 1477.83836716070 *T)
    sumX0  +=       3.221 * cos(1.09261076661  + 78.16111780620 *T)
    sumX0  +=       3.426 * cos(0.06166201047  + 519.85229016070 *T)
    sumX0  +=       4.369 * cos(0.74973637733  + 746.36957977150 *T)
    sumX0  +=       3.160 * cos(2.01821245093  + 664.99569906519 *T)
    sumX0  +=       4.060 * cos(6.06087716530  + 51.87023394000 *T)
    sumX0  +=       3.107 * cos(5.38240469077  + 28.92750015030 *T)
    sumX0  +=       3.259 * cos(5.62260974194  + 657.88215206440 *T)
    sumX0  +=       3.428 * cos(1.24133782529  + 2351.53229427510 *T)
    sumX0  +=       3.235 * cos(1.64692472660  + 406.34695512460 *T)
    sumX0  +=       3.161 * cos(5.69758725685  + 982.87204143010 *T)
    sumX0  +=       4.351 * cos(1.04662835997  + 20388.19423930069 *T)
    sumX0  +=       3.384 * cos(0.30649784029  + 660.85109748100 *T)
    sumX0  +=       3.452 * cos(4.39659352485  + 326.18236048070 *T)
    sumX0  +=       3.298 * cos(0.15489069807  + 1403.84115801359 *T)
    sumX0  +=       3.278 * cos(3.68945780931  + 941.77006037570 *T)
    sumX0  +=       3.723 * cos(5.00962048402  + 451.95723605810 *T)
    sumX0  +=       3.173 * cos(5.46640783518  + 1400.87221259699 *T)
    sumX0  +=       4.113 * cos(1.87439213951  + 1049.31625271919 *T)
    sumX0  +=       4.012 * cos(2.15082049909  + 52.60394712290 *T)
    sumX0  +=       4.142 * cos(4.89782789900  + 978.67925573610 *T)
    sumX0  +=       4.295 * cos(1.37302733197  + 875.58648151749 *T)
    sumX0  +=       3.224 * cos(3.81995287471  + 459.11896716870 *T)
    sumX0  +=       3.151 * cos(3.69005421605  + 381.85605815180 *T)
    sumX0  +=       3.633 * cos(1.38559724652  + 256.78375799000 *T)
    sumX0  +=       4.250 * cos(0.10595516218  + 528.71094230071 *T)
    sumX0  +=       4.186 * cos(2.09187651842  + 943.25453308399 *T)
    sumX0  +=       3.406 * cos(0.25126866750  + 170.46882419179 *T)
    sumX0  +=       3.231 * cos(4.61367643853  + 400.82094669610 *T)
    sumX0  +=       3.726 * cos(0.55318715397  + 1096.48675892331 *T)
    sumX0  +=       3.792 * cos(0.75464081409  + 111.93461141120 *T)
    sumX0  +=       3.651 * cos(4.56933341620  + 154.42718908179 *T)
    sumX0  +=       3.839 * cos(2.45649426115  + 10060.50958617629 *T)
    sumX0  +=       3.356 * cos(0.62546125542  + 1586.34776735071 *T)
    sumX0  +=       3.219 * cos(5.97786590701  + 213.70966926030 *T)
    sumX0  +=       3.671 * cos(1.51743688101  + 57.60237409650 *T)
    sumX0  +=       4.187 * cos(0.29242250575  + 2772.54460848389 *T)
    sumX0  +=       2.960 * cos(2.20142019667  + 2461.73861549450 *T)
    sumX0  +=       3.331 * cos(0.81281655951  + 10133.80671203529 *T)
    sumX0  +=       3.341 * cos(1.17831577639  + 243.76595005080 *T)
    sumX0  +=       3.466 * cos(4.73891819304  + 1150.92455422949 *T)
    sumX0  +=       3.296 * cos(3.49817757911  + 1653.78881638109 *T)
    sumX0  +=       3.014 * cos(1.90092216670  + 1477.39891630350 *T)
    sumX0  +=       4.118 * cos(2.83150543771  + 25596.58902960090 *T)
    sumX0  +=       2.951 * cos(5.04298380276  + 42.78208713641 *T)
    sumX0  +=       2.951 * cos(5.58078877076  + 33.97161910620 *T)
    sumX0  +=       3.830 * cos(4.59720174528  + 323.48860171000 *T)
    sumX0  +=       3.313 * cos(1.64840054600  + 939.10993149980 *T)
    sumX0  +=       3.031 * cos(2.75126158832  + 156450.90896068609 *T)
    sumX0  +=       3.606 * cos(3.92819651217  + 1082.25966492170 *T)
    sumX0  +=       2.967 * cos(0.01380556143  + 6.39415663780 *T)
    sumX0  +=       2.995 * cos(3.55729257964  + 139.70996798570 *T)
    sumX0  +=       3.251 * cos(3.50186784018  + 709.29362807231 *T)
    sumX0  +=       3.480 * cos(0.61716473120  + 518.14081491630 *T)
    sumX0  +=       3.906 * cos(2.84871380483  + 1119.90506559249 *T)
    sumX0  +=       3.406 * cos(1.85522558472  + 148.79811478929 *T)
    sumX0  +=       3.359 * cos(1.74239209634  + 642.84941678320 *T)
    sumX0  +=       3.027 * cos(0.29741383095  + 184.00789699280 *T)
    sumX0  +=       2.918 * cos(2.25866029656  + 83.62343575990 *T)
    sumX0  +=       3.347 * cos(6.10666820526  + 217.68751450571 *T)
    sumX0  +=       3.277 * cos(0.27333269638  + 912.54386098770 *T)
    sumX0  +=       3.277 * cos(0.27333269638  + 913.03149595471 *T)
    sumX0  +=       3.196 * cos(5.84286985933  + 363.10611005610 *T)
    sumX0  +=       2.869 * cos(4.50334436600  + 285.35556607221 *T)
    sumX0  +=       3.158 * cos(1.18152957041  + 540.01727499551 *T)
    sumX0  +=       2.810 * cos(5.14802919795  + 1592.28565818390 *T)
    sumX0  +=       3.471 * cos(6.13160952457  + 144.39038864671 *T)
    sumX0  +=       3.159 * cos(4.14451339136  + 197.55615956570 *T)
    sumX0  +=       3.227 * cos(5.73841253304  + 6203.59701581570 *T)
    sumX0  +=       3.750 * cos(5.81139240481  + 303.35724676999 *T)
    sumX0  +=       3.848 * cos(3.38110828764  + 26048.04181574459 *T)
    sumX0  +=       2.741 * cos(1.70084306406  + 70.83263035680 *T)
    sumX0  +=       2.826 * cos(2.07742210458  + 460.29462333630 *T)
    sumX0  +=       2.748 * cos(0.98378370701  + 600.52359545141 *T)
    sumX0  +=       3.057 * cos(6.13629771077  + 23.81969071961 *T)
    sumX0  +=       3.057 * cos(1.34588220916  + 52.93401552300 *T)
    sumX0  +=       3.446 * cos(3.54046646150  + 500.13913421850 *T)
    sumX0  +=       2.703 * cos(4.69192633180  + 908.09044286280 *T)
    sumX0  +=       2.817 * cos(3.26718539283  + 210.62215161470 *T)
    sumX0  +=       2.848 * cos(5.88127781412  + 450.47276334980 *T)
    sumX0  +=       2.724 * cos(0.93671586048  + 23.18655127321 *T)
    sumX0  +=       2.905 * cos(5.85039527890  + 149.31937965110 *T)
    sumX0  +=       2.848 * cos(6.20081143930  + 622.66987773339 *T)
    sumX0  +=       2.733 * cos(3.50715759295  + 262.72164882321 *T)
    sumX0  +=       2.863 * cos(0.69834580836  + 175.57663362249 *T)
    sumX0  +=       2.681 * cos(1.11809511751  + 25.19228884330 *T)
    sumX0  +=       2.822 * cos(1.57963221264  + 259.75270340660 *T)
    sumX0  +=       3.174 * cos(6.18541771069  + 347.11935670030 *T)
    sumX0  +=       3.271 * cos(1.40248413653  + 458.59770230690 *T)
    sumX0  +=       2.894 * cos(4.18128306427  + 71.82946809809 *T)
    sumX0  +=       3.490 * cos(2.85083291634  + 664.37136833940 *T)
    sumX0  +=       3.506 * cos(5.48691285949  + 771.34811171130 *T)
    sumX0  +=       3.326 * cos(2.12303698267  + 45.22976769120 *T)
    sumX0  +=       2.988 * cos(0.23324807191  + 299.37021175271 *T)
    sumX0  +=       2.916 * cos(3.60780287924  + 6642.84800027830 *T)
    sumX0  +=       2.916 * cos(0.46621022565  + 6643.33563524530 *T)
    sumX0  +=       2.630 * cos(1.12694509764  + 2751.79141717511 *T)
    sumX0  +=       2.903 * cos(4.31055308658  + 477.08701797169 *T)
    sumX0  +=       2.804 * cos(0.26456593020  + 6681.46867088311 *T)
    sumX0  +=       2.622 * cos(2.30179163581  + 521.85802773080 *T)
    sumX0  +=       2.606 * cos(6.15707729666  + 410.85525500370 *T)
    sumX0  +=       3.046 * cos(2.36386768037  + 959.45373476091 *T)
    sumX0  +=       3.127 * cos(3.04512463308  + 225.55182103190 *T)
    sumX0  +=       2.700 * cos(4.45467896965  + 963.64652045490 *T)
    sumX0  +=       2.778 * cos(1.65860124839  + 238.39750818919 *T)
    sumX0  +=       3.029 * cos(4.72630934575  + 473.21855518340 *T)
    sumX0  +=       2.671 * cos(4.60029996028  + 531.94052014820 *T)
    sumX0  +=       2.914 * cos(3.86169076602  + 554.31380496631 *T)
    sumX0  +=       3.087 * cos(6.08851917121  + 340.26644213040 *T)
    sumX0  +=       3.438 * cos(2.32466413132  + 6171.40187101109 *T)
    sumX0  +=       2.879 * cos(5.61809470376  + 218.65072235220 *T)
    sumX0  +=       3.140 * cos(5.02001385281  + 609.12161516050 *T)
    sumX0  +=       3.003 * cos(0.53592571188  + 464.97504399731 *T)
    sumX0  +=       3.257 * cos(1.52476284257  + 305.96249389171 *T)
    sumX0  +=       3.211 * cos(5.64833047248  + 416.53251340600 *T)
    sumX0  +=       3.265 * cos(1.54950325507  + 24.73471445630 *T)
    sumX0  +=       2.644 * cos(1.01963899758  + 508.59414157570 *T)
    sumX0  +=       2.764 * cos(4.98225869197  + 410.59462257279 *T)
    sumX0  +=       3.428 * cos(5.71088563789  + 1012.61950567990 *T)
    sumX0  +=       2.614 * cos(4.07639961382  + 213.59109703130 *T)
    sumX0  +=       3.469 * cos(5.28643352424  + 24.14975911971 *T)
    sumX0  +=       2.606 * cos(0.81160096517  + 213.49472881170 *T)
    sumX0  +=       3.444 * cos(2.56432157215  + 891.57323487331 *T)
    sumX0  +=       2.540 * cos(4.32167771768  + 564.87187026320 *T)
    sumX0  +=       2.540 * cos(4.32167771768  + 565.35950523021 *T)
    sumX0  +=       2.754 * cos(2.69535555411  + 57.55418998670 *T)
    sumX0  +=       2.531 * cos(0.59020723407  + 800.59243462910 *T)
    sumX0  +=       2.557 * cos(0.66999256840  + 341.49028240779 *T)
    sumX0  +=       2.601 * cos(4.54885591305  + 261.23717611490 *T)
    sumX0  +=       3.027 * cos(0.20183300410  + 331.07772159029 *T)
    sumX0  +=       2.494 * cos(0.58142193078  + 203.98168536590 *T)
    sumX0  +=       2.590 * cos(1.76325981719  + 1190.54206257560 *T)
    sumX0  +=       3.494 * cos(2.90876238684  + 534.07938416230 *T)
    sumX0  +=       3.144 * cos(0.01981710217  + 1503.96498681560 *T)
    sumX0  +=       2.818 * cos(3.61898449244  + 49.31067880061 *T)
    sumX0  +=       2.791 * cos(4.48606671949  + 288.32451148881 *T)
    sumX0  +=       2.471 * cos(1.23009614301  + 411.11588743459 *T)
    sumX0  +=       3.059 * cos(3.30977686438  + 172.48911597691 *T)
    sumX0  +=       2.972 * cos(0.30229231666  + 569.29165849331 *T)
    sumX0  +=       3.418 * cos(5.40293550246  + 638.39599865830 *T)
    sumX0  +=       2.541 * cos(4.99016167757  + 1448.09090291091 *T)
    sumX0  +=       2.663 * cos(0.43151826022  + 573.69689250840 *T)
    sumX0  +=       2.439 * cos(4.39632185677  + 1625.96527569680 *T)
    sumX0  +=       2.739 * cos(5.72535305895  + 112.88326504270 *T)
    sumX0  +=       2.821 * cos(5.66863744979  + 402.93606672331 *T)
    sumX0  +=       3.412 * cos(1.27007980380  + 772.83258441960 *T)
    sumX0  +=       2.624 * cos(5.85528852490  + 1624.48080298850 *T)
    sumX0  +=       3.170 * cos(0.53682796950  + 1011.13503297159 *T)
    sumX0  +=       2.908 * cos(4.60949958082  + 635.94831810351 *T)
    sumX0  +=       2.664 * cos(2.68003479349  + 409.41896640519 *T)
    sumX0  +=       3.091 * cos(1.88245278611  + 379.25961975071 *T)
    sumX0  +=       3.301 * cos(1.91350932819  + 19.79366136440 *T)
    sumX0  +=       3.176 * cos(3.29730129609  + 300.90956621520 *T)
    sumX0  +=       3.022 * cos(5.94822554077  + 52.01899178630 *T)
    sumX0  +=       2.890 * cos(1.53549747897  + 293.43232091950 *T)
    sumX0  +=       2.698 * cos(1.69370735844  + 78149.06650032569 *T)
    sumX0  +=       2.558 * cos(0.74578099458  + 1371.33719666830 *T)
    sumX0  +=       2.619 * cos(3.80578981072  + 202.00957769060 *T)
    sumX0  +=       3.176 * cos(3.75055063339  + 10101.61156723069 *T)
    sumX0  +=       3.341 * cos(2.34080319182  + 345.89551642290 *T)
    sumX0  +=       2.373 * cos(4.96475711609  + 130.85131584570 *T)
    sumX0  +=       2.644 * cos(2.68099240015  + 305.10235190919 *T)
    sumX0  +=       3.339 * cos(4.63303989765  + 2849.29831472650 *T)
    sumX0  +=       2.410 * cos(1.58163612779  + 951.85255279310 *T)
    sumX0  +=       3.303 * cos(2.25771292490  + 769.57294599210 *T)
    sumX0  +=       3.302 * cos(4.85894681967  + 90.15202742410 *T)
    sumX0  +=       2.416 * cos(6.00635580174  + 527.92904500800 *T)
    sumX0  +=       2.361 * cos(5.34789183737  + 905.12149744620 *T)
    sumX0  +=       2.737 * cos(4.77190944455  + 1206.21999939070 *T)
    sumX0  +=       2.441 * cos(3.82975575752  + 246.73489546739 *T)
    sumX0  +=       2.441 * cos(0.68816310393  + 247.22253043440 *T)
    sumX0  +=       2.957 * cos(4.25832811500  + 238.23075185041 *T)
    sumX0  +=       3.263 * cos(0.98630889937  + 1506.93393223219 *T)
    sumX0  +=       3.293 * cos(5.93270574395  + 66.15220969580 *T)
    sumX0  +=       3.241 * cos(3.43806050184  + 978.41862330520 *T)
    sumX0  +=       3.149 * cos(3.64971867049  + 271.91036936330 *T)
    sumX0  +=       3.149 * cos(3.64971867050  + 271.42273439630 *T)
    sumX0  +=       2.328 * cos(5.07609916236  + 31.73887900000 *T)
    sumX0  +=       2.372 * cos(0.68652074740  + 309.03450517230 *T)
    sumX0  +=       2.372 * cos(3.82811340099  + 309.52214013930 *T)
    sumX0  +=       2.369 * cos(4.33012817739  + 418.98019396080 *T)
    sumX0  +=       3.007 * cos(4.64009260533  + 1437.78140795740 *T)
    sumX0  +=       3.034 * cos(5.98346126252  + 330.86278114170 *T)
    sumX0  +=       2.345 * cos(2.80677153952  + 453.93183586090 *T)
    sumX0  +=       3.118 * cos(3.73398781358  + 1434.81246254079 *T)
    sumX0  +=       2.324 * cos(3.85931736808  + 495.24626523640 *T)
    sumX0  +=       2.340 * cos(5.41992470939  + 452.43031681009 *T)
    sumX0  +=       2.336 * cos(0.04655833240  + 189.59127930300 *T)
    sumX0  +=       2.920 * cos(3.78758562864  + 1549.69920442121 *T)
    sumX0  +=       2.494 * cos(0.79353025531  + 1187.57311715899 *T)
    sumX0  +=       2.692 * cos(4.17807622816  + 425.13053311509 *T)
    sumX0  +=       2.874 * cos(4.63267401857  + 1654.27645134810 *T)
    sumX0  +=       2.809 * cos(5.67077170621  + 317.58434077160 *T)
    sumX0  +=       2.735 * cos(3.93990204220  + 1513.05064149171 *T)
    sumX0  +=       2.949 * cos(6.26993364897  + 186.71620997851 *T)
    sumX0  +=       2.320 * cos(0.74326897219  + 487.38195871019 *T)
    sumX0  +=       3.113 * cos(6.20902109610  + 353.28425006961 *T)
    sumX0  +=       3.086 * cos(4.87476303199  + 1230.59902177890 *T)
    sumX0  +=       2.722 * cos(2.16494792915  + 49.68318581610 *T)
    sumX0  +=       3.064 * cos(3.68765217385  + 133.13224006171 *T)
    sumX0  +=       3.064 * cos(3.68765217385  + 132.64460509469 *T)
    sumX0  +=       2.470 * cos(2.78243316001  + 532.38246313290 *T)
    sumX0  +=       2.640 * cos(0.51790972890  + 394.33804701421 *T)
    sumX0  +=       2.252 * cos(1.84613004390  + 22.65073219640 *T)
    sumX0  +=       3.151 * cos(5.26039361613  + 859.77184894361 *T)
    sumX0  +=       2.671 * cos(0.92145640556  + 37.36795329250 *T)
    sumX0  +=       2.380 * cos(0.86687455354  + 429.27513469930 *T)
    sumX0  +=       2.655 * cos(2.72088152594  + 484.15238086270 *T)
    sumX0  +=       3.005 * cos(3.02367934874  + 1929.33933741419 *T)
    sumX0  +=       2.550 * cos(5.60497907633  + 496.94318626580 *T)
    sumX0  +=       2.290 * cos(3.41120190653  + 455.18681390559 *T)
    sumX0  +=       2.608 * cos(3.85525903926  + 422.95803920620 *T)
    sumX0  +=       2.226 * cos(2.09977531258  + 47.82620609231 *T)
    sumX0  +=       2.233 * cos(4.94028872789  + 877.34614087170 *T)
    sumX0  +=       2.764 * cos(0.83501700112  + 356.68058425589 *T)
    sumX0  +=       2.719 * cos(1.98953734068  + 177.58237119260 *T)
    sumX0  +=       2.999 * cos(2.06885612973  + 1926.37039199759 *T)
    sumX0  +=       2.693 * cos(3.57972778548  + 6284.80414018320 *T)
    sumX0  +=       2.369 * cos(1.19578023344  + 70.88081446661 *T)
    sumX0  +=       2.498 * cos(3.71851216671  + 315.15121443180 *T)
    sumX0  +=       2.204 * cos(3.20466206592  + 442.88613559700 *T)
    sumX0  +=       2.261 * cos(3.32534753019  + 621.23358913490 *T)
    sumX0  +=       2.213 * cos(6.16263836668  + 1189.05758986730 *T)
    sumX0  +=       2.492 * cos(2.67366070604  + 406.97128585040 *T)
    sumX0  +=       2.976 * cos(1.45402284302  + 1014.10397838820 *T)
    sumX0  +=       2.840 * cos(5.35710509350  + 522.33360061030 *T)
    sumX0  +=       2.340 * cos(1.72448626630  + 440.43845504219 *T)
    sumX0  +=       3.012 * cos(1.13512104183  + 15.90969221110 *T)
    sumX0  +=       3.012 * cos(4.27671369542  + 16.39732717810 *T)
    sumX0  +=       2.372 * cos(0.24227395275  + 132.59642098490 *T)
    sumX0  +=       2.232 * cos(2.42168492591  + 158.12984768129 *T)
    sumX0  +=       2.961 * cos(4.37134416172  + 286.35240381350 *T)
    sumX0  +=       2.961 * cos(4.37134416172  + 286.84003878050 *T)

    # NEPTUNE X Coord
    sumX1  = 0
    sumX1  +=  357822.049 * cos(4.60537437341  + 0.24381748350 *T)
    sumX1  +=  256200.629 * cos(2.01693264233  + 36.89238041300 *T)
    sumX1  +=  242677.799 * cos(5.46293481092  + 39.86132582961 *T)
    sumX1  +=  106073.143 * cos(3.07856435709  + 37.88921815429 *T)
    sumX1  +=  103735.195 * cos(6.08270773807  + 38.37685312130 *T)
    sumX1  +=  118508.231 * cos(2.88623136735  + 76.50988875911 *T)
    sumX1  +=   21930.692 * cos(3.20019569049  + 35.40790770471 *T)
    sumX1  +=   17445.772 * cos(4.26396070854  + 41.34579853790 *T)
    sumX1  +=   13038.843 * cos(5.36684741537  + 3.21276290011 *T)
    sumX1  +=    4928.885 * cos(2.08893204170  + 73.54094334250 *T)
    sumX1  +=    2742.686 * cos(4.06389633495  + 77.99436146740 *T)
    sumX1  +=    2155.134 * cos(4.11881068429  + 4.69723560840 *T)
    sumX1  +=    1882.800 * cos(4.42038284259  + 33.92343499640 *T)
    sumX1  +=    1572.888 * cos(1.07810551784  + 114.64292439690 *T)
    sumX1  +=    1326.507 * cos(6.02985868883  + 75.02541605080 *T)
    sumX1  +=    1343.094 * cos(3.03838214796  + 42.83027124621 *T)
    sumX1  +=     897.979 * cos(4.26993024752  + 426.84200835950 *T)
    sumX1  +=     865.617 * cos(1.66618456177  + 37.85558825950 *T)
    sumX1  +=     849.963 * cos(5.81599535394  + 38.89811798311 *T)
    sumX1  +=     922.754 * cos(3.34516686314  + 72.05647063421 *T)
    sumX1  +=     726.258 * cos(4.24833812431  + 36.40474544600 *T)
    sumX1  +=     778.220 * cos(5.84479856092  + 206.42936592071 *T)
    sumX1  +=     754.025 * cos(5.33205816073  + 220.65645992230 *T)
    sumX1  +=     607.406 * cos(0.10576615596  + 1059.62574767270 *T)
    sumX1  +=     571.831 * cos(2.42930874906  + 522.82123557730 *T)
    sumX1  +=     560.995 * cos(1.91555986158  + 537.04832957890 *T)
    sumX1  +=     501.078 * cos(1.71335109406  + 28.81562556571 *T)
    sumX1  +=     493.238 * cos(5.24702261334  + 39.37369086260 *T)
    sumX1  +=     474.802 * cos(4.40715596351  + 98.65617104110 *T)
    sumX1  +=     453.975 * cos(1.71443209341  + 35.92917256650 *T)
    sumX1  +=     471.731 * cos(4.84217171915  + 1.72829019180 *T)
    sumX1  +=     410.057 * cos(5.76579953705  + 40.82453367610 *T)
    sumX1  +=     366.899 * cos(5.76755572930  + 47.93808067690 *T)
    sumX1  +=     450.109 * cos(1.25670451550  + 76.02225379210 *T)
    sumX1  +=     354.347 * cos(6.27109348494  + 1.24065522479 *T)
    sumX1  +=     300.159 * cos(2.88687992256  + 6.18170831670 *T)
    sumX1  +=     327.501 * cos(4.20479564636  + 33.43580002939 *T)
    sumX1  +=     174.973 * cos(5.64027558321  + 32.43896228810 *T)
    sumX1  +=     171.503 * cos(4.43985554308  + 34.18406742730 *T)
    sumX1  +=     156.749 * cos(2.59545084410  + 79.47883417571 *T)
    sumX1  +=     152.549 * cos(0.58219894744  + 30.30009827400 *T)
    sumX1  +=     150.775 * cos(3.03954929901  + 42.56963881530 *T)
    sumX1  +=     162.280 * cos(0.79977049351  + 31.26330612050 *T)
    sumX1  +=     131.609 * cos(1.62895622934  + 7.83293736379 *T)
    sumX1  +=     136.159 * cos(4.57878446789  + 70.57199792590 *T)
    sumX1  +=     134.616 * cos(0.39184091634  + 45.49040012211 *T)
    sumX1  +=     116.304 * cos(0.61710594601  + 46.45360796860 *T)
    sumX1  +=     115.918 * cos(1.81843338530  + 44.31474395451 *T)
    sumX1  +=     110.293 * cos(6.26561089969  + 35.45609181450 *T)
    sumX1  +=      99.282 * cos(5.06218386285  + 2.72512793310 *T)
    sumX1  +=      99.914 * cos(1.21626942611  + 41.29761442810 *T)
    sumX1  +=     108.706 * cos(3.09142093314  + 113.15845168861 *T)
    sumX1  +=      88.965 * cos(4.26680850699  + 60.52313540329 *T)
    sumX1  +=      86.886 * cos(5.46872067794  + 31.95132732110 *T)
    sumX1  +=      72.232 * cos(3.50587405737  + 640.14110379750 *T)
    sumX1  +=      86.985 * cos(4.80098575405  + 419.72846135871 *T)
    sumX1  +=      73.430 * cos(4.36511226727  + 70.08436295889 *T)
    sumX1  +=      53.395 * cos(4.46520807878  + 433.95555536030 *T)
    sumX1  +=      57.451 * cos(1.08003733120  + 213.54291292150 *T)
    sumX1  +=      51.458 * cos(4.01726374522  + 69.39634175830 *T)
    sumX1  +=      48.797 * cos(6.01365170443  + 111.67397898031 *T)
    sumX1  +=      48.557 * cos(6.24808481100  + 2.67694382330 *T)
    sumX1  +=      42.206 * cos(3.23823062186  + 74.53778108379 *T)
    sumX1  +=      42.550 * cos(1.67247318349  + 7.66618102501 *T)
    sumX1  +=      39.462 * cos(5.84051041865  + 31.78457098230 *T)
    sumX1  +=      39.445 * cos(0.94630986910  + 12.77399045571 *T)
    sumX1  +=      42.389 * cos(5.59019905902  + 110.18950627200 *T)
    sumX1  +=      44.118 * cos(0.44615133445  + 1589.31671276730 *T)
    sumX1  +=      37.988 * cos(2.73850879415  + 6.34846465550 *T)
    sumX1  +=      37.802 * cos(5.98781130211  + 14.25846316400 *T)
    sumX1  +=      36.280 * cos(6.27894536142  + 273.82223084130 *T)
    sumX1  +=      37.247 * cos(4.62968774107  + 73.05330837550 *T)
    sumX1  +=      36.282 * cos(2.85336450932  + 84.58664360640 *T)
    sumX1  +=      40.018 * cos(4.27418471085  + 4.43660317750 *T)
    sumX1  +=      32.400 * cos(1.64328879813  + 44.96913526031 *T)
    sumX1  +=      31.842 * cos(5.16652228087  + 34.92027273770 *T)
    sumX1  +=      32.037 * cos(2.94551844856  + 27.33115285740 *T)
    sumX1  +=      34.456 * cos(3.37152599360  + 529.93478257810 *T)
    sumX1  +=      31.208 * cos(1.73420111381  + 1052.51220067191 *T)
    sumX1  +=      30.002 * cos(2.33639558082  + 1066.73929467350 *T)
    sumX1  +=      33.805 * cos(6.04114496470  + 149.80701461810 *T)
    sumX1  +=      33.096 * cos(2.45794089359  + 116.12739710521 *T)
    sumX1  +=      30.571 * cos(4.02151161164  + 22.39009976550 *T)
    sumX1  +=      24.020 * cos(0.23821463973  + 63.97971578690 *T)
    sumX1  +=      23.780 * cos(4.34619784366  + 105.76971804189 *T)
    sumX1  +=      23.110 * cos(2.51348904373  + 23.87457247379 *T)
    sumX1  +=      22.233 * cos(1.21582777350  + 174.92224231670 *T)
    sumX1  +=      21.377 * cos(3.74139728076  + 316.63568714010 *T)
    sumX1  +=      25.400 * cos(4.52702077197  + 106.73292588839 *T)
    sumX1  +=      20.754 * cos(0.94473828111  + 5.66044345490 *T)
    sumX1  +=      25.572 * cos(5.46068636869  + 529.44714761109 *T)
    sumX1  +=      19.878 * cos(1.01805940417  + 32.96022714990 *T)
    sumX1  +=      19.754 * cos(4.49000124955  + 49.42255338521 *T)
    sumX1  +=      19.241 * cos(6.01413524726  + 7.14491616321 *T)
    sumX1  +=      17.979 * cos(1.48478132886  + 62.49524307860 *T)
    sumX1  +=      19.513 * cos(5.64767624982  + 68.59989025060 *T)
    sumX1  +=      18.273 * cos(5.20130356006  + 227.77000692311 *T)
    sumX1  +=      17.552 * cos(5.80350239388  + 69.08752521760 *T)
    sumX1  +=      16.704 * cos(3.94483252805  + 40.85816357090 *T)
    sumX1  +=      16.996 * cos(1.10449263633  + 91.54262404029 *T)
    sumX1  +=      16.800 * cos(0.51632981099  + 30.46685461280 *T)
    sumX1  +=      16.800 * cos(0.51632981099  + 30.95448957981 *T)
    sumX1  +=      16.400 * cos(5.02426208775  + 33.26904369061 *T)
    sumX1  +=      17.242 * cos(0.18352088447  + 11.55015017831 *T)
    sumX1  +=      15.590 * cos(1.99260946960  + 37.10482873410 *T)
    sumX1  +=      15.590 * cos(5.48957045033  + 39.64887750850 *T)
    sumX1  +=      15.469 * cos(0.19931591320  + 43.79347909271 *T)
    sumX1  +=      16.590 * cos(2.76837508684  + 33.71098667531 *T)
    sumX1  +=      19.347 * cos(5.63498287914  + 152.77596003471 *T)
    sumX1  +=      14.994 * cos(2.71995630411  + 319.06881347989 *T)
    sumX1  +=      14.395 * cos(0.95950619453  + 110.45013870291 *T)
    sumX1  +=      15.528 * cos(1.16348592635  + 79.43065006591 *T)
    sumX1  +=      13.727 * cos(2.45811249773  + 43.48466255200 *T)
    sumX1  +=      13.988 * cos(3.96979676090  + 4.20960064140 *T)
    sumX1  +=      14.467 * cos(0.42164709009  + 108.70503356371 *T)
    sumX1  +=      16.652 * cos(3.17617180933  + 304.84171947829 *T)
    sumX1  +=      15.153 * cos(3.21783791411  + 72.31710306511 *T)
    sumX1  +=      12.810 * cos(2.37864271463  + 11.28951774740 *T)
    sumX1  +=      12.751 * cos(0.61424962834  + 45.79921666280 *T)
    sumX1  +=      13.293 * cos(4.71511827202  + 43.04271956730 *T)
    sumX1  +=      12.751 * cos(2.55685741962  + 515.70768857651 *T)
    sumX1  +=      11.616 * cos(2.50185569269  + 97.17169833279 *T)
    sumX1  +=      11.538 * cos(6.20327139591  + 633.02755679670 *T)
    sumX1  +=      11.046 * cos(1.26188662646  + 25.84668014910 *T)
    sumX1  +=      11.032 * cos(3.82567311622  + 4.86399194720 *T)
    sumX1  +=      11.189 * cos(3.94939417077  + 83.10217089810 *T)
    sumX1  +=      10.860 * cos(0.38739756984  + 9.80504503910 *T)
    sumX1  +=      10.958 * cos(1.48375898962  + 415.04804069769 *T)
    sumX1  +=      10.244 * cos(0.26615444717  + 71.09326278771 *T)
    sumX1  +=      11.427 * cos(1.50679043329  + 129.67565967810 *T)
    sumX1  +=       9.895 * cos(5.62468142972  + 251.67594855930 *T)
    sumX1  +=       9.802 * cos(1.83814841533  + 44.48150029329 *T)
    sumX1  +=      11.029 * cos(4.69741395112  + 143.38148881789 *T)
    sumX1  +=       9.235 * cos(5.99370019321  + 199.31581891990 *T)
    sumX1  +=       8.899 * cos(3.54958918415  + 7.35736448430 *T)
    sumX1  +=       7.746 * cos(5.89688581764  + 103.33659170210 *T)
    sumX1  +=       8.691 * cos(3.86130807292  + 32.74777882880 *T)
    sumX1  +=       7.714 * cos(5.08136079606  + 65.46418849521 *T)
    sumX1  +=       8.007 * cos(1.80952555463  + 544.16187657970 *T)
    sumX1  +=       7.513 * cos(1.47497152514  + 69.60879007940 *T)
    sumX1  +=       7.336 * cos(5.00489054996  + 15.74293587230 *T)
    sumX1  +=       7.195 * cos(6.14029832936  + 949.41942645330 *T)
    sumX1  +=       9.601 * cos(0.96952505042  + 80.96330688400 *T)
    sumX1  +=       8.094 * cos(3.27383873772  + 526.75338884040 *T)
    sumX1  +=       8.109 * cos(1.06293290956  + 533.11617631580 *T)
    sumX1  +=       6.906 * cos(5.27751864757  + 137.27684164590 *T)
    sumX1  +=       7.455 * cos(5.82601593331  + 105.24845318010 *T)
    sumX1  +=       7.826 * cos(4.02401038086  + 77.03115362090 *T)
    sumX1  +=       6.529 * cos(0.86598314454  + 65.20355606430 *T)
    sumX1  +=       7.134 * cos(3.62090018772  + 44.00592741381 *T)
    sumX1  +=       6.186 * cos(2.42103444520  + 31.47575444160 *T)
    sumX1  +=       6.186 * cos(2.42103444520  + 30.98811947460 *T)
    sumX1  +=       7.698 * cos(0.17876132177  + 14.47091148511 *T)
    sumX1  +=       7.434 * cos(5.53413189412  + 146.83806920150 *T)
    sumX1  +=       6.317 * cos(0.53538901275  + 66.94866120350 *T)
    sumX1  +=       6.903 * cos(6.20818943193  + 75.98862389731 *T)
    sumX1  +=       5.591 * cos(1.90701487438  + 448.98829064149 *T)
    sumX1  +=       6.425 * cos(5.04706195455  + 678.27413943531 *T)
    sumX1  +=       5.483 * cos(3.18327336885  + 34.44469985821 *T)
    sumX1  +=       5.483 * cos(4.29890655108  + 42.30900638440 *T)
    sumX1  +=       5.519 * cos(2.84195707915  + 853.44019923550 *T)
    sumX1  +=       5.483 * cos(3.53790147295  + 100.14064374939 *T)
    sumX1  +=       5.483 * cos(3.53790147295  + 100.62827871640 *T)
    sumX1  +=       6.288 * cos(1.03240051727  + 143.90275367970 *T)
    sumX1  +=       6.239 * cos(5.78066710550  + 17.76992530181 *T)
    sumX1  +=       5.246 * cos(5.09114965169  + 209.61075965840 *T)
    sumX1  +=       5.331 * cos(3.26471064810  + 45.96597300160 *T)
    sumX1  +=       5.131 * cos(6.10583196953  + 217.47506618460 *T)
    sumX1  +=       5.325 * cos(4.39759756568  + 19.25439801010 *T)
    sumX1  +=       5.172 * cos(0.86928942503  + 25.35904518210 *T)
    sumX1  +=       5.139 * cos(2.60427452606  + 6.86972951729 *T)
    sumX1  +=       5.992 * cos(1.19287924990  + 9.31741007210 *T)
    sumX1  +=       5.011 * cos(3.34804654196  + 38.85242600079 *T)
    sumX1  +=       4.975 * cos(1.43964900757  + 525.25436191710 *T)
    sumX1  +=       4.910 * cos(5.04787040559  + 45.27795180100 *T)
    sumX1  +=       5.250 * cos(4.87798510402  + 0.71939036300 *T)
    sumX1  +=       4.731 * cos(1.56230403811  + 40.38259069140 *T)
    sumX1  +=       4.731 * cos(2.77828322823  + 36.37111555120 *T)
    sumX1  +=       5.910 * cos(6.11804979728  + 6168.43292559449 *T)
    sumX1  +=       4.700 * cos(6.23394030506  + 50.90702609350 *T)
    sumX1  +=       5.127 * cos(0.06949696047  + 140.93380826310 *T)
    sumX1  +=       5.321 * cos(3.67018745291  + 1104.87233031131 *T)
    sumX1  +=       6.339 * cos(2.59865692618  + 10175.39632805670 *T)
    sumX1  +=       4.983 * cos(3.03193615352  + 1090.64523630970 *T)
    sumX1  +=       5.487 * cos(4.85218420019  + 180.03005174739 *T)
    sumX1  +=       4.560 * cos(3.95095239655  + 323.74923414091 *T)
    sumX1  +=       4.689 * cos(1.24271255508  + 1068.22376738181 *T)
    sumX1  +=       5.562 * cos(1.26401999292  + 10098.64262181409 *T)
    sumX1  +=       4.432 * cos(2.40638908148  + 415.79630809560 *T)
    sumX1  +=       4.456 * cos(6.17485306628  + 235.68919520349 *T)
    sumX1  +=       4.289 * cos(5.97528879519  + 1051.02772796360 *T)
    sumX1  +=       4.145 * cos(3.13378236518  + 33.69643246030 *T)
    sumX1  +=       4.167 * cos(1.72807331665  + 416.53251340600 *T)
    sumX1  +=       4.107 * cos(2.49036069416  + 61.01077037031 *T)
    sumX1  +=       4.088 * cos(5.45739808026  + 423.66061462181 *T)
    sumX1  +=       5.027 * cos(4.43953537205  + 21.70207856490 *T)
    sumX1  +=       4.030 * cos(5.01269280095  + 216.72430665921 *T)
    sumX1  +=       4.278 * cos(4.65333719777  + 310.47079377080 *T)
    sumX1  +=       4.013 * cos(1.39689438468  + 104275.10267753768 *T)
    sumX1  +=       4.505 * cos(1.22507263591  + 291.94784821120 *T)
    sumX1  +=       3.959 * cos(6.18034607330  + 210.36151918381 *T)
    sumX1  +=       3.962 * cos(4.02082978591  + 978.93988816699 *T)
    sumX1  +=       5.561 * cos(1.74843583918  + 1409.47023230609 *T)
    sumX1  +=       5.073 * cos(3.58205424554  + 1498.33591252310 *T)
    sumX1  +=       4.227 * cos(1.48715312131  + 534.38820070301 *T)
    sumX1  +=       4.054 * cos(4.02884108627  + 430.02340209721 *T)
    sumX1  +=       3.863 * cos(2.24264031920  + 1127.50624756031 *T)
    sumX1  +=       4.367 * cos(1.71153359581  + 58.98378094080 *T)
    sumX1  +=       4.694 * cos(3.33961949377  + 77.50672650040 *T)
    sumX1  +=       4.144 * cos(3.59057653937  + 518.14081491630 *T)
    sumX1  +=       4.289 * cos(3.29776439152  + 921.32069912310 *T)
    sumX1  +=       4.039 * cos(3.79987840474  + 1622.76932774409 *T)
    sumX1  +=       5.180 * cos(5.37115331697  + 99.14380600810 *T)
    sumX1  +=       4.845 * cos(6.04321981604  + 136.78920667889 *T)
    sumX1  +=       4.827 * cos(2.78459346340  + 418.24398865040 *T)
    sumX1  +=       3.722 * cos(0.22932453326  + 1065.25482196520 *T)
    sumX1  +=       4.729 * cos(3.76762324044  + 421.21293406700 *T)
    sumX1  +=       3.490 * cos(5.25995346649  + 986.80419469320 *T)
    sumX1  +=       3.715 * cos(6.02151166051  + 254.10907489909 *T)
    sumX1  +=       3.488 * cos(1.23861297869  + 187.94005025590 *T)
    sumX1  +=       3.989 * cos(3.79961685835  + 95.73540973430 *T)
    sumX1  +=       3.603 * cos(0.65230587403  + 67.11541754230 *T)
    sumX1  +=       3.530 * cos(2.51065807549  + 24.36220744081 *T)
    sumX1  +=       3.538 * cos(3.09031960755  + 57.49930823250 *T)
    sumX1  +=       3.838 * cos(1.99815683749  + 979.90309601349 *T)
    sumX1  +=       3.615 * cos(3.17553643085  + 493.28621964860 *T)
    sumX1  +=       3.457 * cos(0.22865254260  + 807.70598162989 *T)
    sumX1  +=       3.648 * cos(3.01000228275  + 647.25465079831 *T)
    sumX1  +=       4.048 * cos(4.68171378592  + 979.69064769239 *T)
    sumX1  +=       4.414 * cos(3.57495606042  + 1062.59469308931 *T)
    sumX1  +=       3.631 * cos(2.31921127570  + 486.17267264780 *T)
    sumX1  +=       3.347 * cos(5.70639780704  + 151.29148732640 *T)
    sumX1  +=       3.305 * cos(0.52158954660  + 1544.07013012871 *T)
    sumX1  +=       3.428 * cos(1.68792809396  + 107.22056085540 *T)
    sumX1  +=       3.286 * cos(5.12949558917  + 1131.69903325430 *T)
    sumX1  +=       3.389 * cos(1.65565102713  + 28.98238190449 *T)
    sumX1  +=       3.353 * cos(3.87388681549  + 10289.79543497010 *T)
    sumX1  +=       3.214 * cos(2.40799794941  + 569.55229092420 *T)
    sumX1  +=       3.210 * cos(5.76688710335  + 114.15528942990 *T)
    sumX1  +=       3.353 * cos(3.19692974184  + 157.88376946540 *T)
    sumX1  +=       3.339 * cos(3.69632773816  + 443.09858391810 *T)
    sumX1  +=       3.188 * cos(1.05807532391  + 361.13400238079 *T)
    sumX1  +=       3.390 * cos(0.82325646834  + 1558.29722413030 *T)
    sumX1  +=       3.933 * cos(0.51543027693  + 313.43973918739 *T)
    sumX1  +=       3.131 * cos(5.23811887945  + 275.30670354960 *T)
    sumX1  +=       3.156 * cos(4.66486358528  + 431.84043533310 *T)
    sumX1  +=       3.993 * cos(3.33001170426  + 67.63668240410 *T)
    sumX1  +=       3.708 * cos(1.81916567333  + 500.39976664941 *T)
    sumX1  +=       4.051 * cos(2.84746860357  + 59.03866269500 *T)
    sumX1  +=       3.757 * cos(3.59917867608  + 296.40126633610 *T)
    sumX1  +=       3.138 * cos(5.35145867078  + 347.11935670030 *T)
    sumX1  +=       3.086 * cos(3.24315098824  + 392.90175841570 *T)
    sumX1  +=       3.466 * cos(0.02941146445  + 215.19414196860 *T)
    sumX1  +=       3.139 * cos(4.67079139650  + 159.36824217371 *T)
    sumX1  +=       3.466 * cos(1.90282193275  + 2145.34674583789 *T)
    sumX1  +=       3.737 * cos(1.95939265626  + 449.03647475130 *T)
    sumX1  +=       3.286 * cos(3.39700619187  + 435.44002806861 *T)
    sumX1  +=       3.043 * cos(3.45909355839  + 2.20386307129 *T)
    sumX1  +=       3.999 * cos(1.21766663097  + 6245.18663183710 *T)
    sumX1  +=       3.999 * cos(4.35925928456  + 6244.69899687009 *T)
    sumX1  +=       2.999 * cos(1.64598289911  + 526.00262931501 *T)
    sumX1  +=       3.014 * cos(3.95092279768  + 1054.94532701169 *T)
    sumX1  +=       3.091 * cos(2.95397758261  + 42.99702758500 *T)
    sumX1  +=       3.274 * cos(1.10162661548  + 736.12033101530 *T)
    sumX1  +=       2.965 * cos(2.69144881941  + 533.86693584120 *T)
    sumX1  +=       3.149 * cos(0.77764778909  + 103.76398047180 *T)
    sumX1  +=       3.610 * cos(3.04477019722  + 55.05162767771 *T)
    sumX1  +=       2.937 * cos(4.36852075939  + 385.25239233810 *T)
    sumX1  +=       2.903 * cos(5.92315544183  + 117.56368570370 *T)
    sumX1  +=       2.968 * cos(2.85171539624  + 613.31440085451 *T)
    sumX1  +=       3.097 * cos(2.85040396879  + 1395.24313830449 *T)
    sumX1  +=       2.931 * cos(5.17875295945  + 202.49721265760 *T)
    sumX1  +=       3.013 * cos(3.06605929280  + 121.23520653590 *T)
    sumX1  +=       3.206 * cos(1.29027400076  + 53.40958840249 *T)
    sumX1  +=       3.269 * cos(5.16847517364  + 480.00777927849 *T)
    sumX1  +=       3.948 * cos(3.85972628729  + 112.88326504270 *T)
    sumX1  +=       2.824 * cos(1.57846497121  + 176.40671502500 *T)
    sumX1  +=       2.827 * cos(1.93940329091  + 429.81095377611 *T)
    sumX1  +=       3.348 * cos(5.12243609352  + 6284.80414018320 *T)
    sumX1  +=       2.862 * cos(0.86276885894  + 384.02855206069 *T)
    sumX1  +=       3.228 * cos(0.42457598020  + 52.60394712290 *T)
    sumX1  +=       3.446 * cos(3.75606585057  + 62.00760811160 *T)
    sumX1  +=       3.096 * cos(3.26760360935  + 71.82946809809 *T)
    sumX1  +=       3.031 * cos(4.66996407487  + 494.23487328010 *T)
    sumX1  +=       3.021 * cos(1.39292760491  + 328.59641114070 *T)
    sumX1  +=       2.731 * cos(2.36952809744  + 432.47108265200 *T)
    sumX1  +=       3.171 * cos(0.25949036332  + 10215.01383640280 *T)
    sumX1  +=       2.674 * cos(0.72177739894  + 158.12984768129 *T)
    sumX1  +=       2.901 * cos(5.80027365050  + 559.46979850680 *T)
    sumX1  +=       2.631 * cos(5.78146252380  + 2008.80135664250 *T)
    sumX1  +=       2.695 * cos(5.11715867535  + 81.61769818981 *T)
    sumX1  +=       2.695 * cos(1.97556602176  + 81.13006322279 *T)
    sumX1  +=       2.721 * cos(2.68965946829  + 326.18236048070 *T)
    sumX1  +=       2.775 * cos(5.84695952836  + 457.86149699650 *T)
    sumX1  +=       3.054 * cos(1.52217085552  + 6281.83519476660 *T)
    sumX1  +=       2.852 * cos(4.47706032032  + 186.71620997851 *T)
    sumX1  +=       2.538 * cos(0.16145086268  + 111.18634401329 *T)
    sumX1  +=       2.835 * cos(4.50055998275  + 419.50145882259 *T)
    sumX1  +=       2.868 * cos(2.09813621145  + 844.56699288049 *T)
    sumX1  +=       2.530 * cos(1.04013419881  + 1050.75254131770 *T)
    sumX1  +=       2.843 * cos(2.58892628620  + 830.33989887890 *T)
    sumX1  +=       2.848 * cos(3.12250765680  + 659.36662477269 *T)
    sumX1  +=       3.031 * cos(4.13022602708  + 406.34695512460 *T)
    sumX1  +=       2.907 * cos(5.28583383732  + 573.69689250840 *T)
    sumX1  +=       2.536 * cos(3.44172011173  + 82.61453593110 *T)
    sumX1  +=       2.957 * cos(0.45041658093  + 947.70795120889 *T)
    sumX1  +=       3.321 * cos(2.30536254604  + 449.99968259780 *T)
    sumX1  +=       3.117 * cos(3.17172140219  + 457.32567791969 *T)
    sumX1  +=       2.902 * cos(2.94761781535  + 10212.04489098620 *T)
    sumX1  +=       2.459 * cos(2.17142711813  + 450.73339578069 *T)
    sumX1  +=       2.557 * cos(2.89791026532  + 525.48136445320 *T)
    sumX1  +=       2.624 * cos(1.78027142371  + 946.22347850060 *T)
    sumX1  +=       2.417 * cos(4.80936350850  + 351.57277482520 *T)
    sumX1  +=       2.454 * cos(4.84992044892  + 196.01680510321 *T)
    sumX1  +=       2.585 * cos(0.99587951695  + 248.70700314271 *T)
    sumX1  +=       2.549 * cos(1.80324449988  + 1062.80714141041 *T)
    sumX1  +=       2.615 * cos(2.49010683388  + 425.13053311509 *T)
    sumX1  +=       2.387 * cos(3.61270640757  + 654.36819779910 *T)
    sumX1  +=       2.439 * cos(6.16957116292  + 462.74230389109 *T)
    sumX1  +=       2.367 * cos(4.32238656757  + 107.52937739611 *T)
    sumX1  +=       2.538 * cos(2.61932649618  + 481.23161955590 *T)
    sumX1  +=       2.479 * cos(1.99143603619  + 205.94173095370 *T)
    sumX1  +=       2.791 * cos(0.73480816467  + 24.14975911971 *T)
    sumX1  +=       2.626 * cos(1.16663444616  + 213.05527795450 *T)
    sumX1  +=       2.445 * cos(4.84988920933  + 146.87169909629 *T)
    sumX1  +=       2.575 * cos(1.89431453798  + 86.07111631471 *T)
    sumX1  +=       3.120 * cos(5.91742201237  + 456.36247007319 *T)
    sumX1  +=       2.587 * cos(0.02224873460  + 400.82094669610 *T)
    sumX1  +=       2.261 * cos(4.82066845150  + 644.33388949151 *T)
    sumX1  +=       2.796 * cos(5.01395178381  + 216.67861467689 *T)
    sumX1  +=       2.896 * cos(2.29072801127  + 1685.29593998510 *T)
    sumX1  +=       2.453 * cos(3.24288673382  + 109.96250373590 *T)
    sumX1  +=       2.325 * cos(4.96633817815  + 442.88613559700 *T)
    sumX1  +=       2.387 * cos(5.32424727918  + 599.08730685290 *T)
    sumX1  +=       2.873 * cos(3.56351208170  + 834.53268457290 *T)
    sumX1  +=       2.963 * cos(0.77021714990  + 2119.00767786191 *T)
    sumX1  +=       2.233 * cos(5.94426610995  + 709.29362807231 *T)
    sumX1  +=       2.337 * cos(4.30558394222  + 210.57396750490 *T)
    sumX1  +=       2.259 * cos(3.67528651606  + 29.50364676630 *T)
    sumX1  +=       2.300 * cos(0.62755545112  + 986.05343516780 *T)
    sumX1  +=       2.199 * cos(2.78438991669  + 606.20085385370 *T)
    sumX1  +=       2.325 * cos(3.26991047297  + 109.70187130500 *T)

    # NEPTUNE X Coord
    sumX2  = 0
    sumX2  +=  1620002.167 * cos(0.60038473142  + 38.37685312130 *T)
    sumX2  +=   28138.323 * cos(5.58440767451  + 0.24381748350 *T)
    sumX2  +=   12318.619 * cos(2.58513114618  + 39.86132582961 *T)
    sumX2  +=    8346.956 * cos(5.13440715484  + 37.88921815429 *T)
    sumX2  +=    5131.003 * cos(5.12974075920  + 76.50988875911 *T)
    sumX2  +=    4109.792 * cos(1.46495026130  + 36.89238041300 *T)
    sumX2  +=    1369.663 * cos(3.55762715050  + 1.72829019180 *T)
    sumX2  +=     633.706 * cos(2.38135108376  + 3.21276290011 *T)
    sumX2  +=     583.006 * cos(1.54592369321  + 41.34579853790 *T)
    sumX2  +=     546.517 * cos(0.70972594452  + 75.02541605080 *T)
    sumX2  +=     246.224 * cos(2.44618778574  + 213.54291292150 *T)
    sumX2  +=     159.773 * cos(1.26414365966  + 206.42936592071 *T)
    sumX2  +=     156.619 * cos(3.61656709171  + 220.65645992230 *T)
    sumX2  +=     191.674 * cos(2.17166123081  + 529.93478257810 *T)
    sumX2  +=     188.212 * cos(4.43184732741  + 35.40790770471 *T)
    sumX2  +=     117.788 * cos(4.12530218101  + 522.82123557730 *T)
    sumX2  +=     114.488 * cos(0.05081176794  + 35.92917256650 *T)
    sumX2  +=     112.666 * cos(0.21220394551  + 537.04832957890 *T)
    sumX2  +=     105.949 * cos(1.13080468733  + 40.82453367610 *T)
    sumX2  +=      77.696 * cos(0.84633184914  + 77.99436146740 *T)
    sumX2  +=      90.798 * cos(2.05479887039  + 73.54094334250 *T)
    sumX2  +=      67.696 * cos(2.44679430551  + 426.84200835950 *T)
    sumX2  +=      74.860 * cos(1.44346648461  + 4.69723560840 *T)
    sumX2  +=      64.717 * cos(6.06001471150  + 34.92027273770 *T)
    sumX2  +=      51.378 * cos(6.13002795973  + 36.40474544600 *T)
    sumX2  +=      50.205 * cos(0.67007332287  + 42.83027124621 *T)
    sumX2  +=      40.929 * cos(6.22049348014  + 33.92343499640 *T)
    sumX2  +=      36.136 * cos(6.15460243008  + 98.65617104110 *T)
    sumX2  +=      33.953 * cos(5.02568302050  + 1059.62574767270 *T)
    sumX2  +=      34.603 * cos(3.26702076082  + 76.02225379210 *T)
    sumX2  +=      35.441 * cos(2.49518827466  + 31.26330612050 *T)
    sumX2  +=      29.614 * cos(3.34814694529  + 28.81562556571 *T)
    sumX2  +=      31.027 * cos(4.97087448592  + 45.49040012211 *T)
    sumX2  +=      35.521 * cos(1.13028002523  + 39.37369086260 *T)
    sumX2  +=      25.488 * cos(4.04320892614  + 47.93808067690 *T)
    sumX2  +=      20.115 * cos(3.93227849656  + 1.24065522479 *T)
    sumX2  +=      14.328 * cos(2.73754899228  + 433.95555536030 *T)
    sumX2  +=      15.503 * cos(3.13160557995  + 114.64292439690 *T)
    sumX2  +=      16.998 * cos(0.93393874821  + 33.43580002939 *T)
    sumX2  +=      13.166 * cos(0.39556817176  + 419.72846135871 *T)
    sumX2  +=      13.053 * cos(6.11304367644  + 60.52313540329 *T)
    sumX2  +=      10.637 * cos(5.94332175983  + 34.18406742730 *T)
    sumX2  +=       9.610 * cos(1.65699013342  + 640.14110379750 *T)
    sumX2  +=       9.354 * cos(1.41415954295  + 42.56963881530 *T)
    sumX2  +=      11.447 * cos(6.19793150566  + 71.56883566720 *T)
    sumX2  +=       8.454 * cos(2.75562169800  + 2.72512793310 *T)
    sumX2  +=       9.012 * cos(4.44112001347  + 72.05647063421 *T)
    sumX2  +=       9.594 * cos(5.79483289118  + 69.39634175830 *T)
    sumX2  +=       7.419 * cos(3.49645345230  + 227.77000692311 *T)
    sumX2  +=       6.800 * cos(5.14250085135  + 113.15845168861 *T)
    sumX2  +=       6.267 * cos(0.79586518424  + 1066.73929467350 *T)
    sumX2  +=       6.895 * cos(4.31090775556  + 111.67397898031 *T)
    sumX2  +=       5.770 * cos(1.23253284004  + 32.43896228810 *T)
    sumX2  +=       5.686 * cos(2.23805923850  + 30.30009827400 *T)
    sumX2  +=       6.679 * cos(4.85529332414  + 258.78949556011 *T)
    sumX2  +=       7.799 * cos(1.58396135295  + 7.35736448430 *T)
    sumX2  +=       5.906 * cos(5.92485931723  + 44.31474395451 *T)
    sumX2  +=       5.606 * cos(5.17941805418  + 46.45360796860 *T)
    sumX2  +=       5.525 * cos(3.61911776351  + 1052.51220067191 *T)
    sumX2  +=       7.257 * cos(0.17315189128  + 1097.75878331050 *T)
    sumX2  +=       5.427 * cos(1.80586256737  + 105.76971804189 *T)
    sumX2  +=       5.179 * cos(4.25986194250  + 515.70768857651 *T)
    sumX2  +=       5.163 * cos(6.27919257182  + 7.83293736379 *T)
    sumX2  +=       4.688 * cos(2.52139212878  + 222.14093263061 *T)
    sumX2  +=       5.379 * cos(5.86341629561  + 22.39009976550 *T)
    sumX2  +=       4.607 * cos(4.89100806572  + 549.16030355330 *T)
    sumX2  +=       4.101 * cos(0.29016053329  + 213.05527795450 *T)
    sumX2  +=       4.262 * cos(5.51395238453  + 204.94489321240 *T)
    sumX2  +=       3.916 * cos(0.21419544093  + 207.91383862900 *T)
    sumX2  +=       4.089 * cos(4.97684127402  + 304.84171947829 *T)
    sumX2  +=       3.729 * cos(1.40848954224  + 199.31581891990 *T)
    sumX2  +=       3.680 * cos(5.54809318630  + 1589.31671276730 *T)
    sumX2  +=       3.702 * cos(1.01863119774  + 319.06881347989 *T)
    sumX2  +=       4.832 * cos(1.26423594188  + 215.02738562980 *T)
    sumX2  +=       3.474 * cos(1.17401543445  + 103.33659170210 *T)
    sumX2  +=       3.298 * cos(0.10619126376  + 544.16187657970 *T)
    sumX2  +=       4.521 * cos(0.07911565781  + 108.21739859670 *T)
    sumX2  +=       3.967 * cos(2.67776762695  + 944.73900579230 *T)
    sumX2  +=       4.059 * cos(3.01104350847  + 149.80701461810 *T)
    sumX2  +=       4.009 * cos(5.61852534309  + 533.11617631580 *T)
    sumX2  +=       3.288 * cos(1.44957894842  + 407.93449369690 *T)
    sumX2  +=       3.976 * cos(5.00221858099  + 526.75338884040 *T)
    sumX2  +=       3.343 * cos(0.65785646071  + 531.41925528640 *T)
    sumX2  +=       3.932 * cos(3.66244745467  + 91.54262404029 *T)
    sumX2  +=       3.478 * cos(6.19876429652  + 6.18170831670 *T)
    sumX2  +=       2.967 * cos(1.04478648324  + 860.55374623631 *T)
    sumX2  +=       3.058 * cos(0.02482940557  + 342.97475511610 *T)
    sumX2  +=       3.974 * cos(2.04910269520  + 335.86120811530 *T)
    sumX2  +=       2.849 * cos(0.86611245106  + 666.48017177350 *T)
    sumX2  +=       2.999 * cos(1.27874757189  + 937.62545879149 *T)
    sumX2  +=       3.008 * cos(5.16783990611  + 74.53778108379 *T)
    sumX2  +=       3.080 * cos(3.04902400148  + 129.67565967810 *T)
    sumX2  +=       3.346 * cos(4.64303639862  + 1162.71852189130 *T)
    sumX2  +=       2.625 * cos(1.69459459452  + 273.82223084130 *T)
    sumX2  +=       2.931 * cos(1.57809055514  + 235.68919520349 *T)
    sumX2  +=       2.579 * cos(0.48473918174  + 1073.85284167431 *T)
    sumX2  +=       2.550 * cos(6.14366282644  + 26.58288545949 *T)
    sumX2  +=       2.542 * cos(4.22297682017  + 1265.81129610991 *T)
    sumX2  +=       2.483 * cos(1.73279038376  + 453.18107633550 *T)
    sumX2  +=       2.732 * cos(1.76626805419  + 563.38739755489 *T)
    sumX2  +=       2.508 * cos(1.67664275102  + 37.85558825950 *T)
    sumX2  +=       2.508 * cos(0.34076894500  + 425.35753565121 *T)
    sumX2  +=       2.680 * cos(5.74609617365  + 454.66554904380 *T)
    sumX2  +=       2.511 * cos(3.15018930028  + 209.61075965840 *T)
    sumX2  +=       2.512 * cos(1.74477024139  + 217.47506618460 *T)
    sumX2  +=       2.552 * cos(5.67032305105  + 79.47883417571 *T)
    sumX2  +=       2.457 * cos(2.67033044982  + 38.89811798311 *T)
    sumX2  +=       2.343 * cos(1.93599816349  + 981.38756872180 *T)
    sumX2  +=       2.501 * cos(1.67412173715  + 669.40093308030 *T)
    sumX2  +=       2.330 * cos(3.95065162563  + 38.32866901151 *T)
    sumX2  +=       2.327 * cos(0.39474993260  + 38.42503723110 *T)
    sumX2  +=       2.481 * cos(5.56752927904  + 655.17383907870 *T)
    sumX2  +=       2.569 * cos(4.22623902188  + 464.97504399731 *T)

    # NEPTUNE X Coord
    sumX3  = 0
    sumX3  +=     985.355 * cos(0.69240373955  + 38.37685312130 *T)
    sumX3  +=     482.798 * cos(0.83271959724  + 37.88921815429 *T)
    sumX3  +=     416.447 * cos(0.37037561694  + 0.24381748350 *T)
    sumX3  +=     303.825 * cos(0.53797420117  + 39.86132582961 *T)
    sumX3  +=      89.203 * cos(1.52338099991  + 36.89238041300 *T)
    sumX3  +=      70.862 * cos(5.83899744010  + 76.50988875911 *T)
    sumX3  +=      28.900 * cos(5.65001946959  + 41.34579853790 *T)
    sumX3  +=      22.279 * cos(2.95886685234  + 206.42936592071 *T)
    sumX3  +=      21.480 * cos(1.87359273442  + 220.65645992230 *T)
    sumX3  +=      16.157 * cos(5.83581915834  + 522.82123557730 *T)
    sumX3  +=      15.714 * cos(4.76719570238  + 537.04832957890 *T)
    sumX3  +=      11.404 * cos(2.25961154910  + 35.40790770471 *T)
    sumX3  +=      13.199 * cos(0.06057423298  + 7.35736448430 *T)
    sumX3  +=       7.024 * cos(0.69890179308  + 3.21276290011 *T)
    sumX3  +=       6.772 * cos(1.19679143435  + 69.39634175830 *T)
    sumX3  +=       4.517 * cos(3.28893027439  + 45.49040012211 *T)
    sumX3  +=       4.523 * cos(4.18705629066  + 31.26330612050 *T)
    sumX3  +=       3.682 * cos(1.47261975259  + 98.65617104110 *T)
    sumX3  +=       3.656 * cos(0.60146659814  + 968.64494742849 *T)
    sumX3  +=       3.927 * cos(0.54740561653  + 426.84200835950 *T)
    sumX3  +=       3.199 * cos(1.67327016260  + 1519.67655352550 *T)
    sumX3  +=       3.498 * cos(3.27476696786  + 407.93449369690 *T)
    sumX3  +=       3.304 * cos(2.24745680549  + 422.16158769850 *T)
    sumX3  +=       3.331 * cos(1.97045240379  + 36.40474544600 *T)
    sumX3  +=       3.244 * cos(2.85168281358  + 484.20056497250 *T)
    sumX3  +=       2.689 * cos(0.89932845788  + 441.06910236111 *T)
    sumX3  +=       3.247 * cos(1.76512860403  + 498.42765897409 *T)
    sumX3  +=       2.651 * cos(0.45669916115  + 304.84171947829 *T)
    sumX3  +=       2.645 * cos(5.61672793904  + 461.77909604459 *T)
    sumX3  +=       2.542 * cos(5.76347018476  + 444.58305662640 *T)
    sumX3  +=       2.524 * cos(0.75739625090  + 433.95555536030 *T)
    sumX3  +=       2.472 * cos(5.63416184223  + 319.06881347989 *T)
    sumX3  +=       2.355 * cos(0.46192754883  + 447.55200204300 *T)
    sumX3  +=       2.876 * cos(5.22513442854  + 853.44019923550 *T)
    sumX3  +=       2.279 * cos(4.63709500769  + 458.81015062800 *T)
    sumX3  +=       2.147 * cos(2.63611189369  + 175.40987728371 *T)
    sumX3  +=       2.637 * cos(3.63693499332  + 73.54094334250 *T)

    # NEPTUNE X Coord
    sumX4  = 0
    sumX4  +=    3455.306 * cos(3.61464892215  + 38.37685312130 *T)
    sumX4  +=      47.405 * cos(2.21390996774  + 0.24381748350 *T)
    sumX4  +=      21.936 * cos(2.72972488197  + 37.88921815429 *T)
    sumX4  +=      15.596 * cos(1.87854121560  + 76.50988875911 *T)
    sumX4  +=      17.186 * cos(5.53785371687  + 39.86132582961 *T)
    sumX4  +=      17.459 * cos(4.82899740364  + 36.89238041300 *T)
    sumX4  +=       4.229 * cos(1.43245860878  + 515.70768857651 *T)
    sumX4  +=       4.334 * cos(5.41648117577  + 433.95555536030 *T)
    sumX4  +=       3.547 * cos(5.75561157107  + 989.98558843089 *T)
    sumX4  +=       3.155 * cos(4.85322840051  + 467.40817033709 *T)
    sumX4  +=       3.017 * cos(0.06479449145  + 227.77000692311 *T)
    sumX4  +=       2.981 * cos(0.29920864811  + 1.72829019180 *T)
    sumX4  +=       2.295 * cos(0.13749342692  + 220.65645992230 *T)
    sumX4  +=       2.296 * cos(4.70646260044  + 206.42936592071 *T)

    # NEPTUNE X Coord
    sumX5  = 0
    sumX5  +=      26.291 * cos(3.71724730200  + 38.37685312130 *T)

    # NEPTUNE Y Coord
    sumY0  = 0
    sumY0  +=  3005973100.580 * cos(3.74109000403  + 38.37685312130 *T)
    sumY0  +=  40567587.218 * cos(2.41070337452  + 0.24381748350 *T)
    sumY0  +=  13506026.414 * cos(1.92976188293  + 76.50988875911 *T)
    sumY0  +=  15716341.901 * cos(4.82548976006  + 36.89238041300 *T)
    sumY0  +=  14935642.614 * cos(5.79716600101  + 39.86132582961 *T)
    sumY0  +=  2590782.232 * cos(0.42530135542  + 1.72829019180 *T)
    sumY0  +=  1073890.204 * cos(3.81397520876  + 75.02541605080 *T)
    sumY0  +=  816388.197 * cos(5.49424416077  + 3.21276290011 *T)
    sumY0  +=  702768.075 * cos(6.16602540157  + 35.40790770471 *T)
    sumY0  +=  687594.822 * cos(2.29155372023  + 37.88921815429 *T)
    sumY0  +=  565555.652 * cos(4.41864141199  + 41.34579853790 *T)
    sumY0  +=  495650.075 * cos(5.31196432386  + 529.93478257810 *T)
    sumY0  +=  306025.380 * cos(5.11155686178  + 73.54094334250 *T)
    sumY0  +=  272446.904 * cos(5.58643013675  + 213.54291292150 *T)
    sumY0  +=  135892.298 * cos(3.97575347243  + 77.99436146740 *T)
    sumY0  +=  122117.697 * cos(2.87943509460  + 34.92027273770 *T)
    sumY0  +=   90968.285 * cos(0.11807115994  + 114.64292439690 *T)
    sumY0  +=   68915.400 * cos(4.26390741720  + 4.69723560840 *T)
    sumY0  +=   40370.680 * cos(1.09050058383  + 33.92343499640 *T)
    sumY0  +=   28891.307 * cos(3.21868082836  + 42.83027124621 *T)
    sumY0  +=   29247.752 * cos(0.05239890051  + 72.05647063421 *T)
    sumY0  +=   25576.289 * cos(3.05422599686  + 71.56883566720 *T)
    sumY0  +=   20517.968 * cos(4.12700709797  + 33.43580002939 *T)
    sumY0  +=   12614.154 * cos(1.99850111659  + 113.15845168861 *T)
    sumY0  +=   12788.929 * cos(1.16690001367  + 111.67397898031 *T)
    sumY0  +=   12013.477 * cos(5.66154697546  + 1059.62574767270 *T)
    sumY0  +=    9854.638 * cos(1.82793273920  + 36.40474544600 *T)
    sumY0  +=    8385.825 * cos(3.22321843541  + 108.21739859670 *T)
    sumY0  +=    7577.585 * cos(4.81209675667  + 426.84200835950 *T)
    sumY0  +=    6452.053 * cos(3.05476893393  + 6.18170831670 *T)
    sumY0  +=    6551.074 * cos(3.48963683470  + 1.24065522479 *T)
    sumY0  +=    4652.534 * cos(4.81582901104  + 37.85558825950 *T)
    sumY0  +=    4732.958 * cos(2.52632268239  + 79.47883417571 *T)
    sumY0  +=    4557.247 * cos(5.80951559837  + 38.89811798311 *T)
    sumY0  +=    4322.550 * cos(0.80665146695  + 38.32866901151 *T)
    sumY0  +=    4315.539 * cos(3.53393508109  + 38.42503723110 *T)
    sumY0  +=    4089.036 * cos(0.42349431022  + 37.41364527480 *T)
    sumY0  +=    4248.658 * cos(4.06300076615  + 28.81562556571 *T)
    sumY0  +=    4622.142 * cos(4.31075084247  + 70.08436295889 *T)
    sumY0  +=    3926.447 * cos(3.91895428213  + 39.34006096781 *T)
    sumY0  +=    3148.422 * cos(0.47516466537  + 76.02225379210 *T)
    sumY0  +=    3940.981 * cos(3.86846009370  + 98.65617104110 *T)
    sumY0  +=    3323.363 * cos(3.11696612599  + 4.43660317750 *T)
    sumY0  +=    3282.964 * cos(4.38630915294  + 39.37369086260 *T)
    sumY0  +=    3110.464 * cos(0.27337264525  + 47.93808067690 *T)
    sumY0  +=    2927.062 * cos(1.26687681282  + 70.57199792590 *T)
    sumY0  +=    2748.919 * cos(2.29910620256  + 32.43896228810 *T)
    sumY0  +=    3316.668 * cos(3.39273716880  + 144.86596152620 *T)
    sumY0  +=    2822.405 * cos(5.35210680933  + 31.95132732110 *T)
    sumY0  +=    2695.972 * cos(2.28196668869  + 110.18950627200 *T)
    sumY0  +=    2522.990 * cos(6.23388252645  + 311.95526647910 *T)
    sumY0  +=    1888.129 * cos(1.63385050550  + 35.92917256650 *T)
    sumY0  +=    1648.229 * cos(2.49960621702  + 30.30009827400 *T)
    sumY0  +=    1826.545 * cos(2.00941496239  + 44.31474395451 *T)
    sumY0  +=    1956.241 * cos(2.57436514192  + 206.42936592071 *T)
    sumY0  +=    1681.257 * cos(2.70480495091  + 40.82453367610 *T)
    sumY0  +=    1533.383 * cos(5.88971111646  + 38.26497853671 *T)
    sumY0  +=    1893.076 * cos(5.46256301015  + 220.65645992230 *T)
    sumY0  +=    1527.526 * cos(4.73412536340  + 38.48872770590 *T)
    sumY0  +=    2085.691 * cos(6.28187170642  + 149.80701461810 *T)
    sumY0  +=    2070.612 * cos(4.39661439400  + 136.78920667889 *T)
    sumY0  +=    1535.699 * cos(2.18492948354  + 73.05330837550 *T)
    sumY0  +=    1667.976 * cos(4.48792091670  + 106.73292588839 *T)
    sumY0  +=    1289.620 * cos(1.82629228420  + 46.45360796860 *T)
    sumY0  +=    1559.811 * cos(5.27109740006  + 38.11622069041 *T)
    sumY0  +=    1545.705 * cos(5.35267674075  + 38.63748555220 *T)
    sumY0  +=    1435.033 * cos(5.44094847718  + 522.82123557730 *T)
    sumY0  +=    1406.206 * cos(2.04637394879  + 537.04832957890 *T)
    sumY0  +=    1256.446 * cos(1.13828126057  + 34.18406742730 *T)
    sumY0  +=    1387.973 * cos(2.14763765402  + 116.12739710521 *T)
    sumY0  +=    1457.739 * cos(3.56061267693  + 181.51452445570 *T)
    sumY0  +=    1228.429 * cos(1.21566711155  + 72.31710306511 *T)
    sumY0  +=    1140.665 * cos(5.53723346032  + 7.83293736379 *T)
    sumY0  +=    1080.801 * cos(3.18403832376  + 42.56963881530 *T)
    sumY0  +=    1201.409 * cos(2.31627619186  + 2.72512793310 *T)
    sumY0  +=    1228.671 * cos(1.08170099047  + 148.32254190981 *T)
    sumY0  +=     722.014 * cos(4.59727081765  + 152.77596003471 *T)
    sumY0  +=     608.545 * cos(2.92457352888  + 35.45609181450 *T)
    sumY0  +=     722.865 * cos(4.66419895504  + 143.38148881789 *T)
    sumY0  +=     632.820 * cos(1.84622497362  + 7.66618102501 *T)
    sumY0  +=     642.369 * cos(5.54570420373  + 68.59989025060 *T)
    sumY0  +=     553.789 * cos(1.41527095431  + 41.29761442810 *T)
    sumY0  +=     682.276 * cos(3.72885979362  + 218.16308738520 *T)
    sumY0  +=     463.186 * cos(1.17340921668  + 31.78457098230 *T)
    sumY0  +=     521.560 * cos(1.91893273311  + 0.71939036300 *T)
    sumY0  +=     437.892 * cos(6.01046620661  + 1589.31671276730 *T)
    sumY0  +=     398.091 * cos(0.79544793472  + 6.34846465550 *T)
    sumY0  +=     384.065 * cos(3.15552603467  + 44.96913526031 *T)
    sumY0  +=     395.583 * cos(3.48448044710  + 108.70503356371 *T)
    sumY0  +=     327.446 * cos(4.26279342171  + 60.52313540329 *T)
    sumY0  +=     358.824 * cos(0.28673200218  + 30.46685461280 *T)
    sumY0  +=     315.179 * cos(1.74548132888  + 74.53778108379 *T)
    sumY0  +=     343.384 * cos(0.17566264278  + 0.76508234530 *T)
    sumY0  +=     399.611 * cos(3.76461168231  + 31.26330612050 *T)
    sumY0  +=     314.611 * cos(1.41723391958  + 419.72846135871 *T)
    sumY0  +=     347.596 * cos(4.83723596339  + 180.03005174739 *T)
    sumY0  +=     382.279 * cos(1.78844211361  + 487.12132627930 *T)
    sumY0  +=     300.918 * cos(2.47842979419  + 69.08752521760 *T)
    sumY0  +=     340.448 * cos(2.33467216950  + 146.83806920150 *T)
    sumY0  +=     298.710 * cos(3.60933906971  + 84.58664360640 *T)
    sumY0  +=     290.629 * cos(0.17794042595  + 110.45013870291 *T)
    sumY0  +=     336.211 * cos(0.57735466050  + 45.49040012211 *T)
    sumY0  +=     305.606 * cos(4.06185849299  + 640.14110379750 *T)
    sumY0  +=     333.702 * cos(3.90017949649  + 254.81165031470 *T)
    sumY0  +=     268.060 * cos(1.73772568979  + 37.00425499760 *T)
    sumY0  +=     264.760 * cos(2.55644426185  + 39.74945124500 *T)
    sumY0  +=     315.240 * cos(1.15162155813  + 388.70897272171 *T)
    sumY0  +=     227.098 * cos(6.16236913832  + 273.82223084130 *T)
    sumY0  +=     306.112 * cos(0.18265553789  + 6283.31966747490 *T)
    sumY0  +=     284.373 * cos(1.79060192705  + 12.77399045571 *T)
    sumY0  +=     221.105 * cos(5.08019996555  + 213.05527795450 *T)
    sumY0  +=     242.568 * cos(0.49358017330  + 14.25846316400 *T)
    sumY0  +=     241.087 * cos(5.73194988554  + 105.24845318010 *T)
    sumY0  +=     226.136 * cos(1.26736306174  + 80.96330688400 *T)
    sumY0  +=     245.904 * cos(5.25701422242  + 27.33115285740 *T)
    sumY0  +=     265.825 * cos(5.68032293038  + 944.73900579230 *T)
    sumY0  +=     207.893 * cos(3.50733218656  + 30.95448957981 *T)
    sumY0  +=     214.661 * cos(1.08322862012  + 316.63568714010 *T)
    sumY0  +=     190.638 * cos(0.75588071076  + 69.39634175830 *T)
    sumY0  +=     246.295 * cos(3.55718121196  + 102.84895673509 *T)
    sumY0  +=     202.915 * cos(2.17108892757  + 415.04804069769 *T)
    sumY0  +=     176.465 * cos(4.85970190916  + 36.78050582840 *T)
    sumY0  +=     193.886 * cos(4.92555932032  + 174.92224231670 *T)
    sumY0  +=     175.209 * cos(5.83814591553  + 39.97320041421 *T)
    sumY0  +=     177.868 * cos(5.01003024093  + 216.67861467689 *T)
    sumY0  +=     138.494 * cos(3.88186287753  + 75.98862389731 *T)
    sumY0  +=     152.234 * cos(3.24582472092  + 11.28951774740 *T)
    sumY0  +=     147.648 * cos(0.11464073993  + 151.29148732640 *T)
    sumY0  +=     156.202 * cos(5.22332207731  + 146.35043423450 *T)
    sumY0  +=     152.289 * cos(1.64425361444  + 23.87457247379 *T)
    sumY0  +=     177.911 * cos(1.60563922042  + 10213.52936369450 *T)
    sumY0  +=     162.474 * cos(2.56271758700  + 63.97971578690 *T)
    sumY0  +=     121.226 * cos(3.53504653517  + 38.16440480021 *T)
    sumY0  +=     129.049 * cos(2.23605274276  + 37.10482873410 *T)
    sumY0  +=     120.334 * cos(0.80557581782  + 38.58930144240 *T)
    sumY0  +=     168.977 * cos(4.06631471177  + 291.46021324420 *T)
    sumY0  +=     121.138 * cos(6.20896007337  + 33.26904369061 *T)
    sumY0  +=     129.366 * cos(0.79823378243  + 45.79921666280 *T)
    sumY0  +=     144.682 * cos(5.34262329825  + 49.42255338521 *T)
    sumY0  +=     122.915 * cos(2.10353894081  + 39.64887750850 *T)
    sumY0  +=     113.400 * cos(5.13399083059  + 83.10217089810 *T)
    sumY0  +=     154.892 * cos(0.17909436973  + 77.47309660560 *T)
    sumY0  +=     106.737 * cos(2.14516700774  + 4.86399194720 *T)
    sumY0  +=     104.756 * cos(4.39192437833  + 43.48466255200 *T)
    sumY0  +=     125.142 * cos(1.11541363958  + 4.20960064140 *T)
    sumY0  +=     103.541 * cos(3.68555108825  + 41.08516610701 *T)
    sumY0  +=     133.573 * cos(5.49226848460  + 182.99899716400 *T)
    sumY0  +=     103.627 * cos(0.72176478921  + 35.66854013560 *T)
    sumY0  +=     116.874 * cos(3.84298763857  + 62.49524307860 *T)
    sumY0  +=      98.063 * cos(1.68574394986  + 9.80504503910 *T)
    sumY0  +=     111.411 * cos(5.91424942326  + 141.89701610960 *T)
    sumY0  +=     114.294 * cos(3.99149302956  + 633.02755679670 *T)
    sumY0  +=     104.705 * cos(4.68992506677  + 433.95555536030 *T)
    sumY0  +=     121.306 * cos(3.01971978017  + 40.85816357090 *T)
    sumY0  +=      96.954 * cos(4.60293836623  + 1052.51220067191 *T)
    sumY0  +=      85.104 * cos(3.21938589681  + 36.67993209190 *T)
    sumY0  +=      85.209 * cos(1.23258290286  + 105.76971804189 *T)
    sumY0  +=      85.291 * cos(4.16574840077  + 109.70187130500 *T)
    sumY0  +=      83.260 * cos(1.57705309557  + 529.44714761109 *T)
    sumY0  +=      80.200 * cos(1.12120137014  + 40.07377415071 *T)
    sumY0  +=     107.927 * cos(4.72809768120  + 1162.71852189130 *T)
    sumY0  +=      95.241 * cos(5.18181889280  + 253.32717760639 *T)
    sumY0  +=      89.535 * cos(1.68098752171  + 32.96022714990 *T)
    sumY0  +=      89.793 * cos(1.19350927545  + 65.46418849521 *T)
    sumY0  +=      72.027 * cos(4.82605474115  + 36.94056452280 *T)
    sumY0  +=      80.381 * cos(0.49818419813  + 67.11541754230 *T)
    sumY0  +=      99.502 * cos(4.10090280616  + 453.18107633550 *T)
    sumY0  +=      88.685 * cos(6.05087292163  + 251.67594855930 *T)
    sumY0  +=      94.971 * cos(5.68681980258  + 219.64756009350 *T)
    sumY0  +=      77.015 * cos(3.73580633492  + 5.66044345490 *T)
    sumY0  +=      69.098 * cos(3.42063825132  + 22.39009976550 *T)
    sumY0  +=      79.079 * cos(5.69904586697  + 44.48150029329 *T)
    sumY0  +=      69.159 * cos(2.38821700872  + 1066.73929467350 *T)
    sumY0  +=      64.446 * cos(2.45996531968  + 66.94866120350 *T)
    sumY0  +=      88.518 * cos(4.23259429373  + 328.10877617370 *T)
    sumY0  +=      65.817 * cos(6.14060374302  + 36.37111555120 *T)
    sumY0  +=      71.422 * cos(2.66025338551  + 43.79347909271 *T)
    sumY0  +=      63.298 * cos(0.64067085772  + 9.15065373330 *T)
    sumY0  +=      77.320 * cos(1.83922353490  + 97.17169833279 *T)
    sumY0  +=      73.912 * cos(3.29477271110  + 2.67694382330 *T)
    sumY0  +=      73.965 * cos(3.98729910569  + 2.95213046920 *T)
    sumY0  +=      56.194 * cos(2.88777806681  + 949.41942645330 *T)
    sumY0  +=      59.173 * cos(2.98452265287  + 100.14064374939 *T)
    sumY0  +=      67.507 * cos(2.37620743833  + 7.14491616321 *T)
    sumY0  +=      71.718 * cos(2.50472239979  + 2.20386307129 *T)
    sumY0  +=      63.606 * cos(3.60095909928  + 25.84668014910 *T)
    sumY0  +=      71.523 * cos(3.62910110768  + 662.28738607949 *T)
    sumY0  +=      57.219 * cos(5.59723911326  + 15.74293587230 *T)
    sumY0  +=      50.322 * cos(5.79549186801  + 37.15301284391 *T)
    sumY0  +=      66.615 * cos(1.85382631980  + 846.32665223470 *T)
    sumY0  +=      56.220 * cos(6.09466556848  + 178.54557903910 *T)
    sumY0  +=      67.883 * cos(2.31467094623  + 224.58861318540 *T)
    sumY0  +=      57.761 * cos(3.59414048269  + 145.35359649321 *T)
    sumY0  +=      53.973 * cos(4.68325129609  + 107.22056085540 *T)
    sumY0  +=      57.588 * cos(0.13600413206  + 25.35904518210 *T)
    sumY0  +=      49.026 * cos(5.99075269953  + 19.25439801010 *T)
    sumY0  +=      63.036 * cos(5.86840206029  + 256.29612302300 *T)
    sumY0  +=      45.304 * cos(5.57731819351  + 4.17597074660 *T)
    sumY0  +=      45.669 * cos(0.60467903265  + 117.61186981350 *T)
    sumY0  +=      52.821 * cos(5.35013106251  + 289.97574053589 *T)
    sumY0  +=      44.016 * cos(0.68418990599  + 32.74777882880 *T)
    sumY0  +=      42.933 * cos(1.50265323283  + 28.98238190449 *T)
    sumY0  +=      38.369 * cos(5.07841615051  + 39.60069339870 *T)
    sumY0  +=      38.805 * cos(2.55324300089  + 103.33659170210 *T)
    sumY0  +=      37.679 * cos(4.97176992254  + 9.31741007210 *T)
    sumY0  +=      40.292 * cos(1.32694372497  + 111.18634401329 *T)
    sumY0  +=      50.011 * cos(4.62887079290  + 221.61966776881 *T)
    sumY0  +=      37.056 * cos(3.05929116522  + 8.32057233081 *T)
    sumY0  +=      36.562 * cos(1.75628268654  + 448.98829064149 *T)
    sumY0  +=      44.628 * cos(5.39841763538  + 525.25436191710 *T)
    sumY0  +=      38.213 * cos(4.99269276748  + 75.54668091261 *T)
    sumY0  +=      45.963 * cos(2.49324091181  + 183.48663213100 *T)
    sumY0  +=      48.222 * cos(4.38408318526  + 364.75733910320 *T)
    sumY0  +=      38.164 * cos(3.66287516322  + 44.00592741381 *T)
    sumY0  +=      47.779 * cos(4.62193118070  + 3340.85624418330 *T)
    sumY0  +=      42.228 * cos(4.07611308238  + 77.03115362090 *T)
    sumY0  +=      35.247 * cos(4.92005743728  + 34.75351639890 *T)
    sumY0  +=      46.804 * cos(5.53981795511  + 33.69643246030 *T)
    sumY0  +=      34.352 * cos(5.79527968049  + 33.71098667531 *T)
    sumY0  +=      34.949 * cos(3.58463727178  + 3.37951923889 *T)
    sumY0  +=      36.030 * cos(0.60196271868  + 71.09326278771 *T)
    sumY0  +=      38.112 * cos(0.94232057009  + 45.96597300160 *T)
    sumY0  +=      33.119 * cos(3.70714424363  + 7.35736448430 *T)
    sumY0  +=      32.049 * cos(3.04761071508  + 34.44469985821 *T)
    sumY0  +=      31.910 * cos(0.20811343013  + 81.61769818981 *T)
    sumY0  +=      38.697 * cos(1.09830424446  + 184.97110483931 *T)
    sumY0  +=      41.486 * cos(4.15630010756  + 310.47079377080 *T)
    sumY0  +=      38.631 * cos(0.74636164144  + 50.90702609350 *T)
    sumY0  +=      42.711 * cos(0.62152472293  + 1021.49271203491 *T)
    sumY0  +=      32.006 * cos(5.68829457469  + 42.00018984371 *T)
    sumY0  +=      38.436 * cos(5.02591476913  + 5.92107588581 *T)
    sumY0  +=      38.880 * cos(1.72301566300  + 76.55807286891 *T)
    sumY0  +=      41.190 * cos(3.00922391966  + 563.87503252191 *T)
    sumY0  +=      29.786 * cos(2.57644898724  + 77.50672650040 *T)
    sumY0  +=      40.604 * cos(6.04591617823  + 292.94468595250 *T)
    sumY0  +=      35.275 * cos(3.24596926614  + 304.84171947829 *T)
    sumY0  +=      38.242 * cos(1.23011716621  + 17.76992530181 *T)
    sumY0  +=      34.445 * cos(6.05203741506  + 319.06881347989 *T)
    sumY0  +=      28.725 * cos(0.80354919578  + 67.63668240410 *T)
    sumY0  +=      32.809 * cos(0.86662032393  + 91.54262404029 *T)
    sumY0  +=      38.880 * cos(5.27893548994  + 76.46170464930 *T)
    sumY0  +=      30.731 * cos(3.65388358465  + 67.60305250931 *T)
    sumY0  +=      28.459 * cos(4.82537806886  + 43.04271956730 *T)
    sumY0  +=      35.368 * cos(5.14016182775  + 313.43973918739 *T)
    sumY0  +=      35.703 * cos(4.78026134196  + 258.26823069831 *T)
    sumY0  +=      32.317 * cos(0.72991843715  + 78.95756931390 *T)
    sumY0  +=      29.243 * cos(5.01962947605  + 61.01077037031 *T)
    sumY0  +=      26.235 * cos(2.30979326374  + 137.27684164590 *T)
    sumY0  +=      26.519 * cos(4.63187110201  + 57.49930823250 *T)
    sumY0  +=      24.931 * cos(1.02449436120  + 42.99702758500 *T)
    sumY0  +=      27.608 * cos(0.68443037331  + 103.76398047180 *T)
    sumY0  +=      28.680 * cos(6.22569747241  + 215.19414196860 *T)
    sumY0  +=      25.052 * cos(0.98956881727  + 350.08830211689 *T)
    sumY0  +=      31.386 * cos(2.53676810018  + 22.22334342671 *T)
    sumY0  +=      27.545 * cos(6.02026727313  + 100.62827871640 *T)
    sumY0  +=      22.617 * cos(1.89172143755  + 36.84419630320 *T)
    sumY0  +=      24.909 * cos(4.92089915310  + 24.36220744081 *T)
    sumY0  +=      26.216 * cos(3.37729185316  + 491.80174694030 *T)
    sumY0  +=      28.040 * cos(1.26215532584  + 11.55015017831 *T)
    sumY0  +=      23.047 * cos(2.67490790904  + 35.51978228931 *T)
    sumY0  +=      27.067 * cos(5.52626880418  + 326.62430346539 *T)
    sumY0  +=      26.192 * cos(0.78880180701  + 20.73887071840 *T)
    sumY0  +=      23.134 * cos(1.02405904726  + 68.43313391180 *T)
    sumY0  +=      21.423 * cos(5.59061648293  + 39.90950993941 *T)
    sumY0  +=      25.696 * cos(5.03652999676  + 186.45557754760 *T)
    sumY0  +=      26.985 * cos(1.96185307311  + 69.60879007940 *T)
    sumY0  +=      23.284 * cos(1.14508397457  + 79.43065006591 *T)
    sumY0  +=      22.894 * cos(5.33085965806  + 227.77000692311 *T)
    sumY0  +=      22.482 * cos(5.43588494929  + 39.81314171980 *T)
    sumY0  +=      23.480 * cos(5.96723336237  + 30.98811947460 *T)
    sumY0  +=      20.858 * cos(1.66497796416  + 41.23392395330 *T)
    sumY0  +=      20.327 * cos(5.86806874135  + 39.03124442710 *T)
    sumY0  +=      20.327 * cos(4.75570383217  + 37.72246181551 *T)
    sumY0  +=      22.639 * cos(1.78594954268  + 0.98002279390 *T)
    sumY0  +=      22.639 * cos(4.92754219627  + 1.46765776091 *T)
    sumY0  +=      19.139 * cos(1.60585998738  + 205.94173095370 *T)
    sumY0  +=      19.118 * cos(0.05485235310  + 2119.00767786191 *T)
    sumY0  +=      25.698 * cos(4.54722652155  + 401.40590203270 *T)
    sumY0  +=      21.582 * cos(5.86612346662  + 81.13006322279 *T)
    sumY0  +=      25.509 * cos(6.21909191790  + 329.59324888200 *T)
    sumY0  +=      24.296 * cos(3.68761645751  + 62.00760811160 *T)
    sumY0  +=      23.969 * cos(2.45967218561  + 135.30473397060 *T)
    sumY0  +=      20.599 * cos(6.09025723810  + 491.31411197330 *T)
    sumY0  +=      16.829 * cos(4.06509805545  + 3.16457879030 *T)
    sumY0  +=      20.030 * cos(2.45066995549  + 217.47506618460 *T)
    sumY0  +=      20.377 * cos(5.60617244490  + 209.61075965840 *T)
    sumY0  +=      17.251 * cos(1.00239992257  + 350.57593708390 *T)
    sumY0  +=      19.625 * cos(1.41143867859  + 129.67565967810 *T)
    sumY0  +=      22.707 * cos(0.97867191772  + 1436.29693524910 *T)
    sumY0  +=      17.142 * cos(4.71740830608  + 29.47001687150 *T)
    sumY0  +=      16.188 * cos(3.33781568208  + 39.00999256771 *T)
    sumY0  +=      16.188 * cos(1.00277158426  + 37.74371367490 *T)
    sumY0  +=      20.858 * cos(3.10425391408  + 58.98378094080 *T)
    sumY0  +=      15.747 * cos(0.31821188942  + 154.26043274300 *T)
    sumY0  +=      19.714 * cos(5.04477015525  + 294.91679362781 *T)
    sumY0  +=      19.078 * cos(1.16675280621  + 202.49721265760 *T)
    sumY0  +=      21.530 * cos(4.95075882360  + 114.15528942990 *T)
    sumY0  +=      19.068 * cos(3.39813326972  + 138.27367938720 *T)
    sumY0  +=      18.723 * cos(4.64325038338  + 323.74923414091 *T)
    sumY0  +=      18.916 * cos(3.89922448206  + 40.38259069140 *T)
    sumY0  +=      15.843 * cos(4.98899291519  + 72.57773549600 *T)
    sumY0  +=      20.695 * cos(3.75000782446  + 86.07111631471 *T)
    sumY0  +=      15.895 * cos(4.16120885988  + 736.12033101530 *T)
    sumY0  +=      14.983 * cos(0.56469438588  + 743.23387801611 *T)
    sumY0  +=      14.928 * cos(5.49703861671  + 34.23225153711 *T)
    sumY0  +=      15.461 * cos(4.47518787654  + 20.85074530300 *T)
    sumY0  +=      16.206 * cos(4.48895168117  + 138.76131435421 *T)
    sumY0  +=      15.978 * cos(5.56972981392  + 515.70768857651 *T)
    sumY0  +=      14.173 * cos(1.42508198977  + 99.14380600810 *T)
    sumY0  +=      18.749 * cos(1.80466304752  + 54.53036281590 *T)
    sumY0  +=      13.971 * cos(3.54176522468  + 76.77052119001 *T)
    sumY0  +=      13.971 * cos(3.46018552739  + 76.24925632820 *T)
    sumY0  +=      14.035 * cos(6.02847994013  + 235.68919520349 *T)
    sumY0  +=      18.894 * cos(3.02786191873  + 31.47575444160 *T)
    sumY0  +=      14.967 * cos(5.68342907224  + 52.39149880180 *T)
    sumY0  +=      17.392 * cos(0.12268817693  + 74.06220820430 *T)
    sumY0  +=      14.788 * cos(3.43864596335  + 56.01483552421 *T)
    sumY0  +=      15.758 * cos(1.26184897401  + 208.86249226050 *T)
    sumY0  +=      12.911 * cos(5.12673395733  + 42.52145470550 *T)
    sumY0  +=      14.356 * cos(0.18539168671  + 251.84270489810 *T)
    sumY0  +=      16.266 * cos(3.39270678896  + 853.44019923550 *T)
    sumY0  +=      15.513 * cos(2.59603540214  + 59.03866269500 *T)
    sumY0  +=      12.783 * cos(0.77187700977  + 107.52937739611 *T)
    sumY0  +=      16.075 * cos(0.02096626523  + 366.24181181149 *T)
    sumY0  +=      14.277 * cos(3.31408666848  + 19.36627259471 *T)
    sumY0  +=      14.742 * cos(6.26354356543  + 82.44777959230 *T)
    sumY0  +=      15.111 * cos(5.70708654477  + 363.27286639489 *T)
    sumY0  +=      14.981 * cos(1.17119164980  + 82.61453593110 *T)
    sumY0  +=      14.840 * cos(5.34075197148  + 44.05411152360 *T)
    sumY0  +=      15.592 * cos(5.74434423333  + 8.62938887150 *T)
    sumY0  +=      14.568 * cos(0.45025790013  + 73.80157577341 *T)
    sumY0  +=      12.251 * cos(5.90063123167  + 47.28368937111 *T)
    sumY0  +=      11.447 * cos(5.62613164770  + 175.40987728371 *T)
    sumY0  +=      13.900 * cos(0.93353054847  + 700.42042171730 *T)
    sumY0  +=      15.583 * cos(5.46046493452  + 837.45344587970 *T)
    sumY0  +=      12.109 * cos(0.53062884942  + 33.00841125970 *T)
    sumY0  +=      12.379 * cos(0.87778018320  + 140.41254340130 *T)
    sumY0  +=      11.481 * cos(3.65591005670  + 39.20693452380 *T)
    sumY0  +=      11.481 * cos(0.68467720964  + 37.54677171881 *T)
    sumY0  +=      11.452 * cos(5.92350892067  + 529.41351771630 *T)
    sumY0  +=      10.981 * cos(1.58931744102  + 63.49208081989 *T)
    sumY0  +=      12.137 * cos(0.75938098769  + 42.30900638440 *T)
    sumY0  +=      13.771 * cos(2.92318261793  + 76.62176334371 *T)
    sumY0  +=      11.036 * cos(1.59378256377  + 530.45604743991 *T)
    sumY0  +=      11.537 * cos(2.72370023352  + 199.31581891990 *T)
    sumY0  +=      11.189 * cos(1.67388131435  + 80.13322548150 *T)
    sumY0  +=      12.835 * cos(2.83910944143  + 38.85242600079 *T)
    sumY0  +=      12.879 * cos(0.03161787960  + 5.69407334969 *T)
    sumY0  +=      13.663 * cos(4.69897705757  + 438.05446496220 *T)
    sumY0  +=      10.132 * cos(2.80479631986  + 187.94005025590 *T)
    sumY0  +=      12.619 * cos(3.09097380707  + 65.20355606430 *T)
    sumY0  +=      10.088 * cos(1.41143864412  + 26.58288545949 *T)
    sumY0  +=      11.959 * cos(1.19714206196  + 64.71592109730 *T)
    sumY0  +=      11.578 * cos(5.81790016857  + 275.30670354960 *T)
    sumY0  +=      12.795 * cos(1.66756565053  + 17.88179988640 *T)
    sumY0  +=      13.771 * cos(4.07876849292  + 76.39801417450 *T)
    sumY0  +=      10.044 * cos(1.67224715151  + 147.83490694279 *T)
    sumY0  +=      13.632 * cos(1.29603813384  + 45.27795180100 *T)
    sumY0  +=      11.660 * cos(4.22880871720  + 143.90275367970 *T)
    sumY0  +=       9.938 * cos(5.79050109000  + 6.86972951729 *T)
    sumY0  +=       9.719 * cos(4.48706829937  + 956.53297345411 *T)
    sumY0  +=      11.441 * cos(5.32553485636  + 533.86693584120 *T)
    sumY0  +=      10.240 * cos(1.34767099242  + 80.70267445310 *T)
    sumY0  +=      10.031 * cos(3.80995841826  + 43.74529498291 *T)
    sumY0  +=      10.063 * cos(1.05825122331  + 0.27744737829 *T)
    sumY0  +=      11.428 * cos(2.19933512981  + 526.00262931501 *T)
    sumY0  +=       9.279 * cos(1.45482205447  + 79.64559051450 *T)
    sumY0  +=      10.172 * cos(0.89461094062  + 568.06781821590 *T)
    sumY0  +=       9.198 * cos(0.36520539350  + 112.67081672160 *T)
    sumY0  +=       9.831 * cos(4.06082180622  + 20.90562705720 *T)
    sumY0  +=       9.830 * cos(1.93960888370  + 544.16187657970 *T)
    sumY0  +=       8.646 * cos(6.06265529598  + 30.77567115350 *T)
    sumY0  +=       9.315 * cos(1.72769398395  + 65.63094483399 *T)
    sumY0  +=       9.201 * cos(1.66299093770  + 184.48346987229 *T)
    sumY0  +=       8.674 * cos(3.58250353029  + 624.15435044170 *T)
    sumY0  +=      10.739 * cos(5.20958133978  + 331.56535655731 *T)
    sumY0  +=       9.612 * cos(3.81549627985  + 182.00215942271 *T)
    sumY0  +=       8.664 * cos(4.05357381243  + 1479.11039154791 *T)
    sumY0  +=       8.092 * cos(4.08843223898  + 6.83609962250 *T)
    sumY0  +=      10.092 * cos(0.00357719036  + 419.24082639170 *T)
    sumY0  +=      10.233 * cos(0.16992310980  + 402.89037474099 *T)
    sumY0  +=       8.502 * cos(3.60646753260  + 17.39416491939 *T)
    sumY0  +=      10.189 * cos(1.01906004060  + 21.70207856490 *T)
    sumY0  +=       9.829 * cos(3.66564448678  + 121.23520653590 *T)
    sumY0  +=       8.406 * cos(4.04270651029  + 376.91500505990 *T)
    sumY0  +=       8.060 * cos(4.05224638436  + 415.79630809560 *T)
    sumY0  +=       9.455 * cos(1.63876624122  + 167.80869531589 *T)
    sumY0  +=       7.941 * cos(6.14526289331  + 526.75338884040 *T)
    sumY0  +=       7.870 * cos(1.33260101318  + 533.11617631580 *T)
    sumY0  +=       7.695 * cos(2.49810660877  + 906.60597015449 *T)
    sumY0  +=       7.862 * cos(5.62722995177  + 1265.81129610991 *T)
    sumY0  +=       8.062 * cos(5.84124471296  + 105.73608814710 *T)
    sumY0  +=       8.904 * cos(5.87904582316  + 399.92142932440 *T)
    sumY0  +=       8.050 * cos(4.85961454632  + 143.86912378490 *T)
    sumY0  +=       9.102 * cos(3.20438608836  + 348.17644063891 *T)
    sumY0  +=       7.137 * cos(5.97349520503  + 117.56368570370 *T)
    sumY0  +=       7.076 * cos(4.77037120491  + 26.84351789039 *T)
    sumY0  +=       8.418 * cos(6.19754313245  + 77.73372903651 *T)
    sumY0  +=       8.257 * cos(6.01515603184  + 117.77862615229 *T)
    sumY0  +=       7.868 * cos(0.36467826737  + 288.49126782760 *T)
    sumY0  +=       8.093 * cos(5.12697881207  + 1692.40948698591 *T)
    sumY0  +=       6.910 * cos(5.16028730720  + 216.72430665921 *T)
    sumY0  +=       7.092 * cos(1.58416634960  + 452.65981147369 *T)
    sumY0  +=       7.060 * cos(3.50187723218  + 453.70234119730 *T)
    sumY0  +=       8.233 * cos(5.07959772857  + 480.00777927849 *T)
    sumY0  +=       6.772 * cos(2.89170457209  + 210.36151918381 *T)
    sumY0  +=       7.025 * cos(6.13907268455  + 55.90296093960 *T)
    sumY0  +=       8.356 * cos(3.67079730328  + 95.73540973430 *T)
    sumY0  +=       7.404 * cos(5.71532443096  + 75.28604848170 *T)
    sumY0  +=       6.839 * cos(2.57023077532  + 41.51255487670 *T)
    sumY0  +=       7.909 * cos(0.07288588503  + 36.63174798211 *T)
    sumY0  +=       7.909 * cos(1.12610872772  + 40.12195826051 *T)
    sumY0  +=       6.362 * cos(4.97586429633  + 29.99128173331 *T)
    sumY0  +=       6.712 * cos(2.41218446093  + 133.82026126229 *T)
    sumY0  +=       7.571 * cos(1.24658605384  + 23.70781613500 *T)
    sumY0  +=       6.677 * cos(4.81403056382  + 1.20702533000 *T)
    sumY0  +=       7.600 * cos(1.64374414108  + 494.23487328010 *T)
    sumY0  +=       8.009 * cos(1.96165940869  + 170.72945662269 *T)
    sumY0  +=       7.584 * cos(1.33750538789  + 119.26309886060 *T)
    sumY0  +=       6.599 * cos(0.68440943828  + 32.22651396700 *T)
    sumY0  +=       6.085 * cos(3.39985070945  + 322.00412900171 *T)
    sumY0  +=       5.953 * cos(0.92774911672  + 52214.18313626970 *T)
    sumY0  +=       7.827 * cos(4.85672910517  + 474.70302789170 *T)
    sumY0  +=       7.907 * cos(6.03373097658  + 485.63685357099 *T)
    sumY0  +=       7.372 * cos(3.31633214824  + 55.05162767771 *T)
    sumY0  +=       6.966 * cos(4.03472609774  + 647.25465079831 *T)
    sumY0  +=       6.266 * cos(1.06894881555  + 177.06110633080 *T)
    sumY0  +=       5.900 * cos(0.21363873876  + 52061.16335875149 *T)
    sumY0  +=       6.221 * cos(0.78444326027  + 602.00806815971 *T)
    sumY0  +=       5.552 * cos(4.30656362928  + 223.10414047710 *T)
    sumY0  +=       5.976 * cos(3.40178743225  + 10.80188278040 *T)
    sumY0  +=       7.600 * cos(0.62565658069  + 488.60579898760 *T)
    sumY0  +=       6.831 * cos(4.75854396498  + 1582.20316576650 *T)
    sumY0  +=       5.654 * cos(1.46952482126  + 12604.52855310410 *T)
    sumY0  +=       5.798 * cos(2.70754675899  + 27.49790919620 *T)
    sumY0  +=       7.216 * cos(4.89431192173  + 739.04109232210 *T)
    sumY0  +=       6.579 * cos(2.37730114095  + 2.69149803831 *T)
    sumY0  +=       5.758 * cos(1.25264555408  + 30.03946584310 *T)
    sumY0  +=       5.270 * cos(5.03822712313  + 6166.94845288619 *T)
    sumY0  +=       7.398 * cos(2.15412967054  + 709.72101684200 *T)
    sumY0  +=       5.679 * cos(4.34696450423  + 17.22740858061 *T)
    sumY0  +=       5.205 * cos(4.18097270804  + 426.35437339250 *T)
    sumY0  +=       5.146 * cos(5.52411562781  + 46.76242450930 *T)
    sumY0  +=       5.694 * cos(4.51992731423  + 168.98435148349 *T)
    sumY0  +=       6.627 * cos(1.36429825841  + 221.13203280179 *T)
    sumY0  +=       5.443 * cos(2.77787969707  + 525.74199688410 *T)
    sumY0  +=       6.475 * cos(0.95284661304  + 591.07424248041 *T)
    sumY0  +=       4.984 * cos(0.17849511014  + 10097.15814910579 *T)
    sumY0  +=       5.318 * cos(3.65617684169  + 44.52719227561 *T)
    sumY0  +=       6.699 * cos(1.37968332714  + 2157.14071349970 *T)
    sumY0  +=       6.443 * cos(4.07988524250  + 675.04456158780 *T)
    sumY0  +=       5.078 * cos(2.53592755854  + 101.62511645769 *T)
    sumY0  +=       5.394 * cos(5.60187660250  + 368.21391948681 *T)
    sumY0  +=       5.072 * cos(4.09677163290  + 272.33775813299 *T)
    sumY0  +=       5.208 * cos(2.96070554414  + 277.27881122490 *T)
    sumY0  +=       5.332 * cos(2.85701594895  + 280.93577784210 *T)
    sumY0  +=       5.989 * cos(1.18032513011  + 93.02709674860 *T)
    sumY0  +=       6.329 * cos(2.06650240521  + 18.87863762769 *T)
    sumY0  +=       5.551 * cos(0.99966130596  + 57.38743364790 *T)
    sumY0  +=       6.471 * cos(4.75702433578  + 68.12431737110 *T)
    sumY0  +=       4.708 * cos(3.81000728156  + 95.68722562449 *T)
    sumY0  +=       5.891 * cos(4.39361748912  + 381.59542572090 *T)
    sumY0  +=       4.717 * cos(5.88762112195  + 104.28524533360 *T)
    sumY0  +=       5.675 * cos(0.14149668499  + 1165.63928319810 *T)
    sumY0  +=       5.888 * cos(2.00299136957  + 42.34263627919 *T)
    sumY0  +=       5.587 * cos(2.52090459839  + 459.60660213570 *T)
    sumY0  +=       5.456 * cos(3.07944464122  + 75.50098893029 *T)
    sumY0  +=       5.940 * cos(4.70996040917  + 6318.48375769610 *T)
    sumY0  +=       5.207 * cos(6.12213701959  + 436.56999225390 *T)
    sumY0  +=       6.160 * cos(3.18966815531  + 749.82616015511 *T)
    sumY0  +=       6.137 * cos(3.02268593798  + 713.17759722561 *T)
    sumY0  +=       4.547 * cos(3.96298179960  + 32.47259218289 *T)
    sumY0  +=       5.246 * cos(0.26649341993  + 109.96250373590 *T)
    sumY0  +=       5.244 * cos(0.76595138199  + 73.58912745230 *T)
    sumY0  +=       5.572 * cos(4.54958395511  + 102.11275142471 *T)
    sumY0  +=       5.638 * cos(6.13292790226  + 10248.69345391570 *T)
    sumY0  +=       4.513 * cos(0.05769066183  + 1272.92484311070 *T)
    sumY0  +=       4.340 * cos(3.93529499489  + 384.02855206069 *T)
    sumY0  +=       4.263 * cos(5.81710901839  + 1577.52274510549 *T)
    sumY0  +=       5.964 * cos(3.35563503899  + 786.47472308461 *T)
    sumY0  +=       4.962 * cos(1.38600480216  + 257.78059573129 *T)
    sumY0  +=       5.327 * cos(4.13135597763  + 107.74182571721 *T)
    sumY0  +=       5.572 * cos(5.58677005833  + 291.29345690540 *T)
    sumY0  +=       4.336 * cos(1.08874295813  + 53.40958840249 *T)
    sumY0  +=       4.427 * cos(1.43077618159  + 189.42452296421 *T)
    sumY0  +=       4.157 * cos(5.03727532307  + 29.50364676630 *T)
    sumY0  +=       4.646 * cos(4.44853801893  + 13285.93981804009 *T)
    sumY0  +=       5.507 * cos(2.70385106164  + 178.11819026941 *T)
    sumY0  +=       5.348 * cos(6.13707191030  + 24.88347230261 *T)
    sumY0  +=       5.339 * cos(5.48920294964  + 314.66357946480 *T)
    sumY0  +=       4.678 * cos(6.00688425085  + 1474.42997088690 *T)
    sumY0  +=       4.090 * cos(4.92713296866  + 765.38016029810 *T)
    sumY0  +=       5.008 * cos(4.28621887979  + 352.06040979221 *T)
    sumY0  +=       5.562 * cos(5.12126233744  + 6248.15557725370 *T)
    sumY0  +=       4.983 * cos(1.59156517574  + 1055.43296197871 *T)
    sumY0  +=       4.566 * cos(0.54461731254  + 325.13983075710 *T)
    sumY0  +=       5.327 * cos(0.54108371123  + 439.53893767049 *T)
    sumY0  +=       5.121 * cos(4.27746071897  + 711.69312451730 *T)
    sumY0  +=       4.181 * cos(2.68829223641  + 6606.19943734880 *T)
    sumY0  +=       4.293 * cos(3.08794166207  + 46.71424039951 *T)
    sumY0  +=       5.532 * cos(2.10559407460  + 320.03202132639 *T)
    sumY0  +=       4.492 * cos(4.81151725335  + 52177.53457334019 *T)
    sumY0  +=       4.312 * cos(6.10122311855  + 22.87773473250 *T)
    sumY0  +=       5.332 * cos(0.25990559894  + 10178.36527347330 *T)
    sumY0  +=       4.593 * cos(4.86059649000  + 1025.68549772890 *T)
    sumY0  +=       5.439 * cos(3.52367947540  + 823.12328601411 *T)
    sumY0  +=       3.870 * cos(2.70915745235  + 1596.43025976811 *T)
    sumY0  +=       3.892 * cos(0.54485159298  + 226.07308589371 *T)
    sumY0  +=       4.891 * cos(4.37893659385  + 8.14175390450 *T)
    sumY0  +=       4.689 * cos(5.09142557332  + 276.79117625789 *T)
    sumY0  +=       4.268 * cos(1.02189794794  + 374.15181032000 *T)
    sumY0  +=       3.828 * cos(3.85156237339  + 2138.23319883710 *T)
    sumY0  +=       4.592 * cos(2.30447944615  + 1376.01761732930 *T)
    sumY0  +=       4.629 * cos(5.68948058955  + 122.71967924421 *T)
    sumY0  +=       3.871 * cos(1.60468692916  + 531.41925528640 *T)
    sumY0  +=       4.995 * cos(5.03302660981  + 32.69959471901 *T)
    sumY0  +=       4.711 * cos(5.14987215661  + 52252.31617190749 *T)
    sumY0  +=       3.893 * cos(1.69554966790  + 116.29415344400 *T)
    sumY0  +=       4.481 * cos(3.09400209140  + 53.04589010760 *T)
    sumY0  +=       4.136 * cos(1.02307294098  + 503.10807963510 *T)
    sumY0  +=       4.508 * cos(2.81495366139  + 562.12992738271 *T)
    sumY0  +=       5.025 * cos(1.96944866339  + 283.38345839689 *T)
    sumY0  +=       4.789 * cos(1.11612617111  + 627.72280540990 *T)
    sumY0  +=       4.021 * cos(1.71534059601  + 6603.23049193219 *T)
    sumY0  +=       5.163 * cos(0.06221778581  + 25519.83532335829 *T)
    sumY0  +=       4.150 * cos(2.29239909221  + 27.44302744200 *T)
    sumY0  +=       3.623 * cos(0.72377687032  + 1665.58278404290 *T)
    sumY0  +=       4.634 * cos(3.36220803589  + 3227.45397501119 *T)
    sumY0  +=       4.060 * cos(4.64578985602  + 304.47802118340 *T)
    sumY0  +=       3.862 * cos(5.22051626712  + 74.50415118900 *T)
    sumY0  +=       3.561 * cos(3.35891592080  + 358.65269193120 *T)
    sumY0  +=       4.557 * cos(1.56282166634  + 25974.74468988559 *T)
    sumY0  +=       4.264 * cos(3.13963744879  + 634.93941827469 *T)
    sumY0  +=       4.482 * cos(0.13471172639  + 342.61105682121 *T)
    sumY0  +=       3.539 * cos(5.28146842802  + 119.75073382760 *T)
    sumY0  +=       4.304 * cos(5.35023544496  + 12567.87999017460 *T)
    sumY0  +=       4.138 * cos(5.60646772527  + 107.25419075020 *T)
    sumY0  +=       4.284 * cos(1.62500514182  + 294.42915866079 *T)
    sumY0  +=       3.723 * cos(0.87405503812  + 987.32545955500 *T)
    sumY0  +=       3.723 * cos(4.01564769171  + 987.81309452200 *T)
    sumY0  +=       4.606 * cos(0.78314632413  + 14.42521950279 *T)
    sumY0  +=       4.236 * cos(1.51001695106  + 155.91166179010 *T)
    sumY0  +=       4.458 * cos(1.07510939804  + 395.82251972250 *T)
    sumY0  +=       4.798 * cos(3.66850235979  + 530.19541500900 *T)
    sumY0  +=       3.640 * cos(3.79814548576  + 2564.83138971310 *T)
    sumY0  +=       3.563 * cos(0.66220700888  + 12451.50877558589 *T)
    sumY0  +=       3.443 * cos(3.70889407011  + 245.25042275910 *T)
    sumY0  +=       3.429 * cos(3.16343780315  + 530.04665716270 *T)
    sumY0  +=       3.872 * cos(5.66297097129  + 308.98632106249 *T)
    sumY0  +=       3.406 * cos(4.31900232100  + 529.82290799351 *T)
    sumY0  +=       4.348 * cos(3.09499292675  + 20311.92816802509 *T)
    sumY0  +=       4.589 * cos(3.67073392808  + 181.08713568601 *T)
    sumY0  +=       3.854 * cos(4.35430550499  + 12564.91104475801 *T)
    sumY0  +=       3.789 * cos(5.86431526205  + 3101.63590180850 *T)
    sumY0  +=       3.783 * cos(1.84016316657  + 1614.17130803499 *T)
    sumY0  +=       3.904 * cos(1.57500723101  + 369.80145805910 *T)
    sumY0  +=       3.765 * cos(3.13810202386  + 1025.94613015981 *T)
    sumY0  +=       4.231 * cos(3.78834664840  + 31.52393855141 *T)
    sumY0  +=       4.303 * cos(3.40265517593  + 396.78572756900 *T)
    sumY0  +=       4.085 * cos(0.23841437878  + 14.47091148511 *T)
    sumY0  +=       4.085 * cos(3.38000703237  + 13.98327651810 *T)
    sumY0  +=       3.346 * cos(0.20283168925  + 20351.54567637119 *T)
    sumY0  +=       4.021 * cos(4.51457854549  + 748.34168744680 *T)
    sumY0  +=       3.753 * cos(2.74283876055  + 524.99372948619 *T)
    sumY0  +=       3.935 * cos(2.81201754924  + 1617.14025345159 *T)
    sumY0  +=       4.432 * cos(5.02857999493  + 511.35159082120 *T)
    sumY0  +=       4.170 * cos(2.85784811733  + 274.87931477991 *T)
    sumY0  +=       3.317 * cos(3.36427187559  + 266.70868384049 *T)
    sumY0  +=       4.545 * cos(2.99451528962  + 244.56240155850 *T)
    sumY0  +=       3.589 * cos(6.26623778468  + 59.52629766200 *T)
    sumY0  +=       3.464 * cos(1.94815791367  + 102.27950776349 *T)
    sumY0  +=       4.526 * cos(2.98322850842  + 525.79018099390 *T)
    sumY0  +=       4.603 * cos(2.83181132811  + 26088.14695905770 *T)
    sumY0  +=       4.021 * cos(3.81502221170  + 52174.56562792359 *T)
    sumY0  +=       3.276 * cos(3.52742657818  + 1306.37745808750 *T)
    sumY0  +=       3.214 * cos(5.51315121035  + 20348.57673095459 *T)
    sumY0  +=       3.706 * cos(3.68281338464  + 27.07052042651 *T)
    sumY0  +=       3.759 * cos(5.89324799399  + 164.83974989929 *T)
    sumY0  +=       3.184 * cos(0.44574677170  + 538.01153742540 *T)
    sumY0  +=       4.430 * cos(3.80837870078  + 529.67415014720 *T)
    sumY0  +=       4.064 * cos(2.60402368915  + 6130.29988995670 *T)
    sumY0  +=       3.918 * cos(5.77655218093  + 375.43053235159 *T)
    sumY0  +=       4.058 * cos(3.56233663362  + 433.43429049850 *T)
    sumY0  +=       3.919 * cos(1.93774102166  + 1092.81773021860 *T)
    sumY0  +=       3.919 * cos(5.07933367525  + 1093.30536518560 *T)
    sumY0  +=       3.175 * cos(2.71648311000  + 241.36645360580 *T)
    sumY0  +=       3.135 * cos(1.09798751738  + 127.22797912329 *T)
    sumY0  +=       3.834 * cos(3.42021462454  + 14.31334491820 *T)
    sumY0  +=       4.022 * cos(0.15000192924  + 1477.83836716070 *T)
    sumY0  +=       3.221 * cos(2.66340709341  + 78.16111780620 *T)
    sumY0  +=       3.426 * cos(4.77405099085  + 519.85229016070 *T)
    sumY0  +=       4.369 * cos(2.32053270413  + 746.36957977150 *T)
    sumY0  +=       3.160 * cos(3.58900877772  + 664.99569906519 *T)
    sumY0  +=       4.060 * cos(4.49008083850  + 51.87023394000 *T)
    sumY0  +=       3.107 * cos(3.81160836398  + 28.92750015030 *T)
    sumY0  +=       3.259 * cos(0.91022076156  + 657.88215206440 *T)
    sumY0  +=       3.428 * cos(2.81213415208  + 2351.53229427510 *T)
    sumY0  +=       3.235 * cos(0.07612839981  + 406.34695512460 *T)
    sumY0  +=       3.161 * cos(0.98519827646  + 982.87204143010 *T)
    sumY0  +=       4.351 * cos(2.61742468676  + 20388.19423930069 *T)
    sumY0  +=       3.384 * cos(1.87729416709  + 660.85109748100 *T)
    sumY0  +=       3.452 * cos(5.96738985165  + 326.18236048070 *T)
    sumY0  +=       3.298 * cos(1.72568702486  + 1403.84115801359 *T)
    sumY0  +=       3.278 * cos(5.26025413610  + 941.77006037570 *T)
    sumY0  +=       3.723 * cos(0.29723150363  + 451.95723605810 *T)
    sumY0  +=       3.173 * cos(0.75401885480  + 1400.87221259699 *T)
    sumY0  +=       4.113 * cos(3.44518846630  + 1049.31625271919 *T)
    sumY0  +=       4.012 * cos(0.58002417229  + 52.60394712290 *T)
    sumY0  +=       4.142 * cos(0.18543891861  + 978.67925573610 *T)
    sumY0  +=       4.295 * cos(2.94382365877  + 875.58648151749 *T)
    sumY0  +=       3.224 * cos(5.39074920150  + 459.11896716870 *T)
    sumY0  +=       3.151 * cos(2.11925788926  + 381.85605815180 *T)
    sumY0  +=       3.633 * cos(6.09798622690  + 256.78375799000 *T)
    sumY0  +=       4.250 * cos(4.81834414256  + 528.71094230071 *T)
    sumY0  +=       4.186 * cos(3.66267284521  + 943.25453308399 *T)
    sumY0  +=       3.406 * cos(1.82206499429  + 170.46882419179 *T)
    sumY0  +=       3.231 * cos(6.18447276533  + 400.82094669610 *T)
    sumY0  +=       3.726 * cos(5.26557613435  + 1096.48675892331 *T)
    sumY0  +=       3.792 * cos(5.46702979448  + 111.93461141120 *T)
    sumY0  +=       3.651 * cos(6.14012974300  + 154.42718908179 *T)
    sumY0  +=       3.839 * cos(4.02729058795  + 10060.50958617629 *T)
    sumY0  +=       3.356 * cos(5.33785023580  + 1586.34776735071 *T)
    sumY0  +=       3.219 * cos(1.26547692662  + 213.70966926030 *T)
    sumY0  +=       3.671 * cos(3.08823320781  + 57.60237409650 *T)
    sumY0  +=       4.187 * cos(1.86321883254  + 2772.54460848389 *T)
    sumY0  +=       2.960 * cos(3.77221652347  + 2461.73861549450 *T)
    sumY0  +=       3.331 * cos(2.38361288630  + 10133.80671203529 *T)
    sumY0  +=       3.341 * cos(2.74911210318  + 243.76595005080 *T)
    sumY0  +=       3.466 * cos(0.02652921265  + 1150.92455422949 *T)
    sumY0  +=       3.296 * cos(5.06897390591  + 1653.78881638109 *T)
    sumY0  +=       3.014 * cos(3.47171849350  + 1477.39891630350 *T)
    sumY0  +=       4.118 * cos(1.26070911091  + 25596.58902960090 *T)
    sumY0  +=       2.951 * cos(3.47218747597  + 42.78208713641 *T)
    sumY0  +=       2.951 * cos(4.00999244396  + 33.97161910620 *T)
    sumY0  +=       3.830 * cos(3.02640541849  + 323.48860171000 *T)
    sumY0  +=       3.313 * cos(3.21919687279  + 939.10993149980 *T)
    sumY0  +=       3.031 * cos(4.32205791511  + 156450.90896068609 *T)
    sumY0  +=       3.606 * cos(2.35740018537  + 1082.25966492170 *T)
    sumY0  +=       2.967 * cos(4.72619454182  + 6.39415663780 *T)
    sumY0  +=       2.995 * cos(5.12808890644  + 139.70996798570 *T)
    sumY0  +=       3.251 * cos(1.93107151339  + 709.29362807231 *T)
    sumY0  +=       3.480 * cos(2.18796105799  + 518.14081491630 *T)
    sumY0  +=       3.906 * cos(4.41951013162  + 1119.90506559249 *T)
    sumY0  +=       3.406 * cos(3.42602191152  + 148.79811478929 *T)
    sumY0  +=       3.359 * cos(0.17159576955  + 642.84941678320 *T)
    sumY0  +=       3.027 * cos(5.00980281133  + 184.00789699280 *T)
    sumY0  +=       2.918 * cos(0.68786396977  + 83.62343575990 *T)
    sumY0  +=       3.347 * cos(4.53587187847  + 217.68751450571 *T)
    sumY0  +=       3.277 * cos(1.84412902317  + 912.54386098770 *T)
    sumY0  +=       3.277 * cos(4.98572167676  + 913.03149595471 *T)
    sumY0  +=       3.196 * cos(4.27207353253  + 363.10611005610 *T)
    sumY0  +=       2.869 * cos(2.93254803921  + 285.35556607221 *T)
    sumY0  +=       3.158 * cos(5.89391855080  + 540.01727499551 *T)
    sumY0  +=       2.810 * cos(3.57723287116  + 1592.28565818390 *T)
    sumY0  +=       3.471 * cos(4.56081319778  + 144.39038864671 *T)
    sumY0  +=       3.159 * cos(5.71530971815  + 197.55615956570 *T)
    sumY0  +=       3.227 * cos(1.02602355265  + 6203.59701581570 *T)
    sumY0  +=       3.750 * cos(1.09900342443  + 303.35724676999 *T)
    sumY0  +=       3.848 * cos(4.95190461444  + 26048.04181574459 *T)
    sumY0  +=       2.741 * cos(0.13004673727  + 70.83263035680 *T)
    sumY0  +=       2.826 * cos(3.64821843137  + 460.29462333630 *T)
    sumY0  +=       2.748 * cos(5.69617268740  + 600.52359545141 *T)
    sumY0  +=       3.057 * cos(4.56550138397  + 23.81969071961 *T)
    sumY0  +=       3.057 * cos(6.05827118955  + 52.93401552300 *T)
    sumY0  +=       3.446 * cos(1.96967013470  + 500.13913421850 *T)
    sumY0  +=       2.703 * cos(6.26272265860  + 908.09044286280 *T)
    sumY0  +=       2.817 * cos(1.69638906604  + 210.62215161470 *T)
    sumY0  +=       2.848 * cos(1.16888883373  + 450.47276334980 *T)
    sumY0  +=       2.724 * cos(5.64910484087  + 23.18655127321 *T)
    sumY0  +=       2.905 * cos(1.13800629851  + 149.31937965110 *T)
    sumY0  +=       2.848 * cos(1.48842245891  + 622.66987773339 *T)
    sumY0  +=       2.733 * cos(1.93636126616  + 262.72164882321 *T)
    sumY0  +=       2.863 * cos(2.26914213515  + 175.57663362249 *T)
    sumY0  +=       2.681 * cos(5.83048409789  + 25.19228884330 *T)
    sumY0  +=       2.822 * cos(0.00883588585  + 259.75270340660 *T)
    sumY0  +=       3.174 * cos(1.47302873031  + 347.11935670030 *T)
    sumY0  +=       3.271 * cos(2.97328046332  + 458.59770230690 *T)
    sumY0  +=       2.894 * cos(5.75207939107  + 71.82946809809 *T)
    sumY0  +=       3.490 * cos(1.28003658954  + 664.37136833940 *T)
    sumY0  +=       3.506 * cos(3.91611653269  + 771.34811171130 *T)
    sumY0  +=       3.326 * cos(0.55224065588  + 45.22976769120 *T)
    sumY0  +=       2.988 * cos(4.94563705230  + 299.37021175271 *T)
    sumY0  +=       2.916 * cos(5.17859920603  + 6642.84800027830 *T)
    sumY0  +=       2.916 * cos(5.17859920603  + 6643.33563524530 *T)
    sumY0  +=       2.630 * cos(5.83933407803  + 2751.79141717511 *T)
    sumY0  +=       2.903 * cos(5.88134941337  + 477.08701797169 *T)
    sumY0  +=       2.804 * cos(4.97695491059  + 6681.46867088311 *T)
    sumY0  +=       2.622 * cos(0.73099530901  + 521.85802773080 *T)
    sumY0  +=       2.606 * cos(1.44468831628  + 410.85525500370 *T)
    sumY0  +=       3.046 * cos(0.79307135358  + 959.45373476091 *T)
    sumY0  +=       3.127 * cos(1.47432830628  + 225.55182103190 *T)
    sumY0  +=       2.700 * cos(2.88388264285  + 963.64652045490 *T)
    sumY0  +=       2.778 * cos(3.22939757518  + 238.39750818919 *T)
    sumY0  +=       3.029 * cos(0.01392036536  + 473.21855518340 *T)
    sumY0  +=       2.671 * cos(3.02950363348  + 531.94052014820 *T)
    sumY0  +=       2.914 * cos(2.29089443923  + 554.31380496631 *T)
    sumY0  +=       3.087 * cos(1.37613019083  + 340.26644213040 *T)
    sumY0  +=       3.438 * cos(3.89546045811  + 6171.40187101109 *T)
    sumY0  +=       2.879 * cos(4.04729837696  + 218.65072235220 *T)
    sumY0  +=       3.140 * cos(3.44921752602  + 609.12161516050 *T)
    sumY0  +=       3.003 * cos(5.24831469226  + 464.97504399731 *T)
    sumY0  +=       3.257 * cos(6.23715182295  + 305.96249389171 *T)
    sumY0  +=       3.211 * cos(0.93594149210  + 416.53251340600 *T)
    sumY0  +=       3.265 * cos(6.26189223545  + 24.73471445630 *T)
    sumY0  +=       2.644 * cos(5.73202797797  + 508.59414157570 *T)
    sumY0  +=       2.764 * cos(0.26986971158  + 410.59462257279 *T)
    sumY0  +=       3.428 * cos(0.99849665750  + 1012.61950567990 *T)
    sumY0  +=       2.614 * cos(2.50560328703  + 213.59109703130 *T)
    sumY0  +=       3.469 * cos(3.71563719744  + 24.14975911971 *T)
    sumY0  +=       2.606 * cos(5.52398994555  + 213.49472881170 *T)
    sumY0  +=       3.444 * cos(0.99352524535  + 891.57323487331 *T)
    sumY0  +=       2.540 * cos(5.89247404448  + 564.87187026320 *T)
    sumY0  +=       2.540 * cos(2.75088139089  + 565.35950523021 *T)
    sumY0  +=       2.754 * cos(4.26615188091  + 57.55418998670 *T)
    sumY0  +=       2.531 * cos(2.16100356086  + 800.59243462910 *T)
    sumY0  +=       2.557 * cos(2.24078889519  + 341.49028240779 *T)
    sumY0  +=       2.601 * cos(2.97805958626  + 261.23717611490 *T)
    sumY0  +=       3.027 * cos(1.77262933089  + 331.07772159029 *T)
    sumY0  +=       2.494 * cos(5.29381091116  + 203.98168536590 *T)
    sumY0  +=       2.590 * cos(3.33405614398  + 1190.54206257560 *T)
    sumY0  +=       3.494 * cos(1.33796606005  + 534.07938416230 *T)
    sumY0  +=       3.144 * cos(1.59061342897  + 1503.96498681560 *T)
    sumY0  +=       2.818 * cos(2.04818816564  + 49.31067880061 *T)
    sumY0  +=       2.791 * cos(2.91527039269  + 288.32451148881 *T)
    sumY0  +=       2.471 * cos(2.80089246981  + 411.11588743459 *T)
    sumY0  +=       3.059 * cos(1.73898053758  + 172.48911597691 *T)
    sumY0  +=       2.972 * cos(5.01468129705  + 569.29165849331 *T)
    sumY0  +=       3.418 * cos(3.83213917567  + 638.39599865830 *T)
    sumY0  +=       2.541 * cos(3.41936535077  + 1448.09090291091 *T)
    sumY0  +=       2.663 * cos(5.14390724060  + 573.69689250840 *T)
    sumY0  +=       2.439 * cos(2.82552552997  + 1625.96527569680 *T)
    sumY0  +=       2.739 * cos(1.01296407856  + 112.88326504270 *T)
    sumY0  +=       2.821 * cos(4.09784112299  + 402.93606672331 *T)
    sumY0  +=       3.412 * cos(5.98246878418  + 772.83258441960 *T)
    sumY0  +=       2.624 * cos(4.28449219810  + 1624.48080298850 *T)
    sumY0  +=       3.170 * cos(2.10762429629  + 1011.13503297159 *T)
    sumY0  +=       2.908 * cos(3.03870325402  + 635.94831810351 *T)
    sumY0  +=       2.664 * cos(4.25083112029  + 409.41896640519 *T)
    sumY0  +=       3.091 * cos(0.31165645931  + 379.25961975071 *T)
    sumY0  +=       3.301 * cos(3.48430565498  + 19.79366136440 *T)
    sumY0  +=       3.176 * cos(4.86809762289  + 300.90956621520 *T)
    sumY0  +=       3.022 * cos(4.37742921398  + 52.01899178630 *T)
    sumY0  +=       2.890 * cos(6.24788645936  + 293.43232091950 *T)
    sumY0  +=       2.698 * cos(3.26450368524  + 78149.06650032569 *T)
    sumY0  +=       2.558 * cos(2.31657732137  + 1371.33719666830 *T)
    sumY0  +=       2.619 * cos(5.37658613752  + 202.00957769060 *T)
    sumY0  +=       3.176 * cos(5.32134696018  + 10101.61156723069 *T)
    sumY0  +=       3.341 * cos(3.91159951862  + 345.89551642290 *T)
    sumY0  +=       2.373 * cos(0.25236813570  + 130.85131584570 *T)
    sumY0  +=       2.644 * cos(4.25178872695  + 305.10235190919 *T)
    sumY0  +=       3.339 * cos(3.06224357085  + 2849.29831472650 *T)
    sumY0  +=       2.410 * cos(3.15243245459  + 951.85255279310 *T)
    sumY0  +=       3.303 * cos(3.82850925169  + 769.57294599210 *T)
    sumY0  +=       3.302 * cos(3.28815049288  + 90.15202742410 *T)
    sumY0  +=       2.416 * cos(4.43555947495  + 527.92904500800 *T)
    sumY0  +=       2.361 * cos(0.63550285699  + 905.12149744620 *T)
    sumY0  +=       2.737 * cos(3.20111311776  + 1206.21999939070 *T)
    sumY0  +=       2.441 * cos(5.40055208431  + 246.73489546739 *T)
    sumY0  +=       2.441 * cos(5.40055208431  + 247.22253043440 *T)
    sumY0  +=       2.957 * cos(2.68753178821  + 238.23075185041 *T)
    sumY0  +=       3.263 * cos(2.55710522617  + 1506.93393223219 *T)
    sumY0  +=       3.293 * cos(1.22031676357  + 66.15220969580 *T)
    sumY0  +=       3.241 * cos(5.00885682863  + 978.41862330520 *T)
    sumY0  +=       3.149 * cos(2.07892234370  + 271.91036936330 *T)
    sumY0  +=       3.149 * cos(5.22051499729  + 271.42273439630 *T)
    sumY0  +=       2.328 * cos(0.36371018198  + 31.73887900000 *T)
    sumY0  +=       2.372 * cos(2.25731707420  + 309.03450517230 *T)
    sumY0  +=       2.372 * cos(2.25731707420  + 309.52214013930 *T)
    sumY0  +=       2.369 * cos(5.90092450419  + 418.98019396080 *T)
    sumY0  +=       3.007 * cos(6.21088893213  + 1437.78140795740 *T)
    sumY0  +=       3.034 * cos(4.41266493573  + 330.86278114170 *T)
    sumY0  +=       2.345 * cos(4.37756786631  + 453.93183586090 *T)
    sumY0  +=       3.118 * cos(5.30478414038  + 1434.81246254079 *T)
    sumY0  +=       2.324 * cos(5.43011369487  + 495.24626523640 *T)
    sumY0  +=       2.340 * cos(0.70753572900  + 452.43031681009 *T)
    sumY0  +=       2.336 * cos(1.61735465920  + 189.59127930300 *T)
    sumY0  +=       2.920 * cos(2.21678930184  + 1549.69920442121 *T)
    sumY0  +=       2.494 * cos(2.36432658211  + 1187.57311715899 *T)
    sumY0  +=       2.692 * cos(5.74887255496  + 425.13053311509 *T)
    sumY0  +=       2.874 * cos(3.06187769177  + 1654.27645134810 *T)
    sumY0  +=       2.809 * cos(0.95838272583  + 317.58434077160 *T)
    sumY0  +=       2.735 * cos(2.36910571540  + 1513.05064149171 *T)
    sumY0  +=       2.949 * cos(4.69913732218  + 186.71620997851 *T)
    sumY0  +=       2.320 * cos(2.31406529898  + 487.38195871019 *T)
    sumY0  +=       3.113 * cos(4.63822476931  + 353.28425006961 *T)
    sumY0  +=       3.086 * cos(3.30396670519  + 1230.59902177890 *T)
    sumY0  +=       2.722 * cos(0.59415160235  + 49.68318581610 *T)
    sumY0  +=       3.064 * cos(2.11685584705  + 133.13224006171 *T)
    sumY0  +=       3.064 * cos(5.25844850064  + 132.64460509469 *T)
    sumY0  +=       2.470 * cos(1.21163683322  + 532.38246313290 *T)
    sumY0  +=       2.640 * cos(5.23029870928  + 394.33804701421 *T)
    sumY0  +=       2.252 * cos(3.41692637069  + 22.65073219640 *T)
    sumY0  +=       3.151 * cos(3.68959728933  + 859.77184894361 *T)
    sumY0  +=       2.671 * cos(2.49225273236  + 37.36795329250 *T)
    sumY0  +=       2.380 * cos(2.43767088034  + 429.27513469930 *T)
    sumY0  +=       2.655 * cos(4.29167785274  + 484.15238086270 *T)
    sumY0  +=       3.005 * cos(4.59447567553  + 1929.33933741419 *T)
    sumY0  +=       2.550 * cos(0.89259009595  + 496.94318626580 *T)
    sumY0  +=       2.290 * cos(4.98199823333  + 455.18681390559 *T)
    sumY0  +=       2.608 * cos(2.28446271246  + 422.95803920620 *T)
    sumY0  +=       2.226 * cos(0.52897898579  + 47.82620609231 *T)
    sumY0  +=       2.233 * cos(3.36949240110  + 877.34614087170 *T)
    sumY0  +=       2.764 * cos(2.40581332791  + 356.68058425589 *T)
    sumY0  +=       2.719 * cos(3.56033366747  + 177.58237119260 *T)
    sumY0  +=       2.999 * cos(3.63965245652  + 1926.37039199759 *T)
    sumY0  +=       2.693 * cos(2.00893145868  + 6284.80414018320 *T)
    sumY0  +=       2.369 * cos(5.90816921383  + 70.88081446661 *T)
    sumY0  +=       2.498 * cos(2.14771583991  + 315.15121443180 *T)
    sumY0  +=       2.204 * cos(4.77545839271  + 442.88613559700 *T)
    sumY0  +=       2.261 * cos(4.89614385698  + 621.23358913490 *T)
    sumY0  +=       2.213 * cos(1.45024938630  + 1189.05758986730 *T)
    sumY0  +=       2.492 * cos(4.24445703283  + 406.97128585040 *T)
    sumY0  +=       2.976 * cos(3.02481916981  + 1014.10397838820 *T)
    sumY0  +=       2.840 * cos(0.64471611311  + 522.33360061030 *T)
    sumY0  +=       2.340 * cos(3.29528259310  + 440.43845504219 *T)
    sumY0  +=       3.012 * cos(2.70591736862  + 15.90969221110 *T)
    sumY0  +=       3.012 * cos(2.70591736862  + 16.39732717810 *T)
    sumY0  +=       2.372 * cos(1.81307027955  + 132.59642098490 *T)
    sumY0  +=       2.232 * cos(3.99248125271  + 158.12984768129 *T)
    sumY0  +=       2.961 * cos(5.94214048852  + 286.35240381350 *T)
    sumY0  +=       2.961 * cos(2.80054783493  + 286.84003878050 *T)

    # NEPTUNE Y Coord
    sumY1  = 0
    sumY1  +=  357822.049 * cos(3.03457804662  + 0.24381748350 *T)
    sumY1  +=  256200.629 * cos(0.44613631554  + 36.89238041300 *T)
    sumY1  +=  242677.799 * cos(3.89213848413  + 39.86132582961 *T)
    sumY1  +=  106073.143 * cos(4.64936068389  + 37.88921815429 *T)
    sumY1  +=  103735.195 * cos(4.51191141127  + 38.37685312130 *T)
    sumY1  +=  118508.231 * cos(1.31543504055  + 76.50988875911 *T)
    sumY1  +=   21930.692 * cos(1.62939936370  + 35.40790770471 *T)
    sumY1  +=   17445.772 * cos(2.69316438174  + 41.34579853790 *T)
    sumY1  +=   13038.843 * cos(3.79605108858  + 3.21276290011 *T)
    sumY1  +=    4928.885 * cos(0.51813571490  + 73.54094334250 *T)
    sumY1  +=    2742.686 * cos(2.49310000815  + 77.99436146740 *T)
    sumY1  +=    2155.134 * cos(2.54801435750  + 4.69723560840 *T)
    sumY1  +=    1882.800 * cos(2.84958651579  + 33.92343499640 *T)
    sumY1  +=    1572.888 * cos(5.79049449823  + 114.64292439690 *T)
    sumY1  +=    1326.507 * cos(4.45906236203  + 75.02541605080 *T)
    sumY1  +=    1343.094 * cos(1.46758582116  + 42.83027124621 *T)
    sumY1  +=     897.979 * cos(2.69913392072  + 426.84200835950 *T)
    sumY1  +=     865.617 * cos(0.09538823497  + 37.85558825950 *T)
    sumY1  +=     849.963 * cos(4.24519902715  + 38.89811798311 *T)
    sumY1  +=     922.754 * cos(1.77437053635  + 72.05647063421 *T)
    sumY1  +=     726.258 * cos(5.81913445111  + 36.40474544600 *T)
    sumY1  +=     778.220 * cos(4.27400223412  + 206.42936592071 *T)
    sumY1  +=     754.025 * cos(3.76126183394  + 220.65645992230 *T)
    sumY1  +=     607.406 * cos(4.81815513635  + 1059.62574767270 *T)
    sumY1  +=     571.831 * cos(0.85851242227  + 522.82123557730 *T)
    sumY1  +=     560.995 * cos(0.34476353479  + 537.04832957890 *T)
    sumY1  +=     501.078 * cos(0.14255476727  + 28.81562556571 *T)
    sumY1  +=     493.238 * cos(0.53463363296  + 39.37369086260 *T)
    sumY1  +=     474.802 * cos(5.97795229031  + 98.65617104110 *T)
    sumY1  +=     453.975 * cos(0.14363576661  + 35.92917256650 *T)
    sumY1  +=     471.731 * cos(3.27137539235  + 1.72829019180 *T)
    sumY1  +=     410.057 * cos(4.19500321025  + 40.82453367610 *T)
    sumY1  +=     366.899 * cos(4.19675940250  + 47.93808067690 *T)
    sumY1  +=     450.109 * cos(2.82750084230  + 76.02225379210 *T)
    sumY1  +=     354.347 * cos(1.55870450456  + 1.24065522479 *T)
    sumY1  +=     300.159 * cos(1.31608359577  + 6.18170831670 *T)
    sumY1  +=     327.501 * cos(5.77559197316  + 33.43580002939 *T)
    sumY1  +=     174.973 * cos(4.06947925642  + 32.43896228810 *T)
    sumY1  +=     171.503 * cos(2.86905921629  + 34.18406742730 *T)
    sumY1  +=     156.749 * cos(1.02465451730  + 79.47883417571 *T)
    sumY1  +=     152.549 * cos(5.29458792782  + 30.30009827400 *T)
    sumY1  +=     150.775 * cos(1.46875297221  + 42.56963881530 *T)
    sumY1  +=     162.280 * cos(5.51215947389  + 31.26330612050 *T)
    sumY1  +=     131.609 * cos(3.19975255613  + 7.83293736379 *T)
    sumY1  +=     136.159 * cos(3.00798814109  + 70.57199792590 *T)
    sumY1  +=     134.616 * cos(5.10422989672  + 45.49040012211 *T)
    sumY1  +=     116.304 * cos(5.32949492640  + 46.45360796860 *T)
    sumY1  +=     115.918 * cos(0.24763705851  + 44.31474395451 *T)
    sumY1  +=     110.293 * cos(4.69481457289  + 35.45609181450 *T)
    sumY1  +=      99.282 * cos(0.34979488247  + 2.72512793310 *T)
    sumY1  +=      99.914 * cos(5.92865840649  + 41.29761442810 *T)
    sumY1  +=     108.706 * cos(1.52062460635  + 113.15845168861 *T)
    sumY1  +=      88.965 * cos(5.83760483379  + 60.52313540329 *T)
    sumY1  +=      86.886 * cos(0.75633169755  + 31.95132732110 *T)
    sumY1  +=      72.232 * cos(1.93507773057  + 640.14110379750 *T)
    sumY1  +=      86.985 * cos(3.23018942725  + 419.72846135871 *T)
    sumY1  +=      73.430 * cos(5.93590859407  + 70.08436295889 *T)
    sumY1  +=      53.395 * cos(2.89441175199  + 433.95555536030 *T)
    sumY1  +=      57.451 * cos(5.79242631159  + 213.54291292150 *T)
    sumY1  +=      51.458 * cos(2.44646741842  + 69.39634175830 *T)
    sumY1  +=      48.797 * cos(4.44285537763  + 111.67397898031 *T)
    sumY1  +=      48.557 * cos(1.53569583062  + 2.67694382330 *T)
    sumY1  +=      42.206 * cos(4.80902694866  + 74.53778108379 *T)
    sumY1  +=      42.550 * cos(0.10167685669  + 7.66618102501 *T)
    sumY1  +=      39.462 * cos(4.26971409186  + 31.78457098230 *T)
    sumY1  +=      39.445 * cos(5.65869884949  + 12.77399045571 *T)
    sumY1  +=      42.389 * cos(4.01940273222  + 110.18950627200 *T)
    sumY1  +=      44.118 * cos(5.15854031484  + 1589.31671276730 *T)
    sumY1  +=      37.988 * cos(4.30930512095  + 6.34846465550 *T)
    sumY1  +=      37.802 * cos(4.41701497532  + 14.25846316400 *T)
    sumY1  +=      36.280 * cos(1.56655638104  + 273.82223084130 *T)
    sumY1  +=      37.247 * cos(6.20048406786  + 73.05330837550 *T)
    sumY1  +=      36.282 * cos(1.28256818253  + 84.58664360640 *T)
    sumY1  +=      40.018 * cos(2.70338838405  + 4.43660317750 *T)
    sumY1  +=      32.400 * cos(0.07249247133  + 44.96913526031 *T)
    sumY1  +=      31.842 * cos(0.45413330049  + 34.92027273770 *T)
    sumY1  +=      32.037 * cos(1.37472212177  + 27.33115285740 *T)
    sumY1  +=      34.456 * cos(1.80072966680  + 529.93478257810 *T)
    sumY1  +=      31.208 * cos(0.16340478702  + 1052.51220067191 *T)
    sumY1  +=      30.002 * cos(0.76559925402  + 1066.73929467350 *T)
    sumY1  +=      33.805 * cos(4.47034863791  + 149.80701461810 *T)
    sumY1  +=      33.096 * cos(0.88714456679  + 116.12739710521 *T)
    sumY1  +=      30.571 * cos(5.59230793843  + 22.39009976550 *T)
    sumY1  +=      24.020 * cos(4.95060362012  + 63.97971578690 *T)
    sumY1  +=      23.780 * cos(5.91699417045  + 105.76971804189 *T)
    sumY1  +=      23.110 * cos(4.08428537053  + 23.87457247379 *T)
    sumY1  +=      22.233 * cos(2.78662410030  + 174.92224231670 *T)
    sumY1  +=      21.377 * cos(2.17060095397  + 316.63568714010 *T)
    sumY1  +=      25.400 * cos(6.09781709877  + 106.73292588839 *T)
    sumY1  +=      20.754 * cos(5.65712726150  + 5.66044345490 *T)
    sumY1  +=      25.572 * cos(0.74829738831  + 529.44714761109 *T)
    sumY1  +=      19.878 * cos(5.73044838456  + 32.96022714990 *T)
    sumY1  +=      19.754 * cos(2.91920492275  + 49.42255338521 *T)
    sumY1  +=      19.241 * cos(4.44333892046  + 7.14491616321 *T)
    sumY1  +=      17.979 * cos(6.19717030924  + 62.49524307860 *T)
    sumY1  +=      19.513 * cos(0.93528726944  + 68.59989025060 *T)
    sumY1  +=      18.273 * cos(3.63050723326  + 227.77000692311 *T)
    sumY1  +=      17.552 * cos(4.23270606709  + 69.08752521760 *T)
    sumY1  +=      16.704 * cos(5.51562885485  + 40.85816357090 *T)
    sumY1  +=      16.996 * cos(2.67528896312  + 91.54262404029 *T)
    sumY1  +=      16.800 * cos(2.08712613778  + 30.46685461280 *T)
    sumY1  +=      16.800 * cos(5.22871879137  + 30.95448957981 *T)
    sumY1  +=      16.400 * cos(3.45346576095  + 33.26904369061 *T)
    sumY1  +=      17.242 * cos(4.89590986485  + 11.55015017831 *T)
    sumY1  +=      15.590 * cos(0.42181314281  + 37.10482873410 *T)
    sumY1  +=      15.590 * cos(3.91877412353  + 39.64887750850 *T)
    sumY1  +=      15.469 * cos(4.91170489358  + 43.79347909271 *T)
    sumY1  +=      16.590 * cos(1.19757876004  + 33.71098667531 *T)
    sumY1  +=      19.347 * cos(4.06418655235  + 152.77596003471 *T)
    sumY1  +=      14.994 * cos(4.29075263091  + 319.06881347989 *T)
    sumY1  +=      14.395 * cos(5.67189517492  + 110.45013870291 *T)
    sumY1  +=      15.528 * cos(5.87587490674  + 79.43065006591 *T)
    sumY1  +=      13.727 * cos(0.88731617093  + 43.48466255200 *T)
    sumY1  +=      13.988 * cos(5.54059308769  + 4.20960064140 *T)
    sumY1  +=      14.467 * cos(5.13403607047  + 108.70503356371 *T)
    sumY1  +=      16.652 * cos(4.74696813612  + 304.84171947829 *T)
    sumY1  +=      15.153 * cos(1.64704158732  + 72.31710306511 *T)
    sumY1  +=      12.810 * cos(0.80784638784  + 11.28951774740 *T)
    sumY1  +=      12.751 * cos(5.32663860873  + 45.79921666280 *T)
    sumY1  +=      13.293 * cos(3.14432194523  + 43.04271956730 *T)
    sumY1  +=      12.751 * cos(0.98606109283  + 515.70768857651 *T)
    sumY1  +=      11.616 * cos(4.07265201948  + 97.17169833279 *T)
    sumY1  +=      11.538 * cos(4.63247506911  + 633.02755679670 *T)
    sumY1  +=      11.046 * cos(5.97427560684  + 25.84668014910 *T)
    sumY1  +=      11.032 * cos(5.39646944302  + 4.86399194720 *T)
    sumY1  +=      11.189 * cos(2.37859784397  + 83.10217089810 *T)
    sumY1  +=      10.860 * cos(5.09978655023  + 9.80504503910 *T)
    sumY1  +=      10.958 * cos(3.05455531642  + 415.04804069769 *T)
    sumY1  +=      10.244 * cos(4.97854342755  + 71.09326278771 *T)
    sumY1  +=      11.427 * cos(3.07758676009  + 129.67565967810 *T)
    sumY1  +=       9.895 * cos(4.05388510292  + 251.67594855930 *T)
    sumY1  +=       9.802 * cos(3.40894474212  + 44.48150029329 *T)
    sumY1  +=      11.029 * cos(6.26821027792  + 143.38148881789 *T)
    sumY1  +=       9.235 * cos(4.42290386641  + 199.31581891990 *T)
    sumY1  +=       8.899 * cos(1.97879285736  + 7.35736448430 *T)
    sumY1  +=       7.746 * cos(4.32608949084  + 103.33659170210 *T)
    sumY1  +=       8.691 * cos(2.29051174612  + 32.74777882880 *T)
    sumY1  +=       7.714 * cos(3.51056446926  + 65.46418849521 *T)
    sumY1  +=       8.007 * cos(0.23872922784  + 544.16187657970 *T)
    sumY1  +=       7.513 * cos(6.18736050552  + 69.60879007940 *T)
    sumY1  +=       7.336 * cos(3.43409422317  + 15.74293587230 *T)
    sumY1  +=       7.195 * cos(4.56950200257  + 949.41942645330 *T)
    sumY1  +=       9.601 * cos(5.68191403081  + 80.96330688400 *T)
    sumY1  +=       8.094 * cos(1.70304241092  + 526.75338884040 *T)
    sumY1  +=       8.109 * cos(5.77532188995  + 533.11617631580 *T)
    sumY1  +=       6.906 * cos(3.70672232078  + 137.27684164590 *T)
    sumY1  +=       7.455 * cos(1.11362695292  + 105.24845318010 *T)
    sumY1  +=       7.826 * cos(2.45321405406  + 77.03115362090 *T)
    sumY1  +=       6.529 * cos(5.57837212493  + 65.20355606430 *T)
    sumY1  +=       7.134 * cos(2.05010386093  + 44.00592741381 *T)
    sumY1  +=       6.186 * cos(0.85023811841  + 31.47575444160 *T)
    sumY1  +=       6.186 * cos(3.99183077200  + 30.98811947460 *T)
    sumY1  +=       7.698 * cos(4.89115030216  + 14.47091148511 *T)
    sumY1  +=       7.434 * cos(3.96333556733  + 146.83806920150 *T)
    sumY1  +=       6.317 * cos(5.24777799313  + 66.94866120350 *T)
    sumY1  +=       6.903 * cos(4.63739310514  + 75.98862389731 *T)
    sumY1  +=       5.591 * cos(3.47781120117  + 448.98829064149 *T)
    sumY1  +=       6.425 * cos(3.47626562775  + 678.27413943531 *T)
    sumY1  +=       5.483 * cos(1.61247704205  + 34.44469985821 *T)
    sumY1  +=       5.483 * cos(2.72811022429  + 42.30900638440 *T)
    sumY1  +=       5.519 * cos(1.27116075236  + 853.44019923550 *T)
    sumY1  +=       5.483 * cos(5.10869779974  + 100.14064374939 *T)
    sumY1  +=       5.483 * cos(1.96710514615  + 100.62827871640 *T)
    sumY1  +=       6.288 * cos(2.60319684406  + 143.90275367970 *T)
    sumY1  +=       6.239 * cos(4.20987077870  + 17.76992530181 *T)
    sumY1  +=       5.246 * cos(3.52035332490  + 209.61075965840 *T)
    sumY1  +=       5.331 * cos(4.83550697489  + 45.96597300160 *T)
    sumY1  +=       5.131 * cos(4.53503564274  + 217.47506618460 *T)
    sumY1  +=       5.325 * cos(2.82680123889  + 19.25439801010 *T)
    sumY1  +=       5.172 * cos(2.44008575183  + 25.35904518210 *T)
    sumY1  +=       5.139 * cos(4.17507085285  + 6.86972951729 *T)
    sumY1  +=       5.992 * cos(2.76367557670  + 9.31741007210 *T)
    sumY1  +=       5.011 * cos(4.91884286875  + 38.85242600079 *T)
    sumY1  +=       4.975 * cos(3.01044533436  + 525.25436191710 *T)
    sumY1  +=       4.910 * cos(3.47707407879  + 45.27795180100 *T)
    sumY1  +=       5.250 * cos(0.16559612363  + 0.71939036300 *T)
    sumY1  +=       4.731 * cos(6.27469301850  + 40.38259069140 *T)
    sumY1  +=       4.731 * cos(1.20748690143  + 36.37111555120 *T)
    sumY1  +=       5.910 * cos(1.40566081690  + 6168.43292559449 *T)
    sumY1  +=       4.700 * cos(4.66314397827  + 50.90702609350 *T)
    sumY1  +=       5.127 * cos(1.64029328726  + 140.93380826310 *T)
    sumY1  +=       5.321 * cos(2.09939112611  + 1104.87233031131 *T)
    sumY1  +=       6.339 * cos(1.02786059939  + 10175.39632805670 *T)
    sumY1  +=       4.983 * cos(1.46113982673  + 1090.64523630970 *T)
    sumY1  +=       5.487 * cos(0.13979521980  + 180.03005174739 *T)
    sumY1  +=       4.560 * cos(2.38015606975  + 323.74923414091 *T)
    sumY1  +=       4.689 * cos(5.95510153546  + 1068.22376738181 *T)
    sumY1  +=       5.562 * cos(2.83481631972  + 10098.64262181409 *T)
    sumY1  +=       4.432 * cos(0.83559275468  + 415.79630809560 *T)
    sumY1  +=       4.456 * cos(1.46246408589  + 235.68919520349 *T)
    sumY1  +=       4.289 * cos(4.40449246840  + 1051.02772796360 *T)
    sumY1  +=       4.145 * cos(4.70457869197  + 33.69643246030 *T)
    sumY1  +=       4.167 * cos(3.29886964345  + 416.53251340600 *T)
    sumY1  +=       4.107 * cos(0.91956436736  + 61.01077037031 *T)
    sumY1  +=       4.088 * cos(3.88660175347  + 423.66061462181 *T)
    sumY1  +=       5.027 * cos(2.86873904525  + 21.70207856490 *T)
    sumY1  +=       4.030 * cos(3.44189647415  + 216.72430665921 *T)
    sumY1  +=       4.278 * cos(6.22413352457  + 310.47079377080 *T)
    sumY1  +=       4.013 * cos(2.96769071148  + 104275.10267753768 *T)
    sumY1  +=       4.505 * cos(5.93746161630  + 291.94784821120 *T)
    sumY1  +=       3.959 * cos(4.60954974650  + 210.36151918381 *T)
    sumY1  +=       3.962 * cos(5.59162611271  + 978.93988816699 *T)
    sumY1  +=       5.561 * cos(3.31923216598  + 1409.47023230609 *T)
    sumY1  +=       5.073 * cos(5.15285057233  + 1498.33591252310 *T)
    sumY1  +=       4.227 * cos(6.19954210169  + 534.38820070301 *T)
    sumY1  +=       4.054 * cos(2.45804475947  + 430.02340209721 *T)
    sumY1  +=       3.863 * cos(0.67184399240  + 1127.50624756031 *T)
    sumY1  +=       4.367 * cos(0.14073726901  + 58.98378094080 *T)
    sumY1  +=       4.694 * cos(4.91041582057  + 77.50672650040 *T)
    sumY1  +=       4.144 * cos(5.16137286617  + 518.14081491630 *T)
    sumY1  +=       4.289 * cos(1.72696806473  + 921.32069912310 *T)
    sumY1  +=       4.039 * cos(5.37067473153  + 1622.76932774409 *T)
    sumY1  +=       5.180 * cos(3.80035699018  + 99.14380600810 *T)
    sumY1  +=       4.845 * cos(1.33083083566  + 136.78920667889 *T)
    sumY1  +=       4.827 * cos(1.21379713661  + 418.24398865040 *T)
    sumY1  +=       3.722 * cos(4.94171351364  + 1065.25482196520 *T)
    sumY1  +=       4.729 * cos(2.19682691364  + 421.21293406700 *T)
    sumY1  +=       3.490 * cos(0.54756448610  + 986.80419469320 *T)
    sumY1  +=       3.715 * cos(1.30912268012  + 254.10907489909 *T)
    sumY1  +=       3.488 * cos(5.95100195908  + 187.94005025590 *T)
    sumY1  +=       3.989 * cos(5.37041318514  + 95.73540973430 *T)
    sumY1  +=       3.603 * cos(2.22310220083  + 67.11541754230 *T)
    sumY1  +=       3.530 * cos(0.93986174870  + 24.36220744081 *T)
    sumY1  +=       3.538 * cos(1.51952328076  + 57.49930823250 *T)
    sumY1  +=       3.838 * cos(3.56895316428  + 979.90309601349 *T)
    sumY1  +=       3.615 * cos(1.60474010406  + 493.28621964860 *T)
    sumY1  +=       3.457 * cos(1.79944886939  + 807.70598162989 *T)
    sumY1  +=       3.648 * cos(1.43920595596  + 647.25465079831 *T)
    sumY1  +=       4.048 * cos(6.25251011272  + 979.69064769239 *T)
    sumY1  +=       4.414 * cos(2.00415973362  + 1062.59469308931 *T)
    sumY1  +=       3.631 * cos(0.74841494891  + 486.17267264780 *T)
    sumY1  +=       3.347 * cos(4.13560148025  + 151.29148732640 *T)
    sumY1  +=       3.305 * cos(5.23397852699  + 1544.07013012871 *T)
    sumY1  +=       3.428 * cos(0.11713176717  + 107.22056085540 *T)
    sumY1  +=       3.286 * cos(3.55869926238  + 1131.69903325430 *T)
    sumY1  +=       3.389 * cos(3.22644735392  + 28.98238190449 *T)
    sumY1  +=       3.353 * cos(2.30309048870  + 10289.79543497010 *T)
    sumY1  +=       3.214 * cos(0.83720162261  + 569.55229092420 *T)
    sumY1  +=       3.210 * cos(1.05449812296  + 114.15528942990 *T)
    sumY1  +=       3.353 * cos(1.62613341505  + 157.88376946540 *T)
    sumY1  +=       3.339 * cos(5.26712406495  + 443.09858391810 *T)
    sumY1  +=       3.188 * cos(2.62887165071  + 361.13400238079 *T)
    sumY1  +=       3.390 * cos(5.53564544873  + 1558.29722413030 *T)
    sumY1  +=       3.933 * cos(2.08622660372  + 313.43973918739 *T)
    sumY1  +=       3.131 * cos(0.52572989907  + 275.30670354960 *T)
    sumY1  +=       3.156 * cos(6.23565991208  + 431.84043533310 *T)
    sumY1  +=       3.993 * cos(4.90080803105  + 67.63668240410 *T)
    sumY1  +=       3.708 * cos(0.24836934654  + 500.39976664941 *T)
    sumY1  +=       4.051 * cos(4.41826493037  + 59.03866269500 *T)
    sumY1  +=       3.757 * cos(2.02838234929  + 296.40126633610 *T)
    sumY1  +=       3.138 * cos(0.63906969040  + 347.11935670030 *T)
    sumY1  +=       3.086 * cos(1.67235466145  + 392.90175841570 *T)
    sumY1  +=       3.466 * cos(1.60020779124  + 215.19414196860 *T)
    sumY1  +=       3.139 * cos(3.09999506971  + 159.36824217371 *T)
    sumY1  +=       3.466 * cos(3.47361825954  + 2145.34674583789 *T)
    sumY1  +=       3.737 * cos(3.53018898305  + 449.03647475130 *T)
    sumY1  +=       3.286 * cos(1.82620986507  + 435.44002806861 *T)
    sumY1  +=       3.043 * cos(5.02988988519  + 2.20386307129 *T)
    sumY1  +=       3.999 * cos(5.93005561135  + 6245.18663183710 *T)
    sumY1  +=       3.999 * cos(5.93005561135  + 6244.69899687009 *T)
    sumY1  +=       2.999 * cos(0.07518657231  + 526.00262931501 *T)
    sumY1  +=       3.014 * cos(5.52171912448  + 1054.94532701169 *T)
    sumY1  +=       3.091 * cos(4.52477390940  + 42.99702758500 *T)
    sumY1  +=       3.274 * cos(5.81401559586  + 736.12033101530 *T)
    sumY1  +=       2.965 * cos(1.12065249261  + 533.86693584120 *T)
    sumY1  +=       3.149 * cos(2.34844411589  + 103.76398047180 *T)
    sumY1  +=       3.610 * cos(1.47397387042  + 55.05162767771 *T)
    sumY1  +=       2.937 * cos(5.93931708618  + 385.25239233810 *T)
    sumY1  +=       2.903 * cos(4.35235911504  + 117.56368570370 *T)
    sumY1  +=       2.968 * cos(1.28091906944  + 613.31440085451 *T)
    sumY1  +=       3.097 * cos(4.42120029558  + 1395.24313830449 *T)
    sumY1  +=       2.931 * cos(3.60795663266  + 202.49721265760 *T)
    sumY1  +=       3.013 * cos(1.49526296600  + 121.23520653590 *T)
    sumY1  +=       3.206 * cos(2.86107032756  + 53.40958840249 *T)
    sumY1  +=       3.269 * cos(0.45608619325  + 480.00777927849 *T)
    sumY1  +=       3.948 * cos(5.43052261409  + 112.88326504270 *T)
    sumY1  +=       2.824 * cos(3.14926129801  + 176.40671502500 *T)
    sumY1  +=       2.827 * cos(0.36860696411  + 429.81095377611 *T)
    sumY1  +=       3.348 * cos(3.55163976673  + 6284.80414018320 *T)
    sumY1  +=       2.862 * cos(2.43356518574  + 384.02855206069 *T)
    sumY1  +=       3.228 * cos(5.13696496058  + 52.60394712290 *T)
    sumY1  +=       3.446 * cos(5.32686217736  + 62.00760811160 *T)
    sumY1  +=       3.096 * cos(4.83839993614  + 71.82946809809 *T)
    sumY1  +=       3.031 * cos(6.24076040166  + 494.23487328010 *T)
    sumY1  +=       3.021 * cos(6.10531658530  + 328.59641114070 *T)
    sumY1  +=       2.731 * cos(0.79873177065  + 432.47108265200 *T)
    sumY1  +=       3.171 * cos(4.97187934370  + 10215.01383640280 *T)
    sumY1  +=       2.674 * cos(2.29257372574  + 158.12984768129 *T)
    sumY1  +=       2.901 * cos(4.22947732371  + 559.46979850680 *T)
    sumY1  +=       2.631 * cos(4.21066619701  + 2008.80135664250 *T)
    sumY1  +=       2.695 * cos(3.54636234855  + 81.61769818981 *T)
    sumY1  +=       2.695 * cos(3.54636234855  + 81.13006322279 *T)
    sumY1  +=       2.721 * cos(4.26045579509  + 326.18236048070 *T)
    sumY1  +=       2.775 * cos(4.27616320157  + 457.86149699650 *T)
    sumY1  +=       3.054 * cos(6.23455983590  + 6281.83519476660 *T)
    sumY1  +=       2.852 * cos(2.90626399353  + 186.71620997851 *T)
    sumY1  +=       2.538 * cos(1.73224718947  + 111.18634401329 *T)
    sumY1  +=       2.835 * cos(6.07135630955  + 419.50145882259 *T)
    sumY1  +=       2.868 * cos(3.66893253825  + 844.56699288049 *T)
    sumY1  +=       2.530 * cos(2.61093052560  + 1050.75254131770 *T)
    sumY1  +=       2.843 * cos(4.15972261299  + 830.33989887890 *T)
    sumY1  +=       2.848 * cos(4.69330398359  + 659.36662477269 *T)
    sumY1  +=       3.031 * cos(2.55942970028  + 406.34695512460 *T)
    sumY1  +=       2.907 * cos(3.71503751053  + 573.69689250840 *T)
    sumY1  +=       2.536 * cos(5.01251643852  + 82.61453593110 *T)
    sumY1  +=       2.957 * cos(2.02121290773  + 947.70795120889 *T)
    sumY1  +=       3.321 * cos(3.87615887284  + 449.99968259780 *T)
    sumY1  +=       3.117 * cos(4.74251772899  + 457.32567791969 *T)
    sumY1  +=       2.902 * cos(1.37682148855  + 10212.04489098620 *T)
    sumY1  +=       2.459 * cos(3.74222344492  + 450.73339578069 *T)
    sumY1  +=       2.557 * cos(1.32711393852  + 525.48136445320 *T)
    sumY1  +=       2.624 * cos(3.35106775051  + 946.22347850060 *T)
    sumY1  +=       2.417 * cos(0.09697452811  + 351.57277482520 *T)
    sumY1  +=       2.454 * cos(3.27912412212  + 196.01680510321 *T)
    sumY1  +=       2.585 * cos(5.70826849733  + 248.70700314271 *T)
    sumY1  +=       2.549 * cos(0.23244817308  + 1062.80714141041 *T)
    sumY1  +=       2.615 * cos(4.06090316067  + 425.13053311509 *T)
    sumY1  +=       2.387 * cos(2.04191008078  + 654.36819779910 *T)
    sumY1  +=       2.439 * cos(1.45718218253  + 462.74230389109 *T)
    sumY1  +=       2.367 * cos(2.75159024078  + 107.52937739611 *T)
    sumY1  +=       2.538 * cos(4.19012282298  + 481.23161955590 *T)
    sumY1  +=       2.479 * cos(3.56223236298  + 205.94173095370 *T)
    sumY1  +=       2.791 * cos(5.44719714506  + 24.14975911971 *T)
    sumY1  +=       2.626 * cos(2.73743077295  + 213.05527795450 *T)
    sumY1  +=       2.445 * cos(0.13750022894  + 146.87169909629 *T)
    sumY1  +=       2.575 * cos(0.32351821119  + 86.07111631471 *T)
    sumY1  +=       3.120 * cos(1.20503303199  + 456.36247007319 *T)
    sumY1  +=       2.587 * cos(1.59304506140  + 400.82094669610 *T)
    sumY1  +=       2.261 * cos(3.24987212470  + 644.33388949151 *T)
    sumY1  +=       2.796 * cos(0.30156280343  + 216.67861467689 *T)
    sumY1  +=       2.896 * cos(0.71993168447  + 1685.29593998510 *T)
    sumY1  +=       2.453 * cos(4.81368306062  + 109.96250373590 *T)
    sumY1  +=       2.325 * cos(0.25394919776  + 442.88613559700 *T)
    sumY1  +=       2.387 * cos(3.75345095238  + 599.08730685290 *T)
    sumY1  +=       2.873 * cos(5.13430840850  + 834.53268457290 *T)
    sumY1  +=       2.963 * cos(5.48260613029  + 2119.00767786191 *T)
    sumY1  +=       2.233 * cos(4.37346978315  + 709.29362807231 *T)
    sumY1  +=       2.337 * cos(2.73478761543  + 210.57396750490 *T)
    sumY1  +=       2.259 * cos(5.24608284286  + 29.50364676630 *T)
    sumY1  +=       2.300 * cos(2.19835177792  + 986.05343516780 *T)
    sumY1  +=       2.199 * cos(1.21359358990  + 606.20085385370 *T)
    sumY1  +=       2.325 * cos(4.84070679976  + 109.70187130500 *T)

    # NEPTUNE Y Coord
    sumY2  = 0
    sumY2  +=  1620002.167 * cos(5.31277371181  + 38.37685312130 *T)
    sumY2  +=   28138.323 * cos(4.01361134771  + 0.24381748350 *T)
    sumY2  +=   12318.619 * cos(1.01433481938  + 39.86132582961 *T)
    sumY2  +=    8346.956 * cos(0.42201817445  + 37.88921815429 *T)
    sumY2  +=    5131.003 * cos(3.55894443240  + 76.50988875911 *T)
    sumY2  +=    4109.792 * cos(6.17733924169  + 36.89238041300 *T)
    sumY2  +=    1369.663 * cos(1.98683082370  + 1.72829019180 *T)
    sumY2  +=     633.706 * cos(0.81055475696  + 3.21276290011 *T)
    sumY2  +=     583.006 * cos(6.25831267359  + 41.34579853790 *T)
    sumY2  +=     546.517 * cos(5.42211492491  + 75.02541605080 *T)
    sumY2  +=     246.224 * cos(0.87539145895  + 213.54291292150 *T)
    sumY2  +=     159.773 * cos(5.97653264005  + 206.42936592071 *T)
    sumY2  +=     156.619 * cos(2.04577076492  + 220.65645992230 *T)
    sumY2  +=     191.674 * cos(0.60086490402  + 529.93478257810 *T)
    sumY2  +=     188.212 * cos(2.86105100061  + 35.40790770471 *T)
    sumY2  +=     117.788 * cos(2.55450585422  + 522.82123557730 *T)
    sumY2  +=     114.488 * cos(4.76320074833  + 35.92917256650 *T)
    sumY2  +=     112.666 * cos(4.92459292590  + 537.04832957890 *T)
    sumY2  +=     105.949 * cos(5.84319366771  + 40.82453367610 *T)
    sumY2  +=      77.696 * cos(5.55872082952  + 77.99436146740 *T)
    sumY2  +=      90.798 * cos(0.48400254359  + 73.54094334250 *T)
    sumY2  +=      67.696 * cos(0.87599797871  + 426.84200835950 *T)
    sumY2  +=      74.860 * cos(6.15585546499  + 4.69723560840 *T)
    sumY2  +=      64.717 * cos(1.34762573111  + 34.92027273770 *T)
    sumY2  +=      51.378 * cos(1.41763897935  + 36.40474544600 *T)
    sumY2  +=      50.205 * cos(5.38246230326  + 42.83027124621 *T)
    sumY2  +=      40.929 * cos(4.64969715335  + 33.92343499640 *T)
    sumY2  +=      36.136 * cos(1.44221344970  + 98.65617104110 *T)
    sumY2  +=      33.953 * cos(3.45488669371  + 1059.62574767270 *T)
    sumY2  +=      34.603 * cos(4.83781708761  + 76.02225379210 *T)
    sumY2  +=      35.441 * cos(0.92439194787  + 31.26330612050 *T)
    sumY2  +=      29.614 * cos(1.77735061850  + 28.81562556571 *T)
    sumY2  +=      31.027 * cos(3.40007815913  + 45.49040012211 *T)
    sumY2  +=      35.521 * cos(2.70107635202  + 39.37369086260 *T)
    sumY2  +=      25.488 * cos(2.47241259934  + 47.93808067690 *T)
    sumY2  +=      20.115 * cos(5.50307482336  + 1.24065522479 *T)
    sumY2  +=      14.328 * cos(1.16675266548  + 433.95555536030 *T)
    sumY2  +=      15.503 * cos(1.56080925316  + 114.64292439690 *T)
    sumY2  +=      16.998 * cos(2.50473507501  + 33.43580002939 *T)
    sumY2  +=      13.166 * cos(5.10795715214  + 419.72846135871 *T)
    sumY2  +=      13.053 * cos(1.40065469605  + 60.52313540329 *T)
    sumY2  +=      10.637 * cos(4.37252543304  + 34.18406742730 *T)
    sumY2  +=       9.610 * cos(0.08619380662  + 640.14110379750 *T)
    sumY2  +=       9.354 * cos(6.12654852334  + 42.56963881530 *T)
    sumY2  +=      11.447 * cos(1.48554252527  + 71.56883566720 *T)
    sumY2  +=       8.454 * cos(4.32641802480  + 2.72512793310 *T)
    sumY2  +=       9.012 * cos(2.87032368668  + 72.05647063421 *T)
    sumY2  +=       9.594 * cos(4.22403656438  + 69.39634175830 *T)
    sumY2  +=       7.419 * cos(1.92565712551  + 227.77000692311 *T)
    sumY2  +=       6.800 * cos(3.57170452455  + 113.15845168861 *T)
    sumY2  +=       6.267 * cos(5.50825416463  + 1066.73929467350 *T)
    sumY2  +=       6.895 * cos(2.74011142877  + 111.67397898031 *T)
    sumY2  +=       5.770 * cos(5.94492182042  + 32.43896228810 *T)
    sumY2  +=       5.686 * cos(0.66726291170  + 30.30009827400 *T)
    sumY2  +=       6.679 * cos(3.28449699734  + 258.78949556011 *T)
    sumY2  +=       7.799 * cos(0.01316502615  + 7.35736448430 *T)
    sumY2  +=       5.906 * cos(4.35406299044  + 44.31474395451 *T)
    sumY2  +=       5.606 * cos(3.60862172739  + 46.45360796860 *T)
    sumY2  +=       5.525 * cos(2.04832143671  + 1052.51220067191 *T)
    sumY2  +=       7.257 * cos(4.88554087166  + 1097.75878331050 *T)
    sumY2  +=       5.427 * cos(3.37665889417  + 105.76971804189 *T)
    sumY2  +=       5.179 * cos(2.68906561571  + 515.70768857651 *T)
    sumY2  +=       5.163 * cos(1.56680359144  + 7.83293736379 *T)
    sumY2  +=       4.688 * cos(0.95059580199  + 222.14093263061 *T)
    sumY2  +=       5.379 * cos(1.15102731522  + 22.39009976550 *T)
    sumY2  +=       4.607 * cos(0.17861908533  + 549.16030355330 *T)
    sumY2  +=       4.101 * cos(1.86095686008  + 213.05527795450 *T)
    sumY2  +=       4.262 * cos(3.94315605774  + 204.94489321240 *T)
    sumY2  +=       3.916 * cos(4.92658442131  + 207.91383862900 *T)
    sumY2  +=       4.089 * cos(0.26445229364  + 304.84171947829 *T)
    sumY2  +=       3.729 * cos(6.12087852262  + 199.31581891990 *T)
    sumY2  +=       3.680 * cos(3.97729685951  + 1589.31671276730 *T)
    sumY2  +=       3.702 * cos(2.58942752453  + 319.06881347989 *T)
    sumY2  +=       4.832 * cos(5.97662492227  + 215.02738562980 *T)
    sumY2  +=       3.474 * cos(5.88640441483  + 103.33659170210 *T)
    sumY2  +=       3.298 * cos(4.81858024415  + 544.16187657970 *T)
    sumY2  +=       4.521 * cos(1.64991198460  + 108.21739859670 *T)
    sumY2  +=       3.967 * cos(4.24856395374  + 944.73900579230 *T)
    sumY2  +=       4.059 * cos(1.44024718167  + 149.80701461810 *T)
    sumY2  +=       4.009 * cos(4.04772901629  + 533.11617631580 *T)
    sumY2  +=       3.288 * cos(3.02037527521  + 407.93449369690 *T)
    sumY2  +=       3.976 * cos(3.43142225420  + 526.75338884040 *T)
    sumY2  +=       3.343 * cos(5.37024544109  + 531.41925528640 *T)
    sumY2  +=       3.932 * cos(5.23324378146  + 91.54262404029 *T)
    sumY2  +=       3.478 * cos(4.62796796973  + 6.18170831670 *T)
    sumY2  +=       2.967 * cos(5.75717546362  + 860.55374623631 *T)
    sumY2  +=       3.058 * cos(1.59562573237  + 342.97475511610 *T)
    sumY2  +=       3.974 * cos(3.61989902199  + 335.86120811530 *T)
    sumY2  +=       2.849 * cos(2.43690877786  + 666.48017177350 *T)
    sumY2  +=       2.999 * cos(2.84954389869  + 937.62545879149 *T)
    sumY2  +=       3.008 * cos(0.45545092573  + 74.53778108379 *T)
    sumY2  +=       3.080 * cos(4.61982032828  + 129.67565967810 *T)
    sumY2  +=       3.346 * cos(3.07224007183  + 1162.71852189130 *T)
    sumY2  +=       2.625 * cos(3.26539092131  + 273.82223084130 *T)
    sumY2  +=       2.931 * cos(3.14888688193  + 235.68919520349 *T)
    sumY2  +=       2.579 * cos(5.19712816213  + 1073.85284167431 *T)
    sumY2  +=       2.550 * cos(1.43127384605  + 26.58288545949 *T)
    sumY2  +=       2.542 * cos(2.65218049337  + 1265.81129610991 *T)
    sumY2  +=       2.483 * cos(3.30358671055  + 453.18107633550 *T)
    sumY2  +=       2.732 * cos(3.33706438099  + 563.38739755489 *T)
    sumY2  +=       2.508 * cos(0.10584642422  + 37.85558825950 *T)
    sumY2  +=       2.508 * cos(5.05315792539  + 425.35753565121 *T)
    sumY2  +=       2.680 * cos(1.03370719327  + 454.66554904380 *T)
    sumY2  +=       2.511 * cos(1.57939297348  + 209.61075965840 *T)
    sumY2  +=       2.512 * cos(0.17397391459  + 217.47506618460 *T)
    sumY2  +=       2.552 * cos(4.09952672426  + 79.47883417571 *T)
    sumY2  +=       2.457 * cos(1.09953412303  + 38.89811798311 *T)
    sumY2  +=       2.343 * cos(3.50679449028  + 981.38756872180 *T)
    sumY2  +=       2.501 * cos(3.24491806395  + 669.40093308030 *T)
    sumY2  +=       2.330 * cos(2.37985529884  + 38.32866901151 *T)
    sumY2  +=       2.327 * cos(5.10713891298  + 38.42503723110 *T)
    sumY2  +=       2.481 * cos(0.85514029866  + 655.17383907870 *T)
    sumY2  +=       2.569 * cos(2.65544269508  + 464.97504399731 *T)

    # NEPTUNE Y Coord
    sumY3  = 0
    sumY3  +=     985.355 * cos(5.40479271994  + 38.37685312130 *T)
    sumY3  +=     482.798 * cos(2.40351592403  + 37.88921815429 *T)
    sumY3  +=     416.447 * cos(5.08276459732  + 0.24381748350 *T)
    sumY3  +=     303.825 * cos(5.25036318156  + 39.86132582961 *T)
    sumY3  +=      89.203 * cos(6.23576998030  + 36.89238041300 *T)
    sumY3  +=      70.862 * cos(4.26820111330  + 76.50988875911 *T)
    sumY3  +=      28.900 * cos(4.07922314280  + 41.34579853790 *T)
    sumY3  +=      22.279 * cos(1.38807052555  + 206.42936592071 *T)
    sumY3  +=      21.480 * cos(0.30279640762  + 220.65645992230 *T)
    sumY3  +=      16.157 * cos(4.26502283154  + 522.82123557730 *T)
    sumY3  +=      15.714 * cos(3.19639937559  + 537.04832957890 *T)
    sumY3  +=      11.404 * cos(0.68881522230  + 35.40790770471 *T)
    sumY3  +=      13.199 * cos(4.77296321336  + 7.35736448430 *T)
    sumY3  +=       7.024 * cos(5.41129077346  + 3.21276290011 *T)
    sumY3  +=       6.772 * cos(5.90918041474  + 69.39634175830 *T)
    sumY3  +=       4.517 * cos(1.71813394760  + 45.49040012211 *T)
    sumY3  +=       4.523 * cos(2.61625996387  + 31.26330612050 *T)
    sumY3  +=       3.682 * cos(3.04341607939  + 98.65617104110 *T)
    sumY3  +=       3.656 * cos(2.17226292493  + 968.64494742849 *T)
    sumY3  +=       3.927 * cos(5.25979459691  + 426.84200835950 *T)
    sumY3  +=       3.199 * cos(3.24406648940  + 1519.67655352550 *T)
    sumY3  +=       3.498 * cos(4.84556329465  + 407.93449369690 *T)
    sumY3  +=       3.304 * cos(3.81825313228  + 422.16158769850 *T)
    sumY3  +=       3.331 * cos(3.54124873059  + 36.40474544600 *T)
    sumY3  +=       3.244 * cos(4.42247914038  + 484.20056497250 *T)
    sumY3  +=       2.689 * cos(5.61171743826  + 441.06910236111 *T)
    sumY3  +=       3.247 * cos(3.33592493083  + 498.42765897409 *T)
    sumY3  +=       2.651 * cos(2.02749548795  + 304.84171947829 *T)
    sumY3  +=       2.645 * cos(0.90433895865  + 461.77909604459 *T)
    sumY3  +=       2.542 * cos(1.05108120438  + 444.58305662640 *T)
    sumY3  +=       2.524 * cos(5.46978523129  + 433.95555536030 *T)
    sumY3  +=       2.472 * cos(0.92177286185  + 319.06881347989 *T)
    sumY3  +=       2.355 * cos(2.03272387563  + 447.55200204300 *T)
    sumY3  +=       2.876 * cos(3.65433810175  + 853.44019923550 *T)
    sumY3  +=       2.279 * cos(6.20789133449  + 458.81015062800 *T)
    sumY3  +=       2.147 * cos(1.06531556689  + 175.40987728371 *T)
    sumY3  +=       2.637 * cos(2.06613866653  + 73.54094334250 *T)

    # NEPTUNE Y Coord
    sumY4  = 0
    sumY4  +=    3455.306 * cos(2.04385259535  + 38.37685312130 *T)
    sumY4  +=      47.405 * cos(0.64311364094  + 0.24381748350 *T)
    sumY4  +=      21.936 * cos(4.30052120876  + 37.88921815429 *T)
    sumY4  +=      15.596 * cos(0.30774488881  + 76.50988875911 *T)
    sumY4  +=      17.186 * cos(3.96705739007  + 39.86132582961 *T)
    sumY4  +=      17.459 * cos(3.25820107685  + 36.89238041300 *T)
    sumY4  +=       4.229 * cos(6.14484758916  + 515.70768857651 *T)
    sumY4  +=       4.334 * cos(3.84568484898  + 433.95555536030 *T)
    sumY4  +=       3.547 * cos(1.04322259069  + 989.98558843089 *T)
    sumY4  +=       3.155 * cos(0.14083942013  + 467.40817033709 *T)
    sumY4  +=       3.017 * cos(4.77718347184  + 227.77000692311 *T)
    sumY4  +=       2.981 * cos(5.01159762849  + 1.72829019180 *T)
    sumY4  +=       2.295 * cos(4.84988240730  + 220.65645992230 *T)
    sumY4  +=       2.296 * cos(3.13566627364  + 206.42936592071 *T)

    # NEPTUNE Y Coord
    sumY5  = 0
    sumY5  +=      26.291 * cos(2.14645097520  + 38.37685312130 *T)

    # NEPTUNE Z Coord
    sumZ0  = 0
    sumZ0  +=  92866054.405 * cos(1.44103930278  + 38.13303563780 *T)
    sumZ0  +=  1245978.462 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ0  +=  474333.567 * cos(2.52218774238  + 36.64856292950 *T)
    sumZ0  +=  451987.936 * cos(3.50949720541  + 39.61750834610 *T)
    sumZ0  +=  417558.068 * cos(5.91310695421  + 76.26607127560 *T)
    sumZ0  +=   84104.329 * cos(4.38928900096  + 1.48447270830 *T)
    sumZ0  +=   32704.958 * cos(1.52048692001  + 74.78159856730 *T)
    sumZ0  +=   30873.335 * cos(3.29017611456  + 35.16409022120 *T)
    sumZ0  +=   25812.584 * cos(3.19303128782  + 2.96894541660 *T)
    sumZ0  +=   16865.319 * cos(2.13251104425  + 41.10198105440 *T)
    sumZ0  +=   11789.909 * cos(3.60001877675  + 213.29909543800 *T)
    sumZ0  +=    9770.125 * cos(2.80133971586  + 73.29712585900 *T)
    sumZ0  +=   11279.680 * cos(3.55816676334  + 529.69096509460 *T)
    sumZ0  +=    4119.873 * cos(1.67934316836  + 77.75054398390 *T)
    sumZ0  +=    2818.034 * cos(4.10661077794  + 114.39910691340 *T)
    sumZ0  +=    2868.677 * cos(4.27011526203  + 33.67961751290 *T)
    sumZ0  +=    2213.464 * cos(1.96045135168  + 4.45341812490 *T)
    sumZ0  +=    1865.650 * cos(5.05540709577  + 71.81265315070 *T)
    sumZ0  +=     840.177 * cos(0.94268885160  + 42.58645376270 *T)
    sumZ0  +=     457.516 * cos(5.71650412080  + 108.46121608020 *T)
    sumZ0  +=     530.252 * cos(0.85800267793  + 111.43016149680 *T)
    sumZ0  +=     490.859 * cos(6.07827301209  + 112.91463420510 *T)
    sumZ0  +=     331.254 * cos(0.29304964526  + 70.32818044240 *T)
    sumZ0  +=     330.045 * cos(2.83839676215  + 426.59819087600 *T)
    sumZ0  +=     273.589 * cos(3.91013681794  + 1059.38193018920 *T)
    sumZ0  +=     277.586 * cos(1.45092010545  + 148.07872442630 *T)
    sumZ0  +=     274.474 * cos(5.42657022437  + 32.19514480460 *T)
    sumZ0  +=     205.306 * cos(0.75818737085  + 5.93789083320 *T)
    sumZ0  +=     173.516 * cos(5.85498030099  + 145.10977900970 *T)
    sumZ0  +=     141.275 * cos(1.73147597657  + 28.57180808220 *T)
    sumZ0  +=     139.093 * cos(1.67466701191  + 184.72728735580 *T)
    sumZ0  +=     143.647 * cos(2.51620047812  + 37.61177077600 *T)
    sumZ0  +=     136.955 * cos(0.20339778664  + 79.23501669220 *T)
    sumZ0  +=     126.296 * cos(4.40661385040  + 37.16982779130 *T)
    sumZ0  +=     120.906 * cos(1.61767636602  + 39.09624348430 *T)
    sumZ0  +=     111.761 * cos(6.20948230785  + 98.89998852460 *T)
    sumZ0  +=     140.758 * cos(3.50944989694  + 38.65430049960 *T)
    sumZ0  +=     111.589 * cos(4.18561395578  + 47.69426319340 *T)
    sumZ0  +=     133.509 * cos(4.78977105547  + 38.08485152800 *T)
    sumZ0  +=     102.622 * cos(0.81673762159  + 4.19278569400 *T)
    sumZ0  +=     133.292 * cos(1.23386935925  + 38.18121974760 *T)
    sumZ0  +=      98.771 * cos(0.72335005782  + 106.97674337190 *T)
    sumZ0  +=      93.919 * cos(0.56607810948  + 206.18554843720 *T)
    sumZ0  +=      81.727 * cos(3.47861315258  + 220.41264243880 *T)
    sumZ0  +=      74.559 * cos(2.31518880439  + 312.19908396260 *T)
    sumZ0  +=      74.401 * cos(5.99935727164  + 181.75834193920 *T)
    sumZ0  +=      73.141 * cos(1.80069951634  + 137.03302416240 *T)
    sumZ0  +=      66.838 * cos(1.87185330904  + 221.37585028530 *T)
    sumZ0  +=      58.303 * cos(5.61662561548  + 35.68535508300 *T)
    sumZ0  +=      51.685 * cos(6.02831347649  + 44.07092647100 *T)
    sumZ0  +=      51.928 * cos(0.40473854286  + 40.58071619260 *T)
    sumZ0  +=      48.182 * cos(2.97141991737  + 37.87240320690 *T)
    sumZ0  +=      49.439 * cos(1.66178905717  + 68.84370773410 *T)
    sumZ0  +=      47.742 * cos(3.05261105371  + 38.39366806870 *T)
    sumZ0  +=      46.428 * cos(2.66596739242  + 146.59425171800 *T)
    sumZ0  +=      50.936 * cos(0.19994012329  + 30.05628079050 *T)
    sumZ0  +=      41.127 * cos(6.05239303825  + 115.88357962170 *T)
    sumZ0  +=      55.537 * cos(2.11977296055  + 109.94568878850 *T)
    sumZ0  +=      41.357 * cos(0.86667380713  + 143.62530630140 *T)
    sumZ0  +=      44.492 * cos(6.00613878606  + 149.56319713460 *T)
    sumZ0  +=      37.856 * cos(5.19945796177  + 72.07328558160 *T)
    sumZ0  +=      47.361 * cos(3.58964541604  + 38.02116105320 *T)
    sumZ0  +=      34.690 * cos(1.47398766326  + 8.07675484730 *T)
    sumZ0  +=      37.349 * cos(5.15067903040  + 33.94024994380 *T)
    sumZ0  +=      34.386 * cos(6.15246630929  + 218.40690486870 *T)
    sumZ0  +=      47.180 * cos(2.43405967107  + 38.24491022240 *T)
    sumZ0  +=      33.382 * cos(0.88396990078  + 42.32582133180 *T)
    sumZ0  +=      40.753 * cos(3.59668759304  + 522.57741809380 *T)
    sumZ0  +=      33.071 * cos(2.02572550598  + 258.02441321480 *T)
    sumZ0  +=      38.849 * cos(5.79294381756  + 46.20979048510 *T)
    sumZ0  +=      31.716 * cos(0.30412624027  + 536.80451209540 *T)
    sumZ0  +=      27.605 * cos(2.81540405940  + 183.24281464750 *T)
    sumZ0  +=      24.616 * cos(0.40597272412  + 30.71067209630 *T)
    sumZ0  +=      21.716 * cos(0.90792314747  + 175.16605980020 *T)
    sumZ0  +=      21.634 * cos(3.25469228647  + 388.46515523820 *T)
    sumZ0  +=      20.384 * cos(5.80954414865  + 7.42236354150 *T)
    sumZ0  +=      18.640 * cos(1.06424642993  + 180.27386923090 *T)
    sumZ0  +=      16.716 * cos(0.02332590259  + 255.05546779820 *T)
    sumZ0  +=      18.790 * cos(0.62408059806  + 35.21227433100 *T)
    sumZ0  +=      16.990 * cos(2.19726523790  + 294.67297614430 *T)
    sumZ0  +=      16.083 * cos(4.20629091415  + 0.96320784650 *T)
    sumZ0  +=      21.571 * cos(2.39337706760  + 152.53214255120 *T)
    sumZ0  +=      17.102 * cos(5.39964889794  + 41.05379694460 *T)
    sumZ0  +=      14.503 * cos(4.66614523797  + 110.20632121940 *T)
    sumZ0  +=      15.218 * cos(2.93182248771  + 219.89137757700 *T)
    sumZ0  +=      14.757 * cos(2.02029526083  + 105.49227066360 *T)
    sumZ0  +=      13.303 * cos(2.09362099250  + 639.89728631400 *T)
    sumZ0  +=      13.582 * cos(0.99222619680  + 44.72531777680 *T)
    sumZ0  +=      11.309 * cos(4.15253392707  + 487.36514376280 *T)
    sumZ0  +=      12.521 * cos(0.41449986025  + 186.21176006410 *T)
    sumZ0  +=      12.363 * cos(3.07599476497  + 6.59228213900 *T)
    sumZ0  +=      10.825 * cos(4.13817476053  + 0.52126486180 *T)
    sumZ0  +=      14.304 * cos(5.15644933777  + 31.54075349880 *T)
    sumZ0  +=      10.056 * cos(4.27077049743  + 1589.07289528380 *T)
    sumZ0  +=       9.355 * cos(1.23360360711  + 216.92243216040 *T)
    sumZ0  +=       8.774 * cos(5.77195684843  + 12.53017297220 *T)
    sumZ0  +=       8.445 * cos(0.17584724644  + 291.70403072770 *T)
    sumZ0  +=       8.927 * cos(2.36869187243  + 331.32153907380 *T)
    sumZ0  +=       9.613 * cos(0.28151591238  + 60.76695288680 *T)
    sumZ0  +=       8.279 * cos(5.72084525545  + 36.76043751410 *T)
    sumZ0  +=       9.111 * cos(3.49027191661  + 45.24658263860 *T)
    sumZ0  +=       8.178 * cos(0.25637861075  + 39.50563376150 *T)
    sumZ0  +=       8.598 * cos(3.10287346009  + 256.53994050650 *T)
    sumZ0  +=       9.489 * cos(4.94205919676  + 151.04766984290 *T)
    sumZ0  +=       8.564 * cos(2.15904622462  + 274.06604832480 *T)
    sumZ0  +=       7.474 * cos(2.93279436008  + 27.08733537390 *T)
    sumZ0  +=       7.944 * cos(0.38277699900  + 10213.28554621100 *T)
    sumZ0  +=       7.132 * cos(1.50971234331  + 944.98282327580 *T)
    sumZ0  +=       9.005 * cos(0.07284074730  + 419.48464387520 *T)
    sumZ0  +=       7.642 * cos(5.56097006597  + 187.69623277240 *T)
    sumZ0  +=       7.705 * cos(1.54152595157  + 84.34282612290 *T)
    sumZ0  +=       6.896 * cos(4.71453324740  + 406.10313764110 *T)
    sumZ0  +=       7.282 * cos(5.81348163927  + 316.39186965660 *T)
    sumZ0  +=       6.215 * cos(5.27967153537  + 7.11354700080 *T)
    sumZ0  +=       6.090 * cos(4.48506819561  + 415.29185818120 *T)
    sumZ0  +=       6.316 * cos(0.59502514374  + 453.42489381900 *T)
    sumZ0  +=       6.177 * cos(5.20115334051  + 80.71948940050 *T)
    sumZ0  +=       6.324 * cos(2.18406254461  + 142.14083359310 *T)
    sumZ0  +=       5.646 * cos(0.76840537906  + 11.04570026390 *T)
    sumZ0  +=       6.102 * cos(5.58764378724  + 662.53120356300 *T)
    sumZ0  +=       5.334 * cos(4.52703538458  + 103.09277421860 *T)
    sumZ0  +=       6.269 * cos(1.95162790177  + 605.95703637020 *T)
    sumZ0  +=       5.591 * cos(5.20631788297  + 14.01464568050 *T)
    sumZ0  +=       4.914 * cos(1.39906038110  + 253.57099508990 *T)
    sumZ0  +=       4.519 * cos(2.07610590431  + 2042.49778910280 *T)
    sumZ0  +=       5.681 * cos(4.80791039970  + 641.12112659140 *T)
    sumZ0  +=       6.294 * cos(0.56936923702  + 31.01948863700 *T)
    sumZ0  +=       4.758 * cos(2.54312258712  + 367.97010200330 *T)
    sumZ0  +=       4.406 * cos(0.33696669222  + 328.35259365720 *T)
    sumZ0  +=       4.390 * cos(5.08949604858  + 286.59622129700 *T)
    sumZ0  +=       4.678 * cos(4.87546696295  + 442.75170057060 *T)
    sumZ0  +=       4.407 * cos(5.58110402011  + 252.08652238160 *T)
    sumZ0  +=       4.305 * cos(1.31724140028  + 493.04240216510 *T)

    # NEPTUNE Z Coord
    sumZ1  = 0
    sumZ1  +=  6832633.707 * cos(3.80782656293  + 38.13303563780 *T)
    sumZ1  +=   64598.028 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ1  +=   42738.331 * cos(4.82540335637  + 36.64856292950 *T)
    sumZ1  +=   31421.638 * cos(6.08083255385  + 39.61750834610 *T)
    sumZ1  +=   27088.623 * cos(1.97557659098  + 76.26607127560 *T)
    sumZ1  +=    5924.197 * cos(0.48500737803  + 1.48447270830 *T)
    sumZ1  +=    2429.056 * cos(3.86784381378  + 74.78159856730 *T)
    sumZ1  +=    2107.258 * cos(6.19720726581  + 35.16409022120 *T)
    sumZ1  +=    1644.542 * cos(5.76041185818  + 2.96894541660 *T)
    sumZ1  +=    1059.588 * cos(4.89687990866  + 41.10198105440 *T)
    sumZ1  +=    1084.464 * cos(5.33722455731  + 213.29909543800 *T)
    sumZ1  +=     880.611 * cos(5.70150456912  + 529.69096509460 *T)
    sumZ1  +=     824.125 * cos(5.04137560987  + 73.29712585900 *T)
    sumZ1  +=     250.821 * cos(4.25953295692  + 77.75054398390 *T)
    sumZ1  +=     158.517 * cos(0.13500997625  + 114.39910691340 *T)
    sumZ1  +=     129.390 * cos(4.76999039957  + 4.45341812490 *T)
    sumZ1  +=     105.033 * cos(0.99234583035  + 33.67961751290 *T)
    sumZ1  +=      54.734 * cos(3.96812588022  + 42.58645376270 *T)
    sumZ1  +=      34.799 * cos(4.30403521625  + 37.61177077600 *T)
    sumZ1  +=      41.340 * cos(2.29314729799  + 206.18554843720 *T)
    sumZ1  +=      28.374 * cos(1.67853213747  + 111.43016149680 *T)
    sumZ1  +=      25.340 * cos(1.70015942799  + 220.41264243880 *T)
    sumZ1  +=      26.021 * cos(6.09040669128  + 426.59819087600 *T)
    sumZ1  +=      27.198 * cos(5.95188476775  + 71.81265315070 *T)
    sumZ1  +=      21.598 * cos(2.17619854748  + 112.91463420510 *T)
    sumZ1  +=      25.154 * cos(4.13631230281  + 28.57180808220 *T)
    sumZ1  +=      20.543 * cos(1.56946393801  + 38.65430049960 *T)
    sumZ1  +=      17.216 * cos(2.81703507859  + 98.89998852460 *T)
    sumZ1  +=      19.926 * cos(3.15301554161  + 108.46121608020 *T)
    sumZ1  +=      15.544 * cos(2.07364839388  + 40.58071619260 *T)
    sumZ1  +=      15.357 * cos(5.45891321516  + 522.57741809380 *T)
    sumZ1  +=      12.266 * cos(3.82427247378  + 5.93789083320 *T)
    sumZ1  +=      13.587 * cos(1.34795861192  + 47.69426319340 *T)
    sumZ1  +=      10.839 * cos(4.75461825384  + 536.80451209540 *T)
    sumZ1  +=      10.785 * cos(2.61566414257  + 79.23501669220 *T)
    sumZ1  +=      10.916 * cos(3.88744647444  + 35.68535508300 *T)
    sumZ1  +=       9.833 * cos(3.62341506247  + 38.18121974760 *T)
    sumZ1  +=       9.849 * cos(0.89613145346  + 38.08485152800 *T)
    sumZ1  +=       9.317 * cos(0.51297445145  + 37.16982779130 *T)
    sumZ1  +=       8.955 * cos(4.00532631258  + 39.09624348430 *T)
    sumZ1  +=       7.089 * cos(2.29610703652  + 137.03302416240 *T)
    sumZ1  +=       7.573 * cos(3.20619266484  + 4.19278569400 *T)
    sumZ1  +=       8.063 * cos(5.86511590361  + 1059.38193018920 *T)
    sumZ1  +=       7.655 * cos(3.39616255650  + 145.10977900970 *T)
    sumZ1  +=       6.061 * cos(2.31235275416  + 109.94568878850 *T)
    sumZ1  +=       5.726 * cos(6.23400745185  + 312.19908396260 *T)
    sumZ1  +=       3.739 * cos(2.50010254963  + 30.05628079050 *T)
    sumZ1  +=       4.145 * cos(2.00938305966  + 44.07092647100 *T)
    sumZ1  +=       3.480 * cos(5.88971113590  + 38.02116105320 *T)
    sumZ1  +=       3.467 * cos(4.73412534395  + 38.24491022240 *T)
    sumZ1  +=       4.736 * cos(6.28313624461  + 149.56319713460 *T)
    sumZ1  +=       3.905 * cos(4.49400805134  + 106.97674337190 *T)
    sumZ1  +=       2.982 * cos(1.80386443362  + 46.20979048510 *T)
    sumZ1  +=       3.538 * cos(5.27114846878  + 37.87240320690 *T)
    sumZ1  +=       3.508 * cos(5.35267669439  + 38.39366806870 *T)
    sumZ1  +=       2.821 * cos(1.10242173566  + 33.94024994380 *T)
    sumZ1  +=       3.150 * cos(2.14792830412  + 115.88357962170 *T)
    sumZ1  +=       3.329 * cos(3.56226463770  + 181.75834193920 *T)
    sumZ1  +=       2.787 * cos(1.21334651843  + 72.07328558160 *T)
    sumZ1  +=       2.453 * cos(3.18401661435  + 42.32582133180 *T)
    sumZ1  +=       2.523 * cos(5.51669920239  + 8.07675484730 *T)

    # NEPTUNE Z Coord
    sumZ2  = 0
    sumZ2  +=  291361.164 * cos(5.57085222635  + 38.13303563780 *T)
    sumZ2  +=    2207.820 * cos(0.45423510946  + 36.64856292950 *T)
    sumZ2  +=    2644.401 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ2  +=    1184.984 * cos(3.62696666572  + 76.26607127560 *T)
    sumZ2  +=     875.873 * cos(1.60783110298  + 39.61750834610 *T)
    sumZ2  +=     253.280 * cos(2.25499665308  + 1.48447270830 *T)
    sumZ2  +=     139.875 * cos(1.66224942942  + 35.16409022120 *T)
    sumZ2  +=     105.837 * cos(5.61746558118  + 74.78159856730 *T)
    sumZ2  +=      55.479 * cos(0.51417309302  + 213.29909543800 *T)
    sumZ2  +=      51.858 * cos(1.20450519900  + 2.96894541660 *T)
    sumZ2  +=      40.183 * cos(1.45604293605  + 529.69096509460 *T)
    sumZ2  +=      41.153 * cos(0.64158589676  + 73.29712585900 *T)
    sumZ2  +=      13.931 * cos(0.27428875486  + 41.10198105440 *T)
    sumZ2  +=       9.181 * cos(2.83312722213  + 33.67961751290 *T)
    sumZ2  +=       7.421 * cos(4.32019600438  + 206.18554843720 *T)
    sumZ2  +=       6.998 * cos(5.87666052849  + 77.75054398390 *T)
    sumZ2  +=       7.085 * cos(1.66037085233  + 71.81265315070 *T)
    sumZ2  +=       6.818 * cos(1.56386946094  + 114.39910691340 *T)
    sumZ2  +=       4.838 * cos(0.83470875824  + 220.41264243880 *T)
    sumZ2  +=       2.545 * cos(0.43712705655  + 4.45341812490 *T)

    # NEPTUNE Z Coord
    sumZ3  = 0
    sumZ3  +=    8221.290 * cos(1.01632472042  + 38.13303563780 *T)
    sumZ3  +=     103.933 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ3  +=      70.154 * cos(2.36502685986  + 36.64856292950 *T)
    sumZ3  +=      31.617 * cos(5.35266161292  + 76.26607127560 *T)
    sumZ3  +=      15.470 * cos(3.21170859085  + 39.61750834610 *T)
    sumZ3  +=       7.111 * cos(3.99067059016  + 1.48447270830 *T)
    sumZ3  +=       4.656 * cos(3.62376309338  + 35.16409022120 *T)
    sumZ3  +=       2.988 * cos(1.03727330540  + 74.78159856730 *T)

    # NEPTUNE Z Coord
    sumZ4  = 0
    sumZ4  +=     172.227 * cos(2.66872693322  + 38.13303563780 *T)

    # NEPTUNE Z Coord
    sumZ5  = 0
    sumZ5  +=       3.394 * cos(4.70646877989  + 38.13303563780 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Neptune",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getNeptuneData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getNeptuneXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getSaturnXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Saturn"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getSaturnXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getSaturnXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # SATURN X Coord
    sumX0  = 0
    sumX0  +=  952312533.591 * cos(0.87401491487  + 213.54291292150 *T)
    sumX0  +=  79501390.398 * cos(4.76580713096  + 0.24381748350 *T)
    sumX0  +=  26427074.351 * cos(0.12339999915  + 426.84200835950 *T)
    sumX0  +=  6836881.382 * cos(4.14537914189  + 206.42936592071 *T)
    sumX0  +=  6628914.946 * cos(0.75057317755  + 220.65645992230 *T)
    sumX0  +=  2340967.916 * cos(2.01979283929  + 7.35736448430 *T)
    sumX0  +=  1250581.159 * cos(2.17392657526  + 110.45013870291 *T)
    sumX0  +=  1141539.711 * cos(3.03345312296  + 419.72846135871 *T)
    sumX0  +=  1098217.124 * cos(5.65720860592  + 640.14110379750 *T)
    sumX0  +=  773784.455 * cos(3.50285419449  + 213.05527795450 *T)
    sumX0  +=  708587.042 * cos(2.70922969272  + 316.63568714010 *T)
    sumX0  +=  434227.030 * cos(0.71624098563  + 529.93478257810 *T)
    sumX0  +=  373110.880 * cos(0.00103492669  + 433.95555536030 *T)
    sumX0  +=  413446.294 * cos(5.12392705512  + 102.84895673509 *T)
    sumX0  +=  169743.473 * cos(6.19449424436  + 205.94173095370 *T)
    sumX0  +=  122184.772 * cos(4.36825124039  + 103.33659170210 *T)
    sumX0  +=  141571.884 * cos(4.84926431365  + 419.24082639170 *T)
    sumX0  +=   97840.616 * cos(1.01545569510  + 323.74923414091 *T)
    sumX0  +=   80586.098 * cos(5.62239507032  + 11.28951774740 *T)
    sumX0  +=   83840.998 * cos(0.62002007163  + 227.77000692311 *T)
    sumX0  +=   72499.120 * cos(2.42001514073  + 633.02755679670 *T)
    sumX0  +=   70188.882 * cos(0.88664463373  + 209.61075965840 *T)
    sumX0  +=   68920.772 * cos(4.01663904384  + 217.47506618460 *T)
    sumX0  +=   65633.219 * cos(2.76404999091  + 202.49721265760 *T)
    sumX0  +=   58340.081 * cos(2.16106790491  + 224.58861318540 *T)
    sumX0  +=   54036.331 * cos(4.90882849644  + 853.44019923550 *T)
    sumX0  +=   45670.301 * cos(1.87822014512  + 14.47091148511 *T)
    sumX0  +=   39957.906 * cos(4.29286414022  + 199.31581891990 *T)
    sumX0  +=   44623.718 * cos(5.61332771211  + 63.49208081989 *T)
    sumX0  +=   25182.378 * cos(0.37762569174  + 216.72430665921 *T)
    sumX0  +=   24597.547 * cos(4.52837189964  + 210.36151918381 *T)
    sumX0  +=   24755.958 * cos(5.60048517853  + 415.79630809560 *T)
    sumX0  +=   25518.626 * cos(1.63814026860  + 117.56368570370 *T)
    sumX0  +=   30420.785 * cos(4.58036001246  + 735.63269604830 *T)
    sumX0  +=   22261.024 * cos(5.81339199632  + 522.33360061030 *T)
    sumX0  +=   23384.603 * cos(5.53454487341  + 647.25465079831 *T)
    sumX0  +=   19921.860 * cos(2.52438740442  + 426.35437339250 *T)
    sumX0  +=   23368.080 * cos(0.18801837763  + 149.80701461810 *T)
    sumX0  +=   24816.783 * cos(5.50044242809  + 75.02541605080 *T)
    sumX0  +=   15057.730 * cos(4.67637379107  + 277.27881122490 *T)
    sumX0  +=   13291.291 * cos(0.78826176081  + 309.03450517230 *T)
    sumX0  +=   12028.365 * cos(1.02934559031  + 1059.62574767270 *T)
    sumX0  +=   12408.238 * cos(4.19692495275  + 490.57790666290 *T)
    sumX0  +=   12037.260 * cos(5.66358935866  + 352.06040979221 *T)
    sumX0  +=    7880.259 * cos(2.60324364809  + 743.23387801611 *T)
    sumX0  +=    7428.823 * cos(4.26427762571  + 6.86972951729 *T)
    sumX0  +=    6985.779 * cos(3.88380796193  + 309.52214013930 *T)
    sumX0  +=    9059.376 * cos(4.31626349433  + 1052.02456570489 *T)
    sumX0  +=    6626.180 * cos(5.53666910204  + 838.72547026689 *T)
    sumX0  +=    8285.289 * cos(2.03028865010  + 95.73540973430 *T)
    sumX0  +=    7908.600 * cos(2.91633632242  + 316.14805217309 *T)
    sumX0  +=    6284.835 * cos(1.10258326823  + 3.68833577960 *T)
    sumX0  +=    6078.527 * cos(6.15871885410  + 441.06910236111 *T)
    sumX0  +=    5628.724 * cos(2.47826941808  + 846.32665223470 *T)
    sumX0  +=    5274.108 * cos(3.40288166935  + 220.16882495529 *T)
    sumX0  +=    5584.926 * cos(3.11230524128  + 412.61491435790 *T)
    sumX0  +=    4423.799 * cos(0.00561711784  + 198.82818395289 *T)
    sumX0  +=    4849.319 * cos(0.26891980614  + 537.04832957890 *T)
    sumX0  +=    4179.381 * cos(0.45517914787  + 4.17597074660 *T)
    sumX0  +=    5772.348 * cos(5.30786967696  + 38.37685312130 *T)
    sumX0  +=    5198.014 * cos(4.54573443270  + 211.09523236671 *T)
    sumX0  +=    5154.525 * cos(0.34613786659  + 215.99059347631 *T)
    sumX0  +=    3792.407 * cos(5.99326135433  + 422.90985509640 *T)
    sumX0  +=    4380.719 * cos(1.15696233721  + 202.00957769060 *T)
    sumX0  +=    4838.174 * cos(2.03403080585  + 109.96250373590 *T)
    sumX0  +=    3765.792 * cos(3.72073369090  + 212.57970507500 *T)
    sumX0  +=    3749.721 * cos(1.16925812858  + 214.50612076800 *T)
    sumX0  +=    3116.757 * cos(0.84596173223  + 213.49472881170 *T)
    sumX0  +=    3115.849 * cos(4.04374864852  + 213.59109703130 *T)
    sumX0  +=    3119.049 * cos(0.50950816465  + 625.42637482889 *T)
    sumX0  +=    3235.633 * cos(1.36565416054  + 632.53992182970 *T)
    sumX0  +=    4077.069 * cos(5.95933511847  + 136.78920667889 *T)
    sumX0  +=    3290.391 * cos(0.32712024501  + 522.82123557730 *T)
    sumX0  +=    2828.736 * cos(0.73223929578  + 330.86278114170 *T)
    sumX0  +=    3365.297 * cos(1.42074759163  + 437.88770862340 *T)
    sumX0  +=    2888.976 * cos(1.14066267703  + 86.07111631471 *T)
    sumX0  +=    2748.221 * cos(5.42931189203  + 138.76131435421 *T)
    sumX0  +=    2742.545 * cos(0.93548005709  + 203.98168536590 *T)
    sumX0  +=    2636.395 * cos(5.40701545041  + 288.32451148881 *T)
    sumX0  +=    3170.535 * cos(5.76636557141  + 76.50988875911 *T)
    sumX0  +=    2530.060 * cos(5.09730517007  + 629.09540353360 *T)
    sumX0  +=    2985.976 * cos(4.15644325876  + 1066.73929467350 *T)
    sumX0  +=    2617.358 * cos(5.45160906944  + 127.22797912329 *T)
    sumX0  +=    2504.577 * cos(1.49069183275  + 9.80504503910 *T)
    sumX0  +=    2459.199 * cos(5.26431553671  + 1155.11733992350 *T)
    sumX0  +=    2429.561 * cos(3.97256605696  + 223.10414047710 *T)
    sumX0  +=    3133.467 * cos(4.05731806661  + 1368.41643536150 *T)
    sumX0  +=    2289.072 * cos(3.68644263049  + 430.77416162260 *T)
    sumX0  +=    2651.422 * cos(5.01550706453  + 412.12727939089 *T)
    sumX0  +=    2791.618 * cos(1.76104978670  + 3.42521122121 *T)
    sumX0  +=    2086.473 * cos(3.54266754065  + 423.66061462181 *T)
    sumX0  +=    2337.517 * cos(5.86744835115  + 388.70897272171 *T)
    sumX0  +=    1735.774 * cos(5.67596672079  + 736.12033101530 *T)
    sumX0  +=    1617.760 * cos(4.44174561688  + 96.22304470130 *T)
    sumX0  +=    1635.420 * cos(4.54438827357  + 12.77399045571 *T)
    sumX0  +=    1637.578 * cos(2.19935832842  + 213.02164805970 *T)
    sumX0  +=    1633.497 * cos(2.69143754521  + 214.06417778330 *T)
    sumX0  +=    1703.439 * cos(2.42595077461  + 529.44714761109 *T)
    sumX0  +=    1499.174 * cos(3.59977145410  + 52.44638055600 *T)
    sumX0  +=    1462.063 * cos(5.92411007321  + 430.02340209721 *T)
    sumX0  +=    1975.850 * cos(2.02313657577  + 1.72829019180 *T)
    sumX0  +=    1422.179 * cos(4.89979449865  + 625.91400979590 *T)
    sumX0  +=    1539.334 * cos(4.78536735793  + 860.55374623631 *T)
    sumX0  +=    1400.317 * cos(3.72256803400  + 341.01470952831 *T)
    sumX0  +=    1337.694 * cos(6.25839826307  + 948.93179148630 *T)
    sumX0  +=    1427.589 * cos(5.41396429448  + 350.57593708390 *T)
    sumX0  +=    1149.795 * cos(0.20404714120  + 941.81824448550 *T)
    sumX0  +=    1112.160 * cos(0.47891380184  + 234.88355392390 *T)
    sumX0  +=    1015.231 * cos(4.99599119065  + 1471.50920958009 *T)
    sumX0  +=    1090.238 * cos(5.82704347671  + 363.10611005610 *T)
    sumX0  +=    1091.460 * cos(1.13704115272  + 174.18603700630 *T)
    sumX0  +=    1152.320 * cos(3.83459893564  + 201.01273994931 *T)
    sumX0  +=     874.317 * cos(5.95988532860  + 515.22005360950 *T)
    sumX0  +=    1180.011 * cos(3.80422455282  + 1684.80830501809 *T)
    sumX0  +=    1058.128 * cos(1.29841878389  + 742.74624304909 *T)
    sumX0  +=     850.822 * cos(4.55296025129  + 949.41942645330 *T)
    sumX0  +=     869.178 * cos(3.23528200552  + 210.62215161470 *T)
    sumX0  +=    1119.594 * cos(3.46591203463  + 703.87700210090 *T)
    sumX0  +=     868.384 * cos(1.65138396031  + 216.46367422830 *T)
    sumX0  +=     830.815 * cos(0.58110190109  + 212.79215339611 *T)
    sumX0  +=     825.441 * cos(4.30721839854  + 214.29367244690 *T)
    sumX0  +=     881.196 * cos(1.46906794476  + 209.35012722751 *T)
    sumX0  +=     866.345 * cos(3.41432821062  + 217.73569861551 *T)
    sumX0  +=     789.515 * cos(0.23961241737  + 223.83785366000 *T)
    sumX0  +=    1091.862 * cos(2.67346837599  + 340.52707456129 *T)
    sumX0  +=    1008.329 * cos(1.08917861648  + 226.07308589371 *T)
    sumX0  +=     919.088 * cos(4.86163770046  + 565.35950523021 *T)
    sumX0  +=     820.108 * cos(2.19486077696  + 88.62186273350 *T)
    sumX0  +=     716.162 * cos(5.81233216238  + 1265.32366114289 *T)
    sumX0  +=     860.928 * cos(3.70614582002  + 252.89978883670 *T)
    sumX0  +=     710.939 * cos(3.41349088844  + 417.28078080390 *T)
    sumX0  +=     701.898 * cos(1.91952059017  + 956.53297345411 *T)
    sumX0  +=     918.630 * cos(4.57690637056  + 563.87503252191 *T)
    sumX0  +=     711.475 * cos(0.90576159835  + 207.91383862900 *T)
    sumX0  +=     797.153 * cos(2.01469397505  + 208.12628695010 *T)
    sumX0  +=     758.732 * cos(4.77442040057  + 728.51914904750 *T)
    sumX0  +=     642.018 * cos(1.92693340665  + 639.65346883050 *T)
    sumX0  +=     769.629 * cos(2.89427987517  + 218.95953889290 *T)
    sumX0  +=     811.029 * cos(6.22246903964  + 1162.71852189130 *T)
    sumX0  +=     692.909 * cos(3.97674040211  + 219.17198721400 *T)
    sumX0  +=     798.757 * cos(1.35424808982  + 22.33521801130 *T)
    sumX0  +=     834.309 * cos(2.96244339722  + 160.85271488200 *T)
    sumX0  +=     655.277 * cos(0.92079690947  + 301.92095817150 *T)
    sumX0  +=     762.632 * cos(1.72094446975  + 21.58445848590 *T)
    sumX0  +=     576.036 * cos(3.10166158268  + 99.40443843901 *T)
    sumX0  +=     615.654 * cos(2.27550746905  + 845.83901726770 *T)
    sumX0  +=     578.589 * cos(0.80405181178  + 265.74547599400 *T)
    sumX0  +=     626.782 * cos(2.79771540666  + 18.40306474821 *T)
    sumX0  +=     589.499 * cos(3.90740268282  + 554.31380496631 *T)
    sumX0  +=     700.182 * cos(0.40018815531  + 62.00760811160 *T)
    sumX0  +=     697.691 * cos(0.57105547911  + 124.67723270450 *T)
    sumX0  +=     613.843 * cos(4.42239495907  + 218.20877936750 *T)
    sumX0  +=     581.263 * cos(2.07871172739  + 231.70216018621 *T)
    sumX0  +=     666.749 * cos(2.82369459311  + 425.35753565121 *T)
    sumX0  +=     516.903 * cos(0.38857203410  + 414.31183538731 *T)
    sumX0  +=     499.031 * cos(2.54039255853  + 408.68276109481 *T)
    sumX0  +=     502.357 * cos(6.19770250859  + 1258.21011414209 *T)
    sumX0  +=     531.983 * cos(2.53130210328  + 215.02738562980 *T)
    sumX0  +=     643.415 * cos(1.39270825866  + 64.97655352819 *T)
    sumX0  +=     518.385 * cos(2.70002198759  + 142.69346761730 *T)
    sumX0  +=     550.972 * cos(0.47047442681  + 208.87704647550 *T)
    sumX0  +=     469.313 * cos(0.32930418795  + 71.09326278771 *T)
    sumX0  +=     456.002 * cos(4.67119057391  + 203.24797218300 *T)
    sumX0  +=     592.362 * cos(1.86789317281  + 266.23311096100 *T)
    sumX0  +=     465.459 * cos(5.40811702089  + 654.36819779910 *T)
    sumX0  +=     444.956 * cos(4.73192251220  + 1787.90107923670 *T)
    sumX0  +=     502.683 * cos(4.37046140723  + 251.67594855930 *T)
    sumX0  +=     522.764 * cos(2.19701769819  + 212.05844021321 *T)
    sumX0  +=     422.458 * cos(2.13565038243  + 479.53220639900 *T)
    sumX0  +=     437.134 * cos(1.32836524720  + 1589.31671276730 *T)
    sumX0  +=     414.039 * cos(3.63548652605  + 148.32254190981 *T)
    sumX0  +=     398.994 * cos(1.64973433253  + 213.65478750610 *T)
    sumX0  +=     398.994 * cos(3.23989183406  + 213.43103833690 *T)
    sumX0  +=     373.088 * cos(5.46010584090  + 1581.71553079950 *T)
    sumX0  +=     469.097 * cos(3.55737355039  + 2001.20017467470 *T)
    sumX0  +=     361.258 * cos(4.90274855427  + 302.40859313851 *T)
    sumX0  +=     345.920 * cos(5.01286731480  + 198.56505939450 *T)
    sumX0  +=     350.777 * cos(5.86079095869  + 98.65617104110 *T)
    sumX0  +=     335.274 * cos(1.03009229022  + 213.75536124260 *T)
    sumX0  +=     335.274 * cos(3.85953387637  + 213.33046460040 *T)
    sumX0  +=     367.346 * cos(5.91088479684  + 750.34742501690 *T)
    sumX0  +=     369.973 * cos(5.39933070480  + 151.29148732640 *T)
    sumX0  +=     399.230 * cos(1.04833113678  + 1059.13811270570 *T)
    sumX0  +=     318.144 * cos(6.15035386620  + 228.52076644850 *T)
    sumX0  +=     315.586 * cos(5.68724533460  + 831.61192326610 *T)
    sumX0  +=     337.834 * cos(0.66798086530  + 175.40987728371 *T)
    sumX0  +=     313.433 * cos(3.54547785742  + 2.20386307129 *T)
    sumX0  +=     330.040 * cos(1.38038349450  + 276.79117625789 *T)
    sumX0  +=     383.783 * cos(2.57160924287  + 56.37853381910 *T)
    sumX0  +=     350.135 * cos(4.75975448170  + 114.38229196601 *T)
    sumX0  +=     401.239 * cos(4.28930925743  + 178.11819026941 *T)
    sumX0  +=     404.631 * cos(6.03696917524  + 1479.11039154791 *T)
    sumX0  +=     404.539 * cos(1.52309497166  + 311.95526647910 *T)
    sumX0  +=     304.876 * cos(6.11535489175  + 195.38366565681 *T)
    sumX0  +=     289.344 * cos(4.48279316993  + 192.20227191910 *T)
    sumX0  +=     270.250 * cos(5.14035236650  + 636.20895053440 *T)
    sumX0  +=     271.661 * cos(1.28934639885  + 278.76328393321 *T)
    sumX0  +=     283.573 * cos(0.04390160982  + 191.45151239371 *T)
    sumX0  +=     294.112 * cos(1.97767959737  + 1162.23088692429 *T)
    sumX0  +=     266.932 * cos(5.65493346800  + 205.46615807421 *T)
    sumX0  +=     269.475 * cos(2.14597987484  + 327.68138740401 *T)
    sumX0  +=     348.053 * cos(0.61890667474  + 248.96763557360 *T)
    sumX0  +=     296.599 * cos(1.51525625285  + 10.53875822200 *T)
    sumX0  +=     252.084 * cos(1.67555502838  + 213.80354535241 *T)
    sumX0  +=     252.084 * cos(3.21407113821  + 213.28228049061 *T)
    sumX0  +=     250.282 * cos(2.61655820302  + 433.46792039329 *T)
    sumX0  +=     294.321 * cos(2.01870159642  + 106.51798543980 *T)
    sumX0  +=     270.973 * cos(4.53462056041  + 1044.91101870410 *T)
    sumX0  +=     263.820 * cos(3.76320728541  + 63.97971578690 *T)
    sumX0  +=     238.275 * cos(5.91316475830  + 1574.60198379870 *T)
    sumX0  +=     305.244 * cos(1.10529286976  + 483.46435966210 *T)
    sumX0  +=     241.411 * cos(5.74870980326  + 221.61966776881 *T)
    sumX0  +=     314.980 * cos(4.37056119892  + 479.04457143200 *T)
    sumX0  +=     223.277 * cos(4.01771641783  + 13.98327651810 *T)
    sumX0  +=     292.846 * cos(3.86376968180  + 424.39432780471 *T)
    sumX0  +=     304.942 * cos(1.75360932249  + 6283.31966747490 *T)
    sumX0  +=     237.167 * cos(4.96019200454  + 235.63431344930 *T)
    sumX0  +=     240.608 * cos(5.75032033143  + 275.79433851660 *T)
    sumX0  +=     263.455 * cos(5.51056104098  + 70.60562782070 *T)
    sumX0  +=     211.236 * cos(0.02154252782  + 544.16187657970 *T)
    sumX0  +=     205.878 * cos(4.46555623121  + 842.39449897160 *T)
    sumX0  +=     282.395 * cos(3.56598534931  + 404.75060783170 *T)
    sumX0  +=     203.077 * cos(4.47228022052  + 2104.29294889330 *T)
    sumX0  +=     267.955 * cos(3.49559055839  + 121.49583896680 *T)
    sumX0  +=     238.411 * cos(4.99775321682  + 138.27367938720 *T)
    sumX0  +=     230.884 * cos(4.03701073645  + 497.69145366371 *T)
    sumX0  +=     203.254 * cos(2.93667823303  + 284.39235822571 *T)
    sumX0  +=     216.423 * cos(2.44114933526  + 8.32057233081 *T)
    sumX0  +=     194.279 * cos(5.27870729321  + 190.96387742670 *T)
    sumX0  +=     216.822 * cos(6.14885127819  + 429.28968891431 *T)
    sumX0  +=     191.411 * cos(5.14394229637  + 1898.10740045609 *T)
    sumX0  +=     212.655 * cos(0.67418560513  + 651.18680406140 *T)
    sumX0  +=     214.969 * cos(2.97753864449  + 425.87880051300 *T)
    sumX0  +=     236.822 * cos(3.57393144724  + 219.69325207580 *T)
    sumX0  +=     229.348 * cos(1.65487599461  + 270.16526422410 *T)
    sumX0  +=     174.163 * cos(3.29103993348  + 426.89019246930 *T)
    sumX0  +=     202.398 * cos(2.42508237402  + 39.60069339870 *T)
    sumX0  +=     179.962 * cos(2.48812846406  + 404.26297286470 *T)
    sumX0  +=     169.610 * cos(4.78044354700  + 501.62360692681 *T)
    sumX0  +=     174.157 * cos(0.09324510966  + 426.79382424970 *T)
    sumX0  +=     165.778 * cos(0.76098657662  + 312.70353387701 *T)
    sumX0  +=     178.798 * cos(4.30679581382  + 355.99256305531 *T)
    sumX0  +=     180.904 * cos(0.63787759163  + 618.31282782810 *T)
    sumX0  +=     182.548 * cos(1.24685071391  + 210.83459993580 *T)
    sumX0  +=     179.529 * cos(3.64276765573  + 216.25122590720 *T)
    sumX0  +=     187.071 * cos(0.50925452549  + 427.80521620600 *T)
    sumX0  +=     180.856 * cos(3.08033197166  + 84.58664360640 *T)
    sumX0  +=     196.496 * cos(5.26678343139  + 489.09343395460 *T)
    sumX0  +=     151.676 * cos(0.75207673924  + 213.38285422711 *T)
    sumX0  +=     151.676 * cos(4.13754942735  + 213.70297161591 *T)
    sumX0  +=     151.430 * cos(3.90513550302  + 220.70464403210 *T)
    sumX0  +=     193.773 * cos(3.31690579880  + 2317.59204433130 *T)
    sumX0  +=     166.996 * cos(3.40695394996  + 1280.03839011150 *T)
    sumX0  +=     189.342 * cos(1.32639271163  + 207.39257376720 *T)
    sumX0  +=     180.605 * cos(0.04955573118  + 492.06237937121 *T)
    sumX0  +=     172.720 * cos(5.73529325856  + 1052.51220067191 *T)
    sumX0  +=     149.398 * cos(4.61667825955  + 839.21310523390 *T)
    sumX0  +=     146.617 * cos(0.72929791759  + 220.60827581251 *T)
    sumX0  +=     193.842 * cos(4.19725789426  + 189.16389053330 *T)
    sumX0  +=     178.668 * cos(2.61010005515  + 344.94686279140 *T)
    sumX0  +=     137.031 * cos(5.42161597844  + 1148.00379292269 *T)
    sumX0  +=     185.345 * cos(5.12406505593  + 602.00806815971 *T)
    sumX0  +=     158.449 * cos(3.42413570789  + 179.60266297770 *T)
    sumX0  +=     167.315 * cos(3.34458985254  + 73.54094334250 *T)
    sumX0  +=     155.890 * cos(4.99214316468  + 289.80898419710 *T)
    sumX0  +=     159.567 * cos(5.53565697352  + 359.17395679300 *T)
    sumX0  +=     138.554 * cos(3.19924734717  + 436.40323591510 *T)
    sumX0  +=     164.796 * cos(0.80382302783  + 1375.52998236229 *T)
    sumX0  +=     167.034 * cos(1.61165618369  + 142.20583265029 *T)
    sumX0  +=     128.473 * cos(5.80762617164  + 107.26874496520 *T)
    sumX0  +=     142.270 * cos(2.75171852816  + 636.95971005981 *T)
    sumX0  +=     177.375 * cos(3.17652608660  + 10213.52936369450 *T)
    sumX0  +=     125.618 * cos(3.34222854569  + 644.07325706060 *T)
    sumX0  +=     131.798 * cos(2.50083511264  + 212.27088853431 *T)
    sumX0  +=     130.711 * cos(2.38850764149  + 214.81493730871 *T)
    sumX0  +=     122.905 * cos(1.72530780843  + 621.98185653281 *T)
    sumX0  +=     165.088 * cos(0.66708846026  + 237.92193530970 *T)
    sumX0  +=     133.438 * cos(5.65562400570  + 4.90968392950 *T)
    sumX0  +=     120.114 * cos(4.72829633676  + 194.89603068979 *T)
    sumX0  +=     161.862 * cos(5.80811726650  + 1795.50226120450 *T)
    sumX0  +=     116.787 * cos(1.61029914973  + 113.63153244060 *T)
    sumX0  +=     118.210 * cos(5.63285421725  + 1890.99385345530 *T)
    sumX0  +=     133.448 * cos(1.56209297419  + 1478.62275658089 *T)
    sumX0  +=     111.758 * cos(0.99483567378  + 206.38118181090 *T)
    sumX0  +=     152.292 * cos(0.56938816562  + 515.70768857651 *T)
    sumX0  +=     128.840 * cos(1.45944172461  + 247.48316286530 *T)
    sumX0  +=     110.215 * cos(5.22553112544  + 5.87289177600 *T)
    sumX0  +=     107.434 * cos(6.23762101588  + 216.98743121760 *T)
    sumX0  +=     115.948 * cos(3.80548274408  + 31.99951143090 *T)
    sumX0  +=     109.063 * cos(5.47169247190  + 114.64292439690 *T)
    sumX0  +=     109.273 * cos(0.74841730668  + 127.71561409031 *T)
    sumX0  +=     107.429 * cos(1.81399388771  + 767.61290040431 *T)
    sumX0  +=     126.160 * cos(4.37110941347  + 181.29958400710 *T)
    sumX0  +=     110.176 * cos(4.18234578108  + 10.80188278040 *T)
    sumX0  +=     123.129 * cos(1.94869447093  + 2.93757625420 *T)
    sumX0  +=     134.189 * cos(2.76336361412  + 35.66854013560 *T)
    sumX0  +=     106.815 * cos(4.16300130274  + 206.47755003051 *T)
    sumX0  +=     105.829 * cos(1.45649238999  + 220.13519506050 *T)
    sumX0  +=     115.432 * cos(1.64246857380  + 415.30867312860 *T)
    sumX0  +=     118.293 * cos(0.52518586111  + 245.78624183590 *T)
    sumX0  +=     104.354 * cos(4.03316323840  + 1073.85284167431 *T)
    sumX0  +=      95.114 * cos(5.17002113254  + 643.32249753521 *T)
    sumX0  +=      99.528 * cos(1.42297994879  + 426.32074349770 *T)
    sumX0  +=      97.935 * cos(4.84536515851  + 2214.49927011270 *T)
    sumX0  +=     106.628 * cos(0.81030389220  + 39.11305843170 *T)
    sumX0  +=      94.853 * cos(4.21773911052  + 2420.68481854989 *T)
    sumX0  +=      96.075 * cos(6.02681746488  + 448.18264936190 *T)
    sumX0  +=     104.203 * cos(4.30869132064  + 1361.30288836070 *T)
    sumX0  +=     107.375 * cos(2.39490064055  + 78.95756931390 *T)
    sumX0  +=      93.096 * cos(2.12427858259  + 209.12312469139 *T)
    sumX0  +=      88.042 * cos(1.47468327901  + 67.42423408300 *T)
    sumX0  +=      93.076 * cos(2.33041461145  + 69.39634175830 *T)
    sumX0  +=      89.875 * cos(5.49491788339  + 134.82916109110 *T)
    sumX0  +=      86.729 * cos(4.33693429388  + 0.71939036300 *T)
    sumX0  +=      88.055 * cos(0.05558729242  + 87.55558902300 *T)
    sumX0  +=      86.045 * cos(4.47959291295  + 525.74199688410 *T)
    sumX0  +=      84.186 * cos(0.68410696577  + 144.39038864671 *T)
    sumX0  +=      81.866 * cos(3.25030344825  + 20.85074530300 *T)
    sumX0  +=      98.341 * cos(1.53715431353  + 4.43660317750 *T)
    sumX0  +=     112.464 * cos(2.14098849805  + 98.91680347199 *T)
    sumX0  +=      91.321 * cos(1.93807777759  + 427.36327322130 *T)
    sumX0  +=      88.237 * cos(0.32719875572  + 209.87388421679 *T)
    sumX0  +=      91.544 * cos(3.59720610688  + 206.95063078250 *T)
    sumX0  +=      79.761 * cos(4.83472817526  + 159.36824217371 *T)
    sumX0  +=     100.440 * cos(3.54881368140  + 140.24578706250 *T)
    sumX0  +=      78.269 * cos(3.24353670896  + 14.25846316400 *T)
    sumX0  +=      93.714 * cos(2.83930960133  + 203.49405039889 *T)
    sumX0  +=      77.457 * cos(0.17608698949  + 191.71463695210 *T)
    sumX0  +=      87.870 * cos(1.80295838343  + 342.49918223660 *T)
    sumX0  +=      80.557 * cos(0.23446317702  + 125.74350641500 *T)
    sumX0  +=      82.279 * cos(4.06810365664  + 91.80325647120 *T)
    sumX0  +=     100.767 * cos(2.20293937930  + 200.52510498229 *T)
    sumX0  +=     102.819 * cos(6.06771895402  + 617.56206830270 *T)
    sumX0  +=      74.922 * cos(6.18045942087  + 421.21293406700 *T)
    sumX0  +=      74.704 * cos(0.69430051027  + 1582.20316576650 *T)
    sumX0  +=      75.306 * cos(2.02562031906  + 1376.01761732930 *T)
    sumX0  +=      71.397 * cos(2.45853219243  + 630.57987624190 *T)
    sumX0  +=      70.299 * cos(5.82879783028  + 5.66044345490 *T)
    sumX0  +=      95.210 * cos(2.04442457798  + 1272.92484311070 *T)
    sumX0  +=      69.326 * cos(5.19530696709  + 214.76675319891 *T)
    sumX0  +=      69.326 * cos(5.97750450668  + 212.31907264410 *T)
    sumX0  +=      76.148 * cos(0.34812529439  + 934.70469748469 *T)
    sumX0  +=      80.760 * cos(2.61290558453  + 15.22167101050 *T)
    sumX0  +=      85.274 * cos(1.61810289570  + 53.93085326429 *T)
    sumX0  +=      79.268 * cos(4.26191238810  + 153.73916788121 *T)
    sumX0  +=      81.890 * cos(1.59839688347  + 353.54488250050 *T)
    sumX0  +=      91.253 * cos(3.81884583659  + 777.17412795991 *T)
    sumX0  +=      70.965 * cos(0.93242756417  + 1265.81129610991 *T)
    sumX0  +=      86.364 * cos(2.72622886800  + 917.17609753890 *T)
    sumX0  +=      81.660 * cos(0.09321047637  + 5.17280848790 *T)
    sumX0  +=      66.327 * cos(3.27691742733  + 227.28237195610 *T)
    sumX0  +=      66.404 * cos(2.35001720349  + 423.92124705270 *T)
    sumX0  +=      68.065 * cos(6.26038360145  + 692.83130183700 *T)
    sumX0  +=      81.793 * cos(3.08308557258  + 2633.98391398789 *T)
    sumX0  +=      70.972 * cos(6.18608914757  + 211.84599189210 *T)
    sumX0  +=      65.212 * cos(4.81833711198  + 6.39415663780 *T)
    sumX0  +=      64.142 * cos(5.15573992685  + 1464.39566257930 *T)
    sumX0  +=      63.961 * cos(0.28228675547  + 185.23173727020 *T)
    sumX0  +=      65.617 * cos(4.17315140812  + 542.78046973539 *T)
    sumX0  +=      62.731 * cos(3.35292072076  + 320.56784040320 *T)
    sumX0  +=      73.401 * cos(1.42622919650  + 286.84003878050 *T)
    sumX0  +=      63.416 * cos(2.84649923006  + 3.16457879030 *T)
    sumX0  +=      61.908 * cos(0.61695293583  + 329.48137429739 *T)
    sumX0  +=      77.192 * cos(1.20473000691  + 46.71424039951 *T)
    sumX0  +=      70.261 * cos(0.02744626680  + 337.97632814250 *T)
    sumX0  +=      58.935 * cos(3.09359530816  + 128.71245183160 *T)
    sumX0  +=      59.359 * cos(4.68839528191  + 50.96190784770 *T)
    sumX0  +=      71.174 * cos(0.94418960401  + 74.53778108379 *T)
    sumX0  +=      59.979 * cos(5.36194030187  + 2207.38572311189 *T)
    sumX0  +=      78.873 * cos(6.18668375333  + 299.37021175271 *T)
    sumX0  +=      60.704 * cos(0.73825309757  + 182.52342428450 *T)
    sumX0  +=      63.502 * cos(2.38997253041  + 391.41728570740 *T)
    sumX0  +=      60.272 * cos(3.53054844171  + 416.54706762100 *T)
    sumX0  +=      56.662 * cos(0.03126623890  + 267.71758366931 *T)
    sumX0  +=      66.008 * cos(4.08045695540  + 778.65860066820 *T)
    sumX0  +=      66.706 * cos(2.56096131990  + 221.17772478411 *T)
    sumX0  +=      66.363 * cos(0.16881593879  + 696.76345510010 *T)
    sumX0  +=      55.245 * cos(5.78088729365  + 437.13694909800 *T)
    sumX0  +=      55.098 * cos(1.58209601288  + 387.48513244430 *T)
    sumX0  +=      71.893 * cos(0.57617376095  + 1691.92185201890 *T)
    sumX0  +=      59.023 * cos(4.97848918827  + 215.23983395090 *T)
    sumX0  +=      54.012 * cos(6.18919211339  + 627.61093082531 *T)
    sumX0  +=      53.837 * cos(4.52951978106  + 241.85408857280 *T)
    sumX0  +=      63.323 * cos(5.07913116250  + 576.40520549410 *T)
    sumX0  +=      67.784 * cos(3.40092497147  + 116.18227885939 *T)
    sumX0  +=      68.327 * cos(1.18421764248  + 1795.01462623750 *T)
    sumX0  +=      74.859 * cos(5.67759501336  + 2.69149803831 *T)
    sumX0  +=      54.888 * cos(5.74041329041  + 312.44290144610 *T)
    sumX0  +=      54.393 * cos(4.49204534403  + 134.34152612410 *T)
    sumX0  +=      56.250 * cos(0.73800257133  + 273.34665796181 *T)
    sumX0  +=      56.864 * cos(1.30791461253  + 445.00125562421 *T)
    sumX0  +=      58.600 * cos(5.77716291851  + 323.26159917390 *T)
    sumX0  +=      50.765 * cos(4.71246376027  + 224.10097821840 *T)
    sumX0  +=      51.202 * cos(1.92483709708  + 7.30918037451 *T)
    sumX0  +=      52.403 * cos(1.22494346070  + 518.88908231420 *T)
    sumX0  +=      64.155 * cos(5.72635780004  + 422.17614191350 *T)
    sumX0  +=      50.427 * cos(6.01983026969  + 214.97920152001 *T)
    sumX0  +=      50.427 * cos(5.15298120408  + 212.10662432301 *T)
    sumX0  +=      62.900 * cos(0.76074425315  + 3.21276290011 *T)
    sumX0  +=      55.902 * cos(3.82234323799  + 1.24065522479 *T)
    sumX0  +=      50.047 * cos(5.69562605727  + 292.25666475190 *T)
    sumX0  +=      49.142 * cos(3.45523655555  + 618.04970326970 *T)
    sumX0  +=      55.589 * cos(5.00243604255  + 221.40721944771 *T)
    sumX0  +=      50.005 * cos(4.56446098577  + 2530.89113976930 *T)
    sumX0  +=      59.759 * cos(0.33324455917  + 439.37218133171 *T)
    sumX0  +=      47.716 * cos(0.39295849664  + 219.90570039690 *T)
    sumX0  +=      63.636 * cos(5.56781141579  + 2111.89413086110 *T)
    sumX0  +=      48.821 * cos(1.11954873085  + 30.77567115350 *T)
    sumX0  +=      49.427 * cos(4.10715723619  + 282.69543719630 *T)
    sumX0  +=      64.684 * cos(1.40586799806  + 305.58998687620 *T)
    sumX0  +=      49.622 * cos(2.64477674752  + 431.03479405351 *T)
    sumX0  +=      58.187 * cos(2.85345969595  + 348.12825652910 *T)
    sumX0  +=      48.661 * cos(3.16552973997  + 97.11681657861 *T)
    sumX0  +=      48.710 * cos(0.89401886908  + 429.76276966630 *T)
    sumX0  +=      65.208 * cos(3.36533374596  + 487.12132627930 *T)
    sumX0  +=      46.209 * cos(6.11163408466  + 426.09124883410 *T)
    sumX0  +=      58.609 * cos(1.53711331943  + 1169.83206889211 *T)
    sumX0  +=      55.906 * cos(2.58196718879  + 418.76525351221 *T)
    sumX0  +=      51.943 * cos(5.13191462514  + 199.52826724100 *T)
    sumX0  +=      54.149 * cos(5.95504378991  + 729.00678401450 *T)
    sumX0  +=      44.466 * cos(3.96307178488  + 2737.07668820650 *T)
    sumX0  +=      55.947 * cos(2.08615316919  + 638.65663108921 *T)
    sumX0  +=      43.290 * cos(4.80484052860  + 339.53023682000 *T)
    sumX0  +=      42.587 * cos(0.60370993867  + 421.42538238810 *T)
    sumX0  +=      52.900 * cos(6.04236572471  + 205.67860639531 *T)
    sumX0  +=      41.439 * cos(2.47464431042  + 120.11443212250 *T)
    sumX0  +=      47.251 * cos(1.55094667019  + 558.24595822940 *T)
    sumX0  +=      48.495 * cos(5.11989933898  + 7.40554859410 *T)
    sumX0  +=      47.908 * cos(4.49005771863  + 192.93598510200 *T)
    sumX0  +=      44.030 * cos(1.76632610072  + 24.13520490470 *T)
    sumX0  +=      45.108 * cos(6.06459427723  + 227.55755860200 *T)
    sumX0  +=      46.355 * cos(3.10853873073  + 432.47108265200 *T)
    sumX0  +=      53.109 * cos(4.45380505884  + 167.07249000551 *T)
    sumX0  +=      45.868 * cos(2.33132835068  + 205.90810105891 *T)
    sumX0  +=      41.458 * cos(4.11197539530  + 1677.69475801730 *T)
    sumX0  +=      37.152 * cos(4.50776996086  + 207.18012544610 *T)
    sumX0  +=      47.404 * cos(3.38391475216  + 6.83609962250 *T)
    sumX0  +=      38.168 * cos(4.08180868423  + 244.56240155850 *T)
    sumX0  +=      36.628 * cos(3.66787249773  + 427.59276788490 *T)
    sumX0  +=      45.029 * cos(2.03211615844  + 428.32648106781 *T)
    sumX0  +=      36.896 * cos(0.49263625994  + 1898.59503542310 *T)
    sumX0  +=      38.453 * cos(5.99529967870  + 209.39831133730 *T)
    sumX0  +=      48.446 * cos(1.21352554632  + 68.90870679129 *T)
    sumX0  +=      38.820 * cos(3.98768004395  + 117.07605073670 *T)
    sumX0  +=      40.434 * cos(0.44979537196  + 234.14984074100 *T)
    sumX0  +=      36.263 * cos(1.46717537926  + 273.82223084130 *T)
    sumX0  +=      49.643 * cos(6.20072079882  + 3340.85624418330 *T)
    sumX0  +=      35.063 * cos(3.13493629794  + 34.18406742730 *T)
    sumX0  +=      37.471 * cos(5.17760779965  + 217.68751450571 *T)
    sumX0  +=      46.113 * cos(0.83857991425  + 212.64914104420 *T)
    sumX0  +=      46.113 * cos(4.05104625234  + 214.43668479881 *T)
    sumX0  +=      37.195 * cos(1.57230760584  + 41.40068029210 *T)
    sumX0  +=      36.077 * cos(4.65555472093  + 867.66729323710 *T)
    sumX0  +=      37.351 * cos(0.06652028564  + 1251.09656714130 *T)
    sumX0  +=      34.230 * cos(1.60559863103  + 182.05034353251 *T)
    sumX0  +=      41.145 * cos(0.98859396839  + 33.69643246030 *T)
    sumX0  +=      40.662 * cos(2.49012270106  + 59.55992755679 *T)
    sumX0  +=      43.402 * cos(3.89634394974  + 135.30473397060 *T)
    sumX0  +=      39.654 * cos(3.59602932397  + 113.89465699899 *T)
    sumX0  +=      36.261 * cos(2.48589063916  + 28.06735816780 *T)
    sumX0  +=      32.687 * cos(2.96330012868  + 956.04533848709 *T)
    sumX0  +=      33.157 * cos(5.57688316935  + 89.10949770051 *T)
    sumX0  +=      36.698 * cos(3.68604273428  + 431.50787480550 *T)
    sumX0  +=      33.531 * cos(2.34303179257  + 81.50831573269 *T)
    sumX0  +=      31.599 * cos(2.62632521648  + 553.82616999929 *T)
    sumX0  +=      31.098 * cos(4.89394578356  + 1780.78753223589 *T)
    sumX0  +=      37.751 * cos(6.02905345714  + 319.81708087780 *T)
    sumX0  +=      35.774 * cos(6.21453795302  + 137.27684164590 *T)
    sumX0  +=      39.730 * cos(0.04188161730  + 739.04109232210 *T)
    sumX0  +=      38.350 * cos(0.00615623981  + 145.87486135500 *T)
    sumX0  +=      34.642 * cos(1.89884834343  + 77.99436146740 *T)
    sumX0  +=      31.570 * cos(3.30002041912  + 466.19888427470 *T)
    sumX0  +=      39.360 * cos(2.19565810576  + 374.15181032000 *T)
    sumX0  +=      31.496 * cos(0.96043262389  + 24.62283987170 *T)
    sumX0  +=      30.207 * cos(5.08475977345  + 2523.77759276850 *T)
    sumX0  +=      29.335 * cos(6.18456452708  + 131.79077970531 *T)
    sumX0  +=      40.419 * cos(0.63548950495  + 1692.40948698591 *T)
    sumX0  +=      31.724 * cos(0.72175533744  + 333.90116252750 *T)
    sumX0  +=      34.923 * cos(2.85187519613  + 2950.37578364450 *T)
    sumX0  +=      28.746 * cos(1.71197970691  + 406.23508054000 *T)
    sumX0  +=      28.755 * cos(0.15837943003  + 193.09604379641 *T)
    sumX0  +=      35.049 * cos(2.84712825648  + 381.59542572090 *T)
    sumX0  +=      27.298 * cos(0.86827415264  + 60.52313540329 *T)
    sumX0  +=      35.019 * cos(0.26058304344  + 23.81969071961 *T)
    sumX0  +=      27.836 * cos(1.17681685755  + 216.51185833811 *T)
    sumX0  +=      36.108 * cos(1.48596443213  + 756.07956517339 *T)
    sumX0  +=      29.306 * cos(3.55420436272  + 464.97504399731 *T)
    sumX0  +=      27.049 * cos(2.92553690245  + 361.62163734781 *T)
    sumX0  +=      34.876 * cos(1.11340331736  + 39.86132582961 *T)
    sumX0  +=      27.398 * cos(3.07953286472  + 411.86415483250 *T)
    sumX0  +=      26.699 * cos(3.34823660722  + 313.45429340240 *T)
    sumX0  +=      31.423 * cos(1.41211589868  + 0.50694204189 *T)
    sumX0  +=      33.616 * cos(5.77820346759  + 395.82251972250 *T)
    sumX0  +=      26.335 * cos(3.73861907459  + 0.76508234530 *T)
    sumX0  +=      26.583 * cos(3.25633746796  + 245.03548231051 *T)
    sumX0  +=      32.446 * cos(4.84330341651  + 423.17297965479 *T)
    sumX0  +=      35.965 * cos(0.81545997217  + 2111.40649589409 *T)
    sumX0  +=      27.132 * cos(3.74886922172  + 60.04756252381 *T)
    sumX0  +=      25.856 * cos(4.81175731698  + 16.91859203990 *T)
    sumX0  +=      29.986 * cos(3.32251668307  + 710.99054910171 *T)
    sumX0  +=      32.943 * cos(3.73798438144  + 339.04260185300 *T)
    sumX0  +=      27.965 * cos(0.70019337757  + 852.95256426850 *T)
    sumX0  +=      26.667 * cos(3.61851114795  + 210.57396750490 *T)
    sumX0  +=      28.346 * cos(1.92868741776  + 281.21096448800 *T)
    sumX0  +=      25.306 * cos(4.73124679745  + 233.98978204660 *T)
    sumX0  +=      25.371 * cos(1.69000982510  + 412.82736267900 *T)
    sumX0  +=      25.734 * cos(2.25578963103  + 459.08533727390 *T)
    sumX0  +=      26.331 * cos(0.45420938431  + 494.51005992600 *T)
    sumX0  +=      30.392 * cos(2.95808144330  + 129.67565967810 *T)
    sumX0  +=      25.998 * cos(5.85713962580  + 681.29796660610 *T)
    sumX0  +=      33.249 * cos(1.88106372870  + 162.33718759030 *T)
    sumX0  +=      34.168 * cos(0.39863588958  + 260.01333583751 *T)
    sumX0  +=      33.256 * cos(4.73243568540  + 342.01154726960 *T)
    sumX0  +=      26.237 * cos(0.63347048675  + 45.33283355520 *T)
    sumX0  +=      24.623 * cos(5.09765340751  + 618.80046279510 *T)
    sumX0  +=      33.518 * cos(2.30165851435  + 26.84351789039 *T)
    sumX0  +=      27.190 * cos(6.13268298688  + 211.89838151880 *T)
    sumX0  +=      27.190 * cos(5.04012848689  + 215.18744432420 *T)
    sumX0  +=      29.507 * cos(1.24794264854  + 455.15318401080 *T)
    sumX0  +=      32.634 * cos(0.34866766006  + 2008.31372167550 *T)
    sumX0  +=      25.187 * cos(2.51192814934  + 1272.43720814370 *T)
    sumX0  +=      25.644 * cos(3.86292975641  + 230.21768747790 *T)
    sumX0  +=      24.787 * cos(4.28010726699  + 2847.28300942589 *T)
    sumX0  +=      23.670 * cos(1.51334180608  + 106.03035047280 *T)
    sumX0  +=      23.511 * cos(4.48384165181  + 214.34606207361 *T)
    sumX0  +=      23.511 * cos(0.40578451478  + 212.73976376941 *T)
    sumX0  +=      23.747 * cos(2.29348970322  + 467.99887116810 *T)
    sumX0  +=      30.107 * cos(3.53377474522  + 408.19512612779 *T)
    sumX0  +=      27.532 * cos(2.92483319573  + 430.28652665560 *T)
    sumX0  +=      24.105 * cos(1.45469526113  + 422.64922266551 *T)
    sumX0  +=      24.814 * cos(1.67549648860  + 329.96900926440 *T)
    sumX0  +=      25.569 * cos(3.07026183070  + 174.92224231670 *T)
    sumX0  +=      24.522 * cos(0.25792931265  + 420.69166920520 *T)
    sumX0  +=      23.164 * cos(2.54403428184  + 245.29860686890 *T)
    sumX0  +=      24.171 * cos(5.58810453278  + 184.97110483931 *T)
    sumX0  +=      22.048 * cos(5.33604766656  + 163.14033674240 *T)
    sumX0  +=      22.198 * cos(2.48883901199  + 426.73013377490 *T)
    sumX0  +=      25.834 * cos(3.42571919082  + 405.50136735710 *T)
    sumX0  +=      21.674 * cos(1.75719521673  + 2119.00767786191 *T)
    sumX0  +=      23.225 * cos(1.32324288348  + 207.32313779801 *T)
    sumX0  +=      21.571 * cos(3.69323223915  + 7.87862934611 *T)
    sumX0  +=      27.080 * cos(0.13937719095  + 1155.60497489050 *T)
    sumX0  +=      21.569 * cos(3.20671779872  + 224.84924561630 *T)
    sumX0  +=      21.638 * cos(2.59969528542  + 432.25863433090 *T)
    sumX0  +=      22.198 * cos(0.89868151046  + 426.95388294410 *T)
    sumX0  +=      21.066 * cos(6.12707563148  + 508.10650660869 *T)
    sumX0  +=      22.956 * cos(4.15344353916  + 93.18466331550 *T)
    sumX0  +=      22.695 * cos(1.52975337199  + 223.57722122910 *T)
    sumX0  +=      27.166 * cos(0.08679484081  + 36.89238041300 *T)
    sumX0  +=      25.768 * cos(5.84872192947  + 143.90275367970 *T)
    sumX0  +=      22.341 * cos(3.56638187846  + 219.76268804501 *T)
    sumX0  +=      20.372 * cos(4.40717131664  + 849.50804597240 *T)
    sumX0  +=      27.791 * cos(5.00076987801  + 106.78110999820 *T)
    sumX0  +=      20.257 * cos(2.36409011623  + 131.64776735340 *T)
    sumX0  +=      24.029 * cos(2.54607748587  + 334.79493440480 *T)
    sumX0  +=      20.321 * cos(3.71713466857  + 3053.46855786309 *T)
    sumX0  +=      21.760 * cos(1.87467800066  + 264.26100328570 *T)
    sumX0  +=      19.818 * cos(2.87228791193  + 357.68948408471 *T)
    sumX0  +=      21.046 * cos(5.57803586395  + 242.11472100371 *T)
    sumX0  +=      22.078 * cos(3.53915870761  + 186.45557754760 *T)
    sumX0  +=      21.330 * cos(2.40736518098  + 343.46239008310 *T)
    sumX0  +=      23.629 * cos(5.94897428321  + 333.41352756050 *T)
    sumX0  +=      24.992 * cos(4.48787502141  + 31.73638687251 *T)
    sumX0  +=      24.756 * cos(4.22954187784  + 46.45360796860 *T)
    sumX0  +=      25.007 * cos(5.30587581385  + 2428.28600051770 *T)
    sumX0  +=      23.761 * cos(5.47300746239  + 536.56069461190 *T)
    sumX0  +=      21.722 * cos(0.38410941425  + 202.76033721600 *T)
    sumX0  +=      18.920 * cos(5.02923263215  + 171.89841514590 *T)
    sumX0  +=      18.653 * cos(0.27903946815  + 427.05445668060 *T)
    sumX0  +=      18.653 * cos(3.10848105430  + 426.62956003840 *T)
    sumX0  +=      21.149 * cos(0.84965595278  + 320.08020543620 *T)
    sumX0  +=      25.426 * cos(0.10718425313  + 481.01667910731 *T)
    sumX0  +=      22.841 * cos(0.14442813021  + 451.22103074770 *T)
    sumX0  +=      23.911 * cos(2.97613048908  + 264.74863825270 *T)
    sumX0  +=      22.993 * cos(1.52199009068  + 556.76148552110 *T)
    sumX0  +=      19.429 * cos(2.06971667626  + 392.90175841570 *T)
    sumX0  +=      22.024 * cos(1.24219628349  + 112.89781925770 *T)
    sumX0  +=      23.549 * cos(6.12523396053  + 593.67068088151 *T)
    sumX0  +=      19.967 * cos(6.07285120389  + 1567.48843679789 *T)
    sumX0  +=      22.876 * cos(2.43650477347  + 67.91186905000 *T)
    sumX0  +=      19.677 * cos(3.01153643104  + 216.23667169220 *T)
    sumX0  +=      23.750 * cos(6.07536125242  + 462.26673101160 *T)
    sumX0  +=      22.555 * cos(2.04140779152  + 78.46993434690 *T)
    sumX0  +=      20.143 * cos(5.23192874336  + 963.64652045490 *T)
    sumX0  +=      21.017 * cos(3.71041617725  + 402.46298597130 *T)
    sumX0  +=      21.934 * cos(2.56964611043  + 347.64062156210 *T)
    sumX0  +=      17.902 * cos(1.45854327689  + 555.79827767460 *T)
    sumX0  +=      20.903 * cos(6.26285263273  + 66.46102623650 *T)
    sumX0  +=      18.918 * cos(3.18987164241  + 561.42735196710 *T)
    sumX0  +=      17.876 * cos(0.23711641897  + 2214.98690507971 *T)
    sumX0  +=      17.731 * cos(6.25218307104  + 460.78225830330 *T)
    sumX0  +=      17.098 * cos(4.61479892752  + 146.83806920150 *T)
    sumX0  +=      24.044 * cos(4.15983188816  + 141.46962733991 *T)
    sumX0  +=      16.884 * cos(3.21168698004  + 478.04773369070 *T)
    sumX0  +=      18.890 * cos(4.70830613247  + 572.47305223100 *T)
    sumX0  +=      18.099 * cos(6.25077296647  + 230.95140066080 *T)
    sumX0  +=      18.165 * cos(0.82492001036  + 533.86693584120 *T)
    sumX0  +=      20.525 * cos(3.30359301299  + 156.92056161890 *T)
    sumX0  +=      16.659 * cos(3.10651814375  + 419.77664546851 *T)
    sumX0  +=      23.314 * cos(0.03668815157  + 280.72332952100 *T)
    sumX0  +=      18.637 * cos(4.27871136207  + 152.77596003471 *T)
    sumX0  +=      16.704 * cos(1.27514451558  + 422.42222012939 *T)
    sumX0  +=      16.529 * cos(5.88739273505  + 533.37930087419 *T)
    sumX0  +=      16.486 * cos(1.18131009643  + 835.28095197081 *T)
    sumX0  +=      17.269 * cos(6.09621582994  + 200.20959079720 *T)
    sumX0  +=      20.803 * cos(3.23661551748  + 175.67050971460 *T)
    sumX0  +=      18.263 * cos(3.29914267725  + 906.13039727500 *T)
    sumX0  +=      17.267 * cos(5.27709845883  + 295.43805848960 *T)
    sumX0  +=      17.953 * cos(0.75321094297  + 547.20025796550 *T)
    sumX0  +=      22.334 * cos(4.06244344780  + 423.87306294291 *T)
    sumX0  +=      20.308 * cos(1.06278645094  + 294.80741117069 *T)
    sumX0  +=      16.381 * cos(5.86040104652  + 263.94548910060 *T)
    sumX0  +=      16.267 * cos(2.15375878438  + 1486.22393854870 *T)
    sumX0  +=      18.540 * cos(4.98517065544  + 552.82933225800 *T)
    sumX0  +=      16.627 * cos(1.54374663198  + 1.20702533000 *T)
    sumX0  +=      16.245 * cos(5.66731880104  + 99.66756299740 *T)
    sumX0  +=      16.399 * cos(5.61888178859  + 757.46097201770 *T)
    sumX0  +=      18.383 * cos(1.33769122464  + 240.63024829540 *T)
    sumX0  +=      16.013 * cos(4.62895816065  + 2097.17940189250 *T)
    sumX0  +=      15.991 * cos(3.77204289063  + 1055.69359440960 *T)
    sumX0  +=      16.379 * cos(4.22158560955  + 692.34366686999 *T)
    sumX0  +=      21.494 * cos(3.37089777455  + 206.64181424181 *T)
    sumX0  +=      15.840 * cos(1.39791514296  + 831.34879870770 *T)
    sumX0  +=      16.271 * cos(0.22203110018  + 526.00262931501 *T)
    sumX0  +=      16.436 * cos(3.95361231238  + 1994.08662767389 *T)
    sumX0  +=      18.303 * cos(5.55336775323  + 100.14064374939 *T)
    sumX0  +=      16.892 * cos(2.18137146685  + 1588.82907780029 *T)
    sumX0  +=      16.255 * cos(5.06377532297  + 226.87623504580 *T)
    sumX0  +=      15.567 * cos(3.11397846224  + 220.54458533771 *T)
    sumX0  +=      15.921 * cos(4.81397888311  + 2840.16946242510 *T)
    sumX0  +=      16.234 * cos(0.62092736572  + 490.09027169589 *T)
    sumX0  +=      15.273 * cos(0.19945798941  + 78.20680978850 *T)
    sumX0  +=      15.648 * cos(1.75194347964  + 215.54865049161 *T)
    sumX0  +=      15.648 * cos(3.13768268695  + 211.53717535141 *T)
    sumX0  +=      18.382 * cos(1.75927851154  + 220.44401160121 *T)
    sumX0  +=      16.593 * cos(4.99590657727  + 109.48693085641 *T)
    sumX0  +=      18.947 * cos(3.07277134880  + 637.69342324271 *T)
    sumX0  +=      16.216 * cos(5.99615199350  + 41.29761442810 *T)
    sumX0  +=      20.572 * cos(0.76699318042  + 285.61619850311 *T)
    sumX0  +=      16.595 * cos(4.77017702394  + 364.59058276440 *T)
    sumX0  +=      17.326 * cos(2.91676943616  + 200.04953210281 *T)
    sumX0  +=      14.971 * cos(1.76753173425  + 646.76701583129 *T)
    sumX0  +=      16.383 * cos(2.67389070738  + 204.14174406030 *T)
    sumX0  +=      16.103 * cos(6.21534908920  + 419.68027724890 *T)
    sumX0  +=      14.886 * cos(5.37199893369  + 416.05943265400 *T)
    sumX0  +=      18.360 * cos(0.30630166172  + 211.57080524620 *T)
    sumX0  +=      19.474 * cos(0.31238053169  + 2008.80135664250 *T)
    sumX0  +=      20.381 * cos(5.74839865243  + 108.00245814810 *T)
    sumX0  +=      14.597 * cos(2.13836272063  + 110.40195459310 *T)
    sumX0  +=      16.369 * cos(2.18957268546  + 172.70156429801 *T)
    sumX0  +=      14.609 * cos(4.75981242464  + 942.30587945250 *T)
    sumX0  +=      14.822 * cos(1.69517172807  + 196.01680510321 *T)
    sumX0  +=      14.578 * cos(6.12937723629  + 255.18741069710 *T)
    sumX0  +=      17.284 * cos(4.99761394207  + 254.69977573010 *T)
    sumX0  +=      15.790 * cos(0.36558450473  + 410.16723380310 *T)
    sumX0  +=      19.120 * cos(0.46326457877  + 2427.79836555070 *T)
    sumX0  +=      17.766 * cos(2.37155421942  + 50.64639366260 *T)
    sumX0  +=      14.974 * cos(1.81004476813  + 65.46418849521 *T)
    sumX0  +=      19.645 * cos(2.85465946217  + 14.73403604350 *T)
    sumX0  +=      15.462 * cos(2.21494463682  + 222.94408178270 *T)
    sumX0  +=      15.819 * cos(2.67749213694  + 82.13896305161 *T)
    sumX0  +=      17.917 * cos(4.36815293872  + 83.10217089810 *T)
    sumX0  +=      14.016 * cos(0.92450223186  + 427.10264079041 *T)
    sumX0  +=      15.063 * cos(5.74509140920  + 441.81986188650 *T)
    sumX0  +=      17.883 * cos(1.34689508017  + 458.59770230690 *T)
    sumX0  +=      15.768 * cos(2.96438388710  + 349.09146437560 *T)
    sumX0  +=      19.463 * cos(2.98264756234  + 72.31710306511 *T)
    sumX0  +=      14.016 * cos(2.46301834168  + 426.58137592861 *T)
    sumX0  +=      13.854 * cos(0.75800637269  + 13.08950464080 *T)
    sumX0  +=      15.926 * cos(1.97285675652  + 227.03629374021 *T)
    sumX0  +=      15.265 * cos(0.04040736838  + 272.85902299479 *T)
    sumX0  +=      14.549 * cos(5.33546747079  + 110.49832281271 *T)
    sumX0  +=      15.012 * cos(1.52188645554  + 220.76833450690 *T)
    sumX0  +=      14.376 * cos(4.16551027176  + 448.93340888730 *T)
    sumX0  +=      15.060 * cos(0.93153679046  + 980.91199584231 *T)
    sumX0  +=      17.967 * cos(4.60399955084  + 702.39252939260 *T)
    sumX0  +=      13.507 * cos(5.23541303763  + 295.29504613770 *T)
    sumX0  +=      15.721 * cos(4.36361091765  + 570.98857952271 *T)
    sumX0  +=      13.819 * cos(5.98054140713  + 457.86149699650 *T)
    sumX0  +=      16.412 * cos(2.75370681275  + 222.14093263061 *T)
    sumX0  +=      16.702 * cos(1.14048158507  + 100.88891114730 *T)
    sumX0  +=      16.229 * cos(2.40723643991  + 111.41334654940 *T)
    sumX0  +=      13.830 * cos(5.00214849060  + 434.91876320681 *T)
    sumX0  +=      18.336 * cos(6.24789499113  + 170.25388374320 *T)
    sumX0  +=      14.088 * cos(1.63837940688  + 540.98048284201 *T)
    sumX0  +=      13.344 * cos(3.19063924466  + 231.06902073981 *T)
    sumX0  +=      13.303 * cos(5.33441711554  + 472.41865939820 *T)
    sumX0  +=      15.477 * cos(1.91986373629  + 543.26810470240 *T)
    sumX0  +=      12.978 * cos(0.29728403273  + 241.99710092470 *T)
    sumX0  +=      15.002 * cos(2.63220592483  + 3266.76765330109 *T)
    sumX0  +=      17.598 * cos(0.09177348842  + 569.29165849331 *T)
    sumX0  +=      15.115 * cos(1.81478587044  + 420.24972622050 *T)
    sumX0  +=      12.874 * cos(4.06856554799  + 258.63192899320 *T)
    sumX0  +=      16.607 * cos(1.08565905389  + 28.32799059869 *T)
    sumX0  +=      15.813 * cos(1.66003803389  + 251.41531612840 *T)
    sumX0  +=      13.957 * cos(6.17562128071  + 864.48589949940 *T)
    sumX0  +=      15.413 * cos(5.17895957732  + 259.11956396020 *T)
    sumX0  +=      16.676 * cos(5.57821764943  + 705.36147480921 *T)
    sumX0  +=      12.604 * cos(4.01665305123  + 3163.67487908250 *T)
    sumX0  +=      12.578 * cos(1.88266412420  + 238.81570718700 *T)
    sumX0  +=      12.841 * cos(3.51894672028  + 149.31937965110 *T)
    sumX0  +=      12.315 * cos(0.48556165492  + 326.93062787860 *T)
    sumX0  +=      13.217 * cos(4.23672411657  + 188.41313100791 *T)
    sumX0  +=      12.147 * cos(4.43773570062  + 267.22994870229 *T)
    sumX0  +=      12.395 * cos(5.58009272761  + 642.58878435231 *T)
    sumX0  +=      14.104 * cos(3.97820147560  + 230.80838830890 *T)
    sumX0  +=      14.269 * cos(5.84355484645  + 42.78208713641 *T)
    sumX0  +=      11.476 * cos(0.33769530075  + 280.24775664150 *T)
    sumX0  +=      11.770 * cos(1.63857719213  + 1685.29593998510 *T)
    sumX0  +=      11.418 * cos(4.11865342199  + 97.91996573070 *T)
    sumX0  +=      12.752 * cos(0.60015594538  + 274.30986580831 *T)
    sumX0  +=      13.026 * cos(0.83195857087  + 193.89919294850 *T)
    sumX0  +=      11.246 * cos(5.33738032742  + 120.01136625851 *T)
    sumX0  +=      11.976 * cos(4.20420231489  + 380.37158544351 *T)
    sumX0  +=      11.206 * cos(4.12139982038  + 714.92270236481 *T)
    sumX0  +=      14.434 * cos(3.77669932178  + 387.22450001341 *T)
    sumX0  +=      11.970 * cos(3.36773444346  + 203.50860461390 *T)
    sumX0  +=      14.741 * cos(0.14243884205  + 2324.70559133209 *T)
    sumX0  +=      12.049 * cos(2.72046266179  + 254.38426154501 *T)
    sumX0  +=      10.842 * cos(2.54339582028  + 640.18928790730 *T)
    sumX0  +=      13.350 * cos(2.22655107118  + 639.17789595100 *T)
    sumX0  +=      14.697 * cos(2.83603289141  + 432.99234751380 *T)
    sumX0  +=      10.885 * cos(5.03796570992  + 17.28229033479 *T)
    sumX0  +=      11.250 * cos(2.69387714803  + 105.76971804189 *T)
    sumX0  +=      13.740 * cos(4.38241554958  + 815.30716359770 *T)
    sumX0  +=      10.769 * cos(4.58820128164  + 19.36627259471 *T)
    sumX0  +=      12.850 * cos(0.91462129399  + 196.27743753410 *T)
    sumX0  +=      10.505 * cos(1.77592799454  + 206.54124050531 *T)
    sumX0  +=      10.385 * cos(0.04027914669  + 197.08058668621 *T)
    sumX0  +=      10.842 * cos(5.62879282141  + 640.09291968770 *T)
    sumX0  +=      10.290 * cos(1.78563521014  + 202.23658022670 *T)
    sumX0  +=      13.234 * cos(3.20279026751  + 894.59706204410 *T)
    sumX0  +=      10.647 * cos(1.24718121251  + 819.81546347680 *T)
    sumX0  +=      10.182 * cos(3.46948577873  + 3369.86042751970 *T)
    sumX0  +=      10.210 * cos(2.71190577157  + 107.52937739611 *T)
    sumX0  +=      11.258 * cos(1.03965055384  + 189.96703968541 *T)
    sumX0  +=      11.951 * cos(1.39915059510  + 732.18817775220 *T)
    sumX0  +=      13.918 * cos(1.27194064881  + 486.64575339981 *T)
    sumX0  +=      10.041 * cos(3.90629251299  + 606.51636803880 *T)
    sumX0  +=      14.076 * cos(4.92923629549  + 256.83194209980 *T)
    sumX0  +=      12.084 * cos(6.20546846687  + 1471.99684454710 *T)
    sumX0  +=      10.842 * cos(5.79198301588  + 1883.88030645450 *T)
    sumX0  +=      11.375 * cos(3.55965858898  + 106.25735300890 *T)
    sumX0  +=      12.596 * cos(4.62753200111  + 403.26613512341 *T)
    sumX0  +=      11.028 * cos(0.09766487939  + 468.48650613510 *T)
    sumX0  +=       9.860 * cos(3.35807537498  + 33.38091827520 *T)
    sumX0  +=      12.108 * cos(0.95598575520  + 355.50492808829 *T)
    sumX0  +=      10.908 * cos(6.07474449340  + 186.71620997851 *T)
    sumX0  +=       9.594 * cos(3.20545536162  + 434.00373947010 *T)
    sumX0  +=      10.453 * cos(1.88602527977  + 1905.22094745690 *T)
    sumX0  +=       9.546 * cos(6.27898280696  + 189.47940471839 *T)
    sumX0  +=      11.233 * cos(6.08951773665  + 641.10431164400 *T)
    sumX0  +=       9.914 * cos(2.69261802946  + 47.93808067690 *T)
    sumX0  +=       9.448 * cos(3.79528802650  + 407.19828838650 *T)
    sumX0  +=       9.546 * cos(2.32156584631  + 219.38443553510 *T)
    sumX0  +=      10.366 * cos(4.21638187695  + 196.86813836510 *T)
    sumX0  +=       9.370 * cos(4.84934706586  + 230.00523915680 *T)
    sumX0  +=      11.272 * cos(1.42491841284  + 731.70054278519 *T)
    sumX0  +=       9.924 * cos(1.76983918803  + 275.30670354960 *T)
    sumX0  +=       9.472 * cos(3.02102603461  + 194.42045781031 *T)
    sumX0  +=      12.896 * cos(5.20406219171  + 662.28738607949 *T)
    sumX0  +=       9.150 * cos(5.91506282258  + 526.26575387340 *T)
    sumX0  +=      11.879 * cos(5.41287240833  + 477.56009872370 *T)
    sumX0  +=      10.766 * cos(5.03028444466  + 240.36961586451 *T)
    sumX0  +=       9.308 * cos(1.90005607885  + 1368.90407032850 *T)
    sumX0  +=       9.735 * cos(0.01733679462  + 433.90737125051 *T)
    sumX0  +=      10.066 * cos(0.49580184666  + 424.13369537380 *T)
    sumX0  +=      10.907 * cos(2.37428912808  + 351.57277482520 *T)
    sumX0  +=      12.189 * cos(0.10561088119  + 480.52904414029 *T)
    sumX0  +=      11.289 * cos(4.81811083501  + 390.19344543000 *T)
    sumX0  +=      11.032 * cos(0.45219265612  + 395.34943897051 *T)
    sumX0  +=       8.918 * cos(1.33335067145  + 19.88753745650 *T)
    sumX0  +=       9.676 * cos(0.65354176165  + 238.67269483510 *T)
    sumX0  +=      10.312 * cos(2.72426681922  + 17.02165790390 *T)
    sumX0  +=       8.636 * cos(1.24833063898  + 92.29089143820 *T)
    sumX0  +=       8.732 * cos(6.24118097005  + 157.88376946540 *T)
    sumX0  +=       9.566 * cos(2.65682037428  + 1493.33748554950 *T)
    sumX0  +=      10.066 * cos(2.89171872688  + 429.55032134520 *T)
    sumX0  +=       9.374 * cos(3.36841689456  + 206.31749133610 *T)
    sumX0  +=       9.533 * cos(3.86186585570  + 237.11878615761 *T)
    sumX0  +=       9.780 * cos(1.98440642041  + 850.25880549781 *T)
    sumX0  +=       8.433 * cos(3.38649663082  + 427.00206705391 *T)
    sumX0  +=       8.433 * cos(0.00102394272  + 426.68194966511 *T)
    sumX0  +=      10.240 * cos(0.08948242268  + 2744.19023520730 *T)
    sumX0  +=       8.356 * cos(5.89204078630  + 824.49837626530 *T)
    sumX0  +=       9.888 * cos(2.77223505114  + 9.31741007210 *T)
    sumX0  +=       8.509 * cos(4.55107024764  + 3156.56133208170 *T)
    sumX0  +=       8.903 * cos(2.43701337444  + 649.70233135310 *T)
    sumX0  +=       8.389 * cos(1.87156206130  + 232.66536803271 *T)
    sumX0  +=       8.197 * cos(4.48510267600  + 90.00326957780 *T)
    sumX0  +=       8.246 * cos(2.28450832726  + 194.63290613140 *T)
    sumX0  +=       8.317 * cos(5.27910556579  + 661.48174479990 *T)
    sumX0  +=       8.295 * cos(6.23450481902  + 2531.37877473630 *T)
    sumX0  +=      11.264 * cos(4.90896177094  + 269.67762925710 *T)
    sumX0  +=      11.219 * cos(0.89543432392  + 453.92934373340 *T)
    sumX0  +=       8.970 * cos(0.94637919422  + 187.67941782501 *T)
    sumX0  +=       8.956 * cos(4.32326904941  + 749.85979004990 *T)
    sumX0  +=      10.013 * cos(5.08881881994  + 2744.67787017430 *T)
    sumX0  +=      10.091 * cos(5.94884155369  + 495.99453263430 *T)
    sumX0  +=       8.906 * cos(2.59926800810  + 518.40144734719 *T)
    sumX0  +=       7.978 * cos(3.25297918872  + 394.59867944510 *T)
    sumX0  +=       7.792 * cos(4.39533597620  + 2413.57127154910 *T)
    sumX0  +=       8.025 * cos(5.87214492419  + 830.86116374070 *T)
    sumX0  +=      10.403 * cos(1.95210095051  + 354.76872277791 *T)
    sumX0  +=       8.025 * cos(2.81881910919  + 857.37235249860 *T)
    sumX0  +=      10.165 * cos(3.22806325066  + 192.44835013500 *T)
    sumX0  +=       9.333 * cos(1.37795990025  + 471.93102443120 *T)
    sumX0  +=       9.111 * cos(5.92051859374  + 167.96626188280 *T)
    sumX0  +=       9.887 * cos(4.36672786548  + 29.44876501211 *T)
    sumX0  +=       7.646 * cos(5.75335152073  + 487.60896124631 *T)
    sumX0  +=       9.194 * cos(1.79966840336  + 594.89452115890 *T)
    sumX0  +=       7.695 * cos(2.58075308553  + 207.70139030790 *T)
    sumX0  +=       7.915 * cos(2.15566286238  + 131.16013238639 *T)
    sumX0  +=       7.999 * cos(5.76892400113  + 256.08118257441 *T)
    sumX0  +=       9.422 * cos(0.15761540408  + 172.48911597691 *T)
    sumX0  +=       8.126 * cos(3.55479241230  + 402.77850015639 *T)
    sumX0  +=      10.105 * cos(1.94180807216  + 204.94489321240 *T)
    sumX0  +=       7.633 * cos(4.41396657448  + 511.28790034640 *T)
    sumX0  +=       9.479 * cos(2.23711171898  + 350.08830211689 *T)
    sumX0  +=       8.422 * cos(4.22542750723  + 503.10807963510 *T)
    sumX0  +=       7.684 * cos(3.51633534377  + 305.10235190919 *T)
    sumX0  +=       9.659 * cos(0.66283481907  + 380.63221787440 *T)
    sumX0  +=       7.375 * cos(1.29638062215  + 367.03826331921 *T)
    sumX0  +=       9.835 * cos(0.05817782312  + 2325.19322629910 *T)
    sumX0  +=       7.261 * cos(1.35829329188  + 69.60879007940 *T)
    sumX0  +=       7.327 * cos(2.60511788764  + 232.45291971160 *T)
    sumX0  +=       8.716 * cos(6.17750630025  + 405.01373239010 *T)
    sumX0  +=       9.076 * cos(3.97326051682  + 237.43430034270 *T)
    sumX0  +=       9.427 * cos(6.07604734850  + 418.24398865040 *T)
    sumX0  +=       7.385 * cos(0.57374541990  + 141.20899490900 *T)
    sumX0  +=       7.235 * cos(1.35532322396  + 2001.68780964171 *T)
    sumX0  +=       7.849 * cos(0.39800874682  + 176.89434999200 *T)
    sumX0  +=       7.786 * cos(4.61643541892  + 100.15519796440 *T)
    sumX0  +=       7.111 * cos(2.35979437658  + 72.57773549600 *T)
    sumX0  +=       9.164 * cos(2.48785028898  + 562.39055981360 *T)
    sumX0  +=       7.298 * cos(1.74992346676  + 425.56998397231 *T)
    sumX0  +=       7.773 * cos(1.62452877443  + 144.17794032560 *T)
    sumX0  +=       7.342 * cos(4.29624613644  + 326.19691469570 *T)
    sumX0  +=       7.093 * cos(2.36284513741  + 291.50590522650 *T)
    sumX0  +=       6.873 * cos(3.87250231635  + 2310.47849733050 *T)
    sumX0  +=       8.660 * cos(0.53084594475  + 343.98365494490 *T)
    sumX0  +=       9.119 * cos(6.14182394487  + 113.15845168861 *T)
    sumX0  +=       7.614 * cos(3.93985644686  + 239.40640801801 *T)
    sumX0  +=       7.081 * cos(3.27322993659  + 448.98829064149 *T)
    sumX0  +=       7.895 * cos(5.03449455718  + 1169.34443392509 *T)
    sumX0  +=       7.696 * cos(2.59606158308  + 746.67839631219 *T)
    sumX0  +=       7.903 * cos(0.76470523272  + 84.68970947040 *T)
    sumX0  +=       7.977 * cos(6.08347168017  + 171.00464326860 *T)
    sumX0  +=       7.298 * cos(1.63759710678  + 428.11403274671 *T)
    sumX0  +=       8.742 * cos(4.49162081829  + 739.56484931140 *T)
    sumX0  +=       7.590 * cos(2.40747801188  + 1802.61580820530 *T)
    sumX0  +=       8.186 * cos(1.53974993955  + 28.69800548670 *T)
    sumX0  +=       7.872 * cos(2.84703772391  + 135.57992061650 *T)
    sumX0  +=       6.935 * cos(3.25946441452  + 1287.15193711231 *T)
    sumX0  +=       6.519 * cos(4.28288912869  + 314.18800658530 *T)
    sumX0  +=       6.628 * cos(4.94895033186  + 269.20205637760 *T)
    sumX0  +=       8.878 * cos(3.02260106254  + 190.64836324160 *T)
    sumX0  +=       8.467 * cos(4.31417662024  + 486.15811843280 *T)
    sumX0  +=       6.757 * cos(1.37932220229  + 419.20719649691 *T)
    sumX0  +=       6.377 * cos(1.22293688706  + 526.75338884040 *T)
    sumX0  +=       7.376 * cos(3.69180714802  + 25.35904518210 *T)
    sumX0  +=       6.333 * cos(1.48800604479  + 117.66675156770 *T)
    sumX0  +=       6.682 * cos(3.91151906705  + 110.97140356470 *T)
    sumX0  +=       6.651 * cos(5.17190850506  + 385.78821141490 *T)
    sumX0  +=       7.569 * cos(5.15022034917  + 196.13442518220 *T)
    sumX0  +=       7.360 * cos(2.13385617362  + 128.60938596760 *T)
    sumX0  +=       6.901 * cos(4.51033589337  + 250.19147585100 *T)
    sumX0  +=       7.381 * cos(2.56818287176  + 376.43943218040 *T)
    sumX0  +=       7.504 * cos(0.59008062582  + 397.63706083091 *T)
    sumX0  +=       7.473 * cos(4.00525270573  + 24.88596443009 *T)
    sumX0  +=       6.275 * cos(1.06369744472  + 329.16586011230 *T)
    sumX0  +=       7.570 * cos(3.02874986191  + 990.47322339791 *T)
    sumX0  +=       6.057 * cos(4.42905792653  + 856.62159297321 *T)
    sumX0  +=       6.364 * cos(1.67678982628  + 2221.61281711350 *T)
    sumX0  +=       5.928 * cos(0.04731871122  + 505.55576018990 *T)
    sumX0  +=       7.836 * cos(5.45653203333  + 129.20008679860 *T)
    sumX0  +=       6.147 * cos(0.67579092414  + 639.61983893570 *T)
    sumX0  +=       6.324 * cos(5.73872165534  + 756.56720014041 *T)
    sumX0  +=       5.898 * cos(2.55127874457  + 73.05330837550 *T)
    sumX0  +=       6.943 * cos(2.32469404402  + 141.73025977080 *T)
    sumX0  +=       5.913 * cos(6.06018875597  + 121.23520653590 *T)
    sumX0  +=       5.823 * cos(1.61871389354  + 843.87897167990 *T)
    sumX0  +=       6.807 * cos(5.39402662205  + 384.30373870660 *T)
    sumX0  +=       6.016 * cos(3.74116326543  + 3480.06674873909 *T)
    sumX0  +=       6.641 * cos(5.02972300030  + 95.63234387030 *T)
    sumX0  +=       6.458 * cos(2.40544724008  + 3583.15952295770 *T)
    sumX0  +=       7.069 * cos(5.79770704301  + 910.06255053810 *T)
    sumX0  +=       7.779 * cos(3.17906379215  + 724.58699578440 *T)
    sumX0  +=       7.221 * cos(0.99307096935  + 312.96665843540 *T)
    sumX0  +=       7.610 * cos(3.02066477143  + 25.51661174900 *T)
    sumX0  +=       7.748 * cos(0.35371355288  + 55.41532597260 *T)
    sumX0  +=       7.664 * cos(1.86348422475  + 236.43746260140 *T)
    sumX0  +=       5.819 * cos(4.79812800922  + 210.88278404561 *T)
    sumX0  +=       5.819 * cos(0.09149815737  + 216.20304179740 *T)
    sumX0  +=       6.261 * cos(3.40399529731  + 109.92887384111 *T)
    sumX0  +=       6.136 * cos(5.77727217314  + 297.98880490840 *T)
    sumX0  +=       5.471 * cos(1.39114507199  + 672.38443271190 *T)
    sumX0  +=       5.971 * cos(5.52081967030  + 2200.27217611110 *T)
    sumX0  +=       6.662 * cos(0.77274894862  + 611.19928082730 *T)
    sumX0  +=       6.604 * cos(0.41941287758  + 393.21727260080 *T)
    sumX0  +=       5.774 * cos(3.77827896207  + 97.17169833279 *T)
    sumX0  +=       5.678 * cos(5.61713898188  + 17.50929287090 *T)
    sumX0  +=       5.771 * cos(0.05335845955  + 164.78486814511 *T)
    sumX0  +=       6.577 * cos(6.26135869387  + 2641.09746098870 *T)
    sumX0  +=       5.341 * cos(1.76885678471  + 104.82106441041 *T)
    sumX0  +=       6.085 * cos(0.58940873359  + 433.43429049850 *T)
    sumX0  +=       6.219 * cos(5.50146110400  + 100.40127618030 *T)
    sumX0  +=       5.285 * cos(4.44903100722  + 37.88921815429 *T)
    sumX0  +=       5.304 * cos(5.79491653729  + 397.14942586389 *T)
    sumX0  +=       6.082 * cos(1.19286005781  + 640.66236865931 *T)
    sumX0  +=       6.287 * cos(0.66549025090  + 636.47207509279 *T)
    sumX0  +=       5.677 * cos(3.09420329388  + 208.43510349080 *T)
    sumX0  +=       5.677 * cos(1.79542287271  + 218.65072235220 *T)
    sumX0  +=       5.284 * cos(5.43816540149  + 760.49935340350 *T)
    sumX0  +=       7.032 * cos(6.12709991872  + 244.07476659150 *T)
    sumX0  +=       5.102 * cos(2.35225677161  + 319.55644844690 *T)
    sumX0  +=       6.572 * cos(2.03290295505  + 1130.47519297691 *T)
    sumX0  +=       5.363 * cos(6.25943152476  + 700.90805668431 *T)
    sumX0  +=       5.071 * cos(0.09529125238  + 806.96977631951 *T)
    sumX0  +=       6.462 * cos(5.68471426394  + 370.21965705690 *T)
    sumX0  +=       5.365 * cos(2.31649860836  + 170.51700830159 *T)
    sumX0  +=       6.508 * cos(0.80289944068  + 566.84397793850 *T)
    sumX0  +=       5.532 * cos(3.51315539521  + 115.86676467430 *T)
    sumX0  +=       5.557 * cos(6.06370454687  + 1788.38871420371 *T)
    sumX0  +=       4.900 * cos(2.05095023209  + 5.38525680900 *T)
    sumX0  +=       4.866 * cos(3.12125648181  + 533.11617631580 *T)
    sumX0  +=       4.838 * cos(3.19533526201  + 2.72512793310 *T)
    sumX0  +=       4.827 * cos(3.21632443448  + 3686.25229717630 *T)
    sumX0  +=       5.563 * cos(0.76539820211  + 508.59414157570 *T)
    sumX0  +=       5.497 * cos(5.67484207544  + 551.27542358051 *T)
    sumX0  +=       6.373 * cos(5.73650670449  + 310.47079377080 *T)
    sumX0  +=       4.825 * cos(3.34774899951  + 184.33796539290 *T)
    sumX0  +=       4.857 * cos(1.54189437048  + 124.25903370670 *T)
    sumX0  +=       5.731 * cos(0.62968033303  + 525.25436191710 *T)
    sumX0  +=       5.593 * cos(1.30288660608  + 320.82847283410 *T)
    sumX0  +=       5.962 * cos(1.91859895835  + 411.37651986550 *T)
    sumX0  +=       4.890 * cos(0.99447912964  + 383.81610373959 *T)
    sumX0  +=       4.978 * cos(4.78098070731  + 178.33063859051 *T)
    sumX0  +=       6.082 * cos(3.68743535734  + 54.89406111079 *T)
    sumX0  +=       5.118 * cos(5.00411789229  + 89.04006173130 *T)
    sumX0  +=       5.637 * cos(0.71124772957  + 500.13913421850 *T)
    sumX0  +=       5.571 * cos(6.04641962947  + 3060.58210486390 *T)
    sumX0  +=       6.493 * cos(4.38605840365  + 26088.14695905770 *T)
    sumX0  +=       5.591 * cos(2.54795843844  + 285.35556607221 *T)
    sumX0  +=       4.800 * cos(0.42879092939  + 1383.13116433011 *T)
    sumX0  +=       5.463 * cos(4.67524250178  + 57.86300652739 *T)
    sumX0  +=       5.035 * cos(0.56276653096  + 658.30035106221 *T)
    sumX0  +=       4.758 * cos(1.51996664773  + 632.06434895021 *T)
    sumX0  +=       5.081 * cos(5.42435496020  + 482.97672469510 *T)
    sumX0  +=       4.612 * cos(5.47321184962  + 634.51202950500 *T)
    sumX0  +=       6.057 * cos(5.77967962177  + 1045.39865367110 *T)
    sumX0  +=       4.578 * cos(4.49110882286  + 271.64973693240 *T)
    sumX0  +=       5.515 * cos(5.87000879666  + 455.41381644171 *T)
    sumX0  +=       5.305 * cos(3.30135414196  + 484.94883237040 *T)
    sumX0  +=       5.440 * cos(0.49478776063  + 770.06058095910 *T)
    sumX0  +=       4.994 * cos(5.98784026769  + 42.88515300039 *T)
    sumX0  +=       4.487 * cos(5.15665153614  + 732.45130231059 *T)
    sumX0  +=       5.536 * cos(4.79114703105  + 304.84171947829 *T)
    sumX0  +=       4.486 * cos(5.55382992269  + 440.17533048380 *T)
    sumX0  +=       4.513 * cos(5.63672980083  + 745.03386490949 *T)
    sumX0  +=       5.772 * cos(2.53139391930  + 32.48714639791 *T)
    sumX0  +=       4.455 * cos(5.63042033123  + 849.77117053080 *T)
    sumX0  +=       5.554 * cos(4.94368042854  + 493.54685207950 *T)
    sumX0  +=       5.487 * cos(1.39236562914  + 319.06881347989 *T)
    sumX0  +=       4.770 * cos(3.19401797350  + 991.95769610620 *T)
    sumX0  +=       4.782 * cos(3.88757447538  + 504.80500066450 *T)
    sumX0  +=       5.299 * cos(4.18648503875  + 200.80029162821 *T)
    sumX0  +=       4.812 * cos(4.61941921678  + 327.94201983490 *T)
    sumX0  +=       4.838 * cos(1.68316643353  + 774.72644740510 *T)
    sumX0  +=       4.398 * cos(1.57789085262  + 637.22034249070 *T)
    sumX0  +=       4.622 * cos(1.37569567707  + 43.04521169480 *T)
    sumX0  +=       4.573 * cos(2.85126820260  + 26.58288545949 *T)
    sumX0  +=       5.909 * cos(0.42178511059  + 313.43973918739 *T)
    sumX0  +=       4.454 * cos(1.49116821094  + 198.72511808891 *T)
    sumX0  +=       4.339 * cos(3.72176257939  + 153.25153291419 *T)
    sumX0  +=       5.432 * cos(3.43242481594  + 199.04063227400 *T)
    sumX0  +=       4.814 * cos(3.18729036913  + 179.91817716280 *T)
    sumX0  +=       5.416 * cos(5.57721164364  + 952.34018776010 *T)
    sumX0  +=       4.417 * cos(2.84130388631  + 113.37090000971 *T)
    sumX0  +=       5.809 * cos(4.81933144237  + 254.59670986610 *T)
    sumX0  +=       4.499 * cos(2.93071896451  + 958.33296034749 *T)
    sumX0  +=       4.316 * cos(5.65077716051  + 840.91002626331 *T)
    sumX0  +=       4.591 * cos(0.70345966091  + 771.54505366740 *T)
    sumX0  +=       4.277 * cos(2.60125938503  + 208.38941150850 *T)
    sumX0  +=       4.839 * cos(4.84673734698  + 635.47523735150 *T)
    sumX0  +=       4.685 * cos(2.16436514227  + 111.67397898031 *T)
    sumX0  +=       5.613 * cos(6.11246793909  + 597.11519917759 *T)
    sumX0  +=       4.574 * cos(3.73747518933  + 201.76349947470 *T)
    sumX0  +=       4.511 * cos(5.04861447966  + 1044.64789414570 *T)
    sumX0  +=       5.229 * cos(6.16508948002  + 108.96566599461 *T)
    sumX0  +=       4.479 * cos(2.28374284200  + 183.85033042589 *T)
    sumX0  +=       4.309 * cos(4.28568350410  + 3472.95320173830 *T)
    sumX0  +=       4.502 * cos(4.84117827365  + 262.30095769790 *T)
    sumX0  +=       4.161 * cos(0.42309281218  + 337.08255626521 *T)
    sumX0  +=       4.124 * cos(2.49237479764  + 440.58146739410 *T)
    sumX0  +=       4.511 * cos(3.18782799716  + 207.42620366200 *T)
    sumX0  +=       4.109 * cos(5.62546102660  + 1140.89024592190 *T)
    sumX0  +=       5.256 * cos(0.36144458092  + 832.09955823310 *T)
    sumX0  +=       4.845 * cos(3.84776415750  + 226.28553421481 *T)
    sumX0  +=       5.044 * cos(4.92819058549  + 1033.11455891480 *T)
    sumX0  +=       4.049 * cos(4.11923173602  + 372.66733761171 *T)
    sumX0  +=       4.077 * cos(5.03795321512  + 650.43604453600 *T)
    sumX0  +=       4.372 * cos(4.29307292418  + 225.32232636830 *T)
    sumX0  +=       4.235 * cos(4.38201838371  + 348.87901605450 *T)
    sumX0  +=       4.119 * cos(1.09022528624  + 2318.07967929830 *T)
    sumX0  +=       5.657 * cos(4.32042192539  + 198.07742442750 *T)
    sumX0  +=       4.022 * cos(5.41283074173  + 623.46632924110 *T)
    sumX0  +=       4.173 * cos(3.03119388396  + 215.83053478190 *T)
    sumX0  +=       4.173 * cos(1.85843228263  + 211.25529106110 *T)
    sumX0  +=       4.084 * cos(5.94461285605  + 2847.77064439291 *T)
    sumX0  +=       3.996 * cos(2.34556628241  + 561.69047652550 *T)
    sumX0  +=       4.060 * cos(3.39845797997  + 228.36070775411 *T)
    sumX0  +=       4.344 * cos(3.41993367044  + 278.27564896620 *T)
    sumX0  +=       4.010 * cos(1.20237501348  + 283.90472325870 *T)
    sumX0  +=       3.948 * cos(4.14830731658  + 2729.96314120569 *T)
    sumX0  +=       5.205 * cos(4.68217317963  + 72.05647063421 *T)
    sumX0  +=       4.634 * cos(4.18140490995  + 527.22646959241 *T)
    sumX0  +=       5.029 * cos(0.09918698761  + 197.50127781150 *T)
    sumX0  +=       4.736 * cos(5.30260514224  + 425.14508733010 *T)
    sumX0  +=       3.969 * cos(3.81127477948  + 21.09682351890 *T)
    sumX0  +=       5.443 * cos(2.20428875290  + 481.97988695381 *T)
    sumX0  +=       5.129 * cos(5.72715613582  + 219.22017132380 *T)
    sumX0  +=       5.129 * cos(5.44565533795  + 207.86565451920 *T)
    sumX0  +=       3.855 * cos(5.22645171015  + 425.61816808210 *T)
    sumX0  +=       4.014 * cos(5.43177787536  + 707.80915536400 *T)
    sumX0  +=       3.860 * cos(4.85679865938  + 512.66930719071 *T)
    sumX0  +=       3.855 * cos(4.44425417057  + 428.06584863690 *T)
    sumX0  +=       3.832 * cos(1.37075793389  + 668.45227944880 *T)
    sumX0  +=       3.833 * cos(2.96960847626  + 969.86629557841 *T)
    sumX0  +=       3.992 * cos(0.10864564758  + 248.75518725251 *T)
    sumX0  +=       4.504 * cos(4.89705379851  + 3061.06973983091 *T)
    sumX0  +=       3.821 * cos(1.05865131896  + 1258.69774910910 *T)
    sumX0  +=       3.915 * cos(1.23637523349  + 969.37866061139 *T)
    sumX0  +=       4.000 * cos(6.03564121200  + 49.47743513939 *T)
    sumX0  +=       4.401 * cos(1.19543136025  + 774.98957196349 *T)
    sumX0  +=       3.903 * cos(2.21363155069  + 379.14774516610 *T)
    sumX0  +=       5.273 * cos(0.72561113063  + 88.35873817511 *T)
    sumX0  +=       4.974 * cos(1.42953521228  + 604.71638114540 *T)
    sumX0  +=       3.770 * cos(0.09007052734  + 212.17031479781 *T)
    sumX0  +=       3.770 * cos(4.79955563925  + 214.91551104521 *T)
    sumX0  +=       4.435 * cos(0.50168099105  + 204.97852310719 *T)
    sumX0  +=       3.735 * cos(4.33435972105  + 1596.43025976811 *T)
    sumX0  +=       3.690 * cos(0.84619898753  + 738.81408978600 *T)
    sumX0  +=       3.728 * cos(0.34905244092  + 407.71955324831 *T)
    sumX0  +=       4.348 * cos(0.29093412404  + 69.12115511239 *T)
    sumX0  +=       4.549 * cos(1.42149182994  + 491.80174694030 *T)
    sumX0  +=       3.710 * cos(0.78421951437  + 1575.08961876571 *T)
    sumX0  +=       3.788 * cos(4.13635893946  + 258.78949556011 *T)
    sumX0  +=       4.244 * cos(5.65635028403  + 319.08336769490 *T)
    sumX0  +=       3.638 * cos(6.22572039721  + 634.72447782610 *T)
    sumX0  +=       4.782 * cos(5.01635723372  + 313.92737415440 *T)
    sumX0  +=       4.614 * cos(4.98407371709  + 318.12015984841 *T)
    sumX0  +=       3.611 * cos(2.69456119616  + 316.68387124990 *T)
    sumX0  +=       3.647 * cos(0.26596132798  + 48.51422729289 *T)
    sumX0  +=       3.812 * cos(5.31105661530  + 6276.20612047410 *T)
    sumX0  +=       4.045 * cos(0.06905219404  + 519.15220687259 *T)
    sumX0  +=       3.837 * cos(0.06938395849  + 28.55499313480 *T)
    sumX0  +=       3.969 * cos(5.57013725657  + 398.53083270820 *T)
    sumX0  +=       4.988 * cos(1.39831832915  + 105.29663728990 *T)
    sumX0  +=       4.187 * cos(2.87641336936  + 1905.70858242391 *T)
    sumX0  +=       3.673 * cos(1.25158403063  + 619.53417597800 *T)
    sumX0  +=       3.529 * cos(2.53741677974  + 215.50295850929 *T)
    sumX0  +=       4.522 * cos(4.78850933616  + 229.58454803150 *T)
    sumX0  +=       3.575 * cos(2.06510339841  + 291.76902978489 *T)
    sumX0  +=       4.154 * cos(5.87953687212  + 721.40560204669 *T)
    sumX0  +=       4.496 * cos(0.57295696271  + 120.74507944141 *T)
    sumX0  +=       3.649 * cos(3.53826337740  + 444.36811617781 *T)
    sumX0  +=       4.846 * cos(4.94677202632  + 25.61967761299 *T)
    sumX0  +=       4.339 * cos(5.05440665939  + 609.12161516050 *T)
    sumX0  +=       3.687 * cos(1.40777379919  + 2538.00468677009 *T)
    sumX0  +=       3.925 * cos(6.25927800152  + 317.59889498661 *T)
    sumX0  +=       4.658 * cos(6.06131115428  + 2641.58509595570 *T)
    sumX0  +=       4.614 * cos(0.83945558776  + 616.07759559439 *T)
    sumX0  +=       3.906 * cos(2.05005363555  + 525.51499434800 *T)
    sumX0  +=       3.627 * cos(0.45858366119  + 10206.41581669370 *T)
    sumX0  +=       3.452 * cos(4.52964007791  + 62.49524307860 *T)
    sumX0  +=       3.889 * cos(2.25920260691  + 103.81216458159 *T)
    sumX0  +=       4.257 * cos(1.23297991754  + 72.09010052900 *T)
    sumX0  +=       4.693 * cos(1.77733414246  + 619.04654101100 *T)
    sumX0  +=       3.405 * cos(1.51170975386  + 393.70490756781 *T)
    sumX0  +=       3.406 * cos(5.10091194872  + 102.89714084490 *T)
    sumX0  +=       3.405 * cos(2.20832557634  + 172.00148100990 *T)
    sumX0  +=       4.000 * cos(1.12916710018  + 1048.09241244180 *T)
    sumX0  +=       4.448 * cos(4.58159314523  + 1485.73630358170 *T)
    sumX0  +=       4.227 * cos(2.90784233814  + 413.82420042030 *T)
    sumX0  +=       3.578 * cos(5.78180514444  + 316.58750303031 *T)
    sumX0  +=       3.352 * cos(3.55976780702  + 298.47643987540 *T)
    sumX0  +=       3.398 * cos(1.90326495607  + 102.80077262530 *T)
    sumX0  +=       3.878 * cos(1.25593176769  + 210.60759739969 *T)
    sumX0  +=       3.282 * cos(3.54688775335  + 215.07977525651 *T)
    sumX0  +=       3.282 * cos(1.34273841324  + 212.00605058650 *T)
    sumX0  +=       3.262 * cos(1.72244772023  + 327.99690158910 *T)
    sumX0  +=       3.353 * cos(2.96079468399  + 766.12842769600 *T)
    sumX0  +=       3.738 * cos(2.23671986356  + 123.29582586020 *T)
    sumX0  +=       3.649 * cos(3.16101908686  + 610.44852130189 *T)
    sumX0  +=       4.126 * cos(2.78507591828  + 187.51935913060 *T)
    sumX0  +=       3.391 * cos(5.28438045270  + 541.29599702710 *T)
    sumX0  +=       3.608 * cos(2.57278849479  + 1119.42949271300 *T)
    sumX0  +=       3.758 * cos(4.81379904529  + 101.88574888859 *T)
    sumX0  +=       3.231 * cos(3.11446745439  + 188.27011865600 *T)
    sumX0  +=       3.787 * cos(5.85679753021  + 652.67127676971 *T)
    sumX0  +=       3.965 * cos(1.83861016216  + 914.20715212231 *T)
    sumX0  +=       3.735 * cos(1.24714047610  + 80.44204202221 *T)
    sumX0  +=       3.177 * cos(1.02908638657  + 691.34682912870 *T)
    sumX0  +=       3.187 * cos(5.12647471004  + 305.85061930710 *T)
    sumX0  +=       4.058 * cos(0.59959878781  + 92.06388890210 *T)
    sumX0  +=       3.123 * cos(5.76483242550  + 307.07445958451 *T)
    sumX0  +=       3.691 * cos(0.48718423619  + 194.53233239490 *T)
    sumX0  +=       3.467 * cos(4.32351780506  + 428.53892938890 *T)
    sumX0  +=       3.270 * cos(4.06613062205  + 282.90788551740 *T)
    sumX0  +=       4.132 * cos(1.72276410779  + 434.47682022211 *T)
    sumX0  +=       3.256 * cos(0.48979574513  + 344.45922782439 *T)
    sumX0  +=       4.289 * cos(2.13145367830  + 187.94005025590 *T)
    sumX0  +=       3.819 * cos(4.52428970078  + 114.69780615110 *T)
    sumX0  +=       3.049 * cos(5.66152224756  + 842.65762352999 *T)
    sumX0  +=       3.360 * cos(5.25154254711  + 2516.66404576769 *T)
    sumX0  +=       3.446 * cos(1.36703927103  + 305.85311143459 *T)
    sumX0  +=       3.964 * cos(4.57298587668  + 101.62511645769 *T)
    sumX0  +=       3.527 * cos(5.16157036991  + 496.20698095540 *T)
    sumX0  +=       3.065 * cos(3.38375104988  + 348.39138108750 *T)
    sumX0  +=       3.500 * cos(4.38434583704  + 789.70430093210 *T)
    sumX0  +=       2.995 * cos(2.81665065444  + 91.54262404029 *T)
    sumX0  +=       3.701 * cos(5.73555133647  + 405.74744557300 *T)
    sumX0  +=       3.539 * cos(0.99888122440  + 233.18663289450 *T)
    sumX0  +=       2.935 * cos(2.61545309154  + 92.55401599659 *T)
    sumX0  +=       3.414 * cos(6.15560691139  + 331.45348197271 *T)
    sumX0  +=       3.976 * cos(2.17943640249  + 31.26330612050 *T)
    sumX0  +=       3.100 * cos(2.25042510426  + 130.19692453990 *T)
    sumX0  +=       3.608 * cos(2.47118725916  + 315.67247929360 *T)
    sumX0  +=       2.909 * cos(1.70469953307  + 109.69937917751 *T)
    sumX0  +=       2.890 * cos(5.51370123281  + 111.20089822830 *T)
    sumX0  +=       2.917 * cos(3.43410376988  + 357.47703576360 *T)
    sumX0  +=       3.278 * cos(1.66981893495  + 126.23114138200 *T)
    sumX0  +=       2.952 * cos(2.02096060057  + 93.28772917949 *T)
    sumX0  +=       2.859 * cos(3.43547714123  + 181.56270856549 *T)
    sumX0  +=       3.008 * cos(4.10197187210  + 434.70631488571 *T)
    sumX0  +=       2.925 * cos(3.52591083414  + 3796.45861839570 *T)
    sumX0  +=       3.751 * cos(6.13610992998  + 624.67561530350 *T)
    sumX0  +=       3.658 * cos(2.79714159952  + 268.68079151581 *T)
    sumX0  +=       3.572 * cos(1.37112649810  + 851.95572652720 *T)
    sumX0  +=       2.923 * cos(5.61293728889  + 252.41215386970 *T)
    sumX0  +=       3.162 * cos(5.36858469376  + 77.71917482150 *T)
    sumX0  +=       3.903 * cos(4.38444688669  + 1066.25165970650 *T)
    sumX0  +=       3.528 * cos(2.16332343815  + 834.79331700380 *T)
    sumX0  +=       3.311 * cos(0.11299799349  + 212.09207010800 *T)
    sumX0  +=       3.007 * cos(5.67168406954  + 699.94484883781 *T)
    sumX0  +=       2.914 * cos(2.44010514833  + 625.16325027050 *T)
    sumX0  +=       2.826 * cos(2.75759654641  + 71.84402231310 *T)
    sumX0  +=       2.804 * cos(4.40192840756  + 425.40571976101 *T)
    sumX0  +=       2.779 * cos(4.75841138753  + 629.84616305900 *T)
    sumX0  +=       3.204 * cos(0.11083820522  + 31.24875190549 *T)
    sumX0  +=       3.277 * cos(4.40244200779  + 232.55349344810 *T)
    sumX0  +=       3.564 * cos(0.55121382371  + 38.11622069041 *T)
    sumX0  +=       3.345 * cos(4.69701413463  + 77.47309660560 *T)
    sumX0  +=       3.005 * cos(1.85424910273  + 135.09228564950 *T)
    sumX0  +=       3.542 * cos(2.10432417506  + 239.56646671240 *T)
    sumX0  +=       2.893 * cos(2.30567809560  + 3899.55139261430 *T)
    sumX0  +=       3.544 * cos(0.37526043126  + 282.20780222929 *T)
    sumX0  +=       2.804 * cos(5.26877747316  + 428.27829695801 *T)
    sumX0  +=       2.850 * cos(5.41873084940  + 177.63055530240 *T)
    sumX0  +=       2.737 * cos(1.21884738750  + 626.12645811700 *T)
    sumX0  +=       2.722 * cos(3.73879814102  + 214.01848580100 *T)
    sumX0  +=       2.716 * cos(3.09434225481  + 620.49738382450 *T)
    sumX0  +=       2.818 * cos(2.46590151702  + 354.50809034700 *T)
    sumX0  +=       3.749 * cos(5.48742211953  + 145.66241303390 *T)
    sumX0  +=       3.118 * cos(5.71514726491  + 3376.97397452050 *T)
    sumX0  +=       2.902 * cos(5.33945085702  + 52.93401552300 *T)
    sumX0  +=       3.000 * cos(3.29647513212  + 206.90493880020 *T)
    sumX0  +=       2.965 * cos(1.58377397053  + 42.56963881530 *T)
    sumX0  +=       2.885 * cos(4.81351175785  + 418.97770183331 *T)
    sumX0  +=       2.693 * cos(5.92886276832  + 433.20479583490 *T)
    sumX0  +=       3.088 * cos(6.09210931343  + 2957.48933064529 *T)
    sumX0  +=       2.922 * cos(2.24591208630  + 1063.07026596880 *T)
    sumX0  +=       3.407 * cos(6.14849506249  + 499.17592637200 *T)
    sumX0  +=       3.270 * cos(2.70578228663  + 924.28964453971 *T)
    sumX0  +=       3.462 * cos(0.16426971475  + 182.78405671541 *T)
    sumX0  +=       3.056 * cos(5.41338556991  + 179.81511129880 *T)
    sumX0  +=       2.751 * cos(3.55568779756  + 322.26476143261 *T)
    sumX0  +=       3.461 * cos(2.20556239264  + 60.31068708220 *T)
    sumX0  +=       3.254 * cos(2.14594096655  + 87.87110320809 *T)
    sumX0  +=       3.320 * cos(2.19525628405  + 122.98031167510 *T)
    sumX0  +=       3.628 * cos(2.76762924657  + 239.14577558710 *T)
    sumX0  +=       2.722 * cos(1.90912932797  + 644.33388949151 *T)
    sumX0  +=       2.866 * cos(2.73211415857  + 315.41184686271 *T)
    sumX0  +=       3.289 * cos(3.51215591822  + 685.71775483620 *T)
    sumX0  +=       2.598 * cos(0.57085232520  + 151.50393564750 *T)
    sumX0  +=       2.931 * cos(4.99737516135  + 7.83293736379 *T)
    sumX0  +=       2.624 * cos(1.60917289432  + 311.21906116870 *T)
    sumX0  +=       2.667 * cos(3.91979165512  + 1080.96638867510 *T)
    sumX0  +=       2.603 * cos(4.71971708944  + 508.73715392760 *T)
    sumX0  +=       2.559 * cos(5.36447720972  + 639.39034427210 *T)
    sumX0  +=       2.562 * cos(4.26277399169  + 827.67977000299 *T)
    sumX0  +=       2.581 * cos(2.63019535055  + 159.20818347930 *T)
    sumX0  +=       2.754 * cos(2.83456084033  + 287.83687652179 *T)
    sumX0  +=       2.531 * cos(4.86434568068  + 1048.84317196720 *T)
    sumX0  +=       2.912 * cos(3.83544556347  + 1382.64352936309 *T)
    sumX0  +=       2.686 * cos(0.15925446852  + 643.06186510430 *T)
    sumX0  +=       2.508 * cos(1.97837844361  + 297.88573904440 *T)
    sumX0  +=       2.488 * cos(3.20204119275  + 95.47228517591 *T)
    sumX0  +=       3.158 * cos(0.65282282373  + 671.89679774490 *T)
    sumX0  +=       2.864 * cos(1.21917033622  + 453.66871130251 *T)
    sumX0  +=       2.518 * cos(3.93587358039  + 2626.87036698710 *T)
    sumX0  +=       2.793 * cos(2.94738171561  + 208.86249226050 *T)
    sumX0  +=       2.660 * cos(3.15325492704  + 443.51678291590 *T)
    sumX0  +=       2.576 * cos(0.14454041746  + 952.86394474940 *T)
    sumX0  +=       2.986 * cos(1.55320502134  + 105.78427225690 *T)
    sumX0  +=       3.018 * cos(5.15417052486  + 141.94270809190 *T)
    sumX0  +=       2.912 * cos(4.22692425727  + 167.33312243640 *T)
    sumX0  +=       2.430 * cos(1.26440767301  + 235.68919520349 *T)
    sumX0  +=       3.277 * cos(4.17742655482  + 1055.95671896799 *T)
    sumX0  +=       2.469 * cos(0.43788807135  + 213.10346206430 *T)
    sumX0  +=       2.430 * cos(5.16272782845  + 112.41018429070 *T)
    sumX0  +=       2.646 * cos(3.36163487366  + 312.21589891000 *T)
    sumX0  +=       2.465 * cos(2.87544505422  + 205.89354684390 *T)
    sumX0  +=       2.655 * cos(5.30662457895  + 440.85665404000 *T)
    sumX0  +=       2.370 * cos(2.29846468477  + 64.71592109730 *T)
    sumX0  +=       2.389 * cos(4.38325797699  + 85.58348134769 *T)
    sumX0  +=       2.369 * cos(0.90337052069  + 220.86890824340 *T)
    sumX0  +=       2.369 * cos(0.84466299231  + 206.21691759960 *T)
    sumX0  +=       2.493 * cos(2.27793769590  + 388.22133775470 *T)
    sumX0  +=       2.974 * cos(2.14395876897  + 531.73476947149 *T)
    sumX0  +=       2.491 * cos(3.11641893226  + 123.78346082721 *T)
    sumX0  +=       2.347 * cos(3.94291944920  + 3046.35501086230 *T)
    sumX0  +=       3.237 * cos(4.15223587316  + 338.72708766791 *T)
    sumX0  +=       2.520 * cos(6.17195337179  + 338.04576411170 *T)
    sumX0  +=       2.377 * cos(1.72451258399  + 628.60776856660 *T)
    sumX0  +=       2.701 * cos(5.31018762949  + 300.85468446100 *T)
    sumX0  +=       2.653 * cos(0.61839059390  + 56.86616878610 *T)
    sumX0  +=       2.732 * cos(4.84630107056  + 244.30176912761 *T)
    sumX0  +=       2.275 * cos(3.01865971091  + 4002.64416683289 *T)
    sumX0  +=       2.952 * cos(2.31378455663  + 645.77017809000 *T)
    sumX0  +=       2.509 * cos(4.52000559703  + 125.42799222991 *T)
    sumX0  +=       2.253 * cos(2.22689163450  + 310.41591201661 *T)
    sumX0  +=       2.632 * cos(5.22086889296  + 416.79314583689 *T)
    sumX0  +=       2.307 * cos(5.65143265211  + 3164.16251404950 *T)
    sumX0  +=       2.532 * cos(5.13544500097  + 90.31878376290 *T)
    sumX0  +=       2.434 * cos(2.93786049007  + 769.09737311260 *T)
    sumX0  +=       2.819 * cos(4.33639017787  + 34.92027273770 *T)
    sumX0  +=       2.827 * cos(5.81289553813  + 782.59075393131 *T)
    sumX0  +=       2.409 * cos(6.23011689079  + 621.49422156580 *T)
    sumX0  +=       2.533 * cos(5.04860084233  + 631.54308408841 *T)
    sumX0  +=       2.542 * cos(2.08891387679  + 176.63371756111 *T)
    sumX0  +=       2.171 * cos(0.62533559840  + 1891.48148842230 *T)
    sumX0  +=       2.552 * cos(1.71107443815  + 12566.39551746630 *T)
    sumX0  +=       2.811 * cos(1.38980153585  + 641.62557650581 *T)
    sumX0  +=       2.910 * cos(4.77695576046  + 330.96584700570 *T)
    sumX0  +=       2.345 * cos(5.85946991886  + 2104.78058386030 *T)
    sumX0  +=       2.188 * cos(5.37105971182  + 1457.28211557850 *T)
    sumX0  +=       2.591 * cos(0.63794997691  + 170.41394243761 *T)
    sumX0  +=       2.615 * cos(0.54729482995  + 143.69030535860 *T)
    sumX0  +=       2.434 * cos(0.55328244480  + 183.48663213100 *T)
    sumX0  +=       3.003 * cos(0.06003858313  + 739.30172475301 *T)
    sumX0  +=       2.177 * cos(2.60567333935  + 67.63668240410 *T)
    sumX0  +=       2.774 * cos(1.24195655197  + 1048.58004740881 *T)
    sumX0  +=       2.462 * cos(5.75946440696  + 247.27071454420 *T)
    sumX0  +=       2.386 * cos(4.37257867425  + 430.98660994371 *T)
    sumX0  +=       2.988 * cos(4.23764191275  + 1802.12817323829 *T)
    sumX0  +=       2.810 * cos(5.72676129521  + 140.72135994200 *T)
    sumX0  +=       2.255 * cos(0.26165543800  + 6681.46867088311 *T)
    sumX0  +=       2.483 * cos(5.73996136268  + 375.95179721340 *T)
    sumX0  +=       2.123 * cos(0.34275870459  + 610.93615626891 *T)
    sumX0  +=       2.154 * cos(1.01602323535  + 444.10748374690 *T)
    sumX0  +=       2.497 * cos(0.09006336245  + 17.91542978119 *T)
    sumX0  +=       2.073 * cos(5.08154035810  + 218.47190392590 *T)
    sumX0  +=       2.063 * cos(4.77672198594  + 616.56523056141 *T)
    sumX0  +=       2.250 * cos(5.37793469323  + 366.28750379380 *T)
    sumX0  +=       2.750 * cos(1.30459159563  + 485.63685357099 *T)
    sumX0  +=       2.252 * cos(2.78563182902  + 49.26498681829 *T)
    sumX0  +=       2.813 * cos(0.73865142296  + 837.45344587970 *T)
    sumX0  +=       2.062 * cos(0.67146633927  + 2634.47154895491 *T)
    sumX0  +=       2.822 * cos(3.09009791052  + 162.65270177540 *T)
    sumX0  +=       2.611 * cos(0.55025625817  + 927.59115048390 *T)
    sumX0  +=       2.212 * cos(5.64460551330  + 193.41155798150 *T)
    sumX0  +=       2.182 * cos(5.06007050383  + 335.38563523581 *T)
    sumX0  +=       2.403 * cos(2.26991708124  + 452.44487102510 *T)
    sumX0  +=       2.564 * cos(3.29999345582  + 427.73578023681 *T)
    sumX0  +=       2.111 * cos(5.24419959900  + 422.69740677530 *T)
    sumX0  +=       2.308 * cos(3.09970342539  + 155.22364058950 *T)
    sumX0  +=       2.054 * cos(1.10667438189  + 883.55136178020 *T)
    sumX0  +=       2.713 * cos(0.18767262914  + 6.34846465550 *T)
    sumX0  +=       2.692 * cos(2.83455718298  + 1040.97886544099 *T)
    sumX0  +=       2.007 * cos(3.46492872207  + 529.18402305270 *T)
    sumX0  +=       2.282 * cos(4.82626097470  + 173.92540457540 *T)
    sumX0  +=       2.783 * cos(3.97708739470  + 165.36101476110 *T)
    sumX0  +=       2.346 * cos(5.96960035551  + 447.44893617900 *T)
    sumX0  +=       2.376 * cos(1.10753711310  + 945.48727319021 *T)
    sumX0  +=       2.081 * cos(2.10189023227  + 90.57941619379 *T)
    sumX0  +=       2.489 * cos(5.97615262196  + 2487.17222746430 *T)
    sumX0  +=       2.124 * cos(0.24199244027  + 133.34468838281 *T)
    sumX0  +=       2.000 * cos(1.71637966221  + 419.19264228190 *T)
    sumX0  +=       1.998 * cos(1.38456498657  + 518.62844988331 *T)
    sumX0  +=       1.969 * cos(0.87882959914  + 530.68554210351 *T)
    sumX0  +=       2.467 * cos(1.16140871474  + 429.81095377611 *T)
    sumX0  +=       1.955 * cos(2.64714107906  + 105.03351273151 *T)
    sumX0  +=       2.054 * cos(4.07410590523  + 3789.34507139489 *T)
    sumX0  +=       2.425 * cos(4.28534276753  + 180.40581212981 *T)
    sumX0  +=       1.995 * cos(5.64222397445  + 409.57653297210 *T)
    sumX0  +=       2.035 * cos(4.31478927336  + 437.40007365639 *T)
    sumX0  +=       1.990 * cos(1.20144466178  + 2854.39655642670 *T)
    sumX0  +=       2.449 * cos(3.51585269206  + 213.00709384469 *T)
    sumX0  +=       2.354 * cos(5.70604510094  + 657.40657918490 *T)
    sumX0  +=       2.466 * cos(1.21196311089  + 421.68850694650 *T)
    sumX0  +=       1.932 * cos(3.95236750027  + 1062.80714141041 *T)
    sumX0  +=       2.448 * cos(6.06475548762  + 205.98991506349 *T)
    sumX0  +=       1.912 * cos(4.86893514364  + 419.28901050149 *T)
    sumX0  +=       2.647 * cos(6.01524298699  + 747.16603127920 *T)
    sumX0  +=       1.873 * cos(5.39710784329  + 345.08987514331 *T)
    sumX0  +=       2.199 * cos(4.03618137033  + 681.78560157310 *T)
    sumX0  +=       2.386 * cos(4.15979534815  + 25234.46294233869 *T)
    sumX0  +=       1.886 * cos(4.74856771776  + 1096.85045721820 *T)
    sumX0  +=       2.200 * cos(3.67398938648  + 784.28767496071 *T)
    sumX0  +=       1.859 * cos(1.27634409007  + 535.66692273460 *T)
    sumX0  +=       2.123 * cos(4.59517360138  + 418.27761854520 *T)
    sumX0  +=       2.326 * cos(5.90565481193  + 166.58485503849 *T)
    sumX0  +=       2.007 * cos(6.24913643216  + 527.01402127131 *T)
    sumX0  +=       2.162 * cos(2.02542462393  + 420.20403423820 *T)
    sumX0  +=       2.185 * cos(3.86722734363  + 110.18950627200 *T)
    sumX0  +=       2.564 * cos(0.08752711773  + 425.94823648220 *T)
    sumX0  +=       1.976 * cos(3.27975163184  + 102.32769187330 *T)
    sumX0  +=       1.951 * cos(3.80946898193  + 168.29633028290 *T)
    sumX0  +=       1.965 * cos(2.72529355918  + 39.91620758380 *T)
    sumX0  +=       1.953 * cos(4.34770725760  + 116.07921299541 *T)
    sumX0  +=       2.118 * cos(0.16994068938  + 409.43352062020 *T)
    sumX0  +=       2.542 * cos(0.28051618477  + 355.02935520881 *T)
    sumX0  +=       2.294 * cos(5.31857052688  + 181.08713568601 *T)
    sumX0  +=       2.378 * cos(0.97161359133  + 385.30057644790 *T)
    sumX0  +=       1.807 * cos(4.30830792029  + 275.58189019550 *T)
    sumX0  +=       1.833 * cos(2.25943126416  + 267.87764236370 *T)
    sumX0  +=       2.302 * cos(5.74355401552  + 43.53284666180 *T)
    sumX0  +=       2.098 * cos(0.08769001353  + 253.16042126761 *T)
    sumX0  +=       2.063 * cos(5.99370738678  + 177.36743074401 *T)
    sumX0  +=       2.261 * cos(1.40921864289  + 635.94831810351 *T)
    sumX0  +=       2.208 * cos(0.06326079599  + 1194.21109128031 *T)
    sumX0  +=       2.053 * cos(3.72658018441  + 103.37022159689 *T)
    sumX0  +=       1.966 * cos(6.03016819317  + 835.54407652920 *T)
    sumX0  +=       2.346 * cos(0.23030073846  + 222.61650551010 *T)
    sumX0  +=       1.855 * cos(3.60752296582  + 23.13836716340 *T)
    sumX0  +=       1.855 * cos(3.60752296582  + 22.65073219640 *T)
    sumX0  +=       2.017 * cos(4.33634384614  + 243.59919371200 *T)
    sumX0  +=       2.346 * cos(5.76125045506  + 2957.97696561230 *T)
    sumX0  +=       1.781 * cos(1.54883325886  + 220.91709235320 *T)
    sumX0  +=       1.781 * cos(0.19920025414  + 206.16873348980 *T)
    sumX0  +=       1.851 * cos(2.90153472874  + 1107.89615748210 *T)
    sumX0  +=       1.906 * cos(0.30653739166  + 586.06949891369 *T)
    sumX0  +=       2.083 * cos(6.04860095971  + 1037.79747170330 *T)
    sumX0  +=       2.271 * cos(6.07739006315  + 234.03796615640 *T)
    sumX0  +=       2.271 * cos(5.09542141062  + 193.04785968660 *T)
    sumX0  +=       2.373 * cos(5.71475634706  + 281.42341280910 *T)
    sumX0  +=       1.873 * cos(1.44953658425  + 1084.00477006090 *T)
    sumX0  +=       2.113 * cos(4.95085993537  + 671.16059243450 *T)
    sumX0  +=       1.854 * cos(5.40160586575  + 614.38067456499 *T)
    sumX0  +=       1.766 * cos(0.87455416846  + 2.67694382330 *T)
    sumX0  +=       1.774 * cos(3.12295108906  + 220.39582749141 *T)
    sumX0  +=       1.774 * cos(4.90826773112  + 206.68999835160 *T)
    sumX0  +=       2.018 * cos(2.80222834013  + 250.45210828191 *T)
    sumX0  +=       1.877 * cos(0.86746878520  + 35.18090516859 *T)
    sumX0  +=       2.293 * cos(6.18682503815  + 912.72267941400 *T)
    sumX0  +=       1.775 * cos(5.64171842529  + 228.73321476960 *T)
    sumX0  +=       1.775 * cos(2.38950039489  + 198.35261107340 *T)
    sumX0  +=       1.908 * cos(4.75541799866  + 3377.46160948750 *T)
    sumX0  +=       1.924 * cos(0.05325859053  + 534.12756827210 *T)
    sumX0  +=       1.740 * cos(3.73739180897  + 3362.74688051889 *T)
    sumX0  +=       2.154 * cos(0.40104075464  + 666.96780674051 *T)
    sumX0  +=       2.288 * cos(0.41986539828  + 103.59722413300 *T)
    sumX0  +=       1.725 * cos(5.50474519843  + 132.85705341579 *T)
    sumX0  +=       1.785 * cos(4.98048380603  + 346.43133549971 *T)
    sumX0  +=       2.259 * cos(2.60980626460  + 700.42042171730 *T)
    sumX0  +=       1.692 * cos(5.01579761567  + 212.53401309270 *T)
    sumX0  +=       1.721 * cos(5.21316754257  + 306.34074640161 *T)
    sumX0  +=       2.004 * cos(4.26613787249  + 256.67188340540 *T)
    sumX0  +=       2.083 * cos(2.24513842982  + 5216.33655531789 *T)
    sumX0  +=       2.234 * cos(5.54983389111  + 799.36859435169 *T)
    sumX0  +=       1.980 * cos(2.56001945144  + 451.97179027311 *T)
    sumX0  +=       2.071 * cos(3.58320191234  + 549.89401673620 *T)
    sumX0  +=       2.071 * cos(2.05739621883  + 45.49040012211 *T)
    sumX0  +=       1.998 * cos(0.36777686598  + 1171.63205578549 *T)
    sumX0  +=       2.007 * cos(0.88372371587  + 1364.48428209840 *T)
    sumX0  +=       1.836 * cos(2.92694247168  + 644.80697024350 *T)
    sumX0  +=       1.687 * cos(5.33007261538  + 520.37355502251 *T)
    sumX0  +=       1.705 * cos(2.31286181244  + 2222.10045208050 *T)
    sumX0  +=       2.080 * cos(1.45709625903  + 587.45090575800 *T)
    sumX0  +=       2.061 * cos(5.09975785401  + 337.55812914470 *T)
    sumX0  +=       1.674 * cos(2.74515959016  + 207.63865198310 *T)
    sumX0  +=       1.723 * cos(1.32784221906  + 6290.43321447570 *T)
    sumX0  +=       1.760 * cos(2.25500156442  + 12.28635548869 *T)
    sumX0  +=       1.697 * cos(4.20915600472  + 552.34169729099 *T)
    sumX0  +=       1.676 * cos(2.04112487415  + 41.08516610701 *T)
    sumX0  +=       1.715 * cos(2.13305723286  + 133.86844537209 *T)
    sumX0  +=       1.707 * cos(2.05213240355  + 98.60128928690 *T)
    sumX0  +=       1.846 * cos(0.24646975759  + 5849.12029463110 *T)
    sumX0  +=       1.707 * cos(1.11053782306  + 895.08469701110 *T)
    sumX0  +=       1.821 * cos(1.68122578438  + 9779.32999085069 *T)
    sumX0  +=       2.189 * cos(0.22703640405  + 66.94866120350 *T)
    sumX0  +=       2.014 * cos(5.22602330033  + 546.71262299849 *T)
    sumX0  +=       1.781 * cos(1.68206567806  + 9772.21644384989 *T)
    sumX0  +=       1.685 * cos(2.77209660580  + 10220.64291069530 *T)
    sumX0  +=       1.942 * cos(0.62509516046  + 246.68001371321 *T)
    sumX0  +=       2.001 * cos(2.79560236728  + 3.94896821049 *T)
    sumX0  +=       1.626 * cos(5.14783884729  + 624.15435044170 *T)
    sumX0  +=       1.942 * cos(3.67591631329  + 9146.54625153749 *T)
    sumX0  +=       1.802 * cos(2.72673271293  + 1389.35093945360 *T)
    sumX0  +=       1.613 * cos(3.84799104836  + 1144.07163965960 *T)
    sumX0  +=       2.160 * cos(6.02089758599  + 336.59492129820 *T)
    sumX0  +=       2.090 * cos(0.77231653699  + 629.35852809199 *T)
    sumX0  +=       1.968 * cos(3.13421198037  + 118.31444522910 *T)
    sumX0  +=       1.822 * cos(5.19742739470  + 1009.22317149361 *T)
    sumX0  +=       1.611 * cos(0.79920795326  + 347.16754081009 *T)
    sumX0  +=       1.728 * cos(1.00799863461  + 1091.38144162010 *T)
    sumX0  +=       2.055 * cos(3.93160223679  + 1372.34858862460 *T)
    sumX0  +=       1.569 * cos(4.53671161147  + 122.71967924421 *T)
    sumX0  +=       1.542 * cos(5.32956278183  + 1166.16304018740 *T)
    sumX0  +=       1.878 * cos(5.27282154972  + 133.82026126229 *T)
    sumX0  +=       1.628 * cos(3.80471189414  + 46.22660543249 *T)
    sumX0  +=       1.987 * cos(5.87487529487  + 913.50706883420 *T)
    sumX0  +=       1.542 * cos(0.20693930936  + 551.34485954970 *T)
    sumX0  +=       1.643 * cos(4.14805608546  + 196.38050339810 *T)
    sumX0  +=       1.715 * cos(5.05123310391  + 709.29362807231 *T)
    sumX0  +=       1.803 * cos(3.56272450751  + 58.07545484850 *T)
    sumX0  +=       1.620 * cos(0.62227051964  + 700.20797339620 *T)
    sumX0  +=       1.584 * cos(5.28418181170  + 3767.45443505930 *T)
    sumX0  +=       2.027 * cos(3.91807332340  + 2118.52004289490 *T)
    sumX0  +=       1.660 * cos(4.88325679293  + 318.43567403349 *T)
    sumX0  +=       1.695 * cos(3.99389886352  + 327.19375243699 *T)
    sumX0  +=       1.617 * cos(2.39206154532  + 83.62343575990 *T)
    sumX0  +=       1.522 * cos(5.52616009652  + 406.39513923441 *T)
    sumX0  +=       1.847 * cos(2.06358167839  + 488.60579898760 *T)
    sumX0  +=       1.505 * cos(4.00358265141  + 447.28887748461 *T)
    sumX0  +=       1.747 * cos(0.03102392120  + 173.69840203930 *T)
    sumX0  +=       1.511 * cos(2.76524724862  + 414.09938706621 *T)
    sumX0  +=       1.932 * cos(1.42317332409  + 40.59753113999 *T)
    sumX0  +=       1.512 * cos(0.91625215222  + 1748.28107876310 *T)
    sumX0  +=       1.775 * cos(2.24504797682  + 2707.58486990309 *T)
    sumX0  +=       1.830 * cos(0.63475720612  + 259.75270340660 *T)
    sumX0  +=       1.574 * cos(3.93642970138  + 322.36782729660 *T)
    sumX0  +=       1.477 * cos(2.05303400646  + 355.24180352990 *T)
    sumX0  +=       1.498 * cos(1.98604024579  + 174.93679653170 *T)
    sumX0  +=       1.487 * cos(4.04993820708  + 141.45507312490 *T)
    sumX0  +=       1.676 * cos(5.43783308850  + 3693.36584417709 *T)
    sumX0  +=       1.461 * cos(1.38524899881  + 105.13657859549 *T)
    sumX0  +=       1.692 * cos(3.46209472083  + 679.49797971270 *T)
    sumX0  +=       1.835 * cos(5.85414378465  + 245.99869015701 *T)
    sumX0  +=       1.449 * cos(5.42328914768  + 1159.04949318660 *T)
    sumX0  +=       1.661 * cos(1.72245289890  + 736.59590389479 *T)
    sumX0  +=       1.443 * cos(3.36134760735  + 466.51439845980 *T)
    sumX0  +=       1.476 * cos(5.38971328268  + 3480.55438370610 *T)
    sumX0  +=       1.645 * cos(5.17910426695  + 249.71839509901 *T)
    sumX0  +=       1.496 * cos(3.22754175149  + 77623.56832092509 *T)
    sumX0  +=       1.890 * cos(3.64866692511  + 764.64395498770 *T)
    sumX0  +=       1.480 * cos(3.14944647582  + 262.77653057739 *T)
    sumX0  +=       1.937 * cos(3.47338953300  + 757.56403788170 *T)
    sumX0  +=       1.587 * cos(0.32048665561  + 6.91791362710 *T)
    sumX0  +=       1.683 * cos(4.01550009219  + 6503.73230991370 *T)
    sumX0  +=       1.770 * cos(5.08423811097  + 482.50115181560 *T)
    sumX0  +=       1.735 * cos(2.28799872431  + 2435.39954751850 *T)
    sumX0  +=       1.457 * cos(3.01927816891  + 256.18424843839 *T)
    sumX0  +=       1.471 * cos(0.70642591223  + 1453.08683775701 *T)
    sumX0  +=       1.613 * cos(1.78409994223  + 271.86218525350 *T)
    sumX0  +=       1.848 * cos(2.55365760167  + 419.94090967981 *T)
    sumX0  +=       1.954 * cos(4.83592161734  + 145.38722638800 *T)
    sumX0  +=       1.662 * cos(0.44336441806  + 25454.87558477749 *T)
    sumX0  +=       1.420 * cos(6.11955516870  + 1354.18934135989 *T)
    sumX0  +=       1.390 * cos(0.11880365257  + 544.75257741070 *T)
    sumX0  +=       1.679 * cos(5.97438658773  + 285.14311775110 *T)
    sumX0  +=       1.607 * cos(0.73844026178  + 544.26494244370 *T)
    sumX0  +=       1.536 * cos(5.89479692757  + 3273.88120030190 *T)
    sumX0  +=       1.536 * cos(5.89479692757  + 3274.36883526891 *T)
    sumX0  +=       1.587 * cos(5.49922746524  + 148.58317434070 *T)
    sumX0  +=       1.794 * cos(2.66160697296  + 754.59509246510 *T)
    sumX0  +=       1.407 * cos(0.85219523429  + 316.11442227830 *T)
    sumX0  +=       1.658 * cos(2.48037578229  + 415.06259491270 *T)
    sumX0  +=       1.512 * cos(4.28907569036  + 428.48653976220 *T)
    sumX0  +=       1.501 * cos(6.18750341180  + 5746.02752041250 *T)
    sumX0  +=       1.696 * cos(2.90084876328  + 59.03866269500 *T)
    sumX0  +=       1.590 * cos(0.23710525653  + 349.61272923740 *T)
    sumX0  +=       1.827 * cos(0.61806717369  + 230.32075334190 *T)
    sumX0  +=       1.891 * cos(5.00801645038  + 1165.63928319810 *T)
    sumX0  +=       1.759 * cos(4.58582597435  + 200.27902676641 *T)
    sumX0  +=       1.759 * cos(3.44539284583  + 226.80679907661 *T)
    sumX0  +=       1.677 * cos(4.23577119544  + 50.15875869559 *T)
    sumX0  +=       1.384 * cos(4.27737882618  + 1176.94561589290 *T)
    sumX0  +=       1.600 * cos(5.43870126795  + 10433.94200613330 *T)
    sumX0  +=       1.491 * cos(1.31312182045  + 9676.23721663209 *T)
    sumX0  +=       1.400 * cos(0.26007468679  + 1950.53447393720 *T)
    sumX0  +=       1.334 * cos(0.72637373445  + 301.65783361311 *T)
    sumX0  +=       1.377 * cos(0.86428517620  + 243.29037717131 *T)
    sumX0  +=       1.377 * cos(4.02534099039  + 183.79544867171 *T)
    sumX0  +=       1.454 * cos(3.28359870251  + 330.37514617470 *T)
    sumX0  +=       1.421 * cos(4.52961540751  + 469.97097884340 *T)
    sumX0  +=       1.426 * cos(3.01357344337  + 934.44157292630 *T)
    sumX0  +=       1.356 * cos(5.87449410633  + 210.13451664770 *T)
    sumX0  +=       1.314 * cos(2.71429002268  + 415.04804069769 *T)
    sumX0  +=       1.338 * cos(3.51731603639  + 118.62995941420 *T)
    sumX0  +=       1.453 * cos(0.57533096444  + 679.81349389780 *T)
    sumX0  +=       1.470 * cos(3.81098724259  + 321.30155358610 *T)
    sumX0  +=       1.314 * cos(2.08783972120  + 878.08234618209 *T)
    sumX0  +=       1.715 * cos(6.08299477871  + 79.43065006591 *T)
    sumX0  +=       1.305 * cos(2.73772701133  + 238.18505986809 *T)
    sumX0  +=       1.382 * cos(0.52411582569  + 718.85485562790 *T)
    sumX0  +=       1.464 * cos(1.92663592308  + 397.90018538930 *T)
    sumX0  +=       1.758 * cos(3.53136981753  + 452.70550345601 *T)
    sumX0  +=       1.356 * cos(4.56256930200  + 280.93577784210 *T)
    sumX0  +=       1.790 * cos(2.61933789345  + 1357.37073509760 *T)
    sumX0  +=       1.578 * cos(5.37892573514  + 383.55297918120 *T)
    sumX0  +=       1.464 * cos(4.41309699814  + 360.13716463950 *T)
    sumX0  +=       1.375 * cos(0.61137430521  + 535.56385687061 *T)
    sumX0  +=       1.620 * cos(0.47405441337  + 109.06873185860 *T)
    sumX0  +=       1.663 * cos(4.95313251291  + 102.73708215049 *T)
    sumX0  +=       1.314 * cos(0.02379514725  + 1670.58121101650 *T)
    sumX0  +=       1.452 * cos(0.95390444904  + 316.89631957100 *T)
    sumX0  +=       1.302 * cos(4.09749727380  + 116.66991382640 *T)
    sumX0  +=       1.421 * cos(2.52300558725  + 166.84548746939 *T)
    sumX0  +=       1.339 * cos(2.72026084734  + 5863.34738863270 *T)
    sumX0  +=       1.685 * cos(6.07730679616  + 214.53975066279 *T)
    sumX0  +=       1.458 * cos(3.04961044313  + 9352.73179997469 *T)
    sumX0  +=       1.341 * cos(0.47332489447  + 1183.16539101641 *T)
    sumX0  +=       1.300 * cos(3.88742356206  + 120.60206708950 *T)
    sumX0  +=       1.281 * cos(4.57888575454  + 1108.38379244910 *T)
    sumX0  +=       1.652 * cos(5.63517331024  + 165.84864972810 *T)
    sumX0  +=       1.259 * cos(0.41488838005  + 2950.86341861150 *T)
    sumX0  +=       1.508 * cos(1.65590855031  + 763.19311217420 *T)
    sumX0  +=       1.251 * cos(0.56339346105  + 1055.20595944260 *T)
    sumX0  +=       1.416 * cos(0.73841384971  + 1063.55790093581 *T)
    sumX0  +=       1.273 * cos(0.60463815018  + 2207.87335807890 *T)
    sumX0  +=       1.755 * cos(0.25702827395  + 557.75832326239 *T)
    sumX0  +=       1.665 * cos(0.62988003125  + 163.82166029861 *T)
    sumX0  +=       1.330 * cos(2.61784978340  + 156.02678974161 *T)
    sumX0  +=       1.472 * cos(3.64179989725  + 346.15614885380 *T)
    sumX0  +=       1.472 * cos(3.64179989725  + 346.64378382081 *T)
    sumX0  +=       1.283 * cos(2.09085692852  + 4215.94326227089 *T)
    sumX0  +=       1.564 * cos(1.45141699498  + 3391.68870348911 *T)
    sumX0  +=       1.232 * cos(2.01286113512  + 767.12526543730 *T)
    sumX0  +=       1.339 * cos(3.55159228628  + 419.76209125350 *T)
    sumX0  +=       1.471 * cos(5.32263840566  + 2693.35777590150 *T)
    sumX0  +=       1.265 * cos(3.33952901594  + 212.30451842910 *T)
    sumX0  +=       1.355 * cos(3.06105542076  + 418.71956152989 *T)
    sumX0  +=       1.236 * cos(1.64676624453  + 409.67959883610 *T)
    sumX0  +=       1.263 * cos(4.89869307825  + 2090.06585489169 *T)
    sumX0  +=       1.580 * cos(4.87533968051  + 1477.83836716070 *T)
    sumX0  +=       1.310 * cos(5.18861487187  + 1012.66768978969 *T)
    sumX0  +=       1.245 * cos(5.08294427560  + 748.96601817260 *T)
    sumX0  +=       1.565 * cos(3.90450910087  + 1268.99268984760 *T)
    sumX0  +=       1.267 * cos(5.32574708258  + 115.11600514891 *T)
    sumX0  +=       1.279 * cos(1.12013287366  + 217.94814693660 *T)
    sumX0  +=       1.279 * cos(0.62790063934  + 209.13767890641 *T)
    sumX0  +=       1.237 * cos(3.25551290814  + 4112.85048805230 *T)
    sumX0  +=       1.208 * cos(1.32712565413  + 116.12739710521 *T)
    sumX0  +=       1.358 * cos(4.00471985016  + 973.79844884150 *T)
    sumX0  +=       1.406 * cos(5.51451351576  + 959.45373476091 *T)
    sumX0  +=       1.365 * cos(0.53588394734  + 988.28866740149 *T)
    sumX0  +=       1.189 * cos(4.92462455483  + 1378.30523918970 *T)
    sumX0  +=       1.200 * cos(0.49702267118  + 614.86830953200 *T)
    sumX0  +=       1.537 * cos(3.54508824305  + 170.72945662269 *T)
    sumX0  +=       1.449 * cos(3.08273741016  + 1020.26887175750 *T)
    sumX0  +=       1.181 * cos(1.33018054312  + 372.77040347569 *T)
    sumX0  +=       1.227 * cos(5.69576717158  + 2421.17245351690 *T)
    sumX0  +=       1.504 * cos(2.57217010618  + 2810.67764412170 *T)
    sumX0  +=       1.191 * cos(1.49579907580  + 487.39651292520 *T)
    sumX0  +=       1.601 * cos(1.72354458106  + 1151.18518666040 *T)
    sumX0  +=       1.207 * cos(0.96898843462  + 3170.78842608329 *T)
    sumX0  +=       1.288 * cos(4.69433760178  + 3149.44778508089 *T)
    sumX0  +=       1.277 * cos(0.27655645720  + 217.24806364849 *T)
    sumX0  +=       1.276 * cos(2.61558896767  + 94.77220188780 *T)
    sumX0  +=       1.451 * cos(1.05411548466  + 514.95692905111 *T)
    sumX0  +=       1.451 * cos(1.05411548466  + 514.46929408409 *T)
    sumX0  +=       1.235 * cos(3.43297201820  + 173.43776960839 *T)
    sumX0  +=       1.353 * cos(3.76027139434  + 530.04665716270 *T)
    sumX0  +=       1.464 * cos(0.02257079027  + 746.15463932291 *T)
    sumX0  +=       1.246 * cos(5.61078961339  + 424.86990068420 *T)
    sumX0  +=       1.236 * cos(2.50843841662  + 1248.38576202810 *T)
    sumX0  +=       1.279 * cos(0.03215301642  + 703.38936713389 *T)
    sumX0  +=       1.236 * cos(4.30746328208  + 538.84831647230 *T)
    sumX0  +=       1.234 * cos(4.16726551204  + 416.32006508490 *T)
    sumX0  +=       1.161 * cos(1.82750814245  + 313.71492583330 *T)
    sumX0  +=       1.414 * cos(1.60384951308  + 2648.69864295650 *T)
    sumX0  +=       1.177 * cos(4.62142425135  + 643.58562209360 *T)
    sumX0  +=       1.187 * cos(1.03828381736  + 1394.98001374610 *T)
    sumX0  +=       1.173 * cos(1.60433079541  + 331.61354066710 *T)
    sumX0  +=       1.176 * cos(2.90358611485  + 252.14902931130 *T)
    sumX0  +=       1.245 * cos(4.65509615846  + 293.74113746021 *T)
    sumX0  +=       1.187 * cos(3.72181414929  + 4105.73694105150 *T)
    sumX0  +=       1.274 * cos(0.56033904615  + 165.10038233019 *T)
    sumX0  +=       1.391 * cos(0.02309515967  + 5860.16599489499 *T)
    sumX0  +=       1.365 * cos(5.22737241965  + 674.08135374130 *T)
    sumX0  +=       1.165 * cos(1.16402753194  + 107.04174242910 *T)
    sumX0  +=       1.169 * cos(1.54612792835  + 104.29979954861 *T)
    sumX0  +=       1.280 * cos(2.60261684467  + 60.79832204921 *T)
    sumX0  +=       1.470 * cos(0.46274502064  + 1147.74066836430 *T)
    sumX0  +=       1.243 * cos(0.16760981697  + 945.75039774860 *T)
    sumX0  +=       1.211 * cos(2.29834110056  + 271.05903610141 *T)
    sumX0  +=       1.378 * cos(4.48082815431  + 880.26690217851 *T)
    sumX0  +=       1.141 * cos(4.65662492035  + 1365.23504162379 *T)
    sumX0  +=       1.452 * cos(3.82619301322  + 670.25226634220 *T)
    sumX0  +=       1.372 * cos(3.12283015772  + 5852.30168836880 *T)
    sumX0  +=       1.152 * cos(5.71979516304  + 253.21530302180 *T)
    sumX0  +=       1.145 * cos(5.20900865265  + 3796.94625336271 *T)
    sumX0  +=       1.147 * cos(5.23246594483  + 2048.15823255770 *T)
    sumX0  +=       1.139 * cos(3.77606446333  + 227.81819103291 *T)
    sumX0  +=       1.139 * cos(4.25515435685  + 199.26763481011 *T)
    sumX0  +=       1.350 * cos(3.52660754473  + 25440.64849077589 *T)
    sumX0  +=       1.152 * cos(5.11322028459  + 475.60005313590 *T)
    sumX0  +=       1.377 * cos(0.50068712612  + 660.85109748100 *T)
    sumX0  +=       1.324 * cos(3.36678513643  + 101.36448402680 *T)
    sumX0  +=       1.189 * cos(2.65397891656  + 1383.93431348220 *T)
    sumX0  +=       1.403 * cos(3.67339271391  + 1699.03539901970 *T)
    sumX0  +=       1.177 * cos(1.82821053485  + 303.30236501580 *T)
    sumX0  +=       1.164 * cos(5.16928967595  + 1773.67398523510 *T)
    sumX0  +=       1.179 * cos(5.67584375514  + 272.33775813299 *T)
    sumX0  +=       1.228 * cos(6.14893570078  + 91.70019060721 *T)
    sumX0  +=       1.431 * cos(3.53241753707  + 1467.31393175860 *T)
    sumX0  +=       1.142 * cos(3.42129382875  + 656.91894421790 *T)
    sumX0  +=       1.157 * cos(3.17352979808  + 469.64340257080 *T)
    sumX0  +=       1.157 * cos(0.03193714449  + 470.13103753780 *T)
    sumX0  +=       1.459 * cos(5.59118854590  + 1128.77827194750 *T)
    sumX0  +=       1.199 * cos(3.32815503030  + 384.51618702770 *T)
    sumX0  +=       1.360 * cos(6.23249225628  + 1560.37488979710 *T)
    sumX0  +=       1.399 * cos(0.26308817006  + 1243.98302014049 *T)
    sumX0  +=       1.154 * cos(3.90218563691  + 5539.84197197530 *T)
    sumX0  +=       1.290 * cos(5.88751484616  + 1403.57803345520 *T)
    sumX0  +=       1.426 * cos(0.71297202771  + 701.69244610450 *T)
    sumX0  +=       1.402 * cos(5.49844037291  + 1112.31594571220 *T)
    sumX0  +=       1.097 * cos(4.39996766325  + 735.68088015810 *T)
    sumX0  +=       1.431 * cos(4.14934397552  + 388.96960515260 *T)
    sumX0  +=       1.187 * cos(1.09475241758  + 6599.71153713150 *T)
    sumX0  +=       1.093 * cos(6.27482624765  + 427.31758123900 *T)
    sumX0  +=       1.537 * cos(2.26997973747  + 2854.88419139370 *T)
    sumX0  +=       1.417 * cos(0.74675814459  + 25557.96835899609 *T)
    sumX0  +=       1.238 * cos(4.45159537738  + 559.94287925881 *T)
    sumX0  +=       1.473 * cos(1.81109515353  + 2751.79141717511 *T)
    sumX0  +=       1.452 * cos(2.08713835729  + 228.03313148150 *T)
    sumX0  +=       1.516 * cos(5.88783767141  + 560.46663624809 *T)
    sumX0  +=       1.103 * cos(4.63060901499  + 531.94052014820 *T)
    sumX0  +=       1.301 * cos(1.44654929805  + 9790.37569111459 *T)
    sumX0  +=       1.099 * cos(5.99645868304  + 527.92904500800 *T)
    sumX0  +=       1.256 * cos(1.98238644435  + 877.81922162370 *T)
    sumX0  +=       1.250 * cos(3.85798606392  + 1226.45442019471 *T)
    sumX0  +=       1.113 * cos(3.14740146804  + 124.18959773749 *T)
    sumX0  +=       1.072 * cos(4.02039116236  + 206.26930722630 *T)
    sumX0  +=       1.072 * cos(4.01082765782  + 220.81651861670 *T)
    sumX0  +=       1.335 * cos(0.17870194529  + 121.00820399980 *T)
    sumX0  +=       1.282 * cos(4.54628381109  + 9782.51138458839 *T)
    sumX0  +=       1.072 * cos(0.62535496972  + 220.49640122790 *T)
    sumX0  +=       1.072 * cos(1.12267854328  + 206.58942461510 *T)
    sumX0  +=       1.303 * cos(3.69503246823  + 1096.48675892331 *T)
    sumX0  +=       1.397 * cos(5.45476565968  + 1461.68485746610 *T)
    sumX0  +=       1.085 * cos(2.72676641341  + 379.88395047650 *T)
    sumX0  +=       1.170 * cos(2.12202833962  + 705.10084237830 *T)
    sumX0  +=       1.054 * cos(3.95603977178  + 155.53915477459 *T)
    sumX0  +=       1.405 * cos(4.62229591705  + 5436.74919775670 *T)
    sumX0  +=       1.097 * cos(4.60268688107  + 476.56326098241 *T)
    sumX0  +=       1.241 * cos(5.11295878390  + 2833.05591542430 *T)
    sumX0  +=       1.288 * cos(0.42374008522  + 1340.80534299830 *T)
    sumX0  +=       1.129 * cos(3.07767826513  + 1222.52226693161 *T)
    sumX0  +=       1.290 * cos(0.79377005415  + 1082.25966492170 *T)
    sumX0  +=       1.166 * cos(5.05925238169  + 2467.64287643290 *T)
    sumX0  +=       1.299 * cos(5.43314879441  + 2509.55049876690 *T)
    sumX0  +=       1.321 * cos(6.10687737458  + 9366.95889397630 *T)
    sumX0  +=       1.051 * cos(4.35040361644  + 959.97749175019 *T)
    sumX0  +=       1.051 * cos(4.35040361644  + 960.46512671720 *T)
    sumX0  +=       1.211 * cos(2.80991523490  + 304.10551416790 *T)
    sumX0  +=       1.052 * cos(3.86792125865  + 645.55772976890 *T)
    sumX0  +=       1.127 * cos(0.31244064224  + 78477.25233764409 *T)
    sumX0  +=       1.258 * cos(4.40841524052  + 5867.27954189579 *T)
    sumX0  +=       1.073 * cos(2.88134377608  + 328.48453655610 *T)
    sumX0  +=       1.290 * cos(4.48821585460  + 71.56883566720 *T)
    sumX0  +=       1.053 * cos(4.63251123701  + 795.43644108859 *T)
    sumX0  +=       1.329 * cos(5.03792511567  + 2740.74571691120 *T)
    sumX0  +=       1.352 * cos(1.20612298463  + 1127.50624756031 *T)
    sumX0  +=       1.032 * cos(5.79901826701  + 774.25336665311 *T)
    sumX0  +=       1.421 * cos(0.04765558923  + 398.38782035631 *T)
    sumX0  +=       1.302 * cos(0.47222793041  + 476.07562601539 *T)
    sumX0  +=       1.018 * cos(1.52577563410  + 135.79236893760 *T)
    sumX0  +=       1.240 * cos(0.29711496611  + 761.98382611181 *T)
    sumX0  +=       1.072 * cos(4.40309497596  + 1257.94698958370 *T)
    sumX0  +=       1.044 * cos(0.40386153641  + 304.36614659881 *T)
    sumX0  +=       1.166 * cos(3.33113338762  + 613.62991503960 *T)
    sumX0  +=       1.025 * cos(0.01753129634  + 121.59890483079 *T)
    sumX0  +=       1.321 * cos(4.43670480817  + 323.48860171000 *T)
    sumX0  +=       1.279 * cos(4.51516906631  + 203.72105293501 *T)
    sumX0  +=       1.279 * cos(3.51604975387  + 223.36477290801 *T)
    sumX0  +=       1.145 * cos(6.17371704189  + 519.88592005549 *T)
    sumX0  +=       1.018 * cos(2.91596221445  + 640.89186332290 *T)
    sumX0  +=       1.389 * cos(5.24370269936  + 1115.76046400830 *T)
    sumX0  +=       1.343 * cos(3.88682161226  + 108.26558270649 *T)
    sumX0  +=       1.195 * cos(0.85808512318  + 99.14380600810 *T)
    sumX0  +=       1.008 * cos(4.30738828647  + 102.58832430420 *T)
    sumX0  +=       1.143 * cos(5.21944528509  + 256.34430713280 *T)
    sumX0  +=       1.019 * cos(4.91718332635  + 6247.26929803929 *T)
    sumX0  +=       1.250 * cos(5.68253927507  + 759.01488069520 *T)
    sumX0  +=       1.000 * cos(2.13999178433  + 1939.48877367330 *T)
    sumX0  +=       1.110 * cos(1.66567300271  + 1397.94895916270 *T)
    sumX0  +=       1.170 * cos(1.26049852836  + 7.39099437910 *T)
    sumX0  +=       1.196 * cos(4.77831019397  + 46.81730626350 *T)
    sumX0  +=       1.151 * cos(2.51852068637  + 76.02225379210 *T)
    sumX0  +=       1.005 * cos(5.76224230474  + 690.85919416170 *T)
    sumX0  +=       1.316 * cos(1.32295596105  + 475.11241816889 *T)
    sumX0  +=       0.995 * cos(4.20379456134  + 99.92819542829 *T)
    sumX0  +=       1.197 * cos(1.26595437757  + 1392.53233319130 *T)
    sumX0  +=       1.110 * cos(2.98882017733  + 523.29680845680 *T)
    sumX0  +=       1.088 * cos(4.97175423832  + 824.23525170691 *T)
    sumX0  +=       1.015 * cos(1.47228772360  + 634.98760238449 *T)
    sumX0  +=       1.012 * cos(1.77286703845  + 689.86235642041 *T)
    sumX0  +=       1.381 * cos(3.30442233560  + 97.43233076370 *T)
    sumX0  +=       1.276 * cos(4.48973341413  + 465.71124930770 *T)
    sumX0  +=       1.233 * cos(5.35220662775  + 1546.29080814740 *T)
    sumX0  +=       1.160 * cos(6.23737046615  + 152.98840835580 *T)
    sumX0  +=       1.247 * cos(3.58622778979  + 2434.91191255150 *T)
    sumX0  +=       1.201 * cos(3.09828888291  + 528.71094230071 *T)
    sumX0  +=       1.045 * cos(1.49017361423  + 608.00084074710 *T)
    sumX0  +=       1.034 * cos(2.46164938759  + 540.49284787499 *T)
    sumX0  +=       0.988 * cos(2.46330736551  + 967.89418790310 *T)
    sumX0  +=       0.973 * cos(3.93273625700  + 173.38288785421 *T)
    sumX0  +=       1.081 * cos(2.34160070624  + 560.93971700010 *T)
    sumX0  +=       1.105 * cos(0.59883609919  + 227.72182281330 *T)
    sumX0  +=       1.105 * cos(1.14919741381  + 199.36400302970 *T)
    sumX0  +=       1.022 * cos(0.99655670520  + 599.40282103799 *T)
    sumX0  +=       1.176 * cos(5.83186908907  + 9797.48923811540 *T)
    sumX0  +=       1.174 * cos(0.61580453745  + 1680.87615175499 *T)
    sumX0  +=       0.980 * cos(3.06958429681  + 944.99963822319 *T)
    sumX0  +=       1.194 * cos(4.57575815213  + 725.07463075140 *T)
    sumX0  +=       1.045 * cos(3.17246710740  + 343.49601997789 *T)
    sumX0  +=       1.227 * cos(3.94543045845  + 1973.37663399041 *T)
    sumX0  +=       1.185 * cos(2.30979572901  + 773.24197469681 *T)
    sumX0  +=       1.222 * cos(3.21080779639  + 776.21092011340 *T)
    sumX0  +=       0.980 * cos(2.09680719136  + 524.78128116509 *T)
    sumX0  +=       1.285 * cos(2.51329534910  + 6.82154540749 *T)
    sumX0  +=       1.225 * cos(2.82333756383  + 25344.66926355809 *T)
    sumX0  +=       0.985 * cos(3.52739874813  + 913.24394427581 *T)
    sumX0  +=       0.995 * cos(0.79227350270  + 557.28275038290 *T)
    sumX0  +=       0.983 * cos(1.11328788816  + 550.38165170320 *T)
    sumX0  +=       1.005 * cos(5.19307115578  + 301.17019864609 *T)
    sumX0  +=       1.208 * cos(4.61381839173  + 2597.37854868370 *T)
    sumX0  +=       0.950 * cos(2.91396525108  + 568.54089896790 *T)
    sumX0  +=       1.220 * cos(4.63270249547  + 1458.71591204951 *T)
    sumX0  +=       1.171 * cos(2.39322957185  + 2538.49232173710 *T)
    sumX0  +=       1.155 * cos(1.80072310652  + 308.77138061390 *T)
    sumX0  +=       0.978 * cos(4.67862862927  + 394.70174530910 *T)
    sumX0  +=       0.967 * cos(0.19601324126  + 1332.99171270940 *T)
    sumX0  +=       1.232 * cos(3.28428140208  + 10.05112325500 *T)
    sumX0  +=       0.962 * cos(5.17770276952  + 416.80770005190 *T)
    sumX0  +=       1.017 * cos(1.74683674400  + 714.29205504590 *T)
    sumX0  +=       1.065 * cos(1.12354067485  + 1961.84329875950 *T)
    sumX0  +=       0.951 * cos(5.15428525662  + 1760.07753855241 *T)
    sumX0  +=       0.990 * cos(0.33566024066  + 6328.08121061550 *T)
    sumX0  +=       1.236 * cos(4.27499899029  + 263.26416554441 *T)
    sumX0  +=       1.104 * cos(2.12495167740  + 394.86180400350 *T)
    sumX0  +=       0.956 * cos(2.78082151468  + 1022.06885865090 *T)
    sumX0  +=       1.275 * cos(3.95835597791  + 104.33342944339 *T)
    sumX0  +=       1.056 * cos(2.41420421180  + 62.28279475750 *T)
    sumX0  +=       1.143 * cos(5.25389269800  + 511.77553531340 *T)
    sumX0  +=       0.941 * cos(4.62379540315  + 555.31064270760 *T)
    sumX0  +=       1.191 * cos(0.13148121604  + 2313.65989106819 *T)
    sumX0  +=       1.170 * cos(0.78792987093  + 331.92905485220 *T)
    sumX0  +=       1.072 * cos(2.19877879267  + 486.90887795820 *T)
    sumX0  +=       1.059 * cos(3.53258689279  + 1471.24608502170 *T)
    sumX0  +=       1.044 * cos(1.75718704182  + 841.90686400460 *T)
    sumX0  +=       0.998 * cos(2.70657811539  + 522.28541650049 *T)
    sumX0  +=       1.221 * cos(6.07798468320  + 473.31243127551 *T)
    sumX0  +=       0.926 * cos(5.46179970939  + 179.11502801070 *T)
    sumX0  +=       1.089 * cos(0.06304982526  + 1986.97308067310 *T)
    sumX0  +=       0.950 * cos(0.83670377886  + 899.01685027420 *T)
    sumX0  +=       1.071 * cos(1.11068546883  + 781.83999440590 *T)
    sumX0  +=       0.998 * cos(5.72763556428  + 1151.93594618579 *T)
    sumX0  +=       1.082 * cos(5.25069693345  + 2151.25100677631 *T)
    sumX0  +=       1.057 * cos(1.31797828563  + 3708.08057314570 *T)
    sumX0  +=       0.916 * cos(2.63738626919  + 386.00065973601 *T)
    sumX0  +=       1.200 * cos(4.51772187194  + 9793.55708485230 *T)
    sumX0  +=       1.139 * cos(5.00077006154  + 5845.18814136799 *T)
    sumX0  +=       1.219 * cos(4.61690528448  + 25771.75508940110 *T)
    sumX0  +=       1.118 * cos(1.34949230498  + 1080.07261679780 *T)
    sumX0  +=       1.043 * cos(1.52618503474  + 6261.49639204090 *T)
    sumX0  +=       1.172 * cos(6.01779016946  + 160.36507991500 *T)
    sumX0  +=       0.906 * cos(4.97943583685  + 1475.44136284320 *T)
    sumX0  +=       0.994 * cos(1.83692702793  + 1467.57705631699 *T)
    sumX0  +=       0.982 * cos(3.40437661999  + 1460.46350931620 *T)
    sumX0  +=       1.193 * cos(0.19326102993  + 1439.75351563271 *T)
    sumX0  +=       1.075 * cos(1.79436431149  + 3171.27606105030 *T)
    sumX0  +=       0.899 * cos(0.95306018744  + 606.20085385370 *T)
    sumX0  +=       1.230 * cos(2.60934825737  + 5334.14405850510 *T)
    sumX0  +=       0.928 * cos(4.83201886555  + 401.29402744810 *T)
    sumX0  +=       1.122 * cos(0.19423881613  + 81.99595069970 *T)
    sumX0  +=       0.901 * cos(4.96802796450  + 428.80205394729 *T)
    sumX0  +=       0.912 * cos(1.41505691677  + 726.07146849269 *T)
    sumX0  +=       1.205 * cos(4.09366056977  + 734.66948820180 *T)
    sumX0  +=       1.047 * cos(0.22623020333  + 1304.68053705810 *T)
    sumX0  +=       1.133 * cos(5.06097777553  + 733.18501549349 *T)
    sumX0  +=       1.109 * cos(6.09594487703  + 1259.43146229200 *T)
    sumX0  +=       0.888 * cos(6.13890404543  + 4.42204896250 *T)
    sumX0  +=       0.924 * cos(2.63079629477  + 633.50312967620 *T)
    sumX0  +=       0.924 * cos(5.77238894836  + 633.99076464320 *T)
    sumX0  +=       0.931 * cos(4.56641553803  + 529.82290799351 *T)
    sumX0  +=       0.956 * cos(2.47364979110  + 1198.14324454341 *T)
    sumX0  +=       1.118 * cos(4.83663685761  + 26.10731258001 *T)
    sumX0  +=       0.962 * cos(1.78028349992  + 770.58184582090 *T)
    sumX0  +=       1.092 * cos(0.85875789744  + 947.28726008359 *T)
    sumX0  +=       0.899 * cos(2.86831464473  + 9896.64985907089 *T)
    sumX0  +=       1.114 * cos(1.71021700338  + 210.08633253789 *T)
    sumX0  +=       0.935 * cos(1.95269050544  + 97.21988244259 *T)
    sumX0  +=       0.909 * cos(2.84341068430  + 981.17512040070 *T)
    sumX0  +=       0.907 * cos(3.53818082973  + 3679.13875017550 *T)
    sumX0  +=       1.078 * cos(2.11362323812  + 429.53576713020 *T)
    sumX0  +=       1.223 * cos(5.59274764012  + 876.85850590470 *T)
    sumX0  +=       1.053 * cos(5.34256956029  + 828.16740497000 *T)
    sumX0  +=       1.159 * cos(1.80620732752  + 979.42752313400 *T)
    sumX0  +=       1.179 * cos(5.56984888168  + 261.23717611490 *T)
    sumX0  +=       0.872 * cos(1.92760134845  + 102.68889804070 *T)
    sumX0  +=       0.902 * cos(6.00836364296  + 1407.51018671831 *T)
    sumX0  +=       0.970 * cos(0.92585247141  + 51322.36608391289 *T)
    sumX0  +=       0.954 * cos(4.49316887147  + 1527.86843632430 *T)
    sumX0  +=       0.870 * cos(5.19978000768  + 103.00901542950 *T)
    sumX0  +=       1.100 * cos(5.62879482377  + 1229.89893849079 *T)
    sumX0  +=       0.980 * cos(1.25124342498  + 1246.90128931980 *T)
    sumX0  +=       0.980 * cos(4.39283607857  + 1246.41365435280 *T)
    sumX0  +=       1.025 * cos(3.23838487665  + 1187.09754427950 *T)
    sumX0  +=       0.899 * cos(0.54202019063  + 663.55941046670 *T)
    sumX0  +=       0.966 * cos(1.79575119893  + 607.00400300581 *T)
    sumX0  +=       0.974 * cos(0.27626387870  + 600.78422788230 *T)
    sumX0  +=       1.052 * cos(6.22307301865  + 1901.77642916080 *T)
    sumX0  +=       0.939 * cos(4.72156953852  + 12573.50906446710 *T)
    sumX0  +=       0.875 * cos(2.22995963009  + 521.85802773080 *T)
    sumX0  +=       0.927 * cos(0.12921366968  + 3583.64715792471 *T)
    sumX0  +=       0.999 * cos(5.03574675925  + 2442.51309451930 *T)
    sumX0  +=       0.868 * cos(2.01963693018  + 449.73655803940 *T)
    sumX0  +=       0.920 * cos(0.32639183017  + 184.60108995130 *T)
    sumX0  +=       0.916 * cos(1.52011458512  + 860.06611126930 *T)
    sumX0  +=       1.151 * cos(1.92144343659  + 775.68965525160 *T)
    sumX0  +=       1.064 * cos(0.14103713575  + 9775.39783758759 *T)
    sumX0  +=       0.991 * cos(1.62300930873  + 447.55200204300 *T)
    sumX0  +=       0.934 * cos(1.28685601483  + 52168.93655363109 *T)
    sumX0  +=       0.852 * cos(6.16455890838  + 4760.34895633411 *T)
    sumX0  +=       0.879 * cos(3.35373437012  + 260.60403666850 *T)
    sumX0  +=       0.961 * cos(2.59929190647  + 10529.92123335110 *T)
    sumX0  +=       0.878 * cos(1.53529030972  + 166.48178917450 *T)
    sumX0  +=       1.023 * cos(3.25205599061  + 517.40460960590 *T)
    sumX0  +=       1.056 * cos(0.94712347846  + 622.24498109120 *T)
    sumX0  +=       0.857 * cos(3.62784678051  + 905.64276230799 *T)
    sumX0  +=       0.858 * cos(4.37974924972  + 866.77352135980 *T)
    sumX0  +=       0.960 * cos(0.24582787127  + 5842.00674763030 *T)
    sumX0  +=       1.006 * cos(1.79633164069  + 1332.72858815100 *T)
    sumX0  +=       0.859 * cos(2.62618934557  + 1886.79857563380 *T)
    sumX0  +=       1.053 * cos(4.19119124113  + 527.48710202330 *T)
    sumX0  +=       0.937 * cos(1.00314188520  + 597.60283414461 *T)
    sumX0  +=       0.858 * cos(0.41037816799  + 1035.24672528451 *T)
    sumX0  +=       1.183 * cos(2.27973890086  + 115.37912970730 *T)
    sumX0  +=       0.920 * cos(1.30633459352  + 589.73852761840 *T)
    sumX0  +=       0.920 * cos(4.44792724711  + 589.25089265140 *T)
    sumX0  +=       0.852 * cos(0.72219554639  + 146.13798591339 *T)
    sumX0  +=       0.852 * cos(3.86378819998  + 146.62562088040 *T)
    sumX0  +=       0.844 * cos(2.53592267944  + 10.27812579110 *T)
    sumX0  +=       0.844 * cos(5.67751533303  + 9.79049082409 *T)
    sumX0  +=       0.959 * cos(1.94261627589  + 177.15498242291 *T)
    sumX0  +=       0.959 * cos(2.94700989070  + 249.93084342011 *T)
    sumX0  +=       1.035 * cos(4.17393929319  + 507.35574708330 *T)
    sumX0  +=       0.988 * cos(5.11539514032  + 67.94549894479 *T)
    sumX0  +=       1.133 * cos(4.92435407900  + 36.40474544600 *T)
    sumX0  +=       0.843 * cos(0.19283077743  + 178.37882270030 *T)
    sumX0  +=       0.843 * cos(4.69679538916  + 248.70700314271 *T)
    sumX0  +=       0.836 * cos(6.26008270853  + 951.21941334670 *T)
    sumX0  +=       1.000 * cos(0.00060532281  + 598.59967188590 *T)
    sumX0  +=       0.870 * cos(1.56871527769  + 2520.33307447240 *T)
    sumX0  +=       1.010 * cos(3.29456379641  + 525.26891613211 *T)
    sumX0  +=       0.869 * cos(1.69757216185  + 803.30074761480 *T)
    sumX0  +=       0.862 * cos(0.89788175589  + 1336.66074141410 *T)
    sumX0  +=       0.865 * cos(2.71043672332  + 1535.24510788349 *T)
    sumX0  +=       1.156 * cos(0.63783988377  + 988.02554284310 *T)
    sumX0  +=       0.854 * cos(3.20756177408  + 3119.95596677750 *T)
    sumX0  +=       0.854 * cos(3.20756177408  + 3120.44360174450 *T)
    sumX0  +=       0.959 * cos(0.01125319100  + 167.80869531589 *T)
    sumX0  +=       0.843 * cos(5.93933593135  + 769.57294599210 *T)
    sumX0  +=       0.819 * cos(5.32247266092  + 20006.84263106329 *T)
    sumX0  +=       0.871 * cos(2.39559363846  + 2015.91490364330 *T)
    sumX0  +=       1.145 * cos(2.59566377091  + 982.39646855060 *T)
    sumX0  +=       0.967 * cos(1.79131153995  + 274.09741748721 *T)
    sumX0  +=       0.811 * cos(3.51067946836  + 310.95842873781 *T)
    sumX0  +=       1.076 * cos(0.55295132771  + 195.64679021520 *T)
    sumX0  +=       0.973 * cos(0.39005913258  + 480.00777927849 *T)
    sumX0  +=       0.927 * cos(2.26192813378  + 221.92848430950 *T)
    sumX0  +=       0.927 * cos(5.76929068640  + 205.15734153350 *T)
    sumX0  +=       1.043 * cos(3.80833299490  + 1054.47224625970 *T)
    sumX0  +=       1.118 * cos(5.22401073053  + 241.36645360580 *T)
    sumX0  +=       1.087 * cos(0.94496532019  + 26404.53882871430 *T)
    sumX0  +=       0.843 * cos(5.30493118625  + 9470.05166819489 *T)
    sumX0  +=       0.799 * cos(2.23950305681  + 556.27385055410 *T)
    sumX0  +=       0.916 * cos(1.67106075742  + 454.66554904380 *T)
    sumX0  +=       0.867 * cos(0.75856039322  + 155.43608891061 *T)
    sumX0  +=       1.078 * cos(5.29407779979  + 903.16145185841 *T)
    sumX0  +=       0.950 * cos(0.01679256709  + 469.48334387639 *T)
    sumX0  +=       0.972 * cos(1.35094632973  + 317.15695200191 *T)
    sumX0  +=       0.793 * cos(3.97038609207  + 529.88659846831 *T)
    sumX0  +=       0.798 * cos(2.46893889261  + 438.14834105430 *T)
    sumX0  +=       0.885 * cos(0.81574627494  + 685.23011986920 *T)
    sumX0  +=       0.791 * cos(4.63298065332  + 206.46299581550 *T)
    sumX0  +=       0.919 * cos(2.21054616872  + 524.25752417581 *T)
    sumX0  +=       0.919 * cos(5.35213882231  + 523.76988920880 *T)
    sumX0  +=       0.841 * cos(5.71772687954  + 3590.76070492550 *T)
    sumX0  +=       0.841 * cos(5.71772687954  + 3590.27306995850 *T)
    sumX0  +=       1.022 * cos(3.33436205114  + 731.92754532131 *T)
    sumX0  +=       0.978 * cos(4.52983537485  + 443.61984877990 *T)
    sumX0  +=       0.923 * cos(3.89253221329  + 261.81332273089 *T)
    sumX0  +=       0.860 * cos(5.19233645312  + 1784.71968549900 *T)
    sumX0  +=       0.803 * cos(2.43021611713  + 19992.61553706169 *T)
    sumX0  +=       0.948 * cos(2.63670367258  + 3280.99474730270 *T)
    sumX0  +=       0.852 * cos(1.68072750753  + 548.09402984280 *T)
    sumX0  +=       0.873 * cos(5.56384706144  + 63.60395540450 *T)
    sumX0  +=       0.956 * cos(4.90457947532  + 108.47803102759 *T)
    sumX0  +=       0.972 * cos(3.15655108462  + 892.79707515070 *T)
    sumX0  +=       0.926 * cos(4.87388996442  + 1025.68549772890 *T)
    sumX0  +=       0.914 * cos(0.82646186068  + 832.34563644899 *T)
    sumX0  +=       1.030 * cos(1.54922251715  + 586.55713388070 *T)
    sumX0  +=       0.809 * cos(0.02318871977  + 1235.36795408890 *T)
    sumX0  +=       0.809 * cos(2.35631924662  + 1310.14955265620 *T)
    sumX0  +=       0.784 * cos(1.33531692413  + 1699.52303398670 *T)
    sumX0  +=       1.020 * cos(3.91431359487  + 102.96083131970 *T)
    sumX0  +=       0.817 * cos(0.58884416481  + 5.90652167079 *T)
    sumX0  +=       1.007 * cos(5.89584195442  + 740.05248427840 *T)
    sumX0  +=       0.872 * cos(2.51565061639  + 63.38020623529 *T)
    sumX0  +=       0.934 * cos(0.00461953261  + 721.89323701370 *T)
    sumX0  +=       0.780 * cos(4.93205218405  + 930.77254422160 *T)
    sumX0  +=       1.031 * cos(0.90633658665  + 738.08037660310 *T)
    sumX0  +=       0.793 * cos(5.83091384800  + 746.41527175380 *T)
    sumX0  +=       1.085 * cos(2.59531505198  + 592.44684060410 *T)
    sumX0  +=       0.961 * cos(4.24333704051  + 1120.39270055950 *T)
    sumX0  +=       0.888 * cos(3.27274826634  + 26514.74514993370 *T)
    sumX0  +=       0.782 * cos(0.97176938732  + 519.63984183961 *T)
    sumX0  +=       0.800 * cos(4.78070775485  + 4113.33812301930 *T)
    sumX0  +=       0.945 * cos(2.48989407033  + 1673.76260475420 *T)
    sumX0  +=       0.850 * cos(1.88127295930  + 580.33735875721 *T)
    sumX0  +=       1.030 * cos(2.74042787073  + 3657.24811383991 *T)
    sumX0  +=       0.775 * cos(1.42999892472  + 735.58451193849 *T)
    sumX0  +=       0.775 * cos(4.57159157831  + 736.07214690550 *T)
    sumX0  +=       0.951 * cos(2.79912666955  + 604.22874617840 *T)
    sumX0  +=       0.760 * cos(0.06105636609  + 3267.25528826810 *T)
    sumX0  +=       0.802 * cos(2.87236079717  + 362.61847508910 *T)
    sumX0  +=       0.866 * cos(3.55253106731  + 1293.63483679419 *T)
    sumX0  +=       0.766 * cos(0.39602652848  + 188.67625556630 *T)
    sumX0  +=       0.768 * cos(3.19838844999  + 845.57589270931 *T)
    sumX0  +=       0.766 * cos(6.03837318319  + 2439.33170078160 *T)
    sumX0  +=       0.769 * cos(4.30008269962  + 1051.97638159510 *T)
    sumX0  +=       0.776 * cos(4.35680451630  + 2014.27037224060 *T)
    sumX0  +=       0.771 * cos(1.21518951297  + 1052.07274981469 *T)
    sumX0  +=       0.822 * cos(4.41913668940  + 582.62498061761 *T)
    sumX0  +=       1.002 * cos(4.18366137124  + 52183.16364763270 *T)
    sumX0  +=       0.828 * cos(2.60868899519  + 631.78916230430 *T)
    sumX0  +=       0.973 * cos(5.48773198848  + 519.14971474510 *T)
    sumX0  +=       0.839 * cos(0.93524853813  + 13362.69352428270 *T)
    sumX0  +=       0.842 * cos(3.25093404707  + 204.45725824539 *T)
    sumX0  +=       0.756 * cos(0.00983905715  + 693.82813957829 *T)
    sumX0  +=       0.756 * cos(3.15143171073  + 694.31577454531 *T)
    sumX0  +=       0.747 * cos(6.11050160069  + 962.75274857760 *T)
    sumX0  +=       0.916 * cos(5.24490716330  + 688.41151360689 *T)
    sumX0  +=       0.769 * cos(1.54076301375  + 6058.48723680599 *T)
    sumX0  +=       0.756 * cos(2.12064124784  + 545.71578525720 *T)
    sumX0  +=       0.830 * cos(1.92350323862  + 595.47066777489 *T)
    sumX0  +=       0.831 * cos(0.77280313250  + 436.87631666710 *T)
    sumX0  +=       0.776 * cos(1.38249726509  + 187.99493201010 *T)
    sumX0  +=       0.757 * cos(4.23072662628  + 5496.22546986590 *T)
    sumX0  +=       0.977 * cos(6.20713264303  + 1055.43296197871 *T)
    sumX0  +=       0.737 * cos(3.54155313450  + 225.58545092669 *T)
    sumX0  +=       0.940 * cos(2.94283899053  + 462.00609858071 *T)
    sumX0  +=       0.881 * cos(5.71084129732  + 157.67132114430 *T)
    sumX0  +=       0.881 * cos(5.46197017645  + 269.41450469871 *T)
    sumX0  +=       0.768 * cos(2.97725228498  + 2093.24724862940 *T)
    sumX0  +=       0.945 * cos(5.11643863868  + 347.65517577710 *T)
    sumX0  +=       0.789 * cos(2.80405308992  + 678.27413943531 *T)
    sumX0  +=       0.894 * cos(5.28864466497  + 185.08872491831 *T)
    sumX0  +=       0.887 * cos(0.33081399998  + 938.63685074779 *T)
    sumX0  +=       0.848 * cos(0.16881037460  + 1016.59984305279 *T)
    sumX0  +=       0.778 * cos(0.01629844158  + 5753.62870238030 *T)
    sumX0  +=       0.778 * cos(3.15789109517  + 5753.14106741329 *T)
    sumX0  +=       0.843 * cos(1.02624495037  + 976.49220761220 *T)
    sumX0  +=       0.900 * cos(2.19862220278  + 655.17383907870 *T)
    sumX0  +=       0.730 * cos(3.17028402119  + 497.95457822210 *T)
    sumX0  +=       0.879 * cos(3.20284714972  + 494.23487328010 *T)
    sumX0  +=       0.728 * cos(4.17683406683  + 995.88984936931 *T)
    sumX0  +=       0.729 * cos(5.48820634686  + 632.58810593949 *T)
    sumX0  +=       0.729 * cos(2.34661369327  + 633.07574090651 *T)
    sumX0  +=       0.841 * cos(0.88660531241  + 1051.06135785839 *T)
    sumX0  +=       0.810 * cos(6.22385696664  + 3900.03902758130 *T)
    sumX0  +=       0.774 * cos(1.63307607705  + 451.48415530609 *T)
    sumX0  +=       0.792 * cos(1.63932414999  + 193.93282284329 *T)
    sumX0  +=       0.894 * cos(1.22695933225  + 103.28840759231 *T)
    sumX0  +=       0.744 * cos(5.76295987616  + 662.77502104650 *T)
    sumX0  +=       1.019 * cos(4.06927095682  + 7232.49527644471 *T)
    sumX0  +=       0.720 * cos(6.27725825901  + 103711.47146249928 *T)
    sumX0  +=       0.776 * cos(1.68634145832  + 1703.45518724980 *T)
    sumX0  +=       1.001 * cos(2.29066167531  + 1126.54303971381 *T)
    sumX0  +=       0.860 * cos(1.85082345148  + 2748.61002343740 *T)
    sumX0  +=       0.716 * cos(2.80780296922  + 4319.03603648950 *T)
    sumX0  +=       0.716 * cos(5.94939562281  + 4319.52367145650 *T)
    sumX0  +=       0.797 * cos(0.88312421788  + 529.98296668790 *T)
    sumX0  +=       0.729 * cos(3.08861395930  + 1649.38358236600 *T)
    sumX0  +=       0.803 * cos(2.56866081755  + 151.01630068050 *T)
    sumX0  +=       0.765 * cos(1.92150517901  + 1379.46213562540 *T)
    sumX0  +=       0.765 * cos(5.06309783260  + 1379.94977059240 *T)
    sumX0  +=       0.749 * cos(3.86728999613  + 785.77214766900 *T)
    sumX0  +=       0.836 * cos(4.09079447369  + 4010.24534880071 *T)
    sumX0  +=       0.727 * cos(1.76200509604  + 608.16089944150 *T)
    sumX0  +=       0.727 * cos(1.76200509604  + 608.64853440851 *T)
    sumX0  +=       0.719 * cos(1.08122225861  + 467.68335698301 *T)
    sumX0  +=       0.709 * cos(2.16805792847  + 103.54904002320 *T)
    sumX0  +=       0.709 * cos(2.16805792847  + 103.06140505620 *T)
    sumX0  +=       0.920 * cos(3.37948839250  + 870.70567462290 *T)
    sumX0  +=       0.716 * cos(3.01984388520  + 664.78325074409 *T)
    sumX0  +=       0.857 * cos(4.73776093669  + 2306.54634406740 *T)
    sumX0  +=       0.724 * cos(4.70354725719  + 740.52556503041 *T)
    sumX0  +=       0.708 * cos(4.72112546056  + 2431.46739425540 *T)
    sumX0  +=       0.706 * cos(4.95679658990  + 102.63650841399 *T)
    sumX0  +=       0.706 * cos(4.95679658990  + 103.12414338101 *T)
    sumX0  +=       0.824 * cos(0.35265072853  + 1163.68172973780 *T)
    sumX0  +=       0.722 * cos(3.07523457584  + 171.21709158970 *T)
    sumX0  +=       0.723 * cos(0.74571067860  + 829.37669103239 *T)
    sumX0  +=       0.720 * cos(1.80911156979  + 255.86873425330 *T)
    sumX0  +=       0.956 * cos(1.61313021087  + 3068.18328683170 *T)
    sumX0  +=       0.800 * cos(4.60599313494  + 279.72649177971 *T)
    sumX0  +=       0.800 * cos(0.28363303165  + 147.35933406330 *T)
    sumX0  +=       0.868 * cos(5.08886004770  + 669.40093308030 *T)
    sumX0  +=       0.769 * cos(4.42409463602  + 3465.83965473750 *T)
    sumX0  +=       0.745 * cos(4.41268882152  + 843.14525849700 *T)
    sumX0  +=       0.848 * cos(4.34921581475  + 103.38477581190 *T)
    sumX0  +=       0.703 * cos(0.03980855697  + 4003.13180179990 *T)
    sumX0  +=       0.719 * cos(2.96421431820  + 9988.69693302559 *T)
    sumX0  +=       0.818 * cos(3.58195947060  + 311.48218572709 *T)
    sumX0  +=       0.901 * cos(5.50487267537  + 937.88609122240 *T)
    sumX0  +=       0.783 * cos(0.40400095682  + 199.83708378171 *T)
    sumX0  +=       0.783 * cos(1.34403255618  + 227.24874206131 *T)
    sumX0  +=       0.909 * cos(5.35487801471  + 4009.75771383370 *T)
    sumX0  +=       0.711 * cos(0.19813010381  + 457.37386202950 *T)
    sumX0  +=       0.705 * cos(1.55599502521  + 5653.98044645779 *T)
    sumX0  +=       0.712 * cos(4.50433897135  + 354.28108781090 *T)
    sumX0  +=       0.733 * cos(4.93048865178  + 37.62858572340 *T)
    sumX0  +=       0.708 * cos(2.84637541601  + 840.69757794220 *T)
    sumX0  +=       0.708 * cos(5.98796806961  + 840.20994297520 *T)
    sumX0  +=       0.695 * cos(4.51136204527  + 812.70191647599 *T)
    sumX0  +=       0.695 * cos(4.51136204527  + 813.18955144300 *T)
    sumX0  +=       0.736 * cos(1.20053248385  + 403.94745867960 *T)
    sumX0  +=       0.733 * cos(0.72827400906  + 182.03578931750 *T)
    sumX0  +=       0.836 * cos(3.44633331041  + 2331.81913833290 *T)
    sumX0  +=       0.928 * cos(2.16436836139  + 842.13386654071 *T)
    sumX0  +=       0.768 * cos(5.27501406783  + 1460.20038475780 *T)
    sumX0  +=       0.685 * cos(5.86089797553  + 455.29619636271 *T)
    sumX0  +=       0.792 * cos(1.62315871295  + 5422.52210375510 *T)
    sumX0  +=       0.837 * cos(5.42340126188  + 570.77613120160 *T)
    sumX0  +=       0.682 * cos(4.88563990270  + 3693.85347914410 *T)
    sumX0  +=       0.752 * cos(2.13635874109  + 413.34862754081 *T)
    sumX0  +=       0.817 * cos(1.53257391407  + 391.73279989249 *T)
    sumX0  +=       0.681 * cos(2.40403440268  + 458.33457774851 *T)
    sumX0  +=       0.803 * cos(5.00204662708  + 418.49006686629 *T)
    sumX0  +=       0.681 * cos(0.38545629971  + 414.57495994570 *T)

    # SATURN X Coord
    sumX1  = 0
    sumX1  +=  9285877.988 * cos(0.61678993503  + 0.24381748350 *T)
    sumX1  +=  3086501.680 * cos(4.27493632359  + 426.84200835950 *T)
    sumX1  +=  2728479.923 * cos(5.84476389020  + 206.42936592071 *T)
    sumX1  +=  2644990.371 * cos(5.33256382404  + 220.65645992230 *T)
    sumX1  +=  629201.988 * cos(0.32457757988  + 7.35736448430 *T)
    sumX1  +=  256626.023 * cos(3.52453971595  + 640.14110379750 *T)
    sumX1  +=  312353.889 * cos(4.82956308970  + 419.72846135871 *T)
    sumX1  +=  189297.041 * cos(4.48614588896  + 433.95555536030 *T)
    sumX1  +=  204465.020 * cos(1.07744722386  + 213.54291292150 *T)
    sumX1  +=  118113.185 * cos(1.16350724193  + 110.45013870291 *T)
    sumX1  +=   67024.295 * cos(3.73239360544  + 316.63568714010 *T)
    sumX1  +=   66948.413 * cos(5.20221291915  + 227.77000692311 *T)
    sumX1  +=   31931.062 * cos(5.99122459014  + 199.31581891990 *T)
    sumX1  +=   24548.187 * cos(1.72318512907  + 205.94173095370 *T)
    sumX1  +=   30521.021 * cos(0.18185692767  + 14.47091148511 *T)
    sumX1  +=   22265.017 * cos(6.19403063806  + 103.33659170210 *T)
    sumX1  +=   18939.520 * cos(2.77554483793  + 853.44019923550 *T)
    sumX1  +=   18082.904 * cos(5.08903997736  + 209.61075965840 *T)
    sumX1  +=   17786.477 * cos(6.10321536300  + 217.47506618460 *T)
    sumX1  +=   16307.253 * cos(4.86906758451  + 216.72430665921 *T)
    sumX1  +=   16919.177 * cos(4.58471330942  + 633.02755679670 *T)
    sumX1  +=   15920.273 * cos(0.03453366907  + 210.36151918381 *T)
    sumX1  +=   16167.186 * cos(5.61042793206  + 323.74923414091 *T)
    sumX1  +=   14472.589 * cos(3.67415332596  + 647.25465079831 *T)
    sumX1  +=   11071.575 * cos(0.03003679044  + 117.56368570370 *T)
    sumX1  +=    9887.534 * cos(5.18637985637  + 202.49721265760 *T)
    sumX1  +=    8713.930 * cos(6.03455153122  + 224.58861318540 *T)
    sumX1  +=    6197.025 * cos(1.36464349110  + 213.05527795450 *T)
    sumX1  +=    5502.258 * cos(4.40320486328  + 441.06910236111 *T)
    sumX1  +=    5466.098 * cos(2.60358233388  + 11.28951774740 *T)
    sumX1  +=    3484.091 * cos(2.54141592195  + 6.86972951729 *T)
    sumX1  +=    3596.198 * cos(2.50280897492  + 309.03450517230 *T)
    sumX1  +=    3029.001 * cos(1.24873358475  + 522.33360061030 *T)
    sumX1  +=    3353.859 * cos(3.74904517566  + 95.73540973430 *T)
    sumX1  +=    2987.715 * cos(5.64325133005  + 309.52214013930 *T)
    sumX1  +=    2732.424 * cos(4.10964270304  + 149.80701461810 *T)
    sumX1  +=    3741.808 * cos(4.82356861468  + 412.61491435790 *T)
    sumX1  +=    2374.342 * cos(1.71597491607  + 198.82818395289 *T)
    sumX1  +=    2875.345 * cos(0.10934318294  + 426.35437339250 *T)
    sumX1  +=    2121.841 * cos(1.70223145921  + 220.16882495529 *T)
    sumX1  +=    2386.687 * cos(0.22569849869  + 109.96250373590 *T)
    sumX1  +=    1755.551 * cos(0.77463091610  + 277.27881122490 *T)
    sumX1  +=    1690.652 * cos(3.97638248269  + 422.90985509640 *T)
    sumX1  +=    1545.396 * cos(5.30260231719  + 330.86278114170 *T)
    sumX1  +=    1447.492 * cos(2.03498842518  + 4.17597074660 *T)
    sumX1  +=    1392.451 * cos(2.03410425344  + 1066.73929467350 *T)
    sumX1  +=    1331.483 * cos(5.06295810929  + 234.88355392390 *T)
    sumX1  +=    1452.348 * cos(6.21365122108  + 3.42521122121 *T)
    sumX1  +=    1285.559 * cos(2.40490814955  + 415.79630809560 *T)
    sumX1  +=    1159.539 * cos(4.60389667301  + 537.04832957890 *T)
    sumX1  +=    1128.476 * cos(5.45092270832  + 423.66061462181 *T)
    sumX1  +=    1131.769 * cos(6.00809348724  + 529.93478257810 *T)
    sumX1  +=    1127.023 * cos(2.88170875021  + 860.55374623631 *T)
    sumX1  +=    1101.803 * cos(4.07656502687  + 430.02340209721 *T)
    sumX1  +=    1061.927 * cos(4.38341564719  + 846.32665223470 *T)
    sumX1  +=    1343.360 * cos(2.98910861966  + 211.09523236671 *T)
    sumX1  +=    1331.640 * cos(1.90897752717  + 215.99059347631 *T)
    sumX1  +=    1287.848 * cos(2.08433845806  + 522.82123557730 *T)
    sumX1  +=    1050.973 * cos(0.41490086140  + 412.12727939089 *T)
    sumX1  +=     900.723 * cos(0.97940540803  + 838.72547026689 *T)
    sumX1  +=     882.663 * cos(5.41577217529  + 437.88770862340 *T)
    sumX1  +=     924.505 * cos(0.31084816096  + 743.23387801611 *T)
    sumX1  +=     876.882 * cos(6.15951573282  + 96.22304470130 *T)
    sumX1  +=     905.253 * cos(3.37316127293  + 3.68833577960 *T)
    sumX1  +=     820.808 * cos(4.74555854872  + 223.83785366000 *T)
    sumX1  +=     844.907 * cos(2.24191891660  + 625.42637482889 *T)
    sumX1  +=     760.731 * cos(0.32645781021  + 625.91400979590 *T)
    sumX1  +=     657.423 * cos(1.75192962523  + 316.14805217309 *T)
    sumX1  +=     813.631 * cos(0.02636548240  + 21.58445848590 *T)
    sumX1  +=     642.768 * cos(0.08669014735  + 1059.62574767270 *T)
    sumX1  +=     657.845 * cos(3.90002282220  + 88.62186273350 *T)
    sumX1  +=     715.042 * cos(6.12383205661  + 430.77416162260 *T)
    sumX1  +=     596.876 * cos(5.09870868637  + 124.67723270450 *T)
    sumX1  +=     472.020 * cos(3.60910799409  + 654.36819779910 *T)
    sumX1  +=     467.656 * cos(1.37893062096  + 515.22005360950 *T)
    sumX1  +=     472.526 * cos(0.17976235838  + 203.24797218300 *T)
    sumX1  +=     487.417 * cos(1.14265056645  + 736.12033101530 *T)
    sumX1  +=     402.114 * cos(4.42612765272  + 86.07111631471 *T)
    sumX1  +=     376.954 * cos(3.71201411962  + 76.50988875911 *T)
    sumX1  +=     419.410 * cos(5.88905869137  + 742.74624304909 *T)
    sumX1  +=     439.384 * cos(2.62527563052  + 301.92095817150 *T)
    sumX1  +=     458.871 * cos(0.66305280938  + 529.44714761109 *T)
    sumX1  +=     345.395 * cos(6.17484774422  + 192.20227191910 *T)
    sumX1  +=     334.209 * cos(0.71189291081  + 1155.11733992350 *T)
    sumX1  +=     330.276 * cos(0.35761199969  + 302.40859313851 *T)
    sumX1  +=     339.713 * cos(2.73817515158  + 218.20877936750 *T)
    sumX1  +=     304.950 * cos(0.18421276814  + 231.70216018621 *T)
    sumX1  +=     312.395 * cos(2.36627641720  + 629.09540353360 *T)
    sumX1  +=     311.746 * cos(1.93452684244  + 941.81824448550 *T)
    sumX1  +=     299.095 * cos(2.15964302278  + 208.87704647550 *T)
    sumX1  +=     299.955 * cos(0.17404785977  + 728.51914904750 *T)
    sumX1  +=     271.105 * cos(2.70363094480  + 288.32451148881 *T)
    sumX1  +=     260.779 * cos(0.87015911392  + 18.40306474821 *T)
    sumX1  +=     278.736 * cos(5.35893413930  + 632.53992182970 *T)
    sumX1  +=     335.527 * cos(3.07953134072  + 203.98168536590 *T)
    sumX1  +=     259.607 * cos(6.00394254983  + 10.53875822200 *T)
    sumX1  +=     263.983 * cos(3.55443649812  + 363.10611005610 *T)
    sumX1  +=     291.544 * cos(1.81131242466  + 223.10414047710 *T)
    sumX1  +=     214.376 * cos(6.03151333634  + 208.12628695010 *T)
    sumX1  +=     208.551 * cos(5.16427411899  + 218.95953889290 *T)
    sumX1  +=     208.257 * cos(4.38726258110  + 408.68276109481 *T)
    sumX1  +=     211.653 * cos(2.10642373554  + 138.76131435421 *T)
    sumX1  +=     199.845 * cos(2.29578299708  + 13.98327651810 *T)
    sumX1  +=     195.781 * cos(0.35428355710  + 52.44638055600 *T)
    sumX1  +=     192.346 * cos(6.21537031577  + 201.01273994931 *T)
    sumX1  +=     191.573 * cos(0.44811077600  + 341.01470952831 *T)
    sumX1  +=     185.995 * cos(4.78683696303  + 948.93179148630 *T)
    sumX1  +=     168.803 * cos(1.10901900891  + 831.61192326610 *T)
    sumX1  +=     170.172 * cos(1.18102994579  + 350.57593708390 *T)
    sumX1  +=     167.553 * cos(5.00268725720  + 226.07308589371 *T)
    sumX1  +=     177.510 * cos(4.94085831586  + 210.62215161470 *T)
    sumX1  +=     177.755 * cos(6.22073652647  + 216.46367422830 *T)
    sumX1  +=     157.494 * cos(1.71852514061  + 195.38366565681 *T)
    sumX1  +=     162.215 * cos(6.01667468309  + 956.53297345411 *T)
    sumX1  +=     160.622 * cos(2.21143597606  + 127.22797912329 *T)
    sumX1  +=     147.298 * cos(0.31012308764  + 9.80504503910 *T)
    sumX1  +=     177.635 * cos(6.27025545843  + 160.85271488200 *T)
    sumX1  +=     170.308 * cos(2.63474062872  + 207.91383862900 *T)
    sumX1  +=     158.062 * cos(5.63276958593  + 1059.13811270570 *T)
    sumX1  +=     138.665 * cos(0.44735566636  + 1471.50920958009 *T)
    sumX1  +=     150.976 * cos(3.13927888955  + 636.20895053440 *T)
    sumX1  +=     136.839 * cos(4.51690712170  + 544.16187657970 *T)
    sumX1  +=     163.932 * cos(2.23903496342  + 219.17198721400 *T)
    sumX1  +=     164.921 * cos(4.61861878322  + 22.33521801130 *T)
    sumX1  +=     134.204 * cos(1.25807922003  + 703.87700210090 *T)
    sumX1  +=     141.934 * cos(0.59018411736  + 202.00957769060 *T)
    sumX1  +=     153.018 * cos(4.27626456939  + 56.37853381910 *T)
    sumX1  +=     132.722 * cos(3.97656514012  + 750.34742501690 *T)
    sumX1  +=     136.617 * cos(1.64500267431  + 1258.21011414209 *T)
    sumX1  +=     124.984 * cos(4.28860125524  + 448.18264936190 *T)
    sumX1  +=     138.310 * cos(4.40216198828  + 142.69346761730 *T)
    sumX1  +=     121.140 * cos(0.16586240985  + 949.41942645330 *T)
    sumX1  +=     146.873 * cos(0.47521375650  + 845.83901726770 *T)
    sumX1  +=     135.623 * cos(0.85079579674  + 490.57790666290 *T)
    sumX1  +=     114.737 * cos(0.83509600385  + 433.46792039329 *T)
    sumX1  +=     107.506 * cos(2.74641093550  + 565.35950523021 *T)
    sumX1  +=     105.407 * cos(3.58312013649  + 209.35012722751 *T)
    sumX1  +=     123.012 * cos(2.85293627235  + 483.46435966210 *T)
    sumX1  +=     104.647 * cos(1.30261134749  + 217.73569861551 *T)
    sumX1  +=     123.445 * cos(4.85723767211  + 266.23311096100 *T)
    sumX1  +=     121.327 * cos(2.34436045601  + 618.31282782810 *T)
    sumX1  +=     121.539 * cos(2.28006635810  + 515.70768857651 *T)
    sumX1  +=      98.330 * cos(1.27153379840  + 1280.03839011150 *T)
    sumX1  +=     123.139 * cos(1.72477244855  + 12.77399045571 *T)
    sumX1  +=     106.461 * cos(6.21307355690  + 1044.91101870410 *T)
    sumX1  +=     134.338 * cos(0.13238469429  + 419.24082639170 *T)
    sumX1  +=     107.530 * cos(2.04781430457  + 424.39432780471 *T)
    sumX1  +=     114.356 * cos(3.50494100764  + 270.16526422410 *T)
    sumX1  +=      94.347 * cos(4.48494950953  + 1265.32366114289 *T)
    sumX1  +=     104.237 * cos(3.80594641363  + 70.60562782070 *T)
    sumX1  +=      88.410 * cos(2.09844529978  + 1073.85284167431 *T)
    sumX1  +=      91.547 * cos(0.19575334067  + 106.51798543980 *T)
    sumX1  +=      84.219 * cos(6.17220906984  + 113.63153244060 *T)
    sumX1  +=      91.641 * cos(2.30922052677  + 497.69145366371 *T)
    sumX1  +=      80.939 * cos(4.70817863540  + 651.18680406140 *T)
    sumX1  +=      82.304 * cos(3.28548883281  + 643.32249753521 *T)
    sumX1  +=      99.237 * cos(5.82968286033  + 639.65346883050 *T)
    sumX1  +=      73.063 * cos(0.84493736921  + 1148.00379292269 *T)
    sumX1  +=      79.913 * cos(3.01047372573  + 191.45151239371 *T)
    sumX1  +=      70.813 * cos(2.40778621473  + 251.67594855930 *T)
    sumX1  +=      72.632 * cos(1.31854571732  + 107.26874496520 *T)
    sumX1  +=      75.817 * cos(1.44164476880  + 284.39235822571 *T)
    sumX1  +=      72.415 * cos(1.87310857657  + 191.71463695210 *T)
    sumX1  +=      84.109 * cos(1.77169250100  + 219.69325207580 *T)
    sumX1  +=      66.814 * cos(4.85072561146  + 636.95971005981 *T)
    sumX1  +=      64.253 * cos(4.88041745139  + 71.09326278771 *T)
    sumX1  +=      73.347 * cos(4.31971264670  + 344.94686279140 *T)
    sumX1  +=      61.578 * cos(5.13231029134  + 479.53220639900 *T)
    sumX1  +=      60.447 * cos(0.18657130674  + 1787.90107923670 *T)
    sumX1  +=      63.039 * cos(3.97913628016  + 437.13694909800 *T)
    sumX1  +=      64.265 * cos(1.35985415674  + 1574.60198379870 *T)
    sumX1  +=      58.169 * cos(3.07003751562  + 75.02541605080 *T)
    sumX1  +=      65.491 * cos(5.37959870394  + 1375.52998236229 *T)
    sumX1  +=      57.055 * cos(1.29146537045  + 98.65617104110 *T)
    sumX1  +=      63.072 * cos(3.82756140289  + 359.17395679300 *T)
    sumX1  +=      76.825 * cos(0.26061032610  + 839.21310523390 *T)
    sumX1  +=      65.944 * cos(3.15806762275  + 207.39257376720 *T)
    sumX1  +=      65.704 * cos(0.38527237466  + 2.93757625420 *T)
    sumX1  +=      59.449 * cos(4.25418200211  + 417.28078080390 *T)
    sumX1  +=      66.539 * cos(0.19553179802  + 1162.23088692429 *T)
    sumX1  +=      53.194 * cos(1.57281678134  + 227.28237195610 *T)
    sumX1  +=      63.890 * cos(1.98911021042  + 235.63431344930 *T)
    sumX1  +=      52.968 * cos(2.48775989817  + 175.40987728371 *T)
    sumX1  +=      51.480 * cos(5.40158452623  + 416.54706762100 *T)
    sumX1  +=      53.788 * cos(0.70800398665  + 114.38229196601 *T)
    sumX1  +=      52.017 * cos(2.01145465470  + 220.70464403210 *T)
    sumX1  +=      48.475 * cos(0.16935246445  + 209.12312469139 *T)
    sumX1  +=      48.625 * cos(0.32855808278  + 63.97971578690 *T)
    sumX1  +=      47.377 * cos(4.19080757740  + 1581.71553079950 *T)
    sumX1  +=      66.468 * cos(4.53149070521  + 337.97632814250 *T)
    sumX1  +=      47.697 * cos(1.91784854250  + 352.06040979221 *T)
    sumX1  +=      50.400 * cos(1.52095286213  + 429.28968891431 *T)
    sumX1  +=      51.302 * cos(1.56161815189  + 121.49583896680 *T)
    sumX1  +=      48.534 * cos(3.08596013281  + 151.29148732640 *T)
    sumX1  +=      50.418 * cos(2.66057863189  + 181.29958400710 *T)
    sumX1  +=      52.765 * cos(3.98236980014  + 4.90968392950 *T)
    sumX1  +=      56.130 * cos(0.33220104012  + 404.75060783170 *T)
    sumX1  +=      53.587 * cos(0.57493074933  + 554.31380496631 *T)
    sumX1  +=      50.974 * cos(2.05704046175  + 934.70469748469 *T)
    sumX1  +=      44.236 * cos(6.08174708544  + 63.49208081989 *T)
    sumX1  +=      46.842 * cos(2.24622013459  + 245.78624183590 *T)
    sumX1  +=      41.530 * cos(6.03788211153  + 99.40443843901 *T)
    sumX1  +=      44.045 * cos(2.25826631267  + 205.46615807421 *T)
    sumX1  +=      40.580 * cos(2.82302665464  + 867.66729323710 *T)
    sumX1  +=      41.030 * cos(4.03479265190  + 81.50831573269 *T)
    sumX1  +=      40.882 * cos(3.76127180855  + 621.98185653281 *T)
    sumX1  +=      38.986 * cos(4.46168372795  + 131.79077970531 *T)
    sumX1  +=      45.904 * cos(0.45235179276  + 1589.31671276730 *T)
    sumX1  +=      53.423 * cos(1.19152844807  + 1052.51220067191 *T)
    sumX1  +=      41.233 * cos(5.98630903924  + 1361.30288836070 *T)
    sumX1  +=      46.867 * cos(2.84588810562  + 10.80188278040 *T)
    sumX1  +=      46.594 * cos(4.62317456797  + 323.26159917390 *T)
    sumX1  +=      37.309 * cos(0.18886259626  + 194.89603068979 *T)
    sumX1  +=      36.647 * cos(1.88978550694  + 148.32254190981 *T)
    sumX1  +=      41.194 * cos(5.88019785775  + 644.07325706060 *T)
    sumX1  +=      36.297 * cos(5.66458901484  + 445.00125562421 *T)
    sumX1  +=      36.204 * cos(2.75001997487  + 214.29367244690 *T)
    sumX1  +=      41.215 * cos(2.50684492522  + 198.56505939450 *T)
    sumX1  +=      45.013 * cos(1.96686946581  + 842.39449897160 *T)
    sumX1  +=      34.449 * cos(0.57010226677  + 1464.39566257930 *T)
    sumX1  +=      34.256 * cos(2.87917030633  + 289.80898419710 *T)
    sumX1  +=      36.820 * cos(2.97826186847  + 206.38118181090 *T)
    sumX1  +=      33.887 * cos(1.08487218987  + 89.10949770051 *T)
    sumX1  +=      36.762 * cos(3.81424116419  + 265.74547599400 *T)
    sumX1  +=      32.982 * cos(5.79106870822  + 276.79117625789 *T)
    sumX1  +=      33.116 * cos(5.24689400325  + 178.11819026941 *T)
    sumX1  +=      32.822 * cos(2.19475458249  + 212.79215339611 *T)
    sumX1  +=      37.519 * cos(2.38353348143  + 228.52076644850 *T)
    sumX1  +=      31.997 * cos(1.80124089820  + 275.79433851660 *T)
    sumX1  +=      40.913 * cos(2.29234559153  + 209.87388421679 *T)
    sumX1  +=      30.291 * cos(4.38172577701  + 78.95756931390 *T)
    sumX1  +=      29.064 * cos(5.92812854477  + 248.96763557360 *T)
    sumX1  +=      31.114 * cos(1.08106077375  + 1890.99385345530 *T)
    sumX1  +=      28.459 * cos(1.12291882969  + 102.84895673509 *T)
    sumX1  +=      35.352 * cos(2.37935763988  + 501.62360692681 *T)
    sumX1  +=      27.208 * cos(6.21367293026  + 2104.29294889330 *T)
    sumX1  +=      37.148 * cos(1.38568578551  + 729.00678401450 *T)
    sumX1  +=      33.504 * cos(1.04515396423  + 436.40323591510 *T)
    sumX1  +=      32.029 * cos(3.21836665934  + 414.31183538731 *T)
    sumX1  +=      27.479 * cos(3.85996509219  + 114.64292439690 *T)
    sumX1  +=      29.371 * cos(5.71075520500  + 215.02738562980 *T)
    sumX1  +=      30.410 * cos(5.54772262940  + 31.99951143090 *T)
    sumX1  +=      33.556 * cos(6.12868949605  + 1478.62275658089 *T)
    sumX1  +=      28.365 * cos(5.81150437735  + 355.99256305531 *T)
    sumX1  +=      24.664 * cos(4.56168548884  + 427.80521620600 *T)
    sumX1  +=      28.864 * cos(5.14212159473  + 1691.92185201890 *T)
    sumX1  +=      33.503 * cos(4.61144380837  + 735.63269604830 *T)
    sumX1  +=      27.468 * cos(5.08884052063  + 405.50136735710 *T)
    sumX1  +=      23.818 * cos(2.06467933184  + 190.96387742670 *T)
    sumX1  +=      23.542 * cos(2.98679769294  + 278.76328393321 *T)
    sumX1  +=      23.335 * cos(3.92749113053  + 1898.10740045609 *T)
    sumX1  +=      24.810 * cos(0.83325589326  + 425.87880051300 *T)
    sumX1  +=      23.968 * cos(1.93756814240  + 431.50787480550 *T)
    sumX1  +=      25.517 * cos(4.02939599784  + 69.39634175830 *T)
    sumX1  +=      22.253 * cos(2.97476505625  + 602.00806815971 *T)
    sumX1  +=      23.190 * cos(0.53219820713  + 618.80046279510 *T)
    sumX1  +=      21.849 * cos(3.74059365574  + 299.37021175271 *T)
    sumX1  +=      21.270 * cos(0.55413991317  + 917.17609753890 *T)
    sumX1  +=      22.382 * cos(2.85390353571  + 576.40520549410 *T)
    sumX1  +=      21.104 * cos(4.01921632968  + 319.81708087780 *T)
    sumX1  +=      24.718 * cos(1.77317212162  + 1251.09656714130 *T)
    sumX1  +=      20.790 * cos(4.86337923369  + 241.99710092470 *T)
    sumX1  +=      25.298 * cos(1.23773943016  + 422.17614191350 *T)
    sumX1  +=      19.818 * cos(1.13017694076  + 5.66044345490 *T)
    sumX1  +=      19.530 * cos(5.09943396528  + 313.45429340240 *T)
    sumX1  +=      23.121 * cos(4.47417203680  + 230.95140066080 *T)
    sumX1  +=      19.612 * cos(1.54092409817  + 508.10650660869 *T)
    sumX1  +=      19.333 * cos(2.88512358020  + 134.34152612410 *T)
    sumX1  +=      20.375 * cos(1.15875306859  + 426.89019246930 *T)
    sumX1  +=      20.742 * cos(4.93204856327  + 312.70353387701 *T)
    sumX1  +=      19.203 * cos(5.31312736000  + 320.56784040320 *T)
    sumX1  +=      18.481 * cos(3.48969850886  + 39.60069339870 *T)
    sumX1  +=      21.067 * cos(4.89382442775  + 421.42538238810 *T)
    sumX1  +=      19.538 * cos(1.46473404121  + 189.16389053330 *T)
    sumX1  +=      20.461 * cos(5.56819424455  + 327.68138740401 *T)
    sumX1  +=      21.171 * cos(2.77619058613  + 221.61966776881 *T)
    sumX1  +=      19.130 * cos(0.13369689085  + 98.91680347199 *T)
    sumX1  +=      18.295 * cos(0.64431243991  + 295.29504613770 *T)
    sumX1  +=      18.160 * cos(5.37336395843  + 205.67860639531 *T)
    sumX1  +=      17.549 * cos(4.33484240447  + 120.11443212250 *T)
    sumX1  +=      17.626 * cos(1.33869269116  + 956.04533848709 *T)
    sumX1  +=      16.792 * cos(0.30316695138  + 1780.78753223589 *T)
    sumX1  +=      16.458 * cos(4.30804482989  + 439.37218133171 *T)
    sumX1  +=      18.239 * cos(0.48112633510  + 6276.20612047410 *T)
    sumX1  +=      16.441 * cos(5.70984800135  + 1169.83206889211 *T)
    sumX1  +=      16.263 * cos(6.02060382308  + 1272.92484311070 *T)
    sumX1  +=      16.077 * cos(1.92413826292  + 778.65860066820 *T)
    sumX1  +=      18.919 * cos(0.91986849601  + 348.12825652910 *T)
    sumX1  +=      20.156 * cos(5.28936051941  + 212.05844021321 *T)
    sumX1  +=      17.804 * cos(6.17596938223  + 6290.43321447570 *T)
    sumX1  +=      22.054 * cos(2.76742279514  + 294.80741117069 *T)
    sumX1  +=      15.573 * cos(4.69738515912  + 74.53778108379 *T)
    sumX1  +=      16.309 * cos(5.78358696622  + 1677.69475801730 *T)
    sumX1  +=      18.944 * cos(3.21110443213  + 127.71561409031 *T)
    sumX1  +=      17.127 * cos(1.90459336979  + 10206.41581669370 *T)
    sumX1  +=      19.690 * cos(2.03308592691  + 696.76345510010 *T)
    sumX1  +=      20.455 * cos(2.11078778994  + 117.07605073670 *T)
    sumX1  +=      16.447 * cos(3.43440454322  + 237.92193530970 *T)
    sumX1  +=      16.503 * cos(0.82013608893  + 2207.38572311189 *T)
    sumX1  +=      20.378 * cos(4.24340498319  + 426.79382424970 *T)
    sumX1  +=      16.713 * cos(1.31716285740  + 10220.64291069530 *T)
    sumX1  +=      17.063 * cos(4.67555598732  + 15.22167101050 *T)
    sumX1  +=      18.187 * cos(6.20263428999  + 167.07249000551 *T)
    sumX1  +=      14.830 * cos(5.67343259467  + 415.30867312860 *T)
    sumX1  +=      15.128 * cos(1.51541058227  + 710.99054910171 *T)
    sumX1  +=      14.162 * cos(6.22052289061  + 525.74199688410 *T)
    sumX1  +=      14.139 * cos(5.13566894741  + 39.11305843170 *T)
    sumX1  +=      18.060 * cos(5.83443986197  + 1795.01462623750 *T)
    sumX1  +=      13.362 * cos(6.01694375584  + 106.03035047280 *T)
    sumX1  +=      13.903 * cos(1.06036130597  + 312.44290144610 *T)
    sumX1  +=      15.359 * cos(5.25985265745  + 182.52342428450 *T)
    sumX1  +=      13.906 * cos(0.65123132262  + 59.55992755679 *T)
    sumX1  +=      13.001 * cos(3.44349383922  + 87.55558902300 *T)
    sumX1  +=      13.054 * cos(4.90885990739  + 196.27743753410 *T)
    sumX1  +=      15.661 * cos(0.24166847150  + 1162.71852189130 *T)
    sumX1  +=      17.412 * cos(2.16826879959  + 202.76033721600 *T)
    sumX1  +=      14.557 * cos(4.55052624968  + 381.59542572090 *T)
    sumX1  +=      17.110 * cos(1.29146334369  + 432.47108265200 *T)
    sumX1  +=      12.702 * cos(5.96286796963  + 2420.68481854989 *T)
    sumX1  +=      17.195 * cos(3.81192777968  + 536.56069461190 *T)
    sumX1  +=      12.663 * cos(0.27572988542  + 1376.01761732930 *T)
    sumX1  +=      15.239 * cos(5.32338258618  + 429.76276966630 *T)
    sumX1  +=      12.562 * cos(4.36343770041  + 852.95256426850 *T)
    sumX1  +=      12.487 * cos(3.78773003979  + 757.46097201770 *T)
    sumX1  +=      12.205 * cos(3.50752925238  + 661.48174479990 *T)
    sumX1  +=      14.071 * cos(4.06346090551  + 395.82251972250 *T)
    sumX1  +=      13.589 * cos(0.17638027363  + 97.11681657861 *T)
    sumX1  +=      13.297 * cos(2.42892875675  + 849.50804597240 *T)
    sumX1  +=      12.950 * cos(0.16804220244  + 91.80325647120 *T)
    sumX1  +=      12.800 * cos(0.91304411194  + 1272.43720814370 *T)
    sumX1  +=      14.060 * cos(0.19022187362  + 116.18227885939 *T)
    sumX1  +=      13.165 * cos(4.65152797704  + 129.67565967810 *T)
    sumX1  +=      11.819 * cos(0.54037105135  + 431.03479405351 *T)
    sumX1  +=      11.579 * cos(5.22220206450  + 144.39038864671 *T)
    sumX1  +=      11.596 * cos(6.22460372490  + 230.80838830890 *T)
    sumX1  +=      12.157 * cos(5.83781129189  + 374.15181032000 *T)
    sumX1  +=      11.328 * cos(3.65554523534  + 2214.49927011270 *T)
    sumX1  +=      13.310 * cos(1.48760075848  + 1567.48843679789 *T)
    sumX1  +=      13.760 * cos(2.11422894625  + 216.98743121760 *T)
    sumX1  +=      11.244 * cos(1.65064078721  + 777.17412795991 *T)
    sumX1  +=      11.256 * cos(2.33090248405  + 1582.20316576650 *T)
    sumX1  +=      11.849 * cos(3.39312849980  + 305.58998687620 *T)
    sumX1  +=      11.016 * cos(0.89921524640  + 416.05943265400 *T)
    sumX1  +=      10.858 * cos(5.70727580993  + 422.42222012939 *T)
    sumX1  +=      14.447 * cos(2.70740169801  + 333.90116252750 *T)
    sumX1  +=      10.877 * cos(3.35090068282  + 67.42423408300 *T)
    sumX1  +=      10.671 * cos(5.59629136646  + 426.32074349770 *T)
    sumX1  +=      10.673 * cos(5.82038644847  + 221.40721944771 *T)
    sumX1  +=      13.923 * cos(3.33425500011  + 558.24595822940 *T)
    sumX1  +=      12.161 * cos(6.15839226574  + 28.69800548670 *T)
    sumX1  +=      11.528 * cos(4.81930434459  + 432.25863433090 *T)
    sumX1  +=      12.972 * cos(5.99735821999  + 196.86813836510 *T)
    sumX1  +=      10.573 * cos(4.30685728121  + 145.87486135500 *T)
    sumX1  +=      11.572 * cos(1.84813289069  + 5849.12029463110 *T)
    sumX1  +=      12.498 * cos(4.92447718078  + 2008.31372167550 *T)
    sumX1  +=      10.593 * cos(4.14913290174  + 343.46239008310 *T)
    sumX1  +=      12.818 * cos(1.19008859693  + 216.23667169220 *T)
    sumX1  +=      10.230 * cos(1.16924254829  + 199.52826724100 *T)
    sumX1  +=      10.672 * cos(0.61641428977  + 84.58664360640 *T)
    sumX1  +=      10.068 * cos(1.36356239980  + 206.95063078250 *T)
    sumX1  +=      11.596 * cos(1.83788003660  + 1155.60497489050 *T)
    sumX1  +=      10.203 * cos(1.09288636884  + 192.93598510200 *T)
    sumX1  +=       9.849 * cos(0.65561240609  + 767.61290040431 *T)
    sumX1  +=      10.252 * cos(0.21594911177  + 618.04970326970 *T)
    sumX1  +=      10.915 * cos(2.79445278511  + 200.20959079720 *T)
    sumX1  +=       9.693 * cos(1.05765509778  + 99.66756299740 *T)
    sumX1  +=       9.672 * cos(6.28030039661  + 238.81570718700 *T)
    sumX1  +=       9.664 * cos(3.32194333896  + 182.05034353251 *T)
    sumX1  +=       9.860 * cos(2.88110794062  + 572.47305223100 *T)
    sumX1  +=      10.186 * cos(0.41648522301  + 20.85074530300 *T)
    sumX1  +=      10.831 * cos(3.27468814796  + 9779.32999085069 *T)
    sumX1  +=       9.746 * cos(4.95654399301  + 185.23173727020 *T)
    sumX1  +=       9.573 * cos(3.93243598582  + 281.21096448800 *T)
    sumX1  +=      11.944 * cos(2.18663468608  + 62.00760811160 *T)
    sumX1  +=       9.779 * cos(2.45540495440  + 188.41313100791 *T)
    sumX1  +=      10.421 * cos(5.51901551995  + 404.26297286470 *T)
    sumX1  +=      10.874 * cos(0.05198472409  + 153.73916788121 *T)
    sumX1  +=      10.712 * cos(6.08785652303  + 427.36327322130 *T)
    sumX1  +=       9.203 * cos(0.58809567989  + 212.27088853431 *T)
    sumX1  +=       9.203 * cos(4.30153048670  + 214.81493730871 *T)
    sumX1  +=      10.194 * cos(1.27958314866  + 5863.34738863270 *T)
    sumX1  +=       9.371 * cos(3.72043342105  + 227.55755860200 *T)
    sumX1  +=      10.238 * cos(5.85231629543  + 244.56240155850 *T)
    sumX1  +=      10.084 * cos(2.37757705902  + 1265.81129610991 *T)
    sumX1  +=      10.456 * cos(4.58738416662  + 41.40068029210 *T)
    sumX1  +=      10.166 * cos(4.11408240826  + 143.90275367970 *T)
    sumX1  +=      11.381 * cos(1.44255806186  + 156.92056161890 *T)
    sumX1  +=      12.113 * cos(1.46915384562  + 563.87503252191 *T)
    sumX1  +=      10.043 * cos(2.09517340450  + 226.87623504580 *T)
    sumX1  +=       9.235 * cos(2.03101070691  + 159.36824217371 *T)
    sumX1  +=       8.818 * cos(1.31229741323  + 50.96190784770 *T)
    sumX1  +=       9.068 * cos(4.25480546013  + 131.64776735340 *T)
    sumX1  +=      10.089 * cos(4.87714594181  + 260.01333583751 *T)
    sumX1  +=      10.283 * cos(5.60568632571  + 46.71424039951 *T)
    sumX1  +=       9.854 * cos(1.36241255057  + 333.41352756050 *T)
    sumX1  +=       9.008 * cos(1.18554265407  + 637.69342324271 *T)
    sumX1  +=       8.926 * cos(3.80008774802  + 234.14984074100 *T)
    sumX1  +=       8.557 * cos(3.22651762138  + 963.64652045490 *T)
    sumX1  +=      10.257 * cos(5.85233716551  + 64.97655352819 *T)
    sumX1  +=       9.526 * cos(2.70626925959  + 9793.55708485230 *T)
    sumX1  +=       9.414 * cos(4.22693590424  + 67.91186905000 *T)
    sumX1  +=       9.628 * cos(4.95110149071  + 326.93062787860 *T)
    sumX1  +=      11.719 * cos(2.59289161193  + 45.33283355520 *T)
    sumX1  +=       8.304 * cos(2.74677901344  + 547.20025796550 *T)
    sumX1  +=       8.224 * cos(0.02907773024  + 2097.17940189250 *T)
    sumX1  +=       8.279 * cos(6.18251689735  + 241.85408857280 *T)
    sumX1  +=       9.189 * cos(3.25644713958  + 556.76148552110 *T)
    sumX1  +=       8.334 * cos(2.43224998073  + 238.67269483510 *T)
    sumX1  +=       8.030 * cos(1.57230846027  + 245.03548231051 *T)
    sumX1  +=      10.377 * cos(2.43734552323  + 212.64914104420 *T)
    sumX1  +=      10.377 * cos(2.45228064336  + 214.43668479881 *T)
    sumX1  +=       9.061 * cos(2.72986467890  + 206.47755003051 *T)
    sumX1  +=       8.325 * cos(1.31204760835  + 824.49837626530 *T)
    sumX1  +=       7.948 * cos(0.66902213114  + 423.17297965479 *T)
    sumX1  +=      10.520 * cos(2.16910803165  + 93.18466331550 *T)
    sumX1  +=       8.773 * cos(0.84825417283  + 347.64062156210 *T)
    sumX1  +=       8.377 * cos(0.53191864317  + 2523.77759276850 *T)
    sumX1  +=       8.448 * cos(0.59207162429  + 1588.82907780029 *T)
    sumX1  +=       8.685 * cos(5.30112505951  + 220.60827581251 *T)
    sumX1  +=       7.716 * cos(4.75842513626  + 78.20680978850 *T)
    sumX1  +=       7.794 * cos(1.04095159566  + 411.86415483250 *T)
    sumX1  +=       8.324 * cos(6.15163980389  + 646.76701583129 *T)
    sumX1  +=       7.401 * cos(3.67486771061  + 329.48137429739 *T)
    sumX1  +=       8.108 * cos(1.26331676970  + 60.52313540329 *T)
    sumX1  +=       7.383 * cos(0.99737896217  + 259.11956396020 *T)
    sumX1  +=       7.791 * cos(2.57812694404  + 174.18603700630 *T)
    sumX1  +=       7.076 * cos(1.51163028004  + 464.97504399731 *T)
    sumX1  +=       9.308 * cos(5.16234685777  + 273.34665796181 *T)
    sumX1  +=       9.741 * cos(5.52835704075  + 2111.40649589409 *T)
    sumX1  +=       7.187 * cos(3.90837528693  + 864.48589949940 *T)
    sumX1  +=       7.014 * cos(3.23027495127  + 627.61093082531 *T)
    sumX1  +=       6.966 * cos(5.87843482101  + 282.69543719630 *T)
    sumX1  +=       7.308 * cos(0.27091093377  + 942.30587945250 *T)
    sumX1  +=       6.845 * cos(0.51779898110  + 1493.33748554950 *T)
    sumX1  +=       6.779 * cos(1.32496197150  + 357.68948408471 *T)
    sumX1  +=       7.128 * cos(3.87942883373  + 134.82916109110 *T)
    sumX1  +=       7.238 * cos(2.66449317354  + 749.85979004990 *T)
    sumX1  +=       6.975 * cos(2.86950431371  + 267.71758366931 *T)
    sumX1  +=       7.167 * cos(3.95268856947  + 459.08533727390 *T)
    sumX1  +=       7.917 * cos(3.50680613163  + 23.81969071961 *T)
    sumX1  +=       6.648 * cos(1.31476044069  + 1287.15193711231 *T)
    sumX1  +=       6.800 * cos(1.45829035132  + 561.42735196710 *T)
    sumX1  +=       8.501 * cos(2.20054323238  + 224.10097821840 *T)
    sumX1  +=       6.569 * cos(1.46335902098  + 339.53023682000 *T)
    sumX1  +=       8.044 * cos(0.78812832104  + 196.13442518220 *T)
    sumX1  +=       6.518 * cos(6.21803736938  + 638.65663108921 *T)
    sumX1  +=       7.597 * cos(4.56291528763  + 186.71620997851 *T)
    sumX1  +=       8.282 * cos(0.90504128787  + 82.13896305161 *T)
    sumX1  +=       7.088 * cos(2.68605579572  + 570.98857952271 *T)
    sumX1  +=       6.908 * cos(3.16579782135  + 526.75338884040 *T)
    sumX1  +=       6.418 * cos(0.83656417761  + 106.78110999820 *T)
    sumX1  +=       6.362 * cos(5.65057614189  + 1994.08662767389 *T)
    sumX1  +=       6.216 * cos(2.02339054092  + 171.89841514590 *T)
    sumX1  +=       6.708 * cos(2.48079717100  + 508.59414157570 *T)
    sumX1  +=       7.258 * cos(1.20191537197  + 1883.88030645450 *T)
    sumX1  +=       7.633 * cos(3.38390891932  + 295.43805848960 *T)
    sumX1  +=       6.841 * cos(4.70448684705  + 329.96900926440 *T)
    sumX1  +=       5.996 * cos(0.45212858596  + 334.79493440480 *T)
    sumX1  +=       5.926 * cos(5.70772732719  + 2737.07668820650 *T)
    sumX1  +=       6.244 * cos(1.25838362517  + 533.11617631580 *T)
    sumX1  +=       6.129 * cos(6.20652183177  + 215.83053478190 *T)
    sumX1  +=       6.129 * cos(4.96628964200  + 211.25529106110 *T)
    sumX1  +=       5.970 * cos(2.73707139983  + 193.89919294850 *T)
    sumX1  +=       5.955 * cos(4.44190853399  + 128.71245183160 *T)
    sumX1  +=       7.108 * cos(1.56333289690  + 427.59276788490 *T)
    sumX1  +=       5.852 * cos(3.12332782599  + 273.82223084130 *T)
    sumX1  +=       5.748 * cos(2.79144830935  + 30.77567115350 *T)
    sumX1  +=       6.470 * cos(5.25484739290  + 162.33718759030 *T)
    sumX1  +=       5.610 * cos(3.91723617761  + 53.93085326429 *T)
    sumX1  +=       5.595 * cos(3.37029838635  + 2530.89113976930 *T)
    sumX1  +=       6.286 * cos(0.94028611990  + 432.99234751380 *T)
    sumX1  +=       5.905 * cos(2.50849975026  + 856.62159297321 *T)
    sumX1  +=       5.420 * cos(4.38832321197  + 692.83130183700 *T)
    sumX1  +=       7.291 * cos(2.45742729955  + 611.19928082730 *T)
    sumX1  +=       5.368 * cos(1.14023208744  + 113.89465699899 *T)
    sumX1  +=       6.762 * cos(2.27424506141  + 421.21293406700 *T)
    sumX1  +=       6.290 * cos(0.32591135640  + 240.36961586451 *T)
    sumX1  +=       6.519 * cos(4.14721310745  + 1052.02456570489 *T)
    sumX1  +=       5.326 * cos(5.94411846893  + 258.63192899320 *T)
    sumX1  +=       5.918 * cos(3.85830083228  + 422.64922266551 *T)
    sumX1  +=       5.209 * cos(2.18808663209  + 137.27684164590 *T)
    sumX1  +=       5.766 * cos(4.77462627159  + 272.85902299479 *T)
    sumX1  +=       6.291 * cos(4.79715993624  + 423.92124705270 *T)
    sumX1  +=       5.107 * cos(2.89180118566  + 255.18741069710 *T)
    sumX1  +=       6.557 * cos(0.12504746114  + 100.15519796440 *T)
    sumX1  +=       5.736 * cos(2.24104597943  + 252.89978883670 *T)
    sumX1  +=       5.147 * cos(1.21443168596  + 489.09343395460 *T)
    sumX1  +=       5.331 * cos(2.09555876606  + 21.09682351890 *T)
    sumX1  +=       5.351 * cos(4.27893371942  + 342.49918223660 *T)
    sumX1  +=       4.900 * cos(2.79017690768  + 533.86693584120 *T)
    sumX1  +=       6.324 * cos(4.29118767258  + 108.00245814810 *T)
    sumX1  +=       6.120 * cos(1.05577091641  + 472.41865939820 *T)
    sumX1  +=       4.858 * cos(2.21950314401  + 29.44876501211 *T)
    sumX1  +=       6.153 * cos(3.06707334061  + 125.74350641500 *T)
    sumX1  +=       4.774 * cos(1.45632160902  + 394.59867944510 *T)
    sumX1  +=       5.096 * cos(4.95196720627  + 24.13520490470 *T)
    sumX1  +=       4.756 * cos(2.19681205319  + 1898.59503542310 *T)
    sumX1  +=       5.064 * cos(0.32711706674  + 1905.22094745690 *T)
    sumX1  +=       4.696 * cos(5.96778289674  + 408.19512612779 *T)
    sumX1  +=       5.004 * cos(3.21047421953  + 650.43604453600 *T)
    sumX1  +=       4.561 * cos(3.08971667737  + 518.88908231420 *T)
    sumX1  +=       5.108 * cos(1.63505259163  + 1471.99684454710 *T)
    sumX1  +=       4.597 * cos(0.85127842430  + 291.50590522650 *T)
    sumX1  +=       4.502 * cos(3.81171113315  + 167.96626188280 *T)
    sumX1  +=       4.559 * cos(1.72604137750  + 418.76525351221 *T)
    sumX1  +=       6.043 * cos(1.62490154530  + 405.01373239010 *T)
    sumX1  +=       5.079 * cos(5.93549950926  + 264.74863825270 *T)
    sumX1  +=       5.333 * cos(5.77629318787  + 486.64575339981 *T)
    sumX1  +=       5.440 * cos(2.06417795899  + 230.21768747790 *T)
    sumX1  +=       5.845 * cos(3.88668079322  + 551.27542358051 *T)
    sumX1  +=       5.448 * cos(4.69757339283  + 2324.70559133209 *T)
    sumX1  +=       5.131 * cos(3.41222829468  + 835.28095197081 *T)
    sumX1  +=       4.374 * cos(3.79830417186  + 138.27367938720 *T)
    sumX1  +=       4.471 * cos(0.86757359303  + 361.62163734781 *T)
    sumX1  +=       4.410 * cos(1.03727724858  + 1140.89024592190 *T)
    sumX1  +=       4.114 * cos(0.51329977335  + 1486.22393854870 *T)
    sumX1  +=       4.269 * cos(2.60295729376  + 225.32232636830 *T)
    sumX1  +=       4.269 * cos(5.42826152642  + 201.76349947470 *T)
    sumX1  +=       4.076 * cos(6.24795212185  + 33.69643246030 *T)
    sumX1  +=       4.308 * cos(0.45279950263  + 430.28652665560 *T)
    sumX1  +=       4.392 * cos(4.27597448362  + 492.06237937121 *T)
    sumX1  +=       4.365 * cos(0.26068379315  + 2840.16946242510 *T)
    sumX1  +=       4.469 * cos(5.64881997462  + 428.32648106781 *T)
    sumX1  +=       4.332 * cos(3.80105116243  + 32.48714639791 *T)
    sumX1  +=       4.016 * cos(5.36343302652  + 425.35753565121 *T)
    sumX1  +=       4.203 * cos(2.42083154646  + 364.59058276440 *T)
    sumX1  +=       3.902 * cos(2.75257407120  + 397.63706083091 *T)
    sumX1  +=       3.955 * cos(4.26771606040  + 850.25880549781 *T)
    sumX1  +=       3.945 * cos(6.13532982514  + 2413.57127154910 *T)
    sumX1  +=       4.359 * cos(2.96124483199  + 312.96665843540 *T)
    sumX1  +=       5.325 * cos(3.23562442738  + 1169.34443392509 *T)
    sumX1  +=       4.743 * cos(1.50902113607  + 1055.69359440960 *T)
    sumX1  +=       5.334 * cos(5.23788081233  + 2427.79836555070 *T)
    sumX1  +=       3.914 * cos(2.20694290059  + 504.80500066450 *T)
    sumX1  +=       3.740 * cos(4.88716400146  + 658.30035106221 *T)
    sumX1  +=       4.015 * cos(1.57344232669  + 526.26575387340 *T)
    sumX1  +=       3.733 * cos(1.21797191197  + 174.92224231670 *T)
    sumX1  +=       3.719 * cos(0.83792596372  + 292.25666475190 *T)
    sumX1  +=       4.998 * cos(1.43326576234  + 448.93340888730 *T)
    sumX1  +=       3.941 * cos(1.19970132537  + 434.00373947010 *T)
    sumX1  +=       3.629 * cos(6.16327394850  + 1692.40948698591 *T)
    sumX1  +=       4.488 * cos(2.13061381956  + 410.16723380310 *T)
    sumX1  +=       4.487 * cos(1.97829463327  + 832.09955823310 *T)
    sumX1  +=       4.724 * cos(0.88455549663  + 205.90810105891 *T)
    sumX1  +=       3.566 * cos(6.23327093123  + 215.23983395090 *T)
    sumX1  +=       3.566 * cos(1.79794788895  + 211.84599189210 *T)
    sumX1  +=       4.241 * cos(3.81199974644  + 370.21965705690 *T)
    sumX1  +=       3.740 * cos(3.21504654325  + 471.93102443120 *T)
    sumX1  +=       3.762 * cos(3.12578966830  + 320.08020543620 *T)
    sumX1  +=       4.157 * cos(0.60528202242  + 2119.00767786191 *T)
    sumX1  +=       3.518 * cos(0.26416908476  + 33.38091827520 *T)
    sumX1  +=       3.584 * cos(5.79171536939  + 77.99436146740 *T)
    sumX1  +=       3.636 * cos(4.51005837077  + 406.23508054000 *T)
    sumX1  +=       3.529 * cos(0.66209885167  + 732.45130231059 *T)
    sumX1  +=       3.573 * cos(0.76302606907  + 440.58146739410 *T)
    sumX1  +=       4.086 * cos(0.94506625757  + 2200.27217611110 *T)
    sumX1  +=       4.527 * cos(0.86326378242  + 221.17772478411 *T)
    sumX1  +=       3.974 * cos(2.85379501585  + 16.91859203990 *T)
    sumX1  +=       3.913 * cos(4.12538485725  + 189.96703968541 *T)
    sumX1  +=       3.397 * cos(4.45063317042  + 245.29860686890 *T)
    sumX1  +=       3.967 * cos(5.18194728645  + 120.74507944141 *T)
    sumX1  +=       3.785 * cos(5.04003954266  + 215.18744432420 *T)
    sumX1  +=       3.785 * cos(6.13277193111  + 211.89838151880 *T)
    sumX1  +=       3.713 * cos(1.92562812532  + 203.50860461390 *T)
    sumX1  +=       3.713 * cos(6.10559069486  + 223.57722122910 *T)
    sumX1  +=       3.167 * cos(1.49545356113  + 224.84924561630 *T)
    sumX1  +=       3.167 * cos(0.25257995187  + 202.23658022670 *T)
    sumX1  +=       4.027 * cos(2.28727319435  + 38.37685312130 *T)
    sumX1  +=       3.454 * cos(0.92726421089  + 542.78046973539 *T)
    sumX1  +=       3.700 * cos(4.89575985300  + 170.25388374320 *T)
    sumX1  +=       3.257 * cos(3.14955384513  + 305.85311143459 *T)
    sumX1  +=       3.858 * cos(4.45826332823  + 135.57992061650 *T)
    sumX1  +=       3.174 * cos(3.30524903499  + 340.52707456129 *T)
    sumX1  +=       3.287 * cos(1.32591701514  + 10.05112325500 *T)
    sumX1  +=       4.371 * cos(1.14794032361  + 1045.39865367110 *T)
    sumX1  +=       3.220 * cos(0.30781065642  + 629.84616305900 *T)
    sumX1  +=       3.146 * cos(6.12951187669  + 553.82616999929 *T)
    sumX1  +=       3.544 * cos(2.54076567072  + 1066.25165970650 *T)
    sumX1  +=       4.149 * cos(6.00042687336  + 136.78920667889 *T)
    sumX1  +=       4.004 * cos(2.56002623387  + 494.51005992600 *T)
    sumX1  +=       3.653 * cos(1.07325080538  + 466.19888427470 *T)
    sumX1  +=       3.162 * cos(0.16615343726  + 2221.61281711350 *T)
    sumX1  +=       3.678 * cos(1.45731309163  + 418.24398865040 *T)
    sumX1  +=       3.135 * cos(1.98132169566  + 409.43352062020 *T)
    sumX1  +=       3.067 * cos(2.46882469796  + 441.81986188650 *T)
    sumX1  +=       2.970 * cos(5.59501410890  + 731.70054278519 *T)
    sumX1  +=       3.379 * cos(0.76892622524  + 237.11878615761 *T)
    sumX1  +=       3.100 * cos(5.70739500607  + 407.19828838650 *T)
    sumX1  +=       3.007 * cos(4.39247880899  + 92.55401599659 *T)
    sumX1  +=       2.898 * cos(2.07796431043  + 1080.96638867510 *T)
    sumX1  +=       3.170 * cos(1.03728235557  + 105.76971804189 *T)
    sumX1  +=       2.943 * cos(4.30277427326  + 256.08118257441 *T)
    sumX1  +=       3.935 * cos(3.71027206200  + 426.09124883410 *T)
    sumX1  +=       3.326 * cos(4.80847422754  + 73.54094334250 *T)
    sumX1  +=       2.837 * cos(5.38571733028  + 467.99887116810 *T)
    sumX1  +=       3.336 * cos(0.12937996666  + 128.60938596760 *T)
    sumX1  +=       2.896 * cos(0.58823826825  + 1802.61580820530 *T)
    sumX1  +=       2.904 * cos(0.36123693139  + 6062.41939006910 *T)
    sumX1  +=       2.774 * cos(3.05473381967  + 2847.28300942589 *T)
    sumX1  +=       2.813 * cos(2.69799483611  + 84.68970947040 *T)
    sumX1  +=       3.318 * cos(1.86340648711  + 254.69977573010 *T)
    sumX1  +=       2.850 * cos(5.39053439064  + 220.76833450690 *T)
    sumX1  +=       3.010 * cos(3.47566151707  + 451.22103074770 *T)
    sumX1  +=       3.559 * cos(1.86339595837  + 519.15220687259 *T)
    sumX1  +=       3.563 * cos(2.74969732330  + 486.15811843280 *T)
    sumX1  +=       3.360 * cos(2.57348311811  + 388.70897272171 *T)
    sumX1  +=       2.703 * cos(3.67104853266  + 345.08987514331 *T)
    sumX1  +=       2.882 * cos(0.29032835089  + 511.28790034640 *T)
    sumX1  +=       3.267 * cos(5.09418923422  + 543.26810470240 *T)
    sumX1  +=       3.197 * cos(5.35873741110  + 54.89406111079 *T)
    sumX1  +=       2.907 * cos(0.41906230884  + 630.57987624190 *T)
    sumX1  +=       2.669 * cos(2.23461202741  + 815.30716359770 *T)
    sumX1  +=       2.869 * cos(0.07993801900  + 57.86300652739 *T)
    sumX1  +=       2.764 * cos(2.67313412640  + 5635.82119919310 *T)
    sumX1  +=       2.610 * cos(5.54553510080  + 857.37235249860 *T)
    sumX1  +=       2.782 * cos(0.10347092925  + 639.17789595100 *T)
    sumX1  +=       2.826 * cos(1.51658586855  + 171.00464326860 *T)
    sumX1  +=       2.568 * cos(5.46900605123  + 3053.46855786309 *T)
    sumX1  +=       2.628 * cos(1.75719236208  + 9992.62908628869 *T)
    sumX1  +=       3.266 * cos(1.28369403361  + 721.40560204669 *T)
    sumX1  +=       2.545 * cos(4.09883259215  + 455.29619636271 *T)
    sumX1  +=       2.510 * cos(3.89386878011  + 641.10431164400 *T)
    sumX1  +=       2.554 * cos(5.33920988436  + 738.81408978600 *T)
    sumX1  +=       2.920 * cos(2.36151426705  + 92.29089143820 *T)
    sumX1  +=       2.992 * cos(4.38095651980  + 462.26673101160 *T)
    sumX1  +=       2.474 * cos(1.99548298969  + 2214.98690507971 *T)
    sumX1  +=       2.474 * cos(1.50083275203  + 419.77664546851 *T)
    sumX1  +=       2.645 * cos(3.34653267179  + 1685.29593998510 *T)
    sumX1  +=       2.597 * cos(0.35423361548  + 426.73013377490 *T)
    sumX1  +=       2.936 * cos(3.12370531281  + 468.48650613510 *T)
    sumX1  +=       2.535 * cos(0.19690988756  + 739.56484931140 *T)
    sumX1  +=       2.575 * cos(3.67638389113  + 77.71917482150 *T)
    sumX1  +=       2.960 * cos(1.43272530301  + 403.26613512341 *T)
    sumX1  +=       2.526 * cos(2.28575142320  + 198.07742442750 *T)
    sumX1  +=       2.819 * cos(6.27705191381  + 256.83194209980 *T)
    sumX1  +=       2.444 * cos(3.25740106622  + 420.69166920520 *T)
    sumX1  +=       3.165 * cos(5.89068478002  + 200.80029162821 *T)
    sumX1  +=       3.165 * cos(2.14053404016  + 226.28553421481 *T)
    sumX1  +=       2.684 * cos(5.47533039513  + 540.98048284201 *T)
    sumX1  +=       2.447 * cos(3.11720914436  + 207.18012544610 *T)
    sumX1  +=       2.447 * cos(4.91400967582  + 219.90570039690 *T)
    sumX1  +=       2.485 * cos(6.02271689683  + 2729.96314120569 *T)
    sumX1  +=       2.810 * cos(1.60047864591  + 297.98880490840 *T)
    sumX1  +=       2.343 * cos(1.03676860662  + 402.46298597130 *T)
    sumX1  +=       3.270 * cos(1.82014094673  + 714.92270236481 *T)
    sumX1  +=       2.983 * cos(2.84636794929  + 1485.73630358170 *T)
    sumX1  +=       2.301 * cos(2.17881414725  + 455.15318401080 *T)
    sumX1  +=       2.295 * cos(2.14340790418  + 503.10807963510 *T)
    sumX1  +=       2.789 * cos(0.36073274237  + 269.67762925710 *T)
    sumX1  +=       2.298 * cos(4.07536270090  + 131.16013238639 *T)
    sumX1  +=       2.626 * cos(2.72737754780  + 112.89781925770 *T)
    sumX1  +=       2.273 * cos(5.75136970440  + 283.90472325870 *T)
    sumX1  +=       2.621 * cos(1.82368596180  + 172.48911597691 *T)
    sumX1  +=       2.285 * cos(0.20921743314  + 525.51499434800 *T)
    sumX1  +=       2.204 * cos(0.87003863856  + 593.67068088151 *T)
    sumX1  +=       2.636 * cos(0.30028112725  + 649.70233135310 *T)
    sumX1  +=       2.222 * cos(6.06903250865  + 478.04773369070 *T)
    sumX1  +=       2.308 * cos(0.82575212401  + 1457.28211557850 *T)
    sumX1  +=       2.203 * cos(2.42859720355  + 1596.43025976811 *T)
    sumX1  +=       2.182 * cos(1.34997153566  + 397.14942586389 *T)
    sumX1  +=       2.502 * cos(0.74653280150  + 9.31741007210 *T)
    sumX1  +=       2.138 * cos(5.57703162616  + 2310.47849733050 *T)
    sumX1  +=       2.142 * cos(4.51014414360  + 211.57080524620 *T)
    sumX1  +=       2.974 * cos(2.23240151481  + 927.59115048390 *T)
    sumX1  +=       2.136 * cos(1.09576811823  + 186.45557754760 *T)
    sumX1  +=       2.091 * cos(2.78262893204  + 1258.69774910910 *T)
    sumX1  +=       2.094 * cos(4.08396200744  + 5.17280848790 *T)
    sumX1  +=       2.460 * cos(2.08164520313  + 5650.04829319470 *T)
    sumX1  +=       2.461 * cos(2.31536038396  + 1382.64352936309 *T)
    sumX1  +=       2.355 * cos(5.56042831680  + 184.33796539290 *T)
    sumX1  +=       2.153 * cos(5.44863811686  + 233.18663289450 *T)
    sumX1  +=       2.597 * cos(5.04726142112  + 426.95388294410 *T)
    sumX1  +=       2.604 * cos(5.20426552840  + 980.91199584231 *T)
    sumX1  +=       2.201 * cos(2.69689845951  + 434.91876320681 *T)
    sumX1  +=       2.162 * cos(4.51403145252  + 1383.13116433011 *T)
    sumX1  +=       2.404 * cos(3.51511061541  + 9580.25798941430 *T)
    sumX1  +=       2.260 * cos(4.11127975780  + 42.78208713641 *T)
    sumX1  +=       2.151 * cos(6.08523383456  + 487.60896124631 *T)
    sumX1  +=       2.123 * cos(1.99177130106  + 149.31937965110 *T)
    sumX1  +=       2.556 * cos(4.74909112241  + 142.20583265029 *T)
    sumX1  +=       1.998 * cos(5.68001469591  + 724.58699578440 *T)
    sumX1  +=       1.973 * cos(2.53052550367  + 481.01667910731 *T)
    sumX1  +=       1.973 * cos(6.13759453722  + 348.87901605450 *T)
    sumX1  +=       1.957 * cos(0.51054293730  + 28.55499313480 *T)
    sumX1  +=       1.957 * cos(3.65213559089  + 28.06735816780 *T)
    sumX1  +=       1.991 * cos(5.56250823493  + 298.47643987540 *T)
    sumX1  +=       1.940 * cos(5.03410941079  + 141.46962733991 *T)
    sumX1  +=       2.507 * cos(5.65120070160  + 6489.50521591210 *T)
    sumX1  +=       2.352 * cos(4.19836125385  + 479.04457143200 *T)
    sumX1  +=       2.048 * cos(1.46373921871  + 1788.38871420371 *T)
    sumX1  +=       2.647 * cos(4.98237868701  + 2744.19023520730 *T)
    sumX1  +=       1.908 * cos(3.37818588512  + 1368.90407032850 *T)
    sumX1  +=       1.900 * cos(6.21516760726  + 271.64973693240 *T)
    sumX1  +=       2.178 * cos(3.04872277319  + 254.59670986610 *T)
    sumX1  +=       1.894 * cos(0.38526664490  + 1048.84317196720 *T)
    sumX1  +=       1.892 * cos(4.53251590605  + 49.26498681829 *T)
    sumX1  +=       1.888 * cos(0.97996534938  + 642.58878435231 *T)
    sumX1  +=       2.182 * cos(4.42761937882  + 427.05445668060 *T)
    sumX1  +=       1.810 * cos(6.24932174424  + 518.40144734719 *T)
    sumX1  +=       1.925 * cos(6.21686018138  + 3156.56133208170 *T)
    sumX1  +=       2.022 * cos(5.64661364195  + 4.42204896250 *T)
    sumX1  +=       1.793 * cos(3.43045021055  + 419.68027724890 *T)
    sumX1  +=       2.336 * cos(0.79346337330  + 10419.71491213170 *T)
    sumX1  +=       1.992 * cos(2.92479817252  + 16.43095707290 *T)
    sumX1  +=       1.790 * cos(4.60289486851  + 71.84402231310 *T)
    sumX1  +=       1.779 * cos(2.30415626764  + 702.39252939260 *T)
    sumX1  +=       1.796 * cos(3.79385866271  + 240.63024829540 *T)
    sumX1  +=       1.974 * cos(0.40863452127  + 640.18928790730 *T)
    sumX1  +=       1.881 * cos(2.58502310856  + 200.52510498229 *T)
    sumX1  +=       1.773 * cos(0.91523739118  + 990.47322339791 *T)
    sumX1  +=       2.183 * cos(0.33182574262  + 411.37651986550 *T)
    sumX1  +=       2.250 * cos(0.80555474376  + 6077.13411903770 *T)
    sumX1  +=       2.182 * cos(0.97387565779  + 426.62956003840 *T)
    sumX1  +=       1.835 * cos(1.94831932983  + 163.14033674240 *T)
    sumX1  +=       2.428 * cos(4.54827834066  + 2641.09746098870 *T)
    sumX1  +=       1.996 * cos(1.88608745432  + 398.38782035631 *T)
    sumX1  +=       1.813 * cos(1.03980179180  + 25.51661174900 *T)
    sumX1  +=       1.980 * cos(6.08475811682  + 206.54124050531 *T)
    sumX1  +=       1.940 * cos(0.04753375650  + 774.72644740510 *T)
    sumX1  +=       1.910 * cos(4.61558010626  + 175.67050971460 *T)
    sumX1  +=       2.379 * cos(3.57535524190  + 594.89452115890 *T)
    sumX1  +=       2.068 * cos(4.09042241645  + 183.85033042589 *T)
    sumX1  +=       1.733 * cos(3.39710168703  + 705.36147480921 *T)
    sumX1  +=       1.824 * cos(0.43158734549  + 420.24972622050 *T)
    sumX1  +=       1.804 * cos(4.09117978828  + 9566.03089541269 *T)
    sumX1  +=       2.037 * cos(3.95444170699  + 481.97988695381 *T)
    sumX1  +=       1.792 * cos(1.26659627713  + 842.65762352999 *T)
    sumX1  +=       1.681 * cos(2.53871998999  + 1575.08961876571 *T)
    sumX1  +=       1.732 * cos(1.55363228273  + 692.34366686999 *T)
    sumX1  +=       1.698 * cos(4.90525724103  + 355.50492808829 *T)
    sumX1  +=       1.828 * cos(4.98959403756  + 484.94883237040 *T)
    sumX1  +=       1.713 * cos(1.30182253021  + 457.86149699650 *T)
    sumX1  +=       1.808 * cos(2.75209469281  + 622.24498109120 *T)
    sumX1  +=       1.911 * cos(3.61821751946  + 5165.60650620771 *T)
    sumX1  +=       1.704 * cos(6.23318457567  + 2538.00468677009 *T)
    sumX1  +=       1.688 * cos(6.03910121927  + 267.22994870229 *T)
    sumX1  +=       1.665 * cos(5.37698735936  + 1127.50624756031 *T)
    sumX1  +=       2.097 * cos(2.22788409731  + 10007.34381525730 *T)
    sumX1  +=       1.691 * cos(2.07854652969  + 789.70430093210 *T)
    sumX1  +=       1.699 * cos(0.77916428304  + 760.49935340350 *T)
    sumX1  +=       1.783 * cos(3.13574966388  + 2001.68780964171 *T)
    sumX1  +=       1.796 * cos(4.27766403083  + 1194.21109128031 *T)
    sumX1  +=       1.619 * cos(3.01440438125  + 189.47940471839 *T)
    sumX1  +=       1.652 * cos(2.95148116143  + 152.77596003471 *T)
    sumX1  +=       1.707 * cos(5.77147938366  + 3046.35501086230 *T)
    sumX1  +=       1.924 * cos(0.43738095399  + 380.37158544351 *T)
    sumX1  +=       1.960 * cos(3.34806236291  + 144.17794032560 *T)
    sumX1  +=       1.599 * cos(5.99560668065  + 2008.80135664250 *T)
    sumX1  +=       1.804 * cos(0.11881932366  + 353.54488250050 *T)
    sumX1  +=       1.586 * cos(0.20398228990  + 6069.53293706989 *T)
    sumX1  +=       1.644 * cos(1.16840005853  + 1112.31594571220 *T)
    sumX1  +=       1.572 * cos(4.52460985815  + 91.54262404029 *T)
    sumX1  +=       2.123 * cos(2.62748087622  + 1802.12817323829 *T)
    sumX1  +=       1.948 * cos(4.60996708702  + 268.68079151581 *T)
    sumX1  +=       1.766 * cos(4.90567167770  + 3333.74269718250 *T)
    sumX1  +=       1.565 * cos(0.79981986869  + 97.91996573070 *T)
    sumX1  +=       1.602 * cos(4.73221068440  + 490.09027169589 *T)
    sumX1  +=       1.546 * cos(2.42254478872  + 258.78949556011 *T)
    sumX1  +=       1.724 * cos(3.10736186704  + 26081.03341205689 *T)
    sumX1  +=       1.531 * cos(4.39157176653  + 5216.33655531789 *T)
    sumX1  +=       1.619 * cos(1.87855507272  + 784.28767496071 *T)
    sumX1  +=       1.871 * cos(0.40711466948  + 206.31749133610 *T)
    sumX1  +=       1.734 * cos(0.20470303542  + 304.84171947829 *T)
    sumX1  +=       1.732 * cos(6.00632565407  + 1332.72858815100 *T)
    sumX1  +=       1.517 * cos(5.25218074419  + 1048.09241244180 *T)
    sumX1  +=       1.563 * cos(3.46526834018  + 555.79827767460 *T)
    sumX1  +=       1.520 * cos(3.35489835249  + 1101.27024544830 *T)
    sumX1  +=       1.687 * cos(4.36557051418  + 3347.96979118411 *T)
    sumX1  +=       1.488 * cos(1.67143354747  + 9999.74263328950 *T)
    sumX1  +=       1.587 * cos(0.21300456338  + 372.77040347569 *T)
    sumX1  +=       1.662 * cos(2.56954470179  + 26095.26050605850 *T)
    sumX1  +=       1.724 * cos(1.33166239973  + 1037.79747170330 *T)
    sumX1  +=       1.477 * cos(4.36052585434  + 208.38941150850 *T)
    sumX1  +=       1.657 * cos(1.83865415601  + 552.82933225800 *T)
    sumX1  +=       1.520 * cos(2.73911257161  + 2318.07967929830 *T)
    sumX1  +=       1.443 * cos(1.98940467508  + 910.06255053810 *T)
    sumX1  +=       1.446 * cos(1.10349996175  + 204.94489321240 *T)
    sumX1  +=       1.446 * cos(0.64453355125  + 222.14093263061 *T)
    sumX1  +=       1.850 * cos(3.79242867450  + 699.94484883781 *T)
    sumX1  +=       1.439 * cos(4.66881004427  + 393.70490756781 *T)
    sumX1  +=       1.452 * cos(6.27116384237  + 1130.47519297691 *T)
    sumX1  +=       1.874 * cos(2.34112858987  + 770.06058095910 *T)
    sumX1  +=       1.782 * cos(1.54447498730  + 835.54407652920 *T)
    sumX1  +=       1.639 * cos(5.07308302027  + 427.10264079041 *T)
    sumX1  +=       1.565 * cos(0.60414645629  + 392.90175841570 *T)
    sumX1  +=       1.395 * cos(5.81501189133  + 9146.54625153749 *T)
    sumX1  +=       1.516 * cos(3.58009173951  + 397.90018538930 *T)
    sumX1  +=       1.684 * cos(6.24638766723  + 1479.11039154791 *T)
    sumX1  +=       1.659 * cos(1.59176755278  + 721.89323701370 *T)
    sumX1  +=       1.514 * cos(5.91645211455  + 319.06881347989 *T)
    sumX1  +=       1.397 * cos(2.98224235085  + 3163.67487908250 *T)
    sumX1  +=       1.747 * cos(2.27656402390  + 37.88921815429 *T)
    sumX1  +=       1.492 * cos(3.04531836360  + 263.94548910060 *T)
    sumX1  +=       1.354 * cos(4.57495887478  + 140.24578706250 *T)
    sumX1  +=       1.506 * cos(0.27404631561  + 251.41531612840 *T)
    sumX1  +=       1.365 * cos(4.69515666017  + 694.31577454531 *T)
    sumX1  +=       1.352 * cos(4.48933959673  + 60.79832204921 *T)
    sumX1  +=       1.352 * cos(4.48933959673  + 60.31068708220 *T)
    sumX1  +=       1.531 * cos(0.64877553761  + 907.09360512150 *T)
    sumX1  +=       1.783 * cos(0.81801801901  + 185.08872491831 *T)
    sumX1  +=       1.550 * cos(4.05055802538  + 170.51700830159 *T)
    sumX1  +=       1.524 * cos(0.15234864850  + 921.32069912310 *T)
    sumX1  +=       1.541 * cos(2.73990779933  + 31.73638687251 *T)
    sumX1  +=       1.578 * cos(5.98549196472  + 190.64836324160 *T)
    sumX1  +=       1.578 * cos(5.18731950905  + 236.43746260140 *T)
    sumX1  +=       1.732 * cos(0.64736011209  + 1126.54303971381 *T)
    sumX1  +=       1.540 * cos(0.36379616132  + 2833.05591542430 *T)
    sumX1  +=       1.323 * cos(4.87522768351  + 460.78225830330 *T)
    sumX1  +=       1.312 * cos(5.10594513264  + 448.98829064149 *T)
    sumX1  +=       1.724 * cos(5.39037296555  + 78.46993434690 *T)
    sumX1  +=       1.472 * cos(4.26339575699  + 34.18406742730 *T)
    sumX1  +=       1.296 * cos(5.62129859108  + 2325.19322629910 *T)
    sumX1  +=       1.309 * cos(5.45522580580  + 349.09146437560 *T)
    sumX1  +=       1.818 * cos(4.16411482533  + 100.40127618030 *T)
    sumX1  +=       1.290 * cos(2.36413128544  + 1176.94561589290 *T)
    sumX1  +=       1.334 * cos(2.44221382511  + 282.90788551740 *T)
    sumX1  +=       1.287 * cos(0.26504708130  + 286.84003878050 *T)
    sumX1  +=       1.392 * cos(4.87830155415  + 69.12115511239 *T)
    sumX1  +=       1.515 * cos(3.62072467848  + 12.28635548869 *T)
    sumX1  +=       1.247 * cos(4.64664167702  + 187.67941782501 *T)
    sumX1  +=       1.247 * cos(0.24298448957  + 239.40640801801 *T)
    sumX1  +=       1.246 * cos(1.79539178116  + 2487.17222746430 *T)
    sumX1  +=       1.511 * cos(1.66615367049  + 213.00709384469 *T)
    sumX1  +=       1.479 * cos(0.68803218820  + 2516.66404576769 *T)
    sumX1  +=       1.233 * cos(1.65114888447  + 2531.37877473630 *T)
    sumX1  +=       1.632 * cos(2.90714941935  + 906.13039727500 *T)
    sumX1  +=       1.387 * cos(0.55195474062  + 401.56921409400 *T)
    sumX1  +=       1.352 * cos(0.63301755055  + 351.57277482520 *T)
    sumX1  +=       1.313 * cos(0.13852930132  + 350.08830211689 *T)
    sumX1  +=       1.236 * cos(3.39882375572  + 11926.01059618529 *T)
    sumX1  +=       1.227 * cos(5.24375477429  + 4127.56521702090 *T)
    sumX1  +=       1.304 * cos(0.28979853169  + 203.49405039889 *T)
    sumX1  +=       1.478 * cos(4.80872386721  + 264.26100328570 *T)
    sumX1  +=       1.254 * cos(6.19414707054  + 827.67977000299 *T)
    sumX1  +=       1.230 * cos(2.44616325751  + 51535.66517935089 *T)
    sumX1  +=       1.385 * cos(4.26128512958  + 41.29761442810 *T)
    sumX1  +=       1.364 * cos(4.18584755012  + 65.46418849521 *T)
    sumX1  +=       1.395 * cos(2.64686968582  + 629.35852809199 *T)
    sumX1  +=       1.284 * cos(5.37115714123  + 168.29633028290 *T)
    sumX1  +=       1.435 * cos(5.84382021042  + 617.56206830270 *T)
    sumX1  +=       1.411 * cos(5.88687181276  + 188.27011865600 *T)
    sumX1  +=       1.191 * cos(0.70400726689  + 606.51636803880 *T)
    sumX1  +=       1.193 * cos(1.77967286420  + 274.30986580831 *T)
    sumX1  +=       1.197 * cos(2.05343251093  + 1062.80714141041 *T)
    sumX1  +=       1.356 * cos(1.22316379022  + 50.64639366260 *T)
    sumX1  +=       1.356 * cos(4.36475644381  + 50.15875869559 *T)
    sumX1  +=       1.276 * cos(0.30928416823  + 2090.06585489169 *T)
    sumX1  +=       1.195 * cos(2.85953206146  + 681.29796660610 *T)
    sumX1  +=       1.603 * cos(1.64227556935  + 1560.37488979710 *T)
    sumX1  +=       1.191 * cos(3.68583487315  + 9594.99527173989 *T)
    sumX1  +=       1.168 * cos(1.08249837425  + 74.39476873190 *T)
    sumX1  +=       1.165 * cos(4.67802137668  + 674.18441960530 *T)
    sumX1  +=       1.242 * cos(5.63232378165  + 1902.52718868620 *T)
    sumX1  +=       1.330 * cos(0.04908988755  + 200.27902676641 *T)
    sumX1  +=       1.330 * cos(1.69894362545  + 226.80679907661 *T)
    sumX1  +=       1.154 * cos(1.32466818176  + 2847.77064439291 *T)
    sumX1  +=       1.150 * cos(5.63834183266  + 17.65230522280 *T)
    sumX1  +=       1.152 * cos(1.92670632990  + 740.05248427840 *T)
    sumX1  +=       1.247 * cos(5.13783491688  + 348.39138108750 *T)
    sumX1  +=       1.387 * cos(5.73885801869  + 527.48710202330 *T)
    sumX1  +=       1.250 * cos(1.98100336387  + 184.60108995130 *T)
    sumX1  +=       1.328 * cos(4.49802504301  + 2957.48933064529 *T)
    sumX1  +=       1.587 * cos(1.96187033145  + 1243.98302014049 *T)
    sumX1  +=       1.176 * cos(1.80593057083  + 747.16603127920 *T)
    sumX1  +=       1.223 * cos(0.59493067511  + 1773.67398523510 *T)
    sumX1  +=       1.161 * cos(0.01398952175  + 2435.39954751850 *T)
    sumX1  +=       1.570 * cos(4.69885872763  + 3060.58210486390 *T)
    sumX1  +=       1.303 * cos(3.48272421278  + 393.21727260080 *T)
    sumX1  +=       1.126 * cos(0.75321492124  + 305.10235190919 *T)
    sumX1  +=       1.534 * cos(1.87537098734  + 81.99595069970 *T)
    sumX1  +=       1.117 * cos(4.61546207470  + 526.00262931501 *T)
    sumX1  +=       1.307 * cos(5.84787741027  + 72.09010052900 *T)
    sumX1  +=       1.114 * cos(0.00680453000  + 25234.46294233869 *T)
    sumX1  +=       1.118 * cos(1.34292799669  + 511.77553531340 *T)
    sumX1  +=       1.242 * cos(2.97258589350  + 145.38722638800 *T)
    sumX1  +=       1.397 * cos(4.30533366512  + 336.59492129820 *T)
    sumX1  +=       1.401 * cos(2.28518305813  + 2118.52004289490 *T)
    sumX1  +=       1.139 * cos(5.01671933650  + 1055.20595944260 *T)
    sumX1  +=       1.096 * cos(3.93031837723  + 739.30172475301 *T)
    sumX1  +=       1.092 * cos(1.72033976580  + 234.39591895690 *T)
    sumX1  +=       1.197 * cos(0.70790588748  + 306.34074640161 *T)
    sumX1  +=       1.126 * cos(5.90717690029  + 1144.07163965960 *T)
    sumX1  +=       1.351 * cos(3.03970702934  + 179.60266297770 *T)
    sumX1  +=       1.165 * cos(0.13814938960  + 746.67839631219 *T)
    sumX1  +=       1.167 * cos(4.32162909656  + 634.72447782610 *T)
    sumX1  +=       1.121 * cos(3.70393028752  + 366.28750379380 *T)
    sumX1  +=       1.051 * cos(2.77078069512  + 383.81610373959 *T)
    sumX1  +=       1.053 * cos(3.41928600573  + 2107.96197759801 *T)
    sumX1  +=       1.051 * cos(2.99910636875  + 10000.23026825651 *T)
    sumX1  +=       1.288 * cos(1.85993541573  + 247.48316286530 *T)
    sumX1  +=       1.045 * cos(4.77749523511  + 1680.87615175499 *T)
    sumX1  +=       1.070 * cos(5.83029249500  + 3472.95320173830 *T)
    sumX1  +=       1.043 * cos(5.10256961747  + 151.50393564750 *T)
    sumX1  +=       1.033 * cos(1.59445268426  + 6070.02057203690 *T)
    sumX1  +=       1.282 * cos(3.59113162781  + 141.94270809190 *T)
    sumX1  +=       1.167 * cos(5.76598251731  + 2920.88396534110 *T)
    sumX1  +=       1.100 * cos(2.52155794298  + 2634.47154895491 *T)
    sumX1  +=       1.035 * cos(5.30771361050  + 3369.86042751970 *T)
    sumX1  +=       1.091 * cos(0.54826165686  + 1314.56934088630 *T)
    sumX1  +=       1.274 * cos(3.38359403787  + 691.34682912870 *T)
    sumX1  +=       1.079 * cos(4.20190952106  + 1044.64789414570 *T)
    sumX1  +=       1.085 * cos(2.22740147584  + 141.20899490900 *T)
    sumX1  +=       1.245 * cos(5.83769374356  + 179.91817716280 *T)
    sumX1  +=       1.146 * cos(0.33774172888  + 220.54458533771 *T)
    sumX1  +=       1.196 * cos(3.17156910173  + 1460.20038475780 *T)
    sumX1  +=       1.136 * cos(2.05466340621  + 480.00777927849 *T)
    sumX1  +=       1.119 * cos(5.41354135574  + 1534.98198332510 *T)
    sumX1  +=       1.056 * cos(3.62871723489  + 437.40007365639 *T)
    sumX1  +=       1.032 * cos(2.32675902580  + 184.97110483931 *T)
    sumX1  +=       1.032 * cos(2.56286714079  + 242.11472100371 *T)
    sumX1  +=       1.039 * cos(2.32832376555  + 1791.08247297440 *T)
    sumX1  +=       1.061 * cos(2.50615202413  + 938.63685074779 *T)
    sumX1  +=       1.123 * cos(3.96404102873  + 25668.17468021549 *T)
    sumX1  +=       1.103 * cos(5.86207947745  + 210.60759739969 *T)
    sumX1  +=       1.015 * cos(4.05475486800  + 252.41215386970 *T)
    sumX1  +=       0.979 * cos(5.22245385112  + 610.44852130189 *T)
    sumX1  +=       0.979 * cos(2.08086119753  + 610.93615626891 *T)
    sumX1  +=       1.043 * cos(3.45408554292  + 17.91542978119 *T)
    sumX1  +=       1.128 * cos(3.92351069480  + 398.53083270820 *T)
    sumX1  +=       1.035 * cos(3.02168366647  + 235.68919520349 *T)
    sumX1  +=       1.072 * cos(4.53150424448  + 746.15463932291 *T)
    sumX1  +=       1.131 * cos(3.37345520465  + 5481.99837586430 *T)
    sumX1  +=       0.986 * cos(5.28602240316  + 192.44835013500 *T)
    sumX1  +=       1.287 * cos(1.95273113714  + 314.18800658530 *T)
    sumX1  +=       0.973 * cos(3.85435039011  + 561.69047652550 *T)
    sumX1  +=       1.298 * cos(1.08374685926  + 614.38067456499 *T)
    sumX1  +=       1.151 * cos(1.20704230584  + 905.64276230799 *T)
    sumX1  +=       1.279 * cos(5.88520431436  + 322.36782729660 *T)
    sumX1  +=       0.963 * cos(3.50426087646  + 782.59075393131 *T)
    sumX1  +=       1.064 * cos(5.75961632886  + 1372.34858862460 *T)
    sumX1  +=       0.973 * cos(3.52289496745  + 24.62283987170 *T)
    sumX1  +=       1.096 * cos(5.06936627625  + 731.92754532131 *T)
    sumX1  +=       1.080 * cos(6.27980868199  + 2906.65687133950 *T)
    sumX1  +=       1.200 * cos(2.31102641105  + 195.64679021520 *T)
    sumX1  +=       0.993 * cos(1.95529510745  + 6062.90702503610 *T)
    sumX1  +=       1.039 * cos(1.07426067099  + 1159.04949318660 *T)
    sumX1  +=       0.952 * cos(2.30173432343  + 239.88198089750 *T)
    sumX1  +=       1.181 * cos(0.98375947358  + 1468.32781584240 *T)
    sumX1  +=       1.137 * cos(2.73354004808  + 2.69149803831 *T)
    sumX1  +=       0.941 * cos(1.23572643410  + 124.18959773749 *T)
    sumX1  +=       0.942 * cos(3.50129097588  + 115.11600514891 *T)
    sumX1  +=       1.020 * cos(1.79568687186  + 196.38050339810 *T)
    sumX1  +=       1.065 * cos(1.53835988752  + 494.23487328010 *T)
    sumX1  +=       0.945 * cos(4.39156950811  + 121.00820399980 *T)
    sumX1  +=       0.941 * cos(3.67244160492  + 120.01136625851 *T)
    sumX1  +=       0.996 * cos(3.06866654160  + 4863.44173055270 *T)
    sumX1  +=       1.059 * cos(4.47767598624  + 25653.94758621389 *T)
    sumX1  +=       0.936 * cos(1.21580989528  + 307.07445958451 *T)
    sumX1  +=       0.986 * cos(1.95558000476  + 5842.00674763030 *T)
    sumX1  +=       1.061 * cos(3.92781056471  + 418.97770183331 *T)
    sumX1  +=       0.941 * cos(5.52302459792  + 97.17169833279 *T)
    sumX1  +=       1.049 * cos(3.59730911706  + 416.79314583689 *T)
    sumX1  +=       0.956 * cos(2.39282306271  + 141.45507312490 *T)
    sumX1  +=       0.922 * cos(4.97372856088  + 1364.48428209840 *T)
    sumX1  +=       1.090 * cos(1.18314786349  + 1151.93594618579 *T)
    sumX1  +=       0.945 * cos(5.48577517665  + 220.86890824340 *T)
    sumX1  +=       0.945 * cos(2.54544364353  + 206.21691759960 *T)
    sumX1  +=       0.906 * cos(2.01210100437  + 12779.69461290430 *T)
    sumX1  +=       0.913 * cos(6.06812377280  + 285.61619850311 *T)
    sumX1  +=       0.993 * cos(0.00718300554  + 73.05330837550 *T)
    sumX1  +=       0.923 * cos(3.95449289739  + 1048.58004740881 *T)
    sumX1  +=       0.910 * cos(2.91944659858  + 1250.83344258291 *T)
    sumX1  +=       1.055 * cos(2.96570583025  + 287.69386416990 *T)
    sumX1  +=       0.891 * cos(4.95995674723  + 4443.95708667750 *T)
    sumX1  +=       0.890 * cos(3.44861046216  + 952.34018776010 *T)
    sumX1  +=       1.011 * cos(3.98082100593  + 138.90432670610 *T)
    sumX1  +=       0.943 * cos(2.88169222264  + 1368.41643536150 *T)
    sumX1  +=       1.240 * cos(2.96642006057  + 105.78427225690 *T)
    sumX1  +=       1.240 * cos(2.96642006057  + 105.29663728990 *T)
    sumX1  +=       0.975 * cos(1.42837247530  + 700.90805668431 *T)
    sumX1  +=       0.963 * cos(2.83983881025  + 5378.90560164571 *T)
    sumX1  +=       1.131 * cos(3.49207338910  + 640.09291968770 *T)
    sumX1  +=       1.014 * cos(3.84499571511  + 246.68001371321 *T)
    sumX1  +=       0.922 * cos(3.37907074746  + 9772.21644384989 *T)
    sumX1  +=       0.896 * cos(4.52407069260  + 1571.15746550260 *T)
    sumX1  +=       1.121 * cos(5.34360557183  + 6467.68194047809 *T)
    sumX1  +=       0.872 * cos(2.88021294314  + 519.63984183961 *T)
    sumX1  +=       1.164 * cos(0.08165804323  + 739.04109232210 *T)
    sumX1  +=       1.090 * cos(2.12969540273  + 1699.03539901970 *T)
    sumX1  +=       1.180 * cos(2.44363933993  + 636.47207509279 *T)
    sumX1  +=       0.912 * cos(4.11625838132  + 31.26330612050 *T)
    sumX1  +=       0.892 * cos(3.93430854425  + 212.57970507500 *T)
    sumX1  +=       0.892 * cos(3.93430854425  + 212.09207010800 *T)
    sumX1  +=       0.985 * cos(5.09403476262  + 988.02554284310 *T)
    sumX1  +=       0.860 * cos(4.51926390946  + 1585.38455950421 *T)
    sumX1  +=       0.990 * cos(0.10745457981  + 3149.44778508089 *T)
    sumX1  +=       1.177 * cos(0.16333880643  + 1328.79643488791 *T)
    sumX1  +=       0.864 * cos(1.23920060185  + 12146.42323862409 *T)
    sumX1  +=       1.007 * cos(2.26868943041  + 685.23011986920 *T)
    sumX1  +=       1.007 * cos(5.41028208400  + 685.71775483620 *T)
    sumX1  +=       0.853 * cos(5.50195803246  + 2641.58509595570 *T)
    sumX1  +=       0.993 * cos(2.58838169596  + 2111.89413086110 *T)
    sumX1  +=       0.844 * cos(4.99931164886  + 1239.78774231901 *T)
    sumX1  +=       0.846 * cos(3.17548903050  + 2115.07552459880 *T)
    sumX1  +=       0.964 * cos(2.75317670197  + 756.56720014041 *T)
    sumX1  +=       0.981 * cos(1.46114906045  + 1471.24608502170 *T)
    sumX1  +=       1.164 * cos(6.03626063674  + 3391.68870348911 *T)
    sumX1  +=       0.885 * cos(3.21468341749  + 609.12161516050 *T)
    sumX1  +=       0.821 * cos(0.43214580939  + 388.22133775470 *T)
    sumX1  +=       0.906 * cos(2.38644851609  + 327.19375243699 *T)
    sumX1  +=       0.903 * cos(3.04044650688  + 477.24458453860 *T)
    sumX1  +=       0.841 * cos(3.88859140273  + 473.31243127551 *T)
    sumX1  +=       0.904 * cos(0.75775511775  + 635.47523735150 *T)
    sumX1  +=       0.822 * cos(0.09352704461  + 1365.23504162379 *T)
    sumX1  +=       0.997 * cos(0.25000097268  + 123.29582586020 *T)
    sumX1  +=       0.879 * cos(5.54528283460  + 3362.74688051889 *T)
    sumX1  +=       0.846 * cos(1.99050608386  + 227.81819103291 *T)
    sumX1  +=       0.846 * cos(6.04071273632  + 199.26763481011 *T)
    sumX1  +=       0.947 * cos(4.80416752233  + 222.61650551010 *T)
    sumX1  +=       0.875 * cos(3.48686422480  + 9993.11672125570 *T)
    sumX1  +=       1.117 * cos(0.69343471577  + 319.08336769490 *T)
    sumX1  +=       0.800 * cos(5.83577165546  + 834.79331700380 *T)
    sumX1  +=       0.839 * cos(6.01134529585  + 206.64181424181 *T)
    sumX1  +=       0.839 * cos(2.01987352433  + 220.44401160121 *T)
    sumX1  +=       0.838 * cos(3.26925163574  + 2442.51309451930 *T)
    sumX1  +=       0.830 * cos(2.56385471112  + 1891.48148842230 *T)
    sumX1  +=       0.841 * cos(1.50960565043  + 60.04756252381 *T)
    sumX1  +=       0.826 * cos(3.67482294787  + 220.13519506050 *T)
    sumX1  +=       0.854 * cos(1.21615128625  + 540.22972331660 *T)
    sumX1  +=       0.842 * cos(5.83341675596  + 2854.39655642670 *T)
    sumX1  +=       0.858 * cos(5.69541304884  + 781.83999440590 *T)
    sumX1  +=       0.860 * cos(2.63629505046  + 5436.74919775670 *T)
    sumX1  +=       0.785 * cos(2.62141651385  + 3494.29384274070 *T)

    # SATURN X Coord
    sumX2  = 0
    sumX2  +=  862721.930 * cos(2.44701762869  + 213.54291292150 *T)
    sumX2  +=  611630.287 * cos(2.28479450822  + 0.24381748350 *T)
    sumX2  +=  561413.805 * cos(1.25489421778  + 206.42936592071 *T)
    sumX2  +=  547271.370 * cos(3.61668298145  + 220.65645992230 *T)
    sumX2  +=  234984.093 * cos(2.44660941367  + 426.84200835950 *T)
    sumX2  +=   88108.317 * cos(4.86838582414  + 7.35736448430 *T)
    sumX2  +=   50142.527 * cos(2.69730354414  + 433.95555536030 *T)
    sumX2  +=   45054.746 * cos(0.35338915884  + 419.72846135871 *T)
    sumX2  +=   33335.009 * cos(1.58127217455  + 640.14110379750 *T)
    sumX2  +=   27185.514 * cos(3.49604874733  + 227.77000692311 *T)
    sumX2  +=   18651.786 * cos(0.02912884251  + 213.05527795450 *T)
    sumX2  +=   12946.617 * cos(1.40691297654  + 199.31581891990 *T)
    sumX2  +=   10383.025 * cos(4.75910342894  + 14.47091148511 *T)
    sumX2  +=    6684.574 * cos(5.66699712629  + 110.45013870291 *T)
    sumX2  +=    4661.903 * cos(1.83447973785  + 647.25465079831 *T)
    sumX2  +=    4933.834 * cos(3.08177201518  + 216.72430665921 *T)
    sumX2  +=    4778.577 * cos(1.81292692641  + 210.36151918381 *T)
    sumX2  +=    4377.928 * cos(5.24025788610  + 316.63568714010 *T)
    sumX2  +=    3532.396 * cos(0.77695513689  + 853.44019923550 *T)
    sumX2  +=    2688.895 * cos(3.15397291653  + 209.61075965840 *T)
    sumX2  +=    2540.548 * cos(2.64953783571  + 441.06910236111 *T)
    sumX2  +=    2540.549 * cos(1.72979147362  + 217.47506618460 *T)
    sumX2  +=    2506.252 * cos(4.68668320836  + 117.56368570370 *T)
    sumX2  +=    2582.753 * cos(1.59861873487  + 103.33659170210 *T)
    sumX2  +=    2111.683 * cos(0.27536451133  + 633.02755679670 *T)
    sumX2  +=    2306.046 * cos(3.67686226266  + 205.94173095370 *T)
    sumX2  +=    1843.706 * cos(3.97050945214  + 323.74923414091 *T)
    sumX2  +=    1272.298 * cos(0.25661618943  + 412.61491435790 *T)
    sumX2  +=     940.336 * cos(0.85421668175  + 6.86972951729 *T)
    sumX2  +=     804.980 * cos(3.35669822169  + 234.88355392390 *T)
    sumX2  +=     698.973 * cos(5.48447806995  + 95.73540973430 *T)
    sumX2  +=     653.010 * cos(3.44302214541  + 198.82818395289 *T)
    sumX2  +=     766.827 * cos(1.08680299509  + 202.49721265760 *T)
    sumX2  +=     662.176 * cos(5.13243854047  + 426.35437339250 *T)
    sumX2  +=     565.841 * cos(6.26658788528  + 220.16882495529 *T)
    sumX2  +=     765.788 * cos(3.85596138373  + 224.58861318540 *T)
    sumX2  +=     672.412 * cos(1.14437192898  + 309.52214013930 *T)
    sumX2  +=     490.992 * cos(4.81907666832  + 109.96250373590 *T)
    sumX2  +=     440.264 * cos(3.59515263445  + 330.86278114170 *T)
    sumX2  +=     426.768 * cos(0.70988535006  + 11.28951774740 *T)
    sumX2  +=     427.613 * cos(1.00345444265  + 860.55374623631 *T)
    sumX2  +=     505.832 * cos(4.26386101303  + 309.03450517230 *T)
    sumX2  +=     559.499 * cos(3.55753813954  + 102.84895673509 *T)
    sumX2  +=     439.569 * cos(4.61666678676  + 21.58445848590 *T)
    sumX2  +=     404.335 * cos(2.95264099159  + 223.83785366000 *T)
    sumX2  +=     400.108 * cos(2.21768387014  + 430.02340209721 *T)
    sumX2  +=     338.086 * cos(0.01109251698  + 1066.73929467350 *T)
    sumX2  +=     310.691 * cos(4.43935743139  + 3.42521122121 *T)
    sumX2  +=     356.340 * cos(2.07173270088  + 422.90985509640 *T)
    sumX2  +=     268.315 * cos(5.61477011539  + 88.62186273350 *T)
    sumX2  +=     265.067 * cos(1.09651380636  + 423.66061462181 *T)
    sumX2  +=     249.792 * cos(3.78368875676  + 4.17597074660 *T)
    sumX2  +=     244.377 * cos(1.81782248050  + 654.36819779910 *T)
    sumX2  +=     305.650 * cos(3.17709889511  + 419.24082639170 *T)
    sumX2  +=     263.490 * cos(3.32799413830  + 124.67723270450 *T)
    sumX2  +=     243.362 * cos(1.58340419486  + 96.22304470130 *T)
    sumX2  +=     228.016 * cos(3.15581836524  + 522.33360061030 *T)
    sumX2  +=     208.205 * cos(1.59689210791  + 192.20227191910 *T)
    sumX2  +=     262.941 * cos(3.81998984240  + 522.82123557730 *T)
    sumX2  +=     217.559 * cos(2.12737362973  + 412.12727939089 *T)
    sumX2  +=     204.218 * cos(2.31819875139  + 149.80701461810 *T)
    sumX2  +=     207.097 * cos(2.04294057461  + 625.91400979590 *T)
    sumX2  +=     226.266 * cos(2.00241821857  + 203.24797218300 *T)
    sumX2  +=     164.818 * cos(2.78185091937  + 537.04832957890 *T)
    sumX2  +=     154.166 * cos(2.09392737387  + 302.40859313851 *T)
    sumX2  +=     159.821 * cos(5.14771420694  + 529.93478257810 *T)
    sumX2  +=     127.987 * cos(3.10053496626  + 515.22005360950 *T)
    sumX2  +=     150.521 * cos(4.33804123124  + 301.92095817150 *T)
    sumX2  +=     109.809 * cos(2.44188223375  + 277.27881122490 *T)
    sumX2  +=     107.126 * cos(4.17749068544  + 10.53875822200 *T)
    sumX2  +=     125.246 * cos(3.27781318747  + 437.88770862340 *T)
    sumX2  +=     119.913 * cos(4.03198023861  + 625.42637482889 *T)
    sumX2  +=      95.870 * cos(0.55536550370  + 13.98327651810 *T)
    sumX2  +=      93.192 * cos(1.91525183876  + 430.77416162260 *T)
    sumX2  +=      82.653 * cos(2.55387341854  + 448.18264936190 *T)
    sumX2  +=      80.395 * cos(4.59254494799  + 231.70216018621 *T)
    sumX2  +=      86.059 * cos(4.21228581841  + 742.74624304909 *T)
    sumX2  +=      78.439 * cos(3.94622515922  + 415.79630809560 *T)
    sumX2  +=      95.719 * cos(0.39732138114  + 846.32665223470 *T)
    sumX2  +=      69.822 * cos(2.93302398470  + 838.72547026689 *T)
    sumX2  +=      88.159 * cos(4.69429588352  + 743.23387801611 *T)
    sumX2  +=      62.703 * cos(1.89462060621  + 728.51914904750 *T)
    sumX2  +=      61.768 * cos(5.02462718406  + 3.68833577960 *T)
    sumX2  +=      59.007 * cos(0.78786498806  + 316.14805217309 *T)
    sumX2  +=      57.483 * cos(5.15275075778  + 18.40306474821 *T)
    sumX2  +=      68.064 * cos(2.90200859132  + 736.12033101530 *T)
    sumX2  +=      73.156 * cos(3.00215539369  + 735.63269604830 *T)
    sumX2  +=      66.143 * cos(5.32689397063  + 529.44714761109 *T)
    sumX2  +=      46.643 * cos(2.82828712050  + 831.61192326610 *T)
    sumX2  +=      46.381 * cos(2.73650889523  + 544.16187657970 *T)
    sumX2  +=      49.127 * cos(3.98526295508  + 515.70768857651 *T)
    sumX2  +=      45.075 * cos(3.61480562395  + 195.38366565681 *T)
    sumX2  +=      45.874 * cos(0.00322886163  + 408.68276109481 *T)
    sumX2  +=      39.597 * cos(0.18910102232  + 1073.85284167431 *T)
    sumX2  +=      37.898 * cos(5.19338156968  + 1059.62574767270 *T)
    sumX2  +=      44.466 * cos(3.71987438426  + 941.81824448550 *T)
    sumX2  +=      34.788 * cos(1.53042898513  + 363.10611005610 *T)
    sumX2  +=      32.598 * cos(5.38947358762  + 433.46792039329 *T)
    sumX2  +=      33.292 * cos(3.55448256672  + 191.71463695210 *T)
    sumX2  +=      41.206 * cos(4.06677518336  + 618.31282782810 *T)
    sumX2  +=      32.714 * cos(3.96597866859  + 1059.13811270570 *T)
    sumX2  +=      41.218 * cos(1.19485367697  + 636.20895053440 *T)
    sumX2  +=      36.847 * cos(0.14902811646  + 223.10414047710 *T)
    sumX2  +=      28.979 * cos(4.45566358804  + 113.63153244060 *T)
    sumX2  +=      30.445 * cos(5.99732067739  + 56.37853381910 *T)
    sumX2  +=      38.514 * cos(2.16683520575  + 86.07111631471 *T)
    sumX2  +=      33.922 * cos(2.16638658532  + 437.13694909800 *T)
    sumX2  +=      29.740 * cos(5.33246443349  + 270.16526422410 *T)
    sumX2  +=      34.418 * cos(1.38531328883  + 643.32249753521 *T)
    sumX2  +=      31.920 * cos(2.71809642095  + 337.97632814250 *T)
    sumX2  +=      35.309 * cos(4.64777361590  + 203.98168536590 *T)
    sumX2  +=      26.923 * cos(2.72104940113  + 131.79077970531 *T)
    sumX2  +=      26.272 * cos(2.14472258006  + 750.34742501690 *T)
    sumX2  +=      25.940 * cos(1.96359688358  + 76.50988875911 *T)
    sumX2  +=      29.423 * cos(5.51595511461  + 1280.03839011150 *T)
    sumX2  +=      26.782 * cos(2.69721459229  + 1155.11733992350 *T)
    sumX2  +=      24.375 * cos(5.77172279789  + 81.50831573269 *T)
    sumX2  +=      25.294 * cos(4.58498242616  + 483.46435966210 *T)
    sumX2  +=      23.209 * cos(1.00054709783  + 867.66729323710 *T)
    sumX2  +=      22.169 * cos(0.86121660984  + 288.32451148881 *T)
    sumX2  +=      23.117 * cos(6.14239842522  + 227.28237195610 *T)
    sumX2  +=      21.350 * cos(3.80984077151  + 160.85271488200 *T)
    sumX2  +=      22.067 * cos(1.65390020558  + 1044.91101870410 *T)
    sumX2  +=      22.088 * cos(2.09313761105  + 70.60562782070 *T)
    sumX2  +=      20.601 * cos(2.14444325102  + 22.33521801130 *T)
    sumX2  +=      20.035 * cos(2.56955927602  + 1148.00379292269 *T)
    sumX2  +=      26.634 * cos(2.72752998143  + 1052.02456570489 *T)
    sumX2  +=      20.249 * cos(1.08828213970  + 416.54706762100 *T)
    sumX2  +=      19.318 * cos(0.56867565113  + 497.69145366371 *T)
    sumX2  +=      17.591 * cos(0.47058042874  + 629.09540353360 *T)
    sumX2  +=      18.938 * cos(3.01811688461  + 107.26874496520 *T)
    sumX2  +=      21.595 * cos(5.64697422334  + 490.57790666290 *T)
    sumX2  +=      20.105 * cos(2.11272613285  + 839.21310523390 *T)
    sumX2  +=      21.565 * cos(4.03058232828  + 956.53297345411 *T)
    sumX2  +=      16.938 * cos(2.88265146092  + 89.10949770051 *T)
    sumX2  +=      16.497 * cos(4.71187682763  + 106.51798543980 *T)
    sumX2  +=      21.360 * cos(5.07283473961  + 845.83901726770 *T)
    sumX2  +=      19.384 * cos(3.45868972702  + 1258.21011414209 *T)
    sumX2  +=      15.417 * cos(3.16624042233  + 241.99710092470 *T)
    sumX2  +=      16.591 * cos(4.54165521178  + 639.65346883050 *T)
    sumX2  +=      16.177 * cos(6.04286742202  + 344.94686279140 *T)
    sumX2  +=      14.854 * cos(1.11858145356  + 266.23311096100 *T)
    sumX2  +=      19.063 * cos(2.07999478622  + 201.01273994931 *T)
    sumX2  +=      18.723 * cos(2.80004753140  + 226.07308589371 *T)
    sumX2  +=      18.567 * cos(6.15802064616  + 142.69346761730 *T)
    sumX2  +=      14.166 * cos(3.57626351086  + 632.53992182970 *T)
    sumX2  +=      17.119 * cos(0.83663196525  + 352.06040979221 *T)
    sumX2  +=      18.888 * cos(2.63389139857  + 52.44638055600 *T)
    sumX2  +=      13.506 * cos(4.03963934050  + 138.76131435421 *T)
    sumX2  +=      17.909 * cos(3.76988751681  + 934.70469748469 *T)
    sumX2  +=      13.747 * cos(2.11060380479  + 359.17395679300 *T)
    sumX2  +=      14.401 * cos(0.42335490960  + 405.50136735710 *T)
    sumX2  +=      16.878 * cos(2.75966318459  + 341.01470952831 *T)
    sumX2  +=      13.527 * cos(6.11265092399  + 284.39235822571 *T)
    sumX2  +=      13.436 * cos(4.50317832794  + 6290.43321447570 *T)
    sumX2  +=      12.962 * cos(0.85501464894  + 181.29958400710 *T)
    sumX2  +=      15.865 * cos(2.58925256281  + 651.18680406140 *T)
    sumX2  +=      13.136 * cos(2.14345131759  + 6276.20612047410 *T)
    sumX2  +=      12.088 * cos(4.09973262355  + 127.22797912329 *T)
    sumX2  +=      12.658 * cos(3.42538380158  + 948.93179148630 *T)
    sumX2  +=      12.303 * cos(3.76145799738  + 445.00125562421 *T)
    sumX2  +=      11.905 * cos(0.05199682564  + 424.39432780471 *T)
    sumX2  +=      12.624 * cos(5.92655665827  + 10220.64291069530 *T)
    sumX2  +=      12.885 * cos(3.71011528810  + 1375.52998236229 *T)
    sumX2  +=      12.454 * cos(2.37524428293  + 295.29504613770 *T)
    sumX2  +=      12.341 * cos(3.56682893682  + 10206.41581669370 *T)
    sumX2  +=      11.182 * cos(4.71695359508  + 209.12312469139 *T)
    sumX2  +=      10.860 * cos(2.79531710541  + 350.57593708390 *T)
    sumX2  +=      10.568 * cos(5.88568250783  + 12.77399045571 *T)
    sumX2  +=      10.726 * cos(5.64879100007  + 703.87700210090 *T)
    sumX2  +=      13.731 * cos(0.78531914787  + 636.95971005981 *T)
    sumX2  +=      11.066 * cos(2.49843137169  + 1471.50920958009 *T)
    sumX2  +=      10.568 * cos(4.02434153212  + 245.78624183590 *T)
    sumX2  +=      11.834 * cos(5.61242569644  + 191.45151239371 *T)
    sumX2  +=      13.045 * cos(2.99603652801  + 323.26159917390 *T)
    sumX2  +=      12.746 * cos(3.09998912111  + 729.00678401450 *T)
    sumX2  +=      11.214 * cos(2.26091951790  + 618.80046279510 *T)
    sumX2  +=      11.539 * cos(0.54965317551  + 75.02541605080 *T)
    sumX2  +=      12.236 * cos(4.46231008485  + 294.80741117069 *T)
    sumX2  +=       9.934 * cos(4.15810018234  + 209.87388421679 *T)
    sumX2  +=      10.182 * cos(5.59563792015  + 235.63431344930 *T)
    sumX2  +=       9.148 * cos(2.31017574698  + 1464.39566257930 *T)
    sumX2  +=       9.436 * cos(3.25931284572  + 508.10650660869 *T)
    sumX2  +=       8.422 * cos(1.74868225923  + 661.48174479990 *T)
    sumX2  +=       8.359 * cos(1.43667117664  + 1361.30288836070 *T)
    sumX2  +=       8.253 * cos(0.86562358245  + 565.35950523021 *T)
    sumX2  +=       9.478 * cos(2.12166784380  + 949.41942645330 *T)
    sumX2  +=      10.785 * cos(2.48347595884  + 1368.41643536150 *T)
    sumX2  +=       8.229 * cos(3.53945903151  + 5849.12029463110 *T)
    sumX2  +=       7.715 * cos(0.79383137585  + 10.80188278040 *T)
    sumX2  +=       8.470 * cos(3.01956396957  + 1052.51220067191 *T)
    sumX2  +=       7.922 * cos(5.89929744937  + 5863.34738863270 *T)
    sumX2  +=       7.332 * cos(2.80163069110  + 404.75060783170 *T)
    sumX2  +=       7.694 * cos(4.96303865474  + 9779.32999085069 *T)
    sumX2  +=       8.936 * cos(3.18123904182  + 1574.60198379870 *T)
    sumX2  +=       9.236 * cos(4.82991847781  + 1162.23088692429 *T)
    sumX2  +=       7.408 * cos(1.03969580778  + 9793.55708485230 *T)
    sumX2  +=       7.006 * cos(0.72365973561  + 208.87704647550 *T)
    sumX2  +=       8.885 * cos(3.49149089495  + 1251.09656714130 *T)
    sumX2  +=       6.858 * cos(1.01645145725  + 218.20877936750 *T)
    sumX2  +=       6.727 * cos(0.58555082022  + 313.45429340240 *T)
    sumX2  +=       6.237 * cos(2.16585608727  + 319.81708087780 *T)
    sumX2  +=       6.440 * cos(5.35585173212  + 429.28968891431 *T)
    sumX2  +=       8.565 * cos(3.30273694005  + 6283.31966747490 *T)
    sumX2  +=       7.708 * cos(5.79722814055  + 621.98185653281 *T)
    sumX2  +=       6.384 * cos(3.26113757491  + 1265.32366114289 *T)
    sumX2  +=       7.986 * cos(4.72604674713  + 10213.52936369450 *T)
    sumX2  +=       5.697 * cos(5.81066083045  + 121.49583896680 *T)
    sumX2  +=       5.771 * cos(2.43214323984  + 114.38229196601 *T)
    sumX2  +=       5.568 * cos(1.85852168121  + 194.89603068979 *T)
    sumX2  +=       6.666 * cos(3.97400153538  + 202.76033721600 *T)
    sumX2  +=       6.258 * cos(0.27664464362  + 117.07605073670 *T)
    sumX2  +=       5.194 * cos(6.09782391086  + 842.39449897160 *T)
    sumX2  +=       5.192 * cos(6.26641739333  + 78.95756931390 *T)
    sumX2  +=       6.155 * cos(5.15951383844  + 196.13442518220 *T)
    sumX2  +=       5.311 * cos(5.92397473787  + 956.04533848709 *T)
    sumX2  +=       6.053 * cos(2.87275161723  + 230.95140066080 *T)
    sumX2  +=       6.283 * cos(3.93024472580  + 63.49208081989 *T)
    sumX2  +=       5.550 * cos(3.50331649787  + 1691.92185201890 *T)
    sumX2  +=       6.169 * cos(2.11657776036  + 536.56069461190 *T)
    sumX2  +=       4.641 * cos(5.22350241766  + 436.40323591510 *T)
    sumX2  +=       4.643 * cos(4.78610214245  + 9.31741007210 *T)
    sumX2  +=       4.944 * cos(2.25390759795  + 1787.90107923670 *T)
    sumX2  +=       4.821 * cos(2.05306967383  + 644.07325706060 *T)
    sumX2  +=       4.551 * cos(2.00498534256  + 757.46097201770 *T)
    sumX2  +=       4.644 * cos(4.49405917272  + 9.80504503910 *T)
    sumX2  +=       4.204 * cos(0.80174207582  + 576.40520549410 *T)
    sumX2  +=       4.172 * cos(0.63505087766  + 251.67594855930 *T)
    sumX2  +=       4.304 * cos(1.00860438730  + 289.80898419710 *T)
    sumX2  +=       4.123 * cos(4.15758653931  + 175.40987728371 *T)
    sumX2  +=       4.456 * cos(2.99760258198  + 824.49837626530 *T)
    sumX2  +=       5.291 * cos(4.58152433636  + 1478.62275658089 *T)
    sumX2  +=       4.026 * cos(6.02985977936  + 710.99054910171 *T)
    sumX2  +=       5.280 * cos(4.11176050624  + 202.00957769060 *T)
    sumX2  +=       4.248 * cos(3.57017670015  + 398.38782035631 *T)
    sumX2  +=       3.877 * cos(3.03469279117  + 98.65617104110 *T)
    sumX2  +=       3.948 * cos(5.52901251368  + 1589.31671276730 *T)
    sumX2  +=       4.047 * cos(2.10751405728  + 1780.78753223589 *T)
    sumX2  +=       3.924 * cos(1.24978633707  + 134.34152612410 *T)
    sumX2  +=       4.377 * cos(6.22486603200  + 10.05112325500 *T)
    sumX2  +=       3.858 * cos(0.27422301004  + 501.62360692681 *T)
    sumX2  +=       3.847 * cos(4.82150225291  + 333.90116252750 *T)
    sumX2  +=       5.256 * cos(4.70444208702  + 5856.23384163189 *T)
    sumX2  +=       5.054 * cos(1.27058229470  + 479.53220639900 *T)
    sumX2  +=       4.175 * cos(2.79949589454  + 63.97971578690 *T)
    sumX2  +=       4.150 * cos(4.07784295914  + 106.03035047280 *T)
    sumX2  +=       3.760 * cos(4.25879904430  + 508.59414157570 *T)
    sumX2  +=       4.966 * cos(5.83309328873  + 216.23667169220 *T)
    sumX2  +=       4.880 * cos(2.13206549479  + 1684.80830501809 *T)
    sumX2  +=       4.042 * cos(6.16882807370  + 120.11443212250 *T)
    sumX2  +=       4.914 * cos(6.12887362514  + 9786.44353785149 *T)
    sumX2  +=       3.672 * cos(3.95123578303  + 422.42222012939 *T)
    sumX2  +=       4.671 * cos(3.22550913459  + 1567.48843679789 *T)
    sumX2  +=       3.483 * cos(5.20582592049  + 348.12825652910 *T)
    sumX2  +=       3.700 * cos(2.70783154700  + 416.05943265400 *T)
    sumX2  +=       4.300 * cos(2.91568424793  + 1890.99385345530 *T)
    sumX2  +=       3.599 * cos(1.40195749079  + 2.93757625420 *T)
    sumX2  +=       3.544 * cos(2.31276351688  + 852.95256426850 *T)
    sumX2  +=       3.267 * cos(1.20456424957  + 1677.69475801730 *T)
    sumX2  +=       3.574 * cos(5.94804944384  + 156.92056161890 *T)
    sumX2  +=       3.416 * cos(1.75912124355  + 167.07249000551 *T)
    sumX2  +=       3.394 * cos(3.09924705896  + 196.27743753410 *T)
    sumX2  +=       3.544 * cos(2.13015052882  + 114.64292439690 *T)
    sumX2  +=       3.999 * cos(3.22701633780  + 326.93062787860 *T)
    sumX2  +=       3.152 * cos(1.19755571878  + 151.29148732640 *T)
    sumX2  +=       3.262 * cos(5.70768338968  + 1272.43720814370 *T)
    sumX2  +=       3.054 * cos(0.93612488497  + 749.85979004990 *T)
    sumX2  +=       3.350 * cos(5.35233532300  + 212.57970507500 *T)
    sumX2  +=       3.340 * cos(2.68186323180  + 214.50612076800 *T)
    sumX2  +=       3.138 * cos(0.69922169045  + 320.56784040320 *T)
    sumX2  +=       3.732 * cos(4.18854031121  + 611.19928082730 *T)
    sumX2  +=       3.741 * cos(0.62303283965  + 849.50804597240 *T)
    sumX2  +=       3.047 * cos(1.79037915226  + 230.80838830890 *T)
    sumX2  +=       3.048 * cos(5.94415826698  + 21.09682351890 *T)
    sumX2  +=       2.851 * cos(4.73186877358  + 6069.53293706989 *T)
    sumX2  +=       2.864 * cos(6.09139978110  + 9999.74263328950 *T)
    sumX2  +=       2.898 * cos(2.37720286549  + 213.49472881170 *T)
    sumX2  +=       2.898 * cos(5.65686211315  + 213.59109703130 *T)
    sumX2  +=       2.819 * cos(6.19576518658  + 431.50787480550 *T)
    sumX2  +=       3.469 * cos(4.96542567369  + 526.75338884040 *T)
    sumX2  +=       3.327 * cos(3.15421993340  + 1581.71553079950 *T)
    sumX2  +=       2.927 * cos(2.32768555098  + 395.82251972250 *T)
    sumX2  +=       3.762 * cos(0.62502984773  + 38.37685312130 *T)
    sumX2  +=       2.893 * cos(6.25258908606  + 381.59542572090 *T)
    sumX2  +=       3.707 * cos(1.85956087078  + 215.99059347631 *T)
    sumX2  +=       2.804 * cos(3.11861161896  + 260.01333583751 *T)
    sumX2  +=       2.799 * cos(0.07347298152  + 129.67565967810 *T)
    sumX2  +=       3.378 * cos(1.60714359099  + 299.37021175271 *T)
    sumX2  +=       2.783 * cos(5.69238885511  + 1287.15193711231 *T)
    sumX2  +=       3.526 * cos(6.21925842594  + 211.09523236671 *T)
    sumX2  +=       2.541 * cos(5.29829781548  + 414.31183538731 *T)
    sumX2  +=       2.607 * cos(3.83630269401  + 276.79117625789 *T)
    sumX2  +=       2.465 * cos(1.10299698027  + 572.47305223100 *T)
    sumX2  +=       2.461 * cos(3.95494769587  + 409.43352062020 *T)
    sumX2  +=       2.462 * cos(5.12829780057  + 59.55992755679 *T)
    sumX2  +=       2.579 * cos(4.98684783023  + 1493.33748554950 *T)
    sumX2  +=       2.414 * cos(5.95835828919  + 343.46239008310 *T)
    sumX2  +=       2.533 * cos(3.91175931336  + 696.76345510010 *T)
    sumX2  +=       2.390 * cos(5.50027750787  + 1588.82907780029 *T)
    sumX2  +=       2.433 * cos(3.53308752088  + 6070.02057203690 *T)
    sumX2  +=       2.747 * cos(3.63141908197  + 216.98743121760 *T)
    sumX2  +=       2.420 * cos(4.62294661622  + 45.33283355520 *T)
    sumX2  +=       2.297 * cos(2.61803220390  + 99.66756299740 *T)
    sumX2  +=       2.813 * cos(5.66780576556  + 422.17614191350 *T)
    sumX2  +=       2.913 * cos(2.14239548593  + 551.27542358051 *T)
    sumX2  +=       2.860 * cos(5.75550548585  + 69.39634175830 *T)
    sumX2  +=       2.387 * cos(3.66176528516  + 1169.83206889211 *T)
    sumX2  +=       2.258 * cos(3.42521115550  + 188.27011865600 *T)
    sumX2  +=       2.293 * cos(4.30779033149  + 98.91680347199 *T)
    sumX2  +=       2.711 * cos(4.38875028285  + 1795.01462623750 *T)
    sumX2  +=       2.337 * cos(4.98618107601  + 10000.23026825651 *T)
    sumX2  +=       2.777 * cos(1.92096933827  + 100.15519796440 *T)
    sumX2  +=       3.022 * cos(4.27657430351  + 190.96387742670 *T)
    sumX2  +=       2.327 * cos(2.43188651345  + 143.90275367970 *T)
    sumX2  +=       2.152 * cos(4.96770554275  + 189.16389053330 *T)
    sumX2  +=       2.147 * cos(5.81368882353  + 265.74547599400 *T)
    sumX2  +=       2.162 * cos(4.60033037919  + 238.81570718700 *T)
    sumX2  +=       2.184 * cos(4.98321991931  + 917.17609753890 *T)
    sumX2  +=       2.297 * cos(5.88374894966  + 533.11617631580 *T)
    sumX2  +=       2.099 * cos(3.63180133150  + 275.79433851660 *T)
    sumX2  +=       2.122 * cos(0.04103472829  + 93.18466331550 *T)
    sumX2  +=       2.068 * cos(3.90345993566  + 221.61966776881 *T)
    sumX2  +=       2.635 * cos(4.32616395006  + 646.76701583129 *T)
    sumX2  +=       2.175 * cos(5.63002332525  + 432.47108265200 *T)
    sumX2  +=       2.049 * cos(4.11329374307  + 205.46615807421 *T)
    sumX2  +=       2.103 * cos(3.07406244371  + 554.31380496631 *T)
    sumX2  +=       2.036 * cos(4.06354766916  + 415.30867312860 *T)
    sumX2  +=       2.045 * cos(4.55001787181  + 547.20025796550 *T)
    sumX2  +=       1.996 * cos(4.73453786942  + 5642.93474619389 *T)
    sumX2  +=       2.709 * cos(2.90208204814  + 174.18603700630 *T)
    sumX2  +=       2.499 * cos(3.91734791042  + 927.59115048390 *T)
    sumX2  +=       2.481 * cos(2.24814867939  + 91.80325647120 *T)
    sumX2  +=       2.555 * cos(1.49254264738  + 196.86813836510 *T)
    sumX2  +=       2.555 * cos(0.25549086562  + 230.21768747790 *T)
    sumX2  +=       2.082 * cos(3.81823920275  + 832.09955823310 *T)
    sumX2  +=       2.231 * cos(2.05013884781  + 2097.17940189250 *T)
    sumX2  +=       2.095 * cos(2.74501129194  + 1140.89024592190 *T)
    sumX2  +=       1.920 * cos(1.50789835321  + 99.40443843901 *T)
    sumX2  +=       1.910 * cos(2.52768436921  + 106.78110999820 *T)
    sumX2  +=       2.301 * cos(3.33597408639  + 2008.31372167550 *T)
    sumX2  +=       2.228 * cos(0.13234013389  + 219.69325207580 *T)
    sumX2  +=       2.219 * cos(1.60488882515  + 207.39257376720 *T)
    sumX2  +=       2.444 * cos(2.58833940963  + 71.09326278771 *T)
    sumX2  +=       1.950 * cos(5.91966225582  + 417.28078080390 *T)
    sumX2  +=       1.941 * cos(3.37602923994  + 374.15181032000 *T)
    sumX2  +=       1.900 * cos(3.19545181564  + 333.41352756050 *T)
    sumX2  +=       2.353 * cos(1.38929393306  + 650.43604453600 *T)
    sumX2  +=       1.819 * cos(3.03945359046  + 120.74507944141 *T)
    sumX2  +=       1.992 * cos(1.67941763290  + 2413.57127154910 *T)
    sumX2  +=       1.805 * cos(6.12365686072  + 9573.14444241349 *T)
    sumX2  +=       2.272 * cos(6.07077392017  + 135.57992061650 *T)
    sumX2  +=       1.823 * cos(5.32795961259  + 425.87880051300 *T)
    sumX2  +=       2.143 * cos(5.66066380057  + 347.64062156210 *T)
    sumX2  +=       1.841 * cos(0.46818772275  + 220.70464403210 *T)
    sumX2  +=       1.834 * cos(1.26884124633  + 206.38118181090 *T)
    sumX2  +=       1.777 * cos(2.67341880565  + 116.18227885939 *T)
    sumX2  +=       2.138 * cos(2.15400406449  + 2104.29294889330 *T)
    sumX2  +=       1.887 * cos(4.94905172419  + 556.76148552110 *T)
    sumX2  +=       2.390 * cos(0.57829722827  + 856.62159297321 *T)
    sumX2  +=       1.825 * cos(1.86446865071  + 345.08987514331 *T)
    sumX2  +=       1.730 * cos(6.28021453291  + 237.92193530970 *T)
    sumX2  +=       2.103 * cos(2.59628252508  + 2207.38572311189 *T)
    sumX2  +=       1.779 * cos(2.78871624346  + 427.80521620600 *T)
    sumX2  +=       1.911 * cos(2.02159497354  + 145.87486135500 *T)
    sumX2  +=       1.811 * cos(1.03042062457  + 570.98857952271 *T)
    sumX2  +=       1.710 * cos(2.50465838575  + 423.17297965479 *T)
    sumX2  +=       2.261 * cos(3.11062462322  + 421.42538238810 *T)
    sumX2  +=       1.681 * cos(4.98403310748  + 305.85311143459 *T)
    sumX2  +=       2.093 * cos(1.52224482636  + 1169.34443392509 *T)
    sumX2  +=       2.071 * cos(2.14859681048  + 942.30587945250 *T)
    sumX2  +=       1.691 * cos(2.43995589654  + 455.29619636271 *T)
    sumX2  +=       2.080 * cos(3.36200869452  + 405.01373239010 *T)
    sumX2  +=       1.764 * cos(2.79871492543  + 186.71620997851 *T)
    sumX2  +=       1.662 * cos(5.22917796645  + 305.58998687620 *T)
    sumX2  +=       1.748 * cos(3.36057980101  + 899.01685027420 *T)
    sumX2  +=       1.696 * cos(1.59892728325  + 85.17734443740 *T)
    sumX2  +=       2.142 * cos(1.99415297086  + 2001.20017467470 *T)
    sumX2  +=       2.282 * cos(2.04372303495  + 439.37218133171 *T)
    sumX2  +=       1.748 * cos(0.82946401625  + 1066.25165970650 *T)
    sumX2  +=       1.908 * cos(6.13391971195  + 131.64776735340 *T)
    sumX2  +=       1.808 * cos(2.51335072478  + 1457.28211557850 *T)
    sumX2  +=       1.794 * cos(5.82136110186  + 913.24394427581 *T)
    sumX2  +=       2.167 * cos(6.18131622353  + 563.87503252191 *T)
    sumX2  +=       1.602 * cos(3.41843479847  + 327.68138740401 *T)
    sumX2  +=       1.891 * cos(3.37025921442  + 1155.60497489050 *T)
    sumX2  +=       2.205 * cos(1.14231592076  + 388.70897272171 *T)
    sumX2  +=       1.597 * cos(5.72890255976  + 281.21096448800 *T)
    sumX2  +=       1.719 * cos(2.72910423965  + 312.44290144610 *T)
    sumX2  +=       2.158 * cos(0.71647260532  + 224.10097821840 *T)
    sumX2  +=       1.659 * cos(0.02762995097  + 561.42735196710 *T)
    sumX2  +=       1.737 * cos(1.74569938692  + 167.96626188280 *T)
    sumX2  +=       1.501 * cos(4.96905771055  + 312.96665843540 *T)
    sumX2  +=       1.587 * cos(2.37695905463  + 732.45130231059 *T)
    sumX2  +=       1.488 * cos(1.15017207061  + 12.28635548869 *T)
    sumX2  +=       1.460 * cos(3.68257556907  + 2627.35800195410 *T)
    sumX2  +=       1.835 * cos(4.46496396099  + 206.47755003051 *T)
    sumX2  +=       1.708 * cos(5.28688437379  + 82.13896305161 *T)
    sumX2  +=       1.842 * cos(3.55522181358  + 220.60827581251 *T)
    sumX2  +=       1.895 * cos(5.04728306015  + 291.50590522650 *T)
    sumX2  +=       1.423 * cos(0.90600055739  + 225.32232636830 *T)
    sumX2  +=       1.423 * cos(0.84203295561  + 201.76349947470 *T)
    sumX2  +=       1.568 * cos(0.17276009703  + 685.71775483620 *T)
    sumX2  +=       1.663 * cos(2.39240670265  + 3597.87425192630 *T)
    sumX2  +=       1.491 * cos(2.09091136066  + 240.36961586451 *T)
    sumX2  +=       1.913 * cos(1.64595427075  + 295.43805848960 *T)
    sumX2  +=       1.496 * cos(3.10700890764  + 1045.39865367110 *T)
    sumX2  +=       1.477 * cos(5.34776911444  + 440.58146739410 *T)
    sumX2  +=       1.410 * cos(3.85254018400  + 137.27684164590 *T)
    sumX2  +=       1.748 * cos(3.62192812061  + 519.15220687259 *T)
    sumX2  +=       1.325 * cos(2.98811767557  + 980.91199584231 *T)
    sumX2  +=       1.352 * cos(5.53324653057  + 5429.63565075589 *T)
    sumX2  +=       1.414 * cos(6.21378141122  + 1371.59782909919 *T)
    sumX2  +=       1.697 * cos(2.96957793002  + 1898.10740045609 *T)
    sumX2  +=       1.424 * cos(3.32164076256  + 721.40560204669 *T)
    sumX2  +=       1.424 * cos(3.32164076256  + 721.89323701370 *T)
    sumX2  +=       1.295 * cos(5.37619224276  + 1905.22094745690 *T)
    sumX2  +=       1.272 * cos(5.72341064428  + 127.71561409031 *T)
    sumX2  +=       1.740 * cos(2.98134401594  + 1883.88030645450 *T)
    sumX2  +=       1.435 * cos(1.83931362023  + 699.94484883781 *T)
    sumX2  +=       1.260 * cos(0.96015330455  + 850.25880549781 *T)
    sumX2  +=       1.316 * cos(0.48574200131  + 1382.64352936309 *T)
    sumX2  +=       1.303 * cos(4.34276081528  + 28.69800548670 *T)
    sumX2  +=       1.701 * cos(3.01361282532  + 312.70353387701 *T)
    sumX2  +=       1.245 * cos(0.67360153650  + 9359.84534697549 *T)
    sumX2  +=       1.240 * cos(1.28978443782  + 1994.08662767389 *T)
    sumX2  +=       1.286 * cos(2.12726478428  + 1048.84317196720 *T)
    sumX2  +=       1.240 * cos(0.05914920282  + 148.32254190981 *T)
    sumX2  +=       1.365 * cos(2.52639842443  + 212.79215339611 *T)
    sumX2  +=       1.362 * cos(5.50511551601  + 214.29367244690 *T)
    sumX2  +=       1.206 * cos(2.27873859730  + 739.56484931140 *T)
    sumX2  +=       1.503 * cos(3.64035751539  + 1243.98302014049 *T)
    sumX2  +=       1.305 * cos(3.14322057429  + 259.11956396020 *T)
    sumX2  +=       1.167 * cos(3.43763509907  + 233.18663289450 *T)
    sumX2  +=       1.167 * cos(4.59358372111  + 193.89919294850 *T)
    sumX2  +=       1.155 * cos(2.09477644020  + 153.73916788121 *T)
    sumX2  +=       1.220 * cos(4.50978581010  + 283.90472325870 *T)
    sumX2  +=       1.153 * cos(1.49807886709  + 525.74199688410 *T)
    sumX2  +=       1.188 * cos(1.37272384469  + 258.63192899320 *T)
    sumX2  +=       1.558 * cos(6.24300347417  + 4216.43089723790 *T)
    sumX2  +=       1.423 * cos(2.20117675779  + 2420.68481854989 *T)
    sumX2  +=       1.138 * cos(5.47453297293  + 17.91542978119 *T)
    sumX2  +=       1.133 * cos(1.27426244635  + 841.90686400460 *T)
    sumX2  +=       1.109 * cos(5.77014938471  + 121.00820399980 *T)
    sumX2  +=       1.109 * cos(1.74252937792  + 629.84616305900 *T)
    sumX2  +=       1.434 * cos(2.33577168517  + 6062.41939006910 *T)
    sumX2  +=       1.122 * cos(4.02937359406  + 128.60938596760 *T)
    sumX2  +=       1.123 * cos(4.63883211615  + 3178.38960805111 *T)
    sumX2  +=       1.154 * cos(5.89138208590  + 67.91186905000 *T)
    sumX2  +=       1.130 * cos(3.68532102062  + 731.70054278519 *T)
    sumX2  +=       1.305 * cos(3.82341882270  + 97.11681657861 *T)
    sumX2  +=       1.248 * cos(2.25171827957  + 2523.77759276850 *T)
    sumX2  +=       1.161 * cos(3.82920355830  + 81.99595069970 *T)
    sumX2  +=       1.178 * cos(2.27177304436  + 1056.44435393500 *T)
    sumX2  +=       1.342 * cos(0.90865804529  + 355.99256305531 *T)
    sumX2  +=       1.249 * cos(3.35095586398  + 526.26575387340 *T)
    sumX2  +=       1.034 * cos(6.13425127848  + 777.17412795991 *T)
    sumX2  +=       1.332 * cos(3.33641634482  + 835.54407652920 *T)
    sumX2  +=       1.024 * cos(4.46551378693  + 5635.82119919310 *T)
    sumX2  +=       1.173 * cos(2.59368230210  + 74.53778108379 *T)
    sumX2  +=       1.173 * cos(1.88536238484  + 2317.59204433130 *T)
    sumX2  +=       1.012 * cos(4.43552013111  + 215.23983395090 *T)
    sumX2  +=       1.186 * cos(4.99759781142  + 1272.92484311070 *T)
    sumX2  +=       1.049 * cos(2.89184213238  + 6062.90702503610 *T)
    sumX2  +=       1.154 * cos(1.74865938292  + 2406.45772454830 *T)
    sumX2  +=       1.062 * cos(4.97332239252  + 906.13039727500 *T)
    sumX2  +=       1.124 * cos(2.19046669320  + 3914.26612158291 *T)
    sumX2  +=       0.993 * cos(3.56282013143  + 211.84599189210 *T)
    sumX2  +=       0.986 * cos(4.64695151983  + 525.51499434800 *T)
    sumX2  +=       0.986 * cos(1.18009463616  + 602.00806815971 *T)
    sumX2  +=       0.992 * cos(1.73454490947  + 306.34074640161 *T)
    sumX2  +=       1.239 * cos(2.83323142368  + 2193.15862911030 *T)
    sumX2  +=       1.134 * cos(1.51692947764  + 2729.96314120569 *T)
    sumX2  +=       1.041 * cos(0.96856014126  + 945.75039774860 *T)
    sumX2  +=       1.025 * cos(4.65212942292  + 32.48714639791 *T)
    sumX2  +=       1.044 * cos(3.13333997041  + 472.41865939820 *T)
    sumX2  +=       1.335 * cos(3.23250354837  + 2324.70559133209 *T)
    sumX2  +=       1.199 * cos(5.97883679965  + 4113.33812301930 *T)
    sumX2  +=       1.174 * cos(6.24416585602  + 1176.94561589290 *T)
    sumX2  +=       1.213 * cos(1.85427016032  + 252.89978883670 *T)
    sumX2  +=       0.973 * cos(3.32855151003  + 245.03548231051 *T)
    sumX2  +=       0.973 * cos(1.56107465656  + 182.05034353251 *T)
    sumX2  +=       1.167 * cos(4.51079844988  + 635.72131556739 *T)
    sumX2  +=       0.936 * cos(4.89205060352  + 220.13519506050 *T)
    sumX2  +=       1.073 * cos(2.28404354413  + 1773.67398523510 *T)
    sumX2  +=       0.931 * cos(3.12796311727  + 206.95063078250 *T)
    sumX2  +=       0.931 * cos(4.22108011413  + 1582.20316576650 *T)
    sumX2  +=       1.237 * cos(5.91198768163  + 2964.60287764609 *T)
    sumX2  +=       1.074 * cos(3.43802549251  + 738.81408978600 *T)
    sumX2  +=       0.976 * cos(4.61137219155  + 31.99951143090 *T)
    sumX2  +=       0.907 * cos(5.15190704681  + 1802.61580820530 *T)
    sumX2  +=       0.925 * cos(4.33653627215  + 3494.78147770770 *T)
    sumX2  +=       0.999 * cos(5.63756377401  + 171.89841514590 *T)
    sumX2  +=       0.931 * cos(3.19108655615  + 1037.79747170330 *T)
    sumX2  +=       1.094 * cos(2.28588137520  + 2516.66404576769 *T)
    sumX2  +=       0.949 * cos(5.39859475117  + 1809.72935520611 *T)
    sumX2  +=       0.964 * cos(4.14879961487  + 2641.58509595570 *T)
    sumX2  +=       0.947 * cos(4.37174355210  + 2111.40649589409 *T)
    sumX2  +=       0.898 * cos(5.87505951681  + 9566.03089541269 *T)
    sumX2  +=       1.029 * cos(4.83213149276  + 519.63984183961 *T)
    sumX2  +=       0.989 * cos(5.52532133338  + 255.18741069710 *T)
    sumX2  +=       1.003 * cos(2.89315844222  + 60.52313540329 *T)
    sumX2  +=       0.886 * cos(5.44090853481  + 305.10235190919 *T)
    sumX2  +=       0.920 * cos(3.26822857927  + 459.08533727390 *T)
    sumX2  +=       0.888 * cos(6.08377383987  + 3259.65410630030 *T)
    sumX2  +=       0.888 * cos(6.08377383987  + 3260.14174126730 *T)
    sumX2  +=       0.960 * cos(4.90164911800  + 84.68970947040 *T)
    sumX2  +=       0.881 * cos(5.34840991118  + 397.90018538930 *T)
    sumX2  +=       0.924 * cos(0.50062273312  + 731.92754532131 *T)
    sumX2  +=       0.960 * cos(4.40263473154  + 5856.72147659890 *T)
    sumX2  +=       0.886 * cos(4.84941446555  + 471.93102443120 *T)
    sumX2  +=       0.909 * cos(2.87462833456  + 746.15463932291 *T)

    # SATURN X Coord
    sumX3  = 0
    sumX3  +=   76886.822 * cos(2.95726755259  + 206.42936592071 *T)
    sumX3  +=   75865.107 * cos(1.87343394188  + 220.65645992230 *T)
    sumX3  +=   28377.246 * cos(3.86557149479  + 0.24381748350 *T)
    sumX3  +=   11451.297 * cos(0.52950959639  + 426.84200835950 *T)
    sumX3  +=    9043.324 * cos(0.90644151923  + 433.95555536030 *T)
    sumX3  +=    7419.530 * cos(1.78113005821  + 227.77000692311 *T)
    sumX3  +=    8292.033 * cos(3.07262247518  + 7.35736448430 *T)
    sumX3  +=    4421.142 * cos(2.21157518590  + 419.72846135871 *T)
    sumX3  +=    3519.324 * cos(3.11036333131  + 199.31581891990 *T)
    sumX3  +=    3064.445 * cos(5.96466087519  + 640.14110379750 *T)
    sumX3  +=    2363.042 * cos(3.04029814628  + 14.47091148511 *T)
    sumX3  +=    1027.777 * cos(0.00466674598  + 647.25465079831 *T)
    sumX3  +=     789.743 * cos(0.89493947314  + 441.06910236111 *T)
    sumX3  +=     810.744 * cos(1.27206560500  + 216.72430665921 *T)
    sumX3  +=     766.386 * cos(3.60274923273  + 210.36151918381 *T)
    sumX3  +=     673.097 * cos(3.76715088659  + 213.05527795450 *T)
    sumX3  +=     506.614 * cos(3.74427289102  + 213.54291292150 *T)
    sumX3  +=     384.956 * cos(3.04387305073  + 117.56368570370 *T)
    sumX3  +=     463.540 * cos(5.12353455374  + 853.44019923550 *T)
    sumX3  +=     411.246 * cos(3.75278799309  + 110.45013870291 *T)
    sumX3  +=     324.142 * cos(1.64931862119  + 234.88355392390 *T)
    sumX3  +=     285.917 * cos(1.98200727487  + 412.61491435790 *T)
    sumX3  +=     238.406 * cos(3.32280489930  + 103.33659170210 *T)
    sumX3  +=     192.441 * cos(0.61468306333  + 316.63568714010 *T)
    sumX3  +=     189.600 * cos(5.40892808690  + 6.86972951729 *T)
    sumX3  +=     189.610 * cos(2.26186337972  + 633.02755679670 *T)
    sumX3  +=     213.874 * cos(5.57078419443  + 205.94173095370 *T)
    sumX3  +=     153.605 * cos(2.27746123288  + 323.74923414091 *T)
    sumX3  +=     154.083 * cos(2.88123748841  + 21.58445848590 *T)
    sumX3  +=     130.169 * cos(1.38079516386  + 209.61075965840 *T)
    sumX3  +=     121.811 * cos(5.19039452998  + 198.82818395289 *T)
    sumX3  +=     112.652 * cos(4.50657670209  + 220.16882495529 *T)
    sumX3  +=     109.203 * cos(3.33018599514  + 217.47506618460 *T)
    sumX3  +=     111.589 * cos(5.41300222633  + 860.55374623631 *T)
    sumX3  +=      97.336 * cos(0.97161200873  + 95.73540973430 *T)
    sumX3  +=     114.728 * cos(1.16141976921  + 223.83785366000 *T)
    sumX3  +=      85.497 * cos(1.88744465241  + 330.86278114170 *T)
    sumX3  +=     103.793 * cos(2.96083959404  + 309.52214013930 *T)
    sumX3  +=      85.243 * cos(0.03105121395  + 654.36819779910 *T)
    sumX3  +=      81.972 * cos(3.29167838280  + 192.20227191910 *T)
    sumX3  +=      93.735 * cos(3.09844453957  + 426.35437339250 *T)
    sumX3  +=      78.232 * cos(1.55716671048  + 124.67723270450 *T)
    sumX3  +=      73.732 * cos(1.05194572850  + 88.62186273350 *T)
    sumX3  +=      68.659 * cos(3.09861369147  + 109.96250373590 *T)
    sumX3  +=      87.657 * cos(0.36385164600  + 430.02340209721 *T)
    sumX3  +=      57.790 * cos(1.70437193383  + 224.58861318540 *T)
    sumX3  +=      50.888 * cos(3.09413498180  + 202.49721265760 *T)
    sumX3  +=      48.029 * cos(3.83068266955  + 302.40859313851 *T)
    sumX3  +=      57.606 * cos(3.83723480361  + 203.24797218300 *T)
    sumX3  +=      56.960 * cos(4.32645734701  + 1066.73929467350 *T)
    sumX3  +=      45.308 * cos(3.25423824441  + 96.22304470130 *T)
    sumX3  +=      48.067 * cos(6.08559382039  + 309.03450517230 *T)
    sumX3  +=      36.732 * cos(0.80231221541  + 448.18264936190 *T)
    sumX3  +=      38.865 * cos(3.77730274172  + 625.91400979590 *T)
    sumX3  +=      43.853 * cos(0.20832317981  + 422.90985509640 *T)
    sumX3  +=      36.680 * cos(5.09355802388  + 13.98327651810 *T)
    sumX3  +=      35.412 * cos(5.54558308698  + 522.82123557730 *T)
    sumX3  +=      30.232 * cos(3.87085631365  + 412.12727939089 *T)
    sumX3  +=      34.292 * cos(6.06837760251  + 301.92095817150 *T)
    sumX3  +=      25.919 * cos(5.31578501663  + 4.17597074660 *T)
    sumX3  +=      26.527 * cos(3.52104928312  + 423.66061462181 *T)
    sumX3  +=      23.623 * cos(4.82567426056  + 515.22005360950 *T)
    sumX3  +=      17.831 * cos(5.24620312943  + 11.28951774740 *T)
    sumX3  +=      17.289 * cos(0.98528821511  + 537.04832957890 *T)
    sumX3  +=      16.881 * cos(5.38329398416  + 195.38366565681 *T)
    sumX3  +=      13.511 * cos(5.72506783696  + 515.70768857651 *T)
    sumX3  +=      12.893 * cos(1.12369483024  + 437.88770862340 *T)
    sumX3  +=      11.197 * cos(2.55648325655  + 742.74624304909 *T)
    sumX3  +=      13.451 * cos(5.19262853301  + 522.33360061030 *T)
    sumX3  +=      10.292 * cos(0.86377999055  + 337.97632814250 *T)
    sumX3  +=      11.481 * cos(5.27139699788  + 191.71463695210 *T)
    sumX3  +=      11.394 * cos(5.93878680228  + 625.42637482889 *T)
    sumX3  +=      11.380 * cos(4.58881936707  + 1073.85284167431 *T)
    sumX3  +=       9.884 * cos(0.96968966802  + 131.79077970531 *T)
    sumX3  +=       9.268 * cos(1.28597973750  + 81.50831573269 *T)
    sumX3  +=      10.374 * cos(0.94317330721  + 544.16187657970 *T)
    sumX3  +=      11.913 * cos(0.32742179509  + 437.13694909800 *T)
    sumX3  +=       8.626 * cos(3.66458258994  + 728.51914904750 *T)
    sumX3  +=       8.671 * cos(4.32466808269  + 419.24082639170 *T)
    sumX3  +=       8.821 * cos(4.57655314696  + 831.61192326610 *T)
    sumX3  +=       8.112 * cos(0.50007270362  + 149.80701461810 *T)
    sumX3  +=       9.768 * cos(5.81884437871  + 618.31282782810 *T)
    sumX3  +=       8.329 * cos(5.47343196835  + 867.66729323710 *T)
    sumX3  +=       7.645 * cos(2.39585916890  + 846.32665223470 *T)
    sumX3  +=       8.659 * cos(5.79445660573  + 643.32249753521 *T)
    sumX3  +=       7.205 * cos(4.39953882951  + 227.28237195610 *T)
    sumX3  +=       7.157 * cos(5.63667649779  + 636.20895053440 *T)
    sumX3  +=       7.036 * cos(2.63239333334  + 231.70216018621 *T)
    sumX3  +=       7.083 * cos(4.84184174066  + 89.10949770051 *T)
    sumX3  +=       6.556 * cos(2.67135559119  + 743.23387801611 *T)
    sumX3  +=       5.903 * cos(0.87375775835  + 270.16526422410 *T)
    sumX3  +=       6.615 * cos(3.63831273611  + 433.46792039329 *T)
    sumX3  +=       6.469 * cos(2.05411382166  + 408.68276109481 *T)
    sumX3  +=       5.426 * cos(1.51765079988  + 241.99710092470 *T)
    sumX3  +=       5.648 * cos(2.55416366478  + 529.93478257810 *T)
    sumX3  +=       7.105 * cos(3.01903914273  + 18.40306474821 *T)
    sumX3  +=       5.052 * cos(5.18252086135  + 416.54706762100 *T)
    sumX3  +=       6.189 * cos(4.68404979693  + 736.12033101530 *T)
    sumX3  +=       5.821 * cos(3.50616955116  + 1280.03839011150 *T)
    sumX3  +=       5.043 * cos(3.76766161323  + 277.27881122490 *T)
    sumX3  +=       4.744 * cos(1.46479508366  + 56.37853381910 *T)
    sumX3  +=       5.705 * cos(4.87015493433  + 430.77416162260 *T)
    sumX3  +=       5.226 * cos(5.29097572967  + 415.79630809560 *T)
    sumX3  +=       5.511 * cos(3.59570699682  + 529.44714761109 *T)
    sumX3  +=       4.520 * cos(2.36588968635  + 1059.13811270570 *T)
    sumX3  +=       4.583 * cos(0.00787099904  + 294.80741117069 *T)
    sumX3  +=       4.789 * cos(4.15071050239  + 295.29504613770 *T)
    sumX3  +=       4.026 * cos(0.04020873612  + 483.46435966210 *T)
    sumX3  +=       3.728 * cos(0.24380479533  + 185.08872491831 *T)
    sumX3  +=       3.578 * cos(2.51284387407  + 405.50136735710 *T)
    sumX3  +=       3.890 * cos(0.29060160988  + 750.34742501690 *T)
    sumX3  +=       4.291 * cos(5.60521207533  + 941.81824448550 *T)
    sumX3  +=       3.758 * cos(6.26818424862  + 661.48174479990 *T)
    sumX3  +=       3.547 * cos(0.40358942005  + 70.60562782070 *T)
    sumX3  +=       3.641 * cos(3.81013693999  + 839.21310523390 *T)
    sumX3  +=       4.196 * cos(3.97690464270  + 618.80046279510 *T)
    sumX3  +=       3.709 * cos(4.96509425454  + 508.10650660869 *T)
    sumX3  +=       4.193 * cos(5.05036487395  + 838.72547026689 *T)
    sumX3  +=       3.571 * cos(4.46056188614  + 1148.00379292269 *T)
    sumX3  +=       3.647 * cos(5.59898306379  + 934.70469748469 *T)
    sumX3  +=       2.978 * cos(3.43782247022  + 1044.91101870410 *T)
    sumX3  +=       2.876 * cos(5.29356968989  + 398.38782035631 *T)
    sumX3  +=       2.643 * cos(5.23820067295  + 497.69145366371 *T)
    sumX3  +=       2.725 * cos(5.89016060143  + 363.10611005610 *T)
    sumX3  +=       3.246 * cos(2.57150474119  + 639.65346883050 *T)
    sumX3  +=       2.721 * cos(4.14564904530  + 1464.39566257930 *T)
    sumX3  +=       2.917 * cos(2.38362703069  + 113.63153244060 *T)
    sumX3  +=       2.560 * cos(4.95049691984  + 729.00678401450 *T)
    sumX3  +=       2.380 * cos(6.24323945584  + 86.07111631471 *T)
    sumX3  +=       2.667 * cos(0.09012233039  + 10.80188278040 *T)
    sumX3  +=       2.238 * cos(2.26341623480  + 10.53875822200 *T)
    sumX3  +=       2.305 * cos(2.02658716121  + 956.53297345411 *T)
    sumX3  +=       2.316 * cos(2.01244422105  + 3.42521122121 *T)
    sumX3  +=       2.282 * cos(5.91510901442  + 611.19928082730 *T)
    sumX3  +=       1.988 * cos(0.28661492755  + 651.18680406140 *T)
    sumX3  +=       2.365 * cos(5.33155242941  + 2111.40649589409 *T)
    sumX3  +=       2.485 * cos(4.32833388537  + 316.14805217309 *T)
    sumX3  +=       2.670 * cos(0.63080955086  + 196.13442518220 *T)
    sumX3  +=       2.670 * cos(1.11722396213  + 230.95140066080 *T)
    sumX3  +=       1.974 * cos(3.29767644397  + 1059.62574767270 *T)
    sumX3  +=       2.440 * cos(2.29714766763  + 106.03035047280 *T)
    sumX3  +=       2.032 * cos(2.08292611476  + 1375.52998236229 *T)
    sumX3  +=       2.151 * cos(4.69823758687  + 824.49837626530 *T)
    sumX3  +=       1.875 * cos(1.50694339603  + 344.94686279140 *T)
    sumX3  +=       2.440 * cos(1.37504875250  + 323.26159917390 *T)
    sumX3  +=       2.034 * cos(5.64959382524  + 927.59115048390 *T)
    sumX3  +=       2.310 * cos(4.84647061416  + 99.66756299740 *T)
    sumX3  +=       1.764 * cos(1.86111058424  + 191.45151239371 *T)
    sumX3  +=       2.032 * cos(3.74188450688  + 636.95971005981 *T)
    sumX3  +=       1.993 * cos(5.41370582068  + 1258.21011414209 *T)
    sumX3  +=       1.921 * cos(4.30051805390  + 2641.09746098870 *T)
    sumX3  +=       1.620 * cos(5.90686570991  + 1596.43025976811 *T)
    sumX3  +=       1.903 * cos(3.15206555826  + 2634.47154895491 *T)
    sumX3  +=       1.568 * cos(3.03962897704  + 235.63431344930 *T)
    sumX3  +=       1.835 * cos(0.46838409843  + 536.56069461190 *T)
    sumX3  +=       1.523 * cos(4.48022908171  + 956.04533848709 *T)
    sumX3  +=       2.086 * cos(1.63830662169  + 445.00125562421 *T)
    sumX3  +=       1.666 * cos(1.70124586487  + 1876.76675945369 *T)
    sumX3  +=       1.479 * cos(2.77539985198  + 6290.43321447570 *T)
    sumX3  +=       1.479 * cos(5.26668359322  + 5849.12029463110 *T)
    sumX3  +=       1.439 * cos(3.22610898757  + 2538.49232173710 *T)
    sumX3  +=       1.464 * cos(4.18005026025  + 5863.34738863270 *T)
    sumX3  +=       1.464 * cos(3.86203316672  + 6276.20612047410 *T)
    sumX3  +=       1.676 * cos(5.34747324592  + 1251.09656714130 *T)
    sumX3  +=       1.945 * cos(3.48549478580  + 845.83901726770 *T)
    sumX3  +=       1.436 * cos(6.23499570787  + 508.59414157570 *T)
    sumX3  +=       1.481 * cos(4.49861202018  + 284.39235822571 *T)
    sumX3  +=       1.709 * cos(3.35714658968  + 2318.07967929830 *T)
    sumX3  +=       1.305 * cos(4.10818846677  + 107.26874496520 *T)
    sumX3  +=       1.548 * cos(2.81939196041  + 2943.74987161071 *T)
    sumX3  +=       1.465 * cos(2.76905277973  + 2531.37877473630 *T)
    sumX3  +=       1.494 * cos(3.94286053931  + 1780.78753223589 *T)
    sumX3  +=       1.243 * cos(5.53319819376  + 218.20877936750 *T)
    sumX3  +=       1.234 * cos(2.48191739575  + 208.87704647550 *T)
    sumX3  +=       1.359 * cos(1.84182551916  + 142.69346761730 *T)
    sumX3  +=       1.217 * cos(5.17100163408  + 1052.51220067191 *T)
    sumX3  +=       1.352 * cos(0.35400553171  + 209.87388421679 *T)
    sumX3  +=       1.391 * cos(5.36503473377  + 1243.98302014049 *T)
    sumX3  +=       1.426 * cos(0.53946388105  + 852.95256426850 *T)
    sumX3  +=       1.501 * cos(2.10731420473  + 1691.92185201890 *T)
    sumX3  +=       1.326 * cos(6.28017468529  + 2751.79141717511 *T)
    sumX3  +=       1.171 * cos(1.32902114795  + 67.91186905000 *T)
    sumX3  +=       1.462 * cos(1.51066849209  + 160.85271488200 *T)
    sumX3  +=       1.164 * cos(4.63302622392  + 117.07605073670 *T)
    sumX3  +=       1.337 * cos(3.11900326329  + 2627.35800195410 *T)
    sumX3  +=       1.509 * cos(4.93045095005  + 1574.60198379870 *T)
    sumX3  +=       1.213 * cos(4.31502095540  + 1354.67697632690 *T)
    sumX3  +=       1.241 * cos(4.40235607417  + 1140.89024592190 *T)
    sumX3  +=       1.288 * cos(2.87027499482  + 2950.86341861150 *T)
    sumX3  +=       1.337 * cos(3.78093252376  + 2421.17245351690 *T)
    sumX3  +=       1.327 * cos(4.98650305262  + 1567.48843679789 *T)
    sumX3  +=       1.118 * cos(3.03705756384  + 209.12312469139 *T)
    sumX3  +=       1.043 * cos(3.94898504220  + 156.92056161890 *T)
    sumX3  +=       1.069 * cos(3.08968284008  + 194.89603068979 *T)
    sumX3  +=       1.088 * cos(2.53876092916  + 1272.92484311070 *T)
    sumX3  +=       1.270 * cos(4.16462693224  + 2516.66404576769 *T)
    sumX3  +=       1.032 * cos(3.44230149136  + 1361.30288836070 *T)
    sumX3  +=       1.009 * cos(4.86702032393  + 52.44638055600 *T)
    sumX3  +=       1.100 * cos(4.02565365863  + 1671.06884598351 *T)
    sumX3  +=       1.071 * cos(2.74654865620  + 1354.18934135989 *T)
    sumX3  +=       1.120 * cos(2.93643432204  + 1037.79747170330 *T)
    sumX3  +=       1.271 * cos(1.45604360323  + 721.40560204669 *T)
    sumX3  +=       1.009 * cos(2.57702574321  + 1670.58121101650 *T)
    sumX3  +=       1.145 * cos(1.75945733161  + 3157.04896704871 *T)
    sumX3  +=       1.072 * cos(0.27481402100  + 2435.39954751850 *T)
    sumX3  +=       1.044 * cos(5.08425663596  + 1560.37488979710 *T)
    sumX3  +=       1.195 * cos(2.65932688138  + 632.53992182970 *T)
    sumX3  +=       1.043 * cos(3.11389723945  + 2729.96314120569 *T)
    sumX3  +=       0.889 * cos(0.42212522834  + 359.17395679300 *T)
    sumX3  +=       0.881 * cos(5.64403825823  + 1148.49142788971 *T)
    sumX3  +=       1.040 * cos(3.92524025620  + 1773.67398523510 *T)
    sumX3  +=       1.118 * cos(4.70881220655  + 1890.99385345530 *T)
    sumX3  +=       1.047 * cos(2.54344637059  + 3260.14174126730 *T)
    sumX3  +=       1.087 * cos(3.60856902528  + 2001.20017467470 *T)
    sumX3  +=       1.087 * cos(3.60856902527  + 2001.68780964171 *T)
    sumX3  +=       0.974 * cos(3.71347864898  + 1987.46071564010 *T)
    sumX3  +=       0.869 * cos(1.74331052904  + 621.98185653281 *T)
    sumX3  +=       1.085 * cos(2.89917981696  + 2214.98690507971 *T)
    sumX3  +=       0.841 * cos(4.92696318078  + 629.09540353360 *T)
    sumX3  +=       0.875 * cos(2.65002330643  + 2847.77064439291 *T)
    sumX3  +=       1.063 * cos(4.38870058058  + 2200.27217611110 *T)
    sumX3  +=       0.853 * cos(5.24982533241  + 288.32451148881 *T)
    sumX3  +=       1.056 * cos(1.76497744402  + 2332.30677329990 *T)
    sumX3  +=       0.888 * cos(5.03418369313  + 1155.11733992350 *T)
    sumX3  +=       1.045 * cos(4.85721663734  + 2.93757625420 *T)
    sumX3  +=       0.793 * cos(4.83325249319  + 341.01470952831 *T)
    sumX3  +=       0.817 * cos(5.39128218727  + 1802.12817323829 *T)
    sumX3  +=       0.780 * cos(5.10185484903  + 1382.64352936309 *T)
    sumX3  +=       0.798 * cos(1.50816477183  + 2015.42726867629 *T)
    sumX3  +=       0.798 * cos(4.64975742542  + 2015.91490364330 *T)
    sumX3  +=       0.834 * cos(3.90981062264  + 1802.61580820530 *T)
    sumX3  +=       0.766 * cos(5.01481205403  + 1471.50920958009 *T)
    sumX3  +=       0.875 * cos(4.28307795111  + 2523.77759276850 *T)
    sumX3  +=       0.829 * cos(2.87306260461  + 1162.23088692429 *T)
    sumX3  +=       0.749 * cos(3.65030777836  + 1809.72935520611 *T)
    sumX3  +=       0.942 * cos(1.83557306265  + 2008.31372167550 *T)
    sumX3  +=       0.730 * cos(5.63130452905  + 138.76131435421 *T)
    sumX3  +=       0.734 * cos(5.54620111084  + 749.85979004990 *T)
    sumX3  +=       0.727 * cos(2.38685285770  + 181.29958400710 *T)
    sumX3  +=       0.727 * cos(2.50277330889  + 245.78624183590 *T)
    sumX3  +=       0.986 * cos(2.62709892303  + 1795.01462623750 *T)
    sumX3  +=       0.914 * cos(3.38925031218  + 266.23311096100 *T)
    sumX3  +=       0.774 * cos(0.28234439952  + 2840.16946242510 *T)
    sumX3  +=       0.774 * cos(3.42393705311  + 2840.65709739210 *T)
    sumX3  +=       0.806 * cos(4.44962138010  + 1912.82212942470 *T)
    sumX3  +=       0.828 * cos(1.60210145147  + 100.15519796440 *T)
    sumX3  +=       0.700 * cos(0.41140636858  + 9779.32999085069 *T)
    sumX3  +=       0.700 * cos(4.19439623716  + 10220.64291069530 *T)
    sumX3  +=       0.687 * cos(4.43661173924  + 102.84895673509 *T)
    sumX3  +=       0.801 * cos(4.16672540134  + 1457.28211557850 *T)

    # SATURN X Coord
    sumX4  = 0
    sumX4  +=    7957.923 * cos(0.10292939767  + 220.65645992230 *T)
    sumX4  +=    7883.034 * cos(4.66633773762  + 206.42936592071 *T)
    sumX4  +=    1517.028 * cos(0.05577219527  + 227.77000692311 *T)
    sumX4  +=    1235.695 * cos(5.39600292264  + 433.95555536030 *T)
    sumX4  +=    1152.216 * cos(5.24396587079  + 213.54291292150 *T)
    sumX4  +=     879.616 * cos(5.34174983557  + 0.24381748350 *T)
    sumX4  +=     709.628 * cos(4.81919971142  + 199.31581891990 *T)
    sumX4  +=     602.013 * cos(1.23511363469  + 7.35736448430 *T)
    sumX4  +=     507.367 * cos(4.78792568609  + 426.84200835950 *T)
    sumX4  +=     386.382 * cos(1.30105603325  + 14.47091148511 *T)
    sumX4  +=     338.827 * cos(4.08599989607  + 419.72846135871 *T)
    sumX4  +=     186.029 * cos(5.42685623154  + 441.06910236111 *T)
    sumX4  +=     220.262 * cos(4.07756916511  + 640.14110379750 *T)
    sumX4  +=     173.210 * cos(4.46238384072  + 647.25465079831 *T)
    sumX4  +=      84.974 * cos(6.19195566472  + 234.88355392390 *T)
    sumX4  +=      85.340 * cos(2.82637979462  + 213.05527795450 *T)
    sumX4  +=      44.264 * cos(1.34743358639  + 117.56368570370 *T)
    sumX4  +=      45.849 * cos(3.59457316555  + 6.86972951729 *T)
    sumX4  +=      47.835 * cos(3.21421968827  + 853.44019923550 *T)
    sumX4  +=      41.104 * cos(3.75480601059  + 412.61491435790 *T)
    sumX4  +=      30.735 * cos(1.14073508300  + 21.58445848590 *T)
    sumX4  +=      22.639 * cos(4.54040624066  + 654.36819779910 *T)
    sumX4  +=      21.067 * cos(4.91324359260  + 103.33659170210 *T)
    sumX4  +=      20.230 * cos(1.74358717590  + 110.45013870291 *T)
    sumX4  +=      21.389 * cos(0.96555928483  + 205.94173095370 *T)
    sumX4  +=      16.854 * cos(5.95101073162  + 124.67723270450 *T)
    sumX4  +=      21.739 * cos(3.56778546142  + 860.55374623631 *T)
    sumX4  +=      17.501 * cos(2.66582523466  + 220.16882495529 *T)
    sumX4  +=      14.506 * cos(2.82815343470  + 88.62186273350 *T)
    sumX4  +=      17.401 * cos(0.69153050158  + 198.82818395289 *T)
    sumX4  +=      13.517 * cos(4.90527630680  + 309.52214013930 *T)
    sumX4  +=      11.814 * cos(0.44837969058  + 323.74923414091 *T)
    sumX4  +=      12.071 * cos(5.31212364211  + 448.18264936190 *T)
    sumX4  +=      11.725 * cos(0.04357660859  + 330.86278114170 *T)
    sumX4  +=      10.174 * cos(2.85692484188  + 95.73540973430 *T)
    sumX4  +=      11.491 * cos(5.22879894174  + 192.20227191910 *T)
    sumX4  +=      10.611 * cos(5.64300353639  + 302.40859313851 *T)
    sumX4  +=      12.147 * cos(4.23365602535  + 633.02755679670 *T)
    sumX4  +=       9.067 * cos(2.69785298236  + 13.98327651810 *T)
    sumX4  +=       8.918 * cos(1.46910827681  + 426.35437339250 *T)
    sumX4  +=       8.311 * cos(4.67543017537  + 430.02340209721 *T)
    sumX4  +=       7.296 * cos(2.34933784290  + 316.63568714010 *T)
    sumX4  +=       7.890 * cos(2.51218274353  + 1066.73929467350 *T)
    sumX4  +=       6.699 * cos(1.49841716625  + 109.96250373590 *T)
    sumX4  +=       6.339 * cos(5.32790301740  + 216.72430665921 *T)
    sumX4  +=       6.254 * cos(2.70033795328  + 210.36151918381 *T)
    sumX4  +=       6.126 * cos(1.64892191331  + 301.92095817150 *T)
    sumX4  +=       6.725 * cos(5.51198561168  + 625.91400979590 *T)
    sumX4  +=       6.784 * cos(5.78945551800  + 223.83785366000 *T)
    sumX4  +=       6.868 * cos(2.22797475448  + 203.24797218300 *T)
    sumX4  +=       6.111 * cos(5.16124898313  + 96.22304470130 *T)
    sumX4  +=       4.468 * cos(1.72149500008  + 309.03450517230 *T)
    sumX4  +=       4.160 * cos(1.03429820392  + 522.82123557730 *T)
    sumX4  +=       3.681 * cos(0.94893010872  + 423.66061462181 *T)
    sumX4  +=       3.615 * cos(0.40952028826  + 405.50136735710 *T)
    sumX4  +=       3.079 * cos(5.63647951312  + 412.12727939089 *T)
    sumX4  +=       3.614 * cos(0.37301132942  + 515.22005360950 *T)
    sumX4  +=       2.896 * cos(1.28921720639  + 515.70768857651 *T)
    sumX4  +=       3.000 * cos(4.72788705186  + 422.90985509640 *T)
    sumX4  +=       2.685 * cos(5.00319133132  + 209.61075965840 *T)
    sumX4  +=       2.918 * cos(0.73261775434  + 89.10949770051 *T)
    sumX4  +=       2.617 * cos(2.98014804776  + 217.47506618460 *T)
    sumX4  +=       2.884 * cos(0.63120048927  + 191.71463695210 *T)
    sumX4  +=       2.281 * cos(0.61985347120  + 316.14805217309 *T)
    sumX4  +=       2.717 * cos(5.35927194820  + 131.79077970531 *T)
    sumX4  +=       2.556 * cos(4.65487521363  + 337.97632814250 *T)
    sumX4  +=       1.782 * cos(6.08157929050  + 241.99710092470 *T)
    sumX4  +=       1.782 * cos(1.94963952968  + 185.08872491831 *T)
    sumX4  +=       1.856 * cos(1.22219785032  + 618.31282782810 *T)
    sumX4  +=       1.713 * cos(1.39863052168  + 625.42637482889 *T)
    sumX4  +=       1.632 * cos(5.57529252306  + 728.51914904750 *T)
    sumX4  +=       1.980 * cos(0.12344871269  + 831.61192326610 *T)
    sumX4  +=       1.607 * cos(4.53783680441  + 611.19928082730 *T)
    sumX4  +=       1.801 * cos(2.70567446780  + 227.28237195610 *T)
    sumX4  +=       1.532 * cos(0.60049717130  + 1052.51220067191 *T)
    sumX4  +=       1.510 * cos(1.01685015001  + 742.74624304909 *T)
    sumX4  +=       1.531 * cos(4.91027438516  + 529.44714761109 *T)
    sumX4  +=       1.394 * cos(2.02930508328  + 1361.30288836070 *T)
    sumX4  +=       1.427 * cos(0.73295227059  + 1251.09656714130 *T)
    sumX4  +=       1.649 * cos(5.83036907174  + 295.29504613770 *T)
    sumX4  +=       1.249 * cos(5.31804281913  + 743.23387801611 *T)
    sumX4  +=       1.353 * cos(5.64535868524  + 1678.18239298430 *T)
    sumX4  +=       1.674 * cos(0.75228934735  + 231.70216018621 *T)
    sumX4  +=       1.646 * cos(0.97572157001  + 195.38366565681 *T)
    sumX4  +=       1.305 * cos(1.00406093433  + 934.70469748469 *T)
    sumX4  +=       1.160 * cos(6.07115885491  + 81.50831573269 *T)
    sumX4  +=       1.537 * cos(0.07816179599  + 21.09682351890 *T)
    sumX4  +=       1.096 * cos(1.15801707287  + 832.09955823310 *T)
    sumX4  +=       1.156 * cos(0.08524385094  + 508.10650660869 *T)
    sumX4  +=       1.062 * cos(1.72268908914  + 294.80741117069 *T)
    sumX4  +=       1.285 * cos(5.87120310708  + 1464.39566257930 *T)
    sumX4  +=       1.079 * cos(6.04015764687  + 1368.90407032850 *T)
    sumX4  +=       1.006 * cos(0.12126175121  + 1361.79052332771 *T)
    sumX4  +=       1.290 * cos(6.26123764935  + 736.12033101530 *T)
    sumX4  +=       1.051 * cos(1.11902862643  + 941.81824448550 *T)
    sumX4  +=       1.027 * cos(2.13247533522  + 433.46792039329 *T)
    sumX4  +=       1.021 * cos(0.91271229263  + 522.33360061030 *T)
    sumX4  +=       1.137 * cos(4.47891183265  + 1891.48148842230 *T)
    sumX4  +=       0.991 * cos(5.35089033439  + 1994.57426264090 *T)
    sumX4  +=       1.038 * cos(6.06703363765  + 1148.00379292269 *T)
    sumX4  +=       0.865 * cos(1.97882029045  + 750.34742501690 *T)
    sumX4  +=       0.859 * cos(1.14491409094  + 416.54706762100 *T)
    sumX4  +=       0.862 * cos(4.75102573016  + 437.13694909800 *T)
    sumX4  +=       0.776 * cos(6.05032387495  + 102.84895673509 *T)
    sumX4  +=       0.797 * cos(0.47611886612  + 1375.52998236229 *T)
    sumX4  +=       0.824 * cos(2.96565494557  + 639.65346883050 *T)
    sumX4  +=       0.881 * cos(5.02815372820  + 1258.69774910910 *T)
    sumX4  +=       0.686 * cos(0.20449337417  + 234.39591895690 *T)
    sumX4  +=       0.684 * cos(5.09702256485  + 661.48174479990 *T)
    sumX4  +=       0.859 * cos(5.20905469515  + 537.04832957890 *T)

    # SATURN X Coord
    sumX5  = 0
    sumX5  +=     605.503 * cos(4.59646585232  + 220.65645992230 *T)
    sumX5  +=     576.439 * cos(0.06437124594  + 206.42936592071 *T)
    sumX5  +=     227.348 * cos(4.59085116990  + 227.77000692311 *T)
    sumX5  +=     133.503 * cos(3.58841633128  + 433.95555536030 *T)
    sumX5  +=      94.765 * cos(0.27215838233  + 199.31581891990 *T)
    sumX5  +=      39.851 * cos(5.86635990391  + 14.47091148511 *T)
    sumX5  +=      34.190 * cos(3.64452029703  + 441.06910236111 *T)
    sumX5  +=      24.941 * cos(5.41093732098  + 0.24381748350 *T)
    sumX5  +=      23.155 * cos(2.64207066157  + 647.25465079831 *T)
    sumX5  +=      16.208 * cos(2.39130620524  + 426.84200835950 *T)
    sumX5  +=      12.333 * cos(2.18222106376  + 640.14110379750 *T)
    sumX5  +=      10.358 * cos(4.50067908167  + 234.88355392390 *T)
    sumX5  +=      10.534 * cos(0.41421931470  + 419.72846135871 *T)
    sumX5  +=       8.668 * cos(3.51972888423  + 192.20227191910 *T)
    sumX5  +=       8.840 * cos(5.55216757332  + 213.54291292150 *T)
    sumX5  +=       4.707 * cos(4.57538541908  + 88.62186273350 *T)
    sumX5  +=       5.904 * cos(4.73107708292  + 7.35736448430 *T)
    sumX5  +=       3.895 * cos(2.30231207024  + 405.50136735710 *T)
    sumX5  +=       3.976 * cos(1.96984083585  + 6.86972951729 *T)
    sumX5  +=       3.675 * cos(0.47302446664  + 213.05527795450 *T)
    sumX5  +=       3.472 * cos(1.35605117518  + 853.44019923550 *T)
    sumX5  +=       4.393 * cos(2.75554415948  + 654.36819779910 *T)
    sumX5  +=       3.290 * cos(3.89078522206  + 124.67723270450 *T)
    sumX5  +=       2.569 * cos(5.91571375440  + 109.96250373590 *T)
    sumX5  +=       2.467 * cos(3.15983568463  + 515.70768857651 *T)
    sumX5  +=       2.708 * cos(1.73056604662  + 860.55374623631 *T)
    sumX5  +=       2.138 * cos(3.58678888073  + 448.18264936190 *T)
    sumX5  +=       2.447 * cos(0.90027079514  + 220.16882495529 *T)
    sumX5  +=       1.979 * cos(0.98519013325  + 302.40859313851 *T)
    sumX5  +=       2.238 * cos(5.57282036056  + 198.82818395289 *T)
    sumX5  +=       1.727 * cos(0.08010225619  + 301.92095817150 *T)
    sumX5  +=       1.398 * cos(1.62923220404  + 309.52214013930 *T)
    sumX5  +=       1.536 * cos(2.37774872925  + 205.94173095370 *T)
    sumX5  +=       1.195 * cos(5.50915521926  + 117.56368570370 *T)
    sumX5  +=       1.000 * cos(4.01529701865  + 223.83785366000 *T)
    sumX5  +=       0.984 * cos(3.95579224093  + 203.24797218300 *T)
    sumX5  +=       0.791 * cos(2.18561862902  + 330.86278114170 *T)
    sumX5  +=       0.979 * cos(5.20989352885  + 95.73540973430 *T)
    sumX5  +=       0.784 * cos(3.19070192516  + 522.82123557730 *T)
    sumX5  +=       0.910 * cos(2.05287501643  + 96.22304470130 *T)

    # SATURN Y Coord
    sumY0  = 0
    sumY0  +=  952312533.591 * cos(5.58640389526  + 213.54291292150 *T)
    sumY0  +=  79501390.398 * cos(3.19501080417  + 0.24381748350 *T)
    sumY0  +=  26427074.351 * cos(4.83578897954  + 426.84200835950 *T)
    sumY0  +=  6836881.382 * cos(2.57458281509  + 206.42936592071 *T)
    sumY0  +=  6628914.946 * cos(5.46296215793  + 220.65645992230 *T)
    sumY0  +=  2340967.916 * cos(0.44899651249  + 7.35736448430 *T)
    sumY0  +=  1250581.159 * cos(0.60313024847  + 110.45013870291 *T)
    sumY0  +=  1141539.711 * cos(1.46265679616  + 419.72846135871 *T)
    sumY0  +=  1098217.124 * cos(4.08641227912  + 640.14110379750 *T)
    sumY0  +=  773784.455 * cos(5.07365052128  + 213.05527795450 *T)
    sumY0  +=  708587.042 * cos(1.13843336592  + 316.63568714010 *T)
    sumY0  +=  434227.030 * cos(5.42862996601  + 529.93478257810 *T)
    sumY0  +=  373110.880 * cos(4.71342390707  + 433.95555536030 *T)
    sumY0  +=  413446.294 * cos(0.41153807474  + 102.84895673509 *T)
    sumY0  +=  169743.473 * cos(1.48210526398  + 205.94173095370 *T)
    sumY0  +=  122184.772 * cos(2.79745491359  + 103.33659170210 *T)
    sumY0  +=  141571.884 * cos(0.13687533326  + 419.24082639170 *T)
    sumY0  +=   97840.616 * cos(5.72784467548  + 323.74923414091 *T)
    sumY0  +=   80586.098 * cos(4.05159874352  + 11.28951774740 *T)
    sumY0  +=   83840.998 * cos(5.33240905202  + 227.77000692311 *T)
    sumY0  +=   72499.120 * cos(0.84921881394  + 633.02755679670 *T)
    sumY0  +=   70188.882 * cos(5.59903361411  + 209.61075965840 *T)
    sumY0  +=   68920.772 * cos(2.44584271705  + 217.47506618460 *T)
    sumY0  +=   65633.219 * cos(1.19325366411  + 202.49721265760 *T)
    sumY0  +=   58340.081 * cos(0.59027157812  + 224.58861318540 *T)
    sumY0  +=   54036.331 * cos(3.33803216964  + 853.44019923550 *T)
    sumY0  +=   45670.301 * cos(0.30742381833  + 14.47091148511 *T)
    sumY0  +=   39957.906 * cos(2.72206781342  + 199.31581891990 *T)
    sumY0  +=   44623.718 * cos(0.90093873172  + 63.49208081989 *T)
    sumY0  +=   25182.378 * cos(5.09001467212  + 216.72430665921 *T)
    sumY0  +=   24597.547 * cos(2.95757557284  + 210.36151918381 *T)
    sumY0  +=   24755.958 * cos(4.02968885174  + 415.79630809560 *T)
    sumY0  +=   25518.626 * cos(0.06734394180  + 117.56368570370 *T)
    sumY0  +=   30420.785 * cos(6.15115633926  + 735.63269604830 *T)
    sumY0  +=   22261.024 * cos(1.10100301594  + 522.33360061030 *T)
    sumY0  +=   23384.603 * cos(3.96374854662  + 647.25465079831 *T)
    sumY0  +=   19921.860 * cos(4.09518373122  + 426.35437339250 *T)
    sumY0  +=   23368.080 * cos(4.90040735801  + 149.80701461810 *T)
    sumY0  +=   24816.783 * cos(3.92964610129  + 75.02541605080 *T)
    sumY0  +=   15057.730 * cos(3.10557746427  + 277.27881122490 *T)
    sumY0  +=   13291.291 * cos(2.35905808760  + 309.03450517230 *T)
    sumY0  +=   12028.365 * cos(5.74173457069  + 1059.62574767270 *T)
    sumY0  +=   12408.238 * cos(2.62612862596  + 490.57790666290 *T)
    sumY0  +=   12037.260 * cos(4.09279303186  + 352.06040979221 *T)
    sumY0  +=    7880.259 * cos(1.03244732130  + 743.23387801611 *T)
    sumY0  +=    7428.823 * cos(5.83507395250  + 6.86972951729 *T)
    sumY0  +=    6985.779 * cos(2.31301163514  + 309.52214013930 *T)
    sumY0  +=    9059.376 * cos(5.88705982112  + 1052.02456570489 *T)
    sumY0  +=    6626.180 * cos(0.82428012166  + 838.72547026689 *T)
    sumY0  +=    8285.289 * cos(3.60108497690  + 95.73540973430 *T)
    sumY0  +=    7908.600 * cos(4.48713264921  + 316.14805217309 *T)
    sumY0  +=    6284.835 * cos(2.67337959503  + 3.68833577960 *T)
    sumY0  +=    6078.527 * cos(4.58792252731  + 441.06910236111 *T)
    sumY0  +=    5628.724 * cos(0.90747309129  + 846.32665223470 *T)
    sumY0  +=    5274.108 * cos(4.97367799615  + 220.16882495529 *T)
    sumY0  +=    5584.926 * cos(1.54150891449  + 412.61491435790 *T)
    sumY0  +=    4423.799 * cos(1.57641344464  + 198.82818395289 *T)
    sumY0  +=    4849.319 * cos(4.98130878652  + 537.04832957890 *T)
    sumY0  +=    4179.381 * cos(5.16756812825  + 4.17597074660 *T)
    sumY0  +=    5772.348 * cos(3.73707335017  + 38.37685312130 *T)
    sumY0  +=    5198.014 * cos(2.97493810590  + 211.09523236671 *T)
    sumY0  +=    5154.525 * cos(5.05852684698  + 215.99059347631 *T)
    sumY0  +=    3792.407 * cos(4.42246502754  + 422.90985509640 *T)
    sumY0  +=    4380.719 * cos(2.72775866401  + 202.00957769060 *T)
    sumY0  +=    4838.174 * cos(3.60482713264  + 109.96250373590 *T)
    sumY0  +=    3765.792 * cos(2.14993736411  + 212.57970507500 *T)
    sumY0  +=    3749.721 * cos(5.88164710896  + 214.50612076800 *T)
    sumY0  +=    3116.757 * cos(5.55835071261  + 213.49472881170 *T)
    sumY0  +=    3115.849 * cos(2.47295232173  + 213.59109703130 *T)
    sumY0  +=    3119.049 * cos(2.08030449144  + 625.42637482889 *T)
    sumY0  +=    3235.633 * cos(2.93645048734  + 632.53992182970 *T)
    sumY0  +=    4077.069 * cos(1.24694613809  + 136.78920667889 *T)
    sumY0  +=    3290.391 * cos(5.03950922540  + 522.82123557730 *T)
    sumY0  +=    2828.736 * cos(5.44462827617  + 330.86278114170 *T)
    sumY0  +=    3365.297 * cos(6.13313657202  + 437.88770862340 *T)
    sumY0  +=    2888.976 * cos(5.85305165742  + 86.07111631471 *T)
    sumY0  +=    2748.221 * cos(3.85851556523  + 138.76131435421 *T)
    sumY0  +=    2742.545 * cos(5.64786903747  + 203.98168536590 *T)
    sumY0  +=    2636.395 * cos(3.83621912362  + 288.32451148881 *T)
    sumY0  +=    3170.535 * cos(4.19556924462  + 76.50988875911 *T)
    sumY0  +=    2530.060 * cos(3.52650884327  + 629.09540353360 *T)
    sumY0  +=    2985.976 * cos(2.58564693197  + 1066.73929467350 *T)
    sumY0  +=    2617.358 * cos(0.73922008905  + 127.22797912329 *T)
    sumY0  +=    2504.577 * cos(6.20308081313  + 9.80504503910 *T)
    sumY0  +=    2459.199 * cos(0.55192655633  + 1155.11733992350 *T)
    sumY0  +=    2429.561 * cos(2.40176973016  + 223.10414047710 *T)
    sumY0  +=    3133.467 * cos(5.62811439341  + 1368.41643536150 *T)
    sumY0  +=    2289.072 * cos(2.11564630370  + 430.77416162260 *T)
    sumY0  +=    2651.422 * cos(0.30311808414  + 412.12727939089 *T)
    sumY0  +=    2791.618 * cos(0.19025345990  + 3.42521122121 *T)
    sumY0  +=    2086.473 * cos(1.97187121386  + 423.66061462181 *T)
    sumY0  +=    2337.517 * cos(4.29665202435  + 388.70897272171 *T)
    sumY0  +=    1735.774 * cos(4.10517039399  + 736.12033101530 *T)
    sumY0  +=    1617.760 * cos(2.87094929008  + 96.22304470130 *T)
    sumY0  +=    1635.420 * cos(2.97359194678  + 12.77399045571 *T)
    sumY0  +=    1637.578 * cos(0.62856200162  + 213.02164805970 *T)
    sumY0  +=    1633.497 * cos(1.12064121841  + 214.06417778330 *T)
    sumY0  +=    1703.439 * cos(3.99674710141  + 529.44714761109 *T)
    sumY0  +=    1499.174 * cos(5.17056778089  + 52.44638055600 *T)
    sumY0  +=    1462.063 * cos(4.35331374641  + 430.02340209721 *T)
    sumY0  +=    1975.850 * cos(0.45234024897  + 1.72829019180 *T)
    sumY0  +=    1422.179 * cos(3.32899817186  + 625.91400979590 *T)
    sumY0  +=    1539.334 * cos(3.21457103113  + 860.55374623631 *T)
    sumY0  +=    1400.317 * cos(2.15177170721  + 341.01470952831 *T)
    sumY0  +=    1337.694 * cos(1.54600928268  + 948.93179148630 *T)
    sumY0  +=    1427.589 * cos(3.84316796768  + 350.57593708390 *T)
    sumY0  +=    1149.795 * cos(1.77484346800  + 941.81824448550 *T)
    sumY0  +=    1112.160 * cos(5.19130278223  + 234.88355392390 *T)
    sumY0  +=    1015.231 * cos(0.28360221027  + 1471.50920958009 *T)
    sumY0  +=    1090.238 * cos(4.25624714992  + 363.10611005610 *T)
    sumY0  +=    1091.460 * cos(5.84943013311  + 174.18603700630 *T)
    sumY0  +=    1152.320 * cos(2.26380260884  + 201.01273994931 *T)
    sumY0  +=     874.317 * cos(1.24749634822  + 515.22005360950 *T)
    sumY0  +=    1180.011 * cos(5.37502087961  + 1684.80830501809 *T)
    sumY0  +=    1058.128 * cos(2.86921511069  + 742.74624304909 *T)
    sumY0  +=     850.822 * cos(2.98216392449  + 949.41942645330 *T)
    sumY0  +=     869.178 * cos(1.66448567873  + 210.62215161470 *T)
    sumY0  +=    1119.594 * cos(1.89511570784  + 703.87700210090 *T)
    sumY0  +=     868.384 * cos(0.08058763351  + 216.46367422830 *T)
    sumY0  +=     830.815 * cos(5.29349088148  + 212.79215339611 *T)
    sumY0  +=     825.441 * cos(2.73642207174  + 214.29367244690 *T)
    sumY0  +=     881.196 * cos(6.18145692514  + 209.35012722751 *T)
    sumY0  +=     866.345 * cos(1.84353188382  + 217.73569861551 *T)
    sumY0  +=     789.515 * cos(4.95200139776  + 223.83785366000 *T)
    sumY0  +=    1091.862 * cos(4.24426470278  + 340.52707456129 *T)
    sumY0  +=    1008.329 * cos(5.80156759687  + 226.07308589371 *T)
    sumY0  +=     919.088 * cos(3.29084137367  + 565.35950523021 *T)
    sumY0  +=     820.108 * cos(3.76565710375  + 88.62186273350 *T)
    sumY0  +=     716.162 * cos(1.09994318200  + 1265.32366114289 *T)
    sumY0  +=     860.928 * cos(2.13534949323  + 252.89978883670 *T)
    sumY0  +=     710.939 * cos(1.84269456164  + 417.28078080390 *T)
    sumY0  +=     701.898 * cos(0.34872426337  + 956.53297345411 *T)
    sumY0  +=     918.630 * cos(3.00611004376  + 563.87503252191 *T)
    sumY0  +=     711.475 * cos(5.61815057873  + 207.91383862900 *T)
    sumY0  +=     797.153 * cos(0.44389764826  + 208.12628695010 *T)
    sumY0  +=     758.732 * cos(0.06203142018  + 728.51914904750 *T)
    sumY0  +=     642.018 * cos(3.49772973344  + 639.65346883050 *T)
    sumY0  +=     769.629 * cos(1.32348354837  + 218.95953889290 *T)
    sumY0  +=     811.029 * cos(4.65167271285  + 1162.71852189130 *T)
    sumY0  +=     692.909 * cos(2.40594407531  + 219.17198721400 *T)
    sumY0  +=     798.757 * cos(6.06663707020  + 22.33521801130 *T)
    sumY0  +=     834.309 * cos(1.39164707042  + 160.85271488200 *T)
    sumY0  +=     655.277 * cos(2.49159323627  + 301.92095817150 *T)
    sumY0  +=     762.632 * cos(0.15014814295  + 21.58445848590 *T)
    sumY0  +=     576.036 * cos(1.53086525589  + 99.40443843901 *T)
    sumY0  +=     615.654 * cos(3.84630379584  + 845.83901726770 *T)
    sumY0  +=     578.589 * cos(2.37484813857  + 265.74547599400 *T)
    sumY0  +=     626.782 * cos(1.22691907986  + 18.40306474821 *T)
    sumY0  +=     589.499 * cos(2.33660635603  + 554.31380496631 *T)
    sumY0  +=     700.182 * cos(1.97098448211  + 62.00760811160 *T)
    sumY0  +=     697.691 * cos(5.28344445950  + 124.67723270450 *T)
    sumY0  +=     613.843 * cos(2.85159863228  + 218.20877936750 *T)
    sumY0  +=     581.263 * cos(0.50791540059  + 231.70216018621 *T)
    sumY0  +=     666.749 * cos(1.25289826631  + 425.35753565121 *T)
    sumY0  +=     516.903 * cos(5.10096101449  + 414.31183538731 *T)
    sumY0  +=     499.031 * cos(0.96959623173  + 408.68276109481 *T)
    sumY0  +=     502.357 * cos(1.48531352820  + 1258.21011414209 *T)
    sumY0  +=     531.983 * cos(0.96050577649  + 215.02738562980 *T)
    sumY0  +=     643.415 * cos(2.96350458546  + 64.97655352819 *T)
    sumY0  +=     518.385 * cos(1.12922566079  + 142.69346761730 *T)
    sumY0  +=     550.972 * cos(5.18286340719  + 208.87704647550 *T)
    sumY0  +=     469.313 * cos(5.04169316833  + 71.09326278771 *T)
    sumY0  +=     456.002 * cos(3.10039424711  + 203.24797218300 *T)
    sumY0  +=     592.362 * cos(0.29709684601  + 266.23311096100 *T)
    sumY0  +=     465.459 * cos(3.83732069410  + 654.36819779910 *T)
    sumY0  +=     444.956 * cos(0.01953353182  + 1787.90107923670 *T)
    sumY0  +=     502.683 * cos(2.79966508043  + 251.67594855930 *T)
    sumY0  +=     522.764 * cos(0.62622137139  + 212.05844021321 *T)
    sumY0  +=     422.458 * cos(0.56485405564  + 479.53220639900 *T)
    sumY0  +=     437.134 * cos(6.04075422758  + 1589.31671276730 *T)
    sumY0  +=     414.039 * cos(2.06469019925  + 148.32254190981 *T)
    sumY0  +=     398.994 * cos(0.07893800573  + 213.65478750610 *T)
    sumY0  +=     398.994 * cos(1.66909550727  + 213.43103833690 *T)
    sumY0  +=     373.088 * cos(0.74771686052  + 1581.71553079950 *T)
    sumY0  +=     469.097 * cos(5.12816987718  + 2001.20017467470 *T)
    sumY0  +=     361.258 * cos(3.33195222747  + 302.40859313851 *T)
    sumY0  +=     345.920 * cos(3.44207098800  + 198.56505939450 *T)
    sumY0  +=     350.777 * cos(1.14840197831  + 98.65617104110 *T)
    sumY0  +=     335.274 * cos(5.74248127061  + 213.75536124260 *T)
    sumY0  +=     335.274 * cos(2.28873754957  + 213.33046460040 *T)
    sumY0  +=     367.346 * cos(4.34008847005  + 750.34742501690 *T)
    sumY0  +=     369.973 * cos(3.82853437800  + 151.29148732640 *T)
    sumY0  +=     399.230 * cos(2.61912746358  + 1059.13811270570 *T)
    sumY0  +=     318.144 * cos(4.57955753940  + 228.52076644850 *T)
    sumY0  +=     315.586 * cos(0.97485635422  + 831.61192326610 *T)
    sumY0  +=     337.834 * cos(5.38036984568  + 175.40987728371 *T)
    sumY0  +=     313.433 * cos(5.11627418422  + 2.20386307129 *T)
    sumY0  +=     330.040 * cos(2.95117982130  + 276.79117625789 *T)
    sumY0  +=     383.783 * cos(4.14240556966  + 56.37853381910 *T)
    sumY0  +=     350.135 * cos(3.18895815490  + 114.38229196601 *T)
    sumY0  +=     401.239 * cos(2.71851293063  + 178.11819026941 *T)
    sumY0  +=     404.631 * cos(4.46617284845  + 1479.11039154791 *T)
    sumY0  +=     404.539 * cos(3.09389129845  + 311.95526647910 *T)
    sumY0  +=     304.876 * cos(4.54455856496  + 195.38366565681 *T)
    sumY0  +=     289.344 * cos(2.91199684314  + 192.20227191910 *T)
    sumY0  +=     270.250 * cos(3.56955603971  + 636.20895053440 *T)
    sumY0  +=     271.661 * cos(6.00173537924  + 278.76328393321 *T)
    sumY0  +=     283.573 * cos(4.75629059021  + 191.45151239371 *T)
    sumY0  +=     294.112 * cos(3.54847592417  + 1162.23088692429 *T)
    sumY0  +=     266.932 * cos(4.08413714120  + 205.46615807421 *T)
    sumY0  +=     269.475 * cos(0.57518354804  + 327.68138740401 *T)
    sumY0  +=     348.053 * cos(5.33129565512  + 248.96763557360 *T)
    sumY0  +=     296.599 * cos(6.22764523323  + 10.53875822200 *T)
    sumY0  +=     252.084 * cos(0.10475870159  + 213.80354535241 *T)
    sumY0  +=     252.084 * cos(1.64327481141  + 213.28228049061 *T)
    sumY0  +=     250.282 * cos(4.18735452982  + 433.46792039329 *T)
    sumY0  +=     294.321 * cos(0.44790526962  + 106.51798543980 *T)
    sumY0  +=     270.973 * cos(6.10541688721  + 1044.91101870410 *T)
    sumY0  +=     263.820 * cos(2.19241095861  + 63.97971578690 *T)
    sumY0  +=     238.275 * cos(1.20077577792  + 1574.60198379870 *T)
    sumY0  +=     305.244 * cos(5.81768185014  + 483.46435966210 *T)
    sumY0  +=     241.411 * cos(4.17791347646  + 221.61966776881 *T)
    sumY0  +=     314.980 * cos(5.94135752572  + 479.04457143200 *T)
    sumY0  +=     223.277 * cos(5.58851274462  + 13.98327651810 *T)
    sumY0  +=     292.846 * cos(2.29297335501  + 424.39432780471 *T)
    sumY0  +=     304.942 * cos(0.18281299570  + 6283.31966747490 *T)
    sumY0  +=     237.167 * cos(3.38939567774  + 235.63431344930 *T)
    sumY0  +=     240.608 * cos(4.17952400464  + 275.79433851660 *T)
    sumY0  +=     263.455 * cos(0.79817206059  + 70.60562782070 *T)
    sumY0  +=     211.236 * cos(4.73393150820  + 544.16187657970 *T)
    sumY0  +=     205.878 * cos(2.89475990441  + 842.39449897160 *T)
    sumY0  +=     282.395 * cos(1.99518902252  + 404.75060783170 *T)
    sumY0  +=     203.077 * cos(6.04307654731  + 2104.29294889330 *T)
    sumY0  +=     267.955 * cos(1.92479423159  + 121.49583896680 *T)
    sumY0  +=     238.411 * cos(0.28536423644  + 138.27367938720 *T)
    sumY0  +=     230.884 * cos(2.46621440966  + 497.69145366371 *T)
    sumY0  +=     203.254 * cos(1.36588190624  + 284.39235822571 *T)
    sumY0  +=     216.423 * cos(0.87035300847  + 8.32057233081 *T)
    sumY0  +=     194.279 * cos(0.56631831282  + 190.96387742670 *T)
    sumY0  +=     216.822 * cos(4.57805495139  + 429.28968891431 *T)
    sumY0  +=     191.411 * cos(0.43155331598  + 1898.10740045609 *T)
    sumY0  +=     212.655 * cos(5.38657458551  + 651.18680406140 *T)
    sumY0  +=     214.969 * cos(1.40674231770  + 425.87880051300 *T)
    sumY0  +=     236.822 * cos(2.00313512044  + 219.69325207580 *T)
    sumY0  +=     229.348 * cos(0.08407966782  + 270.16526422410 *T)
    sumY0  +=     174.163 * cos(1.72024360668  + 426.89019246930 *T)
    sumY0  +=     202.398 * cos(0.85428604723  + 39.60069339870 *T)
    sumY0  +=     179.962 * cos(4.05892479085  + 404.26297286470 *T)
    sumY0  +=     169.610 * cos(3.20964722020  + 501.62360692681 *T)
    sumY0  +=     174.157 * cos(4.80563409005  + 426.79382424970 *T)
    sumY0  +=     165.778 * cos(5.47337555700  + 312.70353387701 *T)
    sumY0  +=     178.798 * cos(2.73599948703  + 355.99256305531 *T)
    sumY0  +=     180.904 * cos(2.20867391842  + 618.31282782810 *T)
    sumY0  +=     182.548 * cos(5.95923969430  + 210.83459993580 *T)
    sumY0  +=     179.529 * cos(2.07197132894  + 216.25122590720 *T)
    sumY0  +=     187.071 * cos(5.22164350587  + 427.80521620600 *T)
    sumY0  +=     180.856 * cos(1.50953564487  + 84.58664360640 *T)
    sumY0  +=     196.496 * cos(3.69598710460  + 489.09343395460 *T)
    sumY0  +=     151.676 * cos(5.46446571963  + 213.38285422711 *T)
    sumY0  +=     151.676 * cos(2.56675310055  + 213.70297161591 *T)
    sumY0  +=     151.430 * cos(2.33433917622  + 220.70464403210 *T)
    sumY0  +=     193.773 * cos(4.88770212559  + 2317.59204433130 *T)
    sumY0  +=     166.996 * cos(1.83615762317  + 1280.03839011150 *T)
    sumY0  +=     189.342 * cos(6.03878169201  + 207.39257376720 *T)
    sumY0  +=     180.605 * cos(4.76194471156  + 492.06237937121 *T)
    sumY0  +=     172.720 * cos(4.16449693176  + 1052.51220067191 *T)
    sumY0  +=     149.398 * cos(3.04588193276  + 839.21310523390 *T)
    sumY0  +=     146.617 * cos(5.44168689797  + 220.60827581251 *T)
    sumY0  +=     193.842 * cos(2.62646156746  + 189.16389053330 *T)
    sumY0  +=     178.668 * cos(1.03930372836  + 344.94686279140 *T)
    sumY0  +=     137.031 * cos(0.70922699806  + 1148.00379292269 *T)
    sumY0  +=     185.345 * cos(3.55326872913  + 602.00806815971 *T)
    sumY0  +=     158.449 * cos(1.85333938109  + 179.60266297770 *T)
    sumY0  +=     167.315 * cos(1.77379352575  + 73.54094334250 *T)
    sumY0  +=     155.890 * cos(3.42134683788  + 289.80898419710 *T)
    sumY0  +=     159.567 * cos(3.96486064673  + 359.17395679300 *T)
    sumY0  +=     138.554 * cos(1.62845102038  + 436.40323591510 *T)
    sumY0  +=     164.796 * cos(2.37461935463  + 1375.52998236229 *T)
    sumY0  +=     167.034 * cos(3.18245251049  + 142.20583265029 *T)
    sumY0  +=     128.473 * cos(4.23682984484  + 107.26874496520 *T)
    sumY0  +=     142.270 * cos(1.18092220136  + 636.95971005981 *T)
    sumY0  +=     177.375 * cos(1.60572975981  + 10213.52936369450 *T)
    sumY0  +=     125.618 * cos(1.77143221890  + 644.07325706060 *T)
    sumY0  +=     131.798 * cos(0.93003878585  + 212.27088853431 *T)
    sumY0  +=     130.711 * cos(0.81771131470  + 214.81493730871 *T)
    sumY0  +=     122.905 * cos(0.15451148163  + 621.98185653281 *T)
    sumY0  +=     165.088 * cos(5.37947744065  + 237.92193530970 *T)
    sumY0  +=     133.438 * cos(4.08482767891  + 4.90968392950 *T)
    sumY0  +=     120.114 * cos(0.01590735638  + 194.89603068979 *T)
    sumY0  +=     161.862 * cos(4.23732093970  + 1795.50226120450 *T)
    sumY0  +=     116.787 * cos(0.03950282293  + 113.63153244060 *T)
    sumY0  +=     118.210 * cos(0.92046523686  + 1890.99385345530 *T)
    sumY0  +=     133.448 * cos(3.13288930099  + 1478.62275658089 *T)
    sumY0  +=     111.758 * cos(5.70722465416  + 206.38118181090 *T)
    sumY0  +=     152.292 * cos(5.28177714601  + 515.70768857651 *T)
    sumY0  +=     128.840 * cos(6.17183070499  + 247.48316286530 *T)
    sumY0  +=     110.215 * cos(3.65473479864  + 5.87289177600 *T)
    sumY0  +=     107.434 * cos(1.52523203549  + 216.98743121760 *T)
    sumY0  +=     115.948 * cos(5.37627907087  + 31.99951143090 *T)
    sumY0  +=     109.063 * cos(3.90089614511  + 114.64292439690 *T)
    sumY0  +=     109.273 * cos(5.46080628706  + 127.71561409031 *T)
    sumY0  +=     107.429 * cos(0.24319756091  + 767.61290040431 *T)
    sumY0  +=     126.160 * cos(2.80031308667  + 181.29958400710 *T)
    sumY0  +=     110.176 * cos(5.75314210788  + 10.80188278040 *T)
    sumY0  +=     123.129 * cos(3.51949079773  + 2.93757625420 *T)
    sumY0  +=     134.189 * cos(1.19256728733  + 35.66854013560 *T)
    sumY0  +=     106.815 * cos(2.59220497594  + 206.47755003051 *T)
    sumY0  +=     105.829 * cos(6.16888137037  + 220.13519506050 *T)
    sumY0  +=     115.432 * cos(3.21326490060  + 415.30867312860 *T)
    sumY0  +=     118.293 * cos(5.23757484150  + 245.78624183590 *T)
    sumY0  +=     104.354 * cos(2.46236691161  + 1073.85284167431 *T)
    sumY0  +=      95.114 * cos(3.59922480574  + 643.32249753521 *T)
    sumY0  +=      99.528 * cos(6.13536892918  + 426.32074349770 *T)
    sumY0  +=      97.935 * cos(0.13297617812  + 2214.49927011270 *T)
    sumY0  +=     106.628 * cos(2.38110021899  + 39.11305843170 *T)
    sumY0  +=      94.853 * cos(5.78853543731  + 2420.68481854989 *T)
    sumY0  +=      96.075 * cos(4.45602113809  + 448.18264936190 *T)
    sumY0  +=     104.203 * cos(5.87948764743  + 1361.30288836070 *T)
    sumY0  +=     107.375 * cos(0.82410431375  + 78.95756931390 *T)
    sumY0  +=      93.096 * cos(3.69507490938  + 209.12312469139 *T)
    sumY0  +=      88.042 * cos(3.04547960580  + 67.42423408300 *T)
    sumY0  +=      93.076 * cos(0.75961828465  + 69.39634175830 *T)
    sumY0  +=      89.875 * cos(3.92412155660  + 134.82916109110 *T)
    sumY0  +=      86.729 * cos(5.90773062067  + 0.71939036300 *T)
    sumY0  +=      88.055 * cos(4.76797627280  + 87.55558902300 *T)
    sumY0  +=      86.045 * cos(2.90879658616  + 525.74199688410 *T)
    sumY0  +=      84.186 * cos(5.39649594615  + 144.39038864671 *T)
    sumY0  +=      81.866 * cos(1.67950712146  + 20.85074530300 *T)
    sumY0  +=      98.341 * cos(6.24954329392  + 4.43660317750 *T)
    sumY0  +=     112.464 * cos(3.71178482485  + 98.91680347199 *T)
    sumY0  +=      91.321 * cos(0.36728145080  + 427.36327322130 *T)
    sumY0  +=      88.237 * cos(1.89799508252  + 209.87388421679 *T)
    sumY0  +=      91.544 * cos(2.02640978009  + 206.95063078250 *T)
    sumY0  +=      79.761 * cos(3.26393184846  + 159.36824217371 *T)
    sumY0  +=     100.440 * cos(1.97801735461  + 140.24578706250 *T)
    sumY0  +=      78.269 * cos(1.67274038217  + 14.25846316400 *T)
    sumY0  +=      93.714 * cos(4.41010592813  + 203.49405039889 *T)
    sumY0  +=      77.457 * cos(1.74688331628  + 191.71463695210 *T)
    sumY0  +=      87.870 * cos(0.23216205664  + 342.49918223660 *T)
    sumY0  +=      80.557 * cos(1.80525950382  + 125.74350641500 *T)
    sumY0  +=      82.279 * cos(5.63889998343  + 91.80325647120 *T)
    sumY0  +=     100.767 * cos(3.77373570610  + 200.52510498229 *T)
    sumY0  +=     102.819 * cos(1.35532997364  + 617.56206830270 *T)
    sumY0  +=      74.922 * cos(4.60966309407  + 421.21293406700 *T)
    sumY0  +=      74.704 * cos(5.40668949066  + 1582.20316576650 *T)
    sumY0  +=      75.306 * cos(0.45482399227  + 1376.01761732930 *T)
    sumY0  +=      71.397 * cos(0.88773586563  + 630.57987624190 *T)
    sumY0  +=      70.299 * cos(4.25800150348  + 5.66044345490 *T)
    sumY0  +=      95.210 * cos(0.47362825119  + 1272.92484311070 *T)
    sumY0  +=      69.326 * cos(3.62451064029  + 214.76675319891 *T)
    sumY0  +=      69.326 * cos(4.40670817988  + 212.31907264410 *T)
    sumY0  +=      76.148 * cos(1.91892162118  + 934.70469748469 *T)
    sumY0  +=      80.760 * cos(1.04210925773  + 15.22167101050 *T)
    sumY0  +=      85.274 * cos(3.18889922249  + 53.93085326429 *T)
    sumY0  +=      79.268 * cos(2.69111606131  + 153.73916788121 *T)
    sumY0  +=      81.890 * cos(0.02760055667  + 353.54488250050 *T)
    sumY0  +=      91.253 * cos(2.24804950980  + 777.17412795991 *T)
    sumY0  +=      70.965 * cos(5.64481654456  + 1265.81129610991 *T)
    sumY0  +=      86.364 * cos(1.15543254121  + 917.17609753890 *T)
    sumY0  +=      81.660 * cos(1.66400680316  + 5.17280848790 *T)
    sumY0  +=      66.327 * cos(4.84771375413  + 227.28237195610 *T)
    sumY0  +=      66.404 * cos(0.77922087670  + 423.92124705270 *T)
    sumY0  +=      68.065 * cos(4.68958727466  + 692.83130183700 *T)
    sumY0  +=      81.793 * cos(4.65388189938  + 2633.98391398789 *T)
    sumY0  +=      70.972 * cos(4.61529282078  + 211.84599189210 *T)
    sumY0  +=      65.212 * cos(3.24754078518  + 6.39415663780 *T)
    sumY0  +=      64.142 * cos(0.44335094646  + 1464.39566257930 *T)
    sumY0  +=      63.961 * cos(4.99467573586  + 185.23173727020 *T)
    sumY0  +=      65.617 * cos(5.74394773491  + 542.78046973539 *T)
    sumY0  +=      62.731 * cos(1.78212439397  + 320.56784040320 *T)
    sumY0  +=      73.401 * cos(6.13861817689  + 286.84003878050 *T)
    sumY0  +=      63.416 * cos(1.27570290327  + 3.16457879030 *T)
    sumY0  +=      61.908 * cos(2.18774926262  + 329.48137429739 *T)
    sumY0  +=      77.192 * cos(5.91711898730  + 46.71424039951 *T)
    sumY0  +=      70.261 * cos(4.73983524718  + 337.97632814250 *T)
    sumY0  +=      58.935 * cos(4.66439163495  + 128.71245183160 *T)
    sumY0  +=      59.359 * cos(6.25919160870  + 50.96190784770 *T)
    sumY0  +=      71.174 * cos(2.51498593080  + 74.53778108379 *T)
    sumY0  +=      59.979 * cos(0.64955132149  + 2207.38572311189 *T)
    sumY0  +=      78.873 * cos(4.61588742653  + 299.37021175271 *T)
    sumY0  +=      60.704 * cos(5.45064207796  + 182.52342428450 *T)
    sumY0  +=      63.502 * cos(0.81917620362  + 391.41728570740 *T)
    sumY0  +=      60.272 * cos(1.95975211492  + 416.54706762100 *T)
    sumY0  +=      56.662 * cos(4.74365521929  + 267.71758366931 *T)
    sumY0  +=      66.008 * cos(2.50966062860  + 778.65860066820 *T)
    sumY0  +=      66.706 * cos(0.99016499310  + 221.17772478411 *T)
    sumY0  +=      66.363 * cos(4.88120491917  + 696.76345510010 *T)
    sumY0  +=      55.245 * cos(4.21009096686  + 437.13694909800 *T)
    sumY0  +=      55.098 * cos(0.01129968609  + 387.48513244430 *T)
    sumY0  +=      71.893 * cos(2.14697008774  + 1691.92185201890 *T)
    sumY0  +=      59.023 * cos(3.40769286148  + 215.23983395090 *T)
    sumY0  +=      54.012 * cos(4.61839578660  + 627.61093082531 *T)
    sumY0  +=      53.837 * cos(2.95872345426  + 241.85408857280 *T)
    sumY0  +=      63.323 * cos(3.50833483571  + 576.40520549410 *T)
    sumY0  +=      67.784 * cos(4.97172129826  + 116.18227885939 *T)
    sumY0  +=      68.327 * cos(2.75501396928  + 1795.01462623750 *T)
    sumY0  +=      74.859 * cos(4.10679868656  + 2.69149803831 *T)
    sumY0  +=      54.888 * cos(4.16961696361  + 312.44290144610 *T)
    sumY0  +=      54.393 * cos(6.06284167082  + 134.34152612410 *T)
    sumY0  +=      56.250 * cos(5.45039155171  + 273.34665796181 *T)
    sumY0  +=      56.864 * cos(6.02030359291  + 445.00125562421 *T)
    sumY0  +=      58.600 * cos(1.06477393813  + 323.26159917390 *T)
    sumY0  +=      50.765 * cos(0.00007477989  + 224.10097821840 *T)
    sumY0  +=      51.202 * cos(0.35404077029  + 7.30918037451 *T)
    sumY0  +=      52.403 * cos(5.93733244108  + 518.88908231420 *T)
    sumY0  +=      64.155 * cos(4.15556147324  + 422.17614191350 *T)
    sumY0  +=      50.427 * cos(4.44903394289  + 214.97920152001 *T)
    sumY0  +=      50.427 * cos(3.58218487728  + 212.10662432301 *T)
    sumY0  +=      62.900 * cos(5.47313323353  + 3.21276290011 *T)
    sumY0  +=      55.902 * cos(5.39313956478  + 1.24065522479 *T)
    sumY0  +=      50.047 * cos(4.12482973047  + 292.25666475190 *T)
    sumY0  +=      49.142 * cos(1.88444022876  + 618.04970326970 *T)
    sumY0  +=      55.589 * cos(3.43163971575  + 221.40721944771 *T)
    sumY0  +=      50.005 * cos(6.13525731257  + 2530.89113976930 *T)
    sumY0  +=      59.759 * cos(5.04563353956  + 439.37218133171 *T)
    sumY0  +=      47.716 * cos(5.10534747703  + 219.90570039690 *T)
    sumY0  +=      63.636 * cos(3.99701508899  + 2111.89413086110 *T)
    sumY0  +=      48.821 * cos(2.69034505764  + 30.77567115350 *T)
    sumY0  +=      49.427 * cos(2.53636090939  + 282.69543719630 *T)
    sumY0  +=      64.684 * cos(6.11825697845  + 305.58998687620 *T)
    sumY0  +=      49.622 * cos(1.07398042072  + 431.03479405351 *T)
    sumY0  +=      58.187 * cos(1.28266336916  + 348.12825652910 *T)
    sumY0  +=      48.661 * cos(1.59473341317  + 97.11681657861 *T)
    sumY0  +=      48.710 * cos(5.60640784947  + 429.76276966630 *T)
    sumY0  +=      65.208 * cos(4.93613007276  + 487.12132627930 *T)
    sumY0  +=      46.209 * cos(4.54083775787  + 426.09124883410 *T)
    sumY0  +=      58.609 * cos(6.24950229981  + 1169.83206889211 *T)
    sumY0  +=      55.906 * cos(1.01117086199  + 418.76525351221 *T)
    sumY0  +=      51.943 * cos(3.56111829834  + 199.52826724100 *T)
    sumY0  +=      54.149 * cos(4.38424746311  + 729.00678401450 *T)
    sumY0  +=      44.466 * cos(5.53386811168  + 2737.07668820650 *T)
    sumY0  +=      55.947 * cos(0.51535684239  + 638.65663108921 *T)
    sumY0  +=      43.290 * cos(3.23404420181  + 339.53023682000 *T)
    sumY0  +=      42.587 * cos(5.31609891905  + 421.42538238810 *T)
    sumY0  +=      52.900 * cos(4.47156939792  + 205.67860639531 *T)
    sumY0  +=      41.439 * cos(4.04544063722  + 120.11443212250 *T)
    sumY0  +=      47.251 * cos(6.26333565058  + 558.24595822940 *T)
    sumY0  +=      48.495 * cos(3.54910301218  + 7.40554859410 *T)
    sumY0  +=      47.908 * cos(2.91926139184  + 192.93598510200 *T)
    sumY0  +=      44.030 * cos(3.33712242752  + 24.13520490470 *T)
    sumY0  +=      45.108 * cos(4.49379795044  + 227.55755860200 *T)
    sumY0  +=      46.355 * cos(1.53774240394  + 432.47108265200 *T)
    sumY0  +=      53.109 * cos(2.88300873205  + 167.07249000551 *T)
    sumY0  +=      45.868 * cos(0.76053202388  + 205.90810105891 *T)
    sumY0  +=      41.458 * cos(5.68277172210  + 1677.69475801730 *T)
    sumY0  +=      37.152 * cos(2.93697363406  + 207.18012544610 *T)
    sumY0  +=      47.404 * cos(1.81311842537  + 6.83609962250 *T)
    sumY0  +=      38.168 * cos(2.51101235743  + 244.56240155850 *T)
    sumY0  +=      36.628 * cos(2.09707617094  + 427.59276788490 *T)
    sumY0  +=      45.029 * cos(0.46131983164  + 428.32648106781 *T)
    sumY0  +=      36.896 * cos(5.20502524032  + 1898.59503542310 *T)
    sumY0  +=      38.453 * cos(4.42450335190  + 209.39831133730 *T)
    sumY0  +=      48.446 * cos(2.78432187312  + 68.90870679129 *T)
    sumY0  +=      38.820 * cos(5.55847637074  + 117.07605073670 *T)
    sumY0  +=      40.434 * cos(5.16218435235  + 234.14984074100 *T)
    sumY0  +=      36.263 * cos(3.03797170606  + 273.82223084130 *T)
    sumY0  +=      49.643 * cos(4.62992447202  + 3340.85624418330 *T)
    sumY0  +=      35.063 * cos(1.56413997114  + 34.18406742730 *T)
    sumY0  +=      37.471 * cos(3.60681147286  + 217.68751450571 *T)
    sumY0  +=      46.113 * cos(5.55096889463  + 212.64914104420 *T)
    sumY0  +=      46.113 * cos(2.48024992554  + 214.43668479881 *T)
    sumY0  +=      37.195 * cos(3.14310393264  + 41.40068029210 *T)
    sumY0  +=      36.077 * cos(3.08475839413  + 867.66729323710 *T)
    sumY0  +=      37.351 * cos(1.63731661243  + 1251.09656714130 *T)
    sumY0  +=      34.230 * cos(0.03480230424  + 182.05034353251 *T)
    sumY0  +=      41.145 * cos(2.55939029519  + 33.69643246030 *T)
    sumY0  +=      40.662 * cos(4.06091902786  + 59.55992755679 *T)
    sumY0  +=      43.402 * cos(5.46714027653  + 135.30473397060 *T)
    sumY0  +=      39.654 * cos(5.16682565077  + 113.89465699899 *T)
    sumY0  +=      36.261 * cos(4.05668696595  + 28.06735816780 *T)
    sumY0  +=      32.687 * cos(4.53409645547  + 956.04533848709 *T)
    sumY0  +=      33.157 * cos(4.00608684255  + 89.10949770051 *T)
    sumY0  +=      36.698 * cos(2.11524640748  + 431.50787480550 *T)
    sumY0  +=      33.531 * cos(3.91382811936  + 81.50831573269 *T)
    sumY0  +=      31.599 * cos(4.19712154327  + 553.82616999929 *T)
    sumY0  +=      31.098 * cos(0.18155680317  + 1780.78753223589 *T)
    sumY0  +=      37.751 * cos(4.45825713034  + 319.81708087780 *T)
    sumY0  +=      35.774 * cos(4.64374162623  + 137.27684164590 *T)
    sumY0  +=      39.730 * cos(4.75427059768  + 739.04109232210 *T)
    sumY0  +=      38.350 * cos(4.71854522019  + 145.87486135500 *T)
    sumY0  +=      34.642 * cos(0.32805201664  + 77.99436146740 *T)
    sumY0  +=      31.570 * cos(1.72922409232  + 466.19888427470 *T)
    sumY0  +=      39.360 * cos(0.62486177896  + 374.15181032000 *T)
    sumY0  +=      31.496 * cos(5.67282160428  + 24.62283987170 *T)
    sumY0  +=      30.207 * cos(0.37237079306  + 2523.77759276850 *T)
    sumY0  +=      29.335 * cos(4.61376820029  + 131.79077970531 *T)
    sumY0  +=      40.419 * cos(5.34787848533  + 1692.40948698591 *T)
    sumY0  +=      31.724 * cos(5.43414431783  + 333.90116252750 *T)
    sumY0  +=      34.923 * cos(4.42267152292  + 2950.37578364450 *T)
    sumY0  +=      28.746 * cos(0.14118338011  + 406.23508054000 *T)
    sumY0  +=      28.755 * cos(4.87076841042  + 193.09604379641 *T)
    sumY0  +=      35.049 * cos(1.27633192968  + 381.59542572090 *T)
    sumY0  +=      27.298 * cos(2.43907047943  + 60.52313540329 *T)
    sumY0  +=      35.019 * cos(4.97297202382  + 23.81969071961 *T)
    sumY0  +=      27.836 * cos(5.88920583793  + 216.51185833811 *T)
    sumY0  +=      36.108 * cos(3.05676075892  + 756.07956517339 *T)
    sumY0  +=      29.306 * cos(1.98340803593  + 464.97504399731 *T)
    sumY0  +=      27.049 * cos(1.35474057566  + 361.62163734781 *T)
    sumY0  +=      34.876 * cos(5.82579229775  + 39.86132582961 *T)
    sumY0  +=      27.398 * cos(1.50873653792  + 411.86415483250 *T)
    sumY0  +=      26.699 * cos(1.77744028043  + 313.45429340240 *T)
    sumY0  +=      31.423 * cos(2.98291222547  + 0.50694204189 *T)
    sumY0  +=      33.616 * cos(4.20740714080  + 395.82251972250 *T)
    sumY0  +=      26.335 * cos(2.16782274779  + 0.76508234530 *T)
    sumY0  +=      26.583 * cos(1.68554114116  + 245.03548231051 *T)
    sumY0  +=      32.446 * cos(0.13091443612  + 423.17297965479 *T)
    sumY0  +=      35.965 * cos(2.38625629897  + 2111.40649589409 *T)
    sumY0  +=      27.132 * cos(2.17807289493  + 60.04756252381 *T)
    sumY0  +=      25.856 * cos(3.24096099019  + 16.91859203990 *T)
    sumY0  +=      29.986 * cos(1.75172035628  + 710.99054910171 *T)
    sumY0  +=      32.943 * cos(5.30878070823  + 339.04260185300 *T)
    sumY0  +=      27.965 * cos(2.27098970437  + 852.95256426850 *T)
    sumY0  +=      26.667 * cos(2.04771482115  + 210.57396750490 *T)
    sumY0  +=      28.346 * cos(0.35789109097  + 281.21096448800 *T)
    sumY0  +=      25.306 * cos(3.16045047065  + 233.98978204660 *T)
    sumY0  +=      25.371 * cos(0.11921349830  + 412.82736267900 *T)
    sumY0  +=      25.734 * cos(0.68499330423  + 459.08533727390 *T)
    sumY0  +=      26.331 * cos(5.16659836470  + 494.51005992600 *T)
    sumY0  +=      30.392 * cos(4.52887777009  + 129.67565967810 *T)
    sumY0  +=      25.998 * cos(1.14475064542  + 681.29796660610 *T)
    sumY0  +=      33.249 * cos(0.31026740190  + 162.33718759030 *T)
    sumY0  +=      34.168 * cos(5.11102486996  + 260.01333583751 *T)
    sumY0  +=      33.256 * cos(0.02004670501  + 342.01154726960 *T)
    sumY0  +=      26.237 * cos(2.20426681355  + 45.33283355520 *T)
    sumY0  +=      24.623 * cos(3.52685708072  + 618.80046279510 *T)
    sumY0  +=      33.518 * cos(3.87245484115  + 26.84351789039 *T)
    sumY0  +=      27.190 * cos(4.56188666009  + 211.89838151880 *T)
    sumY0  +=      27.190 * cos(3.46933216009  + 215.18744432420 *T)
    sumY0  +=      29.507 * cos(5.96033162893  + 455.15318401080 *T)
    sumY0  +=      32.634 * cos(1.91946398686  + 2008.31372167550 *T)
    sumY0  +=      25.187 * cos(4.08272447614  + 1272.43720814370 *T)
    sumY0  +=      25.644 * cos(2.29213342962  + 230.21768747790 *T)
    sumY0  +=      24.787 * cos(5.85090359378  + 2847.28300942589 *T)
    sumY0  +=      23.670 * cos(3.08413813287  + 106.03035047280 *T)
    sumY0  +=      23.511 * cos(2.91304532501  + 214.34606207361 *T)
    sumY0  +=      23.511 * cos(5.11817349516  + 212.73976376941 *T)
    sumY0  +=      23.747 * cos(3.86428603001  + 467.99887116810 *T)
    sumY0  +=      30.107 * cos(5.10457107202  + 408.19512612779 *T)
    sumY0  +=      27.532 * cos(4.49562952252  + 430.28652665560 *T)
    sumY0  +=      24.105 * cos(6.16708424151  + 422.64922266551 *T)
    sumY0  +=      24.814 * cos(0.10470016181  + 329.96900926440 *T)
    sumY0  +=      25.569 * cos(4.64105815749  + 174.92224231670 *T)
    sumY0  +=      24.522 * cos(4.97031829303  + 420.69166920520 *T)
    sumY0  +=      23.164 * cos(4.11483060864  + 245.29860686890 *T)
    sumY0  +=      24.171 * cos(4.01730820598  + 184.97110483931 *T)
    sumY0  +=      22.048 * cos(3.76525133977  + 163.14033674240 *T)
    sumY0  +=      22.198 * cos(0.91804268520  + 426.73013377490 *T)
    sumY0  +=      25.834 * cos(1.85492286402  + 405.50136735710 *T)
    sumY0  +=      21.674 * cos(0.18639888994  + 2119.00767786191 *T)
    sumY0  +=      23.225 * cos(6.03563186387  + 207.32313779801 *T)
    sumY0  +=      21.571 * cos(2.12243591235  + 7.87862934611 *T)
    sumY0  +=      27.080 * cos(4.85176617134  + 1155.60497489050 *T)
    sumY0  +=      21.569 * cos(1.63592147192  + 224.84924561630 *T)
    sumY0  +=      21.638 * cos(1.02889895862  + 432.25863433090 *T)
    sumY0  +=      22.198 * cos(5.61107049084  + 426.95388294410 *T)
    sumY0  +=      21.066 * cos(1.41468665110  + 508.10650660869 *T)
    sumY0  +=      22.956 * cos(2.58264721236  + 93.18466331550 *T)
    sumY0  +=      22.695 * cos(6.24214235238  + 223.57722122910 *T)
    sumY0  +=      27.166 * cos(4.79918382120  + 36.89238041300 *T)
    sumY0  +=      25.768 * cos(1.13633294908  + 143.90275367970 *T)
    sumY0  +=      22.341 * cos(1.99558555167  + 219.76268804501 *T)
    sumY0  +=      20.372 * cos(2.83637498985  + 849.50804597240 *T)
    sumY0  +=      27.791 * cos(0.28838089763  + 106.78110999820 *T)
    sumY0  +=      20.257 * cos(0.79329378943  + 131.64776735340 *T)
    sumY0  +=      24.029 * cos(0.97528115907  + 334.79493440480 *T)
    sumY0  +=      20.321 * cos(5.28793099536  + 3053.46855786309 *T)
    sumY0  +=      21.760 * cos(3.44547432746  + 264.26100328570 *T)
    sumY0  +=      19.818 * cos(1.30149158514  + 357.68948408471 *T)
    sumY0  +=      21.046 * cos(4.00723953716  + 242.11472100371 *T)
    sumY0  +=      22.078 * cos(1.96836238081  + 186.45557754760 *T)
    sumY0  +=      21.330 * cos(0.83656885419  + 343.46239008310 *T)
    sumY0  +=      23.629 * cos(1.23658530283  + 333.41352756050 *T)
    sumY0  +=      24.992 * cos(2.91707869462  + 31.73638687251 *T)
    sumY0  +=      24.756 * cos(2.65874555104  + 46.45360796860 *T)
    sumY0  +=      25.007 * cos(3.73507948705  + 2428.28600051770 *T)
    sumY0  +=      23.761 * cos(0.76061848201  + 536.56069461190 *T)
    sumY0  +=      21.722 * cos(1.95490574105  + 202.76033721600 *T)
    sumY0  +=      18.920 * cos(3.45843630535  + 171.89841514590 *T)
    sumY0  +=      18.653 * cos(4.99142844854  + 427.05445668060 *T)
    sumY0  +=      18.653 * cos(1.53768472750  + 426.62956003840 *T)
    sumY0  +=      21.149 * cos(2.42045227957  + 320.08020543620 *T)
    sumY0  +=      25.426 * cos(4.81957323352  + 481.01667910731 *T)
    sumY0  +=      22.841 * cos(4.85681711059  + 451.22103074770 *T)
    sumY0  +=      23.911 * cos(1.40533416228  + 264.74863825270 *T)
    sumY0  +=      22.993 * cos(6.23437907106  + 556.76148552110 *T)
    sumY0  +=      19.429 * cos(0.49892034946  + 392.90175841570 *T)
    sumY0  +=      22.024 * cos(5.95458526387  + 112.89781925770 *T)
    sumY0  +=      23.549 * cos(4.55443763374  + 593.67068088151 *T)
    sumY0  +=      19.967 * cos(1.36046222351  + 1567.48843679789 *T)
    sumY0  +=      22.876 * cos(0.86570844668  + 67.91186905000 *T)
    sumY0  +=      19.677 * cos(4.58233275783  + 216.23667169220 *T)
    sumY0  +=      23.750 * cos(4.50456492563  + 462.26673101160 *T)
    sumY0  +=      22.555 * cos(3.61220411831  + 78.46993434690 *T)
    sumY0  +=      20.143 * cos(3.66113241657  + 963.64652045490 *T)
    sumY0  +=      21.017 * cos(2.13961985045  + 402.46298597130 *T)
    sumY0  +=      21.934 * cos(4.14044243723  + 347.64062156210 *T)
    sumY0  +=      17.902 * cos(6.17093225727  + 555.79827767460 *T)
    sumY0  +=      20.903 * cos(1.55046365235  + 66.46102623650 *T)
    sumY0  +=      18.918 * cos(1.61907531561  + 561.42735196710 *T)
    sumY0  +=      17.876 * cos(4.94950539935  + 2214.98690507971 *T)
    sumY0  +=      17.731 * cos(4.68138674425  + 460.78225830330 *T)
    sumY0  +=      17.098 * cos(3.04400260073  + 146.83806920150 *T)
    sumY0  +=      24.044 * cos(2.58903556136  + 141.46962733991 *T)
    sumY0  +=      16.884 * cos(1.64089065325  + 478.04773369070 *T)
    sumY0  +=      18.890 * cos(3.13750980567  + 572.47305223100 *T)
    sumY0  +=      18.099 * cos(4.67997663967  + 230.95140066080 *T)
    sumY0  +=      18.165 * cos(5.53730899074  + 533.86693584120 *T)
    sumY0  +=      20.525 * cos(1.73279668620  + 156.92056161890 *T)
    sumY0  +=      16.659 * cos(1.53572181695  + 419.77664546851 *T)
    sumY0  +=      23.314 * cos(1.60748447836  + 280.72332952100 *T)
    sumY0  +=      18.637 * cos(2.70791503528  + 152.77596003471 *T)
    sumY0  +=      16.704 * cos(2.84594084238  + 422.42222012939 *T)
    sumY0  +=      16.529 * cos(1.17500375466  + 533.37930087419 *T)
    sumY0  +=      16.486 * cos(5.89369907681  + 835.28095197081 *T)
    sumY0  +=      17.269 * cos(4.52541950314  + 200.20959079720 *T)
    sumY0  +=      20.803 * cos(1.66581919068  + 175.67050971460 *T)
    sumY0  +=      18.263 * cos(1.72834635046  + 906.13039727500 *T)
    sumY0  +=      17.267 * cos(3.70630213204  + 295.43805848960 *T)
    sumY0  +=      17.953 * cos(5.46559992336  + 547.20025796550 *T)
    sumY0  +=      22.334 * cos(2.49164712100  + 423.87306294291 *T)
    sumY0  +=      20.308 * cos(2.63358277774  + 294.80741117069 *T)
    sumY0  +=      16.381 * cos(4.28960471972  + 263.94548910060 *T)
    sumY0  +=      16.267 * cos(0.58296245759  + 1486.22393854870 *T)
    sumY0  +=      18.540 * cos(3.41437432865  + 552.82933225800 *T)
    sumY0  +=      16.627 * cos(6.25613561237  + 1.20702533000 *T)
    sumY0  +=      16.245 * cos(0.95492982065  + 99.66756299740 *T)
    sumY0  +=      16.399 * cos(4.04808546179  + 757.46097201770 *T)
    sumY0  +=      18.383 * cos(6.05008020502  + 240.63024829540 *T)
    sumY0  +=      16.013 * cos(6.19975448744  + 2097.17940189250 *T)
    sumY0  +=      15.991 * cos(2.20124656384  + 1055.69359440960 *T)
    sumY0  +=      16.379 * cos(5.79238193635  + 692.34366686999 *T)
    sumY0  +=      21.494 * cos(1.80010144776  + 206.64181424181 *T)
    sumY0  +=      15.840 * cos(6.11030412335  + 831.34879870770 *T)
    sumY0  +=      16.271 * cos(4.93442008056  + 526.00262931501 *T)
    sumY0  +=      16.436 * cos(5.52440863918  + 1994.08662767389 *T)
    sumY0  +=      18.303 * cos(0.84097877285  + 100.14064374939 *T)
    sumY0  +=      16.892 * cos(3.75216779365  + 1588.82907780029 *T)
    sumY0  +=      16.255 * cos(3.49297899617  + 226.87623504580 *T)
    sumY0  +=      15.567 * cos(1.54318213545  + 220.54458533771 *T)
    sumY0  +=      15.921 * cos(0.10158990272  + 2840.16946242510 *T)
    sumY0  +=      16.234 * cos(2.19172369252  + 490.09027169589 *T)
    sumY0  +=      15.273 * cos(4.91184696980  + 78.20680978850 *T)
    sumY0  +=      15.648 * cos(0.18114715284  + 215.54865049161 *T)
    sumY0  +=      15.648 * cos(1.56688636016  + 211.53717535141 *T)
    sumY0  +=      18.382 * cos(0.18848218475  + 220.44401160121 *T)
    sumY0  +=      16.593 * cos(3.42511025047  + 109.48693085641 *T)
    sumY0  +=      18.947 * cos(1.50197502200  + 637.69342324271 *T)
    sumY0  +=      16.216 * cos(4.42535566671  + 41.29761442810 *T)
    sumY0  +=      20.572 * cos(5.47938216081  + 285.61619850311 *T)
    sumY0  +=      16.595 * cos(3.19938069715  + 364.59058276440 *T)
    sumY0  +=      17.326 * cos(1.34597310936  + 200.04953210281 *T)
    sumY0  +=      14.971 * cos(3.33832806105  + 646.76701583129 *T)
    sumY0  +=      16.383 * cos(1.10309438059  + 204.14174406030 *T)
    sumY0  +=      16.103 * cos(4.64455276241  + 419.68027724890 *T)
    sumY0  +=      14.886 * cos(0.65960995330  + 416.05943265400 *T)
    sumY0  +=      18.360 * cos(1.87709798852  + 211.57080524620 *T)
    sumY0  +=      19.474 * cos(5.02476951207  + 2008.80135664250 *T)
    sumY0  +=      20.381 * cos(4.17760232563  + 108.00245814810 *T)
    sumY0  +=      14.597 * cos(0.56756639384  + 110.40195459310 *T)
    sumY0  +=      16.369 * cos(0.61877635867  + 172.70156429801 *T)
    sumY0  +=      14.609 * cos(3.18901609785  + 942.30587945250 *T)
    sumY0  +=      14.822 * cos(0.12437540127  + 196.01680510321 *T)
    sumY0  +=      14.578 * cos(4.55858090950  + 255.18741069710 *T)
    sumY0  +=      17.284 * cos(0.28522496168  + 254.69977573010 *T)
    sumY0  +=      15.790 * cos(5.07797348512  + 410.16723380310 *T)
    sumY0  +=      19.120 * cos(2.03406090557  + 2427.79836555070 *T)
    sumY0  +=      17.766 * cos(0.80075789262  + 50.64639366260 *T)
    sumY0  +=      14.974 * cos(0.23924844134  + 65.46418849521 *T)
    sumY0  +=      19.645 * cos(4.42545578897  + 14.73403604350 *T)
    sumY0  +=      15.462 * cos(0.64414831002  + 222.94408178270 *T)
    sumY0  +=      15.819 * cos(1.10669581015  + 82.13896305161 *T)
    sumY0  +=      17.917 * cos(2.79735661192  + 83.10217089810 *T)
    sumY0  +=      14.016 * cos(5.63689121224  + 427.10264079041 *T)
    sumY0  +=      15.063 * cos(4.17429508241  + 441.81986188650 *T)
    sumY0  +=      17.883 * cos(2.91769140697  + 458.59770230690 *T)
    sumY0  +=      15.768 * cos(1.39358756030  + 349.09146437560 *T)
    sumY0  +=      19.463 * cos(1.41185123555  + 72.31710306511 *T)
    sumY0  +=      14.016 * cos(0.89222201489  + 426.58137592861 *T)
    sumY0  +=      13.854 * cos(2.32880269949  + 13.08950464080 *T)
    sumY0  +=      15.926 * cos(0.40206042973  + 227.03629374021 *T)
    sumY0  +=      15.265 * cos(1.61120369517  + 272.85902299479 *T)
    sumY0  +=      14.549 * cos(3.76467114399  + 110.49832281271 *T)
    sumY0  +=      15.012 * cos(6.23427543593  + 220.76833450690 *T)
    sumY0  +=      14.376 * cos(2.59471394496  + 448.93340888730 *T)
    sumY0  +=      15.060 * cos(5.64392577085  + 980.91199584231 *T)
    sumY0  +=      17.967 * cos(3.03320322404  + 702.39252939260 *T)
    sumY0  +=      13.507 * cos(3.66461671083  + 295.29504613770 *T)
    sumY0  +=      15.721 * cos(2.79281459085  + 570.98857952271 *T)
    sumY0  +=      13.819 * cos(4.40974508034  + 457.86149699650 *T)
    sumY0  +=      16.412 * cos(1.18291048596  + 222.14093263061 *T)
    sumY0  +=      16.702 * cos(5.85287056545  + 100.88891114730 *T)
    sumY0  +=      16.229 * cos(0.83644011312  + 111.41334654940 *T)
    sumY0  +=      13.830 * cos(3.43135216380  + 434.91876320681 *T)
    sumY0  +=      18.336 * cos(4.67709866433  + 170.25388374320 *T)
    sumY0  +=      14.088 * cos(0.06758308008  + 540.98048284201 *T)
    sumY0  +=      13.344 * cos(1.61984291787  + 231.06902073981 *T)
    sumY0  +=      13.303 * cos(3.76362078874  + 472.41865939820 *T)
    sumY0  +=      15.477 * cos(0.34906740950  + 543.26810470240 *T)
    sumY0  +=      12.978 * cos(5.00967301311  + 241.99710092470 *T)
    sumY0  +=      15.002 * cos(4.20300225162  + 3266.76765330109 *T)
    sumY0  +=      17.598 * cos(4.80416246880  + 569.29165849331 *T)
    sumY0  +=      15.115 * cos(0.24398954365  + 420.24972622050 *T)
    sumY0  +=      12.874 * cos(5.63936187479  + 258.63192899320 *T)
    sumY0  +=      16.607 * cos(2.65645538069  + 28.32799059869 *T)
    sumY0  +=      15.813 * cos(0.08924170710  + 251.41531612840 *T)
    sumY0  +=      13.957 * cos(4.60482495392  + 864.48589949940 *T)
    sumY0  +=      15.413 * cos(3.60816325052  + 259.11956396020 *T)
    sumY0  +=      16.676 * cos(4.00742132263  + 705.36147480921 *T)
    sumY0  +=      12.604 * cos(5.58744937803  + 3163.67487908250 *T)
    sumY0  +=      12.578 * cos(0.31186779740  + 238.81570718700 *T)
    sumY0  +=      12.841 * cos(5.08974304707  + 149.31937965110 *T)
    sumY0  +=      12.315 * cos(5.19795063531  + 326.93062787860 *T)
    sumY0  +=      13.217 * cos(2.66592778977  + 188.41313100791 *T)
    sumY0  +=      12.147 * cos(6.00853202742  + 267.22994870229 *T)
    sumY0  +=      12.395 * cos(4.00929640081  + 642.58878435231 *T)
    sumY0  +=      14.104 * cos(2.40740514880  + 230.80838830890 *T)
    sumY0  +=      14.269 * cos(4.27275851966  + 42.78208713641 *T)
    sumY0  +=      11.476 * cos(5.05008428113  + 280.24775664150 *T)
    sumY0  +=      11.770 * cos(0.06778086534  + 1685.29593998510 *T)
    sumY0  +=      11.418 * cos(2.54785709519  + 97.91996573070 *T)
    sumY0  +=      12.752 * cos(5.31254492577  + 274.30986580831 *T)
    sumY0  +=      13.026 * cos(5.54434755126  + 193.89919294850 *T)
    sumY0  +=      11.246 * cos(3.76658400062  + 120.01136625851 *T)
    sumY0  +=      11.976 * cos(2.63340598809  + 380.37158544351 *T)
    sumY0  +=      11.206 * cos(2.55060349358  + 714.92270236481 *T)
    sumY0  +=      14.434 * cos(2.20590299498  + 387.22450001341 *T)
    sumY0  +=      11.970 * cos(1.79693811667  + 203.50860461390 *T)
    sumY0  +=      14.741 * cos(1.71323516885  + 2324.70559133209 *T)
    sumY0  +=      12.049 * cos(1.14966633500  + 254.38426154501 *T)
    sumY0  +=      10.842 * cos(0.97259949349  + 640.18928790730 *T)
    sumY0  +=      13.350 * cos(0.65575474438  + 639.17789595100 *T)
    sumY0  +=      14.697 * cos(1.26523656461  + 432.99234751380 *T)
    sumY0  +=      10.885 * cos(0.32557672954  + 17.28229033479 *T)
    sumY0  +=      11.250 * cos(4.26467347482  + 105.76971804189 *T)
    sumY0  +=      13.740 * cos(2.81161922279  + 815.30716359770 *T)
    sumY0  +=      10.769 * cos(3.01740495484  + 19.36627259471 *T)
    sumY0  +=      12.850 * cos(5.62701027438  + 196.27743753410 *T)
    sumY0  +=      10.505 * cos(0.20513166775  + 206.54124050531 *T)
    sumY0  +=      10.385 * cos(4.75266812707  + 197.08058668621 *T)
    sumY0  +=      10.842 * cos(4.05799649461  + 640.09291968770 *T)
    sumY0  +=      10.290 * cos(0.21483888335  + 202.23658022670 *T)
    sumY0  +=      13.234 * cos(4.77358659430  + 894.59706204410 *T)
    sumY0  +=      10.647 * cos(2.81797753930  + 819.81546347680 *T)
    sumY0  +=      10.182 * cos(5.04028210552  + 3369.86042751970 *T)
    sumY0  +=      10.210 * cos(1.14110944477  + 107.52937739611 *T)
    sumY0  +=      11.258 * cos(5.75203953422  + 189.96703968541 *T)
    sumY0  +=      11.951 * cos(6.11153957549  + 732.18817775220 *T)
    sumY0  +=      13.918 * cos(5.98432962919  + 486.64575339981 *T)
    sumY0  +=      10.041 * cos(5.47708883979  + 606.51636803880 *T)
    sumY0  +=      14.076 * cos(3.35843996869  + 256.83194209980 *T)
    sumY0  +=      12.084 * cos(4.63467214008  + 1471.99684454710 *T)
    sumY0  +=      10.842 * cos(1.07959403550  + 1883.88030645450 *T)
    sumY0  +=      11.375 * cos(1.98886226218  + 106.25735300890 *T)
    sumY0  +=      12.596 * cos(3.05673567432  + 403.26613512341 *T)
    sumY0  +=      11.028 * cos(4.81005385977  + 468.48650613510 *T)
    sumY0  +=       9.860 * cos(1.78727904818  + 33.38091827520 *T)
    sumY0  +=      12.108 * cos(2.52678208199  + 355.50492808829 *T)
    sumY0  +=      10.908 * cos(4.50394816660  + 186.71620997851 *T)
    sumY0  +=       9.594 * cos(1.63465903483  + 434.00373947010 *T)
    sumY0  +=      10.453 * cos(3.45682160657  + 1905.22094745690 *T)
    sumY0  +=       9.546 * cos(1.56659382657  + 189.47940471839 *T)
    sumY0  +=      11.233 * cos(4.51872140986  + 641.10431164400 *T)
    sumY0  +=       9.914 * cos(1.12182170266  + 47.93808067690 *T)
    sumY0  +=       9.448 * cos(2.22449169971  + 407.19828838650 *T)
    sumY0  +=       9.546 * cos(0.75076951952  + 219.38443553510 *T)
    sumY0  +=      10.366 * cos(2.64558555015  + 196.86813836510 *T)
    sumY0  +=       9.370 * cos(3.27855073906  + 230.00523915680 *T)
    sumY0  +=      11.272 * cos(2.99571473964  + 731.70054278519 *T)
    sumY0  +=       9.924 * cos(3.34063551482  + 275.30670354960 *T)
    sumY0  +=       9.472 * cos(1.45022970781  + 194.42045781031 *T)
    sumY0  +=      12.896 * cos(0.49167321133  + 662.28738607949 *T)
    sumY0  +=       9.150 * cos(1.20267384220  + 526.26575387340 *T)
    sumY0  +=      11.879 * cos(0.70048342795  + 477.56009872370 *T)
    sumY0  +=      10.766 * cos(3.45948811786  + 240.36961586451 *T)
    sumY0  +=       9.308 * cos(0.32925975206  + 1368.90407032850 *T)
    sumY0  +=       9.735 * cos(4.72972577500  + 433.90737125051 *T)
    sumY0  +=      10.066 * cos(5.20819082705  + 424.13369537380 *T)
    sumY0  +=      10.907 * cos(3.94508545487  + 351.57277482520 *T)
    sumY0  +=      12.189 * cos(1.67640720798  + 480.52904414029 *T)
    sumY0  +=      11.289 * cos(3.24731450821  + 390.19344543000 *T)
    sumY0  +=      11.032 * cos(5.16458163650  + 395.34943897051 *T)
    sumY0  +=       8.918 * cos(6.04573965183  + 19.88753745650 *T)
    sumY0  +=       9.676 * cos(5.36593074203  + 238.67269483510 *T)
    sumY0  +=      10.312 * cos(4.29506314601  + 17.02165790390 *T)
    sumY0  +=       8.636 * cos(5.96071961936  + 92.29089143820 *T)
    sumY0  +=       8.732 * cos(4.67038464325  + 157.88376946540 *T)
    sumY0  +=       9.566 * cos(1.08602404748  + 1493.33748554950 *T)
    sumY0  +=      10.066 * cos(1.32092240008  + 429.55032134520 *T)
    sumY0  +=       9.374 * cos(1.79762056777  + 206.31749133610 *T)
    sumY0  +=       9.533 * cos(2.29106952890  + 237.11878615761 *T)
    sumY0  +=       9.780 * cos(0.41361009361  + 850.25880549781 *T)
    sumY0  +=       8.433 * cos(1.81570030403  + 427.00206705391 *T)
    sumY0  +=       8.433 * cos(4.71341292310  + 426.68194966511 *T)
    sumY0  +=      10.240 * cos(1.66027874947  + 2744.19023520730 *T)
    sumY0  +=       8.356 * cos(1.17965180591  + 824.49837626530 *T)
    sumY0  +=       9.888 * cos(4.34303137793  + 9.31741007210 *T)
    sumY0  +=       8.509 * cos(6.12186657443  + 3156.56133208170 *T)
    sumY0  +=       8.903 * cos(0.86621704764  + 649.70233135310 *T)
    sumY0  +=       8.389 * cos(0.30076573450  + 232.66536803271 *T)
    sumY0  +=       8.197 * cos(2.91430634921  + 90.00326957780 *T)
    sumY0  +=       8.246 * cos(0.71371200046  + 194.63290613140 *T)
    sumY0  +=       8.317 * cos(3.70830923900  + 661.48174479990 *T)
    sumY0  +=       8.295 * cos(4.66370849222  + 2531.37877473630 *T)
    sumY0  +=      11.264 * cos(0.19657279056  + 269.67762925710 *T)
    sumY0  +=      11.219 * cos(5.60782330430  + 453.92934373340 *T)
    sumY0  +=       8.970 * cos(5.65876817460  + 187.67941782501 *T)
    sumY0  +=       8.956 * cos(5.89406537620  + 749.85979004990 *T)
    sumY0  +=      10.013 * cos(3.51802249315  + 2744.67787017430 *T)
    sumY0  +=      10.091 * cos(4.37804522690  + 495.99453263430 *T)
    sumY0  +=       8.906 * cos(4.17006433490  + 518.40144734719 *T)
    sumY0  +=       7.978 * cos(1.68218286192  + 394.59867944510 *T)
    sumY0  +=       7.792 * cos(5.96613230300  + 2413.57127154910 *T)
    sumY0  +=       8.025 * cos(1.15975594380  + 830.86116374070 *T)
    sumY0  +=      10.403 * cos(0.38130462371  + 354.76872277791 *T)
    sumY0  +=       8.025 * cos(1.24802278239  + 857.37235249860 *T)
    sumY0  +=      10.165 * cos(4.79885957746  + 192.44835013500 *T)
    sumY0  +=       9.333 * cos(2.94875622705  + 471.93102443120 *T)
    sumY0  +=       9.111 * cos(4.34972226694  + 167.96626188280 *T)
    sumY0  +=       9.887 * cos(2.79593153869  + 29.44876501211 *T)
    sumY0  +=       7.646 * cos(4.18255519394  + 487.60896124631 *T)
    sumY0  +=       9.194 * cos(0.22887207656  + 594.89452115890 *T)
    sumY0  +=       7.695 * cos(1.00995675874  + 207.70139030790 *T)
    sumY0  +=       7.915 * cos(3.72645918918  + 131.16013238639 *T)
    sumY0  +=       7.999 * cos(4.19812767433  + 256.08118257441 *T)
    sumY0  +=       9.422 * cos(4.87000438446  + 172.48911597691 *T)
    sumY0  +=       8.126 * cos(5.12558873909  + 402.77850015639 *T)
    sumY0  +=      10.105 * cos(0.37101174537  + 204.94489321240 *T)
    sumY0  +=       7.633 * cos(5.98476290128  + 511.28790034640 *T)
    sumY0  +=       9.479 * cos(3.80790804578  + 350.08830211689 *T)
    sumY0  +=       8.422 * cos(2.65463118043  + 503.10807963510 *T)
    sumY0  +=       7.684 * cos(5.08713167056  + 305.10235190919 *T)
    sumY0  +=       9.659 * cos(5.37522379945  + 380.63221787440 *T)
    sumY0  +=       7.375 * cos(6.00876960253  + 367.03826331921 *T)
    sumY0  +=       9.835 * cos(4.77056680350  + 2325.19322629910 *T)
    sumY0  +=       7.261 * cos(6.07068227227  + 69.60879007940 *T)
    sumY0  +=       7.327 * cos(1.03432156084  + 232.45291971160 *T)
    sumY0  +=       8.716 * cos(1.46511731987  + 405.01373239010 *T)
    sumY0  +=       9.076 * cos(5.54405684361  + 237.43430034270 *T)
    sumY0  +=       9.427 * cos(4.50525102171  + 418.24398865040 *T)
    sumY0  +=       7.385 * cos(5.28613440029  + 141.20899490900 *T)
    sumY0  +=       7.235 * cos(6.06771220435  + 2001.68780964171 *T)
    sumY0  +=       7.849 * cos(5.11039772720  + 176.89434999200 *T)
    sumY0  +=       7.786 * cos(3.04563909213  + 100.15519796440 *T)
    sumY0  +=       7.111 * cos(0.78899804979  + 72.57773549600 *T)
    sumY0  +=       9.164 * cos(0.91705396219  + 562.39055981360 *T)
    sumY0  +=       7.298 * cos(0.17912713996  + 425.56998397231 *T)
    sumY0  +=       7.773 * cos(0.05373244763  + 144.17794032560 *T)
    sumY0  +=       7.342 * cos(2.72544980965  + 326.19691469570 *T)
    sumY0  +=       7.093 * cos(0.79204881061  + 291.50590522650 *T)
    sumY0  +=       6.873 * cos(5.44329864315  + 2310.47849733050 *T)
    sumY0  +=       8.660 * cos(5.24323492513  + 343.98365494490 *T)
    sumY0  +=       9.119 * cos(4.57102761808  + 113.15845168861 *T)
    sumY0  +=       7.614 * cos(2.36906012007  + 239.40640801801 *T)
    sumY0  +=       7.081 * cos(4.84402626338  + 448.98829064149 *T)
    sumY0  +=       7.895 * cos(0.32210557679  + 1169.34443392509 *T)
    sumY0  +=       7.696 * cos(4.16685790987  + 746.67839631219 *T)
    sumY0  +=       7.903 * cos(2.33550155951  + 84.68970947040 *T)
    sumY0  +=       7.977 * cos(4.51267535337  + 171.00464326860 *T)
    sumY0  +=       7.298 * cos(0.06680077999  + 428.11403274671 *T)
    sumY0  +=       8.742 * cos(6.06241714509  + 739.56484931140 *T)
    sumY0  +=       7.590 * cos(0.83668168509  + 1802.61580820530 *T)
    sumY0  +=       8.186 * cos(6.25213891994  + 28.69800548670 *T)
    sumY0  +=       7.872 * cos(1.27624139711  + 135.57992061650 *T)
    sumY0  +=       6.935 * cos(1.68866808772  + 1287.15193711231 *T)
    sumY0  +=       6.519 * cos(2.71209280189  + 314.18800658530 *T)
    sumY0  +=       6.628 * cos(3.37815400506  + 269.20205637760 *T)
    sumY0  +=       8.878 * cos(1.45180473574  + 190.64836324160 *T)
    sumY0  +=       8.467 * cos(5.88497294703  + 486.15811843280 *T)
    sumY0  +=       6.757 * cos(6.09171118268  + 419.20719649691 *T)
    sumY0  +=       6.377 * cos(5.93532586744  + 526.75338884040 *T)
    sumY0  +=       7.376 * cos(5.26260347482  + 25.35904518210 *T)
    sumY0  +=       6.333 * cos(3.05880237158  + 117.66675156770 *T)
    sumY0  +=       6.682 * cos(2.34072274026  + 110.97140356470 *T)
    sumY0  +=       6.651 * cos(3.60111217827  + 385.78821141490 *T)
    sumY0  +=       7.569 * cos(3.57942402238  + 196.13442518220 *T)
    sumY0  +=       7.360 * cos(0.56305984682  + 128.60938596760 *T)
    sumY0  +=       6.901 * cos(2.93953956658  + 250.19147585100 *T)
    sumY0  +=       7.381 * cos(0.99738654496  + 376.43943218040 *T)
    sumY0  +=       7.504 * cos(5.30246960621  + 397.63706083091 *T)
    sumY0  +=       7.473 * cos(5.57604903253  + 24.88596443009 *T)
    sumY0  +=       6.275 * cos(5.77608642511  + 329.16586011230 *T)
    sumY0  +=       7.570 * cos(1.45795353512  + 990.47322339791 *T)
    sumY0  +=       6.057 * cos(2.85826159973  + 856.62159297321 *T)
    sumY0  +=       6.364 * cos(3.24758615307  + 2221.61281711350 *T)
    sumY0  +=       5.928 * cos(4.75970769160  + 505.55576018990 *T)
    sumY0  +=       7.836 * cos(3.88573570654  + 129.20008679860 *T)
    sumY0  +=       6.147 * cos(5.38817990453  + 639.61983893570 *T)
    sumY0  +=       6.324 * cos(4.16792532854  + 756.56720014041 *T)
    sumY0  +=       5.898 * cos(4.12207507137  + 73.05330837550 *T)
    sumY0  +=       6.943 * cos(0.75389771722  + 141.73025977080 *T)
    sumY0  +=       5.913 * cos(4.48939242917  + 121.23520653590 *T)
    sumY0  +=       5.823 * cos(0.04791756675  + 843.87897167990 *T)
    sumY0  +=       6.807 * cos(3.82323029526  + 384.30373870660 *T)
    sumY0  +=       6.016 * cos(5.31195959222  + 3480.06674873909 *T)
    sumY0  +=       6.641 * cos(3.45892667351  + 95.63234387030 *T)
    sumY0  +=       6.458 * cos(3.97624356687  + 3583.15952295770 *T)
    sumY0  +=       7.069 * cos(4.22691071621  + 910.06255053810 *T)
    sumY0  +=       7.779 * cos(4.74986011895  + 724.58699578440 *T)
    sumY0  +=       7.221 * cos(2.56386729615  + 312.96665843540 *T)
    sumY0  +=       7.610 * cos(1.44986844464  + 25.51661174900 *T)
    sumY0  +=       7.748 * cos(1.92450987967  + 55.41532597260 *T)
    sumY0  +=       7.664 * cos(0.29268789796  + 236.43746260140 *T)
    sumY0  +=       5.819 * cos(3.22733168242  + 210.88278404561 *T)
    sumY0  +=       5.819 * cos(4.80388713776  + 216.20304179740 *T)
    sumY0  +=       6.261 * cos(1.83319897051  + 109.92887384111 *T)
    sumY0  +=       6.136 * cos(1.06488319276  + 297.98880490840 *T)
    sumY0  +=       5.471 * cos(6.10353405237  + 672.38443271190 *T)
    sumY0  +=       5.971 * cos(0.80843068991  + 2200.27217611110 *T)
    sumY0  +=       6.662 * cos(2.34354527541  + 611.19928082730 *T)
    sumY0  +=       6.604 * cos(1.99020920437  + 393.21727260080 *T)
    sumY0  +=       5.774 * cos(5.34907528886  + 97.17169833279 *T)
    sumY0  +=       5.678 * cos(4.04634265509  + 17.50929287090 *T)
    sumY0  +=       5.771 * cos(4.76574743994  + 164.78486814511 *T)
    sumY0  +=       6.577 * cos(1.54896971348  + 2641.09746098870 *T)
    sumY0  +=       5.341 * cos(0.19806045792  + 104.82106441041 *T)
    sumY0  +=       6.085 * cos(5.30179771397  + 433.43429049850 *T)
    sumY0  +=       6.219 * cos(0.78907212361  + 100.40127618030 *T)
    sumY0  +=       5.285 * cos(6.01982733401  + 37.88921815429 *T)
    sumY0  +=       5.304 * cos(1.08252755690  + 397.14942586389 *T)
    sumY0  +=       6.082 * cos(5.90524903819  + 640.66236865931 *T)
    sumY0  +=       6.287 * cos(2.23628657770  + 636.47207509279 *T)
    sumY0  +=       5.677 * cos(1.52340696709  + 208.43510349080 *T)
    sumY0  +=       5.677 * cos(0.22462654591  + 218.65072235220 *T)
    sumY0  +=       5.284 * cos(3.86736907469  + 760.49935340350 *T)
    sumY0  +=       7.032 * cos(1.41471093834  + 244.07476659150 *T)
    sumY0  +=       5.102 * cos(0.78146044482  + 319.55644844690 *T)
    sumY0  +=       6.572 * cos(0.46210662825  + 1130.47519297691 *T)
    sumY0  +=       5.363 * cos(4.68863519797  + 700.90805668431 *T)
    sumY0  +=       5.071 * cos(4.80768023276  + 806.96977631951 *T)
    sumY0  +=       6.462 * cos(4.11391793715  + 370.21965705690 *T)
    sumY0  +=       5.365 * cos(3.88729493516  + 170.51700830159 *T)
    sumY0  +=       6.508 * cos(5.51528842107  + 566.84397793850 *T)
    sumY0  +=       5.532 * cos(1.94235906842  + 115.86676467430 *T)
    sumY0  +=       5.557 * cos(4.49290822007  + 1788.38871420371 *T)
    sumY0  +=       4.900 * cos(3.62174655888  + 5.38525680900 *T)
    sumY0  +=       4.866 * cos(1.55046015501  + 533.11617631580 *T)
    sumY0  +=       4.838 * cos(4.76613158881  + 2.72512793310 *T)
    sumY0  +=       4.827 * cos(4.78712076128  + 3686.25229717630 *T)
    sumY0  +=       5.563 * cos(5.47778718250  + 508.59414157570 *T)
    sumY0  +=       5.497 * cos(4.10404574865  + 551.27542358051 *T)
    sumY0  +=       6.373 * cos(1.02411772410  + 310.47079377080 *T)
    sumY0  +=       4.825 * cos(1.77695267272  + 184.33796539290 *T)
    sumY0  +=       4.857 * cos(3.11269069727  + 124.25903370670 *T)
    sumY0  +=       5.731 * cos(2.20047665983  + 525.25436191710 *T)
    sumY0  +=       5.593 * cos(6.01527558647  + 320.82847283410 *T)
    sumY0  +=       5.962 * cos(3.48939528514  + 411.37651986550 *T)
    sumY0  +=       4.890 * cos(2.56527545644  + 383.81610373959 *T)
    sumY0  +=       4.978 * cos(3.21018438051  + 178.33063859051 *T)
    sumY0  +=       6.082 * cos(5.25823168414  + 54.89406111079 *T)
    sumY0  +=       5.118 * cos(3.43332156550  + 89.04006173130 *T)
    sumY0  +=       5.637 * cos(5.42363670995  + 500.13913421850 *T)
    sumY0  +=       5.571 * cos(1.33403064909  + 3060.58210486390 *T)
    sumY0  +=       6.493 * cos(2.81526207686  + 26088.14695905770 *T)
    sumY0  +=       5.591 * cos(0.97716211165  + 285.35556607221 *T)
    sumY0  +=       4.800 * cos(5.14117990977  + 1383.13116433011 *T)
    sumY0  +=       5.463 * cos(6.24603882858  + 57.86300652739 *T)
    sumY0  +=       5.035 * cos(5.27515551134  + 658.30035106221 *T)
    sumY0  +=       4.758 * cos(6.23235562812  + 632.06434895021 *T)
    sumY0  +=       5.081 * cos(0.71196597981  + 482.97672469510 *T)
    sumY0  +=       4.612 * cos(3.90241552282  + 634.51202950500 *T)
    sumY0  +=       6.057 * cos(4.20888329498  + 1045.39865367110 *T)
    sumY0  +=       4.578 * cos(2.92031249606  + 271.64973693240 *T)
    sumY0  +=       5.515 * cos(4.29921246987  + 455.41381644171 *T)
    sumY0  +=       5.305 * cos(1.73055781517  + 484.94883237040 *T)
    sumY0  +=       5.440 * cos(5.20717674102  + 770.06058095910 *T)
    sumY0  +=       4.994 * cos(1.27545128731  + 42.88515300039 *T)
    sumY0  +=       4.487 * cos(0.44426255576  + 732.45130231059 *T)
    sumY0  +=       5.536 * cos(0.07875805067  + 304.84171947829 *T)
    sumY0  +=       4.486 * cos(3.98303359589  + 440.17533048380 *T)
    sumY0  +=       4.513 * cos(0.92434082044  + 745.03386490949 *T)
    sumY0  +=       5.772 * cos(0.96059759251  + 32.48714639791 *T)
    sumY0  +=       4.455 * cos(0.91803135084  + 849.77117053080 *T)
    sumY0  +=       5.554 * cos(3.37288410174  + 493.54685207950 *T)
    sumY0  +=       5.487 * cos(2.96316195594  + 319.06881347989 *T)
    sumY0  +=       4.770 * cos(1.62322164671  + 991.95769610620 *T)
    sumY0  +=       4.782 * cos(2.31677814858  + 504.80500066450 *T)
    sumY0  +=       5.299 * cos(2.61568871196  + 200.80029162821 *T)
    sumY0  +=       4.812 * cos(3.04862288999  + 327.94201983490 *T)
    sumY0  +=       4.838 * cos(0.11237010673  + 774.72644740510 *T)
    sumY0  +=       4.398 * cos(0.00709452583  + 637.22034249070 *T)
    sumY0  +=       4.622 * cos(2.94649200387  + 43.04521169480 *T)
    sumY0  +=       4.573 * cos(4.42206452940  + 26.58288545949 *T)
    sumY0  +=       5.909 * cos(1.99258143739  + 313.43973918739 *T)
    sumY0  +=       4.454 * cos(6.20355719133  + 198.72511808891 *T)
    sumY0  +=       4.339 * cos(5.29255890619  + 153.25153291419 *T)
    sumY0  +=       5.432 * cos(5.00322114273  + 199.04063227400 *T)
    sumY0  +=       4.814 * cos(4.75808669593  + 179.91817716280 *T)
    sumY0  +=       5.416 * cos(4.00641531684  + 952.34018776010 *T)
    sumY0  +=       4.417 * cos(1.27050755952  + 113.37090000971 *T)
    sumY0  +=       5.809 * cos(3.24853511558  + 254.59670986610 *T)
    sumY0  +=       4.499 * cos(4.50151529131  + 958.33296034749 *T)
    sumY0  +=       4.316 * cos(4.07998083372  + 840.91002626331 *T)
    sumY0  +=       4.591 * cos(5.41584864129  + 771.54505366740 *T)
    sumY0  +=       4.277 * cos(4.17205571182  + 208.38941150850 *T)
    sumY0  +=       4.839 * cos(3.27594102018  + 635.47523735150 *T)
    sumY0  +=       4.685 * cos(0.59356881548  + 111.67397898031 *T)
    sumY0  +=       5.613 * cos(1.40007895870  + 597.11519917759 *T)
    sumY0  +=       4.574 * cos(2.16667886254  + 201.76349947470 *T)
    sumY0  +=       4.511 * cos(3.47781815287  + 1044.64789414570 *T)
    sumY0  +=       5.229 * cos(4.59429315323  + 108.96566599461 *T)
    sumY0  +=       4.479 * cos(3.85453916879  + 183.85033042589 *T)
    sumY0  +=       4.309 * cos(5.85647983089  + 3472.95320173830 *T)
    sumY0  +=       4.502 * cos(3.27038194685  + 262.30095769790 *T)
    sumY0  +=       4.161 * cos(5.13548179257  + 337.08255626521 *T)
    sumY0  +=       4.124 * cos(4.06317112443  + 440.58146739410 *T)
    sumY0  +=       4.511 * cos(4.75862432396  + 207.42620366200 *T)
    sumY0  +=       4.109 * cos(0.91307204622  + 1140.89024592190 *T)
    sumY0  +=       5.256 * cos(5.07383356130  + 832.09955823310 *T)
    sumY0  +=       4.845 * cos(2.27696783071  + 226.28553421481 *T)
    sumY0  +=       5.044 * cos(0.21580160510  + 1033.11455891480 *T)
    sumY0  +=       4.049 * cos(2.54843540923  + 372.66733761171 *T)
    sumY0  +=       4.077 * cos(3.46715688832  + 650.43604453600 *T)
    sumY0  +=       4.372 * cos(2.72227659738  + 225.32232636830 *T)
    sumY0  +=       4.235 * cos(2.81122205692  + 348.87901605450 *T)
    sumY0  +=       4.119 * cos(5.80261426663  + 2318.07967929830 *T)
    sumY0  +=       5.657 * cos(5.89121825218  + 198.07742442750 *T)
    sumY0  +=       4.022 * cos(3.84203441493  + 623.46632924110 *T)
    sumY0  +=       4.173 * cos(1.46039755717  + 215.83053478190 *T)
    sumY0  +=       4.173 * cos(0.28763595583  + 211.25529106110 *T)
    sumY0  +=       4.084 * cos(4.37381652925  + 2847.77064439291 *T)
    sumY0  +=       3.996 * cos(3.91636260921  + 561.69047652550 *T)
    sumY0  +=       4.060 * cos(1.82766165317  + 228.36070775411 *T)
    sumY0  +=       4.344 * cos(4.99072999724  + 278.27564896620 *T)
    sumY0  +=       4.010 * cos(2.77317134028  + 283.90472325870 *T)
    sumY0  +=       3.948 * cos(5.71910364337  + 2729.96314120569 *T)
    sumY0  +=       5.205 * cos(3.11137685284  + 72.05647063421 *T)
    sumY0  +=       4.634 * cos(2.61060858315  + 527.22646959241 *T)
    sumY0  +=       5.029 * cos(4.81157596800  + 197.50127781150 *T)
    sumY0  +=       4.736 * cos(3.73180881545  + 425.14508733010 *T)
    sumY0  +=       3.969 * cos(5.38207110627  + 21.09682351890 *T)
    sumY0  +=       5.443 * cos(0.63349242611  + 481.97988695381 *T)
    sumY0  +=       5.129 * cos(4.15635980903  + 219.22017132380 *T)
    sumY0  +=       5.129 * cos(3.87485901115  + 207.86565451920 *T)
    sumY0  +=       3.855 * cos(3.65565538336  + 425.61816808210 *T)
    sumY0  +=       4.014 * cos(3.86098154856  + 707.80915536400 *T)
    sumY0  +=       3.860 * cos(3.28600233258  + 512.66930719071 *T)
    sumY0  +=       3.855 * cos(2.87345784377  + 428.06584863690 *T)
    sumY0  +=       3.832 * cos(6.08314691428  + 668.45227944880 *T)
    sumY0  +=       3.833 * cos(1.39881214946  + 969.86629557841 *T)
    sumY0  +=       3.992 * cos(4.82103462797  + 248.75518725251 *T)
    sumY0  +=       4.504 * cos(3.32625747172  + 3061.06973983091 *T)
    sumY0  +=       3.821 * cos(5.77104029935  + 1258.69774910910 *T)
    sumY0  +=       3.915 * cos(2.80717156028  + 969.37866061139 *T)
    sumY0  +=       4.000 * cos(1.32325223162  + 49.47743513939 *T)
    sumY0  +=       4.401 * cos(2.76622768705  + 774.98957196349 *T)
    sumY0  +=       3.903 * cos(0.64283522390  + 379.14774516610 *T)
    sumY0  +=       5.273 * cos(5.43800011102  + 88.35873817511 *T)
    sumY0  +=       4.974 * cos(6.14192419266  + 604.71638114540 *T)
    sumY0  +=       3.770 * cos(4.80245950772  + 212.17031479781 *T)
    sumY0  +=       3.770 * cos(3.22875931246  + 214.91551104521 *T)
    sumY0  +=       4.435 * cos(2.07247731785  + 204.97852310719 *T)
    sumY0  +=       3.735 * cos(2.76356339426  + 1596.43025976811 *T)
    sumY0  +=       3.690 * cos(2.41699531433  + 738.81408978600 *T)
    sumY0  +=       3.728 * cos(5.06144142130  + 407.71955324831 *T)
    sumY0  +=       4.348 * cos(1.86173045083  + 69.12115511239 *T)
    sumY0  +=       4.549 * cos(6.13388081033  + 491.80174694030 *T)
    sumY0  +=       3.710 * cos(5.49660849476  + 1575.08961876571 *T)
    sumY0  +=       3.788 * cos(2.56556261266  + 258.78949556011 *T)
    sumY0  +=       4.244 * cos(4.08555395724  + 319.08336769490 *T)
    sumY0  +=       3.638 * cos(4.65492407042  + 634.72447782610 *T)
    sumY0  +=       4.782 * cos(3.44556090693  + 313.92737415440 *T)
    sumY0  +=       4.614 * cos(3.41327739029  + 318.12015984841 *T)
    sumY0  +=       3.611 * cos(1.12376486937  + 316.68387124990 *T)
    sumY0  +=       3.647 * cos(1.83675765478  + 48.51422729289 *T)
    sumY0  +=       3.812 * cos(3.74026028851  + 6276.20612047410 *T)
    sumY0  +=       4.045 * cos(1.63984852083  + 519.15220687259 *T)
    sumY0  +=       3.837 * cos(4.78177293888  + 28.55499313480 *T)
    sumY0  +=       3.969 * cos(3.99934092978  + 398.53083270820 *T)
    sumY0  +=       4.988 * cos(2.96911465595  + 105.29663728990 *T)
    sumY0  +=       4.187 * cos(1.30561704256  + 1905.70858242391 *T)
    sumY0  +=       3.673 * cos(5.96397301101  + 619.53417597800 *T)
    sumY0  +=       3.529 * cos(4.10821310653  + 215.50295850929 *T)
    sumY0  +=       4.522 * cos(3.21771300936  + 229.58454803150 *T)
    sumY0  +=       3.575 * cos(3.63589972521  + 291.76902978489 *T)
    sumY0  +=       4.154 * cos(1.16714789173  + 721.40560204669 *T)
    sumY0  +=       4.496 * cos(5.28534594309  + 120.74507944141 *T)
    sumY0  +=       3.649 * cos(1.96746705061  + 444.36811617781 *T)
    sumY0  +=       4.846 * cos(0.23438304594  + 25.61967761299 *T)
    sumY0  +=       4.339 * cos(3.48361033260  + 609.12161516050 *T)
    sumY0  +=       3.687 * cos(2.97857012599  + 2538.00468677009 *T)
    sumY0  +=       3.925 * cos(4.68848167473  + 317.59889498661 *T)
    sumY0  +=       4.658 * cos(4.49051482749  + 2641.58509595570 *T)
    sumY0  +=       4.614 * cos(2.41025191456  + 616.07759559439 *T)
    sumY0  +=       3.906 * cos(3.62084996234  + 525.51499434800 *T)
    sumY0  +=       3.627 * cos(5.17097264158  + 10206.41581669370 *T)
    sumY0  +=       3.452 * cos(2.95884375111  + 62.49524307860 *T)
    sumY0  +=       3.889 * cos(3.82999893371  + 103.81216458159 *T)
    sumY0  +=       4.257 * cos(2.80377624434  + 72.09010052900 *T)
    sumY0  +=       4.693 * cos(3.34813046925  + 619.04654101100 *T)
    sumY0  +=       3.405 * cos(6.22409873424  + 393.70490756781 *T)
    sumY0  +=       3.406 * cos(0.38852296833  + 102.89714084490 *T)
    sumY0  +=       3.405 * cos(3.77912190314  + 172.00148100990 *T)
    sumY0  +=       4.000 * cos(2.69996342698  + 1048.09241244180 *T)
    sumY0  +=       4.448 * cos(6.15238947202  + 1485.73630358170 *T)
    sumY0  +=       4.227 * cos(4.47863866494  + 413.82420042030 *T)
    sumY0  +=       3.578 * cos(4.21100881764  + 316.58750303031 *T)
    sumY0  +=       3.352 * cos(1.98897148023  + 298.47643987540 *T)
    sumY0  +=       3.398 * cos(3.47406128287  + 102.80077262530 *T)
    sumY0  +=       3.878 * cos(2.82672809448  + 210.60759739969 *T)
    sumY0  +=       3.282 * cos(1.97609142655  + 215.07977525651 *T)
    sumY0  +=       3.282 * cos(6.05512739363  + 212.00605058650 *T)
    sumY0  +=       3.262 * cos(3.29324404703  + 327.99690158910 *T)
    sumY0  +=       3.353 * cos(1.38999835719  + 766.12842769600 *T)
    sumY0  +=       3.738 * cos(3.80751619036  + 123.29582586020 *T)
    sumY0  +=       3.649 * cos(4.73181541366  + 610.44852130189 *T)
    sumY0  +=       4.126 * cos(1.21427959149  + 187.51935913060 *T)
    sumY0  +=       3.391 * cos(0.57199147232  + 541.29599702710 *T)
    sumY0  +=       3.608 * cos(1.00199216800  + 1119.42949271300 *T)
    sumY0  +=       3.758 * cos(0.10141006490  + 101.88574888859 *T)
    sumY0  +=       3.231 * cos(1.54367112759  + 188.27011865600 *T)
    sumY0  +=       3.787 * cos(4.28600120342  + 652.67127676971 *T)
    sumY0  +=       3.965 * cos(0.26781383536  + 914.20715212231 *T)
    sumY0  +=       3.735 * cos(5.95952945648  + 80.44204202221 *T)
    sumY0  +=       3.177 * cos(5.74147536696  + 691.34682912870 *T)
    sumY0  +=       3.187 * cos(3.55567838325  + 305.85061930710 *T)
    sumY0  +=       4.058 * cos(2.17039511461  + 92.06388890210 *T)
    sumY0  +=       3.123 * cos(4.19403609870  + 307.07445958451 *T)
    sumY0  +=       3.691 * cos(5.19957321657  + 194.53233239490 *T)
    sumY0  +=       3.467 * cos(2.75272147826  + 428.53892938890 *T)
    sumY0  +=       3.270 * cos(2.49533429526  + 282.90788551740 *T)
    sumY0  +=       4.132 * cos(0.15196778100  + 434.47682022211 *T)
    sumY0  +=       3.256 * cos(2.06059207193  + 344.45922782439 *T)
    sumY0  +=       4.289 * cos(0.56065735151  + 187.94005025590 *T)
    sumY0  +=       3.819 * cos(6.09508602758  + 114.69780615110 *T)
    sumY0  +=       3.049 * cos(0.94913326717  + 842.65762352999 *T)
    sumY0  +=       3.360 * cos(0.53915356673  + 2516.66404576769 *T)
    sumY0  +=       3.446 * cos(2.93783559783  + 305.85311143459 *T)
    sumY0  +=       3.964 * cos(6.14378220347  + 101.62511645769 *T)
    sumY0  +=       3.527 * cos(3.59077404312  + 496.20698095540 *T)
    sumY0  +=       3.065 * cos(4.95454737668  + 348.39138108750 *T)
    sumY0  +=       3.500 * cos(2.81354951024  + 789.70430093210 *T)
    sumY0  +=       2.995 * cos(4.38744698123  + 91.54262404029 *T)
    sumY0  +=       3.701 * cos(1.02316235609  + 405.74744557300 *T)
    sumY0  +=       3.539 * cos(5.71127020478  + 233.18663289450 *T)
    sumY0  +=       2.935 * cos(4.18624941833  + 92.55401599659 *T)
    sumY0  +=       3.414 * cos(4.58481058460  + 331.45348197271 *T)
    sumY0  +=       3.976 * cos(0.60864007570  + 31.26330612050 *T)
    sumY0  +=       3.100 * cos(3.82122143105  + 130.19692453990 *T)
    sumY0  +=       3.608 * cos(0.90039093237  + 315.67247929360 *T)
    sumY0  +=       2.909 * cos(0.13390320628  + 109.69937917751 *T)
    sumY0  +=       2.890 * cos(3.94290490601  + 111.20089822830 *T)
    sumY0  +=       2.917 * cos(1.86330744308  + 357.47703576360 *T)
    sumY0  +=       3.278 * cos(0.09902260816  + 126.23114138200 *T)
    sumY0  +=       2.952 * cos(3.59175692736  + 93.28772917949 *T)
    sumY0  +=       2.859 * cos(5.00627346803  + 181.56270856549 *T)
    sumY0  +=       3.008 * cos(2.53117554530  + 434.70631488571 *T)
    sumY0  +=       2.925 * cos(5.09670716093  + 3796.45861839570 *T)
    sumY0  +=       3.751 * cos(1.42372094959  + 624.67561530350 *T)
    sumY0  +=       3.658 * cos(1.22634527273  + 268.68079151581 *T)
    sumY0  +=       3.572 * cos(6.08351547848  + 851.95572652720 *T)
    sumY0  +=       2.923 * cos(0.90054830851  + 252.41215386970 *T)
    sumY0  +=       3.162 * cos(0.65619571338  + 77.71917482150 *T)
    sumY0  +=       3.903 * cos(5.95524321349  + 1066.25165970650 *T)
    sumY0  +=       3.528 * cos(3.73411976495  + 834.79331700380 *T)
    sumY0  +=       3.311 * cos(1.68379432029  + 212.09207010800 *T)
    sumY0  +=       3.007 * cos(4.10088774274  + 699.94484883781 *T)
    sumY0  +=       2.914 * cos(0.86930882153  + 625.16325027050 *T)
    sumY0  +=       2.826 * cos(1.18680021962  + 71.84402231310 *T)
    sumY0  +=       2.804 * cos(2.83113208076  + 425.40571976101 *T)
    sumY0  +=       2.779 * cos(3.18761506074  + 629.84616305900 *T)
    sumY0  +=       3.204 * cos(1.68163453202  + 31.24875190549 *T)
    sumY0  +=       3.277 * cos(2.83164568099  + 232.55349344810 *T)
    sumY0  +=       3.564 * cos(5.26360280410  + 38.11622069041 *T)
    sumY0  +=       3.345 * cos(3.12621780784  + 77.47309660560 *T)
    sumY0  +=       3.005 * cos(3.42504542953  + 135.09228564950 *T)
    sumY0  +=       3.542 * cos(0.53352784827  + 239.56646671240 *T)
    sumY0  +=       2.893 * cos(3.87647442239  + 3899.55139261430 *T)
    sumY0  +=       3.544 * cos(1.94605675805  + 282.20780222929 *T)
    sumY0  +=       2.804 * cos(3.69798114637  + 428.27829695801 *T)
    sumY0  +=       2.850 * cos(0.70634186901  + 177.63055530240 *T)
    sumY0  +=       2.737 * cos(5.93123636788  + 626.12645811700 *T)
    sumY0  +=       2.722 * cos(5.30959446782  + 214.01848580100 *T)
    sumY0  +=       2.716 * cos(1.52354592802  + 620.49738382450 *T)
    sumY0  +=       2.818 * cos(0.89510519023  + 354.50809034700 *T)
    sumY0  +=       3.749 * cos(3.91662579273  + 145.66241303390 *T)
    sumY0  +=       3.118 * cos(1.00275828453  + 3376.97397452050 *T)
    sumY0  +=       2.902 * cos(3.76865453023  + 52.93401552300 *T)
    sumY0  +=       3.000 * cos(4.86727145892  + 206.90493880020 *T)
    sumY0  +=       2.965 * cos(0.01297764373  + 42.56963881530 *T)
    sumY0  +=       2.885 * cos(3.24271543105  + 418.97770183331 *T)
    sumY0  +=       2.693 * cos(4.35806644152  + 433.20479583490 *T)
    sumY0  +=       3.088 * cos(1.37972033305  + 2957.48933064529 *T)
    sumY0  +=       2.922 * cos(3.81670841310  + 1063.07026596880 *T)
    sumY0  +=       3.407 * cos(4.57769873569  + 499.17592637200 *T)
    sumY0  +=       3.270 * cos(1.13498595983  + 924.28964453971 *T)
    sumY0  +=       3.462 * cos(4.87665869514  + 182.78405671541 *T)
    sumY0  +=       3.056 * cos(3.84258924311  + 179.81511129880 *T)
    sumY0  +=       2.751 * cos(1.98489147077  + 322.26476143261 *T)
    sumY0  +=       3.461 * cos(3.77635871944  + 60.31068708220 *T)
    sumY0  +=       3.254 * cos(3.71673729334  + 87.87110320809 *T)
    sumY0  +=       3.320 * cos(0.62445995726  + 122.98031167510 *T)
    sumY0  +=       3.628 * cos(1.19683291978  + 239.14577558710 *T)
    sumY0  +=       2.722 * cos(0.33833300117  + 644.33388949151 *T)
    sumY0  +=       2.866 * cos(1.16131783177  + 315.41184686271 *T)
    sumY0  +=       3.289 * cos(1.94135959143  + 685.71775483620 *T)
    sumY0  +=       2.598 * cos(5.28324130559  + 151.50393564750 *T)
    sumY0  +=       2.931 * cos(0.28498618097  + 7.83293736379 *T)
    sumY0  +=       2.624 * cos(0.03837656753  + 311.21906116870 *T)
    sumY0  +=       2.667 * cos(2.34899532832  + 1080.96638867510 *T)
    sumY0  +=       2.603 * cos(3.14892076264  + 508.73715392760 *T)
    sumY0  +=       2.559 * cos(3.79368088293  + 639.39034427210 *T)
    sumY0  +=       2.562 * cos(5.83357031849  + 827.67977000299 *T)
    sumY0  +=       2.581 * cos(1.05939902375  + 159.20818347930 *T)
    sumY0  +=       2.754 * cos(4.40535716712  + 287.83687652179 *T)
    sumY0  +=       2.531 * cos(0.15195670029  + 1048.84317196720 *T)
    sumY0  +=       2.912 * cos(5.40624189026  + 1382.64352936309 *T)
    sumY0  +=       2.686 * cos(4.87164344891  + 643.06186510430 *T)
    sumY0  +=       2.508 * cos(0.40758211682  + 297.88573904440 *T)
    sumY0  +=       2.488 * cos(1.63124486596  + 95.47228517591 *T)
    sumY0  +=       3.158 * cos(2.22361915053  + 671.89679774490 *T)
    sumY0  +=       2.864 * cos(5.93155931661  + 453.66871130251 *T)
    sumY0  +=       2.518 * cos(5.50666990719  + 2626.87036698710 *T)
    sumY0  +=       2.793 * cos(4.51817804240  + 208.86249226050 *T)
    sumY0  +=       2.660 * cos(1.58245860025  + 443.51678291590 *T)
    sumY0  +=       2.576 * cos(1.71533674425  + 952.86394474940 *T)
    sumY0  +=       2.986 * cos(6.26559400172  + 105.78427225690 *T)
    sumY0  +=       3.018 * cos(3.58337419807  + 141.94270809190 *T)
    sumY0  +=       2.912 * cos(2.65612793047  + 167.33312243640 *T)
    sumY0  +=       2.430 * cos(2.83520399981  + 235.68919520349 *T)
    sumY0  +=       3.277 * cos(5.74822288161  + 1055.95671896799 *T)
    sumY0  +=       2.469 * cos(2.00868439815  + 213.10346206430 *T)
    sumY0  +=       2.430 * cos(0.45033884806  + 112.41018429070 *T)
    sumY0  +=       2.646 * cos(4.93243120045  + 312.21589891000 *T)
    sumY0  +=       2.465 * cos(4.44624138102  + 205.89354684390 *T)
    sumY0  +=       2.655 * cos(3.73582825216  + 440.85665404000 *T)
    sumY0  +=       2.370 * cos(3.86926101157  + 64.71592109730 *T)
    sumY0  +=       2.389 * cos(5.95405430378  + 85.58348134769 *T)
    sumY0  +=       2.369 * cos(5.61575950108  + 220.86890824340 *T)
    sumY0  +=       2.369 * cos(5.55705197269  + 206.21691759960 *T)
    sumY0  +=       2.493 * cos(3.84873402269  + 388.22133775470 *T)
    sumY0  +=       2.974 * cos(3.71475509576  + 531.73476947149 *T)
    sumY0  +=       2.491 * cos(1.54562260546  + 123.78346082721 *T)
    sumY0  +=       2.347 * cos(5.51371577599  + 3046.35501086230 *T)
    sumY0  +=       3.237 * cos(2.58143954637  + 338.72708766791 *T)
    sumY0  +=       2.520 * cos(4.60115704499  + 338.04576411170 *T)
    sumY0  +=       2.377 * cos(3.29530891079  + 628.60776856660 *T)
    sumY0  +=       2.701 * cos(3.73939130270  + 300.85468446100 *T)
    sumY0  +=       2.653 * cos(5.33077957429  + 56.86616878610 *T)
    sumY0  +=       2.732 * cos(3.27550474377  + 244.30176912761 *T)
    sumY0  +=       2.275 * cos(4.58945603770  + 4002.64416683289 *T)
    sumY0  +=       2.952 * cos(0.74298822983  + 645.77017809000 *T)
    sumY0  +=       2.509 * cos(2.94920927023  + 125.42799222991 *T)
    sumY0  +=       2.253 * cos(0.65609530771  + 310.41591201661 *T)
    sumY0  +=       2.632 * cos(0.50847991258  + 416.79314583689 *T)
    sumY0  +=       2.307 * cos(4.08063632531  + 3164.16251404950 *T)
    sumY0  +=       2.532 * cos(0.42305602058  + 90.31878376290 *T)
    sumY0  +=       2.434 * cos(1.36706416328  + 769.09737311260 *T)
    sumY0  +=       2.819 * cos(5.90718650467  + 34.92027273770 *T)
    sumY0  +=       2.827 * cos(4.24209921133  + 782.59075393131 *T)
    sumY0  +=       2.409 * cos(1.51772791041  + 621.49422156580 *T)
    sumY0  +=       2.533 * cos(3.47780451554  + 631.54308408841 *T)
    sumY0  +=       2.542 * cos(0.51811754999  + 176.63371756111 *T)
    sumY0  +=       2.171 * cos(5.33772457878  + 1891.48148842230 *T)
    sumY0  +=       2.552 * cos(0.14027811136  + 12566.39551746630 *T)
    sumY0  +=       2.811 * cos(6.10219051624  + 641.62557650581 *T)
    sumY0  +=       2.910 * cos(0.06456678007  + 330.96584700570 *T)
    sumY0  +=       2.345 * cos(4.28867359206  + 2104.78058386030 *T)
    sumY0  +=       2.188 * cos(0.65867073144  + 1457.28211557850 *T)
    sumY0  +=       2.591 * cos(5.35033895729  + 170.41394243761 *T)
    sumY0  +=       2.615 * cos(2.11809115675  + 143.69030535860 *T)
    sumY0  +=       2.434 * cos(5.26567142518  + 183.48663213100 *T)
    sumY0  +=       3.003 * cos(4.77242756352  + 739.30172475301 *T)
    sumY0  +=       2.177 * cos(4.17646966614  + 67.63668240410 *T)
    sumY0  +=       2.774 * cos(5.95434553236  + 1048.58004740881 *T)
    sumY0  +=       2.462 * cos(4.18866808016  + 247.27071454420 *T)
    sumY0  +=       2.386 * cos(2.80178234746  + 430.98660994371 *T)
    sumY0  +=       2.988 * cos(5.80843823954  + 1802.12817323829 *T)
    sumY0  +=       2.810 * cos(1.01437231483  + 140.72135994200 *T)
    sumY0  +=       2.255 * cos(4.97404441838  + 6681.46867088311 *T)
    sumY0  +=       2.483 * cos(1.02757238229  + 375.95179721340 *T)
    sumY0  +=       2.123 * cos(5.05514768498  + 610.93615626891 *T)
    sumY0  +=       2.154 * cos(5.72841221574  + 444.10748374690 *T)
    sumY0  +=       2.497 * cos(1.66085968924  + 17.91542978119 *T)
    sumY0  +=       2.073 * cos(0.36915137771  + 218.47190392590 *T)
    sumY0  +=       2.063 * cos(3.20592565915  + 616.56523056141 *T)
    sumY0  +=       2.250 * cos(3.80713836644  + 366.28750379380 *T)
    sumY0  +=       2.750 * cos(2.87538792242  + 485.63685357099 *T)
    sumY0  +=       2.252 * cos(4.35642815581  + 49.26498681829 *T)
    sumY0  +=       2.813 * cos(2.30944774975  + 837.45344587970 *T)
    sumY0  +=       2.062 * cos(5.38385531966  + 2634.47154895491 *T)
    sumY0  +=       2.822 * cos(4.66089423732  + 162.65270177540 *T)
    sumY0  +=       2.611 * cos(2.12105258497  + 927.59115048390 *T)
    sumY0  +=       2.212 * cos(0.93221653291  + 193.41155798150 *T)
    sumY0  +=       2.182 * cos(3.48927417704  + 335.38563523581 *T)
    sumY0  +=       2.403 * cos(0.69912075444  + 452.44487102510 *T)
    sumY0  +=       2.564 * cos(1.72919712902  + 427.73578023681 *T)
    sumY0  +=       2.111 * cos(3.67340327220  + 422.69740677530 *T)
    sumY0  +=       2.308 * cos(1.52890709860  + 155.22364058950 *T)
    sumY0  +=       2.054 * cos(2.67747070868  + 883.55136178020 *T)
    sumY0  +=       2.713 * cos(1.75846895594  + 6.34846465550 *T)
    sumY0  +=       2.692 * cos(4.40535350977  + 1040.97886544099 *T)
    sumY0  +=       2.007 * cos(1.89413239528  + 529.18402305270 *T)
    sumY0  +=       2.282 * cos(3.25546464790  + 173.92540457540 *T)
    sumY0  +=       2.783 * cos(5.54788372150  + 165.36101476110 *T)
    sumY0  +=       2.346 * cos(4.39880402871  + 447.44893617900 *T)
    sumY0  +=       2.376 * cos(5.81992609349  + 945.48727319021 *T)
    sumY0  +=       2.081 * cos(3.67268655906  + 90.57941619379 *T)
    sumY0  +=       2.489 * cos(1.26376364157  + 2487.17222746430 *T)
    sumY0  +=       2.124 * cos(4.95438142065  + 133.34468838281 *T)
    sumY0  +=       2.000 * cos(3.28717598901  + 419.19264228190 *T)
    sumY0  +=       1.998 * cos(6.09695396696  + 518.62844988331 *T)
    sumY0  +=       1.969 * cos(5.59121857952  + 530.68554210351 *T)
    sumY0  +=       2.467 * cos(5.87379769512  + 429.81095377611 *T)
    sumY0  +=       1.955 * cos(1.07634475226  + 105.03351273151 *T)
    sumY0  +=       2.054 * cos(5.64490223202  + 3789.34507139489 *T)
    sumY0  +=       2.425 * cos(2.71454644074  + 180.40581212981 *T)
    sumY0  +=       1.995 * cos(4.07142764766  + 409.57653297210 *T)
    sumY0  +=       2.035 * cos(5.88558560015  + 437.40007365639 *T)
    sumY0  +=       1.990 * cos(2.77224098857  + 2854.39655642670 *T)
    sumY0  +=       2.449 * cos(5.08664901886  + 213.00709384469 *T)
    sumY0  +=       2.354 * cos(4.13524877414  + 657.40657918490 *T)
    sumY0  +=       2.466 * cos(2.78275943769  + 421.68850694650 *T)
    sumY0  +=       1.932 * cos(2.38157117348  + 1062.80714141041 *T)
    sumY0  +=       2.448 * cos(1.35236650723  + 205.98991506349 *T)
    sumY0  +=       1.912 * cos(0.15654616325  + 419.28901050149 *T)
    sumY0  +=       2.647 * cos(4.44444666019  + 747.16603127920 *T)
    sumY0  +=       1.873 * cos(3.82631151650  + 345.08987514331 *T)
    sumY0  +=       2.199 * cos(2.46538504354  + 681.78560157310 *T)
    sumY0  +=       2.386 * cos(5.73059167495  + 25234.46294233869 *T)
    sumY0  +=       1.886 * cos(0.03617873737  + 1096.85045721820 *T)
    sumY0  +=       2.200 * cos(2.10319305969  + 784.28767496071 *T)
    sumY0  +=       1.859 * cos(2.84714041687  + 535.66692273460 *T)
    sumY0  +=       2.123 * cos(6.16596992817  + 418.27761854520 *T)
    sumY0  +=       2.326 * cos(1.19326583155  + 166.58485503849 *T)
    sumY0  +=       2.007 * cos(4.67834010536  + 527.01402127131 *T)
    sumY0  +=       2.162 * cos(3.59622095072  + 420.20403423820 *T)
    sumY0  +=       2.185 * cos(2.29643101683  + 110.18950627200 *T)
    sumY0  +=       2.564 * cos(4.79991609811  + 425.94823648220 *T)
    sumY0  +=       1.976 * cos(4.85054795864  + 102.32769187330 *T)
    sumY0  +=       1.951 * cos(2.23867265514  + 168.29633028290 *T)
    sumY0  +=       1.965 * cos(4.29608988597  + 39.91620758380 *T)
    sumY0  +=       1.953 * cos(2.77691093080  + 116.07921299541 *T)
    sumY0  +=       2.118 * cos(4.88232966976  + 409.43352062020 *T)
    sumY0  +=       2.542 * cos(4.99290516515  + 355.02935520881 *T)
    sumY0  +=       2.294 * cos(3.74777420008  + 181.08713568601 *T)
    sumY0  +=       2.378 * cos(2.54240991813  + 385.30057644790 *T)
    sumY0  +=       1.807 * cos(2.73751159349  + 275.58189019550 *T)
    sumY0  +=       1.833 * cos(0.68863493736  + 267.87764236370 *T)
    sumY0  +=       2.302 * cos(4.17275768872  + 43.53284666180 *T)
    sumY0  +=       2.098 * cos(4.80007899391  + 253.16042126761 *T)
    sumY0  +=       2.063 * cos(4.42291105999  + 177.36743074401 *T)
    sumY0  +=       2.261 * cos(6.12160762328  + 635.94831810351 *T)
    sumY0  +=       2.208 * cos(4.77564977638  + 1194.21109128031 *T)
    sumY0  +=       2.053 * cos(5.29737651120  + 103.37022159689 *T)
    sumY0  +=       1.966 * cos(1.31777921278  + 835.54407652920 *T)
    sumY0  +=       2.346 * cos(1.80109706525  + 222.61650551010 *T)
    sumY0  +=       1.855 * cos(2.03672663902  + 23.13836716340 *T)
    sumY0  +=       1.855 * cos(5.17831929261  + 22.65073219640 *T)
    sumY0  +=       2.017 * cos(2.76554751934  + 243.59919371200 *T)
    sumY0  +=       2.346 * cos(4.19045412826  + 2957.97696561230 *T)
    sumY0  +=       1.781 * cos(6.26122223924  + 220.91709235320 *T)
    sumY0  +=       1.781 * cos(4.91158923453  + 206.16873348980 *T)
    sumY0  +=       1.851 * cos(4.47233105553  + 1107.89615748210 *T)
    sumY0  +=       1.906 * cos(1.87733371845  + 586.06949891369 *T)
    sumY0  +=       2.083 * cos(1.33621197933  + 1037.79747170330 *T)
    sumY0  +=       2.271 * cos(4.50659373635  + 234.03796615640 *T)
    sumY0  +=       2.271 * cos(3.52462508383  + 193.04785968660 *T)
    sumY0  +=       2.373 * cos(4.14396002026  + 281.42341280910 *T)
    sumY0  +=       1.873 * cos(6.16192556463  + 1084.00477006090 *T)
    sumY0  +=       2.113 * cos(3.38006360857  + 671.16059243450 *T)
    sumY0  +=       1.854 * cos(0.68921688536  + 614.38067456499 *T)
    sumY0  +=       1.766 * cos(2.44535049525  + 2.67694382330 *T)
    sumY0  +=       1.774 * cos(1.55215476226  + 220.39582749141 *T)
    sumY0  +=       1.774 * cos(3.33747140433  + 206.68999835160 *T)
    sumY0  +=       2.018 * cos(1.23143201334  + 250.45210828191 *T)
    sumY0  +=       1.877 * cos(2.43826511200  + 35.18090516859 *T)
    sumY0  +=       2.293 * cos(4.61602871136  + 912.72267941400 *T)
    sumY0  +=       1.775 * cos(4.07092209849  + 228.73321476960 *T)
    sumY0  +=       1.775 * cos(0.81870406810  + 198.35261107340 *T)
    sumY0  +=       1.908 * cos(3.18462167186  + 3377.46160948750 *T)
    sumY0  +=       1.924 * cos(4.76564757091  + 534.12756827210 *T)
    sumY0  +=       1.740 * cos(5.30818813577  + 3362.74688051889 *T)
    sumY0  +=       2.154 * cos(5.11342973502  + 666.96780674051 *T)
    sumY0  +=       2.288 * cos(5.13225437866  + 103.59722413300 *T)
    sumY0  +=       1.725 * cos(0.79235621805  + 132.85705341579 *T)
    sumY0  +=       1.785 * cos(3.40968747924  + 346.43133549971 *T)
    sumY0  +=       2.259 * cos(4.18060259139  + 700.42042171730 *T)
    sumY0  +=       1.692 * cos(0.30340863528  + 212.53401309270 *T)
    sumY0  +=       1.721 * cos(3.64237121578  + 306.34074640161 *T)
    sumY0  +=       2.004 * cos(2.69534154569  + 256.67188340540 *T)
    sumY0  +=       2.083 * cos(3.81593475661  + 5216.33655531789 *T)
    sumY0  +=       2.234 * cos(0.83744491072  + 799.36859435169 *T)
    sumY0  +=       1.980 * cos(0.98922312465  + 451.97179027311 *T)
    sumY0  +=       2.071 * cos(5.15399823914  + 549.89401673620 *T)
    sumY0  +=       2.071 * cos(0.48659989203  + 45.49040012211 *T)
    sumY0  +=       1.998 * cos(1.93857319277  + 1171.63205578549 *T)
    sumY0  +=       2.007 * cos(2.45452004266  + 1364.48428209840 *T)
    sumY0  +=       1.836 * cos(1.35614614488  + 644.80697024350 *T)
    sumY0  +=       1.687 * cos(3.75927628859  + 520.37355502251 *T)
    sumY0  +=       1.705 * cos(0.74206548564  + 2222.10045208050 *T)
    sumY0  +=       2.080 * cos(6.16948523942  + 587.45090575800 *T)
    sumY0  +=       2.061 * cos(0.38736887363  + 337.55812914470 *T)
    sumY0  +=       1.674 * cos(4.31595591695  + 207.63865198310 *T)
    sumY0  +=       1.723 * cos(6.04023119944  + 6290.43321447570 *T)
    sumY0  +=       1.760 * cos(3.82579789121  + 12.28635548869 *T)
    sumY0  +=       1.697 * cos(5.77995233151  + 552.34169729099 *T)
    sumY0  +=       1.676 * cos(0.47032854735  + 41.08516610701 *T)
    sumY0  +=       1.715 * cos(3.70385355966  + 133.86844537209 *T)
    sumY0  +=       1.707 * cos(0.48133607676  + 98.60128928690 *T)
    sumY0  +=       1.846 * cos(1.81726608439  + 5849.12029463110 *T)
    sumY0  +=       1.707 * cos(5.82292680345  + 895.08469701110 *T)
    sumY0  +=       1.821 * cos(3.25202211117  + 9779.32999085069 *T)
    sumY0  +=       2.189 * cos(4.93942538444  + 66.94866120350 *T)
    sumY0  +=       2.014 * cos(0.51363431995  + 546.71262299849 *T)
    sumY0  +=       1.781 * cos(3.25286200486  + 9772.21644384989 *T)
    sumY0  +=       1.685 * cos(1.20130027900  + 10220.64291069530 *T)
    sumY0  +=       1.942 * cos(5.33748414084  + 246.68001371321 *T)
    sumY0  +=       2.001 * cos(4.36639869407  + 3.94896821049 *T)
    sumY0  +=       1.626 * cos(0.43544986691  + 624.15435044170 *T)
    sumY0  +=       1.942 * cos(5.24671264008  + 9146.54625153749 *T)
    sumY0  +=       1.802 * cos(1.15593638614  + 1389.35093945360 *T)
    sumY0  +=       1.613 * cos(5.41878737515  + 1144.07163965960 *T)
    sumY0  +=       2.160 * cos(1.30850860560  + 336.59492129820 *T)
    sumY0  +=       2.090 * cos(2.34311286379  + 629.35852809199 *T)
    sumY0  +=       1.968 * cos(1.56341565357  + 118.31444522910 *T)
    sumY0  +=       1.822 * cos(3.62663106790  + 1009.22317149361 *T)
    sumY0  +=       1.611 * cos(2.37000428005  + 347.16754081009 *T)
    sumY0  +=       1.728 * cos(2.57879496141  + 1091.38144162010 *T)
    sumY0  +=       2.055 * cos(5.50239856358  + 1372.34858862460 *T)
    sumY0  +=       1.569 * cos(2.96591528468  + 122.71967924421 *T)
    sumY0  +=       1.542 * cos(0.61717380145  + 1166.16304018740 *T)
    sumY0  +=       1.878 * cos(0.56043256933  + 133.82026126229 *T)
    sumY0  +=       1.628 * cos(5.37550822094  + 46.22660543249 *T)
    sumY0  +=       1.987 * cos(1.16248631448  + 913.50706883420 *T)
    sumY0  +=       1.542 * cos(4.91932828975  + 551.34485954970 *T)
    sumY0  +=       1.643 * cos(5.71885241225  + 196.38050339810 *T)
    sumY0  +=       1.715 * cos(3.48043677711  + 709.29362807231 *T)
    sumY0  +=       1.803 * cos(5.13352083431  + 58.07545484850 *T)
    sumY0  +=       1.620 * cos(2.19306684644  + 700.20797339620 *T)
    sumY0  +=       1.584 * cos(3.71338548490  + 3767.45443505930 *T)
    sumY0  +=       2.027 * cos(5.48886965020  + 2118.52004289490 *T)
    sumY0  +=       1.660 * cos(0.17086781255  + 318.43567403349 *T)
    sumY0  +=       1.695 * cos(5.56469519031  + 327.19375243699 *T)
    sumY0  +=       1.617 * cos(0.82126521853  + 83.62343575990 *T)
    sumY0  +=       1.522 * cos(3.95536376972  + 406.39513923441 *T)
    sumY0  +=       1.847 * cos(3.63437800519  + 488.60579898760 *T)
    sumY0  +=       1.505 * cos(2.43278632462  + 447.28887748461 *T)
    sumY0  +=       1.747 * cos(1.60182024800  + 173.69840203930 *T)
    sumY0  +=       1.511 * cos(1.19445092182  + 414.09938706621 *T)
    sumY0  +=       1.932 * cos(2.99396965089  + 40.59753113999 *T)
    sumY0  +=       1.512 * cos(5.62864113260  + 1748.28107876310 *T)
    sumY0  +=       1.775 * cos(3.81584430362  + 2707.58486990309 *T)
    sumY0  +=       1.830 * cos(5.34714618651  + 259.75270340660 *T)
    sumY0  +=       1.574 * cos(5.50722602817  + 322.36782729660 *T)
    sumY0  +=       1.477 * cos(0.48223767966  + 355.24180352990 *T)
    sumY0  +=       1.498 * cos(0.41524391899  + 174.93679653170 *T)
    sumY0  +=       1.487 * cos(5.62073453387  + 141.45507312490 *T)
    sumY0  +=       1.676 * cos(0.72544410812  + 3693.36584417709 *T)
    sumY0  +=       1.461 * cos(2.95604532560  + 105.13657859549 *T)
    sumY0  +=       1.692 * cos(1.89129839404  + 679.49797971270 *T)
    sumY0  +=       1.835 * cos(4.28334745785  + 245.99869015701 *T)
    sumY0  +=       1.449 * cos(0.71090016730  + 1159.04949318660 *T)
    sumY0  +=       1.661 * cos(3.29324922569  + 736.59590389479 *T)
    sumY0  +=       1.443 * cos(4.93214393414  + 466.51439845980 *T)
    sumY0  +=       1.476 * cos(3.81891695589  + 3480.55438370610 *T)
    sumY0  +=       1.645 * cos(3.60830794016  + 249.71839509901 *T)
    sumY0  +=       1.496 * cos(4.79833807829  + 77623.56832092509 *T)
    sumY0  +=       1.890 * cos(2.07787059832  + 764.64395498770 *T)
    sumY0  +=       1.480 * cos(4.72024280261  + 262.77653057739 *T)
    sumY0  +=       1.937 * cos(5.04418585979  + 757.56403788170 *T)
    sumY0  +=       1.587 * cos(1.89128298241  + 6.91791362710 *T)
    sumY0  +=       1.683 * cos(2.44470376540  + 6503.73230991370 *T)
    sumY0  +=       1.770 * cos(3.51344178417  + 482.50115181560 *T)
    sumY0  +=       1.735 * cos(0.71720239752  + 2435.39954751850 *T)
    sumY0  +=       1.457 * cos(4.59007449571  + 256.18424843839 *T)
    sumY0  +=       1.471 * cos(5.41881489262  + 1453.08683775701 *T)
    sumY0  +=       1.613 * cos(0.21330361543  + 271.86218525350 *T)
    sumY0  +=       1.848 * cos(0.98286127487  + 419.94090967981 *T)
    sumY0  +=       1.954 * cos(0.12353263696  + 145.38722638800 *T)
    sumY0  +=       1.662 * cos(2.01416074486  + 25454.87558477749 *T)
    sumY0  +=       1.420 * cos(1.40716618831  + 1354.18934135989 *T)
    sumY0  +=       1.390 * cos(4.83119263295  + 544.75257741070 *T)
    sumY0  +=       1.679 * cos(4.40359026093  + 285.14311775110 *T)
    sumY0  +=       1.607 * cos(2.30923658857  + 544.26494244370 *T)
    sumY0  +=       1.536 * cos(1.18240794718  + 3273.88120030190 *T)
    sumY0  +=       1.536 * cos(4.32400060077  + 3274.36883526891 *T)
    sumY0  +=       1.587 * cos(3.92843113844  + 148.58317434070 *T)
    sumY0  +=       1.794 * cos(4.23240329975  + 754.59509246510 *T)
    sumY0  +=       1.407 * cos(5.56458421467  + 316.11442227830 *T)
    sumY0  +=       1.658 * cos(0.90957945549  + 415.06259491270 *T)
    sumY0  +=       1.512 * cos(2.71827936357  + 428.48653976220 *T)
    sumY0  +=       1.501 * cos(1.47511443142  + 5746.02752041250 *T)
    sumY0  +=       1.696 * cos(4.47164509007  + 59.03866269500 *T)
    sumY0  +=       1.590 * cos(4.94949423692  + 349.61272923740 *T)
    sumY0  +=       1.827 * cos(2.18886350049  + 230.32075334190 *T)
    sumY0  +=       1.891 * cos(3.43722012359  + 1165.63928319810 *T)
    sumY0  +=       1.759 * cos(3.01502964755  + 200.27902676641 *T)
    sumY0  +=       1.759 * cos(1.87459651904  + 226.80679907661 *T)
    sumY0  +=       1.677 * cos(5.80656752223  + 50.15875869559 *T)
    sumY0  +=       1.384 * cos(2.70658249939  + 1176.94561589290 *T)
    sumY0  +=       1.600 * cos(3.86790494115  + 10433.94200613330 *T)
    sumY0  +=       1.491 * cos(2.88391814725  + 9676.23721663209 *T)
    sumY0  +=       1.400 * cos(4.97246366718  + 1950.53447393720 *T)
    sumY0  +=       1.334 * cos(5.43876271483  + 301.65783361311 *T)
    sumY0  +=       1.377 * cos(5.57667415659  + 243.29037717131 *T)
    sumY0  +=       1.377 * cos(2.45454466359  + 183.79544867171 *T)
    sumY0  +=       1.454 * cos(4.85439502931  + 330.37514617470 *T)
    sumY0  +=       1.421 * cos(2.95881908071  + 469.97097884340 *T)
    sumY0  +=       1.426 * cos(1.44277711657  + 934.44157292630 *T)
    sumY0  +=       1.356 * cos(1.16210512595  + 210.13451664770 *T)
    sumY0  +=       1.314 * cos(4.28508634948  + 415.04804069769 *T)
    sumY0  +=       1.338 * cos(5.08811236319  + 118.62995941420 *T)
    sumY0  +=       1.453 * cos(2.14612729124  + 679.81349389780 *T)
    sumY0  +=       1.470 * cos(2.24019091580  + 321.30155358610 *T)
    sumY0  +=       1.314 * cos(3.65863604800  + 878.08234618209 *T)
    sumY0  +=       1.715 * cos(4.51219845192  + 79.43065006591 *T)
    sumY0  +=       1.305 * cos(4.30852333812  + 238.18505986809 *T)
    sumY0  +=       1.382 * cos(5.23650480607  + 718.85485562790 *T)
    sumY0  +=       1.464 * cos(3.49743224988  + 397.90018538930 *T)
    sumY0  +=       1.758 * cos(1.96057349073  + 452.70550345601 *T)
    sumY0  +=       1.356 * cos(6.13336562879  + 280.93577784210 *T)
    sumY0  +=       1.790 * cos(4.19013422025  + 1357.37073509760 *T)
    sumY0  +=       1.578 * cos(3.80812940834  + 383.55297918120 *T)
    sumY0  +=       1.464 * cos(2.84230067135  + 360.13716463950 *T)
    sumY0  +=       1.375 * cos(5.32376328559  + 535.56385687061 *T)
    sumY0  +=       1.620 * cos(2.04485074017  + 109.06873185860 *T)
    sumY0  +=       1.663 * cos(0.24074353253  + 102.73708215049 *T)
    sumY0  +=       1.314 * cos(1.59459147404  + 1670.58121101650 *T)
    sumY0  +=       1.452 * cos(5.66629342942  + 316.89631957100 *T)
    sumY0  +=       1.302 * cos(2.52670094701  + 116.66991382640 *T)
    sumY0  +=       1.421 * cos(4.09380191404  + 166.84548746939 *T)
    sumY0  +=       1.339 * cos(4.29105717413  + 5863.34738863270 *T)
    sumY0  +=       1.685 * cos(1.36491781577  + 214.53975066279 *T)
    sumY0  +=       1.458 * cos(4.62040676992  + 9352.73179997469 *T)
    sumY0  +=       1.341 * cos(5.18571387486  + 1183.16539101641 *T)
    sumY0  +=       1.300 * cos(2.31662723526  + 120.60206708950 *T)
    sumY0  +=       1.281 * cos(3.00808942775  + 1108.38379244910 *T)
    sumY0  +=       1.652 * cos(4.06437698345  + 165.84864972810 *T)
    sumY0  +=       1.259 * cos(5.12727736043  + 2950.86341861150 *T)
    sumY0  +=       1.508 * cos(3.22670487711  + 763.19311217420 *T)
    sumY0  +=       1.251 * cos(2.13418978784  + 1055.20595944260 *T)
    sumY0  +=       1.416 * cos(5.45080283010  + 1063.55790093581 *T)
    sumY0  +=       1.273 * cos(5.31702713057  + 2207.87335807890 *T)
    sumY0  +=       1.755 * cos(1.82782460074  + 557.75832326239 *T)
    sumY0  +=       1.665 * cos(5.34226901164  + 163.82166029861 *T)
    sumY0  +=       1.330 * cos(1.04705345661  + 156.02678974161 *T)
    sumY0  +=       1.472 * cos(5.21259622405  + 346.15614885380 *T)
    sumY0  +=       1.472 * cos(2.07100357046  + 346.64378382081 *T)
    sumY0  +=       1.283 * cos(3.66165325532  + 4215.94326227089 *T)
    sumY0  +=       1.564 * cos(6.16380597537  + 3391.68870348911 *T)
    sumY0  +=       1.232 * cos(3.58365746191  + 767.12526543730 *T)
    sumY0  +=       1.339 * cos(5.12238861307  + 419.76209125350 *T)
    sumY0  +=       1.471 * cos(0.61024942528  + 2693.35777590150 *T)
    sumY0  +=       1.265 * cos(4.91032534273  + 212.30451842910 *T)
    sumY0  +=       1.355 * cos(4.63185174755  + 418.71956152989 *T)
    sumY0  +=       1.236 * cos(3.21756257132  + 409.67959883610 *T)
    sumY0  +=       1.263 * cos(0.18630409787  + 2090.06585489169 *T)
    sumY0  +=       1.580 * cos(3.30454335371  + 1477.83836716070 *T)
    sumY0  +=       1.310 * cos(0.47622589149  + 1012.66768978969 *T)
    sumY0  +=       1.245 * cos(0.37055529521  + 748.96601817260 *T)
    sumY0  +=       1.565 * cos(2.33371277407  + 1268.99268984760 *T)
    sumY0  +=       1.267 * cos(3.75495075578  + 115.11600514891 *T)
    sumY0  +=       1.279 * cos(5.83252185405  + 217.94814693660 *T)
    sumY0  +=       1.279 * cos(5.34028961972  + 209.13767890641 *T)
    sumY0  +=       1.237 * cos(4.82630923494  + 4112.85048805230 *T)
    sumY0  +=       1.208 * cos(6.03951463451  + 116.12739710521 *T)
    sumY0  +=       1.358 * cos(2.43392352337  + 973.79844884150 *T)
    sumY0  +=       1.406 * cos(3.94371718897  + 959.45373476091 *T)
    sumY0  +=       1.365 * cos(2.10668027413  + 988.28866740149 *T)
    sumY0  +=       1.189 * cos(3.35382822803  + 1378.30523918970 *T)
    sumY0  +=       1.200 * cos(5.20941165157  + 614.86830953200 *T)
    sumY0  +=       1.537 * cos(5.11588456984  + 170.72945662269 *T)
    sumY0  +=       1.449 * cos(1.51194108336  + 1020.26887175750 *T)
    sumY0  +=       1.181 * cos(2.90097686991  + 372.77040347569 *T)
    sumY0  +=       1.227 * cos(4.12497084479  + 2421.17245351690 *T)
    sumY0  +=       1.504 * cos(4.14296643297  + 2810.67764412170 *T)
    sumY0  +=       1.191 * cos(6.20818805618  + 487.39651292520 *T)
    sumY0  +=       1.601 * cos(3.29434090785  + 1151.18518666040 *T)
    sumY0  +=       1.207 * cos(2.53978476142  + 3170.78842608329 *T)
    sumY0  +=       1.288 * cos(6.26513392858  + 3149.44778508089 *T)
    sumY0  +=       1.277 * cos(1.84735278399  + 217.24806364849 *T)
    sumY0  +=       1.276 * cos(4.18638529447  + 94.77220188780 *T)
    sumY0  +=       1.451 * cos(5.76650446504  + 514.95692905111 *T)
    sumY0  +=       1.451 * cos(2.62491181145  + 514.46929408409 *T)
    sumY0  +=       1.235 * cos(5.00376834500  + 173.43776960839 *T)
    sumY0  +=       1.353 * cos(2.18947506754  + 530.04665716270 *T)
    sumY0  +=       1.464 * cos(4.73495977065  + 746.15463932291 *T)
    sumY0  +=       1.246 * cos(0.89840063300  + 424.86990068420 *T)
    sumY0  +=       1.236 * cos(0.93764208982  + 1248.38576202810 *T)
    sumY0  +=       1.279 * cos(1.60294934322  + 703.38936713389 *T)
    sumY0  +=       1.236 * cos(5.87825960887  + 538.84831647230 *T)
    sumY0  +=       1.234 * cos(5.73806183883  + 416.32006508490 *T)
    sumY0  +=       1.161 * cos(0.25671181566  + 313.71492583330 *T)
    sumY0  +=       1.414 * cos(0.03305318629  + 2648.69864295650 *T)
    sumY0  +=       1.177 * cos(6.19222057815  + 643.58562209360 *T)
    sumY0  +=       1.187 * cos(5.75067279775  + 1394.98001374610 *T)
    sumY0  +=       1.173 * cos(0.03353446861  + 331.61354066710 *T)
    sumY0  +=       1.176 * cos(1.33278978806  + 252.14902931130 *T)
    sumY0  +=       1.245 * cos(3.08429983167  + 293.74113746021 *T)
    sumY0  +=       1.187 * cos(5.29261047609  + 4105.73694105150 *T)
    sumY0  +=       1.274 * cos(2.13113537295  + 165.10038233019 *T)
    sumY0  +=       1.391 * cos(1.59389148646  + 5860.16599489499 *T)
    sumY0  +=       1.365 * cos(3.65657609285  + 674.08135374130 *T)
    sumY0  +=       1.165 * cos(2.73482385874  + 107.04174242910 *T)
    sumY0  +=       1.169 * cos(6.25851690874  + 104.29979954861 *T)
    sumY0  +=       1.280 * cos(1.03182051788  + 60.79832204921 *T)
    sumY0  +=       1.470 * cos(5.17513400102  + 1147.74066836430 *T)
    sumY0  +=       1.243 * cos(1.73840614377  + 945.75039774860 *T)
    sumY0  +=       1.211 * cos(0.72754477376  + 271.05903610141 *T)
    sumY0  +=       1.378 * cos(2.91003182751  + 880.26690217851 *T)
    sumY0  +=       1.141 * cos(6.22742124714  + 1365.23504162379 *T)
    sumY0  +=       1.452 * cos(5.39698934001  + 670.25226634220 *T)
    sumY0  +=       1.372 * cos(4.69362648452  + 5852.30168836880 *T)
    sumY0  +=       1.152 * cos(1.00740618265  + 253.21530302180 *T)
    sumY0  +=       1.145 * cos(3.63821232586  + 3796.94625336271 *T)
    sumY0  +=       1.147 * cos(3.66166961803  + 2048.15823255770 *T)
    sumY0  +=       1.139 * cos(2.20526813654  + 227.81819103291 *T)
    sumY0  +=       1.139 * cos(2.68435803005  + 199.26763481011 *T)
    sumY0  +=       1.350 * cos(5.09740387152  + 25440.64849077589 *T)
    sumY0  +=       1.152 * cos(3.54242395780  + 475.60005313590 *T)
    sumY0  +=       1.377 * cos(2.07148345291  + 660.85109748100 *T)
    sumY0  +=       1.324 * cos(4.93758146323  + 101.36448402680 *T)
    sumY0  +=       1.189 * cos(1.08318258977  + 1383.93431348220 *T)
    sumY0  +=       1.403 * cos(5.24418904071  + 1699.03539901970 *T)
    sumY0  +=       1.177 * cos(0.25741420805  + 303.30236501580 *T)
    sumY0  +=       1.164 * cos(0.45690069556  + 1773.67398523510 *T)
    sumY0  +=       1.179 * cos(0.96345477476  + 272.33775813299 *T)
    sumY0  +=       1.228 * cos(4.57813937399  + 91.70019060721 *T)
    sumY0  +=       1.431 * cos(1.96162121028  + 1467.31393175860 *T)
    sumY0  +=       1.142 * cos(4.99209015555  + 656.91894421790 *T)
    sumY0  +=       1.157 * cos(4.74432612488  + 469.64340257080 *T)
    sumY0  +=       1.157 * cos(4.74432612488  + 470.13103753780 *T)
    sumY0  +=       1.459 * cos(4.02039221910  + 1128.77827194750 *T)
    sumY0  +=       1.199 * cos(1.75735870350  + 384.51618702770 *T)
    sumY0  +=       1.360 * cos(1.52010327589  + 1560.37488979710 *T)
    sumY0  +=       1.399 * cos(1.83388449686  + 1243.98302014049 *T)
    sumY0  +=       1.154 * cos(5.47298196371  + 5539.84197197530 *T)
    sumY0  +=       1.290 * cos(4.31671851937  + 1403.57803345520 *T)
    sumY0  +=       1.426 * cos(2.28376835451  + 701.69244610450 *T)
    sumY0  +=       1.402 * cos(3.92764404612  + 1112.31594571220 *T)
    sumY0  +=       1.097 * cos(5.97076399005  + 735.68088015810 *T)
    sumY0  +=       1.431 * cos(2.57854764872  + 388.96960515260 *T)
    sumY0  +=       1.187 * cos(5.80714139797  + 6599.71153713150 *T)
    sumY0  +=       1.093 * cos(1.56243726726  + 427.31758123900 *T)
    sumY0  +=       1.537 * cos(0.69918341067  + 2854.88419139370 *T)
    sumY0  +=       1.417 * cos(2.31755447139  + 25557.96835899609 *T)
    sumY0  +=       1.238 * cos(2.88079905059  + 559.94287925881 *T)
    sumY0  +=       1.473 * cos(0.24029882674  + 2751.79141717511 *T)
    sumY0  +=       1.452 * cos(3.65793468409  + 228.03313148150 *T)
    sumY0  +=       1.516 * cos(1.17544869102  + 560.46663624809 *T)
    sumY0  +=       1.103 * cos(3.05981268820  + 531.94052014820 *T)
    sumY0  +=       1.301 * cos(3.01734562484  + 9790.37569111459 *T)
    sumY0  +=       1.099 * cos(4.42566235624  + 527.92904500800 *T)
    sumY0  +=       1.256 * cos(0.41159011755  + 877.81922162370 *T)
    sumY0  +=       1.250 * cos(2.28718973712  + 1226.45442019471 *T)
    sumY0  +=       1.113 * cos(4.71819779483  + 124.18959773749 *T)
    sumY0  +=       1.072 * cos(2.44959483556  + 206.26930722630 *T)
    sumY0  +=       1.072 * cos(2.44003133103  + 220.81651861670 *T)
    sumY0  +=       1.335 * cos(1.74949827208  + 121.00820399980 *T)
    sumY0  +=       1.282 * cos(6.11708013789  + 9782.51138458839 *T)
    sumY0  +=       1.072 * cos(5.33774395010  + 220.49640122790 *T)
    sumY0  +=       1.072 * cos(5.83506752367  + 206.58942461510 *T)
    sumY0  +=       1.303 * cos(2.12423614143  + 1096.48675892331 *T)
    sumY0  +=       1.397 * cos(3.88396933288  + 1461.68485746610 *T)
    sumY0  +=       1.085 * cos(4.29756274020  + 379.88395047650 *T)
    sumY0  +=       1.170 * cos(0.55123201282  + 705.10084237830 *T)
    sumY0  +=       1.054 * cos(5.52683609857  + 155.53915477459 *T)
    sumY0  +=       1.405 * cos(6.19309224384  + 5436.74919775670 *T)
    sumY0  +=       1.097 * cos(3.03189055427  + 476.56326098241 *T)
    sumY0  +=       1.241 * cos(0.40056980352  + 2833.05591542430 *T)
    sumY0  +=       1.288 * cos(5.13612906560  + 1340.80534299830 *T)
    sumY0  +=       1.129 * cos(1.50688193833  + 1222.52226693161 *T)
    sumY0  +=       1.290 * cos(5.50615903453  + 1082.25966492170 *T)
    sumY0  +=       1.166 * cos(3.48845605490  + 2467.64287643290 *T)
    sumY0  +=       1.299 * cos(0.72075981403  + 2509.55049876690 *T)
    sumY0  +=       1.321 * cos(1.39448839419  + 9366.95889397630 *T)
    sumY0  +=       1.051 * cos(5.92119994324  + 959.97749175019 *T)
    sumY0  +=       1.051 * cos(2.77960728965  + 960.46512671720 *T)
    sumY0  +=       1.211 * cos(1.23911890811  + 304.10551416790 *T)
    sumY0  +=       1.052 * cos(2.29712493185  + 645.55772976890 *T)
    sumY0  +=       1.127 * cos(5.02482962262  + 78477.25233764409 *T)
    sumY0  +=       1.258 * cos(5.97921156731  + 5867.27954189579 *T)
    sumY0  +=       1.073 * cos(1.31054744928  + 328.48453655610 *T)
    sumY0  +=       1.290 * cos(6.05901218139  + 71.56883566720 *T)
    sumY0  +=       1.053 * cos(6.20330756381  + 795.43644108859 *T)
    sumY0  +=       1.329 * cos(3.46712878887  + 2740.74571691120 *T)
    sumY0  +=       1.352 * cos(5.91851196501  + 1127.50624756031 *T)
    sumY0  +=       1.032 * cos(4.22822194022  + 774.25336665311 *T)
    sumY0  +=       1.421 * cos(4.76004456961  + 398.38782035631 *T)
    sumY0  +=       1.302 * cos(2.04302425721  + 476.07562601539 *T)
    sumY0  +=       1.018 * cos(6.23816461448  + 135.79236893760 *T)
    sumY0  +=       1.240 * cos(5.00950394649  + 761.98382611181 *T)
    sumY0  +=       1.072 * cos(2.83229864917  + 1257.94698958370 *T)
    sumY0  +=       1.044 * cos(5.11625051680  + 304.36614659881 *T)
    sumY0  +=       1.166 * cos(4.90192971442  + 613.62991503960 *T)
    sumY0  +=       1.025 * cos(1.58832762314  + 121.59890483079 *T)
    sumY0  +=       1.321 * cos(2.86590848137  + 323.48860171000 *T)
    sumY0  +=       1.279 * cos(2.94437273951  + 203.72105293501 *T)
    sumY0  +=       1.279 * cos(1.94525342708  + 223.36477290801 *T)
    sumY0  +=       1.145 * cos(1.46132806151  + 519.88592005549 *T)
    sumY0  +=       1.018 * cos(1.34516588766  + 640.89186332290 *T)
    sumY0  +=       1.389 * cos(0.53131371898  + 1115.76046400830 *T)
    sumY0  +=       1.343 * cos(5.45761793906  + 108.26558270649 *T)
    sumY0  +=       1.195 * cos(5.57047410357  + 99.14380600810 *T)
    sumY0  +=       1.008 * cos(5.87818461327  + 102.58832430420 *T)
    sumY0  +=       1.143 * cos(0.50705630471  + 256.34430713280 *T)
    sumY0  +=       1.019 * cos(0.20479434596  + 6247.26929803929 *T)
    sumY0  +=       1.250 * cos(4.11174294828  + 759.01488069520 *T)
    sumY0  +=       1.000 * cos(0.56919545753  + 1939.48877367330 *T)
    sumY0  +=       1.110 * cos(0.09487667592  + 1397.94895916270 *T)
    sumY0  +=       1.170 * cos(2.83129485516  + 7.39099437910 *T)
    sumY0  +=       1.196 * cos(0.06592121358  + 46.81730626350 *T)
    sumY0  +=       1.151 * cos(4.08931701317  + 76.02225379210 *T)
    sumY0  +=       1.005 * cos(1.04985332435  + 690.85919416170 *T)
    sumY0  +=       1.316 * cos(2.89375228785  + 475.11241816889 *T)
    sumY0  +=       0.995 * cos(5.77459088813  + 99.92819542829 *T)
    sumY0  +=       1.197 * cos(5.97834335795  + 1392.53233319130 *T)
    sumY0  +=       1.110 * cos(4.55961650413  + 523.29680845680 *T)
    sumY0  +=       1.088 * cos(3.40095791153  + 824.23525170691 *T)
    sumY0  +=       1.015 * cos(3.04308405040  + 634.98760238449 *T)
    sumY0  +=       1.012 * cos(0.20207071166  + 689.86235642041 *T)
    sumY0  +=       1.381 * cos(4.87521866240  + 97.43233076370 *T)
    sumY0  +=       1.276 * cos(6.06052974092  + 465.71124930770 *T)
    sumY0  +=       1.233 * cos(0.63981764737  + 1546.29080814740 *T)
    sumY0  +=       1.160 * cos(4.66657413935  + 152.98840835580 *T)
    sumY0  +=       1.247 * cos(5.15702411658  + 2434.91191255150 *T)
    sumY0  +=       1.201 * cos(1.52749255612  + 528.71094230071 *T)
    sumY0  +=       1.045 * cos(3.06096994103  + 608.00084074710 *T)
    sumY0  +=       1.034 * cos(4.03244571439  + 540.49284787499 *T)
    sumY0  +=       0.988 * cos(4.03410369230  + 967.89418790310 *T)
    sumY0  +=       0.973 * cos(2.36193993021  + 173.38288785421 *T)
    sumY0  +=       1.081 * cos(3.91239703303  + 560.93971700010 *T)
    sumY0  +=       1.105 * cos(5.31122507957  + 227.72182281330 *T)
    sumY0  +=       1.105 * cos(5.86158639420  + 199.36400302970 *T)
    sumY0  +=       1.022 * cos(2.56735303200  + 599.40282103799 *T)
    sumY0  +=       1.176 * cos(1.11948010869  + 9797.48923811540 *T)
    sumY0  +=       1.174 * cos(2.18660086424  + 1680.87615175499 *T)
    sumY0  +=       0.980 * cos(4.64038062361  + 944.99963822319 *T)
    sumY0  +=       1.194 * cos(3.00496182534  + 725.07463075140 *T)
    sumY0  +=       1.045 * cos(4.74326343420  + 343.49601997789 *T)
    sumY0  +=       1.227 * cos(2.37463413166  + 1973.37663399041 *T)
    sumY0  +=       1.185 * cos(0.73899940221  + 773.24197469681 *T)
    sumY0  +=       1.222 * cos(1.64001146959  + 776.21092011340 *T)
    sumY0  +=       0.980 * cos(3.66760351816  + 524.78128116509 *T)
    sumY0  +=       1.285 * cos(4.08409167589  + 6.82154540749 *T)
    sumY0  +=       1.225 * cos(4.39413389062  + 25344.66926355809 *T)
    sumY0  +=       0.985 * cos(1.95660242133  + 913.24394427581 *T)
    sumY0  +=       0.995 * cos(5.50466248309  + 557.28275038290 *T)
    sumY0  +=       0.983 * cos(5.82567686855  + 550.38165170320 *T)
    sumY0  +=       1.005 * cos(0.48068217539  + 301.17019864609 *T)
    sumY0  +=       1.208 * cos(6.18461471853  + 2597.37854868370 *T)
    sumY0  +=       0.950 * cos(1.34316892428  + 568.54089896790 *T)
    sumY0  +=       1.220 * cos(3.06190616867  + 1458.71591204951 *T)
    sumY0  +=       1.171 * cos(0.82243324505  + 2538.49232173710 *T)
    sumY0  +=       1.155 * cos(0.22992677973  + 308.77138061390 *T)
    sumY0  +=       0.978 * cos(6.24942495607  + 394.70174530910 *T)
    sumY0  +=       0.967 * cos(1.76680956806  + 1332.99171270940 *T)
    sumY0  +=       1.232 * cos(4.85507772888  + 10.05112325500 *T)
    sumY0  +=       0.962 * cos(3.60690644272  + 416.80770005190 *T)
    sumY0  +=       1.017 * cos(3.31763307079  + 714.29205504590 *T)
    sumY0  +=       1.065 * cos(2.69433700165  + 1961.84329875950 *T)
    sumY0  +=       0.951 * cos(3.58348892983  + 1760.07753855241 *T)
    sumY0  +=       0.990 * cos(5.04804922105  + 6328.08121061550 *T)
    sumY0  +=       1.236 * cos(2.70420266349  + 263.26416554441 *T)
    sumY0  +=       1.104 * cos(3.69574800420  + 394.86180400350 *T)
    sumY0  +=       0.956 * cos(4.35161784148  + 1022.06885865090 *T)
    sumY0  +=       1.275 * cos(5.52915230471  + 104.33342944339 *T)
    sumY0  +=       1.056 * cos(0.84340788500  + 62.28279475750 *T)
    sumY0  +=       1.143 * cos(3.68309637121  + 511.77553531340 *T)
    sumY0  +=       0.941 * cos(6.19459172994  + 555.31064270760 *T)
    sumY0  +=       1.191 * cos(1.70227754284  + 2313.65989106819 *T)
    sumY0  +=       1.170 * cos(2.35872619773  + 331.92905485220 *T)
    sumY0  +=       1.072 * cos(3.76957511946  + 486.90887795820 *T)
    sumY0  +=       1.059 * cos(1.96179056600  + 1471.24608502170 *T)
    sumY0  +=       1.044 * cos(3.32798336862  + 841.90686400460 *T)
    sumY0  +=       0.998 * cos(4.27737444218  + 522.28541650049 *T)
    sumY0  +=       1.221 * cos(4.50718835641  + 473.31243127551 *T)
    sumY0  +=       0.926 * cos(0.74941072900  + 179.11502801070 *T)
    sumY0  +=       1.089 * cos(1.63384615205  + 1986.97308067310 *T)
    sumY0  +=       0.950 * cos(5.54909275925  + 899.01685027420 *T)
    sumY0  +=       1.071 * cos(5.82307444922  + 781.83999440590 *T)
    sumY0  +=       0.998 * cos(1.01524658390  + 1151.93594618579 *T)
    sumY0  +=       1.082 * cos(3.67990060666  + 2151.25100677631 *T)
    sumY0  +=       1.057 * cos(6.03036726602  + 3708.08057314570 *T)
    sumY0  +=       0.916 * cos(1.06658994239  + 386.00065973601 *T)
    sumY0  +=       1.200 * cos(6.08851819874  + 9793.55708485230 *T)
    sumY0  +=       1.139 * cos(0.28838108115  + 5845.18814136799 *T)
    sumY0  +=       1.219 * cos(3.04610895768  + 25771.75508940110 *T)
    sumY0  +=       1.118 * cos(6.06188128537  + 1080.07261679780 *T)
    sumY0  +=       1.043 * cos(3.09698136153  + 6261.49639204090 *T)
    sumY0  +=       1.172 * cos(1.30540118907  + 160.36507991500 *T)
    sumY0  +=       0.906 * cos(0.26704685647  + 1475.44136284320 *T)
    sumY0  +=       0.994 * cos(3.40772335472  + 1467.57705631699 *T)
    sumY0  +=       0.982 * cos(4.97517294679  + 1460.46350931620 *T)
    sumY0  +=       1.193 * cos(4.90565001031  + 1439.75351563271 *T)
    sumY0  +=       1.075 * cos(0.22356798470  + 3171.27606105030 *T)
    sumY0  +=       0.899 * cos(5.66544916782  + 606.20085385370 *T)
    sumY0  +=       1.230 * cos(1.03855193058  + 5334.14405850510 *T)
    sumY0  +=       0.928 * cos(0.11962988517  + 401.29402744810 *T)
    sumY0  +=       1.122 * cos(4.90662779651  + 81.99595069970 *T)
    sumY0  +=       0.901 * cos(0.25563898411  + 428.80205394729 *T)
    sumY0  +=       0.912 * cos(2.98585324356  + 726.07146849269 *T)
    sumY0  +=       1.205 * cos(5.66445689657  + 734.66948820180 *T)
    sumY0  +=       1.047 * cos(1.79702653012  + 1304.68053705810 *T)
    sumY0  +=       1.133 * cos(0.34858879514  + 733.18501549349 *T)
    sumY0  +=       1.109 * cos(4.52514855024  + 1259.43146229200 *T)
    sumY0  +=       0.888 * cos(1.42651506504  + 4.42204896250 *T)
    sumY0  +=       0.924 * cos(4.20159262156  + 633.50312967620 *T)
    sumY0  +=       0.924 * cos(4.20159262156  + 633.99076464320 *T)
    sumY0  +=       0.931 * cos(2.99561921123  + 529.82290799351 *T)
    sumY0  +=       0.956 * cos(0.90285346431  + 1198.14324454341 *T)
    sumY0  +=       1.118 * cos(3.26584053082  + 26.10731258001 *T)
    sumY0  +=       0.962 * cos(0.20948717312  + 770.58184582090 *T)
    sumY0  +=       1.092 * cos(2.42955422424  + 947.28726008359 *T)
    sumY0  +=       0.899 * cos(4.43911097152  + 9896.64985907089 *T)
    sumY0  +=       1.114 * cos(3.28101333017  + 210.08633253789 *T)
    sumY0  +=       0.935 * cos(3.52348683224  + 97.21988244259 *T)
    sumY0  +=       0.909 * cos(4.41420701109  + 981.17512040070 *T)
    sumY0  +=       0.907 * cos(5.10897715653  + 3679.13875017550 *T)
    sumY0  +=       1.078 * cos(3.68441956491  + 429.53576713020 *T)
    sumY0  +=       1.223 * cos(0.88035865973  + 876.85850590470 *T)
    sumY0  +=       1.053 * cos(3.77177323349  + 828.16740497000 *T)
    sumY0  +=       1.159 * cos(0.23541100073  + 979.42752313400 *T)
    sumY0  +=       1.179 * cos(3.99905255489  + 261.23717611490 *T)
    sumY0  +=       0.872 * cos(3.49839767525  + 102.68889804070 *T)
    sumY0  +=       0.902 * cos(4.43756731616  + 1407.51018671831 *T)
    sumY0  +=       0.970 * cos(2.49664879820  + 51322.36608391289 *T)
    sumY0  +=       0.954 * cos(2.92237254467  + 1527.86843632430 *T)
    sumY0  +=       0.870 * cos(0.48739102730  + 103.00901542950 *T)
    sumY0  +=       1.100 * cos(0.91640584339  + 1229.89893849079 *T)
    sumY0  +=       0.980 * cos(5.96363240537  + 1246.90128931980 *T)
    sumY0  +=       0.980 * cos(5.96363240537  + 1246.41365435280 *T)
    sumY0  +=       1.025 * cos(1.66758854985  + 1187.09754427950 *T)
    sumY0  +=       0.899 * cos(2.11281651742  + 663.55941046670 *T)
    sumY0  +=       0.966 * cos(0.22495487214  + 607.00400300581 *T)
    sumY0  +=       0.974 * cos(4.98865285908  + 600.78422788230 *T)
    sumY0  +=       1.052 * cos(4.65227669185  + 1901.77642916080 *T)
    sumY0  +=       0.939 * cos(3.15077321172  + 12573.50906446710 *T)
    sumY0  +=       0.875 * cos(0.65916330329  + 521.85802773080 *T)
    sumY0  +=       0.927 * cos(4.84160265007  + 3583.64715792471 *T)
    sumY0  +=       0.999 * cos(3.46495043245  + 2442.51309451930 *T)
    sumY0  +=       0.868 * cos(0.44884060339  + 449.73655803940 *T)
    sumY0  +=       0.920 * cos(1.89718815697  + 184.60108995130 *T)
    sumY0  +=       0.916 * cos(3.09091091191  + 860.06611126930 *T)
    sumY0  +=       1.151 * cos(0.35064710980  + 775.68965525160 *T)
    sumY0  +=       1.064 * cos(1.71183346254  + 9775.39783758759 *T)
    sumY0  +=       0.991 * cos(3.19380563552  + 447.55200204300 *T)
    sumY0  +=       0.934 * cos(5.99924499521  + 52168.93655363109 *T)
    sumY0  +=       0.852 * cos(4.59376258158  + 4760.34895633411 *T)
    sumY0  +=       0.879 * cos(1.78293804332  + 260.60403666850 *T)
    sumY0  +=       0.961 * cos(1.02849557968  + 10529.92123335110 *T)
    sumY0  +=       0.878 * cos(6.24767929010  + 166.48178917450 *T)
    sumY0  +=       1.023 * cos(1.68125966381  + 517.40460960590 *T)
    sumY0  +=       1.056 * cos(2.51791980526  + 622.24498109120 *T)
    sumY0  +=       0.857 * cos(5.19864310731  + 905.64276230799 *T)
    sumY0  +=       0.858 * cos(2.80895292292  + 866.77352135980 *T)
    sumY0  +=       0.960 * cos(1.81662419806  + 5842.00674763030 *T)
    sumY0  +=       1.006 * cos(0.22553531389  + 1332.72858815100 *T)
    sumY0  +=       0.859 * cos(1.05539301878  + 1886.79857563380 *T)
    sumY0  +=       1.053 * cos(2.62039491434  + 527.48710202330 *T)
    sumY0  +=       0.937 * cos(5.71553086559  + 597.60283414461 *T)
    sumY0  +=       0.858 * cos(5.12276714838  + 1035.24672528451 *T)
    sumY0  +=       1.183 * cos(3.85053522766  + 115.37912970730 *T)
    sumY0  +=       0.920 * cos(6.01872357390  + 589.73852761840 *T)
    sumY0  +=       0.920 * cos(6.01872357391  + 589.25089265140 *T)
    sumY0  +=       0.852 * cos(2.29299187318  + 146.13798591339 *T)
    sumY0  +=       0.852 * cos(2.29299187318  + 146.62562088040 *T)
    sumY0  +=       0.844 * cos(0.96512635264  + 10.27812579110 *T)
    sumY0  +=       0.844 * cos(0.96512635264  + 9.79049082409 *T)
    sumY0  +=       0.959 * cos(0.37181994910  + 177.15498242291 *T)
    sumY0  +=       0.959 * cos(1.37621356390  + 249.93084342011 *T)
    sumY0  +=       1.035 * cos(5.74473561998  + 507.35574708330 *T)
    sumY0  +=       0.988 * cos(0.40300615994  + 67.94549894479 *T)
    sumY0  +=       1.133 * cos(0.21196509862  + 36.40474544600 *T)
    sumY0  +=       0.843 * cos(4.90521975782  + 178.37882270030 *T)
    sumY0  +=       0.843 * cos(3.12599906236  + 248.70700314271 *T)
    sumY0  +=       0.836 * cos(1.54769372814  + 951.21941334670 *T)
    sumY0  +=       1.000 * cos(1.57140164961  + 598.59967188590 *T)
    sumY0  +=       0.870 * cos(6.28110425807  + 2520.33307447240 *T)
    sumY0  +=       1.010 * cos(1.72376746961  + 525.26891613211 *T)
    sumY0  +=       0.869 * cos(3.26836848865  + 803.30074761480 *T)
    sumY0  +=       0.862 * cos(5.61027073627  + 1336.66074141410 *T)
    sumY0  +=       0.865 * cos(4.28123305011  + 1535.24510788349 *T)
    sumY0  +=       1.156 * cos(5.35022886415  + 988.02554284310 *T)
    sumY0  +=       0.854 * cos(4.77835810088  + 3119.95596677750 *T)
    sumY0  +=       0.854 * cos(1.63676544729  + 3120.44360174450 *T)
    sumY0  +=       0.959 * cos(1.58204951779  + 167.80869531589 *T)
    sumY0  +=       0.843 * cos(1.22694695097  + 769.57294599210 *T)
    sumY0  +=       0.819 * cos(0.61008368054  + 20006.84263106329 *T)
    sumY0  +=       0.871 * cos(0.82479731167  + 2015.91490364330 *T)
    sumY0  +=       1.145 * cos(1.02486744412  + 982.39646855060 *T)
    sumY0  +=       0.967 * cos(0.22051521316  + 274.09741748721 *T)
    sumY0  +=       0.811 * cos(1.93988314156  + 310.95842873781 *T)
    sumY0  +=       1.076 * cos(2.12374765450  + 195.64679021520 *T)
    sumY0  +=       0.973 * cos(1.96085545938  + 480.00777927849 *T)
    sumY0  +=       0.927 * cos(0.69113180699  + 221.92848430950 *T)
    sumY0  +=       0.927 * cos(4.19849435960  + 205.15734153350 *T)
    sumY0  +=       1.043 * cos(5.37912932170  + 1054.47224625970 *T)
    sumY0  +=       1.118 * cos(0.51162175015  + 241.36645360580 *T)
    sumY0  +=       1.087 * cos(5.65735430058  + 26404.53882871430 *T)
    sumY0  +=       0.843 * cos(0.59254220586  + 9470.05166819489 *T)
    sumY0  +=       0.799 * cos(3.81029938361  + 556.27385055410 *T)
    sumY0  +=       0.916 * cos(3.24185708422  + 454.66554904380 *T)
    sumY0  +=       0.867 * cos(5.47094937361  + 155.43608891061 *T)
    sumY0  +=       1.078 * cos(3.72328147300  + 903.16145185841 *T)
    sumY0  +=       0.950 * cos(1.58758889388  + 469.48334387639 *T)
    sumY0  +=       0.972 * cos(6.06333531011  + 317.15695200191 *T)
    sumY0  +=       0.793 * cos(2.39958976528  + 529.88659846831 *T)
    sumY0  +=       0.798 * cos(0.89814256581  + 438.14834105430 *T)
    sumY0  +=       0.885 * cos(2.38654260174  + 685.23011986920 *T)
    sumY0  +=       0.791 * cos(6.20377698011  + 206.46299581550 *T)
    sumY0  +=       0.919 * cos(0.63974984193  + 524.25752417581 *T)
    sumY0  +=       0.919 * cos(0.63974984193  + 523.76988920880 *T)
    sumY0  +=       0.841 * cos(4.14693055275  + 3590.76070492550 *T)
    sumY0  +=       0.841 * cos(1.00533789916  + 3590.27306995850 *T)
    sumY0  +=       1.022 * cos(1.76356572435  + 731.92754532131 *T)
    sumY0  +=       0.978 * cos(6.10063170165  + 443.61984877990 *T)
    sumY0  +=       0.923 * cos(5.46332854009  + 261.81332273089 *T)
    sumY0  +=       0.860 * cos(0.47994747274  + 1784.71968549900 *T)
    sumY0  +=       0.803 * cos(4.00101244392  + 19992.61553706169 *T)
    sumY0  +=       0.948 * cos(4.20749999937  + 3280.99474730270 *T)
    sumY0  +=       0.852 * cos(0.10993118074  + 548.09402984280 *T)
    sumY0  +=       0.873 * cos(0.85145808105  + 63.60395540450 *T)
    sumY0  +=       0.956 * cos(0.19219049494  + 108.47803102759 *T)
    sumY0  +=       0.972 * cos(1.58575475783  + 892.79707515070 *T)
    sumY0  +=       0.926 * cos(3.30309363763  + 1025.68549772890 *T)
    sumY0  +=       0.914 * cos(2.39725818747  + 832.34563644899 *T)
    sumY0  +=       1.030 * cos(6.26161149754  + 586.55713388070 *T)
    sumY0  +=       0.809 * cos(1.59398504657  + 1235.36795408890 *T)
    sumY0  +=       0.809 * cos(3.92711557341  + 1310.14955265620 *T)
    sumY0  +=       0.784 * cos(6.04770590451  + 1699.52303398670 *T)
    sumY0  +=       1.020 * cos(5.48510992166  + 102.96083131970 *T)
    sumY0  +=       0.817 * cos(2.15964049160  + 5.90652167079 *T)
    sumY0  +=       1.007 * cos(4.32504562762  + 740.05248427840 *T)
    sumY0  +=       0.872 * cos(4.08644694319  + 63.38020623529 *T)
    sumY0  +=       0.934 * cos(4.71700851300  + 721.89323701370 *T)
    sumY0  +=       0.780 * cos(0.21966320366  + 930.77254422160 *T)
    sumY0  +=       1.031 * cos(2.47713291345  + 738.08037660310 *T)
    sumY0  +=       0.793 * cos(4.26011752121  + 746.41527175380 *T)
    sumY0  +=       1.085 * cos(1.02451872519  + 592.44684060410 *T)
    sumY0  +=       0.961 * cos(2.67254071371  + 1120.39270055950 *T)
    sumY0  +=       0.888 * cos(1.70195193954  + 26514.74514993370 *T)
    sumY0  +=       0.782 * cos(5.68415836770  + 519.63984183961 *T)
    sumY0  +=       0.800 * cos(3.20991142806  + 4113.33812301930 *T)
    sumY0  +=       0.945 * cos(4.06069039712  + 1673.76260475420 *T)
    sumY0  +=       0.850 * cos(0.31047663250  + 580.33735875721 *T)
    sumY0  +=       1.030 * cos(1.16963154394  + 3657.24811383991 *T)
    sumY0  +=       0.775 * cos(3.00079525151  + 735.58451193849 *T)
    sumY0  +=       0.775 * cos(3.00079525151  + 736.07214690550 *T)
    sumY0  +=       0.951 * cos(4.36992299634  + 604.22874617840 *T)
    sumY0  +=       0.760 * cos(4.77344534648  + 3267.25528826810 *T)
    sumY0  +=       0.802 * cos(4.44315712396  + 362.61847508910 *T)
    sumY0  +=       0.866 * cos(5.12332739411  + 1293.63483679419 *T)
    sumY0  +=       0.766 * cos(1.96682285527  + 188.67625556630 *T)
    sumY0  +=       0.768 * cos(1.62759212320  + 845.57589270931 *T)
    sumY0  +=       0.766 * cos(4.46757685640  + 2439.33170078160 *T)
    sumY0  +=       0.769 * cos(5.87087902642  + 1051.97638159510 *T)
    sumY0  +=       0.776 * cos(2.78600818951  + 2014.27037224060 *T)
    sumY0  +=       0.771 * cos(2.78598583977  + 1052.07274981469 *T)
    sumY0  +=       0.822 * cos(2.84834036260  + 582.62498061761 *T)
    sumY0  +=       1.002 * cos(2.61286504444  + 52183.16364763270 *T)
    sumY0  +=       0.828 * cos(4.17948532199  + 631.78916230430 *T)
    sumY0  +=       0.973 * cos(3.91693566169  + 519.14971474510 *T)
    sumY0  +=       0.839 * cos(5.64763751852  + 13362.69352428270 *T)
    sumY0  +=       0.842 * cos(4.82173037387  + 204.45725824539 *T)
    sumY0  +=       0.756 * cos(1.58063538394  + 693.82813957829 *T)
    sumY0  +=       0.756 * cos(1.58063538394  + 694.31577454531 *T)
    sumY0  +=       0.747 * cos(4.53970527389  + 962.75274857760 *T)
    sumY0  +=       0.916 * cos(0.53251818292  + 688.41151360689 *T)
    sumY0  +=       0.769 * cos(3.11155934055  + 6058.48723680599 *T)
    sumY0  +=       0.756 * cos(0.54984492104  + 545.71578525720 *T)
    sumY0  +=       0.830 * cos(3.49429956542  + 595.47066777489 *T)
    sumY0  +=       0.831 * cos(5.48519211288  + 436.87631666710 *T)
    sumY0  +=       0.776 * cos(2.95329359189  + 187.99493201010 *T)
    sumY0  +=       0.757 * cos(2.65993029948  + 5496.22546986590 *T)
    sumY0  +=       0.977 * cos(4.63633631623  + 1055.43296197871 *T)
    sumY0  +=       0.737 * cos(5.11234946130  + 225.58545092669 *T)
    sumY0  +=       0.940 * cos(1.37204266374  + 462.00609858071 *T)
    sumY0  +=       0.881 * cos(4.14004497053  + 157.67132114430 *T)
    sumY0  +=       0.881 * cos(3.89117384965  + 269.41450469871 *T)
    sumY0  +=       0.768 * cos(4.54804861178  + 2093.24724862940 *T)
    sumY0  +=       0.945 * cos(3.54564231189  + 347.65517577710 *T)
    sumY0  +=       0.789 * cos(1.23325676313  + 678.27413943531 *T)
    sumY0  +=       0.894 * cos(3.71784833817  + 185.08872491831 *T)
    sumY0  +=       0.887 * cos(1.90161032677  + 938.63685074779 *T)
    sumY0  +=       0.848 * cos(1.73960670139  + 1016.59984305279 *T)
    sumY0  +=       0.778 * cos(4.72868742196  + 5753.62870238030 *T)
    sumY0  +=       0.778 * cos(4.72868742196  + 5753.14106741329 *T)
    sumY0  +=       0.843 * cos(2.59704127717  + 976.49220761220 *T)
    sumY0  +=       0.900 * cos(3.76941852958  + 655.17383907870 *T)
    sumY0  +=       0.730 * cos(4.74108034798  + 497.95457822210 *T)
    sumY0  +=       0.879 * cos(4.77364347651  + 494.23487328010 *T)
    sumY0  +=       0.728 * cos(2.60603774003  + 995.88984936931 *T)
    sumY0  +=       0.729 * cos(0.77581736648  + 632.58810593949 *T)
    sumY0  +=       0.729 * cos(0.77581736648  + 633.07574090651 *T)
    sumY0  +=       0.841 * cos(2.45740163920  + 1051.06135785839 *T)
    sumY0  +=       0.810 * cos(4.65306063985  + 3900.03902758130 *T)
    sumY0  +=       0.774 * cos(3.20387240385  + 451.48415530609 *T)
    sumY0  +=       0.792 * cos(3.21012047679  + 193.93282284329 *T)
    sumY0  +=       0.894 * cos(5.93934831263  + 103.28840759231 *T)
    sumY0  +=       0.744 * cos(4.19216354937  + 662.77502104650 *T)
    sumY0  +=       1.019 * cos(2.49847463002  + 7232.49527644471 *T)
    sumY0  +=       0.720 * cos(1.56486927863  + 103711.47146249928 *T)
    sumY0  +=       0.776 * cos(0.11554513153  + 1703.45518724980 *T)
    sumY0  +=       1.001 * cos(0.71986534852  + 1126.54303971381 *T)
    sumY0  +=       0.860 * cos(0.28002712469  + 2748.61002343740 *T)
    sumY0  +=       0.716 * cos(4.37859929602  + 4319.03603648950 *T)
    sumY0  +=       0.716 * cos(4.37859929602  + 4319.52367145650 *T)
    sumY0  +=       0.797 * cos(5.59551319827  + 529.98296668790 *T)
    sumY0  +=       0.729 * cos(4.65941028609  + 1649.38358236600 *T)
    sumY0  +=       0.803 * cos(4.13945714434  + 151.01630068050 *T)
    sumY0  +=       0.765 * cos(3.49230150581  + 1379.46213562540 *T)
    sumY0  +=       0.765 * cos(3.49230150581  + 1379.94977059240 *T)
    sumY0  +=       0.749 * cos(2.29649366933  + 785.77214766900 *T)
    sumY0  +=       0.836 * cos(2.51999814689  + 4010.24534880071 *T)
    sumY0  +=       0.727 * cos(3.33280142284  + 608.16089944150 *T)
    sumY0  +=       0.727 * cos(0.19120876925  + 608.64853440851 *T)
    sumY0  +=       0.719 * cos(5.79361123899  + 467.68335698301 *T)
    sumY0  +=       0.709 * cos(0.59726160167  + 103.54904002320 *T)
    sumY0  +=       0.709 * cos(3.73885425526  + 103.06140505620 *T)
    sumY0  +=       0.920 * cos(1.80869206570  + 870.70567462290 *T)
    sumY0  +=       0.716 * cos(4.59064021199  + 664.78325074409 *T)
    sumY0  +=       0.857 * cos(0.02537195630  + 2306.54634406740 *T)
    sumY0  +=       0.724 * cos(3.13275093039  + 740.52556503041 *T)
    sumY0  +=       0.708 * cos(3.15032913377  + 2431.46739425540 *T)
    sumY0  +=       0.706 * cos(0.24440760951  + 102.63650841399 *T)
    sumY0  +=       0.706 * cos(3.38600026310  + 103.12414338101 *T)
    sumY0  +=       0.824 * cos(5.06503970891  + 1163.68172973780 *T)
    sumY0  +=       0.722 * cos(1.50443824904  + 171.21709158970 *T)
    sumY0  +=       0.723 * cos(2.31650700540  + 829.37669103239 *T)
    sumY0  +=       0.720 * cos(0.23831524300  + 255.86873425330 *T)
    sumY0  +=       0.956 * cos(0.04233388408  + 3068.18328683170 *T)
    sumY0  +=       0.800 * cos(3.03519680815  + 279.72649177971 *T)
    sumY0  +=       0.800 * cos(4.99602201203  + 147.35933406330 *T)
    sumY0  +=       0.868 * cos(0.37647106732  + 669.40093308030 *T)
    sumY0  +=       0.769 * cos(5.99489096281  + 3465.83965473750 *T)
    sumY0  +=       0.745 * cos(2.84189249473  + 843.14525849700 *T)
    sumY0  +=       0.848 * cos(2.77841948796  + 103.38477581190 *T)
    sumY0  +=       0.703 * cos(4.75219753735  + 4003.13180179990 *T)
    sumY0  +=       0.719 * cos(4.53501064500  + 9988.69693302559 *T)
    sumY0  +=       0.818 * cos(5.15275579739  + 311.48218572709 *T)
    sumY0  +=       0.901 * cos(0.79248369499  + 937.88609122240 *T)
    sumY0  +=       0.783 * cos(5.11638993720  + 199.83708378171 *T)
    sumY0  +=       0.783 * cos(6.05642153657  + 227.24874206131 *T)
    sumY0  +=       0.909 * cos(0.64248903432  + 4009.75771383370 *T)
    sumY0  +=       0.711 * cos(1.76892643060  + 457.37386202950 *T)
    sumY0  +=       0.705 * cos(3.12679135200  + 5653.98044645779 *T)
    sumY0  +=       0.712 * cos(6.07513529815  + 354.28108781090 *T)
    sumY0  +=       0.733 * cos(0.21809967139  + 37.62858572340 *T)
    sumY0  +=       0.708 * cos(1.27557908922  + 840.69757794220 *T)
    sumY0  +=       0.708 * cos(1.27557908922  + 840.20994297520 *T)
    sumY0  +=       0.695 * cos(6.08215837207  + 812.70191647599 *T)
    sumY0  +=       0.695 * cos(2.94056571848  + 813.18955144300 *T)
    sumY0  +=       0.736 * cos(5.91292146424  + 403.94745867960 *T)
    sumY0  +=       0.733 * cos(2.29907033586  + 182.03578931750 *T)
    sumY0  +=       0.836 * cos(5.01712963721  + 2331.81913833290 *T)
    sumY0  +=       0.928 * cos(0.59357203459  + 842.13386654071 *T)
    sumY0  +=       0.768 * cos(3.70421774103  + 1460.20038475780 *T)
    sumY0  +=       0.685 * cos(4.29010164874  + 455.29619636271 *T)
    sumY0  +=       0.792 * cos(3.19395503975  + 5422.52210375510 *T)
    sumY0  +=       0.837 * cos(3.85260493508  + 570.77613120160 *T)
    sumY0  +=       0.682 * cos(3.31484357591  + 3693.85347914410 *T)
    sumY0  +=       0.752 * cos(0.56556241430  + 413.34862754081 *T)
    sumY0  +=       0.817 * cos(3.10337024087  + 391.73279989249 *T)
    sumY0  +=       0.681 * cos(0.83323807589  + 458.33457774851 *T)
    sumY0  +=       0.803 * cos(0.28965764669  + 418.49006686629 *T)
    sumY0  +=       0.681 * cos(1.95625262651  + 414.57495994570 *T)

    # SATURN Y Coord
    sumY1  = 0
    sumY1  +=  9285877.988 * cos(5.32917891541  + 0.24381748350 *T)
    sumY1  +=  3086501.680 * cos(2.70413999679  + 426.84200835950 *T)
    sumY1  +=  2728479.923 * cos(4.27396756341  + 206.42936592071 *T)
    sumY1  +=  2644990.371 * cos(3.76176749725  + 220.65645992230 *T)
    sumY1  +=  629201.988 * cos(5.03696656027  + 7.35736448430 *T)
    sumY1  +=  256626.023 * cos(1.95374338915  + 640.14110379750 *T)
    sumY1  +=  312353.889 * cos(3.25876676291  + 419.72846135871 *T)
    sumY1  +=  189297.041 * cos(2.91534956216  + 433.95555536030 *T)
    sumY1  +=  204465.020 * cos(5.78983620425  + 213.54291292150 *T)
    sumY1  +=  118113.185 * cos(5.87589622232  + 110.45013870291 *T)
    sumY1  +=   67024.295 * cos(2.16159727865  + 316.63568714010 *T)
    sumY1  +=   66948.413 * cos(3.63141659236  + 227.77000692311 *T)
    sumY1  +=   31931.062 * cos(4.42042826334  + 199.31581891990 *T)
    sumY1  +=   24548.187 * cos(3.29398145587  + 205.94173095370 *T)
    sumY1  +=   30521.021 * cos(4.89424590805  + 14.47091148511 *T)
    sumY1  +=   22265.017 * cos(4.62323431127  + 103.33659170210 *T)
    sumY1  +=   18939.520 * cos(1.20474851114  + 853.44019923550 *T)
    sumY1  +=   18082.904 * cos(3.51824365057  + 209.61075965840 *T)
    sumY1  +=   17786.477 * cos(4.53241903621  + 217.47506618460 *T)
    sumY1  +=   16307.253 * cos(3.29827125771  + 216.72430665921 *T)
    sumY1  +=   16919.177 * cos(3.01391698262  + 633.02755679670 *T)
    sumY1  +=   15920.273 * cos(4.74692264945  + 210.36151918381 *T)
    sumY1  +=   16167.186 * cos(4.03963160527  + 323.74923414091 *T)
    sumY1  +=   14472.589 * cos(2.10335699916  + 647.25465079831 *T)
    sumY1  +=   11071.575 * cos(4.74242577082  + 117.56368570370 *T)
    sumY1  +=    9887.534 * cos(3.61558352957  + 202.49721265760 *T)
    sumY1  +=    8713.930 * cos(4.46375520442  + 224.58861318540 *T)
    sumY1  +=    6197.025 * cos(2.93543981790  + 213.05527795450 *T)
    sumY1  +=    5502.258 * cos(2.83240853648  + 441.06910236111 *T)
    sumY1  +=    5466.098 * cos(1.03278600709  + 11.28951774740 *T)
    sumY1  +=    3484.091 * cos(4.11221224874  + 6.86972951729 *T)
    sumY1  +=    3596.198 * cos(4.07360530172  + 309.03450517230 *T)
    sumY1  +=    3029.001 * cos(2.81952991155  + 522.33360061030 *T)
    sumY1  +=    3353.859 * cos(5.31984150246  + 95.73540973430 *T)
    sumY1  +=    2987.715 * cos(4.07245500326  + 309.52214013930 *T)
    sumY1  +=    2732.424 * cos(2.53884637625  + 149.80701461810 *T)
    sumY1  +=    3741.808 * cos(3.25277228789  + 412.61491435790 *T)
    sumY1  +=    2374.342 * cos(3.28677124286  + 198.82818395289 *T)
    sumY1  +=    2875.345 * cos(1.68013950973  + 426.35437339250 *T)
    sumY1  +=    2121.841 * cos(3.27302778600  + 220.16882495529 *T)
    sumY1  +=    2386.687 * cos(1.79649482549  + 109.96250373590 *T)
    sumY1  +=    1755.551 * cos(5.48701989649  + 277.27881122490 *T)
    sumY1  +=    1690.652 * cos(2.40558615589  + 422.90985509640 *T)
    sumY1  +=    1545.396 * cos(3.73180599039  + 330.86278114170 *T)
    sumY1  +=    1447.492 * cos(0.46419209839  + 4.17597074660 *T)
    sumY1  +=    1392.451 * cos(0.46330792664  + 1066.73929467350 *T)
    sumY1  +=    1331.483 * cos(3.49216178249  + 234.88355392390 *T)
    sumY1  +=    1452.348 * cos(4.64285489428  + 3.42521122121 *T)
    sumY1  +=    1285.559 * cos(0.83411182275  + 415.79630809560 *T)
    sumY1  +=    1159.539 * cos(3.03310034622  + 537.04832957890 *T)
    sumY1  +=    1128.476 * cos(3.88012638153  + 423.66061462181 *T)
    sumY1  +=    1131.769 * cos(4.43729716045  + 529.93478257810 *T)
    sumY1  +=    1127.023 * cos(1.31091242342  + 860.55374623631 *T)
    sumY1  +=    1101.803 * cos(2.50576870008  + 430.02340209721 *T)
    sumY1  +=    1061.927 * cos(2.81261932039  + 846.32665223470 *T)
    sumY1  +=    1343.360 * cos(1.41831229286  + 211.09523236671 *T)
    sumY1  +=    1331.640 * cos(0.33818120038  + 215.99059347631 *T)
    sumY1  +=    1287.848 * cos(0.51354213126  + 522.82123557730 *T)
    sumY1  +=    1050.973 * cos(1.98569718820  + 412.12727939089 *T)
    sumY1  +=     900.723 * cos(2.55020173482  + 838.72547026689 *T)
    sumY1  +=     882.663 * cos(3.84497584849  + 437.88770862340 *T)
    sumY1  +=     924.505 * cos(5.02323714134  + 743.23387801611 *T)
    sumY1  +=     876.882 * cos(4.58871940602  + 96.22304470130 *T)
    sumY1  +=     905.253 * cos(4.94395759972  + 3.68833577960 *T)
    sumY1  +=     820.808 * cos(3.17476222193  + 223.83785366000 *T)
    sumY1  +=     844.907 * cos(3.81271524339  + 625.42637482889 *T)
    sumY1  +=     760.731 * cos(5.03884679059  + 625.91400979590 *T)
    sumY1  +=     657.423 * cos(3.32272595203  + 316.14805217309 *T)
    sumY1  +=     813.631 * cos(4.73875446278  + 21.58445848590 *T)
    sumY1  +=     642.768 * cos(4.79907912774  + 1059.62574767270 *T)
    sumY1  +=     657.845 * cos(5.47081914900  + 88.62186273350 *T)
    sumY1  +=     715.042 * cos(4.55303572982  + 430.77416162260 *T)
    sumY1  +=     596.876 * cos(3.52791235957  + 124.67723270450 *T)
    sumY1  +=     472.020 * cos(2.03831166729  + 654.36819779910 *T)
    sumY1  +=     467.656 * cos(2.94972694776  + 515.22005360950 *T)
    sumY1  +=     472.526 * cos(4.89215133877  + 203.24797218300 *T)
    sumY1  +=     487.417 * cos(5.85503954683  + 736.12033101530 *T)
    sumY1  +=     402.114 * cos(2.85533132593  + 86.07111631471 *T)
    sumY1  +=     376.954 * cos(2.14121779282  + 76.50988875911 *T)
    sumY1  +=     419.410 * cos(1.17666971099  + 742.74624304909 *T)
    sumY1  +=     439.384 * cos(4.19607195732  + 301.92095817150 *T)
    sumY1  +=     458.871 * cos(2.23384913618  + 529.44714761109 *T)
    sumY1  +=     345.395 * cos(4.60405141742  + 192.20227191910 *T)
    sumY1  +=     334.209 * cos(2.28268923760  + 1155.11733992350 *T)
    sumY1  +=     330.276 * cos(5.07000098007  + 302.40859313851 *T)
    sumY1  +=     339.713 * cos(1.16737882478  + 218.20877936750 *T)
    sumY1  +=     304.950 * cos(4.89660174852  + 231.70216018621 *T)
    sumY1  +=     312.395 * cos(0.79548009041  + 629.09540353360 *T)
    sumY1  +=     311.746 * cos(3.50532316923  + 941.81824448550 *T)
    sumY1  +=     299.095 * cos(0.58884669599  + 208.87704647550 *T)
    sumY1  +=     299.955 * cos(1.74484418656  + 728.51914904750 *T)
    sumY1  +=     271.105 * cos(1.13283461800  + 288.32451148881 *T)
    sumY1  +=     260.779 * cos(5.58254809430  + 18.40306474821 *T)
    sumY1  +=     278.736 * cos(0.64654515892  + 632.53992182970 *T)
    sumY1  +=     335.527 * cos(1.50873501393  + 203.98168536590 *T)
    sumY1  +=     259.607 * cos(4.43314622304  + 10.53875822200 *T)
    sumY1  +=     263.983 * cos(1.98364017133  + 363.10611005610 *T)
    sumY1  +=     291.544 * cos(0.24051609787  + 223.10414047710 *T)
    sumY1  +=     214.376 * cos(4.46071700955  + 208.12628695010 *T)
    sumY1  +=     208.551 * cos(3.59347779220  + 218.95953889290 *T)
    sumY1  +=     208.257 * cos(2.81646625431  + 408.68276109481 *T)
    sumY1  +=     211.653 * cos(0.53562740874  + 138.76131435421 *T)
    sumY1  +=     199.845 * cos(3.86657932388  + 13.98327651810 *T)
    sumY1  +=     195.781 * cos(1.92507988390  + 52.44638055600 *T)
    sumY1  +=     192.346 * cos(4.64457398897  + 201.01273994931 *T)
    sumY1  +=     191.573 * cos(5.16049975639  + 341.01470952831 *T)
    sumY1  +=     185.995 * cos(0.07444798265  + 948.93179148630 *T)
    sumY1  +=     168.803 * cos(2.67981533571  + 831.61192326610 *T)
    sumY1  +=     170.172 * cos(5.89341892618  + 350.57593708390 *T)
    sumY1  +=     167.553 * cos(3.43189093041  + 226.07308589371 *T)
    sumY1  +=     177.510 * cos(3.37006198906  + 210.62215161470 *T)
    sumY1  +=     177.755 * cos(4.64994019967  + 216.46367422830 *T)
    sumY1  +=     157.494 * cos(0.14772881381  + 195.38366565681 *T)
    sumY1  +=     162.215 * cos(4.44587835630  + 956.53297345411 *T)
    sumY1  +=     160.622 * cos(3.78223230286  + 127.22797912329 *T)
    sumY1  +=     147.298 * cos(5.02251206802  + 9.80504503910 *T)
    sumY1  +=     177.635 * cos(4.69945913164  + 160.85271488200 *T)
    sumY1  +=     170.308 * cos(1.06394430192  + 207.91383862900 *T)
    sumY1  +=     158.062 * cos(0.92038060554  + 1059.13811270570 *T)
    sumY1  +=     138.665 * cos(2.01815199316  + 1471.50920958009 *T)
    sumY1  +=     150.976 * cos(1.56848256275  + 636.20895053440 *T)
    sumY1  +=     136.839 * cos(2.94611079491  + 544.16187657970 *T)
    sumY1  +=     163.932 * cos(0.66823863663  + 219.17198721400 *T)
    sumY1  +=     164.921 * cos(3.04782245643  + 22.33521801130 *T)
    sumY1  +=     134.204 * cos(5.97046820042  + 703.87700210090 *T)
    sumY1  +=     141.934 * cos(2.16098044416  + 202.00957769060 *T)
    sumY1  +=     153.018 * cos(5.84706089618  + 56.37853381910 *T)
    sumY1  +=     132.722 * cos(2.40576881332  + 750.34742501690 *T)
    sumY1  +=     136.617 * cos(3.21579900111  + 1258.21011414209 *T)
    sumY1  +=     124.984 * cos(2.71780492845  + 448.18264936190 *T)
    sumY1  +=     138.310 * cos(2.83136566148  + 142.69346761730 *T)
    sumY1  +=     121.140 * cos(4.87825139023  + 949.41942645330 *T)
    sumY1  +=     146.873 * cos(2.04601008330  + 845.83901726770 *T)
    sumY1  +=     135.623 * cos(5.56318477713  + 490.57790666290 *T)
    sumY1  +=     114.737 * cos(2.40589233064  + 433.46792039329 *T)
    sumY1  +=     107.506 * cos(1.17561460871  + 565.35950523021 *T)
    sumY1  +=     105.407 * cos(2.01232380969  + 209.35012722751 *T)
    sumY1  +=     123.012 * cos(1.28213994555  + 483.46435966210 *T)
    sumY1  +=     104.647 * cos(6.01500032787  + 217.73569861551 *T)
    sumY1  +=     123.445 * cos(3.28644134532  + 266.23311096100 *T)
    sumY1  +=     121.327 * cos(3.91515678280  + 618.31282782810 *T)
    sumY1  +=     121.539 * cos(0.70927003130  + 515.70768857651 *T)
    sumY1  +=      98.330 * cos(5.98392277878  + 1280.03839011150 *T)
    sumY1  +=     123.139 * cos(0.15397612176  + 12.77399045571 *T)
    sumY1  +=     106.461 * cos(1.50068457652  + 1044.91101870410 *T)
    sumY1  +=     134.338 * cos(1.70318102108  + 419.24082639170 *T)
    sumY1  +=     107.530 * cos(0.47701797778  + 424.39432780471 *T)
    sumY1  +=     114.356 * cos(1.93414468084  + 270.16526422410 *T)
    sumY1  +=      94.347 * cos(6.05574583632  + 1265.32366114289 *T)
    sumY1  +=     104.237 * cos(5.37674274043  + 70.60562782070 *T)
    sumY1  +=      88.410 * cos(0.52764897299  + 1073.85284167431 *T)
    sumY1  +=      91.547 * cos(4.90814232105  + 106.51798543980 *T)
    sumY1  +=      84.219 * cos(4.60141274305  + 113.63153244060 *T)
    sumY1  +=      91.641 * cos(0.73842419997  + 497.69145366371 *T)
    sumY1  +=      80.939 * cos(3.13738230860  + 651.18680406140 *T)
    sumY1  +=      82.304 * cos(1.71469250601  + 643.32249753521 *T)
    sumY1  +=      99.237 * cos(1.11729387995  + 639.65346883050 *T)
    sumY1  +=      73.063 * cos(2.41573369601  + 1148.00379292269 *T)
    sumY1  +=      79.913 * cos(1.43967739894  + 191.45151239371 *T)
    sumY1  +=      70.813 * cos(0.83698988794  + 251.67594855930 *T)
    sumY1  +=      72.632 * cos(6.03093469770  + 107.26874496520 *T)
    sumY1  +=      75.817 * cos(6.15403374919  + 284.39235822571 *T)
    sumY1  +=      72.415 * cos(3.44390490336  + 191.71463695210 *T)
    sumY1  +=      84.109 * cos(0.20089617421  + 219.69325207580 *T)
    sumY1  +=      66.814 * cos(3.27992928467  + 636.95971005981 *T)
    sumY1  +=      64.253 * cos(3.30962112459  + 71.09326278771 *T)
    sumY1  +=      73.347 * cos(2.74891631991  + 344.94686279140 *T)
    sumY1  +=      61.578 * cos(3.56151396455  + 479.53220639900 *T)
    sumY1  +=      60.447 * cos(1.75736763354  + 1787.90107923670 *T)
    sumY1  +=      63.039 * cos(2.40833995336  + 437.13694909800 *T)
    sumY1  +=      64.265 * cos(2.93065048354  + 1574.60198379870 *T)
    sumY1  +=      58.169 * cos(1.49924118883  + 75.02541605080 *T)
    sumY1  +=      65.491 * cos(0.66720972356  + 1375.52998236229 *T)
    sumY1  +=      57.055 * cos(2.86226169724  + 98.65617104110 *T)
    sumY1  +=      63.072 * cos(2.25676507609  + 359.17395679300 *T)
    sumY1  +=      76.825 * cos(4.97299930649  + 839.21310523390 *T)
    sumY1  +=      65.944 * cos(1.58727129595  + 207.39257376720 *T)
    sumY1  +=      65.704 * cos(1.95606870146  + 2.93757625420 *T)
    sumY1  +=      59.449 * cos(2.68338567532  + 417.28078080390 *T)
    sumY1  +=      66.539 * cos(1.76632812482  + 1162.23088692429 *T)
    sumY1  +=      53.194 * cos(3.14361310813  + 227.28237195610 *T)
    sumY1  +=      63.890 * cos(0.41831388362  + 235.63431344930 *T)
    sumY1  +=      52.968 * cos(0.91696357138  + 175.40987728371 *T)
    sumY1  +=      51.480 * cos(3.83078819944  + 416.54706762100 *T)
    sumY1  +=      53.788 * cos(5.42039296703  + 114.38229196601 *T)
    sumY1  +=      52.017 * cos(0.44065832790  + 220.70464403210 *T)
    sumY1  +=      48.475 * cos(1.74014879125  + 209.12312469139 *T)
    sumY1  +=      48.625 * cos(5.04094706317  + 63.97971578690 *T)
    sumY1  +=      47.377 * cos(5.76160390420  + 1581.71553079950 *T)
    sumY1  +=      66.468 * cos(2.96069437842  + 337.97632814250 *T)
    sumY1  +=      47.697 * cos(0.34705221570  + 352.06040979221 *T)
    sumY1  +=      50.400 * cos(6.23334184252  + 429.28968891431 *T)
    sumY1  +=      51.302 * cos(6.27400713228  + 121.49583896680 *T)
    sumY1  +=      48.534 * cos(1.51516380602  + 151.29148732640 *T)
    sumY1  +=      50.418 * cos(1.08978230509  + 181.29958400710 *T)
    sumY1  +=      52.765 * cos(2.41157347334  + 4.90968392950 *T)
    sumY1  +=      56.130 * cos(5.04459002050  + 404.75060783170 *T)
    sumY1  +=      53.587 * cos(5.28731972971  + 554.31380496631 *T)
    sumY1  +=      50.974 * cos(3.62783678854  + 934.70469748469 *T)
    sumY1  +=      44.236 * cos(1.36935810506  + 63.49208081989 *T)
    sumY1  +=      46.842 * cos(0.67542380780  + 245.78624183590 *T)
    sumY1  +=      41.530 * cos(4.46708578473  + 99.40443843901 *T)
    sumY1  +=      44.045 * cos(0.68746998587  + 205.46615807421 *T)
    sumY1  +=      40.580 * cos(1.25223032785  + 867.66729323710 *T)
    sumY1  +=      41.030 * cos(5.60558897869  + 81.50831573269 *T)
    sumY1  +=      40.882 * cos(2.19047548175  + 621.98185653281 *T)
    sumY1  +=      38.986 * cos(2.89088740116  + 131.79077970531 *T)
    sumY1  +=      45.904 * cos(5.16474077315  + 1589.31671276730 *T)
    sumY1  +=      53.423 * cos(5.90391742845  + 1052.51220067191 *T)
    sumY1  +=      41.233 * cos(1.27392005885  + 1361.30288836070 *T)
    sumY1  +=      46.867 * cos(4.41668443242  + 10.80188278040 *T)
    sumY1  +=      46.594 * cos(6.19397089476  + 323.26159917390 *T)
    sumY1  +=      37.309 * cos(1.75965892305  + 194.89603068979 *T)
    sumY1  +=      36.647 * cos(0.31898918015  + 148.32254190981 *T)
    sumY1  +=      41.194 * cos(4.30940153096  + 644.07325706060 *T)
    sumY1  +=      36.297 * cos(4.09379268805  + 445.00125562421 *T)
    sumY1  +=      36.204 * cos(1.17922364807  + 214.29367244690 *T)
    sumY1  +=      41.215 * cos(0.93604859842  + 198.56505939450 *T)
    sumY1  +=      45.013 * cos(0.39607313901  + 842.39449897160 *T)
    sumY1  +=      34.449 * cos(2.14089859356  + 1464.39566257930 *T)
    sumY1  +=      34.256 * cos(1.30837397954  + 289.80898419710 *T)
    sumY1  +=      36.820 * cos(1.40746554167  + 206.38118181090 *T)
    sumY1  +=      33.887 * cos(5.79726117026  + 89.10949770051 *T)
    sumY1  +=      36.762 * cos(5.38503749099  + 265.74547599400 *T)
    sumY1  +=      32.982 * cos(1.07867972783  + 276.79117625789 *T)
    sumY1  +=      33.116 * cos(3.67609767645  + 178.11819026941 *T)
    sumY1  +=      32.822 * cos(0.62395825570  + 212.79215339611 *T)
    sumY1  +=      37.519 * cos(0.81273715463  + 228.52076644850 *T)
    sumY1  +=      31.997 * cos(0.23044457140  + 275.79433851660 *T)
    sumY1  +=      40.913 * cos(3.86314191833  + 209.87388421679 *T)
    sumY1  +=      30.291 * cos(2.81092945022  + 78.95756931390 *T)
    sumY1  +=      29.064 * cos(4.35733221797  + 248.96763557360 *T)
    sumY1  +=      31.114 * cos(2.65185710054  + 1890.99385345530 *T)
    sumY1  +=      28.459 * cos(2.69371515649  + 102.84895673509 *T)
    sumY1  +=      35.352 * cos(0.80856131309  + 501.62360692681 *T)
    sumY1  +=      27.208 * cos(1.50128394987  + 2104.29294889330 *T)
    sumY1  +=      37.148 * cos(6.09807476590  + 729.00678401450 *T)
    sumY1  +=      33.504 * cos(5.75754294461  + 436.40323591510 *T)
    sumY1  +=      32.029 * cos(1.64757033255  + 414.31183538731 *T)
    sumY1  +=      27.479 * cos(2.28916876539  + 114.64292439690 *T)
    sumY1  +=      29.371 * cos(4.13995887820  + 215.02738562980 *T)
    sumY1  +=      30.410 * cos(0.83533364902  + 31.99951143090 *T)
    sumY1  +=      33.556 * cos(1.41630051566  + 1478.62275658089 *T)
    sumY1  +=      28.365 * cos(4.24070805056  + 355.99256305531 *T)
    sumY1  +=      24.664 * cos(2.99088916204  + 427.80521620600 *T)
    sumY1  +=      28.864 * cos(0.42973261435  + 1691.92185201890 *T)
    sumY1  +=      33.503 * cos(6.18224013517  + 735.63269604830 *T)
    sumY1  +=      27.468 * cos(3.51804419384  + 405.50136735710 *T)
    sumY1  +=      23.818 * cos(3.63547565864  + 190.96387742670 *T)
    sumY1  +=      23.542 * cos(1.41600136615  + 278.76328393321 *T)
    sumY1  +=      23.335 * cos(5.49828745732  + 1898.10740045609 *T)
    sumY1  +=      24.810 * cos(5.54564487364  + 425.87880051300 *T)
    sumY1  +=      23.968 * cos(0.36677181560  + 431.50787480550 *T)
    sumY1  +=      25.517 * cos(2.45859967105  + 69.39634175830 *T)
    sumY1  +=      22.253 * cos(1.40396872946  + 602.00806815971 *T)
    sumY1  +=      23.190 * cos(5.24458718751  + 618.80046279510 *T)
    sumY1  +=      21.849 * cos(2.16979732894  + 299.37021175271 *T)
    sumY1  +=      21.270 * cos(5.26652889356  + 917.17609753890 *T)
    sumY1  +=      22.382 * cos(1.28310720891  + 576.40520549410 *T)
    sumY1  +=      21.104 * cos(2.44842000288  + 319.81708087780 *T)
    sumY1  +=      24.718 * cos(3.34396844841  + 1251.09656714130 *T)
    sumY1  +=      20.790 * cos(3.29258290690  + 241.99710092470 *T)
    sumY1  +=      25.298 * cos(5.95012841054  + 422.17614191350 *T)
    sumY1  +=      19.818 * cos(5.84256592114  + 5.66044345490 *T)
    sumY1  +=      19.530 * cos(3.52863763848  + 313.45429340240 *T)
    sumY1  +=      23.121 * cos(2.90337571001  + 230.95140066080 *T)
    sumY1  +=      19.612 * cos(3.11172042496  + 508.10650660869 *T)
    sumY1  +=      19.333 * cos(4.45591990700  + 134.34152612410 *T)
    sumY1  +=      20.375 * cos(5.87114204897  + 426.89019246930 *T)
    sumY1  +=      20.742 * cos(3.36125223647  + 312.70353387701 *T)
    sumY1  +=      19.203 * cos(3.74233103320  + 320.56784040320 *T)
    sumY1  +=      18.481 * cos(1.91890218206  + 39.60069339870 *T)
    sumY1  +=      21.067 * cos(3.32302810096  + 421.42538238810 *T)
    sumY1  +=      19.538 * cos(6.17712302160  + 189.16389053330 *T)
    sumY1  +=      20.461 * cos(3.99739791776  + 327.68138740401 *T)
    sumY1  +=      21.171 * cos(1.20539425934  + 221.61966776881 *T)
    sumY1  +=      19.130 * cos(1.70449321765  + 98.91680347199 *T)
    sumY1  +=      18.295 * cos(5.35670142029  + 295.29504613770 *T)
    sumY1  +=      18.160 * cos(3.80256763164  + 205.67860639531 *T)
    sumY1  +=      17.549 * cos(5.90563873127  + 120.11443212250 *T)
    sumY1  +=      17.626 * cos(2.90948901796  + 956.04533848709 *T)
    sumY1  +=      16.792 * cos(1.87396327818  + 1780.78753223589 *T)
    sumY1  +=      16.458 * cos(2.73724850309  + 439.37218133171 *T)
    sumY1  +=      18.239 * cos(5.19351531548  + 6276.20612047410 *T)
    sumY1  +=      16.441 * cos(4.13905167455  + 1169.83206889211 *T)
    sumY1  +=      16.263 * cos(4.44980749629  + 1272.92484311070 *T)
    sumY1  +=      16.077 * cos(0.35334193612  + 778.65860066820 *T)
    sumY1  +=      18.919 * cos(5.63225747640  + 348.12825652910 *T)
    sumY1  +=      20.156 * cos(3.71856419261  + 212.05844021321 *T)
    sumY1  +=      17.804 * cos(4.60517305544  + 6290.43321447570 *T)
    sumY1  +=      22.054 * cos(4.33821912194  + 294.80741117069 *T)
    sumY1  +=      15.573 * cos(6.26818148592  + 74.53778108379 *T)
    sumY1  +=      16.309 * cos(1.07119798583  + 1677.69475801730 *T)
    sumY1  +=      18.944 * cos(1.64030810533  + 127.71561409031 *T)
    sumY1  +=      17.127 * cos(0.33379704300  + 10206.41581669370 *T)
    sumY1  +=      19.690 * cos(0.46228960012  + 696.76345510010 *T)
    sumY1  +=      20.455 * cos(3.68158411673  + 117.07605073670 *T)
    sumY1  +=      16.447 * cos(1.86360821642  + 237.92193530970 *T)
    sumY1  +=      16.503 * cos(2.39093241573  + 2207.38572311189 *T)
    sumY1  +=      20.378 * cos(2.67260865639  + 426.79382424970 *T)
    sumY1  +=      16.713 * cos(6.02955183779  + 10220.64291069530 *T)
    sumY1  +=      17.063 * cos(3.10475966052  + 15.22167101050 *T)
    sumY1  +=      18.187 * cos(4.63183796319  + 167.07249000551 *T)
    sumY1  +=      14.830 * cos(0.96104361429  + 415.30867312860 *T)
    sumY1  +=      15.128 * cos(6.22779956266  + 710.99054910171 *T)
    sumY1  +=      14.162 * cos(4.64972656382  + 525.74199688410 *T)
    sumY1  +=      14.139 * cos(0.42327996702  + 39.11305843170 *T)
    sumY1  +=      18.060 * cos(1.12205088158  + 1795.01462623750 *T)
    sumY1  +=      13.362 * cos(1.30455477545  + 106.03035047280 *T)
    sumY1  +=      13.903 * cos(5.77275028636  + 312.44290144610 *T)
    sumY1  +=      15.359 * cos(3.68905633066  + 182.52342428450 *T)
    sumY1  +=      13.906 * cos(2.22202764942  + 59.55992755679 *T)
    sumY1  +=      13.001 * cos(1.87269751242  + 87.55558902300 *T)
    sumY1  +=      13.054 * cos(3.33806358060  + 196.27743753410 *T)
    sumY1  +=      15.661 * cos(4.95405745189  + 1162.71852189130 *T)
    sumY1  +=      17.412 * cos(3.73906512638  + 202.76033721600 *T)
    sumY1  +=      14.557 * cos(2.97972992288  + 381.59542572090 *T)
    sumY1  +=      17.110 * cos(6.00385232407  + 432.47108265200 *T)
    sumY1  +=      12.702 * cos(1.25047898924  + 2420.68481854989 *T)
    sumY1  +=      17.195 * cos(5.38272410647  + 536.56069461190 *T)
    sumY1  +=      12.663 * cos(4.98811886581  + 1376.01761732930 *T)
    sumY1  +=      15.239 * cos(3.75258625939  + 429.76276966630 *T)
    sumY1  +=      12.562 * cos(5.93423402721  + 852.95256426850 *T)
    sumY1  +=      12.487 * cos(2.21693371299  + 757.46097201770 *T)
    sumY1  +=      12.205 * cos(1.93673292559  + 661.48174479990 *T)
    sumY1  +=      14.071 * cos(2.49266457872  + 395.82251972250 *T)
    sumY1  +=      13.589 * cos(4.88876925401  + 97.11681657861 *T)
    sumY1  +=      13.297 * cos(0.85813242995  + 849.50804597240 *T)
    sumY1  +=      12.950 * cos(1.73883852923  + 91.80325647120 *T)
    sumY1  +=      12.800 * cos(2.48384043874  + 1272.43720814370 *T)
    sumY1  +=      14.060 * cos(1.76101820041  + 116.18227885939 *T)
    sumY1  +=      13.165 * cos(6.22232430384  + 129.67565967810 *T)
    sumY1  +=      11.819 * cos(5.25276003174  + 431.03479405351 *T)
    sumY1  +=      11.579 * cos(3.65140573770  + 144.39038864671 *T)
    sumY1  +=      11.596 * cos(4.65380739810  + 230.80838830890 *T)
    sumY1  +=      12.157 * cos(4.26701496510  + 374.15181032000 *T)
    sumY1  +=      11.328 * cos(5.22634156214  + 2214.49927011270 *T)
    sumY1  +=      13.310 * cos(3.05839708528  + 1567.48843679789 *T)
    sumY1  +=      13.760 * cos(3.68502527304  + 216.98743121760 *T)
    sumY1  +=      11.244 * cos(0.07984446041  + 777.17412795991 *T)
    sumY1  +=      11.256 * cos(0.76010615725  + 1582.20316576650 *T)
    sumY1  +=      11.849 * cos(1.82233217301  + 305.58998687620 *T)
    sumY1  +=      11.016 * cos(2.47001157319  + 416.05943265400 *T)
    sumY1  +=      10.858 * cos(0.99488682955  + 422.42222012939 *T)
    sumY1  +=      14.447 * cos(1.13660537122  + 333.90116252750 *T)
    sumY1  +=      10.877 * cos(4.92169700962  + 67.42423408300 *T)
    sumY1  +=      10.671 * cos(4.02549503966  + 426.32074349770 *T)
    sumY1  +=      10.673 * cos(4.24959012167  + 221.40721944771 *T)
    sumY1  +=      13.923 * cos(1.76345867331  + 558.24595822940 *T)
    sumY1  +=      12.161 * cos(4.58759593895  + 28.69800548670 *T)
    sumY1  +=      11.528 * cos(3.24850801780  + 432.25863433090 *T)
    sumY1  +=      12.972 * cos(4.42656189319  + 196.86813836510 *T)
    sumY1  +=      10.573 * cos(2.73606095442  + 145.87486135500 *T)
    sumY1  +=      11.572 * cos(3.41892921748  + 5849.12029463110 *T)
    sumY1  +=      12.498 * cos(0.21208820040  + 2008.31372167550 *T)
    sumY1  +=      10.593 * cos(2.57833657495  + 343.46239008310 *T)
    sumY1  +=      12.818 * cos(2.76088492373  + 216.23667169220 *T)
    sumY1  +=      10.230 * cos(5.88163152868  + 199.52826724100 *T)
    sumY1  +=      10.672 * cos(5.32880327015  + 84.58664360640 *T)
    sumY1  +=      10.068 * cos(6.07595138018  + 206.95063078250 *T)
    sumY1  +=      11.596 * cos(0.26708370981  + 1155.60497489050 *T)
    sumY1  +=      10.203 * cos(5.80527534922  + 192.93598510200 *T)
    sumY1  +=       9.849 * cos(5.36800138647  + 767.61290040431 *T)
    sumY1  +=      10.252 * cos(4.92833809216  + 618.04970326970 *T)
    sumY1  +=      10.915 * cos(1.22365645831  + 200.20959079720 *T)
    sumY1  +=       9.693 * cos(2.62845142458  + 99.66756299740 *T)
    sumY1  +=       9.672 * cos(4.70950406981  + 238.81570718700 *T)
    sumY1  +=       9.664 * cos(1.75114701217  + 182.05034353251 *T)
    sumY1  +=       9.860 * cos(1.31031161383  + 572.47305223100 *T)
    sumY1  +=      10.186 * cos(5.12887420340  + 20.85074530300 *T)
    sumY1  +=      10.831 * cos(4.84548447475  + 9779.32999085069 *T)
    sumY1  +=       9.746 * cos(3.38574766621  + 185.23173727020 *T)
    sumY1  +=       9.573 * cos(2.36163965903  + 281.21096448800 *T)
    sumY1  +=      11.944 * cos(3.75743101287  + 62.00760811160 *T)
    sumY1  +=       9.779 * cos(0.88460862761  + 188.41313100791 *T)
    sumY1  +=      10.421 * cos(0.80662653956  + 404.26297286470 *T)
    sumY1  +=      10.874 * cos(4.76437370447  + 153.73916788121 *T)
    sumY1  +=      10.712 * cos(4.51706019623  + 427.36327322130 *T)
    sumY1  +=       9.203 * cos(5.30048466028  + 212.27088853431 *T)
    sumY1  +=       9.203 * cos(2.73073415990  + 214.81493730871 *T)
    sumY1  +=      10.194 * cos(2.85037947546  + 5863.34738863270 *T)
    sumY1  +=       9.371 * cos(2.14963709425  + 227.55755860200 *T)
    sumY1  +=      10.238 * cos(4.28151996864  + 244.56240155850 *T)
    sumY1  +=      10.084 * cos(0.80678073222  + 1265.81129610991 *T)
    sumY1  +=      10.456 * cos(6.15818049342  + 41.40068029210 *T)
    sumY1  +=      10.166 * cos(5.68487873505  + 143.90275367970 *T)
    sumY1  +=      11.381 * cos(6.15494704224  + 156.92056161890 *T)
    sumY1  +=      12.113 * cos(6.18154282600  + 563.87503252191 *T)
    sumY1  +=      10.043 * cos(0.52437707771  + 226.87623504580 *T)
    sumY1  +=       9.235 * cos(0.46021438011  + 159.36824217371 *T)
    sumY1  +=       8.818 * cos(2.88309374002  + 50.96190784770 *T)
    sumY1  +=       9.068 * cos(2.68400913334  + 131.64776735340 *T)
    sumY1  +=      10.089 * cos(3.30634961501  + 260.01333583751 *T)
    sumY1  +=      10.283 * cos(4.03488999891  + 46.71424039951 *T)
    sumY1  +=       9.854 * cos(2.93320887736  + 333.41352756050 *T)
    sumY1  +=       9.008 * cos(5.89793163445  + 637.69342324271 *T)
    sumY1  +=       8.926 * cos(2.22929142122  + 234.14984074100 *T)
    sumY1  +=       8.557 * cos(1.65572129459  + 963.64652045490 *T)
    sumY1  +=      10.257 * cos(1.13994818513  + 64.97655352819 *T)
    sumY1  +=       9.526 * cos(4.27706558639  + 9793.55708485230 *T)
    sumY1  +=       9.414 * cos(2.65613957745  + 67.91186905000 *T)
    sumY1  +=       9.628 * cos(3.38030516392  + 326.93062787860 *T)
    sumY1  +=      11.719 * cos(4.16368793873  + 45.33283355520 *T)
    sumY1  +=       8.304 * cos(1.17598268665  + 547.20025796550 *T)
    sumY1  +=       8.224 * cos(1.59987405703  + 2097.17940189250 *T)
    sumY1  +=       8.279 * cos(4.61172057056  + 241.85408857280 *T)
    sumY1  +=       9.189 * cos(1.68565081278  + 556.76148552110 *T)
    sumY1  +=       8.334 * cos(0.86145365394  + 238.67269483510 *T)
    sumY1  +=       8.030 * cos(0.00151213347  + 245.03548231051 *T)
    sumY1  +=      10.377 * cos(0.86654919644  + 212.64914104420 *T)
    sumY1  +=      10.377 * cos(0.88148431656  + 214.43668479881 *T)
    sumY1  +=       9.061 * cos(1.15906835210  + 206.47755003051 *T)
    sumY1  +=       8.325 * cos(2.88284393514  + 824.49837626530 *T)
    sumY1  +=       7.948 * cos(2.23981845794  + 423.17297965479 *T)
    sumY1  +=      10.520 * cos(0.59831170486  + 93.18466331550 *T)
    sumY1  +=       8.773 * cos(2.41905049963  + 347.64062156210 *T)
    sumY1  +=       8.377 * cos(2.10271496997  + 2523.77759276850 *T)
    sumY1  +=       8.448 * cos(2.16286795108  + 1588.82907780029 *T)
    sumY1  +=       8.685 * cos(3.73032873272  + 220.60827581251 *T)
    sumY1  +=       7.716 * cos(3.18762880947  + 78.20680978850 *T)
    sumY1  +=       7.794 * cos(5.75334057605  + 411.86415483250 *T)
    sumY1  +=       8.324 * cos(1.43925082351  + 646.76701583129 *T)
    sumY1  +=       7.401 * cos(5.24566403741  + 329.48137429739 *T)
    sumY1  +=       8.108 * cos(2.83411309649  + 60.52313540329 *T)
    sumY1  +=       7.383 * cos(5.70976794256  + 259.11956396020 *T)
    sumY1  +=       7.791 * cos(1.00733061725  + 174.18603700630 *T)
    sumY1  +=       7.076 * cos(6.22401926042  + 464.97504399731 *T)
    sumY1  +=       9.308 * cos(3.59155053097  + 273.34665796181 *T)
    sumY1  +=       9.741 * cos(0.81596806036  + 2111.40649589409 *T)
    sumY1  +=       7.187 * cos(2.33757896013  + 864.48589949940 *T)
    sumY1  +=       7.014 * cos(1.65947862448  + 627.61093082531 *T)
    sumY1  +=       6.966 * cos(4.30763849421  + 282.69543719630 *T)
    sumY1  +=       7.308 * cos(4.98329991415  + 942.30587945250 *T)
    sumY1  +=       6.845 * cos(5.23018796148  + 1493.33748554950 *T)
    sumY1  +=       6.779 * cos(6.03735095188  + 357.68948408471 *T)
    sumY1  +=       7.128 * cos(2.30863250694  + 134.82916109110 *T)
    sumY1  +=       7.238 * cos(4.23528950033  + 749.85979004990 *T)
    sumY1  +=       6.975 * cos(1.29870798692  + 267.71758366931 *T)
    sumY1  +=       7.167 * cos(2.38189224267  + 459.08533727390 *T)
    sumY1  +=       7.917 * cos(1.93600980484  + 23.81969071961 *T)
    sumY1  +=       6.648 * cos(6.02714942107  + 1287.15193711231 *T)
    sumY1  +=       6.800 * cos(6.17067933170  + 561.42735196710 *T)
    sumY1  +=       8.501 * cos(3.77133955917  + 224.10097821840 *T)
    sumY1  +=       6.569 * cos(6.17574800136  + 339.53023682000 *T)
    sumY1  +=       8.044 * cos(5.50051730143  + 196.13442518220 *T)
    sumY1  +=       6.518 * cos(4.64724104258  + 638.65663108921 *T)
    sumY1  +=       7.597 * cos(2.99211896084  + 186.71620997851 *T)
    sumY1  +=       8.282 * cos(5.61743026825  + 82.13896305161 *T)
    sumY1  +=       7.088 * cos(1.11525946892  + 570.98857952271 *T)
    sumY1  +=       6.908 * cos(1.59500149455  + 526.75338884040 *T)
    sumY1  +=       6.418 * cos(2.40736050441  + 106.78110999820 *T)
    sumY1  +=       6.362 * cos(0.93818716151  + 1994.08662767389 *T)
    sumY1  +=       6.216 * cos(0.45259421413  + 171.89841514590 *T)
    sumY1  +=       6.708 * cos(0.91000084421  + 508.59414157570 *T)
    sumY1  +=       7.258 * cos(2.77271169877  + 1883.88030645450 *T)
    sumY1  +=       7.633 * cos(1.81311259252  + 295.43805848960 *T)
    sumY1  +=       6.841 * cos(3.13369052026  + 329.96900926440 *T)
    sumY1  +=       5.996 * cos(5.16451756634  + 334.79493440480 *T)
    sumY1  +=       5.926 * cos(0.99533834681  + 2737.07668820650 *T)
    sumY1  +=       6.244 * cos(5.97077260556  + 533.11617631580 *T)
    sumY1  +=       6.129 * cos(4.63572550497  + 215.83053478190 *T)
    sumY1  +=       6.129 * cos(3.39549331521  + 211.25529106110 *T)
    sumY1  +=       5.970 * cos(1.16627507303  + 193.89919294850 *T)
    sumY1  +=       5.955 * cos(6.01270486078  + 128.71245183160 *T)
    sumY1  +=       7.108 * cos(6.27572187728  + 427.59276788490 *T)
    sumY1  +=       5.852 * cos(4.69412415278  + 273.82223084130 *T)
    sumY1  +=       5.748 * cos(4.36224463614  + 30.77567115350 *T)
    sumY1  +=       6.470 * cos(3.68405106611  + 162.33718759030 *T)
    sumY1  +=       5.610 * cos(5.48803250441  + 53.93085326429 *T)
    sumY1  +=       5.595 * cos(4.94109471314  + 2530.89113976930 *T)
    sumY1  +=       6.286 * cos(5.65267510028  + 432.99234751380 *T)
    sumY1  +=       5.905 * cos(0.93770342346  + 856.62159297321 *T)
    sumY1  +=       5.420 * cos(2.81752688518  + 692.83130183700 *T)
    sumY1  +=       7.291 * cos(4.02822362635  + 611.19928082730 *T)
    sumY1  +=       5.368 * cos(2.71102841423  + 113.89465699899 *T)
    sumY1  +=       6.762 * cos(0.70344873462  + 421.21293406700 *T)
    sumY1  +=       6.290 * cos(5.03830033679  + 240.36961586451 *T)
    sumY1  +=       6.519 * cos(5.71800943425  + 1052.02456570489 *T)
    sumY1  +=       5.326 * cos(1.23172948855  + 258.63192899320 *T)
    sumY1  +=       5.918 * cos(2.28750450549  + 422.64922266551 *T)
    sumY1  +=       5.209 * cos(0.61729030530  + 137.27684164590 *T)
    sumY1  +=       5.766 * cos(0.06223729120  + 272.85902299479 *T)
    sumY1  +=       6.291 * cos(3.22636360944  + 423.92124705270 *T)
    sumY1  +=       5.107 * cos(1.32100485887  + 255.18741069710 *T)
    sumY1  +=       6.557 * cos(4.83743644153  + 100.15519796440 *T)
    sumY1  +=       5.736 * cos(0.67024965264  + 252.89978883670 *T)
    sumY1  +=       5.147 * cos(5.92682066634  + 489.09343395460 *T)
    sumY1  +=       5.331 * cos(3.66635509285  + 21.09682351890 *T)
    sumY1  +=       5.351 * cos(2.70813739263  + 342.49918223660 *T)
    sumY1  +=       4.900 * cos(1.21938058089  + 533.86693584120 *T)
    sumY1  +=       6.324 * cos(2.72039134579  + 108.00245814810 *T)
    sumY1  +=       6.120 * cos(5.76815989679  + 472.41865939820 *T)
    sumY1  +=       4.858 * cos(0.64870681722  + 29.44876501211 *T)
    sumY1  +=       6.153 * cos(4.63786966741  + 125.74350641500 *T)
    sumY1  +=       4.774 * cos(6.16871058940  + 394.59867944510 *T)
    sumY1  +=       5.096 * cos(0.23957822589  + 24.13520490470 *T)
    sumY1  +=       4.756 * cos(0.62601572639  + 1898.59503542310 *T)
    sumY1  +=       5.064 * cos(1.89791339354  + 1905.22094745690 *T)
    sumY1  +=       4.696 * cos(1.25539391636  + 408.19512612779 *T)
    sumY1  +=       5.004 * cos(1.63967789274  + 650.43604453600 *T)
    sumY1  +=       4.561 * cos(1.51892035057  + 518.88908231420 *T)
    sumY1  +=       5.108 * cos(0.06425626484  + 1471.99684454710 *T)
    sumY1  +=       4.597 * cos(5.56366740468  + 291.50590522650 *T)
    sumY1  +=       4.502 * cos(2.24091480636  + 167.96626188280 *T)
    sumY1  +=       4.559 * cos(0.15524505071  + 418.76525351221 *T)
    sumY1  +=       6.043 * cos(3.19569787210  + 405.01373239010 *T)
    sumY1  +=       5.079 * cos(4.36470318246  + 264.74863825270 *T)
    sumY1  +=       5.333 * cos(4.20549686107  + 486.64575339981 *T)
    sumY1  +=       5.440 * cos(0.49338163220  + 230.21768747790 *T)
    sumY1  +=       5.845 * cos(2.31588446643  + 551.27542358051 *T)
    sumY1  +=       5.448 * cos(6.26836971962  + 2324.70559133209 *T)
    sumY1  +=       5.131 * cos(1.84143196789  + 835.28095197081 *T)
    sumY1  +=       4.374 * cos(5.36910049865  + 138.27367938720 *T)
    sumY1  +=       4.471 * cos(5.57996257341  + 361.62163734781 *T)
    sumY1  +=       4.410 * cos(2.60807357538  + 1140.89024592190 *T)
    sumY1  +=       4.114 * cos(5.22568875373  + 1486.22393854870 *T)
    sumY1  +=       4.269 * cos(1.03216096696  + 225.32232636830 *T)
    sumY1  +=       4.269 * cos(3.85746519963  + 201.76349947470 *T)
    sumY1  +=       4.076 * cos(1.53556314147  + 33.69643246030 *T)
    sumY1  +=       4.308 * cos(2.02359582943  + 430.28652665560 *T)
    sumY1  +=       4.392 * cos(2.70517815683  + 492.06237937121 *T)
    sumY1  +=       4.365 * cos(1.83148011994  + 2840.16946242510 *T)
    sumY1  +=       4.469 * cos(4.07802364782  + 428.32648106781 *T)
    sumY1  +=       4.332 * cos(2.23025483563  + 32.48714639791 *T)
    sumY1  +=       4.016 * cos(3.79263669972  + 425.35753565121 *T)
    sumY1  +=       4.203 * cos(0.85003521966  + 364.59058276440 *T)
    sumY1  +=       3.902 * cos(1.18177774441  + 397.63706083091 *T)
    sumY1  +=       3.955 * cos(2.69691973360  + 850.25880549781 *T)
    sumY1  +=       3.945 * cos(1.42294084476  + 2413.57127154910 *T)
    sumY1  +=       4.359 * cos(4.53204115879  + 312.96665843540 *T)
    sumY1  +=       5.325 * cos(4.80642075418  + 1169.34443392509 *T)
    sumY1  +=       4.743 * cos(6.22141011646  + 1055.69359440960 *T)
    sumY1  +=       5.334 * cos(0.52549183195  + 2427.79836555070 *T)
    sumY1  +=       3.914 * cos(0.63614657379  + 504.80500066450 *T)
    sumY1  +=       3.740 * cos(3.31636767466  + 658.30035106221 *T)
    sumY1  +=       4.015 * cos(3.14423865348  + 526.26575387340 *T)
    sumY1  +=       3.733 * cos(2.78876823876  + 174.92224231670 *T)
    sumY1  +=       3.719 * cos(5.55031494411  + 292.25666475190 *T)
    sumY1  +=       4.998 * cos(6.14565474272  + 448.93340888730 *T)
    sumY1  +=       3.941 * cos(5.91209030576  + 434.00373947010 *T)
    sumY1  +=       3.629 * cos(4.59247762170  + 1692.40948698591 *T)
    sumY1  +=       4.488 * cos(0.55981749277  + 410.16723380310 *T)
    sumY1  +=       4.487 * cos(0.40749830647  + 832.09955823310 *T)
    sumY1  +=       4.724 * cos(5.59694447702  + 205.90810105891 *T)
    sumY1  +=       3.566 * cos(4.66247460443  + 215.23983395090 *T)
    sumY1  +=       3.566 * cos(0.22715156216  + 211.84599189210 *T)
    sumY1  +=       4.241 * cos(2.24120341965  + 370.21965705690 *T)
    sumY1  +=       3.740 * cos(4.78584287004  + 471.93102443120 *T)
    sumY1  +=       3.762 * cos(4.69658599509  + 320.08020543620 *T)
    sumY1  +=       4.157 * cos(5.31767100280  + 2119.00767786191 *T)
    sumY1  +=       3.518 * cos(4.97655806515  + 33.38091827520 *T)
    sumY1  +=       3.584 * cos(4.22091904260  + 77.99436146740 *T)
    sumY1  +=       3.636 * cos(2.93926204397  + 406.23508054000 *T)
    sumY1  +=       3.529 * cos(2.23289517847  + 732.45130231059 *T)
    sumY1  +=       3.573 * cos(2.33382239586  + 440.58146739410 *T)
    sumY1  +=       4.086 * cos(2.51586258437  + 2200.27217611110 *T)
    sumY1  +=       4.527 * cos(5.57565276280  + 221.17772478411 *T)
    sumY1  +=       3.974 * cos(1.28299868905  + 16.91859203990 *T)
    sumY1  +=       3.913 * cos(2.55458853045  + 189.96703968541 *T)
    sumY1  +=       3.397 * cos(6.02142949721  + 245.29860686890 *T)
    sumY1  +=       3.967 * cos(3.61115095966  + 120.74507944141 *T)
    sumY1  +=       3.785 * cos(3.46924321586  + 215.18744432420 *T)
    sumY1  +=       3.785 * cos(4.56197560432  + 211.89838151880 *T)
    sumY1  +=       3.713 * cos(0.35483179852  + 203.50860461390 *T)
    sumY1  +=       3.713 * cos(4.53479436807  + 223.57722122910 *T)
    sumY1  +=       3.167 * cos(6.20784254152  + 224.84924561630 *T)
    sumY1  +=       3.167 * cos(4.96496893225  + 202.23658022670 *T)
    sumY1  +=       4.027 * cos(0.71647686755  + 38.37685312130 *T)
    sumY1  +=       3.454 * cos(2.49806053768  + 542.78046973539 *T)
    sumY1  +=       3.700 * cos(3.32496352620  + 170.25388374320 *T)
    sumY1  +=       3.257 * cos(4.72035017193  + 305.85311143459 *T)
    sumY1  +=       3.858 * cos(2.88746700144  + 135.57992061650 *T)
    sumY1  +=       3.174 * cos(4.87604536178  + 340.52707456129 *T)
    sumY1  +=       3.287 * cos(2.89671334193  + 10.05112325500 *T)
    sumY1  +=       4.371 * cos(5.86032930399  + 1045.39865367110 *T)
    sumY1  +=       3.220 * cos(5.02019963681  + 629.84616305900 *T)
    sumY1  +=       3.146 * cos(1.41712289631  + 553.82616999929 *T)
    sumY1  +=       3.544 * cos(4.11156199752  + 1066.25165970650 *T)
    sumY1  +=       4.149 * cos(1.28803789298  + 136.78920667889 *T)
    sumY1  +=       4.004 * cos(0.98922990708  + 494.51005992600 *T)
    sumY1  +=       3.653 * cos(5.78563978576  + 466.19888427470 *T)
    sumY1  +=       3.162 * cos(1.73694976405  + 2221.61281711350 *T)
    sumY1  +=       3.678 * cos(6.16970207201  + 418.24398865040 *T)
    sumY1  +=       3.135 * cos(0.41052536886  + 409.43352062020 *T)
    sumY1  +=       3.067 * cos(0.89802837117  + 441.81986188650 *T)
    sumY1  +=       2.970 * cos(0.88262512852  + 731.70054278519 *T)
    sumY1  +=       3.379 * cos(5.48131520562  + 237.11878615761 *T)
    sumY1  +=       3.100 * cos(4.13659867927  + 407.19828838650 *T)
    sumY1  +=       3.007 * cos(5.96327513578  + 92.55401599659 *T)
    sumY1  +=       2.898 * cos(0.50716798363  + 1080.96638867510 *T)
    sumY1  +=       3.170 * cos(2.60807868236  + 105.76971804189 *T)
    sumY1  +=       2.943 * cos(2.73197794646  + 256.08118257441 *T)
    sumY1  +=       3.935 * cos(2.13947573521  + 426.09124883410 *T)
    sumY1  +=       3.326 * cos(3.23767790075  + 73.54094334250 *T)
    sumY1  +=       2.837 * cos(0.67332834990  + 467.99887116810 *T)
    sumY1  +=       3.336 * cos(4.84176894704  + 128.60938596760 *T)
    sumY1  +=       2.896 * cos(5.30062724864  + 1802.61580820530 *T)
    sumY1  +=       2.904 * cos(1.93203325818  + 6062.41939006910 *T)
    sumY1  +=       2.774 * cos(4.62553014647  + 2847.28300942589 *T)
    sumY1  +=       2.813 * cos(4.26879116291  + 84.68970947040 *T)
    sumY1  +=       3.318 * cos(3.43420281390  + 254.69977573010 *T)
    sumY1  +=       2.850 * cos(3.81973806384  + 220.76833450690 *T)
    sumY1  +=       3.010 * cos(1.90486519028  + 451.22103074770 *T)
    sumY1  +=       3.559 * cos(3.43419228516  + 519.15220687259 *T)
    sumY1  +=       3.563 * cos(4.32049365009  + 486.15811843280 *T)
    sumY1  +=       3.360 * cos(1.00268679132  + 388.70897272171 *T)
    sumY1  +=       2.703 * cos(2.10025220587  + 345.08987514331 *T)
    sumY1  +=       2.882 * cos(1.86112467769  + 511.28790034640 *T)
    sumY1  +=       3.267 * cos(3.52339290743  + 543.26810470240 *T)
    sumY1  +=       3.197 * cos(0.64634843072  + 54.89406111079 *T)
    sumY1  +=       2.907 * cos(5.13145128922  + 630.57987624190 *T)
    sumY1  +=       2.669 * cos(0.66381570062  + 815.30716359770 *T)
    sumY1  +=       2.869 * cos(1.65073434579  + 57.86300652739 *T)
    sumY1  +=       2.764 * cos(4.24393045320  + 5635.82119919310 *T)
    sumY1  +=       2.610 * cos(3.97473877400  + 857.37235249860 *T)
    sumY1  +=       2.782 * cos(4.81585990964  + 639.17789595100 *T)
    sumY1  +=       2.826 * cos(6.22897484893  + 171.00464326860 *T)
    sumY1  +=       2.568 * cos(0.75661707085  + 3053.46855786309 *T)
    sumY1  +=       2.628 * cos(3.32798868887  + 9992.62908628869 *T)
    sumY1  +=       3.266 * cos(2.85449036040  + 721.40560204669 *T)
    sumY1  +=       2.545 * cos(2.52803626535  + 455.29619636271 *T)
    sumY1  +=       2.510 * cos(2.32307245332  + 641.10431164400 *T)
    sumY1  +=       2.554 * cos(0.62682090398  + 738.81408978600 *T)
    sumY1  +=       2.920 * cos(0.79071794025  + 92.29089143820 *T)
    sumY1  +=       2.992 * cos(2.81016019301  + 462.26673101160 *T)
    sumY1  +=       2.474 * cos(0.42468666289  + 2214.98690507971 *T)
    sumY1  +=       2.474 * cos(6.21322173241  + 419.77664546851 *T)
    sumY1  +=       2.645 * cos(1.77573634500  + 1685.29593998510 *T)
    sumY1  +=       2.597 * cos(5.06662259586  + 426.73013377490 *T)
    sumY1  +=       2.936 * cos(1.55290898601  + 468.48650613510 *T)
    sumY1  +=       2.535 * cos(1.76770621435  + 739.56484931140 *T)
    sumY1  +=       2.575 * cos(5.24718021793  + 77.71917482150 *T)
    sumY1  +=       2.960 * cos(6.14511428340  + 403.26613512341 *T)
    sumY1  +=       2.526 * cos(3.85654775000  + 198.07742442750 *T)
    sumY1  +=       2.819 * cos(4.70625558701  + 256.83194209980 *T)
    sumY1  +=       2.444 * cos(1.68660473943  + 420.69166920520 *T)
    sumY1  +=       3.165 * cos(4.31988845323  + 200.80029162821 *T)
    sumY1  +=       3.165 * cos(0.56973771336  + 226.28553421481 *T)
    sumY1  +=       2.684 * cos(3.90453406833  + 540.98048284201 *T)
    sumY1  +=       2.447 * cos(1.54641281756  + 207.18012544610 *T)
    sumY1  +=       2.447 * cos(3.34321334903  + 219.90570039690 *T)
    sumY1  +=       2.485 * cos(1.31032791644  + 2729.96314120569 *T)
    sumY1  +=       2.810 * cos(3.17127497271  + 297.98880490840 *T)
    sumY1  +=       2.343 * cos(5.74915758700  + 402.46298597130 *T)
    sumY1  +=       3.270 * cos(0.24934461993  + 714.92270236481 *T)
    sumY1  +=       2.983 * cos(4.41716427608  + 1485.73630358170 *T)
    sumY1  +=       2.301 * cos(0.60801782045  + 455.15318401080 *T)
    sumY1  +=       2.295 * cos(0.57261157738  + 503.10807963510 *T)
    sumY1  +=       2.789 * cos(1.93152906917  + 269.67762925710 *T)
    sumY1  +=       2.298 * cos(5.64615902769  + 131.16013238639 *T)
    sumY1  +=       2.626 * cos(1.15658122100  + 112.89781925770 *T)
    sumY1  +=       2.273 * cos(1.03898072401  + 283.90472325870 *T)
    sumY1  +=       2.621 * cos(0.25288963500  + 172.48911597691 *T)
    sumY1  +=       2.285 * cos(1.78001375993  + 525.51499434800 *T)
    sumY1  +=       2.204 * cos(5.58242761895  + 593.67068088151 *T)
    sumY1  +=       2.636 * cos(5.01267010764  + 649.70233135310 *T)
    sumY1  +=       2.222 * cos(4.49823618185  + 478.04773369070 *T)
    sumY1  +=       2.308 * cos(2.39654845081  + 1457.28211557850 *T)
    sumY1  +=       2.203 * cos(0.85780087675  + 1596.43025976811 *T)
    sumY1  +=       2.182 * cos(2.92076786246  + 397.14942586389 *T)
    sumY1  +=       2.502 * cos(2.31732912830  + 9.31741007210 *T)
    sumY1  +=       2.138 * cos(0.86464264578  + 2310.47849733050 *T)
    sumY1  +=       2.142 * cos(6.08094047039  + 211.57080524620 *T)
    sumY1  +=       2.974 * cos(3.80319784161  + 927.59115048390 *T)
    sumY1  +=       2.136 * cos(5.80815709862  + 186.45557754760 *T)
    sumY1  +=       2.091 * cos(1.21183260525  + 1258.69774910910 *T)
    sumY1  +=       2.094 * cos(5.65475833423  + 5.17280848790 *T)
    sumY1  +=       2.460 * cos(3.65244152992  + 5650.04829319470 *T)
    sumY1  +=       2.461 * cos(3.88615671076  + 1382.64352936309 *T)
    sumY1  +=       2.355 * cos(3.98963199001  + 184.33796539290 *T)
    sumY1  +=       2.153 * cos(3.87784179007  + 233.18663289450 *T)
    sumY1  +=       2.597 * cos(3.47646509433  + 426.95388294410 *T)
    sumY1  +=       2.604 * cos(3.63346920160  + 980.91199584231 *T)
    sumY1  +=       2.201 * cos(1.12610213271  + 434.91876320681 *T)
    sumY1  +=       2.162 * cos(2.94323512572  + 1383.13116433011 *T)
    sumY1  +=       2.404 * cos(5.08590694221  + 9580.25798941430 *T)
    sumY1  +=       2.260 * cos(2.54048343101  + 42.78208713641 *T)
    sumY1  +=       2.151 * cos(4.51443750777  + 487.60896124631 *T)
    sumY1  +=       2.123 * cos(3.56256762785  + 149.31937965110 *T)
    sumY1  +=       2.556 * cos(0.03670214203  + 142.20583265029 *T)
    sumY1  +=       1.998 * cos(0.96762571552  + 724.58699578440 *T)
    sumY1  +=       1.973 * cos(0.95972917687  + 481.01667910731 *T)
    sumY1  +=       1.973 * cos(4.56679821042  + 348.87901605450 *T)
    sumY1  +=       1.957 * cos(5.22293191768  + 28.55499313480 *T)
    sumY1  +=       1.957 * cos(5.22293191768  + 28.06735816780 *T)
    sumY1  +=       1.991 * cos(3.99171190814  + 298.47643987540 *T)
    sumY1  +=       1.940 * cos(3.46331308399  + 141.46962733991 *T)
    sumY1  +=       2.507 * cos(4.08040437480  + 6489.50521591210 *T)
    sumY1  +=       2.352 * cos(5.76915758065  + 479.04457143200 *T)
    sumY1  +=       2.048 * cos(6.17612819910  + 1788.38871420371 *T)
    sumY1  +=       2.647 * cos(0.26998970662  + 2744.19023520730 *T)
    sumY1  +=       1.908 * cos(1.80738955833  + 1368.90407032850 *T)
    sumY1  +=       1.900 * cos(4.64437128047  + 271.64973693240 *T)
    sumY1  +=       2.178 * cos(1.47792644640  + 254.59670986610 *T)
    sumY1  +=       1.894 * cos(1.95606297169  + 1048.84317196720 *T)
    sumY1  +=       1.892 * cos(6.10331223285  + 49.26498681829 *T)
    sumY1  +=       1.888 * cos(5.69235432976  + 642.58878435231 *T)
    sumY1  +=       2.182 * cos(2.85682305202  + 427.05445668060 *T)
    sumY1  +=       1.810 * cos(1.53693276386  + 518.40144734719 *T)
    sumY1  +=       1.925 * cos(1.50447120099  + 3156.56133208170 *T)
    sumY1  +=       2.022 * cos(0.93422466156  + 4.42204896250 *T)
    sumY1  +=       1.793 * cos(1.85965388376  + 419.68027724890 *T)
    sumY1  +=       2.336 * cos(5.50585235368  + 10419.71491213170 *T)
    sumY1  +=       1.992 * cos(4.49559449931  + 16.43095707290 *T)
    sumY1  +=       1.790 * cos(3.03209854172  + 71.84402231310 *T)
    sumY1  +=       1.779 * cos(0.73335994085  + 702.39252939260 *T)
    sumY1  +=       1.796 * cos(2.22306233592  + 240.63024829540 *T)
    sumY1  +=       1.974 * cos(5.12102350165  + 640.18928790730 *T)
    sumY1  +=       1.881 * cos(4.15581943535  + 200.52510498229 *T)
    sumY1  +=       1.773 * cos(5.62762637156  + 990.47322339791 *T)
    sumY1  +=       2.183 * cos(1.90262206941  + 411.37651986550 *T)
    sumY1  +=       2.250 * cos(5.51794372414  + 6077.13411903770 *T)
    sumY1  +=       2.182 * cos(5.68626463817  + 426.62956003840 *T)
    sumY1  +=       1.835 * cos(0.37752300303  + 163.14033674240 *T)
    sumY1  +=       2.428 * cos(6.11907466746  + 2641.09746098870 *T)
    sumY1  +=       1.996 * cos(0.31529112752  + 398.38782035631 *T)
    sumY1  +=       1.813 * cos(5.75219077219  + 25.51661174900 *T)
    sumY1  +=       1.980 * cos(4.51396179003  + 206.54124050531 *T)
    sumY1  +=       1.940 * cos(4.75992273689  + 774.72644740510 *T)
    sumY1  +=       1.910 * cos(3.04478377946  + 175.67050971460 *T)
    sumY1  +=       2.379 * cos(2.00455891511  + 594.89452115890 *T)
    sumY1  +=       2.068 * cos(5.66121874325  + 183.85033042589 *T)
    sumY1  +=       1.733 * cos(1.82630536023  + 705.36147480921 *T)
    sumY1  +=       1.824 * cos(5.14397632587  + 420.24972622050 *T)
    sumY1  +=       1.804 * cos(5.66197611508  + 9566.03089541269 *T)
    sumY1  +=       2.037 * cos(2.38364538019  + 481.97988695381 *T)
    sumY1  +=       1.792 * cos(2.83739260393  + 842.65762352999 *T)
    sumY1  +=       1.681 * cos(0.96792366320  + 1575.08961876571 *T)
    sumY1  +=       1.732 * cos(3.12442860953  + 692.34366686999 *T)
    sumY1  +=       1.698 * cos(0.19286826064  + 355.50492808829 *T)
    sumY1  +=       1.828 * cos(3.41879771076  + 484.94883237040 *T)
    sumY1  +=       1.713 * cos(6.01421151060  + 457.86149699650 *T)
    sumY1  +=       1.808 * cos(4.32289101961  + 622.24498109120 *T)
    sumY1  +=       1.911 * cos(2.04742119266  + 5165.60650620771 *T)
    sumY1  +=       1.704 * cos(1.52079559528  + 2538.00468677009 *T)
    sumY1  +=       1.688 * cos(1.32671223888  + 267.22994870229 *T)
    sumY1  +=       1.665 * cos(3.80619103257  + 1127.50624756031 *T)
    sumY1  +=       2.097 * cos(0.65708777051  + 10007.34381525730 *T)
    sumY1  +=       1.691 * cos(0.50775020289  + 789.70430093210 *T)
    sumY1  +=       1.699 * cos(5.49155326342  + 760.49935340350 *T)
    sumY1  +=       1.783 * cos(1.56495333708  + 2001.68780964171 *T)
    sumY1  +=       1.796 * cos(2.70686770403  + 1194.21109128031 *T)
    sumY1  +=       1.619 * cos(4.58520070805  + 189.47940471839 *T)
    sumY1  +=       1.652 * cos(1.38068483464  + 152.77596003471 *T)
    sumY1  +=       1.707 * cos(1.05909040328  + 3046.35501086230 *T)
    sumY1  +=       1.924 * cos(5.14976993437  + 380.37158544351 *T)
    sumY1  +=       1.960 * cos(1.77726603611  + 144.17794032560 *T)
    sumY1  +=       1.599 * cos(4.42481035386  + 2008.80135664250 *T)
    sumY1  +=       1.804 * cos(4.83120830404  + 353.54488250050 *T)
    sumY1  +=       1.586 * cos(1.77477861670  + 6069.53293706989 *T)
    sumY1  +=       1.644 * cos(5.88078903892  + 1112.31594571220 *T)
    sumY1  +=       1.572 * cos(6.09540618494  + 91.54262404029 *T)
    sumY1  +=       2.123 * cos(4.19827720301  + 1802.12817323829 *T)
    sumY1  +=       1.948 * cos(3.03917076022  + 268.68079151581 *T)
    sumY1  +=       1.766 * cos(3.33487535090  + 3333.74269718250 *T)
    sumY1  +=       1.565 * cos(5.51220884907  + 97.91996573070 *T)
    sumY1  +=       1.602 * cos(0.01982170402  + 490.09027169589 *T)
    sumY1  +=       1.546 * cos(0.85174846193  + 258.78949556011 *T)
    sumY1  +=       1.724 * cos(1.53656554024  + 26081.03341205689 *T)
    sumY1  +=       1.531 * cos(5.96236809333  + 5216.33655531789 *T)
    sumY1  +=       1.619 * cos(0.30775874593  + 784.28767496071 *T)
    sumY1  +=       1.871 * cos(5.11950364986  + 206.31749133610 *T)
    sumY1  +=       1.734 * cos(1.77549936222  + 304.84171947829 *T)
    sumY1  +=       1.732 * cos(4.43552932727  + 1332.72858815100 *T)
    sumY1  +=       1.517 * cos(0.53979176380  + 1048.09241244180 *T)
    sumY1  +=       1.563 * cos(1.89447201338  + 555.79827767460 *T)
    sumY1  +=       1.520 * cos(1.78410202569  + 1101.27024544830 *T)
    sumY1  +=       1.687 * cos(2.79477418739  + 3347.96979118411 *T)
    sumY1  +=       1.488 * cos(3.24222987427  + 9999.74263328950 *T)
    sumY1  +=       1.587 * cos(1.78380089017  + 372.77040347569 *T)
    sumY1  +=       1.662 * cos(0.99874837499  + 26095.26050605850 *T)
    sumY1  +=       1.724 * cos(2.90245872652  + 1037.79747170330 *T)
    sumY1  +=       1.477 * cos(5.93132218114  + 208.38941150850 *T)
    sumY1  +=       1.657 * cos(0.26785782921  + 552.82933225800 *T)
    sumY1  +=       1.520 * cos(1.16831624482  + 2318.07967929830 *T)
    sumY1  +=       1.443 * cos(0.41860834829  + 910.06255053810 *T)
    sumY1  +=       1.446 * cos(5.81588894214  + 204.94489321240 *T)
    sumY1  +=       1.446 * cos(5.35692253163  + 222.14093263061 *T)
    sumY1  +=       1.850 * cos(2.22163234770  + 699.94484883781 *T)
    sumY1  +=       1.439 * cos(3.09801371747  + 393.70490756781 *T)
    sumY1  +=       1.452 * cos(4.70036751558  + 1130.47519297691 *T)
    sumY1  +=       1.874 * cos(0.77033226307  + 770.06058095910 *T)
    sumY1  +=       1.782 * cos(3.11527131409  + 835.54407652920 *T)
    sumY1  +=       1.639 * cos(3.50228669347  + 427.10264079041 *T)
    sumY1  +=       1.565 * cos(5.31653543667  + 392.90175841570 *T)
    sumY1  +=       1.395 * cos(1.10262291095  + 9146.54625153749 *T)
    sumY1  +=       1.516 * cos(5.15088806631  + 397.90018538930 *T)
    sumY1  +=       1.684 * cos(4.67559134043  + 1479.11039154791 *T)
    sumY1  +=       1.659 * cos(0.02097122599  + 721.89323701370 *T)
    sumY1  +=       1.514 * cos(1.20406313416  + 319.06881347989 *T)
    sumY1  +=       1.397 * cos(4.55303867764  + 3163.67487908250 *T)
    sumY1  +=       1.747 * cos(3.84736035069  + 37.88921815429 *T)
    sumY1  +=       1.492 * cos(1.47452203681  + 263.94548910060 *T)
    sumY1  +=       1.354 * cos(3.00416254799  + 140.24578706250 *T)
    sumY1  +=       1.506 * cos(4.98643529599  + 251.41531612840 *T)
    sumY1  +=       1.365 * cos(3.12436033337  + 694.31577454531 *T)
    sumY1  +=       1.352 * cos(2.91854326993  + 60.79832204921 *T)
    sumY1  +=       1.352 * cos(6.06013592352  + 60.31068708220 *T)
    sumY1  +=       1.531 * cos(5.36116451799  + 907.09360512150 *T)
    sumY1  +=       1.783 * cos(5.53040699939  + 185.08872491831 *T)
    sumY1  +=       1.550 * cos(5.62135435218  + 170.51700830159 *T)
    sumY1  +=       1.524 * cos(4.86473762888  + 921.32069912310 *T)
    sumY1  +=       1.541 * cos(1.16911147253  + 31.73638687251 *T)
    sumY1  +=       1.578 * cos(4.41469563792  + 190.64836324160 *T)
    sumY1  +=       1.578 * cos(3.61652318226  + 236.43746260140 *T)
    sumY1  +=       1.732 * cos(5.35974909247  + 1126.54303971381 *T)
    sumY1  +=       1.540 * cos(1.93459248812  + 2833.05591542430 *T)
    sumY1  +=       1.323 * cos(3.30443135672  + 460.78225830330 *T)
    sumY1  +=       1.312 * cos(0.39355615225  + 448.98829064149 *T)
    sumY1  +=       1.724 * cos(0.67798398516  + 78.46993434690 *T)
    sumY1  +=       1.472 * cos(2.69259943020  + 34.18406742730 *T)
    sumY1  +=       1.296 * cos(4.05050226429  + 2325.19322629910 *T)
    sumY1  +=       1.309 * cos(3.88442947901  + 349.09146437560 *T)
    sumY1  +=       1.818 * cos(5.73491115213  + 100.40127618030 *T)
    sumY1  +=       1.290 * cos(0.79333495865  + 1176.94561589290 *T)
    sumY1  +=       1.334 * cos(0.87141749832  + 282.90788551740 *T)
    sumY1  +=       1.287 * cos(4.97743606169  + 286.84003878050 *T)
    sumY1  +=       1.392 * cos(0.16591257376  + 69.12115511239 *T)
    sumY1  +=       1.515 * cos(5.19152100527  + 12.28635548869 *T)
    sumY1  +=       1.247 * cos(3.07584535023  + 187.67941782501 *T)
    sumY1  +=       1.247 * cos(4.95537346995  + 239.40640801801 *T)
    sumY1  +=       1.246 * cos(3.36618810796  + 2487.17222746430 *T)
    sumY1  +=       1.511 * cos(3.23694999728  + 213.00709384469 *T)
    sumY1  +=       1.479 * cos(2.25882851500  + 2516.66404576769 *T)
    sumY1  +=       1.233 * cos(0.08035255767  + 2531.37877473630 *T)
    sumY1  +=       1.632 * cos(1.33635309255  + 906.13039727500 *T)
    sumY1  +=       1.387 * cos(5.26434372100  + 401.56921409400 *T)
    sumY1  +=       1.352 * cos(2.20381387735  + 351.57277482520 *T)
    sumY1  +=       1.313 * cos(1.70932562812  + 350.08830211689 *T)
    sumY1  +=       1.236 * cos(4.96962008251  + 11926.01059618529 *T)
    sumY1  +=       1.227 * cos(3.67295844750  + 4127.56521702090 *T)
    sumY1  +=       1.304 * cos(1.86059485849  + 203.49405039889 *T)
    sumY1  +=       1.478 * cos(0.09633488682  + 264.26100328570 *T)
    sumY1  +=       1.254 * cos(1.48175809016  + 827.67977000299 *T)
    sumY1  +=       1.230 * cos(4.01695958431  + 51535.66517935089 *T)
    sumY1  +=       1.385 * cos(2.69048880279  + 41.29761442810 *T)
    sumY1  +=       1.364 * cos(2.61505122332  + 65.46418849521 *T)
    sumY1  +=       1.395 * cos(4.21766601261  + 629.35852809199 *T)
    sumY1  +=       1.284 * cos(3.80036081443  + 168.29633028290 *T)
    sumY1  +=       1.435 * cos(1.13143123004  + 617.56206830270 *T)
    sumY1  +=       1.411 * cos(4.31607548596  + 188.27011865600 *T)
    sumY1  +=       1.191 * cos(2.27480359368  + 606.51636803880 *T)
    sumY1  +=       1.193 * cos(0.20887653741  + 274.30986580831 *T)
    sumY1  +=       1.197 * cos(0.48263618413  + 1062.80714141041 *T)
    sumY1  +=       1.356 * cos(5.93555277060  + 50.64639366260 *T)
    sumY1  +=       1.356 * cos(5.93555277060  + 50.15875869559 *T)
    sumY1  +=       1.276 * cos(1.88008049503  + 2090.06585489169 *T)
    sumY1  +=       1.195 * cos(4.43032838826  + 681.29796660610 *T)
    sumY1  +=       1.603 * cos(3.21307189614  + 1560.37488979710 *T)
    sumY1  +=       1.191 * cos(5.25663119995  + 9594.99527173989 *T)
    sumY1  +=       1.168 * cos(2.65329470104  + 74.39476873190 *T)
    sumY1  +=       1.165 * cos(6.24881770348  + 674.18441960530 *T)
    sumY1  +=       1.242 * cos(4.06152745485  + 1902.52718868620 *T)
    sumY1  +=       1.330 * cos(4.76147886793  + 200.27902676641 *T)
    sumY1  +=       1.330 * cos(0.12814729866  + 226.80679907661 *T)
    sumY1  +=       1.154 * cos(6.03705716215  + 2847.77064439291 *T)
    sumY1  +=       1.150 * cos(4.06754550587  + 17.65230522280 *T)
    sumY1  +=       1.152 * cos(0.35591000310  + 740.05248427840 *T)
    sumY1  +=       1.247 * cos(0.42544593650  + 348.39138108750 *T)
    sumY1  +=       1.387 * cos(4.16806169190  + 527.48710202330 *T)
    sumY1  +=       1.250 * cos(3.55179969067  + 184.60108995130 *T)
    sumY1  +=       1.328 * cos(6.06882136981  + 2957.48933064529 *T)
    sumY1  +=       1.587 * cos(3.53266665825  + 1243.98302014049 *T)
    sumY1  +=       1.176 * cos(0.23513424404  + 747.16603127920 *T)
    sumY1  +=       1.223 * cos(2.16572700191  + 1773.67398523510 *T)
    sumY1  +=       1.161 * cos(4.72637850213  + 2435.39954751850 *T)
    sumY1  +=       1.570 * cos(6.26965505442  + 3060.58210486390 *T)
    sumY1  +=       1.303 * cos(5.05352053957  + 393.21727260080 *T)
    sumY1  +=       1.126 * cos(2.32401124803  + 305.10235190919 *T)
    sumY1  +=       1.534 * cos(0.30457466054  + 81.99595069970 *T)
    sumY1  +=       1.117 * cos(3.04466574791  + 526.00262931501 *T)
    sumY1  +=       1.307 * cos(1.13548842988  + 72.09010052900 *T)
    sumY1  +=       1.114 * cos(1.57760085679  + 25234.46294233869 *T)
    sumY1  +=       1.118 * cos(6.05531697707  + 511.77553531340 *T)
    sumY1  +=       1.242 * cos(4.54338222029  + 145.38722638800 *T)
    sumY1  +=       1.397 * cos(5.87612999192  + 336.59492129820 *T)
    sumY1  +=       1.401 * cos(3.85597938492  + 2118.52004289490 *T)
    sumY1  +=       1.139 * cos(0.30433035611  + 1055.20595944260 *T)
    sumY1  +=       1.096 * cos(2.35952205044  + 739.30172475301 *T)
    sumY1  +=       1.092 * cos(3.29113609259  + 234.39591895690 *T)
    sumY1  +=       1.197 * cos(5.42029486787  + 306.34074640161 *T)
    sumY1  +=       1.126 * cos(1.19478791991  + 1144.07163965960 *T)
    sumY1  +=       1.351 * cos(1.46891070255  + 179.60266297770 *T)
    sumY1  +=       1.165 * cos(1.70894571639  + 746.67839631219 *T)
    sumY1  +=       1.167 * cos(2.75083276977  + 634.72447782610 *T)
    sumY1  +=       1.121 * cos(2.13313396073  + 366.28750379380 *T)
    sumY1  +=       1.051 * cos(4.34157702192  + 383.81610373959 *T)
    sumY1  +=       1.053 * cos(1.84848967894  + 2107.96197759801 *T)
    sumY1  +=       1.051 * cos(1.42831004195  + 10000.23026825651 *T)
    sumY1  +=       1.288 * cos(0.28913908894  + 247.48316286530 *T)
    sumY1  +=       1.045 * cos(0.06510625472  + 1680.87615175499 *T)
    sumY1  +=       1.070 * cos(1.11790351462  + 3472.95320173830 *T)
    sumY1  +=       1.043 * cos(3.53177329067  + 151.50393564750 *T)
    sumY1  +=       1.033 * cos(0.02365635747  + 6070.02057203690 *T)
    sumY1  +=       1.282 * cos(2.02033530101  + 141.94270809190 *T)
    sumY1  +=       1.167 * cos(1.05359353693  + 2920.88396534110 *T)
    sumY1  +=       1.100 * cos(0.95076161618  + 2634.47154895491 *T)
    sumY1  +=       1.035 * cos(0.59532463011  + 3369.86042751970 *T)
    sumY1  +=       1.091 * cos(5.26065063725  + 1314.56934088630 *T)
    sumY1  +=       1.274 * cos(1.81279771107  + 691.34682912870 *T)
    sumY1  +=       1.079 * cos(2.63111319426  + 1044.64789414570 *T)
    sumY1  +=       1.085 * cos(0.65660514905  + 141.20899490900 *T)
    sumY1  +=       1.245 * cos(1.12530476318  + 179.91817716280 *T)
    sumY1  +=       1.146 * cos(5.05013070926  + 220.54458533771 *T)
    sumY1  +=       1.196 * cos(1.60077277494  + 1460.20038475780 *T)
    sumY1  +=       1.136 * cos(3.62545973301  + 480.00777927849 *T)
    sumY1  +=       1.119 * cos(3.84274502894  + 1534.98198332510 *T)
    sumY1  +=       1.056 * cos(5.19951356169  + 437.40007365639 *T)
    sumY1  +=       1.032 * cos(0.75596269901  + 184.97110483931 *T)
    sumY1  +=       1.032 * cos(0.99207081399  + 242.11472100371 *T)
    sumY1  +=       1.039 * cos(3.89912009234  + 1791.08247297440 *T)
    sumY1  +=       1.061 * cos(4.07694835092  + 938.63685074779 *T)
    sumY1  +=       1.123 * cos(5.53483735553  + 25668.17468021549 *T)
    sumY1  +=       1.103 * cos(1.14969049707  + 210.60759739969 *T)
    sumY1  +=       1.015 * cos(5.62555119479  + 252.41215386970 *T)
    sumY1  +=       0.979 * cos(0.51006487073  + 610.44852130189 *T)
    sumY1  +=       0.979 * cos(0.51006487073  + 610.93615626891 *T)
    sumY1  +=       1.043 * cos(5.02488186971  + 17.91542978119 *T)
    sumY1  +=       1.128 * cos(2.35271436800  + 398.53083270820 *T)
    sumY1  +=       1.035 * cos(4.59247999326  + 235.68919520349 *T)
    sumY1  +=       1.072 * cos(2.96070791769  + 746.15463932291 *T)
    sumY1  +=       1.131 * cos(1.80265887785  + 5481.99837586430 *T)
    sumY1  +=       0.986 * cos(0.57363342277  + 192.44835013500 *T)
    sumY1  +=       1.287 * cos(0.38193481035  + 314.18800658530 *T)
    sumY1  +=       0.973 * cos(5.42514671691  + 561.69047652550 *T)
    sumY1  +=       1.298 * cos(2.65454318605  + 614.38067456499 *T)
    sumY1  +=       1.151 * cos(2.77783863264  + 905.64276230799 *T)
    sumY1  +=       1.279 * cos(1.17281533398  + 322.36782729660 *T)
    sumY1  +=       0.963 * cos(1.93346454966  + 782.59075393131 *T)
    sumY1  +=       1.064 * cos(1.04722734848  + 1372.34858862460 *T)
    sumY1  +=       0.973 * cos(1.95209864065  + 24.62283987170 *T)
    sumY1  +=       1.096 * cos(3.49856994945  + 731.92754532131 *T)
    sumY1  +=       1.080 * cos(1.56741970160  + 2906.65687133950 *T)
    sumY1  +=       1.200 * cos(3.88182273784  + 195.64679021520 *T)
    sumY1  +=       0.993 * cos(0.38449878066  + 6062.90702503610 *T)
    sumY1  +=       1.039 * cos(2.64505699779  + 1159.04949318660 *T)
    sumY1  +=       0.952 * cos(3.87253065022  + 239.88198089750 *T)
    sumY1  +=       1.181 * cos(2.55455580038  + 1468.32781584240 *T)
    sumY1  +=       1.137 * cos(1.16274372128  + 2.69149803831 *T)
    sumY1  +=       0.941 * cos(2.80652276090  + 124.18959773749 *T)
    sumY1  +=       0.942 * cos(1.93049464908  + 115.11600514891 *T)
    sumY1  +=       1.020 * cos(3.36648319866  + 196.38050339810 *T)
    sumY1  +=       1.065 * cos(3.10915621431  + 494.23487328010 *T)
    sumY1  +=       0.945 * cos(5.96236583490  + 121.00820399980 *T)
    sumY1  +=       0.941 * cos(2.10164527812  + 120.01136625851 *T)
    sumY1  +=       0.996 * cos(1.49787021480  + 4863.44173055270 *T)
    sumY1  +=       1.059 * cos(6.04847231304  + 25653.94758621389 *T)
    sumY1  +=       0.936 * cos(5.92819887567  + 307.07445958451 *T)
    sumY1  +=       0.986 * cos(3.52637633155  + 5842.00674763030 *T)
    sumY1  +=       1.061 * cos(2.35701423791  + 418.97770183331 *T)
    sumY1  +=       0.941 * cos(0.81063561753  + 97.17169833279 *T)
    sumY1  +=       1.049 * cos(5.16810544385  + 416.79314583689 *T)
    sumY1  +=       0.956 * cos(3.96361938951  + 141.45507312490 *T)
    sumY1  +=       0.922 * cos(0.26133958050  + 1364.48428209840 *T)
    sumY1  +=       1.090 * cos(2.75394419028  + 1151.93594618579 *T)
    sumY1  +=       0.945 * cos(3.91497884985  + 220.86890824340 *T)
    sumY1  +=       0.945 * cos(0.97464731674  + 206.21691759960 *T)
    sumY1  +=       0.906 * cos(0.44130467758  + 12779.69461290430 *T)
    sumY1  +=       0.913 * cos(4.49732744600  + 285.61619850311 *T)
    sumY1  +=       0.993 * cos(1.57797933233  + 73.05330837550 *T)
    sumY1  +=       0.923 * cos(2.38369657059  + 1048.58004740881 *T)
    sumY1  +=       0.910 * cos(1.34865027178  + 1250.83344258291 *T)
    sumY1  +=       1.055 * cos(4.53650215705  + 287.69386416990 *T)
    sumY1  +=       0.891 * cos(3.38916042044  + 4443.95708667750 *T)
    sumY1  +=       0.890 * cos(1.87781413537  + 952.34018776010 *T)
    sumY1  +=       1.011 * cos(2.41002467914  + 138.90432670610 *T)
    sumY1  +=       0.943 * cos(4.45248854943  + 1368.41643536150 *T)
    sumY1  +=       1.240 * cos(1.39562373377  + 105.78427225690 *T)
    sumY1  +=       1.240 * cos(4.53721638736  + 105.29663728990 *T)
    sumY1  +=       0.975 * cos(6.14076145568  + 700.90805668431 *T)
    sumY1  +=       0.963 * cos(1.26904248346  + 5378.90560164571 *T)
    sumY1  +=       1.131 * cos(1.92127706230  + 640.09291968770 *T)
    sumY1  +=       1.014 * cos(2.27419938832  + 246.68001371321 *T)
    sumY1  +=       0.922 * cos(4.94986707425  + 9772.21644384989 *T)
    sumY1  +=       0.896 * cos(2.95327436580  + 1571.15746550260 *T)
    sumY1  +=       1.121 * cos(0.63121659144  + 6467.68194047809 *T)
    sumY1  +=       0.872 * cos(1.30941661635  + 519.63984183961 *T)
    sumY1  +=       1.164 * cos(4.79404702362  + 739.04109232210 *T)
    sumY1  +=       1.090 * cos(3.70049172953  + 1699.03539901970 *T)
    sumY1  +=       1.180 * cos(4.01443566672  + 636.47207509279 *T)
    sumY1  +=       0.912 * cos(2.54546205453  + 31.26330612050 *T)
    sumY1  +=       0.892 * cos(2.36351221746  + 212.57970507500 *T)
    sumY1  +=       0.892 * cos(5.50510487105  + 212.09207010800 *T)
    sumY1  +=       0.985 * cos(3.52323843583  + 988.02554284310 *T)
    sumY1  +=       0.860 * cos(2.94846758267  + 1585.38455950421 *T)
    sumY1  +=       0.990 * cos(1.67825090661  + 3149.44778508089 *T)
    sumY1  +=       1.177 * cos(4.87572778682  + 1328.79643488791 *T)
    sumY1  +=       0.864 * cos(2.80999692865  + 12146.42323862409 *T)
    sumY1  +=       1.007 * cos(3.83948575721  + 685.23011986920 *T)
    sumY1  +=       1.007 * cos(3.83948575721  + 685.71775483620 *T)
    sumY1  +=       0.853 * cos(3.93116170567  + 2641.58509595570 *T)
    sumY1  +=       0.993 * cos(1.01758536916  + 2111.89413086110 *T)
    sumY1  +=       0.844 * cos(3.42851532207  + 1239.78774231901 *T)
    sumY1  +=       0.846 * cos(1.60469270371  + 2115.07552459880 *T)
    sumY1  +=       0.964 * cos(1.18238037517  + 756.56720014041 *T)
    sumY1  +=       0.981 * cos(6.17353804084  + 1471.24608502170 *T)
    sumY1  +=       1.164 * cos(4.46546430995  + 3391.68870348911 *T)
    sumY1  +=       0.885 * cos(1.64388709070  + 609.12161516050 *T)
    sumY1  +=       0.821 * cos(2.00294213618  + 388.22133775470 *T)
    sumY1  +=       0.906 * cos(3.95724484289  + 327.19375243699 *T)
    sumY1  +=       0.903 * cos(1.46965018009  + 477.24458453860 *T)
    sumY1  +=       0.841 * cos(2.31779507594  + 473.31243127551 *T)
    sumY1  +=       0.904 * cos(5.47014409814  + 635.47523735150 *T)
    sumY1  +=       0.822 * cos(1.66432337141  + 1365.23504162379 *T)
    sumY1  +=       0.997 * cos(1.82079729948  + 123.29582586020 *T)
    sumY1  +=       0.879 * cos(0.83289385422  + 3362.74688051889 *T)
    sumY1  +=       0.846 * cos(0.41970975707  + 227.81819103291 *T)
    sumY1  +=       0.846 * cos(4.46991640952  + 199.26763481011 *T)
    sumY1  +=       0.947 * cos(0.09177854194  + 222.61650551010 *T)
    sumY1  +=       0.875 * cos(1.91606789801  + 9993.11672125570 *T)
    sumY1  +=       1.117 * cos(5.40582369615  + 319.08336769490 *T)
    sumY1  +=       0.800 * cos(1.12338267507  + 834.79331700380 *T)
    sumY1  +=       0.839 * cos(4.44054896906  + 206.64181424181 *T)
    sumY1  +=       0.839 * cos(0.44907719753  + 220.44401160121 *T)
    sumY1  +=       0.838 * cos(1.69845530894  + 2442.51309451930 *T)
    sumY1  +=       0.830 * cos(0.99305838433  + 1891.48148842230 *T)
    sumY1  +=       0.841 * cos(6.22199463081  + 60.04756252381 *T)
    sumY1  +=       0.826 * cos(2.10402662108  + 220.13519506050 *T)
    sumY1  +=       0.854 * cos(5.92854026664  + 540.22972331660 *T)
    sumY1  +=       0.842 * cos(1.12102777557  + 2854.39655642670 *T)
    sumY1  +=       0.858 * cos(4.12461672205  + 781.83999440590 *T)
    sumY1  +=       0.860 * cos(4.20709137726  + 5436.74919775670 *T)
    sumY1  +=       0.785 * cos(4.19221284065  + 3494.29384274070 *T)

    # SATURN Y Coord
    sumY2  = 0
    sumY2  +=  862721.930 * cos(0.87622130189  + 213.54291292150 *T)
    sumY2  +=  611630.287 * cos(0.71399818143  + 0.24381748350 *T)
    sumY2  +=  561413.805 * cos(5.96728319816  + 206.42936592071 *T)
    sumY2  +=  547271.370 * cos(2.04588665465  + 220.65645992230 *T)
    sumY2  +=  234984.093 * cos(0.87581308687  + 426.84200835950 *T)
    sumY2  +=   88108.317 * cos(3.29758949735  + 7.35736448430 *T)
    sumY2  +=   50142.527 * cos(1.12650721735  + 433.95555536030 *T)
    sumY2  +=   45054.746 * cos(5.06577813923  + 419.72846135871 *T)
    sumY2  +=   33335.009 * cos(0.01047584776  + 640.14110379750 *T)
    sumY2  +=   27185.514 * cos(1.92525242054  + 227.77000692311 *T)
    sumY2  +=   18651.786 * cos(1.59992516931  + 213.05527795450 *T)
    sumY2  +=   12946.617 * cos(6.11930195692  + 199.31581891990 *T)
    sumY2  +=   10383.025 * cos(3.18830710215  + 14.47091148511 *T)
    sumY2  +=    6684.574 * cos(4.09620079950  + 110.45013870291 *T)
    sumY2  +=    4661.903 * cos(0.26368341105  + 647.25465079831 *T)
    sumY2  +=    4933.834 * cos(1.51097568839  + 216.72430665921 *T)
    sumY2  +=    4778.577 * cos(0.24213059962  + 210.36151918381 *T)
    sumY2  +=    4377.928 * cos(3.66946155930  + 316.63568714010 *T)
    sumY2  +=    3532.396 * cos(5.48934411728  + 853.44019923550 *T)
    sumY2  +=    2688.895 * cos(1.58317658973  + 209.61075965840 *T)
    sumY2  +=    2540.548 * cos(1.07874150891  + 441.06910236111 *T)
    sumY2  +=    2540.549 * cos(0.15899514683  + 217.47506618460 *T)
    sumY2  +=    2506.252 * cos(3.11588688157  + 117.56368570370 *T)
    sumY2  +=    2582.753 * cos(0.02782240807  + 103.33659170210 *T)
    sumY2  +=    2111.683 * cos(4.98775349171  + 633.02755679670 *T)
    sumY2  +=    2306.046 * cos(5.24765858946  + 205.94173095370 *T)
    sumY2  +=    1843.706 * cos(2.39971312535  + 323.74923414091 *T)
    sumY2  +=    1272.298 * cos(4.96900516981  + 412.61491435790 *T)
    sumY2  +=     940.336 * cos(2.42501300855  + 6.86972951729 *T)
    sumY2  +=     804.980 * cos(1.78590189490  + 234.88355392390 *T)
    sumY2  +=     698.973 * cos(0.77208908956  + 95.73540973430 *T)
    sumY2  +=     653.010 * cos(5.01381847221  + 198.82818395289 *T)
    sumY2  +=     766.827 * cos(5.79919197548  + 202.49721265760 *T)
    sumY2  +=     662.176 * cos(0.42004956009  + 426.35437339250 *T)
    sumY2  +=     565.841 * cos(1.55419890490  + 220.16882495529 *T)
    sumY2  +=     765.788 * cos(2.28516505694  + 224.58861318540 *T)
    sumY2  +=     672.412 * cos(5.85676090936  + 309.52214013930 *T)
    sumY2  +=     490.992 * cos(0.10668768793  + 109.96250373590 *T)
    sumY2  +=     440.264 * cos(2.02435630765  + 330.86278114170 *T)
    sumY2  +=     426.768 * cos(5.42227433044  + 11.28951774740 *T)
    sumY2  +=     427.613 * cos(5.71584342304  + 860.55374623631 *T)
    sumY2  +=     505.832 * cos(5.83465733982  + 309.03450517230 *T)
    sumY2  +=     559.499 * cos(5.12833446634  + 102.84895673509 *T)
    sumY2  +=     439.569 * cos(3.04587045996  + 21.58445848590 *T)
    sumY2  +=     404.335 * cos(1.38184466479  + 223.83785366000 *T)
    sumY2  +=     400.108 * cos(0.64688754335  + 430.02340209721 *T)
    sumY2  +=     338.086 * cos(4.72348149737  + 1066.73929467350 *T)
    sumY2  +=     310.691 * cos(2.86856110459  + 3.42521122121 *T)
    sumY2  +=     356.340 * cos(0.50093637409  + 422.90985509640 *T)
    sumY2  +=     268.315 * cos(0.90238113501  + 88.62186273350 *T)
    sumY2  +=     265.067 * cos(5.80890278675  + 423.66061462181 *T)
    sumY2  +=     249.792 * cos(2.21289242997  + 4.17597074660 *T)
    sumY2  +=     244.377 * cos(0.24702615370  + 654.36819779910 *T)
    sumY2  +=     305.650 * cos(4.74789522190  + 419.24082639170 *T)
    sumY2  +=     263.490 * cos(1.75719781150  + 124.67723270450 *T)
    sumY2  +=     243.362 * cos(0.01260786807  + 96.22304470130 *T)
    sumY2  +=     228.016 * cos(4.72661469203  + 522.33360061030 *T)
    sumY2  +=     208.205 * cos(0.02609578112  + 192.20227191910 *T)
    sumY2  +=     262.941 * cos(2.24919351561  + 522.82123557730 *T)
    sumY2  +=     217.559 * cos(3.69816995653  + 412.12727939089 *T)
    sumY2  +=     204.218 * cos(0.74740242459  + 149.80701461810 *T)
    sumY2  +=     207.097 * cos(0.47214424781  + 625.91400979590 *T)
    sumY2  +=     226.266 * cos(0.43162189178  + 203.24797218300 *T)
    sumY2  +=     164.818 * cos(1.21105459258  + 537.04832957890 *T)
    sumY2  +=     154.166 * cos(0.52313104707  + 302.40859313851 *T)
    sumY2  +=     159.821 * cos(3.57691788014  + 529.93478257810 *T)
    sumY2  +=     127.987 * cos(4.67133129306  + 515.22005360950 *T)
    sumY2  +=     150.521 * cos(5.90883755804  + 301.92095817150 *T)
    sumY2  +=     109.809 * cos(0.87108590696  + 277.27881122490 *T)
    sumY2  +=     107.126 * cos(2.60669435864  + 10.53875822200 *T)
    sumY2  +=     125.246 * cos(1.70701686068  + 437.88770862340 *T)
    sumY2  +=     119.913 * cos(5.60277656541  + 625.42637482889 *T)
    sumY2  +=      95.870 * cos(2.12616183050  + 13.98327651810 *T)
    sumY2  +=      93.192 * cos(0.34445551196  + 430.77416162260 *T)
    sumY2  +=      82.653 * cos(0.98307709174  + 448.18264936190 *T)
    sumY2  +=      80.395 * cos(3.02174862119  + 231.70216018621 *T)
    sumY2  +=      86.059 * cos(5.78308214520  + 742.74624304909 *T)
    sumY2  +=      78.439 * cos(2.37542883243  + 415.79630809560 *T)
    sumY2  +=      95.719 * cos(5.10971036152  + 846.32665223470 *T)
    sumY2  +=      69.822 * cos(4.50382031149  + 838.72547026689 *T)
    sumY2  +=      88.159 * cos(3.12349955672  + 743.23387801611 *T)
    sumY2  +=      62.703 * cos(3.46541693301  + 728.51914904750 *T)
    sumY2  +=      61.768 * cos(0.31223820367  + 3.68833577960 *T)
    sumY2  +=      59.007 * cos(2.35866131486  + 316.14805217309 *T)
    sumY2  +=      57.483 * cos(3.58195443098  + 18.40306474821 *T)
    sumY2  +=      68.064 * cos(1.33121226452  + 736.12033101530 *T)
    sumY2  +=      73.156 * cos(4.57295172048  + 735.63269604830 *T)
    sumY2  +=      66.143 * cos(0.61450499024  + 529.44714761109 *T)
    sumY2  +=      46.643 * cos(4.39908344730  + 831.61192326610 *T)
    sumY2  +=      46.381 * cos(1.16571256843  + 544.16187657970 *T)
    sumY2  +=      49.127 * cos(2.41446662828  + 515.70768857651 *T)
    sumY2  +=      45.075 * cos(2.04400929715  + 195.38366565681 *T)
    sumY2  +=      45.874 * cos(4.71561784201  + 408.68276109481 *T)
    sumY2  +=      39.597 * cos(4.90149000270  + 1073.85284167431 *T)
    sumY2  +=      37.898 * cos(3.62258524288  + 1059.62574767270 *T)
    sumY2  +=      44.466 * cos(5.29067071105  + 941.81824448550 *T)
    sumY2  +=      34.788 * cos(6.24281796552  + 363.10611005610 *T)
    sumY2  +=      32.598 * cos(0.67708460723  + 433.46792039329 *T)
    sumY2  +=      33.292 * cos(5.12527889351  + 191.71463695210 *T)
    sumY2  +=      41.206 * cos(5.63757151016  + 618.31282782810 *T)
    sumY2  +=      32.714 * cos(5.53677499538  + 1059.13811270570 *T)
    sumY2  +=      41.218 * cos(5.90724265736  + 636.20895053440 *T)
    sumY2  +=      36.847 * cos(4.86141709685  + 223.10414047710 *T)
    sumY2  +=      28.979 * cos(2.88486726124  + 113.63153244060 *T)
    sumY2  +=      30.445 * cos(1.28493169701  + 56.37853381910 *T)
    sumY2  +=      38.514 * cos(0.59603887895  + 86.07111631471 *T)
    sumY2  +=      33.922 * cos(0.59559025853  + 437.13694909800 *T)
    sumY2  +=      29.740 * cos(3.76166810670  + 270.16526422410 *T)
    sumY2  +=      34.418 * cos(6.09770226922  + 643.32249753521 *T)
    sumY2  +=      31.920 * cos(1.14730009415  + 337.97632814250 *T)
    sumY2  +=      35.309 * cos(3.07697728911  + 203.98168536590 *T)
    sumY2  +=      26.923 * cos(1.15025307434  + 131.79077970531 *T)
    sumY2  +=      26.272 * cos(0.57392625326  + 750.34742501690 *T)
    sumY2  +=      25.940 * cos(0.39280055678  + 76.50988875911 *T)
    sumY2  +=      29.423 * cos(3.94515878782  + 1280.03839011150 *T)
    sumY2  +=      26.782 * cos(4.26801091908  + 1155.11733992350 *T)
    sumY2  +=      24.375 * cos(1.05933381750  + 81.50831573269 *T)
    sumY2  +=      25.294 * cos(3.01418609936  + 483.46435966210 *T)
    sumY2  +=      23.209 * cos(5.71293607821  + 867.66729323710 *T)
    sumY2  +=      22.169 * cos(5.57360559022  + 288.32451148881 *T)
    sumY2  +=      23.117 * cos(1.43000944484  + 227.28237195610 *T)
    sumY2  +=      21.350 * cos(2.23904444472  + 160.85271488200 *T)
    sumY2  +=      22.067 * cos(3.22469653238  + 1044.91101870410 *T)
    sumY2  +=      22.088 * cos(3.66393393785  + 70.60562782070 *T)
    sumY2  +=      20.601 * cos(0.57364692423  + 22.33521801130 *T)
    sumY2  +=      20.035 * cos(4.14035560282  + 1148.00379292269 *T)
    sumY2  +=      26.634 * cos(4.29832630822  + 1052.02456570489 *T)
    sumY2  +=      20.249 * cos(5.80067112008  + 416.54706762100 *T)
    sumY2  +=      19.318 * cos(5.28106463151  + 497.69145366371 *T)
    sumY2  +=      17.591 * cos(5.18296940913  + 629.09540353360 *T)
    sumY2  +=      18.938 * cos(1.44732055781  + 107.26874496520 *T)
    sumY2  +=      21.595 * cos(4.07617789654  + 490.57790666290 *T)
    sumY2  +=      20.105 * cos(0.54192980606  + 839.21310523390 *T)
    sumY2  +=      21.565 * cos(2.45978600148  + 956.53297345411 *T)
    sumY2  +=      16.938 * cos(1.31185513413  + 89.10949770051 *T)
    sumY2  +=      16.497 * cos(3.14108050084  + 106.51798543980 *T)
    sumY2  +=      21.360 * cos(0.36044575923  + 845.83901726770 *T)
    sumY2  +=      19.384 * cos(5.02948605381  + 1258.21011414209 *T)
    sumY2  +=      15.417 * cos(1.59544409554  + 241.99710092470 *T)
    sumY2  +=      16.591 * cos(6.11245153858  + 639.65346883050 *T)
    sumY2  +=      16.177 * cos(4.47207109523  + 344.94686279140 *T)
    sumY2  +=      14.854 * cos(5.83097043394  + 266.23311096100 *T)
    sumY2  +=      19.063 * cos(0.50919845943  + 201.01273994931 *T)
    sumY2  +=      18.723 * cos(1.22925120460  + 226.07308589371 *T)
    sumY2  +=      18.567 * cos(4.58722431937  + 142.69346761730 *T)
    sumY2  +=      14.166 * cos(5.14705983766  + 632.53992182970 *T)
    sumY2  +=      17.119 * cos(5.54902094563  + 352.06040979221 *T)
    sumY2  +=      18.888 * cos(4.20468772536  + 52.44638055600 *T)
    sumY2  +=      13.506 * cos(2.46884301371  + 138.76131435421 *T)
    sumY2  +=      17.909 * cos(5.34068384361  + 934.70469748469 *T)
    sumY2  +=      13.747 * cos(0.53980747800  + 359.17395679300 *T)
    sumY2  +=      14.401 * cos(5.13574388998  + 405.50136735710 *T)
    sumY2  +=      16.878 * cos(1.18886685780  + 341.01470952831 *T)
    sumY2  +=      13.527 * cos(4.54185459719  + 284.39235822571 *T)
    sumY2  +=      13.436 * cos(2.93238200114  + 6290.43321447570 *T)
    sumY2  +=      12.962 * cos(5.56740362933  + 181.29958400710 *T)
    sumY2  +=      15.865 * cos(1.01845623601  + 651.18680406140 *T)
    sumY2  +=      13.136 * cos(0.57265499079  + 6276.20612047410 *T)
    sumY2  +=      12.088 * cos(5.67052895035  + 127.22797912329 *T)
    sumY2  +=      12.658 * cos(4.99618012837  + 948.93179148630 *T)
    sumY2  +=      12.303 * cos(2.19066167059  + 445.00125562421 *T)
    sumY2  +=      11.905 * cos(4.76438580603  + 424.39432780471 *T)
    sumY2  +=      12.624 * cos(4.35576033147  + 10220.64291069530 *T)
    sumY2  +=      12.885 * cos(5.28091161489  + 1375.52998236229 *T)
    sumY2  +=      12.454 * cos(0.80444795613  + 295.29504613770 *T)
    sumY2  +=      12.341 * cos(1.99603261002  + 10206.41581669370 *T)
    sumY2  +=      11.182 * cos(0.00456461469  + 209.12312469139 *T)
    sumY2  +=      10.860 * cos(1.22452077861  + 350.57593708390 *T)
    sumY2  +=      10.568 * cos(4.31488618104  + 12.77399045571 *T)
    sumY2  +=      10.726 * cos(4.07799467328  + 703.87700210090 *T)
    sumY2  +=      13.731 * cos(5.49770812826  + 636.95971005981 *T)
    sumY2  +=      11.066 * cos(4.06922769849  + 1471.50920958009 *T)
    sumY2  +=      10.568 * cos(2.45354520533  + 245.78624183590 *T)
    sumY2  +=      11.834 * cos(4.04162936965  + 191.45151239371 *T)
    sumY2  +=      13.045 * cos(4.56683285480  + 323.26159917390 *T)
    sumY2  +=      12.746 * cos(1.52919279431  + 729.00678401450 *T)
    sumY2  +=      11.214 * cos(0.69012319111  + 618.80046279510 *T)
    sumY2  +=      11.539 * cos(5.26204215590  + 75.02541605080 *T)
    sumY2  +=      12.236 * cos(6.03310641164  + 294.80741117069 *T)
    sumY2  +=       9.934 * cos(5.72889650913  + 209.87388421679 *T)
    sumY2  +=      10.182 * cos(4.02484159335  + 235.63431344930 *T)
    sumY2  +=       9.148 * cos(3.88097207377  + 1464.39566257930 *T)
    sumY2  +=       9.436 * cos(4.83010917251  + 508.10650660869 *T)
    sumY2  +=       8.422 * cos(0.17788593244  + 661.48174479990 *T)
    sumY2  +=       8.359 * cos(3.00746750344  + 1361.30288836070 *T)
    sumY2  +=       8.253 * cos(5.57801256284  + 565.35950523021 *T)
    sumY2  +=       9.478 * cos(0.55087151701  + 949.41942645330 *T)
    sumY2  +=      10.785 * cos(4.05427228564  + 1368.41643536150 *T)
    sumY2  +=       8.229 * cos(5.11025535831  + 5849.12029463110 *T)
    sumY2  +=       7.715 * cos(2.36462770264  + 10.80188278040 *T)
    sumY2  +=       8.470 * cos(1.44876764278  + 1052.51220067191 *T)
    sumY2  +=       7.922 * cos(1.18690846899  + 5863.34738863270 *T)
    sumY2  +=       7.332 * cos(1.23083436431  + 404.75060783170 *T)
    sumY2  +=       7.694 * cos(0.25064967436  + 9779.32999085069 *T)
    sumY2  +=       8.936 * cos(4.75203536861  + 1574.60198379870 *T)
    sumY2  +=       9.236 * cos(0.11752949743  + 1162.23088692429 *T)
    sumY2  +=       7.408 * cos(2.61049213458  + 9793.55708485230 *T)
    sumY2  +=       7.006 * cos(5.43604871599  + 208.87704647550 *T)
    sumY2  +=       8.885 * cos(5.06228722175  + 1251.09656714130 *T)
    sumY2  +=       6.858 * cos(5.72884043764  + 218.20877936750 *T)
    sumY2  +=       6.727 * cos(5.29793980060  + 313.45429340240 *T)
    sumY2  +=       6.237 * cos(0.59505976047  + 319.81708087780 *T)
    sumY2  +=       6.440 * cos(3.78505540532  + 429.28968891431 *T)
    sumY2  +=       8.565 * cos(1.73194061326  + 6283.31966747490 *T)
    sumY2  +=       7.708 * cos(4.22643181375  + 621.98185653281 *T)
    sumY2  +=       6.384 * cos(4.83193390171  + 1265.32366114289 *T)
    sumY2  +=       7.986 * cos(3.15525042033  + 10213.52936369450 *T)
    sumY2  +=       5.697 * cos(4.23986450365  + 121.49583896680 *T)
    sumY2  +=       5.771 * cos(0.86134691305  + 114.38229196601 *T)
    sumY2  +=       5.568 * cos(3.42931800801  + 194.89603068979 *T)
    sumY2  +=       6.666 * cos(5.54479786218  + 202.76033721600 *T)
    sumY2  +=       6.258 * cos(1.84744097041  + 117.07605073670 *T)
    sumY2  +=       5.194 * cos(4.52702758406  + 842.39449897160 *T)
    sumY2  +=       5.192 * cos(4.69562106653  + 78.95756931390 *T)
    sumY2  +=       6.155 * cos(3.58871751165  + 196.13442518220 *T)
    sumY2  +=       5.311 * cos(1.21158575749  + 956.04533848709 *T)
    sumY2  +=       6.053 * cos(1.30195529044  + 230.95140066080 *T)
    sumY2  +=       6.283 * cos(5.50104105259  + 63.49208081989 *T)
    sumY2  +=       5.550 * cos(5.07411282467  + 1691.92185201890 *T)
    sumY2  +=       6.169 * cos(3.68737408715  + 536.56069461190 *T)
    sumY2  +=       4.641 * cos(3.65270609087  + 436.40323591510 *T)
    sumY2  +=       4.643 * cos(0.07371316206  + 9.31741007210 *T)
    sumY2  +=       4.944 * cos(3.82470392475  + 1787.90107923670 *T)
    sumY2  +=       4.821 * cos(0.48227334704  + 644.07325706060 *T)
    sumY2  +=       4.551 * cos(0.43418901577  + 757.46097201770 *T)
    sumY2  +=       4.644 * cos(2.92326284592  + 9.80504503910 *T)
    sumY2  +=       4.204 * cos(5.51413105620  + 576.40520549410 *T)
    sumY2  +=       4.172 * cos(5.34743985805  + 251.67594855930 *T)
    sumY2  +=       4.304 * cos(5.72099336768  + 289.80898419710 *T)
    sumY2  +=       4.123 * cos(2.58679021252  + 175.40987728371 *T)
    sumY2  +=       4.456 * cos(4.56839890878  + 824.49837626530 *T)
    sumY2  +=       5.291 * cos(6.15232066316  + 1478.62275658089 *T)
    sumY2  +=       4.026 * cos(4.45906345257  + 710.99054910171 *T)
    sumY2  +=       5.280 * cos(5.68255683303  + 202.00957769060 *T)
    sumY2  +=       4.248 * cos(1.99938037336  + 398.38782035631 *T)
    sumY2  +=       3.877 * cos(4.60548911797  + 98.65617104110 *T)
    sumY2  +=       3.948 * cos(3.95821618688  + 1589.31671276730 *T)
    sumY2  +=       4.047 * cos(3.67831038408  + 1780.78753223589 *T)
    sumY2  +=       3.924 * cos(2.82058266387  + 134.34152612410 *T)
    sumY2  +=       4.377 * cos(1.51247705161  + 10.05112325500 *T)
    sumY2  +=       3.858 * cos(4.98661199042  + 501.62360692681 *T)
    sumY2  +=       3.847 * cos(3.25070592611  + 333.90116252750 *T)
    sumY2  +=       5.256 * cos(6.27523841381  + 5856.23384163189 *T)
    sumY2  +=       5.054 * cos(5.98297127509  + 479.53220639900 *T)
    sumY2  +=       4.175 * cos(1.22869956774  + 63.97971578690 *T)
    sumY2  +=       4.150 * cos(5.64863928594  + 106.03035047280 *T)
    sumY2  +=       3.760 * cos(2.68800271750  + 508.59414157570 *T)
    sumY2  +=       4.966 * cos(1.12070430834  + 216.23667169220 *T)
    sumY2  +=       4.880 * cos(3.70286182159  + 1684.80830501809 *T)
    sumY2  +=       4.042 * cos(1.45643909331  + 120.11443212250 *T)
    sumY2  +=       4.914 * cos(1.41648464476  + 9786.44353785149 *T)
    sumY2  +=       3.672 * cos(5.52203210982  + 422.42222012939 *T)
    sumY2  +=       4.671 * cos(4.79630546138  + 1567.48843679789 *T)
    sumY2  +=       3.483 * cos(3.63502959369  + 348.12825652910 *T)
    sumY2  +=       3.700 * cos(4.27862787379  + 416.05943265400 *T)
    sumY2  +=       4.300 * cos(4.48648057472  + 1890.99385345530 *T)
    sumY2  +=       3.599 * cos(2.97275381758  + 2.93757625420 *T)
    sumY2  +=       3.544 * cos(3.88355984367  + 852.95256426850 *T)
    sumY2  +=       3.267 * cos(2.77536057636  + 1677.69475801730 *T)
    sumY2  +=       3.574 * cos(4.37725311704  + 156.92056161890 *T)
    sumY2  +=       3.416 * cos(0.18832491676  + 167.07249000551 *T)
    sumY2  +=       3.394 * cos(1.52845073217  + 196.27743753410 *T)
    sumY2  +=       3.544 * cos(0.55935420202  + 114.64292439690 *T)
    sumY2  +=       3.999 * cos(1.65622001101  + 326.93062787860 *T)
    sumY2  +=       3.152 * cos(5.90994469916  + 151.29148732640 *T)
    sumY2  +=       3.262 * cos(0.99529440929  + 1272.43720814370 *T)
    sumY2  +=       3.054 * cos(2.50692121176  + 749.85979004990 *T)
    sumY2  +=       3.350 * cos(3.78153899621  + 212.57970507500 *T)
    sumY2  +=       3.340 * cos(1.11106690501  + 214.50612076800 *T)
    sumY2  +=       3.138 * cos(5.41161067084  + 320.56784040320 *T)
    sumY2  +=       3.732 * cos(5.75933663800  + 611.19928082730 *T)
    sumY2  +=       3.741 * cos(5.33542182003  + 849.50804597240 *T)
    sumY2  +=       3.047 * cos(0.21958282546  + 230.80838830890 *T)
    sumY2  +=       3.048 * cos(1.23176928659  + 21.09682351890 *T)
    sumY2  +=       2.851 * cos(0.01947979319  + 6069.53293706989 *T)
    sumY2  +=       2.864 * cos(1.37901080071  + 9999.74263328950 *T)
    sumY2  +=       2.898 * cos(0.80640653869  + 213.49472881170 *T)
    sumY2  +=       2.898 * cos(4.08606578636  + 213.59109703130 *T)
    sumY2  +=       2.819 * cos(4.62496885979  + 431.50787480550 *T)
    sumY2  +=       3.469 * cos(3.39462934689  + 526.75338884040 *T)
    sumY2  +=       3.327 * cos(4.72501626019  + 1581.71553079950 *T)
    sumY2  +=       2.927 * cos(0.75688922418  + 395.82251972250 *T)
    sumY2  +=       3.762 * cos(5.33741882811  + 38.37685312130 *T)
    sumY2  +=       2.893 * cos(4.68179275926  + 381.59542572090 *T)
    sumY2  +=       3.707 * cos(0.28876454398  + 215.99059347631 *T)
    sumY2  +=       2.804 * cos(1.54781529217  + 260.01333583751 *T)
    sumY2  +=       2.799 * cos(1.64426930831  + 129.67565967810 *T)
    sumY2  +=       3.378 * cos(0.03634726419  + 299.37021175271 *T)
    sumY2  +=       2.783 * cos(4.12159252831  + 1287.15193711231 *T)
    sumY2  +=       3.526 * cos(4.64846209915  + 211.09523236671 *T)
    sumY2  +=       2.541 * cos(3.72750148868  + 414.31183538731 *T)
    sumY2  +=       2.607 * cos(5.40709902081  + 276.79117625789 *T)
    sumY2  +=       2.465 * cos(5.81538596066  + 572.47305223100 *T)
    sumY2  +=       2.461 * cos(2.38415136907  + 409.43352062020 *T)
    sumY2  +=       2.462 * cos(0.41590882019  + 59.55992755679 *T)
    sumY2  +=       2.579 * cos(3.41605150344  + 1493.33748554950 *T)
    sumY2  +=       2.414 * cos(4.38756196239  + 343.46239008310 *T)
    sumY2  +=       2.533 * cos(2.34096298657  + 696.76345510010 *T)
    sumY2  +=       2.390 * cos(0.78788852748  + 1588.82907780029 *T)
    sumY2  +=       2.433 * cos(1.96229119408  + 6070.02057203690 *T)
    sumY2  +=       2.747 * cos(5.20221540877  + 216.98743121760 *T)
    sumY2  +=       2.420 * cos(6.19374294302  + 45.33283355520 *T)
    sumY2  +=       2.297 * cos(4.18882853069  + 99.66756299740 *T)
    sumY2  +=       2.813 * cos(4.09700943877  + 422.17614191350 *T)
    sumY2  +=       2.913 * cos(0.57159915913  + 551.27542358051 *T)
    sumY2  +=       2.860 * cos(4.18470915906  + 69.39634175830 *T)
    sumY2  +=       2.387 * cos(2.09096895837  + 1169.83206889211 *T)
    sumY2  +=       2.258 * cos(1.85441482871  + 188.27011865600 *T)
    sumY2  +=       2.293 * cos(5.87858665828  + 98.91680347199 *T)
    sumY2  +=       2.711 * cos(5.95954660965  + 1795.01462623750 *T)
    sumY2  +=       2.337 * cos(3.41538474921  + 10000.23026825651 *T)
    sumY2  +=       2.777 * cos(0.35017301148  + 100.15519796440 *T)
    sumY2  +=       3.022 * cos(5.84737063031  + 190.96387742670 *T)
    sumY2  +=       2.327 * cos(4.00268284025  + 143.90275367970 *T)
    sumY2  +=       2.152 * cos(3.39690921596  + 189.16389053330 *T)
    sumY2  +=       2.147 * cos(1.10129984315  + 265.74547599400 *T)
    sumY2  +=       2.162 * cos(3.02953405239  + 238.81570718700 *T)
    sumY2  +=       2.184 * cos(3.41242359251  + 917.17609753890 *T)
    sumY2  +=       2.297 * cos(4.31295262287  + 533.11617631580 *T)
    sumY2  +=       2.099 * cos(2.06100500470  + 275.79433851660 *T)
    sumY2  +=       2.122 * cos(4.75342370868  + 93.18466331550 *T)
    sumY2  +=       2.068 * cos(2.33266360886  + 221.61966776881 *T)
    sumY2  +=       2.635 * cos(5.89696027686  + 646.76701583129 *T)
    sumY2  +=       2.175 * cos(4.05922699845  + 432.47108265200 *T)
    sumY2  +=       2.049 * cos(2.54249741628  + 205.46615807421 *T)
    sumY2  +=       2.103 * cos(1.50326611692  + 554.31380496631 *T)
    sumY2  +=       2.036 * cos(5.63434399596  + 415.30867312860 *T)
    sumY2  +=       2.045 * cos(2.97922154502  + 547.20025796550 *T)
    sumY2  +=       1.996 * cos(0.02214888903  + 5642.93474619389 *T)
    sumY2  +=       2.709 * cos(1.33128572135  + 174.18603700630 *T)
    sumY2  +=       2.499 * cos(5.48814423722  + 927.59115048390 *T)
    sumY2  +=       2.481 * cos(3.81894500619  + 91.80325647120 *T)
    sumY2  +=       2.555 * cos(6.20493162777  + 196.86813836510 *T)
    sumY2  +=       2.555 * cos(4.96787984600  + 230.21768747790 *T)
    sumY2  +=       2.082 * cos(2.24744287595  + 832.09955823310 *T)
    sumY2  +=       2.231 * cos(3.62093517460  + 2097.17940189250 *T)
    sumY2  +=       2.095 * cos(4.31580761873  + 1140.89024592190 *T)
    sumY2  +=       1.920 * cos(6.22028733359  + 99.40443843901 *T)
    sumY2  +=       1.910 * cos(4.09848069601  + 106.78110999820 *T)
    sumY2  +=       2.301 * cos(4.90677041319  + 2008.31372167550 *T)
    sumY2  +=       2.228 * cos(4.84472911427  + 219.69325207580 *T)
    sumY2  +=       2.219 * cos(0.03409249836  + 207.39257376720 *T)
    sumY2  +=       2.444 * cos(1.01754308284  + 71.09326278771 *T)
    sumY2  +=       1.950 * cos(4.34886592902  + 417.28078080390 *T)
    sumY2  +=       1.941 * cos(1.80523291314  + 374.15181032000 *T)
    sumY2  +=       1.900 * cos(4.76624814243  + 333.41352756050 *T)
    sumY2  +=       2.353 * cos(6.10168291345  + 650.43604453600 *T)
    sumY2  +=       1.819 * cos(1.46865726367  + 120.74507944141 *T)
    sumY2  +=       1.992 * cos(3.25021395969  + 2413.57127154910 *T)
    sumY2  +=       1.805 * cos(1.41126788033  + 9573.14444241349 *T)
    sumY2  +=       2.272 * cos(4.49997759337  + 135.57992061650 *T)
    sumY2  +=       1.823 * cos(3.75716328579  + 425.87880051300 *T)
    sumY2  +=       2.143 * cos(0.94827482019  + 347.64062156210 *T)
    sumY2  +=       1.841 * cos(5.18057670314  + 220.70464403210 *T)
    sumY2  +=       1.834 * cos(5.98123022671  + 206.38118181090 *T)
    sumY2  +=       1.777 * cos(4.24421513244  + 116.18227885939 *T)
    sumY2  +=       2.138 * cos(3.72480039128  + 2104.29294889330 *T)
    sumY2  +=       1.887 * cos(3.37825539740  + 556.76148552110 *T)
    sumY2  +=       2.390 * cos(5.29068620866  + 856.62159297321 *T)
    sumY2  +=       1.825 * cos(0.29367232391  + 345.08987514331 *T)
    sumY2  +=       1.730 * cos(4.70941820611  + 237.92193530970 *T)
    sumY2  +=       2.103 * cos(4.16707885187  + 2207.38572311189 *T)
    sumY2  +=       1.779 * cos(1.21791991667  + 427.80521620600 *T)
    sumY2  +=       1.911 * cos(0.45079864675  + 145.87486135500 *T)
    sumY2  +=       1.811 * cos(5.74280960496  + 570.98857952271 *T)
    sumY2  +=       1.710 * cos(4.07545471254  + 423.17297965479 *T)
    sumY2  +=       2.261 * cos(1.53982829642  + 421.42538238810 *T)
    sumY2  +=       1.681 * cos(0.27164412710  + 305.85311143459 *T)
    sumY2  +=       2.093 * cos(3.09304115315  + 1169.34443392509 *T)
    sumY2  +=       2.071 * cos(0.57780048368  + 942.30587945250 *T)
    sumY2  +=       1.691 * cos(0.86915956975  + 455.29619636271 *T)
    sumY2  +=       2.080 * cos(4.93280502131  + 405.01373239010 *T)
    sumY2  +=       1.764 * cos(1.22791859864  + 186.71620997851 *T)
    sumY2  +=       1.662 * cos(3.65838163965  + 305.58998687620 *T)
    sumY2  +=       1.748 * cos(1.78978347421  + 899.01685027420 *T)
    sumY2  +=       1.696 * cos(0.02813095646  + 85.17734443740 *T)
    sumY2  +=       2.142 * cos(3.56494929765  + 2001.20017467470 *T)
    sumY2  +=       2.282 * cos(0.47292670815  + 439.37218133171 *T)
    sumY2  +=       1.748 * cos(2.40026034305  + 1066.25165970650 *T)
    sumY2  +=       1.908 * cos(4.56312338515  + 131.64776735340 *T)
    sumY2  +=       1.808 * cos(4.08414705157  + 1457.28211557850 *T)
    sumY2  +=       1.794 * cos(4.25056477507  + 913.24394427581 *T)
    sumY2  +=       2.167 * cos(4.61051989673  + 563.87503252191 *T)
    sumY2  +=       1.602 * cos(1.84763847168  + 327.68138740401 *T)
    sumY2  +=       1.891 * cos(1.79946288763  + 1155.60497489050 *T)
    sumY2  +=       2.205 * cos(5.85470490114  + 388.70897272171 *T)
    sumY2  +=       1.597 * cos(4.15810623296  + 281.21096448800 *T)
    sumY2  +=       1.719 * cos(1.15830791285  + 312.44290144610 *T)
    sumY2  +=       2.158 * cos(2.28726893212  + 224.10097821840 *T)
    sumY2  +=       1.659 * cos(4.74001893135  + 561.42735196710 *T)
    sumY2  +=       1.737 * cos(0.17490306013  + 167.96626188280 *T)
    sumY2  +=       1.501 * cos(0.25666873016  + 312.96665843540 *T)
    sumY2  +=       1.587 * cos(3.94775538143  + 732.45130231059 *T)
    sumY2  +=       1.488 * cos(2.72096839740  + 12.28635548869 *T)
    sumY2  +=       1.460 * cos(2.11177924228  + 2627.35800195410 *T)
    sumY2  +=       1.835 * cos(2.89416763419  + 206.47755003051 *T)
    sumY2  +=       1.708 * cos(3.71608804699  + 82.13896305161 *T)
    sumY2  +=       1.842 * cos(1.98442548678  + 220.60827581251 *T)
    sumY2  +=       1.895 * cos(3.47648673336  + 291.50590522650 *T)
    sumY2  +=       1.423 * cos(5.61838953778  + 225.32232636830 *T)
    sumY2  +=       1.423 * cos(5.55442193599  + 201.76349947470 *T)
    sumY2  +=       1.568 * cos(4.88514907741  + 685.71775483620 *T)
    sumY2  +=       1.663 * cos(0.82161037585  + 3597.87425192630 *T)
    sumY2  +=       1.491 * cos(0.52011503386  + 240.36961586451 *T)
    sumY2  +=       1.913 * cos(0.07515794396  + 295.43805848960 *T)
    sumY2  +=       1.496 * cos(1.53621258084  + 1045.39865367110 *T)
    sumY2  +=       1.477 * cos(0.63538013406  + 440.58146739410 *T)
    sumY2  +=       1.410 * cos(2.28174385720  + 137.27684164590 *T)
    sumY2  +=       1.748 * cos(5.19272444741  + 519.15220687259 *T)
    sumY2  +=       1.325 * cos(1.41732134877  + 980.91199584231 *T)
    sumY2  +=       1.352 * cos(0.82085755018  + 5429.63565075589 *T)
    sumY2  +=       1.414 * cos(1.50139243083  + 1371.59782909919 *T)
    sumY2  +=       1.697 * cos(4.54037425682  + 1898.10740045609 *T)
    sumY2  +=       1.424 * cos(4.89243708936  + 721.40560204669 *T)
    sumY2  +=       1.424 * cos(1.75084443577  + 721.89323701370 *T)
    sumY2  +=       1.295 * cos(0.66380326238  + 1905.22094745690 *T)
    sumY2  +=       1.272 * cos(4.15261431749  + 127.71561409031 *T)
    sumY2  +=       1.740 * cos(4.55214034273  + 1883.88030645450 *T)
    sumY2  +=       1.435 * cos(0.26851729343  + 699.94484883781 *T)
    sumY2  +=       1.260 * cos(5.67254228493  + 850.25880549781 *T)
    sumY2  +=       1.316 * cos(2.05653832811  + 1382.64352936309 *T)
    sumY2  +=       1.303 * cos(2.77196448848  + 28.69800548670 *T)
    sumY2  +=       1.701 * cos(1.44281649853  + 312.70353387701 *T)
    sumY2  +=       1.245 * cos(2.24439786330  + 9359.84534697549 *T)
    sumY2  +=       1.240 * cos(2.86058076462  + 1994.08662767389 *T)
    sumY2  +=       1.286 * cos(3.69806111108  + 1048.84317196720 *T)
    sumY2  +=       1.240 * cos(4.77153818320  + 148.32254190981 *T)
    sumY2  +=       1.365 * cos(0.95560209763  + 212.79215339611 *T)
    sumY2  +=       1.362 * cos(3.93431918921  + 214.29367244690 *T)
    sumY2  +=       1.206 * cos(3.84953492409  + 739.56484931140 *T)
    sumY2  +=       1.503 * cos(5.21115384219  + 1243.98302014049 *T)
    sumY2  +=       1.305 * cos(1.57242424750  + 259.11956396020 *T)
    sumY2  +=       1.167 * cos(1.86683877227  + 233.18663289450 *T)
    sumY2  +=       1.167 * cos(3.02278739432  + 193.89919294850 *T)
    sumY2  +=       1.155 * cos(0.52398011340  + 153.73916788121 *T)
    sumY2  +=       1.220 * cos(6.08058213690  + 283.90472325870 *T)
    sumY2  +=       1.153 * cos(6.21046784748  + 525.74199688410 *T)
    sumY2  +=       1.188 * cos(2.94352017149  + 258.63192899320 *T)
    sumY2  +=       1.558 * cos(4.67220714738  + 4216.43089723790 *T)
    sumY2  +=       1.423 * cos(3.77197308459  + 2420.68481854989 *T)
    sumY2  +=       1.138 * cos(0.76214399254  + 17.91542978119 *T)
    sumY2  +=       1.133 * cos(2.84505877314  + 841.90686400460 *T)
    sumY2  +=       1.109 * cos(1.05776040432  + 121.00820399980 *T)
    sumY2  +=       1.109 * cos(0.17173305113  + 629.84616305900 *T)
    sumY2  +=       1.434 * cos(3.90656801196  + 6062.41939006910 *T)
    sumY2  +=       1.122 * cos(2.45857726727  + 128.60938596760 *T)
    sumY2  +=       1.123 * cos(3.06803578936  + 3178.38960805111 *T)
    sumY2  +=       1.154 * cos(4.32058575911  + 67.91186905000 *T)
    sumY2  +=       1.130 * cos(5.25611734741  + 731.70054278519 *T)
    sumY2  +=       1.305 * cos(2.25262249590  + 97.11681657861 *T)
    sumY2  +=       1.248 * cos(3.82251460637  + 2523.77759276850 *T)
    sumY2  +=       1.161 * cos(2.25840723150  + 81.99595069970 *T)
    sumY2  +=       1.178 * cos(0.70097671757  + 1056.44435393500 *T)
    sumY2  +=       1.342 * cos(5.62104702567  + 355.99256305531 *T)
    sumY2  +=       1.249 * cos(4.92175219077  + 526.26575387340 *T)
    sumY2  +=       1.034 * cos(4.56345495168  + 777.17412795991 *T)
    sumY2  +=       1.332 * cos(4.90721267162  + 835.54407652920 *T)
    sumY2  +=       1.024 * cos(6.03631011372  + 5635.82119919310 *T)
    sumY2  +=       1.173 * cos(4.16447862889  + 74.53778108379 *T)
    sumY2  +=       1.173 * cos(3.45615871164  + 2317.59204433130 *T)
    sumY2  +=       1.012 * cos(2.86472380431  + 215.23983395090 *T)
    sumY2  +=       1.186 * cos(3.42680148463  + 1272.92484311070 *T)
    sumY2  +=       1.049 * cos(1.32104580559  + 6062.90702503610 *T)
    sumY2  +=       1.154 * cos(3.31945570971  + 2406.45772454830 *T)
    sumY2  +=       1.062 * cos(3.40252606572  + 906.13039727500 *T)
    sumY2  +=       1.124 * cos(0.61967036641  + 3914.26612158291 *T)
    sumY2  +=       0.993 * cos(1.99202380463  + 211.84599189210 *T)
    sumY2  +=       0.986 * cos(6.21774784663  + 525.51499434800 *T)
    sumY2  +=       0.986 * cos(5.89248361654  + 602.00806815971 *T)
    sumY2  +=       0.992 * cos(0.16374858268  + 306.34074640161 *T)
    sumY2  +=       1.239 * cos(4.40402775047  + 2193.15862911030 *T)
    sumY2  +=       1.134 * cos(3.08772580443  + 2729.96314120569 *T)
    sumY2  +=       1.041 * cos(2.53935646805  + 945.75039774860 *T)
    sumY2  +=       1.025 * cos(3.08133309612  + 32.48714639791 *T)
    sumY2  +=       1.044 * cos(1.56254364362  + 472.41865939820 *T)
    sumY2  +=       1.335 * cos(4.80329987516  + 2324.70559133209 *T)
    sumY2  +=       1.199 * cos(4.40804047285  + 4113.33812301930 *T)
    sumY2  +=       1.174 * cos(4.67336952922  + 1176.94561589290 *T)
    sumY2  +=       1.213 * cos(0.28347383353  + 252.89978883670 *T)
    sumY2  +=       0.973 * cos(1.75775518324  + 245.03548231051 *T)
    sumY2  +=       0.973 * cos(6.27346363694  + 182.05034353251 *T)
    sumY2  +=       1.167 * cos(6.08159477667  + 635.72131556739 *T)
    sumY2  +=       0.936 * cos(3.32125427673  + 220.13519506050 *T)
    sumY2  +=       1.073 * cos(3.85483987093  + 1773.67398523510 *T)
    sumY2  +=       0.931 * cos(1.55716679047  + 206.95063078250 *T)
    sumY2  +=       0.931 * cos(2.65028378733  + 1582.20316576650 *T)
    sumY2  +=       1.237 * cos(1.19959870124  + 2964.60287764609 *T)
    sumY2  +=       1.074 * cos(5.00882181931  + 738.81408978600 *T)
    sumY2  +=       0.976 * cos(6.18216851835  + 31.99951143090 *T)
    sumY2  +=       0.907 * cos(3.58111072001  + 1802.61580820530 *T)
    sumY2  +=       0.925 * cos(2.76573994536  + 3494.78147770770 *T)
    sumY2  +=       0.999 * cos(4.06676744721  + 171.89841514590 *T)
    sumY2  +=       0.931 * cos(4.76188288294  + 1037.79747170330 *T)
    sumY2  +=       1.094 * cos(3.85667770199  + 2516.66404576769 *T)
    sumY2  +=       0.949 * cos(3.82779842438  + 1809.72935520611 *T)
    sumY2  +=       0.964 * cos(2.57800328808  + 2641.58509595570 *T)
    sumY2  +=       0.947 * cos(5.94253987889  + 2111.40649589409 *T)
    sumY2  +=       0.898 * cos(1.16267053643  + 9566.03089541269 *T)
    sumY2  +=       1.029 * cos(3.26133516597  + 519.63984183961 *T)
    sumY2  +=       0.989 * cos(3.95452500658  + 255.18741069710 *T)
    sumY2  +=       1.003 * cos(4.46395476902  + 60.52313540329 *T)
    sumY2  +=       0.886 * cos(0.72851955442  + 305.10235190919 *T)
    sumY2  +=       0.920 * cos(1.69743225247  + 459.08533727390 *T)
    sumY2  +=       0.888 * cos(1.37138485948  + 3259.65410630030 *T)
    sumY2  +=       0.888 * cos(4.51297751307  + 3260.14174126730 *T)
    sumY2  +=       0.960 * cos(0.18926013762  + 84.68970947040 *T)
    sumY2  +=       0.881 * cos(0.63602093080  + 397.90018538930 *T)
    sumY2  +=       0.924 * cos(5.21301171350  + 731.92754532131 *T)
    sumY2  +=       0.960 * cos(2.83183840474  + 5856.72147659890 *T)
    sumY2  +=       0.886 * cos(0.13702548516  + 471.93102443120 *T)
    sumY2  +=       0.909 * cos(1.30383200776  + 746.15463932291 *T)

    # SATURN Y Coord
    sumY3  = 0
    sumY3  +=   76886.822 * cos(1.38647122579  + 206.42936592071 *T)
    sumY3  +=   75865.107 * cos(0.30263761509  + 220.65645992230 *T)
    sumY3  +=   28377.246 * cos(2.29477516800  + 0.24381748350 *T)
    sumY3  +=   11451.297 * cos(5.24189857677  + 426.84200835950 *T)
    sumY3  +=    9043.324 * cos(5.61883049962  + 433.95555536030 *T)
    sumY3  +=    7419.530 * cos(0.21033373142  + 227.77000692311 *T)
    sumY3  +=    8292.033 * cos(1.50182614839  + 7.35736448430 *T)
    sumY3  +=    4421.142 * cos(0.64077885911  + 419.72846135871 *T)
    sumY3  +=    3519.324 * cos(1.53956700451  + 199.31581891990 *T)
    sumY3  +=    3064.445 * cos(4.39386454839  + 640.14110379750 *T)
    sumY3  +=    2363.042 * cos(1.46950181949  + 14.47091148511 *T)
    sumY3  +=    1027.777 * cos(4.71705572637  + 647.25465079831 *T)
    sumY3  +=     789.743 * cos(5.60732845352  + 441.06910236111 *T)
    sumY3  +=     810.744 * cos(5.98445458539  + 216.72430665921 *T)
    sumY3  +=     766.386 * cos(2.03195290594  + 210.36151918381 *T)
    sumY3  +=     673.097 * cos(5.33794721338  + 213.05527795450 *T)
    sumY3  +=     506.614 * cos(2.17347656423  + 213.54291292150 *T)
    sumY3  +=     384.956 * cos(1.47307672393  + 117.56368570370 *T)
    sumY3  +=     463.540 * cos(3.55273822694  + 853.44019923550 *T)
    sumY3  +=     411.246 * cos(2.18199166630  + 110.45013870291 *T)
    sumY3  +=     324.142 * cos(0.07852229440  + 234.88355392390 *T)
    sumY3  +=     285.917 * cos(0.41121094808  + 412.61491435790 *T)
    sumY3  +=     238.406 * cos(1.75200857251  + 103.33659170210 *T)
    sumY3  +=     192.441 * cos(5.32707204371  + 316.63568714010 *T)
    sumY3  +=     189.600 * cos(0.69653910652  + 6.86972951729 *T)
    sumY3  +=     189.610 * cos(0.69106705292  + 633.02755679670 *T)
    sumY3  +=     213.874 * cos(0.85839521405  + 205.94173095370 *T)
    sumY3  +=     153.605 * cos(0.70666490608  + 323.74923414091 *T)
    sumY3  +=     154.083 * cos(1.31044116162  + 21.58445848590 *T)
    sumY3  +=     130.169 * cos(6.09318414425  + 209.61075965840 *T)
    sumY3  +=     121.811 * cos(0.47800554959  + 198.82818395289 *T)
    sumY3  +=     112.652 * cos(6.07737302889  + 220.16882495529 *T)
    sumY3  +=     109.203 * cos(1.75938966834  + 217.47506618460 *T)
    sumY3  +=     111.589 * cos(3.84220589954  + 860.55374623631 *T)
    sumY3  +=      97.336 * cos(2.54240833553  + 95.73540973430 *T)
    sumY3  +=     114.728 * cos(5.87380874959  + 223.83785366000 *T)
    sumY3  +=      85.497 * cos(0.31664832561  + 330.86278114170 *T)
    sumY3  +=     103.793 * cos(1.39004326725  + 309.52214013930 *T)
    sumY3  +=      85.243 * cos(4.74344019433  + 654.36819779910 *T)
    sumY3  +=      81.972 * cos(1.72088205600  + 192.20227191910 *T)
    sumY3  +=      93.735 * cos(4.66924086636  + 426.35437339250 *T)
    sumY3  +=      78.232 * cos(6.26955569087  + 124.67723270450 *T)
    sumY3  +=      73.732 * cos(2.62274205530  + 88.62186273350 *T)
    sumY3  +=      68.659 * cos(4.66941001827  + 109.96250373590 *T)
    sumY3  +=      87.657 * cos(5.07624062638  + 430.02340209721 *T)
    sumY3  +=      57.790 * cos(0.13357560704  + 224.58861318540 *T)
    sumY3  +=      50.888 * cos(1.52333865500  + 202.49721265760 *T)
    sumY3  +=      48.029 * cos(2.25988634275  + 302.40859313851 *T)
    sumY3  +=      57.606 * cos(2.26643847681  + 203.24797218300 *T)
    sumY3  +=      56.960 * cos(2.75566102022  + 1066.73929467350 *T)
    sumY3  +=      45.308 * cos(1.68344191762  + 96.22304470130 *T)
    sumY3  +=      48.067 * cos(1.37320484001  + 309.03450517230 *T)
    sumY3  +=      36.732 * cos(5.51470119579  + 448.18264936190 *T)
    sumY3  +=      38.865 * cos(2.20650641493  + 625.91400979590 *T)
    sumY3  +=      43.853 * cos(4.92071216020  + 422.90985509640 *T)
    sumY3  +=      36.680 * cos(0.38116904350  + 13.98327651810 *T)
    sumY3  +=      35.412 * cos(3.97478676019  + 522.82123557730 *T)
    sumY3  +=      30.232 * cos(5.44165264045  + 412.12727939089 *T)
    sumY3  +=      34.292 * cos(1.35598862213  + 301.92095817150 *T)
    sumY3  +=      25.919 * cos(3.74498868983  + 4.17597074660 *T)
    sumY3  +=      26.527 * cos(1.95025295633  + 423.66061462181 *T)
    sumY3  +=      23.623 * cos(0.11328528018  + 515.22005360950 *T)
    sumY3  +=      17.831 * cos(3.67540680263  + 11.28951774740 *T)
    sumY3  +=      17.289 * cos(5.69767719550  + 537.04832957890 *T)
    sumY3  +=      16.881 * cos(3.81249765736  + 195.38366565681 *T)
    sumY3  +=      13.511 * cos(4.15427151016  + 515.70768857651 *T)
    sumY3  +=      12.893 * cos(5.83608381063  + 437.88770862340 *T)
    sumY3  +=      11.197 * cos(4.12727958334  + 742.74624304909 *T)
    sumY3  +=      13.451 * cos(0.48023955262  + 522.33360061030 *T)
    sumY3  +=      10.292 * cos(5.57616897093  + 337.97632814250 *T)
    sumY3  +=      11.481 * cos(0.55900801750  + 191.71463695210 *T)
    sumY3  +=      11.394 * cos(1.22639782189  + 625.42637482889 *T)
    sumY3  +=      11.380 * cos(3.01802304028  + 1073.85284167431 *T)
    sumY3  +=       9.884 * cos(5.68207864841  + 131.79077970531 *T)
    sumY3  +=       9.268 * cos(2.85677606430  + 81.50831573269 *T)
    sumY3  +=      10.374 * cos(5.65556228759  + 544.16187657970 *T)
    sumY3  +=      11.913 * cos(5.03981077547  + 437.13694909800 *T)
    sumY3  +=       8.626 * cos(5.23537891674  + 728.51914904750 *T)
    sumY3  +=       8.671 * cos(5.89546440948  + 419.24082639170 *T)
    sumY3  +=       8.821 * cos(6.14734947376  + 831.61192326610 *T)
    sumY3  +=       8.112 * cos(5.21246168400  + 149.80701461810 *T)
    sumY3  +=       9.768 * cos(1.10645539833  + 618.31282782810 *T)
    sumY3  +=       8.329 * cos(3.90263564155  + 867.66729323710 *T)
    sumY3  +=       7.645 * cos(0.82506284210  + 846.32665223470 *T)
    sumY3  +=       8.659 * cos(4.22366027894  + 643.32249753521 *T)
    sumY3  +=       7.205 * cos(5.97033515631  + 227.28237195610 *T)
    sumY3  +=       7.157 * cos(4.06588017100  + 636.20895053440 *T)
    sumY3  +=       7.036 * cos(1.06159700654  + 231.70216018621 *T)
    sumY3  +=       7.083 * cos(3.27104541387  + 89.10949770051 *T)
    sumY3  +=       6.556 * cos(1.10055926440  + 743.23387801611 *T)
    sumY3  +=       5.903 * cos(5.58614673874  + 270.16526422410 *T)
    sumY3  +=       6.615 * cos(5.20910906290  + 433.46792039329 *T)
    sumY3  +=       6.469 * cos(0.48331749487  + 408.68276109481 *T)
    sumY3  +=       5.426 * cos(6.23003978027  + 241.99710092470 *T)
    sumY3  +=       5.648 * cos(0.98336733799  + 529.93478257810 *T)
    sumY3  +=       7.105 * cos(1.44824281594  + 18.40306474821 *T)
    sumY3  +=       5.052 * cos(3.61172453455  + 416.54706762100 *T)
    sumY3  +=       6.189 * cos(3.11325347013  + 736.12033101530 *T)
    sumY3  +=       5.821 * cos(1.93537322437  + 1280.03839011150 *T)
    sumY3  +=       5.043 * cos(2.19686528643  + 277.27881122490 *T)
    sumY3  +=       4.744 * cos(3.03559141046  + 56.37853381910 *T)
    sumY3  +=       5.705 * cos(3.29935860753  + 430.77416162260 *T)
    sumY3  +=       5.226 * cos(3.72017940288  + 415.79630809560 *T)
    sumY3  +=       5.511 * cos(5.16650332361  + 529.44714761109 *T)
    sumY3  +=       4.520 * cos(3.93668601315  + 1059.13811270570 *T)
    sumY3  +=       4.583 * cos(1.57866732584  + 294.80741117069 *T)
    sumY3  +=       4.789 * cos(2.57991417560  + 295.29504613770 *T)
    sumY3  +=       4.026 * cos(4.75259771651  + 483.46435966210 *T)
    sumY3  +=       3.728 * cos(4.95619377571  + 185.08872491831 *T)
    sumY3  +=       3.578 * cos(0.94204754728  + 405.50136735710 *T)
    sumY3  +=       3.890 * cos(5.00299059027  + 750.34742501690 *T)
    sumY3  +=       4.291 * cos(0.89282309495  + 941.81824448550 *T)
    sumY3  +=       3.758 * cos(4.69738792182  + 661.48174479990 *T)
    sumY3  +=       3.547 * cos(1.97438574685  + 70.60562782070 *T)
    sumY3  +=       3.641 * cos(2.23934061320  + 839.21310523390 *T)
    sumY3  +=       4.196 * cos(2.40610831590  + 618.80046279510 *T)
    sumY3  +=       3.709 * cos(0.25270527416  + 508.10650660869 *T)
    sumY3  +=       4.193 * cos(0.33797589356  + 838.72547026689 *T)
    sumY3  +=       3.571 * cos(6.03135821293  + 1148.00379292269 *T)
    sumY3  +=       3.647 * cos(0.88659408340  + 934.70469748469 *T)
    sumY3  +=       2.978 * cos(5.00861879701  + 1044.91101870410 *T)
    sumY3  +=       2.876 * cos(3.72277336310  + 398.38782035631 *T)
    sumY3  +=       2.643 * cos(3.66740434616  + 497.69145366371 *T)
    sumY3  +=       2.725 * cos(4.31936427464  + 363.10611005610 *T)
    sumY3  +=       3.246 * cos(4.14230106799  + 639.65346883050 *T)
    sumY3  +=       2.721 * cos(5.71644537210  + 1464.39566257930 *T)
    sumY3  +=       2.917 * cos(0.81283070390  + 113.63153244060 *T)
    sumY3  +=       2.560 * cos(3.37970059304  + 729.00678401450 *T)
    sumY3  +=       2.380 * cos(4.67244312905  + 86.07111631471 *T)
    sumY3  +=       2.667 * cos(1.66091865719  + 10.80188278040 *T)
    sumY3  +=       2.238 * cos(0.69261990800  + 10.53875822200 *T)
    sumY3  +=       2.305 * cos(0.45579083441  + 956.53297345411 *T)
    sumY3  +=       2.316 * cos(0.44164789425  + 3.42521122121 *T)
    sumY3  +=       2.282 * cos(1.20272003403  + 611.19928082730 *T)
    sumY3  +=       1.988 * cos(4.99900390794  + 651.18680406140 *T)
    sumY3  +=       2.365 * cos(0.61916344902  + 2111.40649589409 *T)
    sumY3  +=       2.485 * cos(5.89913021216  + 316.14805217309 *T)
    sumY3  +=       2.670 * cos(5.34319853125  + 196.13442518220 *T)
    sumY3  +=       2.670 * cos(5.82961294252  + 230.95140066080 *T)
    sumY3  +=       1.974 * cos(1.72688011717  + 1059.62574767270 *T)
    sumY3  +=       2.440 * cos(3.86794399442  + 106.03035047280 *T)
    sumY3  +=       2.032 * cos(3.65372244155  + 1375.52998236229 *T)
    sumY3  +=       2.151 * cos(6.26903391366  + 824.49837626530 *T)
    sumY3  +=       1.875 * cos(6.21933237642  + 344.94686279140 *T)
    sumY3  +=       2.440 * cos(2.94584507929  + 323.26159917390 *T)
    sumY3  +=       2.034 * cos(0.93720484486  + 927.59115048390 *T)
    sumY3  +=       2.310 * cos(0.13408163378  + 99.66756299740 *T)
    sumY3  +=       1.764 * cos(0.29031425745  + 191.45151239371 *T)
    sumY3  +=       2.032 * cos(2.17108818008  + 636.95971005981 *T)
    sumY3  +=       1.993 * cos(0.70131684029  + 1258.21011414209 *T)
    sumY3  +=       1.921 * cos(5.87131438069  + 2641.09746098870 *T)
    sumY3  +=       1.620 * cos(4.33606938312  + 1596.43025976811 *T)
    sumY3  +=       1.903 * cos(1.58126923147  + 2634.47154895491 *T)
    sumY3  +=       1.568 * cos(1.46883265025  + 235.63431344930 *T)
    sumY3  +=       1.835 * cos(2.03918042522  + 536.56069461190 *T)
    sumY3  +=       1.523 * cos(6.05102540851  + 956.04533848709 *T)
    sumY3  +=       2.086 * cos(0.06751029490  + 445.00125562421 *T)
    sumY3  +=       1.666 * cos(3.27204219166  + 1876.76675945369 *T)
    sumY3  +=       1.479 * cos(1.20460352519  + 6290.43321447570 *T)
    sumY3  +=       1.479 * cos(0.55429461283  + 5849.12029463110 *T)
    sumY3  +=       1.439 * cos(1.65531266077  + 2538.49232173710 *T)
    sumY3  +=       1.464 * cos(5.75084658705  + 5863.34738863270 *T)
    sumY3  +=       1.464 * cos(2.29123683992  + 6276.20612047410 *T)
    sumY3  +=       1.676 * cos(0.63508426553  + 1251.09656714130 *T)
    sumY3  +=       1.945 * cos(5.05629111259  + 845.83901726770 *T)
    sumY3  +=       1.436 * cos(4.66419938107  + 508.59414157570 *T)
    sumY3  +=       1.481 * cos(2.92781569338  + 284.39235822571 *T)
    sumY3  +=       1.709 * cos(1.78635026289  + 2318.07967929830 *T)
    sumY3  +=       1.305 * cos(2.53739213998  + 107.26874496520 *T)
    sumY3  +=       1.548 * cos(1.24859563362  + 2943.74987161071 *T)
    sumY3  +=       1.465 * cos(1.19825645294  + 2531.37877473630 *T)
    sumY3  +=       1.494 * cos(5.51365686610  + 1780.78753223589 *T)
    sumY3  +=       1.243 * cos(3.96240186696  + 218.20877936750 *T)
    sumY3  +=       1.234 * cos(0.91112106896  + 208.87704647550 *T)
    sumY3  +=       1.359 * cos(0.27102919236  + 142.69346761730 *T)
    sumY3  +=       1.217 * cos(3.60020530728  + 1052.51220067191 *T)
    sumY3  +=       1.352 * cos(1.92480185850  + 209.87388421679 *T)
    sumY3  +=       1.391 * cos(0.65264575339  + 1243.98302014049 *T)
    sumY3  +=       1.426 * cos(2.11026020785  + 852.95256426850 *T)
    sumY3  +=       1.501 * cos(3.67811053153  + 1691.92185201890 *T)
    sumY3  +=       1.326 * cos(4.70937835849  + 2751.79141717511 *T)
    sumY3  +=       1.171 * cos(6.04141012834  + 67.91186905000 *T)
    sumY3  +=       1.462 * cos(6.22305747248  + 160.85271488200 *T)
    sumY3  +=       1.164 * cos(6.20382255072  + 117.07605073670 *T)
    sumY3  +=       1.337 * cos(1.54820693649  + 2627.35800195410 *T)
    sumY3  +=       1.509 * cos(0.21806196966  + 1574.60198379870 *T)
    sumY3  +=       1.213 * cos(2.74422462860  + 1354.67697632690 *T)
    sumY3  +=       1.241 * cos(5.97315240097  + 1140.89024592190 *T)
    sumY3  +=       1.288 * cos(1.29947866802  + 2950.86341861150 *T)
    sumY3  +=       1.337 * cos(2.21013619697  + 2421.17245351690 *T)
    sumY3  +=       1.327 * cos(0.27411407224  + 1567.48843679789 *T)
    sumY3  +=       1.118 * cos(4.60785389063  + 209.12312469139 *T)
    sumY3  +=       1.043 * cos(2.37818871541  + 156.92056161890 *T)
    sumY3  +=       1.069 * cos(4.66047916687  + 194.89603068979 *T)
    sumY3  +=       1.088 * cos(0.96796460237  + 1272.92484311070 *T)
    sumY3  +=       1.270 * cos(5.73542325903  + 2516.66404576769 *T)
    sumY3  +=       1.032 * cos(5.01309781815  + 1361.30288836070 *T)
    sumY3  +=       1.009 * cos(0.15463134354  + 52.44638055600 *T)
    sumY3  +=       1.100 * cos(2.45485733183  + 1671.06884598351 *T)
    sumY3  +=       1.071 * cos(4.31734498299  + 1354.18934135989 *T)
    sumY3  +=       1.120 * cos(4.50723064884  + 1037.79747170330 *T)
    sumY3  +=       1.271 * cos(3.02683993003  + 721.40560204669 *T)
    sumY3  +=       1.009 * cos(4.14782207001  + 1670.58121101650 *T)
    sumY3  +=       1.145 * cos(0.18866100481  + 3157.04896704871 *T)
    sumY3  +=       1.072 * cos(4.98720300139  + 2435.39954751850 *T)
    sumY3  +=       1.044 * cos(0.37186765557  + 1560.37488979710 *T)
    sumY3  +=       1.195 * cos(4.23012320818  + 632.53992182970 *T)
    sumY3  +=       1.043 * cos(4.68469356625  + 2729.96314120569 *T)
    sumY3  +=       0.889 * cos(5.13451420872  + 359.17395679300 *T)
    sumY3  +=       0.881 * cos(4.07324193143  + 1148.49142788971 *T)
    sumY3  +=       1.040 * cos(5.49603658299  + 1773.67398523510 *T)
    sumY3  +=       1.118 * cos(6.27960853334  + 1890.99385345530 *T)
    sumY3  +=       1.047 * cos(0.97265004380  + 3260.14174126730 *T)
    sumY3  +=       1.087 * cos(5.17936535207  + 2001.20017467470 *T)
    sumY3  +=       1.087 * cos(2.03777269848  + 2001.68780964171 *T)
    sumY3  +=       0.974 * cos(2.14268232218  + 1987.46071564010 *T)
    sumY3  +=       0.869 * cos(0.17251420225  + 621.98185653281 *T)
    sumY3  +=       1.085 * cos(1.32838349016  + 2214.98690507971 *T)
    sumY3  +=       0.841 * cos(3.35616685398  + 629.09540353360 *T)
    sumY3  +=       0.875 * cos(1.07922697964  + 2847.77064439291 *T)
    sumY3  +=       1.063 * cos(5.95949690738  + 2200.27217611110 *T)
    sumY3  +=       0.853 * cos(3.67902900562  + 288.32451148881 *T)
    sumY3  +=       1.056 * cos(0.19418111722  + 2332.30677329990 *T)
    sumY3  +=       0.888 * cos(0.32179471274  + 1155.11733992350 *T)
    sumY3  +=       1.045 * cos(0.14482765695  + 2.93757625420 *T)
    sumY3  +=       0.793 * cos(3.26245616640  + 341.01470952831 *T)
    sumY3  +=       0.817 * cos(0.67889320689  + 1802.12817323829 *T)
    sumY3  +=       0.780 * cos(0.38946586865  + 1382.64352936309 *T)
    sumY3  +=       0.798 * cos(3.07896109863  + 2015.42726867629 *T)
    sumY3  +=       0.798 * cos(3.07896109863  + 2015.91490364330 *T)
    sumY3  +=       0.834 * cos(2.33901429584  + 1802.61580820530 *T)
    sumY3  +=       0.766 * cos(0.30242307365  + 1471.50920958009 *T)
    sumY3  +=       0.875 * cos(5.85387427791  + 2523.77759276850 *T)
    sumY3  +=       0.829 * cos(4.44385893140  + 1162.23088692429 *T)
    sumY3  +=       0.749 * cos(2.07951145157  + 1809.72935520611 *T)
    sumY3  +=       0.942 * cos(3.40636938944  + 2008.31372167550 *T)
    sumY3  +=       0.730 * cos(4.06050820225  + 138.76131435421 *T)
    sumY3  +=       0.734 * cos(0.83381213045  + 749.85979004990 *T)
    sumY3  +=       0.727 * cos(0.81605653091  + 181.29958400710 *T)
    sumY3  +=       0.727 * cos(0.93197698209  + 245.78624183590 *T)
    sumY3  +=       0.986 * cos(4.19789524983  + 1795.01462623750 *T)
    sumY3  +=       0.914 * cos(1.81845398539  + 266.23311096100 *T)
    sumY3  +=       0.774 * cos(1.85314072632  + 2840.16946242510 *T)
    sumY3  +=       0.774 * cos(1.85314072632  + 2840.65709739210 *T)
    sumY3  +=       0.806 * cos(2.87882505331  + 1912.82212942470 *T)
    sumY3  +=       0.828 * cos(0.03130512468  + 100.15519796440 *T)
    sumY3  +=       0.700 * cos(1.98220269537  + 9779.32999085069 *T)
    sumY3  +=       0.700 * cos(2.62359991037  + 10220.64291069530 *T)
    sumY3  +=       0.687 * cos(6.00740806604  + 102.84895673509 *T)
    sumY3  +=       0.801 * cos(5.73752172813  + 1457.28211557850 *T)

    # SATURN Y Coord
    sumY4  = 0
    sumY4  +=    7957.923 * cos(4.81531837805  + 220.65645992230 *T)
    sumY4  +=    7883.034 * cos(3.09554141083  + 206.42936592071 *T)
    sumY4  +=    1517.028 * cos(4.76816117565  + 227.77000692311 *T)
    sumY4  +=    1235.695 * cos(3.82520659584  + 433.95555536030 *T)
    sumY4  +=    1152.216 * cos(3.67316954399  + 213.54291292150 *T)
    sumY4  +=     879.616 * cos(3.77095350877  + 0.24381748350 *T)
    sumY4  +=     709.628 * cos(3.24840338462  + 199.31581891990 *T)
    sumY4  +=     602.013 * cos(5.94750261507  + 7.35736448430 *T)
    sumY4  +=     507.367 * cos(3.21712935929  + 426.84200835950 *T)
    sumY4  +=     386.382 * cos(6.01344501364  + 14.47091148511 *T)
    sumY4  +=     338.827 * cos(2.51520356928  + 419.72846135871 *T)
    sumY4  +=     186.029 * cos(3.85605990474  + 441.06910236111 *T)
    sumY4  +=     220.262 * cos(2.50677283832  + 640.14110379750 *T)
    sumY4  +=     173.210 * cos(2.89158751393  + 647.25465079831 *T)
    sumY4  +=      84.974 * cos(4.62115933793  + 234.88355392390 *T)
    sumY4  +=      85.340 * cos(4.39717612141  + 213.05527795450 *T)
    sumY4  +=      44.264 * cos(6.05982256678  + 117.56368570370 *T)
    sumY4  +=      45.849 * cos(5.16536949235  + 6.86972951729 *T)
    sumY4  +=      47.835 * cos(1.64342336148  + 853.44019923550 *T)
    sumY4  +=      41.104 * cos(2.18400968380  + 412.61491435790 *T)
    sumY4  +=      30.735 * cos(5.85312406338  + 21.58445848590 *T)
    sumY4  +=      22.639 * cos(2.96960991387  + 654.36819779910 *T)
    sumY4  +=      21.067 * cos(3.34244726581  + 103.33659170210 *T)
    sumY4  +=      20.230 * cos(0.17279084911  + 110.45013870291 *T)
    sumY4  +=      21.389 * cos(2.53635561163  + 205.94173095370 *T)
    sumY4  +=      16.854 * cos(4.38021440482  + 124.67723270450 *T)
    sumY4  +=      21.739 * cos(1.99698913462  + 860.55374623631 *T)
    sumY4  +=      17.501 * cos(4.23662156145  + 220.16882495529 *T)
    sumY4  +=      14.506 * cos(4.39894976150  + 88.62186273350 *T)
    sumY4  +=      17.401 * cos(2.26232682838  + 198.82818395289 *T)
    sumY4  +=      13.517 * cos(3.33447998001  + 309.52214013930 *T)
    sumY4  +=      11.814 * cos(5.16076867097  + 323.74923414091 *T)
    sumY4  +=      12.071 * cos(3.74132731532  + 448.18264936190 *T)
    sumY4  +=      11.725 * cos(4.75596558897  + 330.86278114170 *T)
    sumY4  +=      10.174 * cos(4.42772116867  + 95.73540973430 *T)
    sumY4  +=      11.491 * cos(3.65800261495  + 192.20227191910 *T)
    sumY4  +=      10.611 * cos(4.07220720960  + 302.40859313851 *T)
    sumY4  +=      12.147 * cos(2.66285969856  + 633.02755679670 *T)
    sumY4  +=       9.067 * cos(4.26864930915  + 13.98327651810 *T)
    sumY4  +=       8.918 * cos(3.03990460360  + 426.35437339250 *T)
    sumY4  +=       8.311 * cos(3.10463384858  + 430.02340209721 *T)
    sumY4  +=       7.296 * cos(0.77854151610  + 316.63568714010 *T)
    sumY4  +=       7.890 * cos(0.94138641674  + 1066.73929467350 *T)
    sumY4  +=       6.699 * cos(3.06921349305  + 109.96250373590 *T)
    sumY4  +=       6.339 * cos(3.75710669061  + 216.72430665921 *T)
    sumY4  +=       6.254 * cos(1.12954162648  + 210.36151918381 *T)
    sumY4  +=       6.126 * cos(3.21971824010  + 301.92095817150 *T)
    sumY4  +=       6.725 * cos(3.94118928489  + 625.91400979590 *T)
    sumY4  +=       6.784 * cos(4.21865919120  + 223.83785366000 *T)
    sumY4  +=       6.868 * cos(0.65717842769  + 203.24797218300 *T)
    sumY4  +=       6.111 * cos(3.59045265634  + 96.22304470130 *T)
    sumY4  +=       4.468 * cos(3.29229132687  + 309.03450517230 *T)
    sumY4  +=       4.160 * cos(5.74668718431  + 522.82123557730 *T)
    sumY4  +=       3.681 * cos(5.66131908910  + 423.66061462181 *T)
    sumY4  +=       3.615 * cos(5.12190926865  + 405.50136735710 *T)
    sumY4  +=       3.079 * cos(0.92409053273  + 412.12727939089 *T)
    sumY4  +=       3.614 * cos(1.94380765622  + 515.22005360950 *T)
    sumY4  +=       2.896 * cos(6.00160618678  + 515.70768857651 *T)
    sumY4  +=       3.000 * cos(3.15709072507  + 422.90985509640 *T)
    sumY4  +=       2.685 * cos(3.43239500452  + 209.61075965840 *T)
    sumY4  +=       2.918 * cos(5.44500673473  + 89.10949770051 *T)
    sumY4  +=       2.617 * cos(1.40935172097  + 217.47506618460 *T)
    sumY4  +=       2.884 * cos(2.20199681607  + 191.71463695210 *T)
    sumY4  +=       2.281 * cos(2.19064979800  + 316.14805217309 *T)
    sumY4  +=       2.717 * cos(3.78847562141  + 131.79077970531 *T)
    sumY4  +=       2.556 * cos(3.08407888684  + 337.97632814250 *T)
    sumY4  +=       1.782 * cos(4.51078296370  + 241.99710092470 *T)
    sumY4  +=       1.782 * cos(0.37884320289  + 185.08872491831 *T)
    sumY4  +=       1.856 * cos(2.79299417712  + 618.31282782810 *T)
    sumY4  +=       1.713 * cos(2.96942684848  + 625.42637482889 *T)
    sumY4  +=       1.632 * cos(0.86290354268  + 728.51914904750 *T)
    sumY4  +=       1.980 * cos(1.69424503948  + 831.61192326610 *T)
    sumY4  +=       1.607 * cos(6.10863313121  + 611.19928082730 *T)
    sumY4  +=       1.801 * cos(4.27647079459  + 227.28237195610 *T)
    sumY4  +=       1.532 * cos(5.31288615169  + 1052.51220067191 *T)
    sumY4  +=       1.510 * cos(2.58764647681  + 742.74624304909 *T)
    sumY4  +=       1.531 * cos(0.19788540477  + 529.44714761109 *T)
    sumY4  +=       1.394 * cos(3.60010141007  + 1361.30288836070 *T)
    sumY4  +=       1.427 * cos(2.30374859738  + 1251.09656714130 *T)
    sumY4  +=       1.649 * cos(4.25957274495  + 295.29504613770 *T)
    sumY4  +=       1.249 * cos(3.74724649234  + 743.23387801611 *T)
    sumY4  +=       1.353 * cos(4.07456235845  + 1678.18239298430 *T)
    sumY4  +=       1.674 * cos(5.46467832773  + 231.70216018621 *T)
    sumY4  +=       1.646 * cos(5.68811055040  + 195.38366565681 *T)
    sumY4  +=       1.305 * cos(2.57485726113  + 934.70469748469 *T)
    sumY4  +=       1.160 * cos(1.35876987453  + 81.50831573269 *T)
    sumY4  +=       1.537 * cos(1.64895812279  + 21.09682351890 *T)
    sumY4  +=       1.096 * cos(5.87040605326  + 832.09955823310 *T)
    sumY4  +=       1.156 * cos(1.65604017773  + 508.10650660869 *T)
    sumY4  +=       1.062 * cos(3.29348541593  + 294.80741117069 *T)
    sumY4  +=       1.285 * cos(1.15881412669  + 1464.39566257930 *T)
    sumY4  +=       1.079 * cos(4.46936132008  + 1368.90407032850 *T)
    sumY4  +=       1.006 * cos(4.83365073160  + 1361.79052332771 *T)
    sumY4  +=       1.290 * cos(4.69044132255  + 736.12033101530 *T)
    sumY4  +=       1.051 * cos(2.68982495323  + 941.81824448550 *T)
    sumY4  +=       1.027 * cos(3.70327166201  + 433.46792039329 *T)
    sumY4  +=       1.021 * cos(2.48350861942  + 522.33360061030 *T)
    sumY4  +=       1.137 * cos(2.90811550586  + 1891.48148842230 *T)
    sumY4  +=       0.991 * cos(3.78009400759  + 1994.57426264090 *T)
    sumY4  +=       1.038 * cos(1.35464465726  + 1148.00379292269 *T)
    sumY4  +=       0.865 * cos(0.40802396366  + 750.34742501690 *T)
    sumY4  +=       0.859 * cos(5.85730307132  + 416.54706762100 *T)
    sumY4  +=       0.862 * cos(3.18022940336  + 437.13694909800 *T)
    sumY4  +=       0.776 * cos(1.33793489456  + 102.84895673509 *T)
    sumY4  +=       0.797 * cos(2.04691519292  + 1375.52998236229 *T)
    sumY4  +=       0.824 * cos(4.53645127236  + 639.65346883050 *T)
    sumY4  +=       0.881 * cos(3.45735740140  + 1258.69774910910 *T)
    sumY4  +=       0.686 * cos(1.77528970097  + 234.39591895690 *T)
    sumY4  +=       0.684 * cos(3.52622623805  + 661.48174479990 *T)
    sumY4  +=       0.859 * cos(3.63825836835  + 537.04832957890 *T)

    # SATURN Y Coord
    sumY5  = 0
    sumY5  +=     605.503 * cos(3.02566952553  + 220.65645992230 *T)
    sumY5  +=     576.439 * cos(4.77676022633  + 206.42936592071 *T)
    sumY5  +=     227.348 * cos(3.02005484310  + 227.77000692311 *T)
    sumY5  +=     133.503 * cos(2.01762000448  + 433.95555536030 *T)
    sumY5  +=      94.765 * cos(4.98454736271  + 199.31581891990 *T)
    sumY5  +=      39.851 * cos(4.29556357711  + 14.47091148511 *T)
    sumY5  +=      34.190 * cos(2.07372397023  + 441.06910236111 *T)
    sumY5  +=      24.941 * cos(3.84014099419  + 0.24381748350 *T)
    sumY5  +=      23.155 * cos(1.07127433478  + 647.25465079831 *T)
    sumY5  +=      16.208 * cos(0.82050987845  + 426.84200835950 *T)
    sumY5  +=      12.333 * cos(0.61142473697  + 640.14110379750 *T)
    sumY5  +=      10.358 * cos(2.92988275487  + 234.88355392390 *T)
    sumY5  +=      10.534 * cos(5.12660829509  + 419.72846135871 *T)
    sumY5  +=       8.668 * cos(1.94893255744  + 192.20227191910 *T)
    sumY5  +=       8.840 * cos(3.98137124653  + 213.54291292150 *T)
    sumY5  +=       4.707 * cos(6.14618174587  + 88.62186273350 *T)
    sumY5  +=       5.904 * cos(3.16028075613  + 7.35736448430 *T)
    sumY5  +=       3.895 * cos(0.73151574345  + 405.50136735710 *T)
    sumY5  +=       3.976 * cos(3.54063716264  + 6.86972951729 *T)
    sumY5  +=       3.675 * cos(2.04382079343  + 213.05527795450 *T)
    sumY5  +=       3.472 * cos(6.06844015557  + 853.44019923550 *T)
    sumY5  +=       4.393 * cos(1.18474783269  + 654.36819779910 *T)
    sumY5  +=       3.290 * cos(2.31998889526  + 124.67723270450 *T)
    sumY5  +=       2.569 * cos(1.20332477401  + 109.96250373590 *T)
    sumY5  +=       2.467 * cos(1.58903935784  + 515.70768857651 *T)
    sumY5  +=       2.708 * cos(0.15976971982  + 860.55374623631 *T)
    sumY5  +=       2.138 * cos(2.01599255393  + 448.18264936190 *T)
    sumY5  +=       2.447 * cos(2.47106712193  + 220.16882495529 *T)
    sumY5  +=       1.979 * cos(5.69757911363  + 302.40859313851 *T)
    sumY5  +=       2.238 * cos(0.86043138018  + 198.82818395289 *T)
    sumY5  +=       1.727 * cos(1.65089858299  + 301.92095817150 *T)
    sumY5  +=       1.398 * cos(0.05843587724  + 309.52214013930 *T)
    sumY5  +=       1.536 * cos(3.94854505604  + 205.94173095370 *T)
    sumY5  +=       1.195 * cos(3.93835889247  + 117.56368570370 *T)
    sumY5  +=       1.000 * cos(2.44450069185  + 223.83785366000 *T)
    sumY5  +=       0.984 * cos(2.38499591413  + 203.24797218300 *T)
    sumY5  +=       0.791 * cos(0.61482230222  + 330.86278114170 *T)
    sumY5  +=       0.979 * cos(0.49750454847  + 95.73540973430 *T)
    sumY5  +=       0.784 * cos(1.61990559836  + 522.82123557730 *T)
    sumY5  +=       0.910 * cos(0.48207868964  + 96.22304470130 *T)

    # SATURN Z Coord
    sumZ0  = 0
    sumZ0  +=  41356950.940 * cos(3.60234142982  + 213.29909543800 *T)
    sumZ0  +=  1148283.576 * cos(2.85128367469  + 426.59819087600 *T)
    sumZ0  +=  1214249.867 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ0  +=  329280.791 * cos(0.57121407104  + 206.18554843720 *T)
    sumZ0  +=  286934.048 * cos(3.48073526693  + 220.41264243880 *T)
    sumZ0  +=   99076.584 * cos(4.73369511264  + 7.11354700080 *T)
    sumZ0  +=   57361.820 * cos(4.92611225093  + 110.20632121940 *T)
    sumZ0  +=   47738.127 * cos(2.10039779728  + 639.89728631400 *T)
    sumZ0  +=   43458.803 * cos(5.84904978051  + 419.48464387520 *T)
    sumZ0  +=   34565.673 * cos(5.42614229590  + 316.39186965660 *T)
    sumZ0  +=   16185.391 * cos(2.72987173675  + 433.71173787680 *T)
    sumZ0  +=    9001.270 * cos(1.38140102737  + 103.09277421860 *T)
    sumZ0  +=   11433.574 * cos(3.71662021072  + 529.69096509460 *T)
    sumZ0  +=    5398.708 * cos(5.13204892363  + 202.25339517410 *T)
    sumZ0  +=    3902.467 * cos(3.71499738796  + 323.50541665740 *T)
    sumZ0  +=    3709.212 * cos(5.05549348785  + 632.78373931320 *T)
    sumZ0  +=    3614.100 * cos(3.35210451276  + 227.52618943960 *T)
    sumZ0  +=    3379.953 * cos(2.13868919206  + 11.04570026390 *T)
    sumZ0  +=    3089.874 * cos(3.62572857085  + 209.36694217490 *T)
    sumZ0  +=    2683.064 * cos(4.87689555581  + 224.34479570190 *T)
    sumZ0  +=    2963.493 * cos(0.46490184985  + 217.23124870110 *T)
    sumZ0  +=    2343.367 * cos(1.34558278340  + 853.19638175200 *T)
    sumZ0  +=    2423.663 * cos(2.92907094760  + 63.73589830340 *T)
    sumZ0  +=    1701.916 * cos(1.89892525654  + 735.87651353180 *T)
    sumZ0  +=    1941.205 * cos(4.59421314662  + 14.22709400160 *T)
    sumZ0  +=    1990.145 * cos(0.73166053611  + 199.07200143640 *T)
    sumZ0  +=    1460.265 * cos(3.12851339724  + 522.57741809380 *T)
    sumZ0  +=    1148.341 * cos(4.41139213915  + 117.31986822020 *T)
    sumZ0  +=    1092.809 * cos(3.10679381209  + 216.48048917570 *T)
    sumZ0  +=    1015.179 * cos(1.97897195994  + 647.01083331480 *T)
    sumZ0  +=    1098.254 * cos(0.96097709156  + 210.11770170030 *T)
    sumZ0  +=     991.030 * cos(2.99610026682  + 846.08283475120 *T)
    sumZ0  +=    1028.743 * cos(2.11933059243  + 415.55249061210 *T)
    sumZ0  +=     907.817 * cos(4.68576278610  + 309.27832265580 *T)
    sumZ0  +=     818.092 * cos(2.91497656196  + 149.56319713460 *T)
    sumZ0  +=     733.443 * cos(2.10018715614  + 74.78159856730 *T)
    sumZ0  +=     631.275 * cos(1.30557255814  + 277.03499374140 *T)
    sumZ0  +=     574.266 * cos(0.63353925382  + 490.33408917940 *T)
    sumZ0  +=     471.645 * cos(1.62329782277  + 1052.26838318840 *T)
    sumZ0  +=     416.292 * cos(5.88171380023  + 95.97922721780 *T)
    sumZ0  +=     412.467 * cos(2.56322277221  + 1162.47470440780 *T)
    sumZ0  +=     381.934 * cos(2.83850532211  + 838.96928775040 *T)
    sumZ0  +=     370.902 * cos(2.22356953874  + 351.81659230870 *T)
    sumZ0  +=     337.338 * cos(4.12487690436  + 3.93215326310 *T)
    sumZ0  +=     263.173 * cos(2.60594324261  + 440.82528487760 *T)
    sumZ0  +=     245.414 * cos(3.87716027424  + 1059.38193018920 *T)
    sumZ0  +=     231.186 * cos(2.79137518143  + 127.47179660680 *T)
    sumZ0  +=     232.835 * cos(1.69429632635  + 38.13303563780 *T)
    sumZ0  +=     226.218 * cos(0.98981012262  + 210.85141488320 *T)
    sumZ0  +=     252.287 * cos(3.04205345262  + 137.03302416240 *T)
    sumZ0  +=     213.656 * cos(3.00046819587  + 536.80451209540 *T)
    sumZ0  +=     223.924 * cos(3.07498346391  + 215.74677599280 *T)
    sumZ0  +=     189.917 * cos(5.75255233102  + 742.99006053260 *T)
    sumZ0  +=     162.873 * cos(3.89762690585  + 214.26230328450 *T)
    sumZ0  +=     164.510 * cos(2.45697452905  + 422.66603761290 *T)
    sumZ0  +=     158.304 * cos(2.31306407231  + 1478.86657406440 *T)
    sumZ0  +=     156.454 * cos(1.35680198123  + 1368.66025284500 *T)
    sumZ0  +=     202.699 * cos(3.85581899880  + 949.17560896980 *T)
    sumZ0  +=     150.079 * cos(4.14045329244  + 437.64389113990 *T)
    sumZ0  +=     142.705 * cos(2.58383947868  + 1155.36115740700 *T)
    sumZ0  +=     163.551 * cos(0.16578976592  + 212.33588759150 *T)
    sumZ0  +=     135.335 * cos(0.48883402314  + 213.34727954780 *T)
    sumZ0  +=     129.140 * cos(2.17035714121  + 76.26607127560 *T)
    sumZ0  +=     128.206 * cos(0.59415255592  + 1066.49547719000 *T)
    sumZ0  +=     140.534 * cos(3.97289861748  + 625.67019231240 *T)
    sumZ0  +=     121.564 * cos(4.51381049808  + 3.18139373770 *T)
    sumZ0  +=     119.190 * cos(1.61391416177  + 628.85158605010 *T)
    sumZ0  +=     115.147 * cos(3.43459040245  + 330.61896365820 *T)
    sumZ0  +=     135.375 * cos(3.57424025701  + 213.25091132820 *T)
    sumZ0  +=     103.727 * cos(6.17639441368  + 200.76892246580 *T)
    sumZ0  +=     107.498 * cos(3.52498564069  + 138.51749687070 *T)
    sumZ0  +=     107.560 * cos(3.99837836470  + 85.82729883120 *T)
    sumZ0  +=     113.808 * cos(0.43922136441  + 222.86032299360 *T)
    sumZ0  +=      94.920 * cos(2.26222536028  + 388.46515523820 *T)
    sumZ0  +=     110.037 * cos(3.51152757685  + 1265.56747862640 *T)
    sumZ0  +=     101.425 * cos(0.14126238472  + 430.53034413910 *T)
    sumZ0  +=     101.928 * cos(0.21047731069  + 412.37109687440 *T)
    sumZ0  +=     104.973 * cos(4.19518394917  + 9.56122755560 *T)
    sumZ0  +=      83.620 * cos(4.82013591775  + 1.48447270830 *T)
    sumZ0  +=      71.104 * cos(4.92741306047  + 212.77783057620 *T)
    sumZ0  +=      70.089 * cos(1.05108025259  + 12.53017297220 *T)
    sumZ0  +=      66.662 * cos(1.22617152895  + 860.30992875280 *T)
    sumZ0  +=      77.048 * cos(0.80511279723  + 52.69019803950 *T)
    sumZ0  +=      90.986 * cos(6.28198480720  + 423.41679713830 *T)
    sumZ0  +=      62.746 * cos(2.06105893244  + 1795.25844372100 *T)
    sumZ0  +=      63.467 * cos(2.36959372262  + 429.77958461370 *T)
    sumZ0  +=      58.999 * cos(2.31882650737  + 1471.75302706360 *T)
    sumZ0  +=      76.181 * cos(2.67435759741  + 191.20769491020 *T)
    sumZ0  +=      63.389 * cos(1.81556177589  + 350.33211960040 *T)
    sumZ0  +=      56.601 * cos(1.11303957705  + 1685.05212250160 *T)
    sumZ0  +=      57.837 * cos(3.29043420072  + 515.46387109300 *T)
    sumZ0  +=      70.959 * cos(5.41979944279  + 213.82036029980 *T)
    sumZ0  +=      70.150 * cos(3.82574373735  + 942.06206196900 *T)
    sumZ0  +=      51.149 * cos(3.86030131690  + 173.94221952280 *T)
    sumZ0  +=      51.666 * cos(3.49111344070  + 203.73786788240 *T)
    sumZ0  +=      50.128 * cos(1.98273693937  + 195.13984817330 *T)
    sumZ0  +=      48.091 * cos(3.20971334548  + 234.63973644040 *T)
    sumZ0  +=      49.761 * cos(6.18808156638  + 703.63318461740 *T)
    sumZ0  +=      46.488 * cos(3.80768518798  + 225.82926841020 *T)
    sumZ0  +=      54.298 * cos(3.18076013146  + 1581.95934828300 *T)
    sumZ0  +=      47.483 * cos(1.53655093485  + 265.98929347750 *T)
    sumZ0  +=      39.354 * cos(2.10537295968  + 728.76296653100 *T)
    sumZ0  +=      39.602 * cos(2.27507746067  + 288.08069400530 *T)
    sumZ0  +=      41.716 * cos(1.05111221485  + 563.63121503840 *T)
    sumZ0  +=      37.614 * cos(6.14179074602  + 217.49188113200 *T)
    sumZ0  +=      42.785 * cos(0.11273484212  + 252.65597135320 *T)
    sumZ0  +=      37.718 * cos(5.96333653993  + 210.37833413120 *T)
    sumZ0  +=      36.792 * cos(1.35292393469  + 565.11568774670 *T)
    sumZ0  +=      37.862 * cos(4.00773869969  + 62.25142559510 *T)
    sumZ0  +=      36.057 * cos(3.30899953753  + 212.54833591260 *T)
    sumZ0  +=      35.460 * cos(5.01105830246  + 65.22037101170 *T)
    sumZ0  +=      34.852 * cos(4.75445504646  + 207.88246946660 *T)
    sumZ0  +=      40.506 * cos(2.20073759144  + 362.86229257260 *T)
    sumZ0  +=      32.560 * cos(6.14806301843  + 417.03696332040 *T)
    sumZ0  +=      37.967 * cos(4.57075294928  + 302.16477565500 *T)
    sumZ0  +=      32.818 * cos(5.79538785685  + 99.16062095550 *T)
    sumZ0  +=      30.744 * cos(3.64317302294  + 207.67002114550 *T)
    sumZ0  +=      32.548 * cos(2.95740474962  + 70.84944530420 *T)
    sumZ0  +=      35.830 * cos(0.75230512173  + 214.04985496340 *T)
    sumZ0  +=      34.253 * cos(2.96835157474  + 223.59403617650 *T)
    sumZ0  +=      32.939 * cos(4.45144712576  + 21.34064100240 *T)
    sumZ0  +=      37.689 * cos(4.37977418287  + 216.21985674480 *T)
    sumZ0  +=      32.421 * cos(5.78559756829  + 160.60889739850 *T)
    sumZ0  +=      28.902 * cos(5.81529149124  + 88.86568021700 *T)
    sumZ0  +=      33.335 * cos(5.62401622695  + 218.71572140940 *T)
    sumZ0  +=      38.185 * cos(4.19997816879  + 209.10630974400 *T)
    sumZ0  +=      27.693 * cos(5.55618950213  + 18.15924726470 *T)
    sumZ0  +=      28.922 * cos(5.58858980738  + 425.11371816770 *T)
    sumZ0  +=      32.330 * cos(4.22794944092  + 22.09140052780 *T)
    sumZ0  +=      29.913 * cos(3.31182307103  + 124.43341522100 *T)
    sumZ0  +=      29.455 * cos(5.83042047243  + 956.28915597060 *T)
    sumZ0  +=      26.648 * cos(0.86666603224  + 217.96496188400 *T)
    sumZ0  +=      25.659 * cos(5.10869894674  + 479.28838891550 *T)
    sumZ0  +=      25.652 * cos(4.80215156054  + 231.45834270270 *T)
    sumZ0  +=      26.011 * cos(2.05589181933  + 1788.14489672020 *T)
    sumZ0  +=      25.663 * cos(1.81564756635  + 2111.65031337760 *T)
    sumZ0  +=      22.977 * cos(5.44037839738  + 142.44965013380 *T)
    sumZ0  +=      31.283 * cos(3.54881062563  + 1258.45393162560 *T)
    sumZ0  +=      30.192 * cos(0.42929684303  + 218.92816973050 *T)
    sumZ0  +=      26.495 * cos(2.86288115466  + 1898.35121793960 *T)
    sumZ0  +=      29.523 * cos(0.37961341545  + 554.06998748280 *T)
    sumZ0  +=      22.228 * cos(5.48142509504  + 408.43894361130 *T)
    sumZ0  +=      21.237 * cos(0.87139429027  + 2001.44399215820 *T)
    sumZ0  +=      20.174 * cos(1.85380647370  + 654.12438031560 *T)
    sumZ0  +=      20.521 * cos(6.16117414568  + 56.62235130260 *T)
    sumZ0  +=      23.744 * cos(3.20046265011  + 208.63322899200 *T)
    sumZ0  +=      21.685 * cos(3.17871482760  + 414.06801790380 *T)
    sumZ0  +=      19.856 * cos(3.13857572740  + 175.16605980020 *T)
    sumZ0  +=      17.347 * cos(5.96946168051  + 213.18722085340 *T)
    sumZ0  +=      17.925 * cos(1.14556340660  + 203.00415469950 *T)
    sumZ0  +=      18.247 * cos(3.00127995342  + 831.85574074960 *T)
    sumZ0  +=      17.378 * cos(0.74127655202  + 177.87437278590 *T)
    sumZ0  +=      16.401 * cos(1.36402971878  + 211.81462272970 *T)
    sumZ0  +=      16.228 * cos(3.17582194781  + 98.89998852460 *T)
    sumZ0  +=      16.107 * cos(0.54971362869  + 251.43213107580 *T)
    sumZ0  +=      15.546 * cos(2.29108586908  + 750.10360753340 *T)
    sumZ0  +=      14.801 * cos(0.20111341190  + 148.07872442630 *T)
    sumZ0  +=      15.320 * cos(3.80296373340  + 483.22054217860 *T)
    sumZ0  +=      14.467 * cos(1.32549449124  + 114.13847448250 *T)
    sumZ0  +=      14.146 * cos(4.85276668809  + 327.43756992050 *T)
    sumZ0  +=      15.436 * cos(1.48161033604  + 198.32124191100 *T)
    sumZ0  +=      14.576 * cos(3.75966213667  + 213.51154375910 *T)
    sumZ0  +=      14.820 * cos(3.34957428396  + 248.72381809010 *T)
    sumZ0  +=      13.743 * cos(4.77276011938  + 106.27416795630 *T)
    sumZ0  +=      17.347 * cos(4.37930417897  + 213.41097002260 *T)
    sumZ0  +=      14.576 * cos(0.30591841563  + 213.08664711690 *T)
    sumZ0  +=      13.755 * cos(2.59733474149  + 228.27694896500 *T)
    sumZ0  +=      12.945 * cos(1.84552645547  + 1045.15483618760 *T)
    sumZ0  +=      12.834 * cos(4.24597463890  + 10.29494073850 *T)
    sumZ0  +=      12.885 * cos(4.86095700310  + 39.35687591520 *T)
    sumZ0  +=      16.969 * cos(1.27721722241  + 214.78356814630 *T)
    sumZ0  +=      12.706 * cos(0.31024660308  + 424.15051032120 *T)
    sumZ0  +=      12.558 * cos(0.03128357867  + 1272.68102562720 *T)
    sumZ0  +=      14.310 * cos(2.34661407283  + 1375.77379984580 *T)
    sumZ0  +=      11.808 * cos(1.79617614059  + 2104.53676637680 *T)
    sumZ0  +=      13.258 * cos(2.56724178887  + 2214.74308759620 *T)
    sumZ0  +=      11.158 * cos(0.45949551763  + 497.44763618020 *T)
    sumZ0  +=      14.798 * cos(1.03056978581  + 2.44768055480 *T)
    sumZ0  +=      14.861 * cos(3.26343762010  + 1574.84580128220 *T)
    sumZ0  +=      11.978 * cos(6.22486801109  + 121.25202148330 *T)
    sumZ0  +=      10.959 * cos(1.31570222417  + 235.39049596580 *T)
    sumZ0  +=      11.901 * cos(1.58846496072  + 635.96513305090 *T)
    sumZ0  +=      11.080 * cos(5.37041305935  + 1169.58825140860 *T)
    sumZ0  +=      10.475 * cos(2.19368288733  + 221.37585028530 *T)
    sumZ0  +=      12.129 * cos(4.63783219973  + 269.92144674060 *T)
    sumZ0  +=      10.959 * cos(4.40512358221  + 213.55972786890 *T)
    sumZ0  +=      10.673 * cos(1.60141035946  + 2428.04218303420 *T)
    sumZ0  +=      13.041 * cos(1.87126265057  + 151.04766984290 *T)
    sumZ0  +=      11.521 * cos(2.11579255167  + 205.22234059070 *T)
    sumZ0  +=      10.228 * cos(2.41342516828  + 275.55052103310 *T)
    sumZ0  +=      10.229 * cos(0.96548776504  + 842.15068148810 *T)
    sumZ0  +=      10.959 * cos(5.94363969204  + 213.03846300710 *T)
    sumZ0  +=       9.406 * cos(3.39740548756  + 650.94298657790 *T)
    sumZ0  +=      10.284 * cos(0.01930234298  + 219.44943459230 *T)
    sumZ0  +=       9.424 * cos(1.68112701238  + 488.84961647110 *T)
    sumZ0  +=       8.939 * cos(5.28697277625  + 340.77089204480 *T)
    sumZ0  +=       9.017 * cos(2.70864683669  + 1692.16566950240 *T)
    sumZ0  +=       9.979 * cos(3.97256303244  + 278.51946644970 *T)
    sumZ0  +=       8.356 * cos(0.65143790392  + 2317.83586181480 *T)
    sumZ0  +=       9.107 * cos(0.33774330802  + 10213.28554621100 *T)
    sumZ0  +=       8.169 * cos(3.23957036879  + 427.56139872250 *T)
    sumZ0  +=       9.044 * cos(0.23054654773  + 312.19908396260 *T)
    sumZ0  +=       7.927 * cos(3.97539997718  + 210.59078245230 *T)
    sumZ0  +=      10.063 * cos(4.21498692120  + 618.55664531160 *T)
    sumZ0  +=       9.335 * cos(5.70779396792  + 425.63498302950 *T)
    sumZ0  +=       8.385 * cos(0.64196694648  + 188.92007304980 *T)
    sumZ0  +=       8.670 * cos(2.75601079050  + 491.81856188770 *T)
    sumZ0  +=       8.100 * cos(4.06039030284  + 207.14875628370 *T)
    sumZ0  +=       9.452 * cos(2.59277908129  + 429.04587143080 *T)
    sumZ0  +=       7.738 * cos(1.55361395079  + 601.76425067620 *T)
    sumZ0  +=       7.577 * cos(6.02114380426  + 426.64637498580 *T)
    sumZ0  +=       9.275 * cos(5.16642875471  + 8.07675484730 *T)
    sumZ0  +=       7.638 * cos(2.70476229552  + 543.91805909620 *T)
    sumZ0  +=       7.663 * cos(2.73399784358  + 1148.24761040620 *T)
    sumZ0  +=       7.158 * cos(5.84039229615  + 84.34282612290 *T)
    sumZ0  +=       8.033 * cos(4.20065885029  + 767.36908292080 *T)
    sumZ0  +=       6.822 * cos(6.12047591103  + 1279.79457262800 *T)
    sumZ0  +=       8.887 * cos(4.87842237223  + 1589.07289528380 *T)
    sumZ0  +=       6.961 * cos(3.34826318442  + 237.67811782620 *T)
    sumZ0  +=       6.580 * cos(5.81874339759  + 284.14854074220 *T)
    sumZ0  +=       6.629 * cos(1.53732206945  + 289.56516671360 *T)
    sumZ0  +=       6.283 * cos(3.84796221759  + 125.98732389850 *T)
    sumZ0  +=       7.796 * cos(0.08813581424  + 216.00740842370 *T)
    sumZ0  +=       8.282 * cos(1.90022750359  + 134.58534360760 *T)
    sumZ0  +=       6.576 * cos(0.35050721605  + 220.46082654860 *T)
    sumZ0  +=       6.587 * cos(0.58291546593  + 213.45915413240 *T)
    sumZ0  +=       6.378 * cos(6.14309601319  + 73.29712585900 *T)
    sumZ0  +=       6.626 * cos(3.37506241952  + 312.45971639350 *T)
    sumZ0  +=       6.587 * cos(3.48062808501  + 213.13903674360 *T)
    sumZ0  +=       7.576 * cos(2.82334902636  + 426.55000676620 *T)
    sumZ0  +=       5.988 * cos(5.47551169354  + 35.42472265210 *T)
    sumZ0  +=       5.723 * cos(5.22938546212  + 212.02707105080 *T)
    sumZ0  +=       6.565 * cos(5.18423527980  + 67.66805156650 *T)
    sumZ0  +=       5.704 * cos(5.31164037710  + 1485.98012106520 *T)
    sumZ0  +=       6.842 * cos(6.15429926594  + 179.35884549420 *T)
    sumZ0  +=       7.349 * cos(0.73079523905  + 355.74874557180 *T)
    sumZ0  +=       5.590 * cos(4.19082216094  + 247.23934538180 *T)
    sumZ0  +=       6.488 * cos(2.29038019982  + 2531.13495725280 *T)
    sumZ0  +=       7.413 * cos(2.98595847215  + 1891.23767093880 *T)
    sumZ0  +=       5.491 * cos(2.70527999156  + 2008.55753915900 *T)
    sumZ0  +=       5.820 * cos(2.98199777797  + 5.41662597140 *T)
    sumZ0  +=       5.622 * cos(1.45901882320  + 501.37978944330 *T)
    sumZ0  +=       5.473 * cos(0.81641753389  + 181.05576652360 *T)
    sumZ0  +=       5.454 * cos(6.12211624341  + 643.82943957710 *T)
    sumZ0  +=       5.407 * cos(1.53066536154  + 2420.92863603340 *T)
    sumZ0  +=       5.442 * cos(0.02368001914  + 404.50679034820 *T)
    sumZ0  +=       5.436 * cos(4.67977840196  + 621.73803904930 *T)
    sumZ0  +=       6.267 * cos(5.43877709236  + 636.71589257630 *T)
    sumZ0  +=       5.068 * cos(4.33978904750  + 113.38771495710 *T)
    sumZ0  +=       4.865 * cos(5.10809843245  + 69.15252427480 *T)
    sumZ0  +=       5.070 * cos(3.27536823314  + 144.14657116320 *T)
    sumZ0  +=       4.738 * cos(0.53669462342  + 1073.60902419080 *T)
    sumZ0  +=       5.183 * cos(2.18841260581  + 107.02492748170 *T)
    sumZ0  +=       5.310 * cos(0.66732715963  + 116.42609634290 *T)
    sumZ0  +=       5.141 * cos(2.42998599018  + 692.58748435350 *T)
    sumZ0  +=       6.231 * cos(5.93759865879  + 436.15941843160 *T)
    sumZ0  +=       6.367 * cos(3.45785496152  + 220.36445832900 *T)
    sumZ0  +=       4.541 * cos(0.61777794179  + 206.23373254700 *T)
    sumZ0  +=       4.528 * cos(1.65756596310  + 1361.54670584420 *T)
    sumZ0  +=       5.172 * cos(4.96799695932  + 342.25536475310 *T)
    sumZ0  +=       5.805 * cos(2.10740836405  + 4.66586644600 *T)
    sumZ0  +=       4.639 * cos(1.63821364676  + 5.62907429250 *T)
    sumZ0  +=       4.731 * cos(1.44641985940  + 2744.43405269080 *T)
    sumZ0  +=       4.756 * cos(3.73287086188  + 206.13736432740 *T)
    sumZ0  +=       4.322 * cos(4.15153138812  + 426.07692601420 *T)
    sumZ0  +=       4.167 * cos(1.65241802351  + 191.95845443560 *T)
    sumZ0  +=       4.164 * cos(2.47912208303  + 447.93883187840 *T)
    sumZ0  +=       5.676 * cos(5.11706007934  + 214.57111982520 *T)
    sumZ0  +=       3.966 * cos(0.05086647098  + 206.70681329900 *T)
    sumZ0  +=       4.603 * cos(5.52562693778  + 78.71375183040 *T)
    sumZ0  +=       3.802 * cos(0.43628874694  + 617.80588578620 *T)
    sumZ0  +=       4.800 * cos(5.54041183899  + 344.70304530790 *T)
    sumZ0  +=       4.743 * cos(0.53760335914  + 128.95626931510 *T)
    sumZ0  +=       4.130 * cos(1.61538042826  + 643.07868005170 *T)
    sumZ0  +=       4.686 * cos(3.97383030786  + 934.94851496820 *T)
    sumZ0  +=       4.043 * cos(2.18819710124  + 358.93013930950 *T)
    sumZ0  +=       3.906 * cos(5.19970623343  + 630.33605875840 *T)
    sumZ0  +=       3.883 * cos(5.39147665291  + 320.32402291970 *T)
    sumZ0  +=       4.913 * cos(1.03288629446  + 32.24332891440 *T)
    sumZ0  +=       3.689 * cos(5.45822055399  + 916.93228005540 *T)
    sumZ0  +=       4.228 * cos(4.28360558940  + 4.19278569400 *T)
    sumZ0  +=       4.020 * cos(4.88073386744  + 54.17467074780 *T)
    sumZ0  +=       4.185 * cos(3.32405342181  + 245.54242435240 *T)
    sumZ0  +=       4.583 * cos(1.85097979187  + 114.39910691340 *T)
    sumZ0  +=       4.596 * cos(4.18504944347  + 219.89137757700 *T)
    sumZ0  +=       3.552 * cos(5.98005739127  + 20.60692781950 *T)
    sumZ0  +=       3.975 * cos(5.80157206371  + 14.97785352700 *T)
    sumZ0  +=       3.475 * cos(2.47178960968  + 1464.63948006280 *T)
    sumZ0  +=       4.395 * cos(0.85418906454  + 525.49817940060 *T)
    sumZ0  +=       3.187 * cos(2.61148353061  + 2324.94940881560 *T)
    sumZ0  +=       3.966 * cos(4.66662911986  + 427.11945573780 *T)
    sumZ0  +=       3.124 * cos(0.10655817305  + 14.01464568050 *T)
    sumZ0  +=       3.104 * cos(0.38849805992  + 2634.22773147140 *T)
    sumZ0  +=       3.508 * cos(2.85265213502  + 87.31177153950 *T)
    sumZ0  +=       2.979 * cos(3.04084830124  + 420.96911658350 *T)
    sumZ0  +=       3.143 * cos(1.71420147928  + 0.96320784650 *T)
    sumZ0  +=       3.169 * cos(1.34499488466  + 159.12442469020 *T)
    sumZ0  +=       3.871 * cos(2.71348462732  + 2207.62954059540 *T)
    sumZ0  +=       2.814 * cos(0.52649713800  + 778.41478318470 *T)
    sumZ0  +=       2.890 * cos(1.29415320006  + 2737.32050569000 *T)
    sumZ0  +=       2.831 * cos(4.45595764630  + 518.64526483070 *T)
    sumZ0  +=       2.757 * cos(2.05371431350  + 214.52293571540 *T)
    sumZ0  +=       2.757 * cos(2.83591185309  + 212.07525516060 *T)
    sumZ0  +=       3.187 * cos(2.00772012188  + 2847.52682690940 *T)
    sumZ0  +=       2.712 * cos(5.27492060596  + 305.34616939270 *T)
    sumZ0  +=       2.771 * cos(2.54360038772  + 319.57326339430 *T)
    sumZ0  +=       2.630 * cos(5.50124179878  + 423.67742956920 *T)
    sumZ0  +=       2.626 * cos(0.49212317679  + 153.49535039770 *T)
    sumZ0  +=       2.828 * cos(3.04457317761  + 211.60217440860 *T)
    sumZ0  +=       2.650 * cos(2.71849987794  + 145.63104387150 *T)
    sumZ0  +=       2.589 * cos(5.18248239650  + 1802.37199072180 *T)
    sumZ0  +=       2.579 * cos(1.68789924795  + 6.15033915430 *T)
    sumZ0  +=       3.452 * cos(0.69033243447  + 776.93031047640 *T)
    sumZ0  +=       2.527 * cos(3.47384786002  + 184.98791978670 *T)
    sumZ0  +=       2.496 * cos(5.96218386110  + 2.92076130680 *T)
    sumZ0  +=       3.222 * cos(3.60809233593  + 1251.34038462480 *T)
    sumZ0  +=       2.684 * cos(5.48972047825  + 391.17346822390 *T)
    sumZ0  +=       2.370 * cos(1.54660575926  + 51.20572533120 *T)
    sumZ0  +=       2.310 * cos(0.63846192912  + 416.30325013750 *T)
    sumZ0  +=       2.444 * cos(3.87972666346  + 182.27960680100 *T)
    sumZ0  +=       2.435 * cos(2.71829697900  + 627.36711334180 *T)
    sumZ0  +=       2.236 * cos(1.31003501645  + 241.61027108930 *T)
    sumZ0  +=       2.653 * cos(5.70255270090  + 220.93390730060 *T)
    sumZ0  +=       2.646 * cos(3.30185009916  + 696.51963761660 *T)
    sumZ0  +=       2.211 * cos(4.72699183302  + 387.24131496080 *T)
    sumZ0  +=       2.313 * cos(2.81517281649  + 738.79727483860 *T)
    sumZ0  +=       2.197 * cos(2.63929129920  + 436.89313161450 *T)
    sumZ0  +=       2.348 * cos(1.83689902078  + 214.99601646740 *T)
    sumZ0  +=       2.842 * cos(2.88171478117  + 337.73251065900 *T)
    sumZ0  +=       2.195 * cos(1.58125142363  + 1677.93857550080 *T)
    sumZ0  +=       2.609 * cos(1.41922918177  + 329.72519178090 *T)
    sumZ0  +=       2.207 * cos(5.67197607780  + 905.88657979150 *T)
    sumZ0  +=       2.182 * cos(3.15263816837  + 439.12836384820 *T)
    sumZ0  +=       2.060 * cos(5.78919528834  + 120.35824960600 *T)
    sumZ0  +=       2.079 * cos(5.08007114480  + 7.06536289100 *T)
    sumZ0  +=       2.351 * cos(0.26849523535  + 140.00196957900 *T)
    sumZ0  +=       2.770 * cos(4.57559729641  + 24.37902238820 *T)
    sumZ0  +=       2.705 * cos(4.32491993913  + 46.47042291600 *T)
    sumZ0  +=       2.052 * cos(1.22649156182  + 92.04707395470 *T)
    sumZ0  +=       2.046 * cos(5.19173373776  + 347.88443904560 *T)
    sumZ0  +=       2.036 * cos(2.88971262431  + 980.66817835880 *T)
    sumZ0  +=       2.481 * cos(1.55318712059  + 576.16138801060 *T)
    sumZ0  +=       2.024 * cos(4.32935289944  + 31.01948863700 *T)
    sumZ0  +=       2.006 * cos(2.01138855049  + 211.86280683950 *T)
    sumZ0  +=       2.006 * cos(2.87823761610  + 214.73538403650 *T)
    sumZ0  +=       2.026 * cos(2.42235724251  + 2641.34127847220 *T)
    sumZ0  +=       1.938 * cos(2.48016602910  + 487.36514376280 *T)
    sumZ0  +=       2.065 * cos(1.16081214330  + 192.69216761850 *T)
    sumZ0  +=       2.254 * cos(5.99623174740  + 1905.46476494040 *T)
    sumZ0  +=       1.922 * cos(5.98223148262  + 96.87299909510 *T)
    sumZ0  +=       2.534 * cos(2.62272549680  + 421.93232443000 *T)
    sumZ0  +=       2.106 * cos(1.84883841490  + 299.12639426920 *T)
    sumZ0  +=       2.211 * cos(1.86083014117  + 221.16340196420 *T)
    sumZ0  +=       2.083 * cos(6.14443432076  + 59.80374504030 *T)
    sumZ0  +=       1.909 * cos(1.13231253998  + 282.45161971280 *T)
    sumZ0  +=       1.898 * cos(3.53454927143  + 219.66188291340 *T)
    sumZ0  +=       2.007 * cos(4.35167828920  + 444.75743814070 *T)
    sumZ0  +=       1.974 * cos(5.78637079670  + 430.79097657000 *T)
    sumZ0  +=       1.937 * cos(4.03561100629  + 429.51895218280 *T)
    sumZ0  +=       1.838 * cos(2.97004270413  + 425.84743135060 *T)
    sumZ0  +=       1.880 * cos(3.59058053053  + 3340.61242669980 *T)
    sumZ0  +=       1.858 * cos(5.83605607376  + 280.96714700450 *T)
    sumZ0  +=       1.768 * cos(4.86059839675  + 2118.76386037840 *T)
    sumZ0  +=       2.128 * cos(3.95375781114  + 831.10498122420 *T)
    sumZ0  +=       2.111 * cos(5.51034879271  + 428.08266358430 *T)
    sumZ0  +=       2.359 * cos(3.95534711749  + 2.96894541660 *T)
    sumZ0  +=       2.084 * cos(4.99656154667  + 254.94359321360 *T)
    sumZ0  +=       2.189 * cos(5.75843317256  + 418.52143602870 *T)
    sumZ0  +=       2.225 * cos(5.22774546552  + 638.41281360570 *T)
    sumZ0  +=       2.362 * cos(0.10821349524  + 131.40394986990 *T)
    sumZ0  +=       1.938 * cos(3.62212314488  + 267.47376618580 *T)
    sumZ0  +=       2.349 * cos(2.45602769901  + 2524.02141025200 *T)
    sumZ0  +=       1.752 * cos(0.17721513623  + 969.62247809490 *T)
    sumZ0  +=       2.216 * cos(4.46378373321  + 353.30106501700 *T)
    sumZ0  +=       1.697 * cos(3.74680430424  + 421.18156490460 *T)
    sumZ0  +=       1.692 * cos(1.29673821700  + 2840.41327990860 *T)
    sumZ0  +=       1.698 * cos(2.09387250037  + 292.01284726840 *T)
    sumZ0  +=       2.118 * cos(2.90133577623  + 205.43478891180 *T)
    sumZ0  +=       2.049 * cos(1.88559014939  + 199.28444975750 *T)
    sumZ0  +=       1.712 * cos(2.97477489664  + 273.10284047830 *T)
    sumZ0  +=       1.828 * cos(2.23147050409  + 1781.03134971940 *T)
    sumZ0  +=       1.933 * cos(3.33016281296  + 533.62311835770 *T)
    sumZ0  +=       1.737 * cos(2.59432238776  + 264.50482076920 *T)
    sumZ0  +=       1.676 * cos(2.73737546375  + 3480.31056622260 *T)
    sumZ0  +=       1.935 * cos(2.01080991907  + 7.16173111060 *T)
    sumZ0  +=       1.602 * cos(4.81017819939  + 1382.88734684660 *T)
    sumZ0  +=       1.894 * cos(4.74702697079  + 558.00214074590 *T)
    sumZ0  +=       1.589 * cos(0.85622467383  + 1044.40407666220 *T)
    sumZ0  +=       1.565 * cos(5.03064453644  + 458.84151979040 *T)
    sumZ0  +=       1.648 * cos(5.00766073214  + 2648.45482547300 *T)
    sumZ0  +=       1.562 * cos(4.62304098893  + 543.02428721890 *T)
    sumZ0  +=       1.794 * cos(2.92299868481  + 227.31374111850 *T)
    sumZ0  +=       1.844 * cos(6.25012647811  + 432.22726516850 *T)
    sumZ0  +=       1.821 * cos(5.47892889469  + 205.66428357540 *T)
    sumZ0  +=       1.605 * cos(2.57251520836  + 3127.31333126180 *T)
    sumZ0  +=       1.477 * cos(1.36669731999  + 206.93630796260 *T)
    sumZ0  +=       1.468 * cos(4.67912895132  + 339.28641933650 *T)
    sumZ0  +=       1.481 * cos(0.92457754588  + 1994.33044515740 *T)
    sumZ0  +=       1.547 * cos(0.13419340100  + 465.95506679120 *T)
    sumZ0  +=       1.457 * cos(0.52628058703  + 427.34895040140 *T)
    sumZ0  +=       1.902 * cos(1.24600006636  + 166.82867252200 *T)
    sumZ0  +=       1.529 * cos(2.85370771355  + 209.15449385380 *T)
    sumZ0  +=       1.494 * cos(3.66845408834  + 1182.92157353290 *T)
    sumZ0  +=       1.588 * cos(1.63988438811  + 3163.91869656600 *T)
    sumZ0  +=       1.502 * cos(4.96348459770  + 373.90799283650 *T)
    sumZ0  +=       1.608 * cos(3.59138122093  + 233.90602325750 *T)
    sumZ0  +=       1.460 * cos(1.17818401345  + 92.94084583200 *T)
    sumZ0  +=       1.462 * cos(5.05747764334  + 33.94024994380 *T)
    sumZ0  +=       1.490 * cos(2.03601440129  + 217.44369702220 *T)
    sumZ0  +=       1.688 * cos(1.50928274793  + 815.06334611420 *T)
    sumZ0  +=       1.778 * cos(0.24563726151  + 6.59228213900 *T)
    sumZ0  +=       1.834 * cos(3.98017256784  + 212.40532356070 *T)
    sumZ0  +=       1.834 * cos(0.90945359875  + 214.19286731530 *T)
    sumZ0  +=       1.451 * cos(3.59438611623  + 494.26624244250 *T)
    sumZ0  +=       1.474 * cos(4.97772142978  + 1699.27921650320 *T)
    sumZ0  +=       1.388 * cos(4.77847833406  + 2854.64037391020 *T)
    sumZ0  +=       1.589 * cos(5.43527416407  + 2538.24850425360 *T)
    sumZ0  +=       1.390 * cos(4.08155384651  + 894.84087952760 *T)
    sumZ0  +=       1.435 * cos(1.51395699369  + 867.42347575360 *T)
    sumZ0  +=       1.404 * cos(4.02343833901  + 300.61086697750 *T)

    # SATURN Z Coord
    sumZ1  = 0
    sumZ1  +=  3810308.320 * cos(5.33520316778  + 213.29909543800 *T)
    sumZ1  +=  707598.508 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ1  +=  144960.439 * cos(2.30149367100  + 206.18554843720 *T)
    sumZ1  +=   89591.263 * cos(1.69286730496  + 220.41264243880 *T)
    sumZ1  +=   88982.080 * cos(6.09698706089  + 426.59819087600 *T)
    sumZ1  +=   18761.941 * cos(1.23200112053  + 419.48464387520 *T)
    sumZ1  +=   18103.257 * cos(2.90933498453  + 7.11354700080 *T)
    sumZ1  +=    6860.076 * cos(0.84862244997  + 433.71173787680 *T)
    sumZ1  +=    8398.809 * cos(5.89715948074  + 639.89728631400 *T)
    sumZ1  +=    5939.993 * cos(0.42396560640  + 316.39186965660 *T)
    sumZ1  +=    2571.420 * cos(1.61235344206  + 227.52618943960 *T)
    sumZ1  +=    2030.372 * cos(5.19792292680  + 110.20632121940 *T)
    sumZ1  +=    1632.449 * cos(2.43938017095  + 199.07200143640 *T)
    sumZ1  +=    1126.250 * cos(2.85739547477  + 14.22709400160 *T)
    sumZ1  +=    1025.703 * cos(0.77676708535  + 632.78373931320 *T)
    sumZ1  +=    1035.641 * cos(2.45710665385  + 217.23124870110 *T)
    sumZ1  +=     919.338 * cos(5.69490380320  + 529.69096509460 *T)
    sumZ1  +=    1025.261 * cos(3.40455226241  + 103.09277421860 *T)
    sumZ1  +=     788.013 * cos(2.75643752556  + 210.11770170030 *T)
    sumZ1  +=     652.471 * cos(1.66305264153  + 202.25339517410 *T)
    sumZ1  +=     678.403 * cos(5.28303371104  + 853.19638175200 *T)
    sumZ1  +=     546.618 * cos(0.04231005132  + 647.01083331480 *T)
    sumZ1  +=     616.068 * cos(1.25272437511  + 216.48048917570 *T)
    sumZ1  +=     589.338 * cos(1.25138980225  + 209.36694217490 *T)
    sumZ1  +=     398.211 * cos(2.75660806272  + 117.31986822020 *T)
    sumZ1  +=     342.895 * cos(1.80589354512  + 224.34479570190 *T)
    sumZ1  +=     383.087 * cos(4.43945576495  + 11.04570026390 *T)
    sumZ1  +=     315.596 * cos(1.62252525888  + 323.50541665740 *T)
    sumZ1  +=     215.621 * cos(0.81176116742  + 440.82528487760 *T)
    sumZ1  +=     208.871 * cos(1.21651777785  + 63.73589830340 *T)
    sumZ1  +=     180.311 * cos(0.93555501458  + 309.27832265580 *T)
    sumZ1  +=     172.362 * cos(0.22156069668  + 735.87651353180 *T)
    sumZ1  +=     132.928 * cos(1.49015865779  + 95.97922721780 *T)
    sumZ1  +=     141.627 * cos(1.38605092643  + 412.37109687440 *T)
    sumZ1  +=     114.025 * cos(4.13401798874  + 415.55249061210 *T)
    sumZ1  +=     107.829 * cos(4.67440616464  + 846.08283475120 *T)
    sumZ1  +=     113.250 * cos(3.23797737561  + 277.03499374140 *T)
    sumZ1  +=      96.663 * cos(5.13321545536  + 522.57741809380 *T)
    sumZ1  +=      80.905 * cos(5.97207598010  + 149.56319713460 *T)
    sumZ1  +=      78.248 * cos(4.68235800719  + 215.74677599280 *T)
    sumZ1  +=      73.767 * cos(3.76912541093  + 74.78159856730 *T)
    sumZ1  +=      58.294 * cos(4.12074427198  + 1162.47470440780 *T)
    sumZ1  +=      59.154 * cos(0.00909087388  + 1052.26838318840 *T)
    sumZ1  +=      54.000 * cos(2.63292711570  + 3.18139373770 *T)
    sumZ1  +=      53.374 * cos(1.48573267831  + 234.63973644040 *T)
    sumZ1  +=      57.043 * cos(1.87312646111  + 423.41679713830 *T)
    sumZ1  +=      52.955 * cos(1.65323274690  + 330.61896365820 *T)
    sumZ1  +=      50.894 * cos(2.68041943186  + 742.99006053260 *T)
    sumZ1  +=      49.329 * cos(4.95750749572  + 3.93215326310 *T)
    sumZ1  +=      61.047 * cos(0.29238361571  + 422.66603761290 *T)
    sumZ1  +=      53.719 * cos(2.46146211746  + 490.33408917940 *T)
    sumZ1  +=      51.705 * cos(4.60792983851  + 1066.49547719000 *T)
    sumZ1  +=      43.576 * cos(5.53983423151  + 860.30992875280 *T)
    sumZ1  +=      38.478 * cos(2.42479516896  + 430.53034413910 *T)
    sumZ1  +=      37.913 * cos(5.63475884335  + 210.85141488320 *T)
    sumZ1  +=      42.636 * cos(0.46334621133  + 429.77958461370 *T)
    sumZ1  +=      38.288 * cos(3.94527342286  + 351.81659230870 *T)
    sumZ1  +=      32.052 * cos(2.73490028872  + 21.34064100240 *T)
    sumZ1  +=      37.226 * cos(0.77640416852  + 536.80451209540 *T)
    sumZ1  +=      34.303 * cos(5.53964443183  + 949.17560896980 *T)
    sumZ1  +=      32.486 * cos(1.51592537935  + 437.64389113990 *T)
    sumZ1  +=      25.209 * cos(5.61832461480  + 203.73786788240 *T)
    sumZ1  +=      32.600 * cos(1.15985496536  + 223.59403617650 *T)
    sumZ1  +=      28.157 * cos(5.03992835097  + 515.46387109300 *T)
    sumZ1  +=      25.365 * cos(1.63357066215  + 137.03302416240 *T)
    sumZ1  +=      22.772 * cos(3.87097220542  + 1478.86657406440 *T)
    sumZ1  +=      22.874 * cos(1.52589628823  + 124.43341522100 *T)
    sumZ1  +=      21.781 * cos(1.25479228248  + 88.86568021700 *T)
    sumZ1  +=      21.071 * cos(2.91498706203  + 203.00415469950 *T)
    sumZ1  +=      22.077 * cos(6.03637528612  + 1368.66025284500 *T)
    sumZ1  +=      21.851 * cos(3.47826348846  + 38.13303563780 *T)
    sumZ1  +=      22.636 * cos(0.42502415739  + 127.47179660680 *T)
    sumZ1  +=      18.793 * cos(0.02824698856  + 654.12438031560 *T)
    sumZ1  +=      18.132 * cos(0.35264011886  + 85.82729883120 *T)
    sumZ1  +=      19.127 * cos(4.85094487248  + 838.96928775040 *T)
    sumZ1  +=      15.811 * cos(0.06549781548  + 302.16477565500 *T)
    sumZ1  +=      14.993 * cos(5.63243246457  + 214.26230328450 *T)
    sumZ1  +=      16.073 * cos(5.40920567275  + 1265.56747862640 *T)
    sumZ1  +=      14.177 * cos(2.06003944064  + 138.51749687070 *T)
    sumZ1  +=      14.570 * cos(5.68093007156  + 942.06206196900 *T)
    sumZ1  +=      12.460 * cos(2.22381807104  + 213.34727954780 *T)
    sumZ1  +=      12.340 * cos(5.42069314117  + 217.96496188400 *T)
    sumZ1  +=      11.994 * cos(3.89295080099  + 728.76296653100 *T)
    sumZ1  +=      15.172 * cos(4.88775469732  + 208.63322899200 *T)
    sumZ1  +=      14.560 * cos(4.02541618429  + 195.13984817330 *T)
    sumZ1  +=      12.463 * cos(5.30921040178  + 213.25091132820 *T)
    sumZ1  +=      11.186 * cos(1.39522203481  + 210.37833413120 *T)
    sumZ1  +=      11.627 * cos(2.81135897939  + 231.45834270270 *T)
    sumZ1  +=      15.326 * cos(4.31191275758  + 628.85158605010 *T)
    sumZ1  +=      12.272 * cos(2.46356221718  + 200.76892246580 *T)
    sumZ1  +=      15.060 * cos(1.90080442413  + 212.33588759150 *T)
    sumZ1  +=      11.799 * cos(1.46416357606  + 218.71572140940 *T)
    sumZ1  +=      10.790 * cos(0.87460319256  + 408.43894361130 *T)
    sumZ1  +=      10.282 * cos(5.35193913607  + 207.67002114550 *T)
    sumZ1  +=      13.027 * cos(3.77267768832  + 350.33211960040 *T)
    sumZ1  +=      10.032 * cos(2.41897316373  + 10.29494073850 *T)
    sumZ1  +=      12.485 * cos(2.56988226727  + 340.77089204480 *T)
    sumZ1  +=      11.155 * cos(4.46475107308  + 52.69019803950 *T)
    sumZ1  +=      11.678 * cos(5.99973973452  + 191.20769491020 *T)
    sumZ1  +=       9.985 * cos(5.55354944984  + 76.26607127560 *T)
    sumZ1  +=       9.222 * cos(3.47198476691  + 18.15924726470 *T)
    sumZ1  +=       8.684 * cos(3.59341160996  + 1795.25844372100 *T)
    sumZ1  +=       8.772 * cos(2.84157598451  + 191.95845443560 *T)
    sumZ1  +=       8.792 * cos(5.77952076653  + 1685.05212250160 *T)
    sumZ1  +=      10.212 * cos(4.67919299694  + 1155.36115740700 *T)
    sumZ1  +=       7.416 * cos(2.34743956282  + 160.60889739850 *T)
    sumZ1  +=       8.681 * cos(3.66196489035  + 222.86032299360 *T)
    sumZ1  +=       8.204 * cos(4.00167578990  + 388.46515523820 *T)
    sumZ1  +=       7.229 * cos(0.85694899722  + 142.44965013380 *T)
    sumZ1  +=       7.962 * cos(6.12384852532  + 209.10630974400 *T)
    sumZ1  +=       8.152 * cos(0.17850883770  + 1.48447270830 *T)
    sumZ1  +=       7.462 * cos(4.54643529511  + 625.67019231240 *T)
    sumZ1  +=       8.149 * cos(4.75191185936  + 831.85574074960 *T)
    sumZ1  +=       7.464 * cos(5.87758421420  + 362.86229257260 *T)
    sumZ1  +=       6.550 * cos(0.37942895236  + 212.77783057620 *T)
    sumZ1  +=       7.220 * cos(5.21944812420  + 1581.95934828300 *T)
    sumZ1  +=       6.436 * cos(0.63886589467  + 22.09140052780 *T)
    sumZ1  +=       7.311 * cos(0.07712493176  + 265.98929347750 *T)
    sumZ1  +=       7.040 * cos(5.39023352132  + 1258.45393162560 *T)
    sumZ1  +=       6.246 * cos(6.10782448635  + 618.55664531160 *T)
    sumZ1  +=       5.540 * cos(1.63194253275  + 56.62235130260 *T)
    sumZ1  +=       7.460 * cos(2.16123164874  + 207.88246946660 *T)
    sumZ1  +=       6.689 * cos(5.61849505042  + 483.22054217860 *T)
    sumZ1  +=       5.366 * cos(2.54918207389  + 554.06998748280 *T)
    sumZ1  +=       7.245 * cos(3.44291004853  + 12.53017297220 *T)
    sumZ1  +=       5.769 * cos(0.85039589871  + 225.82926841020 *T)
    sumZ1  +=       6.530 * cos(0.87136105335  + 213.82036029980 *T)
    sumZ1  +=       5.249 * cos(0.71690429208  + 447.93883187840 *T)
    sumZ1  +=       4.793 * cos(4.97423814367  + 212.54833591260 *T)
    sumZ1  +=       5.801 * cos(0.02624779337  + 269.92144674060 *T)
    sumZ1  +=       4.710 * cos(3.07134751984  + 1375.77379984580 *T)
    sumZ1  +=       5.649 * cos(5.76570118321  + 635.96513305090 *T)
    sumZ1  +=       4.366 * cos(4.43717368507  + 235.39049596580 *T)
    sumZ1  +=       5.595 * cos(1.00573470142  + 70.84944530420 *T)
    sumZ1  +=       5.094 * cos(0.10066001708  + 750.10360753340 *T)
    sumZ1  +=       4.487 * cos(4.79200734443  + 218.92816973050 *T)
    sumZ1  +=       4.220 * cos(3.11074964024  + 703.63318461740 *T)
    sumZ1  +=       4.530 * cos(2.43215667228  + 216.21985674480 *T)
    sumZ1  +=       4.076 * cos(1.21943503644  + 417.03696332040 *T)
    sumZ1  +=       4.075 * cos(5.52140305637  + 2001.44399215820 *T)
    sumZ1  +=       4.489 * cos(5.12835463313  + 1574.84580128220 *T)
    sumZ1  +=       3.835 * cos(4.59484102776  + 424.15051032120 *T)
    sumZ1  +=       4.075 * cos(4.07912904262  + 1279.79457262800 *T)
    sumZ1  +=       4.918 * cos(4.36928157904  + 1471.75302706360 *T)
    sumZ1  +=       3.719 * cos(3.42064399733  + 1045.15483618760 *T)
    sumZ1  +=       4.734 * cos(5.61204474661  + 9.56122755560 *T)
    sumZ1  +=       4.537 * cos(0.79869612235  + 543.91805909620 *T)
    sumZ1  +=       3.610 * cos(5.94058751363  + 934.94851496820 *T)
    sumZ1  +=       4.230 * cos(5.90345986772  + 173.94221952280 *T)
    sumZ1  +=       3.444 * cos(2.86327846707  + 106.27416795630 *T)
    sumZ1  +=       3.432 * cos(4.89216582171  + 497.44763618020 *T)
    sumZ1  +=       3.553 * cos(4.80761740206  + 1073.60902419080 *T)
    sumZ1  +=       4.601 * cos(4.77897976995  + 288.08069400530 *T)
    sumZ1  +=       3.314 * cos(3.41625676576  + 2111.65031337760 *T)
    sumZ1  +=       3.264 * cos(4.45520175360  + 107.02492748170 *T)
    sumZ1  +=       3.442 * cos(2.57824148394  + 113.38771495710 *T)
    sumZ1  +=       3.664 * cos(4.49847588517  + 1148.24761040620 *T)
    sumZ1  +=       3.349 * cos(1.27990783834  + 344.70304530790 *T)
    sumZ1  +=       2.971 * cos(4.31371705105  + 219.44943459230 *T)
    sumZ1  +=       2.920 * cos(1.64503980864  + 636.71589257630 *T)
    sumZ1  +=       2.971 * cos(2.61742682364  + 404.50679034820 *T)
    sumZ1  +=       3.051 * cos(5.10422803245  + 1898.35121793960 *T)
    sumZ1  +=       2.774 * cos(3.14101665470  + 217.49188113200 *T)
    sumZ1  +=       2.950 * cos(5.37168308363  + 175.16605980020 *T)
    sumZ1  +=       2.822 * cos(1.33413781392  + 650.94298657790 *T)
    sumZ1  +=       3.174 * cos(1.19063213567  + 337.73251065900 *T)
    sumZ1  +=       3.041 * cos(6.26099119193  + 207.14875628370 *T)
    sumZ1  +=       2.591 * cos(4.52751512849  + 565.11568774670 *T)
    sumZ1  +=       3.416 * cos(5.97202994093  + 643.07868005170 *T)
    sumZ1  +=       2.507 * cos(0.83753919585  + 436.89313161450 *T)
    sumZ1  +=       2.350 * cos(2.16713733034  + 177.87437278590 *T)
    sumZ1  +=       2.755 * cos(3.61967991567  + 114.13847448250 *T)
    sumZ1  +=       2.495 * cos(4.64668211731  + 429.04587143080 *T)
    sumZ1  +=       2.768 * cos(4.77439427919  + 1891.23767093880 *T)
    sumZ1  +=       2.238 * cos(4.00121267561  + 284.14854074220 *T)
    sumZ1  +=       2.251 * cos(4.90436209991  + 2214.74308759620 *T)
    sumZ1  +=       2.409 * cos(3.56267761888  + 1169.58825140860 *T)
    sumZ1  +=       2.157 * cos(3.02442510484  + 1692.16566950240 *T)
    sumZ1  +=       2.003 * cos(2.47564274875  + 416.30325013750 *T)
    sumZ1  +=       2.085 * cos(4.09277149039  + 1272.68102562720 *T)
    sumZ1  +=       1.945 * cos(0.97841451672  + 114.39910691340 *T)
    sumZ1  +=       1.907 * cos(3.36673783584  + 1059.38193018920 *T)
    sumZ1  +=       1.861 * cos(4.45402970588  + 508.35032409220 *T)
    sumZ1  +=       1.874 * cos(2.73629728926  + 214.04985496340 *T)
    sumZ1  +=       1.905 * cos(2.99492664264  + 1361.54670584420 *T)
    sumZ1  +=       2.064 * cos(5.39674851006  + 245.54242435240 *T)
    sumZ1  +=       1.878 * cos(0.56676085326  + 621.73803904930 *T)
    sumZ1  +=       2.525 * cos(1.19036969669  + 479.28838891550 *T)
    sumZ1  +=       1.966 * cos(2.58393370207  + 99.16062095550 *T)
    sumZ1  +=       1.908 * cos(2.85404393589  + 355.74874557180 *T)
    sumZ1  +=       2.056 * cos(2.39618093351  + 2435.15573003500 *T)
    sumZ1  +=       2.158 * cos(4.97710925643  + 2317.83586181480 *T)
    sumZ1  +=       2.202 * cos(0.84567094644  + 358.93013930950 *T)
    sumZ1  +=       1.881 * cos(1.13062277248  + 757.21715453420 *T)
    sumZ1  +=       1.707 * cos(1.57935987501  + 6069.77675455340 *T)
    sumZ1  +=       2.151 * cos(4.06616159617  + 1788.14489672020 *T)
    sumZ1  +=       1.746 * cos(3.06455575770  + 2008.55753915900 *T)
    sumZ1  +=       1.746 * cos(5.09056895026  + 220.46082654860 *T)
    sumZ1  +=       1.795 * cos(2.95316519456  + 1781.03134971940 *T)
    sumZ1  +=       1.700 * cos(6.05669760153  + 206.13736432740 *T)
    sumZ1  +=       1.614 * cos(5.96461527769  + 867.42347575360 *T)
    sumZ1  +=       1.956 * cos(2.13214810643  + 252.65597135320 *T)
    sumZ1  +=       1.739 * cos(5.77930712282  + 181.05576652360 *T)
    sumZ1  +=       1.553 * cos(1.32350368373  + 131.54696222180 *T)
    sumZ1  +=       1.599 * cos(0.86145918975  + 81.75213321620 *T)
    sumZ1  +=       1.792 * cos(0.86879017108  + 4.66586644600 *T)
    sumZ1  +=       1.548 * cos(4.47085375766  + 121.25202148330 *T)
    sumZ1  +=       1.619 * cos(4.79891754382  + 842.15068148810 *T)
    sumZ1  +=       1.513 * cos(3.01415732754  + 2428.04218303420 *T)
    sumZ1  +=       1.822 * cos(2.62978911169  + 643.82943957710 *T)
    sumZ1  +=       2.085 * cos(3.00604517701  + 563.63121503840 *T)
    sumZ1  +=       1.939 * cos(2.25124681713  + 1677.93857550080 *T)
    sumZ1  +=       1.472 * cos(3.39986301851  + 956.28915597060 *T)
    sumZ1  +=       1.467 * cos(4.24274724035  + 98.89998852460 *T)
    sumZ1  +=       1.384 * cos(2.33959628851  + 444.75743814070 *T)
    sumZ1  +=       1.810 * cos(2.02611183018  + 2420.92863603340 *T)
    sumZ1  +=       1.628 * cos(4.47984535594  + 251.43213107580 *T)
    sumZ1  +=       1.593 * cos(1.97508934614  + 62.25142559510 *T)
    sumZ1  +=       1.516 * cos(1.25225067905  + 425.11371816770 *T)
    sumZ1  +=       1.181 * cos(5.09622518839  + 414.06801790380 *T)
    sumZ1  +=       1.211 * cos(0.96174816662  + 214.78356814630 *T)
    sumZ1  +=       1.429 * cos(2.95386700663  + 65.22037101170 *T)
    sumZ1  +=       1.042 * cos(5.99222596459  + 151.04766984290 *T)
    sumZ1  +=       1.200 * cos(0.65918491217  + 211.81462272970 *T)
    sumZ1  +=       0.967 * cos(6.01089625149  + 1589.07289528380 *T)
    sumZ1  +=       0.904 * cos(0.07909774752  + 213.41097002260 *T)
    sumZ1  +=       0.904 * cos(1.66925524906  + 213.18722085340 *T)
    sumZ1  +=       1.071 * cos(3.60798747502  + 248.72381809010 *T)
    sumZ1  +=       0.815 * cos(4.93878023673  + 2.44768055480 *T)
    sumZ1  +=       0.775 * cos(3.45264409045  + 198.32124191100 *T)
    sumZ1  +=       0.983 * cos(3.20558404998  + 312.19908396260 *T)
    sumZ1  +=       0.760 * cos(2.28889729136  + 213.08664711690 *T)
    sumZ1  +=       0.760 * cos(5.74264101239  + 213.51154375910 *T)
    sumZ1  +=       0.724 * cos(4.57593199965  + 228.27694896500 *T)

    # SATURN Z Coord
    sumZ2  = 0
    sumZ2  +=  195192.259 * cos(0.50946874402  + 213.29909543800 *T)
    sumZ2  +=   36491.216 * cos(3.99604491617  + 206.18554843720 *T)
    sumZ2  +=   15091.412 * cos(6.17687148383  + 220.41264243880 *T)
    sumZ2  +=   19236.496 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ2  +=    3420.562 * cos(3.04256365132  + 419.48464387520 *T)
    sumZ2  +=    1903.731 * cos(1.01852901874  + 7.11354700080 *T)
    sumZ2  +=    1553.807 * cos(5.27527652115  + 433.71173787680 *T)
    sumZ2  +=    1658.586 * cos(5.05753354015  + 426.59819087600 *T)
    sumZ2  +=     939.968 * cos(6.15093957163  + 227.52618943960 *T)
    sumZ2  +=     665.875 * cos(1.95823182523  + 316.39186965660 *T)
    sumZ2  +=     727.405 * cos(3.86689219213  + 639.89728631400 *T)
    sumZ2  +=     708.634 * cos(4.14159824503  + 199.07200143640 *T)
    sumZ2  +=     332.785 * cos(1.11013701815  + 14.22709400160 *T)
    sumZ2  +=     276.129 * cos(4.52772200630  + 210.11770170030 *T)
    sumZ2  +=     246.881 * cos(4.50860976295  + 103.09277421860 *T)
    sumZ2  +=     236.419 * cos(1.02568314619  + 110.20632121940 *T)
    sumZ2  +=     193.411 * cos(4.35154385844  + 217.23124870110 *T)
    sumZ2  +=     170.756 * cos(2.72871130039  + 632.78373931320 *T)
    sumZ2  +=     154.811 * cos(4.42812217496  + 647.01083331480 *T)
    sumZ2  +=     159.221 * cos(5.67806982371  + 216.48048917570 *T)
    sumZ2  +=     100.473 * cos(3.17382026227  + 853.19638175200 *T)
    sumZ2  +=      90.850 * cos(5.30427538537  + 440.82528487760 *T)
    sumZ2  +=      77.600 * cos(3.28475417716  + 202.25339517410 *T)
    sumZ2  +=      72.917 * cos(1.06686891373  + 117.31986822020 *T)
    sumZ2  +=      61.331 * cos(1.61343080245  + 529.69096509460 *T)
    sumZ2  +=      64.246 * cos(3.02694204128  + 412.37109687440 *T)
    sumZ2  +=      63.778 * cos(5.55207588103  + 209.36694217490 *T)
    sumZ2  +=      41.451 * cos(3.33188411884  + 309.27832265580 *T)
    sumZ2  +=      33.437 * cos(0.11560358087  + 323.50541665740 *T)
    sumZ2  +=      29.893 * cos(6.03701243746  + 234.63973644040 *T)
    sumZ2  +=      25.877 * cos(3.50813571651  + 95.97922721780 *T)
    sumZ2  +=      17.574 * cos(0.22180355628  + 3.93215326310 *T)
    sumZ2  +=      16.154 * cos(3.77501334464  + 423.41679713830 *T)
    sumZ2  +=      15.793 * cos(1.02173188447  + 21.34064100240 *T)
    sumZ2  +=      14.438 * cos(5.60523358660  + 223.59403617650 *T)
    sumZ2  +=      13.542 * cos(4.81528922933  + 429.77958461370 *T)
    sumZ2  +=      15.398 * cos(5.52110630605  + 224.34479570190 *T)
    sumZ2  +=      11.511 * cos(4.71374175861  + 203.00415469950 *T)
    sumZ2  +=      14.090 * cos(5.07886462834  + 277.03499374140 *T)
    sumZ2  +=      14.469 * cos(3.60157016903  + 860.30992875280 *T)
    sumZ2  +=      12.120 * cos(0.05423484257  + 415.55249061210 *T)
    sumZ2  +=      13.400 * cos(0.51901182147  + 522.57741809380 *T)
    sumZ2  +=      10.074 * cos(4.55389948936  + 422.66603761290 *T)
    sumZ2  +=      12.614 * cos(6.15368985726  + 330.61896365820 *T)
    sumZ2  +=      10.793 * cos(2.51124542537  + 1066.49547719000 *T)
    sumZ2  +=       8.990 * cos(6.00345025989  + 124.43341522100 *T)
    sumZ2  +=      10.308 * cos(6.00841164198  + 11.04570026390 *T)
    sumZ2  +=       8.308 * cos(0.68581626528  + 3.18139373770 *T)
    sumZ2  +=       9.597 * cos(5.80785895962  + 63.73589830340 *T)
    sumZ2  +=       7.930 * cos(5.51468008833  + 735.87651353180 *T)
    sumZ2  +=       8.053 * cos(3.03363150224  + 88.86568021700 *T)
    sumZ2  +=       8.834 * cos(4.47705749563  + 654.12438031560 *T)
    sumZ2  +=       7.356 * cos(4.43315310489  + 191.95845443560 *T)
    sumZ2  +=       6.543 * cos(4.88541538755  + 625.67019231240 *T)
    sumZ2  +=       8.411 * cos(1.49318798287  + 742.99006053260 *T)
    sumZ2  +=       6.909 * cos(4.48978383769  + 430.53034413910 *T)
    sumZ2  +=       6.388 * cos(0.08457977809  + 215.74677599280 *T)
    sumZ2  +=       7.107 * cos(0.70220678131  + 949.17560896980 *T)
    sumZ2  +=       5.669 * cos(0.46049062500  + 515.46387109300 *T)
    sumZ2  +=       5.034 * cos(5.23865643179  + 74.78159856730 *T)
    sumZ2  +=       4.315 * cos(1.63053592014  + 1059.38193018920 *T)
    sumZ2  +=       4.299 * cos(1.18161521312  + 210.85141488320 *T)
    sumZ2  +=       3.314 * cos(5.42600783265  + 437.64389113990 *T)
    sumZ2  +=       3.489 * cos(0.53174386810  + 10.29494073850 *T)
    sumZ2  +=       3.355 * cos(5.23729212209  + 447.93883187840 *T)
    sumZ2  +=       3.056 * cos(1.18250311873  + 942.06206196900 *T)
    sumZ2  +=       2.996 * cos(5.62814654221  + 728.76296653100 *T)
    sumZ2  +=       3.361 * cos(5.11461252953  + 536.80451209540 *T)
    sumZ2  +=       2.396 * cos(1.84582366198  + 1471.75302706360 *T)
    sumZ2  +=       2.280 * cos(0.09630660324  + 195.13984817330 *T)
    sumZ2  +=       2.351 * cos(1.27223026559  + 231.45834270270 *T)
    sumZ2  +=       2.515 * cos(0.03375128717  + 838.96928775040 *T)
    sumZ2  +=       2.748 * cos(1.32449995239  + 203.73786788240 *T)
    sumZ2  +=       2.290 * cos(3.07698270979  + 408.43894361130 *T)
    sumZ2  +=       2.040 * cos(5.80760536835  + 1162.47470440780 *T)
    sumZ2  +=       2.084 * cos(0.74674443126  + 1265.56747862640 *T)
    sumZ2  +=       2.283 * cos(1.77539793334  + 846.08283475120 *T)
    sumZ2  +=       1.868 * cos(0.23989474559  + 1898.35121793960 *T)
    sumZ2  +=       1.720 * cos(1.98614962644  + 18.15924726470 *T)
    sumZ2  +=       1.849 * cos(0.68415972849  + 1258.45393162560 *T)
    sumZ2  +=       1.575 * cos(2.31138086283  + 302.16477565500 *T)
    sumZ2  +=       1.874 * cos(0.84722174143  + 1375.77379984580 *T)
    sumZ2  +=       1.401 * cos(3.28871251562  + 1073.60902419080 *T)
    sumZ2  +=       1.369 * cos(4.66184064286  + 490.33408917940 *T)
    sumZ2  +=       1.319 * cos(0.58131415395  + 618.55664531160 *T)
    sumZ2  +=       1.430 * cos(2.32626814584  + 1368.66025284500 *T)
    sumZ2  +=       1.519 * cos(0.21611173188  + 831.85574074960 *T)
    sumZ2  +=       1.250 * cos(5.97999895371  + 351.81659230870 *T)
    sumZ2  +=       0.966 * cos(2.64913416683  + 85.82729883120 *T)
    sumZ2  +=       0.771 * cos(1.16732337173  + 217.96496188400 *T)
    sumZ2  +=       0.697 * cos(0.53637719953  + 628.85158605010 *T)

    # SATURN Z Coord
    sumZ3  = 0
    sumZ3  +=    6552.316 * cos(1.98918006303  + 213.29909543800 *T)
    sumZ3  +=    6205.120 * cos(5.69682939522  + 206.18554843720 *T)
    sumZ3  +=    5692.884 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ3  +=    1736.879 * cos(4.32959476351  + 220.41264243880 *T)
    sumZ3  +=     450.555 * cos(4.83582005306  + 419.48464387520 *T)
    sumZ3  +=     231.881 * cos(4.39490786964  + 227.52618943960 *T)
    sumZ3  +=     199.865 * cos(2.33322152046  + 426.59819087600 *T)
    sumZ3  +=     205.985 * cos(5.84743845970  + 199.07200143640 *T)
    sumZ3  +=     242.710 * cos(3.41690207168  + 433.71173787680 *T)
    sumZ3  +=     139.339 * cos(5.28736351227  + 7.11354700080 *T)
    sumZ3  +=      67.206 * cos(5.62860360986  + 14.22709400160 *T)
    sumZ3  +=      56.299 * cos(1.93019307649  + 639.89728631400 *T)
    sumZ3  +=      55.780 * cos(0.01778616215  + 210.11770170030 *T)
    sumZ3  +=      50.857 * cos(3.44381337767  + 316.39186965660 *T)
    sumZ3  +=      29.135 * cos(0.23714193891  + 103.09277421860 *T)
    sumZ3  +=      30.392 * cos(2.54675335277  + 647.01083331480 *T)
    sumZ3  +=      25.847 * cos(3.51464713160  + 440.82528487760 *T)
    sumZ3  +=      20.480 * cos(3.72108337650  + 216.48048917570 *T)
    sumZ3  +=      18.853 * cos(6.10978933980  + 217.23124870110 *T)
    sumZ3  +=      18.881 * cos(4.62069633304  + 632.78373931320 *T)
    sumZ3  +=      17.464 * cos(4.72219191592  + 412.37109687440 *T)
    sumZ3  +=      10.821 * cos(1.21499195673  + 853.19638175200 *T)
    sumZ3  +=      10.927 * cos(4.30569213091  + 234.63973644040 *T)
    sumZ3  +=       9.297 * cos(5.60250826661  + 117.31986822020 *T)
    sumZ3  +=       7.656 * cos(5.66078016942  + 309.27832265580 *T)
    sumZ3  +=       7.303 * cos(5.47485612817  + 202.25339517410 *T)
    sumZ3  +=       5.000 * cos(5.51413752856  + 21.34064100240 *T)
    sumZ3  +=       4.259 * cos(5.49035642189  + 95.97922721780 *T)
    sumZ3  +=       4.838 * cos(0.29619340824  + 110.20632121940 *T)
    sumZ3  +=       3.957 * cos(6.09695713906  + 191.95845443560 *T)
    sumZ3  +=       3.467 * cos(3.73810181850  + 223.59403617650 *T)
    sumZ3  +=       3.322 * cos(1.65077341966  + 860.30992875280 *T)
    sumZ3  +=       2.905 * cos(4.28100382746  + 323.50541665740 *T)
    sumZ3  +=       2.380 * cos(2.83328075768  + 429.77958461370 *T)
    sumZ3  +=       2.398 * cos(4.14927979178  + 124.43341522100 *T)
    sumZ3  +=       2.780 * cos(2.64718847967  + 654.12438031560 *T)
    sumZ3  +=       2.787 * cos(0.65101567800  + 203.00415469950 *T)
    sumZ3  +=       1.947 * cos(1.53880195116  + 209.36694217490 *T)
    sumZ3  +=       1.856 * cos(5.62026724615  + 302.16477565500 *T)
    sumZ3  +=       1.819 * cos(4.80294039631  + 88.86568021700 *T)
    sumZ3  +=       1.959 * cos(4.38771818896  + 330.61896365820 *T)
    sumZ3  +=       1.487 * cos(3.12393556594  + 529.69096509460 *T)
    sumZ3  +=       1.276 * cos(3.91936027558  + 447.93883187840 *T)
    sumZ3  +=       1.374 * cos(1.28276260217  + 1066.49547719000 *T)
    sumZ3  +=       0.804 * cos(4.10158366806  + 224.34479570190 *T)

    # SATURN Z Coord
    sumZ4  = 0
    sumZ4  +=     789.033 * cos(1.11928782644  + 206.18554843720 *T)
    sumZ4  +=     281.365 * cos(3.13274851308  + 213.29909543800 *T)
    sumZ4  +=     169.515 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ4  +=     157.511 * cos(2.46634171520  + 220.41264243880 *T)
    sumZ4  +=      46.995 * cos(1.27259130638  + 199.07200143640 *T)
    sumZ4  +=      43.842 * cos(0.38587829145  + 419.48464387520 *T)
    sumZ4  +=      43.114 * cos(2.62919688921  + 227.52618943960 *T)
    sumZ4  +=      28.988 * cos(1.55482750505  + 433.71173787680 *T)
    sumZ4  +=       9.447 * cos(3.85114048856  + 14.22709400160 *T)
    sumZ4  +=       6.018 * cos(1.29220324867  + 426.59819087600 *T)
    sumZ4  +=       7.462 * cos(3.20599530963  + 7.11354700080 *T)
    sumZ4  +=       4.617 * cos(0.67712478942  + 647.01083331480 *T)
    sumZ4  +=       5.282 * cos(1.71952902559  + 440.82528487760 *T)
    sumZ4  +=       3.522 * cos(6.19992669453  + 639.89728631400 *T)
    sumZ4  +=       2.258 * cos(0.44292528032  + 412.37109687440 *T)
    sumZ4  +=       2.256 * cos(2.11084590211  + 210.11770170030 *T)
    sumZ4  +=       2.460 * cos(2.50380847653  + 234.63973644040 *T)
    sumZ4  +=       1.239 * cos(5.78004218066  + 110.20632121940 *T)
    sumZ4  +=       1.449 * cos(5.96384491391  + 216.48048917570 *T)
    sumZ4  +=       1.322 * cos(0.54271370335  + 853.19638175200 *T)
    sumZ4  +=       1.151 * cos(4.86229315211  + 117.31986822020 *T)
    sumZ4  +=       1.045 * cos(5.61649418233  + 316.39186965660 *T)

    # SATURN Z Coord
    sumZ5  = 0
    sumZ5  +=      77.433 * cos(2.81859989913  + 206.18554843720 *T)
    sumZ5  +=      13.826 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ5  +=       9.303 * cos(0.47844462514  + 220.41264243880 *T)
    sumZ5  +=       7.840 * cos(2.99415198633  + 199.07200143640 *T)
    sumZ5  +=       5.604 * cos(0.80222427065  + 227.52618943960 *T)
    sumZ5  +=       2.881 * cos(6.00054088318  + 433.71173787680 *T)
    sumZ5  +=       1.362 * cos(2.58324584516  + 419.48464387520 *T)
    sumZ5  +=       0.968 * cos(5.23916441152  + 213.29909543800 *T)
    sumZ5  +=       0.867 * cos(5.81659714144  + 7.11354700080 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Saturn",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getSaturnData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getSaturnXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getUranusXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Uranus"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getUranusXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getUranusXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # URANUS X Coord
    sumX0  = 0
    sumX0  +=  1916944479.396 * cos(5.48129363987  + 75.02541605080 *T)
    sumX0  +=  133267708.718 * cos(6.16089978558  + 0.24381748350 *T)
    sumX0  +=  44396480.992 * cos(1.65965632053  + 149.80701461810 *T)
    sumX0  +=  14712072.726 * cos(3.42449547672  + 73.54094334250 *T)
    sumX0  +=  14127113.794 * cos(4.39569319388  + 76.50988875911 *T)
    sumX0  +=  6225592.204 * cos(5.14041718059  + 1.72829019180 *T)
    sumX0  +=  1542809.804 * cos(4.12121005059  + 224.58861318540 *T)
    sumX0  +=  1443286.598 * cos(2.65100655909  + 148.32254190981 *T)
    sumX0  +=  941982.751 * cos(1.66111566598  + 11.28951774740 *T)
    sumX0  +=  657433.967 * cos(0.57593488766  + 151.29148732640 *T)
    sumX0  +=  637676.334 * cos(4.21540624166  + 63.97971578690 *T)
    sumX0  +=  621475.723 * cos(3.05878846167  + 77.99436146740 *T)
    sumX0  +=  563553.759 * cos(4.80714363531  + 72.05647063421 *T)
    sumX0  +=  547585.874 * cos(3.63122642205  + 86.07111631471 *T)
    sumX0  +=  433273.658 * cos(5.66160493313  + 74.53778108379 *T)
    sumX0  +=  458904.544 * cos(3.90806753942  + 3.21276290011 *T)
    sumX0  +=  496011.875 * cos(0.59957126795  + 529.93478257810 *T)
    sumX0  +=  385773.622 * cos(6.18935243085  + 138.76131435421 *T)
    sumX0  +=  268150.781 * cos(0.96866014994  + 213.54291292150 *T)
    sumX0  +=  215803.817 * cos(5.30879023629  + 38.37685312130 *T)
    sumX0  +=  144768.728 * cos(2.31926691782  + 71.09326278771 *T)
    sumX0  +=  135315.426 * cos(5.51056178199  + 78.95756931390 *T)
    sumX0  +=  119632.236 * cos(4.10138480343  + 39.86132582961 *T)
    sumX0  +=  124987.110 * cos(2.51425439859  + 111.67397898031 *T)
    sumX0  +=  111232.552 * cos(5.12247819448  + 223.10414047710 *T)
    sumX0  +=  104563.861 * cos(3.90482142702  + 146.83806920150 *T)
    sumX0  +=  109354.782 * cos(4.45315493564  + 34.92027273770 *T)
    sumX0  +=   63579.167 * cos(0.29966151527  + 299.37021175271 *T)
    sumX0  +=   76241.847 * cos(2.53654184182  + 63.49208081989 *T)
    sumX0  +=   53591.167 * cos(3.94045434980  + 4.17597074660 *T)
    sumX0  +=   64420.500 * cos(3.72849455609  + 110.18950627200 *T)
    sumX0  +=   44566.570 * cos(0.46052856039  + 73.05330837550 *T)
    sumX0  +=   39200.127 * cos(2.68881030068  + 4.69723560840 *T)
    sumX0  +=   33660.673 * cos(2.51806070888  + 65.46418849521 *T)
    sumX0  +=   34338.030 * cos(3.03778732537  + 226.07308589371 *T)
    sumX0  +=   34546.984 * cos(1.84696372059  + 79.47883417571 *T)
    sumX0  +=   37552.003 * cos(4.14039081989  + 202.00957769060 *T)
    sumX0  +=   29546.886 * cos(6.00042096798  + 70.57199792590 *T)
    sumX0  +=   25683.031 * cos(5.30733054298  + 9.80504503910 *T)
    sumX0  +=   23447.563 * cos(4.09774770482  + 145.87486135500 *T)
    sumX0  +=   22948.537 * cos(5.51505103176  + 84.58664360640 *T)
    sumX0  +=   31820.127 * cos(5.53944193731  + 152.77596003471 *T)
    sumX0  +=   22832.266 * cos(2.29382969939  + 127.22797912329 *T)
    sumX0  +=   28378.722 * cos(6.01789844072  + 184.97110483931 *T)
    sumX0  +=   26655.018 * cos(6.11025984035  + 160.85271488200 *T)
    sumX0  +=   19671.485 * cos(5.53430548402  + 74.91354146621 *T)
    sumX0  +=   19648.859 * cos(2.28659171687  + 75.13729063540 *T)
    sumX0  +=   19896.586 * cos(0.57662115081  + 12.77399045571 *T)
    sumX0  +=   24486.297 * cos(1.99412405993  + 1.24065522479 *T)
    sumX0  +=   16842.912 * cos(0.47911269541  + 52.44638055600 *T)
    sumX0  +=   22087.252 * cos(4.59910433514  + 71.56883566720 *T)
    sumX0  +=   20068.191 * cos(4.47381667533  + 22.33521801130 *T)
    sumX0  +=   19913.902 * cos(1.39857362559  + 113.15845168861 *T)
    sumX0  +=   18351.147 * cos(5.69970669527  + 33.43580002939 *T)
    sumX0  +=   16887.001 * cos(0.21283887905  + 36.89238041300 *T)
    sumX0  +=   15173.078 * cos(2.88404559166  + 41.34579853790 *T)
    sumX0  +=   11242.410 * cos(6.11592492783  + 71.84402231310 *T)
    sumX0  +=   13948.685 * cos(6.27550136313  + 221.61966776881 *T)
    sumX0  +=   10809.036 * cos(1.69959188016  + 78.20680978850 *T)
    sumX0  +=   13591.810 * cos(2.55400587762  + 87.55558902300 *T)
    sumX0  +=   11997.310 * cos(0.94882794460  + 1059.62574767270 *T)
    sumX0  +=   12401.284 * cos(6.21959182172  + 72.57773549600 *T)
    sumX0  +=   11534.891 * cos(1.77255746394  + 77.47309660560 *T)
    sumX0  +=    9194.297 * cos(5.13684548839  + 145.35359649321 *T)
    sumX0  +=    8737.279 * cos(4.96955634864  + 186.45557754760 *T)
    sumX0  +=    9136.217 * cos(3.36421968897  + 127.71561409031 *T)
    sumX0  +=    9219.832 * cos(4.79646192202  + 108.21739859670 *T)
    sumX0  +=    7422.726 * cos(4.95959892966  + 108.70503356371 *T)
    sumX0  +=    7094.598 * cos(1.30379670348  + 297.88573904440 *T)
    sumX0  +=    6261.877 * cos(1.71362082374  + 153.73916788121 *T)
    sumX0  +=    7490.650 * cos(0.11367653768  + 426.84200835950 *T)
    sumX0  +=    7827.357 * cos(5.83003917523  + 340.52707456129 *T)
    sumX0  +=    6073.898 * cos(1.79808383323  + 149.31937965110 *T)
    sumX0  +=    5510.674 * cos(3.31059818622  + 140.24578706250 *T)
    sumX0  +=    6397.195 * cos(5.29054037494  + 62.49524307860 *T)
    sumX0  +=    5370.666 * cos(4.12500435843  + 75.54668091261 *T)
    sumX0  +=    5352.595 * cos(3.69257552001  + 74.50415118900 *T)
    sumX0  +=    4493.418 * cos(1.11410223517  + 66.94866120350 *T)
    sumX0  +=    4261.392 * cos(3.96170766253  + 265.74547599400 *T)
    sumX0  +=    5026.106 * cos(4.68825486100  + 18.40306474821 *T)
    sumX0  +=    4641.163 * cos(3.40278945922  + 36.40474544600 *T)
    sumX0  +=    4448.201 * cos(0.98197743832  + 183.48663213100 *T)
    sumX0  +=    3781.540 * cos(1.01094840328  + 10.80188278040 *T)
    sumX0  +=    4750.672 * cos(4.95760164342  + 74.06220820430 *T)
    sumX0  +=    4350.588 * cos(5.87369411626  + 70.08436295889 *T)
    sumX0  +=    4446.999 * cos(0.29417874414  + 114.64292439690 *T)
    sumX0  +=    3313.232 * cos(0.52117217508  + 83.10217089810 *T)
    sumX0  +=    4512.500 * cos(2.88455231924  + 75.98862389731 *T)
    sumX0  +=    3556.191 * cos(1.47647595802  + 6.18170831670 *T)
    sumX0  +=    3266.875 * cos(0.51825947223  + 220.65645992230 *T)
    sumX0  +=    3059.638 * cos(1.52537517070  + 52.93401552300 *T)
    sumX0  +=    3576.768 * cos(1.12142187124  + 137.27684164590 *T)
    sumX0  +=    2881.563 * cos(2.76127323120  + 374.15181032000 *T)
    sumX0  +=    3196.605 * cos(4.55588965060  + 76.02225379210 *T)
    sumX0  +=    2944.255 * cos(5.09789115860  + 7.35736448430 *T)
    sumX0  +=    3729.115 * cos(4.96516637585  + 144.86596152620 *T)
    sumX0  +=    2521.158 * cos(1.85148242342  + 202.49721265760 *T)
    sumX0  +=    2595.960 * cos(0.22431012776  + 97.11681657861 *T)
    sumX0  +=    2560.051 * cos(4.44250043160  + 80.44204202221 *T)
    sumX0  +=    2653.876 * cos(0.04778187588  + 305.10235190919 *T)
    sumX0  +=    2508.094 * cos(0.64297274411  + 31.95132732110 *T)
    sumX0  +=    1961.925 * cos(0.09387541521  + 20.85074530300 *T)
    sumX0  +=    2229.736 * cos(0.63549038852  + 80.96330688400 *T)
    sumX0  +=    2202.407 * cos(5.20629983391  + 2.93757625420 *T)
    sumX0  +=    1935.845 * cos(2.06583555933  + 75.07360016060 *T)
    sumX0  +=    1935.843 * cos(5.75516004795  + 74.97723194101 *T)
    sumX0  +=    2138.399 * cos(4.20897433466  + 74.76478361991 *T)
    sumX0  +=    2128.668 * cos(3.61197532719  + 75.28604848170 *T)
    sumX0  +=    2214.188 * cos(2.18592026080  + 259.75270340660 *T)
    sumX0  +=    1890.291 * cos(5.49936650879  + 300.85468446100 *T)
    sumX0  +=    2249.087 * cos(3.34771929967  + 138.27367938720 *T)
    sumX0  +=    1926.813 * cos(1.29258285090  + 159.36824217371 *T)
    sumX0  +=    1799.526 * cos(5.73284739685  + 74.86535735640 *T)
    sumX0  +=    1794.407 * cos(2.08815174296  + 75.18547474520 *T)
    sumX0  +=    1908.592 * cos(0.91765709548  + 69.08752521760 *T)
    sumX0  +=    2175.205 * cos(3.14101293062  + 131.64776735340 *T)
    sumX0  +=    1898.923 * cos(4.22364589697  + 206.42936592071 *T)
    sumX0  +=    1654.495 * cos(3.89578357216  + 277.27881122490 *T)
    sumX0  +=    1785.247 * cos(1.20558646138  + 56.86616878610 *T)
    sumX0  +=    1617.939 * cos(1.98998613398  + 2.69149803831 *T)
    sumX0  +=    1860.708 * cos(1.67531170047  + 42.83027124621 *T)
    sumX0  +=    2180.164 * cos(1.23556883050  + 479.04457143200 *T)
    sumX0  +=    1576.477 * cos(2.09193074602  + 190.96387742670 *T)
    sumX0  +=    1904.361 * cos(5.83571650342  + 15.22167101050 *T)
    sumX0  +=    2054.711 * cos(3.60264621769  + 834.79331700380 *T)
    sumX0  +=    1834.581 * cos(6.05975047580  + 106.73292588839 *T)
    sumX0  +=    1890.085 * cos(4.32104275484  + 154.26043274300 *T)
    sumX0  +=    1795.901 * cos(1.73393986559  + 227.55755860200 *T)
    sumX0  +=    1670.623 * cos(1.20607419790  + 220.13519506050 *T)
    sumX0  +=    1347.353 * cos(5.46539883456  + 288.32451148881 *T)
    sumX0  +=    1425.992 * cos(0.38776026293  + 93.18466331550 *T)
    sumX0  +=    1265.397 * cos(5.63850209574  + 404.26297286470 *T)
    sumX0  +=    1581.901 * cos(5.72246324321  + 39.60069339870 *T)
    sumX0  +=    1269.733 * cos(2.66547740633  + 142.69346761730 *T)
    sumX0  +=    1653.676 * cos(5.13281153725  + 181.51452445570 *T)
    sumX0  +=    1437.856 * cos(0.72853365638  + 522.82123557730 *T)
    sumX0  +=    1407.281 * cos(3.61754461658  + 537.04832957890 *T)
    sumX0  +=    1359.462 * cos(3.81802154525  + 62.00760811160 *T)
    sumX0  +=    1417.609 * cos(2.29697805296  + 235.63431344930 *T)
    sumX0  +=    1122.290 * cos(0.14112673022  + 60.04756252381 *T)
    sumX0  +=    1037.982 * cos(2.76177625247  + 5.66044345490 *T)
    sumX0  +=    1164.766 * cos(2.08226620054  + 82.13896305161 *T)
    sumX0  +=    1009.353 * cos(2.02387614580  + 74.27465652540 *T)
    sumX0  +=    1216.357 * cos(3.41823698867  + 212.05844021321 *T)
    sumX0  +=     996.588 * cos(5.79711941494  + 75.77617557621 *T)
    sumX0  +=     983.856 * cos(1.56441672942  + 129.20008679860 *T)
    sumX0  +=    1223.267 * cos(3.84379621710  + 187.94005025590 *T)
    sumX0  +=    1335.581 * cos(2.17715635407  + 379.88395047650 *T)
    sumX0  +=    1203.897 * cos(5.60031851266  + 67.91186905000 *T)
    sumX0  +=    1146.840 * cos(2.42164879835  + 296.40126633610 *T)
    sumX0  +=    1000.425 * cos(0.10902349258  + 8.32057233081 *T)
    sumX0  +=    1191.037 * cos(5.58759862804  + 50.64639366260 *T)
    sumX0  +=    1128.718 * cos(3.94259181404  + 230.32075334190 *T)
    sumX0  +=     930.244 * cos(3.76250779613  + 276.79117625789 *T)
    sumX0  +=     857.081 * cos(6.06089726967  + 68.43313391180 *T)
    sumX0  +=     912.408 * cos(1.69363364406  + 149.69514003350 *T)
    sumX0  +=    1170.925 * cos(2.02859165449  + 110.45013870291 *T)
    sumX0  +=     857.595 * cos(1.50487359225  + 90.00326957780 *T)
    sumX0  +=     989.091 * cos(2.61645612487  + 35.66854013560 *T)
    sumX0  +=    1054.223 * cos(5.53713388048  + 14.25846316400 *T)
    sumX0  +=     959.939 * cos(2.90123772360  + 147.83490694279 *T)
    sumX0  +=     957.455 * cos(4.68244742095  + 64.97655352819 *T)
    sumX0  +=     909.789 * cos(4.72919527392  + 149.91888920271 *T)
    sumX0  +=     943.313 * cos(4.77502421061  + 53.93085326429 *T)
    sumX0  +=     891.565 * cos(6.16624008939  + 107.22056085540 *T)
    sumX0  +=     775.480 * cos(2.00068126371  + 70.60562782070 *T)
    sumX0  +=     861.039 * cos(0.07017362216  + 143.86912378490 *T)
    sumX0  +=     851.667 * cos(2.07466532945  + 134.82916109110 *T)
    sumX0  +=     842.290 * cos(2.74294938604  + 3.68833577960 *T)
    sumX0  +=     893.222 * cos(3.38882447852  + 23.81969071961 *T)
    sumX0  +=     790.626 * cos(0.82969306173  + 7.83293736379 *T)
    sumX0  +=     683.506 * cos(1.80707483601  + 81.61769818981 *T)
    sumX0  +=     789.879 * cos(4.67531119051  + 35.40790770471 *T)
    sumX0  +=     669.882 * cos(1.56521412524  + 50.96190784770 *T)
    sumX0  +=     853.775 * cos(5.20540536482  + 200.52510498229 *T)
    sumX0  +=     698.863 * cos(3.36651984265  + 125.74350641500 *T)
    sumX0  +=     814.817 * cos(6.23437914876  + 143.38148881789 *T)
    sumX0  +=     862.909 * cos(5.03099273884  + 162.33718759030 *T)
    sumX0  +=     845.548 * cos(6.20605203929  + 203.49405039889 *T)
    sumX0  +=     797.191 * cos(2.22086891871  + 99.40443843901 *T)
    sumX0  +=     766.426 * cos(2.46431012048  + 68.80564092731 *T)
    sumX0  +=     609.673 * cos(4.95156945766  + 24.13520490470 *T)
    sumX0  +=     654.550 * cos(3.88721455407  + 109.70187130500 *T)
    sumX0  +=     783.834 * cos(5.30108995878  + 218.16308738520 *T)
    sumX0  +=     744.901 * cos(1.22651434994  + 760.01171843649 *T)
    sumX0  +=     643.611 * cos(1.27158746409  + 89.04006173130 *T)
    sumX0  +=     525.538 * cos(6.19695292650  + 128.71245183160 *T)
    sumX0  +=     703.451 * cos(5.35556821897  + 81.24519117430 *T)
    sumX0  +=     590.282 * cos(3.34460813817  + 69.60879007940 *T)
    sumX0  +=     528.305 * cos(2.28084002864  + 146.62562088040 *T)
    sumX0  +=     621.414 * cos(0.82570374928  + 68.59989025060 *T)
    sumX0  +=     677.104 * cos(5.66627508966  + 56.37853381910 *T)
    sumX0  +=     697.582 * cos(2.93128584821  + 617.56206830270 *T)
    sumX0  +=     597.125 * cos(2.20653215822  + 182.00215942271 *T)
    sumX0  +=     689.780 * cos(0.88947967570  + 2.72512793310 *T)
    sumX0  +=     578.345 * cos(2.48116102766  + 28.81562556571 *T)
    sumX0  +=     587.454 * cos(3.90812210311  + 152.25469517290 *T)
    sumX0  +=     670.562 * cos(3.06119354039  + 5.97595764000 *T)
    sumX0  +=     596.736 * cos(0.23847810121  + 116.18227885939 *T)
    sumX0  +=     480.108 * cos(4.14247181684  + 415.30867312860 *T)
    sumX0  +=     505.417 * cos(2.53324420248  + 157.88376946540 *T)
    sumX0  +=     551.335 * cos(1.14403413933  + 261.23717611490 *T)
    sumX0  +=     584.839 * cos(6.18920235362  + 258.26823069831 *T)
    sumX0  +=     571.647 * cos(1.12166883220  + 194.89603068979 *T)
    sumX0  +=     473.414 * cos(1.32843094724  + 134.34152612410 *T)
    sumX0  +=     551.129 * cos(5.62854491664  + 59.55992755679 *T)
    sumX0  +=     518.895 * cos(2.30620424068  + 352.06040979221 *T)
    sumX0  +=     444.205 * cos(3.75797687497  + 329.48137429739 *T)
    sumX0  +=     531.507 * cos(6.02129628860  + 54.41848823131 *T)
    sumX0  +=     449.312 * cos(1.03442910633  + 542.78046973539 *T)
    sumX0  +=     486.453 * cos(0.33128845125  + 73.65281792710 *T)
    sumX0  +=     437.062 * cos(1.29933406121  + 1589.31671276730 *T)
    sumX0  +=     488.485 * cos(2.40589686793  + 9.31741007210 *T)
    sumX0  +=     431.252 * cos(4.69836222247  + 41.40068029210 *T)
    sumX0  +=     435.419 * cos(3.76505789238  + 372.66733761171 *T)
    sumX0  +=     484.495 * cos(4.03440276126  + 209.12312469139 *T)
    sumX0  +=     418.899 * cos(1.97412934294  + 95.63234387030 *T)
    sumX0  +=     514.102 * cos(5.40821050729  + 116.12739710521 *T)
    sumX0  +=     388.869 * cos(5.96632177370  + 152.98840835580 *T)
    sumX0  +=     361.158 * cos(5.56439421115  + 120.11443212250 *T)
    sumX0  +=     371.870 * cos(5.67993821009  + 144.39038864671 *T)
    sumX0  +=     339.152 * cos(0.64656047679  + 46.45360796860 *T)
    sumX0  +=     373.018 * cos(3.47226296213  + 73.42906875790 *T)
    sumX0  +=     382.930 * cos(5.21294716380  + 114.38229196601 *T)
    sumX0  +=     426.834 * cos(2.46975859086  + 141.73025977080 *T)
    sumX0  +=     396.238 * cos(5.45517176193  + 98.91680347199 *T)
    sumX0  +=     396.936 * cos(0.12474175505  + 180.03005174739 *T)
    sumX0  +=     318.113 * cos(0.22602316909  + 33.38091827520 *T)
    sumX0  +=     325.989 * cos(4.18961191162  + 228.52076644850 *T)
    sumX0  +=     435.822 * cos(2.17402798911  + 37.88921815429 *T)
    sumX0  +=     392.309 * cos(1.55188964788  + 155.53915477459 *T)
    sumX0  +=     353.063 * cos(6.23671226488  + 215.02738562980 *T)
    sumX0  +=     410.664 * cos(4.20954401513  + 490.57790666290 *T)
    sumX0  +=     325.464 * cos(0.26699789914  + 7.66618102501 *T)
    sumX0  +=     299.890 * cos(3.75278581215  + 45.33283355520 *T)
    sumX0  +=     319.736 * cos(1.86158355595  + 30.46685461280 *T)
    sumX0  +=     278.968 * cos(2.37528917719  + 6.34846465550 *T)
    sumX0  +=     312.880 * cos(2.99684581751  + 419.72846135871 *T)
    sumX0  +=     386.565 * cos(5.46895169467  + 254.81165031470 *T)
    sumX0  +=     311.788 * cos(2.37646745038  + 147.35933406330 *T)
    sumX0  +=     300.229 * cos(3.46713251313  + 0.71939036300 *T)
    sumX0  +=     337.396 * cos(5.17753753648  + 150.77022246460 *T)
    sumX0  +=     253.789 * cos(1.48413722958  + 19.36627259471 *T)
    sumX0  +=     308.097 * cos(5.63479635439  + 640.14110379750 *T)
    sumX0  +=     252.403 * cos(0.29915428385  + 150.32827947990 *T)
    sumX0  +=     238.679 * cos(5.71090385224  + 553.82616999929 *T)
    sumX0  +=     266.427 * cos(1.01726382392  + 105.24845318010 *T)
    sumX0  +=     278.375 * cos(1.17449170388  + 29.44876501211 *T)
    sumX0  +=     306.043 * cos(1.75345400361  + 6283.31966747490 *T)
    sumX0  +=     265.739 * cos(5.77720505204  + 28.06735816780 *T)
    sumX0  +=     280.569 * cos(4.44867676294  + 126.23114138200 *T)
    sumX0  +=     300.771 * cos(1.99978996741  + 984.84414910540 *T)
    sumX0  +=     229.063 * cos(5.90121206501  + 33.92343499640 *T)
    sumX0  +=     226.799 * cos(5.35680022107  + 121.23520653590 *T)
    sumX0  +=     238.148 * cos(4.30420589870  + 2.20386307129 *T)
    sumX0  +=     198.420 * cos(4.76118534410  + 69.91760662011 *T)
    sumX0  +=     220.174 * cos(0.46900159784  + 44.31474395451 *T)
    sumX0  +=     220.857 * cos(5.01281205014  + 47.93808067690 *T)
    sumX0  +=     211.834 * cos(4.91315491890  + 0.76508234530 *T)
    sumX0  +=     257.979 * cos(5.01665522627  + 1289.70268353110 *T)
    sumX0  +=     240.510 * cos(4.63405186786  + 756.07956517339 *T)
    sumX0  +=     213.870 * cos(2.65681249594  + 316.63568714010 *T)
    sumX0  +=     231.462 * cos(4.35067672591  + 46.71424039951 *T)
    sumX0  +=     232.197 * cos(0.60568782891  + 339.04260185300 *T)
    sumX0  +=     214.419 * cos(2.75101540987  + 189.42452296421 *T)
    sumX0  +=     219.049 * cos(3.86758077039  + 213.05527795450 *T)
    sumX0  +=     241.635 * cos(1.97515700042  + 102.84895673509 *T)
    sumX0  +=     198.526 * cos(0.31867538279  + 61.01077037031 *T)
    sumX0  +=     237.145 * cos(1.57461819411  + 342.01154726960 *T)
    sumX0  +=     173.496 * cos(1.31235751133  + 116.66991382640 *T)
    sumX0  +=     194.828 * cos(2.42005289802  + 218.65072235220 *T)
    sumX0  +=     175.785 * cos(3.10552908409  + 57.49930823250 *T)
    sumX0  +=     167.277 * cos(0.90426589163  + 30.30009827400 *T)
    sumX0  +=     173.400 * cos(2.71724777154  + 681.29796660610 *T)
    sumX0  +=     214.483 * cos(2.48131614069  + 135.79236893760 *T)
    sumX0  +=     205.110 * cos(0.29778918320  + 216.67861467689 *T)
    sumX0  +=     168.458 * cos(3.08665425457  + 80.13322548150 *T)
    sumX0  +=     164.308 * cos(5.42212132676  + 467.99887116810 *T)
    sumX0  +=     203.489 * cos(1.11698481512  + 148.84380677161 *T)
    sumX0  +=     177.178 * cos(0.65075297933  + 103.59722413300 *T)
    sumX0  +=     165.274 * cos(0.65702158328  + 155.22364058950 *T)
    sumX0  +=     185.802 * cos(5.56639953504  + 74.13164417351 *T)
    sumX0  +=     185.802 * cos(2.25459555455  + 75.91918792810 *T)
    sumX0  +=     187.376 * cos(6.19888809652  + 149.28574975631 *T)
    sumX0  +=     153.285 * cos(1.88338215978  + 17.28229033479 *T)
    sumX0  +=     161.572 * cos(5.03486059863  + 264.26100328570 *T)
    sumX0  +=     185.820 * cos(1.04769820785  + 76.39801417450 *T)
    sumX0  +=     181.668 * cos(4.68053051052  + 455.15318401080 *T)
    sumX0  +=     175.514 * cos(2.84020133855  + 333.41352756050 *T)
    sumX0  +=     157.908 * cos(5.70734619944  + 82.44777959230 *T)
    sumX0  +=     169.475 * cos(4.34257763319  + 454.66554904380 *T)
    sumX0  +=     196.145 * cos(5.63627762153  + 291.46021324420 *T)
    sumX0  +=     177.970 * cos(5.10767427277  + 685.23011986920 *T)
    sumX0  +=     137.738 * cos(5.22315408886  + 448.93340888730 *T)
    sumX0  +=     152.620 * cos(3.60399534591  + 234.14984074100 *T)
    sumX0  +=     137.407 * cos(2.06690682589  + 16.91859203990 *T)
    sumX0  +=     150.365 * cos(6.22507689893  + 294.91679362781 *T)
    sumX0  +=     131.039 * cos(3.99587345412  + 156.39929675710 *T)
    sumX0  +=     180.038 * cos(3.58352924163  + 103.33659170210 *T)
    sumX0  +=     129.725 * cos(6.13605280603  + 76.72233708021 *T)
    sumX0  +=     142.122 * cos(4.21177012685  + 224.10097821840 *T)
    sumX0  +=     128.982 * cos(0.15395575367  + 130.68455950690 *T)
    sumX0  +=     146.399 * cos(4.63496132213  + 334.53430197390 *T)
    sumX0  +=     177.104 * cos(3.17890757818  + 10213.52936369450 *T)
    sumX0  +=     129.816 * cos(4.71661836260  + 92.55152386910 *T)
    sumX0  +=     171.451 * cos(4.58548960528  + 156.02678974161 *T)
    sumX0  +=     155.457 * cos(5.73264918423  + 347.64062156210 *T)
    sumX0  +=     134.222 * cos(2.12839585652  + 67.60305250931 *T)
    sumX0  +=     167.287 * cos(5.44444937854  + 80.75755620730 *T)
    sumX0  +=     126.618 * cos(4.54441723736  + 55.90296093960 *T)
    sumX0  +=     127.892 * cos(1.20412251150  + 61.69209392650 *T)
    sumX0  +=     146.027 * cos(6.20700368410  + 174.18603700630 *T)
    sumX0  +=     143.517 * cos(3.15339611893  + 23.87457247379 *T)
    sumX0  +=     139.674 * cos(6.08093937474  + 286.84003878050 *T)
    sumX0  +=     114.667 * cos(3.63547708752  + 13.08950464080 *T)
    sumX0  +=     114.401 * cos(1.06004300555  + 692.34366686999 *T)
    sumX0  +=     139.103 * cos(0.34979944278  + 120.60206708950 *T)
    sumX0  +=     124.306 * cos(2.81921095084  + 111.18634401329 *T)
    sumX0  +=     121.491 * cos(0.17904284099  + 79.69128249680 *T)
    sumX0  +=     106.720 * cos(0.32743265674  + 67.16110952460 *T)
    sumX0  +=     124.445 * cos(1.20141389812  + 141.89701610960 *T)
    sumX0  +=     150.445 * cos(3.19975890072  + 77.50672650040 *T)
    sumX0  +=     128.748 * cos(1.88192968286  + 254.69977573010 *T)
    sumX0  +=     135.145 * cos(2.72523252904  + 171.89841514590 *T)
    sumX0  +=     125.741 * cos(5.43013824571  + 98.60128928690 *T)
    sumX0  +=     106.860 * cos(1.67581101594  + 375.63628302830 *T)
    sumX0  +=     128.791 * cos(3.10899754760  + 155.74490545130 *T)
    sumX0  +=     104.092 * cos(1.66738591591  + 54.57854692570 *T)
    sumX0  +=     142.443 * cos(6.18206932531  + 203.98168536590 *T)
    sumX0  +=     109.153 * cos(3.31853315915  + 1.20702533000 *T)
    sumX0  +=     111.788 * cos(3.20988788864  + 272.85902299479 *T)
    sumX0  +=     125.849 * cos(0.73952694764  + 182.99899716400 *T)
    sumX0  +=     128.119 * cos(1.25401600728  + 256.78375799000 *T)
    sumX0  +=     101.034 * cos(1.92789553777  + 363.10611005610 *T)
    sumX0  +=     104.496 * cos(5.98306685829  + 131.16013238639 *T)
    sumX0  +=     135.112 * cos(5.89965158304  + 629.09540353360 *T)
    sumX0  +=      96.070 * cos(0.96004206819  + 258.63192899320 *T)
    sumX0  +=     102.131 * cos(0.32621877592  + 88.35873817511 *T)
    sumX0  +=     101.524 * cos(2.58765430562  + 51.44954281471 *T)
    sumX0  +=     114.745 * cos(5.56311265795  + 633.02755679670 *T)
    sumX0  +=     105.009 * cos(1.08881999355  + 105.73608814710 *T)
    sumX0  +=     106.158 * cos(4.21108150573  + 302.33915716930 *T)
    sumX0  +=     114.559 * cos(2.79618530706  + 604.71638114540 *T)
    sumX0  +=      97.687 * cos(3.08538191339  + 77.03115362090 *T)
    sumX0  +=      92.981 * cos(3.37398230043  + 94.14787116200 *T)
    sumX0  +=      92.956 * cos(3.04748798907  + 43.53284666180 *T)
    sumX0  +=      99.120 * cos(0.38719985960  + 149.54638218720 *T)
    sumX0  +=     116.555 * cos(3.87240492076  + 85.58348134769 *T)
    sumX0  +=      89.934 * cos(4.52722147760  + 149.85519872791 *T)
    sumX0  +=      89.708 * cos(4.70315697530  + 217.47506618460 *T)
    sumX0  +=     111.015 * cos(0.47609598811  + 253.32717760639 *T)
    sumX0  +=      92.676 * cos(1.28379383832  + 267.22994870229 *T)
    sumX0  +=     111.165 * cos(3.60094431144  + 19.88753745650 *T)
    sumX0  +=      89.583 * cos(2.61632627473  + 295.43805848960 *T)
    sumX0  +=      89.934 * cos(1.93336144673  + 149.75883050830 *T)
    sumX0  +=      89.829 * cos(0.03255236313  + 529.44714761109 *T)
    sumX0  +=     104.822 * cos(0.00646324626  + 433.95555536030 *T)
    sumX0  +=      97.705 * cos(6.18106020040  + 1052.51220067191 *T)
    sumX0  +=     107.796 * cos(1.21569499408  + 100.62827871640 *T)
    sumX0  +=      87.921 * cos(5.89750281232  + 74.39227660440 *T)
    sumX0  +=      87.921 * cos(1.92349227727  + 75.65855549721 *T)
    sumX0  +=      86.241 * cos(6.10142938790  + 76.66994745350 *T)
    sumX0  +=     115.960 * cos(2.63074702292  + 1214.92108496379 *T)
    sumX0  +=     112.324 * cos(0.51418420733  + 229.04203131030 *T)
    sumX0  +=      92.049 * cos(5.96349988177  + 6.86972951729 *T)
    sumX0  +=     106.860 * cos(3.13973519271  + 81.92651473050 *T)
    sumX0  +=      82.022 * cos(1.12030839416  + 82.88972257700 *T)
    sumX0  +=      83.480 * cos(4.54952756372  + 149.96707331250 *T)
    sumX0  +=      81.598 * cos(4.72010043651  + 146.35043423450 *T)
    sumX0  +=      91.518 * cos(0.69706776322  + 150.80385235940 *T)
    sumX0  +=      78.950 * cos(3.66249859834  + 4.86399194720 *T)
    sumX0  +=     107.950 * cos(0.01759566775  + 1162.71852189130 *T)
    sumX0  +=      80.076 * cos(2.03790828937  + 73.38088464811 *T)
    sumX0  +=      93.643 * cos(2.41659050778  + 99.14380600810 *T)
    sumX0  +=      99.120 * cos(6.07338306473  + 150.06764704901 *T)
    sumX0  +=      82.453 * cos(1.09950958918  + 7.14491616321 *T)
    sumX0  +=      76.677 * cos(1.12697999111  + 73.80157577341 *T)
    sumX0  +=      75.891 * cos(1.38853558210  + 219.17198721400 *T)
    sumX0  +=      74.887 * cos(0.96555468752  + 6.83609962250 *T)
    sumX0  +=     102.486 * cos(4.69327511390  + 210.57396750490 *T)
    sumX0  +=     100.859 * cos(0.96653178552  + 219.64756009350 *T)
    sumX0  +=      83.480 * cos(1.91105536061  + 149.64695592370 *T)
    sumX0  +=      83.970 * cos(4.85630817630  + 371.18286490340 *T)
    sumX0  +=      77.201 * cos(1.59554506373  + 26.26737127440 *T)
    sumX0  +=      73.643 * cos(6.15300341975  + 95.47228517591 *T)
    sumX0  +=     102.191 * cos(5.80314332396  + 328.10877617370 *T)
    sumX0  +=      80.117 * cos(1.27552322935  + 142.38465107660 *T)
    sumX0  +=      80.972 * cos(2.05291971561  + 67.11541754230 *T)
    sumX0  +=      71.543 * cos(3.15877821167  + 189.47940471839 *T)
    sumX0  +=      82.381 * cos(5.77613777190  + 58.35064149440 *T)
    sumX0  +=      90.613 * cos(6.17999848530  + 17.02165790390 *T)
    sumX0  +=      82.892 * cos(3.21408807945  + 175.40987728371 *T)
    sumX0  +=      69.078 * cos(0.68364339720  + 119.75073382760 *T)
    sumX0  +=      77.925 * cos(6.13231295495  + 191.45151239371 *T)
    sumX0  +=      78.846 * cos(0.01322136724  + 262.72164882321 *T)
    sumX0  +=      87.497 * cos(0.06139857943  + 894.59706204410 *T)
    sumX0  +=      70.199 * cos(4.41025707210  + 819.81546347680 *T)
    sumX0  +=      68.577 * cos(3.42724370622  + 15.79781762650 *T)
    sumX0  +=      76.274 * cos(3.42240995180  + 180.51768671441 *T)
    sumX0  +=      67.854 * cos(0.81718753405  + 606.51636803880 *T)
    sumX0  +=      77.388 * cos(4.76490704968  + 310.41591201661 *T)
    sumX0  +=      65.829 * cos(5.89081437901  + 75.82856520290 *T)
    sumX0  +=      65.829 * cos(1.93018071058  + 74.22226689870 *T)
    sumX0  +=      78.426 * cos(5.41224911130  + 50.90702609350 *T)
    sumX0  +=      71.406 * cos(6.22865606356  + 12.28635548869 *T)
    sumX0  +=      77.423 * cos(1.46638312463  + 70.29011363560 *T)
    sumX0  +=      69.368 * cos(3.96790790686  + 1066.73929467350 *T)
    sumX0  +=      80.391 * cos(4.69665260602  + 16.70614371880 *T)
    sumX0  +=      85.613 * cos(4.60014385205  + 173.92540457540 *T)
    sumX0  +=      72.675 * cos(1.00235377097  + 44.48150029329 *T)
    sumX0  +=      84.563 * cos(3.52774210442  + 55.41532597260 *T)
    sumX0  +=      62.255 * cos(0.31856134826  + 209.61075965840 *T)
    sumX0  +=      84.357 * cos(0.35171178066  + 331.56535655731 *T)
    sumX0  +=      70.498 * cos(4.37183477775  + 76.62176334371 *T)
    sumX0  +=      79.016 * cos(2.29035112753  + 477.56009872370 *T)
    sumX0  +=      82.673 * cos(3.24272077192  + 480.52904414029 *T)
    sumX0  +=      82.513 * cos(0.02775890883  + 192.44835013500 *T)
    sumX0  +=      80.452 * cos(2.41833994886  + 333.04982926560 *T)
    sumX0  +=      58.904 * cos(0.47702801943  + 24.62283987170 *T)
    sumX0  +=      72.407 * cos(2.95570874153  + 909.57491557110 *T)
    sumX0  +=      76.612 * cos(4.19475982131  + 82.72966388260 *T)
    sumX0  +=      75.510 * cos(3.63824154759  + 67.32116821901 *T)
    sumX0  +=      75.555 * cos(4.23824479480  + 15.74293587230 *T)
    sumX0  +=      70.086 * cos(2.85435934542  + 167.96626188280 *T)
    sumX0  +=      77.989 * cos(3.27374124663  + 70.35954960481 *T)
    sumX0  +=      66.531 * cos(4.44069547850  + 61.53203523210 *T)
    sumX0  +=      64.336 * cos(1.38579260364  + 178.54557903910 *T)
    sumX0  +=      56.673 * cos(4.77020753909  + 106.51798543980 *T)
    sumX0  +=      77.640 * cos(3.06328562010  + 71.40207932840 *T)
    sumX0  +=      62.431 * cos(1.58038361055  + 25.35904518210 *T)
    sumX0  +=      66.444 * cos(1.53111301715  + 125.42799222991 *T)
    sumX0  +=      56.738 * cos(5.97516509720  + 67.42423408300 *T)
    sumX0  +=      55.917 * cos(4.46821485097  + 949.41942645330 *T)
    sumX0  +=      76.032 * cos(4.77182769338  + 60.52313540329 *T)
    sumX0  +=      59.000 * cos(0.40199703694  + 402.77850015639 *T)
    sumX0  +=      73.138 * cos(5.40480984208  + 26.84351789039 *T)
    sumX0  +=      54.636 * cos(1.47333332256  + 72.10465474401 *T)
    sumX0  +=      64.610 * cos(4.53004325815  + 471.93102443120 *T)
    sumX0  +=      53.942 * cos(4.57598306501  + 117.66675156770 *T)
    sumX0  +=      54.371 * cos(3.70705139945  + 341.01470952831 *T)
    sumX0  +=      54.836 * cos(6.00184655368  + 3.42521122121 *T)
    sumX0  +=      72.732 * cos(4.75709331353  + 78.64875277320 *T)
    sumX0  +=      69.223 * cos(6.00657895059  + 8.10812400971 *T)
    sumX0  +=      70.778 * cos(0.72528535802  + 237.43430034270 *T)
    sumX0  +=      53.991 * cos(4.58226336725  + 149.05625509271 *T)
    sumX0  +=      51.866 * cos(0.06447646319  + 77.94617735760 *T)
    sumX0  +=      60.750 * cos(0.05642817034  + 79.76071846601 *T)
    sumX0  +=      69.434 * cos(1.15052233162  + 256.29612302300 *T)
    sumX0  +=      52.587 * cos(2.71406262711  + 830.86116374070 *T)
    sumX0  +=      56.127 * cos(4.11138293624  + 65.62424718960 *T)
    sumX0  +=      53.925 * cos(2.05138349837  + 13.73719830221 *T)
    sumX0  +=      50.216 * cos(5.16661515376  + 195.38366565681 *T)
    sumX0  +=      57.142 * cos(2.07134229657  + 31.89644556690 *T)
    sumX0  +=      55.249 * cos(1.26051053222  + 75.39792306630 *T)
    sumX0  +=      55.249 * cos(0.27729925019  + 74.65290903530 *T)
    sumX0  +=      49.301 * cos(0.68505345474  + 76.24925632820 *T)
    sumX0  +=      60.539 * cos(0.62920341226  + 289.97574053589 *T)
    sumX0  +=      64.921 * cos(3.42395602826  + 846.32665223470 *T)
    sumX0  +=      59.034 * cos(3.72320060052  + 27.33115285740 *T)
    sumX0  +=      51.079 * cos(1.42491352478  + 0.35569206810 *T)
    sumX0  +=      51.079 * cos(4.85827178240  + 0.13194289890 *T)
    sumX0  +=      56.906 * cos(1.59844453309  + 143.21473247911 *T)
    sumX0  +=      51.926 * cos(4.71905052001  + 185.23173727020 *T)
    sumX0  +=      48.626 * cos(2.61494512788  + 49.16192095430 *T)
    sumX0  +=      51.388 * cos(2.81765408295  + 42.88515300039 *T)
    sumX0  +=      64.661 * cos(5.92230270405  + 910.06255053810 *T)
    sumX0  +=      48.885 * cos(2.28326794620  + 91.70019060721 *T)
    sumX0  +=      50.130 * cos(2.68135977613  + 102.11275142471 *T)
    sumX0  +=      47.590 * cos(4.15725769893  + 224.47673860080 *T)
    sumX0  +=      57.461 * cos(1.13975618226  + 303.61787920090 *T)
    sumX0  +=      50.603 * cos(4.72637033906  + 35.18090516859 *T)
    sumX0  +=      48.745 * cos(2.56799692881  + 201.01273994931 *T)
    sumX0  +=      58.382 * cos(1.14571117545  + 486.15811843280 *T)
    sumX0  +=      49.531 * cos(1.94026883146  + 44.96913526031 *T)
    sumX0  +=      54.462 * cos(1.20363826712  + 237.11878615761 *T)
    sumX0  +=      54.466 * cos(4.07199773375  + 75.67980735660 *T)
    sumX0  +=      54.466 * cos(3.74899735584  + 74.37102474501 *T)
    sumX0  +=      60.438 * cos(1.89700849107  + 266.23311096100 *T)
    sumX0  +=      48.443 * cos(0.85823284795  + 178.11819026941 *T)
    sumX0  +=      55.698 * cos(2.09947437284  + 278.76328393321 *T)
    sumX0  +=      44.294 * cos(5.09757382517  + 57.75994066341 *T)
    sumX0  +=      48.861 * cos(0.49608314902  + 122.71967924421 *T)
    sumX0  +=      46.778 * cos(4.39128238600  + 14.47091148511 *T)
    sumX0  +=      52.405 * cos(1.63380007153  + 58.98378094080 *T)
    sumX0  +=      47.669 * cos(3.70988562108  + 84.42658491201 *T)
    sumX0  +=      59.409 * cos(4.21497001188  + 117.61186981350 *T)
    sumX0  +=      42.464 * cos(1.91841049036  + 150.55777414350 *T)
    sumX0  +=      42.527 * cos(5.16925942986  + 69.39634175830 *T)
    sumX0  +=      44.136 * cos(4.05039656128  + 554.31380496631 *T)
    sumX0  +=      56.899 * cos(2.10899559438  + 25.30416342790 *T)
    sumX0  +=      53.440 * cos(4.60001445986  + 275.30670354960 *T)
    sumX0  +=      48.371 * cos(1.92839725116  + 56.01483552421 *T)
    sumX0  +=      51.224 * cos(4.05673094488  + 1364.48428209840 *T)
    sumX0  +=      44.203 * cos(6.07482944330  + 70.83263035680 *T)
    sumX0  +=      41.205 * cos(4.61128278499  + 124.25903370670 *T)
    sumX0  +=      41.202 * cos(2.82827687638  + 49.47743513939 *T)
    sumX0  +=      53.218 * cos(1.97620392270  + 7.46043034830 *T)
    sumX0  +=      39.879 * cos(3.39793772882  + 48.51422729289 *T)
    sumX0  +=      39.626 * cos(3.19265463758  + 72.73779419041 *T)
    sumX0  +=      42.269 * cos(4.84099240112  + 232.66536803271 *T)
    sumX0  +=      39.724 * cos(3.77191165523  + 88.51879686950 *T)
    sumX0  +=      54.178 * cos(5.97029601300  + 364.75733910320 *T)
    sumX0  +=      47.398 * cos(0.63156941487  + 74.17408278890 *T)
    sumX0  +=      47.398 * cos(0.90624036754  + 75.87674931271 *T)
    sumX0  +=      45.255 * cos(3.53774254616  + 393.21727260080 *T)
    sumX0  +=      41.021 * cos(1.74616565416  + 79.21820174480 *T)
    sumX0  +=      40.961 * cos(2.22598462877  + 306.58682461749 *T)
    sumX0  +=      48.986 * cos(1.09858650309  + 198.07742442750 *T)
    sumX0  +=      42.523 * cos(4.44089809382  + 22.39009976550 *T)
    sumX0  +=      37.414 * cos(2.64496729780  + 397.14942586389 *T)
    sumX0  +=      43.272 * cos(4.08248091850  + 28.32799059869 *T)
    sumX0  +=      47.477 * cos(0.90974768312  + 224.70048777000 *T)
    sumX0  +=      47.665 * cos(5.63221090280  + 73.32849502140 *T)
    sumX0  +=      40.139 * cos(4.02740827147  + 164.78486814511 *T)
    sumX0  +=      37.770 * cos(0.39022145039  + 91.48774228611 *T)
    sumX0  +=      37.294 * cos(5.02963292331  + 275.79433851660 *T)
    sumX0  +=      39.829 * cos(3.07388783910  + 28.98238190449 *T)
    sumX0  +=      42.462 * cos(1.60057894892  + 25.61967761299 *T)
    sumX0  +=      38.602 * cos(1.73856338598  + 91.80325647120 *T)
    sumX0  +=      49.996 * cos(4.52551797919  + 52.13086637091 *T)
    sumX0  +=      37.120 * cos(3.30294172342  + 167.33312243640 *T)
    sumX0  +=      35.657 * cos(4.61504038625  + 77.31303791120 *T)
    sumX0  +=      46.968 * cos(5.66194598133  + 836.27778971210 *T)
    sumX0  +=      36.770 * cos(6.23330863408  + 123.78346082721 *T)
    sumX0  +=      45.438 * cos(0.21594288443  + 199.04063227400 *T)
    sumX0  +=      48.093 * cos(6.20387971062  + 3340.85624418330 *T)
    sumX0  +=      34.133 * cos(5.28355572528  + 154.91482404881 *T)
    sumX0  +=      40.987 * cos(5.00683283281  + 411.37651986550 *T)
    sumX0  +=      37.167 * cos(6.06382239637  + 227.03629374021 *T)
    sumX0  +=      35.365 * cos(5.35602526661  + 183.85033042589 *T)
    sumX0  +=      34.349 * cos(1.22331331983  + 58.56308981550 *T)
    sumX0  +=      44.963 * cos(4.68970697967  + 833.30884429549 *T)
    sumX0  +=      33.485 * cos(1.71788340603  + 227.77000692311 *T)
    sumX0  +=      32.838 * cos(2.93232570660  + 69.12115511239 *T)
    sumX0  +=      37.229 * cos(2.21578270703  + 103.76398047180 *T)
    sumX0  +=      33.396 * cos(2.69897029979  + 92.29089143820 *T)
    sumX0  +=      43.914 * cos(1.40663545062  + 293.43232091950 *T)
    sumX0  +=      44.941 * cos(1.33050790583  + 292.94468595250 *T)
    sumX0  +=      37.327 * cos(3.07443834918  + 0.55933166859 *T)
    sumX0  +=      32.513 * cos(0.07648721797  + 136.31363379940 *T)
    sumX0  +=      34.574 * cos(0.03012597295  + 35.92917256650 *T)
    sumX0  +=      41.936 * cos(3.73124645338  + 92.06388890210 *T)
    sumX0  +=      33.257 * cos(1.06857797769  + 248.96763557360 *T)
    sumX0  +=      44.206 * cos(1.31564656912  + 767.61290040431 *T)
    sumX0  +=      33.118 * cos(2.87749681035  + 284.39235822571 *T)
    sumX0  +=      34.173 * cos(0.10260218436  + 133.13224006171 *T)
    sumX0  +=      35.868 * cos(0.01050739027  + 179.91817716280 *T)
    sumX0  +=      37.508 * cos(5.59184408484  + 94.66913602380 *T)
    sumX0  +=      41.561 * cos(3.77043120506  + 163.82166029861 *T)
    sumX0  +=      34.188 * cos(3.60117869846  + 336.01877468221 *T)
    sumX0  +=      33.991 * cos(0.79141238881  + 32.43896228810 *T)
    sumX0  +=      30.436 * cos(5.91304525500  + 31.78457098230 *T)
    sumX0  +=      34.876 * cos(1.56285963223  + 215.19414196860 *T)
    sumX0  +=      37.259 * cos(6.20129626897  + 91.06705116081 *T)
    sumX0  +=      30.353 * cos(0.09307819419  + 32.48714639791 *T)
    sumX0  +=      31.931 * cos(0.43513268629  + 205.94173095370 *T)
    sumX0  +=      30.008 * cos(2.49973464553  + 745.03386490949 *T)
    sumX0  +=      36.558 * cos(2.58993639058  + 45.96597300160 *T)
    sumX0  +=      29.428 * cos(0.89723240240  + 270.16526422410 *T)
    sumX0  +=      36.881 * cos(4.37161776762  + 75.61611688180 *T)
    sumX0  +=      36.881 * cos(3.44937732197  + 74.43471521981 *T)
    sumX0  +=      39.559 * cos(1.15354419544  + 58.87190635621 *T)
    sumX0  +=      38.377 * cos(0.95780009117  + 39.37369086260 *T)
    sumX0  +=      32.329 * cos(2.26521239763  + 80.65449034331 *T)
    sumX0  +=      40.007 * cos(0.14301803428  + 1744.61205005840 *T)
    sumX0  +=      29.388 * cos(0.57900535447  + 530.89799042461 *T)
    sumX0  +=      31.193 * cos(5.34416408985  + 9.15065373330 *T)
    sumX0  +=      34.894 * cos(4.92668697632  + 143.58723949460 *T)
    sumX0  +=      29.512 * cos(6.10629740660  + 958.33296034749 *T)
    sumX0  +=      29.255 * cos(2.41173436056  + 415.79630809560 *T)
    sumX0  +=      29.888 * cos(5.08597902881  + 163.14033674240 *T)
    sumX0  +=      32.864 * cos(5.89146148240  + 94.03599657740 *T)
    sumX0  +=      35.403 * cos(1.30282799370  + 31.73638687251 *T)
    sumX0  +=      36.501 * cos(0.77005331472  + 142.90591593840 *T)
    sumX0  +=      32.416 * cos(1.28513183872  + 114.69780615110 *T)
    sumX0  +=      28.338 * cos(3.75891762209  + 528.97157473160 *T)
    sumX0  +=      31.160 * cos(5.81959220361  + 118.15438653471 *T)
    sumX0  +=      34.324 * cos(5.62039019983  + 57.38743364790 *T)
    sumX0  +=      33.749 * cos(5.83827641206  + 42.78208713641 *T)
    sumX0  +=      34.375 * cos(0.78267663790  + 326.62430346539 *T)
    sumX0  +=      29.984 * cos(0.03281857450  + 90.52453443961 *T)
    sumX0  +=      35.105 * cos(0.74094997285  + 230.80838830890 *T)
    sumX0  +=      31.999 * cos(5.07107605631  + 216.51185833811 *T)
    sumX0  +=      30.853 * cos(2.28294479233  + 8.84183719261 *T)
    sumX0  +=      34.254 * cos(5.04640211853  + 228.83628063359 *T)
    sumX0  +=      27.044 * cos(3.03125650197  + 27.96429230381 *T)
    sumX0  +=      27.412 * cos(4.74407451560  + 221.40721944771 *T)
    sumX0  +=      26.388 * cos(4.77835419912  + 45.82046852221 *T)
    sumX0  +=      32.494 * cos(1.78574998112  + 1033.11455891480 *T)
    sumX0  +=      35.994 * cos(4.20173334239  + 4.49148493169 *T)
    sumX0  +=      29.345 * cos(5.38380888199  + 123.29582586020 *T)
    sumX0  +=      26.034 * cos(5.84327450882  + 78.46993434690 *T)
    sumX0  +=      26.650 * cos(4.89722523811  + 151.50393564750 *T)
    sumX0  +=      31.030 * cos(3.60400127791  + 66.42739634171 *T)
    sumX0  +=      28.238 * cos(5.31937121882  + 222.61650551010 *T)
    sumX0  +=      26.518 * cos(4.41892897104  + 11.40139233200 *T)
    sumX0  +=      26.014 * cos(6.23229383455  + 447.44893617900 *T)
    sumX0  +=      25.836 * cos(5.11134834182  + 68.54500849640 *T)
    sumX0  +=      24.852 * cos(2.56533658439  + 42.99702758500 *T)
    sumX0  +=      33.654 * cos(0.21042031029  + 54.53036281590 *T)
    sumX0  +=      24.448 * cos(0.97333931967  + 269.67762925710 *T)
    sumX0  +=      33.190 * cos(1.90416729889  + 141.94270809190 *T)
    sumX0  +=      24.034 * cos(4.76114773513  + 327.99690158910 *T)
    sumX0  +=      31.856 * cos(0.25852732885  + 1140.13948639650 *T)
    sumX0  +=      26.275 * cos(5.82738497494  + 39.91620758380 *T)
    sumX0  +=      25.887 * cos(1.29578545514  + 2.04380437690 *T)
    sumX0  +=      23.665 * cos(3.67797605271  + 21.84758304429 *T)
    sumX0  +=      25.701 * cos(5.62367875742  + 73.91345035801 *T)
    sumX0  +=      25.701 * cos(2.19731633217  + 76.13738174360 *T)
    sumX0  +=      24.158 * cos(0.24782459423  + 4.20960064140 *T)
    sumX0  +=      32.085 * cos(0.52321152462  + 368.21391948681 *T)
    sumX0  +=      25.732 * cos(5.22545182629  + 3.37951923889 *T)
    sumX0  +=      24.270 * cos(5.33004436579  + 130.19692453990 *T)
    sumX0  +=      23.775 * cos(4.38064538443  + 969.37866061139 *T)
    sumX0  +=      22.911 * cos(1.37498134000  + 117.56368570370 *T)
    sumX0  +=      27.148 * cos(1.36880911424  + 11.17764316280 *T)
    sumX0  +=      31.420 * cos(4.91847220813  + 619.04654101100 *T)
    sumX0  +=      30.627 * cos(0.38544249631  + 91.17892574540 *T)
    sumX0  +=      29.507 * cos(1.92973953754  + 565.35950523021 *T)
    sumX0  +=      25.369 * cos(1.20472379976  + 81.13006322279 *T)
    sumX0  +=      27.172 * cos(4.46412260278  + 25.51661174900 *T)
    sumX0  +=      26.354 * cos(4.80557558326  + 204.97852310719 *T)
    sumX0  +=      22.571 * cos(3.73766875284  + 70.88081446661 *T)
    sumX0  +=      22.904 * cos(2.70921472298  + 81.50582360521 *T)
    sumX0  +=      30.825 * cos(3.40941027630  + 1439.75351563271 *T)
    sumX0  +=      25.426 * cos(3.23853983642  + 133.34468838281 *T)
    sumX0  +=      27.083 * cos(5.66221904271  + 105.08169684130 *T)
    sumX0  +=      23.195 * cos(5.23000166632  + 90.57941619379 *T)
    sumX0  +=      22.062 * cos(1.70024849278  + 189.16389053330 *T)
    sumX0  +=      22.242 * cos(0.31042361770  + 86.18299089930 *T)
    sumX0  +=      23.869 * cos(5.80511989792  + 413.82420042030 *T)
    sumX0  +=      21.820 * cos(5.43668576549  + 34.86539098351 *T)
    sumX0  +=      29.143 * cos(3.98193402828  + 616.07759559439 *T)
    sumX0  +=      26.632 * cos(5.56229885820  + 100.88891114730 *T)
    sumX0  +=      23.677 * cos(2.06825442969  + 541.29599702710 *T)
    sumX0  +=      30.371 * cos(6.14019922693  + 401.40590203270 *T)
    sumX0  +=      23.717 * cos(5.53655558593  + 45.79921666280 *T)
    sumX0  +=      27.663 * cos(2.55223795831  + 405.74744557300 *T)
    sumX0  +=      24.084 * cos(4.27906130982  + 170.41394243761 *T)
    sumX0  +=      25.601 * cos(2.80411540512  + 31.24875190549 *T)
    sumX0  +=      22.288 * cos(2.36593798640  + 108.16251684250 *T)
    sumX0  +=      22.857 * cos(3.29724344840  + 104.23036357940 *T)
    sumX0  +=      28.908 * cos(1.50679345449  + 329.59324888200 *T)
    sumX0  +=      24.430 * cos(0.22851414496  + 49.42255338521 *T)
    sumX0  +=      23.008 * cos(3.64071080504  + 217.16624964391 *T)
    sumX0  +=      20.835 * cos(0.36990037219  + 141.20899490900 *T)
    sumX0  +=      22.817 * cos(3.61760173513  + 145.38722638800 *T)
    sumX0  +=      20.957 * cos(4.08353082157  + 79.17001763500 *T)
    sumX0  +=      27.954 * cos(2.22651254688  + 68.17250148091 *T)
    sumX0  +=      23.022 * cos(5.92938841563  + 162.65270177540 *T)
    sumX0  +=      25.411 * cos(5.76699581868  + 41.88831525910 *T)
    sumX0  +=      25.079 * cos(2.20069442464  + 92.66339845371 *T)
    sumX0  +=      22.779 * cos(5.52671292313  + 107.74182571721 *T)
    sumX0  +=      25.092 * cos(5.30025519045  + 87.87110320809 *T)
    sumX0  +=      27.016 * cos(2.72752005167  + 0.99457700890 *T)
    sumX0  +=      20.414 * cos(5.00634045533  + 222.14093263061 *T)
    sumX0  +=      20.185 * cos(1.39247008619  + 144.17794032560 *T)
    sumX0  +=      27.094 * cos(2.53571464342  + 369.69839219510 *T)
    sumX0  +=      19.466 * cos(0.94793502304  + 9.15735137770 *T)
    sumX0  +=      20.822 * cos(2.59012459788  + 72.31710306511 *T)
    sumX0  +=      20.822 * cos(5.23087049171  + 77.73372903651 *T)
    sumX0  +=      23.312 * cos(4.27157982827  + 52.39149880180 *T)
    sumX0  +=      19.238 * cos(3.43567333582  + 145.14114817210 *T)
    sumX0  +=      22.156 * cos(0.30949345254  + 64.73047531231 *T)
    sumX0  +=      19.648 * cos(5.73502411840  + 692.83130183700 *T)
    sumX0  +=      18.748 * cos(2.81185581123  + 289.80898419710 *T)
    sumX0  +=      19.125 * cos(6.24717703663  + 22.65073219640 *T)
    sumX0  +=      24.581 * cos(4.62574787093  + 20.20305164159 *T)
    sumX0  +=      21.617 * cos(1.70409750584  + 251.84270489810 *T)
    sumX0  +=      21.860 * cos(3.58883523958  + 38.11622069041 *T)
    sumX0  +=      21.658 * cos(1.25465688439  + 30.98811947460 *T)
    sumX0  +=      21.231 * cos(2.27353854359  + 55.38169607781 *T)
    sumX0  +=      25.483 * cos(4.64764517239  + 41.08516610701 *T)
    sumX0  +=      21.046 * cos(0.38705533625  + 58.07545484850 *T)
    sumX0  +=      22.854 * cos(1.45857912338  + 167.07249000551 *T)
    sumX0  +=      18.322 * cos(4.94540136145  + 165.84864972810 *T)
    sumX0  +=      18.466 * cos(5.32494059479  + 380.84715832300 *T)
    sumX0  +=      24.739 * cos(5.59448807386  + 81.87833062070 *T)
    sumX0  +=      22.782 * cos(3.12770425841  + 305.58998687620 *T)
    sumX0  +=      18.449 * cos(2.20918468220  + 378.92074262999 *T)
    sumX0  +=      20.493 * cos(2.67165332200  + 148.21066732521 *T)
    sumX0  +=      24.545 * cos(3.35975251109  + 66.46102623650 *T)
    sumX0  +=      20.079 * cos(4.42522434046  + 156.92056161890 *T)
    sumX0  +=      17.465 * cos(2.28393103769  + 312.21589891000 *T)
    sumX0  +=      21.567 * cos(1.12414642509  + 225.55182103190 *T)
    sumX0  +=      20.030 * cos(3.79354487434  + 134.30789622931 *T)
    sumX0  +=      17.391 * cos(4.43131947107  + 73.01967848070 *T)
    sumX0  +=      19.537 * cos(2.39762567547  + 140.41254340130 *T)
    sumX0  +=      22.517 * cos(0.52473596856  + 248.70700314271 *T)
    sumX0  +=      17.589 * cos(0.35927102062  + 303.30236501580 *T)
    sumX0  +=      20.366 * cos(3.56741326841  + 97.65933329981 *T)
    sumX0  +=      22.843 * cos(2.35344003630  + 758.52724572820 *T)
    sumX0  +=      21.264 * cos(1.64752588071  + 330.08088384900 *T)
    sumX0  +=      16.980 * cos(2.73160214009  + 74.56784166380 *T)
    sumX0  +=      16.980 * cos(5.08939294950  + 75.48299043781 *T)
    sumX0  +=      19.715 * cos(5.18582165447  + 531.73476947149 *T)
    sumX0  +=      21.989 * cos(5.89750482501  + 136.78920667889 *T)
    sumX0  +=      21.616 * cos(0.96280708434  + 10.32630990090 *T)
    sumX0  +=      16.752 * cos(4.31140364382  + 19.25439801010 *T)
    sumX0  +=      21.941 * cos(2.87264335658  + 624.67561530350 *T)
    sumX0  +=      16.246 * cos(1.61745332834  + 2119.00767786191 *T)
    sumX0  +=      16.279 * cos(4.16287111598  + 1.12878064019 *T)
    sumX0  +=      16.279 * cos(1.02127846239  + 1.61641560721 *T)
    sumX0  +=      22.328 * cos(3.27882453633  + 97.91996573070 *T)
    sumX0  +=      19.188 * cos(4.94451945419  + 14.31334491820 *T)
    sumX0  +=      22.519 * cos(0.85578283496  + 165.36101476110 *T)
    sumX0  +=      20.956 * cos(1.59412571256  + 330.96584700570 *T)
    sumX0  +=      18.599 * cos(5.83146793262  + 148.43441649440 *T)
    sumX0  +=      18.917 * cos(2.54214892451  + 255.29928528170 *T)
    sumX0  +=      21.299 * cos(1.30861186159  + 95.52046928571 *T)
    sumX0  +=      19.398 * cos(0.43941419672  + 408.19512612779 *T)
    sumX0  +=      21.480 * cos(4.04620670995  + 1669.83045149109 *T)
    sumX0  +=      15.730 * cos(0.96946804568  + 17.88179988640 *T)
    sumX0  +=      15.989 * cos(2.45562453742  + 132.85705341579 *T)
    sumX0  +=      16.213 * cos(1.36728837925  + 5.17280848790 *T)
    sumX0  +=      15.494 * cos(2.11918907538  + 743.23387801611 *T)
    sumX0  +=      15.346 * cos(2.13158900960  + 610.44852130189 *T)
    sumX0  +=      19.338 * cos(0.96053560010  + 363.27286639489 *T)
    sumX0  +=      17.882 * cos(1.66188033865  + 190.90899567250 *T)
    sumX0  +=      16.009 * cos(5.51847022593  + 74.19533464831 *T)
    sumX0  +=      16.009 * cos(2.30252486366  + 75.85549745330 *T)
    sumX0  +=      18.450 * cos(3.22255276270  + 135.57992061650 *T)
    sumX0  +=      15.066 * cos(1.89059622287  + 118.26626111931 *T)
    sumX0  +=      19.811 * cos(2.43871507941  + 375.95179721340 *T)
    sumX0  +=      17.887 * cos(1.19130006535  + 85.32035678930 *T)
    sumX0  +=      15.552 * cos(1.65433821047  + 5.87289177600 *T)
    sumX0  +=      15.552 * cos(1.65433821047  + 5.38525680900 *T)
    sumX0  +=      14.946 * cos(4.66006612006  + 76.46170464930 *T)
    sumX0  +=      14.946 * cos(0.01933631594  + 73.58912745230 *T)
    sumX0  +=      18.660 * cos(1.95091917300  + 107.26874496520 *T)
    sumX0  +=      14.573 * cos(3.69875469397  + 73.49275923270 *T)
    sumX0  +=      14.573 * cos(0.98064774203  + 76.55807286891 *T)
    sumX0  +=      15.019 * cos(0.65840143476  + 64.09159037150 *T)
    sumX0  +=      17.892 * cos(2.82013610923  + 231.80522605020 *T)
    sumX0  +=      15.543 * cos(3.30320904812  + 66.11857980100 *T)
    sumX0  +=      18.436 * cos(0.47890956330  + 549.89401673620 *T)
    sumX0  +=      15.371 * cos(5.73614840902  + 736.12033101530 *T)
    sumX0  +=      17.668 * cos(3.18092258141  + 378.39947776819 *T)
    sumX0  +=      14.688 * cos(4.10270741432  + 17.91542978119 *T)
    sumX0  +=      14.089 * cos(1.33160825911  + 63.86784120231 *T)
    sumX0  +=      15.759 * cos(2.23574097001  + 104.25161543881 *T)
    sumX0  +=      16.874 * cos(5.09404028824  + 84.09900863940 *T)
    sumX0  +=      15.468 * cos(0.86301025949  + 515.70768857651 *T)
    sumX0  +=      17.006 * cos(0.51683233019  + 106.25735300890 *T)
    sumX0  +=      15.683 * cos(1.71904566531  + 46.81730626350 *T)
    sumX0  +=      15.035 * cos(3.00271084463  + 490.09027169589 *T)
    sumX0  +=      13.767 * cos(4.61271175472  + 535.66692273460 *T)
    sumX0  +=      14.034 * cos(3.68802555946  + 281.21096448800 *T)
    sumX0  +=      13.418 * cos(3.63644963778  + 70.02948120471 *T)
    sumX0  +=      13.418 * cos(4.18460952894  + 80.02135089690 *T)
    sumX0  +=      18.602 * cos(1.68483266458  + 366.24181181149 *T)
    sumX0  +=      16.062 * cos(2.15259310684  + 73.28031091161 *T)
    sumX0  +=      16.062 * cos(2.52680932916  + 76.77052119001 *T)
    sumX0  +=      18.697 * cos(4.35467835660  + 130.16329464510 *T)
    sumX0  +=      13.830 * cos(5.82941711627  + 278.27564896620 *T)
    sumX0  +=      16.797 * cos(3.58460178095  + 157.51126244990 *T)
    sumX0  +=      16.797 * cos(0.44300912736  + 157.02362748290 *T)
    sumX0  +=      13.389 * cos(0.37621077797  + 118.62995941420 *T)
    sumX0  +=      14.382 * cos(2.64523963261  + 194.53233239490 *T)
    sumX0  +=      17.090 * cos(5.34605161061  + 60.79832204921 *T)
    sumX0  +=      13.114 * cos(4.17991617574  + 883.55136178020 *T)
    sumX0  +=      16.641 * cos(0.22443011643  + 118.31444522910 *T)
    sumX0  +=      13.074 * cos(4.16045920009  + 12.25272559390 *T)
    sumX0  +=      13.153 * cos(2.76878972668  + 225.10987804721 *T)
    sumX0  +=      13.157 * cos(4.80246349684  + 122.08653979781 *T)
    sumX0  +=      16.504 * cos(0.88543342905  + 285.35556607221 *T)
    sumX0  +=      13.716 * cos(1.39024809565  + 1096.85045721820 *T)
    sumX0  +=      12.748 * cos(2.48416087520  + 18.76676304310 *T)
    sumX0  +=      16.958 * cos(4.29317977301  + 49.00186225990 *T)
    sumX0  +=      14.158 * cos(0.89623255371  + 147.80127704800 *T)
    sumX0  +=      14.666 * cos(4.09150859076  + 0.50694204189 *T)
    sumX0  +=      16.964 * cos(4.64489049535  + 64.94292363340 *T)
    sumX0  +=      16.167 * cos(0.81441085963  + 176.89434999200 *T)
    sumX0  +=      17.434 * cos(4.18185648024  + 381.36842318480 *T)
    sumX0  +=      12.702 * cos(0.95868773338  + 141.45507312490 *T)
    sumX0  +=      12.494 * cos(3.48097632144  + 196.01680510321 *T)
    sumX0  +=      13.527 * cos(0.03173550378  + 73.70100203691 *T)
    sumX0  +=      13.527 * cos(4.64766693222  + 76.34983006470 *T)
    sumX0  +=      14.805 * cos(3.56965005519  + 97.17169833279 *T)
    sumX0  +=      12.844 * cos(0.72145574815  + 322.36782729660 *T)
    sumX0  +=      13.194 * cos(1.14309437857  + 552.34169729099 *T)
    sumX0  +=      12.712 * cos(1.93345626951  + 748.96601817260 *T)
    sumX0  +=      13.959 * cos(6.05671842797  + 60.35637906450 *T)
    sumX0  +=      16.442 * cos(4.32045457497  + 83.62343575990 *T)
    sumX0  +=      12.305 * cos(4.99419427175  + 14.73403604350 *T)
    sumX0  +=      16.319 * cos(3.56734156984  + 703.87700210090 *T)
    sumX0  +=      12.524 * cos(4.56012598550  + 501.62360692681 *T)
    sumX0  +=      13.101 * cos(1.72098134216  + 112.67081672160 *T)
    sumX0  +=      13.451 * cos(3.28134459176  + 196.38050339810 *T)
    sumX0  +=      16.729 * cos(2.85149875163  + 336.59492129820 *T)
    sumX0  +=      15.662 * cos(5.74090488191  + 124.74666867370 *T)
    sumX0  +=      16.361 * cos(5.04073870288  + 853.44019923550 *T)
    sumX0  +=      15.016 * cos(2.24876100706  + 193.41155798150 *T)
    sumX0  +=      13.452 * cos(5.98440620710  + 172.44093186710 *T)
    sumX0  +=      14.832 * cos(2.47767910416  + 115.60613224340 *T)
    sumX0  +=      11.802 * cos(1.77590529063  + 34.28713329129 *T)
    sumX0  +=      11.640 * cos(4.47571714689  + 105.13657859549 *T)
    sumX0  +=      12.330 * cos(4.49348945560  + 83.93225230061 *T)
    sumX0  +=      11.536 * cos(6.01601210702  + 241.36645360580 *T)
    sumX0  +=      11.526 * cos(3.51836993689  + 628.60776856660 *T)
    sumX0  +=      13.571 * cos(4.79339886213  + 205.46615807421 *T)
    sumX0  +=      11.579 * cos(2.24790593657  + 65.57606307981 *T)
    sumX0  +=      11.468 * cos(0.00191176159  + 416.79314583689 *T)
    sumX0  +=      11.443 * cos(3.94479633517  + 20.69068660860 *T)
    sumX0  +=      12.105 * cos(4.85907022662  + 43.84836084689 *T)
    sumX0  +=      13.131 * cos(1.89137553565  + 288.49126782760 *T)
    sumX0  +=      14.050 * cos(6.17885749061  + 323.74923414091 *T)
    sumX0  +=      12.885 * cos(3.64902131797  + 283.90472325870 *T)
    sumX0  +=      15.266 * cos(2.99363853985  + 906.13039727500 *T)
    sumX0  +=      11.090 * cos(3.73542178033  + 56.11540926071 *T)
    sumX0  +=      15.022 * cos(0.00720412281  + 438.05446496220 *T)
    sumX0  +=      11.690 * cos(6.00028447424  + 256.18424843839 *T)
    sumX0  +=      10.863 * cos(3.14224264687  + 230.00523915680 *T)
    sumX0  +=      10.598 * cos(5.07035234152  + 11.60503193250 *T)
    sumX0  +=      10.559 * cos(0.12894495110  + 466.51439845980 *T)
    sumX0  +=      11.670 * cos(5.99338710443  + 308.93143930831 *T)
    sumX0  +=      14.228 * cos(3.18332327007  + 85.10790846821 *T)
    sumX0  +=      14.820 * cos(2.60745843389  + 154.05468206630 *T)
    sumX0  +=      11.614 * cos(2.42007432028  + 140.90017836831 *T)
    sumX0  +=      12.388 * cos(3.52030869410  + 1171.63205578549 *T)
    sumX0  +=      12.580 * cos(2.08601686331  + 114.12165953510 *T)
    sumX0  +=      12.667 * cos(4.44809124758  + 38.85242600079 *T)
    sumX0  +=      12.598 * cos(5.43444858317  + 1819.39364862570 *T)
    sumX0  +=      13.260 * cos(4.20791801425  + 77.71917482150 *T)
    sumX0  +=      14.442 * cos(4.27374278803  + 82.24202891559 *T)
    sumX0  +=      13.709 * cos(3.54703874690  + 109.06873185860 *T)
    sumX0  +=      10.566 * cos(6.07201741563  + 1107.89615748210 *T)
    sumX0  +=      12.906 * cos(0.41462458624  + 757.56403788170 *T)
    sumX0  +=      12.022 * cos(5.10417154765  + 207.63865198310 *T)
    sumX0  +=      11.539 * cos(0.97518052335  + 75.59486502240 *T)
    sumX0  +=      11.539 * cos(0.56262925906  + 74.45596707921 *T)
    sumX0  +=      10.478 * cos(3.97956030158  + 72.09010052900 *T)
    sumX0  +=      13.623 * cos(2.77502843360  + 406.34695512460 *T)
    sumX0  +=      10.221 * cos(5.18369595497  + 66.67347455760 *T)
    sumX0  +=      13.400 * cos(5.08609743155  + 25.84668014910 *T)
    sumX0  +=      12.805 * cos(3.30229416427  + 124.53422035260 *T)
    sumX0  +=      10.035 * cos(2.18243817807  + 82.61453593110 *T)
    sumX0  +=       9.987 * cos(2.36226793901  + 115.18544111810 *T)
    sumX0  +=      10.797 * cos(4.34736910404  + 199.31581891990 *T)
    sumX0  +=       9.996 * cos(5.23327380756  + 59.03866269500 *T)
    sumX0  +=      10.812 * cos(2.03480650379  + 89.25251005240 *T)
    sumX0  +=      13.348 * cos(5.29214999926  + 139.75815209550 *T)
    sumX0  +=      12.606 * cos(4.27203364948  + 28.55499313480 *T)
    sumX0  +=       9.781 * cos(5.57308917944  + 84.47476902181 *T)
    sumX0  +=       9.902 * cos(0.05888868084  + 262.77653057739 *T)
    sumX0  +=      10.352 * cos(2.38232189099  + 361.62163734781 *T)
    sumX0  +=      10.581 * cos(3.71988895032  + 679.81349389780 *T)
    sumX0  +=      10.976 * cos(6.13965346235  + 209.08949479660 *T)
    sumX0  +=      11.881 * cos(3.72150920911  + 827.67977000299 *T)
    sumX0  +=      10.568 * cos(2.60664396849  + 177.06110633080 *T)
    sumX0  +=      11.169 * cos(1.62829652213  + 318.43567403349 *T)
    sumX0  +=      10.655 * cos(5.22088985892  + 264.20612153150 *T)
    sumX0  +=      10.248 * cos(3.29930797837  + 65.63094483399 *T)
    sumX0  +=       9.768 * cos(4.02937576762  + 422.42222012939 *T)
    sumX0  +=      11.423 * cos(5.09245873399  + 679.49797971270 *T)
    sumX0  +=      11.319 * cos(1.06014170278  + 43.79347909271 *T)
    sumX0  +=      10.444 * cos(3.65292109900  + 64.24034821780 *T)
    sumX0  +=      10.733 * cos(1.76427675291  + 89.69445303710 *T)
    sumX0  +=       9.435 * cos(4.23041304178  + 59.83511420271 *T)
    sumX0  +=      10.593 * cos(4.05752516876  + 71.94459604960 *T)
    sumX0  +=      10.332 * cos(3.41106834601  + 61.04440026510 *T)
    sumX0  +=      10.962 * cos(3.68227156047  + 526.00262931501 *T)
    sumX0  +=      11.184 * cos(4.54016329366  + 150.61016377020 *T)
    sumX0  +=      10.812 * cos(2.19768748461  + 546.71262299849 *T)
    sumX0  +=      10.427 * cos(5.03449006418  + 106.03035047280 *T)
    sumX0  +=       9.926 * cos(4.82121666504  + 767.12526543730 *T)
    sumX0  +=      12.114 * cos(0.65339570824  + 404.86248241631 *T)
    sumX0  +=       9.406 * cos(1.01182317286  + 69.55640045270 *T)
    sumX0  +=       9.406 * cos(0.52598660955  + 80.49443164890 *T)
    sumX0  +=      10.953 * cos(1.50406511728  + 57.86300652739 *T)
    sumX0  +=      12.472 * cos(1.85646878831  + 337.55812914470 *T)
    sumX0  +=       9.300 * cos(1.19008258747  + 831.34879870770 *T)
    sumX0  +=      10.211 * cos(3.49614631238  + 544.16187657970 *T)
    sumX0  +=       9.945 * cos(4.38409935647  + 192.93598510200 *T)
    sumX0  +=      11.250 * cos(4.00375643631  + 544.26494244370 *T)
    sumX0  +=       9.524 * cos(2.76840570335  + 217.68751450571 *T)
    sumX0  +=       8.854 * cos(4.52273760683  + 210.36151918381 *T)
    sumX0  +=      10.193 * cos(1.28740045689  + 443.61984877990 *T)
    sumX0  +=       9.954 * cos(0.40697067708  + 79.85134119120 *T)
    sumX0  +=       9.954 * cos(1.13082001641  + 70.19949091040 *T)
    sumX0  +=       9.930 * cos(2.19978678204  + 63.01650794040 *T)
    sumX0  +=       8.650 * cos(3.04198602011  + 533.11617631580 *T)
    sumX0  +=      10.719 * cos(1.14850073361  + 399.92142932440 *T)
    sumX0  +=      12.021 * cos(0.45310721940  + 54.89406111079 *T)
    sumX0  +=      11.050 * cos(6.08929460929  + 210.60759739969 *T)
    sumX0  +=      10.384 * cos(2.71667758871  + 30.35498002820 *T)
    sumX0  +=      10.014 * cos(4.63045642556  + 419.24082639170 *T)
    sumX0  +=      11.868 * cos(4.54047117988  + 353.54488250050 *T)
    sumX0  +=      11.500 * cos(1.83332132826  + 402.89037474099 *T)
    sumX0  +=      10.255 * cos(0.01704274302  + 30.93323772040 *T)
    sumX0  +=       8.996 * cos(4.66913463180  + 179.03321400610 *T)
    sumX0  +=      11.583 * cos(5.69985889860  + 754.59509246510 *T)
    sumX0  +=      10.299 * cos(3.14377404278  + 112.89781925770 *T)
    sumX0  +=       8.391 * cos(0.26029466491  + 231.18089532440 *T)
    sumX0  +=       9.702 * cos(2.79614338200  + 181.03895157621 *T)
    sumX0  +=       8.169 * cos(6.13604295471  + 674.18441960530 *T)
    sumX0  +=       9.580 * cos(0.13644851421  + 292.25666475190 *T)
    sumX0  +=       8.174 * cos(3.07915003163  + 121.59890483079 *T)
    sumX0  +=      10.408 * cos(0.54237327390  + 533.86693584120 *T)
    sumX0  +=       9.226 * cos(0.54709114702  + 74.11039231410 *T)
    sumX0  +=       9.226 * cos(0.99071863539  + 75.94043978751 *T)
    sumX0  +=       8.015 * cos(1.33546656788  + 350.57593708390 *T)
    sumX0  +=       9.330 * cos(1.56226315938  + 53.04589010760 *T)
    sumX0  +=       9.204 * cos(3.60245234214  + 223.62540533890 *T)
    sumX0  +=       8.485 * cos(3.91173741300  + 212.57970507500 *T)
    sumX0  +=       8.003 * cos(5.62485208216  + 1022.06885865090 *T)
    sumX0  +=       8.044 * cos(2.40245968826  + 77.36122202100 *T)
    sumX0  +=       8.044 * cos(5.41853540133  + 72.68961008061 *T)
    sumX0  +=      10.452 * cos(0.41418724093  + 841.90686400460 *T)
    sumX0  +=       7.961 * cos(3.88224898072  + 43.37278796740 *T)
    sumX0  +=       8.333 * cos(2.75421683880  + 152.09463647851 *T)
    sumX0  +=       9.216 * cos(0.57128331684  + 670.25226634220 *T)
    sumX0  +=       9.064 * cos(0.56424328108  + 153.43035134050 *T)
    sumX0  +=       7.893 * cos(4.08557344211  + 93.93542284090 *T)
    sumX0  +=       9.501 * cos(5.55196701361  + 156.70811329780 *T)
    sumX0  +=       9.347 * cos(1.91675181106  + 157.22937815961 *T)
    sumX0  +=       8.617 * cos(4.16810964272  + 85.81048388380 *T)
    sumX0  +=       7.939 * cos(3.63926584848  + 109.22629842550 *T)
    sumX0  +=       8.607 * cos(6.26788179913  + 63.17656663481 *T)
    sumX0  +=      10.579 * cos(6.03372568697  + 17.76992530181 *T)
    sumX0  +=       7.543 * cos(4.71193192925  + 77.26064828450 *T)
    sumX0  +=       7.543 * cos(6.25065581393  + 72.79018381711 *T)
    sumX0  +=       8.486 * cos(4.73310682076  + 78.10623605200 *T)
    sumX0  +=       9.532 * cos(3.48105051561  + 835.28095197081 *T)
    sumX0  +=       7.509 * cos(0.24778282724  + 216.72430665921 *T)
    sumX0  +=       8.683 * cos(1.33880616649  + 117.77862615229 *T)
    sumX0  +=       7.488 * cos(3.66693161684  + 90.21571789890 *T)
    sumX0  +=       9.008 * cos(3.17075001642  + 175.67050971460 *T)
    sumX0  +=       7.543 * cos(3.74071360489  + 74.29170286790 *T)
    sumX0  +=       7.543 * cos(0.93868883111  + 75.75912923371 *T)
    sumX0  +=       8.913 * cos(6.15081816716  + 683.74564716089 *T)
    sumX0  +=       8.280 * cos(3.88721456088  + 216.23667169220 *T)
    sumX0  +=      10.165 * cos(1.03073080986  + 1364.97191706540 *T)
    sumX0  +=       7.591 * cos(1.52968766452  + 351.57277482520 *T)
    sumX0  +=       8.205 * cos(5.09744900361  + 37.41364527480 *T)
    sumX0  +=       7.588 * cos(4.85567964296  + 198.56505939450 *T)
    sumX0  +=       8.473 * cos(4.68984634698  + 71.25332148210 *T)
    sumX0  +=       8.473 * cos(3.13114874261  + 78.79751061951 *T)
    sumX0  +=       9.107 * cos(4.67464118916  + 37.15301284391 *T)
    sumX0  +=       7.766 * cos(5.62963711474  + 1479.11039154791 *T)
    sumX0  +=       8.403 * cos(2.94304661637  + 253.21530302180 *T)
    sumX0  +=       7.642 * cos(3.78920936733  + 49.96507010641 *T)
    sumX0  +=       9.234 * cos(5.87328409600  + 166.84548746939 *T)
    sumX0  +=       8.874 * cos(1.21562855803  + 483.46435966210 *T)
    sumX0  +=       7.128 * cos(5.26168401372  + 173.69840203930 *T)
    sumX0  +=       7.872 * cos(4.40559016099  + 437.88770862340 *T)
    sumX0  +=       9.185 * cos(3.56691829389  + 101.04896984171 *T)
    sumX0  +=       7.312 * cos(1.42644648376  + 526.75338884040 *T)
    sumX0  +=       9.194 * cos(0.24395127549  + 97.43233076370 *T)
    sumX0  +=       8.182 * cos(4.10243639384  + 115.86676467430 *T)
    sumX0  +=       9.663 * cos(4.58316566558  + 763.19311217420 *T)
    sumX0  +=       8.486 * cos(5.36038512781  + 242.11472100371 *T)
    sumX0  +=       8.799 * cos(1.82084099731  + 366.72944677850 *T)
    sumX0  +=       8.458 * cos(0.01382575407  + 297.98880490840 *T)
    sumX0  +=       8.904 * cos(3.41133622364  + 60.20762121820 *T)
    sumX0  +=       7.575 * cos(1.04449883197  + 214.50612076800 *T)
    sumX0  +=       6.999 * cos(0.89579766070  + 144.69920518740 *T)
    sumX0  +=       8.000 * cos(2.06536324132  + 325.13983075710 *T)
    sumX0  +=       8.498 * cos(5.46871328448  + 20.73887071840 *T)
    sumX0  +=       9.641 * cos(2.32432883600  + 56.89979868090 *T)
    sumX0  +=       7.094 * cos(1.39205091254  + 49.75262178531 *T)
    sumX0  +=       7.954 * cos(5.63302185515  + 87.03432416121 *T)
    sumX0  +=       6.970 * cos(3.53190960252  + 130.79643409150 *T)
    sumX0  +=       7.608 * cos(0.76489562356  + 409.31590054121 *T)
    sumX0  +=       7.168 * cos(0.93279559301  + 1265.81129610991 *T)
    sumX0  +=       8.728 * cos(1.96431386488  + 78.99119920869 *T)
    sumX0  +=       7.572 * cos(1.74214777176  + 148.91324274080 *T)
    sumX0  +=       6.783 * cos(1.46116350403  + 555.31064270760 *T)
    sumX0  +=       8.481 * cos(2.96221922852  + 119.26309886060 *T)
    sumX0  +=       7.247 * cos(4.39792712384  + 33.26904369061 *T)
    sumX0  +=       6.911 * cos(4.10374012893  + 41.51255487670 *T)
    sumX0  +=       6.813 * cos(1.07636826238  + 301.17019864609 *T)
    sumX0  +=       8.470 * cos(5.65128185788  + 250.19147585100 *T)
    sumX0  +=       6.549 * cos(3.94896050121  + 129.36014549300 *T)
    sumX0  +=       6.543 * cos(1.39714714677  + 44.36962570870 *T)
    sumX0  +=       8.617 * cos(4.71598219531  + 150.70078649541 *T)
    sumX0  +=       7.676 * cos(0.12204540495  + 60.46825364911 *T)
    sumX0  +=       8.905 * cos(1.60527300986  + 749.82616015511 *T)
    sumX0  +=       7.175 * cos(0.91189106023  + 445.96446347071 *T)
    sumX0  +=       6.465 * cos(3.89256895528  + 75.50098893029 *T)
    sumX0  +=       7.106 * cos(5.68687043451  + 6456.63624021419 *T)
    sumX0  +=       7.220 * cos(0.39640769723  + 1692.40948698591 *T)
    sumX0  +=       6.926 * cos(5.63023317193  + 20.90562705720 *T)
    sumX0  +=       7.786 * cos(1.14887986313  + 100.40127618030 *T)
    sumX0  +=       6.164 * cos(3.36294623636  + 1376.01761732930 *T)
    sumX0  +=       7.312 * cos(6.02023970623  + 188.90325810240 *T)
    sumX0  +=       6.788 * cos(1.55308559495  + 86.87426546680 *T)
    sumX0  +=       6.246 * cos(1.42620877338  + 14.57397734909 *T)
    sumX0  +=       6.057 * cos(0.06309705992  + 267.71758366931 *T)
    sumX0  +=       7.767 * cos(2.05936273306  + 95.73540973430 *T)
    sumX0  +=       7.632 * cos(1.14078660914  + 22.22334342671 *T)
    sumX0  +=       6.881 * cos(2.23862156704  + 163.87654205279 *T)
    sumX0  +=       6.523 * cos(5.12980083900  + 368.83825021259 *T)
    sumX0  +=       6.006 * cos(5.77254836616  + 560.93971700010 *T)
    sumX0  +=       6.916 * cos(4.01629596963  + 497.69145366371 *T)
    sumX0  +=       7.042 * cos(2.46549785002  + 39.11305843170 *T)
    sumX0  +=       7.444 * cos(1.82983001446  + 396.78572756900 *T)
    sumX0  +=       7.949 * cos(0.17119410374  + 474.70302789170 *T)
    sumX0  +=       7.441 * cos(0.32295076565  + 260.01333583751 *T)
    sumX0  +=       8.032 * cos(2.40925890235  + 224.06734832360 *T)
    sumX0  +=       5.746 * cos(4.13473916751  + 193.04785968660 *T)
    sumX0  +=       5.743 * cos(3.19578471830  + 20201.98247923659 *T)
    sumX0  +=       7.757 * cos(2.22105969633  + 688.41151360689 *T)
    sumX0  +=       7.329 * cos(5.57096867484  + 237.17366791179 *T)
    sumX0  +=       5.919 * cos(5.22221186676  + 17.39416491939 *T)
    sumX0  +=       6.252 * cos(0.31443956867  + 132.16903221521 *T)
    sumX0  +=       5.870 * cos(2.71837669898  + 690.85919416170 *T)
    sumX0  +=       6.964 * cos(1.45493516622  + 197.55615956570 *T)
    sumX0  +=       7.219 * cos(3.86376797010  + 266.96931627140 *T)
    sumX0  +=       7.947 * cos(4.04273198078  + 514.46929408409 *T)
    sumX0  +=       7.626 * cos(5.24503341118  + 842.39449897160 *T)
    sumX0  +=       6.895 * cos(3.05037238160  + 425.35753565121 *T)
    sumX0  +=       6.665 * cos(1.38387522554  + 436.56999225390 *T)
    sumX0  +=       6.291 * cos(0.10212516125  + 343.49601997789 *T)
    sumX0  +=       6.140 * cos(2.78780439407  + 213.70966926030 *T)
    sumX0  +=       5.915 * cos(2.43173952853  + 62.75587550951 *T)
    sumX0  +=       5.915 * cos(5.38925556105  + 87.29495659210 *T)
    sumX0  +=       6.781 * cos(5.62139692952  + 647.25465079831 *T)
    sumX0  +=       6.309 * cos(4.56219828589  + 178.37882270030 *T)
    sumX0  +=       5.782 * cos(1.65730371744  + 401.29402744810 *T)
    sumX0  +=       7.776 * cos(0.40754452945  + 204.94489321240 *T)
    sumX0  +=       6.809 * cos(2.65807547458  + 291.94784821120 *T)
    sumX0  +=       5.680 * cos(2.60329405000  + 79.64559051450 *T)
    sumX0  +=       5.958 * cos(2.08802033162  + 808.76976321290 *T)
    sumX0  +=       5.850 * cos(1.55616178293  + 358.65269193120 *T)
    sumX0  +=       5.624 * cos(2.57073308836  + 42.30900638440 *T)
    sumX0  +=       7.759 * cos(0.26906707780  + 79.27308349900 *T)
    sumX0  +=       5.495 * cos(2.40092261645  + 159.20818347930 *T)
    sumX0  +=       5.488 * cos(3.62340702832  + 293.95358578131 *T)
    sumX0  +=       7.423 * cos(0.60231124353  + 139.72452220071 *T)
    sumX0  +=       5.570 * cos(4.40848654037  + 187.99493201010 *T)
    sumX0  +=       6.894 * cos(4.85514543558  + 407.83142783290 *T)
    sumX0  +=       5.539 * cos(3.27596723365  + 1235.36795408890 *T)
    sumX0  +=       6.394 * cos(5.72210287246  + 956.53297345411 *T)
    sumX0  +=       5.577 * cos(5.67561703294  + 168.92946972931 *T)
    sumX0  +=       6.276 * cos(6.25873212243  + 97.00494199400 *T)
    sumX0  +=       7.541 * cos(2.51146519598  + 303.87851163179 *T)
    sumX0  +=       6.343 * cos(0.19031564137  + 164.36417701981 *T)
    sumX0  +=       6.353 * cos(5.36295948561  + 370.21965705690 *T)
    sumX0  +=       5.390 * cos(1.55675166817  + 818.01547658340 *T)
    sumX0  +=       5.777 * cos(4.25785415604  + 271.37455028650 *T)
    sumX0  +=       6.126 * cos(0.94260816975  + 45.22976769120 *T)
    sumX0  +=       5.488 * cos(3.66616025422  + 136.47369249381 *T)
    sumX0  +=       6.742 * cos(4.40828850034  + 89.84321088341 *T)
    sumX0  +=       6.582 * cos(0.85603229747  + 1291.18715623940 *T)
    sumX0  +=       5.219 * cos(5.16576359992  + 7.62048904269 *T)
    sumX0  +=       6.692 * cos(1.57725367967  + 5994.75133850259 *T)
    sumX0  +=       6.355 * cos(5.81946862067  + 184.48346987229 *T)
    sumX0  +=       7.139 * cos(2.55929105921  + 359.17395679300 *T)
    sumX0  +=       6.215 * cos(0.02868656789  + 1582.20316576650 *T)
    sumX0  +=       5.605 * cos(3.11246665067  + 43.48466255200 *T)
    sumX0  +=       5.449 * cos(0.56711967932  + 229.09691306450 *T)
    sumX0  +=       5.787 * cos(0.20970024792  + 492.06237937121 *T)
    sumX0  +=       6.553 * cos(4.18583913829  + 1065.35788782919 *T)
    sumX0  +=       5.340 * cos(4.29293326576  + 27.49790919620 *T)
    sumX0  +=       5.124 * cos(2.24404925860  + 479.53220639900 *T)
    sumX0  +=       5.421 * cos(0.30923710851  + 298.88257678570 *T)
    sumX0  +=       5.261 * cos(0.87883152563  + 211.57080524620 *T)
    sumX0  +=       6.420 * cos(5.27556145965  + 489.09343395460 *T)
    sumX0  +=       5.889 * cos(2.39433492737  + 1514.53511420000 *T)
    sumX0  +=       4.956 * cos(4.14479114010  + 450.41788159561 *T)
    sumX0  +=       6.450 * cos(3.02138558238  + 9924.96103472219 *T)
    sumX0  +=       4.940 * cos(0.52935323621  + 20.95381116700 *T)
    sumX0  +=       5.847 * cos(1.42358003244  + 89.58257845250 *T)
    sumX0  +=       5.600 * cos(3.40624847077  + 241.85408857280 *T)
    sumX0  +=       6.152 * cos(4.92625022847  + 524.51815660670 *T)
    sumX0  +=       6.088 * cos(2.72225507513  + 711.69312451730 *T)
    sumX0  +=       5.407 * cos(3.68956638902  + 148.16248321540 *T)
    sumX0  +=       6.327 * cos(3.36352822659  + 300.64893378430 *T)
    sumX0  +=       5.490 * cos(0.79611377485  + 261.81332273089 *T)
    sumX0  +=       5.280 * cos(5.56598623472  + 74.42590649920 *T)
    sumX0  +=       4.835 * cos(5.88731588561  + 122.77456099839 *T)
    sumX0  +=       5.417 * cos(3.67541639447  + 34.18406742730 *T)
    sumX0  +=       5.465 * cos(5.75476347267  + 49.89563413720 *T)
    sumX0  +=       4.962 * cos(3.25176869821  + 40.49446527601 *T)
    sumX0  +=       5.202 * cos(3.43757897939  + 102.27950776349 *T)
    sumX0  +=       5.768 * cos(0.87260894941  + 237.92193530970 *T)
    sumX0  +=       5.794 * cos(0.10891997847  + 215.99059347631 *T)
    sumX0  +=       6.057 * cos(1.62279256892  + 1595.04885292380 *T)
    sumX0  +=       6.144 * cos(3.15358005008  + 456.95317090420 *T)
    sumX0  +=       5.439 * cos(3.24285098245  + 164.13717448370 *T)
    sumX0  +=       6.169 * cos(1.60041197456  + 2199.52141658570 *T)
    sumX0  +=       5.803 * cos(3.65374412104  + 476.07562601539 *T)
    sumX0  +=       4.737 * cos(0.78451360563  + 905.64276230799 *T)
    sumX0  +=       4.927 * cos(1.69388966240  + 752.89817143570 *T)
    sumX0  +=       4.656 * cos(0.34442537000  + 89.09494348550 *T)
    sumX0  +=       5.228 * cos(2.74942181286  + 112.63718682681 *T)
    sumX0  +=       5.103 * cos(2.15939141750  + 93.54836161040 *T)
    sumX0  +=       4.793 * cos(1.96914093772  + 257.14745628489 *T)
    sumX0  +=       5.517 * cos(1.72811028560  + 85.95924173011 *T)
    sumX0  +=       5.174 * cos(2.58855661317  + 460.88532416730 *T)
    sumX0  +=       4.600 * cos(3.32053950440  + 693.82813957829 *T)
    sumX0  +=       5.568 * cos(0.53088329446  + 346.15614885380 *T)
    sumX0  +=       5.750 * cos(4.66927784024  + 475.11241816889 *T)
    sumX0  +=       5.821 * cos(3.92004054025  + 331.92905485220 *T)
    sumX0  +=       5.474 * cos(2.88263190253  + 119.09634252180 *T)
    sumX0  +=       5.555 * cos(3.07944899399  + 166.58485503849 *T)
    sumX0  +=       4.721 * cos(5.57665667104  + 818.33099076849 *T)
    sumX0  +=       4.598 * cos(3.62412165981  + 155.43608891061 *T)
    sumX0  +=       4.983 * cos(1.00895911835  + 305.36298434010 *T)
    sumX0  +=       4.604 * cos(0.67475144027  + 240.63024829540 *T)
    sumX0  +=       4.534 * cos(2.51553633780  + 426.35437339250 *T)
    sumX0  +=       6.030 * cos(4.20090275052  + 928.22179780281 *T)
    sumX0  +=       5.073 * cos(3.12942565704  + 119.15122427600 *T)
    sumX0  +=       4.800 * cos(3.95967152002  + 133.82026126229 *T)
    sumX0  +=       5.861 * cos(0.80317428777  + 441.51104534581 *T)
    sumX0  +=       4.509 * cos(3.96144988696  + 106.67804413421 *T)
    sumX0  +=       4.990 * cos(1.91385887833  + 53.40958840249 *T)
    sumX0  +=       5.764 * cos(2.06812354087  + 896.08153475239 *T)
    sumX0  +=       5.407 * cos(3.17620848565  + 469.48334387639 *T)
    sumX0  +=       5.823 * cos(2.53279154851  + 29.50364676630 *T)
    sumX0  +=       5.944 * cos(1.89466199570  + 235.94982763440 *T)
    sumX0  +=       6.147 * cos(6.16340894908  + 1288.21821082279 *T)
    sumX0  +=       4.598 * cos(2.84824896257  + 224.32798075451 *T)
    sumX0  +=       5.709 * cos(1.19300025955  + 893.11258933580 *T)
    sumX0  +=       4.523 * cos(0.42084353764  + 530.09484127250 *T)
    sumX0  +=       5.024 * cos(3.74456662203  + 314.66357946480 *T)
    sumX0  +=       4.527 * cos(0.18441552056  + 100.29821031630 *T)
    sumX0  +=       4.510 * cos(3.91932536657  + 529.77472388370 *T)
    sumX0  +=       5.026 * cos(2.35475688005  + 74.64965566840 *T)
    sumX0  +=       5.303 * cos(5.80155174455  + 273.82223084130 *T)
    sumX0  +=       4.380 * cos(2.56083534926  + 1578.27101250340 *T)
    sumX0  +=       5.525 * cos(5.57587445203  + 453.18107633550 *T)
    sumX0  +=       4.962 * cos(5.09899961225  + 810.90192958261 *T)
    sumX0  +=       4.972 * cos(4.91118746618  + 334.89800026879 *T)
    sumX0  +=       5.401 * cos(0.36756701975  + 193.67219041240 *T)
    sumX0  +=       5.498 * cos(5.42712164709  + 230.58138577279 *T)
    sumX0  +=       4.501 * cos(1.13664286212  + 523.71500745461 *T)
    sumX0  +=       4.303 * cos(1.61642146352  + 1246.41365435280 *T)
    sumX0  +=       5.261 * cos(0.11599202303  + 170.25388374320 *T)
    sumX0  +=       4.481 * cos(3.32785486855  + 5.35162691421 *T)
    sumX0  +=       4.655 * cos(1.22893719366  + 178.43370445449 *T)
    sumX0  +=       5.029 * cos(4.93286736415  + 193.93282284329 *T)
    sumX0  +=       5.743 * cos(5.13195285730  + 246.68001371321 *T)
    sumX0  +=       5.652 * cos(2.43693211690  + 315.41184686271 *T)
    sumX0  +=       4.604 * cos(1.52839326823  + 1072.47143483000 *T)
    sumX0  +=       5.494 * cos(4.72792390506  + 1044.64789414570 *T)
    sumX0  +=       4.789 * cos(2.23605317652  + 361.78839368660 *T)
    sumX0  +=       5.139 * cos(3.92903030642  + 223.20720634109 *T)
    sumX0  +=       5.310 * cos(5.00194088840  + 444.47999076240 *T)
    sumX0  +=       4.815 * cos(2.88473476458  + 3041.24221494710 *T)
    sumX0  +=       4.119 * cos(3.70589854610  + 2120.49215057020 *T)
    sumX0  +=       4.552 * cos(5.13940535692  + 158.50810019120 *T)
    sumX0  +=       4.594 * cos(5.30513223240  + 1310.14955265620 *T)
    sumX0  +=       4.173 * cos(4.27937119249  + 181.40264987110 *T)
    sumX0  +=       4.090 * cos(3.85142033448  + 138.64943976960 *T)
    sumX0  +=       4.884 * cos(3.63854055554  + 6.65728119619 *T)
    sumX0  +=       4.740 * cos(1.08537988908  + 25788.53292982149 *T)
    sumX0  +=       5.699 * cos(4.42014972948  + 772.83258441960 *T)
    sumX0  +=       5.296 * cos(6.13051223469  + 694.79134742480 *T)
    sumX0  +=       4.555 * cos(1.54916256463  + 154.42718908179 *T)
    sumX0  +=       4.669 * cos(0.25884741077  + 127.82748867490 *T)
    sumX0  +=       4.509 * cos(0.11998259133  + 227.35180792530 *T)
    sumX0  +=       4.302 * cos(4.07489237646  + 100.08576199520 *T)
    sumX0  +=       4.027 * cos(2.46764881744  + 294.05665164530 *T)
    sumX0  +=       4.114 * cos(1.36897521941  + 383.06534421420 *T)
    sumX0  +=       4.552 * cos(1.51276211559  + 1361.03976380230 *T)
    sumX0  +=       4.653 * cos(3.06693329591  + 1019.52060435960 *T)
    sumX0  +=       4.849 * cos(1.52566036879  + 119.11759438120 *T)
    sumX0  +=       4.617 * cos(3.31831252020  + 887.48351504330 *T)
    sumX0  +=       5.166 * cos(5.64256709119  + 49.26498681829 *T)
    sumX0  +=       5.543 * cos(0.97951125744  + 296.50433220009 *T)
    sumX0  +=       3.936 * cos(3.52768574626  + 137.79810650771 *T)
    sumX0  +=       3.959 * cos(5.50934383834  + 151.17961274180 *T)
    sumX0  +=       4.546 * cos(2.33233821317  + 44.42661853911 *T)
    sumX0  +=       4.195 * cos(6.00162241896  + 305.21422649380 *T)
    sumX0  +=       3.911 * cos(0.86277832354  + 1020.26887175750 *T)
    sumX0  +=       3.906 * cos(3.18695974498  + 36.41929966101 *T)
    sumX0  +=       4.980 * cos(5.69236181156  + 682.78243931439 *T)
    sumX0  +=       3.894 * cos(4.70455775456  + 13.77082819700 *T)
    sumX0  +=       4.225 * cos(3.79984325344  + 310.73142620169 *T)
    sumX0  +=       3.999 * cos(0.85861235733  + 93.98781246761 *T)
    sumX0  +=       3.999 * cos(0.67919742508  + 56.06301963400 *T)
    sumX0  +=       4.072 * cos(0.89973310426  + 956.84848763920 *T)
    sumX0  +=       4.160 * cos(3.63151901615  + 138.92807069299 *T)
    sumX0  +=       3.870 * cos(1.95474602032  + 316.14805217309 *T)
    sumX0  +=       5.282 * cos(2.92111334017  + 442.99551805410 *T)
    sumX0  +=       3.856 * cos(3.41183001590  + 778.17096570120 *T)
    sumX0  +=       5.026 * cos(3.61371361415  + 252.89978883670 *T)
    sumX0  +=       4.872 * cos(2.20405135166  + 196.59295171920 *T)
    sumX0  +=       4.088 * cos(4.88665302875  + 770.58184582090 *T)
    sumX0  +=       3.976 * cos(1.86804354686  + 34.44469985821 *T)
    sumX0  +=       3.976 * cos(1.86804354686  + 33.95706489120 *T)
    sumX0  +=       4.425 * cos(2.95639722479  + 621.98185653281 *T)
    sumX0  +=       3.827 * cos(2.58907561634  + 29.81246330700 *T)
    sumX0  +=       4.303 * cos(3.92882697682  + 302.39403892350 *T)
    sumX0  +=       4.082 * cos(1.36715410600  + 76.98546163860 *T)
    sumX0  +=       4.775 * cos(5.01224539274  + 379.25961975071 *T)
    sumX0  +=       4.866 * cos(1.48866583048  + 349.12509427039 *T)
    sumX0  +=       4.558 * cos(2.22201452659  + 484.67364572450 *T)
    sumX0  +=       3.896 * cos(5.05170610646  + 525.74199688410 *T)
    sumX0  +=       3.790 * cos(4.25929018143  + 168.81759514471 *T)
    sumX0  +=       4.584 * cos(3.59480552867  + 30.41197285861 *T)
    sumX0  +=       3.789 * cos(3.30651741587  + 225.58545092669 *T)
    sumX0  +=       3.919 * cos(2.30050922510  + 367.61440993520 *T)
    sumX0  +=       4.495 * cos(5.49223496970  + 823.74761673989 *T)
    sumX0  +=       4.071 * cos(5.88966518721  + 152.28832506769 *T)
    sumX0  +=       4.431 * cos(4.29789318653  + 781.83999440590 *T)
    sumX0  +=       3.787 * cos(4.05895470835  + 96.22304470130 *T)
    sumX0  +=       3.738 * cos(2.49934308138  + 909.68679015570 *T)
    sumX0  +=       3.824 * cos(1.72982544585  + 81.01818863820 *T)
    sumX0  +=       5.100 * cos(1.77501816666  + 786.47472308461 *T)
    sumX0  +=       4.678 * cos(3.46326949476  + 294.42915866079 *T)
    sumX0  +=       4.456 * cos(0.22032826250  + 473.41549713950 *T)
    sumX0  +=       4.182 * cos(4.89705342604  + 109.96250373590 *T)
    sumX0  +=       4.721 * cos(3.16039895369  + 487.64259114110 *T)
    sumX0  +=       4.194 * cos(2.53407050665  + 135.82599883240 *T)
    sumX0  +=       4.150 * cos(5.59000281909  + 208.60185982960 *T)
    sumX0  +=       3.881 * cos(4.75443738028  + 527.48710202330 *T)
    sumX0  +=       5.003 * cos(4.37961745221  + 26088.14695905770 *T)
    sumX0  +=       4.925 * cos(0.81517935481  + 104.82106441041 *T)
    sumX0  +=       4.629 * cos(0.34855620937  + 324.97307441831 *T)
    sumX0  +=       3.764 * cos(5.45647898824  + 81.17575520511 *T)
    sumX0  +=       3.764 * cos(5.50610875494  + 68.87507689650 *T)
    sumX0  +=       4.634 * cos(1.92952542549  + 439.53893767049 *T)
    sumX0  +=       4.661 * cos(2.87667311621  + 1031.63008620649 *T)
    sumX0  +=       4.246 * cos(0.02482008071  + 290.87525790760 *T)
    sumX0  +=       4.701 * cos(2.63355711377  + 292.57217893700 *T)
    sumX0  +=       4.168 * cos(1.50787160276  + 347.68880567190 *T)
    sumX0  +=       3.574 * cos(2.38256827172  + 101.85211899380 *T)
    sumX0  +=       4.316 * cos(3.83309032671  + 1034.59903162310 *T)
    sumX0  +=       4.648 * cos(6.16877789702  + 701.01993126891 *T)
    sumX0  +=       5.000 * cos(3.39683093071  + 636.42389098300 *T)
    sumX0  +=       4.950 * cos(0.01286873425  + 901.71060904489 *T)
    sumX0  +=       4.343 * cos(2.34864924381  + 372.77040347569 *T)
    sumX0  +=       3.612 * cos(5.08316127068  + 154.31531449720 *T)
    sumX0  +=       4.180 * cos(2.96623594157  + 304.99047732459 *T)
    sumX0  +=       3.767 * cos(0.18019206646  + 1040.97886544099 *T)
    sumX0  +=       4.105 * cos(4.59107231627  + 1169.57143646120 *T)
    sumX0  +=       4.389 * cos(1.49290014038  + 67.63668240410 *T)
    sumX0  +=       3.610 * cos(2.89127345548  + 969.86629557841 *T)
    sumX0  +=       3.589 * cos(0.27875915694  + 1340.59289467720 *T)
    sumX0  +=       4.184 * cos(6.23478790622  + 1575.08961876571 *T)
    sumX0  +=       4.327 * cos(5.64145082781  + 470.44655172289 *T)
    sumX0  +=       3.635 * cos(3.93501297739  + 618.04970326970 *T)
    sumX0  +=       4.568 * cos(3.41794989423  + 374.46732450509 *T)
    sumX0  +=       4.014 * cos(1.70084166838  + 191.07575201129 *T)
    sumX0  +=       3.518 * cos(1.41112167500  + 812.70191647599 *T)
    sumX0  +=       4.543 * cos(3.48436789856  + 18.87863762769 *T)
    sumX0  +=       4.255 * cos(3.36331514891  + 257.78059573129 *T)
    sumX0  +=       3.536 * cos(2.41329229079  + 1158.78636862821 *T)
    sumX0  +=       4.782 * cos(5.54428360596  + 20.36311033600 *T)
    sumX0  +=       3.580 * cos(3.54221700449  + 155.79978720550 *T)
    sumX0  +=       4.461 * cos(3.14107597585  + 155.91166179010 *T)
    sumX0  +=       4.866 * cos(4.52211926632  + 16.85490156510 *T)
    sumX0  +=       4.353 * cos(5.88270989811  + 154.54231703330 *T)
    sumX0  +=       3.770 * cos(3.06440812943  + 238.91877305099 *T)
    sumX0  +=       3.780 * cos(2.96795333847  + 250.35823218980 *T)
    sumX0  +=       4.854 * cos(1.95565618643  + 823.12328601411 *T)
    sumX0  +=       4.833 * cos(4.80754178031  + 1894.66288216001 *T)
    sumX0  +=       3.662 * cos(6.11416816246  + 319.92014674179 *T)
    sumX0  +=       3.977 * cos(0.97335262015  + 326.19691469570 *T)
    sumX0  +=       3.861 * cos(4.17358736891  + 47.99296243110 *T)
    sumX0  +=       4.433 * cos(0.71074152760  + 1685.29593998510 *T)
    sumX0  +=       4.182 * cos(1.56605335566  + 548.40954402789 *T)
    sumX0  +=       3.421 * cos(1.94916325782  + 605.03189533049 *T)
    sumX0  +=       3.876 * cos(2.42283968740  + 302.13340649259 *T)
    sumX0  +=       4.008 * cos(0.26728290195  + 6137.20098863639 *T)
    sumX0  +=       4.500 * cos(2.95199565386  + 47.30494123050 *T)
    sumX0  +=       4.050 * cos(2.87043107244  + 6.46359260701 *T)
    sumX0  +=       3.411 * cos(6.02869083070  + 997.68983626269 *T)
    sumX0  +=       4.312 * cos(5.58120507194  + 847.81112494300 *T)
    sumX0  +=       3.428 * cos(6.26485948451  + 71.35638734610 *T)
    sumX0  +=       3.928 * cos(1.52298026279  + 395.30125486070 *T)
    sumX0  +=       3.650 * cos(4.95207182494  + 1254.27796087899 *T)
    sumX0  +=       3.844 * cos(0.57682019240  + 121.49583896680 *T)
    sumX0  +=       3.966 * cos(6.20228771414  + 761.49619114479 *T)
    sumX0  +=       3.497 * cos(2.60162047401  + 319.32944591080 *T)
    sumX0  +=       3.510 * cos(0.82870221909  + 273.34665796181 *T)
    sumX0  +=       3.408 * cos(0.81896528567  + 1016.33671849440 *T)
    sumX0  +=       4.206 * cos(2.08946991165  + 435.44002806861 *T)
    sumX0  +=       3.372 * cos(0.84549509997  + 213.49472881170 *T)
    sumX0  +=       3.369 * cos(4.04202536168  + 213.59109703130 *T)
    sumX0  +=       4.154 * cos(5.53330486226  + 357.16821922291 *T)
    sumX0  +=       3.966 * cos(3.31508277427  + 185.33480313420 *T)
    sumX0  +=       4.636 * cos(3.53025378242  + 6169.91739830280 *T)
    sumX0  +=       4.042 * cos(1.50838388855  + 432.47108265200 *T)
    sumX0  +=       3.366 * cos(1.08933178457  + 699.45721387080 *T)
    sumX0  +=       3.632 * cos(3.02688066661  + 771.05741870039 *T)
    sumX0  +=       3.967 * cos(1.24490650084  + 280.24775664150 *T)
    sumX0  +=       3.436 * cos(0.23792026514  + 182.36585771760 *T)
    sumX0  +=       4.588 * cos(4.60587020989  + 25483.18676042879 *T)
    sumX0  +=       3.825 * cos(3.29754629533  + 6129.33668211020 *T)
    sumX0  +=       3.243 * cos(2.79388610647  + 279.97256999560 *T)
    sumX0  +=       4.579 * cos(0.12286146246  + 2735.89604555439 *T)
    sumX0  +=       4.222 * cos(5.57461283603  + 344.94686279140 *T)
    sumX0  +=       4.222 * cos(2.43302018244  + 344.45922782439 *T)
    sumX0  +=       3.528 * cos(1.32293950034  + 187.78248368899 *T)
    sumX0  +=       3.976 * cos(2.04915986842  + 433.43429049850 *T)
    sumX0  +=       3.885 * cos(3.16317291930  + 254.33607743521 *T)
    sumX0  +=       3.256 * cos(5.33196090740  + 1756.14538528930 *T)
    sumX0  +=       3.450 * cos(0.99685451963  + 1336.66074141410 *T)
    sumX0  +=       3.763 * cos(4.59733292662  + 598.77849031221 *T)
    sumX0  +=       3.661 * cos(5.44386390628  + 523.99689174490 *T)
    sumX0  +=       3.764 * cos(1.88881171881  + 498.65466151020 *T)
    sumX0  +=       3.218 * cos(4.24420447682  + 829.37669103239 *T)
    sumX0  +=       3.783 * cos(1.68999507602  + 10067.41068485599 *T)
    sumX0  +=       4.463 * cos(4.05993678503  + 686.71459257749 *T)
    sumX0  +=       3.955 * cos(5.22637518019  + 386.99749747730 *T)
    sumX0  +=       3.269 * cos(3.81995058967  + 90.80641872990 *T)
    sumX0  +=       3.268 * cos(4.00080141209  + 59.24441337170 *T)
    sumX0  +=       3.540 * cos(4.82265466943  + 313.17910675650 *T)
    sumX0  +=       3.535 * cos(3.36280834282  + 78039.12081153718 *T)
    sumX0  +=       3.318 * cos(5.75025984322  + 522.33360061030 *T)
    sumX0  +=       3.173 * cos(2.42451995202  + 1475.17823828480 *T)
    sumX0  +=       3.156 * cos(4.25625589255  + 1596.43025976811 *T)
    sumX0  +=       4.282 * cos(4.65731144409  + 269.31393096221 *T)
    sumX0  +=       3.178 * cos(5.76350607566  + 1886.79857563380 *T)
    sumX0  +=       4.407 * cos(2.37962716553  + 236.74627914210 *T)
    sumX0  +=       3.698 * cos(1.89962562145  + 762.98066385310 *T)
    sumX0  +=       3.368 * cos(3.39410364596  + 196.86813836510 *T)
    sumX0  +=       4.221 * cos(4.85565822083  + 102.74589087111 *T)
    sumX0  +=       3.725 * cos(5.21185176662  + 511.28790034640 *T)
    sumX0  +=       3.791 * cos(3.67873273608  + 1147.74066836430 *T)
    sumX0  +=       3.211 * cos(5.53866872898  + 151.81275218821 *T)
    sumX0  +=       3.516 * cos(0.78503990832  + 470.24960976680 *T)
    sumX0  +=       3.947 * cos(3.66915150826  + 757.04277301989 *T)
    sumX0  +=       3.990 * cos(4.04987769223  + 908.09044286280 *T)
    sumX0  +=       3.674 * cos(2.73741527477  + 631.78916230430 *T)
    sumX0  +=       3.787 * cos(0.46761206888  + 798.14475407429 *T)
    sumX0  +=       4.369 * cos(4.95294795394  + 10100.12709452239 *T)
    sumX0  +=       3.536 * cos(5.60494639394  + 354.75416856289 *T)
    sumX0  +=       3.391 * cos(2.25906474385  + 5908.92403967139 *T)
    sumX0  +=       3.083 * cos(4.48757923108  + 829.06117684730 *T)
    sumX0  +=       4.306 * cos(4.50722020801  + 503.10807963510 *T)
    sumX0  +=       3.547 * cos(1.18953274954  + 38.84993387330 *T)
    sumX0  +=       3.214 * cos(2.73020363368  + 660.85109748100 *T)
    sumX0  +=       3.069 * cos(3.57500923483  + 53.56715496940 *T)
    sumX0  +=       3.103 * cos(4.11072552350  + 367.35377750430 *T)
    sumX0  +=       3.753 * cos(2.02140889521  + 486.86069384840 *T)
    sumX0  +=       3.982 * cos(5.27127178209  + 614.59312288610 *T)
    sumX0  +=       3.567 * cos(1.58596030710  + 473.21855518340 *T)
    sumX0  +=       3.603 * cos(4.72020715629  + 10059.54637832979 *T)
    sumX0  +=       3.339 * cos(0.78854853830  + 15.25530090530 *T)
    sumX0  +=       3.342 * cos(4.51414863179  + 391.73279989249 *T)
    sumX0  +=       3.638 * cos(1.61128601114  + 104.77288030060 *T)
    sumX0  +=       3.557 * cos(0.14967035207  + 365.86930479600 *T)
    sumX0  +=       3.047 * cos(0.96929818074  + 75.23786437190 *T)
    sumX0  +=       3.047 * cos(3.71010425526  + 74.81296772971 *T)
    sumX0  +=       3.632 * cos(1.90931698611  + 2175.14239419750 *T)
    sumX0  +=       3.727 * cos(1.77889452064  + 668.76779363389 *T)
    sumX0  +=       3.110 * cos(5.05640701071  + 95.68722562449 *T)
    sumX0  +=       3.110 * cos(5.05640701071  + 96.17486059150 *T)
    sumX0  +=       3.034 * cos(2.90532445467  + 98.65617104110 *T)
    sumX0  +=       3.292 * cos(1.54349283302  + 68.90870679129 *T)
    sumX0  +=       3.166 * cos(1.24980683639  + 5.69407334969 *T)
    sumX0  +=       3.043 * cos(2.01164982401  + 100.14064374939 *T)
    sumX0  +=       3.150 * cos(3.91957783792  + 6532.90231148980 *T)
    sumX0  +=       3.284 * cos(0.73960261885  + 138.87318893880 *T)
    sumX0  +=       3.008 * cos(0.77075221627  + 287.83687652179 *T)
    sumX0  +=       3.451 * cos(1.61115939195  + 1206.21999939070 *T)
    sumX0  +=       3.235 * cos(3.99969523787  + 302.65467135440 *T)
    sumX0  +=       4.224 * cos(4.39410979413  + 527.22646959241 *T)
    sumX0  +=       2.990 * cos(1.05588985579  + 182.52342428450 *T)
    sumX0  +=       3.383 * cos(1.88782670005  + 30.95448957981 *T)
    sumX0  +=       4.137 * cos(2.73198697887  + 436.40323591510 *T)
    sumX0  +=       3.071 * cos(1.35070695195  + 198.82818395289 *T)
    sumX0  +=       3.076 * cos(3.98325926368  + 608.00084074710 *T)
    sumX0  +=       3.557 * cos(1.81225163838  + 731.70054278519 *T)
    sumX0  +=       3.557 * cos(1.81225163838  + 732.18817775220 *T)
    sumX0  +=       3.552 * cos(0.07771515960  + 262.20038396140 *T)
    sumX0  +=       4.135 * cos(5.62625626157  + 195.15666312070 *T)
    sumX0  +=       2.951 * cos(0.30762412341  + 161.65586403411 *T)
    sumX0  +=       2.963 * cos(2.39442029962  + 1044.16025917870 *T)
    sumX0  +=       3.762 * cos(3.72993621639  + 751.31063286340 *T)
    sumX0  +=       3.317 * cos(1.95763203162  + 29.55183087609 *T)
    sumX0  +=       3.109 * cos(4.88578496396  + 1061.42573456610 *T)
    sumX0  +=       2.930 * cos(6.21731762577  + 60.31068708220 *T)
    sumX0  +=       2.938 * cos(5.68333516250  + 380.63471000190 *T)
    sumX0  +=       3.292 * cos(1.40459725923  + 576.40520549410 *T)
    sumX0  +=       2.939 * cos(4.59146101512  + 47.02975458460 *T)
    sumX0  +=       2.922 * cos(0.92284575293  + 55.05162767771 *T)
    sumX0  +=       2.922 * cos(0.61496402948  + 94.99920442390 *T)
    sumX0  +=       3.190 * cos(3.68158030074  + 9839.13373589100 *T)
    sumX0  +=       4.110 * cos(4.80248129129  + 2885.94687765600 *T)
    sumX0  +=       3.635 * cos(0.86054546097  + 189.96703968541 *T)
    sumX0  +=       3.962 * cos(0.33365956631  + 511.35159082120 *T)
    sumX0  +=       2.907 * cos(5.93117000490  + 373.29166833749 *T)
    sumX0  +=       2.911 * cos(1.86560039627  + 379.13319095109 *T)
    sumX0  +=       3.057 * cos(3.85149288925  + 517.66524203681 *T)
    sumX0  +=       3.748 * cos(3.72709697779  + 5985.19011094699 *T)
    sumX0  +=       3.250 * cos(0.47551640281  + 385.82184130970 *T)
    sumX0  +=       3.250 * cos(0.47551640281  + 386.30947627670 *T)
    sumX0  +=       3.118 * cos(3.20537698563  + 404.75060783170 *T)
    sumX0  +=       3.966 * cos(4.01667977047  + 243.59919371200 *T)
    sumX0  +=       2.935 * cos(4.99774337126  + 1190.54206257560 *T)
    sumX0  +=       3.497 * cos(0.14584351149  + 3114.53934080610 *T)
    sumX0  +=       2.871 * cos(3.51272688765  + 743.54939220120 *T)
    sumX0  +=       2.948 * cos(3.53416941444  + 78338.73484077338 *T)
    sumX0  +=       3.905 * cos(4.31421378967  + 669.93675215711 *T)
    sumX0  +=       2.937 * cos(0.29648642653  + 590.48928714380 *T)
    sumX0  +=       4.039 * cos(3.00223087435  + 25633.23759253039 *T)
    sumX0  +=       3.590 * cos(0.30018156799  + 376.70255673880 *T)
    sumX0  +=       2.945 * cos(6.16606883810  + 274.30986580831 *T)
    sumX0  +=       3.853 * cos(2.68183358602  + 224.69167904940 *T)
    sumX0  +=       3.805 * cos(4.38433280473  + 3117.50828622269 *T)
    sumX0  +=       2.874 * cos(2.09548901827  + 89.35557591640 *T)
    sumX0  +=       3.560 * cos(2.89480416655  + 748.34168744680 *T)
    sumX0  +=       2.905 * cos(0.11587901518  + 872.66572021069 *T)
    sumX0  +=       3.186 * cos(3.01587158229  + 53.87597151011 *T)
    sumX0  +=       3.567 * cos(1.99291047052  + 403.37800970800 *T)
    sumX0  +=       3.734 * cos(5.74700811308  + 909.46304098649 *T)
    sumX0  +=       3.317 * cos(3.32499008438  + 204.45725824539 *T)
    sumX0  +=       3.185 * cos(5.79949438148  + 5982.22116553039 *T)
    sumX0  +=       2.821 * cos(2.97445175627  + 116.78178841100 *T)
    sumX0  +=       2.821 * cos(2.97445175627  + 116.29415344400 *T)
    sumX0  +=       2.924 * cos(5.47085827815  + 143.90275367970 *T)
    sumX0  +=       3.219 * cos(3.73062920727  + 76.82540294419 *T)
    sumX0  +=       2.902 * cos(6.02969827171  + 102.58583217670 *T)
    sumX0  +=       2.902 * cos(1.79129681788  + 47.46499992491 *T)
    sumX0  +=       3.255 * cos(1.03586814463  + 482.61302640021 *T)
    sumX0  +=       2.994 * cos(3.55361890559  + 162.39206934450 *T)
    sumX0  +=       3.717 * cos(4.95057674227  + 6061.45618222260 *T)
    sumX0  +=       3.942 * cos(0.04160332713  + 6096.62027244380 *T)
    sumX0  +=       3.586 * cos(3.03514042359  + 983.35967639710 *T)
    sumX0  +=       2.803 * cos(3.13310905329  + 644.07325706060 *T)
    sumX0  +=       2.802 * cos(1.97945311635  + 1592.28565818390 *T)
    sumX0  +=       3.498 * cos(5.94376047053  + 232.29286101720 *T)
    sumX0  +=       3.387 * cos(4.62847819573  + 25861.83005568049 *T)
    sumX0  +=       3.049 * cos(2.94693150681  + 6529.93336607319 *T)
    sumX0  +=       2.910 * cos(4.43764615840  + 355.99256305531 *T)
    sumX0  +=       3.266 * cos(4.88740891863  + 265.48484356309 *T)
    sumX0  +=       3.868 * cos(5.48362456906  + 2124.73981801839 *T)
    sumX0  +=       3.667 * cos(5.22566104677  + 415.90818268021 *T)
    sumX0  +=       2.825 * cos(5.24697205919  + 294.31728407620 *T)
    sumX0  +=       2.775 * cos(4.33278706900  + 552.02618310591 *T)
    sumX0  +=       3.099 * cos(2.76412505763  + 436.50630177909 *T)
    sumX0  +=       2.858 * cos(3.66116683929  + 529.88659846831 *T)
    sumX0  +=       3.700 * cos(2.58340975985  + 25864.79900109709 *T)
    sumX0  +=       2.855 * cos(0.67966744335  + 529.98296668790 *T)
    sumX0  +=       2.805 * cos(5.17536800845  + 2089.05197080790 *T)
    sumX0  +=       2.766 * cos(2.53840571741  + 1600.36241303120 *T)
    sumX0  +=       3.634 * cos(3.00439582111  + 77.88248688280 *T)
    sumX0  +=       3.634 * cos(1.67500661489  + 72.16834521881 *T)
    sumX0  +=       2.849 * cos(4.00550189531  + 323.48860171000 *T)
    sumX0  +=       2.774 * cos(3.80513617353  + 1606.09455318770 *T)
    sumX0  +=       2.788 * cos(5.07162517305  + 187.03172416360 *T)
    sumX0  +=       2.971 * cos(3.10045016404  + 586.13318938849 *T)
    sumX0  +=       2.901 * cos(2.38225620488  + 62.52887297340 *T)
    sumX0  +=       3.856 * cos(2.65955588851  + 113.21333344280 *T)
    sumX0  +=       2.957 * cos(3.89188234418  + 53.82778740031 *T)
    sumX0  +=       2.741 * cos(4.81447095359  + 274.34349570310 *T)
    sumX0  +=       2.895 * cos(5.22229317128  + 151.60700151150 *T)
    sumX0  +=       2.827 * cos(4.49360284571  + 979.21507481291 *T)
    sumX0  +=       3.178 * cos(2.47353230626  + 529.67415014720 *T)
    sumX0  +=       3.613 * cos(1.91993731589  + 583.65187893891 *T)
    sumX0  +=       3.096 * cos(1.82324058715  + 530.19541500900 *T)
    sumX0  +=       3.630 * cos(0.09561817703  + 9991.66587844219 *T)
    sumX0  +=       2.778 * cos(1.48524423298  + 465.02992575149 *T)
    sumX0  +=       2.899 * cos(5.75092160965  + 428.32648106781 *T)
    sumX0  +=       2.977 * cos(2.67097238665  + 74.37772238940 *T)
    sumX0  +=       3.543 * cos(0.23405199750  + 764.67758488250 *T)
    sumX0  +=       2.975 * cos(4.20884044975  + 224.42855449101 *T)
    sumX0  +=       3.288 * cos(0.65434532542  + 201.04636984410 *T)
    sumX0  +=       3.824 * cos(3.93883782891  + 826.92901047760 *T)
    sumX0  +=       3.019 * cos(3.67922796466  + 829.16424271130 *T)
    sumX0  +=       2.780 * cos(5.61085980857  + 64.45528866639 *T)
    sumX0  +=       3.345 * cos(3.85271375822  + 170.46882419179 *T)
    sumX0  +=       3.345 * cos(0.71112110463  + 170.95645915881 *T)
    sumX0  +=       3.500 * cos(5.14891828595  + 9915.39980716659 *T)
    sumX0  +=       3.048 * cos(6.03618648483  + 409.89204715720 *T)
    sumX0  +=       2.854 * cos(3.24258254247  + 112.25012559630 *T)
    sumX0  +=       2.775 * cos(1.35827435388  + 395.34943897051 *T)
    sumX0  +=       2.715 * cos(5.94386708396  + 1533.49751061680 *T)
    sumX0  +=       2.941 * cos(1.40585778661  + 148.11009358871 *T)
    sumX0  +=       2.946 * cos(1.94087119383  + 430.77416162260 *T)
    sumX0  +=       3.426 * cos(1.74548379376  + 914.72841698410 *T)
    sumX0  +=       3.161 * cos(5.61863271879  + 554.42567955090 *T)
    sumX0  +=       2.968 * cos(4.51598961684  + 1542.09553032591 *T)
    sumX0  +=       3.071 * cos(1.13353110856  + 1062.80714141041 *T)
    sumX0  +=       2.999 * cos(4.97086723334  + 132.82342352100 *T)
    sumX0  +=       3.514 * cos(3.30332079475  + 6355.13232062560 *T)
    sumX0  +=       2.745 * cos(1.55758551038  + 482.01351684860 *T)
    sumX0  +=       2.850 * cos(3.54616507843  + 395.66495315560 *T)
    sumX0  +=       2.663 * cos(4.35706294828  + 96.48367713221 *T)
    sumX0  +=       3.000 * cos(3.14530124801  + 519.85229016070 *T)
    sumX0  +=       2.655 * cos(1.19029556060  + 1160.58635552159 *T)
    sumX0  +=       3.001 * cos(3.75883031715  + 253.81481257341 *T)
    sumX0  +=       2.983 * cos(0.93935356188  + 9912.43086174999 *T)
    sumX0  +=       2.641 * cos(4.81011847825  + 487.39651292520 *T)
    sumX0  +=       3.712 * cos(1.46420419890  + 10026.82996866339 *T)
    sumX0  +=       3.062 * cos(3.91142249915  + 789.70430093210 *T)
    sumX0  +=       3.147 * cos(1.22318621690  + 394.70174530910 *T)
    sumX0  +=       2.712 * cos(4.23397123966  + 78.16111780620 *T)
    sumX0  +=       2.692 * cos(1.62673300086  + 161.16822906709 *T)
    sumX0  +=       3.303 * cos(1.79828372577  + 7.94806531530 *T)
    sumX0  +=       2.747 * cos(2.81204346965  + 1372.40097825130 *T)
    sumX0  +=       3.011 * cos(2.22464119257  + 1605.83142862930 *T)
    sumX0  +=       2.807 * cos(5.73736217285  + 208.12628695010 *T)
    sumX0  +=       2.631 * cos(5.99241143684  + 271.64973693240 *T)
    sumX0  +=       3.688 * cos(2.66246767657  + 26543.05632558499 *T)
    sumX0  +=       3.416 * cos(3.60662433619  + 331.07772159029 *T)
    sumX0  +=       2.951 * cos(2.78355300655  + 12564.91104475801 *T)
    sumX0  +=       2.727 * cos(0.43905805953  + 1064.07916579760 *T)
    sumX0  +=       2.902 * cos(3.07188287999  + 1540.87418217599 *T)
    sumX0  +=       2.728 * cos(4.59132510744  + 109.55636682560 *T)
    sumX0  +=       2.999 * cos(5.77659280033  + 219.27505307800 *T)
    sumX0  +=       2.764 * cos(1.07670855016  + 142.20583265029 *T)
    sumX0  +=       2.586 * cos(2.07825454401  + 959.81743305580 *T)
    sumX0  +=       2.719 * cos(0.45061432224  + 52325.61329776649 *T)
    sumX0  +=       3.042 * cos(5.84610070395  + 752.79510557170 *T)
    sumX0  +=       3.089 * cos(0.78770097881  + 93.28772917949 *T)
    sumX0  +=       3.647 * cos(5.21958279042  + 697.56335088530 *T)
    sumX0  +=       3.492 * cos(2.77297949310  + 580.91350537320 *T)
    sumX0  +=       3.597 * cos(0.24187824789  + 132.28090679980 *T)
    sumX0  +=       2.974 * cos(3.91006313529  + 444.58305662640 *T)
    sumX0  +=       2.572 * cos(2.67254064624  + 2117.52320515360 *T)
    sumX0  +=       2.572 * cos(2.59845311288  + 23.33205575259 *T)
    sumX0  +=       3.167 * cos(3.77971757660  + 12567.87999017460 *T)
    sumX0  +=       3.382 * cos(4.72616841526  + 10285.34201684520 *T)
    sumX0  +=       2.555 * cos(4.34804949360  + 678.11657286839 *T)
    sumX0  +=       3.300 * cos(2.15244356490  + 440.65090336330 *T)
    sumX0  +=       2.831 * cos(4.29539679381  + 540.01727499551 *T)
    sumX0  +=       3.189 * cos(3.69749797847  + 769.86363900301 *T)
    sumX0  +=       3.190 * cos(5.54179110621  + 2424.35384725460 *T)
    sumX0  +=       3.280 * cos(0.39508288314  + 86.38663049980 *T)
    sumX0  +=       2.780 * cos(3.71874153588  + 1586.34776735071 *T)
    sumX0  +=       2.594 * cos(4.89957913206  + 1373.88545095959 *T)
    sumX0  +=       2.782 * cos(1.02745709962  + 1268.99268984760 *T)
    sumX0  +=       3.010 * cos(4.04370467440  + 1543.84312759260 *T)
    sumX0  +=       3.022 * cos(4.32583593317  + 181.29958400710 *T)
    sumX0  +=       3.022 * cos(1.18424327958  + 180.81194904009 *T)
    sumX0  +=       2.833 * cos(2.10884741514  + 555.79827767460 *T)
    sumX0  +=       3.293 * cos(0.26036305872  + 463.49057128900 *T)
    sumX0  +=       3.316 * cos(2.51263252063  + 807.99667464080 *T)
    sumX0  +=       3.300 * cos(0.73760471241  + 649.80539721709 *T)
    sumX0  +=       2.983 * cos(0.24641616415  + 936.40161851410 *T)
    sumX0  +=       2.865 * cos(4.91220221036  + 2192.40786958489 *T)
    sumX0  +=       3.356 * cos(5.11523694385  + 247.58622872930 *T)
    sumX0  +=       2.506 * cos(5.30544660690  + 40.85816357090 *T)
    sumX0  +=       3.114 * cos(5.11220642057  + 1553.89199011520 *T)
    sumX0  +=       3.268 * cos(0.04975503988  + 737.66849419841 *T)
    sumX0  +=       3.100 * cos(4.07790490748  + 657.66721161581 *T)
    sumX0  +=       2.510 * cos(1.43422723324  + 20278.24855051219 *T)
    sumX0  +=       2.833 * cos(3.12964283648  + 539.81152431880 *T)
    sumX0  +=       2.554 * cos(5.25801939266  + 826.19529729470 *T)
    sumX0  +=       2.806 * cos(5.71817066352  + 351.94853520760 *T)
    sumX0  +=       2.633 * cos(0.82117751455  + 1448.66704952690 *T)
    sumX0  +=       2.482 * cos(2.35351578734  + 1159.10188281330 *T)
    sumX0  +=       3.088 * cos(3.64065375900  + 245.78624183590 *T)
    sumX0  +=       3.458 * cos(4.58072564831  + 809.48114734910 *T)
    sumX0  +=       3.385 * cos(0.97492846896  + 1056.44435393500 *T)
    sumX0  +=       3.050 * cos(2.24338368884  + 52174.56562792359 *T)
    sumX0  +=       2.571 * cos(2.13979409339  + 584.58502620540 *T)
    sumX0  +=       2.506 * cos(6.23219506154  + 381.88968804659 *T)
    sumX0  +=       2.718 * cos(2.48384899075  + 352.17228437681 *T)
    sumX0  +=       3.112 * cos(5.17420301794  + 481.12855369190 *T)
    sumX0  +=       2.427 * cos(0.46160779810  + 20275.27960509559 *T)
    sumX0  +=       2.660 * cos(3.11456525410  + 587.55397162200 *T)
    sumX0  +=       3.031 * cos(2.65730851569  + 106.62105130379 *T)
    sumX0  +=       2.876 * cos(4.25326612876  + 837.76226242039 *T)
    sumX0  +=       2.876 * cos(4.25326612876  + 838.24989738740 *T)
    sumX0  +=       2.493 * cos(4.86549248982  + 339.53023682000 *T)
    sumX0  +=       3.071 * cos(2.94134266397  + 264.74863825270 *T)
    sumX0  +=       2.956 * cos(4.36010712190  + 561.64229241569 *T)
    sumX0  +=       2.956 * cos(4.36010712190  + 562.12992738271 *T)
    sumX0  +=       2.675 * cos(4.82615405372  + 257.66872114670 *T)
    sumX0  +=       2.412 * cos(0.21644968354  + 537.15139544289 *T)
    sumX0  +=       2.425 * cos(5.98792108916  + 243.65407546620 *T)
    sumX0  +=       3.094 * cos(0.89516329895  + 911.75947156750 *T)
    sumX0  +=       2.916 * cos(5.65344305695  + 627.12329585830 *T)
    sumX0  +=       2.461 * cos(0.76674192255  + 672.69994689699 *T)
    sumX0  +=       3.109 * cos(6.09716177778  + 45.27795180100 *T)
    sumX0  +=       3.239 * cos(0.98686936775  + 412.33972771199 *T)
    sumX0  +=       2.930 * cos(1.53043345203  + 173.38288785421 *T)
    sumX0  +=       2.377 * cos(2.36768270315  + 1927.95543844240 *T)
    sumX0  +=       2.783 * cos(2.17600324391  + 888.96798775160 *T)
    sumX0  +=       2.572 * cos(0.26643251840  + 1536.46645603341 *T)
    sumX0  +=       2.447 * cos(1.98626338075  + 1327.57508673799 *T)
    sumX0  +=       2.533 * cos(2.95752298614  + 1330.54403215460 *T)
    sumX0  +=       2.951 * cos(2.80925054018  + 328.59641114070 *T)
    sumX0  +=       2.474 * cos(2.81613373048  + 1407.51018671831 *T)
    sumX0  +=       2.518 * cos(3.82616998392  + 175.57663362249 *T)
    sumX0  +=       2.678 * cos(4.15340303192  + 180.40581212981 *T)
    sumX0  +=       3.251 * cos(1.90510984540  + 83.41098743880 *T)
    sumX0  +=       2.327 * cos(0.66127447858  + 325.54922103430 *T)
    sumX0  +=       2.802 * cos(0.16897738467  + 121.81135315189 *T)
    sumX0  +=       2.675 * cos(5.49343148274  + 105.62421356250 *T)
    sumX0  +=       2.909 * cos(0.78041431577  + 651.18680406140 *T)
    sumX0  +=       2.358 * cos(1.30452169405  + 377.87821290640 *T)
    sumX0  +=       2.780 * cos(3.88186933893  + 291.29345690540 *T)
    sumX0  +=       2.876 * cos(5.63710538659  + 6813.01063256950 *T)
    sumX0  +=       2.523 * cos(2.55630283191  + 440.91153579420 *T)
    sumX0  +=       3.239 * cos(3.10119943091  + 724.84762821529 *T)
    sumX0  +=       2.532 * cos(2.35901341471  + 39.34006096781 *T)
    sumX0  +=       2.503 * cos(3.49357090208  + 360.30392097830 *T)
    sumX0  +=       2.568 * cos(3.49872926031  + 1699.52303398670 *T)
    sumX0  +=       2.570 * cos(5.29329193342  + 299.47327761669 *T)
    sumX0  +=       3.223 * cos(4.19015476099  + 442.39600850250 *T)
    sumX0  +=       2.316 * cos(2.75922843554  + 714.92270236481 *T)
    sumX0  +=       2.457 * cos(1.15216748085  + 552.82933225800 *T)
    sumX0  +=       2.319 * cos(0.77064178822  + 290.12449838219 *T)
    sumX0  +=       2.922 * cos(4.03553902122  + 1600.20235433680 *T)
    sumX0  +=       2.297 * cos(1.88640861087  + 1811.52934209949 *T)
    sumX0  +=       2.674 * cos(1.18404012894  + 519.26158932971 *T)
    sumX0  +=       2.585 * cos(1.04419941319  + 754.27957828001 *T)
    sumX0  +=       2.277 * cos(6.00831997160  + 386.10372559999 *T)
    sumX0  +=       2.326 * cos(3.39783484366  + 172.80463016199 *T)
    sumX0  +=       3.163 * cos(2.18762574676  + 476.18750059999 *T)
    sumX0  +=       3.127 * cos(5.74767332186  + 652.98679095480 *T)
    sumX0  +=       3.015 * cos(4.60083877410  + 3190.80541208169 *T)
    sumX0  +=       3.015 * cos(4.60083877410  + 3191.29304704871 *T)
    sumX0  +=       2.772 * cos(4.21896384049  + 431.08967580769 *T)
    sumX0  +=       2.329 * cos(1.38342611870  + 759.01488069520 *T)
    sumX0  +=       2.334 * cos(2.76557699620  + 837.97471074149 *T)
    sumX0  +=       2.991 * cos(0.49950728589  + 458.81015062800 *T)
    sumX0  +=       3.166 * cos(4.75068391337  + 26161.44408491669 *T)
    sumX0  +=       2.566 * cos(4.39543350128  + 63.23144838899 *T)
    sumX0  +=       2.282 * cos(5.65971495881  + 811.21744376770 *T)
    sumX0  +=       2.423 * cos(3.92143505564  + 103.10958916600 *T)
    sumX0  +=       3.125 * cos(4.47876249023  + 939.37056393069 *T)
    sumX0  +=       2.277 * cos(3.94882643258  + 0.27744737829 *T)
    sumX0  +=       2.982 * cos(3.78210055675  + 37.85558825950 *T)
    sumX0  +=       2.982 * cos(3.78210055675  + 37.36795329250 *T)
    sumX0  +=       2.437 * cos(0.38702218640  + 377.12075573661 *T)
    sumX0  +=       2.672 * cos(6.00980899009  + 412.07909528109 *T)
    sumX0  +=       2.289 * cos(1.03807449021  + 1258.69774910910 *T)
    sumX0  +=       2.259 * cos(4.68066906641  + 113.63153244060 *T)
    sumX0  +=       3.098 * cos(4.89446319562  + 612.42062897720 *T)
    sumX0  +=       3.098 * cos(1.75287054203  + 611.93299401020 *T)
    sumX0  +=       3.157 * cos(4.46324746103  + 3795.76561071060 *T)
    sumX0  +=       2.795 * cos(0.77242359455  + 10743.22032878910 *T)
    sumX0  +=       2.777 * cos(4.75646431289  + 211.09523236671 *T)
    sumX0  +=       3.088 * cos(5.58047807216  + 230.52650401861 *T)
    sumX0  +=       2.940 * cos(3.51648675050  + 714.66206993390 *T)
    sumX0  +=       2.685 * cos(2.97936141724  + 303.82362987761 *T)
    sumX0  +=       2.260 * cos(4.95371521477  + 124.20415195250 *T)
    sumX0  +=       3.092 * cos(0.26856203052  + 3414.15337004230 *T)
    sumX0  +=       3.015 * cos(1.58292445535  + 342.49918223660 *T)
    sumX0  +=       3.083 * cos(3.86343122317  + 206.95063078250 *T)
    sumX0  +=       3.083 * cos(0.72183856958  + 206.46299581550 *T)
    sumX0  +=       2.331 * cos(5.55613356200  + 820.56622300220 *T)
    sumX0  +=       2.712 * cos(1.33807664445  + 770.30665917500 *T)
    sumX0  +=       2.587 * cos(5.71194398470  + 636.20895053440 *T)
    sumX0  +=       2.587 * cos(2.57035133111  + 635.72131556739 *T)
    sumX0  +=       2.803 * cos(2.76934913754  + 25938.09612695609 *T)
    sumX0  +=       2.803 * cos(2.76934913754  + 25938.58376192310 *T)
    sumX0  +=       2.405 * cos(5.05357423584  + 488.60579898760 *T)
    sumX0  +=       2.359 * cos(2.98384806596  + 87.06795405600 *T)
    sumX0  +=       2.160 * cos(3.91324321531  + 199.52826724100 *T)
    sumX0  +=       2.362 * cos(5.50921508396  + 1222.03463196460 *T)
    sumX0  +=       2.362 * cos(5.50921508396  + 1222.52226693161 *T)
    sumX0  +=       2.740 * cos(0.85654894869  + 627.61093082531 *T)
    sumX0  +=       2.798 * cos(2.35367531012  + 771.34811171130 *T)
    sumX0  +=       2.798 * cos(5.49526796371  + 770.86047674430 *T)
    sumX0  +=       2.435 * cos(2.36446205622  + 350.08830211689 *T)
    sumX0  +=       2.357 * cos(1.46842067912  + 355.24180352990 *T)
    sumX0  +=       2.485 * cos(3.78369829663  + 142.41828097139 *T)
    sumX0  +=       2.521 * cos(1.03497131510  + 129.67565967810 *T)
    sumX0  +=       2.089 * cos(5.45017034742  + 74.58596519360 *T)
    sumX0  +=       2.302 * cos(3.23730533014  + 52177.04693837319 *T)
    sumX0  +=       2.302 * cos(3.23730533014  + 52177.53457334019 *T)
    sumX0  +=       2.619 * cos(4.76291733303  + 1321.68288788710 *T)
    sumX0  +=       2.572 * cos(3.16528157166  + 659.43031524749 *T)
    sumX0  +=       2.225 * cos(1.53417972319  + 59.52629766200 *T)
    sumX0  +=       2.835 * cos(0.16315666564  + 59.34747923569 *T)
    sumX0  +=       2.282 * cos(5.64376298083  + 295.91363136910 *T)
    sumX0  +=       1.992 * cos(0.17905928659  + 765.64079272899 *T)
    sumX0  +=       1.992 * cos(3.32065194018  + 766.12842769600 *T)
    sumX0  +=       2.574 * cos(3.32754474496  + 192.56022471960 *T)
    sumX0  +=       2.264 * cos(4.07388420311  + 75.05904594560 *T)
    sumX0  +=       2.089 * cos(2.85631372853  + 74.48959697400 *T)
    sumX0  +=       2.032 * cos(4.97430667920  + 215.68177693560 *T)
    sumX0  +=       1.982 * cos(3.59359575315  + 898.52921530720 *T)
    sumX0  +=       1.982 * cos(0.45200309956  + 899.01685027420 *T)
    sumX0  +=       2.374 * cos(4.72352640183  + 550.96909916730 *T)
    sumX0  +=       2.374 * cos(4.72352640183  + 551.45673413431 *T)
    sumX0  +=       2.352 * cos(0.06708062177  + 24.36220744081 *T)
    sumX0  +=       2.591 * cos(3.75715375752  + 367.72628451979 *T)
    sumX0  +=       2.230 * cos(1.52395191838  + 329.96900926440 *T)
    sumX0  +=       2.610 * cos(2.73849324439  + 66.15220969580 *T)
    sumX0  +=       2.610 * cos(2.73849324439  + 66.63984466281 *T)
    sumX0  +=       1.942 * cos(5.47218005624  + 74.69783977820 *T)
    sumX0  +=       2.217 * cos(2.71427205276  + 353.05724753350 *T)
    sumX0  +=       2.034 * cos(1.30593331214  + 527.96267490280 *T)
    sumX0  +=       2.034 * cos(1.30593331214  + 528.45030986980 *T)
    sumX0  +=       2.025 * cos(1.90306188675  + 94.98465020890 *T)
    sumX0  +=       2.220 * cos(0.06407550248  + 160.36507991500 *T)
    sumX0  +=       2.205 * cos(5.46838417109  + 460.78225830330 *T)
    sumX0  +=       2.205 * cos(5.46838417109  + 460.29462333630 *T)
    sumX0  +=       1.903 * cos(5.26788824520  + 170.51700830159 *T)
    sumX0  +=       1.903 * cos(5.26788824520  + 171.00464326860 *T)
    sumX0  +=       1.880 * cos(5.91981669441  + 845.83901726770 *T)
    sumX0  +=       1.879 * cos(1.52960331705  + 810.41429461559 *T)
    sumX0  +=       1.918 * cos(4.29742245145  + 980.42436087530 *T)
    sumX0  +=       1.918 * cos(1.15582979786  + 980.91199584231 *T)
    sumX0  +=       2.012 * cos(4.58948418435  + 482.97672469510 *T)
    sumX0  +=       2.408 * cos(3.65417851763  + 673.56008887950 *T)
    sumX0  +=       2.408 * cos(3.65417851763  + 673.07245391250 *T)
    sumX0  +=       1.917 * cos(4.50882706137  + 464.22677659940 *T)
    sumX0  +=       1.917 * cos(4.50882706137  + 464.71441156640 *T)
    sumX0  +=       2.523 * cos(3.82677029026  + 207.11738712130 *T)
    sumX0  +=       2.235 * cos(4.73973227208  + 156.28742217250 *T)
    sumX0  +=       1.889 * cos(6.20777223595  + 96.62918161160 *T)
    sumX0  +=       2.163 * cos(4.45487815975  + 1325.12740618320 *T)
    sumX0  +=       2.163 * cos(1.31328550616  + 1325.61504115020 *T)
    sumX0  +=       2.041 * cos(4.71231705875  + 753.38580640270 *T)
    sumX0  +=       2.385 * cos(0.51948475769  + 589.10213480510 *T)
    sumX0  +=       2.385 * cos(3.66107741128  + 589.58976977210 *T)
    sumX0  +=       2.320 * cos(4.52095470178  + 153.77279777600 *T)
    sumX0  +=       1.825 * cos(4.34973286546  + 241.62708603669 *T)
    sumX0  +=       1.938 * cos(5.90510178691  + 989.98558843089 *T)
    sumX0  +=       1.938 * cos(5.90510178691  + 990.47322339791 *T)
    sumX0  +=       1.801 * cos(1.59469951538  + 386.47623261550 *T)
    sumX0  +=       1.801 * cos(1.59469951538  + 386.96386758251 *T)
    sumX0  +=       1.870 * cos(0.33398234764  + 814.08332332030 *T)
    sumX0  +=       1.870 * cos(0.33398234764  + 813.59568835330 *T)
    sumX0  +=       1.820 * cos(2.95417643509  + 194.42045781031 *T)
    sumX0  +=       2.331 * cos(3.49571982266  + 229.20878764910 *T)
    sumX0  +=       1.846 * cos(5.53844865127  + 1310.63718762320 *T)
    sumX0  +=       1.813 * cos(3.46850679276  + 242.85092631409 *T)
    sumX0  +=       1.813 * cos(0.32691413917  + 243.33856128110 *T)
    sumX0  +=       2.125 * cos(0.90702543941  + 927.73416283579 *T)
    sumX0  +=       1.841 * cos(0.75141045449  + 114.15528942990 *T)
    sumX0  +=       1.770 * cos(5.79168127546  + 113.89465699899 *T)
    sumX0  +=       1.797 * cos(1.81953146289  + 286.35240381350 *T)
    sumX0  +=       2.348 * cos(0.72588619615  + 838.46234570850 *T)
    sumX0  +=       2.341 * cos(0.85296576676  + 80.47567191700 *T)
    sumX0  +=       2.214 * cos(5.39273615306  + 355.50492808829 *T)
    sumX0  +=       1.986 * cos(5.86961068002  + 180.55131660920 *T)
    sumX0  +=       1.742 * cos(6.12132593481  + 380.37158544351 *T)
    sumX0  +=       1.858 * cos(5.61326911538  + 1471.50920958009 *T)
    sumX0  +=       1.858 * cos(5.61326911538  + 1471.99684454710 *T)
    sumX0  +=       1.883 * cos(3.44892554698  + 102.76714273050 *T)
    sumX0  +=       2.209 * cos(4.52058460762  + 658.40341692619 *T)
    sumX0  +=       2.209 * cos(1.37899195403  + 658.89105189320 *T)
    sumX0  +=       1.942 * cos(2.25463899399  + 703.38936713389 *T)
    sumX0  +=       1.915 * cos(3.44323535772  + 423.17297965479 *T)
    sumX0  +=       1.915 * cos(3.44323535772  + 423.66061462181 *T)
    sumX0  +=       2.166 * cos(3.23067903187  + 79.53371592989 *T)
    sumX0  +=       1.755 * cos(2.27265833368  + 152.50077338879 *T)
    sumX0  +=       1.700 * cos(4.83117512138  + 50.15875869559 *T)
    sumX0  +=       2.051 * cos(2.99945701352  + 807.96661406080 *T)
    sumX0  +=       2.051 * cos(2.99945701352  + 808.45424902780 *T)
    sumX0  +=       1.839 * cos(5.75757356379  + 756.56720014041 *T)
    sumX0  +=       1.682 * cos(4.06333599913  + 1049.57688515010 *T)
    sumX0  +=       1.682 * cos(0.92174334554  + 1050.06452011710 *T)
    sumX0  +=       1.706 * cos(4.77079031354  + 832.34563644899 *T)
    sumX0  +=       1.706 * cos(4.77079031354  + 832.83327141600 *T)
    sumX0  +=       1.776 * cos(3.76195690472  + 1384.93115122349 *T)
    sumX0  +=       1.776 * cos(0.62036425113  + 1385.41878619050 *T)
    sumX0  +=       1.811 * cos(4.64985456293  + 454.55367445919 *T)
    sumX0  +=       1.811 * cos(4.64985456293  + 455.04130942620 *T)
    sumX0  +=       1.692 * cos(4.72965893305  + 271.16210196539 *T)
    sumX0  +=       2.206 * cos(6.17339592873  + 132.33578855399 *T)
    sumX0  +=       1.658 * cos(3.45028228993  + 184.33796539290 *T)
    sumX0  +=       1.875 * cos(5.27331155633  + 1282.58913653029 *T)
    sumX0  +=       1.875 * cos(2.13171890274  + 1283.07677149730 *T)
    sumX0  +=       1.686 * cos(0.45014900270  + 406.71065341949 *T)
    sumX0  +=       1.686 * cos(0.45014900270  + 407.19828838650 *T)
    sumX0  +=       2.083 * cos(4.25600154596  + 119.63885924300 *T)
    sumX0  +=       2.240 * cos(3.90434509762  + 404.37484744930 *T)
    sumX0  +=       1.952 * cos(0.24163677359  + 458.33457774851 *T)
    sumX0  +=       1.952 * cos(0.24163677359  + 457.84694278150 *T)
    sumX0  +=       2.288 * cos(5.51294457419  + 135.30473397060 *T)
    sumX0  +=       1.949 * cos(3.26830580660  + 1056.16916728910 *T)
    sumX0  +=       1.949 * cos(0.12671315301  + 1056.65680225610 *T)
    sumX0  +=       1.974 * cos(1.88730039824  + 269.93826168799 *T)
    sumX0  +=       1.974 * cos(5.02889305183  + 270.42589665501 *T)
    sumX0  +=       1.659 * cos(5.45360392276  + 1450.31158092960 *T)
    sumX0  +=       1.659 * cos(5.45360392276  + 1450.79921589661 *T)
    sumX0  +=       2.309 * cos(0.71286416918  + 74.79841351470 *T)
    sumX0  +=       1.644 * cos(3.45713257957  + 321.51649403470 *T)
    sumX0  +=       1.644 * cos(0.31553992598  + 322.00412900171 *T)
    sumX0  +=       1.656 * cos(0.89488112289  + 516.91697463890 *T)
    sumX0  +=       1.656 * cos(0.89488112289  + 517.40460960590 *T)
    sumX0  +=       2.034 * cos(1.27659765222  + 274.87931477991 *T)
    sumX0  +=       2.034 * cos(1.27659765222  + 274.39167981289 *T)
    sumX0  +=       2.154 * cos(3.29965728228  + 516.29264391310 *T)
    sumX0  +=       2.154 * cos(0.15806462869  + 515.80500894610 *T)
    sumX0  +=       1.711 * cos(1.53733627594  + 885.99904233499 *T)
    sumX0  +=       1.711 * cos(1.53733627594  + 886.48667730201 *T)
    sumX0  +=       1.826 * cos(2.44969882307  + 398.43695661610 *T)
    sumX0  +=       1.826 * cos(2.44969882307  + 398.92459158310 *T)
    sumX0  +=       1.966 * cos(4.10423736234  + 15.90969221110 *T)
    sumX0  +=       1.966 * cos(0.96264470875  + 16.39732717810 *T)
    sumX0  +=       1.678 * cos(3.69519426870  + 1182.67775604939 *T)
    sumX0  +=       1.678 * cos(0.55360161511  + 1183.16539101641 *T)
    sumX0  +=       2.098 * cos(4.85173961477  + 630.09224127489 *T)
    sumX0  +=       2.098 * cos(1.71014696118  + 630.57987624190 *T)
    sumX0  +=       2.239 * cos(2.53095817704  + 446.55516430170 *T)
    sumX0  +=       2.239 * cos(2.53095817704  + 446.06752933469 *T)
    sumX0  +=       1.889 * cos(1.07863762669  + 86.69544704049 *T)
    sumX0  +=       1.889 * cos(1.07863762669  + 87.18308200751 *T)
    sumX0  +=       1.759 * cos(3.42836121936  + 614.11755000660 *T)
    sumX0  +=       1.759 * cos(0.28676856577  + 613.62991503960 *T)
    sumX0  +=       1.673 * cos(3.25241882006  + 231.06902073981 *T)
    sumX0  +=       1.607 * cos(5.76189875081  + 412.12727939089 *T)
    sumX0  +=       1.607 * cos(2.62030609722  + 412.61491435790 *T)
    sumX0  +=       2.070 * cos(3.82797801425  + 824.23525170691 *T)
    sumX0  +=       1.835 * cos(3.95891079221  + 146.87169909629 *T)
    sumX0  +=       1.580 * cos(4.75553561047  + 2002.73703700969 *T)
    sumX0  +=       1.580 * cos(1.61394295688  + 2003.22467197670 *T)
    sumX0  +=       1.979 * cos(5.88730379340  + 407.23191828129 *T)
    sumX0  +=       1.979 * cos(5.88730379340  + 407.71955324831 *T)
    sumX0  +=       1.764 * cos(5.38315884955  + 544.75257741070 *T)
    sumX0  +=       1.815 * cos(2.11042594430  + 174.92224231670 *T)
    sumX0  +=       1.808 * cos(1.40200347339  + 454.77742362840 *T)
    sumX0  +=       1.808 * cos(1.40200347339  + 455.26505859540 *T)
    sumX0  +=       1.667 * cos(4.30345024075  + 409.67959883610 *T)
    sumX0  +=       1.667 * cos(1.16185758716  + 410.16723380310 *T)
    sumX0  +=       1.739 * cos(0.21471955910  + 271.42273439630 *T)
    sumX0  +=       1.739 * cos(0.21471955910  + 271.91036936330 *T)
    sumX0  +=       1.764 * cos(2.87177593963  + 625.42637482889 *T)
    sumX0  +=       1.764 * cos(6.01336859322  + 625.91400979590 *T)
    sumX0  +=       1.578 * cos(5.77840141165  + 1321.19525292009 *T)
    sumX0  +=       2.105 * cos(2.76721779750  + 348.12825652910 *T)
    sumX0  +=       1.602 * cos(4.08797383059  + 1133.91971127299 *T)
    sumX0  +=       1.602 * cos(0.94638117700  + 1134.40734624000 *T)
    sumX0  +=       1.690 * cos(0.70280224085  + 2.24955505361 *T)
    sumX0  +=       1.690 * cos(0.70280224085  + 1.76192008660 *T)
    sumX0  +=       1.611 * cos(3.21423971704  + 1442.23482608230 *T)
    sumX0  +=       1.611 * cos(3.21423971704  + 1442.72246104930 *T)
    sumX0  +=       1.553 * cos(0.42212828385  + 550.85722458270 *T)
    sumX0  +=       1.553 * cos(0.42212828385  + 551.34485954970 *T)
    sumX0  +=       1.801 * cos(0.30318768980  + 450.47276334980 *T)
    sumX0  +=       1.801 * cos(3.44478034339  + 450.96039831681 *T)
    sumX0  +=       2.000 * cos(4.23137559040  + 239.14577558710 *T)
    sumX0  +=       2.000 * cos(4.23137559040  + 238.65814062010 *T)
    sumX0  +=       1.610 * cos(2.64143743565  + 1274.72483000410 *T)
    sumX0  +=       1.610 * cos(2.64143743565  + 1275.21246497110 *T)
    sumX0  +=       1.669 * cos(2.95148684971  + 839.24673512870 *T)
    sumX0  +=       1.669 * cos(2.95148684971  + 839.73437009571 *T)
    sumX0  +=       1.618 * cos(1.79025814071  + 376.91500505990 *T)
    sumX0  +=       1.618 * cos(1.79025814071  + 377.40264002690 *T)
    sumX0  +=       1.986 * cos(2.19226936272  + 980.39073098050 *T)
    sumX0  +=       1.986 * cos(5.33386201631  + 979.90309601349 *T)
    sumX0  +=       1.523 * cos(6.09198787084  + 67.94549894479 *T)
    sumX0  +=       1.625 * cos(0.48907210790  + 72.94143379089 *T)
    sumX0  +=       1.699 * cos(3.06470005459  + 418.27761854520 *T)
    sumX0  +=       1.699 * cos(3.06470005459  + 418.76525351221 *T)
    sumX0  +=       1.699 * cos(4.85438710621  + 991.47006113920 *T)
    sumX0  +=       1.699 * cos(4.85438710621  + 991.95769610620 *T)
    sumX0  +=       2.140 * cos(4.60009132464  + 468.89264304540 *T)
    sumX0  +=       2.140 * cos(1.45849867105  + 469.38027801241 *T)
    sumX0  +=       1.943 * cos(3.84559085346  + 55.52720055719 *T)
    sumX0  +=       1.887 * cos(1.39886028887  + 250.92768116140 *T)
    sumX0  +=       1.887 * cos(1.39886028887  + 251.41531612840 *T)
    sumX0  +=       1.500 * cos(2.05004005131  + 63.60395540450 *T)
    sumX0  +=       1.595 * cos(2.85027634129  + 440.17533048380 *T)
    sumX0  +=       1.595 * cos(2.85027634129  + 439.68769551680 *T)
    sumX0  +=       1.499 * cos(0.58296033612  + 309.03450517230 *T)
    sumX0  +=       1.499 * cos(3.72455298971  + 309.52214013930 *T)
    sumX0  +=       1.554 * cos(1.79217885440  + 759.04851059000 *T)
    sumX0  +=       1.554 * cos(4.93377150799  + 759.53614555700 *T)
    sumX0  +=       1.834 * cos(3.07143289100  + 1.45310354590 *T)
    sumX0  +=       1.834 * cos(3.07143289100  + 1.94073851290 *T)
    sumX0  +=       2.054 * cos(5.70588193100  + 549.97226142600 *T)
    sumX0  +=       2.054 * cos(2.56428927741  + 549.48462645899 *T)
    sumX0  +=       1.488 * cos(2.83790861898  + 1095.36598450990 *T)
    sumX0  +=       1.488 * cos(5.97950127257  + 1095.85361947691 *T)
    sumX0  +=       1.484 * cos(3.96898705243  + 68.12431737110 *T)
    sumX0  +=       1.818 * cos(4.87903340292  + 587.61766209680 *T)
    sumX0  +=       1.818 * cos(4.87903340292  + 588.10529706381 *T)
    sumX0  +=       1.714 * cos(3.60683578110  + 245.03548231051 *T)
    sumX0  +=       1.714 * cos(0.46524312751  + 244.54784734349 *T)
    sumX0  +=       1.920 * cos(2.50542705831  + 675.04456158780 *T)
    sumX0  +=       1.920 * cos(2.50542705831  + 674.55692662080 *T)
    sumX0  +=       1.504 * cos(1.02910554577  + 833.79647926250 *T)
    sumX0  +=       2.044 * cos(3.94263526988  + 6662.95980046789 *T)
    sumX0  +=       2.044 * cos(0.80104261629  + 6663.44743543490 *T)
    sumX0  +=       1.648 * cos(3.48244866747  + 438.54209992920 *T)
    sumX0  +=       1.466 * cos(0.25731804451  + 304.36614659881 *T)
    sumX0  +=       1.472 * cos(1.73963754908  + 1257.45935461670 *T)
    sumX0  +=       1.472 * cos(4.88123020267  + 1257.94698958370 *T)
    sumX0  +=       1.615 * cos(5.47975458652  + 984.35651413839 *T)
    sumX0  +=       1.625 * cos(3.50840756791  + 949.19242391720 *T)
    sumX0  +=       1.625 * cos(3.50840756791  + 949.68005888420 *T)
    sumX0  +=       1.715 * cos(1.35099907342  + 349.61272923740 *T)
    sumX0  +=       1.931 * cos(2.40467498858  + 1819.88128359270 *T)
    sumX0  +=       1.450 * cos(3.78546043220  + 493.75930040060 *T)
    sumX0  +=       1.450 * cos(0.64386777861  + 493.27166543360 *T)
    sumX0  +=       1.436 * cos(5.92306997669  + 430.02340209721 *T)
    sumX0  +=       1.436 * cos(2.78147732310  + 429.53576713020 *T)
    sumX0  +=       1.945 * cos(4.63182275077  + 131.68139724819 *T)
    sumX0  +=       1.431 * cos(4.50116477345  + 157.39613449839 *T)
    sumX0  +=       1.662 * cos(4.80337367540  + 6209.53490664889 *T)
    sumX0  +=       1.662 * cos(4.80337367540  + 6210.02254161590 *T)
    sumX0  +=       1.819 * cos(3.50918818351  + 1008.73553652659 *T)
    sumX0  +=       1.819 * cos(3.50918818351  + 1009.22317149361 *T)
    sumX0  +=       1.518 * cos(1.15354810143  + 328.67822514530 *T)
    sumX0  +=       1.518 * cos(4.29514075502  + 329.16586011230 *T)
    sumX0  +=       1.638 * cos(1.43556938440  + 519.88592005549 *T)
    sumX0  +=       1.638 * cos(4.57716203799  + 520.37355502251 *T)
    sumX0  +=       1.680 * cos(4.46803343205  + 359.64952967250 *T)
    sumX0  +=       1.680 * cos(4.46803343205  + 360.13716463950 *T)
    sumX0  +=       1.843 * cos(0.96654134734  + 478.15960827531 *T)
    sumX0  +=       1.843 * cos(4.10813400093  + 477.67197330830 *T)
    sumX0  +=       1.489 * cos(1.97136538075  + 400.49087829599 *T)
    sumX0  +=       1.489 * cos(1.97136538075  + 400.97851326301 *T)
    sumX0  +=       1.752 * cos(4.88043742079  + 6271.78633224399 *T)
    sumX0  +=       1.752 * cos(1.73884476720  + 6272.27396721100 *T)
    sumX0  +=       1.895 * cos(4.56971312882  + 1055.17232954780 *T)
    sumX0  +=       1.895 * cos(1.42812047523  + 1054.68469458080 *T)
    sumX0  +=       1.717 * cos(3.59655571468  + 536.56069461190 *T)
    sumX0  +=       1.758 * cos(1.02130969978  + 1095.05047032480 *T)
    sumX0  +=       1.758 * cos(4.16290235337  + 1094.56283535780 *T)
    sumX0  +=       1.721 * cos(4.15981655427  + 104.33342944339 *T)
    sumX0  +=       1.713 * cos(4.47285479740  + 621.49422156580 *T)
    sumX0  +=       1.460 * cos(3.44943623014  + 341.12658411290 *T)
    sumX0  +=       1.460 * cos(0.30784357655  + 340.63894914589 *T)
    sumX0  +=       1.780 * cos(5.09342203154  + 1668.83361374980 *T)
    sumX0  +=       1.780 * cos(5.09342203154  + 1668.34597878280 *T)
    sumX0  +=       1.401 * cos(2.09081885651  + 451.48415530609 *T)
    sumX0  +=       1.401 * cos(2.09081885651  + 451.97179027311 *T)
    sumX0  +=       1.394 * cos(1.22796389725  + 252.41215386970 *T)
    sumX0  +=       1.942 * cos(5.36682846266  + 10593.16949668749 *T)
    sumX0  +=       1.942 * cos(2.22523580907  + 10593.65713165450 *T)
    sumX0  +=       1.427 * cos(5.28628705446  + 228.21194990781 *T)
    sumX0  +=       1.427 * cos(2.14469440087  + 227.72431494079 *T)
    sumX0  +=       1.395 * cos(0.11161595772  + 697.76029284139 *T)
    sumX0  +=       1.395 * cos(3.25320861131  + 698.24792780840 *T)
    sumX0  +=       1.403 * cos(2.45154329677  + 281.24459438279 *T)
    sumX0  +=       1.403 * cos(5.59313595036  + 281.73222934980 *T)
    sumX0  +=       1.507 * cos(0.19523124567  + 193.89919294850 *T)
    sumX0  +=       1.401 * cos(2.18053804948  + 47.45044570990 *T)
    sumX0  +=       1.458 * cos(2.87243635260  + 214.19730422731 *T)
    sumX0  +=       1.408 * cos(3.27849545936  + 105.76971804189 *T)
    sumX0  +=       1.545 * cos(3.36846576185  + 912.54386098770 *T)
    sumX0  +=       1.545 * cos(3.36846576185  + 913.03149595471 *T)
    sumX0  +=       1.780 * cos(3.98299757592  + 623.19114259520 *T)
    sumX0  +=       1.780 * cos(0.84140492233  + 623.67877756220 *T)
    sumX0  +=       1.421 * cos(4.99710506357  + 1349.50642857140 *T)
    sumX0  +=       1.421 * cos(4.99710506357  + 1349.99406353840 *T)
    sumX0  +=       1.421 * cos(2.01327407687  + 1447.18257681859 *T)
    sumX0  +=       1.421 * cos(5.15486673046  + 1447.67021178560 *T)

    # URANUS X Coord
    sumX1  = 0
    sumX1  +=  2225113.750 * cos(1.80968682072  + 0.24381748350 *T)
    sumX1  +=  738897.387 * cos(6.01007226305  + 149.80701461810 *T)
    sumX1  +=  239408.407 * cos(5.33797172897  + 73.54094334250 *T)
    sumX1  +=  229359.688 * cos(2.48132547665  + 76.50988875911 *T)
    sumX1  +=  110583.560 * cos(5.57527703898  + 11.28951774740 *T)
    sumX1  +=   95657.865 * cos(0.35447716429  + 63.97971578690 *T)
    sumX1  +=   81488.096 * cos(1.21038214498  + 86.07111631471 *T)
    sumX1  +=   45577.241 * cos(2.29785938115  + 138.76131435421 *T)
    sumX1  +=   51354.528 * cos(2.18905169391  + 224.58861318540 *T)
    sumX1  +=   38568.000 * cos(0.30675960989  + 71.09326278771 *T)
    sumX1  +=   37681.305 * cos(0.07581737994  + 75.02541605080 *T)
    sumX1  +=   41920.157 * cos(4.16572993198  + 74.53778108379 *T)
    sumX1  +=   36154.364 * cos(1.23626010875  + 78.95756931390 *T)
    sumX1  +=   21671.035 * cos(4.93673342225  + 151.29148732640 *T)
    sumX1  +=   19425.371 * cos(1.30586488933  + 77.99436146740 *T)
    sumX1  +=   17377.170 * cos(0.24648496590  + 72.05647063421 *T)
    sumX1  +=   15101.686 * cos(5.53695908680  + 4.17597074660 *T)
    sumX1  +=    7734.175 * cos(1.61342909349  + 71.84402231310 *T)
    sumX1  +=    7431.784 * cos(6.20301472189  + 78.20680978850 *T)
    sumX1  +=    6987.545 * cos(2.40718117644  + 145.87486135500 *T)
    sumX1  +=    7298.813 * cos(2.23506175232  + 3.21276290011 *T)
    sumX1  +=    7589.358 * cos(2.76776928200  + 148.32254190981 *T)
    sumX1  +=    4377.580 * cos(3.74279342397  + 160.85271488200 *T)
    sumX1  +=    3947.976 * cos(4.74495586893  + 65.46418849521 *T)
    sumX1  +=    4122.455 * cos(1.48583809368  + 22.33521801130 *T)
    sumX1  +=    2797.759 * cos(4.90690786558  + 213.54291292150 *T)
    sumX1  +=    2617.647 * cos(3.36386956187  + 9.80504503910 *T)
    sumX1  +=    2637.599 * cos(4.54174637500  + 12.77399045571 *T)
    sumX1  +=    3175.281 * cos(4.65202774219  + 299.37021175271 *T)
    sumX1  +=    2245.280 * cos(3.50167279614  + 52.44638055600 *T)
    sumX1  +=    2227.953 * cos(0.18128254666  + 87.55558902300 *T)
    sumX1  +=    2822.936 * cos(3.40176644427  + 84.58664360640 *T)
    sumX1  +=    2517.942 * cos(2.84033057389  + 18.40306474821 *T)
    sumX1  +=    2220.819 * cos(4.78480068974  + 72.57773549600 *T)
    sumX1  +=    1622.656 * cos(3.75786744204  + 153.73916788121 *T)
    sumX1  +=    1925.669 * cos(2.39938534926  + 39.86132582961 *T)
    sumX1  +=    1555.527 * cos(4.13751627934  + 74.06220820430 *T)
    sumX1  +=    1901.819 * cos(0.07719535510  + 127.71561409031 *T)
    sumX1  +=    1598.284 * cos(0.08347749781  + 79.47883417571 *T)
    sumX1  +=    1476.433 * cos(3.67237467967  + 75.98862389731 *T)
    sumX1  +=    1532.412 * cos(3.71778823462  + 152.77596003471 *T)
    sumX1  +=    1440.833 * cos(5.32095275949  + 127.22797912329 *T)
    sumX1  +=    1745.915 * cos(3.24923390980  + 77.47309660560 *T)
    sumX1  +=    1713.773 * cos(3.39380823601  + 223.10414047710 *T)
    sumX1  +=    1706.344 * cos(1.11295016517  + 226.07308589371 *T)
    sumX1  +=    1357.973 * cos(1.46850463244  + 70.57199792590 *T)
    sumX1  +=    1738.426 * cos(5.39623683669  + 146.83806920150 *T)
    sumX1  +=    1467.571 * cos(3.53057221654  + 2.93757625420 *T)
    sumX1  +=    1179.806 * cos(5.13920013032  + 220.65645992230 *T)
    sumX1  +=    1236.409 * cos(0.98112016066  + 4.69723560840 *T)
    sumX1  +=    1067.152 * cos(1.38177304653  + 62.49524307860 *T)
    sumX1  +=     977.123 * cos(1.01847642495  + 74.91354146621 *T)
    sumX1  +=     977.123 * cos(0.51933335746  + 75.13729063540 *T)
    sumX1  +=     937.717 * cos(3.09490503051  + 56.86616878610 *T)
    sumX1  +=    1087.876 * cos(4.99233413497  + 131.64776735340 *T)
    sumX1  +=     890.295 * cos(4.27308702685  + 426.84200835950 *T)
    sumX1  +=    1040.906 * cos(5.41835035760  + 110.18950627200 *T)
    sumX1  +=     935.342 * cos(0.35022350257  + 149.31937965110 *T)
    sumX1  +=     862.611 * cos(4.50957284182  + 52.93401552300 *T)
    sumX1  +=     777.859 * cos(6.15240216459  + 63.49208081989 *T)
    sumX1  +=     746.826 * cos(4.78455196521  + 93.18466331550 *T)
    sumX1  +=     747.709 * cos(5.93971377228  + 206.42936592071 *T)
    sumX1  +=     648.200 * cos(0.20345060980  + 80.44204202221 *T)
    sumX1  +=     640.350 * cos(6.10288292956  + 1.24065522479 *T)
    sumX1  +=     794.750 * cos(3.28971980546  + 7.35736448430 *T)
    sumX1  +=     605.254 * cos(0.10549865294  + 1059.62574767270 *T)
    sumX1  +=     615.326 * cos(4.84518689972  + 10.80188278040 *T)
    sumX1  +=     569.613 * cos(4.62358405921  + 73.05330837550 *T)
    sumX1  +=     665.795 * cos(3.49783402516  + 97.11681657861 *T)
    sumX1  +=     575.875 * cos(2.42879950405  + 522.82123557730 *T)
    sumX1  +=     563.041 * cos(1.91595270042  + 537.04832957890 *T)
    sumX1  +=     493.956 * cos(3.50296492994  + 137.27684164590 *T)
    sumX1  +=     477.372 * cos(1.15766007580  + 41.34579853790 *T)
    sumX1  +=     445.046 * cos(3.65108277717  + 2.69149803831 *T)
    sumX1  +=     441.333 * cos(1.07480656636  + 67.91186905000 *T)
    sumX1  +=     413.028 * cos(3.78904491456  + 66.94866120350 *T)
    sumX1  +=     408.718 * cos(0.33871031527  + 82.13896305161 *T)
    sumX1  +=     393.354 * cos(1.58833529143  + 15.22167101050 *T)
    sumX1  +=     347.208 * cos(4.09740245991  + 146.62562088040 *T)
    sumX1  +=     350.848 * cos(3.06941415154  + 76.02225379210 *T)
    sumX1  +=     311.934 * cos(1.33402230480  + 1.72829019180 *T)
    sumX1  +=     307.527 * cos(4.70379249427  + 202.49721265760 *T)
    sumX1  +=     343.437 * cos(4.42179382418  + 142.69346761730 *T)
    sumX1  +=     291.412 * cos(4.41796460350  + 5.66044345490 *T)
    sumX1  +=     327.071 * cos(5.84215236142  + 113.15845168861 *T)
    sumX1  +=     283.053 * cos(1.12281740456  + 138.27367938720 *T)
    sumX1  +=     276.688 * cos(0.37637007386  + 70.60562782070 *T)
    sumX1  +=     279.650 * cos(4.09523616178  + 83.10217089810 *T)
    sumX1  +=     294.524 * cos(0.50203802366  + 145.35359649321 *T)
    sumX1  +=     254.666 * cos(6.25426492940  + 235.63431344930 *T)
    sumX1  +=     294.078 * cos(1.06082719798  + 33.43580002939 *T)
    sumX1  +=     275.813 * cos(1.12612561421  + 56.37853381910 *T)
    sumX1  +=     255.740 * cos(0.71777505890  + 265.74547599400 *T)
    sumX1  +=     273.184 * cos(1.96036359915  + 36.89238041300 *T)
    sumX1  +=     246.560 * cos(5.32039073182  + 159.36824217371 *T)
    sumX1  +=     235.614 * cos(3.63128451947  + 20.85074530300 *T)
    sumX1  +=     217.300 * cos(4.25375853714  + 152.98840835580 *T)
    sumX1  +=     233.552 * cos(0.39323553377  + 108.70503356371 *T)
    sumX1  +=     209.062 * cos(4.64314195946  + 140.24578706250 *T)
    sumX1  +=     229.061 * cos(2.81847049652  + 194.89603068979 *T)
    sumX1  +=     197.502 * cos(5.13942008569  + 190.96387742670 *T)
    sumX1  +=     194.913 * cos(2.32796285072  + 209.12312469139 *T)
    sumX1  +=     192.549 * cos(0.45945144783  + 23.81969071961 *T)
    sumX1  +=     228.047 * cos(5.73826756793  + 297.88573904440 *T)
    sumX1  +=     167.442 * cos(6.01751680724  + 134.34152612410 *T)
    sumX1  +=     169.219 * cos(3.75680402327  + 59.55992755679 *T)
    sumX1  +=     182.817 * cos(0.00711215290  + 134.82916109110 *T)
    sumX1  +=     154.035 * cos(0.91123613328  + 288.32451148881 *T)
    sumX1  +=     165.598 * cos(6.03173268563  + 6.18170831670 *T)
    sumX1  +=     153.435 * cos(2.70033880330  + 162.33718759030 *T)
    sumX1  +=     191.930 * cos(0.83163790645  + 374.15181032000 *T)
    sumX1  +=     154.136 * cos(3.25486116088  + 14.25846316400 *T)
    sumX1  +=     152.205 * cos(1.12920055542  + 120.11443212250 *T)
    sumX1  +=     179.814 * cos(4.00847058700  + 60.04756252381 *T)
    sumX1  +=     143.557 * cos(5.80170507906  + 212.05844021321 *T)
    sumX1  +=     141.934 * cos(1.31057170313  + 69.60879007940 *T)
    sumX1  +=     140.696 * cos(5.35688361188  + 29.44876501211 *T)
    sumX1  +=     140.277 * cos(5.15925499975  + 80.96330688400 *T)
    sumX1  +=     134.345 * cos(0.43302660999  + 39.60069339870 *T)
    sumX1  +=     148.026 * cos(3.01875910211  + 186.45557754760 *T)
    sumX1  +=     138.037 * cos(4.87862613958  + 3.68833577960 *T)
    sumX1  +=     122.748 * cos(4.43345211218  + 129.20008679860 *T)
    sumX1  +=     132.977 * cos(1.16919209454  + 68.80564092731 *T)
    sumX1  +=     141.987 * cos(4.72628475671  + 114.64292439690 *T)
    sumX1  +=     132.318 * cos(3.93114176785  + 90.00326957780 *T)
    sumX1  +=     116.010 * cos(3.50661004031  + 33.38091827520 *T)
    sumX1  +=     119.775 * cos(2.67306209327  + 69.08752521760 *T)
    sumX1  +=     120.027 * cos(0.98279379047  + 305.10235190919 *T)
    sumX1  +=     117.279 * cos(0.70638047167  + 277.27881122490 *T)
    sumX1  +=     120.086 * cos(2.88322008541  + 50.64639366260 *T)
    sumX1  +=     109.599 * cos(5.23026989145  + 89.04006173130 *T)
    sumX1  +=     121.041 * cos(2.51714662354  + 154.26043274300 *T)
    sumX1  +=     118.779 * cos(0.36869567507  + 81.24519117430 *T)
    sumX1  +=     103.250 * cos(3.71412105335  + 144.39038864671 *T)
    sumX1  +=     118.351 * cos(1.42159512933  + 41.40068029210 *T)
    sumX1  +=     116.406 * cos(3.28187465983  + 116.18227885939 *T)
    sumX1  +=     136.248 * cos(5.73853005968  + 45.33283355520 *T)
    sumX1  +=     104.859 * cos(2.65494895117  + 54.41848823131 *T)
    sumX1  +=      96.225 * cos(4.45558403980  + 50.96190784770 *T)
    sumX1  +=     116.279 * cos(6.16263593711  + 227.55755860200 *T)
    sumX1  +=     122.913 * cos(3.57188437917  + 300.85468446100 *T)
    sumX1  +=      93.159 * cos(1.08143702108  + 110.45013870291 *T)
    sumX1  +=     102.738 * cos(5.34762722745  + 152.25469517290 *T)
    sumX1  +=      86.724 * cos(4.81341834361  + 7.83293736379 *T)
    sumX1  +=      81.743 * cos(6.27044958652  + 228.52076644850 *T)
    sumX1  +=      84.795 * cos(6.21044885379  + 42.83027124621 *T)
    sumX1  +=      75.252 * cos(0.08318536070  + 2.20386307129 *T)
    sumX1  +=      74.096 * cos(0.25759751910  + 148.84380677161 *T)
    sumX1  +=      74.141 * cos(6.12651431950  + 150.77022246460 *T)
    sumX1  +=      80.549 * cos(4.94365526521  + 99.40443843901 *T)
    sumX1  +=      73.326 * cos(3.51729985830  + 640.14110379750 *T)
    sumX1  +=      81.866 * cos(2.33249180877  + 31.95132732110 *T)
    sumX1  +=      73.049 * cos(0.08899713731  + 9.31741007210 *T)
    sumX1  +=      74.334 * cos(4.73265453188  + 8.32057233081 *T)
    sumX1  +=      73.323 * cos(5.35768135167  + 95.63234387030 *T)
    sumX1  +=      69.283 * cos(2.45721425737  + 120.60206708950 *T)
    sumX1  +=      67.771 * cos(0.24277166907  + 16.91859203990 *T)
    sumX1  +=      71.013 * cos(1.85476079726  + 24.13520490470 *T)
    sumX1  +=      75.802 * cos(1.43483468727  + 36.40474544600 *T)
    sumX1  +=      79.127 * cos(4.41984040421  + 5.97595764000 *T)
    sumX1  +=      72.170 * cos(2.40889219421  + 404.26297286470 *T)
    sumX1  +=      67.065 * cos(4.84209324295  + 28.81562556571 *T)
    sumX1  +=      72.371 * cos(5.06657435408  + 230.32075334190 *T)
    sumX1  +=      86.876 * cos(4.81292247832  + 419.72846135871 *T)
    sumX1  +=      62.861 * cos(4.86097820142  + 202.00957769060 *T)
    sumX1  +=      70.935 * cos(2.67374366940  + 183.48663213100 *T)
    sumX1  +=      69.304 * cos(4.51347223027  + 333.41352756050 *T)
    sumX1  +=      60.612 * cos(4.31236105540  + 116.66991382640 *T)
    sumX1  +=      60.708 * cos(5.84738885486  + 70.35954960481 *T)
    sumX1  +=      70.577 * cos(1.23478501009  + 70.08436295889 *T)
    sumX1  +=      60.413 * cos(2.93879448663  + 149.91888920271 *T)
    sumX1  +=      56.872 * cos(0.78979170331  + 53.93085326429 *T)
    sumX1  +=      56.665 * cos(5.11299964404  + 79.69128249680 *T)
    sumX1  +=      56.672 * cos(0.54290142404  + 73.32849502140 *T)
    sumX1  +=      53.385 * cos(0.53928334859  + 329.48137429739 *T)
    sumX1  +=      60.881 * cos(1.12038902225  + 126.23114138200 *T)
    sumX1  +=      55.720 * cos(4.13440106405  + 76.72233708021 *T)
    sumX1  +=      55.712 * cos(0.10603963574  + 157.88376946540 *T)
    sumX1  +=      60.449 * cos(4.04220409451  + 347.64062156210 *T)
    sumX1  +=      51.883 * cos(1.22311412430  + 147.35933406330 *T)
    sumX1  +=      53.514 * cos(4.48532885411  + 433.95555536030 *T)
    sumX1  +=      49.248 * cos(3.09071508357  + 68.43313391180 *T)
    sumX1  +=      48.173 * cos(1.33565042584  + 128.71245183160 *T)
    sumX1  +=      57.429 * cos(1.74516578840  + 19.88753745650 *T)
    sumX1  +=      45.855 * cos(1.83493312947  + 8.10812400971 *T)
    sumX1  +=      45.035 * cos(4.03985804524  + 6.86972951729 *T)
    sumX1  +=      45.638 * cos(1.36951355984  + 415.30867312860 *T)
    sumX1  +=      43.941 * cos(4.22289924597  + 75.77617557621 *T)
    sumX1  +=      43.941 * cos(3.59810035804  + 74.27465652540 *T)
    sumX1  +=      52.767 * cos(1.19129920562  + 147.83490694279 *T)
    sumX1  +=      40.257 * cos(2.70867032624  + 155.22364058950 *T)
    sumX1  +=      42.861 * cos(1.61841225493  + 107.22056085540 *T)
    sumX1  +=      39.791 * cos(2.79302647335  + 258.63192899320 *T)
    sumX1  +=      39.522 * cos(4.43426945929  + 98.91680347199 *T)
    sumX1  +=      39.676 * cos(2.43777338987  + 276.79117625789 *T)
    sumX1  +=      41.680 * cos(1.56627991706  + 272.85902299479 *T)
    sumX1  +=      49.985 * cos(6.15572302327  + 125.74350641500 *T)
    sumX1  +=      39.779 * cos(1.76579064544  + 143.86912378490 *T)
    sumX1  +=      47.193 * cos(4.34084191086  + 67.16110952460 *T)
    sumX1  +=      42.716 * cos(0.45899696552  + 1589.31671276730 *T)
    sumX1  +=      36.533 * cos(4.68392419441  + 81.61769818981 *T)
    sumX1  +=      46.283 * cos(0.57532354052  + 62.00760811160 *T)
    sumX1  +=      36.873 * cos(2.17201422211  + 46.71424039951 *T)
    sumX1  +=      35.335 * cos(4.81939505521  + 135.79236893760 *T)
    sumX1  +=      36.570 * cos(0.97919259316  + 167.96626188280 *T)
    sumX1  +=      33.878 * cos(2.56857021871  + 14.47091148511 *T)
    sumX1  +=      33.444 * cos(3.72411302136  + 149.69514003350 *T)
    sumX1  +=      40.027 * cos(1.96239717900  + 187.94005025590 *T)
    sumX1  +=      33.511 * cos(2.62937481855  + 61.01077037031 *T)
    sumX1  +=      37.402 * cos(0.21188675855  + 259.75270340660 *T)
    sumX1  +=      34.209 * cos(3.38154632289  + 529.93478257810 *T)
    sumX1  +=      31.617 * cos(2.82544568008  + 224.10097821840 *T)
    sumX1  +=      42.780 * cos(1.30383338598  + 58.35064149440 *T)
    sumX1  +=      30.892 * cos(1.73608557912  + 1052.51220067191 *T)
    sumX1  +=      30.724 * cos(5.69827217624  + 103.33659170210 *T)
    sumX1  +=      37.429 * cos(0.79282523594  + 74.13164417351 *T)
    sumX1  +=      37.429 * cos(0.74498454647  + 75.91918792810 *T)
    sumX1  +=      30.708 * cos(0.24881030437  + 72.73779419041 *T)
    sumX1  +=      30.708 * cos(1.28899947804  + 77.31303791120 *T)
    sumX1  +=      35.364 * cos(3.42228139592  + 82.88972257700 *T)
    sumX1  +=      28.175 * cos(0.90754643164  + 295.43805848960 *T)
    sumX1  +=      29.578 * cos(2.34251637992  + 1066.73929467350 *T)
    sumX1  +=      29.606 * cos(3.07274976796  + 121.23520653590 *T)
    sumX1  +=      28.770 * cos(5.93764446818  + 6.34846465550 *T)
    sumX1  +=      32.637 * cos(2.82392002724  + 155.53915477459 *T)
    sumX1  +=      31.706 * cos(1.92242975115  + 38.37685312130 *T)
    sumX1  +=      26.723 * cos(4.75188318045  + 43.53284666180 *T)
    sumX1  +=      27.442 * cos(3.02987836624  + 220.13519506050 *T)
    sumX1  +=      36.094 * cos(5.82060940307  + 340.52707456129 *T)
    sumX1  +=      28.308 * cos(1.87170058153  + 85.58348134769 *T)
    sumX1  +=      30.388 * cos(3.04166414826  + 184.97110483931 *T)
    sumX1  +=      33.909 * cos(6.02618077188  + 171.89841514590 *T)
    sumX1  +=      28.512 * cos(5.60849102784  + 51.44954281471 *T)
    sumX1  +=      28.765 * cos(2.64529774267  + 64.97655352819 *T)
    sumX1  +=      31.960 * cos(3.10259498469  + 221.61966776881 *T)
    sumX1  +=      26.763 * cos(0.36818165170  + 217.47506618460 *T)
    sumX1  +=      24.737 * cos(2.60957039175  + 47.93808067690 *T)
    sumX1  +=      31.097 * cos(2.39335860911  + 98.60128928690 *T)
    sumX1  +=      23.919 * cos(5.91490770857  + 219.17198721400 *T)
    sumX1  +=      29.056 * cos(1.38484676611  + 106.73292588839 *T)
    sumX1  +=      23.488 * cos(0.57419956660  + 28.06735816780 *T)
    sumX1  +=      23.868 * cos(0.00422826975  + 57.49930823250 *T)
    sumX1  +=      26.140 * cos(0.51337177220  + 91.70019060721 *T)
    sumX1  +=      24.837 * cos(2.69236553781  + 25.51661174900 *T)
    sumX1  +=      28.239 * cos(1.50901261149  + 131.16013238639 *T)
    sumX1  +=      25.242 * cos(6.21367764022  + 471.93102443120 *T)
    sumX1  +=      23.194 * cos(3.05542590412  + 553.82616999929 *T)
    sumX1  +=      23.342 * cos(4.42887541532  + 61.69209392650 *T)
    sumX1  +=      23.751 * cos(4.09923036421  + 542.78046973539 *T)
    sumX1  +=      22.911 * cos(1.29984079692  + 234.14984074100 *T)
    sumX1  +=      22.260 * cos(6.14564001783  + 227.77000692311 *T)
    sumX1  +=      21.787 * cos(4.61743304596  + 19.36627259471 *T)
    sumX1  +=      20.991 * cos(3.73437162386  + 316.63568714010 *T)
    sumX1  +=      24.652 * cos(5.52981795171  + 529.44714761109 *T)
    sumX1  +=      23.384 * cos(4.90996648905  + 254.69977573010 *T)
    sumX1  +=      21.352 * cos(4.81173625387  + 7.66618102501 *T)
    sumX1  +=      20.300 * cos(2.08272474890  + 467.99887116810 *T)
    sumX1  +=      21.096 * cos(6.10072201079  + 141.94270809190 *T)
    sumX1  +=      19.957 * cos(4.95453887653  + 209.61075965840 *T)
    sumX1  +=      22.156 * cos(2.51690657915  + 68.59989025060 *T)
    sumX1  +=      27.098 * cos(5.67372992575  + 111.67397898031 *T)
    sumX1  +=      26.932 * cos(6.04273389615  + 35.40790770471 *T)
    sumX1  +=      23.041 * cos(5.74991815262  + 486.15811843280 *T)
    sumX1  +=      23.558 * cos(0.87109649175  + 195.38366565681 *T)
    sumX1  +=      19.352 * cos(2.46209752291  + 17.02165790390 *T)
    sumX1  +=      18.710 * cos(2.07001816127  + 69.91760662011 *T)
    sumX1  +=      18.292 * cos(3.74192034003  + 46.45360796860 *T)
    sumX1  +=      18.185 * cos(1.53550218920  + 92.55152386910 *T)
    sumX1  +=      20.489 * cos(3.55935460916  + 116.12739710521 *T)
    sumX1  +=      18.124 * cos(3.38460274859  + 88.35873817511 *T)
    sumX1  +=      17.311 * cos(3.10967389974  + 106.51798543980 *T)
    sumX1  +=      19.722 * cos(0.37725290787  + 296.40126633610 *T)
    sumX1  +=      17.322 * cos(3.58150916447  + 100.62827871640 *T)
    sumX1  +=      17.131 * cos(4.71716216087  + 99.14380600810 *T)
    sumX1  +=      21.450 * cos(2.84014990759  + 191.45151239371 *T)
    sumX1  +=      16.758 * cos(0.69775418663  + 210.57396750490 *T)
    sumX1  +=      17.082 * cos(5.75097694023  + 80.13322548150 *T)
    sumX1  +=      19.774 * cos(0.48633289327  + 80.75755620730 *T)
    sumX1  +=      18.159 * cos(3.24022378066  + 151.50393564750 *T)
    sumX1  +=      16.166 * cos(4.43388054341  + 397.14942586389 *T)
    sumX1  +=      16.332 * cos(3.65718289491  + 94.66913602380 *T)
    sumX1  +=      17.336 * cos(0.34233495912  + 13.08950464080 *T)
    sumX1  +=      16.300 * cos(5.58189501765  + 23.87457247379 *T)
    sumX1  +=      18.165 * cos(5.60556739075  + 198.07742442750 *T)
    sumX1  +=      16.506 * cos(4.69351492032  + 73.38088464811 *T)
    sumX1  +=      16.506 * cos(3.12748016927  + 76.66994745350 *T)
    sumX1  +=      18.503 * cos(5.93375894816  + 156.02678974161 *T)
    sumX1  +=      17.934 * cos(3.85540024938  + 182.00215942271 *T)
    sumX1  +=      15.839 * cos(3.12530503781  + 13.73719830221 *T)
    sumX1  +=      15.946 * cos(1.30849298209  + 55.90296093960 *T)
    sumX1  +=      16.546 * cos(5.55355094997  + 150.80385235940 *T)
    sumX1  +=      17.819 * cos(0.27595854333  + 221.40721944771 *T)
    sumX1  +=      15.752 * cos(3.89703455623  + 67.60305250931 *T)
    sumX1  +=      21.162 * cos(1.87088890253  + 372.66733761171 *T)
    sumX1  +=      19.517 * cos(1.11006773425  + 479.04457143200 *T)
    sumX1  +=      17.205 * cos(2.71359950788  + 213.05527795450 *T)
    sumX1  +=      15.267 * cos(5.79293754979  + 68.12431737110 *T)
    sumX1  +=      15.236 * cos(3.72448728433  + 30.30009827400 *T)
    sumX1  +=      15.262 * cos(1.79781862562  + 67.42423408300 *T)
    sumX1  +=      14.562 * cos(3.11386923384  + 50.90702609350 *T)
    sumX1  +=      15.149 * cos(5.07527743555  + 135.57992061650 *T)
    sumX1  +=      15.898 * cos(3.22516048145  + 174.18603700630 *T)
    sumX1  +=      15.347 * cos(0.45031815988  + 16.70614371880 *T)
    sumX1  +=      14.302 * cos(2.42986138277  + 310.41591201661 *T)
    sumX1  +=      17.345 * cos(5.49231862434  + 261.23717611490 *T)
    sumX1  +=      17.208 * cos(0.99198084016  + 183.85033042589 *T)
    sumX1  +=      17.305 * cos(1.74803865372  + 77.50672650040 *T)
    sumX1  +=      14.612 * cos(3.56947733384  + 30.46685461280 *T)
    sumX1  +=      13.603 * cos(2.07215319428  + 286.84003878050 *T)
    sumX1  +=      15.382 * cos(3.33099430620  + 411.37651986550 *T)
    sumX1  +=      13.708 * cos(3.51368533450  + 352.06040979221 *T)
    sumX1  +=      13.683 * cos(3.92221596093  + 82.44777959230 *T)
    sumX1  +=      16.002 * cos(1.88175647624  + 133.13224006171 *T)
    sumX1  +=      13.766 * cos(4.97884966758  + 44.31474395451 *T)
    sumX1  +=      13.284 * cos(2.82193356482  + 7.14491616321 *T)
    sumX1  +=      12.988 * cos(0.75322804380  + 173.92540457540 *T)
    sumX1  +=      17.441 * cos(0.75477790662  + 60.79832204921 *T)
    sumX1  +=      16.261 * cos(1.99911970336  + 64.73047531231 *T)
    sumX1  +=      12.734 * cos(2.97731330561  + 75.54668091261 *T)
    sumX1  +=      12.734 * cos(1.70208913039  + 74.50415118900 *T)
    sumX1  +=      12.700 * cos(1.71027593315  + 57.75994066341 *T)
    sumX1  +=      14.302 * cos(0.72178151343  + 45.82046852221 *T)
    sumX1  +=      12.678 * cos(5.21256052974  + 81.92651473050 *T)
    sumX1  +=      16.549 * cos(2.78519168402  + 760.01171843649 *T)
    sumX1  +=      12.720 * cos(2.81709441807  + 270.16526422410 *T)
    sumX1  +=      13.453 * cos(3.75052891994  + 12.28635548869 *T)
    sumX1  +=      12.225 * cos(1.22978831724  + 284.39235822571 *T)
    sumX1  +=      11.971 * cos(1.77290743684  + 32.48714639791 *T)
    sumX1  +=      12.357 * cos(2.56239278409  + 515.70768857651 *T)
    sumX1  +=      12.069 * cos(6.21182865124  + 633.02755679670 *T)
    sumX1  +=      13.087 * cos(1.80415855132  + 109.70187130500 *T)
    sumX1  +=      11.341 * cos(6.09688210298  + 189.47940471839 *T)
    sumX1  +=      15.361 * cos(0.60509201896  + 200.52510498229 *T)
    sumX1  +=      12.491 * cos(1.18576138349  + 33.92343499640 *T)
    sumX1  +=      14.227 * cos(4.47775362086  + 203.49405039889 *T)
    sumX1  +=      11.094 * cos(2.98683187494  + 267.22994870229 *T)
    sumX1  +=      12.080 * cos(2.65683685185  + 75.67980735660 *T)
    sumX1  +=      12.080 * cos(5.16415823774  + 74.37102474501 *T)
    sumX1  +=      13.196 * cos(1.55647786503  + 143.38148881789 *T)
    sumX1  +=      13.530 * cos(5.82185773287  + 85.32035678930 *T)
    sumX1  +=      10.739 * cos(0.24603800620  + 94.14787116200 *T)
    sumX1  +=      10.888 * cos(4.54493098814  + 2.72512793310 *T)
    sumX1  +=      13.830 * cos(5.36492110746  + 31.89644556690 *T)
    sumX1  +=      11.969 * cos(4.16293798397  + 42.78208713641 *T)
    sumX1  +=      11.212 * cos(2.68331450691  + 61.53203523210 *T)
    sumX1  +=      11.583 * cos(5.07324345893  + 124.53422035260 *T)
    sumX1  +=      12.891 * cos(4.94162673731  + 266.23311096100 *T)
    sumX1  +=      10.738 * cos(3.25988549780  + 35.66854013560 *T)
    sumX1  +=      13.803 * cos(4.23791127410  + 26.26737127440 *T)
    sumX1  +=      10.459 * cos(5.12191321154  + 145.14114817210 *T)
    sumX1  +=      10.474 * cos(4.75504662292  + 692.34366686999 *T)
    sumX1  +=      10.123 * cos(5.18571320702  + 237.11878615761 *T)
    sumX1  +=       9.800 * cos(6.07581528627  + 92.29089143820 *T)
    sumX1  +=      10.121 * cos(1.17089552104  + 156.39929675710 *T)
    sumX1  +=       9.888 * cos(5.73883644068  + 42.88515300039 *T)
    sumX1  +=      10.092 * cos(3.96119852956  + 215.02738562980 *T)
    sumX1  +=      10.098 * cos(1.26158239056  + 533.11617631580 *T)
    sumX1  +=      10.667 * cos(3.37369147427  + 91.80325647120 *T)
    sumX1  +=      10.954 * cos(0.56363643422  + 89.25251005240 *T)
    sumX1  +=       9.630 * cos(4.20178830544  + 55.38169607781 *T)
    sumX1  +=       9.631 * cos(3.24507538918  + 130.68455950690 *T)
    sumX1  +=       9.832 * cos(2.83130973684  + 54.57854692570 *T)
    sumX1  +=       9.739 * cos(5.68100585654  + 76.62176334371 *T)
    sumX1  +=       9.739 * cos(5.28158188664  + 73.42906875790 *T)
    sumX1  +=      10.658 * cos(1.14021240327  + 31.24875190549 *T)
    sumX1  +=       9.264 * cos(0.06980096659  + 27.33115285740 *T)
    sumX1  +=      11.861 * cos(1.56797219001  + 264.26100328570 *T)
    sumX1  +=       9.015 * cos(2.20482574279  + 17.91542978119 *T)
    sumX1  +=       9.170 * cos(2.03311846559  + 203.98168536590 *T)
    sumX1  +=       8.989 * cos(1.27026973997  + 69.12115511239 *T)
    sumX1  +=       9.028 * cos(5.58743285195  + 141.45507312490 *T)
    sumX1  +=       9.664 * cos(2.39178937690  + 41.88831525910 *T)
    sumX1  +=       8.940 * cos(5.01256462910  + 17.28229033479 *T)
    sumX1  +=      10.494 * cos(1.17131913276  + 624.67561530350 *T)
    sumX1  +=      10.035 * cos(3.29686320022  + 448.93340888730 *T)
    sumX1  +=      10.076 * cos(0.25849068019  + 685.23011986920 *T)
    sumX1  +=       8.857 * cos(4.96386104548  + 44.48150029329 *T)
    sumX1  +=       8.598 * cos(6.06606880520  + 199.31581891990 *T)
    sumX1  +=       9.805 * cos(6.15923742380  + 130.16329464510 *T)
    sumX1  +=       8.882 * cos(2.75549341652  + 105.24845318010 *T)
    sumX1  +=       8.882 * cos(2.75549341652  + 105.73608814710 *T)
    sumX1  +=       8.493 * cos(1.04668491872  + 104.23036357940 *T)
    sumX1  +=       8.845 * cos(5.13692463782  + 88.51879686950 *T)
    sumX1  +=       8.246 * cos(2.68751890723  + 269.67762925710 *T)
    sumX1  +=       8.157 * cos(1.41280465589  + 163.82166029861 *T)
    sumX1  +=       8.392 * cos(5.83503369760  + 681.29796660610 *T)
    sumX1  +=       9.056 * cos(2.62794754205  + 617.56206830270 *T)
    sumX1  +=       8.249 * cos(3.44796441477  + 454.66554904380 *T)
    sumX1  +=       8.521 * cos(1.30623024759  + 155.74490545130 *T)
    sumX1  +=      10.211 * cos(1.78694722678  + 145.38722638800 *T)
    sumX1  +=       7.859 * cos(4.05422578891  + 119.75073382760 *T)
    sumX1  +=       8.216 * cos(1.79455339352  + 544.16187657970 *T)
    sumX1  +=       8.454 * cos(0.22947250972  + 393.21727260080 *T)
    sumX1  +=       8.005 * cos(0.46907750935  + 141.73025977080 *T)
    sumX1  +=       8.848 * cos(3.95115689507  + 193.41155798150 *T)
    sumX1  +=       7.843 * cos(3.54311700587  + 46.81730626350 *T)
    sumX1  +=       7.556 * cos(3.58717505818  + 66.67347455760 *T)
    sumX1  +=       7.436 * cos(1.91784517551  + 178.11819026941 *T)
    sumX1  +=       9.047 * cos(3.51401146870  + 3.42521122121 *T)
    sumX1  +=       7.519 * cos(0.50643097415  + 341.01470952831 *T)
    sumX1  +=       7.416 * cos(3.89859514324  + 606.51636803880 *T)
    sumX1  +=       7.453 * cos(5.63849048468  + 175.40987728371 *T)
    sumX1  +=       7.475 * cos(4.80425272291  + 58.07545484850 *T)
    sumX1  +=       8.741 * cos(4.78639260740  + 58.98378094080 *T)
    sumX1  +=       8.749 * cos(3.03655779976  + 179.91817716280 *T)
    sumX1  +=       7.892 * cos(3.40393008709  + 207.63865198310 *T)
    sumX1  +=       8.402 * cos(6.21287302099  + 70.29011363560 *T)
    sumX1  +=       8.402 * cos(1.60812206860  + 79.76071846601 *T)
    sumX1  +=       8.251 * cos(5.10906071911  + 58.56308981550 *T)
    sumX1  +=       8.140 * cos(2.58466669254  + 39.91620758380 *T)
    sumX1  +=       7.045 * cos(6.16222014109  + 949.41942645330 *T)
    sumX1  +=       7.070 * cos(1.85983109705  + 610.44852130189 *T)
    sumX1  +=       7.791 * cos(4.85392265583  + 196.38050339810 *T)
    sumX1  +=       6.907 * cos(3.58477670994  + 107.74182571721 *T)
    sumX1  +=       7.730 * cos(0.65447053669  + 111.18634401329 *T)
    sumX1  +=       7.449 * cos(2.87976997493  + 142.38465107660 *T)
    sumX1  +=       6.733 * cos(3.14579027954  + 49.75262178531 *T)
    sumX1  +=       7.683 * cos(5.01616370789  + 201.01273994931 *T)
    sumX1  +=       7.711 * cos(4.34156872682  + 210.60759739969 *T)
    sumX1  +=       6.780 * cos(0.65490353399  + 78.64875277320 *T)
    sumX1  +=       6.780 * cos(0.88290624842  + 71.40207932840 *T)
    sumX1  +=       6.908 * cos(0.77950341628  + 21.84758304429 *T)
    sumX1  +=       6.741 * cos(3.00729837927  + 363.10611005610 *T)
    sumX1  +=       6.686 * cos(5.64257628839  + 109.06873185860 *T)
    sumX1  +=       6.572 * cos(4.85673617912  + 216.72430665921 *T)
    sumX1  +=       6.803 * cos(4.98968550869  + 95.47228517591 *T)
    sumX1  +=       6.560 * cos(5.82455888933  + 117.56368570370 *T)
    sumX1  +=       7.044 * cos(4.28026207681  + 114.69780615110 *T)
    sumX1  +=       6.442 * cos(5.74121457797  + 49.47743513939 *T)
    sumX1  +=       6.451 * cos(2.26015709937  + 118.62995941420 *T)
    sumX1  +=       6.473 * cos(0.00551268255  + 37.88921815429 *T)
    sumX1  +=       6.192 * cos(3.54435368886  + 380.84715832300 *T)
    sumX1  +=       6.192 * cos(5.54162291567  + 528.97157473160 *T)
    sumX1  +=       6.479 * cos(0.93518082353  + 133.34468838281 *T)
    sumX1  +=       6.479 * cos(0.93518082353  + 132.85705341579 *T)
    sumX1  +=       6.351 * cos(0.15072906836  + 45.96597300160 *T)
    sumX1  +=       6.402 * cos(0.01370420311  + 535.66692273460 *T)
    sumX1  +=       7.995 * cos(2.91398983172  + 35.18090516859 *T)
    sumX1  +=       6.129 * cos(4.68773030835  + 136.31363379940 *T)
    sumX1  +=       6.460 * cos(2.19114059601  + 216.23667169220 *T)
    sumX1  +=       6.081 * cos(5.07641530470  + 530.89799042461 *T)
    sumX1  +=       6.081 * cos(4.00956079286  + 378.92074262999 *T)
    sumX1  +=       6.809 * cos(5.72263495940  + 237.43430034270 *T)
    sumX1  +=       7.057 * cos(3.30830104174  + 526.75338884040 *T)
    sumX1  +=       7.100 * cos(4.41290399848  + 71.56883566720 *T)
    sumX1  +=       6.155 * cos(5.05323034597  + 142.90591593840 *T)
    sumX1  +=       6.607 * cos(1.89479128516  + 192.44835013500 *T)
    sumX1  +=       6.635 * cos(3.99115479596  + 34.28713329129 *T)
    sumX1  +=       6.319 * cos(3.57711620849  + 143.58723949460 *T)
    sumX1  +=       6.200 * cos(5.67320056167  + 455.15318401080 *T)
    sumX1  +=       6.495 * cos(3.28083835466  + 24.62283987170 *T)
    sumX1  +=       7.911 * cos(2.50242282775  + 77.71917482150 *T)
    sumX1  +=       5.852 * cos(4.64831077945  + 405.74744557300 *T)
    sumX1  +=       6.837 * cos(6.06157563262  + 375.63628302830 *T)
    sumX1  +=       5.864 * cos(5.60462038691  + 108.16251684250 *T)
    sumX1  +=       7.485 * cos(1.15010416646  + 48.51422729289 *T)
    sumX1  +=       5.956 * cos(4.74593134277  + 28.98238190449 *T)
    sumX1  +=       6.512 * cos(5.04955849957  + 549.89401673620 *T)
    sumX1  +=       6.744 * cos(1.08072004510  + 27.96429230381 *T)
    sumX1  +=       5.515 * cos(2.08339216933  + 232.66536803271 *T)
    sumX1  +=       5.532 * cos(0.31890406514  + 164.78486814511 *T)
    sumX1  +=       6.345 * cos(2.71756249318  + 91.48774228611 *T)
    sumX1  +=       6.490 * cos(1.76744850978  + 180.03005174739 *T)
    sumX1  +=       5.431 * cos(0.43190359607  + 28.55499313480 *T)
    sumX1  +=       5.425 * cos(4.51559573958  + 73.65281792710 *T)
    sumX1  +=       5.425 * cos(0.16380669642  + 76.39801417450 *T)
    sumX1  +=       5.562 * cos(5.40559445337  + 125.42799222991 *T)
    sumX1  +=       5.415 * cos(0.23454091581  + 15.79781762650 *T)
    sumX1  +=       6.477 * cos(3.03489812985  + 91.06705116081 *T)
    sumX1  +=       5.588 * cos(2.33528771267  + 408.19512612779 *T)
    sumX1  +=       5.776 * cos(4.10655662159  + 218.65072235220 *T)
    sumX1  +=       5.350 * cos(5.53830738326  + 30.98811947460 *T)
    sumX1  +=       5.255 * cos(4.37451988017  + 49.42255338521 *T)
    sumX1  +=       5.759 * cos(4.69888388132  + 35.92917256650 *T)
    sumX1  +=       5.412 * cos(0.49313502322  + 22.39009976550 *T)
    sumX1  +=       6.770 * cos(2.47861246297  + 118.15438653471 *T)
    sumX1  +=       5.380 * cos(2.58879305933  + 234.25290660500 *T)
    sumX1  +=       5.209 * cos(6.26374262154  + 210.36151918381 *T)
    sumX1  +=       5.177 * cos(2.42407399027  + 34.86539098351 *T)
    sumX1  +=       5.160 * cos(0.76362512485  + 312.21589891000 *T)
    sumX1  +=       5.191 * cos(2.89795037802  + 853.44019923550 *T)
    sumX1  +=       5.099 * cos(5.96759889760  + 248.96763557360 *T)
    sumX1  +=       5.267 * cos(3.95527262737  + 167.07249000551 *T)
    sumX1  +=       5.553 * cos(4.22057884827  + 756.07956517339 *T)
    sumX1  +=       5.467 * cos(4.39069254467  + 301.17019864609 *T)
    sumX1  +=       4.979 * cos(0.39632287951  + 82.61453593110 *T)
    sumX1  +=       4.890 * cos(4.28174925638  + 114.38229196601 *T)
    sumX1  +=       4.949 * cos(2.32780688360  + 72.09010052900 *T)
    sumX1  +=       6.602 * cos(3.50824637161  + 123.29582586020 *T)
    sumX1  +=       6.602 * cos(3.50824637161  + 123.78346082721 *T)
    sumX1  +=       5.370 * cos(2.46922478726  + 293.95358578131 *T)
    sumX1  +=       5.201 * cos(3.74659188258  + 67.11541754230 *T)
    sumX1  +=       5.872 * cos(0.96808951269  + 406.34695512460 *T)
    sumX1  +=       5.250 * cos(0.24305505446  + 830.86116374070 *T)
    sumX1  +=       4.889 * cos(0.96967547082  + 319.32944591080 *T)
    sumX1  +=       5.168 * cos(2.97658265378  + 148.11009358871 *T)
    sumX1  +=       6.514 * cos(2.35520859919  + 302.33915716930 *T)
    sumX1  +=       5.080 * cos(1.86039634355  + 163.14033674240 *T)
    sumX1  +=       4.695 * cos(1.08205169196  + 105.13657859549 *T)
    sumX1  +=       5.583 * cos(1.97933911711  + 283.90472325870 *T)
    sumX1  +=       4.645 * cos(1.73959984601  + 87.87110320809 *T)
    sumX1  +=       5.916 * cos(0.89454328694  + 117.66675156770 *T)
    sumX1  +=       4.506 * cos(4.68119964918  + 100.29821031630 *T)
    sumX1  +=       4.545 * cos(1.88213954631  + 52.39149880180 *T)
    sumX1  +=       4.927 * cos(1.84093010860  + 78.46993434690 *T)
    sumX1  +=       5.394 * cos(4.06857911814  + 231.80522605020 *T)
    sumX1  +=       5.223 * cos(1.80728120886  + 376.70255673880 *T)
    sumX1  +=       4.438 * cos(3.54909482913  + 748.96601817260 *T)
    sumX1  +=       5.514 * cos(4.87130555529  + 281.21096448800 *T)
    sumX1  +=       5.215 * cos(1.09888503990  + 336.59492129820 *T)
    sumX1  +=       5.879 * cos(3.19764906565  + 402.77850015639 *T)
    sumX1  +=       5.420 * cos(2.66878339450  + 322.36782729660 *T)
    sumX1  +=       4.308 * cos(4.56314648517  + 122.71967924421 *T)
    sumX1  +=       4.264 * cos(3.90016849877  + 257.14745628489 *T)
    sumX1  +=       4.520 * cos(1.63248491800  + 49.00186225990 *T)
    sumX1  +=       4.775 * cos(3.17635591657  + 57.86300652739 *T)
    sumX1  +=       4.774 * cos(5.43704799780  + 827.67977000299 *T)
    sumX1  +=       4.196 * cos(3.75278965356  + 222.14093263061 *T)
    sumX1  +=       4.391 * cos(3.97300698545  + 103.59722413300 *T)
    sumX1  +=       5.698 * cos(0.61058889672  + 43.84836084689 *T)
    sumX1  +=       4.161 * cos(3.08172700119  + 43.79347909271 *T)
    sumX1  +=       4.435 * cos(5.04812695927  + 102.11275142471 *T)
    sumX1  +=       4.326 * cos(3.86651430250  + 25.35904518210 *T)
    sumX1  +=       4.344 * cos(4.06256630733  + 49.89563413720 *T)
    sumX1  +=       5.029 * cos(0.85995505548  + 189.42452296421 *T)
    sumX1  +=       4.148 * cos(2.58868726006  + 271.37455028650 *T)
    sumX1  +=       4.154 * cos(1.32034817135  + 170.41394243761 *T)
    sumX1  +=       5.131 * cos(3.49343840091  + 107.26874496520 *T)
    sumX1  +=       4.194 * cos(5.10678907780  + 383.06534421420 *T)
    sumX1  +=       4.307 * cos(3.42628980232  + 106.03035047280 *T)
    sumX1  +=       4.419 * cos(2.19876844402  + 422.42222012939 *T)
    sumX1  +=       4.369 * cos(5.90195130694  + 596.11836143630 *T)
    sumX1  +=       4.317 * cos(5.38430845148  + 610.34545543790 *T)
    sumX1  +=       3.889 * cos(3.08485433387  + 501.62360692681 *T)
    sumX1  +=       3.870 * cos(1.71101422686  + 196.01680510321 *T)
    sumX1  +=       4.439 * cos(3.94878447089  + 302.13340649259 *T)
    sumX1  +=       5.072 * cos(5.15076461582  + 841.90686400460 *T)
    sumX1  +=       4.570 * cos(5.48306410732  + 25.30416342790 *T)
    sumX1  +=       3.853 * cos(1.39776017808  + 327.99690158910 *T)
    sumX1  +=       4.468 * cos(2.54860791860  + 20.95381116700 *T)
    sumX1  +=       3.811 * cos(4.72636380146  + 121.59890483079 *T)
    sumX1  +=       4.526 * cos(5.54057803341  + 223.20720634109 *T)
    sumX1  +=       4.482 * cos(5.83135585848  + 30.35498002820 *T)
    sumX1  +=       3.991 * cos(1.28311117198  + 227.03629374021 *T)
    sumX1  +=       3.731 * cos(6.12095489341  + 379.88395047650 *T)
    sumX1  +=       3.956 * cos(4.56824830678  + 995.88984936931 *T)
    sumX1  +=       3.718 * cos(3.65666608721  + 323.74923414091 *T)
    sumX1  +=       4.018 * cos(3.19895348695  + 136.78920667889 *T)
    sumX1  +=       4.001 * cos(3.56100095575  + 1887.54933515920 *T)
    sumX1  +=       3.691 * cos(1.41771281486  + 219.27505307800 *T)
    sumX1  +=       4.238 * cos(1.44245928396  + 487.64259114110 *T)
    sumX1  +=       4.319 * cos(0.96541270679  + 470.44655172289 *T)
    sumX1  +=       3.696 * cos(4.06633350945  + 95.73540973430 *T)
    sumX1  +=       3.585 * cos(3.51009891487  + 647.25465079831 *T)
    sumX1  +=       3.615 * cos(0.20591103864  + 38.85242600079 *T)
    sumX1  +=       4.508 * cos(2.90546350140  + 763.19311217420 *T)
    sumX1  +=       4.774 * cos(2.46832397929  + 205.94173095370 *T)
    sumX1  +=       3.585 * cos(1.34007514641  + 819.81546347680 *T)
    sumX1  +=       3.631 * cos(2.10367195809  + 310.73142620169 *T)
    sumX1  +=       3.713 * cos(4.67112895523  + 906.60597015449 *T)
    sumX1  +=       4.069 * cos(5.59978544475  + 331.92905485220 *T)
    sumX1  +=       3.909 * cos(2.27593546539  + 204.94489321240 *T)
    sumX1  +=       3.826 * cos(6.01055375896  + 253.21530302180 *T)
    sumX1  +=       4.459 * cos(0.93566957933  + 49.26498681829 *T)
    sumX1  +=       3.398 * cos(2.96372890711  + 548.09402984280 *T)
    sumX1  +=       4.440 * cos(1.94179155953  + 54.89406111079 *T)
    sumX1  +=       3.614 * cos(0.30233234299  + 511.28790034640 *T)
    sumX1  +=       3.347 * cos(1.29886958053  + 672.38443271190 *T)
    sumX1  +=       3.809 * cos(4.51178738730  + 328.59641114070 *T)
    sumX1  +=       3.555 * cos(3.61696917256  + 1368.90407032850 *T)
    sumX1  +=       4.016 * cos(1.13955629148  + 442.99551805410 *T)
    sumX1  +=       3.311 * cos(3.53242079712  + 185.23173727020 *T)
    sumX1  +=       3.338 * cos(2.31470330186  + 90.21571789890 *T)
    sumX1  +=       3.346 * cos(5.43284733631  + 10022.08109758290 *T)
    sumX1  +=       3.379 * cos(2.85060290289  + 1051.02772796360 *T)
    sumX1  +=       3.359 * cos(5.72116664423  + 40.82453367610 *T)
    sumX1  +=       3.472 * cos(2.97502666152  + 141.89701610960 *T)
    sumX1  +=       3.470 * cos(4.38415952735  + 1068.22376738181 *T)
    sumX1  +=       3.246 * cos(4.22017131096  + 30.93323772040 *T)
    sumX1  +=       3.583 * cos(1.73082339999  + 297.98880490840 *T)
    sumX1  +=       4.055 * cos(0.36258365907  + 228.83628063359 *T)
    sumX1  +=       3.241 * cos(0.13111768083  + 34.44469985821 *T)
    sumX1  +=       3.616 * cos(1.14515654777  + 736.12033101530 *T)
    sumX1  +=       3.560 * cos(0.27670779076  + 334.89800026879 *T)
    sumX1  +=       4.091 * cos(1.17441098516  + 913.50706883420 *T)
    sumX1  +=       3.191 * cos(1.16792482000  + 374.46732450509 *T)
    sumX1  +=       3.544 * cos(0.49055376796  + 484.67364572450 *T)
    sumX1  +=       3.559 * cos(2.60397878803  + 781.83999440590 *T)
    sumX1  +=       3.813 * cos(6.05894981328  + 342.01154726960 *T)
    sumX1  +=       3.256 * cos(5.96135741097  + 97.17169833279 *T)
    sumX1  +=       3.256 * cos(5.96135741097  + 97.65933329981 *T)
    sumX1  +=       3.853 * cos(0.47025651359  + 122.08653979781 *T)
    sumX1  +=       4.122 * cos(2.34292957523  + 339.04260185300 *T)
    sumX1  +=       3.135 * cos(1.41353750131  + 674.18441960530 *T)
    sumX1  +=       3.359 * cos(0.15532049803  + 369.80145805910 *T)
    sumX1  +=       4.038 * cos(2.34720338930  + 146.35043423450 *T)
    sumX1  +=       4.010 * cos(6.22343326461  + 60.52313540329 *T)
    sumX1  +=       3.941 * cos(1.29255732924  + 729.00678401450 *T)
    sumX1  +=       3.097 * cos(5.41601825571  + 115.18544111810 *T)
    sumX1  +=       3.291 * cos(1.11683648084  + 2424.35384725460 *T)
    sumX1  +=       3.582 * cos(6.04348598778  + 349.12509427039 *T)
    sumX1  +=       3.393 * cos(1.72165904584  + 290.87525790760 *T)
    sumX1  +=       3.127 * cos(2.23294920715  + 582.13734565059 *T)
    sumX1  +=       3.177 * cos(0.19583207281  + 194.53233239490 *T)
    sumX1  +=       3.592 * cos(2.40358939188  + 1512.56300652469 *T)
    sumX1  +=       3.448 * cos(2.88793959605  + 630.57987624190 *T)
    sumX1  +=       3.156 * cos(5.03361932450  + 555.31064270760 *T)
    sumX1  +=       3.587 * cos(0.58158301767  + 333.04982926560 *T)
    sumX1  +=       3.671 * cos(1.45655426897  + 480.52904414029 *T)
    sumX1  +=       3.505 * cos(1.64838416510  + 52.13086637091 *T)
    sumX1  +=       3.028 * cos(6.26872130718  + 260.11640170150 *T)
    sumX1  +=       3.021 * cos(5.57108092759  + 745.03386490949 *T)
    sumX1  +=       3.015 * cos(3.89416639328  + 1053.99667338020 *T)
    sumX1  +=       3.255 * cos(3.12462674297  + 114.12165953510 *T)
    sumX1  +=       3.582 * cos(5.88445674452  + 292.94468595250 *T)
    sumX1  +=       4.061 * cos(3.95340643388  + 758.52724572820 *T)
    sumX1  +=       3.736 * cos(0.36613852781  + 5983.70563823870 *T)
    sumX1  +=       3.303 * cos(1.28155244731  + 532.38246313290 *T)
    sumX1  +=       4.088 * cos(5.55402436617  + 1265.06053658450 *T)
    sumX1  +=       2.960 * cos(1.72706602311  + 15.74293587230 *T)
    sumX1  +=       3.705 * cos(1.47926124922  + 604.71638114540 *T)
    sumX1  +=       3.156 * cos(5.92521270548  + 657.40657918490 *T)
    sumX1  +=       3.133 * cos(5.04870054324  + 346.15614885380 *T)
    sumX1  +=       3.653 * cos(0.63048648313  + 601.74743572880 *T)
    sumX1  +=       3.119 * cos(1.91614513517  + 969.37866061139 *T)
    sumX1  +=       2.876 * cos(2.35938981628  + 61.04440026510 *T)
    sumX1  +=       2.838 * cos(2.74193803313  + 595.15515358980 *T)
    sumX1  +=       3.359 * cos(4.95035468643  + 229.04203131030 *T)
    sumX1  +=       3.882 * cos(1.38177660207  + 275.79433851660 *T)
    sumX1  +=       3.259 * cos(1.01400972497  + 184.48346987229 *T)
    sumX1  +=       3.003 * cos(1.83095192956  + 531.73476947149 *T)
    sumX1  +=       2.904 * cos(4.01411287559  + 90.52453443961 *T)
    sumX1  +=       3.653 * cos(2.96914294937  + 34.92027273770 *T)
    sumX1  +=       3.037 * cos(5.20432039185  + 432.57414851600 *T)
    sumX1  +=       2.813 * cos(4.72918594081  + 106.25735300890 *T)
    sumX1  +=       3.621 * cos(3.22439526217  + 156.92056161890 *T)
    sumX1  +=       3.602 * cos(3.15871536713  + 162.65270177540 *T)
    sumX1  +=       3.499 * cos(2.67172150269  + 296.50433220009 *T)
    sumX1  +=       3.773 * cos(4.08436936009  + 357.79254994869 *T)
    sumX1  +=       3.503 * cos(1.78954865420  + 9913.91533445829 *T)
    sumX1  +=       3.442 * cos(3.60672329365  + 689.16227313230 *T)
    sumX1  +=       2.741 * cos(1.22562559151  + 279.97256999560 *T)
    sumX1  +=       2.811 * cos(1.78185455227  + 526.00262931501 *T)
    sumX1  +=       2.811 * cos(4.92344720586  + 525.51499434800 *T)
    sumX1  +=       2.855 * cos(5.71771141107  + 49.16192095430 *T)
    sumX1  +=       3.389 * cos(4.67259709450  + 1471.24608502170 *T)
    sumX1  +=       2.713 * cos(5.56428665950  + 395.66495315560 *T)
    sumX1  +=       2.786 * cos(1.34350771854  + 121.49583896680 *T)
    sumX1  +=       2.898 * cos(6.21681698735  + 101.04896984171 *T)
    sumX1  +=       2.865 * cos(6.01164299992  + 525.48136445320 *T)
    sumX1  +=       3.424 * cos(5.73901886785  + 1065.35788782919 *T)
    sumX1  +=       2.851 * cos(0.54853216378  + 344.94686279140 *T)
    sumX1  +=       3.251 * cos(1.45328025834  + 4.20960064140 *T)
    sumX1  +=       2.737 * cos(3.42458700601  + 1065.25482196520 *T)
    sumX1  +=       2.950 * cos(4.98027465475  + 256.08118257441 *T)
    sumX1  +=       2.631 * cos(0.78267027234  + 833.30884429549 *T)
    sumX1  +=       3.298 * cos(5.06593115911  + 469.48334387639 *T)
    sumX1  +=       3.109 * cos(6.28117696143  + 626.16008801179 *T)
    sumX1  +=       3.056 * cos(3.68463096730  + 222.61650551010 *T)
    sumX1  +=       2.887 * cos(2.43495358759  + 241.36645360580 *T)
    sumX1  +=       2.880 * cos(3.68589687520  + 370.21965705690 *T)
    sumX1  +=       2.756 * cos(3.92430653721  + 103.76398047180 *T)
    sumX1  +=       2.886 * cos(2.58668498001  + 668.45227944880 *T)
    sumX1  +=       3.592 * cos(0.47803095441  + 1062.59469308931 *T)
    sumX1  +=       3.123 * cos(5.91118856963  + 384.02855206069 *T)
    sumX1  +=       2.750 * cos(1.59375992485  + 628.60776856660 *T)
    sumX1  +=       2.877 * cos(2.89340756487  + 6058.48723680599 *T)
    sumX1  +=       2.805 * cos(2.02053605529  + 2111.89413086110 *T)
    sumX1  +=       2.563 * cos(3.56308780594  + 440.43845504219 *T)
    sumX1  +=       2.563 * cos(0.42149515235  + 440.92609000920 *T)
    sumX1  +=       2.696 * cos(2.96689095835  + 187.78248368899 *T)
    sumX1  +=       2.619 * cos(4.26306664613  + 218.16308738520 *T)
    sumX1  +=       3.250 * cos(2.54781834108  + 608.96404859360 *T)
    sumX1  +=       2.535 * cos(4.28939637077  + 546.71262299849 *T)
    sumX1  +=       2.863 * cos(2.93725449249  + 637.69342324271 *T)
    sumX1  +=       2.510 * cos(2.72205634566  + 1258.69774910910 *T)
    sumX1  +=       2.502 * cos(0.12173637336  + 766.12842769600 *T)
    sumX1  +=       2.501 * cos(3.95752847987  + 108.21739859670 *T)
    sumX1  +=       2.522 * cos(4.54787198105  + 115.60613224340 *T)
    sumX1  +=       2.695 * cos(1.46434656788  + 592.18620817320 *T)
    sumX1  +=       2.760 * cos(1.05454060387  + 631.78916230430 *T)
    sumX1  +=       2.903 * cos(2.53838707555  + 620.49738382450 *T)
    sumX1  +=       2.884 * cos(4.72874609806  + 5.17280848790 *T)
    sumX1  +=       2.890 * cos(3.31519233579  + 361.62163734781 *T)
    sumX1  +=       2.464 * cos(3.09248623775  + 416.79314583689 *T)
    sumX1  +=       3.268 * cos(2.53292811337  + 39.37369086260 *T)
    sumX1  +=       2.473 * cos(0.08779249635  + 368.83825021259 *T)
    sumX1  +=       3.222 * cos(0.65044985311  + 421.21293406700 *T)
    sumX1  +=       2.626 * cos(1.10132862735  + 135.82599883240 *T)
    sumX1  +=       2.497 * cos(2.08712349497  + 109.22629842550 *T)
    sumX1  +=       2.634 * cos(0.11373538142  + 159.47130803769 *T)
    sumX1  +=       3.221 * cos(0.89035575469  + 124.25903370670 *T)
    sumX1  +=       2.517 * cos(4.16137803488  + 389.96644289390 *T)
    sumX1  +=       2.921 * cos(3.40506832572  + 109.96250373590 *T)
    sumX1  +=       3.208 * cos(4.58577380818  + 436.50630177909 *T)
    sumX1  +=       2.416 * cos(3.47591283723  + 531.41925528640 *T)
    sumX1  +=       3.214 * cos(2.54161801429  + 256.18424843839 *T)
    sumX1  +=       2.505 * cos(3.99534621400  + 460.88532416730 *T)
    sumX1  +=       2.681 * cos(4.31478466061  + 9988.69693302559 *T)
    sumX1  +=       2.782 * cos(0.78003425566  + 947.28726008359 *T)
    sumX1  +=       2.409 * cos(5.18542514143  + 1228.25440708809 *T)
    sumX1  +=       2.353 * cos(5.50653116579  + 75.23786437190 *T)
    sumX1  +=       2.353 * cos(5.45605657739  + 74.81296772971 *T)
    sumX1  +=       2.440 * cos(5.74974436995  + 894.59706204410 *T)
    sumX1  +=       2.656 * cos(3.45452000558  + 623.46632924110 *T)
    sumX1  +=       2.337 * cos(2.77425938866  + 533.86693584120 *T)
    sumX1  +=       2.402 * cos(1.96391904397  + 843.87897167990 *T)
    sumX1  +=       3.239 * cos(5.50110511166  + 956.53297345411 *T)
    sumX1  +=       3.017 * cos(1.71602977265  + 2043.73844432759 *T)
    sumX1  +=       3.017 * cos(4.85762242624  + 2044.22607929460 *T)
    sumX1  +=       2.405 * cos(0.69226746482  + 59.03866269500 *T)
    sumX1  +=       2.405 * cos(0.69226746482  + 59.52629766200 *T)
    sumX1  +=       2.976 * cos(4.98106512921  + 80.65449034331 *T)
    sumX1  +=       2.378 * cos(2.81529183944  + 223.62540533890 *T)
    sumX1  +=       2.317 * cos(1.17777954553  + 247.22253043440 *T)
    sumX1  +=       2.571 * cos(1.89573446400  + 595.47066777489 *T)
    sumX1  +=       2.545 * cos(0.42139208830  + 450.21213091890 *T)
    sumX1  +=       2.346 * cos(1.76423659089  + 60.31068708220 *T)
    sumX1  +=       2.905 * cos(5.55552157075  + 344.45922782439 *T)
    sumX1  +=       2.979 * cos(1.27048702580  + 370.32272292090 *T)
    sumX1  +=       2.400 * cos(1.68813814874  + 14.73403604350 *T)
    sumX1  +=       2.717 * cos(1.31138458381  + 479.64408098360 *T)
    sumX1  +=       2.611 * cos(0.99388014320  + 3116.02381351440 *T)
    sumX1  +=       2.621 * cos(0.25841040278  + 78.99119920869 *T)
    sumX1  +=       2.223 * cos(1.90165447185  + 53.82778740031 *T)
    sumX1  +=       2.402 * cos(5.03538910572  + 298.88257678570 *T)
    sumX1  +=       2.990 * cos(3.95961598379  + 1812.28010162490 *T)
    sumX1  +=       2.451 * cos(1.91418865833  + 230.80838830890 *T)
    sumX1  +=       2.234 * cos(2.22641223106  + 6144.31453563720 *T)
    sumX1  +=       2.327 * cos(1.84267006744  + 756.56720014041 *T)
    sumX1  +=       2.505 * cos(5.12227466803  + 98.65617104110 *T)
    sumX1  +=       2.409 * cos(1.84825229471  + 248.48000060659 *T)
    sumX1  +=       2.564 * cos(0.59724352209  + 554.31380496631 *T)
    sumX1  +=       2.094 * cos(5.43602058182  + 451.22103074770 *T)
    sumX1  +=       2.094 * cos(5.43602058182  + 450.73339578069 *T)
    sumX1  +=       2.295 * cos(2.50743217001  + 767.12526543730 *T)
    sumX1  +=       2.295 * cos(2.50743217001  + 767.61290040431 *T)
    sumX1  +=       2.066 * cos(1.46552725443  + 518.40144734719 *T)
    sumX1  +=       2.066 * cos(1.46552725442  + 518.88908231420 *T)
    sumX1  +=       2.137 * cos(1.06481249549  + 192.93598510200 *T)
    sumX1  +=       2.874 * cos(1.42201513450  + 1336.17310644710 *T)
    sumX1  +=       2.325 * cos(5.15816119219  + 217.16624964391 *T)
    sumX1  +=       2.325 * cos(2.01656853860  + 216.67861467689 *T)
    sumX1  +=       2.567 * cos(2.28998100536  + 980.42436087530 *T)
    sumX1  +=       2.617 * cos(5.90003632196  + 130.19692453990 *T)
    sumX1  +=       2.128 * cos(5.11811294529  + 180.51768671441 *T)
    sumX1  +=       2.474 * cos(2.38064815315  + 124.74666867370 *T)
    sumX1  +=       2.195 * cos(5.26620078093  + 449.52410971830 *T)
    sumX1  +=       2.195 * cos(5.26620078093  + 449.03647475130 *T)
    sumX1  +=       2.135 * cos(5.71566208623  + 112.67081672160 *T)
    sumX1  +=       1.942 * cos(5.33502835135  + 196.86813836510 *T)
    sumX1  +=       1.912 * cos(3.62510337339  + 119.11759438120 *T)
    sumX1  +=       2.139 * cos(6.13537221304  + 179.03321400610 *T)
    sumX1  +=       2.139 * cos(2.99377955945  + 178.54557903910 *T)
    sumX1  +=       1.950 * cos(5.27659816322  + 39.11305843170 *T)
    sumX1  +=       1.854 * cos(0.40784304859  + 355.24180352990 *T)
    sumX1  +=       1.854 * cos(3.54943570218  + 354.75416856289 *T)
    sumX1  +=       1.854 * cos(3.15413689271  + 198.82818395289 *T)
    sumX1  +=       1.965 * cos(2.90831551828  + 84.09900863940 *T)
    sumX1  +=       2.359 * cos(6.02456565449  + 31.73638687251 *T)
    sumX1  +=       1.903 * cos(4.02163891252  + 153.25153291419 *T)
    sumX1  +=       2.125 * cos(4.46822209107  + 490.57790666290 *T)
    sumX1  +=       2.125 * cos(1.32662943748  + 490.09027169589 *T)
    sumX1  +=       1.914 * cos(2.80943361629  + 331.45348197271 *T)
    sumX1  +=       1.914 * cos(2.80943361629  + 330.96584700570 *T)
    sumX1  +=       2.417 * cos(3.62587768999  + 139.75815209550 *T)
    sumX1  +=       2.146 * cos(4.80060249860  + 463.26356875290 *T)
    sumX1  +=       2.146 * cos(4.80060249860  + 463.75120371991 *T)
    sumX1  +=       1.907 * cos(0.29413995146  + 458.12212942741 *T)
    sumX1  +=       1.907 * cos(0.29413995146  + 457.63449446040 *T)
    sumX1  +=       1.967 * cos(5.78363534283  + 180.81194904009 *T)
    sumX1  +=       1.967 * cos(2.64204268924  + 181.29958400710 *T)
    sumX1  +=       2.145 * cos(0.14136422951  + 66.42739634171 *T)
    sumX1  +=       2.145 * cos(3.28295688310  + 65.93976137470 *T)
    sumX1  +=       1.797 * cos(2.39944522660  + 757.56403788170 *T)
    sumX1  +=       1.797 * cos(2.39944522659  + 758.05167284870 *T)
    sumX1  +=       1.750 * cos(4.28600192557  + 144.86596152620 *T)
    sumX1  +=       2.036 * cos(2.34392479937  + 304.10551416790 *T)
    sumX1  +=       2.036 * cos(2.34392479937  + 303.61787920090 *T)
    sumX1  +=       1.797 * cos(0.32134026487  + 704.87383984220 *T)
    sumX1  +=       1.797 * cos(0.32134026487  + 705.36147480921 *T)
    sumX1  +=       1.894 * cos(0.87966092565  + 910.06255053810 *T)
    sumX1  +=       1.894 * cos(0.87966092565  + 909.57491557110 *T)
    sumX1  +=       1.722 * cos(5.49208647911  + 67.63668240410 *T)
    sumX1  +=       1.837 * cos(6.09149650375  + 544.26494244370 *T)
    sumX1  +=       1.837 * cos(2.94990385016  + 544.75257741070 *T)
    sumX1  +=       1.726 * cos(6.28245115060  + 1507.42156719921 *T)
    sumX1  +=       1.726 * cos(3.14085849702  + 1506.93393223219 *T)
    sumX1  +=       1.730 * cos(6.03213855973  + 565.35950523021 *T)
    sumX1  +=       1.730 * cos(2.89054590614  + 564.87187026320 *T)
    sumX1  +=       1.877 * cos(0.39026248475  + 582.62498061761 *T)
    sumX1  +=       1.795 * cos(0.24875732404  + 681.78560157310 *T)
    sumX1  +=       1.921 * cos(5.53736958606  + 699.45721387080 *T)
    sumX1  +=       1.983 * cos(2.54851132238  + 256.78375799000 *T)
    sumX1  +=       1.983 * cos(5.69010397597  + 256.29612302300 *T)
    sumX1  +=       1.763 * cos(0.62072025706  + 1521.64866120080 *T)
    sumX1  +=       1.763 * cos(3.76231291065  + 1521.16102623379 *T)
    sumX1  +=       2.238 * cos(3.66188523153  + 699.94484883781 *T)
    sumX1  +=       1.965 * cos(0.50815651093  + 1317.26309965699 *T)
    sumX1  +=       1.965 * cos(0.50815651092  + 1317.75073462401 *T)
    sumX1  +=       1.641 * cos(2.64526934415  + 76.98546163860 *T)
    sumX1  +=       1.645 * cos(3.22138588883  + 237.92193530970 *T)
    sumX1  +=       1.880 * cos(1.89024853051  + 71.35638734610 *T)
    sumX1  +=       2.123 * cos(3.24947752102  + 528.45030986980 *T)
    sumX1  +=       2.123 * cos(0.10788486743  + 527.96267490280 *T)
    sumX1  +=       2.071 * cos(0.60877130052  + 688.41151360689 *T)
    sumX1  +=       2.071 * cos(3.75036395411  + 688.89914857391 *T)
    sumX1  +=       2.066 * cos(5.47736990001  + 182.99899716400 *T)
    sumX1  +=       2.080 * cos(4.96628697356  + 415.79630809560 *T)
    sumX1  +=       1.610 * cos(5.00948478905  + 173.69840203930 *T)
    sumX1  +=       1.624 * cos(2.10483095182  + 683.74564716089 *T)
    sumX1  +=       1.624 * cos(2.10483095182  + 684.23328212791 *T)
    sumX1  +=       1.582 * cos(0.64155462640  + 349.09146437560 *T)
    sumX1  +=       1.582 * cos(0.64155462640  + 348.60382940860 *T)
    sumX1  +=       1.525 * cos(3.24207905299  + 555.79827767460 *T)
    sumX1  +=       1.823 * cos(5.40830199356  + 290.46337550291 *T)
    sumX1  +=       1.823 * cos(5.40830199356  + 289.97574053589 *T)
    sumX1  +=       1.524 * cos(1.14769098014  + 1303.03600565540 *T)
    sumX1  +=       1.524 * cos(1.14769098014  + 1303.52364062241 *T)
    sumX1  +=       1.902 * cos(5.79795457847  + 351.57277482520 *T)
    sumX1  +=       1.683 * cos(3.65104607633  + 477.56009872370 *T)
    sumX1  +=       1.683 * cos(0.50945342274  + 478.04773369070 *T)
    sumX1  +=       1.652 * cos(5.08038518118  + 473.90313210650 *T)
    sumX1  +=       1.652 * cos(1.93879252759  + 473.41549713950 *T)
    sumX1  +=       1.475 * cos(2.15788763961  + 406.71065341949 *T)
    sumX1  +=       1.475 * cos(5.29948029320  + 407.19828838650 *T)
    sumX1  +=       1.473 * cos(4.76013578848  + 376.43943218040 *T)
    sumX1  +=       1.473 * cos(1.61854313489  + 375.95179721340 *T)
    sumX1  +=       1.646 * cos(3.41010170818  + 330.08088384900 *T)
    sumX1  +=       1.646 * cos(0.26850905460  + 329.59324888200 *T)
    sumX1  +=       2.036 * cos(3.51391687030  + 402.46298597130 *T)
    sumX1  +=       2.036 * cos(3.51391687030  + 401.97535100430 *T)
    sumX1  +=       1.806 * cos(1.95017105531  + 105.62421356250 *T)
    sumX1  +=       1.436 * cos(4.41744820869  + 384.30373870660 *T)
    sumX1  +=       1.436 * cos(4.41744820869  + 383.81610373959 *T)
    sumX1  +=       1.436 * cos(4.16082568297  + 192.56022471960 *T)
    sumX1  +=       1.436 * cos(4.16082568297  + 193.04785968660 *T)
    sumX1  +=       1.579 * cos(3.84792179756  + 74.64965566840 *T)
    sumX1  +=       1.983 * cos(5.87547482444  + 28.32799059869 *T)
    sumX1  +=       1.517 * cos(3.89748087769  + 696.76345510010 *T)
    sumX1  +=       1.517 * cos(3.89748087769  + 696.27582013310 *T)
    sumX1  +=       1.512 * cos(5.37436432179  + 482.97672469510 *T)
    sumX1  +=       1.512 * cos(2.23277166820  + 483.46435966210 *T)
    sumX1  +=       1.507 * cos(1.08636660761  + 420.93774742110 *T)
    sumX1  +=       1.507 * cos(4.22795926120  + 421.42538238810 *T)
    sumX1  +=       1.461 * cos(3.02594666056  + 6208.53806890760 *T)
    sumX1  +=       1.461 * cos(6.16753931415  + 6208.05043394060 *T)
    sumX1  +=       1.593 * cos(4.09159297487  + 1441.23798834100 *T)
    sumX1  +=       1.593 * cos(4.09159297487  + 1440.75035337400 *T)
    sumX1  +=       1.473 * cos(5.46192054865  + 818.33099076849 *T)
    sumX1  +=       1.473 * cos(5.46192054865  + 818.81862573550 *T)
    sumX1  +=       1.672 * cos(2.10061200490  + 1598.71788162851 *T)
    sumX1  +=       1.672 * cos(2.10061200490  + 1598.23024666149 *T)
    sumX1  +=       1.531 * cos(5.12947393362  + 1737.98613802460 *T)
    sumX1  +=       1.531 * cos(1.98788128003  + 1737.49850305759 *T)
    sumX1  +=       1.912 * cos(5.97684430410  + 486.64575339981 *T)

    # URANUS X Coord
    sumX2  = 0
    sumX2  +=  1016618.950 * cos(0.77056492682  + 75.02541605080 *T)
    sumX2  +=   38040.244 * cos(1.27157972349  + 0.24381748350 *T)
    sumX2  +=   34677.856 * cos(3.16763307609  + 149.80701461810 *T)
    sumX2  +=   12364.786 * cos(5.85713505816  + 76.50988875911 *T)
    sumX2  +=    7910.801 * cos(3.80553695138  + 11.28951774740 *T)
    sumX2  +=    7671.667 * cos(2.54030927271  + 63.97971578690 *T)
    sumX2  +=    5406.463 * cos(4.76691460989  + 71.09326278771 *T)
    sumX2  +=    7007.650 * cos(5.30773139768  + 86.07111631471 *T)
    sumX2  +=    4975.036 * cos(3.03470499387  + 78.95756931390 *T)
    sumX2  +=    3201.017 * cos(6.01127265738  + 74.53778108379 *T)
    sumX2  +=    3151.344 * cos(4.54344132198  + 73.54094334250 *T)
    sumX2  +=    2824.019 * cos(4.03409221161  + 138.76131435421 *T)
    sumX2  +=    3389.631 * cos(0.43008591381  + 1.72829019180 *T)
    sumX2  +=    2518.149 * cos(3.35779148690  + 71.84402231310 *T)
    sumX2  +=    2418.298 * cos(4.45586602451  + 78.20680978850 *T)
    sumX2  +=    2201.666 * cos(0.86216177570  + 4.17597074660 *T)
    sumX2  +=    1870.514 * cos(5.78094816572  + 224.58861318540 *T)
    sumX2  +=    1040.809 * cos(0.72139231899  + 145.87486135500 *T)
    sumX2  +=     825.679 * cos(4.57404708700  + 148.32254190981 *T)
    sumX2  +=     856.424 * cos(2.19300534976  + 151.29148732640 *T)
    sumX2  +=     661.761 * cos(1.01323912739  + 18.40306474821 *T)
    sumX2  +=     687.998 * cos(4.93411449420  + 77.99436146740 *T)
    sumX2  +=     504.683 * cos(5.29540359205  + 22.33521801130 *T)
    sumX2  +=     477.418 * cos(3.17731628846  + 72.57773549600 *T)
    sumX2  +=     476.558 * cos(1.85582546952  + 2.93757625420 *T)
    sumX2  +=     417.277 * cos(1.55648119928  + 160.85271488200 *T)
    sumX2  +=     385.934 * cos(4.71654989644  + 77.47309660560 *T)
    sumX2  +=     325.076 * cos(0.00844871452  + 65.46418849521 *T)
    sumX2  +=     284.902 * cos(0.53316205138  + 131.64776735340 *T)
    sumX2  +=     257.868 * cos(4.97326431366  + 56.86616878610 *T)
    sumX2  +=     334.260 * cos(5.52076636592  + 3.21276290011 *T)
    sumX2  +=     227.240 * cos(2.58050408018  + 127.71561409031 *T)
    sumX2  +=     308.306 * cos(1.74334590495  + 84.58664360640 *T)
    sumX2  +=     220.448 * cos(3.46259685462  + 220.65645992230 *T)
    sumX2  +=     207.595 * cos(2.68871981891  + 12.77399045571 *T)
    sumX2  +=     206.410 * cos(2.89659870733  + 93.18466331550 *T)
    sumX2  +=     207.522 * cos(1.83273630503  + 213.54291292150 *T)
    sumX2  +=     209.061 * cos(5.56996462509  + 153.73916788121 *T)
    sumX2  +=     215.886 * cos(1.45032322858  + 72.05647063421 *T)
    sumX2  +=     212.084 * cos(4.24531706023  + 87.55558902300 *T)
    sumX2  +=     208.906 * cos(5.76818344719  + 52.44638055600 *T)
    sumX2  +=     151.875 * cos(1.37030417137  + 206.42936592071 *T)
    sumX2  +=     148.298 * cos(1.59631432822  + 9.80504503910 *T)
    sumX2  +=     192.262 * cos(2.16356274415  + 529.93478257810 *T)
    sumX2  +=     119.051 * cos(2.12938419254  + 299.37021175271 *T)
    sumX2  +=     132.408 * cos(0.84833790409  + 52.93401552300 *T)
    sumX2  +=     117.583 * cos(4.12588726472  + 522.82123557730 *T)
    sumX2  +=     114.530 * cos(0.20480870043  + 537.04832957890 *T)
    sumX2  +=     105.238 * cos(0.95165369617  + 127.22797912329 *T)
    sumX2  +=      98.089 * cos(2.01372348892  + 80.44204202221 *T)
    sumX2  +=     114.250 * cos(1.45343138159  + 7.35736448430 *T)
    sumX2  +=      97.881 * cos(0.88047242223  + 97.11681657861 *T)
    sumX2  +=     106.194 * cos(5.89144138305  + 146.62562088040 *T)
    sumX2  +=      89.216 * cos(2.74785227203  + 67.91186905000 *T)
    sumX2  +=     116.340 * cos(0.59881354910  + 38.37685312130 *T)
    sumX2  +=      77.574 * cos(4.98371791372  + 82.13896305161 *T)
    sumX2  +=      84.483 * cos(0.53008500607  + 223.10414047710 *T)
    sumX2  +=      98.707 * cos(3.60232211559  + 62.49524307860 *T)
    sumX2  +=      68.606 * cos(2.46240164068  + 426.84200835950 *T)
    sumX2  +=      66.709 * cos(5.05989525693  + 2.69149803831 *T)
    sumX2  +=      65.809 * cos(3.54607627462  + 15.22167101050 *T)
    sumX2  +=      67.096 * cos(4.82707195481  + 226.07308589371 *T)
    sumX2  +=      87.378 * cos(5.70834290687  + 39.86132582961 *T)
    sumX2  +=      62.935 * cos(2.33622388716  + 149.31937965110 *T)
    sumX2  +=      61.638 * cos(2.56105139517  + 152.98840835580 *T)
    sumX2  +=      65.914 * cos(3.96068940502  + 79.47883417571 *T)
    sumX2  +=      59.769 * cos(2.39173625701  + 10.80188278040 *T)
    sumX2  +=      59.691 * cos(2.89597337716  + 56.37853381910 *T)
    sumX2  +=      56.598 * cos(1.25875671662  + 152.77596003471 *T)
    sumX2  +=      52.271 * cos(5.01620205161  + 70.60562782070 *T)
    sumX2  +=      67.381 * cos(4.08745585432  + 111.67397898031 *T)
    sumX2  +=      48.944 * cos(1.98442603431  + 63.49208081989 *T)
    sumX2  +=      48.935 * cos(4.54101314556  + 194.89603068979 *T)
    sumX2  +=      58.953 * cos(2.87996709683  + 34.92027273770 *T)
    sumX2  +=      42.178 * cos(5.74275407313  + 146.83806920150 *T)
    sumX2  +=      41.131 * cos(0.64777803124  + 209.12312469139 *T)
    sumX2  +=      48.341 * cos(6.17208956617  + 142.69346761730 *T)
    sumX2  +=      39.938 * cos(3.21570509181  + 29.44876501211 *T)
    sumX2  +=      35.504 * cos(4.35546434528  + 73.05330837550 *T)
    sumX2  +=      35.998 * cos(5.79742738847  + 69.60879007940 *T)
    sumX2  +=      35.709 * cos(1.54618860768  + 45.33283355520 *T)
    sumX2  +=      34.365 * cos(3.33620594444  + 70.57199792590 *T)
    sumX2  +=      40.938 * cos(4.61497618218  + 4.69723560840 *T)
    sumX2  +=      33.246 * cos(5.01331392267  + 1059.62574767270 *T)
    sumX2  +=      34.076 * cos(5.46577508001  + 137.27684164590 *T)
    sumX2  +=      34.342 * cos(6.08604337798  + 5.66044345490 *T)
    sumX2  +=      34.750 * cos(2.99991330176  + 120.11443212250 *T)
    sumX2  +=      29.609 * cos(4.31333582780  + 134.82916109110 *T)
    sumX2  +=      29.214 * cos(4.33169131521  + 134.34152612410 *T)
    sumX2  +=      27.728 * cos(0.58899684764  + 202.49721265760 *T)
    sumX2  +=      26.015 * cos(1.44667967680  + 20.85074530300 *T)
    sumX2  +=      25.423 * cos(4.18101350243  + 23.81969071961 *T)
    sumX2  +=      25.360 * cos(1.97190987769  + 60.04756252381 *T)
    sumX2  +=      27.639 * cos(0.63901286173  + 152.25469517290 *T)
    sumX2  +=      24.633 * cos(1.98871425552  + 59.55992755679 *T)
    sumX2  +=      26.033 * cos(4.07056770292  + 235.63431344930 *T)
    sumX2  +=      22.264 * cos(5.20786731153  + 76.02225379210 *T)
    sumX2  +=      20.549 * cos(5.58659383272  + 138.27367938720 *T)
    sumX2  +=      22.051 * cos(0.79574466032  + 33.38091827520 *T)
    sumX2  +=      19.406 * cos(3.47975672795  + 159.36824217371 *T)
    sumX2  +=      20.190 * cos(1.31242826249  + 70.35954960481 *T)
    sumX2  +=      22.751 * cos(5.38605025058  + 110.18950627200 *T)
    sumX2  +=      20.572 * cos(4.61541897614  + 120.60206708950 *T)
    sumX2  +=      19.010 * cos(1.13202069249  + 14.47091148511 *T)
    sumX2  +=      18.238 * cos(0.76330726538  + 3.68833577960 *T)
    sumX2  +=      18.044 * cos(2.65389705259  + 265.74547599400 *T)
    sumX2  +=      18.132 * cos(5.82700586339  + 90.00326957780 *T)
    sumX2  +=      18.837 * cos(3.36592790957  + 79.69128249680 *T)
    sumX2  +=      17.966 * cos(2.30154733724  + 73.32849502140 *T)
    sumX2  +=      17.619 * cos(2.37679027189  + 76.72233708021 *T)
    sumX2  +=      15.431 * cos(5.76818140157  + 1.24065522479 *T)
    sumX2  +=      17.688 * cos(4.07867858977  + 41.40068029210 *T)
    sumX2  +=      15.101 * cos(4.88508430106  + 16.91859203990 *T)
    sumX2  +=      15.248 * cos(6.07362345688  + 3.42521122121 *T)
    sumX2  +=      14.771 * cos(6.23530181561  + 333.41352756050 *T)
    sumX2  +=      14.325 * cos(6.07731251478  + 19.88753745650 *T)
    sumX2  +=      13.703 * cos(5.74950760055  + 116.18227885939 *T)
    sumX2  +=      15.398 * cos(0.55165142555  + 162.33718759030 *T)
    sumX2  +=      13.837 * cos(2.81469028445  + 433.95555536030 *T)
    sumX2  +=      13.525 * cos(1.82049184184  + 144.39038864671 *T)
    sumX2  +=      15.125 * cos(5.47781650962  + 147.35933406330 *T)
    sumX2  +=      15.394 * cos(4.75630456170  + 41.34579853790 *T)
    sumX2  +=      13.487 * cos(2.33365398351  + 347.64062156210 *T)
    sumX2  +=      17.344 * cos(1.14035051184  + 190.96387742670 *T)
    sumX2  +=      12.741 * cos(0.49316850236  + 419.72846135871 *T)
    sumX2  +=      15.299 * cos(1.30791646464  + 184.97110483931 *T)
    sumX2  +=      14.796 * cos(2.98132078729  + 113.15845168861 *T)
    sumX2  +=      11.042 * cos(0.89444511328  + 25.51661174900 *T)
    sumX2  +=      10.803 * cos(6.10809462182  + 140.24578706250 *T)
    sumX2  +=      10.475 * cos(2.21239506785  + 6.86972951729 *T)
    sumX2  +=      11.185 * cos(0.75741549234  + 116.66991382640 *T)
    sumX2  +=      11.147 * cos(4.98644870698  + 54.41848823131 *T)
    sumX2  +=      13.346 * cos(4.08247597684  + 33.43580002939 *T)
    sumX2  +=       9.592 * cos(1.22752159642  + 212.05844021321 *T)
    sumX2  +=      10.605 * cos(0.82432336974  + 74.91354146621 *T)
    sumX2  +=      10.593 * cos(3.85979493686  + 75.13729063540 *T)
    sumX2  +=      11.833 * cos(2.92629549746  + 89.04006173130 *T)
    sumX2  +=       9.674 * cos(0.15663016702  + 129.20008679860 *T)
    sumX2  +=       9.261 * cos(5.39380503857  + 167.96626188280 *T)
    sumX2  +=       9.431 * cos(1.62655798238  + 640.14110379750 *T)
    sumX2  +=      10.948 * cos(5.84704956948  + 202.00957769060 *T)
    sumX2  +=      10.470 * cos(3.21728620233  + 71.56883566720 *T)
    sumX2  +=       8.486 * cos(2.71064949590  + 277.27881122490 *T)
    sumX2  +=       9.140 * cos(1.81085828444  + 228.52076644850 *T)
    sumX2  +=       8.941 * cos(4.65640678923  + 258.63192899320 *T)
    sumX2  +=       8.140 * cos(2.93786869433  + 60.79832204921 *T)
    sumX2  +=       7.518 * cos(3.72398919720  + 126.23114138200 *T)
    sumX2  +=       7.532 * cos(4.84171458960  + 89.25251005240 *T)
    sumX2  +=       7.736 * cos(6.14746801790  + 272.85902299479 *T)
    sumX2  +=       8.693 * cos(1.27494883068  + 221.61966776881 *T)
    sumX2  +=       6.861 * cos(3.28548607175  + 195.38366565681 *T)
    sumX2  +=       8.704 * cos(0.49510828703  + 50.96190784770 *T)
    sumX2  +=       7.676 * cos(3.18883684041  + 58.35064149440 *T)
    sumX2  +=       7.545 * cos(2.72894616317  + 53.93085326429 *T)
    sumX2  +=       7.394 * cos(2.88313029524  + 95.63234387030 *T)
    sumX2  +=       6.681 * cos(5.42279212849  + 66.94866120350 *T)
    sumX2  +=       6.264 * cos(1.99748477436  + 39.60069339870 *T)
    sumX2  +=       7.026 * cos(1.90422724178  + 51.44954281471 *T)
    sumX2  +=       6.150 * cos(4.80885665481  + 374.15181032000 *T)
    sumX2  +=       6.356 * cos(4.52111917010  + 155.22364058950 *T)
    sumX2  +=       5.982 * cos(2.01491416824  + 125.74350641500 *T)
    sumX2  +=       6.068 * cos(0.40878107182  + 135.57992061650 *T)
    sumX2  +=       6.766 * cos(1.53737992411  + 151.50393564750 *T)
    sumX2  +=       7.931 * cos(3.18839879732  + 288.32451148881 *T)
    sumX2  +=       6.447 * cos(2.96047986448  + 297.88573904440 *T)
    sumX2  +=       6.043 * cos(0.77132741272  + 1066.73929467350 *T)
    sumX2  +=       5.769 * cos(0.22669318121  + 213.05527795450 *T)
    sumX2  +=       5.884 * cos(0.89035454231  + 24.13520490470 *T)
    sumX2  +=       5.760 * cos(0.98271147838  + 103.33659170210 *T)
    sumX2  +=       5.890 * cos(0.68025338272  + 46.71424039951 *T)
    sumX2  +=       6.180 * cos(2.84467959186  + 45.82046852221 *T)
    sumX2  +=       6.347 * cos(4.26496745804  + 227.77000692311 *T)
    sumX2  +=       5.285 * cos(2.38608402940  + 305.10235190919 *T)
    sumX2  +=       5.176 * cos(3.56268834670  + 1052.51220067191 *T)
    sumX2  +=       5.443 * cos(1.65371041202  + 471.93102443120 *T)
    sumX2  +=       5.222 * cos(0.59481167464  + 124.53422035260 *T)
    sumX2  +=       5.424 * cos(4.52802960517  + 91.70019060721 *T)
    sumX2  +=       5.814 * cos(1.36394366099  + 340.52707456129 *T)
    sumX2  +=       5.051 * cos(4.26199970891  + 515.70768857651 *T)
    sumX2  +=       4.924 * cos(4.31596399692  + 404.26297286470 *T)
    sumX2  +=       5.332 * cos(2.58586292511  + 83.10217089810 *T)
    sumX2  +=       5.189 * cos(5.10948461536  + 61.01077037031 *T)
    sumX2  +=       5.198 * cos(4.05963240944  + 486.15811843280 *T)
    sumX2  +=       4.404 * cos(3.47905723669  + 171.89841514590 *T)
    sumX2  +=       5.159 * cos(4.18184292472  + 383.06534421420 *T)
    sumX2  +=       4.708 * cos(5.52448195325  + 110.45013870291 *T)
    sumX2  +=       4.387 * cos(3.15889315302  + 209.61075965840 *T)
    sumX2  +=       4.748 * cos(5.94194450321  + 98.60128928690 *T)
    sumX2  +=       4.248 * cos(1.30011269774  + 230.32075334190 *T)
    sumX2  +=       5.145 * cos(1.65634530989  + 74.06220820430 *T)
    sumX2  +=       4.114 * cos(5.01485142282  + 2.20386307129 *T)
    sumX2  +=       5.063 * cos(2.99988089960  + 75.98862389731 *T)
    sumX2  +=       3.940 * cos(5.91769063408  + 55.38169607781 *T)
    sumX2  +=       4.565 * cos(0.81573485480  + 223.20720634109 *T)
    sumX2  +=       3.744 * cos(0.15879098262  + 43.53284666180 *T)
    sumX2  +=       4.710 * cos(0.25957456206  + 186.45557754760 *T)
    sumX2  +=       3.710 * cos(1.73137944713  + 94.66913602380 *T)
    sumX2  +=       3.673 * cos(2.68570227818  + 80.65449034331 *T)
    sumX2  +=       3.673 * cos(5.13529281141  + 69.39634175830 *T)
    sumX2  +=       3.912 * cos(3.10734552053  + 183.85033042589 *T)
    sumX2  +=       3.744 * cos(4.08326458184  + 155.53915477459 *T)
    sumX2  +=       4.397 * cos(6.10187638345  + 902.19824401191 *T)
    sumX2  +=       4.340 * cos(3.39324127850  + 131.16013238639 *T)
    sumX2  +=       3.600 * cos(1.76146036337  + 61.69209392650 *T)
    sumX2  +=       4.967 * cos(3.22509645861  + 108.21739859670 *T)
    sumX2  +=       3.476 * cos(1.54136017705  + 199.31581891990 *T)
    sumX2  +=       3.658 * cos(4.79206377689  + 849.50804597240 *T)
    sumX2  +=       3.523 * cos(3.69437187603  + 451.22103074770 *T)
    sumX2  +=       4.488 * cos(2.74277410385  + 329.48137429739 *T)
    sumX2  +=       3.980 * cos(1.95932786654  + 36.89238041300 *T)
    sumX2  +=       3.375 * cos(5.91720302986  + 415.30867312860 *T)
    sumX2  +=       3.410 * cos(1.87951443558  + 4.20960064140 *T)
    sumX2  +=       3.568 * cos(6.07468170604  + 446.06752933469 *T)
    sumX2  +=       4.379 * cos(1.89351588041  + 145.35359649321 *T)
    sumX2  +=       3.342 * cos(5.45803791124  + 295.43805848960 *T)
    sumX2  +=       3.570 * cos(6.04818504684  + 50.64639366260 *T)
    sumX2  +=       3.501 * cos(2.14476824270  + 460.29462333630 *T)
    sumX2  +=       3.555 * cos(0.03884741871  + 397.14942586389 *T)
    sumX2  +=       3.345 * cos(0.09439746136  + 544.16187657970 *T)
    sumX2  +=       3.245 * cos(4.46873398006  + 428.32648106781 *T)
    sumX2  +=       3.326 * cos(2.20835247016  + 354.75416856289 *T)
    sumX2  +=       3.517 * cos(2.35139644667  + 204.94489321240 *T)
    sumX2  +=       3.209 * cos(2.99492291709  + 237.43430034270 *T)
    sumX2  +=       3.283 * cos(1.61893390630  + 32.48714639791 *T)
    sumX2  +=       3.195 * cos(6.27817675624  + 1058.14127496441 *T)
    sumX2  +=       3.291 * cos(3.75210973869  + 504.80500066450 *T)
    sumX2  +=       3.210 * cos(1.96285885557  + 323.74923414091 *T)
    sumX2  +=       3.632 * cos(5.42847493228  + 57.75994066341 *T)
    sumX2  +=       3.617 * cos(4.97056941864  + 104.23036357940 *T)
    sumX2  +=       3.794 * cos(6.10025129416  + 294.05665164530 *T)
    sumX2  +=       3.164 * cos(6.16087797760  + 1819.39364862570 *T)
    sumX2  +=       3.165 * cos(3.65964152483  + 216.72430665921 *T)
    sumX2  +=       3.165 * cos(0.51804887124  + 216.23667169220 *T)
    sumX2  +=       3.073 * cos(5.18145172576  + 41.88831525910 *T)
    sumX2  +=       3.059 * cos(5.35132993250  + 905.64276230799 *T)
    sumX2  +=       3.362 * cos(5.69676404259  + 85.58348134769 *T)
    sumX2  +=       3.541 * cos(3.69183627724  + 956.53297345411 *T)
    sumX2  +=       3.541 * cos(5.63801834999  + 191.45151239371 *T)
    sumX2  +=       3.260 * cos(3.54321329291  + 227.55755860200 *T)
    sumX2  +=       3.047 * cos(0.50860086202  + 531.41925528640 *T)
    sumX2  +=       2.983 * cos(5.52362556155  + 1589.31671276730 *T)
    sumX2  +=       3.264 * cos(5.67637422985  + 336.59492129820 *T)
    sumX2  +=       3.273 * cos(3.54065334147  + 60.31068708220 *T)
    sumX2  +=       3.052 * cos(1.58893349421  + 411.37651986550 *T)
    sumX2  +=       2.957 * cos(3.38662804009  + 207.91383862900 *T)
    sumX2  +=       3.082 * cos(1.86722969180  + 217.47506618460 *T)
    sumX2  +=       3.147 * cos(6.00497294061  + 1169.83206889211 *T)
    sumX2  +=       3.055 * cos(5.34963236447  + 381.36842318480 *T)
    sumX2  +=       3.006 * cos(3.41332999286  + 290.87525790760 *T)
    sumX2  +=       2.859 * cos(0.87412339321  + 276.79117625789 *T)
    sumX2  +=       2.836 * cos(6.05953478189  + 88.35873817511 *T)
    sumX2  +=       2.790 * cos(2.75758754465  + 62.00760811160 *T)
    sumX2  +=       3.077 * cos(1.58786085482  + 109.06873185860 *T)
    sumX2  +=       2.839 * cos(1.72539635996  + 942.30587945250 *T)
    sumX2  +=       2.822 * cos(3.28498999245  + 128.71245183160 *T)
    sumX2  +=       3.374 * cos(2.87502037309  + 736.12033101530 *T)
    sumX2  +=       2.933 * cos(2.23835900787  + 306.58682461749 *T)
    sumX2  +=       2.763 * cos(5.42861535964  + 1282.58913653029 *T)
    sumX2  +=       2.895 * cos(5.69820720276  + 75.54668091261 *T)
    sumX2  +=       3.459 * cos(2.93102396110  + 610.44852130189 *T)
    sumX2  +=       2.886 * cos(5.26577859528  + 74.50415118900 *T)
    sumX2  +=       2.638 * cos(3.75521976789  + 198.07742442750 *T)
    sumX2  +=       2.939 * cos(5.73905357521  + 179.91817716280 *T)
    sumX2  +=       2.556 * cos(1.29879665561  + 430.02340209721 *T)
    sumX2  +=       2.743 * cos(3.82938065661  + 141.94270809190 *T)
    sumX2  +=       2.743 * cos(3.82938065661  + 141.45507312490 *T)
    sumX2  +=       2.525 * cos(4.29423148270  + 219.17198721400 *T)
    sumX2  +=       2.654 * cos(4.61997728072  + 475.86317769430 *T)
    sumX2  +=       3.424 * cos(3.63115669579  + 133.13224006171 *T)
    sumX2  +=       2.448 * cos(3.34099164118  + 358.68632182600 *T)
    sumX2  +=       2.613 * cos(3.56416070170  + 1155.60497489050 *T)
    sumX2  +=       2.428 * cos(0.92534350100  + 714.92270236481 *T)
    sumX2  +=       3.038 * cos(0.02659248152  + 145.38722638800 *T)
    sumX2  +=       2.558 * cos(0.15484452507  + 460.88532416730 *T)
    sumX2  +=       2.602 * cos(0.97639397784  + 4.90968392950 *T)
    sumX2  +=       2.602 * cos(0.56141580457  + 145.14114817210 *T)
    sumX2  +=       2.603 * cos(4.00253973196  + 343.46239008310 *T)
    sumX2  +=       3.095 * cos(4.96470959339  + 269.67762925710 *T)
    sumX2  +=       3.095 * cos(4.96470959339  + 270.16526422410 *T)
    sumX2  +=       2.576 * cos(0.29697709235  + 988.02554284310 *T)
    sumX2  +=       2.387 * cos(5.04233641349  + 346.43133549971 *T)
    sumX2  +=       2.643 * cos(3.31153040747  + 376.70255673880 *T)
    sumX2  +=       2.290 * cos(5.05227824278  + 275.30670354960 *T)
    sumX2  +=       3.061 * cos(6.15240912134  + 606.51636803880 *T)
    sumX2  +=       2.264 * cos(0.83314819323  + 323.26159917390 *T)
    sumX2  +=       2.638 * cos(0.79175765980  + 77.71917482150 *T)
    sumX2  +=       2.246 * cos(5.55551202459  + 143.90275367970 *T)
    sumX2  +=       2.334 * cos(2.10308783101  + 49.26498681829 *T)
    sumX2  +=       2.334 * cos(5.24468048460  + 49.75262178531 *T)
    sumX2  +=       2.696 * cos(4.28534164626  + 70.08436295889 *T)
    sumX2  +=       2.327 * cos(3.06369666366  + 479.04457143200 *T)
    sumX2  +=       2.398 * cos(2.55511918128  + 183.48663213100 *T)
    sumX2  +=       2.040 * cos(1.27681238416  + 1968.95684576030 *T)
    sumX2  +=       2.040 * cos(1.27681238416  + 1969.44448072730 *T)
    sumX2  +=       1.960 * cos(5.72020707325  + 894.59706204410 *T)
    sumX2  +=       1.960 * cos(2.57861441966  + 895.08469701110 *T)
    sumX2  +=       2.286 * cos(2.84819442391  + 147.83490694279 *T)
    sumX2  +=       2.536 * cos(1.61730292038  + 108.70503356371 *T)
    sumX2  +=       1.832 * cos(6.16198387079  + 379.88395047650 *T)
    sumX2  +=       1.977 * cos(3.03212060303  + 5.17280848790 *T)
    sumX2  +=       2.397 * cos(1.86738161703  + 114.64292439690 *T)
    sumX2  +=       2.450 * cos(4.54363281495  + 203.98168536590 *T)
    sumX2  +=       2.258 * cos(2.03199421502  + 201.01273994931 *T)
    sumX2  +=       1.767 * cos(0.29075268819  + 1514.04747923300 *T)
    sumX2  +=       1.767 * cos(3.43234534178  + 1514.53511420000 *T)
    sumX2  +=       1.978 * cos(2.12934272436  + 2.72512793310 *T)
    sumX2  +=       1.763 * cos(3.56951955484  + 275.79433851660 *T)
    sumX2  +=       1.714 * cos(5.14400730282  + 440.43845504219 *T)
    sumX2  +=       1.714 * cos(5.14400730282  + 440.92609000920 *T)
    sumX2  +=       1.710 * cos(0.74103984530  + 309.03450517230 *T)
    sumX2  +=       1.710 * cos(0.74103984530  + 309.52214013930 *T)
    sumX2  +=       1.917 * cos(3.04967782266  + 6.18170831670 *T)
    sumX2  +=       1.613 * cos(0.68793280970  + 1044.16025917870 *T)
    sumX2  +=       1.613 * cos(0.68793280970  + 1044.64789414570 *T)
    sumX2  +=       1.722 * cos(4.76110262034  + 752.89817143570 *T)
    sumX2  +=       1.722 * cos(1.61950996675  + 753.38580640270 *T)
    sumX2  +=       1.656 * cos(3.95388550783  + 21.84758304429 *T)
    sumX2  +=       2.072 * cos(1.38441205546  + 203.49405039889 *T)
    sumX2  +=       1.528 * cos(1.43661391386  + 106.51798543980 *T)
    sumX2  +=       1.528 * cos(4.57820656745  + 106.03035047280 *T)
    sumX2  +=       1.802 * cos(0.72583106929  + 344.45922782439 *T)
    sumX2  +=       1.802 * cos(0.72583106929  + 344.94686279140 *T)
    sumX2  +=       1.610 * cos(0.77273753295  + 266.23311096100 *T)
    sumX2  +=       1.509 * cos(2.15678911314  + 924.28964453971 *T)
    sumX2  +=       1.509 * cos(2.15678911314  + 923.80200957270 *T)
    sumX2  +=       1.727 * cos(2.66552748343  + 625.16325027050 *T)
    sumX2  +=       1.727 * cos(5.80712013702  + 624.67561530350 *T)
    sumX2  +=       1.491 * cos(4.05516091832  + 422.90985509640 *T)
    sumX2  +=       1.491 * cos(0.91356826473  + 422.42222012939 *T)
    sumX2  +=       1.766 * cos(5.63078579097  + 224.10097821840 *T)
    sumX2  +=       2.006 * cos(3.39755981160  + 144.86596152620 *T)
    sumX2  +=       1.470 * cos(5.47870687186  + 284.39235822571 *T)
    sumX2  +=       1.470 * cos(2.33711421827  + 283.90472325870 *T)
    sumX2  +=       1.482 * cos(0.38832219784  + 17.91542978119 *T)
    sumX2  +=       1.585 * cos(3.24031630932  + 710.99054910171 *T)
    sumX2  +=       1.585 * cos(0.09872365573  + 710.50291413470 *T)
    sumX2  +=       1.415 * cos(0.78027163232  + 13.98327651810 *T)
    sumX2  +=       1.408 * cos(1.20663638110  + 254.69977573010 *T)
    sumX2  +=       1.408 * cos(4.34822903469  + 255.18741069710 *T)
    sumX2  +=       1.517 * cos(2.68660553420  + 976.97984257920 *T)
    sumX2  +=       1.517 * cos(2.68660553420  + 976.49220761220 *T)

    # URANUS X Coord
    sumX3  = 0
    sumX3  +=     937.647 * cos(1.00888984070  + 149.80701461810 *T)
    sumX3  +=     967.548 * cos(1.42552804346  + 0.24381748350 *T)
    sumX3  +=     892.089 * cos(0.79608202303  + 75.02541605080 *T)
    sumX3  +=     684.660 * cos(2.90453472816  + 71.09326278771 *T)
    sumX3  +=     623.727 * cos(4.86286247873  + 78.95756931390 *T)
    sumX3  +=     525.597 * cos(3.81206896615  + 76.50988875911 *T)
    sumX3  +=     507.791 * cos(4.64109877567  + 63.97971578690 *T)
    sumX3  +=     496.551 * cos(3.13919316435  + 86.07111631471 *T)
    sumX3  +=     445.874 * cos(4.99904086064  + 71.84402231310 *T)
    sumX3  +=     424.392 * cos(2.80391306791  + 78.20680978850 *T)
    sumX3  +=     366.041 * cos(2.07755748247  + 11.28951774740 *T)
    sumX3  +=     313.023 * cos(4.23932450493  + 73.54094334250 *T)
    sumX3  +=     249.030 * cos(2.48148146374  + 4.17597074660 *T)
    sumX3  +=     209.150 * cos(1.40893371766  + 74.53778108379 *T)
    sumX3  +=     127.465 * cos(5.54018220786  + 138.76131435421 *T)
    sumX3  +=     120.647 * cos(5.46876561078  + 18.40306474821 *T)
    sumX3  +=     121.784 * cos(5.29001670826  + 145.87486135500 *T)
    sumX3  +=      78.251 * cos(3.48103273393  + 224.58861318540 *T)
    sumX3  +=      84.577 * cos(0.33560464127  + 2.93757625420 *T)
    sumX3  +=      50.491 * cos(2.35623662812  + 131.64776735340 *T)
    sumX3  +=      45.951 * cos(6.24540196810  + 151.29148732640 *T)
    sumX3  +=      46.649 * cos(0.55435911040  + 56.86616878610 *T)
    sumX3  +=      48.261 * cos(2.93721017694  + 22.33521801130 *T)
    sumX3  +=      34.791 * cos(1.00162978861  + 93.18466331550 *T)
    sumX3  +=      29.328 * cos(1.75294089893  + 220.65645992230 *T)
    sumX3  +=      30.594 * cos(5.68097375758  + 160.85271488200 *T)
    sumX3  +=      25.988 * cos(1.11718356254  + 153.73916788121 *T)
    sumX3  +=      20.587 * cos(3.04578219289  + 206.42936592071 *T)
    sumX3  +=      18.718 * cos(3.43630968031  + 82.13896305161 *T)
    sumX3  +=      18.634 * cos(4.35949046286  + 67.91186905000 *T)
    sumX3  +=      16.697 * cos(2.01487108353  + 87.55558902300 *T)
    sumX3  +=      20.883 * cos(0.13997936753  + 84.58664360640 *T)
    sumX3  +=      20.072 * cos(1.42313946401  + 65.46418849521 *T)
    sumX3  +=      15.729 * cos(5.83349115204  + 522.82123557730 *T)
    sumX3  +=      15.573 * cos(4.75214345059  + 537.04832957890 *T)
    sumX3  +=      19.429 * cos(4.90609476184  + 127.71561409031 *T)
    sumX3  +=      14.513 * cos(1.55811447849  + 52.44638055600 *T)
    sumX3  +=      14.273 * cos(3.15874513195  + 52.93401552300 *T)
    sumX3  +=      17.225 * cos(1.23506274281  + 146.62562088040 *T)
    sumX3  +=      14.135 * cos(2.83531205367  + 77.99436146740 *T)
    sumX3  +=      13.035 * cos(1.86446595495  + 72.05647063421 *T)
    sumX3  +=      14.349 * cos(5.92118700781  + 148.32254190981 *T)
    sumX3  +=       9.002 * cos(4.74037798245  + 97.11681657861 *T)
    sumX3  +=       8.798 * cos(4.52661827695  + 56.37853381910 *T)
    sumX3  +=       8.208 * cos(5.82435206434  + 62.49524307860 *T)
    sumX3  +=       7.603 * cos(1.13885911665  + 152.98840835580 *T)
    sumX3  +=       7.140 * cos(6.25024559600  + 194.89603068979 *T)
    sumX3  +=       6.892 * cos(1.11490592145  + 80.65449034331 *T)
    sumX3  +=       6.892 * cos(0.42290386096  + 69.39634175830 *T)
    sumX3  +=       6.511 * cos(3.55251442874  + 45.33283355520 *T)
    sumX3  +=       6.477 * cos(5.17331047084  + 209.12312469139 *T)
    sumX3  +=       5.804 * cos(3.28871244212  + 70.60562782070 *T)
    sumX3  +=       5.326 * cos(2.66683839862  + 127.22797912329 *T)
    sumX3  +=       5.029 * cos(4.10585397027  + 149.31937965110 *T)
    sumX3  +=       5.009 * cos(5.03201863198  + 297.98880490840 *T)
    sumX3  +=       5.452 * cos(1.62029053066  + 72.57773549600 *T)
    sumX3  +=       6.294 * cos(6.09382218854  + 10.80188278040 *T)
    sumX3  +=       5.251 * cos(3.00689879881  + 77.47309660560 *T)
    sumX3  +=       4.684 * cos(3.18410902200  + 202.49721265760 *T)
    sumX3  +=       4.291 * cos(2.72886332432  + 134.34152612410 *T)
    sumX3  +=       5.421 * cos(4.86504346481  + 120.11443212250 *T)
    sumX3  +=       4.058 * cos(3.84771327094  + 312.21589891000 *T)
    sumX3  +=       3.926 * cos(0.20635932899  + 59.55992755679 *T)
    sumX3  +=       4.419 * cos(0.79082946257  + 69.60879007940 *T)
    sumX3  +=       4.250 * cos(3.81801389997  + 80.44204202221 *T)
    sumX3  +=       4.556 * cos(1.09930492481  + 29.44876501211 *T)
    sumX3  +=       3.592 * cos(3.12402973242  + 199.31581891990 *T)
    sumX3  +=       3.523 * cos(3.84292492170  + 636.20895053440 *T)
    sumX3  +=       3.428 * cos(4.80416729358  + 621.98185653281 *T)
    sumX3  +=       3.942 * cos(3.83644807496  + 3.21276290011 *T)
    sumX3  +=       3.620 * cos(3.64168879156  + 3.42521122121 *T)
    sumX3  +=       3.227 * cos(2.36250033807  + 3.68833577960 *T)
    sumX3  +=       3.564 * cos(2.97403981009  + 70.35954960481 *T)
    sumX3  +=       3.564 * cos(1.70536262591  + 79.69128249680 *T)
    sumX3  +=       3.552 * cos(0.48986762414  + 426.84200835950 *T)
    sumX3  +=       3.245 * cos(4.05425635175  + 611.82992814620 *T)
    sumX3  +=       3.090 * cos(2.27444674300  + 486.15811843280 *T)
    sumX3  +=       3.149 * cos(3.23221666282  + 471.93102443120 *T)
    sumX3  +=       2.864 * cos(4.40034684781  + 272.85902299479 *T)
    sumX3  +=       3.756 * cos(0.28897263531  + 258.63192899320 *T)
    sumX3  +=       2.808 * cos(6.27609619511  + 647.25465079831 *T)
    sumX3  +=       2.640 * cos(2.22066140796  + 288.32451148881 *T)
    sumX3  +=       2.544 * cos(0.39091843469  + 76.02225379210 *T)
    sumX3  +=       2.670 * cos(2.78782481368  + 375.95179721340 *T)
    sumX3  +=       2.998 * cos(0.71551145066  + 76.72233708021 *T)
    sumX3  +=       2.998 * cos(3.96389098534  + 73.32849502140 *T)
    sumX3  +=       3.125 * cos(1.93972182076  + 408.68276109481 *T)
    sumX3  +=       2.330 * cos(0.87729499827  + 433.95555536030 *T)
    sumX3  +=       2.645 * cos(0.46205095531  + 120.60206708950 *T)
    sumX3  +=       2.219 * cos(3.00549424798  + 390.92965074040 *T)
    sumX3  +=       2.517 * cos(1.04902375688  + 213.54291292150 *T)
    sumX3  +=       2.074 * cos(3.43164566806  + 63.49208081989 *T)
    sumX3  +=       1.894 * cos(1.69172880285  + 213.05527795450 *T)
    sumX3  +=       1.849 * cos(2.61330239059  + 134.82916109110 *T)
    sumX3  +=       1.838 * cos(5.57764928554  + 73.05330837550 *T)
    sumX3  +=       2.214 * cos(0.39891007096  + 1.72829019180 *T)
    sumX3  +=       1.644 * cos(4.66534694390  + 384.30373870660 *T)
    sumX3  +=       1.644 * cos(4.66534694390  + 383.81610373959 *T)
    sumX3  +=       1.553 * cos(5.52374457077  + 462.26673101160 *T)
    sumX3  +=       1.553 * cos(2.38215191718  + 461.77909604459 *T)
    sumX3  +=       1.419 * cos(5.77169479025  + 422.90985509640 *T)
    sumX3  +=       1.419 * cos(2.63010213666  + 422.42222012939 *T)
    sumX3  +=       1.411 * cos(4.93713727855  + 9.80504503910 *T)
    sumX3  +=       1.443 * cos(5.13235693386  + 195.38366565681 *T)
    sumX3  +=       1.402 * cos(4.46431718602  + 984.84414910540 *T)
    sumX3  +=       1.402 * cos(4.46431718602  + 984.35651413839 *T)

    # URANUS X Coord
    sumX4  = 0
    sumX4  +=    2203.535 * cos(3.78800316770  + 75.02541605080 *T)
    sumX4  +=     128.370 * cos(4.00299274369  + 0.24381748350 *T)
    sumX4  +=      72.353 * cos(5.87048427972  + 149.80701461810 *T)
    sumX4  +=      30.194 * cos(0.37339777366  + 63.97971578690 *T)
    sumX4  +=      30.655 * cos(1.07320659320  + 86.07111631471 *T)
    sumX4  +=      16.471 * cos(3.60870796009  + 18.40306474821 *T)
    sumX4  +=      16.067 * cos(2.53660745454  + 76.50988875911 *T)
    sumX4  +=      16.064 * cos(1.88841261109  + 73.54094334250 *T)
    sumX4  +=      10.244 * cos(3.15336566765  + 74.53778108379 *T)
    sumX4  +=       6.697 * cos(4.11208833064  + 131.64776735340 *T)
    sumX4  +=       7.164 * cos(3.44355611982  + 1.72829019180 *T)
    sumX4  +=       6.834 * cos(2.38206416682  + 56.86616878610 *T)
    sumX4  +=       4.835 * cos(5.58205045904  + 11.28951774740 *T)
    sumX4  +=       5.294 * cos(5.39389664747  + 93.18466331550 *T)
    sumX4  +=       3.031 * cos(0.09710279040  + 71.09326278771 *T)
    sumX4  +=       2.817 * cos(4.57947190523  + 78.95756931390 *T)
    sumX4  +=       2.343 * cos(0.07235611013  + 220.65645992230 *T)
    sumX4  +=       2.389 * cos(0.73257610170  + 283.90472325870 *T)
    sumX4  +=       2.294 * cos(3.03614566251  + 365.65685647490 *T)
    sumX4  +=       2.237 * cos(4.80048540778  + 206.42936592071 *T)
    sumX4  +=       1.842 * cos(5.42341768633  + 138.76131435421 *T)
    sumX4  +=       2.009 * cos(2.15212981432  + 224.58861318540 *T)
    sumX4  +=       1.813 * cos(0.96578276317  + 148.32254190981 *T)

    # URANUS X Coord
    sumX5  = 0
    sumX5  +=      16.766 * cos(3.88666258140  + 75.02541605080 *T)
    sumX5  +=       1.843 * cos(6.11677355708  + 0.24381748350 *T)

    # URANUS Y Coord
    sumY0  = 0
    sumY0  +=  1916944479.396 * cos(3.91049731307  + 75.02541605080 *T)
    sumY0  +=  133267708.718 * cos(4.59010345878  + 0.24381748350 *T)
    sumY0  +=  44396480.992 * cos(0.08885999374  + 149.80701461810 *T)
    sumY0  +=  14712072.726 * cos(1.85369914992  + 73.54094334250 *T)
    sumY0  +=  14127113.794 * cos(2.82489686708  + 76.50988875911 *T)
    sumY0  +=  6225592.204 * cos(3.56962085379  + 1.72829019180 *T)
    sumY0  +=  1542809.804 * cos(2.55041372379  + 224.58861318540 *T)
    sumY0  +=  1443286.598 * cos(1.08021023229  + 148.32254190981 *T)
    sumY0  +=  941982.751 * cos(0.09031933919  + 11.28951774740 *T)
    sumY0  +=  657433.967 * cos(5.28832386804  + 151.29148732640 *T)
    sumY0  +=  637676.334 * cos(2.64460991486  + 63.97971578690 *T)
    sumY0  +=  621475.723 * cos(1.48799213488  + 77.99436146740 *T)
    sumY0  +=  563553.759 * cos(3.23634730851  + 72.05647063421 *T)
    sumY0  +=  547585.874 * cos(2.06043009525  + 86.07111631471 *T)
    sumY0  +=  433273.658 * cos(0.94921595274  + 74.53778108379 *T)
    sumY0  +=  458904.544 * cos(2.33727121263  + 3.21276290011 *T)
    sumY0  +=  496011.875 * cos(5.31196024834  + 529.93478257810 *T)
    sumY0  +=  385773.622 * cos(4.61855610406  + 138.76131435421 *T)
    sumY0  +=  268150.781 * cos(5.68104913033  + 213.54291292150 *T)
    sumY0  +=  215803.817 * cos(3.73799390950  + 38.37685312130 *T)
    sumY0  +=  144768.728 * cos(0.74847059103  + 71.09326278771 *T)
    sumY0  +=  135315.426 * cos(3.93976545519  + 78.95756931390 *T)
    sumY0  +=  119632.236 * cos(2.53058847664  + 39.86132582961 *T)
    sumY0  +=  124987.110 * cos(0.94345807180  + 111.67397898031 *T)
    sumY0  +=  111232.552 * cos(3.55168186769  + 223.10414047710 *T)
    sumY0  +=  104563.861 * cos(2.33402510023  + 146.83806920150 *T)
    sumY0  +=  109354.782 * cos(6.02395126243  + 34.92027273770 *T)
    sumY0  +=   63579.167 * cos(5.01205049565  + 299.37021175271 *T)
    sumY0  +=   76241.847 * cos(4.10733816861  + 63.49208081989 *T)
    sumY0  +=   53591.167 * cos(2.36965802300  + 4.17597074660 *T)
    sumY0  +=   64420.500 * cos(2.15769822929  + 110.18950627200 *T)
    sumY0  +=   44566.570 * cos(2.03132488718  + 73.05330837550 *T)
    sumY0  +=   39200.127 * cos(1.11801397388  + 4.69723560840 *T)
    sumY0  +=   33660.673 * cos(0.94726438208  + 65.46418849521 *T)
    sumY0  +=   34338.030 * cos(1.46699099857  + 226.07308589371 *T)
    sumY0  +=   34546.984 * cos(0.27616739379  + 79.47883417571 *T)
    sumY0  +=   37552.003 * cos(5.71118714669  + 202.00957769060 *T)
    sumY0  +=   29546.886 * cos(4.42962464119  + 70.57199792590 *T)
    sumY0  +=   25683.031 * cos(3.73653421618  + 9.80504503910 *T)
    sumY0  +=   23447.563 * cos(2.52695137802  + 145.87486135500 *T)
    sumY0  +=   22948.537 * cos(3.94425470497  + 84.58664360640 *T)
    sumY0  +=   31820.127 * cos(3.96864561052  + 152.77596003471 *T)
    sumY0  +=   22832.266 * cos(3.86462602619  + 127.22797912329 *T)
    sumY0  +=   28378.722 * cos(4.44710211392  + 184.97110483931 *T)
    sumY0  +=   26655.018 * cos(4.53946351355  + 160.85271488200 *T)
    sumY0  +=   19671.485 * cos(3.96350915722  + 74.91354146621 *T)
    sumY0  +=   19648.859 * cos(0.71579539008  + 75.13729063540 *T)
    sumY0  +=   19896.586 * cos(5.28901013119  + 12.77399045571 *T)
    sumY0  +=   24486.297 * cos(3.56492038672  + 1.24065522479 *T)
    sumY0  +=   16842.912 * cos(2.04990902220  + 52.44638055600 *T)
    sumY0  +=   22087.252 * cos(6.16990066193  + 71.56883566720 *T)
    sumY0  +=   20068.191 * cos(2.90302034853  + 22.33521801130 *T)
    sumY0  +=   19913.902 * cos(6.11096260597  + 113.15845168861 *T)
    sumY0  +=   18351.147 * cos(0.98731771489  + 33.43580002939 *T)
    sumY0  +=   16887.001 * cos(4.92522785944  + 36.89238041300 *T)
    sumY0  +=   15173.078 * cos(1.31324926487  + 41.34579853790 *T)
    sumY0  +=   11242.410 * cos(4.54512860104  + 71.84402231310 *T)
    sumY0  +=   13948.685 * cos(4.70470503634  + 221.61966776881 *T)
    sumY0  +=   10809.036 * cos(0.12879555337  + 78.20680978850 *T)
    sumY0  +=   13591.810 * cos(0.98320955083  + 87.55558902300 *T)
    sumY0  +=   11997.310 * cos(5.66121692499  + 1059.62574767270 *T)
    sumY0  +=   12401.284 * cos(4.64879549493  + 72.57773549600 *T)
    sumY0  +=   11534.891 * cos(0.20176113714  + 77.47309660560 *T)
    sumY0  +=    9194.297 * cos(3.56604916160  + 145.35359649321 *T)
    sumY0  +=    8737.279 * cos(3.39876002185  + 186.45557754760 *T)
    sumY0  +=    9136.217 * cos(1.79342336217  + 127.71561409031 *T)
    sumY0  +=    9219.832 * cos(0.08407294164  + 108.21739859670 *T)
    sumY0  +=    7422.726 * cos(3.38880260286  + 108.70503356371 *T)
    sumY0  +=    7094.598 * cos(6.01618568387  + 297.88573904440 *T)
    sumY0  +=    6261.877 * cos(0.14282449694  + 153.73916788121 *T)
    sumY0  +=    7490.650 * cos(4.82606551806  + 426.84200835950 *T)
    sumY0  +=    7827.357 * cos(1.11765019484  + 340.52707456129 *T)
    sumY0  +=    6073.898 * cos(3.36888016003  + 149.31937965110 *T)
    sumY0  +=    5510.674 * cos(1.73980185942  + 140.24578706250 *T)
    sumY0  +=    6397.195 * cos(3.71974404814  + 62.49524307860 *T)
    sumY0  +=    5370.666 * cos(2.55420803164  + 75.54668091261 *T)
    sumY0  +=    5352.595 * cos(2.12177919322  + 74.50415118900 *T)
    sumY0  +=    4493.418 * cos(5.82649121556  + 66.94866120350 *T)
    sumY0  +=    4261.392 * cos(5.53250398932  + 265.74547599400 *T)
    sumY0  +=    5026.106 * cos(3.11745853421  + 18.40306474821 *T)
    sumY0  +=    4641.163 * cos(4.97358578602  + 36.40474544600 *T)
    sumY0  +=    4448.201 * cos(5.69436641871  + 183.48663213100 *T)
    sumY0  +=    3781.540 * cos(2.58174473007  + 10.80188278040 *T)
    sumY0  +=    4750.672 * cos(3.38680531662  + 74.06220820430 *T)
    sumY0  +=    4350.588 * cos(1.16130513588  + 70.08436295889 *T)
    sumY0  +=    4446.999 * cos(5.00656772452  + 114.64292439690 *T)
    sumY0  +=    3313.232 * cos(5.23356115546  + 83.10217089810 *T)
    sumY0  +=    4512.500 * cos(1.31375599245  + 75.98862389731 *T)
    sumY0  +=    3556.191 * cos(6.18886493840  + 6.18170831670 *T)
    sumY0  +=    3266.875 * cos(5.23064845262  + 220.65645992230 *T)
    sumY0  +=    3059.638 * cos(6.23776415109  + 52.93401552300 *T)
    sumY0  +=    3576.768 * cos(5.83381085162  + 137.27684164590 *T)
    sumY0  +=    2881.563 * cos(1.19047690440  + 374.15181032000 *T)
    sumY0  +=    3196.605 * cos(6.12668597740  + 76.02225379210 *T)
    sumY0  +=    2944.255 * cos(3.52709483180  + 7.35736448430 *T)
    sumY0  +=    3729.115 * cos(0.25277739547  + 144.86596152620 *T)
    sumY0  +=    2521.158 * cos(0.28068609662  + 202.49721265760 *T)
    sumY0  +=    2595.960 * cos(4.93669910815  + 97.11681657861 *T)
    sumY0  +=    2560.051 * cos(2.87170410480  + 80.44204202221 *T)
    sumY0  +=    2653.876 * cos(1.61857820268  + 305.10235190919 *T)
    sumY0  +=    2508.094 * cos(2.21376907090  + 31.95132732110 *T)
    sumY0  +=    1961.925 * cos(4.80626439560  + 20.85074530300 *T)
    sumY0  +=    2229.736 * cos(5.34787936891  + 80.96330688400 *T)
    sumY0  +=    2202.407 * cos(0.49391085353  + 2.93757625420 *T)
    sumY0  +=    1935.845 * cos(0.49503923254  + 75.07360016060 *T)
    sumY0  +=    1935.843 * cos(4.18436372115  + 74.97723194101 *T)
    sumY0  +=    2138.399 * cos(2.63817800787  + 74.76478361991 *T)
    sumY0  +=    2128.668 * cos(2.04117900040  + 75.28604848170 *T)
    sumY0  +=    2214.188 * cos(0.61512393401  + 259.75270340660 *T)
    sumY0  +=    1890.291 * cos(3.92857018200  + 300.85468446100 *T)
    sumY0  +=    2249.087 * cos(4.91851562646  + 138.27367938720 *T)
    sumY0  +=    1926.813 * cos(6.00497183129  + 159.36824217371 *T)
    sumY0  +=    1799.526 * cos(4.16205107006  + 74.86535735640 *T)
    sumY0  +=    1794.407 * cos(0.51735541616  + 75.18547474520 *T)
    sumY0  +=    1908.592 * cos(5.63004607587  + 69.08752521760 *T)
    sumY0  +=    2175.205 * cos(1.57021660383  + 131.64776735340 *T)
    sumY0  +=    1898.923 * cos(2.65284957017  + 206.42936592071 *T)
    sumY0  +=    1654.495 * cos(2.32498724537  + 277.27881122490 *T)
    sumY0  +=    1785.247 * cos(5.91797544176  + 56.86616878610 *T)
    sumY0  +=    1617.939 * cos(0.41918980719  + 2.69149803831 *T)
    sumY0  +=    1860.708 * cos(0.10451537367  + 42.83027124621 *T)
    sumY0  +=    2180.164 * cos(2.80636515730  + 479.04457143200 *T)
    sumY0  +=    1576.477 * cos(3.66272707282  + 190.96387742670 *T)
    sumY0  +=    1904.361 * cos(4.26492017662  + 15.22167101050 *T)
    sumY0  +=    2054.711 * cos(5.17344254449  + 834.79331700380 *T)
    sumY0  +=    1834.581 * cos(1.34736149542  + 106.73292588839 *T)
    sumY0  +=    1890.085 * cos(2.75024642804  + 154.26043274300 *T)
    sumY0  +=    1795.901 * cos(0.16314353880  + 227.55755860200 *T)
    sumY0  +=    1670.623 * cos(5.91846317829  + 220.13519506050 *T)
    sumY0  +=    1347.353 * cos(3.89460250777  + 288.32451148881 *T)
    sumY0  +=    1425.992 * cos(5.10014924331  + 93.18466331550 *T)
    sumY0  +=    1265.397 * cos(0.92611311536  + 404.26297286470 *T)
    sumY0  +=    1581.901 * cos(4.15166691641  + 39.60069339870 *T)
    sumY0  +=    1269.733 * cos(1.09468107953  + 142.69346761730 *T)
    sumY0  +=    1653.676 * cos(0.42042255686  + 181.51452445570 *T)
    sumY0  +=    1437.856 * cos(5.44092263677  + 522.82123557730 *T)
    sumY0  +=    1407.281 * cos(2.04674828978  + 537.04832957890 *T)
    sumY0  +=    1359.462 * cos(5.38881787204  + 62.00760811160 *T)
    sumY0  +=    1417.609 * cos(0.72618172617  + 235.63431344930 *T)
    sumY0  +=    1122.290 * cos(4.85351571060  + 60.04756252381 *T)
    sumY0  +=    1037.982 * cos(1.19097992567  + 5.66044345490 *T)
    sumY0  +=    1164.766 * cos(0.51146987374  + 82.13896305161 *T)
    sumY0  +=    1009.353 * cos(0.45307981900  + 74.27465652540 *T)
    sumY0  +=    1216.357 * cos(1.84744066187  + 212.05844021321 *T)
    sumY0  +=     996.588 * cos(4.22632308814  + 75.77617557621 *T)
    sumY0  +=     983.856 * cos(6.27680570981  + 129.20008679860 *T)
    sumY0  +=    1223.267 * cos(2.27299989031  + 187.94005025590 *T)
    sumY0  +=    1335.581 * cos(3.74795268087  + 379.88395047650 *T)
    sumY0  +=    1203.897 * cos(4.02952218586  + 67.91186905000 *T)
    sumY0  +=    1146.840 * cos(0.85085247156  + 296.40126633610 *T)
    sumY0  +=    1000.425 * cos(4.82141247296  + 8.32057233081 *T)
    sumY0  +=    1191.037 * cos(4.01680230124  + 50.64639366260 *T)
    sumY0  +=    1128.718 * cos(5.51338814083  + 230.32075334190 *T)
    sumY0  +=     930.244 * cos(5.33330412293  + 276.79117625789 *T)
    sumY0  +=     857.081 * cos(4.49010094287  + 68.43313391180 *T)
    sumY0  +=     912.408 * cos(0.12283731726  + 149.69514003350 *T)
    sumY0  +=    1170.925 * cos(0.45779532770  + 110.45013870291 *T)
    sumY0  +=     857.595 * cos(6.21726257263  + 90.00326957780 *T)
    sumY0  +=     989.091 * cos(1.04565979808  + 35.66854013560 *T)
    sumY0  +=    1054.223 * cos(3.96633755369  + 14.25846316400 *T)
    sumY0  +=     959.939 * cos(4.47203405039  + 147.83490694279 *T)
    sumY0  +=     957.455 * cos(6.25324374775  + 64.97655352819 *T)
    sumY0  +=     909.789 * cos(3.15839894712  + 149.91888920271 *T)
    sumY0  +=     943.313 * cos(0.06263523023  + 53.93085326429 *T)
    sumY0  +=     891.565 * cos(4.59544376260  + 107.22056085540 *T)
    sumY0  +=     775.480 * cos(3.57147759051  + 70.60562782070 *T)
    sumY0  +=     861.039 * cos(4.78256260255  + 143.86912378490 *T)
    sumY0  +=     851.667 * cos(0.50386900265  + 134.82916109110 *T)
    sumY0  +=     842.290 * cos(4.31374571284  + 3.68833577960 *T)
    sumY0  +=     893.222 * cos(1.81802815173  + 23.81969071961 *T)
    sumY0  +=     790.626 * cos(2.40048938853  + 7.83293736379 *T)
    sumY0  +=     683.506 * cos(0.23627850921  + 81.61769818981 *T)
    sumY0  +=     789.879 * cos(3.10451486371  + 35.40790770471 *T)
    sumY0  +=     669.882 * cos(3.13601045204  + 50.96190784770 *T)
    sumY0  +=     853.775 * cos(0.49301638444  + 200.52510498229 *T)
    sumY0  +=     698.863 * cos(4.93731616945  + 125.74350641500 *T)
    sumY0  +=     814.817 * cos(1.52199016837  + 143.38148881789 *T)
    sumY0  +=     862.909 * cos(3.46019641204  + 162.33718759030 *T)
    sumY0  +=     845.548 * cos(1.49366305890  + 203.49405039889 *T)
    sumY0  +=     797.191 * cos(0.65007259191  + 99.40443843901 *T)
    sumY0  +=     766.426 * cos(0.89351379369  + 68.80564092731 *T)
    sumY0  +=     609.673 * cos(0.23918047727  + 24.13520490470 *T)
    sumY0  +=     654.550 * cos(5.45801088086  + 109.70187130500 *T)
    sumY0  +=     783.834 * cos(0.58870097840  + 218.16308738520 *T)
    sumY0  +=     744.901 * cos(2.79731067673  + 760.01171843649 *T)
    sumY0  +=     643.611 * cos(5.98397644447  + 89.04006173130 *T)
    sumY0  +=     525.538 * cos(1.48456394611  + 128.71245183160 *T)
    sumY0  +=     703.451 * cos(3.78477189218  + 81.24519117430 *T)
    sumY0  +=     590.282 * cos(1.77381181137  + 69.60879007940 *T)
    sumY0  +=     528.305 * cos(0.71004370185  + 146.62562088040 *T)
    sumY0  +=     621.414 * cos(2.39650007607  + 68.59989025060 *T)
    sumY0  +=     677.104 * cos(0.95388610928  + 56.37853381910 *T)
    sumY0  +=     697.582 * cos(4.50208217501  + 617.56206830270 *T)
    sumY0  +=     597.125 * cos(0.63573583143  + 182.00215942271 *T)
    sumY0  +=     689.780 * cos(2.46027600250  + 2.72512793310 *T)
    sumY0  +=     578.345 * cos(0.91036470086  + 28.81562556571 *T)
    sumY0  +=     587.454 * cos(2.33732577632  + 152.25469517290 *T)
    sumY0  +=     670.562 * cos(4.63198986719  + 5.97595764000 *T)
    sumY0  +=     596.736 * cos(1.80927442800  + 116.18227885939 *T)
    sumY0  +=     480.108 * cos(5.71326814363  + 415.30867312860 *T)
    sumY0  +=     505.417 * cos(0.96244787568  + 157.88376946540 *T)
    sumY0  +=     551.335 * cos(5.85642311971  + 261.23717611490 *T)
    sumY0  +=     584.839 * cos(4.61840602682  + 258.26823069831 *T)
    sumY0  +=     571.647 * cos(2.69246515899  + 194.89603068979 *T)
    sumY0  +=     473.414 * cos(2.89922727404  + 134.34152612410 *T)
    sumY0  +=     551.129 * cos(0.91615593626  + 59.55992755679 *T)
    sumY0  +=     518.895 * cos(0.73540791389  + 352.06040979221 *T)
    sumY0  +=     444.205 * cos(5.32877320177  + 329.48137429739 *T)
    sumY0  +=     531.507 * cos(4.45049996181  + 54.41848823131 *T)
    sumY0  +=     449.312 * cos(2.60522543313  + 542.78046973539 *T)
    sumY0  +=     486.453 * cos(5.04367743163  + 73.65281792710 *T)
    sumY0  +=     437.062 * cos(6.01172304160  + 1589.31671276730 *T)
    sumY0  +=     488.485 * cos(3.97669319472  + 9.31741007210 *T)
    sumY0  +=     431.252 * cos(6.26915854927  + 41.40068029210 *T)
    sumY0  +=     435.419 * cos(2.19426156558  + 372.66733761171 *T)
    sumY0  +=     484.495 * cos(5.60519908806  + 209.12312469139 *T)
    sumY0  +=     418.899 * cos(0.40333301614  + 95.63234387030 *T)
    sumY0  +=     514.102 * cos(3.83741418050  + 116.12739710521 *T)
    sumY0  +=     388.869 * cos(4.39552544690  + 152.98840835580 *T)
    sumY0  +=     361.158 * cos(0.85200523076  + 120.11443212250 *T)
    sumY0  +=     371.870 * cos(4.10914188330  + 144.39038864671 *T)
    sumY0  +=     339.152 * cos(5.35894945718  + 46.45360796860 *T)
    sumY0  +=     373.018 * cos(1.90146663534  + 73.42906875790 *T)
    sumY0  +=     382.930 * cos(3.64215083700  + 114.38229196601 *T)
    sumY0  +=     426.834 * cos(0.89896226407  + 141.73025977080 *T)
    sumY0  +=     396.238 * cos(0.74278278155  + 98.91680347199 *T)
    sumY0  +=     396.936 * cos(1.69553808184  + 180.03005174739 *T)
    sumY0  +=     318.113 * cos(4.93841214947  + 33.38091827520 *T)
    sumY0  +=     325.989 * cos(2.61881558482  + 228.52076644850 *T)
    sumY0  +=     435.822 * cos(3.74482431590  + 37.88921815429 *T)
    sumY0  +=     392.309 * cos(3.12268597468  + 155.53915477459 *T)
    sumY0  +=     353.063 * cos(4.66591593809  + 215.02738562980 *T)
    sumY0  +=     410.664 * cos(2.63874768833  + 490.57790666290 *T)
    sumY0  +=     325.464 * cos(4.97938687953  + 7.66618102501 *T)
    sumY0  +=     299.890 * cos(5.32358213894  + 45.33283355520 *T)
    sumY0  +=     319.736 * cos(3.43237988274  + 30.46685461280 *T)
    sumY0  +=     278.968 * cos(3.94608550399  + 6.34846465550 *T)
    sumY0  +=     312.880 * cos(1.42604949071  + 419.72846135871 *T)
    sumY0  +=     386.565 * cos(0.75656271428  + 254.81165031470 *T)
    sumY0  +=     311.788 * cos(0.80567112359  + 147.35933406330 *T)
    sumY0  +=     300.229 * cos(5.03792883993  + 0.71939036300 *T)
    sumY0  +=     337.396 * cos(3.60674120968  + 150.77022246460 *T)
    sumY0  +=     253.789 * cos(6.19652620996  + 19.36627259471 *T)
    sumY0  +=     308.097 * cos(4.06400002760  + 640.14110379750 *T)
    sumY0  +=     252.403 * cos(5.01154326423  + 150.32827947990 *T)
    sumY0  +=     238.679 * cos(0.99851487185  + 553.82616999929 *T)
    sumY0  +=     266.427 * cos(2.58806015071  + 105.24845318010 *T)
    sumY0  +=     278.375 * cos(5.88688068426  + 29.44876501211 *T)
    sumY0  +=     306.043 * cos(0.18265767681  + 6283.31966747490 *T)
    sumY0  +=     265.739 * cos(1.06481607166  + 28.06735816780 *T)
    sumY0  +=     280.569 * cos(2.87788043615  + 126.23114138200 *T)
    sumY0  +=     300.771 * cos(0.42899364062  + 984.84414910540 *T)
    sumY0  +=     229.063 * cos(4.33041573821  + 33.92343499640 *T)
    sumY0  +=     226.799 * cos(3.78600389427  + 121.23520653590 *T)
    sumY0  +=     238.148 * cos(5.87500222549  + 2.20386307129 *T)
    sumY0  +=     198.420 * cos(3.19038901730  + 69.91760662011 *T)
    sumY0  +=     220.174 * cos(5.18139057822  + 44.31474395451 *T)
    sumY0  +=     220.857 * cos(3.44201572334  + 47.93808067690 *T)
    sumY0  +=     211.834 * cos(3.34235859210  + 0.76508234530 *T)
    sumY0  +=     257.979 * cos(0.30426624589  + 1289.70268353110 *T)
    sumY0  +=     240.510 * cos(6.20484819466  + 756.07956517339 *T)
    sumY0  +=     213.870 * cos(1.08601616914  + 316.63568714010 *T)
    sumY0  +=     231.462 * cos(2.77988039912  + 46.71424039951 *T)
    sumY0  +=     232.197 * cos(2.17648415571  + 339.04260185300 *T)
    sumY0  +=     214.419 * cos(1.18021908307  + 189.42452296421 *T)
    sumY0  +=     219.049 * cos(5.43837709719  + 213.05527795450 *T)
    sumY0  +=     241.635 * cos(3.54595332721  + 102.84895673509 *T)
    sumY0  +=     198.526 * cos(5.03106436317  + 61.01077037031 *T)
    sumY0  +=     237.145 * cos(3.14541452090  + 342.01154726960 *T)
    sumY0  +=     173.496 * cos(6.02474649172  + 116.66991382640 *T)
    sumY0  +=     194.828 * cos(0.84925657122  + 218.65072235220 *T)
    sumY0  +=     175.785 * cos(1.53473275730  + 57.49930823250 *T)
    sumY0  +=     167.277 * cos(5.61665487201  + 30.30009827400 *T)
    sumY0  +=     173.400 * cos(4.28804409833  + 681.29796660610 *T)
    sumY0  +=     214.483 * cos(0.91051981390  + 135.79236893760 *T)
    sumY0  +=     205.110 * cos(1.86858550999  + 216.67861467689 *T)
    sumY0  +=     168.458 * cos(1.51585792778  + 80.13322548150 *T)
    sumY0  +=     164.308 * cos(0.70973234637  + 467.99887116810 *T)
    sumY0  +=     203.489 * cos(5.82937379550  + 148.84380677161 *T)
    sumY0  +=     177.178 * cos(5.36314195972  + 103.59722413300 *T)
    sumY0  +=     165.274 * cos(5.36941056367  + 155.22364058950 *T)
    sumY0  +=     185.802 * cos(3.99560320824  + 74.13164417351 *T)
    sumY0  +=     185.802 * cos(0.68379922776  + 75.91918792810 *T)
    sumY0  +=     187.376 * cos(4.62809176973  + 149.28574975631 *T)
    sumY0  +=     153.285 * cos(3.45417848657  + 17.28229033479 *T)
    sumY0  +=     161.572 * cos(0.32247161825  + 264.26100328570 *T)
    sumY0  +=     185.820 * cos(5.76008718824  + 76.39801417450 *T)
    sumY0  +=     181.668 * cos(3.10973418373  + 455.15318401080 *T)
    sumY0  +=     175.514 * cos(4.41099766534  + 333.41352756050 *T)
    sumY0  +=     157.908 * cos(4.13654987265  + 82.44777959230 *T)
    sumY0  +=     169.475 * cos(5.91337395999  + 454.66554904380 *T)
    sumY0  +=     196.145 * cos(0.92388864115  + 291.46021324420 *T)
    sumY0  +=     177.970 * cos(0.39528529238  + 685.23011986920 *T)
    sumY0  +=     137.738 * cos(3.65235776206  + 448.93340888730 *T)
    sumY0  +=     152.620 * cos(2.03319901912  + 234.14984074100 *T)
    sumY0  +=     137.407 * cos(0.49611049910  + 16.91859203990 *T)
    sumY0  +=     150.365 * cos(4.65428057213  + 294.91679362781 *T)
    sumY0  +=     131.039 * cos(2.42507712732  + 156.39929675710 *T)
    sumY0  +=     180.038 * cos(2.01273291483  + 103.33659170210 *T)
    sumY0  +=     129.725 * cos(4.56525647923  + 76.72233708021 *T)
    sumY0  +=     142.122 * cos(5.78256645364  + 224.10097821840 *T)
    sumY0  +=     128.982 * cos(4.86634473406  + 130.68455950690 *T)
    sumY0  +=     146.399 * cos(3.06416499534  + 334.53430197390 *T)
    sumY0  +=     177.104 * cos(1.60811125139  + 10213.52936369450 *T)
    sumY0  +=     129.816 * cos(3.14582203581  + 92.55152386910 *T)
    sumY0  +=     171.451 * cos(3.01469327848  + 156.02678974161 *T)
    sumY0  +=     155.457 * cos(1.02026020385  + 347.64062156210 *T)
    sumY0  +=     134.222 * cos(0.55759952973  + 67.60305250931 *T)
    sumY0  +=     167.287 * cos(0.73206039815  + 80.75755620730 *T)
    sumY0  +=     126.618 * cos(2.97362091056  + 55.90296093960 *T)
    sumY0  +=     127.892 * cos(5.91651149188  + 61.69209392650 *T)
    sumY0  +=     146.027 * cos(4.63620735731  + 174.18603700630 *T)
    sumY0  +=     143.517 * cos(4.72419244572  + 23.87457247379 *T)
    sumY0  +=     139.674 * cos(4.51014304795  + 286.84003878050 *T)
    sumY0  +=     114.667 * cos(5.20627341431  + 13.08950464080 *T)
    sumY0  +=     114.401 * cos(2.63083933234  + 692.34366686999 *T)
    sumY0  +=     139.103 * cos(5.06218842316  + 120.60206708950 *T)
    sumY0  +=     124.306 * cos(4.39000727764  + 111.18634401329 *T)
    sumY0  +=     121.491 * cos(4.89143182138  + 79.69128249680 *T)
    sumY0  +=     106.720 * cos(5.03982163712  + 67.16110952460 *T)
    sumY0  +=     124.445 * cos(2.77221022491  + 141.89701610960 *T)
    sumY0  +=     150.445 * cos(4.77055522751  + 77.50672650040 *T)
    sumY0  +=     128.748 * cos(3.45272600965  + 254.69977573010 *T)
    sumY0  +=     135.145 * cos(1.15443620224  + 171.89841514590 *T)
    sumY0  +=     125.741 * cos(3.85934191891  + 98.60128928690 *T)
    sumY0  +=     106.860 * cos(0.10501468915  + 375.63628302830 *T)
    sumY0  +=     128.791 * cos(1.53820122080  + 155.74490545130 *T)
    sumY0  +=     104.092 * cos(0.09658958912  + 54.57854692570 *T)
    sumY0  +=     142.443 * cos(4.61127299852  + 203.98168536590 *T)
    sumY0  +=     109.153 * cos(1.74773683236  + 1.20702533000 *T)
    sumY0  +=     111.788 * cos(4.78068421543  + 272.85902299479 *T)
    sumY0  +=     125.849 * cos(2.31032327443  + 182.99899716400 *T)
    sumY0  +=     128.119 * cos(5.96640498766  + 256.78375799000 *T)
    sumY0  +=     101.034 * cos(0.35709921098  + 363.10611005610 *T)
    sumY0  +=     104.496 * cos(1.27067787791  + 131.16013238639 *T)
    sumY0  +=     135.112 * cos(4.32885525625  + 629.09540353360 *T)
    sumY0  +=      96.070 * cos(2.53083839499  + 258.63192899320 *T)
    sumY0  +=     102.131 * cos(5.03860775630  + 88.35873817511 *T)
    sumY0  +=     101.524 * cos(1.01685797883  + 51.44954281471 *T)
    sumY0  +=     114.745 * cos(3.99231633115  + 633.02755679670 *T)
    sumY0  +=     105.009 * cos(5.80120897394  + 105.73608814710 *T)
    sumY0  +=     106.158 * cos(2.64028517894  + 302.33915716930 *T)
    sumY0  +=     114.559 * cos(1.22538898026  + 604.71638114540 *T)
    sumY0  +=      97.687 * cos(1.51458558660  + 77.03115362090 *T)
    sumY0  +=      92.981 * cos(1.80318597364  + 94.14787116200 *T)
    sumY0  +=      92.956 * cos(1.47669166228  + 43.53284666180 *T)
    sumY0  +=      99.120 * cos(5.09958883998  + 149.54638218720 *T)
    sumY0  +=     116.555 * cos(5.44320124755  + 85.58348134769 *T)
    sumY0  +=      89.934 * cos(2.95642515081  + 149.85519872791 *T)
    sumY0  +=      89.708 * cos(3.13236064850  + 217.47506618460 *T)
    sumY0  +=     111.015 * cos(2.04689231490  + 253.32717760639 *T)
    sumY0  +=      92.676 * cos(2.85459016512  + 267.22994870229 *T)
    sumY0  +=     111.165 * cos(2.03014798464  + 19.88753745650 *T)
    sumY0  +=      89.583 * cos(1.04552994794  + 295.43805848960 *T)
    sumY0  +=      89.934 * cos(0.36256511993  + 149.75883050830 *T)
    sumY0  +=      89.829 * cos(1.60334868993  + 529.44714761109 *T)
    sumY0  +=     104.822 * cos(4.71885222665  + 433.95555536030 *T)
    sumY0  +=      97.705 * cos(4.61026387361  + 1052.51220067191 *T)
    sumY0  +=     107.796 * cos(5.92808397447  + 100.62827871640 *T)
    sumY0  +=      87.921 * cos(4.32670648553  + 74.39227660440 *T)
    sumY0  +=      87.921 * cos(0.35269595047  + 75.65855549721 *T)
    sumY0  +=      86.241 * cos(4.53063306110  + 76.66994745350 *T)
    sumY0  +=     115.960 * cos(4.20154334971  + 1214.92108496379 *T)
    sumY0  +=     112.324 * cos(5.22657318772  + 229.04203131030 *T)
    sumY0  +=      92.049 * cos(1.25111090139  + 6.86972951729 *T)
    sumY0  +=     106.860 * cos(1.56893886591  + 81.92651473050 *T)
    sumY0  +=      82.022 * cos(5.83269737454  + 82.88972257700 *T)
    sumY0  +=      83.480 * cos(2.97873123693  + 149.96707331250 *T)
    sumY0  +=      81.598 * cos(0.00771145613  + 146.35043423450 *T)
    sumY0  +=      91.518 * cos(2.26786409002  + 150.80385235940 *T)
    sumY0  +=      78.950 * cos(5.23329492514  + 4.86399194720 *T)
    sumY0  +=     107.950 * cos(4.72998464814  + 1162.71852189130 *T)
    sumY0  +=      80.076 * cos(0.46711196257  + 73.38088464811 *T)
    sumY0  +=      93.643 * cos(0.84579418099  + 99.14380600810 *T)
    sumY0  +=      99.120 * cos(4.50258673794  + 150.06764704901 *T)
    sumY0  +=      82.453 * cos(5.81189856956  + 7.14491616321 *T)
    sumY0  +=      76.677 * cos(5.83936897150  + 73.80157577341 *T)
    sumY0  +=      75.891 * cos(6.10092456248  + 219.17198721400 *T)
    sumY0  +=      74.887 * cos(5.67794366791  + 6.83609962250 *T)
    sumY0  +=     102.486 * cos(3.12247878711  + 210.57396750490 *T)
    sumY0  +=     100.859 * cos(2.53732811231  + 219.64756009350 *T)
    sumY0  +=      83.480 * cos(0.34025903382  + 149.64695592370 *T)
    sumY0  +=      83.970 * cos(3.28551184950  + 371.18286490340 *T)
    sumY0  +=      77.201 * cos(0.02474873694  + 26.26737127440 *T)
    sumY0  +=      73.643 * cos(4.58220709295  + 95.47228517591 *T)
    sumY0  +=     102.191 * cos(1.09075434358  + 328.10877617370 *T)
    sumY0  +=      80.117 * cos(5.98791220974  + 142.38465107660 *T)
    sumY0  +=      80.972 * cos(3.62371604241  + 67.11541754230 *T)
    sumY0  +=      71.543 * cos(4.72957453846  + 189.47940471839 *T)
    sumY0  +=      82.381 * cos(4.20534144511  + 58.35064149440 *T)
    sumY0  +=      90.613 * cos(1.46760950491  + 17.02165790390 *T)
    sumY0  +=      82.892 * cos(1.64329175266  + 175.40987728371 *T)
    sumY0  +=      69.078 * cos(5.39603237759  + 119.75073382760 *T)
    sumY0  +=      77.925 * cos(4.56151662815  + 191.45151239371 *T)
    sumY0  +=      78.846 * cos(4.72561034763  + 262.72164882321 *T)
    sumY0  +=      87.497 * cos(1.63219490623  + 894.59706204410 *T)
    sumY0  +=      70.199 * cos(5.98105339890  + 819.81546347680 *T)
    sumY0  +=      68.577 * cos(4.99804003301  + 15.79781762650 *T)
    sumY0  +=      76.274 * cos(1.85161362500  + 180.51768671441 *T)
    sumY0  +=      67.854 * cos(2.38798386084  + 606.51636803880 *T)
    sumY0  +=      77.388 * cos(3.19411072289  + 310.41591201661 *T)
    sumY0  +=      65.829 * cos(4.32001805221  + 75.82856520290 *T)
    sumY0  +=      65.829 * cos(0.35938438379  + 74.22226689870 *T)
    sumY0  +=      78.426 * cos(3.84145278450  + 50.90702609350 *T)
    sumY0  +=      71.406 * cos(1.51626708317  + 12.28635548869 *T)
    sumY0  +=      77.423 * cos(6.17877210501  + 70.29011363560 *T)
    sumY0  +=      69.368 * cos(2.39711158006  + 1066.73929467350 *T)
    sumY0  +=      80.391 * cos(3.12585627922  + 16.70614371880 *T)
    sumY0  +=      85.613 * cos(3.02934752525  + 173.92540457540 *T)
    sumY0  +=      72.675 * cos(2.57315009777  + 44.48150029329 *T)
    sumY0  +=      84.563 * cos(5.09853843122  + 55.41532597260 *T)
    sumY0  +=      62.255 * cos(5.03095032865  + 209.61075965840 *T)
    sumY0  +=      84.357 * cos(5.06410076104  + 331.56535655731 *T)
    sumY0  +=      70.498 * cos(2.80103845095  + 76.62176334371 *T)
    sumY0  +=      79.016 * cos(3.86114745433  + 477.56009872370 *T)
    sumY0  +=      82.673 * cos(4.81351709872  + 480.52904414029 *T)
    sumY0  +=      82.513 * cos(1.59855523562  + 192.44835013500 *T)
    sumY0  +=      80.452 * cos(0.84754362206  + 333.04982926560 *T)
    sumY0  +=      58.904 * cos(5.18941699981  + 24.62283987170 *T)
    sumY0  +=      72.407 * cos(4.52650506832  + 909.57491557110 *T)
    sumY0  +=      76.612 * cos(2.62396349451  + 82.72966388260 *T)
    sumY0  +=      75.510 * cos(2.06744522079  + 67.32116821901 *T)
    sumY0  +=      75.555 * cos(2.66744846800  + 15.74293587230 *T)
    sumY0  +=      70.086 * cos(1.28356301862  + 167.96626188280 *T)
    sumY0  +=      77.989 * cos(1.70294491984  + 70.35954960481 *T)
    sumY0  +=      66.531 * cos(2.86989915171  + 61.53203523210 *T)
    sumY0  +=      64.336 * cos(2.95658893043  + 178.54557903910 *T)
    sumY0  +=      56.673 * cos(3.19941121229  + 106.51798543980 *T)
    sumY0  +=      77.640 * cos(1.49248929330  + 71.40207932840 *T)
    sumY0  +=      62.431 * cos(3.15117993734  + 25.35904518210 *T)
    sumY0  +=      66.444 * cos(6.24350199754  + 125.42799222991 *T)
    sumY0  +=      56.738 * cos(1.26277611682  + 67.42423408300 *T)
    sumY0  +=      55.917 * cos(2.89741852417  + 949.41942645330 *T)
    sumY0  +=      76.032 * cos(0.05943871300  + 60.52313540329 *T)
    sumY0  +=      59.000 * cos(1.97279336373  + 402.77850015639 *T)
    sumY0  +=      73.138 * cos(0.69242086169  + 26.84351789039 *T)
    sumY0  +=      54.636 * cos(6.18572230294  + 72.10465474401 *T)
    sumY0  +=      64.610 * cos(6.10083958494  + 471.93102443120 *T)
    sumY0  +=      53.942 * cos(6.14677939181  + 117.66675156770 *T)
    sumY0  +=      54.371 * cos(2.13625507265  + 341.01470952831 *T)
    sumY0  +=      54.836 * cos(4.43105022688  + 3.42521122121 *T)
    sumY0  +=      72.732 * cos(3.18629698674  + 78.64875277320 *T)
    sumY0  +=      69.223 * cos(4.43578262380  + 8.10812400971 *T)
    sumY0  +=      70.778 * cos(2.29608168481  + 237.43430034270 *T)
    sumY0  +=      53.991 * cos(3.01146704045  + 149.05625509271 *T)
    sumY0  +=      51.866 * cos(4.77686544357  + 77.94617735760 *T)
    sumY0  +=      60.750 * cos(4.76881715073  + 79.76071846601 *T)
    sumY0  +=      69.434 * cos(2.72131865841  + 256.29612302300 *T)
    sumY0  +=      52.587 * cos(4.28485895391  + 830.86116374070 *T)
    sumY0  +=      56.127 * cos(2.54058660944  + 65.62424718960 *T)
    sumY0  +=      53.925 * cos(0.48058717157  + 13.73719830221 *T)
    sumY0  +=      50.216 * cos(3.59581882696  + 195.38366565681 *T)
    sumY0  +=      57.142 * cos(0.50054596978  + 31.89644556690 *T)
    sumY0  +=      55.249 * cos(5.97289951260  + 75.39792306630 *T)
    sumY0  +=      55.249 * cos(4.98968823058  + 74.65290903530 *T)
    sumY0  +=      49.301 * cos(5.39744243512  + 76.24925632820 *T)
    sumY0  +=      60.539 * cos(2.19999973905  + 289.97574053589 *T)
    sumY0  +=      64.921 * cos(1.85315970147  + 846.32665223470 *T)
    sumY0  +=      59.034 * cos(2.15240427373  + 27.33115285740 *T)
    sumY0  +=      51.079 * cos(6.13730250517  + 0.35569206810 *T)
    sumY0  +=      51.079 * cos(3.28747545560  + 0.13194289890 *T)
    sumY0  +=      56.906 * cos(0.02764820630  + 143.21473247911 *T)
    sumY0  +=      51.926 * cos(3.14825419321  + 185.23173727020 *T)
    sumY0  +=      48.626 * cos(1.04414880109  + 49.16192095430 *T)
    sumY0  +=      51.388 * cos(4.38845040975  + 42.88515300039 *T)
    sumY0  +=      64.661 * cos(4.35150637725  + 910.06255053810 *T)
    sumY0  +=      48.885 * cos(0.71247161940  + 91.70019060721 *T)
    sumY0  +=      50.130 * cos(1.11056344933  + 102.11275142471 *T)
    sumY0  +=      47.590 * cos(2.58646137213  + 224.47673860080 *T)
    sumY0  +=      57.461 * cos(2.71055250906  + 303.61787920090 *T)
    sumY0  +=      50.603 * cos(0.01398135868  + 35.18090516859 *T)
    sumY0  +=      48.745 * cos(0.99720060201  + 201.01273994931 *T)
    sumY0  +=      58.382 * cos(2.71650750224  + 486.15811843280 *T)
    sumY0  +=      49.531 * cos(0.36947250466  + 44.96913526031 *T)
    sumY0  +=      54.462 * cos(5.91602724751  + 237.11878615761 *T)
    sumY0  +=      54.466 * cos(2.50120140695  + 75.67980735660 *T)
    sumY0  +=      54.466 * cos(2.17820102905  + 74.37102474501 *T)
    sumY0  +=      60.438 * cos(0.32621216428  + 266.23311096100 *T)
    sumY0  +=      48.443 * cos(5.57062182834  + 178.11819026941 *T)
    sumY0  +=      55.698 * cos(0.52867804605  + 278.76328393321 *T)
    sumY0  +=      44.294 * cos(3.52677749838  + 57.75994066341 *T)
    sumY0  +=      48.861 * cos(5.20847212940  + 122.71967924421 *T)
    sumY0  +=      46.778 * cos(2.82048605921  + 14.47091148511 *T)
    sumY0  +=      52.405 * cos(0.06300374474  + 58.98378094080 *T)
    sumY0  +=      47.669 * cos(2.13908929429  + 84.42658491201 *T)
    sumY0  +=      59.409 * cos(2.64417368509  + 117.61186981350 *T)
    sumY0  +=      42.464 * cos(0.34761416357  + 150.55777414350 *T)
    sumY0  +=      42.527 * cos(3.59846310307  + 69.39634175830 *T)
    sumY0  +=      44.136 * cos(2.47960023448  + 554.31380496631 *T)
    sumY0  +=      56.899 * cos(0.53819926759  + 25.30416342790 *T)
    sumY0  +=      53.440 * cos(6.17081078665  + 275.30670354960 *T)
    sumY0  +=      48.371 * cos(0.35760092436  + 56.01483552421 *T)
    sumY0  +=      51.224 * cos(5.62752727167  + 1364.48428209840 *T)
    sumY0  +=      44.203 * cos(4.50403311651  + 70.83263035680 *T)
    sumY0  +=      41.205 * cos(6.18207911178  + 124.25903370670 *T)
    sumY0  +=      41.202 * cos(4.39907320318  + 49.47743513939 *T)
    sumY0  +=      53.218 * cos(3.54700024950  + 7.46043034830 *T)
    sumY0  +=      39.879 * cos(4.96873405562  + 48.51422729289 *T)
    sumY0  +=      39.626 * cos(1.62185831078  + 72.73779419041 *T)
    sumY0  +=      42.269 * cos(3.27019607433  + 232.66536803271 *T)
    sumY0  +=      39.724 * cos(2.20111532844  + 88.51879686950 *T)
    sumY0  +=      54.178 * cos(1.25790703261  + 364.75733910320 *T)
    sumY0  +=      47.398 * cos(5.34395839525  + 74.17408278890 *T)
    sumY0  +=      47.398 * cos(5.61862934792  + 75.87674931271 *T)
    sumY0  +=      45.255 * cos(5.10853887295  + 393.21727260080 *T)
    sumY0  +=      41.021 * cos(0.17536932736  + 79.21820174480 *T)
    sumY0  +=      40.961 * cos(3.79678095557  + 306.58682461749 *T)
    sumY0  +=      48.986 * cos(2.66938282989  + 198.07742442750 *T)
    sumY0  +=      42.523 * cos(6.01169442061  + 22.39009976550 *T)
    sumY0  +=      37.414 * cos(4.21576362460  + 397.14942586389 *T)
    sumY0  +=      43.272 * cos(5.65327724529  + 28.32799059869 *T)
    sumY0  +=      47.477 * cos(5.62213666350  + 224.70048777000 *T)
    sumY0  +=      47.665 * cos(4.06141457601  + 73.32849502140 *T)
    sumY0  +=      40.139 * cos(2.45661194468  + 164.78486814511 *T)
    sumY0  +=      37.770 * cos(5.10261043077  + 91.48774228611 *T)
    sumY0  +=      37.294 * cos(3.45883659652  + 275.79433851660 *T)
    sumY0  +=      39.829 * cos(4.64468416590  + 28.98238190449 *T)
    sumY0  +=      42.462 * cos(3.17137527571  + 25.61967761299 *T)
    sumY0  +=      38.602 * cos(3.30935971278  + 91.80325647120 *T)
    sumY0  +=      49.996 * cos(2.95472165240  + 52.13086637091 *T)
    sumY0  +=      37.120 * cos(1.73214539662  + 167.33312243640 *T)
    sumY0  +=      35.657 * cos(3.04424405945  + 77.31303791120 *T)
    sumY0  +=      46.968 * cos(0.94955700094  + 836.27778971210 *T)
    sumY0  +=      36.770 * cos(4.66251230729  + 123.78346082721 *T)
    sumY0  +=      45.438 * cos(1.78673921122  + 199.04063227400 *T)
    sumY0  +=      48.093 * cos(4.63308338383  + 3340.85624418330 *T)
    sumY0  +=      34.133 * cos(3.71275939849  + 154.91482404881 *T)
    sumY0  +=      40.987 * cos(0.29444385242  + 411.37651986550 *T)
    sumY0  +=      37.167 * cos(4.49302606958  + 227.03629374021 *T)
    sumY0  +=      35.365 * cos(0.64363628623  + 183.85033042589 *T)
    sumY0  +=      34.349 * cos(5.93570230021  + 58.56308981550 *T)
    sumY0  +=      44.963 * cos(6.26050330647  + 833.30884429549 *T)
    sumY0  +=      33.485 * cos(0.14708707923  + 227.77000692311 *T)
    sumY0  +=      32.838 * cos(4.50312203340  + 69.12115511239 *T)
    sumY0  +=      37.229 * cos(3.78657903383  + 103.76398047180 *T)
    sumY0  +=      33.396 * cos(1.12817397300  + 92.29089143820 *T)
    sumY0  +=      43.914 * cos(6.11902443100  + 293.43232091950 *T)
    sumY0  +=      44.941 * cos(2.90130423263  + 292.94468595250 *T)
    sumY0  +=      37.327 * cos(4.64523467597  + 0.55933166859 *T)
    sumY0  +=      32.513 * cos(4.78887619836  + 136.31363379940 *T)
    sumY0  +=      34.574 * cos(4.74251495333  + 35.92917256650 *T)
    sumY0  +=      41.936 * cos(5.30204278018  + 92.06388890210 *T)
    sumY0  +=      33.257 * cos(5.78096695807  + 248.96763557360 *T)
    sumY0  +=      44.206 * cos(6.02803554950  + 767.61290040431 *T)
    sumY0  +=      33.118 * cos(1.30670048356  + 284.39235822571 *T)
    sumY0  +=      34.173 * cos(4.81499116474  + 133.13224006171 *T)
    sumY0  +=      35.868 * cos(1.58130371706  + 179.91817716280 *T)
    sumY0  +=      37.508 * cos(4.02104775804  + 94.66913602380 *T)
    sumY0  +=      41.561 * cos(2.19963487826  + 163.82166029861 *T)
    sumY0  +=      34.188 * cos(2.03038237166  + 336.01877468221 *T)
    sumY0  +=      33.991 * cos(5.50380136919  + 32.43896228810 *T)
    sumY0  +=      30.436 * cos(4.34224892820  + 31.78457098230 *T)
    sumY0  +=      34.876 * cos(3.13365595903  + 215.19414196860 *T)
    sumY0  +=      37.259 * cos(4.63049994218  + 91.06705116081 *T)
    sumY0  +=      30.353 * cos(4.80546717458  + 32.48714639791 *T)
    sumY0  +=      31.931 * cos(2.00592901309  + 205.94173095370 *T)
    sumY0  +=      30.008 * cos(4.07053097233  + 745.03386490949 *T)
    sumY0  +=      36.558 * cos(4.16073271737  + 45.96597300160 *T)
    sumY0  +=      29.428 * cos(5.60962138278  + 270.16526422410 *T)
    sumY0  +=      36.881 * cos(2.80082144083  + 75.61611688180 *T)
    sumY0  +=      36.881 * cos(1.87858099517  + 74.43471521981 *T)
    sumY0  +=      39.559 * cos(5.86593317582  + 58.87190635621 *T)
    sumY0  +=      38.377 * cos(2.52859641796  + 39.37369086260 *T)
    sumY0  +=      32.329 * cos(0.69441607084  + 80.65449034331 *T)
    sumY0  +=      40.007 * cos(1.71381436107  + 1744.61205005840 *T)
    sumY0  +=      29.388 * cos(5.29139433485  + 530.89799042461 *T)
    sumY0  +=      31.193 * cos(3.77336776305  + 9.15065373330 *T)
    sumY0  +=      34.894 * cos(3.35589064953  + 143.58723949460 *T)
    sumY0  +=      29.512 * cos(1.39390842622  + 958.33296034749 *T)
    sumY0  +=      29.255 * cos(0.84093803376  + 415.79630809560 *T)
    sumY0  +=      29.888 * cos(3.51518270201  + 163.14033674240 *T)
    sumY0  +=      32.864 * cos(4.32066515560  + 94.03599657740 *T)
    sumY0  +=      35.403 * cos(6.01521697408  + 31.73638687251 *T)
    sumY0  +=      36.501 * cos(5.48244229511  + 142.90591593840 *T)
    sumY0  +=      32.416 * cos(2.85592816552  + 114.69780615110 *T)
    sumY0  +=      28.338 * cos(2.18812129530  + 528.97157473160 *T)
    sumY0  +=      31.160 * cos(4.24879587681  + 118.15438653471 *T)
    sumY0  +=      34.324 * cos(4.04959387303  + 57.38743364790 *T)
    sumY0  +=      33.749 * cos(4.26748008527  + 42.78208713641 *T)
    sumY0  +=      34.375 * cos(2.35347296470  + 326.62430346539 *T)
    sumY0  +=      29.984 * cos(4.74520755489  + 90.52453443961 *T)
    sumY0  +=      35.105 * cos(5.45333895324  + 230.80838830890 *T)
    sumY0  +=      31.999 * cos(3.50027972951  + 216.51185833811 *T)
    sumY0  +=      30.853 * cos(0.71214846554  + 8.84183719261 *T)
    sumY0  +=      34.254 * cos(0.33401313814  + 228.83628063359 *T)
    sumY0  +=      27.044 * cos(1.46046017518  + 27.96429230381 *T)
    sumY0  +=      27.412 * cos(3.17327818880  + 221.40721944771 *T)
    sumY0  +=      26.388 * cos(3.20755787232  + 45.82046852221 *T)
    sumY0  +=      32.494 * cos(3.35654630791  + 1033.11455891480 *T)
    sumY0  +=      35.994 * cos(5.77252966919  + 4.49148493169 *T)
    sumY0  +=      29.345 * cos(0.67141990161  + 123.29582586020 *T)
    sumY0  +=      26.034 * cos(1.13088552844  + 78.46993434690 *T)
    sumY0  +=      26.650 * cos(3.32642891132  + 151.50393564750 *T)
    sumY0  +=      31.030 * cos(2.03320495111  + 66.42739634171 *T)
    sumY0  +=      28.238 * cos(0.60698223844  + 222.61650551010 *T)
    sumY0  +=      26.518 * cos(2.84813264424  + 11.40139233200 *T)
    sumY0  +=      26.014 * cos(4.66149750775  + 447.44893617900 *T)
    sumY0  +=      25.836 * cos(3.54055201502  + 68.54500849640 *T)
    sumY0  +=      24.852 * cos(4.13613291119  + 42.99702758500 *T)
    sumY0  +=      33.654 * cos(4.92280929068  + 54.53036281590 *T)
    sumY0  +=      24.448 * cos(2.54413564647  + 269.67762925710 *T)
    sumY0  +=      33.190 * cos(0.33337097209  + 141.94270809190 *T)
    sumY0  +=      24.034 * cos(0.04875875474  + 327.99690158910 *T)
    sumY0  +=      31.856 * cos(1.82932365564  + 1140.13948639650 *T)
    sumY0  +=      26.275 * cos(1.11499599456  + 39.91620758380 *T)
    sumY0  +=      25.887 * cos(2.86658178193  + 2.04380437690 *T)
    sumY0  +=      23.665 * cos(5.24877237951  + 21.84758304429 *T)
    sumY0  +=      25.701 * cos(4.05288243063  + 73.91345035801 *T)
    sumY0  +=      25.701 * cos(0.62652000537  + 76.13738174360 *T)
    sumY0  +=      24.158 * cos(1.81862092102  + 4.20960064140 *T)
    sumY0  +=      32.085 * cos(5.23560050501  + 368.21391948681 *T)
    sumY0  +=      25.732 * cos(0.51306284591  + 3.37951923889 *T)
    sumY0  +=      24.270 * cos(0.61765538541  + 130.19692453990 *T)
    sumY0  +=      23.775 * cos(5.95144171122  + 969.37866061139 *T)
    sumY0  +=      22.911 * cos(6.08737032038  + 117.56368570370 *T)
    sumY0  +=      27.148 * cos(6.08119809462  + 11.17764316280 *T)
    sumY0  +=      31.420 * cos(0.20608322775  + 619.04654101100 *T)
    sumY0  +=      30.627 * cos(5.09783147670  + 91.17892574540 *T)
    sumY0  +=      29.507 * cos(0.35894321075  + 565.35950523021 *T)
    sumY0  +=      25.369 * cos(2.77552012656  + 81.13006322279 *T)
    sumY0  +=      27.172 * cos(2.89332627598  + 25.51661174900 *T)
    sumY0  +=      26.354 * cos(0.09318660287  + 204.97852310719 *T)
    sumY0  +=      22.571 * cos(2.16687242605  + 70.88081446661 *T)
    sumY0  +=      22.904 * cos(1.13841839619  + 81.50582360521 *T)
    sumY0  +=      30.825 * cos(1.83861394951  + 1439.75351563271 *T)
    sumY0  +=      25.426 * cos(1.66774350963  + 133.34468838281 *T)
    sumY0  +=      27.083 * cos(4.09142271592  + 105.08169684130 *T)
    sumY0  +=      23.195 * cos(0.51761268593  + 90.57941619379 *T)
    sumY0  +=      22.062 * cos(0.12945216599  + 189.16389053330 *T)
    sumY0  +=      22.242 * cos(5.02281259808  + 86.18299089930 *T)
    sumY0  +=      23.869 * cos(1.09273091754  + 413.82420042030 *T)
    sumY0  +=      21.820 * cos(3.86588943870  + 34.86539098351 *T)
    sumY0  +=      29.143 * cos(5.55273035507  + 616.07759559439 *T)
    sumY0  +=      26.632 * cos(3.99150253140  + 100.88891114730 *T)
    sumY0  +=      23.677 * cos(3.63905075648  + 541.29599702710 *T)
    sumY0  +=      30.371 * cos(1.42781024655  + 401.40590203270 *T)
    sumY0  +=      23.717 * cos(3.96575925914  + 45.79921666280 *T)
    sumY0  +=      27.663 * cos(4.12303428511  + 405.74744557300 *T)
    sumY0  +=      24.084 * cos(2.70826498302  + 170.41394243761 *T)
    sumY0  +=      25.601 * cos(4.37491173191  + 31.24875190549 *T)
    sumY0  +=      22.288 * cos(0.79514165960  + 108.16251684250 *T)
    sumY0  +=      22.857 * cos(1.72644712161  + 104.23036357940 *T)
    sumY0  +=      28.908 * cos(3.07758978128  + 329.59324888200 *T)
    sumY0  +=      24.430 * cos(4.94090312535  + 49.42255338521 *T)
    sumY0  +=      23.008 * cos(2.06991447825  + 217.16624964391 *T)
    sumY0  +=      20.835 * cos(5.08228935258  + 141.20899490900 *T)
    sumY0  +=      22.817 * cos(5.18839806192  + 145.38722638800 *T)
    sumY0  +=      20.957 * cos(2.51273449478  + 79.17001763500 *T)
    sumY0  +=      27.954 * cos(0.65571622008  + 68.17250148091 *T)
    sumY0  +=      23.022 * cos(1.21699943524  + 162.65270177540 *T)
    sumY0  +=      25.411 * cos(4.19619949189  + 41.88831525910 *T)
    sumY0  +=      25.079 * cos(0.62989809784  + 92.66339845371 *T)
    sumY0  +=      22.779 * cos(3.95591659633  + 107.74182571721 *T)
    sumY0  +=      25.092 * cos(0.58786621007  + 87.87110320809 *T)
    sumY0  +=      27.016 * cos(1.15672372488  + 0.99457700890 *T)
    sumY0  +=      20.414 * cos(3.43554412854  + 222.14093263061 *T)
    sumY0  +=      20.185 * cos(6.10485906657  + 144.17794032560 *T)
    sumY0  +=      27.094 * cos(0.96491831662  + 369.69839219510 *T)
    sumY0  +=      19.466 * cos(2.51873134984  + 9.15735137770 *T)
    sumY0  +=      20.822 * cos(1.01932827109  + 72.31710306511 *T)
    sumY0  +=      20.822 * cos(3.66007416491  + 77.73372903651 *T)
    sumY0  +=      23.312 * cos(2.70078350147  + 52.39149880180 *T)
    sumY0  +=      19.238 * cos(1.86487700903  + 145.14114817210 *T)
    sumY0  +=      22.156 * cos(5.02188243292  + 64.73047531231 *T)
    sumY0  +=      19.648 * cos(4.16422779161  + 692.83130183700 *T)
    sumY0  +=      18.748 * cos(1.24105948443  + 289.80898419710 *T)
    sumY0  +=      19.125 * cos(1.53478805624  + 22.65073219640 *T)
    sumY0  +=      24.581 * cos(6.19654419773  + 20.20305164159 *T)
    sumY0  +=      21.617 * cos(3.27489383263  + 251.84270489810 *T)
    sumY0  +=      21.860 * cos(2.01803891279  + 38.11622069041 *T)
    sumY0  +=      21.658 * cos(2.82545321119  + 30.98811947460 *T)
    sumY0  +=      21.231 * cos(0.70274221680  + 55.38169607781 *T)
    sumY0  +=      25.483 * cos(3.07684884560  + 41.08516610701 *T)
    sumY0  +=      21.046 * cos(1.95785166305  + 58.07545484850 *T)
    sumY0  +=      22.854 * cos(6.17096810376  + 167.07249000551 *T)
    sumY0  +=      18.322 * cos(3.37460503465  + 165.84864972810 *T)
    sumY0  +=      18.466 * cos(0.61255161441  + 380.84715832300 *T)
    sumY0  +=      24.739 * cos(4.02369174706  + 81.87833062070 *T)
    sumY0  +=      22.782 * cos(1.55690793162  + 305.58998687620 *T)
    sumY0  +=      18.449 * cos(3.77998100899  + 378.92074262999 *T)
    sumY0  +=      20.493 * cos(1.10085699520  + 148.21066732521 *T)
    sumY0  +=      24.545 * cos(4.93054883789  + 66.46102623650 *T)
    sumY0  +=      20.079 * cos(2.85442801367  + 156.92056161890 *T)
    sumY0  +=      17.465 * cos(3.85472736449  + 312.21589891000 *T)
    sumY0  +=      21.567 * cos(5.83653540547  + 225.55182103190 *T)
    sumY0  +=      20.030 * cos(2.22274854754  + 134.30789622931 *T)
    sumY0  +=      17.391 * cos(2.86052314428  + 73.01967848070 *T)
    sumY0  +=      19.537 * cos(3.96842200227  + 140.41254340130 *T)
    sumY0  +=      22.517 * cos(5.23712494895  + 248.70700314271 *T)
    sumY0  +=      17.589 * cos(5.07166000101  + 303.30236501580 *T)
    sumY0  +=      20.366 * cos(1.99661694162  + 97.65933329981 *T)
    sumY0  +=      22.843 * cos(3.92423636309  + 758.52724572820 *T)
    sumY0  +=      21.264 * cos(0.07672955391  + 330.08088384900 *T)
    sumY0  +=      16.980 * cos(1.16080581329  + 74.56784166380 *T)
    sumY0  +=      16.980 * cos(3.51859662271  + 75.48299043781 *T)
    sumY0  +=      19.715 * cos(0.47343267408  + 531.73476947149 *T)
    sumY0  +=      21.989 * cos(1.18511584463  + 136.78920667889 *T)
    sumY0  +=      21.616 * cos(5.67519606472  + 10.32630990090 *T)
    sumY0  +=      16.752 * cos(2.74060731702  + 19.25439801010 *T)
    sumY0  +=      21.941 * cos(4.44343968337  + 624.67561530350 *T)
    sumY0  +=      16.246 * cos(0.04665700154  + 2119.00767786191 *T)
    sumY0  +=      16.279 * cos(5.73366744278  + 1.12878064019 *T)
    sumY0  +=      16.279 * cos(5.73366744278  + 1.61641560721 *T)
    sumY0  +=      22.328 * cos(1.70802820953  + 97.91996573070 *T)
    sumY0  +=      19.188 * cos(0.23213047381  + 14.31334491820 *T)
    sumY0  +=      22.519 * cos(2.42657916175  + 165.36101476110 *T)
    sumY0  +=      20.956 * cos(3.16492203936  + 330.96584700570 *T)
    sumY0  +=      18.599 * cos(4.26067160583  + 148.43441649440 *T)
    sumY0  +=      18.917 * cos(0.97135259771  + 255.29928528170 *T)
    sumY0  +=      21.299 * cos(6.02100084197  + 95.52046928571 *T)
    sumY0  +=      19.398 * cos(2.01021052352  + 408.19512612779 *T)
    sumY0  +=      21.480 * cos(5.61700303674  + 1669.83045149109 *T)
    sumY0  +=      15.730 * cos(5.68185702607  + 17.88179988640 *T)
    sumY0  +=      15.989 * cos(4.02642086421  + 132.85705341579 *T)
    sumY0  +=      16.213 * cos(2.93808470604  + 5.17280848790 *T)
    sumY0  +=      15.494 * cos(0.54839274859  + 743.23387801611 *T)
    sumY0  +=      15.346 * cos(3.70238533639  + 610.44852130189 *T)
    sumY0  +=      19.338 * cos(2.53133192689  + 363.27286639489 *T)
    sumY0  +=      17.882 * cos(0.09108401185  + 190.90899567250 *T)
    sumY0  +=      16.009 * cos(3.94767389914  + 74.19533464831 *T)
    sumY0  +=      16.009 * cos(0.73172853686  + 75.85549745330 *T)
    sumY0  +=      18.450 * cos(1.65175643591  + 135.57992061650 *T)
    sumY0  +=      15.066 * cos(0.31979989608  + 118.26626111931 *T)
    sumY0  +=      19.811 * cos(4.00951140620  + 375.95179721340 *T)
    sumY0  +=      17.887 * cos(5.90368904573  + 85.32035678930 *T)
    sumY0  +=      15.552 * cos(0.08354188367  + 5.87289177600 *T)
    sumY0  +=      15.552 * cos(3.22513453726  + 5.38525680900 *T)
    sumY0  +=      14.946 * cos(3.08926979326  + 76.46170464930 *T)
    sumY0  +=      14.946 * cos(4.73172529633  + 73.58912745230 *T)
    sumY0  +=      18.660 * cos(0.38012284620  + 107.26874496520 *T)
    sumY0  +=      14.573 * cos(2.12795836718  + 73.49275923270 *T)
    sumY0  +=      14.573 * cos(5.69303672241  + 76.55807286891 *T)
    sumY0  +=      15.019 * cos(5.37079041514  + 64.09159037150 *T)
    sumY0  +=      17.892 * cos(4.39093243602  + 231.80522605020 *T)
    sumY0  +=      15.543 * cos(1.73241272133  + 66.11857980100 *T)
    sumY0  +=      18.436 * cos(2.04970589009  + 549.89401673620 *T)
    sumY0  +=      15.371 * cos(4.16535208222  + 736.12033101530 *T)
    sumY0  +=      17.668 * cos(4.75171890820  + 378.39947776819 *T)
    sumY0  +=      14.688 * cos(5.67350374112  + 17.91542978119 *T)
    sumY0  +=      14.089 * cos(6.04399723949  + 63.86784120231 *T)
    sumY0  +=      15.759 * cos(0.66494464321  + 104.25161543881 *T)
    sumY0  +=      16.874 * cos(0.38165130786  + 84.09900863940 *T)
    sumY0  +=      15.468 * cos(5.57539923988  + 515.70768857651 *T)
    sumY0  +=      17.006 * cos(5.22922131058  + 106.25735300890 *T)
    sumY0  +=      15.683 * cos(3.28984199210  + 46.81730626350 *T)
    sumY0  +=      15.035 * cos(4.57350717142  + 490.09027169589 *T)
    sumY0  +=      13.767 * cos(6.18350808152  + 535.66692273460 *T)
    sumY0  +=      14.034 * cos(2.11722923267  + 281.21096448800 *T)
    sumY0  +=      13.418 * cos(2.06565331098  + 70.02948120471 *T)
    sumY0  +=      13.418 * cos(2.61381320215  + 80.02135089690 *T)
    sumY0  +=      18.602 * cos(3.25562899138  + 366.24181181149 *T)
    sumY0  +=      16.062 * cos(0.58179678004  + 73.28031091161 *T)
    sumY0  +=      16.062 * cos(0.95601300237  + 76.77052119001 *T)
    sumY0  +=      18.697 * cos(2.78388202981  + 130.16329464510 *T)
    sumY0  +=      13.830 * cos(1.11702813589  + 278.27564896620 *T)
    sumY0  +=      16.797 * cos(2.01380545416  + 157.51126244990 *T)
    sumY0  +=      16.797 * cos(2.01380545416  + 157.02362748290 *T)
    sumY0  +=      13.389 * cos(1.94700710476  + 118.62995941420 *T)
    sumY0  +=      14.382 * cos(1.07444330581  + 194.53233239490 *T)
    sumY0  +=      17.090 * cos(3.77525528381  + 60.79832204921 *T)
    sumY0  +=      13.114 * cos(5.75071250253  + 883.55136178020 *T)
    sumY0  +=      16.641 * cos(4.93681909681  + 118.31444522910 *T)
    sumY0  +=      13.074 * cos(2.58966287329  + 12.25272559390 *T)
    sumY0  +=      13.153 * cos(1.19799339989  + 225.10987804721 *T)
    sumY0  +=      13.157 * cos(3.23166717004  + 122.08653979781 *T)
    sumY0  +=      16.504 * cos(5.59782240944  + 285.35556607221 *T)
    sumY0  +=      13.716 * cos(2.96104442245  + 1096.85045721820 *T)
    sumY0  +=      12.748 * cos(4.05495720199  + 18.76676304310 *T)
    sumY0  +=      16.958 * cos(2.72238344622  + 49.00186225990 *T)
    sumY0  +=      14.158 * cos(5.60862153409  + 147.80127704800 *T)
    sumY0  +=      14.666 * cos(5.66230491756  + 0.50694204189 *T)
    sumY0  +=      16.964 * cos(3.07409416856  + 64.94292363340 *T)
    sumY0  +=      16.167 * cos(5.52679984001  + 176.89434999200 *T)
    sumY0  +=      17.434 * cos(5.75265280703  + 381.36842318480 *T)
    sumY0  +=      12.702 * cos(2.52948406017  + 141.45507312490 *T)
    sumY0  +=      12.494 * cos(1.91017999465  + 196.01680510321 *T)
    sumY0  +=      13.527 * cos(4.74412448417  + 73.70100203691 *T)
    sumY0  +=      13.527 * cos(3.07687060542  + 76.34983006470 *T)
    sumY0  +=      14.805 * cos(5.14044638199  + 97.17169833279 *T)
    sumY0  +=      12.844 * cos(2.29225207494  + 322.36782729660 *T)
    sumY0  +=      13.194 * cos(2.71389070536  + 552.34169729099 *T)
    sumY0  +=      12.712 * cos(3.50425259630  + 748.96601817260 *T)
    sumY0  +=      13.959 * cos(4.48592210118  + 60.35637906450 *T)
    sumY0  +=      16.442 * cos(2.74965824817  + 83.62343575990 *T)
    sumY0  +=      12.305 * cos(0.28180529137  + 14.73403604350 *T)
    sumY0  +=      16.319 * cos(1.99654524305  + 703.87700210090 *T)
    sumY0  +=      12.524 * cos(2.98932965871  + 501.62360692681 *T)
    sumY0  +=      13.101 * cos(3.29177766895  + 112.67081672160 *T)
    sumY0  +=      13.451 * cos(4.85214091855  + 196.38050339810 *T)
    sumY0  +=      16.729 * cos(4.42229507842  + 336.59492129820 *T)
    sumY0  +=      15.662 * cos(4.17010855511  + 124.74666867370 *T)
    sumY0  +=      16.361 * cos(3.46994237608  + 853.44019923550 *T)
    sumY0  +=      15.016 * cos(3.81955733385  + 193.41155798150 *T)
    sumY0  +=      13.452 * cos(4.41360988031  + 172.44093186710 *T)
    sumY0  +=      14.832 * cos(0.90688277736  + 115.60613224340 *T)
    sumY0  +=      11.802 * cos(3.34670161742  + 34.28713329129 *T)
    sumY0  +=      11.640 * cos(6.04651347369  + 105.13657859549 *T)
    sumY0  +=      12.330 * cos(2.92269312880  + 83.93225230061 *T)
    sumY0  +=      11.536 * cos(1.30362312663  + 241.36645360580 *T)
    sumY0  +=      11.526 * cos(5.08916626368  + 628.60776856660 *T)
    sumY0  +=      13.571 * cos(3.22260253533  + 205.46615807421 *T)
    sumY0  +=      11.579 * cos(0.67710960978  + 65.57606307981 *T)
    sumY0  +=      11.468 * cos(1.57270808839  + 416.79314583689 *T)
    sumY0  +=      11.443 * cos(2.37400000837  + 20.69068660860 *T)
    sumY0  +=      12.105 * cos(0.14668124624  + 43.84836084689 *T)
    sumY0  +=      13.131 * cos(3.46217186244  + 288.49126782760 *T)
    sumY0  +=      14.050 * cos(4.60806116381  + 323.74923414091 *T)
    sumY0  +=      12.885 * cos(5.21981764477  + 283.90472325870 *T)
    sumY0  +=      15.266 * cos(1.42284221306  + 906.13039727500 *T)
    sumY0  +=      11.090 * cos(2.16462545354  + 56.11540926071 *T)
    sumY0  +=      15.022 * cos(1.57800044960  + 438.05446496220 *T)
    sumY0  +=      11.690 * cos(1.28789549386  + 256.18424843839 *T)
    sumY0  +=      10.863 * cos(1.57144632008  + 230.00523915680 *T)
    sumY0  +=      10.598 * cos(0.35796336114  + 11.60503193250 *T)
    sumY0  +=      10.559 * cos(1.69974127789  + 466.51439845980 *T)
    sumY0  +=      11.670 * cos(4.42259077763  + 308.93143930831 *T)
    sumY0  +=      14.228 * cos(1.61252694327  + 85.10790846821 *T)
    sumY0  +=      14.820 * cos(4.17825476069  + 154.05468206630 *T)
    sumY0  +=      11.614 * cos(0.84927799349  + 140.90017836831 *T)
    sumY0  +=      12.388 * cos(5.09110502089  + 1171.63205578549 *T)
    sumY0  +=      12.580 * cos(0.51522053651  + 114.12165953510 *T)
    sumY0  +=      12.667 * cos(6.01888757437  + 38.85242600079 *T)
    sumY0  +=      12.598 * cos(0.72205960279  + 1819.39364862570 *T)
    sumY0  +=      13.260 * cos(5.77871434104  + 77.71917482150 *T)
    sumY0  +=      14.442 * cos(5.84453911482  + 82.24202891559 *T)
    sumY0  +=      13.709 * cos(5.11783507369  + 109.06873185860 *T)
    sumY0  +=      10.566 * cos(1.35962843525  + 1107.89615748210 *T)
    sumY0  +=      12.906 * cos(1.98542091304  + 757.56403788170 *T)
    sumY0  +=      12.022 * cos(0.39178256727  + 207.63865198310 *T)
    sumY0  +=      11.539 * cos(5.68756950373  + 75.59486502240 *T)
    sumY0  +=      11.539 * cos(5.27501823945  + 74.45596707921 *T)
    sumY0  +=      10.478 * cos(5.55035662837  + 72.09010052900 *T)
    sumY0  +=      13.623 * cos(1.20423210680  + 406.34695512460 *T)
    sumY0  +=      10.221 * cos(0.47130697459  + 66.67347455760 *T)
    sumY0  +=      13.400 * cos(3.51530110476  + 25.84668014910 *T)
    sumY0  +=      12.805 * cos(1.73149783747  + 124.53422035260 *T)
    sumY0  +=      10.035 * cos(3.75323450486  + 82.61453593110 *T)
    sumY0  +=       9.987 * cos(0.79147161221  + 115.18544111810 *T)
    sumY0  +=      10.797 * cos(2.77657277725  + 199.31581891990 *T)
    sumY0  +=       9.996 * cos(0.52088482718  + 59.03866269500 *T)
    sumY0  +=      10.812 * cos(0.46401017699  + 89.25251005240 *T)
    sumY0  +=      13.348 * cos(0.57976101887  + 139.75815209550 *T)
    sumY0  +=      12.606 * cos(2.70123732269  + 28.55499313480 *T)
    sumY0  +=       9.781 * cos(4.00229285265  + 84.47476902181 *T)
    sumY0  +=       9.902 * cos(1.62968500763  + 262.77653057739 *T)
    sumY0  +=      10.352 * cos(0.81152556420  + 361.62163734781 *T)
    sumY0  +=      10.581 * cos(5.29068527712  + 679.81349389780 *T)
    sumY0  +=      10.976 * cos(4.56885713555  + 209.08949479660 *T)
    sumY0  +=      11.881 * cos(5.29230553590  + 827.67977000299 *T)
    sumY0  +=      10.568 * cos(4.17744029529  + 177.06110633080 *T)
    sumY0  +=      11.169 * cos(3.19909284892  + 318.43567403349 *T)
    sumY0  +=      10.655 * cos(3.65009353213  + 264.20612153150 *T)
    sumY0  +=      10.248 * cos(4.87010430516  + 65.63094483399 *T)
    sumY0  +=       9.768 * cos(5.60017209442  + 422.42222012939 *T)
    sumY0  +=      11.423 * cos(3.52166240719  + 679.49797971270 *T)
    sumY0  +=      11.319 * cos(5.77253068317  + 43.79347909271 *T)
    sumY0  +=      10.444 * cos(2.08212477220  + 64.24034821780 *T)
    sumY0  +=      10.733 * cos(0.19348042611  + 89.69445303710 *T)
    sumY0  +=       9.435 * cos(2.65961671499  + 59.83511420271 *T)
    sumY0  +=      10.593 * cos(2.48672884196  + 71.94459604960 *T)
    sumY0  +=      10.332 * cos(4.98186467280  + 61.04440026510 *T)
    sumY0  +=      10.962 * cos(2.11147523367  + 526.00262931501 *T)
    sumY0  +=      11.184 * cos(2.96936696687  + 150.61016377020 *T)
    sumY0  +=      10.812 * cos(3.76848381140  + 546.71262299849 *T)
    sumY0  +=      10.427 * cos(0.32210108380  + 106.03035047280 *T)
    sumY0  +=       9.926 * cos(0.10882768465  + 767.12526543730 *T)
    sumY0  +=      12.114 * cos(5.36578468862  + 404.86248241631 *T)
    sumY0  +=       9.406 * cos(5.72421215324  + 69.55640045270 *T)
    sumY0  +=       9.406 * cos(5.23837558994  + 80.49443164890 *T)
    sumY0  +=      10.953 * cos(3.07486144408  + 57.86300652739 *T)
    sumY0  +=      12.472 * cos(3.42726511510  + 337.55812914470 *T)
    sumY0  +=       9.300 * cos(5.90247156785  + 831.34879870770 *T)
    sumY0  +=      10.211 * cos(1.92534998559  + 544.16187657970 *T)
    sumY0  +=       9.945 * cos(2.81330302968  + 192.93598510200 *T)
    sumY0  +=      11.250 * cos(5.57455276310  + 544.26494244370 *T)
    sumY0  +=       9.524 * cos(1.19760937655  + 217.68751450571 *T)
    sumY0  +=       8.854 * cos(2.95194128004  + 210.36151918381 *T)
    sumY0  +=      10.193 * cos(2.85819678368  + 443.61984877990 *T)
    sumY0  +=       9.954 * cos(5.11935965746  + 79.85134119120 *T)
    sumY0  +=       9.954 * cos(5.84320899680  + 70.19949091040 *T)
    sumY0  +=       9.930 * cos(0.62899045525  + 63.01650794040 *T)
    sumY0  +=       8.650 * cos(1.47118969332  + 533.11617631580 *T)
    sumY0  +=      10.719 * cos(2.71929706040  + 399.92142932440 *T)
    sumY0  +=      12.021 * cos(2.02390354619  + 54.89406111079 *T)
    sumY0  +=      11.050 * cos(1.37690562890  + 210.60759739969 *T)
    sumY0  +=      10.384 * cos(4.28747391550  + 30.35498002820 *T)
    sumY0  +=      10.014 * cos(6.20125275236  + 419.24082639170 *T)
    sumY0  +=      11.868 * cos(2.96967485309  + 353.54488250050 *T)
    sumY0  +=      11.500 * cos(3.40411765505  + 402.89037474099 *T)
    sumY0  +=      10.255 * cos(4.72943172340  + 30.93323772040 *T)
    sumY0  +=       8.996 * cos(3.09833830500  + 179.03321400610 *T)
    sumY0  +=      11.583 * cos(0.98746991822  + 754.59509246510 *T)
    sumY0  +=      10.299 * cos(1.57297771598  + 112.89781925770 *T)
    sumY0  +=       8.391 * cos(4.97268364529  + 231.18089532440 *T)
    sumY0  +=       9.702 * cos(1.22534705521  + 181.03895157621 *T)
    sumY0  +=       8.169 * cos(1.42365397432  + 674.18441960530 *T)
    sumY0  +=       9.580 * cos(4.84883749459  + 292.25666475190 *T)
    sumY0  +=       8.174 * cos(4.64994635843  + 121.59890483079 *T)
    sumY0  +=      10.408 * cos(5.25476225428  + 533.86693584120 *T)
    sumY0  +=       9.226 * cos(5.25948012740  + 74.11039231410 *T)
    sumY0  +=       9.226 * cos(5.70310761578  + 75.94043978751 *T)
    sumY0  +=       8.015 * cos(6.04785554826  + 350.57593708390 *T)
    sumY0  +=       9.330 * cos(6.27465213976  + 53.04589010760 *T)
    sumY0  +=       9.204 * cos(2.03165601534  + 223.62540533890 *T)
    sumY0  +=       8.485 * cos(2.34094108621  + 212.57970507500 *T)
    sumY0  +=       8.003 * cos(0.91246310178  + 1022.06885865090 *T)
    sumY0  +=       8.044 * cos(0.83166336147  + 77.36122202100 *T)
    sumY0  +=       8.044 * cos(3.84773907453  + 72.68961008061 *T)
    sumY0  +=      10.452 * cos(1.98498356773  + 841.90686400460 *T)
    sumY0  +=       7.961 * cos(2.31145265393  + 43.37278796740 *T)
    sumY0  +=       8.333 * cos(1.18342051201  + 152.09463647851 *T)
    sumY0  +=       9.216 * cos(2.14207964363  + 670.25226634220 *T)
    sumY0  +=       9.064 * cos(5.27663226147  + 153.43035134050 *T)
    sumY0  +=       7.893 * cos(2.51477711531  + 93.93542284090 *T)
    sumY0  +=       9.501 * cos(3.98117068682  + 156.70811329780 *T)
    sumY0  +=       9.347 * cos(0.34595548426  + 157.22937815961 *T)
    sumY0  +=       8.617 * cos(2.59731331592  + 85.81048388380 *T)
    sumY0  +=       7.939 * cos(2.06846952168  + 109.22629842550 *T)
    sumY0  +=       8.607 * cos(4.69708547234  + 63.17656663481 *T)
    sumY0  +=      10.579 * cos(4.46292936018  + 17.76992530181 *T)
    sumY0  +=       7.543 * cos(3.14113560246  + 77.26064828450 *T)
    sumY0  +=       7.543 * cos(4.67985948713  + 72.79018381711 *T)
    sumY0  +=       8.486 * cos(3.16231049397  + 78.10623605200 *T)
    sumY0  +=       9.532 * cos(1.91025418881  + 835.28095197081 *T)
    sumY0  +=       7.509 * cos(4.96017180762  + 216.72430665921 *T)
    sumY0  +=       8.683 * cos(2.90960249329  + 117.77862615229 *T)
    sumY0  +=       7.488 * cos(2.09613529005  + 90.21571789890 *T)
    sumY0  +=       9.008 * cos(1.59995368963  + 175.67050971460 *T)
    sumY0  +=       7.543 * cos(2.16991727809  + 74.29170286790 *T)
    sumY0  +=       7.543 * cos(5.65107781150  + 75.75912923371 *T)
    sumY0  +=       8.913 * cos(1.43842918678  + 683.74564716089 *T)
    sumY0  +=       8.280 * cos(5.45801088768  + 216.23667169220 *T)
    sumY0  +=      10.165 * cos(5.74311979024  + 1364.97191706540 *T)
    sumY0  +=       7.591 * cos(3.10048399132  + 351.57277482520 *T)
    sumY0  +=       8.205 * cos(3.52665267682  + 37.41364527480 *T)
    sumY0  +=       7.588 * cos(3.28488331617  + 198.56505939450 *T)
    sumY0  +=       8.473 * cos(3.11905002019  + 71.25332148210 *T)
    sumY0  +=       8.473 * cos(1.56035241581  + 78.79751061951 *T)
    sumY0  +=       9.107 * cos(3.10384486237  + 37.15301284391 *T)
    sumY0  +=       7.766 * cos(4.05884078794  + 1479.11039154791 *T)
    sumY0  +=       8.403 * cos(4.51384294317  + 253.21530302180 *T)
    sumY0  +=       7.642 * cos(2.21841304053  + 49.96507010641 *T)
    sumY0  +=       9.234 * cos(1.16089511561  + 166.84548746939 *T)
    sumY0  +=       8.874 * cos(5.92801753842  + 483.46435966210 *T)
    sumY0  +=       7.128 * cos(0.54929503334  + 173.69840203930 *T)
    sumY0  +=       7.872 * cos(2.83479383420  + 437.88770862340 *T)
    sumY0  +=       9.185 * cos(1.99612196710  + 101.04896984171 *T)
    sumY0  +=       7.312 * cos(6.13883546415  + 526.75338884040 *T)
    sumY0  +=       9.194 * cos(1.81474760228  + 97.43233076370 *T)
    sumY0  +=       8.182 * cos(2.53164006705  + 115.86676467430 *T)
    sumY0  +=       9.663 * cos(6.15396199238  + 763.19311217420 *T)
    sumY0  +=       8.486 * cos(3.78958880101  + 242.11472100371 *T)
    sumY0  +=       8.799 * cos(0.25004467051  + 366.72944677850 *T)
    sumY0  +=       8.458 * cos(1.58462208087  + 297.98880490840 *T)
    sumY0  +=       8.904 * cos(1.84053989684  + 60.20762121820 *T)
    sumY0  +=       7.575 * cos(5.75688781236  + 214.50612076800 *T)
    sumY0  +=       6.999 * cos(5.60818664108  + 144.69920518740 *T)
    sumY0  +=       8.000 * cos(3.63615956812  + 325.13983075710 *T)
    sumY0  +=       8.498 * cos(3.89791695768  + 20.73887071840 *T)
    sumY0  +=       9.641 * cos(3.89512516279  + 56.89979868090 *T)
    sumY0  +=       7.094 * cos(6.10443989292  + 49.75262178531 *T)
    sumY0  +=       7.954 * cos(4.06222552835  + 87.03432416121 *T)
    sumY0  +=       6.970 * cos(1.96111327573  + 130.79643409150 *T)
    sumY0  +=       7.608 * cos(5.47728460394  + 409.31590054121 *T)
    sumY0  +=       7.168 * cos(5.64518457340  + 1265.81129610991 *T)
    sumY0  +=       8.728 * cos(3.53511019168  + 78.99119920869 *T)
    sumY0  +=       7.572 * cos(0.17135144497  + 148.91324274080 *T)
    sumY0  +=       6.783 * cos(3.03195983082  + 555.31064270760 *T)
    sumY0  +=       8.481 * cos(4.53301555531  + 119.26309886060 *T)
    sumY0  +=       7.247 * cos(2.82713079705  + 33.26904369061 *T)
    sumY0  +=       6.911 * cos(5.67453645573  + 41.51255487670 *T)
    sumY0  +=       6.813 * cos(2.64716458917  + 301.17019864609 *T)
    sumY0  +=       8.470 * cos(4.08048553109  + 250.19147585100 *T)
    sumY0  +=       6.549 * cos(2.37816417441  + 129.36014549300 *T)
    sumY0  +=       6.543 * cos(2.96794347356  + 44.36962570870 *T)
    sumY0  +=       8.617 * cos(3.14518586851  + 150.70078649541 *T)
    sumY0  +=       7.676 * cos(4.83443438533  + 60.46825364911 *T)
    sumY0  +=       8.905 * cos(0.03447668306  + 749.82616015511 *T)
    sumY0  +=       7.175 * cos(5.62428004061  + 445.96446347071 *T)
    sumY0  +=       6.465 * cos(5.46336528208  + 75.50098893029 *T)
    sumY0  +=       7.106 * cos(0.97448145412  + 6456.63624021419 *T)
    sumY0  +=       7.220 * cos(5.10879667762  + 1692.40948698591 *T)
    sumY0  +=       6.926 * cos(0.91784419155  + 20.90562705720 *T)
    sumY0  +=       7.786 * cos(2.71967618992  + 100.40127618030 *T)
    sumY0  +=       6.164 * cos(1.79214990956  + 1376.01761732930 *T)
    sumY0  +=       7.312 * cos(4.44944337943  + 188.90325810240 *T)
    sumY0  +=       6.788 * cos(6.26547457533  + 86.87426546680 *T)
    sumY0  +=       6.246 * cos(2.99700510018  + 14.57397734909 *T)
    sumY0  +=       6.057 * cos(4.77548604030  + 267.71758366931 *T)
    sumY0  +=       7.767 * cos(3.63015905985  + 95.73540973430 *T)
    sumY0  +=       7.632 * cos(5.85317558953  + 22.22334342671 *T)
    sumY0  +=       6.881 * cos(3.80941789384  + 163.87654205279 *T)
    sumY0  +=       6.523 * cos(0.41741185861  + 368.83825021259 *T)
    sumY0  +=       6.006 * cos(1.06015938578  + 560.93971700010 *T)
    sumY0  +=       6.916 * cos(2.44549964284  + 497.69145366371 *T)
    sumY0  +=       7.042 * cos(4.03629417682  + 39.11305843170 *T)
    sumY0  +=       7.444 * cos(0.25903368767  + 396.78572756900 *T)
    sumY0  +=       7.949 * cos(1.74199043054  + 474.70302789170 *T)
    sumY0  +=       7.441 * cos(5.03533974603  + 260.01333583751 *T)
    sumY0  +=       8.032 * cos(0.83846257555  + 224.06734832360 *T)
    sumY0  +=       5.746 * cos(2.56394284072  + 193.04785968660 *T)
    sumY0  +=       5.743 * cos(4.76658104509  + 20201.98247923659 *T)
    sumY0  +=       7.757 * cos(3.79185602312  + 688.41151360689 *T)
    sumY0  +=       7.329 * cos(0.85857969446  + 237.17366791179 *T)
    sumY0  +=       5.919 * cos(0.50982288637  + 17.39416491939 *T)
    sumY0  +=       6.252 * cos(5.02682854905  + 132.16903221521 *T)
    sumY0  +=       5.870 * cos(4.28917302578  + 690.85919416170 *T)
    sumY0  +=       6.964 * cos(3.02573149302  + 197.55615956570 *T)
    sumY0  +=       7.219 * cos(5.43456429689  + 266.96931627140 *T)
    sumY0  +=       7.947 * cos(5.61352830758  + 514.46929408409 *T)
    sumY0  +=       7.626 * cos(3.67423708438  + 842.39449897160 *T)
    sumY0  +=       6.895 * cos(1.47957605481  + 425.35753565121 *T)
    sumY0  +=       6.665 * cos(2.95467155233  + 436.56999225390 *T)
    sumY0  +=       6.291 * cos(1.67292148805  + 343.49601997789 *T)
    sumY0  +=       6.140 * cos(4.35860072087  + 213.70966926030 *T)
    sumY0  +=       5.915 * cos(0.86094320174  + 62.75587550951 *T)
    sumY0  +=       5.915 * cos(3.81845923426  + 87.29495659210 *T)
    sumY0  +=       6.781 * cos(4.05060060272  + 647.25465079831 *T)
    sumY0  +=       6.309 * cos(2.99140195909  + 178.37882270030 *T)
    sumY0  +=       5.782 * cos(3.22810004423  + 401.29402744810 *T)
    sumY0  +=       7.776 * cos(5.11993350983  + 204.94489321240 *T)
    sumY0  +=       6.809 * cos(1.08727914779  + 291.94784821120 *T)
    sumY0  +=       5.680 * cos(4.17409037680  + 79.64559051450 *T)
    sumY0  +=       5.958 * cos(3.65881665841  + 808.76976321290 *T)
    sumY0  +=       5.850 * cos(6.26855076332  + 358.65269193120 *T)
    sumY0  +=       5.624 * cos(0.99993676157  + 42.30900638440 *T)
    sumY0  +=       7.759 * cos(1.83986340459  + 79.27308349900 *T)
    sumY0  +=       5.495 * cos(0.83012628966  + 159.20818347930 *T)
    sumY0  +=       5.488 * cos(2.05261070153  + 293.95358578131 *T)
    sumY0  +=       7.423 * cos(5.31470022391  + 139.72452220071 *T)
    sumY0  +=       5.570 * cos(5.97928286717  + 187.99493201010 *T)
    sumY0  +=       6.894 * cos(3.28434910878  + 407.83142783290 *T)
    sumY0  +=       5.539 * cos(4.84676356044  + 1235.36795408890 *T)
    sumY0  +=       6.394 * cos(4.15130654566  + 956.53297345411 *T)
    sumY0  +=       5.577 * cos(4.10482070615  + 168.92946972931 *T)
    sumY0  +=       6.276 * cos(4.68793579564  + 97.00494199400 *T)
    sumY0  +=       7.541 * cos(4.08226152277  + 303.87851163179 *T)
    sumY0  +=       6.343 * cos(4.90270462176  + 164.36417701981 *T)
    sumY0  +=       6.353 * cos(3.79216315882  + 370.21965705690 *T)
    sumY0  +=       5.390 * cos(6.26914064855  + 818.01547658340 *T)
    sumY0  +=       5.777 * cos(5.82865048284  + 271.37455028650 *T)
    sumY0  +=       6.126 * cos(5.65499715014  + 45.22976769120 *T)
    sumY0  +=       5.488 * cos(2.09536392742  + 136.47369249381 *T)
    sumY0  +=       6.742 * cos(2.83749217354  + 89.84321088341 *T)
    sumY0  +=       6.582 * cos(2.42682862426  + 1291.18715623940 *T)
    sumY0  +=       5.219 * cos(0.45337461954  + 7.62048904269 *T)
    sumY0  +=       6.692 * cos(3.14805000647  + 5994.75133850259 *T)
    sumY0  +=       6.355 * cos(1.10707964029  + 184.48346987229 *T)
    sumY0  +=       7.139 * cos(0.98849473241  + 359.17395679300 *T)
    sumY0  +=       6.215 * cos(4.74107554828  + 1582.20316576650 *T)
    sumY0  +=       5.605 * cos(1.54167032387  + 43.48466255200 *T)
    sumY0  +=       5.449 * cos(2.13791600611  + 229.09691306450 *T)
    sumY0  +=       5.787 * cos(4.92208922830  + 492.06237937121 *T)
    sumY0  +=       6.553 * cos(5.75663546509  + 1065.35788782919 *T)
    sumY0  +=       5.340 * cos(5.86372959256  + 27.49790919620 *T)
    sumY0  +=       5.124 * cos(0.67325293181  + 479.53220639900 *T)
    sumY0  +=       5.421 * cos(1.88003343531  + 298.88257678570 *T)
    sumY0  +=       5.261 * cos(2.44962785242  + 211.57080524620 *T)
    sumY0  +=       6.420 * cos(3.70476513286  + 489.09343395460 *T)
    sumY0  +=       5.889 * cos(0.82353860057  + 1514.53511420000 *T)
    sumY0  +=       4.956 * cos(2.57399481330  + 450.41788159561 *T)
    sumY0  +=       6.450 * cos(4.59218190917  + 9924.96103472219 *T)
    sumY0  +=       4.940 * cos(2.10014956300  + 20.95381116700 *T)
    sumY0  +=       5.847 * cos(6.13596901282  + 89.58257845250 *T)
    sumY0  +=       5.600 * cos(1.83545214397  + 241.85408857280 *T)
    sumY0  +=       6.152 * cos(3.35545390167  + 524.51815660670 *T)
    sumY0  +=       6.088 * cos(1.15145874834  + 711.69312451730 *T)
    sumY0  +=       5.407 * cos(2.11877006223  + 148.16248321540 *T)
    sumY0  +=       6.327 * cos(4.93432455338  + 300.64893378430 *T)
    sumY0  +=       5.490 * cos(2.36691010165  + 261.81332273089 *T)
    sumY0  +=       5.280 * cos(0.85359725433  + 74.42590649920 *T)
    sumY0  +=       4.835 * cos(1.17492690523  + 122.77456099839 *T)
    sumY0  +=       5.417 * cos(2.10462006768  + 34.18406742730 *T)
    sumY0  +=       5.465 * cos(4.18396714587  + 49.89563413720 *T)
    sumY0  +=       4.962 * cos(1.68097237142  + 40.49446527601 *T)
    sumY0  +=       5.202 * cos(5.00837530618  + 102.27950776349 *T)
    sumY0  +=       5.768 * cos(5.58499792979  + 237.92193530970 *T)
    sumY0  +=       5.794 * cos(4.82130895886  + 215.99059347631 *T)
    sumY0  +=       6.057 * cos(3.19358889572  + 1595.04885292380 *T)
    sumY0  +=       6.144 * cos(4.72437637687  + 456.95317090420 *T)
    sumY0  +=       5.439 * cos(4.81364730924  + 164.13717448370 *T)
    sumY0  +=       6.169 * cos(3.17120830135  + 2199.52141658570 *T)
    sumY0  +=       5.803 * cos(5.22454044783  + 476.07562601539 *T)
    sumY0  +=       4.737 * cos(2.35530993243  + 905.64276230799 *T)
    sumY0  +=       4.927 * cos(3.26468598920  + 752.89817143570 *T)
    sumY0  +=       4.656 * cos(1.91522169679  + 89.09494348550 *T)
    sumY0  +=       5.228 * cos(1.17862548607  + 112.63718682681 *T)
    sumY0  +=       5.103 * cos(3.73018774429  + 93.54836161040 *T)
    sumY0  +=       4.793 * cos(3.53993726451  + 257.14745628489 *T)
    sumY0  +=       5.517 * cos(0.15731395881  + 85.95924173011 *T)
    sumY0  +=       5.174 * cos(4.15935293996  + 460.88532416730 *T)
    sumY0  +=       4.600 * cos(4.89133583119  + 693.82813957829 *T)
    sumY0  +=       5.568 * cos(2.10167962126  + 346.15614885380 *T)
    sumY0  +=       5.750 * cos(6.24007416703  + 475.11241816889 *T)
    sumY0  +=       5.821 * cos(5.49083686705  + 331.92905485220 *T)
    sumY0  +=       5.474 * cos(1.31183557573  + 119.09634252180 *T)
    sumY0  +=       5.555 * cos(4.65024532079  + 166.58485503849 *T)
    sumY0  +=       4.721 * cos(0.86426769065  + 818.33099076849 *T)
    sumY0  +=       4.598 * cos(2.05332533302  + 155.43608891061 *T)
    sumY0  +=       4.983 * cos(2.57975544514  + 305.36298434010 *T)
    sumY0  +=       4.604 * cos(5.38714042066  + 240.63024829540 *T)
    sumY0  +=       4.534 * cos(4.08633266459  + 426.35437339250 *T)
    sumY0  +=       6.030 * cos(2.63010642372  + 928.22179780281 *T)
    sumY0  +=       5.073 * cos(4.70022198384  + 119.15122427600 *T)
    sumY0  +=       4.800 * cos(5.53046784681  + 133.82026126229 *T)
    sumY0  +=       5.861 * cos(5.51556326816  + 441.51104534581 *T)
    sumY0  +=       4.509 * cos(2.39065356016  + 106.67804413421 *T)
    sumY0  +=       4.990 * cos(3.48465520513  + 53.40958840249 *T)
    sumY0  +=       5.764 * cos(3.63891986766  + 896.08153475239 *T)
    sumY0  +=       5.407 * cos(4.74700481244  + 469.48334387639 *T)
    sumY0  +=       5.823 * cos(4.10358787531  + 29.50364676630 *T)
    sumY0  +=       5.944 * cos(3.46545832250  + 235.94982763440 *T)
    sumY0  +=       6.147 * cos(1.45101996869  + 1288.21821082279 *T)
    sumY0  +=       4.598 * cos(1.27745263577  + 224.32798075451 *T)
    sumY0  +=       5.709 * cos(2.76379658634  + 893.11258933580 *T)
    sumY0  +=       4.523 * cos(5.13323251802  + 530.09484127250 *T)
    sumY0  +=       5.024 * cos(5.31536294883  + 314.66357946480 *T)
    sumY0  +=       4.527 * cos(4.89680450095  + 100.29821031630 *T)
    sumY0  +=       4.510 * cos(2.34852903978  + 529.77472388370 *T)
    sumY0  +=       5.026 * cos(3.92555320684  + 74.64965566840 *T)
    sumY0  +=       5.303 * cos(1.08916276416  + 273.82223084130 *T)
    sumY0  +=       4.380 * cos(0.99003902246  + 1578.27101250340 *T)
    sumY0  +=       5.525 * cos(0.86348547164  + 453.18107633550 *T)
    sumY0  +=       4.962 * cos(3.52820328545  + 810.90192958261 *T)
    sumY0  +=       4.972 * cos(0.19879848580  + 334.89800026879 *T)
    sumY0  +=       5.401 * cos(1.93836334655  + 193.67219041240 *T)
    sumY0  +=       5.498 * cos(0.71473266670  + 230.58138577279 *T)
    sumY0  +=       4.501 * cos(5.84903184251  + 523.71500745461 *T)
    sumY0  +=       4.303 * cos(3.18721779031  + 1246.41365435280 *T)
    sumY0  +=       5.261 * cos(4.82838100342  + 170.25388374320 *T)
    sumY0  +=       4.481 * cos(1.75705854176  + 5.35162691421 *T)
    sumY0  +=       4.655 * cos(2.79973352045  + 178.43370445449 *T)
    sumY0  +=       5.029 * cos(0.22047838377  + 193.93282284329 *T)
    sumY0  +=       5.743 * cos(3.56115653050  + 246.68001371321 *T)
    sumY0  +=       5.652 * cos(0.86613579011  + 315.41184686271 *T)
    sumY0  +=       4.604 * cos(3.09918959502  + 1072.47143483000 *T)
    sumY0  +=       5.494 * cos(3.15712757827  + 1044.64789414570 *T)
    sumY0  +=       4.789 * cos(3.80684950331  + 361.78839368660 *T)
    sumY0  +=       5.139 * cos(5.49982663321  + 223.20720634109 *T)
    sumY0  +=       5.310 * cos(3.43114456160  + 444.47999076240 *T)
    sumY0  +=       4.815 * cos(4.45553109137  + 3041.24221494710 *T)
    sumY0  +=       4.119 * cos(2.13510221930  + 2120.49215057020 *T)
    sumY0  +=       4.552 * cos(0.42701637654  + 158.50810019120 *T)
    sumY0  +=       4.594 * cos(0.59274325201  + 1310.14955265620 *T)
    sumY0  +=       4.173 * cos(5.85016751929  + 181.40264987110 *T)
    sumY0  +=       4.090 * cos(2.28062400769  + 138.64943976960 *T)
    sumY0  +=       4.884 * cos(5.20933688234  + 6.65728119619 *T)
    sumY0  +=       4.740 * cos(2.65617621588  + 25788.53292982149 *T)
    sumY0  +=       5.699 * cos(2.84935340269  + 772.83258441960 *T)
    sumY0  +=       5.296 * cos(1.41812325430  + 694.79134742480 *T)
    sumY0  +=       4.555 * cos(3.11995889142  + 154.42718908179 *T)
    sumY0  +=       4.669 * cos(4.97123639115  + 127.82748867490 *T)
    sumY0  +=       4.509 * cos(1.69077891813  + 227.35180792530 *T)
    sumY0  +=       4.302 * cos(2.50409604966  + 100.08576199520 *T)
    sumY0  +=       4.027 * cos(4.03844514423  + 294.05665164530 *T)
    sumY0  +=       4.114 * cos(2.93977154621  + 383.06534421420 *T)
    sumY0  +=       4.552 * cos(6.22515109598  + 1361.03976380230 *T)
    sumY0  +=       4.653 * cos(4.63772962271  + 1019.52060435960 *T)
    sumY0  +=       4.849 * cos(6.23804934917  + 119.11759438120 *T)
    sumY0  +=       4.617 * cos(4.88910884699  + 887.48351504330 *T)
    sumY0  +=       5.166 * cos(0.93017811080  + 49.26498681829 *T)
    sumY0  +=       5.543 * cos(2.55030758424  + 296.50433220009 *T)
    sumY0  +=       3.936 * cos(1.95688941946  + 137.79810650771 *T)
    sumY0  +=       3.959 * cos(3.93854751155  + 151.17961274180 *T)
    sumY0  +=       4.546 * cos(0.76154188637  + 44.42661853911 *T)
    sumY0  +=       4.195 * cos(1.28923343857  + 305.21422649380 *T)
    sumY0  +=       3.911 * cos(5.57516730393  + 1020.26887175750 *T)
    sumY0  +=       3.906 * cos(1.61616341819  + 36.41929966101 *T)
    sumY0  +=       4.980 * cos(0.97997283118  + 682.78243931439 *T)
    sumY0  +=       3.894 * cos(6.27535408135  + 13.77082819700 *T)
    sumY0  +=       4.225 * cos(5.37063958024  + 310.73142620169 *T)
    sumY0  +=       3.999 * cos(5.57100133771  + 93.98781246761 *T)
    sumY0  +=       3.999 * cos(5.39158640547  + 56.06301963400 *T)
    sumY0  +=       4.072 * cos(2.47052943106  + 956.84848763920 *T)
    sumY0  +=       4.160 * cos(5.20231534295  + 138.92807069299 *T)
    sumY0  +=       3.870 * cos(3.52554234711  + 316.14805217309 *T)
    sumY0  +=       5.282 * cos(1.35031701337  + 442.99551805410 *T)
    sumY0  +=       3.856 * cos(4.98262634270  + 778.17096570120 *T)
    sumY0  +=       5.026 * cos(2.04291728735  + 252.89978883670 *T)
    sumY0  +=       4.872 * cos(3.77484767846  + 196.59295171920 *T)
    sumY0  +=       4.088 * cos(3.31585670195  + 770.58184582090 *T)
    sumY0  +=       3.976 * cos(0.29724722006  + 34.44469985821 *T)
    sumY0  +=       3.976 * cos(3.43883987365  + 33.95706489120 *T)
    sumY0  +=       4.425 * cos(1.38560089800  + 621.98185653281 *T)
    sumY0  +=       3.827 * cos(4.15987194313  + 29.81246330700 *T)
    sumY0  +=       4.303 * cos(5.49962330361  + 302.39403892350 *T)
    sumY0  +=       4.082 * cos(2.93795043280  + 76.98546163860 *T)
    sumY0  +=       4.775 * cos(3.44144906594  + 379.25961975071 *T)
    sumY0  +=       4.866 * cos(3.05946215728  + 349.12509427039 *T)
    sumY0  +=       4.558 * cos(3.79281085339  + 484.67364572450 *T)
    sumY0  +=       3.896 * cos(3.48090977966  + 525.74199688410 *T)
    sumY0  +=       3.790 * cos(2.68849385464  + 168.81759514471 *T)
    sumY0  +=       4.584 * cos(2.02400920188  + 30.41197285861 *T)
    sumY0  +=       3.789 * cos(4.87731374266  + 225.58545092669 *T)
    sumY0  +=       3.919 * cos(3.87130555190  + 367.61440993520 *T)
    sumY0  +=       4.495 * cos(0.77984598931  + 823.74761673989 *T)
    sumY0  +=       4.071 * cos(1.17727620682  + 152.28832506769 *T)
    sumY0  +=       4.431 * cos(2.72709685973  + 781.83999440590 *T)
    sumY0  +=       3.787 * cos(2.48815838156  + 96.22304470130 *T)
    sumY0  +=       3.738 * cos(4.07013940817  + 909.68679015570 *T)
    sumY0  +=       3.824 * cos(3.30062177264  + 81.01818863820 *T)
    sumY0  +=       5.100 * cos(0.20422183986  + 786.47472308461 *T)
    sumY0  +=       4.678 * cos(5.03406582155  + 294.42915866079 *T)
    sumY0  +=       4.456 * cos(1.79112458929  + 473.41549713950 *T)
    sumY0  +=       4.182 * cos(0.18466444565  + 109.96250373590 *T)
    sumY0  +=       4.721 * cos(4.73119528048  + 487.64259114110 *T)
    sumY0  +=       4.194 * cos(4.10486683345  + 135.82599883240 *T)
    sumY0  +=       4.150 * cos(0.87761383870  + 208.60185982960 *T)
    sumY0  +=       3.881 * cos(3.18364105348  + 527.48710202330 *T)
    sumY0  +=       5.003 * cos(2.80882112541  + 26088.14695905770 *T)
    sumY0  +=       4.925 * cos(5.52756833519  + 104.82106441041 *T)
    sumY0  +=       4.629 * cos(5.06094518975  + 324.97307441831 *T)
    sumY0  +=       3.764 * cos(3.88568266144  + 81.17575520511 *T)
    sumY0  +=       3.764 * cos(3.93531242815  + 68.87507689650 *T)
    sumY0  +=       4.634 * cos(3.50032175228  + 439.53893767049 *T)
    sumY0  +=       4.661 * cos(4.44746944301  + 1031.63008620649 *T)
    sumY0  +=       4.246 * cos(1.59561640750  + 290.87525790760 *T)
    sumY0  +=       4.701 * cos(4.20435344056  + 292.57217893700 *T)
    sumY0  +=       4.168 * cos(3.07866792955  + 347.68880567190 *T)
    sumY0  +=       3.574 * cos(0.81177194493  + 101.85211899380 *T)
    sumY0  +=       4.316 * cos(5.40388665351  + 1034.59903162310 *T)
    sumY0  +=       4.648 * cos(4.59798157023  + 701.01993126891 *T)
    sumY0  +=       5.000 * cos(4.96762725751  + 636.42389098300 *T)
    sumY0  +=       4.950 * cos(1.58366506104  + 901.71060904489 *T)
    sumY0  +=       4.343 * cos(3.91944557060  + 372.77040347569 *T)
    sumY0  +=       3.612 * cos(0.37077229030  + 154.31531449720 *T)
    sumY0  +=       4.180 * cos(4.53703226837  + 304.99047732459 *T)
    sumY0  +=       3.767 * cos(1.75098839325  + 1040.97886544099 *T)
    sumY0  +=       4.105 * cos(3.02027598948  + 1169.57143646120 *T)
    sumY0  +=       4.389 * cos(3.06369646718  + 67.63668240410 *T)
    sumY0  +=       3.610 * cos(1.32047712869  + 969.86629557841 *T)
    sumY0  +=       3.589 * cos(4.99114813732  + 1340.59289467720 *T)
    sumY0  +=       4.184 * cos(4.66399157943  + 1575.08961876571 *T)
    sumY0  +=       4.327 * cos(0.92906184743  + 470.44655172289 *T)
    sumY0  +=       3.635 * cos(2.36421665060  + 618.04970326970 *T)
    sumY0  +=       4.568 * cos(4.98874622103  + 374.46732450509 *T)
    sumY0  +=       4.014 * cos(3.27163799518  + 191.07575201129 *T)
    sumY0  +=       3.518 * cos(2.98191800179  + 812.70191647599 *T)
    sumY0  +=       4.543 * cos(5.05516422535  + 18.87863762769 *T)
    sumY0  +=       4.255 * cos(4.93411147571  + 257.78059573129 *T)
    sumY0  +=       3.536 * cos(0.84249596400  + 1158.78636862821 *T)
    sumY0  +=       4.782 * cos(0.83189462557  + 20.36311033600 *T)
    sumY0  +=       3.580 * cos(5.11301333129  + 155.79978720550 *T)
    sumY0  +=       4.461 * cos(4.71187230264  + 155.91166179010 *T)
    sumY0  +=       4.866 * cos(2.95132293953  + 16.85490156510 *T)
    sumY0  +=       4.353 * cos(4.31191357131  + 154.54231703330 *T)
    sumY0  +=       3.770 * cos(4.63520445622  + 238.91877305099 *T)
    sumY0  +=       3.780 * cos(4.53874966526  + 250.35823218980 *T)
    sumY0  +=       4.854 * cos(0.38485985963  + 823.12328601411 *T)
    sumY0  +=       4.833 * cos(3.23674545351  + 1894.66288216001 *T)
    sumY0  +=       3.662 * cos(1.40177918208  + 319.92014674179 *T)
    sumY0  +=       3.977 * cos(5.68574160053  + 326.19691469570 *T)
    sumY0  +=       3.861 * cos(5.74438369570  + 47.99296243110 *T)
    sumY0  +=       4.433 * cos(5.42313050798  + 1685.29593998510 *T)
    sumY0  +=       4.182 * cos(3.13684968245  + 548.40954402789 *T)
    sumY0  +=       3.421 * cos(3.51995958462  + 605.03189533049 *T)
    sumY0  +=       3.876 * cos(3.99363601420  + 302.13340649259 *T)
    sumY0  +=       4.008 * cos(1.83807922875  + 6137.20098863639 *T)
    sumY0  +=       4.500 * cos(1.38119932706  + 47.30494123050 *T)
    sumY0  +=       4.050 * cos(1.29963474565  + 6.46359260701 *T)
    sumY0  +=       3.411 * cos(1.31630185032  + 997.68983626269 *T)
    sumY0  +=       4.312 * cos(4.01040874514  + 847.81112494300 *T)
    sumY0  +=       3.428 * cos(1.55247050413  + 71.35638734610 *T)
    sumY0  +=       3.928 * cos(6.23536924317  + 395.30125486070 *T)
    sumY0  +=       3.650 * cos(0.23968284456  + 1254.27796087899 *T)
    sumY0  +=       3.844 * cos(5.28920917279  + 121.49583896680 *T)
    sumY0  +=       3.966 * cos(1.48989873375  + 761.49619114479 *T)
    sumY0  +=       3.497 * cos(4.17241680080  + 319.32944591080 *T)
    sumY0  +=       3.510 * cos(5.54109119947  + 273.34665796181 *T)
    sumY0  +=       3.408 * cos(5.53135426605  + 1016.33671849440 *T)
    sumY0  +=       4.206 * cos(0.51867358485  + 435.44002806861 *T)
    sumY0  +=       3.372 * cos(5.55788408036  + 213.49472881170 *T)
    sumY0  +=       3.369 * cos(2.47122903489  + 213.59109703130 *T)
    sumY0  +=       4.154 * cos(3.96250853547  + 357.16821922291 *T)
    sumY0  +=       3.966 * cos(4.88587910106  + 185.33480313420 *T)
    sumY0  +=       4.636 * cos(5.10105010921  + 6169.91739830280 *T)
    sumY0  +=       4.042 * cos(6.22077286894  + 432.47108265200 *T)
    sumY0  +=       3.366 * cos(2.66012811137  + 699.45721387080 *T)
    sumY0  +=       3.632 * cos(4.59767699340  + 771.05741870039 *T)
    sumY0  +=       3.967 * cos(5.95729548122  + 280.24775664150 *T)
    sumY0  +=       3.436 * cos(1.80871659193  + 182.36585771760 *T)
    sumY0  +=       4.588 * cos(6.17666653669  + 25483.18676042879 *T)
    sumY0  +=       3.825 * cos(4.86834262213  + 6129.33668211020 *T)
    sumY0  +=       3.243 * cos(4.36468243327  + 279.97256999560 *T)
    sumY0  +=       4.579 * cos(1.69365778926  + 2735.89604555439 *T)
    sumY0  +=       4.222 * cos(4.00381650923  + 344.94686279140 *T)
    sumY0  +=       4.222 * cos(4.00381650923  + 344.45922782439 *T)
    sumY0  +=       3.528 * cos(2.89373582714  + 187.78248368899 *T)
    sumY0  +=       3.976 * cos(0.47836354163  + 433.43429049850 *T)
    sumY0  +=       3.885 * cos(1.59237659251  + 254.33607743521 *T)
    sumY0  +=       3.256 * cos(3.76116458060  + 1756.14538528930 *T)
    sumY0  +=       3.450 * cos(5.70924350001  + 1336.66074141410 *T)
    sumY0  +=       3.763 * cos(3.02653659982  + 598.77849031221 *T)
    sumY0  +=       3.661 * cos(3.87306757949  + 523.99689174490 *T)
    sumY0  +=       3.764 * cos(0.31801539201  + 498.65466151020 *T)
    sumY0  +=       3.218 * cos(5.81500080361  + 829.37669103239 *T)
    sumY0  +=       3.783 * cos(3.26079140282  + 10067.41068485599 *T)
    sumY0  +=       4.463 * cos(5.63073311182  + 686.71459257749 *T)
    sumY0  +=       3.955 * cos(0.51398619980  + 386.99749747730 *T)
    sumY0  +=       3.269 * cos(2.24915426287  + 90.80641872990 *T)
    sumY0  +=       3.268 * cos(2.43000508529  + 59.24441337170 *T)
    sumY0  +=       3.540 * cos(0.11026568904  + 313.17910675650 *T)
    sumY0  +=       3.535 * cos(4.93360466961  + 78039.12081153718 *T)
    sumY0  +=       3.318 * cos(1.03787086284  + 522.33360061030 *T)
    sumY0  +=       3.173 * cos(0.85372362523  + 1475.17823828480 *T)
    sumY0  +=       3.156 * cos(2.68545956575  + 1596.43025976811 *T)
    sumY0  +=       4.282 * cos(3.08651511729  + 269.31393096221 *T)
    sumY0  +=       3.178 * cos(4.19270974886  + 1886.79857563380 *T)
    sumY0  +=       4.407 * cos(0.80883083873  + 236.74627914210 *T)
    sumY0  +=       3.698 * cos(3.47042194825  + 762.98066385310 *T)
    sumY0  +=       3.368 * cos(1.82330731917  + 196.86813836510 *T)
    sumY0  +=       4.221 * cos(3.28486189404  + 102.74589087111 *T)
    sumY0  +=       3.725 * cos(0.49946278623  + 511.28790034640 *T)
    sumY0  +=       3.791 * cos(2.10793640929  + 1147.74066836430 *T)
    sumY0  +=       3.211 * cos(3.96787240218  + 151.81275218821 *T)
    sumY0  +=       3.516 * cos(2.35583623511  + 470.24960976680 *T)
    sumY0  +=       3.947 * cos(5.23994783505  + 757.04277301989 *T)
    sumY0  +=       3.990 * cos(5.62067401902  + 908.09044286280 *T)
    sumY0  +=       3.674 * cos(4.30821160157  + 631.78916230430 *T)
    sumY0  +=       3.787 * cos(2.03840839568  + 798.14475407429 *T)
    sumY0  +=       4.369 * cos(0.24055897355  + 10100.12709452239 *T)
    sumY0  +=       3.536 * cos(0.89255741356  + 354.75416856289 *T)
    sumY0  +=       3.391 * cos(3.82986107065  + 5908.92403967139 *T)
    sumY0  +=       3.083 * cos(2.91678290429  + 829.06117684730 *T)
    sumY0  +=       4.306 * cos(2.93642388121  + 503.10807963510 *T)
    sumY0  +=       3.547 * cos(5.90192172992  + 38.84993387330 *T)
    sumY0  +=       3.214 * cos(4.30099996048  + 660.85109748100 *T)
    sumY0  +=       3.069 * cos(2.00421290803  + 53.56715496940 *T)
    sumY0  +=       3.103 * cos(5.68152185030  + 367.35377750430 *T)
    sumY0  +=       3.753 * cos(3.59220522200  + 486.86069384840 *T)
    sumY0  +=       3.982 * cos(0.55888280170  + 614.59312288610 *T)
    sumY0  +=       3.567 * cos(3.15675663390  + 473.21855518340 *T)
    sumY0  +=       3.603 * cos(0.00781817590  + 10059.54637832979 *T)
    sumY0  +=       3.339 * cos(2.35934486509  + 15.25530090530 *T)
    sumY0  +=       3.342 * cos(6.08494495859  + 391.73279989249 *T)
    sumY0  +=       3.638 * cos(0.04048968434  + 104.77288030060 *T)
    sumY0  +=       3.557 * cos(1.72046667887  + 365.86930479600 *T)
    sumY0  +=       3.047 * cos(5.68168716112  + 75.23786437190 *T)
    sumY0  +=       3.047 * cos(2.13930792847  + 74.81296772971 *T)
    sumY0  +=       3.632 * cos(3.48011331291  + 2175.14239419750 *T)
    sumY0  +=       3.727 * cos(3.34969084744  + 668.76779363389 *T)
    sumY0  +=       3.110 * cos(0.34401803032  + 95.68722562449 *T)
    sumY0  +=       3.110 * cos(3.48561068391  + 96.17486059150 *T)
    sumY0  +=       3.034 * cos(4.47612078147  + 98.65617104110 *T)
    sumY0  +=       3.292 * cos(3.11428915982  + 68.90870679129 *T)
    sumY0  +=       3.166 * cos(2.82060316319  + 5.69407334969 *T)
    sumY0  +=       3.043 * cos(3.58244615080  + 100.14064374939 *T)
    sumY0  +=       3.150 * cos(5.49037416472  + 6532.90231148980 *T)
    sumY0  +=       3.284 * cos(5.45199159924  + 138.87318893880 *T)
    sumY0  +=       3.008 * cos(2.34154854306  + 287.83687652179 *T)
    sumY0  +=       3.451 * cos(0.04036306515  + 1206.21999939070 *T)
    sumY0  +=       3.235 * cos(5.57049156467  + 302.65467135440 *T)
    sumY0  +=       4.224 * cos(2.82331346733  + 527.22646959241 *T)
    sumY0  +=       2.990 * cos(5.76827883617  + 182.52342428450 *T)
    sumY0  +=       3.383 * cos(0.31703037326  + 30.95448957981 *T)
    sumY0  +=       4.137 * cos(1.16119065208  + 436.40323591510 *T)
    sumY0  +=       3.071 * cos(2.92150327875  + 198.82818395289 *T)
    sumY0  +=       3.076 * cos(5.55405559047  + 608.00084074710 *T)
    sumY0  +=       3.557 * cos(3.38304796517  + 731.70054278519 *T)
    sumY0  +=       3.557 * cos(0.24145531158  + 732.18817775220 *T)
    sumY0  +=       3.552 * cos(4.79010413998  + 262.20038396140 *T)
    sumY0  +=       4.135 * cos(0.91386728119  + 195.15666312070 *T)
    sumY0  +=       2.951 * cos(5.02001310379  + 161.65586403411 *T)
    sumY0  +=       2.963 * cos(3.96521662641  + 1044.16025917870 *T)
    sumY0  +=       3.762 * cos(2.15913988960  + 751.31063286340 *T)
    sumY0  +=       3.317 * cos(3.52842835842  + 29.55183087609 *T)
    sumY0  +=       3.109 * cos(0.17339598358  + 1061.42573456610 *T)
    sumY0  +=       2.930 * cos(1.50492864538  + 60.31068708220 *T)
    sumY0  +=       2.938 * cos(0.97094618212  + 380.63471000190 *T)
    sumY0  +=       3.292 * cos(6.11698623962  + 576.40520549410 *T)
    sumY0  +=       2.939 * cos(6.16225734191  + 47.02975458460 *T)
    sumY0  +=       2.922 * cos(5.63523473331  + 55.05162767771 *T)
    sumY0  +=       2.922 * cos(5.32735300987  + 94.99920442390 *T)
    sumY0  +=       3.190 * cos(5.25237662753  + 9839.13373589100 *T)
    sumY0  +=       4.110 * cos(3.23168496449  + 2885.94687765600 *T)
    sumY0  +=       3.635 * cos(5.57293444136  + 189.96703968541 *T)
    sumY0  +=       3.962 * cos(1.90445589310  + 511.35159082120 *T)
    sumY0  +=       2.907 * cos(1.21878102451  + 373.29166833749 *T)
    sumY0  +=       2.911 * cos(3.43639672306  + 379.13319095109 *T)
    sumY0  +=       3.057 * cos(2.28069656246  + 517.66524203681 *T)
    sumY0  +=       3.748 * cos(5.29789330458  + 5985.19011094699 *T)
    sumY0  +=       3.250 * cos(2.04631272961  + 385.82184130970 *T)
    sumY0  +=       3.250 * cos(5.18790538319  + 386.30947627670 *T)
    sumY0  +=       3.118 * cos(1.63458065883  + 404.75060783170 *T)
    sumY0  +=       3.966 * cos(2.44588344367  + 243.59919371200 *T)
    sumY0  +=       2.935 * cos(0.28535439087  + 1190.54206257560 *T)
    sumY0  +=       3.497 * cos(1.71663983829  + 3114.53934080610 *T)
    sumY0  +=       2.871 * cos(5.08352321445  + 743.54939220120 *T)
    sumY0  +=       2.948 * cos(1.96337308764  + 78338.73484077338 *T)
    sumY0  +=       3.905 * cos(2.74341746288  + 669.93675215711 *T)
    sumY0  +=       2.937 * cos(5.00887540692  + 590.48928714380 *T)
    sumY0  +=       4.039 * cos(1.43143454755  + 25633.23759253039 *T)
    sumY0  +=       3.590 * cos(1.87097789478  + 376.70255673880 *T)
    sumY0  +=       2.945 * cos(4.59527251130  + 274.30986580831 *T)
    sumY0  +=       3.853 * cos(4.25262991282  + 224.69167904940 *T)
    sumY0  +=       3.805 * cos(5.95512913152  + 3117.50828622269 *T)
    sumY0  +=       2.874 * cos(3.66628534507  + 89.35557591640 *T)
    sumY0  +=       3.560 * cos(1.32400783975  + 748.34168744680 *T)
    sumY0  +=       2.905 * cos(1.68667534198  + 872.66572021069 *T)
    sumY0  +=       3.186 * cos(1.44507525550  + 53.87597151011 *T)
    sumY0  +=       3.567 * cos(0.42211414373  + 403.37800970800 *T)
    sumY0  +=       3.734 * cos(1.03461913269  + 909.46304098649 *T)
    sumY0  +=       3.317 * cos(4.89578641117  + 204.45725824539 *T)
    sumY0  +=       3.185 * cos(1.08710540110  + 5982.22116553039 *T)
    sumY0  +=       2.821 * cos(1.40365542947  + 116.78178841100 *T)
    sumY0  +=       2.821 * cos(4.54524808306  + 116.29415344400 *T)
    sumY0  +=       2.924 * cos(0.75846929777  + 143.90275367970 *T)
    sumY0  +=       3.219 * cos(5.30142553407  + 76.82540294419 *T)
    sumY0  +=       2.902 * cos(4.45890194491  + 102.58583217670 *T)
    sumY0  +=       2.902 * cos(0.22050049109  + 47.46499992491 *T)
    sumY0  +=       3.255 * cos(5.74825712501  + 482.61302640021 *T)
    sumY0  +=       2.994 * cos(5.12441523239  + 162.39206934450 *T)
    sumY0  +=       3.717 * cos(0.23818776188  + 6061.45618222260 *T)
    sumY0  +=       3.942 * cos(1.61239965392  + 6096.62027244380 *T)
    sumY0  +=       3.586 * cos(1.46434409680  + 983.35967639710 *T)
    sumY0  +=       2.803 * cos(1.56231272650  + 644.07325706060 *T)
    sumY0  +=       2.802 * cos(0.40865678955  + 1592.28565818390 *T)
    sumY0  +=       3.498 * cos(4.37296414373  + 232.29286101720 *T)
    sumY0  +=       3.387 * cos(6.19927452252  + 25861.83005568049 *T)
    sumY0  +=       3.049 * cos(4.51772783361  + 6529.93336607319 *T)
    sumY0  +=       2.910 * cos(2.86684983161  + 355.99256305531 *T)
    sumY0  +=       3.266 * cos(0.17501993824  + 265.48484356309 *T)
    sumY0  +=       3.868 * cos(0.77123558868  + 2124.73981801839 *T)
    sumY0  +=       3.667 * cos(3.65486471998  + 415.90818268021 *T)
    sumY0  +=       2.825 * cos(0.53458307880  + 294.31728407620 *T)
    sumY0  +=       2.775 * cos(2.76199074220  + 552.02618310591 *T)
    sumY0  +=       3.099 * cos(4.33492138442  + 436.50630177909 *T)
    sumY0  +=       2.858 * cos(2.09037051250  + 529.88659846831 *T)
    sumY0  +=       3.700 * cos(4.15420608664  + 25864.79900109709 *T)
    sumY0  +=       2.855 * cos(5.39205642374  + 529.98296668790 *T)
    sumY0  +=       2.805 * cos(3.60457168165  + 2089.05197080790 *T)
    sumY0  +=       2.766 * cos(0.96760939061  + 1600.36241303120 *T)
    sumY0  +=       3.634 * cos(1.43359949431  + 77.88248688280 *T)
    sumY0  +=       3.634 * cos(0.10421028810  + 72.16834521881 *T)
    sumY0  +=       2.849 * cos(2.43470556852  + 323.48860171000 *T)
    sumY0  +=       2.774 * cos(5.37593250032  + 1606.09455318770 *T)
    sumY0  +=       2.788 * cos(0.35923619267  + 187.03172416360 *T)
    sumY0  +=       2.971 * cos(4.67124649084  + 586.13318938849 *T)
    sumY0  +=       2.901 * cos(3.95305253167  + 62.52887297340 *T)
    sumY0  +=       3.856 * cos(4.23035221531  + 113.21333344280 *T)
    sumY0  +=       2.957 * cos(2.32108601739  + 53.82778740031 *T)
    sumY0  +=       2.741 * cos(0.10208197320  + 274.34349570310 *T)
    sumY0  +=       2.895 * cos(0.50990419090  + 151.60700151150 *T)
    sumY0  +=       2.827 * cos(2.92280651892  + 979.21507481291 *T)
    sumY0  +=       3.178 * cos(0.90273597947  + 529.67415014720 *T)
    sumY0  +=       3.613 * cos(0.34914098910  + 583.65187893891 *T)
    sumY0  +=       3.096 * cos(0.25244426036  + 530.19541500900 *T)
    sumY0  +=       3.630 * cos(1.66641450382  + 9991.66587844219 *T)
    sumY0  +=       2.778 * cos(3.05604055978  + 465.02992575149 *T)
    sumY0  +=       2.899 * cos(4.18012528286  + 428.32648106781 *T)
    sumY0  +=       2.977 * cos(4.24176871345  + 74.37772238940 *T)
    sumY0  +=       3.543 * cos(1.80484832429  + 764.67758488250 *T)
    sumY0  +=       2.975 * cos(2.63804412296  + 224.42855449101 *T)
    sumY0  +=       3.288 * cos(2.22514165222  + 201.04636984410 *T)
    sumY0  +=       3.824 * cos(5.50963415570  + 826.92901047760 *T)
    sumY0  +=       3.019 * cos(5.25002429145  + 829.16424271130 *T)
    sumY0  +=       2.780 * cos(0.89847082818  + 64.45528866639 *T)
    sumY0  +=       3.345 * cos(5.42351008501  + 170.46882419179 *T)
    sumY0  +=       3.345 * cos(5.42351008501  + 170.95645915881 *T)
    sumY0  +=       3.500 * cos(0.43652930557  + 9915.39980716659 *T)
    sumY0  +=       3.048 * cos(1.32379750445  + 409.89204715720 *T)
    sumY0  +=       2.854 * cos(4.81337886926  + 112.25012559630 *T)
    sumY0  +=       2.775 * cos(6.07066333426  + 395.34943897051 *T)
    sumY0  +=       2.715 * cos(4.37307075717  + 1533.49751061680 *T)
    sumY0  +=       2.941 * cos(6.11824676700  + 148.11009358871 *T)
    sumY0  +=       2.946 * cos(0.37007486704  + 430.77416162260 *T)
    sumY0  +=       3.426 * cos(0.17468746696  + 914.72841698410 *T)
    sumY0  +=       3.161 * cos(4.04783639199  + 554.42567955090 *T)
    sumY0  +=       2.968 * cos(2.94519329004  + 1542.09553032591 *T)
    sumY0  +=       3.071 * cos(5.84592008895  + 1062.80714141041 *T)
    sumY0  +=       2.999 * cos(3.40007090655  + 132.82342352100 *T)
    sumY0  +=       3.514 * cos(1.73252446796  + 6355.13232062560 *T)
    sumY0  +=       2.745 * cos(3.12838183717  + 482.01351684860 *T)
    sumY0  +=       2.850 * cos(5.11696140522  + 395.66495315560 *T)
    sumY0  +=       2.663 * cos(2.78626662149  + 96.48367713221 *T)
    sumY0  +=       3.000 * cos(1.57450492122  + 519.85229016070 *T)
    sumY0  +=       2.655 * cos(2.76109188739  + 1160.58635552159 *T)
    sumY0  +=       3.001 * cos(2.18803399036  + 253.81481257341 *T)
    sumY0  +=       2.983 * cos(2.51014988868  + 9912.43086174999 *T)
    sumY0  +=       2.641 * cos(3.23932215146  + 487.39651292520 *T)
    sumY0  +=       3.712 * cos(3.03500052569  + 10026.82996866339 *T)
    sumY0  +=       3.062 * cos(2.34062617235  + 789.70430093210 *T)
    sumY0  +=       3.147 * cos(2.79398254369  + 394.70174530910 *T)
    sumY0  +=       2.712 * cos(5.80476756645  + 78.16111780620 *T)
    sumY0  +=       2.692 * cos(3.19752932766  + 161.16822906709 *T)
    sumY0  +=       3.303 * cos(0.22748739897  + 7.94806531530 *T)
    sumY0  +=       2.747 * cos(4.38283979645  + 1372.40097825130 *T)
    sumY0  +=       3.011 * cos(0.65384486578  + 1605.83142862930 *T)
    sumY0  +=       2.807 * cos(4.16656584606  + 208.12628695010 *T)
    sumY0  +=       2.631 * cos(4.42161511005  + 271.64973693240 *T)
    sumY0  +=       3.688 * cos(1.09167134977  + 26543.05632558499 *T)
    sumY0  +=       3.416 * cos(5.17742066298  + 331.07772159029 *T)
    sumY0  +=       2.951 * cos(1.21275667976  + 12564.91104475801 *T)
    sumY0  +=       2.727 * cos(5.15144703991  + 1064.07916579760 *T)
    sumY0  +=       2.902 * cos(4.64267920679  + 1540.87418217599 *T)
    sumY0  +=       2.728 * cos(3.02052878065  + 109.55636682560 *T)
    sumY0  +=       2.999 * cos(1.06420381994  + 219.27505307800 *T)
    sumY0  +=       2.764 * cos(2.64750487696  + 142.20583265029 *T)
    sumY0  +=       2.586 * cos(3.64905087080  + 959.81743305580 *T)
    sumY0  +=       2.719 * cos(5.16300330262  + 52325.61329776649 *T)
    sumY0  +=       3.042 * cos(4.27530437715  + 752.79510557170 *T)
    sumY0  +=       3.089 * cos(2.35849730561  + 93.28772917949 *T)
    sumY0  +=       3.647 * cos(0.50719381004  + 697.56335088530 *T)
    sumY0  +=       3.492 * cos(4.34377581989  + 580.91350537320 *T)
    sumY0  +=       3.597 * cos(4.95426722828  + 132.28090679980 *T)
    sumY0  +=       2.974 * cos(5.48085946209  + 444.58305662640 *T)
    sumY0  +=       2.572 * cos(1.10174431945  + 2117.52320515360 *T)
    sumY0  +=       2.572 * cos(4.16924943967  + 23.33205575259 *T)
    sumY0  +=       3.167 * cos(2.20892124981  + 12567.87999017460 *T)
    sumY0  +=       3.382 * cos(3.15537208847  + 10285.34201684520 *T)
    sumY0  +=       2.555 * cos(5.91884582039  + 678.11657286839 *T)
    sumY0  +=       3.300 * cos(3.72323989169  + 440.65090336330 *T)
    sumY0  +=       2.831 * cos(2.72460046701  + 540.01727499551 *T)
    sumY0  +=       3.189 * cos(2.12670165168  + 769.86363900301 *T)
    sumY0  +=       3.190 * cos(3.97099477941  + 2424.35384725460 *T)
    sumY0  +=       3.280 * cos(1.96587920993  + 86.38663049980 *T)
    sumY0  +=       2.780 * cos(2.14794520908  + 1586.34776735071 *T)
    sumY0  +=       2.594 * cos(0.18719015167  + 1373.88545095959 *T)
    sumY0  +=       2.782 * cos(5.73984608000  + 1268.99268984760 *T)
    sumY0  +=       3.010 * cos(5.61450100119  + 1543.84312759260 *T)
    sumY0  +=       3.022 * cos(2.75503960638  + 181.29958400710 *T)
    sumY0  +=       3.022 * cos(2.75503960638  + 180.81194904009 *T)
    sumY0  +=       2.833 * cos(0.53805108835  + 555.79827767460 *T)
    sumY0  +=       3.293 * cos(4.97275203910  + 463.49057128900 *T)
    sumY0  +=       3.316 * cos(0.94183619384  + 807.99667464080 *T)
    sumY0  +=       3.300 * cos(2.30840103920  + 649.80539721709 *T)
    sumY0  +=       2.983 * cos(1.81721249094  + 936.40161851410 *T)
    sumY0  +=       2.865 * cos(0.19981322997  + 2192.40786958489 *T)
    sumY0  +=       3.356 * cos(0.40284796347  + 247.58622872930 *T)
    sumY0  +=       2.506 * cos(0.59305762652  + 40.85816357090 *T)
    sumY0  +=       3.114 * cos(3.54141009377  + 1553.89199011520 *T)
    sumY0  +=       3.268 * cos(4.76214402026  + 737.66849419841 *T)
    sumY0  +=       3.100 * cos(2.50710858069  + 657.66721161581 *T)
    sumY0  +=       2.510 * cos(3.00502356004  + 20278.24855051219 *T)
    sumY0  +=       2.833 * cos(4.70043916327  + 539.81152431880 *T)
    sumY0  +=       2.554 * cos(0.54563041227  + 826.19529729470 *T)
    sumY0  +=       2.806 * cos(4.14737433673  + 351.94853520760 *T)
    sumY0  +=       2.633 * cos(2.39197384134  + 1448.66704952690 *T)
    sumY0  +=       2.482 * cos(3.92431211413  + 1159.10188281330 *T)
    sumY0  +=       3.088 * cos(2.06985743220  + 245.78624183590 *T)
    sumY0  +=       3.458 * cos(3.00992932151  + 809.48114734910 *T)
    sumY0  +=       3.385 * cos(5.68731744934  + 1056.44435393500 *T)
    sumY0  +=       3.050 * cos(0.67258736204  + 52174.56562792359 *T)
    sumY0  +=       2.571 * cos(3.71059042018  + 584.58502620540 *T)
    sumY0  +=       2.506 * cos(1.51980608116  + 381.88968804659 *T)
    sumY0  +=       2.718 * cos(0.91305266395  + 352.17228437681 *T)
    sumY0  +=       3.112 * cos(3.60340669115  + 481.12855369190 *T)
    sumY0  +=       2.427 * cos(2.03240412489  + 20275.27960509559 *T)
    sumY0  +=       2.660 * cos(4.68536158090  + 587.55397162200 *T)
    sumY0  +=       3.031 * cos(4.22810484249  + 106.62105130379 *T)
    sumY0  +=       2.876 * cos(5.82406245556  + 837.76226242039 *T)
    sumY0  +=       2.876 * cos(2.68246980196  + 838.24989738740 *T)
    sumY0  +=       2.493 * cos(3.29469616302  + 339.53023682000 *T)
    sumY0  +=       3.071 * cos(1.37054633718  + 264.74863825270 *T)
    sumY0  +=       2.956 * cos(5.93090344870  + 561.64229241569 *T)
    sumY0  +=       2.956 * cos(2.78931079511  + 562.12992738271 *T)
    sumY0  +=       2.675 * cos(0.11376507333  + 257.66872114670 *T)
    sumY0  +=       2.412 * cos(1.78724601033  + 537.15139544289 *T)
    sumY0  +=       2.425 * cos(1.27553210878  + 243.65407546620 *T)
    sumY0  +=       3.094 * cos(5.60755227933  + 911.75947156750 *T)
    sumY0  +=       2.916 * cos(0.94105407657  + 627.12329585830 *T)
    sumY0  +=       2.461 * cos(2.33753824934  + 672.69994689699 *T)
    sumY0  +=       3.109 * cos(4.52636545098  + 45.27795180100 *T)
    sumY0  +=       3.239 * cos(2.55766569454  + 412.33972771199 *T)
    sumY0  +=       2.930 * cos(6.24282243242  + 173.38288785421 *T)
    sumY0  +=       2.377 * cos(3.93847902994  + 1927.95543844240 *T)
    sumY0  +=       2.783 * cos(3.74679957070  + 888.96798775160 *T)
    sumY0  +=       2.572 * cos(4.97882149879  + 1536.46645603341 *T)
    sumY0  +=       2.447 * cos(3.55705970755  + 1327.57508673799 *T)
    sumY0  +=       2.533 * cos(4.52831931294  + 1330.54403215460 *T)
    sumY0  +=       2.951 * cos(1.23845421338  + 328.59641114070 *T)
    sumY0  +=       2.474 * cos(1.24533740369  + 1407.51018671831 *T)
    sumY0  +=       2.518 * cos(5.39696631071  + 175.57663362249 *T)
    sumY0  +=       2.678 * cos(2.58260670513  + 180.40581212981 *T)
    sumY0  +=       3.251 * cos(0.33431351861  + 83.41098743880 *T)
    sumY0  +=       2.327 * cos(2.23207080538  + 325.54922103430 *T)
    sumY0  +=       2.802 * cos(1.73977371146  + 121.81135315189 *T)
    sumY0  +=       2.675 * cos(3.92263515595  + 105.62421356250 *T)
    sumY0  +=       2.909 * cos(5.49280329616  + 651.18680406140 *T)
    sumY0  +=       2.358 * cos(2.87531802084  + 377.87821290640 *T)
    sumY0  +=       2.780 * cos(2.31107301213  + 291.29345690540 *T)
    sumY0  +=       2.876 * cos(4.06630905979  + 6813.01063256950 *T)
    sumY0  +=       2.523 * cos(4.12709915871  + 440.91153579420 *T)
    sumY0  +=       3.239 * cos(4.67199575771  + 724.84762821529 *T)
    sumY0  +=       2.532 * cos(0.78821708791  + 39.34006096781 *T)
    sumY0  +=       2.503 * cos(5.06436722887  + 360.30392097830 *T)
    sumY0  +=       2.568 * cos(1.92793293351  + 1699.52303398670 *T)
    sumY0  +=       2.570 * cos(0.58090295304  + 299.47327761669 *T)
    sumY0  +=       3.223 * cos(5.76095108778  + 442.39600850250 *T)
    sumY0  +=       2.316 * cos(1.18843210875  + 714.92270236481 *T)
    sumY0  +=       2.457 * cos(5.86455646124  + 552.82933225800 *T)
    sumY0  +=       2.319 * cos(2.34143811502  + 290.12449838219 *T)
    sumY0  +=       2.922 * cos(2.46474269442  + 1600.20235433680 *T)
    sumY0  +=       2.297 * cos(3.45720493767  + 1811.52934209949 *T)
    sumY0  +=       2.674 * cos(5.89642910933  + 519.26158932971 *T)
    sumY0  +=       2.585 * cos(5.75658839358  + 754.27957828001 *T)
    sumY0  +=       2.277 * cos(1.29593099122  + 386.10372559999 *T)
    sumY0  +=       2.326 * cos(4.96863117046  + 172.80463016199 *T)
    sumY0  +=       3.163 * cos(3.75842207356  + 476.18750059999 *T)
    sumY0  +=       3.127 * cos(1.03528434148  + 652.98679095480 *T)
    sumY0  +=       3.015 * cos(6.17163510089  + 3190.80541208169 *T)
    sumY0  +=       3.015 * cos(3.03004244730  + 3191.29304704871 *T)
    sumY0  +=       2.772 * cos(5.78976016729  + 431.08967580769 *T)
    sumY0  +=       2.329 * cos(6.09581509908  + 759.01488069520 *T)
    sumY0  +=       2.334 * cos(4.33637332299  + 837.97471074149 *T)
    sumY0  +=       2.991 * cos(2.07030361268  + 458.81015062800 *T)
    sumY0  +=       3.166 * cos(3.17988758658  + 26161.44408491669 *T)
    sumY0  +=       2.566 * cos(5.96622982807  + 63.23144838899 *T)
    sumY0  +=       2.282 * cos(0.94732597842  + 811.21744376770 *T)
    sumY0  +=       2.423 * cos(5.49223138244  + 103.10958916600 *T)
    sumY0  +=       3.125 * cos(6.04955881702  + 939.37056393069 *T)
    sumY0  +=       2.277 * cos(5.51962275937  + 0.27744737829 *T)
    sumY0  +=       2.982 * cos(2.21130422996  + 37.85558825950 *T)
    sumY0  +=       2.982 * cos(5.35289688355  + 37.36795329250 *T)
    sumY0  +=       2.437 * cos(5.09941116678  + 377.12075573661 *T)
    sumY0  +=       2.672 * cos(1.29742000970  + 412.07909528109 *T)
    sumY0  +=       2.289 * cos(5.75046347059  + 1258.69774910910 *T)
    sumY0  +=       2.259 * cos(3.10987273961  + 113.63153244060 *T)
    sumY0  +=       3.098 * cos(3.32366686883  + 612.42062897720 *T)
    sumY0  +=       3.098 * cos(3.32366686883  + 611.93299401020 *T)
    sumY0  +=       3.157 * cos(2.89245113423  + 3795.76561071060 *T)
    sumY0  +=       2.795 * cos(5.48481257493  + 10743.22032878910 *T)
    sumY0  +=       2.777 * cos(3.18566798610  + 211.09523236671 *T)
    sumY0  +=       3.088 * cos(4.00968174537  + 230.52650401861 *T)
    sumY0  +=       2.940 * cos(1.94569042371  + 714.66206993390 *T)
    sumY0  +=       2.685 * cos(1.40856509045  + 303.82362987761 *T)
    sumY0  +=       2.260 * cos(3.38291888798  + 124.20415195250 *T)
    sumY0  +=       3.092 * cos(4.98095101091  + 3414.15337004230 *T)
    sumY0  +=       3.015 * cos(0.01212812855  + 342.49918223660 *T)
    sumY0  +=       3.083 * cos(2.29263489637  + 206.95063078250 *T)
    sumY0  +=       3.083 * cos(2.29263489637  + 206.46299581550 *T)
    sumY0  +=       2.331 * cos(0.84374458161  + 820.56622300220 *T)
    sumY0  +=       2.712 * cos(2.90887297124  + 770.30665917500 *T)
    sumY0  +=       2.587 * cos(4.14114765790  + 636.20895053440 *T)
    sumY0  +=       2.587 * cos(4.14114765790  + 635.72131556739 *T)
    sumY0  +=       2.803 * cos(4.34014546433  + 25938.09612695609 *T)
    sumY0  +=       2.803 * cos(1.19855281074  + 25938.58376192310 *T)
    sumY0  +=       2.405 * cos(0.34118525545  + 488.60579898760 *T)
    sumY0  +=       2.359 * cos(4.55464439275  + 87.06795405600 *T)
    sumY0  +=       2.160 * cos(2.34244688851  + 199.52826724100 *T)
    sumY0  +=       2.362 * cos(0.79682610358  + 1222.03463196460 *T)
    sumY0  +=       2.362 * cos(3.93841875717  + 1222.52226693161 *T)
    sumY0  +=       2.740 * cos(5.56893792908  + 627.61093082531 *T)
    sumY0  +=       2.798 * cos(0.78287898333  + 771.34811171130 *T)
    sumY0  +=       2.798 * cos(0.78287898333  + 770.86047674430 *T)
    sumY0  +=       2.435 * cos(3.93525838301  + 350.08830211689 *T)
    sumY0  +=       2.357 * cos(6.18080965950  + 355.24180352990 *T)
    sumY0  +=       2.485 * cos(5.35449462342  + 142.41828097139 *T)
    sumY0  +=       2.521 * cos(2.60576764189  + 129.67565967810 *T)
    sumY0  +=       2.089 * cos(0.73778136703  + 74.58596519360 *T)
    sumY0  +=       2.302 * cos(4.80810165693  + 52177.04693837319 *T)
    sumY0  +=       2.302 * cos(1.66650900334  + 52177.53457334019 *T)
    sumY0  +=       2.619 * cos(3.19212100623  + 1321.68288788710 *T)
    sumY0  +=       2.572 * cos(4.73607789846  + 659.43031524749 *T)
    sumY0  +=       2.225 * cos(6.24656870358  + 59.52629766200 *T)
    sumY0  +=       2.835 * cos(1.73395299243  + 59.34747923569 *T)
    sumY0  +=       2.282 * cos(0.93137400045  + 295.91363136910 *T)
    sumY0  +=       1.992 * cos(1.74985561339  + 765.64079272899 *T)
    sumY0  +=       1.992 * cos(1.74985561339  + 766.12842769600 *T)
    sumY0  +=       2.574 * cos(4.89834107176  + 192.56022471960 *T)
    sumY0  +=       2.264 * cos(5.64468052990  + 75.05904594560 *T)
    sumY0  +=       2.089 * cos(4.42711005532  + 74.48959697400 *T)
    sumY0  +=       2.032 * cos(3.40351035240  + 215.68177693560 *T)
    sumY0  +=       1.982 * cos(5.16439207994  + 898.52921530720 *T)
    sumY0  +=       1.982 * cos(5.16439207994  + 899.01685027420 *T)
    sumY0  +=       2.374 * cos(0.01113742145  + 550.96909916730 *T)
    sumY0  +=       2.374 * cos(3.15273007504  + 551.45673413431 *T)
    sumY0  +=       2.352 * cos(4.77946960216  + 24.36220744081 *T)
    sumY0  +=       2.591 * cos(5.32795008432  + 367.72628451979 *T)
    sumY0  +=       2.230 * cos(6.23634089876  + 329.96900926440 *T)
    sumY0  +=       2.610 * cos(4.30928957119  + 66.15220969580 *T)
    sumY0  +=       2.610 * cos(1.16769691760  + 66.63984466281 *T)
    sumY0  +=       1.942 * cos(0.75979107585  + 74.69783977820 *T)
    sumY0  +=       2.217 * cos(4.28506837956  + 353.05724753350 *T)
    sumY0  +=       2.034 * cos(2.87672963893  + 527.96267490280 *T)
    sumY0  +=       2.034 * cos(6.01832229252  + 528.45030986980 *T)
    sumY0  +=       2.025 * cos(3.47385821354  + 94.98465020890 *T)
    sumY0  +=       2.220 * cos(1.63487182927  + 160.36507991500 *T)
    sumY0  +=       2.205 * cos(3.89758784429  + 460.78225830330 *T)
    sumY0  +=       2.205 * cos(0.75599519071  + 460.29462333630 *T)
    sumY0  +=       1.903 * cos(0.55549926481  + 170.51700830159 *T)
    sumY0  +=       1.903 * cos(3.69709191840  + 171.00464326860 *T)
    sumY0  +=       1.880 * cos(1.20742771402  + 845.83901726770 *T)
    sumY0  +=       1.879 * cos(3.10039964385  + 810.41429461559 *T)
    sumY0  +=       1.918 * cos(5.86821877824  + 980.42436087530 *T)
    sumY0  +=       1.918 * cos(5.86821877824  + 980.91199584231 *T)
    sumY0  +=       2.012 * cos(6.16028051115  + 482.97672469510 *T)
    sumY0  +=       2.408 * cos(2.08338219083  + 673.56008887950 *T)
    sumY0  +=       2.408 * cos(5.22497484442  + 673.07245391250 *T)
    sumY0  +=       1.917 * cos(6.07962338817  + 464.22677659940 *T)
    sumY0  +=       1.917 * cos(2.93803073458  + 464.71441156640 *T)
    sumY0  +=       2.523 * cos(5.39756661705  + 207.11738712130 *T)
    sumY0  +=       2.235 * cos(3.16893594529  + 156.28742217250 *T)
    sumY0  +=       1.889 * cos(1.49538325556  + 96.62918161160 *T)
    sumY0  +=       2.163 * cos(6.02567448655  + 1325.12740618320 *T)
    sumY0  +=       2.163 * cos(6.02567448655  + 1325.61504115020 *T)
    sumY0  +=       2.041 * cos(3.14152073195  + 753.38580640270 *T)
    sumY0  +=       2.385 * cos(2.09028108448  + 589.10213480510 *T)
    sumY0  +=       2.385 * cos(2.09028108448  + 589.58976977210 *T)
    sumY0  +=       2.320 * cos(6.09175102857  + 153.77279777600 *T)
    sumY0  +=       1.825 * cos(5.92052919225  + 241.62708603669 *T)
    sumY0  +=       1.938 * cos(1.19271280652  + 989.98558843089 *T)
    sumY0  +=       1.938 * cos(4.33430546011  + 990.47322339791 *T)
    sumY0  +=       1.801 * cos(3.16549584218  + 386.47623261550 *T)
    sumY0  +=       1.801 * cos(0.02390318859  + 386.96386758251 *T)
    sumY0  +=       1.870 * cos(5.04637132802  + 814.08332332030 *T)
    sumY0  +=       1.870 * cos(1.90477867443  + 813.59568835330 *T)
    sumY0  +=       1.820 * cos(1.38338010829  + 194.42045781031 *T)
    sumY0  +=       2.331 * cos(5.06651614946  + 229.20878764910 *T)
    sumY0  +=       1.846 * cos(3.96765232448  + 1310.63718762320 *T)
    sumY0  +=       1.813 * cos(5.03930311956  + 242.85092631409 *T)
    sumY0  +=       1.813 * cos(5.03930311956  + 243.33856128110 *T)
    sumY0  +=       2.125 * cos(2.47782176621  + 927.73416283579 *T)
    sumY0  +=       1.841 * cos(2.32220678128  + 114.15528942990 *T)
    sumY0  +=       1.770 * cos(1.07929229507  + 113.89465699899 *T)
    sumY0  +=       1.797 * cos(3.39032778968  + 286.35240381350 *T)
    sumY0  +=       2.348 * cos(5.43827517653  + 838.46234570850 *T)
    sumY0  +=       2.341 * cos(2.42376209355  + 80.47567191700 *T)
    sumY0  +=       2.214 * cos(0.68034717267  + 355.50492808829 *T)
    sumY0  +=       1.986 * cos(1.15722169964  + 180.55131660920 *T)
    sumY0  +=       1.742 * cos(4.55052960802  + 380.37158544351 *T)
    sumY0  +=       1.858 * cos(0.90088013499  + 1471.50920958009 *T)
    sumY0  +=       1.858 * cos(4.04247278858  + 1471.99684454710 *T)
    sumY0  +=       1.883 * cos(1.87812922018  + 102.76714273050 *T)
    sumY0  +=       2.209 * cos(6.09138093441  + 658.40341692619 *T)
    sumY0  +=       2.209 * cos(6.09138093441  + 658.89105189320 *T)
    sumY0  +=       1.942 * cos(3.82543532079  + 703.38936713389 *T)
    sumY0  +=       1.915 * cos(5.01403168451  + 423.17297965479 *T)
    sumY0  +=       1.915 * cos(1.87243903092  + 423.66061462181 *T)
    sumY0  +=       2.166 * cos(4.80147535867  + 79.53371592989 *T)
    sumY0  +=       1.755 * cos(3.84345466047  + 152.50077338879 *T)
    sumY0  +=       1.700 * cos(0.11878614099  + 50.15875869559 *T)
    sumY0  +=       2.051 * cos(4.57025334031  + 807.96661406080 *T)
    sumY0  +=       2.051 * cos(1.42866068672  + 808.45424902780 *T)
    sumY0  +=       1.839 * cos(4.18677723700  + 756.56720014041 *T)
    sumY0  +=       1.682 * cos(5.63413232593  + 1049.57688515010 *T)
    sumY0  +=       1.682 * cos(5.63413232593  + 1050.06452011710 *T)
    sumY0  +=       1.706 * cos(0.05840133316  + 832.34563644899 *T)
    sumY0  +=       1.706 * cos(3.19999398675  + 832.83327141600 *T)
    sumY0  +=       1.776 * cos(5.33275323151  + 1384.93115122349 *T)
    sumY0  +=       1.776 * cos(5.33275323151  + 1385.41878619050 *T)
    sumY0  +=       1.811 * cos(6.22065088973  + 454.55367445919 *T)
    sumY0  +=       1.811 * cos(3.07905823614  + 455.04130942620 *T)
    sumY0  +=       1.692 * cos(0.01726995267  + 271.16210196539 *T)
    sumY0  +=       2.206 * cos(1.46100694835  + 132.33578855399 *T)
    sumY0  +=       1.658 * cos(1.87948596313  + 184.33796539290 *T)
    sumY0  +=       1.875 * cos(0.56092257594  + 1282.58913653029 *T)
    sumY0  +=       1.875 * cos(0.56092257594  + 1283.07677149730 *T)
    sumY0  +=       1.686 * cos(2.02094532949  + 406.71065341949 *T)
    sumY0  +=       1.686 * cos(5.16253798308  + 407.19828838650 *T)
    sumY0  +=       2.083 * cos(2.68520521916  + 119.63885924300 *T)
    sumY0  +=       2.240 * cos(5.47514142442  + 404.37484744930 *T)
    sumY0  +=       1.952 * cos(4.95402575397  + 458.33457774851 *T)
    sumY0  +=       1.952 * cos(1.81243310038  + 457.84694278150 *T)
    sumY0  +=       2.288 * cos(0.80055559380  + 135.30473397060 *T)
    sumY0  +=       1.949 * cos(4.83910213339  + 1056.16916728910 *T)
    sumY0  +=       1.949 * cos(4.83910213339  + 1056.65680225610 *T)
    sumY0  +=       1.974 * cos(3.45809672503  + 269.93826168799 *T)
    sumY0  +=       1.974 * cos(3.45809672503  + 270.42589665501 *T)
    sumY0  +=       1.659 * cos(0.74121494237  + 1450.31158092960 *T)
    sumY0  +=       1.659 * cos(3.88280759596  + 1450.79921589661 *T)
    sumY0  +=       2.309 * cos(2.28366049597  + 74.79841351470 *T)
    sumY0  +=       1.644 * cos(5.02792890637  + 321.51649403470 *T)
    sumY0  +=       1.644 * cos(5.02792890637  + 322.00412900171 *T)
    sumY0  +=       1.656 * cos(2.46567744969  + 516.91697463890 *T)
    sumY0  +=       1.656 * cos(5.60727010328  + 517.40460960590 *T)
    sumY0  +=       2.034 * cos(5.98898663260  + 274.87931477991 *T)
    sumY0  +=       2.034 * cos(2.84739397901  + 274.39167981289 *T)
    sumY0  +=       2.154 * cos(1.72886095548  + 516.29264391310 *T)
    sumY0  +=       2.154 * cos(1.72886095548  + 515.80500894610 *T)
    sumY0  +=       1.711 * cos(3.10813260274  + 885.99904233499 *T)
    sumY0  +=       1.711 * cos(6.24972525633  + 886.48667730201 *T)
    sumY0  +=       1.826 * cos(4.02049514986  + 398.43695661610 *T)
    sumY0  +=       1.826 * cos(0.87890249627  + 398.92459158310 *T)
    sumY0  +=       1.966 * cos(5.67503368913  + 15.90969221110 *T)
    sumY0  +=       1.966 * cos(5.67503368913  + 16.39732717810 *T)
    sumY0  +=       1.678 * cos(5.26599059550  + 1182.67775604939 *T)
    sumY0  +=       1.678 * cos(5.26599059550  + 1183.16539101641 *T)
    sumY0  +=       2.098 * cos(0.13935063439  + 630.09224127489 *T)
    sumY0  +=       2.098 * cos(0.13935063439  + 630.57987624190 *T)
    sumY0  +=       2.239 * cos(0.96016185025  + 446.55516430170 *T)
    sumY0  +=       2.239 * cos(4.10175450384  + 446.06752933469 *T)
    sumY0  +=       1.889 * cos(2.64943395349  + 86.69544704049 *T)
    sumY0  +=       1.889 * cos(5.79102660707  + 87.18308200751 *T)
    sumY0  +=       1.759 * cos(1.85756489257  + 614.11755000660 *T)
    sumY0  +=       1.759 * cos(1.85756489257  + 613.62991503960 *T)
    sumY0  +=       1.673 * cos(1.68162249326  + 231.06902073981 *T)
    sumY0  +=       1.607 * cos(1.04950977043  + 412.12727939089 *T)
    sumY0  +=       1.607 * cos(1.04950977043  + 412.61491435790 *T)
    sumY0  +=       2.070 * cos(2.25718168745  + 824.23525170691 *T)
    sumY0  +=       1.835 * cos(5.52970711900  + 146.87169909629 *T)
    sumY0  +=       1.580 * cos(0.04314663008  + 2002.73703700969 *T)
    sumY0  +=       1.580 * cos(0.04314663008  + 2003.22467197670 *T)
    sumY0  +=       1.979 * cos(1.17491481301  + 407.23191828129 *T)
    sumY0  +=       1.979 * cos(4.31650746660  + 407.71955324831 *T)
    sumY0  +=       1.764 * cos(3.81236252275  + 544.75257741070 *T)
    sumY0  +=       1.815 * cos(3.68122227110  + 174.92224231670 *T)
    sumY0  +=       1.808 * cos(2.97279980018  + 454.77742362840 *T)
    sumY0  +=       1.808 * cos(6.11439245377  + 455.26505859540 *T)
    sumY0  +=       1.667 * cos(5.87424656754  + 409.67959883610 *T)
    sumY0  +=       1.667 * cos(5.87424656754  + 410.16723380310 *T)
    sumY0  +=       1.739 * cos(1.78551588590  + 271.42273439630 *T)
    sumY0  +=       1.739 * cos(4.92710853949  + 271.91036936330 *T)
    sumY0  +=       1.764 * cos(4.44257226642  + 625.42637482889 *T)
    sumY0  +=       1.764 * cos(4.44257226642  + 625.91400979590 *T)
    sumY0  +=       1.578 * cos(1.06601243126  + 1321.19525292009 *T)
    sumY0  +=       2.105 * cos(1.19642147071  + 348.12825652910 *T)
    sumY0  +=       1.602 * cos(5.65877015739  + 1133.91971127299 *T)
    sumY0  +=       1.602 * cos(5.65877015738  + 1134.40734624000 *T)
    sumY0  +=       1.690 * cos(5.41519122123  + 2.24955505361 *T)
    sumY0  +=       1.690 * cos(2.27359856764  + 1.76192008660 *T)
    sumY0  +=       1.611 * cos(4.78503604384  + 1442.23482608230 *T)
    sumY0  +=       1.611 * cos(1.64344339025  + 1442.72246104930 *T)
    sumY0  +=       1.553 * cos(1.99292461065  + 550.85722458270 *T)
    sumY0  +=       1.553 * cos(5.13451726423  + 551.34485954970 *T)
    sumY0  +=       1.801 * cos(1.87398401660  + 450.47276334980 *T)
    sumY0  +=       1.801 * cos(1.87398401659  + 450.96039831681 *T)
    sumY0  +=       2.000 * cos(2.66057926360  + 239.14577558710 *T)
    sumY0  +=       2.000 * cos(5.80217191719  + 238.65814062010 *T)
    sumY0  +=       1.610 * cos(4.21223376244  + 1274.72483000410 *T)
    sumY0  +=       1.610 * cos(1.07064110885  + 1275.21246497110 *T)
    sumY0  +=       1.669 * cos(4.52228317651  + 839.24673512870 *T)
    sumY0  +=       1.669 * cos(1.38069052292  + 839.73437009571 *T)
    sumY0  +=       1.618 * cos(3.36105446751  + 376.91500505990 *T)
    sumY0  +=       1.618 * cos(0.21946181392  + 377.40264002690 *T)
    sumY0  +=       1.986 * cos(0.62147303592  + 980.39073098050 *T)
    sumY0  +=       1.986 * cos(0.62147303592  + 979.90309601349 *T)
    sumY0  +=       1.523 * cos(1.37959889046  + 67.94549894479 *T)
    sumY0  +=       1.625 * cos(2.05986843470  + 72.94143379089 *T)
    sumY0  +=       1.699 * cos(4.63549638138  + 418.27761854520 *T)
    sumY0  +=       1.699 * cos(1.49390372779  + 418.76525351221 *T)
    sumY0  +=       1.699 * cos(0.14199812582  + 991.47006113920 *T)
    sumY0  +=       1.699 * cos(3.28359077941  + 991.95769610620 *T)
    sumY0  +=       2.140 * cos(6.17088765144  + 468.89264304540 *T)
    sumY0  +=       2.140 * cos(6.17088765144  + 469.38027801241 *T)
    sumY0  +=       1.943 * cos(5.41638718025  + 55.52720055719 *T)
    sumY0  +=       1.887 * cos(2.96965661566  + 250.92768116140 *T)
    sumY0  +=       1.887 * cos(6.11124926925  + 251.41531612840 *T)
    sumY0  +=       1.500 * cos(3.62083637810  + 63.60395540450 *T)
    sumY0  +=       1.595 * cos(1.27948001449  + 440.17533048380 *T)
    sumY0  +=       1.595 * cos(4.42107266808  + 439.68769551680 *T)
    sumY0  +=       1.499 * cos(2.15375666291  + 309.03450517230 *T)
    sumY0  +=       1.499 * cos(2.15375666291  + 309.52214013930 *T)
    sumY0  +=       1.554 * cos(3.36297518119  + 759.04851059000 *T)
    sumY0  +=       1.554 * cos(3.36297518119  + 759.53614555700 *T)
    sumY0  +=       1.834 * cos(4.64222921780  + 1.45310354590 *T)
    sumY0  +=       1.834 * cos(1.50063656421  + 1.94073851290 *T)
    sumY0  +=       2.054 * cos(4.13508560420  + 549.97226142600 *T)
    sumY0  +=       2.054 * cos(4.13508560420  + 549.48462645899 *T)
    sumY0  +=       1.488 * cos(4.40870494577  + 1095.36598450990 *T)
    sumY0  +=       1.488 * cos(4.40870494577  + 1095.85361947691 *T)
    sumY0  +=       1.484 * cos(2.39819072564  + 68.12431737110 *T)
    sumY0  +=       1.818 * cos(0.16664442254  + 587.61766209680 *T)
    sumY0  +=       1.818 * cos(3.30823707613  + 588.10529706381 *T)
    sumY0  +=       1.714 * cos(2.03603945431  + 245.03548231051 *T)
    sumY0  +=       1.714 * cos(2.03603945431  + 244.54784734349 *T)
    sumY0  +=       1.920 * cos(0.93463073151  + 675.04456158780 *T)
    sumY0  +=       1.920 * cos(4.07622338510  + 674.55692662080 *T)
    sumY0  +=       1.504 * cos(5.74149452616  + 833.79647926250 *T)
    sumY0  +=       2.044 * cos(5.51343159668  + 6662.95980046789 *T)
    sumY0  +=       2.044 * cos(5.51343159668  + 6663.44743543490 *T)
    sumY0  +=       1.648 * cos(1.91165234067  + 438.54209992920 *T)
    sumY0  +=       1.466 * cos(4.96970702490  + 304.36614659881 *T)
    sumY0  +=       1.472 * cos(3.31043387588  + 1257.45935461670 *T)
    sumY0  +=       1.472 * cos(3.31043387588  + 1257.94698958370 *T)
    sumY0  +=       1.615 * cos(0.76736560614  + 984.35651413839 *T)
    sumY0  +=       1.625 * cos(5.07920389470  + 949.19242391720 *T)
    sumY0  +=       1.625 * cos(1.93761124111  + 949.68005888420 *T)
    sumY0  +=       1.715 * cos(6.06338805381  + 349.61272923740 *T)
    sumY0  +=       1.931 * cos(0.83387866179  + 1819.88128359270 *T)
    sumY0  +=       1.450 * cos(2.21466410541  + 493.75930040060 *T)
    sumY0  +=       1.450 * cos(2.21466410541  + 493.27166543360 *T)
    sumY0  +=       1.436 * cos(4.35227364990  + 430.02340209721 *T)
    sumY0  +=       1.436 * cos(4.35227364990  + 429.53576713020 *T)
    sumY0  +=       1.945 * cos(6.20261907757  + 131.68139724819 *T)
    sumY0  +=       1.431 * cos(6.07196110025  + 157.39613449839 *T)
    sumY0  +=       1.662 * cos(0.09098469501  + 6209.53490664889 *T)
    sumY0  +=       1.662 * cos(3.23257734860  + 6210.02254161590 *T)
    sumY0  +=       1.819 * cos(5.07998451030  + 1008.73553652659 *T)
    sumY0  +=       1.819 * cos(1.93839185671  + 1009.22317149361 *T)
    sumY0  +=       1.518 * cos(2.72434442822  + 328.67822514530 *T)
    sumY0  +=       1.518 * cos(2.72434442822  + 329.16586011230 *T)
    sumY0  +=       1.638 * cos(3.00636571119  + 519.88592005549 *T)
    sumY0  +=       1.638 * cos(3.00636571119  + 520.37355502251 *T)
    sumY0  +=       1.680 * cos(6.03882975885  + 359.64952967250 *T)
    sumY0  +=       1.680 * cos(2.89723710526  + 360.13716463950 *T)
    sumY0  +=       1.843 * cos(5.67893032773  + 478.15960827531 *T)
    sumY0  +=       1.843 * cos(5.67893032773  + 477.67197330830 *T)
    sumY0  +=       1.489 * cos(3.54216170755  + 400.49087829599 *T)
    sumY0  +=       1.489 * cos(0.40056905396  + 400.97851326301 *T)
    sumY0  +=       1.752 * cos(0.16804844040  + 6271.78633224399 *T)
    sumY0  +=       1.752 * cos(0.16804844040  + 6272.27396721100 *T)
    sumY0  +=       1.895 * cos(2.99891680203  + 1055.17232954780 *T)
    sumY0  +=       1.895 * cos(2.99891680203  + 1054.68469458080 *T)
    sumY0  +=       1.717 * cos(5.16735204147  + 536.56069461190 *T)
    sumY0  +=       1.758 * cos(5.73369868016  + 1095.05047032480 *T)
    sumY0  +=       1.758 * cos(5.73369868016  + 1094.56283535780 *T)
    sumY0  +=       1.721 * cos(5.73061288107  + 104.33342944339 *T)
    sumY0  +=       1.713 * cos(6.04365112420  + 621.49422156580 *T)
    sumY0  +=       1.460 * cos(1.87863990334  + 341.12658411290 *T)
    sumY0  +=       1.460 * cos(1.87863990334  + 340.63894914589 *T)
    sumY0  +=       1.780 * cos(3.52262570474  + 1668.83361374980 *T)
    sumY0  +=       1.780 * cos(0.38103305115  + 1668.34597878280 *T)
    sumY0  +=       1.401 * cos(3.66161518331  + 451.48415530609 *T)
    sumY0  +=       1.401 * cos(0.52002252972  + 451.97179027311 *T)
    sumY0  +=       1.394 * cos(2.79876022404  + 252.41215386970 *T)
    sumY0  +=       1.942 * cos(0.65443948228  + 10593.16949668749 *T)
    sumY0  +=       1.942 * cos(0.65443948228  + 10593.65713165450 *T)
    sumY0  +=       1.427 * cos(3.71549072767  + 228.21194990781 *T)
    sumY0  +=       1.427 * cos(3.71549072767  + 227.72431494079 *T)
    sumY0  +=       1.395 * cos(1.68241228451  + 697.76029284139 *T)
    sumY0  +=       1.395 * cos(1.68241228451  + 698.24792780840 *T)
    sumY0  +=       1.403 * cos(4.02233962356  + 281.24459438279 *T)
    sumY0  +=       1.403 * cos(4.02233962356  + 281.73222934980 *T)
    sumY0  +=       1.507 * cos(4.90762022605  + 193.89919294850 *T)
    sumY0  +=       1.401 * cos(3.75133437627  + 47.45044570990 *T)
    sumY0  +=       1.458 * cos(1.30164002580  + 214.19730422731 *T)
    sumY0  +=       1.408 * cos(4.84929178616  + 105.76971804189 *T)
    sumY0  +=       1.545 * cos(4.93926208865  + 912.54386098770 *T)
    sumY0  +=       1.545 * cos(1.79766943506  + 913.03149595471 *T)
    sumY0  +=       1.780 * cos(5.55379390271  + 623.19114259520 *T)
    sumY0  +=       1.780 * cos(5.55379390271  + 623.67877756220 *T)
    sumY0  +=       1.421 * cos(0.28471608319  + 1349.50642857140 *T)
    sumY0  +=       1.421 * cos(3.42630873678  + 1349.99406353840 *T)
    sumY0  +=       1.421 * cos(3.58407040366  + 1447.18257681859 *T)
    sumY0  +=       1.421 * cos(3.58407040366  + 1447.67021178560 *T)

    # URANUS Y Coord
    sumY1  = 0
    sumY1  +=  2225113.750 * cos(0.23889049392  + 0.24381748350 *T)
    sumY1  +=  738897.387 * cos(4.43927593626  + 149.80701461810 *T)
    sumY1  +=  239408.407 * cos(3.76717540218  + 73.54094334250 *T)
    sumY1  +=  229359.688 * cos(0.91052914986  + 76.50988875911 *T)
    sumY1  +=  110583.560 * cos(4.00448071218  + 11.28951774740 *T)
    sumY1  +=   95657.865 * cos(5.06686614467  + 63.97971578690 *T)
    sumY1  +=   81488.096 * cos(5.92277112536  + 86.07111631471 *T)
    sumY1  +=   45577.241 * cos(0.72706305435  + 138.76131435421 *T)
    sumY1  +=   51354.528 * cos(0.61825536711  + 224.58861318540 *T)
    sumY1  +=   38568.000 * cos(5.01914859027  + 71.09326278771 *T)
    sumY1  +=   37681.305 * cos(4.78820636032  + 75.02541605080 *T)
    sumY1  +=   41920.157 * cos(5.73652625878  + 74.53778108379 *T)
    sumY1  +=   36154.364 * cos(5.94864908914  + 78.95756931390 *T)
    sumY1  +=   21671.035 * cos(3.36593709546  + 151.29148732640 *T)
    sumY1  +=   19425.371 * cos(6.01825386971  + 77.99436146740 *T)
    sumY1  +=   17377.170 * cos(4.95887394628  + 72.05647063421 *T)
    sumY1  +=   15101.686 * cos(3.96616276000  + 4.17597074660 *T)
    sumY1  +=    7734.175 * cos(0.04263276670  + 71.84402231310 *T)
    sumY1  +=    7431.784 * cos(4.63221839509  + 78.20680978850 *T)
    sumY1  +=    6987.545 * cos(0.83638484964  + 145.87486135500 *T)
    sumY1  +=    7298.813 * cos(0.66426542553  + 3.21276290011 *T)
    sumY1  +=    7589.358 * cos(1.19697295521  + 148.32254190981 *T)
    sumY1  +=    4377.580 * cos(2.17199709718  + 160.85271488200 *T)
    sumY1  +=    3947.976 * cos(3.17415954213  + 65.46418849521 *T)
    sumY1  +=    4122.455 * cos(6.19822707406  + 22.33521801130 *T)
    sumY1  +=    2797.759 * cos(3.33611153878  + 213.54291292150 *T)
    sumY1  +=    2617.647 * cos(1.79307323508  + 9.80504503910 *T)
    sumY1  +=    2637.599 * cos(2.97095004820  + 12.77399045571 *T)
    sumY1  +=    3175.281 * cos(3.08123141539  + 299.37021175271 *T)
    sumY1  +=    2245.280 * cos(5.07246912294  + 52.44638055600 *T)
    sumY1  +=    2227.953 * cos(4.89367152704  + 87.55558902300 *T)
    sumY1  +=    2822.936 * cos(1.83097011748  + 84.58664360640 *T)
    sumY1  +=    2517.942 * cos(1.26953424709  + 18.40306474821 *T)
    sumY1  +=    2220.819 * cos(3.21400436294  + 72.57773549600 *T)
    sumY1  +=    1622.656 * cos(2.18707111524  + 153.73916788121 *T)
    sumY1  +=    1925.669 * cos(0.82858902246  + 39.86132582961 *T)
    sumY1  +=    1555.527 * cos(2.56671995255  + 74.06220820430 *T)
    sumY1  +=    1901.819 * cos(4.78958433549  + 127.71561409031 *T)
    sumY1  +=    1598.284 * cos(4.79586647820  + 79.47883417571 *T)
    sumY1  +=    1476.433 * cos(2.10157835287  + 75.98862389731 *T)
    sumY1  +=    1532.412 * cos(2.14699190783  + 152.77596003471 *T)
    sumY1  +=    1440.833 * cos(0.60856377911  + 127.22797912329 *T)
    sumY1  +=    1745.915 * cos(1.67843758301  + 77.47309660560 *T)
    sumY1  +=    1713.773 * cos(1.82301190922  + 223.10414047710 *T)
    sumY1  +=    1706.344 * cos(5.82533914555  + 226.07308589371 *T)
    sumY1  +=    1357.973 * cos(6.18089361282  + 70.57199792590 *T)
    sumY1  +=    1738.426 * cos(3.82544050989  + 146.83806920150 *T)
    sumY1  +=    1467.571 * cos(5.10136854333  + 2.93757625420 *T)
    sumY1  +=    1179.806 * cos(3.56840380352  + 220.65645992230 *T)
    sumY1  +=    1236.409 * cos(5.69350914104  + 4.69723560840 *T)
    sumY1  +=    1067.152 * cos(6.09416202691  + 62.49524307860 *T)
    sumY1  +=     977.123 * cos(5.73086540533  + 74.91354146621 *T)
    sumY1  +=     977.123 * cos(5.23172233785  + 75.13729063540 *T)
    sumY1  +=     937.717 * cos(1.52410870371  + 56.86616878610 *T)
    sumY1  +=    1087.876 * cos(3.42153780818  + 131.64776735340 *T)
    sumY1  +=     890.295 * cos(2.70229070005  + 426.84200835950 *T)
    sumY1  +=    1040.906 * cos(3.84755403081  + 110.18950627200 *T)
    sumY1  +=     935.342 * cos(1.92101982937  + 149.31937965110 *T)
    sumY1  +=     862.611 * cos(2.93877651502  + 52.93401552300 *T)
    sumY1  +=     777.859 * cos(1.44001318420  + 63.49208081989 *T)
    sumY1  +=     746.826 * cos(3.21375563841  + 93.18466331550 *T)
    sumY1  +=     747.709 * cos(4.36891744549  + 206.42936592071 *T)
    sumY1  +=     648.200 * cos(4.91583959018  + 80.44204202221 *T)
    sumY1  +=     640.350 * cos(1.39049394918  + 1.24065522479 *T)
    sumY1  +=     794.750 * cos(1.71892347867  + 7.35736448430 *T)
    sumY1  +=     605.254 * cos(4.81788763333  + 1059.62574767270 *T)
    sumY1  +=     615.326 * cos(0.13279791934  + 10.80188278040 *T)
    sumY1  +=     569.613 * cos(6.19438038601  + 73.05330837550 *T)
    sumY1  +=     665.795 * cos(1.92703769837  + 97.11681657861 *T)
    sumY1  +=     575.875 * cos(0.85800317726  + 522.82123557730 *T)
    sumY1  +=     563.041 * cos(0.34515637363  + 537.04832957890 *T)
    sumY1  +=     493.956 * cos(1.93216860314  + 137.27684164590 *T)
    sumY1  +=     477.372 * cos(5.87004905618  + 41.34579853790 *T)
    sumY1  +=     445.046 * cos(2.08028645037  + 2.69149803831 *T)
    sumY1  +=     441.333 * cos(5.78719554674  + 67.91186905000 *T)
    sumY1  +=     413.028 * cos(2.21824858776  + 66.94866120350 *T)
    sumY1  +=     408.718 * cos(5.05109929565  + 82.13896305161 *T)
    sumY1  +=     393.354 * cos(0.01753896463  + 15.22167101050 *T)
    sumY1  +=     347.208 * cos(2.52660613312  + 146.62562088040 *T)
    sumY1  +=     350.848 * cos(4.64021047834  + 76.02225379210 *T)
    sumY1  +=     311.934 * cos(6.04641128519  + 1.72829019180 *T)
    sumY1  +=     307.527 * cos(3.13299616747  + 202.49721265760 *T)
    sumY1  +=     343.437 * cos(2.85099749738  + 142.69346761730 *T)
    sumY1  +=     291.412 * cos(2.84716827670  + 5.66044345490 *T)
    sumY1  +=     327.071 * cos(4.27135603462  + 113.15845168861 *T)
    sumY1  +=     283.053 * cos(2.69361373135  + 138.27367938720 *T)
    sumY1  +=     276.688 * cos(1.94716640065  + 70.60562782070 *T)
    sumY1  +=     279.650 * cos(2.52443983499  + 83.10217089810 *T)
    sumY1  +=     294.524 * cos(5.21442700405  + 145.35359649321 *T)
    sumY1  +=     254.666 * cos(4.68346860260  + 235.63431344930 *T)
    sumY1  +=     294.078 * cos(2.63162352478  + 33.43580002939 *T)
    sumY1  +=     275.813 * cos(2.69692194100  + 56.37853381910 *T)
    sumY1  +=     255.740 * cos(2.28857138569  + 265.74547599400 *T)
    sumY1  +=     273.184 * cos(0.38956727235  + 36.89238041300 *T)
    sumY1  +=     246.560 * cos(3.74959440503  + 159.36824217371 *T)
    sumY1  +=     235.614 * cos(2.06048819267  + 20.85074530300 *T)
    sumY1  +=     217.300 * cos(2.68296221034  + 152.98840835580 *T)
    sumY1  +=     233.552 * cos(5.10562451416  + 108.70503356371 *T)
    sumY1  +=     209.062 * cos(3.07234563267  + 140.24578706250 *T)
    sumY1  +=     229.061 * cos(4.38926682332  + 194.89603068979 *T)
    sumY1  +=     197.502 * cos(0.42703110531  + 190.96387742670 *T)
    sumY1  +=     194.913 * cos(3.89875917751  + 209.12312469139 *T)
    sumY1  +=     192.549 * cos(5.17184042822  + 23.81969071961 *T)
    sumY1  +=     228.047 * cos(4.16747124114  + 297.88573904440 *T)
    sumY1  +=     167.442 * cos(1.30512782686  + 134.34152612410 *T)
    sumY1  +=     169.219 * cos(5.32760035006  + 59.55992755679 *T)
    sumY1  +=     182.817 * cos(4.71950113329  + 134.82916109110 *T)
    sumY1  +=     154.035 * cos(5.62362511366  + 288.32451148881 *T)
    sumY1  +=     165.598 * cos(4.46093635884  + 6.18170831670 *T)
    sumY1  +=     153.435 * cos(1.12954247651  + 162.33718759030 *T)
    sumY1  +=     191.930 * cos(5.54402688683  + 374.15181032000 *T)
    sumY1  +=     154.136 * cos(1.68406483408  + 14.25846316400 *T)
    sumY1  +=     152.205 * cos(2.69999688221  + 120.11443212250 *T)
    sumY1  +=     179.814 * cos(2.43767426020  + 60.04756252381 *T)
    sumY1  +=     143.557 * cos(4.23090875226  + 212.05844021321 *T)
    sumY1  +=     141.934 * cos(6.02296068352  + 69.60879007940 *T)
    sumY1  +=     140.696 * cos(3.78608728509  + 29.44876501211 *T)
    sumY1  +=     140.277 * cos(3.58845867295  + 80.96330688400 *T)
    sumY1  +=     134.345 * cos(5.14541559038  + 39.60069339870 *T)
    sumY1  +=     148.026 * cos(1.44796277531  + 186.45557754760 *T)
    sumY1  +=     138.037 * cos(0.16623715920  + 3.68833577960 *T)
    sumY1  +=     122.748 * cos(2.86265578538  + 129.20008679860 *T)
    sumY1  +=     132.977 * cos(5.88158107492  + 68.80564092731 *T)
    sumY1  +=     141.987 * cos(3.15548842991  + 114.64292439690 *T)
    sumY1  +=     132.318 * cos(2.36034544105  + 90.00326957780 *T)
    sumY1  +=     116.010 * cos(1.93581371352  + 33.38091827520 *T)
    sumY1  +=     119.775 * cos(1.10226576648  + 69.08752521760 *T)
    sumY1  +=     120.027 * cos(2.55359011727  + 305.10235190919 *T)
    sumY1  +=     117.279 * cos(5.41876945206  + 277.27881122490 *T)
    sumY1  +=     120.086 * cos(1.31242375861  + 50.64639366260 *T)
    sumY1  +=     109.599 * cos(3.65947356465  + 89.04006173130 *T)
    sumY1  +=     121.041 * cos(0.94635029675  + 154.26043274300 *T)
    sumY1  +=     118.779 * cos(5.08108465545  + 81.24519117430 *T)
    sumY1  +=     103.250 * cos(2.14332472656  + 144.39038864671 *T)
    sumY1  +=     118.351 * cos(2.99239145613  + 41.40068029210 *T)
    sumY1  +=     116.406 * cos(4.85267098663  + 116.18227885939 *T)
    sumY1  +=     136.248 * cos(1.02614107930  + 45.33283355520 *T)
    sumY1  +=     104.859 * cos(1.08415262437  + 54.41848823131 *T)
    sumY1  +=      96.225 * cos(6.02638036659  + 50.96190784770 *T)
    sumY1  +=     116.279 * cos(4.59183961032  + 227.55755860200 *T)
    sumY1  +=     122.913 * cos(2.00108805237  + 300.85468446100 *T)
    sumY1  +=      93.159 * cos(5.79382600147  + 110.45013870291 *T)
    sumY1  +=     102.738 * cos(3.77683090065  + 152.25469517290 *T)
    sumY1  +=      86.724 * cos(0.10102936322  + 7.83293736379 *T)
    sumY1  +=      81.743 * cos(4.69965325972  + 228.52076644850 *T)
    sumY1  +=      84.795 * cos(4.63965252699  + 42.83027124621 *T)
    sumY1  +=      75.252 * cos(1.65398168749  + 2.20386307129 *T)
    sumY1  +=      74.096 * cos(4.96998649948  + 148.84380677161 *T)
    sumY1  +=      74.141 * cos(4.55571799271  + 150.77022246460 *T)
    sumY1  +=      80.549 * cos(3.37285893842  + 99.40443843901 *T)
    sumY1  +=      73.326 * cos(1.94650353151  + 640.14110379750 *T)
    sumY1  +=      81.866 * cos(3.90328813556  + 31.95132732110 *T)
    sumY1  +=      73.049 * cos(1.65979346411  + 9.31741007210 *T)
    sumY1  +=      74.334 * cos(3.16185820508  + 8.32057233081 *T)
    sumY1  +=      73.323 * cos(3.78688502488  + 95.63234387030 *T)
    sumY1  +=      69.283 * cos(0.88641793058  + 120.60206708950 *T)
    sumY1  +=      67.771 * cos(4.95516064946  + 16.91859203990 *T)
    sumY1  +=      71.013 * cos(3.42555712405  + 24.13520490470 *T)
    sumY1  +=      75.802 * cos(3.00563101406  + 36.40474544600 *T)
    sumY1  +=      79.127 * cos(5.99063673101  + 5.97595764000 *T)
    sumY1  +=      72.170 * cos(3.97968852101  + 404.26297286470 *T)
    sumY1  +=      67.065 * cos(3.27129691616  + 28.81562556571 *T)
    sumY1  +=      72.371 * cos(0.35418537369  + 230.32075334190 *T)
    sumY1  +=      86.876 * cos(3.24212615152  + 419.72846135871 *T)
    sumY1  +=      62.861 * cos(0.14858922104  + 202.00957769060 *T)
    sumY1  +=      70.935 * cos(1.10294734261  + 183.48663213100 *T)
    sumY1  +=      69.304 * cos(6.08426855707  + 333.41352756050 *T)
    sumY1  +=      60.612 * cos(2.74156472861  + 116.66991382640 *T)
    sumY1  +=      60.708 * cos(4.27659252807  + 70.35954960481 *T)
    sumY1  +=      70.577 * cos(2.80558133688  + 70.08436295889 *T)
    sumY1  +=      60.413 * cos(1.36799815983  + 149.91888920271 *T)
    sumY1  +=      56.872 * cos(2.36058803010  + 53.93085326429 *T)
    sumY1  +=      56.665 * cos(3.54220331724  + 79.69128249680 *T)
    sumY1  +=      56.672 * cos(5.25529040442  + 73.32849502140 *T)
    sumY1  +=      53.385 * cos(2.11007967538  + 329.48137429739 *T)
    sumY1  +=      60.881 * cos(5.83277800264  + 126.23114138200 *T)
    sumY1  +=      55.720 * cos(2.56360473726  + 76.72233708021 *T)
    sumY1  +=      55.712 * cos(4.81842861613  + 157.88376946540 *T)
    sumY1  +=      60.449 * cos(5.61300042131  + 347.64062156210 *T)
    sumY1  +=      51.883 * cos(5.93550310468  + 147.35933406330 *T)
    sumY1  +=      53.514 * cos(2.91453252731  + 433.95555536030 *T)
    sumY1  +=      49.248 * cos(1.51991875678  + 68.43313391180 *T)
    sumY1  +=      48.173 * cos(2.90644675264  + 128.71245183160 *T)
    sumY1  +=      57.429 * cos(0.17436946160  + 19.88753745650 *T)
    sumY1  +=      45.855 * cos(0.26413680268  + 8.10812400971 *T)
    sumY1  +=      45.035 * cos(5.61065437203  + 6.86972951729 *T)
    sumY1  +=      45.638 * cos(2.94030988664  + 415.30867312860 *T)
    sumY1  +=      43.941 * cos(2.65210291918  + 75.77617557621 *T)
    sumY1  +=      43.941 * cos(2.02730403124  + 74.27465652540 *T)
    sumY1  +=      52.767 * cos(2.76209553241  + 147.83490694279 *T)
    sumY1  +=      40.257 * cos(1.13787399944  + 155.22364058950 *T)
    sumY1  +=      42.861 * cos(0.04761592814  + 107.22056085540 *T)
    sumY1  +=      39.791 * cos(4.36382280015  + 258.63192899320 *T)
    sumY1  +=      39.522 * cos(6.00506578609  + 98.91680347199 *T)
    sumY1  +=      39.676 * cos(4.00856971667  + 276.79117625789 *T)
    sumY1  +=      41.680 * cos(3.13707624385  + 272.85902299479 *T)
    sumY1  +=      49.985 * cos(1.44333404289  + 125.74350641500 *T)
    sumY1  +=      39.779 * cos(0.19499431864  + 143.86912378490 *T)
    sumY1  +=      47.193 * cos(2.77004558406  + 67.16110952460 *T)
    sumY1  +=      42.716 * cos(5.17138594591  + 1589.31671276730 *T)
    sumY1  +=      36.533 * cos(3.11312786761  + 81.61769818981 *T)
    sumY1  +=      46.283 * cos(2.14611986732  + 62.00760811160 *T)
    sumY1  +=      36.873 * cos(0.60121789532  + 46.71424039951 *T)
    sumY1  +=      35.335 * cos(3.24859872841  + 135.79236893760 *T)
    sumY1  +=      36.570 * cos(5.69158157354  + 167.96626188280 *T)
    sumY1  +=      33.878 * cos(0.99777389192  + 14.47091148511 *T)
    sumY1  +=      33.444 * cos(2.15331669456  + 149.69514003350 *T)
    sumY1  +=      40.027 * cos(0.39160085220  + 187.94005025590 *T)
    sumY1  +=      33.511 * cos(1.05857849175  + 61.01077037031 *T)
    sumY1  +=      37.402 * cos(4.92427573893  + 259.75270340660 *T)
    sumY1  +=      34.209 * cos(1.81074999609  + 529.93478257810 *T)
    sumY1  +=      31.617 * cos(4.39624200688  + 224.10097821840 *T)
    sumY1  +=      42.780 * cos(6.01622236636  + 58.35064149440 *T)
    sumY1  +=      30.892 * cos(0.16528925233  + 1052.51220067191 *T)
    sumY1  +=      30.724 * cos(4.12747584944  + 103.33659170210 *T)
    sumY1  +=      37.429 * cos(5.50521421633  + 74.13164417351 *T)
    sumY1  +=      37.429 * cos(5.45737352685  + 75.91918792810 *T)
    sumY1  +=      30.708 * cos(4.96119928476  + 72.73779419041 *T)
    sumY1  +=      30.708 * cos(6.00138845842  + 77.31303791120 *T)
    sumY1  +=      35.364 * cos(1.85148506912  + 82.88972257700 *T)
    sumY1  +=      28.175 * cos(5.61993541202  + 295.43805848960 *T)
    sumY1  +=      29.578 * cos(0.77172005313  + 1066.73929467350 *T)
    sumY1  +=      29.606 * cos(1.50195344116  + 121.23520653590 *T)
    sumY1  +=      28.770 * cos(1.22525548779  + 6.34846465550 *T)
    sumY1  +=      32.637 * cos(4.39471635403  + 155.53915477459 *T)
    sumY1  +=      31.706 * cos(0.35163342435  + 38.37685312130 *T)
    sumY1  +=      26.723 * cos(3.18108685366  + 43.53284666180 *T)
    sumY1  +=      27.442 * cos(1.45908203944  + 220.13519506050 *T)
    sumY1  +=      36.094 * cos(1.10822042269  + 340.52707456129 *T)
    sumY1  +=      28.308 * cos(3.44249690833  + 85.58348134769 *T)
    sumY1  +=      30.388 * cos(1.47086782147  + 184.97110483931 *T)
    sumY1  +=      33.909 * cos(4.45538444508  + 171.89841514590 *T)
    sumY1  +=      28.512 * cos(4.03769470105  + 51.44954281471 *T)
    sumY1  +=      28.765 * cos(4.21609406946  + 64.97655352819 *T)
    sumY1  +=      31.960 * cos(1.53179865789  + 221.61966776881 *T)
    sumY1  +=      26.763 * cos(5.08057063209  + 217.47506618460 *T)
    sumY1  +=      24.737 * cos(1.03877406496  + 47.93808067690 *T)
    sumY1  +=      31.097 * cos(0.82256228232  + 98.60128928690 *T)
    sumY1  +=      23.919 * cos(4.34411138178  + 219.17198721400 *T)
    sumY1  +=      29.056 * cos(2.95564309290  + 106.73292588839 *T)
    sumY1  +=      23.488 * cos(2.14499589339  + 28.06735816780 *T)
    sumY1  +=      23.868 * cos(4.71661725014  + 57.49930823250 *T)
    sumY1  +=      26.140 * cos(5.22576075259  + 91.70019060721 *T)
    sumY1  +=      24.837 * cos(1.12156921102  + 25.51661174900 *T)
    sumY1  +=      28.239 * cos(3.07980893828  + 131.16013238639 *T)
    sumY1  +=      25.242 * cos(1.50128865984  + 471.93102443120 *T)
    sumY1  +=      23.194 * cos(4.62622223092  + 553.82616999929 *T)
    sumY1  +=      23.342 * cos(2.85807908852  + 61.69209392650 *T)
    sumY1  +=      23.751 * cos(5.67002669100  + 542.78046973539 *T)
    sumY1  +=      22.911 * cos(6.01222977731  + 234.14984074100 *T)
    sumY1  +=      22.260 * cos(4.57484369104  + 227.77000692311 *T)
    sumY1  +=      21.787 * cos(3.04663671917  + 19.36627259471 *T)
    sumY1  +=      20.991 * cos(2.16357529707  + 316.63568714010 *T)
    sumY1  +=      24.652 * cos(0.81742897132  + 529.44714761109 *T)
    sumY1  +=      23.384 * cos(0.19757750866  + 254.69977573010 *T)
    sumY1  +=      21.352 * cos(3.24093992707  + 7.66618102501 *T)
    sumY1  +=      20.300 * cos(3.65352107570  + 467.99887116810 *T)
    sumY1  +=      21.096 * cos(4.52992568399  + 141.94270809190 *T)
    sumY1  +=      19.957 * cos(3.38374254974  + 209.61075965840 *T)
    sumY1  +=      22.156 * cos(4.08770290595  + 68.59989025060 *T)
    sumY1  +=      27.098 * cos(4.10293359896  + 111.67397898031 *T)
    sumY1  +=      26.932 * cos(4.47193756936  + 35.40790770471 *T)
    sumY1  +=      23.041 * cos(1.03752917223  + 486.15811843280 *T)
    sumY1  +=      23.558 * cos(5.58348547213  + 195.38366565681 *T)
    sumY1  +=      19.352 * cos(4.03289384970  + 17.02165790390 *T)
    sumY1  +=      18.710 * cos(0.49922183447  + 69.91760662011 *T)
    sumY1  +=      18.292 * cos(2.17112401323  + 46.45360796860 *T)
    sumY1  +=      18.185 * cos(6.24789116958  + 92.55152386910 *T)
    sumY1  +=      20.489 * cos(1.98855828236  + 116.12739710521 *T)
    sumY1  +=      18.124 * cos(1.81380642180  + 88.35873817511 *T)
    sumY1  +=      17.311 * cos(1.53887757294  + 106.51798543980 *T)
    sumY1  +=      19.722 * cos(5.08964188825  + 296.40126633610 *T)
    sumY1  +=      17.322 * cos(2.01071283768  + 100.62827871640 *T)
    sumY1  +=      17.131 * cos(3.14636583407  + 99.14380600810 *T)
    sumY1  +=      21.450 * cos(1.26935358079  + 191.45151239371 *T)
    sumY1  +=      16.758 * cos(5.41014316701  + 210.57396750490 *T)
    sumY1  +=      17.082 * cos(4.18018061343  + 80.13322548150 *T)
    sumY1  +=      19.774 * cos(2.05712922006  + 80.75755620730 *T)
    sumY1  +=      18.159 * cos(1.66942745387  + 151.50393564750 *T)
    sumY1  +=      16.166 * cos(6.00467687020  + 397.14942586389 *T)
    sumY1  +=      16.332 * cos(2.08638656812  + 94.66913602380 *T)
    sumY1  +=      17.336 * cos(1.91313128591  + 13.08950464080 *T)
    sumY1  +=      16.300 * cos(0.86950603727  + 23.87457247379 *T)
    sumY1  +=      18.165 * cos(0.89317841037  + 198.07742442750 *T)
    sumY1  +=      16.506 * cos(3.12271859353  + 73.38088464811 *T)
    sumY1  +=      16.506 * cos(1.55668384247  + 76.66994745350 *T)
    sumY1  +=      18.503 * cos(4.36296262136  + 156.02678974161 *T)
    sumY1  +=      17.934 * cos(2.28460392259  + 182.00215942271 *T)
    sumY1  +=      15.839 * cos(1.55450871101  + 13.73719830221 *T)
    sumY1  +=      15.946 * cos(6.02088196247  + 55.90296093960 *T)
    sumY1  +=      16.546 * cos(0.84116196958  + 150.80385235940 *T)
    sumY1  +=      17.819 * cos(4.98834752372  + 221.40721944771 *T)
    sumY1  +=      15.752 * cos(2.32623822944  + 67.60305250931 *T)
    sumY1  +=      21.162 * cos(0.30009257573  + 372.66733761171 *T)
    sumY1  +=      19.517 * cos(2.68086406105  + 479.04457143200 *T)
    sumY1  +=      17.205 * cos(4.28439583468  + 213.05527795450 *T)
    sumY1  +=      15.267 * cos(4.22214122299  + 68.12431737110 *T)
    sumY1  +=      15.236 * cos(2.15369095754  + 30.30009827400 *T)
    sumY1  +=      15.262 * cos(3.36861495242  + 67.42423408300 *T)
    sumY1  +=      14.562 * cos(1.54307290705  + 50.90702609350 *T)
    sumY1  +=      15.149 * cos(3.50448110876  + 135.57992061650 *T)
    sumY1  +=      15.898 * cos(1.65436415466  + 174.18603700630 *T)
    sumY1  +=      15.347 * cos(5.16270714026  + 16.70614371880 *T)
    sumY1  +=      14.302 * cos(0.85906505598  + 310.41591201661 *T)
    sumY1  +=      17.345 * cos(3.92152229754  + 261.23717611490 *T)
    sumY1  +=      17.208 * cos(2.56277716695  + 183.85033042589 *T)
    sumY1  +=      17.305 * cos(3.31883498051  + 77.50672650040 *T)
    sumY1  +=      14.612 * cos(5.14027366063  + 30.46685461280 *T)
    sumY1  +=      13.603 * cos(0.50135686748  + 286.84003878050 *T)
    sumY1  +=      15.382 * cos(4.90179063300  + 411.37651986550 *T)
    sumY1  +=      13.708 * cos(1.94288900771  + 352.06040979221 *T)
    sumY1  +=      13.683 * cos(2.35141963414  + 82.44777959230 *T)
    sumY1  +=      16.002 * cos(0.31096014945  + 133.13224006171 *T)
    sumY1  +=      13.766 * cos(3.40805334079  + 44.31474395451 *T)
    sumY1  +=      13.284 * cos(1.25113723802  + 7.14491616321 *T)
    sumY1  +=      12.988 * cos(5.46561702418  + 173.92540457540 *T)
    sumY1  +=      17.441 * cos(5.46716688700  + 60.79832204921 *T)
    sumY1  +=      16.261 * cos(0.42832337657  + 64.73047531231 *T)
    sumY1  +=      12.734 * cos(1.40651697882  + 75.54668091261 *T)
    sumY1  +=      12.734 * cos(0.13129280359  + 74.50415118900 *T)
    sumY1  +=      12.700 * cos(0.13947960635  + 57.75994066341 *T)
    sumY1  +=      14.302 * cos(5.43417049382  + 45.82046852221 *T)
    sumY1  +=      12.678 * cos(3.64176420295  + 81.92651473050 *T)
    sumY1  +=      16.549 * cos(4.35598801081  + 760.01171843649 *T)
    sumY1  +=      12.720 * cos(1.24629809128  + 270.16526422410 *T)
    sumY1  +=      13.453 * cos(5.32132524673  + 12.28635548869 *T)
    sumY1  +=      12.225 * cos(5.94217729762  + 284.39235822571 *T)
    sumY1  +=      11.971 * cos(0.20211111004  + 32.48714639791 *T)
    sumY1  +=      12.357 * cos(0.99159645729  + 515.70768857651 *T)
    sumY1  +=      12.069 * cos(4.64103232444  + 633.02755679670 *T)
    sumY1  +=      13.087 * cos(3.37495487812  + 109.70187130500 *T)
    sumY1  +=      11.341 * cos(1.38449312260  + 189.47940471839 *T)
    sumY1  +=      15.361 * cos(2.17588834575  + 200.52510498229 *T)
    sumY1  +=      12.491 * cos(5.89815036388  + 33.92343499640 *T)
    sumY1  +=      14.227 * cos(6.04854994765  + 203.49405039889 *T)
    sumY1  +=      11.094 * cos(4.55762820174  + 267.22994870229 *T)
    sumY1  +=      12.080 * cos(1.08604052506  + 75.67980735660 *T)
    sumY1  +=      12.080 * cos(3.59336191094  + 74.37102474501 *T)
    sumY1  +=      13.196 * cos(3.12727419182  + 143.38148881789 *T)
    sumY1  +=      13.530 * cos(4.25106140608  + 85.32035678930 *T)
    sumY1  +=      10.739 * cos(4.95842698659  + 94.14787116200 *T)
    sumY1  +=      10.888 * cos(6.11572731494  + 2.72512793310 *T)
    sumY1  +=      13.830 * cos(3.79412478066  + 31.89644556690 *T)
    sumY1  +=      11.969 * cos(2.59214165717  + 42.78208713641 *T)
    sumY1  +=      11.212 * cos(1.11251818011  + 61.53203523210 *T)
    sumY1  +=      11.583 * cos(3.50244713214  + 124.53422035260 *T)
    sumY1  +=      12.891 * cos(3.37083041051  + 266.23311096100 *T)
    sumY1  +=      10.738 * cos(1.68908917101  + 35.66854013560 *T)
    sumY1  +=      13.803 * cos(2.66711494730  + 26.26737127440 *T)
    sumY1  +=      10.459 * cos(3.55111688475  + 145.14114817210 *T)
    sumY1  +=      10.474 * cos(0.04265764253  + 692.34366686999 *T)
    sumY1  +=      10.123 * cos(3.61491688022  + 237.11878615761 *T)
    sumY1  +=       9.800 * cos(4.50501895948  + 92.29089143820 *T)
    sumY1  +=      10.121 * cos(5.88328450143  + 156.39929675710 *T)
    sumY1  +=       9.888 * cos(1.02644746029  + 42.88515300039 *T)
    sumY1  +=      10.092 * cos(2.39040220276  + 215.02738562980 *T)
    sumY1  +=      10.098 * cos(5.97397137094  + 533.11617631580 *T)
    sumY1  +=      10.667 * cos(4.94448780106  + 91.80325647120 *T)
    sumY1  +=      10.954 * cos(5.27602541461  + 89.25251005240 *T)
    sumY1  +=       9.630 * cos(2.63099197865  + 55.38169607781 *T)
    sumY1  +=       9.631 * cos(1.67427906238  + 130.68455950690 *T)
    sumY1  +=       9.832 * cos(1.26051341005  + 54.57854692570 *T)
    sumY1  +=       9.739 * cos(4.11020952975  + 76.62176334371 *T)
    sumY1  +=       9.739 * cos(3.71078555984  + 73.42906875790 *T)
    sumY1  +=      10.658 * cos(2.71100873007  + 31.24875190549 *T)
    sumY1  +=       9.264 * cos(4.78218994698  + 27.33115285740 *T)
    sumY1  +=      11.861 * cos(3.13876851681  + 264.26100328570 *T)
    sumY1  +=       9.015 * cos(3.77562206958  + 17.91542978119 *T)
    sumY1  +=       9.170 * cos(0.46232213879  + 203.98168536590 *T)
    sumY1  +=       8.989 * cos(2.84106606677  + 69.12115511239 *T)
    sumY1  +=       9.028 * cos(0.87504387157  + 141.45507312490 *T)
    sumY1  +=       9.664 * cos(0.82099305011  + 41.88831525910 *T)
    sumY1  +=       8.940 * cos(0.30017564872  + 17.28229033479 *T)
    sumY1  +=      10.494 * cos(2.74211545956  + 624.67561530350 *T)
    sumY1  +=      10.035 * cos(1.72606687342  + 448.93340888730 *T)
    sumY1  +=      10.076 * cos(1.82928700699  + 685.23011986920 *T)
    sumY1  +=       8.857 * cos(0.25147206510  + 44.48150029329 *T)
    sumY1  +=       8.598 * cos(4.49527247841  + 199.31581891990 *T)
    sumY1  +=       9.805 * cos(4.58844109700  + 130.16329464510 *T)
    sumY1  +=       8.882 * cos(4.32628974331  + 105.24845318010 *T)
    sumY1  +=       8.882 * cos(1.18469708972  + 105.73608814710 *T)
    sumY1  +=       8.493 * cos(5.75907389910  + 104.23036357940 *T)
    sumY1  +=       8.845 * cos(3.56612831103  + 88.51879686950 *T)
    sumY1  +=       8.246 * cos(4.25831523402  + 269.67762925710 *T)
    sumY1  +=       8.157 * cos(6.12519363628  + 163.82166029861 *T)
    sumY1  +=       8.392 * cos(1.12264471721  + 681.29796660610 *T)
    sumY1  +=       9.056 * cos(4.19874386884  + 617.56206830270 *T)
    sumY1  +=       8.249 * cos(5.01876074157  + 454.66554904380 *T)
    sumY1  +=       8.521 * cos(6.01861922798  + 155.74490545130 *T)
    sumY1  +=      10.211 * cos(3.35774355357  + 145.38722638800 *T)
    sumY1  +=       7.859 * cos(2.48342946211  + 119.75073382760 *T)
    sumY1  +=       8.216 * cos(0.22375706672  + 544.16187657970 *T)
    sumY1  +=       8.454 * cos(1.80026883651  + 393.21727260080 *T)
    sumY1  +=       8.005 * cos(5.18146648973  + 141.73025977080 *T)
    sumY1  +=       8.848 * cos(5.52195322186  + 193.41155798150 *T)
    sumY1  +=       7.843 * cos(5.11391333266  + 46.81730626350 *T)
    sumY1  +=       7.556 * cos(5.15797138498  + 66.67347455760 *T)
    sumY1  +=       7.436 * cos(0.34704884871  + 178.11819026941 *T)
    sumY1  +=       9.047 * cos(1.94321514190  + 3.42521122121 *T)
    sumY1  +=       7.519 * cos(5.21881995453  + 341.01470952831 *T)
    sumY1  +=       7.416 * cos(5.46939147004  + 606.51636803880 *T)
    sumY1  +=       7.453 * cos(4.06769415789  + 175.40987728371 *T)
    sumY1  +=       7.475 * cos(0.09186374252  + 58.07545484850 *T)
    sumY1  +=       8.741 * cos(3.21559628060  + 58.98378094080 *T)
    sumY1  +=       8.749 * cos(4.60735412655  + 179.91817716280 *T)
    sumY1  +=       7.892 * cos(4.97472641388  + 207.63865198310 *T)
    sumY1  +=       8.402 * cos(4.64207669420  + 70.29011363560 *T)
    sumY1  +=       8.402 * cos(0.03732574180  + 79.76071846601 *T)
    sumY1  +=       8.251 * cos(3.53826439231  + 58.56308981550 *T)
    sumY1  +=       8.140 * cos(4.15546301933  + 39.91620758380 *T)
    sumY1  +=       7.045 * cos(4.59142381429  + 949.41942645330 *T)
    sumY1  +=       7.070 * cos(3.43062742384  + 610.44852130189 *T)
    sumY1  +=       7.791 * cos(0.14153367545  + 196.38050339810 *T)
    sumY1  +=       6.907 * cos(2.01398038315  + 107.74182571721 *T)
    sumY1  +=       7.730 * cos(2.22526686348  + 111.18634401329 *T)
    sumY1  +=       7.449 * cos(1.30897364813  + 142.38465107660 *T)
    sumY1  +=       6.733 * cos(1.57499395274  + 49.75262178531 *T)
    sumY1  +=       7.683 * cos(3.44536738109  + 201.01273994931 *T)
    sumY1  +=       7.711 * cos(5.91236505361  + 210.60759739969 *T)
    sumY1  +=       6.780 * cos(5.36729251438  + 78.64875277320 *T)
    sumY1  +=       6.780 * cos(5.59529522880  + 71.40207932840 *T)
    sumY1  +=       6.908 * cos(2.35029974307  + 21.84758304429 *T)
    sumY1  +=       6.741 * cos(1.43650205247  + 363.10611005610 *T)
    sumY1  +=       6.686 * cos(0.93018730801  + 109.06873185860 *T)
    sumY1  +=       6.572 * cos(3.28593985233  + 216.72430665921 *T)
    sumY1  +=       6.803 * cos(3.41888918190  + 95.47228517591 *T)
    sumY1  +=       6.560 * cos(4.25376256253  + 117.56368570370 *T)
    sumY1  +=       7.044 * cos(5.85105840360  + 114.69780615110 *T)
    sumY1  +=       6.442 * cos(1.02882559759  + 49.47743513939 *T)
    sumY1  +=       6.451 * cos(3.83095342617  + 118.62995941420 *T)
    sumY1  +=       6.473 * cos(1.57630900935  + 37.88921815429 *T)
    sumY1  +=       6.192 * cos(5.11515001565  + 380.84715832300 *T)
    sumY1  +=       6.192 * cos(3.97082658887  + 528.97157473160 *T)
    sumY1  +=       6.479 * cos(5.64756980391  + 133.34468838281 *T)
    sumY1  +=       6.479 * cos(2.50597715032  + 132.85705341579 *T)
    sumY1  +=       6.351 * cos(1.72152539515  + 45.96597300160 *T)
    sumY1  +=       6.402 * cos(1.58450052990  + 535.66692273460 *T)
    sumY1  +=       7.995 * cos(4.48478615852  + 35.18090516859 *T)
    sumY1  +=       6.129 * cos(3.11693398156  + 136.31363379940 *T)
    sumY1  +=       6.460 * cos(3.76193692281  + 216.23667169220 *T)
    sumY1  +=       6.081 * cos(3.50561897791  + 530.89799042461 *T)
    sumY1  +=       6.081 * cos(5.58035711965  + 378.92074262999 *T)
    sumY1  +=       6.809 * cos(1.01024597901  + 237.43430034270 *T)
    sumY1  +=       7.057 * cos(1.73750471495  + 526.75338884040 *T)
    sumY1  +=       7.100 * cos(5.98370032528  + 71.56883566720 *T)
    sumY1  +=       6.155 * cos(3.48243401917  + 142.90591593840 *T)
    sumY1  +=       6.607 * cos(3.46558761195  + 192.44835013500 *T)
    sumY1  +=       6.635 * cos(5.56195112276  + 34.28713329129 *T)
    sumY1  +=       6.319 * cos(2.00631988170  + 143.58723949460 *T)
    sumY1  +=       6.200 * cos(4.10240423487  + 455.15318401080 *T)
    sumY1  +=       6.495 * cos(1.71004202786  + 24.62283987170 *T)
    sumY1  +=       7.911 * cos(4.07321915455  + 77.71917482150 *T)
    sumY1  +=       5.852 * cos(6.21910710624  + 405.74744557300 *T)
    sumY1  +=       6.837 * cos(4.49077930583  + 375.63628302830 *T)
    sumY1  +=       5.864 * cos(4.03382406011  + 108.16251684250 *T)
    sumY1  +=       7.485 * cos(2.72090049326  + 48.51422729289 *T)
    sumY1  +=       5.956 * cos(0.03354236238  + 28.98238190449 *T)
    sumY1  +=       6.512 * cos(0.33716951919  + 549.89401673620 *T)
    sumY1  +=       6.744 * cos(5.79310902549  + 27.96429230381 *T)
    sumY1  +=       5.515 * cos(0.51259584253  + 232.66536803271 *T)
    sumY1  +=       5.532 * cos(5.03129304553  + 164.78486814511 *T)
    sumY1  +=       6.345 * cos(1.14676616638  + 91.48774228611 *T)
    sumY1  +=       6.490 * cos(3.33824483658  + 180.03005174739 *T)
    sumY1  +=       5.431 * cos(5.14429257646  + 28.55499313480 *T)
    sumY1  +=       5.425 * cos(2.94479941279  + 73.65281792710 *T)
    sumY1  +=       5.425 * cos(4.87619567680  + 76.39801417450 *T)
    sumY1  +=       5.562 * cos(3.83479812657  + 125.42799222991 *T)
    sumY1  +=       5.415 * cos(1.80533724260  + 15.79781762650 *T)
    sumY1  +=       6.477 * cos(1.46410180306  + 91.06705116081 *T)
    sumY1  +=       5.588 * cos(3.90608403947  + 408.19512612779 *T)
    sumY1  +=       5.776 * cos(2.53576029480  + 218.65072235220 *T)
    sumY1  +=       5.350 * cos(0.82591840288  + 30.98811947460 *T)
    sumY1  +=       5.255 * cos(2.80372355337  + 49.42255338521 *T)
    sumY1  +=       5.759 * cos(3.12808755452  + 35.92917256650 *T)
    sumY1  +=       5.412 * cos(2.06393135001  + 22.39009976550 *T)
    sumY1  +=       6.770 * cos(0.90781613617  + 118.15438653471 *T)
    sumY1  +=       5.380 * cos(4.15958938612  + 234.25290660500 *T)
    sumY1  +=       5.209 * cos(4.69294629475  + 210.36151918381 *T)
    sumY1  +=       5.177 * cos(0.85327766347  + 34.86539098351 *T)
    sumY1  +=       5.160 * cos(2.33442145165  + 312.21589891000 *T)
    sumY1  +=       5.191 * cos(1.32715405123  + 853.44019923550 *T)
    sumY1  +=       5.099 * cos(4.39680257081  + 248.96763557360 *T)
    sumY1  +=       5.267 * cos(2.38447630057  + 167.07249000551 *T)
    sumY1  +=       5.553 * cos(5.79137517506  + 756.07956517339 *T)
    sumY1  +=       5.467 * cos(5.96148887147  + 301.17019864609 *T)
    sumY1  +=       4.979 * cos(1.96711920630  + 82.61453593110 *T)
    sumY1  +=       4.890 * cos(2.71095292959  + 114.38229196601 *T)
    sumY1  +=       4.949 * cos(3.89860321039  + 72.09010052900 *T)
    sumY1  +=       6.602 * cos(5.07904269840  + 123.29582586020 *T)
    sumY1  +=       6.602 * cos(1.93745004481  + 123.78346082721 *T)
    sumY1  +=       5.370 * cos(0.89842846047  + 293.95358578131 *T)
    sumY1  +=       5.201 * cos(5.31738820937  + 67.11541754230 *T)
    sumY1  +=       5.872 * cos(5.68047849308  + 406.34695512460 *T)
    sumY1  +=       5.250 * cos(1.81385138125  + 830.86116374070 *T)
    sumY1  +=       4.889 * cos(2.54047179761  + 319.32944591080 *T)
    sumY1  +=       5.168 * cos(1.40578632698  + 148.11009358871 *T)
    sumY1  +=       6.514 * cos(0.78441227240  + 302.33915716930 *T)
    sumY1  +=       5.080 * cos(0.28960001676  + 163.14033674240 *T)
    sumY1  +=       4.695 * cos(2.65284801875  + 105.13657859549 *T)
    sumY1  +=       5.583 * cos(3.55013544390  + 283.90472325870 *T)
    sumY1  +=       4.645 * cos(3.31039617280  + 87.87110320809 *T)
    sumY1  +=       5.916 * cos(2.46533961374  + 117.66675156770 *T)
    sumY1  +=       4.506 * cos(3.11040332239  + 100.29821031630 *T)
    sumY1  +=       4.545 * cos(0.31134321952  + 52.39149880180 *T)
    sumY1  +=       4.927 * cos(3.41172643540  + 78.46993434690 *T)
    sumY1  +=       5.394 * cos(5.63937544494  + 231.80522605020 *T)
    sumY1  +=       5.223 * cos(3.37807753565  + 376.70255673880 *T)
    sumY1  +=       4.438 * cos(5.11989115592  + 748.96601817260 *T)
    sumY1  +=       5.514 * cos(3.30050922849  + 281.21096448800 *T)
    sumY1  +=       5.215 * cos(2.66968136670  + 336.59492129820 *T)
    sumY1  +=       5.879 * cos(4.76844539244  + 402.77850015639 *T)
    sumY1  +=       5.420 * cos(4.23957972129  + 322.36782729660 *T)
    sumY1  +=       4.308 * cos(2.99235015837  + 122.71967924421 *T)
    sumY1  +=       4.264 * cos(5.47096482557  + 257.14745628489 *T)
    sumY1  +=       4.520 * cos(0.06168859120  + 49.00186225990 *T)
    sumY1  +=       4.775 * cos(4.74715224337  + 57.86300652739 *T)
    sumY1  +=       4.774 * cos(0.72465901741  + 827.67977000299 *T)
    sumY1  +=       4.196 * cos(2.18199332677  + 222.14093263061 *T)
    sumY1  +=       4.391 * cos(2.40221065865  + 103.59722413300 *T)
    sumY1  +=       5.698 * cos(2.18138522351  + 43.84836084689 *T)
    sumY1  +=       4.161 * cos(1.51093067440  + 43.79347909271 *T)
    sumY1  +=       4.435 * cos(3.47733063247  + 102.11275142471 *T)
    sumY1  +=       4.326 * cos(5.43731062929  + 25.35904518210 *T)
    sumY1  +=       4.344 * cos(2.49176998053  + 49.89563413720 *T)
    sumY1  +=       5.029 * cos(5.57234403587  + 189.42452296421 *T)
    sumY1  +=       4.148 * cos(4.15948358685  + 271.37455028650 *T)
    sumY1  +=       4.154 * cos(6.03273715174  + 170.41394243761 *T)
    sumY1  +=       5.131 * cos(1.92264207412  + 107.26874496520 *T)
    sumY1  +=       4.194 * cos(0.39440009742  + 383.06534421420 *T)
    sumY1  +=       4.307 * cos(4.99708612912  + 106.03035047280 *T)
    sumY1  +=       4.419 * cos(3.76956477081  + 422.42222012939 *T)
    sumY1  +=       4.369 * cos(4.33115498014  + 596.11836143630 *T)
    sumY1  +=       4.317 * cos(3.81351212469  + 610.34545543790 *T)
    sumY1  +=       3.889 * cos(1.51405800707  + 501.62360692681 *T)
    sumY1  +=       3.870 * cos(0.14021790007  + 196.01680510321 *T)
    sumY1  +=       4.439 * cos(5.51958079768  + 302.13340649259 *T)
    sumY1  +=       5.072 * cos(0.43837563544  + 841.90686400460 *T)
    sumY1  +=       4.570 * cos(3.91226778053  + 25.30416342790 *T)
    sumY1  +=       3.853 * cos(2.96855650488  + 327.99690158910 *T)
    sumY1  +=       4.468 * cos(4.11940424540  + 20.95381116700 *T)
    sumY1  +=       3.811 * cos(0.01397482107  + 121.59890483079 *T)
    sumY1  +=       4.526 * cos(0.82818905303  + 223.20720634109 *T)
    sumY1  +=       4.482 * cos(1.11896687809  + 30.35498002820 *T)
    sumY1  +=       3.991 * cos(5.99550015236  + 227.03629374021 *T)
    sumY1  +=       3.731 * cos(1.40856591302  + 379.88395047650 *T)
    sumY1  +=       3.956 * cos(2.99745197998  + 995.88984936931 *T)
    sumY1  +=       3.718 * cos(2.08586976042  + 323.74923414091 *T)
    sumY1  +=       4.018 * cos(4.76974981374  + 136.78920667889 *T)
    sumY1  +=       4.001 * cos(1.99020462896  + 1887.54933515920 *T)
    sumY1  +=       3.691 * cos(2.98850914165  + 219.27505307800 *T)
    sumY1  +=       4.238 * cos(3.01325561075  + 487.64259114110 *T)
    sumY1  +=       4.319 * cos(2.53620903359  + 470.44655172289 *T)
    sumY1  +=       3.696 * cos(5.63712983624  + 95.73540973430 *T)
    sumY1  +=       3.585 * cos(1.93930258808  + 647.25465079831 *T)
    sumY1  +=       3.615 * cos(1.77670736543  + 38.85242600079 *T)
    sumY1  +=       4.508 * cos(4.47625982819  + 763.19311217420 *T)
    sumY1  +=       4.774 * cos(4.03912030608  + 205.94173095370 *T)
    sumY1  +=       3.585 * cos(2.91087147320  + 819.81546347680 *T)
    sumY1  +=       3.631 * cos(3.67446828489  + 310.73142620169 *T)
    sumY1  +=       3.713 * cos(6.24192528203  + 906.60597015449 *T)
    sumY1  +=       4.069 * cos(0.88739646436  + 331.92905485220 *T)
    sumY1  +=       3.909 * cos(0.70513913859  + 204.94489321240 *T)
    sumY1  +=       3.826 * cos(1.29816477857  + 253.21530302180 *T)
    sumY1  +=       4.459 * cos(2.50646590612  + 49.26498681829 *T)
    sumY1  +=       3.398 * cos(1.39293258032  + 548.09402984280 *T)
    sumY1  +=       4.440 * cos(3.51258788632  + 54.89406111079 *T)
    sumY1  +=       3.614 * cos(1.87312866978  + 511.28790034640 *T)
    sumY1  +=       3.347 * cos(6.01125856091  + 672.38443271190 *T)
    sumY1  +=       3.809 * cos(2.94099106051  + 328.59641114070 *T)
    sumY1  +=       3.555 * cos(2.04617284577  + 1368.90407032850 *T)
    sumY1  +=       4.016 * cos(5.85194527186  + 442.99551805410 *T)
    sumY1  +=       3.311 * cos(1.96162447032  + 185.23173727020 *T)
    sumY1  +=       3.338 * cos(0.74390697506  + 90.21571789890 *T)
    sumY1  +=       3.346 * cos(3.86205100951  + 10022.08109758290 *T)
    sumY1  +=       3.379 * cos(1.27980657609  + 1051.02772796360 *T)
    sumY1  +=       3.359 * cos(4.15037031744  + 40.82453367610 *T)
    sumY1  +=       3.472 * cos(4.54582298831  + 141.89701610960 *T)
    sumY1  +=       3.470 * cos(2.81336320056  + 1068.22376738181 *T)
    sumY1  +=       3.246 * cos(2.64937498416  + 30.93323772040 *T)
    sumY1  +=       3.583 * cos(3.30161972679  + 297.98880490840 *T)
    sumY1  +=       4.055 * cos(1.93337998586  + 228.83628063359 *T)
    sumY1  +=       3.241 * cos(4.84350666121  + 34.44469985821 *T)
    sumY1  +=       3.616 * cos(5.85754552815  + 736.12033101530 *T)
    sumY1  +=       3.560 * cos(1.84750411756  + 334.89800026879 *T)
    sumY1  +=       4.091 * cos(2.74520731196  + 913.50706883420 *T)
    sumY1  +=       3.191 * cos(2.73872114679  + 374.46732450509 *T)
    sumY1  +=       3.544 * cos(2.06135009476  + 484.67364572450 *T)
    sumY1  +=       3.559 * cos(1.03318246123  + 781.83999440590 *T)
    sumY1  +=       3.813 * cos(1.34656083290  + 342.01154726960 *T)
    sumY1  +=       3.256 * cos(1.24896843058  + 97.17169833279 *T)
    sumY1  +=       3.256 * cos(4.39056108417  + 97.65933329981 *T)
    sumY1  +=       3.853 * cos(5.18264549398  + 122.08653979781 *T)
    sumY1  +=       4.122 * cos(3.91372590203  + 339.04260185300 *T)
    sumY1  +=       3.135 * cos(2.98433382810  + 674.18441960530 *T)
    sumY1  +=       3.359 * cos(1.72611682483  + 369.80145805910 *T)
    sumY1  +=       4.038 * cos(3.91799971610  + 146.35043423450 *T)
    sumY1  +=       4.010 * cos(1.51104428423  + 60.52313540329 *T)
    sumY1  +=       3.941 * cos(6.00494630962  + 729.00678401450 *T)
    sumY1  +=       3.097 * cos(3.84522192892  + 115.18544111810 *T)
    sumY1  +=       3.291 * cos(5.82922546122  + 2424.35384725460 *T)
    sumY1  +=       3.582 * cos(1.33109700739  + 349.12509427039 *T)
    sumY1  +=       3.393 * cos(3.29245537264  + 290.87525790760 *T)
    sumY1  +=       3.127 * cos(3.80374553395  + 582.13734565059 *T)
    sumY1  +=       3.177 * cos(4.90822105320  + 194.53233239490 *T)
    sumY1  +=       3.592 * cos(3.97438571867  + 1512.56300652469 *T)
    sumY1  +=       3.448 * cos(1.31714326925  + 630.57987624190 *T)
    sumY1  +=       3.156 * cos(0.32123034412  + 555.31064270760 *T)
    sumY1  +=       3.587 * cos(5.29397199806  + 333.04982926560 *T)
    sumY1  +=       3.671 * cos(3.02735059577  + 480.52904414029 *T)
    sumY1  +=       3.505 * cos(0.07758783831  + 52.13086637091 *T)
    sumY1  +=       3.028 * cos(1.55633232680  + 260.11640170150 *T)
    sumY1  +=       3.021 * cos(0.85869194720  + 745.03386490949 *T)
    sumY1  +=       3.015 * cos(2.32337006649  + 1053.99667338020 *T)
    sumY1  +=       3.255 * cos(1.55383041618  + 114.12165953510 *T)
    sumY1  +=       3.582 * cos(1.17206776414  + 292.94468595250 *T)
    sumY1  +=       4.061 * cos(5.52420276067  + 758.52724572820 *T)
    sumY1  +=       3.736 * cos(1.93693485461  + 5983.70563823870 *T)
    sumY1  +=       3.303 * cos(5.99394142769  + 532.38246313290 *T)
    sumY1  +=       4.088 * cos(3.98322803937  + 1265.06053658450 *T)
    sumY1  +=       2.960 * cos(0.15626969632  + 15.74293587230 *T)
    sumY1  +=       3.705 * cos(6.19165022961  + 604.71638114540 *T)
    sumY1  +=       3.156 * cos(4.35441637869  + 657.40657918490 *T)
    sumY1  +=       3.133 * cos(0.33631156286  + 346.15614885380 *T)
    sumY1  +=       3.653 * cos(5.34287546352  + 601.74743572880 *T)
    sumY1  +=       3.119 * cos(3.48694146196  + 969.37866061139 *T)
    sumY1  +=       2.876 * cos(3.93018614308  + 61.04440026510 *T)
    sumY1  +=       2.838 * cos(1.17114170634  + 595.15515358980 *T)
    sumY1  +=       3.359 * cos(3.37955835963  + 229.04203131030 *T)
    sumY1  +=       3.882 * cos(6.09416558245  + 275.79433851660 *T)
    sumY1  +=       3.259 * cos(2.58480605176  + 184.48346987229 *T)
    sumY1  +=       3.003 * cos(3.40174825635  + 531.73476947149 *T)
    sumY1  +=       2.904 * cos(2.44331654879  + 90.52453443961 *T)
    sumY1  +=       3.653 * cos(4.53993927617  + 34.92027273770 *T)
    sumY1  +=       3.037 * cos(0.49193141146  + 432.57414851600 *T)
    sumY1  +=       2.813 * cos(3.15838961401  + 106.25735300890 *T)
    sumY1  +=       3.621 * cos(1.65359893537  + 156.92056161890 *T)
    sumY1  +=       3.602 * cos(4.72951169393  + 162.65270177540 *T)
    sumY1  +=       3.499 * cos(4.24251782948  + 296.50433220009 *T)
    sumY1  +=       3.773 * cos(5.65516568689  + 357.79254994869 *T)
    sumY1  +=       3.503 * cos(3.36034498099  + 9913.91533445829 *T)
    sumY1  +=       3.442 * cos(5.17751962044  + 689.16227313230 *T)
    sumY1  +=       2.741 * cos(2.79642191830  + 279.97256999560 *T)
    sumY1  +=       2.811 * cos(0.21105822548  + 526.00262931501 *T)
    sumY1  +=       2.811 * cos(0.21105822548  + 525.51499434800 *T)
    sumY1  +=       2.855 * cos(4.14691508428  + 49.16192095430 *T)
    sumY1  +=       3.389 * cos(3.10180076770  + 1471.24608502170 *T)
    sumY1  +=       2.713 * cos(0.85189767911  + 395.66495315560 *T)
    sumY1  +=       2.786 * cos(6.05589669892  + 121.49583896680 *T)
    sumY1  +=       2.898 * cos(4.64602066056  + 101.04896984171 *T)
    sumY1  +=       2.865 * cos(4.44084667312  + 525.48136445320 *T)
    sumY1  +=       3.424 * cos(1.02662988746  + 1065.35788782919 *T)
    sumY1  +=       2.851 * cos(5.26092114416  + 344.94686279140 *T)
    sumY1  +=       3.251 * cos(3.02407658514  + 4.20960064140 *T)
    sumY1  +=       2.737 * cos(1.85379067922  + 1065.25482196520 *T)
    sumY1  +=       2.950 * cos(3.40947832795  + 256.08118257441 *T)
    sumY1  +=       2.631 * cos(2.35346659913  + 833.30884429549 *T)
    sumY1  +=       3.298 * cos(0.35354217872  + 469.48334387639 *T)
    sumY1  +=       3.109 * cos(1.56878798104  + 626.16008801179 *T)
    sumY1  +=       3.056 * cos(5.25542729410  + 222.61650551010 *T)
    sumY1  +=       2.887 * cos(4.00574991438  + 241.36645360580 *T)
    sumY1  +=       2.880 * cos(2.11510054841  + 370.21965705690 *T)
    sumY1  +=       2.756 * cos(5.49510286400  + 103.76398047180 *T)
    sumY1  +=       2.886 * cos(1.01588865321  + 668.45227944880 *T)
    sumY1  +=       3.592 * cos(5.19041993480  + 1062.59469308931 *T)
    sumY1  +=       3.123 * cos(1.19879958924  + 384.02855206069 *T)
    sumY1  +=       2.750 * cos(3.16455625165  + 628.60776856660 *T)
    sumY1  +=       2.877 * cos(4.46420389167  + 6058.48723680599 *T)
    sumY1  +=       2.805 * cos(0.44973972850  + 2111.89413086110 *T)
    sumY1  +=       2.563 * cos(5.13388413274  + 440.43845504219 *T)
    sumY1  +=       2.563 * cos(5.13388413274  + 440.92609000920 *T)
    sumY1  +=       2.696 * cos(4.53768728515  + 187.78248368899 *T)
    sumY1  +=       2.619 * cos(5.83386297293  + 218.16308738520 *T)
    sumY1  +=       3.250 * cos(4.11861466788  + 608.96404859360 *T)
    sumY1  +=       2.535 * cos(5.86019269756  + 546.71262299849 *T)
    sumY1  +=       2.863 * cos(1.36645816570  + 637.69342324271 *T)
    sumY1  +=       2.510 * cos(1.15126001886  + 1258.69774910910 *T)
    sumY1  +=       2.502 * cos(4.83412535374  + 766.12842769600 *T)
    sumY1  +=       2.501 * cos(5.52832480667  + 108.21739859670 *T)
    sumY1  +=       2.522 * cos(2.97707565426  + 115.60613224340 *T)
    sumY1  +=       2.695 * cos(6.17673554826  + 592.18620817320 *T)
    sumY1  +=       2.760 * cos(2.62533693066  + 631.78916230430 *T)
    sumY1  +=       2.903 * cos(0.96759074876  + 620.49738382450 *T)
    sumY1  +=       2.884 * cos(0.01635711768  + 5.17280848790 *T)
    sumY1  +=       2.890 * cos(1.74439600900  + 361.62163734781 *T)
    sumY1  +=       2.464 * cos(4.66328256455  + 416.79314583689 *T)
    sumY1  +=       3.268 * cos(4.10372444017  + 39.37369086260 *T)
    sumY1  +=       2.473 * cos(1.65858882314  + 368.83825021259 *T)
    sumY1  +=       3.222 * cos(5.36283883350  + 421.21293406700 *T)
    sumY1  +=       2.626 * cos(2.67212495415  + 135.82599883240 *T)
    sumY1  +=       2.497 * cos(0.51632716818  + 109.22629842550 *T)
    sumY1  +=       2.634 * cos(1.68453170822  + 159.47130803769 *T)
    sumY1  +=       3.221 * cos(2.46115208149  + 124.25903370670 *T)
    sumY1  +=       2.517 * cos(5.73217436168  + 389.96644289390 *T)
    sumY1  +=       2.921 * cos(4.97586465251  + 109.96250373590 *T)
    sumY1  +=       3.208 * cos(6.15657013497  + 436.50630177909 *T)
    sumY1  +=       2.416 * cos(1.90511651044  + 531.41925528640 *T)
    sumY1  +=       3.214 * cos(4.11241434108  + 256.18424843839 *T)
    sumY1  +=       2.505 * cos(5.56614254079  + 460.88532416730 *T)
    sumY1  +=       2.681 * cos(5.88558098740  + 9988.69693302559 *T)
    sumY1  +=       2.782 * cos(2.35083058245  + 947.28726008359 *T)
    sumY1  +=       2.409 * cos(0.47303616104  + 1228.25440708809 *T)
    sumY1  +=       2.353 * cos(3.93573483899  + 75.23786437190 *T)
    sumY1  +=       2.353 * cos(3.88526025060  + 74.81296772971 *T)
    sumY1  +=       2.440 * cos(1.03735538956  + 894.59706204410 *T)
    sumY1  +=       2.656 * cos(1.88372367878  + 623.46632924110 *T)
    sumY1  +=       2.337 * cos(1.20346306186  + 533.86693584120 *T)
    sumY1  +=       2.402 * cos(0.39312271717  + 843.87897167990 *T)
    sumY1  +=       3.239 * cos(3.93030878487  + 956.53297345411 *T)
    sumY1  +=       3.017 * cos(3.28682609944  + 2043.73844432759 *T)
    sumY1  +=       3.017 * cos(3.28682609944  + 2044.22607929460 *T)
    sumY1  +=       2.405 * cos(2.26306379161  + 59.03866269500 *T)
    sumY1  +=       2.405 * cos(5.40465644520  + 59.52629766200 *T)
    sumY1  +=       2.976 * cos(3.41026880242  + 80.65449034331 *T)
    sumY1  +=       2.378 * cos(1.24449551265  + 223.62540533890 *T)
    sumY1  +=       2.317 * cos(5.89016852592  + 247.22253043440 *T)
    sumY1  +=       2.571 * cos(3.46653079079  + 595.47066777489 *T)
    sumY1  +=       2.545 * cos(1.99218841509  + 450.21213091890 *T)
    sumY1  +=       2.346 * cos(3.33503291769  + 60.31068708220 *T)
    sumY1  +=       2.905 * cos(0.84313259036  + 344.45922782439 *T)
    sumY1  +=       2.979 * cos(2.84128335260  + 370.32272292090 *T)
    sumY1  +=       2.400 * cos(3.25893447553  + 14.73403604350 *T)
    sumY1  +=       2.717 * cos(6.02377356419  + 479.64408098360 *T)
    sumY1  +=       2.611 * cos(2.56467646999  + 3116.02381351440 *T)
    sumY1  +=       2.621 * cos(1.82920672957  + 78.99119920869 *T)
    sumY1  +=       2.223 * cos(0.33085814506  + 53.82778740031 *T)
    sumY1  +=       2.402 * cos(0.32300012534  + 298.88257678570 *T)
    sumY1  +=       2.990 * cos(5.53041231058  + 1812.28010162490 *T)
    sumY1  +=       2.451 * cos(0.34339233154  + 230.80838830890 *T)
    sumY1  +=       2.234 * cos(3.79720855786  + 6144.31453563720 *T)
    sumY1  +=       2.327 * cos(0.27187374064  + 756.56720014041 *T)
    sumY1  +=       2.505 * cos(0.40988568764  + 98.65617104110 *T)
    sumY1  +=       2.409 * cos(3.41904862150  + 248.48000060659 *T)
    sumY1  +=       2.564 * cos(5.30963250248  + 554.31380496631 *T)
    sumY1  +=       2.094 * cos(3.86522425502  + 451.22103074770 *T)
    sumY1  +=       2.094 * cos(0.72363160144  + 450.73339578069 *T)
    sumY1  +=       2.295 * cos(4.07822849681  + 767.12526543730 *T)
    sumY1  +=       2.295 * cos(0.93663584322  + 767.61290040431 *T)
    sumY1  +=       2.066 * cos(3.03632358122  + 518.40144734719 *T)
    sumY1  +=       2.066 * cos(6.17791623481  + 518.88908231420 *T)
    sumY1  +=       2.137 * cos(5.77720147587  + 192.93598510200 *T)
    sumY1  +=       2.874 * cos(2.99281146129  + 1336.17310644710 *T)
    sumY1  +=       2.325 * cos(3.58736486540  + 217.16624964391 *T)
    sumY1  +=       2.325 * cos(3.58736486540  + 216.67861467689 *T)
    sumY1  +=       2.567 * cos(3.86077733216  + 980.42436087530 *T)
    sumY1  +=       2.617 * cos(1.18764734158  + 130.19692453990 *T)
    sumY1  +=       2.128 * cos(3.54731661850  + 180.51768671441 *T)
    sumY1  +=       2.474 * cos(0.80985182636  + 124.74666867370 *T)
    sumY1  +=       2.195 * cos(3.69540445414  + 449.52410971830 *T)
    sumY1  +=       2.195 * cos(0.55381180055  + 449.03647475130 *T)
    sumY1  +=       2.135 * cos(1.00327310585  + 112.67081672160 *T)
    sumY1  +=       1.942 * cos(3.76423202456  + 196.86813836510 *T)
    sumY1  +=       1.912 * cos(2.05430704660  + 119.11759438120 *T)
    sumY1  +=       2.139 * cos(4.56457588625  + 179.03321400610 *T)
    sumY1  +=       2.139 * cos(4.56457588625  + 178.54557903910 *T)
    sumY1  +=       1.950 * cos(0.56420918284  + 39.11305843170 *T)
    sumY1  +=       1.854 * cos(5.12023202898  + 355.24180352990 *T)
    sumY1  +=       1.854 * cos(5.12023202898  + 354.75416856289 *T)
    sumY1  +=       1.854 * cos(4.72493321950  + 198.82818395289 *T)
    sumY1  +=       1.965 * cos(4.47911184507  + 84.09900863940 *T)
    sumY1  +=       2.359 * cos(4.45376932769  + 31.73638687251 *T)
    sumY1  +=       1.903 * cos(5.59243523932  + 153.25153291419 *T)
    sumY1  +=       2.125 * cos(2.89742576427  + 490.57790666290 *T)
    sumY1  +=       2.125 * cos(2.89742576428  + 490.09027169589 *T)
    sumY1  +=       1.914 * cos(1.23863728949  + 331.45348197271 *T)
    sumY1  +=       1.914 * cos(4.38022994308  + 330.96584700570 *T)
    sumY1  +=       2.417 * cos(5.19667401678  + 139.75815209550 *T)
    sumY1  +=       2.146 * cos(0.08821351821  + 463.26356875290 *T)
    sumY1  +=       2.146 * cos(3.22980617180  + 463.75120371991 *T)
    sumY1  +=       1.907 * cos(5.00652893184  + 458.12212942741 *T)
    sumY1  +=       1.907 * cos(1.86493627825  + 457.63449446040 *T)
    sumY1  +=       1.967 * cos(1.07124636244  + 180.81194904009 *T)
    sumY1  +=       1.967 * cos(1.07124636244  + 181.29958400710 *T)
    sumY1  +=       2.145 * cos(4.85375320990  + 66.42739634171 *T)
    sumY1  +=       2.145 * cos(4.85375320990  + 65.93976137470 *T)
    sumY1  +=       1.797 * cos(3.97024155339  + 757.56403788170 *T)
    sumY1  +=       1.797 * cos(0.82864889980  + 758.05167284870 *T)
    sumY1  +=       1.750 * cos(5.85679825237  + 144.86596152620 *T)
    sumY1  +=       2.036 * cos(0.77312847257  + 304.10551416790 *T)
    sumY1  +=       2.036 * cos(3.91472112616  + 303.61787920090 *T)
    sumY1  +=       1.797 * cos(1.89213659166  + 704.87383984220 *T)
    sumY1  +=       1.797 * cos(5.03372924525  + 705.36147480921 *T)
    sumY1  +=       1.894 * cos(5.59204990603  + 910.06255053810 *T)
    sumY1  +=       1.894 * cos(2.45045725245  + 909.57491557110 *T)
    sumY1  +=       1.722 * cos(0.77969749873  + 67.63668240410 *T)
    sumY1  +=       1.837 * cos(1.37910752336  + 544.26494244370 *T)
    sumY1  +=       1.837 * cos(1.37910752336  + 544.75257741070 *T)
    sumY1  +=       1.726 * cos(4.71165482381  + 1507.42156719921 *T)
    sumY1  +=       1.726 * cos(4.71165482381  + 1506.93393223219 *T)
    sumY1  +=       1.730 * cos(4.46134223293  + 565.35950523021 *T)
    sumY1  +=       1.730 * cos(4.46134223293  + 564.87187026320 *T)
    sumY1  +=       1.877 * cos(5.10265146514  + 582.62498061761 *T)
    sumY1  +=       1.795 * cos(4.96114630442  + 681.78560157310 *T)
    sumY1  +=       1.921 * cos(0.82498060567  + 699.45721387080 *T)
    sumY1  +=       1.983 * cos(0.97771499558  + 256.78375799000 *T)
    sumY1  +=       1.983 * cos(0.97771499559  + 256.29612302300 *T)
    sumY1  +=       1.763 * cos(5.33310923745  + 1521.64866120080 *T)
    sumY1  +=       1.763 * cos(5.33310923745  + 1521.16102623379 *T)
    sumY1  +=       2.238 * cos(2.09108890473  + 699.94484883781 *T)
    sumY1  +=       1.965 * cos(2.07895283772  + 1317.26309965699 *T)
    sumY1  +=       1.965 * cos(5.22054549131  + 1317.75073462401 *T)
    sumY1  +=       1.641 * cos(4.21606567095  + 76.98546163860 *T)
    sumY1  +=       1.645 * cos(1.65058956203  + 237.92193530970 *T)
    sumY1  +=       1.880 * cos(3.46104485730  + 71.35638734610 *T)
    sumY1  +=       2.123 * cos(1.67868119422  + 528.45030986980 *T)
    sumY1  +=       2.123 * cos(1.67868119422  + 527.96267490280 *T)
    sumY1  +=       2.071 * cos(2.17956762731  + 688.41151360689 *T)
    sumY1  +=       2.071 * cos(2.17956762731  + 688.89914857391 *T)
    sumY1  +=       2.066 * cos(0.76498091962  + 182.99899716400 *T)
    sumY1  +=       2.080 * cos(3.39549064677  + 415.79630809560 *T)
    sumY1  +=       1.610 * cos(0.29709580866  + 173.69840203930 *T)
    sumY1  +=       1.624 * cos(3.67562727861  + 683.74564716089 *T)
    sumY1  +=       1.624 * cos(0.53403462502  + 684.23328212791 *T)
    sumY1  +=       1.582 * cos(5.35394360679  + 349.09146437560 *T)
    sumY1  +=       1.582 * cos(2.21235095320  + 348.60382940860 *T)
    sumY1  +=       1.525 * cos(1.67128272620  + 555.79827767460 *T)
    sumY1  +=       1.823 * cos(3.83750566677  + 290.46337550291 *T)
    sumY1  +=       1.823 * cos(0.69591301318  + 289.97574053589 *T)
    sumY1  +=       1.524 * cos(2.71848730694  + 1303.03600565540 *T)
    sumY1  +=       1.524 * cos(5.86007996053  + 1303.52364062241 *T)
    sumY1  +=       1.902 * cos(1.08556559808  + 351.57277482520 *T)
    sumY1  +=       1.683 * cos(5.22184240312  + 477.56009872370 *T)
    sumY1  +=       1.683 * cos(5.22184240312  + 478.04773369070 *T)
    sumY1  +=       1.652 * cos(3.50958885438  + 473.90313210650 *T)
    sumY1  +=       1.652 * cos(3.50958885438  + 473.41549713950 *T)
    sumY1  +=       1.475 * cos(3.72868396640  + 406.71065341949 *T)
    sumY1  +=       1.475 * cos(3.72868396640  + 407.19828838650 *T)
    sumY1  +=       1.473 * cos(3.18933946169  + 376.43943218040 *T)
    sumY1  +=       1.473 * cos(3.18933946169  + 375.95179721340 *T)
    sumY1  +=       1.646 * cos(1.83930538139  + 330.08088384900 *T)
    sumY1  +=       1.646 * cos(1.83930538139  + 329.59324888200 *T)
    sumY1  +=       2.036 * cos(1.94312054350  + 402.46298597130 *T)
    sumY1  +=       2.036 * cos(5.08471319709  + 401.97535100430 *T)
    sumY1  +=       1.806 * cos(0.37937472852  + 105.62421356250 *T)
    sumY1  +=       1.436 * cos(2.84665188190  + 384.30373870660 *T)
    sumY1  +=       1.436 * cos(5.98824453549  + 383.81610373959 *T)
    sumY1  +=       1.436 * cos(5.73162200977  + 192.56022471960 *T)
    sumY1  +=       1.436 * cos(2.59002935618  + 193.04785968660 *T)
    sumY1  +=       1.579 * cos(5.41871812435  + 74.64965566840 *T)
    sumY1  +=       1.983 * cos(1.16308584406  + 28.32799059869 *T)
    sumY1  +=       1.517 * cos(2.32668455089  + 696.76345510010 *T)
    sumY1  +=       1.517 * cos(5.46827720448  + 696.27582013310 *T)
    sumY1  +=       1.512 * cos(0.66197534140  + 482.97672469510 *T)
    sumY1  +=       1.512 * cos(0.66197534140  + 483.46435966210 *T)
    sumY1  +=       1.507 * cos(2.65716293440  + 420.93774742110 *T)
    sumY1  +=       1.507 * cos(2.65716293440  + 421.42538238810 *T)
    sumY1  +=       1.461 * cos(1.45515033376  + 6208.53806890760 *T)
    sumY1  +=       1.461 * cos(1.45515033376  + 6208.05043394060 *T)
    sumY1  +=       1.593 * cos(2.52079664808  + 1441.23798834100 *T)
    sumY1  +=       1.593 * cos(5.66238930167  + 1440.75035337400 *T)
    sumY1  +=       1.473 * cos(0.74953156826  + 818.33099076849 *T)
    sumY1  +=       1.473 * cos(3.89112422185  + 818.81862573550 *T)
    sumY1  +=       1.672 * cos(0.52981567811  + 1598.71788162851 *T)
    sumY1  +=       1.672 * cos(3.67140833170  + 1598.23024666149 *T)
    sumY1  +=       1.531 * cos(3.55867760683  + 1737.98613802460 *T)
    sumY1  +=       1.531 * cos(3.55867760683  + 1737.49850305759 *T)
    sumY1  +=       1.912 * cos(4.40604797730  + 486.64575339981 *T)

    # URANUS Y Coord
    sumY2  = 0
    sumY2  +=  1016618.950 * cos(5.48295390720  + 75.02541605080 *T)
    sumY2  +=   38040.244 * cos(5.98396870387  + 0.24381748350 *T)
    sumY2  +=   34677.856 * cos(1.59683674929  + 149.80701461810 *T)
    sumY2  +=   12364.786 * cos(4.28633873136  + 76.50988875911 *T)
    sumY2  +=    7910.801 * cos(2.23474062458  + 11.28951774740 *T)
    sumY2  +=    7671.667 * cos(0.96951294592  + 63.97971578690 *T)
    sumY2  +=    5406.463 * cos(3.19611828309  + 71.09326278771 *T)
    sumY2  +=    7007.650 * cos(3.73693507089  + 86.07111631471 *T)
    sumY2  +=    4975.036 * cos(1.46390866707  + 78.95756931390 *T)
    sumY2  +=    3201.017 * cos(1.29888367700  + 74.53778108379 *T)
    sumY2  +=    3151.344 * cos(2.97264499518  + 73.54094334250 *T)
    sumY2  +=    2824.019 * cos(2.46329588481  + 138.76131435421 *T)
    sumY2  +=    3389.631 * cos(5.14247489420  + 1.72829019180 *T)
    sumY2  +=    2518.149 * cos(1.78699516010  + 71.84402231310 *T)
    sumY2  +=    2418.298 * cos(2.88506969771  + 78.20680978850 *T)
    sumY2  +=    2201.666 * cos(5.57455075609  + 4.17597074660 *T)
    sumY2  +=    1870.514 * cos(4.21015183892  + 224.58861318540 *T)
    sumY2  +=    1040.809 * cos(5.43378129938  + 145.87486135500 *T)
    sumY2  +=     825.679 * cos(3.00325076020  + 148.32254190981 *T)
    sumY2  +=     856.424 * cos(0.62220902296  + 151.29148732640 *T)
    sumY2  +=     661.761 * cos(5.72562810777  + 18.40306474821 *T)
    sumY2  +=     687.998 * cos(3.36331816740  + 77.99436146740 *T)
    sumY2  +=     504.683 * cos(3.72460726525  + 22.33521801130 *T)
    sumY2  +=     477.418 * cos(1.60651996166  + 72.57773549600 *T)
    sumY2  +=     476.558 * cos(3.42662179631  + 2.93757625420 *T)
    sumY2  +=     417.277 * cos(6.26887017967  + 160.85271488200 *T)
    sumY2  +=     385.934 * cos(3.14575356964  + 77.47309660560 *T)
    sumY2  +=     325.076 * cos(4.72083769491  + 65.46418849521 *T)
    sumY2  +=     284.902 * cos(5.24555103176  + 131.64776735340 *T)
    sumY2  +=     257.868 * cos(3.40246798687  + 56.86616878610 *T)
    sumY2  +=     334.260 * cos(3.94997003912  + 3.21276290011 *T)
    sumY2  +=     227.240 * cos(1.00970775338  + 127.71561409031 *T)
    sumY2  +=     308.306 * cos(0.17254957815  + 84.58664360640 *T)
    sumY2  +=     220.448 * cos(1.89180052783  + 220.65645992230 *T)
    sumY2  +=     207.595 * cos(1.11792349211  + 12.77399045571 *T)
    sumY2  +=     206.410 * cos(1.32580238054  + 93.18466331550 *T)
    sumY2  +=     207.522 * cos(0.26193997823  + 213.54291292150 *T)
    sumY2  +=     209.061 * cos(3.99916829829  + 153.73916788121 *T)
    sumY2  +=     215.886 * cos(6.16271220897  + 72.05647063421 *T)
    sumY2  +=     212.084 * cos(2.67452073343  + 87.55558902300 *T)
    sumY2  +=     208.906 * cos(1.05579446681  + 52.44638055600 *T)
    sumY2  +=     151.875 * cos(6.08269315176  + 206.42936592071 *T)
    sumY2  +=     148.298 * cos(0.02551800142  + 9.80504503910 *T)
    sumY2  +=     192.262 * cos(0.59276641736  + 529.93478257810 *T)
    sumY2  +=     119.051 * cos(0.55858786575  + 299.37021175271 *T)
    sumY2  +=     132.408 * cos(5.56072688448  + 52.93401552300 *T)
    sumY2  +=     117.583 * cos(2.55509093793  + 522.82123557730 *T)
    sumY2  +=     114.530 * cos(4.91719768081  + 537.04832957890 *T)
    sumY2  +=     105.238 * cos(2.52245002296  + 127.22797912329 *T)
    sumY2  +=      98.089 * cos(0.44292716212  + 80.44204202221 *T)
    sumY2  +=     114.250 * cos(6.16582036197  + 7.35736448430 *T)
    sumY2  +=      97.881 * cos(5.59286140261  + 97.11681657861 *T)
    sumY2  +=     106.194 * cos(4.32064505626  + 146.62562088040 *T)
    sumY2  +=      89.216 * cos(1.17705594524  + 67.91186905000 *T)
    sumY2  +=     116.340 * cos(5.31120252949  + 38.37685312130 *T)
    sumY2  +=      77.574 * cos(3.41292158693  + 82.13896305161 *T)
    sumY2  +=      84.483 * cos(5.24247398645  + 223.10414047710 *T)
    sumY2  +=      98.707 * cos(2.03152578880  + 62.49524307860 *T)
    sumY2  +=      68.606 * cos(0.89160531388  + 426.84200835950 *T)
    sumY2  +=      66.709 * cos(3.48909893013  + 2.69149803831 *T)
    sumY2  +=      65.809 * cos(1.97527994783  + 15.22167101050 *T)
    sumY2  +=      67.096 * cos(3.25627562801  + 226.07308589371 *T)
    sumY2  +=      87.378 * cos(4.13754658007  + 39.86132582961 *T)
    sumY2  +=      62.935 * cos(3.90702021396  + 149.31937965110 *T)
    sumY2  +=      61.638 * cos(0.99025506838  + 152.98840835580 *T)
    sumY2  +=      65.914 * cos(2.38989307822  + 79.47883417571 *T)
    sumY2  +=      59.769 * cos(3.96253258380  + 10.80188278040 *T)
    sumY2  +=      59.691 * cos(4.46676970396  + 56.37853381910 *T)
    sumY2  +=      56.598 * cos(5.97114569701  + 152.77596003471 *T)
    sumY2  +=      52.271 * cos(0.30381307123  + 70.60562782070 *T)
    sumY2  +=      67.381 * cos(2.51665952752  + 111.67397898031 *T)
    sumY2  +=      48.944 * cos(3.55522236110  + 63.49208081989 *T)
    sumY2  +=      48.935 * cos(6.11180947236  + 194.89603068979 *T)
    sumY2  +=      58.953 * cos(4.45076342362  + 34.92027273770 *T)
    sumY2  +=      42.178 * cos(4.17195774633  + 146.83806920150 *T)
    sumY2  +=      41.131 * cos(2.21857435803  + 209.12312469139 *T)
    sumY2  +=      48.341 * cos(4.60129323937  + 142.69346761730 *T)
    sumY2  +=      39.938 * cos(1.64490876502  + 29.44876501211 *T)
    sumY2  +=      35.504 * cos(5.92626067208  + 73.05330837550 *T)
    sumY2  +=      35.998 * cos(4.22663106168  + 69.60879007940 *T)
    sumY2  +=      35.709 * cos(3.11698493448  + 45.33283355520 *T)
    sumY2  +=      34.365 * cos(1.76540961765  + 70.57199792590 *T)
    sumY2  +=      40.938 * cos(3.04417985538  + 4.69723560840 *T)
    sumY2  +=      33.246 * cos(3.44251759587  + 1059.62574767270 *T)
    sumY2  +=      34.076 * cos(3.89497875321  + 137.27684164590 *T)
    sumY2  +=      34.342 * cos(4.51524705118  + 5.66044345490 *T)
    sumY2  +=      34.750 * cos(4.57070962856  + 120.11443212250 *T)
    sumY2  +=      29.609 * cos(2.74253950101  + 134.82916109110 *T)
    sumY2  +=      29.214 * cos(5.90248764200  + 134.34152612410 *T)
    sumY2  +=      27.728 * cos(5.30138582803  + 202.49721265760 *T)
    sumY2  +=      26.015 * cos(6.15906865719  + 20.85074530300 *T)
    sumY2  +=      25.423 * cos(2.61021717564  + 23.81969071961 *T)
    sumY2  +=      25.360 * cos(0.40111355089  + 60.04756252381 *T)
    sumY2  +=      27.639 * cos(5.35140184211  + 152.25469517290 *T)
    sumY2  +=      24.633 * cos(3.55951058232  + 59.55992755679 *T)
    sumY2  +=      26.033 * cos(2.49977137612  + 235.63431344930 *T)
    sumY2  +=      22.264 * cos(0.49547833115  + 76.02225379210 *T)
    sumY2  +=      20.549 * cos(0.87420485234  + 138.27367938720 *T)
    sumY2  +=      22.051 * cos(5.50813364070  + 33.38091827520 *T)
    sumY2  +=      19.406 * cos(1.90896040115  + 159.36824217371 *T)
    sumY2  +=      20.190 * cos(6.02481724288  + 70.35954960481 *T)
    sumY2  +=      22.751 * cos(3.81525392379  + 110.18950627200 *T)
    sumY2  +=      20.572 * cos(3.04462264934  + 120.60206708950 *T)
    sumY2  +=      19.010 * cos(5.84440967288  + 14.47091148511 *T)
    sumY2  +=      18.238 * cos(2.33410359218  + 3.68833577960 *T)
    sumY2  +=      18.044 * cos(4.22469337938  + 265.74547599400 *T)
    sumY2  +=      18.132 * cos(4.25620953660  + 90.00326957780 *T)
    sumY2  +=      18.837 * cos(1.79513158278  + 79.69128249680 *T)
    sumY2  +=      17.966 * cos(0.73075101045  + 73.32849502140 *T)
    sumY2  +=      17.619 * cos(0.80599394510  + 76.72233708021 *T)
    sumY2  +=      15.431 * cos(1.05579242119  + 1.24065522479 *T)
    sumY2  +=      17.688 * cos(5.64947491656  + 41.40068029210 *T)
    sumY2  +=      15.101 * cos(3.31428797427  + 16.91859203990 *T)
    sumY2  +=      15.248 * cos(4.50282713009  + 3.42521122121 *T)
    sumY2  +=      14.771 * cos(1.52291283523  + 333.41352756050 *T)
    sumY2  +=      14.325 * cos(4.50651618798  + 19.88753745650 *T)
    sumY2  +=      13.703 * cos(1.03711862016  + 116.18227885939 *T)
    sumY2  +=      15.398 * cos(5.26404040593  + 162.33718759030 *T)
    sumY2  +=      13.837 * cos(1.24389395766  + 433.95555536030 *T)
    sumY2  +=      13.525 * cos(0.24969551504  + 144.39038864671 *T)
    sumY2  +=      15.125 * cos(3.90702018283  + 147.35933406330 *T)
    sumY2  +=      15.394 * cos(3.18550823490  + 41.34579853790 *T)
    sumY2  +=      13.487 * cos(3.90445031030  + 347.64062156210 *T)
    sumY2  +=      17.344 * cos(2.71114683863  + 190.96387742670 *T)
    sumY2  +=      12.741 * cos(5.20555748274  + 419.72846135871 *T)
    sumY2  +=      15.299 * cos(6.02030544502  + 184.97110483931 *T)
    sumY2  +=      14.796 * cos(1.41052446050  + 113.15845168861 *T)
    sumY2  +=      11.042 * cos(5.60683409366  + 25.51661174900 *T)
    sumY2  +=      10.803 * cos(4.53729829502  + 140.24578706250 *T)
    sumY2  +=      10.475 * cos(3.78319139464  + 6.86972951729 *T)
    sumY2  +=      11.185 * cos(5.46980447272  + 116.66991382640 *T)
    sumY2  +=      11.147 * cos(3.41565238019  + 54.41848823131 *T)
    sumY2  +=      13.346 * cos(5.65327230364  + 33.43580002939 *T)
    sumY2  +=       9.592 * cos(5.93991057680  + 212.05844021321 *T)
    sumY2  +=      10.605 * cos(5.53671235012  + 74.91354146621 *T)
    sumY2  +=      10.593 * cos(2.28899861007  + 75.13729063540 *T)
    sumY2  +=      11.833 * cos(1.35549917067  + 89.04006173130 *T)
    sumY2  +=       9.674 * cos(4.86901914741  + 129.20008679860 *T)
    sumY2  +=       9.261 * cos(3.82300871178  + 167.96626188280 *T)
    sumY2  +=       9.431 * cos(0.05576165559  + 640.14110379750 *T)
    sumY2  +=      10.948 * cos(1.13466058909  + 202.00957769060 *T)
    sumY2  +=      10.470 * cos(4.78808252913  + 71.56883566720 *T)
    sumY2  +=       8.486 * cos(1.13985316911  + 277.27881122490 *T)
    sumY2  +=       9.140 * cos(0.24006195765  + 228.52076644850 *T)
    sumY2  +=       8.941 * cos(6.22720311602  + 258.63192899320 *T)
    sumY2  +=       8.140 * cos(1.36707236754  + 60.79832204921 *T)
    sumY2  +=       7.518 * cos(2.15319287040  + 126.23114138200 *T)
    sumY2  +=       7.532 * cos(3.27091826281  + 89.25251005240 *T)
    sumY2  +=       7.736 * cos(1.43507903752  + 272.85902299479 *T)
    sumY2  +=       8.693 * cos(5.98733781106  + 221.61966776881 *T)
    sumY2  +=       6.861 * cos(1.71468974496  + 195.38366565681 *T)
    sumY2  +=       8.704 * cos(2.06590461383  + 50.96190784770 *T)
    sumY2  +=       7.676 * cos(1.61804051361  + 58.35064149440 *T)
    sumY2  +=       7.545 * cos(4.29974248997  + 53.93085326429 *T)
    sumY2  +=       7.394 * cos(1.31233396844  + 95.63234387030 *T)
    sumY2  +=       6.681 * cos(3.85199580169  + 66.94866120350 *T)
    sumY2  +=       6.264 * cos(0.42668844756  + 39.60069339870 *T)
    sumY2  +=       7.026 * cos(0.33343091499  + 51.44954281471 *T)
    sumY2  +=       6.150 * cos(3.23806032802  + 374.15181032000 *T)
    sumY2  +=       6.356 * cos(2.95032284330  + 155.22364058950 *T)
    sumY2  +=       5.982 * cos(3.58571049504  + 125.74350641500 *T)
    sumY2  +=       6.068 * cos(5.12117005220  + 135.57992061650 *T)
    sumY2  +=       6.766 * cos(6.24976890449  + 151.50393564750 *T)
    sumY2  +=       7.931 * cos(1.61760247052  + 288.32451148881 *T)
    sumY2  +=       6.447 * cos(1.38968353769  + 297.88573904440 *T)
    sumY2  +=       6.043 * cos(5.48371639310  + 1066.73929467350 *T)
    sumY2  +=       5.769 * cos(1.79748950801  + 213.05527795450 *T)
    sumY2  +=       5.884 * cos(2.46115086910  + 24.13520490470 *T)
    sumY2  +=       5.760 * cos(5.69510045876  + 103.33659170210 *T)
    sumY2  +=       5.890 * cos(5.39264236310  + 46.71424039951 *T)
    sumY2  +=       6.180 * cos(1.27388326507  + 45.82046852221 *T)
    sumY2  +=       6.347 * cos(2.69417113124  + 227.77000692311 *T)
    sumY2  +=       5.285 * cos(3.95688035619  + 305.10235190919 *T)
    sumY2  +=       5.176 * cos(1.99189201991  + 1052.51220067191 *T)
    sumY2  +=       5.443 * cos(3.22450673881  + 471.93102443120 *T)
    sumY2  +=       5.222 * cos(5.30720065503  + 124.53422035260 *T)
    sumY2  +=       5.424 * cos(2.95723327837  + 91.70019060721 *T)
    sumY2  +=       5.814 * cos(2.93473998779  + 340.52707456129 *T)
    sumY2  +=       5.051 * cos(2.69120338212  + 515.70768857651 *T)
    sumY2  +=       4.924 * cos(5.88676032371  + 404.26297286470 *T)
    sumY2  +=       5.332 * cos(1.01506659832  + 83.10217089810 *T)
    sumY2  +=       5.189 * cos(3.53868828856  + 61.01077037031 *T)
    sumY2  +=       5.198 * cos(5.63042873624  + 486.15811843280 *T)
    sumY2  +=       4.404 * cos(1.90826090989  + 171.89841514590 *T)
    sumY2  +=       5.159 * cos(5.75263925151  + 383.06534421420 *T)
    sumY2  +=       4.708 * cos(3.95368562646  + 110.45013870291 *T)
    sumY2  +=       4.387 * cos(1.58809682622  + 209.61075965840 *T)
    sumY2  +=       4.748 * cos(4.37114817642  + 98.60128928690 *T)
    sumY2  +=       4.248 * cos(2.87090902454  + 230.32075334190 *T)
    sumY2  +=       5.145 * cos(0.08554898309  + 74.06220820430 *T)
    sumY2  +=       4.114 * cos(0.30246244243  + 2.20386307129 *T)
    sumY2  +=       5.063 * cos(1.42908457281  + 75.98862389731 *T)
    sumY2  +=       3.940 * cos(4.34689430728  + 55.38169607781 *T)
    sumY2  +=       4.565 * cos(2.38653118160  + 223.20720634109 *T)
    sumY2  +=       3.744 * cos(4.87117996301  + 43.53284666180 *T)
    sumY2  +=       4.710 * cos(4.97196354244  + 186.45557754760 *T)
    sumY2  +=       3.710 * cos(0.16058312034  + 94.66913602380 *T)
    sumY2  +=       3.673 * cos(1.11490595138  + 80.65449034331 *T)
    sumY2  +=       3.673 * cos(3.56449648462  + 69.39634175830 *T)
    sumY2  +=       3.912 * cos(4.67814184732  + 183.85033042589 *T)
    sumY2  +=       3.744 * cos(5.65406090864  + 155.53915477459 *T)
    sumY2  +=       4.397 * cos(4.53108005666  + 902.19824401191 *T)
    sumY2  +=       4.340 * cos(4.96403760529  + 131.16013238639 *T)
    sumY2  +=       3.600 * cos(0.19066403657  + 61.69209392650 *T)
    sumY2  +=       4.967 * cos(4.79589278540  + 108.21739859670 *T)
    sumY2  +=       3.476 * cos(6.25374915743  + 199.31581891990 *T)
    sumY2  +=       3.658 * cos(3.22126745009  + 849.50804597240 *T)
    sumY2  +=       3.523 * cos(2.12357554923  + 451.22103074770 *T)
    sumY2  +=       4.488 * cos(4.31357043065  + 329.48137429739 *T)
    sumY2  +=       3.980 * cos(0.38853153974  + 36.89238041300 *T)
    sumY2  +=       3.375 * cos(1.20481404948  + 415.30867312860 *T)
    sumY2  +=       3.410 * cos(3.45031076238  + 4.20960064140 *T)
    sumY2  +=       3.568 * cos(1.36229272566  + 446.06752933469 *T)
    sumY2  +=       4.379 * cos(0.32271955362  + 145.35359649321 *T)
    sumY2  +=       3.342 * cos(3.88724158445  + 295.43805848960 *T)
    sumY2  +=       3.570 * cos(4.47738872004  + 50.64639366260 *T)
    sumY2  +=       3.501 * cos(3.71556456950  + 460.29462333630 *T)
    sumY2  +=       3.555 * cos(1.60964374551  + 397.14942586389 *T)
    sumY2  +=       3.345 * cos(4.80678644174  + 544.16187657970 *T)
    sumY2  +=       3.245 * cos(2.89793765327  + 428.32648106781 *T)
    sumY2  +=       3.326 * cos(3.77914879695  + 354.75416856289 *T)
    sumY2  +=       3.517 * cos(0.78060011988  + 204.94489321240 *T)
    sumY2  +=       3.209 * cos(4.56571924389  + 237.43430034270 *T)
    sumY2  +=       3.283 * cos(0.04813757950  + 32.48714639791 *T)
    sumY2  +=       3.195 * cos(4.70738042945  + 1058.14127496441 *T)
    sumY2  +=       3.291 * cos(2.18131341190  + 504.80500066450 *T)
    sumY2  +=       3.210 * cos(0.39206252877  + 323.74923414091 *T)
    sumY2  +=       3.632 * cos(3.85767860548  + 57.75994066341 *T)
    sumY2  +=       3.617 * cos(3.39977309184  + 104.23036357940 *T)
    sumY2  +=       3.794 * cos(1.38786231377  + 294.05665164530 *T)
    sumY2  +=       3.164 * cos(1.44848899722  + 1819.39364862570 *T)
    sumY2  +=       3.165 * cos(2.08884519803  + 216.72430665921 *T)
    sumY2  +=       3.165 * cos(2.08884519803  + 216.23667169220 *T)
    sumY2  +=       3.073 * cos(3.61065539896  + 41.88831525910 *T)
    sumY2  +=       3.059 * cos(0.63894095212  + 905.64276230799 *T)
    sumY2  +=       3.362 * cos(0.98437506221  + 85.58348134769 *T)
    sumY2  +=       3.541 * cos(2.12103995044  + 956.53297345411 *T)
    sumY2  +=       3.541 * cos(4.06722202320  + 191.45151239371 *T)
    sumY2  +=       3.260 * cos(1.97241696611  + 227.55755860200 *T)
    sumY2  +=       3.047 * cos(5.22098984240  + 531.41925528640 *T)
    sumY2  +=       2.983 * cos(3.95282923476  + 1589.31671276730 *T)
    sumY2  +=       3.264 * cos(0.96398524947  + 336.59492129820 *T)
    sumY2  +=       3.273 * cos(5.11144966826  + 60.31068708220 *T)
    sumY2  +=       3.052 * cos(3.15972982100  + 411.37651986550 *T)
    sumY2  +=       2.957 * cos(1.81583171330  + 207.91383862900 *T)
    sumY2  +=       3.082 * cos(0.29643336501  + 217.47506618460 *T)
    sumY2  +=       3.147 * cos(4.43417661382  + 1169.83206889211 *T)
    sumY2  +=       3.055 * cos(0.63724338409  + 381.36842318480 *T)
    sumY2  +=       3.006 * cos(4.98412631966  + 290.87525790760 *T)
    sumY2  +=       2.859 * cos(2.44491972000  + 276.79117625789 *T)
    sumY2  +=       2.836 * cos(4.48873845510  + 88.35873817511 *T)
    sumY2  +=       2.790 * cos(4.32838387144  + 62.00760811160 *T)
    sumY2  +=       3.077 * cos(3.15865718162  + 109.06873185860 *T)
    sumY2  +=       2.839 * cos(0.15460003317  + 942.30587945250 *T)
    sumY2  +=       2.822 * cos(4.85578631925  + 128.71245183160 *T)
    sumY2  +=       3.374 * cos(1.30422404629  + 736.12033101530 *T)
    sumY2  +=       2.933 * cos(3.80915533467  + 306.58682461749 *T)
    sumY2  +=       2.763 * cos(0.71622637926  + 1282.58913653029 *T)
    sumY2  +=       2.895 * cos(4.12741087597  + 75.54668091261 *T)
    sumY2  +=       3.459 * cos(4.50182028789  + 610.44852130189 *T)
    sumY2  +=       2.886 * cos(3.69498226849  + 74.50415118900 *T)
    sumY2  +=       2.638 * cos(5.32601609468  + 198.07742442750 *T)
    sumY2  +=       2.939 * cos(1.02666459482  + 179.91817716280 *T)
    sumY2  +=       2.556 * cos(6.01118563599  + 430.02340209721 *T)
    sumY2  +=       2.743 * cos(2.25858432981  + 141.94270809190 *T)
    sumY2  +=       2.743 * cos(5.40017698340  + 141.45507312490 *T)
    sumY2  +=       2.525 * cos(2.72343515591  + 219.17198721400 *T)
    sumY2  +=       2.654 * cos(6.19077360752  + 475.86317769430 *T)
    sumY2  +=       3.424 * cos(2.06036036899  + 133.13224006171 *T)
    sumY2  +=       2.448 * cos(4.91178796797  + 358.68632182600 *T)
    sumY2  +=       2.613 * cos(1.99336437491  + 1155.60497489050 *T)
    sumY2  +=       2.428 * cos(5.63773248138  + 714.92270236481 *T)
    sumY2  +=       3.038 * cos(1.59738880832  + 145.38722638800 *T)
    sumY2  +=       2.558 * cos(1.72564085187  + 460.88532416730 *T)
    sumY2  +=       2.602 * cos(5.68878295822  + 4.90968392950 *T)
    sumY2  +=       2.602 * cos(5.27380478496  + 145.14114817210 *T)
    sumY2  +=       2.603 * cos(2.43174340516  + 343.46239008310 *T)
    sumY2  +=       3.095 * cos(0.25232061301  + 269.67762925710 *T)
    sumY2  +=       3.095 * cos(3.39391326660  + 270.16526422410 *T)
    sumY2  +=       2.576 * cos(5.00936607273  + 988.02554284310 *T)
    sumY2  +=       2.387 * cos(3.47154008670  + 346.43133549971 *T)
    sumY2  +=       2.643 * cos(4.88232673426  + 376.70255673880 *T)
    sumY2  +=       2.290 * cos(0.33988926240  + 275.30670354960 *T)
    sumY2  +=       3.061 * cos(1.44002014095  + 606.51636803880 *T)
    sumY2  +=       2.264 * cos(2.40394452003  + 323.26159917390 *T)
    sumY2  +=       2.638 * cos(2.36255398659  + 77.71917482150 *T)
    sumY2  +=       2.246 * cos(0.84312304421  + 143.90275367970 *T)
    sumY2  +=       2.334 * cos(3.67388415780  + 49.26498681829 *T)
    sumY2  +=       2.334 * cos(3.67388415780  + 49.75262178531 *T)
    sumY2  +=       2.696 * cos(5.85613797305  + 70.08436295889 *T)
    sumY2  +=       2.327 * cos(4.63449299045  + 479.04457143200 *T)
    sumY2  +=       2.398 * cos(0.98432285449  + 183.48663213100 *T)
    sumY2  +=       2.040 * cos(2.84760871096  + 1968.95684576030 *T)
    sumY2  +=       2.040 * cos(5.98920136455  + 1969.44448072730 *T)
    sumY2  +=       1.960 * cos(1.00781809287  + 894.59706204410 *T)
    sumY2  +=       1.960 * cos(1.00781809287  + 895.08469701110 *T)
    sumY2  +=       2.286 * cos(4.41899075071  + 147.83490694279 *T)
    sumY2  +=       2.536 * cos(0.04650659359  + 108.70503356371 *T)
    sumY2  +=       1.832 * cos(1.44959489041  + 379.88395047650 *T)
    sumY2  +=       1.977 * cos(4.60291692983  + 5.17280848790 *T)
    sumY2  +=       2.397 * cos(0.29658529023  + 114.64292439690 *T)
    sumY2  +=       2.450 * cos(2.97283648816  + 203.98168536590 *T)
    sumY2  +=       2.258 * cos(0.46119788822  + 201.01273994931 *T)
    sumY2  +=       1.767 * cos(1.86154901498  + 1514.04747923300 *T)
    sumY2  +=       1.767 * cos(1.86154901498  + 1514.53511420000 *T)
    sumY2  +=       1.978 * cos(3.70013905116  + 2.72512793310 *T)
    sumY2  +=       1.763 * cos(1.99872322805  + 275.79433851660 *T)
    sumY2  +=       1.714 * cos(0.43161832244  + 440.43845504219 *T)
    sumY2  +=       1.714 * cos(3.57321097603  + 440.92609000920 *T)
    sumY2  +=       1.710 * cos(2.31183617210  + 309.03450517230 *T)
    sumY2  +=       1.710 * cos(5.45342882568  + 309.52214013930 *T)
    sumY2  +=       1.917 * cos(1.47888149586  + 6.18170831670 *T)
    sumY2  +=       1.613 * cos(2.25872913649  + 1044.16025917870 *T)
    sumY2  +=       1.613 * cos(5.40032179008  + 1044.64789414570 *T)
    sumY2  +=       1.722 * cos(0.04871363996  + 752.89817143570 *T)
    sumY2  +=       1.722 * cos(0.04871363996  + 753.38580640270 *T)
    sumY2  +=       1.656 * cos(5.52468183462  + 21.84758304429 *T)
    sumY2  +=       2.072 * cos(2.95520838225  + 203.49405039889 *T)
    sumY2  +=       1.528 * cos(6.14900289425  + 106.51798543980 *T)
    sumY2  +=       1.528 * cos(6.14900289425  + 106.03035047280 *T)
    sumY2  +=       1.802 * cos(2.29662739609  + 344.45922782439 *T)
    sumY2  +=       1.802 * cos(5.43822004968  + 344.94686279140 *T)
    sumY2  +=       1.610 * cos(5.48512651333  + 266.23311096100 *T)
    sumY2  +=       1.509 * cos(0.58599278635  + 924.28964453971 *T)
    sumY2  +=       1.509 * cos(3.72758543994  + 923.80200957270 *T)
    sumY2  +=       1.727 * cos(1.09473115664  + 625.16325027050 *T)
    sumY2  +=       1.727 * cos(1.09473115664  + 624.67561530350 *T)
    sumY2  +=       1.491 * cos(2.48436459152  + 422.90985509640 *T)
    sumY2  +=       1.491 * cos(2.48436459152  + 422.42222012939 *T)
    sumY2  +=       1.766 * cos(0.91839681059  + 224.10097821840 *T)
    sumY2  +=       2.006 * cos(4.96835613839  + 144.86596152620 *T)
    sumY2  +=       1.470 * cos(3.90791054506  + 284.39235822571 *T)
    sumY2  +=       1.470 * cos(3.90791054506  + 283.90472325870 *T)
    sumY2  +=       1.482 * cos(1.95911852464  + 17.91542978119 *T)
    sumY2  +=       1.585 * cos(1.66951998253  + 710.99054910171 *T)
    sumY2  +=       1.585 * cos(1.66951998253  + 710.50291413470 *T)
    sumY2  +=       1.415 * cos(2.35106795911  + 13.98327651810 *T)
    sumY2  +=       1.408 * cos(2.77743270790  + 254.69977573010 *T)
    sumY2  +=       1.408 * cos(2.77743270790  + 255.18741069710 *T)
    sumY2  +=       1.517 * cos(1.11580920740  + 976.97984257920 *T)
    sumY2  +=       1.517 * cos(4.25740186099  + 976.49220761220 *T)

    # URANUS Y Coord
    sumY3  = 0
    sumY3  +=     937.647 * cos(5.72127882108  + 149.80701461810 *T)
    sumY3  +=     967.548 * cos(6.13791702385  + 0.24381748350 *T)
    sumY3  +=     892.089 * cos(5.50847100342  + 75.02541605080 *T)
    sumY3  +=     684.660 * cos(1.33373840136  + 71.09326278771 *T)
    sumY3  +=     623.727 * cos(3.29206615194  + 78.95756931390 *T)
    sumY3  +=     525.597 * cos(2.24127263936  + 76.50988875911 *T)
    sumY3  +=     507.791 * cos(3.07030244888  + 63.97971578690 *T)
    sumY3  +=     496.551 * cos(1.56839683755  + 86.07111631471 *T)
    sumY3  +=     445.874 * cos(3.42824453385  + 71.84402231310 *T)
    sumY3  +=     424.392 * cos(1.23311674112  + 78.20680978850 *T)
    sumY3  +=     366.041 * cos(0.50676115568  + 11.28951774740 *T)
    sumY3  +=     313.023 * cos(2.66852817813  + 73.54094334250 *T)
    sumY3  +=     249.030 * cos(0.91068513695  + 4.17597074660 *T)
    sumY3  +=     209.150 * cos(2.97973004445  + 74.53778108379 *T)
    sumY3  +=     127.465 * cos(3.96938588107  + 138.76131435421 *T)
    sumY3  +=     120.647 * cos(3.89796928399  + 18.40306474821 *T)
    sumY3  +=     121.784 * cos(3.71922038147  + 145.87486135500 *T)
    sumY3  +=      78.251 * cos(1.91023640713  + 224.58861318540 *T)
    sumY3  +=      84.577 * cos(1.90640096807  + 2.93757625420 *T)
    sumY3  +=      50.491 * cos(0.78544030133  + 131.64776735340 *T)
    sumY3  +=      45.951 * cos(4.67460564130  + 151.29148732640 *T)
    sumY3  +=      46.649 * cos(5.26674809079  + 56.86616878610 *T)
    sumY3  +=      48.261 * cos(1.36641385015  + 22.33521801130 *T)
    sumY3  +=      34.791 * cos(5.71401876900  + 93.18466331550 *T)
    sumY3  +=      29.328 * cos(0.18214457213  + 220.65645992230 *T)
    sumY3  +=      30.594 * cos(4.11017743079  + 160.85271488200 *T)
    sumY3  +=      25.988 * cos(5.82957254292  + 153.73916788121 *T)
    sumY3  +=      20.587 * cos(1.47498586610  + 206.42936592071 *T)
    sumY3  +=      18.718 * cos(1.86551335351  + 82.13896305161 *T)
    sumY3  +=      18.634 * cos(2.78869413607  + 67.91186905000 *T)
    sumY3  +=      16.697 * cos(0.44407475674  + 87.55558902300 *T)
    sumY3  +=      20.883 * cos(4.85236834792  + 84.58664360640 *T)
    sumY3  +=      20.072 * cos(6.13552844440  + 65.46418849521 *T)
    sumY3  +=      15.729 * cos(4.26269482524  + 522.82123557730 *T)
    sumY3  +=      15.573 * cos(3.18134712379  + 537.04832957890 *T)
    sumY3  +=      19.429 * cos(3.33529843504  + 127.71561409031 *T)
    sumY3  +=      14.513 * cos(3.12891080528  + 52.44638055600 *T)
    sumY3  +=      14.273 * cos(1.58794880516  + 52.93401552300 *T)
    sumY3  +=      17.225 * cos(5.94745172319  + 146.62562088040 *T)
    sumY3  +=      14.135 * cos(1.26451572688  + 77.99436146740 *T)
    sumY3  +=      13.035 * cos(0.29366962816  + 72.05647063421 *T)
    sumY3  +=      14.349 * cos(4.35039068102  + 148.32254190981 *T)
    sumY3  +=       9.002 * cos(3.16958165565  + 97.11681657861 *T)
    sumY3  +=       8.798 * cos(6.09741460374  + 56.37853381910 *T)
    sumY3  +=       8.208 * cos(4.25355573755  + 62.49524307860 *T)
    sumY3  +=       7.603 * cos(5.85124809704  + 152.98840835580 *T)
    sumY3  +=       7.140 * cos(1.53785661561  + 194.89603068979 *T)
    sumY3  +=       6.892 * cos(5.82729490183  + 80.65449034331 *T)
    sumY3  +=       6.892 * cos(5.13529284134  + 69.39634175830 *T)
    sumY3  +=       6.511 * cos(5.12331075553  + 45.33283355520 *T)
    sumY3  +=       6.477 * cos(0.46092149045  + 209.12312469139 *T)
    sumY3  +=       5.804 * cos(4.85950876891  + 70.60562782070 *T)
    sumY3  +=       5.326 * cos(4.23763472541  + 127.22797912329 *T)
    sumY3  +=       5.029 * cos(5.67665029706  + 149.31937965110 *T)
    sumY3  +=       5.009 * cos(0.31962965160  + 297.98880490840 *T)
    sumY3  +=       5.452 * cos(0.04949420386  + 72.57773549600 *T)
    sumY3  +=       6.294 * cos(1.38143320816  + 10.80188278040 *T)
    sumY3  +=       5.251 * cos(1.43610247201  + 77.47309660560 *T)
    sumY3  +=       4.684 * cos(1.61331269521  + 202.49721265760 *T)
    sumY3  +=       4.291 * cos(4.29965965111  + 134.34152612410 *T)
    sumY3  +=       5.421 * cos(0.15265448442  + 120.11443212250 *T)
    sumY3  +=       4.058 * cos(5.41850959774  + 312.21589891000 *T)
    sumY3  +=       3.926 * cos(1.77715565579  + 59.55992755679 *T)
    sumY3  +=       4.419 * cos(5.50321844296  + 69.60879007940 *T)
    sumY3  +=       4.250 * cos(2.24721757318  + 80.44204202221 *T)
    sumY3  +=       4.556 * cos(5.81169390520  + 29.44876501211 *T)
    sumY3  +=       3.592 * cos(1.55323340563  + 199.31581891990 *T)
    sumY3  +=       3.523 * cos(2.27212859490  + 636.20895053440 *T)
    sumY3  +=       3.428 * cos(3.23337096678  + 621.98185653281 *T)
    sumY3  +=       3.942 * cos(2.26565174816  + 3.21276290011 *T)
    sumY3  +=       3.620 * cos(2.07089246477  + 3.42521122121 *T)
    sumY3  +=       3.227 * cos(3.93329666486  + 3.68833577960 *T)
    sumY3  +=       3.564 * cos(1.40324348330  + 70.35954960481 *T)
    sumY3  +=       3.564 * cos(0.13456629911  + 79.69128249680 *T)
    sumY3  +=       3.552 * cos(5.20225660452  + 426.84200835950 *T)
    sumY3  +=       3.245 * cos(2.48346002496  + 611.82992814620 *T)
    sumY3  +=       3.090 * cos(3.84524306979  + 486.15811843280 *T)
    sumY3  +=       3.149 * cos(4.80301298962  + 471.93102443120 *T)
    sumY3  +=       2.864 * cos(5.97114317460  + 272.85902299479 *T)
    sumY3  +=       3.756 * cos(1.85976896210  + 258.63192899320 *T)
    sumY3  +=       2.808 * cos(4.70529986831  + 647.25465079831 *T)
    sumY3  +=       2.640 * cos(0.64986508116  + 288.32451148881 *T)
    sumY3  +=       2.544 * cos(1.96171476149  + 76.02225379210 *T)
    sumY3  +=       2.670 * cos(4.35862114048  + 375.95179721340 *T)
    sumY3  +=       2.998 * cos(5.42790043104  + 76.72233708021 *T)
    sumY3  +=       2.998 * cos(2.39309465855  + 73.32849502140 *T)
    sumY3  +=       3.125 * cos(0.36892549396  + 408.68276109481 *T)
    sumY3  +=       2.330 * cos(5.58968397865  + 433.95555536030 *T)
    sumY3  +=       2.645 * cos(5.17443993569  + 120.60206708950 *T)
    sumY3  +=       2.219 * cos(4.57629057477  + 390.92965074040 *T)
    sumY3  +=       2.517 * cos(5.76141273726  + 213.54291292150 *T)
    sumY3  +=       2.074 * cos(5.00244199485  + 63.49208081989 *T)
    sumY3  +=       1.894 * cos(3.26252512964  + 213.05527795450 *T)
    sumY3  +=       1.849 * cos(1.04250606380  + 134.82916109110 *T)
    sumY3  +=       1.838 * cos(0.86526030515  + 73.05330837550 *T)
    sumY3  +=       2.214 * cos(5.11129905135  + 1.72829019180 *T)
    sumY3  +=       1.644 * cos(3.09455061710  + 384.30373870660 *T)
    sumY3  +=       1.644 * cos(6.23614327069  + 383.81610373959 *T)
    sumY3  +=       1.553 * cos(3.95294824397  + 462.26673101160 *T)
    sumY3  +=       1.553 * cos(3.95294824397  + 461.77909604459 *T)
    sumY3  +=       1.419 * cos(4.20089846346  + 422.90985509640 *T)
    sumY3  +=       1.419 * cos(4.20089846346  + 422.42222012939 *T)
    sumY3  +=       1.411 * cos(3.36634095175  + 9.80504503910 *T)
    sumY3  +=       1.443 * cos(3.56156060706  + 195.38366565681 *T)
    sumY3  +=       1.402 * cos(2.89352085922  + 984.84414910540 *T)
    sumY3  +=       1.402 * cos(6.03511351281  + 984.35651413839 *T)

    # URANUS Y Coord
    sumY4  = 0
    sumY4  +=    2203.535 * cos(2.21720684091  + 75.02541605080 *T)
    sumY4  +=     128.370 * cos(2.43219641689  + 0.24381748350 *T)
    sumY4  +=      72.353 * cos(4.29968795292  + 149.80701461810 *T)
    sumY4  +=      30.194 * cos(5.08578675405  + 63.97971578690 *T)
    sumY4  +=      30.655 * cos(5.78559557359  + 86.07111631471 *T)
    sumY4  +=      16.471 * cos(2.03791163330  + 18.40306474821 *T)
    sumY4  +=      16.067 * cos(0.96581112775  + 76.50988875911 *T)
    sumY4  +=      16.064 * cos(0.31761628429  + 73.54094334250 *T)
    sumY4  +=      10.244 * cos(4.72416199445  + 74.53778108379 *T)
    sumY4  +=       6.697 * cos(2.54129200384  + 131.64776735340 *T)
    sumY4  +=       7.164 * cos(1.87275979303  + 1.72829019180 *T)
    sumY4  +=       6.834 * cos(0.81126784003  + 56.86616878610 *T)
    sumY4  +=       4.835 * cos(4.01125413225  + 11.28951774740 *T)
    sumY4  +=       5.294 * cos(3.82310032068  + 93.18466331550 *T)
    sumY4  +=       3.031 * cos(4.80949177079  + 71.09326278771 *T)
    sumY4  +=       2.817 * cos(3.00867557843  + 78.95756931390 *T)
    sumY4  +=       2.343 * cos(4.78474509052  + 220.65645992230 *T)
    sumY4  +=       2.389 * cos(2.30337242849  + 283.90472325870 *T)
    sumY4  +=       2.294 * cos(4.60694198931  + 365.65685647490 *T)
    sumY4  +=       2.237 * cos(3.22968908099  + 206.42936592071 *T)
    sumY4  +=       1.842 * cos(3.85262135953  + 138.76131435421 *T)
    sumY4  +=       2.009 * cos(0.58133348752  + 224.58861318540 *T)
    sumY4  +=       1.813 * cos(5.67817174356  + 148.32254190981 *T)

    # URANUS Y Coord
    sumY5  = 0
    sumY5  +=      16.766 * cos(2.31586625461  + 75.02541605080 *T)
    sumY5  +=       1.843 * cos(4.54597723029  + 0.24381748350 *T)

    # URANUS Z Coord
    sumZ0  = 0
    sumZ0  +=  25878127.698 * cos(2.61861272578  + 74.78159856730 *T)
    sumZ0  +=  1774318.778 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ0  +=  599316.131 * cos(5.08119500585  + 149.56319713460 *T)
    sumZ0  +=  190281.890 * cos(1.61643841193  + 76.26607127560 *T)
    sumZ0  +=  190881.685 * cos(0.57869575952  + 73.29712585900 *T)
    sumZ0  +=   84626.761 * cos(2.26030150166  + 1.48447270830 *T)
    sumZ0  +=   30734.257 * cos(0.23571721555  + 63.73589830340 *T)
    sumZ0  +=   20842.052 * cos(1.26054208091  + 224.34479570190 *T)
    sumZ0  +=   19734.273 * cos(6.04314677688  + 148.07872442630 *T)
    sumZ0  +=   12537.530 * cos(5.17169051466  + 11.04570026390 *T)
    sumZ0  +=   14582.864 * cos(6.14852037212  + 71.81265315070 *T)
    sumZ0  +=   10407.529 * cos(3.65320417038  + 213.29909543800 *T)
    sumZ0  +=   11261.541 * cos(3.55973769686  + 529.69096509460 *T)
    sumZ0  +=    8855.669 * cos(4.03774505739  + 151.04766984290 *T)
    sumZ0  +=    8239.460 * cos(0.34225652715  + 77.75054398390 *T)
    sumZ0  +=    7950.169 * cos(0.72564903051  + 85.82729883120 *T)
    sumZ0  +=    6867.469 * cos(0.81417174224  + 2.96894541660 *T)
    sumZ0  +=    5648.720 * cos(3.45324719543  + 138.51749687070 *T)
    sumZ0  +=    4581.938 * cos(1.69668682344  + 38.13303563780 *T)
    sumZ0  +=    2578.399 * cos(5.19696447390  + 111.43016149680 *T)
    sumZ0  +=    2964.070 * cos(6.14338802239  + 35.16409022120 *T)
    sumZ0  +=    1884.104 * cos(2.61192472648  + 78.71375183040 *T)
    sumZ0  +=    2330.304 * cos(5.72640226150  + 70.84944530420 *T)
    sumZ0  +=    1985.215 * cos(0.76408839812  + 39.61750834610 *T)
    sumZ0  +=    1743.154 * cos(1.20586281789  + 146.59425171800 *T)
    sumZ0  +=    2037.011 * cos(0.95353587037  + 70.32818044240 *T)
    sumZ0  +=    1508.924 * cos(2.26195448553  + 222.86032299360 *T)
    sumZ0  +=    1082.736 * cos(2.62872874057  + 108.46121608020 *T)
    sumZ0  +=    1051.968 * cos(0.75560753840  + 109.94568878850 *T)
    sumZ0  +=     860.142 * cos(3.72285572975  + 299.12639426920 *T)
    sumZ0  +=     816.182 * cos(3.95303779460  + 52.69019803950 *T)
    sumZ0  +=     661.249 * cos(3.83459160939  + 36.64856292950 *T)
    sumZ0  +=     712.421 * cos(1.08343941878  + 3.93215326310 *T)
    sumZ0  +=     561.959 * cos(1.87601204328  + 351.81659230870 *T)
    sumZ0  +=     618.657 * cos(5.77033241076  + 4.45341812490 *T)
    sumZ0  +=     526.378 * cos(3.88178280084  + 112.91463420510 *T)
    sumZ0  +=     624.327 * cos(5.55998063360  + 202.25339517410 *T)
    sumZ0  +=     464.130 * cos(0.20598822419  + 225.82926841020 *T)
    sumZ0  +=     448.110 * cos(2.70547740286  + 145.10977900970 *T)
    sumZ0  +=     421.562 * cos(3.66721349765  + 184.72728735580 *T)
    sumZ0  +=     458.133 * cos(5.49113221489  + 79.23501669220 *T)
    sumZ0  +=     437.386 * cos(1.24892718921  + 33.67961751290 *T)
    sumZ0  +=     393.617 * cos(1.25018492386  + 62.25142559510 *T)
    sumZ0  +=     424.721 * cos(2.73945218102  + 152.53214255120 *T)
    sumZ0  +=     465.359 * cos(3.24015058631  + 127.47179660680 *T)
    sumZ0  +=     350.450 * cos(2.57788261436  + 84.34282612290 *T)
    sumZ0  +=     346.833 * cos(2.91261394620  + 426.59819087600 *T)
    sumZ0  +=     375.747 * cos(3.20459646801  + 160.60889739850 *T)
    sumZ0  +=     313.152 * cos(1.32117543131  + 145.63104387150 *T)
    sumZ0  +=     284.456 * cos(1.70356835837  + 22.09140052780 *T)
    sumZ0  +=     265.517 * cos(2.67172684401  + 74.66972398270 *T)
    sumZ0  +=     258.948 * cos(4.12921582346  + 12.53017297220 *T)
    sumZ0  +=     274.419 * cos(3.90646620441  + 1059.38193018920 *T)
    sumZ0  +=     325.409 * cos(2.48214857847  + 9.56122755560 *T)
    sumZ0  +=     260.666 * cos(5.78699886075  + 41.10198105440 *T)
    sumZ0  +=     237.721 * cos(0.80288483705  + 65.22037101170 *T)
    sumZ0  +=     223.998 * cos(3.91250165407  + 221.37585028530 *T)
    sumZ0  +=     220.626 * cos(3.93529110558  + 106.97674337190 *T)
    sumZ0  +=     265.210 * cos(5.70721287265  + 74.89347315190 *T)
    sumZ0  +=     260.727 * cos(0.26432051819  + 277.03499374140 *T)
    sumZ0  +=     201.268 * cos(5.94906398723  + 87.31177153950 *T)
    sumZ0  +=     165.472 * cos(3.54108848226  + 490.33408917940 *T)
    sumZ0  +=     216.763 * cos(2.14355016573  + 68.84370773410 *T)
    sumZ0  +=     166.510 * cos(3.37208863092  + 72.33391801250 *T)
    sumZ0  +=     144.942 * cos(3.24374054293  + 56.62235130260 *T)
    sumZ0  +=     151.752 * cos(3.25335981894  + 71.60020482960 *T)
    sumZ0  +=     145.795 * cos(5.12147352849  + 77.96299230500 *T)
    sumZ0  +=     155.998 * cos(5.19284812595  + 77.22927912210 *T)
    sumZ0  +=     123.397 * cos(2.06100256833  + 186.21176006410 *T)
    sumZ0  +=     123.620 * cos(1.96019194023  + 288.08069400530 *T)
    sumZ0  +=     100.136 * cos(2.82301109906  + 114.39910691340 *T)
    sumZ0  +=      95.939 * cos(4.72085119752  + 297.64192156090 *T)
    sumZ0  +=      93.120 * cos(2.91989514468  + 181.75834193920 *T)
    sumZ0  +=      95.640 * cos(3.53906186354  + 220.41264243880 *T)
    sumZ0  +=     109.824 * cos(3.74385247499  + 67.66805156650 *T)
    sumZ0  +=      85.882 * cos(5.10268841297  + 153.49535039770 *T)
    sumZ0  +=      87.546 * cos(5.06531594976  + 183.24281464750 *T)
    sumZ0  +=     107.550 * cos(1.08025777979  + 340.77089204480 *T)
    sumZ0  +=      81.507 * cos(4.03891440133  + 143.62530630140 *T)
    sumZ0  +=      72.499 * cos(1.26231939492  + 75.30286342910 *T)
    sumZ0  +=      80.746 * cos(0.64185046064  + 206.18554843720 *T)
    sumZ0  +=      66.852 * cos(5.37638250776  + 265.98929347750 *T)
    sumZ0  +=      61.441 * cos(1.93112217884  + 18.15924726470 *T)
    sumZ0  +=      62.039 * cos(4.48334461215  + 66.70484372000 *T)
    sumZ0  +=      73.016 * cos(1.75514255931  + 415.55249061210 *T)
    sumZ0  +=      64.039 * cos(2.09527812031  + 73.81839072080 *T)
    sumZ0  +=      56.535 * cos(5.23288988461  + 628.85158605010 *T)
    sumZ0  +=      70.290 * cos(4.25243387975  + 137.03302416240 *T)
    sumZ0  +=      62.161 * cos(0.01243451360  + 75.74480641380 *T)
    sumZ0  +=      72.251 * cos(0.83002325085  + 74.26033370550 *T)
    sumZ0  +=      54.171 * cos(2.56946172290  + 32.19514480460 *T)
    sumZ0  +=      51.593 * cos(2.16146254563  + 54.17467074780 *T)
    sumZ0  +=      57.081 * cos(4.52969105049  + 5.93789083320 *T)
    sumZ0  +=      50.182 * cos(3.95193155421  + 82.85835341460 *T)
    sumZ0  +=      59.321 * cos(0.59787435734  + 140.00196957900 *T)
    sumZ0  +=      48.938 * cos(3.50893151896  + 7.11354700080 *T)
    sumZ0  +=      46.178 * cos(0.49809404707  + 131.40394986990 *T)
    sumZ0  +=      46.517 * cos(3.03683183762  + 59.80374504030 *T)
    sumZ0  +=      42.340 * cos(3.04902220404  + 218.40690486870 *T)
    sumZ0  +=      39.090 * cos(6.17926332920  + 373.90799283650 *T)
    sumZ0  +=      43.400 * cos(4.99216593131  + 258.02441321480 *T)
    sumZ0  +=      39.572 * cos(5.35887385465  + 219.89137757700 *T)
    sumZ0  +=      34.900 * cos(1.59769199954  + 80.19822453870 *T)
    sumZ0  +=      34.920 * cos(3.23679252932  + 380.12776796000 *T)
    sumZ0  +=      36.403 * cos(1.98850437514  + 81.89514556810 *T)
    sumZ0  +=      40.711 * cos(3.59792341256  + 522.57741809380 *T)
    sumZ0  +=      32.126 * cos(3.27298565306  + 191.20769491020 *T)
    sumZ0  +=      34.349 * cos(5.21152504802  + 105.49227066360 *T)
    sumZ0  +=      32.231 * cos(3.61065476446  + 20.60692781950 *T)
    sumZ0  +=      36.837 * cos(3.48760382716  + 96.87299909510 *T)
    sumZ0  +=      30.915 * cos(4.60433376269  + 42.58645376270 *T)
    sumZ0  +=      28.644 * cos(5.75510160651  + 259.50888592310 *T)
    sumZ0  +=      28.856 * cos(1.34653259405  + 74.52096613640 *T)
    sumZ0  +=      28.208 * cos(4.66642323140  + 159.12442469020 *T)
    sumZ0  +=      29.731 * cos(4.37509990489  + 80.71948940050 *T)
    sumZ0  +=      29.315 * cos(5.03629852507  + 51.20572533120 *T)
    sumZ0  +=      25.879 * cos(3.05776566988  + 8.07675484730 *T)
    sumZ0  +=      26.096 * cos(5.48691240270  + 74.82978267710 *T)
    sumZ0  +=      25.646 * cos(2.63747057104  + 300.61086697750 *T)
    sumZ0  +=      25.720 * cos(2.42332626332  + 305.34616939270 *T)
    sumZ0  +=      25.191 * cos(6.16170714892  + 99.16062095550 *T)
    sumZ0  +=      26.096 * cos(2.89305157919  + 74.73341445750 *T)
    sumZ0  +=      25.722 * cos(4.26142211768  + 180.27386923090 *T)
    sumZ0  +=      24.190 * cos(5.50922979275  + 74.94165726170 *T)
    sumZ0  +=      26.128 * cos(0.13440940381  + 835.03713448730 *T)
    sumZ0  +=      25.366 * cos(0.26521632573  + 536.80451209540 *T)
    sumZ0  +=      24.161 * cos(5.19499722816  + 227.31374111850 *T)
    sumZ0  +=      27.066 * cos(3.36936938207  + 554.06998748280 *T)
    sumZ0  +=      29.888 * cos(1.80973859163  + 3.18139373770 *T)
    sumZ0  +=      25.184 * cos(1.53219969036  + 154.01661525950 *T)
    sumZ0  +=      29.189 * cos(2.73786672090  + 479.28838891550 *T)
    sumZ0  +=      26.685 * cos(2.98667762458  + 14.97785352700 *T)
    sumZ0  +=      22.295 * cos(3.75337744344  + 39.35687591520 *T)
    sumZ0  +=      28.696 * cos(0.74986873597  + 75.04223099820 *T)
    sumZ0  +=      20.972 * cos(3.17916896424  + 255.05546779820 *T)
    sumZ0  +=      21.014 * cos(0.66302313851  + 767.36908292080 *T)
    sumZ0  +=      20.577 * cos(0.98634295841  + 116.42609634290 *T)
    sumZ0  +=      24.248 * cos(2.87116663203  + 74.62153987290 *T)
    sumZ0  +=      21.752 * cos(5.26140008123  + 294.67297614430 *T)
    sumZ0  +=      23.545 * cos(3.16204155741  + 67.35923502580 *T)
    sumZ0  +=      19.430 * cos(5.66067717228  + 235.39049596580 *T)
    sumZ0  +=      22.220 * cos(3.72262913359  + 92.94084583200 *T)
    sumZ0  +=      21.468 * cos(5.59192826712  + 2.44768055480 *T)
    sumZ0  +=      16.392 * cos(0.73044725202  + 142.44965013380 *T)
    sumZ0  +=      17.210 * cos(0.87217558692  + 404.50679034820 *T)
    sumZ0  +=      15.780 * cos(5.95091393024  + 296.15744885260 *T)
    sumZ0  +=      18.777 * cos(0.60348363638  + 211.81462272970 *T)
    sumZ0  +=      15.523 * cos(1.76138162430  + 41.64449777560 *T)
    sumZ0  +=      17.311 * cos(0.80301716069  + 187.69623277240 *T)
    sumZ0  +=      16.318 * cos(3.13602300249  + 28.31117565130 *T)
    sumZ0  +=      15.443 * cos(2.76173273635  + 14.01464568050 *T)
    sumZ0  +=      14.941 * cos(1.86976831313  + 203.73786788240 *T)
    sumZ0  +=      13.508 * cos(1.11474597705  + 984.60033162190 *T)
    sumZ0  +=      14.962 * cos(4.21276743909  + 454.90936652730 *T)
    sumZ0  +=      13.681 * cos(2.06636726694  + 639.89728631400 *T)
    sumZ0  +=      15.178 * cos(2.79287665181  + 50.40257617910 *T)
    sumZ0  +=      13.092 * cos(5.16479782395  + 74.03083904190 *T)
    sumZ0  +=      13.024 * cos(1.54167247979  + 115.88357962170 *T)
    sumZ0  +=      13.756 * cos(4.44130044058  + 216.92243216040 *T)
    sumZ0  +=      12.912 * cos(2.65544175682  + 75.53235809270 *T)
    sumZ0  +=      14.560 * cos(0.31573630495  + 200.76892246580 *T)
    sumZ0  +=      16.675 * cos(4.45252615374  + 125.98732389850 *T)
    sumZ0  +=      13.750 * cos(6.21844937284  + 5.41662597140 *T)
    sumZ0  +=      13.992 * cos(2.89612356381  + 60.76695288680 *T)
    sumZ0  +=      13.034 * cos(0.55043059925  + 23.57587323610 *T)
    sumZ0  +=      11.409 * cos(3.39650826181  + 291.70403072770 *T)
    sumZ0  +=      11.138 * cos(2.91935975246  + 68.18931642830 *T)
    sumZ0  +=      11.850 * cos(4.83551999164  + 149.45132255000 *T)
    sumZ0  +=      11.862 * cos(5.50185243281  + 331.32153907380 *T)
    sumZ0  +=      12.269 * cos(2.11880702802  + 162.09337010680 *T)
    sumZ0  +=      11.814 * cos(5.31010267521  + 142.14083359310 *T)
    sumZ0  +=      10.875 * cos(0.14632073584  + 128.95626931510 *T)
    sumZ0  +=      10.636 * cos(3.41354811558  + 286.59622129700 *T)
    sumZ0  +=      10.932 * cos(0.17212866351  + 230.56457082540 *T)
    sumZ0  +=      13.012 * cos(6.00108633545  + 256.53994050650 *T)
    sumZ0  +=      10.662 * cos(5.52797089925  + 35.42472265210 *T)
    sumZ0  +=      11.802 * cos(1.58760278423  + 149.67507171920 *T)
    sumZ0  +=       9.878 * cos(5.76116968524  + 110.20632121940 *T)
    sumZ0  +=       9.583 * cos(4.26743788620  + 1589.07289528380 *T)
    sumZ0  +=      11.677 * cos(2.40931590777  + 81.00137369080 *T)
    sumZ0  +=       9.947 * cos(4.00367169516  + 760.25553592000 *T)
    sumZ0  +=      11.956 * cos(5.08786325344  + 89.75945209430 *T)
    sumZ0  +=       9.140 * cos(4.92826513743  + 81.37388070630 *T)
    sumZ0  +=       9.041 * cos(5.71300476219  + 419.48464387520 *T)
    sumZ0  +=       8.569 * cos(5.36608418229  + 28.57180808220 *T)
    sumZ0  +=       9.334 * cos(0.45928779675  + 120.35824960600 *T)
    sumZ0  +=       9.162 * cos(3.71396614183  + 362.86229257260 *T)
    sumZ0  +=       8.487 * cos(1.72924017920  + 24.37902238820 *T)
    sumZ0  +=       7.594 * cos(6.10246512358  + 103.09277421860 *T)
    sumZ0  +=       9.947 * cos(5.60590635849  + 68.56182344380 *T)
    sumZ0  +=       7.461 * cos(2.32031732682  + 905.88657979150 *T)
    sumZ0  +=       8.997 * cos(2.37965677543  + 195.13984817330 *T)
    sumZ0  +=       8.243 * cos(4.74708765747  + 316.39186965660 *T)
    sumZ0  +=       7.366 * cos(6.15565920052  + 135.54855145410 *T)
    sumZ0  +=       7.592 * cos(0.05052252659  + 69.36497259590 *T)
    sumZ0  +=       9.926 * cos(4.65530352311  + 88.79624424780 *T)
    sumZ0  +=       7.471 * cos(5.67091902966  + 367.97010200330 *T)
    sumZ0  +=       9.010 * cos(5.15786385091  + 692.58748435350 *T)
    sumZ0  +=       6.852 * cos(5.42203185877  + 146.38180339690 *T)
    sumZ0  +=       6.805 * cos(3.79379417792  + 214.78356814630 *T)
    sumZ0  +=       7.423 * cos(0.10794595769  + 293.18850343600 *T)
    sumZ0  +=       7.905 * cos(4.25870316750  + 260.99335863140 *T)
    sumZ0  +=       8.094 * cos(4.30080188984  + 617.80588578620 *T)
    sumZ0  +=       7.684 * cos(1.67030681336  + 114.13847448250 *T)
    sumZ0  +=       7.620 * cos(0.76643908895  + 152.01087768940 *T)
    sumZ0  +=       6.240 * cos(3.84416768971  + 30.71067209630 *T)
    sumZ0  +=       8.334 * cos(0.23758565081  + 565.11568774670 *T)
    sumZ0  +=       6.097 * cos(5.10863235559  + 329.72519178090 *T)
    sumZ0  +=       5.988 * cos(1.11565079117  + 372.42352012820 *T)
    sumZ0  +=       7.252 * cos(6.05773736503  + 6.21977512350 *T)
    sumZ0  +=       6.028 * cos(5.24025259251  + 95.38852638680 *T)
    sumZ0  +=       6.289 * cos(3.47252263966  + 73.40900044360 *T)
    sumZ0  +=       6.000 * cos(5.61195390119  + 157.63995198190 *T)
    sumZ0  +=       5.425 * cos(1.81773059325  + 647.01083331480 *T)
    sumZ0  +=       5.756 * cos(0.60729000613  + 45.57665103870 *T)
    sumZ0  +=       5.331 * cos(0.57774785878  + 462.02291352810 *T)
    sumZ0  +=       5.807 * cos(2.22669544081  + 358.93013930950 *T)
    sumZ0  +=       5.826 * cos(5.23125320066  + 284.14854074220 *T)
    sumZ0  +=       6.314 * cos(0.18733033209  + 10213.28554621100 *T)
    sumZ0  +=       6.532 * cos(3.68538415222  + 278.51946644970 *T)
    sumZ0  +=       5.337 * cos(2.55697146693  + 543.02428721890 *T)
    sumZ0  +=       5.214 * cos(3.23506131011  + 7.42236354150 *T)
    sumZ0  +=       5.726 * cos(0.79025128961  + 483.22054217860 *T)
    sumZ0  +=       6.176 * cos(4.09929403804  + 328.35259365720 *T)
    sumZ0  +=       6.858 * cos(5.69002837482  + 141.48644228730 *T)
    sumZ0  +=       5.025 * cos(2.82710574399  + 152.74459087230 *T)
    sumZ0  +=       4.960 * cos(3.86366490265  + 46.20979048510 *T)
    sumZ0  +=       5.299 * cos(0.28596890733  + 0.96320784650 *T)
    sumZ0  +=       4.711 * cos(1.04008534095  + 1332.48477066750 *T)
    sumZ0  +=       5.322 * cos(5.53930777799  + 477.91579079180 *T)
    sumZ0  +=       4.560 * cos(5.18525746864  + 1215.16490244730 *T)
    sumZ0  +=       4.528 * cos(5.38100183232  + 983.11585891360 *T)
    sumZ0  +=       4.640 * cos(1.37720813333  + 121.25202148330 *T)
    sumZ0  +=       4.414 * cos(5.22232716227  + 909.81873305460 *T)
    sumZ0  +=       4.365 * cos(2.15062971711  + 611.58611066270 *T)

    # URANUS Z Coord
    sumZ1  = 0
    sumZ1  +=  3962262.983 * cos(4.12418900865  + 74.78159856730 *T)
    sumZ1  +=   82241.017 * cos(0.33841633701  + 149.56319713460 *T)
    sumZ1  +=   33563.127 * cos(2.12025871831  + 73.29712585900 *T)
    sumZ1  +=   39689.344 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ1  +=   26107.421 * cos(3.06937675063  + 76.26607127560 *T)
    sumZ1  +=   12908.640 * cos(3.77705052097  + 1.48447270830 *T)
    sumZ1  +=    3540.856 * cos(3.75158221944  + 63.73589830340 *T)
    sumZ1  +=    2531.606 * cos(2.84542573949  + 224.34479570190 *T)
    sumZ1  +=    3003.205 * cos(1.25505343184  + 148.07872442630 *T)
    sumZ1  +=    1622.011 * cos(3.59519265641  + 71.81265315070 *T)
    sumZ1  +=    1113.725 * cos(5.37716047615  + 138.51749687070 *T)
    sumZ1  +=    1221.565 * cos(1.20355053543  + 11.04570026390 *T)
    sumZ1  +=    1076.837 * cos(5.57432235156  + 151.04766984290 *T)
    sumZ1  +=     957.872 * cos(3.39246978049  + 85.82729883120 *T)
    sumZ1  +=     881.547 * cos(5.69920992505  + 529.69096509460 *T)
    sumZ1  +=     942.010 * cos(5.40707153970  + 213.29909543800 *T)
    sumZ1  +=    1015.398 * cos(1.70749191167  + 77.75054398390 *T)
    sumZ1  +=     762.361 * cos(4.44942686053  + 78.71375183040 *T)
    sumZ1  +=     861.921 * cos(2.56138068523  + 2.96894541660 *T)
    sumZ1  +=     414.164 * cos(3.67975318141  + 70.84944530420 *T)
    sumZ1  +=     468.225 * cos(3.85273975522  + 38.13303563780 *T)
    sumZ1  +=     312.200 * cos(2.64302370891  + 3.93215326310 *T)
    sumZ1  +=     270.537 * cos(1.07042111099  + 111.43016149680 *T)
    sumZ1  +=     229.316 * cos(2.48137482804  + 146.59425171800 *T)
    sumZ1  +=     211.208 * cos(5.63209083386  + 35.16409022120 *T)
    sumZ1  +=     206.882 * cos(3.77563841652  + 222.86032299360 *T)
    sumZ1  +=     226.153 * cos(2.76382998968  + 39.61750834610 *T)
    sumZ1  +=     152.516 * cos(1.15795669733  + 65.22037101170 *T)
    sumZ1  +=     143.635 * cos(2.29883148753  + 109.94568878850 *T)
    sumZ1  +=     137.496 * cos(1.27325091137  + 52.69019803950 *T)
    sumZ1  +=     126.912 * cos(4.98433527702  + 71.60020482960 *T)
    sumZ1  +=      91.423 * cos(5.37075683718  + 299.12639426920 *T)
    sumZ1  +=      78.168 * cos(3.30117226546  + 77.96299230500 *T)
    sumZ1  +=      71.984 * cos(4.77848755964  + 70.32818044240 *T)
    sumZ1  +=      53.505 * cos(4.24118032837  + 74.66972398270 *T)
    sumZ1  +=      56.461 * cos(4.74607474496  + 184.72728735580 *T)
    sumZ1  +=      64.129 * cos(1.40401800165  + 4.45341812490 *T)
    sumZ1  +=      60.190 * cos(4.25563172540  + 22.09140052780 *T)
    sumZ1  +=      49.324 * cos(1.81501026461  + 225.82926841020 *T)
    sumZ1  +=      51.044 * cos(4.39586820753  + 36.64856292950 *T)
    sumZ1  +=      52.135 * cos(5.22811836344  + 56.62235130260 *T)
    sumZ1  +=      45.079 * cos(5.97279472842  + 145.63104387150 *T)
    sumZ1  +=      47.486 * cos(0.40482409134  + 77.22927912210 *T)
    sumZ1  +=      42.020 * cos(5.25755630615  + 202.25339517410 *T)
    sumZ1  +=      45.284 * cos(4.24607325536  + 152.53214255120 *T)
    sumZ1  +=      38.647 * cos(3.45355985008  + 351.81659230870 *T)
    sumZ1  +=      50.009 * cos(0.53909767456  + 79.23501669220 *T)
    sumZ1  +=      39.778 * cos(4.81127483225  + 62.25142559510 *T)
    sumZ1  +=      40.898 * cos(0.10596214023  + 112.91463420510 *T)
    sumZ1  +=      45.145 * cos(5.98712490700  + 160.60889739850 *T)
    sumZ1  +=      35.920 * cos(2.17013246060  + 127.47179660680 *T)
    sumZ1  +=      31.263 * cos(0.58715748016  + 153.49535039770 *T)
    sumZ1  +=      31.754 * cos(2.81745075870  + 277.03499374140 *T)
    sumZ1  +=      36.970 * cos(2.35222730494  + 206.18554843720 *T)
    sumZ1  +=      31.783 * cos(4.79422665078  + 221.37585028530 *T)
    sumZ1  +=      36.288 * cos(0.84576342160  + 33.67961751290 *T)
    sumZ1  +=      28.318 * cos(6.06676080756  + 426.59819087600 *T)
    sumZ1  +=      29.023 * cos(1.00910858838  + 75.74480641380 *T)
    sumZ1  +=      25.619 * cos(1.75329899148  + 220.41264243880 *T)
    sumZ1  +=      28.552 * cos(5.54335981183  + 67.66805156650 *T)
    sumZ1  +=      25.386 * cos(5.13375412515  + 84.34282612290 *T)
    sumZ1  +=      25.695 * cos(1.62197123657  + 41.10198105440 *T)
    sumZ1  +=      27.612 * cos(0.99170994448  + 74.89347315190 *T)
    sumZ1  +=      20.608 * cos(2.21883575454  + 131.40394986990 *T)
    sumZ1  +=      24.067 * cos(4.33103426983  + 9.56122755560 *T)
    sumZ1  +=      23.899 * cos(2.43498266206  + 87.31177153950 *T)
    sumZ1  +=      23.625 * cos(6.27443196421  + 18.15924726470 *T)
    sumZ1  +=      24.706 * cos(0.26779002293  + 12.53017297220 *T)
    sumZ1  +=      24.563 * cos(0.18402440195  + 3.18139373770 *T)
    sumZ1  +=      19.905 * cos(3.48825148521  + 186.21176006410 *T)
    sumZ1  +=      15.547 * cos(0.20397814976  + 137.03302416240 *T)
    sumZ1  +=      15.494 * cos(4.31583400109  + 288.08069400530 *T)
    sumZ1  +=      13.835 * cos(0.02174450243  + 81.89514556810 *T)
    sumZ1  +=      13.597 * cos(0.15357754975  + 297.64192156090 *T)
    sumZ1  +=      13.359 * cos(3.17889804773  + 80.19822453870 *T)
    sumZ1  +=      13.294 * cos(3.87688117315  + 145.10977900970 *T)
    sumZ1  +=      17.024 * cos(1.81428109960  + 73.81839072080 *T)
    sumZ1  +=      14.748 * cos(0.03189985172  + 66.70484372000 *T)
    sumZ1  +=      12.232 * cos(2.64378420008  + 75.30286342910 *T)
    sumZ1  +=      11.124 * cos(5.08426227737  + 490.33408917940 *T)
    sumZ1  +=      12.563 * cos(1.82692572459  + 140.00196957900 *T)
    sumZ1  +=      13.411 * cos(5.39551143522  + 522.57741809380 *T)
    sumZ1  +=      12.192 * cos(2.21125075603  + 74.26033370550 *T)
    sumZ1  +=      11.266 * cos(0.79673017651  + 340.77089204480 *T)
    sumZ1  +=       9.590 * cos(6.09094318692  + 96.87299909510 *T)
    sumZ1  +=       9.224 * cos(2.14550163776  + 7.11354700080 *T)
    sumZ1  +=       8.877 * cos(0.46929853551  + 2.44768055480 *T)
    sumZ1  +=       9.874 * cos(5.79572540180  + 183.24281464750 *T)
    sumZ1  +=      10.128 * cos(5.09593396273  + 114.39910691340 *T)
    sumZ1  +=       8.837 * cos(4.70446704205  + 14.97785352700 *T)
    sumZ1  +=       9.469 * cos(6.21249255999  + 108.46121608020 *T)
    sumZ1  +=       7.547 * cos(5.32002357346  + 82.85835341460 *T)
    sumZ1  +=       7.560 * cos(4.69622310205  + 536.80451209540 *T)
    sumZ1  +=       7.550 * cos(4.37352817861  + 415.55249061210 *T)
    sumZ1  +=       7.800 * cos(1.81108816316  + 92.94084583200 *T)
    sumZ1  +=       8.448 * cos(5.75611469277  + 1059.38193018920 *T)
    sumZ1  +=       8.261 * cos(2.80861697520  + 72.33391801250 *T)
    sumZ1  +=       8.096 * cos(6.27869649292  + 5.93789083320 *T)
    sumZ1  +=       6.564 * cos(1.27991694625  + 373.90799283650 *T)
    sumZ1  +=       5.973 * cos(3.96626028042  + 617.80588578620 *T)
    sumZ1  +=       5.775 * cos(2.43822838062  + 142.44965013380 *T)
    sumZ1  +=       5.356 * cos(4.89393300983  + 41.64449777560 *T)
    sumZ1  +=       7.123 * cos(5.82803873586  + 191.20769491020 *T)
    sumZ1  +=       6.074 * cos(1.61805362044  + 305.34616939270 *T)
    sumZ1  +=       5.023 * cos(1.26471022025  + 703.63318461740 *T)
    sumZ1  +=       4.868 * cos(0.30672295767  + 33.13710079170 *T)
    sumZ1  +=       4.986 * cos(4.65490725829  + 195.13984817330 *T)
    sumZ1  +=       5.616 * cos(2.21173751117  + 32.19514480460 *T)
    sumZ1  +=       4.445 * cos(4.80807045815  + 20.60692781950 *T)
    sumZ1  +=       5.055 * cos(5.34765264557  + 80.71948940050 *T)
    sumZ1  +=       4.398 * cos(0.49529839431  + 74.82978267710 *T)
    sumZ1  +=       4.398 * cos(4.18462288684  + 74.73341445750 *T)
    sumZ1  +=       4.853 * cos(2.63817804331  + 74.52096613640 *T)
    sumZ1  +=       4.836 * cos(2.04143912495  + 75.04223099820 *T)
    sumZ1  +=       5.022 * cos(0.61491281306  + 259.50888592310 *T)
    sumZ1  +=       4.288 * cos(3.92852240171  + 300.61086697750 *T)
    sumZ1  +=       4.371 * cos(6.00527473515  + 159.12442469020 *T)
    sumZ1  +=       4.091 * cos(4.16218259902  + 74.62153987290 *T)
    sumZ1  +=       4.076 * cos(0.51761494342  + 74.94165726170 *T)
    sumZ1  +=       5.085 * cos(0.47739127862  + 181.75834193920 *T)
    sumZ1  +=       4.223 * cos(0.10445996392  + 42.58645376270 *T)
    sumZ1  +=       4.941 * cos(2.80437422101  + 479.28838891550 *T)
    sumZ1  +=       3.301 * cos(4.32752251719  + 116.42609634290 *T)
    sumZ1  +=       4.284 * cos(2.75000237791  + 154.01661525950 *T)
    sumZ1  +=       4.073 * cos(0.16290844853  + 227.31374111850 *T)
    sumZ1  +=       3.569 * cos(5.93367812903  + 219.89137757700 *T)
    sumZ1  +=       2.877 * cos(0.92771324396  + 404.50679034820 *T)
    sumZ1  +=       3.606 * cos(4.15115668974  + 39.35687591520 *T)
    sumZ1  +=       2.930 * cos(5.67425699047  + 68.84370773410 *T)
    sumZ1  +=       3.216 * cos(0.72597245494  + 235.39049596580 *T)
    sumZ1  +=       2.349 * cos(1.20639876458  + 5.41662597140 *T)
    sumZ1  +=       2.291 * cos(0.45375065997  + 74.03083904190 *T)
    sumZ1  +=       2.770 * cos(1.84988185963  + 211.81462272970 *T)
    sumZ1  +=       2.264 * cos(4.22640788890  + 75.53235809270 *T)
    sumZ1  +=       2.775 * cos(2.27306099902  + 187.69623277240 *T)
    sumZ1  +=       3.034 * cos(3.74888989756  + 380.12776796000 *T)
    sumZ1  +=       2.614 * cos(0.46579056125  + 128.95626931510 *T)
    sumZ1  +=       2.608 * cos(0.85101218281  + 296.15744885260 *T)
    sumZ1  +=       2.641 * cos(5.51157783762  + 230.56457082540 *T)
    sumZ1  +=       2.427 * cos(0.21201331744  + 835.03713448730 *T)
    sumZ1  +=       2.706 * cos(0.68084398426  + 99.16062095550 *T)
    sumZ1  +=       2.700 * cos(4.01778306134  + 50.40257617910 *T)
    sumZ1  +=       2.306 * cos(1.00290501307  + 35.42472265210 *T)
    sumZ1  +=       2.095 * cos(5.76874685766  + 54.17467074780 *T)
    sumZ1  +=       1.942 * cos(4.49004561030  + 68.18931642830 *T)
    sumZ1  +=       2.068 * cos(0.12254293382  + 149.45132255000 *T)
    sumZ1  +=       2.666 * cos(0.45683512473  + 110.20632121940 *T)
    sumZ1  +=       1.945 * cos(6.21758865027  + 89.75945209430 *T)
    sumZ1  +=       2.386 * cos(3.96907647535  + 14.01464568050 *T)
    sumZ1  +=       1.896 * cos(5.34437105960  + 59.80374504030 *T)
    sumZ1  +=       2.065 * cos(3.15839878322  + 149.67507171920 *T)
    sumZ1  +=       1.777 * cos(4.00647819013  + 265.98929347750 *T)
    sumZ1  +=       2.209 * cos(0.64028748776  + 218.40690486870 *T)
    sumZ1  +=       2.162 * cos(1.24773830053  + 106.97674337190 *T)
    sumZ1  +=       2.023 * cos(1.82007872482  + 23.57587323610 *T)
    sumZ1  +=       2.035 * cos(0.51927845768  + 200.76892246580 *T)
    sumZ1  +=       1.504 * cos(0.25796780548  + 81.37388070630 *T)
    sumZ1  +=       1.960 * cos(3.46041862039  + 162.09337010680 *T)
    sumZ1  +=       1.414 * cos(2.99649402595  + 51.20572533120 *T)

    # URANUS Z Coord
    sumZ2  = 0
    sumZ2  +=  177117.836 * cos(5.80022816704  + 74.78159856730 *T)
    sumZ2  +=   16035.660 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ2  +=    2755.730 * cos(2.17676848750  + 149.56319713460 *T)
    sumZ2  +=    1828.256 * cos(3.84635676237  + 73.29712585900 *T)
    sumZ2  +=     868.510 * cos(4.87897232400  + 76.26607127560 *T)
    sumZ2  +=     572.523 * cos(5.46404733048  + 1.48447270830 *T)
    sumZ2  +=     285.628 * cos(5.08160639995  + 63.73589830340 *T)
    sumZ2  +=     157.189 * cos(0.91294082225  + 138.51749687070 *T)
    sumZ2  +=     152.299 * cos(6.26275017065  + 78.71375183040 *T)
    sumZ2  +=     141.686 * cos(2.84004707284  + 148.07872442630 *T)
    sumZ2  +=      93.961 * cos(3.52447729329  + 11.04570026390 *T)
    sumZ2  +=      92.321 * cos(5.16432077685  + 71.81265315070 *T)
    sumZ2  +=      65.470 * cos(0.00153662660  + 85.82729883120 *T)
    sumZ2  +=      66.123 * cos(4.23405260860  + 3.93215326310 *T)
    sumZ2  +=      59.402 * cos(5.03486983197  + 224.34479570190 *T)
    sumZ2  +=      50.738 * cos(0.41363094297  + 71.60020482960 *T)
    sumZ2  +=      39.303 * cos(1.43398963413  + 529.69096509460 *T)
    sumZ2  +=      41.691 * cos(0.60405160402  + 213.29909543800 *T)
    sumZ2  +=      28.845 * cos(4.32177165196  + 2.96894541660 *T)
    sumZ2  +=      24.452 * cos(1.45233443794  + 151.04766984290 *T)
    sumZ2  +=      22.558 * cos(3.68734701846  + 77.75054398390 *T)
    sumZ2  +=      20.948 * cos(5.56707525527  + 38.13303563780 *T)
    sumZ2  +=      18.328 * cos(1.52738789991  + 77.96299230500 *T)
    sumZ2  +=      14.050 * cos(4.09291571100  + 146.59425171800 *T)
    sumZ2  +=      12.166 * cos(2.77134894443  + 111.43016149680 *T)
    sumZ2  +=      12.047 * cos(3.14714635296  + 65.22037101170 *T)
    sumZ2  +=      10.691 * cos(4.20068087609  + 35.16409022120 *T)
    sumZ2  +=       8.575 * cos(1.32099488216  + 70.84944530420 *T)
    sumZ2  +=       8.712 * cos(3.96782908270  + 109.94568878850 *T)
    sumZ2  +=       8.425 * cos(2.12601661788  + 56.62235130260 *T)
    sumZ2  +=       9.384 * cos(5.01391767290  + 3.18139373770 *T)
    sumZ2  +=       7.268 * cos(5.60415789247  + 127.47179660680 *T)
    sumZ2  +=       7.110 * cos(3.80523625157  + 454.90936652730 *T)
    sumZ2  +=       7.227 * cos(5.55318237104  + 222.86032299360 *T)
    sumZ2  +=       6.941 * cos(0.53699361276  + 22.09140052780 *T)
    sumZ2  +=       7.547 * cos(4.53131604233  + 39.61750834610 *T)
    sumZ2  +=       7.103 * cos(3.99578891327  + 52.69019803950 *T)
    sumZ2  +=       6.166 * cos(6.23459907887  + 70.32818044240 *T)
    sumZ2  +=       5.037 * cos(2.75663731170  + 12.53017297220 *T)
    sumZ2  +=       5.681 * cos(2.88511085602  + 160.60889739850 *T)
    sumZ2  +=       4.463 * cos(4.73686825389  + 87.31177153950 *T)
    sumZ2  +=       4.239 * cos(1.56497120608  + 9.56122755560 *T)
    sumZ2  +=       4.139 * cos(5.80700510713  + 74.66972398270 *T)
    sumZ2  +=       4.953 * cos(1.62290961095  + 84.34282612290 *T)
    sumZ2  +=       3.989 * cos(3.71609335244  + 18.15924726470 *T)
    sumZ2  +=       5.068 * cos(1.82574036051  + 77.22927912210 *T)
    sumZ2  +=       4.044 * cos(3.72554698482  + 202.25339517410 *T)
    sumZ2  +=       3.751 * cos(2.22353174536  + 153.49535039770 *T)
    sumZ2  +=       3.700 * cos(2.26416938160  + 75.74480641380 *T)
    sumZ2  +=       3.670 * cos(2.32544430817  + 145.63104387150 *T)
    sumZ2  +=       3.643 * cos(2.14599111308  + 299.12639426920 *T)
    sumZ2  +=       3.263 * cos(2.75844544119  + 73.81839072080 *T)
    sumZ2  +=       2.446 * cos(6.17908901556  + 62.25142559510 *T)
    sumZ2  +=       2.232 * cos(6.26646620658  + 220.41264243880 *T)
    sumZ2  +=       2.529 * cos(3.41304662369  + 131.40394986990 *T)
    sumZ2  +=       2.669 * cos(6.01785430306  + 184.72728735580 *T)
    sumZ2  +=       1.937 * cos(4.88080111657  + 225.82926841020 *T)
    sumZ2  +=       2.394 * cos(6.16425420001  + 72.33391801250 *T)
    sumZ2  +=       1.693 * cos(3.21375560117  + 92.94084583200 *T)
    sumZ2  +=       1.730 * cos(0.61525482065  + 36.64856292950 *T)
    sumZ2  +=       1.868 * cos(5.20707362099  + 145.10977900970 *T)
    sumZ2  +=       1.707 * cos(4.36686229005  + 206.18554843720 *T)
    sumZ2  +=       1.555 * cos(2.47637875795  + 426.59819087600 *T)
    sumZ2  +=       1.473 * cos(4.91621635684  + 80.19822453870 *T)
    sumZ2  +=       1.409 * cos(3.59004232634  + 4.45341812490 *T)

    # URANUS Z Coord
    sumZ3  = 0
    sumZ3  +=    5141.166 * cos(1.25274291045  + 74.78159856730 *T)
    sumZ3  +=     316.609 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ3  +=      65.271 * cos(5.77590506888  + 73.29712585900 *T)
    sumZ3  +=      59.071 * cos(4.00961080499  + 149.56319713460 *T)
    sumZ3  +=      34.006 * cos(1.04273395751  + 63.73589830340 *T)
    sumZ3  +=      23.565 * cos(1.66204476861  + 78.71375183040 *T)
    sumZ3  +=      20.797 * cos(0.28508462597  + 76.26607127560 *T)
    sumZ3  +=      17.194 * cos(0.82951296239  + 1.48447270830 *T)
    sumZ3  +=      12.086 * cos(2.51385690941  + 138.51749687070 *T)
    sumZ3  +=      12.289 * cos(1.87238784591  + 71.60020482960 *T)
    sumZ3  +=       6.582 * cos(5.70015947451  + 3.93215326310 *T)
    sumZ3  +=       4.929 * cos(4.05567973712  + 85.82729883120 *T)
    sumZ3  +=       3.901 * cos(0.60406901282  + 71.81265315070 *T)
    sumZ3  +=       4.647 * cos(4.41312919070  + 148.07872442630 *T)
    sumZ3  +=       3.312 * cos(4.65337102453  + 70.84944530420 *T)
    sumZ3  +=       3.697 * cos(2.79328856468  + 11.04570026390 *T)
    sumZ3  +=       1.714 * cos(5.43238781955  + 145.63104387150 *T)
    sumZ3  +=       1.496 * cos(5.73048296712  + 18.15924726470 *T)

    # URANUS Z Coord
    sumZ4  = 0
    sumZ4  +=     109.765 * cos(2.85499529315  + 74.78159856730 *T)
    sumZ4  +=       7.890 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ4  +=       2.125 * cos(2.94706053740  + 63.73589830340 *T)
    sumZ4  +=       2.092 * cos(1.68142886919  + 73.29712585900 *T)
    sumZ4  +=       1.559 * cos(5.50762023733  + 149.56319713460 *T)
    sumZ4  +=       1.558 * cos(1.27604541567  + 70.84944530420 *T)

    # URANUS Z Coord
    sumZ5  = 0
    sumZ5  +=       2.165 * cos(4.87604439769  + 74.78159856730 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Uranus",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getUranusData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getUranusXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#!/usr/bin/env python
# coding: utf-8
#   getVenusXYZData, based on Meeus, chapter 30
#   and VSOP87 data tables, 
#   ftp://ftp.imcce.fr/pub/ephem/planets/vsop87/ 
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#       NEW: hvfTime object
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   Dict        Dictionary with specific output values for each planet
#       Object:  "Venus"
#       X     :  X coord (unit: au)
#       Y     :  Y coord (unit: au)
#       Z     :  Z coord (unit: au)
#       Long  :  heliocentric longitude, degrees
#       Lat   :  heliocentric latitude, degrees
#       Rau   :  radius, distance Object - Sun (unit: au)
#
#   Similarly for each planet
#   data from VSOP87
#   replace TDT by TT (terrestrial time) - hvf 06.04.2023

Version = "V2.1 hvf 2022-11-03 getVenusXYZData.py created by mkPlanetXYZData.py"
VersionList.append(Version)

import math

def getVenusXYZData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    T = TT / 365250.    # eq. 31.1 Julian Millennia

#######################################################################

    # VENUS X Coord
    sumX0  = 0
    sumX0  +=  72268045.621 * cos(3.17614669179  + 10213.52936369450 *T)
    sumX0  +=  733886.107 * cos(5.43699242686  + 0.24381748350 *T)
    sumX0  +=  244692.613 * cos(4.05605630888  + 20426.81490990550 *T)
    sumX0  +=   63317.702 * cos(0.49335411201  + 10213.04172872750 *T)
    sumX0  +=    2795.043 * cos(0.33061720033  + 2353.10997125530 *T)
    sumX0  +=    1941.766 * cos(4.23356531450  + 1577.09972496430 *T)
    sumX0  +=    1242.687 * cos(4.93613589610  + 30640.10045611650 *T)
    sumX0  +=    1163.369 * cos(2.88004293502  + 18073.94875613370 *T)
    sumX0  +=    1044.280 * cos(1.75322105845  + 6283.31966747490 *T)
    sumX0  +=     767.420 * cos(0.59253172189  + 529.93478257810 *T)
    sumX0  +=     669.994 * cos(1.45757409332  + 14143.73905991410 *T)
    sumX0  +=     657.435 * cos(0.50103996032  + 8636.18582124671 *T)
    sumX0  +=     476.814 * cos(5.84349432310  + 10187.23104389470 *T)
    sumX0  +=     475.078 * cos(3.65048802648  + 10239.82768349430 *T)
    sumX0  +=     559.353 * cos(1.16579177696  + 22004.15845235330 *T)
    sumX0  +=     544.577 * cos(2.71804028246  + 11790.87290614230 *T)
    sumX0  +=     408.020 * cos(3.92608771235  + 775.76642880750 *T)
    sumX0  +=     281.345 * cos(2.80298571306  + 9683.83839859990 *T)
    sumX0  +=     270.035 * cos(3.67770752796  + 9438.00675237050 *T)
    sumX0  +=     268.860 * cos(0.41990030277  + 10743.22032878910 *T)
    sumX0  +=     299.980 * cos(5.65601998711  + 5507.30942118390 *T)
    sumX0  +=     214.307 * cos(0.74923674788  + 10022.08109758290 *T)
    sumX0  +=     241.347 * cos(5.80682329886  + 10989.05197501850 *T)
    sumX0  +=     208.372 * cos(2.46285909236  + 10404.97762980610 *T)
    sumX0  +=     271.224 * cos(2.62417096158  + 19897.12394481090 *T)
    sumX0  +=     264.276 * cos(5.12064283406  + 9153.65979853830 *T)
    sumX0  +=     214.535 * cos(1.37526779747  + 20426.32727493849 *T)
    sumX0  +=     175.488 * cos(6.12331394777  + 191.20444862810 *T)
    sumX0  +=     123.055 * cos(1.58857792634  + 1059.62574767270 *T)
    sumX0  +=     154.230 * cos(5.35560161427  + 4705.97612502710 *T)
    sumX0  +=     147.037 * cos(5.55854050911  + 19651.29229858150 *T)
    sumX0  +=     100.153 * cos(0.79332045392  + 9437.51911740350 *T)
    sumX0  +=      69.555 * cos(0.09311754829  + 9104.15081160110 *T)
    sumX0  +=      60.317 * cos(3.11675324526  + 11322.90791578790 *T)
    sumX0  +=      81.789 * cos(4.14338512686  + 15721.08260236190 *T)
    sumX0  +=      58.874 * cos(3.05764615349  + 20618.26317601710 *T)
    sumX0  +=      60.776 * cos(3.78470873239  + 4552.19731454230 *T)
    sumX0  +=      46.411 * cos(0.52070218977  + 10192.75396820210 *T)
    sumX0  +=      46.285 * cos(2.68960619882  + 10234.30475918690 *T)
    sumX0  +=      62.288 * cos(2.59109725374  + 25934.36814857290 *T)
    sumX0  +=      50.365 * cos(4.75194275844  + 8623.96883344369 *T)
    sumX0  +=      57.275 * cos(0.64239189151  + 3154.44326741210 *T)
    sumX0  +=      44.993 * cos(1.63208209330  + 19367.43297971630 *T)
    sumX0  +=      40.167 * cos(3.21167128301  + 9411.70843257070 *T)
    sumX0  +=      47.716 * cos(0.86304453306  + 213.54291292150 *T)
    sumX0  +=      39.763 * cos(1.64025153322  + 801.57711364030 *T)
    sumX0  +=      36.106 * cos(0.00294457034  + 11015.35029481830 *T)
    sumX0  +=      45.150 * cos(2.05121968981  + 29580.71852592730 *T)
    sumX0  +=      36.973 * cos(4.39488677348  + 7058.84227879890 *T)
    sumX0  +=      44.928 * cos(2.21932487684  + 13367.72881362310 *T)
    sumX0  +=      27.822 * cos(1.80340755563  + 1109.13473460990 *T)
    sumX0  +=      31.714 * cos(5.71016699928  + 15874.86141284670 *T)
    sumX0  +=      36.279 * cos(3.91641576039  + 10000.23026825651 *T)
    sumX0  +=      35.360 * cos(5.58217966591  + 10426.82845913250 *T)
    sumX0  +=      25.530 * cos(5.55124903584  + 7860.66320992270 *T)
    sumX0  +=      24.951 * cos(5.10701160912  + 13368.21644859010 *T)
    sumX0  +=      20.285 * cos(2.70230185537  + 10220.64291069530 *T)
    sumX0  +=      20.275 * cos(0.50470014586  + 10206.41581669370 *T)
    sumX0  +=      21.291 * cos(2.94920881997  + 3128.63258257930 *T)
    sumX0  +=      19.173 * cos(2.04473251084  + 7.35736448430 *T)
    sumX0  +=      18.901 * cos(5.32807743689  + 26.05450231630 *T)
    sumX0  +=      21.874 * cos(3.64157746468  + 17297.93850984270 *T)
    sumX0  +=      19.888 * cos(2.97802303514  + 3930.45351370310 *T)
    sumX0  +=      15.237 * cos(4.38451579542  + 9830.63283147130 *T)
    sumX0  +=      17.432 * cos(0.19412655939  + 3531.81687532789 *T)
    sumX0  +=      14.522 * cos(5.10931707627  + 10596.42589591770 *T)
    sumX0  +=      16.153 * cos(2.06702370588  + 7084.65296363170 *T)
    sumX0  +=      16.091 * cos(2.35723470212  + 20213.51581446750 *T)
    sumX0  +=      14.842 * cos(4.06993785667  + 29864.57784479250 *T)
    sumX0  +=      11.112 * cos(3.77156689319  + 28287.23430234470 *T)
    sumX0  +=      12.828 * cos(2.04993681745  + 1589.31671276730 *T)
    sumX0  +=      11.212 * cos(2.47660347576  + 9154.14743350530 *T)
    sumX0  +=      14.376 * cos(3.86295510498  + 23581.50199480110 *T)
    sumX0  +=      12.175 * cos(4.86624913254  + 21228.63584102931 *T)
    sumX0  +=       9.393 * cos(0.62322150367  + 11272.91129388371 *T)
    sumX0  +=      12.824 * cos(1.41267695017  + 18837.25437965469 *T)
    sumX0  +=       9.207 * cos(0.63567894131  + 5661.57586663570 *T)
    sumX0  +=       8.622 * cos(4.96988848671  + 21536.19346199890 *T)
    sumX0  +=      11.653 * cos(6.20176223180  + 3340.85624418330 *T)
    sumX0  +=      11.496 * cos(5.06383150221  + 21228.14820606230 *T)
    sumX0  +=       9.511 * cos(0.26562072904  + 17298.42614480970 *T)
    sumX0  +=      11.281 * cos(4.58316811302  + 9786.44353785149 *T)
    sumX0  +=       8.498 * cos(3.85640563749  + 1577.58735993130 *T)
    sumX0  +=      10.763 * cos(1.23302415918  + 9929.67004482930 *T)
    sumX0  +=       8.847 * cos(1.67978485166  + 29051.02756083270 *T)
    sumX0  +=      10.378 * cos(1.97768679369  + 10497.38868255971 *T)
    sumX0  +=       7.498 * cos(5.81617198925  + 40853.38600232750 *T)
    sumX0  +=       7.375 * cos(0.80020428978  + 16496.60521368590 *T)
    sumX0  +=       8.792 * cos(0.46879154707  + 10208.00643938710 *T)
    sumX0  +=       8.786 * cos(2.74189610123  + 10219.05228800190 *T)
    sumX0  +=       7.423 * cos(4.16147581059  + 10175.50169105870 *T)
    sumX0  +=       7.386 * cos(5.33241732891  + 10251.55703633031 *T)
    sumX0  +=       6.692 * cos(4.38326911706  + 8094.27786834910 *T)
    sumX0  +=       6.531 * cos(1.17040611406  + 20419.70136290470 *T)
    sumX0  +=       5.876 * cos(2.02888635328  + 32217.44399856430 *T)
    sumX0  +=       7.945 * cos(3.46479056876  + 12566.39551746630 *T)
    sumX0  +=       5.723 * cos(2.30960586444  + 24357.02460612510 *T)
    sumX0  +=       5.960 * cos(0.33603622534  + 6681.46867088311 *T)
    sumX0  +=       5.974 * cos(2.16510174683  + 426.84200835950 *T)
    sumX0  +=       5.761 * cos(0.76064746748  + 2352.62233628830 *T)
    sumX0  +=       7.279 * cos(3.49868729640  + 11014.86265985130 *T)
    sumX0  +=       4.942 * cos(3.65960075191  + 3149.40797807170 *T)
    sumX0  +=       5.638 * cos(2.37982075471  + 382.65271473970 *T)
    sumX0  +=       6.069 * cos(2.45567436191  + 9683.35076363289 *T)
    sumX0  +=       4.581 * cos(2.63348121114  + 14765.48286075331 *T)
    sumX0  +=       6.372 * cos(0.20312515446  + 25158.35790228190 *T)
    sumX0  +=       4.902 * cos(1.17139302409  + 18837.74201462170 *T)
    sumX0  +=       4.698 * cos(0.27470947352  + 18849.47136745770 *T)
    sumX0  +=       5.371 * cos(4.44585316652  + 26088.14695905770 *T)
    sumX0  +=       5.495 * cos(3.29251136036  + 17086.20248320570 *T)
    sumX0  +=       5.381 * cos(5.41096356380  + 33794.78754101210 *T)
    sumX0  +=       4.870 * cos(2.78487952662  + 20453.11322970530 *T)
    sumX0  +=       3.922 * cos(2.87932422455  + 13745.59005650590 *T)
    sumX0  +=       4.608 * cos(4.34320749688  + 5661.08823166870 *T)
    sumX0  +=       3.970 * cos(3.08739367796  + 4731.78680985990 *T)
    sumX0  +=       3.556 * cos(0.34805582338  + 1551.28904013150 *T)
    sumX0  +=       4.329 * cos(3.40536513934  + 23958.87560271690 *T)
    sumX0  +=       3.274 * cos(3.43453378712  + 529.44714761109 *T)
    sumX0  +=       3.328 * cos(1.04776147690  + 18307.56341456009 *T)
    sumX0  +=       2.787 * cos(2.05633948048  + 11790.38527117530 *T)
    sumX0  +=       3.603 * cos(4.92117852022  + 14945.07235607090 *T)
    sumX0  +=       2.949 * cos(2.82363129219  + 10198.27689250950 *T)
    sumX0  +=       2.944 * cos(0.38707318449  + 10228.78183487951 *T)
    sumX0  +=       3.221 * cos(0.43638422425  + 20400.51659010570 *T)
    sumX0  +=       3.653 * cos(1.62294150568  + 29088.56759850150 *T)
    sumX0  +=       2.683 * cos(2.20899209567  + 6282.83203250789 *T)
    sumX0  +=       2.848 * cos(0.52334662875  + 11322.42028082089 *T)
    sumX0  +=       2.745 * cos(1.78121493204  + 5481.49873635110 *T)
    sumX0  +=       2.426 * cos(2.81466312460  + 6872.91693699471 *T)
    sumX0  +=       2.262 * cos(3.48344118262  + 7860.17557495569 *T)
    sumX0  +=       2.431 * cos(3.85595069475  + 30110.40949102190 *T)
    sumX0  +=       2.218 * cos(5.83392361752  + 17277.65074931730 *T)
    sumX0  +=       2.955 * cos(5.30083680831  + 27511.71169102070 *T)
    sumX0  +=       2.300 * cos(3.89687853440  + 3442.81876244891 *T)
    sumX0  +=       2.424 * cos(0.96902068311  + 20956.50587500010 *T)
    sumX0  +=       2.622 * cos(3.74514898210  + 9580.74562438130 *T)
    sumX0  +=       2.032 * cos(1.05319678961  + 20000.21671902950 *T)
    sumX0  +=       2.206 * cos(3.60536125643  + 537.04832957890 *T)
    sumX0  +=       2.010 * cos(0.56779046000  + 21202.33752122950 *T)
    sumX0  +=       2.471 * cos(5.74733670645  + 10846.31310300770 *T)
    sumX0  +=       2.315 * cos(5.33473916529  + 9573.14444241349 *T)
    sumX0  +=       2.188 * cos(5.98208918864  + 9967.69771746511 *T)
    sumX0  +=       2.278 * cos(4.39189942690  + 10110.43658947590 *T)
    sumX0  +=       2.278 * cos(5.10198662047  + 10316.62213791310 *T)
    sumX0  +=       1.808 * cos(4.00370769986  + 22779.68106367730 *T)
    sumX0  +=       2.244 * cos(2.44298085439  + 5507.79705615090 *T)
    sumX0  +=       1.796 * cos(3.09913214862  + 7063.87756813929 *T)
    sumX0  +=       2.118 * cos(3.51159951397  + 10459.36100992390 *T)
    sumX0  +=       1.908 * cos(5.61152019917  + 20809.71144212870 *T)
    sumX0  +=       2.093 * cos(0.69713061337  + 522.82123557730 *T)
    sumX0  +=       1.753 * cos(0.39797312506  + 13554.14179039430 *T)
    sumX0  +=       2.072 * cos(4.26049243578  + 9690.95194560070 *T)
    sumX0  +=       1.993 * cos(3.55209979851  + 10404.48999483909 *T)
    sumX0  +=       2.160 * cos(3.04497475534  + 33018.77729472110 *T)
    sumX0  +=       1.962 * cos(5.23488234630  + 10736.10678178830 *T)
    sumX0  +=       1.484 * cos(2.74179945598  + 3723.26514143949 *T)
    sumX0  +=       1.997 * cos(0.06145988645  + 18875.28205229050 *T)
    sumX0  +=       2.057 * cos(1.98619862343  + 102.84895673509 *T)
    sumX0  +=       2.016 * cos(0.55406656183  + 37724.99723723170 *T)
    sumX0  +=       1.618 * cos(4.87669871338  + 7085.14059859870 *T)
    sumX0  +=       1.659 * cos(1.57309690432  + 20235.36664379391 *T)
    sumX0  +=       1.553 * cos(1.42749385034  + 14945.55999103790 *T)
    sumX0  +=       1.873 * cos(1.50254912801  + 30213.50226524050 *T)
    sumX0  +=       1.542 * cos(4.02240332746  + 6490.02040477150 *T)
    sumX0  +=       1.360 * cos(5.79998899684  + 10233.19912459430 *T)
    sumX0  +=       1.360 * cos(3.69389705053  + 10193.85960279471 *T)
    sumX0  +=       1.592 * cos(0.72118861666  + 10138.74776512720 *T)
    sumX0  +=       1.592 * cos(2.48951212353  + 10288.31096226180 *T)
    sumX0  +=       1.368 * cos(2.85917999065  + 18875.76968725751 *T)
    sumX0  +=       1.235 * cos(3.79205714927  + 20.53157800890 *T)
    sumX0  +=       1.485 * cos(4.99009504257  + 9160.77334553909 *T)
    sumX0  +=       1.516 * cos(2.06896537087  + 9146.54625153749 *T)
    sumX0  +=       1.233 * cos(0.18739466393  + 18073.46112116669 *T)
    sumX0  +=       1.560 * cos(0.36404331783  + 9897.13749403790 *T)
    sumX0  +=       1.560 * cos(2.84665742236  + 10529.92123335110 *T)
    sumX0  +=       1.179 * cos(6.01432548499  + 14919.26167123810 *T)
    sumX0  +=       1.213 * cos(1.30457787805  + 28521.33659573810 *T)
    sumX0  +=       1.182 * cos(5.63437933479  + 9794.04471981930 *T)
    sumX0  +=       1.402 * cos(4.53367387536  + 8661.99650607950 *T)
    sumX0  +=       1.158 * cos(3.85902457723  + 10633.01400756970 *T)
    sumX0  +=       1.257 * cos(4.23637984727  + 2378.92065608810 *T)
    sumX0  +=       1.089 * cos(2.26544798668  + 30639.61282114949 *T)
    sumX0  +=       1.402 * cos(5.00395460415  + 10640.12755457050 *T)
    sumX0  +=       1.483 * cos(5.56946826365  + 21535.70582703189 *T)
    sumX0  +=       1.079 * cos(4.38415798738  + 775.27879384050 *T)
    sumX0  +=       1.370 * cos(4.61207624622  + 9786.93117281850 *T)
    sumX0  +=       1.064 * cos(5.60167497040  + 16984.23996494010 *T)
    sumX0  +=       1.158 * cos(4.33523311867  + 3929.96587873609 *T)
    sumX0  +=       0.967 * cos(1.59661459457  + 640.14110379750 *T)
    sumX0  +=       0.995 * cos(1.12539751877  + 8662.48414104651 *T)
    sumX0  +=       1.305 * cos(4.46692235351  + 36948.98699094070 *T)
    sumX0  +=       1.191 * cos(4.52742138803  + 39264.31310704370 *T)
    sumX0  +=       1.010 * cos(1.09565332862  + 10007.34381525730 *T)
    sumX0  +=       1.010 * cos(2.11504740292  + 10419.71491213170 *T)
    sumX0  +=       0.950 * cos(4.36909872600  + 1109.62236957690 *T)
    sumX0  +=       1.221 * cos(5.50189153252  + 75.02541605080 *T)
    sumX0  +=       0.987 * cos(5.46991186463  + 13937.03832261750 *T)
    sumX0  +=       1.164 * cos(1.47388964891  + 22805.49174851010 *T)
    sumX0  +=       0.845 * cos(4.27255310755  + 6770.46678376209 *T)
    sumX0  +=       0.996 * cos(4.53838005508  + 13341.91812879030 *T)
    sumX0  +=       1.123 * cos(1.83377276818  + 246.07546371290 *T)
    sumX0  +=       1.139 * cos(2.43198497783  + 2119.00767786191 *T)
    sumX0  +=       1.046 * cos(0.31602811545  + 31441.92138724031 *T)
    sumX0  +=       0.879 * cos(3.85735462170  + 15720.59496739490 *T)
    sumX0  +=       0.761 * cos(4.01867784780  + 7564.58690325449 *T)
    sumX0  +=       0.814 * cos(6.07006670549  + 9676.72485159910 *T)
    sumX0  +=       0.786 * cos(2.08315748724  + 11764.57458634250 *T)
    sumX0  +=       0.883 * cos(0.01404050923  + 25158.84553724890 *T)
    sumX0  +=       0.795 * cos(3.42432291896  + 10750.33387578991 *T)
    sumX0  +=       0.770 * cos(0.68478801852  + 7834.36489012290 *T)
    sumX0  +=       0.825 * cos(2.06778751923  + 191.69208359510 *T)
    sumX0  +=       0.696 * cos(3.39721023806  + 36147.65369478390 *T)
    sumX0  +=       0.941 * cos(3.99222118800  + 28520.84896077109 *T)
    sumX0  +=       0.787 * cos(5.89002484288  + 19890.01039781010 *T)
    sumX0  +=       0.788 * cos(0.18085455233  + 7994.77225950771 *T)
    sumX0  +=       0.624 * cos(2.84762425105  + 39794.00407213830 *T)
    sumX0  +=       0.609 * cos(2.15716574588  + 8624.45646841071 *T)
    sumX0  +=       0.836 * cos(1.97255748877  + 41655.20693345130 *T)
    sumX0  +=       0.744 * cos(2.51526655235  + 19904.23749181170 *T)
    sumX0  +=       0.689 * cos(4.60442929527  + 1066.73929467350 *T)
    sumX0  +=       0.803 * cos(5.88878266978  + 40879.19668716030 *T)
    sumX0  +=       0.692 * cos(0.54636447385  + 6770.95441872910 *T)
    sumX0  +=       0.585 * cos(5.02593807559  + 14143.25142494710 *T)
    sumX0  +=       0.678 * cos(5.96525738921  + 12592.20620229910 *T)
    sumX0  +=       0.573 * cos(2.52291476260  + 12592.69383726610 *T)
    sumX0  +=       0.670 * cos(0.92908106036  + 10239.34004852730 *T)
    sumX0  +=       0.655 * cos(4.98008645111  + 19366.94534474929 *T)
    sumX0  +=       0.626 * cos(1.17708461387  + 52176.05010063190 *T)
    sumX0  +=       0.548 * cos(0.67624288629  + 17777.87244946549 *T)
    sumX0  +=       0.709 * cos(2.90326183951  + 26735.70144472970 *T)
    sumX0  +=       0.531 * cos(3.09187361677  + 110.45013870291 *T)
    sumX0  +=       0.605 * cos(4.63445748441  + 10059.75055320970 *T)
    sumX0  +=       0.676 * cos(5.20222552429  + 20529.90768412410 *T)
    sumX0  +=       0.598 * cos(3.03085156721  + 12432.28646788130 *T)
    sumX0  +=       0.593 * cos(4.85950569906  + 10367.30817417930 *T)
    sumX0  +=       0.663 * cos(0.05389382712  + 1162.71852189130 *T)
    sumX0  +=       0.466 * cos(5.52080701144  + 8631.08238044450 *T)
    sumX0  +=       0.468 * cos(3.02762841033  + 16983.75232997309 *T)
    sumX0  +=       0.454 * cos(0.79241866785  + 10042.85649307530 *T)
    sumX0  +=       0.462 * cos(5.82955181248  + 6.86972951729 *T)
    sumX0  +=       0.444 * cos(2.41828213333  + 10384.20223431370 *T)
    sumX0  +=       0.460 * cos(0.69082886296  + 11802.60225897830 *T)
    sumX0  +=       0.436 * cos(1.85693465963  + 9639.18456535971 *T)
    sumX0  +=       0.565 * cos(5.70677906646  + 8520.87605922509 *T)
    sumX0  +=       0.416 * cos(1.72605691131  + 10225.25871653050 *T)
    sumX0  +=       0.416 * cos(1.48464382888  + 10201.80001085851 *T)
    sumX0  +=       0.414 * cos(3.17556892991  + 10186.74340892770 *T)
    sumX0  +=       0.431 * cos(1.73079752624  + 8616.85528644289 *T)
    sumX0  +=       0.393 * cos(2.42848511919  + 3532.30451029490 *T)
    sumX0  +=       0.406 * cos(1.35364773044  + 10787.87416202930 *T)
    sumX0  +=       0.380 * cos(0.22739133734  + 19317.43635781211 *T)
    sumX0  +=       0.458 * cos(1.71416676029  + 2218.51328670329 *T)
    sumX0  +=       0.433 * cos(5.30346380915  + 19650.80466361450 *T)
    sumX0  +=       0.377 * cos(5.67314793766  + 149.80701461810 *T)
    sumX0  +=       0.390 * cos(0.06912441711  + 13936.55068765050 *T)
    sumX0  +=       0.360 * cos(5.95509031178  + 24150.32386882850 *T)
    sumX0  +=       0.359 * cos(4.26088296989  + 30831.54872222810 *T)
    sumX0  +=       0.501 * cos(1.02737770483  + 44809.40638337990 *T)
    sumX0  +=       0.409 * cos(1.77188267484  + 27197.03787618409 *T)
    sumX0  +=       0.354 * cos(2.84306437211  + 10323.73568491390 *T)
    sumX0  +=       0.354 * cos(0.36763636808  + 10103.32304247510 *T)
    sumX0  +=       0.428 * cos(2.67107743725  + 13656.10430865991 *T)
    sumX0  +=       0.467 * cos(5.30307476925  + 38.37685312130 *T)
    sumX0  +=       0.343 * cos(1.61330776142  + 19786.91762359150 *T)
    sumX0  +=       0.457 * cos(0.79025821347  + 18308.05104952710 *T)
    sumX0  +=       0.399 * cos(0.98531427210  + 1052.51220067191 *T)
    sumX0  +=       0.445 * cos(4.32811724206  + 30665.91114094930 *T)
    sumX0  +=       0.344 * cos(0.03854207192  + 27490.93629552830 *T)
    sumX0  +=       0.315 * cos(0.01870495544  + 20640.11400534350 *T)
    sumX0  +=       0.399 * cos(4.24912842655  + 10212.04489098620 *T)
    sumX0  +=       0.399 * cos(5.24475762082  + 10215.01383640280 *T)
    sumX0  +=       0.390 * cos(2.28211006759  + 30000.20316980250 *T)
    sumX0  +=       0.340 * cos(1.17434519565  + 22805.97938347710 *T)
    sumX0  +=       0.398 * cos(3.64836715501  + 9836.15575577870 *T)
    sumX0  +=       0.311 * cos(4.76207516886  + 22003.67081738630 *T)
    sumX0  +=       0.404 * cos(3.98593080978  + 20352.03331133820 *T)
    sumX0  +=       0.366 * cos(1.12867710261  + 16522.41589851870 *T)
    sumX0  +=       0.350 * cos(2.99077020749  + 15874.37377787969 *T)
    sumX0  +=       0.385 * cos(1.36867906917  + 20180.98326367610 *T)
    sumX0  +=       0.379 * cos(5.84551902141  + 10590.90297161030 *T)
    sumX0  +=       0.395 * cos(0.41318733253  + 17277.16311435029 *T)
    sumX0  +=       0.383 * cos(1.78033504479  + 35372.13108345990 *T)
    sumX0  +=       0.277 * cos(3.16062357954  + 10988.56434005150 *T)
    sumX0  +=       0.365 * cos(1.42467845102  + 10349.15468870450 *T)
    sumX0  +=       0.365 * cos(1.78602228917  + 10077.90403868451 *T)
    sumX0  +=       0.336 * cos(0.74013259732  + 220.65645992230 *T)
    sumX0  +=       0.313 * cos(5.66612603850  + 6309.13035230770 *T)
    sumX0  +=       0.276 * cos(5.59087067643  + 26709.89075989690 *T)
    sumX0  +=       0.283 * cos(5.53432657905  + 3904.15519390330 *T)
    sumX0  +=       0.271 * cos(3.85648651665  + 9745.56437334010 *T)
    sumX0  +=       0.321 * cos(4.20436094748  + 38734.62214194910 *T)
    sumX0  +=       0.327 * cos(6.11821528186  + 9359.84534697549 *T)
    sumX0  +=       0.264 * cos(5.16095616548  + 76.50988875911 *T)
    sumX0  +=       0.310 * cos(3.61551590451  + 27991.15799567649 *T)
    sumX0  +=       0.359 * cos(3.38306212619  + 45585.41662967090 *T)
    sumX0  +=       0.326 * cos(5.59092073268  + 633.02755679670 *T)
    sumX0  +=       0.314 * cos(1.40080599663  + 20406.03951441310 *T)
    sumX0  +=       0.306 * cos(1.12609450630  + 2957.95971196010 *T)
    sumX0  +=       0.266 * cos(4.09083877114  + 8635.69818627969 *T)
    sumX0  +=       0.254 * cos(5.63745242067  + 10681.49435404891 *T)
    sumX0  +=       0.343 * cos(4.64295996667  + 8106.49485615210 *T)
    sumX0  +=       0.300 * cos(4.12771906867  + 206.42936592071 *T)
    sumX0  +=       0.250 * cos(3.04073965857  + 283.61550138170 *T)
    sumX0  +=       0.336 * cos(0.68541596989  + 23581.01435983410 *T)
    sumX0  +=       0.236 * cos(5.97871163742  + 10595.93826095070 *T)
    sumX0  +=       0.234 * cos(3.82015244691  + 19624.99397878170 *T)
    sumX0  +=       0.255 * cos(2.78388035745  + 213.05527795450 *T)
    sumX0  +=       0.281 * cos(0.78332195796  + 9399.97907973470 *T)
    sumX0  +=       0.314 * cos(0.56852266937  + 19786.42998862449 *T)
    sumX0  +=       0.229 * cos(4.91536328985  + 574.10098085129 *T)
    sumX0  +=       0.295 * cos(3.93680452053  + 31749.47900820990 *T)
    sumX0  +=       0.316 * cos(2.45044363511  + 48739.61607959950 *T)
    sumX0  +=       0.229 * cos(4.45667675327  + 949.41942645330 *T)
    sumX0  +=       0.229 * cos(6.25019131093  + 19896.63630984389 *T)
    sumX0  +=       0.257 * cos(4.89479846572  + 19360.31943271550 *T)
    sumX0  +=       0.231 * cos(4.38656769900  + 9264.35375472471 *T)
    sumX0  +=       0.255 * cos(4.03464641753  + 9993.11672125570 *T)
    sumX0  +=       0.255 * cos(5.45923962984  + 10433.94200613330 *T)
    sumX0  +=       0.291 * cos(0.75030918802  + 9999.74263328950 *T)
    sumX0  +=       0.285 * cos(5.75179047399  + 34596.12083716890 *T)
    sumX0  +=       0.211 * cos(0.18276464350  + 10249.55723037191 *T)
    sumX0  +=       0.211 * cos(3.02793609669  + 10177.50149701710 *T)
    sumX0  +=       0.255 * cos(2.42745847719  + 11027.07964765431 *T)
    sumX0  +=       0.220 * cos(0.56223178729  + 37.78385515230 *T)
    sumX0  +=       0.280 * cos(0.52583614761  + 37410.32342239509 *T)
    sumX0  +=       0.258 * cos(5.27008173054  + 29573.60497892650 *T)
    sumX0  +=       0.247 * cos(1.90810058449  + 29587.83207292810 *T)
    sumX0  +=       0.206 * cos(3.37472645930  + 30426.80136067850 *T)
    sumX0  +=       0.263 * cos(3.16130275876  + 20433.92845690630 *T)
    sumX0  +=       0.212 * cos(3.16045353079  + 10206.90080479450 *T)
    sumX0  +=       0.212 * cos(0.05024720940  + 10220.15792259451 *T)
    sumX0  +=       0.202 * cos(0.82632204894  + 16894.75421709410 *T)
    sumX0  +=       0.272 * cos(4.85092349120  + 12320.56387123691 *T)
    sumX0  +=       0.211 * cos(5.11529334983  + 11162.70497266430 *T)
    sumX0  +=       0.197 * cos(1.28459142755  + 8734.17515466309 *T)
    sumX0  +=       0.267 * cos(4.30013889854  + 377.61742539930 *T)
    sumX0  +=       0.224 * cos(3.35322427637  + 9051.05465928671 *T)
    sumX0  +=       0.186 * cos(4.53704574212  + 10251.66239933230 *T)
    sumX0  +=       0.186 * cos(4.95684030525  + 10175.39632805670 *T)
    sumX0  +=       0.258 * cos(2.52339474378  + 9864.60494324651 *T)
    sumX0  +=       0.186 * cos(4.77906943206  + 40077.86339100350 *T)
    sumX0  +=       0.233 * cos(1.49242697658  + 29089.05523346850 *T)
    sumX0  +=       0.186 * cos(4.86984152157  + 17272.12782500990 *T)
    sumX0  +=       0.254 * cos(3.15931132763  + 19264.34020549770 *T)
    sumX0  +=       0.246 * cos(0.68740453037  + 10562.45378414250 *T)
    sumX0  +=       0.211 * cos(6.16254771086  + 11376.00406810230 *T)
    sumX0  +=       0.178 * cos(0.50910605505  + 10949.40587722630 *T)
    sumX0  +=       0.178 * cos(2.70159468514  + 9477.65285016271 *T)
    sumX0  +=       0.197 * cos(0.95969377699  + 1692.40948698591 *T)
    sumX0  +=       0.177 * cos(0.12290786955  + 20316.60858868610 *T)
    sumX0  +=       0.216 * cos(2.56868513559  + 20452.62559473830 *T)
    sumX0  +=       0.197 * cos(0.19093764884  + 9050.56702431969 *T)
    sumX0  +=       0.159 * cos(1.16783306054  + 27197.52551115110 *T)
    sumX0  +=       0.167 * cos(5.81167464279  + 4705.48849006010 *T)
    sumX0  +=       0.213 * cos(2.13389038075  + 27511.22405605370 *T)
    sumX0  +=       0.154 * cos(0.72358681867  + 10199.30226969290 *T)
    sumX0  +=       0.154 * cos(2.48711392152  + 10227.75645769610 *T)
    sumX0  +=       0.174 * cos(1.89882712841  + 14.47091148511 *T)
    sumX0  +=       0.209 * cos(1.65966829596  + 10063.47853159290 *T)
    sumX0  +=       0.167 * cos(3.57558694059  + 6872.42930202769 *T)
    sumX0  +=       0.185 * cos(0.68349269134  + 9256.75257275689 *T)
    sumX0  +=       0.166 * cos(3.75162140471  + 10263.28638916630 *T)
    sumX0  +=       0.166 * cos(5.74226464266  + 10163.77233822270 *T)
    sumX0  +=       0.168 * cos(5.33662556430  + 7991.18509413049 *T)
    sumX0  +=       0.174 * cos(3.76838134579  + 9161.26098050610 *T)
    sumX0  +=       0.197 * cos(3.44221896080  + 846.32665223470 *T)
    sumX0  +=       0.176 * cos(4.69215492733  + 9779.32999085069 *T)
    sumX0  +=       0.198 * cos(3.88553571350  + 52669.82577581910 *T)
    sumX0  +=       0.169 * cos(5.71918201676  + 11265.79774688290 *T)
    sumX0  +=       0.188 * cos(0.90560199287  + 38526.33053338850 *T)
    sumX0  +=       0.140 * cos(0.96666880637  + 27991.64563064350 *T)
    sumX0  +=       0.194 * cos(2.26842216085  + 8837.26792888169 *T)
    sumX0  +=       0.173 * cos(1.30662690057  + 9793.55708485230 *T)
    sumX0  +=       0.140 * cos(0.33029139370  + 10191.67853436810 *T)
    sumX0  +=       0.140 * cos(2.88040934649  + 10235.38019302090 *T)
    sumX0  +=       0.196 * cos(2.27385726594  + 1293.24040609909 *T)
    sumX0  +=       0.180 * cos(4.38252963041  + 10021.59346261590 *T)
    sumX0  +=       0.138 * cos(0.38495137130  + 10217.46151695760 *T)
    sumX0  +=       0.138 * cos(2.82574936889  + 10209.59721043140 *T)
    sumX0  +=       0.148 * cos(2.31505492408  + 19374.54652671710 *T)
    sumX0  +=       0.137 * cos(3.95061609635  + 16522.90353348570 *T)
    sumX0  +=       0.143 * cos(1.40332876782  + 20447.59030539790 *T)
    sumX0  +=       0.139 * cos(5.72624016073  + 7255.32583425090 *T)
    sumX0  +=       0.126 * cos(3.82073811345  + 20277.25171277090 *T)
    sumX0  +=       0.164 * cos(4.79111675703  + 49515.62632589050 *T)
    sumX0  +=       0.149 * cos(2.18564631641  + 10135.85559326650 *T)
    sumX0  +=       0.149 * cos(1.02505442378  + 10291.20313412251 *T)
    sumX0  +=       0.118 * cos(1.12766998844  + 4551.70967957530 *T)
    sumX0  +=       0.117 * cos(0.93680170584  + 1582.20316576650 *T)
    sumX0  +=       0.151 * cos(0.84241735172  + 10246.06191448590 *T)
    sumX0  +=       0.151 * cos(2.36828338847  + 10180.99681290311 *T)
    sumX0  +=       0.152 * cos(3.18276216864  + 39302.34077967950 *T)
    sumX0  +=       0.144 * cos(3.82694942167  + 9986.05323090550 *T)
    sumX0  +=       0.144 * cos(5.66693662570  + 10441.00549648351 *T)
    sumX0  +=       0.136 * cos(2.08219853379  + 17469.09901542890 *T)
    sumX0  +=       0.119 * cos(3.32943340593  + 20809.22380716170 *T)
    sumX0  +=       0.130 * cos(5.57171941495  + 10236.30456514530 *T)
    sumX0  +=       0.130 * cos(3.92216663242  + 10190.75416224370 *T)
    sumX0  +=       0.156 * cos(4.19081128787  + 20388.68187426770 *T)
    sumX0  +=       0.132 * cos(2.35919234539  + 9580.25798941430 *T)
    sumX0  +=       0.155 * cos(0.85940280671  + 8238.03681783850 *T)
    sumX0  +=       0.142 * cos(3.58430646945  + 31441.43375227330 *T)
    sumX0  +=       0.131 * cos(5.39629404064  + 8101.39141534989 *T)
    sumX0  +=       0.133 * cos(4.00547577050  + 24382.83529095790 *T)
    sumX0  +=       0.105 * cos(4.23077158650  + 32858.36992533629 *T)
    sumX0  +=       0.105 * cos(3.82110715168  + 1059.13811270570 *T)
    sumX0  +=       0.101 * cos(4.35357648238  + 24978.76840696430 *T)
    sumX0  +=       0.126 * cos(4.90495452705  + 22645.57201409229 *T)
    sumX0  +=       0.101 * cos(5.26437713251  + 20043.91837768230 *T)
    sumX0  +=       0.110 * cos(2.97625385917  + 43071.65547154729 *T)
    sumX0  +=       0.103 * cos(0.46440823619  + 12431.79883291429 *T)
    sumX0  +=       0.136 * cos(6.07687689635  + 5999.46034860970 *T)
    sumX0  +=       0.097 * cos(0.92055841324  + 853.44019923550 *T)
    sumX0  +=       0.130 * cos(3.21780638112  + 4176.28515993250 *T)
    sumX0  +=       0.113 * cos(3.93650904588  + 19794.03117059230 *T)
    sumX0  +=       0.093 * cos(5.48927852030  + 22645.08437912529 *T)
    sumX0  +=       0.124 * cos(2.32988224429  + 42456.54022960810 *T)
    sumX0  +=       0.107 * cos(4.30116108617  + 3.68833577960 *T)
    sumX0  +=       0.127 * cos(3.54954178661  + 284.10313634870 *T)
    sumX0  +=       0.112 * cos(2.47056814408  + 20206.40226746670 *T)
    sumX0  +=       0.097 * cos(4.21617446768  + 78263.95324220609 *T)
    sumX0  +=       0.093 * cos(0.23349714961  + 8204.48418956850 *T)
    sumX0  +=       0.127 * cos(5.30922609526  + 56600.03547203870 *T)
    sumX0  +=       0.111 * cos(4.68533396525  + 18830.14083265389 *T)
    sumX0  +=       0.098 * cos(3.12531862236  + 11059.61219844570 *T)
    sumX0  +=       0.098 * cos(0.08567718624  + 9367.44652894330 *T)
    sumX0  +=       0.103 * cos(5.15240336002  + 10244.44348932950 *T)
    sumX0  +=       0.103 * cos(4.34148268734  + 10182.61523805950 *T)
    sumX0  +=       0.111 * cos(1.30087873837  + 18844.36792665549 *T)
    sumX0  +=       0.125 * cos(2.35130115448  + 12189.02190955050 *T)
    sumX0  +=       0.088 * cos(4.33292988106  + 20350.54883862990 *T)
    sumX0  +=       0.088 * cos(2.43603338900  + 29058.14110783350 *T)
    sumX0  +=       0.096 * cos(3.91338437973  + 2333.44021035551 *T)
    sumX0  +=       0.122 * cos(5.43206767869  + 2636.96929012050 *T)
    sumX0  +=       0.087 * cos(1.36103886488  + 8087.16432134830 *T)
    sumX0  +=       0.086 * cos(3.64796343622  + 7034.89593815990 *T)
    sumX0  +=       0.095 * cos(2.03086124785  + 10202.48366343060 *T)
    sumX0  +=       0.095 * cos(1.17983949234  + 10224.57506395840 *T)
    sumX0  +=       0.084 * cos(1.22799614537  + 10401.45451145710 *T)
    sumX0  +=       0.084 * cos(1.98270459482  + 10025.60421593190 *T)
    sumX0  +=       0.085 * cos(5.03727591626  + 19477.63930093570 *T)
    sumX0  +=       0.086 * cos(4.48909813316  + 29864.09020982550 *T)
    sumX0  +=       0.099 * cos(2.80223352440  + 2648.69864295650 *T)
    sumX0  +=       0.082 * cos(1.04973324929  + 802.06474860730 *T)
    sumX0  +=       0.087 * cos(0.60735009545  + 13745.10242153889 *T)
    sumX0  +=       0.081 * cos(0.59637598254  + 1550.80140516450 *T)
    sumX0  +=       0.083 * cos(3.04079971271  + 25933.88051360590 *T)
    sumX0  +=       0.079 * cos(6.14325401892  + 44008.07308722310 *T)
    sumX0  +=       0.078 * cos(6.11312046046  + 9573.63207738050 *T)
    sumX0  +=       0.098 * cos(5.37191506696  + 20220.62936146830 *T)
    sumX0  +=       0.108 * cos(3.66133732652  + 24149.83623386150 *T)
    sumX0  +=       0.100 * cos(5.07064547333  + 1596.43025976811 *T)
    sumX0  +=       0.095 * cos(3.35135945156  + 29477.62575170870 *T)
    sumX0  +=       0.078 * cos(3.03881688108  + 205.94173095370 *T)
    sumX0  +=       0.096 * cos(1.62154425800  + 30206.38871823969 *T)
    sumX0  +=       0.079 * cos(4.07943333749  + 3340.36860921629 *T)
    sumX0  +=       0.095 * cos(4.51771749335  + 30220.61581224130 *T)
    sumX0  +=       0.082 * cos(4.95983920368  + 29043.91401383189 *T)
    sumX0  +=       0.077 * cos(2.97497430839  + 17468.61138046190 *T)
    sumX0  +=       0.100 * cos(4.43773436133  + 38500.51984855570 *T)
    sumX0  +=       0.107 * cos(4.65407793574  + 2069.25065239010 *T)
    sumX0  +=       0.077 * cos(3.58586905737  + 10853.42665000850 *T)
    sumX0  +=       0.087 * cos(4.37137980512  + 5.27910682390 *T)
    sumX0  +=       0.076 * cos(4.87466102622  + 10243.35080184331 *T)
    sumX0  +=       0.076 * cos(4.61922502115  + 10183.70792554570 *T)
    sumX0  +=       0.076 * cos(6.08417794379  + 956.53297345411 *T)
    sumX0  +=       0.100 * cos(2.63929709585  + 28947.93478661410 *T)
    sumX0  +=       0.096 * cos(4.86017307249  + 35371.64344849290 *T)
    sumX0  +=       0.073 * cos(0.32640489167  + 17248.18148437089 *T)
    sumX0  +=       0.073 * cos(1.84890257629  + 21001.15970824030 *T)
    sumX0  +=       0.081 * cos(2.26124411922  + 18734.64924040310 *T)
    sumX0  +=       0.076 * cos(5.51994215698  + 103.33659170210 *T)
    sumX0  +=       0.085 * cos(5.27676110314  + 3914.71340755109 *T)
    sumX0  +=       0.074 * cos(1.92420691865  + 8827.63408735830 *T)
    sumX0  +=       0.070 * cos(1.49155151362  + 6298.57213865991 *T)
    sumX0  +=       0.087 * cos(2.91040927813  + 33019.26492968810 *T)
    sumX0  +=       0.071 * cos(3.95908921272  + 34172.16114892790 *T)
    sumX0  +=       0.079 * cos(2.65919074354  + 10018.55797923390 *T)
    sumX0  +=       0.079 * cos(0.55150999665  + 10408.50074815510 *T)
    sumX0  +=       0.069 * cos(6.05632855781  + 33794.29990604510 *T)
    sumX0  +=       0.087 * cos(1.54753819207  + 29050.53992586569 *T)
    sumX0  +=       0.069 * cos(0.90459565864  + 41044.83426843910 *T)
    sumX0  +=       0.069 * cos(2.25427054035  + 9644.70748966710 *T)
    sumX0  +=       0.069 * cos(0.95643019984  + 10782.35123772190 *T)
    sumX0  +=       0.087 * cos(5.46685094593  + 28313.04498717750 *T)
    sumX0  +=       0.085 * cos(3.77767527899  + 46386.74992582770 *T)
    sumX0  +=       0.083 * cos(5.79338403589  + 9721.86607123571 *T)
    sumX0  +=       0.083 * cos(3.70050201148  + 10705.19265615330 *T)
    sumX0  +=       0.090 * cos(5.18770728347  + 20049.44130198970 *T)
    sumX0  +=       0.064 * cos(5.86320987239  + 9147.03388650451 *T)
    sumX0  +=       0.076 * cos(1.05201656383  + 21486.19684009470 *T)
    sumX0  +=       0.085 * cos(1.34427023572  + 19573.13089318649 *T)
    sumX0  +=       0.063 * cos(1.61314796366  + 18314.67696156089 *T)
    sumX0  +=       0.071 * cos(6.14562384557  + 1573.57660661529 *T)
    sumX0  +=       0.070 * cos(5.46351961493  + 1580.62284331330 *T)
    sumX0  +=       0.080 * cos(6.26741884849  + 39301.85314471250 *T)
    sumX0  +=       0.067 * cos(0.48701132675  + 3956.26419853590 *T)
    sumX0  +=       0.087 * cos(5.47908021981  + 29264.32665627070 *T)
    sumX0  +=       0.063 * cos(3.37584641365  + 27461.46703058189 *T)
    sumX0  +=       0.070 * cos(5.75487124049  + 1479.11039154791 *T)
    sumX0  +=       0.081 * cos(0.05555806146  + 10205.92818172669 *T)
    sumX0  +=       0.080 * cos(0.23956461858  + 38204.44354188749 *T)
    sumX0  +=       0.060 * cos(1.34871003737  + 20421.29198559810 *T)
    sumX0  +=       0.064 * cos(4.30292059849  + 18300.44986755929 *T)
    sumX0  +=       0.076 * cos(3.93986568359  + 29683.81130014589 *T)
    sumX0  +=       0.082 * cos(0.44981983060  + 60530.24516825830 *T)
    sumX0  +=       0.077 * cos(6.19200892946  + 53445.83602211010 *T)
    sumX0  +=       0.062 * cos(1.28649446141  + 11599.42464003070 *T)
    sumX0  +=       0.058 * cos(3.08211963404  + 7576.80389105750 *T)
    sumX0  +=       0.057 * cos(3.67915137516  + 11280.02484088450 *T)
    sumX0  +=       0.059 * cos(1.26040809465  + 10255.08015467930 *T)
    sumX0  +=       0.059 * cos(1.95029264554  + 10171.97857270970 *T)
    sumX0  +=       0.057 * cos(4.06263183264  + 20464.84258254131 *T)
    sumX0  +=       0.079 * cos(1.97866126955  + 10136.77565745190 *T)
    sumX0  +=       0.066 * cos(2.33925448441  + 14169.54974474690 *T)
    sumX0  +=       0.063 * cos(3.68749019288  + 18099.75944096650 *T)
    sumX0  +=       0.060 * cos(3.19857198519  + 29786.90407436450 *T)
    sumX0  +=       0.064 * cos(5.57155183718  + 30007.31671680330 *T)
    sumX0  +=       0.054 * cos(3.67263578488  + 30666.39877591630 *T)
    sumX0  +=       0.058 * cos(3.05733553465  + 75449.75065697990 *T)
    sumX0  +=       0.073 * cos(3.85067322628  + 51449.56808076170 *T)
    sumX0  +=       0.074 * cos(5.48458129044  + 21720.29913348810 *T)
    sumX0  +=       0.063 * cos(2.67652037589  + 26736.18907969670 *T)
    sumX0  +=       0.054 * cos(3.91509939932  + 38204.93117685450 *T)
    sumX0  +=       0.072 * cos(4.47140350205  + 100608.35237674530 *T)
    sumX0  +=       0.053 * cos(3.80804701502  + 419.72846135871 *T)
    sumX0  +=       0.052 * cos(2.09262980178  + 40426.78781145150 *T)
    sumX0  +=       0.052 * cos(5.73445119519  + 29580.23089096029 *T)
    sumX0  +=       0.053 * cos(1.24051301926  + 419.24082639170 *T)
    sumX0  +=       0.058 * cos(1.31187367842  + 20412.58781590390 *T)
    sumX0  +=       0.054 * cos(3.26242420831  + 9366.95889397630 *T)
    sumX0  +=       0.055 * cos(3.59360279993  + 9103.66317663410 *T)
    sumX0  +=       0.050 * cos(2.37895606630  + 34570.31015233610 *T)
    sumX0  +=       0.067 * cos(2.64494428691  + 632.53992182970 *T)
    sumX0  +=       0.050 * cos(1.18004520961  + 37724.50960226470 *T)
    sumX0  +=       0.069 * cos(1.73296667148  + 53284.94101775829 *T)
    sumX0  +=       0.057 * cos(1.89823008385  + 8307.57696378710 *T)
    sumX0  +=       0.050 * cos(4.69860484886  + 8417.78328500650 *T)
    sumX0  +=       0.049 * cos(1.91804881602  + 26087.65932409069 *T)
    sumX0  +=       0.064 * cos(2.02816930105  + 7064.36520310630 *T)
    sumX0  +=       0.049 * cos(1.71661686491  + 14128.48658872910 *T)
    sumX0  +=       0.056 * cos(2.23445542422  + 28418.24382151950 *T)
    sumX0  +=       0.066 * cos(6.05166710265  + 19580.73207515430 *T)
    sumX0  +=       0.047 * cos(2.93855195986  + 10742.73269382209 *T)
    sumX0  +=       0.061 * cos(1.41947943365  + 43232.06284093210 *T)
    sumX0  +=       0.061 * cos(4.56107208724  + 43232.55047589910 *T)
    sumX0  +=       0.047 * cos(4.81852576439  + 38741.73568894990 *T)
    sumX0  +=       0.050 * cos(1.33989551475  + 59161.19323437110 *T)
    sumX0  +=       0.046 * cos(2.23137394571  + 18844.85556162250 *T)
    sumX0  +=       0.054 * cos(0.69428675399  + 48947.90768816010 *T)
    sumX0  +=       0.046 * cos(4.63842079862  + 3154.93090237910 *T)
    sumX0  +=       0.046 * cos(4.21170546636  + 194.72756697710 *T)
    sumX0  +=       0.056 * cos(1.07866938616  + 9936.78359183010 *T)
    sumX0  +=       0.056 * cos(2.13203135403  + 10490.27513555891 *T)
    sumX0  +=       0.054 * cos(4.49664827025  + 9922.55649782850 *T)
    sumX0  +=       0.054 * cos(4.99723777712  + 10504.50222956051 *T)
    sumX0  +=       0.059 * cos(2.01423634465  + 18204.47064034149 *T)
    sumX0  +=       0.048 * cos(1.22734670146  + 38727.50859494830 *T)
    sumX0  +=       0.057 * cos(5.20275648643  + 50316.95962204730 *T)
    sumX0  +=       0.061 * cos(5.98664723657  + 10140.23223783550 *T)
    sumX0  +=       0.061 * cos(3.50723881080  + 10286.82648955350 *T)
    sumX0  +=       0.061 * cos(0.57175194941  + 6567.17898634010 *T)
    sumX0  +=       0.048 * cos(3.35196537073  + 9246.55848625890 *T)
    sumX0  +=       0.060 * cos(2.77045228898  + 3684.13423250670 *T)
    sumX0  +=       0.060 * cos(3.41701763721  + 14427.59837877930 *T)
    sumX0  +=       0.051 * cos(2.29628619637  + 47623.60896860609 *T)
    sumX0  +=       0.051 * cos(5.43142352332  + 9566.03089541269 *T)
    sumX0  +=       0.052 * cos(0.06151623743  + 8727.06160766230 *T)
    sumX0  +=       0.057 * cos(0.62091438086  + 32243.25468339710 *T)
    sumX0  +=       0.048 * cos(2.05408316319  + 433.95555536030 *T)
    sumX0  +=       0.045 * cos(5.52086086974  + 10263.52598559870 *T)
    sumX0  +=       0.045 * cos(3.97302517763  + 10163.53274179031 *T)
    sumX0  +=       0.049 * cos(1.65627933457  + 9794.60496479390 *T)
    sumX0  +=       0.049 * cos(1.55442140562  + 10632.45376259511 *T)
    sumX0  +=       0.058 * cos(6.19349919374  + 1860.95904382949 *T)
    sumX0  +=       0.047 * cos(1.50917280490  + 20142.95559104030 *T)
    sumX0  +=       0.046 * cos(5.38397167897  + 2349.58685290630 *T)
    sumX0  +=       0.044 * cos(2.04804189558  + 30632.98690911569 *T)
    sumX0  +=       0.046 * cos(4.70186674108  + 2356.63308960431 *T)
    sumX0  +=       0.041 * cos(0.12858083101  + 12850.25483633150 *T)
    sumX0  +=       0.041 * cos(1.71656432171  + 8094.76550331610 *T)
    sumX0  +=       0.042 * cos(5.08304675769  + 77802.61681075170 *T)
    sumX0  +=       0.042 * cos(2.81186817738  + 25132.54721744910 *T)
    sumX0  +=       0.043 * cos(2.37147053321  + 19359.83179774849 *T)
    sumX0  +=       0.040 * cos(3.89496191683  + 10192.26633323510 *T)
    sumX0  +=       0.042 * cos(3.28822065771  + 34363.60941503950 *T)
    sumX0  +=       0.039 * cos(4.35089840905  + 109.96250373590 *T)
    sumX0  +=       0.044 * cos(4.46863580810  + 79379.96035319949 *T)
    sumX0  +=       0.043 * cos(4.38742720144  + 38734.13450698209 *T)
    sumX0  +=       0.046 * cos(0.42463075467  + 17778.36008443250 *T)
    sumX0  +=       0.039 * cos(6.01839872141  + 12565.90788249930 *T)
    sumX0  +=       0.041 * cos(0.30408875222  + 27681.89692667289 *T)
    sumX0  +=       0.053 * cos(1.87369351531  + 64460.45486447790 *T)
    sumX0  +=       0.039 * cos(3.44426638934  + 28631.54291695750 *T)
    sumX0  +=       0.052 * cos(5.20963176413  + 38631.52936773050 *T)
    sumX0  +=       0.043 * cos(6.14192058418  + 11180.50024113010 *T)
    sumX0  +=       0.039 * cos(1.02239033953  + 1265.81129610991 *T)
    sumX0  +=       0.039 * cos(3.80918822366  + 7879.84533585549 *T)
    sumX0  +=       0.039 * cos(0.66759557007  + 7880.33297082250 *T)
    sumX0  +=       0.052 * cos(1.99450435156  + 1.72829019180 *T)
    sumX0  +=       0.039 * cos(0.47445196924  + 54981.62877357310 *T)
    sumX0  +=       0.047 * cos(4.89376316808  + 187.68133027909 *T)
    sumX0  +=       0.044 * cos(1.17091466926  + 10168.38814405790 *T)
    sumX0  +=       0.044 * cos(2.03978607093  + 10258.67058333111 *T)
    sumX0  +=       0.048 * cos(4.68507697395  + 10063.96616655990 *T)
    sumX0  +=       0.043 * cos(5.12159413101  + 22029.96913718610 *T)
    sumX0  +=       0.051 * cos(5.89313258670  + 104538.56207296488 *T)
    sumX0  +=       0.051 * cos(6.27593957045  + 16250.77356745650 *T)
    sumX0  +=       0.037 * cos(4.48693750961  + 29161.23388205210 *T)
    sumX0  +=       0.043 * cos(4.92338640427  + 10363.09256082911 *T)
    sumX0  +=       0.038 * cos(4.13589046896  + 47917.50738795030 *T)
    sumX0  +=       0.049 * cos(4.56271718314  + 66037.79840692569 *T)
    sumX0  +=       0.037 * cos(4.41298098667  + 6257.02134767510 *T)
    sumX0  +=       0.036 * cos(2.97028336578  + 8674.21349388250 *T)
    sumX0  +=       0.045 * cos(2.00823427492  + 522.33360061030 *T)
    sumX0  +=       0.049 * cos(1.96278700412  + 491.90710994230 *T)
    sumX0  +=       0.042 * cos(2.95264046208  + 115674.52048208468 *T)
    sumX0  +=       0.040 * cos(6.06388400822  + 10233.81712421990 *T)
    sumX0  +=       0.036 * cos(1.27578830807  + 47938.28278344270 *T)
    sumX0  +=       0.035 * cos(2.75316747179  + 12546.72575656651 *T)
    sumX0  +=       0.048 * cos(2.91217379208  + 90395.06683053430 *T)
    sumX0  +=       0.040 * cos(2.61778472956  + 14127.99895376209 *T)
    sumX0  +=       0.044 * cos(4.30648713686  + 9698.06549260150 *T)
    sumX0  +=       0.044 * cos(5.18739891051  + 10728.99323478751 *T)
    sumX0  +=       0.036 * cos(2.23668112498  + 69942.19741831250 *T)
    sumX0  +=       0.035 * cos(2.59567778333  + 41654.71929848429 *T)
    sumX0  +=       0.034 * cos(4.26526324958  + 55798.70217588190 *T)
    sumX0  +=       0.037 * cos(3.25219443320  + 35192.05395317530 *T)
    sumX0  +=       0.046 * cos(2.90936826705  + 36301.43250526870 *T)
    sumX0  +=       0.047 * cos(1.48071490743  + 86464.85713431469 *T)
    sumX0  +=       0.035 * cos(5.44812994648  + 48732.98752069949 *T)
    sumX0  +=       0.035 * cos(4.79214932619  + 10011.27596852040 *T)
    sumX0  +=       0.035 * cos(4.70173672118  + 10415.78275886861 *T)
    sumX0  +=       0.034 * cos(6.17898420868  + 37410.81105736210 *T)
    sumX0  +=       0.033 * cos(5.38008686554  + 49477.59865325469 *T)
    sumX0  +=       0.042 * cos(4.39944587348  + 18830.62846762090 *T)
    sumX0  +=       0.035 * cos(5.27547059803  + 49515.13869092350 *T)
    sumX0  +=       0.047 * cos(0.65352099169  + 31022.50935337269 *T)
    sumX0  +=       0.034 * cos(4.90000292698  + 100909.28143881488 *T)
    sumX0  +=       0.033 * cos(3.74889730790  + 10117.55013647670 *T)
    sumX0  +=       0.033 * cos(5.74498873947  + 10309.50859091230 *T)
    sumX0  +=       0.034 * cos(4.93693556191  + 7461.49412903589 *T)
    sumX0  +=       0.034 * cos(3.87169740614  + 45584.92899470390 *T)
    sumX0  +=       0.032 * cos(0.10155743123  + 9676.23721663209 *T)
    sumX0  +=       0.039 * cos(2.72509244693  + 316.63568714010 *T)
    sumX0  +=       0.033 * cos(4.45732662556  + 31748.99137324289 *T)
    sumX0  +=       0.031 * cos(3.44787752858  + 30103.29594402109 *T)
    sumX0  +=       0.035 * cos(4.20052948408  + 58946.27306691049 *T)
    sumX0  +=       0.032 * cos(3.76438941046  + 21977.86013255350 *T)
    sumX0  +=       0.036 * cos(1.22867623820  + 20430.74706316860 *T)
    sumX0  +=       0.032 * cos(0.22256920805  + 81732.82650697129 *T)
    sumX0  +=       0.030 * cos(2.28012493395  + 19573.61852815350 *T)
    sumX0  +=       0.031 * cos(2.48999281964  + 11.28951774740 *T)
    sumX0  +=       0.038 * cos(4.94526723480  + 20617.77554105009 *T)
    sumX0  +=       0.033 * cos(5.88114654745  + 83310.17004941910 *T)
    sumX0  +=       0.029 * cos(0.24041771274  + 11752.84523350650 *T)
    sumX0  +=       0.029 * cos(5.51688017902  + 9801.15826682010 *T)
    sumX0  +=       0.029 * cos(3.97700586835  + 10625.90046056890 *T)
    sumX0  +=       0.030 * cos(3.62180540908  + 20432.33783421290 *T)
    sumX0  +=       0.031 * cos(4.94266626401  + 7255.81346921790 *T)
    sumX0  +=       0.029 * cos(4.41071491793  + 9889.53631207009 *T)
    sumX0  +=       0.029 * cos(1.77595651133  + 224.58861318540 *T)
    sumX0  +=       0.028 * cos(0.27335698936  + 10099.09007500929 *T)
    sumX0  +=       0.028 * cos(2.93734375083  + 10327.96865237972 *T)
    sumX0  +=       0.037 * cos(0.01012570970  + 47964.09346827550 *T)
    sumX0  +=       0.029 * cos(1.93091525734  + 10426.34082416550 *T)
    sumX0  +=       0.034 * cos(4.22987244572  + 1539.55968729550 *T)
    sumX0  +=       0.039 * cos(5.18936044570  + 30563.83438484090 *T)
    sumX0  +=       0.039 * cos(1.33268449559  + 5791.16874004909 *T)
    sumX0  +=       0.028 * cos(5.43649843683  + 1178.95072155610 *T)
    sumX0  +=       0.031 * cos(5.46650886390  + 39786.89052513750 *T)
    sumX0  +=       0.028 * cos(3.65949505965  + 73872.40711453209 *T)
    sumX0  +=       0.027 * cos(0.53018129778  + 9411.22079760369 *T)
    sumX0  +=       0.028 * cos(0.41914594254  + 15694.78428256210 *T)
    sumX0  +=       0.034 * cos(3.67856303130  + 40213.48871601350 *T)
    sumX0  +=       0.037 * cos(4.44226497286  + 36949.47462590770 *T)
    sumX0  +=       0.035 * cos(3.56301256331  + 37704.22184173930 *T)
    sumX0  +=       0.028 * cos(5.98328888184  + 4732.27444482690 *T)
    sumX0  +=       0.037 * cos(2.63889784222  + 13859.87974104890 *T)
    sumX0  +=       0.035 * cos(2.72288729195  + 18093.13088206649 *T)
    sumX0  +=       0.035 * cos(5.86447994554  + 18093.61851703350 *T)
    sumX0  +=       0.031 * cos(2.99968207636  + 42430.72954477530 *T)
    sumX0  +=       0.026 * cos(4.09923425337  + 8734.66278963010 *T)
    sumX0  +=       0.028 * cos(1.92482741603  + 7886.47389475550 *T)
    sumX0  +=       0.031 * cos(0.27204438582  + 25960.17883340570 *T)

    # VENUS X Coord
    sumX1  = 0
    sumX1  +=   51880.979 * cos(2.28175279050  + 0.24381748350 *T)
    sumX1  +=   17260.820 * cos(0.92558492780  + 20426.81490990550 *T)
    sumX1  +=   11049.864 * cos(2.03460665124  + 10213.04172872750 *T)
    sumX1  +=    4831.725 * cos(4.77621954273  + 10213.52936369450 *T)
    sumX1  +=     175.352 * cos(1.80600884943  + 30640.10045611650 *T)
    sumX1  +=      60.293 * cos(4.92800632264  + 8636.18582124671 *T)
    sumX1  +=      62.982 * cos(2.09113961987  + 10187.23104389470 *T)
    sumX1  +=      62.707 * cos(1.11859002578  + 10239.82768349430 *T)
    sumX1  +=      49.342 * cos(4.57448643960  + 11790.87290614230 *T)
    sumX1  +=      36.826 * cos(2.07066164376  + 775.76642880750 *T)
    sumX1  +=      39.689 * cos(3.30583760303  + 20426.32727493849 *T)
    sumX1  +=      30.101 * cos(5.32708077931  + 10022.08109758290 *T)
    sumX1  +=      29.330 * cos(4.16541909519  + 10404.97762980610 *T)
    sumX1  +=      19.990 * cos(1.15893210018  + 9438.00675237050 *T)
    sumX1  +=      17.877 * cos(2.07736309994  + 10989.05197501850 *T)
    sumX1  +=      14.175 * cos(3.70557013555  + 9683.83839859990 *T)
    sumX1  +=      13.388 * cos(1.33844459770  + 191.20444862810 *T)
    sumX1  +=      13.513 * cos(5.77342390577  + 10743.22032878910 *T)
    sumX1  +=      14.034 * cos(3.49932313102  + 4705.97612502710 *T)
    sumX1  +=      13.385 * cos(1.15013591011  + 19651.29229858150 *T)
    sumX1  +=       6.873 * cos(4.83631203632  + 9104.15081160110 *T)
    sumX1  +=       7.547 * cos(5.57551333367  + 1059.62574767270 *T)
    sumX1  +=       7.388 * cos(6.00037338177  + 15721.08260236190 *T)
    sumX1  +=       5.971 * cos(4.65746818072  + 11322.90791578790 *T)
    sumX1  +=       5.966 * cos(1.10731222776  + 10220.64291069530 *T)
    sumX1  +=       5.965 * cos(2.10389967659  + 10206.41581669370 *T)
    sumX1  +=       5.155 * cos(2.49611854501  + 3154.44326741210 *T)
    sumX1  +=       4.571 * cos(4.59065584437  + 20618.26317601710 *T)
    sumX1  +=       4.246 * cos(6.05417599890  + 10000.23026825651 *T)
    sumX1  +=       4.157 * cos(3.43981648730  + 10426.82845913250 *T)
    sumX1  +=       4.649 * cos(2.86570769042  + 9830.63283147130 *T)
    sumX1  +=       4.431 * cos(0.34499590120  + 10596.42589591770 *T)
    sumX1  +=       3.938 * cos(0.35777611387  + 7.35736448430 *T)
    sumX1  +=       3.399 * cos(2.39541457575  + 7860.66320992270 *T)
    sumX1  +=       2.758 * cos(3.91101238225  + 19367.43297971630 *T)
    sumX1  +=       2.699 * cos(0.65662263221  + 7058.84227879890 *T)
    sumX1  +=       2.515 * cos(5.65192320935  + 8623.96883344369 *T)
    sumX1  +=       2.725 * cos(4.96638871958  + 801.57711364030 *T)
    sumX1  +=       2.539 * cos(5.96953617871  + 9411.70843257070 *T)
    sumX1  +=       1.822 * cos(3.32629702713  + 1109.13473460990 *T)
    sumX1  +=       2.300 * cos(3.51596869326  + 11015.35029481830 *T)
    sumX1  +=       1.816 * cos(2.53054405922  + 13368.21644859010 *T)
    sumX1  +=       1.814 * cos(4.85812202292  + 3930.45351370310 *T)
    sumX1  +=       1.586 * cos(2.69305924606  + 40853.38600232750 *T)
    sumX1  +=       1.999 * cos(2.52624662737  + 26.05450231630 *T)
    sumX1  +=       1.745 * cos(2.26492854011  + 4552.19731454230 *T)
    sumX1  +=       1.503 * cos(0.63999417521  + 12566.39551746630 *T)
    sumX1  +=       1.446 * cos(3.91665476010  + 7084.65296363170 *T)
    sumX1  +=       1.400 * cos(6.24730767768  + 11272.91129388371 *T)
    sumX1  +=       1.282 * cos(4.66310117053  + 10219.05228800190 *T)
    sumX1  +=       1.282 * cos(4.83078487684  + 10208.00643938710 *T)
    sumX1  +=       1.352 * cos(2.86394052328  + 20419.70136290470 *T)
    sumX1  +=       1.475 * cos(5.68356978914  + 3128.63258257930 *T)
    sumX1  +=       1.235 * cos(2.12083036469  + 1577.58735993130 *T)
    sumX1  +=       1.286 * cos(5.69421578243  + 23581.50199480110 *T)
    sumX1  +=       1.283 * cos(3.88408958331  + 382.65271473970 *T)
    sumX1  +=       1.171 * cos(3.28848400510  + 9154.14743350530 *T)
    sumX1  +=       1.084 * cos(4.35524384671  + 10198.27689250950 *T)
    sumX1  +=       1.084 * cos(5.13864220066  + 10228.78183487951 *T)
    sumX1  +=       1.198 * cos(4.95985546788  + 16496.60521368590 *T)
    sumX1  +=       0.869 * cos(2.39926268528  + 2352.62233628830 *T)
    sumX1  +=       1.112 * cos(0.28321011044  + 1589.31671276730 *T)
    sumX1  +=       0.922 * cos(0.95607730643  + 15874.86141284670 *T)
    sumX1  +=       0.743 * cos(0.68161454240  + 28287.23430234470 *T)
    sumX1  +=       0.746 * cos(2.15375090404  + 3149.40797807170 *T)
    sumX1  +=       0.797 * cos(5.11193244020  + 426.84200835950 *T)
    sumX1  +=       0.625 * cos(3.18531726606  + 20400.51659010570 *T)
    sumX1  +=       0.615 * cos(0.22278918989  + 21536.19346199890 *T)
    sumX1  +=       0.832 * cos(1.92436994458  + 21228.63584102931 *T)
    sumX1  +=       0.656 * cos(5.34274867064  + 11014.86265985130 *T)
    sumX1  +=       0.586 * cos(2.16720042671  + 5661.57586663570 *T)
    sumX1  +=       0.648 * cos(1.91289793451  + 537.04832957890 *T)
    sumX1  +=       0.674 * cos(5.27695637784  + 8094.27786834910 *T)
    sumX1  +=       0.636 * cos(0.25111675779  + 18849.47136745770 *T)
    sumX1  +=       0.655 * cos(3.80934378726  + 17298.42614480970 *T)
    sumX1  +=       0.615 * cos(2.39412529972  + 522.82123557730 *T)
    sumX1  +=       0.507 * cos(0.55272313227  + 20453.11322970530 *T)
    sumX1  +=       0.451 * cos(5.12854697169  + 32217.44399856430 *T)
    sumX1  +=       0.480 * cos(2.46182518074  + 529.44714761109 *T)
    sumX1  +=       0.439 * cos(3.76528392741  + 6282.83203250789 *T)
    sumX1  +=       0.439 * cos(2.58129029648  + 29051.02756083270 *T)
    sumX1  +=       0.410 * cos(5.46936106504  + 24357.02460612510 *T)
    sumX1  +=       0.452 * cos(3.76215759010  + 9146.54625153749 *T)
    sumX1  +=       0.417 * cos(4.91991688397  + 6681.46867088311 *T)
    sumX1  +=       0.434 * cos(3.29962044494  + 9160.77334553909 *T)
    sumX1  +=       0.364 * cos(2.71580018519  + 10640.12755457050 *T)
    sumX1  +=       0.363 * cos(0.47492285209  + 9786.93117281850 *T)
    sumX1  +=       0.428 * cos(0.81939944524  + 20809.71144212870 *T)
    sumX1  +=       0.440 * cos(2.47378943911  + 9690.95194560070 *T)
    sumX1  +=       0.440 * cos(0.73691128332  + 10736.10678178830 *T)
    sumX1  +=       0.372 * cos(0.50265145827  + 10251.55703633031 *T)
    sumX1  +=       0.372 * cos(2.70804928192  + 10175.50169105870 *T)
    sumX1  +=       0.319 * cos(1.05799277323  + 17277.65074931730 *T)
    sumX1  +=       0.427 * cos(2.95582453815  + 18837.74201462170 *T)
    sumX1  +=       0.319 * cos(0.18812390399  + 7860.17557495569 *T)
    sumX1  +=       0.303 * cos(1.07261427828  + 14765.48286075331 *T)
    sumX1  +=       0.351 * cos(2.50811815863  + 6490.02040477150 *T)
    sumX1  +=       0.287 * cos(0.13453811680  + 1577.09972496430 *T)
    sumX1  +=       0.318 * cos(0.48117169551  + 14945.07235607090 *T)
    sumX1  +=       0.297 * cos(1.06831804424  + 10233.19912459430 *T)
    sumX1  +=       0.297 * cos(2.14238269595  + 10193.85960279471 *T)
    sumX1  +=       0.372 * cos(5.93165702933  + 22004.15845235330 *T)
    sumX1  +=       0.311 * cos(1.29581017314  + 9683.35076363289 *T)
    sumX1  +=       0.284 * cos(2.36025817240  + 3442.81876244891 *T)
    sumX1  +=       0.269 * cos(1.16222638968  + 9573.14444241349 *T)
    sumX1  +=       0.262 * cos(3.22424327301  + 21202.33752122950 *T)
    sumX1  +=       0.285 * cos(4.56673097639  + 13745.59005650590 *T)
    sumX1  +=       0.276 * cos(1.42552468909  + 9676.72485159910 *T)
    sumX1  +=       0.289 * cos(5.78612487399  + 20235.36664379391 *T)
    sumX1  +=       0.263 * cos(5.63102915308  + 22779.68106367730 *T)
    sumX1  +=       0.280 * cos(5.09544909614  + 14919.26167123810 *T)
    sumX1  +=       0.262 * cos(1.80492967584  + 10750.33387578991 *T)
    sumX1  +=       0.256 * cos(4.41502652709  + 10234.30475918690 *T)
    sumX1  +=       0.239 * cos(4.74729006212  + 5507.79705615090 *T)
    sumX1  +=       0.253 * cos(5.07523059777  + 10192.75396820210 *T)
    sumX1  +=       0.268 * cos(0.81556729812  + 27511.71169102070 *T)
    sumX1  +=       0.282 * cos(6.23856358633  + 4731.78680985990 *T)
    sumX1  +=       0.232 * cos(4.24398296621  + 3723.26514143949 *T)
    sumX1  +=       0.285 * cos(5.52678704563  + 20956.50587500010 *T)
    sumX1  +=       0.244 * cos(3.89280464375  + 2353.10997125530 *T)
    sumX1  +=       0.272 * cos(4.39654535286  + 20000.21671902950 *T)
    sumX1  +=       0.254 * cos(4.68153309834  + 6283.31966747490 *T)
    sumX1  +=       0.238 * cos(4.49319488898  + 30639.61282114949 *T)
    sumX1  +=       0.232 * cos(1.29928686548  + 19890.01039781010 *T)
    sumX1  +=       0.210 * cos(1.76685272207  + 18073.46112116669 *T)
    sumX1  +=       0.216 * cos(4.76667969992  + 18073.94875613370 *T)
    sumX1  +=       0.199 * cos(0.33116574732  + 9639.18456535971 *T)
    sumX1  +=       0.218 * cos(0.81835363281  + 19904.23749181170 *T)
    sumX1  +=       0.225 * cos(0.70271153327  + 13937.03832261750 *T)
    sumX1  +=       0.235 * cos(3.30083410129  + 1551.28904013150 *T)
    sumX1  +=       0.191 * cos(4.33367506839  + 6872.91693699471 *T)
    sumX1  +=       0.179 * cos(0.13296789433  + 30110.40949102190 *T)
    sumX1  +=       0.185 * cos(2.87950187226  + 10787.87416202930 *T)
    sumX1  +=       0.235 * cos(2.81621171423  + 1066.73929467350 *T)
    sumX1  +=       0.208 * cos(5.61842788521  + 10007.34381525730 *T)
    sumX1  +=       0.208 * cos(3.87545816216  + 10419.71491213170 *T)
    sumX1  +=       0.179 * cos(1.93004567854  + 18875.28205229050 *T)
    sumX1  +=       0.184 * cos(1.66108282172  + 9967.69771746511 *T)
    sumX1  +=       0.166 * cos(1.92550152422  + 18307.56341456009 *T)
    sumX1  +=       0.163 * cos(4.95974146969  + 10225.25871653050 *T)
    sumX1  +=       0.163 * cos(4.53414457768  + 10201.80001085851 *T)
    sumX1  +=       0.156 * cos(6.06699266861  + 10316.62213791310 *T)
    sumX1  +=       0.156 * cos(3.42689337876  + 10110.43658947590 *T)
    sumX1  +=       0.178 * cos(1.54946223064  + 10459.36100992390 *T)
    sumX1  +=       0.172 * cos(4.28587752330  + 14143.73905991410 *T)
    sumX1  +=       0.163 * cos(3.33734155709  + 8616.85528644289 *T)
    sumX1  +=       0.163 * cos(4.34265322302  + 6.86972951729 *T)
    sumX1  +=       0.157 * cos(5.15568499427  + 13554.14179039430 *T)
    sumX1  +=       0.152 * cos(5.19097050991  + 640.14110379750 *T)
    sumX1  +=       0.149 * cos(4.59659252448  + 7063.87756813929 *T)
    sumX1  +=       0.179 * cos(2.57671187060  + 29864.57784479250 *T)
    sumX1  +=       0.140 * cos(3.74361509193  + 8631.08238044450 *T)
    sumX1  +=       0.140 * cos(1.47790992779  + 529.93478257810 *T)
    sumX1  +=       0.131 * cos(4.02274509510  + 9794.04471981930 *T)
    sumX1  +=       0.170 * cos(1.92390663066  + 9153.65979853830 *T)
    sumX1  +=       0.128 * cos(5.46982761491  + 10633.01400756970 *T)
    sumX1  +=       0.177 * cos(4.60925492420  + 5481.49873635110 *T)
    sumX1  +=       0.154 * cos(0.98130891951  + 2119.00767786191 *T)
    sumX1  +=       0.131 * cos(0.85435524802  + 16984.23996494010 *T)
    sumX1  +=       0.131 * cos(2.78432814801  + 9437.51911740350 *T)
    sumX1  +=       0.118 * cos(0.33770416850  + 11790.38527117530 *T)
    sumX1  +=       0.140 * cos(3.57821725968  + 191.69208359510 *T)
    sumX1  +=       0.143 * cos(4.89084114003  + 7994.77225950771 *T)
    sumX1  +=       0.123 * cos(2.80558057017  + 213.54291292150 *T)
    sumX1  +=       0.111 * cos(1.17808071549  + 8661.99650607950 *T)
    sumX1  +=       0.121 * cos(2.21290119959  + 28521.33659573810 *T)
    sumX1  +=       0.119 * cos(1.63506882892  + 3929.96587873609 *T)
    sumX1  +=       0.106 * cos(5.90822125685  + 1109.62236957690 *T)
    sumX1  +=       0.099 * cos(2.30048706584  + 20.53157800890 *T)
    sumX1  +=       0.104 * cos(6.17530863971  + 36147.65369478390 *T)
    sumX1  +=       0.107 * cos(3.35632619038  + 22805.49174851010 *T)
    sumX1  +=       0.125 * cos(4.66656170776  + 12432.28646788130 *T)
    sumX1  +=       0.101 * cos(3.14294490163  + 1052.51220067191 *T)
    sumX1  +=       0.093 * cos(0.14824204033  + 574.10098085129 *T)
    sumX1  +=       0.129 * cos(4.90531185126  + 7564.58690325449 *T)
    sumX1  +=       0.092 * cos(1.92722181403  + 9359.84534697549 *T)
    sumX1  +=       0.099 * cos(2.32246660439  + 31441.92138724031 *T)
    sumX1  +=       0.091 * cos(6.18685264626  + 775.27879384050 *T)
    sumX1  +=       0.112 * cos(5.12786964903  + 10186.74340892770 *T)
    sumX1  +=       0.095 * cos(3.66782801204  + 10433.94200613330 *T)
    sumX1  +=       0.095 * cos(5.82605803532  + 9993.11672125570 *T)
    sumX1  +=       0.092 * cos(0.35578303465  + 14143.25142494710 *T)
    sumX1  +=       0.096 * cos(5.32806010513  + 220.65645992230 *T)
    sumX1  +=       0.084 * cos(5.80809189896  + 6770.46678376209 *T)
    sumX1  +=       0.088 * cos(1.99015323508  + 11322.42028082089 *T)
    sumX1  +=       0.110 * cos(0.30263518591  + 11802.60225897830 *T)
    sumX1  +=       0.108 * cos(5.13400336320  + 29580.71852592730 *T)
    sumX1  +=       0.093 * cos(5.89321185295  + 2957.95971196010 *T)
    sumX1  +=       0.101 * cos(2.09186878730  + 6770.95441872910 *T)
    sumX1  +=       0.105 * cos(4.75827384051  + 8662.48414104651 *T)
    sumX1  +=       0.094 * cos(1.47978763830  + 20433.92845690630 *T)
    sumX1  +=       0.089 * cos(2.27805851709  + 15720.59496739490 *T)
    sumX1  +=       0.088 * cos(5.79703525886  + 206.42936592071 *T)
    sumX1  +=       0.101 * cos(2.08027263650  + 7085.14059859870 *T)
    sumX1  +=       0.077 * cos(4.50520624769  + 10177.50149701710 *T)
    sumX1  +=       0.077 * cos(4.98867979968  + 10249.55723037191 *T)
    sumX1  +=       0.087 * cos(1.15473517110  + 2378.92065608810 *T)
    sumX1  +=       0.080 * cos(1.30578005980  + 5507.30942118390 *T)
    sumX1  +=       0.100 * cos(4.88512859986  + 14945.55999103790 *T)
    sumX1  +=       0.096 * cos(2.91515384440  + 8624.45646841071 *T)
    sumX1  +=       0.071 * cos(1.17266085434  + 24150.32386882850 *T)
    sumX1  +=       0.089 * cos(6.18976687503  + 18875.76968725751 *T)
    sumX1  +=       0.071 * cos(0.26607557139  + 29043.91401383189 *T)
    sumX1  +=       0.075 * cos(6.05522319041  + 19897.12394481090 *T)
    sumX1  +=       0.068 * cos(3.96046123295  + 10384.20223431370 *T)
    sumX1  +=       0.068 * cos(5.53342481442  + 10042.85649307530 *T)
    sumX1  +=       0.077 * cos(0.65929723000  + 29573.60497892650 *T)
    sumX1  +=       0.073 * cos(0.21801456964  + 29587.83207292810 *T)
    sumX1  +=       0.063 * cos(3.39603171816  + 10239.34004852730 *T)
    sumX1  +=       0.086 * cos(3.93300049009  + 19650.80466361450 *T)
    sumX1  +=       0.086 * cos(0.41942430117  + 19360.31943271550 *T)
    sumX1  +=       0.061 * cos(0.66096242131  + 29058.14110783350 *T)
    sumX1  +=       0.081 * cos(4.74585662186  + 11764.57458634250 *T)
    sumX1  +=       0.066 * cos(4.80481512048  + 26735.70144472970 *T)
    sumX1  +=       0.075 * cos(1.14917066460  + 25934.36814857290 *T)
    sumX1  +=       0.067 * cos(1.15917986378  + 40077.86339100350 *T)
    sumX1  +=       0.059 * cos(4.95335370099  + 10988.56434005150 *T)
    sumX1  +=       0.077 * cos(0.71010883429  + 20213.51581446750 *T)
    sumX1  +=       0.077 * cos(3.31402335732  + 7834.36489012290 *T)
    sumX1  +=       0.057 * cos(1.06828917998  + 10404.48999483909 *T)
    sumX1  +=       0.057 * cos(1.40187066473  + 110.45013870291 *T)
    sumX1  +=       0.067 * cos(1.57273164979  + 17777.87244946549 *T)
    sumX1  +=       0.060 * cos(6.21424049877  + 30831.54872222810 *T)
    sumX1  +=       0.058 * cos(2.47733285182  + 12592.20620229910 *T)
    sumX1  +=       0.054 * cos(4.90109601900  + 19317.43635781211 *T)
    sumX1  +=       0.065 * cos(3.16919569918  + 25158.84553724890 *T)
    sumX1  +=       0.064 * cos(2.20893920610  + 18308.05104952710 *T)
    sumX1  +=       0.059 * cos(0.10111317126  + 9779.32999085069 *T)
    sumX1  +=       0.053 * cos(1.51934336072  + 27490.93629552830 *T)
    sumX1  +=       0.050 * cos(1.75539922827  + 10103.32304247510 *T)
    sumX1  +=       0.050 * cos(1.45530151192  + 10323.73568491390 *T)
    sumX1  +=       0.050 * cos(6.21018693293  + 39794.00407213830 *T)
    sumX1  +=       0.055 * cos(5.90009200198  + 9793.55708485230 *T)
    sumX1  +=       0.048 * cos(4.23138520787  + 29089.05523346850 *T)
    sumX1  +=       0.051 * cos(0.59694211544  + 26709.89075989690 *T)
    sumX1  +=       0.047 * cos(3.85157632106  + 16522.41589851870 *T)
    sumX1  +=       0.060 * cos(6.18581659700  + 12592.69383726610 *T)
    sumX1  +=       0.057 * cos(3.26039666468  + 2218.51328670329 *T)
    sumX1  +=       0.047 * cos(3.62423252693  + 20640.11400534350 *T)
    sumX1  +=       0.046 * cos(4.31718986476  + 19786.91762359150 *T)
    sumX1  +=       0.047 * cos(2.61830291086  + 57836.89451481709 *T)
    sumX1  +=       0.045 * cos(3.93917249780  + 3532.30451029490 *T)
    sumX1  +=       0.047 * cos(0.02859294876  + 22003.67081738630 *T)
    sumX1  +=       0.045 * cos(4.38533780662  + 30000.20316980250 *T)
    sumX1  +=       0.043 * cos(5.13871521232  + 29786.90407436450 *T)
    sumX1  +=       0.051 * cos(2.50512066478  + 44008.07308722310 *T)
    sumX1  +=       0.042 * cos(4.64511163321  + 9999.74263328950 *T)
    sumX1  +=       0.050 * cos(5.23051268807  + 23581.01435983410 *T)
    sumX1  +=       0.051 * cos(1.29476911793  + 13341.91812879030 *T)
    sumX1  +=       0.040 * cos(5.56163575062  + 33019.26492968810 *T)
    sumX1  +=       0.051 * cos(1.12826192024  + 13656.10430865991 *T)
    sumX1  +=       0.048 * cos(5.72986884882  + 213.05527795450 *T)
    sumX1  +=       0.039 * cos(5.77845089286  + 20447.59030539790 *T)
    sumX1  +=       0.039 * cos(1.18630935880  + 10595.93826095070 *T)
    sumX1  +=       0.042 * cos(6.23472824067  + 30665.91114094930 *T)
    sumX1  +=       0.042 * cos(0.50469059035  + 3914.71340755109 *T)
    sumX1  +=       0.042 * cos(4.77756101392  + 38204.93117685450 *T)
    sumX1  +=       0.041 * cos(3.60025537560  + 17469.09901542890 *T)
    sumX1  +=       0.036 * cos(5.95301096335  + 10077.90403868451 *T)
    sumX1  +=       0.036 * cos(3.54087508402  + 10349.15468870450 *T)
    sumX1  +=       0.040 * cos(0.08435320836  + 18830.14083265389 *T)
    sumX1  +=       0.039 * cos(5.88165362901  + 18844.36792665549 *T)
    sumX1  +=       0.034 * cos(3.36376891454  + 21001.15970824030 *T)
    sumX1  +=       0.034 * cos(1.57325955329  + 13936.55068765050 *T)
    sumX1  +=       0.032 * cos(0.62824969555  + 36949.47462590770 *T)
    sumX1  +=       0.033 * cos(4.53469110076  + 16983.75232997309 *T)
    sumX1  +=       0.040 * cos(3.88665965181  + 47938.28278344270 *T)
    sumX1  +=       0.032 * cos(5.46497273946  + 39301.85314471250 *T)
    sumX1  +=       0.032 * cos(5.46497273946  + 39302.34077967950 *T)
    sumX1  +=       0.036 * cos(0.13872965215  + 19374.54652671710 *T)
    sumX1  +=       0.032 * cos(5.49940747482  + 16894.75421709410 *T)
    sumX1  +=       0.042 * cos(0.37006377215  + 55798.70217588190 *T)
    sumX1  +=       0.031 * cos(3.51648880247  + 20043.91837768230 *T)
    sumX1  +=       0.033 * cos(1.44339512827  + 48418.21672306549 *T)
    sumX1  +=       0.031 * cos(1.24271036891  + 9573.63207738050 *T)
    sumX1  +=       0.034 * cos(5.27191504099  + 54247.16931826690 *T)
    sumX1  +=       0.032 * cos(2.12080485437  + 10191.67853436810 *T)
    sumX1  +=       0.032 * cos(1.08989588581  + 10235.38019302090 *T)
    sumX1  +=       0.031 * cos(2.88701937696  + 8087.16432134830 *T)
    sumX1  +=       0.032 * cos(1.43989537444  + 19896.63630984389 *T)
    sumX1  +=       0.036 * cos(0.86097595428  + 7255.32583425090 *T)
    sumX1  +=       0.030 * cos(1.96661961760  + 10853.42665000850 *T)
    sumX1  +=       0.037 * cos(0.40709954927  + 27511.22405605370 *T)
    sumX1  +=       0.032 * cos(3.94523183143  + 19366.94534474929 *T)
    sumX1  +=       0.029 * cos(5.21162982204  + 10291.20313412251 *T)
    sumX1  +=       0.029 * cos(4.28225622533  + 10135.85559326650 *T)
    sumX1  +=       0.028 * cos(5.74069429835  + 104351.85638378028 *T)
    sumX1  +=       0.032 * cos(4.16582591800  + 20206.40226746670 *T)
    sumX1  +=       0.029 * cos(6.04399510869  + 29864.09020982550 *T)
    sumX1  +=       0.029 * cos(3.70427099153  + 49515.13869092350 *T)
    sumX1  +=       0.027 * cos(5.20809241529  + 5661.08823166870 *T)
    sumX1  +=       0.034 * cos(2.13914627667  + 9161.26098050610 *T)
    sumX1  +=       0.027 * cos(2.35964005449  + 6872.42930202769 *T)
    sumX1  +=       0.032 * cos(5.37255836333  + 9836.15575577870 *T)
    sumX1  +=       0.032 * cos(4.12132768404  + 10590.90297161030 *T)
    sumX1  +=       0.027 * cos(6.25293311610  + 6298.57213865991 *T)
    sumX1  +=       0.035 * cos(5.35069959846  + 9986.05323090550 *T)
    sumX1  +=       0.035 * cos(4.14318644891  + 10441.00549648351 *T)
    sumX1  +=       0.033 * cos(2.06705744681  + 31441.43375227330 *T)
    sumX1  +=       0.026 * cos(1.99269274261  + 40879.68432212730 *T)
    sumX1  +=       0.034 * cos(2.86141693506  + 1059.13811270570 *T)
    sumX1  +=       0.028 * cos(3.27143415157  + 27197.03787618409 *T)
    sumX1  +=       0.026 * cos(4.77604105797  + 34363.60941503950 *T)
    sumX1  +=       0.032 * cos(1.06669637529  + 11265.79774688290 *T)
    sumX1  +=       0.029 * cos(3.31179813455  + 30206.38871823969 *T)
    sumX1  +=       0.032 * cos(5.28518925051  + 51868.49247966230 *T)
    sumX1  +=       0.029 * cos(3.60960317836  + 35371.64344849290 *T)
    sumX1  +=       0.029 * cos(3.60960317836  + 35372.13108345990 *T)
    sumX1  +=       0.027 * cos(1.39820608512  + 34596.12083716890 *T)
    sumX1  +=       0.033 * cos(1.79087945691  + 59728.91187210151 *T)
    sumX1  +=       0.029 * cos(3.70229059266  + 20220.62936146830 *T)
    sumX1  +=       0.025 * cos(4.71719114041  + 25933.88051360590 *T)
    sumX1  +=       0.024 * cos(5.85674327230  + 8734.17515466309 *T)
    sumX1  +=       0.027 * cos(2.50446352876  + 66813.80865321669 *T)
    sumX1  +=       0.024 * cos(1.21352192025  + 36173.46437961670 *T)
    sumX1  +=       0.024 * cos(4.42282415616  + 31022.99698833970 *T)
    sumX1  +=       0.023 * cos(0.26441959495  + 24382.83529095790 *T)
    sumX1  +=       0.028 * cos(1.01958443856  + 27461.46703058189 *T)
    sumX1  +=       0.027 * cos(2.82757763572  + 30220.61581224130 *T)
    sumX1  +=       0.023 * cos(2.71148449796  + 10367.30817417930 *T)
    sumX1  +=       0.023 * cos(0.49921624223  + 10059.75055320970 *T)
    sumX1  +=       0.023 * cos(5.14356401989  + 20452.62559473830 *T)
    sumX1  +=       0.026 * cos(2.79748800924  + 39271.42665404450 *T)
    sumX1  +=       0.026 * cos(3.27996977321  + 39257.19956004290 *T)
    sumX1  +=       0.022 * cos(0.64539022706  + 43232.06284093210 *T)
    sumX1  +=       0.022 * cos(0.64539022706  + 43232.55047589910 *T)
    sumX1  +=       0.026 * cos(3.68377277974  + 8101.39141534989 *T)
    sumX1  +=       0.022 * cos(2.92850972756  + 24978.76840696430 *T)
    sumX1  +=       0.022 * cos(2.54378881371  + 25158.35790228190 *T)
    sumX1  +=       0.030 * cos(4.35689894819  + 17248.18148437089 *T)
    sumX1  +=       0.022 * cos(4.40135816223  + 27991.15799567649 *T)
    sumX1  +=       0.022 * cos(1.75419561511  + 30426.80136067850 *T)
    sumX1  +=       0.022 * cos(5.61039853516  + 802.06474860730 *T)
    sumX1  +=       0.028 * cos(5.29619670984  + 17297.93850984270 *T)
    sumX1  +=       0.021 * cos(1.01517564916  + 9147.03388650451 *T)
    sumX1  +=       0.024 * cos(5.57258032099  + 46386.74992582770 *T)
    sumX1  +=       0.021 * cos(2.72351924355  + 27197.52551115110 *T)
    sumX1  +=       0.021 * cos(3.36801059218  + 44809.89401834690 *T)
    sumX1  +=       0.024 * cos(5.90015079750  + 52176.05010063190 *T)
    sumX1  +=       0.025 * cos(4.38058629438  + 22805.97938347710 *T)
    sumX1  +=       0.022 * cos(0.93569574784  + 9566.03089541269 *T)
    sumX1  +=       0.023 * cos(2.47187255983  + 6309.13035230770 *T)
    sumX1  +=       0.021 * cos(2.05775290145  + 10949.40587722630 *T)
    sumX1  +=       0.021 * cos(1.15294783874  + 9477.65285016271 *T)
    sumX1  +=       0.024 * cos(0.56697771725  + 10288.31096226180 *T)
    sumX1  +=       0.024 * cos(2.64372302294  + 10138.74776512720 *T)
    sumX1  +=       0.027 * cos(0.54831840561  + 9580.25798941430 *T)
    sumX1  +=       0.020 * cos(5.88803150614  + 8635.69818627969 *T)
    sumX1  +=       0.020 * cos(2.21258610934  + 11280.02484088450 *T)
    sumX1  +=       0.021 * cos(3.74792739506  + 3531.81687532789 *T)
    sumX1  +=       0.022 * cos(1.10471815077  + 62883.59895699710 *T)
    sumX1  +=       0.020 * cos(4.47242869756  + 17468.61138046190 *T)
    sumX1  +=       0.023 * cos(2.67526294292  + 10263.28638916630 *T)
    sumX1  +=       0.023 * cos(0.53543779727  + 10163.77233822270 *T)
    sumX1  +=       0.027 * cos(4.35265351175  + 32216.95636359729 *T)
    sumX1  +=       0.021 * cos(0.20196459716  + 47624.09660357310 *T)
    sumX1  +=       0.023 * cos(4.44238440406  + 55022.69192959090 *T)
    sumX1  +=       0.021 * cos(5.91319275029  + 18300.44986755929 *T)
    sumX1  +=       0.026 * cos(3.21324974687  + 63659.12156832110 *T)
    sumX1  +=       0.021 * cos(1.29300419185  + 43071.65547154729 *T)
    sumX1  +=       0.019 * cos(1.80883011026  + 20316.60858868610 *T)
    sumX1  +=       0.027 * cos(2.91239597616  + 28286.74666737769 *T)
    sumX1  +=       0.020 * cos(0.92201397306  + 62081.77802587330 *T)
    sumX1  +=       0.021 * cos(3.71520409064  + 24341.77213494010 *T)
    sumX1  +=       0.020 * cos(2.70818412083  + 49477.59865325469 *T)
    sumX1  +=       0.020 * cos(3.92540543269  + 70744.01834943630 *T)
    sumX1  +=       0.018 * cos(3.24042993488  + 14128.48658872910 *T)
    sumX1  +=       0.022 * cos(1.96762524944  + 27991.64563064350 *T)
    sumX1  +=       0.018 * cos(5.40557069193  + 9050.56702431969 *T)
    sumX1  +=       0.018 * cos(1.21555571929  + 4705.48849006010 *T)
    sumX1  +=       0.017 * cos(1.63173823644  + 28313.04498717750 *T)
    sumX1  +=       0.017 * cos(4.46941749243  + 10497.38868255971 *T)
    sumX1  +=       0.017 * cos(5.02446855494  + 9929.67004482930 *T)
    sumX1  +=       0.017 * cos(0.77261247444  + 19624.99397878170 *T)
    sumX1  +=       0.018 * cos(5.82240394487  + 29050.53992586569 *T)
    sumX1  +=       0.017 * cos(6.13439373448  + 18314.67696156089 *T)
    sumX1  +=       0.016 * cos(4.75152522759  + 48740.10371456650 *T)
    sumX1  +=       0.018 * cos(2.82811097731  + 38526.33053338850 *T)
    sumX1  +=       0.018 * cos(0.02896271913  + 42430.72954477530 *T)
    sumX1  +=       0.018 * cos(4.17302923646  + 42456.54022960810 *T)
    sumX1  +=       0.018 * cos(1.01249884206  + 33794.29990604510 *T)
    sumX1  +=       0.018 * cos(1.01249884206  + 33794.78754101210 *T)
    sumX1  +=       0.016 * cos(5.43519924864  + 9745.56437334010 *T)
    sumX1  +=       0.016 * cos(4.05868679873  + 10681.49435404891 *T)
    sumX1  +=       0.016 * cos(4.91307958316  + 38734.62214194910 *T)
    sumX1  +=       0.016 * cos(1.98563978946  + 47162.27253715170 *T)
    sumX1  +=       0.016 * cos(1.98563978946  + 47162.76017211870 *T)
    sumX1  +=       0.015 * cos(5.40287159543  + 10199.30226969290 *T)
    sumX1  +=       0.015 * cos(4.09101445194  + 10227.75645769610 *T)
    sumX1  +=       0.015 * cos(4.40286421658  + 41654.71929848429 *T)
    sumX1  +=       0.017 * cos(0.71063741169  + 50316.95962204730 *T)
    sumX1  +=       0.015 * cos(2.70406423385  + 53285.42865272530 *T)
    sumX1  +=       0.015 * cos(2.34568334494  + 66011.98772209289 *T)
    sumX1  +=       0.015 * cos(2.83728676171  + 38727.50859494830 *T)
    sumX1  +=       0.015 * cos(1.51871369727  + 46360.93924099490 *T)
    sumX1  +=       0.019 * cos(4.60379625756  + 67589.33126454070 *T)
    sumX1  +=       0.014 * cos(1.35244180937  + 21228.14820606230 *T)
    sumX1  +=       0.014 * cos(0.37440411348  + 40000.18962057550 *T)
    sumX1  +=       0.014 * cos(1.52734358006  + 11027.07964765431 *T)
    sumX1  +=       0.014 * cos(1.68335716013  + 9399.97907973470 *T)
    sumX1  +=       0.014 * cos(1.39514541459  + 37895.67010785090 *T)
    sumX1  +=       0.014 * cos(1.44888424284  + 37410.81105736210 *T)
    sumX1  +=       0.014 * cos(2.94215733290  + 50291.14893721450 *T)
    sumX1  +=       0.013 * cos(4.17219621815  + 12565.90788249930 *T)
    sumX1  +=       0.015 * cos(6.22361957178  + 21486.19684009470 *T)
    sumX1  +=       0.013 * cos(6.14132227253  + 52670.31341078610 *T)

    # VENUS X Coord
    sumX2  = 0
    sumX2  +=   39116.576 * cos(4.74721085665  + 10213.52936369450 *T)
    sumX2  +=    1782.208 * cos(4.51495547304  + 0.24381748350 *T)
    sumX2  +=     953.335 * cos(3.63516488566  + 10213.04172872750 *T)
    sumX2  +=     818.754 * cos(5.17913474548  + 20426.81490990550 *T)
    sumX2  +=      12.014 * cos(5.55504004602  + 30640.10045611650 *T)
    sumX2  +=       8.598 * cos(0.48030487437  + 10239.82768349430 *T)
    sumX2  +=       8.636 * cos(2.67172999291  + 10187.23104389470 *T)
    sumX2  +=       3.631 * cos(5.13091719462  + 20426.32727493849 *T)
    sumX2  +=       2.965 * cos(3.11794464067  + 8636.18582124671 *T)
    sumX2  +=       3.797 * cos(0.73505869096  + 9438.00675237050 *T)
    sumX2  +=       3.421 * cos(2.39489149839  + 10989.05197501850 *T)
    sumX2  +=       2.218 * cos(3.69616749811  + 10022.08109758290 *T)
    sumX2  +=       2.134 * cos(5.69388419448  + 10404.97762980610 *T)
    sumX2  +=       2.187 * cos(6.14628758046  + 11790.87290614230 *T)
    sumX2  +=       1.794 * cos(0.26131314165  + 775.76642880750 *T)
    sumX2  +=       1.421 * cos(1.89983182794  + 2353.10997125530 *T)
    sumX2  +=       0.972 * cos(2.65726077013  + 1577.09972496430 *T)
    sumX2  +=       0.682 * cos(1.68817935727  + 4705.97612502710 *T)
    sumX2  +=       0.658 * cos(4.84602149726  + 9683.83839859990 *T)
    sumX2  +=       0.715 * cos(1.34290194955  + 9830.63283147130 *T)
    sumX2  +=       0.683 * cos(1.84498400268  + 10596.42589591770 *T)
    sumX2  +=       0.568 * cos(2.68352094618  + 19651.29229858150 *T)
    sumX2  +=       0.544 * cos(3.07694729173  + 191.20444862810 *T)
    sumX2  +=       0.678 * cos(4.45588465096  + 18073.94875613370 *T)
    sumX2  +=       0.574 * cos(5.09706238916  + 801.57711364030 *T)
    sumX2  +=       0.418 * cos(4.31488544265  + 10743.22032878910 *T)
    sumX2  +=       0.561 * cos(3.32645295084  + 6283.31966747490 *T)
    sumX2  +=       0.531 * cos(0.97430911335  + 7058.84227879890 *T)
    sumX2  +=       0.371 * cos(5.86501171146  + 9411.70843257070 *T)
    sumX2  +=       0.414 * cos(2.16574465889  + 529.93478257810 *T)
    sumX2  +=       0.287 * cos(3.12423584879  + 9104.15081160110 *T)
    sumX2  +=       0.314 * cos(1.28037944612  + 15721.08260236190 *T)
    sumX2  +=       0.301 * cos(2.30568402177  + 26.05450231630 *T)
    sumX2  +=       0.354 * cos(2.15592461692  + 13368.21644859010 *T)
    sumX2  +=       0.361 * cos(3.03077999717  + 14143.73905991410 *T)
    sumX2  +=       0.349 * cos(2.69866467882  + 22004.15845235330 *T)
    sumX2  +=       0.253 * cos(3.49550693096  + 1059.62574767270 *T)
    sumX2  +=       0.253 * cos(6.11019925201  + 11322.90791578790 *T)
    sumX2  +=       0.253 * cos(4.31046856084  + 3154.44326741210 *T)
    sumX2  +=       0.321 * cos(3.51926077378  + 11015.35029481830 *T)
    sumX2  +=       0.305 * cos(5.79665591896  + 3128.63258257930 *T)
    sumX2  +=       0.202 * cos(1.70860223309  + 10426.82845913250 *T)
    sumX2  +=       0.195 * cos(6.03063629123  + 20618.26317601710 *T)
    sumX2  +=       0.178 * cos(1.34005224381  + 10000.23026825651 *T)
    sumX2  +=       0.157 * cos(6.22467540109  + 40853.38600232750 *T)
    sumX2  +=       0.159 * cos(5.43308713463  + 382.65271473970 *T)
    sumX2  +=       0.177 * cos(1.96418161194  + 21228.63584102931 *T)
    sumX2  +=       0.126 * cos(2.62766766451  + 18849.47136745770 *T)
    sumX2  +=       0.131 * cos(0.73006046049  + 1577.58735993130 *T)
    sumX2  +=       0.148 * cos(4.01019073124  + 5507.30942118390 *T)
    sumX2  +=       0.139 * cos(4.55765675380  + 7860.66320992270 *T)
    sumX2  +=       0.146 * cos(4.19737673183  + 19897.12394481090 *T)
    sumX2  +=       0.133 * cos(3.60550981906  + 17298.42614480970 *T)
    sumX2  +=       0.088 * cos(0.74174526538  + 8623.96883344369 *T)
    sumX2  +=       0.116 * cos(3.56002594290  + 9153.65979853830 *T)
    sumX2  +=       0.111 * cos(5.81554919476  + 20453.11322970530 *T)
    sumX2  +=       0.077 * cos(5.72141298699  + 7084.65296363170 *T)
    sumX2  +=       0.081 * cos(3.85700349092  + 2352.62233628830 *T)
    sumX2  +=       0.088 * cos(0.02333507058  + 3930.45351370310 *T)
    sumX2  +=       0.077 * cos(3.12509816452  + 16496.60521368590 *T)
    sumX2  +=       0.069 * cos(1.03046490507  + 23581.50199480110 *T)
    sumX2  +=       0.058 * cos(4.44021603496  + 9154.14743350530 *T)
    sumX2  +=       0.062 * cos(5.17807799754  + 1109.13473460990 *T)
    sumX2  +=       0.063 * cos(5.42244602531  + 9437.51911740350 *T)
    sumX2  +=       0.047 * cos(3.73566941706  + 29051.02756083270 *T)
    sumX2  +=       0.054 * cos(2.34590128189  + 20809.71144212870 *T)
    sumX2  +=       0.055 * cos(0.27212302992  + 4731.78680985990 *T)
    sumX2  +=       0.046 * cos(5.56053620908  + 19367.43297971630 *T)
    sumX2  +=       0.044 * cos(4.38854627107  + 10787.87416202930 *T)
    sumX2  +=       0.044 * cos(5.10533977629  + 9639.18456535971 *T)
    sumX2  +=       0.040 * cos(0.19406822983  + 8094.27786834910 *T)
    sumX2  +=       0.041 * cos(0.23997896197  + 3929.96587873609 *T)
    sumX2  +=       0.047 * cos(4.99296496795  + 12566.39551746630 *T)
    sumX2  +=       0.038 * cos(5.33087019939  + 6282.83203250789 *T)
    sumX2  +=       0.045 * cos(4.21174903018  + 25934.36814857290 *T)
    sumX2  +=       0.034 * cos(4.95079079946  + 28287.23430234470 *T)
    sumX2  +=       0.033 * cos(0.84076094156  + 11014.86265985130 *T)
    sumX2  +=       0.034 * cos(1.04665276837  + 14919.26167123810 *T)
    sumX2  +=       0.032 * cos(3.42124102189  + 24357.02460612510 *T)
    sumX2  +=       0.028 * cos(3.83338891686  + 20235.36664379391 *T)
    sumX2  +=       0.040 * cos(3.61668312604  + 29580.71852592730 *T)
    sumX2  +=       0.033 * cos(5.35791701104  + 4552.19731454230 *T)
    sumX2  +=       0.033 * cos(0.97035705954  + 6490.02040477150 *T)
    sumX2  +=       0.027 * cos(0.80253867623  + 10640.12755457050 *T)
    sumX2  +=       0.026 * cos(2.37794131191  + 9786.93117281850 *T)
    sumX2  +=       0.026 * cos(0.52950336946  + 3149.40797807170 *T)
    sumX2  +=       0.025 * cos(2.09390807595  + 10192.75396820210 *T)
    sumX2  +=       0.025 * cos(4.26281208500  + 10234.30475918690 *T)
    sumX2  +=       0.024 * cos(1.66809456061  + 21536.19346199890 *T)
    sumX2  +=       0.026 * cos(5.00411416070  + 14143.25142494710 *T)
    sumX2  +=       0.024 * cos(6.24637715602  + 18307.56341456009 *T)
    sumX2  +=       0.024 * cos(6.23872379023  + 5507.79705615090 *T)
    sumX2  +=       0.025 * cos(0.42768275114  + 18875.28205229050 *T)
    sumX2  +=       0.030 * cos(3.82426253554  + 20400.51659010570 *T)
    sumX2  +=       0.025 * cos(2.48440063286  + 27511.71169102070 *T)
    sumX2  +=       0.025 * cos(1.71524233724  + 8661.99650607950 *T)
    sumX2  +=       0.019 * cos(5.30003473361  + 11272.91129388371 *T)
    sumX2  +=       0.022 * cos(0.30636210636  + 30639.61282114949 *T)
    sumX2  +=       0.026 * cos(2.43625052079  + 213.54291292150 *T)
    sumX2  +=       0.018 * cos(3.34143186430  + 28521.33659573810 *T)
    sumX2  +=       0.023 * cos(2.19310017282  + 13937.03832261750 *T)
    sumX2  +=       0.024 * cos(0.64478935102  + 13367.72881362310 *T)
    sumX2  +=       0.021 * cos(0.38789416530  + 10186.74340892770 *T)
    sumX2  +=       0.019 * cos(2.13540095490  + 10988.56434005150 *T)
    sumX2  +=       0.018 * cos(3.02877957634  + 32217.44399856430 *T)
    sumX2  +=       0.020 * cos(5.05892110724  + 191.69208359510 *T)
    sumX2  +=       0.017 * cos(1.00021130316  + 15874.86141284670 *T)
    sumX2  +=       0.014 * cos(3.13205445097  + 21202.33752122950 *T)
    sumX2  +=       0.014 * cos(2.49765103801  + 17277.65074931730 *T)
    sumX2  +=       0.015 * cos(0.11837091281  + 18073.46112116669 *T)
    sumX2  +=       0.014 * cos(4.77259175609  + 10239.34004852730 *T)
    sumX2  +=       0.014 * cos(2.38979050420  + 14945.07235607090 *T)
    sumX2  +=       0.014 * cos(4.18962126273  + 8662.48414104651 *T)
    sumX2  +=       0.014 * cos(5.23413049702  + 11764.57458634250 *T)
    sumX2  +=       0.013 * cos(1.93488532828  + 34570.31015233610 *T)
    sumX2  +=       0.013 * cos(0.61222694816  + 22779.68106367730 *T)
    sumX2  +=       0.012 * cos(3.85463106546  + 31441.92138724031 *T)
    sumX2  +=       0.011 * cos(4.27550517900  + 10220.64291069530 *T)
    sumX2  +=       0.011 * cos(2.07789625538  + 10206.41581669370 *T)
    sumX2  +=       0.010 * cos(3.61798826402  + 7.35736448430 *T)
    sumX2  +=       0.010 * cos(1.57784029028  + 22003.67081738630 *T)
    sumX2  +=       0.012 * cos(3.46936245747  + 25158.84553724890 *T)
    sumX2  +=       0.009 * cos(3.35948007941  + 38500.51984855570 *T)
    sumX2  +=       0.010 * cos(2.93060547107  + 11790.38527117530 *T)
    sumX2  +=       0.012 * cos(2.06733085074  + 17297.93850984270 *T)
    sumX2  +=       0.008 * cos(2.77292844888  + 10404.48999483909 *T)
    sumX2  +=       0.009 * cos(4.94797813171  + 29089.05523346850 *T)
    sumX2  +=       0.007 * cos(1.94867772097  + 43232.55047589910 *T)
    sumX2  +=       0.007 * cos(6.01401025413  + 30000.20316980250 *T)
    sumX2  +=       0.009 * cos(4.47573265145  + 7860.17557495569 *T)
    sumX2  +=       0.009 * cos(4.90406740797  + 3531.81687532789 *T)
    sumX2  +=       0.007 * cos(1.80027626747  + 22805.49174851010 *T)
    sumX2  +=       0.007 * cos(1.18159563994  + 529.44714761109 *T)

    # VENUS X Coord
    sumX3  = 0
    sumX3  +=      55.518 * cos(5.27636048864  + 10213.04172872750 *T)
    sumX3  +=      57.047 * cos(5.87936770190  + 0.24381748350 *T)
    sumX3  +=      32.093 * cos(3.05612494200  + 20426.81490990550 *T)
    sumX3  +=      18.867 * cos(4.86548938285  + 10213.52936369450 *T)
    sumX3  +=       0.716 * cos(3.26370970167  + 30640.10045611650 *T)
    sumX3  +=       0.218 * cos(0.66337297711  + 20426.32727493849 *T)
    sumX3  +=       0.032 * cos(0.22308629368  + 8636.18582124671 *T)
    sumX3  +=       0.034 * cos(2.69459006102  + 10239.82768349430 *T)
    sumX3  +=       0.034 * cos(3.66311278763  + 10187.23104389470 *T)
    sumX3  +=       0.027 * cos(6.14322148767  + 11790.87290614230 *T)
    sumX3  +=       0.020 * cos(3.64902570031  + 775.76642880750 *T)
    sumX3  +=       0.016 * cos(0.61723457662  + 10022.08109758290 *T)
    sumX3  +=       0.016 * cos(5.73871486925  + 10404.97762980610 *T)
    sumX3  +=       0.010 * cos(3.76695027338  + 40853.38600232750 *T)
    sumX3  +=       0.010 * cos(2.80431500839  + 9438.00675237050 *T)
    sumX3  +=       0.010 * cos(3.65133831937  + 10989.05197501850 *T)
    sumX3  +=       0.008 * cos(5.27853417078  + 9683.83839859990 *T)
    sumX3  +=       0.007 * cos(6.04834530223  + 191.20444862810 *T)
    sumX3  +=       0.007 * cos(1.06297780672  + 10743.22032878910 *T)
    sumX3  +=       0.008 * cos(5.07274318566  + 4705.97612502710 *T)
    sumX3  +=       0.007 * cos(2.72342831598  + 19651.29229858150 *T)
    sumX3  +=       0.004 * cos(0.12637968025  + 9104.15081160110 *T)
    sumX3  +=       0.004 * cos(0.86625328320  + 1059.62574767270 *T)
    sumX3  +=       0.004 * cos(1.29045440101  + 15721.08260236190 *T)
    sumX3  +=       0.003 * cos(6.23075905972  + 11322.90791578790 *T)

    # VENUS X Coord
    sumX4  = 0
    sumX4  +=      83.081 * cos(1.47755047593  + 10213.52936369450 *T)
    sumX4  +=       2.486 * cos(0.71777141614  + 10213.04172872750 *T)
    sumX4  +=       0.657 * cos(1.07128766548  + 20426.81490990550 *T)
    sumX4  +=       0.217 * cos(3.60535760515  + 0.24381748350 *T)
    sumX4  +=       0.035 * cos(0.98556560154  + 30640.10045611650 *T)
    sumX4  +=       0.011 * cos(2.44309012697  + 20426.32727493849 *T)
    sumX4  +=       0.005 * cos(2.06924437235  + 10239.82768349430 *T)
    sumX4  +=       0.005 * cos(4.26079266220  + 10187.23104389470 *T)
    sumX4  +=       0.003 * cos(4.92755555887  + 2353.10997125530 *T)
    sumX4  +=       0.002 * cos(4.89604781036  + 8636.18582124671 *T)
    sumX4  +=       0.002 * cos(5.91399711778  + 1577.09972496430 *T)
    sumX4  +=       0.002 * cos(2.29946508014  + 9438.00675237050 *T)

    # VENUS X Coord
    sumX5  = 0
    sumX5  +=       0.631 * cos(1.59925460860  + 10213.52936369450 *T)
    sumX5  +=       0.095 * cos(2.54586723363  + 10213.04172872750 *T)
    sumX5  +=       0.088 * cos(1.72183649952  + 0.24381748350 *T)
    sumX5  +=       0.040 * cos(5.81128931245  + 20426.81490990550 *T)

    # VENUS Y Coord
    sumY0  = 0
    sumY0  +=  72268045.621 * cos(1.60535036499  + 10213.52936369450 *T)
    sumY0  +=  733886.107 * cos(3.86619610007  + 0.24381748350 *T)
    sumY0  +=  244692.613 * cos(2.48525998209  + 20426.81490990550 *T)
    sumY0  +=   63317.702 * cos(2.06415043880  + 10213.04172872750 *T)
    sumY0  +=    2795.043 * cos(5.04300618072  + 2353.10997125530 *T)
    sumY0  +=    1941.766 * cos(5.80436164129  + 1577.09972496430 *T)
    sumY0  +=    1242.687 * cos(3.36533956931  + 30640.10045611650 *T)
    sumY0  +=    1163.369 * cos(1.30924660823  + 18073.94875613370 *T)
    sumY0  +=    1044.280 * cos(0.18242473165  + 6283.31966747490 *T)
    sumY0  +=     767.420 * cos(5.30492070227  + 529.93478257810 *T)
    sumY0  +=     669.994 * cos(6.16996307370  + 14143.73905991410 *T)
    sumY0  +=     657.435 * cos(5.21342894071  + 8636.18582124671 *T)
    sumY0  +=     476.814 * cos(4.27269799631  + 10187.23104389470 *T)
    sumY0  +=     475.078 * cos(2.07969169968  + 10239.82768349430 *T)
    sumY0  +=     559.353 * cos(5.87818075735  + 22004.15845235330 *T)
    sumY0  +=     544.577 * cos(1.14724395566  + 11790.87290614230 *T)
    sumY0  +=     408.020 * cos(2.35529138555  + 775.76642880750 *T)
    sumY0  +=     281.345 * cos(1.23218938626  + 9683.83839859990 *T)
    sumY0  +=     270.035 * cos(2.10691120117  + 9438.00675237050 *T)
    sumY0  +=     268.860 * cos(5.13228928315  + 10743.22032878910 *T)
    sumY0  +=     299.980 * cos(0.94363100672  + 5507.30942118390 *T)
    sumY0  +=     214.307 * cos(5.46162572826  + 10022.08109758290 *T)
    sumY0  +=     241.347 * cos(4.23602697206  + 10989.05197501850 *T)
    sumY0  +=     208.372 * cos(0.89206276557  + 10404.97762980610 *T)
    sumY0  +=     271.224 * cos(1.05337463479  + 19897.12394481090 *T)
    sumY0  +=     264.276 * cos(0.40825385367  + 9153.65979853830 *T)
    sumY0  +=     214.535 * cos(2.94606412427  + 20426.32727493849 *T)
    sumY0  +=     175.488 * cos(1.41092496739  + 191.20444862810 *T)
    sumY0  +=     123.055 * cos(0.01778159955  + 1059.62574767270 *T)
    sumY0  +=     154.230 * cos(3.78480528747  + 4705.97612502710 *T)
    sumY0  +=     147.037 * cos(3.98774418232  + 19651.29229858150 *T)
    sumY0  +=     100.153 * cos(2.36411678071  + 9437.51911740350 *T)
    sumY0  +=      69.555 * cos(4.80550652867  + 9104.15081160110 *T)
    sumY0  +=      60.317 * cos(1.54595691846  + 11322.90791578790 *T)
    sumY0  +=      81.789 * cos(2.57258880007  + 15721.08260236190 *T)
    sumY0  +=      58.874 * cos(1.48684982669  + 20618.26317601710 *T)
    sumY0  +=      60.776 * cos(2.21391240560  + 4552.19731454230 *T)
    sumY0  +=      46.411 * cos(5.23309117015  + 10192.75396820210 *T)
    sumY0  +=      46.285 * cos(1.11880987202  + 10234.30475918690 *T)
    sumY0  +=      62.288 * cos(1.02030092694  + 25934.36814857290 *T)
    sumY0  +=      50.365 * cos(0.03955377805  + 8623.96883344369 *T)
    sumY0  +=      57.275 * cos(2.21318821830  + 3154.44326741210 *T)
    sumY0  +=      44.993 * cos(0.06128576651  + 19367.43297971630 *T)
    sumY0  +=      40.167 * cos(1.64087495621  + 9411.70843257070 *T)
    sumY0  +=      47.716 * cos(5.57543351345  + 213.54291292150 *T)
    sumY0  +=      39.763 * cos(3.21104786001  + 801.57711364030 *T)
    sumY0  +=      36.106 * cos(4.71533355073  + 11015.35029481830 *T)
    sumY0  +=      45.150 * cos(0.48042336302  + 29580.71852592730 *T)
    sumY0  +=      36.973 * cos(2.82409044668  + 7058.84227879890 *T)
    sumY0  +=      44.928 * cos(3.79012120364  + 13367.72881362310 *T)
    sumY0  +=      27.822 * cos(3.37420388243  + 1109.13473460990 *T)
    sumY0  +=      31.714 * cos(4.13937067249  + 15874.86141284670 *T)
    sumY0  +=      36.279 * cos(2.34561943360  + 10000.23026825651 *T)
    sumY0  +=      35.360 * cos(4.01138333911  + 10426.82845913250 *T)
    sumY0  +=      25.530 * cos(3.98045270905  + 7860.66320992270 *T)
    sumY0  +=      24.951 * cos(3.53621528232  + 13368.21644859010 *T)
    sumY0  +=      20.285 * cos(1.13150552857  + 10220.64291069530 *T)
    sumY0  +=      20.275 * cos(5.21708912625  + 10206.41581669370 *T)
    sumY0  +=      21.291 * cos(1.37841249317  + 3128.63258257930 *T)
    sumY0  +=      19.173 * cos(0.47393618404  + 7.35736448430 *T)
    sumY0  +=      18.901 * cos(0.61568845651  + 26.05450231630 *T)
    sumY0  +=      21.874 * cos(5.21237379147  + 17297.93850984270 *T)
    sumY0  +=      19.888 * cos(1.40722670834  + 3930.45351370310 *T)
    sumY0  +=      15.237 * cos(2.81371946863  + 9830.63283147130 *T)
    sumY0  +=      17.432 * cos(1.76492288618  + 3531.81687532789 *T)
    sumY0  +=      14.522 * cos(3.53852074947  + 10596.42589591770 *T)
    sumY0  +=      16.153 * cos(3.63782003268  + 7084.65296363170 *T)
    sumY0  +=      16.091 * cos(0.78643837533  + 20213.51581446750 *T)
    sumY0  +=      14.842 * cos(2.49914152987  + 29864.57784479250 *T)
    sumY0  +=      11.112 * cos(2.20077056640  + 28287.23430234470 *T)
    sumY0  +=      12.828 * cos(0.47914049065  + 1589.31671276730 *T)
    sumY0  +=      11.212 * cos(0.90580714897  + 9154.14743350530 *T)
    sumY0  +=      14.376 * cos(2.29215877818  + 23581.50199480110 *T)
    sumY0  +=      12.175 * cos(3.29545280575  + 21228.63584102931 *T)
    sumY0  +=       9.393 * cos(5.33561048405  + 11272.91129388371 *T)
    sumY0  +=      12.824 * cos(2.98347327697  + 18837.25437965469 *T)
    sumY0  +=       9.207 * cos(5.34806792169  + 5661.57586663570 *T)
    sumY0  +=       8.622 * cos(3.39909215992  + 21536.19346199890 *T)
    sumY0  +=      11.653 * cos(4.63096590501  + 3340.85624418330 *T)
    sumY0  +=      11.496 * cos(0.35144252183  + 21228.14820606230 *T)
    sumY0  +=       9.511 * cos(4.97800970942  + 17298.42614480970 *T)
    sumY0  +=      11.281 * cos(6.15396443981  + 9786.44353785149 *T)
    sumY0  +=       8.498 * cos(2.28560931070  + 1577.58735993130 *T)
    sumY0  +=      10.763 * cos(5.94541313957  + 9929.67004482930 *T)
    sumY0  +=       8.847 * cos(0.10898852487  + 29051.02756083270 *T)
    sumY0  +=      10.378 * cos(0.40689046689  + 10497.38868255971 *T)
    sumY0  +=       7.498 * cos(4.24537566245  + 40853.38600232750 *T)
    sumY0  +=       7.375 * cos(5.51259327017  + 16496.60521368590 *T)
    sumY0  +=       8.792 * cos(5.18118052745  + 10208.00643938710 *T)
    sumY0  +=       8.786 * cos(1.17109977444  + 10219.05228800190 *T)
    sumY0  +=       7.423 * cos(2.59067948380  + 10175.50169105870 *T)
    sumY0  +=       7.386 * cos(3.76162100212  + 10251.55703633031 *T)
    sumY0  +=       6.692 * cos(5.95406544386  + 8094.27786834910 *T)
    sumY0  +=       6.531 * cos(5.88279509444  + 20419.70136290470 *T)
    sumY0  +=       5.876 * cos(0.45809002649  + 32217.44399856430 *T)
    sumY0  +=       7.945 * cos(1.89399424197  + 12566.39551746630 *T)
    sumY0  +=       5.723 * cos(0.73880953765  + 24357.02460612510 *T)
    sumY0  +=       5.960 * cos(5.04842520573  + 6681.46867088311 *T)
    sumY0  +=       5.974 * cos(0.59430542004  + 426.84200835950 *T)
    sumY0  +=       5.761 * cos(2.33144379428  + 2352.62233628830 *T)
    sumY0  +=       7.279 * cos(5.06948362320  + 11014.86265985130 *T)
    sumY0  +=       4.942 * cos(2.08880442511  + 3149.40797807170 *T)
    sumY0  +=       5.638 * cos(3.95061708151  + 382.65271473970 *T)
    sumY0  +=       6.069 * cos(4.02647068871  + 9683.35076363289 *T)
    sumY0  +=       4.581 * cos(1.06268488434  + 14765.48286075331 *T)
    sumY0  +=       6.372 * cos(1.77392148125  + 25158.35790228190 *T)
    sumY0  +=       4.902 * cos(5.88378200448  + 18837.74201462170 *T)
    sumY0  +=       4.698 * cos(4.98709845390  + 18849.47136745770 *T)
    sumY0  +=       5.371 * cos(2.87505683972  + 26088.14695905770 *T)
    sumY0  +=       5.495 * cos(1.72171503357  + 17086.20248320570 *T)
    sumY0  +=       5.381 * cos(3.84016723700  + 33794.78754101210 *T)
    sumY0  +=       4.870 * cos(1.21408319983  + 20453.11322970530 *T)
    sumY0  +=       3.922 * cos(1.30852789776  + 13745.59005650590 *T)
    sumY0  +=       4.608 * cos(5.91400382368  + 5661.08823166870 *T)
    sumY0  +=       3.970 * cos(4.65819000475  + 4731.78680985990 *T)
    sumY0  +=       3.556 * cos(5.06044480376  + 1551.28904013150 *T)
    sumY0  +=       4.329 * cos(1.83456881254  + 23958.87560271690 *T)
    sumY0  +=       3.274 * cos(5.00533011391  + 529.44714761109 *T)
    sumY0  +=       3.328 * cos(2.61855780370  + 18307.56341456009 *T)
    sumY0  +=       2.787 * cos(3.62713580728  + 11790.38527117530 *T)
    sumY0  +=       3.603 * cos(0.20878953983  + 14945.07235607090 *T)
    sumY0  +=       2.949 * cos(1.25283496540  + 10198.27689250950 *T)
    sumY0  +=       2.944 * cos(5.09946216488  + 10228.78183487951 *T)
    sumY0  +=       3.221 * cos(5.14877320464  + 20400.51659010570 *T)
    sumY0  +=       3.653 * cos(3.19373783247  + 29088.56759850150 *T)
    sumY0  +=       2.683 * cos(3.77978842246  + 6282.83203250789 *T)
    sumY0  +=       2.848 * cos(2.09414295555  + 11322.42028082089 *T)
    sumY0  +=       2.745 * cos(0.21041860525  + 5481.49873635110 *T)
    sumY0  +=       2.426 * cos(1.24386679781  + 6872.91693699471 *T)
    sumY0  +=       2.262 * cos(5.05423750942  + 7860.17557495569 *T)
    sumY0  +=       2.431 * cos(2.28515436796  + 30110.40949102190 *T)
    sumY0  +=       2.218 * cos(4.26312729072  + 17277.65074931730 *T)
    sumY0  +=       2.955 * cos(3.73004048151  + 27511.71169102070 *T)
    sumY0  +=       2.300 * cos(2.32608220761  + 3442.81876244891 *T)
    sumY0  +=       2.424 * cos(5.68140966349  + 20956.50587500010 *T)
    sumY0  +=       2.622 * cos(2.17435265531  + 9580.74562438130 *T)
    sumY0  +=       2.032 * cos(5.76558577000  + 20000.21671902950 *T)
    sumY0  +=       2.206 * cos(2.03456492963  + 537.04832957890 *T)
    sumY0  +=       2.010 * cos(5.28017944038  + 21202.33752122950 *T)
    sumY0  +=       2.471 * cos(4.17654037966  + 10846.31310300770 *T)
    sumY0  +=       2.315 * cos(0.62235018491  + 9573.14444241349 *T)
    sumY0  +=       2.188 * cos(4.41129286184  + 9967.69771746511 *T)
    sumY0  +=       2.278 * cos(2.82110310010  + 10110.43658947590 *T)
    sumY0  +=       2.278 * cos(3.53119029368  + 10316.62213791310 *T)
    sumY0  +=       1.808 * cos(2.43291137306  + 22779.68106367730 *T)
    sumY0  +=       2.244 * cos(0.87218452760  + 5507.79705615090 *T)
    sumY0  +=       1.796 * cos(4.66992847541  + 7063.87756813929 *T)
    sumY0  +=       2.118 * cos(1.94080318718  + 10459.36100992390 *T)
    sumY0  +=       1.908 * cos(4.04072387238  + 20809.71144212870 *T)
    sumY0  +=       2.093 * cos(5.40951959375  + 522.82123557730 *T)
    sumY0  +=       1.753 * cos(5.11036210545  + 13554.14179039430 *T)
    sumY0  +=       2.072 * cos(2.68969610898  + 9690.95194560070 *T)
    sumY0  +=       1.993 * cos(5.12289612530  + 10404.48999483909 *T)
    sumY0  +=       2.160 * cos(4.61577108214  + 33018.77729472110 *T)
    sumY0  +=       1.962 * cos(3.66408601951  + 10736.10678178830 *T)
    sumY0  +=       1.484 * cos(4.31259578278  + 3723.26514143949 *T)
    sumY0  +=       1.997 * cos(1.63225621325  + 18875.28205229050 *T)
    sumY0  +=       2.057 * cos(3.55699495023  + 102.84895673509 *T)
    sumY0  +=       2.016 * cos(5.26645554222  + 37724.99723723170 *T)
    sumY0  +=       1.618 * cos(3.30590238659  + 7085.14059859870 *T)
    sumY0  +=       1.659 * cos(0.00230057752  + 20235.36664379391 *T)
    sumY0  +=       1.553 * cos(6.13988283072  + 14945.55999103790 *T)
    sumY0  +=       1.873 * cos(6.21493810839  + 30213.50226524050 *T)
    sumY0  +=       1.542 * cos(2.45160700066  + 6490.02040477150 *T)
    sumY0  +=       1.360 * cos(4.22919267004  + 10233.19912459430 *T)
    sumY0  +=       1.360 * cos(2.12310072374  + 10193.85960279471 *T)
    sumY0  +=       1.592 * cos(5.43357759705  + 10138.74776512720 *T)
    sumY0  +=       1.592 * cos(0.91871579673  + 10288.31096226180 *T)
    sumY0  +=       1.368 * cos(1.28838366385  + 18875.76968725751 *T)
    sumY0  +=       1.235 * cos(5.36285347606  + 20.53157800890 *T)
    sumY0  +=       1.485 * cos(0.27770606219  + 9160.77334553909 *T)
    sumY0  +=       1.516 * cos(3.63976169766  + 9146.54625153749 *T)
    sumY0  +=       1.233 * cos(1.75819099072  + 18073.46112116669 *T)
    sumY0  +=       1.560 * cos(5.07643229822  + 9897.13749403790 *T)
    sumY0  +=       1.560 * cos(1.27586109556  + 10529.92123335110 *T)
    sumY0  +=       1.179 * cos(4.44352915820  + 14919.26167123810 *T)
    sumY0  +=       1.213 * cos(6.01696685843  + 28521.33659573810 *T)
    sumY0  +=       1.182 * cos(4.06358300800  + 9794.04471981930 *T)
    sumY0  +=       1.402 * cos(6.10447020216  + 8661.99650607950 *T)
    sumY0  +=       1.158 * cos(2.28822825043  + 10633.01400756970 *T)
    sumY0  +=       1.257 * cos(5.80717617406  + 2378.92065608810 *T)
    sumY0  +=       1.089 * cos(3.83624431348  + 30639.61282114949 *T)
    sumY0  +=       1.402 * cos(3.43315827736  + 10640.12755457050 *T)
    sumY0  +=       1.483 * cos(0.85707928326  + 21535.70582703189 *T)
    sumY0  +=       1.079 * cos(5.95495431418  + 775.27879384050 *T)
    sumY0  +=       1.370 * cos(3.04127991943  + 9786.93117281850 *T)
    sumY0  +=       1.064 * cos(4.03087864361  + 16984.23996494010 *T)
    sumY0  +=       1.158 * cos(5.90602944547  + 3929.96587873609 *T)
    sumY0  +=       0.967 * cos(0.02581826777  + 640.14110379750 *T)
    sumY0  +=       0.995 * cos(5.83778649915  + 8662.48414104651 *T)
    sumY0  +=       1.305 * cos(6.03771868030  + 36948.98699094070 *T)
    sumY0  +=       1.191 * cos(2.95662506124  + 39264.31310704370 *T)
    sumY0  +=       1.010 * cos(5.80804230901  + 10007.34381525730 *T)
    sumY0  +=       1.010 * cos(0.54425107612  + 10419.71491213170 *T)
    sumY0  +=       0.950 * cos(2.79830239920  + 1109.62236957690 *T)
    sumY0  +=       1.221 * cos(3.93109520572  + 75.02541605080 *T)
    sumY0  +=       0.987 * cos(3.89911553784  + 13937.03832261750 *T)
    sumY0  +=       1.164 * cos(3.04468597570  + 22805.49174851010 *T)
    sumY0  +=       0.845 * cos(5.84334943435  + 6770.46678376209 *T)
    sumY0  +=       0.996 * cos(2.96758372829  + 13341.91812879030 *T)
    sumY0  +=       1.123 * cos(0.26297644138  + 246.07546371290 *T)
    sumY0  +=       1.139 * cos(0.86118865103  + 2119.00767786191 *T)
    sumY0  +=       1.046 * cos(5.02841709583  + 31441.92138724031 *T)
    sumY0  +=       0.879 * cos(5.42815094849  + 15720.59496739490 *T)
    sumY0  +=       0.761 * cos(5.58947417459  + 7564.58690325449 *T)
    sumY0  +=       0.814 * cos(4.49927037870  + 9676.72485159910 *T)
    sumY0  +=       0.786 * cos(0.51236116044  + 11764.57458634250 *T)
    sumY0  +=       0.883 * cos(4.72642948961  + 25158.84553724890 *T)
    sumY0  +=       0.795 * cos(1.85352659217  + 10750.33387578991 *T)
    sumY0  +=       0.770 * cos(5.39717699890  + 7834.36489012290 *T)
    sumY0  +=       0.825 * cos(0.49699119244  + 191.69208359510 *T)
    sumY0  +=       0.696 * cos(1.82641391127  + 36147.65369478390 *T)
    sumY0  +=       0.941 * cos(5.56301751479  + 28520.84896077109 *T)
    sumY0  +=       0.787 * cos(4.31922851608  + 19890.01039781010 *T)
    sumY0  +=       0.788 * cos(4.89324353272  + 7994.77225950771 *T)
    sumY0  +=       0.624 * cos(1.27682792425  + 39794.00407213830 *T)
    sumY0  +=       0.609 * cos(0.58636941909  + 8624.45646841071 *T)
    sumY0  +=       0.836 * cos(0.40176116197  + 41655.20693345130 *T)
    sumY0  +=       0.744 * cos(0.94447022556  + 19904.23749181170 *T)
    sumY0  +=       0.689 * cos(3.03363296847  + 1066.73929467350 *T)
    sumY0  +=       0.803 * cos(1.17639368940  + 40879.19668716030 *T)
    sumY0  +=       0.692 * cos(5.25875345423  + 6770.95441872910 *T)
    sumY0  +=       0.585 * cos(0.31354909520  + 14143.25142494710 *T)
    sumY0  +=       0.678 * cos(1.25286840883  + 12592.20620229910 *T)
    sumY0  +=       0.573 * cos(0.95211843581  + 12592.69383726610 *T)
    sumY0  +=       0.670 * cos(2.49987738716  + 10239.34004852730 *T)
    sumY0  +=       0.655 * cos(0.26769747072  + 19366.94534474929 *T)
    sumY0  +=       0.626 * cos(5.88947359425  + 52176.05010063190 *T)
    sumY0  +=       0.548 * cos(2.24703921308  + 17777.87244946549 *T)
    sumY0  +=       0.709 * cos(4.47405816630  + 26735.70144472970 *T)
    sumY0  +=       0.531 * cos(1.52107728998  + 110.45013870291 *T)
    sumY0  +=       0.605 * cos(3.06366115762  + 10059.75055320970 *T)
    sumY0  +=       0.676 * cos(3.63142919749  + 20529.90768412410 *T)
    sumY0  +=       0.598 * cos(1.46005524042  + 12432.28646788130 *T)
    sumY0  +=       0.593 * cos(3.28870937227  + 10367.30817417930 *T)
    sumY0  +=       0.663 * cos(4.76628280750  + 1162.71852189130 *T)
    sumY0  +=       0.466 * cos(0.80841803105  + 8631.08238044450 *T)
    sumY0  +=       0.468 * cos(4.59842473713  + 16983.75232997309 *T)
    sumY0  +=       0.454 * cos(5.50480764823  + 10042.85649307530 *T)
    sumY0  +=       0.462 * cos(1.11716283210  + 6.86972951729 *T)
    sumY0  +=       0.444 * cos(0.84748580654  + 10384.20223431370 *T)
    sumY0  +=       0.460 * cos(5.40321784334  + 11802.60225897830 *T)
    sumY0  +=       0.436 * cos(0.28613833283  + 9639.18456535971 *T)
    sumY0  +=       0.565 * cos(0.99439008608  + 8520.87605922509 *T)
    sumY0  +=       0.416 * cos(0.15526058452  + 10225.25871653050 *T)
    sumY0  +=       0.416 * cos(6.19703280926  + 10201.80001085851 *T)
    sumY0  +=       0.414 * cos(4.74636525670  + 10186.74340892770 *T)
    sumY0  +=       0.431 * cos(3.30159385303  + 8616.85528644289 *T)
    sumY0  +=       0.393 * cos(0.85768879239  + 3532.30451029490 *T)
    sumY0  +=       0.406 * cos(6.06603671083  + 10787.87416202930 *T)
    sumY0  +=       0.380 * cos(4.93978031772  + 19317.43635781211 *T)
    sumY0  +=       0.458 * cos(3.28496308708  + 2218.51328670329 *T)
    sumY0  +=       0.433 * cos(0.59107482876  + 19650.80466361450 *T)
    sumY0  +=       0.377 * cos(4.10235161087  + 149.80701461810 *T)
    sumY0  +=       0.390 * cos(1.63992074391  + 13936.55068765050 *T)
    sumY0  +=       0.360 * cos(4.38429398498  + 24150.32386882850 *T)
    sumY0  +=       0.359 * cos(2.69008664310  + 30831.54872222810 *T)
    sumY0  +=       0.501 * cos(2.59817403162  + 44809.40638337990 *T)
    sumY0  +=       0.409 * cos(3.34267900163  + 27197.03787618409 *T)
    sumY0  +=       0.354 * cos(1.27226804531  + 10323.73568491390 *T)
    sumY0  +=       0.354 * cos(5.08002534847  + 10103.32304247510 *T)
    sumY0  +=       0.428 * cos(1.10028111046  + 13656.10430865991 *T)
    sumY0  +=       0.467 * cos(3.73227844246  + 38.37685312130 *T)
    sumY0  +=       0.343 * cos(0.04251143462  + 19786.91762359150 *T)
    sumY0  +=       0.457 * cos(5.50264719385  + 18308.05104952710 *T)
    sumY0  +=       0.399 * cos(5.69770325248  + 1052.51220067191 *T)
    sumY0  +=       0.445 * cos(5.89891356885  + 30665.91114094930 *T)
    sumY0  +=       0.344 * cos(4.75093105230  + 27490.93629552830 *T)
    sumY0  +=       0.315 * cos(4.73109393582  + 20640.11400534350 *T)
    sumY0  +=       0.399 * cos(2.67833209975  + 10212.04489098620 *T)
    sumY0  +=       0.399 * cos(3.67396129403  + 10215.01383640280 *T)
    sumY0  +=       0.390 * cos(0.71131374079  + 30000.20316980250 *T)
    sumY0  +=       0.340 * cos(5.88673417604  + 22805.97938347710 *T)
    sumY0  +=       0.398 * cos(2.07757082821  + 9836.15575577870 *T)
    sumY0  +=       0.311 * cos(0.04968618848  + 22003.67081738630 *T)
    sumY0  +=       0.404 * cos(2.41513448299  + 20352.03331133820 *T)
    sumY0  +=       0.366 * cos(2.69947342940  + 16522.41589851870 *T)
    sumY0  +=       0.350 * cos(4.56156653428  + 15874.37377787969 *T)
    sumY0  +=       0.385 * cos(6.08106804956  + 20180.98326367610 *T)
    sumY0  +=       0.379 * cos(4.27472269461  + 10590.90297161030 *T)
    sumY0  +=       0.395 * cos(1.98398365933  + 17277.16311435029 *T)
    sumY0  +=       0.383 * cos(0.20953871800  + 35372.13108345990 *T)
    sumY0  +=       0.277 * cos(4.73141990633  + 10988.56434005150 *T)
    sumY0  +=       0.365 * cos(6.13706743141  + 10349.15468870450 *T)
    sumY0  +=       0.365 * cos(0.21522596237  + 10077.90403868451 *T)
    sumY0  +=       0.336 * cos(5.45252157770  + 220.65645992230 *T)
    sumY0  +=       0.313 * cos(0.95373705811  + 6309.13035230770 *T)
    sumY0  +=       0.276 * cos(4.02007434964  + 26709.89075989690 *T)
    sumY0  +=       0.283 * cos(3.96353025226  + 3904.15519390330 *T)
    sumY0  +=       0.271 * cos(2.28569018986  + 9745.56437334010 *T)
    sumY0  +=       0.321 * cos(2.63356462069  + 38734.62214194910 *T)
    sumY0  +=       0.327 * cos(1.40582630148  + 9359.84534697549 *T)
    sumY0  +=       0.264 * cos(3.59015983869  + 76.50988875911 *T)
    sumY0  +=       0.310 * cos(5.18631223131  + 27991.15799567649 *T)
    sumY0  +=       0.359 * cos(1.81226579939  + 45585.41662967090 *T)
    sumY0  +=       0.326 * cos(4.02012440588  + 633.02755679670 *T)
    sumY0  +=       0.314 * cos(6.11319497702  + 20406.03951441310 *T)
    sumY0  +=       0.306 * cos(5.83848348668  + 2957.95971196010 *T)
    sumY0  +=       0.266 * cos(5.66163509794  + 8635.69818627969 *T)
    sumY0  +=       0.254 * cos(4.06665609388  + 10681.49435404891 *T)
    sumY0  +=       0.343 * cos(3.07216363987  + 8106.49485615210 *T)
    sumY0  +=       0.300 * cos(2.55692274187  + 206.42936592071 *T)
    sumY0  +=       0.250 * cos(4.61153598536  + 283.61550138170 *T)
    sumY0  +=       0.336 * cos(2.25621229669  + 23581.01435983410 *T)
    sumY0  +=       0.236 * cos(1.26632265704  + 10595.93826095070 *T)
    sumY0  +=       0.234 * cos(2.24935612012  + 19624.99397878170 *T)
    sumY0  +=       0.255 * cos(4.35467668425  + 213.05527795450 *T)
    sumY0  +=       0.281 * cos(5.49571093835  + 9399.97907973470 *T)
    sumY0  +=       0.314 * cos(2.13931899616  + 19786.42998862449 *T)
    sumY0  +=       0.229 * cos(0.20297430947  + 574.10098085129 *T)
    sumY0  +=       0.295 * cos(2.36600819373  + 31749.47900820990 *T)
    sumY0  +=       0.316 * cos(4.02123996191  + 48739.61607959950 *T)
    sumY0  +=       0.229 * cos(2.88588042648  + 949.41942645330 *T)
    sumY0  +=       0.229 * cos(1.53780233055  + 19896.63630984389 *T)
    sumY0  +=       0.257 * cos(3.32400213893  + 19360.31943271550 *T)
    sumY0  +=       0.231 * cos(2.81577137220  + 9264.35375472471 *T)
    sumY0  +=       0.255 * cos(2.46385009073  + 9993.11672125570 *T)
    sumY0  +=       0.255 * cos(3.88844330305  + 10433.94200613330 *T)
    sumY0  +=       0.291 * cos(2.32110551482  + 9999.74263328950 *T)
    sumY0  +=       0.285 * cos(1.03940149360  + 34596.12083716890 *T)
    sumY0  +=       0.211 * cos(4.89515362389  + 10249.55723037191 *T)
    sumY0  +=       0.211 * cos(1.45713976989  + 10177.50149701710 *T)
    sumY0  +=       0.255 * cos(0.85666215039  + 11027.07964765431 *T)
    sumY0  +=       0.220 * cos(2.13302811409  + 37.78385515230 *T)
    sumY0  +=       0.280 * cos(2.09663247440  + 37410.32342239509 *T)
    sumY0  +=       0.258 * cos(3.69928540375  + 29573.60497892650 *T)
    sumY0  +=       0.247 * cos(0.33730425769  + 29587.83207292810 *T)
    sumY0  +=       0.206 * cos(1.80393013250  + 30426.80136067850 *T)
    sumY0  +=       0.263 * cos(1.59050643196  + 20433.92845690630 *T)
    sumY0  +=       0.212 * cos(1.58965720399  + 10206.90080479450 *T)
    sumY0  +=       0.212 * cos(4.76263618979  + 10220.15792259451 *T)
    sumY0  +=       0.202 * cos(5.53871102932  + 16894.75421709410 *T)
    sumY0  +=       0.272 * cos(3.28012716441  + 12320.56387123691 *T)
    sumY0  +=       0.211 * cos(3.54449702303  + 11162.70497266430 *T)
    sumY0  +=       0.197 * cos(2.85538775435  + 8734.17515466309 *T)
    sumY0  +=       0.267 * cos(2.72934257174  + 377.61742539930 *T)
    sumY0  +=       0.224 * cos(1.78242794958  + 9051.05465928671 *T)
    sumY0  +=       0.186 * cos(2.96624941532  + 10251.66239933230 *T)
    sumY0  +=       0.186 * cos(3.38604397846  + 10175.39632805670 *T)
    sumY0  +=       0.258 * cos(0.95259841698  + 9864.60494324651 *T)
    sumY0  +=       0.186 * cos(3.20827310527  + 40077.86339100350 *T)
    sumY0  +=       0.233 * cos(6.20481595696  + 29089.05523346850 *T)
    sumY0  +=       0.186 * cos(3.29904519478  + 17272.12782500990 *T)
    sumY0  +=       0.254 * cos(1.58851500083  + 19264.34020549770 *T)
    sumY0  +=       0.246 * cos(5.39979351075  + 10562.45378414250 *T)
    sumY0  +=       0.211 * cos(4.59175138406  + 11376.00406810230 *T)
    sumY0  +=       0.178 * cos(5.22149503543  + 10949.40587722630 *T)
    sumY0  +=       0.178 * cos(1.13079835835  + 9477.65285016271 *T)
    sumY0  +=       0.197 * cos(5.67208275737  + 1692.40948698591 *T)
    sumY0  +=       0.177 * cos(4.83529684993  + 20316.60858868610 *T)
    sumY0  +=       0.216 * cos(4.13948146239  + 20452.62559473830 *T)
    sumY0  +=       0.197 * cos(1.76173397564  + 9050.56702431969 *T)
    sumY0  +=       0.159 * cos(5.88022204093  + 27197.52551115110 *T)
    sumY0  +=       0.167 * cos(1.09928566241  + 4705.48849006010 *T)
    sumY0  +=       0.213 * cos(3.70468670755  + 27511.22405605370 *T)
    sumY0  +=       0.154 * cos(5.43597579906  + 10199.30226969290 *T)
    sumY0  +=       0.154 * cos(0.91631759472  + 10227.75645769610 *T)
    sumY0  +=       0.174 * cos(0.32803080161  + 14.47091148511 *T)
    sumY0  +=       0.209 * cos(3.23046462275  + 10063.47853159290 *T)
    sumY0  +=       0.167 * cos(5.14638326738  + 6872.42930202769 *T)
    sumY0  +=       0.185 * cos(2.25428901814  + 9256.75257275689 *T)
    sumY0  +=       0.166 * cos(2.18082507791  + 10263.28638916630 *T)
    sumY0  +=       0.166 * cos(4.17146831587  + 10163.77233822270 *T)
    sumY0  +=       0.168 * cos(0.62423658392  + 7991.18509413049 *T)
    sumY0  +=       0.174 * cos(2.19758501900  + 9161.26098050610 *T)
    sumY0  +=       0.197 * cos(1.87142263401  + 846.32665223470 *T)
    sumY0  +=       0.176 * cos(6.26295125412  + 9779.32999085069 *T)
    sumY0  +=       0.198 * cos(5.45633204029  + 52669.82577581910 *T)
    sumY0  +=       0.169 * cos(4.14838568997  + 11265.79774688290 *T)
    sumY0  +=       0.188 * cos(2.47639831966  + 38526.33053338850 *T)
    sumY0  +=       0.140 * cos(5.67905778675  + 27991.64563064350 *T)
    sumY0  +=       0.194 * cos(3.83921848765  + 8837.26792888169 *T)
    sumY0  +=       0.173 * cos(2.87742322736  + 9793.55708485230 *T)
    sumY0  +=       0.140 * cos(5.04268037408  + 10191.67853436810 *T)
    sumY0  +=       0.140 * cos(1.30961301970  + 10235.38019302090 *T)
    sumY0  +=       0.196 * cos(3.84465359273  + 1293.24040609909 *T)
    sumY0  +=       0.180 * cos(5.95332595720  + 10021.59346261590 *T)
    sumY0  +=       0.138 * cos(5.09734035169  + 10217.46151695760 *T)
    sumY0  +=       0.138 * cos(1.25495304209  + 10209.59721043140 *T)
    sumY0  +=       0.148 * cos(0.74425859729  + 19374.54652671710 *T)
    sumY0  +=       0.137 * cos(2.37981976955  + 16522.90353348570 *T)
    sumY0  +=       0.143 * cos(6.11571774821  + 20447.59030539790 *T)
    sumY0  +=       0.139 * cos(1.01385118035  + 7255.32583425090 *T)
    sumY0  +=       0.126 * cos(2.24994178665  + 20277.25171277090 *T)
    sumY0  +=       0.164 * cos(3.22032043024  + 49515.62632589050 *T)
    sumY0  +=       0.149 * cos(0.61484998961  + 10135.85559326650 *T)
    sumY0  +=       0.149 * cos(5.73744340417  + 10291.20313412251 *T)
    sumY0  +=       0.118 * cos(2.69846631523  + 4551.70967957530 *T)
    sumY0  +=       0.117 * cos(5.64919068623  + 1582.20316576650 *T)
    sumY0  +=       0.151 * cos(5.55480633210  + 10246.06191448590 *T)
    sumY0  +=       0.151 * cos(0.79748706168  + 10180.99681290311 *T)
    sumY0  +=       0.152 * cos(1.61196584184  + 39302.34077967950 *T)
    sumY0  +=       0.144 * cos(2.25615309488  + 9986.05323090550 *T)
    sumY0  +=       0.144 * cos(4.09614029890  + 10441.00549648351 *T)
    sumY0  +=       0.136 * cos(0.51140220699  + 17469.09901542890 *T)
    sumY0  +=       0.119 * cos(4.90022973272  + 20809.22380716170 *T)
    sumY0  +=       0.130 * cos(4.00092308815  + 10236.30456514530 *T)
    sumY0  +=       0.130 * cos(2.35137030563  + 10190.75416224370 *T)
    sumY0  +=       0.156 * cos(2.62001496107  + 20388.68187426770 *T)
    sumY0  +=       0.132 * cos(3.92998867218  + 9580.25798941430 *T)
    sumY0  +=       0.155 * cos(5.57179178709  + 8238.03681783850 *T)
    sumY0  +=       0.142 * cos(5.15510279625  + 31441.43375227330 *T)
    sumY0  +=       0.131 * cos(0.68390506026  + 8101.39141534989 *T)
    sumY0  +=       0.133 * cos(5.57627209730  + 24382.83529095790 *T)
    sumY0  +=       0.105 * cos(5.80156791329  + 32858.36992533629 *T)
    sumY0  +=       0.105 * cos(5.39190347848  + 1059.13811270570 *T)
    sumY0  +=       0.101 * cos(2.78278015558  + 24978.76840696430 *T)
    sumY0  +=       0.126 * cos(3.33415820026  + 22645.57201409229 *T)
    sumY0  +=       0.101 * cos(3.69358080572  + 20043.91837768230 *T)
    sumY0  +=       0.110 * cos(4.54705018597  + 43071.65547154729 *T)
    sumY0  +=       0.103 * cos(2.03520456299  + 12431.79883291429 *T)
    sumY0  +=       0.136 * cos(4.50608056956  + 5999.46034860970 *T)
    sumY0  +=       0.097 * cos(5.63294739362  + 853.44019923550 *T)
    sumY0  +=       0.130 * cos(1.64701005433  + 4176.28515993250 *T)
    sumY0  +=       0.113 * cos(2.36571271909  + 19794.03117059230 *T)
    sumY0  +=       0.093 * cos(0.77688953992  + 22645.08437912529 *T)
    sumY0  +=       0.124 * cos(3.90067857109  + 42456.54022960810 *T)
    sumY0  +=       0.107 * cos(5.87195741297  + 3.68833577960 *T)
    sumY0  +=       0.127 * cos(1.97874545982  + 284.10313634870 *T)
    sumY0  +=       0.112 * cos(0.89977181729  + 20206.40226746670 *T)
    sumY0  +=       0.097 * cos(2.64537814089  + 78263.95324220609 *T)
    sumY0  +=       0.093 * cos(1.80429347640  + 8204.48418956850 *T)
    sumY0  +=       0.127 * cos(0.59683711488  + 56600.03547203870 *T)
    sumY0  +=       0.111 * cos(6.25613029204  + 18830.14083265389 *T)
    sumY0  +=       0.098 * cos(1.55452229557  + 11059.61219844570 *T)
    sumY0  +=       0.098 * cos(4.79806616663  + 9367.44652894330 *T)
    sumY0  +=       0.103 * cos(3.58160703323  + 10244.44348932950 *T)
    sumY0  +=       0.103 * cos(2.77068636055  + 10182.61523805950 *T)
    sumY0  +=       0.111 * cos(2.87167506517  + 18844.36792665549 *T)
    sumY0  +=       0.125 * cos(0.78050482769  + 12189.02190955050 *T)
    sumY0  +=       0.088 * cos(2.76213355426  + 20350.54883862990 *T)
    sumY0  +=       0.088 * cos(0.86523706220  + 29058.14110783350 *T)
    sumY0  +=       0.096 * cos(2.34258805293  + 2333.44021035551 *T)
    sumY0  +=       0.122 * cos(3.86127135190  + 2636.96929012050 *T)
    sumY0  +=       0.087 * cos(2.93183519168  + 8087.16432134830 *T)
    sumY0  +=       0.086 * cos(5.21875976302  + 7034.89593815990 *T)
    sumY0  +=       0.095 * cos(0.46006492106  + 10202.48366343060 *T)
    sumY0  +=       0.095 * cos(5.89222847272  + 10224.57506395840 *T)
    sumY0  +=       0.084 * cos(5.94038512576  + 10401.45451145710 *T)
    sumY0  +=       0.084 * cos(0.41190826802  + 10025.60421593190 *T)
    sumY0  +=       0.085 * cos(3.46647958947  + 19477.63930093570 *T)
    sumY0  +=       0.086 * cos(6.05989445995  + 29864.09020982550 *T)
    sumY0  +=       0.099 * cos(1.23143719760  + 2648.69864295650 *T)
    sumY0  +=       0.082 * cos(5.76212222967  + 802.06474860730 *T)
    sumY0  +=       0.087 * cos(2.17814642225  + 13745.10242153889 *T)
    sumY0  +=       0.081 * cos(2.16717230933  + 1550.80140516450 *T)
    sumY0  +=       0.083 * cos(4.61159603950  + 25933.88051360590 *T)
    sumY0  +=       0.079 * cos(4.57245769213  + 44008.07308722310 *T)
    sumY0  +=       0.078 * cos(4.54232413366  + 9573.63207738050 *T)
    sumY0  +=       0.098 * cos(3.80111874017  + 20220.62936146830 *T)
    sumY0  +=       0.108 * cos(5.23213365332  + 24149.83623386150 *T)
    sumY0  +=       0.100 * cos(3.49984914654  + 1596.43025976811 *T)
    sumY0  +=       0.095 * cos(1.78056312476  + 29477.62575170870 *T)
    sumY0  +=       0.078 * cos(4.60961320788  + 205.94173095370 *T)
    sumY0  +=       0.096 * cos(0.05074793120  + 30206.38871823969 *T)
    sumY0  +=       0.079 * cos(5.65022966428  + 3340.36860921629 *T)
    sumY0  +=       0.095 * cos(2.94692116655  + 30220.61581224130 *T)
    sumY0  +=       0.082 * cos(3.38904287688  + 29043.91401383189 *T)
    sumY0  +=       0.077 * cos(4.54577063518  + 17468.61138046190 *T)
    sumY0  +=       0.100 * cos(2.86693803454  + 38500.51984855570 *T)
    sumY0  +=       0.107 * cos(3.08328160895  + 2069.25065239010 *T)
    sumY0  +=       0.077 * cos(2.01507273057  + 10853.42665000850 *T)
    sumY0  +=       0.087 * cos(5.94217613191  + 5.27910682390 *T)
    sumY0  +=       0.076 * cos(3.30386469943  + 10243.35080184331 *T)
    sumY0  +=       0.076 * cos(3.04842869435  + 10183.70792554570 *T)
    sumY0  +=       0.076 * cos(4.51338161700  + 956.53297345411 *T)
    sumY0  +=       0.100 * cos(1.06850076905  + 28947.93478661410 *T)
    sumY0  +=       0.096 * cos(0.14778409211  + 35371.64344849290 *T)
    sumY0  +=       0.073 * cos(1.89720121846  + 17248.18148437089 *T)
    sumY0  +=       0.073 * cos(0.27810624949  + 21001.15970824030 *T)
    sumY0  +=       0.081 * cos(0.69044779243  + 18734.64924040310 *T)
    sumY0  +=       0.076 * cos(3.94914583018  + 103.33659170210 *T)
    sumY0  +=       0.085 * cos(0.56437212275  + 3914.71340755109 *T)
    sumY0  +=       0.074 * cos(0.35341059185  + 8827.63408735830 *T)
    sumY0  +=       0.070 * cos(6.20394049400  + 6298.57213865991 *T)
    sumY0  +=       0.087 * cos(1.33961295133  + 33019.26492968810 *T)
    sumY0  +=       0.071 * cos(2.38829288593  + 34172.16114892790 *T)
    sumY0  +=       0.079 * cos(1.08839441675  + 10018.55797923390 *T)
    sumY0  +=       0.079 * cos(5.26389897703  + 10408.50074815510 *T)
    sumY0  +=       0.069 * cos(1.34393957743  + 33794.29990604510 *T)
    sumY0  +=       0.087 * cos(3.11833451886  + 29050.53992586569 *T)
    sumY0  +=       0.069 * cos(5.61698463903  + 41044.83426843910 *T)
    sumY0  +=       0.069 * cos(0.68347421356  + 9644.70748966710 *T)
    sumY0  +=       0.069 * cos(5.66881918022  + 10782.35123772190 *T)
    sumY0  +=       0.087 * cos(0.75446196555  + 28313.04498717750 *T)
    sumY0  +=       0.085 * cos(5.34847160578  + 46386.74992582770 *T)
    sumY0  +=       0.083 * cos(4.22258770909  + 9721.86607123571 *T)
    sumY0  +=       0.083 * cos(2.12970568469  + 10705.19265615330 *T)
    sumY0  +=       0.090 * cos(3.61691095667  + 20049.44130198970 *T)
    sumY0  +=       0.064 * cos(4.29241354560  + 9147.03388650451 *T)
    sumY0  +=       0.076 * cos(5.76440554422  + 21486.19684009470 *T)
    sumY0  +=       0.085 * cos(2.91506656251  + 19573.13089318649 *T)
    sumY0  +=       0.063 * cos(3.18394429045  + 18314.67696156089 *T)
    sumY0  +=       0.071 * cos(1.43323486518  + 1573.57660661529 *T)
    sumY0  +=       0.070 * cos(0.75113063455  + 1580.62284331330 *T)
    sumY0  +=       0.080 * cos(1.55502986811  + 39301.85314471250 *T)
    sumY0  +=       0.067 * cos(2.05780765354  + 3956.26419853590 *T)
    sumY0  +=       0.087 * cos(3.90828389301  + 29264.32665627070 *T)
    sumY0  +=       0.063 * cos(4.94664274045  + 27461.46703058189 *T)
    sumY0  +=       0.070 * cos(4.18407491369  + 1479.11039154791 *T)
    sumY0  +=       0.081 * cos(1.62635438826  + 10205.92818172669 *T)
    sumY0  +=       0.080 * cos(1.81036094537  + 38204.44354188749 *T)
    sumY0  +=       0.060 * cos(6.06109901775  + 20421.29198559810 *T)
    sumY0  +=       0.064 * cos(5.87371692529  + 18300.44986755929 *T)
    sumY0  +=       0.076 * cos(2.36906935680  + 29683.81130014589 *T)
    sumY0  +=       0.082 * cos(2.02061615739  + 60530.24516825830 *T)
    sumY0  +=       0.077 * cos(4.62121260267  + 53445.83602211010 *T)
    sumY0  +=       0.062 * cos(5.99888344180  + 11599.42464003070 *T)
    sumY0  +=       0.058 * cos(1.51132330725  + 7576.80389105750 *T)
    sumY0  +=       0.057 * cos(2.10835504837  + 11280.02484088450 *T)
    sumY0  +=       0.059 * cos(5.97279707503  + 10255.08015467930 *T)
    sumY0  +=       0.059 * cos(0.37949631875  + 10171.97857270970 *T)
    sumY0  +=       0.057 * cos(2.49183550585  + 20464.84258254131 *T)
    sumY0  +=       0.079 * cos(3.54945759635  + 10136.77565745190 *T)
    sumY0  +=       0.066 * cos(3.91005081121  + 14169.54974474690 *T)
    sumY0  +=       0.063 * cos(5.25828651967  + 18099.75944096650 *T)
    sumY0  +=       0.060 * cos(1.62777565840  + 29786.90407436450 *T)
    sumY0  +=       0.064 * cos(4.00075551038  + 30007.31671680330 *T)
    sumY0  +=       0.054 * cos(2.10183945809  + 30666.39877591630 *T)
    sumY0  +=       0.058 * cos(1.48653920785  + 75449.75065697990 *T)
    sumY0  +=       0.073 * cos(2.27987689949  + 51449.56808076170 *T)
    sumY0  +=       0.074 * cos(3.91378496364  + 21720.29913348810 *T)
    sumY0  +=       0.063 * cos(1.10572404909  + 26736.18907969670 *T)
    sumY0  +=       0.054 * cos(2.34430307252  + 38204.93117685450 *T)
    sumY0  +=       0.072 * cos(2.90060717525  + 100608.35237674530 *T)
    sumY0  +=       0.053 * cos(2.23725068823  + 419.72846135871 *T)
    sumY0  +=       0.052 * cos(0.52183347499  + 40426.78781145150 *T)
    sumY0  +=       0.052 * cos(1.02206221481  + 29580.23089096029 *T)
    sumY0  +=       0.053 * cos(2.81130934606  + 419.24082639170 *T)
    sumY0  +=       0.058 * cos(6.02426265881  + 20412.58781590390 *T)
    sumY0  +=       0.054 * cos(4.83322053510  + 9366.95889397630 *T)
    sumY0  +=       0.055 * cos(5.16439912672  + 9103.66317663410 *T)
    sumY0  +=       0.050 * cos(0.80815973951  + 34570.31015233610 *T)
    sumY0  +=       0.067 * cos(4.21574061370  + 632.53992182970 *T)
    sumY0  +=       0.050 * cos(2.75084153641  + 37724.50960226470 *T)
    sumY0  +=       0.069 * cos(3.30376299828  + 53284.94101775829 *T)
    sumY0  +=       0.057 * cos(3.46902641065  + 8307.57696378710 *T)
    sumY0  +=       0.050 * cos(6.26940117566  + 8417.78328500650 *T)
    sumY0  +=       0.049 * cos(3.48884514281  + 26087.65932409069 *T)
    sumY0  +=       0.064 * cos(0.45737297425  + 7064.36520310630 *T)
    sumY0  +=       0.049 * cos(0.14582053812  + 14128.48658872910 *T)
    sumY0  +=       0.056 * cos(0.66365909742  + 28418.24382151950 *T)
    sumY0  +=       0.066 * cos(4.48087077585  + 19580.73207515430 *T)
    sumY0  +=       0.047 * cos(4.50934828666  + 10742.73269382209 *T)
    sumY0  +=       0.061 * cos(2.99027576045  + 43232.06284093210 *T)
    sumY0  +=       0.061 * cos(2.99027576045  + 43232.55047589910 *T)
    sumY0  +=       0.047 * cos(3.24772943759  + 38741.73568894990 *T)
    sumY0  +=       0.050 * cos(6.05228449513  + 59161.19323437110 *T)
    sumY0  +=       0.046 * cos(0.66057761892  + 18844.85556162250 *T)
    sumY0  +=       0.054 * cos(5.40667573437  + 48947.90768816010 *T)
    sumY0  +=       0.046 * cos(3.06762447182  + 3154.93090237910 *T)
    sumY0  +=       0.046 * cos(5.78250179315  + 194.72756697710 *T)
    sumY0  +=       0.056 * cos(5.79105836654  + 9936.78359183010 *T)
    sumY0  +=       0.056 * cos(0.56123502724  + 10490.27513555891 *T)
    sumY0  +=       0.054 * cos(2.92585194345  + 9922.55649782850 *T)
    sumY0  +=       0.054 * cos(3.42644145033  + 10504.50222956051 *T)
    sumY0  +=       0.059 * cos(3.58503267144  + 18204.47064034149 *T)
    sumY0  +=       0.048 * cos(5.93973568185  + 38727.50859494830 *T)
    sumY0  +=       0.057 * cos(0.49036750604  + 50316.95962204730 *T)
    sumY0  +=       0.061 * cos(4.41585090977  + 10140.23223783550 *T)
    sumY0  +=       0.061 * cos(1.93644248401  + 10286.82648955350 *T)
    sumY0  +=       0.061 * cos(5.28414092980  + 6567.17898634010 *T)
    sumY0  +=       0.048 * cos(1.78116904393  + 9246.55848625890 *T)
    sumY0  +=       0.060 * cos(4.34124861578  + 3684.13423250670 *T)
    sumY0  +=       0.060 * cos(1.84622131041  + 14427.59837877930 *T)
    sumY0  +=       0.051 * cos(3.86708252316  + 47623.60896860609 *T)
    sumY0  +=       0.051 * cos(0.71903454294  + 9566.03089541269 *T)
    sumY0  +=       0.052 * cos(1.63231256422  + 8727.06160766230 *T)
    sumY0  +=       0.057 * cos(2.19171070766  + 32243.25468339710 *T)
    sumY0  +=       0.048 * cos(0.48328683640  + 433.95555536030 *T)
    sumY0  +=       0.045 * cos(3.95006454295  + 10263.52598559870 *T)
    sumY0  +=       0.045 * cos(2.40222885083  + 10163.53274179031 *T)
    sumY0  +=       0.049 * cos(0.08548300777  + 9794.60496479390 *T)
    sumY0  +=       0.049 * cos(6.26681038600  + 10632.45376259511 *T)
    sumY0  +=       0.058 * cos(1.48111021336  + 1860.95904382949 *T)
    sumY0  +=       0.047 * cos(6.22156178529  + 20142.95559104030 *T)
    sumY0  +=       0.046 * cos(3.81317535217  + 2349.58685290630 *T)
    sumY0  +=       0.044 * cos(0.47724556879  + 30632.98690911569 *T)
    sumY0  +=       0.046 * cos(3.13107041429  + 2356.63308960431 *T)
    sumY0  +=       0.041 * cos(4.84096981140  + 12850.25483633150 *T)
    sumY0  +=       0.041 * cos(0.14576799491  + 8094.76550331610 *T)
    sumY0  +=       0.042 * cos(3.51225043090  + 77802.61681075170 *T)
    sumY0  +=       0.042 * cos(1.24107185058  + 25132.54721744910 *T)
    sumY0  +=       0.043 * cos(3.94226686000  + 19359.83179774849 *T)
    sumY0  +=       0.040 * cos(5.46575824362  + 10192.26633323510 *T)
    sumY0  +=       0.042 * cos(1.71742433091  + 34363.60941503950 *T)
    sumY0  +=       0.039 * cos(5.92169473585  + 109.96250373590 *T)
    sumY0  +=       0.044 * cos(2.89783948131  + 79379.96035319949 *T)
    sumY0  +=       0.043 * cos(5.95822352823  + 38734.13450698209 *T)
    sumY0  +=       0.046 * cos(5.13701973505  + 17778.36008443250 *T)
    sumY0  +=       0.039 * cos(1.30600974103  + 12565.90788249930 *T)
    sumY0  +=       0.041 * cos(1.87488507902  + 27681.89692667289 *T)
    sumY0  +=       0.053 * cos(3.44448984210  + 64460.45486447790 *T)
    sumY0  +=       0.039 * cos(1.87347006255  + 28631.54291695750 *T)
    sumY0  +=       0.052 * cos(3.63883543733  + 38631.52936773050 *T)
    sumY0  +=       0.043 * cos(4.57112425738  + 11180.50024113010 *T)
    sumY0  +=       0.039 * cos(5.73477931992  + 1265.81129610991 *T)
    sumY0  +=       0.039 * cos(5.37998455046  + 7879.84533585549 *T)
    sumY0  +=       0.039 * cos(5.37998455046  + 7880.33297082250 *T)
    sumY0  +=       0.052 * cos(0.42370802477  + 1.72829019180 *T)
    sumY0  +=       0.039 * cos(5.18684094963  + 54981.62877357310 *T)
    sumY0  +=       0.047 * cos(0.18137418769  + 187.68133027909 *T)
    sumY0  +=       0.044 * cos(5.88330364964  + 10168.38814405790 *T)
    sumY0  +=       0.044 * cos(0.46898974414  + 10258.67058333111 *T)
    sumY0  +=       0.048 * cos(3.11428064716  + 10063.96616655990 *T)
    sumY0  +=       0.043 * cos(0.40920515062  + 22029.96913718610 *T)
    sumY0  +=       0.051 * cos(4.32233625991  + 104538.56207296488 *T)
    sumY0  +=       0.051 * cos(4.70514324366  + 16250.77356745650 *T)
    sumY0  +=       0.037 * cos(2.91614118281  + 29161.23388205210 *T)
    sumY0  +=       0.043 * cos(3.35259007747  + 10363.09256082911 *T)
    sumY0  +=       0.038 * cos(2.56509414217  + 47917.50738795030 *T)
    sumY0  +=       0.049 * cos(6.13351350993  + 66037.79840692569 *T)
    sumY0  +=       0.037 * cos(2.84218465987  + 6257.02134767510 *T)
    sumY0  +=       0.036 * cos(1.39948703899  + 8674.21349388250 *T)
    sumY0  +=       0.045 * cos(3.57903060172  + 522.33360061030 *T)
    sumY0  +=       0.049 * cos(0.39199067732  + 491.90710994230 *T)
    sumY0  +=       0.042 * cos(1.38184413529  + 115674.52048208468 *T)
    sumY0  +=       0.040 * cos(1.35149502784  + 10233.81712421990 *T)
    sumY0  +=       0.036 * cos(5.98817728846  + 47938.28278344270 *T)
    sumY0  +=       0.035 * cos(1.18237114499  + 12546.72575656651 *T)
    sumY0  +=       0.048 * cos(1.34137746528  + 90395.06683053430 *T)
    sumY0  +=       0.040 * cos(4.18858105636  + 14127.99895376209 *T)
    sumY0  +=       0.044 * cos(2.73569081007  + 9698.06549260150 *T)
    sumY0  +=       0.044 * cos(3.61660258371  + 10728.99323478751 *T)
    sumY0  +=       0.036 * cos(0.66588479818  + 69942.19741831250 *T)
    sumY0  +=       0.035 * cos(4.16647411013  + 41654.71929848429 *T)
    sumY0  +=       0.034 * cos(2.69446692279  + 55798.70217588190 *T)
    sumY0  +=       0.037 * cos(1.68139810641  + 35192.05395317530 *T)
    sumY0  +=       0.046 * cos(1.33857194026  + 36301.43250526870 *T)
    sumY0  +=       0.047 * cos(6.19310388781  + 86464.85713431469 *T)
    sumY0  +=       0.035 * cos(0.73574096610  + 48732.98752069949 *T)
    sumY0  +=       0.035 * cos(3.22135299939  + 10011.27596852040 *T)
    sumY0  +=       0.035 * cos(3.13094039439  + 10415.78275886861 *T)
    sumY0  +=       0.034 * cos(4.60818788188  + 37410.81105736210 *T)
    sumY0  +=       0.033 * cos(3.80929053875  + 49477.59865325469 *T)
    sumY0  +=       0.042 * cos(2.82864954668  + 18830.62846762090 *T)
    sumY0  +=       0.035 * cos(0.56308161765  + 49515.13869092350 *T)
    sumY0  +=       0.047 * cos(2.22431731848  + 31022.50935337269 *T)
    sumY0  +=       0.034 * cos(3.32920660018  + 100909.28143881488 *T)
    sumY0  +=       0.033 * cos(2.17810098111  + 10117.55013647670 *T)
    sumY0  +=       0.033 * cos(4.17419241267  + 10309.50859091230 *T)
    sumY0  +=       0.034 * cos(0.22454658153  + 7461.49412903589 *T)
    sumY0  +=       0.034 * cos(5.44249373293  + 45584.92899470390 *T)
    sumY0  +=       0.032 * cos(1.67235375803  + 9676.23721663209 *T)
    sumY0  +=       0.039 * cos(1.15429612014  + 316.63568714010 *T)
    sumY0  +=       0.033 * cos(6.02812295236  + 31748.99137324289 *T)
    sumY0  +=       0.031 * cos(1.87708120179  + 30103.29594402109 *T)
    sumY0  +=       0.035 * cos(5.77132581087  + 58946.27306691049 *T)
    sumY0  +=       0.032 * cos(2.19359308367  + 21977.86013255350 *T)
    sumY0  +=       0.036 * cos(5.94106521859  + 20430.74706316860 *T)
    sumY0  +=       0.032 * cos(4.93495818843  + 81732.82650697129 *T)
    sumY0  +=       0.030 * cos(0.70932860715  + 19573.61852815350 *T)
    sumY0  +=       0.031 * cos(0.91919649284  + 11.28951774740 *T)
    sumY0  +=       0.038 * cos(0.23287825442  + 20617.77554105009 *T)
    sumY0  +=       0.033 * cos(4.31035022066  + 83310.17004941910 *T)
    sumY0  +=       0.029 * cos(4.95280669313  + 11752.84523350650 *T)
    sumY0  +=       0.029 * cos(3.94608385223  + 9801.15826682010 *T)
    sumY0  +=       0.029 * cos(2.40620954155  + 10625.90046056890 *T)
    sumY0  +=       0.030 * cos(2.05100908229  + 20432.33783421290 *T)
    sumY0  +=       0.031 * cos(3.37186993722  + 7255.81346921790 *T)
    sumY0  +=       0.029 * cos(5.98151124473  + 9889.53631207009 *T)
    sumY0  +=       0.029 * cos(0.20516018454  + 224.58861318540 *T)
    sumY0  +=       0.028 * cos(4.98574596975  + 10099.09007500929 *T)
    sumY0  +=       0.028 * cos(1.36654742403  + 10327.96865237972 *T)
    sumY0  +=       0.037 * cos(1.58092203649  + 47964.09346827550 *T)
    sumY0  +=       0.029 * cos(3.50171158414  + 10426.34082416550 *T)
    sumY0  +=       0.034 * cos(2.65907611892  + 1539.55968729550 *T)
    sumY0  +=       0.039 * cos(3.61856411890  + 30563.83438484090 *T)
    sumY0  +=       0.039 * cos(2.90348082239  + 5791.16874004909 *T)
    sumY0  +=       0.028 * cos(0.72410945645  + 1178.95072155610 *T)
    sumY0  +=       0.031 * cos(3.89571253711  + 39786.89052513750 *T)
    sumY0  +=       0.028 * cos(2.08869873285  + 73872.40711453209 *T)
    sumY0  +=       0.027 * cos(2.10097762457  + 9411.22079760369 *T)
    sumY0  +=       0.028 * cos(5.13153492293  + 15694.78428256210 *T)
    sumY0  +=       0.034 * cos(2.10776670451  + 40213.48871601350 *T)
    sumY0  +=       0.037 * cos(2.87146864607  + 36949.47462590770 *T)
    sumY0  +=       0.035 * cos(1.99221623651  + 37704.22184173930 *T)
    sumY0  +=       0.028 * cos(4.41249255504  + 4732.27444482690 *T)
    sumY0  +=       0.037 * cos(1.06810151543  + 13859.87974104890 *T)
    sumY0  +=       0.035 * cos(4.29368361875  + 18093.13088206649 *T)
    sumY0  +=       0.035 * cos(4.29368361875  + 18093.61851703350 *T)
    sumY0  +=       0.031 * cos(1.42888574956  + 42430.72954477530 *T)
    sumY0  +=       0.026 * cos(2.52843792658  + 8734.66278963010 *T)
    sumY0  +=       0.028 * cos(3.49562374283  + 7886.47389475550 *T)
    sumY0  +=       0.031 * cos(1.84284071261  + 25960.17883340570 *T)

    # VENUS Y Coord
    sumY1  = 0
    sumY1  +=   51880.979 * cos(0.71095646371  + 0.24381748350 *T)
    sumY1  +=   17260.820 * cos(5.63797390819  + 20426.81490990550 *T)
    sumY1  +=   11049.864 * cos(3.60540297803  + 10213.04172872750 *T)
    sumY1  +=    4831.725 * cos(3.20542321593  + 10213.52936369450 *T)
    sumY1  +=     175.352 * cos(0.23521252264  + 30640.10045611650 *T)
    sumY1  +=      60.293 * cos(3.35720999584  + 8636.18582124671 *T)
    sumY1  +=      62.982 * cos(0.52034329307  + 10187.23104389470 *T)
    sumY1  +=      62.707 * cos(5.83097900617  + 10239.82768349430 *T)
    sumY1  +=      49.342 * cos(3.00369011281  + 11790.87290614230 *T)
    sumY1  +=      36.826 * cos(0.49986531697  + 775.76642880750 *T)
    sumY1  +=      39.689 * cos(4.87663392982  + 20426.32727493849 *T)
    sumY1  +=      30.101 * cos(3.75628445251  + 10022.08109758290 *T)
    sumY1  +=      29.330 * cos(2.59462276840  + 10404.97762980610 *T)
    sumY1  +=      19.990 * cos(5.87132108056  + 9438.00675237050 *T)
    sumY1  +=      17.877 * cos(0.50656677315  + 10989.05197501850 *T)
    sumY1  +=      14.175 * cos(2.13477380875  + 9683.83839859990 *T)
    sumY1  +=      13.388 * cos(2.90924092450  + 191.20444862810 *T)
    sumY1  +=      13.513 * cos(4.20262757897  + 10743.22032878910 *T)
    sumY1  +=      14.034 * cos(1.92852680423  + 4705.97612502710 *T)
    sumY1  +=      13.385 * cos(5.86252489050  + 19651.29229858150 *T)
    sumY1  +=       6.873 * cos(3.26551570952  + 9104.15081160110 *T)
    sumY1  +=       7.547 * cos(4.00471700687  + 1059.62574767270 *T)
    sumY1  +=       7.388 * cos(4.42957705498  + 15721.08260236190 *T)
    sumY1  +=       5.971 * cos(3.08667185392  + 11322.90791578790 *T)
    sumY1  +=       5.966 * cos(5.81970120815  + 10220.64291069530 *T)
    sumY1  +=       5.965 * cos(0.53310334979  + 10206.41581669370 *T)
    sumY1  +=       5.155 * cos(4.06691487180  + 3154.44326741210 *T)
    sumY1  +=       4.571 * cos(3.01985951758  + 20618.26317601710 *T)
    sumY1  +=       4.246 * cos(4.48337967211  + 10000.23026825651 *T)
    sumY1  +=       4.157 * cos(1.86902016050  + 10426.82845913250 *T)
    sumY1  +=       4.649 * cos(1.29491136363  + 9830.63283147130 *T)
    sumY1  +=       4.431 * cos(5.05738488158  + 10596.42589591770 *T)
    sumY1  +=       3.938 * cos(5.07016509425  + 7.35736448430 *T)
    sumY1  +=       3.399 * cos(0.82461824896  + 7860.66320992270 *T)
    sumY1  +=       2.758 * cos(2.34021605546  + 19367.43297971630 *T)
    sumY1  +=       2.699 * cos(5.36901161259  + 7058.84227879890 *T)
    sumY1  +=       2.515 * cos(0.93953422896  + 8623.96883344369 *T)
    sumY1  +=       2.725 * cos(0.25399973919  + 801.57711364030 *T)
    sumY1  +=       2.539 * cos(4.39873985192  + 9411.70843257070 *T)
    sumY1  +=       1.822 * cos(4.89709335392  + 1109.13473460990 *T)
    sumY1  +=       2.300 * cos(1.94517236646  + 11015.35029481830 *T)
    sumY1  +=       1.816 * cos(0.95974773242  + 13368.21644859010 *T)
    sumY1  +=       1.814 * cos(3.28732569612  + 3930.45351370310 *T)
    sumY1  +=       1.586 * cos(1.12226291927  + 40853.38600232750 *T)
    sumY1  +=       1.999 * cos(4.09704295416  + 26.05450231630 *T)
    sumY1  +=       1.745 * cos(0.69413221331  + 4552.19731454230 *T)
    sumY1  +=       1.503 * cos(5.35238315560  + 12566.39551746630 *T)
    sumY1  +=       1.446 * cos(5.48745108690  + 7084.65296363170 *T)
    sumY1  +=       1.400 * cos(4.67651135088  + 11272.91129388371 *T)
    sumY1  +=       1.282 * cos(3.09230484373  + 10219.05228800190 *T)
    sumY1  +=       1.282 * cos(3.25998855005  + 10208.00643938710 *T)
    sumY1  +=       1.352 * cos(1.29314419649  + 20419.70136290470 *T)
    sumY1  +=       1.475 * cos(4.11277346234  + 3128.63258257930 *T)
    sumY1  +=       1.235 * cos(0.55003403789  + 1577.58735993130 *T)
    sumY1  +=       1.286 * cos(4.12341945563  + 23581.50199480110 *T)
    sumY1  +=       1.283 * cos(5.45488591011  + 382.65271473970 *T)
    sumY1  +=       1.171 * cos(1.71768767831  + 9154.14743350530 *T)
    sumY1  +=       1.084 * cos(2.78444751991  + 10198.27689250950 *T)
    sumY1  +=       1.084 * cos(3.56784587387  + 10228.78183487951 *T)
    sumY1  +=       1.198 * cos(3.38905914109  + 16496.60521368590 *T)
    sumY1  +=       0.869 * cos(3.97005901207  + 2352.62233628830 *T)
    sumY1  +=       1.112 * cos(4.99559909082  + 1589.31671276730 *T)
    sumY1  +=       0.922 * cos(5.66846628682  + 15874.86141284670 *T)
    sumY1  +=       0.743 * cos(5.39400352278  + 28287.23430234470 *T)
    sumY1  +=       0.746 * cos(0.58295457724  + 3149.40797807170 *T)
    sumY1  +=       0.797 * cos(3.54113611340  + 426.84200835950 *T)
    sumY1  +=       0.625 * cos(1.61452093926  + 20400.51659010570 *T)
    sumY1  +=       0.615 * cos(4.93517817028  + 21536.19346199890 *T)
    sumY1  +=       0.832 * cos(0.35357361779  + 21228.63584102931 *T)
    sumY1  +=       0.656 * cos(0.63035969025  + 11014.86265985130 *T)
    sumY1  +=       0.586 * cos(0.59640409991  + 5661.57586663570 *T)
    sumY1  +=       0.648 * cos(0.34210160772  + 537.04832957890 *T)
    sumY1  +=       0.674 * cos(0.56456739745  + 8094.27786834910 *T)
    sumY1  +=       0.636 * cos(4.96350573817  + 18849.47136745770 *T)
    sumY1  +=       0.655 * cos(2.23854746047  + 17298.42614480970 *T)
    sumY1  +=       0.615 * cos(0.82332897293  + 522.82123557730 *T)
    sumY1  +=       0.507 * cos(5.26511211265  + 20453.11322970530 *T)
    sumY1  +=       0.451 * cos(3.55775064489  + 32217.44399856430 *T)
    sumY1  +=       0.480 * cos(4.03262150754  + 529.44714761109 *T)
    sumY1  +=       0.439 * cos(5.33608025421  + 6282.83203250789 *T)
    sumY1  +=       0.439 * cos(1.01049396969  + 29051.02756083270 *T)
    sumY1  +=       0.410 * cos(3.89856473824  + 24357.02460612510 *T)
    sumY1  +=       0.452 * cos(5.33295391690  + 9146.54625153749 *T)
    sumY1  +=       0.417 * cos(3.34912055717  + 6681.46867088311 *T)
    sumY1  +=       0.434 * cos(4.87041677173  + 9160.77334553909 *T)
    sumY1  +=       0.364 * cos(1.14500385840  + 10640.12755457050 *T)
    sumY1  +=       0.363 * cos(5.18731183247  + 9786.93117281850 *T)
    sumY1  +=       0.428 * cos(5.53178842562  + 20809.71144212870 *T)
    sumY1  +=       0.440 * cos(0.90299311231  + 9690.95194560070 *T)
    sumY1  +=       0.440 * cos(5.44930026371  + 10736.10678178830 *T)
    sumY1  +=       0.372 * cos(5.21504043865  + 10251.55703633031 *T)
    sumY1  +=       0.372 * cos(1.13725295513  + 10175.50169105870 *T)
    sumY1  +=       0.319 * cos(5.77038175361  + 17277.65074931730 *T)
    sumY1  +=       0.427 * cos(1.38502821135  + 18837.74201462170 *T)
    sumY1  +=       0.319 * cos(1.75892023078  + 7860.17557495569 *T)
    sumY1  +=       0.303 * cos(5.78500325867  + 14765.48286075331 *T)
    sumY1  +=       0.351 * cos(0.93732183183  + 6490.02040477150 *T)
    sumY1  +=       0.287 * cos(1.70533444360  + 1577.09972496430 *T)
    sumY1  +=       0.318 * cos(2.05196802231  + 14945.07235607090 *T)
    sumY1  +=       0.297 * cos(5.78070702462  + 10233.19912459430 *T)
    sumY1  +=       0.297 * cos(0.57158636916  + 10193.85960279471 *T)
    sumY1  +=       0.372 * cos(4.36086070254  + 22004.15845235330 *T)
    sumY1  +=       0.311 * cos(2.86660649993  + 9683.35076363289 *T)
    sumY1  +=       0.284 * cos(0.78946184560  + 3442.81876244891 *T)
    sumY1  +=       0.269 * cos(2.73302271647  + 9573.14444241349 *T)
    sumY1  +=       0.262 * cos(1.65344694622  + 21202.33752122950 *T)
    sumY1  +=       0.285 * cos(2.99593464960  + 13745.59005650590 *T)
    sumY1  +=       0.276 * cos(6.13791366947  + 9676.72485159910 *T)
    sumY1  +=       0.289 * cos(4.21532854719  + 20235.36664379391 *T)
    sumY1  +=       0.263 * cos(4.06023282629  + 22779.68106367730 *T)
    sumY1  +=       0.280 * cos(3.52465276935  + 14919.26167123810 *T)
    sumY1  +=       0.262 * cos(0.23413334905  + 10750.33387578991 *T)
    sumY1  +=       0.256 * cos(2.84423020029  + 10234.30475918690 *T)
    sumY1  +=       0.239 * cos(3.17649373533  + 5507.79705615090 *T)
    sumY1  +=       0.253 * cos(3.50443427097  + 10192.75396820210 *T)
    sumY1  +=       0.268 * cos(5.52795627850  + 27511.71169102070 *T)
    sumY1  +=       0.282 * cos(1.52617460595  + 4731.78680985990 *T)
    sumY1  +=       0.232 * cos(5.81477929300  + 3723.26514143949 *T)
    sumY1  +=       0.285 * cos(3.95599071883  + 20956.50587500010 *T)
    sumY1  +=       0.244 * cos(2.32200831696  + 2353.10997125530 *T)
    sumY1  +=       0.272 * cos(2.82574902606  + 20000.21671902950 *T)
    sumY1  +=       0.254 * cos(3.11073677155  + 6283.31966747490 *T)
    sumY1  +=       0.238 * cos(6.06399121577  + 30639.61282114949 *T)
    sumY1  +=       0.232 * cos(6.01167584587  + 19890.01039781010 *T)
    sumY1  +=       0.210 * cos(3.33764904887  + 18073.46112116669 *T)
    sumY1  +=       0.216 * cos(3.19588337312  + 18073.94875613370 *T)
    sumY1  +=       0.199 * cos(5.04355472771  + 9639.18456535971 *T)
    sumY1  +=       0.218 * cos(5.53074261320  + 19904.23749181170 *T)
    sumY1  +=       0.225 * cos(5.41510051366  + 13937.03832261750 *T)
    sumY1  +=       0.235 * cos(1.73003777450  + 1551.28904013150 *T)
    sumY1  +=       0.191 * cos(2.76287874160  + 6872.91693699471 *T)
    sumY1  +=       0.179 * cos(4.84535687472  + 30110.40949102190 *T)
    sumY1  +=       0.185 * cos(1.30870554546  + 10787.87416202930 *T)
    sumY1  +=       0.235 * cos(1.24541538744  + 1066.73929467350 *T)
    sumY1  +=       0.208 * cos(4.04763155841  + 10007.34381525730 *T)
    sumY1  +=       0.208 * cos(2.30466183537  + 10419.71491213170 *T)
    sumY1  +=       0.179 * cos(3.50084200533  + 18875.28205229050 *T)
    sumY1  +=       0.184 * cos(0.09028649493  + 9967.69771746511 *T)
    sumY1  +=       0.166 * cos(3.49629785102  + 18307.56341456009 *T)
    sumY1  +=       0.163 * cos(3.38894514289  + 10225.25871653050 *T)
    sumY1  +=       0.163 * cos(2.96334825089  + 10201.80001085851 *T)
    sumY1  +=       0.156 * cos(4.49619634182  + 10316.62213791310 *T)
    sumY1  +=       0.156 * cos(1.85609705196  + 10110.43658947590 *T)
    sumY1  +=       0.178 * cos(6.26185121103  + 10459.36100992390 *T)
    sumY1  +=       0.172 * cos(2.71508119650  + 14143.73905991410 *T)
    sumY1  +=       0.163 * cos(4.90813788389  + 8616.85528644289 *T)
    sumY1  +=       0.163 * cos(5.91344954982  + 6.86972951729 *T)
    sumY1  +=       0.157 * cos(3.58488866748  + 13554.14179039430 *T)
    sumY1  +=       0.152 * cos(3.62017418311  + 640.14110379750 *T)
    sumY1  +=       0.149 * cos(6.16738885127  + 7063.87756813929 *T)
    sumY1  +=       0.179 * cos(1.00591554381  + 29864.57784479250 *T)
    sumY1  +=       0.140 * cos(5.31441141872  + 8631.08238044450 *T)
    sumY1  +=       0.140 * cos(6.19029890817  + 529.93478257810 *T)
    sumY1  +=       0.131 * cos(2.45194876830  + 9794.04471981930 *T)
    sumY1  +=       0.170 * cos(3.49470295745  + 9153.65979853830 *T)
    sumY1  +=       0.128 * cos(3.89903128812  + 10633.01400756970 *T)
    sumY1  +=       0.177 * cos(3.03845859741  + 5481.49873635110 *T)
    sumY1  +=       0.154 * cos(5.69369789989  + 2119.00767786191 *T)
    sumY1  +=       0.131 * cos(5.56674422840  + 16984.23996494010 *T)
    sumY1  +=       0.131 * cos(4.35512447481  + 9437.51911740350 *T)
    sumY1  +=       0.118 * cos(1.90850049529  + 11790.38527117530 *T)
    sumY1  +=       0.140 * cos(2.00742093288  + 191.69208359510 *T)
    sumY1  +=       0.143 * cos(3.32004481323  + 7994.77225950771 *T)
    sumY1  +=       0.123 * cos(1.23478424338  + 213.54291292150 *T)
    sumY1  +=       0.111 * cos(2.74887704228  + 8661.99650607950 *T)
    sumY1  +=       0.121 * cos(0.64210487280  + 28521.33659573810 *T)
    sumY1  +=       0.119 * cos(3.20586515571  + 3929.96587873609 *T)
    sumY1  +=       0.106 * cos(4.33742493006  + 1109.62236957690 *T)
    sumY1  +=       0.099 * cos(3.87128339263  + 20.53157800890 *T)
    sumY1  +=       0.104 * cos(4.60451231291  + 36147.65369478390 *T)
    sumY1  +=       0.107 * cos(4.92712251717  + 22805.49174851010 *T)
    sumY1  +=       0.125 * cos(3.09576538097  + 12432.28646788130 *T)
    sumY1  +=       0.101 * cos(1.57214857483  + 1052.51220067191 *T)
    sumY1  +=       0.093 * cos(1.71903836713  + 574.10098085129 *T)
    sumY1  +=       0.129 * cos(0.19292287088  + 7564.58690325449 *T)
    sumY1  +=       0.092 * cos(3.49801814083  + 9359.84534697549 *T)
    sumY1  +=       0.099 * cos(0.75167027759  + 31441.92138724031 *T)
    sumY1  +=       0.091 * cos(1.47446366588  + 775.27879384050 *T)
    sumY1  +=       0.112 * cos(0.41548066865  + 10186.74340892770 *T)
    sumY1  +=       0.095 * cos(2.09703168525  + 10433.94200613330 *T)
    sumY1  +=       0.095 * cos(4.25526170853  + 9993.11672125570 *T)
    sumY1  +=       0.092 * cos(1.92657936145  + 14143.25142494710 *T)
    sumY1  +=       0.096 * cos(3.75726377834  + 220.65645992230 *T)
    sumY1  +=       0.084 * cos(1.09570291858  + 6770.46678376209 *T)
    sumY1  +=       0.088 * cos(3.56094956187  + 11322.42028082089 *T)
    sumY1  +=       0.110 * cos(5.01502416629  + 11802.60225897830 *T)
    sumY1  +=       0.108 * cos(3.56320703640  + 29580.71852592730 *T)
    sumY1  +=       0.093 * cos(4.32241552615  + 2957.95971196010 *T)
    sumY1  +=       0.101 * cos(0.52107246051  + 6770.95441872910 *T)
    sumY1  +=       0.105 * cos(3.18747751371  + 8662.48414104651 *T)
    sumY1  +=       0.094 * cos(6.19217661869  + 20433.92845690630 *T)
    sumY1  +=       0.089 * cos(3.84885484388  + 15720.59496739490 *T)
    sumY1  +=       0.088 * cos(4.22623893206  + 206.42936592071 *T)
    sumY1  +=       0.101 * cos(0.50947630970  + 7085.14059859870 *T)
    sumY1  +=       0.077 * cos(2.93440992089  + 10177.50149701710 *T)
    sumY1  +=       0.077 * cos(3.41788347289  + 10249.55723037191 *T)
    sumY1  +=       0.087 * cos(2.72553149790  + 2378.92065608810 *T)
    sumY1  +=       0.080 * cos(2.87657638659  + 5507.30942118390 *T)
    sumY1  +=       0.100 * cos(3.31433227306  + 14945.55999103790 *T)
    sumY1  +=       0.096 * cos(1.34435751761  + 8624.45646841071 *T)
    sumY1  +=       0.071 * cos(5.88504983472  + 24150.32386882850 *T)
    sumY1  +=       0.089 * cos(4.61897054824  + 18875.76968725751 *T)
    sumY1  +=       0.071 * cos(4.97846455178  + 29043.91401383189 *T)
    sumY1  +=       0.075 * cos(4.48442686361  + 19897.12394481090 *T)
    sumY1  +=       0.068 * cos(2.38966490616  + 10384.20223431370 *T)
    sumY1  +=       0.068 * cos(3.96262848762  + 10042.85649307530 *T)
    sumY1  +=       0.077 * cos(5.37168621038  + 29573.60497892650 *T)
    sumY1  +=       0.073 * cos(4.93040355002  + 29587.83207292810 *T)
    sumY1  +=       0.063 * cos(4.96682804495  + 10239.34004852730 *T)
    sumY1  +=       0.086 * cos(5.50379681688  + 19650.80466361450 *T)
    sumY1  +=       0.086 * cos(5.13181328155  + 19360.31943271550 *T)
    sumY1  +=       0.061 * cos(5.37335140170  + 29058.14110783350 *T)
    sumY1  +=       0.081 * cos(3.17506029506  + 11764.57458634250 *T)
    sumY1  +=       0.066 * cos(0.09242614010  + 26735.70144472970 *T)
    sumY1  +=       0.075 * cos(5.86155964498  + 25934.36814857290 *T)
    sumY1  +=       0.067 * cos(5.87156884416  + 40077.86339100350 *T)
    sumY1  +=       0.059 * cos(0.24096472060  + 10988.56434005150 *T)
    sumY1  +=       0.077 * cos(5.42249781467  + 20213.51581446750 *T)
    sumY1  +=       0.077 * cos(1.74322703053  + 7834.36489012290 *T)
    sumY1  +=       0.057 * cos(2.63908550678  + 10404.48999483909 *T)
    sumY1  +=       0.057 * cos(6.11425964512  + 110.45013870291 *T)
    sumY1  +=       0.067 * cos(3.14352797658  + 17777.87244946549 *T)
    sumY1  +=       0.060 * cos(4.64344417198  + 30831.54872222810 *T)
    sumY1  +=       0.058 * cos(4.04812917862  + 12592.20620229910 *T)
    sumY1  +=       0.054 * cos(3.33029969221  + 19317.43635781211 *T)
    sumY1  +=       0.065 * cos(1.59839937239  + 25158.84553724890 *T)
    sumY1  +=       0.064 * cos(0.63814287930  + 18308.05104952710 *T)
    sumY1  +=       0.059 * cos(1.67190949806  + 9779.32999085069 *T)
    sumY1  +=       0.053 * cos(6.23173234110  + 27490.93629552830 *T)
    sumY1  +=       0.050 * cos(0.18460290148  + 10103.32304247510 *T)
    sumY1  +=       0.050 * cos(6.16769049230  + 10323.73568491390 *T)
    sumY1  +=       0.050 * cos(4.63939060613  + 39794.00407213830 *T)
    sumY1  +=       0.055 * cos(1.18770302160  + 9793.55708485230 *T)
    sumY1  +=       0.048 * cos(2.66058888107  + 29089.05523346850 *T)
    sumY1  +=       0.051 * cos(5.30933109582  + 26709.89075989690 *T)
    sumY1  +=       0.047 * cos(5.42237264786  + 16522.41589851870 *T)
    sumY1  +=       0.060 * cos(4.61502027021  + 12592.69383726610 *T)
    sumY1  +=       0.057 * cos(4.83119299148  + 2218.51328670329 *T)
    sumY1  +=       0.047 * cos(2.05343620014  + 20640.11400534350 *T)
    sumY1  +=       0.046 * cos(2.74639353796  + 19786.91762359150 *T)
    sumY1  +=       0.047 * cos(4.18909923765  + 57836.89451481709 *T)
    sumY1  +=       0.045 * cos(2.36837617101  + 3532.30451029490 *T)
    sumY1  +=       0.047 * cos(1.59938927555  + 22003.67081738630 *T)
    sumY1  +=       0.045 * cos(2.81454147982  + 30000.20316980250 *T)
    sumY1  +=       0.043 * cos(3.56791888552  + 29786.90407436450 *T)
    sumY1  +=       0.051 * cos(0.93432433798  + 44008.07308722310 *T)
    sumY1  +=       0.042 * cos(6.21590796001  + 9999.74263328950 *T)
    sumY1  +=       0.050 * cos(0.51812370768  + 23581.01435983410 *T)
    sumY1  +=       0.051 * cos(6.00715809832  + 13341.91812879030 *T)
    sumY1  +=       0.040 * cos(3.99083942383  + 33019.26492968810 *T)
    sumY1  +=       0.051 * cos(5.84065090063  + 13656.10430865991 *T)
    sumY1  +=       0.048 * cos(1.01747986844  + 213.05527795450 *T)
    sumY1  +=       0.039 * cos(4.20765456607  + 20447.59030539790 *T)
    sumY1  +=       0.039 * cos(2.75710568559  + 10595.93826095070 *T)
    sumY1  +=       0.042 * cos(1.52233926028  + 30665.91114094930 *T)
    sumY1  +=       0.042 * cos(2.07548691715  + 3914.71340755109 *T)
    sumY1  +=       0.042 * cos(3.20676468713  + 38204.93117685450 *T)
    sumY1  +=       0.041 * cos(2.02945904881  + 17469.09901542890 *T)
    sumY1  +=       0.036 * cos(4.38221463656  + 10077.90403868451 *T)
    sumY1  +=       0.036 * cos(1.97007875722  + 10349.15468870450 *T)
    sumY1  +=       0.040 * cos(1.65514953516  + 18830.14083265389 *T)
    sumY1  +=       0.039 * cos(1.16926464863  + 18844.36792665549 *T)
    sumY1  +=       0.034 * cos(1.79297258774  + 21001.15970824030 *T)
    sumY1  +=       0.034 * cos(3.14405588009  + 13936.55068765050 *T)
    sumY1  +=       0.032 * cos(5.34063867594  + 36949.47462590770 *T)
    sumY1  +=       0.033 * cos(6.10548742756  + 16983.75232997309 *T)
    sumY1  +=       0.040 * cos(2.31586332501  + 47938.28278344270 *T)
    sumY1  +=       0.032 * cos(0.75258375907  + 39301.85314471250 *T)
    sumY1  +=       0.032 * cos(3.89417641266  + 39302.34077967950 *T)
    sumY1  +=       0.036 * cos(4.85111863254  + 19374.54652671710 *T)
    sumY1  +=       0.032 * cos(3.92861114802  + 16894.75421709410 *T)
    sumY1  +=       0.042 * cos(5.08245275253  + 55798.70217588190 *T)
    sumY1  +=       0.031 * cos(1.94569247568  + 20043.91837768230 *T)
    sumY1  +=       0.033 * cos(6.15578410865  + 48418.21672306549 *T)
    sumY1  +=       0.031 * cos(5.95509934930  + 9573.63207738050 *T)
    sumY1  +=       0.034 * cos(0.55952606060  + 54247.16931826690 *T)
    sumY1  +=       0.032 * cos(0.55000852758  + 10191.67853436810 *T)
    sumY1  +=       0.032 * cos(5.80228486620  + 10235.38019302090 *T)
    sumY1  +=       0.031 * cos(4.45781570376  + 8087.16432134830 *T)
    sumY1  +=       0.032 * cos(3.01069170124  + 19896.63630984389 *T)
    sumY1  +=       0.036 * cos(2.43177228108  + 7255.32583425090 *T)
    sumY1  +=       0.030 * cos(0.39582329080  + 10853.42665000850 *T)
    sumY1  +=       0.037 * cos(1.97789587606  + 27511.22405605370 *T)
    sumY1  +=       0.032 * cos(5.51602815822  + 19366.94534474929 *T)
    sumY1  +=       0.029 * cos(3.64083349524  + 10291.20313412251 *T)
    sumY1  +=       0.029 * cos(2.71145989854  + 10135.85559326650 *T)
    sumY1  +=       0.028 * cos(4.16989797155  + 104351.85638378028 *T)
    sumY1  +=       0.032 * cos(2.59502959120  + 20206.40226746670 *T)
    sumY1  +=       0.029 * cos(1.33160612830  + 29864.09020982550 *T)
    sumY1  +=       0.029 * cos(5.27506731832  + 49515.13869092350 *T)
    sumY1  +=       0.027 * cos(0.49570343491  + 5661.08823166870 *T)
    sumY1  +=       0.034 * cos(0.56834994987  + 9161.26098050610 *T)
    sumY1  +=       0.027 * cos(3.93043638128  + 6872.42930202769 *T)
    sumY1  +=       0.032 * cos(3.80176203654  + 9836.15575577870 *T)
    sumY1  +=       0.032 * cos(2.55053135724  + 10590.90297161030 *T)
    sumY1  +=       0.027 * cos(4.68213678931  + 6298.57213865991 *T)
    sumY1  +=       0.035 * cos(3.77990327166  + 9986.05323090550 *T)
    sumY1  +=       0.035 * cos(2.57239012212  + 10441.00549648351 *T)
    sumY1  +=       0.033 * cos(3.63785377361  + 31441.43375227330 *T)
    sumY1  +=       0.026 * cos(0.42189641581  + 40879.68432212730 *T)
    sumY1  +=       0.034 * cos(4.43221326186  + 1059.13811270570 *T)
    sumY1  +=       0.028 * cos(4.84223047836  + 27197.03787618409 *T)
    sumY1  +=       0.026 * cos(3.20524473117  + 34363.60941503950 *T)
    sumY1  +=       0.032 * cos(5.77908535567  + 11265.79774688290 *T)
    sumY1  +=       0.029 * cos(1.74100180776  + 30206.38871823969 *T)
    sumY1  +=       0.032 * cos(3.71439292372  + 51868.49247966230 *T)
    sumY1  +=       0.029 * cos(5.18039950515  + 35371.64344849290 *T)
    sumY1  +=       0.029 * cos(2.03880685156  + 35372.13108345990 *T)
    sumY1  +=       0.027 * cos(2.96900241192  + 34596.12083716890 *T)
    sumY1  +=       0.033 * cos(0.22008313011  + 59728.91187210151 *T)
    sumY1  +=       0.029 * cos(2.13149426586  + 20220.62936146830 *T)
    sumY1  +=       0.025 * cos(0.00480216002  + 25933.88051360590 *T)
    sumY1  +=       0.024 * cos(1.14435429192  + 8734.17515466309 *T)
    sumY1  +=       0.027 * cos(0.93366720196  + 66813.80865321669 *T)
    sumY1  +=       0.024 * cos(2.78431824704  + 36173.46437961670 *T)
    sumY1  +=       0.024 * cos(2.85202782937  + 31022.99698833970 *T)
    sumY1  +=       0.023 * cos(1.83521592175  + 24382.83529095790 *T)
    sumY1  +=       0.028 * cos(2.59038076535  + 27461.46703058189 *T)
    sumY1  +=       0.027 * cos(1.25678130892  + 30220.61581224130 *T)
    sumY1  +=       0.023 * cos(1.14068817117  + 10367.30817417930 *T)
    sumY1  +=       0.023 * cos(5.21160522261  + 10059.75055320970 *T)
    sumY1  +=       0.023 * cos(0.43117503950  + 20452.62559473830 *T)
    sumY1  +=       0.026 * cos(1.22669168244  + 39271.42665404450 *T)
    sumY1  +=       0.026 * cos(1.70917344642  + 39257.19956004290 *T)
    sumY1  +=       0.022 * cos(2.21618655385  + 43232.06284093210 *T)
    sumY1  +=       0.022 * cos(5.35777920744  + 43232.55047589910 *T)
    sumY1  +=       0.026 * cos(5.25456910654  + 8101.39141534989 *T)
    sumY1  +=       0.022 * cos(1.35771340076  + 24978.76840696430 *T)
    sumY1  +=       0.022 * cos(4.11458514050  + 25158.35790228190 *T)
    sumY1  +=       0.030 * cos(5.92769527498  + 17248.18148437089 *T)
    sumY1  +=       0.022 * cos(5.97215448902  + 27991.15799567649 *T)
    sumY1  +=       0.022 * cos(0.18339928832  + 30426.80136067850 *T)
    sumY1  +=       0.022 * cos(4.03960220837  + 802.06474860730 *T)
    sumY1  +=       0.028 * cos(0.58380772945  + 17297.93850984270 *T)
    sumY1  +=       0.021 * cos(5.72756462954  + 9147.03388650451 *T)
    sumY1  +=       0.024 * cos(0.86019134061  + 46386.74992582770 *T)
    sumY1  +=       0.021 * cos(1.15272291675  + 27197.52551115110 *T)
    sumY1  +=       0.021 * cos(1.79721426538  + 44809.89401834690 *T)
    sumY1  +=       0.024 * cos(4.32935447070  + 52176.05010063190 *T)
    sumY1  +=       0.025 * cos(2.80978996759  + 22805.97938347710 *T)
    sumY1  +=       0.022 * cos(2.50649207464  + 9566.03089541269 *T)
    sumY1  +=       0.023 * cos(4.04266888662  + 6309.13035230770 *T)
    sumY1  +=       0.021 * cos(0.48695657465  + 10949.40587722630 *T)
    sumY1  +=       0.021 * cos(5.86533681913  + 9477.65285016271 *T)
    sumY1  +=       0.024 * cos(5.27936669763  + 10288.31096226180 *T)
    sumY1  +=       0.024 * cos(1.07292669615  + 10138.74776512720 *T)
    sumY1  +=       0.027 * cos(2.11911473240  + 9580.25798941430 *T)
    sumY1  +=       0.020 * cos(1.17564252576  + 8635.69818627969 *T)
    sumY1  +=       0.020 * cos(0.64178978255  + 11280.02484088450 *T)
    sumY1  +=       0.021 * cos(5.31872372185  + 3531.81687532789 *T)
    sumY1  +=       0.022 * cos(5.81710713116  + 62883.59895699710 *T)
    sumY1  +=       0.020 * cos(6.04322502436  + 17468.61138046190 *T)
    sumY1  +=       0.023 * cos(1.10446661612  + 10263.28638916630 *T)
    sumY1  +=       0.023 * cos(5.24782677766  + 10163.77233822270 *T)
    sumY1  +=       0.027 * cos(5.92344983854  + 32216.95636359729 *T)
    sumY1  +=       0.021 * cos(4.91435357755  + 47624.09660357310 *T)
    sumY1  +=       0.023 * cos(6.01318073085  + 55022.69192959090 *T)
    sumY1  +=       0.021 * cos(1.20080376991  + 18300.44986755929 *T)
    sumY1  +=       0.026 * cos(1.64245342007  + 63659.12156832110 *T)
    sumY1  +=       0.021 * cos(2.86380051865  + 43071.65547154729 *T)
    sumY1  +=       0.019 * cos(0.23803378346  + 20316.60858868610 *T)
    sumY1  +=       0.027 * cos(4.48319230296  + 28286.74666737769 *T)
    sumY1  +=       0.020 * cos(5.63440295344  + 62081.77802587330 *T)
    sumY1  +=       0.021 * cos(2.14440776384  + 24341.77213494010 *T)
    sumY1  +=       0.020 * cos(1.13738779403  + 49477.59865325469 *T)
    sumY1  +=       0.020 * cos(2.35460910589  + 70744.01834943630 *T)
    sumY1  +=       0.018 * cos(1.66963360809  + 14128.48658872910 *T)
    sumY1  +=       0.022 * cos(0.39682892264  + 27991.64563064350 *T)
    sumY1  +=       0.018 * cos(0.69318171155  + 9050.56702431969 *T)
    sumY1  +=       0.018 * cos(2.78635204608  + 4705.48849006010 *T)
    sumY1  +=       0.017 * cos(3.20253456324  + 28313.04498717750 *T)
    sumY1  +=       0.017 * cos(2.89862116564  + 10497.38868255971 *T)
    sumY1  +=       0.017 * cos(3.45367222814  + 9929.67004482930 *T)
    sumY1  +=       0.017 * cos(5.48500145482  + 19624.99397878170 *T)
    sumY1  +=       0.018 * cos(1.11001496449  + 29050.53992586569 *T)
    sumY1  +=       0.017 * cos(1.42200475410  + 18314.67696156089 *T)
    sumY1  +=       0.016 * cos(3.18072890080  + 48740.10371456650 *T)
    sumY1  +=       0.018 * cos(4.39890730410  + 38526.33053338850 *T)
    sumY1  +=       0.018 * cos(4.74135169951  + 42430.72954477530 *T)
    sumY1  +=       0.018 * cos(5.74382556326  + 42456.54022960810 *T)
    sumY1  +=       0.018 * cos(2.58329516885  + 33794.29990604510 *T)
    sumY1  +=       0.018 * cos(5.72488782244  + 33794.78754101210 *T)
    sumY1  +=       0.016 * cos(3.86440292184  + 9745.56437334010 *T)
    sumY1  +=       0.016 * cos(2.48789047194  + 10681.49435404891 *T)
    sumY1  +=       0.016 * cos(3.34228325637  + 38734.62214194910 *T)
    sumY1  +=       0.016 * cos(3.55643611626  + 47162.27253715170 *T)
    sumY1  +=       0.016 * cos(0.41484346267  + 47162.76017211870 *T)
    sumY1  +=       0.015 * cos(3.83207526863  + 10199.30226969290 *T)
    sumY1  +=       0.015 * cos(2.52021812515  + 10227.75645769610 *T)
    sumY1  +=       0.015 * cos(5.97366054337  + 41654.71929848429 *T)
    sumY1  +=       0.017 * cos(2.28143373849  + 50316.95962204730 *T)
    sumY1  +=       0.015 * cos(1.13326790706  + 53285.42865272530 *T)
    sumY1  +=       0.015 * cos(0.77488701815  + 66011.98772209289 *T)
    sumY1  +=       0.015 * cos(1.26649043492  + 38727.50859494830 *T)
    sumY1  +=       0.015 * cos(6.23110267766  + 46360.93924099490 *T)
    sumY1  +=       0.019 * cos(3.03299993077  + 67589.33126454070 *T)
    sumY1  +=       0.014 * cos(2.92323813617  + 21228.14820606230 *T)
    sumY1  +=       0.014 * cos(5.08679309387  + 40000.18962057550 *T)
    sumY1  +=       0.014 * cos(6.23973256044  + 11027.07964765431 *T)
    sumY1  +=       0.014 * cos(0.11256083334  + 9399.97907973470 *T)
    sumY1  +=       0.014 * cos(6.10753439498  + 37895.67010785090 *T)
    sumY1  +=       0.014 * cos(6.16127322322  + 37410.81105736210 *T)
    sumY1  +=       0.014 * cos(1.37136100611  + 50291.14893721450 *T)
    sumY1  +=       0.013 * cos(5.74299254495  + 12565.90788249930 *T)
    sumY1  +=       0.015 * cos(4.65282324499  + 21486.19684009470 *T)
    sumY1  +=       0.013 * cos(4.57052594573  + 52670.31341078610 *T)

    # VENUS Y Coord
    sumY2  = 0
    sumY2  +=   39116.576 * cos(3.17641452985  + 10213.52936369450 *T)
    sumY2  +=    1782.208 * cos(2.94415914625  + 0.24381748350 *T)
    sumY2  +=     953.335 * cos(5.20596121245  + 10213.04172872750 *T)
    sumY2  +=     818.754 * cos(3.60833841869  + 20426.81490990550 *T)
    sumY2  +=      12.014 * cos(3.98424371923  + 30640.10045611650 *T)
    sumY2  +=       8.598 * cos(5.19269385475  + 10239.82768349430 *T)
    sumY2  +=       8.636 * cos(1.10093366611  + 10187.23104389470 *T)
    sumY2  +=       3.631 * cos(0.41852821424  + 20426.32727493849 *T)
    sumY2  +=       2.965 * cos(1.54714831387  + 8636.18582124671 *T)
    sumY2  +=       3.797 * cos(5.44744767135  + 9438.00675237050 *T)
    sumY2  +=       3.421 * cos(0.82409517159  + 10989.05197501850 *T)
    sumY2  +=       2.218 * cos(2.12537117132  + 10022.08109758290 *T)
    sumY2  +=       2.134 * cos(4.12308786769  + 10404.97762980610 *T)
    sumY2  +=       2.187 * cos(4.57549125366  + 11790.87290614230 *T)
    sumY2  +=       1.794 * cos(4.97370212204  + 775.76642880750 *T)
    sumY2  +=       1.421 * cos(0.32903550115  + 2353.10997125530 *T)
    sumY2  +=       0.972 * cos(4.22805709693  + 1577.09972496430 *T)
    sumY2  +=       0.682 * cos(0.11738303047  + 4705.97612502710 *T)
    sumY2  +=       0.658 * cos(3.27522517046  + 9683.83839859990 *T)
    sumY2  +=       0.715 * cos(6.05529092994  + 9830.63283147130 *T)
    sumY2  +=       0.683 * cos(0.27418767588  + 10596.42589591770 *T)
    sumY2  +=       0.568 * cos(1.11272461938  + 19651.29229858150 *T)
    sumY2  +=       0.544 * cos(4.64774361853  + 191.20444862810 *T)
    sumY2  +=       0.678 * cos(2.88508832416  + 18073.94875613370 *T)
    sumY2  +=       0.574 * cos(0.38467340878  + 801.57711364030 *T)
    sumY2  +=       0.418 * cos(2.74408911585  + 10743.22032878910 *T)
    sumY2  +=       0.561 * cos(1.75565662404  + 6283.31966747490 *T)
    sumY2  +=       0.531 * cos(5.68669809373  + 7058.84227879890 *T)
    sumY2  +=       0.371 * cos(4.29421538467  + 9411.70843257070 *T)
    sumY2  +=       0.414 * cos(0.59494833209  + 529.93478257810 *T)
    sumY2  +=       0.287 * cos(1.55343952200  + 9104.15081160110 *T)
    sumY2  +=       0.314 * cos(5.99276842650  + 15721.08260236190 *T)
    sumY2  +=       0.301 * cos(3.87648034857  + 26.05450231630 *T)
    sumY2  +=       0.354 * cos(0.58512829013  + 13368.21644859010 *T)
    sumY2  +=       0.361 * cos(1.45998367037  + 14143.73905991410 *T)
    sumY2  +=       0.349 * cos(1.12786835203  + 22004.15845235330 *T)
    sumY2  +=       0.253 * cos(1.92471060417  + 1059.62574767270 *T)
    sumY2  +=       0.253 * cos(4.53940292522  + 11322.90791578790 *T)
    sumY2  +=       0.253 * cos(5.88126488763  + 3154.44326741210 *T)
    sumY2  +=       0.321 * cos(1.94846444699  + 11015.35029481830 *T)
    sumY2  +=       0.305 * cos(4.22585959217  + 3128.63258257930 *T)
    sumY2  +=       0.202 * cos(0.13780590630  + 10426.82845913250 *T)
    sumY2  +=       0.195 * cos(4.45983996443  + 20618.26317601710 *T)
    sumY2  +=       0.178 * cos(6.05244122419  + 10000.23026825651 *T)
    sumY2  +=       0.157 * cos(4.65387907430  + 40853.38600232750 *T)
    sumY2  +=       0.159 * cos(0.72069815424  + 382.65271473970 *T)
    sumY2  +=       0.177 * cos(0.39338528515  + 21228.63584102931 *T)
    sumY2  +=       0.126 * cos(1.05687133771  + 18849.47136745770 *T)
    sumY2  +=       0.131 * cos(5.44244944088  + 1577.58735993130 *T)
    sumY2  +=       0.148 * cos(5.58098705803  + 5507.30942118390 *T)
    sumY2  +=       0.139 * cos(2.98686042700  + 7860.66320992270 *T)
    sumY2  +=       0.146 * cos(2.62658040504  + 19897.12394481090 *T)
    sumY2  +=       0.133 * cos(2.03471349227  + 17298.42614480970 *T)
    sumY2  +=       0.088 * cos(2.31254159217  + 8623.96883344369 *T)
    sumY2  +=       0.116 * cos(5.13082226969  + 9153.65979853830 *T)
    sumY2  +=       0.111 * cos(4.24475286797  + 20453.11322970530 *T)
    sumY2  +=       0.077 * cos(1.00902400661  + 7084.65296363170 *T)
    sumY2  +=       0.081 * cos(5.42779981771  + 2352.62233628830 *T)
    sumY2  +=       0.088 * cos(4.73572405097  + 3930.45351370310 *T)
    sumY2  +=       0.077 * cos(1.55430183773  + 16496.60521368590 *T)
    sumY2  +=       0.069 * cos(5.74285388545  + 23581.50199480110 *T)
    sumY2  +=       0.058 * cos(2.86941970816  + 9154.14743350530 *T)
    sumY2  +=       0.062 * cos(0.46568901715  + 1109.13473460990 *T)
    sumY2  +=       0.063 * cos(0.71005704492  + 9437.51911740350 *T)
    sumY2  +=       0.047 * cos(2.16487309027  + 29051.02756083270 *T)
    sumY2  +=       0.054 * cos(0.77510495510  + 20809.71144212870 *T)
    sumY2  +=       0.055 * cos(1.84291935671  + 4731.78680985990 *T)
    sumY2  +=       0.046 * cos(3.98973988228  + 19367.43297971630 *T)
    sumY2  +=       0.044 * cos(2.81774994428  + 10787.87416202930 *T)
    sumY2  +=       0.044 * cos(3.53454344950  + 9639.18456535971 *T)
    sumY2  +=       0.040 * cos(1.76486455663  + 8094.27786834910 *T)
    sumY2  +=       0.041 * cos(1.81077528877  + 3929.96587873609 *T)
    sumY2  +=       0.047 * cos(3.42216864116  + 12566.39551746630 *T)
    sumY2  +=       0.038 * cos(0.61848121900  + 6282.83203250789 *T)
    sumY2  +=       0.045 * cos(2.64095270339  + 25934.36814857290 *T)
    sumY2  +=       0.034 * cos(3.37999447267  + 28287.23430234470 *T)
    sumY2  +=       0.033 * cos(2.41155726835  + 11014.86265985130 *T)
    sumY2  +=       0.034 * cos(5.75904174875  + 14919.26167123810 *T)
    sumY2  +=       0.032 * cos(1.85044469509  + 24357.02460612510 *T)
    sumY2  +=       0.028 * cos(2.26259259006  + 20235.36664379391 *T)
    sumY2  +=       0.040 * cos(2.04588679924  + 29580.71852592730 *T)
    sumY2  +=       0.033 * cos(3.78712068425  + 4552.19731454230 *T)
    sumY2  +=       0.033 * cos(5.68274603993  + 6490.02040477150 *T)
    sumY2  +=       0.027 * cos(5.51492765662  + 10640.12755457050 *T)
    sumY2  +=       0.026 * cos(0.80714498512  + 9786.93117281850 *T)
    sumY2  +=       0.026 * cos(5.24189234984  + 3149.40797807170 *T)
    sumY2  +=       0.025 * cos(0.52311174916  + 10192.75396820210 *T)
    sumY2  +=       0.025 * cos(2.69201575821  + 10234.30475918690 *T)
    sumY2  +=       0.024 * cos(0.09729823382  + 21536.19346199890 *T)
    sumY2  +=       0.026 * cos(0.29172518031  + 14143.25142494710 *T)
    sumY2  +=       0.024 * cos(1.53398817563  + 18307.56341456009 *T)
    sumY2  +=       0.024 * cos(4.66792746343  + 5507.79705615090 *T)
    sumY2  +=       0.025 * cos(1.99847907794  + 18875.28205229050 *T)
    sumY2  +=       0.030 * cos(2.25346620875  + 20400.51659010570 *T)
    sumY2  +=       0.025 * cos(0.91360430606  + 27511.71169102070 *T)
    sumY2  +=       0.025 * cos(3.28603866404  + 8661.99650607950 *T)
    sumY2  +=       0.019 * cos(3.72923840682  + 11272.91129388371 *T)
    sumY2  +=       0.022 * cos(1.87715843315  + 30639.61282114949 *T)
    sumY2  +=       0.026 * cos(0.86545419400  + 213.54291292150 *T)
    sumY2  +=       0.018 * cos(1.77063553751  + 28521.33659573810 *T)
    sumY2  +=       0.023 * cos(0.62230384603  + 13937.03832261750 *T)
    sumY2  +=       0.024 * cos(2.21558567781  + 13367.72881362310 *T)
    sumY2  +=       0.021 * cos(1.95869049210  + 10186.74340892770 *T)
    sumY2  +=       0.019 * cos(3.70619728169  + 10988.56434005150 *T)
    sumY2  +=       0.018 * cos(1.45798324955  + 32217.44399856430 *T)
    sumY2  +=       0.020 * cos(3.48812478045  + 191.69208359510 *T)
    sumY2  +=       0.017 * cos(5.71260028354  + 15874.86141284670 *T)
    sumY2  +=       0.014 * cos(1.56125812418  + 21202.33752122950 *T)
    sumY2  +=       0.014 * cos(0.92685471122  + 17277.65074931730 *T)
    sumY2  +=       0.015 * cos(1.68916723960  + 18073.46112116669 *T)
    sumY2  +=       0.014 * cos(0.06020277571  + 10239.34004852730 *T)
    sumY2  +=       0.014 * cos(3.96058683099  + 14945.07235607090 *T)
    sumY2  +=       0.014 * cos(2.61882493594  + 8662.48414104651 *T)
    sumY2  +=       0.014 * cos(3.66333417023  + 11764.57458634250 *T)
    sumY2  +=       0.013 * cos(0.36408900148  + 34570.31015233610 *T)
    sumY2  +=       0.013 * cos(5.32461592855  + 22779.68106367730 *T)
    sumY2  +=       0.012 * cos(2.28383473867  + 31441.92138724031 *T)
    sumY2  +=       0.011 * cos(2.70470885221  + 10220.64291069530 *T)
    sumY2  +=       0.011 * cos(0.50709992859  + 10206.41581669370 *T)
    sumY2  +=       0.010 * cos(2.04719193722  + 7.35736448430 *T)
    sumY2  +=       0.010 * cos(3.14863661708  + 22003.67081738630 *T)
    sumY2  +=       0.012 * cos(1.89856613068  + 25158.84553724890 *T)
    sumY2  +=       0.009 * cos(1.78868375262  + 38500.51984855570 *T)
    sumY2  +=       0.010 * cos(4.50140179786  + 11790.38527117530 *T)
    sumY2  +=       0.012 * cos(3.63812717753  + 17297.93850984270 *T)
    sumY2  +=       0.008 * cos(4.34372477567  + 10404.48999483909 *T)
    sumY2  +=       0.009 * cos(3.37718180492  + 29089.05523346850 *T)
    sumY2  +=       0.007 * cos(0.37788139417  + 43232.55047589910 *T)
    sumY2  +=       0.007 * cos(4.44321392733  + 30000.20316980250 *T)
    sumY2  +=       0.009 * cos(6.04652897824  + 7860.17557495569 *T)
    sumY2  +=       0.009 * cos(0.19167842758  + 3531.81687532789 *T)
    sumY2  +=       0.007 * cos(3.37107259426  + 22805.49174851010 *T)
    sumY2  +=       0.007 * cos(2.75239196673  + 529.44714761109 *T)

    # VENUS Y Coord
    sumY3  = 0
    sumY3  +=      55.518 * cos(0.56397150825  + 10213.04172872750 *T)
    sumY3  +=      57.047 * cos(4.30857137510  + 0.24381748350 *T)
    sumY3  +=      32.093 * cos(1.48532861520  + 20426.81490990550 *T)
    sumY3  +=      18.867 * cos(3.29469305606  + 10213.52936369450 *T)
    sumY3  +=       0.716 * cos(1.69291337487  + 30640.10045611650 *T)
    sumY3  +=       0.218 * cos(2.23416930390  + 20426.32727493849 *T)
    sumY3  +=       0.032 * cos(4.93547527406  + 8636.18582124671 *T)
    sumY3  +=       0.034 * cos(1.12379373423  + 10239.82768349430 *T)
    sumY3  +=       0.034 * cos(2.09231646083  + 10187.23104389470 *T)
    sumY3  +=       0.027 * cos(4.57242516087  + 11790.87290614230 *T)
    sumY3  +=       0.020 * cos(2.07822937351  + 775.76642880750 *T)
    sumY3  +=       0.016 * cos(5.32962355701  + 10022.08109758290 *T)
    sumY3  +=       0.016 * cos(4.16791854245  + 10404.97762980610 *T)
    sumY3  +=       0.010 * cos(2.19615394659  + 40853.38600232750 *T)
    sumY3  +=       0.010 * cos(1.23351868160  + 9438.00675237050 *T)
    sumY3  +=       0.010 * cos(2.08054199257  + 10989.05197501850 *T)
    sumY3  +=       0.008 * cos(3.70773784399  + 9683.83839859990 *T)
    sumY3  +=       0.007 * cos(1.33595632185  + 191.20444862810 *T)
    sumY3  +=       0.007 * cos(5.77536678710  + 10743.22032878910 *T)
    sumY3  +=       0.008 * cos(3.50194685886  + 4705.97612502710 *T)
    sumY3  +=       0.007 * cos(1.15263198918  + 19651.29229858150 *T)
    sumY3  +=       0.004 * cos(4.83876866063  + 9104.15081160110 *T)
    sumY3  +=       0.004 * cos(5.57864226359  + 1059.62574767270 *T)
    sumY3  +=       0.004 * cos(6.00284338140  + 15721.08260236190 *T)
    sumY3  +=       0.003 * cos(4.65996273292  + 11322.90791578790 *T)

    # VENUS Y Coord
    sumY4  = 0
    sumY4  +=      83.081 * cos(6.18993945631  + 10213.52936369450 *T)
    sumY4  +=       2.486 * cos(2.28856774293  + 10213.04172872750 *T)
    sumY4  +=       0.657 * cos(5.78367664587  + 20426.81490990550 *T)
    sumY4  +=       0.217 * cos(2.03456127835  + 0.24381748350 *T)
    sumY4  +=       0.035 * cos(5.69795458192  + 30640.10045611650 *T)
    sumY4  +=       0.011 * cos(4.01388645377  + 20426.32727493849 *T)
    sumY4  +=       0.005 * cos(0.49844804556  + 10239.82768349430 *T)
    sumY4  +=       0.005 * cos(2.68999633541  + 10187.23104389470 *T)
    sumY4  +=       0.003 * cos(3.35675923207  + 2353.10997125530 *T)
    sumY4  +=       0.002 * cos(3.32525148356  + 8636.18582124671 *T)
    sumY4  +=       0.002 * cos(1.20160813739  + 1577.09972496430 *T)
    sumY4  +=       0.002 * cos(0.72866875334  + 9438.00675237050 *T)

    # VENUS Y Coord
    sumY5  = 0
    sumY5  +=       0.631 * cos(0.02845828180  + 10213.52936369450 *T)
    sumY5  +=       0.095 * cos(4.11666356042  + 10213.04172872750 *T)
    sumY5  +=       0.088 * cos(0.15104017272  + 0.24381748350 *T)
    sumY5  +=       0.040 * cos(4.24049298565  + 20426.81490990550 *T)

    # VENUS Z Coord
    sumZ0  = 0
    sumZ0  +=  4282990.302 * cos(0.26703856476  + 10213.28554621100 *T)
    sumZ0  +=   35588.343 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ0  +=   14501.879 * cos(1.14696911390  + 20426.57109242200 *T)
    sumZ0  +=     140.675 * cos(0.85984113219  + 1577.34354244780 *T)
    sumZ0  +=     134.921 * cos(3.70465787853  + 2352.86615377180 *T)
    sumZ0  +=      95.750 * cos(3.66962547073  + 9437.76293488700 *T)
    sumZ0  +=      72.910 * cos(1.55750278240  + 9153.90361602180 *T)
    sumZ0  +=      73.654 * cos(2.02778434780  + 30639.85663863300 *T)
    sumZ0  +=      59.714 * cos(6.25390371649  + 18073.70493865020 *T)
    sumZ0  +=      45.307 * cos(2.28075620268  + 5507.55323866740 *T)
    sumZ0  +=      43.023 * cos(3.95620233472  + 529.69096509460 *T)
    sumZ0  +=      37.415 * cos(4.53900359933  + 22003.91463486980 *T)
    sumZ0  +=      32.943 * cos(0.72316171903  + 10239.58386601080 *T)
    sumZ0  +=      28.156 * cos(2.93415727239  + 10186.98722641120 *T)
    sumZ0  +=      32.447 * cos(3.86196252330  + 8635.94200376320 *T)
    sumZ0  +=      27.853 * cos(6.12790511758  + 11790.62908865880 *T)
    sumZ0  +=      23.013 * cos(0.98699428893  + 775.52261132400 *T)
    sumZ0  +=      19.425 * cos(5.12759462747  + 6283.07584999140 *T)
    sumZ0  +=      16.289 * cos(2.90852685195  + 10988.80815753500 *T)
    sumZ0  +=      15.382 * cos(5.98588709813  + 19896.88012732740 *T)
    sumZ0  +=      15.507 * cos(2.00401259454  + 13367.97263110660 *T)
    sumZ0  +=      12.412 * cos(4.13644030917  + 10021.83728009940 *T)
    sumZ0  +=      13.004 * cos(4.83195492470  + 14143.49524243060 *T)
    sumZ0  +=      10.730 * cos(5.61235409382  + 10404.73381232260 *T)
    sumZ0  +=      10.336 * cos(2.75167531541  + 191.44826611160 *T)
    sumZ0  +=      10.848 * cos(3.29419910254  + 11015.10647733480 *T)
    sumZ0  +=       9.775 * cos(1.21418881204  + 8624.21265092720 *T)
    sumZ0  +=      10.158 * cos(5.71954183827  + 9683.59458111640 *T)
    sumZ0  +=       8.080 * cos(2.69831049093  + 19651.04848109800 *T)
    sumZ0  +=       8.209 * cos(3.55371021890  + 10742.97651130560 *T)
    sumZ0  +=       6.416 * cos(0.46264840234  + 9786.68735533500 *T)
    sumZ0  +=       6.429 * cos(4.91481073656  + 1059.38193018920 *T)
    sumZ0  +=       6.722 * cos(2.41848486045  + 4705.73230754360 *T)
    sumZ0  +=       5.903 * cos(5.96215831403  + 25934.12433108940 *T)
    sumZ0  +=       5.647 * cos(3.57594715836  + 3154.68708489560 *T)
    sumZ0  +=       4.044 * cos(6.24650537722  + 11322.66409830440 *T)
    sumZ0  +=       4.033 * cos(3.42887820794  + 9103.90699411760 *T)
    sumZ0  +=       4.210 * cos(4.99232132925  + 7084.89678111520 *T)
    sumZ0  +=       4.135 * cos(3.43782832503  + 17298.18232732620 *T)
    sumZ0  +=       3.747 * cos(1.30318722019  + 15720.83878487840 *T)
    sumZ0  +=       3.417 * cos(3.95474989164  + 7860.41939243920 *T)
    sumZ0  +=       3.440 * cos(0.77140273510  + 4551.95349705880 *T)
    sumZ0  +=       3.488 * cos(0.14633368219  + 20618.01935853360 *T)
    sumZ0  +=       2.851 * cos(5.68591706153  + 10206.17199921020 *T)
    sumZ0  +=       2.750 * cos(3.89497095361  + 10192.51015071860 *T)
    sumZ0  +=       2.734 * cos(5.62996472233  + 12566.15169998280 *T)
    sumZ0  +=       2.682 * cos(4.58018033496  + 801.82093112380 *T)
    sumZ0  +=       2.743 * cos(6.06387494803  + 10234.06094170340 *T)
    sumZ0  +=       2.832 * cos(4.24241233220  + 213.29909543800 *T)
    sumZ0  +=       2.756 * cos(5.43236957664  + 29580.47470844380 *T)
    sumZ0  +=       2.155 * cos(0.29731467355  + 9411.46461508720 *T)
    sumZ0  +=       2.403 * cos(4.90533210845  + 19367.18916223280 *T)
    sumZ0  +=       2.103 * cos(1.12514771936  + 29864.33402730900 *T)
    sumZ0  +=       1.632 * cos(4.67123397256  + 1109.37855209340 *T)
    sumZ0  +=       1.787 * cos(2.52506338741  + 15874.61759536320 *T)
    sumZ0  +=       1.613 * cos(1.47525165805  + 7058.59846131540 *T)
    sumZ0  +=       1.598 * cos(2.21757389250  + 10426.58464164900 *T)
    sumZ0  +=       1.154 * cos(1.98818396175  + 26.29831979980 *T)
    sumZ0  +=       1.132 * cos(5.38517603552  + 7.11354700080 *T)
    sumZ0  +=       1.468 * cos(0.75541488791  + 9999.98645077300 *T)
    sumZ0  +=       1.118 * cos(0.96571599573  + 23581.25817731760 *T)
    sumZ0  +=       1.210 * cos(3.17717646354  + 3532.06069281140 *T)
    sumZ0  +=       0.965 * cos(0.01161931809  + 3128.38876509580 *T)
    sumZ0  +=       1.027 * cos(2.51592691800  + 33794.54372352860 *T)
    sumZ0  +=       0.847 * cos(1.24888706950  + 9573.38825989700 *T)
    sumZ0  +=       0.889 * cos(1.48214942791  + 9830.38901398780 *T)
    sumZ0  +=       0.953 * cos(5.73118460251  + 20213.27199698400 *T)
    sumZ0  +=       0.987 * cos(4.70495944427  + 14945.31617355440 *T)
    sumZ0  +=       1.030 * cos(0.84929643011  + 8094.52168583260 *T)
    sumZ0  +=       0.770 * cos(4.54196353475  + 21228.39202354580 *T)
    sumZ0  +=       0.778 * cos(1.16378510983  + 8662.24032356300 *T)
    sumZ0  +=       0.860 * cos(5.74281955136  + 5661.33204915220 *T)
    sumZ0  +=       0.670 * cos(5.36864204589  + 1589.07289528380 *T)
    sumZ0  +=       0.615 * cos(1.96289781619  + 26087.90314157420 *T)
    sumZ0  +=       0.740 * cos(1.97614280765  + 10596.18207843420 *T)
    sumZ0  +=       0.606 * cos(6.22070159607  + 25158.60171976540 *T)
    sumZ0  +=       0.533 * cos(0.73765107493  + 28286.99048486120 *T)
    sumZ0  +=       0.547 * cos(5.06653964506  + 29050.78374334920 *T)
    sumZ0  +=       0.592 * cos(3.32632513186  + 3340.61242669980 *T)
    sumZ0  +=       0.591 * cos(4.62374022159  + 9929.42622734580 *T)
    sumZ0  +=       0.537 * cos(6.00327539255  + 4732.03062734340 *T)
    sumZ0  +=       0.569 * cos(5.37276655331  + 10497.14486507620 *T)
    sumZ0  +=       0.520 * cos(6.11596800207  + 10218.80847051840 *T)
    sumZ0  +=       0.542 * cos(3.94171539558  + 37724.75341974820 *T)
    sumZ0  +=       0.575 * cos(2.29073875885  + 21535.94964451540 *T)
    sumZ0  +=       0.437 * cos(2.42330391120  + 10251.31321884680 *T)
    sumZ0  +=       0.439 * cos(2.90237828484  + 40853.14218484400 *T)
    sumZ0  +=       0.459 * cos(1.69153061401  + 3930.20969621960 *T)
    sumZ0  +=       0.405 * cos(1.42927875873  + 9161.01716302260 *T)
    sumZ0  +=       0.418 * cos(5.48110803243  + 32217.20018108080 *T)
    sumZ0  +=       0.521 * cos(3.84286345199  + 10207.76262190360 *T)
    sumZ0  +=       0.457 * cos(6.16974742379  + 10220.39909321180 *T)
    sumZ0  +=       0.440 * cos(1.25236241968  + 10175.25787357520 *T)
    sumZ0  +=       0.394 * cos(4.81650684498  + 9146.79006902100 *T)
    sumZ0  +=       0.362 * cos(0.02259543121  + 14765.23904326980 *T)
    sumZ0  +=       0.396 * cos(4.56209231282  + 10063.72234907640 *T)
    sumZ0  +=       0.407 * cos(6.15731262846  + 18875.52586977400 *T)
    sumZ0  +=       0.349 * cos(5.78507522549  + 12592.45001978260 *T)
    sumZ0  +=       0.379 * cos(4.56511543982  + 20419.45754542119 *T)
    sumZ0  +=       0.314 * cos(5.08259050420  + 18837.49819713819 *T)
    sumZ0  +=       0.343 * cos(2.36516734231  + 27511.46787353720 *T)
    sumZ0  +=       0.292 * cos(3.76618455555  + 6681.22485339960 *T)
    sumZ0  +=       0.289 * cos(3.59841384736  + 11272.66747640020 *T)
    sumZ0  +=       0.314 * cos(5.23220905264  + 426.59819087600 *T)
    sumZ0  +=       0.301 * cos(4.24971815099  + 14919.01785375460 *T)
    sumZ0  +=       0.322 * cos(1.36846420067  + 29088.81141598500 *T)
    sumZ0  +=       0.292 * cos(0.74997511570  + 3149.16416058820 *T)
    sumZ0  +=       0.306 * cos(0.47092243462  + 23958.63178523340 *T)
    sumZ0  +=       0.254 * cos(4.34026574490  + 10137.01947493540 *T)
    sumZ0  +=       0.304 * cos(5.33181632059  + 41654.96311596780 *T)
    sumZ0  +=       0.264 * cos(0.01153452145  + 20452.86941222180 *T)
    sumZ0  +=       0.334 * cos(5.28920083729  + 382.89653222320 *T)
    sumZ0  +=       0.236 * cos(1.03442196051  + 22779.43724619380 *T)
    sumZ0  +=       0.250 * cos(5.05395469123  + 24356.78078864160 *T)
    sumZ0  +=       0.278 * cos(0.10022232617  + 16496.36139620240 *T)
    sumZ0  +=       0.232 * cos(0.48064662619  + 17085.95866572220 *T)
    sumZ0  +=       0.205 * cos(6.19639409918  + 7064.12138562280 *T)
    sumZ0  +=       0.197 * cos(3.71380377240  + 1551.04522264800 *T)
    sumZ0  +=       0.200 * cos(4.06516779510  + 18849.22754997420 *T)
    sumZ0  +=       0.170 * cos(5.96522395118  + 10198.03307502600 *T)
    sumZ0  +=       0.170 * cos(3.52866583267  + 10228.53801739600 *T)
    sumZ0  +=       0.180 * cos(0.48168923159  + 45585.17281218740 *T)
    sumZ0  +=       0.144 * cos(2.81280885697  + 17277.40693183380 *T)
    sumZ0  +=       0.177 * cos(3.58248029244  + 20400.27277262220 *T)
    sumZ0  +=       0.170 * cos(3.72282190488  + 31441.67756975680 *T)
    sumZ0  +=       0.166 * cos(0.03569768750  + 13745.34623902240 *T)
    sumZ0  +=       0.127 * cos(5.14389302386  + 5481.25491886760 *T)
    sumZ0  +=       0.138 * cos(1.19253977825  + 22805.73556599360 *T)
    sumZ0  +=       0.126 * cos(4.21594185374  + 20956.26205751660 *T)
    sumZ0  +=       0.118 * cos(1.71706550416  + 13341.67431130680 *T)
    sumZ0  +=       0.131 * cos(0.92317994329  + 3442.57494496540 *T)
    sumZ0  +=       0.137 * cos(2.89407950366  + 33019.02111220460 *T)
    sumZ0  +=       0.104 * cos(2.47347358674  + 26709.64694241340 *T)
    sumZ0  +=       0.127 * cos(0.46833130674  + 536.80451209540 *T)
    sumZ0  +=       0.126 * cos(2.84049662693  + 9967.45389998160 *T)
    sumZ0  +=       0.107 * cos(4.19045394315  + 19999.97290154599 *T)
    sumZ0  +=       0.110 * cos(0.73944778374  + 30110.16567353840 *T)
    sumZ0  +=       0.138 * cos(4.01403107693  + 21202.09370374600 *T)
    sumZ0  +=       0.121 * cos(3.85928223786  + 522.57741809380 *T)
    sumZ0  +=       0.122 * cos(0.37000676558  + 10459.11719244040 *T)
    sumZ0  +=       0.095 * cos(5.25435135477  + 30213.25844775700 *T)
    sumZ0  +=       0.099 * cos(5.61149803116  + 10529.67741586760 *T)
    sumZ0  +=       0.106 * cos(2.42064996718  + 20809.46762464520 *T)
    sumZ0  +=       0.095 * cos(3.11317801589  + 9896.89367655440 *T)
    sumZ0  +=       0.085 * cos(5.88248062550  + 3723.50895892300 *T)
    sumZ0  +=       0.099 * cos(1.93329637732  + 49515.38250840700 *T)
    sumZ0  +=       0.084 * cos(4.93745373910  + 52175.80628314840 *T)
    sumZ0  +=       0.086 * cos(0.85755223960  + 2379.16447357160 *T)
    sumZ0  +=       0.114 * cos(5.11309629210  + 103.09277421860 *T)
    sumZ0  +=       0.110 * cos(1.44699918003  + 43232.30665841560 *T)
    sumZ0  +=       0.088 * cos(1.59017183044  + 10110.19277199240 *T)
    sumZ0  +=       0.086 * cos(0.87055324194  + 16522.65971600220 *T)
    sumZ0  +=       0.089 * cos(2.33714216061  + 10316.37832042960 *T)
    sumZ0  +=       0.098 * cos(4.66807552803  + 20235.12282631040 *T)
    sumZ0  +=       0.104 * cos(5.02724740402  + 35371.88726597640 *T)
    sumZ0  +=       0.093 * cos(4.21356144460  + 36301.18868778519 *T)
    sumZ0  +=       0.089 * cos(0.88060757437  + 6489.77658728800 *T)
    sumZ0  +=       0.083 * cos(2.07543775594  + 31749.23519072640 *T)
    sumZ0  +=       0.096 * cos(2.90682399304  + 9793.80090233580 *T)
    sumZ0  +=       0.078 * cos(2.65839634325  + 10232.95530711079 *T)
    sumZ0  +=       0.078 * cos(0.55230439694  + 10193.61578531121 *T)
    sumZ0  +=       0.092 * cos(5.63110477712  + 10288.06714477830 *T)
    sumZ0  +=       0.092 * cos(3.86278127025  + 10138.50394764370 *T)
    sumZ0  +=       0.090 * cos(4.34324193073  + 36949.23080842420 *T)
    sumZ0  +=       0.073 * cos(2.56775245907  + 13936.79450513400 *T)
    sumZ0  +=       0.071 * cos(0.64943236639  + 20.77539549240 *T)
    sumZ0  +=       0.090 * cos(0.07319663423  + 6872.67311951120 *T)
    sumZ0  +=       0.069 * cos(1.52914774412  + 9690.70812811720 *T)
    sumZ0  +=       0.086 * cos(1.02476175384  + 65236.22129328540 *T)
    sumZ0  +=       0.079 * cos(1.65446145032  + 39264.06928956020 *T)
    sumZ0  +=       0.071 * cos(0.70714638996  + 36147.40987730040 *T)
    sumZ0  +=       0.071 * cos(2.52879760001  + 26735.94526221320 *T)
    sumZ0  +=       0.061 * cos(5.47943297658  + 11764.33076885900 *T)
    sumZ0  +=       0.080 * cos(2.87220602049  + 47162.51635463520 *T)
    sumZ0  +=       0.065 * cos(2.53312656681  + 10735.86296430480 *T)
    sumZ0  +=       0.068 * cos(2.28501865578  + 8521.11987670860 *T)
    sumZ0  +=       0.065 * cos(2.12423586627  + 28313.28880466100 *T)
    sumZ0  +=       0.062 * cos(2.28587915309  + 3903.91137641980 *T)
    sumZ0  +=       0.069 * cos(0.42270579845  + 9580.50180689780 *T)
    sumZ0  +=       0.063 * cos(5.80076678078  + 2118.76386037840 *T)
    sumZ0  +=       0.056 * cos(4.73821324745  + 639.89728631400 *T)
    sumZ0  +=       0.063 * cos(3.35690006984  + 53445.59220462660 *T)
    sumZ0  +=       0.053 * cos(2.14583523296  + 8631.32619792800 *T)
    sumZ0  +=       0.066 * cos(4.12244312057  + 28521.09277825460 *T)
    sumZ0  +=       0.058 * cos(5.25045057985  + 10419.47109464820 *T)
    sumZ0  +=       0.058 * cos(4.23724598221  + 10007.09999777380 *T)
    sumZ0  +=       0.053 * cos(3.02375632858  + 19317.19254032860 *T)
    sumZ0  +=       0.071 * cos(2.36609582367  + 74.78159856730 *T)
    sumZ0  +=       0.062 * cos(2.46239069210  + 10846.06928552420 *T)
    sumZ0  +=       0.061 * cos(3.77333230443  + 51868.24866217880 *T)

    # VENUS Z Coord
    sumZ1  = 0
    sumZ1  +=  371157.532 * cos(1.80370484107  + 10213.28554621100 *T)
    sumZ1  +=    1583.350 * cos(3.38513432885  + 20426.57109242200 *T)
    sumZ1  +=     216.150 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ1  +=      12.186 * cos(5.60527972670  + 1577.34354244780 *T)
    sumZ1  +=      11.953 * cos(4.61859507922  + 30639.85663863300 *T)
    sumZ1  +=      11.662 * cos(5.24208633854  + 2352.86615377180 *T)
    sumZ1  +=      10.398 * cos(2.27393994810  + 9437.76293488700 *T)
    sumZ1  +=       6.381 * cos(0.28617884662  + 9153.90361602180 *T)
    sumZ1  +=       5.445 * cos(5.08079565040  + 10186.98722641120 *T)
    sumZ1  +=       4.769 * cos(1.50400058537  + 11790.62908865880 *T)
    sumZ1  +=       5.179 * cos(1.50697716140  + 18073.70493865020 *T)
    sumZ1  +=       3.981 * cos(0.75197593763  + 5507.55323866740 *T)
    sumZ1  +=       3.814 * cos(5.50447607190  + 529.69096509460 *T)
    sumZ1  +=       3.372 * cos(3.87807635377  + 10239.58386601080 *T)
    sumZ1  +=       2.840 * cos(1.28497223665  + 10404.73381232260 *T)
    sumZ1  +=       3.220 * cos(6.07567189618  + 22003.91463486980 *T)
    sumZ1  +=       2.336 * cos(4.88957218376  + 10988.80815753500 *T)
    sumZ1  +=       1.669 * cos(0.38548673512  + 6283.07584999140 *T)
    sumZ1  +=       1.381 * cos(4.32769653919  + 19651.04848109800 *T)
    sumZ1  +=       1.296 * cos(1.26608932430  + 19896.88012732740 *T)
    sumZ1  +=       1.203 * cos(0.38472694833  + 13367.97263110660 *T)
    sumZ1  +=       1.126 * cos(0.08391401839  + 14143.49524243060 *T)
    sumZ1  +=       0.867 * cos(6.12475712244  + 7860.41939243920 *T)
    sumZ1  +=       0.737 * cos(1.36372254530  + 9683.59458111640 *T)
    sumZ1  +=       0.673 * cos(2.25480471785  + 10021.83728009940 *T)
    sumZ1  +=       0.673 * cos(1.74966742013  + 11322.66409830440 *T)
    sumZ1  +=       0.654 * cos(0.52171368417  + 8624.21265092720 *T)
    sumZ1  +=       0.739 * cos(0.71134257549  + 8635.94200376320 *T)
    sumZ1  +=       0.592 * cos(2.89903060772  + 15720.83878487840 *T)
    sumZ1  +=       0.573 * cos(1.68337600461  + 20618.01935853360 *T)
    sumZ1  +=       0.509 * cos(1.21775184841  + 25934.12433108940 *T)
    sumZ1  +=       0.402 * cos(5.15486544320  + 9786.68735533500 *T)
    sumZ1  +=       0.522 * cos(3.93344256885  + 775.52261132400 *T)
    sumZ1  +=       0.360 * cos(0.93512209325  + 1059.38193018920 *T)
    sumZ1  +=       0.331 * cos(6.18983417760  + 11015.10647733480 *T)
    sumZ1  +=       0.390 * cos(0.55127102590  + 19367.18916223280 *T)
    sumZ1  +=       0.293 * cos(2.40304140450  + 801.82093112380 *T)
    sumZ1  +=       0.311 * cos(5.00264213987  + 10742.97651130560 *T)
    sumZ1  +=       0.303 * cos(1.76024438478  + 17298.18232732620 *T)
    sumZ1  +=       0.266 * cos(2.34165306562  + 9411.46461508720 *T)
    sumZ1  +=       0.336 * cos(2.31707280345  + 9999.98645077300 *T)
    sumZ1  +=       0.337 * cos(3.71507493522  + 10596.18207843420 *T)
    sumZ1  +=       0.238 * cos(3.48412611241  + 7058.59846131540 *T)
    sumZ1  +=       0.252 * cos(1.31468836511  + 10234.06094170340 *T)
    sumZ1  +=       0.225 * cos(5.45422332781  + 10192.51015071860 *T)
    sumZ1  +=       0.224 * cos(2.62853874690  + 23581.25817731760 *T)
    sumZ1  +=       0.245 * cos(5.76868176406  + 213.29909543800 *T)
    sumZ1  +=       0.240 * cos(4.88817120361  + 10206.17199921020 *T)
    sumZ1  +=       0.210 * cos(2.37015222881  + 4551.95349705880 *T)
    sumZ1  +=       0.198 * cos(0.00551647295  + 9830.38901398780 *T)
    sumZ1  +=       0.235 * cos(0.61645223949  + 29580.47470844380 *T)
    sumZ1  +=       0.190 * cos(5.02902672388  + 3930.20969621960 *T)
    sumZ1  +=       0.219 * cos(4.33456063754  + 15874.61759536320 *T)
    sumZ1  +=       0.171 * cos(3.12200105851  + 12566.15169998280 *T)
    sumZ1  +=       0.174 * cos(5.99537915416  + 26.29831979980 *T)
    sumZ1  +=       0.180 * cos(3.65160439089  + 21228.39202354580 *T)
    sumZ1  +=       0.172 * cos(2.65343695006  + 29864.33402730900 *T)
    sumZ1  +=       0.138 * cos(2.09789826211  + 3128.38876509580 *T)
    sumZ1  +=       0.138 * cos(6.13843515618  + 9161.01716302260 *T)
    sumZ1  +=       0.163 * cos(5.40807514906  + 4705.73230754360 *T)
    sumZ1  +=       0.125 * cos(3.63191548504  + 7.11354700080 *T)
    sumZ1  +=       0.146 * cos(1.77971392072  + 191.44826611160 *T)
    sumZ1  +=       0.099 * cos(5.66016472051  + 40853.14218484400 *T)
    sumZ1  +=       0.117 * cos(5.44747244034  + 10426.58464164900 *T)
    sumZ1  +=       0.095 * cos(1.51959002513  + 10218.80847051840 *T)
    sumZ1  +=       0.101 * cos(4.42330401713  + 10220.39909321180 *T)
    sumZ1  +=       0.100 * cos(5.54541542103  + 5661.33204915220 *T)
    sumZ1  +=       0.093 * cos(6.04217729850  + 20419.45754542119 *T)
    sumZ1  +=       0.082 * cos(5.21733543718  + 8662.24032356300 *T)
    sumZ1  +=       0.092 * cos(0.17084455774  + 9146.79006902100 *T)
    sumZ1  +=       0.085 * cos(4.60308698783  + 25158.60171976540 *T)
    sumZ1  +=       0.090 * cos(5.47130878539  + 37724.75341974820 *T)
    sumZ1  +=       0.074 * cos(1.54547841954  + 7084.89678111520 *T)
    sumZ1  +=       0.087 * cos(3.99154062553  + 33794.54372352860 *T)
    sumZ1  +=       0.091 * cos(0.63814500882  + 3154.68708489560 *T)
    sumZ1  +=       0.065 * cos(3.27192235582  + 1109.37855209340 *T)
    sumZ1  +=       0.089 * cos(1.74739410911  + 3532.06069281140 *T)
    sumZ1  +=       0.069 * cos(1.21744230443  + 10198.03307502600 *T)
    sumZ1  +=       0.059 * cos(1.45863866349  + 10207.76262190360 *T)
    sumZ1  +=       0.081 * cos(0.78654424919  + 20213.27199698400 *T)
    sumZ1  +=       0.056 * cos(0.16985541402  + 14919.01785375460 *T)
    sumZ1  +=       0.052 * cos(2.57737005542  + 22779.43724619380 *T)
    sumZ1  +=       0.061 * cos(0.72963813902  + 382.89653222320 *T)
    sumZ1  +=       0.052 * cos(3.49206526650  + 21535.94964451540 *T)
    sumZ1  +=       0.062 * cos(3.04943681345  + 28286.99048486120 *T)
    sumZ1  +=       0.056 * cos(2.00173927326  + 10228.53801739600 *T)
    sumZ1  +=       0.057 * cos(0.62155717729  + 41654.96311596780 *T)
    sumZ1  +=       0.048 * cos(0.21226498407  + 29050.78374334920 *T)
    sumZ1  +=       0.055 * cos(6.20318924174  + 18837.49819713819 *T)
    sumZ1  +=       0.056 * cos(2.35632004464  + 9103.90699411760 *T)
    sumZ1  +=       0.044 * cos(4.12172817384  + 27511.46787353720 *T)
    sumZ1  +=       0.039 * cos(6.15465517048  + 20400.27277262220 *T)
    sumZ1  +=       0.040 * cos(3.85673830469  + 11272.66747640020 *T)
    sumZ1  +=       0.053 * cos(1.80084365566  + 45585.17281218740 *T)
    sumZ1  +=       0.051 * cos(1.70380845885  + 426.59819087600 *T)
    sumZ1  +=       0.040 * cos(5.52599682847  + 522.57741809380 *T)
    sumZ1  +=       0.034 * cos(3.92659473780  + 4732.03062734340 *T)
    sumZ1  +=       0.036 * cos(5.94006075380  + 21202.09370374600 *T)
    sumZ1  +=       0.032 * cos(5.31625288376  + 3149.16416058820 *T)
    sumZ1  +=       0.039 * cos(1.26146315948  + 32217.20018108080 *T)
    sumZ1  +=       0.032 * cos(5.03769459810  + 536.80451209540 *T)
    sumZ1  +=       0.034 * cos(4.19685270992  + 13341.67431130680 *T)
    sumZ1  +=       0.031 * cos(1.20392434392  + 12592.45001978260 *T)
    sumZ1  +=       0.036 * cos(3.22618486892  + 49515.38250840700 *T)
    sumZ1  +=       0.031 * cos(6.15462442323  + 29088.81141598500 *T)
    sumZ1  +=       0.029 * cos(0.47911260035  + 1589.07289528380 *T)
    sumZ1  +=       0.027 * cos(1.23709297168  + 33019.02111220460 *T)
    sumZ1  +=       0.022 * cos(3.33096598037  + 20452.86941222180 *T)
    sumZ1  +=       0.028 * cos(0.89703139186  + 8094.52168583260 *T)
    sumZ1  +=       0.027 * cos(4.63672266528  + 3340.61242669980 *T)
    sumZ1  +=       0.024 * cos(5.94541303751  + 9929.42622734580 *T)
    sumZ1  +=       0.024 * cos(0.40689057274  + 10497.14486507620 *T)
    sumZ1  +=       0.017 * cos(2.88410168794  + 16496.36139620240 *T)
    sumZ1  +=       0.017 * cos(2.59067946967  + 10175.25787357520 *T)
    sumZ1  +=       0.017 * cos(3.76162101633  + 10251.31321884680 *T)
    sumZ1  +=       0.013 * cos(0.74065551442  + 24356.78078864160 *T)
    sumZ1  +=       0.014 * cos(5.04938942644  + 6681.22485339960 *T)

    # VENUS Z Coord
    sumZ2  = 0
    sumZ2  +=   16179.656 * cos(3.38519723238  + 10213.28554621100 *T)
    sumZ2  +=     305.549 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ2  +=      62.577 * cos(5.25189799706  + 20426.57109242200 *T)
    sumZ2  +=       0.850 * cos(0.09737945971  + 10186.98722641120 *T)
    sumZ2  +=       0.592 * cos(0.95308466065  + 30639.85663863300 *T)
    sumZ2  +=       0.573 * cos(0.56552197979  + 2352.86615377180 *T)
    sumZ2  +=       0.569 * cos(4.36593284769  + 10239.58386601080 *T)
    sumZ2  +=       0.438 * cos(3.18384728320  + 11790.62908865880 *T)
    sumZ2  +=       0.468 * cos(4.01105575472  + 1577.34354244780 *T)
    sumZ2  +=       0.356 * cos(6.12814990238  + 10988.80815753500 *T)
    sumZ2  +=       0.324 * cos(2.79347039249  + 10404.73381232260 *T)
    sumZ2  +=       0.277 * cos(6.08171605083  + 9437.76293488700 *T)
    sumZ2  +=       0.281 * cos(5.10638363837  + 9153.90361602180 *T)
    sumZ2  +=       0.208 * cos(3.14696384347  + 18073.70493865020 *T)
    sumZ2  +=       0.167 * cos(0.70476486230  + 529.69096509460 *T)
    sumZ2  +=       0.135 * cos(6.00194499808  + 19651.04848109800 *T)
    sumZ2  +=       0.145 * cos(5.41776609230  + 5507.55323866740 *T)
    sumZ2  +=       0.117 * cos(2.16955801658  + 6283.07584999140 *T)
    sumZ2  +=       0.121 * cos(1.25671231625  + 22003.91463486980 *T)
    sumZ2  +=       0.118 * cos(2.56442724367  + 9683.59458111640 *T)
    sumZ2  +=       0.074 * cos(4.57489383229  + 15720.83878487840 *T)
    sumZ2  +=       0.083 * cos(5.29424644857  + 13367.97263110660 *T)
    sumZ2  +=       0.059 * cos(3.28182139103  + 11322.66409830440 *T)
    sumZ2  +=       0.064 * cos(5.23563988056  + 10596.18207843420 *T)
    sumZ2  +=       0.060 * cos(2.73672115191  + 19896.88012732740 *T)
    sumZ2  +=       0.047 * cos(4.80208170598  + 7058.59846131540 *T)
    sumZ2  +=       0.042 * cos(2.36980390290  + 14143.49524243060 *T)
    sumZ2  +=       0.041 * cos(1.68366169373  + 801.82093112380 *T)
    sumZ2  +=       0.032 * cos(0.58444963462  + 10206.17199921020 *T)
    sumZ2  +=       0.035 * cos(3.21932079821  + 20618.01935853360 *T)
    sumZ2  +=       0.028 * cos(4.55053233088  + 9999.98645077300 *T)
    sumZ2  +=       0.028 * cos(5.84340580032  + 10220.39909321180 *T)
    sumZ2  +=       0.029 * cos(3.13488740448  + 9411.46461508720 *T)
    sumZ2  +=       0.024 * cos(4.56262782130  + 10742.97651130560 *T)
    sumZ2  +=       0.025 * cos(0.54239218952  + 11015.10647733480 *T)
    sumZ2  +=       0.020 * cos(4.83254240300  + 9830.38901398780 *T)
    sumZ2  +=       0.025 * cos(1.23260224812  + 10021.83728009940 *T)
    sumZ2  +=       0.019 * cos(5.15537838399  + 26.29831979980 *T)
    sumZ2  +=       0.022 * cos(2.31119555973  + 18849.22754997420 *T)
    sumZ2  +=       0.021 * cos(5.31579555879  + 775.52261132400 *T)
    sumZ2  +=       0.020 * cos(2.90919452314  + 3128.38876509580 *T)
    sumZ2  +=       0.023 * cos(2.08236048283  + 8635.94200376320 *T)
    sumZ2  +=       0.014 * cos(3.33582216202  + 1059.38193018920 *T)
    sumZ2  +=       0.009 * cos(3.25236762648  + 9103.90699411760 *T)
    sumZ2  +=       0.009 * cos(0.70904281096  + 7860.41939243920 *T)
    sumZ2  +=       0.010 * cos(2.21224313590  + 19367.18916223280 *T)
    sumZ2  +=       0.009 * cos(5.04826725887  + 7.11354700080 *T)
    sumZ2  +=       0.007 * cos(4.98138067490  + 10426.58464164900 *T)
    sumZ2  +=       0.007 * cos(4.31777952738  + 3154.68708489560 *T)

    # VENUS Z Coord
    sumZ3  = 0
    sumZ3  +=     467.466 * cos(4.99201374907  + 10213.28554621100 *T)
    sumZ3  +=      21.639 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ3  +=       2.002 * cos(0.76147646688  + 20426.57109242200 *T)
    sumZ3  +=       0.057 * cos(1.51447613604  + 10186.98722641120 *T)
    sumZ3  +=       0.042 * cos(5.70731176550  + 10239.58386601080 *T)
    sumZ3  +=       0.020 * cos(4.71416496562  + 11790.62908865880 *T)
    sumZ3  +=       0.019 * cos(5.61164867192  + 9437.76293488700 *T)
    sumZ3  +=       0.018 * cos(1.02002338896  + 10988.80815753500 *T)
    sumZ3  +=       0.018 * cos(2.08106024829  + 2352.86615377180 *T)
    sumZ3  +=       0.008 * cos(4.33056892256  + 10404.73381232260 *T)
    sumZ3  +=       0.007 * cos(3.07196293928  + 30639.85663863300 *T)
    sumZ3  +=       0.005 * cos(2.22179616897  + 1577.34354244780 *T)

    # VENUS Z Coord
    sumZ4  = 0
    sumZ4  +=      10.192 * cos(0.31526267335  + 10213.28554621100 *T)
    sumZ4  +=       0.233 * cos(3.14159265359  + 0.00000000000 *T)
    sumZ4  +=       0.069 * cos(2.32746005974  + 20426.57109242200 *T)

    # VENUS Z Coord
    sumZ5  = 0
    sumZ5  +=       0.173 * cos(2.05270625526  + 10213.28554621100 *T)
    sumZ5  +=       0.042 * cos(0.00000000000  + 0.00000000000 *T)
    sumZ5  +=       0.004 * cos(3.77138066766  + 20426.57109242200 *T)

    # calculate and scale the sums
    X = sumX0 + sumX1*T**1 + sumX2*T**2 + sumX3*T**3 + sumX4*T**4 + sumX5*T**5
    X /= 1e8
    Y = sumY0 + sumY1*T**1 + sumY2*T**2 + sumY3*T**3 + sumY4*T**4 + sumY5*T**5
    Y /= 1e8
    Z = sumZ0 + sumZ1*T**1 + sumZ2*T**2 + sumZ3*T**3 + sumZ4*T**4 + sumZ5*T**5
    Z /= 1e8

    # calculate Rau, Long Lat
    Rau = math.sqrt(X*X+Y*Y+Z*Z)
    Lon = Mod(degrees(atan2(Y, X)), 360.)
    Lat = degrees(atan2(Z, math.sqrt(X*X+Y*Y)))
    
    # create dictionary
    Dict = {"Object"  : "Venus",
            "Mode"  : "XYZ",
            "X"     : X,
            "Y"     : Y,
            "Z"     : Z,
            "Long"  : Lon,
            "Lat"   : Lat,
            "Rau"   : Rau,
            "Alt"   : -34./60.
           }
    Data = (X, Y, Z, Dict)


    return Data
#######################################################################



def getVenusData(dtCivicTime, zone, verbose = False):
    X, Y, Z, Dict = getVenusXYZData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    dummy, dummy, epsilon = getNutation(dtCivicTime, zone, verbose)
    # create coordinates relative to Earth: eX, eY, eZ, eL, eB, eRau
    if Dict["Object"] != "Earth":
        Xe, Ye, Ze, Dicte = getEarthXYZData(dtCivicTime, zone, verbose)
        eX = X - Xe; eY = Y - Ye; eZ = Z - Ze
        eRau = sqrt(eX*eX+eY*eY+eZ*eZ)
        eL = Mod(degrees(atan2(eY, eX)), 360.)
        eB = degrees(atan2(eZ, sqrt(eX*eX+eY*eY)))
        Dict["eX"]   = eX;   Dict["eY"]    = eY; Dict["eZ"]   = eZ
        Dict["eRau"] = eRau; Dict["eLong"] = eL; Dict["eLat"] = eB

        # Rektaszension, Deklination relativ zur Erde
        RA = atan2(Cos(epsilon)*Sin(eL) - Tan(eB)*Sin(epsilon), Cos(eL))
        Decl = asin(Sin(eB)*Cos(epsilon) + Cos(eB)*Sin(epsilon)*Sin(eL))
        RA = Mod(degrees(RA), 360.); Decl = degrees(Decl)
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
    else:   # Earth
        RA = 0; Decl = 0; 
        Dict["RA"] = RA
        Dict["Decl"] = Decl
        Dict["epsilon"] = epsilon
        return (RA, Decl, Dict)
#######################################################################


#######################################################################
#
#   getSunData, based on Meeus, chapter 24
#   Epoch J2000.0, JDE 2451545.0 (2000 Jan 1.5 TD)
#   Parameters:
#   dtCivicTime datetime object, date and time in local time zone
#   zone        time zone offset in hours, positive EAST
#   verbose     default False, used for debugging
#
#   Returns:
#   SunRA       Rektaszension in degrees
#   SunDecl     Declination above equator in degrees
#   SunDict     Dictionary with Sun-specific output values:
#       Object:     "Sun"
#       Ltrue       True (ecliptic) Longitude in degrees
#       Lattrue     True ecliptic latitude, always 0.0
#       R           Distance Earth - Sun in km
#       epsilon     Schiefe der Ekliptik (oblique) in degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getSunData.py"
VersionList.append(Version)

def getSunData(dtCivicTime, zone, verbose=False):
    # use VSOP87 data for Earth
    # use getNutation to include nutation
    deltaLong, deltaLat, epsilon = getNutation(dtCivicTime, zone, False)    
    # we do not need the nutation effects, just epsilon
    deltaLong = 0; deltaLat = 0

    # ecliptic latitude of Earth, chapter 31 - use negative value for Sun,
    # add 180 to longitude
    # Meeus p. 154
    dummy, dummy, Dict = getEarthData(dtCivicTime, zone, verbose)
    L = Dict["Long"]
    B = Dict["Lat"]
    SunLat = -B + deltaLat
    Ltrue = Mod(L + deltaLong + 180., 360.)
    Rau = Dict["Rau"]
    Rabs = Rau * 149597870.700    # km

    # Rektaszension, Deklination
    SunRA = atan2(Cos(epsilon)*Sin(Ltrue) - Tan(SunLat)*Sin(epsilon), \
            Cos(Ltrue))
    SunDecl = asin(Sin(SunLat)*Cos(epsilon) \
            + Cos(SunLat)*Sin(epsilon)*Sin(Ltrue))
    SunRA = Mod(degrees(SunRA), 360.)
    SunDecl = degrees(SunDecl)

    # altitude at Sunrise, Sunset (degrees), standard value
    Alt = -50./60.    # degrees, 50'

    if (verbose):
        print ("SunData:")
        print ("Rau = ", Rau)
        print ("Rabs = ", Rabs)
        print ("epsilon = ", epsilon)
        print ("SunRA = ", SunRA)
        print ("SunDecl = ", SunDecl)
        print ("Alt = ", Alt)
#       Object:     "Sun"
#       SunLong     Ltrue   True (ecliptic) Longitude in degrees
#       SunLat      True ecliptic latitude, always 0.0
#       R           Rabs    Distance Earth - Sun in km
#       epsilon     Schiefe der Ekliptik (oblique) in degrees
#       Alt         Altitude at rise, set (degrees)
    SunDict = { "Object" : "Sun",
                "Long"  : Ltrue,
                "SunLong"  : Ltrue,
                "Lat"   : SunLat,
                "SunLat"   : SunLat,
                "RA"    : SunRA,
                "Decl"  : SunDecl,
                "R"       : Rabs,
                "Rau"     : Rau,
                "epsilon" : epsilon,
                "Alt"     : Alt,
                "X"     : 0,
                "Y"     : 0,
                "Z"     : 0,
                "eX"     : -Dict["X"],
                "eY"     : -Dict["Y"],
                "eZ"     : -Dict["Z"],
                "eRau"     : Rau,
                "eLong"  : Ltrue,
                "eLat"   : SunLat,
                }
    return (SunRA, SunDecl, SunDict)

############## End of getSunData() #####################################
########################################################################
#
#   getMoonData(dtCivicTime, zone, verbose=False):
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       MoonDist    distance Earth - Moon, km
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonHP      horizontal parallax, degrees
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)

Version = "V4.1 hvf 2022-05-19 getMoonData.py"
VersionList.append(Version)

def getMoonData(dtCivicTime, zone, verbose=False):
    # calculate Ecliptic Longitude (degrees)
    Epoch2000 = 2451545.0
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.

    T = TT / 36525.    # eq. 21.1

    # obliquity of the ecliptic, eq. 21.2, degrees
    epsilon = 23.+26./60+21.448/3600. - 46.8150/3600.*T \
        - 0.00059/3600.*T*T + 0.001813/3600.*T*T*T

    # Moon mean longitude (mean equinox of the date, incl, effect of light-time)
    # eq. 45.1
    Lprime  = 218.3164591 + T * (481267.88134236 + T * (-0.0013268 \
            + T * (1./538841. - T / 65194000)))

    # Mean elongation of the Moon
    # eq. 45.2
    D       = 297.8502042 + T * (445267.1115168 + T * (-0.0016300 \
            + T * (1./545868.  -T / 113065000.)))

    # Sun's mean anomaly
    # eq. 45.3
    M       = 357.5291092 + T * (35999.0502909 + T * (-0.0001536 \
            + T / 24490000))

    # Moon's mean anomaly
    # eq 45.4
    MM      = 134.9634114 + T * (477198.8676313 + T * (0.0089970 \
            + T * (1./69699.  - T / 14712000.)))

    # Moon's argument of latitude (mean distance of the Moon
    # from its ascending node)
    # eq 45.5
    F       = 93.2720993 + T * (483202.0175273 + T * (-0.0034029 \
            + T * (-1./3526000. + T / 863310000.)))

    # three further arguments are needed
    A1      = 119.75 + 131.849 * T          # action of Venus
    A2      =  53.09 + 479264.290 * T       # action of Jupiter
    A3      = 313.45 + 481266.484 * T

    # effect of decreasing eccentricity of Earth's orbit
    E       = 1. + T * (-0.002516 + T * -0.0000074)
    EE      = E*E

    # Longitude, distance: now calculate sums sigL, sigR
    # see Table 45.A, p.309
    arg = MM;           sigL = 6288774*Sin(arg);    sigR = -20905355*Cos(arg)
    arg = 2*D - MM;     sigL += 1274027*Sin(arg);   sigR += -3699111*Cos(arg)
    arg = 2*D;          sigL += 658314*Sin(arg);    sigR += -2955968*Cos(arg)
    arg = 2*MM;         sigL += 213618*Sin(arg);    sigR += -569925*Cos(arg)  
    arg = M;            sigL += -185116*Sin(arg)*E; sigR += 48888*Cos(arg)*E
    arg = 2*F;          sigL += -114332*Sin(arg);   sigR += -3149*Cos(arg)
    arg = 2*D - 2*MM;   sigL += 58793*Sin(arg);     sigR += 246158*Cos(arg)
    arg = 2*D - M - MM; sigL += 57066*Sin(arg)*E;   sigR += -152138*Cos(arg)*E
    arg = 2*D + MM;     sigL += 53322*Sin(arg);     sigR += -170733*Cos(arg)
    arg = 2*D - M;      sigL += 45758*Sin(arg)*E;   sigR += -204586*Cos(arg)*E
    arg = M - MM;       sigL += -40923*Sin(arg)*E;  sigR += -129620*Cos(arg)*E
    arg = D;            sigL += -34720*Sin(arg);    sigR += 108743*Cos(arg)
    arg = M + MM;       sigL += -30383*Sin(arg)*E;  sigR +=  104755*Cos(arg)*E
    arg = 2*D - 2*F;    sigL += 15327*Sin(arg);     sigR += 10321*Cos(arg)
    arg = MM + 2*F;     sigL += -12528*Sin(arg)    
    arg = MM - 2*F;     sigL += 10980*Sin(arg);     sigR += 79661*Cos(arg)
    arg = 4*D-MM;       sigL += 10675*Sin(arg);     sigR += -34782*Cos(arg)
    arg = 3*MM;         sigL += 10034*Sin(arg);     sigR += -23210*Cos(arg)
    arg = 4*D-2*MM;     sigL +=  8548*Sin(arg);     sigR += -21636*Cos(arg)
    arg = 2*D + M - MM; sigL += -7888*Sin(arg)*E;   sigR += 24208*Cos(arg)*E
    arg = 2*D + M ;     sigL += -6766*Sin(arg)*E;   sigR += 30824*Cos(arg)*E
    arg = D - MM;       sigL += -5163*Sin(arg);     sigR += -8379*Cos(arg)
    arg = D + M;        sigL += 4987*Sin(arg)*E;    sigR += -16675*Cos(arg)*E
    arg = 2*D - M + MM; sigL += 4036*Sin(arg)*E;    sigR += -12831*Cos(arg)*E
    arg = 2*D + 2*MM;   sigL += 3994*Sin(arg);      sigR += -10445*Cos(arg)
    arg = 4*D;          sigL += 3861*Sin(arg);      sigR += -11650*Cos(arg)
    arg = 2*D - 3*MM;   sigL += 3665*Sin(arg);      sigR += 14403*Cos(arg)
    arg = M - 2*MM;     sigL += -2689*Sin(arg)*E;   sigR += -7003*Cos(arg)*E
    arg = 2*D - MM + 2*F; sigL += -2602*Sin(arg)    
    arg = 2*D - M - 2*MM; sigL += 2390*Sin(arg)*E;  sigR += 10056*Cos(arg)*E
    arg = D + MM;       sigL += -2348*Sin(arg);     sigR += 6322*Cos(arg)
    arg = 2*D - 2*M ;   sigL += 2236*Sin(arg)*EE;   sigR += -9884*Cos(arg)*EE
    arg = M + 2*MM;     sigL += -2120*Sin(arg)*E;   sigR += 5751*Cos(arg)*E
    arg = 2*M ;         sigL += -2069*Sin(arg)*EE
    arg = 2*D - 2*M - MM; sigL += 2048*Sin(arg)*EE; sigR += -4950*Cos(arg)*EE
    arg = 2*D + MM - 2*F;   sigL += -1773*Sin(arg); sigR += 4130*Cos(arg)
    arg = 2*D + 2*F;    sigL += -1595*Sin(arg)
    arg = 4*D - M - MM; sigL += 1215*Sin(arg)*E;    sigR += -3958*Cos(arg)*E
    arg = 2*MM + 2*F;   sigL += -1110*Sin(arg)
    arg = 3*D - MM;     sigL += -892*Sin(arg);      sigR += 3258*Cos(arg)
    arg = 2*D + M + MM; sigL += -810*Sin(arg)*E;    sigR += 2616*Cos(arg)*E
    arg = 4*D - M - 2*MM; sigL += 759*Sin(arg)*E;   sigR += -1897*Cos(arg)*E
    arg = 2*M - MM;     sigL += -713*Sin(arg)*EE;   sigR += -2117*Cos(arg)*EE
    arg = 2*D + 2*M - MM; sigL += -700*Sin(arg)*EE; sigR += 2354*Cos(arg)*EE
    arg = 2*D + M - 2*MM;   sigL += 691*Sin(arg)*E
    arg = 2*D - M - 2*F;    sigL += 596*Sin(arg)*E
    arg = 4*D + MM;     sigL += 549*Sin(arg);       sigR += -1423*Cos(arg)
    arg = 4*MM;         sigL += 537*Sin(arg);       sigR += -1117*Cos(arg)
    arg = 4*D - M;      sigL += 520*Sin(arg)*E;     sigR += -1571*Cos(arg)*E
    arg = D - 2*MM;     sigL += -487*Sin(arg);      sigR += -1739*Cos(arg)
    arg = 2*D + M - 2*F; sigL += -399*Sin(arg)*E
    arg = 2*MM - 2*F;   sigL += -381*Sin(arg);      sigR += -4421*Cos(arg)
    arg = D + M + MM ;  sigL += 351*Sin(arg)*E
    arg = 3*D - 2*MM;   sigL += -340*Sin(arg)
    arg = 4*D - 3*MM;   sigL += 330*Sin(arg)
    arg = 2*D - M + 2*MM ;  sigL += 327*Sin(arg)*E
    arg = 2*M + MM;     sigL += -323*Sin(arg)*EE;   sigR += 1165*Cos(arg)*EE
    arg = D + M - MM ;  sigL += 299*Sin(arg)*E
    arg = 2*D + 3*MM;   sigL += 294*Sin(arg)
    arg = 2*D - MM - 2*F;                           sigR += 8752*Cos(arg)

    # Latitude: calculate sumB, see Table 45.B, p.311
    arg = F;                    sigB = 5128122*Sin(arg)
    arg = MM + F;               sigB += 280602*Sin(arg)
    arg = MM - F;               sigB += 277693*Sin(arg)
    arg = 2*D - F;              sigB += 173237*Sin(arg)
    arg = 2*D - MM + F;         sigB += 55413*Sin(arg)
    arg = 2*D - MM - F;         sigB += 46271*Sin(arg)
    arg = 2*D + F;              sigB += 32573*Sin(arg)
    arg = 2*MM + F;             sigB += 17198*Sin(arg)
    arg = 2*D + MM - F;         sigB += 9266*Sin(arg)
    arg = 2*MM - F;             sigB += 8822*Sin(arg)
    arg = 2*D - M - F;          sigB += 8216*Sin(arg)*E
    arg = 2*D - 2*MM - F;       sigB += 4324*Sin(arg)
    arg = 2*D + MM + F;         sigB += 4200*Sin(arg)
    arg = 2*D + M - F;          sigB += -3359*Sin(arg)*E
    arg = 2*D - M - MM + F;     sigB += 2463*Sin(arg)*E
    arg = 2*D - M + F;          sigB += 2211*Sin(arg)*E
    arg = 2*D - M - MM - F;     sigB += 2065*Sin(arg)*E
    arg = M - MM - F;           sigB += -1870*Sin(arg)*E
    arg = 4*D - MM - F;         sigB += 1828*Sin(arg)
    arg = M + F;                sigB += -1794*Sin(arg)*E
    arg = 3*F;                  sigB += -1749*Sin(arg)
    arg = M - MM + F;           sigB += -1565*Sin(arg)*E
    arg = D + F;                sigB += -1491*Sin(arg)
    arg = M + MM + F;           sigB += -1475*Sin(arg)*E
    arg = M + MM - F;           sigB += -1410*Sin(arg)*E
    arg = M - F;                sigB += -1344*Sin(arg)*E
    arg = D - F;                sigB += -1335*Sin(arg)
    arg = 3*MM + F;             sigB += 1107*Sin(arg)
    arg = 4*D - F;              sigB += 1021*Sin(arg)
    arg = 4*D - M + F;          sigB += 833*Sin(arg)
    arg = MM - 3*F;             sigB += 777*Sin(arg)
    arg = 4*D - 2*MM + F;       sigB += 671*Sin(arg)
    arg = 2*D - 3*F;            sigB += 607*Sin(arg)
    arg = 2*D + 2*MM - F;       sigB += 596*Sin(arg)
    arg = 2*D - M + MM - F;     sigB += 491*Sin(arg)*E
    arg = 2*D - 2*MM + F;       sigB += -451*Sin(arg)
    arg = 3*MM - F;             sigB += 439*Sin(arg)
    arg = 2*D + 2*MM + F;       sigB += 422*Sin(arg)
    arg = 2*D - 3*MM - F;       sigB += 421*Sin(arg)
    arg = 2*D + M - MM + F;     sigB += -366*Sin(arg)*E
    arg = 2*D + M + F;          sigB += -351*Sin(arg)*E
    arg = 4*D + F;              sigB += 331*Sin(arg)
    arg = 2*D - M + MM + F;     sigB += 315*Sin(arg)*E
    arg = 2*D - 2*M - F;        sigB += 302*Sin(arg)*EE
    arg = MM + 3*F;             sigB += -283*Sin(arg)
    arg = 2*D + M + MM - F;     sigB += -229*Sin(arg)*E
    arg = D + M - F;            sigB += 223*Sin(arg)*E
    arg = D + M + F;            sigB += 223*Sin(arg)*E
    arg = M - 2*MM - F;         sigB += -220*Sin(arg)*E
    arg = 2*D + M - MM - F;     sigB += -220*Sin(arg)*E
    arg = D + MM + F;           sigB += -185*Sin(arg)
    arg = 2*D - M - 2*MM - F;   sigB += 181*Sin(arg)*E
    arg = M + 2*MM + F;         sigB += -177*Sin(arg)*E
    arg = 4*D - 2*MM - F;       sigB += 176*Sin(arg)
    arg = 4*D - M - MM - F;     sigB += 166*Sin(arg)*E
    arg = D + MM - F;           sigB += -164*Sin(arg)
    arg = 4*D + MM - F;         sigB += 132*Sin(arg)
    arg = D - MM - F;           sigB += -119*Sin(arg)
    arg = 4*D - M - F;          sigB += 115*Sin(arg)*E
    arg = 2*D - 2*M + F;        sigB += 107*Sin(arg)*EE

    # Additives to sigL (p.112)
    sigL +=   3958*Sin(A1) \
            + 1962*Sin(Lprime-F) \
            + 318*Sin(A2)

    # Additives to sigB
    sigB += -2235*Sin(Lprime) \
            + 382*Sin(A3) \
            + 175*Sin(A1-F) \
            + 175*Sin(A1+F) \
            + 127*Sin(Lprime-MM) \
            - 115*Sin(Lprime+MM)

    MoonLong = Lprime + sigL/1e6                # degrees
    MoonLong = Mod(MoonLong, 360.)              # degrees, normnalized
    MoonLat  = sigB/1e6                         # degrees
    MoonDist = 385000.56 + sigR/1e3             # km
    MoonHP   = degrees(asin(6378.14/MoonDist))  # degrees
    MoonDistAU = MoonDist / 149597870.700       # au
    X = MoonDistAU * Cos(MoonLat) * Cos(MoonLong) # au, geocentric X
    Y = MoonDistAU * Cos(MoonLat) * Sin(MoonLong) # au, geocentric y
    Z = MoonDistAU * Sin(MoonLat)               # au, geocentric Z

    # Lunar Node and lunar Perigee
    # p.313, eq. 45.7

    MoonNode = 125.0445550 + T * (-1934.1361849 + T * (0.0020762 \
        + T * (1./467410. - T / 60616000.)))
    MoonNode = Mod(MoonNode, 360.)
    MoonPerigee = 83.3532430 + T * (4069.0137111 + T * (-0.0103238 \
        + T * (-1./80053 + T / 18999000.)))
    MoonPerigee = Mod(MoonPerigee, 360.)

    MoonRA = atan2(Sin(MoonLong)*Cos(epsilon)-Tan(MoonLat)*Sin(epsilon), \
                Cos(MoonLong))
    MoonDecl = asin(Sin(MoonLat)*Cos(epsilon) \
            + Cos(MoonLat)*Sin(epsilon)*Sin(MoonLong))
    MoonRA = Mod(degrees(MoonRA), 360.); MoonDecl = degrees(MoonDecl)

    # Moon altitude at rise, set (see p.98)
    Alt = 0.7275 * MoonHP - 34./60. # degrees

    if verbose:
        print ("JD  = ", JD)
        print ("T   = ", T)
        print ("Lprime = ", Deg(Lprime))
        print ("D   = ", Deg(D))
        print ("M   = ", Deg(M))
        print ("MM  = ", Deg(MM))
        print ("F   = ", Deg(F))
        print ("A1  = ", Deg(A1))
        print ("A2  = ", Deg(A2))
        print ("A3  = ", Deg(A3))
        print ("E   = ", E)
        print ("sigL= %.1f" % sigL)
        print ("sigB= %.1f" % sigB)
        print ("sigR= %.1f" % sigR)
        print ("MoonLong    = %10.6f degrees" % MoonLong)
        print ("MoonLat     = %10.6f degrees" % MoonLat )
        print ("MoonDist    =   %.1f km" % MoonDist)
        print ("MoonHP      = %10.6f degrees" % MoonHP  )
        print ("MoonNode    = %10.6f degrees" % MoonNode)
        print ("MoonPerigee = %10.6f degrees" % MoonPerigee)
        print ("MoonRA      = %10.6f degrees" % MoonRA     )
        print ("MoonDecl    = %10.6f degrees" % MoonDecl)
        print ("Alt         = %10.6f degrees" % Alt     )
#   Returns:    
#   MoonRA      Rektaszension in degrees
#   MoonDecl    Declination above equator in degrees
#   MoonDict    Dictionary with Moon-specific output values:
#       Object:     "Moon"
#       MoonLong    ecliptic longitude, degrees
#       MoonLat     ecliptic latitude, degrees
#       RA          Rektaszension
#       Decl        Deklination
#       MoonDist    distance Earth - Moon, km
#       MoonHP      horizontal parallax, degrees
#       MoonDistAU  distance Earth - Moon, au
#       X           geocentric X, au
#       Y           geocentric Y, au
#       Z           geocentric Z, au
#       MoonNode    ecliptic longitude of ascending node, degrees
#       MoonPerigee longitude of perigee og Moon's orbit, degrees
#       Alt         Altitude at rise, set (degrees)
#       suffix      "++" if MoonDist <= 360000, "--" if MoonDist >= 405000
    suffix = '  '
    if MoonDist <= 360000. : suffix = "++" # Super Moon
    if MoonDist >= 405000. : suffix = "--" # Mini Moon

    MoonDict = {
        "Object"    : "Moon",
        "MoonLong"  : MoonLong,
        "MoonLat"   : MoonLat,
        "RA"        : MoonRA,
        "Decl"      : MoonDecl,
        "MoonDist"  : MoonDist,
        "MoonHP"    : MoonHP,
        "MoonDistAU": MoonDistAU,
        "X"         : X,
        "Y"         : Y,
        "Z"         : Z,
        "MoonNode"  : MoonNode,
        "MoonPerigee": MoonPerigee,
        "Alt"       : Alt,
        "suffix"    : suffix}
    

    return (MoonRA, MoonDecl, MoonDict)

################ End of getMoonData() #################################
Version = "V5.1 hvf 2023-03-13 getObjectTimes.py"
VersionList.append(Version)
# steps of 2 hours except for Sun and Moon
# I realized that Moon south passing on 2022-05-14 for Singapur gave wrong
# azimut values, therefore I change stepsize for Sun and Moon to 15 minutes
# hvf 16.05.2022
# third order Chebysheff polynomial interpolation
# use hvfTime - hvf 09.11.2022
# interpolate x and y instead of azimut, avoids problem near Zenit
# e.g. Singapur 2023-03-09
# hvf 13.03.2023

# important libraries
import sys
# Python Cookbook p.22 sorting a list of dicts by a common key
from operator import itemgetter
from datetime import datetime, date, timedelta
import numpy as np
import scipy.optimize


# some small functions
def isUp(alt, Alt):
    return alt > Alt
def isDown(alt, Alt):
    return alt <= Alt
def isNE(azimut):
    return azimut >= 0 and azimut < 90.
def isSE(azimut):
    return azimut >= 90. and azimut < 180.
def isSW(azimut):
    return azimut >= 180. and azimut < 270.
def isNW(azimut):
    return azimut >= 270. and azimut < 360.
# using x, y, z coords: see https://de.wikipedia.org/wiki/Astronomische_Koordinatensysteme#Umrechnungen
# x points South, y points West, z points Zenit
def isNExy(x, y):
    return (x <= 0. and y <= 0.)
def isSExy(x, y):
    return (x > 0. and y <= 0.)
def isSWxy(x, y):
    return (x > 0. and y > 0.)
def isNWxy(x, y):
    return (x <= 0. and y > 0.)
def isNxy(x, y):
    return (x <= 0.)
def isSxy(x, y):
    return (x > 0.)
def isExy(x, y):
    return (y <= 0.)
def isWxy(x, y):
    return (y > 0.)

# create coefficients for a polynome through (xi, yi)
def getPolyCoeffs(xi, yi):
    # return coeffs of polynom of degree n-1 going through (xi, yi)
    xarr = np.array(xi)
    yarr = np.array(yi)
    n = len(xarr)
    if n != len(yarr):
        print ("arrays xi and yi differ in length: ", n, len(yarr))
        sys.exit(1)
    A = np.ndarray((n, n))
    for row in range(n):
        for col in range(n):
            a = xarr[col]**row
            A[row, col] = a
    Ainv = np.linalg.inv(A)
    coeffs = np.dot(yarr, Ainv)
    check = np.dot(coeffs, A)
    delta = yarr-check
    return coeffs

# return value of polynom with coefficients coeffs for value x
def getPoly(x, coeffs):
    # return f(x), value of polynom with coefficients coeffs at value x
    f = 0.0
    for n in range(len(coeffs)):
        f += x**n * coeffs[n]
    return f

# function getLMST() calculates local sideral time
# for given civic time dt and time zone and longitude
def getLMST(dt, zone, longitude, verbose = False):
    # day at 00:00 civic time
    dtUTC = getUTfromCivicTime(dt, zone)
    JD = getJulianDay(dtUTC)
    GreenwichST = getGMSTfromJD(JD)
    LocalST = getLMSTfromGMST(GreenwichST, longitude)
    if verbose:
        print ("Day            = ", nicedatetime(dt))
        print ("UTC            = ", nicedatetime(dtUTC))
        print ("JD             = ", JD)
        print ("Greenwich Sideral Time = ", nicedatetime (GreenwichST))
        print ("Local     Sideral Time = ", nicedatetime (LocalST))
        print ("Longitude      = ", longitude)
    return LocalST

# get azimut and altitude for given civic time, zone, object data, 
#   longitude, latitude
def getAziAltfromCivicTimeLongLat(dtCivicTime, zone, getData, 
        longitude, latitude, verbose = False):

    (RA, Decl, Dict) = getData(dtCivicTime, zone, verbose)
    localST = getLMST(dtCivicTime, zone, longitude, verbose)
    RAh = RA/15.    # convert to hours
    HourAngle = localST - timedelta(hours=RAh)   # hour angle
    # use hvfTime()
    (h, m, s) = HourAngle.getTime()
    HourAngledecimal = h + m/60. + s/3600.
    HourAngledecimalstring = "%02d:%02d:%02d" % (h, m, s)

    #HourAngledecimal = HourAngle.time().hour + HourAngle.time().minute/60. \
    #    + HourAngle.time().second/3600 + HourAngle.time().microsecond/3600e6
    HourAnglegrad = HourAngledecimal*15.
    # get azimut, altitude from hour angle, declination, latitude
    (azi, alt, xyzDict) = getAziAltfromHourDecl(HourAnglegrad, Decl, latitude)

    verbose = False
    if verbose:
        print ("getAziAltfromCivicTimeLongLat:")
        print ("    dt = ", nicedatetime(dtCivicTime), ", zone = ", zone)
        print ("    Declination, latitude = ", round(Decl, 2),round(latitude,2))
        print ("    Hour Angle (deg) = ", HourAnglegrad)
        print ("    Azi, alt = ", round(azi, 2), round(alt, 2))
        print ("    x, y, z    = ", xyzDict["x"], xyzDict["y"], xyzDict["z"])
        print ("    xx, yy, zz = ", xyzDict["xx"], xyzDict["yy"], xyzDict["zz"])
    verbose = False

    # add xyzDict to Dict
    Dict["xyzDict"] = xyzDict
    return (azi, alt, Dict)

# interpolate altitude between times dt1 and dt2 for zone, use Npoints
# Chebychev points between dt1 and dt2
# only used for rise and set, where Alt is given in the object Dict (Moon)
# Alt is None if Dict["Alt"] value is used, else Alt is used (for twilight)
def interpolateAlt(dt1, dt2, zone, Npoints, getData, Alt,
        longitude, latitude, verbose = False):

    altList = []
    xList = []
    DictList = []
    AltList = []
    dtspan = (dt2-dt1)/2
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude)
        altList.append(alt)
        xList.append(xj)
        if Alt is None:
            AltList.append(Dict["Alt"])
        else:
            AltList.append(Alt)
            

    # create polynom for altitude and target Altitude
    # Altitude varies for Moon only
    coeffs = getPolyCoeffs(xList, altList)
    altCoeffs = getPolyCoeffs(xList, AltList)

    # apply target Alt - Alt varies for Moon only
    coeffs -= altCoeffs

    # get zero of polynom - catch ValueError
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # extend interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("xList = ", xList)
        print ("AltList = ", AltList)
        print ("coeffs = ", coeffs)
        print ("altCoeffs = ", altCoeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero

# interpolate azimut between times dt1 and dt1 for zone, use Npoints
# Chebychev points between dt1 and dt2
def interpolateAzi(dt1, dt2, zone, targetAzi, Npoints, getData,
        longitude, latitude, verbose = False):

    aziList = []
    xList = []
    DictList = []
    dtspan = (dt2-dt1)/2
    #print ("interpolateAzi dt1, dt2 = ", dt1, dt2)
    #print ("interpolateAzi dtspan = ", dtspan, type(dtspan))
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen, zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        #print("j ", j, "dt ", nicedatetime(dt))
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude, verbose)
        if targetAzi >= 0. and targetAzi < 45.:
            if azi > 225.: azi -= 360.
        aziList.append(azi)
        xList.append(xj)

    # create polynom
    coeffs = getPolyCoeffs(xList, aziList)

    # apply target Azimut
    coeffs[0] -= targetAzi

    # get zero of polynom, catch ValueError (Clansayes, 2022-06-02)
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # we need to enlarge the interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("targetAzi = ", targetAzi)
        print ("xList = ", xList)
        print ("aziList = ", aziList)
        print ("coeffs = ", coeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero

# interpolate y (transit over meridian)  between times dt1 and dt1 for zone, i
# use Npoints Chebychev points between dt1 and dt2
def interpolateY(dt1, dt2, zone, targetY, Npoints, getData,
        longitude, latitude, verbose = False):

    YList = []
    xList = []  # Stuetzstellen
    DictList = []
    dtspan = (dt2-dt1)/2
    #print ("interpolateY dt1, dt2 = ", dt1, dt2)
    #print ("interpolateY dtspan = ", dtspan, type(dtspan))
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen, zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        #print("j ", j, "dt ", nicedatetime(dt))
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude, verbose)
        xtemp = Dict["xyzDict"]["x"]   
        ytemp = Dict["xyzDict"]["y"]   
        YList.append(ytemp)
        xList.append(xj)

    # create polynom
    coeffs = getPolyCoeffs(xList, YList)

    # apply target Azimut
    coeffs[0] -= targetY

    # get zero of polynom, catch ValueError (Clansayes, 2022-06-02)
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # we need to enlarge the interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("targetY = ", targetY)
        print ("xList = ", xList)
        print ("YList = ", YList)
        print ("coeffs = ", coeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero

# interpolate x (transit over East-West)  between times dt1 and dt1 for zone,
# use Npoints Chebychev points between dt1 and dt2
def interpolateX(dt1, dt2, zone, targetX, Npoints, getData,
        longitude, latitude, verbose = False):

    XList = []
    xList = []  # Stuetzstellen
    DictList = []
    dtspan = (dt2-dt1)/2
    #print ("interpolateX dt1, dt2 = ", dt1, dt2)
    #print ("interpolateX dtspan = ", dtspan, type(dtspan))
    dtmean = dt1 + dtspan
    for j in range(Npoints):
        # Tschebyscheff Stuetzstellen, zwischen -1 und +1
        xj = cos((2.*j+1.)/(2.*Npoints)*pi)
        dt = dtmean + dtspan * xj       # zwischen dt1 und dt2
        #print("j ", j, "dt ", nicedatetime(dt))
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
            getData, longitude, latitude, verbose)
        xtemp = Dict["xyzDict"]["x"]   
        ytemp = Dict["xyzDict"]["y"]   
        XList.append(xtemp)
        xList.append(xj)

    # create polynom
    coeffs = getPolyCoeffs(xList, XList)

    # apply target Azimut
    coeffs[0] -= targetX

    # get zero of polynom, catch ValueError (Clansayes, 2022-06-02)
    try:
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1., 1.), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
    except ValueError:  # we need to enlarge the interval a little bit
        xzero = scipy.optimize.root_scalar(getPoly, args=coeffs, \
                method=None, bracket=(-1.01, 1.01), fprime=None, fprime2=None, \
                x0=0, xtol=None, rtol=None, maxiter=None, \
                options=None)
        if verbose:
            fa = getPoly(-1.0, coeffs)
            fb = getPoly(1.0, coeffs)
            print ("f(a), f(b) = ", fa, fb)
            print ("xzero = ", xzero.root)

    dtzero = dtmean + dtspan*xzero.root
    if verbose:
        print ("targetX = ", targetX)
        print ("xList = ", xList)
        print ("XList = ", XList)
        print ("coeffs = ", coeffs)
        print ("xzero = ", xzero)
        print ("xzero.root = ", xzero.root)
        print ("dtzero = ", dtzero)

    return dtzero


#######################################################################
#   getObjectTimes  Parameters:
#   dtCivicTime hvfTime object containing date and time of request
#   longitude   geographic longitude, degrees, positive if East
#   latitude    geographic latitude, degrees, positive if North
#   zone        time zone, hours East of Greenwich
#   getData     Function that supplies the necessary parameters
#               like getMoonData, getSunData
#   getData Returns:
#       RA       Rektaszension in degrees
#       Decl     Declination above equator in degrees
#       Dict     Dictionary with object-specific output values
#   Alt         altitude above horizon defining rise and set,
#               default value is None, Alt is taken from Object Dict
#   verbose     default False, use if debugging output is required
#
#   getObjectTimes returns a List of dictionaries with events containing: 
#   Object:     Dict["Object"],  # Sun, Moon, planets
#   dt    :     dt,     # hvfTime object with date and time of event
#   azi   :     azi,    # azimut of object at time of event, degrees
#   alt   :     alt,    # altitude above horizon of event, degrees
#   event :             # event type, see below
#   Alt   :     Dict["Alt"],    # altitude at time of event, taken from Dict
#   Dict  :     Dict,   # object dictionary got from getData
#   Event types:                    
#   "Rise"                rise time
#   "Set"                 set time 
#   "East  Passing (N-S)" time at first vertical (East, azimut 90 degrees)
#   "East  Passing (S-N)" time at first vertical (East, azimut 90 degrees)
#   "Transit South (E-W)" time of transit on South meridian, above horizon
#   "Transit South (W-E)" time of transit on South meridian, above horizon
#   "Transit North (E-W)" time of transit on North meridian, above horizon
#   "Transit North (W-E)" time of transit on North meridian, above horizon
#   "West  Passing (S-N)" time at second vertical (West, azimut 270 degrees)
#   "West  Passing (N-S)" time at second vertical (West, azimut 270 degrees)
#   "North Passing (W-E)" time of transit on North meridian, below horizon
#   "North Passing (E-W)" time of transit on North meridian, below horizon
#   "South Passing (E-W)" time of transit on South meridian, below horizon
#   "South Passing (W-E)" time of transit on South meridian, below horizon
#
#   all quantities are not defined (not in the dict) if there is no such event
#   NB. calling sequence is identical to getRTS()
#
#   The returned List of Dictionaries is sorted by the event time
#
#   Method: interpolation on the specified day
#
#   hvf 03.07.2021 09.06.2022
#   interpolate x, y (horizontal coords, x pointing South, y West)
#   avoids inaccuacies near Zenit
#   hvf 13.03.2023§

def getObjectTimes(dtCivicTime, longitude, latitude, zone, getData, 
                    Alt=None, verbose=False):
    
    if verbose: print ("getObjectTimes dt = ", dtCivicTime, type(dtCivicTime))
    # North Pole, South Pole: avoid singularities
    if latitude >=  90: latitude =  89.99
    if latitude <= -90: latitude = -89.99

    # create list of object data during 24 hours, steps+1 elements

    #dtMidnight = datetime(dtCivicTime.year, dtCivicTime.month, dtCivicTime.day)
    (y, m, d) = dtCivicTime.getDate()
    d = floor(d)    # strip fractions, i.e. hours, mins, secs
    dtMidnight = hvfTime(y, m, d, 0, 0, 0)
    if verbose: print ("dtMidnight = ", dtMidnight)
    Steps = 12  # one step every 2 hour
    #Steps = 96  # one step every 15 min, necessary for singapur ?????????
    # we want finer steps for Sun and Moon
    (dummy, dummy, Dict) = getAziAltfromCivicTimeLongLat(dtMidnight, zone, 
            getData, longitude, latitude)
    if (Dict["Object"] == "Sun" or Dict["Object"] == "Moon"):
        Steps = 96  # one step every 15 minutes
    intervalHours = 24. / Steps
    Npoints = 4 # interpolate between 4 points
    statusList = []
    EventList = []
    statusListXY = []
    useAzi = False  # do not use Azimut interpolation

    if verbose: print ("Steps = ", Steps, ", interval = ", intervalHours)

    for i in range(Steps+1):
        hour = intervalHours * i
        dtDayHour = dtMidnight + timedelta(hours=hour)
        if verbose: print ("dtDayHour = ", dtDayHour)
        (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dtDayHour, zone, 
            getData, longitude, latitude)
        if verbose: print ("azi, alt = ", azi, alt)
        # use Alt if a value is given, not None
        if Alt is None:
            useAlt = Dict["Alt"]
        else:
            useAlt = Alt
        x = Dict["xyzDict"]["x"]
        y = Dict["xyzDict"]["y"]
        # store azi, alt and rest of stuff
        statusList.append((i, hour, isUp(alt, useAlt), 
            isDown(alt, useAlt), isNE(azi), isSE(azi), 
            isSW(azi), isNW(azi)))
        statusListXY.append((i, hour, isUp(alt, useAlt), 
            isDown(alt, useAlt), isNxy(x, y), isSxy(x, y), 
            isWxy(x, y), isExy(x, y), round(x,3), round(y,3),
            round(azi, 1), round(alt, 2)))
    
    if verbose:
        print ("getObjectData stored events:")
        print (" i  hour isUP   isDown isNE  isSE  isSW  isNW")
        for status in statusList:
            print (status)
        print (" i  hour isUP   isDown isN   isS   isW   isE    x       y")
        for status in statusListXY:
            print (status)
    

    for n in range(len(statusList)-1):
        (i, hour, UP, DOWN, NE, SE, SW, NW ) = statusList[n]
        (i1, hour1, UP1, DOWN1, NE1, SE1, SW1, NW1) = statusList[n+1]
        dt1 = dtMidnight + timedelta(hours=hour)
        dt2 = dtMidnight + timedelta(hours=hour1)

        # Set
        if UP and not UP1 or not DOWN and DOWN1:
            if verbose:
                print("Setting between %.1f and %.1f" % (hour, hour1))
            dt = interpolateAlt(dt1, dt2, zone, Npoints, getData, Alt,
                    longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': "Set",
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            EventList.append(Event)

        # Rise
        if not UP and UP1 or DOWN and not DOWN1:
            if verbose:
                print("Rising  between %.1f and %.1f" % (hour, hour1))
            dt = interpolateAlt(dt1, dt2, zone, Npoints, getData, Alt,
                    longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': "Rise",
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            EventList.append(Event)
            
    for n in range(len(statusListXY)-1):
        (i, hour, UP, DOWN, N, S, W, E, xx, yy, aa, hh ) = statusListXY[n]
        (i1, hour1, UP1, DOWN1, N1, S1, W1, E1, xx1, yy1, aa1, hh1) = statusListXY[n+1]
        dt1 = dtMidnight + timedelta(hours=hour)
        dt2 = dtMidnight + timedelta(hours=hour1)

        # East or West Passing North to South, South to North: x = 0
        NS1 = N and S1  # from N to S
        SN1 = S and N1  # from S to N
        if NS1 or SN1:
            if NS1 and E:
                sevent = "Passing  East (N-S)"
            elif NS1 and W:
                sevent = "Passing  West (N-S)"
            elif SN1 and E:
                sevent = "East  Passing (S-N)"
            elif SN1 and W:
                sevent = "Passing  West (S-N)"
            if verbose:
                print(sevent, " between %.1f and %.1f" % (hour, hour1))
            dt = interpolateX(dt1, dt2, zone, 0., Npoints, 
                    getData, longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': sevent,
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            EventList.append(Event)

        # North or South passing or transit Est to West, West to East: y = 0.
        EW1 = E and W1  # from E to W
        WE1 = W and E1  # from W to E
        if EW1 or WE1:
            dt = interpolateY(dt1, dt2, zone, 0., Npoints, 
                    getData, longitude, latitude, verbose)
            (azi, alt, Dict) = getAziAltfromCivicTimeLongLat(dt, zone, 
                getData, longitude, latitude)
            if alt > Dict["Alt"]:
                if EW1 and S:
                    sevent = "Transit South (E-W)"
                elif EW1 and N:
                    sevent = "Transit North (E-W)"
                elif WE1 and S:
                    sevent = "Transit South (W-E)"
                elif WE1 and N:
                    sevent = "Transit North (W-E)"
            else:
                if EW1 and S:
                    sevent = "Passing South (E-W)"
                elif EW1 and N:
                    sevent = "Passing North (E-W)"
                elif WE1 and S:
                    sevent = "Passing South (W-E)"
                elif WE1 and N:
                    sevent = "Passing North (W-E)"
                    
            Event = {'Object': Dict["Object"],  # Sun, Moon
                    'dt': dt,
                    'azi': azi,
                    'alt': alt,
                    'event': sevent,
                    'Alt': Dict["Alt"],
                    'Dict': Dict,
                    }
            if verbose:
                print(sevent, " between %.1f and %.1f" % (hour, hour1))
            EventList.append(Event)

    # sort events by time dt
    SortedEventList = sorted(EventList, key = itemgetter('dt'))
    if verbose:
        print ("Steps = %d, Npoints = %d" % (Steps, Npoints))
        for e in SortedEventList:
            print ("%-4s %20s %s Azimut %3d Altitude %7.3f" % (e["Object"], e["event"], 
                nicedatetime(e['dt']), round(e["azi"], 0), e["alt"]))


    return SortedEventList


'''
EventList = getObjectTimes(dt, longitude, latitude, Zone, getMoonData, None,False)
print ("%-4s Events:" % EventList[0]["Object"])
for e in EventList:
    print ("%-4s %20s %s Azimut %3d Altitude %7.3f" % (e["Object"], e["event"], 
                nicedatetime(e['dt']), round(e["azi"], 0), e["alt"]))
print ()
EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, None, False)
print ("%-4s Events:" % EventList[0]["Object"])
for e in EventList:
    print ("%-4s %s %s Azimut %3d Altitude %7.3f" % (e["Object"], e["event"], 
                nicedatetime(e['dt']), round(e["azi"], 0), e["alt"]))

print ()
'''

Version = "V3.1 hvf 2023-11-29 MoonPhasesMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(Version)
# Phases of the Moon, Meeus Chapter 47
# hvf 08.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87
# includes Solar Eclipses - hvf 28.06.2023

# getPhasesJDEbyK (k, verbose = False)
# k counts the number of new moons since 2000-01-06
# Meeus chapter 47
# returns the Julian Ephemeris Days (JDE) time of the mean phase of the Moon
# returns also the elements for solar eclipse (for new moon)

def getPhasesJDEbyK(k, verbose = False):
    # k integer:        New Moon    - Solar eclipse possible
    # k integer + 0.25: First Quarter
    # k integer + 0.5 : Full Moon   - Lunar eclipse possible
    # k integer + 0.75: Last Quarter

    T  = k / 1236.85            # Julian centuries after Epoch2000 (47.3)
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # Sun's Mean anomaly at time JDE (47.4)
    M = 2.5534 + 29.10535669*k - 0.0000218*T*T - 0.00000011*T*T*T
    M = Mod(M, 360.)

    # Moon's Mean anomaly (47.5)
    MM = 201.5643 + 385.81693528*k + 0.0107438*T*T \
            + 0.00001239*T**3 - 0.000000058*T**4
    MM = Mod(MM, 360.)

    # Moon's argument of Latitude (47.6)
    F = 160.7108 + 390.67050274*k - 0.0016341*T*T \
            - 0.00000227*T**3 + 0.000000011*T**4
    F = Mod(F, 360.)

    # Longitude of the Ascending Node of the lunar orbit (47.7)
    Omega = 124.7746 - 1.56375580*k \
            + 0.0020691*T*T + 0.00000215*T**3
    Omega = Mod(Omega, 360.)

    # chapter 52 Solar Eclipses  p.350
    F1 = F - 0.02665 * Sin(Omega)

    # Planetary Arguments

    A1  = 299.77 +  0.107408*k - 0.009173*T*T
    A2  = 251.88 +  0.016321*k
    A3  = 251.83 + 26.651886*k
    A4  = 349.42 + 36.412478*k
    A5  =  84.66 + 18.206239*k
    A6  = 141.74 + 53.303771*k
    A7  = 207.14 +  2.453732*k
    A8  = 154.84 +  7.306860*k
    A9  =  34.52 + 27.261239*k
    A10 = 207.19 +  0.121824*k
    A11 = 291.34 +  1.844379*k
    A12 = 161.72 + 24.198154*k
    A13 = 239.56 + 25.513099*k
    A14 = 331.55 +  3.592518*k

    # Corrections (in days) to JDE (p. 321)
    # for New Moon and Full Moon

    arg = MM;       sNew =  -0.40720*Sin(arg);   sFull =  -0.40614*Sin(arg)
    arg = M;        sNew +=  0.17241*E*Sin(arg); sFull +=  0.17302*E*Sin(arg)
    arg = 2*MM;     sNew +=  0.01608*Sin(arg);   sFull +=  0.01614*Sin(arg)
    arg = 2*F;      sNew +=  0.01039*Sin(arg);   sFull +=  0.01043*Sin(arg)
    arg = MM-M;     sNew +=  0.00739*E*Sin(arg); sFull +=  0.00734*E*Sin(arg)
    arg = MM+M;     sNew += -0.00514*E*Sin(arg); sFull += -0.00515*E*Sin(arg)
    arg = 2*M;      sNew +=  0.00208*EE*Sin(arg); sFull += 0.00209*EE*Sin(arg)
    arg = MM-2*F;   sNew += -0.00111*Sin(arg);   sFull += -0.00111*Sin(arg)
    arg = MM+2*F;   sNew += -0.00057*Sin(arg);   sFull += -0.00057*Sin(arg)
    arg = 2*MM+M;   sNew +=  0.00056*E*Sin(arg); sFull +=  0.00056*E*Sin(arg)
    arg = 3*MM;     sNew += -0.00042*Sin(arg);   sFull += -0.00042*Sin(arg)
    arg = M+2*F;    sNew +=  0.00042*E*Sin(arg); sFull +=  0.00042*E*Sin(arg)
    arg = M-2*F;    sNew +=  0.00038*E*Sin(arg); sFull +=  0.00038*E*Sin(arg)
    arg = 2*MM-M;   sNew += -0.00024*E*Sin(arg); sFull += -0.00024*E*Sin(arg)
    arg = Omega;    sNew += -0.00017*Sin(arg);   sFull += -0.00017*Sin(arg)
    arg = MM+2*M;   sNew += -0.00007*Sin(arg);   sFull += -0.00007*Sin(arg)
    arg = 2*MM-2*F; sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = 3*M;      sNew +=  0.00004*Sin(arg);   sFull +=  0.00004*Sin(arg)
    arg = MM+M-2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = 2*MM+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM+M+2*F; sNew += -0.00003*Sin(arg);   sFull += -0.00003*Sin(arg)
    arg = MM-M+2*F; sNew +=  0.00003*Sin(arg);   sFull +=  0.00003*Sin(arg)
    arg = MM-M-2*F; sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 3*MM+M;   sNew += -0.00002*Sin(arg);   sFull += -0.00002*Sin(arg)
    arg = 4*MM;     sNew +=  0.00002*Sin(arg);   sFull +=  0.00002*Sin(arg)

    # Correction (in days) for First and Last Quarter (p.322)

    arg = MM;       sQ =  -0.62801*Sin(arg); 
    arg = M;        sQ +=  0.17172*E*Sin(arg); 
    arg = MM+M;     sQ += -0.01183*E*Sin(arg); 
    arg = 2*MM;     sQ +=  0.00862*Sin(arg); 
    arg = 2*F;      sQ +=  0.00804*Sin(arg); 
    arg = MM-M;     sQ +=  0.00454*E*Sin(arg); 
    arg = 2*M;      sQ +=  0.00204*EE*Sin(arg); 
    arg = MM-2*F;   sQ += -0.00180*Sin(arg); 
    arg = MM+2*F;   sQ += -0.00070*Sin(arg); 
    arg = 3*MM;     sQ += -0.00040*Sin(arg); 
    arg = 2*MM-M;   sQ += -0.00034*E*Sin(arg); 
    arg = M+2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = M-2*F;    sQ +=  0.00032*E*Sin(arg); 
    arg = 2*M+MM;   sQ += -0.00028*EE*Sin(arg); 
    arg = 2*MM+M;   sQ +=  0.00027*E*Sin(arg); 
    arg = Omega;    sQ += -0.00017*Sin(arg); 
    arg = MM-M-2*F; sQ += -0.00005*Sin(arg); 
    arg = 2*MM+2*F; sQ +=  0.00004*Sin(arg); 
    arg = MM+M+2*F; sQ += -0.00004*Sin(arg); 
    arg = MM-2*M;   sQ +=  0.00004*Sin(arg); 
    arg = MM+M-2*F; sQ +=  0.00003*Sin(arg); 
    arg = 3*M;      sQ +=  0.00003*Sin(arg); 
    arg = 2*MM-2*F; sQ +=  0.00002*Sin(arg); 
    arg = MM-M+2*F; sQ +=  0.00002*Sin(arg); 
    arg = 3*MM+M;   sQ += -0.00002*Sin(arg); 

    # additional correction for Quarter phases only

    W = 0.00306 - 0.00038*E*Cos(M) + 0.00026*Cos(MM) \
        - 0.00002*Cos(MM-M) + 0.00002*Cos(MM+M) + 0.00002*Cos(2*F)
    # for first Quarter: +W
    # for last Quarter:  -W

    # additional correction for all phases

    addCorr = 0.000325*Sin(A1)  + 0.000165*Sin(A2)  + 0.000164*Sin(A3) \
            + 0.000126*Sin(A4)  + 0.000110*Sin(A5)  + 0.000062*Sin(A6) \
            + 0.000060*Sin(A7)  + 0.000056*Sin(A8)  + 0.000047*Sin(A9) \
            + 0.000042*Sin(A10) + 0.000040*Sin(A11) + 0.000037*Sin(A12) \
            + 0.000035*Sin(A13) + 0.000023*Sin(A14)

    # times of the mean phases of the Moon, already affected by the Sun's
    # aberration and by the Moon's light-time (eq. 47.1)

    JDE = 2451550.09765 + 29.530588853*k \
            + 0.0001337*T*T \
            - 0.000000150*T*T*T \
            + 0.00000000073*T**4
    JDEsave = JDE

    # Eclipses correction to time of max eclipse (52.1)
    # sN: New Moon; sF: Full Moon
    arg = 2*MM;     sN  = 0.0161*Sin(arg); 
    arg = 2*F1;     sN += -0.0097*Sin(arg)
    arg = MM-M;     sN += 0.0073*E*Sin(arg)
    arg = MM+M;     sN += -0.0050*E*Sin(arg)
    arg = MM-2*F1;  sN += -0.0023*Sin(arg)
    arg = 2*M;      sN += 0.0021*E*Sin(arg)
    arg = MM+2*F1;  sN += 0.0012*Sin(arg)
    arg = 2*MM+M;   sN += 0.0006*E*Sin(arg)
    arg = 3*MM;     sN += -0.0004*Sin(arg)
    arg = M+2*F1;   sN += -0.0003*E*Sin(arg)
    arg = A1;       sN += 0.0003*Sin(arg)
    arg = M-2*F1;   sN += -0.0002*E*Sin(arg)
    arg = 2*MM-M;   sN += -0.0002*E*Sin(arg)
    arg = Omega;    sN += -0.0002*Sin(arg)

    sF = sN

    arg = MM;   sN += -0.4075*Sin(arg);     sF += -0.4065*Sin(arg)
    arg = M;    sN += 0.1721*E*Sin(arg);    sF += 0.1727*E*Sin(arg)

    # P and Q - p. 351
    Q = 5.2207
    arg = M;    P  = 0.2070*E*Sin(arg);     Q += -0.0048*E*Cos(arg)
    arg = 2*M;  P += 0.0024*E*Sin(arg);     Q += 0.0020*E*Cos(arg)
    arg = MM;   P += -0.0392*Sin(arg);      Q += -0.3299*Cos(arg)
    arg = 2*MM; P += 0.0116*Sin(arg)
    arg = MM+M; P += -0.0073*E*Sin(arg);    Q += -0.0060*E*Cos(arg)
    arg = MM-M; P += 0.0067*E*Sin(arg);     Q += 0.0041*E*Cos(arg)
    arg = 2*F1; P += 0.0118*Sin(arg)

    # WW, gamma, u
    # For a Solar eclipse, gamma is the least distance from the axis 
    # of the Moon's shadow to the center of the Earth, 
    # in units of the equatorial radius of the Earth.
    # gamma is positive if the shadow passes North of the center.
    # when gamma is between +0.9972 and -0.9972, the solar eclipse is central:
    # there exists a line of central eclipse on the Earth's surface.
    WW = abs(Cos(F1))
    gamma = (P*Cos(F1) + Q*Sin(F1)) * (1. - 0.0048*WW)

    # u denotes the radius of the Moon's umbral cone in the fundamental plane.
    # The radius of the penumbral cone in the fundamental plane is
    #   u + 0.5461

    u = 0.0059
    u += 0.0046*E*Cos(M)
    u += -0.0182*Cos(MM)
    u += 0.0004*Cos(2*MM)
    u += -0.0005*Cos(M+MM)

    # if abs(gamma) is between 0.9972 and 1.5433+u, the eclipse is not central.
    # If abs(gamma) is between 0.9972 and 1.5433 + u, the eclipse is not
    # central. In most cases, it is then a partial eclipse. However, when
    # abs(gamma) is between 0.9972 and 1.0260, a part of the umbral cone may
    # touch the surface of the Earth (within the polar regions), while the
    # axis of the cone does not touch the Earth. These non-central total
    # or annular eclipses occur when 0.9972 < abs(gamma) < 0.9972 + |u|.


    # CAREFUL: these tests must work for negative years as well!

    def isNewMoon(k):   # k must be integer
        test = (abs(k - int(k)) < 0.001)
        # if verbose: print ("isNewMoon k, test = ", k, test)
        return test     # True or False

    def isFullMoon(k):  # k must be half-integer
        test = (abs(abs(k - int(k)) - 0.5) < 0.001)
        # if verbose: print ("isFullMoon k, test = ", k, test)
        return test     # True or False 

    def isFirstQuarter(k):  # k must be .25
        test = abs(abs(k - floor(k)) - 0.25) < 0.001
        # if verbose: print ("isFirstQuarter k, test = ", k, test)
        return test     # True or False 

    def isLastQuarter(k):  # k must be .75
        test = abs(abs(ceil(k) - k) - 0.25) < 0.001
        # if verbose: print ("isLastQuarter k, test = ", k, test)
        return test     # True or False 

    JDEeclipse = None
    if isNewMoon(k):
        JDE += sNew
        JDEeclipse = hvfTime(JDEsave + sN)
        # if verbose: print ("New Moon")
    elif isFullMoon(k):
        JDE += sFull
        # if verbose: print ("Full Moon")
    elif isFirstQuarter(k):
        JDE += sQ
        JDE += W
        # if verbose: print ("First Quarter")
    elif isLastQuarter(k):
        JDE += sQ
        JDE -= W
        # if verbose: print ("Last Quarter")
    else:
        print ("k not for New or Full Moon: ", k)
    JDE += addCorr
    eclType = None
    omega = None
    SEpossible = False
    if isNewMoon(k) and abs(gamma) <= 1.5433+u: 
        SEpossible = True
        if abs(gamma) < 0.9972:
            if u < 0:
                eclType = ("Total Central Solar Eclipse")
            elif u < 0.0047:
                omega = 0.00464*sqrt(1-gamma**2)
                if u < omega:
                    eclType = ("Annular-total Central Solar Eclipse")
                else:
                    eclType = ("Annular Central Solar Eclipse")
            elif u >= 0.0047:
                eclType = ("Annular Central Solar Eclipse")
        elif abs(gamma) < 0.9972+abs(u):
            eclType = ("Non-central total or annular Solar eclipse")
        elif abs(gamma) < 1.5433+u:
            eclType = ("Non-central Solar Eclipse")
        if verbose:
            print ("getPhasesJDEbyK: k = %.2f" % k)
            print ("JDE = %.6f, %s" % (JDE, nicedatetime(hvfTime(JDE))))
            print ("JDEsave = %.6f, %s" % (JDEsave, nicedatetime(hvfTime(JDEsave))))
            print ("corrections sN, sF = ", sN, sF)
            #if JDEeclipse: print ("JDEeclipse = %.6f, %s" % (JDEeclipse, nicedatetime(hvfTime(JDEeclipse))))
            if JDEeclipse: print ("JDEeclipse = %s" % (nicedatetime(hvfTime(JDEeclipse))))
            print ("E, M(Sun), MM(Moon), F, Omega = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (E, M, MM, F, Omega)) 
            print ("F1, WW, gamma, u = %.4f, %.4f, %.4f, %.4f" % (F1, WW, gamma, u))
            print ("gamma = %.4f, closest distance of shadow to Earth center"%gamma)
            print ("Radius of umbral cone u = %.4f" % u)
            print ()
    
    EclipseDict = { "Object" : "Sun",
            "possible"  : SEpossible,
            "JDE"   : JDE,
            "JDEeclipse"   : JDEeclipse,
            "dt"    : JDEeclipse,
            "u"     : u,
            "gamma" : gamma,
            "omega" : omega,
            "event" : eclType,
            }
    if SEpossible and verbose: 
        print (EclipseDict)
        print ("JDEclipse = ", nicedatetime(hvfTime(JDEeclipse)))
        print ()


    return JDE, EclipseDict

# getMoonPhases(year, useUT)
#   returns:
#   newMoon
#   firstQuarter
#   fullMoon
#   lastQuarter
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
from operator import itemgetter

def getMoonPhases(year, useUT):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.)*12.3685  # relative moon phase at new year
    if (k0 == int(k0)): k = k0
    else: k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85            # Julian centuries after Epoch2000
    y  = year
    EventList = []
    EventDistList = []
    eventType = ["New Moon", "First Quarter", "Full Moon", "Last Quarter"]
    EventTypeList = []
    # starting values
    #print ("getMoonPhases start: k, T, y =", k, T, y)

    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    while y == year:
        #print ("y, k, ", y, k)
        dateList = ['', '', '', ''] # new,first q, full, last q
        distList = ['', '']         # new, full moon
        for i in range(4):
            #verb = k == int(k)      # verbose
            verb = False
            JDE, EclipseDict = getPhasesJDEbyK(k, verb)  # TT
            dtE = getGreenwichDate(JDE)      # TT
            deltaT = timedelta(seconds=getDeltaT(JDE))
            dtUTC = dtE - deltaT
            dtDisplay = dtE
            if useUT: dtDisplay = dtUTC

            # getMoonData expects UTC time
            (MoonRA, MoonDecl, MoonDict) = getMoonData(dtUTC, 0, False)
            MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "  "
            if MoonDist <= 360000. : suffix = "++" # Super Moon
            if MoonDist >= 405000. : suffix = "--" # Mini Moon
            (yyyy, dummy, dummy) = dtUTC.getDate()

            # Event
            Event = {'Object': MoonDict["Object"],  # Moon
                    'dt'    : dtDisplay,
                    'event' : eventType[i],
                    'mode'  : eventType[i],
                    'Dict'  : MoonDict,
                    }
            EventTypeList.append(Event)
            # just pass eclipse dict, adjust dt for UTC or TT
            if EclipseDict["possible"] :
                JDEeclipseDisplay = EclipseDict["JDEeclipse"]
                if useUT:
                    JDEeclipseDisplay = JDEeclipseDisplay - deltaT    
                EclipseDict["dt"] = JDEeclipseDisplay        

                EventTypeList.append(EclipseDict)
            #if (dt.year == year): 
            if (yyyy == year): 
                dateList[i] = niceshortdatetime(dtDisplay)
                if i == 0:
                    distList[0] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
                if i == 2:
                    distList[1] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
            k += 0.25
            T = k / 1236.85
            #print ("k, T, yyyy =", k, T, yyyy)
        EventList.append(dateList)
        if len(distList[0]+distList[1]) > 0: EventDistList.append(distList)
        T = k / 1236.85
        #y = int(T * 100 + 2000)
        y = floor(T * 100 + 2000)
        #print ("k, T, y =", k, T, y)
   # sort events by time dt in EventTypeList (solar eclipses)
    SortedEventTypeList = sorted(EventTypeList, key = itemgetter('dt'))

    return (EventList, EventDistList, SortedEventTypeList)

#########################################################      

# getMoonPhasesxxx(dtCivicTime, zone, verbose=False)
#   
#   returns:
#   newMoon
#   firstQuarter
#   fullMoon
#   lastQuarter
#   for about 30 days before and after dtCivicTime
#
from operator import itemgetter

def getMoonPhasesxxx(dtCivicTime, zone, verbose=False):
    # chapter 47 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 
    #k0 = (yearxxx-2000.)*12.3685  # relative moon phase at given 
    k0 = Ty * 12.3685  # relative moon phase at given (47.2)
    k  = ceil(k0) - 1     # start with new moon
    T  = k / 1236.85      # Julian centuries after Epoch2000
    kmin = k - 2
    kmax = k + 3
    EventList = []
    EventDistList = []
    eventType = ["New Moon", "First Quarter", "Full Moon", "Last Quarter"]
    EventTypeList = []
    # starting values
    #print ("getMoonPhases start: k, T, y =", k, T, y)
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    k = kmin
    for ki in range(kmin, kmax+1):
    #while y == year:
        #print ("y, k, ", y, k)
        dateList = ['', '', '', ''] # new,first q, full, last q
        distList = ['', '']         # new, full moon
        for i in range(4):
            #verb = k == int(k)      # verbose
            verb = False
            JDE, EclipseDict = getPhasesJDEbyK(k, verb)  # TT
            # only +- 3 weeks
            
            dtE = getGreenwichDate(JDE)      # TT
            deltaT = timedelta(seconds=getDeltaT(JDE))
            dtUTC = dtE - deltaT
            dtDisplay = dtE
            if useUT: dtDisplay = dtUTC

            # getMoonData expects UTC time
            (MoonRA, MoonDecl, MoonDict) = getMoonData(dtUTC, 0, False)
            MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "  "
            if MoonDist <= 360000. : suffix = "++" # Super Moon
            if MoonDist >= 405000. : suffix = "--" # Mini Moon
            (yyyy, dummy, dummy) = dtUTC.getDate()

            dtCT = getCivicTimefromUT(dtUTC, Zone)
            # Event
            #Event = {'Object': MoonDict["Object"],  # Moon
            Event = {'Object': "Moon Phases",
                    'dt'    : dtUTC,
                    'dtCivicTime' : dtCT,
                    'event' : eventType[i],
                    'mode'  : eventType[i],
                    'decl'  : MoonDecl,
                    'dist'  : MoonDist,
                    'suffix': suffix,
   
                    'Dict'  : MoonDict,
                    }
            EventTypeList.append(Event)
            # just pass eclipse dict, adjust dt for UTC or TT
            if EclipseDict["possible"] :
                JDEeclipseDisplay = EclipseDict["JDEeclipse"]
                if useUT:
                    JDEeclipseDisplay = JDEeclipseDisplay - deltaT    
                EclipseDict["dt"] = JDEeclipseDisplay        
                EclipseDict["dtCivicTime"] = \
                    getCivicTimefromUT(JDEeclipseDisplay, Zone)        
                EventTypeList.append(EclipseDict)
            if True:
                dateList[i] = niceshortdatetime(dtDisplay)
                if i == 0:
                    distList[0] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
                if i == 2:
                    distList[1] = niceshortdatetime(dtDisplay) + \
                        ("  %8.0f" % MoonDist) + ' ' + suffix
            k += 0.25
            T = k / 1236.85
        EventList.append(dateList)
        if len(distList[0]+distList[1]) > 0: EventDistList.append(distList)
        T = k / 1236.85
        #y = int(T * 100 + 2000)
        y = floor(T * 100 + 2000)
        #print ("k, T, y =", k, T, y)
   # sort events by time dt in EventTypeList (solar eclipses)
    SortedEventTypeList = sorted(EventTypeList, key = itemgetter('dt'))

    return (EventList, EventDistList, SortedEventTypeList)

#########################################################      
# Maximum Declinations of the Moon, Meeus Chapter 50
# hvf 02.05.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87
# replace TDT by TT (terrestrial time) - hvf 06.04.2023

import sys

VersionMoon = "V3.1 hvf 2023-11-29 MoonDecliMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(VersionMoon)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    sys.exit(0)

# getDecliJDEbyK (k, verbose = False)
# k counts the number of revolutions since 2000-01-01
# Meeus chapter 50
# returns the Julian Ephemeris Days (JDE) time and maximum declination the Moon
# both for Northern and Southern maxima
# return (JDEN, deltaN, JDES, deltaS)
def getDecliJDEbyK(k, verbose = False):
    # k integer:        Maximum Declination N, S since 2000-01-01

    T  = k / 1336.86            # (50.1), Julian centuries after Epoch2000
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # DN, DS for North and South extrema
    DN = 152.2029 + 333.0705546*k - 0.0004025*T*T \
        + 0.00000011*T**3
    DN = Mod(DN, 360.)
    DS = 345.6676 + 333.0705546*k - 0.0004025*T*T \
        + 0.00000011*T**3
    DS = Mod(DS, 360.)

    # Sun's Mean anomaly 
    MN = 14.8591 + 26.9281592*k - 0.0000544*T*T - 0.00000010*T*T*T
    MN = Mod(MN, 360.)
    MS =  1.3981 + 26.9281592*k - 0.0000544*T*T - 0.00000010*T*T*T
    MS = Mod(MS, 360.)

    # Moon's Mean anomaly 
    MMN =  4.6881 + 356.9562795*k + 0.0103126*T*T \
            + 0.00001251*T**3
    MMN = Mod(MMN, 360.)
    MMS = 186.2100 + 356.9562795*k + 0.0103126*T*T \
            + 0.00001251*T**3
    MMS = Mod(MMS, 360.)

    FN = 325.8867 +1.4467806*k \
            - 0.0020708*T*T - 0.00000215*T**3 
    FN = Mod(FN, 360.)
    FS = 145.1633 +1.4467806*k \
            - 0.0020708*T*T - 0.00000215*T**3 
    FS = Mod(FS, 360.)

    JDEN = 2451562.5897 + 27.321582241*k \
            + 0.000100695*T*T \
            - 0.000000141*T*T*T 

    JDES = 2451548.9289 + 27.321582241*k \
            + 0.000100695*T*T \
            - 0.000000141*T*T*T 

    # Table 50.A periodic terms for the time of the Moon's maximum declination
    sumTimNorth = 0.8975 * Cos(FN) \
                - 0.4726 * Sin(MMN) \
                - 0.1030 * Sin(2*FN) \
                - 0.0976 * Sin(2*DN-MMN) \
                - 0.0462 * Cos(MMN-FN) \
                - 0.0461 * Cos(MMN+FN) \
                - 0.0438 * Sin(2*DN) \
                + 0.0162*E*Sin(MN) \
                - 0.0157 * Cos(3*FN) \
                + 0.0145 * Sin(MMN+2*FN) \
                + 0.0136 * Cos(2*DN-FN) \
                - 0.0095 * Cos(2*DN-MMN-FN) \
                - 0.0091 * Cos(2*DN-MMN+FN) \
                - 0.0089 * Cos(2*DN+FN) \
                + 0.0075 * Sin(2*MMN) \
                - 0.0068 * Sin(MMN-2*FN) \
                + 0.0061 * Cos(2*MMN-FN) \
                - 0.0047 * Sin(MMN+3*FN) \
                - 0.0043*E*Sin(2*DN-MN-MMN) \
                - 0.0040 * Cos(MMN-2*FN) \
                - 0.0037 * Sin(2*DN-2*MMN) \
                + 0.0031 * Sin(FN) \
                + 0.0030 * Sin(2*DN+MMN) \
                - 0.0029 * Cos(MMN+2*FN) \
                - 0.0029*E*Sin(2*DN-MN) \
                - 0.0027 * Sin(MMN+FN) \
                + 0.0024*E*Sin(MN-MMN) \
                - 0.0021 * Sin(MMN-3*FN) \
                + 0.0019 * Sin(2*MMN+FN) \
                + 0.0018 * Cos(2*DN-2*MMN-FN) \
                + 0.0018 * Sin(3*FN) \
                + 0.0017 * Cos(MMN+3*FN) \
                + 0.0017 * Cos(2*MMN) \
                - 0.0014 * Cos(2*DN-MMN) \
                + 0.0013 * Cos(2*DN+MMN+FN) \
                + 0.0013 * Cos(MMN) \
                + 0.0012 * Sin(3*MMN+FN) \
                + 0.0011 * Sin(2*DN-MMN+FN) \
                - 0.0011 * Cos(2*DN-2*MMN) \
                + 0.0010 * Cos(DN+FN) \
                + 0.0010*E*Sin(MN+MMN) \
                - 0.0009 * Sin(2*DN-2*FN) \
                + 0.0007 * Cos(2*MMN+FN) \
                - 0.0007 * Cos(3*MMN+FN) 
                
    sumTimSouth =-0.8975 * Cos(FS) \
                - 0.4726 * Sin(MMS) \
                - 0.1030 * Sin(2*FS) \
                - 0.0976 * Sin(2*DS-MMS) \
                + 0.0541 * Cos(MMS-FS) \
                + 0.0516 * Cos(MMS+FS) \
                - 0.0438 * Sin(2*DS) \
                + 0.0112*E*Sin(MS) \
                + 0.0157 * Cos(3*FS) \
                + 0.0023 * Sin(MMS+2*FS) \
                - 0.0136 * Cos(2*DS-FS) \
                + 0.0110 * Cos(2*DS-MMS-FS) \
                + 0.0091 * Cos(2*DS-MMS+FS) \
                + 0.0089 * Cos(2*DS+FS) \
                + 0.0075 * Sin(2*MMS) \
                - 0.0030 * Sin(MMS-2*FS) \
                - 0.0061 * Cos(2*MMS-FS) \
                - 0.0047 * Sin(MMS+3*FS) \
                - 0.0043*E*Sin(2*DS-MS-MMS) \
                + 0.0040 * Cos(MMS-2*FS) \
                - 0.0037 * Sin(2*DS-2*MMS) \
                - 0.0031 * Sin(FS) \
                + 0.0030 * Sin(2*DS+MMS) \
                + 0.0029 * Cos(MMS+2*FS) \
                - 0.0029*E*Sin(2*DS-MS) \
                - 0.0027 * Sin(MMS+FS) \
                + 0.0024*E*Sin(MS-MMS) \
                - 0.0021 * Sin(MMS-3*FS) \
                - 0.0019 * Sin(2*MMS+FS) \
                - 0.0006 * Cos(2*DS-2*MMS-FS) \
                - 0.0018 * Sin(3*FS) \
                - 0.0017 * Cos(MMS+3*FS) \
                + 0.0017 * Cos(2*MMS) \
                + 0.0014 * Cos(2*DS-MMS) \
                - 0.0013 * Cos(2*DS+MMS+FS) \
                - 0.0013 * Cos(MMS) \
                + 0.0012 * Sin(3*MMS+FS) \
                + 0.0011 * Sin(2*DS-MMS+FS) \
                + 0.0011 * Cos(2*DS-2*MMS) \
                + 0.0010 * Cos(DS+FS) \
                + 0.0010*E*Sin(MS+MMS) \
                - 0.0009 * Sin(2*DS-2*FS) \
                - 0.0007 * Cos(2*MMS+FS) \
                - 0.0007 * Cos(3*MMS+FS) 
                
    JDEN += sumTimNorth
    JDES += sumTimSouth

    # Table 50.B Periodic terms for the value of the Moon's maximum declination

    delta = 23.6961 - 0.013004*T

    sumDelNorth =  5.1093 * Sin(FN) \
                + 0.2658 * Cos(2*FN) \
                + 0.1448 * Sin(2*DN-FN) \
                - 0.0322 * Sin(3*FN) \
                + 0.0133 * Cos(2*DN-2*FN) \
                + 0.0125 * Cos(2*DN) \
                - 0.0124 * Sin(MMN-FN) \
                - 0.0101 * Sin(MMN+2*FN) \
                + 0.0097 * Cos(FN) \
                - 0.0087*E*Sin(2*DN+MN-FN) \
                + 0.0074 * Sin(MMN+3*FN) \
                + 0.0067 * Sin(DN+FN) \
                + 0.0063 * Sin(MMN-2*FN) \
                + 0.0060*E*Sin(2*DN-MN-FN) \
                - 0.0057 * Sin(2*DN-MMN-FN) \
                - 0.0056 * Cos(MMN+FN) \
                + 0.0052 * Cos(MMN+2*FN) \
                + 0.0041 * Cos(2*MMN+FN) \
                - 0.0040 * Cos(MMN-3*FN) \
                + 0.0038 * Cos(2*MMN-FN) \
                - 0.0034 * Cos(MMN-2*FN) \
                - 0.0029 * Sin(2*MMN) \
                + 0.0029 * Sin(3*MMN+FN) \
                - 0.0028*E*Cos(2*DN+MN-FN) \
                - 0.0028 * Cos(MMN-FN) \
                - 0.0023 * Cos(3*FN) \
                - 0.0021 * Sin(2*DN+FN) \
                + 0.0019 * Cos(MMN+3*FN) \
                + 0.0018 * Cos(DN+FN) \
                + 0.0017 * Sin(2*MMN-FN) \
                + 0.0015 * Cos(3*MMN+FN) \
                + 0.0014 * Cos(2*DN+2*MMN+FN) \
                - 0.0012 * Sin(2*DN-2*MMN-FN) \
                - 0.0012 * Cos(2*MMN) \
                - 0.0010 * Cos(MMN) \
                - 0.0010 * Sin(2*FN) \
                + 0.0006 * Sin(MMN+FN)


    sumDelSouth = -5.1093 * Sin(FS) \
                + 0.2658 * Cos(2*FS) \
                - 0.1448 * Sin(2*DS-FS) \
                + 0.0322 * Sin(3*FS) \
                + 0.0133 * Cos(2*DS-2*FS) \
                + 0.0125 * Cos(2*DS) \
                - 0.0015 * Sin(MMS-FS) \
                + 0.0101 * Sin(MMS+2*FS) \
                - 0.0097 * Cos(FS) \
                + 0.0087*E*Sin(2*DS+MS-FS) \
                + 0.0074 * Sin(MMS+3*FS) \
                + 0.0067 * Sin(DS+FS) \
                - 0.0063 * Sin(MMS-2*FS) \
                - 0.0060*E*Sin(2*DS-MS-FS) \
                + 0.0057 * Sin(2*DS-MMS-FS) \
                - 0.0056 * Cos(MMS+FS) \
                - 0.0052 * Cos(MMS+2*FS) \
                - 0.0041 * Cos(2*MMS+FS) \
                - 0.0040 * Cos(MMS-3*FS) \
                - 0.0038 * Cos(2*MMS-FS) \
                + 0.0034 * Cos(MMS-2*FS) \
                - 0.0029 * Sin(2*MMS) \
                + 0.0029 * Sin(3*MMS+FS) \
                + 0.0028*E*Cos(2*DS+MS-FS) \
                - 0.0028 * Cos(MMS-FS) \
                + 0.0023 * Cos(3*FS) \
                + 0.0021 * Sin(2*DS+FS) \
                + 0.0019 * Cos(MMS+3*FS) \
                + 0.0018 * Cos(DS+FS) \
                - 0.0017 * Sin(2*MMS-FS) \
                + 0.0015 * Cos(3*MMS+FS) \
                + 0.0014 * Cos(2*DS+2*MMS+FS) \
                + 0.0012 * Sin(2*DS-2*MMS-FS) \
                - 0.0012 * Cos(2*MMS) \
                + 0.0010 * Cos(MMS) \
                - 0.0010 * Sin(2*FS) \
                + 0.0037 * Sin(MMS+FS)

    deltaN = delta + sumDelNorth
    deltaS = -(delta + sumDelSouth)

    if verbose:
        print ("getDecliJDEbyK: k = %.2f" % k)
        dtN = getGreenwichDate(JDEN)
        dtS = getGreenwichDate(JDES)
        print ("JDEN = %.6f" % JDEN, nicedatetime(dtN), " deltaN = %.6f"%deltaN)
        print ("JDES = %.6f" % JDES, nicedatetime(dtS), " deltaS = %.6f"%deltaS)

    return (JDEN, deltaN, JDES, deltaS)

# getMoonDecli(year, useUT)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonDecli(year, useUT):
    # chapter 50 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.03)*13.3686  # k0 sideric months since 2000-01-10
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Ascending Node
    y  = year
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for k in range(kstart, kstart+16):
        dateList = ['', '']
        foundDate = False
        T  = k / 1336.86            # Julian centuries after Epoch2000
        y = int(T * 100 + 2000)
        (JDEN, deltaN, JDES, deltaS) = getDecliJDEbyK(k, False)
        dtN = getGreenwichDate(JDEN)
        dtS = getGreenwichDate(JDES)
        deltaTN = timedelta(seconds=getDeltaT(JDEN))
        deltaTS = timedelta(seconds=getDeltaT(JDES))
        if useUT: 
            dtN -= deltaTN
            dtS -= deltaTS
        # South is before North for a given k, so we display South first

        (yyyy, dummy, dummy) = dtS.getDate()
        #if dtS.year == year: 
        if yyyy == year: 
            dateList[0] = nicedateshorttime(dtS) + " %9.4f"%deltaS
            foundDate = True

        (yyyy, dummy, dummy) = dtN.getDate()
        #if dtN.year == year: 
        if yyyy == year: 
            dateList[1] = nicedateshorttime(dtN) + " %9.4f"%deltaN
            foundDate = True

        if foundDate: EventList.append(dateList)
    return EventList
#########################################################     

# getMoonDeclixxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonDeclixxx(dtCivicTime, zone, verbose=False):
    # chapter 50 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 

    #k0 = (year-2000.03)*13.3686  # k0 sideric months since 2000-01-10
    k0 = (Ty-0.03)*13.3686  # k0 sideric months since 2000-01-10
    k  = ceil(k0) - 2     # start with Ascending Node
    kmin = k 
    kmax = k + 3
    EventList = []
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for k in range(kmin, kmax+1):
        dateList = ['', '']
        foundDate = False
        T  = k / 1336.86            # Julian centuries after Epoch2000
        y = int(T * 100 + 2000)
        (JDEN, deltaN, JDES, deltaS) = getDecliJDEbyK(k, False)
        dtN = getGreenwichDate(JDEN)
        dtS = getGreenwichDate(JDES)
        deltaTN = timedelta(seconds=getDeltaT(JDEN))
        deltaTS = timedelta(seconds=getDeltaT(JDES))
        if useUT: 
            dtN -= deltaTN
            dtS -= deltaTS
        dtCTN = getCivicTimefromUT(dtN, zone)
        dtCTS = getCivicTimefromUT(dtS, zone)


        mode = ''

        (yyyy, dummy, dummy) = dtS.getDate()
        #if dtS.year == year: 
        if True:
            #dateList[0] = nicedateshorttime(dtS) + " %9.4f"%deltaS
            #foundDate = True
            (dummy, dummy, MoonDict) = getMoonData(dtS, 0, False)
            MoonDist = MoonDict["MoonDist"]
            suffix = MoonDict["suffix"]
            mode = "South Decl"
            Event = {'Object': 'Declination',
                    'dt': dtS,
                    'dtCivicTime' : dtCTS,
                    'dist': MoonDist,
                    'decl': deltaS,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)

        (yyyy, dummy, dummy) = dtN.getDate()
        #if dtN.year == year: 
        if True:
            #dateList[1] = nicedateshorttime(dtN) + " %9.4f"%deltaN
            (dummy, MoonDecl, MoonDict) = getMoonData(dtN, 0, False)
            MoonDist = MoonDict["MoonDist"]
            suffix = MoonDict["suffix"]
            mode = "North Decl"
            Event = {'Object': 'Declination',
                    'dt': dtN,
                    'dtCivicTime' : dtCTN,
                    'dist': MoonDist,
                    'decl': deltaN,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)

    return EventList
#########################################################     
#! /usr/bin/env python3
# Perigee and Apogee of the Moon, Meeus Chapter 48
# hvf 28.02.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

from datetime import datetime, date, timedelta
from math import sin, acos, cos, pi, radians, degrees, atan, atan2
from math import asin, tan, ceil, floor, sqrt, fabs
import sys

VersionMoonPeriApo ="V3.1 hvf 2023-11-29 MoonPeriMain.py based on Meeus, Astronomical Algorithms"
VersionList.append(VersionMoonPeriApo)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    print ("Usage: %s [UTC|TT] year-year ...]" % progname)
    sys.exit(0)

def isPerigee(k):   # k must be integer
    test = (abs(k - int(k)) < 0.001)
    #if verbose: print ("isPerigee k, test = ", k, test)
    return test     # True or False

def isApogee(k):  # k must be half-integer
    test = (abs(abs(k - int(k)) - 0.5) < 0.001)
    #if verbose: print ("isApogee k, test = ", k, test)
    return test     # True or False 

# getPeriJDEbyK (k, verbose = False)
# k counts the number of perigees since 1999-12-22
# Meeus chapter 48
# returns the Julian Ephemeris Days (JDE) time of the perigee, apogee
def getPeriJDEbyK(k, verbose = False):
    # k integer:        Perigee
    # k integer + 0.5 : Apogee

    T  = k / 1325.55            # 48.3, Julian centuries after Epoch2000

    # mean perigee, apogee (48.1)
    JDE = 2451534.6698 + 27.55454988 * k \
                        - 0.0006886 * T*T \
                        - 0.000001098 * T**3 \
                        + 0.0000000052 * T**4
    JDE0 = JDE

    # Moon's mean elongation at time JDE 
    D = 171.9179  + 335.9106046 * k \
                    - 0.0100250 * T*T \
                    - 0.00001156 * T**3 \
                    + 0.000000055 * T**4
    D = Mod(D, 360.)

    # Sun's Mean anomaly
    M = 347.3477   + 27.1577721 * k \
                    - 0.0008323 * T*T \
                    - 0.0000010 * T**3 
    M = Mod(M, 360.)

    # Moon's argument of Latitude
    F = 316.6109  + 364.5287911 * k \
                    - 0.0125131 * T*T \
                    - 0.0000148 *T**3 
    F = Mod(F, 360.)

    # sumP: perigee sum (in days)
    # sumA: apogee sum (in days)
    # sum HPP: perigee parallax (in arcsec), table 48.B
    # sum HPA: apogee parallax (in arcsec), table 48.B
    sumHPP = 3629.215;      sumHPA = 3245.251
    sumP = 0.;                                  sumA = 0.
    arg = 2*D;      sumP += -1.6769 * Sin(arg); sumA += 0.4392 * Sin(arg)
    sumHPP +=  63.224 * Cos(arg);    sumHPA += -9.147 * Cos(arg)
    arg = 4*D;      sumP +=  0.4589 * Sin(arg); sumA += 0.0684 * Sin(arg)
    sumHPP +=  -6.990 * Cos(arg);    sumHPA +=  0.355 * Cos(arg)
    arg = 6*D;      sumP += -0.1856 * Sin(arg); sumA += 0.0144 * Sin(arg)
    sumHPP +=   1.927 * Cos(arg);    sumHPA +=  0.052 * Cos(arg)
    arg = 8*D;      sumP +=  0.0883 * Sin(arg); sumA += 0.0035 * Sin(arg)
    sumHPP +=  -0.702 * Cos(arg);    sumHPA +=  0.010 * Cos(arg)
    arg = 2*D-M;    sumP += (-0.0773 + 0.00019 * T) * Sin(arg); \
                                sumA += (0.0426 - 0.00011 * T) * Sin(arg)
    sumHPP += (2.834 - 0.0071 * T) * Cos(arg);    sumHPA +=  0.159 * Cos(arg)
    arg = M;        sumP += ( 0.0502 - 0.00013 * T) * Sin(arg); \
                                sumA += (0.0456 - 0.00011 * T) * Sin(arg)
    sumHPP += ( 0.696 - 0.0017 * T) * Cos(arg);    
    sumHPA += (-0.656 + 0.0016 * T) * Cos(arg)
    arg = 10*D;     sumP += -0.0460 * Sin(arg); sumA +=  0.0009 * Sin(arg)
    sumHPP +=  0.297 * Cos(arg)
    arg = 4*D-M;    sumP += ( 0.0422 - 0.00011 * T) * Sin(arg); \
                                sumA +=  0.0113 * Sin(arg)
    sumHPP += (-0.629 + 0.0016 * T) * Cos(arg);    sumHPA += 0.065 * Cos(arg)
    arg = 6*D-M;    sumP += -0.0256 * Sin(arg); sumA +=  0.0034 * Sin(arg)
    sumHPP +=  0.260 * Cos(arg);    sumHPA +=  0.014 * Cos(arg)
    arg = 12*D;     sumP +=  0.0253 * Sin(arg); sumA +=  0.0003 * Sin(arg)
    sumHPP += -0.138 * Cos(arg)
    arg = D;        sumP +=  0.0237 * Sin(arg); sumA += -0.0189 * Sin(arg)
    sumHPP += -1.263 * Cos(arg);    sumHPA += -0.841 * Cos(arg)
    arg = 8*D-M;    sumP +=  0.0162 * Sin(arg); sumA +=  0.0011 * Sin(arg)
    sumHPP += -0.127 * Cos(arg)
    arg = 14*D;     sumP += -0.0145 * Sin(arg)
    sumHPP +=  0.068 * Cos(arg)
    arg = 2*F;      sumP +=  0.0129 * Sin(arg); sumA +=  0.0212 * Sin(arg)
    sumHPP += -0.690 * Cos(arg);    sumHPA +=  0.697 * Cos(arg)
    arg = 3*D;      sumP += -0.0112 * Sin(arg); sumA += -0.0017 * Sin(arg)
    sumHPP +=  0.201 * Cos(arg)
    arg = 10*D-M;   sumP += -0.0104 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.067 * Cos(arg)
    arg = 16*D;     sumP +=  0.0086 * Sin(arg)
    sumHPP += -0.035 * Cos(arg)
    arg = 12*D-M;   sumP +=  0.0069 * Sin(arg)
    sumHPP += -0.038 * Cos(arg)
    arg = 5*D;      sumP +=  0.0066 * Sin(arg); sumA += -0.0004 * Sin(arg)
    sumHPP += -0.079 * Cos(arg)
    arg = 2*D+2*F;  sumP += -0.0053 * Sin(arg); sumA +=  0.0047 * Sin(arg)
    sumHPP +=  0.104 * Cos(arg);    sumHPA +=  0.031 * Cos(arg)
    arg = 18*D;     sumP += -0.0052 * Sin(arg)
    sumHPP +=  0.019 * Cos(arg)
    arg = 14*D-M;   sumP += -0.0046 * Sin(arg)
    sumHPP +=  0.023 * Cos(arg)
    arg = 7*D;      sumP += -0.0041 * Sin(arg)
    sumHPP +=  0.037 * Cos(arg)
    arg = 2*D+M;    sumP +=  0.0040 * Sin(arg); sumA +=  0.0005 * Sin(arg)
    sumHPP += -0.161 * Cos(arg);    sumHPA +=  0.043 * Cos(arg)
    arg = 20*D;     sumP +=  0.0032 * Sin(arg)
    sumHPP += -0.010 * Cos(arg)
    arg = D+M;      sumP += -0.0032 * Sin(arg); sumA +=  0.0036 * Sin(arg)
    sumHPP +=  0.157 * Cos(arg);    sumHPA +=  0.127 * Cos(arg)
    arg = 16*D-M;   sumP +=  0.0031 * Sin(arg)
    sumHPP += -0.014 * Cos(arg)
    arg = 4*D+M;    sumP += -0.0029 * Sin(arg)
    sumHPP +=  0.054 * Cos(arg)
    arg = 9*D;      sumP +=  0.0027 * Sin(arg)
    sumHPP += -0.020 * Cos(arg)
    arg = 4*D+2*F;  sumP +=  0.0027 * Sin(arg); sumA +=  0.0013 * Sin(arg)
    sumHPP += -0.037 * Cos(arg)

    arg = 2*D-2*M;  sumP += -0.0027 * Sin(arg); sumA +=  0.0022 * Sin(arg)
    sumHPP +=  0.104 * Cos(arg);    sumHPA +=  0.022 * Cos(arg)
    arg = 4*D-2*M;  sumP +=  0.0024 * Sin(arg); sumA +=  0.0010 * Sin(arg)
    sumHPP += -0.038 * Cos(arg)
    arg = 6*D-2*M;  sumP += -0.0021 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.022 * Cos(arg)
    arg = 22*D;     sumP += -0.0021 * Sin(arg)
    arg = 18*D-M;   sumP += -0.0021 * Sin(arg)
    arg = 6*D+M;    sumP +=  0.0019 * Sin(arg)
    sumHPP += -0.025 * Cos(arg)
    arg = 11*D;     sumP += -0.0018 * Sin(arg)
    sumHPP +=  0.011 * Cos(arg)
    arg = 8*D+M;    sumP += -0.0014 * Sin(arg)
    sumHPP +=  0.012 * Cos(arg)
    arg = 4*D-2*F;  sumP += -0.0014 * Sin(arg); sumA += -0.0004 * Sin(arg)
    sumHPP +=  0.013 * Cos(arg)
    arg = 6*D+2*F;  sumP += -0.0014 * Sin(arg); sumA +=  0.0004 * Sin(arg)
    sumHPP +=  0.017 * Cos(arg)
    arg = 3*D+M;    sumP +=  0.0014 * Sin(arg); sumA +=  0.0007 * Sin(arg)
    sumHPP += -0.030 * Cos(arg)
    arg = 5*D+M;    sumP += -0.0014 * Sin(arg)
    sumHPP +=  0.010 * Cos(arg)
    arg = 13*D;     sumP +=  0.0013 * Sin(arg)
    arg = 20*D-M;   sumP +=  0.0013 * Sin(arg)
    arg = 3*D+2*M;  sumP +=  0.0011 * Sin(arg)
    arg = 2*D-2*F+M;sumP += -0.0011 * Sin(arg)
    arg = D+2*M;    sumP += -0.0010 * Sin(arg)
    arg = 2*F-M;    sumP +=  0.0008 * Sin(arg)
    sumHPP +=  0.014 * Cos(arg)
    arg = 2*D+4*F;  sumP +=  0.0007 * Sin(arg)
    arg = 2*F-2*M;  sumP += -0.0006 * Sin(arg)
    arg = 2*D-2*F+2*M;sumP += -0.0006 * Sin(arg)
    arg = 24*D;     sumP +=  0.0006 * Sin(arg)
    arg = 4*D-4*F;  sumP +=  0.0005 * Sin(arg)
    arg = 2*D+2*M;  sumP +=  0.0005 * Sin(arg); sumA +=  0.0005 * Sin(arg)
    sumHPP += -0.023 * Cos(arg);    sumHPA +=  0.019 * Cos(arg)
    arg = D-M;      sumP += -0.0004 * Sin(arg); sumA += -0.0003 * Sin(arg)
    sumHPP +=  0.029 * Cos(arg)
    arg = 2*D-2*F;                              sumA += -0.0034 * Sin(arg)
    sumHPP += -0.392 * Cos(arg);    sumHPA += -0.023 * Cos(arg)
    arg = 2*M;                                  sumA +=  0.0006 * Sin(arg)
    sumHPP +=  0.023 * Cos(arg);    sumHPA += -0.016 * Cos(arg)
    arg = 2*F+M;                                sumA +=  0.0003 * Sin(arg)
    arg = 2*D+2*F-M;                            sumA +=  0.0003 * Sin(arg)
    arg = 2*D-2*F-M;sumHPP += -0.021 * Sin(arg)

    MoonDistP = 6378.14 / Sin(sumHPP/3600.)
    MoonDistA = 6378.14 / Sin(sumHPA/3600.)

    if isPerigee(k):
        JDE += sumP
        MoonDist = MoonDistP
        if verbose: print ("Perigee")
    elif isApogee(k):
        JDE += sumA
        MoonDist = MoonDistA
        if verbose: print ("Apogee")
    else:
        print ("k not for Perigee or Apogee: ", k)
    

    if verbose:
        print ("getPeriJDEbyK: k = %.2f" % k)
        print ("JDE0 = %.6f" % JDE0)
        print ("M(Sun), D, F, sumP, sumA = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (M, D, F, sumP, sumA)) 
        print ("sumP = %.5f, sumA = %.5f" % (sumP, sumA))
        print ("sumHPP = %.5f, sumHPA = %.5f" % (sumHPP, sumHPA))
        print ("JDE = %.5f" % JDE)
    return (JDE, MoonDist)

# getMoonPeriApo(year, useUT)
#   returns:
#   Perigee
#   Apogee
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonPeriApo(year, useUT):
    # chapter 48 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-1999.97)*13.2555  # counting perigee zero 48.2
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Perigee
    
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    print (" Moon Perigees and Apogees for the Year %4d" % year, comment)
    print (" -------------------------------------------", "-"*len(comment))
    print ()
    print ("             Perigee                          Apogee")
    
    for kint in range(kstart, kstart+16):
        dateList = ['', '']
        for k in (kint, kint+0.5):
            T  = k / 1325.55            # Julian centuries
            #y  = year
            y = int(T * 100 + 1999.97)
            (JDE, MoonDist) = getPeriJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
        
            # call getMoonData for distance of moon
            #(MoonRa, MoonDecl, MoonDict) = getMoonData(dt-deltaT, 0.0, False)
            #MoonDist = MoonDict["MoonDist"]
            # Super Moon, Mini Moon
            suffix = "   "
            if MoonDist <= 360000. : suffix = " ++" # Super Moon
            if MoonDist >= 405000. : suffix = " --" # Mini Moon

    
            (yyyy, dummy, dummy) = dt.getDate()
            if isPerigee(k) and yyyy == year: 
                #print ("Perigee  ", niceshortdatetime(dt))
                dateList[0] = niceshortdatetime(dt) + " %.0f"%MoonDist + suffix
            if isApogee(k): 
                #print ("Apogee   ", niceshortdatetime(dt))
                #if dt.year == year: 
                if yyyy == year: 
                    dateList[1] = niceshortdatetime(dt) + " %.0f"%MoonDist + suffix
                if len (dateList[0]) + len(dateList[1]) > 0: 
                    EventList.append(dateList)
                dateList = ['', '']
    
    return EventList
#########################################################      

# getMoonPeriApoxxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Perigee
#   Apogee
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonPeriApoxxx(dtCivicTime, zone, verbose=False):
    # chapter 48 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 

    #k0 = (year-1999.97)*13.2555  # counting perigee zero 48.2
    k0 = (Ty+0.03)*13.2555  # counting perigee zero 48.2
    k  = ceil(k0) - 2     # start with Perigee
    kmin = k
    kmax = k + 4
    EventList = []
    useUT = True
    
    for kint in range(kmin, kmax):
        for k in (kint, kint+0.5):
            T  = k / 1325.55            # Julian centuries
            y = int(T * 100 + 1999.97)
            (JDE, MoonDist) = getPeriJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
            dtCT = getCivicTimefromUT(dt, zone)

            # call getMoonData for declination of Moon
            (dummy, dummy, MoonDict) = getMoonData(dt, 0, False)
            MoonDecl = MoonDict["Decl"]
            # Super Moon, Mini Moon
            suffix = MoonDict["suffix"]
            mode = ''
            if isPerigee(k): 
                mode = 'Perigee'
            if isApogee(k): 
                mode = 'Apogee'
            Event = {'Object': 'Perigee',
                    'dt': dt,
                    'dtCivicTime' : dtCT,
                    'dist': MoonDist,
                    'decl': MoonDecl,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}
            EventList.append(Event)
    return EventList
#########################################################      
# Nodes of the Moon, Meeus Chapter 49
# hvf 02.05.2021
# based on Chapront's ELP - 2000/82 theory for the Moon and VSOP87

import sys

VersionMoon = "V3.1 hvf 2023-11-29 MoonNodesMain.py using ELP-2000/82 and VSOP Ephemerides"
VersionList.append(VersionMoon)

def Usage(progname):
    print ("Usage: %s [UTC|TT] year [year ...]" % progname)
    sys.exit(0)

def isAscendingNode(k):   # k must be integer
    test = (abs(k - int(k)) < 0.001)
    #if verbose: print ("isAscendingNode k, test = ", k, test)
    return test     # True or False

def isDescendingNode(k):  # k must be half-integer
    test = (abs(abs(k - int(k)) - 0.5) < 0.001)
    #if verbose: print ("isDescendingNode k, test = ", k, test)
    return test     # True or False 

# getNodesJDEbyK (k, verbose = False)
# k counts the number of new moons since 2000-01-06
# Meeus chapter 49
# returns the Julian Ephemeris Days (JDE) time of the node passing  of the Moon
def getNodesJDEbyK(k, verbose = False):
    # k integer:        Ascending Node
    # k integer + 0.5 : Descending Node

    T  = k / 1342.23            # (49.1), Julian centuries after Epoch2000
    E = 1 - 0.002516*T - 0.0000074*T*T      # (45.6)
    EE = E*E

    # D
    D = 183.6380 + 331.73735691*k + 0.0015057*T*T \
        + 0.00000209*T**3 - 0.000000010*T**4
    D = Mod(D, 360.)

    # Sun's Mean anomaly 
    M = 17.4006 + 26.82037250*k - 0.0000999*T*T - 0.00000006*T*T*T
    M = Mod(M, 360.)

    # Moon's Mean anomaly 
    MM = 38.3776 + 355.52747322*k + 0.0123577*T*T \
            + 0.000014628*T**3 - 0.000000016*T**4
    MM = Mod(MM, 360.)

    # Longitude of the Ascending Node of the lunar orbit 
    Omega = 123.9767 - 1.44098949*k \
            + 0.0020625*T*T + 0.00000214*T**3 - 0.000000016*T**4
    Omega = Mod(Omega, 360.)

    V = 299.75 + 132.85*T - 0.009173*T*T

    P = Omega + 272.75 - 2.3*T

    JDE = 2451565.16195 + 27.212220817*k \
            + 0.0002572*T*T \
            + 0.000000021*T*T*T \
            - 0.00000000088*T**4 \
            - 0.4721 * Sin(MM) \
            - 0.1649 * Sin(2*D) \
            - 0.0868 * Sin(2*D-MM) \
            + 0.0084 * Sin(2*D+MM) \
            - 0.0083*E*Sin(2*D-M) \
            - 0.0039*E*Sin(2*D-M-MM) \
            + 0.0034 * Sin(2*MM) \
            - 0.0031 * Sin(2*D-2*MM) \
            + 0.0030*E*Sin(2*D+M) \
            + 0.0028*E*Sin(M-MM) \
            + 0.0026*E*Sin(M) \
            + 0.0025 * Sin(4*D) \
            + 0.0024 * Sin(D) \
            + 0.0022*E*Sin(M+MM) \
            + 0.0017 * Sin(Omega) \
            + 0.0014 * Sin(4*D-MM) \
            + 0.0005*E*Sin(2*D+M-MM) \
            + 0.0004*E*Sin(2*D-M+MM) \
            - 0.0003*E*Sin(2*D-2*M) \
            + 0.0003*E*Sin(4*D-M) \
            + 0.0003 * Sin(V) \
            + 0.0003 * Sin(P)

    if isAscendingNode(k):
        if verbose: print ("Ascending Node")
    elif isDescendingNode(k):
        if verbose: print ("Descending Node")
    else:
        print ("k not for Ascending or descending Node: ", k)

    if verbose:
        print ("getNodesJDEbyK: k = %.2f" % k)
        dt = getGreenwichDate(JDE)
        print ("JDE = %.6f" % JDE, nicedatetime(dt))
        print ("E, M(Sun), MM(Moon), D, Omega = %.7f, %.4f, %.4f, %.4f, %.4f" \
            % (E, M, MM, D, Omega)) 
        print ("V, P = %.7f, %.7f" % (V, P))
    return JDE

# getMoonNodes(year, useUT)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonNodes(year, useUT):
    # chapter 49 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    intyear = int(year)
    if intyear != year:
        print ("Year must be integer: ", year)
        sys.exit(1)

    k0 = (year-2000.05)*13.4223  # k0 = 0 is ascending node on 2000-01-21
    if (k0 == int(k0)): kstart = k0
    else: kstart  = ceil(k0) - 1     # start with Ascending Node
    y  = year
    EventList = []
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for kint in range(kstart, kstart+16):
        dateList = ['', '']
        foundDate = False
        for k in(kint, kint+0.5):
            T  = k / 1342.23            # Julian centuries after Epoch2000
            y = int(T * 100 + 2000)
            JDE = getNodesJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT

            (yyyy, dummy, dummy) = dt.getDate()
            #if isAscendingNode(k) and dt.year == year: 
            if isAscendingNode(k) and yyyy == year: 
                dateList[0] = nicedatetime(dt)
                foundDate = True

            #if isDescendingNode(k) and dt.year == year: 
            if isDescendingNode(k) and yyyy == year: 
                dateList[1] = nicedatetime(dt)
                foundDate = True

        if foundDate: EventList.append(dateList)
        #print ("k, T, y =", k, T, y)
    return EventList
#########################################################     

# getMoonNodesxxx(dtCivicTime, zone, verbose=False)
#   returns:
#   Ascending Nodes
#   Descending Nodes
#   for the calender year 
#   useUT: convert ephemerides to UTC
#
def getMoonNodesxxx(dtCivicTime, zone, verbose=False):
    # chapter 49 in  Meeus
    Epoch2000 = 2451545.0   # just to remember
    dtUT = getUTfromCivicTime(dtCivicTime, zone)
    JD = getJulianDay(dtUT)
    TT = JD - Epoch2000 + getDeltaT(JD)/86400.
    #T = TT / 365250.    # eq. (31.1) Julian Millennia
    Ty = TT / 365.25    # years and fractions after Epoch2000
    # 
    k0 = (Ty-0.05)*13.4223  # k0 = 0 is ascending node on 2000-01-21
    kstart  = ceil(k0) - 1     # start with Ascending Node
    kmin = kstart - 2
    kmax = kstart + 3
    EventList = []
    useUT = True
    if useUT: comment = "Universal Time Coordinated (UTC)"
    else: comment = "Terrestrial Time (TT)"
    for kint in range(kmin, kmax+1):
        dateList = ['', '']
        foundDate = False
        for k in(kint, kint+0.5):
            T  = k / 1342.23            # Julian centuries after Epoch2000
            y = int(T * 100 + 2000)
            JDE = getNodesJDEbyK(k, False)
            dt = getGreenwichDate(JDE)
            deltaT = timedelta(seconds=getDeltaT(JDE))
            if useUT: dt -= deltaT
            dtCT = getCivicTimefromUT(dt, zone)

           # call getMoonData for declination, distance of Moon
            (dummy, MoonDecl, MoonDict) = getMoonData(dt, 0, False)
            MoonDist = MoonDict['MoonDist']
            # Super Moon, Mini Moon
            suffix = MoonDict["suffix"]
            MoonDecl = MoonDict["Decl"]
            mode = '  '
            if isAscendingNode(k):
                mode = 'Ascending'
            if isDescendingNode(k):
                mode = 'Descending'

            # here we cheat a little bit: since getMoonData does not predict
            # latitude zero for the nodes, we set it here explicitly
            MoonDict['MoonLat'] = 0.0

            Event = {'Object': 'Nodes',
                    'dt': dt,
                    'dtCivicTime': dtCT,
                    'dist': MoonDist,
                    'decl': MoonDecl,
                    'mode': mode,
                    'suffix': suffix,
                    'Dict': MoonDict}

            EventList.append(Event)

    return EventList
#########################################################     

Version = "V3.1 hvf 2023-11-30 SunPlanetsTail.py with Planets Data, Events, Magnitudes"
VersionList.append(Version)
#######################################################################
# uses timezone data from pytz
# hvf 07.04.2021
# replace getRTS() by getObjectTimes() for more precise rise and set times
# hvf 15.06.2022 07.08.2022
# use hvfTime() instead of datetime()
# add Terrestrial Time TT to output - hvf 4.12.2022
# add duration of twilight phases - hvf 19.08.2023
# distinguish time zones of server, client - hvf 08.05.2024
import pytz
from itertools import chain

Version = "V8.9 hvf 2024-05-08 using VSOP87 ephemerides \n  and known timezones, flexible date input format, years from -9999 to 9999"
longVersion = "V8.9 hvf 2024-05-08 SunPlanetsMain-TZ.py using VSOP87 ephemerides \n  and known timezones, flexible date input format, years from -9999 to 9999"
VersionList.append(longVersion)

def shortUsage(progname):
    print ("Run %s --help for more explanations." % progname)

def Usage(progname):
    global Version
    #print ("%s %s" % (progname, Version))
    print ()
    print ("Run %s --help for this explanation." % progname)
    print ("Run %s --list for a list of the predefined locations." % progname)
    print ("Run %s --listtimezones for a list of all known timezones." %
            progname)
    print ("Run %s --check location to check if location is known or not." %
            progname)
    print ("Run %s --new newlocation --lon xx.xx --lat yy.yy for a new location, longitude and latitude must also be specified." % progname)
    print ("Run %s for default location, current date and time." % progname)
    print ("Run %s knownlocation for known location, current date and time." %
            progname)
    print ("  Known locations supply longitude, latitude, and timezone.")
    print ("  Longitude is in degrees East of Greenwich.")
    print ("  Latitude is in degrees, positive for North, negative for South.")
    print ("Additional Options:")
    print ("--date yyyy-mm-dd  Date in ISO format, e.g. 2021-05-19")
    print ("  also dd.mm.yyyy, -dd.mm.yyyy, -yyyy-mm-dd, yyyy between -9999 and 9999")
    print ("--time HH:MM:SS.sss Time in ISO format and variants, e.g. 13:15:23")
    print ("  also HH, HHMM, HHMMSS, HHMMSS:sss")
    print ("--zone hoursEast Timezone in hours East of Greenwich (UTC), e.g. 2")
    print ("--timezone knownTimezone  Timezone, e.g. Europe/Zurich")
    print ("  Date and Time default to current day and local time unless specified.")
    print ()
    return

# getUTfromCivicTimezone (dt, tdUTCOffset)
# dt datetime object
# tdUTCOffset   timedelta object
# returns a datetime object

def getUTfromCivicTimezone (dt, tdUTCOffset):
    return dt - tdUTCOffset

# niceTimeDeltaHHMM() returns a string +-HH:MM
def niceTimeDeltaHHMM(td):  
    # td is a timedelta, positive or negative
    if td.total_seconds() < 0.0:  sign = '-'; td = - td
    else:       sign = '+'
    hours = int(td.total_seconds() / 3600.)
    td -= timedelta(hours=hours)
    mins = int(td.total_seconds() / 60.)
    #secs = int(Mod(td.total_seconds(), 60.))
    tdstr = sign + "%2.2d:%2.2d" % (hours, mins)
    return tdstr

# getTZoffset(Timezone)
# returns the offset in seconds of dtCivicTime for Timezone compared to UTC
def getTZoffset(dtCivicTime, Timezone):
    #print (" dt = ", dtCivicTime, ", Timezone = ", Timezone)
    dt_with_tz = pytz.timezone(Timezone).localize(dtCivicTime)
    tdUTCOffset = dt_with_tz.utcoffset()
    #Zone = tdUTCOffset.total_seconds() / 3600.
    return tdUTCOffset   

def getTimezoneDetails(year):
# prints all timezones with their offsets
    jan = str(year) + "-01-10 1200"
    jul = str(year) + "-07-10 1200"
    print ("Timezones and their Offsets for year ", year)
    nlines = 0
    # timezones per country
    country_timezones_list = [i for i in pytz.country_timezones]
    for c in country_timezones_list:
        #print ("Country ", c, sorted(pytz.country_timezones(c)))
        for ctz in sorted(pytz.country_timezones(c)):
            if Mod(nlines, 50) == 0:
                print ()
                print ("Country   Timezone                    Offset in ", year)
                print ("                                      January  July")
            nlines += 1
            dtjan = dateutil.parser.isoparse(jan)
            dtjul = dateutil.parser.isoparse(jul)
            janOffset = getTZoffset(dtjan, ctz)
            julOffset = getTZoffset(dtjul, ctz)
            janStr = niceTimeDeltaHHMM(janOffset)
            julStr = niceTimeDeltaHHMM(julOffset)
            print ("%2.2s    %-30s  %6.6s   %6.6s" % (c, ctz, janStr,julStr))
    print ("%d timezones found." % nlines)
    nlines = 0

    # timezones not assigned to a country
    tz_notusedList = []
    for tz in pytz.all_timezones:
        country_in_tz = [c for c in pytz.country_timezones if tz in pytz.country_timezones(c)]
        if len(country_in_tz) > 0:
            #print ("Timezone", tz, country_in_tz)
            pass
        else:
            tz_notusedList.append(tz)
    for tz in sorted(tz_notusedList):
        if Mod(nlines, 50) == 0:
            print ()
            print ("Timezone                              Offset in ", year)
            print ("                                      January  July")
        nlines += 1
        dtjan = dateutil.parser.isoparse(jan)
        dtjul = dateutil.parser.isoparse(jul)
        janOffset = getTZoffset(dtjan, tz)
        julOffset = getTZoffset(dtjul, tz)
        janStr = niceTimeDeltaHHMM(janOffset)
        julStr = niceTimeDeltaHHMM(julOffset)
        print ("%-34s    %6.6s   %6.6s" % (tz, janStr,julStr))
    print ("%d timezones found." % nlines)
    return


# Main program
if __name__ == '__main__':
    
    # preset some locations
    Locations = {
        "Zurich": {"lon": 8.55, "lat": 47.3767, "zone": 1, "tz" : "Europe/Zurich"},
        "Gockhausen": {"lon": 8.592861, "lat": 47.381643, "zone": 1, "tz" : "Europe/Zurich"},
        "Lenk": {"lon": 7.44810, "lat": 46.46025, "zone": 1, "tz" : "Europe/Zurich"},
        "Clansayes": {"lon": 4.8273, "lat": 44.3923, "zone": 1, "tz" : "Europe/Paris"},
        "Paris": {"lon": 2.550, "lat": 49.013, "zone": 1, "tz" : "Europe/Paris"},
        "Greenwich": {"lon": 0.0, "lat": 51.48125, "zone": 0, "tz" : "GMT"},
        "Auckland": {"lon": 174.7667, "lat": -36.85, "zone": 12, "tz" : "Pacific/Auckland"},
        "Wellington": {"lon": 174.767, "lat": -41.283, "zone": 12, "tz" : "Pacific/Auckland"},
        "Christchurch": {"lon": 172.532, "lat": -43.489, "zone": 12, "tz" : "Pacific/Auckland"},
        "KapHoorn": {"lon": -67.2892, "lat": -55.9678, "zone": -3, "tz" : "America/Punta_Arenas"},
        "Hammerfest": {"lon": 23.6833, "lat": 70.6667, "zone": 1, "tz" : "Europe/Oslo"},
        "Barentsburg": {"lon": 14.2167, "lat": 78.0667, "zone": 1, "tz" : "Arctic/Longyearbyen"},
        "Nordpol": {"lon": 0.0, "lat": 90, "zone": 0, "tz" : "GMT"},
        "Suedpol": {"lon": 0.0, "lat": -90, "zone": 13, "tz" : "Antarctica/South_Pole"},
        "McMurdoStation": {"lon": 166.6667, "lat": -77.85, "zone": 13, "tz" : "Antarctica/McMurdo"},
        "HongKong": {"lon": 114.1667, "lat": 22.3, "zone": 8, "tz" : "Asia/Hong_Kong"},
        "Singapur": {"lon": 103.8333, "lat": 1.2833, "zone": 8, "tz" : "Asia/Singapore"},
        "Anchorage": {"lon": -149.8833, "lat": 61.1833, "zone": -8, "tz" : "America/Anchorage"},
        "Hawaii": {"lon": -157.85, "lat": 21.3167, "zone": -10, "tz" : "Pacific/Honolulu"},
        "Honolulu": {"lon": -157.85, "lat": 21.3167, "zone": -10, "tz" : "Pacific/Honolulu"},
        "Adelaide": {"lon": 138.600, "lat": -34.917, "zone": 9.5, "tz" : "Australia/Adelaide"},
        "Brisbane": {"lon": 153.033, "lat": -27.467, "zone": 10, "tz" : "Australia/Brisbane"},
        "BuenosAires": {"lon": -58.536, "lat": -34.822, "zone": -3, "tz" : "America/Argentina/Buenos_Aires"},
        "Canberra": {"lon": 149.117, "lat": -35.300, "zone": 10, "tz" : "Australia/Canberra"},
        "Sydney": {"lon": 151.200, "lat": -33.850, "zone": 10, "tz" : "Australia/Sydney"},
        "Chicago": {"lon": -87.905, "lat": 41.979, "zone": -5, "tz" : "America/Chicago"},
        "Doha": {"lon": 51.565, "lat": 25.261, "zone": 3, "tz" : "Asia/Qatar"},
        "Dubai": {"lon": 55.364, "lat": 25.253, "zone": 4, "tz" : "Asia/Dubai"},
        "Hobart": {"lon": 147.317, "lat": -42.867, "zone": 10, "tz" : "Australia/Hobart"},
        "Melbourne": {"lon": 144.950, "lat": -37.800, "zone": 10, "tz" : "Australia/Melbourne"},
        "Perth": {"lon": 115.850, "lat": -31.950, "zone": 8, "tz" : "Australia/Perth"},
        "Johannesburg": {"lon": 28.067, "lat": -26.200, "zone": 2, "tz" : "Africa/Johannesburg"},
        "Kapstadt": {"lon": 18.597, "lat": -33.969, "zone": 2, "tz" : "Africa/Johannesburg"},
        "LosAngeles": {"lon": -118.408, "lat": 33.943, "zone": -7, "tz" : "America/Los_Angeles"},
        "SanFrancisco": {"lon": -122.375, "lat": 37.619, "zone": -7, "tz" : "America/Los_Angeles"},
        "NewYork": {"lon": -73.779, "lat": 40.640, "zone": -4, "tz" : "America/New_York"},
        "Montreal": {"lon": -73.567, "lat": 45.500, "zone": -4, "tz" : "America/Montreal"},
        "Vancouver": {"lon": -123.184, "lat": 49.194, "zone": -7, "tz" : "America/Vancouver"},
        "Santiago": {"lon": -70.667, "lat": -33.450, "zone": -3, "tz" : "America/Santiago"},
        "Ushuaia": {"lon": -68.300, "lat": -54.800, "zone": -3, "tz" : "America/Argentina/Ushuaia"},
        "Tokyo": {"lon": 139.7667, "lat": 35.6833, "zone": 9, "tz" : "Asia/Tokyo"},
        "Peking": {"lon": 116.3833, "lat": 39.9333, "zone": 8, "tz" : "Asia/Shanghai"},
        "Moskau": {"lon": 37.6167, "lat": 55.7500, "zone": 3, "tz" : "Europe/Moscow"},
        "Kiev": {"lon": 30.5164, "lat": 50.4536, "zone": 2, "tz" : "Europe/Kiev"},
        "East_Cape_NZ":  {"lon": 178.548, "lat": -37.689, "tz": "Pacific/Auckland", "zone": 12 }, 
        "Gisborne_NZ":  {"lon": 178.018, "lat": -38.6625, "tz": "Pacific/Auckland", "zone": 12 }, 
        "Hastings_NZ":  {"lon": 176.85, "lat": -39.63, "tz": "Pacific/Auckland", "zone": 12 }, 
        "Cape_Reinga_NZ":  {"lon": 172.677, "lat": -34.427, "tz": "Pacific/Auckland", "zone": 12 }, 
        "Cape_Tribulation_QLD":  {"lon": 145.463, "lat": -16.087, "tz": "Australia/Brisbane", "zone": 10 }, 
        "Cairns_QLD":  {"lon": 145.783, "lat": -16.916, "tz": "Australia/Brisbane", "zone": 10 }, 
        "Emerald_QLD":  {"lon": 148.158, "lat": -23.525, "tz": "Australia/Brisbane", "zone": 10 }, 
        "Agnes_Water_QLD":  {"lon": 151.903, "lat": -24.209, "tz": "Australia/Brisbane", "zone": 10 }, 
        "Brisbane_QLD":  {"lon": 153.033, "lat": -27.467, "tz": "Australia/Brisbane", "zone": 10 }, 
        "Bangkok":  {"lon": 100.5, "lat": 13.75, "tz": "Asia/Bangkok", "zone": 7 }, 

    }

############### input reading and parsing ######################    
    print ("###########################################################")
    progname = os.path.basename(sys.argv[0])
    for arg in sys.argv:
        print ("%s " % arg, end='')
    print ()
    print ("%s Version %s" % (progname, Version))
    # use current day and time as starting point unless explicit data are given
    # UTC as default
    locationFlag = False
    newlocFlag = False
    lonFlag = False
    latFlag = False
    Zone = 0
    Timezone = "UTC"   # timezone
    zoneFlag = False
    tzFlag = False
    dateFlag = False
    timeFlag = False
    day = None
    daytime = None
    verboseFlag = False
    verbose = False
    # read input parameters: location latitude longitude [Zone yyyy-mm-dd HH:MM:SS]
    # --help
    # --list
    # --listtimezones
    # --check location
    # location
    # --new location
    # --lon longitude
    # --lat latitude
    # --zone hours
    # --timezone timezone
    # --date date
    # --time time
    # --verbose

    # set default location and current date and time
    if len(sys.argv) <= 1:
        # Gockhausen (from Google Maps)
        location = "Gockhausen (CH)"
        longitude = 8.592861
        latitude =  47.381643
        Timezone = "Europe/Zurich"
        tzFlag = True
        lonFlag = True
        latFlag = True

    while len(sys.argv) > 1:
        param = sys.argv[1]
        # split into words if necessary
        paramList = param.split()
        # print ("ParamList = ", paramList)
        if len(paramList) > 1:
            sys.argv[2:2] = paramList
            # print ("sys.argv = ", sys.argv)
            param = sys.argv[2]
            # print ("param = ", param)
            del sys.argv[1:2]
            # print ("sys.argv = ", sys.argv)
            # print ("param = ", param)
        del sys.argv[1]

        if param == '--help':
            for v in VersionList:
                print (v)
            Usage(progname)
            sys.exit(0)

        elif param == '--list':
            print ("Location      Latitude  Longitude    Zone Timezone")
            for location in sorted(Locations):
                longitude = Locations[location]["lon"]
                latitude  = Locations[location]["lat"]
                Zone      = Locations[location]["zone"]
                Timezone  = Locations[location]["tz"]
                print ("%-12s  %8.3f   %8.3f   %5.1f %s" % \
                    (location, latitude, longitude, Zone, Timezone))
            sys.exit(0)

        elif param == '--listtimezones':
            dt = dateutil.parser.isoparse(str(date.today()))
            getTimezoneDetails(dt.year)
            sys.exit(0)

        elif param == '--check':
            try:
                location = sys.argv[1]
                del sys.argv[1]
            except Exception as e:
                print ("Missing Parameter: Location\n") 
                shortUsage(progname)
                sys.exit(0)

            if location in Locations:
                longitude = Locations[location]["lon"]
                latitude  = Locations[location]["lat"]
                Zone      = Locations[location]["zone"]
                # CAREFUL: Zone does not account for daylight saving time!!
                Timezone  = Locations[location]["tz"]
                print ("%-12s  %8.3f   %8.3f   %5.1f %s" % \
                    (location, latitude, longitude, Zone, Timezone))
                lonFlag = True
                latFlag = True
                tzFlag  = True
            else:
                print ("Location %s is not known.\n" % location)
                shortUsage(progname)
            sys.exit(0)

        elif param == '--new':
            try:
                location = sys.argv[1]
                del sys.argv[1]
                newlocFlag = True
                locationFlag = True
            except Exception as e:
                print ("Missing Parameter: Location\n") 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--lon':
            try:
                longitude = float(sys.argv[1])
                del sys.argv[1]
                lonFlag = True
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--lat':
            try:
                latitude = float(sys.argv[1])
                del sys.argv[1]
                latFlag = True
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)
            
        elif param == '--zone':
            try:
                Zone = float(sys.argv[1])
                del sys.argv[1]
                zoneFlag = True
                Timezone = 'GMT'
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--timezone':
            try:
                Timezone = sys.argv[1]
                del sys.argv[1]
                if Timezone not in pytz.all_timezones_set:
                    print ("Timezone %s is not valid." % Timezone)
                    Usage(progname)
                    sys.exit(0)
                zoneFlag = False
                tzFlag = True
            except Exception as e:
                print ("Error ", e)
                shortUsage(progname)
                sys.exit(0)

        elif param == '--date':
            try:
                day = sys.argv[1]
                del sys.argv[1]
                dateFlag = True
                # print ("date day = ", day)
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        elif param == '--time':
            try:
                daytime = sys.argv[1]
                del sys.argv[1]
                timeFlag = True
            except Exception as e:
                print ("Error ", e) 
                shortUsage(progname)
                sys.exit(0)

        # Lookup known locations
        elif param in Locations:
            location  = param
            longitude = Locations[location]["lon"]
            latitude  = Locations[location]["lat"]
            Zone      = Locations[location]["zone"]
            Timezone  = Locations[location]["tz"]
            tzFlag = True
            lonFlag = True
            latFlag = True
            zoneFlag = False
            locationFlag = True

        # see verbose flag
        elif param == "--verbose":
            verboseFlag = True
            verbose = True

        else:
        # Unknown location, the whole set
            print ("Parameter %s not known." % param)
            shortUsage(progname)
            sys.exit(0)
    # All parameters have been read, now process them

    # no location set, use our default
        if not locationFlag:
        # Gockhausen (from Google Maps)
            location = "Gockhausen (CH)"
            longitude = 8.592861
            latitude =  47.381643
            Timezone = "Europe/Zurich"
            tzFlag = True
            lonFlag = True
            latFlag = True
            locationFlag = True

    # yyyy-mm-dd
    today = hvfTime.today() # (y, m, d), d has fractions
    #print ("today = ", today)
    # HH:MM:SS
    now = hvfTime.now()     # (h, m, s), localtime of server
    if dateFlag: 
        td = day    # string
    else:
        #td = today  # tupel
        td = "%04d-%02d-%02d" % today

    if timeFlag:
        tt = daytime    # string
    else:
        #tt = now        # tupel, localtime of server
        tt = "%02d:%02d:%02d" % now
    t1 = str(td) + " " + tt
    # dt is a hvfTime object
    try:
        #dttemp = dateutil.parser.isoparse(t1)
        dttemp = parseDateTime(t1)
        dt = hvfTime(dttemp)
    except Exception as e:
        print ("Error ", e)
        print ("Cannot parse '%s' into datetime." % t1)
        shortUsage(progname)
        sys.exit(0)
    dtServer = dt   # if daytime not specified, we want to keep local and
                    # server time separate
########################################################################        
    # check server timezone here
    ServerTimezone = getTimezone(verbose)
    dtServer_with_tz = pytz.timezone(ServerTimezone).localize(dttemp.getDatetime())
    ServerTzinfo = dtServer_with_tz.tzinfo
    Server_tdUTCOffset = dtServer_with_tz.utcoffset()
    ServerZone = Server_tdUTCOffset.total_seconds() / 3600.
    Server_tdstring = niceTimeDeltaHHMM(Server_tdUTCOffset)
    Server_zonetime_with_tz = dtServer_with_tz
    if verbose:
        print ("Server utcoffset = ", Server_tdstring)
        print ("Server Timezone = ", ServerTimezone)
        print ("Server zone = ", ServerZone)
        print ("Server time = ", dtServer)

    # dt is the hvfTime object of the current time at the object location
    # we correct by the server time if we use the current time clock
    # i.e. --time not set
    # PROBLEM: Zone is the static value, does not account for tzflag DST 
    # apply timezone
    if zoneFlag:    # offset in hours
        dt_with_tz  = getUTfromCivicTime(dt, Zone)
        dt_with_tz  = dt
        tdUTCOffset = timedelta(hours=Zone)
        tdstring = niceTimeDeltaHHMM(tdUTCOffset)
        Timezone = tdstring
        zonetime_with_tz = str(dt_with_tz) + tdstring
    elif tzFlag:    # offset by name
        try:
            dt_with_tz = pytz.timezone(Timezone).localize(dttemp.getDatetime())
            tzinfo = dt_with_tz.tzinfo
            #print ("tzinfo = ", tzinfo)
            tdUTCOffset = dt_with_tz.utcoffset()
            Zone = tdUTCOffset.total_seconds() / 3600.
            if not timeFlag:
                timeCorr = Zone-ServerZone
                dt = dtServer + timedelta(hours=timeCorr)
                dttemp = dt
                if verbose:
                    print ("time correction (hours) = ", timeCorr)
                    print ("new dt = ", dt)
            tdstring = niceTimeDeltaHHMM(tdUTCOffset)
            dt_with_tz = pytz.timezone(Timezone).localize(dttemp.getDatetime())
            tzinfo = dt_with_tz.tzinfo
            zonetime_with_tz = dt_with_tz
        except:     # year outside datetime range
            Timezone = 'UTC'
            Zone = 0
            dt_with_tz  = dt
            tdUTCOffset = timedelta(hours=Zone)
            tdstring = niceTimeDeltaHHMM(tdUTCOffset)
            zonetime_with_tz = str(dt_with_tz) + tdstring

    #print ("utcoffset = ", tdstring)
    #print ("Timezone = ", Timezone)

    # check longitude and latitude
    if not lonFlag:
        print ("No Longitude specified.\n")
        shortUsage(progname)
        sys.exit(0)
    if not latFlag:
        print ("No Latitude specified.\n")
        shortUsage(progname)
        sys.exit(0)

    if abs(latitude) > 90.0:
        print ("Latitude ", latitude, " must be between -90 and +90 degrees.")
        shortUsage(progname)
        sys.exit(0)
########################################################################
# hier geht es richtig los
    #verbose = False
    verbose = verboseFlag
    print ("Location                    =", location)
    print ("Latitude                    =", latitude, " degrees N")
    print ("Longitude                   =", longitude, " degrees E")
    print ("Timezone                    =", Timezone)
    print ("Zonetime, Day of Week       =", dt, getDayofWeek(dt))
    print ("Zonetime with TZ            =", zonetime_with_tz)
    print ("Zone                        =", Zone, "hours East of Greenwich") 
    dtLocalCivicTime = dt
    # convert local time zone to Greenwich mean time (UTC)
    #UTC = getUTfromCivicTime(dt, Zone)
    UTC = getUTfromCivicTimezone(dt, tdUTCOffset)
    # mean local time
    meanLT = getMeanLocalTime(UTC, longitude)
    print ("UTC Universal Time          =", UTC)
    print ("Mean Local Time             =", meanLT)

    # Zeitgleichung, wahre Ortszeit
    # Julian Day für Greenwich mean time
    #dtGMT = dt-timedelta(hours=Zone)
    dtGMT = UTC
    JD0 = getJulianDay(dtGMT)
    # find Julian Date for midnight GMT
    #dtGMTmidnight = datetime(dtGMT.year, dtGMT.month, dtGMT.day)
    (y, m, d) = dtGMT.getDate()
    d = floor(d)
    dtGMTmidnight = hvfTime(y, m, d, 0, 0, 0)
    JD1 = getJulianDay(dtGMTmidnight)
    # td is time eq. as a timedelta object, zgstr is a printable 
    (td, zgstr, SunStuffDict) = getZeitgleichung(JD0, False)
    # wahre Ortszeit
    trueLT = meanLT + td
    # Greenwich date
    greenwichDate = getGreenwichDate(JD0)
    greenwichDateJD1 = getGreenwichDate(JD1)
    # Greenwich mean sideral time
    greenwichST = getGMSTfromJD(JD0)
    localST = getLMSTfromGMST(greenwichST, longitude)
    deltaT = getDeltaT(JD0)
    TT = greenwichDate + timedelta(seconds=deltaT)
    JDE = getJulianDay(TT)    # Julian Ephemeris Day

    print ("Time Equation               = %s" % zgstr, " minutes")
    print ("True Local Time             = %s" % nicedatetime(trueLT))
    print ("Julian Day  at midnight GMT = ", JD1)
    print ("Greenwich Date midnight GMT = ", nicedatetime(greenwichDateJD1))
    print ()
    print ("Julian Day JD               (current time) = ", JD0)
    print ("Greenwich Date              (current time) = ", nicedatetime(greenwichDate))
    print ("Greenwich Mean Sideral Time (current time) = ", nicedatetime(greenwichST))
    print ("Local Mean Sideral Time     (current time) = ", nicedatetime(localST))
    print ("Longitude (hours)           = ", hmsstring(longitude/15.))
    print ("Mean Local Noon             = ", hmsstring(12.0+Zone-longitude/15.))
    print ("deltaT (sec)                = ", round(deltaT, 1))
    print ("Terrestrial Time TT         (current time) = ", nicedatetime(TT))
    print ("Julian Ephemeris Day JDE    (current time) = ", JDE)
    print ()
    # get Sun Decl, RA, hour angle, altitude, azimut
    # at current local civic time
    (RA, Decl, SunDict) = getSunData(dt, Zone, False)
    RAh = RA/15. # convert to hours
    # Stundenwinkel, hour angle
    Hour = localST - timedelta(hours=RAh)
    # use hvfTime()
    (h, m, s) = Hour.getTime()
    Hourdecimal = h + m/60. + s/3600.
    Hourdecimalstring = "%02d:%02d:%02d" % (h, m, s)
    #Hourdecimal = Hour.time().hour + Hour.time().minute/60. \
    #            + Hour.time().second/3600 + Hour.time().microsecond/3600e6
    Hourgrad = Hourdecimal*15.
    # get azimut, altitude from hour angle, declination, latitude
    (azi, alt, xyzDict) = getAziAltfromHourDecl(Hourgrad, Decl, latitude)
    SunSign = getSign(SunDict["SunLong"])

    print ("Sun Longitude     (degrees) = %8.3f" % SunDict["SunLong"])
    print ("Sun Latitude      (degrees) = %8.5f" % SunDict["SunLat"])
    print ("Sun RA            (hours)   = ", hmsstring(RAh))
    print ("Sun RA            (degrees) = %8.3f" % RA)
    print ("Sun Declination   (degrees) = %8.5f" % Decl)
    print ("Sun Hour Angle    (hours)   = ", Hourdecimalstring)
    print ("Sun Hour Angle    (degrees) = ", round(Hourgrad, 3))
    print ("Azimut            (degrees) = %6.1f" % azi)
    print ("Altitude          (degrees) = %6.1f (no refraction)" % alt)
    print ("SunSign                     = ", SunSign)
    print ("Distance Sun-Earth     (km) =  %d" % SunDict["R"])
    print ("Distance Sun-Earth     (au) =  %.8f" % SunDict["Rau"])
    print ()

#############  Sun Rise time, Set time ##############################
    # get transit time, altitude, declination
    SunEventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                None, False)

    #print ("SunEventList = ", SunEventList)
    # Sun  Rise, Set, and twilight events
    TwilightList = []
    RiseTime = None; SetTime = None # initialize for later
    RiseList = [e for e in SunEventList if e["event"] == "Rise"]
    for e in RiseList:
        RiseTime = e["dt"]
        azi = e["azi"]
        alt = e["alt"]
        TwilightList.append((RiseTime, "Sun Rise", azi, alt))
    SetList = [e for e in SunEventList if e["event"] == "Set"]
    for e in SetList:
        SetTime = e["dt"]
        azi = e["azi"]
        alt = e["alt"]
        TwilightList.append((SetTime, "Sun Set", azi, alt))
 
    # find lowest and highest altitude SunMinAlt and SunMaxAlt
    # we will need this to find out if we have twilight, midnight sun, etc.
    Alt = SunEventList[0]["Dict"]["Alt"]
    AltList = []
    # changed search string - hvf 13.03.2023
    for e in SunEventList:
        if "North " in e["event"]:
            AltList.append(e["alt"])
        if "South " in e["event"]:
            AltList.append(e["alt"])
            

    SunMinAlt = min(AltList)
    SunMaxAlt = max(AltList)
    #################3
    #print ("SunMinAlt = ", SunMinAlt)
    #print ("SunMaxAlt = ", SunMaxAlt)

    # civil twilight, 6 degrees below horizon
    Sun06EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                -6., False)
    # find Rise event
    Rise06List = [e for e in Sun06EventList if e["event"] == "Rise"]
    for e in Rise06List:
        Rise06Time = e["dt"]
        TwilightList.append ((Rise06Time, "Civil Twilight Begin"))

    # find Set event - may have 2 Set events, e.g 2022-06-09 Gockhausen
    Set06List = [e for e in Sun06EventList if e["event"] == "Set"]
    for e in Set06List:
        Set06Time = e["dt"]
        TwilightList.append ((Set06Time, "Civil Twilight End"))

    # nautical twilight, 12 degrees below horizon
    Sun12EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                -12., False)
    # find Rise event
    Rise12List = [e for e in Sun12EventList if e["event"] == "Rise"]
    for e in Rise12List:
        Rise12Time = e["dt"]
        TwilightList.append ((Rise12Time, "Nautical Twilight Begin"))

    # find Set event - may have 2 Set events, e.g 2022-06-09 Gockhausen
    Set12List = [e for e in Sun12EventList if e["event"] == "Set"]
    for e in Set12List:
        Set12Time = e["dt"]
        TwilightList.append ((Set12Time, "Nautical Twilight End"))


    # astronomical twilight, 18 degrees below horizon
    Sun18EventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                -18., False)
    # find Rise event
    Rise18List = [e for e in Sun18EventList if e["event"] == "Rise"]
    for e in Rise18List:
        Rise18Time = e["dt"]
        TwilightList.append ((Rise18Time, "Astronomical Twilight Begin"))

    # find Set event - may have 2 Set events, e.g 2022-06-09 Gockhausen
    Set18List = [e for e in Sun18EventList if e["event"] == "Set"]
    for e in Set18List:
        Set18Time = e["dt"]
        TwilightList.append ((Set18Time, "Astronomical Twilight End"))

    # find Sun alt at midnight 
    #dtMidnight = datetime(dt.year, dt.month, dt.day)
    (y, m, d) = dt.getDate()
    d = floor(d)
    dtMidnight = hvfTime(y, m, d, 0, 0, 0)
 
    (dummy, altMidnight, MidnightDict) = getAziAltfromCivicTimeLongLat \
            (dtMidnight, Zone, getSunData, longitude, latitude)
    dtMidnight1 = dtMidnight + timedelta(hours=24.)
    (dummy, altMidnight1, MidnightDict1) = getAziAltfromCivicTimeLongLat \
            (dtMidnight1, Zone, getSunData, longitude, latitude)

    Alt = MidnightDict["Alt"]
    Alt1 = MidnightDict1["Alt"]
    if altMidnight < -18.:
        TwilightList.append ((dtMidnight, "Deep Night"))
    elif altMidnight < -12.:
        TwilightList.append ((dtMidnight, "Astronomical Twilight"))
    elif altMidnight < -6.:
        TwilightList.append ((dtMidnight, "Nautical Twilight"))
    elif altMidnight < Alt:
        TwilightList.append ((dtMidnight, "Civil Twilight"))

    if altMidnight1 < -18.:
        TwilightList.append ((dtMidnight1, "Deep Night"))
    elif altMidnight1 < -12.:
        TwilightList.append ((dtMidnight1, "Astronomical Twilight"))
    elif altMidnight1 < -6.:
        TwilightList.append ((dtMidnight1, "Nautical Twilight"))
    elif altMidnight1 < Alt:
        TwilightList.append ((dtMidnight1, "Civil Twilight"))

    # print TwilightList
    #print ("Twilight List = ", TwilightList)
    SortedTwilightList = sorted(TwilightList, key = itemgetter(0))
    #print ("Sorted Twilight List = ", SortedTwilightList)

    # get length of each twilight phase
    TwilightLength = []
    for i in range(1, len(SortedTwilightList)):
        e1 = SortedTwilightList[i-1]
        e = SortedTwilightList[i]
        tt = (e[0] - e1[0])
        nicet = nicetimedelta(tt)
        TwilightLength.append(nicet)
    #print ("TwilightLength = ", TwilightLength)

    # print Twilight
    SunSeen = False
    for e in SortedTwilightList:
        try:
            twiLen = TwilightLength.pop(0)
        except:
            twiLen = ''
        if not SunSeen and "Sun" in e[1]: 
            print ()        # blank line before Sun Rise
            SunSeen = True
        elif SunSeen and "Sun" in e[1]: 
            # keep twiLen
            TwilightLength.insert(0, twiLen)
        if SunSeen and "Sun" not in e[1]: 
            print ()        # blank line after Sun Set
            SunSeen = False
            if RiseTime and SetTime:
                daylength = SetTime - RiseTime
                print ("Day Length (hours)          = ", 
                    nicetimedelta(daylength))
                print ()
        if len(e) > 2:      # Sun gets also Azimut: Sun Rise, Sun Set
            print ("%-27s = %s   Azimut = %3d" % (e[1], nicedatetime(e[0]),
                round(e[2], 0)))
        else:               # twilight
            print ("%-27s = %s   Dauer %s" % (e[1], nicedatetime(e[0]), twiLen))

    # Midnight Sun
    if SunMinAlt >  Alt:
        print ("Sun never sets, Midnight Sun")
    if SunMaxAlt <= Alt:
        print ()
        print ("Sun never rises above horizon, Polar Night")
    print ()
############## SUN EVENTS, ORDERED BY TIME OF DAY ######################
    SunEventList = getObjectTimes(dt, longitude, latitude, Zone, getSunData, 
                None, False)
    print ("%-4s Events:" % SunEventList[0]["Object"])
    for e in SunEventList:
        print ("%-4s %-20s %s Azi %3d Alt %7.3f Di %.0f" % (e["Object"], 
                e["event"], nicedatetime(e['dt']), round(e["azi"], 0), 
                e["alt"], e["Dict"]["R"]))

    print ()

############## MOON DATA ####################
    print ("Moon Position")
    (MoonRA, MoonDecl, MoonDict) = getMoonData(dt, Zone, False)
    MoonLong = MoonDict["MoonLong"]
    MoonLat  = MoonDict["MoonLat"]
    MoonDist = MoonDict["MoonDist"]
    MoonHP   = MoonDict["MoonHP"]
    MoonAlt  = MoonDict["Alt"]

    print ("Moon Longitude    (degrees) = %8.3f" % MoonLong)
    print ("Moon Latitude     (degrees) = %8.3f" % MoonLat)
    print ("Moon RA           (hours)   = ", hmsstring(MoonRA/15.))
    print ("Moon RA           (degrees) = %8.3f" % MoonRA)
    print ("Moon Declination  (degrees) = %8.3f" % MoonDecl)

    # hour angle
    MoonHour = localST - timedelta(hours=MoonRA/15.)
    # use hvfTime()
    (h, m, s) = MoonHour.getTime()
    MoonHourdecimal = h + m/60. + s/3600.
    MoonHourdecimalstring = "%02d:%02d:%02d" % (h, m, s)
    print ("Moon Hour Angle   (hours)   = ", MoonHourdecimalstring)
    #MoonHourdecimal = MoonHour.time().hour + MoonHour.time().minute/60. \
    #     + MoonHour.time().second/3600 + MoonHour.time().microsecond/3600e6
    MoonHourgrad = MoonHourdecimal*15.
    print ("Moon Hour Angle   (degrees) = %8.3f" % MoonHourgrad)
    (azi, alt, xyzDict) = getAziAltfromHourDecl(MoonHourgrad, MoonDecl,latitude)
    print ("Azimut            (degrees) = %6.1f" % azi)
    print ("Altitude          (degrees) = %6.1f (no refraction)" % alt)
    
    suffix = ""
    if MoonDist <= 360000. : suffix = " ++"
    if MoonDist >= 405000. : suffix = " --"
    print ("Distance Moon-Earth    (km) =  %.0f" % MoonDist + suffix)
    print ("MoonHP            (degrees) = %8.3f" %  MoonHP)

    #Alt = 0.7275 * MoonHP - 34./60.     # Meeus p. 98
    print ("Altitude Moonrise, Moonset  = %8.3f" %  MoonAlt)
    print()

    # Mondphasen, Sternbilder etc.
    MoonAge = Mod(MoonLong - SunDict["SunLong"], 360.)
    MoonPhase = 0.5 * (1. - Cos(MoonAge))   # between 0 and 1
    Phases = ("Neumond", "Zunehmende Sichel", "Erstes Viertel",\
             "Zunehmender Mond", "Vollmond", "Abnehmender Mond",\
             "Letztes Viertel", "Abnehmende Sichel", "Neumond")
    mainPhase = 360./29.53
    p = Mod(MoonAge, 90.)
    if p<mainPhase or p>(90.-mainPhase):
        p = 2 * round(MoonAge/90.)
    else:
        p = 2 * floor(MoonAge/90.) + 1
    ShowMoonPhase = Phases[int(p)]
    print ("MoonAge           (degrees) = %8.3f" % MoonAge  )
    print ("MoonPhase (between 0 and 1) = %8.3f" % MoonPhase)
    print ("Show MoonPhase              = ", ShowMoonPhase)
    MoonSign = getSign(MoonLong)
    print ("MoonSign                    = ", MoonSign)
    # Distance (angle) between Sun and Moon
    D = AngularDistance(getSunData(dt, Zone,False), getMoonData(dt, Zone,False))
    print ("Angular Distance Sun-Moon   = %8.3f" % D)

    print ()

    # Moon Parallax - Meeus chapter 39
    # (39.2), assume Earth radius is 1
    deltaRA = atan2(-Cos(latitude)*Sin(MoonHP)*Sin(MoonHourgrad), 
                Cos(MoonDecl)-Cos(latitude)*Sin(MoonHP)*Cos(MoonHourgrad))
    if verbose: print ("deltaRA (rad) = ", deltaRA)
    deltaRA = degrees(deltaRA)
    if verbose: print ("deltaRA (deg) = ", deltaRA)
    # (39.3), assume Earth radius is 1
    MDprime = atan2((Sin(MoonDecl)-Sin(latitude)*Sin(MoonHP))*Cos(deltaRA),
                 Cos(MoonDecl)-Cos(latitude)*Sin(MoonHP)*Cos(MoonHourgrad))
    if verbose: print ("MDprime (rad) = ", MDprime)
    MDprime = degrees(MDprime)
    if verbose: print ("MDprime (deg) = ", MDprime)
    RAprime = MoonRA + deltaRA
    if verbose: print ("RAprime (deg) = ", RAprime)
    print ("Moon RA parallax  (degrees) = %8.3f" % RAprime)
    print ("Moon Decl parallx (degrees) = %8.3f" % MDprime)
    print ()
############## MOON EVENTS, ORDERED BY TIME OF DAY ######################
    MoonEventList = getObjectTimes(dt, longitude, latitude, Zone, getMoonData, 
                        None,False)
    print ("%-4s Events:" % MoonEventList[0]["Object"])
    for e in MoonEventList:
        print ("%-4s %-20s %s Azi %3d Alt %7.3f Di %.0f %2s" % (e["Object"], 
            e["event"], nicedatetime(e['dt']), round(e["azi"], 0), e["alt"],
            e["Dict"]["MoonDist"], e["Dict"]["suffix"]))

    print ()

#   sys.exit(0)
#######################################################################
    useUT = True
    PhasesEventList, EDList, SETList = getMoonPhasesxxx(dt, Zone)
    DecliEventList = getMoonDeclixxx(dt, Zone)
    PeriEventList  = getMoonPeriApoxxx(dt, Zone)
    NodesEventList = getMoonNodesxxx(dt, Zone)
    SynopticEventList = chain(SETList, DecliEventList, PeriEventList, 
                            NodesEventList)

    SortedSynopticEventList = sorted(SynopticEventList,key=itemgetter('dt'))

    print ("Synopsis of Moon Events within %2d days:" % 21, \
   "       Distance   Declination  Latitude")
    print (
   "                                               km         degrees      degrees")
    isBefore = True
    for i in range(len(SortedSynopticEventList)):
        Event = SortedSynopticEventList[i]
        # use only events within +- 21 days of dt
        dtCT = Event['dtCivicTime']
        dtdiff = dt - dtCT
        if abs(dtdiff) > 21.: continue
        if isBefore and dtdiff <= 0.:
            print ()
            isBefore = False
        if Event['Object'] == 'Sun':
            print (" %-11s %17s %-36s %s" % (
                    "Sun", 
                    nicedatetime(Event['dtCivicTime']), Event['event'], ''))
        else:
            print (" %-11s %17s %-13s %6d %2s %8.4f    %8.4f" % (
            Event['Object'], 
            nicedatetime(Event['dtCivicTime']), 
                    Event['mode'], Event['dist'], Event['suffix'],
                    Event['decl'], Event['Dict']['MoonLat']))
    print()
Version = "V3.1 hvf 2023-11-30 SunPlanetsTail.py with Planets Data, Events, Magnitudes"
# use hvfTime - hvf 09.11.2022
VersionList.append(Version)

from math import log10

#######################################################################
# Planet Data

# print data for a list of planets
def printPlanetList(PlanetList, dtCivicTime, zone, verbose=False):
    # Dictionary with the functions to get planet data
    PlanetGet  = {  
                "Venus"     : getVenusData,
                "Mars"      : getMarsData,
                "Jupiter"   : getJupiterData,
                "Saturn"    : getSaturnData,
                "Mercury"   : getMercuryData,
                "Earth"     : getEarthData,
                "Uranus"    : getUranusData,
                "Neptune"   : getNeptuneData,
                "Sun"       : getSunData,
                "Moon"      : getMoonData,
                }

    # apparent brightness: Vzero is visual magnitude at 1 au
    # Practical Astronomy with your Calculator or Spreadsheet,
    # Peter Duffet-Smith and Jonathan Zwart, page 123 and 140
    # Values agree with Astronomical Almanach 1984 (see below)
    # the table is not used, but we keep it for documentation purposes
    PlanetVzero = { 
                "Mercury"   : -0.42,
                "Venus"     : -4.40,
                "Mars"      : -1.52,
                "Jupiter"   : -9.40,
                "Saturn"    : -8.88,    # without rings
                "Uranus"    : -7.19,
                "Neptune"   : -6.87,
                }

    LongList = []; LatList = []; XList = []; YList = []; ZList = []
    eXList = []; eYList = []; eZList = [];
    eLongList = []; eLatList = [];
    RAList = []; DeclList = []; AziList = []; AltList = []
    TransDateList = []; TransTimeList = []; TransAltList = []
    RiseDateList = []; RiseList = []; RiseAziList = []
    SetDateList = []; SetList = []; SetAziList = []
    SignList = []; RauList = []; eRauList = []

    RADeclList = []     # used later for distance between planets
    RADeclDict = {}     # used later for distance between planets

    for p in PlanetList:
        getdata = PlanetGet[p]
        RA, Decl, D = getdata(dt, Zone, False)
        X = D["X"]; XList.append(X)
        Y = D["Y"]; YList.append(Y)
        Z = D["Z"]; ZList.append(Z)
        # XYZ relative to Earth
        eX = D["eX"]; eXList.append(eX)
        eY = D["eY"]; eYList.append(eY)
        eZ = D["eZ"]; eZList.append(eZ)
        Rau = D["Rau"]; RauList.append(Rau)
        eRau = D["eRau"]; eRauList.append(eRau)
        LongList.append(D["Long"])
        LatList.append(D["Lat"])
        eLongList.append(D["eLong"])
        eLatList.append(D["eLat"])
        RAList.append(RA)
        DeclList.append(Decl)
        # here we need the geocentric data for rise, transit, set
        '''
        RiseTime, TransitTime, SetTime, RiseAzimut, SetAzimut, transAlt, \
            NoSet, NoRise, flag = \
            getRTS(dt, longitude, latitude, Zone, getdata, None, False)
        '''
        # get azi, alt from hour angle, declination
        RAh = RA/15. # convert to hours
        Hour = localST - timedelta(hours=RAh)
        # use hvfTime()
        (h, m, s) = Hour.getTime()
        Hourdecimal = h + m/60. + s/3600.
        Hourdecimalstring = "%02d:%02d:%02d" % (h, m, s)
   
        #Hourdecimal = Hour.time().hour + Hour.time().minute/60. \
        #            + Hour.time().second/3600 + Hour.time().microsecond/3600e6
        Hourgrad = Hourdecimal*15.
        (azi, alt, xyzDict) = getAziAltfromHourDecl(Hourgrad, Decl, latitude)
        Sign = getSign(D["eLong"])
        AziList.append(round(azi, 0))
        AltList.append(alt)
        SignList.append(Sign)
        RADeclList.append([p, RA, Decl, Rau, eRau])
        RADeclDict[p] = [RA, Decl, Rau, eRau]
    

    printZeile ("Planet      ", PlanetList, "%10s")
    print      ("Heliocentric Coordinates")
    printZeile ("Longitude   ", LongList, "%8.3f")
    printZeile ("Latitude    ", LatList, "%8.3f")
    printZeile ("X (au)      ", XList, "%13.8f")
    printZeile ("Y (au)      ", YList, "%13.8f")
    printZeile ("Z (au)      ", ZList, "%13.8f")
    printZeile ("Radius (au) ", RauList, "%12.8f")
    print ()
    print      ("Geocentric Coordinates")
    printZeile ("eLongitude  ", eLongList, "%8.3f")
    printZeile ("eLatitude   ", eLatList, "%8.3f")
    printZeile ("eX (au)     ", eXList, "%13.8f")
    printZeile ("eY (au)     ", eYList, "%13.8f")
    printZeile ("eZ (au)     ", eZList, "%13.8f")
    printZeile ("eRadius (au)", eRauList, "%12.8f")
    printZeile ("RA (hours)  ", [hmsstring(RA/15.) for RA in RAList], "%10s")
    printZeile ("RA (degrees)", RAList, "%8.3f")
    printZeile ("Declination ", DeclList, "%8.3f")
    printZeile ("Azimut      ", AziList, "%8.0f")
    printZeile ("Altitude    ", AltList, "%8.1f")
    printZeile ("Sign        ", SignList, "%10s")
    print ()
    return RADeclDict
    
# calculate angular distance between two objects with (RA1, Decl1), (RA2, Decl2)
# all angles in degrees
# see Meeus, chapter 16, p. 105, Angular Separation
def AngularDistance(star1, star2, verbose=False):
    RA1 = star1[0]; Decl1 = star1[1]
    RA2 = star2[0]; Decl2 = star2[1]
    cosD = Sin(Decl1)*Sin(Decl2) + Cos(Decl1)*Cos(Decl2)*Cos(RA1-RA2)
    D = degrees(acos(cosD))
    # check if angular separation is small
    if D < 0.2:     # below 0.2 degrees, 12'
        Dmean = (Decl1+Decl2)/2.
        Ddelta = Decl1 - Decl2
        D = sqrt(((RA1-RA2)*Cos(Dmean))**2 + Ddelta**2)
    return D
    
# print data for first group of planets
PlanetList1 = ["Venus", "Mars", "Jupiter", "Saturn"]
RADeclDict1 = printPlanetList(PlanetList1, dt, Zone, False)

# print data for second group of planets
PlanetList2 = ["Sun", "Mercury", "Uranus", "Neptune"]
RADeclDict2 = printPlanetList(PlanetList2, dt, Zone, False)
# update 
RADeclDict1.update(RADeclDict2)

# print angular distances between Sun and all planets
#
# List of planets, horizontally
PList1 = ["Mercury", "Venus", "Mars", "Jupiter", "Saturn", "Uranus", "Neptune"]
# List of Sun and planets, vertically
PList2 = ["Sun"]
for p in PList1[::-1]: PList2.append(p)
print ("Angular Distances (degrees) between Sun and Planets")
# title row
title = " "*8
for p in PList1:
    title += "%10s" % p
print (title)

# calculate and print each line
for j in range(len(PList2)-1):
    p1 = PList2[j]
    str = "%-8s" % p1
    Results = []
    star1 = RADeclDict1[p1]
    for i in range(len(PList1)-j):
        p2 = PList1[i]
        star2 = RADeclDict1[p2]
        D = AngularDistance(star1, star2, False)
        Results.append(D)
        str += "%10.5f" % D
    print (str)

# calculate angular distance to Moon
MoonRA, MoonDecl, MoonD = getMoonData(dt, Zone, False)
str = "Moon    "
star1 = (MoonRA, MoonDecl)
Results = []
for i in range(len(PList1)):
    p2 = PList1[i]
    star2 = RADeclDict1[p2]
    D = AngularDistance(star1, star2, False)
    Results.append(D)
    str += "%10.5f" % D
print (str)

# calculate angular distance to Moon with parallax
#MoonRA, MoonDecl, MoonD = getMoonData(dt, Zone, False)
str = "Moon par"
star1 = (RAprime, MDprime)
Results = []
for i in range(len(PList1)):
    p2 = PList1[i]
    star2 = RADeclDict1[p2]
    D = AngularDistance(star1, star2, False)
    Results.append(D)
    str += "%10.5f" % D
print (str)
print ("Note: Moon par includes the local parallax of the observer when calculating\nangular distances by Moon to planets. This allows for a better assessment \nof planet occultations by Moon. \nSee http://lunar-occultations.com/iota/planets/planets.htm")
######################### Calculate apparent brightness (Magnitude)

#   getMagnitude
#   Parameters:
#   p   String of Planet Name
#   RADeclDict  dict with RA, Declination, radius from Sun, radius from Earth
#   Magnitude is not a strong value, its definition and calculation varies.
#   We use 
def getMagnitude (p, RADeclDict):

    RA, Decl, Rau, eRau = RADeclDict[p] 
    RASun, DeclSun, SunRau, eSunRau = RADeclDict["Sun"]
    # eRau  planet-earth distance, delta
    # Rau   planet-Sun, r
    # SunRau    earth-Sun, R

    # Meeus (1991) p 272, 274
    R = SunRau
    r = Rau
    delta = eRau

    # Phase Angle i , Meeus (1991) 32.12 p. 221, Cosinussatz ebene Trigo
    cosi = (r**2 + delta**2 - R**2)/(2 * r * delta)
    i = degrees(acos(cosi))

    # dito p. 272, Phase
    k = (1. + cosi)/2.

    # print ("k, R, r, delta, i = ", p, "%.3f"%k, "%.3f"%R, "%.3f"%r, "%.3f"%delta, "%.3f"%i)

    """
    # Meeus p. 274
    if p == "Mercury":
        Mag = 1.16 + 5*log10(r*delta) + 0.02838*(i-50) + 0.0001023*(i-50)**2
    elif p == "Venus":
        Mag = -4.00 + 5*log10(r*delta) + 0.01322*i + 0.0000004247*i**3
    elif p == "Mars":
        Mag = -1.30 + 5*log10(r*delta) + 0.01486*i
    elif p == "Jupiter":
        Mag = -8.93 + 5*log10(r*delta)
    elif p == "Saturn":
        Mag = -8.68 + 5*log10(r*delta)  # ohne Ring, siehe p.274
    elif p == "Uranus":
        Mag = -6.85 + 5*log10(r*delta)
    elif p == "Neptune":
        Mag = -7.05 + 5*log10(r*delta)
    """

    # Meeus p. 275, using Astronomical Almanach since 1984
    # James L. Hilton, Improving the Visual Magnitudes of the Planets, in 
    # The Astronomical Almanac. I. Mercury and Venus, 
    # the Astronomical Journal, 129:2902-2906, 2005, June, 
    # https://sci-hub.ru/10.1086/430212

    # imin, imax are the minimal and maximal phase angles valid for 
    # Venus and Mercury

    if p == "Mercury":                      # eq. (2)
        #Mag = -0.42 + 5*log10(r*delta) + 3.80*(i/100) - 2.73*(i/100)**2 +2*(i/100)**3
        # here eq (3), p. 2904
        imin = 2.1; imax = 169.5
        if i >= imin and i <= imax: 
            Mag = -0.60 + 5*log10(r*delta) + 4.98*(i/100) \
                    - 4.88*(i/100)**2 +3.02*(i/100)**3
        else: Mag = None
    elif p == "Venus":                      # eq. (4)
        #Mag = -4.40 + 5*log10(r*delta) + 0.0009*i + 0.000239*i**2 -0.65e-6*i**3
        # here eq (5), p. 2906
        imin = 2.2; imax = 170.2; imid = 163.6
        if i >= imin and i <= imid: 
            Mag = -4.47 + 5*log10(r*delta) + 1.03*(i/100) \
                    + 0.57*(i/100)**2 +0.13*(i/100)**3
        elif i > imid and i <= imax: 
            Mag =  0.98 + 5*log10(r*delta) - 1.02*(i/100) 
        else: Mag = None
    elif p == "Mars":
        Mag = -1.52 + 5*log10(r*delta) + 0.016*i
    elif p == "Jupiter":
        Mag = -9.40 + 5*log10(r*delta) + 0.005*i
    elif p == "Saturn":
        Mag = -8.88 + 5*log10(r*delta)  # ohne Ring, siehe p.274
    elif p == "Uranus":
        Mag = -7.19 + 5*log10(r*delta)
    elif p == "Neptune":
        Mag = -6.87 + 5*log10(r*delta)

    return (k, Mag)

str1 = "Phase   "
str2 = "Magnitude"

print ()
print ("Phases (between 0 and 1) and Apparent Brightness (Magnitude) of Planets")
print ("Note: Magnitude of Saturn neglects its Ring")
# title row
title = " "*8
for p in PList1:
    title += "%10s" % p
    F, Mag = getMagnitude(p, RADeclDict1)
    str1 += "%10.3f"%F
    if Mag:
        str2 += "%9.2f "%Mag
    else:
        str2 += "   invisib"
print (title)
print (str1)
print (str2)


############### PLANET EVENTS, ORDERED BY TIME OF DAY #################
print ()
for getData in [getVenusData, getMarsData, getJupiterData, getSaturnData, 
                getMercuryData, getUranusData, getNeptuneData, ]:
    EventList = getObjectTimes(dt, longitude, latitude, Zone, getData,
                        None,False)
    print ("%-7s Events:" % EventList[0]["Object"])
    for e in EventList:
        print ("%-7s %-20s %s Azi %3d Alt %7.3f" % (e["Object"],
            e["event"], nicedatetime(e['dt']), round(e["azi"], 0), e["alt"],
            ))

    print ()
sys.exit()
